import{aJ as e}from"./@ui-core-DKdciH1Z.js";import{K as t,y as r,b as l}from"./index-D7KLAbvI.js";function i({err:a,isAlertCircle:s=!1}){return e.jsx(t,{icon:s?e.jsx(r,{className:"h-[120px] w-[120px] fill-neutral-300"}):e.jsx(l,{className:"h-[120px] w-[120px] fill-neutral-300"}),children:e.jsx("div",{className:"text-center",children:e.jsx("p",{className:"max-w-full whitespace-pre-wrap break-words text-text-lg text-theme-text-secondary",children:a.message})})})}export{i as E};
