import{aM as l,aJ as s}from"./@ui-core-DKdciH1Z.js";import{bQ as c,e as a,A as n,h as x,a6 as d}from"./index-D7KLAbvI.js";import{s as m}from"./index-C1wOIU40.js";import{a as o,O as p}from"./@react-router-DaIuhn3X.js";import"./@reactflow-CJzATMj6.js";import"./@zod-DKyD3xVW.js";import"./@monaco-editor-CSMpaClf.js";import"./stack-detail-query-IhXZFDdF.js";function u(){const{stackId:i}=o(),e=l({...m.stackDetail(i)});if(e.isPending)return s.jsx(c,{children:s.jsxs("div",{className:"flex items-center gap-3",children:[s.jsx(a,{className:"size-7 shrink-0 rounded-md"}),s.jsxs("div",{className:"space-y-0.5",children:[s.jsx(a,{className:"h-4 w-[100px]"}),s.jsx(a,{className:"h-6 w-[150px]"})]})]})});if(e.isError)return null;const{name:t,id:r}=e.data;return s.jsx(c,{children:s.jsxs("div",{className:"group/copybutton flex items-center gap-3",children:[s.jsx(n,{type:"square",size:"lg",className:"shrink-0",children:s.jsx(x,{size:"lg",children:t[0]??"S"})}),s.jsxs("div",{className:"space-y-0.5 overflow-hidden",children:[s.jsxs("div",{className:"flex items-center gap-1",children:[s.jsx("p",{className:"truncate text-text-sm text-theme-text-secondary",children:r}),s.jsx(d,{copyText:r})]}),s.jsx("h1",{className:"truncate text-display-xs font-semibold",children:s.jsx("span",{className:"truncate",children:t})})]})]})})}function b(){return s.jsxs("div",{className:"flex h-full flex-col",children:[s.jsx(u,{}),s.jsx(p,{})]})}export{b as default};
