"""Release [0.83.1].

Revision ID: 0.83.1
Revises: 3d7e39f3ac92
Create Date: 2025-06-23 07:26:40.361406

"""

# revision identifiers, used by Alembic.
revision = "0.83.1"
down_revision = "3d7e39f3ac92"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    pass


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    pass
