"""Release [0.42.0].

Revision ID: 0.42.0
Revises: 0e4735b23577
Create Date: 2023-07-20 09:46:06.914807

"""

# revision identifiers, used by Alembic.
revision = "0.42.0"
down_revision = "0e4735b23577"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    pass


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    pass
