pre[class*=language-],code[class*=language-]{color:#24292f;text-shadow:none;direction:ltr;text-align:left;white-space:pre;word-spacing:normal;word-break:normal;line-height:1.5;-moz-tab-size:4;-o-tab-size:4;tab-size:4;-webkit-hyphens:none;hyphens:none}pre[class*=language-]::-moz-selection,code[class*=language-]::-moz-selection{text-shadow:none;background:#9fc6e9}pre[class*=language-]::selection,code[class*=language-]::selection,pre[class*=language-]::mozselection,code[class*=language-]::mozselection{text-shadow:none;background:#9fc6e9}@media print{pre[class*=language-],code[class*=language-]{text-shadow:none}}:not(pre)>code[class*=language-]{padding:.1em .3em;border-radius:.3em;color:#24292f;background:#eff1f3}pre[data-line]{position:relative}pre[class*=language-]>code[class*=language-]{position:relative;z-index:1}.line-highlight{position:absolute;left:0;right:0;padding:inherit 0;margin-top:1em;background:#fff8c5;box-shadow:inset 5px 0 #eed888;z-index:0;pointer-events:none;line-height:inherit;white-space:pre}.namespace{opacity:.7}.token.comment,.token.prolog,.token.doctype,.token.cdata{color:#6e7781}.token.punctuation{color:#24292f}.token.property,.token.tag,.token.boolean,.token.number,.token.constant,.token.symbol,.token.deleted{color:#0550ae}.token.selector,.token.attr-name,.token.string,.token.char,.token.builtin,.token.inserted{color:#0a3069}.token.operator,.token.entity,.token.url,.language-css .token.string,.style .token.string{color:#0550ae}.token.atrule,.token.attr-value,.token.keyword{color:#cf222e}.token.function{color:#8250df}.token.regex,.token.important,.token.variable{color:#0a3069}.token.important,.token.bold{font-weight:700}.token.italic{font-style:italic}.token.entity{cursor:help}
