"""Release [0.93.1].

Revision ID: 0.93.1
Revises: 40a700c16141
Create Date: 2026-01-13 13:18:35.733568

"""

# revision identifiers, used by Alembic.
revision = "0.93.1"
down_revision = "40a700c16141"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    pass


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    pass
