"""Release [0.61.0].

Revision ID: 0.61.0
Revises: 0d707865f404
Create Date: 2024-07-08 15:52:54.765307

"""

# revision identifiers, used by Alembic.
revision = "0.61.0"
down_revision = "0d707865f404"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    pass


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    pass
