"""Release [0.56.3].

Revision ID: 0.56.3
Revises: 1d74e596abb8
Create Date: 2024-04-09 14:16:22.179577

"""

# revision identifiers, used by Alembic.
revision = "0.56.3"
down_revision = "1d74e596abb8"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    pass


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    pass
