import{aJ as r}from"./@ui-core-DKdciH1Z.js";import{k as a,c as o}from"./index-D7KLAbvI.js";function m({children:e}){return r.jsx(a,{className:"w-full",children:e})}function i({children:e,className:s,...t}){return r.jsx("div",{className:o("border-b border-theme-border-moderate px-5 py-3 text-display-xs font-semibold",s),...t,children:e})}function p({children:e,className:s,...t}){return r.jsx("div",{...t,className:o("p-5",s),children:e})}function x({children:e}){return r.jsx("div",{className:"flex items-center justify-end gap-2 border-t border-theme-border-moderate p-5",children:e})}export{p as B,x as F,i as H,m as W};
