"""Release [0.85.0].

Revision ID: 0.85.0
Revises: 502b4fa5fa88
Create Date: 2025-09-11 14:36:30.276055

"""

# revision identifiers, used by Alembic.
revision = "0.85.0"
down_revision = "502b4fa5fa88"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    pass


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    pass
