import{r as i,aJ as e}from"./@ui-core-DKdciH1Z.js";import{R as a}from"./ButtonGroup-BirxIr-W.js";import{a9 as n}from"./index-D7KLAbvI.js";import{u as m,g as p,R as u,P as c}from"./service-hZJMUIot.js";import{a as l}from"./@react-router-DaIuhn3X.js";import"./helper-DWeonard.js";import"./bulk-delete-C2Plj5jD.js";import"./@reactflow-CJzATMj6.js";import"./@zod-DKyD3xVW.js";import"./@monaco-editor-CSMpaClf.js";import"./action-cell-CsLz-mf4.js";import"./schedule-tag-Cpi_jTxz.js";import"./clock-DcaJYqTc.js";import"./SearchField-DEer4oQ6.js";import"./all-pipeline-runs-query-CQbINLBU.js";import"./Pagination-Bl0X6mOG.js";import"./chevron-right-double-CNOR0jlH.js";function d(r){const{setBreadcrumbs:s}=n();i.useEffect(()=>{},[s,r])}function x(){const{pipelineId:r}=l(),s=m(),o=i.useMemo(()=>p(),[]),t={pipeline_id:r,...s,sort_by:"desc:updated"};return e.jsxs(e.Fragment,{children:[e.jsx(u,{params:t}),e.jsx(c,{columns:o,params:t})]})}function J(){return d(),e.jsx("div",{className:"space-y-5",children:e.jsx(a,{children:e.jsx(x,{})})})}export{J as default};
