"""Pipeline source code [514b9fae1500].

Revision ID: 514b9fae1500
Revises: 2837a7ab3d77
Create Date: 2025-12-09 11:13:39.173696

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "514b9fae1500"
down_revision = "2837a7ab3d77"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("pipeline_snapshot", schema=None) as batch_op:
        batch_op.add_column(sa.Column("source_code", sa.TEXT(), nullable=True))

    # ### end Alembic commands ###


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("pipeline_snapshot", schema=None) as batch_op:
        batch_op.drop_column("source_code")

    # ### end Alembic commands ###
