import{aJ as s}from"./@ui-core-DKdciH1Z.js";import{T as r,ay as t,ag as o,r as i}from"./index-D7KLAbvI.js";import{L as l}from"./@react-router-DaIuhn3X.js";function c({runId:a,status:e}){return s.jsx(l,{to:i.projects.runs.detail(a),children:s.jsxs(r,{emphasis:"subtle",rounded:!1,className:"inline-flex items-center gap-0.5",color:o(e),children:[s.jsx(t,{className:"h-3 w-3 fill-current"}),a.split("-")[0]]})})}export{c as R};
