import{r as a,aJ as e}from"./@ui-core-DKdciH1Z.js";import{c as n,a as m,b as p,d as i,e as l,u,D as c,P as y,f as D}from"./column-definitions-BIORRFlg.js";import{a as d}from"./@react-router-DaIuhn3X.js";import"./Pagination-Bl0X6mOG.js";import"./index-D7KLAbvI.js";import"./@reactflow-CJzATMj6.js";import"./@zod-DKyD3xVW.js";import"./@monaco-editor-CSMpaClf.js";import"./chevron-right-double-CNOR0jlH.js";import"./helper-DWeonard.js";import"./SearchField-DEer4oQ6.js";import"./button-group-basic-1Ecitaf4.js";import"./deployment-status-tag-9Q9AOMJF.js";import"./pipeline-link-DyLNQTVa.js";function x(){return a.useMemo(()=>[n(),m(),p(),i(),l()],[])}function P(){const{pipelineId:o}=d(),r=u(),s=x(),t={...r,pipeline:o,sort_by:"desc:updated"};return e.jsxs(e.Fragment,{children:[e.jsx(c,{params:t}),e.jsx(y,{params:t,columns:s})]})}function J(){return e.jsx("div",{className:"space-y-5",children:e.jsx(D,{children:e.jsx(P,{})})})}export{J as default};
