"""Release [0.56.2].

Revision ID: 0.56.2
Revises: 0701da9951a0
Create Date: 2024-03-25 14:49:49.021147

"""

# revision identifiers, used by Alembic.
revision = "0.56.2"
down_revision = "0701da9951a0"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    pass


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    pass
