import{aJ as e}from"./@ui-core-DKdciH1Z.js";import{P as r,a as n,b as o,c as i,d as l,e as c}from"./ProCta-Lhqw0IMw.js";import{C as m}from"./CodeSnippet-Dwuwo3ZI.js";import{k as d}from"./index-D7KLAbvI.js";import{f}from"./@react-router-DaIuhn3X.js";import"./check-tR7EFZ1z.js";import"./@reactflow-CJzATMj6.js";import"./@zod-DKyD3xVW.js";import"./@monaco-editor-CSMpaClf.js";const p="/assets/acp-DOsXjFc7.webp";function u(){const[a]=f(),t=a.get("artifact");function s(){return t?`zenml artifact list --name='contains:${t}'`:"zenml artifact list"}return e.jsxs(d,{className:"flex flex-wrap justify-between p-5",children:[e.jsxs("div",{children:[e.jsx("p",{className:"text-text-xl font-semibold",children:"Staying Open Source? "}),e.jsx("p",{className:"text-theme-text-secondary",children:"No problem! Use this CLI command to list your artifacts"})]}),e.jsx(m,{code:s()})]})}function M(){return e.jsxs("div",{className:"layout-container space-y-5 py-5",children:[e.jsxs(r,{className:"relative overflow-y-hidden",children:[e.jsxs("div",{className:"w-full max-w-none space-y-5 lg:max-w-[900px]",children:[e.jsx(n,{children:"Advanced Artifact Management Features with ZenML Pro"}),e.jsx(o,{}),e.jsx(i,{features:[{title:"Artifact Control Plane Dashboard",subtitle:"Artifact management and monitoring"},{title:"Enterprise Security",subtitle:"Social SSO, RBAC, and User Management"},{title:"Managed ZenML Server",subtitle:"On your VPC or hosted on our infrastructure"},{title:"Advanced MLOps",subtitle:"CI/CD/CT, Model Control Plane and more"}]}),e.jsx(l,{})]}),e.jsx(c,{className:"absolute translate-x-[30%] translate-y-[15%] scale-[40%]",src:p,alt:"Screenshot of the ZenML Pro Artifact Control Plane"})]}),e.jsx(u,{})]})}export{M as default};
