import{aJ as s}from"./@ui-core-DKdciH1Z.js";import{a1 as e,a2 as n,a3 as l,a5 as a}from"./index-D7KLAbvI.js";function h({connectorType:o}){const r=o.logo_url,i=o.name;return s.jsx(e,{children:s.jsxs(n,{children:[s.jsx(l,{children:s.jsxs("div",{className:"size-5 shrink-0",children:[s.jsx("img",{width:24,height:24,className:"size-5 shrink-0",src:r??"",alt:`Logo of ${i}`}),s.jsx("p",{className:"sr-only",children:i})]})}),s.jsx(a,{children:i})]})})}function d({resourceType:o}){const r=o.logo_url,i=o.name;return s.jsx(e,{children:s.jsxs(n,{children:[s.jsx(l,{children:s.jsxs("div",{className:"size-5 shrink-0",children:[s.jsx("img",{width:24,height:24,className:"size-5 shrink-0",src:r??"",alt:`Logo of ${i}`}),s.jsx("p",{className:"sr-only",children:i})]})}),s.jsx(a,{children:i})]})})}export{h as C,d as R};
