import{aJ as r}from"./@ui-core-DKdciH1Z.js";import{R as o}from"./RunsBody-DCI-S_ke.js";import{R as t}from"./ButtonGroup-BirxIr-W.js";import"./index-D7KLAbvI.js";import"./@react-router-DaIuhn3X.js";import"./@reactflow-CJzATMj6.js";import"./@zod-DKyD3xVW.js";import"./@monaco-editor-CSMpaClf.js";import"./Pagination-Bl0X6mOG.js";import"./chevron-right-double-CNOR0jlH.js";import"./SearchField-DEer4oQ6.js";import"./all-pipeline-runs-query-CQbINLBU.js";import"./pipeline-link-DyLNQTVa.js";import"./action-cell-CsLz-mf4.js";import"./schedule-tag-Cpi_jTxz.js";import"./clock-DcaJYqTc.js";import"./helper-DWeonard.js";import"./bulk-delete-C2Plj5jD.js";function v(){return r.jsx(t,{children:r.jsx(o,{})})}export{v as default};
