import{r as t,aJ as s}from"./@ui-core-DKdciH1Z.js";import{C as i}from"./form-DzivnZia.js";import{a9 as o,bi as n,bj as p,r as a,aS as m,ab as u}from"./index-D7KLAbvI.js";import{a as c}from"./@react-router-DaIuhn3X.js";import"./@zod-DKyD3xVW.js";import"./name-Bg7MBXU_.js";import"./all-pipeline-runs-query-CQbINLBU.js";import"./update-snapshot-D62dGIXu.js";import"./@reactflow-CJzATMj6.js";import"./@monaco-editor-CSMpaClf.js";function d(e){const{setBreadcrumbs:r}=o();t.useEffect(()=>{e&&r([...e.resources?.pipeline?[n,{label:e.resources.pipeline.name||"",href:a.projects.pipelines.detail.runs(e.resources.pipeline.id)}]:[p],{label:s.jsx(m,{name:e.name,index:e.body?.index}),href:a.projects.runs.detail(e.id)},{label:"Create Snapshot",href:a.projects.runs.createSnapshot(e.id)}])},[r,e])}function R(){const{runId:e}=c(),r=u({runId:e});return d(r.data),r.isPending?s.jsx("p",{children:"Loading..."}):r.isError?s.jsx("p",{children:"Error"}):s.jsx("div",{className:"mx-auto w-full max-w-4xl p-5",children:s.jsx(i,{run:r.data})})}export{R as default};
