import{r as e,aJ as t}from"./@ui-core-DKdciH1Z.js";import{C as s}from"./form-DzivnZia.js";import{a9 as a,bk as o,r as m}from"./index-D7KLAbvI.js";import"./@zod-DKyD3xVW.js";import"./@react-router-DaIuhn3X.js";import"./name-Bg7MBXU_.js";import"./all-pipeline-runs-query-CQbINLBU.js";import"./update-snapshot-D62dGIXu.js";import"./@reactflow-CJzATMj6.js";import"./@monaco-editor-CSMpaClf.js";function p(){const{setBreadcrumbs:r}=a();e.useEffect(()=>{r([o,{label:"Create",href:m.projects.snapshots.create}])},[r])}function C(){return p(),t.jsx("div",{className:"mx-auto w-full max-w-4xl p-5",children:t.jsx(s,{run:void 0})})}export{C as default};
