import{r as s,aJ as t}from"./@ui-core-DKdciH1Z.js";import{c as e,a as r,b as n,i as p,d as m,e as i,f as l,g as u,u as c,S as h,P as S,h as d}from"./column-definitions-CRMIPc2j.js";import"./index-D7KLAbvI.js";import"./@react-router-DaIuhn3X.js";import"./@reactflow-CJzATMj6.js";import"./@zod-DKyD3xVW.js";import"./@monaco-editor-CSMpaClf.js";import"./SearchField-DEer4oQ6.js";import"./button-group-basic-1Ecitaf4.js";import"./Pagination-Bl0X6mOG.js";import"./chevron-right-double-CNOR0jlH.js";import"./deployment-status-tag-9Q9AOMJF.js";import"./pipeline-link-DyLNQTVa.js";import"./run-status-tag-DzXqsiXY.js";import"./action-cell-CsLz-mf4.js";import"./helper-DWeonard.js";import"./update-snapshot-D62dGIXu.js";import"./bulk-delete-C2Plj5jD.js";function C(){return s.useMemo(()=>[e(),r(),n(),p(),m(),i(),l(),u()],[])}function x(){const o=C(),a={named_only:!0,...c(),sort_by:"desc:updated"};return t.jsxs(t.Fragment,{children:[t.jsx(h,{displayCreateButton:!0,params:a}),t.jsx(S,{params:a,columns:o})]})}function J(){return t.jsx("div",{className:"space-y-5 py-5",children:t.jsx(d,{children:t.jsx(x,{})})})}export{J as default};
