"""Release [0.91.2].

Revision ID: 0.91.2
Revises: 4dd9d3afd2c0
Create Date: 2025-11-19 00:54:27.184728

"""

# revision identifiers, used by Alembic.
revision = "0.91.2"
down_revision = "4dd9d3afd2c0"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    pass


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    pass
