"""Release [0.80.1].

Revision ID: 0.80.1
Revises: 1f1105204aaf
Create Date: 2025-03-27 14:31:15.476115

"""

# revision identifiers, used by Alembic.
revision = "0.80.1"
down_revision = "1f1105204aaf"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    pass


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    pass
