import{aJ as o}from"./@ui-core-DKdciH1Z.js";import{S as r}from"./Tabs-CQ_r9Hd8.js";import{R as m}from"./RunsBody-DCI-S_ke.js";import{R as i}from"./ButtonGroup-BirxIr-W.js";import{S as p}from"./StackList-BnoJDftj.js";import{a as s}from"./@react-router-DaIuhn3X.js";import"./index-D7KLAbvI.js";import"./@reactflow-CJzATMj6.js";import"./@zod-DKyD3xVW.js";import"./@monaco-editor-CSMpaClf.js";import"./scrolling-tabs-list-BGYSkrRR.js";import"./transform-Bq7p9csj.js";import"./KeyValue-DF0Uxg2V.js";import"./NestedCollapsible-k41ntPyA.js";import"./eye-off-T0xdenj_.js";import"./eye-0EFEzbgA.js";import"./CollapsibleCard-3zV5aMxC.js";import"./CodeSnippet-Dwuwo3ZI.js";import"./Pagination-Bl0X6mOG.js";import"./chevron-right-double-CNOR0jlH.js";import"./SearchField-DEer4oQ6.js";import"./all-pipeline-runs-query-CQbINLBU.js";import"./pipeline-link-DyLNQTVa.js";import"./action-cell-CsLz-mf4.js";import"./schedule-tag-Cpi_jTxz.js";import"./clock-DcaJYqTc.js";import"./helper-DWeonard.js";import"./bulk-delete-C2Plj5jD.js";import"./index-C1wOIU40.js";import"./stack-detail-query-IhXZFDdF.js";import"./resizable-sheet-D7Mw_0Os.js";import"./react-resizable-panels.browser-DzY2Iih7.js";import"./components-Br2ezRib.js";import"./NumberBox-rSe9DK5q.js";import"./ComponentBadge-BKlKYVYF.js";import"./ComponentIcon-BOKrWHiD.js";import"./bar-chart-square-check-Lt3fLSJ-.js";import"./layout-BRN9AeDt.js";function M(){const{componentId:t}=s();return o.jsx("div",{className:"@container",children:o.jsx(r,{isPanel:!1,stacksTabContent:o.jsx(p,{fixedQueryParams:{component_id:t}}),runsTabContent:o.jsx(i,{children:o.jsx(m,{fixedQueryParams:{stack_component:t}})}),componentId:t})})}export{M as default};
