"""Release [0.63.0].

Revision ID: 0.63.0
Revises: 7d1919bb1ef0
Create Date: 2024-07-29 16:59:59.891939

"""

# revision identifiers, used by Alembic.
revision = "0.63.0"
down_revision = "7d1919bb1ef0"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    pass


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    pass
