import{aJ as r}from"./@ui-core-DKdciH1Z.js";import{R as m}from"./ButtonGroup-BirxIr-W.js";import{u as i,R as a,P as n,g as p}from"./service-hZJMUIot.js";import{a as l}from"./@react-router-DaIuhn3X.js";import"./index-D7KLAbvI.js";import"./@reactflow-CJzATMj6.js";import"./@zod-DKyD3xVW.js";import"./@monaco-editor-CSMpaClf.js";import"./helper-DWeonard.js";import"./bulk-delete-C2Plj5jD.js";import"./action-cell-CsLz-mf4.js";import"./schedule-tag-Cpi_jTxz.js";import"./clock-DcaJYqTc.js";import"./SearchField-DEer4oQ6.js";import"./all-pipeline-runs-query-CQbINLBU.js";import"./Pagination-Bl0X6mOG.js";import"./chevron-right-double-CNOR0jlH.js";function u(){const o=i(),e=p(),{deploymentId:s}=l(),t={triggered_by_deployment_id:s,...o};return r.jsxs(r.Fragment,{children:[r.jsx(a,{params:t}),r.jsx(n,{columns:e,params:t})]})}function F(){return r.jsx("div",{className:"space-y-5 p-5 lg:px-[80px]",children:r.jsx(m,{children:r.jsx(u,{})})})}export{F as default};
