"""Release [0.91.0].

Revision ID: 0.91.0
Revises: 24552f3be1f2
Create Date: 2025-10-24 09:00:18.541933

"""

# revision identifiers, used by Alembic.
revision = "0.91.0"
down_revision = "24552f3be1f2"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    pass


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    pass
