import{r as o,aJ as t}from"./@ui-core-DKdciH1Z.js";import{c as n,a as i,b as p,d as m,e as u,f as c,g as l,u as h,S as d,P as S,h as x}from"./column-definitions-CRMIPc2j.js";import{a as C}from"./@react-router-DaIuhn3X.js";import{a9 as f}from"./index-D7KLAbvI.js";import"./SearchField-DEer4oQ6.js";import"./button-group-basic-1Ecitaf4.js";import"./Pagination-Bl0X6mOG.js";import"./chevron-right-double-CNOR0jlH.js";import"./@zod-DKyD3xVW.js";import"./deployment-status-tag-9Q9AOMJF.js";import"./pipeline-link-DyLNQTVa.js";import"./run-status-tag-DzXqsiXY.js";import"./action-cell-CsLz-mf4.js";import"./helper-DWeonard.js";import"./update-snapshot-D62dGIXu.js";import"./bulk-delete-C2Plj5jD.js";import"./@reactflow-CJzATMj6.js";import"./@monaco-editor-CSMpaClf.js";function P(){return o.useMemo(()=>[n(),i(),p(),m(),u(),c(),l()],[])}function b(){const{pipelineId:e}=C(),s=h(),r=P(),a={pipeline:e,named_only:!0,...s,sort_by:"desc:updated"};return t.jsxs(t.Fragment,{children:[t.jsx(d,{params:a}),t.jsx(S,{params:a,columns:r})]})}function j(e){const{setBreadcrumbs:s}=f();o.useEffect(()=>{},[s,e])}function q(){return j(),t.jsx("div",{className:"space-y-5",children:t.jsx(x,{children:t.jsx(b,{})})})}export{q as default};
