import{aJ as e}from"./@ui-core-DKdciH1Z.js";import{k as t,B as i,bM as n,P as a}from"./index-D7KLAbvI.js";import{C as c}from"./CodeSnippet-Dwuwo3ZI.js";import{H as d}from"./Helpbox-BvyemBlY.js";import{I as o}from"./Infobox-DUoxxn6u.js";import"./@react-router-DaIuhn3X.js";import"./@reactflow-CJzATMj6.js";import"./@zod-DKyD3xVW.js";import"./@monaco-editor-CSMpaClf.js";import"./alert-triangle-DWngjzfY.js";function l(s){return e.jsxs("div",{className:"space-y-1",children:[e.jsx("p",{className:"text-text-sm text-theme-text-secondary",children:s.description}),e.jsx(c,{codeClasses:"whitespace-pre-wrap",wrap:!0,code:s.command})]})}const m=[{command:"zenml code-repository register <NAME> --type=<CODE_REPOSITORY_TYPE] [--CODE_REPOSITORY_OPTIONS]",description:"Register a code repository"},{command:"zenml code-repository register <NAME> --type=custom --source=<CODE_REPOSITORY_SOURCE> [--CODE_REPOSITORY_OPTIONS]",description:"Use a custom repository "},{command:"zenml code-repository list",description:"List your registered code repositories"},{command:"zenml code-repository delete <REPOSITORY_NAME_OR_ID>",description:"Delete a code repository that you have previously registered"}];function p(){return e.jsx(o,{children:e.jsxs("div",{className:"flex w-full flex-wrap items-center gap-x-2 gap-y-0.5 text-text-md",children:[e.jsx("p",{className:"font-semibold",children:"We are creating a new Repositories experience"}),e.jsx("p",{children:"In the meanwhile you can use the CLI to add and manage your repos."})]})})}function x(){return e.jsx(t,{className:"flex flex-col-reverse items-stretch overflow-hidden md:flex-row",children:e.jsxs("div",{className:"w-full p-7 md:w-2/3",children:[e.jsx("h2",{className:"text-display-xs font-semibold",children:"Learn More about Repositories"}),e.jsx("p",{className:"mt-2 text-text-lg text-theme-text-secondary",children:"Get started with ZenML Repositories for streamlined pipeline versioning and faster Docker builds."}),e.jsx(i,{asChild:!0,className:"mt-5 flex w-fit items-center gap-1 px-2 py-1",size:"md",children:e.jsxs("a",{href:"https://docs.zenml.io/concepts/code-repositories",target:"_blank",rel:"noopener noreferrer",children:[e.jsx(n,{className:"size-4 shrink-0 fill-white"}),"Read the Documentation"]})})]})})}function h(){return e.jsxs("section",{className:"space-y-5 pl-8 pr-5",children:[e.jsx(o,{className:"text-text-md",intent:"neutral",children:"Code repositories enable ZenML to keep track of the code version that you use for your pipeline runs. Additionally, running a pipeline which is tracked in a registered code repository can decrease the time it takes Docker to build images for containerized stack components."}),m.map((s,r)=>e.jsx(e.Fragment,{children:l(s)},r)),e.jsx(d,{text:"Check all the commands and find more about Repositories in our Docs",link:"https://docs.zenml.io/how-to/setting-up-a-project-repository/connect-your-git-repository"})]})}function E(){return e.jsxs(t,{className:"space-y-4 p-5",children:[e.jsx("h1",{className:"text-text-xl font-semibold",children:"Repositories"}),e.jsx(p,{}),e.jsx(x,{}),e.jsxs("div",{className:"flex items-center gap-2",children:[e.jsx(a,{}),"Administering your Code Repositories"]}),e.jsx(h,{})]})}export{E as default};
