import{F as a,ao as r,G as n,H as o,J as u}from"./index-D7KLAbvI.js";import{aM as i}from"./@ui-core-DKdciH1Z.js";function c({stackId:t}){return["stacks",t]}async function f({stackId:t}){const s=o(u.stacks.detail(t)),e=await a(s,{method:"GET",headers:{"Content-Type":"application/json"}});if(e.status===404&&r(),!e.ok)throw new n({message:`Error while fetching stack ${t}`,status:e.status,statusText:e.statusText});return e.json()}function y(t,s){return i({queryKey:c(t),queryFn:()=>f(t),...s})}export{f,y as u};
