import{aJ as e}from"./@ui-core-DKdciH1Z.js";import{P as r,a as t,b as s,c as a,d as i,e as o}from"./ProCta-Lhqw0IMw.js";import"./index-D7KLAbvI.js";import"./@react-router-DaIuhn3X.js";import"./@reactflow-CJzATMj6.js";import"./@zod-DKyD3xVW.js";import"./@monaco-editor-CSMpaClf.js";import"./check-tR7EFZ1z.js";const n="/assets/triggers-DVX4pEGY.avif";function f(){return e.jsx("div",{children:e.jsx("div",{className:"layout-container space-y-5 py-5",children:e.jsxs(r,{className:"relative overflow-y-hidden",children:[e.jsxs("div",{className:"w-full max-w-none space-y-5 lg:max-w-[900px]",children:[e.jsx(t,{children:"Advanced Native Scheduling, Webhooks, and Triggers"}),e.jsx(s,{}),e.jsx(a,{features:[{title:"Automate recurring training and evaluation jobs",subtitle:"Schedule pipelines to run at specific times or intervals"},{title:"Enterprise Security",subtitle:"Social SSO, RBAC, and User Management"},{title:"Managed ZenML Server",subtitle:"On your VPC or hosted on our infrastructure"},{title:"Advanced MLOps",subtitle:"CI/CD/CT, Artifact Control Plane and more"}]}),e.jsx(i,{})]}),e.jsx(o,{className:"absolute translate-x-[30%] translate-y-[15%] scale-[40%]",src:n,alt:"Screenshot of the ZenML Pro Triggers"})]})})})}export{f as default};
