"""Release [0.93.3].

Revision ID: 0.93.3
Revises: 8e36b8227160
Create Date: 2026-02-18 13:41:36.511218

"""

# revision identifiers, used by Alembic.
revision = "0.93.3"
down_revision = "8e36b8227160"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    pass


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    pass
