import{aJ as r}from"./@ui-core-DKdciH1Z.js";import{c1 as a}from"./index-D7KLAbvI.js";const n=a("flex h-7 w-7 items-center justify-center rounded-sm text-text-lg font-semibold",{variants:{intent:{default:"bg-primary-100 text-theme-text-brand",disabled:"bg-neutral-100 text-theme-text-tertiary"}},defaultVariants:{intent:"default"}});function m({children:t,intent:e}){return r.jsx("div",{className:n({intent:e}),children:t})}export{m as N};
