"""Release [0.94.1].

Revision ID: 0.94.1
Revises: 24771f985eac
Create Date: 2026-03-19 11:15:07.142707

"""

# revision identifiers, used by Alembic.
revision = "0.94.1"
down_revision = "24771f985eac"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    pass


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    pass
