import{r,aJ as e}from"./@ui-core-DKdciH1Z.js";import{c as a,a as s,g as m,b as n,d as l,e as p,u as i,D as u,P as c,f as y}from"./column-definitions-BIORRFlg.js";import"./Pagination-Bl0X6mOG.js";import"./index-D7KLAbvI.js";import"./@react-router-DaIuhn3X.js";import"./@reactflow-CJzATMj6.js";import"./@zod-DKyD3xVW.js";import"./@monaco-editor-CSMpaClf.js";import"./chevron-right-double-CNOR0jlH.js";import"./helper-DWeonard.js";import"./SearchField-DEer4oQ6.js";import"./button-group-basic-1Ecitaf4.js";import"./deployment-status-tag-9Q9AOMJF.js";import"./pipeline-link-DyLNQTVa.js";function D(){return r.useMemo(()=>[a(),s(),m(),n(),l(),p()],[])}function x(){const o=D(),t={...i(),sort_by:"desc:updated"};return e.jsxs(e.Fragment,{children:[e.jsx(u,{params:t}),e.jsx(c,{params:t,columns:o})]})}function F(){return e.jsx("div",{className:"space-y-5 py-5",children:e.jsx(y,{children:e.jsx(x,{})})})}export{F as default};
