"""Release [0.56.4].

Revision ID: 0.56.4
Revises: 2d201872e23c
Create Date: 2024-04-24 08:05:30.664289

"""

# revision identifiers, used by Alembic.
revision = "0.56.4"
down_revision = "2d201872e23c"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    pass


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    pass
