import{r as t,aJ as a}from"./@ui-core-DKdciH1Z.js";import{c as s}from"./index-D7KLAbvI.js";const d=t.forwardRef(({className:e,...r},o)=>a.jsx("label",{ref:o,...r,className:s("flex w-full items-center gap-2 space-x-2 rounded-md border border-theme-border-moderate p-4 font-semibold transition-shadow duration-150 hover:cursor-pointer hover:shadow-sm data-[state=selected]:border-theme-surface-strong",e)}));d.displayName="RadioItemLabel";const m=t.forwardRef(({className:e,...r},o)=>a.jsx("input",{type:"radio",ref:o,...r,className:s("h-3 w-3 border !border-theme-border-bold text-theme-surface-primary focus:ring-2 focus:ring-theme-surface-strong",e)}));m.displayName="RadioItem";function c(e){return new Promise(r=>setTimeout(r,e))}export{d as R,m as a,c as s};
