"""Release [0.93.2].

Revision ID: 0.93.2
Revises: d1e8c2c4a9ef
Create Date: 2026-01-29 10:14:59.243786

"""

# revision identifiers, used by Alembic.
revision = "0.93.2"
down_revision = "d1e8c2c4a9ef"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    pass


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    pass
