"""Release [0.93.0].

Revision ID: 0.93.0
Revises: 514b9fae1500
Create Date: 2025-12-15 11:32:21.261040

"""

# revision identifiers, used by Alembic.
revision = "0.93.0"
down_revision = "514b9fae1500"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    pass


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    pass
