import{aJ as e}from"./@ui-core-DKdciH1Z.js";import{K as s,B as t,r as a,bW as r}from"./index-D7KLAbvI.js";import{L as i}from"./@react-router-DaIuhn3X.js";import"./@reactflow-CJzATMj6.js";import"./@zod-DKyD3xVW.js";import"./@monaco-editor-CSMpaClf.js";function p(){return e.jsx("div",{className:"flex min-h-screen w-full flex-col",children:e.jsx(s,{icon:e.jsx(r,{className:"h-[120px] w-[120px] fill-neutral-300"}),children:e.jsxs("div",{className:"text-center",children:[e.jsx("h1",{className:"mb-2 text-display-xs font-semibold",children:"We can't find the page you are looking for"}),e.jsx("p",{className:"text-lg text-theme-text-secondary",children:"You can try typing a different URL or we can bring you back to your Homepage."}),e.jsx("div",{className:"mt-5 flex justify-center",children:e.jsx(t,{size:"md",asChild:!0,children:e.jsx(i,{className:"w-min self-center whitespace-nowrap",to:a.projects.overview,children:e.jsx("span",{className:"px-0.5",children:"Go to Home"})})})})]})})})}export{p as default};
