import{r as i,aJ as s}from"./@ui-core-DKdciH1Z.js";import{R as n}from"./ButtonGroup-BirxIr-W.js";import{u as m,g as p,R as u,P as l}from"./service-hZJMUIot.js";import{a as c}from"./@react-router-DaIuhn3X.js";import"./index-D7KLAbvI.js";import"./@reactflow-CJzATMj6.js";import"./@zod-DKyD3xVW.js";import"./@monaco-editor-CSMpaClf.js";import"./helper-DWeonard.js";import"./bulk-delete-C2Plj5jD.js";import"./action-cell-CsLz-mf4.js";import"./schedule-tag-Cpi_jTxz.js";import"./clock-DcaJYqTc.js";import"./SearchField-DEer4oQ6.js";import"./all-pipeline-runs-query-CQbINLBU.js";import"./Pagination-Bl0X6mOG.js";import"./chevron-right-double-CNOR0jlH.js";function d(){const t=m(),o=i.useMemo(()=>p().filter(e=>e.id!=="Snapshot"),[]),{snapshotId:a}=c(),r={source_snapshot_id:a,sort_by:"desc:updated",...t};return s.jsxs(s.Fragment,{children:[s.jsx(u,{params:r}),s.jsx(l,{columns:o,params:r})]})}function I(){return s.jsx("div",{className:"space-y-5",children:s.jsx(n,{children:s.jsx(d,{})})})}export{I as default};
