"""Release [0.92.0].

Revision ID: 0.92.0
Revises: d203788f82b9
Create Date: 2025-12-02 01:27:33.886780

"""

# revision identifiers, used by Alembic.
revision = "0.92.0"
down_revision = "d203788f82b9"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    pass


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    pass
