import{aJ as e,r as o}from"./@ui-core-DKdciH1Z.js";import{B as l,r as i,d as n,I as m,c as d,aR as x}from"./index-D7KLAbvI.js";import{c as p}from"./persist-C7PD3b0n.js";import{c as u}from"./@react-router-DaIuhn3X.js";import{a as f}from"./@zod-DKyD3xVW.js";function v(){const r=u();function s(){p(),r(i.stacks.create.index)}return e.jsx(l,{onClick:()=>s(),intent:"secondary",size:"md",children:"Cancel"})}function k(r){switch(r){case"aws":return"AWS";case"azure":return"Azure";case"gcp":return"GCP";default:return"Provider"}}function C(){const{register:r,formState:{errors:s}}=f();return e.jsxs("section",{className:"space-y-5 border-b border-theme-border-moderate pb-5",children:[e.jsxs("div",{className:"space-y-1",children:[e.jsxs("p",{className:"flex items-center gap-1 text-text-lg font-semibold",children:[e.jsx(n,{className:"h-5 w-5 fill-primary-400"}),"Select a name for your Stack"]}),e.jsx("p",{className:"text-theme-text-secondary",children:"Please select a name for your stack, that is not used already."})]}),e.jsx("div",{className:"flex items-center space-x-2",children:e.jsxs("div",{className:"space-y-1",children:[e.jsx(m,{className:`${s.stackName?"border-red-500":""}`,placeholder:"zenml-remote-stack",...r("stackName")}),s.stackName&&e.jsx("p",{className:"text-text-xs text-red-500",children:s.stackName.message?.toString()})]})})]})}const h=o.forwardRef(({children:r,className:s,id:t,...a},c)=>e.jsxs("div",{className:"min-h-[160px] min-w-[160px]",children:[e.jsx("input",{id:t,...a,ref:c,className:d("peer sr-only",s),type:"radio"}),e.jsx("label",{htmlFor:t,className:"flex h-full w-full flex-col items-start justify-center space-y-5 rounded-md border border-theme-border-minimal bg-theme-surface-primary p-5 text-text-lg text-theme-text-secondary hover:cursor-pointer	peer-checked:border-primary-100 peer-checked:bg-primary-25 peer-focus-visible:border-primary-100 peer-disabled:cursor-default peer-disabled:bg-neutral-50",children:r})]}));h.displayName="CloudProviderRadioButton";function S({icon:r,title:s,subtitle:t,comingSoon:a}){return e.jsxs("div",{className:"space-y-1 text-left",children:[r,e.jsxs("div",{className:"flex items-center gap-1",children:[e.jsx("p",{className:"text-text-lg font-semibold text-theme-text-primary",children:s}),a&&e.jsx(x,{className:"font-semibold",color:"purple",size:"sm",children:"Coming Soon"})]}),e.jsx("p",{className:"text-text-sm text-theme-text-secondary",children:t})]})}export{h as C,S as P,C as S,v as a,k as g};
