import{aJ as t}from"./@ui-core-DKdciH1Z.js";import{C as o,S as r}from"./StackComponentList-B1Sb0tkR.js";import"./index-D7KLAbvI.js";import"./@react-router-DaIuhn3X.js";import"./@reactflow-CJzATMj6.js";import"./@zod-DKyD3xVW.js";import"./@monaco-editor-CSMpaClf.js";import"./Pagination-Bl0X6mOG.js";import"./chevron-right-double-CNOR0jlH.js";import"./SearchField-DEer4oQ6.js";import"./expand-full-aWqoyakM.js";import"./resizable-sheet-D7Mw_0Os.js";import"./react-resizable-panels.browser-DzY2Iih7.js";import"./Tabs-CQ_r9Hd8.js";import"./scrolling-tabs-list-BGYSkrRR.js";import"./transform-Bq7p9csj.js";import"./KeyValue-DF0Uxg2V.js";import"./NestedCollapsible-k41ntPyA.js";import"./eye-off-T0xdenj_.js";import"./eye-0EFEzbgA.js";import"./CollapsibleCard-3zV5aMxC.js";import"./CodeSnippet-Dwuwo3ZI.js";import"./all-pipeline-runs-query-CQbINLBU.js";import"./index-C1wOIU40.js";import"./stack-detail-query-IhXZFDdF.js";import"./ComponentBadge-BKlKYVYF.js";import"./ComponentIcon-BOKrWHiD.js";import"./bar-chart-square-check-Lt3fLSJ-.js";import"./layout-BRN9AeDt.js";import"./action-cell-CsLz-mf4.js";import"./Infobox-DUoxxn6u.js";import"./alert-triangle-DWngjzfY.js";import"./helper-DWeonard.js";import"./bulk-delete-C2Plj5jD.js";function F(){return t.jsx("div",{className:"pt-5",children:t.jsx(o,{children:t.jsx(r,{})})})}export{F as default};
