"""Cache key index [4dd9d3afd2c0].

Revision ID: 4dd9d3afd2c0
Revises: 0.91.1
Create Date: 2025-11-13 11:43:34.498878

"""

from alembic import op

# revision identifiers, used by Alembic.
revision = "4dd9d3afd2c0"
down_revision = "0.91.1"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("step_run", schema=None) as batch_op:
        batch_op.create_index(
            "ix_step_run_cache_key", ["cache_key"], unique=False
        )

    # ### end Alembic commands ###


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("step_run", schema=None) as batch_op:
        batch_op.drop_index("ix_step_run_cache_key")

    # ### end Alembic commands ###
