"""Release [0.94.0].

Revision ID: 0.94.0
Revises: 0.94.0rc0
Create Date: 2026-03-04 14:32:50.581985

"""

# revision identifiers, used by Alembic.
revision = "0.94.0"
down_revision = "0.94.0rc0"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    pass


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    pass
