import{aJ as e}from"./@ui-core-DKdciH1Z.js";import{k as a,bW as t,bX as l}from"./index-D7KLAbvI.js";function o({link:s,text:r="Do you need help?"}){return e.jsxs(a,{className:"flex w-full flex-wrap items-center justify-between p-2",children:[e.jsxs("div",{className:"flex items-center gap-[10px]",children:[e.jsx("div",{className:"flex h-7 w-7 items-center justify-center rounded-sm bg-teal-25",children:e.jsx(t,{className:"h-5 w-5 fill-teal-400"})}),e.jsx("p",{children:r})]}),e.jsx("a",{target:"_blank",rel:"noopener noreferrer",className:l({intent:"primary",emphasis:"subtle",size:"md"}),href:s,children:"Browse our docs"})]})}export{o as H};
