import{aJ as r}from"./@ui-core-DKdciH1Z.js";import{c as t,aj as o,y as m,c1 as l}from"./index-D7KLAbvI.js";import{S as c}from"./alert-triangle-DWngjzfY.js";const d=l("flex items-center text-text-sm rounded-md border px-4 py-3",{variants:{intent:{primary:"border-primary-400 bg-primary-25",warning:"border-warning-300 text-warning-900 bg-warning-50",neutral:"border-theme-border-moderate",error:"bg-error-50 border-error-300"}},defaultVariants:{intent:"primary"}});function w({children:e,className:s,intent:a,button:n,...i}){return r.jsxs("div",{...i,className:t(d({intent:a}),s),children:[r.jsx(x,{intent:a}),r.jsx("div",{className:"w-full min-w-0",children:e}),n&&r.jsx("div",{className:"ml-auto",children:n})]})}function x({intent:e}){switch(e){case"warning":return r.jsx(c,{className:"mr-4 h-5 w-5 shrink-0 fill-warning-700"});case"error":return r.jsx(m,{className:"mr-4 h-5 w-5 shrink-0 fill-error-700"});default:return r.jsx(o,{className:"mr-4 h-5 w-5 shrink-0 fill-theme-text-brand"})}}export{w as I};
