import{aJ as e,r as d}from"./@ui-core-DKdciH1Z.js";import{S as m}from"./chevron-right-double-CNOR0jlH.js";import{bY as l,B as c,c as n,bZ as h}from"./index-D7KLAbvI.js";import{P as b,a as t,b as u}from"./react-resizable-panels.browser-DzY2Iih7.js";function g({className:a,children:s,...r}){return e.jsxs("div",{className:n("flex h-9 items-center gap-1 border-b border-theme-border-moderate bg-theme-surface-primary px-4 py-3",a),...r,children:[e.jsx(l,{asChild:!0,children:e.jsxs(c,{type:"button",className:"flex aspect-square items-center justify-center p-0",intent:"secondary",emphasis:"minimal",children:[e.jsx(m,{className:"h-5 w-5 fill-neutral-500"}),e.jsx("span",{className:"sr-only",children:"Close"})]})}),s]})}const p=d.forwardRef(({children:a,className:s,autoSaveId:r,...o},i)=>e.jsx(h,{...o,ref:i,className:n("h-full w-full !max-w-full !border-0 !bg-transparent p-0 !shadow-none",s),children:e.jsxs(b,{autoSaveId:r,direction:"horizontal",className:"h-full",children:[e.jsx(t,{defaultSize:30,minSize:10,className:"relative bg-transparent",children:e.jsx(l,{className:"absolute inset-0 hover:cursor-default"})}),e.jsx(u,{className:"w-[1px] bg-theme-border-moderate transition-colors duration-200 data-[resize-handle-state=drag]:bg-theme-border-bold data-[resize-handle-state=hover]:bg-theme-border-bold"}),e.jsx(t,{className:"!overflow-y-auto bg-theme-surface-secondary",minSize:25,defaultSize:50,children:a})]})}));p.displayName="ResizableSheetContent";export{p as R,g as S};
