import{aJ as e}from"./@ui-core-DKdciH1Z.js";import{c as s}from"./index-D7KLAbvI.js";function a({children:t,className:n,...r}){return e.jsx("dt",{...r,className:s("col-span-1 flex min-w-0 items-center truncate text-theme-text-secondary",n),children:t})}function c({children:t,className:n,...r}){return e.jsx("dd",{...r,className:s("col-span-2 flex h-6 w-full min-w-0 items-center truncate text-neutral-700",n),children:t})}function m({label:t,value:n}){return e.jsxs(e.Fragment,{children:[e.jsx(a,{children:t}),e.jsx(c,{children:n})]})}export{m as K,c as V,a};
