import{aJ as e}from"./@ui-core-DKdciH1Z.js";import{T as l,$ as p,c as s,r as o}from"./index-D7KLAbvI.js";import{L as m}from"./@react-router-DaIuhn3X.js";function d({pipelineId:i,pipelineName:r,iconClassname:a,className:n,...t}){return e.jsx(m,{to:o.projects.pipelines.detail.runs(i),children:e.jsxs(l,{color:"purple",className:s("inline-flex items-center gap-0.5 text-primary-400",n),rounded:!1,emphasis:"subtle",...t,children:[e.jsx(p,{className:s("size-4 shrink-0 fill-primary-400",a)}),e.jsx("div",{className:"truncate",children:r})]})})}export{d as P};
