import{aJ as s}from"./@ui-core-DKdciH1Z.js";import{T as r,bn as n,bo as i,c as o}from"./index-D7KLAbvI.js";function c({status:e,className:a,...t}){return s.jsxs(r,{size:"xs",emphasis:"subtle",rounded:!1,className:o("inline-flex items-center gap-1 text-theme-text-primary",a),color:"grey",...t,children:[s.jsx("div",{className:`size-1 shrink-0 rounded-rounded ${n(e)}`}),i(e??"")]})}export{c as D};
