"""Cache expiration [124b57b8c7b1].

Revision ID: 124b57b8c7b1
Revises: 7497d2ff5731
Create Date: 2025-10-06 14:35:53.052890

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "124b57b8c7b1"
down_revision = "7497d2ff5731"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("step_run", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column("cache_expires_at", sa.DateTime(), nullable=True)
        )

    # ### end Alembic commands ###


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("step_run", schema=None) as batch_op:
        batch_op.drop_column("cache_expires_at")

    # ### end Alembic commands ###
