"""Release [0.91.1].

Revision ID: 0.91.1
Revises: a5a17015b681
Create Date: 2025-11-11 01:00:37.460524

"""

# revision identifiers, used by Alembic.
revision = "0.91.1"
down_revision = "a5a17015b681"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    pass


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    pass
