import{r as $t,W as ow}from"./@ui-core-DKdciH1Z.js";const Dle="modulepreload",Tle=function(s){return"/"+s},fz={},Se=function(e,t,i){let n=Promise.resolve();if(t&&t.length>0){let l=function(c){return Promise.all(c.map(d=>Promise.resolve(d).then(h=>({status:"fulfilled",value:h}),h=>({status:"rejected",reason:h}))))};document.getElementsByTagName("link");const r=document.querySelector("meta[property=csp-nonce]"),a=r?.nonce||r?.getAttribute("nonce");n=l(t.map(c=>{if(c=Tle(c),c in fz)return;fz[c]=!0;const d=c.endsWith(".css"),h=d?'[rel="stylesheet"]':"";if(document.querySelector(`link[href="${c}"]${h}`))return;const u=document.createElement("link");if(u.rel=d?"stylesheet":Dle,d||(u.as="script"),u.crossOrigin="",u.href=c,a&&u.setAttribute("nonce",a),document.head.appendChild(u),d)return new Promise((g,f)=>{u.addEventListener("load",g),u.addEventListener("error",()=>f(new Error(`Unable to preload CSS for ${c}`)))})}))}function o(r){const a=new Event("vite:preloadError",{cancelable:!0});if(a.payload=r,window.dispatchEvent(a),!a.defaultPrevented)throw r}return n.then(r=>{for(const a of r||[])a.status==="rejected"&&o(a.reason);return e().catch(o)})};function Nle(s,e,t){return e in s?Object.defineProperty(s,e,{value:t,enumerable:!0,configurable:!0,writable:!0}):s[e]=t,s}function pz(s,e){var t=Object.keys(s);if(Object.getOwnPropertySymbols){var i=Object.getOwnPropertySymbols(s);e&&(i=i.filter(function(n){return Object.getOwnPropertyDescriptor(s,n).enumerable})),t.push.apply(t,i)}return t}function mz(s){for(var e=1;e<arguments.length;e++){var t=arguments[e]!=null?arguments[e]:{};e%2?pz(Object(t),!0).forEach(function(i){Nle(s,i,t[i])}):Object.getOwnPropertyDescriptors?Object.defineProperties(s,Object.getOwnPropertyDescriptors(t)):pz(Object(t)).forEach(function(i){Object.defineProperty(s,i,Object.getOwnPropertyDescriptor(t,i))})}return s}function Rle(s,e){if(s==null)return{};var t={},i=Object.keys(s),n,o;for(o=0;o<i.length;o++)n=i[o],!(e.indexOf(n)>=0)&&(t[n]=s[n]);return t}function Ale(s,e){if(s==null)return{};var t=Rle(s,e),i,n;if(Object.getOwnPropertySymbols){var o=Object.getOwnPropertySymbols(s);for(n=0;n<o.length;n++)i=o[n],!(e.indexOf(i)>=0)&&Object.prototype.propertyIsEnumerable.call(s,i)&&(t[i]=s[i])}return t}function Mle(s,e){return Ole(s)||Ple(s,e)||Fle(s,e)||Ble()}function Ole(s){if(Array.isArray(s))return s}function Ple(s,e){if(!(typeof Symbol>"u"||!(Symbol.iterator in Object(s)))){var t=[],i=!0,n=!1,o=void 0;try{for(var r=s[Symbol.iterator](),a;!(i=(a=r.next()).done)&&(t.push(a.value),!(e&&t.length===e));i=!0);}catch(l){n=!0,o=l}finally{try{!i&&r.return!=null&&r.return()}finally{if(n)throw o}}return t}}function Fle(s,e){if(s){if(typeof s=="string")return _z(s,e);var t=Object.prototype.toString.call(s).slice(8,-1);if(t==="Object"&&s.constructor&&(t=s.constructor.name),t==="Map"||t==="Set")return Array.from(s);if(t==="Arguments"||/^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(t))return _z(s,e)}}function _z(s,e){(e==null||e>s.length)&&(e=s.length);for(var t=0,i=new Array(e);t<e;t++)i[t]=s[t];return i}function Ble(){throw new TypeError(`Invalid attempt to destructure non-iterable instance.
In order to be iterable, non-array objects must have a [Symbol.iterator]() method.`)}function Wle(s,e,t){return e in s?Object.defineProperty(s,e,{value:t,enumerable:!0,configurable:!0,writable:!0}):s[e]=t,s}function bz(s,e){var t=Object.keys(s);if(Object.getOwnPropertySymbols){var i=Object.getOwnPropertySymbols(s);e&&(i=i.filter(function(n){return Object.getOwnPropertyDescriptor(s,n).enumerable})),t.push.apply(t,i)}return t}function wz(s){for(var e=1;e<arguments.length;e++){var t=arguments[e]!=null?arguments[e]:{};e%2?bz(Object(t),!0).forEach(function(i){Wle(s,i,t[i])}):Object.getOwnPropertyDescriptors?Object.defineProperties(s,Object.getOwnPropertyDescriptors(t)):bz(Object(t)).forEach(function(i){Object.defineProperty(s,i,Object.getOwnPropertyDescriptor(t,i))})}return s}function Hle(){for(var s=arguments.length,e=new Array(s),t=0;t<s;t++)e[t]=arguments[t];return function(i){return e.reduceRight(function(n,o){return o(n)},i)}}function s1(s){return function e(){for(var t=this,i=arguments.length,n=new Array(i),o=0;o<i;o++)n[o]=arguments[o];return n.length>=s.length?s.apply(this,n):function(){for(var r=arguments.length,a=new Array(r),l=0;l<r;l++)a[l]=arguments[l];return e.apply(t,[].concat(n,a))}}}function lE(s){return{}.toString.call(s).includes("Object")}function Vle(s){return!Object.keys(s).length}function Hy(s){return typeof s=="function"}function zle(s,e){return Object.prototype.hasOwnProperty.call(s,e)}function $le(s,e){return lE(e)||Nu("changeType"),Object.keys(e).some(function(t){return!zle(s,t)})&&Nu("changeField"),e}function Ule(s){Hy(s)||Nu("selectorType")}function qle(s){Hy(s)||lE(s)||Nu("handlerType"),lE(s)&&Object.values(s).some(function(e){return!Hy(e)})&&Nu("handlersType")}function jle(s){s||Nu("initialIsRequired"),lE(s)||Nu("initialType"),Vle(s)&&Nu("initialContent")}function Gle(s,e){throw new Error(s[e]||s.default)}var Kle={initialIsRequired:"initial state is required",initialType:"initial state should be an object",initialContent:"initial state shouldn't be an empty object",handlerType:"handler should be an object or a function",handlersType:"all handlers should be a functions",selectorType:"selector should be a function",changeType:"provided value of changes should be an object",changeField:'it seams you want to change a field in the state which is not specified in the "initial" state',default:"an unknown error accured in `state-local` package"},Nu=s1(Gle)(Kle),Hx={changes:$le,selector:Ule,handler:qle,initial:jle};function Yle(s){var e=arguments.length>1&&arguments[1]!==void 0?arguments[1]:{};Hx.initial(s),Hx.handler(e);var t={current:s},i=s1(Qle)(t,e),n=s1(Xle)(t),o=s1(Hx.changes)(s),r=s1(Zle)(t);function a(){var c=arguments.length>0&&arguments[0]!==void 0?arguments[0]:function(d){return d};return Hx.selector(c),c(t.current)}function l(c){Hle(i,n,o,r)(c)}return[a,l]}function Zle(s,e){return Hy(e)?e(s.current):e}function Xle(s,e){return s.current=wz(wz({},s.current),e),e}function Qle(s,e,t){return Hy(e)?e(s.current):Object.keys(t).forEach(function(i){var n;return(n=e[i])===null||n===void 0?void 0:n.call(e,s.current[i])}),t}var Jle={create:Yle},ece={paths:{vs:"https://cdn.jsdelivr.net/npm/monaco-editor@0.52.2/min/vs"}};function tce(s){return function e(){for(var t=this,i=arguments.length,n=new Array(i),o=0;o<i;o++)n[o]=arguments[o];return n.length>=s.length?s.apply(this,n):function(){for(var r=arguments.length,a=new Array(r),l=0;l<r;l++)a[l]=arguments[l];return e.apply(t,[].concat(n,a))}}}function ice(s){return{}.toString.call(s).includes("Object")}function nce(s){return s||vz("configIsRequired"),ice(s)||vz("configType"),s.urls?(sce(),{paths:{vs:s.urls.monacoBase}}):s}function sce(){console.warn(kX.deprecation)}function oce(s,e){throw new Error(s[e]||s.default)}var kX={configIsRequired:"the configuration object is required",configType:"the configuration object should be an object",default:"an unknown error accured in `@monaco-editor/loader` package",deprecation:`Deprecation warning!
    You are using deprecated way of configuration.

    Instead of using
      monaco.config({ urls: { monacoBase: '...' } })
    use
      monaco.config({ paths: { vs: '...' } })

    For more please check the link https://github.com/suren-atoyan/monaco-loader#config
  `},vz=tce(oce)(kX),rce={config:nce},ace=function(){for(var e=arguments.length,t=new Array(e),i=0;i<e;i++)t[i]=arguments[i];return function(n){return t.reduceRight(function(o,r){return r(o)},n)}};function EX(s,e){return Object.keys(e).forEach(function(t){e[t]instanceof Object&&s[t]&&Object.assign(e[t],EX(s[t],e[t]))}),mz(mz({},s),e)}var lce={type:"cancelation",msg:"operation is manually canceled"};function L2(s){var e=!1,t=new Promise(function(i,n){s.then(function(o){return e?n(lce):i(o)}),s.catch(n)});return t.cancel=function(){return e=!0},t}var cce=Jle.create({config:ece,isInitialized:!1,resolve:null,reject:null,monaco:null}),IX=Mle(cce,2),F0=IX[0],SR=IX[1];function dce(s){var e=rce.config(s),t=e.monaco,i=Ale(e,["monaco"]);SR(function(n){return{config:EX(n.config,i),monaco:t}})}function hce(){var s=F0(function(e){var t=e.monaco,i=e.isInitialized,n=e.resolve;return{monaco:t,isInitialized:i,resolve:n}});if(!s.isInitialized){if(SR({isInitialized:!0}),s.monaco)return s.resolve(s.monaco),L2(k2);if(window.monaco&&window.monaco.editor)return DX(window.monaco),s.resolve(window.monaco),L2(k2);ace(uce,fce)(pce)}return L2(k2)}function uce(s){return document.body.appendChild(s)}function gce(s){var e=document.createElement("script");return s&&(e.src=s),e}function fce(s){var e=F0(function(i){var n=i.config,o=i.reject;return{config:n,reject:o}}),t=gce("".concat(e.config.paths.vs,"/loader.js"));return t.onload=function(){return s()},t.onerror=e.reject,t}function pce(){var s=F0(function(t){var i=t.config,n=t.resolve,o=t.reject;return{config:i,resolve:n,reject:o}}),e=window.require;e.config(s.config),e(["vs/editor/editor.main"],function(t){DX(t),s.resolve(t)},function(t){s.reject(t)})}function DX(s){F0().monaco||SR({monaco:s})}function mce(){return F0(function(s){var e=s.monaco;return e})}var k2=new Promise(function(s,e){return SR({resolve:s,reject:e})}),TX={config:dce,init:hce,__getMonacoInstance:mce},_ce={wrapper:{display:"flex",position:"relative",textAlign:"initial"},fullWidth:{width:"100%"},hide:{display:"none"}},E2=_ce,bce={container:{display:"flex",height:"100%",width:"100%",justifyContent:"center",alignItems:"center"}},wce=bce;function vce({children:s}){return ow.createElement("div",{style:wce.container},s)}var Cce=vce,yce=Cce;function Sce({width:s,height:e,isEditorReady:t,loading:i,_ref:n,className:o,wrapperProps:r}){return ow.createElement("section",{style:{...E2.wrapper,width:s,height:e},...r},!t&&ow.createElement(yce,null,i),ow.createElement("div",{ref:n,style:{...E2.fullWidth,...!t&&E2.hide},className:o}))}var xce=Sce,NX=$t.memo(xce);function Lce(s){$t.useEffect(s,[])}var RX=Lce;function kce(s,e,t=!0){let i=$t.useRef(!0);$t.useEffect(i.current||!t?()=>{i.current=!1}:s,e)}var Hr=kce;function L1(){}function vb(s,e,t,i){return Ece(s,i)||Ice(s,e,t,i)}function Ece(s,e){return s.editor.getModel(AX(s,e))}function Ice(s,e,t,i){return s.editor.createModel(e,t,i?AX(s,i):void 0)}function AX(s,e){return s.Uri.parse(e)}function Dce({original:s,modified:e,language:t,originalLanguage:i,modifiedLanguage:n,originalModelPath:o,modifiedModelPath:r,keepCurrentOriginalModel:a=!1,keepCurrentModifiedModel:l=!1,theme:c="light",loading:d="Loading...",options:h={},height:u="100%",width:g="100%",className:f,wrapperProps:p={},beforeMount:_=L1,onMount:b=L1}){let[w,C]=$t.useState(!1),[v,y]=$t.useState(!0),x=$t.useRef(null),L=$t.useRef(null),E=$t.useRef(null),I=$t.useRef(b),R=$t.useRef(_),O=$t.useRef(!1);RX(()=>{let z=TX.init();return z.then(Q=>(L.current=Q)&&y(!1)).catch(Q=>Q?.type!=="cancelation"&&console.error("Monaco initialization: error:",Q)),()=>x.current?$():z.cancel()}),Hr(()=>{if(x.current&&L.current){let z=x.current.getOriginalEditor(),Q=vb(L.current,s||"",i||t||"text",o||"");Q!==z.getModel()&&z.setModel(Q)}},[o],w),Hr(()=>{if(x.current&&L.current){let z=x.current.getModifiedEditor(),Q=vb(L.current,e||"",n||t||"text",r||"");Q!==z.getModel()&&z.setModel(Q)}},[r],w),Hr(()=>{let z=x.current.getModifiedEditor();z.getOption(L.current.editor.EditorOption.readOnly)?z.setValue(e||""):e!==z.getValue()&&(z.executeEdits("",[{range:z.getModel().getFullModelRange(),text:e||"",forceMoveMarkers:!0}]),z.pushUndoStop())},[e],w),Hr(()=>{x.current?.getModel()?.original.setValue(s||"")},[s],w),Hr(()=>{let{original:z,modified:Q}=x.current.getModel();L.current.editor.setModelLanguage(z,i||t||"text"),L.current.editor.setModelLanguage(Q,n||t||"text")},[t,i,n],w),Hr(()=>{L.current?.editor.setTheme(c)},[c],w),Hr(()=>{x.current?.updateOptions(h)},[h],w);let P=$t.useCallback(()=>{if(!L.current)return;R.current(L.current);let z=vb(L.current,s||"",i||t||"text",o||""),Q=vb(L.current,e||"",n||t||"text",r||"");x.current?.setModel({original:z,modified:Q})},[t,e,n,s,i,o,r]),B=$t.useCallback(()=>{!O.current&&E.current&&(x.current=L.current.editor.createDiffEditor(E.current,{automaticLayout:!0,...h}),P(),L.current?.editor.setTheme(c),C(!0),O.current=!0)},[h,c,P]);$t.useEffect(()=>{w&&I.current(x.current,L.current)},[w]),$t.useEffect(()=>{!v&&!w&&B()},[v,w,B]);function $(){let z=x.current?.getModel();a||z?.original?.dispose(),l||z?.modified?.dispose(),x.current?.dispose()}return ow.createElement(NX,{width:g,height:u,isEditorReady:w,loading:d,_ref:E,className:f,wrapperProps:p})}var Tce=Dce;$t.memo(Tce);function Nce(s){let e=$t.useRef();return $t.useEffect(()=>{e.current=s},[s]),e.current}var Rce=Nce,Vx=new Map;function Ace({defaultValue:s,defaultLanguage:e,defaultPath:t,value:i,language:n,path:o,theme:r="light",line:a,loading:l="Loading...",options:c={},overrideServices:d={},saveViewState:h=!0,keepCurrentModel:u=!1,width:g="100%",height:f="100%",className:p,wrapperProps:_={},beforeMount:b=L1,onMount:w=L1,onChange:C,onValidate:v=L1}){let[y,x]=$t.useState(!1),[L,E]=$t.useState(!0),I=$t.useRef(null),R=$t.useRef(null),O=$t.useRef(null),P=$t.useRef(w),B=$t.useRef(b),$=$t.useRef(),z=$t.useRef(i),Q=Rce(o),J=$t.useRef(!1),ne=$t.useRef(!1);RX(()=>{let Me=TX.init();return Me.then(Ze=>(I.current=Ze)&&E(!1)).catch(Ze=>Ze?.type!=="cancelation"&&console.error("Monaco initialization: error:",Ze)),()=>R.current?ke():Me.cancel()}),Hr(()=>{let Me=vb(I.current,s||i||"",e||n||"",o||t||"");Me!==R.current?.getModel()&&(h&&Vx.set(Q,R.current?.saveViewState()),R.current?.setModel(Me),h&&R.current?.restoreViewState(Vx.get(o)))},[o],y),Hr(()=>{R.current?.updateOptions(c)},[c],y),Hr(()=>{!R.current||i===void 0||(R.current.getOption(I.current.editor.EditorOption.readOnly)?R.current.setValue(i):i!==R.current.getValue()&&(ne.current=!0,R.current.executeEdits("",[{range:R.current.getModel().getFullModelRange(),text:i,forceMoveMarkers:!0}]),R.current.pushUndoStop(),ne.current=!1))},[i],y),Hr(()=>{let Me=R.current?.getModel();Me&&n&&I.current?.editor.setModelLanguage(Me,n)},[n],y),Hr(()=>{a!==void 0&&R.current?.revealLine(a)},[a],y),Hr(()=>{I.current?.editor.setTheme(r)},[r],y);let de=$t.useCallback(()=>{if(!(!O.current||!I.current)&&!J.current){B.current(I.current);let Me=o||t,Ze=vb(I.current,i||s||"",e||n||"",Me||"");R.current=I.current?.editor.create(O.current,{model:Ze,automaticLayout:!0,...c},d),h&&R.current.restoreViewState(Vx.get(Me)),I.current.editor.setTheme(r),a!==void 0&&R.current.revealLine(a),x(!0),J.current=!0}},[s,e,t,i,n,o,c,d,h,r,a]);$t.useEffect(()=>{y&&P.current(R.current,I.current)},[y]),$t.useEffect(()=>{!L&&!y&&de()},[L,y,de]),z.current=i,$t.useEffect(()=>{y&&C&&($.current?.dispose(),$.current=R.current?.onDidChangeModelContent(Me=>{ne.current||C(R.current.getValue(),Me)}))},[y,C]),$t.useEffect(()=>{if(y){let Me=I.current.editor.onDidChangeMarkers(Ze=>{let Xe=R.current.getModel()?.uri;if(Xe&&Ze.find(At=>At.path===Xe.path)){let At=I.current.editor.getModelMarkers({resource:Xe});v?.(At)}});return()=>{Me?.dispose()}}return()=>{}},[y,v]);function ke(){$.current?.dispose(),u?h&&Vx.set(o,R.current.saveViewState()):R.current.getModel()?.dispose(),R.current.dispose()}return ow.createElement(NX,{width:g,height:f,isEditorReady:y,loading:l,_ref:O,className:p,wrapperProps:_})}var Mce=Ace,Oce=$t.memo(Mce),XGe=Oce;function MX(){return globalThis._VSCODE_NLS_MESSAGES}function A6(){return globalThis._VSCODE_NLS_LANGUAGE}const Pce=A6()==="pseudo"||typeof document<"u"&&document.location&&typeof document.location.hash=="string"&&document.location.hash.indexOf("pseudo=true")>=0;function cE(s,e){let t;return e.length===0?t=s:t=s.replace(/\{(\d+)\}/g,(i,n)=>{const o=n[0],r=e[o];let a=i;return typeof r=="string"?a=r:(typeof r=="number"||typeof r=="boolean"||r===void 0||r===null)&&(a=String(r)),a}),Pce&&(t="［"+t.replace(/[aouei]/g,"$&$&")+"］"),t}function m(s,e,...t){return cE(typeof s=="number"?OX(s,e):e,t)}function OX(s,e){const t=MX()?.[s];if(typeof t!="string"){if(typeof e=="string")return e;throw new Error(`!!! NLS MISSING: ${s} !!!`)}return t}function V(s,e,...t){let i;typeof s=="number"?i=OX(s,e):i=e;const n=cE(i,t);return{value:n,original:e===i?n:cE(e,t)}}function Fce(s,e){const t=s;typeof t.vscodeWindowId!="number"&&Object.defineProperty(t,"vscodeWindowId",{get:()=>e})}const Nt=window,BT=class BT{constructor(){this.mapWindowIdToZoomFactor=new Map}getZoomFactor(e){return this.mapWindowIdToZoomFactor.get(this.getWindowId(e))??1}getWindowId(e){return e.vscodeWindowId}};BT.INSTANCE=new BT;let QO=BT;function PX(s,e,t){typeof e=="string"&&(e=s.matchMedia(e)),e.addEventListener("change",t)}function JO(s){return QO.INSTANCE.getZoomFactor(s)}const $v=navigator.userAgent,wo=$v.indexOf("Firefox")>=0,xR=$v.indexOf("AppleWebKit")>=0,B0=$v.indexOf("Chrome")>=0,wm=!B0&&$v.indexOf("Safari")>=0,FX=!B0&&!wm&&xR;$v.indexOf("Electron/")>=0;const Cz=$v.indexOf("Android")>=0;let I2=!1;if(typeof Nt.matchMedia=="function"){const s=Nt.matchMedia("(display-mode: standalone) or (display-mode: window-controls-overlay)"),e=Nt.matchMedia("(display-mode: fullscreen)");I2=s.matches,PX(Nt,s,({matches:t})=>{I2&&e.matches||(I2=t)})}function M6(){return globalThis.MonacoEnvironment}class Bce{constructor(){this.listeners=[],this.unexpectedErrorHandler=function(e){setTimeout(()=>{throw e.stack?ev.isErrorNoTelemetry(e)?new ev(e.message+`

`+e.stack):new Error(e.message+`

`+e.stack):e},0)}}emit(e){this.listeners.forEach(t=>{t(e)})}onUnexpectedError(e){this.unexpectedErrorHandler(e),this.emit(e)}onUnexpectedExternalError(e){this.unexpectedErrorHandler(e)}}const O6=new Bce;function dE(s){O6.onUnexpectedError(s)}function Oe(s){Dr(s)||O6.onUnexpectedError(s)}function ji(s){Dr(s)||O6.onUnexpectedExternalError(s)}function eP(s){if(s instanceof Error){const{name:e,message:t,cause:i}=s,n=s.stacktrace||s.stack;return{$isError:!0,name:e,message:t,stack:n,noTelemetry:ev.isErrorNoTelemetry(s),cause:i?eP(i):void 0,code:s.code}}return s}const hE="Canceled";function Dr(s){return s instanceof ia?!0:s instanceof Error&&s.name===hE&&s.message===hE}class ia extends Error{constructor(){super(hE),this.name=this.message}}function Wce(){const s=new Error(hE);return s.name=s.message,s}function Qr(s){return s?new Error(`Illegal argument: ${s}`):new Error("Illegal argument")}function uE(s){return s?new Error(`Illegal state: ${s}`):new Error("Illegal state")}class Hce extends Error{constructor(e){super("NotSupported"),e&&(this.message=e)}}class ev extends Error{constructor(e){super(e),this.name="CodeExpectedError"}static fromError(e){if(e instanceof ev)return e;const t=new ev;return t.message=e.message,t.stack=e.stack,t}static isErrorNoTelemetry(e){return e.name==="CodeExpectedError"}}class Ce extends Error{constructor(e){super(e||"An unexpected bug occurred."),Object.setPrototypeOf(this,Ce.prototype)}}function D2(s,e){if(!s)throw new Error(e?`Assertion failed (${e})`:"Assertion Failed")}function W0(s,e="Unreachable"){throw new Error(e)}function Vy(s,e="unexpected state"){if(!s)throw typeof e=="string"?new Ce(`Assertion Failed: ${e}`):e}function yz(s,e="Soft Assertion Failed"){s||Oe(new Ce(e))}function rg(s){if(!s()){debugger;s(),Oe(new Ce("Assertion Failed"))}}function H0(s,e){let t=0;for(;t<s.length-1;){const i=s[t],n=s[t+1];if(!e(i,n))return!1;t++}return!0}function bn(s){return typeof s=="string"}function Vce(s,e){return Array.isArray(s)&&s.every(e)}function rn(s){return typeof s=="object"&&s!==null&&!Array.isArray(s)&&!(s instanceof RegExp)&&!(s instanceof Date)}function zce(s){const e=Object.getPrototypeOf(Uint8Array);return typeof s=="object"&&s instanceof e}function ch(s){return typeof s=="number"&&!isNaN(s)}function tP(s){return!!s&&typeof s[Symbol.iterator]=="function"}function iP(s){return s===!0||s===!1}function fs(s){return typeof s>"u"}function Ln(s){return!Ur(s)}function Ur(s){return fs(s)||s===null}function dt(s,e){if(!s)throw new Error(e?`Unexpected type, expected '${e}'`:"Unexpected type")}function gu(s){return Vy(s!=null,"Argument is `undefined` or `null`."),s}function vm(s){return typeof s=="function"}function $ce(s,e){const t=Math.min(s.length,e.length);for(let i=0;i<t;i++)Uce(s[i],e[i])}function Uce(s,e){if(bn(e)){if(typeof s!==e)throw new Error(`argument does not match constraint: typeof ${e}`)}else if(vm(e)){try{if(s instanceof e)return}catch{}if(!Ur(s)&&s.constructor===e||e.length===1&&e.call(void 0,s)===!0)return;throw new Error("argument does not match one of these constraints: arg instanceof constraint, arg.constructor === constraint, nor constraint(arg) === true")}}const Wf="en";let zy=!1,$y=!1,k1=!1,BX=!1,P6=!1,F6=!1,WX=!1,zx,mk=Wf,Sz=Wf,qce,ud;const eh=globalThis;let dr;typeof eh.vscode<"u"&&typeof eh.vscode.process<"u"?dr=eh.vscode.process:typeof process<"u"&&typeof process?.versions?.node=="string"&&(dr=process);const jce=typeof dr?.versions?.electron=="string",Gce=jce&&dr?.type==="renderer";if(typeof dr=="object"){zy=dr.platform==="win32",$y=dr.platform==="darwin",k1=dr.platform==="linux",k1&&dr.env.SNAP&&dr.env.SNAP_REVISION,dr.env.CI||dr.env.BUILD_ARTIFACTSTAGINGDIRECTORY||dr.env.GITHUB_WORKSPACE,zx=Wf,mk=Wf;const s=dr.env.VSCODE_NLS_CONFIG;if(s)try{const e=JSON.parse(s);zx=e.userLocale,Sz=e.osLocale,mk=e.resolvedLanguage||Wf,qce=e.languagePack?.translationsConfigFile}catch{}BX=!0}else typeof navigator=="object"&&!Gce?(ud=navigator.userAgent,zy=ud.indexOf("Windows")>=0,$y=ud.indexOf("Macintosh")>=0,F6=(ud.indexOf("Macintosh")>=0||ud.indexOf("iPad")>=0||ud.indexOf("iPhone")>=0)&&!!navigator.maxTouchPoints&&navigator.maxTouchPoints>0,k1=ud.indexOf("Linux")>=0,WX=ud?.indexOf("Mobi")>=0,P6=!0,mk=A6()||Wf,zx=navigator.language.toLowerCase(),Sz=zx):console.error("Unable to resolve platform.");let _k=0;$y?_k=1:zy?_k=3:k1&&(_k=2);const Bn=zy,Ye=$y,vo=k1,Sl=BX,Vc=P6,Kce=P6&&typeof eh.importScripts=="function",Yce=Kce?eh.origin:void 0,na=F6,HX=WX,T2=_k,Dc=ud,Zce=mk,Xce=typeof eh.postMessage=="function"&&!eh.importScripts,V0=(()=>{if(Xce){const s=[];eh.addEventListener("message",t=>{if(t.data&&t.data.vscodeScheduleAsyncWork)for(let i=0,n=s.length;i<n;i++){const o=s[i];if(o.id===t.data.vscodeScheduleAsyncWork){s.splice(i,1),o.callback();return}}});let e=0;return t=>{const i=++e;s.push({id:i,callback:t}),eh.postMessage({vscodeScheduleAsyncWork:i},"*")}}return s=>setTimeout(s)})(),Fo=$y||F6?2:zy?1:3;let xz=!0,Lz=!1;function VX(){if(!Lz){Lz=!0;const s=new Uint8Array(2);s[0]=1,s[1]=2,xz=new Uint16Array(s.buffer)[0]===513}return xz}const B6=!!(Dc&&Dc.indexOf("Chrome")>=0),Qce=!!(Dc&&Dc.indexOf("Firefox")>=0),Jce=!!(!B6&&Dc&&Dc.indexOf("Safari")>=0),zX=!!(Dc&&Dc.indexOf("Edg/")>=0),ede=!!(Dc&&Dc.indexOf("Android")>=0),z0={clipboard:{writeText:Sl||document.queryCommandSupported&&document.queryCommandSupported("copy")||!!(navigator&&navigator.clipboard&&navigator.clipboard.writeText),readText:Sl||!!(navigator&&navigator.clipboard&&navigator.clipboard.readText)},pointerEvents:Nt.PointerEvent&&("ontouchstart"in Nt||navigator.maxTouchPoints>0)};class W6{constructor(){this._keyCodeToStr=[],this._strToKeyCode=Object.create(null)}define(e,t){this._keyCodeToStr[e]=t,this._strToKeyCode[t.toLowerCase()]=e}keyCodeToStr(e){return this._keyCodeToStr[e]}strToKeyCode(e){return this._strToKeyCode[e.toLowerCase()]||0}}const bk=new W6,nP=new W6,sP=new W6,$X=new Array(230),tde=Object.create(null),ide=Object.create(null),H6=[];for(let s=0;s<=193;s++)H6[s]=-1;(function(){const e=[[1,0,"None",0,"unknown",0,"VK_UNKNOWN","",""],[1,1,"Hyper",0,"",0,"","",""],[1,2,"Super",0,"",0,"","",""],[1,3,"Fn",0,"",0,"","",""],[1,4,"FnLock",0,"",0,"","",""],[1,5,"Suspend",0,"",0,"","",""],[1,6,"Resume",0,"",0,"","",""],[1,7,"Turbo",0,"",0,"","",""],[1,8,"Sleep",0,"",0,"VK_SLEEP","",""],[1,9,"WakeUp",0,"",0,"","",""],[0,10,"KeyA",31,"A",65,"VK_A","",""],[0,11,"KeyB",32,"B",66,"VK_B","",""],[0,12,"KeyC",33,"C",67,"VK_C","",""],[0,13,"KeyD",34,"D",68,"VK_D","",""],[0,14,"KeyE",35,"E",69,"VK_E","",""],[0,15,"KeyF",36,"F",70,"VK_F","",""],[0,16,"KeyG",37,"G",71,"VK_G","",""],[0,17,"KeyH",38,"H",72,"VK_H","",""],[0,18,"KeyI",39,"I",73,"VK_I","",""],[0,19,"KeyJ",40,"J",74,"VK_J","",""],[0,20,"KeyK",41,"K",75,"VK_K","",""],[0,21,"KeyL",42,"L",76,"VK_L","",""],[0,22,"KeyM",43,"M",77,"VK_M","",""],[0,23,"KeyN",44,"N",78,"VK_N","",""],[0,24,"KeyO",45,"O",79,"VK_O","",""],[0,25,"KeyP",46,"P",80,"VK_P","",""],[0,26,"KeyQ",47,"Q",81,"VK_Q","",""],[0,27,"KeyR",48,"R",82,"VK_R","",""],[0,28,"KeyS",49,"S",83,"VK_S","",""],[0,29,"KeyT",50,"T",84,"VK_T","",""],[0,30,"KeyU",51,"U",85,"VK_U","",""],[0,31,"KeyV",52,"V",86,"VK_V","",""],[0,32,"KeyW",53,"W",87,"VK_W","",""],[0,33,"KeyX",54,"X",88,"VK_X","",""],[0,34,"KeyY",55,"Y",89,"VK_Y","",""],[0,35,"KeyZ",56,"Z",90,"VK_Z","",""],[0,36,"Digit1",22,"1",49,"VK_1","",""],[0,37,"Digit2",23,"2",50,"VK_2","",""],[0,38,"Digit3",24,"3",51,"VK_3","",""],[0,39,"Digit4",25,"4",52,"VK_4","",""],[0,40,"Digit5",26,"5",53,"VK_5","",""],[0,41,"Digit6",27,"6",54,"VK_6","",""],[0,42,"Digit7",28,"7",55,"VK_7","",""],[0,43,"Digit8",29,"8",56,"VK_8","",""],[0,44,"Digit9",30,"9",57,"VK_9","",""],[0,45,"Digit0",21,"0",48,"VK_0","",""],[1,46,"Enter",3,"Enter",13,"VK_RETURN","",""],[1,47,"Escape",9,"Escape",27,"VK_ESCAPE","",""],[1,48,"Backspace",1,"Backspace",8,"VK_BACK","",""],[1,49,"Tab",2,"Tab",9,"VK_TAB","",""],[1,50,"Space",10,"Space",32,"VK_SPACE","",""],[0,51,"Minus",88,"-",189,"VK_OEM_MINUS","-","OEM_MINUS"],[0,52,"Equal",86,"=",187,"VK_OEM_PLUS","=","OEM_PLUS"],[0,53,"BracketLeft",92,"[",219,"VK_OEM_4","[","OEM_4"],[0,54,"BracketRight",94,"]",221,"VK_OEM_6","]","OEM_6"],[0,55,"Backslash",93,"\\",220,"VK_OEM_5","\\","OEM_5"],[0,56,"IntlHash",0,"",0,"","",""],[0,57,"Semicolon",85,";",186,"VK_OEM_1",";","OEM_1"],[0,58,"Quote",95,"'",222,"VK_OEM_7","'","OEM_7"],[0,59,"Backquote",91,"`",192,"VK_OEM_3","`","OEM_3"],[0,60,"Comma",87,",",188,"VK_OEM_COMMA",",","OEM_COMMA"],[0,61,"Period",89,".",190,"VK_OEM_PERIOD",".","OEM_PERIOD"],[0,62,"Slash",90,"/",191,"VK_OEM_2","/","OEM_2"],[1,63,"CapsLock",8,"CapsLock",20,"VK_CAPITAL","",""],[1,64,"F1",59,"F1",112,"VK_F1","",""],[1,65,"F2",60,"F2",113,"VK_F2","",""],[1,66,"F3",61,"F3",114,"VK_F3","",""],[1,67,"F4",62,"F4",115,"VK_F4","",""],[1,68,"F5",63,"F5",116,"VK_F5","",""],[1,69,"F6",64,"F6",117,"VK_F6","",""],[1,70,"F7",65,"F7",118,"VK_F7","",""],[1,71,"F8",66,"F8",119,"VK_F8","",""],[1,72,"F9",67,"F9",120,"VK_F9","",""],[1,73,"F10",68,"F10",121,"VK_F10","",""],[1,74,"F11",69,"F11",122,"VK_F11","",""],[1,75,"F12",70,"F12",123,"VK_F12","",""],[1,76,"PrintScreen",0,"",0,"","",""],[1,77,"ScrollLock",84,"ScrollLock",145,"VK_SCROLL","",""],[1,78,"Pause",7,"PauseBreak",19,"VK_PAUSE","",""],[1,79,"Insert",19,"Insert",45,"VK_INSERT","",""],[1,80,"Home",14,"Home",36,"VK_HOME","",""],[1,81,"PageUp",11,"PageUp",33,"VK_PRIOR","",""],[1,82,"Delete",20,"Delete",46,"VK_DELETE","",""],[1,83,"End",13,"End",35,"VK_END","",""],[1,84,"PageDown",12,"PageDown",34,"VK_NEXT","",""],[1,85,"ArrowRight",17,"RightArrow",39,"VK_RIGHT","Right",""],[1,86,"ArrowLeft",15,"LeftArrow",37,"VK_LEFT","Left",""],[1,87,"ArrowDown",18,"DownArrow",40,"VK_DOWN","Down",""],[1,88,"ArrowUp",16,"UpArrow",38,"VK_UP","Up",""],[1,89,"NumLock",83,"NumLock",144,"VK_NUMLOCK","",""],[1,90,"NumpadDivide",113,"NumPad_Divide",111,"VK_DIVIDE","",""],[1,91,"NumpadMultiply",108,"NumPad_Multiply",106,"VK_MULTIPLY","",""],[1,92,"NumpadSubtract",111,"NumPad_Subtract",109,"VK_SUBTRACT","",""],[1,93,"NumpadAdd",109,"NumPad_Add",107,"VK_ADD","",""],[1,94,"NumpadEnter",3,"",0,"","",""],[1,95,"Numpad1",99,"NumPad1",97,"VK_NUMPAD1","",""],[1,96,"Numpad2",100,"NumPad2",98,"VK_NUMPAD2","",""],[1,97,"Numpad3",101,"NumPad3",99,"VK_NUMPAD3","",""],[1,98,"Numpad4",102,"NumPad4",100,"VK_NUMPAD4","",""],[1,99,"Numpad5",103,"NumPad5",101,"VK_NUMPAD5","",""],[1,100,"Numpad6",104,"NumPad6",102,"VK_NUMPAD6","",""],[1,101,"Numpad7",105,"NumPad7",103,"VK_NUMPAD7","",""],[1,102,"Numpad8",106,"NumPad8",104,"VK_NUMPAD8","",""],[1,103,"Numpad9",107,"NumPad9",105,"VK_NUMPAD9","",""],[1,104,"Numpad0",98,"NumPad0",96,"VK_NUMPAD0","",""],[1,105,"NumpadDecimal",112,"NumPad_Decimal",110,"VK_DECIMAL","",""],[0,106,"IntlBackslash",97,"OEM_102",226,"VK_OEM_102","",""],[1,107,"ContextMenu",58,"ContextMenu",93,"","",""],[1,108,"Power",0,"",0,"","",""],[1,109,"NumpadEqual",0,"",0,"","",""],[1,110,"F13",71,"F13",124,"VK_F13","",""],[1,111,"F14",72,"F14",125,"VK_F14","",""],[1,112,"F15",73,"F15",126,"VK_F15","",""],[1,113,"F16",74,"F16",127,"VK_F16","",""],[1,114,"F17",75,"F17",128,"VK_F17","",""],[1,115,"F18",76,"F18",129,"VK_F18","",""],[1,116,"F19",77,"F19",130,"VK_F19","",""],[1,117,"F20",78,"F20",131,"VK_F20","",""],[1,118,"F21",79,"F21",132,"VK_F21","",""],[1,119,"F22",80,"F22",133,"VK_F22","",""],[1,120,"F23",81,"F23",134,"VK_F23","",""],[1,121,"F24",82,"F24",135,"VK_F24","",""],[1,122,"Open",0,"",0,"","",""],[1,123,"Help",0,"",0,"","",""],[1,124,"Select",0,"",0,"","",""],[1,125,"Again",0,"",0,"","",""],[1,126,"Undo",0,"",0,"","",""],[1,127,"Cut",0,"",0,"","",""],[1,128,"Copy",0,"",0,"","",""],[1,129,"Paste",0,"",0,"","",""],[1,130,"Find",0,"",0,"","",""],[1,131,"AudioVolumeMute",117,"AudioVolumeMute",173,"VK_VOLUME_MUTE","",""],[1,132,"AudioVolumeUp",118,"AudioVolumeUp",175,"VK_VOLUME_UP","",""],[1,133,"AudioVolumeDown",119,"AudioVolumeDown",174,"VK_VOLUME_DOWN","",""],[1,134,"NumpadComma",110,"NumPad_Separator",108,"VK_SEPARATOR","",""],[0,135,"IntlRo",115,"ABNT_C1",193,"VK_ABNT_C1","",""],[1,136,"KanaMode",0,"",0,"","",""],[0,137,"IntlYen",0,"",0,"","",""],[1,138,"Convert",0,"",0,"","",""],[1,139,"NonConvert",0,"",0,"","",""],[1,140,"Lang1",0,"",0,"","",""],[1,141,"Lang2",0,"",0,"","",""],[1,142,"Lang3",0,"",0,"","",""],[1,143,"Lang4",0,"",0,"","",""],[1,144,"Lang5",0,"",0,"","",""],[1,145,"Abort",0,"",0,"","",""],[1,146,"Props",0,"",0,"","",""],[1,147,"NumpadParenLeft",0,"",0,"","",""],[1,148,"NumpadParenRight",0,"",0,"","",""],[1,149,"NumpadBackspace",0,"",0,"","",""],[1,150,"NumpadMemoryStore",0,"",0,"","",""],[1,151,"NumpadMemoryRecall",0,"",0,"","",""],[1,152,"NumpadMemoryClear",0,"",0,"","",""],[1,153,"NumpadMemoryAdd",0,"",0,"","",""],[1,154,"NumpadMemorySubtract",0,"",0,"","",""],[1,155,"NumpadClear",131,"Clear",12,"VK_CLEAR","",""],[1,156,"NumpadClearEntry",0,"",0,"","",""],[1,0,"",5,"Ctrl",17,"VK_CONTROL","",""],[1,0,"",4,"Shift",16,"VK_SHIFT","",""],[1,0,"",6,"Alt",18,"VK_MENU","",""],[1,0,"",57,"Meta",91,"VK_COMMAND","",""],[1,157,"ControlLeft",5,"",0,"VK_LCONTROL","",""],[1,158,"ShiftLeft",4,"",0,"VK_LSHIFT","",""],[1,159,"AltLeft",6,"",0,"VK_LMENU","",""],[1,160,"MetaLeft",57,"",0,"VK_LWIN","",""],[1,161,"ControlRight",5,"",0,"VK_RCONTROL","",""],[1,162,"ShiftRight",4,"",0,"VK_RSHIFT","",""],[1,163,"AltRight",6,"",0,"VK_RMENU","",""],[1,164,"MetaRight",57,"",0,"VK_RWIN","",""],[1,165,"BrightnessUp",0,"",0,"","",""],[1,166,"BrightnessDown",0,"",0,"","",""],[1,167,"MediaPlay",0,"",0,"","",""],[1,168,"MediaRecord",0,"",0,"","",""],[1,169,"MediaFastForward",0,"",0,"","",""],[1,170,"MediaRewind",0,"",0,"","",""],[1,171,"MediaTrackNext",124,"MediaTrackNext",176,"VK_MEDIA_NEXT_TRACK","",""],[1,172,"MediaTrackPrevious",125,"MediaTrackPrevious",177,"VK_MEDIA_PREV_TRACK","",""],[1,173,"MediaStop",126,"MediaStop",178,"VK_MEDIA_STOP","",""],[1,174,"Eject",0,"",0,"","",""],[1,175,"MediaPlayPause",127,"MediaPlayPause",179,"VK_MEDIA_PLAY_PAUSE","",""],[1,176,"MediaSelect",128,"LaunchMediaPlayer",181,"VK_MEDIA_LAUNCH_MEDIA_SELECT","",""],[1,177,"LaunchMail",129,"LaunchMail",180,"VK_MEDIA_LAUNCH_MAIL","",""],[1,178,"LaunchApp2",130,"LaunchApp2",183,"VK_MEDIA_LAUNCH_APP2","",""],[1,179,"LaunchApp1",0,"",0,"VK_MEDIA_LAUNCH_APP1","",""],[1,180,"SelectTask",0,"",0,"","",""],[1,181,"LaunchScreenSaver",0,"",0,"","",""],[1,182,"BrowserSearch",120,"BrowserSearch",170,"VK_BROWSER_SEARCH","",""],[1,183,"BrowserHome",121,"BrowserHome",172,"VK_BROWSER_HOME","",""],[1,184,"BrowserBack",122,"BrowserBack",166,"VK_BROWSER_BACK","",""],[1,185,"BrowserForward",123,"BrowserForward",167,"VK_BROWSER_FORWARD","",""],[1,186,"BrowserStop",0,"",0,"VK_BROWSER_STOP","",""],[1,187,"BrowserRefresh",0,"",0,"VK_BROWSER_REFRESH","",""],[1,188,"BrowserFavorites",0,"",0,"VK_BROWSER_FAVORITES","",""],[1,189,"ZoomToggle",0,"",0,"","",""],[1,190,"MailReply",0,"",0,"","",""],[1,191,"MailForward",0,"",0,"","",""],[1,192,"MailSend",0,"",0,"","",""],[1,0,"",114,"KeyInComposition",229,"","",""],[1,0,"",116,"ABNT_C2",194,"VK_ABNT_C2","",""],[1,0,"",96,"OEM_8",223,"VK_OEM_8","",""],[1,0,"",0,"",0,"VK_KANA","",""],[1,0,"",0,"",0,"VK_HANGUL","",""],[1,0,"",0,"",0,"VK_JUNJA","",""],[1,0,"",0,"",0,"VK_FINAL","",""],[1,0,"",0,"",0,"VK_HANJA","",""],[1,0,"",0,"",0,"VK_KANJI","",""],[1,0,"",0,"",0,"VK_CONVERT","",""],[1,0,"",0,"",0,"VK_NONCONVERT","",""],[1,0,"",0,"",0,"VK_ACCEPT","",""],[1,0,"",0,"",0,"VK_MODECHANGE","",""],[1,0,"",0,"",0,"VK_SELECT","",""],[1,0,"",0,"",0,"VK_PRINT","",""],[1,0,"",0,"",0,"VK_EXECUTE","",""],[1,0,"",0,"",0,"VK_SNAPSHOT","",""],[1,0,"",0,"",0,"VK_HELP","",""],[1,0,"",0,"",0,"VK_APPS","",""],[1,0,"",0,"",0,"VK_PROCESSKEY","",""],[1,0,"",0,"",0,"VK_PACKET","",""],[1,0,"",0,"",0,"VK_DBE_SBCSCHAR","",""],[1,0,"",0,"",0,"VK_DBE_DBCSCHAR","",""],[1,0,"",0,"",0,"VK_ATTN","",""],[1,0,"",0,"",0,"VK_CRSEL","",""],[1,0,"",0,"",0,"VK_EXSEL","",""],[1,0,"",0,"",0,"VK_EREOF","",""],[1,0,"",0,"",0,"VK_PLAY","",""],[1,0,"",0,"",0,"VK_ZOOM","",""],[1,0,"",0,"",0,"VK_NONAME","",""],[1,0,"",0,"",0,"VK_PA1","",""],[1,0,"",0,"",0,"VK_OEM_CLEAR","",""]],t=[],i=[];for(const n of e){const[o,r,a,l,c,d,h,u,g]=n;if(i[r]||(i[r]=!0,tde[a]=r,ide[a.toLowerCase()]=r,o&&(H6[r]=l)),!t[l]){if(t[l]=!0,!c)throw new Error(`String representation missing for key code ${l} around scan code ${a}`);bk.define(l,c),nP.define(l,u||c),sP.define(l,g||u||c)}d&&($X[d]=l)}})();var Ad;(function(s){function e(a){return bk.keyCodeToStr(a)}s.toString=e;function t(a){return bk.strToKeyCode(a)}s.fromString=t;function i(a){return nP.keyCodeToStr(a)}s.toUserSettingsUS=i;function n(a){return sP.keyCodeToStr(a)}s.toUserSettingsGeneral=n;function o(a){return nP.strToKeyCode(a)||sP.strToKeyCode(a)}s.fromUserSettings=o;function r(a){if(a>=98&&a<=113)return null;switch(a){case 16:return"Up";case 18:return"Down";case 15:return"Left";case 17:return"Right"}return bk.keyCodeToStr(a)}s.toElectronAccelerator=r})(Ad||(Ad={}));function Gi(s,e){const t=(e&65535)<<16>>>0;return(s|t)>>>0}function oP(s,e){if(typeof s=="number"){if(s===0)return null;const t=(s&65535)>>>0,i=(s&4294901760)>>>16;return i!==0?new N2([$x(t,e),$x(i,e)]):new N2([$x(t,e)])}else{const t=[];for(let i=0;i<s.length;i++)t.push($x(s[i],e));return new N2(t)}}function $x(s,e){const t=!!(s&2048),i=!!(s&256),n=e===2?i:t,o=!!(s&1024),r=!!(s&512),a=e===2?t:i,l=s&255;return new dh(n,o,r,a,l)}class dh{constructor(e,t,i,n,o){this.ctrlKey=e,this.shiftKey=t,this.altKey=i,this.metaKey=n,this.keyCode=o}equals(e){return e instanceof dh&&this.ctrlKey===e.ctrlKey&&this.shiftKey===e.shiftKey&&this.altKey===e.altKey&&this.metaKey===e.metaKey&&this.keyCode===e.keyCode}isModifierKey(){return this.keyCode===0||this.keyCode===5||this.keyCode===57||this.keyCode===6||this.keyCode===4}isDuplicateModifierCase(){return this.ctrlKey&&this.keyCode===5||this.shiftKey&&this.keyCode===4||this.altKey&&this.keyCode===6||this.metaKey&&this.keyCode===57}}class N2{constructor(e){if(e.length===0)throw Qr("chords");this.chords=e}}class nde{constructor(e,t,i,n,o,r){this.ctrlKey=e,this.shiftKey=t,this.altKey=i,this.metaKey=n,this.keyLabel=o,this.keyAriaLabel=r}}class sde{}function ode(s){if(s.charCode){const t=String.fromCharCode(s.charCode).toUpperCase();return Ad.fromString(t)}const e=s.keyCode;if(e===3)return 7;if(wo)switch(e){case 59:return 85;case 60:if(vo)return 97;break;case 61:return 86;case 107:return 109;case 109:return 111;case 173:return 88;case 224:if(Ye)return 57;break}else if(xR){if(Ye&&e===93)return 57;if(!Ye&&e===92)return 57}return $X[e]||0}const rde=Ye?256:2048,ade=512,lde=1024,cde=Ye?2048:256;class Pt{constructor(e){this._standardKeyboardEventBrand=!0;const t=e;this.browserEvent=t,this.target=t.target,this.ctrlKey=t.ctrlKey,this.shiftKey=t.shiftKey,this.altKey=t.altKey,this.metaKey=t.metaKey,this.altGraphKey=t.getModifierState?.("AltGraph"),this.keyCode=ode(t),this.code=t.code,this.ctrlKey=this.ctrlKey||this.keyCode===5,this.altKey=this.altKey||this.keyCode===6,this.shiftKey=this.shiftKey||this.keyCode===4,this.metaKey=this.metaKey||this.keyCode===57,this._asKeybinding=this._computeKeybinding(),this._asKeyCodeChord=this._computeKeyCodeChord()}preventDefault(){this.browserEvent&&this.browserEvent.preventDefault&&this.browserEvent.preventDefault()}stopPropagation(){this.browserEvent&&this.browserEvent.stopPropagation&&this.browserEvent.stopPropagation()}toKeyCodeChord(){return this._asKeyCodeChord}equals(e){return this._asKeybinding===e}_computeKeybinding(){let e=0;this.keyCode!==5&&this.keyCode!==4&&this.keyCode!==6&&this.keyCode!==57&&(e=this.keyCode);let t=0;return this.ctrlKey&&(t|=rde),this.altKey&&(t|=ade),this.shiftKey&&(t|=lde),this.metaKey&&(t|=cde),t|=e,t}_computeKeyCodeChord(){let e=0;return this.keyCode!==5&&this.keyCode!==4&&this.keyCode!==6&&this.keyCode!==57&&(e=this.keyCode),new dh(this.ctrlKey,this.shiftKey,this.altKey,this.metaKey,e)}}const kz=new WeakMap;function dde(s){if(!s.parent||s.parent===s)return null;try{const e=s.location,t=s.parent.location;if(e.origin!=="null"&&t.origin!=="null"&&e.origin!==t.origin)return null}catch{return null}return s.parent}class hde{static getSameOriginWindowChain(e){let t=kz.get(e);if(!t){t=[],kz.set(e,t);let i=e,n;do n=dde(i),n?t.push({window:new WeakRef(i),iframeElement:i.frameElement||null}):t.push({window:new WeakRef(i),iframeElement:null}),i=n;while(i)}return t.slice(0)}static getPositionOfChildWindowRelativeToAncestorWindow(e,t){if(!t||e===t)return{top:0,left:0};let i=0,n=0;const o=this.getSameOriginWindowChain(e);for(const r of o){const a=r.window.deref();if(i+=a?.scrollY??0,n+=a?.scrollX??0,a===t||!r.iframeElement)break;const l=r.iframeElement.getBoundingClientRect();i+=l.top,n+=l.left}return{top:i,left:n}}}class is{constructor(e,t){this.timestamp=Date.now(),this.browserEvent=t,this.leftButton=t.button===0,this.middleButton=t.button===1,this.rightButton=t.button===2,this.buttons=t.buttons,this.defaultPrevented=t.defaultPrevented,this.target=t.target,this.detail=t.detail||1,t.type==="dblclick"&&(this.detail=2),this.ctrlKey=t.ctrlKey,this.shiftKey=t.shiftKey,this.altKey=t.altKey,this.metaKey=t.metaKey,typeof t.pageX=="number"?(this.posx=t.pageX,this.posy=t.pageY):(this.posx=t.clientX+this.target.ownerDocument.body.scrollLeft+this.target.ownerDocument.documentElement.scrollLeft,this.posy=t.clientY+this.target.ownerDocument.body.scrollTop+this.target.ownerDocument.documentElement.scrollTop);const i=hde.getPositionOfChildWindowRelativeToAncestorWindow(e,t.view);this.posx-=i.left,this.posy-=i.top}preventDefault(){this.browserEvent.preventDefault()}stopPropagation(){this.browserEvent.stopPropagation()}}class Cm{constructor(e,t=0,i=0){this.browserEvent=e||null,this.target=e?e.target||e.targetNode||e.srcElement:null,this.deltaY=i,this.deltaX=t;let n=!1;if(B0){const o=navigator.userAgent.match(/Chrome\/(\d+)/);n=(o?parseInt(o[1]):123)<=122}if(e){const o=e,r=e,a=e.view?.devicePixelRatio||1;if(typeof o.wheelDeltaY<"u")n?this.deltaY=o.wheelDeltaY/(120*a):this.deltaY=o.wheelDeltaY/120;else if(typeof r.VERTICAL_AXIS<"u"&&r.axis===r.VERTICAL_AXIS)this.deltaY=-r.detail/3;else if(e.type==="wheel"){const l=e;l.deltaMode===l.DOM_DELTA_LINE?wo&&!Ye?this.deltaY=-e.deltaY/3:this.deltaY=-e.deltaY:this.deltaY=-e.deltaY/40}if(typeof o.wheelDeltaX<"u")wm&&Bn?this.deltaX=-(o.wheelDeltaX/120):n?this.deltaX=o.wheelDeltaX/(120*a):this.deltaX=o.wheelDeltaX/120;else if(typeof r.HORIZONTAL_AXIS<"u"&&r.axis===r.HORIZONTAL_AXIS)this.deltaX=-e.detail/3;else if(e.type==="wheel"){const l=e;l.deltaMode===l.DOM_DELTA_LINE?wo&&!Ye?this.deltaX=-e.deltaX/3:this.deltaX=-e.deltaX:this.deltaX=-e.deltaX/40}this.deltaY===0&&this.deltaX===0&&e.wheelDelta&&(n?this.deltaY=e.wheelDelta/(120*a):this.deltaY=e.wheelDelta/120)}}preventDefault(){this.browserEvent?.preventDefault()}stopPropagation(){this.browserEvent?.stopPropagation()}}function ym(s,e){const t=this;let i=!1,n;return function(){return i||(i=!0,n=s.apply(t,arguments)),n}}var ot;(function(s){function e(L){return!!L&&typeof L=="object"&&typeof L[Symbol.iterator]=="function"}s.is=e;const t=Object.freeze([]);function i(){return t}s.empty=i;function*n(L){yield L}s.single=n;function o(L){return e(L)?L:n(L)}s.wrap=o;function r(L){return L||t}s.from=r;function*a(L){for(let E=L.length-1;E>=0;E--)yield L[E]}s.reverse=a;function l(L){return!L||L[Symbol.iterator]().next().done===!0}s.isEmpty=l;function c(L){return L[Symbol.iterator]().next().value}s.first=c;function d(L,E){let I=0;for(const R of L)if(E(R,I++))return!0;return!1}s.some=d;function h(L,E){let I=0;for(const R of L)if(!E(R,I++))return!1;return!0}s.every=h;function u(L,E){for(const I of L)if(E(I))return I}s.find=u;function*g(L,E){for(const I of L)E(I)&&(yield I)}s.filter=g;function*f(L,E){let I=0;for(const R of L)yield E(R,I++)}s.map=f;function*p(L,E){let I=0;for(const R of L)yield*E(R,I++)}s.flatMap=p;function*_(...L){for(const E of L)tP(E)?yield*E:yield E}s.concat=_;function b(L,E,I){let R=I;for(const O of L)R=E(R,O);return R}s.reduce=b;function w(L){let E=0;for(const I of L)E++;return E}s.length=w;function*C(L,E,I=L.length){for(E<-L.length&&(E=0),E<0&&(E+=L.length),I<0?I+=L.length:I>L.length&&(I=L.length);E<I;E++)yield L[E]}s.slice=C;function v(L,E=Number.POSITIVE_INFINITY){const I=[];if(E===0)return[I,L];const R=L[Symbol.iterator]();for(let O=0;O<E;O++){const P=R.next();if(P.done)return[I,s.empty()];I.push(P.value)}return[I,{[Symbol.iterator](){return R}}]}s.consume=v;async function y(L){const E=[];for await(const I of L)E.push(I);return E}s.asyncToArray=y;async function x(L){let E=[];for await(const I of L)E=E.concat(I);return E}s.asyncToArrayFlat=x})(ot||(ot={}));function i_(s){return typeof s=="object"&&s!==null&&typeof s.dispose=="function"&&s.dispose.length===0}function xt(s){if(ot.is(s)){const e=[];for(const t of s)if(t)try{t.dispose()}catch(i){e.push(i)}if(e.length===1)throw e[0];if(e.length>1)throw new AggregateError(e,"Encountered errors while disposing of store");return Array.isArray(s)?[]:s}else if(s)return s.dispose(),s}function Ra(...s){return pe(()=>xt(s))}class ude{constructor(e){this._isDisposed=!1,this._fn=e}dispose(){if(!this._isDisposed){if(!this._fn)throw new Error("Unbound disposable context: Need to use an arrow function to preserve the value of this");this._isDisposed=!0,this._fn()}}}function pe(s){return new ude(s)}var Cp;let U=(Cp=class{constructor(){this._toDispose=new Set,this._isDisposed=!1}dispose(){this._isDisposed||(this._isDisposed=!0,this.clear())}get isDisposed(){return this._isDisposed}clear(){if(this._toDispose.size!==0)try{xt(this._toDispose)}finally{this._toDispose.clear()}}add(e){if(!e||e===F.None)return e;if(e===this)throw new Error("Cannot register a disposable on itself!");return this._isDisposed?Cp.DISABLE_DISPOSED_WARNING||console.warn(new Error("Trying to add a disposable to a DisposableStore that has already been disposed of. The added object will be leaked!").stack):this._toDispose.add(e),e}delete(e){if(e){if(e===this)throw new Error("Cannot dispose a disposable on itself!");this._toDispose.delete(e),e.dispose()}}},Cp.DISABLE_DISPOSED_WARNING=!1,Cp);var Ww;let F=(Ww=class{constructor(){this._store=new U,this._store}dispose(){this._store.dispose()}_register(e){if(e===this)throw new Error("Cannot register a disposable on itself!");return this._store.add(e)}},Ww.None=Object.freeze({dispose(){}}),Ww);class vt{constructor(){this._isDisposed=!1}get value(){return this._isDisposed?void 0:this._value}set value(e){this._isDisposed||e===this._value||(this._value?.dispose(),this._value=e)}clear(){this.value=void 0}dispose(){this._isDisposed=!0,this._value?.dispose(),this._value=void 0}}class gde{constructor(e){this._disposable=e,this._counter=1}acquire(){return this._counter++,this}release(){return--this._counter===0&&this._disposable.dispose(),this}}class fde{constructor(e){this.object=e}dispose(){}}class LR{constructor(){this._store=new Map,this._isDisposed=!1}dispose(){this._isDisposed=!0,this.clearAndDisposeAll()}clearAndDisposeAll(){if(this._store.size)try{xt(this._store.values())}finally{this._store.clear()}}get(e){return this._store.get(e)}set(e,t,i=!1){this._isDisposed&&console.warn(new Error("Trying to add a disposable to a DisposableMap that has already been disposed of. The added object will be leaked!").stack),i||this._store.get(e)?.dispose(),this._store.set(e,t)}deleteAndDispose(e){this._store.get(e)?.dispose(),this._store.delete(e)}values(){return this._store.values()}[Symbol.iterator](){return this._store[Symbol.iterator]()}}var Kd;let gn=(Kd=class{constructor(e){this.element=e,this.next=Kd.Undefined,this.prev=Kd.Undefined}},Kd.Undefined=new Kd(void 0),Kd);class Ns{constructor(){this._first=gn.Undefined,this._last=gn.Undefined,this._size=0}get size(){return this._size}isEmpty(){return this._first===gn.Undefined}clear(){let e=this._first;for(;e!==gn.Undefined;){const t=e.next;e.prev=gn.Undefined,e.next=gn.Undefined,e=t}this._first=gn.Undefined,this._last=gn.Undefined,this._size=0}unshift(e){return this._insert(e,!1)}push(e){return this._insert(e,!0)}_insert(e,t){const i=new gn(e);if(this._first===gn.Undefined)this._first=i,this._last=i;else if(t){const o=this._last;this._last=i,i.prev=o,o.next=i}else{const o=this._first;this._first=i,i.next=o,o.prev=i}this._size+=1;let n=!1;return()=>{n||(n=!0,this._remove(i))}}shift(){if(this._first!==gn.Undefined){const e=this._first.element;return this._remove(this._first),e}}pop(){if(this._last!==gn.Undefined){const e=this._last.element;return this._remove(this._last),e}}_remove(e){if(e.prev!==gn.Undefined&&e.next!==gn.Undefined){const t=e.prev;t.next=e.next,e.next.prev=t}else e.prev===gn.Undefined&&e.next===gn.Undefined?(this._first=gn.Undefined,this._last=gn.Undefined):e.next===gn.Undefined?(this._last=this._last.prev,this._last.next=gn.Undefined):e.prev===gn.Undefined&&(this._first=this._first.next,this._first.prev=gn.Undefined);this._size-=1}*[Symbol.iterator](){let e=this._first;for(;e!==gn.Undefined;)yield e.element,e=e.next}}const pde=globalThis.performance.now.bind(globalThis.performance);class Cn{static create(e){return new Cn(e)}constructor(e){this._now=e===!1?Date.now:pde,this._startTime=this._now(),this._stopTime=-1}stop(){this._stopTime=this._now()}reset(){this._startTime=this._now(),this._stopTime=-1}elapsed(){return this._stopTime!==-1?this._stopTime-this._startTime:this._now()-this._startTime}}var oe;(function(s){s.None=()=>F.None;function e(P,B){return u(P,()=>{},0,void 0,!0,void 0,B)}s.defer=e;function t(P){return(B,$=null,z)=>{let Q=!1,J;return J=P(ne=>{if(!Q)return J?J.dispose():Q=!0,B.call($,ne)},null,z),Q&&J.dispose(),J}}s.once=t;function i(P,B){return s.once(s.filter(P,B))}s.onceIf=i;function n(P,B,$){return d((z,Q=null,J)=>P(ne=>z.call(Q,B(ne)),null,J),$)}s.map=n;function o(P,B,$){return d((z,Q=null,J)=>P(ne=>{B(ne),z.call(Q,ne)},null,J),$)}s.forEach=o;function r(P,B,$){return d((z,Q=null,J)=>P(ne=>B(ne)&&z.call(Q,ne),null,J),$)}s.filter=r;function a(P){return P}s.signal=a;function l(...P){return(B,$=null,z)=>{const Q=Ra(...P.map(J=>J(ne=>B.call($,ne))));return h(Q,z)}}s.any=l;function c(P,B,$,z){let Q=$;return n(P,J=>(Q=B(Q,J),Q),z)}s.reduce=c;function d(P,B){let $;const z={onWillAddFirstListener(){$=P(Q.fire,Q)},onDidRemoveLastListener(){$?.dispose()}},Q=new M(z);return B?.add(Q),Q.event}function h(P,B){return B instanceof Array?B.push(P):B&&B.add(P),P}function u(P,B,$=100,z=!1,Q=!1,J,ne){let de,ke,Me,Ze=0,Xe;const At={leakWarningThreshold:J,onWillAddFirstListener(){de=P(un=>{Ze++,ke=B(ke,un),z&&!Me&&(Et.fire(ke),ke=void 0),Xe=()=>{const Vi=ke;ke=void 0,Me=void 0,(!z||Ze>1)&&Et.fire(Vi),Ze=0},typeof $=="number"?(Me&&clearTimeout(Me),Me=setTimeout(Xe,$)):Me===void 0&&(Me=null,queueMicrotask(Xe))})},onWillRemoveListener(){Q&&Ze>0&&Xe?.()},onDidRemoveLastListener(){Xe=void 0,de.dispose()}},Et=new M(At);return ne?.add(Et),Et.event}s.debounce=u;function g(P,B=0,$){return s.debounce(P,(z,Q)=>z?(z.push(Q),z):[Q],B,void 0,!0,void 0,$)}s.accumulate=g;function f(P,B=(z,Q)=>z===Q,$){let z=!0,Q;return r(P,J=>{const ne=z||!B(J,Q);return z=!1,Q=J,ne},$)}s.latch=f;function p(P,B,$){return[s.filter(P,B,$),s.filter(P,z=>!B(z),$)]}s.split=p;function _(P,B=!1,$=[],z){let Q=$.slice(),J=P(ke=>{Q?Q.push(ke):de.fire(ke)});z&&z.add(J);const ne=()=>{Q?.forEach(ke=>de.fire(ke)),Q=null},de=new M({onWillAddFirstListener(){J||(J=P(ke=>de.fire(ke)),z&&z.add(J))},onDidAddFirstListener(){Q&&(B?setTimeout(ne):ne())},onDidRemoveLastListener(){J&&J.dispose(),J=null}});return z&&z.add(de),de.event}s.buffer=_;function b(P,B){return(z,Q,J)=>{const ne=B(new C);return P(function(de){const ke=ne.evaluate(de);ke!==w&&z.call(Q,ke)},void 0,J)}}s.chain=b;const w=Symbol("HaltChainable");class C{constructor(){this.steps=[]}map(B){return this.steps.push(B),this}forEach(B){return this.steps.push($=>(B($),$)),this}filter(B){return this.steps.push($=>B($)?$:w),this}reduce(B,$){let z=$;return this.steps.push(Q=>(z=B(z,Q),z)),this}latch(B=($,z)=>$===z){let $=!0,z;return this.steps.push(Q=>{const J=$||!B(Q,z);return $=!1,z=Q,J?Q:w}),this}evaluate(B){for(const $ of this.steps)if(B=$(B),B===w)break;return B}}function v(P,B,$=z=>z){const z=(...de)=>ne.fire($(...de)),Q=()=>P.on(B,z),J=()=>P.removeListener(B,z),ne=new M({onWillAddFirstListener:Q,onDidRemoveLastListener:J});return ne.event}s.fromNodeEventEmitter=v;function y(P,B,$=z=>z){const z=(...de)=>ne.fire($(...de)),Q=()=>P.addEventListener(B,z),J=()=>P.removeEventListener(B,z),ne=new M({onWillAddFirstListener:Q,onDidRemoveLastListener:J});return ne.event}s.fromDOMEventEmitter=y;function x(P,B){let $;const z=new Promise((Q,J)=>{const ne=t(P)(Q,null,B);$=()=>ne.dispose()});return z.cancel=$,z}s.toPromise=x;function L(P,B){return P($=>B.fire($))}s.forward=L;function E(P,B,$){return B($),P(z=>B(z))}s.runAndSubscribe=E;class I{constructor(B,$){this._observable=B,this._counter=0,this._hasChanged=!1;const z={onWillAddFirstListener:()=>{B.addObserver(this),this._observable.reportChanges()},onDidRemoveLastListener:()=>{B.removeObserver(this)}};this.emitter=new M(z),$&&$.add(this.emitter)}beginUpdate(B){this._counter++}handlePossibleChange(B){}handleChange(B,$){this._hasChanged=!0}endUpdate(B){this._counter--,this._counter===0&&(this._observable.reportChanges(),this._hasChanged&&(this._hasChanged=!1,this.emitter.fire(this._observable.get())))}}function R(P,B){return new I(P,B).emitter.event}s.fromObservable=R;function O(P){return(B,$,z)=>{let Q=0,J=!1;const ne={beginUpdate(){Q++},endUpdate(){Q--,Q===0&&(P.reportChanges(),J&&(J=!1,B.call($)))},handlePossibleChange(){},handleChange(){J=!0}};P.addObserver(ne),P.reportChanges();const de={dispose(){P.removeObserver(ne)}};return z instanceof U?z.add(de):Array.isArray(z)&&z.push(de),de}}s.fromObservableLight=O})(oe||(oe={}));const Ub=class Ub{constructor(e){this.listenerCount=0,this.invocationCount=0,this.elapsedOverall=0,this.durations=[],this.name=`${e}_${Ub._idPool++}`,Ub.all.add(this)}start(e){this._stopWatch=new Cn,this.listenerCount=e}stop(){if(this._stopWatch){const e=this._stopWatch.elapsed();this.durations.push(e),this.elapsedOverall+=e,this.invocationCount+=1,this._stopWatch=void 0}}};Ub.all=new Set,Ub._idPool=0;let rP=Ub,mde=-1;const WT=class WT{constructor(e,t,i=(WT._idPool++).toString(16).padStart(3,"0")){this._errorHandler=e,this.threshold=t,this.name=i,this._warnCountdown=0}dispose(){this._stacks?.clear()}check(e,t){const i=this.threshold;if(i<=0||t<i)return;this._stacks||(this._stacks=new Map);const n=this._stacks.get(e.value)||0;if(this._stacks.set(e.value,n+1),this._warnCountdown-=1,this._warnCountdown<=0){this._warnCountdown=i*.5;const[o,r]=this.getMostFrequentStack(),a=`[${this.name}] potential listener LEAK detected, having ${t} listeners already. MOST frequent listener (${r}):`;console.warn(a),console.warn(o);const l=new _de(a,o);this._errorHandler(l)}return()=>{const o=this._stacks.get(e.value)||0;this._stacks.set(e.value,o-1)}}getMostFrequentStack(){if(!this._stacks)return;let e,t=0;for(const[i,n]of this._stacks)(!e||t<n)&&(e=[i,n],t=n);return e}};WT._idPool=1;let aP=WT;class V6{static create(){const e=new Error;return new V6(e.stack??"")}constructor(e){this.value=e}print(){console.warn(this.value.split(`
`).slice(2).join(`
`))}}class _de extends Error{constructor(e,t){super(e),this.name="ListenerLeakError",this.stack=t}}class bde extends Error{constructor(e,t){super(e),this.name="ListenerRefusalError",this.stack=t}}class R2{constructor(e){this.value=e}}const wde=2;let M=class{constructor(e){this._size=0,this._options=e,this._leakageMon=this._options?.leakWarningThreshold?new aP(e?.onListenerError??Oe,this._options?.leakWarningThreshold??mde):void 0,this._perfMon=this._options?._profName?new rP(this._options._profName):void 0,this._deliveryQueue=this._options?.deliveryQueue}dispose(){this._disposed||(this._disposed=!0,this._deliveryQueue?.current===this&&this._deliveryQueue.reset(),this._listeners&&(this._listeners=void 0,this._size=0),this._options?.onDidRemoveLastListener?.(),this._leakageMon?.dispose())}get event(){return this._event??=(e,t,i)=>{if(this._leakageMon&&this._size>this._leakageMon.threshold**2){const a=`[${this._leakageMon.name}] REFUSES to accept new listeners because it exceeded its threshold by far (${this._size} vs ${this._leakageMon.threshold})`;console.warn(a);const l=this._leakageMon.getMostFrequentStack()??["UNKNOWN stack",-1],c=new bde(`${a}. HINT: Stack shows most frequent listener (${l[1]}-times)`,l[0]);return(this._options?.onListenerError||Oe)(c),F.None}if(this._disposed)return F.None;t&&(e=e.bind(t));const n=new R2(e);let o;this._leakageMon&&this._size>=Math.ceil(this._leakageMon.threshold*.2)&&(n.stack=V6.create(),o=this._leakageMon.check(n.stack,this._size+1)),this._listeners?this._listeners instanceof R2?(this._deliveryQueue??=new UX,this._listeners=[this._listeners,n]):this._listeners.push(n):(this._options?.onWillAddFirstListener?.(this),this._listeners=n,this._options?.onDidAddFirstListener?.(this)),this._options?.onDidAddListener?.(this),this._size++;const r=pe(()=>{o?.(),this._removeListener(n)});return i instanceof U?i.add(r):Array.isArray(i)&&i.push(r),r},this._event}_removeListener(e){if(this._options?.onWillRemoveListener?.(this),!this._listeners)return;if(this._size===1){this._listeners=void 0,this._options?.onDidRemoveLastListener?.(this),this._size=0;return}const t=this._listeners,i=t.indexOf(e);if(i===-1)throw console.log("disposed?",this._disposed),console.log("size?",this._size),console.log("arr?",JSON.stringify(this._listeners)),new Error("Attempted to dispose unknown listener");this._size--,t[i]=void 0;const n=this._deliveryQueue.current===this;if(this._size*wde<=t.length){let o=0;for(let r=0;r<t.length;r++)t[r]?t[o++]=t[r]:n&&o<this._deliveryQueue.end&&(this._deliveryQueue.end--,o<this._deliveryQueue.i&&this._deliveryQueue.i--);t.length=o}}_deliver(e,t){if(!e)return;const i=this._options?.onListenerError||Oe;if(!i){e.value(t);return}try{e.value(t)}catch(n){i(n)}}_deliverQueue(e){const t=e.current._listeners;for(;e.i<e.end;)this._deliver(t[e.i++],e.value);e.reset()}fire(e){if(this._deliveryQueue?.current&&(this._deliverQueue(this._deliveryQueue),this._perfMon?.stop()),this._perfMon?.start(this._size),this._listeners)if(this._listeners instanceof R2)this._deliver(this._listeners,e);else{const t=this._deliveryQueue;t.enqueue(this,e,this._listeners.length),this._deliverQueue(t)}this._perfMon?.stop()}hasListeners(){return this._size>0}};const vde=()=>new UX;class UX{constructor(){this.i=-1,this.end=0}enqueue(e,t,i){this.i=0,this.end=i,this.current=e,this.value=t}reset(){this.i=this.end,this.current=void 0,this.value=void 0}}class Sm extends M{constructor(e){super(e),this._isPaused=0,this._eventQueue=new Ns,this._mergeFn=e?.merge}pause(){this._isPaused++}resume(){if(this._isPaused!==0&&--this._isPaused===0)if(this._mergeFn){if(this._eventQueue.size>0){const e=Array.from(this._eventQueue);this._eventQueue.clear(),super.fire(this._mergeFn(e))}}else for(;!this._isPaused&&this._eventQueue.size!==0;)super.fire(this._eventQueue.shift())}fire(e){this._size&&(this._isPaused!==0?this._eventQueue.push(e):super.fire(e))}}class qX extends Sm{constructor(e){super(e),this._delay=e.delay??100}fire(e){this._handle||(this.pause(),this._handle=setTimeout(()=>{this._handle=void 0,this.resume()},this._delay)),super.fire(e)}}class Cde extends M{constructor(e){super(e),this._queuedEvents=[],this._mergeFn=e?.merge}fire(e){this.hasListeners()&&(this._queuedEvents.push(e),this._queuedEvents.length===1&&queueMicrotask(()=>{this._mergeFn?super.fire(this._mergeFn(this._queuedEvents)):this._queuedEvents.forEach(t=>super.fire(t)),this._queuedEvents=[]}))}}class yde{constructor(){this.hasListeners=!1,this.events=[],this.emitter=new M({onWillAddFirstListener:()=>this.onFirstListenerAdd(),onDidRemoveLastListener:()=>this.onLastListenerRemove()})}get event(){return this.emitter.event}add(e){const t={event:e,listener:null};return this.events.push(t),this.hasListeners&&this.hook(t),pe(ym(()=>{this.hasListeners&&this.unhook(t);const n=this.events.indexOf(t);this.events.splice(n,1)}))}onFirstListenerAdd(){this.hasListeners=!0,this.events.forEach(e=>this.hook(e))}onLastListenerRemove(){this.hasListeners=!1,this.events.forEach(e=>this.unhook(e))}hook(e){e.listener=e.event(t=>this.emitter.fire(t))}unhook(e){e.listener?.dispose(),e.listener=null}dispose(){this.emitter.dispose();for(const e of this.events)e.listener?.dispose();this.events=[]}}class $0{constructor(){this.data=[]}wrapEvent(e,t,i){return(n,o,r)=>e(a=>{const l=this.data[this.data.length-1];if(!t){l?l.buffers.push(()=>n.call(o,a)):n.call(o,a);return}const c=l;if(!c){n.call(o,t(i,a));return}c.items??=[],c.items.push(a),c.buffers.length===0&&l.buffers.push(()=>{c.reducedResult??=i?c.items.reduce(t,i):c.items.reduce(t),n.call(o,c.reducedResult)})},void 0,r)}bufferEvents(e){const t={buffers:new Array};this.data.push(t);const i=e();return this.data.pop(),t.buffers.forEach(n=>n()),i}}class yC{constructor(){this.listening=!1,this.inputEvent=oe.None,this.inputEventListener=F.None,this.emitter=new M({onDidAddFirstListener:()=>{this.listening=!0,this.inputEventListener=this.inputEvent(this.emitter.fire,this.emitter)},onDidRemoveLastListener:()=>{this.listening=!1,this.inputEventListener.dispose()}}),this.event=this.emitter.event}set input(e){this.inputEvent=e,this.listening&&(this.inputEventListener.dispose(),this.inputEventListener=e(this.emitter.fire,this.emitter))}dispose(){this.inputEventListener.dispose(),this.emitter.dispose()}}const jX=Object.freeze(function(s,e){const t=setTimeout(s.bind(e),0);return{dispose(){clearTimeout(t)}}});var Ke;(function(s){function e(t){return t===s.None||t===s.Cancelled||t instanceof wk?!0:!t||typeof t!="object"?!1:typeof t.isCancellationRequested=="boolean"&&typeof t.onCancellationRequested=="function"}s.isCancellationToken=e,s.None=Object.freeze({isCancellationRequested:!1,onCancellationRequested:oe.None}),s.Cancelled=Object.freeze({isCancellationRequested:!0,onCancellationRequested:jX})})(Ke||(Ke={}));class wk{constructor(){this._isCancelled=!1,this._emitter=null}cancel(){this._isCancelled||(this._isCancelled=!0,this._emitter&&(this._emitter.fire(void 0),this.dispose()))}get isCancellationRequested(){return this._isCancelled}get onCancellationRequested(){return this._isCancelled?jX:(this._emitter||(this._emitter=new M),this._emitter.event)}dispose(){this._emitter&&(this._emitter.dispose(),this._emitter=null)}}let si=class{constructor(e){this._token=void 0,this._parentListener=void 0,this._parentListener=e&&e.onCancellationRequested(this.cancel,this)}get token(){return this._token||(this._token=new wk),this._token}cancel(){this._token?this._token instanceof wk&&this._token.cancel():this._token=Ke.Cancelled}dispose(e=!1){e&&this.cancel(),this._parentListener?.dispose(),this._token?this._token instanceof wk&&this._token.dispose():this._token=Ke.None}};function lP(s){const e=new si;return s.add({dispose(){e.cancel()}}),e.token}const GX=Symbol("MicrotaskDelay");function cP(s){return!!s&&typeof s.then=="function"}function an(s){const e=new si,t=s(e.token);let i=!1;const n=new Promise((o,r)=>{const a=e.token.onCancellationRequested(()=>{i=!0,a.dispose(),r(new ia)});Promise.resolve(t).then(l=>{a.dispose(),e.dispose(),i?i_(l)&&l.dispose():o(l)},l=>{a.dispose(),e.dispose(),r(l)})});return new class{cancel(){e.cancel(),e.dispose()}then(o,r){return n.then(o,r)}catch(o){return this.then(void 0,o)}finally(o){return n.finally(o)}}}function tv(s,e,t){return new Promise((i,n)=>{const o=e.onCancellationRequested(()=>{o.dispose(),i(t)});s.then(i,n).finally(()=>o.dispose())})}function Sde(s,e){return new Promise((t,i)=>{const n=e.onCancellationRequested(()=>{n.dispose(),i(new ia)});s.then(t,i).finally(()=>n.dispose())})}let xde=class{constructor(){this.activePromise=null,this.queuedPromise=null,this.queuedPromiseFactory=null,this.cancellationTokenSource=new si}queue(e){if(this.cancellationTokenSource.token.isCancellationRequested)return Promise.reject(new Error("Throttler is disposed"));if(this.activePromise){if(this.queuedPromiseFactory=e,!this.queuedPromise){const t=()=>{if(this.queuedPromise=null,this.cancellationTokenSource.token.isCancellationRequested)return;const i=this.queue(this.queuedPromiseFactory);return this.queuedPromiseFactory=null,i};this.queuedPromise=new Promise(i=>{this.activePromise.then(t,t).then(i)})}return new Promise((t,i)=>{this.queuedPromise.then(t,i)})}return this.activePromise=e(this.cancellationTokenSource.token),new Promise((t,i)=>{this.activePromise.then(n=>{this.activePromise=null,t(n)},n=>{this.activePromise=null,i(n)})})}dispose(){this.cancellationTokenSource.cancel()}};const Lde=(s,e)=>{let t=!0;const i=setTimeout(()=>{t=!1,e()},s);return{isTriggered:()=>t,dispose:()=>{clearTimeout(i),t=!1}}},kde=s=>{let e=!0;return queueMicrotask(()=>{e&&(e=!1,s())}),{isTriggered:()=>e,dispose:()=>{e=!1}}};class sa{constructor(e){this.defaultDelay=e,this.deferred=null,this.completionPromise=null,this.doResolve=null,this.doReject=null,this.task=null}trigger(e,t=this.defaultDelay){this.task=e,this.cancelTimeout(),this.completionPromise||(this.completionPromise=new Promise((n,o)=>{this.doResolve=n,this.doReject=o}).then(()=>{if(this.completionPromise=null,this.doResolve=null,this.task){const n=this.task;return this.task=null,n()}}));const i=()=>{this.deferred=null,this.doResolve?.(null)};return this.deferred=t===GX?kde(i):Lde(t,i),this.completionPromise}isTriggered(){return!!this.deferred?.isTriggered()}cancel(){this.cancelTimeout(),this.completionPromise&&(this.doReject?.(new ia),this.completionPromise=null)}cancelTimeout(){this.deferred?.dispose(),this.deferred=null}dispose(){this.cancel()}}class KX{constructor(e){this.delayer=new sa(e),this.throttler=new xde}trigger(e,t){return this.delayer.trigger(()=>this.throttler.queue(e),t)}cancel(){this.delayer.cancel()}dispose(){this.delayer.dispose(),this.throttler.dispose()}}function Tc(s,e){return e?new Promise((t,i)=>{const n=setTimeout(()=>{o.dispose(),t()},s),o=e.onCancellationRequested(()=>{clearTimeout(n),o.dispose(),i(new ia)})}):an(t=>Tc(s,t))}function hh(s,e=0,t){const i=setTimeout(()=>{s(),t&&n.dispose()},e),n=pe(()=>{clearTimeout(i),t?.delete(n)});return t?.add(n),n}function z6(s,e=i=>!!i,t=null){let i=0;const n=s.length,o=()=>{if(i>=n)return Promise.resolve(t);const r=s[i++];return Promise.resolve(r()).then(l=>e(l)?Promise.resolve(l):o())};return o()}let Ede=class{constructor(){this._runningTask=void 0,this._pendingTasks=[]}schedule(e){const t=new n_;return this._pendingTasks.push({task:e,deferred:t,setUndefinedWhenCleared:!1}),this._runIfNotRunning(),t.p}_runIfNotRunning(){this._runningTask===void 0&&this._processQueue()}async _processQueue(){if(this._pendingTasks.length===0)return;const e=this._pendingTasks.shift();if(e){if(this._runningTask)throw new Ce;this._runningTask=e.task;try{const t=await e.task();e.deferred.complete(t)}catch(t){e.deferred.error(t)}finally{this._runningTask=void 0,this._processQueue()}}}clearPending(){const e=this._pendingTasks;this._pendingTasks=[];for(const t of e)t.setUndefinedWhenCleared?t.deferred.complete(void 0):t.deferred.error(new ia)}};class $o{constructor(e,t){this._isDisposed=!1,this._token=void 0,typeof e=="function"&&typeof t=="number"&&this.setIfNotSet(e,t)}dispose(){this.cancel(),this._isDisposed=!0}cancel(){this._token!==void 0&&(clearTimeout(this._token),this._token=void 0)}cancelAndSet(e,t){if(this._isDisposed)throw new Ce("Calling 'cancelAndSet' on a disposed TimeoutTimer");this.cancel(),this._token=setTimeout(()=>{this._token=void 0,e()},t)}setIfNotSet(e,t){if(this._isDisposed)throw new Ce("Calling 'setIfNotSet' on a disposed TimeoutTimer");this._token===void 0&&(this._token=setTimeout(()=>{this._token=void 0,e()},t))}}class $6{constructor(){this.disposable=void 0,this.isDisposed=!1}cancel(){this.disposable?.dispose(),this.disposable=void 0}cancelAndSet(e,t,i=globalThis){if(this.isDisposed)throw new Ce("Calling 'cancelAndSet' on a disposed IntervalTimer");this.cancel();const n=i.setInterval(()=>{e()},t);this.disposable=pe(()=>{i.clearInterval(n),this.disposable=void 0})}dispose(){this.cancel(),this.isDisposed=!0}}class Rt{constructor(e,t){this.timeoutToken=void 0,this.runner=e,this.timeout=t,this.timeoutHandler=this.onTimeout.bind(this)}dispose(){this.cancel(),this.runner=null}cancel(){this.isScheduled()&&(clearTimeout(this.timeoutToken),this.timeoutToken=void 0)}schedule(e=this.timeout){this.cancel(),this.timeoutToken=setTimeout(this.timeoutHandler,e)}get delay(){return this.timeout}set delay(e){this.timeout=e}isScheduled(){return this.timeoutToken!==void 0}onTimeout(){this.timeoutToken=void 0,this.runner&&this.doRun()}doRun(){this.runner?.()}}let U6,E1;(function(){const s=globalThis;typeof s.requestIdleCallback!="function"||typeof s.cancelIdleCallback!="function"?E1=(e,t,i)=>{V0(()=>{if(n)return;const o=Date.now()+15;t(Object.freeze({didTimeout:!0,timeRemaining(){return Math.max(0,o-Date.now())}}))});let n=!1;return{dispose(){n||(n=!0)}}}:E1=(e,t,i)=>{const n=e.requestIdleCallback(t,typeof i=="number"?{timeout:i}:void 0);let o=!1;return{dispose(){o||(o=!0,e.cancelIdleCallback(n))}}},U6=(e,t)=>E1(globalThis,e,t)})();class YX{constructor(e,t){this._didRun=!1,this._executor=()=>{try{this._value=t()}catch(i){this._error=i}finally{this._didRun=!0}},this._handle=E1(e,()=>this._executor())}dispose(){this._handle.dispose()}get value(){if(this._didRun||(this._handle.dispose(),this._executor()),this._error)throw this._error;return this._value}get isInitialized(){return this._didRun}}class Ide extends YX{constructor(e){super(globalThis,e)}}class n_{get isRejected(){return this.outcome?.outcome===1}get isSettled(){return!!this.outcome}constructor(){this.p=new Promise((e,t)=>{this.completeCallback=e,this.errorCallback=t})}complete(e){return this.isSettled?Promise.resolve():new Promise(t=>{this.completeCallback(e),this.outcome={outcome:0,value:e},t()})}error(e){return this.isSettled?Promise.resolve():new Promise(t=>{this.errorCallback(e),this.outcome={outcome:1,value:e},t()})}cancel(){return this.error(new ia)}}var Uy;(function(s){async function e(i){let n;const o=await Promise.all(i.map(r=>r.then(a=>a,a=>{n||(n=a)})));if(typeof n<"u")throw n;return o}s.settled=e;function t(i){return new Promise(async(n,o)=>{try{await i(n,o)}catch(r){o(r)}})}s.withAsyncBody=t})(Uy||(Uy={}));function Dde(s){const e=new si,t=s(e.token);return new Nde(e,async i=>{const n=e.token.onCancellationRequested(()=>{n.dispose(),e.dispose(),i.reject(new ia)});try{for await(const o of t){if(e.token.isCancellationRequested)return;i.emitOne(o)}n.dispose(),e.dispose()}catch(o){n.dispose(),e.dispose(),i.reject(o)}})}class Tde{constructor(){this._unsatisfiedConsumers=[],this._unconsumedValues=[]}get hasFinalValue(){return!!this._finalValue}produce(e){if(this._ensureNoFinalValue(),this._unsatisfiedConsumers.length>0){const t=this._unsatisfiedConsumers.shift();this._resolveOrRejectDeferred(t,e)}else this._unconsumedValues.push(e)}produceFinal(e){this._ensureNoFinalValue(),this._finalValue=e;for(const t of this._unsatisfiedConsumers)this._resolveOrRejectDeferred(t,e);this._unsatisfiedConsumers.length=0}_ensureNoFinalValue(){if(this._finalValue)throw new Ce("ProducerConsumer: cannot produce after final value has been set")}_resolveOrRejectDeferred(e,t){t.ok?e.complete(t.value):e.error(t.error)}consume(){if(this._unconsumedValues.length>0||this._finalValue){const e=this._unconsumedValues.length>0?this._unconsumedValues.shift():this._finalValue;return e.ok?Promise.resolve(e.value):Promise.reject(e.error)}else{const e=new n_;return this._unsatisfiedConsumers.push(e),e.p}}}const or=class or{constructor(e,t){this._onReturn=t,this._producerConsumer=new Tde,this._iterator={next:()=>this._producerConsumer.consume(),return:()=>(this._onReturn?.(),Promise.resolve({done:!0,value:void 0})),throw:async i=>(this._finishError(i),{done:!0,value:void 0})},queueMicrotask(async()=>{const i=e({emitOne:n=>this._producerConsumer.produce({ok:!0,value:{done:!1,value:n}}),emitMany:n=>{for(const o of n)this._producerConsumer.produce({ok:!0,value:{done:!1,value:o}})},reject:n=>this._finishError(n)});if(!this._producerConsumer.hasFinalValue)try{await i,this._finishOk()}catch(n){this._finishError(n)}})}static fromArray(e){return new or(t=>{t.emitMany(e)})}static fromPromise(e){return new or(async t=>{t.emitMany(await e)})}static fromPromisesResolveOrder(e){return new or(async t=>{await Promise.all(e.map(async i=>t.emitOne(await i)))})}static merge(e){return new or(async t=>{await Promise.all(e.map(async i=>{for await(const n of i)t.emitOne(n)}))})}static map(e,t){return new or(async i=>{for await(const n of e)i.emitOne(t(n))})}map(e){return or.map(this,e)}static coalesce(e){return or.filter(e,t=>!!t)}coalesce(){return or.coalesce(this)}static filter(e,t){return new or(async i=>{for await(const n of e)t(n)&&i.emitOne(n)})}filter(e){return or.filter(this,e)}_finishOk(){this._producerConsumer.hasFinalValue||this._producerConsumer.produceFinal({ok:!0,value:{done:!0,value:void 0}})}_finishError(e){this._producerConsumer.hasFinalValue||this._producerConsumer.produceFinal({ok:!1,error:e})}[Symbol.asyncIterator](){return this._iterator}};or.EMPTY=or.fromArray([]);let Jr=or;class Nde extends Jr{constructor(e,t){super(t),this._source=e}cancel(){this._source.cancel()}}function ZX(s){return s}class Rde{constructor(e,t){this.lastCache=void 0,this.lastArgKey=void 0,typeof e=="function"?(this._fn=e,this._computeKey=ZX):(this._fn=t,this._computeKey=e.getCacheKey)}get(e){const t=this._computeKey(e);return this.lastArgKey!==t&&(this.lastArgKey=t,this.lastCache=this._fn(e)),this.lastCache}}class dP{get cachedValues(){return this._map}constructor(e,t){this._map=new Map,this._map2=new Map,typeof e=="function"?(this._fn=e,this._computeKey=ZX):(this._fn=t,this._computeKey=e.getCacheKey)}get(e){const t=this._computeKey(e);if(this._map2.has(t))return this._map2.get(t);const i=this._fn(e);return this._map.set(e,i),this._map2.set(t,i),i}}var hf;(function(s){s[s.Uninitialized=0]="Uninitialized",s[s.Running=1]="Running",s[s.Completed=2]="Completed"})(hf||(hf={}));class es{constructor(e){this.executor=e,this._state=hf.Uninitialized}get value(){if(this._state===hf.Uninitialized){this._state=hf.Running;try{this._value=this.executor()}catch(e){this._error=e}finally{this._state=hf.Completed}}else if(this._state===hf.Running)throw new Error("Cannot read the value of a lazy that is being initialized");if(this._error)throw this._error;return this._value}get rawValue(){return this._value}}function XX(s){return!s||typeof s!="string"?!0:s.trim().length===0}const Ade=/{(\d+)}/g;function xm(s,...e){return e.length===0?s:s.replace(Ade,function(t,i){const n=parseInt(i,10);return isNaN(n)||n<0||n>=e.length?t:e[n]})}function Mde(s){return s.replace(/[<>"'&]/g,e=>{switch(e){case"<":return"&lt;";case">":return"&gt;";case'"':return"&quot;";case"'":return"&apos;";case"&":return"&amp;"}return e})}function Aa(s){return s.replace(/[<>&]/g,function(e){switch(e){case"<":return"&lt;";case">":return"&gt;";case"&":return"&amp;";default:return e}})}function Uo(s){return s.replace(/[\\\{\}\*\+\?\|\^\$\.\[\]\(\)]/g,"\\$&")}function Ode(s,e=" "){const t=U0(s,e);return QX(t,e)}function U0(s,e){if(!s||!e)return s;const t=e.length;if(t===0||s.length===0)return s;let i=0;for(;s.indexOf(e,i)===i;)i=i+t;return s.substring(i)}function QX(s,e){if(!s||!e)return s;const t=e.length,i=s.length;if(t===0||i===0)return s;let n=i,o=-1;for(;o=s.lastIndexOf(e,n-1),!(o===-1||o+t!==n);){if(o===0)return"";n=o}return s.substring(0,n)}function Pde(s){return s.replace(/[\-\\\{\}\+\?\|\^\$\.\,\[\]\(\)\#\s]/g,"\\$&").replace(/[\*]/g,".*")}function JX(s,e,t={}){if(!s)throw new Error("Cannot create regex from empty string");e||(s=Uo(s)),t.wholeWord&&(/\B/.test(s.charAt(0))||(s="\\b"+s),/\B/.test(s.charAt(s.length-1))||(s=s+"\\b"));let i="";return t.global&&(i+="g"),t.matchCase||(i+="i"),t.multiline&&(i+="m"),t.unicode&&(i+="u"),new RegExp(s,i)}function eQ(s){return s.source==="^"||s.source==="^$"||s.source==="$"||s.source==="^\\s*$"?!1:!!(s.exec("")&&s.lastIndex===0)}function qo(s){return s.split(/\r\n|\r|\n/)}function As(s){for(let e=0,t=s.length;e<t;e++){const i=s.charCodeAt(e);if(i!==32&&i!==9)return e}return-1}function Bt(s,e=0,t=s.length){for(let i=e;i<t;i++){const n=s.charCodeAt(i);if(n!==32&&n!==9)return s.substring(e,i)}return s.substring(e,t)}function Ma(s,e=s.length-1){for(let t=e;t>=0;t--){const i=s.charCodeAt(t);if(i!==32&&i!==9)return t}return-1}function qy(s,e){return s<e?-1:s>e?1:0}function q6(s,e,t=0,i=s.length,n=0,o=e.length){for(;t<i&&n<o;t++,n++){const l=s.charCodeAt(t),c=e.charCodeAt(n);if(l<c)return-1;if(l>c)return 1}const r=i-t,a=o-n;return r<a?-1:r>a?1:0}function hP(s,e){return Uv(s,e,0,s.length,0,e.length)}function Uv(s,e,t=0,i=s.length,n=0,o=e.length){for(;t<i&&n<o;t++,n++){let l=s.charCodeAt(t),c=e.charCodeAt(n);if(l===c)continue;if(l>=128||c>=128)return q6(s.toLowerCase(),e.toLowerCase(),t,i,n,o);fu(l)&&(l-=32),fu(c)&&(c-=32);const d=l-c;if(d!==0)return d}const r=i-t,a=o-n;return r<a?-1:r>a?1:0}function Ux(s){return s>=48&&s<=57}function fu(s){return s>=97&&s<=122}function oc(s){return s>=65&&s<=90}function Hf(s,e){return s.length===e.length&&Uv(s,e)===0}function j6(s,e){const t=e.length;return t<=s.length&&Uv(s,e,0,t)===0}function Fde(s,e){const t=s.length,i=t-e.length;return i>=0&&Uv(s,e,i,t)===0}function Fa(s,e){const t=Math.min(s.length,e.length);let i;for(i=0;i<t;i++)if(s.charCodeAt(i)!==e.charCodeAt(i))return i;return t}function uh(s,e){const t=Math.min(s.length,e.length);let i;const n=s.length-1,o=e.length-1;for(i=0;i<t;i++)if(s.charCodeAt(n-i)!==e.charCodeAt(o-i))return i;return t}function sn(s){return 55296<=s&&s<=56319}function ag(s){return 56320<=s&&s<=57343}function G6(s,e){return(s-55296<<10)+(e-56320)+65536}function gE(s,e,t){const i=s.charCodeAt(t);if(sn(i)&&t+1<e){const n=s.charCodeAt(t+1);if(ag(n))return G6(i,n)}return i}function Bde(s,e){const t=s.charCodeAt(e-1);if(ag(t)&&e>1){const i=s.charCodeAt(e-2);if(sn(i))return G6(i,t)}return t}class K6{get offset(){return this._offset}constructor(e,t=0){this._str=e,this._len=e.length,this._offset=t}setOffset(e){this._offset=e}prevCodePoint(){const e=Bde(this._str,this._offset);return this._offset-=e>=65536?2:1,e}nextCodePoint(){const e=gE(this._str,this._len,this._offset);return this._offset+=e>=65536?2:1,e}eol(){return this._offset>=this._len}}class fE{get offset(){return this._iterator.offset}constructor(e,t=0){this._iterator=new K6(e,t)}nextGraphemeLength(){const e=pE.getInstance(),t=this._iterator,i=t.offset;let n=e.getGraphemeBreakType(t.nextCodePoint());for(;!t.eol();){const o=t.offset,r=e.getGraphemeBreakType(t.nextCodePoint());if(Ez(n,r)){t.setOffset(o);break}n=r}return t.offset-i}prevGraphemeLength(){const e=pE.getInstance(),t=this._iterator,i=t.offset;let n=e.getGraphemeBreakType(t.prevCodePoint());for(;t.offset>0;){const o=t.offset,r=e.getGraphemeBreakType(t.prevCodePoint());if(Ez(r,n)){t.setOffset(o);break}n=r}return i-t.offset}eol(){return this._iterator.eol()}}function Y6(s,e){return new fE(s,e).nextGraphemeLength()}function tQ(s,e){return new fE(s,e).prevGraphemeLength()}function Wde(s,e){e>0&&ag(s.charCodeAt(e))&&e--;const t=e+Y6(s,e);return[t-tQ(s,t),t]}let A2;function Hde(){return/(?:[\u05BE\u05C0\u05C3\u05C6\u05D0-\u05F4\u0608\u060B\u060D\u061B-\u064A\u066D-\u066F\u0671-\u06D5\u06E5\u06E6\u06EE\u06EF\u06FA-\u0710\u0712-\u072F\u074D-\u07A5\u07B1-\u07EA\u07F4\u07F5\u07FA\u07FE-\u0815\u081A\u0824\u0828\u0830-\u0858\u085E-\u088E\u08A0-\u08C9\u200F\uFB1D\uFB1F-\uFB28\uFB2A-\uFD3D\uFD50-\uFDC7\uFDF0-\uFDFC\uFE70-\uFEFC]|\uD802[\uDC00-\uDD1B\uDD20-\uDE00\uDE10-\uDE35\uDE40-\uDEE4\uDEEB-\uDF35\uDF40-\uDFFF]|\uD803[\uDC00-\uDD23\uDE80-\uDEA9\uDEAD-\uDF45\uDF51-\uDF81\uDF86-\uDFF6]|\uD83A[\uDC00-\uDCCF\uDD00-\uDD43\uDD4B-\uDFFF]|\uD83B[\uDC00-\uDEBB])/}function iv(s){return A2||(A2=Hde()),A2.test(s)}const Vde=/^[\t\n\r\x20-\x7E]*$/;function q0(s){return Vde.test(s)}const iQ=/[\u2028\u2029]/;function nQ(s){return iQ.test(s)}function gh(s){return s>=11904&&s<=55215||s>=63744&&s<=64255||s>=65281&&s<=65374}function Z6(s){return s>=127462&&s<=127487||s===8986||s===8987||s===9200||s===9203||s>=9728&&s<=10175||s===11088||s===11093||s>=127744&&s<=128591||s>=128640&&s<=128764||s>=128992&&s<=129008||s>=129280&&s<=129535||s>=129648&&s<=129782}const zde="\uFEFF";function X6(s){return!!(s&&s.length>0&&s.charCodeAt(0)===65279)}function $de(s,e=!1){return s?(e&&(s=s.replace(/\\./g,"")),s.toLowerCase()!==s):!1}function sQ(s){return s=s%52,s<26?String.fromCharCode(97+s):String.fromCharCode(65+s-26)}function Ez(s,e){return s===0?e!==5&&e!==7:s===2&&e===3?!1:s===4||s===2||s===3||e===4||e===2||e===3?!0:!(s===8&&(e===8||e===9||e===11||e===12)||(s===11||s===9)&&(e===9||e===10)||(s===12||s===10)&&e===10||e===5||e===13||e===7||s===1||s===13&&e===14||s===6&&e===6)}const If=class If{static getInstance(){return If._INSTANCE||(If._INSTANCE=new If),If._INSTANCE}constructor(){this._data=Ude()}getGraphemeBreakType(e){if(e<32)return e===10?3:e===13?2:4;if(e<127)return 0;const t=this._data,i=t.length/3;let n=1;for(;n<=i;)if(e<t[3*n])n=2*n;else if(e>t[3*n+1])n=2*n+1;else return t[3*n+2];return 0}};If._INSTANCE=null;let pE=If;function Ude(){return JSON.parse("[0,0,0,51229,51255,12,44061,44087,12,127462,127487,6,7083,7085,5,47645,47671,12,54813,54839,12,128678,128678,14,3270,3270,5,9919,9923,14,45853,45879,12,49437,49463,12,53021,53047,12,71216,71218,7,128398,128399,14,129360,129374,14,2519,2519,5,4448,4519,9,9742,9742,14,12336,12336,14,44957,44983,12,46749,46775,12,48541,48567,12,50333,50359,12,52125,52151,12,53917,53943,12,69888,69890,5,73018,73018,5,127990,127990,14,128558,128559,14,128759,128760,14,129653,129655,14,2027,2035,5,2891,2892,7,3761,3761,5,6683,6683,5,8293,8293,4,9825,9826,14,9999,9999,14,43452,43453,5,44509,44535,12,45405,45431,12,46301,46327,12,47197,47223,12,48093,48119,12,48989,49015,12,49885,49911,12,50781,50807,12,51677,51703,12,52573,52599,12,53469,53495,12,54365,54391,12,65279,65279,4,70471,70472,7,72145,72147,7,119173,119179,5,127799,127818,14,128240,128244,14,128512,128512,14,128652,128652,14,128721,128722,14,129292,129292,14,129445,129450,14,129734,129743,14,1476,1477,5,2366,2368,7,2750,2752,7,3076,3076,5,3415,3415,5,4141,4144,5,6109,6109,5,6964,6964,5,7394,7400,5,9197,9198,14,9770,9770,14,9877,9877,14,9968,9969,14,10084,10084,14,43052,43052,5,43713,43713,5,44285,44311,12,44733,44759,12,45181,45207,12,45629,45655,12,46077,46103,12,46525,46551,12,46973,46999,12,47421,47447,12,47869,47895,12,48317,48343,12,48765,48791,12,49213,49239,12,49661,49687,12,50109,50135,12,50557,50583,12,51005,51031,12,51453,51479,12,51901,51927,12,52349,52375,12,52797,52823,12,53245,53271,12,53693,53719,12,54141,54167,12,54589,54615,12,55037,55063,12,69506,69509,5,70191,70193,5,70841,70841,7,71463,71467,5,72330,72342,5,94031,94031,5,123628,123631,5,127763,127765,14,127941,127941,14,128043,128062,14,128302,128317,14,128465,128467,14,128539,128539,14,128640,128640,14,128662,128662,14,128703,128703,14,128745,128745,14,129004,129007,14,129329,129330,14,129402,129402,14,129483,129483,14,129686,129704,14,130048,131069,14,173,173,4,1757,1757,1,2200,2207,5,2434,2435,7,2631,2632,5,2817,2817,5,3008,3008,5,3201,3201,5,3387,3388,5,3542,3542,5,3902,3903,7,4190,4192,5,6002,6003,5,6439,6440,5,6765,6770,7,7019,7027,5,7154,7155,7,8205,8205,13,8505,8505,14,9654,9654,14,9757,9757,14,9792,9792,14,9852,9853,14,9890,9894,14,9937,9937,14,9981,9981,14,10035,10036,14,11035,11036,14,42654,42655,5,43346,43347,7,43587,43587,5,44006,44007,7,44173,44199,12,44397,44423,12,44621,44647,12,44845,44871,12,45069,45095,12,45293,45319,12,45517,45543,12,45741,45767,12,45965,45991,12,46189,46215,12,46413,46439,12,46637,46663,12,46861,46887,12,47085,47111,12,47309,47335,12,47533,47559,12,47757,47783,12,47981,48007,12,48205,48231,12,48429,48455,12,48653,48679,12,48877,48903,12,49101,49127,12,49325,49351,12,49549,49575,12,49773,49799,12,49997,50023,12,50221,50247,12,50445,50471,12,50669,50695,12,50893,50919,12,51117,51143,12,51341,51367,12,51565,51591,12,51789,51815,12,52013,52039,12,52237,52263,12,52461,52487,12,52685,52711,12,52909,52935,12,53133,53159,12,53357,53383,12,53581,53607,12,53805,53831,12,54029,54055,12,54253,54279,12,54477,54503,12,54701,54727,12,54925,54951,12,55149,55175,12,68101,68102,5,69762,69762,7,70067,70069,7,70371,70378,5,70720,70721,7,71087,71087,5,71341,71341,5,71995,71996,5,72249,72249,7,72850,72871,5,73109,73109,5,118576,118598,5,121505,121519,5,127245,127247,14,127568,127569,14,127777,127777,14,127872,127891,14,127956,127967,14,128015,128016,14,128110,128172,14,128259,128259,14,128367,128368,14,128424,128424,14,128488,128488,14,128530,128532,14,128550,128551,14,128566,128566,14,128647,128647,14,128656,128656,14,128667,128673,14,128691,128693,14,128715,128715,14,128728,128732,14,128752,128752,14,128765,128767,14,129096,129103,14,129311,129311,14,129344,129349,14,129394,129394,14,129413,129425,14,129466,129471,14,129511,129535,14,129664,129666,14,129719,129722,14,129760,129767,14,917536,917631,5,13,13,2,1160,1161,5,1564,1564,4,1807,1807,1,2085,2087,5,2307,2307,7,2382,2383,7,2497,2500,5,2563,2563,7,2677,2677,5,2763,2764,7,2879,2879,5,2914,2915,5,3021,3021,5,3142,3144,5,3263,3263,5,3285,3286,5,3398,3400,7,3530,3530,5,3633,3633,5,3864,3865,5,3974,3975,5,4155,4156,7,4229,4230,5,5909,5909,7,6078,6085,7,6277,6278,5,6451,6456,7,6744,6750,5,6846,6846,5,6972,6972,5,7074,7077,5,7146,7148,7,7222,7223,5,7416,7417,5,8234,8238,4,8417,8417,5,9000,9000,14,9203,9203,14,9730,9731,14,9748,9749,14,9762,9763,14,9776,9783,14,9800,9811,14,9831,9831,14,9872,9873,14,9882,9882,14,9900,9903,14,9929,9933,14,9941,9960,14,9974,9974,14,9989,9989,14,10006,10006,14,10062,10062,14,10160,10160,14,11647,11647,5,12953,12953,14,43019,43019,5,43232,43249,5,43443,43443,5,43567,43568,7,43696,43696,5,43765,43765,7,44013,44013,5,44117,44143,12,44229,44255,12,44341,44367,12,44453,44479,12,44565,44591,12,44677,44703,12,44789,44815,12,44901,44927,12,45013,45039,12,45125,45151,12,45237,45263,12,45349,45375,12,45461,45487,12,45573,45599,12,45685,45711,12,45797,45823,12,45909,45935,12,46021,46047,12,46133,46159,12,46245,46271,12,46357,46383,12,46469,46495,12,46581,46607,12,46693,46719,12,46805,46831,12,46917,46943,12,47029,47055,12,47141,47167,12,47253,47279,12,47365,47391,12,47477,47503,12,47589,47615,12,47701,47727,12,47813,47839,12,47925,47951,12,48037,48063,12,48149,48175,12,48261,48287,12,48373,48399,12,48485,48511,12,48597,48623,12,48709,48735,12,48821,48847,12,48933,48959,12,49045,49071,12,49157,49183,12,49269,49295,12,49381,49407,12,49493,49519,12,49605,49631,12,49717,49743,12,49829,49855,12,49941,49967,12,50053,50079,12,50165,50191,12,50277,50303,12,50389,50415,12,50501,50527,12,50613,50639,12,50725,50751,12,50837,50863,12,50949,50975,12,51061,51087,12,51173,51199,12,51285,51311,12,51397,51423,12,51509,51535,12,51621,51647,12,51733,51759,12,51845,51871,12,51957,51983,12,52069,52095,12,52181,52207,12,52293,52319,12,52405,52431,12,52517,52543,12,52629,52655,12,52741,52767,12,52853,52879,12,52965,52991,12,53077,53103,12,53189,53215,12,53301,53327,12,53413,53439,12,53525,53551,12,53637,53663,12,53749,53775,12,53861,53887,12,53973,53999,12,54085,54111,12,54197,54223,12,54309,54335,12,54421,54447,12,54533,54559,12,54645,54671,12,54757,54783,12,54869,54895,12,54981,55007,12,55093,55119,12,55243,55291,10,66045,66045,5,68325,68326,5,69688,69702,5,69817,69818,5,69957,69958,7,70089,70092,5,70198,70199,5,70462,70462,5,70502,70508,5,70750,70750,5,70846,70846,7,71100,71101,5,71230,71230,7,71351,71351,5,71737,71738,5,72000,72000,7,72160,72160,5,72273,72278,5,72752,72758,5,72882,72883,5,73031,73031,5,73461,73462,7,94192,94193,7,119149,119149,7,121403,121452,5,122915,122916,5,126980,126980,14,127358,127359,14,127535,127535,14,127759,127759,14,127771,127771,14,127792,127793,14,127825,127867,14,127897,127899,14,127945,127945,14,127985,127986,14,128000,128007,14,128021,128021,14,128066,128100,14,128184,128235,14,128249,128252,14,128266,128276,14,128335,128335,14,128379,128390,14,128407,128419,14,128444,128444,14,128481,128481,14,128499,128499,14,128526,128526,14,128536,128536,14,128543,128543,14,128556,128556,14,128564,128564,14,128577,128580,14,128643,128645,14,128649,128649,14,128654,128654,14,128660,128660,14,128664,128664,14,128675,128675,14,128686,128689,14,128695,128696,14,128705,128709,14,128717,128719,14,128725,128725,14,128736,128741,14,128747,128748,14,128755,128755,14,128762,128762,14,128981,128991,14,129009,129023,14,129160,129167,14,129296,129304,14,129320,129327,14,129340,129342,14,129356,129356,14,129388,129392,14,129399,129400,14,129404,129407,14,129432,129442,14,129454,129455,14,129473,129474,14,129485,129487,14,129648,129651,14,129659,129660,14,129671,129679,14,129709,129711,14,129728,129730,14,129751,129753,14,129776,129782,14,917505,917505,4,917760,917999,5,10,10,3,127,159,4,768,879,5,1471,1471,5,1536,1541,1,1648,1648,5,1767,1768,5,1840,1866,5,2070,2073,5,2137,2139,5,2274,2274,1,2363,2363,7,2377,2380,7,2402,2403,5,2494,2494,5,2507,2508,7,2558,2558,5,2622,2624,7,2641,2641,5,2691,2691,7,2759,2760,5,2786,2787,5,2876,2876,5,2881,2884,5,2901,2902,5,3006,3006,5,3014,3016,7,3072,3072,5,3134,3136,5,3157,3158,5,3260,3260,5,3266,3266,5,3274,3275,7,3328,3329,5,3391,3392,7,3405,3405,5,3457,3457,5,3536,3537,7,3551,3551,5,3636,3642,5,3764,3772,5,3895,3895,5,3967,3967,7,3993,4028,5,4146,4151,5,4182,4183,7,4226,4226,5,4253,4253,5,4957,4959,5,5940,5940,7,6070,6070,7,6087,6088,7,6158,6158,4,6432,6434,5,6448,6449,7,6679,6680,5,6742,6742,5,6754,6754,5,6783,6783,5,6912,6915,5,6966,6970,5,6978,6978,5,7042,7042,7,7080,7081,5,7143,7143,7,7150,7150,7,7212,7219,5,7380,7392,5,7412,7412,5,8203,8203,4,8232,8232,4,8265,8265,14,8400,8412,5,8421,8432,5,8617,8618,14,9167,9167,14,9200,9200,14,9410,9410,14,9723,9726,14,9733,9733,14,9745,9745,14,9752,9752,14,9760,9760,14,9766,9766,14,9774,9774,14,9786,9786,14,9794,9794,14,9823,9823,14,9828,9828,14,9833,9850,14,9855,9855,14,9875,9875,14,9880,9880,14,9885,9887,14,9896,9897,14,9906,9916,14,9926,9927,14,9935,9935,14,9939,9939,14,9962,9962,14,9972,9972,14,9978,9978,14,9986,9986,14,9997,9997,14,10002,10002,14,10017,10017,14,10055,10055,14,10071,10071,14,10133,10135,14,10548,10549,14,11093,11093,14,12330,12333,5,12441,12442,5,42608,42610,5,43010,43010,5,43045,43046,5,43188,43203,7,43302,43309,5,43392,43394,5,43446,43449,5,43493,43493,5,43571,43572,7,43597,43597,7,43703,43704,5,43756,43757,5,44003,44004,7,44009,44010,7,44033,44059,12,44089,44115,12,44145,44171,12,44201,44227,12,44257,44283,12,44313,44339,12,44369,44395,12,44425,44451,12,44481,44507,12,44537,44563,12,44593,44619,12,44649,44675,12,44705,44731,12,44761,44787,12,44817,44843,12,44873,44899,12,44929,44955,12,44985,45011,12,45041,45067,12,45097,45123,12,45153,45179,12,45209,45235,12,45265,45291,12,45321,45347,12,45377,45403,12,45433,45459,12,45489,45515,12,45545,45571,12,45601,45627,12,45657,45683,12,45713,45739,12,45769,45795,12,45825,45851,12,45881,45907,12,45937,45963,12,45993,46019,12,46049,46075,12,46105,46131,12,46161,46187,12,46217,46243,12,46273,46299,12,46329,46355,12,46385,46411,12,46441,46467,12,46497,46523,12,46553,46579,12,46609,46635,12,46665,46691,12,46721,46747,12,46777,46803,12,46833,46859,12,46889,46915,12,46945,46971,12,47001,47027,12,47057,47083,12,47113,47139,12,47169,47195,12,47225,47251,12,47281,47307,12,47337,47363,12,47393,47419,12,47449,47475,12,47505,47531,12,47561,47587,12,47617,47643,12,47673,47699,12,47729,47755,12,47785,47811,12,47841,47867,12,47897,47923,12,47953,47979,12,48009,48035,12,48065,48091,12,48121,48147,12,48177,48203,12,48233,48259,12,48289,48315,12,48345,48371,12,48401,48427,12,48457,48483,12,48513,48539,12,48569,48595,12,48625,48651,12,48681,48707,12,48737,48763,12,48793,48819,12,48849,48875,12,48905,48931,12,48961,48987,12,49017,49043,12,49073,49099,12,49129,49155,12,49185,49211,12,49241,49267,12,49297,49323,12,49353,49379,12,49409,49435,12,49465,49491,12,49521,49547,12,49577,49603,12,49633,49659,12,49689,49715,12,49745,49771,12,49801,49827,12,49857,49883,12,49913,49939,12,49969,49995,12,50025,50051,12,50081,50107,12,50137,50163,12,50193,50219,12,50249,50275,12,50305,50331,12,50361,50387,12,50417,50443,12,50473,50499,12,50529,50555,12,50585,50611,12,50641,50667,12,50697,50723,12,50753,50779,12,50809,50835,12,50865,50891,12,50921,50947,12,50977,51003,12,51033,51059,12,51089,51115,12,51145,51171,12,51201,51227,12,51257,51283,12,51313,51339,12,51369,51395,12,51425,51451,12,51481,51507,12,51537,51563,12,51593,51619,12,51649,51675,12,51705,51731,12,51761,51787,12,51817,51843,12,51873,51899,12,51929,51955,12,51985,52011,12,52041,52067,12,52097,52123,12,52153,52179,12,52209,52235,12,52265,52291,12,52321,52347,12,52377,52403,12,52433,52459,12,52489,52515,12,52545,52571,12,52601,52627,12,52657,52683,12,52713,52739,12,52769,52795,12,52825,52851,12,52881,52907,12,52937,52963,12,52993,53019,12,53049,53075,12,53105,53131,12,53161,53187,12,53217,53243,12,53273,53299,12,53329,53355,12,53385,53411,12,53441,53467,12,53497,53523,12,53553,53579,12,53609,53635,12,53665,53691,12,53721,53747,12,53777,53803,12,53833,53859,12,53889,53915,12,53945,53971,12,54001,54027,12,54057,54083,12,54113,54139,12,54169,54195,12,54225,54251,12,54281,54307,12,54337,54363,12,54393,54419,12,54449,54475,12,54505,54531,12,54561,54587,12,54617,54643,12,54673,54699,12,54729,54755,12,54785,54811,12,54841,54867,12,54897,54923,12,54953,54979,12,55009,55035,12,55065,55091,12,55121,55147,12,55177,55203,12,65024,65039,5,65520,65528,4,66422,66426,5,68152,68154,5,69291,69292,5,69633,69633,5,69747,69748,5,69811,69814,5,69826,69826,5,69932,69932,7,70016,70017,5,70079,70080,7,70095,70095,5,70196,70196,5,70367,70367,5,70402,70403,7,70464,70464,5,70487,70487,5,70709,70711,7,70725,70725,7,70833,70834,7,70843,70844,7,70849,70849,7,71090,71093,5,71103,71104,5,71227,71228,7,71339,71339,5,71344,71349,5,71458,71461,5,71727,71735,5,71985,71989,7,71998,71998,5,72002,72002,7,72154,72155,5,72193,72202,5,72251,72254,5,72281,72283,5,72344,72345,5,72766,72766,7,72874,72880,5,72885,72886,5,73023,73029,5,73104,73105,5,73111,73111,5,92912,92916,5,94095,94098,5,113824,113827,4,119142,119142,7,119155,119162,4,119362,119364,5,121476,121476,5,122888,122904,5,123184,123190,5,125252,125258,5,127183,127183,14,127340,127343,14,127377,127386,14,127491,127503,14,127548,127551,14,127744,127756,14,127761,127761,14,127769,127769,14,127773,127774,14,127780,127788,14,127796,127797,14,127820,127823,14,127869,127869,14,127894,127895,14,127902,127903,14,127943,127943,14,127947,127950,14,127972,127972,14,127988,127988,14,127992,127994,14,128009,128011,14,128019,128019,14,128023,128041,14,128064,128064,14,128102,128107,14,128174,128181,14,128238,128238,14,128246,128247,14,128254,128254,14,128264,128264,14,128278,128299,14,128329,128330,14,128348,128359,14,128371,128377,14,128392,128393,14,128401,128404,14,128421,128421,14,128433,128434,14,128450,128452,14,128476,128478,14,128483,128483,14,128495,128495,14,128506,128506,14,128519,128520,14,128528,128528,14,128534,128534,14,128538,128538,14,128540,128542,14,128544,128549,14,128552,128555,14,128557,128557,14,128560,128563,14,128565,128565,14,128567,128576,14,128581,128591,14,128641,128642,14,128646,128646,14,128648,128648,14,128650,128651,14,128653,128653,14,128655,128655,14,128657,128659,14,128661,128661,14,128663,128663,14,128665,128666,14,128674,128674,14,128676,128677,14,128679,128685,14,128690,128690,14,128694,128694,14,128697,128702,14,128704,128704,14,128710,128714,14,128716,128716,14,128720,128720,14,128723,128724,14,128726,128727,14,128733,128735,14,128742,128744,14,128746,128746,14,128749,128751,14,128753,128754,14,128756,128758,14,128761,128761,14,128763,128764,14,128884,128895,14,128992,129003,14,129008,129008,14,129036,129039,14,129114,129119,14,129198,129279,14,129293,129295,14,129305,129310,14,129312,129319,14,129328,129328,14,129331,129338,14,129343,129343,14,129351,129355,14,129357,129359,14,129375,129387,14,129393,129393,14,129395,129398,14,129401,129401,14,129403,129403,14,129408,129412,14,129426,129431,14,129443,129444,14,129451,129453,14,129456,129465,14,129472,129472,14,129475,129482,14,129484,129484,14,129488,129510,14,129536,129647,14,129652,129652,14,129656,129658,14,129661,129663,14,129667,129670,14,129680,129685,14,129705,129708,14,129712,129718,14,129723,129727,14,129731,129733,14,129744,129750,14,129754,129759,14,129768,129775,14,129783,129791,14,917504,917504,4,917506,917535,4,917632,917759,4,918000,921599,4,0,9,4,11,12,4,14,31,4,169,169,14,174,174,14,1155,1159,5,1425,1469,5,1473,1474,5,1479,1479,5,1552,1562,5,1611,1631,5,1750,1756,5,1759,1764,5,1770,1773,5,1809,1809,5,1958,1968,5,2045,2045,5,2075,2083,5,2089,2093,5,2192,2193,1,2250,2273,5,2275,2306,5,2362,2362,5,2364,2364,5,2369,2376,5,2381,2381,5,2385,2391,5,2433,2433,5,2492,2492,5,2495,2496,7,2503,2504,7,2509,2509,5,2530,2531,5,2561,2562,5,2620,2620,5,2625,2626,5,2635,2637,5,2672,2673,5,2689,2690,5,2748,2748,5,2753,2757,5,2761,2761,7,2765,2765,5,2810,2815,5,2818,2819,7,2878,2878,5,2880,2880,7,2887,2888,7,2893,2893,5,2903,2903,5,2946,2946,5,3007,3007,7,3009,3010,7,3018,3020,7,3031,3031,5,3073,3075,7,3132,3132,5,3137,3140,7,3146,3149,5,3170,3171,5,3202,3203,7,3262,3262,7,3264,3265,7,3267,3268,7,3271,3272,7,3276,3277,5,3298,3299,5,3330,3331,7,3390,3390,5,3393,3396,5,3402,3404,7,3406,3406,1,3426,3427,5,3458,3459,7,3535,3535,5,3538,3540,5,3544,3550,7,3570,3571,7,3635,3635,7,3655,3662,5,3763,3763,7,3784,3789,5,3893,3893,5,3897,3897,5,3953,3966,5,3968,3972,5,3981,3991,5,4038,4038,5,4145,4145,7,4153,4154,5,4157,4158,5,4184,4185,5,4209,4212,5,4228,4228,7,4237,4237,5,4352,4447,8,4520,4607,10,5906,5908,5,5938,5939,5,5970,5971,5,6068,6069,5,6071,6077,5,6086,6086,5,6089,6099,5,6155,6157,5,6159,6159,5,6313,6313,5,6435,6438,7,6441,6443,7,6450,6450,5,6457,6459,5,6681,6682,7,6741,6741,7,6743,6743,7,6752,6752,5,6757,6764,5,6771,6780,5,6832,6845,5,6847,6862,5,6916,6916,7,6965,6965,5,6971,6971,7,6973,6977,7,6979,6980,7,7040,7041,5,7073,7073,7,7078,7079,7,7082,7082,7,7142,7142,5,7144,7145,5,7149,7149,5,7151,7153,5,7204,7211,7,7220,7221,7,7376,7378,5,7393,7393,7,7405,7405,5,7415,7415,7,7616,7679,5,8204,8204,5,8206,8207,4,8233,8233,4,8252,8252,14,8288,8292,4,8294,8303,4,8413,8416,5,8418,8420,5,8482,8482,14,8596,8601,14,8986,8987,14,9096,9096,14,9193,9196,14,9199,9199,14,9201,9202,14,9208,9210,14,9642,9643,14,9664,9664,14,9728,9729,14,9732,9732,14,9735,9741,14,9743,9744,14,9746,9746,14,9750,9751,14,9753,9756,14,9758,9759,14,9761,9761,14,9764,9765,14,9767,9769,14,9771,9773,14,9775,9775,14,9784,9785,14,9787,9791,14,9793,9793,14,9795,9799,14,9812,9822,14,9824,9824,14,9827,9827,14,9829,9830,14,9832,9832,14,9851,9851,14,9854,9854,14,9856,9861,14,9874,9874,14,9876,9876,14,9878,9879,14,9881,9881,14,9883,9884,14,9888,9889,14,9895,9895,14,9898,9899,14,9904,9905,14,9917,9918,14,9924,9925,14,9928,9928,14,9934,9934,14,9936,9936,14,9938,9938,14,9940,9940,14,9961,9961,14,9963,9967,14,9970,9971,14,9973,9973,14,9975,9977,14,9979,9980,14,9982,9985,14,9987,9988,14,9992,9996,14,9998,9998,14,10000,10001,14,10004,10004,14,10013,10013,14,10024,10024,14,10052,10052,14,10060,10060,14,10067,10069,14,10083,10083,14,10085,10087,14,10145,10145,14,10175,10175,14,11013,11015,14,11088,11088,14,11503,11505,5,11744,11775,5,12334,12335,5,12349,12349,14,12951,12951,14,42607,42607,5,42612,42621,5,42736,42737,5,43014,43014,5,43043,43044,7,43047,43047,7,43136,43137,7,43204,43205,5,43263,43263,5,43335,43345,5,43360,43388,8,43395,43395,7,43444,43445,7,43450,43451,7,43454,43456,7,43561,43566,5,43569,43570,5,43573,43574,5,43596,43596,5,43644,43644,5,43698,43700,5,43710,43711,5,43755,43755,7,43758,43759,7,43766,43766,5,44005,44005,5,44008,44008,5,44012,44012,7,44032,44032,11,44060,44060,11,44088,44088,11,44116,44116,11,44144,44144,11,44172,44172,11,44200,44200,11,44228,44228,11,44256,44256,11,44284,44284,11,44312,44312,11,44340,44340,11,44368,44368,11,44396,44396,11,44424,44424,11,44452,44452,11,44480,44480,11,44508,44508,11,44536,44536,11,44564,44564,11,44592,44592,11,44620,44620,11,44648,44648,11,44676,44676,11,44704,44704,11,44732,44732,11,44760,44760,11,44788,44788,11,44816,44816,11,44844,44844,11,44872,44872,11,44900,44900,11,44928,44928,11,44956,44956,11,44984,44984,11,45012,45012,11,45040,45040,11,45068,45068,11,45096,45096,11,45124,45124,11,45152,45152,11,45180,45180,11,45208,45208,11,45236,45236,11,45264,45264,11,45292,45292,11,45320,45320,11,45348,45348,11,45376,45376,11,45404,45404,11,45432,45432,11,45460,45460,11,45488,45488,11,45516,45516,11,45544,45544,11,45572,45572,11,45600,45600,11,45628,45628,11,45656,45656,11,45684,45684,11,45712,45712,11,45740,45740,11,45768,45768,11,45796,45796,11,45824,45824,11,45852,45852,11,45880,45880,11,45908,45908,11,45936,45936,11,45964,45964,11,45992,45992,11,46020,46020,11,46048,46048,11,46076,46076,11,46104,46104,11,46132,46132,11,46160,46160,11,46188,46188,11,46216,46216,11,46244,46244,11,46272,46272,11,46300,46300,11,46328,46328,11,46356,46356,11,46384,46384,11,46412,46412,11,46440,46440,11,46468,46468,11,46496,46496,11,46524,46524,11,46552,46552,11,46580,46580,11,46608,46608,11,46636,46636,11,46664,46664,11,46692,46692,11,46720,46720,11,46748,46748,11,46776,46776,11,46804,46804,11,46832,46832,11,46860,46860,11,46888,46888,11,46916,46916,11,46944,46944,11,46972,46972,11,47000,47000,11,47028,47028,11,47056,47056,11,47084,47084,11,47112,47112,11,47140,47140,11,47168,47168,11,47196,47196,11,47224,47224,11,47252,47252,11,47280,47280,11,47308,47308,11,47336,47336,11,47364,47364,11,47392,47392,11,47420,47420,11,47448,47448,11,47476,47476,11,47504,47504,11,47532,47532,11,47560,47560,11,47588,47588,11,47616,47616,11,47644,47644,11,47672,47672,11,47700,47700,11,47728,47728,11,47756,47756,11,47784,47784,11,47812,47812,11,47840,47840,11,47868,47868,11,47896,47896,11,47924,47924,11,47952,47952,11,47980,47980,11,48008,48008,11,48036,48036,11,48064,48064,11,48092,48092,11,48120,48120,11,48148,48148,11,48176,48176,11,48204,48204,11,48232,48232,11,48260,48260,11,48288,48288,11,48316,48316,11,48344,48344,11,48372,48372,11,48400,48400,11,48428,48428,11,48456,48456,11,48484,48484,11,48512,48512,11,48540,48540,11,48568,48568,11,48596,48596,11,48624,48624,11,48652,48652,11,48680,48680,11,48708,48708,11,48736,48736,11,48764,48764,11,48792,48792,11,48820,48820,11,48848,48848,11,48876,48876,11,48904,48904,11,48932,48932,11,48960,48960,11,48988,48988,11,49016,49016,11,49044,49044,11,49072,49072,11,49100,49100,11,49128,49128,11,49156,49156,11,49184,49184,11,49212,49212,11,49240,49240,11,49268,49268,11,49296,49296,11,49324,49324,11,49352,49352,11,49380,49380,11,49408,49408,11,49436,49436,11,49464,49464,11,49492,49492,11,49520,49520,11,49548,49548,11,49576,49576,11,49604,49604,11,49632,49632,11,49660,49660,11,49688,49688,11,49716,49716,11,49744,49744,11,49772,49772,11,49800,49800,11,49828,49828,11,49856,49856,11,49884,49884,11,49912,49912,11,49940,49940,11,49968,49968,11,49996,49996,11,50024,50024,11,50052,50052,11,50080,50080,11,50108,50108,11,50136,50136,11,50164,50164,11,50192,50192,11,50220,50220,11,50248,50248,11,50276,50276,11,50304,50304,11,50332,50332,11,50360,50360,11,50388,50388,11,50416,50416,11,50444,50444,11,50472,50472,11,50500,50500,11,50528,50528,11,50556,50556,11,50584,50584,11,50612,50612,11,50640,50640,11,50668,50668,11,50696,50696,11,50724,50724,11,50752,50752,11,50780,50780,11,50808,50808,11,50836,50836,11,50864,50864,11,50892,50892,11,50920,50920,11,50948,50948,11,50976,50976,11,51004,51004,11,51032,51032,11,51060,51060,11,51088,51088,11,51116,51116,11,51144,51144,11,51172,51172,11,51200,51200,11,51228,51228,11,51256,51256,11,51284,51284,11,51312,51312,11,51340,51340,11,51368,51368,11,51396,51396,11,51424,51424,11,51452,51452,11,51480,51480,11,51508,51508,11,51536,51536,11,51564,51564,11,51592,51592,11,51620,51620,11,51648,51648,11,51676,51676,11,51704,51704,11,51732,51732,11,51760,51760,11,51788,51788,11,51816,51816,11,51844,51844,11,51872,51872,11,51900,51900,11,51928,51928,11,51956,51956,11,51984,51984,11,52012,52012,11,52040,52040,11,52068,52068,11,52096,52096,11,52124,52124,11,52152,52152,11,52180,52180,11,52208,52208,11,52236,52236,11,52264,52264,11,52292,52292,11,52320,52320,11,52348,52348,11,52376,52376,11,52404,52404,11,52432,52432,11,52460,52460,11,52488,52488,11,52516,52516,11,52544,52544,11,52572,52572,11,52600,52600,11,52628,52628,11,52656,52656,11,52684,52684,11,52712,52712,11,52740,52740,11,52768,52768,11,52796,52796,11,52824,52824,11,52852,52852,11,52880,52880,11,52908,52908,11,52936,52936,11,52964,52964,11,52992,52992,11,53020,53020,11,53048,53048,11,53076,53076,11,53104,53104,11,53132,53132,11,53160,53160,11,53188,53188,11,53216,53216,11,53244,53244,11,53272,53272,11,53300,53300,11,53328,53328,11,53356,53356,11,53384,53384,11,53412,53412,11,53440,53440,11,53468,53468,11,53496,53496,11,53524,53524,11,53552,53552,11,53580,53580,11,53608,53608,11,53636,53636,11,53664,53664,11,53692,53692,11,53720,53720,11,53748,53748,11,53776,53776,11,53804,53804,11,53832,53832,11,53860,53860,11,53888,53888,11,53916,53916,11,53944,53944,11,53972,53972,11,54000,54000,11,54028,54028,11,54056,54056,11,54084,54084,11,54112,54112,11,54140,54140,11,54168,54168,11,54196,54196,11,54224,54224,11,54252,54252,11,54280,54280,11,54308,54308,11,54336,54336,11,54364,54364,11,54392,54392,11,54420,54420,11,54448,54448,11,54476,54476,11,54504,54504,11,54532,54532,11,54560,54560,11,54588,54588,11,54616,54616,11,54644,54644,11,54672,54672,11,54700,54700,11,54728,54728,11,54756,54756,11,54784,54784,11,54812,54812,11,54840,54840,11,54868,54868,11,54896,54896,11,54924,54924,11,54952,54952,11,54980,54980,11,55008,55008,11,55036,55036,11,55064,55064,11,55092,55092,11,55120,55120,11,55148,55148,11,55176,55176,11,55216,55238,9,64286,64286,5,65056,65071,5,65438,65439,5,65529,65531,4,66272,66272,5,68097,68099,5,68108,68111,5,68159,68159,5,68900,68903,5,69446,69456,5,69632,69632,7,69634,69634,7,69744,69744,5,69759,69761,5,69808,69810,7,69815,69816,7,69821,69821,1,69837,69837,1,69927,69931,5,69933,69940,5,70003,70003,5,70018,70018,7,70070,70078,5,70082,70083,1,70094,70094,7,70188,70190,7,70194,70195,7,70197,70197,7,70206,70206,5,70368,70370,7,70400,70401,5,70459,70460,5,70463,70463,7,70465,70468,7,70475,70477,7,70498,70499,7,70512,70516,5,70712,70719,5,70722,70724,5,70726,70726,5,70832,70832,5,70835,70840,5,70842,70842,5,70845,70845,5,70847,70848,5,70850,70851,5,71088,71089,7,71096,71099,7,71102,71102,7,71132,71133,5,71219,71226,5,71229,71229,5,71231,71232,5,71340,71340,7,71342,71343,7,71350,71350,7,71453,71455,5,71462,71462,7,71724,71726,7,71736,71736,7,71984,71984,5,71991,71992,7,71997,71997,7,71999,71999,1,72001,72001,1,72003,72003,5,72148,72151,5,72156,72159,7,72164,72164,7,72243,72248,5,72250,72250,1,72263,72263,5,72279,72280,7,72324,72329,1,72343,72343,7,72751,72751,7,72760,72765,5,72767,72767,5,72873,72873,7,72881,72881,7,72884,72884,7,73009,73014,5,73020,73021,5,73030,73030,1,73098,73102,7,73107,73108,7,73110,73110,7,73459,73460,5,78896,78904,4,92976,92982,5,94033,94087,7,94180,94180,5,113821,113822,5,118528,118573,5,119141,119141,5,119143,119145,5,119150,119154,5,119163,119170,5,119210,119213,5,121344,121398,5,121461,121461,5,121499,121503,5,122880,122886,5,122907,122913,5,122918,122922,5,123566,123566,5,125136,125142,5,126976,126979,14,126981,127182,14,127184,127231,14,127279,127279,14,127344,127345,14,127374,127374,14,127405,127461,14,127489,127490,14,127514,127514,14,127538,127546,14,127561,127567,14,127570,127743,14,127757,127758,14,127760,127760,14,127762,127762,14,127766,127768,14,127770,127770,14,127772,127772,14,127775,127776,14,127778,127779,14,127789,127791,14,127794,127795,14,127798,127798,14,127819,127819,14,127824,127824,14,127868,127868,14,127870,127871,14,127892,127893,14,127896,127896,14,127900,127901,14,127904,127940,14,127942,127942,14,127944,127944,14,127946,127946,14,127951,127955,14,127968,127971,14,127973,127984,14,127987,127987,14,127989,127989,14,127991,127991,14,127995,127999,5,128008,128008,14,128012,128014,14,128017,128018,14,128020,128020,14,128022,128022,14,128042,128042,14,128063,128063,14,128065,128065,14,128101,128101,14,128108,128109,14,128173,128173,14,128182,128183,14,128236,128237,14,128239,128239,14,128245,128245,14,128248,128248,14,128253,128253,14,128255,128258,14,128260,128263,14,128265,128265,14,128277,128277,14,128300,128301,14,128326,128328,14,128331,128334,14,128336,128347,14,128360,128366,14,128369,128370,14,128378,128378,14,128391,128391,14,128394,128397,14,128400,128400,14,128405,128406,14,128420,128420,14,128422,128423,14,128425,128432,14,128435,128443,14,128445,128449,14,128453,128464,14,128468,128475,14,128479,128480,14,128482,128482,14,128484,128487,14,128489,128494,14,128496,128498,14,128500,128505,14,128507,128511,14,128513,128518,14,128521,128525,14,128527,128527,14,128529,128529,14,128533,128533,14,128535,128535,14,128537,128537,14]")}function qde(s,e){if(s===0)return 0;const t=jde(s,e);if(t!==void 0)return t;const i=new K6(e,s);return i.prevCodePoint(),i.offset}function jde(s,e){const t=new K6(e,s);let i=t.prevCodePoint();for(;Gde(i)||i===65039||i===8419;){if(t.offset===0)return;i=t.prevCodePoint()}if(!Z6(i))return;let n=t.offset;return n>0&&t.prevCodePoint()===8205&&(n=t.offset),n}function Gde(s){return 127995<=s&&s<=127999}const oQ=" ",Zl=class Zl{static getInstance(e){return Zl.cache.get(Array.from(e))}static getLocales(){return Zl._locales.value}constructor(e){this.confusableDictionary=e}isAmbiguous(e){return this.confusableDictionary.has(e)}getPrimaryConfusable(e){return this.confusableDictionary.get(e)}getConfusableCodePoints(){return new Set(this.confusableDictionary.keys())}};Zl.ambiguousCharacterData=new es(()=>JSON.parse('{"_common":[8232,32,8233,32,5760,32,8192,32,8193,32,8194,32,8195,32,8196,32,8197,32,8198,32,8200,32,8201,32,8202,32,8287,32,8199,32,8239,32,2042,95,65101,95,65102,95,65103,95,8208,45,8209,45,8210,45,65112,45,1748,45,8259,45,727,45,8722,45,10134,45,11450,45,1549,44,1643,44,184,44,42233,44,894,59,2307,58,2691,58,1417,58,1795,58,1796,58,5868,58,65072,58,6147,58,6153,58,8282,58,1475,58,760,58,42889,58,8758,58,720,58,42237,58,451,33,11601,33,660,63,577,63,2429,63,5038,63,42731,63,119149,46,8228,46,1793,46,1794,46,42510,46,68176,46,1632,46,1776,46,42232,46,1373,96,65287,96,8219,96,1523,96,8242,96,1370,96,8175,96,65344,96,900,96,8189,96,8125,96,8127,96,8190,96,697,96,884,96,712,96,714,96,715,96,756,96,699,96,701,96,700,96,702,96,42892,96,1497,96,2036,96,2037,96,5194,96,5836,96,94033,96,94034,96,65339,91,10088,40,10098,40,12308,40,64830,40,65341,93,10089,41,10099,41,12309,41,64831,41,10100,123,119060,123,10101,125,65342,94,8270,42,1645,42,8727,42,66335,42,5941,47,8257,47,8725,47,8260,47,9585,47,10187,47,10744,47,119354,47,12755,47,12339,47,11462,47,20031,47,12035,47,65340,92,65128,92,8726,92,10189,92,10741,92,10745,92,119311,92,119355,92,12756,92,20022,92,12034,92,42872,38,708,94,710,94,5869,43,10133,43,66203,43,8249,60,10094,60,706,60,119350,60,5176,60,5810,60,5120,61,11840,61,12448,61,42239,61,8250,62,10095,62,707,62,119351,62,5171,62,94015,62,8275,126,732,126,8128,126,8764,126,65372,124,65293,45,118002,50,120784,50,120794,50,120804,50,120814,50,120824,50,130034,50,42842,50,423,50,1000,50,42564,50,5311,50,42735,50,119302,51,118003,51,120785,51,120795,51,120805,51,120815,51,120825,51,130035,51,42923,51,540,51,439,51,42858,51,11468,51,1248,51,94011,51,71882,51,118004,52,120786,52,120796,52,120806,52,120816,52,120826,52,130036,52,5070,52,71855,52,118005,53,120787,53,120797,53,120807,53,120817,53,120827,53,130037,53,444,53,71867,53,118006,54,120788,54,120798,54,120808,54,120818,54,120828,54,130038,54,11474,54,5102,54,71893,54,119314,55,118007,55,120789,55,120799,55,120809,55,120819,55,120829,55,130039,55,66770,55,71878,55,2819,56,2538,56,2666,56,125131,56,118008,56,120790,56,120800,56,120810,56,120820,56,120830,56,130040,56,547,56,546,56,66330,56,2663,57,2920,57,2541,57,3437,57,118009,57,120791,57,120801,57,120811,57,120821,57,120831,57,130041,57,42862,57,11466,57,71884,57,71852,57,71894,57,9082,97,65345,97,119834,97,119886,97,119938,97,119990,97,120042,97,120094,97,120146,97,120198,97,120250,97,120302,97,120354,97,120406,97,120458,97,593,97,945,97,120514,97,120572,97,120630,97,120688,97,120746,97,65313,65,117974,65,119808,65,119860,65,119912,65,119964,65,120016,65,120068,65,120120,65,120172,65,120224,65,120276,65,120328,65,120380,65,120432,65,913,65,120488,65,120546,65,120604,65,120662,65,120720,65,5034,65,5573,65,42222,65,94016,65,66208,65,119835,98,119887,98,119939,98,119991,98,120043,98,120095,98,120147,98,120199,98,120251,98,120303,98,120355,98,120407,98,120459,98,388,98,5071,98,5234,98,5551,98,65314,66,8492,66,117975,66,119809,66,119861,66,119913,66,120017,66,120069,66,120121,66,120173,66,120225,66,120277,66,120329,66,120381,66,120433,66,42932,66,914,66,120489,66,120547,66,120605,66,120663,66,120721,66,5108,66,5623,66,42192,66,66178,66,66209,66,66305,66,65347,99,8573,99,119836,99,119888,99,119940,99,119992,99,120044,99,120096,99,120148,99,120200,99,120252,99,120304,99,120356,99,120408,99,120460,99,7428,99,1010,99,11429,99,43951,99,66621,99,128844,67,71913,67,71922,67,65315,67,8557,67,8450,67,8493,67,117976,67,119810,67,119862,67,119914,67,119966,67,120018,67,120174,67,120226,67,120278,67,120330,67,120382,67,120434,67,1017,67,11428,67,5087,67,42202,67,66210,67,66306,67,66581,67,66844,67,8574,100,8518,100,119837,100,119889,100,119941,100,119993,100,120045,100,120097,100,120149,100,120201,100,120253,100,120305,100,120357,100,120409,100,120461,100,1281,100,5095,100,5231,100,42194,100,8558,68,8517,68,117977,68,119811,68,119863,68,119915,68,119967,68,120019,68,120071,68,120123,68,120175,68,120227,68,120279,68,120331,68,120383,68,120435,68,5024,68,5598,68,5610,68,42195,68,8494,101,65349,101,8495,101,8519,101,119838,101,119890,101,119942,101,120046,101,120098,101,120150,101,120202,101,120254,101,120306,101,120358,101,120410,101,120462,101,43826,101,1213,101,8959,69,65317,69,8496,69,117978,69,119812,69,119864,69,119916,69,120020,69,120072,69,120124,69,120176,69,120228,69,120280,69,120332,69,120384,69,120436,69,917,69,120492,69,120550,69,120608,69,120666,69,120724,69,11577,69,5036,69,42224,69,71846,69,71854,69,66182,69,119839,102,119891,102,119943,102,119995,102,120047,102,120099,102,120151,102,120203,102,120255,102,120307,102,120359,102,120411,102,120463,102,43829,102,42905,102,383,102,7837,102,1412,102,119315,70,8497,70,117979,70,119813,70,119865,70,119917,70,120021,70,120073,70,120125,70,120177,70,120229,70,120281,70,120333,70,120385,70,120437,70,42904,70,988,70,120778,70,5556,70,42205,70,71874,70,71842,70,66183,70,66213,70,66853,70,65351,103,8458,103,119840,103,119892,103,119944,103,120048,103,120100,103,120152,103,120204,103,120256,103,120308,103,120360,103,120412,103,120464,103,609,103,7555,103,397,103,1409,103,117980,71,119814,71,119866,71,119918,71,119970,71,120022,71,120074,71,120126,71,120178,71,120230,71,120282,71,120334,71,120386,71,120438,71,1292,71,5056,71,5107,71,42198,71,65352,104,8462,104,119841,104,119945,104,119997,104,120049,104,120101,104,120153,104,120205,104,120257,104,120309,104,120361,104,120413,104,120465,104,1211,104,1392,104,5058,104,65320,72,8459,72,8460,72,8461,72,117981,72,119815,72,119867,72,119919,72,120023,72,120179,72,120231,72,120283,72,120335,72,120387,72,120439,72,919,72,120494,72,120552,72,120610,72,120668,72,120726,72,11406,72,5051,72,5500,72,42215,72,66255,72,731,105,9075,105,65353,105,8560,105,8505,105,8520,105,119842,105,119894,105,119946,105,119998,105,120050,105,120102,105,120154,105,120206,105,120258,105,120310,105,120362,105,120414,105,120466,105,120484,105,618,105,617,105,953,105,8126,105,890,105,120522,105,120580,105,120638,105,120696,105,120754,105,1110,105,42567,105,1231,105,43893,105,5029,105,71875,105,65354,106,8521,106,119843,106,119895,106,119947,106,119999,106,120051,106,120103,106,120155,106,120207,106,120259,106,120311,106,120363,106,120415,106,120467,106,1011,106,1112,106,65322,74,117983,74,119817,74,119869,74,119921,74,119973,74,120025,74,120077,74,120129,74,120181,74,120233,74,120285,74,120337,74,120389,74,120441,74,42930,74,895,74,1032,74,5035,74,5261,74,42201,74,119844,107,119896,107,119948,107,120000,107,120052,107,120104,107,120156,107,120208,107,120260,107,120312,107,120364,107,120416,107,120468,107,8490,75,65323,75,117984,75,119818,75,119870,75,119922,75,119974,75,120026,75,120078,75,120130,75,120182,75,120234,75,120286,75,120338,75,120390,75,120442,75,922,75,120497,75,120555,75,120613,75,120671,75,120729,75,11412,75,5094,75,5845,75,42199,75,66840,75,1472,108,8739,73,9213,73,65512,73,1633,108,1777,73,66336,108,125127,108,118001,108,120783,73,120793,73,120803,73,120813,73,120823,73,130033,73,65321,73,8544,73,8464,73,8465,73,117982,108,119816,73,119868,73,119920,73,120024,73,120128,73,120180,73,120232,73,120284,73,120336,73,120388,73,120440,73,65356,108,8572,73,8467,108,119845,108,119897,108,119949,108,120001,108,120053,108,120105,73,120157,73,120209,73,120261,73,120313,73,120365,73,120417,73,120469,73,448,73,120496,73,120554,73,120612,73,120670,73,120728,73,11410,73,1030,73,1216,73,1493,108,1503,108,1575,108,126464,108,126592,108,65166,108,65165,108,1994,108,11599,73,5825,73,42226,73,93992,73,66186,124,66313,124,119338,76,8556,76,8466,76,117985,76,119819,76,119871,76,119923,76,120027,76,120079,76,120131,76,120183,76,120235,76,120287,76,120339,76,120391,76,120443,76,11472,76,5086,76,5290,76,42209,76,93974,76,71843,76,71858,76,66587,76,66854,76,65325,77,8559,77,8499,77,117986,77,119820,77,119872,77,119924,77,120028,77,120080,77,120132,77,120184,77,120236,77,120288,77,120340,77,120392,77,120444,77,924,77,120499,77,120557,77,120615,77,120673,77,120731,77,1018,77,11416,77,5047,77,5616,77,5846,77,42207,77,66224,77,66321,77,119847,110,119899,110,119951,110,120003,110,120055,110,120107,110,120159,110,120211,110,120263,110,120315,110,120367,110,120419,110,120471,110,1400,110,1404,110,65326,78,8469,78,117987,78,119821,78,119873,78,119925,78,119977,78,120029,78,120081,78,120185,78,120237,78,120289,78,120341,78,120393,78,120445,78,925,78,120500,78,120558,78,120616,78,120674,78,120732,78,11418,78,42208,78,66835,78,3074,111,3202,111,3330,111,3458,111,2406,111,2662,111,2790,111,3046,111,3174,111,3302,111,3430,111,3664,111,3792,111,4160,111,1637,111,1781,111,65359,111,8500,111,119848,111,119900,111,119952,111,120056,111,120108,111,120160,111,120212,111,120264,111,120316,111,120368,111,120420,111,120472,111,7439,111,7441,111,43837,111,959,111,120528,111,120586,111,120644,111,120702,111,120760,111,963,111,120532,111,120590,111,120648,111,120706,111,120764,111,11423,111,4351,111,1413,111,1505,111,1607,111,126500,111,126564,111,126596,111,65259,111,65260,111,65258,111,65257,111,1726,111,64428,111,64429,111,64427,111,64426,111,1729,111,64424,111,64425,111,64423,111,64422,111,1749,111,3360,111,4125,111,66794,111,71880,111,71895,111,66604,111,1984,79,2534,79,2918,79,12295,79,70864,79,71904,79,118000,79,120782,79,120792,79,120802,79,120812,79,120822,79,130032,79,65327,79,117988,79,119822,79,119874,79,119926,79,119978,79,120030,79,120082,79,120134,79,120186,79,120238,79,120290,79,120342,79,120394,79,120446,79,927,79,120502,79,120560,79,120618,79,120676,79,120734,79,11422,79,1365,79,11604,79,4816,79,2848,79,66754,79,42227,79,71861,79,66194,79,66219,79,66564,79,66838,79,9076,112,65360,112,119849,112,119901,112,119953,112,120005,112,120057,112,120109,112,120161,112,120213,112,120265,112,120317,112,120369,112,120421,112,120473,112,961,112,120530,112,120544,112,120588,112,120602,112,120646,112,120660,112,120704,112,120718,112,120762,112,120776,112,11427,112,65328,80,8473,80,117989,80,119823,80,119875,80,119927,80,119979,80,120031,80,120083,80,120187,80,120239,80,120291,80,120343,80,120395,80,120447,80,929,80,120504,80,120562,80,120620,80,120678,80,120736,80,11426,80,5090,80,5229,80,42193,80,66197,80,119850,113,119902,113,119954,113,120006,113,120058,113,120110,113,120162,113,120214,113,120266,113,120318,113,120370,113,120422,113,120474,113,1307,113,1379,113,1382,113,8474,81,117990,81,119824,81,119876,81,119928,81,119980,81,120032,81,120084,81,120188,81,120240,81,120292,81,120344,81,120396,81,120448,81,11605,81,119851,114,119903,114,119955,114,120007,114,120059,114,120111,114,120163,114,120215,114,120267,114,120319,114,120371,114,120423,114,120475,114,43847,114,43848,114,7462,114,11397,114,43905,114,119318,82,8475,82,8476,82,8477,82,117991,82,119825,82,119877,82,119929,82,120033,82,120189,82,120241,82,120293,82,120345,82,120397,82,120449,82,422,82,5025,82,5074,82,66740,82,5511,82,42211,82,94005,82,65363,115,119852,115,119904,115,119956,115,120008,115,120060,115,120112,115,120164,115,120216,115,120268,115,120320,115,120372,115,120424,115,120476,115,42801,115,445,115,1109,115,43946,115,71873,115,66632,115,65331,83,117992,83,119826,83,119878,83,119930,83,119982,83,120034,83,120086,83,120138,83,120190,83,120242,83,120294,83,120346,83,120398,83,120450,83,1029,83,1359,83,5077,83,5082,83,42210,83,94010,83,66198,83,66592,83,119853,116,119905,116,119957,116,120009,116,120061,116,120113,116,120165,116,120217,116,120269,116,120321,116,120373,116,120425,116,120477,116,8868,84,10201,84,128872,84,65332,84,117993,84,119827,84,119879,84,119931,84,119983,84,120035,84,120087,84,120139,84,120191,84,120243,84,120295,84,120347,84,120399,84,120451,84,932,84,120507,84,120565,84,120623,84,120681,84,120739,84,11430,84,5026,84,42196,84,93962,84,71868,84,66199,84,66225,84,66325,84,119854,117,119906,117,119958,117,120010,117,120062,117,120114,117,120166,117,120218,117,120270,117,120322,117,120374,117,120426,117,120478,117,42911,117,7452,117,43854,117,43858,117,651,117,965,117,120534,117,120592,117,120650,117,120708,117,120766,117,1405,117,66806,117,71896,117,8746,85,8899,85,117994,85,119828,85,119880,85,119932,85,119984,85,120036,85,120088,85,120140,85,120192,85,120244,85,120296,85,120348,85,120400,85,120452,85,1357,85,4608,85,66766,85,5196,85,42228,85,94018,85,71864,85,8744,118,8897,118,65366,118,8564,118,119855,118,119907,118,119959,118,120011,118,120063,118,120115,118,120167,118,120219,118,120271,118,120323,118,120375,118,120427,118,120479,118,7456,118,957,118,120526,118,120584,118,120642,118,120700,118,120758,118,1141,118,1496,118,71430,118,43945,118,71872,118,119309,86,1639,86,1783,86,8548,86,117995,86,119829,86,119881,86,119933,86,119985,86,120037,86,120089,86,120141,86,120193,86,120245,86,120297,86,120349,86,120401,86,120453,86,1140,86,11576,86,5081,86,5167,86,42719,86,42214,86,93960,86,71840,86,66845,86,623,119,119856,119,119908,119,119960,119,120012,119,120064,119,120116,119,120168,119,120220,119,120272,119,120324,119,120376,119,120428,119,120480,119,7457,119,1121,119,1309,119,1377,119,71434,119,71438,119,71439,119,43907,119,71910,87,71919,87,117996,87,119830,87,119882,87,119934,87,119986,87,120038,87,120090,87,120142,87,120194,87,120246,87,120298,87,120350,87,120402,87,120454,87,1308,87,5043,87,5076,87,42218,87,5742,120,10539,120,10540,120,10799,120,65368,120,8569,120,119857,120,119909,120,119961,120,120013,120,120065,120,120117,120,120169,120,120221,120,120273,120,120325,120,120377,120,120429,120,120481,120,5441,120,5501,120,5741,88,9587,88,66338,88,71916,88,65336,88,8553,88,117997,88,119831,88,119883,88,119935,88,119987,88,120039,88,120091,88,120143,88,120195,88,120247,88,120299,88,120351,88,120403,88,120455,88,42931,88,935,88,120510,88,120568,88,120626,88,120684,88,120742,88,11436,88,11613,88,5815,88,42219,88,66192,88,66228,88,66327,88,66855,88,611,121,7564,121,65369,121,119858,121,119910,121,119962,121,120014,121,120066,121,120118,121,120170,121,120222,121,120274,121,120326,121,120378,121,120430,121,120482,121,655,121,7935,121,43866,121,947,121,8509,121,120516,121,120574,121,120632,121,120690,121,120748,121,1199,121,4327,121,71900,121,65337,89,117998,89,119832,89,119884,89,119936,89,119988,89,120040,89,120092,89,120144,89,120196,89,120248,89,120300,89,120352,89,120404,89,120456,89,933,89,978,89,120508,89,120566,89,120624,89,120682,89,120740,89,11432,89,1198,89,5033,89,5053,89,42220,89,94019,89,71844,89,66226,89,119859,122,119911,122,119963,122,120015,122,120067,122,120119,122,120171,122,120223,122,120275,122,120327,122,120379,122,120431,122,120483,122,7458,122,43923,122,71876,122,71909,90,66293,90,65338,90,8484,90,8488,90,117999,90,119833,90,119885,90,119937,90,119989,90,120041,90,120197,90,120249,90,120301,90,120353,90,120405,90,120457,90,918,90,120493,90,120551,90,120609,90,120667,90,120725,90,5059,90,42204,90,71849,90,65282,34,65283,35,65284,36,65285,37,65286,38,65290,42,65291,43,65294,46,65295,47,65296,48,65298,50,65299,51,65300,52,65301,53,65302,54,65303,55,65304,56,65305,57,65308,60,65309,61,65310,62,65312,64,65316,68,65318,70,65319,71,65324,76,65329,81,65330,82,65333,85,65334,86,65335,87,65343,95,65346,98,65348,100,65350,102,65355,107,65357,109,65358,110,65361,113,65362,114,65364,116,65365,117,65367,119,65370,122,65371,123,65373,125,119846,109],"_default":[160,32,8211,45,65374,126,8218,44,65306,58,65281,33,8216,96,8217,96,8245,96,180,96,12494,47,1047,51,1073,54,1072,97,1040,65,1068,98,1042,66,1089,99,1057,67,1077,101,1045,69,1053,72,305,105,1050,75,921,73,1052,77,1086,111,1054,79,1009,112,1088,112,1056,80,1075,114,1058,84,215,120,1093,120,1061,88,1091,121,1059,89,65288,40,65289,41,65292,44,65297,49,65307,59,65311,63],"cs":[65374,126,8218,44,65306,58,65281,33,8216,96,8245,96,180,96,12494,47,1047,51,1073,54,1072,97,1040,65,1068,98,1042,66,1089,99,1057,67,1077,101,1045,69,1053,72,305,105,1050,75,921,73,1052,77,1086,111,1054,79,1009,112,1088,112,1056,80,1075,114,1058,84,1093,120,1061,88,1091,121,1059,89,65288,40,65289,41,65292,44,65297,49,65307,59,65311,63],"de":[65374,126,65306,58,65281,33,8245,96,180,96,12494,47,1047,51,1073,54,1072,97,1040,65,1068,98,1042,66,1089,99,1057,67,1077,101,1045,69,1053,72,305,105,1050,75,921,73,1052,77,1086,111,1054,79,1009,112,1088,112,1056,80,1075,114,1058,84,1093,120,1061,88,1091,121,1059,89,65288,40,65289,41,65292,44,65297,49,65307,59,65311,63],"es":[8211,45,65374,126,8218,44,65306,58,65281,33,8245,96,180,96,12494,47,1047,51,1073,54,1072,97,1040,65,1068,98,1042,66,1089,99,1057,67,1077,101,1045,69,1053,72,305,105,1050,75,1052,77,1086,111,1054,79,1009,112,1088,112,1056,80,1075,114,1058,84,215,120,1093,120,1061,88,1091,121,1059,89,65288,40,65289,41,65292,44,65297,49,65307,59,65311,63],"fr":[65374,126,8218,44,65306,58,65281,33,8216,96,8245,96,12494,47,1047,51,1073,54,1072,97,1040,65,1068,98,1042,66,1089,99,1057,67,1077,101,1045,69,1053,72,305,105,1050,75,921,73,1052,77,1086,111,1054,79,1009,112,1088,112,1056,80,1075,114,1058,84,215,120,1093,120,1061,88,1091,121,1059,89,65288,40,65289,41,65292,44,65297,49,65307,59,65311,63],"it":[160,32,8211,45,65374,126,8218,44,65306,58,65281,33,8245,96,180,96,12494,47,1047,51,1073,54,1072,97,1040,65,1068,98,1042,66,1089,99,1057,67,1077,101,1045,69,1053,72,305,105,1050,75,921,73,1052,77,1086,111,1054,79,1009,112,1088,112,1056,80,1075,114,1058,84,215,120,1093,120,1061,88,1091,121,1059,89,65288,40,65289,41,65292,44,65297,49,65307,59,65311,63],"ja":[8211,45,8218,44,65281,33,8216,96,8245,96,180,96,1047,51,1073,54,1072,97,1040,65,1068,98,1042,66,1089,99,1057,67,1077,101,1045,69,1053,72,305,105,1050,75,921,73,1052,77,1086,111,1054,79,1009,112,1088,112,1056,80,1075,114,1058,84,215,120,1093,120,1061,88,1091,121,1059,89,65292,44,65297,49,65307,59],"ko":[8211,45,65374,126,8218,44,65306,58,65281,33,8245,96,180,96,12494,47,1047,51,1073,54,1072,97,1040,65,1068,98,1042,66,1089,99,1057,67,1077,101,1045,69,1053,72,305,105,1050,75,921,73,1052,77,1086,111,1054,79,1009,112,1088,112,1056,80,1075,114,1058,84,215,120,1093,120,1061,88,1091,121,1059,89,65288,40,65289,41,65292,44,65297,49,65307,59,65311,63],"pl":[65374,126,65306,58,65281,33,8216,96,8245,96,180,96,12494,47,1047,51,1073,54,1072,97,1040,65,1068,98,1042,66,1089,99,1057,67,1077,101,1045,69,1053,72,305,105,1050,75,921,73,1052,77,1086,111,1054,79,1009,112,1088,112,1056,80,1075,114,1058,84,215,120,1093,120,1061,88,1091,121,1059,89,65288,40,65289,41,65292,44,65297,49,65307,59,65311,63],"pt-BR":[65374,126,8218,44,65306,58,65281,33,8216,96,8245,96,180,96,12494,47,1047,51,1073,54,1072,97,1040,65,1068,98,1042,66,1089,99,1057,67,1077,101,1045,69,1053,72,305,105,1050,75,921,73,1052,77,1086,111,1054,79,1009,112,1088,112,1056,80,1075,114,1058,84,215,120,1093,120,1061,88,1091,121,1059,89,65288,40,65289,41,65292,44,65297,49,65307,59,65311,63],"qps-ploc":[160,32,8211,45,65374,126,8218,44,65306,58,65281,33,8216,96,8245,96,180,96,12494,47,1047,51,1073,54,1072,97,1040,65,1068,98,1042,66,1089,99,1057,67,1077,101,1045,69,1053,72,305,105,1050,75,921,73,1052,77,1086,111,1054,79,1088,112,1056,80,1075,114,1058,84,215,120,1093,120,1061,88,1091,121,1059,89,65288,40,65289,41,65292,44,65297,49,65307,59,65311,63],"ru":[65374,126,8218,44,65306,58,65281,33,8216,96,8245,96,180,96,12494,47,305,105,921,73,1009,112,215,120,65288,40,65289,41,65292,44,65297,49,65307,59,65311,63],"tr":[160,32,8211,45,65374,126,8218,44,65306,58,65281,33,8245,96,180,96,12494,47,1047,51,1073,54,1072,97,1040,65,1068,98,1042,66,1089,99,1057,67,1077,101,1045,69,1053,72,1050,75,921,73,1052,77,1086,111,1054,79,1009,112,1088,112,1056,80,1075,114,1058,84,215,120,1093,120,1061,88,1091,121,1059,89,65288,40,65289,41,65292,44,65297,49,65307,59,65311,63],"zh-hans":[160,32,65374,126,8218,44,8245,96,180,96,12494,47,1047,51,1073,54,1072,97,1040,65,1068,98,1042,66,1089,99,1057,67,1077,101,1045,69,1053,72,305,105,1050,75,921,73,1052,77,1086,111,1054,79,1009,112,1088,112,1056,80,1075,114,1058,84,215,120,1093,120,1061,88,1091,121,1059,89,65297,49],"zh-hant":[8211,45,65374,126,8218,44,180,96,12494,47,1047,51,1073,54,1072,97,1040,65,1068,98,1042,66,1089,99,1057,67,1077,101,1045,69,1053,72,305,105,1050,75,921,73,1052,77,1086,111,1054,79,1009,112,1088,112,1056,80,1075,114,1058,84,215,120,1093,120,1061,88,1091,121,1059,89]}')),Zl.cache=new Rde({getCacheKey:JSON.stringify},e=>{function t(d){const h=new Map;for(let u=0;u<d.length;u+=2)h.set(d[u],d[u+1]);return h}function i(d,h){const u=new Map(d);for(const[g,f]of h)u.set(g,f);return u}function n(d,h){if(!d)return h;const u=new Map;for(const[g,f]of d)h.has(g)&&u.set(g,f);return u}const o=Zl.ambiguousCharacterData.value;let r=e.filter(d=>!d.startsWith("_")&&Object.hasOwn(o,d));r.length===0&&(r=["_default"]);let a;for(const d of r){const h=t(o[d]);a=n(a,h)}const l=t(o._common),c=i(l,a);return new Zl(c)}),Zl._locales=new es(()=>Object.keys(Zl.ambiguousCharacterData.value).filter(e=>!e.startsWith("_")));let jy=Zl;const qb=class qb{static getRawData(){return JSON.parse('{"_common":[11,12,13,127,847,1564,4447,4448,6068,6069,6155,6156,6157,6158,7355,7356,8192,8193,8194,8195,8196,8197,8198,8199,8200,8201,8202,8204,8205,8206,8207,8234,8235,8236,8237,8238,8239,8287,8288,8289,8290,8291,8292,8293,8294,8295,8296,8297,8298,8299,8300,8301,8302,8303,10240,12644,65024,65025,65026,65027,65028,65029,65030,65031,65032,65033,65034,65035,65036,65037,65038,65039,65279,65440,65520,65521,65522,65523,65524,65525,65526,65527,65528,65532,78844,119155,119156,119157,119158,119159,119160,119161,119162,917504,917505,917506,917507,917508,917509,917510,917511,917512,917513,917514,917515,917516,917517,917518,917519,917520,917521,917522,917523,917524,917525,917526,917527,917528,917529,917530,917531,917532,917533,917534,917535,917536,917537,917538,917539,917540,917541,917542,917543,917544,917545,917546,917547,917548,917549,917550,917551,917552,917553,917554,917555,917556,917557,917558,917559,917560,917561,917562,917563,917564,917565,917566,917567,917568,917569,917570,917571,917572,917573,917574,917575,917576,917577,917578,917579,917580,917581,917582,917583,917584,917585,917586,917587,917588,917589,917590,917591,917592,917593,917594,917595,917596,917597,917598,917599,917600,917601,917602,917603,917604,917605,917606,917607,917608,917609,917610,917611,917612,917613,917614,917615,917616,917617,917618,917619,917620,917621,917622,917623,917624,917625,917626,917627,917628,917629,917630,917631,917760,917761,917762,917763,917764,917765,917766,917767,917768,917769,917770,917771,917772,917773,917774,917775,917776,917777,917778,917779,917780,917781,917782,917783,917784,917785,917786,917787,917788,917789,917790,917791,917792,917793,917794,917795,917796,917797,917798,917799,917800,917801,917802,917803,917804,917805,917806,917807,917808,917809,917810,917811,917812,917813,917814,917815,917816,917817,917818,917819,917820,917821,917822,917823,917824,917825,917826,917827,917828,917829,917830,917831,917832,917833,917834,917835,917836,917837,917838,917839,917840,917841,917842,917843,917844,917845,917846,917847,917848,917849,917850,917851,917852,917853,917854,917855,917856,917857,917858,917859,917860,917861,917862,917863,917864,917865,917866,917867,917868,917869,917870,917871,917872,917873,917874,917875,917876,917877,917878,917879,917880,917881,917882,917883,917884,917885,917886,917887,917888,917889,917890,917891,917892,917893,917894,917895,917896,917897,917898,917899,917900,917901,917902,917903,917904,917905,917906,917907,917908,917909,917910,917911,917912,917913,917914,917915,917916,917917,917918,917919,917920,917921,917922,917923,917924,917925,917926,917927,917928,917929,917930,917931,917932,917933,917934,917935,917936,917937,917938,917939,917940,917941,917942,917943,917944,917945,917946,917947,917948,917949,917950,917951,917952,917953,917954,917955,917956,917957,917958,917959,917960,917961,917962,917963,917964,917965,917966,917967,917968,917969,917970,917971,917972,917973,917974,917975,917976,917977,917978,917979,917980,917981,917982,917983,917984,917985,917986,917987,917988,917989,917990,917991,917992,917993,917994,917995,917996,917997,917998,917999],"cs":[173,8203,12288],"de":[173,8203,12288],"es":[8203,12288],"fr":[173,8203,12288],"it":[160,173,12288],"ja":[173],"ko":[173,12288],"pl":[173,8203,12288],"pt-BR":[173,8203,12288],"qps-ploc":[160,173,8203,12288],"ru":[173,12288],"tr":[160,173,8203,12288],"zh-hans":[160,173,8203,12288],"zh-hant":[173,12288]}')}static getData(){return this._data||(this._data=new Set([...Object.values(qb.getRawData())].flat())),this._data}static isInvisibleCharacter(e){return qb.getData().has(e)}static get codePoints(){return qb.getData()}};qb._data=void 0;let np=qb;var Iz={};let rw;const M2=globalThis.vscode;if(typeof M2<"u"&&typeof M2.process<"u"){const s=M2.process;rw={get platform(){return s.platform},get arch(){return s.arch},get env(){return s.env},cwd(){return s.cwd()}}}else typeof process<"u"&&typeof process?.versions?.node=="string"?rw={get platform(){return process.platform},get arch(){return process.arch},get env(){return Iz},cwd(){return Iz.VSCODE_CWD||process.cwd()}}:rw={get platform(){return Bn?"win32":Ye?"darwin":"linux"},get arch(){},get env(){return{}},cwd(){return"/"}};const mE=rw.cwd,_E=rw.env,Kde=rw.platform,Yde=65,Zde=97,Xde=90,Qde=122,sp=46,xs=47,ir=92,ed=58,Jde=63;class rQ extends Error{constructor(e,t,i){let n;typeof t=="string"&&t.indexOf("not ")===0?(n="must not be",t=t.replace(/^not /,"")):n="must be";const o=e.indexOf(".")!==-1?"property":"argument";let r=`The "${e}" ${o} ${n} of type ${t}`;r+=`. Received type ${typeof i}`,super(r),this.code="ERR_INVALID_ARG_TYPE"}}function ehe(s,e){if(s===null||typeof s!="object")throw new rQ(e,"Object",s)}function xn(s,e){if(typeof s!="string")throw new rQ(e,"string",s)}const kg=Kde==="win32";function Dt(s){return s===xs||s===ir}function uP(s){return s===xs}function td(s){return s>=Yde&&s<=Xde||s>=Zde&&s<=Qde}function bE(s,e,t,i){let n="",o=0,r=-1,a=0,l=0;for(let c=0;c<=s.length;++c){if(c<s.length)l=s.charCodeAt(c);else{if(i(l))break;l=xs}if(i(l)){if(!(r===c-1||a===1))if(a===2){if(n.length<2||o!==2||n.charCodeAt(n.length-1)!==sp||n.charCodeAt(n.length-2)!==sp){if(n.length>2){const d=n.lastIndexOf(t);d===-1?(n="",o=0):(n=n.slice(0,d),o=n.length-1-n.lastIndexOf(t)),r=c,a=0;continue}else if(n.length!==0){n="",o=0,r=c,a=0;continue}}e&&(n+=n.length>0?`${t}..`:"..",o=2)}else n.length>0?n+=`${t}${s.slice(r+1,c)}`:n=s.slice(r+1,c),o=c-r-1;r=c,a=0}else l===sp&&a!==-1?++a:a=-1}return n}function the(s){return s?`${s[0]==="."?"":"."}${s}`:""}function aQ(s,e){ehe(e,"pathObject");const t=e.dir||e.root,i=e.base||`${e.name||""}${the(e.ext)}`;return t?t===e.root?`${t}${i}`:`${t}${s}${i}`:i}const Bo={resolve(...s){let e="",t="",i=!1;for(let n=s.length-1;n>=-1;n--){let o;if(n>=0){if(o=s[n],xn(o,`paths[${n}]`),o.length===0)continue}else e.length===0?o=mE():(o=_E[`=${e}`]||mE(),(o===void 0||o.slice(0,2).toLowerCase()!==e.toLowerCase()&&o.charCodeAt(2)===ir)&&(o=`${e}\\`));const r=o.length;let a=0,l="",c=!1;const d=o.charCodeAt(0);if(r===1)Dt(d)&&(a=1,c=!0);else if(Dt(d))if(c=!0,Dt(o.charCodeAt(1))){let h=2,u=h;for(;h<r&&!Dt(o.charCodeAt(h));)h++;if(h<r&&h!==u){const g=o.slice(u,h);for(u=h;h<r&&Dt(o.charCodeAt(h));)h++;if(h<r&&h!==u){for(u=h;h<r&&!Dt(o.charCodeAt(h));)h++;(h===r||h!==u)&&(l=`\\\\${g}\\${o.slice(u,h)}`,a=h)}}}else a=1;else td(d)&&o.charCodeAt(1)===ed&&(l=o.slice(0,2),a=2,r>2&&Dt(o.charCodeAt(2))&&(c=!0,a=3));if(l.length>0)if(e.length>0){if(l.toLowerCase()!==e.toLowerCase())continue}else e=l;if(i){if(e.length>0)break}else if(t=`${o.slice(a)}\\${t}`,i=c,c&&e.length>0)break}return t=bE(t,!i,"\\",Dt),i?`${e}\\${t}`:`${e}${t}`||"."},normalize(s){xn(s,"path");const e=s.length;if(e===0)return".";let t=0,i,n=!1;const o=s.charCodeAt(0);if(e===1)return uP(o)?"\\":s;if(Dt(o))if(n=!0,Dt(s.charCodeAt(1))){let a=2,l=a;for(;a<e&&!Dt(s.charCodeAt(a));)a++;if(a<e&&a!==l){const c=s.slice(l,a);for(l=a;a<e&&Dt(s.charCodeAt(a));)a++;if(a<e&&a!==l){for(l=a;a<e&&!Dt(s.charCodeAt(a));)a++;if(a===e)return`\\\\${c}\\${s.slice(l)}\\`;a!==l&&(i=`\\\\${c}\\${s.slice(l,a)}`,t=a)}}}else t=1;else td(o)&&s.charCodeAt(1)===ed&&(i=s.slice(0,2),t=2,e>2&&Dt(s.charCodeAt(2))&&(n=!0,t=3));let r=t<e?bE(s.slice(t),!n,"\\",Dt):"";if(r.length===0&&!n&&(r="."),r.length>0&&Dt(s.charCodeAt(e-1))&&(r+="\\"),!n&&i===void 0&&s.includes(":")){if(r.length>=2&&td(r.charCodeAt(0))&&r.charCodeAt(1)===ed)return`.\\${r}`;let a=s.indexOf(":");do if(a===e-1||Dt(s.charCodeAt(a+1)))return`.\\${r}`;while((a=s.indexOf(":",a+1))!==-1)}return i===void 0?n?`\\${r}`:r:n?`${i}\\${r}`:`${i}${r}`},isAbsolute(s){xn(s,"path");const e=s.length;if(e===0)return!1;const t=s.charCodeAt(0);return Dt(t)||e>2&&td(t)&&s.charCodeAt(1)===ed&&Dt(s.charCodeAt(2))},join(...s){if(s.length===0)return".";let e,t;for(let o=0;o<s.length;++o){const r=s[o];xn(r,"path"),r.length>0&&(e===void 0?e=t=r:e+=`\\${r}`)}if(e===void 0)return".";let i=!0,n=0;if(typeof t=="string"&&Dt(t.charCodeAt(0))){++n;const o=t.length;o>1&&Dt(t.charCodeAt(1))&&(++n,o>2&&(Dt(t.charCodeAt(2))?++n:i=!1))}if(i){for(;n<e.length&&Dt(e.charCodeAt(n));)n++;n>=2&&(e=`\\${e.slice(n)}`)}return Bo.normalize(e)},relative(s,e){if(xn(s,"from"),xn(e,"to"),s===e)return"";const t=Bo.resolve(s),i=Bo.resolve(e);if(t===i||(s=t.toLowerCase(),e=i.toLowerCase(),s===e))return"";if(t.length!==s.length||i.length!==e.length){const f=t.split("\\"),p=i.split("\\");f[f.length-1]===""&&f.pop(),p[p.length-1]===""&&p.pop();const _=f.length,b=p.length,w=_<b?_:b;let C;for(C=0;C<w&&f[C].toLowerCase()===p[C].toLowerCase();C++);return C===0?i:C===w?b>w?p.slice(C).join("\\"):_>w?"..\\".repeat(_-1-C)+"..":"":"..\\".repeat(_-C)+p.slice(C).join("\\")}let n=0;for(;n<s.length&&s.charCodeAt(n)===ir;)n++;let o=s.length;for(;o-1>n&&s.charCodeAt(o-1)===ir;)o--;const r=o-n;let a=0;for(;a<e.length&&e.charCodeAt(a)===ir;)a++;let l=e.length;for(;l-1>a&&e.charCodeAt(l-1)===ir;)l--;const c=l-a,d=r<c?r:c;let h=-1,u=0;for(;u<d;u++){const f=s.charCodeAt(n+u);if(f!==e.charCodeAt(a+u))break;f===ir&&(h=u)}if(u!==d){if(h===-1)return i}else{if(c>d){if(e.charCodeAt(a+u)===ir)return i.slice(a+u+1);if(u===2)return i.slice(a+u)}r>d&&(s.charCodeAt(n+u)===ir?h=u:u===2&&(h=3)),h===-1&&(h=0)}let g="";for(u=n+h+1;u<=o;++u)(u===o||s.charCodeAt(u)===ir)&&(g+=g.length===0?"..":"\\..");return a+=h,g.length>0?`${g}${i.slice(a,l)}`:(i.charCodeAt(a)===ir&&++a,i.slice(a,l))},toNamespacedPath(s){if(typeof s!="string"||s.length===0)return s;const e=Bo.resolve(s);if(e.length<=2)return s;if(e.charCodeAt(0)===ir){if(e.charCodeAt(1)===ir){const t=e.charCodeAt(2);if(t!==Jde&&t!==sp)return`\\\\?\\UNC\\${e.slice(2)}`}}else if(td(e.charCodeAt(0))&&e.charCodeAt(1)===ed&&e.charCodeAt(2)===ir)return`\\\\?\\${e}`;return e},dirname(s){xn(s,"path");const e=s.length;if(e===0)return".";let t=-1,i=0;const n=s.charCodeAt(0);if(e===1)return Dt(n)?s:".";if(Dt(n)){if(t=i=1,Dt(s.charCodeAt(1))){let a=2,l=a;for(;a<e&&!Dt(s.charCodeAt(a));)a++;if(a<e&&a!==l){for(l=a;a<e&&Dt(s.charCodeAt(a));)a++;if(a<e&&a!==l){for(l=a;a<e&&!Dt(s.charCodeAt(a));)a++;if(a===e)return s;a!==l&&(t=i=a+1)}}}}else td(n)&&s.charCodeAt(1)===ed&&(t=e>2&&Dt(s.charCodeAt(2))?3:2,i=t);let o=-1,r=!0;for(let a=e-1;a>=i;--a)if(Dt(s.charCodeAt(a))){if(!r){o=a;break}}else r=!1;if(o===-1){if(t===-1)return".";o=t}return s.slice(0,o)},basename(s,e){e!==void 0&&xn(e,"suffix"),xn(s,"path");let t=0,i=-1,n=!0,o;if(s.length>=2&&td(s.charCodeAt(0))&&s.charCodeAt(1)===ed&&(t=2),e!==void 0&&e.length>0&&e.length<=s.length){if(e===s)return"";let r=e.length-1,a=-1;for(o=s.length-1;o>=t;--o){const l=s.charCodeAt(o);if(Dt(l)){if(!n){t=o+1;break}}else a===-1&&(n=!1,a=o+1),r>=0&&(l===e.charCodeAt(r)?--r===-1&&(i=o):(r=-1,i=a))}return t===i?i=a:i===-1&&(i=s.length),s.slice(t,i)}for(o=s.length-1;o>=t;--o)if(Dt(s.charCodeAt(o))){if(!n){t=o+1;break}}else i===-1&&(n=!1,i=o+1);return i===-1?"":s.slice(t,i)},extname(s){xn(s,"path");let e=0,t=-1,i=0,n=-1,o=!0,r=0;s.length>=2&&s.charCodeAt(1)===ed&&td(s.charCodeAt(0))&&(e=i=2);for(let a=s.length-1;a>=e;--a){const l=s.charCodeAt(a);if(Dt(l)){if(!o){i=a+1;break}continue}n===-1&&(o=!1,n=a+1),l===sp?t===-1?t=a:r!==1&&(r=1):t!==-1&&(r=-1)}return t===-1||n===-1||r===0||r===1&&t===n-1&&t===i+1?"":s.slice(t,n)},format:aQ.bind(null,"\\"),parse(s){xn(s,"path");const e={root:"",dir:"",base:"",ext:"",name:""};if(s.length===0)return e;const t=s.length;let i=0,n=s.charCodeAt(0);if(t===1)return Dt(n)?(e.root=e.dir=s,e):(e.base=e.name=s,e);if(Dt(n)){if(i=1,Dt(s.charCodeAt(1))){let h=2,u=h;for(;h<t&&!Dt(s.charCodeAt(h));)h++;if(h<t&&h!==u){for(u=h;h<t&&Dt(s.charCodeAt(h));)h++;if(h<t&&h!==u){for(u=h;h<t&&!Dt(s.charCodeAt(h));)h++;h===t?i=h:h!==u&&(i=h+1)}}}}else if(td(n)&&s.charCodeAt(1)===ed){if(t<=2)return e.root=e.dir=s,e;if(i=2,Dt(s.charCodeAt(2))){if(t===3)return e.root=e.dir=s,e;i=3}}i>0&&(e.root=s.slice(0,i));let o=-1,r=i,a=-1,l=!0,c=s.length-1,d=0;for(;c>=i;--c){if(n=s.charCodeAt(c),Dt(n)){if(!l){r=c+1;break}continue}a===-1&&(l=!1,a=c+1),n===sp?o===-1?o=c:d!==1&&(d=1):o!==-1&&(d=-1)}return a!==-1&&(o===-1||d===0||d===1&&o===a-1&&o===r+1?e.base=e.name=s.slice(r,a):(e.name=s.slice(r,o),e.base=s.slice(r,a),e.ext=s.slice(o,a))),r>0&&r!==i?e.dir=s.slice(0,r-1):e.dir=e.root,e},sep:"\\",delimiter:";",win32:null,posix:null},ihe=(()=>{if(kg){const s=/\\/g;return()=>{const e=mE().replace(s,"/");return e.slice(e.indexOf("/"))}}return()=>mE()})(),Zi={resolve(...s){let e="",t=!1;for(let i=s.length-1;i>=0&&!t;i--){const n=s[i];xn(n,`paths[${i}]`),n.length!==0&&(e=`${n}/${e}`,t=n.charCodeAt(0)===xs)}if(!t){const i=ihe();e=`${i}/${e}`,t=i.charCodeAt(0)===xs}return e=bE(e,!t,"/",uP),t?`/${e}`:e.length>0?e:"."},normalize(s){if(xn(s,"path"),s.length===0)return".";const e=s.charCodeAt(0)===xs,t=s.charCodeAt(s.length-1)===xs;return s=bE(s,!e,"/",uP),s.length===0?e?"/":t?"./":".":(t&&(s+="/"),e?`/${s}`:s)},isAbsolute(s){return xn(s,"path"),s.length>0&&s.charCodeAt(0)===xs},join(...s){if(s.length===0)return".";const e=[];for(let t=0;t<s.length;++t){const i=s[t];xn(i,"path"),i.length>0&&e.push(i)}return e.length===0?".":Zi.normalize(e.join("/"))},relative(s,e){if(xn(s,"from"),xn(e,"to"),s===e||(s=Zi.resolve(s),e=Zi.resolve(e),s===e))return"";const t=1,i=s.length,n=i-t,o=1,r=e.length-o,a=n<r?n:r;let l=-1,c=0;for(;c<a;c++){const h=s.charCodeAt(t+c);if(h!==e.charCodeAt(o+c))break;h===xs&&(l=c)}if(c===a)if(r>a){if(e.charCodeAt(o+c)===xs)return e.slice(o+c+1);if(c===0)return e.slice(o+c)}else n>a&&(s.charCodeAt(t+c)===xs?l=c:c===0&&(l=0));let d="";for(c=t+l+1;c<=i;++c)(c===i||s.charCodeAt(c)===xs)&&(d+=d.length===0?"..":"/..");return`${d}${e.slice(o+l)}`},toNamespacedPath(s){return s},dirname(s){if(xn(s,"path"),s.length===0)return".";const e=s.charCodeAt(0)===xs;let t=-1,i=!0;for(let n=s.length-1;n>=1;--n)if(s.charCodeAt(n)===xs){if(!i){t=n;break}}else i=!1;return t===-1?e?"/":".":e&&t===1?"//":s.slice(0,t)},basename(s,e){e!==void 0&&xn(e,"suffix"),xn(s,"path");let t=0,i=-1,n=!0,o;if(e!==void 0&&e.length>0&&e.length<=s.length){if(e===s)return"";let r=e.length-1,a=-1;for(o=s.length-1;o>=0;--o){const l=s.charCodeAt(o);if(l===xs){if(!n){t=o+1;break}}else a===-1&&(n=!1,a=o+1),r>=0&&(l===e.charCodeAt(r)?--r===-1&&(i=o):(r=-1,i=a))}return t===i?i=a:i===-1&&(i=s.length),s.slice(t,i)}for(o=s.length-1;o>=0;--o)if(s.charCodeAt(o)===xs){if(!n){t=o+1;break}}else i===-1&&(n=!1,i=o+1);return i===-1?"":s.slice(t,i)},extname(s){xn(s,"path");let e=-1,t=0,i=-1,n=!0,o=0;for(let r=s.length-1;r>=0;--r){const a=s[r];if(a==="/"){if(!n){t=r+1;break}continue}i===-1&&(n=!1,i=r+1),a==="."?e===-1?e=r:o!==1&&(o=1):e!==-1&&(o=-1)}return e===-1||i===-1||o===0||o===1&&e===i-1&&e===t+1?"":s.slice(e,i)},format:aQ.bind(null,"/"),parse(s){xn(s,"path");const e={root:"",dir:"",base:"",ext:"",name:""};if(s.length===0)return e;const t=s.charCodeAt(0)===xs;let i;t?(e.root="/",i=1):i=0;let n=-1,o=0,r=-1,a=!0,l=s.length-1,c=0;for(;l>=i;--l){const d=s.charCodeAt(l);if(d===xs){if(!a){o=l+1;break}continue}r===-1&&(a=!1,r=l+1),d===sp?n===-1?n=l:c!==1&&(c=1):n!==-1&&(c=-1)}if(r!==-1){const d=o===0&&t?1:o;n===-1||c===0||c===1&&n===r-1&&n===o+1?e.base=e.name=s.slice(d,r):(e.name=s.slice(d,n),e.base=s.slice(d,r),e.ext=s.slice(n,r))}return o>0?e.dir=s.slice(0,o-1):t&&(e.dir="/"),e},sep:"/",delimiter:":",win32:null,posix:null};Zi.win32=Bo.win32=Bo;Zi.posix=Bo.posix=Zi;const lQ=kg?Bo.normalize:Zi.normalize,nhe=kg?Bo.resolve:Zi.resolve,she=kg?Bo.relative:Zi.relative,vk=kg?Bo.dirname:Zi.dirname,th=kg?Bo.basename:Zi.basename,ohe=kg?Bo.extname:Zi.extname,bl=kg?Bo.sep:Zi.sep,rhe=/^\w[\w\d+.-]*$/,ahe=/^\//,lhe=/^\/\//;function che(s,e){if(!s.scheme&&e)throw new Error(`[UriError]: Scheme is missing: {scheme: "", authority: "${s.authority}", path: "${s.path}", query: "${s.query}", fragment: "${s.fragment}"}`);if(s.scheme&&!rhe.test(s.scheme))throw new Error("[UriError]: Scheme contains illegal characters.");if(s.path){if(s.authority){if(!ahe.test(s.path))throw new Error('[UriError]: If a URI contains an authority component, then the path component must either be empty or begin with a slash ("/") character')}else if(lhe.test(s.path))throw new Error('[UriError]: If a URI does not contain an authority component, then the path cannot begin with two slash characters ("//")')}}function dhe(s,e){return!s&&!e?"file":s}function hhe(s,e){switch(s){case"https":case"http":case"file":e?e[0]!==rl&&(e=rl+e):e=rl;break}return e}const zi="",rl="/",uhe=/^(([^:/?#]+?):)?(\/\/([^/?#]*))?([^?#]*)(\?([^#]*))?(#(.*))?/;let Le=class Ck{static isUri(e){return e instanceof Ck?!0:!e||typeof e!="object"?!1:typeof e.authority=="string"&&typeof e.fragment=="string"&&typeof e.path=="string"&&typeof e.query=="string"&&typeof e.scheme=="string"&&typeof e.fsPath=="string"&&typeof e.with=="function"&&typeof e.toString=="function"}constructor(e,t,i,n,o,r=!1){typeof e=="object"?(this.scheme=e.scheme||zi,this.authority=e.authority||zi,this.path=e.path||zi,this.query=e.query||zi,this.fragment=e.fragment||zi):(this.scheme=dhe(e,r),this.authority=t||zi,this.path=hhe(this.scheme,i||zi),this.query=n||zi,this.fragment=o||zi,che(this,r))}get fsPath(){return wE(this,!1)}with(e){if(!e)return this;let{scheme:t,authority:i,path:n,query:o,fragment:r}=e;return t===void 0?t=this.scheme:t===null&&(t=zi),i===void 0?i=this.authority:i===null&&(i=zi),n===void 0?n=this.path:n===null&&(n=zi),o===void 0?o=this.query:o===null&&(o=zi),r===void 0?r=this.fragment:r===null&&(r=zi),t===this.scheme&&i===this.authority&&n===this.path&&o===this.query&&r===this.fragment?this:new F_(t,i,n,o,r)}static parse(e,t=!1){const i=uhe.exec(e);return i?new F_(i[2]||zi,qx(i[4]||zi),qx(i[5]||zi),qx(i[7]||zi),qx(i[9]||zi),t):new F_(zi,zi,zi,zi,zi)}static file(e){let t=zi;if(Bn&&(e=e.replace(/\\/g,rl)),e[0]===rl&&e[1]===rl){const i=e.indexOf(rl,2);i===-1?(t=e.substring(2),e=rl):(t=e.substring(2,i),e=e.substring(i)||rl)}return new F_("file",t,e,zi,zi)}static from(e,t){return new F_(e.scheme,e.authority,e.path,e.query,e.fragment,t)}static joinPath(e,...t){if(!e.path)throw new Error("[UriError]: cannot call joinPath on URI without path");let i;return Bn&&e.scheme==="file"?i=Ck.file(Bo.join(wE(e,!0),...t)).path:i=Zi.join(e.path,...t),e.with({path:i})}toString(e=!1){return gP(this,e)}toJSON(){return this}static revive(e){if(e){if(e instanceof Ck)return e;{const t=new F_(e);return t._formatted=e.external??null,t._fsPath=e._sep===cQ?e.fsPath??null:null,t}}else return e}};const cQ=Bn?1:void 0;let F_=class extends Le{constructor(){super(...arguments),this._formatted=null,this._fsPath=null}get fsPath(){return this._fsPath||(this._fsPath=wE(this,!1)),this._fsPath}toString(e=!1){return e?gP(this,!0):(this._formatted||(this._formatted=gP(this,!1)),this._formatted)}toJSON(){const e={$mid:1};return this._fsPath&&(e.fsPath=this._fsPath,e._sep=cQ),this._formatted&&(e.external=this._formatted),this.path&&(e.path=this.path),this.scheme&&(e.scheme=this.scheme),this.authority&&(e.authority=this.authority),this.query&&(e.query=this.query),this.fragment&&(e.fragment=this.fragment),e}};const dQ={58:"%3A",47:"%2F",63:"%3F",35:"%23",91:"%5B",93:"%5D",64:"%40",33:"%21",36:"%24",38:"%26",39:"%27",40:"%28",41:"%29",42:"%2A",43:"%2B",44:"%2C",59:"%3B",61:"%3D",32:"%20"};function Dz(s,e,t){let i,n=-1;for(let o=0;o<s.length;o++){const r=s.charCodeAt(o);if(r>=97&&r<=122||r>=65&&r<=90||r>=48&&r<=57||r===45||r===46||r===95||r===126||e&&r===47||t&&r===91||t&&r===93||t&&r===58)n!==-1&&(i+=encodeURIComponent(s.substring(n,o)),n=-1),i!==void 0&&(i+=s.charAt(o));else{i===void 0&&(i=s.substr(0,o));const a=dQ[r];a!==void 0?(n!==-1&&(i+=encodeURIComponent(s.substring(n,o)),n=-1),i+=a):n===-1&&(n=o)}}return n!==-1&&(i+=encodeURIComponent(s.substring(n))),i!==void 0?i:s}function ghe(s){let e;for(let t=0;t<s.length;t++){const i=s.charCodeAt(t);i===35||i===63?(e===void 0&&(e=s.substr(0,t)),e+=dQ[i]):e!==void 0&&(e+=s[t])}return e!==void 0?e:s}function wE(s,e){let t;return s.authority&&s.path.length>1&&s.scheme==="file"?t=`//${s.authority}${s.path}`:s.path.charCodeAt(0)===47&&(s.path.charCodeAt(1)>=65&&s.path.charCodeAt(1)<=90||s.path.charCodeAt(1)>=97&&s.path.charCodeAt(1)<=122)&&s.path.charCodeAt(2)===58?e?t=s.path.substr(1):t=s.path[1].toLowerCase()+s.path.substr(2):t=s.path,Bn&&(t=t.replace(/\//g,"\\")),t}function gP(s,e){const t=e?ghe:Dz;let i="",{scheme:n,authority:o,path:r,query:a,fragment:l}=s;if(n&&(i+=n,i+=":"),(o||n==="file")&&(i+=rl,i+=rl),o){let c=o.indexOf("@");if(c!==-1){const d=o.substr(0,c);o=o.substr(c+1),c=d.lastIndexOf(":"),c===-1?i+=t(d,!1,!1):(i+=t(d.substr(0,c),!1,!1),i+=":",i+=t(d.substr(c+1),!1,!0)),i+="@"}o=o.toLowerCase(),c=o.lastIndexOf(":"),c===-1?i+=t(o,!1,!0):(i+=t(o.substr(0,c),!1,!0),i+=o.substr(c))}if(r){if(r.length>=3&&r.charCodeAt(0)===47&&r.charCodeAt(2)===58){const c=r.charCodeAt(1);c>=65&&c<=90&&(r=`/${String.fromCharCode(c+32)}:${r.substr(3)}`)}else if(r.length>=2&&r.charCodeAt(1)===58){const c=r.charCodeAt(0);c>=65&&c<=90&&(r=`${String.fromCharCode(c+32)}:${r.substr(2)}`)}i+=t(r,!0,!1)}return a&&(i+="?",i+=t(a,!1,!1)),l&&(i+="#",i+=e?l:Dz(l,!1,!1)),i}function hQ(s){try{return decodeURIComponent(s)}catch{return s.length>3?s.substr(0,3)+hQ(s.substr(3)):s}}const Tz=/(%[0-9A-Za-z][0-9A-Za-z])+/g;function qx(s){return s.match(Tz)?s.replace(Tz,e=>hQ(e)):s}var Te;(function(s){s.inMemory="inmemory",s.vscode="vscode",s.internal="private",s.walkThrough="walkThrough",s.walkThroughSnippet="walkThroughSnippet",s.http="http",s.https="https",s.file="file",s.mailto="mailto",s.untitled="untitled",s.data="data",s.command="command",s.vscodeRemote="vscode-remote",s.vscodeRemoteResource="vscode-remote-resource",s.vscodeManagedRemoteResource="vscode-managed-remote-resource",s.vscodeUserData="vscode-userdata",s.vscodeCustomEditor="vscode-custom-editor",s.vscodeNotebookCell="vscode-notebook-cell",s.vscodeNotebookCellMetadata="vscode-notebook-cell-metadata",s.vscodeNotebookCellMetadataDiff="vscode-notebook-cell-metadata-diff",s.vscodeNotebookCellOutput="vscode-notebook-cell-output",s.vscodeNotebookCellOutputDiff="vscode-notebook-cell-output-diff",s.vscodeNotebookMetadata="vscode-notebook-metadata",s.vscodeInteractiveInput="vscode-interactive-input",s.vscodeSettings="vscode-settings",s.vscodeWorkspaceTrust="vscode-workspace-trust",s.vscodeTerminal="vscode-terminal",s.vscodeChatCodeBlock="vscode-chat-code-block",s.vscodeChatCodeCompareBlock="vscode-chat-code-compare-block",s.vscodeChatEditor="vscode-chat-editor",s.vscodeChatInput="chatSessionInput",s.vscodeLocalChatSession="vscode-chat-session",s.webviewPanel="webview-panel",s.vscodeWebview="vscode-webview",s.extension="extension",s.vscodeFileResource="vscode-file",s.tmp="tmp",s.vsls="vsls",s.vscodeSourceControl="vscode-scm",s.commentsInput="comment",s.codeSetting="code-setting",s.outputChannel="output",s.accessibleView="accessible-view",s.chatEditingSnapshotScheme="chat-editing-snapshot-text-model",s.chatEditingModel="chat-editing-text-model",s.copilotPr="copilot-pr"})(Te||(Te={}));function kR(s,e){return Le.isUri(s)?Hf(s.scheme,e):j6(s,e+":")}function fP(s,...e){return e.some(t=>kR(s,t))}const fhe="tkn";class phe{constructor(){this._hosts=Object.create(null),this._ports=Object.create(null),this._connectionTokens=Object.create(null),this._preferredWebSchema="http",this._delegate=null,this._serverRootPath="/"}setPreferredWebSchema(e){this._preferredWebSchema=e}get _remoteResourcesPath(){return Zi.join(this._serverRootPath,Te.vscodeRemoteResource)}rewrite(e){if(this._delegate)try{return this._delegate(e)}catch(a){return Oe(a),e}const t=e.authority;let i=this._hosts[t];i&&i.indexOf(":")!==-1&&i.indexOf("[")===-1&&(i=`[${i}]`);const n=this._ports[t],o=this._connectionTokens[t];let r=`path=${encodeURIComponent(e.path)}`;return typeof o=="string"&&(r+=`&${fhe}=${encodeURIComponent(o)}`),Le.from({scheme:Vc?this._preferredWebSchema:Te.vscodeRemoteResource,authority:`${i}:${n}`,path:this._remoteResourcesPath,query:r})}}const uQ=new phe,mhe="vscode-app",Cy=class Cy{uriToBrowserUri(e){return e.scheme===Te.vscodeRemote?uQ.rewrite(e):e.scheme===Te.file&&(Sl||Yce===`${Te.vscodeFileResource}://${Cy.FALLBACK_AUTHORITY}`)?e.with({scheme:Te.vscodeFileResource,authority:e.authority||Cy.FALLBACK_AUTHORITY,query:null,fragment:null}):e}};Cy.FALLBACK_AUTHORITY=mhe;let pP=Cy;const gQ=new pP;var mP;(function(s){const e=new Map([["1",{"Cross-Origin-Opener-Policy":"same-origin"}],["2",{"Cross-Origin-Embedder-Policy":"require-corp"}],["3",{"Cross-Origin-Opener-Policy":"same-origin","Cross-Origin-Embedder-Policy":"require-corp"}]]);s.CoopAndCoep=Object.freeze(e.get("3"));const t="vscode-coi";function i(o){let r;typeof o=="string"?r=new URL(o).searchParams:o instanceof URL?r=o.searchParams:Le.isUri(o)&&(r=new URL(o.toString(!0)).searchParams);const a=r?.get(t);if(a)return e.get(a)}s.getHeadersFromQuery=i;function n(o,r,a){if(!globalThis.crossOriginIsolated)return;const l=r&&a?"3":a?"2":"1";o instanceof URLSearchParams?o.set(t,l):o[t]=l}s.addSearchParam=n})(mP||(mP={}));const Nz=typeof Buffer<"u";new es(()=>new Uint8Array(256));let O2;class Ru{static wrap(e){return Nz&&!Buffer.isBuffer(e)&&(e=Buffer.from(e.buffer,e.byteOffset,e.byteLength)),new Ru(e)}constructor(e){this.buffer=e,this.byteLength=this.buffer.byteLength}toString(){return Nz?this.buffer.toString():(O2||(O2=new TextDecoder),O2.decode(this.buffer))}}function _he(s,e){return s[e+0]<<0>>>0|s[e+1]<<8>>>0}function bhe(s,e,t){s[t+0]=e&255,e=e>>>8,s[t+1]=e&255}function tl(s,e){return s[e]*2**24+s[e+1]*2**16+s[e+2]*2**8+s[e+3]}function il(s,e,t){s[t+3]=e,e=e>>>8,s[t+2]=e,e=e>>>8,s[t+1]=e,e=e>>>8,s[t]=e}function Rz(s,e){return s[e]}function Az(s,e,t){s[t]=e}const Mz="0123456789abcdef";function whe({buffer:s}){let e="";for(let t=0;t<s.length;t++){const i=s[t];e+=Mz[i>>>4],e+=Mz[i&15]}return e}function j0(s){return ER(s,0)}function ER(s,e){switch(typeof s){case"object":return s===null?Bd(349,e):Array.isArray(s)?Che(s,e):yhe(s,e);case"string":return Q6(s,e);case"boolean":return vhe(s,e);case"number":return Bd(s,e);case"undefined":return Bd(937,e);default:return Bd(617,e)}}function Bd(s,e){return(e<<5)-e+s|0}function vhe(s,e){return Bd(s?433:863,e)}function Q6(s,e){e=Bd(149417,e);for(let t=0,i=s.length;t<i;t++)e=Bd(s.charCodeAt(t),e);return e}function Che(s,e){return e=Bd(104579,e),s.reduce((t,i)=>ER(i,t),e)}function yhe(s,e){return e=Bd(181387,e),Object.keys(s).sort().reduce((t,i)=>(t=Q6(i,t),ER(s[i],t)),e)}function P2(s,e,t=32){const i=t-e,n=~((1<<i)-1);return(s<<e|(n&s)>>>i)>>>0}function SC(s,e=32){return s instanceof ArrayBuffer?whe(Ru.wrap(new Uint8Array(s))):(s>>>0).toString(16).padStart(e/4,"0")}const HT=class HT{constructor(){this._h0=1732584193,this._h1=4023233417,this._h2=2562383102,this._h3=271733878,this._h4=3285377520,this._buff=new Uint8Array(67),this._buffDV=new DataView(this._buff.buffer),this._buffLen=0,this._totalLen=0,this._leftoverHighSurrogate=0,this._finished=!1}update(e){const t=e.length;if(t===0)return;const i=this._buff;let n=this._buffLen,o=this._leftoverHighSurrogate,r,a;for(o!==0?(r=o,a=-1,o=0):(r=e.charCodeAt(0),a=0);;){let l=r;if(sn(r))if(a+1<t){const c=e.charCodeAt(a+1);ag(c)?(a++,l=G6(r,c)):l=65533}else{o=r;break}else ag(r)&&(l=65533);if(n=this._push(i,n,l),a++,a<t)r=e.charCodeAt(a);else break}this._buffLen=n,this._leftoverHighSurrogate=o}_push(e,t,i){return i<128?e[t++]=i:i<2048?(e[t++]=192|(i&1984)>>>6,e[t++]=128|(i&63)>>>0):i<65536?(e[t++]=224|(i&61440)>>>12,e[t++]=128|(i&4032)>>>6,e[t++]=128|(i&63)>>>0):(e[t++]=240|(i&1835008)>>>18,e[t++]=128|(i&258048)>>>12,e[t++]=128|(i&4032)>>>6,e[t++]=128|(i&63)>>>0),t>=64&&(this._step(),t-=64,this._totalLen+=64,e[0]=e[64],e[1]=e[65],e[2]=e[66]),t}digest(){return this._finished||(this._finished=!0,this._leftoverHighSurrogate&&(this._leftoverHighSurrogate=0,this._buffLen=this._push(this._buff,this._buffLen,65533)),this._totalLen+=this._buffLen,this._wrapUp()),SC(this._h0)+SC(this._h1)+SC(this._h2)+SC(this._h3)+SC(this._h4)}_wrapUp(){this._buff[this._buffLen++]=128,this._buff.subarray(this._buffLen).fill(0),this._buffLen>56&&(this._step(),this._buff.fill(0));const e=8*this._totalLen;this._buffDV.setUint32(56,Math.floor(e/4294967296),!1),this._buffDV.setUint32(60,e%4294967296,!1),this._step()}_step(){const e=HT._bigBlock32,t=this._buffDV;for(let h=0;h<64;h+=4)e.setUint32(h,t.getUint32(h,!1),!1);for(let h=64;h<320;h+=4)e.setUint32(h,P2(e.getUint32(h-12,!1)^e.getUint32(h-32,!1)^e.getUint32(h-56,!1)^e.getUint32(h-64,!1),1),!1);let i=this._h0,n=this._h1,o=this._h2,r=this._h3,a=this._h4,l,c,d;for(let h=0;h<80;h++)h<20?(l=n&o|~n&r,c=1518500249):h<40?(l=n^o^r,c=1859775393):h<60?(l=n&o|n&r|o&r,c=2400959708):(l=n^o^r,c=3395469782),d=P2(i,5)+l+a+c+e.getUint32(h*4,!1)&4294967295,a=r,r=o,o=P2(n,30),n=i,i=d;this._h0=this._h0+i&4294967295,this._h1=this._h1+n&4294967295,this._h2=this._h2+o&4294967295,this._h3=this._h3+r&4294967295,this._h4=this._h4+a&4294967295}};HT._bigBlock32=new DataView(new ArrayBuffer(320));let _P=HT;function She(s){if(s.length===0)throw new Error("Invalid tail call");return[s.slice(0,s.length-1),s[s.length-1]]}function ni(s,e,t=(i,n)=>i===n){if(s===e)return!0;if(!s||!e||s.length!==e.length)return!1;for(let i=0,n=s.length;i<n;i++)if(!t(s[i],e[i]))return!1;return!0}function xhe(s,e){const t=s.length-1;e<t&&(s[e]=s[t]),s.pop()}function vE(s,e,t){return J6(s.length,i=>t(s[i],e))}function J6(s,e){let t=0,i=s-1;for(;t<=i;){const n=(t+i)/2|0,o=e(n);if(o<0)t=n+1;else if(o>0)i=n-1;else return n}return-(t+1)}function bP(s,e,t){if(s=s|0,s>=e.length)throw new TypeError("invalid index");const i=e[Math.floor(e.length*Math.random())],n=[],o=[],r=[];for(const a of e){const l=t(a,i);l<0?n.push(a):l>0?o.push(a):r.push(a)}return s<n.length?bP(s,n,t):s<n.length+r.length?r[0]:bP(s-(n.length+r.length),o,t)}function Oz(s,e){const t=[];let i;for(const n of s.slice(0).sort(e))!i||e(i[0],n)!==0?(i=[n],t.push(i)):i.push(n);return t}function*eW(s,e){let t,i;for(const n of s)i!==void 0&&e(i,n)?t.push(n):(t&&(yield t),t=[n]),i=n;t&&(yield t)}function fQ(s,e){for(let t=0;t<=s.length;t++)e(t===0?void 0:s[t-1],t===s.length?void 0:s[t])}function Lhe(s,e){for(let t=0;t<s.length;t++)e(t===0?void 0:s[t-1],s[t],t+1===s.length?void 0:s[t+1])}function Al(s){return s.filter(e=>!!e)}function Pz(s){let e=0;for(let t=0;t<s.length;t++)s[t]&&(s[e]=s[t],e+=1);s.length=e}function pQ(s){return!Array.isArray(s)||s.length===0}function Ms(s){return Array.isArray(s)&&s.length>0}function fh(s,e=t=>t){const t=new Set;return s.filter(i=>{const n=e(i);return t.has(n)?!1:(t.add(n),!0)})}function Us(s,e){let t=typeof e=="number"?s:0;typeof e=="number"?t=s:(t=0,e=s);const i=[];if(t<=e)for(let n=t;n<e;n++)i.push(n);else for(let n=t;n>e;n--)i.push(n);return i}function IR(s,e,t){const i=s.slice(0,e),n=s.slice(e);return i.concat(t,n)}function F2(s,e){const t=s.indexOf(e);t>-1&&(s.splice(t,1),s.unshift(e))}function jx(s,e){const t=s.indexOf(e);t>-1&&(s.splice(t,1),s.push(e))}function wP(s,e){for(const t of e)s.push(t)}function khe(s,e){const t=[];for(const i of s){const n=e(i);n!==void 0&&t.push(n)}return t}function tW(s){return Array.isArray(s)?s:[s]}function Ehe(s,e,t){const i=mQ(s,e),n=s.length,o=t.length;s.length=n+o;for(let r=n-1;r>=i;r--)s[r+o]=s[r];for(let r=0;r<o;r++)s[r+i]=t[r]}function CE(s,e,t,i){const n=mQ(s,e);let o=s.splice(n,t);return o===void 0&&(o=[]),Ehe(s,n,i),o}function mQ(s,e){return e<0?Math.max(e+s.length,0):Math.min(e,s.length)}var Au;(function(s){function e(o){return o<0}s.isLessThan=e;function t(o){return o<=0}s.isLessThanOrEqual=t;function i(o){return o>0}s.isGreaterThan=i;function n(o){return o===0}s.isNeitherLessOrGreaterThan=n,s.greaterThan=1,s.lessThan=-1,s.neitherLessOrGreaterThan=0})(Au||(Au={}));function ns(s,e){return(t,i)=>e(s(t),s(i))}function Ihe(...s){return(e,t)=>{for(const i of s){const n=i(e,t);if(!Au.isNeitherLessOrGreaterThan(n))return n}return Au.neitherLessOrGreaterThan}}const Wo=(s,e)=>s-e,_Q=(s,e)=>Wo(s?1:0,e?1:0);function bQ(s){return(e,t)=>-s(e,t)}function Dhe(s){return(e,t)=>e===void 0?t===void 0?Au.neitherLessOrGreaterThan:Au.lessThan:t===void 0?Au.greaterThan:s(e,t)}class ph{constructor(e){this.firstIdx=0,this.items=e,this.lastIdx=this.items.length-1}get length(){return this.lastIdx-this.firstIdx+1}takeWhile(e){let t=this.firstIdx;for(;t<this.items.length&&e(this.items[t]);)t++;const i=t===this.firstIdx?null:this.items.slice(this.firstIdx,t);return this.firstIdx=t,i}takeFromEndWhile(e){let t=this.lastIdx;for(;t>=0&&e(this.items[t]);)t--;const i=t===this.lastIdx?null:this.items.slice(t+1,this.lastIdx+1);return this.lastIdx=t,i}peek(){if(this.length!==0)return this.items[this.firstIdx]}dequeue(){const e=this.items[this.firstIdx];return this.firstIdx++,e}takeCount(e){const t=this.items.slice(this.firstIdx,this.firstIdx+e);return this.firstIdx+=e,t}}const jb=class jb{constructor(e){this.iterate=e}toArray(){const e=[];return this.iterate(t=>(e.push(t),!0)),e}filter(e){return new jb(t=>this.iterate(i=>e(i)?t(i):!0))}map(e){return new jb(t=>this.iterate(i=>t(e(i))))}findLast(e){let t;return this.iterate(i=>(e(i)&&(t=i),!0)),t}findLastMaxBy(e){let t,i=!0;return this.iterate(n=>((i||Au.isGreaterThan(e(n,t)))&&(i=!1,t=n),!0)),t}};jb.empty=new jb(e=>{});let op=jb;class yE{constructor(e){this._indexMap=e}static createSortPermutation(e,t){const i=Array.from(e.keys()).sort((n,o)=>t(e[n],e[o]));return new yE(i)}apply(e){return e.map((t,i)=>e[this._indexMap[i]])}inverse(){const e=this._indexMap.slice();for(let t=0;t<this._indexMap.length;t++)e[this._indexMap[t]]=t;return new yE(e)}}function SE(s){return s.reduce((e,t)=>e+t,0)}let Jg;function The(s){Jg?Jg instanceof Fz?Jg.loggers.push(s):Jg=new Fz([Jg,s]):Jg=s}function gs(){return Jg}class Fz{constructor(e){this.loggers=e}handleObservableCreated(e,t){for(const i of this.loggers)i.handleObservableCreated(e,t)}handleOnListenerCountChanged(e,t){for(const i of this.loggers)i.handleOnListenerCountChanged(e,t)}handleObservableUpdated(e,t){for(const i of this.loggers)i.handleObservableUpdated(e,t)}handleAutorunCreated(e,t){for(const i of this.loggers)i.handleAutorunCreated(e,t)}handleAutorunDisposed(e){for(const t of this.loggers)t.handleAutorunDisposed(e)}handleAutorunDependencyChanged(e,t,i){for(const n of this.loggers)n.handleAutorunDependencyChanged(e,t,i)}handleAutorunStarted(e){for(const t of this.loggers)t.handleAutorunStarted(e)}handleAutorunFinished(e){for(const t of this.loggers)t.handleAutorunFinished(e)}handleDerivedDependencyChanged(e,t,i){for(const n of this.loggers)n.handleDerivedDependencyChanged(e,t,i)}handleDerivedCleared(e){for(const t of this.loggers)t.handleDerivedCleared(e)}handleBeginTransaction(e){for(const t of this.loggers)t.handleBeginTransaction(e)}handleEndTransaction(e){for(const t of this.loggers)t.handleEndTransaction(e)}}var dn;(function(s){let e=!1;function t(){e=!0}s.enable=t;function i(){if(!e)return;const n=Error,o=n.stackTraceLimit;n.stackTraceLimit=3;const r=new Error().stack;return n.stackTraceLimit=o,iW.fromStack(r,2)}s.ofCaller=i})(dn||(dn={}));class iW{static fromStack(e,t){const i=e.split(`
`),n=Nhe(i[t+1]);if(n)return new iW(n.fileName,n.line,n.column,n.id)}constructor(e,t,i,n){this.fileName=e,this.line=t,this.column=i,this.id=n}}function Nhe(s){const e=s.match(/\((.*):(\d+):(\d+)\)/);if(e)return{fileName:e[1],line:parseInt(e[2]),column:parseInt(e[3]),id:s};const t=s.match(/at ([^\(\)]*):(\d+):(\d+)/);if(t)return{fileName:t[1],line:parseInt(t[2]),column:parseInt(t[3]),id:s}}const kr=(s,e)=>s===e;function Gy(s=kr){return(e,t)=>ni(e,t,s)}function nW(){return(s,e)=>s.equals(e)}function Vf(s,e,t){if(t!==void 0){const i=s;return i==null||e===void 0||e===null?e===i:t(i,e)}else{const i=s;return(n,o)=>n==null||o===void 0||o===null?o===n:i(n,o)}}function vP(s,e){if(s===e)return!0;if(Array.isArray(s)&&Array.isArray(e)){if(s.length!==e.length)return!1;for(let t=0;t<s.length;t++)if(!vP(s[t],e[t]))return!1;return!0}if(s&&typeof s=="object"&&e&&typeof e=="object"&&Object.getPrototypeOf(s)===Object.prototype&&Object.getPrototypeOf(e)===Object.prototype){const t=s,i=e,n=Object.keys(t),o=Object.keys(i),r=new Set(o);if(n.length!==o.length)return!1;for(const a of n)if(!r.has(a)||!vP(t[a],i[a]))return!1;return!0}return!1}class os{constructor(e,t,i){this.owner=e,this.debugNameSource=t,this.referenceFn=i}getDebugName(e){return Rhe(e,this)}}const Bz=new Map,CP=new WeakMap;function Rhe(s,e){const t=CP.get(s);if(t)return t;const i=Ahe(s,e);if(i){let n=Bz.get(i)??0;n++,Bz.set(i,n);const o=n===1?i:`${i}#${n}`;return CP.set(s,o),o}}function Ahe(s,e){const t=CP.get(s);if(t)return t;const i=e.owner?Ohe(e.owner)+".":"";let n;const o=e.debugNameSource;if(o!==void 0)if(typeof o=="function"){if(n=o(),n!==void 0)return i+n}else return i+o;const r=e.referenceFn;if(r!==void 0&&(n=sW(r),n!==void 0))return i+n;if(e.owner!==void 0){const a=Mhe(e.owner,s);if(a!==void 0)return i+a}}function Mhe(s,e){for(const t in s)if(s[t]===e)return t}const Wz=new Map,Hz=new WeakMap;function Ohe(s){const e=Hz.get(s);if(e)return e;const t=wQ(s)??"Object";let i=Wz.get(t)??0;i++,Wz.set(t,i);const n=i===1?t:`${t}#${i}`;return Hz.set(s,n),n}function wQ(s){const e=s.constructor;if(e)return e.name==="Object"?void 0:e.name}function sW(s){const e=s.toString(),i=/\/\*\*\s*@description\s*([^*]*)\*\//.exec(e);return(i?i[1]:void 0)?.trim()}let yP;function Phe(s){yP=s}let vQ;function Fhe(s){vQ=s}class CQ{get TChange(){return null}reportChanges(){this.get()}read(e){return e?e.readObservable(this):this.get()}map(e,t,i=dn.ofCaller()){const n=t===void 0?void 0:e,o=t===void 0?e:t;return yP({owner:n,debugName:()=>{const r=sW(o);if(r!==void 0)return r;const l=/^\s*\(?\s*([a-zA-Z_$][a-zA-Z_$0-9]*)\s*\)?\s*=>\s*\1(?:\??)\.([a-zA-Z_$][a-zA-Z_$0-9]*)\s*$/.exec(o.toString());if(l)return`${this.debugName}.${l[2]}`;if(!n)return`${this.debugName} (mapped)`},debugReferenceFn:o},r=>o(this.read(r),r),i)}flatten(){return yP({owner:void 0,debugName:()=>`${this.debugName} (flattened)`},e=>this.read(e).read(e))}recomputeInitiallyAndOnChange(e,t){return e.add(vQ(this,t)),this}}class qv extends CQ{constructor(e){super(),this._observers=new Set,gs()?.handleObservableCreated(this,e)}addObserver(e){const t=this._observers.size;this._observers.add(e),t===0&&this.onFirstObserverAdded(),t!==this._observers.size&&gs()?.handleOnListenerCountChanged(this,this._observers.size)}removeObserver(e){const t=this._observers.delete(e);t&&this._observers.size===0&&this.onLastObserverRemoved(),t&&gs()?.handleOnListenerCountChanged(this,this._observers.size)}onFirstObserverAdded(){}onLastObserverRemoved(){}debugGetObservers(){return this._observers}}function Bhe(s){switch(s){case 0:return"initial";case 1:return"dependenciesMightHaveChanged";case 2:return"stale";case 3:return"upToDate";default:return"<unknown>"}}class Vr extends qv{get debugName(){return this._debugNameData.getDebugName(this)??"(anonymous)"}constructor(e,t,i,n=void 0,o,r){super(r),this._debugNameData=e,this._computeFn=t,this._changeTracker=i,this._handleLastObserverRemoved=n,this._equalityComparator=o,this._state=0,this._value=void 0,this._updateCount=0,this._dependencies=new Set,this._dependenciesToBeRemoved=new Set,this._changeSummary=void 0,this._isUpdating=!1,this._isComputing=!1,this._didReportChange=!1,this._isInBeforeUpdate=!1,this._isReaderValid=!1,this._store=void 0,this._delayedStore=void 0,this._removedObserverToCallEndUpdateOn=null,this._changeSummary=this._changeTracker?.createChangeSummary(void 0)}onLastObserverRemoved(){this._state=0,this._value=void 0,gs()?.handleDerivedCleared(this);for(const e of this._dependencies)e.removeObserver(this);this._dependencies.clear(),this._store!==void 0&&(this._store.dispose(),this._store=void 0),this._delayedStore!==void 0&&(this._delayedStore.dispose(),this._delayedStore=void 0),this._handleLastObserverRemoved?.()}get(){if(this._isComputing,this._observers.size===0){let t;try{this._isReaderValid=!0;let i;this._changeTracker&&(i=this._changeTracker.createChangeSummary(void 0),this._changeTracker.beforeUpdate?.(this,i)),t=this._computeFn(this,i)}finally{this._isReaderValid=!1}return this.onLastObserverRemoved(),t}else{do{if(this._state===1){for(const t of this._dependencies)if(t.reportChanges(),this._state===2)break}this._state===1&&(this._state=3),this._state!==3&&this._recompute()}while(this._state!==3);return this._value}}_recompute(){let e=!1;this._isComputing=!0,this._didReportChange=!1;const t=this._dependenciesToBeRemoved;this._dependenciesToBeRemoved=this._dependencies,this._dependencies=t;try{const i=this._changeSummary;this._isReaderValid=!0,this._changeTracker&&(this._isInBeforeUpdate=!0,this._changeTracker.beforeUpdate?.(this,i),this._isInBeforeUpdate=!1,this._changeSummary=this._changeTracker?.createChangeSummary(i));const n=this._state!==0,o=this._value;this._state=3;const r=this._delayedStore;r!==void 0&&(this._delayedStore=void 0);try{this._store!==void 0&&(this._store.dispose(),this._store=void 0),this._value=this._computeFn(this,i)}finally{this._isReaderValid=!1;for(const a of this._dependenciesToBeRemoved)a.removeObserver(this);this._dependenciesToBeRemoved.clear(),r!==void 0&&r.dispose()}e=this._didReportChange||n&&!this._equalityComparator(o,this._value),gs()?.handleObservableUpdated(this,{oldValue:o,newValue:this._value,change:void 0,didChange:e,hadValue:n})}catch(i){dE(i)}if(this._isComputing=!1,!this._didReportChange&&e)for(const i of this._observers)i.handleChange(this,void 0);else this._didReportChange=!1}toString(){return`LazyDerived<${this.debugName}>`}beginUpdate(e){if(this._isUpdating)throw new Ce("Cyclic deriveds are not supported yet!");this._updateCount++,this._isUpdating=!0;try{const t=this._updateCount===1;if(this._state===3&&(this._state=1,!t))for(const i of this._observers)i.handlePossibleChange(this);if(t)for(const i of this._observers)i.beginUpdate(this)}finally{this._isUpdating=!1}}endUpdate(e){if(this._updateCount--,this._updateCount===0){const t=[...this._observers];for(const i of t)i.endUpdate(this);if(this._removedObserverToCallEndUpdateOn){const i=[...this._removedObserverToCallEndUpdateOn];this._removedObserverToCallEndUpdateOn=null;for(const n of i)n.endUpdate(this)}}rg(()=>this._updateCount>=0)}handlePossibleChange(e){if(this._state===3&&this._dependencies.has(e)&&!this._dependenciesToBeRemoved.has(e)){this._state=1;for(const t of this._observers)t.handlePossibleChange(this)}}handleChange(e,t){if(this._dependencies.has(e)&&!this._dependenciesToBeRemoved.has(e)||this._isInBeforeUpdate){gs()?.handleDerivedDependencyChanged(this,e,t);let i=!1;try{i=this._changeTracker?this._changeTracker.handleChange({changedObservable:e,change:t,didChange:o=>o===e},this._changeSummary):!0}catch(o){dE(o)}const n=this._state===3;if(i&&(this._state===1||n)&&(this._state=2,n))for(const o of this._observers)o.handlePossibleChange(this)}}_ensureReaderValid(){if(!this._isReaderValid)throw new Ce("The reader object cannot be used outside its compute function!")}readObservable(e){this._ensureReaderValid(),e.addObserver(this);const t=e.get();return this._dependencies.add(e),this._dependenciesToBeRemoved.delete(e),t}get store(){return this._ensureReaderValid(),this._store===void 0&&(this._store=new U),this._store}addObserver(e){const t=!this._observers.has(e)&&this._updateCount>0;super.addObserver(e),t&&(this._removedObserverToCallEndUpdateOn&&this._removedObserverToCallEndUpdateOn.has(e)?this._removedObserverToCallEndUpdateOn.delete(e):e.beginUpdate(this))}removeObserver(e){this._observers.has(e)&&this._updateCount>0&&(this._removedObserverToCallEndUpdateOn||(this._removedObserverToCallEndUpdateOn=new Set),this._removedObserverToCallEndUpdateOn.add(e)),super.removeObserver(e)}debugGetState(){return{state:this._state,stateStr:Bhe(this._state),updateCount:this._updateCount,isComputing:this._isComputing,dependencies:this._dependencies,value:this._value}}debugSetValue(e){this._value=e}debugRecompute(){this._isComputing?this._state=2:this._recompute()}setValue(e,t,i){this._value=e;const n=this._observers;t.updateObserver(this,this);for(const o of n)o.handleChange(this,i)}}class yQ extends Vr{constructor(e,t,i,n=void 0,o,r,a){super(e,t,i,n,o,a),this.set=r}}function j(s,e,t=dn.ofCaller()){return e!==void 0?new Vr(new os(s,void 0,e),e,void 0,void 0,kr,t):new Vr(new os(void 0,void 0,s),s,void 0,void 0,kr,t)}function oW(s,e,t,i=dn.ofCaller()){return new yQ(new os(s,void 0,e),e,void 0,void 0,kr,t,i)}function Jn(s,e,t=dn.ofCaller()){return new Vr(new os(s.owner,s.debugName,s.debugReferenceFn),e,void 0,s.onLastObserverRemoved,s.equalsFn??kr,t)}Phe(Jn);function Vz(s,e,t=dn.ofCaller()){return new Vr(new os(s.owner,s.debugName,void 0),e,s.changeTracker,void 0,s.equalityComparer??kr,t)}function Wr(s,e,t=dn.ofCaller()){let i,n;e===void 0?(i=s,n=void 0):(n=s,i=e);let o;return new Vr(new os(n,void 0,i),r=>{o?o.clear():o=new U;const a=i(r);return a&&o.add(a),a},void 0,()=>{o&&(o.dispose(),o=void 0)},kr,t)}function Whe(s){switch(s){case 1:return"dependenciesMightHaveChanged";case 2:return"stale";case 3:return"upToDate";default:return"<unknown>"}}class Ky{get debugName(){return this._debugNameData.getDebugName(this)??"(anonymous)"}constructor(e,t,i,n){this._debugNameData=e,this._runFn=t,this._changeTracker=i,this._state=2,this._updateCount=0,this._disposed=!1,this._dependencies=new Set,this._dependenciesToBeRemoved=new Set,this._isRunning=!1,this._store=void 0,this._delayedStore=void 0,this._changeSummary=this._changeTracker?.createChangeSummary(void 0),gs()?.handleAutorunCreated(this,n),this._run()}dispose(){if(!this._disposed){this._disposed=!0;for(const e of this._dependencies)e.removeObserver(this);this._dependencies.clear(),this._store!==void 0&&this._store.dispose(),this._delayedStore!==void 0&&this._delayedStore.dispose(),gs()?.handleAutorunDisposed(this)}}_run(){const e=this._dependenciesToBeRemoved;this._dependenciesToBeRemoved=this._dependencies,this._dependencies=e,this._state=3;try{if(!this._disposed){gs()?.handleAutorunStarted(this);const t=this._changeSummary,i=this._delayedStore;i!==void 0&&(this._delayedStore=void 0);try{this._isRunning=!0,this._changeTracker&&(this._changeTracker.beforeUpdate?.(this,t),this._changeSummary=this._changeTracker.createChangeSummary(t)),this._store!==void 0&&(this._store.dispose(),this._store=void 0),this._runFn(this,t)}catch(n){dE(n)}finally{this._isRunning=!1,i!==void 0&&i.dispose()}}}finally{this._disposed||gs()?.handleAutorunFinished(this);for(const t of this._dependenciesToBeRemoved)t.removeObserver(this);this._dependenciesToBeRemoved.clear()}}toString(){return`Autorun<${this.debugName}>`}beginUpdate(e){this._state===3&&(this._state=1),this._updateCount++}endUpdate(e){try{if(this._updateCount===1)do{if(this._state===1){this._state=3;for(const t of this._dependencies)if(t.reportChanges(),this._state===2)break}this._state!==3&&this._run()}while(this._state!==3)}finally{this._updateCount--}rg(()=>this._updateCount>=0)}handlePossibleChange(e){this._state===3&&this._isDependency(e)&&(this._state=1)}handleChange(e,t){if(this._isDependency(e)){gs()?.handleAutorunDependencyChanged(this,e,t);try{(!this._changeTracker||this._changeTracker.handleChange({changedObservable:e,change:t,didChange:n=>n===e},this._changeSummary))&&(this._state=2)}catch(i){dE(i)}}}_isDependency(e){return this._dependencies.has(e)&&!this._dependenciesToBeRemoved.has(e)}_ensureNoRunning(){if(!this._isRunning)throw new Ce("The reader object cannot be used outside its compute function!")}readObservable(e){if(this._ensureNoRunning(),this._disposed)return e.get();e.addObserver(this);const t=e.get();return this._dependencies.add(e),this._dependenciesToBeRemoved.delete(e),t}get store(){if(this._ensureNoRunning(),this._disposed)throw new Ce("Cannot access store after dispose");return this._store===void 0&&(this._store=new U),this._store}debugGetState(){return{isRunning:this._isRunning,updateCount:this._updateCount,dependencies:this._dependencies,state:this._state,stateStr:Whe(this._state)}}debugRerun(){this._isRunning?this._state=2:this._run()}}function De(s,e=dn.ofCaller()){return new Ky(new os(void 0,void 0,s),s,void 0,e)}function DR(s,e,t=dn.ofCaller()){return new Ky(new os(s.owner,s.debugName,s.debugReferenceFn??e),e,void 0,t)}function jv(s,e,t=dn.ofCaller()){return new Ky(new os(s.owner,s.debugName,s.debugReferenceFn??e),e,s.changeTracker,t)}function Hhe(s,e){const t=new U,i=jv({owner:s.owner,debugName:s.debugName,debugReferenceFn:s.debugReferenceFn??e,changeTracker:s.changeTracker},(n,o)=>{t.clear(),e(n,o,t)});return pe(()=>{i.dispose(),t.dispose()})}function ps(s){const e=new U,t=DR({owner:void 0,debugName:void 0,debugReferenceFn:s},i=>{e.clear(),s(i,e)});return pe(()=>{t.dispose(),e.dispose()})}function Vhe(s,e){let t;return DR({debugReferenceFn:e},i=>{const n=s.read(i),o=t;t=n,e({lastValue:o,newValue:n})})}function zz(s){const e=new Error("BugIndicatingErrorRecovery: "+s);Oe(e),console.error("recovered from an error that indicates a bug",e)}function Vt(s,e){const t=new Gv(s,e);try{s(t)}finally{t.finish()}}let Gx;function o1(s){if(Gx)s(Gx);else{const e=new Gv(s,void 0);Gx=e;try{s(e)}finally{e.finish(),Gx=void 0}}}async function zhe(s,e){const t=new Gv(s,e);try{await s(t)}finally{t.finish()}}function nv(s,e,t){s?e(s):Vt(e,t)}class Gv{constructor(e,t){this._fn=e,this._getDebugName=t,this._updatingObservers=[],gs()?.handleBeginTransaction(this)}getDebugName(){return this._getDebugName?this._getDebugName():sW(this._fn)}updateObserver(e,t){if(!this._updatingObservers){zz("Transaction already finished!"),Vt(i=>{i.updateObserver(e,t)});return}this._updatingObservers.push({observer:e,observable:t}),e.beginUpdate(t)}finish(){const e=this._updatingObservers;if(!e){zz("transaction.finish() has already been called!");return}for(let t=0;t<e.length;t++){const{observer:i,observable:n}=e[t];i.endUpdate(n)}this._updatingObservers=null,gs()?.handleEndTransaction(this)}debugGetUpdatingObservers(){return this._updatingObservers}}function bt(...s){let e,t,i,n;return s.length===2?[t,i]=s:[e,t,i,n]=s,new Wd(new os(e,void 0,i),t,i,()=>Wd.globalTransaction,kr,n??dn.ofCaller())}function SQ(s,e,t,i=dn.ofCaller()){return new Wd(new os(s.owner,s.debugName,s.debugReferenceFn??t),e,t,()=>Wd.globalTransaction,s.equalsFn??kr,i)}class Wd extends qv{constructor(e,t,i,n,o,r){super(r),this._debugNameData=e,this.event=t,this._getValue=i,this._getTransaction=n,this._equalityComparator=o,this._hasValue=!1,this.handleEvent=a=>{const l=this._getValue(a),c=this._value,d=!this._hasValue||!this._equalityComparator(c,l);let h=!1;d&&(this._value=l,this._hasValue&&(h=!0,nv(this._getTransaction(),u=>{gs()?.handleObservableUpdated(this,{oldValue:c,newValue:l,change:void 0,didChange:d,hadValue:this._hasValue});for(const g of this._observers)u.updateObserver(g,this),g.handleChange(this,void 0)},()=>{const u=this.getDebugName();return"Event fired"+(u?`: ${u}`:"")})),this._hasValue=!0),h||gs()?.handleObservableUpdated(this,{oldValue:c,newValue:l,change:void 0,didChange:d,hadValue:this._hasValue})}}getDebugName(){return this._debugNameData.getDebugName(this)}get debugName(){const e=this.getDebugName();return"From Event"+(e?`: ${e}`:"")}onFirstObserverAdded(){this._subscription=this.event(this.handleEvent)}onLastObserverRemoved(){this._subscription.dispose(),this._subscription=void 0,this._hasValue=!1,this._value=void 0}get(){return this._subscription?(this._hasValue||this.handleEvent(void 0),this._value):this._getValue(void 0)}debugSetValue(e){this._value=e}debugGetState(){return{value:this._value,hasValue:this._hasValue}}}(function(s){s.Observer=Wd;function e(t,i){let n=!1;Wd.globalTransaction===void 0&&(Wd.globalTransaction=t,n=!0);try{i()}finally{n&&(Wd.globalTransaction=void 0)}}s.batchEventsGlobally=e})(bt||(bt={}));function $z(s,e){let t=!1,i,n;return bt(o=>{const r=De(a=>{const l=s.read(a);t?(n&&clearTimeout(n),n=setTimeout(()=>{i=l,o()},e)):(t=!0,i=l)});return{dispose(){r.dispose(),t=!1,i=void 0}}},()=>t?i:s.get())}function sv(s,e){const t=new $he(!0,e);s.addObserver(t);try{t.beginUpdate(s)}finally{t.endUpdate(s)}return pe(()=>{s.removeObserver(t)})}Fhe(sv);class $he{constructor(e,t){this._forceRecompute=e,this._handleValue=t,this._counter=0}beginUpdate(e){this._counter++}endUpdate(e){this._counter===1&&this._forceRecompute&&(this._handleValue?this._handleValue(e.get()):e.reportChanges()),this._counter--}handlePossibleChange(e){}handleChange(e,t){}}function Ih(s,e){let t;return Jn({owner:s,debugReferenceFn:e},n=>(t=e(n,t),t))}function rW(s,e,t,i){let n=new Uz(t,i);return Jn({debugReferenceFn:t,owner:s,onLastObserverRemoved:()=>{n.dispose(),n=new Uz(t)}},r=>(n.setItems(e.read(r)),n.getItems()))}let Uz=class{constructor(e,t){this._map=e,this._keySelector=t,this._cache=new Map,this._items=[]}dispose(){this._cache.forEach(e=>e.store.dispose()),this._cache.clear()}setItems(e){const t=[],i=new Set(this._cache.keys());for(const n of e){const o=this._keySelector?this._keySelector(n):n;let r=this._cache.get(o);if(r)i.delete(o);else{const a=new U;r={out:this._map(n,a),store:a},this._cache.set(o,r)}t.push(r.out)}for(const n of i)this._cache.get(n).store.dispose(),this._cache.delete(n);this._items=t}getItems(){return this._items}};function Cb(s,e){switch(typeof s){case"number":return""+s;case"string":return s.length+2<=e?`"${s}"`:`"${s.substr(0,e-7)}"+...`;case"boolean":return s?"true":"false";case"undefined":return"undefined";case"object":return s===null?"null":Array.isArray(s)?Uhe(s,e):qhe(s,e);case"symbol":return s.toString();case"function":return`[[Function${s.name?" "+s.name:""}]]`;default:return""+s}}function Uhe(s,e){let t="[ ",i=!0;for(const n of s){if(i||(t+=", "),t.length-5>e){t+="...";break}i=!1,t+=`${Cb(n,e-t.length)}`}return t+=" ]",t}function qhe(s,e){if(typeof s.toString=="function"&&s.toString!==Object.prototype.toString){const o=s.toString();return o.length<=e?o:o.substring(0,e-3)+"..."}const t=wQ(s);let i=t?t+"(":"{ ",n=!0;for(const[o,r]of Object.entries(s)){if(n||(i+=", "),i.length-5>e){i+="...";break}n=!1,i+=`${o}: ${Cb(r,e-i.length)}`}return i+=t?")":" }",i}class aW{static createClient(e,t){return new aW(e,t)}constructor(e,t){this._channelFactory=e,this._getHandler=t,this._channel=this._channelFactory({handleNotification:o=>{const r=o,a=this._getHandler().notifications[r[0]];if(!a)throw new Error(`Unknown notification "${r[0]}"!`);a(...r[1])},handleRequest:o=>{const r=o;try{return{type:"result",value:this._getHandler().requests[r[0]](...r[1])}}catch(a){return{type:"error",value:a}}}});const i=new Proxy({},{get:(o,r)=>async(...a)=>{const l=await this._channel.sendRequest([r,a]);if(l.type==="error")throw l.value;return l.value}}),n=new Proxy({},{get:(o,r)=>(...a)=>{this._channel.sendNotification([r,a])}});this.api={notifications:n,requests:i}}}function jhe(s,e){const t=globalThis;let i=[],n;const{channel:o,handler:r}=Ghe({sendNotification:l=>{n?n.sendNotification(l):i.push(l)}});let a;return(t.$$debugValueEditor_debugChannels??(t.$$debugValueEditor_debugChannels={}))[s]=l=>{a=e(),n=l;for(const c of i)l.sendNotification(c);return i=[],r},aW.createClient(o,()=>{if(!a)throw new Error("Not supported");return a})}function Ghe(s){let e;return{channel:i=>(e=i,{sendNotification:n=>{s.sendNotification(n)},sendRequest:n=>{throw new Error("not supported")}}),handler:{handleRequest:i=>i.type==="notification"?e?.handleNotification(i.data):e?.handleRequest(i.data)}}}class Khe{constructor(){this._timeout=void 0}throttle(e,t){this._timeout===void 0&&(this._timeout=setTimeout(()=>{this._timeout=void 0,e()},t))}dispose(){this._timeout!==void 0&&clearTimeout(this._timeout)}}function xQ(s,e){for(const t in e)s[t]&&typeof s[t]=="object"&&e[t]&&typeof e[t]=="object"?xQ(s[t],e[t]):s[t]=e[t]}function LQ(s,e){for(const t in e)e[t]===null?delete s[t]:s[t]&&typeof s[t]=="object"&&e[t]&&typeof e[t]=="object"?LQ(s[t],e[t]):s[t]=e[t]}function Re(s,e,t=dn.ofCaller()){let i;return typeof s=="string"?i=new os(void 0,s,void 0):i=new os(s,void 0,void 0),new Yy(i,e,kr,t)}class Yy extends qv{get debugName(){return this._debugNameData.getDebugName(this)??"ObservableValue"}constructor(e,t,i,n){super(n),this._debugNameData=e,this._equalityComparator=i,this._value=t,gs()?.handleObservableUpdated(this,{hadValue:!1,newValue:t,change:void 0,didChange:!0,oldValue:void 0})}get(){return this._value}set(e,t,i){if(i===void 0&&this._equalityComparator(this._value,e))return;let n;t||(t=n=new Gv(()=>{},()=>`Setting ${this.debugName}`));try{const o=this._value;this._setValue(e),gs()?.handleObservableUpdated(this,{oldValue:o,newValue:e,change:i,didChange:!0,hadValue:!0});for(const r of this._observers)t.updateObserver(r,this),r.handleChange(this,i)}finally{n&&n.finish()}}toString(){return`${this.debugName}: ${this._value}`}_setValue(e){this._value=e}debugGetState(){return{value:this._value}}debugSetValue(e){this._value=e}}function lW(s,e,t=dn.ofCaller()){let i;return typeof s=="string"?i=new os(void 0,s,void 0):i=new os(s,void 0,void 0),new Yhe(i,e,kr,t)}class Yhe extends Yy{_setValue(e){this._value!==e&&(this._value&&this._value.dispose(),this._value=e)}dispose(){this._value?.dispose()}}const Df=class Df{static getInstance(){return Df._instance===void 0&&(Df._instance=new Df),Df._instance}getTransactionState(){const e=[],t=[...this._activeTransactions];if(t.length===0)return;const i=t.flatMap(o=>o.debugGetUpdatingObservers()??[]).map(o=>o.observer),n=new Set;for(;i.length>0;){const o=i.shift();if(n.has(o))continue;n.add(o);const r=this._getInfo(o,a=>{n.has(a)||i.push(a)});r&&e.push(r)}return{names:t.map(o=>o.getDebugName()??"tx"),affected:e}}_getObservableInfo(e){const t=this._instanceInfos.get(e);if(!t){Oe(new Ce("No info found"));return}return t}_getAutorunInfo(e){const t=this._instanceInfos.get(e);if(!t){Oe(new Ce("No info found"));return}return t}_getInfo(e,t){if(e instanceof Vr){const i=[...e.debugGetObservers()];for(const l of i)t(l);const n=this._getObservableInfo(e);if(!n)return;const o=e.debugGetState(),r={name:e.debugName,instanceId:n.instanceId,updateCount:o.updateCount},a=[...n.changedObservables].map(l=>this._instanceInfos.get(l)?.instanceId).filter(Ln);if(o.isComputing)return{...r,type:"observable/derived",state:"updating",changedDependencies:a,initialComputation:!1};switch(o.state){case 0:return{...r,type:"observable/derived",state:"noValue"};case 3:return{...r,type:"observable/derived",state:"upToDate"};case 2:return{...r,type:"observable/derived",state:"stale",changedDependencies:a};case 1:return{...r,type:"observable/derived",state:"possiblyStale"}}}else if(e instanceof Ky){const i=this._getAutorunInfo(e);if(!i)return;const n={name:e.debugName,instanceId:i.instanceId,updateCount:i.updateCount},o=[...i.changedObservables].map(r=>this._instanceInfos.get(r).instanceId);if(e.debugGetState().isRunning)return{...n,type:"autorun",state:"updating",changedDependencies:o};switch(e.debugGetState().state){case 3:return{...n,type:"autorun",state:"upToDate"};case 2:return{...n,type:"autorun",state:"stale",changedDependencies:o};case 1:return{...n,type:"autorun",state:"possiblyStale"}}}}_formatObservable(e){const t=this._getObservableInfo(e);if(t)return{name:e.debugName,instanceId:t.instanceId}}_formatObserver(e){if(e instanceof Vr)return{name:e.toString(),instanceId:this._getObservableInfo(e)?.instanceId};const t=this._getAutorunInfo(e);if(t)return{name:e.toString(),instanceId:t.instanceId}}constructor(){this._declarationId=0,this._instanceId=0,this._declarations=new Map,this._instanceInfos=new WeakMap,this._aliveInstances=new Map,this._activeTransactions=new Set,this._channel=jhe("observableDevTools",()=>({notifications:{setDeclarationIdFilter:e=>{},logObservableValue:e=>{console.log("logObservableValue",e)},flushUpdates:()=>{this._flushUpdates()},resetUpdates:()=>{this._pendingChanges=null,this._channel.api.notifications.handleChange(this._fullState,!0)}},requests:{getDeclarations:()=>{const e={};for(const t of this._declarations.values())e[t.id]=t;return{decls:e}},getSummarizedInstances:()=>null,getObservableValueInfo:e=>({observers:[...this._aliveInstances.get(e).debugGetObservers()].map(i=>this._formatObserver(i)).filter(Ln)}),getDerivedInfo:e=>{const t=this._aliveInstances.get(e);return{dependencies:[...t.debugGetState().dependencies].map(i=>this._formatObservable(i)).filter(Ln),observers:[...t.debugGetObservers()].map(i=>this._formatObserver(i)).filter(Ln)}},getAutorunInfo:e=>({dependencies:[...this._aliveInstances.get(e).debugGetState().dependencies].map(i=>this._formatObservable(i)).filter(Ln)}),getTransactionState:()=>this.getTransactionState(),setValue:(e,t)=>{const i=this._aliveInstances.get(e);if(i instanceof Vr)i.debugSetValue(t);else if(i instanceof Yy)i.debugSetValue(t);else if(i instanceof Wd)i.debugSetValue(t);else throw new Ce("Observable is not supported");const n=[...i.debugGetObservers()];for(const o of n)o.beginUpdate(i);for(const o of n)o.handleChange(i,void 0);for(const o of n)o.endUpdate(i)},getValue:e=>{const t=this._aliveInstances.get(e);if(t instanceof Vr)return Cb(t.debugGetState().value,200);if(t instanceof Yy)return Cb(t.debugGetState().value,200)},logValue:e=>{const t=this._aliveInstances.get(e);if(t&&"get"in t)console.log("Logged Value:",t.get());else throw new Ce("Observable is not supported")},rerun:e=>{const t=this._aliveInstances.get(e);if(t instanceof Vr)t.debugRecompute();else if(t instanceof Ky)t.debugRerun();else throw new Ce("Observable is not supported")}}})),this._pendingChanges=null,this._changeThrottler=new Khe,this._fullState={},this._flushUpdates=()=>{this._pendingChanges!==null&&(this._channel.api.notifications.handleChange(this._pendingChanges,!1),this._pendingChanges=null)},dn.enable()}_handleChange(e){LQ(this._fullState,e),this._pendingChanges===null?this._pendingChanges=e:xQ(this._pendingChanges,e),this._changeThrottler.throttle(this._flushUpdates,10)}_getDeclarationId(e,t){if(!t)return-1;let i=this._declarations.get(t.id);return i===void 0&&(i={id:this._declarationId++,type:e,url:t.fileName,line:t.line,column:t.column},this._declarations.set(t.id,i),this._handleChange({decls:{[i.id]:i}})),i.id}handleObservableCreated(e,t){const n={declarationId:this._getDeclarationId("observable/value",t),instanceId:this._instanceId++,listenerCount:0,lastValue:void 0,updateCount:0,changedObservables:new Set};this._instanceInfos.set(e,n)}handleOnListenerCountChanged(e,t){const i=this._getObservableInfo(e);if(i){if(i.listenerCount===0&&t>0){const n=e instanceof Vr?"observable/derived":"observable/value";this._aliveInstances.set(i.instanceId,e),this._handleChange({instances:{[i.instanceId]:{instanceId:i.instanceId,declarationId:i.declarationId,formattedValue:i.lastValue,type:n,name:e.debugName}}})}else i.listenerCount>0&&t===0&&(this._handleChange({instances:{[i.instanceId]:null}}),this._aliveInstances.delete(i.instanceId));i.listenerCount=t}}handleObservableUpdated(e,t){if(e instanceof Vr){this._handleDerivedRecomputed(e,t);return}const i=this._getObservableInfo(e);i&&t.didChange&&(i.lastValue=Cb(t.newValue,30),i.listenerCount>0&&this._handleChange({instances:{[i.instanceId]:{formattedValue:i.lastValue}}}))}handleAutorunCreated(e,t){const n={declarationId:this._getDeclarationId("autorun",t),instanceId:this._instanceId++,updateCount:0,changedObservables:new Set};this._instanceInfos.set(e,n),this._aliveInstances.set(n.instanceId,e),n&&this._handleChange({instances:{[n.instanceId]:{instanceId:n.instanceId,declarationId:n.declarationId,runCount:0,type:"autorun",name:e.debugName}}})}handleAutorunDisposed(e){const t=this._getAutorunInfo(e);t&&(this._handleChange({instances:{[t.instanceId]:null}}),this._instanceInfos.delete(e),this._aliveInstances.delete(t.instanceId))}handleAutorunDependencyChanged(e,t,i){const n=this._getAutorunInfo(e);n&&n.changedObservables.add(t)}handleAutorunStarted(e){}handleAutorunFinished(e){const t=this._getAutorunInfo(e);t&&(t.changedObservables.clear(),t.updateCount++,this._handleChange({instances:{[t.instanceId]:{runCount:t.updateCount}}}))}handleDerivedDependencyChanged(e,t,i){const n=this._getObservableInfo(e);n&&n.changedObservables.add(t)}_handleDerivedRecomputed(e,t){const i=this._getObservableInfo(e);if(!i)return;const n=Cb(t.newValue,30);i.updateCount++,i.changedObservables.clear(),i.lastValue=n,i.listenerCount>0&&this._handleChange({instances:{[i.instanceId]:{formattedValue:n,recomputationCount:i.updateCount}}})}handleDerivedCleared(e){const t=this._getObservableInfo(e);t&&(t.lastValue=void 0,t.changedObservables.clear(),t.listenerCount>0&&this._handleChange({instances:{[t.instanceId]:{formattedValue:void 0}}}))}handleBeginTransaction(e){this._activeTransactions.add(e)}handleEndTransaction(e){this._activeTransactions.delete(e)}};Df._instance=void 0;let SP=Df;_E&&_E.VSCODE_DEV_DEBUG_OBSERVABLES&&The(SP.getInstance());const{getWindow:be,getDocument:Zhe,getWindows:kQ,getWindowsCount:Xhe,getWindowId:Zy,getWindowById:qz,onDidRegisterWindow:G0,onWillUnregisterWindow:Qhe,onDidUnregisterWindow:Jhe}=(function(){const s=new Map;Fce(Nt,1);const e={window:Nt,disposables:new U};s.set(Nt.vscodeWindowId,e);const t=new M,i=new M,n=new M;function o(r,a){return(typeof r=="number"?s.get(r):void 0)??(a?e:void 0)}return{onDidRegisterWindow:t.event,onWillUnregisterWindow:n.event,onDidUnregisterWindow:i.event,registerWindow(r){if(s.has(r.vscodeWindowId))return F.None;const a=new U,l={window:r,disposables:a.add(new U)};return s.set(r.vscodeWindowId,l),a.add(pe(()=>{s.delete(r.vscodeWindowId),i.fire(r)})),a.add(W(r,se.BEFORE_UNLOAD,()=>{n.fire(r)})),t.fire(l),a},getWindows(){return s.values()},getWindowsCount(){return s.size},getWindowId(r){return r.vscodeWindowId},hasWindow(r){return s.has(r)},getWindowById:o,getWindow(r){const a=r;if(a?.ownerDocument?.defaultView)return a.ownerDocument.defaultView.window;const l=r;return l?.view?l.view.window:Nt},getDocument(r){return be(r).document}}})();function Pn(s){for(;s.firstChild;)s.firstChild.remove()}class eue{constructor(e,t,i,n){this._node=e,this._type=t,this._handler=i,this._options=n||!1,this._node.addEventListener(this._type,this._handler,this._options)}dispose(){this._handler&&(this._node.removeEventListener(this._type,this._handler,this._options),this._node=null,this._handler=null)}}function W(s,e,t,i){return new eue(s,e,t,i)}function EQ(s,e){return function(t){return e(new is(s,t))}}function tue(s){return function(e){return s(new Pt(e))}}const Fi=function(e,t,i,n){let o=i;return t==="click"||t==="mousedown"||t==="contextmenu"?o=EQ(be(e),i):(t==="keydown"||t==="keypress"||t==="keyup")&&(o=tue(i)),W(e,t,o,n)},iue=function(e,t,i){const n=EQ(be(e),t);return IQ(e,n,i)};function IQ(s,e,t){return W(s,na&&z0.pointerEvents?se.POINTER_DOWN:se.MOUSE_DOWN,e,t)}function nue(s,e,t){return W(s,na&&z0.pointerEvents?se.POINTER_MOVE:se.MOUSE_MOVE,e,t)}function jz(s,e,t){return W(s,na&&z0.pointerEvents?se.POINTER_UP:se.MOUSE_UP,e,t)}function r1(s,e,t){return E1(s,e,t)}class B2 extends YX{constructor(e,t){super(e,t)}}let xE,Co;class cW extends $6{constructor(e){super(),this.defaultTarget=e&&be(e)}cancelAndSet(e,t,i){return super.cancelAndSet(e,t,i??this.defaultTarget)}}class W2{constructor(e,t=0){this._runner=e,this.priority=t,this._canceled=!1}dispose(){this._canceled=!0}execute(){if(!this._canceled)try{this._runner()}catch(e){Oe(e)}}static sort(e,t){return t.priority-e.priority}}(function(){const s=new Map,e=new Map,t=new Map,i=new Map,n=o=>{t.set(o,!1);const r=s.get(o)??[];for(e.set(o,r),s.set(o,[]),i.set(o,!0);r.length>0;)r.sort(W2.sort),r.shift().execute();i.set(o,!1)};Co=(o,r,a=0)=>{const l=Zy(o),c=new W2(r,a);let d=s.get(l);return d||(d=[],s.set(l,d)),d.push(c),t.get(l)||(t.set(l,!0),o.requestAnimationFrame(()=>n(l))),c},xE=(o,r,a)=>{const l=Zy(o);if(i.get(l)){const c=new W2(r,a);let d=e.get(l);return d||(d=[],e.set(l,d)),d.push(c),c}else return Co(o,r,a)}})();function TR(s){return be(s).getComputedStyle(s,null)}function lg(s,e,t){const i=be(s),n=i.document;if(s!==n.body)return new St(s.clientWidth,s.clientHeight);if(na&&i?.visualViewport)return new St(i.visualViewport.width,i.visualViewport.height);if(i?.innerWidth&&i.innerHeight)return new St(i.innerWidth,i.innerHeight);if(n.body&&n.body.clientWidth&&n.body.clientHeight)return new St(n.body.clientWidth,n.body.clientHeight);if(n.documentElement&&n.documentElement.clientWidth&&n.documentElement.clientHeight)return new St(n.documentElement.clientWidth,n.documentElement.clientHeight);throw new Error("Unable to figure out browser width and height")}class Pi{static convertToPixels(e,t){return parseFloat(t)||0}static getDimension(e,t){const i=TR(e),n=i?i.getPropertyValue(t):"0";return Pi.convertToPixels(e,n)}static getBorderLeftWidth(e){return Pi.getDimension(e,"border-left-width")}static getBorderRightWidth(e){return Pi.getDimension(e,"border-right-width")}static getBorderTopWidth(e){return Pi.getDimension(e,"border-top-width")}static getBorderBottomWidth(e){return Pi.getDimension(e,"border-bottom-width")}static getPaddingLeft(e){return Pi.getDimension(e,"padding-left")}static getPaddingRight(e){return Pi.getDimension(e,"padding-right")}static getPaddingTop(e){return Pi.getDimension(e,"padding-top")}static getPaddingBottom(e){return Pi.getDimension(e,"padding-bottom")}static getMarginLeft(e){return Pi.getDimension(e,"margin-left")}static getMarginTop(e){return Pi.getDimension(e,"margin-top")}static getMarginRight(e){return Pi.getDimension(e,"margin-right")}static getMarginBottom(e){return Pi.getDimension(e,"margin-bottom")}}const Tf=class Tf{constructor(e,t){this.width=e,this.height=t}with(e=this.width,t=this.height){return e!==this.width||t!==this.height?new Tf(e,t):this}static is(e){return typeof e=="object"&&typeof e.height=="number"&&typeof e.width=="number"}static lift(e){return e instanceof Tf?e:new Tf(e.width,e.height)}static equals(e,t){return e===t?!0:!e||!t?!1:e.width===t.width&&e.height===t.height}};Tf.None=new Tf(0,0);let St=Tf;function DQ(s){let e=s.offsetParent,t=s.offsetTop,i=s.offsetLeft;for(;(s=s.parentNode)!==null&&s!==s.ownerDocument.body&&s!==s.ownerDocument.documentElement;){t-=s.scrollTop;const n=NQ(s)?null:TR(s);n&&(i-=n.direction!=="rtl"?s.scrollLeft:-s.scrollLeft),s===e&&(i+=Pi.getBorderLeftWidth(s),t+=Pi.getBorderTopWidth(s),t+=s.offsetTop,i+=s.offsetLeft,e=s.offsetParent)}return{left:i,top:t}}function sue(s,e,t){typeof e=="number"&&(s.style.width=`${e}px`),typeof t=="number"&&(s.style.height=`${t}px`)}function Li(s){const e=s.getBoundingClientRect(),t=be(s);return{left:e.left+t.scrollX,top:e.top+t.scrollY,width:e.width,height:e.height}}function TQ(s){let e=s,t=1;do{const i=TR(e).zoom;i!=null&&i!=="1"&&(t*=i),e=e.parentElement}while(e!==null&&e!==e.ownerDocument.documentElement);return t}function Ro(s){const e=Pi.getMarginLeft(s)+Pi.getMarginRight(s);return s.offsetWidth+e}function H2(s){const e=Pi.getBorderLeftWidth(s)+Pi.getBorderRightWidth(s),t=Pi.getPaddingLeft(s)+Pi.getPaddingRight(s);return s.offsetWidth-e-t}function oue(s){const e=Pi.getBorderTopWidth(s)+Pi.getBorderBottomWidth(s),t=Pi.getPaddingTop(s)+Pi.getPaddingBottom(s);return s.offsetHeight-e-t}function Hd(s){const e=Pi.getMarginTop(s)+Pi.getMarginBottom(s);return s.offsetHeight+e}function wn(s,e){return!!e?.contains(s)}function rue(s,e,t){for(;s&&s.nodeType===s.ELEMENT_NODE;){if(s.classList.contains(e))return s;if(t){if(typeof t=="string"){if(s.classList.contains(t))return null}else if(s===t)return null}s=s.parentNode}return null}function V2(s,e,t){return!!rue(s,e,t)}function NQ(s){return s&&!!s.host&&!!s.mode}function LE(s){return!!cg(s)}function cg(s){for(;s.parentNode;){if(s===s.ownerDocument?.body)return null;s=s.parentNode}return NQ(s)?s:null}function ln(){let s=K0().activeElement;for(;s?.shadowRoot;)s=s.shadowRoot.activeElement;return s}function NR(s){return ln()===s}function RQ(s){return wn(ln(),s)}function K0(){return Xhe()<=1?Nt.document:Array.from(kQ()).map(({window:e})=>e.document).find(e=>e.hasFocus())??Nt.document}function kt(){return K0().defaultView?.window??Nt}const aue=new class{constructor(){this.mutationObservers=new Map}observe(s,e,t){let i=this.mutationObservers.get(s);i||(i=new Map,this.mutationObservers.set(s,i));const n=j0(t);let o=i.get(n);if(o)o.users+=1;else{const r=new M,a=new MutationObserver(c=>r.fire(c));a.observe(s,t);const l=o={users:1,observer:a,onDidMutate:r.event};e.add(pe(()=>{l.users-=1,l.users===0&&(r.dispose(),a.disconnect(),i?.delete(n),i?.size===0&&this.mutationObservers.delete(s))})),i.set(n,o)}return o.onDidMutate}};function ki(s){return s instanceof HTMLElement||s instanceof be(s).HTMLElement}function Gz(s){return s instanceof HTMLAnchorElement||s instanceof be(s).HTMLAnchorElement}function AQ(s){return s instanceof SVGElement||s instanceof be(s).SVGElement}function dW(s){return s instanceof MouseEvent||s instanceof be(s).MouseEvent}function Id(s){return s instanceof KeyboardEvent||s instanceof be(s).KeyboardEvent}const se={CLICK:"click",AUXCLICK:"auxclick",DBLCLICK:"dblclick",MOUSE_UP:"mouseup",MOUSE_DOWN:"mousedown",MOUSE_OVER:"mouseover",MOUSE_MOVE:"mousemove",MOUSE_OUT:"mouseout",MOUSE_ENTER:"mouseenter",MOUSE_LEAVE:"mouseleave",MOUSE_WHEEL:"wheel",POINTER_UP:"pointerup",POINTER_DOWN:"pointerdown",POINTER_MOVE:"pointermove",POINTER_LEAVE:"pointerleave",CONTEXT_MENU:"contextmenu",KEY_DOWN:"keydown",KEY_UP:"keyup",BEFORE_UNLOAD:"beforeunload",FOCUS:"focus",FOCUS_IN:"focusin",FOCUS_OUT:"focusout",BLUR:"blur",INPUT:"input",DRAG_START:"dragstart",DRAG:"drag",DRAG_ENTER:"dragenter",DRAG_LEAVE:"dragleave",DRAG_OVER:"dragover",DROP:"drop",DRAG_END:"dragend"};function lue(s){const e=s;return!!(e&&typeof e.preventDefault=="function"&&typeof e.stopPropagation=="function")}const gt={stop:(s,e)=>(s.preventDefault(),e&&s.stopPropagation(),s)};function cue(s){const e=[];for(let t=0;s&&s.nodeType===s.ELEMENT_NODE;t++)e[t]=s.scrollTop,s=s.parentNode;return e}function due(s,e){for(let t=0;s&&s.nodeType===s.ELEMENT_NODE;t++)s.scrollTop!==e[t]&&(s.scrollTop=e[t]),s=s.parentNode}let hue=class xP extends F{get onDidFocus(){return this._onDidFocus.event}get onDidBlur(){return this._onDidBlur.event}static hasFocusWithin(e){if(ki(e)){const t=cg(e),i=t?t.activeElement:e.ownerDocument.activeElement;return wn(i,e)}else{const t=e;return wn(t.document.activeElement,t.document)}}constructor(e){super(),this._onDidFocus=this._register(new M),this._onDidBlur=this._register(new M);let t=xP.hasFocusWithin(e),i=!1;const n=()=>{i=!1,t||(t=!0,this._onDidFocus.fire())},o=()=>{t&&(i=!0,(ki(e)?be(e):e).setTimeout(()=>{i&&(i=!1,t=!1,this._onDidBlur.fire())},0))};this._refreshStateHandler=()=>{xP.hasFocusWithin(e)!==t&&(t?o():n())},this._register(W(e,se.FOCUS,n,!0)),this._register(W(e,se.BLUR,o,!0)),ki(e)&&(this._register(W(e,se.FOCUS_IN,()=>this._refreshStateHandler())),this._register(W(e,se.FOCUS_OUT,()=>this._refreshStateHandler())))}};function oa(s){return new hue(s)}function uue(s,e){return s.after(e),e}function X(s,...e){if(s.append(...e),e.length===1&&typeof e[0]!="string")return e[0]}function RR(s,e){return s.insertBefore(e,s.firstChild),e}function vn(s,...e){s.textContent="",X(s,...e)}const gue=/([\w\-]+)?(#([\w\-]+))?((\.([\w\-]+))*)/;var Xy;(function(s){s.HTML="http://www.w3.org/1999/xhtml",s.SVG="http://www.w3.org/2000/svg"})(Xy||(Xy={}));function MQ(s,e,t,...i){const n=gue.exec(e);if(!n)throw new Error("Bad use of emmet");const o=n[1]||"div";let r;return s!==Xy.HTML?r=document.createElementNS(s,o):r=document.createElement(o),n[3]&&(r.id=n[3]),n[4]&&(r.className=n[4].replace(/\./g," ").trim()),t&&Object.entries(t).forEach(([a,l])=>{typeof l>"u"||(/^on\w+$/.test(a)?r[a]=l:a==="selected"?l&&r.setAttribute(a,"true"):r.setAttribute(a,l))}),r.append(...i),r}function ie(s,e,...t){return MQ(Xy.HTML,s,e,...t)}ie.SVG=function(s,e,...t){return MQ(Xy.SVG,s,e,...t)};function fue(s,...e){s?Mo(...e):js(...e)}function Mo(...s){for(const e of s)e.style.display="",e.removeAttribute("aria-hidden")}function js(...s){for(const e of s)e.style.display="none",e.setAttribute("aria-hidden","true")}function Kz(s,e){const t=s.devicePixelRatio*e;return Math.max(1,Math.floor(t))/s.devicePixelRatio}function OQ(s){Nt.open(s,"_blank","noopener")}function pue(s,e){const t=()=>{e(),i=Co(s,t)};let i=Co(s,t);return pe(()=>i.dispose())}uQ.setPreferredWebSchema(/^https:/.test(Nt.location.href)?"https":"http");class Vd extends M{constructor(){super(),this._subscriptions=new U,this._keyStatus={altKey:!1,shiftKey:!1,ctrlKey:!1,metaKey:!1},this._subscriptions.add(oe.runAndSubscribe(G0,({window:e,disposables:t})=>this.registerListeners(e,t),{window:Nt,disposables:this._subscriptions}))}registerListeners(e,t){t.add(W(e,"keydown",i=>{if(i.defaultPrevented)return;const n=new Pt(i);if(!(n.keyCode===6&&i.repeat)){if(i.altKey&&!this._keyStatus.altKey)this._keyStatus.lastKeyPressed="alt";else if(i.ctrlKey&&!this._keyStatus.ctrlKey)this._keyStatus.lastKeyPressed="ctrl";else if(i.metaKey&&!this._keyStatus.metaKey)this._keyStatus.lastKeyPressed="meta";else if(i.shiftKey&&!this._keyStatus.shiftKey)this._keyStatus.lastKeyPressed="shift";else if(n.keyCode!==6)this._keyStatus.lastKeyPressed=void 0;else return;this._keyStatus.altKey=i.altKey,this._keyStatus.ctrlKey=i.ctrlKey,this._keyStatus.metaKey=i.metaKey,this._keyStatus.shiftKey=i.shiftKey,this._keyStatus.lastKeyPressed&&(this._keyStatus.event=i,this.fire(this._keyStatus))}},!0)),t.add(W(e,"keyup",i=>{i.defaultPrevented||(!i.altKey&&this._keyStatus.altKey?this._keyStatus.lastKeyReleased="alt":!i.ctrlKey&&this._keyStatus.ctrlKey?this._keyStatus.lastKeyReleased="ctrl":!i.metaKey&&this._keyStatus.metaKey?this._keyStatus.lastKeyReleased="meta":!i.shiftKey&&this._keyStatus.shiftKey?this._keyStatus.lastKeyReleased="shift":this._keyStatus.lastKeyReleased=void 0,this._keyStatus.lastKeyPressed!==this._keyStatus.lastKeyReleased&&(this._keyStatus.lastKeyPressed=void 0),this._keyStatus.altKey=i.altKey,this._keyStatus.ctrlKey=i.ctrlKey,this._keyStatus.metaKey=i.metaKey,this._keyStatus.shiftKey=i.shiftKey,this._keyStatus.lastKeyReleased&&(this._keyStatus.event=i,this.fire(this._keyStatus)))},!0)),t.add(W(e.document.body,"mousedown",()=>{this._keyStatus.lastKeyPressed=void 0},!0)),t.add(W(e.document.body,"mouseup",()=>{this._keyStatus.lastKeyPressed=void 0},!0)),t.add(W(e.document.body,"mousemove",i=>{i.buttons&&(this._keyStatus.lastKeyPressed=void 0)},!0)),t.add(W(e,"blur",()=>{this.resetKeyStatus()}))}get keyStatus(){return this._keyStatus}resetKeyStatus(){this.doResetKeyStatus(),this.fire(this._keyStatus)}doResetKeyStatus(){this._keyStatus={altKey:!1,shiftKey:!1,ctrlKey:!1,metaKey:!1}}static getInstance(){return Vd.instance||(Vd.instance=new Vd),Vd.instance}dispose(){super.dispose(),this._subscriptions.dispose()}}class mue extends F{constructor(e,t){super(),this.element=e,this.callbacks=t,this.counter=0,this.dragStartTime=0,this.registerListeners()}registerListeners(){this.callbacks.onDragStart&&this._register(W(this.element,se.DRAG_START,e=>{this.callbacks.onDragStart?.(e)})),this.callbacks.onDrag&&this._register(W(this.element,se.DRAG,e=>{this.callbacks.onDrag?.(e)})),this._register(W(this.element,se.DRAG_ENTER,e=>{this.counter++,this.dragStartTime=e.timeStamp,this.callbacks.onDragEnter?.(e)})),this._register(W(this.element,se.DRAG_OVER,e=>{e.preventDefault(),this.callbacks.onDragOver?.(e,e.timeStamp-this.dragStartTime)})),this._register(W(this.element,se.DRAG_LEAVE,e=>{this.counter--,this.counter===0&&(this.dragStartTime=0,this.callbacks.onDragLeave?.(e))})),this._register(W(this.element,se.DRAG_END,e=>{this.counter=0,this.dragStartTime=0,this.callbacks.onDragEnd?.(e)})),this._register(W(this.element,se.DROP,e=>{this.counter=0,this.dragStartTime=0,this.callbacks.onDrop?.(e)}))}}const _ue=/(?<tag>[\w\-]+)?(?:#(?<id>[\w\-]+))?(?<class>(?:\.(?:[\w\-]+))*)(?:@(?<name>(?:[\w\_])+))?/;function ct(s,...e){let t,i;Array.isArray(e[0])?(t={},i=e[0]):(t=e[0]||{},i=e[1]);const n=_ue.exec(s);if(!n||!n.groups)throw new Error("Bad use of h");const o=n.groups.tag||"div",r=document.createElement(o);n.groups.id&&(r.id=n.groups.id);const a=[];if(n.groups.class)for(const c of n.groups.class.split("."))c!==""&&a.push(c);if(t.className!==void 0)for(const c of t.className.split("."))c!==""&&a.push(c);a.length>0&&(r.className=a.join(" "));const l={};if(n.groups.name&&(l[n.groups.name]=r),i)for(const c of i)ki(c)?r.appendChild(c):typeof c=="string"?r.append(c):"root"in c&&(Object.assign(l,c),r.appendChild(c.root));for(const[c,d]of Object.entries(t))if(c!=="className")if(c==="style")for(const[h,u]of Object.entries(d))r.style.setProperty(Qy(h),typeof u=="number"?u+"px":""+u);else c==="tabIndex"?r.tabIndex=d:r.setAttribute(Qy(c),d.toString());return l.root=r,l}function Qy(s){return s.replace(/([a-z])([A-Z])/g,"$1-$2").toLowerCase()}function wl(s){return s.tagName.toLowerCase()==="input"||s.tagName.toLowerCase()==="textarea"||ki(s)&&!!s.editContext}var Ve;(function(s){function e(n=void 0){return(o,r,a)=>{const l=r.class;delete r.class;const c=r.ref;delete r.ref;const d=r.obsRef;return delete r.obsRef,new wue(o,c,d,n,l,r,a)}}function t(n,o=void 0){const r=e(o);return(a,l)=>r(n,a,l)}s.div=t("div"),s.elem=e(void 0),s.svg=t("svg","http://www.w3.org/2000/svg"),s.svgElem=e("http://www.w3.org/2000/svg");function i(){let n;const o=function(r){n=r};return Object.defineProperty(o,"element",{get(){if(!n)throw new Ce("Make sure the ref is set before accessing the element. Maybe wrong initialization order?");return n}}),o}s.ref=i})(Ve||(Ve={}));class hW{constructor(e,t,i,n,o,r,a){this._deriveds=[],this._element=n?document.createElementNS(n,e):document.createElement(e),t&&t(this._element),i&&this._deriveds.push(j(l=>{i(this),l.store.add({dispose:()=>{i(null)}})})),o&&(FQ(o)?this._deriveds.push(j(this,l=>{Yz(this._element,Zz(o,l))})):Yz(this._element,Zz(o,void 0)));for(const[l,c]of Object.entries(r))if(l==="style")for(const[d,h]of Object.entries(c)){const u=Qy(d);zf(h)?this._deriveds.push(Jn({owner:this,debugName:()=>`set.style.${u}`},g=>{this._element.style.setProperty(u,Xz(h.read(g)))})):this._element.style.setProperty(u,Xz(h))}else l==="tabIndex"?zf(c)?this._deriveds.push(j(this,d=>{this._element.tabIndex=c.read(d)})):this._element.tabIndex=c:l.startsWith("on")?this._element[l]=c:zf(c)?this._deriveds.push(Jn({owner:this,debugName:()=>`set.${l}`},d=>{Qz(this._element,l,c.read(d))})):Qz(this._element,l,c);if(a){let l=function(d,h){return zf(h)?l(d,h.read(d)):Array.isArray(h)?h.flatMap(u=>l(d,u)):h instanceof hW?(d&&h.readEffect(d),[h._element]):h?[h]:[]};const c=j(this,d=>{this._element.replaceChildren(...l(d,a))});this._deriveds.push(c),BQ(a)||c.get()}}readEffect(e){for(const t of this._deriveds)t.read(e)}keepUpdated(e){return j(t=>{this.readEffect(t)}).recomputeInitiallyAndOnChange(e),this}toDisposableLiveElement(){const e=new U;return this.keepUpdated(e),new bue(this._element,e)}}function Yz(s,e){AQ(s)?s.setAttribute("class",e):s.className=e}function PQ(s,e,t){if(zf(s)){t(s.read(e));return}if(Array.isArray(s)){for(const i of s)PQ(i,e,t);return}t(s)}function Zz(s,e){let t="";return PQ(s,e,i=>{i&&(t.length===0?t=i:t+=" "+i)}),t}function FQ(s){return zf(s)?!0:Array.isArray(s)?s.some(e=>FQ(e)):!1}function Xz(s){return typeof s=="number"?s+"px":s}function BQ(s){return zf(s)?!0:Array.isArray(s)?s.some(e=>BQ(e)):!1}class bue{constructor(e,t){this.element=e,this._disposable=t}dispose(){this._disposable.dispose()}}class wue extends hW{constructor(){super(...arguments),this._isHovered=void 0,this._didMouseMoveDuringHover=void 0}get element(){return this._element}get isHovered(){if(!this._isHovered){const e=Re("hovered",!1);this._element.addEventListener("mouseenter",t=>e.set(!0,void 0)),this._element.addEventListener("mouseleave",t=>e.set(!1,void 0)),this._isHovered=e}return this._isHovered}get didMouseMoveDuringHover(){if(!this._didMouseMoveDuringHover){let e=!1;const t=Re("didMouseMoveDuringHover",!1);this._element.addEventListener("mouseenter",i=>{e=!0}),this._element.addEventListener("mousemove",i=>{e&&t.set(!0,void 0)}),this._element.addEventListener("mouseleave",i=>{e=!1,t.set(!1,void 0)}),this._didMouseMoveDuringHover=t}return this._didMouseMoveDuringHover}}function Qz(s,e,t){t==null?s.removeAttribute(Qy(e)):s.setAttribute(Qy(e),String(t))}function zf(s){return!!s&&s.read!==void 0&&s.reportChanges!==void 0}const Jz=2e4;let uf,yk,LP,Sk,kP;function vue(s){uf=document.createElement("div"),uf.className="monaco-aria-container";const e=()=>{const i=document.createElement("div");return i.className="monaco-alert",i.setAttribute("role","alert"),i.setAttribute("aria-atomic","true"),uf.appendChild(i),i};yk=e(),LP=e();const t=()=>{const i=document.createElement("div");return i.className="monaco-status",i.setAttribute("aria-live","polite"),i.setAttribute("aria-atomic","true"),uf.appendChild(i),i};Sk=t(),kP=t(),s.appendChild(uf)}function Js(s){uf&&(yk.textContent!==s?(Pn(LP),kE(yk,s)):(Pn(yk),kE(LP,s)))}function xl(s){uf&&(Sk.textContent!==s?(Pn(kP),kE(Sk,s)):(Pn(Sk),kE(kP,s)))}function kE(s,e){Pn(s),e.length>Jz&&(e=e.substr(0,Jz)),s.textContent=e,s.style.visibility="hidden",s.style.visibility="visible"}var pl;(function(s){s.serviceIds=new Map,s.DI_TARGET="$di$target",s.DI_DEPENDENCIES="$di$dependencies";function e(t){return t[s.DI_DEPENDENCIES]||[]}s.getServiceDependencies=e})(pl||(pl={}));const _e=qe("instantiationService");function Cue(s,e,t){e[pl.DI_TARGET]===e?e[pl.DI_DEPENDENCIES].push({id:s,index:t}):(e[pl.DI_DEPENDENCIES]=[{id:s,index:t}],e[pl.DI_TARGET]=e)}function qe(s){if(pl.serviceIds.has(s))return pl.serviceIds.get(s);const e=function(t,i,n){if(arguments.length!==3)throw new Error("@IServiceName-decorator can only be used to decorate a parameter");Cue(e,t,n)};return e.toString=()=>s,pl.serviceIds.set(s,e),e}const ht=qe("codeEditorService");let A=class ef{constructor(e,t){this.lineNumber=e,this.column=t}with(e=this.lineNumber,t=this.column){return e===this.lineNumber&&t===this.column?this:new ef(e,t)}delta(e=0,t=0){return this.with(Math.max(1,this.lineNumber+e),Math.max(1,this.column+t))}equals(e){return ef.equals(this,e)}static equals(e,t){return!e&&!t?!0:!!e&&!!t&&e.lineNumber===t.lineNumber&&e.column===t.column}isBefore(e){return ef.isBefore(this,e)}static isBefore(e,t){return e.lineNumber<t.lineNumber?!0:t.lineNumber<e.lineNumber?!1:e.column<t.column}isBeforeOrEqual(e){return ef.isBeforeOrEqual(this,e)}static isBeforeOrEqual(e,t){return e.lineNumber<t.lineNumber?!0:t.lineNumber<e.lineNumber?!1:e.column<=t.column}static compare(e,t){const i=e.lineNumber|0,n=t.lineNumber|0;if(i===n){const o=e.column|0,r=t.column|0;return o-r}return i-n}clone(){return new ef(this.lineNumber,this.column)}toString(){return"("+this.lineNumber+","+this.column+")"}static lift(e){return new ef(e.lineNumber,e.column)}static isIPosition(e){return!!e&&typeof e.lineNumber=="number"&&typeof e.column=="number"}toJSON(){return{lineNumber:this.lineNumber,column:this.column}}};const hi=qe("modelService"),Fs=qe("textModelService");class Cr extends F{get onDidChange(){return this._onDidChange.event}constructor(e,t="",i="",n=!0,o){super(),this._onDidChange=this._register(new M),this._enabled=!0,this._id=e,this._label=t,this._cssClass=i,this._enabled=n,this._actionCallback=o}get id(){return this._id}get label(){return this._label}set label(e){this._setLabel(e)}_setLabel(e){this._label!==e&&(this._label=e,this._onDidChange.fire({label:e}))}get tooltip(){return this._tooltip||""}set tooltip(e){this._setTooltip(e)}_setTooltip(e){this._tooltip!==e&&(this._tooltip=e,this._onDidChange.fire({tooltip:e}))}get class(){return this._cssClass}set class(e){this._setClass(e)}_setClass(e){this._cssClass!==e&&(this._cssClass=e,this._onDidChange.fire({class:e}))}get enabled(){return this._enabled}set enabled(e){this._setEnabled(e)}_setEnabled(e){this._enabled!==e&&(this._enabled=e,this._onDidChange.fire({enabled:e}))}get checked(){return this._checked}set checked(e){this._setChecked(e)}_setChecked(e){this._checked!==e&&(this._checked=e,this._onDidChange.fire({checked:e}))}async run(e,t){this._actionCallback&&await this._actionCallback(e)}}class Lm extends F{constructor(){super(...arguments),this._onWillRun=this._register(new M),this._onDidRun=this._register(new M)}get onWillRun(){return this._onWillRun.event}get onDidRun(){return this._onDidRun.event}async run(e,t){if(!e.enabled)return;this._onWillRun.fire({action:e});let i;try{await this.runAction(e,t)}catch(n){i=n}this._onDidRun.fire({action:e,error:i})}async runAction(e,t){await e.run(t)}}const yy=class yy{constructor(){this.id=yy.ID,this.label="",this.tooltip="",this.class="separator",this.enabled=!1,this.checked=!1}static join(...e){let t=[];for(const i of e)i.length&&(t.length?t=[...t,new yy,...i]:t=i);return t}async run(){}};yy.ID="vs.actions.separator";let tn=yy;class ov{get actions(){return this._actions}constructor(e,t,i,n){this.tooltip="",this.enabled=!0,this.checked=void 0,this.id=e,this.label=t,this.class=n,this._actions=i}async run(){}}const VT=class VT extends Cr{constructor(){super(VT.ID,m(28,"(empty)"),void 0,!1)}};VT.ID="vs.actions.empty";let EP=VT;function rp(s){return{id:s.id,label:s.label,tooltip:s.tooltip??s.label,class:s.class,enabled:s.enabled??!0,checked:s.checked,run:async(...e)=>s.run(...e)}}const IP=Object.create(null);function S(s,e){if(bn(e)){const t=IP[e];if(t===void 0)throw new Error(`${s} references an unknown codicon: ${e}`);e=t}return IP[s]=e,{id:s}}function WQ(){return IP}const yue={add:S("add",6e4),plus:S("plus",6e4),gistNew:S("gist-new",6e4),repoCreate:S("repo-create",6e4),lightbulb:S("lightbulb",60001),lightBulb:S("light-bulb",60001),repo:S("repo",60002),repoDelete:S("repo-delete",60002),gistFork:S("gist-fork",60003),repoForked:S("repo-forked",60003),gitPullRequest:S("git-pull-request",60004),gitPullRequestAbandoned:S("git-pull-request-abandoned",60004),recordKeys:S("record-keys",60005),keyboard:S("keyboard",60005),tag:S("tag",60006),gitPullRequestLabel:S("git-pull-request-label",60006),tagAdd:S("tag-add",60006),tagRemove:S("tag-remove",60006),person:S("person",60007),personFollow:S("person-follow",60007),personOutline:S("person-outline",60007),personFilled:S("person-filled",60007),sourceControl:S("source-control",60008),mirror:S("mirror",60009),mirrorPublic:S("mirror-public",60009),star:S("star",60010),starAdd:S("star-add",60010),starDelete:S("star-delete",60010),starEmpty:S("star-empty",60010),comment:S("comment",60011),commentAdd:S("comment-add",60011),alert:S("alert",60012),warning:S("warning",60012),search:S("search",60013),searchSave:S("search-save",60013),logOut:S("log-out",60014),signOut:S("sign-out",60014),logIn:S("log-in",60015),signIn:S("sign-in",60015),eye:S("eye",60016),eyeUnwatch:S("eye-unwatch",60016),eyeWatch:S("eye-watch",60016),circleFilled:S("circle-filled",60017),primitiveDot:S("primitive-dot",60017),closeDirty:S("close-dirty",60017),debugBreakpoint:S("debug-breakpoint",60017),debugBreakpointDisabled:S("debug-breakpoint-disabled",60017),debugHint:S("debug-hint",60017),terminalDecorationSuccess:S("terminal-decoration-success",60017),primitiveSquare:S("primitive-square",60018),edit:S("edit",60019),pencil:S("pencil",60019),info:S("info",60020),issueOpened:S("issue-opened",60020),gistPrivate:S("gist-private",60021),gitForkPrivate:S("git-fork-private",60021),lock:S("lock",60021),mirrorPrivate:S("mirror-private",60021),close:S("close",60022),removeClose:S("remove-close",60022),x:S("x",60022),repoSync:S("repo-sync",60023),sync:S("sync",60023),clone:S("clone",60024),desktopDownload:S("desktop-download",60024),beaker:S("beaker",60025),microscope:S("microscope",60025),vm:S("vm",60026),deviceDesktop:S("device-desktop",60026),file:S("file",60027),more:S("more",60028),ellipsis:S("ellipsis",60028),kebabHorizontal:S("kebab-horizontal",60028),mailReply:S("mail-reply",60029),reply:S("reply",60029),organization:S("organization",60030),organizationFilled:S("organization-filled",60030),organizationOutline:S("organization-outline",60030),newFile:S("new-file",60031),fileAdd:S("file-add",60031),newFolder:S("new-folder",60032),fileDirectoryCreate:S("file-directory-create",60032),trash:S("trash",60033),trashcan:S("trashcan",60033),history:S("history",60034),clock:S("clock",60034),folder:S("folder",60035),fileDirectory:S("file-directory",60035),symbolFolder:S("symbol-folder",60035),logoGithub:S("logo-github",60036),markGithub:S("mark-github",60036),github:S("github",60036),terminal:S("terminal",60037),console:S("console",60037),repl:S("repl",60037),zap:S("zap",60038),symbolEvent:S("symbol-event",60038),error:S("error",60039),stop:S("stop",60039),variable:S("variable",60040),symbolVariable:S("symbol-variable",60040),array:S("array",60042),symbolArray:S("symbol-array",60042),symbolModule:S("symbol-module",60043),symbolPackage:S("symbol-package",60043),symbolNamespace:S("symbol-namespace",60043),symbolObject:S("symbol-object",60043),symbolMethod:S("symbol-method",60044),symbolFunction:S("symbol-function",60044),symbolConstructor:S("symbol-constructor",60044),symbolBoolean:S("symbol-boolean",60047),symbolNull:S("symbol-null",60047),symbolNumeric:S("symbol-numeric",60048),symbolNumber:S("symbol-number",60048),symbolStructure:S("symbol-structure",60049),symbolStruct:S("symbol-struct",60049),symbolParameter:S("symbol-parameter",60050),symbolTypeParameter:S("symbol-type-parameter",60050),symbolKey:S("symbol-key",60051),symbolText:S("symbol-text",60051),symbolReference:S("symbol-reference",60052),goToFile:S("go-to-file",60052),symbolEnum:S("symbol-enum",60053),symbolValue:S("symbol-value",60053),symbolRuler:S("symbol-ruler",60054),symbolUnit:S("symbol-unit",60054),activateBreakpoints:S("activate-breakpoints",60055),archive:S("archive",60056),arrowBoth:S("arrow-both",60057),arrowDown:S("arrow-down",60058),arrowLeft:S("arrow-left",60059),arrowRight:S("arrow-right",60060),arrowSmallDown:S("arrow-small-down",60061),arrowSmallLeft:S("arrow-small-left",60062),arrowSmallRight:S("arrow-small-right",60063),arrowSmallUp:S("arrow-small-up",60064),arrowUp:S("arrow-up",60065),bell:S("bell",60066),bold:S("bold",60067),book:S("book",60068),bookmark:S("bookmark",60069),debugBreakpointConditionalUnverified:S("debug-breakpoint-conditional-unverified",60070),debugBreakpointConditional:S("debug-breakpoint-conditional",60071),debugBreakpointConditionalDisabled:S("debug-breakpoint-conditional-disabled",60071),debugBreakpointDataUnverified:S("debug-breakpoint-data-unverified",60072),debugBreakpointData:S("debug-breakpoint-data",60073),debugBreakpointDataDisabled:S("debug-breakpoint-data-disabled",60073),debugBreakpointLogUnverified:S("debug-breakpoint-log-unverified",60074),debugBreakpointLog:S("debug-breakpoint-log",60075),debugBreakpointLogDisabled:S("debug-breakpoint-log-disabled",60075),briefcase:S("briefcase",60076),broadcast:S("broadcast",60077),browser:S("browser",60078),bug:S("bug",60079),calendar:S("calendar",60080),caseSensitive:S("case-sensitive",60081),check:S("check",60082),checklist:S("checklist",60083),chevronDown:S("chevron-down",60084),chevronLeft:S("chevron-left",60085),chevronRight:S("chevron-right",60086),chevronUp:S("chevron-up",60087),chromeClose:S("chrome-close",60088),chromeMaximize:S("chrome-maximize",60089),chromeMinimize:S("chrome-minimize",60090),chromeRestore:S("chrome-restore",60091),circleOutline:S("circle-outline",60092),circle:S("circle",60092),debugBreakpointUnverified:S("debug-breakpoint-unverified",60092),terminalDecorationIncomplete:S("terminal-decoration-incomplete",60092),circleSlash:S("circle-slash",60093),circuitBoard:S("circuit-board",60094),clearAll:S("clear-all",60095),clippy:S("clippy",60096),closeAll:S("close-all",60097),cloudDownload:S("cloud-download",60098),cloudUpload:S("cloud-upload",60099),code:S("code",60100),collapseAll:S("collapse-all",60101),colorMode:S("color-mode",60102),commentDiscussion:S("comment-discussion",60103),creditCard:S("credit-card",60105),dash:S("dash",60108),dashboard:S("dashboard",60109),database:S("database",60110),debugContinue:S("debug-continue",60111),debugDisconnect:S("debug-disconnect",60112),debugPause:S("debug-pause",60113),debugRestart:S("debug-restart",60114),debugStart:S("debug-start",60115),debugStepInto:S("debug-step-into",60116),debugStepOut:S("debug-step-out",60117),debugStepOver:S("debug-step-over",60118),debugStop:S("debug-stop",60119),debug:S("debug",60120),deviceCameraVideo:S("device-camera-video",60121),deviceCamera:S("device-camera",60122),deviceMobile:S("device-mobile",60123),diffAdded:S("diff-added",60124),diffIgnored:S("diff-ignored",60125),diffModified:S("diff-modified",60126),diffRemoved:S("diff-removed",60127),diffRenamed:S("diff-renamed",60128),diff:S("diff",60129),diffSidebyside:S("diff-sidebyside",60129),discard:S("discard",60130),editorLayout:S("editor-layout",60131),emptyWindow:S("empty-window",60132),exclude:S("exclude",60133),extensions:S("extensions",60134),eyeClosed:S("eye-closed",60135),fileBinary:S("file-binary",60136),fileCode:S("file-code",60137),fileMedia:S("file-media",60138),filePdf:S("file-pdf",60139),fileSubmodule:S("file-submodule",60140),fileSymlinkDirectory:S("file-symlink-directory",60141),fileSymlinkFile:S("file-symlink-file",60142),fileZip:S("file-zip",60143),files:S("files",60144),filter:S("filter",60145),flame:S("flame",60146),foldDown:S("fold-down",60147),foldUp:S("fold-up",60148),fold:S("fold",60149),folderActive:S("folder-active",60150),folderOpened:S("folder-opened",60151),gear:S("gear",60152),gift:S("gift",60153),gistSecret:S("gist-secret",60154),gist:S("gist",60155),gitCommit:S("git-commit",60156),gitCompare:S("git-compare",60157),compareChanges:S("compare-changes",60157),gitMerge:S("git-merge",60158),githubAction:S("github-action",60159),githubAlt:S("github-alt",60160),globe:S("globe",60161),grabber:S("grabber",60162),graph:S("graph",60163),gripper:S("gripper",60164),heart:S("heart",60165),home:S("home",60166),horizontalRule:S("horizontal-rule",60167),hubot:S("hubot",60168),inbox:S("inbox",60169),issueReopened:S("issue-reopened",60171),issues:S("issues",60172),italic:S("italic",60173),jersey:S("jersey",60174),json:S("json",60175),kebabVertical:S("kebab-vertical",60176),key:S("key",60177),law:S("law",60178),lightbulbAutofix:S("lightbulb-autofix",60179),linkExternal:S("link-external",60180),link:S("link",60181),listOrdered:S("list-ordered",60182),listUnordered:S("list-unordered",60183),liveShare:S("live-share",60184),loading:S("loading",60185),location:S("location",60186),mailRead:S("mail-read",60187),mail:S("mail",60188),markdown:S("markdown",60189),megaphone:S("megaphone",60190),mention:S("mention",60191),milestone:S("milestone",60192),gitPullRequestMilestone:S("git-pull-request-milestone",60192),mortarBoard:S("mortar-board",60193),move:S("move",60194),multipleWindows:S("multiple-windows",60195),mute:S("mute",60196),noNewline:S("no-newline",60197),note:S("note",60198),octoface:S("octoface",60199),openPreview:S("open-preview",60200),package:S("package",60201),paintcan:S("paintcan",60202),pin:S("pin",60203),play:S("play",60204),run:S("run",60204),plug:S("plug",60205),preserveCase:S("preserve-case",60206),preview:S("preview",60207),project:S("project",60208),pulse:S("pulse",60209),question:S("question",60210),quote:S("quote",60211),radioTower:S("radio-tower",60212),reactions:S("reactions",60213),references:S("references",60214),refresh:S("refresh",60215),regex:S("regex",60216),remoteExplorer:S("remote-explorer",60217),remote:S("remote",60218),remove:S("remove",60219),replaceAll:S("replace-all",60220),replace:S("replace",60221),repoClone:S("repo-clone",60222),repoForcePush:S("repo-force-push",60223),repoPull:S("repo-pull",60224),repoPush:S("repo-push",60225),report:S("report",60226),requestChanges:S("request-changes",60227),rocket:S("rocket",60228),rootFolderOpened:S("root-folder-opened",60229),rootFolder:S("root-folder",60230),rss:S("rss",60231),ruby:S("ruby",60232),saveAll:S("save-all",60233),saveAs:S("save-as",60234),save:S("save",60235),screenFull:S("screen-full",60236),screenNormal:S("screen-normal",60237),searchStop:S("search-stop",60238),server:S("server",60240),settingsGear:S("settings-gear",60241),settings:S("settings",60242),shield:S("shield",60243),smiley:S("smiley",60244),sortPrecedence:S("sort-precedence",60245),splitHorizontal:S("split-horizontal",60246),splitVertical:S("split-vertical",60247),squirrel:S("squirrel",60248),starFull:S("star-full",60249),starHalf:S("star-half",60250),symbolClass:S("symbol-class",60251),symbolColor:S("symbol-color",60252),symbolConstant:S("symbol-constant",60253),symbolEnumMember:S("symbol-enum-member",60254),symbolField:S("symbol-field",60255),symbolFile:S("symbol-file",60256),symbolInterface:S("symbol-interface",60257),symbolKeyword:S("symbol-keyword",60258),symbolMisc:S("symbol-misc",60259),symbolOperator:S("symbol-operator",60260),symbolProperty:S("symbol-property",60261),wrench:S("wrench",60261),wrenchSubaction:S("wrench-subaction",60261),symbolSnippet:S("symbol-snippet",60262),tasklist:S("tasklist",60263),telescope:S("telescope",60264),textSize:S("text-size",60265),threeBars:S("three-bars",60266),thumbsdown:S("thumbsdown",60267),thumbsup:S("thumbsup",60268),tools:S("tools",60269),triangleDown:S("triangle-down",60270),triangleLeft:S("triangle-left",60271),triangleRight:S("triangle-right",60272),triangleUp:S("triangle-up",60273),twitter:S("twitter",60274),unfold:S("unfold",60275),unlock:S("unlock",60276),unmute:S("unmute",60277),unverified:S("unverified",60278),verified:S("verified",60279),versions:S("versions",60280),vmActive:S("vm-active",60281),vmOutline:S("vm-outline",60282),vmRunning:S("vm-running",60283),watch:S("watch",60284),whitespace:S("whitespace",60285),wholeWord:S("whole-word",60286),window:S("window",60287),wordWrap:S("word-wrap",60288),zoomIn:S("zoom-in",60289),zoomOut:S("zoom-out",60290),listFilter:S("list-filter",60291),listFlat:S("list-flat",60292),listSelection:S("list-selection",60293),selection:S("selection",60293),listTree:S("list-tree",60294),debugBreakpointFunctionUnverified:S("debug-breakpoint-function-unverified",60295),debugBreakpointFunction:S("debug-breakpoint-function",60296),debugBreakpointFunctionDisabled:S("debug-breakpoint-function-disabled",60296),debugStackframeActive:S("debug-stackframe-active",60297),circleSmallFilled:S("circle-small-filled",60298),debugStackframeDot:S("debug-stackframe-dot",60298),terminalDecorationMark:S("terminal-decoration-mark",60298),debugStackframe:S("debug-stackframe",60299),debugStackframeFocused:S("debug-stackframe-focused",60299),debugBreakpointUnsupported:S("debug-breakpoint-unsupported",60300),symbolString:S("symbol-string",60301),debugReverseContinue:S("debug-reverse-continue",60302),debugStepBack:S("debug-step-back",60303),debugRestartFrame:S("debug-restart-frame",60304),debugAlt:S("debug-alt",60305),callIncoming:S("call-incoming",60306),callOutgoing:S("call-outgoing",60307),menu:S("menu",60308),expandAll:S("expand-all",60309),feedback:S("feedback",60310),gitPullRequestReviewer:S("git-pull-request-reviewer",60310),groupByRefType:S("group-by-ref-type",60311),ungroupByRefType:S("ungroup-by-ref-type",60312),account:S("account",60313),gitPullRequestAssignee:S("git-pull-request-assignee",60313),bellDot:S("bell-dot",60314),debugConsole:S("debug-console",60315),library:S("library",60316),output:S("output",60317),runAll:S("run-all",60318),syncIgnored:S("sync-ignored",60319),pinned:S("pinned",60320),githubInverted:S("github-inverted",60321),serverProcess:S("server-process",60322),serverEnvironment:S("server-environment",60323),pass:S("pass",60324),issueClosed:S("issue-closed",60324),stopCircle:S("stop-circle",60325),playCircle:S("play-circle",60326),record:S("record",60327),debugAltSmall:S("debug-alt-small",60328),vmConnect:S("vm-connect",60329),cloud:S("cloud",60330),merge:S("merge",60331),export:S("export",60332),graphLeft:S("graph-left",60333),magnet:S("magnet",60334),notebook:S("notebook",60335),redo:S("redo",60336),checkAll:S("check-all",60337),pinnedDirty:S("pinned-dirty",60338),passFilled:S("pass-filled",60339),circleLargeFilled:S("circle-large-filled",60340),circleLarge:S("circle-large",60341),circleLargeOutline:S("circle-large-outline",60341),combine:S("combine",60342),gather:S("gather",60342),table:S("table",60343),variableGroup:S("variable-group",60344),typeHierarchy:S("type-hierarchy",60345),typeHierarchySub:S("type-hierarchy-sub",60346),typeHierarchySuper:S("type-hierarchy-super",60347),gitPullRequestCreate:S("git-pull-request-create",60348),runAbove:S("run-above",60349),runBelow:S("run-below",60350),notebookTemplate:S("notebook-template",60351),debugRerun:S("debug-rerun",60352),workspaceTrusted:S("workspace-trusted",60353),workspaceUntrusted:S("workspace-untrusted",60354),workspaceUnknown:S("workspace-unknown",60355),terminalCmd:S("terminal-cmd",60356),terminalDebian:S("terminal-debian",60357),terminalLinux:S("terminal-linux",60358),terminalPowershell:S("terminal-powershell",60359),terminalTmux:S("terminal-tmux",60360),terminalUbuntu:S("terminal-ubuntu",60361),terminalBash:S("terminal-bash",60362),arrowSwap:S("arrow-swap",60363),copy:S("copy",60364),personAdd:S("person-add",60365),filterFilled:S("filter-filled",60366),wand:S("wand",60367),debugLineByLine:S("debug-line-by-line",60368),inspect:S("inspect",60369),layers:S("layers",60370),layersDot:S("layers-dot",60371),layersActive:S("layers-active",60372),compass:S("compass",60373),compassDot:S("compass-dot",60374),compassActive:S("compass-active",60375),azure:S("azure",60376),issueDraft:S("issue-draft",60377),gitPullRequestClosed:S("git-pull-request-closed",60378),gitPullRequestDraft:S("git-pull-request-draft",60379),debugAll:S("debug-all",60380),debugCoverage:S("debug-coverage",60381),runErrors:S("run-errors",60382),folderLibrary:S("folder-library",60383),debugContinueSmall:S("debug-continue-small",60384),beakerStop:S("beaker-stop",60385),graphLine:S("graph-line",60386),graphScatter:S("graph-scatter",60387),pieChart:S("pie-chart",60388),bracket:S("bracket",60175),bracketDot:S("bracket-dot",60389),bracketError:S("bracket-error",60390),lockSmall:S("lock-small",60391),azureDevops:S("azure-devops",60392),verifiedFilled:S("verified-filled",60393),newline:S("newline",60394),layout:S("layout",60395),layoutActivitybarLeft:S("layout-activitybar-left",60396),layoutActivitybarRight:S("layout-activitybar-right",60397),layoutPanelLeft:S("layout-panel-left",60398),layoutPanelCenter:S("layout-panel-center",60399),layoutPanelJustify:S("layout-panel-justify",60400),layoutPanelRight:S("layout-panel-right",60401),layoutPanel:S("layout-panel",60402),layoutSidebarLeft:S("layout-sidebar-left",60403),layoutSidebarRight:S("layout-sidebar-right",60404),layoutStatusbar:S("layout-statusbar",60405),layoutMenubar:S("layout-menubar",60406),layoutCentered:S("layout-centered",60407),target:S("target",60408),indent:S("indent",60409),recordSmall:S("record-small",60410),errorSmall:S("error-small",60411),terminalDecorationError:S("terminal-decoration-error",60411),arrowCircleDown:S("arrow-circle-down",60412),arrowCircleLeft:S("arrow-circle-left",60413),arrowCircleRight:S("arrow-circle-right",60414),arrowCircleUp:S("arrow-circle-up",60415),layoutSidebarRightOff:S("layout-sidebar-right-off",60416),layoutPanelOff:S("layout-panel-off",60417),layoutSidebarLeftOff:S("layout-sidebar-left-off",60418),blank:S("blank",60419),heartFilled:S("heart-filled",60420),map:S("map",60421),mapHorizontal:S("map-horizontal",60421),foldHorizontal:S("fold-horizontal",60421),mapFilled:S("map-filled",60422),mapHorizontalFilled:S("map-horizontal-filled",60422),foldHorizontalFilled:S("fold-horizontal-filled",60422),circleSmall:S("circle-small",60423),bellSlash:S("bell-slash",60424),bellSlashDot:S("bell-slash-dot",60425),commentUnresolved:S("comment-unresolved",60426),gitPullRequestGoToChanges:S("git-pull-request-go-to-changes",60427),gitPullRequestNewChanges:S("git-pull-request-new-changes",60428),searchFuzzy:S("search-fuzzy",60429),commentDraft:S("comment-draft",60430),send:S("send",60431),sparkle:S("sparkle",60432),insert:S("insert",60433),mic:S("mic",60434),thumbsdownFilled:S("thumbsdown-filled",60435),thumbsupFilled:S("thumbsup-filled",60436),coffee:S("coffee",60437),snake:S("snake",60438),game:S("game",60439),vr:S("vr",60440),chip:S("chip",60441),piano:S("piano",60442),music:S("music",60443),micFilled:S("mic-filled",60444),repoFetch:S("repo-fetch",60445),copilot:S("copilot",60446),lightbulbSparkle:S("lightbulb-sparkle",60447),robot:S("robot",60448),sparkleFilled:S("sparkle-filled",60449),diffSingle:S("diff-single",60450),diffMultiple:S("diff-multiple",60451),surroundWith:S("surround-with",60452),share:S("share",60453),gitStash:S("git-stash",60454),gitStashApply:S("git-stash-apply",60455),gitStashPop:S("git-stash-pop",60456),vscode:S("vscode",60457),vscodeInsiders:S("vscode-insiders",60458),codeOss:S("code-oss",60459),runCoverage:S("run-coverage",60460),runAllCoverage:S("run-all-coverage",60461),coverage:S("coverage",60462),githubProject:S("github-project",60463),mapVertical:S("map-vertical",60464),foldVertical:S("fold-vertical",60464),mapVerticalFilled:S("map-vertical-filled",60465),foldVerticalFilled:S("fold-vertical-filled",60465),goToSearch:S("go-to-search",60466),percentage:S("percentage",60467),sortPercentage:S("sort-percentage",60467),attach:S("attach",60468),goToEditingSession:S("go-to-editing-session",60469),editSession:S("edit-session",60470),codeReview:S("code-review",60471),copilotWarning:S("copilot-warning",60472),python:S("python",60473),copilotLarge:S("copilot-large",60474),copilotWarningLarge:S("copilot-warning-large",60475),keyboardTab:S("keyboard-tab",60476),copilotBlocked:S("copilot-blocked",60477),copilotNotConnected:S("copilot-not-connected",60478),flag:S("flag",60479),lightbulbEmpty:S("lightbulb-empty",60480),symbolMethodArrow:S("symbol-method-arrow",60481),copilotUnavailable:S("copilot-unavailable",60482),repoPinned:S("repo-pinned",60483),keyboardTabAbove:S("keyboard-tab-above",60484),keyboardTabBelow:S("keyboard-tab-below",60485),gitPullRequestDone:S("git-pull-request-done",60486),mcp:S("mcp",60487),extensionsLarge:S("extensions-large",60488),layoutPanelDock:S("layout-panel-dock",60489),layoutSidebarLeftDock:S("layout-sidebar-left-dock",60490),layoutSidebarRightDock:S("layout-sidebar-right-dock",60491),copilotInProgress:S("copilot-in-progress",60492),copilotError:S("copilot-error",60493),copilotSuccess:S("copilot-success",60494),chatSparkle:S("chat-sparkle",60495),searchSparkle:S("search-sparkle",60496),editSparkle:S("edit-sparkle",60497),copilotSnooze:S("copilot-snooze",60498),sendToRemoteAgent:S("send-to-remote-agent",60499),commentDiscussionSparkle:S("comment-discussion-sparkle",60500),chatSparkleWarning:S("chat-sparkle-warning",60501),chatSparkleError:S("chat-sparkle-error",60502),collection:S("collection",60503),newCollection:S("new-collection",60504),thinking:S("thinking",60505),build:S("build",60506),commentDiscussionQuote:S("comment-discussion-quote",60507),cursor:S("cursor",60508),eraser:S("eraser",60509),fileText:S("file-text",60510),gitLens:S("git-lens",60511),quotes:S("quotes",60512),rename:S("rename",60513),runWithDeps:S("run-with-deps",60514),debugConnected:S("debug-connected",60515),strikethrough:S("strikethrough",60516),openInProduct:S("open-in-product",60517),indexZero:S("index-zero",60518),agent:S("agent",60519),editCode:S("edit-code",60520),repoSelected:S("repo-selected",60521),skip:S("skip",60522),mergeInto:S("merge-into",60523),gitBranchChanges:S("git-branch-changes",60524),gitBranchStagedChanges:S("git-branch-staged-changes",60525),gitBranchConflicts:S("git-branch-conflicts",60526),gitBranch:S("git-branch",60527),gitBranchCreate:S("git-branch-create",60527),gitBranchDelete:S("git-branch-delete",60527),searchLarge:S("search-large",60528),terminalGitBash:S("terminal-git-bash",60529)},Sue={dialogError:S("dialog-error","error"),dialogWarning:S("dialog-warning","warning"),dialogInfo:S("dialog-info","info"),dialogClose:S("dialog-close","close"),treeItemExpanded:S("tree-item-expanded","chevron-down"),treeFilterOnTypeOn:S("tree-filter-on-type-on","list-filter"),treeFilterOnTypeOff:S("tree-filter-on-type-off","list-selection"),treeFilterClear:S("tree-filter-clear","close"),treeItemLoading:S("tree-item-loading","loading"),menuSelection:S("menu-selection","check"),menuSubmenu:S("menu-submenu","chevron-right"),menuBarMore:S("menubar-more","more"),scrollbarButtonLeft:S("scrollbar-button-left","triangle-left"),scrollbarButtonRight:S("scrollbar-button-right","triangle-right"),scrollbarButtonUp:S("scrollbar-button-up","triangle-up"),scrollbarButtonDown:S("scrollbar-button-down","triangle-down"),toolBarMore:S("toolbar-more","more"),quickInputBack:S("quick-input-back","arrow-left"),dropDownButton:S("drop-down-button",60084),symbolCustomColor:S("symbol-customcolor",60252),exportIcon:S("export",60332),workspaceUnspecified:S("workspace-unspecified",60355),newLine:S("newline",60394),thumbsDownFilled:S("thumbsdown-filled",60435),thumbsUpFilled:S("thumbsup-filled",60436),gitFetch:S("git-fetch",60445),lightbulbSparkleAutofix:S("lightbulb-sparkle-autofix",60447),debugBreakpointPending:S("debug-breakpoint-pending",60377)},Z={...yue,...Sue};var DP;(function(s){function e(t){return!!t&&typeof t=="object"&&typeof t.id=="string"}s.isThemeColor=e})(DP||(DP={}));var xe;(function(s){s.iconNameSegment="[A-Za-z0-9]+",s.iconNameExpression="[A-Za-z0-9-]+",s.iconModifierExpression="~[A-Za-z]+",s.iconNameCharacter="[A-Za-z0-9~-]";const e=new RegExp(`^(${s.iconNameExpression})(${s.iconModifierExpression})?$`);function t(f){const p=e.exec(f.id);if(!p)return t(Z.error);const[,_,b]=p,w=["codicon","codicon-"+_];return b&&w.push("codicon-modifier-"+b.substring(1)),w}s.asClassNameArray=t;function i(f){return t(f).join(" ")}s.asClassName=i;function n(f){return"."+t(f).join(".")}s.asCSSSelector=n;function o(f){return!!f&&typeof f=="object"&&typeof f.id=="string"&&(typeof f.color>"u"||DP.isThemeColor(f.color))}s.isThemeIcon=o;const r=new RegExp(`^\\$\\((${s.iconNameExpression}(?:${s.iconModifierExpression})?)\\)$`);function a(f){const p=r.exec(f);if(!p)return;const[,_]=p;return{id:_}}s.fromString=a;function l(f){return{id:f}}s.fromId=l;function c(f,p){let _=f.id;const b=_.lastIndexOf("~");return b!==-1&&(_=_.substring(0,b)),p&&(_=`${_}~${p}`),{id:_}}s.modify=c;function d(f){const p=f.id.lastIndexOf("~");if(p!==-1)return f.id.substring(p+1)}s.getModifier=d;function h(f,p){return f.id===p.id&&f.color?.id===p.color?.id}s.isEqual=h;function u(f){return f?.id===Z.file.id}s.isFile=u;function g(f){return f?.id===Z.folder.id}s.isFolder=g})(xe||(xe={}));const Gt=qe("commandService"),at=new class{constructor(){this._commands=new Map,this._onDidRegisterCommand=new M,this.onDidRegisterCommand=this._onDidRegisterCommand.event}registerCommand(s,e){if(!s)throw new Error("invalid command");if(typeof s=="string"){if(!e)throw new Error("invalid command");return this.registerCommand({id:s,handler:e})}if(s.metadata&&Array.isArray(s.metadata.args)){const r=[];for(const l of s.metadata.args)r.push(l.constraint);const a=s.handler;s.handler=function(l,...c){return $ce(c,r),a(l,...c)}}const{id:t}=s;let i=this._commands.get(t);i||(i=new Ns,this._commands.set(t,i));const n=i.unshift(s),o=pe(()=>{n(),this._commands.get(t)?.isEmpty()&&this._commands.delete(t)});return this._onDidRegisterCommand.fire(t),o}registerCommandAlias(s,e){return at.registerCommand(s,(t,...i)=>t.get(Gt).executeCommand(e,...i))}getCommand(s){const e=this._commands.get(s);if(!(!e||e.isEmpty()))return ot.first(e)}getCommands(){const s=new Map;for(const e of this._commands.keys()){const t=this.getCommand(e);t&&s.set(e,t)}return s}};at.registerCommand("noop",()=>{});function z2(...s){switch(s.length){case 1:return m(1693,"Did you mean {0}?",s[0]);case 2:return m(1694,"Did you mean {0} or {1}?",s[0],s[1]);case 3:return m(1695,"Did you mean {0}, {1} or {2}?",s[0],s[1],s[2]);default:return}}const xue=m(1696,"Did you forget to open or close the quote?"),Lue=m(1697,"Did you forget to escape the '/' (slash) character? Put two backslashes before it to escape, e.g., '\\\\/'.");var Yd;let xC=(Yd=class{constructor(){this._input="",this._start=0,this._current=0,this._tokens=[],this._errors=[],this.stringRe=/[a-zA-Z0-9_<>\-\./\\:\*\?\+\[\]\^,#@;"%\$\p{L}-]+/uy}static getLexeme(e){switch(e.type){case 0:return"(";case 1:return")";case 2:return"!";case 3:return e.isTripleEq?"===":"==";case 4:return e.isTripleEq?"!==":"!=";case 5:return"<";case 6:return"<=";case 7:return">=";case 8:return">=";case 9:return"=~";case 10:return e.lexeme;case 11:return"true";case 12:return"false";case 13:return"in";case 14:return"not";case 15:return"&&";case 16:return"||";case 17:return e.lexeme;case 18:return e.lexeme;case 19:return e.lexeme;case 20:return"EOF";default:throw uE(`unhandled token type: ${JSON.stringify(e)}; have you forgotten to add a case?`)}}reset(e){return this._input=e,this._start=0,this._current=0,this._tokens=[],this._errors=[],this}scan(){for(;!this._isAtEnd();)switch(this._start=this._current,this._advance()){case 40:this._addToken(0);break;case 41:this._addToken(1);break;case 33:if(this._match(61)){const t=this._match(61);this._tokens.push({type:4,offset:this._start,isTripleEq:t})}else this._addToken(2);break;case 39:this._quotedString();break;case 47:this._regex();break;case 61:if(this._match(61)){const t=this._match(61);this._tokens.push({type:3,offset:this._start,isTripleEq:t})}else this._match(126)?this._addToken(9):this._error(z2("==","=~"));break;case 60:this._addToken(this._match(61)?6:5);break;case 62:this._addToken(this._match(61)?8:7);break;case 38:this._match(38)?this._addToken(15):this._error(z2("&&"));break;case 124:this._match(124)?this._addToken(16):this._error(z2("||"));break;case 32:case 13:case 9:case 10:case 160:break;default:this._string()}return this._start=this._current,this._addToken(20),Array.from(this._tokens)}_match(e){return this._isAtEnd()||this._input.charCodeAt(this._current)!==e?!1:(this._current++,!0)}_advance(){return this._input.charCodeAt(this._current++)}_peek(){return this._isAtEnd()?0:this._input.charCodeAt(this._current)}_addToken(e){this._tokens.push({type:e,offset:this._start})}_error(e){const t=this._start,i=this._input.substring(this._start,this._current),n={type:19,offset:this._start,lexeme:i};this._errors.push({offset:t,lexeme:i,additionalInfo:e}),this._tokens.push(n)}_string(){this.stringRe.lastIndex=this._start;const e=this.stringRe.exec(this._input);if(e){this._current=this._start+e[0].length;const t=this._input.substring(this._start,this._current),i=Yd._keywords.get(t);i?this._addToken(i):this._tokens.push({type:17,lexeme:t,offset:this._start})}}_quotedString(){for(;this._peek()!==39&&!this._isAtEnd();)this._advance();if(this._isAtEnd()){this._error(xue);return}this._advance(),this._tokens.push({type:18,lexeme:this._input.substring(this._start+1,this._current-1),offset:this._start+1})}_regex(){let e=this._current,t=!1,i=!1;for(;;){if(e>=this._input.length){this._current=e,this._error(Lue);return}const o=this._input.charCodeAt(e);if(t)t=!1;else if(o===47&&!i){e++;break}else o===91?i=!0:o===92?t=!0:o===93&&(i=!1);e++}for(;e<this._input.length&&Yd._regexFlags.has(this._input.charCodeAt(e));)e++;this._current=e;const n=this._input.substring(this._start,this._current);this._tokens.push({type:10,lexeme:n,offset:this._start})}_isAtEnd(){return this._current>=this._input.length}},Yd._regexFlags=new Set(["i","g","s","m","y","u"].map(e=>e.charCodeAt(0))),Yd._keywords=new Map([["not",14],["in",13],["false",12],["true",11]]),Yd);const ms=new Map;ms.set("false",!1);ms.set("true",!0);ms.set("isMac",Ye);ms.set("isLinux",vo);ms.set("isWindows",Bn);ms.set("isWeb",Vc);ms.set("isMacNative",Ye&&!Vc);ms.set("isEdge",zX);ms.set("isFirefox",Qce);ms.set("isChrome",B6);ms.set("isSafari",Jce);const kue=Object.prototype.hasOwnProperty,Eue={regexParsingWithErrorRecovery:!0},Iue=m(1675,"Empty context key expression"),Due=m(1676,"Did you forget to write an expression? You can also put 'false' or 'true' to always evaluate to false or true, respectively."),Tue=m(1677,"'in' after 'not'."),e$=m(1678,"closing parenthesis ')'"),Nue=m(1679,"Unexpected token"),Rue=m(1680,"Did you forget to put && or || before the token?"),Aue=m(1681,"Unexpected end of expression"),Mue=m(1682,"Did you forget to put a context key?");var pc;let Oue=(pc=class{constructor(e=Eue){this._config=e,this._scanner=new xC,this._tokens=[],this._current=0,this._parsingErrors=[],this._flagsGYRe=/g|y/g}parse(e){if(e===""){this._parsingErrors.push({message:Iue,offset:0,lexeme:"",additionalInfo:Due});return}this._tokens=this._scanner.reset(e).scan(),this._current=0,this._parsingErrors=[];try{const t=this._expr();if(!this._isAtEnd()){const i=this._peek(),n=i.type===17?Rue:void 0;throw this._parsingErrors.push({message:Nue,offset:i.offset,lexeme:xC.getLexeme(i),additionalInfo:n}),pc._parseError}return t}catch(t){if(t!==pc._parseError)throw t;return}}_expr(){return this._or()}_or(){const e=[this._and()];for(;this._matchOne(16);){const t=this._and();e.push(t)}return e.length===1?e[0]:K.or(...e)}_and(){const e=[this._term()];for(;this._matchOne(15);){const t=this._term();e.push(t)}return e.length===1?e[0]:K.and(...e)}_term(){if(this._matchOne(2)){const e=this._peek();switch(e.type){case 11:return this._advance(),eo.INSTANCE;case 12:return this._advance(),yo.INSTANCE;case 0:{this._advance();const t=this._expr();return this._consume(1,e$),t?.negate()}case 17:return this._advance(),o_.create(e.lexeme);default:throw this._errExpectedButGot("KEY | true | false | '(' expression ')'",e)}}return this._primary()}_primary(){const e=this._peek();switch(e.type){case 11:return this._advance(),K.true();case 12:return this._advance(),K.false();case 0:{this._advance();const t=this._expr();return this._consume(1,e$),t}case 17:{const t=e.lexeme;if(this._advance(),this._matchOne(9)){const n=this._peek();if(!this._config.regexParsingWithErrorRecovery){if(this._advance(),n.type!==10)throw this._errExpectedButGot("REGEX",n);const o=n.lexeme,r=o.lastIndexOf("/"),a=r===o.length-1?void 0:this._removeFlagsGY(o.substring(r+1));let l;try{l=new RegExp(o.substring(1,r),a)}catch{throw this._errExpectedButGot("REGEX",n)}return Jy.create(t,l)}switch(n.type){case 10:case 19:{const o=[n.lexeme];this._advance();let r=this._peek(),a=0;for(let u=0;u<n.lexeme.length;u++)n.lexeme.charCodeAt(u)===40?a++:n.lexeme.charCodeAt(u)===41&&a--;for(;!this._isAtEnd()&&r.type!==15&&r.type!==16;){switch(r.type){case 0:a++;break;case 1:a--;break;case 10:case 18:for(let u=0;u<r.lexeme.length;u++)r.lexeme.charCodeAt(u)===40?a++:n.lexeme.charCodeAt(u)===41&&a--}if(a<0)break;o.push(xC.getLexeme(r)),this._advance(),r=this._peek()}const l=o.join(""),c=l.lastIndexOf("/"),d=c===l.length-1?void 0:this._removeFlagsGY(l.substring(c+1));let h;try{h=new RegExp(l.substring(1,c),d)}catch{throw this._errExpectedButGot("REGEX",n)}return K.regex(t,h)}case 18:{const o=n.lexeme;this._advance();let r=null;if(!XX(o)){const a=o.indexOf("/"),l=o.lastIndexOf("/");if(a!==l&&a>=0){const c=o.slice(a+1,l),d=o[l+1]==="i"?"i":"";try{r=new RegExp(c,d)}catch{throw this._errExpectedButGot("REGEX",n)}}}if(r===null)throw this._errExpectedButGot("REGEX",n);return Jy.create(t,r)}default:throw this._errExpectedButGot("REGEX",this._peek())}}if(this._matchOne(14)){this._consume(13,Tue);const n=this._value();return K.notIn(t,n)}switch(this._peek().type){case 3:{this._advance();const n=this._value();if(this._previous().type===18)return K.equals(t,n);switch(n){case"true":return K.has(t);case"false":return K.not(t);default:return K.equals(t,n)}}case 4:{this._advance();const n=this._value();if(this._previous().type===18)return K.notEquals(t,n);switch(n){case"true":return K.not(t);case"false":return K.has(t);default:return K.notEquals(t,n)}}case 5:return this._advance(),WR.create(t,this._value());case 6:return this._advance(),HR.create(t,this._value());case 7:return this._advance(),FR.create(t,this._value());case 8:return this._advance(),BR.create(t,this._value());case 13:return this._advance(),K.in(t,this._value());default:return K.has(t)}}case 20:throw this._parsingErrors.push({message:Aue,offset:e.offset,lexeme:"",additionalInfo:Mue}),pc._parseError;default:throw this._errExpectedButGot(`true | false | KEY 
	| KEY '=~' REGEX 
	| KEY ('==' | '!=' | '<' | '<=' | '>' | '>=' | 'in' | 'not' 'in') value`,this._peek())}}_value(){const e=this._peek();switch(e.type){case 17:case 18:return this._advance(),e.lexeme;case 11:return this._advance(),"true";case 12:return this._advance(),"false";case 13:return this._advance(),"in";default:return""}}_removeFlagsGY(e){return e.replaceAll(this._flagsGYRe,"")}_previous(){return this._tokens[this._current-1]}_matchOne(e){return this._check(e)?(this._advance(),!0):!1}_advance(){return this._isAtEnd()||this._current++,this._previous()}_consume(e,t){if(this._check(e))return this._advance();throw this._errExpectedButGot(t,this._peek())}_errExpectedButGot(e,t,i){const n=m(1683,`Expected: {0}
Received: '{1}'.`,e,xC.getLexeme(t)),o=t.offset,r=xC.getLexeme(t);return this._parsingErrors.push({message:n,offset:o,lexeme:r,additionalInfo:i}),pc._parseError}_check(e){return this._peek().type===e}_peek(){return this._tokens[this._current]}_isAtEnd(){return this._peek().type===20}},pc._parseError=new Error,pc);const FV=class FV{static false(){return eo.INSTANCE}static true(){return yo.INSTANCE}static has(e){return s_.create(e)}static equals(e,t){return Kv.create(e,t)}static notEquals(e,t){return OR.create(e,t)}static regex(e,t){return Jy.create(e,t)}static in(e,t){return AR.create(e,t)}static notIn(e,t){return MR.create(e,t)}static not(e){return o_.create(e)}static and(...e){return $f.create(e,null,!0)}static or(...e){return Md.create(e,null,!0)}static deserialize(e){return e==null?void 0:this._parser.parse(e)}};FV._parser=new Oue({regexParsingWithErrorRecovery:!1});let K=FV;function Pue(s,e){const t=s?s.substituteConstants():void 0,i=e?e.substituteConstants():void 0;return!t&&!i?!0:!t||!i?!1:t.equals(i)}function I1(s,e){return s.cmp(e)}const zT=class zT{constructor(){this.type=0}cmp(e){return this.type-e.type}equals(e){return e.type===this.type}substituteConstants(){return this}evaluate(e){return!1}serialize(){return"false"}keys(){return[]}negate(){return yo.INSTANCE}};zT.INSTANCE=new zT;let eo=zT;const $T=class $T{constructor(){this.type=1}cmp(e){return this.type-e.type}equals(e){return e.type===this.type}substituteConstants(){return this}evaluate(e){return!0}serialize(){return"true"}keys(){return[]}negate(){return eo.INSTANCE}};$T.INSTANCE=new $T;let yo=$T;class s_{static create(e,t=null){const i=ms.get(e);return typeof i=="boolean"?i?yo.INSTANCE:eo.INSTANCE:new s_(e,t)}constructor(e,t){this.key=e,this.negated=t,this.type=2}cmp(e){return e.type!==this.type?this.type-e.type:VQ(this.key,e.key)}equals(e){return e.type===this.type?this.key===e.key:!1}substituteConstants(){const e=ms.get(this.key);return typeof e=="boolean"?e?yo.INSTANCE:eo.INSTANCE:this}evaluate(e){return!!e.getValue(this.key)}serialize(){return this.key}keys(){return[this.key]}negate(){return this.negated||(this.negated=o_.create(this.key,this)),this.negated}}class Kv{static create(e,t,i=null){if(typeof t=="boolean")return t?s_.create(e,i):o_.create(e,i);const n=ms.get(e);return typeof n=="boolean"?t===(n?"true":"false")?yo.INSTANCE:eo.INSTANCE:new Kv(e,t,i)}constructor(e,t,i){this.key=e,this.value=t,this.negated=i,this.type=4}cmp(e){return e.type!==this.type?this.type-e.type:r_(this.key,this.value,e.key,e.value)}equals(e){return e.type===this.type?this.key===e.key&&this.value===e.value:!1}substituteConstants(){const e=ms.get(this.key);if(typeof e=="boolean"){const t=e?"true":"false";return this.value===t?yo.INSTANCE:eo.INSTANCE}return this}evaluate(e){return e.getValue(this.key)==this.value}serialize(){return`${this.key} == '${this.value}'`}keys(){return[this.key]}negate(){return this.negated||(this.negated=OR.create(this.key,this.value,this)),this.negated}}class AR{static create(e,t){return new AR(e,t)}constructor(e,t){this.key=e,this.valueKey=t,this.type=10,this.negated=null}cmp(e){return e.type!==this.type?this.type-e.type:r_(this.key,this.valueKey,e.key,e.valueKey)}equals(e){return e.type===this.type?this.key===e.key&&this.valueKey===e.valueKey:!1}substituteConstants(){return this}evaluate(e){const t=e.getValue(this.valueKey),i=e.getValue(this.key);return Array.isArray(t)?t.includes(i):typeof i=="string"&&typeof t=="object"&&t!==null?kue.call(t,i):!1}serialize(){return`${this.key} in '${this.valueKey}'`}keys(){return[this.key,this.valueKey]}negate(){return this.negated||(this.negated=MR.create(this.key,this.valueKey)),this.negated}}class MR{static create(e,t){return new MR(e,t)}constructor(e,t){this.key=e,this.valueKey=t,this.type=11,this._negated=AR.create(e,t)}cmp(e){return e.type!==this.type?this.type-e.type:this._negated.cmp(e._negated)}equals(e){return e.type===this.type?this._negated.equals(e._negated):!1}substituteConstants(){return this}evaluate(e){return!this._negated.evaluate(e)}serialize(){return`${this.key} not in '${this.valueKey}'`}keys(){return this._negated.keys()}negate(){return this._negated}}class OR{static create(e,t,i=null){if(typeof t=="boolean")return t?o_.create(e,i):s_.create(e,i);const n=ms.get(e);return typeof n=="boolean"?t===(n?"true":"false")?eo.INSTANCE:yo.INSTANCE:new OR(e,t,i)}constructor(e,t,i){this.key=e,this.value=t,this.negated=i,this.type=5}cmp(e){return e.type!==this.type?this.type-e.type:r_(this.key,this.value,e.key,e.value)}equals(e){return e.type===this.type?this.key===e.key&&this.value===e.value:!1}substituteConstants(){const e=ms.get(this.key);if(typeof e=="boolean"){const t=e?"true":"false";return this.value===t?eo.INSTANCE:yo.INSTANCE}return this}evaluate(e){return e.getValue(this.key)!=this.value}serialize(){return`${this.key} != '${this.value}'`}keys(){return[this.key]}negate(){return this.negated||(this.negated=Kv.create(this.key,this.value,this)),this.negated}}class o_{static create(e,t=null){const i=ms.get(e);return typeof i=="boolean"?i?eo.INSTANCE:yo.INSTANCE:new o_(e,t)}constructor(e,t){this.key=e,this.negated=t,this.type=3}cmp(e){return e.type!==this.type?this.type-e.type:VQ(this.key,e.key)}equals(e){return e.type===this.type?this.key===e.key:!1}substituteConstants(){const e=ms.get(this.key);return typeof e=="boolean"?e?eo.INSTANCE:yo.INSTANCE:this}evaluate(e){return!e.getValue(this.key)}serialize(){return`!${this.key}`}keys(){return[this.key]}negate(){return this.negated||(this.negated=s_.create(this.key,this)),this.negated}}function PR(s,e){if(typeof s=="string"){const t=parseFloat(s);isNaN(t)||(s=t)}return typeof s=="string"||typeof s=="number"?e(s):eo.INSTANCE}class FR{static create(e,t,i=null){return PR(t,n=>new FR(e,n,i))}constructor(e,t,i){this.key=e,this.value=t,this.negated=i,this.type=12}cmp(e){return e.type!==this.type?this.type-e.type:r_(this.key,this.value,e.key,e.value)}equals(e){return e.type===this.type?this.key===e.key&&this.value===e.value:!1}substituteConstants(){return this}evaluate(e){return typeof this.value=="string"?!1:parseFloat(e.getValue(this.key))>this.value}serialize(){return`${this.key} > ${this.value}`}keys(){return[this.key]}negate(){return this.negated||(this.negated=HR.create(this.key,this.value,this)),this.negated}}class BR{static create(e,t,i=null){return PR(t,n=>new BR(e,n,i))}constructor(e,t,i){this.key=e,this.value=t,this.negated=i,this.type=13}cmp(e){return e.type!==this.type?this.type-e.type:r_(this.key,this.value,e.key,e.value)}equals(e){return e.type===this.type?this.key===e.key&&this.value===e.value:!1}substituteConstants(){return this}evaluate(e){return typeof this.value=="string"?!1:parseFloat(e.getValue(this.key))>=this.value}serialize(){return`${this.key} >= ${this.value}`}keys(){return[this.key]}negate(){return this.negated||(this.negated=WR.create(this.key,this.value,this)),this.negated}}class WR{static create(e,t,i=null){return PR(t,n=>new WR(e,n,i))}constructor(e,t,i){this.key=e,this.value=t,this.negated=i,this.type=14}cmp(e){return e.type!==this.type?this.type-e.type:r_(this.key,this.value,e.key,e.value)}equals(e){return e.type===this.type?this.key===e.key&&this.value===e.value:!1}substituteConstants(){return this}evaluate(e){return typeof this.value=="string"?!1:parseFloat(e.getValue(this.key))<this.value}serialize(){return`${this.key} < ${this.value}`}keys(){return[this.key]}negate(){return this.negated||(this.negated=BR.create(this.key,this.value,this)),this.negated}}class HR{static create(e,t,i=null){return PR(t,n=>new HR(e,n,i))}constructor(e,t,i){this.key=e,this.value=t,this.negated=i,this.type=15}cmp(e){return e.type!==this.type?this.type-e.type:r_(this.key,this.value,e.key,e.value)}equals(e){return e.type===this.type?this.key===e.key&&this.value===e.value:!1}substituteConstants(){return this}evaluate(e){return typeof this.value=="string"?!1:parseFloat(e.getValue(this.key))<=this.value}serialize(){return`${this.key} <= ${this.value}`}keys(){return[this.key]}negate(){return this.negated||(this.negated=FR.create(this.key,this.value,this)),this.negated}}class Jy{static create(e,t){return new Jy(e,t)}constructor(e,t){this.key=e,this.regexp=t,this.type=7,this.negated=null}cmp(e){if(e.type!==this.type)return this.type-e.type;if(this.key<e.key)return-1;if(this.key>e.key)return 1;const t=this.regexp?this.regexp.source:"",i=e.regexp?e.regexp.source:"";return t<i?-1:t>i?1:0}equals(e){if(e.type===this.type){const t=this.regexp?this.regexp.source:"",i=e.regexp?e.regexp.source:"";return this.key===e.key&&t===i}return!1}substituteConstants(){return this}evaluate(e){const t=e.getValue(this.key);return this.regexp?this.regexp.test(t):!1}serialize(){const e=this.regexp?`/${this.regexp.source}/${this.regexp.flags}`:"/invalid/";return`${this.key} =~ ${e}`}keys(){return[this.key]}negate(){return this.negated||(this.negated=uW.create(this)),this.negated}}class uW{static create(e){return new uW(e)}constructor(e){this._actual=e,this.type=8}cmp(e){return e.type!==this.type?this.type-e.type:this._actual.cmp(e._actual)}equals(e){return e.type===this.type?this._actual.equals(e._actual):!1}substituteConstants(){return this}evaluate(e){return!this._actual.evaluate(e)}serialize(){return`!(${this._actual.serialize()})`}keys(){return this._actual.keys()}negate(){return this._actual}}function HQ(s){let e=null;for(let t=0,i=s.length;t<i;t++){const n=s[t].substituteConstants();if(s[t]!==n&&e===null){e=[];for(let o=0;o<t;o++)e[o]=s[o]}e!==null&&(e[t]=n)}return e===null?s:e}class $f{static create(e,t,i){return $f._normalizeArr(e,t,i)}constructor(e,t){this.expr=e,this.negated=t,this.type=6}cmp(e){if(e.type!==this.type)return this.type-e.type;if(this.expr.length<e.expr.length)return-1;if(this.expr.length>e.expr.length)return 1;for(let t=0,i=this.expr.length;t<i;t++){const n=I1(this.expr[t],e.expr[t]);if(n!==0)return n}return 0}equals(e){if(e.type===this.type){if(this.expr.length!==e.expr.length)return!1;for(let t=0,i=this.expr.length;t<i;t++)if(!this.expr[t].equals(e.expr[t]))return!1;return!0}return!1}substituteConstants(){const e=HQ(this.expr);return e===this.expr?this:$f.create(e,this.negated,!1)}evaluate(e){for(let t=0,i=this.expr.length;t<i;t++)if(!this.expr[t].evaluate(e))return!1;return!0}static _normalizeArr(e,t,i){const n=[];let o=!1;for(const r of e)if(r){if(r.type===1){o=!0;continue}if(r.type===0)return eo.INSTANCE;if(r.type===6){n.push(...r.expr);continue}n.push(r)}if(n.length===0&&o)return yo.INSTANCE;if(n.length!==0){if(n.length===1)return n[0];n.sort(I1);for(let r=1;r<n.length;r++)n[r-1].equals(n[r])&&(n.splice(r,1),r--);if(n.length===1)return n[0];for(;n.length>1;){const r=n[n.length-1];if(r.type!==9)break;n.pop();const a=n.pop(),l=n.length===0,c=Md.create(r.expr.map(d=>$f.create([d,a],null,i)),null,l);c&&(n.push(c),n.sort(I1))}if(n.length===1)return n[0];if(i){for(let r=0;r<n.length;r++)for(let a=r+1;a<n.length;a++)if(n[r].negate().equals(n[a]))return eo.INSTANCE;if(n.length===1)return n[0]}return new $f(n,t)}}serialize(){return this.expr.map(e=>e.serialize()).join(" && ")}keys(){const e=[];for(const t of this.expr)e.push(...t.keys());return e}negate(){if(!this.negated){const e=[];for(const t of this.expr)e.push(t.negate());this.negated=Md.create(e,this,!0)}return this.negated}}class Md{static create(e,t,i){return Md._normalizeArr(e,t,i)}constructor(e,t){this.expr=e,this.negated=t,this.type=9}cmp(e){if(e.type!==this.type)return this.type-e.type;if(this.expr.length<e.expr.length)return-1;if(this.expr.length>e.expr.length)return 1;for(let t=0,i=this.expr.length;t<i;t++){const n=I1(this.expr[t],e.expr[t]);if(n!==0)return n}return 0}equals(e){if(e.type===this.type){if(this.expr.length!==e.expr.length)return!1;for(let t=0,i=this.expr.length;t<i;t++)if(!this.expr[t].equals(e.expr[t]))return!1;return!0}return!1}substituteConstants(){const e=HQ(this.expr);return e===this.expr?this:Md.create(e,this.negated,!1)}evaluate(e){for(let t=0,i=this.expr.length;t<i;t++)if(this.expr[t].evaluate(e))return!0;return!1}static _normalizeArr(e,t,i){let n=[],o=!1;if(e){for(let r=0,a=e.length;r<a;r++){const l=e[r];if(l){if(l.type===0){o=!0;continue}if(l.type===1)return yo.INSTANCE;if(l.type===9){n=n.concat(l.expr);continue}n.push(l)}}if(n.length===0&&o)return eo.INSTANCE;n.sort(I1)}if(n.length!==0){if(n.length===1)return n[0];for(let r=1;r<n.length;r++)n[r-1].equals(n[r])&&(n.splice(r,1),r--);if(n.length===1)return n[0];if(i){for(let r=0;r<n.length;r++)for(let a=r+1;a<n.length;a++)if(n[r].negate().equals(n[a]))return yo.INSTANCE;if(n.length===1)return n[0]}return new Md(n,t)}}serialize(){return this.expr.map(e=>e.serialize()).join(" || ")}keys(){const e=[];for(const t of this.expr)e.push(...t.keys());return e}negate(){if(!this.negated){const e=[];for(const t of this.expr)e.push(t.negate());for(;e.length>1;){const t=e.shift(),i=e.shift(),n=[];for(const o of i$(t))for(const r of i$(i))n.push($f.create([o,r],null,!1));e.unshift(Md.create(n,null,!1))}this.negated=Md.create(e,this,!0)}return this.negated}}const Gb=class Gb extends s_{static all(){return Gb._info.values()}constructor(e,t,i){super(e,null),this._defaultValue=t,typeof i=="object"?Gb._info.push({...i,key:e}):i!==!0&&Gb._info.push({key:e,description:i,type:t!=null?typeof t:void 0})}bindTo(e){return e.createKey(this.key,this._defaultValue)}getValue(e){return e.getContextKeyValue(this.key)}toNegated(){return this.negate()}isEqualTo(e){return Kv.create(this.key,e)}};Gb._info=[];let ce=Gb;const Ae=qe("contextKeyService");function VQ(s,e){return s<e?-1:s>e?1:0}function r_(s,e,t,i){return s<t?-1:s>t?1:e<i?-1:e>i?1:0}function TP(s,e){if(s.type===0||e.type===1)return!0;if(s.type===9)return e.type===9?t$(s.expr,e.expr):!1;if(e.type===9){for(const t of e.expr)if(TP(s,t))return!0;return!1}if(s.type===6){if(e.type===6)return t$(e.expr,s.expr);for(const t of s.expr)if(TP(t,e))return!0;return!1}return s.equals(e)}function t$(s,e){let t=0,i=0;for(;t<s.length&&i<e.length;){const n=s[t].cmp(e[i]);if(n<0)return!1;n===0&&t++,i++}return t===s.length}function i$(s){return s.type===9?s.expr:[s]}class Fue{constructor(){this.data=new Map}add(e,t){D2(bn(e)),D2(rn(t)),D2(!this.data.has(e),"There is already an extension with this id"),this.data.set(e,t)}as(e){return this.data.get(e)||null}dispose(){this.data.forEach(e=>{vm(e.dispose)&&e.dispose()}),this.data.clear()}}const bi=new Fue;class gW{constructor(){this._coreKeybindings=new Ns,this._extensionKeybindings=[],this._cachedMergedKeybindings=null}static bindToCurrentPlatform(e){if(Fo===1){if(e&&e.win)return e.win}else if(Fo===2){if(e&&e.mac)return e.mac}else if(e&&e.linux)return e.linux;return e}registerKeybindingRule(e){const t=gW.bindToCurrentPlatform(e),i=new U;if(t&&t.primary){const n=oP(t.primary,Fo);n&&i.add(this._registerDefaultKeybinding(n,e.id,e.args,e.weight,0,e.when))}if(t&&Array.isArray(t.secondary))for(let n=0,o=t.secondary.length;n<o;n++){const r=t.secondary[n],a=oP(r,Fo);a&&i.add(this._registerDefaultKeybinding(a,e.id,e.args,e.weight,-n-1,e.when))}return i}registerCommandAndKeybindingRule(e){return Ra(this.registerKeybindingRule(e),at.registerCommand(e))}_registerDefaultKeybinding(e,t,i,n,o,r){const a=this._coreKeybindings.push({keybinding:e,command:t,commandArgs:i,when:r,weight1:n,weight2:o,extensionId:null,isBuiltinExtension:!1});return this._cachedMergedKeybindings=null,pe(()=>{a(),this._cachedMergedKeybindings=null})}getDefaultKeybindings(){return this._cachedMergedKeybindings||(this._cachedMergedKeybindings=Array.from(this._coreKeybindings).concat(this._extensionKeybindings),this._cachedMergedKeybindings.sort(Wue)),this._cachedMergedKeybindings.slice(0)}}const Dn=new gW,Bue={EditorModes:"platform.keybindingsRegistry"};bi.add(Bue.EditorModes,Dn);function Wue(s,e){if(s.weight1!==e.weight1)return s.weight1-e.weight1;if(s.command&&e.command){if(s.command<e.command)return-1;if(s.command>e.command)return 1}return s.weight2-e.weight2}var Hue=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},n$=function(s,e){return function(t,i){e(t,i,s)}},xk;function aw(s){return s.command!==void 0}function Vue(s){return s.submenu!==void 0}const D=class D{constructor(e){if(D._instances.has(e))throw new TypeError(`MenuId with identifier '${e}' already exists. Use MenuId.for(ident) or a unique identifier`);D._instances.set(e,this),this.id=e}};D._instances=new Map,D.CommandPalette=new D("CommandPalette"),D.DebugBreakpointsContext=new D("DebugBreakpointsContext"),D.DebugCallStackContext=new D("DebugCallStackContext"),D.DebugConsoleContext=new D("DebugConsoleContext"),D.DebugVariablesContext=new D("DebugVariablesContext"),D.NotebookVariablesContext=new D("NotebookVariablesContext"),D.DebugHoverContext=new D("DebugHoverContext"),D.DebugWatchContext=new D("DebugWatchContext"),D.DebugToolBar=new D("DebugToolBar"),D.DebugToolBarStop=new D("DebugToolBarStop"),D.DebugDisassemblyContext=new D("DebugDisassemblyContext"),D.DebugCallStackToolbar=new D("DebugCallStackToolbar"),D.DebugCreateConfiguration=new D("DebugCreateConfiguration"),D.EditorContext=new D("EditorContext"),D.SimpleEditorContext=new D("SimpleEditorContext"),D.EditorContent=new D("EditorContent"),D.EditorLineNumberContext=new D("EditorLineNumberContext"),D.EditorContextCopy=new D("EditorContextCopy"),D.EditorContextPeek=new D("EditorContextPeek"),D.EditorContextShare=new D("EditorContextShare"),D.EditorTitle=new D("EditorTitle"),D.CompactWindowEditorTitle=new D("CompactWindowEditorTitle"),D.EditorTitleRun=new D("EditorTitleRun"),D.EditorTitleContext=new D("EditorTitleContext"),D.EditorTitleContextShare=new D("EditorTitleContextShare"),D.EmptyEditorGroup=new D("EmptyEditorGroup"),D.EmptyEditorGroupContext=new D("EmptyEditorGroupContext"),D.EditorTabsBarContext=new D("EditorTabsBarContext"),D.EditorTabsBarShowTabsSubmenu=new D("EditorTabsBarShowTabsSubmenu"),D.EditorTabsBarShowTabsZenModeSubmenu=new D("EditorTabsBarShowTabsZenModeSubmenu"),D.EditorActionsPositionSubmenu=new D("EditorActionsPositionSubmenu"),D.EditorSplitMoveSubmenu=new D("EditorSplitMoveSubmenu"),D.ExplorerContext=new D("ExplorerContext"),D.ExplorerContextShare=new D("ExplorerContextShare"),D.ExtensionContext=new D("ExtensionContext"),D.ExtensionEditorContextMenu=new D("ExtensionEditorContextMenu"),D.GlobalActivity=new D("GlobalActivity"),D.CommandCenter=new D("CommandCenter"),D.CommandCenterCenter=new D("CommandCenterCenter"),D.LayoutControlMenuSubmenu=new D("LayoutControlMenuSubmenu"),D.LayoutControlMenu=new D("LayoutControlMenu"),D.MenubarMainMenu=new D("MenubarMainMenu"),D.MenubarAppearanceMenu=new D("MenubarAppearanceMenu"),D.MenubarDebugMenu=new D("MenubarDebugMenu"),D.MenubarEditMenu=new D("MenubarEditMenu"),D.MenubarCopy=new D("MenubarCopy"),D.MenubarFileMenu=new D("MenubarFileMenu"),D.MenubarGoMenu=new D("MenubarGoMenu"),D.MenubarHelpMenu=new D("MenubarHelpMenu"),D.MenubarLayoutMenu=new D("MenubarLayoutMenu"),D.MenubarNewBreakpointMenu=new D("MenubarNewBreakpointMenu"),D.PanelAlignmentMenu=new D("PanelAlignmentMenu"),D.PanelPositionMenu=new D("PanelPositionMenu"),D.ActivityBarPositionMenu=new D("ActivityBarPositionMenu"),D.MenubarPreferencesMenu=new D("MenubarPreferencesMenu"),D.MenubarRecentMenu=new D("MenubarRecentMenu"),D.MenubarSelectionMenu=new D("MenubarSelectionMenu"),D.MenubarShare=new D("MenubarShare"),D.MenubarSwitchEditorMenu=new D("MenubarSwitchEditorMenu"),D.MenubarSwitchGroupMenu=new D("MenubarSwitchGroupMenu"),D.MenubarTerminalMenu=new D("MenubarTerminalMenu"),D.MenubarTerminalSuggestStatusMenu=new D("MenubarTerminalSuggestStatusMenu"),D.MenubarViewMenu=new D("MenubarViewMenu"),D.MenubarHomeMenu=new D("MenubarHomeMenu"),D.OpenEditorsContext=new D("OpenEditorsContext"),D.OpenEditorsContextShare=new D("OpenEditorsContextShare"),D.ProblemsPanelContext=new D("ProblemsPanelContext"),D.SCMInputBox=new D("SCMInputBox"),D.SCMChangeContext=new D("SCMChangeContext"),D.SCMResourceContext=new D("SCMResourceContext"),D.SCMResourceContextShare=new D("SCMResourceContextShare"),D.SCMResourceFolderContext=new D("SCMResourceFolderContext"),D.SCMResourceGroupContext=new D("SCMResourceGroupContext"),D.SCMSourceControl=new D("SCMSourceControl"),D.SCMSourceControlInline=new D("SCMSourceControlInline"),D.SCMSourceControlTitle=new D("SCMSourceControlTitle"),D.SCMHistoryTitle=new D("SCMHistoryTitle"),D.SCMHistoryItemContext=new D("SCMHistoryItemContext"),D.SCMHistoryItemChangeContext=new D("SCMHistoryItemChangeContext"),D.SCMHistoryItemRefContext=new D("SCMHistoryItemRefContext"),D.SCMArtifactGroupContext=new D("SCMArtifactGroupContext"),D.SCMArtifactContext=new D("SCMArtifactContext"),D.SCMQuickDiffDecorations=new D("SCMQuickDiffDecorations"),D.SCMTitle=new D("SCMTitle"),D.SearchContext=new D("SearchContext"),D.SearchActionMenu=new D("SearchActionContext"),D.StatusBarWindowIndicatorMenu=new D("StatusBarWindowIndicatorMenu"),D.StatusBarRemoteIndicatorMenu=new D("StatusBarRemoteIndicatorMenu"),D.StickyScrollContext=new D("StickyScrollContext"),D.TestItem=new D("TestItem"),D.TestItemGutter=new D("TestItemGutter"),D.TestProfilesContext=new D("TestProfilesContext"),D.TestMessageContext=new D("TestMessageContext"),D.TestMessageContent=new D("TestMessageContent"),D.TestPeekElement=new D("TestPeekElement"),D.TestPeekTitle=new D("TestPeekTitle"),D.TestCallStack=new D("TestCallStack"),D.TestCoverageFilterItem=new D("TestCoverageFilterItem"),D.TouchBarContext=new D("TouchBarContext"),D.TitleBar=new D("TitleBar"),D.TitleBarContext=new D("TitleBarContext"),D.TitleBarTitleContext=new D("TitleBarTitleContext"),D.TunnelContext=new D("TunnelContext"),D.TunnelPrivacy=new D("TunnelPrivacy"),D.TunnelProtocol=new D("TunnelProtocol"),D.TunnelPortInline=new D("TunnelInline"),D.TunnelTitle=new D("TunnelTitle"),D.TunnelLocalAddressInline=new D("TunnelLocalAddressInline"),D.TunnelOriginInline=new D("TunnelOriginInline"),D.ViewItemContext=new D("ViewItemContext"),D.ViewContainerTitle=new D("ViewContainerTitle"),D.ViewContainerTitleContext=new D("ViewContainerTitleContext"),D.ViewTitle=new D("ViewTitle"),D.ViewTitleContext=new D("ViewTitleContext"),D.CommentEditorActions=new D("CommentEditorActions"),D.CommentThreadTitle=new D("CommentThreadTitle"),D.CommentThreadActions=new D("CommentThreadActions"),D.CommentThreadAdditionalActions=new D("CommentThreadAdditionalActions"),D.CommentThreadTitleContext=new D("CommentThreadTitleContext"),D.CommentThreadCommentContext=new D("CommentThreadCommentContext"),D.CommentTitle=new D("CommentTitle"),D.CommentActions=new D("CommentActions"),D.CommentsViewThreadActions=new D("CommentsViewThreadActions"),D.InteractiveToolbar=new D("InteractiveToolbar"),D.InteractiveCellTitle=new D("InteractiveCellTitle"),D.InteractiveCellDelete=new D("InteractiveCellDelete"),D.InteractiveCellExecute=new D("InteractiveCellExecute"),D.InteractiveInputExecute=new D("InteractiveInputExecute"),D.InteractiveInputConfig=new D("InteractiveInputConfig"),D.ReplInputExecute=new D("ReplInputExecute"),D.IssueReporter=new D("IssueReporter"),D.NotebookToolbar=new D("NotebookToolbar"),D.NotebookToolbarContext=new D("NotebookToolbarContext"),D.NotebookStickyScrollContext=new D("NotebookStickyScrollContext"),D.NotebookCellTitle=new D("NotebookCellTitle"),D.NotebookCellDelete=new D("NotebookCellDelete"),D.NotebookCellInsert=new D("NotebookCellInsert"),D.NotebookCellBetween=new D("NotebookCellBetween"),D.NotebookCellListTop=new D("NotebookCellTop"),D.NotebookCellExecute=new D("NotebookCellExecute"),D.NotebookCellExecuteGoTo=new D("NotebookCellExecuteGoTo"),D.NotebookCellExecutePrimary=new D("NotebookCellExecutePrimary"),D.NotebookDiffCellInputTitle=new D("NotebookDiffCellInputTitle"),D.NotebookDiffDocumentMetadata=new D("NotebookDiffDocumentMetadata"),D.NotebookDiffCellMetadataTitle=new D("NotebookDiffCellMetadataTitle"),D.NotebookDiffCellOutputsTitle=new D("NotebookDiffCellOutputsTitle"),D.NotebookOutputToolbar=new D("NotebookOutputToolbar"),D.NotebookOutlineFilter=new D("NotebookOutlineFilter"),D.NotebookOutlineActionMenu=new D("NotebookOutlineActionMenu"),D.NotebookEditorLayoutConfigure=new D("NotebookEditorLayoutConfigure"),D.NotebookKernelSource=new D("NotebookKernelSource"),D.BulkEditTitle=new D("BulkEditTitle"),D.BulkEditContext=new D("BulkEditContext"),D.TimelineItemContext=new D("TimelineItemContext"),D.TimelineTitle=new D("TimelineTitle"),D.TimelineTitleContext=new D("TimelineTitleContext"),D.TimelineFilterSubMenu=new D("TimelineFilterSubMenu"),D.AccountsContext=new D("AccountsContext"),D.SidebarTitle=new D("SidebarTitle"),D.PanelTitle=new D("PanelTitle"),D.AuxiliaryBarTitle=new D("AuxiliaryBarTitle"),D.TerminalInstanceContext=new D("TerminalInstanceContext"),D.TerminalEditorInstanceContext=new D("TerminalEditorInstanceContext"),D.TerminalNewDropdownContext=new D("TerminalNewDropdownContext"),D.TerminalTabContext=new D("TerminalTabContext"),D.TerminalTabEmptyAreaContext=new D("TerminalTabEmptyAreaContext"),D.TerminalStickyScrollContext=new D("TerminalStickyScrollContext"),D.WebviewContext=new D("WebviewContext"),D.InlineCompletionsActions=new D("InlineCompletionsActions"),D.InlineEditsActions=new D("InlineEditsActions"),D.NewFile=new D("NewFile"),D.MergeInput1Toolbar=new D("MergeToolbar1Toolbar"),D.MergeInput2Toolbar=new D("MergeToolbar2Toolbar"),D.MergeBaseToolbar=new D("MergeBaseToolbar"),D.MergeInputResultToolbar=new D("MergeToolbarResultToolbar"),D.InlineSuggestionToolbar=new D("InlineSuggestionToolbar"),D.InlineEditToolbar=new D("InlineEditToolbar"),D.ChatContext=new D("ChatContext"),D.ChatCodeBlock=new D("ChatCodeblock"),D.ChatCompareBlock=new D("ChatCompareBlock"),D.ChatMessageTitle=new D("ChatMessageTitle"),D.ChatHistory=new D("ChatHistory"),D.ChatWelcomeContext=new D("ChatWelcomeContext"),D.ChatMessageFooter=new D("ChatMessageFooter"),D.ChatExecute=new D("ChatExecute"),D.ChatInput=new D("ChatInput"),D.ChatInputSide=new D("ChatInputSide"),D.ChatModePicker=new D("ChatModePicker"),D.ChatEditingWidgetToolbar=new D("ChatEditingWidgetToolbar"),D.ChatEditingEditorContent=new D("ChatEditingEditorContent"),D.ChatEditingEditorHunk=new D("ChatEditingEditorHunk"),D.ChatEditingDeletedNotebookCell=new D("ChatEditingDeletedNotebookCell"),D.ChatInputAttachmentToolbar=new D("ChatInputAttachmentToolbar"),D.ChatEditingWidgetModifiedFilesToolbar=new D("ChatEditingWidgetModifiedFilesToolbar"),D.ChatInputResourceAttachmentContext=new D("ChatInputResourceAttachmentContext"),D.ChatInputSymbolAttachmentContext=new D("ChatInputSymbolAttachmentContext"),D.ChatInlineResourceAnchorContext=new D("ChatInlineResourceAnchorContext"),D.ChatInlineSymbolAnchorContext=new D("ChatInlineSymbolAnchorContext"),D.ChatMessageCheckpoint=new D("ChatMessageCheckpoint"),D.ChatMessageRestoreCheckpoint=new D("ChatMessageRestoreCheckpoint"),D.ChatNewMenu=new D("ChatNewMenu"),D.ChatEditingCodeBlockContext=new D("ChatEditingCodeBlockContext"),D.ChatTitleBarMenu=new D("ChatTitleBarMenu"),D.ChatAttachmentsContext=new D("ChatAttachmentsContext"),D.ChatToolOutputResourceToolbar=new D("ChatToolOutputResourceToolbar"),D.ChatTextEditorMenu=new D("ChatTextEditorMenu"),D.ChatToolOutputResourceContext=new D("ChatToolOutputResourceContext"),D.ChatMultiDiffContext=new D("ChatMultiDiffContext"),D.ChatSessionsMenu=new D("ChatSessionsMenu"),D.ChatSessionsCreateSubMenu=new D("ChatSessionsCreateSubMenu"),D.ChatConfirmationMenu=new D("ChatConfirmationMenu"),D.ChatEditorInlineExecute=new D("ChatEditorInputExecute"),D.ChatEditorInlineInputSide=new D("ChatEditorInputSide"),D.AccessibleView=new D("AccessibleView"),D.MultiDiffEditorFileToolbar=new D("MultiDiffEditorFileToolbar"),D.DiffEditorHunkToolbar=new D("DiffEditorHunkToolbar"),D.DiffEditorSelectionToolbar=new D("DiffEditorSelectionToolbar");let fe=D;const da=qe("menuService"),Sy=class Sy{static for(e){let t=this._all.get(e);return t||(t=new Sy(e),this._all.set(e,t)),t}static merge(e){const t=new Set;for(const i of e)i instanceof Sy&&t.add(i.id);return t}constructor(e){this.id=e,this.has=t=>t===e}};Sy._all=new Map;let gf=Sy;const En=new class{constructor(){this._commands=new Map,this._menuItems=new Map,this._onDidChangeMenu=new Cde({merge:gf.merge}),this.onDidChangeMenu=this._onDidChangeMenu.event}addCommand(s){return this._commands.set(s.id,s),this._onDidChangeMenu.fire(gf.for(fe.CommandPalette)),pe(()=>{this._commands.delete(s.id)&&this._onDidChangeMenu.fire(gf.for(fe.CommandPalette))})}getCommand(s){return this._commands.get(s)}getCommands(){const s=new Map;return this._commands.forEach((e,t)=>s.set(t,e)),s}appendMenuItem(s,e){let t=this._menuItems.get(s);t||(t=new Ns,this._menuItems.set(s,t));const i=t.push(e);return this._onDidChangeMenu.fire(gf.for(s)),pe(()=>{i(),this._onDidChangeMenu.fire(gf.for(s))})}appendMenuItems(s){const e=new U;for(const{id:t,item:i}of s)e.add(this.appendMenuItem(t,i));return e}getMenuItems(s){let e;return this._menuItems.has(s)?e=[...this._menuItems.get(s)]:e=[],s===fe.CommandPalette&&this._appendImplicitItems(e),e}_appendImplicitItems(s){const e=new Set;for(const t of s)aw(t)&&(e.add(t.command.id),t.alt&&e.add(t.alt.id));this._commands.forEach((t,i)=>{e.has(i)||s.push({command:t})})}};class ap extends ov{constructor(e,t,i){super(`submenuitem.${e.submenu.id}`,typeof e.title=="string"?e.title:e.title.value,i,"submenu"),this.item=e,this.hideActions=t}}let yr=xk=class{static label(e,t){return t?.renderShortTitle&&e.shortTitle?typeof e.shortTitle=="string"?e.shortTitle:e.shortTitle.value:typeof e.title=="string"?e.title:e.title.value}constructor(e,t,i,n,o,r,a){this.hideActions=n,this.menuKeybinding=o,this._commandService=a,this.id=e.id,this.label=xk.label(e,i),this.tooltip=(typeof e.tooltip=="string"?e.tooltip:e.tooltip?.value)??"",this.enabled=!e.precondition||r.contextMatchesRules(e.precondition),this.checked=void 0;let l;if(e.toggled){const c=e.toggled.condition?e.toggled:{condition:e.toggled};this.checked=r.contextMatchesRules(c.condition),this.checked&&c.tooltip&&(this.tooltip=typeof c.tooltip=="string"?c.tooltip:c.tooltip.value),this.checked&&xe.isThemeIcon(c.icon)&&(l=c.icon),this.checked&&c.title&&(this.label=typeof c.title=="string"?c.title:c.title.value)}l||(l=xe.isThemeIcon(e.icon)?e.icon:void 0),this.item=e,this.alt=t?new xk(t,void 0,i,n,void 0,r,a):void 0,this._options=i,this.class=l&&xe.asClassName(l)}run(...e){let t=[];return this._options?.arg&&(t=[...t,this._options.arg]),this._options?.shouldForwardArgs&&(t=[...t,...e]),this._commandService.executeCommand(this.id,...t)}};yr=xk=Hue([n$(5,Ae),n$(6,Gt)],yr);class Tn{constructor(e){this.desc=e}}function It(s){const e=[],t=new s,{f1:i,menu:n,keybinding:o,...r}=t.desc;if(at.getCommand(r.id))throw new Error(`Cannot register two commands with the same id: ${r.id}`);if(e.push(at.registerCommand({id:r.id,handler:(a,...l)=>t.run(a,...l),metadata:r.metadata??{description:t.desc.title}})),Array.isArray(n))for(const a of n)e.push(En.appendMenuItem(a.id,{command:{...r,precondition:a.precondition===null?void 0:r.precondition},...a}));else n&&e.push(En.appendMenuItem(n.id,{command:{...r,precondition:n.precondition===null?void 0:r.precondition},...n}));if(i&&(e.push(En.appendMenuItem(fe.CommandPalette,{command:r,when:r.precondition})),e.push(En.addCommand(r))),Array.isArray(o))for(const a of o)e.push(Dn.registerKeybindingRule({...a,id:r.id,when:r.precondition?K.and(r.precondition,a.when):a.when}));else o&&e.push(Dn.registerKeybindingRule({...o,id:r.id,when:r.precondition?K.and(r.precondition,o.when):o.when}));return{dispose(){xt(e)}}}const vs=qe("telemetryService");var s$,o$,r$;class zue{constructor(e,t){this.uri=e,this.value=t}}function $ue(s){return Array.isArray(s)}const Nf=class Nf{constructor(e,t){if(this[s$]="ResourceMap",e instanceof Nf)this.map=new Map(e.map),this.toKey=t??Nf.defaultToKey;else if($ue(e)){this.map=new Map,this.toKey=t??Nf.defaultToKey;for(const[i,n]of e)this.set(i,n)}else this.map=new Map,this.toKey=e??Nf.defaultToKey}set(e,t){return this.map.set(this.toKey(e),new zue(e,t)),this}get(e){return this.map.get(this.toKey(e))?.value}has(e){return this.map.has(this.toKey(e))}get size(){return this.map.size}clear(){this.map.clear()}delete(e){return this.map.delete(this.toKey(e))}forEach(e,t){typeof t<"u"&&(e=e.bind(t));for(const[i,n]of this.map)e(n.value,n.uri,this)}*values(){for(const e of this.map.values())yield e.value}*keys(){for(const e of this.map.values())yield e.uri}*entries(){for(const e of this.map.values())yield[e.uri,e.value]}*[(s$=Symbol.toStringTag,Symbol.iterator)](){for(const[,e]of this.map)yield[e.uri,e.value]}};Nf.defaultToKey=e=>e.toString();let Wi=Nf;class Uue{constructor(e,t){this[o$]="ResourceSet",!e||typeof e=="function"?this._map=new Wi(e):(this._map=new Wi(t),e.forEach(this.add,this))}get size(){return this._map.size}add(e){return this._map.set(e,e),this}clear(){this._map.clear()}delete(e){return this._map.delete(e)}forEach(e,t){this._map.forEach((i,n)=>e.call(t,n,n,this))}has(e){return this._map.has(e)}entries(){return this._map.entries()}keys(){return this._map.keys()}values(){return this._map.keys()}[(o$=Symbol.toStringTag,Symbol.iterator)](){return this.keys()}}class que{constructor(){this[r$]="LinkedMap",this._map=new Map,this._head=void 0,this._tail=void 0,this._size=0,this._state=0}clear(){this._map.clear(),this._head=void 0,this._tail=void 0,this._size=0,this._state++}isEmpty(){return!this._head&&!this._tail}get size(){return this._size}get first(){return this._head?.value}get last(){return this._tail?.value}has(e){return this._map.has(e)}get(e,t=0){const i=this._map.get(e);if(i)return t!==0&&this.touch(i,t),i.value}set(e,t,i=0){let n=this._map.get(e);if(n)n.value=t,i!==0&&this.touch(n,i);else{switch(n={key:e,value:t,next:void 0,previous:void 0},i){case 0:this.addItemLast(n);break;case 1:this.addItemFirst(n);break;case 2:this.addItemLast(n);break;default:this.addItemLast(n);break}this._map.set(e,n),this._size++}return this}delete(e){return!!this.remove(e)}remove(e){const t=this._map.get(e);if(t)return this._map.delete(e),this.removeItem(t),this._size--,t.value}shift(){if(!this._head&&!this._tail)return;if(!this._head||!this._tail)throw new Error("Invalid list");const e=this._head;return this._map.delete(e.key),this.removeItem(e),this._size--,e.value}forEach(e,t){const i=this._state;let n=this._head;for(;n;){if(t?e.bind(t)(n.value,n.key,this):e(n.value,n.key,this),this._state!==i)throw new Error("LinkedMap got modified during iteration.");n=n.next}}keys(){const e=this,t=this._state;let i=this._head;const n={[Symbol.iterator](){return n},next(){if(e._state!==t)throw new Error("LinkedMap got modified during iteration.");if(i){const o={value:i.key,done:!1};return i=i.next,o}else return{value:void 0,done:!0}}};return n}values(){const e=this,t=this._state;let i=this._head;const n={[Symbol.iterator](){return n},next(){if(e._state!==t)throw new Error("LinkedMap got modified during iteration.");if(i){const o={value:i.value,done:!1};return i=i.next,o}else return{value:void 0,done:!0}}};return n}entries(){const e=this,t=this._state;let i=this._head;const n={[Symbol.iterator](){return n},next(){if(e._state!==t)throw new Error("LinkedMap got modified during iteration.");if(i){const o={value:[i.key,i.value],done:!1};return i=i.next,o}else return{value:void 0,done:!0}}};return n}[(r$=Symbol.toStringTag,Symbol.iterator)](){return this.entries()}trimOld(e){if(e>=this.size)return;if(e===0){this.clear();return}let t=this._head,i=this.size;for(;t&&i>e;)this._map.delete(t.key),t=t.next,i--;this._head=t,this._size=i,t&&(t.previous=void 0),this._state++}trimNew(e){if(e>=this.size)return;if(e===0){this.clear();return}let t=this._tail,i=this.size;for(;t&&i>e;)this._map.delete(t.key),t=t.previous,i--;this._tail=t,this._size=i,t&&(t.next=void 0),this._state++}addItemFirst(e){if(!this._head&&!this._tail)this._tail=e;else if(this._head)e.next=this._head,this._head.previous=e;else throw new Error("Invalid list");this._head=e,this._state++}addItemLast(e){if(!this._head&&!this._tail)this._head=e;else if(this._tail)e.previous=this._tail,this._tail.next=e;else throw new Error("Invalid list");this._tail=e,this._state++}removeItem(e){if(e===this._head&&e===this._tail)this._head=void 0,this._tail=void 0;else if(e===this._head){if(!e.next)throw new Error("Invalid list");e.next.previous=void 0,this._head=e.next}else if(e===this._tail){if(!e.previous)throw new Error("Invalid list");e.previous.next=void 0,this._tail=e.previous}else{const t=e.next,i=e.previous;if(!t||!i)throw new Error("Invalid list");t.previous=i,i.next=t}e.next=void 0,e.previous=void 0,this._state++}touch(e,t){if(!this._head||!this._tail)throw new Error("Invalid list");if(!(t!==1&&t!==2)){if(t===1){if(e===this._head)return;const i=e.next,n=e.previous;e===this._tail?(n.next=void 0,this._tail=n):(i.previous=n,n.next=i),e.previous=void 0,e.next=this._head,this._head.previous=e,this._head=e,this._state++}else if(t===2){if(e===this._tail)return;const i=e.next,n=e.previous;e===this._head?(i.previous=void 0,this._head=i):(i.previous=n,n.next=i),e.next=void 0,e.previous=this._tail,this._tail.next=e,this._tail=e,this._state++}}}toJSON(){const e=[];return this.forEach((t,i)=>{e.push([i,t])}),e}fromJSON(e){this.clear();for(const[t,i]of e)this.set(t,i)}}class jue extends que{constructor(e,t=1){super(),this._limit=e,this._ratio=Math.min(Math.max(0,t),1)}get limit(){return this._limit}set limit(e){this._limit=e,this.checkTrim()}get(e,t=2){return super.get(e,t)}peek(e){return super.get(e,0)}set(e,t){return super.set(e,t,2),this}checkTrim(){this.size>this._limit&&this.trim(Math.round(this._limit*this._ratio))}}class Va extends jue{constructor(e,t=1){super(e,t)}trim(e){this.trimOld(e)}set(e,t){return super.set(e,t),this.checkTrim(),this}}class Gue{constructor(e){if(this._m1=new Map,this._m2=new Map,e)for(const[t,i]of e)this.set(t,i)}clear(){this._m1.clear(),this._m2.clear()}set(e,t){this._m1.set(e,t),this._m2.set(t,e)}get(e){return this._m1.get(e)}getKey(e){return this._m2.get(e)}delete(e){const t=this._m1.get(e);return t===void 0?!1:(this._m1.delete(e),this._m2.delete(t),!0)}keys(){return this._m1.keys()}values(){return this._m1.values()}}class zQ{constructor(){this.map=new Map}add(e,t){let i=this.map.get(e);i||(i=new Set,this.map.set(e,i)),i.add(t)}delete(e,t){const i=this.map.get(e);i&&(i.delete(t),i.size===0&&this.map.delete(e))}forEach(e,t){const i=this.map.get(e);i&&i.forEach(t)}}class VR{constructor(){this._data=new Map}set(e,...t){let i=this._data;for(let n=0;n<t.length-1;n++)i.has(t[n])||i.set(t[n],new Map),i=i.get(t[n]);i.set(t[t.length-1],e)}get(...e){let t=this._data;for(let i=0;i<e.length-1;i++){if(!t.has(e[i]))return;t=t.get(e[i])}return t.get(e[e.length-1])}clear(){this._data.clear()}toString(){const e=(t,i)=>{let n="";for(const[o,r]of t)n+=`${"  ".repeat(i)}${o}: `,r instanceof Map?n+=`
`+e(r,i+1):n+=`${r}
`;return n};return e(this._data,0)}}function Wh(s){return s===47||s===92}function $Q(s){return s.replace(/[\\/]/g,Zi.sep)}function Kue(s){return s.indexOf("/")===-1&&(s=$Q(s)),/^[a-zA-Z]:(\/|$)/.test(s)&&(s="/"+s),s}function a$(s,e=Zi.sep){if(!s)return"";const t=s.length,i=s.charCodeAt(0);if(Wh(i)){if(Wh(s.charCodeAt(1))&&!Wh(s.charCodeAt(2))){let o=3;const r=o;for(;o<t&&!Wh(s.charCodeAt(o));o++);if(r!==o&&!Wh(s.charCodeAt(o+1))){for(o+=1;o<t;o++)if(Wh(s.charCodeAt(o)))return s.slice(0,o+1).replace(/[\\/]/g,e)}}return e}else if(UQ(i)&&s.charCodeAt(1)===58)return Wh(s.charCodeAt(2))?s.slice(0,2)+e:s.slice(0,2);let n=s.indexOf("://");if(n!==-1){for(n+=3;n<t;n++)if(Wh(s.charCodeAt(n)))return s.slice(0,n+1)}return""}function NP(s,e,t,i=bl){if(s===e)return!0;if(!s||!e||e.length>s.length)return!1;if(t){if(!j6(s,e))return!1;if(e.length===s.length)return!0;let o=e.length;return e.charAt(e.length-1)===i&&o--,s.charAt(o)===i}return e.charAt(e.length-1)!==i&&(e+=i),s.indexOf(e)===0}function UQ(s){return s>=65&&s<=90||s>=97&&s<=122}function Yue(s,e=Bn){return e?UQ(s.charCodeAt(0))&&s.charCodeAt(1)===58:!1}function $l(s){return wE(s,!0)}class Zue{constructor(e){this._ignorePathCasing=e}compare(e,t,i=!1){return e===t?0:qy(this.getComparisonKey(e,i),this.getComparisonKey(t,i))}isEqual(e,t,i=!1){return e===t?!0:!e||!t?!1:this.getComparisonKey(e,i)===this.getComparisonKey(t,i)}getComparisonKey(e,t=!1){return e.with({path:this._ignorePathCasing(e)?e.path.toLowerCase():void 0,fragment:t?null:void 0}).toString()}isEqualOrParent(e,t,i=!1){if(e.scheme===t.scheme){if(e.scheme===Te.file)return NP($l(e),$l(t),this._ignorePathCasing(e))&&e.query===t.query&&(i||e.fragment===t.fragment);if(c$(e.authority,t.authority))return NP(e.path,t.path,this._ignorePathCasing(e),"/")&&e.query===t.query&&(i||e.fragment===t.fragment)}return!1}joinPath(e,...t){return Le.joinPath(e,...t)}basenameOrAuthority(e){return ra(e)||e.authority}basename(e){return Zi.basename(e.path)}extname(e){return Zi.extname(e.path)}dirname(e){if(e.path.length===0)return e;let t;return e.scheme===Te.file?t=Le.file(vk($l(e))).path:(t=Zi.dirname(e.path),e.authority&&t.length&&t.charCodeAt(0)!==47&&(console.error(`dirname("${e.toString})) resulted in a relative path`),t="/")),e.with({path:t})}normalizePath(e){if(!e.path.length)return e;let t;return e.scheme===Te.file?t=Le.file(lQ($l(e))).path:t=Zi.normalize(e.path),e.with({path:t})}relativePath(e,t){if(e.scheme!==t.scheme||!c$(e.authority,t.authority))return;if(e.scheme===Te.file){const o=she($l(e),$l(t));return Bn?$Q(o):o}let i=e.path||"/";const n=t.path||"/";if(this._ignorePathCasing(e)){let o=0;for(const r=Math.min(i.length,n.length);o<r&&!(i.charCodeAt(o)!==n.charCodeAt(o)&&i.charAt(o).toLowerCase()!==n.charAt(o).toLowerCase());o++);i=n.substr(0,o)+i.substr(o)}return Zi.relative(i,n)}resolvePath(e,t){if(e.scheme===Te.file){const i=Le.file(nhe($l(e),t));return e.with({authority:i.authority,path:i.path})}return t=Kue(t),e.with({path:Zi.resolve(e.path,t)})}isAbsolutePath(e){return!!e.path&&e.path[0]==="/"}isEqualAuthority(e,t){return e===t||e!==void 0&&t!==void 0&&Hf(e,t)}hasTrailingPathSeparator(e,t=bl){if(e.scheme===Te.file){const i=$l(e);return i.length>a$(i).length&&i[i.length-1]===t}else{const i=e.path;return i.length>1&&i.charCodeAt(i.length-1)===47&&!/^[a-zA-Z]:(\/$|\\$)/.test(e.fsPath)}}removeTrailingPathSeparator(e,t=bl){return d$(e,t)?e.with({path:e.path.substr(0,e.path.length-1)}):e}addTrailingPathSeparator(e,t=bl){let i=!1;if(e.scheme===Te.file){const n=$l(e);i=n!==void 0&&n.length===a$(n).length&&n[n.length-1]===t}else{t="/";const n=e.path;i=n.length===1&&n.charCodeAt(n.length-1)===47}return!i&&!d$(e,t)?e.with({path:e.path+"/"}):e}}const oi=new Zue(()=>!1),dg=oi.isEqual.bind(oi);oi.isEqualOrParent.bind(oi);oi.getComparisonKey.bind(oi);const Xue=oi.basenameOrAuthority.bind(oi),ra=oi.basename.bind(oi),Que=oi.extname.bind(oi),zR=oi.dirname.bind(oi),qQ=oi.joinPath.bind(oi),Jue=oi.normalizePath.bind(oi),ege=oi.relativePath.bind(oi),l$=oi.resolvePath.bind(oi);oi.isAbsolutePath.bind(oi);const c$=oi.isEqualAuthority.bind(oi),d$=oi.hasTrailingPathSeparator.bind(oi);oi.removeTrailingPathSeparator.bind(oi);oi.addTrailingPathSeparator.bind(oi);var hg;(function(s){s.META_DATA_LABEL="label",s.META_DATA_DESCRIPTION="description",s.META_DATA_SIZE="size",s.META_DATA_MIME="mime";function e(t){const i=new Map;t.path.substring(t.path.indexOf(";")+1,t.path.lastIndexOf(";")).split(";").forEach(r=>{const[a,l]=r.split(":");a&&l&&i.set(a,l)});const o=t.path.substring(0,t.path.indexOf(";"));return o&&i.set(s.META_DATA_MIME,o),i}s.parseMetaData=e})(hg||(hg={}));const jt=qe("logService"),jQ=qe("loggerService");var pn;(function(s){s[s.Off=0]="Off",s[s.Trace=1]="Trace",s[s.Debug=2]="Debug",s[s.Info=3]="Info",s[s.Warning=4]="Warning",s[s.Error=5]="Error"})(pn||(pn={}));const GQ=pn.Info;function KQ(s,e){return s!==pn.Off&&s<=e}class YQ extends F{constructor(){super(...arguments),this.level=GQ,this._onDidChangeLogLevel=this._register(new M)}get onDidChangeLogLevel(){return this._onDidChangeLogLevel.event}setLevel(e){this.level!==e&&(this.level=e,this._onDidChangeLogLevel.fire(this.level))}getLevel(){return this.level}checkLogLevel(e){return KQ(this.level,e)}canLog(e){return this._store.isDisposed?!1:this.checkLogLevel(e)}}class tge extends YQ{constructor(e=GQ,t=!0){super(),this.useColors=t,this.setLevel(e)}trace(e,...t){this.canLog(pn.Trace)&&(this.useColors?console.log("%cTRACE","color: #888",e,...t):console.log(e,...t))}debug(e,...t){this.canLog(pn.Debug)&&(this.useColors?console.log("%cDEBUG","background: #eee; color: #888",e,...t):console.log(e,...t))}info(e,...t){this.canLog(pn.Info)&&(this.useColors?console.log("%c INFO","color: #33f",e,...t):console.log(e,...t))}warn(e,...t){this.canLog(pn.Warning)&&(this.useColors?console.warn("%c WARN","color: #993",e,...t):console.log(e,...t))}error(e,...t){this.canLog(pn.Error)&&(this.useColors?console.error("%c  ERR","color: #f33",e,...t):console.error(e,...t))}}class ige extends YQ{constructor(e){super(),this.loggers=e,e.length&&this.setLevel(e[0].getLevel())}setLevel(e){for(const t of this.loggers)t.setLevel(e);super.setLevel(e)}trace(e,...t){for(const i of this.loggers)i.trace(e,...t)}debug(e,...t){for(const i of this.loggers)i.debug(e,...t)}info(e,...t){for(const i of this.loggers)i.info(e,...t)}warn(e,...t){for(const i of this.loggers)i.warn(e,...t)}error(e,...t){for(const i of this.loggers)i.error(e,...t)}dispose(){for(const e of this.loggers)e.dispose();super.dispose()}}class nge extends F{constructor(e,t,i){if(super(),this.logLevel=e,this.logsHome=t,this._loggers=new Wi,this._onDidChangeLoggers=this._register(new M),this._onDidChangeVisibility=this._register(new M),i)for(const n of i)this._loggers.set(n.resource,{logger:void 0,info:n})}getLoggerEntry(e){return bn(e)?[...this._loggers.values()].find(t=>t.info.id===e):this._loggers.get(e)}createLogger(e,t){const i=this.toResource(e),n=bn(e)?e:t?.id??j0(i.toString()).toString(16);let o=this._loggers.get(i)?.logger;const r=t?.logLevel==="always"?pn.Trace:t?.logLevel;o||(o=this.doCreateLogger(i,r??this.getLogLevel(i)??this.logLevel,{...t,id:n}));const a={logger:o,info:{resource:i,id:n,logLevel:r,name:t?.name,hidden:t?.hidden,group:t?.group,extensionId:t?.extensionId,when:t?.when}};return this.registerLogger(a.info),this._loggers.set(i,a),o}toResource(e){return bn(e)?qQ(this.logsHome,`${e}.log`):e}setVisibility(e,t){const i=this.getLoggerEntry(e);i&&t!==!i.info.hidden&&(i.info.hidden=!t,this._loggers.set(i.info.resource,i),this._onDidChangeVisibility.fire([i.info.resource,t]))}getLogLevel(e){let t;return e&&(t=this._loggers.get(e)?.info.logLevel),t??this.logLevel}registerLogger(e){const t=this._loggers.get(e.resource);t?t.info.hidden!==e.hidden&&this.setVisibility(e.resource,!e.hidden):(this._loggers.set(e.resource,{info:e,logger:void 0}),this._onDidChangeLoggers.fire({added:[e],removed:[]}))}dispose(){this._loggers.forEach(e=>e.logger?.dispose()),this._loggers.clear(),super.dispose()}}class sge{constructor(){this.onDidChangeLogLevel=new M().event}setLevel(e){}getLevel(){return pn.Info}trace(e,...t){}debug(e,...t){}info(e,...t){}warn(e,...t){}error(e,...t){}dispose(){}}class oge extends nge{constructor(){super(pn.Off,Le.parse("log:///log"))}doCreateLogger(e,t,i){return new sge}}function rge(s){switch(s){case pn.Trace:return"trace";case pn.Debug:return"debug";case pn.Info:return"info";case pn.Warning:return"warn";case pn.Error:return"error";case pn.Off:return"off"}}new ce("logLevel",rge(pn.Info));const xy=class xy{static getRegisteredCommands(){return[...xy.REGISTERED_COMMANDS]}static registerCommand(e){xy.REGISTERED_COMMANDS.add(e)}};xy.REGISTERED_COMMANDS=new Set;let EE=xy,$R=class{constructor(e){this.id=e.id,this.precondition=e.precondition,this._kbOpts=e.kbOpts,this._menuOpts=e.menuOpts,this.metadata=e.metadata,this.canTriggerInlineEdits=e.canTriggerInlineEdits}register(){if(Array.isArray(this._menuOpts)?this._menuOpts.forEach(this._registerMenuItem,this):this._menuOpts&&this._registerMenuItem(this._menuOpts),this._kbOpts){const e=Array.isArray(this._kbOpts)?this._kbOpts:[this._kbOpts];for(const t of e){let i=t.kbExpr;this.precondition&&(i?i=K.and(i,this.precondition):i=this.precondition);const n={id:this.id,weight:t.weight,args:t.args,when:i,primary:t.primary,secondary:t.secondary,win:t.win,linux:t.linux,mac:t.mac};Dn.registerKeybindingRule(n)}}at.registerCommand({id:this.id,handler:(e,t)=>this.runCommand(e,t),metadata:this.metadata}),this.canTriggerInlineEdits&&EE.registerCommand(this.id)}_registerMenuItem(e){En.appendMenuItem(e.menuId,{group:e.group,command:{id:this.id,title:e.title,icon:e.icon,precondition:this.precondition},when:e.when,order:e.order})}};class Yv extends $R{constructor(){super(...arguments),this._implementations=[]}addImplementation(e,t,i,n){return this._implementations.push({priority:e,name:t,implementation:i,when:n}),this._implementations.sort((o,r)=>r.priority-o.priority),{dispose:()=>{for(let o=0;o<this._implementations.length;o++)if(this._implementations[o].implementation===i){this._implementations.splice(o,1);return}}}}runCommand(e,t){const i=e.get(jt),n=e.get(Ae);i.trace(`Executing Command '${this.id}' which has ${this._implementations.length} bound.`);for(const o of this._implementations){if(o.when){const a=n.getContext(ln());if(!o.when.evaluate(a))continue}const r=o.implementation(e,t);if(r)return i.trace(`Command '${this.id}' was handled by '${o.name}'.`),typeof r=="boolean"?void 0:r}i.trace(`The Command '${this.id}' was not handled by any implementation.`)}}class ZQ extends $R{constructor(e,t){super(t),this.command=e}runCommand(e,t){return this.command.runCommand(e,t)}}class hn extends $R{static bindToContribution(e){return class extends hn{constructor(i){super(i),this._callback=i.handler}runEditorCommand(i,n,o){const r=e(n);r&&this._callback(r,o)}}}static runEditorCommand(e,t,i,n){const o=e.get(ht),r=o.getFocusedCodeEditor()||o.getActiveCodeEditor();if(r)return r.invokeWithinContext(a=>{if(a.get(Ae).contextMatchesRules(i??void 0))return n(a,r,t)})}runCommand(e,t){return hn.runEditorCommand(e,t,this.precondition,(i,n,o)=>this.runEditorCommand(i,n,o))}}class ue extends hn{static convertOptions(e){let t;Array.isArray(e.menuOpts)?t=e.menuOpts:e.menuOpts?t=[e.menuOpts]:t=[];function i(n){return n.menuId||(n.menuId=fe.EditorContext),n.title||(n.title=typeof e.label=="string"?e.label:e.label.value),n.when=K.and(e.precondition,n.when),n}return Array.isArray(e.contextMenuOpts)?t.push(...e.contextMenuOpts.map(i)):e.contextMenuOpts&&t.push(i(e.contextMenuOpts)),e.menuOpts=t,e}constructor(e){super(ue.convertOptions(e)),typeof e.label=="string"?(this.label=e.label,this.alias=e.alias??e.label):(this.label=e.label.value,this.alias=e.alias??e.label.original)}runEditorCommand(e,t,i){return this.reportTelemetry(e,t),this.run(e,t,i||{})}reportTelemetry(e,t){e.get(vs).publicLog2("editorActionInvoked",{name:this.label,id:this.id})}}class UR extends ue{constructor(){super(...arguments),this._implementations=[]}addImplementation(e,t){return this._implementations.push([e,t]),this._implementations.sort((i,n)=>n[0]-i[0]),{dispose:()=>{for(let i=0;i<this._implementations.length;i++)if(this._implementations[i][1]===t){this._implementations.splice(i,1);return}}}}run(e,t,i){for(const n of this._implementations){const o=n[1](e,t,i);if(o)return typeof o=="boolean"?void 0:o}}}class za extends Tn{run(e,...t){const i=e.get(ht),n=i.getFocusedCodeEditor()||i.getActiveCodeEditor();if(n)return n.invokeWithinContext(o=>{const r=o.get(Ae),a=o.get(jt);if(!r.contextMatchesRules(this.desc.precondition??void 0)){a.debug("[EditorAction2] NOT running command because its precondition is FALSE",this.desc.id,this.desc.precondition?.serialize());return}return this.runEditorCommand(o,n,...t)})}}function Lo(s,e){at.registerCommand(s,function(t,...i){const n=t.get(_e),[o,r]=i;dt(Le.isUri(o)),dt(A.isIPosition(r));const a=t.get(hi).getModel(o);if(a){const l=A.lift(r);return n.invokeFunction(e,a,l,...i.slice(2))}return t.get(Fs).createModelReference(o).then(l=>new Promise((c,d)=>{try{const h=n.invokeFunction(e,l.object.textEditorModel,A.lift(r),i.slice(2));c(h)}catch(h){d(h)}}).finally(()=>{l.dispose()}))})}function le(s){return qr.INSTANCE.registerEditorCommand(s),s}function ae(s){const e=new s;return qr.INSTANCE.registerEditorAction(e),e}function qR(s){return qr.INSTANCE.registerEditorAction(s),s}function age(s){qr.INSTANCE.registerEditorAction(s)}function lt(s,e,t){qr.INSTANCE.registerEditorContribution(s,e,t)}var lw;(function(s){function e(r){return qr.INSTANCE.getEditorCommand(r)}s.getEditorCommand=e;function t(){return qr.INSTANCE.getEditorActions()}s.getEditorActions=t;function i(){return qr.INSTANCE.getEditorContributions()}s.getEditorContributions=i;function n(r){return qr.INSTANCE.getEditorContributions().filter(a=>r.indexOf(a.id)>=0)}s.getSomeEditorContributions=n;function o(){return qr.INSTANCE.getDiffEditorContributions()}s.getDiffEditorContributions=o})(lw||(lw={}));const lge={EditorCommonContributions:"editor.contributions"},UT=class UT{constructor(){this.editorContributions=[],this.diffEditorContributions=[],this.editorActions=[],this.editorCommands=Object.create(null)}registerEditorContribution(e,t,i){this.editorContributions.push({id:e,ctor:t,instantiation:i})}getEditorContributions(){return this.editorContributions.slice(0)}getDiffEditorContributions(){return this.diffEditorContributions.slice(0)}registerEditorAction(e){e.register(),this.editorActions.push(e)}getEditorActions(){return this.editorActions}registerEditorCommand(e){e.register(),this.editorCommands[e.id]=e}getEditorCommand(e){return this.editorCommands[e]||null}};UT.INSTANCE=new UT;let qr=UT;bi.add(lge.EditorCommonContributions,qr.INSTANCE);function Y0(s){return s.register(),s}const XQ=Y0(new Yv({id:"undo",precondition:void 0,kbOpts:{weight:0,primary:2104},menuOpts:[{menuId:fe.MenubarEditMenu,group:"1_do",title:m(69,"&&Undo"),order:1},{menuId:fe.CommandPalette,group:"",title:m(70,"Undo"),order:1},{menuId:fe.SimpleEditorContext,group:"1_do",title:m(71,"Undo"),order:1}]}));Y0(new ZQ(XQ,{id:"default:undo",precondition:void 0}));const QQ=Y0(new Yv({id:"redo",precondition:void 0,kbOpts:{weight:0,primary:2103,secondary:[3128],mac:{primary:3128}},menuOpts:[{menuId:fe.MenubarEditMenu,group:"1_do",title:m(72,"&&Redo"),order:2},{menuId:fe.CommandPalette,group:"",title:m(73,"Redo"),order:1},{menuId:fe.SimpleEditorContext,group:"1_do",title:m(74,"Redo"),order:2}]}));Y0(new ZQ(QQ,{id:"default:redo",precondition:void 0}));const cge=Y0(new Yv({id:"editor.action.selectAll",precondition:void 0,kbOpts:{weight:0,kbExpr:null,primary:2079},menuOpts:[{menuId:fe.MenubarSelectionMenu,group:"1_basic",title:m(75,"&&Select All"),order:1},{menuId:fe.CommandPalette,group:"",title:m(76,"Select All"),order:1},{menuId:fe.SimpleEditorContext,group:"9_select",title:m(77,"Select All"),order:1}]}));let k=class Nn{constructor(e,t,i,n){e>i||e===i&&t>n?(this.startLineNumber=i,this.startColumn=n,this.endLineNumber=e,this.endColumn=t):(this.startLineNumber=e,this.startColumn=t,this.endLineNumber=i,this.endColumn=n)}isEmpty(){return Nn.isEmpty(this)}static isEmpty(e){return e.startLineNumber===e.endLineNumber&&e.startColumn===e.endColumn}containsPosition(e){return Nn.containsPosition(this,e)}static containsPosition(e,t){return!(t.lineNumber<e.startLineNumber||t.lineNumber>e.endLineNumber||t.lineNumber===e.startLineNumber&&t.column<e.startColumn||t.lineNumber===e.endLineNumber&&t.column>e.endColumn)}static strictContainsPosition(e,t){return!(t.lineNumber<e.startLineNumber||t.lineNumber>e.endLineNumber||t.lineNumber===e.startLineNumber&&t.column<=e.startColumn||t.lineNumber===e.endLineNumber&&t.column>=e.endColumn)}containsRange(e){return Nn.containsRange(this,e)}static containsRange(e,t){return!(t.startLineNumber<e.startLineNumber||t.endLineNumber<e.startLineNumber||t.startLineNumber>e.endLineNumber||t.endLineNumber>e.endLineNumber||t.startLineNumber===e.startLineNumber&&t.startColumn<e.startColumn||t.endLineNumber===e.endLineNumber&&t.endColumn>e.endColumn)}strictContainsRange(e){return Nn.strictContainsRange(this,e)}static strictContainsRange(e,t){return!(t.startLineNumber<e.startLineNumber||t.endLineNumber<e.startLineNumber||t.startLineNumber>e.endLineNumber||t.endLineNumber>e.endLineNumber||t.startLineNumber===e.startLineNumber&&t.startColumn<=e.startColumn||t.endLineNumber===e.endLineNumber&&t.endColumn>=e.endColumn)}plusRange(e){return Nn.plusRange(this,e)}static plusRange(e,t){let i,n,o,r;return t.startLineNumber<e.startLineNumber?(i=t.startLineNumber,n=t.startColumn):t.startLineNumber===e.startLineNumber?(i=t.startLineNumber,n=Math.min(t.startColumn,e.startColumn)):(i=e.startLineNumber,n=e.startColumn),t.endLineNumber>e.endLineNumber?(o=t.endLineNumber,r=t.endColumn):t.endLineNumber===e.endLineNumber?(o=t.endLineNumber,r=Math.max(t.endColumn,e.endColumn)):(o=e.endLineNumber,r=e.endColumn),new Nn(i,n,o,r)}intersectRanges(e){return Nn.intersectRanges(this,e)}static intersectRanges(e,t){let i=e.startLineNumber,n=e.startColumn,o=e.endLineNumber,r=e.endColumn;const a=t.startLineNumber,l=t.startColumn,c=t.endLineNumber,d=t.endColumn;return i<a?(i=a,n=l):i===a&&(n=Math.max(n,l)),o>c?(o=c,r=d):o===c&&(r=Math.min(r,d)),i>o||i===o&&n>r?null:new Nn(i,n,o,r)}equalsRange(e){return Nn.equalsRange(this,e)}static equalsRange(e,t){return!e&&!t?!0:!!e&&!!t&&e.startLineNumber===t.startLineNumber&&e.startColumn===t.startColumn&&e.endLineNumber===t.endLineNumber&&e.endColumn===t.endColumn}getEndPosition(){return Nn.getEndPosition(this)}static getEndPosition(e){return new A(e.endLineNumber,e.endColumn)}getStartPosition(){return Nn.getStartPosition(this)}static getStartPosition(e){return new A(e.startLineNumber,e.startColumn)}toString(){return"["+this.startLineNumber+","+this.startColumn+" -> "+this.endLineNumber+","+this.endColumn+"]"}setEndPosition(e,t){return new Nn(this.startLineNumber,this.startColumn,e,t)}setStartPosition(e,t){return new Nn(e,t,this.endLineNumber,this.endColumn)}collapseToStart(){return Nn.collapseToStart(this)}static collapseToStart(e){return new Nn(e.startLineNumber,e.startColumn,e.startLineNumber,e.startColumn)}collapseToEnd(){return Nn.collapseToEnd(this)}static collapseToEnd(e){return new Nn(e.endLineNumber,e.endColumn,e.endLineNumber,e.endColumn)}delta(e){return new Nn(this.startLineNumber+e,this.startColumn,this.endLineNumber+e,this.endColumn)}isSingleLine(){return this.startLineNumber===this.endLineNumber}static fromPositions(e,t=e){return new Nn(e.lineNumber,e.column,t.lineNumber,t.column)}static lift(e){return e?new Nn(e.startLineNumber,e.startColumn,e.endLineNumber,e.endColumn):null}static isIRange(e){return!!e&&typeof e.startLineNumber=="number"&&typeof e.startColumn=="number"&&typeof e.endLineNumber=="number"&&typeof e.endColumn=="number"}static areIntersectingOrTouching(e,t){return!(e.endLineNumber<t.startLineNumber||e.endLineNumber===t.startLineNumber&&e.endColumn<t.startColumn||t.endLineNumber<e.startLineNumber||t.endLineNumber===e.startLineNumber&&t.endColumn<e.startColumn)}static areIntersecting(e,t){return!(e.endLineNumber<t.startLineNumber||e.endLineNumber===t.startLineNumber&&e.endColumn<=t.startColumn||t.endLineNumber<e.startLineNumber||t.endLineNumber===e.startLineNumber&&t.endColumn<=e.startColumn)}static areOnlyIntersecting(e,t){return!(e.endLineNumber<t.startLineNumber-1||e.endLineNumber===t.startLineNumber&&e.endColumn<t.startColumn-1||t.endLineNumber<e.startLineNumber-1||t.endLineNumber===e.startLineNumber&&t.endColumn<e.startColumn-1)}static compareRangesUsingStarts(e,t){if(e&&t){const o=e.startLineNumber|0,r=t.startLineNumber|0;if(o===r){const a=e.startColumn|0,l=t.startColumn|0;if(a===l){const c=e.endLineNumber|0,d=t.endLineNumber|0;if(c===d){const h=e.endColumn|0,u=t.endColumn|0;return h-u}return c-d}return a-l}return o-r}return(e?1:0)-(t?1:0)}static compareRangesUsingEnds(e,t){return e.endLineNumber===t.endLineNumber?e.endColumn===t.endColumn?e.startLineNumber===t.startLineNumber?e.startColumn-t.startColumn:e.startLineNumber-t.startLineNumber:e.endColumn-t.endColumn:e.endLineNumber-t.endLineNumber}static spansMultipleLines(e){return e.endLineNumber>e.startLineNumber}toJSON(){return this}},he=class ma extends k{constructor(e,t,i,n){super(e,t,i,n),this.selectionStartLineNumber=e,this.selectionStartColumn=t,this.positionLineNumber=i,this.positionColumn=n}toString(){return"["+this.selectionStartLineNumber+","+this.selectionStartColumn+" -> "+this.positionLineNumber+","+this.positionColumn+"]"}equalsSelection(e){return ma.selectionsEqual(this,e)}static selectionsEqual(e,t){return e.selectionStartLineNumber===t.selectionStartLineNumber&&e.selectionStartColumn===t.selectionStartColumn&&e.positionLineNumber===t.positionLineNumber&&e.positionColumn===t.positionColumn}getDirection(){return this.selectionStartLineNumber===this.startLineNumber&&this.selectionStartColumn===this.startColumn?0:1}setEndPosition(e,t){return this.getDirection()===0?new ma(this.startLineNumber,this.startColumn,e,t):new ma(e,t,this.startLineNumber,this.startColumn)}getPosition(){return new A(this.positionLineNumber,this.positionColumn)}getSelectionStart(){return new A(this.selectionStartLineNumber,this.selectionStartColumn)}setStartPosition(e,t){return this.getDirection()===0?new ma(e,t,this.endLineNumber,this.endColumn):new ma(this.endLineNumber,this.endColumn,e,t)}static fromPositions(e,t=e){return new ma(e.lineNumber,e.column,t.lineNumber,t.column)}static fromRange(e,t){return t===0?new ma(e.startLineNumber,e.startColumn,e.endLineNumber,e.endColumn):new ma(e.endLineNumber,e.endColumn,e.startLineNumber,e.startColumn)}static liftSelection(e){return new ma(e.selectionStartLineNumber,e.selectionStartColumn,e.positionLineNumber,e.positionColumn)}static selectionsArrEqual(e,t){if(e&&!t||!e&&t)return!1;if(!e&&!t)return!0;if(e.length!==t.length)return!1;for(let i=0,n=e.length;i<n;i++)if(!this.selectionsEqual(e[i],t[i]))return!1;return!0}static isISelection(e){return!!e&&typeof e.selectionStartLineNumber=="number"&&typeof e.selectionStartColumn=="number"&&typeof e.positionLineNumber=="number"&&typeof e.positionColumn=="number"}static createWithDirection(e,t,i,n,o){return o===0?new ma(e,t,i,n):new ma(i,n,e,t)}};function Uf(s,e){const t=s.getCount(),i=s.findTokenIndexAtOffset(e),n=s.getLanguageId(i);let o=i;for(;o+1<t&&s.getLanguageId(o+1)===n;)o++;let r=i;for(;r>0&&s.getLanguageId(r-1)===n;)r--;return new dge(s,n,r,o+1,s.getStartOffset(r),s.getEndOffset(o))}class dge{constructor(e,t,i,n,o,r){this._scopedLineTokensBrand=void 0,this._actual=e,this.languageId=t,this._firstTokenIndex=i,this._lastTokenIndex=n,this.firstCharOffset=o,this._lastCharOffset=r,this.languageIdCodec=e.languageIdCodec}getLineContent(){return this._actual.getLineContent().substring(this.firstCharOffset,this._lastCharOffset)}getLineLength(){return this._lastCharOffset-this.firstCharOffset}getActualLineContentBefore(e){return this._actual.getLineContent().substring(0,this.firstCharOffset+e)}getTokenCount(){return this._lastTokenIndex-this._firstTokenIndex}findTokenIndexAtOffset(e){return this._actual.findTokenIndexAtOffset(e+this.firstCharOffset)-this._firstTokenIndex}getStandardTokenType(e){return this._actual.getStandardTokenType(e+this._firstTokenIndex)}toIViewLineTokens(){return this._actual.sliceAndInflate(this.firstCharOffset,this._lastCharOffset,0)}}function Ul(s){return(s&3)!==0}class pi{static _nextVisibleColumn(e,t,i){return e===9?pi.nextRenderTabStop(t,i):gh(e)||Z6(e)?t+2:t+1}static visibleColumnFromColumn(e,t,i){const n=Math.min(t-1,e.length),o=e.substring(0,n),r=new fE(o);let a=0;for(;!r.eol();){const l=gE(o,n,r.offset);r.nextGraphemeLength(),a=this._nextVisibleColumn(l,a,i)}return a}static columnFromVisibleColumn(e,t,i){if(t<=0)return 1;const n=e.length,o=new fE(e);let r=0,a=1;for(;!o.eol();){const l=gE(e,n,o.offset);o.nextGraphemeLength();const c=this._nextVisibleColumn(l,r,i),d=o.offset+1;if(c>=t){const h=t-r;return c-t<h?d:a}r=c,a=d}return n+1}static nextRenderTabStop(e,t){return e+t-e%t}static nextIndentTabStop(e,t){return pi.nextRenderTabStop(e,t)}static prevRenderTabStop(e,t){return Math.max(0,e-1-(e-1)%t)}static prevIndentTabStop(e,t){return pi.prevRenderTabStop(e,t)}}function hge(s,e,t){let i=0;for(let o=0;o<s.length;o++)s.charAt(o)==="	"?i=pi.nextIndentTabStop(i,e):i++;let n="";if(!t){const o=Math.floor(i/e);i=i%e;for(let r=0;r<o;r++)n+="	"}for(let o=0;o<i;o++)n+=" ";return n}function fW(s,e,t){let i=As(s);return i===-1&&(i=s.length),hge(s.substring(0,i),e,t)+s.substring(i)}class uge{constructor(){this._inputMode="insert",this._onDidChangeInputMode=new M,this.onDidChangeInputMode=this._onDidChangeInputMode.event}getInputMode(){return this._inputMode}}const RP=new uge,gge=()=>!0,fge=()=>!1,pge=s=>s===" "||s==="	";class B_{static shouldRecreate(e){return e.hasChanged(165)||e.hasChanged(148)||e.hasChanged(45)||e.hasChanged(85)||e.hasChanged(88)||e.hasChanged(89)||e.hasChanged(10)||e.hasChanged(11)||e.hasChanged(15)||e.hasChanged(13)||e.hasChanged(14)||e.hasChanged(20)||e.hasChanged(145)||e.hasChanged(141)||e.hasChanged(59)||e.hasChanged(104)||e.hasChanged(147)||e.hasChanged(93)}constructor(e,t,i,n){this.languageConfigurationService=n,this._cursorMoveConfigurationBrand=void 0,this._languageId=e;const o=i.options,r=o.get(165),a=o.get(59);this.readOnly=o.get(104),this.tabSize=t.tabSize,this.indentSize=t.indentSize,this.insertSpaces=t.insertSpaces,this.stickyTabStops=o.get(132),this.lineHeight=a.lineHeight,this.typicalHalfwidthCharacterWidth=a.typicalHalfwidthCharacterWidth,this.pageSize=Math.max(1,Math.floor(r.height/this.lineHeight)-2),this.useTabStops=o.get(145),this.trimWhitespaceOnDelete=o.get(141),this.wordSeparators=o.get(148),this.emptySelectionClipboard=o.get(45),this.copyWithSyntaxHighlighting=o.get(31),this.multiCursorMergeOverlapping=o.get(85),this.multiCursorPaste=o.get(88),this.multiCursorLimit=o.get(89),this.autoClosingBrackets=o.get(10),this.autoClosingComments=o.get(11),this.autoClosingQuotes=o.get(15),this.autoClosingDelete=o.get(13),this.autoClosingOvertype=o.get(14),this.autoSurround=o.get(20),this.autoIndent=o.get(16),this.wordSegmenterLocales=o.get(147),this.overtypeOnPaste=o.get(93),this.surroundingPairs={},this._electricChars=null,this.shouldAutoCloseBefore={quote:this._getShouldAutoClose(e,this.autoClosingQuotes,!0),comment:this._getShouldAutoClose(e,this.autoClosingComments,!1),bracket:this._getShouldAutoClose(e,this.autoClosingBrackets,!1)},this.autoClosingPairs=this.languageConfigurationService.getLanguageConfiguration(e).getAutoClosingPairs();const l=this.languageConfigurationService.getLanguageConfiguration(e).getSurroundingPairs();if(l)for(const d of l)this.surroundingPairs[d.open]=d.close;const c=this.languageConfigurationService.getLanguageConfiguration(e).comments;this.blockCommentStartToken=c?.blockCommentStartToken??null}get electricChars(){if(!this._electricChars){this._electricChars={};const e=this.languageConfigurationService.getLanguageConfiguration(this._languageId).electricCharacter?.getElectricCharacters();if(e)for(const t of e)this._electricChars[t]=!0}return this._electricChars}get inputMode(){return RP.getInputMode()}onElectricCharacter(e,t,i){const n=Uf(t,i-1),o=this.languageConfigurationService.getLanguageConfiguration(n.languageId).electricCharacter;return o?o.onElectricCharacter(e,n,i-n.firstCharOffset):null}normalizeIndentation(e){return fW(e,this.indentSize,this.insertSpaces)}_getShouldAutoClose(e,t,i){switch(t){case"beforeWhitespace":return pge;case"languageDefined":return this._getLanguageDefinedShouldAutoClose(e,i);case"always":return gge;case"never":return fge}}_getLanguageDefinedShouldAutoClose(e,t){const i=this.languageConfigurationService.getLanguageConfiguration(e).getAutoCloseBeforeSet(t);return n=>i.indexOf(n)!==-1}visibleColumnFromColumn(e,t){return pi.visibleColumnFromColumn(e.getLineContent(t.lineNumber),t.column,this.tabSize)}columnFromVisibleColumn(e,t,i){const n=pi.columnFromVisibleColumn(e.getLineContent(t),i,this.tabSize),o=e.getLineMinColumn(t);if(n<o)return o;const r=e.getLineMaxColumn(t);return n>r?r:n}}let ut=class JQ{static fromModelState(e){return new mge(e)}static fromViewState(e){return new _ge(e)}static fromModelSelection(e){const t=he.liftSelection(e),i=new An(k.fromPositions(t.getSelectionStart()),0,0,t.getPosition(),0);return JQ.fromModelState(i)}static fromModelSelections(e){const t=[];for(let i=0,n=e.length;i<n;i++)t[i]=this.fromModelSelection(e[i]);return t}constructor(e,t){this._cursorStateBrand=void 0,this.modelState=e,this.viewState=t}equals(e){return this.viewState.equals(e.viewState)&&this.modelState.equals(e.modelState)}};class mge{constructor(e){this.modelState=e,this.viewState=null}}class _ge{constructor(e){this.modelState=null,this.viewState=e}}class An{constructor(e,t,i,n,o){this.selectionStart=e,this.selectionStartKind=t,this.selectionStartLeftoverVisibleColumns=i,this.position=n,this.leftoverVisibleColumns=o,this._singleCursorStateBrand=void 0,this.selection=An._computeSelection(this.selectionStart,this.position)}equals(e){return this.selectionStartLeftoverVisibleColumns===e.selectionStartLeftoverVisibleColumns&&this.leftoverVisibleColumns===e.leftoverVisibleColumns&&this.selectionStartKind===e.selectionStartKind&&this.position.equals(e.position)&&this.selectionStart.equalsRange(e.selectionStart)}hasSelection(){return!this.selection.isEmpty()||!this.selectionStart.isEmpty()}move(e,t,i,n){return e?new An(this.selectionStart,this.selectionStartKind,this.selectionStartLeftoverVisibleColumns,new A(t,i),n):new An(new k(t,i,t,i),0,n,new A(t,i),n)}static _computeSelection(e,t){return e.isEmpty()||!t.isBeforeOrEqual(e.getStartPosition())?he.fromPositions(e.getStartPosition(),t):he.fromPositions(e.getEndPosition(),t)}}class to{constructor(e,t,i){this._editOperationResultBrand=void 0,this.type=e,this.commands=t,this.shouldPushStackElementBefore=i.shouldPushStackElementBefore,this.shouldPushStackElementAfter=i.shouldPushStackElementAfter}}function ug(s){return s==="'"||s==='"'||s==="`"}class ff{static columnSelect(e,t,i,n,o,r){const a=Math.abs(o-i)+1,l=i>o,c=n>r,d=n<r,h=[];for(let u=0;u<a;u++){const g=i+(l?-u:u),f=e.columnFromVisibleColumn(t,g,n),p=e.columnFromVisibleColumn(t,g,r),_=e.visibleColumnFromColumn(t,new A(g,f)),b=e.visibleColumnFromColumn(t,new A(g,p));d&&(_>r||b<n)||c&&(b>n||_<r)||h.push(new An(new k(g,f,g,f),0,0,new A(g,p),0))}if(h.length===0)for(let u=0;u<a;u++){const g=i+(l?-u:u),f=t.getLineMaxColumn(g);h.push(new An(new k(g,f,g,f),0,0,new A(g,f),0))}return{viewStates:h,reversed:l,fromLineNumber:i,fromVisualColumn:n,toLineNumber:o,toVisualColumn:r}}static columnSelectLeft(e,t,i){let n=i.toViewVisualColumn;return n>0&&n--,ff.columnSelect(e,t,i.fromViewLineNumber,i.fromViewVisualColumn,i.toViewLineNumber,n)}static columnSelectRight(e,t,i){let n=0;const o=Math.min(i.fromViewLineNumber,i.toViewLineNumber),r=Math.max(i.fromViewLineNumber,i.toViewLineNumber);for(let l=o;l<=r;l++){const c=t.getLineMaxColumn(l),d=e.visibleColumnFromColumn(t,new A(l,c));n=Math.max(n,d)}let a=i.toViewVisualColumn;return a<n&&a++,this.columnSelect(e,t,i.fromViewLineNumber,i.fromViewVisualColumn,i.toViewLineNumber,a)}static columnSelectUp(e,t,i,n){const o=n?e.pageSize:1,r=Math.max(1,i.toViewLineNumber-o);return this.columnSelect(e,t,i.fromViewLineNumber,i.fromViewVisualColumn,r,i.toViewVisualColumn)}static columnSelectDown(e,t,i,n){const o=n?e.pageSize:1,r=Math.min(t.getLineCount(),i.toViewLineNumber+o);return this.columnSelect(e,t,i.fromViewLineNumber,i.fromViewVisualColumn,r,i.toViewVisualColumn)}}class ts{constructor(e,t,i=!1){this._range=e,this._text=t,this.insertsAutoWhitespace=i}getEditOperations(e,t){t.addTrackedEditOperation(this._range,this._text)}computeCursorState(e,t){const n=t.getInverseEditOperations()[0].range;return he.fromPositions(n.getEndPosition())}}class AP{constructor(e,t,i=!1){this._range=e,this._text=t,this.insertsAutoWhitespace=i}getEditOperations(e,t){const i=this._range.getStartPosition(),n=this._range.getEndPosition(),o=n.lineNumber,r=this._text.length+(this._range.isEmpty()?0:-1);let a=eJ(e,n,r);a.lineNumber>o&&(a=new A(o,e.getLineMaxColumn(o)));const l=k.fromPositions(i,a);t.addTrackedEditOperation(l,this._text)}computeCursorState(e,t){const n=t.getInverseEditOperations()[0].range;return he.fromPositions(n.getEndPosition())}}class bge{constructor(e,t){this._range=e,this._text=t}getEditOperations(e,t){t.addTrackedEditOperation(this._range,this._text)}computeCursorState(e,t){const n=t.getInverseEditOperations()[0].range;return he.fromRange(n,0)}}class Lk{constructor(e,t,i=!1){this._range=e,this._text=t,this.insertsAutoWhitespace=i}getEditOperations(e,t){t.addTrackedEditOperation(this._range,this._text)}computeCursorState(e,t){const n=t.getInverseEditOperations()[0].range;return he.fromPositions(n.getStartPosition())}}class IE{constructor(e,t,i,n,o=!1){this._range=e,this._text=t,this._columnDeltaOffset=n,this._lineNumberDeltaOffset=i,this.insertsAutoWhitespace=o}getEditOperations(e,t){t.addTrackedEditOperation(this._range,this._text)}computeCursorState(e,t){const n=t.getInverseEditOperations()[0].range;return he.fromPositions(n.getEndPosition().delta(this._lineNumberDeltaOffset,this._columnDeltaOffset))}}class wge{constructor(e){this._range=e}getEditOperations(e,t){const i=e.getValueInRange(this._range),n=this._range.getEndPosition(),o=n.lineNumber;let r=eJ(e,n,i.length);r.lineNumber>o&&(r=new A(o,e.getLineMaxColumn(o)));const a=k.fromPositions(n,r);t.addTrackedEditOperation(a,"")}computeCursorState(e,t){const n=t.getInverseEditOperations()[0].range;return he.fromPositions(n.getEndPosition())}}class pW{constructor(e,t,i,n=!1){this._range=e,this._text=t,this._initialSelection=i,this._forceMoveMarkers=n,this._selectionId=null}getEditOperations(e,t){t.addTrackedEditOperation(this._range,this._text,this._forceMoveMarkers),this._selectionId=t.trackSelection(this._initialSelection)}computeCursorState(e,t){return t.getTrackedSelection(this._selectionId)}}function eJ(s,e,t){if(t<0)throw new Error("Unexpected negative delta");const i=s.getLineCount();let n=new A(i,s.getLineMaxColumn(i));for(let o=e.lineNumber;o<=i;o++)if(o===e.lineNumber){const r=t-s.getLineMaxColumn(e.lineNumber)+e.column;if(r<=0){n=new A(e.lineNumber,e.column+t);break}t=r}else{const r=t-s.getLineMaxColumn(o);if(r<=0){n=new A(o,t);break}t=r}return n}class eS{static whitespaceVisibleColumn(e,t,i){const n=e.length;let o=0,r=-1,a=-1;for(let l=0;l<n;l++){if(l===t)return[r,a,o];switch(o%i===0&&(r=l,a=o),e.charCodeAt(l)){case 32:o+=1;break;case 9:o=pi.nextRenderTabStop(o,i);break;default:return[-1,-1,-1]}}return t===n?[r,a,o]:[-1,-1,-1]}static atomicPosition(e,t,i,n){const o=e.length,[r,a,l]=eS.whitespaceVisibleColumn(e,t,i);if(l===-1)return-1;let c;switch(n){case 0:c=!0;break;case 1:c=!1;break;case 2:if(l%i===0)return t;c=l%i<=i/2;break}if(c){if(r===-1)return-1;let u=a;for(let g=r;g<o;++g){if(u===a+i)return r;switch(e.charCodeAt(g)){case 32:u+=1;break;case 9:u=pi.nextRenderTabStop(u,i);break;default:return-1}}return u===a+i?r:-1}const d=pi.nextRenderTabStop(l,i);let h=l;for(let u=t;u<o;u++){if(h===d)return u;switch(e.charCodeAt(u)){case 32:h+=1;break;case 9:h=pi.nextRenderTabStop(h,i);break;default:return-1}}return h===d?o:-1}}class $2{constructor(e,t,i){this._cursorPositionBrand=void 0,this.lineNumber=e,this.column=t,this.leftoverVisibleColumns=i}}class Ct{static leftPosition(e,t){if(t.column>e.getLineMinColumn(t.lineNumber))return t.delta(void 0,-tQ(e.getLineContent(t.lineNumber),t.column-1));if(t.lineNumber>1){const i=t.lineNumber-1;return new A(i,e.getLineMaxColumn(i))}else return t}static leftPositionAtomicSoftTabs(e,t,i){if(t.column<=e.getLineIndentColumn(t.lineNumber)){const n=e.getLineMinColumn(t.lineNumber),o=e.getLineContent(t.lineNumber),r=eS.atomicPosition(o,t.column-1,i,0);if(r!==-1&&r+1>=n)return new A(t.lineNumber,r+1)}return this.leftPosition(e,t)}static left(e,t,i){const n=e.stickyTabStops?Ct.leftPositionAtomicSoftTabs(t,i,e.tabSize):Ct.leftPosition(t,i);return new $2(n.lineNumber,n.column,0)}static moveLeft(e,t,i,n,o){let r,a;if(i.hasSelection()&&!n)r=i.selection.startLineNumber,a=i.selection.startColumn;else{const l=i.position.delta(void 0,-(o-1)),c=t.normalizePosition(Ct.clipPositionColumn(l,t),0),d=Ct.left(e,t,c);r=d.lineNumber,a=d.column}return i.move(n,r,a,0)}static clipPositionColumn(e,t){return new A(e.lineNumber,Ct.clipRange(e.column,t.getLineMinColumn(e.lineNumber),t.getLineMaxColumn(e.lineNumber)))}static clipRange(e,t,i){return e<t?t:e>i?i:e}static rightPosition(e,t,i){return i<e.getLineMaxColumn(t)?i=i+Y6(e.getLineContent(t),i-1):t<e.getLineCount()&&(t=t+1,i=e.getLineMinColumn(t)),new A(t,i)}static rightPositionAtomicSoftTabs(e,t,i,n,o){if(i<e.getLineIndentColumn(t)){const r=e.getLineContent(t),a=eS.atomicPosition(r,i-1,n,1);if(a!==-1)return new A(t,a+1)}return this.rightPosition(e,t,i)}static right(e,t,i){const n=e.stickyTabStops?Ct.rightPositionAtomicSoftTabs(t,i.lineNumber,i.column,e.tabSize,e.indentSize):Ct.rightPosition(t,i.lineNumber,i.column);return new $2(n.lineNumber,n.column,0)}static moveRight(e,t,i,n,o){let r,a;if(i.hasSelection()&&!n)r=i.selection.endLineNumber,a=i.selection.endColumn;else{const l=i.position.delta(void 0,o-1),c=t.normalizePosition(Ct.clipPositionColumn(l,t),1),d=Ct.right(e,t,c);r=d.lineNumber,a=d.column}return i.move(n,r,a,0)}static vertical(e,t,i,n,o,r,a,l){const c=pi.visibleColumnFromColumn(t.getLineContent(i),n,e.tabSize)+o,d=t.getLineCount(),h=i===1&&n===1,u=i===d&&n===t.getLineMaxColumn(i),g=r<i?h:u;if(i=r,i<1?(i=1,a?n=t.getLineMinColumn(i):n=Math.min(t.getLineMaxColumn(i),n)):i>d?(i=d,a?n=t.getLineMaxColumn(i):n=Math.min(t.getLineMaxColumn(i),n)):n=e.columnFromVisibleColumn(t,i,c),g?o=0:o=c-pi.visibleColumnFromColumn(t.getLineContent(i),n,e.tabSize),l!==void 0){const f=new A(i,n),p=t.normalizePosition(f,l);o=o+(n-p.column),i=p.lineNumber,n=p.column}return new $2(i,n,o)}static down(e,t,i,n,o,r,a){return this.vertical(e,t,i,n,o,i+r,a,4)}static moveDown(e,t,i,n,o){let r,a;i.hasSelection()&&!n?(r=i.selection.endLineNumber,a=i.selection.endColumn):(r=i.position.lineNumber,a=i.position.column);let l=0,c;do if(c=Ct.down(e,t,r+l,a,i.leftoverVisibleColumns,o,!0),t.normalizePosition(new A(c.lineNumber,c.column),2).lineNumber>r)break;while(l++<10&&r+l<t.getLineCount());return i.move(n,c.lineNumber,c.column,c.leftoverVisibleColumns)}static translateDown(e,t,i){const n=i.selection,o=Ct.down(e,t,n.selectionStartLineNumber,n.selectionStartColumn,i.selectionStartLeftoverVisibleColumns,1,!1),r=Ct.down(e,t,n.positionLineNumber,n.positionColumn,i.leftoverVisibleColumns,1,!1);return new An(new k(o.lineNumber,o.column,o.lineNumber,o.column),0,o.leftoverVisibleColumns,new A(r.lineNumber,r.column),r.leftoverVisibleColumns)}static up(e,t,i,n,o,r,a){return this.vertical(e,t,i,n,o,i-r,a,3)}static moveUp(e,t,i,n,o){let r,a;i.hasSelection()&&!n?(r=i.selection.startLineNumber,a=i.selection.startColumn):(r=i.position.lineNumber,a=i.position.column);const l=Ct.up(e,t,r,a,i.leftoverVisibleColumns,o,!0);return i.move(n,l.lineNumber,l.column,l.leftoverVisibleColumns)}static translateUp(e,t,i){const n=i.selection,o=Ct.up(e,t,n.selectionStartLineNumber,n.selectionStartColumn,i.selectionStartLeftoverVisibleColumns,1,!1),r=Ct.up(e,t,n.positionLineNumber,n.positionColumn,i.leftoverVisibleColumns,1,!1);return new An(new k(o.lineNumber,o.column,o.lineNumber,o.column),0,o.leftoverVisibleColumns,new A(r.lineNumber,r.column),r.leftoverVisibleColumns)}static _isBlankLine(e,t){return e.getLineFirstNonWhitespaceColumn(t)===0}static moveToPrevBlankLine(e,t,i,n){let o=i.position.lineNumber;for(;o>1&&this._isBlankLine(t,o);)o--;for(;o>1&&!this._isBlankLine(t,o);)o--;return i.move(n,o,t.getLineMinColumn(o),0)}static moveToNextBlankLine(e,t,i,n){const o=t.getLineCount();let r=i.position.lineNumber;for(;r<o&&this._isBlankLine(t,r);)r++;for(;r<o&&!this._isBlankLine(t,r);)r++;return i.move(n,r,t.getLineMinColumn(r),0)}static moveToBeginningOfLine(e,t,i,n){const o=i.position.lineNumber,r=t.getLineMinColumn(o),a=t.getLineFirstNonWhitespaceColumn(o)||r;let l;return i.position.column===a?l=r:l=a,i.move(n,o,l,0)}static moveToEndOfLine(e,t,i,n,o){const r=i.position.lineNumber,a=t.getLineMaxColumn(r);return i.move(n,r,a,o?1073741824-a:0)}static moveToBeginningOfBuffer(e,t,i,n){return i.move(n,1,1,0)}static moveToEndOfBuffer(e,t,i,n){const o=t.getLineCount(),r=t.getLineMaxColumn(o);return i.move(n,o,r,0)}}class km{static deleteRight(e,t,i,n){const o=[];let r=e!==3;for(let a=0,l=n.length;a<l;a++){const c=n[a],d=this.getDeleteRightRange(c,i,t);if(d.isEmpty()){o[a]=null;continue}d.startLineNumber!==d.endLineNumber&&(r=!0),o[a]=new ts(d,"")}return[r,o]}static getDeleteRightRange(e,t,i){if(!e.isEmpty())return e;const n=e.getPosition(),o=Ct.right(i,t,n);if(i.trimWhitespaceOnDelete&&o.lineNumber!==n.lineNumber){const r=t.getLineFirstNonWhitespaceColumn(n.lineNumber)>0,a=t.getLineFirstNonWhitespaceColumn(o.lineNumber);if(r&&a>0)return new k(o.lineNumber,a,n.lineNumber,n.column)}return new k(o.lineNumber,o.column,n.lineNumber,n.column)}static isAutoClosingPairDelete(e,t,i,n,o,r,a){if(t==="never"&&i==="never"||e==="never")return!1;for(let l=0,c=r.length;l<c;l++){const d=r[l],h=d.getPosition();if(!d.isEmpty())return!1;const u=o.getLineContent(h.lineNumber);if(h.column<2||h.column>=u.length+1)return!1;const g=u.charAt(h.column-2),f=n.get(g);if(!f)return!1;if(ug(g)){if(i==="never")return!1}else if(t==="never")return!1;const p=u.charAt(h.column-1);let _=!1;for(const b of f)b.open===g&&b.close===p&&(_=!0);if(!_)return!1;if(e==="auto"){let b=!1;for(let w=0,C=a.length;w<C;w++){const v=a[w];if(h.lineNumber===v.startLineNumber&&h.column===v.startColumn){b=!0;break}}if(!b)return!1}}return!0}static _runAutoClosingPairDelete(e,t,i){const n=[];for(let o=0,r=i.length;o<r;o++){const a=i[o].getPosition(),l=new k(a.lineNumber,a.column-1,a.lineNumber,a.column+1);n[o]=new ts(l,"")}return[!0,n]}static deleteLeft(e,t,i,n,o){if(this.isAutoClosingPairDelete(t.autoClosingDelete,t.autoClosingBrackets,t.autoClosingQuotes,t.autoClosingPairs.autoClosingPairsOpenByEnd,i,n,o))return this._runAutoClosingPairDelete(t,i,n);const r=[];let a=e!==2;for(let l=0,c=n.length;l<c;l++){const d=km.getDeleteLeftRange(n[l],i,t);if(d.isEmpty()){r[l]=null;continue}d.startLineNumber!==d.endLineNumber&&(a=!0),r[l]=new ts(d,"")}return[a,r]}static getDeleteLeftRange(e,t,i){if(!e.isEmpty())return e;const n=e.getPosition();if(i.useTabStops&&n.column>1){const o=t.getLineContent(n.lineNumber),r=As(o),a=r===-1?o.length+1:r+1;if(n.column<=a){const l=i.visibleColumnFromColumn(t,n),c=pi.prevIndentTabStop(l,i.indentSize),d=i.columnFromVisibleColumn(t,n.lineNumber,c);return new k(n.lineNumber,d,n.lineNumber,n.column)}}return k.fromPositions(km.getPositionAfterDeleteLeft(n,t),n)}static getPositionAfterDeleteLeft(e,t){if(e.column>1){const i=qde(e.column-1,t.getLineContent(e.lineNumber));return e.with(void 0,i+1)}else if(e.lineNumber>1){const i=e.lineNumber-1;return new A(i,t.getLineMaxColumn(i))}else return e}static cut(e,t,i){const n=[];let o=null;i.sort((r,a)=>A.compare(r.getStartPosition(),a.getEndPosition()));for(let r=0,a=i.length;r<a;r++){const l=i[r];if(l.isEmpty())if(e.emptySelectionClipboard){const c=l.getPosition();let d,h,u,g;c.lineNumber<t.getLineCount()?(d=c.lineNumber,h=1,u=c.lineNumber+1,g=1):c.lineNumber>1&&o?.endLineNumber!==c.lineNumber?(d=c.lineNumber-1,h=t.getLineMaxColumn(c.lineNumber-1),u=c.lineNumber,g=t.getLineMaxColumn(c.lineNumber)):(d=c.lineNumber,h=1,u=c.lineNumber,g=t.getLineMaxColumn(c.lineNumber));const f=new k(d,h,u,g);o=f,f.isEmpty()?n[r]=null:n[r]=new ts(f,"")}else n[r]=null;else n[r]=new ts(l,"")}return new to(0,n,{shouldPushStackElementBefore:!0,shouldPushStackElementAfter:!0})}}const a_={DateTimeFormat(s,e){return new es(()=>{try{return new Intl.DateTimeFormat(s,e)}catch{return new Intl.DateTimeFormat(void 0,e)}})},Collator(s,e){return new es(()=>{try{return new Intl.Collator(s,e)}catch{return new Intl.Collator(void 0,e)}})},Segmenter(s,e){return new es(()=>{try{return new Intl.Segmenter(s,e)}catch{return new Intl.Segmenter(void 0,e)}})},Locale(s,e){return new es(()=>{try{return new Intl.Locale(s,e)}catch{return new Intl.Locale(Wf,e)}})},NumberFormat(s,e){return new es(()=>{try{return new Intl.NumberFormat(s,e)}catch{return new Intl.NumberFormat(void 0,e)}})}};function DE(s){return s<0?0:s>255?255:s|0}function W_(s){return s<0?0:s>4294967295?4294967295:s|0}class Zv{constructor(e){const t=DE(e);this._defaultValue=t,this._asciiMap=Zv._createAsciiMap(t),this._map=new Map}static _createAsciiMap(e){const t=new Uint8Array(256);return t.fill(e),t}set(e,t){const i=DE(t);e>=0&&e<256?this._asciiMap[e]=i:this._map.set(e,i)}get(e){return e>=0&&e<256?this._asciiMap[e]:this._map.get(e)||this._defaultValue}clear(){this._asciiMap.fill(this._defaultValue),this._map.clear()}}class TE{constructor(){this._actual=new Zv(0)}add(e){this._actual.set(e,1)}has(e){return this._actual.get(e)===1}clear(){return this._actual.clear()}}class vge extends Zv{constructor(e,t){super(0),this._segmenter=null,this._cachedLine=null,this._cachedSegments=[],this.intlSegmenterLocales=t,this.intlSegmenterLocales.length>0?this._segmenter=a_.Segmenter(this.intlSegmenterLocales,{granularity:"word"}):this._segmenter=null;for(let i=0,n=e.length;i<n;i++)this.set(e.charCodeAt(i),2);this.set(32,1),this.set(9,1)}findPrevIntlWordBeforeOrAtOffset(e,t){let i=null;for(const n of this._getIntlSegmenterWordsOnLine(e)){if(n.index>t)break;i=n}return i}findNextIntlWordAtOrAfterOffset(e,t){for(const i of this._getIntlSegmenterWordsOnLine(e))if(!(i.index<t))return i;return null}_getIntlSegmenterWordsOnLine(e){return this._segmenter?this._cachedLine===e?this._cachedSegments:(this._cachedLine=e,this._cachedSegments=this._filterWordSegments(this._segmenter.value.segment(e)),this._cachedSegments):[]}_filterWordSegments(e){const t=[];for(const i of e)this._isWordLike(i)&&t.push(i);return t}_isWordLike(e){return!!e.isWordLike}}const h$=new Va(10);function aa(s,e){const t=`${s}/${e.join(",")}`;let i=h$.get(t);return i||(i=new vge(s,e),h$.set(t,i)),i}class yt{static _createWord(e,t,i,n,o){return{start:n,end:o,wordType:t,nextCharClass:i}}static _createIntlWord(e,t){return{start:e.index,end:e.index+e.segment.length,wordType:1,nextCharClass:t}}static _findPreviousWordOnLine(e,t,i){const n=t.getLineContent(i.lineNumber);return this._doFindPreviousWordOnLine(n,e,i)}static _doFindPreviousWordOnLine(e,t,i){let n=0;const o=t.findPrevIntlWordBeforeOrAtOffset(e,i.column-2);for(let r=i.column-2;r>=0;r--){const a=e.charCodeAt(r),l=t.get(a);if(o&&r===o.index)return this._createIntlWord(o,l);if(l===0){if(n===2)return this._createWord(e,n,l,r+1,this._findEndOfWord(e,t,n,r+1));n=1}else if(l===2){if(n===1)return this._createWord(e,n,l,r+1,this._findEndOfWord(e,t,n,r+1));n=2}else if(l===1&&n!==0)return this._createWord(e,n,l,r+1,this._findEndOfWord(e,t,n,r+1))}return n!==0?this._createWord(e,n,1,0,this._findEndOfWord(e,t,n,0)):null}static _findEndOfWord(e,t,i,n){const o=t.findNextIntlWordAtOrAfterOffset(e,n),r=e.length;for(let a=n;a<r;a++){const l=e.charCodeAt(a),c=t.get(l);if(o&&a===o.index+o.segment.length||c===1||i===1&&c===2||i===2&&c===0)return a}return r}static _findNextWordOnLine(e,t,i){const n=t.getLineContent(i.lineNumber);return this._doFindNextWordOnLine(n,e,i)}static _doFindNextWordOnLine(e,t,i){let n=0;const o=e.length,r=t.findNextIntlWordAtOrAfterOffset(e,i.column-1);for(let a=i.column-1;a<o;a++){const l=e.charCodeAt(a),c=t.get(l);if(r&&a===r.index)return this._createIntlWord(r,c);if(c===0){if(n===2)return this._createWord(e,n,c,this._findStartOfWord(e,t,n,a-1),a);n=1}else if(c===2){if(n===1)return this._createWord(e,n,c,this._findStartOfWord(e,t,n,a-1),a);n=2}else if(c===1&&n!==0)return this._createWord(e,n,c,this._findStartOfWord(e,t,n,a-1),a)}return n!==0?this._createWord(e,n,1,this._findStartOfWord(e,t,n,o-1),o):null}static _findStartOfWord(e,t,i,n){const o=t.findPrevIntlWordBeforeOrAtOffset(e,n);for(let r=n;r>=0;r--){const a=e.charCodeAt(r),l=t.get(a);if(o&&r===o.index)return r;if(l===1||i===1&&l===2||i===2&&l===0)return r+1}return 0}static moveWordLeft(e,t,i,n,o){let r=i.lineNumber,a=i.column;a===1&&r>1&&(r=r-1,a=t.getLineMaxColumn(r));let l=yt._findPreviousWordOnLine(e,t,new A(r,a));if(n===0)return new A(r,l?l.start+1:1);if(n===1)return!o&&l&&l.wordType===2&&l.end-l.start===1&&l.nextCharClass===0&&(l=yt._findPreviousWordOnLine(e,t,new A(r,l.start+1))),new A(r,l?l.start+1:1);if(n===3){for(;l&&l.wordType===2;)l=yt._findPreviousWordOnLine(e,t,new A(r,l.start+1));return new A(r,l?l.start+1:1)}return l&&a<=l.end+1&&(l=yt._findPreviousWordOnLine(e,t,new A(r,l.start+1))),new A(r,l?l.end+1:1)}static _moveWordPartLeft(e,t){const i=t.lineNumber,n=e.getLineMaxColumn(i);if(t.column===1)return i>1?new A(i-1,e.getLineMaxColumn(i-1)):t;const o=e.getLineContent(i);for(let r=t.column-1;r>1;r--){const a=o.charCodeAt(r-2),l=o.charCodeAt(r-1);if(a===95&&l!==95)return new A(i,r);if(a===45&&l!==45)return new A(i,r);if((fu(a)||Ux(a))&&oc(l))return new A(i,r);if(oc(a)&&oc(l)&&r+1<n){const c=o.charCodeAt(r);if(fu(c)||Ux(c))return new A(i,r)}}return new A(i,1)}static moveWordRight(e,t,i,n){let o=i.lineNumber,r=i.column,a=!1;r===t.getLineMaxColumn(o)&&o<t.getLineCount()&&(a=!0,o=o+1,r=1);let l=yt._findNextWordOnLine(e,t,new A(o,r));if(n===2)l&&l.wordType===2&&l.end-l.start===1&&l.nextCharClass===0&&(l=yt._findNextWordOnLine(e,t,new A(o,l.end+1))),l?r=l.end+1:r=t.getLineMaxColumn(o);else if(n===3){for(a&&(r=0);l&&(l.wordType===2||l.start+1<=r);)l=yt._findNextWordOnLine(e,t,new A(o,l.end+1));l?r=l.start+1:r=t.getLineMaxColumn(o)}else l&&!a&&r>=l.start+1&&(l=yt._findNextWordOnLine(e,t,new A(o,l.end+1))),l?r=l.start+1:r=t.getLineMaxColumn(o);return new A(o,r)}static _moveWordPartRight(e,t){const i=t.lineNumber,n=e.getLineMaxColumn(i);if(t.column===n)return i<e.getLineCount()?new A(i+1,1):t;const o=e.getLineContent(i);for(let r=t.column+1;r<n;r++){const a=o.charCodeAt(r-2),l=o.charCodeAt(r-1);if(a!==95&&l===95)return new A(i,r);if(a!==45&&l===45)return new A(i,r);if((fu(a)||Ux(a))&&oc(l))return new A(i,r);if(oc(a)&&oc(l)&&r+1<n){const c=o.charCodeAt(r);if(fu(c)||Ux(c))return new A(i,r)}}return new A(i,n)}static _deleteWordLeftWhitespace(e,t){const i=e.getLineContent(t.lineNumber),n=t.column-2,o=Ma(i,n);return o+1<n?new k(t.lineNumber,o+2,t.lineNumber,t.column):null}static deleteWordLeft(e,t){const i=e.wordSeparators,n=e.model,o=e.selection,r=e.whitespaceHeuristics;if(!o.isEmpty())return o;if(km.isAutoClosingPairDelete(e.autoClosingDelete,e.autoClosingBrackets,e.autoClosingQuotes,e.autoClosingPairs.autoClosingPairsOpenByEnd,e.model,[e.selection],e.autoClosedCharacters)){const h=e.selection.getPosition();return new k(h.lineNumber,h.column-1,h.lineNumber,h.column+1)}const a=new A(o.positionLineNumber,o.positionColumn);let l=a.lineNumber,c=a.column;if(l===1&&c===1)return null;if(r){const h=this._deleteWordLeftWhitespace(n,a);if(h)return h}let d=yt._findPreviousWordOnLine(i,n,a);return t===0?d?c=d.start+1:c>1?c=1:(l--,c=n.getLineMaxColumn(l)):(d&&c<=d.end+1&&(d=yt._findPreviousWordOnLine(i,n,new A(l,d.start+1))),d?c=d.end+1:c>1?c=1:(l--,c=n.getLineMaxColumn(l))),new k(l,c,a.lineNumber,a.column)}static deleteInsideWord(e,t,i){if(!i.isEmpty())return i;const n=new A(i.positionLineNumber,i.positionColumn),o=this._deleteInsideWordWhitespace(t,n);return o||this._deleteInsideWordDetermineDeleteRange(e,t,n)}static _charAtIsWhitespace(e,t){const i=e.charCodeAt(t);return i===32||i===9}static _deleteInsideWordWhitespace(e,t){const i=e.getLineContent(t.lineNumber),n=i.length;if(n===0)return null;let o=Math.max(t.column-2,0);if(!this._charAtIsWhitespace(i,o))return null;let r=Math.min(t.column-1,n-1);if(!this._charAtIsWhitespace(i,r))return null;for(;o>0&&this._charAtIsWhitespace(i,o-1);)o--;for(;r+1<n&&this._charAtIsWhitespace(i,r+1);)r++;return new k(t.lineNumber,o+1,t.lineNumber,r+2)}static _deleteInsideWordDetermineDeleteRange(e,t,i){const n=t.getLineContent(i.lineNumber),o=n.length;if(o===0)return i.lineNumber>1?new k(i.lineNumber-1,t.getLineMaxColumn(i.lineNumber-1),i.lineNumber,1):i.lineNumber<t.getLineCount()?new k(i.lineNumber,1,i.lineNumber+1,1):new k(i.lineNumber,1,i.lineNumber,1);const r=h=>h.start+1<=i.column&&i.column<=h.end+1,a=(h,u)=>(h=Math.min(h,i.column),u=Math.max(u,i.column),new k(i.lineNumber,h,i.lineNumber,u)),l=h=>{let u=h.start+1,g=h.end+1,f=!1;for(;g-1<o&&this._charAtIsWhitespace(n,g-1);)f=!0,g++;if(!f)for(;u>1&&this._charAtIsWhitespace(n,u-2);)u--;return a(u,g)},c=yt._findPreviousWordOnLine(e,t,i);if(c&&r(c))return l(c);const d=yt._findNextWordOnLine(e,t,i);return d&&r(d)?l(d):c&&d?a(c.end+1,d.start+1):c?a(c.start+1,c.end+1):d?a(d.start+1,d.end+1):a(1,o+1)}static _deleteWordPartLeft(e,t){if(!t.isEmpty())return t;const i=t.getPosition(),n=yt._moveWordPartLeft(e,i);return new k(i.lineNumber,i.column,n.lineNumber,n.column)}static _findFirstNonWhitespaceChar(e,t){const i=e.length;for(let n=t;n<i;n++){const o=e.charAt(n);if(o!==" "&&o!=="	")return n}return i}static _deleteWordRightWhitespace(e,t){const i=e.getLineContent(t.lineNumber),n=t.column-1,o=this._findFirstNonWhitespaceChar(i,n);return n+1<o?new k(t.lineNumber,t.column,t.lineNumber,o+1):null}static deleteWordRight(e,t){const i=e.wordSeparators,n=e.model,o=e.selection,r=e.whitespaceHeuristics;if(!o.isEmpty())return o;const a=new A(o.positionLineNumber,o.positionColumn);let l=a.lineNumber,c=a.column;const d=n.getLineCount(),h=n.getLineMaxColumn(l);if(l===d&&c===h)return null;if(r){const g=this._deleteWordRightWhitespace(n,a);if(g)return g}let u=yt._findNextWordOnLine(i,n,a);return t===2?u?c=u.end+1:c<h||l===d?c=h:(l++,u=yt._findNextWordOnLine(i,n,new A(l,1)),u?c=u.start+1:c=n.getLineMaxColumn(l)):(u&&c>=u.start+1&&(u=yt._findNextWordOnLine(i,n,new A(l,u.end+1))),u?c=u.start+1:c<h||l===d?c=h:(l++,u=yt._findNextWordOnLine(i,n,new A(l,1)),u?c=u.start+1:c=n.getLineMaxColumn(l))),new k(l,c,a.lineNumber,a.column)}static _deleteWordPartRight(e,t){if(!t.isEmpty())return t;const i=t.getPosition(),n=yt._moveWordPartRight(e,i);return new k(i.lineNumber,i.column,n.lineNumber,n.column)}static _createWordAtPosition(e,t,i){const n=new k(t,i.start+1,t,i.end+1);return{word:e.getValueInRange(n),startColumn:n.startColumn,endColumn:n.endColumn}}static getWordAtPosition(e,t,i,n){const o=aa(t,i),r=yt._findPreviousWordOnLine(o,e,n);if(r&&r.wordType===1&&r.start<=n.column-1&&n.column-1<=r.end)return yt._createWordAtPosition(e,n.lineNumber,r);const a=yt._findNextWordOnLine(o,e,n);return a&&a.wordType===1&&a.start<=n.column-1&&n.column-1<=a.end?yt._createWordAtPosition(e,n.lineNumber,a):null}static word(e,t,i,n,o){const r=aa(e.wordSeparators,e.wordSegmenterLocales),a=yt._findPreviousWordOnLine(r,t,o),l=yt._findNextWordOnLine(r,t,o);if(!n){let g,f;return a&&a.wordType===1&&a.start<=o.column-1&&o.column-1<=a.end||a&&a.wordType===2&&a.start<=o.column-1&&o.column-1<a.end?(g=a.start+1,f=a.end+1):l&&l.wordType===1&&l.start<=o.column-1&&o.column-1<=l.end||l&&l.wordType===2&&l.start<=o.column-1&&o.column-1<l.end?(g=l.start+1,f=l.end+1):(a?g=a.end+1:g=1,l?f=l.start+1:f=t.getLineMaxColumn(o.lineNumber)),new An(new k(o.lineNumber,g,o.lineNumber,f),1,0,new A(o.lineNumber,f),0)}let c,d;a&&a.wordType===1&&a.start<o.column-1&&o.column-1<a.end?(c=a.start+1,d=a.end+1):l&&l.wordType===1&&l.start<o.column-1&&o.column-1<l.end?(c=l.start+1,d=l.end+1):(c=o.column,d=o.column);const h=o.lineNumber;let u;if(i.selectionStart.containsPosition(o))u=i.selectionStart.endColumn;else if(o.isBeforeOrEqual(i.selectionStart.getStartPosition())){u=c;const g=new A(h,u);i.selectionStart.containsPosition(g)&&(u=i.selectionStart.endColumn)}else{u=d;const g=new A(h,u);i.selectionStart.containsPosition(g)&&(u=i.selectionStart.startColumn)}return i.move(!0,h,u,0)}}class jR extends yt{static deleteWordPartLeft(e){const t=Kx([yt.deleteWordLeft(e,0),yt.deleteWordLeft(e,2),yt._deleteWordPartLeft(e.model,e.selection)]);return t.sort(k.compareRangesUsingEnds),t[2]}static deleteWordPartRight(e){const t=Kx([yt.deleteWordRight(e,0),yt.deleteWordRight(e,2),yt._deleteWordPartRight(e.model,e.selection)]);return t.sort(k.compareRangesUsingStarts),t[0]}static moveWordPartLeft(e,t,i,n){const o=Kx([yt.moveWordLeft(e,t,i,0,n),yt.moveWordLeft(e,t,i,2,n),yt._moveWordPartLeft(t,i)]);return o.sort(A.compare),o[2]}static moveWordPartRight(e,t,i){const n=Kx([yt.moveWordRight(e,t,i,0),yt.moveWordRight(e,t,i,2),yt._moveWordPartRight(t,i)]);return n.sort(A.compare),n[0]}}function Kx(s){return s.filter(e=>!!e)}function rc(s){if(!s||typeof s!="object"||s instanceof RegExp)return s;const e=Array.isArray(s)?[]:{};return Object.entries(s).forEach(([t,i])=>{e[t]=i&&typeof i=="object"?rc(i):i}),e}function Cge(s){if(!s||typeof s!="object")return s;const e=[s];for(;e.length>0;){const t=e.shift();Object.freeze(t);for(const i in t)if(tJ.call(t,i)){const n=t[i];typeof n=="object"&&!Object.isFrozen(n)&&!zce(n)&&e.push(n)}}return s}const tJ=Object.prototype.hasOwnProperty;function iJ(s,e){return MP(s,e,new Set)}function MP(s,e,t){if(Ur(s))return s;const i=e(s);if(typeof i<"u")return i;if(Array.isArray(s)){const n=[];for(const o of s)n.push(MP(o,e,t));return n}if(rn(s)){if(t.has(s))throw new Error("Cannot clone recursive data-structure");t.add(s);const n={};for(const o in s)tJ.call(s,o)&&(n[o]=MP(s[o],e,t));return t.delete(s),n}return s}function GR(s,e,t=!0){return rn(s)?(rn(e)&&Object.keys(e).forEach(i=>{i in s?t&&(rn(s[i])&&rn(e[i])?GR(s[i],e[i],t):s[i]=e[i]):s[i]=e[i]}),s):e}function Ho(s,e){if(s===e)return!0;if(s==null||e===null||e===void 0||typeof s!=typeof e||typeof s!="object"||Array.isArray(s)!==Array.isArray(e))return!1;let t,i;if(Array.isArray(s)){if(s.length!==e.length)return!1;for(t=0;t<s.length;t++)if(!Ho(s[t],e[t]))return!1}else{const n=[];for(i in s)n.push(i);n.sort();const o=[];for(i in e)o.push(i);if(o.sort(),!Ho(n,o))return!1;for(t=0;t<n.length;t++)if(!Ho(s[n[t]],e[n[t]]))return!1}return!0}var Sr;(function(s){s[s.Left=1]="Left",s[s.Center=2]="Center",s[s.Right=4]="Right",s[s.Full=7]="Full"})(Sr||(Sr={}));var Ll;(function(s){s[s.Left=1]="Left",s[s.Center=2]="Center",s[s.Right=3]="Right"})(Ll||(Ll={}));var br;(function(s){s[s.LTR=0]="LTR",s[s.RTL=1]="RTL"})(br||(br={}));var jr;(function(s){s[s.Both=0]="Both",s[s.Right=1]="Right",s[s.Left=2]="Left",s[s.None=3]="None"})(jr||(jr={}));class kk{get originalIndentSize(){return this._indentSizeIsTabSize?"tabSize":this.indentSize}constructor(e){this._textModelResolvedOptionsBrand=void 0,this.tabSize=Math.max(1,e.tabSize|0),e.indentSize==="tabSize"?(this.indentSize=this.tabSize,this._indentSizeIsTabSize=!0):(this.indentSize=Math.max(1,e.indentSize|0),this._indentSizeIsTabSize=!1),this.insertSpaces=!!e.insertSpaces,this.defaultEOL=e.defaultEOL|0,this.trimAutoWhitespace=!!e.trimAutoWhitespace,this.bracketPairColorizationOptions=e.bracketPairColorizationOptions}equals(e){return this.tabSize===e.tabSize&&this._indentSizeIsTabSize===e._indentSizeIsTabSize&&this.indentSize===e.indentSize&&this.insertSpaces===e.insertSpaces&&this.defaultEOL===e.defaultEOL&&this.trimAutoWhitespace===e.trimAutoWhitespace&&Ho(this.bracketPairColorizationOptions,e.bracketPairColorizationOptions)}createChangeEvent(e){return{tabSize:this.tabSize!==e.tabSize,indentSize:this.indentSize!==e.indentSize,insertSpaces:this.insertSpaces!==e.insertSpaces,trimAutoWhitespace:this.trimAutoWhitespace!==e.trimAutoWhitespace}}}class tS{constructor(e,t){this._findMatchBrand=void 0,this.range=e,this.matches=t}}function yge(s){return!!s&&typeof s.read=="function"}class U2{constructor(e,t,i,n,o,r){this.identifier=e,this.range=t,this.text=i,this.forceMoveMarkers=n,this.isAutoWhitespaceEdit=o,this._isTracked=r}}class Sge{constructor(e,t,i){this.regex=e,this.wordSeparators=t,this.simpleSearch=i}}class xge{constructor(e,t,i){this.reverseEdits=e,this.changes=t,this.trimAutoWhitespaceLineNumbers=i}}function nJ(s){return!s.isTooLargeForSyncing()&&!s.isForSimpleWidget}class Rn{static addCursorDown(e,t,i){const n=[];let o=0;for(let r=0,a=t.length;r<a;r++){const l=t[r];n[o++]=new ut(l.modelState,l.viewState),i?n[o++]=ut.fromModelState(Ct.translateDown(e.cursorConfig,e.model,l.modelState)):n[o++]=ut.fromViewState(Ct.translateDown(e.cursorConfig,e,l.viewState))}return n}static addCursorUp(e,t,i){const n=[];let o=0;for(let r=0,a=t.length;r<a;r++){const l=t[r];n[o++]=new ut(l.modelState,l.viewState),i?n[o++]=ut.fromModelState(Ct.translateUp(e.cursorConfig,e.model,l.modelState)):n[o++]=ut.fromViewState(Ct.translateUp(e.cursorConfig,e,l.viewState))}return n}static moveToBeginningOfLine(e,t,i){const n=[];for(let o=0,r=t.length;o<r;o++){const a=t[o];n[o]=this._moveToLineStart(e,a,i)}return n}static _moveToLineStart(e,t,i){const n=t.viewState.position.column,o=t.modelState.position.column,r=n===o,a=t.viewState.position.lineNumber,l=e.getLineFirstNonWhitespaceColumn(a);return!r&&!(n===l)?this._moveToLineStartByView(e,t,i):this._moveToLineStartByModel(e,t,i)}static _moveToLineStartByView(e,t,i){return ut.fromViewState(Ct.moveToBeginningOfLine(e.cursorConfig,e,t.viewState,i))}static _moveToLineStartByModel(e,t,i){return ut.fromModelState(Ct.moveToBeginningOfLine(e.cursorConfig,e.model,t.modelState,i))}static moveToEndOfLine(e,t,i,n){const o=[];for(let r=0,a=t.length;r<a;r++){const l=t[r];o[r]=this._moveToLineEnd(e,l,i,n)}return o}static _moveToLineEnd(e,t,i,n){const o=t.viewState.position,r=e.getLineMaxColumn(o.lineNumber),a=o.column===r,l=t.modelState.position,c=e.model.getLineMaxColumn(l.lineNumber),d=r-o.column===c-l.column;return a||d?this._moveToLineEndByModel(e,t,i,n):this._moveToLineEndByView(e,t,i,n)}static _moveToLineEndByView(e,t,i,n){return ut.fromViewState(Ct.moveToEndOfLine(e.cursorConfig,e,t.viewState,i,n))}static _moveToLineEndByModel(e,t,i,n){return ut.fromModelState(Ct.moveToEndOfLine(e.cursorConfig,e.model,t.modelState,i,n))}static expandLineSelection(e,t){const i=[];for(let n=0,o=t.length;n<o;n++){const r=t[n],a=r.modelState.selection.startLineNumber,l=e.model.getLineCount();let c=r.modelState.selection.endLineNumber,d;c===l?d=e.model.getLineMaxColumn(l):(c++,d=1),i[n]=ut.fromModelState(new An(new k(a,1,a,1),0,0,new A(c,d),0))}return i}static moveToBeginningOfBuffer(e,t,i){const n=[];for(let o=0,r=t.length;o<r;o++){const a=t[o];n[o]=ut.fromModelState(Ct.moveToBeginningOfBuffer(e.cursorConfig,e.model,a.modelState,i))}return n}static moveToEndOfBuffer(e,t,i){const n=[];for(let o=0,r=t.length;o<r;o++){const a=t[o];n[o]=ut.fromModelState(Ct.moveToEndOfBuffer(e.cursorConfig,e.model,a.modelState,i))}return n}static selectAll(e,t){const i=e.model.getLineCount(),n=e.model.getLineMaxColumn(i);return ut.fromModelState(new An(new k(1,1,1,1),0,0,new A(i,n),0))}static line(e,t,i,n,o){const r=e.model.validatePosition(n),a=o?e.coordinatesConverter.validateViewPosition(new A(o.lineNumber,o.column),r):e.coordinatesConverter.convertModelPositionToViewPosition(r);if(!i){const c=e.model.getLineCount();let d=r.lineNumber+1,h=1;return d>c&&(d=c,h=e.model.getLineMaxColumn(d)),ut.fromModelState(new An(new k(r.lineNumber,1,d,h),2,0,new A(d,h),0))}const l=t.modelState.selectionStart.getStartPosition().lineNumber;if(r.lineNumber<l)return ut.fromViewState(t.viewState.move(!0,a.lineNumber,1,0));if(r.lineNumber>l){const c=e.getLineCount();let d=a.lineNumber+1,h=1;return d>c&&(d=c,h=e.getLineMaxColumn(d)),ut.fromViewState(t.viewState.move(!0,d,h,0))}else{const c=t.modelState.selectionStart.getEndPosition();return ut.fromModelState(t.modelState.move(!0,c.lineNumber,c.column,0))}}static word(e,t,i,n){const o=e.model.validatePosition(n);return ut.fromModelState(yt.word(e.cursorConfig,e.model,t.modelState,i,o))}static cancelSelection(e,t){if(!t.modelState.hasSelection())return new ut(t.modelState,t.viewState);const i=t.viewState.position.lineNumber,n=t.viewState.position.column;return ut.fromViewState(new An(new k(i,n,i,n),0,0,new A(i,n),0))}static moveTo(e,t,i,n,o){if(i){if(t.modelState.selectionStartKind===1)return this.word(e,t,i,n);if(t.modelState.selectionStartKind===2)return this.line(e,t,i,n,o)}const r=e.model.validatePosition(n),a=o?e.coordinatesConverter.validateViewPosition(new A(o.lineNumber,o.column),r):e.coordinatesConverter.convertModelPositionToViewPosition(r);return ut.fromViewState(t.viewState.move(i,a.lineNumber,a.column,0))}static simpleMove(e,t,i,n,o,r){switch(i){case 0:return r===4?this._moveHalfLineLeft(e,t,n):this._moveLeft(e,t,n,o);case 1:return r===4?this._moveHalfLineRight(e,t,n):this._moveRight(e,t,n,o);case 2:return r===2?this._moveUpByViewLines(e,t,n,o):this._moveUpByModelLines(e,t,n,o);case 3:return r===2?this._moveDownByViewLines(e,t,n,o):this._moveDownByModelLines(e,t,n,o);case 4:return r===2?t.map(a=>ut.fromViewState(Ct.moveToPrevBlankLine(e.cursorConfig,e,a.viewState,n))):t.map(a=>ut.fromModelState(Ct.moveToPrevBlankLine(e.cursorConfig,e.model,a.modelState,n)));case 5:return r===2?t.map(a=>ut.fromViewState(Ct.moveToNextBlankLine(e.cursorConfig,e,a.viewState,n))):t.map(a=>ut.fromModelState(Ct.moveToNextBlankLine(e.cursorConfig,e.model,a.modelState,n)));case 6:return this._moveToViewMinColumn(e,t,n);case 7:return this._moveToViewFirstNonWhitespaceColumn(e,t,n);case 8:return this._moveToViewCenterColumn(e,t,n);case 9:return this._moveToViewMaxColumn(e,t,n);case 10:return this._moveToViewLastNonWhitespaceColumn(e,t,n);default:return null}}static viewportMove(e,t,i,n,o){const r=e.getCompletelyVisibleViewRange(),a=e.coordinatesConverter.convertViewRangeToModelRange(r);switch(i){case 11:{const l=this._firstLineNumberInRange(e.model,a,o),c=e.model.getLineFirstNonWhitespaceColumn(l);return[this._moveToModelPosition(e,t[0],n,l,c)]}case 13:{const l=this._lastLineNumberInRange(e.model,a,o),c=e.model.getLineFirstNonWhitespaceColumn(l);return[this._moveToModelPosition(e,t[0],n,l,c)]}case 12:{const l=Math.round((a.startLineNumber+a.endLineNumber)/2),c=e.model.getLineFirstNonWhitespaceColumn(l);return[this._moveToModelPosition(e,t[0],n,l,c)]}case 14:{const l=[];for(let c=0,d=t.length;c<d;c++){const h=t[c];l[c]=this.findPositionInViewportIfOutside(e,h,r,n)}return l}default:return null}}static findPositionInViewportIfOutside(e,t,i,n){const o=t.viewState.position.lineNumber;if(i.startLineNumber<=o&&o<=i.endLineNumber-1)return new ut(t.modelState,t.viewState);{let r;o>i.endLineNumber-1?r=i.endLineNumber-1:o<i.startLineNumber?r=i.startLineNumber:r=o;const a=Ct.vertical(e.cursorConfig,e,o,t.viewState.position.column,t.viewState.leftoverVisibleColumns,r,!1);return ut.fromViewState(t.viewState.move(n,a.lineNumber,a.column,a.leftoverVisibleColumns))}}static _firstLineNumberInRange(e,t,i){let n=t.startLineNumber;return t.startColumn!==e.getLineMinColumn(n)&&n++,Math.min(t.endLineNumber,n+i-1)}static _lastLineNumberInRange(e,t,i){let n=t.startLineNumber;return t.startColumn!==e.getLineMinColumn(n)&&n++,Math.max(n,t.endLineNumber-i+1)}static _moveLeft(e,t,i,n){return t.map(o=>{const a=e.getTextDirection(o.viewState.position.lineNumber)===br.RTL;return ut.fromViewState(a?Ct.moveRight(e.cursorConfig,e,o.viewState,i,n):Ct.moveLeft(e.cursorConfig,e,o.viewState,i,n))})}static _moveHalfLineLeft(e,t,i){const n=[];for(let o=0,r=t.length;o<r;o++){const a=t[o],l=a.viewState.position.lineNumber,c=Math.round(e.getLineLength(l)/2);n[o]=ut.fromViewState(Ct.moveLeft(e.cursorConfig,e,a.viewState,i,c))}return n}static _moveRight(e,t,i,n){return t.map(o=>{const a=e.getTextDirection(o.viewState.position.lineNumber)===br.RTL;return ut.fromViewState(a?Ct.moveLeft(e.cursorConfig,e,o.viewState,i,n):Ct.moveRight(e.cursorConfig,e,o.viewState,i,n))})}static _moveHalfLineRight(e,t,i){const n=[];for(let o=0,r=t.length;o<r;o++){const a=t[o],l=a.viewState.position.lineNumber,c=Math.round(e.getLineLength(l)/2);n[o]=ut.fromViewState(Ct.moveRight(e.cursorConfig,e,a.viewState,i,c))}return n}static _moveDownByViewLines(e,t,i,n){const o=[];for(let r=0,a=t.length;r<a;r++){const l=t[r];o[r]=ut.fromViewState(Ct.moveDown(e.cursorConfig,e,l.viewState,i,n))}return o}static _moveDownByModelLines(e,t,i,n){const o=[];for(let r=0,a=t.length;r<a;r++){const l=t[r];o[r]=ut.fromModelState(Ct.moveDown(e.cursorConfig,e.model,l.modelState,i,n))}return o}static _moveUpByViewLines(e,t,i,n){const o=[];for(let r=0,a=t.length;r<a;r++){const l=t[r];o[r]=ut.fromViewState(Ct.moveUp(e.cursorConfig,e,l.viewState,i,n))}return o}static _moveUpByModelLines(e,t,i,n){const o=[];for(let r=0,a=t.length;r<a;r++){const l=t[r];o[r]=ut.fromModelState(Ct.moveUp(e.cursorConfig,e.model,l.modelState,i,n))}return o}static _moveToViewPosition(e,t,i,n,o){return ut.fromViewState(t.viewState.move(i,n,o,0))}static _moveToModelPosition(e,t,i,n,o){return ut.fromModelState(t.modelState.move(i,n,o,0))}static _moveToViewMinColumn(e,t,i){const n=[];for(let o=0,r=t.length;o<r;o++){const a=t[o],l=a.viewState.position.lineNumber,c=e.getLineMinColumn(l);n[o]=this._moveToViewPosition(e,a,i,l,c)}return n}static _moveToViewFirstNonWhitespaceColumn(e,t,i){const n=[];for(let o=0,r=t.length;o<r;o++){const a=t[o],l=a.viewState.position.lineNumber,c=e.getLineFirstNonWhitespaceColumn(l);n[o]=this._moveToViewPosition(e,a,i,l,c)}return n}static _moveToViewCenterColumn(e,t,i){const n=[];for(let o=0,r=t.length;o<r;o++){const a=t[o],l=a.viewState.position.lineNumber,c=Math.round((e.getLineMaxColumn(l)+e.getLineMinColumn(l))/2);n[o]=this._moveToViewPosition(e,a,i,l,c)}return n}static _moveToViewMaxColumn(e,t,i){const n=[];for(let o=0,r=t.length;o<r;o++){const a=t[o],l=a.viewState.position.lineNumber,c=e.getLineMaxColumn(l);n[o]=this._moveToViewPosition(e,a,i,l,c)}return n}static _moveToViewLastNonWhitespaceColumn(e,t,i){const n=[];for(let o=0,r=t.length;o<r;o++){const a=t[o],l=a.viewState.position.lineNumber,c=e.getLineLastNonWhitespaceColumn(l);n[o]=this._moveToViewPosition(e,a,i,l,c)}return n}}var NE;(function(s){const e=function(i){if(!rn(i))return!1;const n=i;return!(!bn(n.to)||!fs(n.select)&&!iP(n.select)||!fs(n.by)&&!bn(n.by)||!fs(n.value)&&!ch(n.value)||!fs(n.noHistory)&&!iP(n.noHistory))};s.metadata={description:"Move cursor to a logical position in the view",args:[{name:"Cursor move argument object",description:`Property-value pairs that can be passed through this argument:
					* 'to': A mandatory logical position value providing where to move the cursor.
						\`\`\`
						'left', 'right', 'up', 'down', 'prevBlankLine', 'nextBlankLine',
						'wrappedLineStart', 'wrappedLineEnd', 'wrappedLineColumnCenter'
						'wrappedLineFirstNonWhitespaceCharacter', 'wrappedLineLastNonWhitespaceCharacter'
						'viewPortTop', 'viewPortCenter', 'viewPortBottom', 'viewPortIfOutside'
						\`\`\`
					* 'by': Unit to move. Default is computed based on 'to' value.
						\`\`\`
						'line', 'wrappedLine', 'character', 'halfLine'
						\`\`\`
					* 'value': Number of units to move. Default is '1'.
					* 'select': If 'true' makes the selection. Default is 'false'.
					* 'noHistory': If 'true' does not add the movement to navigation history. Default is 'false'.
				`,constraint:e,schema:{type:"object",required:["to"],properties:{to:{type:"string",enum:["left","right","up","down","prevBlankLine","nextBlankLine","wrappedLineStart","wrappedLineEnd","wrappedLineColumnCenter","wrappedLineFirstNonWhitespaceCharacter","wrappedLineLastNonWhitespaceCharacter","viewPortTop","viewPortCenter","viewPortBottom","viewPortIfOutside"]},by:{type:"string",enum:["line","wrappedLine","character","halfLine"]},value:{type:"number",default:1},select:{type:"boolean",default:!1},noHistory:{type:"boolean",default:!1}}}}]},s.RawDirection={Left:"left",Right:"right",Up:"up",Down:"down",PrevBlankLine:"prevBlankLine",NextBlankLine:"nextBlankLine",WrappedLineStart:"wrappedLineStart",WrappedLineFirstNonWhitespaceCharacter:"wrappedLineFirstNonWhitespaceCharacter",WrappedLineColumnCenter:"wrappedLineColumnCenter",WrappedLineEnd:"wrappedLineEnd",WrappedLineLastNonWhitespaceCharacter:"wrappedLineLastNonWhitespaceCharacter",ViewPortTop:"viewPortTop",ViewPortCenter:"viewPortCenter",ViewPortBottom:"viewPortBottom",ViewPortIfOutside:"viewPortIfOutside"},s.RawUnit={Line:"line",WrappedLine:"wrappedLine",Character:"character",HalfLine:"halfLine"};function t(i){if(!i.to)return null;let n;switch(i.to){case s.RawDirection.Left:n=0;break;case s.RawDirection.Right:n=1;break;case s.RawDirection.Up:n=2;break;case s.RawDirection.Down:n=3;break;case s.RawDirection.PrevBlankLine:n=4;break;case s.RawDirection.NextBlankLine:n=5;break;case s.RawDirection.WrappedLineStart:n=6;break;case s.RawDirection.WrappedLineFirstNonWhitespaceCharacter:n=7;break;case s.RawDirection.WrappedLineColumnCenter:n=8;break;case s.RawDirection.WrappedLineEnd:n=9;break;case s.RawDirection.WrappedLineLastNonWhitespaceCharacter:n=10;break;case s.RawDirection.ViewPortTop:n=11;break;case s.RawDirection.ViewPortBottom:n=13;break;case s.RawDirection.ViewPortCenter:n=12;break;case s.RawDirection.ViewPortIfOutside:n=14;break;default:return null}let o=0;switch(i.by){case s.RawUnit.Line:o=1;break;case s.RawUnit.WrappedLine:o=2;break;case s.RawUnit.Character:o=3;break;case s.RawUnit.HalfLine:o=4;break}return{direction:n,unit:o,select:!!i.select,value:i.value||1,noHistory:!!i.noHistory}}s.parse=t})(NE||(NE={}));var Xi;(function(s){s[s.None=0]="None",s[s.Indent=1]="Indent",s[s.IndentOutdent=2]="IndentOutdent",s[s.Outdent=3]="Outdent"})(Xi||(Xi={}));class q2{constructor(e){if(this._neutralCharacter=null,this._neutralCharacterSearched=!1,this.open=e.open,this.close=e.close,this._inString=!0,this._inComment=!0,this._inRegEx=!0,Array.isArray(e.notIn))for(let t=0,i=e.notIn.length;t<i;t++)switch(e.notIn[t]){case"string":this._inString=!1;break;case"comment":this._inComment=!1;break;case"regex":this._inRegEx=!1;break}}isOK(e){switch(e){case 0:return!0;case 1:return this._inComment;case 2:return this._inString;case 3:return this._inRegEx}}shouldAutoClose(e,t){if(e.getTokenCount()===0)return!0;const i=e.findTokenIndexAtOffset(t-2),n=e.getStandardTokenType(i);return this.isOK(n)}_findNeutralCharacterInRange(e,t){for(let i=e;i<=t;i++){const n=String.fromCharCode(i);if(!this.open.includes(n)&&!this.close.includes(n))return n}return null}findNeutralCharacter(){return this._neutralCharacterSearched||(this._neutralCharacterSearched=!0,this._neutralCharacter||(this._neutralCharacter=this._findNeutralCharacterInRange(48,57)),this._neutralCharacter||(this._neutralCharacter=this._findNeutralCharacterInRange(97,122)),this._neutralCharacter||(this._neutralCharacter=this._findNeutralCharacterInRange(65,90))),this._neutralCharacter}}class Lge{constructor(e){this.autoClosingPairsOpenByStart=new Map,this.autoClosingPairsOpenByEnd=new Map,this.autoClosingPairsCloseByStart=new Map,this.autoClosingPairsCloseByEnd=new Map,this.autoClosingPairsCloseSingleChar=new Map;for(const t of e)LC(this.autoClosingPairsOpenByStart,t.open.charAt(0),t),LC(this.autoClosingPairsOpenByEnd,t.open.charAt(t.open.length-1),t),LC(this.autoClosingPairsCloseByStart,t.close.charAt(0),t),LC(this.autoClosingPairsCloseByEnd,t.close.charAt(t.close.length-1),t),t.close.length===1&&t.open.length===1&&LC(this.autoClosingPairsCloseSingleChar,t.close,t)}}function LC(s,e,t){s.has(e)?s.get(e).push(t):s.set(e,[t])}const RE="`~!@#$%^&*()-=+[{]}\\|;:'\",.<>/?";function kge(s=""){let e="(-?\\d*\\.\\d\\w*)|([^";for(const t of RE)s.indexOf(t)>=0||(e+="\\"+t);return e+="\\s]+)",new RegExp(e,"g")}const mW=kge();function _W(s){let e=mW;if(s&&s instanceof RegExp)if(s.global)e=s;else{let t="g";s.ignoreCase&&(t+="i"),s.multiline&&(t+="m"),s.unicode&&(t+="u"),e=new RegExp(s.source,t)}return e.lastIndex=0,e}const sJ=new Ns;sJ.unshift({maxLen:1e3,windowSize:15,timeBudget:150});function iS(s,e,t,i,n){if(e=_W(e),n||(n=ot.first(sJ)),t.length>n.maxLen){let c=s-n.maxLen/2;return c<0?c=0:i+=c,t=t.substring(c,s+n.maxLen/2),iS(s,e,t,i,n)}const o=Date.now(),r=s-1-i;let a=-1,l=null;for(let c=1;!(Date.now()-o>=n.timeBudget);c++){const d=r-n.windowSize*c;e.lastIndex=Math.max(0,d);const h=Ege(e,t,r,a);if(!h&&l||(l=h,d<=0))break;a=d}if(l){const c={word:l[0],startColumn:i+1+l.index,endColumn:i+1+l.index+l[0].length};return e.lastIndex=0,c}return null}function Ege(s,e,t,i){let n;for(;n=s.exec(e);){const o=n.index||0;if(o<=t&&s.lastIndex>=t)return n;if(i>0&&o>i)return null}return null}const Kb=class Kb{constructor(e){if(e.autoClosingPairs?this._autoClosingPairs=e.autoClosingPairs.map(t=>new q2(t)):e.brackets?this._autoClosingPairs=e.brackets.map(t=>new q2({open:t[0],close:t[1]})):this._autoClosingPairs=[],e.__electricCharacterSupport&&e.__electricCharacterSupport.docComment){const t=e.__electricCharacterSupport.docComment;this._autoClosingPairs.push(new q2({open:t.open,close:t.close||""}))}this._autoCloseBeforeForQuotes=typeof e.autoCloseBefore=="string"?e.autoCloseBefore:Kb.DEFAULT_AUTOCLOSE_BEFORE_LANGUAGE_DEFINED_QUOTES,this._autoCloseBeforeForBrackets=typeof e.autoCloseBefore=="string"?e.autoCloseBefore:Kb.DEFAULT_AUTOCLOSE_BEFORE_LANGUAGE_DEFINED_BRACKETS,this._surroundingPairs=e.surroundingPairs||this._autoClosingPairs}getAutoClosingPairs(){return this._autoClosingPairs}getAutoCloseBeforeSet(e){return e?this._autoCloseBeforeForQuotes:this._autoCloseBeforeForBrackets}getSurroundingPairs(){return this._surroundingPairs}};Kb.DEFAULT_AUTOCLOSE_BEFORE_LANGUAGE_DEFINED_QUOTES=`;:.,=}])> 
	`,Kb.DEFAULT_AUTOCLOSE_BEFORE_LANGUAGE_DEFINED_BRACKETS=`'"\`;:.,=}])> 
	`;let OP=Kb,j2;function oJ(){return j2||(j2=new TextDecoder("UTF-16LE")),j2}let G2;function Ige(){return G2||(G2=new TextDecoder("UTF-16BE")),G2}let K2;function rJ(){return K2||(K2=VX()?oJ():Ige()),K2}function Dge(s,e,t){const i=new Uint16Array(s.buffer,e,t);return t>0&&(i[0]===65279||i[0]===65534)?Tge(s,e,t):oJ().decode(i)}function Tge(s,e,t){const i=[];let n=0;for(let o=0;o<t;o++){const r=_he(s,e);e+=2,i[n++]=String.fromCharCode(r)}return i.join("")}class Eg{constructor(e){this._capacity=e|0,this._buffer=new Uint16Array(this._capacity),this._completedStrings=null,this._bufferLength=0}reset(){this._completedStrings=null,this._bufferLength=0}build(){return this._completedStrings!==null?(this._flushBuffer(),this._completedStrings.join("")):this._buildBuffer()}_buildBuffer(){if(this._bufferLength===0)return"";const e=new Uint16Array(this._buffer.buffer,0,this._bufferLength);return rJ().decode(e)}_flushBuffer(){const e=this._buildBuffer();this._bufferLength=0,this._completedStrings===null?this._completedStrings=[e]:this._completedStrings[this._completedStrings.length]=e}appendCharCode(e){const t=this._capacity-this._bufferLength;t<=1&&(t===0||sn(e))&&this._flushBuffer(),this._buffer[this._bufferLength++]=e}appendASCIICharCode(e){this._bufferLength===this._capacity&&this._flushBuffer(),this._buffer[this._bufferLength++]=e}appendString(e){const t=e.length;if(this._bufferLength+t>=this._capacity){this._flushBuffer(),this._completedStrings[this._completedStrings.length]=e;return}for(let i=0;i<t;i++)this._buffer[this._bufferLength++]=e.charCodeAt(i)}}class AE{constructor(e,t,i,n,o,r){this._richEditBracketBrand=void 0,this.languageId=e,this.index=t,this.open=i,this.close=n,this.forwardRegex=o,this.reversedRegex=r,this._openSet=AE._toSet(this.open),this._closeSet=AE._toSet(this.close)}isOpen(e){return this._openSet.has(e)}isClose(e){return this._closeSet.has(e)}static _toSet(e){const t=new Set;for(const i of e)t.add(i);return t}}function Nge(s){const e=s.length;s=s.map(r=>[r[0].toLowerCase(),r[1].toLowerCase()]);const t=[];for(let r=0;r<e;r++)t[r]=r;const i=(r,a)=>{const[l,c]=r,[d,h]=a;return l===d||l===h||c===d||c===h},n=(r,a)=>{const l=Math.min(r,a),c=Math.max(r,a);for(let d=0;d<e;d++)t[d]===c&&(t[d]=l)};for(let r=0;r<e;r++){const a=s[r];for(let l=r+1;l<e;l++){const c=s[l];i(a,c)&&n(t[r],t[l])}}const o=[];for(let r=0;r<e;r++){const a=[],l=[];for(let c=0;c<e;c++)if(t[c]===r){const[d,h]=s[c];a.push(d),l.push(h)}a.length>0&&o.push({open:a,close:l})}return o}class Rge{constructor(e,t){this._richEditBracketsBrand=void 0;const i=Nge(t);this.brackets=i.map((n,o)=>new AE(e,o,n.open,n.close,Age(n.open,n.close,i,o),Mge(n.open,n.close,i,o))),this.forwardRegex=Oge(this.brackets),this.reversedRegex=Pge(this.brackets),this.textIsBracket={},this.textIsOpenBracket={},this.maxBracketLength=0;for(const n of this.brackets){for(const o of n.open)this.textIsBracket[o]=n,this.textIsOpenBracket[o]=!0,this.maxBracketLength=Math.max(this.maxBracketLength,o.length);for(const o of n.close)this.textIsBracket[o]=n,this.textIsOpenBracket[o]=!1,this.maxBracketLength=Math.max(this.maxBracketLength,o.length)}}}function aJ(s,e,t,i){for(let n=0,o=e.length;n<o;n++){if(n===t)continue;const r=e[n];for(const a of r.open)a.indexOf(s)>=0&&i.push(a);for(const a of r.close)a.indexOf(s)>=0&&i.push(a)}}function lJ(s,e){return s.length-e.length}function KR(s){if(s.length<=1)return s;const e=[],t=new Set;for(const i of s)t.has(i)||(e.push(i),t.add(i));return e}function Age(s,e,t,i){let n=[];n=n.concat(s),n=n.concat(e);for(let o=0,r=n.length;o<r;o++)aJ(n[o],t,i,n);return n=KR(n),n.sort(lJ),n.reverse(),Z0(n)}function Mge(s,e,t,i){let n=[];n=n.concat(s),n=n.concat(e);for(let o=0,r=n.length;o<r;o++)aJ(n[o],t,i,n);return n=KR(n),n.sort(lJ),n.reverse(),Z0(n.map(bW))}function Oge(s){let e=[];for(const t of s){for(const i of t.open)e.push(i);for(const i of t.close)e.push(i)}return e=KR(e),Z0(e)}function Pge(s){let e=[];for(const t of s){for(const i of t.open)e.push(i);for(const i of t.close)e.push(i)}return e=KR(e),Z0(e.map(bW))}function Fge(s){const e=/^[\w ]+$/.test(s);return s=Uo(s),e?`\\b${s}\\b`:s}function Z0(s,e){const t=`(${s.map(Fge).join(")|(")})`;return JX(t,!0,e)}const bW=(function(){function s(i){const n=new Uint16Array(i.length);let o=0;for(let r=i.length-1;r>=0;r--)n[o++]=i.charCodeAt(r);return rJ().decode(n)}let e=null,t=null;return function(n){return e!==n&&(e=n,t=s(e)),t}})();class _a{static _findPrevBracketInText(e,t,i,n){const o=i.match(e);if(!o)return null;const r=i.length-(o.index||0),a=o[0].length,l=n+r;return new k(t,l-a+1,t,l+1)}static findPrevBracketInRange(e,t,i,n,o){const a=bW(i).substring(i.length-o,i.length-n);return this._findPrevBracketInText(e,t,a,n)}static findNextBracketInText(e,t,i,n){const o=i.match(e);if(!o)return null;const r=o.index||0,a=o[0].length;if(a===0)return null;const l=n+r;return new k(t,l+1,t,l+1+a)}static findNextBracketInRange(e,t,i,n,o){const r=i.substring(n,o);return this.findNextBracketInText(e,t,r,n)}}class Bge{constructor(e){this._richEditBrackets=e}getElectricCharacters(){const e=[];if(this._richEditBrackets)for(const t of this._richEditBrackets.brackets)for(const i of t.close){const n=i.charAt(i.length-1);e.push(n)}return fh(e)}onElectricCharacter(e,t,i){if(!this._richEditBrackets||this._richEditBrackets.brackets.length===0)return null;const n=t.findTokenIndexAtOffset(i-1);if(Ul(t.getStandardTokenType(n)))return null;const o=this._richEditBrackets.reversedRegex,r=t.getLineContent().substring(0,i-1)+e,a=_a.findPrevBracketInRange(o,1,r,0,r.length);if(!a)return null;const l=r.substring(a.startColumn-1,a.endColumn-1).toLowerCase();if(this._richEditBrackets.textIsOpenBracket[l])return null;const d=t.getActualLineContentBefore(a.startColumn-1);return/^\s*$/.test(d)?{matchOpenBracket:l}:null}}function Yx(s){return s.global&&(s.lastIndex=0),!0}class Wge{constructor(e){this._indentationRules=e}shouldIncrease(e){return!!(this._indentationRules&&this._indentationRules.increaseIndentPattern&&Yx(this._indentationRules.increaseIndentPattern)&&this._indentationRules.increaseIndentPattern.test(e))}shouldDecrease(e){return!!(this._indentationRules&&this._indentationRules.decreaseIndentPattern&&Yx(this._indentationRules.decreaseIndentPattern)&&this._indentationRules.decreaseIndentPattern.test(e))}shouldIndentNextLine(e){return!!(this._indentationRules&&this._indentationRules.indentNextLinePattern&&Yx(this._indentationRules.indentNextLinePattern)&&this._indentationRules.indentNextLinePattern.test(e))}shouldIgnore(e){return!!(this._indentationRules&&this._indentationRules.unIndentedLinePattern&&Yx(this._indentationRules.unIndentedLinePattern)&&this._indentationRules.unIndentedLinePattern.test(e))}getIndentMetadata(e){let t=0;return this.shouldIncrease(e)&&(t+=1),this.shouldDecrease(e)&&(t+=2),this.shouldIndentNextLine(e)&&(t+=4),this.shouldIgnore(e)&&(t+=8),t}}class yb{constructor(e){e=e||{},e.brackets=e.brackets||[["(",")"],["{","}"],["[","]"]],this._brackets=[],e.brackets.forEach(t=>{const i=yb._createOpenBracketRegExp(t[0]),n=yb._createCloseBracketRegExp(t[1]);i&&n&&this._brackets.push({open:t[0],openRegExp:i,close:t[1],closeRegExp:n})}),this._regExpRules=e.onEnterRules||[]}onEnter(e,t,i,n){if(e>=3)for(let o=0,r=this._regExpRules.length;o<r;o++){const a=this._regExpRules[o];if([{reg:a.beforeText,text:i},{reg:a.afterText,text:n},{reg:a.previousLineText,text:t}].every(c=>c.reg?(c.reg.lastIndex=0,c.reg.test(c.text)):!0))return a.action}if(e>=2&&i.length>0&&n.length>0)for(let o=0,r=this._brackets.length;o<r;o++){const a=this._brackets[o];if(a.openRegExp.test(i)&&a.closeRegExp.test(n))return{indentAction:Xi.IndentOutdent}}if(e>=2&&i.length>0){for(let o=0,r=this._brackets.length;o<r;o++)if(this._brackets[o].openRegExp.test(i))return{indentAction:Xi.Indent}}return null}static _createOpenBracketRegExp(e){let t=Uo(e);return/\B/.test(t.charAt(0))||(t="\\b"+t),t+="\\s*$",yb._safeRegExp(t)}static _createCloseBracketRegExp(e){let t=Uo(e);return/\B/.test(t.charAt(t.length-1))||(t=t+"\\b"),t="^\\s*"+t,yb._safeRegExp(t)}static _safeRegExp(e){try{return new RegExp(e)}catch(t){return Oe(t),null}}}const He=qe("configurationService");function PP(s,e){const t=Object.create(null);for(const i in s)cJ(t,i,s[i],e);return t}function cJ(s,e,t,i){const n=e.split("."),o=n.pop();let r=s;for(let a=0;a<n.length;a++){const l=n[a];let c=r[l];switch(typeof c){case"undefined":c=r[l]=Object.create(null);break;case"object":if(c===null){i(`Ignoring ${e} as ${n.slice(0,a+1).join(".")} is null`);return}break;default:i(`Ignoring ${e} as ${n.slice(0,a+1).join(".")} is ${JSON.stringify(c)}`);return}r=c}if(typeof r=="object"&&r!==null)try{r[o]=t}catch{i(`Ignoring ${e} as ${n.join(".")} is ${JSON.stringify(r)}`)}else i(`Ignoring ${e} as ${n.join(".")} is ${JSON.stringify(r)}`)}function Hge(s,e){const t=e.split(".");dJ(s,t)}function dJ(s,e){if(!s)return;const t=e.shift();if(e.length===0){delete s[t];return}if(Object.keys(s).indexOf(t)!==-1){const i=s[t];typeof i=="object"&&!Array.isArray(i)&&(dJ(i,e),Object.keys(i).length===0&&delete s[t])}}function u$(s,e,t){function i(r,a){let l=r;for(const c of a){if(typeof l!="object"||l===null)return;l=l[c]}return l}const n=e.split("."),o=i(s,n);return typeof o>"u"?t:o}function Vge(s){return s.replace(/^\[/,"").replace(/]$/g,"").replace(/\]\[/g,", ")}const Ei=qe("languageService");class ac{constructor(e,t=[],i=!1){this.ctor=e,this.staticArguments=t,this.supportsDelayedInstantiation=i}}const hJ=[];function tt(s,e,t){e instanceof ac||(e=new ac(e,[],!!t)),hJ.push([s,e])}function g$(){return hJ}const Ti=Object.freeze({text:"text/plain",binary:"application/octet-stream",unknown:"application/unknown",markdown:"text/markdown",latex:"text/latex",uriList:"text/uri-list",html:"text/html"}),YR={JSONContribution:"base.contributions.json"};function zge(s){return s.length>0&&s.charAt(s.length-1)==="#"?s.substring(0,s.length-1):s}class $ge extends F{constructor(){super(...arguments),this.schemasById={},this._onDidChangeSchema=this._register(new M)}registerSchema(e,t,i){const n=zge(e);this.schemasById[n]=t,this._onDidChangeSchema.fire(e),i&&i.add(pe(()=>{delete this.schemasById[n],this._onDidChangeSchema.fire(e)}))}notifySchemaChanged(e){this._onDidChangeSchema.fire(e)}}const Uge=new $ge;bi.add(YR.JSONContribution,Uge);let nr;const Y2=globalThis.vscode;if(typeof Y2<"u"&&typeof Y2.context<"u"){const s=Y2.context.configuration();if(s)nr=s.product;else throw new Error("Sandbox: unable to resolve product configuration from preload script.")}else if(globalThis._VSCODE_PRODUCT_JSON&&globalThis._VSCODE_PACKAGE_JSON){if(nr=globalThis._VSCODE_PRODUCT_JSON,_E.VSCODE_DEV&&Object.assign(nr,{nameShort:`${nr.nameShort} Dev`,nameLong:`${nr.nameLong} Dev`,dataFolderName:`${nr.dataFolderName}-dev`,serverDataFolderName:nr.serverDataFolderName?`${nr.serverDataFolderName}-dev`:void 0}),!nr.version){const s=globalThis._VSCODE_PACKAGE_JSON;Object.assign(nr,{version:s.version})}}else nr={},Object.keys(nr).length===0&&Object.assign(nr,{version:"1.104.0-dev",nameShort:"Code - OSS Dev",nameLong:"Code - OSS Dev",applicationName:"code-oss",dataFolderName:".vscode-oss",urlProtocol:"code-oss",reportIssueUrl:"https://github.com/microsoft/vscode/issues/new",licenseName:"MIT",licenseUrl:"https://github.com/microsoft/vscode/blob/main/LICENSE.txt",serverLicenseUrl:"https://github.com/microsoft/vscode/blob/main/LICENSE.txt"});var Ek=nr;const Ml={Configuration:"base.contributions.configuration"},Zx="vscode://schemas/settings/resourceLanguage",f$=bi.as(YR.JSONContribution);class qge extends F{constructor(){super(),this.registeredConfigurationDefaults=[],this.overrideIdentifiers=new Set,this._onDidSchemaChange=this._register(new M),this._onDidUpdateConfiguration=this._register(new M),this.configurationDefaultsOverrides=new Map,this.defaultLanguageConfigurationOverridesNode={id:"defaultOverrides",title:m(1664,"Default Language Configuration Overrides"),properties:{}},this.configurationContributors=[this.defaultLanguageConfigurationOverridesNode],this.resourceLanguageSettingsSchema={properties:{},patternProperties:{},additionalProperties:!0,allowTrailingCommas:!0,allowComments:!0},this.configurationProperties={},this.policyConfigurations=new Map,this.excludedConfigurationProperties={},f$.registerSchema(Zx,this.resourceLanguageSettingsSchema),this.registerOverridePropertyPatternKey()}registerConfiguration(e,t=!0){return this.registerConfigurations([e],t),e}registerConfigurations(e,t=!0){const i=new Set;this.doRegisterConfigurations(e,t,i),f$.registerSchema(Zx,this.resourceLanguageSettingsSchema),this._onDidSchemaChange.fire(),this._onDidUpdateConfiguration.fire({properties:i})}registerDefaultConfigurations(e){const t=new Set;this.doRegisterDefaultConfigurations(e,t),this._onDidSchemaChange.fire(),this._onDidUpdateConfiguration.fire({properties:t,defaultsOverrides:!0})}doRegisterDefaultConfigurations(e,t){this.registeredConfigurationDefaults.push(...e);const i=[];for(const{overrides:n,source:o}of e)for(const r in n){t.add(r);const a=this.configurationDefaultsOverrides.get(r)??this.configurationDefaultsOverrides.set(r,{configurationDefaultOverrides:[]}).get(r),l=n[r];if(a.configurationDefaultOverrides.push({value:l,source:o}),gg.test(r)){const c=this.mergeDefaultConfigurationsForOverrideIdentifier(r,l,o,a.configurationDefaultOverrideValue);if(!c)continue;a.configurationDefaultOverrideValue=c,this.updateDefaultOverrideProperty(r,c,o),i.push(...ME(r))}else{const c=this.mergeDefaultConfigurationsForConfigurationProperty(r,l,o,a.configurationDefaultOverrideValue);if(!c)continue;a.configurationDefaultOverrideValue=c;const d=this.configurationProperties[r];d&&(this.updatePropertyDefaultValue(r,d),this.updateSchema(r,d))}}this.doRegisterOverrideIdentifiers(i)}updateDefaultOverrideProperty(e,t,i){const n={section:{id:this.defaultLanguageConfigurationOverridesNode.id,title:this.defaultLanguageConfigurationOverridesNode.title,order:this.defaultLanguageConfigurationOverridesNode.order,extensionInfo:this.defaultLanguageConfigurationOverridesNode.extensionInfo},type:"object",default:t.value,description:m(1665,"Configure settings to be overridden for {0}.",Vge(e)),$ref:Zx,defaultDefaultValue:t.value,source:i,defaultValueSource:i};this.configurationProperties[e]=n,this.defaultLanguageConfigurationOverridesNode.properties[e]=n}mergeDefaultConfigurationsForOverrideIdentifier(e,t,i,n){const o=n?.value||{},r=n?.source??new Map;if(!(r instanceof Map)){console.error("objectConfigurationSources is not a Map");return}for(const a of Object.keys(t)){const l=t[a];if(rn(l)&&(fs(o[a])||rn(o[a]))){if(o[a]={...o[a]??{},...l},i)for(const d in l)r.set(`${a}.${d}`,i)}else o[a]=l,i?r.set(a,i):r.delete(a)}return{value:o,source:r}}mergeDefaultConfigurationsForConfigurationProperty(e,t,i,n){const o=this.configurationProperties[e],r=n?.value??o?.defaultDefaultValue;let a=i;if(rn(t)&&(o!==void 0&&o.type==="object"||o===void 0&&(fs(r)||rn(r)))){if(a=n?.source??new Map,!(a instanceof Map)){console.error("defaultValueSource is not a Map");return}for(const c in t)i&&a.set(`${e}.${c}`,i);t={...rn(r)?r:{},...t}}return{value:t,source:a}}registerOverrideIdentifiers(e){this.doRegisterOverrideIdentifiers(e),this._onDidSchemaChange.fire()}doRegisterOverrideIdentifiers(e){for(const t of e)this.overrideIdentifiers.add(t);this.updateOverridePropertyPatternKey()}doRegisterConfigurations(e,t,i){e.forEach(n=>{this.validateAndRegisterProperties(n,t,n.extensionInfo,n.restrictedProperties,void 0,i),this.configurationContributors.push(n),this.registerJSONConfiguration(n)})}validateAndRegisterProperties(e,t=!0,i,n,o=4,r){o=Ur(e.scope)?o:e.scope;const a=e.properties;if(a)for(const c in a){const d=a[c];if(d.section={id:e.id,title:e.title,order:e.order,extensionInfo:e.extensionInfo},t&&Kge(c,d,i?.id)){delete a[c];continue}d.source=i,d.defaultDefaultValue=a[c].default,this.updatePropertyDefaultValue(c,d),gg.test(c)?d.scope=void 0:(d.scope=Ur(d.scope)?o:d.scope,d.restricted=Ur(d.restricted)?!!n?.includes(c):d.restricted),d.experiment?d.tags?.some(g=>g.toLowerCase()==="onexp")||(d.tags=d.tags??[],d.tags.push("onExP")):d.tags?.some(g=>g.toLowerCase()==="onexp")&&(console.error(`Invalid tag 'onExP' found for property '${c}'. Please use 'experiment' property instead.`),d.experiment={mode:"startup"});const h=a[c].hasOwnProperty("included")&&!a[c].included,u=a[c].policy?.name;h?(this.excludedConfigurationProperties[c]=a[c],u&&(this.policyConfigurations.set(u,c),r.add(c)),delete a[c]):(r.add(c),u&&this.policyConfigurations.set(u,c),this.configurationProperties[c]=a[c],!a[c].deprecationMessage&&a[c].markdownDeprecationMessage&&(a[c].deprecationMessage=a[c].markdownDeprecationMessage))}const l=e.allOf;if(l)for(const c of l)this.validateAndRegisterProperties(c,t,i,n,o,r)}getConfigurationProperties(){return this.configurationProperties}getPolicyConfigurations(){return this.policyConfigurations}getExcludedConfigurationProperties(){return this.excludedConfigurationProperties}registerJSONConfiguration(e){const t=i=>{const n=i.properties;if(n)for(const r in n)this.updateSchema(r,n[r]);i.allOf?.forEach(t)};t(e)}updateSchema(e,t){switch(t.scope){case 1:break;case 2:break;case 3:break;case 7:break;case 4:break;case 5:break;case 6:this.resourceLanguageSettingsSchema.properties[e]=t;break}}updateOverridePropertyPatternKey(){for(const e of this.overrideIdentifiers.values()){const t=`[${e}]`,i={type:"object",description:m(1666,"Configure editor settings to be overridden for a language."),errorMessage:m(1667,"This setting does not support per-language configuration."),$ref:Zx};this.updatePropertyDefaultValue(t,i)}}registerOverridePropertyPatternKey(){m(1668,"Configure editor settings to be overridden for a language."),m(1669,"This setting does not support per-language configuration."),this._onDidSchemaChange.fire()}updatePropertyDefaultValue(e,t){const i=this.configurationDefaultsOverrides.get(e)?.configurationDefaultOverrideValue;let n,o;i&&(!t.disallowConfigurationDefault||!i.source)&&(n=i.value,o=i.source),fs(n)&&(n=t.defaultDefaultValue,o=void 0),fs(n)&&(n=Gge(t.type)),t.default=n,t.defaultValueSource=o}}const uJ="\\[([^\\]]+)\\]",p$=new RegExp(uJ,"g"),jge=`^(${uJ})+$`,gg=new RegExp(jge);function ME(s){const e=[];if(gg.test(s)){let t=p$.exec(s);for(;t?.length;){const i=t[1].trim();i&&e.push(i),t=p$.exec(s)}}return fh(e)}function Gge(s){switch(Array.isArray(s)?s[0]:s){case"boolean":return!1;case"integer":case"number":return 0;case"string":return"";case"array":return[];case"object":return{};default:return null}}const Ik=new qge;bi.add(Ml.Configuration,Ik);function Kge(s,e,t){return s.trim()?gg.test(s)?m(1671,"Cannot register '{0}'. This matches property pattern '\\\\[.*\\\\]$' for describing language specific editor settings. Use 'configurationDefaults' contribution.",s):Ik.getConfigurationProperties()[s]!==void 0&&(!t||!Yge.has(t.toLowerCase()))?m(1672,"Cannot register '{0}'. This property is already registered.",s):e.policy?.name&&Ik.getPolicyConfigurations().get(e.policy?.name)!==void 0?m(1673,"Cannot register '{0}'. The associated policy {1} is already registered with {2}.",s,e.policy?.name,Ik.getPolicyConfigurations().get(e.policy?.name)):null:m(1670,"Cannot register an empty property")}const Yge=new Set(Ek.defaultChatAgent?[Ek.defaultChatAgent.extensionId,Ek.defaultChatAgent.chatExtensionId].map(s=>s.toLowerCase()):[]),Zge={ModesRegistry:"editor.modesRegistry"};class Xge extends F{constructor(){super(),this._onDidChangeLanguages=this._register(new M),this.onDidChangeLanguages=this._onDidChangeLanguages.event,this._languages=[]}registerLanguage(e){return this._languages.push(e),this._onDidChangeLanguages.fire(void 0),{dispose:()=>{for(let t=0,i=this._languages.length;t<i;t++)if(this._languages[t]===e){this._languages.splice(t,1);return}}}}getLanguages(){return this._languages}}const rv=new Xge;bi.add(Zge.ModesRegistry,rv);const xr="plaintext",Qge=".txt";rv.registerLanguage({id:xr,extensions:[Qge],aliases:[m(784,"Plain Text"),"text"],mimetypes:[Ti.text]});bi.as(Ml.Configuration).registerDefaultConfigurations([{overrides:{"[plaintext]":{"editor.unicodeHighlight.ambiguousCharacters":!1,"editor.unicodeHighlight.invisibleCharacters":!1},"[go]":{"editor.insertSpaces":!1},"[makefile]":{"editor.insertSpaces":!1},"[shellscript]":{"files.eol":`
`},"[yaml]":{"editor.insertSpaces":!0,"editor.tabSize":2}}}]);class Jge{constructor(e,t){this.languageId=e;const i=t.brackets?m$(t.brackets):[],n=new dP(a=>{const l=new Set;return{info:new efe(this,a,l),closing:l}}),o=new dP(a=>{const l=new Set,c=new Set;return{info:new tfe(this,a,l,c),opening:l,openingColorized:c}});for(const[a,l]of i){const c=n.get(a),d=o.get(l);c.closing.add(d.info),d.opening.add(c.info)}const r=t.colorizedBracketPairs?m$(t.colorizedBracketPairs):i.filter(a=>!(a[0]==="<"&&a[1]===">"));for(const[a,l]of r){const c=n.get(a),d=o.get(l);c.closing.add(d.info),d.openingColorized.add(c.info),d.opening.add(c.info)}this._openingBrackets=new Map([...n.cachedValues].map(([a,l])=>[a,l.info])),this._closingBrackets=new Map([...o.cachedValues].map(([a,l])=>[a,l.info]))}get openingBrackets(){return[...this._openingBrackets.values()]}get closingBrackets(){return[...this._closingBrackets.values()]}getOpeningBracketInfo(e){return this._openingBrackets.get(e)}getClosingBracketInfo(e){return this._closingBrackets.get(e)}getBracketInfo(e){return this.getOpeningBracketInfo(e)||this.getClosingBracketInfo(e)}getBracketRegExp(e){const t=Array.from([...this._openingBrackets.keys(),...this._closingBrackets.keys()]);return Z0(t,e)}}function m$(s){return s.filter(([e,t])=>e!==""&&t!=="")}class gJ{constructor(e,t){this.config=e,this.bracketText=t}get languageId(){return this.config.languageId}}class efe extends gJ{constructor(e,t,i){super(e,t),this.openedBrackets=i,this.isOpeningBracket=!0}}class tfe extends gJ{constructor(e,t,i,n){super(e,t),this.openingBrackets=i,this.openingColorizedBrackets=n,this.isOpeningBracket=!1}closes(e){return e.config!==this.config?!1:this.openingBrackets.has(e)}closesColorized(e){return e.config!==this.config?!1:this.openingColorizedBrackets.has(e)}getOpeningBrackets(){return[...this.openingBrackets]}}var ife=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},_$=function(s,e){return function(t,i){e(t,i,s)}};class Z2{constructor(e){this.languageId=e}affects(e){return this.languageId?this.languageId===e:!0}}const ui=qe("languageConfigurationService");let FP=class extends F{constructor(e,t){super(),this.configurationService=e,this.languageService=t,this._registry=this._register(new rfe),this.onDidChangeEmitter=this._register(new M),this.onDidChange=this.onDidChangeEmitter.event,this.configurations=new Map;const i=new Set(Object.values(BP));this._register(this.configurationService.onDidChangeConfiguration(n=>{const o=n.change.keys.some(a=>i.has(a)),r=n.change.overrides.filter(([a,l])=>l.some(c=>i.has(c))).map(([a])=>a);if(o)this.configurations.clear(),this.onDidChangeEmitter.fire(new Z2(void 0));else for(const a of r)this.languageService.isRegisteredLanguageId(a)&&(this.configurations.delete(a),this.onDidChangeEmitter.fire(new Z2(a)))})),this._register(this._registry.onDidChange(n=>{this.configurations.delete(n.languageId),this.onDidChangeEmitter.fire(new Z2(n.languageId))}))}register(e,t,i){return this._registry.register(e,t,i)}getLanguageConfiguration(e){let t=this.configurations.get(e);return t||(t=nfe(e,this._registry,this.configurationService,this.languageService),this.configurations.set(e,t)),t}};FP=ife([_$(0,He),_$(1,Ei)],FP);function nfe(s,e,t,i){let n=e.getLanguageConfiguration(s);if(!n){if(!i.isRegisteredLanguageId(s))return new cw(s,{});n=new cw(s,{})}const o=sfe(n.languageId,t),r=pJ([n.underlyingConfig,o]);return new cw(n.languageId,r)}const BP={brackets:"editor.language.brackets",colorizedBracketPairs:"editor.language.colorizedBracketPairs"};function sfe(s,e){const t=e.getValue(BP.brackets,{overrideIdentifier:s}),i=e.getValue(BP.colorizedBracketPairs,{overrideIdentifier:s});return{brackets:b$(t),colorizedBracketPairs:b$(i)}}function b$(s){if(Array.isArray(s))return s.map(e=>{if(!(!Array.isArray(e)||e.length!==2))return[e[0],e[1]]}).filter(e=>!!e)}function fJ(s,e,t){const i=s.getLineContent(e);let n=Bt(i);return n.length>t-1&&(n=n.substring(0,t-1)),n}class ofe{constructor(e){this.languageId=e,this._resolved=null,this._entries=[],this._order=0,this._resolved=null}register(e,t){const i=new w$(e,t,++this._order);return this._entries.push(i),this._resolved=null,pe(()=>{for(let n=0;n<this._entries.length;n++)if(this._entries[n]===i){this._entries.splice(n,1),this._resolved=null;break}})}getResolvedConfiguration(){if(!this._resolved){const e=this._resolve();e&&(this._resolved=new cw(this.languageId,e))}return this._resolved}_resolve(){return this._entries.length===0?null:(this._entries.sort(w$.cmp),pJ(this._entries.map(e=>e.configuration)))}}function pJ(s){let e={comments:void 0,brackets:void 0,wordPattern:void 0,indentationRules:void 0,onEnterRules:void 0,autoClosingPairs:void 0,surroundingPairs:void 0,autoCloseBefore:void 0,folding:void 0,colorizedBracketPairs:void 0,__electricCharacterSupport:void 0};for(const t of s)e={comments:t.comments||e.comments,brackets:t.brackets||e.brackets,wordPattern:t.wordPattern||e.wordPattern,indentationRules:t.indentationRules||e.indentationRules,onEnterRules:t.onEnterRules||e.onEnterRules,autoClosingPairs:t.autoClosingPairs||e.autoClosingPairs,surroundingPairs:t.surroundingPairs||e.surroundingPairs,autoCloseBefore:t.autoCloseBefore||e.autoCloseBefore,folding:t.folding||e.folding,colorizedBracketPairs:t.colorizedBracketPairs||e.colorizedBracketPairs,__electricCharacterSupport:t.__electricCharacterSupport||e.__electricCharacterSupport};return e}class w${constructor(e,t,i){this.configuration=e,this.priority=t,this.order=i}static cmp(e,t){return e.priority===t.priority?e.order-t.order:e.priority-t.priority}}class v${constructor(e){this.languageId=e}}class rfe extends F{constructor(){super(),this._entries=new Map,this._onDidChange=this._register(new M),this.onDidChange=this._onDidChange.event,this._register(this.register(xr,{brackets:[["(",")"],["[","]"],["{","}"]],surroundingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:"<",close:">"},{open:'"',close:'"'},{open:"'",close:"'"},{open:"`",close:"`"}],colorizedBracketPairs:[],folding:{offSide:!0}},0))}register(e,t,i=0){let n=this._entries.get(e);n||(n=new ofe(e),this._entries.set(e,n));const o=n.register(t,i);return this._onDidChange.fire(new v$(e)),pe(()=>{o.dispose(),this._onDidChange.fire(new v$(e))})}getLanguageConfiguration(e){return this._entries.get(e)?.getResolvedConfiguration()||null}}class cw{constructor(e,t){this.languageId=e,this.underlyingConfig=t,this._brackets=null,this._electricCharacter=null,this._onEnterSupport=this.underlyingConfig.brackets||this.underlyingConfig.indentationRules||this.underlyingConfig.onEnterRules?new yb(this.underlyingConfig):null,this.comments=cw._handleComments(this.underlyingConfig),this.characterPair=new OP(this.underlyingConfig),this.wordDefinition=this.underlyingConfig.wordPattern||mW,this.indentationRules=this.underlyingConfig.indentationRules,this.underlyingConfig.indentationRules?this.indentRulesSupport=new Wge(this.underlyingConfig.indentationRules):this.indentRulesSupport=null,this.foldingRules=this.underlyingConfig.folding||{},this.bracketsNew=new Jge(e,this.underlyingConfig)}getWordDefinition(){return _W(this.wordDefinition)}get brackets(){return!this._brackets&&this.underlyingConfig.brackets&&(this._brackets=new Rge(this.languageId,this.underlyingConfig.brackets)),this._brackets}get electricCharacter(){return this._electricCharacter||(this._electricCharacter=new Bge(this.brackets)),this._electricCharacter}onEnter(e,t,i,n){return this._onEnterSupport?this._onEnterSupport.onEnter(e,t,i,n):null}getAutoClosingPairs(){return new Lge(this.characterPair.getAutoClosingPairs())}getAutoCloseBeforeSet(e){return this.characterPair.getAutoCloseBeforeSet(e)}getSurroundingPairs(){return this.characterPair.getSurroundingPairs()}static _handleComments(e){const t=e.comments;if(!t)return null;const i={};if(t.lineComment&&(typeof t.lineComment=="string"?i.lineCommentToken=t.lineComment:(i.lineCommentToken=t.lineComment.comment,i.lineCommentNoIndent=t.lineComment.noIndent)),t.blockComment){const[n,o]=t.blockComment;i.blockCommentStartToken=n,i.blockCommentEndToken=o}return i}}tt(ui,FP,1);class ds{static getLanguageId(e){return(e&255)>>>0}static getTokenType(e){return(e&768)>>>8}static containsBalancedBrackets(e){return(e&1024)!==0}static getFontStyle(e){return(e&30720)>>>11}static getForeground(e){return(e&16744448)>>>15}static getBackground(e){return(e&4278190080)>>>24}static getClassNameFromMetadata(e){let i="mtk"+this.getForeground(e);const n=this.getFontStyle(e);return n&1&&(i+=" mtki"),n&2&&(i+=" mtkb"),n&4&&(i+=" mtku"),n&8&&(i+=" mtks"),i}static getInlineStyleFromMetadata(e,t){const i=this.getForeground(e),n=this.getFontStyle(e);let o=`color: ${t[i]};`;n&1&&(o+="font-style: italic;"),n&2&&(o+="font-weight: bold;");let r="";return n&4&&(r+=" underline"),n&8&&(r+=" line-through"),r&&(o+=`text-decoration:${r};`),o}static getPresentationFromMetadata(e){const t=this.getForeground(e),i=this.getFontStyle(e);return{foreground:t,italic:!!(i&1),bold:!!(i&2),underline:!!(i&4),strikethrough:!!(i&8)}}}class me{static fromTo(e,t){return new me(e,t)}static addRange(e,t){let i=0;for(;i<t.length&&t[i].endExclusive<e.start;)i++;let n=i;for(;n<t.length&&t[n].start<=e.endExclusive;)n++;if(i===n)t.splice(i,0,e);else{const o=Math.min(e.start,t[i].start),r=Math.max(e.endExclusive,t[n-1].endExclusive);t.splice(i,n-i,new me(o,r))}}static tryCreate(e,t){if(!(e>t))return new me(e,t)}static ofLength(e){return new me(0,e)}static ofStartAndLength(e,t){return new me(e,e+t)}static emptyAt(e){return new me(e,e)}constructor(e,t){if(this.start=e,this.endExclusive=t,e>t)throw new Ce(`Invalid range: ${this.toString()}`)}get isEmpty(){return this.start===this.endExclusive}delta(e){return new me(this.start+e,this.endExclusive+e)}deltaStart(e){return new me(this.start+e,this.endExclusive)}deltaEnd(e){return new me(this.start,this.endExclusive+e)}get length(){return this.endExclusive-this.start}toString(){return`[${this.start}, ${this.endExclusive})`}equals(e){return this.start===e.start&&this.endExclusive===e.endExclusive}contains(e){return this.start<=e&&e<this.endExclusive}join(e){return new me(Math.min(this.start,e.start),Math.max(this.endExclusive,e.endExclusive))}intersect(e){const t=Math.max(this.start,e.start),i=Math.min(this.endExclusive,e.endExclusive);if(t<=i)return new me(t,i)}intersectionLength(e){const t=Math.max(this.start,e.start),i=Math.min(this.endExclusive,e.endExclusive);return Math.max(0,i-t)}intersects(e){const t=Math.max(this.start,e.start),i=Math.min(this.endExclusive,e.endExclusive);return t<i}intersectsOrTouches(e){const t=Math.max(this.start,e.start),i=Math.min(this.endExclusive,e.endExclusive);return t<=i}isBefore(e){return this.endExclusive<=e.start}isAfter(e){return this.start>=e.endExclusive}slice(e){return e.slice(this.start,this.endExclusive)}substring(e){return e.substring(this.start,this.endExclusive)}clip(e){if(this.isEmpty)throw new Ce(`Invalid clipping range: ${this.toString()}`);return Math.max(this.start,Math.min(this.endExclusive-1,e))}clipCyclic(e){if(this.isEmpty)throw new Ce(`Invalid clipping range: ${this.toString()}`);return e<this.start?this.endExclusive-(this.start-e)%this.length:e>=this.endExclusive?this.start+(e-this.start)%this.length:e}forEach(e){for(let t=this.start;t<this.endExclusive;t++)e(t)}joinRightTouching(e){if(this.endExclusive!==e.start)throw new Ce(`Invalid join: ${this.toString()} and ${e.toString()}`);return new me(this.start,e.endExclusive)}}class wW{constructor(){this._sortedRanges=[]}addRange(e){let t=0;for(;t<this._sortedRanges.length&&this._sortedRanges[t].endExclusive<e.start;)t++;let i=t;for(;i<this._sortedRanges.length&&this._sortedRanges[i].start<=e.endExclusive;)i++;if(t===i)this._sortedRanges.splice(t,0,e);else{const n=Math.min(e.start,this._sortedRanges[t].start),o=Math.max(e.endExclusive,this._sortedRanges[i-1].endExclusive);this._sortedRanges.splice(t,i-t,new me(n,o))}}toString(){return this._sortedRanges.map(e=>e.toString()).join(", ")}intersectsStrict(e){let t=0;for(;t<this._sortedRanges.length&&this._sortedRanges[t].endExclusive<=e.start;)t++;return t<this._sortedRanges.length&&this._sortedRanges[t].start<e.endExclusive}intersectWithRange(e){const t=new wW;for(const i of this._sortedRanges){const n=i.intersect(e);n&&t.addRange(n)}return t}intersectWithRangeLength(e){return this.intersectWithRange(e).length}get length(){return this._sortedRanges.reduce((e,t)=>e+t.length,0)}}const Sd=class Sd{static createEmpty(e,t){const i=Sd.defaultTokenMetadata,n=new Uint32Array(2);return n[0]=e.length,n[1]=i,new Sd(n,e,t)}static createFromTextAndMetadata(e,t){let i=0,n="";const o=new Array;for(const{text:r,metadata:a}of e)o.push(i+r.length,a),i+=r.length,n+=r;return new Sd(new Uint32Array(o),n,t)}static convertToEndOffset(e,t){const n=(e.length>>>1)-1;for(let o=0;o<n;o++)e[o<<1]=e[o+1<<1];e[n<<1]=t}static findIndexInTokensArray(e,t){if(e.length<=2)return 0;let i=0,n=(e.length>>>1)-1;for(;i<n;){const o=i+Math.floor((n-i)/2),r=e[o<<1];if(r===t)return o+1;r<t?i=o+1:r>t&&(n=o)}return i}constructor(e,t,i){this._lineTokensBrand=void 0,(e.length>1?e[e.length-2]:0)!==t.length&&Oe(new Error("Token length and text length do not match!")),this._tokens=e,this._tokensCount=this._tokens.length>>>1,this._text=t,this.languageIdCodec=i}getTextLength(){return this._text.length}equals(e){return e instanceof Sd?this.slicedEquals(e,0,this._tokensCount):!1}slicedEquals(e,t,i){if(this._text!==e._text||this._tokensCount!==e._tokensCount)return!1;const n=t<<1,o=n+(i<<1);for(let r=n;r<o;r++)if(this._tokens[r]!==e._tokens[r])return!1;return!0}getLineContent(){return this._text}getCount(){return this._tokensCount}getStartOffset(e){return e>0?this._tokens[e-1<<1]:0}getMetadata(e){return this._tokens[(e<<1)+1]}getLanguageId(e){const t=this._tokens[(e<<1)+1],i=ds.getLanguageId(t);return this.languageIdCodec.decodeLanguageId(i)}getStandardTokenType(e){const t=this._tokens[(e<<1)+1];return ds.getTokenType(t)}getForeground(e){const t=this._tokens[(e<<1)+1];return ds.getForeground(t)}getClassName(e){const t=this._tokens[(e<<1)+1];return ds.getClassNameFromMetadata(t)}getInlineStyle(e,t){const i=this._tokens[(e<<1)+1];return ds.getInlineStyleFromMetadata(i,t)}getPresentation(e){const t=this._tokens[(e<<1)+1];return ds.getPresentationFromMetadata(t)}getEndOffset(e){return this._tokens[e<<1]}findTokenIndexAtOffset(e){return Sd.findIndexInTokensArray(this._tokens,e)}inflate(){return this}sliceAndInflate(e,t,i){return new vW(this,e,t,i)}sliceZeroCopy(e){return this.sliceAndInflate(e.start,e.endExclusive,0)}withInserted(e){if(e.length===0)return this;let t=0,i=0,n="";const o=new Array;let r=0;for(;;){const a=t<this._tokensCount?this._tokens[t<<1]:-1,l=i<e.length?e[i]:null;if(a!==-1&&(l===null||a<=l.offset)){n+=this._text.substring(r,a);const c=this._tokens[(t<<1)+1];o.push(n.length,c),t++,r=a}else if(l){if(l.offset>r){n+=this._text.substring(r,l.offset);const c=this._tokens[(t<<1)+1];o.push(n.length,c),r=l.offset}n+=l.text,o.push(n.length,l.tokenMetadata),i++}else break}return new Sd(new Uint32Array(o),n,this.languageIdCodec)}getTokensInRange(e){const t=new lfe,i=this.findTokenIndexAtOffset(e.start),n=this.findTokenIndexAtOffset(e.endExclusive);for(let o=i;o<=n;o++){const a=new me(this.getStartOffset(o),this.getEndOffset(o)).intersectionLength(e);a>0&&t.add(a,this.getMetadata(o))}return t.build()}getTokenText(e){const t=this.getStartOffset(e),i=this.getEndOffset(e);return this._text.substring(t,i)}forEach(e){const t=this.getCount();for(let i=0;i<t;i++)e(i)}toString(){let e="";return this.forEach(t=>{e+=`[${this.getTokenText(t)}]{${this.getClassName(t)}}`}),e}};Sd.defaultTokenMetadata=(32768|2<<24)>>>0;let Ri=Sd;class vW{constructor(e,t,i,n){this._source=e,this._startOffset=t,this._endOffset=i,this._deltaOffset=n,this._firstTokenIndex=e.findTokenIndexAtOffset(t),this.languageIdCodec=e.languageIdCodec,this._tokensCount=0;for(let o=this._firstTokenIndex,r=e.getCount();o<r&&!(e.getStartOffset(o)>=i);o++)this._tokensCount++}getMetadata(e){return this._source.getMetadata(this._firstTokenIndex+e)}getLanguageId(e){return this._source.getLanguageId(this._firstTokenIndex+e)}getLineContent(){return this._source.getLineContent().substring(this._startOffset,this._endOffset)}equals(e){return e instanceof vW?this._startOffset===e._startOffset&&this._endOffset===e._endOffset&&this._deltaOffset===e._deltaOffset&&this._source.slicedEquals(e._source,this._firstTokenIndex,this._tokensCount):!1}getCount(){return this._tokensCount}getStandardTokenType(e){return this._source.getStandardTokenType(this._firstTokenIndex+e)}getForeground(e){return this._source.getForeground(this._firstTokenIndex+e)}getEndOffset(e){const t=this._source.getEndOffset(this._firstTokenIndex+e);return Math.min(this._endOffset,t)-this._startOffset+this._deltaOffset}getClassName(e){return this._source.getClassName(this._firstTokenIndex+e)}getInlineStyle(e,t){return this._source.getInlineStyle(this._firstTokenIndex+e,t)}getPresentation(e){return this._source.getPresentation(this._firstTokenIndex+e)}findTokenIndexAtOffset(e){return this._source.findTokenIndexAtOffset(e+this._startOffset-this._deltaOffset)-this._firstTokenIndex}getTokenText(e){const t=this._firstTokenIndex+e,i=this._source.getStartOffset(t),n=this._source.getEndOffset(t);let o=this._source.getTokenText(t);return i<this._startOffset&&(o=o.substring(this._startOffset-i)),n>this._endOffset&&(o=o.substring(0,o.length-(n-this._endOffset))),o}forEach(e){for(let t=0;t<this.getCount();t++)e(t)}}function afe(s,e){const t=e.lineNumber;if(!s.tokenization.isCheapToTokenize(t))return;s.tokenization.forceTokenization(t);const i=s.tokenization.getLineTokens(t),n=i.findTokenIndexAtOffset(e.column-1);return i.getStandardTokenType(n)}class ih{static fromLineTokens(e){const t=[];for(let i=0;i<e.getCount();i++)t.push(new WP(e.getEndOffset(i)-e.getStartOffset(i),e.getMetadata(i)));return ih.create(t)}static create(e){return new ih(e)}constructor(e){this._tokenInfo=e}toLineTokens(e,t){return Ri.createFromTextAndMetadata(this.map((i,n)=>({text:i.substring(e),metadata:n.metadata})),t)}forEach(e){let t=0;for(const i of this._tokenInfo){const n=new me(t,t+i.length);e(n,i),t+=i.length}}map(e){const t=[];let i=0;for(const n of this._tokenInfo){const o=new me(i,i+n.length);t.push(e(o,n)),i+=n.length}return t}slice(e){const t=[];let i=0;for(const n of this._tokenInfo){const o=i,r=o+n.length;if(r>e.start){if(o>=e.endExclusive)break;const a=Math.max(0,e.start-o),l=Math.max(0,r-e.endExclusive);t.push(new WP(n.length-a-l,n.metadata))}i+=n.length}return ih.create(t)}}class WP{constructor(e,t){this.length=e,this.metadata=t}}class lfe{constructor(){this._tokens=[]}add(e,t){this._tokens.push(new WP(e,t))}build(){return ih.create(this._tokens)}}class CW{constructor(e,t,i){this._indentRulesSupport=t,this._indentationLineProcessor=new mJ(e,i)}shouldIncrease(e,t){const i=this._indentationLineProcessor.getProcessedLine(e,t);return this._indentRulesSupport.shouldIncrease(i)}shouldDecrease(e,t){const i=this._indentationLineProcessor.getProcessedLine(e,t);return this._indentRulesSupport.shouldDecrease(i)}shouldIgnore(e,t){const i=this._indentationLineProcessor.getProcessedLine(e,t);return this._indentRulesSupport.shouldIgnore(i)}shouldIndentNextLine(e,t){const i=this._indentationLineProcessor.getProcessedLine(e,t);return this._indentRulesSupport.shouldIndentNextLine(i)}}class yW{constructor(e,t){this.model=e,this.indentationLineProcessor=new mJ(e,t)}getProcessedTokenContextAroundRange(e){const t=this._getProcessedTokensBeforeRange(e),i=this._getProcessedTokensAfterRange(e),n=this._getProcessedPreviousLineTokens(e);return{beforeRangeProcessedTokens:t,afterRangeProcessedTokens:i,previousLineProcessedTokens:n}}_getProcessedTokensBeforeRange(e){this.model.tokenization.forceTokenization(e.startLineNumber);const t=this.model.tokenization.getLineTokens(e.startLineNumber),i=Uf(t,e.startColumn-1);let n;if(SW(this.model,e.getStartPosition())){const r=e.startColumn-1-i.firstCharOffset,a=i.firstCharOffset,l=a+r;n=t.sliceAndInflate(a,l,0)}else{const r=e.startColumn-1;n=t.sliceAndInflate(0,r,0)}return this.indentationLineProcessor.getProcessedTokens(n)}_getProcessedTokensAfterRange(e){const t=e.isEmpty()?e.getStartPosition():e.getEndPosition();this.model.tokenization.forceTokenization(t.lineNumber);const i=this.model.tokenization.getLineTokens(t.lineNumber),n=Uf(i,t.column-1),o=t.column-1-n.firstCharOffset,r=n.firstCharOffset+o,a=n.firstCharOffset+n.getLineLength(),l=i.sliceAndInflate(r,a,0);return this.indentationLineProcessor.getProcessedTokens(l)}_getProcessedPreviousLineTokens(e){const t=g=>{this.model.tokenization.forceTokenization(g);const f=this.model.tokenization.getLineTokens(g),p=this.model.getLineMaxColumn(g)-1;return Uf(f,p)};this.model.tokenization.forceTokenization(e.startLineNumber);const i=this.model.tokenization.getLineTokens(e.startLineNumber),n=Uf(i,e.startColumn-1),o=Ri.createEmpty("",n.languageIdCodec),r=e.startLineNumber-1;if(r===0||!(n.firstCharOffset===0))return o;const c=t(r);if(!(n.languageId===c.languageId))return o;const h=c.toIViewLineTokens();return this.indentationLineProcessor.getProcessedTokens(h)}}class mJ{constructor(e,t){this.model=e,this.languageConfigurationService=t}getProcessedLine(e,t){const i=(r,a)=>{const l=Bt(r);return a+r.substring(l.length)};this.model.tokenization.forceTokenization?.(e);const n=this.model.tokenization.getLineTokens(e);let o=this.getProcessedTokens(n).getLineContent();return t!==void 0&&(o=i(o,t)),o}getProcessedTokens(e){const t=l=>l===2||l===3||l===1,i=e.getLanguageId(0),o=this.languageConfigurationService.getLanguageConfiguration(i).bracketsNew.getBracketRegExp({global:!0}),r=[];return e.forEach(l=>{const c=e.getStandardTokenType(l);let d=e.getTokenText(l);t(c)&&(d=d.replace(o,""));const h=e.getMetadata(l);r.push({text:d,metadata:h})}),Ri.createFromTextAndMetadata(r,e.languageIdCodec)}}function SW(s,e){s.tokenization.forceTokenization(e.lineNumber);const t=s.tokenization.getLineTokens(e.lineNumber),i=Uf(t,e.column-1),n=i.firstCharOffset===0,o=t.getLanguageId(0)===i.languageId;return!n&&!o}function dw(s,e,t,i){e.tokenization.forceTokenization(t.startLineNumber);const n=e.getLanguageIdAtPosition(t.startLineNumber,t.startColumn),o=i.getLanguageConfiguration(n);if(!o)return null;const a=new yW(e,i).getProcessedTokenContextAroundRange(t),l=a.previousLineProcessedTokens.getLineContent(),c=a.beforeRangeProcessedTokens.getLineContent(),d=a.afterRangeProcessedTokens.getLineContent(),h=o.onEnter(s,l,c,d);if(!h)return null;const u=h.indentAction;let g=h.appendText;const f=h.removeText||0;g?u===Xi.Indent&&(g="	"+g):u===Xi.Indent||u===Xi.IndentOutdent?g="	":g="";let p=fJ(e,t.startLineNumber,t.startColumn);return f&&(p=p.substring(0,p.length-f)),{indentAction:u,appendText:g,removeText:f,indentation:p}}var cfe=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},dfe=function(s,e){return function(t,i){e(t,i,s)}},Dk;const X2=Object.create(null);function Bg(s,e){if(e<=0)return"";X2[s]||(X2[s]=["",s]);const t=X2[s];for(let i=t.length;i<=e;i++)t[i]=t[i-1]+s;return t[e]}let la=Dk=class{static unshiftIndent(e,t,i,n,o){const r=pi.visibleColumnFromColumn(e,t,i);if(o){const a=Bg(" ",n),c=pi.prevIndentTabStop(r,n)/n;return Bg(a,c)}else{const c=pi.prevRenderTabStop(r,i)/i;return Bg("	",c)}}static shiftIndent(e,t,i,n,o){const r=pi.visibleColumnFromColumn(e,t,i);if(o){const a=Bg(" ",n),c=pi.nextIndentTabStop(r,n)/n;return Bg(a,c)}else{const c=pi.nextRenderTabStop(r,i)/i;return Bg("	",c)}}constructor(e,t,i){this._languageConfigurationService=i,this._opts=t,this._selection=e,this._selectionId=null,this._useLastEditRangeForCursorEndPosition=!1,this._selectionStartColumnStaysPut=!1}_addEditOperation(e,t,i){this._useLastEditRangeForCursorEndPosition?e.addTrackedEditOperation(t,i):e.addEditOperation(t,i)}getEditOperations(e,t){const i=this._selection.startLineNumber;let n=this._selection.endLineNumber;this._selection.endColumn===1&&i!==n&&(n=n-1);const{tabSize:o,indentSize:r,insertSpaces:a}=this._opts,l=i===n;if(this._opts.useTabStops){this._selection.isEmpty()&&/^\s*$/.test(e.getLineContent(i))&&(this._useLastEditRangeForCursorEndPosition=!0);let c=0,d=0;for(let h=i;h<=n;h++,c=d){d=0;const u=e.getLineContent(h);let g=As(u);if(this._opts.isUnshift&&(u.length===0||g===0)||!l&&!this._opts.isUnshift&&u.length===0)continue;if(g===-1&&(g=u.length),h>1&&pi.visibleColumnFromColumn(u,g+1,o)%r!==0&&e.tokenization.isCheapToTokenize(h-1)){const _=dw(this._opts.autoIndent,e,new k(h-1,e.getLineMaxColumn(h-1),h-1,e.getLineMaxColumn(h-1)),this._languageConfigurationService);if(_){if(d=c,_.appendText)for(let b=0,w=_.appendText.length;b<w&&d<r&&_.appendText.charCodeAt(b)===32;b++)d++;_.removeText&&(d=Math.max(0,d-_.removeText));for(let b=0;b<d&&!(g===0||u.charCodeAt(g-1)!==32);b++)g--}}if(this._opts.isUnshift&&g===0)continue;let f;this._opts.isUnshift?f=Dk.unshiftIndent(u,g+1,o,r,a):f=Dk.shiftIndent(u,g+1,o,r,a),this._addEditOperation(t,new k(h,1,h,g+1),f),h===i&&!this._selection.isEmpty()&&(this._selectionStartColumnStaysPut=this._selection.startColumn<=g+1)}}else{!this._opts.isUnshift&&this._selection.isEmpty()&&e.getLineLength(i)===0&&(this._useLastEditRangeForCursorEndPosition=!0);const c=a?Bg(" ",r):"	";for(let d=i;d<=n;d++){const h=e.getLineContent(d);let u=As(h);if(!(this._opts.isUnshift&&(h.length===0||u===0))&&!(!l&&!this._opts.isUnshift&&h.length===0)&&(u===-1&&(u=h.length),!(this._opts.isUnshift&&u===0)))if(this._opts.isUnshift){u=Math.min(u,r);for(let g=0;g<u;g++)if(h.charCodeAt(g)===9){u=g+1;break}this._addEditOperation(t,new k(d,1,d,u+1),"")}else this._addEditOperation(t,new k(d,1,d,1),c),d===i&&!this._selection.isEmpty()&&(this._selectionStartColumnStaysPut=this._selection.startColumn===1)}}this._selectionId=t.trackSelection(this._selection)}computeCursorState(e,t){if(this._useLastEditRangeForCursorEndPosition){const n=t.getInverseEditOperations()[0];return new he(n.range.endLineNumber,n.range.endColumn,n.range.endLineNumber,n.range.endColumn)}const i=t.getTrackedSelection(this._selectionId);if(this._selectionStartColumnStaysPut){const n=this._selection.startColumn;return i.startColumn<=n?i:i.getDirection()===0?new he(i.startLineNumber,n,i.endLineNumber,i.endColumn):new he(i.endLineNumber,i.endColumn,i.startLineNumber,n)}return i}};la=Dk=cfe([dfe(2,ui)],la);class hfe{constructor(e,t,i){this._range=e,this._charBeforeSelection=t,this._charAfterSelection=i}getEditOperations(e,t){t.addTrackedEditOperation(new k(this._range.startLineNumber,this._range.startColumn,this._range.startLineNumber,this._range.startColumn),this._charBeforeSelection),t.addTrackedEditOperation(new k(this._range.endLineNumber,this._range.endColumn,this._range.endLineNumber,this._range.endColumn),this._charAfterSelection||null)}computeCursorState(e,t){const i=t.getInverseEditOperations(),n=i[0].range,o=i[1].range;return new he(n.endLineNumber,n.endColumn,o.endLineNumber,o.endColumn-this._charAfterSelection.length)}}class ufe{constructor(e,t,i){this._position=e,this._text=t,this._charAfter=i}getEditOperations(e,t){t.addTrackedEditOperation(new k(this._position.lineNumber,this._position.column,this._position.lineNumber,this._position.column),this._text+this._charAfter)}computeCursorState(e,t){const n=t.getInverseEditOperations()[0].range;return new he(n.endLineNumber,n.startColumn,n.endLineNumber,n.endColumn-this._charAfter.length)}}function gfe(s,e,t){const i=s.tokenization.getLanguageIdAtPosition(e,0);if(e>1){let n,o=-1;for(n=e-1;n>=1;n--){if(s.tokenization.getLanguageIdAtPosition(n,0)!==i)return o;const r=s.getLineContent(n);if(t.shouldIgnore(n)||/^\s+$/.test(r)||r===""){o=n;continue}return n}}return-1}function nS(s,e,t,i=!0,n){if(s<4)return null;const o=n.getLanguageConfiguration(e.tokenization.getLanguageId()).indentRulesSupport;if(!o)return null;const r=new CW(e,o,n);if(t<=1)return{indentation:"",action:null};for(let l=t-1;l>0&&e.getLineContent(l)==="";l--)if(l===1)return{indentation:"",action:null};const a=gfe(e,t,r);if(a<0)return null;if(a<1)return{indentation:"",action:null};if(r.shouldIncrease(a)||r.shouldIndentNextLine(a)){const l=e.getLineContent(a);return{indentation:Bt(l),action:Xi.Indent,line:a}}else if(r.shouldDecrease(a)){const l=e.getLineContent(a);return{indentation:Bt(l),action:null,line:a}}else{if(a===1)return{indentation:Bt(e.getLineContent(a)),action:null,line:a};const l=a-1,c=o.getIndentMetadata(e.getLineContent(l));if(!(c&3)&&c&4){let d=0;for(let h=l-1;h>0;h--)if(!r.shouldIndentNextLine(h)){d=h;break}return{indentation:Bt(e.getLineContent(d+1)),action:null,line:d+1}}if(i)return{indentation:Bt(e.getLineContent(a)),action:null,line:a};for(let d=a;d>0;d--){if(r.shouldIncrease(d))return{indentation:Bt(e.getLineContent(d)),action:Xi.Indent,line:d};if(r.shouldIndentNextLine(d)){let h=0;for(let u=d-1;u>0;u--)if(!r.shouldIndentNextLine(d)){h=u;break}return{indentation:Bt(e.getLineContent(h+1)),action:null,line:h+1}}else if(r.shouldDecrease(d))return{indentation:Bt(e.getLineContent(d)),action:null,line:d}}return{indentation:Bt(e.getLineContent(1)),action:null,line:1}}}function D1(s,e,t,i,n,o){if(s<4)return null;const r=o.getLanguageConfiguration(t);if(!r)return null;const a=o.getLanguageConfiguration(t).indentRulesSupport;if(!a)return null;const l=new CW(e,a,o),c=nS(s,e,i,void 0,o);if(c){const d=c.line;if(d!==void 0){let h=!0;for(let u=d;u<i-1;u++)if(!/^\s*$/.test(e.getLineContent(u))){h=!1;break}if(h){const u=r.onEnter(s,"",e.getLineContent(d),"");if(u){let g=Bt(e.getLineContent(d));return u.removeText&&(g=g.substring(0,g.length-u.removeText)),u.indentAction===Xi.Indent||u.indentAction===Xi.IndentOutdent?g=n.shiftIndent(g):u.indentAction===Xi.Outdent&&(g=n.unshiftIndent(g)),l.shouldDecrease(i)&&(g=n.unshiftIndent(g)),u.appendText&&(g+=u.appendText),Bt(g)}}}return l.shouldDecrease(i)?c.action===Xi.Indent?c.indentation:n.unshiftIndent(c.indentation):c.action===Xi.Indent?n.shiftIndent(c.indentation):c.indentation}return null}function ffe(s,e,t,i,n){if(s<4)return null;const o=e.getLanguageIdAtPosition(t.startLineNumber,t.startColumn),r=n.getLanguageConfiguration(o).indentRulesSupport;if(!r)return null;e.tokenization.forceTokenization(t.startLineNumber);const l=new yW(e,n).getProcessedTokenContextAroundRange(t),c=l.afterRangeProcessedTokens,d=l.beforeRangeProcessedTokens,h=Bt(d.getLineContent()),u=mfe(e,t.startLineNumber,d),g=SW(e,t.getStartPosition()),f=e.getLineContent(t.startLineNumber),p=Bt(f),_=nS(s,u,t.startLineNumber+1,void 0,n);if(!_){const w=g?p:h;return{beforeEnter:w,afterEnter:w}}let b=g?p:_.indentation;return _.action===Xi.Indent&&(b=i.shiftIndent(b)),r.shouldDecrease(c.getLineContent())&&(b=i.unshiftIndent(b)),{beforeEnter:g?p:h,afterEnter:b}}function pfe(s,e,t,i,n,o){const r=s.autoIndent;if(r<4||SW(e,t.getStartPosition()))return null;const l=e.getLanguageIdAtPosition(t.startLineNumber,t.startColumn),c=o.getLanguageConfiguration(l).indentRulesSupport;if(!c)return null;const h=new yW(e,o).getProcessedTokenContextAroundRange(t),u=h.beforeRangeProcessedTokens.getLineContent(),g=h.afterRangeProcessedTokens.getLineContent(),f=u+g,p=u+i+g;if(!c.shouldDecrease(f)&&c.shouldDecrease(p)){const b=nS(r,e,t.startLineNumber,!1,o);if(!b)return null;let w=b.indentation;return b.action!==Xi.Indent&&(w=n.unshiftIndent(w)),w}const _=t.startLineNumber-1;if(_>0){const b=e.getLineContent(_);if(c.shouldIndentNextLine(b)&&c.shouldIncrease(p)){const C=nS(r,e,t.startLineNumber,!1,o)?.indentation;if(C!==void 0){const v=e.getLineContent(t.startLineNumber),y=Bt(v),L=n.shiftIndent(C)===y,E=/^\s*$/.test(f),I=s.autoClosingPairs.autoClosingPairsOpenByEnd.get(i),O=I&&I.length>0&&E;if(L&&O)return C}}}return null}function _J(s,e,t){const i=t.getLanguageConfiguration(s.getLanguageId()).indentRulesSupport;return!i||e<1||e>s.getLineCount()?null:i.getIndentMetadata(s.getLineContent(e))}function mfe(s,e,t){return{tokenization:{getLineTokens:n=>n===e?t:s.tokenization.getLineTokens(n),getLanguageId:()=>s.getLanguageId(),getLanguageIdAtPosition:(n,o)=>s.getLanguageIdAtPosition(n,o)},getLineContent:n=>n===e?t.getLineContent():s.getLineContent(n)}}class _fe{static getEdits(e,t,i,n,o){if(!o&&this._isAutoIndentType(e,t,i)){const r=[];for(const l of i){const c=this._findActualIndentationForSelection(e,t,l,n);if(c===null)return;r.push({selection:l,indentation:c})}const a=HP.getAutoClosingPairClose(e,t,i,n,!1);return this._getIndentationAndAutoClosingPairEdits(e,t,r,n,a)}}static _isAutoIndentType(e,t,i){if(e.autoIndent<4)return!1;for(let n=0,o=i.length;n<o;n++)if(!t.tokenization.isCheapToTokenize(i[n].getEndPosition().lineNumber))return!1;return!0}static _findActualIndentationForSelection(e,t,i,n){const o=pfe(e,t,i,n,{shiftIndent:a=>kW(e,a),unshiftIndent:a=>OE(e,a)},e.languageConfigurationService);if(o===null)return null;const r=fJ(t,i.startLineNumber,i.startColumn);return o===e.normalizeIndentation(r)?null:o}static _getIndentationAndAutoClosingPairEdits(e,t,i,n,o){const r=i.map(({selection:l,indentation:c})=>{if(o!==null){const d=this._getEditFromIndentationAndSelection(e,t,c,l,n,!1);return new Ife(d,l,n,o)}else{const d=this._getEditFromIndentationAndSelection(e,t,c,l,n,!0);return tf(d.range,d.text,!1)}}),a={shouldPushStackElementBefore:!0,shouldPushStackElementAfter:!1};return new to(4,r,a)}static _getEditFromIndentationAndSelection(e,t,i,n,o,r=!0){const a=n.startLineNumber,l=t.getLineFirstNonWhitespaceColumn(a);let c=e.normalizeIndentation(i);if(l!==0){const h=t.getLineContent(a);c+=h.substring(l-1,n.startColumn-1)}return c+=r?o:"",{range:new k(a,1,n.endLineNumber,n.endColumn),text:c}}}class bfe{static getEdits(e,t,i,n,o,r){if(bJ(t,i,n,o,r))return this._runAutoClosingOvertype(e,n,r)}static _runAutoClosingOvertype(e,t,i){const n=[];for(let o=0,r=t.length;o<r;o++){const l=t[o].getPosition(),c=new k(l.lineNumber,l.column,l.lineNumber,l.column+1);n[o]=new ts(c,i)}return new to(4,n,{shouldPushStackElementBefore:XR(e,4),shouldPushStackElementAfter:!1})}}class wfe{static getEdits(e,t,i,n,o){if(bJ(e,t,i,n,o)){const r=i.map(a=>new ts(new k(a.positionLineNumber,a.positionColumn,a.positionLineNumber,a.positionColumn+1),"",!1));return new to(4,r,{shouldPushStackElementBefore:!0,shouldPushStackElementAfter:!1})}}}class HP{static getEdits(e,t,i,n,o,r){if(!r){const a=this.getAutoClosingPairClose(e,t,i,n,o);if(a!==null)return this._runAutoClosingOpenCharType(i,n,o,a)}}static _runAutoClosingOpenCharType(e,t,i,n){const o=[];for(let r=0,a=e.length;r<a;r++){const l=e[r];o[r]=new Efe(l,t,!i,n)}return new to(4,o,{shouldPushStackElementBefore:!0,shouldPushStackElementAfter:!1})}static getAutoClosingPairClose(e,t,i,n,o){for(const f of i)if(!f.isEmpty())return null;const r=i.map(f=>{const p=f.getPosition();return o?{lineNumber:p.lineNumber,beforeColumn:p.column-n.length,afterColumn:p.column}:{lineNumber:p.lineNumber,beforeColumn:p.column,afterColumn:p.column}}),a=this._findAutoClosingPairOpen(e,t,r.map(f=>new A(f.lineNumber,f.beforeColumn)),n);if(!a)return null;let l,c;if(ug(n)?(l=e.autoClosingQuotes,c=e.shouldAutoCloseBefore.quote):(e.blockCommentStartToken?a.open.includes(e.blockCommentStartToken):!1)?(l=e.autoClosingComments,c=e.shouldAutoCloseBefore.comment):(l=e.autoClosingBrackets,c=e.shouldAutoCloseBefore.bracket),l==="never")return null;const h=this._findContainedAutoClosingPair(e,a),u=h?h.close:"";let g=!0;for(const f of r){const{lineNumber:p,beforeColumn:_,afterColumn:b}=f,w=t.getLineContent(p),C=w.substring(0,_-1),v=w.substring(b-1);if(v.startsWith(u)||(g=!1),v.length>0){const E=v.charAt(0);if(!this._isBeforeClosingBrace(e,v)&&!c(E))return null}if(a.open.length===1&&(n==="'"||n==='"')&&l!=="always"){const E=aa(e.wordSeparators,[]);if(C.length>0){const I=C.charCodeAt(C.length-1);if(E.get(I)===0)return null}}if(!t.tokenization.isCheapToTokenize(p))return null;t.tokenization.forceTokenization(p);const y=t.tokenization.getLineTokens(p),x=Uf(y,_-1);if(!a.shouldAutoClose(x,_-x.firstCharOffset))return null;const L=a.findNeutralCharacter();if(L){const E=t.tokenization.getTokenTypeIfInsertingCharacter(p,_,L);if(!a.isOK(E))return null}}return g?a.close.substring(0,a.close.length-u.length):a.close}static _findContainedAutoClosingPair(e,t){if(t.open.length<=1)return null;const i=t.close.charAt(t.close.length-1),n=e.autoClosingPairs.autoClosingPairsCloseByEnd.get(i)||[];let o=null;for(const r of n)r.open!==t.open&&t.open.includes(r.open)&&t.close.endsWith(r.close)&&(!o||r.open.length>o.open.length)&&(o=r);return o}static _findAutoClosingPairOpen(e,t,i,n){const o=e.autoClosingPairs.autoClosingPairsOpenByEnd.get(n);if(!o)return null;let r=null;for(const a of o)if(r===null||a.open.length>r.open.length){let l=!0;for(const c of i)if(t.getValueInRange(new k(c.lineNumber,c.column-a.open.length+1,c.lineNumber,c.column))+n!==a.open){l=!1;break}l&&(r=a)}return r}static _isBeforeClosingBrace(e,t){const i=t.charAt(0),n=e.autoClosingPairs.autoClosingPairsOpenByStart.get(i)||[],o=e.autoClosingPairs.autoClosingPairsCloseByStart.get(i)||[],r=n.some(l=>t.startsWith(l.open)),a=o.some(l=>t.startsWith(l.close));return!r&&a}}class C${static getEdits(e,t){if(!(e.inputMode==="overtype"))return null;const n=t.map(o=>new wge(o.insertedTextRange));return new to(4,n,{shouldPushStackElementBefore:!0,shouldPushStackElementAfter:!1})}}class vfe{static getEdits(e,t,i,n,o){if(!o&&this._isSurroundSelectionType(e,t,i,n))return this._runSurroundSelectionType(e,i,n)}static _runSurroundSelectionType(e,t,i){const n=[];for(let o=0,r=t.length;o<r;o++){const a=t[o],l=e.surroundingPairs[i];n[o]=new hfe(a,i,l)}return new to(0,n,{shouldPushStackElementBefore:!0,shouldPushStackElementAfter:!0})}static _isSurroundSelectionType(e,t,i,n){if(!wJ(e,n)||!e.surroundingPairs.hasOwnProperty(n))return!1;const o=ug(n);for(const r of i){if(r.isEmpty())return!1;let a=!0;for(let l=r.startLineNumber;l<=r.endLineNumber;l++){const c=t.getLineContent(l),d=l===r.startLineNumber?r.startColumn-1:0,h=l===r.endLineNumber?r.endColumn-1:c.length,u=c.substring(d,h);if(/[^ \t]/.test(u)){a=!1;break}}if(a)return!1;if(o&&r.startLineNumber===r.endLineNumber&&r.startColumn+1===r.endColumn){const l=t.getValueInRange(r);if(ug(l))return!1}}return!0}}class Cfe{static getEdits(e,t,i,n,o,r){if(!r&&this._isTypeInterceptorElectricChar(t,i,n)){const a=this._typeInterceptorElectricChar(e,t,i,n[0],o);if(a)return a}}static _isTypeInterceptorElectricChar(e,t,i){return!!(i.length===1&&t.tokenization.isCheapToTokenize(i[0].getEndPosition().lineNumber))}static _typeInterceptorElectricChar(e,t,i,n,o){if(!t.electricChars.hasOwnProperty(o)||!n.isEmpty())return null;const r=n.getPosition();i.tokenization.forceTokenization(r.lineNumber);const a=i.tokenization.getLineTokens(r.lineNumber);let l;try{l=t.onElectricCharacter(o,a,r.column)}catch(c){return Oe(c),null}if(!l)return null;if(l.matchOpenBracket){const c=(a.getLineContent()+o).lastIndexOf(l.matchOpenBracket)+1,d=i.bracketPairs.findMatchingBracketUp(l.matchOpenBracket,{lineNumber:r.lineNumber,column:c},500);if(d){if(d.startLineNumber===r.lineNumber)return null;const h=i.getLineContent(d.startLineNumber),u=Bt(h),g=t.normalizeIndentation(u),f=i.getLineContent(r.lineNumber),p=i.getLineFirstNonWhitespaceColumn(r.lineNumber)||r.column,_=f.substring(p-1,r.column-1),b=g+_+o,w=new k(r.lineNumber,1,r.lineNumber,r.column),C=new ts(w,b);return new to(LW(b,e),[C],{shouldPushStackElementBefore:!1,shouldPushStackElementAfter:!0})}}return null}}class yfe{static getEdits(e,t,i,n,o){const r=[];for(let l=0,c=i.length;l<c;l++){const d=e.inputMode==="overtype"&&!o?AP:ts;r[l]=new d(i[l],n)}const a=LW(n,t);return new to(a,r,{shouldPushStackElementBefore:XR(t,a),shouldPushStackElementAfter:!1})}}class ZR{static getEdits(e,t,i,n,o){if(!o&&n===`
`){const r=[];for(let a=0,l=i.length;a<l;a++)r[a]=this._enter(e,t,!1,i[a]);return new to(4,r,{shouldPushStackElementBefore:!0,shouldPushStackElementAfter:!1})}}static _enter(e,t,i,n){if(e.autoIndent===0)return tf(n,`
`,i);if(!t.tokenization.isCheapToTokenize(n.getStartPosition().lineNumber)||e.autoIndent===1){const l=t.getLineContent(n.startLineNumber),c=Bt(l).substring(0,n.startColumn-1);return tf(n,`
`+e.normalizeIndentation(c),i)}const o=dw(e.autoIndent,t,n,e.languageConfigurationService);if(o){if(o.indentAction===Xi.None)return tf(n,`
`+e.normalizeIndentation(o.indentation+o.appendText),i);if(o.indentAction===Xi.Indent)return tf(n,`
`+e.normalizeIndentation(o.indentation+o.appendText),i);if(o.indentAction===Xi.IndentOutdent){const l=e.normalizeIndentation(o.indentation),c=e.normalizeIndentation(o.indentation+o.appendText),d=`
`+c+`
`+l;return i?new Lk(n,d,!0):new IE(n,d,-1,c.length-l.length,!0)}else if(o.indentAction===Xi.Outdent){const l=OE(e,o.indentation);return tf(n,`
`+e.normalizeIndentation(l+o.appendText),i)}}const r=t.getLineContent(n.startLineNumber),a=Bt(r).substring(0,n.startColumn-1);if(e.autoIndent>=4){const l=ffe(e.autoIndent,t,n,{unshiftIndent:c=>OE(e,c),shiftIndent:c=>kW(e,c),normalizeIndentation:c=>e.normalizeIndentation(c)},e.languageConfigurationService);if(l){let c=e.visibleColumnFromColumn(t,n.getEndPosition());const d=n.endColumn,h=t.getLineContent(n.endLineNumber),u=As(h);if(u>=0?n=n.setEndPosition(n.endLineNumber,Math.max(n.endColumn,u+1)):n=n.setEndPosition(n.endLineNumber,t.getLineMaxColumn(n.endLineNumber)),i)return new Lk(n,`
`+e.normalizeIndentation(l.afterEnter),!0);{let g=0;return d<=u+1&&(e.insertSpaces||(c=Math.ceil(c/e.indentSize)),g=Math.min(c+1-e.normalizeIndentation(l.afterEnter).length-1,0)),new IE(n,`
`+e.normalizeIndentation(l.afterEnter),0,g,!0)}}}return tf(n,`
`+e.normalizeIndentation(a),i)}static lineInsertBefore(e,t,i){if(t===null||i===null)return[];const n=[];for(let o=0,r=i.length;o<r;o++){let a=i[o].positionLineNumber;if(a===1)n[o]=new Lk(new k(1,1,1,1),`
`);else{a--;const l=t.getLineMaxColumn(a);n[o]=this._enter(e,t,!1,new k(a,l,a,l))}}return n}static lineInsertAfter(e,t,i){if(t===null||i===null)return[];const n=[];for(let o=0,r=i.length;o<r;o++){const a=i[o].positionLineNumber,l=t.getLineMaxColumn(a);n[o]=this._enter(e,t,!1,new k(a,l,a,l))}return n}static lineBreakInsert(e,t,i){const n=[];for(let o=0,r=i.length;o<r;o++)n[o]=this._enter(e,t,!0,i[o]);return n}}class Sfe{static getEdits(e,t,i,n,o,r){const a=this._distributePasteToCursors(e,i,n,o,r);return a?(i=i.sort(k.compareRangesUsingStarts),this._distributedPaste(e,t,i,a)):this._simplePaste(e,t,i,n,o)}static _distributePasteToCursors(e,t,i,n,o){if(n||t.length===1)return null;if(o&&o.length===t.length)return o;if(e.multiCursorPaste==="spread"){i.charCodeAt(i.length-1)===10&&(i=i.substring(0,i.length-1)),i.charCodeAt(i.length-1)===13&&(i=i.substring(0,i.length-1));const r=qo(i);if(r.length===t.length)return r}return null}static _distributedPaste(e,t,i,n){const o=[];for(let r=0,a=i.length;r<a;r++){const c=e.overtypeOnPaste&&e.inputMode==="overtype"?AP:ts;o[r]=new c(i[r],n[r])}return new to(0,o,{shouldPushStackElementBefore:!0,shouldPushStackElementAfter:!0})}static _simplePaste(e,t,i,n,o){const r=[];for(let a=0,l=i.length;a<l;a++){const c=i[a],d=c.getPosition();if(o&&!c.isEmpty()&&(o=!1),o&&n.indexOf(`
`)!==n.length-1&&(o=!1),o){const h=new k(d.lineNumber,1,d.lineNumber,1);r[a]=new pW(h,n,c,!0)}else{const u=e.overtypeOnPaste&&e.inputMode==="overtype"?AP:ts;r[a]=new u(c,n)}}return new to(0,r,{shouldPushStackElementBefore:!0,shouldPushStackElementAfter:!0})}}class xfe{static getEdits(e,t,i,n,o,r,a,l){const c=n.map(d=>this._compositionType(i,d,o,r,a,l));return new to(4,c,{shouldPushStackElementBefore:XR(e,4),shouldPushStackElementAfter:!1})}static _compositionType(e,t,i,n,o,r){if(!t.isEmpty())return null;const a=t.getPosition(),l=Math.max(1,a.column-n),c=Math.min(e.getLineMaxColumn(a.lineNumber),a.column+o),d=new k(a.lineNumber,l,a.lineNumber,c);return new IE(d,i,0,r)}}class Lfe{static getEdits(e,t,i){const n=[];for(let r=0,a=t.length;r<a;r++)n[r]=new ts(t[r],i);const o=LW(i,e);return new to(o,n,{shouldPushStackElementBefore:XR(e,o),shouldPushStackElementAfter:!1})}}class kfe{static getCommands(e,t,i){const n=[];for(let o=0,r=i.length;o<r;o++){const a=i[o];if(a.isEmpty()){const l=t.getLineContent(a.startLineNumber);if(/^\s*$/.test(l)&&t.tokenization.isCheapToTokenize(a.startLineNumber)){let c=this._goodIndentForLine(e,t,a.startLineNumber);c=c||"	";const d=e.normalizeIndentation(c);if(!l.startsWith(d)){n[o]=new ts(new k(a.startLineNumber,1,a.startLineNumber,l.length+1),d,!0);continue}}n[o]=this._replaceJumpToNextIndent(e,t,a,!0)}else{if(a.startLineNumber===a.endLineNumber){const l=t.getLineMaxColumn(a.startLineNumber);if(a.startColumn!==1||a.endColumn!==l){n[o]=this._replaceJumpToNextIndent(e,t,a,!1);continue}}n[o]=new la(a,{isUnshift:!1,tabSize:e.tabSize,indentSize:e.indentSize,insertSpaces:e.insertSpaces,useTabStops:e.useTabStops,autoIndent:e.autoIndent},e.languageConfigurationService)}}return n}static _goodIndentForLine(e,t,i){let n=null,o="";const r=nS(e.autoIndent,t,i,!1,e.languageConfigurationService);if(r)n=r.action,o=r.indentation;else if(i>1){let a;for(a=i-1;a>=1;a--){const d=t.getLineContent(a);if(Ma(d)>=0)break}if(a<1)return null;const l=t.getLineMaxColumn(a),c=dw(e.autoIndent,t,new k(a,l,a,l),e.languageConfigurationService);c&&(o=c.indentation+c.appendText)}return n&&(n===Xi.Indent&&(o=kW(e,o)),n===Xi.Outdent&&(o=OE(e,o)),o=e.normalizeIndentation(o)),o||null}static _replaceJumpToNextIndent(e,t,i,n){let o="";const r=i.getStartPosition();if(e.insertSpaces){const a=e.visibleColumnFromColumn(t,r),l=e.indentSize,c=l-a%l;for(let d=0;d<c;d++)o+=" "}else o="	";return new ts(i,o,n)}}class xW extends IE{constructor(e,t,i,n,o,r){super(e,t,i,n),this._openCharacter=o,this._closeCharacter=r,this.closeCharacterRange=null,this.enclosingRange=null}_computeCursorStateWithRange(e,t,i){return this.closeCharacterRange=new k(t.startLineNumber,t.endColumn-this._closeCharacter.length,t.endLineNumber,t.endColumn),this.enclosingRange=new k(t.startLineNumber,t.endColumn-this._openCharacter.length-this._closeCharacter.length,t.endLineNumber,t.endColumn),super.computeCursorState(e,i)}}class Efe extends xW{constructor(e,t,i,n){const o=(i?t:"")+n,r=0,a=-n.length;super(e,o,r,a,t,n)}computeCursorState(e,t){const n=t.getInverseEditOperations()[0].range;return this._computeCursorStateWithRange(e,n,t)}}class Ife extends xW{constructor(e,t,i,n){const o=i+n,r=0,a=i.length;super(t,o,r,a,i,n),this._autoIndentationEdit=e,this._autoClosingEdit={range:t,text:o}}getEditOperations(e,t){t.addTrackedEditOperation(this._autoIndentationEdit.range,this._autoIndentationEdit.text),t.addTrackedEditOperation(this._autoClosingEdit.range,this._autoClosingEdit.text)}computeCursorState(e,t){const i=t.getInverseEditOperations();if(i.length!==2)throw new Error("There should be two inverse edit operations!");const n=i[0].range,o=i[1].range,r=n.plusRange(o);return this._computeCursorStateWithRange(e,r,t)}}function LW(s,e){return s===" "?e===5||e===6?6:5:4}function XR(s,e){return S$(s)&&!S$(e)?!0:s===5?!1:y$(s)!==y$(e)}function y$(s){return s===6||s===5?"space":s}function S$(s){return s===4||s===5||s===6}function bJ(s,e,t,i,n){if(s.autoClosingOvertype==="never"||!s.autoClosingPairs.autoClosingPairsCloseSingleChar.has(n))return!1;for(let o=0,r=t.length;o<r;o++){const a=t[o];if(!a.isEmpty())return!1;const l=a.getPosition(),c=e.getLineContent(l.lineNumber);if(c.charAt(l.column-1)!==n)return!1;const h=ug(n);if((l.column>2?c.charCodeAt(l.column-2):0)===92&&h)return!1;if(s.autoClosingOvertype==="auto"){let g=!1;for(let f=0,p=i.length;f<p;f++){const _=i[f];if(l.lineNumber===_.startLineNumber&&l.column===_.startColumn){g=!0;break}}if(!g)return!1}}return!0}function tf(s,e,t){return t?new Lk(s,e,!0):new ts(s,e,!0)}function kW(s,e,t){return t=t||1,la.shiftIndent(e,e.length+t,s.tabSize,s.indentSize,s.insertSpaces)}function OE(s,e,t){return t=t||1,la.unshiftIndent(e,e.length+t,s.tabSize,s.indentSize,s.insertSpaces)}function wJ(s,e){return ug(e)?s.autoSurround==="quotes"||s.autoSurround==="languageDefined":s.autoSurround==="brackets"||s.autoSurround==="languageDefined"}class ou{static indent(e,t,i){if(t===null||i===null)return[];const n=[];for(let o=0,r=i.length;o<r;o++)n[o]=new la(i[o],{isUnshift:!1,tabSize:e.tabSize,indentSize:e.indentSize,insertSpaces:e.insertSpaces,useTabStops:e.useTabStops,autoIndent:e.autoIndent},e.languageConfigurationService);return n}static outdent(e,t,i){const n=[];for(let o=0,r=i.length;o<r;o++)n[o]=new la(i[o],{isUnshift:!0,tabSize:e.tabSize,indentSize:e.indentSize,insertSpaces:e.insertSpaces,useTabStops:e.useTabStops,autoIndent:e.autoIndent},e.languageConfigurationService);return n}static paste(e,t,i,n,o,r){return Sfe.getEdits(e,t,i,n,o,r)}static tab(e,t,i){return kfe.getCommands(e,t,i)}static compositionType(e,t,i,n,o,r,a,l){return xfe.getEdits(e,t,i,n,o,r,a,l)}static compositionEndWithInterceptors(e,t,i,n,o,r){if(!n)return null;let a=null;for(const u of n)if(a===null)a=u.insertedText;else if(a!==u.insertedText)return null;if(!a||a.length!==1)return C$.getEdits(t,n);const l=a;let c=!1;for(const u of n)if(u.deletedText.length!==0){c=!0;break}if(c){if(!wJ(t,l)||!t.surroundingPairs.hasOwnProperty(l))return null;const u=ug(l);for(const p of n)if(p.deletedSelectionStart!==0||p.deletedSelectionEnd!==p.deletedText.length||/^[ \t]+$/.test(p.deletedText)||u&&ug(p.deletedText))return null;const g=[];for(const p of o){if(!p.isEmpty())return null;g.push(p.getPosition())}if(g.length!==n.length)return null;const f=[];for(let p=0,_=g.length;p<_;p++)f.push(new ufe(g[p],n[p].deletedText,t.surroundingPairs[l]));return new to(4,f,{shouldPushStackElementBefore:!0,shouldPushStackElementAfter:!1})}const d=wfe.getEdits(t,i,o,r,l);if(d!==void 0)return d;const h=HP.getEdits(t,i,o,l,!0,!1);return h!==void 0?h:C$.getEdits(t,n)}static typeWithInterceptors(e,t,i,n,o,r,a){const l=ZR.getEdits(i,n,o,a,e);if(l!==void 0)return l;const c=_fe.getEdits(i,n,o,a,e);if(c!==void 0)return c;const d=bfe.getEdits(t,i,n,o,r,a);if(d!==void 0)return d;const h=HP.getEdits(i,n,o,a,!1,e);if(h!==void 0)return h;const u=vfe.getEdits(i,n,o,a,e);if(u!==void 0)return u;const g=Cfe.getEdits(t,i,n,o,a,e);return g!==void 0?g:yfe.getEdits(i,t,o,a,e)}static typeWithoutInterceptors(e,t,i,n,o){return Lfe.getEdits(e,n,o)}}class Dfe{constructor(e,t,i,n,o,r,a){this.deletedText=e,this.deletedSelectionStart=t,this.deletedSelectionEnd=i,this.insertedText=n,this.insertedSelectionStart=o,this.insertedSelectionEnd=r,this.insertedTextRange=a}}var N;(function(s){s.editorSimpleInput=new ce("editorSimpleInput",!1,!0),s.editorTextFocus=new ce("editorTextFocus",!1,m(681,"Whether the editor text has focus (cursor is blinking)")),s.focus=new ce("editorFocus",!1,m(682,"Whether the editor or an editor widget has focus (e.g. focus is in the find widget)")),s.textInputFocus=new ce("textInputFocus",!1,m(683,"Whether an editor or a rich text input has focus (cursor is blinking)")),s.readOnly=new ce("editorReadonly",!1,m(684,"Whether the editor is read-only")),s.inDiffEditor=new ce("inDiffEditor",!1,m(685,"Whether the context is a diff editor")),s.isEmbeddedDiffEditor=new ce("isEmbeddedDiffEditor",!1,m(686,"Whether the context is an embedded diff editor")),s.inMultiDiffEditor=new ce("inMultiDiffEditor",!1,m(687,"Whether the context is a multi diff editor")),s.multiDiffEditorAllCollapsed=new ce("multiDiffEditorAllCollapsed",void 0,m(688,"Whether all files in multi diff editor are collapsed")),s.hasChanges=new ce("diffEditorHasChanges",!1,m(689,"Whether the diff editor has changes")),s.comparingMovedCode=new ce("comparingMovedCode",!1,m(690,"Whether a moved code block is selected for comparison")),s.accessibleDiffViewerVisible=new ce("accessibleDiffViewerVisible",!1,m(691,"Whether the accessible diff viewer is visible")),s.diffEditorRenderSideBySideInlineBreakpointReached=new ce("diffEditorRenderSideBySideInlineBreakpointReached",!1,m(692,"Whether the diff editor render side by side inline breakpoint is reached")),s.diffEditorInlineMode=new ce("diffEditorInlineMode",!1,m(693,"Whether inline mode is active")),s.diffEditorOriginalWritable=new ce("diffEditorOriginalWritable",!1,m(694,"Whether modified is writable in the diff editor")),s.diffEditorModifiedWritable=new ce("diffEditorModifiedWritable",!1,m(695,"Whether modified is writable in the diff editor")),s.diffEditorOriginalUri=new ce("diffEditorOriginalUri","",m(696,"The uri of the original document")),s.diffEditorModifiedUri=new ce("diffEditorModifiedUri","",m(697,"The uri of the modified document")),s.columnSelection=new ce("editorColumnSelection",!1,m(698,"Whether `editor.columnSelection` is enabled")),s.writable=s.readOnly.toNegated(),s.hasNonEmptySelection=new ce("editorHasSelection",!1,m(699,"Whether the editor has text selected")),s.hasOnlyEmptySelection=s.hasNonEmptySelection.toNegated(),s.hasMultipleSelections=new ce("editorHasMultipleSelections",!1,m(700,"Whether the editor has multiple selections")),s.hasSingleSelection=s.hasMultipleSelections.toNegated(),s.tabMovesFocus=new ce("editorTabMovesFocus",!1,m(701,"Whether `Tab` will move focus out of the editor")),s.tabDoesNotMoveFocus=s.tabMovesFocus.toNegated(),s.isInEmbeddedEditor=new ce("isInEmbeddedEditor",!1,!0),s.canUndo=new ce("canUndo",!1,!0),s.canRedo=new ce("canRedo",!1,!0),s.hoverVisible=new ce("editorHoverVisible",!1,m(702,"Whether the editor hover is visible")),s.hoverFocused=new ce("editorHoverFocused",!1,m(703,"Whether the editor hover is focused")),s.stickyScrollFocused=new ce("stickyScrollFocused",!1,m(704,"Whether the sticky scroll is focused")),s.stickyScrollVisible=new ce("stickyScrollVisible",!1,m(705,"Whether the sticky scroll is visible")),s.standaloneColorPickerVisible=new ce("standaloneColorPickerVisible",!1,m(706,"Whether the standalone color picker is visible")),s.standaloneColorPickerFocused=new ce("standaloneColorPickerFocused",!1,m(707,"Whether the standalone color picker is focused")),s.inCompositeEditor=new ce("inCompositeEditor",void 0,m(708,"Whether the editor is part of a larger editor (e.g. notebooks)")),s.notInCompositeEditor=s.inCompositeEditor.toNegated(),s.languageId=new ce("editorLangId","",m(709,"The language identifier of the editor")),s.hasCompletionItemProvider=new ce("editorHasCompletionItemProvider",!1,m(710,"Whether the editor has a completion item provider")),s.hasCodeActionsProvider=new ce("editorHasCodeActionsProvider",!1,m(711,"Whether the editor has a code actions provider")),s.hasCodeLensProvider=new ce("editorHasCodeLensProvider",!1,m(712,"Whether the editor has a code lens provider")),s.hasDefinitionProvider=new ce("editorHasDefinitionProvider",!1,m(713,"Whether the editor has a definition provider")),s.hasDeclarationProvider=new ce("editorHasDeclarationProvider",!1,m(714,"Whether the editor has a declaration provider")),s.hasImplementationProvider=new ce("editorHasImplementationProvider",!1,m(715,"Whether the editor has an implementation provider")),s.hasTypeDefinitionProvider=new ce("editorHasTypeDefinitionProvider",!1,m(716,"Whether the editor has a type definition provider")),s.hasHoverProvider=new ce("editorHasHoverProvider",!1,m(717,"Whether the editor has a hover provider")),s.hasDocumentHighlightProvider=new ce("editorHasDocumentHighlightProvider",!1,m(718,"Whether the editor has a document highlight provider")),s.hasDocumentSymbolProvider=new ce("editorHasDocumentSymbolProvider",!1,m(719,"Whether the editor has a document symbol provider")),s.hasReferenceProvider=new ce("editorHasReferenceProvider",!1,m(720,"Whether the editor has a reference provider")),s.hasRenameProvider=new ce("editorHasRenameProvider",!1,m(721,"Whether the editor has a rename provider")),s.hasSignatureHelpProvider=new ce("editorHasSignatureHelpProvider",!1,m(722,"Whether the editor has a signature help provider")),s.hasInlayHintsProvider=new ce("editorHasInlayHintsProvider",!1,m(723,"Whether the editor has an inline hints provider")),s.hasDocumentFormattingProvider=new ce("editorHasDocumentFormattingProvider",!1,m(724,"Whether the editor has a document formatting provider")),s.hasDocumentSelectionFormattingProvider=new ce("editorHasDocumentSelectionFormattingProvider",!1,m(725,"Whether the editor has a document selection formatting provider")),s.hasMultipleDocumentFormattingProvider=new ce("editorHasMultipleDocumentFormattingProvider",!1,m(726,"Whether the editor has multiple document formatting providers")),s.hasMultipleDocumentSelectionFormattingProvider=new ce("editorHasMultipleDocumentSelectionFormattingProvider",!1,m(727,"Whether the editor has multiple document selection formatting providers"))})(N||(N={}));const _t=0;class Di extends hn{runEditorCommand(e,t,i){const n=t._getViewModel();n&&this.runCoreEditorCommand(n,i||{})}}var Un;(function(s){const e=function(i){if(!rn(i))return!1;const n=i;return!(!bn(n.to)||!fs(n.by)&&!bn(n.by)||!fs(n.value)&&!ch(n.value)||!fs(n.revealCursor)&&!iP(n.revealCursor))};s.metadata={description:"Scroll editor in the given direction",args:[{name:"Editor scroll argument object",description:"Property-value pairs that can be passed through this argument:\n					* 'to': A mandatory direction value.\n						```\n						'up', 'down'\n						```\n					* 'by': Unit to move. Default is computed based on 'to' value.\n						```\n						'line', 'wrappedLine', 'page', 'halfPage', 'editor'\n						```\n					* 'value': Number of units to move. Default is '1'.\n					* 'revealCursor': If 'true' reveals the cursor if it is outside view port.\n				",constraint:e,schema:{type:"object",required:["to"],properties:{to:{type:"string",enum:["up","down"]},by:{type:"string",enum:["line","wrappedLine","page","halfPage","editor"]},value:{type:"number",default:1},revealCursor:{type:"boolean"}}}}]},s.RawDirection={Up:"up",Right:"right",Down:"down",Left:"left"},s.RawUnit={Line:"line",WrappedLine:"wrappedLine",Page:"page",HalfPage:"halfPage",Editor:"editor",Column:"column"};function t(i){let n;switch(i.to){case s.RawDirection.Up:n=1;break;case s.RawDirection.Right:n=2;break;case s.RawDirection.Down:n=3;break;case s.RawDirection.Left:n=4;break;default:return null}let o;switch(i.by){case s.RawUnit.Line:o=1;break;case s.RawUnit.WrappedLine:o=2;break;case s.RawUnit.Page:o=3;break;case s.RawUnit.HalfPage:o=4;break;case s.RawUnit.Editor:o=5;break;case s.RawUnit.Column:o=6;break;default:o=2}const r=Math.floor(i.value||1),a=!!i.revealCursor;return{direction:n,unit:o,value:r,revealCursor:a,select:!!i.select}}s.parse=t})(Un||(Un={}));var Sb;(function(s){const e=function(t){if(!rn(t))return!1;const i=t;return!(!ch(i.lineNumber)&&!bn(i.lineNumber)||!fs(i.at)&&!bn(i.at))};s.metadata={description:"Reveal the given line at the given logical position",args:[{name:"Reveal line argument object",description:"Property-value pairs that can be passed through this argument:\n					* 'lineNumber': A mandatory line number value.\n					* 'at': Logical position at which line has to be revealed.\n						```\n						'top', 'center', 'bottom'\n						```\n				",constraint:e,schema:{type:"object",required:["lineNumber"],properties:{lineNumber:{type:["number","string"]},at:{type:"string",enum:["top","center","bottom"]}}}}]},s.RawAtArgument={Top:"top",Center:"center",Bottom:"bottom"}})(Sb||(Sb={}));class VP{constructor(e){e.addImplementation(1e4,"code-editor",(t,i)=>{const n=t.get(ht).getFocusedCodeEditor();return n&&n.hasTextFocus()?this._runEditorCommand(t,n,i):!1}),e.addImplementation(1e3,"generic-dom-input-textarea",(t,i)=>{const n=ln();return n&&wl(n)?(this.runDOMCommand(n),!0):!1}),e.addImplementation(0,"generic-dom",(t,i)=>{const n=t.get(ht).getActiveCodeEditor();return n?(n.focus(),this._runEditorCommand(t,n,i)):!1})}_runEditorCommand(e,t,i){const n=this.runEditorCommand(e,t,i);return n||!0}}var yn;(function(s){class e extends Di{constructor(w){super(w),this._inSelectionMode=w.inSelectionMode}runCoreEditorCommand(w,C){if(!C.position)return;w.model.pushStackElement(),w.setCursorStates(C.source,3,[Rn.moveTo(w,w.getPrimaryCursorState(),this._inSelectionMode,C.position,C.viewPosition)])&&C.revealType!==2&&w.revealAllCursors(C.source,!0,!0)}}s.MoveTo=le(new e({id:"_moveTo",inSelectionMode:!1,precondition:void 0})),s.MoveToSelect=le(new e({id:"_moveToSelect",inSelectionMode:!0,precondition:void 0}));class t extends Di{runCoreEditorCommand(w,C){w.model.pushStackElement();const v=this._getColumnSelectResult(w,w.getPrimaryCursorState(),w.getCursorColumnSelectData(),C);v!==null&&(w.setCursorStates(C.source,3,v.viewStates.map(y=>ut.fromViewState(y))),w.setCursorColumnSelectData({isReal:!0,fromViewLineNumber:v.fromLineNumber,fromViewVisualColumn:v.fromVisualColumn,toViewLineNumber:v.toLineNumber,toViewVisualColumn:v.toVisualColumn}),v.reversed?w.revealTopMostCursor(C.source):w.revealBottomMostCursor(C.source))}}s.ColumnSelect=le(new class extends t{constructor(){super({id:"columnSelect",precondition:void 0})}_getColumnSelectResult(b,w,C,v){if(typeof v.position>"u"||typeof v.viewPosition>"u"||typeof v.mouseColumn>"u")return null;const y=b.model.validatePosition(v.position),x=b.coordinatesConverter.validateViewPosition(new A(v.viewPosition.lineNumber,v.viewPosition.column),y),L=v.doColumnSelect?C.fromViewLineNumber:x.lineNumber,E=v.doColumnSelect?C.fromViewVisualColumn:v.mouseColumn-1;return ff.columnSelect(b.cursorConfig,b,L,E,x.lineNumber,v.mouseColumn-1)}}),s.CursorColumnSelectLeft=le(new class extends t{constructor(){super({id:"cursorColumnSelectLeft",precondition:void 0,kbOpts:{weight:_t,kbExpr:N.textInputFocus,primary:3599,linux:{primary:0}}})}_getColumnSelectResult(b,w,C,v){return ff.columnSelectLeft(b.cursorConfig,b,C)}}),s.CursorColumnSelectRight=le(new class extends t{constructor(){super({id:"cursorColumnSelectRight",precondition:void 0,kbOpts:{weight:_t,kbExpr:N.textInputFocus,primary:3601,linux:{primary:0}}})}_getColumnSelectResult(b,w,C,v){return ff.columnSelectRight(b.cursorConfig,b,C)}});class i extends t{constructor(w){super(w),this._isPaged=w.isPaged}_getColumnSelectResult(w,C,v,y){return ff.columnSelectUp(w.cursorConfig,w,v,this._isPaged)}}s.CursorColumnSelectUp=le(new i({isPaged:!1,id:"cursorColumnSelectUp",precondition:void 0,kbOpts:{weight:_t,kbExpr:N.textInputFocus,primary:3600,linux:{primary:0}}})),s.CursorColumnSelectPageUp=le(new i({isPaged:!0,id:"cursorColumnSelectPageUp",precondition:void 0,kbOpts:{weight:_t,kbExpr:N.textInputFocus,primary:3595,linux:{primary:0}}}));class n extends t{constructor(w){super(w),this._isPaged=w.isPaged}_getColumnSelectResult(w,C,v,y){return ff.columnSelectDown(w.cursorConfig,w,v,this._isPaged)}}s.CursorColumnSelectDown=le(new n({isPaged:!1,id:"cursorColumnSelectDown",precondition:void 0,kbOpts:{weight:_t,kbExpr:N.textInputFocus,primary:3602,linux:{primary:0}}})),s.CursorColumnSelectPageDown=le(new n({isPaged:!0,id:"cursorColumnSelectPageDown",precondition:void 0,kbOpts:{weight:_t,kbExpr:N.textInputFocus,primary:3596,linux:{primary:0}}}));class o extends Di{constructor(){super({id:"cursorMove",precondition:void 0,metadata:NE.metadata})}runCoreEditorCommand(w,C){const v=NE.parse(C);v&&this._runCursorMove(w,C.source,v)}_runCursorMove(w,C,v){const y=v.noHistory?"api":C;w.model.pushStackElement(),w.setCursorStates(y,3,o._move(w,w.getCursorStates(),v)),w.revealAllCursors(y,!0)}static _move(w,C,v){const y=v.select,x=v.value;switch(v.direction){case 0:case 1:case 2:case 3:case 4:case 5:case 6:case 7:case 8:case 9:case 10:return Rn.simpleMove(w,C,v.direction,y,x,v.unit);case 11:case 13:case 12:case 14:return Rn.viewportMove(w,C,v.direction,y,x);default:return null}}}s.CursorMoveImpl=o,s.CursorMove=le(new o);class r extends Di{constructor(w){super(w),this._staticArgs=w.args}runCoreEditorCommand(w,C){let v=this._staticArgs;this._staticArgs.value===-1&&(v={direction:this._staticArgs.direction,unit:this._staticArgs.unit,select:this._staticArgs.select,value:C.pageSize||w.cursorConfig.pageSize}),w.model.pushStackElement(),w.setCursorStates(C.source,3,Rn.simpleMove(w,w.getCursorStates(),v.direction,v.select,v.value,v.unit)),w.revealAllCursors(C.source,!0)}}s.CursorLeft=le(new r({args:{direction:0,unit:0,select:!1,value:1},id:"cursorLeft",precondition:void 0,kbOpts:{weight:_t,kbExpr:N.textInputFocus,primary:15,mac:{primary:15,secondary:[288]}}})),s.CursorLeftSelect=le(new r({args:{direction:0,unit:0,select:!0,value:1},id:"cursorLeftSelect",precondition:void 0,kbOpts:{weight:_t,kbExpr:N.textInputFocus,primary:1039}})),s.CursorRight=le(new r({args:{direction:1,unit:0,select:!1,value:1},id:"cursorRight",precondition:void 0,kbOpts:{weight:_t,kbExpr:N.textInputFocus,primary:17,mac:{primary:17,secondary:[292]}}})),s.CursorRightSelect=le(new r({args:{direction:1,unit:0,select:!0,value:1},id:"cursorRightSelect",precondition:void 0,kbOpts:{weight:_t,kbExpr:N.textInputFocus,primary:1041}})),s.CursorUp=le(new r({args:{direction:2,unit:2,select:!1,value:1},id:"cursorUp",precondition:void 0,kbOpts:{weight:_t,kbExpr:N.textInputFocus,primary:16,mac:{primary:16,secondary:[302]}}})),s.CursorUpSelect=le(new r({args:{direction:2,unit:2,select:!0,value:1},id:"cursorUpSelect",precondition:void 0,kbOpts:{weight:_t,kbExpr:N.textInputFocus,primary:1040,secondary:[3088],mac:{primary:1040},linux:{primary:1040}}})),s.CursorPageUp=le(new r({args:{direction:2,unit:2,select:!1,value:-1},id:"cursorPageUp",precondition:void 0,kbOpts:{weight:_t,kbExpr:N.textInputFocus,primary:11}})),s.CursorPageUpSelect=le(new r({args:{direction:2,unit:2,select:!0,value:-1},id:"cursorPageUpSelect",precondition:void 0,kbOpts:{weight:_t,kbExpr:N.textInputFocus,primary:1035}})),s.CursorDown=le(new r({args:{direction:3,unit:2,select:!1,value:1},id:"cursorDown",precondition:void 0,kbOpts:{weight:_t,kbExpr:N.textInputFocus,primary:18,mac:{primary:18,secondary:[300]}}})),s.CursorDownSelect=le(new r({args:{direction:3,unit:2,select:!0,value:1},id:"cursorDownSelect",precondition:void 0,kbOpts:{weight:_t,kbExpr:N.textInputFocus,primary:1042,secondary:[3090],mac:{primary:1042},linux:{primary:1042}}})),s.CursorPageDown=le(new r({args:{direction:3,unit:2,select:!1,value:-1},id:"cursorPageDown",precondition:void 0,kbOpts:{weight:_t,kbExpr:N.textInputFocus,primary:12}})),s.CursorPageDownSelect=le(new r({args:{direction:3,unit:2,select:!0,value:-1},id:"cursorPageDownSelect",precondition:void 0,kbOpts:{weight:_t,kbExpr:N.textInputFocus,primary:1036}})),s.CreateCursor=le(new class extends Di{constructor(){super({id:"createCursor",precondition:void 0})}runCoreEditorCommand(b,w){if(!w.position)return;let C;w.wholeLine?C=Rn.line(b,b.getPrimaryCursorState(),!1,w.position,w.viewPosition):C=Rn.moveTo(b,b.getPrimaryCursorState(),!1,w.position,w.viewPosition);const v=b.getCursorStates();if(v.length>1){const y=C.modelState?C.modelState.position:null,x=C.viewState?C.viewState.position:null;for(let L=0,E=v.length;L<E;L++){const I=v[L];if(!(y&&!I.modelState.selection.containsPosition(y))&&!(x&&!I.viewState.selection.containsPosition(x))){v.splice(L,1),b.model.pushStackElement(),b.setCursorStates(w.source,3,v);return}}}v.push(C),b.model.pushStackElement(),b.setCursorStates(w.source,3,v)}}),s.LastCursorMoveToSelect=le(new class extends Di{constructor(){super({id:"_lastCursorMoveToSelect",precondition:void 0})}runCoreEditorCommand(b,w){if(!w.position)return;const C=b.getLastAddedCursorIndex(),v=b.getCursorStates(),y=v.slice(0);y[C]=Rn.moveTo(b,v[C],!0,w.position,w.viewPosition),b.model.pushStackElement(),b.setCursorStates(w.source,3,y)}});class a extends Di{constructor(w){super(w),this._inSelectionMode=w.inSelectionMode}runCoreEditorCommand(w,C){w.model.pushStackElement(),w.setCursorStates(C.source,3,Rn.moveToBeginningOfLine(w,w.getCursorStates(),this._inSelectionMode)),w.revealAllCursors(C.source,!0)}}s.CursorHome=le(new a({inSelectionMode:!1,id:"cursorHome",precondition:void 0,kbOpts:{weight:_t,kbExpr:N.textInputFocus,primary:14,mac:{primary:14,secondary:[2063]}}})),s.CursorHomeSelect=le(new a({inSelectionMode:!0,id:"cursorHomeSelect",precondition:void 0,kbOpts:{weight:_t,kbExpr:N.textInputFocus,primary:1038,mac:{primary:1038,secondary:[3087]}}}));class l extends Di{constructor(w){super(w),this._inSelectionMode=w.inSelectionMode}runCoreEditorCommand(w,C){w.model.pushStackElement(),w.setCursorStates(C.source,3,this._exec(w.getCursorStates())),w.revealAllCursors(C.source,!0)}_exec(w){const C=[];for(let v=0,y=w.length;v<y;v++){const x=w[v],L=x.modelState.position.lineNumber;C[v]=ut.fromModelState(x.modelState.move(this._inSelectionMode,L,1,0))}return C}}s.CursorLineStart=le(new l({inSelectionMode:!1,id:"cursorLineStart",precondition:void 0,kbOpts:{weight:_t,kbExpr:N.textInputFocus,primary:0,mac:{primary:287}}})),s.CursorLineStartSelect=le(new l({inSelectionMode:!0,id:"cursorLineStartSelect",precondition:void 0,kbOpts:{weight:_t,kbExpr:N.textInputFocus,primary:0,mac:{primary:1311}}}));class c extends Di{constructor(w){super(w),this._inSelectionMode=w.inSelectionMode}runCoreEditorCommand(w,C){w.model.pushStackElement(),w.setCursorStates(C.source,3,Rn.moveToEndOfLine(w,w.getCursorStates(),this._inSelectionMode,C.sticky||!1)),w.revealAllCursors(C.source,!0)}}s.CursorEnd=le(new c({inSelectionMode:!1,id:"cursorEnd",precondition:void 0,kbOpts:{args:{sticky:!1},weight:_t,kbExpr:N.textInputFocus,primary:13,mac:{primary:13,secondary:[2065]}},metadata:{description:"Go to End",args:[{name:"args",schema:{type:"object",properties:{sticky:{description:m(66,"Stick to the end even when going to longer lines"),type:"boolean",default:!1}}}}]}})),s.CursorEndSelect=le(new c({inSelectionMode:!0,id:"cursorEndSelect",precondition:void 0,kbOpts:{args:{sticky:!1},weight:_t,kbExpr:N.textInputFocus,primary:1037,mac:{primary:1037,secondary:[3089]}},metadata:{description:"Select to End",args:[{name:"args",schema:{type:"object",properties:{sticky:{description:m(67,"Stick to the end even when going to longer lines"),type:"boolean",default:!1}}}}]}}));class d extends Di{constructor(w){super(w),this._inSelectionMode=w.inSelectionMode}runCoreEditorCommand(w,C){w.model.pushStackElement(),w.setCursorStates(C.source,3,this._exec(w,w.getCursorStates())),w.revealAllCursors(C.source,!0)}_exec(w,C){const v=[];for(let y=0,x=C.length;y<x;y++){const L=C[y],E=L.modelState.position.lineNumber,I=w.model.getLineMaxColumn(E);v[y]=ut.fromModelState(L.modelState.move(this._inSelectionMode,E,I,0))}return v}}s.CursorLineEnd=le(new d({inSelectionMode:!1,id:"cursorLineEnd",precondition:void 0,kbOpts:{weight:_t,kbExpr:N.textInputFocus,primary:0,mac:{primary:291}}})),s.CursorLineEndSelect=le(new d({inSelectionMode:!0,id:"cursorLineEndSelect",precondition:void 0,kbOpts:{weight:_t,kbExpr:N.textInputFocus,primary:0,mac:{primary:1315}}}));class h extends Di{constructor(w){super(w),this._inSelectionMode=w.inSelectionMode}runCoreEditorCommand(w,C){w.model.pushStackElement(),w.setCursorStates(C.source,3,Rn.moveToBeginningOfBuffer(w,w.getCursorStates(),this._inSelectionMode)),w.revealAllCursors(C.source,!0)}}s.CursorTop=le(new h({inSelectionMode:!1,id:"cursorTop",precondition:void 0,kbOpts:{weight:_t,kbExpr:N.textInputFocus,primary:2062,mac:{primary:2064}}})),s.CursorTopSelect=le(new h({inSelectionMode:!0,id:"cursorTopSelect",precondition:void 0,kbOpts:{weight:_t,kbExpr:N.textInputFocus,primary:3086,mac:{primary:3088}}}));class u extends Di{constructor(w){super(w),this._inSelectionMode=w.inSelectionMode}runCoreEditorCommand(w,C){w.model.pushStackElement(),w.setCursorStates(C.source,3,Rn.moveToEndOfBuffer(w,w.getCursorStates(),this._inSelectionMode)),w.revealAllCursors(C.source,!0)}}s.CursorBottom=le(new u({inSelectionMode:!1,id:"cursorBottom",precondition:void 0,kbOpts:{weight:_t,kbExpr:N.textInputFocus,primary:2061,mac:{primary:2066}}})),s.CursorBottomSelect=le(new u({inSelectionMode:!0,id:"cursorBottomSelect",precondition:void 0,kbOpts:{weight:_t,kbExpr:N.textInputFocus,primary:3085,mac:{primary:3090}}}));class g extends Di{constructor(){super({id:"editorScroll",precondition:void 0,metadata:Un.metadata})}determineScrollMethod(w){const C=[6],v=[1,2,3,4,5,6],y=[4,2],x=[1,3];return C.includes(w.unit)&&y.includes(w.direction)?this._runHorizontalEditorScroll.bind(this):v.includes(w.unit)&&x.includes(w.direction)?this._runVerticalEditorScroll.bind(this):null}runCoreEditorCommand(w,C){const v=Un.parse(C);if(!v)return;const y=this.determineScrollMethod(v);y&&y(w,C.source,v)}_runVerticalEditorScroll(w,C,v){const y=this._computeDesiredScrollTop(w,v);if(v.revealCursor){const x=w.getCompletelyVisibleViewRangeAtScrollTop(y);w.setCursorStates(C,3,[Rn.findPositionInViewportIfOutside(w,w.getPrimaryCursorState(),x,v.select)])}w.viewLayout.setScrollPosition({scrollTop:y},0)}_computeDesiredScrollTop(w,C){if(C.unit===1){const x=w.viewLayout.getFutureViewport(),L=w.getCompletelyVisibleViewRangeAtScrollTop(x.top),E=w.coordinatesConverter.convertViewRangeToModelRange(L);let I;C.direction===1?I=Math.max(1,E.startLineNumber-C.value):I=Math.min(w.model.getLineCount(),E.startLineNumber+C.value);const R=w.coordinatesConverter.convertModelPositionToViewPosition(new A(I,1));return w.viewLayout.getVerticalOffsetForLineNumber(R.lineNumber)}if(C.unit===5){let x=0;return C.direction===3&&(x=w.model.getLineCount()-w.cursorConfig.pageSize),w.viewLayout.getVerticalOffsetForLineNumber(x)}let v;C.unit===3?v=w.cursorConfig.pageSize*C.value:C.unit===4?v=Math.round(w.cursorConfig.pageSize/2)*C.value:v=C.value;const y=(C.direction===1?-1:1)*v;return w.viewLayout.getCurrentScrollTop()+y*w.cursorConfig.lineHeight}_runHorizontalEditorScroll(w,C,v){const y=this._computeDesiredScrollLeft(w,v);w.viewLayout.setScrollPosition({scrollLeft:y},0)}_computeDesiredScrollLeft(w,C){const v=(C.direction===4?-1:1)*C.value;return w.viewLayout.getCurrentScrollLeft()+v*w.cursorConfig.typicalHalfwidthCharacterWidth}}s.EditorScrollImpl=g,s.EditorScroll=le(new g),s.ScrollLineUp=le(new class extends Di{constructor(){super({id:"scrollLineUp",precondition:void 0,kbOpts:{weight:_t,kbExpr:N.textInputFocus,primary:2064,mac:{primary:267}}})}runCoreEditorCommand(b,w){s.EditorScroll.runCoreEditorCommand(b,{to:Un.RawDirection.Up,by:Un.RawUnit.WrappedLine,value:1,revealCursor:!1,select:!1,source:w.source})}}),s.ScrollPageUp=le(new class extends Di{constructor(){super({id:"scrollPageUp",precondition:void 0,kbOpts:{weight:_t,kbExpr:N.textInputFocus,primary:2059,win:{primary:523},linux:{primary:523}}})}runCoreEditorCommand(b,w){s.EditorScroll.runCoreEditorCommand(b,{to:Un.RawDirection.Up,by:Un.RawUnit.Page,value:1,revealCursor:!1,select:!1,source:w.source})}}),s.ScrollEditorTop=le(new class extends Di{constructor(){super({id:"scrollEditorTop",precondition:void 0,kbOpts:{weight:_t,kbExpr:N.textInputFocus}})}runCoreEditorCommand(b,w){s.EditorScroll.runCoreEditorCommand(b,{to:Un.RawDirection.Up,by:Un.RawUnit.Editor,value:1,revealCursor:!1,select:!1,source:w.source})}}),s.ScrollLineDown=le(new class extends Di{constructor(){super({id:"scrollLineDown",precondition:void 0,kbOpts:{weight:_t,kbExpr:N.textInputFocus,primary:2066,mac:{primary:268}}})}runCoreEditorCommand(b,w){s.EditorScroll.runCoreEditorCommand(b,{to:Un.RawDirection.Down,by:Un.RawUnit.WrappedLine,value:1,revealCursor:!1,select:!1,source:w.source})}}),s.ScrollPageDown=le(new class extends Di{constructor(){super({id:"scrollPageDown",precondition:void 0,kbOpts:{weight:_t,kbExpr:N.textInputFocus,primary:2060,win:{primary:524},linux:{primary:524}}})}runCoreEditorCommand(b,w){s.EditorScroll.runCoreEditorCommand(b,{to:Un.RawDirection.Down,by:Un.RawUnit.Page,value:1,revealCursor:!1,select:!1,source:w.source})}}),s.ScrollEditorBottom=le(new class extends Di{constructor(){super({id:"scrollEditorBottom",precondition:void 0,kbOpts:{weight:_t,kbExpr:N.textInputFocus}})}runCoreEditorCommand(b,w){s.EditorScroll.runCoreEditorCommand(b,{to:Un.RawDirection.Down,by:Un.RawUnit.Editor,value:1,revealCursor:!1,select:!1,source:w.source})}}),s.ScrollLeft=le(new class extends Di{constructor(){super({id:"scrollLeft",precondition:void 0,kbOpts:{weight:_t,kbExpr:N.textInputFocus}})}runCoreEditorCommand(b,w){s.EditorScroll.runCoreEditorCommand(b,{to:Un.RawDirection.Left,by:Un.RawUnit.Column,value:2,revealCursor:!1,select:!1,source:w.source})}}),s.ScrollRight=le(new class extends Di{constructor(){super({id:"scrollRight",precondition:void 0,kbOpts:{weight:_t,kbExpr:N.textInputFocus}})}runCoreEditorCommand(b,w){s.EditorScroll.runCoreEditorCommand(b,{to:Un.RawDirection.Right,by:Un.RawUnit.Column,value:2,revealCursor:!1,select:!1,source:w.source})}});class f extends Di{constructor(w){super(w),this._inSelectionMode=w.inSelectionMode}runCoreEditorCommand(w,C){C.position&&(w.model.pushStackElement(),w.setCursorStates(C.source,3,[Rn.word(w,w.getPrimaryCursorState(),this._inSelectionMode,C.position)]),C.revealType!==2&&w.revealAllCursors(C.source,!0,!0))}}s.WordSelect=le(new f({inSelectionMode:!1,id:"_wordSelect",precondition:void 0})),s.WordSelectDrag=le(new f({inSelectionMode:!0,id:"_wordSelectDrag",precondition:void 0})),s.LastCursorWordSelect=le(new class extends Di{constructor(){super({id:"lastCursorWordSelect",precondition:void 0})}runCoreEditorCommand(b,w){if(!w.position)return;const C=b.getLastAddedCursorIndex(),v=b.getCursorStates(),y=v.slice(0),x=v[C];y[C]=Rn.word(b,x,x.modelState.hasSelection(),w.position),b.model.pushStackElement(),b.setCursorStates(w.source,3,y)}});class p extends Di{constructor(w){super(w),this._inSelectionMode=w.inSelectionMode}runCoreEditorCommand(w,C){C.position&&(w.model.pushStackElement(),w.setCursorStates(C.source,3,[Rn.line(w,w.getPrimaryCursorState(),this._inSelectionMode,C.position,C.viewPosition)]),C.revealType!==2&&w.revealAllCursors(C.source,!1,!0))}}s.LineSelect=le(new p({inSelectionMode:!1,id:"_lineSelect",precondition:void 0})),s.LineSelectDrag=le(new p({inSelectionMode:!0,id:"_lineSelectDrag",precondition:void 0}));class _ extends Di{constructor(w){super(w),this._inSelectionMode=w.inSelectionMode}runCoreEditorCommand(w,C){if(!C.position)return;const v=w.getLastAddedCursorIndex(),y=w.getCursorStates(),x=y.slice(0);x[v]=Rn.line(w,y[v],this._inSelectionMode,C.position,C.viewPosition),w.model.pushStackElement(),w.setCursorStates(C.source,3,x)}}s.LastCursorLineSelect=le(new _({inSelectionMode:!1,id:"lastCursorLineSelect",precondition:void 0})),s.LastCursorLineSelectDrag=le(new _({inSelectionMode:!0,id:"lastCursorLineSelectDrag",precondition:void 0})),s.CancelSelection=le(new class extends Di{constructor(){super({id:"cancelSelection",precondition:N.hasNonEmptySelection,kbOpts:{weight:_t,kbExpr:N.textInputFocus,primary:9,secondary:[1033]}})}runCoreEditorCommand(b,w){b.model.pushStackElement(),b.setCursorStates(w.source,3,[Rn.cancelSelection(b,b.getPrimaryCursorState())]),b.revealAllCursors(w.source,!0)}}),s.RemoveSecondaryCursors=le(new class extends Di{constructor(){super({id:"removeSecondaryCursors",precondition:N.hasMultipleSelections,kbOpts:{weight:_t+1,kbExpr:N.textInputFocus,primary:9,secondary:[1033]}})}runCoreEditorCommand(b,w){b.model.pushStackElement(),b.setCursorStates(w.source,3,[b.getPrimaryCursorState()]),b.revealAllCursors(w.source,!0),xl(m(68,"Removed secondary cursors"))}}),s.RevealLine=le(new class extends Di{constructor(){super({id:"revealLine",precondition:void 0,metadata:Sb.metadata})}runCoreEditorCommand(b,w){const C=w,v=C.lineNumber||0;let y=typeof v=="number"?v+1:parseInt(v)+1;y<1&&(y=1);const x=b.model.getLineCount();y>x&&(y=x);const L=new k(y,1,y,b.model.getLineMaxColumn(y));let E=0;if(C.at)switch(C.at){case Sb.RawAtArgument.Top:E=3;break;case Sb.RawAtArgument.Center:E=1;break;case Sb.RawAtArgument.Bottom:E=4;break}const I=b.coordinatesConverter.convertModelRangeToViewRange(L);b.revealRange(w.source,!1,I,E,0)}}),s.SelectAll=new class extends VP{constructor(){super(cge)}runDOMCommand(b){wo&&(b.focus(),b.select()),b.ownerDocument.execCommand("selectAll")}runEditorCommand(b,w,C){const v=w._getViewModel();v&&this.runCoreEditorCommand(v,C)}runCoreEditorCommand(b,w){b.model.pushStackElement(),b.setCursorStates("keyboard",3,[Rn.selectAll(b,b.getPrimaryCursorState())])}},s.SetSelection=le(new class extends Di{constructor(){super({id:"setSelection",precondition:void 0})}runCoreEditorCommand(b,w){w.selection&&(b.model.pushStackElement(),b.setCursorStates(w.source,3,[ut.fromModelSelection(w.selection)]))}})})(yn||(yn={}));const Tfe=K.and(N.textInputFocus,N.columnSelection);function Xv(s,e){Dn.registerKeybindingRule({id:s,primary:e,when:Tfe,weight:_t+1})}Xv(yn.CursorColumnSelectLeft.id,1039);Xv(yn.CursorColumnSelectRight.id,1041);Xv(yn.CursorColumnSelectUp.id,1040);Xv(yn.CursorColumnSelectPageUp.id,1035);Xv(yn.CursorColumnSelectDown.id,1042);Xv(yn.CursorColumnSelectPageDown.id,1036);function x$(s){return s.register(),s}var hw;(function(s){class e extends hn{runEditorCommand(i,n,o){const r=n._getViewModel();r&&this.runCoreEditingCommand(n,r,o||{})}}s.CoreEditingCommand=e,s.LineBreakInsert=le(new class extends e{constructor(){super({id:"lineBreakInsert",precondition:N.writable,kbOpts:{weight:_t,kbExpr:N.textInputFocus,primary:0,mac:{primary:301}}})}runCoreEditingCommand(t,i,n){t.pushUndoStop(),t.executeCommands(this.id,ZR.lineBreakInsert(i.cursorConfig,i.model,i.getCursorStates().map(o=>o.modelState.selection)))}}),s.Outdent=le(new class extends e{constructor(){super({id:"outdent",precondition:N.writable,kbOpts:{weight:_t,kbExpr:K.and(N.editorTextFocus,N.tabDoesNotMoveFocus),primary:1026}})}runCoreEditingCommand(t,i,n){t.pushUndoStop(),t.executeCommands(this.id,ou.outdent(i.cursorConfig,i.model,i.getCursorStates().map(o=>o.modelState.selection))),t.pushUndoStop()}}),s.Tab=le(new class extends e{constructor(){super({id:"tab",precondition:N.writable,kbOpts:{weight:_t,kbExpr:K.and(N.editorTextFocus,N.tabDoesNotMoveFocus),primary:2}})}runCoreEditingCommand(t,i,n){t.pushUndoStop(),t.executeCommands(this.id,ou.tab(i.cursorConfig,i.model,i.getCursorStates().map(o=>o.modelState.selection))),t.pushUndoStop()}}),s.DeleteLeft=le(new class extends e{constructor(){super({id:"deleteLeft",precondition:void 0,kbOpts:{weight:_t,kbExpr:N.textInputFocus,primary:1,secondary:[1025],mac:{primary:1,secondary:[1025,294,257]}}})}runCoreEditingCommand(t,i,n){const[o,r]=km.deleteLeft(i.getPrevEditOperationType(),i.cursorConfig,i.model,i.getCursorStates().map(a=>a.modelState.selection),i.getCursorAutoClosedCharacters());o&&t.pushUndoStop(),t.executeCommands(this.id,r),i.setPrevEditOperationType(2)}}),s.DeleteRight=le(new class extends e{constructor(){super({id:"deleteRight",precondition:void 0,kbOpts:{weight:_t,kbExpr:N.textInputFocus,primary:20,mac:{primary:20,secondary:[290,276]}}})}runCoreEditingCommand(t,i,n){const[o,r]=km.deleteRight(i.getPrevEditOperationType(),i.cursorConfig,i.model,i.getCursorStates().map(a=>a.modelState.selection));o&&t.pushUndoStop(),t.executeCommands(this.id,r),i.setPrevEditOperationType(3)}}),s.Undo=new class extends VP{constructor(){super(XQ)}runDOMCommand(t){t.ownerDocument.execCommand("undo")}runEditorCommand(t,i,n){if(!(!i.hasModel()||i.getOption(104)===!0))return i.getModel().undo()}},s.Redo=new class extends VP{constructor(){super(QQ)}runDOMCommand(t){t.ownerDocument.execCommand("redo")}runEditorCommand(t,i,n){if(!(!i.hasModel()||i.getOption(104)===!0))return i.getModel().redo()}}})(hw||(hw={}));class L$ extends $R{constructor(e,t,i){super({id:e,precondition:void 0,metadata:i}),this._handlerId=t}runCommand(e,t){const i=e.get(ht).getFocusedCodeEditor();i&&i.trigger("keyboard",this._handlerId,t)}}function l_(s,e){x$(new L$("default:"+s,s)),x$(new L$(s,s,e))}l_("type",{description:"Type",args:[{name:"args",schema:{type:"object",required:["text"],properties:{text:{type:"string"}}}}]});l_("replacePreviousChar");l_("compositionType");l_("compositionStart");l_("compositionEnd");l_("paste");l_("cut");const EW=qe("markerDecorationsService");var Nfe=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},Rfe=function(s,e){return function(t,i){e(t,i,s)}},Hw;let PE=(Hw=class{constructor(e,t){}dispose(){}},Hw.ID="editor.contrib.markerDecorations",Hw);PE=Nfe([Rfe(1,EW)],PE);lt(PE.ID,PE,0);class FE{constructor(e){this.domNode=e,this._maxWidth="",this._width="",this._height="",this._top="",this._left="",this._bottom="",this._right="",this._paddingLeft="",this._fontFamily="",this._fontWeight="",this._fontSize="",this._fontStyle="",this._fontFeatureSettings="",this._fontVariationSettings="",this._textDecoration="",this._lineHeight="",this._letterSpacing="",this._className="",this._display="",this._position="",this._visibility="",this._color="",this._backgroundColor="",this._layerHint=!1,this._contain="none",this._boxShadow=""}focus(){this.domNode.focus()}setMaxWidth(e){const t=ja(e);this._maxWidth!==t&&(this._maxWidth=t,this.domNode.style.maxWidth=this._maxWidth)}setWidth(e){const t=ja(e);this._width!==t&&(this._width=t,this.domNode.style.width=this._width)}setHeight(e){const t=ja(e);this._height!==t&&(this._height=t,this.domNode.style.height=this._height)}setTop(e){const t=ja(e);this._top!==t&&(this._top=t,this.domNode.style.top=this._top)}setLeft(e){const t=ja(e);this._left!==t&&(this._left=t,this.domNode.style.left=this._left)}setBottom(e){const t=ja(e);this._bottom!==t&&(this._bottom=t,this.domNode.style.bottom=this._bottom)}setRight(e){const t=ja(e);this._right!==t&&(this._right=t,this.domNode.style.right=this._right)}setPaddingLeft(e){const t=ja(e);this._paddingLeft!==t&&(this._paddingLeft=t,this.domNode.style.paddingLeft=this._paddingLeft)}setFontFamily(e){this._fontFamily!==e&&(this._fontFamily=e,this.domNode.style.fontFamily=this._fontFamily)}setFontWeight(e){this._fontWeight!==e&&(this._fontWeight=e,this.domNode.style.fontWeight=this._fontWeight)}setFontSize(e){const t=ja(e);this._fontSize!==t&&(this._fontSize=t,this.domNode.style.fontSize=this._fontSize)}setFontStyle(e){this._fontStyle!==e&&(this._fontStyle=e,this.domNode.style.fontStyle=this._fontStyle)}setFontFeatureSettings(e){this._fontFeatureSettings!==e&&(this._fontFeatureSettings=e,this.domNode.style.fontFeatureSettings=this._fontFeatureSettings)}setFontVariationSettings(e){this._fontVariationSettings!==e&&(this._fontVariationSettings=e,this.domNode.style.fontVariationSettings=this._fontVariationSettings)}setTextDecoration(e){this._textDecoration!==e&&(this._textDecoration=e,this.domNode.style.textDecoration=this._textDecoration)}setLineHeight(e){const t=ja(e);this._lineHeight!==t&&(this._lineHeight=t,this.domNode.style.lineHeight=this._lineHeight)}setLetterSpacing(e){const t=ja(e);this._letterSpacing!==t&&(this._letterSpacing=t,this.domNode.style.letterSpacing=this._letterSpacing)}setClassName(e){this._className!==e&&(this._className=e,this.domNode.className=this._className)}toggleClassName(e,t){this.domNode.classList.toggle(e,t),this._className=this.domNode.className}setDisplay(e){this._display!==e&&(this._display=e,this.domNode.style.display=this._display)}setPosition(e){this._position!==e&&(this._position=e,this.domNode.style.position=this._position)}setVisibility(e){this._visibility!==e&&(this._visibility=e,this.domNode.style.visibility=this._visibility)}setColor(e){this._color!==e&&(this._color=e,this.domNode.style.color=this._color)}setBackgroundColor(e){this._backgroundColor!==e&&(this._backgroundColor=e,this.domNode.style.backgroundColor=this._backgroundColor)}setLayerHinting(e){this._layerHint!==e&&(this._layerHint=e,this.domNode.style.transform=this._layerHint?"translate3d(0px, 0px, 0px)":"")}setBoxShadow(e){this._boxShadow!==e&&(this._boxShadow=e,this.domNode.style.boxShadow=e)}setContain(e){this._contain!==e&&(this._contain=e,this.domNode.style.contain=this._contain)}setAttribute(e,t){this.domNode.setAttribute(e,t)}removeAttribute(e){this.domNode.removeAttribute(e)}appendChild(e){this.domNode.appendChild(e.domNode)}removeChild(e){this.domNode.removeChild(e.domNode)}}function ja(s){return typeof s=="number"?`${s}px`:s}function wt(s){return new FE(s)}function In(s,e){s instanceof FE?(s.setFontFamily(e.getMassagedFontFamily()),s.setFontWeight(e.fontWeight),s.setFontSize(e.fontSize),s.setFontFeatureSettings(e.fontFeatureSettings),s.setFontVariationSettings(e.fontVariationSettings),s.setLineHeight(e.lineHeight),s.setLetterSpacing(e.letterSpacing)):(s.style.fontFamily=e.getMassagedFontFamily(),s.style.fontWeight=e.fontWeight,s.style.fontSize=e.fontSize+"px",s.style.fontFeatureSettings=e.fontFeatureSettings,s.style.fontVariationSettings=e.fontVariationSettings,s.style.lineHeight=e.lineHeight+"px",s.style.letterSpacing=e.letterSpacing+"px")}class vJ extends F{constructor(e,t){super(),this._onDidChange=this._register(new M),this.onDidChange=this._onDidChange.event,this._referenceDomElement=e,this._width=-1,this._height=-1,this._resizeObserver=null,this.measureReferenceDomElement(!1,t)}dispose(){this.stopObserving(),super.dispose()}getWidth(){return this._width}getHeight(){return this._height}startObserving(){if(!this._resizeObserver&&this._referenceDomElement){let e=null;const t=()=>{e?this.observe({width:e.width,height:e.height}):this.observe()};let i=!1,n=!1;const o=()=>{if(i&&!n)try{i=!1,n=!0,t()}finally{Co(be(this._referenceDomElement),()=>{n=!1,o()})}};this._resizeObserver=new ResizeObserver(r=>{r&&r[0]&&r[0].contentRect?e={width:r[0].contentRect.width,height:r[0].contentRect.height}:e=null,i=!0,o()}),this._resizeObserver.observe(this._referenceDomElement)}}stopObserving(){this._resizeObserver&&(this._resizeObserver.disconnect(),this._resizeObserver=null)}observe(e){this.measureReferenceDomElement(!0,e)}measureReferenceDomElement(e,t){let i=0,n=0;t?(i=t.width,n=t.height):this._referenceDomElement&&(i=this._referenceDomElement.clientWidth,n=this._referenceDomElement.clientHeight),i=Math.max(5,i),n=Math.max(5,n),(this._width!==i||this._height!==n)&&(this._width=i,this._height=n,e&&this._onDidChange.fire())}}class Afe extends F{constructor(e){super(),this._onDidChange=this._register(new M),this.onDidChange=this._onDidChange.event,this._listener=()=>this._handleChange(e,!0),this._mediaQueryList=null,this._handleChange(e,!1)}_handleChange(e,t){this._mediaQueryList?.removeEventListener("change",this._listener),this._mediaQueryList=e.matchMedia(`(resolution: ${e.devicePixelRatio}dppx)`),this._mediaQueryList.addEventListener("change",this._listener),t&&this._onDidChange.fire()}}class Mfe extends F{get value(){return this._value}constructor(e){super(),this._onDidChange=this._register(new M),this.onDidChange=this._onDidChange.event,this._value=this._getPixelRatio(e);const t=this._register(new Afe(e));this._register(t.onDidChange(()=>{this._value=this._getPixelRatio(e),this._onDidChange.fire(this._value)}))}_getPixelRatio(e){const t=document.createElement("canvas").getContext("2d"),i=e.devicePixelRatio||1,n=t.webkitBackingStorePixelRatio||t.mozBackingStorePixelRatio||t.msBackingStorePixelRatio||t.oBackingStorePixelRatio||t.backingStorePixelRatio||1;return i/n}}class Ofe{constructor(){this.mapWindowIdToPixelRatioMonitor=new Map}_getOrCreatePixelRatioMonitor(e){const t=Zy(e);let i=this.mapWindowIdToPixelRatioMonitor.get(t);return i||(i=new Mfe(e),this.mapWindowIdToPixelRatioMonitor.set(t,i),oe.once(Jhe)(({vscodeWindowId:n})=>{n===t&&(i?.dispose(),this.mapWindowIdToPixelRatioMonitor.delete(t))})),i}getInstance(e){return this._getOrCreatePixelRatioMonitor(e)}}const sS=new Ofe;class Pfe{constructor(e,t){this.chr=e,this.type=t,this.width=0}fulfill(e){this.width=e}}class IW{constructor(e,t){this._bareFontInfo=e,this._requests=t,this._container=null,this._testElements=null}read(e){this._createDomElements(),e.document.body.appendChild(this._container),this._readFromDomElements(),this._container?.remove(),this._container=null,this._testElements=null}_createDomElements(){const e=document.createElement("div");e.style.position="absolute",e.style.top="-50000px",e.style.width="50000px";const t=document.createElement("div");In(t,this._bareFontInfo),e.appendChild(t);const i=document.createElement("div");In(i,this._bareFontInfo),i.style.fontWeight="bold",e.appendChild(i);const n=document.createElement("div");In(n,this._bareFontInfo),n.style.fontStyle="italic",e.appendChild(n);const o=[];for(const r of this._requests){let a;r.type===0&&(a=t),r.type===2&&(a=i),r.type===1&&(a=n),a.appendChild(document.createElement("br"));const l=document.createElement("span");IW._render(l,r),a.appendChild(l),o.push(l)}this._container=e,this._testElements=o}static _render(e,t){if(t.chr===" "){let i=" ";for(let n=0;n<8;n++)i+=i;e.innerText=i}else{let i=t.chr;for(let n=0;n<8;n++)i+=i;e.textContent=i}}_readFromDomElements(){for(let e=0,t=this._requests.length;e<t;e++){const i=this._requests[e],n=this._testElements[e];i.fulfill(n.offsetWidth/256)}}}function Ffe(s,e,t){new IW(e,t).read(s)}const Gr=new class{constructor(){this._zoomLevel=0,this._onDidChangeZoomLevel=new M,this.onDidChangeZoomLevel=this._onDidChangeZoomLevel.event}getZoomLevel(){return this._zoomLevel}setZoomLevel(s){s=Math.min(Math.max(-5,s),20),this._zoomLevel!==s&&(this._zoomLevel=s,this._onDidChangeZoomLevel.fire(this._zoomLevel))}},Bfe=Ye?1.5:1.35,Q2=8;class Em{static _create(e,t,i,n,o,r,a,l,c){r===0?r=Bfe*i:r<Q2&&(r=r*i),r=Math.round(r),r<Q2&&(r=Q2);const d=1+(c?0:Gr.getZoomLevel()*.1);return i*=d,r*=d,o===yJ&&(t==="normal"||t==="bold"?o=CJ:(o=`'wght' ${parseInt(t,10)}`,t="normal")),new Em({pixelRatio:l,fontFamily:e,fontWeight:t,fontSize:i,fontFeatureSettings:n,fontVariationSettings:o,lineHeight:r,letterSpacing:a})}constructor(e){this._bareFontInfoBrand=void 0,this.pixelRatio=e.pixelRatio,this.fontFamily=String(e.fontFamily),this.fontWeight=String(e.fontWeight),this.fontSize=e.fontSize,this.fontFeatureSettings=e.fontFeatureSettings,this.fontVariationSettings=e.fontVariationSettings,this.lineHeight=e.lineHeight|0,this.letterSpacing=e.letterSpacing}getId(){return`${this.pixelRatio}-${this.fontFamily}-${this.fontWeight}-${this.fontSize}-${this.fontFeatureSettings}-${this.fontVariationSettings}-${this.lineHeight}-${this.letterSpacing}`}getMassagedFontFamily(){const e=_o.fontFamily,t=Em._wrapInQuotes(this.fontFamily);return this.fontFamily!==e?`${t}, ${e}`:t}static _wrapInQuotes(e){return/[,"']/.test(e)?e:/[+ ]/.test(e)?`"${e}"`:e}}const Wfe=2;class BE extends Em{constructor(e,t){super(e),this._editorStylingBrand=void 0,this.version=Wfe,this.isTrusted=t,this.isMonospace=e.isMonospace,this.typicalHalfwidthCharacterWidth=e.typicalHalfwidthCharacterWidth,this.typicalFullwidthCharacterWidth=e.typicalFullwidthCharacterWidth,this.canUseHalfwidthRightwardsArrow=e.canUseHalfwidthRightwardsArrow,this.spaceWidth=e.spaceWidth,this.middotWidth=e.middotWidth,this.wsmiddotWidth=e.wsmiddotWidth,this.maxDigitWidth=e.maxDigitWidth}equals(e){return this.fontFamily===e.fontFamily&&this.fontWeight===e.fontWeight&&this.fontSize===e.fontSize&&this.fontFeatureSettings===e.fontFeatureSettings&&this.fontVariationSettings===e.fontVariationSettings&&this.lineHeight===e.lineHeight&&this.letterSpacing===e.letterSpacing&&this.typicalHalfwidthCharacterWidth===e.typicalHalfwidthCharacterWidth&&this.typicalFullwidthCharacterWidth===e.typicalFullwidthCharacterWidth&&this.canUseHalfwidthRightwardsArrow===e.canUseHalfwidthRightwardsArrow&&this.spaceWidth===e.spaceWidth&&this.middotWidth===e.middotWidth&&this.wsmiddotWidth===e.wsmiddotWidth&&this.maxDigitWidth===e.maxDigitWidth}}const CJ="normal",yJ="translate",Hfe="Consolas, 'Courier New', monospace",Vfe="Menlo, Monaco, 'Courier New', monospace",zfe="'Droid Sans Mono', 'monospace', monospace",_o={fontFamily:Ye?Vfe:Bn?Hfe:zfe,fontWeight:"normal",fontSize:Ye?12:14,lineHeight:0,letterSpacing:0},Qn={tabSize:4,indentSize:4,insertSpaces:!0,detectIndentation:!0,trimAutoWhitespace:!0,largeFileOptimizations:!0,bracketPairColorizationOptions:{enabled:!0,independentColorPoolPerBracketType:!1}},ql=8;class SJ{constructor(e){this._values=e}hasChanged(e){return this._values[e]}}class xJ{constructor(){this.stableMinimapLayoutInput=null,this.stableFitMaxMinimapScale=0,this.stableFitRemainingWidth=0}}class ci{constructor(e,t,i,n){this.id=e,this.name=t,this.defaultValue=i,this.schema=n}applyUpdate(e,t){return QR(e,t)}compute(e,t,i){return i}}class T1{constructor(e,t){this.newValue=e,this.didChange=t}}function QR(s,e){if(typeof s!="object"||typeof e!="object"||!s||!e)return new T1(e,s!==e);if(Array.isArray(s)||Array.isArray(e)){const i=Array.isArray(s)&&Array.isArray(e)&&ni(s,e);return new T1(e,!i)}let t=!1;for(const i in e)if(e.hasOwnProperty(i)){const n=QR(s[i],e[i]);n.didChange&&(s[i]=n.newValue,t=!0)}return new T1(s,t)}class Ig{constructor(e,t){this.schema=void 0,this.id=e,this.name="_never_",this.defaultValue=t}applyUpdate(e,t){return QR(e,t)}validate(e){return this.defaultValue}}class Qv{constructor(e,t,i,n){this.id=e,this.name=t,this.defaultValue=i,this.schema=n}applyUpdate(e,t){return QR(e,t)}compute(e,t,i){return i}}function we(s,e){return typeof s>"u"?e:s==="false"?!1:!!s}class Ge extends Qv{constructor(e,t,i,n=void 0){typeof n<"u"&&(n.type="boolean",n.default=i),super(e,t,i,n)}validate(e){return we(e,this.defaultValue)}}function Dd(s,e,t,i){if(typeof s=="string"&&(s=parseInt(s,10)),typeof s!="number"||isNaN(s))return e;let n=s;return n=Math.max(t,n),n=Math.min(i,n),n|0}class Tt extends Qv{static clampedInt(e,t,i,n){return Dd(e,t,i,n)}constructor(e,t,i,n,o,r=void 0){typeof r<"u"&&(r.type="integer",r.default=i,r.minimum=n,r.maximum=o),super(e,t,i,r),this.minimum=n,this.maximum=o}validate(e){return Tt.clampedInt(e,this.defaultValue,this.minimum,this.maximum)}}function $fe(s,e,t,i){if(typeof s>"u")return e;const n=po.float(s,e);return po.clamp(n,t,i)}class po extends Qv{static clamp(e,t,i){return e<t?t:e>i?i:e}static float(e,t){return typeof e=="string"&&(e=parseFloat(e)),typeof e!="number"||isNaN(e)?t:e}constructor(e,t,i,n,o,r,a){typeof o<"u"&&(o.type="number",o.default=i,o.minimum=r,o.maximum=a),super(e,t,i,o),this.validationFn=n,this.minimum=r,this.maximum=a}validate(e){return this.validationFn(po.float(e,this.defaultValue))}}class Es extends Qv{static string(e,t){return typeof e!="string"?t:e}constructor(e,t,i,n=void 0){typeof n<"u"&&(n.type="string",n.default=i),super(e,t,i,n)}validate(e){return Es.string(e,this.defaultValue)}}function Yt(s,e,t,i){return typeof s!="string"?e:i&&s in i?i[s]:t.indexOf(s)===-1?e:s}class ai extends Qv{constructor(e,t,i,n,o=void 0){typeof o<"u"&&(o.type="string",o.enum=n.slice(0),o.default=i),super(e,t,i,o),this._allowedValues=n}validate(e){return Yt(e,this.defaultValue,this._allowedValues)}}class kC extends ci{constructor(e,t,i,n,o,r,a=void 0){typeof a<"u"&&(a.type="string",a.enum=o,a.default=n),super(e,t,i,a),this._allowedValues=o,this._convert=r}validate(e){return typeof e!="string"?this.defaultValue:this._allowedValues.indexOf(e)===-1?this.defaultValue:this._convert(e)}}function Ufe(s){switch(s){case"none":return 0;case"keep":return 1;case"brackets":return 2;case"advanced":return 3;case"full":return 4}}class qfe extends ci{constructor(){super(2,"accessibilitySupport",0,{type:"string",enum:["auto","on","off"],enumDescriptions:[m(201,"Use platform APIs to detect when a Screen Reader is attached."),m(202,"Optimize for usage with a Screen Reader."),m(203,"Assume a screen reader is not attached.")],default:"auto",tags:["accessibility"],description:m(204,"Controls if the UI should run in a mode where it is optimized for screen readers.")})}validate(e){switch(e){case"auto":return 0;case"off":return 1;case"on":return 2}return this.defaultValue}compute(e,t,i){return i===0?e.accessibilitySupport:i}}class jfe extends ci{constructor(){const e={insertSpace:!0,ignoreEmptyLines:!0};super(29,"comments",e,{"editor.comments.insertSpace":{type:"boolean",default:e.insertSpace,description:m(205,"Controls whether a space character is inserted when commenting.")},"editor.comments.ignoreEmptyLines":{type:"boolean",default:e.ignoreEmptyLines,description:m(206,"Controls if empty lines should be ignored with toggle, add or remove actions for line comments.")}})}validate(e){if(!e||typeof e!="object")return this.defaultValue;const t=e;return{insertSpace:we(t.insertSpace,this.defaultValue.insertSpace),ignoreEmptyLines:we(t.ignoreEmptyLines,this.defaultValue.ignoreEmptyLines)}}}function Gfe(s){switch(s){case"blink":return 1;case"smooth":return 2;case"phase":return 3;case"expand":return 4;case"solid":return 5}}var mn;(function(s){s[s.Line=1]="Line",s[s.Block=2]="Block",s[s.Underline=3]="Underline",s[s.LineThin=4]="LineThin",s[s.BlockOutline=5]="BlockOutline",s[s.UnderlineThin=6]="UnderlineThin"})(mn||(mn={}));function k$(s){switch(s){case"line":return mn.Line;case"block":return mn.Block;case"underline":return mn.Underline;case"line-thin":return mn.LineThin;case"block-outline":return mn.BlockOutline;case"underline-thin":return mn.UnderlineThin}}class Kfe extends Ig{constructor(){super(162,"")}compute(e,t,i){const n=["monaco-editor"];return t.get(48)&&n.push(t.get(48)),e.extraEditorClassName&&n.push(e.extraEditorClassName),t.get(82)==="default"?n.push("mouse-default"):t.get(82)==="copy"&&n.push("mouse-copy"),t.get(127)&&n.push("showUnused"),t.get(157)&&n.push("showDeprecated"),n.join(" ")}}class Yfe extends Ge{constructor(){super(45,"emptySelectionClipboard",!0,{description:m(207,"Controls whether copying without a selection copies the current line.")})}compute(e,t,i){return i&&e.emptySelectionClipboard}}class Zfe extends ci{constructor(){const e={cursorMoveOnType:!0,findOnType:!0,seedSearchStringFromSelection:"always",autoFindInSelection:"never",globalFindClipboard:!1,addExtraSpaceOnTop:!0,loop:!0,history:"workspace",replaceHistory:"workspace"};super(50,"find",e,{"editor.find.cursorMoveOnType":{type:"boolean",default:e.cursorMoveOnType,description:m(208,"Controls whether the cursor should jump to find matches while typing.")},"editor.find.seedSearchStringFromSelection":{type:"string",enum:["never","always","selection"],default:e.seedSearchStringFromSelection,enumDescriptions:[m(209,"Never seed search string from the editor selection."),m(210,"Always seed search string from the editor selection, including word at cursor position."),m(211,"Only seed search string from the editor selection.")],description:m(212,"Controls whether the search string in the Find Widget is seeded from the editor selection.")},"editor.find.autoFindInSelection":{type:"string",enum:["never","always","multiline"],default:e.autoFindInSelection,enumDescriptions:[m(213,"Never turn on Find in Selection automatically (default)."),m(214,"Always turn on Find in Selection automatically."),m(215,"Turn on Find in Selection automatically when multiple lines of content are selected.")],description:m(216,"Controls the condition for turning on Find in Selection automatically.")},"editor.find.globalFindClipboard":{type:"boolean",default:e.globalFindClipboard,description:m(217,"Controls whether the Find Widget should read or modify the shared find clipboard on macOS."),included:Ye},"editor.find.addExtraSpaceOnTop":{type:"boolean",default:e.addExtraSpaceOnTop,description:m(218,"Controls whether the Find Widget should add extra lines on top of the editor. When true, you can scroll beyond the first line when the Find Widget is visible.")},"editor.find.loop":{type:"boolean",default:e.loop,description:m(219,"Controls whether the search automatically restarts from the beginning (or the end) when no further matches can be found.")},"editor.find.history":{type:"string",enum:["never","workspace"],default:"workspace",enumDescriptions:[m(220,"Do not store search history from the find widget."),m(221,"Store search history across the active workspace")],description:m(222,"Controls how the find widget history should be stored")},"editor.find.replaceHistory":{type:"string",enum:["never","workspace"],default:"workspace",enumDescriptions:[m(223,"Do not store history from the replace widget."),m(224,"Store replace history across the active workspace")],description:m(225,"Controls how the replace widget history should be stored")},"editor.find.findOnType":{type:"boolean",default:e.findOnType,description:m(226,"Controls whether the Find Widget should search as you type.")}})}validate(e){if(!e||typeof e!="object")return this.defaultValue;const t=e;return{cursorMoveOnType:we(t.cursorMoveOnType,this.defaultValue.cursorMoveOnType),findOnType:we(t.findOnType,this.defaultValue.findOnType),seedSearchStringFromSelection:typeof t.seedSearchStringFromSelection=="boolean"?t.seedSearchStringFromSelection?"always":"never":Yt(t.seedSearchStringFromSelection,this.defaultValue.seedSearchStringFromSelection,["never","always","selection"]),autoFindInSelection:typeof t.autoFindInSelection=="boolean"?t.autoFindInSelection?"always":"never":Yt(t.autoFindInSelection,this.defaultValue.autoFindInSelection,["never","always","multiline"]),globalFindClipboard:we(t.globalFindClipboard,this.defaultValue.globalFindClipboard),addExtraSpaceOnTop:we(t.addExtraSpaceOnTop,this.defaultValue.addExtraSpaceOnTop),loop:we(t.loop,this.defaultValue.loop),history:Yt(t.history,this.defaultValue.history,["never","workspace"]),replaceHistory:Yt(t.replaceHistory,this.defaultValue.replaceHistory,["never","workspace"])}}}const xd=class xd extends ci{constructor(){super(60,"fontLigatures",xd.OFF,{anyOf:[{type:"boolean",description:m(227,"Enables/Disables font ligatures ('calt' and 'liga' font features). Change this to a string for fine-grained control of the 'font-feature-settings' CSS property.")},{type:"string",description:m(228,"Explicit 'font-feature-settings' CSS property. A boolean can be passed instead if one only needs to turn on/off ligatures.")}],description:m(229,"Configures font ligatures or font features. Can be either a boolean to enable/disable ligatures or a string for the value of the CSS 'font-feature-settings' property."),default:!1})}validate(e){return typeof e>"u"?this.defaultValue:typeof e=="string"?e==="false"||e.length===0?xd.OFF:e==="true"?xd.ON:e:e?xd.ON:xd.OFF}};xd.OFF='"liga" off, "calt" off',xd.ON='"liga" on, "calt" on';let mh=xd;const Ld=class Ld extends ci{constructor(){super(63,"fontVariations",Ld.OFF,{anyOf:[{type:"boolean",description:m(230,"Enables/Disables the translation from font-weight to font-variation-settings. Change this to a string for fine-grained control of the 'font-variation-settings' CSS property.")},{type:"string",description:m(231,"Explicit 'font-variation-settings' CSS property. A boolean can be passed instead if one only needs to translate font-weight to font-variation-settings.")}],description:m(232,"Configures font variations. Can be either a boolean to enable/disable the translation from font-weight to font-variation-settings or a string for the value of the CSS 'font-variation-settings' property."),default:!1})}validate(e){return typeof e>"u"?this.defaultValue:typeof e=="string"?e==="false"?Ld.OFF:e==="true"?Ld.TRANSLATE:e:e?Ld.TRANSLATE:Ld.OFF}compute(e,t,i){return e.fontInfo.fontVariationSettings}};Ld.OFF=CJ,Ld.TRANSLATE=yJ;let zP=Ld;class Xfe extends Ig{constructor(){super(59,new BE({pixelRatio:0,fontFamily:"",fontWeight:"",fontSize:0,fontFeatureSettings:"",fontVariationSettings:"",lineHeight:0,letterSpacing:0,isMonospace:!1,typicalHalfwidthCharacterWidth:0,typicalFullwidthCharacterWidth:0,canUseHalfwidthRightwardsArrow:!1,spaceWidth:0,middotWidth:0,wsmiddotWidth:0,maxDigitWidth:0},!1))}compute(e,t,i){return e.fontInfo}}class Qfe extends Ig{constructor(){super(161,mn.Line)}compute(e,t,i){return e.inputMode==="overtype"?t.get(92):t.get(34)}}class Jfe extends Ig{constructor(){super(170,!1)}compute(e,t){return e.editContextSupported&&t.get(44)}}class epe extends Ig{constructor(){super(172,!1)}compute(e,t){return e.accessibilitySupport===2?t.get(7):t.get(6)}}class tpe extends Qv{constructor(){super(61,"fontSize",_o.fontSize,{type:"number",minimum:6,maximum:100,default:_o.fontSize,description:m(233,"Controls the font size in pixels.")})}validate(e){const t=po.float(e,this.defaultValue);return t===0?_o.fontSize:po.clamp(t,6,100)}compute(e,t,i){return e.fontInfo.fontSize}}const Xl=class Xl extends ci{constructor(){super(62,"fontWeight",_o.fontWeight,{anyOf:[{type:"number",minimum:Xl.MINIMUM_VALUE,maximum:Xl.MAXIMUM_VALUE,errorMessage:m(234,'Only "normal" and "bold" keywords or numbers between 1 and 1000 are allowed.')},{type:"string",pattern:"^(normal|bold|1000|[1-9][0-9]{0,2})$"},{enum:Xl.SUGGESTION_VALUES}],default:_o.fontWeight,description:m(235,'Controls the font weight. Accepts "normal" and "bold" keywords or numbers between 1 and 1000.')})}validate(e){return e==="normal"||e==="bold"?e:String(Tt.clampedInt(e,_o.fontWeight,Xl.MINIMUM_VALUE,Xl.MAXIMUM_VALUE))}};Xl.SUGGESTION_VALUES=["normal","bold","100","200","300","400","500","600","700","800","900"],Xl.MINIMUM_VALUE=1,Xl.MAXIMUM_VALUE=1e3;let $P=Xl;class ipe extends ci{constructor(){const e={multiple:"peek",multipleDefinitions:"peek",multipleTypeDefinitions:"peek",multipleDeclarations:"peek",multipleImplementations:"peek",multipleReferences:"peek",multipleTests:"peek",alternativeDefinitionCommand:"editor.action.goToReferences",alternativeTypeDefinitionCommand:"editor.action.goToReferences",alternativeDeclarationCommand:"editor.action.goToReferences",alternativeImplementationCommand:"",alternativeReferenceCommand:"",alternativeTestsCommand:""},t={type:"string",enum:["peek","gotoAndPeek","goto"],default:e.multiple,enumDescriptions:[m(236,"Show Peek view of the results (default)"),m(237,"Go to the primary result and show a Peek view"),m(238,"Go to the primary result and enable Peek-less navigation to others")]},i=["","editor.action.referenceSearch.trigger","editor.action.goToReferences","editor.action.peekImplementation","editor.action.goToImplementation","editor.action.peekTypeDefinition","editor.action.goToTypeDefinition","editor.action.peekDeclaration","editor.action.revealDeclaration","editor.action.peekDefinition","editor.action.revealDefinitionAside","editor.action.revealDefinition"];super(67,"gotoLocation",e,{"editor.gotoLocation.multiple":{deprecationMessage:m(239,"This setting is deprecated, please use separate settings like 'editor.editor.gotoLocation.multipleDefinitions' or 'editor.editor.gotoLocation.multipleImplementations' instead.")},"editor.gotoLocation.multipleDefinitions":{description:m(240,"Controls the behavior the 'Go to Definition'-command when multiple target locations exist."),...t},"editor.gotoLocation.multipleTypeDefinitions":{description:m(241,"Controls the behavior the 'Go to Type Definition'-command when multiple target locations exist."),...t},"editor.gotoLocation.multipleDeclarations":{description:m(242,"Controls the behavior the 'Go to Declaration'-command when multiple target locations exist."),...t},"editor.gotoLocation.multipleImplementations":{description:m(243,"Controls the behavior the 'Go to Implementations'-command when multiple target locations exist."),...t},"editor.gotoLocation.multipleReferences":{description:m(244,"Controls the behavior the 'Go to References'-command when multiple target locations exist."),...t},"editor.gotoLocation.alternativeDefinitionCommand":{type:"string",default:e.alternativeDefinitionCommand,enum:i,description:m(245,"Alternative command id that is being executed when the result of 'Go to Definition' is the current location.")},"editor.gotoLocation.alternativeTypeDefinitionCommand":{type:"string",default:e.alternativeTypeDefinitionCommand,enum:i,description:m(246,"Alternative command id that is being executed when the result of 'Go to Type Definition' is the current location.")},"editor.gotoLocation.alternativeDeclarationCommand":{type:"string",default:e.alternativeDeclarationCommand,enum:i,description:m(247,"Alternative command id that is being executed when the result of 'Go to Declaration' is the current location.")},"editor.gotoLocation.alternativeImplementationCommand":{type:"string",default:e.alternativeImplementationCommand,enum:i,description:m(248,"Alternative command id that is being executed when the result of 'Go to Implementation' is the current location.")},"editor.gotoLocation.alternativeReferenceCommand":{type:"string",default:e.alternativeReferenceCommand,enum:i,description:m(249,"Alternative command id that is being executed when the result of 'Go to Reference' is the current location.")}})}validate(e){if(!e||typeof e!="object")return this.defaultValue;const t=e;return{multiple:Yt(t.multiple,this.defaultValue.multiple,["peek","gotoAndPeek","goto"]),multipleDefinitions:Yt(t.multipleDefinitions,"peek",["peek","gotoAndPeek","goto"]),multipleTypeDefinitions:Yt(t.multipleTypeDefinitions,"peek",["peek","gotoAndPeek","goto"]),multipleDeclarations:Yt(t.multipleDeclarations,"peek",["peek","gotoAndPeek","goto"]),multipleImplementations:Yt(t.multipleImplementations,"peek",["peek","gotoAndPeek","goto"]),multipleReferences:Yt(t.multipleReferences,"peek",["peek","gotoAndPeek","goto"]),multipleTests:Yt(t.multipleTests,"peek",["peek","gotoAndPeek","goto"]),alternativeDefinitionCommand:Es.string(t.alternativeDefinitionCommand,this.defaultValue.alternativeDefinitionCommand),alternativeTypeDefinitionCommand:Es.string(t.alternativeTypeDefinitionCommand,this.defaultValue.alternativeTypeDefinitionCommand),alternativeDeclarationCommand:Es.string(t.alternativeDeclarationCommand,this.defaultValue.alternativeDeclarationCommand),alternativeImplementationCommand:Es.string(t.alternativeImplementationCommand,this.defaultValue.alternativeImplementationCommand),alternativeReferenceCommand:Es.string(t.alternativeReferenceCommand,this.defaultValue.alternativeReferenceCommand),alternativeTestsCommand:Es.string(t.alternativeTestsCommand,this.defaultValue.alternativeTestsCommand)}}}class npe extends ci{constructor(){const e={enabled:!0,delay:300,hidingDelay:300,sticky:!0,above:!0};super(69,"hover",e,{"editor.hover.enabled":{type:"boolean",default:e.enabled,description:m(250,"Controls whether the hover is shown.")},"editor.hover.delay":{type:"number",default:e.delay,minimum:0,maximum:1e4,description:m(251,"Controls the delay in milliseconds after which the hover is shown.")},"editor.hover.sticky":{type:"boolean",default:e.sticky,description:m(252,"Controls whether the hover should remain visible when mouse is moved over it.")},"editor.hover.hidingDelay":{type:"integer",minimum:0,default:e.hidingDelay,markdownDescription:m(253,"Controls the delay in milliseconds after which the hover is hidden. Requires `#editor.hover.sticky#` to be enabled.")},"editor.hover.above":{type:"boolean",default:e.above,description:m(254,"Prefer showing hovers above the line, if there's space.")}})}validate(e){if(!e||typeof e!="object")return this.defaultValue;const t=e;return{enabled:we(t.enabled,this.defaultValue.enabled),delay:Tt.clampedInt(t.delay,this.defaultValue.delay,0,1e4),sticky:we(t.sticky,this.defaultValue.sticky),hidingDelay:Tt.clampedInt(t.hidingDelay,this.defaultValue.hidingDelay,0,6e5),above:we(t.above,this.defaultValue.above)}}}class uw extends Ig{constructor(){super(165,{width:0,height:0,glyphMarginLeft:0,glyphMarginWidth:0,glyphMarginDecorationLaneCount:0,lineNumbersLeft:0,lineNumbersWidth:0,decorationsLeft:0,decorationsWidth:0,contentLeft:0,contentWidth:0,minimap:{renderMinimap:0,minimapLeft:0,minimapWidth:0,minimapHeightIsEditorHeight:!1,minimapIsSampling:!1,minimapScale:1,minimapLineHeight:1,minimapCanvasInnerWidth:0,minimapCanvasInnerHeight:0,minimapCanvasOuterWidth:0,minimapCanvasOuterHeight:0},viewportColumn:0,isWordWrapMinified:!1,isViewportWrapping:!1,wrappingColumn:-1,verticalScrollbarWidth:0,horizontalScrollbarHeight:0,overviewRuler:{top:0,width:0,height:0,right:0}})}compute(e,t,i){return uw.computeLayout(t,{memory:e.memory,outerWidth:e.outerWidth,outerHeight:e.outerHeight,isDominatedByLongLines:e.isDominatedByLongLines,lineHeight:e.fontInfo.lineHeight,viewLineCount:e.viewLineCount,lineNumbersDigitCount:e.lineNumbersDigitCount,typicalHalfwidthCharacterWidth:e.fontInfo.typicalHalfwidthCharacterWidth,maxDigitWidth:e.fontInfo.maxDigitWidth,pixelRatio:e.pixelRatio,glyphMarginDecorationLaneCount:e.glyphMarginDecorationLaneCount})}static computeContainedMinimapLineCount(e){const t=e.height/e.lineHeight,i=Math.floor(e.paddingTop/e.lineHeight);let n=Math.floor(e.paddingBottom/e.lineHeight);e.scrollBeyondLastLine&&(n=Math.max(n,t-1));const o=(i+e.viewLineCount+n)/(e.pixelRatio*e.height),r=Math.floor(e.viewLineCount/o);return{typicalViewportLineCount:t,extraLinesBeforeFirstLine:i,extraLinesBeyondLastLine:n,desiredRatio:o,minimapLineCount:r}}static _computeMinimapLayout(e,t){const i=e.outerWidth,n=e.outerHeight,o=e.pixelRatio;if(!e.minimap.enabled)return{renderMinimap:0,minimapLeft:0,minimapWidth:0,minimapHeightIsEditorHeight:!1,minimapIsSampling:!1,minimapScale:1,minimapLineHeight:1,minimapCanvasInnerWidth:0,minimapCanvasInnerHeight:Math.floor(o*n),minimapCanvasOuterWidth:0,minimapCanvasOuterHeight:n};const r=t.stableMinimapLayoutInput,a=r&&e.outerHeight===r.outerHeight&&e.lineHeight===r.lineHeight&&e.typicalHalfwidthCharacterWidth===r.typicalHalfwidthCharacterWidth&&e.pixelRatio===r.pixelRatio&&e.scrollBeyondLastLine===r.scrollBeyondLastLine&&e.paddingTop===r.paddingTop&&e.paddingBottom===r.paddingBottom&&e.minimap.enabled===r.minimap.enabled&&e.minimap.side===r.minimap.side&&e.minimap.size===r.minimap.size&&e.minimap.showSlider===r.minimap.showSlider&&e.minimap.renderCharacters===r.minimap.renderCharacters&&e.minimap.maxColumn===r.minimap.maxColumn&&e.minimap.scale===r.minimap.scale&&e.verticalScrollbarWidth===r.verticalScrollbarWidth&&e.isViewportWrapping===r.isViewportWrapping,l=e.lineHeight,c=e.typicalHalfwidthCharacterWidth,d=e.scrollBeyondLastLine,h=e.minimap.renderCharacters;let u=o>=2?Math.round(e.minimap.scale*2):e.minimap.scale;const g=e.minimap.maxColumn,f=e.minimap.size,p=e.minimap.side,_=e.verticalScrollbarWidth,b=e.viewLineCount,w=e.remainingWidth,C=e.isViewportWrapping,v=h?2:3;let y=Math.floor(o*n);const x=y/o;let L=!1,E=!1,I=v*u,R=u/o,O=1;if(f==="fill"||f==="fit"){const{typicalViewportLineCount:ne,extraLinesBeforeFirstLine:de,extraLinesBeyondLastLine:ke,desiredRatio:Me,minimapLineCount:Ze}=uw.computeContainedMinimapLineCount({viewLineCount:b,scrollBeyondLastLine:d,paddingTop:e.paddingTop,paddingBottom:e.paddingBottom,height:n,lineHeight:l,pixelRatio:o});if(b/Ze>1)L=!0,E=!0,u=1,I=1,R=u/o;else{let At=!1,Et=u+1;if(f==="fit"){const un=Math.ceil((de+b+ke)*I);C&&a&&w<=t.stableFitRemainingWidth?(At=!0,Et=t.stableFitMaxMinimapScale):At=un>y}if(f==="fill"||At){L=!0;const un=u;I=Math.min(l*o,Math.max(1,Math.floor(1/Me))),C&&a&&w<=t.stableFitRemainingWidth&&(Et=t.stableFitMaxMinimapScale),u=Math.min(Et,Math.max(1,Math.floor(I/v))),u>un&&(O=Math.min(2,u/un)),R=u/o/O,y=Math.ceil(Math.max(ne,de+b+ke)*I),C?(t.stableMinimapLayoutInput=e,t.stableFitRemainingWidth=w,t.stableFitMaxMinimapScale=u):(t.stableMinimapLayoutInput=null,t.stableFitRemainingWidth=0)}}}const P=Math.floor(g*R),B=Math.min(P,Math.max(0,Math.floor((w-_-2)*R/(c+R)))+ql);let $=Math.floor(o*B);const z=$/o;$=Math.floor($*O);const Q=h?1:2,J=p==="left"?0:i-B-_;return{renderMinimap:Q,minimapLeft:J,minimapWidth:B,minimapHeightIsEditorHeight:L,minimapIsSampling:E,minimapScale:u,minimapLineHeight:I,minimapCanvasInnerWidth:$,minimapCanvasInnerHeight:y,minimapCanvasOuterWidth:z,minimapCanvasOuterHeight:x}}static computeLayout(e,t){const i=t.outerWidth|0,n=t.outerHeight|0,o=t.lineHeight|0,r=t.lineNumbersDigitCount|0,a=t.typicalHalfwidthCharacterWidth,l=t.maxDigitWidth,c=t.pixelRatio,d=t.viewLineCount,h=e.get(154),u=h==="inherit"?e.get(153):h,g=u==="inherit"?e.get(149):u,f=e.get(152),p=t.isDominatedByLongLines,_=e.get(66),b=e.get(76).renderType!==0,w=e.get(77),C=e.get(119),v=e.get(96),y=e.get(81),x=e.get(117),L=x.verticalScrollbarSize,E=x.verticalHasArrows,I=x.arrowSize,R=x.horizontalScrollbarSize,O=e.get(52),P=e.get(126)!=="never";let B=e.get(74);O&&P&&(B+=16);let $=0;if(b){const Ws=Math.max(r,w);$=Math.round(Ws*l)}let z=0;_&&(z=o*t.glyphMarginDecorationLaneCount);let Q=0,J=Q+z,ne=J+$,de=ne+B;const ke=i-z-$-B;let Me=!1,Ze=!1,Xe=-1;e.get(2)===2&&u==="inherit"&&p?(Me=!0,Ze=!0):g==="on"||g==="bounded"?Ze=!0:g==="wordWrapColumn"&&(Xe=f);const At=uw._computeMinimapLayout({outerWidth:i,outerHeight:n,lineHeight:o,typicalHalfwidthCharacterWidth:a,pixelRatio:c,scrollBeyondLastLine:C,paddingTop:v.top,paddingBottom:v.bottom,minimap:y,verticalScrollbarWidth:L,viewLineCount:d,remainingWidth:ke,isViewportWrapping:Ze},t.memory||new xJ);At.renderMinimap!==0&&At.minimapLeft===0&&(Q+=At.minimapWidth,J+=At.minimapWidth,ne+=At.minimapWidth,de+=At.minimapWidth);const Et=ke-At.minimapWidth,un=Math.max(1,Math.floor((Et-L-2)/a)),Vi=E?I:0;return Ze&&(Xe=Math.max(1,un),g==="bounded"&&(Xe=Math.min(Xe,f))),{width:i,height:n,glyphMarginLeft:Q,glyphMarginWidth:z,glyphMarginDecorationLaneCount:t.glyphMarginDecorationLaneCount,lineNumbersLeft:J,lineNumbersWidth:$,decorationsLeft:ne,decorationsWidth:B,contentLeft:de,contentWidth:Et,minimap:At,viewportColumn:un,isWordWrapMinified:Me,isViewportWrapping:Ze,wrappingColumn:Xe,verticalScrollbarWidth:L,horizontalScrollbarHeight:R,overviewRuler:{top:Vi,width:L,height:n-2*Vi,right:0}}}}class spe extends ci{constructor(){super(156,"wrappingStrategy","simple",{"editor.wrappingStrategy":{enumDescriptions:[m(255,"Assumes that all characters are of the same width. This is a fast algorithm that works correctly for monospace fonts and certain scripts (like Latin characters) where glyphs are of equal width."),m(256,"Delegates wrapping points computation to the browser. This is a slow algorithm, that might cause freezes for large files, but it works correctly in all cases.")],type:"string",enum:["simple","advanced"],default:"simple",description:m(257,"Controls the algorithm that computes wrapping points. Note that when in accessibility mode, advanced will be used for the best experience.")}})}validate(e){return Yt(e,"simple",["simple","advanced"])}compute(e,t,i){return t.get(2)===2?"advanced":i}}var xa;(function(s){s.Off="off",s.OnCode="onCode",s.On="on"})(xa||(xa={}));class ope extends ci{constructor(){const e={enabled:xa.OnCode};super(73,"lightbulb",e,{"editor.lightbulb.enabled":{type:"string",enum:[xa.Off,xa.OnCode,xa.On],default:e.enabled,enumDescriptions:[m(258,"Disable the code action menu."),m(259,"Show the code action menu when the cursor is on lines with code."),m(260,"Show the code action menu when the cursor is on lines with code or on empty lines.")],description:m(261,"Enables the Code Action lightbulb in the editor.")}})}validate(e){return!e||typeof e!="object"?this.defaultValue:{enabled:Yt(e.enabled,this.defaultValue.enabled,[xa.Off,xa.OnCode,xa.On])}}}class rpe extends ci{constructor(){const e={enabled:!0,maxLineCount:5,defaultModel:"outlineModel",scrollWithEditor:!0};super(131,"stickyScroll",e,{"editor.stickyScroll.enabled":{type:"boolean",default:e.enabled,description:m(262,"Shows the nested current scopes during the scroll at the top of the editor.")},"editor.stickyScroll.maxLineCount":{type:"number",default:e.maxLineCount,minimum:1,maximum:20,description:m(263,"Defines the maximum number of sticky lines to show.")},"editor.stickyScroll.defaultModel":{type:"string",enum:["outlineModel","foldingProviderModel","indentationModel"],default:e.defaultModel,description:m(264,"Defines the model to use for determining which lines to stick. If the outline model does not exist, it will fall back on the folding provider model which falls back on the indentation model. This order is respected in all three cases.")},"editor.stickyScroll.scrollWithEditor":{type:"boolean",default:e.scrollWithEditor,description:m(265,"Enable scrolling of Sticky Scroll with the editor's horizontal scrollbar.")}})}validate(e){if(!e||typeof e!="object")return this.defaultValue;const t=e;return{enabled:we(t.enabled,this.defaultValue.enabled),maxLineCount:Tt.clampedInt(t.maxLineCount,this.defaultValue.maxLineCount,1,20),defaultModel:Yt(t.defaultModel,this.defaultValue.defaultModel,["outlineModel","foldingProviderModel","indentationModel"]),scrollWithEditor:we(t.scrollWithEditor,this.defaultValue.scrollWithEditor)}}}class ape extends ci{constructor(){const e={enabled:"on",fontSize:0,fontFamily:"",padding:!1,maximumLength:43};super(159,"inlayHints",e,{"editor.inlayHints.enabled":{type:"string",default:e.enabled,description:m(266,"Enables the inlay hints in the editor."),enum:["on","onUnlessPressed","offUnlessPressed","off"],markdownEnumDescriptions:[m(267,"Inlay hints are enabled"),m(268,"Inlay hints are showing by default and hide when holding {0}",Ye?"Ctrl+Option":"Ctrl+Alt"),m(269,"Inlay hints are hidden by default and show when holding {0}",Ye?"Ctrl+Option":"Ctrl+Alt"),m(270,"Inlay hints are disabled")]},"editor.inlayHints.fontSize":{type:"number",default:e.fontSize,markdownDescription:m(271,"Controls font size of inlay hints in the editor. As default the {0} is used when the configured value is less than {1} or greater than the editor font size.","`#editor.fontSize#`","`5`")},"editor.inlayHints.fontFamily":{type:"string",default:e.fontFamily,markdownDescription:m(272,"Controls font family of inlay hints in the editor. When set to empty, the {0} is used.","`#editor.fontFamily#`")},"editor.inlayHints.padding":{type:"boolean",default:e.padding,description:m(273,"Enables the padding around the inlay hints in the editor.")},"editor.inlayHints.maximumLength":{type:"number",default:e.maximumLength,markdownDescription:m(274,"Maximum overall length of inlay hints, for a single line, before they get truncated by the editor. Set to `0` to never truncate")}})}validate(e){if(!e||typeof e!="object")return this.defaultValue;const t=e;return typeof t.enabled=="boolean"&&(t.enabled=t.enabled?"on":"off"),{enabled:Yt(t.enabled,this.defaultValue.enabled,["on","off","offUnlessPressed","onUnlessPressed"]),fontSize:Tt.clampedInt(t.fontSize,this.defaultValue.fontSize,0,100),fontFamily:Es.string(t.fontFamily,this.defaultValue.fontFamily),padding:we(t.padding,this.defaultValue.padding),maximumLength:Tt.clampedInt(t.maximumLength,this.defaultValue.maximumLength,0,Number.MAX_SAFE_INTEGER)}}}class lpe extends ci{constructor(){super(74,"lineDecorationsWidth",10)}validate(e){return typeof e=="string"&&/^\d+(\.\d+)?ch$/.test(e)?-parseFloat(e.substring(0,e.length-2)):Tt.clampedInt(e,this.defaultValue,0,1e3)}compute(e,t,i){return i<0?Tt.clampedInt(-i*e.fontInfo.typicalHalfwidthCharacterWidth,this.defaultValue,0,1e3):i}}class cpe extends po{constructor(){super(75,"lineHeight",_o.lineHeight,e=>po.clamp(e,0,150),{markdownDescription:m(275,`Controls the line height. 
 - Use 0 to automatically compute the line height from the font size.
 - Values between 0 and 8 will be used as a multiplier with the font size.
 - Values greater than or equal to 8 will be used as effective values.`)},0,150)}compute(e,t,i){return e.fontInfo.lineHeight}}class dpe extends ci{constructor(){const e={enabled:!0,size:"proportional",side:"right",showSlider:"mouseover",autohide:"none",renderCharacters:!0,maxColumn:120,scale:1,showRegionSectionHeaders:!0,showMarkSectionHeaders:!0,markSectionHeaderRegex:"\\bMARK:\\s*(?<separator>-?)\\s*(?<label>.*)$",sectionHeaderFontSize:9,sectionHeaderLetterSpacing:1};super(81,"minimap",e,{"editor.minimap.enabled":{type:"boolean",default:e.enabled,description:m(276,"Controls whether the minimap is shown.")},"editor.minimap.autohide":{type:"string",enum:["none","mouseover","scroll"],enumDescriptions:[m(277,"The minimap is always shown."),m(278,"The minimap is hidden when mouse is not over the minimap and shown when mouse is over the minimap."),m(279,"The minimap is only shown when the editor is scrolled")],default:e.autohide,description:m(280,"Controls whether the minimap is hidden automatically.")},"editor.minimap.size":{type:"string",enum:["proportional","fill","fit"],enumDescriptions:[m(281,"The minimap has the same size as the editor contents (and might scroll)."),m(282,"The minimap will stretch or shrink as necessary to fill the height of the editor (no scrolling)."),m(283,"The minimap will shrink as necessary to never be larger than the editor (no scrolling).")],default:e.size,description:m(284,"Controls the size of the minimap.")},"editor.minimap.side":{type:"string",enum:["left","right"],default:e.side,description:m(285,"Controls the side where to render the minimap.")},"editor.minimap.showSlider":{type:"string",enum:["always","mouseover"],default:e.showSlider,description:m(286,"Controls when the minimap slider is shown.")},"editor.minimap.scale":{type:"number",default:e.scale,minimum:1,maximum:3,enum:[1,2,3],description:m(287,"Scale of content drawn in the minimap: 1, 2 or 3.")},"editor.minimap.renderCharacters":{type:"boolean",default:e.renderCharacters,description:m(288,"Render the actual characters on a line as opposed to color blocks.")},"editor.minimap.maxColumn":{type:"number",default:e.maxColumn,description:m(289,"Limit the width of the minimap to render at most a certain number of columns.")},"editor.minimap.showRegionSectionHeaders":{type:"boolean",default:e.showRegionSectionHeaders,description:m(290,"Controls whether named regions are shown as section headers in the minimap.")},"editor.minimap.showMarkSectionHeaders":{type:"boolean",default:e.showMarkSectionHeaders,description:m(291,"Controls whether MARK: comments are shown as section headers in the minimap.")},"editor.minimap.markSectionHeaderRegex":{type:"string",default:e.markSectionHeaderRegex,description:m(292,"Defines the regular expression used to find section headers in comments. The regex must contain a named match group `label` (written as `(?<label>.+)`) that encapsulates the section header, otherwise it will not work. Optionally you can include another match group named `separator`. Use \\n in the pattern to match multi-line headers.")},"editor.minimap.sectionHeaderFontSize":{type:"number",default:e.sectionHeaderFontSize,description:m(293,"Controls the font size of section headers in the minimap.")},"editor.minimap.sectionHeaderLetterSpacing":{type:"number",default:e.sectionHeaderLetterSpacing,description:m(294,"Controls the amount of space (in pixels) between characters of section header. This helps the readability of the header in small font sizes.")}})}validate(e){if(!e||typeof e!="object")return this.defaultValue;const t=e;let i=this.defaultValue.markSectionHeaderRegex;const n=t.markSectionHeaderRegex;if(typeof n=="string")try{new RegExp(n,"d"),i=n}catch{}return{enabled:we(t.enabled,this.defaultValue.enabled),autohide:Yt(t.autohide,this.defaultValue.autohide,["none","mouseover","scroll"]),size:Yt(t.size,this.defaultValue.size,["proportional","fill","fit"]),side:Yt(t.side,this.defaultValue.side,["right","left"]),showSlider:Yt(t.showSlider,this.defaultValue.showSlider,["always","mouseover"]),renderCharacters:we(t.renderCharacters,this.defaultValue.renderCharacters),scale:Tt.clampedInt(t.scale,1,1,3),maxColumn:Tt.clampedInt(t.maxColumn,this.defaultValue.maxColumn,1,1e4),showRegionSectionHeaders:we(t.showRegionSectionHeaders,this.defaultValue.showRegionSectionHeaders),showMarkSectionHeaders:we(t.showMarkSectionHeaders,this.defaultValue.showMarkSectionHeaders),markSectionHeaderRegex:i,sectionHeaderFontSize:po.clamp(po.float(t.sectionHeaderFontSize,this.defaultValue.sectionHeaderFontSize),4,32),sectionHeaderLetterSpacing:po.clamp(po.float(t.sectionHeaderLetterSpacing,this.defaultValue.sectionHeaderLetterSpacing),0,5)}}}function hpe(s){return s==="ctrlCmd"?Ye?"metaKey":"ctrlKey":"altKey"}class upe extends ci{constructor(){super(96,"padding",{top:0,bottom:0},{"editor.padding.top":{type:"number",default:0,minimum:0,maximum:1e3,description:m(295,"Controls the amount of space between the top edge of the editor and the first line.")},"editor.padding.bottom":{type:"number",default:0,minimum:0,maximum:1e3,description:m(296,"Controls the amount of space between the bottom edge of the editor and the last line.")}})}validate(e){if(!e||typeof e!="object")return this.defaultValue;const t=e;return{top:Tt.clampedInt(t.top,0,0,1e3),bottom:Tt.clampedInt(t.bottom,0,0,1e3)}}}class gpe extends ci{constructor(){const e={enabled:!0,cycle:!0};super(98,"parameterHints",e,{"editor.parameterHints.enabled":{type:"boolean",default:e.enabled,description:m(297,"Enables a pop-up that shows parameter documentation and type information as you type.")},"editor.parameterHints.cycle":{type:"boolean",default:e.cycle,description:m(298,"Controls whether the parameter hints menu cycles or closes when reaching the end of the list.")}})}validate(e){if(!e||typeof e!="object")return this.defaultValue;const t=e;return{enabled:we(t.enabled,this.defaultValue.enabled),cycle:we(t.cycle,this.defaultValue.cycle)}}}class fpe extends Ig{constructor(){super(163,1)}compute(e,t,i){return e.pixelRatio}}class ppe extends ci{constructor(){super(100,"placeholder",void 0)}validate(e){return typeof e>"u"?this.defaultValue:typeof e=="string"?e:this.defaultValue}}class mpe extends ci{constructor(){const e={other:"on",comments:"off",strings:"off"},t=[{type:"boolean"},{type:"string",enum:["on","inline","off"],enumDescriptions:[m(299,"Quick suggestions show inside the suggest widget"),m(300,"Quick suggestions show as ghost text"),m(301,"Quick suggestions are disabled")]}];super(102,"quickSuggestions",e,{type:"object",additionalProperties:!1,properties:{strings:{anyOf:t,default:e.strings,description:m(302,"Enable quick suggestions inside strings.")},comments:{anyOf:t,default:e.comments,description:m(303,"Enable quick suggestions inside comments.")},other:{anyOf:t,default:e.other,description:m(304,"Enable quick suggestions outside of strings and comments.")}},default:e,markdownDescription:m(305,"Controls whether suggestions should automatically show up while typing. This can be controlled for typing in comments, strings, and other code. Quick suggestion can be configured to show as ghost text or with the suggest widget. Also be aware of the {0}-setting which controls if suggestions are triggered by special characters.","`#editor.suggestOnTriggerCharacters#`")}),this.defaultValue=e}validate(e){if(typeof e=="boolean"){const c=e?"on":"off";return{comments:c,strings:c,other:c}}if(!e||typeof e!="object")return this.defaultValue;const{other:t,comments:i,strings:n}=e,o=["on","inline","off"];let r,a,l;return typeof t=="boolean"?r=t?"on":"off":r=Yt(t,this.defaultValue.other,o),typeof i=="boolean"?a=i?"on":"off":a=Yt(i,this.defaultValue.comments,o),typeof n=="boolean"?l=n?"on":"off":l=Yt(n,this.defaultValue.strings,o),{other:r,comments:a,strings:l}}}class _pe extends ci{constructor(){super(76,"lineNumbers",{renderType:1,renderFn:null},{type:"string",enum:["off","on","relative","interval"],enumDescriptions:[m(306,"Line numbers are not rendered."),m(307,"Line numbers are rendered as absolute number."),m(308,"Line numbers are rendered as distance in lines to cursor position."),m(309,"Line numbers are rendered every 10 lines.")],default:"on",description:m(310,"Controls the display of line numbers.")})}validate(e){let t=this.defaultValue.renderType,i=this.defaultValue.renderFn;return typeof e<"u"&&(typeof e=="function"?(t=4,i=e):e==="interval"?t=3:e==="relative"?t=2:e==="on"?t=1:t=0),{renderType:t,renderFn:i}}}function WE(s){const e=s.get(112);return e==="editable"?s.get(104):e!=="on"}function HE(s){return!s.get(172)}class bpe extends ci{constructor(){const e=[],t={type:"number",description:m(311,"Number of monospace characters at which this editor ruler will render.")};super(116,"rulers",e,{type:"array",items:{anyOf:[t,{type:["object"],properties:{column:t,color:{type:"string",description:m(312,"Color of this editor ruler."),format:"color-hex"}}}]},default:e,description:m(313,"Render vertical rulers after a certain number of monospace characters. Use multiple values for multiple rulers. No rulers are drawn if array is empty.")})}validate(e){if(Array.isArray(e)){const t=[];for(const i of e)if(typeof i=="number")t.push({column:Tt.clampedInt(i,0,0,1e4),color:null});else if(i&&typeof i=="object"){const n=i;t.push({column:Tt.clampedInt(n.column,0,0,1e4),color:n.color})}return t.sort((i,n)=>i.column-n.column),t}return this.defaultValue}}class wpe extends ci{constructor(){super(105,"readOnlyMessage",void 0)}validate(e){return!e||typeof e!="object"?this.defaultValue:e}}function E$(s,e){if(typeof s!="string")return e;switch(s){case"hidden":return 2;case"visible":return 3;default:return 1}}let vpe=class extends ci{constructor(){const e={vertical:1,horizontal:1,arrowSize:11,useShadows:!0,verticalHasArrows:!1,horizontalHasArrows:!1,horizontalScrollbarSize:12,horizontalSliderSize:12,verticalScrollbarSize:14,verticalSliderSize:14,handleMouseWheel:!0,alwaysConsumeMouseWheel:!0,scrollByPage:!1,ignoreHorizontalScrollbarInContentHeight:!1};super(117,"scrollbar",e,{"editor.scrollbar.vertical":{type:"string",enum:["auto","visible","hidden"],enumDescriptions:[m(314,"The vertical scrollbar will be visible only when necessary."),m(315,"The vertical scrollbar will always be visible."),m(316,"The vertical scrollbar will always be hidden.")],default:"auto",description:m(317,"Controls the visibility of the vertical scrollbar.")},"editor.scrollbar.horizontal":{type:"string",enum:["auto","visible","hidden"],enumDescriptions:[m(318,"The horizontal scrollbar will be visible only when necessary."),m(319,"The horizontal scrollbar will always be visible."),m(320,"The horizontal scrollbar will always be hidden.")],default:"auto",description:m(321,"Controls the visibility of the horizontal scrollbar.")},"editor.scrollbar.verticalScrollbarSize":{type:"number",default:e.verticalScrollbarSize,description:m(322,"The width of the vertical scrollbar.")},"editor.scrollbar.horizontalScrollbarSize":{type:"number",default:e.horizontalScrollbarSize,description:m(323,"The height of the horizontal scrollbar.")},"editor.scrollbar.scrollByPage":{type:"boolean",default:e.scrollByPage,description:m(324,"Controls whether clicks scroll by page or jump to click position.")},"editor.scrollbar.ignoreHorizontalScrollbarInContentHeight":{type:"boolean",default:e.ignoreHorizontalScrollbarInContentHeight,description:m(325,"When set, the horizontal scrollbar will not increase the size of the editor's content.")}})}validate(e){if(!e||typeof e!="object")return this.defaultValue;const t=e,i=Tt.clampedInt(t.horizontalScrollbarSize,this.defaultValue.horizontalScrollbarSize,0,1e3),n=Tt.clampedInt(t.verticalScrollbarSize,this.defaultValue.verticalScrollbarSize,0,1e3);return{arrowSize:Tt.clampedInt(t.arrowSize,this.defaultValue.arrowSize,0,1e3),vertical:E$(t.vertical,this.defaultValue.vertical),horizontal:E$(t.horizontal,this.defaultValue.horizontal),useShadows:we(t.useShadows,this.defaultValue.useShadows),verticalHasArrows:we(t.verticalHasArrows,this.defaultValue.verticalHasArrows),horizontalHasArrows:we(t.horizontalHasArrows,this.defaultValue.horizontalHasArrows),handleMouseWheel:we(t.handleMouseWheel,this.defaultValue.handleMouseWheel),alwaysConsumeMouseWheel:we(t.alwaysConsumeMouseWheel,this.defaultValue.alwaysConsumeMouseWheel),horizontalScrollbarSize:i,horizontalSliderSize:Tt.clampedInt(t.horizontalSliderSize,i,0,1e3),verticalScrollbarSize:n,verticalSliderSize:Tt.clampedInt(t.verticalSliderSize,n,0,1e3),scrollByPage:we(t.scrollByPage,this.defaultValue.scrollByPage),ignoreHorizontalScrollbarInContentHeight:we(t.ignoreHorizontalScrollbarInContentHeight,this.defaultValue.ignoreHorizontalScrollbarInContentHeight)}}};const lr="inUntrustedWorkspace",go={allowedCharacters:"editor.unicodeHighlight.allowedCharacters",invisibleCharacters:"editor.unicodeHighlight.invisibleCharacters",nonBasicASCII:"editor.unicodeHighlight.nonBasicASCII",ambiguousCharacters:"editor.unicodeHighlight.ambiguousCharacters",includeComments:"editor.unicodeHighlight.includeComments",includeStrings:"editor.unicodeHighlight.includeStrings",allowedLocales:"editor.unicodeHighlight.allowedLocales"};class Cpe extends ci{constructor(){const e={nonBasicASCII:lr,invisibleCharacters:!0,ambiguousCharacters:!0,includeComments:lr,includeStrings:!0,allowedCharacters:{},allowedLocales:{_os:!0,_vscode:!0}};super(142,"unicodeHighlight",e,{[go.nonBasicASCII]:{restricted:!0,type:["boolean","string"],enum:[!0,!1,lr],default:e.nonBasicASCII,description:m(326,"Controls whether all non-basic ASCII characters are highlighted. Only characters between U+0020 and U+007E, tab, line-feed and carriage-return are considered basic ASCII.")},[go.invisibleCharacters]:{restricted:!0,type:"boolean",default:e.invisibleCharacters,description:m(327,"Controls whether characters that just reserve space or have no width at all are highlighted.")},[go.ambiguousCharacters]:{restricted:!0,type:"boolean",default:e.ambiguousCharacters,description:m(328,"Controls whether characters are highlighted that can be confused with basic ASCII characters, except those that are common in the current user locale.")},[go.includeComments]:{restricted:!0,type:["boolean","string"],enum:[!0,!1,lr],default:e.includeComments,description:m(329,"Controls whether characters in comments should also be subject to Unicode highlighting.")},[go.includeStrings]:{restricted:!0,type:["boolean","string"],enum:[!0,!1,lr],default:e.includeStrings,description:m(330,"Controls whether characters in strings should also be subject to Unicode highlighting.")},[go.allowedCharacters]:{restricted:!0,type:"object",default:e.allowedCharacters,description:m(331,"Defines allowed characters that are not being highlighted."),additionalProperties:{type:"boolean"}},[go.allowedLocales]:{restricted:!0,type:"object",additionalProperties:{type:"boolean"},default:e.allowedLocales,description:m(332,"Unicode characters that are common in allowed locales are not being highlighted.")}})}applyUpdate(e,t){let i=!1;t.allowedCharacters&&e&&(Ho(e.allowedCharacters,t.allowedCharacters)||(e={...e,allowedCharacters:t.allowedCharacters},i=!0)),t.allowedLocales&&e&&(Ho(e.allowedLocales,t.allowedLocales)||(e={...e,allowedLocales:t.allowedLocales},i=!0));const n=super.applyUpdate(e,t);return i?new T1(n.newValue,!0):n}validate(e){if(!e||typeof e!="object")return this.defaultValue;const t=e;return{nonBasicASCII:gw(t.nonBasicASCII,lr,[!0,!1,lr]),invisibleCharacters:we(t.invisibleCharacters,this.defaultValue.invisibleCharacters),ambiguousCharacters:we(t.ambiguousCharacters,this.defaultValue.ambiguousCharacters),includeComments:gw(t.includeComments,lr,[!0,!1,lr]),includeStrings:gw(t.includeStrings,lr,[!0,!1,lr]),allowedCharacters:this.validateBooleanMap(t.allowedCharacters,this.defaultValue.allowedCharacters),allowedLocales:this.validateBooleanMap(t.allowedLocales,this.defaultValue.allowedLocales)}}validateBooleanMap(e,t){if(typeof e!="object"||!e)return t;const i={};for(const[n,o]of Object.entries(e))o===!0&&(i[n]=!0);return i}}class ype extends ci{constructor(){const e={enabled:!0,mode:"subwordSmart",showToolbar:"onHover",suppressSuggestions:!1,keepOnBlur:!1,fontFamily:"default",syntaxHighlightingEnabled:!0,minShowDelay:0,suppressInSnippetMode:!0,edits:{enabled:!0,showCollapsed:!1,renderSideBySide:"auto",allowCodeShifting:"always"},triggerCommandOnProviderChange:!1,experimental:{suppressInlineSuggestions:"",showOnSuggestConflict:"never",emptyResponseInformation:!0}};super(71,"inlineSuggest",e,{"editor.inlineSuggest.enabled":{type:"boolean",default:e.enabled,description:m(333,"Controls whether to automatically show inline suggestions in the editor.")},"editor.inlineSuggest.showToolbar":{type:"string",default:e.showToolbar,enum:["always","onHover","never"],enumDescriptions:[m(334,"Show the inline suggestion toolbar whenever an inline suggestion is shown."),m(335,"Show the inline suggestion toolbar when hovering over an inline suggestion."),m(336,"Never show the inline suggestion toolbar.")],description:m(337,"Controls when to show the inline suggestion toolbar.")},"editor.inlineSuggest.syntaxHighlightingEnabled":{type:"boolean",default:e.syntaxHighlightingEnabled,description:m(338,"Controls whether to show syntax highlighting for inline suggestions in the editor.")},"editor.inlineSuggest.suppressSuggestions":{type:"boolean",default:e.suppressSuggestions,description:m(339,"Controls how inline suggestions interact with the suggest widget. If enabled, the suggest widget is not shown automatically when inline suggestions are available.")},"editor.inlineSuggest.suppressInSnippetMode":{type:"boolean",default:e.suppressInSnippetMode,description:m(340,"Controls whether inline suggestions are suppressed when in snippet mode.")},"editor.inlineSuggest.minShowDelay":{type:"number",default:0,minimum:0,maximum:1e4,description:m(341,"Controls the minimal delay in milliseconds after which inline suggestions are shown after typing.")},"editor.inlineSuggest.experimental.suppressInlineSuggestions":{type:"string",default:e.experimental.suppressInlineSuggestions,tags:["experimental"],description:m(342,"Suppresses inline completions for specified extension IDs -- comma separated."),experiment:{mode:"auto"}},"editor.inlineSuggest.experimental.emptyResponseInformation":{type:"boolean",default:e.experimental.emptyResponseInformation,tags:["experimental"],description:m(343,"Controls whether to send request information from the inline suggestion provider."),experiment:{mode:"auto"}},"editor.inlineSuggest.triggerCommandOnProviderChange":{type:"boolean",default:e.triggerCommandOnProviderChange,tags:["experimental"],description:m(344,"Controls whether to trigger a command when the inline suggestion provider changes."),experiment:{mode:"auto"}},"editor.inlineSuggest.experimental.showOnSuggestConflict":{type:"string",default:e.experimental.showOnSuggestConflict,tags:["experimental"],enum:["always","never","whenSuggestListIsIncomplete"],description:m(345,"Controls whether to show inline suggestions when there is a suggest conflict."),experiment:{mode:"auto"}},"editor.inlineSuggest.fontFamily":{type:"string",default:e.fontFamily,description:m(346,"Controls the font family of the inline suggestions.")},"editor.inlineSuggest.edits.allowCodeShifting":{type:"string",default:e.edits.allowCodeShifting,description:m(347,"Controls whether showing a suggestion will shift the code to make space for the suggestion inline."),enum:["always","horizontal","never"],tags:["nextEditSuggestions"]},"editor.inlineSuggest.edits.renderSideBySide":{type:"string",default:e.edits.renderSideBySide,description:m(348,"Controls whether larger suggestions can be shown side by side."),enum:["auto","never"],enumDescriptions:[m(349,"Larger suggestions will show side by side if there is enough space, otherwise they will be shown below."),m(350,"Larger suggestions are never shown side by side and will always be shown below.")],tags:["nextEditSuggestions"]},"editor.inlineSuggest.edits.showCollapsed":{type:"boolean",default:e.edits.showCollapsed,description:m(351,"Controls whether the suggestion will show as collapsed until jumping to it."),tags:["nextEditSuggestions"]}})}validate(e){if(!e||typeof e!="object")return this.defaultValue;const t=e;return{enabled:we(t.enabled,this.defaultValue.enabled),mode:Yt(t.mode,this.defaultValue.mode,["prefix","subword","subwordSmart"]),showToolbar:Yt(t.showToolbar,this.defaultValue.showToolbar,["always","onHover","never"]),suppressSuggestions:we(t.suppressSuggestions,this.defaultValue.suppressSuggestions),keepOnBlur:we(t.keepOnBlur,this.defaultValue.keepOnBlur),fontFamily:Es.string(t.fontFamily,this.defaultValue.fontFamily),syntaxHighlightingEnabled:we(t.syntaxHighlightingEnabled,this.defaultValue.syntaxHighlightingEnabled),minShowDelay:Tt.clampedInt(t.minShowDelay,0,0,1e4),suppressInSnippetMode:we(t.suppressInSnippetMode,this.defaultValue.suppressInSnippetMode),edits:this._validateEdits(t.edits),triggerCommandOnProviderChange:we(t.triggerCommandOnProviderChange,this.defaultValue.triggerCommandOnProviderChange),experimental:this._validateExperimental(t.experimental)}}_validateEdits(e){if(!e||typeof e!="object")return this.defaultValue.edits;const t=e;return{enabled:we(t.enabled,this.defaultValue.edits.enabled),showCollapsed:we(t.showCollapsed,this.defaultValue.edits.showCollapsed),allowCodeShifting:Yt(t.allowCodeShifting,this.defaultValue.edits.allowCodeShifting,["always","horizontal","never"]),renderSideBySide:Yt(t.renderSideBySide,this.defaultValue.edits.renderSideBySide,["never","auto"])}}_validateExperimental(e){if(!e||typeof e!="object")return this.defaultValue.experimental;const t=e;return{suppressInlineSuggestions:Es.string(t.suppressInlineSuggestions,this.defaultValue.experimental.suppressInlineSuggestions),showOnSuggestConflict:Yt(t.showOnSuggestConflict,this.defaultValue.experimental.showOnSuggestConflict,["always","never","whenSuggestListIsIncomplete"]),emptyResponseInformation:we(t.emptyResponseInformation,this.defaultValue.experimental.emptyResponseInformation)}}}class Spe extends ci{constructor(){const e={enabled:Qn.bracketPairColorizationOptions.enabled,independentColorPoolPerBracketType:Qn.bracketPairColorizationOptions.independentColorPoolPerBracketType};super(21,"bracketPairColorization",e,{"editor.bracketPairColorization.enabled":{type:"boolean",default:e.enabled,markdownDescription:m(352,"Controls whether bracket pair colorization is enabled or not. Use {0} to override the bracket highlight colors.","`#workbench.colorCustomizations#`")},"editor.bracketPairColorization.independentColorPoolPerBracketType":{type:"boolean",default:e.independentColorPoolPerBracketType,description:m(353,"Controls whether each bracket type has its own independent color pool.")}})}validate(e){if(!e||typeof e!="object")return this.defaultValue;const t=e;return{enabled:we(t.enabled,this.defaultValue.enabled),independentColorPoolPerBracketType:we(t.independentColorPoolPerBracketType,this.defaultValue.independentColorPoolPerBracketType)}}}class xpe extends ci{constructor(){const e={bracketPairs:!1,bracketPairsHorizontal:"active",highlightActiveBracketPair:!0,indentation:!0,highlightActiveIndentation:!0};super(22,"guides",e,{"editor.guides.bracketPairs":{type:["boolean","string"],enum:[!0,"active",!1],enumDescriptions:[m(354,"Enables bracket pair guides."),m(355,"Enables bracket pair guides only for the active bracket pair."),m(356,"Disables bracket pair guides.")],default:e.bracketPairs,description:m(357,"Controls whether bracket pair guides are enabled or not.")},"editor.guides.bracketPairsHorizontal":{type:["boolean","string"],enum:[!0,"active",!1],enumDescriptions:[m(358,"Enables horizontal guides as addition to vertical bracket pair guides."),m(359,"Enables horizontal guides only for the active bracket pair."),m(360,"Disables horizontal bracket pair guides.")],default:e.bracketPairsHorizontal,description:m(361,"Controls whether horizontal bracket pair guides are enabled or not.")},"editor.guides.highlightActiveBracketPair":{type:"boolean",default:e.highlightActiveBracketPair,description:m(362,"Controls whether the editor should highlight the active bracket pair.")},"editor.guides.indentation":{type:"boolean",default:e.indentation,description:m(363,"Controls whether the editor should render indent guides.")},"editor.guides.highlightActiveIndentation":{type:["boolean","string"],enum:[!0,"always",!1],enumDescriptions:[m(364,"Highlights the active indent guide."),m(365,"Highlights the active indent guide even if bracket guides are highlighted."),m(366,"Do not highlight the active indent guide.")],default:e.highlightActiveIndentation,description:m(367,"Controls whether the editor should highlight the active indent guide.")}})}validate(e){if(!e||typeof e!="object")return this.defaultValue;const t=e;return{bracketPairs:gw(t.bracketPairs,this.defaultValue.bracketPairs,[!0,!1,"active"]),bracketPairsHorizontal:gw(t.bracketPairsHorizontal,this.defaultValue.bracketPairsHorizontal,[!0,!1,"active"]),highlightActiveBracketPair:we(t.highlightActiveBracketPair,this.defaultValue.highlightActiveBracketPair),indentation:we(t.indentation,this.defaultValue.indentation),highlightActiveIndentation:gw(t.highlightActiveIndentation,this.defaultValue.highlightActiveIndentation,[!0,!1,"always"])}}}function gw(s,e,t){const i=t.indexOf(s);return i===-1?e:t[i]}class Lpe extends ci{constructor(){const e={insertMode:"insert",filterGraceful:!0,snippetsPreventQuickSuggestions:!1,localityBonus:!1,shareSuggestSelections:!1,selectionMode:"always",showIcons:!0,showStatusBar:!1,preview:!1,previewMode:"subwordSmart",showInlineDetails:!0,showMethods:!0,showFunctions:!0,showConstructors:!0,showDeprecated:!0,matchOnWordStartOnly:!0,showFields:!0,showVariables:!0,showClasses:!0,showStructs:!0,showInterfaces:!0,showModules:!0,showProperties:!0,showEvents:!0,showOperators:!0,showUnits:!0,showValues:!0,showConstants:!0,showEnums:!0,showEnumMembers:!0,showKeywords:!0,showWords:!0,showColors:!0,showFiles:!0,showReferences:!0,showFolders:!0,showTypeParameters:!0,showSnippets:!0,showUsers:!0,showIssues:!0};super(134,"suggest",e,{"editor.suggest.insertMode":{type:"string",enum:["insert","replace"],enumDescriptions:[m(368,"Insert suggestion without overwriting text right of the cursor."),m(369,"Insert suggestion and overwrite text right of the cursor.")],default:e.insertMode,description:m(370,"Controls whether words are overwritten when accepting completions. Note that this depends on extensions opting into this feature.")},"editor.suggest.filterGraceful":{type:"boolean",default:e.filterGraceful,description:m(371,"Controls whether filtering and sorting suggestions accounts for small typos.")},"editor.suggest.localityBonus":{type:"boolean",default:e.localityBonus,description:m(372,"Controls whether sorting favors words that appear close to the cursor.")},"editor.suggest.shareSuggestSelections":{type:"boolean",default:e.shareSuggestSelections,markdownDescription:m(373,"Controls whether remembered suggestion selections are shared between multiple workspaces and windows (needs `#editor.suggestSelection#`).")},"editor.suggest.selectionMode":{type:"string",enum:["always","never","whenTriggerCharacter","whenQuickSuggestion"],enumDescriptions:[m(374,"Always select a suggestion when automatically triggering IntelliSense."),m(375,"Never select a suggestion when automatically triggering IntelliSense."),m(376,"Select a suggestion only when triggering IntelliSense from a trigger character."),m(377,"Select a suggestion only when triggering IntelliSense as you type.")],default:e.selectionMode,markdownDescription:m(378,"Controls whether a suggestion is selected when the widget shows. Note that this only applies to automatically triggered suggestions ({0} and {1}) and that a suggestion is always selected when explicitly invoked, e.g via `Ctrl+Space`.","`#editor.quickSuggestions#`","`#editor.suggestOnTriggerCharacters#`")},"editor.suggest.snippetsPreventQuickSuggestions":{type:"boolean",default:e.snippetsPreventQuickSuggestions,description:m(379,"Controls whether an active snippet prevents quick suggestions.")},"editor.suggest.showIcons":{type:"boolean",default:e.showIcons,description:m(380,"Controls whether to show or hide icons in suggestions.")},"editor.suggest.showStatusBar":{type:"boolean",default:e.showStatusBar,description:m(381,"Controls the visibility of the status bar at the bottom of the suggest widget.")},"editor.suggest.preview":{type:"boolean",default:e.preview,description:m(382,"Controls whether to preview the suggestion outcome in the editor.")},"editor.suggest.showInlineDetails":{type:"boolean",default:e.showInlineDetails,description:m(383,"Controls whether suggest details show inline with the label or only in the details widget.")},"editor.suggest.maxVisibleSuggestions":{type:"number",deprecationMessage:m(384,"This setting is deprecated. The suggest widget can now be resized.")},"editor.suggest.filteredTypes":{type:"object",deprecationMessage:m(385,"This setting is deprecated, please use separate settings like 'editor.suggest.showKeywords' or 'editor.suggest.showSnippets' instead.")},"editor.suggest.showMethods":{type:"boolean",default:!0,markdownDescription:m(386,"When enabled IntelliSense shows `method`-suggestions.")},"editor.suggest.showFunctions":{type:"boolean",default:!0,markdownDescription:m(387,"When enabled IntelliSense shows `function`-suggestions.")},"editor.suggest.showConstructors":{type:"boolean",default:!0,markdownDescription:m(388,"When enabled IntelliSense shows `constructor`-suggestions.")},"editor.suggest.showDeprecated":{type:"boolean",default:!0,markdownDescription:m(389,"When enabled IntelliSense shows `deprecated`-suggestions.")},"editor.suggest.matchOnWordStartOnly":{type:"boolean",default:!0,markdownDescription:m(390,"When enabled IntelliSense filtering requires that the first character matches on a word start. For example, `c` on `Console` or `WebContext` but _not_ on `description`. When disabled IntelliSense will show more results but still sorts them by match quality.")},"editor.suggest.showFields":{type:"boolean",default:!0,markdownDescription:m(391,"When enabled IntelliSense shows `field`-suggestions.")},"editor.suggest.showVariables":{type:"boolean",default:!0,markdownDescription:m(392,"When enabled IntelliSense shows `variable`-suggestions.")},"editor.suggest.showClasses":{type:"boolean",default:!0,markdownDescription:m(393,"When enabled IntelliSense shows `class`-suggestions.")},"editor.suggest.showStructs":{type:"boolean",default:!0,markdownDescription:m(394,"When enabled IntelliSense shows `struct`-suggestions.")},"editor.suggest.showInterfaces":{type:"boolean",default:!0,markdownDescription:m(395,"When enabled IntelliSense shows `interface`-suggestions.")},"editor.suggest.showModules":{type:"boolean",default:!0,markdownDescription:m(396,"When enabled IntelliSense shows `module`-suggestions.")},"editor.suggest.showProperties":{type:"boolean",default:!0,markdownDescription:m(397,"When enabled IntelliSense shows `property`-suggestions.")},"editor.suggest.showEvents":{type:"boolean",default:!0,markdownDescription:m(398,"When enabled IntelliSense shows `event`-suggestions.")},"editor.suggest.showOperators":{type:"boolean",default:!0,markdownDescription:m(399,"When enabled IntelliSense shows `operator`-suggestions.")},"editor.suggest.showUnits":{type:"boolean",default:!0,markdownDescription:m(400,"When enabled IntelliSense shows `unit`-suggestions.")},"editor.suggest.showValues":{type:"boolean",default:!0,markdownDescription:m(401,"When enabled IntelliSense shows `value`-suggestions.")},"editor.suggest.showConstants":{type:"boolean",default:!0,markdownDescription:m(402,"When enabled IntelliSense shows `constant`-suggestions.")},"editor.suggest.showEnums":{type:"boolean",default:!0,markdownDescription:m(403,"When enabled IntelliSense shows `enum`-suggestions.")},"editor.suggest.showEnumMembers":{type:"boolean",default:!0,markdownDescription:m(404,"When enabled IntelliSense shows `enumMember`-suggestions.")},"editor.suggest.showKeywords":{type:"boolean",default:!0,markdownDescription:m(405,"When enabled IntelliSense shows `keyword`-suggestions.")},"editor.suggest.showWords":{type:"boolean",default:!0,markdownDescription:m(406,"When enabled IntelliSense shows `text`-suggestions.")},"editor.suggest.showColors":{type:"boolean",default:!0,markdownDescription:m(407,"When enabled IntelliSense shows `color`-suggestions.")},"editor.suggest.showFiles":{type:"boolean",default:!0,markdownDescription:m(408,"When enabled IntelliSense shows `file`-suggestions.")},"editor.suggest.showReferences":{type:"boolean",default:!0,markdownDescription:m(409,"When enabled IntelliSense shows `reference`-suggestions.")},"editor.suggest.showCustomcolors":{type:"boolean",default:!0,markdownDescription:m(410,"When enabled IntelliSense shows `customcolor`-suggestions.")},"editor.suggest.showFolders":{type:"boolean",default:!0,markdownDescription:m(411,"When enabled IntelliSense shows `folder`-suggestions.")},"editor.suggest.showTypeParameters":{type:"boolean",default:!0,markdownDescription:m(412,"When enabled IntelliSense shows `typeParameter`-suggestions.")},"editor.suggest.showSnippets":{type:"boolean",default:!0,markdownDescription:m(413,"When enabled IntelliSense shows `snippet`-suggestions.")},"editor.suggest.showUsers":{type:"boolean",default:!0,markdownDescription:m(414,"When enabled IntelliSense shows `user`-suggestions.")},"editor.suggest.showIssues":{type:"boolean",default:!0,markdownDescription:m(415,"When enabled IntelliSense shows `issues`-suggestions.")}})}validate(e){if(!e||typeof e!="object")return this.defaultValue;const t=e;return{insertMode:Yt(t.insertMode,this.defaultValue.insertMode,["insert","replace"]),filterGraceful:we(t.filterGraceful,this.defaultValue.filterGraceful),snippetsPreventQuickSuggestions:we(t.snippetsPreventQuickSuggestions,this.defaultValue.filterGraceful),localityBonus:we(t.localityBonus,this.defaultValue.localityBonus),shareSuggestSelections:we(t.shareSuggestSelections,this.defaultValue.shareSuggestSelections),selectionMode:Yt(t.selectionMode,this.defaultValue.selectionMode,["always","never","whenQuickSuggestion","whenTriggerCharacter"]),showIcons:we(t.showIcons,this.defaultValue.showIcons),showStatusBar:we(t.showStatusBar,this.defaultValue.showStatusBar),preview:we(t.preview,this.defaultValue.preview),previewMode:Yt(t.previewMode,this.defaultValue.previewMode,["prefix","subword","subwordSmart"]),showInlineDetails:we(t.showInlineDetails,this.defaultValue.showInlineDetails),showMethods:we(t.showMethods,this.defaultValue.showMethods),showFunctions:we(t.showFunctions,this.defaultValue.showFunctions),showConstructors:we(t.showConstructors,this.defaultValue.showConstructors),showDeprecated:we(t.showDeprecated,this.defaultValue.showDeprecated),matchOnWordStartOnly:we(t.matchOnWordStartOnly,this.defaultValue.matchOnWordStartOnly),showFields:we(t.showFields,this.defaultValue.showFields),showVariables:we(t.showVariables,this.defaultValue.showVariables),showClasses:we(t.showClasses,this.defaultValue.showClasses),showStructs:we(t.showStructs,this.defaultValue.showStructs),showInterfaces:we(t.showInterfaces,this.defaultValue.showInterfaces),showModules:we(t.showModules,this.defaultValue.showModules),showProperties:we(t.showProperties,this.defaultValue.showProperties),showEvents:we(t.showEvents,this.defaultValue.showEvents),showOperators:we(t.showOperators,this.defaultValue.showOperators),showUnits:we(t.showUnits,this.defaultValue.showUnits),showValues:we(t.showValues,this.defaultValue.showValues),showConstants:we(t.showConstants,this.defaultValue.showConstants),showEnums:we(t.showEnums,this.defaultValue.showEnums),showEnumMembers:we(t.showEnumMembers,this.defaultValue.showEnumMembers),showKeywords:we(t.showKeywords,this.defaultValue.showKeywords),showWords:we(t.showWords,this.defaultValue.showWords),showColors:we(t.showColors,this.defaultValue.showColors),showFiles:we(t.showFiles,this.defaultValue.showFiles),showReferences:we(t.showReferences,this.defaultValue.showReferences),showFolders:we(t.showFolders,this.defaultValue.showFolders),showTypeParameters:we(t.showTypeParameters,this.defaultValue.showTypeParameters),showSnippets:we(t.showSnippets,this.defaultValue.showSnippets),showUsers:we(t.showUsers,this.defaultValue.showUsers),showIssues:we(t.showIssues,this.defaultValue.showIssues)}}}class kpe extends ci{constructor(){super(129,"smartSelect",{selectLeadingAndTrailingWhitespace:!0,selectSubwords:!0},{"editor.smartSelect.selectLeadingAndTrailingWhitespace":{description:m(416,"Whether leading and trailing whitespace should always be selected."),default:!0,type:"boolean"},"editor.smartSelect.selectSubwords":{description:m(417,"Whether subwords (like 'foo' in 'fooBar' or 'foo_bar') should be selected."),default:!0,type:"boolean"}})}validate(e){return!e||typeof e!="object"?this.defaultValue:{selectLeadingAndTrailingWhitespace:we(e.selectLeadingAndTrailingWhitespace,this.defaultValue.selectLeadingAndTrailingWhitespace),selectSubwords:we(e.selectSubwords,this.defaultValue.selectSubwords)}}}class Epe extends ci{constructor(){const e=[];super(147,"wordSegmenterLocales",e,{anyOf:[{type:"string"},{type:"array",items:{type:"string"}}],description:m(418,"Locales to be used for word segmentation when doing word related navigations or operations. Specify the BCP 47 language tag of the word you wish to recognize (e.g., ja, zh-CN, zh-Hant-TW, etc.)."),type:"array",items:{type:"string"},default:e})}validate(e){if(typeof e=="string"&&(e=[e]),Array.isArray(e)){const t=[];for(const i of e)if(typeof i=="string")try{Intl.Segmenter.supportedLocalesOf(i).length>0&&t.push(i)}catch{}return t}return this.defaultValue}}class Ipe extends ci{constructor(){super(155,"wrappingIndent",1,{"editor.wrappingIndent":{type:"string",enum:["none","same","indent","deepIndent"],enumDescriptions:[m(419,"No indentation. Wrapped lines begin at column 1."),m(420,"Wrapped lines get the same indentation as the parent."),m(421,"Wrapped lines get +1 indentation toward the parent."),m(422,"Wrapped lines get +2 indentation toward the parent.")],description:m(423,"Controls the indentation of wrapped lines."),default:"same"}})}validate(e){switch(e){case"none":return 0;case"same":return 1;case"indent":return 2;case"deepIndent":return 3}return 1}compute(e,t,i){return t.get(2)===2?0:i}}class Dpe extends Ig{constructor(){super(166,{isDominatedByLongLines:!1,isWordWrapMinified:!1,isViewportWrapping:!1,wrappingColumn:-1})}compute(e,t,i){const n=t.get(165);return{isDominatedByLongLines:e.isDominatedByLongLines,isWordWrapMinified:n.isWordWrapMinified,isViewportWrapping:n.isViewportWrapping,wrappingColumn:n.wrappingColumn}}}class Tpe extends ci{constructor(){const e={enabled:!0,showDropSelector:"afterDrop"};super(43,"dropIntoEditor",e,{"editor.dropIntoEditor.enabled":{type:"boolean",default:e.enabled,markdownDescription:m(424,"Controls whether you can drag and drop a file into a text editor by holding down the `Shift` key (instead of opening the file in an editor).")},"editor.dropIntoEditor.showDropSelector":{type:"string",markdownDescription:m(425,"Controls if a widget is shown when dropping files into the editor. This widget lets you control how the file is dropped."),enum:["afterDrop","never"],enumDescriptions:[m(426,"Show the drop selector widget after a file is dropped into the editor."),m(427,"Never show the drop selector widget. Instead the default drop provider is always used.")],default:"afterDrop"}})}validate(e){if(!e||typeof e!="object")return this.defaultValue;const t=e;return{enabled:we(t.enabled,this.defaultValue.enabled),showDropSelector:Yt(t.showDropSelector,this.defaultValue.showDropSelector,["afterDrop","never"])}}}class Npe extends ci{constructor(){const e={enabled:!0,showPasteSelector:"afterPaste"};super(97,"pasteAs",e,{"editor.pasteAs.enabled":{type:"boolean",default:e.enabled,markdownDescription:m(428,"Controls whether you can paste content in different ways.")},"editor.pasteAs.showPasteSelector":{type:"string",markdownDescription:m(429,"Controls if a widget is shown when pasting content in to the editor. This widget lets you control how the file is pasted."),enum:["afterPaste","never"],enumDescriptions:[m(430,"Show the paste selector widget after content is pasted into the editor."),m(431,"Never show the paste selector widget. Instead the default pasting behavior is always used.")],default:"afterPaste"}})}validate(e){if(!e||typeof e!="object")return this.defaultValue;const t=e;return{enabled:we(t.enabled,this.defaultValue.enabled),showPasteSelector:Yt(t.showPasteSelector,this.defaultValue.showPasteSelector,["afterPaste","never"])}}}const xb=[];function ee(s){return xb[s.id]=s,s}const Is={acceptSuggestionOnCommitCharacter:ee(new Ge(0,"acceptSuggestionOnCommitCharacter",!0,{markdownDescription:m(432,"Controls whether suggestions should be accepted on commit characters. For example, in JavaScript, the semi-colon (`;`) can be a commit character that accepts a suggestion and types that character.")})),acceptSuggestionOnEnter:ee(new ai(1,"acceptSuggestionOnEnter","on",["on","smart","off"],{markdownEnumDescriptions:["",m(433,"Only accept a suggestion with `Enter` when it makes a textual change."),""],markdownDescription:m(434,"Controls whether suggestions should be accepted on `Enter`, in addition to `Tab`. Helps to avoid ambiguity between inserting new lines or accepting suggestions.")})),accessibilitySupport:ee(new qfe),accessibilityPageSize:ee(new Tt(3,"accessibilityPageSize",500,1,1073741824,{description:m(435,"Controls the number of lines in the editor that can be read out by a screen reader at once. When we detect a screen reader we automatically set the default to be 500. Warning: this has a performance implication for numbers larger than the default."),tags:["accessibility"]})),allowOverflow:ee(new Ge(4,"allowOverflow",!0)),allowVariableLineHeights:ee(new Ge(5,"allowVariableLineHeights",!0,{description:m(436,"Controls whether to allow using variable line heights in the editor.")})),allowVariableFonts:ee(new Ge(6,"allowVariableFonts",!0,{description:m(437,"Controls whether to allow using variable fonts in the editor.")})),allowVariableFontsInAccessibilityMode:ee(new Ge(7,"allowVariableFontsInAccessibilityMode",!1,{description:m(438,"Controls whether to allow using variable fonts in the editor in the accessibility mode."),tags:["accessibility"]})),ariaLabel:ee(new Es(8,"ariaLabel",m(439,"Editor content"))),ariaRequired:ee(new Ge(9,"ariaRequired",!1,void 0)),screenReaderAnnounceInlineSuggestion:ee(new Ge(12,"screenReaderAnnounceInlineSuggestion",!0,{description:m(440,"Control whether inline suggestions are announced by a screen reader."),tags:["accessibility"]})),autoClosingBrackets:ee(new ai(10,"autoClosingBrackets","languageDefined",["always","languageDefined","beforeWhitespace","never"],{enumDescriptions:["",m(441,"Use language configurations to determine when to autoclose brackets."),m(442,"Autoclose brackets only when the cursor is to the left of whitespace."),""],description:m(443,"Controls whether the editor should automatically close brackets after the user adds an opening bracket.")})),autoClosingComments:ee(new ai(11,"autoClosingComments","languageDefined",["always","languageDefined","beforeWhitespace","never"],{enumDescriptions:["",m(444,"Use language configurations to determine when to autoclose comments."),m(445,"Autoclose comments only when the cursor is to the left of whitespace."),""],description:m(446,"Controls whether the editor should automatically close comments after the user adds an opening comment.")})),autoClosingDelete:ee(new ai(13,"autoClosingDelete","auto",["always","auto","never"],{enumDescriptions:["",m(447,"Remove adjacent closing quotes or brackets only if they were automatically inserted."),""],description:m(448,"Controls whether the editor should remove adjacent closing quotes or brackets when deleting.")})),autoClosingOvertype:ee(new ai(14,"autoClosingOvertype","auto",["always","auto","never"],{enumDescriptions:["",m(449,"Type over closing quotes or brackets only if they were automatically inserted."),""],description:m(450,"Controls whether the editor should type over closing quotes or brackets.")})),autoClosingQuotes:ee(new ai(15,"autoClosingQuotes","languageDefined",["always","languageDefined","beforeWhitespace","never"],{enumDescriptions:["",m(451,"Use language configurations to determine when to autoclose quotes."),m(452,"Autoclose quotes only when the cursor is to the left of whitespace."),""],description:m(453,"Controls whether the editor should automatically close quotes after the user adds an opening quote.")})),autoIndent:ee(new kC(16,"autoIndent",4,"full",["none","keep","brackets","advanced","full"],Ufe,{enumDescriptions:[m(454,"The editor will not insert indentation automatically."),m(455,"The editor will keep the current line's indentation."),m(456,"The editor will keep the current line's indentation and honor language defined brackets."),m(457,"The editor will keep the current line's indentation, honor language defined brackets and invoke special onEnterRules defined by languages."),m(458,"The editor will keep the current line's indentation, honor language defined brackets, invoke special onEnterRules defined by languages, and honor indentationRules defined by languages.")],description:m(459,"Controls whether the editor should automatically adjust the indentation when users type, paste, move or indent lines.")})),autoIndentOnPaste:ee(new Ge(17,"autoIndentOnPaste",!1,{description:m(460,"Controls whether the editor should automatically auto-indent the pasted content.")})),autoIndentOnPasteWithinString:ee(new Ge(18,"autoIndentOnPasteWithinString",!0,{description:m(461,"Controls whether the editor should automatically auto-indent the pasted content when pasted within a string. This takes effect when autoIndentOnPaste is true.")})),automaticLayout:ee(new Ge(19,"automaticLayout",!1)),autoSurround:ee(new ai(20,"autoSurround","languageDefined",["languageDefined","quotes","brackets","never"],{enumDescriptions:[m(462,"Use language configurations to determine when to automatically surround selections."),m(463,"Surround with quotes but not brackets."),m(464,"Surround with brackets but not quotes."),""],description:m(465,"Controls whether the editor should automatically surround selections when typing quotes or brackets.")})),bracketPairColorization:ee(new Spe),bracketPairGuides:ee(new xpe),stickyTabStops:ee(new Ge(132,"stickyTabStops",!1,{description:m(466,"Emulate selection behavior of tab characters when using spaces for indentation. Selection will stick to tab stops.")})),codeLens:ee(new Ge(23,"codeLens",!0,{description:m(467,"Controls whether the editor shows CodeLens.")})),codeLensFontFamily:ee(new Es(24,"codeLensFontFamily","",{description:m(468,"Controls the font family for CodeLens.")})),codeLensFontSize:ee(new Tt(25,"codeLensFontSize",0,0,100,{type:"number",default:0,minimum:0,maximum:100,markdownDescription:m(469,"Controls the font size in pixels for CodeLens. When set to 0, 90% of `#editor.fontSize#` is used.")})),colorDecorators:ee(new Ge(26,"colorDecorators",!0,{description:m(470,"Controls whether the editor should render the inline color decorators and color picker.")})),colorDecoratorActivatedOn:ee(new ai(168,"colorDecoratorsActivatedOn","clickAndHover",["clickAndHover","hover","click"],{enumDescriptions:[m(471,"Make the color picker appear both on click and hover of the color decorator"),m(472,"Make the color picker appear on hover of the color decorator"),m(473,"Make the color picker appear on click of the color decorator")],description:m(474,"Controls the condition to make a color picker appear from a color decorator.")})),colorDecoratorsLimit:ee(new Tt(27,"colorDecoratorsLimit",500,1,1e6,{markdownDescription:m(475,"Controls the max number of color decorators that can be rendered in an editor at once.")})),columnSelection:ee(new Ge(28,"columnSelection",!1,{description:m(476,"Enable that the selection with the mouse and keys is doing column selection.")})),comments:ee(new jfe),contextmenu:ee(new Ge(30,"contextmenu",!0)),copyWithSyntaxHighlighting:ee(new Ge(31,"copyWithSyntaxHighlighting",!0,{description:m(477,"Controls whether syntax highlighting should be copied into the clipboard.")})),cursorBlinking:ee(new kC(32,"cursorBlinking",1,"blink",["blink","smooth","phase","expand","solid"],Gfe,{description:m(478,"Control the cursor animation style.")})),cursorSmoothCaretAnimation:ee(new ai(33,"cursorSmoothCaretAnimation","off",["off","explicit","on"],{enumDescriptions:[m(479,"Smooth caret animation is disabled."),m(480,"Smooth caret animation is enabled only when the user moves the cursor with an explicit gesture."),m(481,"Smooth caret animation is always enabled.")],description:m(482,"Controls whether the smooth caret animation should be enabled.")})),cursorStyle:ee(new kC(34,"cursorStyle",mn.Line,"line",["line","block","underline","line-thin","block-outline","underline-thin"],k$,{description:m(483,"Controls the cursor style in insert input mode.")})),overtypeCursorStyle:ee(new kC(92,"overtypeCursorStyle",mn.Block,"block",["line","block","underline","line-thin","block-outline","underline-thin"],k$,{description:m(484,"Controls the cursor style in overtype input mode.")})),cursorSurroundingLines:ee(new Tt(35,"cursorSurroundingLines",0,0,1073741824,{description:m(485,"Controls the minimal number of visible leading lines (minimum 0) and trailing lines (minimum 1) surrounding the cursor. Known as 'scrollOff' or 'scrollOffset' in some other editors.")})),cursorSurroundingLinesStyle:ee(new ai(36,"cursorSurroundingLinesStyle","default",["default","all"],{enumDescriptions:[m(486,"`cursorSurroundingLines` is enforced only when triggered via the keyboard or API."),m(487,"`cursorSurroundingLines` is enforced always.")],markdownDescription:m(488,"Controls when `#editor.cursorSurroundingLines#` should be enforced.")})),cursorWidth:ee(new Tt(37,"cursorWidth",0,0,1073741824,{markdownDescription:m(489,"Controls the width of the cursor when `#editor.cursorStyle#` is set to `line`.")})),cursorHeight:ee(new Tt(38,"cursorHeight",0,0,1073741824,{markdownDescription:m(490,"Controls the height of the cursor when `#editor.cursorStyle#` is set to `line`. Cursor's max height depends on line height.")})),disableLayerHinting:ee(new Ge(39,"disableLayerHinting",!1)),disableMonospaceOptimizations:ee(new Ge(40,"disableMonospaceOptimizations",!1)),domReadOnly:ee(new Ge(41,"domReadOnly",!1)),dragAndDrop:ee(new Ge(42,"dragAndDrop",!0,{description:m(491,"Controls whether the editor should allow moving selections via drag and drop.")})),emptySelectionClipboard:ee(new Yfe),dropIntoEditor:ee(new Tpe),editContext:ee(new Ge(44,"editContext",!0,{description:m(492,"Sets whether the EditContext API should be used instead of the text area to power input in the editor."),included:B6||zX||Sl})),renderRichScreenReaderContent:ee(new Ge(107,"renderRichScreenReaderContent",!1,{markdownDescription:m(493,"Whether to render rich screen reader content when the `#editor.editContext#` setting is enabled.")})),stickyScroll:ee(new rpe),experimentalGpuAcceleration:ee(new ai(46,"experimentalGpuAcceleration","off",["off","on"],{tags:["experimental"],enumDescriptions:[m(494,"Use regular DOM-based rendering."),m(495,"Use GPU acceleration.")],description:m(496,"Controls whether to use the experimental GPU acceleration to render the editor.")})),experimentalWhitespaceRendering:ee(new ai(47,"experimentalWhitespaceRendering","svg",["svg","font","off"],{enumDescriptions:[m(497,"Use a new rendering method with svgs."),m(498,"Use a new rendering method with font characters."),m(499,"Use the stable rendering method.")],description:m(500,"Controls whether whitespace is rendered with a new, experimental method.")})),extraEditorClassName:ee(new Es(48,"extraEditorClassName","")),fastScrollSensitivity:ee(new po(49,"fastScrollSensitivity",5,s=>s<=0?5:s,{markdownDescription:m(501,"Scrolling speed multiplier when pressing `Alt`.")})),find:ee(new Zfe),fixedOverflowWidgets:ee(new Ge(51,"fixedOverflowWidgets",!1)),folding:ee(new Ge(52,"folding",!0,{description:m(502,"Controls whether the editor has code folding enabled.")})),foldingStrategy:ee(new ai(53,"foldingStrategy","auto",["auto","indentation"],{enumDescriptions:[m(503,"Use a language-specific folding strategy if available, else the indentation-based one."),m(504,"Use the indentation-based folding strategy.")],description:m(505,"Controls the strategy for computing folding ranges.")})),foldingHighlight:ee(new Ge(54,"foldingHighlight",!0,{description:m(506,"Controls whether the editor should highlight folded ranges.")})),foldingImportsByDefault:ee(new Ge(55,"foldingImportsByDefault",!1,{description:m(507,"Controls whether the editor automatically collapses import ranges.")})),foldingMaximumRegions:ee(new Tt(56,"foldingMaximumRegions",5e3,10,65e3,{description:m(508,"The maximum number of foldable regions. Increasing this value may result in the editor becoming less responsive when the current source has a large number of foldable regions.")})),unfoldOnClickAfterEndOfLine:ee(new Ge(57,"unfoldOnClickAfterEndOfLine",!1,{description:m(509,"Controls whether clicking on the empty content after a folded line will unfold the line.")})),fontFamily:ee(new Es(58,"fontFamily",_o.fontFamily,{description:m(510,"Controls the font family.")})),fontInfo:ee(new Xfe),fontLigatures2:ee(new mh),fontSize:ee(new tpe),fontWeight:ee(new $P),fontVariations:ee(new zP),formatOnPaste:ee(new Ge(64,"formatOnPaste",!1,{description:m(511,"Controls whether the editor should automatically format the pasted content. A formatter must be available and the formatter should be able to format a range in a document.")})),formatOnType:ee(new Ge(65,"formatOnType",!1,{description:m(512,"Controls whether the editor should automatically format the line after typing.")})),glyphMargin:ee(new Ge(66,"glyphMargin",!0,{description:m(513,"Controls whether the editor should render the vertical glyph margin. Glyph margin is mostly used for debugging.")})),gotoLocation:ee(new ipe),hideCursorInOverviewRuler:ee(new Ge(68,"hideCursorInOverviewRuler",!1,{description:m(514,"Controls whether the cursor should be hidden in the overview ruler.")})),hover:ee(new npe),inDiffEditor:ee(new Ge(70,"inDiffEditor",!1)),inertialScroll:ee(new Ge(158,"inertialScroll",!1,{description:m(515,"Make scrolling inertial - mostly useful with touchpad on linux.")})),letterSpacing:ee(new po(72,"letterSpacing",_o.letterSpacing,s=>po.clamp(s,-5,20),{description:m(516,"Controls the letter spacing in pixels.")})),lightbulb:ee(new ope),lineDecorationsWidth:ee(new lpe),lineHeight:ee(new cpe),lineNumbers:ee(new _pe),lineNumbersMinChars:ee(new Tt(77,"lineNumbersMinChars",5,1,300)),linkedEditing:ee(new Ge(78,"linkedEditing",!1,{description:m(517,"Controls whether the editor has linked editing enabled. Depending on the language, related symbols such as HTML tags, are updated while editing.")})),links:ee(new Ge(79,"links",!0,{description:m(518,"Controls whether the editor should detect links and make them clickable.")})),matchBrackets:ee(new ai(80,"matchBrackets","always",["always","near","never"],{description:m(519,"Highlight matching brackets.")})),minimap:ee(new dpe),mouseStyle:ee(new ai(82,"mouseStyle","text",["text","default","copy"])),mouseWheelScrollSensitivity:ee(new po(83,"mouseWheelScrollSensitivity",1,s=>s===0?1:s,{markdownDescription:m(520,"A multiplier to be used on the `deltaX` and `deltaY` of mouse wheel scroll events.")})),mouseWheelZoom:ee(new Ge(84,"mouseWheelZoom",!1,{markdownDescription:Ye?m(521,"Zoom the font of the editor when using mouse wheel and holding `Cmd`."):m(522,"Zoom the font of the editor when using mouse wheel and holding `Ctrl`.")})),multiCursorMergeOverlapping:ee(new Ge(85,"multiCursorMergeOverlapping",!0,{description:m(523,"Merge multiple cursors when they are overlapping.")})),multiCursorModifier:ee(new kC(86,"multiCursorModifier","altKey","alt",["ctrlCmd","alt"],hpe,{markdownEnumDescriptions:[m(524,"Maps to `Control` on Windows and Linux and to `Command` on macOS."),m(525,"Maps to `Alt` on Windows and Linux and to `Option` on macOS.")],markdownDescription:m(526,"The modifier to be used to add multiple cursors with the mouse. The Go to Definition and Open Link mouse gestures will adapt such that they do not conflict with the [multicursor modifier](https://code.visualstudio.com/docs/editor/codebasics#_multicursor-modifier).")})),mouseMiddleClickAction:ee(new ai(87,"mouseMiddleClickAction","default",["default","openLink","ctrlLeftClick"],{description:m(527,"Controls what happens when middle mouse button is clicked in the editor.")})),multiCursorPaste:ee(new ai(88,"multiCursorPaste","spread",["spread","full"],{markdownEnumDescriptions:[m(528,"Each cursor pastes a single line of the text."),m(529,"Each cursor pastes the full text.")],markdownDescription:m(530,"Controls pasting when the line count of the pasted text matches the cursor count.")})),multiCursorLimit:ee(new Tt(89,"multiCursorLimit",1e4,1,1e5,{markdownDescription:m(531,"Controls the max number of cursors that can be in an active editor at once.")})),occurrencesHighlight:ee(new ai(90,"occurrencesHighlight","singleFile",["off","singleFile","multiFile"],{markdownEnumDescriptions:[m(532,"Does not highlight occurrences."),m(533,"Highlights occurrences only in the current file."),m(534,"Experimental: Highlights occurrences across all valid open files.")],markdownDescription:m(535,"Controls whether occurrences should be highlighted across open files.")})),occurrencesHighlightDelay:ee(new Tt(91,"occurrencesHighlightDelay",0,0,2e3,{description:m(536,"Controls the delay in milliseconds after which occurrences are highlighted."),tags:["preview"]})),overtypeOnPaste:ee(new Ge(93,"overtypeOnPaste",!0,{description:m(537,"Controls whether pasting should overtype.")})),overviewRulerBorder:ee(new Ge(94,"overviewRulerBorder",!0,{description:m(538,"Controls whether a border should be drawn around the overview ruler.")})),overviewRulerLanes:ee(new Tt(95,"overviewRulerLanes",3,0,3)),padding:ee(new upe),pasteAs:ee(new Npe),parameterHints:ee(new gpe),peekWidgetDefaultFocus:ee(new ai(99,"peekWidgetDefaultFocus","tree",["tree","editor"],{enumDescriptions:[m(539,"Focus the tree when opening peek"),m(540,"Focus the editor when opening peek")],description:m(541,"Controls whether to focus the inline editor or the tree in the peek widget.")})),placeholder:ee(new ppe),definitionLinkOpensInPeek:ee(new Ge(101,"definitionLinkOpensInPeek",!1,{description:m(542,"Controls whether the Go to Definition mouse gesture always opens the peek widget.")})),quickSuggestions:ee(new mpe),quickSuggestionsDelay:ee(new Tt(103,"quickSuggestionsDelay",10,0,1073741824,{description:m(543,"Controls the delay in milliseconds after which quick suggestions will show up."),experiment:{mode:"auto"}})),readOnly:ee(new Ge(104,"readOnly",!1)),readOnlyMessage:ee(new wpe),renameOnType:ee(new Ge(106,"renameOnType",!1,{description:m(544,"Controls whether the editor auto renames on type."),markdownDeprecationMessage:m(545,"Deprecated, use `#editor.linkedEditing#` instead.")})),renderControlCharacters:ee(new Ge(108,"renderControlCharacters",!0,{description:m(546,"Controls whether the editor should render control characters."),restricted:!0})),renderFinalNewline:ee(new ai(109,"renderFinalNewline",vo?"dimmed":"on",["off","on","dimmed"],{description:m(547,"Render last line number when the file ends with a newline.")})),renderLineHighlight:ee(new ai(110,"renderLineHighlight","line",["none","gutter","line","all"],{enumDescriptions:["","","",m(548,"Highlights both the gutter and the current line.")],description:m(549,"Controls how the editor should render the current line highlight.")})),renderLineHighlightOnlyWhenFocus:ee(new Ge(111,"renderLineHighlightOnlyWhenFocus",!1,{description:m(550,"Controls if the editor should render the current line highlight only when the editor is focused.")})),renderValidationDecorations:ee(new ai(112,"renderValidationDecorations","editable",["editable","on","off"])),renderWhitespace:ee(new ai(113,"renderWhitespace","selection",["none","boundary","selection","trailing","all"],{enumDescriptions:["",m(551,"Render whitespace characters except for single spaces between words."),m(552,"Render whitespace characters only on selected text."),m(553,"Render only trailing whitespace characters."),""],description:m(554,"Controls how the editor should render whitespace characters.")})),revealHorizontalRightPadding:ee(new Tt(114,"revealHorizontalRightPadding",15,0,1e3)),roundedSelection:ee(new Ge(115,"roundedSelection",!0,{description:m(555,"Controls whether selections should have rounded corners.")})),rulers:ee(new bpe),scrollbar:ee(new vpe),scrollBeyondLastColumn:ee(new Tt(118,"scrollBeyondLastColumn",4,0,1073741824,{description:m(556,"Controls the number of extra characters beyond which the editor will scroll horizontally.")})),scrollBeyondLastLine:ee(new Ge(119,"scrollBeyondLastLine",!0,{description:m(557,"Controls whether the editor will scroll beyond the last line.")})),scrollOnMiddleClick:ee(new Ge(171,"scrollOnMiddleClick",!1,{description:m(558,"Controls whether the editor will scroll when the middle button is pressed.")})),scrollPredominantAxis:ee(new Ge(120,"scrollPredominantAxis",!0,{description:m(559,"Scroll only along the predominant axis when scrolling both vertically and horizontally at the same time. Prevents horizontal drift when scrolling vertically on a trackpad.")})),selectionClipboard:ee(new Ge(121,"selectionClipboard",!0,{description:m(560,"Controls whether the Linux primary clipboard should be supported."),included:vo})),selectionHighlight:ee(new Ge(122,"selectionHighlight",!0,{description:m(561,"Controls whether the editor should highlight matches similar to the selection.")})),selectionHighlightMaxLength:ee(new Tt(123,"selectionHighlightMaxLength",200,0,1073741824,{description:m(562,"Controls how many characters can be in the selection before similiar matches are not highlighted. Set to zero for unlimited.")})),selectionHighlightMultiline:ee(new Ge(124,"selectionHighlightMultiline",!1,{description:m(563,"Controls whether the editor should highlight selection matches that span multiple lines.")})),selectOnLineNumbers:ee(new Ge(125,"selectOnLineNumbers",!0)),showFoldingControls:ee(new ai(126,"showFoldingControls","mouseover",["always","never","mouseover"],{enumDescriptions:[m(564,"Always show the folding controls."),m(565,"Never show the folding controls and reduce the gutter size."),m(566,"Only show the folding controls when the mouse is over the gutter.")],description:m(567,"Controls when the folding controls on the gutter are shown.")})),showUnused:ee(new Ge(127,"showUnused",!0,{description:m(568,"Controls fading out of unused code.")})),showDeprecated:ee(new Ge(157,"showDeprecated",!0,{description:m(569,"Controls strikethrough deprecated variables.")})),inlayHints:ee(new ape),snippetSuggestions:ee(new ai(128,"snippetSuggestions","inline",["top","bottom","inline","none"],{enumDescriptions:[m(570,"Show snippet suggestions on top of other suggestions."),m(571,"Show snippet suggestions below other suggestions."),m(572,"Show snippets suggestions with other suggestions."),m(573,"Do not show snippet suggestions.")],description:m(574,"Controls whether snippets are shown with other suggestions and how they are sorted.")})),smartSelect:ee(new kpe),smoothScrolling:ee(new Ge(130,"smoothScrolling",!1,{description:m(575,"Controls whether the editor will scroll using an animation.")})),stopRenderingLineAfter:ee(new Tt(133,"stopRenderingLineAfter",1e4,-1,1073741824)),suggest:ee(new Lpe),inlineSuggest:ee(new ype),inlineCompletionsAccessibilityVerbose:ee(new Ge(169,"inlineCompletionsAccessibilityVerbose",!1,{description:m(576,"Controls whether the accessibility hint should be provided to screen reader users when an inline completion is shown.")})),suggestFontSize:ee(new Tt(135,"suggestFontSize",0,0,1e3,{markdownDescription:m(577,"Font size for the suggest widget. When set to {0}, the value of {1} is used.","`0`","`#editor.fontSize#`")})),suggestLineHeight:ee(new Tt(136,"suggestLineHeight",0,0,1e3,{markdownDescription:m(578,"Line height for the suggest widget. When set to {0}, the value of {1} is used. The minimum value is 8.","`0`","`#editor.lineHeight#`")})),suggestOnTriggerCharacters:ee(new Ge(137,"suggestOnTriggerCharacters",!0,{description:m(579,"Controls whether suggestions should automatically show up when typing trigger characters.")})),suggestSelection:ee(new ai(138,"suggestSelection","first",["first","recentlyUsed","recentlyUsedByPrefix"],{markdownEnumDescriptions:[m(580,"Always select the first suggestion."),m(581,"Select recent suggestions unless further typing selects one, e.g. `console.| -> console.log` because `log` has been completed recently."),m(582,"Select suggestions based on previous prefixes that have completed those suggestions, e.g. `co -> console` and `con -> const`.")],description:m(583,"Controls how suggestions are pre-selected when showing the suggest list.")})),tabCompletion:ee(new ai(139,"tabCompletion","off",["on","off","onlySnippets"],{enumDescriptions:[m(584,"Tab complete will insert the best matching suggestion when pressing tab."),m(585,"Disable tab completions."),m(586,"Tab complete snippets when their prefix match. Works best when 'quickSuggestions' aren't enabled.")],description:m(587,"Enables tab completions.")})),tabIndex:ee(new Tt(140,"tabIndex",0,-1,1073741824)),trimWhitespaceOnDelete:ee(new Ge(141,"trimWhitespaceOnDelete",!1,{description:m(588,"Controls whether the editor will also delete the next line's indentation whitespace when deleting a newline.")})),unicodeHighlight:ee(new Cpe),unusualLineTerminators:ee(new ai(143,"unusualLineTerminators","prompt",["auto","off","prompt"],{enumDescriptions:[m(589,"Unusual line terminators are automatically removed."),m(590,"Unusual line terminators are ignored."),m(591,"Unusual line terminators prompt to be removed.")],description:m(592,"Remove unusual line terminators that might cause problems.")})),useShadowDOM:ee(new Ge(144,"useShadowDOM",!0)),useTabStops:ee(new Ge(145,"useTabStops",!0,{description:m(593,"Spaces and tabs are inserted and deleted in alignment with tab stops.")})),wordBreak:ee(new ai(146,"wordBreak","normal",["normal","keepAll"],{markdownEnumDescriptions:[m(594,"Use the default line break rule."),m(595,"Word breaks should not be used for Chinese/Japanese/Korean (CJK) text. Non-CJK text behavior is the same as for normal.")],description:m(596,"Controls the word break rules used for Chinese/Japanese/Korean (CJK) text.")})),wordSegmenterLocales:ee(new Epe),wordSeparators:ee(new Es(148,"wordSeparators",RE,{description:m(597,"Characters that will be used as word separators when doing word related navigations or operations.")})),wordWrap:ee(new ai(149,"wordWrap","off",["off","on","wordWrapColumn","bounded"],{markdownEnumDescriptions:[m(598,"Lines will never wrap."),m(599,"Lines will wrap at the viewport width."),m(600,"Lines will wrap at `#editor.wordWrapColumn#`."),m(601,"Lines will wrap at the minimum of viewport and `#editor.wordWrapColumn#`.")],description:m(602,"Controls how lines should wrap.")})),wordWrapBreakAfterCharacters:ee(new Es(150,"wordWrapBreakAfterCharacters"," 	})]?|/&.,;¢°′″‰℃、。｡､￠，．：；？！％・･ゝゞヽヾーァィゥェォッャュョヮヵヶぁぃぅぇぉっゃゅょゎゕゖㇰㇱㇲㇳㇴㇵㇶㇷㇸㇹㇺㇻㇼㇽㇾㇿ々〻ｧｨｩｪｫｬｭｮｯｰ”〉》」』】〕）］｝｣")),wordWrapBreakBeforeCharacters:ee(new Es(151,"wordWrapBreakBeforeCharacters","([{‘“〈《「『【〔（［｛｢£¥＄￡￥+＋")),wordWrapColumn:ee(new Tt(152,"wordWrapColumn",80,1,1073741824,{markdownDescription:m(603,"Controls the wrapping column of the editor when `#editor.wordWrap#` is `wordWrapColumn` or `bounded`.")})),wordWrapOverride1:ee(new ai(153,"wordWrapOverride1","inherit",["off","on","inherit"])),wordWrapOverride2:ee(new ai(154,"wordWrapOverride2","inherit",["off","on","inherit"])),wrapOnEscapedLineFeeds:ee(new Ge(160,"wrapOnEscapedLineFeeds",!1,{markdownDescription:m(604,'Controls whether literal `\\n` shall trigger a wordWrap when `#editor.wordWrap#` is enabled.\n\nFor example:\n```c\nchar* str="hello\\nworld"\n```\nwill be displayed as\n```c\nchar* str="hello\\n\n           world"\n```')})),effectiveCursorStyle:ee(new Qfe),editorClassName:ee(new Kfe),defaultColorDecorators:ee(new ai(167,"defaultColorDecorators","auto",["auto","always","never"],{enumDescriptions:[m(605,"Show default color decorators only when no extension provides colors decorators."),m(606,"Always show default color decorators."),m(607,"Never show default color decorators.")],description:m(608,"Controls whether inline color decorations should be shown using the default document color provider.")})),pixelRatio:ee(new fpe),tabFocusMode:ee(new Ge(164,"tabFocusMode",!1,{markdownDescription:m(609,"Controls whether the editor receives tabs or defers them to the workbench for navigation.")})),layoutInfo:ee(new uw),wrappingInfo:ee(new Dpe),wrappingIndent:ee(new Ipe),wrappingStrategy:ee(new spe),effectiveEditContextEnabled:ee(new Jfe),effectiveAllowVariableFonts:ee(new epe)};class Rpe extends F{constructor(){super(...arguments),this._cache=new Map,this._evictUntrustedReadingsTimeout=-1,this._onDidChange=this._register(new M),this.onDidChange=this._onDidChange.event}dispose(){this._evictUntrustedReadingsTimeout!==-1&&(clearTimeout(this._evictUntrustedReadingsTimeout),this._evictUntrustedReadingsTimeout=-1),super.dispose()}clearAllFontInfos(){this._cache.clear(),this._onDidChange.fire()}_ensureCache(e){const t=Zy(e);let i=this._cache.get(t);return i||(i=new Ape,this._cache.set(t,i)),i}_writeToCache(e,t,i){this._ensureCache(e).put(t,i),!i.isTrusted&&this._evictUntrustedReadingsTimeout===-1&&(this._evictUntrustedReadingsTimeout=e.setTimeout(()=>{this._evictUntrustedReadingsTimeout=-1,this._evictUntrustedReadings(e)},5e3))}_evictUntrustedReadings(e){const t=this._ensureCache(e),i=t.getValues();let n=!1;for(const o of i)o.isTrusted||(n=!0,t.remove(o));n&&this._onDidChange.fire()}readFontInfo(e,t){const i=this._ensureCache(e);if(!i.has(t)){let n=this._actualReadFontInfo(e,t);(n.typicalHalfwidthCharacterWidth<=2||n.typicalFullwidthCharacterWidth<=2||n.spaceWidth<=2||n.maxDigitWidth<=2)&&(n=new BE({pixelRatio:sS.getInstance(e).value,fontFamily:n.fontFamily,fontWeight:n.fontWeight,fontSize:n.fontSize,fontFeatureSettings:n.fontFeatureSettings,fontVariationSettings:n.fontVariationSettings,lineHeight:n.lineHeight,letterSpacing:n.letterSpacing,isMonospace:n.isMonospace,typicalHalfwidthCharacterWidth:Math.max(n.typicalHalfwidthCharacterWidth,5),typicalFullwidthCharacterWidth:Math.max(n.typicalFullwidthCharacterWidth,5),canUseHalfwidthRightwardsArrow:n.canUseHalfwidthRightwardsArrow,spaceWidth:Math.max(n.spaceWidth,5),middotWidth:Math.max(n.middotWidth,5),wsmiddotWidth:Math.max(n.wsmiddotWidth,5),maxDigitWidth:Math.max(n.maxDigitWidth,5)},!1)),this._writeToCache(e,t,n)}return i.get(t)}_createRequest(e,t,i,n){const o=new Pfe(e,t);return i.push(o),n?.push(o),o}_actualReadFontInfo(e,t){const i=[],n=[],o=this._createRequest("n",0,i,n),r=this._createRequest("ｍ",0,i,null),a=this._createRequest(" ",0,i,n),l=this._createRequest("0",0,i,n),c=this._createRequest("1",0,i,n),d=this._createRequest("2",0,i,n),h=this._createRequest("3",0,i,n),u=this._createRequest("4",0,i,n),g=this._createRequest("5",0,i,n),f=this._createRequest("6",0,i,n),p=this._createRequest("7",0,i,n),_=this._createRequest("8",0,i,n),b=this._createRequest("9",0,i,n),w=this._createRequest("→",0,i,n),C=this._createRequest("￫",0,i,null),v=this._createRequest("·",0,i,n),y=this._createRequest("⸱",0,i,null),x="|/-_ilm%";for(let O=0,P=x.length;O<P;O++)this._createRequest(x.charAt(O),0,i,n),this._createRequest(x.charAt(O),1,i,n),this._createRequest(x.charAt(O),2,i,n);Ffe(e,t,i);const L=Math.max(l.width,c.width,d.width,h.width,u.width,g.width,f.width,p.width,_.width,b.width);let E=t.fontFeatureSettings===mh.OFF;const I=n[0].width;for(let O=1,P=n.length;E&&O<P;O++){const B=I-n[O].width;if(B<-.001||B>.001){E=!1;break}}let R=!0;return E&&C.width!==I&&(R=!1),C.width>w.width&&(R=!1),new BE({pixelRatio:sS.getInstance(e).value,fontFamily:t.fontFamily,fontWeight:t.fontWeight,fontSize:t.fontSize,fontFeatureSettings:t.fontFeatureSettings,fontVariationSettings:t.fontVariationSettings,lineHeight:t.lineHeight,letterSpacing:t.letterSpacing,isMonospace:E,typicalHalfwidthCharacterWidth:o.width,typicalFullwidthCharacterWidth:r.width,canUseHalfwidthRightwardsArrow:R,spaceWidth:a.width,middotWidth:v.width,wsmiddotWidth:y.width,maxDigitWidth:L},!0)}}class Ape{constructor(){this._keys=Object.create(null),this._values=Object.create(null)}has(e){const t=e.getId();return!!this._values[t]}get(e){const t=e.getId();return this._values[t]}put(e,t){const i=e.getId();this._keys[i]=e,this._values[i]=t}remove(e){const t=e.getId();delete this._keys[t],delete this._values[t]}getValues(){return Object.keys(this._keys).map(e=>this._values[e])}}const UP=new Rpe,Yb=class Yb{constructor(e,t){this.key=e,this.migrate=t}apply(e){const t=Yb._read(e,this.key),i=o=>Yb._read(e,o),n=(o,r)=>Yb._write(e,o,r);this.migrate(t,i,n)}static _read(e,t){if(typeof e>"u"||e===null)return;const i=t.indexOf(".");if(i>=0){const n=t.substring(0,i);return this._read(e[n],t.substring(i+1))}return e[t]}static _write(e,t,i){const n=t.indexOf(".");if(n>=0){const o=t.substring(0,n);e[o]=e[o]||{},this._write(e[o],t.substring(n+1),i);return}e[t]=i}};Yb.items=[];let oS=Yb;function ha(s,e){oS.items.push(new oS(s,e))}function io(s,e){ha(s,(t,i,n)=>{if(typeof t<"u"){for(const[o,r]of e)if(t===o){n(s,r);return}}})}function Mpe(s){oS.items.forEach(e=>e.apply(s))}io("wordWrap",[[!0,"on"],[!1,"off"]]);io("lineNumbers",[[!0,"on"],[!1,"off"]]);io("cursorBlinking",[["visible","solid"]]);io("renderWhitespace",[[!0,"boundary"],[!1,"none"]]);io("renderLineHighlight",[[!0,"line"],[!1,"none"]]);io("acceptSuggestionOnEnter",[[!0,"on"],[!1,"off"]]);io("tabCompletion",[[!1,"off"],[!0,"onlySnippets"]]);io("hover",[[!0,{enabled:!0}],[!1,{enabled:!1}]]);io("parameterHints",[[!0,{enabled:!0}],[!1,{enabled:!1}]]);io("autoIndent",[[!1,"advanced"],[!0,"full"]]);io("matchBrackets",[[!0,"always"],[!1,"never"]]);io("renderFinalNewline",[[!0,"on"],[!1,"off"]]);io("cursorSmoothCaretAnimation",[[!0,"on"],[!1,"off"]]);io("occurrencesHighlight",[[!0,"singleFile"],[!1,"off"]]);io("wordBasedSuggestions",[[!0,"matchingDocuments"],[!1,"off"]]);io("defaultColorDecorators",[[!0,"auto"],[!1,"never"]]);io("minimap.autohide",[[!0,"mouseover"],[!1,"none"]]);ha("autoClosingBrackets",(s,e,t)=>{s===!1&&(t("autoClosingBrackets","never"),typeof e("autoClosingQuotes")>"u"&&t("autoClosingQuotes","never"),typeof e("autoSurround")>"u"&&t("autoSurround","never"))});ha("renderIndentGuides",(s,e,t)=>{typeof s<"u"&&(t("renderIndentGuides",void 0),typeof e("guides.indentation")>"u"&&t("guides.indentation",!!s))});ha("highlightActiveIndentGuide",(s,e,t)=>{typeof s<"u"&&(t("highlightActiveIndentGuide",void 0),typeof e("guides.highlightActiveIndentation")>"u"&&t("guides.highlightActiveIndentation",!!s))});const Ope={method:"showMethods",function:"showFunctions",constructor:"showConstructors",deprecated:"showDeprecated",field:"showFields",variable:"showVariables",class:"showClasses",struct:"showStructs",interface:"showInterfaces",module:"showModules",property:"showProperties",event:"showEvents",operator:"showOperators",unit:"showUnits",value:"showValues",constant:"showConstants",enum:"showEnums",enumMember:"showEnumMembers",keyword:"showKeywords",text:"showWords",color:"showColors",file:"showFiles",reference:"showReferences",folder:"showFolders",typeParameter:"showTypeParameters",snippet:"showSnippets"};ha("suggest.filteredTypes",(s,e,t)=>{if(s&&typeof s=="object"){for(const i of Object.entries(Ope))s[i[0]]===!1&&typeof e(`suggest.${i[1]}`)>"u"&&t(`suggest.${i[1]}`,!1);t("suggest.filteredTypes",void 0)}});ha("quickSuggestions",(s,e,t)=>{if(typeof s=="boolean"){const i=s?"on":"off";t("quickSuggestions",{comments:i,strings:i,other:i})}});ha("experimental.stickyScroll.enabled",(s,e,t)=>{typeof s=="boolean"&&(t("experimental.stickyScroll.enabled",void 0),typeof e("stickyScroll.enabled")>"u"&&t("stickyScroll.enabled",s))});ha("experimental.stickyScroll.maxLineCount",(s,e,t)=>{typeof s=="number"&&(t("experimental.stickyScroll.maxLineCount",void 0),typeof e("stickyScroll.maxLineCount")>"u"&&t("stickyScroll.maxLineCount",s))});ha("editor.experimentalEditContextEnabled",(s,e,t)=>{typeof s=="boolean"&&(t("editor.experimentalEditContextEnabled",void 0),typeof e("editor.editContext")>"u"&&t("editor.editContext",s))});ha("codeActionsOnSave",(s,e,t)=>{if(s&&typeof s=="object"){let i=!1;const n={};for(const o of Object.entries(s))typeof o[1]=="boolean"?(i=!0,n[o[0]]=o[1]?"explicit":"never"):n[o[0]]=o[1];i&&t("codeActionsOnSave",n)}});ha("codeActionWidget.includeNearbyQuickfixes",(s,e,t)=>{typeof s=="boolean"&&(t("codeActionWidget.includeNearbyQuickfixes",void 0),typeof e("codeActionWidget.includeNearbyQuickFixes")>"u"&&t("codeActionWidget.includeNearbyQuickFixes",s))});ha("lightbulb.enabled",(s,e,t)=>{typeof s=="boolean"&&t("lightbulb.enabled",s?void 0:"off")});ha("inlineSuggest.edits.codeShifting",(s,e,t)=>{typeof s=="boolean"&&(t("inlineSuggest.edits.codeShifting",void 0),t("inlineSuggest.edits.allowCodeShifting",s?"always":"never"))});class Ppe{constructor(){this._tabFocus=!1,this._onDidChangeTabFocus=new M,this.onDidChangeTabFocus=this._onDidChangeTabFocus.event}getTabFocusMode(){return this._tabFocus}setTabFocusMode(e){this._tabFocus=e,this._onDidChangeTabFocus.fire(this._tabFocus)}}const av=new Ppe;function Fpe(s,e,t){const i=s.get(58),n=s.get(62),o=s.get(61),r=s.get(60),a=s.get(63),l=s.get(75),c=s.get(72);return Em._create(i,n,o,r,a,l,c,e,t)}function Bpe(s,e,t=!1){const i=Is.fontFamily.validate(s.fontFamily),n=Is.fontWeight.validate(s.fontWeight),o=Is.fontSize.validate(s.fontSize),r=Is.fontLigatures2.validate(s.fontLigatures),a=Is.fontVariations.validate(s.fontVariations),l=Is.lineHeight.validate(s.lineHeight),c=Is.letterSpacing.validate(s.letterSpacing);return Em._create(i,n,o,r,a,l,c,e,t)}const Wn=qe("accessibilityService"),Jv=new ce("accessibilityModeEnabled",!1);var Wpe=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},Hpe=function(s,e){return function(t,i){e(t,i,s)}};let qP=class extends F{constructor(e,t,i,n,o){super(),this._accessibilityService=o,this._onDidChange=this._register(new M),this.onDidChange=this._onDidChange.event,this._onDidChangeFast=this._register(new M),this.onDidChangeFast=this._onDidChangeFast.event,this._isDominatedByLongLines=!1,this._viewLineCount=1,this._lineNumbersDigitCount=1,this._reservedHeight=0,this._glyphMarginDecorationLaneCount=1,this._computeOptionsMemory=new xJ,this.isSimpleWidget=e,this.contextMenuId=t,this._containerObserver=this._register(new vJ(n,i.dimension)),this._targetWindowId=be(n).vscodeWindowId,this._rawOptions=I$(i),this._validatedOptions=ru.validateOptions(this._rawOptions),this.options=this._computeOptions(),this.options.get(19)&&this._containerObserver.startObserving(),this._register(Gr.onDidChangeZoomLevel(()=>this._recomputeOptions())),this._register(av.onDidChangeTabFocus(()=>this._recomputeOptions())),this._register(this._containerObserver.onDidChange(()=>this._recomputeOptions())),this._register(UP.onDidChange(()=>this._recomputeOptions())),this._register(sS.getInstance(be(n)).onDidChange(()=>this._recomputeOptions())),this._register(this._accessibilityService.onDidChangeScreenReaderOptimized(()=>this._recomputeOptions())),this._register(RP.onDidChangeInputMode(()=>this._recomputeOptions()))}_recomputeOptions(){const e=this._computeOptions(),t=ru.checkEquals(this.options,e);t!==null&&(this.options=e,this._onDidChangeFast.fire(t),this._onDidChange.fire(t))}_computeOptions(){const e=this._readEnvConfiguration(),t=Fpe(this._validatedOptions,e.pixelRatio,this.isSimpleWidget),i=this._readFontInfo(t),n={memory:this._computeOptionsMemory,outerWidth:e.outerWidth,outerHeight:e.outerHeight-this._reservedHeight,fontInfo:i,extraEditorClassName:e.extraEditorClassName,isDominatedByLongLines:this._isDominatedByLongLines,viewLineCount:this._viewLineCount,lineNumbersDigitCount:this._lineNumbersDigitCount,emptySelectionClipboard:e.emptySelectionClipboard,pixelRatio:e.pixelRatio,tabFocusMode:this._validatedOptions.get(164)||av.getTabFocusMode(),inputMode:RP.getInputMode(),accessibilitySupport:e.accessibilitySupport,glyphMarginDecorationLaneCount:this._glyphMarginDecorationLaneCount,editContextSupported:e.editContextSupported};return ru.computeOptions(this._validatedOptions,n)}_readEnvConfiguration(){return{extraEditorClassName:zpe(),outerWidth:this._containerObserver.getWidth(),outerHeight:this._containerObserver.getHeight(),emptySelectionClipboard:xR||wo,pixelRatio:sS.getInstance(qz(this._targetWindowId,!0).window).value,editContextSupported:typeof globalThis.EditContext=="function",accessibilitySupport:this._accessibilityService.isScreenReaderOptimized()?2:this._accessibilityService.getAccessibilitySupport()}}_readFontInfo(e){return UP.readFontInfo(qz(this._targetWindowId,!0).window,e)}getRawOptions(){return this._rawOptions}updateOptions(e){const t=I$(e);ru.applyUpdate(this._rawOptions,t)&&(this._validatedOptions=ru.validateOptions(this._rawOptions),this._recomputeOptions())}observeContainer(e){this._containerObserver.observe(e)}setIsDominatedByLongLines(e){this._isDominatedByLongLines!==e&&(this._isDominatedByLongLines=e,this._recomputeOptions())}setModelLineCount(e){const t=Vpe(e);this._lineNumbersDigitCount!==t&&(this._lineNumbersDigitCount=t,this._recomputeOptions())}setViewLineCount(e){this._viewLineCount!==e&&(this._viewLineCount=e,this._recomputeOptions())}setReservedHeight(e){this._reservedHeight!==e&&(this._reservedHeight=e,this._recomputeOptions())}setGlyphMarginDecorationLaneCount(e){this._glyphMarginDecorationLaneCount!==e&&(this._glyphMarginDecorationLaneCount=e,this._recomputeOptions())}};qP=Wpe([Hpe(4,Wn)],qP);function Vpe(s){let e=0;for(;s;)s=Math.floor(s/10),e++;return e||1}function zpe(){let s="";return wm||FX?(s+="no-minimap-shadow ",s+="enable-user-select "):s+="no-user-select ",Ye&&(s+="mac "),s}class $pe{constructor(){this._values=[]}_read(e){return this._values[e]}get(e){return this._values[e]}_write(e,t){this._values[e]=t}}class Upe{constructor(){this._values=[]}_read(e){if(e>=this._values.length)throw new Error("Cannot read uninitialized value");return this._values[e]}get(e){return this._read(e)}_write(e,t){this._values[e]=t}}class ru{static validateOptions(e){const t=new $pe;for(const i of xb){const n=i.name==="_never_"?void 0:e[i.name];t._write(i.id,i.validate(n))}return t}static computeOptions(e,t){const i=new Upe;for(const n of xb)i._write(n.id,n.compute(t,i,e._read(n.id)));return i}static _deepEquals(e,t){if(typeof e!="object"||typeof t!="object"||!e||!t)return e===t;if(Array.isArray(e)||Array.isArray(t))return Array.isArray(e)&&Array.isArray(t)?ni(e,t):!1;if(Object.keys(e).length!==Object.keys(t).length)return!1;for(const i in e)if(!ru._deepEquals(e[i],t[i]))return!1;return!0}static checkEquals(e,t){const i=[];let n=!1;for(const o of xb){const r=!ru._deepEquals(e._read(o.id),t._read(o.id));i[o.id]=r,r&&(n=!0)}return n?new SJ(i):null}static applyUpdate(e,t){let i=!1;for(const n of xb)if(t.hasOwnProperty(n.name)){const o=n.applyUpdate(e[n.name],t[n.name]);e[n.name]=o.newValue,i=i||o.didChange}return i}}function I$(s){const e=rc(s);return Mpe(e),e}var pu;(function(s){const e={total:0,min:Number.MAX_VALUE,max:0},t={...e},i={...e},n={...e};let o=0;const r={keydown:0,input:0,render:0};function a(){b(),performance.mark("inputlatency/start"),performance.mark("keydown/start"),r.keydown=1,queueMicrotask(l)}s.onKeyDown=a;function l(){r.keydown===1&&(performance.mark("keydown/end"),r.keydown=2)}function c(){performance.mark("input/start"),r.input=1,_()}s.onBeforeInput=c;function d(){r.input===0&&c(),queueMicrotask(h)}s.onInput=d;function h(){r.input===1&&(performance.mark("input/end"),r.input=2)}function u(){b()}s.onKeyUp=u;function g(){b()}s.onSelectionChange=g;function f(){r.keydown===2&&r.input===2&&r.render===0&&(performance.mark("render/start"),r.render=1,queueMicrotask(p),_())}s.onRenderStart=f;function p(){r.render===1&&(performance.mark("render/end"),r.render=2)}function _(){setTimeout(b)}function b(){r.keydown===2&&r.input===2&&r.render===2&&(performance.mark("inputlatency/end"),performance.measure("keydown","keydown/start","keydown/end"),performance.measure("input","input/start","input/end"),performance.measure("render","render/start","render/end"),performance.measure("inputlatency","inputlatency/start","inputlatency/end"),w("keydown",e),w("input",t),w("render",i),w("inputlatency",n),o++,C())}function w(L,E){const I=performance.getEntriesByName(L)[0].duration;E.total+=I,E.min=Math.min(E.min,I),E.max=Math.max(E.max,I)}function C(){performance.clearMarks("keydown/start"),performance.clearMarks("keydown/end"),performance.clearMarks("input/start"),performance.clearMarks("input/end"),performance.clearMarks("render/start"),performance.clearMarks("render/end"),performance.clearMarks("inputlatency/start"),performance.clearMarks("inputlatency/end"),performance.clearMeasures("keydown"),performance.clearMeasures("input"),performance.clearMeasures("render"),performance.clearMeasures("inputlatency"),r.keydown=0,r.input=0,r.render=0}function v(){if(o===0)return;const L={keydown:y(e),input:y(t),render:y(i),total:y(n),sampleCount:o};return x(e),x(t),x(i),x(n),o=0,L}s.getAndClearMeasurements=v;function y(L){return{average:L.total/o,max:L.max,min:L.min}}function x(L){L.total=0,L.min=Number.MAX_VALUE,L.max=0}})(pu||(pu={}));const VE=new Map;function qpe(){return new jpe}class jpe{constructor(){this._currentCssStyle="",this._styleSheet=void 0}setStyle(e){e!==this._currentCssStyle&&(this._currentCssStyle=e,this._styleSheet?this._styleSheet.textContent=e:this._styleSheet=ca(Nt.document.head,t=>t.textContent=e))}dispose(){this._styleSheet&&(this._styleSheet.remove(),this._styleSheet=void 0)}}function ca(s=Nt.document.head,e,t){const i=document.createElement("style");if(i.type="text/css",i.media="screen",e?.(i),s.appendChild(i),t&&t.add(pe(()=>i.remove())),s===Nt.document.head){const n=new Set;VE.set(i,n),t&&t.add(pe(()=>VE.delete(i)));for(const{window:o,disposables:r}of kQ()){if(o===Nt)continue;const a=r.add(Gpe(i,n,o));t?.add(a)}}return i}function Gpe(s,e,t){const i=new U,n=s.cloneNode(!0);t.document.head.appendChild(n),i.add(pe(()=>n.remove()));for(const o of kJ(s))n.sheet?.insertRule(o.cssText,n.sheet?.cssRules.length);return i.add(aue.observe(s,i,{childList:!0,subtree:wo,characterData:wo})(()=>{n.textContent=s.textContent})),e.add(n),i.add(pe(()=>e.delete(n))),i}let J2=null;function LJ(){return J2||(J2=ca()),J2}function kJ(s){return s?.sheet?.rules?s.sheet.rules:s?.sheet?.cssRules?s.sheet.cssRules:[]}function zE(s,e,t=LJ()){if(!(!t||!e)){t.sheet?.insertRule(`${s} {${e}}`,0);for(const i of VE.get(t)??[])zE(s,e,i)}}function jP(s,e=LJ()){if(!e)return;const t=kJ(e),i=[];for(let n=0;n<t.length;n++){const o=t[n];Kpe(o)&&o.selectorText.indexOf(s)!==-1&&i.push(n)}for(let n=i.length-1;n>=0;n--)e.sheet?.deleteRule(i[n]);for(const n of VE.get(e)??[])jP(s,n)}function Kpe(s){return typeof s.selectorText=="string"}function Ype(s){const e=new U,t=e.add(qpe());return e.add(De(i=>{t.setStyle(s.read(i))})),e}class eC{constructor(){this._hooks=new U,this._pointerMoveCallback=null,this._onStopCallback=null}dispose(){this.stopMonitoring(!1),this._hooks.dispose()}stopMonitoring(e,t){if(!this.isMonitoring())return;this._hooks.clear(),this._pointerMoveCallback=null;const i=this._onStopCallback;this._onStopCallback=null,e&&i&&i(t)}isMonitoring(){return!!this._pointerMoveCallback}startMonitoring(e,t,i,n,o){this.isMonitoring()&&this.stopMonitoring(!1),this._pointerMoveCallback=n,this._onStopCallback=o;let r=e;try{e.setPointerCapture(t),this._hooks.add(pe(()=>{try{e.releasePointerCapture(t)}catch{}}))}catch{r=be(e)}this._hooks.add(W(r,se.POINTER_MOVE,a=>{if(a.buttons!==i){this.stopMonitoring(!0);return}a.preventDefault(),this._pointerMoveCallback(a)})),this._hooks.add(W(r,se.POINTER_UP,a=>this.stopMonitoring(!0)))}}function Mu(s,e){const t=Math.pow(10,e);return Math.round(s*t)/t}class G{constructor(e,t,i,n=1){this._rgbaBrand=void 0,this.r=Math.min(255,Math.max(0,e))|0,this.g=Math.min(255,Math.max(0,t))|0,this.b=Math.min(255,Math.max(0,i))|0,this.a=Mu(Math.max(Math.min(1,n),0),3)}static equals(e,t){return e.r===t.r&&e.g===t.g&&e.b===t.b&&e.a===t.a}}class Ea{constructor(e,t,i,n){this._hslaBrand=void 0,this.h=Math.max(Math.min(360,e),0)|0,this.s=Mu(Math.max(Math.min(1,t),0),3),this.l=Mu(Math.max(Math.min(1,i),0),3),this.a=Mu(Math.max(Math.min(1,n),0),3)}static equals(e,t){return e.h===t.h&&e.s===t.s&&e.l===t.l&&e.a===t.a}static fromRGBA(e){const t=e.r/255,i=e.g/255,n=e.b/255,o=e.a,r=Math.max(t,i,n),a=Math.min(t,i,n);let l=0,c=0;const d=(a+r)/2,h=r-a;if(h>0){switch(c=Math.min(d<=.5?h/(2*d):h/(2-2*d),1),r){case t:l=(i-n)/h+(i<n?6:0);break;case i:l=(n-t)/h+2;break;case n:l=(t-i)/h+4;break}l*=60,l=Math.round(l)}return new Ea(l,c,d,o)}static _hue2rgb(e,t,i){return i<0&&(i+=1),i>1&&(i-=1),i<1/6?e+(t-e)*6*i:i<1/2?t:i<2/3?e+(t-e)*(2/3-i)*6:e}static toRGBA(e){const t=e.h/360,{s:i,l:n,a:o}=e;let r,a,l;if(i===0)r=a=l=n;else{const c=n<.5?n*(1+i):n+i-n*i,d=2*n-c;r=Ea._hue2rgb(d,c,t+1/3),a=Ea._hue2rgb(d,c,t),l=Ea._hue2rgb(d,c,t-1/3)}return new G(Math.round(r*255),Math.round(a*255),Math.round(l*255),o)}}class vc{constructor(e,t,i,n){this._hsvaBrand=void 0,this.h=Math.max(Math.min(360,e),0)|0,this.s=Mu(Math.max(Math.min(1,t),0),3),this.v=Mu(Math.max(Math.min(1,i),0),3),this.a=Mu(Math.max(Math.min(1,n),0),3)}static equals(e,t){return e.h===t.h&&e.s===t.s&&e.v===t.v&&e.a===t.a}static fromRGBA(e){const t=e.r/255,i=e.g/255,n=e.b/255,o=Math.max(t,i,n),r=Math.min(t,i,n),a=o-r,l=o===0?0:a/o;let c;return a===0?c=0:o===t?c=((i-n)/a%6+6)%6:o===i?c=(n-t)/a+2:c=(t-i)/a+4,new vc(Math.round(c*60),l,o,e.a)}static toRGBA(e){const{h:t,s:i,v:n,a:o}=e,r=n*i,a=r*(1-Math.abs(t/60%2-1)),l=n-r;let[c,d,h]=[0,0,0];return t<60?(c=r,d=a):t<120?(c=a,d=r):t<180?(d=r,h=a):t<240?(d=a,h=r):t<300?(c=a,h=r):t<=360&&(c=r,h=a),c=Math.round((c+l)*255),d=Math.round((d+l)*255),h=Math.round((h+l)*255),new G(c,d,h,o)}}var li;let q=(li=class{static fromHex(e){return li.Format.CSS.parseHex(e)||li.red}static equals(e,t){return!e&&!t?!0:!e||!t?!1:e.equals(t)}get hsla(){return this._hsla?this._hsla:Ea.fromRGBA(this.rgba)}get hsva(){return this._hsva?this._hsva:vc.fromRGBA(this.rgba)}constructor(e){if(e)if(e instanceof G)this.rgba=e;else if(e instanceof Ea)this._hsla=e,this.rgba=Ea.toRGBA(e);else if(e instanceof vc)this._hsva=e,this.rgba=vc.toRGBA(e);else throw new Error("Invalid color ctor argument");else throw new Error("Color needs a value")}equals(e){return!!e&&G.equals(this.rgba,e.rgba)&&Ea.equals(this.hsla,e.hsla)&&vc.equals(this.hsva,e.hsva)}getRelativeLuminance(){const e=li._relativeLuminanceForComponent(this.rgba.r),t=li._relativeLuminanceForComponent(this.rgba.g),i=li._relativeLuminanceForComponent(this.rgba.b),n=.2126*e+.7152*t+.0722*i;return Mu(n,4)}static _relativeLuminanceForComponent(e){const t=e/255;return t<=.03928?t/12.92:Math.pow((t+.055)/1.055,2.4)}isLighter(){return(this.rgba.r*299+this.rgba.g*587+this.rgba.b*114)/1e3>=128}isLighterThan(e){const t=this.getRelativeLuminance(),i=e.getRelativeLuminance();return t>i}isDarkerThan(e){const t=this.getRelativeLuminance(),i=e.getRelativeLuminance();return t<i}lighten(e){return new li(new Ea(this.hsla.h,this.hsla.s,this.hsla.l+this.hsla.l*e,this.hsla.a))}darken(e){return new li(new Ea(this.hsla.h,this.hsla.s,this.hsla.l-this.hsla.l*e,this.hsla.a))}transparent(e){const{r:t,g:i,b:n,a:o}=this.rgba;return new li(new G(t,i,n,o*e))}isTransparent(){return this.rgba.a===0}isOpaque(){return this.rgba.a===1}opposite(){return new li(new G(255-this.rgba.r,255-this.rgba.g,255-this.rgba.b,this.rgba.a))}mix(e,t=.5){const i=Math.min(Math.max(t,0),1),n=this.rgba,o=e.rgba,r=n.r+(o.r-n.r)*i,a=n.g+(o.g-n.g)*i,l=n.b+(o.b-n.b)*i,c=n.a+(o.a-n.a)*i;return new li(new G(r,a,l,c))}makeOpaque(e){if(this.isOpaque()||e.rgba.a!==1)return this;const{r:t,g:i,b:n,a:o}=this.rgba;return new li(new G(e.rgba.r-o*(e.rgba.r-t),e.rgba.g-o*(e.rgba.g-i),e.rgba.b-o*(e.rgba.b-n),1))}toString(){return this._toString||(this._toString=li.Format.CSS.format(this)),this._toString}toNumber32Bit(){return this._toNumber32Bit||(this._toNumber32Bit=(this.rgba.r<<24|this.rgba.g<<16|this.rgba.b<<8|this.rgba.a*255<<0)>>>0),this._toNumber32Bit}static getLighterColor(e,t,i){if(e.isLighterThan(t))return e;i=i||.5;const n=e.getRelativeLuminance(),o=t.getRelativeLuminance();return i=i*(o-n)/o,e.lighten(i)}static getDarkerColor(e,t,i){if(e.isDarkerThan(t))return e;i=i||.5;const n=e.getRelativeLuminance(),o=t.getRelativeLuminance();return i=i*(n-o)/n,e.darken(i)}},li.white=new li(new G(255,255,255,1)),li.black=new li(new G(0,0,0,1)),li.red=new li(new G(255,0,0,1)),li.blue=new li(new G(0,0,255,1)),li.green=new li(new G(0,255,0,1)),li.cyan=new li(new G(0,255,255,1)),li.lightgrey=new li(new G(211,211,211,1)),li.transparent=new li(new G(0,0,0,0)),li);(function(s){(function(e){(function(t){function i(p){return p.rgba.a===1?`rgb(${p.rgba.r}, ${p.rgba.g}, ${p.rgba.b})`:s.Format.CSS.formatRGBA(p)}t.formatRGB=i;function n(p){return`rgba(${p.rgba.r}, ${p.rgba.g}, ${p.rgba.b}, ${+p.rgba.a.toFixed(2)})`}t.formatRGBA=n;function o(p){return p.hsla.a===1?`hsl(${p.hsla.h}, ${Math.round(p.hsla.s*100)}%, ${Math.round(p.hsla.l*100)}%)`:s.Format.CSS.formatHSLA(p)}t.formatHSL=o;function r(p){return`hsla(${p.hsla.h}, ${Math.round(p.hsla.s*100)}%, ${Math.round(p.hsla.l*100)}%, ${p.hsla.a.toFixed(2)})`}t.formatHSLA=r;function a(p){const _=p.toString(16);return _.length!==2?"0"+_:_}function l(p){return`#${a(p.rgba.r)}${a(p.rgba.g)}${a(p.rgba.b)}`}t.formatHex=l;function c(p,_=!1){return _&&p.rgba.a===1?s.Format.CSS.formatHex(p):`#${a(p.rgba.r)}${a(p.rgba.g)}${a(p.rgba.b)}${a(Math.round(p.rgba.a*255))}`}t.formatHexA=c;function d(p){return p.isOpaque()?s.Format.CSS.formatHex(p):s.Format.CSS.formatRGBA(p)}t.format=d;function h(p){if(p==="transparent")return s.transparent;if(p.startsWith("#"))return g(p);if(p.startsWith("rgba(")){const _=p.match(/rgba\((?<r>(?:\+|-)?\d+), *(?<g>(?:\+|-)?\d+), *(?<b>(?:\+|-)?\d+), *(?<a>(?:\+|-)?\d+(\.\d+)?)\)/);if(!_)throw new Error("Invalid color format "+p);const b=parseInt(_.groups?.r??"0"),w=parseInt(_.groups?.g??"0"),C=parseInt(_.groups?.b??"0"),v=parseFloat(_.groups?.a??"0");return new s(new G(b,w,C,v))}if(p.startsWith("rgb(")){const _=p.match(/rgb\((?<r>(?:\+|-)?\d+), *(?<g>(?:\+|-)?\d+), *(?<b>(?:\+|-)?\d+)\)/);if(!_)throw new Error("Invalid color format "+p);const b=parseInt(_.groups?.r??"0"),w=parseInt(_.groups?.g??"0"),C=parseInt(_.groups?.b??"0");return new s(new G(b,w,C))}return u(p)}t.parse=h;function u(p){switch(p){case"aliceblue":return new s(new G(240,248,255,1));case"antiquewhite":return new s(new G(250,235,215,1));case"aqua":return new s(new G(0,255,255,1));case"aquamarine":return new s(new G(127,255,212,1));case"azure":return new s(new G(240,255,255,1));case"beige":return new s(new G(245,245,220,1));case"bisque":return new s(new G(255,228,196,1));case"black":return new s(new G(0,0,0,1));case"blanchedalmond":return new s(new G(255,235,205,1));case"blue":return new s(new G(0,0,255,1));case"blueviolet":return new s(new G(138,43,226,1));case"brown":return new s(new G(165,42,42,1));case"burlywood":return new s(new G(222,184,135,1));case"cadetblue":return new s(new G(95,158,160,1));case"chartreuse":return new s(new G(127,255,0,1));case"chocolate":return new s(new G(210,105,30,1));case"coral":return new s(new G(255,127,80,1));case"cornflowerblue":return new s(new G(100,149,237,1));case"cornsilk":return new s(new G(255,248,220,1));case"crimson":return new s(new G(220,20,60,1));case"cyan":return new s(new G(0,255,255,1));case"darkblue":return new s(new G(0,0,139,1));case"darkcyan":return new s(new G(0,139,139,1));case"darkgoldenrod":return new s(new G(184,134,11,1));case"darkgray":return new s(new G(169,169,169,1));case"darkgreen":return new s(new G(0,100,0,1));case"darkgrey":return new s(new G(169,169,169,1));case"darkkhaki":return new s(new G(189,183,107,1));case"darkmagenta":return new s(new G(139,0,139,1));case"darkolivegreen":return new s(new G(85,107,47,1));case"darkorange":return new s(new G(255,140,0,1));case"darkorchid":return new s(new G(153,50,204,1));case"darkred":return new s(new G(139,0,0,1));case"darksalmon":return new s(new G(233,150,122,1));case"darkseagreen":return new s(new G(143,188,143,1));case"darkslateblue":return new s(new G(72,61,139,1));case"darkslategray":return new s(new G(47,79,79,1));case"darkslategrey":return new s(new G(47,79,79,1));case"darkturquoise":return new s(new G(0,206,209,1));case"darkviolet":return new s(new G(148,0,211,1));case"deeppink":return new s(new G(255,20,147,1));case"deepskyblue":return new s(new G(0,191,255,1));case"dimgray":return new s(new G(105,105,105,1));case"dimgrey":return new s(new G(105,105,105,1));case"dodgerblue":return new s(new G(30,144,255,1));case"firebrick":return new s(new G(178,34,34,1));case"floralwhite":return new s(new G(255,250,240,1));case"forestgreen":return new s(new G(34,139,34,1));case"fuchsia":return new s(new G(255,0,255,1));case"gainsboro":return new s(new G(220,220,220,1));case"ghostwhite":return new s(new G(248,248,255,1));case"gold":return new s(new G(255,215,0,1));case"goldenrod":return new s(new G(218,165,32,1));case"gray":return new s(new G(128,128,128,1));case"green":return new s(new G(0,128,0,1));case"greenyellow":return new s(new G(173,255,47,1));case"grey":return new s(new G(128,128,128,1));case"honeydew":return new s(new G(240,255,240,1));case"hotpink":return new s(new G(255,105,180,1));case"indianred":return new s(new G(205,92,92,1));case"indigo":return new s(new G(75,0,130,1));case"ivory":return new s(new G(255,255,240,1));case"khaki":return new s(new G(240,230,140,1));case"lavender":return new s(new G(230,230,250,1));case"lavenderblush":return new s(new G(255,240,245,1));case"lawngreen":return new s(new G(124,252,0,1));case"lemonchiffon":return new s(new G(255,250,205,1));case"lightblue":return new s(new G(173,216,230,1));case"lightcoral":return new s(new G(240,128,128,1));case"lightcyan":return new s(new G(224,255,255,1));case"lightgoldenrodyellow":return new s(new G(250,250,210,1));case"lightgray":return new s(new G(211,211,211,1));case"lightgreen":return new s(new G(144,238,144,1));case"lightgrey":return new s(new G(211,211,211,1));case"lightpink":return new s(new G(255,182,193,1));case"lightsalmon":return new s(new G(255,160,122,1));case"lightseagreen":return new s(new G(32,178,170,1));case"lightskyblue":return new s(new G(135,206,250,1));case"lightslategray":return new s(new G(119,136,153,1));case"lightslategrey":return new s(new G(119,136,153,1));case"lightsteelblue":return new s(new G(176,196,222,1));case"lightyellow":return new s(new G(255,255,224,1));case"lime":return new s(new G(0,255,0,1));case"limegreen":return new s(new G(50,205,50,1));case"linen":return new s(new G(250,240,230,1));case"magenta":return new s(new G(255,0,255,1));case"maroon":return new s(new G(128,0,0,1));case"mediumaquamarine":return new s(new G(102,205,170,1));case"mediumblue":return new s(new G(0,0,205,1));case"mediumorchid":return new s(new G(186,85,211,1));case"mediumpurple":return new s(new G(147,112,219,1));case"mediumseagreen":return new s(new G(60,179,113,1));case"mediumslateblue":return new s(new G(123,104,238,1));case"mediumspringgreen":return new s(new G(0,250,154,1));case"mediumturquoise":return new s(new G(72,209,204,1));case"mediumvioletred":return new s(new G(199,21,133,1));case"midnightblue":return new s(new G(25,25,112,1));case"mintcream":return new s(new G(245,255,250,1));case"mistyrose":return new s(new G(255,228,225,1));case"moccasin":return new s(new G(255,228,181,1));case"navajowhite":return new s(new G(255,222,173,1));case"navy":return new s(new G(0,0,128,1));case"oldlace":return new s(new G(253,245,230,1));case"olive":return new s(new G(128,128,0,1));case"olivedrab":return new s(new G(107,142,35,1));case"orange":return new s(new G(255,165,0,1));case"orangered":return new s(new G(255,69,0,1));case"orchid":return new s(new G(218,112,214,1));case"palegoldenrod":return new s(new G(238,232,170,1));case"palegreen":return new s(new G(152,251,152,1));case"paleturquoise":return new s(new G(175,238,238,1));case"palevioletred":return new s(new G(219,112,147,1));case"papayawhip":return new s(new G(255,239,213,1));case"peachpuff":return new s(new G(255,218,185,1));case"peru":return new s(new G(205,133,63,1));case"pink":return new s(new G(255,192,203,1));case"plum":return new s(new G(221,160,221,1));case"powderblue":return new s(new G(176,224,230,1));case"purple":return new s(new G(128,0,128,1));case"rebeccapurple":return new s(new G(102,51,153,1));case"red":return new s(new G(255,0,0,1));case"rosybrown":return new s(new G(188,143,143,1));case"royalblue":return new s(new G(65,105,225,1));case"saddlebrown":return new s(new G(139,69,19,1));case"salmon":return new s(new G(250,128,114,1));case"sandybrown":return new s(new G(244,164,96,1));case"seagreen":return new s(new G(46,139,87,1));case"seashell":return new s(new G(255,245,238,1));case"sienna":return new s(new G(160,82,45,1));case"silver":return new s(new G(192,192,192,1));case"skyblue":return new s(new G(135,206,235,1));case"slateblue":return new s(new G(106,90,205,1));case"slategray":return new s(new G(112,128,144,1));case"slategrey":return new s(new G(112,128,144,1));case"snow":return new s(new G(255,250,250,1));case"springgreen":return new s(new G(0,255,127,1));case"steelblue":return new s(new G(70,130,180,1));case"tan":return new s(new G(210,180,140,1));case"teal":return new s(new G(0,128,128,1));case"thistle":return new s(new G(216,191,216,1));case"tomato":return new s(new G(255,99,71,1));case"turquoise":return new s(new G(64,224,208,1));case"violet":return new s(new G(238,130,238,1));case"wheat":return new s(new G(245,222,179,1));case"white":return new s(new G(255,255,255,1));case"whitesmoke":return new s(new G(245,245,245,1));case"yellow":return new s(new G(255,255,0,1));case"yellowgreen":return new s(new G(154,205,50,1));default:return null}}function g(p){const _=p.length;if(_===0||p.charCodeAt(0)!==35)return null;if(_===7){const b=16*f(p.charCodeAt(1))+f(p.charCodeAt(2)),w=16*f(p.charCodeAt(3))+f(p.charCodeAt(4)),C=16*f(p.charCodeAt(5))+f(p.charCodeAt(6));return new s(new G(b,w,C,1))}if(_===9){const b=16*f(p.charCodeAt(1))+f(p.charCodeAt(2)),w=16*f(p.charCodeAt(3))+f(p.charCodeAt(4)),C=16*f(p.charCodeAt(5))+f(p.charCodeAt(6)),v=16*f(p.charCodeAt(7))+f(p.charCodeAt(8));return new s(new G(b,w,C,v/255))}if(_===4){const b=f(p.charCodeAt(1)),w=f(p.charCodeAt(2)),C=f(p.charCodeAt(3));return new s(new G(16*b+b,16*w+w,16*C+C))}if(_===5){const b=f(p.charCodeAt(1)),w=f(p.charCodeAt(2)),C=f(p.charCodeAt(3)),v=f(p.charCodeAt(4));return new s(new G(16*b+b,16*w+w,16*C+C,(16*v+v)/255))}return null}t.parseHex=g;function f(p){switch(p){case 48:return 0;case 49:return 1;case 50:return 2;case 51:return 3;case 52:return 4;case 53:return 5;case 54:return 6;case 55:return 7;case 56:return 8;case 57:return 9;case 97:return 10;case 65:return 10;case 98:return 11;case 66:return 11;case 99:return 12;case 67:return 12;case 100:return 13;case 68:return 13;case 101:return 14;case 69:return 14;case 102:return 15;case 70:return 15}return 0}})(e.CSS||(e.CSS={}))})(s.Format||(s.Format={}))})(q||(q={}));function DW(s){return`--vscode-${s.replace(/\./g,"-")}`}function te(s){return`var(${DW(s)})`}function EJ(s,e){return`var(${DW(s)}, ${e})`}function Zpe(s){return s!==null&&typeof s=="object"&&"light"in s&&"dark"in s}const IJ={ColorContribution:"base.contributions.colors"},Xpe="default";class Qpe extends F{constructor(){super(),this._onDidChangeSchema=this._register(new M),this.onDidChangeSchema=this._onDidChangeSchema.event,this.colorSchema={type:"object",properties:{}},this.colorReferenceSchema={type:"string",enum:[],enumDescriptions:[]},this.colorsById={}}registerColor(e,t,i,n=!1,o){const r={id:e,description:i,defaults:t,needsTransparency:n,deprecationMessage:o};this.colorsById[e]=r;const a={type:"string",format:"color-hex",defaultSnippets:[{body:"${1:#ff0000}"}]};return o&&(a.deprecationMessage=o),n&&(a.pattern="^#(?:(?<rgba>[0-9a-fA-f]{3}[0-9a-eA-E])|(?:[0-9a-fA-F]{6}(?:(?![fF]{2})(?:[0-9a-fA-F]{2}))))?$",a.patternErrorMessage=m(2022,"This color must be transparent or it will obscure content")),this.colorSchema.properties[e]={description:i,oneOf:[a,{type:"string",const:Xpe,description:m(2023,"Use the default color.")}]},this.colorReferenceSchema.enum.push(e),this.colorReferenceSchema.enumDescriptions.push(i),this._onDidChangeSchema.fire(),e}getColors(){return Object.keys(this.colorsById).map(e=>this.colorsById[e])}resolveDefaultColor(e,t){const i=this.colorsById[e];if(i?.defaults){const n=Zpe(i.defaults)?i.defaults[t.type]:i.defaults;return Ar(n,t)}}getColorSchema(){return this.colorSchema}toString(){const e=(t,i)=>{const n=t.indexOf(".")===-1?0:1,o=i.indexOf(".")===-1?0:1;return n!==o?n-o:t.localeCompare(i)};return Object.keys(this.colorsById).sort(e).map(t=>`- \`${t}\`: ${this.colorsById[t].description}`).join(`
`)}}const JR=new Qpe;bi.add(IJ.ColorContribution,JR);function T(s,e,t,i,n){return JR.registerColor(s,e,t,i,n)}function Jpe(s,e){switch(s.op){case 0:return Ar(s.value,e)?.darken(s.factor);case 1:return Ar(s.value,e)?.lighten(s.factor);case 2:return Ar(s.value,e)?.transparent(s.factor);case 7:{const t=Ar(s.color,e)||q.transparent,i=Ar(s.with,e)||q.transparent;return t.mix(i,s.ratio)}case 3:{const t=Ar(s.background,e);return t?Ar(s.value,e)?.makeOpaque(t):Ar(s.value,e)}case 4:for(const t of s.values){const i=Ar(t,e);if(i)return i}return;case 6:return Ar(e.defines(s.if)?s.then:s.else,e);case 5:{const t=Ar(s.value,e);if(!t)return;const i=Ar(s.background,e);return i?t.isDarkerThan(i)?q.getLighterColor(t,i,s.factor).transparent(s.transparency):q.getDarkerColor(t,i,s.factor).transparent(s.transparency):t.transparent(s.factor*s.transparency)}default:throw W0()}}function mo(s,e){return{op:0,value:s,factor:e}}function vl(s,e){return{op:1,value:s,factor:e}}function Be(s,e){return{op:2,value:s,factor:e}}function rS(...s){return{op:4,values:s}}function eme(s,e,t){return{op:6,if:s,then:e,else:t}}function D$(s,e,t,i){return{op:5,value:s,background:e,factor:t,transparency:i}}function Ar(s,e){if(s!==null){if(typeof s=="string")return s[0]==="#"?q.fromHex(s):e.getColor(s);if(s instanceof q)return s;if(typeof s=="object")return Jpe(s,e)}}const DJ="vscode://schemas/workbench-colors",TJ=bi.as(YR.JSONContribution);TJ.registerSchema(DJ,JR.getColorSchema());const T$=new Rt(()=>TJ.notifySchemaChanged(DJ),200);JR.onDidChangeSchema(()=>{T$.isScheduled()||T$.schedule()});const et=T("foreground",{dark:"#CCCCCC",light:"#616161",hcDark:"#FFFFFF",hcLight:"#292929"},m(1773,"Overall foreground color. This color is only used if not overridden by a component."));T("disabledForeground",{dark:"#CCCCCC80",light:"#61616180",hcDark:"#A5A5A5",hcLight:"#7F7F7F"},m(1774,"Overall foreground for disabled elements. This color is only used if not overridden by a component."));T("errorForeground",{dark:"#F48771",light:"#A1260D",hcDark:"#F48771",hcLight:"#B5200D"},m(1775,"Overall foreground color for error messages. This color is only used if not overridden by a component."));const NJ=T("descriptionForeground",{light:"#717171",dark:Be(et,.7),hcDark:Be(et,.7),hcLight:Be(et,.7)},m(1776,"Foreground color for description text providing additional information, for example for a label.")),$E=T("icon.foreground",{dark:"#C5C5C5",light:"#424242",hcDark:"#FFFFFF",hcLight:"#292929"},m(1777,"The default color for icons in the workbench.")),Nc=T("focusBorder",{dark:"#007FD4",light:"#0090F1",hcDark:"#F38518",hcLight:"#006BBD"},m(1778,"Overall border color for focused elements. This color is only used if not overridden by a component.")),rt=T("contrastBorder",{light:null,dark:null,hcDark:"#6FC3DF",hcLight:"#0F4A85"},m(1779,"An extra border around elements to separate them from others for greater contrast.")),ri=T("contrastActiveBorder",{light:null,dark:null,hcDark:Nc,hcLight:Nc},m(1780,"An extra border around active elements to separate them from others for greater contrast."));T("selection.background",null,m(1781,"The background color of text selections in the workbench (e.g. for input fields or text areas). Note that this does not apply to selections within the editor."));const tme=T("textLink.foreground",{light:"#006AB1",dark:"#3794FF",hcDark:"#21A6FF",hcLight:"#0F4A85"},m(1782,"Foreground color for links in text."));T("textLink.activeForeground",{light:"#006AB1",dark:"#3794FF",hcDark:"#21A6FF",hcLight:"#0F4A85"},m(1783,"Foreground color for links in text when clicked on and on mouse hover."));T("textSeparator.foreground",{light:"#0000002e",dark:"#ffffff2e",hcDark:q.black,hcLight:"#292929"},m(1784,"Color for text separators."));T("textPreformat.foreground",{light:"#A31515",dark:"#D7BA7D",hcDark:"#000000",hcLight:"#FFFFFF"},m(1785,"Foreground color for preformatted text segments."));T("textPreformat.background",{light:"#0000001A",dark:"#FFFFFF1A",hcDark:"#FFFFFF",hcLight:"#09345f"},m(1786,"Background color for preformatted text segments."));T("textBlockQuote.background",{light:"#f2f2f2",dark:"#222222",hcDark:null,hcLight:"#F2F2F2"},m(1787,"Background color for block quotes in text."));T("textBlockQuote.border",{light:"#007acc80",dark:"#007acc80",hcDark:q.white,hcLight:"#292929"},m(1788,"Border color for block quotes in text."));T("textCodeBlock.background",{light:"#dcdcdc66",dark:"#0a0a0a66",hcDark:q.black,hcLight:"#F2F2F2"},m(1789,"Background color for code blocks in text."));T("sash.hoverBorder",Nc,m(1994,"Border color of active sashes."));const Tk=T("badge.background",{dark:"#4D4D4D",light:"#C4C4C4",hcDark:q.black,hcLight:"#0F4A85"},m(1995,"Badge background color. Badges are small information labels, e.g. for search results count.")),ime=T("badge.foreground",{dark:q.white,light:"#333",hcDark:q.white,hcLight:q.white},m(1996,"Badge foreground color. Badges are small information labels, e.g. for search results count."));T("activityWarningBadge.foreground",{dark:q.black.lighten(.2),light:q.white,hcDark:null,hcLight:q.black.lighten(.2)},m(1997,"Foreground color of the warning activity badge"));T("activityWarningBadge.background",{dark:"#CCA700",light:"#BF8803",hcDark:null,hcLight:"#CCA700"},m(1998,"Background color of the warning activity badge"));T("activityErrorBadge.foreground",{dark:q.black.lighten(.2),light:q.white,hcDark:null,hcLight:q.black.lighten(.2)},m(1999,"Foreground color of the error activity badge"));T("activityErrorBadge.background",{dark:"#F14C4C",light:"#E51400",hcDark:null,hcLight:"#F14C4C"},m(2e3,"Background color of the error activity badge"));const eA=T("scrollbar.shadow",{dark:"#000000",light:"#DDDDDD",hcDark:null,hcLight:null},m(2001,"Scrollbar shadow to indicate that the view is scrolled.")),RJ=T("scrollbarSlider.background",{dark:q.fromHex("#797979").transparent(.4),light:q.fromHex("#646464").transparent(.4),hcDark:Be(rt,.6),hcLight:Be(rt,.4)},m(2002,"Scrollbar slider background color.")),AJ=T("scrollbarSlider.hoverBackground",{dark:q.fromHex("#646464").transparent(.7),light:q.fromHex("#646464").transparent(.7),hcDark:Be(rt,.8),hcLight:Be(rt,.8)},m(2003,"Scrollbar slider background color when hovering.")),MJ=T("scrollbarSlider.activeBackground",{dark:q.fromHex("#BFBFBF").transparent(.4),light:q.fromHex("#000000").transparent(.6),hcDark:rt,hcLight:rt},m(2004,"Scrollbar slider background color when clicked on."));T("scrollbar.background",null,m(2005,"Scrollbar track background color."));const nme=T("progressBar.background",{dark:q.fromHex("#0E70C0"),light:q.fromHex("#0E70C0"),hcDark:rt,hcLight:rt},m(2006,"Background color of the progress bar that can show for long running operations."));T("chart.line",{dark:"#236B8E",light:"#236B8E",hcDark:"#236B8E",hcLight:"#236B8E"},m(2007,"Line color for the chart."));T("chart.axis",{dark:q.fromHex("#BFBFBF").transparent(.4),light:q.fromHex("#000000").transparent(.6),hcDark:rt,hcLight:rt},m(2008,"Axis color for the chart."));T("chart.guide",{dark:q.fromHex("#BFBFBF").transparent(.2),light:q.fromHex("#000000").transparent(.2),hcDark:rt,hcLight:rt},m(2009,"Guide line for the chart."));const Bi=T("editor.background",{light:"#ffffff",dark:"#1E1E1E",hcDark:q.black,hcLight:q.white},m(1798,"Editor background color.")),zc=T("editor.foreground",{light:"#333333",dark:"#BBBBBB",hcDark:q.white,hcLight:et},m(1799,"Editor default foreground color."));T("editorStickyScroll.background",Bi,m(1800,"Background color of sticky scroll in the editor"));T("editorStickyScrollGutter.background",Bi,m(1801,"Background color of the gutter part of sticky scroll in the editor"));T("editorStickyScrollHover.background",{dark:"#2A2D2E",light:"#F0F0F0",hcDark:null,hcLight:q.fromHex("#0F4A85").transparent(.1)},m(1802,"Background color of sticky scroll on hover in the editor"));T("editorStickyScroll.border",{dark:null,light:null,hcDark:rt,hcLight:rt},m(1803,"Border color of sticky scroll in the editor"));T("editorStickyScroll.shadow",eA,m(1804," Shadow color of sticky scroll in the editor"));const wr=T("editorWidget.background",{dark:"#252526",light:"#F3F3F3",hcDark:"#0C141F",hcLight:q.white},m(1805,"Background color of editor widgets, such as find/replace.")),tA=T("editorWidget.foreground",et,m(1806,"Foreground color of editor widgets, such as find/replace.")),TW=T("editorWidget.border",{dark:"#454545",light:"#C8C8C8",hcDark:rt,hcLight:rt},m(1807,"Border color of editor widgets. The color is only used if the widget chooses to have a border and if the color is not overridden by a widget."));T("editorWidget.resizeBorder",null,m(1808,"Border color of the resize bar of editor widgets. The color is only used if the widget chooses to have a resize border and if the color is not overridden by a widget."));T("editorError.background",null,m(1809,"Background color of error text in the editor. The color must not be opaque so as not to hide underlying decorations."),!0);const iA=T("editorError.foreground",{dark:"#F14C4C",light:"#E51400",hcDark:"#F48771",hcLight:"#B5200D"},m(1810,"Foreground color of error squigglies in the editor.")),sme=T("editorError.border",{dark:null,light:null,hcDark:q.fromHex("#E47777").transparent(.8),hcLight:"#B5200D"},m(1811,"If set, color of double underlines for errors in the editor.")),ome=T("editorWarning.background",null,m(1812,"Background color of warning text in the editor. The color must not be opaque so as not to hide underlying decorations."),!0),_h=T("editorWarning.foreground",{dark:"#CCA700",light:"#BF8803",hcDark:"#FFD370",hcLight:"#895503"},m(1813,"Foreground color of warning squigglies in the editor.")),aS=T("editorWarning.border",{dark:null,light:null,hcDark:q.fromHex("#FFCC00").transparent(.8),hcLight:q.fromHex("#FFCC00").transparent(.8)},m(1814,"If set, color of double underlines for warnings in the editor."));T("editorInfo.background",null,m(1815,"Background color of info text in the editor. The color must not be opaque so as not to hide underlying decorations."),!0);const Rc=T("editorInfo.foreground",{dark:"#3794FF",light:"#1a85ff",hcDark:"#3794FF",hcLight:"#1a85ff"},m(1816,"Foreground color of info squigglies in the editor.")),lS=T("editorInfo.border",{dark:null,light:null,hcDark:q.fromHex("#3794FF").transparent(.8),hcLight:"#292929"},m(1817,"If set, color of double underlines for infos in the editor.")),rme=T("editorHint.foreground",{dark:q.fromHex("#eeeeee").transparent(.7),light:"#6c6c6c",hcDark:null,hcLight:null},m(1818,"Foreground color of hint squigglies in the editor."));T("editorHint.border",{dark:null,light:null,hcDark:q.fromHex("#eeeeee").transparent(.8),hcLight:"#292929"},m(1819,"If set, color of double underlines for hints in the editor."));const ame=T("editorLink.activeForeground",{dark:"#4E94CE",light:q.blue,hcDark:q.cyan,hcLight:"#292929"},m(1820,"Color of active links.")),mu=T("editor.selectionBackground",{light:"#ADD6FF",dark:"#264F78",hcDark:"#f3f518",hcLight:"#0F4A85"},m(1821,"Color of the editor selection.")),lme=T("editor.selectionForeground",{light:null,dark:null,hcDark:"#000000",hcLight:q.white},m(1822,"Color of the selected text for high contrast.")),OJ=T("editor.inactiveSelectionBackground",{light:Be(mu,.5),dark:Be(mu,.5),hcDark:Be(mu,.7),hcLight:Be(mu,.5)},m(1823,"Color of the selection in an inactive editor. The color must not be opaque so as not to hide underlying decorations."),!0),NW=T("editor.selectionHighlightBackground",{light:D$(mu,Bi,.3,.6),dark:D$(mu,Bi,.3,.6),hcDark:null,hcLight:null},m(1824,"Color for regions with the same content as the selection. The color must not be opaque so as not to hide underlying decorations."),!0);T("editor.selectionHighlightBorder",{light:null,dark:null,hcDark:ri,hcLight:ri},m(1825,"Border color for regions with the same content as the selection."));T("editor.compositionBorder",{light:"#000000",dark:"#ffffff",hcLight:"#000000",hcDark:"#ffffff"},m(1826,"The border color for an IME composition."));T("editor.findMatchBackground",{light:"#A8AC94",dark:"#515C6A",hcDark:null,hcLight:null},m(1827,"Color of the current search match."));const cme=T("editor.findMatchForeground",null,m(1828,"Text color of the current search match.")),zd=T("editor.findMatchHighlightBackground",{light:"#EA5C0055",dark:"#EA5C0055",hcDark:null,hcLight:null},m(1829,"Color of the other search matches. The color must not be opaque so as not to hide underlying decorations."),!0),dme=T("editor.findMatchHighlightForeground",null,m(1830,"Foreground color of the other search matches."),!0);T("editor.findRangeHighlightBackground",{dark:"#3a3d4166",light:"#b4b4b44d",hcDark:null,hcLight:null},m(1831,"Color of the range limiting the search. The color must not be opaque so as not to hide underlying decorations."),!0);T("editor.findMatchBorder",{light:null,dark:null,hcDark:ri,hcLight:ri},m(1832,"Border color of the current search match."));const _u=T("editor.findMatchHighlightBorder",{light:null,dark:null,hcDark:ri,hcLight:ri},m(1833,"Border color of the other search matches.")),hme=T("editor.findRangeHighlightBorder",{dark:null,light:null,hcDark:Be(ri,.4),hcLight:Be(ri,.4)},m(1834,"Border color of the range limiting the search. The color must not be opaque so as not to hide underlying decorations."),!0);T("editor.hoverHighlightBackground",{light:"#ADD6FF26",dark:"#264f7840",hcDark:"#ADD6FF26",hcLight:null},m(1835,"Highlight below the word for which a hover is shown. The color must not be opaque so as not to hide underlying decorations."),!0);const UE=T("editorHoverWidget.background",wr,m(1836,"Background color of the editor hover.")),N$=T("editorHoverWidget.foreground",tA,m(1837,"Foreground color of the editor hover.")),RW=T("editorHoverWidget.border",TW,m(1838,"Border color of the editor hover."));T("editorHoverWidget.statusBarBackground",{dark:vl(UE,.2),light:mo(UE,.05),hcDark:wr,hcLight:wr},m(1839,"Background color of the editor hover status bar."));const AW=T("editorInlayHint.foreground",{dark:"#969696",light:"#969696",hcDark:q.white,hcLight:q.black},m(1840,"Foreground color of inline hints")),MW=T("editorInlayHint.background",{dark:Be(Tk,.1),light:Be(Tk,.1),hcDark:Be(q.white,.1),hcLight:Be(Tk,.1)},m(1841,"Background color of inline hints")),ume=T("editorInlayHint.typeForeground",AW,m(1842,"Foreground color of inline hints for types")),gme=T("editorInlayHint.typeBackground",MW,m(1843,"Background color of inline hints for types")),fme=T("editorInlayHint.parameterForeground",AW,m(1844,"Foreground color of inline hints for parameters")),pme=T("editorInlayHint.parameterBackground",MW,m(1845,"Background color of inline hints for parameters")),mme=T("editorLightBulb.foreground",{dark:"#FFCC00",light:"#DDB100",hcDark:"#FFCC00",hcLight:"#007ACC"},m(1846,"The color used for the lightbulb actions icon."));T("editorLightBulbAutoFix.foreground",{dark:"#75BEFF",light:"#007ACC",hcDark:"#75BEFF",hcLight:"#007ACC"},m(1847,"The color used for the lightbulb auto fix actions icon."));T("editorLightBulbAi.foreground",mme,m(1848,"The color used for the lightbulb AI icon."));T("editor.snippetTabstopHighlightBackground",{dark:new q(new G(124,124,124,.3)),light:new q(new G(10,50,100,.2)),hcDark:new q(new G(124,124,124,.3)),hcLight:new q(new G(10,50,100,.2))},m(1849,"Highlight background color of a snippet tabstop."));T("editor.snippetTabstopHighlightBorder",null,m(1850,"Highlight border color of a snippet tabstop."));T("editor.snippetFinalTabstopHighlightBackground",null,m(1851,"Highlight background color of the final tabstop of a snippet."));T("editor.snippetFinalTabstopHighlightBorder",{dark:"#525252",light:new q(new G(10,50,100,.5)),hcDark:"#525252",hcLight:"#292929"},m(1852,"Highlight border color of the final tabstop of a snippet."));const GP=new q(new G(155,185,85,.2)),KP=new q(new G(255,0,0,.2)),qf=T("diffEditor.insertedTextBackground",{dark:"#9ccc2c33",light:"#9ccc2c40",hcDark:null,hcLight:null},m(1853,"Background color for text that got inserted. The color must not be opaque so as not to hide underlying decorations."),!0),bu=T("diffEditor.removedTextBackground",{dark:"#ff000033",light:"#ff000033",hcDark:null,hcLight:null},m(1854,"Background color for text that got removed. The color must not be opaque so as not to hide underlying decorations."),!0),Xx=T("diffEditor.insertedLineBackground",{dark:GP,light:GP,hcDark:null,hcLight:null},m(1855,"Background color for lines that got inserted. The color must not be opaque so as not to hide underlying decorations."),!0);T("diffEditor.removedLineBackground",{dark:KP,light:KP,hcDark:null,hcLight:null},m(1856,"Background color for lines that got removed. The color must not be opaque so as not to hide underlying decorations."),!0);T("diffEditorGutter.insertedLineBackground",null,m(1857,"Background color for the margin where lines got inserted."));T("diffEditorGutter.removedLineBackground",null,m(1858,"Background color for the margin where lines got removed."));const _me=T("diffEditorOverview.insertedForeground",null,m(1859,"Diff overview ruler foreground for inserted content.")),bme=T("diffEditorOverview.removedForeground",null,m(1860,"Diff overview ruler foreground for removed content."));T("diffEditor.insertedTextBorder",{dark:null,light:null,hcDark:"#33ff2eff",hcLight:"#374E06"},m(1861,"Outline color for the text that got inserted."));T("diffEditor.removedTextBorder",{dark:null,light:null,hcDark:"#FF008F",hcLight:"#AD0707"},m(1862,"Outline color for text that got removed."));T("diffEditor.border",{dark:null,light:null,hcDark:rt,hcLight:rt},m(1863,"Border color between the two text editors."));T("diffEditor.diagonalFill",{dark:"#cccccc33",light:"#22222233",hcDark:null,hcLight:null},m(1864,"Color of the diff editor's diagonal fill. The diagonal fill is used in side-by-side diff views."));T("diffEditor.unchangedRegionBackground","sideBar.background",m(1865,"The background color of unchanged blocks in the diff editor."));T("diffEditor.unchangedRegionForeground","foreground",m(1866,"The foreground color of unchanged blocks in the diff editor."));T("diffEditor.unchangedCodeBackground",{dark:"#74747429",light:"#b8b8b829",hcDark:null,hcLight:null},m(1867,"The background color of unchanged code in the diff editor."));const tC=T("widget.shadow",{dark:Be(q.black,.36),light:Be(q.black,.16),hcDark:null,hcLight:null},m(1868,"Shadow color of widgets such as find/replace inside the editor.")),OW=T("widget.border",{dark:null,light:null,hcDark:rt,hcLight:rt},m(1869,"Border color of widgets such as find/replace inside the editor.")),R$=T("toolbar.hoverBackground",{dark:"#5a5d5e50",light:"#b8b8b850",hcDark:null,hcLight:null},m(1870,"Toolbar background when hovering over actions using the mouse"));T("toolbar.hoverOutline",{dark:null,light:null,hcDark:ri,hcLight:ri},m(1871,"Toolbar outline when hovering over actions using the mouse"));T("toolbar.activeBackground",{dark:vl(R$,.1),light:mo(R$,.1),hcDark:null,hcLight:null},m(1872,"Toolbar background when holding the mouse over actions"));const wme=T("breadcrumb.foreground",Be(et,.8),m(1873,"Color of focused breadcrumb items.")),vme=T("breadcrumb.background",Bi,m(1874,"Background color of breadcrumb items.")),A$=T("breadcrumb.focusForeground",{light:mo(et,.2),dark:vl(et,.1),hcDark:vl(et,.1),hcLight:vl(et,.1)},m(1875,"Color of focused breadcrumb items.")),Cme=T("breadcrumb.activeSelectionForeground",{light:mo(et,.2),dark:vl(et,.1),hcDark:vl(et,.1),hcLight:vl(et,.1)},m(1876,"Color of selected breadcrumb items."));T("breadcrumbPicker.background",wr,m(1877,"Background color of breadcrumb item picker."));const PJ=.5,M$=q.fromHex("#40C8AE").transparent(PJ),O$=q.fromHex("#40A6FF").transparent(PJ),P$=q.fromHex("#606060").transparent(.4),PW=.4,lv=1,YP=T("merge.currentHeaderBackground",{dark:M$,light:M$,hcDark:null,hcLight:null},m(1878,"Current header background in inline merge-conflicts. The color must not be opaque so as not to hide underlying decorations."),!0);T("merge.currentContentBackground",Be(YP,PW),m(1879,"Current content background in inline merge-conflicts. The color must not be opaque so as not to hide underlying decorations."),!0);const ZP=T("merge.incomingHeaderBackground",{dark:O$,light:O$,hcDark:null,hcLight:null},m(1880,"Incoming header background in inline merge-conflicts. The color must not be opaque so as not to hide underlying decorations."),!0);T("merge.incomingContentBackground",Be(ZP,PW),m(1881,"Incoming content background in inline merge-conflicts. The color must not be opaque so as not to hide underlying decorations."),!0);const XP=T("merge.commonHeaderBackground",{dark:P$,light:P$,hcDark:null,hcLight:null},m(1882,"Common ancestor header background in inline merge-conflicts. The color must not be opaque so as not to hide underlying decorations."),!0);T("merge.commonContentBackground",Be(XP,PW),m(1883,"Common ancestor content background in inline merge-conflicts. The color must not be opaque so as not to hide underlying decorations."),!0);const cv=T("merge.border",{dark:null,light:null,hcDark:"#C3DF6F",hcLight:"#007ACC"},m(1884,"Border color on headers and the splitter in inline merge-conflicts."));T("editorOverviewRuler.currentContentForeground",{dark:Be(YP,lv),light:Be(YP,lv),hcDark:cv,hcLight:cv},m(1885,"Current overview ruler foreground for inline merge-conflicts."));T("editorOverviewRuler.incomingContentForeground",{dark:Be(ZP,lv),light:Be(ZP,lv),hcDark:cv,hcLight:cv},m(1886,"Incoming overview ruler foreground for inline merge-conflicts."));T("editorOverviewRuler.commonContentForeground",{dark:Be(XP,lv),light:Be(XP,lv),hcDark:cv,hcLight:cv},m(1887,"Common ancestor overview ruler foreground for inline merge-conflicts."));const eM=T("editorOverviewRuler.findMatchForeground",{dark:"#d186167e",light:"#d186167e",hcDark:"#AB5A00",hcLight:"#AB5A00"},m(1888,"Overview ruler marker color for find matches. The color must not be opaque so as not to hide underlying decorations."),!0),FJ=T("editorOverviewRuler.selectionHighlightForeground","#A0A0A0CC",m(1889,"Overview ruler marker color for selection highlights. The color must not be opaque so as not to hide underlying decorations."),!0),yme=T("problemsErrorIcon.foreground",iA,m(1890,"The color used for the problems error icon.")),Sme=T("problemsWarningIcon.foreground",_h,m(1891,"The color used for the problems warning icon.")),xme=T("problemsInfoIcon.foreground",Rc,m(1892,"The color used for the problems info icon.")),QP=T("minimap.findMatchHighlight",{light:"#d18616",dark:"#d18616",hcDark:"#AB5A00",hcLight:"#0F4A85"},m(1983,"Minimap marker color for find matches."),!0),nA=T("minimap.selectionOccurrenceHighlight",{light:"#c9c9c9",dark:"#676767",hcDark:"#ffffff",hcLight:"#0F4A85"},m(1984,"Minimap marker color for repeating editor selections."),!0),F$=T("minimap.selectionHighlight",{light:"#ADD6FF",dark:"#264F78",hcDark:"#ffffff",hcLight:"#0F4A85"},m(1985,"Minimap marker color for the editor selection."),!0),Lme=T("minimap.infoHighlight",{dark:Rc,light:Rc,hcDark:lS,hcLight:lS},m(1986,"Minimap marker color for infos.")),kme=T("minimap.warningHighlight",{dark:_h,light:_h,hcDark:aS,hcLight:aS},m(1987,"Minimap marker color for warnings.")),Eme=T("minimap.errorHighlight",{dark:new q(new G(255,18,18,.7)),light:new q(new G(255,18,18,.7)),hcDark:new q(new G(255,50,50,1)),hcLight:"#B5200D"},m(1988,"Minimap marker color for errors.")),Ime=T("minimap.background",null,m(1989,"Minimap background color.")),Dme=T("minimap.foregroundOpacity",q.fromHex("#000f"),m(1990,'Opacity of foreground elements rendered in the minimap. For example, "#000000c0" will render the elements with 75% opacity.'));T("minimapSlider.background",Be(RJ,.5),m(1991,"Minimap slider background color."));T("minimapSlider.hoverBackground",Be(AJ,.5),m(1992,"Minimap slider background color when hovering."));T("minimapSlider.activeBackground",Be(MJ,.5),m(1993,"Minimap slider background color when clicked on."));T("charts.foreground",et,m(1790,"The foreground color used in charts."));T("charts.lines",Be(et,.5),m(1791,"The color used for horizontal lines in charts."));T("charts.red",iA,m(1792,"The red color used in chart visualizations."));T("charts.blue",Rc,m(1793,"The blue color used in chart visualizations."));T("charts.yellow",_h,m(1794,"The yellow color used in chart visualizations."));T("charts.orange",QP,m(1795,"The orange color used in chart visualizations."));T("charts.green",{dark:"#89D185",light:"#388A34",hcDark:"#89D185",hcLight:"#374e06"},m(1796,"The green color used in chart visualizations."));T("charts.purple",{dark:"#B180D7",light:"#652D90",hcDark:"#B180D7",hcLight:"#652D90"},m(1797,"The purple color used in chart visualizations."));const JP=T("input.background",{dark:"#3C3C3C",light:q.white,hcDark:q.black,hcLight:q.white},m(1893,"Input box background.")),BJ=T("input.foreground",et,m(1894,"Input box foreground.")),WJ=T("input.border",{dark:null,light:null,hcDark:rt,hcLight:rt},m(1895,"Input box border.")),X0=T("inputOption.activeBorder",{dark:"#007ACC",light:"#007ACC",hcDark:rt,hcLight:rt},m(1896,"Border color of activated options in input fields.")),Tme=T("inputOption.hoverBackground",{dark:"#5a5d5e80",light:"#b8b8b850",hcDark:null,hcLight:null},m(1897,"Background color of activated options in input fields.")),iC=T("inputOption.activeBackground",{dark:Be(Nc,.4),light:Be(Nc,.2),hcDark:q.transparent,hcLight:q.transparent},m(1898,"Background hover color of options in input fields.")),Q0=T("inputOption.activeForeground",{dark:q.white,light:q.black,hcDark:et,hcLight:et},m(1899,"Foreground color of activated options in input fields."));T("input.placeholderForeground",{light:Be(et,.5),dark:Be(et,.5),hcDark:Be(et,.7),hcLight:Be(et,.7)},m(1900,"Input box foreground color for placeholder text."));const Nme=T("inputValidation.infoBackground",{dark:"#063B49",light:"#D6ECF2",hcDark:q.black,hcLight:q.white},m(1901,"Input validation background color for information severity.")),Rme=T("inputValidation.infoForeground",{dark:null,light:null,hcDark:null,hcLight:et},m(1902,"Input validation foreground color for information severity.")),Ame=T("inputValidation.infoBorder",{dark:"#007acc",light:"#007acc",hcDark:rt,hcLight:rt},m(1903,"Input validation border color for information severity.")),Mme=T("inputValidation.warningBackground",{dark:"#352A05",light:"#F6F5D2",hcDark:q.black,hcLight:q.white},m(1904,"Input validation background color for warning severity.")),Ome=T("inputValidation.warningForeground",{dark:null,light:null,hcDark:null,hcLight:et},m(1905,"Input validation foreground color for warning severity.")),Pme=T("inputValidation.warningBorder",{dark:"#B89500",light:"#B89500",hcDark:rt,hcLight:rt},m(1906,"Input validation border color for warning severity.")),Fme=T("inputValidation.errorBackground",{dark:"#5A1D1D",light:"#F2DEDE",hcDark:q.black,hcLight:q.white},m(1907,"Input validation background color for error severity.")),Bme=T("inputValidation.errorForeground",{dark:null,light:null,hcDark:null,hcLight:et},m(1908,"Input validation foreground color for error severity.")),Wme=T("inputValidation.errorBorder",{dark:"#BE1100",light:"#BE1100",hcDark:rt,hcLight:rt},m(1909,"Input validation border color for error severity.")),sA=T("dropdown.background",{dark:"#3C3C3C",light:q.white,hcDark:q.black,hcLight:q.white},m(1910,"Dropdown background.")),Hme=T("dropdown.listBackground",{dark:null,light:null,hcDark:q.black,hcLight:q.white},m(1911,"Dropdown list background.")),FW=T("dropdown.foreground",{dark:"#F0F0F0",light:et,hcDark:q.white,hcLight:et},m(1912,"Dropdown foreground.")),BW=T("dropdown.border",{dark:sA,light:"#CECECE",hcDark:rt,hcLight:rt},m(1913,"Dropdown border.")),oA=T("button.foreground",q.white,m(1914,"Button foreground color.")),Vme=T("button.separator",Be(oA,.4),m(1915,"Button separator color.")),jf=T("button.background",{dark:"#0E639C",light:"#007ACC",hcDark:q.black,hcLight:"#0F4A85"},m(1916,"Button background color.")),zme=T("button.hoverBackground",{dark:vl(jf,.2),light:mo(jf,.2),hcDark:jf,hcLight:jf},m(1917,"Button background color when hovering.")),$me=T("button.border",rt,m(1918,"Button border color.")),HJ=T("button.secondaryForeground",{dark:q.white,light:q.white,hcDark:q.white,hcLight:et},m(1919,"Secondary button foreground color.")),qE=T("button.secondaryBackground",{dark:"#3A3D41",light:"#5F6A79",hcDark:null,hcLight:q.white},m(1920,"Secondary button background color.")),Ume=T("button.secondaryHoverBackground",{dark:vl(qE,.2),light:mo(qE,.2),hcDark:null,hcLight:null},m(1921,"Secondary button background color when hovering.")),a1=T("radio.activeForeground",Q0,m(1922,"Foreground color of active radio option.")),qme=T("radio.activeBackground",iC,m(1923,"Background color of active radio option.")),jme=T("radio.activeBorder",X0,m(1924,"Border color of the active radio option.")),Gme=T("radio.inactiveForeground",null,m(1925,"Foreground color of inactive radio option.")),Kme=T("radio.inactiveBackground",null,m(1926,"Background color of inactive radio option.")),Yme=T("radio.inactiveBorder",{light:Be(a1,.2),dark:Be(a1,.2),hcDark:Be(a1,.4),hcLight:Be(a1,.2)},m(1927,"Border color of the inactive radio option.")),Zme=T("radio.inactiveHoverBackground",Tme,m(1928,"Background color of inactive active radio option when hovering.")),WW=T("checkbox.background",sA,m(1929,"Background color of checkbox widget."));T("checkbox.selectBackground",wr,m(1930,"Background color of checkbox widget when the element it's in is selected."));const HW=T("checkbox.foreground",FW,m(1931,"Foreground color of checkbox widget.")),Xme=T("checkbox.border",BW,m(1932,"Border color of checkbox widget."));T("checkbox.selectBorder",$E,m(1933,"Border color of checkbox widget when the element it's in is selected."));const Qme=T("checkbox.disabled.background",{op:7,color:WW,with:HW,ratio:.33},m(1934,"Background of a disabled checkbox.")),Jme=T("checkbox.disabled.foreground",{op:7,color:HW,with:WW,ratio:.33},m(1935,"Foreground of a disabled checkbox.")),e_e=T("keybindingLabel.background",{dark:new q(new G(128,128,128,.17)),light:new q(new G(221,221,221,.4)),hcDark:q.transparent,hcLight:q.transparent},m(1936,"Keybinding label background color. The keybinding label is used to represent a keyboard shortcut.")),t_e=T("keybindingLabel.foreground",{dark:q.fromHex("#CCCCCC"),light:q.fromHex("#555555"),hcDark:q.white,hcLight:et},m(1937,"Keybinding label foreground color. The keybinding label is used to represent a keyboard shortcut.")),i_e=T("keybindingLabel.border",{dark:new q(new G(51,51,51,.6)),light:new q(new G(204,204,204,.4)),hcDark:new q(new G(111,195,223)),hcLight:rt},m(1938,"Keybinding label border color. The keybinding label is used to represent a keyboard shortcut.")),n_e=T("keybindingLabel.bottomBorder",{dark:new q(new G(68,68,68,.6)),light:new q(new G(187,187,187,.4)),hcDark:new q(new G(111,195,223)),hcLight:et},m(1939,"Keybinding label border bottom color. The keybinding label is used to represent a keyboard shortcut.")),s_e=T("list.focusBackground",null,m(1940,"List/Tree background color for the focused item when the list/tree is active. An active list/tree has keyboard focus, an inactive does not.")),o_e=T("list.focusForeground",null,m(1941,"List/Tree foreground color for the focused item when the list/tree is active. An active list/tree has keyboard focus, an inactive does not.")),r_e=T("list.focusOutline",{dark:Nc,light:Nc,hcDark:ri,hcLight:ri},m(1942,"List/Tree outline color for the focused item when the list/tree is active. An active list/tree has keyboard focus, an inactive does not.")),a_e=T("list.focusAndSelectionOutline",null,m(1943,"List/Tree outline color for the focused item when the list/tree is active and selected. An active list/tree has keyboard focus, an inactive does not.")),Im=T("list.activeSelectionBackground",{dark:"#04395E",light:"#0060C0",hcDark:null,hcLight:q.fromHex("#0F4A85").transparent(.1)},m(1944,"List/Tree background color for the selected item when the list/tree is active. An active list/tree has keyboard focus, an inactive does not.")),cS=T("list.activeSelectionForeground",{dark:q.white,light:q.white,hcDark:null,hcLight:null},m(1945,"List/Tree foreground color for the selected item when the list/tree is active. An active list/tree has keyboard focus, an inactive does not.")),VJ=T("list.activeSelectionIconForeground",null,m(1946,"List/Tree icon foreground color for the selected item when the list/tree is active. An active list/tree has keyboard focus, an inactive does not.")),l_e=T("list.inactiveSelectionBackground",{dark:"#37373D",light:"#E4E6F1",hcDark:null,hcLight:q.fromHex("#0F4A85").transparent(.1)},m(1947,"List/Tree background color for the selected item when the list/tree is inactive. An active list/tree has keyboard focus, an inactive does not.")),c_e=T("list.inactiveSelectionForeground",null,m(1948,"List/Tree foreground color for the selected item when the list/tree is inactive. An active list/tree has keyboard focus, an inactive does not.")),d_e=T("list.inactiveSelectionIconForeground",null,m(1949,"List/Tree icon foreground color for the selected item when the list/tree is inactive. An active list/tree has keyboard focus, an inactive does not.")),h_e=T("list.inactiveFocusBackground",null,m(1950,"List/Tree background color for the focused item when the list/tree is inactive. An active list/tree has keyboard focus, an inactive does not.")),u_e=T("list.inactiveFocusOutline",null,m(1951,"List/Tree outline color for the focused item when the list/tree is inactive. An active list/tree has keyboard focus, an inactive does not.")),zJ=T("list.hoverBackground",{dark:"#2A2D2E",light:"#F0F0F0",hcDark:q.white.transparent(.1),hcLight:q.fromHex("#0F4A85").transparent(.1)},m(1952,"List/Tree background when hovering over items using the mouse.")),$J=T("list.hoverForeground",null,m(1953,"List/Tree foreground when hovering over items using the mouse.")),g_e=T("list.dropBackground",{dark:"#062F4A",light:"#D6EBFF",hcDark:null,hcLight:null},m(1954,"List/Tree drag and drop background when moving items over other items when using the mouse.")),f_e=T("list.dropBetweenBackground",{dark:$E,light:$E,hcDark:null,hcLight:null},m(1955,"List/Tree drag and drop border color when moving items between items when using the mouse.")),Lb=T("list.highlightForeground",{dark:"#2AAAFF",light:"#0066BF",hcDark:Nc,hcLight:Nc},m(1956,"List/Tree foreground color of the match highlights when searching inside the list/tree.")),p_e=T("list.focusHighlightForeground",{dark:Lb,light:eme(Im,Lb,"#BBE7FF"),hcDark:Lb,hcLight:Lb},m(1957,"List/Tree foreground color of the match highlights on actively focused items when searching inside the list/tree."));T("list.invalidItemForeground",{dark:"#B89500",light:"#B89500",hcDark:"#B89500",hcLight:"#B5200D"},m(1958,"List/Tree foreground color for invalid items, for example an unresolved root in explorer."));T("list.errorForeground",{dark:"#F88070",light:"#B01011",hcDark:null,hcLight:null},m(1959,"Foreground color of list items containing errors."));T("list.warningForeground",{dark:"#CCA700",light:"#855F00",hcDark:null,hcLight:null},m(1960,"Foreground color of list items containing warnings."));const m_e=T("listFilterWidget.background",{light:mo(wr,0),dark:vl(wr,0),hcDark:wr,hcLight:wr},m(1961,"Background color of the type filter widget in lists and trees.")),__e=T("listFilterWidget.outline",{dark:q.transparent,light:q.transparent,hcDark:"#f38518",hcLight:"#007ACC"},m(1962,"Outline color of the type filter widget in lists and trees.")),b_e=T("listFilterWidget.noMatchesOutline",{dark:"#BE1100",light:"#BE1100",hcDark:rt,hcLight:rt},m(1963,"Outline color of the type filter widget in lists and trees, when there are no matches.")),w_e=T("listFilterWidget.shadow",tC,m(1964,"Shadow color of the type filter widget in lists and trees."));T("list.filterMatchBackground",{dark:zd,light:zd,hcDark:null,hcLight:null},m(1965,"Background color of the filtered match."));T("list.filterMatchBorder",{dark:_u,light:_u,hcDark:rt,hcLight:ri},m(1966,"Border color of the filtered match."));T("list.deemphasizedForeground",{dark:"#8C8C8C",light:"#8E8E90",hcDark:"#A7A8A9",hcLight:"#666666"},m(1967,"List/Tree foreground color for items that are deemphasized."));const UJ=T("tree.indentGuidesStroke",{dark:"#585858",light:"#a9a9a9",hcDark:"#a9a9a9",hcLight:"#a5a5a5"},m(1968,"Tree stroke color for the indentation guides.")),v_e=T("tree.inactiveIndentGuidesStroke",Be(UJ,.4),m(1969,"Tree stroke color for the indentation guides that are not active.")),C_e=T("tree.tableColumnsBorder",{dark:"#CCCCCC20",light:"#61616120",hcDark:null,hcLight:null},m(1970,"Table border color between columns.")),y_e=T("tree.tableOddRowsBackground",{dark:Be(et,.04),light:Be(et,.04),hcDark:null,hcLight:null},m(1971,"Background color for odd table rows."));T("editorActionList.background",wr,m(1972,"Action List background color."));const S_e=T("editorActionList.foreground",tA,m(1973,"Action List foreground color."));T("editorActionList.focusForeground",cS,m(1974,"Action List foreground color for the focused item."));T("editorActionList.focusBackground",Im,m(1975,"Action List background color for the focused item."));const x_e=T("menu.border",{dark:null,light:null,hcDark:rt,hcLight:rt},m(1976,"Border color of menus.")),L_e=T("menu.foreground",FW,m(1977,"Foreground color of menu items.")),k_e=T("menu.background",sA,m(1978,"Background color of menu items.")),E_e=T("menu.selectionForeground",cS,m(1979,"Foreground color of the selected menu item in menus.")),I_e=T("menu.selectionBackground",Im,m(1980,"Background color of the selected menu item in menus.")),D_e=T("menu.selectionBorder",{dark:null,light:null,hcDark:ri,hcLight:ri},m(1981,"Border color of the selected menu item in menus.")),T_e=T("menu.separatorBackground",{dark:"#606060",light:"#D4D4D4",hcDark:rt,hcLight:rt},m(1982,"Color of a separator menu item in menus.")),tM=T("quickInput.background",wr,m(2010,"Quick picker background color. The quick picker widget is the container for pickers like the command palette.")),N_e=T("quickInput.foreground",tA,m(2011,"Quick picker foreground color. The quick picker widget is the container for pickers like the command palette.")),R_e=T("quickInputTitle.background",{dark:new q(new G(255,255,255,.105)),light:new q(new G(0,0,0,.06)),hcDark:"#000000",hcLight:q.white},m(2012,"Quick picker title background color. The quick picker widget is the container for pickers like the command palette.")),qJ=T("pickerGroup.foreground",{dark:"#3794FF",light:"#0066BF",hcDark:q.white,hcLight:"#0F4A85"},m(2013,"Quick picker color for grouping labels.")),A_e=T("pickerGroup.border",{dark:"#3F3F46",light:"#CCCEDB",hcDark:q.white,hcLight:"#0F4A85"},m(2014,"Quick picker color for grouping borders.")),B$=T("quickInput.list.focusBackground",null,"",void 0,m(2015,"Please use quickInputList.focusBackground instead")),dS=T("quickInputList.focusForeground",cS,m(2016,"Quick picker foreground color for the focused item.")),VW=T("quickInputList.focusIconForeground",VJ,m(2017,"Quick picker icon foreground color for the focused item.")),hS=T("quickInputList.focusBackground",{dark:rS(B$,Im),light:rS(B$,Im),hcDark:null,hcLight:null},m(2018,"Quick picker background color for the focused item."));T("search.resultsInfoForeground",{light:et,dark:Be(et,.65),hcDark:et,hcLight:et},m(2019,"Color of the text in the search viewlet's completion message."));T("searchEditor.findMatchBackground",{light:Be(zd,.66),dark:Be(zd,.66),hcDark:zd,hcLight:zd},m(2020,"Color of the Search Editor query matches."));T("searchEditor.findMatchBorder",{light:Be(_u,.66),dark:Be(_u,.66),hcDark:_u,hcLight:_u},m(2021,"Border color of the Search Editor query matches."));class J0{constructor(e,t){this.x=e,this.y=t,this._pageCoordinatesBrand=void 0}toClientCoordinates(e){return new jJ(this.x-e.scrollX,this.y-e.scrollY)}}class jJ{constructor(e,t){this.clientX=e,this.clientY=t,this._clientCoordinatesBrand=void 0}toPageCoordinates(e){return new J0(this.clientX+e.scrollX,this.clientY+e.scrollY)}}class M_e{constructor(e,t,i,n){this.x=e,this.y=t,this.width=i,this.height=n,this._editorPagePositionBrand=void 0}}class O_e{constructor(e,t){this.x=e,this.y=t,this._positionRelativeToEditorBrand=void 0}}function rA(s){const e=Li(s);return new M_e(e.left,e.top,e.width,e.height)}function aA(s,e,t){const i=e.width/s.offsetWidth,n=e.height/s.offsetHeight,o=(t.x-e.x)/i,r=(t.y-e.y)/n;return new O_e(o,r)}class bh extends is{constructor(e,t,i){super(be(i),e),this._editorMouseEventBrand=void 0,this.isFromPointerCapture=t,this.pos=new J0(this.posx,this.posy),this.editorPos=rA(i),this.relativePos=aA(i,this.editorPos,this.pos)}}class P_e{constructor(e){this._editorViewDomNode=e}_create(e){return new bh(e,!1,this._editorViewDomNode)}onContextMenu(e,t){return W(e,se.CONTEXT_MENU,i=>{t(this._create(i))})}onMouseUp(e,t){return W(e,se.MOUSE_UP,i=>{t(this._create(i))})}onMouseDown(e,t){return W(e,se.MOUSE_DOWN,i=>{t(this._create(i))})}onPointerDown(e,t){return W(e,se.POINTER_DOWN,i=>{t(this._create(i),i.pointerId)})}onMouseLeave(e,t){return W(e,se.MOUSE_LEAVE,i=>{t(this._create(i))})}onMouseMove(e,t){return W(e,se.MOUSE_MOVE,i=>t(this._create(i)))}}class F_e{constructor(e){this._editorViewDomNode=e}_create(e){return new bh(e,!1,this._editorViewDomNode)}onPointerUp(e,t){return W(e,"pointerup",i=>{t(this._create(i))})}onPointerDown(e,t){return W(e,se.POINTER_DOWN,i=>{t(this._create(i),i.pointerId)})}onPointerLeave(e,t){return W(e,se.POINTER_LEAVE,i=>{t(this._create(i))})}onPointerMove(e,t){return W(e,"pointermove",i=>t(this._create(i)))}}class B_e extends F{constructor(e){super(),this._editorViewDomNode=e,this._globalPointerMoveMonitor=this._register(new eC),this._keydownListener=null}startMonitoring(e,t,i,n,o){this._keydownListener=Fi(e.ownerDocument,"keydown",r=>{r.toKeyCodeChord().isModifierKey()||this._globalPointerMoveMonitor.stopMonitoring(!0,r.browserEvent)},!0),this._globalPointerMoveMonitor.startMonitoring(e,t,i,r=>{n(new bh(r,!0,this._editorViewDomNode))},r=>{this._keydownListener.dispose(),o(r)})}stopMonitoring(){this._globalPointerMoveMonitor.stopMonitoring(!0)}}const qT=class qT{constructor(e){this._editor=e,this._instanceId=++qT._idPool,this._counter=0,this._rules=new LR,this._garbageCollectionScheduler=new Rt(()=>this.garbageCollect(),1e3)}dispose(){this._rules.dispose(),this._garbageCollectionScheduler.dispose()}createClassNameRef(e){const t=this.getOrCreateRule(e);return t.increaseRefCount(),{className:t.className,dispose:()=>{t.decreaseRefCount(),this._garbageCollectionScheduler.schedule()}}}getOrCreateRule(e){const t=this.computeUniqueKey(e);let i=this._rules.get(t);if(!i){const n=this._counter++;i=new W_e(t,`dyn-rule-${this._instanceId}-${n}`,LE(this._editor.getContainerDomNode())?this._editor.getContainerDomNode():void 0,e),this._rules.set(t,i)}return i}computeUniqueKey(e){return JSON.stringify(e)}garbageCollect(){for(const e of this._rules.values())e.hasReferences()||this._rules.deleteAndDispose(e.key)}};qT._idPool=0;let jE=qT;class W_e{constructor(e,t,i,n){this.key=e,this.className=t,this.properties=n,this._referenceCount=0,this._styleElementDisposables=new U,this._styleElement=ca(i,void 0,this._styleElementDisposables),this._styleElement.textContent=this.getCssText(this.className,this.properties)}getCssText(e,t){let i=`.${e} {`;for(const n in t){const o=t[n];let r;typeof o=="object"?r=te(o.id):r=o;const a=H_e(n);i+=`
	${a}: ${r};`}return i+=`
}`,i}dispose(){this._styleElementDisposables.dispose(),this._styleElement=void 0}increaseRefCount(){this._referenceCount++}decreaseRefCount(){this._referenceCount--}hasReferences(){return this._referenceCount>0}}function H_e(s){return s.replace(/(^[A-Z])/,([e])=>e.toLowerCase()).replace(/([A-Z])/g,([e])=>`-${e.toLowerCase()}`)}class c_ extends F{constructor(){super(),this._shouldRender=!0}shouldRender(){return this._shouldRender}forceShouldRender(){this._shouldRender=!0}setShouldRender(){this._shouldRender=!0}onDidRender(){this._shouldRender=!1}onCompositionStart(e){return!1}onCompositionEnd(e){return!1}onConfigurationChanged(e){return!1}onCursorStateChanged(e){return!1}onDecorationsChanged(e){return!1}onFlushed(e){return!1}onFocusChanged(e){return!1}onLanguageConfigurationChanged(e){return!1}onLineMappingChanged(e){return!1}onLinesChanged(e){return!1}onLinesDeleted(e){return!1}onLinesInserted(e){return!1}onRevealRangeRequest(e){return!1}onScrollChanged(e){return!1}onThemeChanged(e){return!1}onTokensChanged(e){return!1}onTokensColorsChanged(e){return!1}onZonesChanged(e){return!1}handleEvents(e){let t=!1;for(let i=0,n=e.length;i<n;i++){const o=e[i];switch(o.type){case 0:this.onCompositionStart(o)&&(t=!0);break;case 1:this.onCompositionEnd(o)&&(t=!0);break;case 2:this.onConfigurationChanged(o)&&(t=!0);break;case 3:this.onCursorStateChanged(o)&&(t=!0);break;case 4:this.onDecorationsChanged(o)&&(t=!0);break;case 5:this.onFlushed(o)&&(t=!0);break;case 6:this.onFocusChanged(o)&&(t=!0);break;case 7:this.onLanguageConfigurationChanged(o)&&(t=!0);break;case 8:this.onLineMappingChanged(o)&&(t=!0);break;case 9:this.onLinesChanged(o)&&(t=!0);break;case 10:this.onLinesDeleted(o)&&(t=!0);break;case 11:this.onLinesInserted(o)&&(t=!0);break;case 12:this.onRevealRangeRequest(o)&&(t=!0);break;case 13:this.onScrollChanged(o)&&(t=!0);break;case 15:this.onTokensChanged(o)&&(t=!0);break;case 14:this.onThemeChanged(o)&&(t=!0);break;case 16:this.onTokensColorsChanged(o)&&(t=!0);break;case 17:this.onZonesChanged(o)&&(t=!0);break;default:console.info("View received unknown event: "),console.info(o)}}t&&(this._shouldRender=!0)}}class no extends c_{constructor(e){super(),this._context=e,this._context.addEventHandler(this)}dispose(){this._context.removeEventHandler(this),super.dispose()}}class kl{static write(e,t){e.setAttribute("data-mprt",String(t))}static read(e){const t=e.getAttribute("data-mprt");return t===null?0:parseInt(t,10)}static collect(e,t){const i=[];let n=0;for(;e&&e!==e.ownerDocument.body&&e!==t;)e.nodeType===e.ELEMENT_NODE&&(i[n++]=this.read(e)),e=e.parentElement;const o=new Uint8Array(n);for(let r=0;r<n;r++)o[r]=i[n-r-1];return o}}class V_e{constructor(e,t){this._restrictedRenderingContextBrand=void 0,this._viewLayout=e,this.viewportData=t,this.scrollWidth=this._viewLayout.getScrollWidth(),this.scrollHeight=this._viewLayout.getScrollHeight(),this.visibleRange=this.viewportData.visibleRange,this.bigNumbersDelta=this.viewportData.bigNumbersDelta;const i=this._viewLayout.getCurrentViewport();this.scrollTop=i.top,this.scrollLeft=i.left,this.viewportWidth=i.width,this.viewportHeight=i.height}getScrolledTopFromAbsoluteTop(e){return e-this.scrollTop}getVerticalOffsetForLineNumber(e,t){return this._viewLayout.getVerticalOffsetForLineNumber(e,t)}getVerticalOffsetAfterLineNumber(e,t){return this._viewLayout.getVerticalOffsetAfterLineNumber(e,t)}getLineHeightForLineNumber(e){return this._viewLayout.getLineHeightForLineNumber(e)}getDecorationsInViewport(){return this.viewportData.getDecorationsInViewport()}}class z_e extends V_e{constructor(e,t,i,n){super(e,t),this._renderingContextBrand=void 0,this._viewLines=i,this._viewLinesGpu=n}linesVisibleRangesForRange(e,t){const i=this._viewLines.linesVisibleRangesForRange(e,t);if(!this._viewLinesGpu)return i??null;const n=this._viewLinesGpu.linesVisibleRangesForRange(e,t);return i?n?i.concat(n).sort((o,r)=>o.lineNumber-r.lineNumber):i:n}visibleRangeForPosition(e){return this._viewLines.visibleRangeForPosition(e)??this._viewLinesGpu?.visibleRangeForPosition(e)??null}}class GJ{constructor(e,t,i,n){this.outsideRenderedLine=e,this.lineNumber=t,this.ranges=i,this.continuesOnNextLine=n}}class ex{static from(e){const t=new Array(e.length);for(let i=0,n=e.length;i<n;i++){const o=e[i];t[i]=new ex(o.left,o.width)}return t}constructor(e,t){this._horizontalRangeBrand=void 0,this.left=Math.round(e),this.width=Math.round(t)}toString(){return`[${this.left},${this.width}]`}}class Ou{constructor(e,t){this._floatHorizontalRangeBrand=void 0,this.left=e,this.width=t}toString(){return`[${this.left},${this.width}]`}static compare(e,t){return e.left-t.left}}class KJ{constructor(e,t){this.outsideRenderedLine=e,this.originalLeft=t,this.left=Math.round(this.originalLeft)}}class eF{constructor(e,t){this.outsideRenderedLine=e,this.ranges=t}}class Nk{static _createRange(){return this._handyReadyRange||(this._handyReadyRange=document.createRange()),this._handyReadyRange}static _detachRange(e,t){e.selectNodeContents(t)}static _readClientRects(e,t,i,n,o){const r=this._createRange();try{return r.setStart(e,t),r.setEnd(i,n),r.getClientRects()}catch{return null}finally{this._detachRange(r,o)}}static _mergeAdjacentRanges(e){if(e.length===1)return e;e.sort(Ou.compare);const t=[];let i=0,n=e[0];for(let o=1,r=e.length;o<r;o++){const a=e[o];n.left+n.width+.9>=a.left?n.width=Math.max(n.width,a.left+a.width-n.left):(t[i++]=n,n=a)}return t[i++]=n,t}static _createHorizontalRangesFromClientRects(e,t,i){if(!e||e.length===0)return null;const n=[];for(let o=0,r=e.length;o<r;o++){const a=e[o];n[o]=new Ou(Math.max(0,(a.left-t)/i),a.width/i)}return this._mergeAdjacentRanges(n)}static readHorizontalRanges(e,t,i,n,o,r){const l=e.children.length-1;if(0>l)return null;if(t=Math.min(l,Math.max(0,t)),n=Math.min(l,Math.max(0,n)),t===n&&i===o&&i===0&&!e.children[t].firstChild){const u=e.children[t].getClientRects();return r.markDidDomLayout(),this._createHorizontalRangesFromClientRects(u,r.clientRectDeltaLeft,r.clientRectScale)}t!==n&&n>0&&o===0&&(n--,o=1073741824);let c=e.children[t].firstChild,d=e.children[n].firstChild;if((!c||!d)&&(!c&&i===0&&t>0&&(c=e.children[t-1].firstChild,i=1073741824),!d&&o===0&&n>0&&(d=e.children[n-1].firstChild,o=1073741824)),!c||!d)return null;i=Math.min(c.textContent.length,Math.max(0,i)),o=Math.min(d.textContent.length,Math.max(0,o));const h=this._readClientRects(c,i,d,o,r.endNode);return r.markDidDomLayout(),this._createHorizontalRangesFromClientRects(h,r.clientRectDeltaLeft,r.clientRectScale)}}class Rs{constructor(e,t,i,n){this.startColumn=e,this.endColumn=t,this.className=i,this.type=n,this._lineDecorationBrand=void 0}static _equals(e,t){return e.startColumn===t.startColumn&&e.endColumn===t.endColumn&&e.className===t.className&&e.type===t.type}static equalsArr(e,t){const i=e.length,n=t.length;if(i!==n)return!1;for(let o=0;o<i;o++)if(!Rs._equals(e[o],t[o]))return!1;return!0}static extractWrapped(e,t,i){if(e.length===0)return e;const n=t+1,o=i+1,r=i-t,a=[];let l=0;for(const c of e)c.endColumn<=n||c.startColumn>=o||(a[l++]=new Rs(Math.max(1,c.startColumn-n+1),Math.min(r+1,c.endColumn-n+1),c.className,c.type));return a}static filter(e,t,i,n){if(e.length===0)return[];const o=[];let r=0;for(let a=0,l=e.length;a<l;a++){const c=e[a],d=c.range;if(d.endLineNumber<t||d.startLineNumber>t||d.isEmpty()&&(c.type===0||c.type===3))continue;const h=d.startLineNumber===t?d.startColumn:i,u=d.endLineNumber===t?d.endColumn:n;o[r++]=new Rs(h,u,c.inlineClassName,c.type)}return o}static _typeCompare(e,t){const i=[2,0,1,3];return i[e]-i[t]}static compare(e,t){if(e.startColumn!==t.startColumn)return e.startColumn-t.startColumn;if(e.endColumn!==t.endColumn)return e.endColumn-t.endColumn;const i=Rs._typeCompare(e.type,t.type);return i!==0?i:e.className!==t.className?e.className<t.className?-1:1:0}}class W${constructor(e,t,i,n){this.startOffset=e,this.endOffset=t,this.className=i,this.metadata=n}}class GE{constructor(){this.stopOffsets=[],this.classNames=[],this.metadata=[],this.count=0}static _metadata(e){let t=0;for(let i=0,n=e.length;i<n;i++)t|=e[i];return t}consumeLowerThan(e,t,i){for(;this.count>0&&this.stopOffsets[0]<e;){let n=0;for(;n+1<this.count&&this.stopOffsets[n]===this.stopOffsets[n+1];)n++;i.push(new W$(t,this.stopOffsets[n],this.classNames.join(" "),GE._metadata(this.metadata))),t=this.stopOffsets[n]+1,this.stopOffsets.splice(0,n+1),this.classNames.splice(0,n+1),this.metadata.splice(0,n+1),this.count-=n+1}return this.count>0&&t<e&&(i.push(new W$(t,e-1,this.classNames.join(" "),GE._metadata(this.metadata))),t=e),t}insert(e,t,i){if(this.count===0||this.stopOffsets[this.count-1]<=e)this.stopOffsets.push(e),this.classNames.push(t),this.metadata.push(i);else for(let n=0;n<this.count;n++)if(this.stopOffsets[n]>=e){this.stopOffsets.splice(n,0,e),this.classNames.splice(n,0,t),this.metadata.splice(n,0,i);break}this.count++}}class $_e{static normalize(e,t){if(t.length===0)return[];const i=[],n=new GE;let o=0;for(let r=0,a=t.length;r<a;r++){const l=t[r];let c=l.startColumn,d=l.endColumn;const h=l.className,u=l.type===1?2:l.type===2?4:0;if(c>1){const p=e.charCodeAt(c-2);sn(p)&&c--}if(d>1){const p=e.charCodeAt(d-2);sn(p)&&d--}const g=c-1,f=d-2;o=n.consumeLowerThan(g,o,i),n.count===0&&(o=g),n.insert(f,h,u)}return n.consumeLowerThan(1073741824,o,i),i}}class on{constructor(e,t,i,n){this.endIndex=e,this.type=t,this.metadata=i,this.containsRTL=n,this._linePartBrand=void 0}isWhitespace(){return!!(this.metadata&1)}isPseudoAfter(){return!!(this.metadata&4)}}class Dh{get isLTR(){return!this.containsRTL&&this.textDirection!==br.RTL}constructor(e,t,i,n,o,r,a,l,c,d,h,u,g,f,p,_,b,w,C,v,y,x=!1){this.useMonospaceOptimizations=e,this.canUseHalfwidthRightwardsArrow=t,this.lineContent=i,this.continuesWithWrappedLine=n,this.isBasicASCII=o,this.containsRTL=r,this.fauxIndentLength=a,this.lineTokens=l,this.lineDecorations=c.sort(Rs.compare),this.tabSize=d,this.startVisibleColumn=h,this.spaceWidth=u,this.stopRenderingLineAfter=p,this.renderWhitespace=_==="all"?4:_==="boundary"?1:_==="selection"?2:_==="trailing"?3:0,this.renderControlCharacters=b,this.fontLigatures=w,this.selectionsOnLine=C&&C.sort((I,R)=>I.start<R.start?-1:1),this.renderNewLineWhenEmpty=x,this.textDirection=v,this.verticalScrollbarSize=y;const L=Math.abs(f-u),E=Math.abs(g-u);L<E?(this.renderSpaceWidth=f,this.renderSpaceCharCode=11825):(this.renderSpaceWidth=g,this.renderSpaceCharCode=183)}sameSelection(e){if(this.selectionsOnLine===null)return e===null;if(e===null||e.length!==this.selectionsOnLine.length)return!1;for(let t=0;t<this.selectionsOnLine.length;t++)if(!this.selectionsOnLine[t].equals(e[t]))return!1;return!0}equals(e){return this.useMonospaceOptimizations===e.useMonospaceOptimizations&&this.canUseHalfwidthRightwardsArrow===e.canUseHalfwidthRightwardsArrow&&this.lineContent===e.lineContent&&this.continuesWithWrappedLine===e.continuesWithWrappedLine&&this.isBasicASCII===e.isBasicASCII&&this.containsRTL===e.containsRTL&&this.fauxIndentLength===e.fauxIndentLength&&this.tabSize===e.tabSize&&this.startVisibleColumn===e.startVisibleColumn&&this.spaceWidth===e.spaceWidth&&this.renderSpaceWidth===e.renderSpaceWidth&&this.renderSpaceCharCode===e.renderSpaceCharCode&&this.stopRenderingLineAfter===e.stopRenderingLineAfter&&this.renderWhitespace===e.renderWhitespace&&this.renderControlCharacters===e.renderControlCharacters&&this.fontLigatures===e.fontLigatures&&Rs.equalsArr(this.lineDecorations,e.lineDecorations)&&this.lineTokens.equals(e.lineTokens)&&this.sameSelection(e.selectionsOnLine)&&this.textDirection===e.textDirection&&this.verticalScrollbarSize===e.verticalScrollbarSize&&this.renderNewLineWhenEmpty===e.renderNewLineWhenEmpty}}class YJ{constructor(e,t){this.partIndex=e,this.charIndex=t}}class lc{static getPartIndex(e){return(e&4294901760)>>>16}static getCharIndex(e){return(e&65535)>>>0}constructor(e,t){this.length=e,this._data=new Uint32Array(this.length),this._horizontalOffset=new Uint32Array(this.length)}setColumnInfo(e,t,i,n){const o=(t<<16|i<<0)>>>0;this._data[e-1]=o,this._horizontalOffset[e-1]=n}getHorizontalOffset(e){return this._horizontalOffset.length===0?0:this._horizontalOffset[e-1]}charOffsetToPartData(e){return this.length===0?0:e<0?this._data[0]:e>=this.length?this._data[this.length-1]:this._data[e]}getDomPosition(e){const t=this.charOffsetToPartData(e-1),i=lc.getPartIndex(t),n=lc.getCharIndex(t);return new YJ(i,n)}getColumn(e,t){return this.partDataToCharOffset(e.partIndex,t,e.charIndex)+1}partDataToCharOffset(e,t,i){if(this.length===0)return 0;const n=(e<<16|i<<0)>>>0;let o=0,r=this.length-1;for(;o+1<r;){const p=o+r>>>1,_=this._data[p];if(_===n)return p;_>n?r=p:o=p}if(o===r)return o;const a=this._data[o],l=this._data[r];if(a===n)return o;if(l===n)return r;const c=lc.getPartIndex(a),d=lc.getCharIndex(a),h=lc.getPartIndex(l);let u;c!==h?u=t:u=lc.getCharIndex(l);const g=i-d,f=u-i;return g<=f?o:r}}class KE{constructor(e,t){this._renderLineOutputBrand=void 0,this.characterMapping=e,this.containsForeignElements=t}}function nC(s,e){if(s.lineContent.length===0){if(s.lineDecorations.length>0){e.appendString("<span>");let t=0,i=0,n=0;for(const r of s.lineDecorations)(r.type===1||r.type===2)&&(e.appendString('<span class="'),e.appendString(r.className),e.appendString('"></span>'),r.type===1&&(n|=1,t++),r.type===2&&(n|=2,i++));e.appendString("</span>");const o=new lc(1,t+i);return o.setColumnInfo(1,t,0,0),new KE(o,n)}return s.renderNewLineWhenEmpty?e.appendString(`<span><span>
</span></span>`):e.appendString("<span><span></span></span>"),new KE(new lc(0,0),0)}return J_e(j_e(s),e)}class U_e{constructor(e,t,i){this.characterMapping=e,this.html=t,this.containsForeignElements=i}}function lA(s){const e=new Eg(1e4),t=nC(s,e);return new U_e(t.characterMapping,e.build(),t.containsForeignElements)}class q_e{constructor(e,t,i,n,o,r,a,l,c,d,h,u,g,f,p){this.fontIsMonospace=e,this.canUseHalfwidthRightwardsArrow=t,this.lineContent=i,this.len=n,this.isOverflowing=o,this.overflowingCharCount=r,this.parts=a,this.containsForeignElements=l,this.fauxIndentLength=c,this.tabSize=d,this.startVisibleColumn=h,this.spaceWidth=u,this.renderSpaceCharCode=g,this.renderWhitespace=f,this.renderControlCharacters=p}}function j_e(s){const e=s.lineContent;let t,i,n;s.stopRenderingLineAfter!==-1&&s.stopRenderingLineAfter<e.length?(t=!0,i=e.length-s.stopRenderingLineAfter,n=s.stopRenderingLineAfter):(t=!1,i=0,n=e.length);let o=G_e(e,s.containsRTL,s.lineTokens,s.fauxIndentLength,n);s.renderControlCharacters&&!s.isBasicASCII&&(o=Z_e(e,o)),(s.renderWhitespace===4||s.renderWhitespace===1||s.renderWhitespace===2&&s.selectionsOnLine||s.renderWhitespace===3&&!s.continuesWithWrappedLine)&&(o=X_e(s,e,n,o));let r=0;if(s.lineDecorations.length>0){for(let a=0,l=s.lineDecorations.length;a<l;a++){const c=s.lineDecorations[a];c.type===3||c.type===1?r|=1:c.type===2&&(r|=2)}o=Q_e(e,n,o,s.lineDecorations)}return s.containsRTL?o=Y_e(e,o):o=K_e(e,o,!s.isBasicASCII||s.fontLigatures),new q_e(s.useMonospaceOptimizations,s.canUseHalfwidthRightwardsArrow,e,n,t,i,o,r,s.fauxIndentLength,s.tabSize,s.startVisibleColumn,s.spaceWidth,s.renderSpaceCharCode,s.renderWhitespace,s.renderControlCharacters)}function G_e(s,e,t,i,n){const o=[];let r=0;i>0&&(o[r++]=new on(i,"",0,!1));let a=i;for(let l=0,c=t.getCount();l<c;l++){const d=t.getEndOffset(l);if(d<=i)continue;const h=t.getClassName(l);if(d>=n){const g=e?iv(s.substring(a,n)):!1;o[r++]=new on(n,h,0,g);break}const u=e?iv(s.substring(a,d)):!1;o[r++]=new on(d,h,0,u),a=d}return o}function K_e(s,e,t){let i=0;const n=[];let o=0;if(t)for(let r=0,a=e.length;r<a;r++){const l=e[r],c=l.endIndex;if(i+50<c){const d=l.type,h=l.metadata,u=l.containsRTL;let g=-1,f=i;for(let p=i;p<c;p++)s.charCodeAt(p)===32&&(g=p),g!==-1&&p-f>=50&&(n[o++]=new on(g+1,d,h,u),f=g+1,g=-1);f!==c&&(n[o++]=new on(c,d,h,u))}else n[o++]=l;i=c}else for(let r=0,a=e.length;r<a;r++){const l=e[r],c=l.endIndex,d=c-i;if(d>50){const h=l.type,u=l.metadata,g=l.containsRTL,f=Math.ceil(d/50);for(let p=1;p<f;p++){const _=i+p*50;n[o++]=new on(_,h,u,g)}n[o++]=new on(c,h,u,g)}else n[o++]=l;i=c}return n}function Y_e(s,e){if(e.length===0)return e;const t=e[0];if(!t.containsRTL)return e;const i=t.endIndex;let n=0;for(let r=0;r<i;r++){const a=s.charCodeAt(r);if(a!==32&&a!==9){n=r;break}}if(n===0)return e;const o=[];o.push(new on(n,t.type,t.metadata,!1)),o.push(new on(i,t.type,t.metadata,t.containsRTL));for(let r=1;r<e.length;r++)o.push(e[r]);return o}function ZJ(s){return s<32?s!==9:s===127||s>=8234&&s<=8238||s>=8294&&s<=8297||s>=8206&&s<=8207||s===1564}function Z_e(s,e){const t=[];let i=new on(0,"",0,!1),n=0;for(const o of e){const r=o.endIndex;for(;n<r;n++){const a=s.charCodeAt(n);ZJ(a)&&(n>i.endIndex&&(i=new on(n,o.type,o.metadata,o.containsRTL),t.push(i)),i=new on(n+1,"mtkcontrol",o.metadata,!1),t.push(i))}n>i.endIndex&&(i=new on(r,o.type,o.metadata,o.containsRTL),t.push(i))}return t}function X_e(s,e,t,i){const n=s.continuesWithWrappedLine,o=s.fauxIndentLength,r=s.tabSize,a=s.startVisibleColumn,l=s.useMonospaceOptimizations,c=s.selectionsOnLine,d=s.renderWhitespace===1,h=s.renderWhitespace===3,u=s.renderSpaceWidth!==s.spaceWidth,g=[];let f=0,p=0,_=i[p].type,b=i[p].containsRTL,w=i[p].endIndex;const C=i.length;let v=!1,y=As(e),x;y===-1?(v=!0,y=t,x=t):x=Ma(e);let L=!1,E=0,I=c&&c[E],R=a%r;for(let P=o;P<t;P++){const B=e.charCodeAt(P);I&&I.endExclusive<=P&&(E++,I=c&&c[E]);let $;if(P<y||P>x)$=!0;else if(B===9)$=!0;else if(B===32)if(d)if(L)$=!0;else{const z=P+1<t?e.charCodeAt(P+1):0;$=z===32||z===9}else $=!0;else $=!1;if($&&c&&($=!!I&&I.start<=P&&P<I.endExclusive),$&&h&&($=v||P>x),$&&b&&P>=y&&P<=x&&($=!1),L){if(!$||!l&&R>=r){if(u){const z=f>0?g[f-1].endIndex:o;for(let Q=z+1;Q<=P;Q++)g[f++]=new on(Q,"mtkw",1,!1)}else g[f++]=new on(P,"mtkw",1,!1);R=R%r}}else(P===w||$&&P>o)&&(g[f++]=new on(P,_,0,b),R=R%r);for(B===9?R=r:gh(B)?R+=2:R++,L=$;P===w&&(p++,p<C);)_=i[p].type,b=i[p].containsRTL,w=i[p].endIndex}let O=!1;if(L)if(n&&d){const P=t>0?e.charCodeAt(t-1):0,B=t>1?e.charCodeAt(t-2):0;P===32&&B!==32&&B!==9||(O=!0)}else O=!0;if(O)if(u){const P=f>0?g[f-1].endIndex:o;for(let B=P+1;B<=t;B++)g[f++]=new on(B,"mtkw",1,!1)}else g[f++]=new on(t,"mtkw",1,!1);else g[f++]=new on(t,_,0,b);return g}function Q_e(s,e,t,i){i.sort(Rs.compare);const n=$_e.normalize(s,i),o=n.length;let r=0;const a=[];let l=0,c=0;for(let h=0,u=t.length;h<u;h++){const g=t[h],f=g.endIndex,p=g.type,_=g.metadata,b=g.containsRTL;for(;r<o&&n[r].startOffset<f;){const w=n[r];if(w.startOffset>c&&(c=w.startOffset,a[l++]=new on(c,p,_,b)),w.endOffset+1<=f)c=w.endOffset+1,a[l++]=new on(c,p+" "+w.className,_|w.metadata,b),r++;else{c=f,a[l++]=new on(c,p+" "+w.className,_|w.metadata,b);break}}f>c&&(c=f,a[l++]=new on(c,p,_,b))}const d=t[t.length-1].endIndex;if(r<o&&n[r].startOffset===d)for(;r<o&&n[r].startOffset===d;){const h=n[r];a[l++]=new on(c,h.className,h.metadata,!1),r++}return a}function J_e(s,e){const t=s.fontIsMonospace,i=s.canUseHalfwidthRightwardsArrow,n=s.containsForeignElements,o=s.lineContent,r=s.len,a=s.isOverflowing,l=s.overflowingCharCount,c=s.parts,d=s.fauxIndentLength,h=s.tabSize,u=s.startVisibleColumn,g=s.spaceWidth,f=s.renderSpaceCharCode,p=s.renderWhitespace,_=s.renderControlCharacters,b=new lc(r+1,c.length);let w=!1,C=0,v=u,y=0,x=0,L=0;e.appendString("<span>");for(let E=0,I=c.length;E<I;E++){const R=c[E],O=R.endIndex,P=R.type,B=R.containsRTL,$=p!==0&&R.isWhitespace(),z=$&&!t&&(P==="mtkw"||!n),Q=C===O&&R.isPseudoAfter();if(y=0,e.appendString("<span "),B&&e.appendString('dir="rtl" style="unicode-bidi:isolate" '),e.appendString('class="'),e.appendString(z?"mtkz":P),e.appendASCIICharCode(34),$){let J=0;{let ne=C,de=v;for(;ne<O;ne++){const Me=(o.charCodeAt(ne)===9?h-de%h:1)|0;J+=Me,ne>=d&&(de+=Me)}}for(z&&(e.appendString(' style="width:'),e.appendString(String(g*J)),e.appendString('px"')),e.appendASCIICharCode(62);C<O;C++){b.setColumnInfo(C+1,E-L,y,x),L=0;const ne=o.charCodeAt(C);let de,ke;if(ne===9){de=h-v%h|0,ke=de,!i||ke>1?e.appendCharCode(8594):e.appendCharCode(65515);for(let Me=2;Me<=ke;Me++)e.appendCharCode(160)}else de=2,ke=1,e.appendCharCode(f),e.appendCharCode(8204);y+=de,x+=ke,C>=d&&(v+=ke)}}else for(e.appendASCIICharCode(62);C<O;C++){b.setColumnInfo(C+1,E-L,y,x),L=0;const J=o.charCodeAt(C);let ne=1,de=1;switch(J){case 9:ne=h-v%h,de=ne;for(let ke=1;ke<=ne;ke++)e.appendCharCode(160);break;case 32:e.appendCharCode(160);break;case 60:e.appendString("&lt;");break;case 62:e.appendString("&gt;");break;case 38:e.appendString("&amp;");break;case 0:_?e.appendCharCode(9216):e.appendString("&#00;");break;case 65279:case 8232:case 8233:case 133:e.appendCharCode(65533);break;default:gh(J)&&de++,_&&J<32?e.appendCharCode(9216+J):_&&J===127?e.appendCharCode(9249):_&&ZJ(J)?(e.appendString("[U+"),e.appendString(ebe(J)),e.appendString("]"),ne=8,de=ne):e.appendCharCode(J)}y+=ne,x+=de,C>=d&&(v+=de)}Q?L++:L=0,C>=r&&!w&&R.isPseudoAfter()&&(w=!0,b.setColumnInfo(C+1,E,y,x)),e.appendString("</span>")}return w||b.setColumnInfo(r+1,c.length-1,y,x),a&&(e.appendString('<span class="mtkoverflow">'),e.appendString(m(796,"Show more ({0})",tbe(l))),e.appendString("</span>")),e.appendString("</span>"),new KE(b,n)}function ebe(s){return s.toString(16).toUpperCase().padStart(4,"0")}function tbe(s){return s<1024?m(797,"{0} chars",s):s<1024*1024?`${(s/1024).toFixed(1)} KB`:`${(s/1024/1024).toFixed(1)} MB`}var Xr;(function(s){s.DARK="dark",s.LIGHT="light",s.HIGH_CONTRAST_DARK="hcDark",s.HIGH_CONTRAST_LIGHT="hcLight"})(Xr||(Xr={}));var kb;(function(s){s.VS="vs",s.VS_DARK="vs-dark",s.HC_BLACK="hc-black",s.HC_LIGHT="hc-light"})(kb||(kb={}));function Cl(s){return s===Xr.HIGH_CONTRAST_DARK||s===Xr.HIGH_CONTRAST_LIGHT}function wh(s){return s===Xr.DARK||s===Xr.HIGH_CONTRAST_DARK}const ibe=(function(){return Sl?!0:!(vo||wo||wm)})();let fw=!0;const jT=class jT{constructor(e,t){this._viewGpuContext=e,this._options=t,this._isMaybeInvalid=!0,this._renderedViewLine=null}getDomNode(){return this._renderedViewLine&&this._renderedViewLine.domNode?this._renderedViewLine.domNode.domNode:null}setDomNode(e){if(this._renderedViewLine)this._renderedViewLine.domNode=wt(e);else throw new Error("I have no rendered view line to set the dom node to...")}onContentChanged(){this._isMaybeInvalid=!0}onTokensChanged(){this._isMaybeInvalid=!0}onDecorationsChanged(){this._isMaybeInvalid=!0}onOptionsChanged(e){this._isMaybeInvalid=!0,this._options=e}onSelectionChanged(){return Cl(this._options.themeType)||this._renderedViewLine?.input.renderWhitespace===2?(this._isMaybeInvalid=!0,!0):!1}renderLine(e,t,i,n,o){if(this._options.useGpu&&this._viewGpuContext?.canRender(this._options,n,e))return this._renderedViewLine?.domNode?.domNode.remove(),this._renderedViewLine=null,!1;if(this._isMaybeInvalid===!1)return!1;this._isMaybeInvalid=!1;const r=n.getViewLineRenderingData(e),a=this._options,l=Rs.filter(r.inlineDecorations,e,r.minColumn,r.maxColumn),c=r.hasVariableFonts||a.experimentalWhitespaceRendering==="off"?a.renderWhitespace:"none",d=!r.hasVariableFonts;let h=null;if(Cl(a.themeType)||c==="selection"){const p=n.selections;for(const _ of p){if(_.endLineNumber<e||_.startLineNumber>e)continue;const b=_.startLineNumber===e?_.startColumn:r.minColumn,w=_.endLineNumber===e?_.endColumn:r.maxColumn;b<w&&(Cl(a.themeType)&&l.push(new Rs(b,w,"inline-selected-text",0)),c==="selection"&&(h||(h=[]),h.push(new me(b-1,w-1))))}}const u=new Dh(a.useMonospaceOptimizations,a.canUseHalfwidthRightwardsArrow,r.content,r.continuesWithWrappedLine,r.isBasicASCII,r.containsRTL,r.minColumn-1,r.tokens,l,r.tabSize,r.startVisibleColumn,a.spaceWidth,a.middotWidth,a.wsmiddotWidth,a.stopRenderingLineAfter,c,a.renderControlCharacters,a.fontLigatures!==mh.OFF,h,r.textDirection,a.verticalScrollbarSize);if(this._renderedViewLine&&this._renderedViewLine.input.equals(u))return!1;o.appendString("<div "),r.textDirection===br.RTL?o.appendString('dir="rtl" '):r.containsRTL&&o.appendString('dir="ltr" '),o.appendString('style="top:'),o.appendString(String(t)),o.appendString("px;height:"),o.appendString(String(i)),o.appendString("px;line-height:"),o.appendString(String(i)),r.textDirection===br.RTL&&(o.appendString("px;padding-right:"),o.appendString(String(a.verticalScrollbarSize))),o.appendString('px;" class="'),o.appendString(jT.CLASS_NAME),o.appendString('">');const g=nC(u,o);o.appendString("</div>");let f=null;return d&&fw&&ibe&&r.isBasicASCII&&u.isLTR&&a.useMonospaceOptimizations&&g.containsForeignElements===0&&(f=new Qx(this._renderedViewLine?this._renderedViewLine.domNode:null,u,g.characterMapping)),f||(f=QJ(this._renderedViewLine?this._renderedViewLine.domNode:null,u,g.characterMapping,g.containsForeignElements)),this._renderedViewLine=f,!0}layoutLine(e,t,i){this._renderedViewLine&&this._renderedViewLine.domNode&&(this._renderedViewLine.domNode.setTop(t),this._renderedViewLine.domNode.setHeight(i),this._renderedViewLine.domNode.setLineHeight(i))}isRenderedRTL(){return this._renderedViewLine?this._renderedViewLine.input.textDirection===br.RTL:!1}getWidth(e){return this._renderedViewLine?this._renderedViewLine.getWidth(e):0}getWidthIsFast(){return this._renderedViewLine?this._renderedViewLine.getWidthIsFast():!0}needsMonospaceFontCheck(){return this._renderedViewLine?this._renderedViewLine instanceof Qx:!1}monospaceAssumptionsAreValid(){return this._renderedViewLine&&this._renderedViewLine instanceof Qx?this._renderedViewLine.monospaceAssumptionsAreValid():fw}onMonospaceAssumptionsInvalidated(){this._renderedViewLine&&this._renderedViewLine instanceof Qx&&(this._renderedViewLine=this._renderedViewLine.toSlowRenderedLine())}getVisibleRangesForRange(e,t,i,n){if(!this._renderedViewLine)return null;t=Math.min(this._renderedViewLine.input.lineContent.length+1,Math.max(1,t)),i=Math.min(this._renderedViewLine.input.lineContent.length+1,Math.max(1,i));const o=this._renderedViewLine.input.stopRenderingLineAfter;if(o!==-1&&t>o+1&&i>o+1)return new eF(!0,[new Ou(this.getWidth(n),0)]);o!==-1&&t>o+1&&(t=o+1),o!==-1&&i>o+1&&(i=o+1);const r=this._renderedViewLine.getVisibleRangesForRange(e,t,i,n);return r&&r.length>0?new eF(!1,r):null}getColumnOfNodeOffset(e,t){return this._renderedViewLine?this._renderedViewLine.getColumnOfNodeOffset(e,t):1}};jT.CLASS_NAME="view-line";let Od=jT;class Qx{constructor(e,t,i){this._cachedWidth=-1,this.domNode=e,this.input=t;const n=Math.floor(t.lineContent.length/300);if(n>0){this._keyColumnPixelOffsetCache=new Float32Array(n);for(let o=0;o<n;o++)this._keyColumnPixelOffsetCache[o]=-1}else this._keyColumnPixelOffsetCache=null;this._characterMapping=i,this._charWidth=t.spaceWidth}getWidth(e){if(!this.domNode||this.input.lineContent.length<300){const t=this._characterMapping.getHorizontalOffset(this._characterMapping.length);return Math.round(this._charWidth*t)}return this._cachedWidth===-1&&(this._cachedWidth=this._getReadingTarget(this.domNode).offsetWidth,e?.markDidDomLayout()),this._cachedWidth}getWidthIsFast(){return this.input.lineContent.length<300||this._cachedWidth!==-1}monospaceAssumptionsAreValid(){if(!this.domNode)return fw;if(this.input.lineContent.length<300){const e=this.getWidth(null),t=this.domNode.domNode.firstChild.offsetWidth;Math.abs(e-t)>=2&&(console.warn("monospace assumptions have been violated, therefore disabling monospace optimizations!"),fw=!1)}return fw}toSlowRenderedLine(){return QJ(this.domNode,this.input,this._characterMapping,0)}getVisibleRangesForRange(e,t,i,n){const o=this._getColumnPixelOffset(e,t,n),r=this._getColumnPixelOffset(e,i,n);return[new Ou(o,r-o)]}_getColumnPixelOffset(e,t,i){if(t<=300){const c=this._characterMapping.getHorizontalOffset(t);return this._charWidth*c}const n=Math.floor((t-1)/300)-1,o=(n+1)*300+1;let r=-1;if(this._keyColumnPixelOffsetCache&&(r=this._keyColumnPixelOffsetCache[n],r===-1&&(r=this._actualReadPixelOffset(e,o,i),this._keyColumnPixelOffsetCache[n]=r)),r===-1){const c=this._characterMapping.getHorizontalOffset(t);return this._charWidth*c}const a=this._characterMapping.getHorizontalOffset(o),l=this._characterMapping.getHorizontalOffset(t);return r+this._charWidth*(l-a)}_getReadingTarget(e){return e.domNode.firstChild}_actualReadPixelOffset(e,t,i){if(!this.domNode)return-1;const n=this._characterMapping.getDomPosition(t),o=Nk.readHorizontalRanges(this._getReadingTarget(this.domNode),n.partIndex,n.charIndex,n.partIndex,n.charIndex,i);return!o||o.length===0?-1:o[0].left}getColumnOfNodeOffset(e,t){return dv(this._characterMapping,e,t)}}class XJ{constructor(e,t,i,n){if(this.domNode=e,this.input=t,this._characterMapping=i,this._isWhitespaceOnly=/^\s*$/.test(t.lineContent),this._containsForeignElements=n,this._cachedWidth=-1,this._pixelOffsetCache=null,t.isLTR){this._pixelOffsetCache=new Float32Array(Math.max(2,this._characterMapping.length+1));for(let o=0,r=this._characterMapping.length;o<=r;o++)this._pixelOffsetCache[o]=-1}}_getReadingTarget(e){return e.domNode.firstChild}getWidth(e){return this.domNode?(this._cachedWidth===-1&&(this._cachedWidth=this._getReadingTarget(this.domNode).offsetWidth,e?.markDidDomLayout()),this._cachedWidth):0}getWidthIsFast(){return this._cachedWidth!==-1}getVisibleRangesForRange(e,t,i,n){if(!this.domNode)return null;if(this._pixelOffsetCache!==null){const o=this._readPixelOffset(this.domNode,e,t,n);if(o===-1)return null;const r=this._readPixelOffset(this.domNode,e,i,n);return r===-1?null:[new Ou(o,r-o)]}return this._readVisibleRangesForRange(this.domNode,e,t,i,n)}_readVisibleRangesForRange(e,t,i,n,o){if(i===n){const r=this._readPixelOffset(e,t,i,o);return r===-1?null:[new Ou(r,0)]}else return this._readRawVisibleRangesForRange(e,i,n,o)}_readPixelOffset(e,t,i,n){if(this.input.isLTR&&this._characterMapping.length===0){if(this._containsForeignElements===0||this._containsForeignElements===2)return 0;if(this._containsForeignElements===1)return this.getWidth(n);const o=this._getReadingTarget(e);return o.firstChild?(n.markDidDomLayout(),o.firstChild.offsetWidth):0}if(this._pixelOffsetCache!==null){const o=this._pixelOffsetCache[i];if(o!==-1)return o;const r=this._actualReadPixelOffset(e,t,i,n);return this._pixelOffsetCache[i]=r,r}return this._actualReadPixelOffset(e,t,i,n)}_actualReadPixelOffset(e,t,i,n){if(this._characterMapping.length===0){const l=Nk.readHorizontalRanges(this._getReadingTarget(e),0,0,0,0,n);return!l||l.length===0?-1:l[0].left}if(this.input.isLTR&&i===this._characterMapping.length&&this._isWhitespaceOnly&&this._containsForeignElements===0)return this.getWidth(n);const o=this._characterMapping.getDomPosition(i),r=Nk.readHorizontalRanges(this._getReadingTarget(e),o.partIndex,o.charIndex,o.partIndex,o.charIndex,n);if(!r||r.length===0)return-1;const a=r[0].left;if(this.input.isBasicASCII){const l=this._characterMapping.getHorizontalOffset(i),c=Math.round(this.input.spaceWidth*l);if(Math.abs(c-a)<=1)return c}return a}_readRawVisibleRangesForRange(e,t,i,n){if(this.input.isLTR&&t===1&&i===this._characterMapping.length)return[new Ou(0,this.getWidth(n))];const o=this._characterMapping.getDomPosition(t),r=this._characterMapping.getDomPosition(i);return Nk.readHorizontalRanges(this._getReadingTarget(e),o.partIndex,o.charIndex,r.partIndex,r.charIndex,n)}getColumnOfNodeOffset(e,t){return dv(this._characterMapping,e,t)}}class nbe extends XJ{_readVisibleRangesForRange(e,t,i,n,o){const r=super._readVisibleRangesForRange(e,t,i,n,o);if(!r||r.length===0||i===n||i===1&&n===this._characterMapping.length)return r;if(this.input.isLTR){const a=this._readPixelOffset(e,t,n,o);if(a!==-1){const l=r[r.length-1];l.left<a&&(l.width=a-l.left)}}return r}}const QJ=(function(){return xR?sbe:obe})();function sbe(s,e,t,i){return new nbe(s,e,t,i)}function obe(s,e,t,i){return new XJ(s,e,t,i)}function dv(s,e,t){const i=e.textContent.length;let n=-1;for(;e;)e=e.previousSibling,n++;return s.getColumn(new YJ(n,t),i)}class $h{constructor(e=null){this.hitTarget=e,this.type=0}}class JJ{get hitTarget(){return this.spanNode}constructor(e,t,i){this.position=e,this.spanNode=t,this.injectedText=i,this.type=1}}var pf;(function(s){function e(t,i,n){const o=t.getPositionFromDOMInfo(i,n);return o?new JJ(o,i,null):new $h(i)}s.createFromDOMInfo=e})(pf||(pf={}));class rbe{constructor(e,t){this.lastViewCursorsRenderData=e,this.lastTextareaPosition=t}}class Zn{static _deduceRage(e,t=null){return!t&&e?new k(e.lineNumber,e.column,e.lineNumber,e.column):t??null}static createUnknown(e,t,i){return{type:0,element:e,mouseColumn:t,position:i,range:this._deduceRage(i)}}static createTextarea(e,t){return{type:1,element:e,mouseColumn:t,position:null,range:null}}static createMargin(e,t,i,n,o,r){return{type:e,element:t,mouseColumn:i,position:n,range:o,detail:r}}static createViewZone(e,t,i,n,o){return{type:e,element:t,mouseColumn:i,position:n,range:this._deduceRage(n),detail:o}}static createContentText(e,t,i,n,o){return{type:6,element:e,mouseColumn:t,position:i,range:this._deduceRage(i,n),detail:o}}static createContentEmpty(e,t,i,n){return{type:7,element:e,mouseColumn:t,position:i,range:this._deduceRage(i),detail:n}}static createContentWidget(e,t,i){return{type:9,element:e,mouseColumn:t,position:null,range:null,detail:i}}static createScrollbar(e,t,i){return{type:11,element:e,mouseColumn:t,position:i,range:this._deduceRage(i)}}static createOverlayWidget(e,t,i){return{type:12,element:e,mouseColumn:t,position:null,range:null,detail:i}}static createOutsideEditor(e,t,i,n){return{type:13,element:null,mouseColumn:e,position:t,range:this._deduceRage(t),outsidePosition:i,outsideDistance:n}}static _typeToString(e){return e===1?"TEXTAREA":e===2?"GUTTER_GLYPH_MARGIN":e===3?"GUTTER_LINE_NUMBERS":e===4?"GUTTER_LINE_DECORATIONS":e===5?"GUTTER_VIEW_ZONE":e===6?"CONTENT_TEXT":e===7?"CONTENT_EMPTY":e===8?"CONTENT_VIEW_ZONE":e===9?"CONTENT_WIDGET":e===10?"OVERVIEW_RULER":e===11?"SCROLLBAR":e===12?"OVERLAY_WIDGET":"UNKNOWN"}static toString(e){return this._typeToString(e.type)+": "+e.position+" - "+e.range+" - "+JSON.stringify(e.detail)}}class Vs{static isTextArea(e){return e.length===2&&e[0]===3&&e[1]===7}static isChildOfViewLines(e){return e.length>=4&&e[0]===3&&e[3]===8}static isStrictChildOfViewLines(e){return e.length>4&&e[0]===3&&e[3]===8}static isChildOfScrollableElement(e){return e.length>=2&&e[0]===3&&e[1]===6}static isChildOfMinimap(e){return e.length>=2&&e[0]===3&&e[1]===9}static isChildOfContentWidgets(e){return e.length>=4&&e[0]===3&&e[3]===1}static isChildOfOverflowGuard(e){return e.length>=1&&e[0]===3}static isChildOfOverflowingContentWidgets(e){return e.length>=1&&e[0]===2}static isChildOfOverlayWidgets(e){return e.length>=2&&e[0]===3&&e[1]===4}static isChildOfOverflowingOverlayWidgets(e){return e.length>=1&&e[0]===5}}class hv{constructor(e,t,i){this.viewModel=e.viewModel;const n=e.configuration.options;this.layoutInfo=n.get(165),this.viewDomNode=t.viewDomNode,this.viewLinesGpu=t.viewLinesGpu,this.lineHeight=n.get(75),this.stickyTabStops=n.get(132),this.typicalHalfwidthCharacterWidth=n.get(59).typicalHalfwidthCharacterWidth,this.lastRenderData=i,this._context=e,this._viewHelper=t}getZoneAtCoord(e){return hv.getZoneAtCoord(this._context,e)}static getZoneAtCoord(e,t){const i=e.viewLayout.getWhitespaceAtVerticalOffset(t);if(i){const n=i.verticalOffset+i.height/2,o=e.viewModel.getLineCount();let r=null,a,l=null;return i.afterLineNumber!==o&&(l=new A(i.afterLineNumber+1,1)),i.afterLineNumber>0&&(r=new A(i.afterLineNumber,e.viewModel.getLineMaxColumn(i.afterLineNumber))),l===null?a=r:r===null?a=l:t<n?a=r:a=l,{viewZoneId:i.id,afterLineNumber:i.afterLineNumber,positionBefore:r,positionAfter:l,position:a}}return null}getFullLineRangeAtCoord(e){if(this._context.viewLayout.isAfterLines(e)){const n=this._context.viewModel.getLineCount(),o=this._context.viewModel.getLineMaxColumn(n);return{range:new k(n,o,n,o),isAfterLines:!0}}const t=this._context.viewLayout.getLineNumberAtVerticalOffset(e),i=this._context.viewModel.getLineMaxColumn(t);return{range:new k(t,1,t,i),isAfterLines:!1}}getLineNumberAtVerticalOffset(e){return this._context.viewLayout.getLineNumberAtVerticalOffset(e)}isAfterLines(e){return this._context.viewLayout.isAfterLines(e)}isInTopPadding(e){return this._context.viewLayout.isInTopPadding(e)}isInBottomPadding(e){return this._context.viewLayout.isInBottomPadding(e)}getVerticalOffsetForLineNumber(e){return this._context.viewLayout.getVerticalOffsetForLineNumber(e)}findAttribute(e,t){return hv._findAttribute(e,t,this._viewHelper.viewDomNode)}static _findAttribute(e,t,i){for(;e&&e!==e.ownerDocument.body;){if(e.hasAttribute&&e.hasAttribute(t))return e.getAttribute(t);if(e===i)return null;e=e.parentNode}return null}getLineWidth(e){return this._viewHelper.getLineWidth(e)}isRtl(e){return this.viewModel.getTextDirection(e)===br.RTL}visibleRangeForPosition(e,t){return this._viewHelper.visibleRangeForPosition(e,t)}getPositionFromDOMInfo(e,t){return this._viewHelper.getPositionFromDOMInfo(e,t)}getCurrentScrollTop(){return this._context.viewLayout.getCurrentScrollTop()}getCurrentScrollLeft(){return this._context.viewLayout.getCurrentScrollLeft()}}class abe{constructor(e,t,i,n){this.editorPos=t,this.pos=i,this.relativePos=n,this.mouseVerticalOffset=Math.max(0,e.getCurrentScrollTop()+this.relativePos.y),this.mouseContentHorizontalOffset=e.getCurrentScrollLeft()+this.relativePos.x-e.layoutInfo.contentLeft,this.isInMarginArea=this.relativePos.x<e.layoutInfo.contentLeft&&this.relativePos.x>=e.layoutInfo.glyphMarginLeft,this.isInContentArea=!this.isInMarginArea,this.mouseColumn=Math.max(0,Ls._getMouseColumn(this.mouseContentHorizontalOffset,e.typicalHalfwidthCharacterWidth))}}class lbe extends abe{get target(){return this._useHitTestTarget?this.hitTestResult.value.hitTarget:this._eventTarget}get targetPath(){return this._targetPathCacheElement!==this.target&&(this._targetPathCacheElement=this.target,this._targetPathCacheValue=kl.collect(this.target,this._ctx.viewDomNode)),this._targetPathCacheValue}constructor(e,t,i,n,o){super(e,t,i,n),this.hitTestResult=new es(()=>Ls.doHitTest(this._ctx,this)),this._targetPathCacheElement=null,this._targetPathCacheValue=new Uint8Array(0),this._ctx=e,this._eventTarget=o;const r=!!this._eventTarget;this._useHitTestTarget=!r}toString(){return`pos(${this.pos.x},${this.pos.y}), editorPos(${this.editorPos.x},${this.editorPos.y}), relativePos(${this.relativePos.x},${this.relativePos.y}), mouseVerticalOffset: ${this.mouseVerticalOffset}, mouseContentHorizontalOffset: ${this.mouseContentHorizontalOffset}
	target: ${this.target?this.target.outerHTML:null}`}get wouldBenefitFromHitTestTargetSwitch(){return!this._useHitTestTarget&&this.hitTestResult.value.hitTarget!==null&&this.target!==this.hitTestResult.value.hitTarget}switchToHitTestTarget(){this._useHitTestTarget=!0}_getMouseColumn(e=null){return e&&e.column<this._ctx.viewModel.getLineMaxColumn(e.lineNumber)?pi.visibleColumnFromColumn(this._ctx.viewModel.getLineContent(e.lineNumber),e.column,this._ctx.viewModel.model.getOptions().tabSize)+1:this.mouseColumn}fulfillUnknown(e=null){return Zn.createUnknown(this.target,this._getMouseColumn(e),e)}fulfillTextarea(){return Zn.createTextarea(this.target,this._getMouseColumn())}fulfillMargin(e,t,i,n){return Zn.createMargin(e,this.target,this._getMouseColumn(t),t,i,n)}fulfillViewZone(e,t,i){return Zn.createViewZone(e,this.target,this._getMouseColumn(),t,i)}fulfillContentText(e,t,i){return Zn.createContentText(this.target,this._getMouseColumn(e),e,t,i)}fulfillContentEmpty(e,t){return Zn.createContentEmpty(this.target,this._getMouseColumn(e),e,t)}fulfillContentWidget(e){return Zn.createContentWidget(this.target,this._getMouseColumn(),e)}fulfillScrollbar(e){return Zn.createScrollbar(this.target,this._getMouseColumn(e),e)}fulfillOverlayWidget(e){return Zn.createOverlayWidget(this.target,this._getMouseColumn(),e)}}const H$={isAfterLines:!0};function Wg(s){return{isAfterLines:!1,horizontalDistanceToText:s}}class Ls{constructor(e,t){this._context=e,this._viewHelper=t}mouseTargetIsWidget(e){const t=e.target,i=kl.collect(t,this._viewHelper.viewDomNode);return!!(Vs.isChildOfContentWidgets(i)||Vs.isChildOfOverflowingContentWidgets(i)||Vs.isChildOfOverlayWidgets(i)||Vs.isChildOfOverflowingOverlayWidgets(i))}createMouseTarget(e,t,i,n,o){const r=new hv(this._context,this._viewHelper,e),a=new lbe(r,t,i,n,o);try{const l=Ls._createMouseTarget(r,a);if(l.type===6&&r.stickyTabStops&&l.position!==null){const c=Ls._snapToSoftTabBoundary(l.position,r.viewModel),d=k.fromPositions(c,c).plusRange(l.range);return a.fulfillContentText(c,d,l.detail)}return l}catch{return a.fulfillUnknown()}}static _createMouseTarget(e,t){if(t.target===null)return t.fulfillUnknown();const i=t;let n=null;return!Vs.isChildOfOverflowGuard(t.targetPath)&&!Vs.isChildOfOverflowingContentWidgets(t.targetPath)&&!Vs.isChildOfOverflowingOverlayWidgets(t.targetPath)&&(n=n||t.fulfillUnknown()),n=n||Ls._hitTestContentWidget(e,i),n=n||Ls._hitTestOverlayWidget(e,i),n=n||Ls._hitTestMinimap(e,i),n=n||Ls._hitTestScrollbarSlider(e,i),n=n||Ls._hitTestViewZone(e,i),n=n||Ls._hitTestMargin(e,i),n=n||Ls._hitTestViewCursor(e,i),n=n||Ls._hitTestTextArea(e,i),n=n||Ls._hitTestViewLines(e,i),n=n||Ls._hitTestScrollbar(e,i),n||t.fulfillUnknown()}static _hitTestContentWidget(e,t){if(Vs.isChildOfContentWidgets(t.targetPath)||Vs.isChildOfOverflowingContentWidgets(t.targetPath)){const i=e.findAttribute(t.target,"widgetId");return i?t.fulfillContentWidget(i):t.fulfillUnknown()}return null}static _hitTestOverlayWidget(e,t){if(Vs.isChildOfOverlayWidgets(t.targetPath)||Vs.isChildOfOverflowingOverlayWidgets(t.targetPath)){const i=e.findAttribute(t.target,"widgetId");return i?t.fulfillOverlayWidget(i):t.fulfillUnknown()}return null}static _hitTestViewCursor(e,t){if(t.target){const i=e.lastRenderData.lastViewCursorsRenderData;for(const n of i)if(t.target===n.domNode)return t.fulfillContentText(n.position,null,{mightBeForeignElement:!1,injectedText:null})}if(t.isInContentArea){const i=e.lastRenderData.lastViewCursorsRenderData,n=t.mouseContentHorizontalOffset,o=t.mouseVerticalOffset;for(const r of i){if(n<r.contentLeft||n>r.contentLeft+r.width)continue;const a=e.getVerticalOffsetForLineNumber(r.position.lineNumber);if(a<=o&&o<=a+r.height)return t.fulfillContentText(r.position,null,{mightBeForeignElement:!1,injectedText:null})}}return null}static _hitTestViewZone(e,t){const i=e.getZoneAtCoord(t.mouseVerticalOffset);if(i){const n=t.isInContentArea?8:5;return t.fulfillViewZone(n,i.position,i)}return null}static _hitTestTextArea(e,t){return Vs.isTextArea(t.targetPath)?e.lastRenderData.lastTextareaPosition?t.fulfillContentText(e.lastRenderData.lastTextareaPosition,null,{mightBeForeignElement:!1,injectedText:null}):t.fulfillTextarea():null}static _hitTestMargin(e,t){if(t.isInMarginArea){const i=e.getFullLineRangeAtCoord(t.mouseVerticalOffset),n=i.range.getStartPosition();let o=Math.abs(t.relativePos.x);const r={isAfterLines:i.isAfterLines,glyphMarginLeft:e.layoutInfo.glyphMarginLeft,glyphMarginWidth:e.layoutInfo.glyphMarginWidth,lineNumbersWidth:e.layoutInfo.lineNumbersWidth,offsetX:o};if(o-=e.layoutInfo.glyphMarginLeft,o<=e.layoutInfo.glyphMarginWidth){const a=e.viewModel.coordinatesConverter.convertViewPositionToModelPosition(i.range.getStartPosition()),l=e.viewModel.glyphLanes.getLanesAtLine(a.lineNumber);return r.glyphMarginLane=l[Math.floor(o/e.lineHeight)],t.fulfillMargin(2,n,i.range,r)}return o-=e.layoutInfo.glyphMarginWidth,o<=e.layoutInfo.lineNumbersWidth?t.fulfillMargin(3,n,i.range,r):(o-=e.layoutInfo.lineNumbersWidth,t.fulfillMargin(4,n,i.range,r))}return null}static _hitTestViewLines(e,t){if(!Vs.isChildOfViewLines(t.targetPath))return null;if(e.isInTopPadding(t.mouseVerticalOffset))return t.fulfillContentEmpty(new A(1,1),H$);if(e.isAfterLines(t.mouseVerticalOffset)||e.isInBottomPadding(t.mouseVerticalOffset)){const n=e.viewModel.getLineCount(),o=e.viewModel.getLineMaxColumn(n);return t.fulfillContentEmpty(new A(n,o),H$)}if(Vs.isStrictChildOfViewLines(t.targetPath)){const n=e.getLineNumberAtVerticalOffset(t.mouseVerticalOffset),o=e.viewModel.getLineLength(n),r=e.getLineWidth(n);if(o===0){const l=Wg(t.mouseContentHorizontalOffset-r);return t.fulfillContentEmpty(new A(n,1),l)}if(e.isRtl(n)){if(t.mouseContentHorizontalOffset+r<=e.layoutInfo.contentWidth-e.layoutInfo.verticalScrollbarWidth){const l=Wg(t.mouseContentHorizontalOffset-r),c=new A(n,e.viewModel.getLineMaxColumn(n));return t.fulfillContentEmpty(c,l)}}else if(t.mouseContentHorizontalOffset>=r){const l=Wg(t.mouseContentHorizontalOffset-r),c=new A(n,e.viewModel.getLineMaxColumn(n));return t.fulfillContentEmpty(c,l)}}else if(e.viewLinesGpu){const n=e.getLineNumberAtVerticalOffset(t.mouseVerticalOffset);if(e.viewModel.getLineLength(n)===0){const l=e.getLineWidth(n),c=Wg(t.mouseContentHorizontalOffset-l);return t.fulfillContentEmpty(new A(n,1),c)}const o=e.getLineWidth(n);if(e.isRtl(n)){if(t.mouseContentHorizontalOffset+o<=e.layoutInfo.contentWidth-e.layoutInfo.verticalScrollbarWidth){const l=Wg(t.mouseContentHorizontalOffset-o),c=new A(n,e.viewModel.getLineMaxColumn(n));return t.fulfillContentEmpty(c,l)}}else if(t.mouseContentHorizontalOffset>=o){const l=Wg(t.mouseContentHorizontalOffset-o),c=new A(n,e.viewModel.getLineMaxColumn(n));return t.fulfillContentEmpty(c,l)}const a=e.viewLinesGpu.getPositionAtCoordinate(n,t.mouseContentHorizontalOffset);if(a){const l={injectedText:null,mightBeForeignElement:!1};return t.fulfillContentText(a,k.fromPositions(a,a),l)}}const i=t.hitTestResult.value;return i.type===1?Ls.createMouseTargetFromHitTestPosition(e,t,i.spanNode,i.position,i.injectedText):t.wouldBenefitFromHitTestTargetSwitch?(t.switchToHitTestTarget(),this._createMouseTarget(e,t)):t.fulfillUnknown()}static _hitTestMinimap(e,t){if(Vs.isChildOfMinimap(t.targetPath)){const i=e.getLineNumberAtVerticalOffset(t.mouseVerticalOffset),n=e.viewModel.getLineMaxColumn(i);return t.fulfillScrollbar(new A(i,n))}return null}static _hitTestScrollbarSlider(e,t){if(Vs.isChildOfScrollableElement(t.targetPath)&&t.target&&t.target.nodeType===1){const i=t.target.className;if(i&&/\b(slider|scrollbar)\b/.test(i)){const n=e.getLineNumberAtVerticalOffset(t.mouseVerticalOffset),o=e.viewModel.getLineMaxColumn(n);return t.fulfillScrollbar(new A(n,o))}}return null}static _hitTestScrollbar(e,t){if(Vs.isChildOfScrollableElement(t.targetPath)){const i=e.getLineNumberAtVerticalOffset(t.mouseVerticalOffset),n=e.viewModel.getLineMaxColumn(i);return t.fulfillScrollbar(new A(i,n))}return null}getMouseColumn(e){const t=this._context.configuration.options,i=t.get(165),n=this._context.viewLayout.getCurrentScrollLeft()+e.x-i.contentLeft;return Ls._getMouseColumn(n,t.get(59).typicalHalfwidthCharacterWidth)}static _getMouseColumn(e,t){return e<0?1:Math.round(e/t)+1}static createMouseTargetFromHitTestPosition(e,t,i,n,o){const r=n.lineNumber,a=n.column,l=e.getLineWidth(r);if(t.mouseContentHorizontalOffset>l){const b=Wg(t.mouseContentHorizontalOffset-l);return t.fulfillContentEmpty(n,b)}const c=e.visibleRangeForPosition(r,a);if(!c)return t.fulfillUnknown(n);const d=c.left;if(Math.abs(t.mouseContentHorizontalOffset-d)<1)return t.fulfillContentText(n,null,{mightBeForeignElement:!!o,injectedText:o});const h=[];if(h.push({offset:c.left,column:a}),a>1){const b=e.visibleRangeForPosition(r,a-1);b&&h.push({offset:b.left,column:a-1})}const u=e.viewModel.getLineMaxColumn(r);if(a<u){const b=e.visibleRangeForPosition(r,a+1);b&&h.push({offset:b.left,column:a+1})}h.sort((b,w)=>b.offset-w.offset);const g=t.pos.toClientCoordinates(be(e.viewDomNode)),f=i.getBoundingClientRect(),p=f.left<=g.clientX&&g.clientX<=f.right;let _=null;for(let b=1;b<h.length;b++){const w=h[b-1],C=h[b];if(w.offset<=t.mouseContentHorizontalOffset&&t.mouseContentHorizontalOffset<=C.offset){_=new k(r,w.column,r,C.column);const v=Math.abs(w.offset-t.mouseContentHorizontalOffset),y=Math.abs(C.offset-t.mouseContentHorizontalOffset);n=v<y?new A(r,w.column):new A(r,C.column);break}}return t.fulfillContentText(n,_,{mightBeForeignElement:!p||!!o,injectedText:o})}static _doHitTestWithCaretRangeFromPoint(e,t){const i=e.getLineNumberAtVerticalOffset(t.mouseVerticalOffset),n=e.getVerticalOffsetForLineNumber(i),o=n+e.lineHeight;if(!(i===e.viewModel.getLineCount()&&t.mouseVerticalOffset>o)){const a=Math.floor((n+o)/2);let l=t.pos.y+(a-t.mouseVerticalOffset);l<=t.editorPos.y&&(l=t.editorPos.y+1),l>=t.editorPos.y+t.editorPos.height&&(l=t.editorPos.y+t.editorPos.height-1);const c=new J0(t.pos.x,l),d=this._actualDoHitTestWithCaretRangeFromPoint(e,c.toClientCoordinates(be(e.viewDomNode)));if(d.type===1)return d}return this._actualDoHitTestWithCaretRangeFromPoint(e,t.pos.toClientCoordinates(be(e.viewDomNode)))}static _actualDoHitTestWithCaretRangeFromPoint(e,t){const i=cg(e.viewDomNode);let n;if(i?typeof i.caretRangeFromPoint>"u"?n=cbe(i,t.clientX,t.clientY):n=i.caretRangeFromPoint(t.clientX,t.clientY):n=e.viewDomNode.ownerDocument.caretRangeFromPoint(t.clientX,t.clientY),!n||!n.startContainer)return new $h;const o=n.startContainer;if(o.nodeType===o.TEXT_NODE){const r=o.parentNode,a=r?r.parentNode:null,l=a?a.parentNode:null;return(l&&l.nodeType===l.ELEMENT_NODE?l.className:null)===Od.CLASS_NAME?pf.createFromDOMInfo(e,r,n.startOffset):new $h(o.parentNode)}else if(o.nodeType===o.ELEMENT_NODE){const r=o.parentNode,a=r?r.parentNode:null;return(a&&a.nodeType===a.ELEMENT_NODE?a.className:null)===Od.CLASS_NAME?pf.createFromDOMInfo(e,o,o.textContent.length):new $h(o)}return new $h}static _doHitTestWithCaretPositionFromPoint(e,t){const i=e.viewDomNode.ownerDocument.caretPositionFromPoint(t.clientX,t.clientY);if(i.offsetNode.nodeType===i.offsetNode.TEXT_NODE){const n=i.offsetNode.parentNode,o=n?n.parentNode:null,r=o?o.parentNode:null;return(r&&r.nodeType===r.ELEMENT_NODE?r.className:null)===Od.CLASS_NAME?pf.createFromDOMInfo(e,i.offsetNode.parentNode,i.offset):new $h(i.offsetNode.parentNode)}if(i.offsetNode.nodeType===i.offsetNode.ELEMENT_NODE){const n=i.offsetNode.parentNode,o=n&&n.nodeType===n.ELEMENT_NODE?n.className:null,r=n?n.parentNode:null,a=r&&r.nodeType===r.ELEMENT_NODE?r.className:null;if(o===Od.CLASS_NAME){const l=i.offsetNode.childNodes[Math.min(i.offset,i.offsetNode.childNodes.length-1)];if(l)return pf.createFromDOMInfo(e,l,0)}else if(a===Od.CLASS_NAME)return pf.createFromDOMInfo(e,i.offsetNode,0)}return new $h(i.offsetNode)}static _snapToSoftTabBoundary(e,t){const i=t.getLineContent(e.lineNumber),{tabSize:n}=t.model.getOptions(),o=eS.atomicPosition(i,e.column-1,n,2);return o!==-1?new A(e.lineNumber,o+1):e}static doHitTest(e,t){let i=new $h;if(typeof e.viewDomNode.ownerDocument.caretRangeFromPoint=="function"?i=this._doHitTestWithCaretRangeFromPoint(e,t):e.viewDomNode.ownerDocument.caretPositionFromPoint&&(i=this._doHitTestWithCaretPositionFromPoint(e,t.pos.toClientCoordinates(be(e.viewDomNode)))),i.type===1){const n=e.viewModel.getInjectedTextAt(i.position),o=e.viewModel.normalizePosition(i.position,2);(n||!o.equals(i.position))&&(i=new JJ(o,i.spanNode,n))}return i}}function cbe(s,e,t){const i=document.createRange();let n=s.elementFromPoint(e,t);if(n?.hasChildNodes()){for(;n&&n.firstChild&&n.firstChild.nodeType!==n.firstChild.TEXT_NODE&&n.lastChild&&n.lastChild.firstChild;)n=n.lastChild;const o=n.getBoundingClientRect(),r=be(n),a=r.getComputedStyle(n,null).getPropertyValue("font-style"),l=r.getComputedStyle(n,null).getPropertyValue("font-variant"),c=r.getComputedStyle(n,null).getPropertyValue("font-weight"),d=r.getComputedStyle(n,null).getPropertyValue("font-size"),h=r.getComputedStyle(n,null).getPropertyValue("line-height"),u=r.getComputedStyle(n,null).getPropertyValue("font-family"),g=`${a} ${l} ${c} ${d}/${h} ${u}`,f=n.innerText;let p=o.left,_=0,b;if(e>o.left+o.width)_=f.length;else{const w=tF.getInstance();for(let C=0;C<f.length+1;C++){if(b=w.getCharWidth(f.charAt(C),g)/2,p+=b,e<p){_=C;break}p+=b}}i.setStart(n.firstChild,_),i.setEnd(n.firstChild,_)}return i}const Rf=class Rf{static getInstance(){return Rf._INSTANCE||(Rf._INSTANCE=new Rf),Rf._INSTANCE}constructor(){this._cache={},this._canvas=document.createElement("canvas")}getCharWidth(e,t){const i=e+t;if(this._cache[i])return this._cache[i];const n=this._canvas.getContext("2d");n.font=t;const r=n.measureText(e).width;return this._cache[i]=r,r}};Rf._INSTANCE=null;let tF=Rf;function Ai(s,e,t){let i=null,n=null;if(typeof t.value=="function"?(i="value",n=t.value,n.length!==0&&console.warn("Memoize should only be used in functions with zero parameters")):typeof t.get=="function"&&(i="get",n=t.get),!n)throw new Error("not supported");const o=`$memoize$${e}`;t[i]=function(...r){return this.hasOwnProperty(o)||Object.defineProperty(this,o,{configurable:!1,enumerable:!1,writable:!1,value:n.apply(this,r)}),this[o]}}var dbe=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},qt;(function(s){s.Tap="-monaco-gesturetap",s.Change="-monaco-gesturechange",s.Start="-monaco-gesturestart",s.End="-monaco-gesturesend",s.Contextmenu="-monaco-gesturecontextmenu"})(qt||(qt={}));const Kn=class Kn extends F{constructor(){super(),this.dispatched=!1,this.targets=new Ns,this.ignoreTargets=new Ns,this.activeTouches={},this.handle=null,this._lastSetTapCountTime=0,this._register(oe.runAndSubscribe(G0,({window:e,disposables:t})=>{t.add(W(e.document,"touchstart",i=>this.onTouchStart(i),{passive:!1})),t.add(W(e.document,"touchend",i=>this.onTouchEnd(e,i))),t.add(W(e.document,"touchmove",i=>this.onTouchMove(i),{passive:!1}))},{window:Nt,disposables:this._store}))}static addTarget(e){if(!Kn.isTouchDevice())return F.None;Kn.INSTANCE||(Kn.INSTANCE=new Kn);const t=Kn.INSTANCE.targets.push(e);return pe(t)}static ignoreTarget(e){if(!Kn.isTouchDevice())return F.None;Kn.INSTANCE||(Kn.INSTANCE=new Kn);const t=Kn.INSTANCE.ignoreTargets.push(e);return pe(t)}static isTouchDevice(){return"ontouchstart"in Nt||navigator.maxTouchPoints>0}dispose(){this.handle&&(this.handle.dispose(),this.handle=null),super.dispose()}onTouchStart(e){const t=Date.now();this.handle&&(this.handle.dispose(),this.handle=null);for(let i=0,n=e.targetTouches.length;i<n;i++){const o=e.targetTouches.item(i);this.activeTouches[o.identifier]={id:o.identifier,initialTarget:o.target,initialTimeStamp:t,initialPageX:o.pageX,initialPageY:o.pageY,rollingTimestamps:[t],rollingPageX:[o.pageX],rollingPageY:[o.pageY]};const r=this.newGestureEvent(qt.Start,o.target);r.pageX=o.pageX,r.pageY=o.pageY,this.dispatchEvent(r)}this.dispatched&&(e.preventDefault(),e.stopPropagation(),this.dispatched=!1)}onTouchEnd(e,t){const i=Date.now(),n=Object.keys(this.activeTouches).length;for(let o=0,r=t.changedTouches.length;o<r;o++){const a=t.changedTouches.item(o);if(!this.activeTouches.hasOwnProperty(String(a.identifier))){console.warn("move of an UNKNOWN touch",a);continue}const l=this.activeTouches[a.identifier],c=Date.now()-l.initialTimeStamp;if(c<Kn.HOLD_DELAY&&Math.abs(l.initialPageX-l.rollingPageX.at(-1))<30&&Math.abs(l.initialPageY-l.rollingPageY.at(-1))<30){const d=this.newGestureEvent(qt.Tap,l.initialTarget);d.pageX=l.rollingPageX.at(-1),d.pageY=l.rollingPageY.at(-1),this.dispatchEvent(d)}else if(c>=Kn.HOLD_DELAY&&Math.abs(l.initialPageX-l.rollingPageX.at(-1))<30&&Math.abs(l.initialPageY-l.rollingPageY.at(-1))<30){const d=this.newGestureEvent(qt.Contextmenu,l.initialTarget);d.pageX=l.rollingPageX.at(-1),d.pageY=l.rollingPageY.at(-1),this.dispatchEvent(d)}else if(n===1){const d=l.rollingPageX.at(-1),h=l.rollingPageY.at(-1),u=l.rollingTimestamps.at(-1)-l.rollingTimestamps[0],g=d-l.rollingPageX[0],f=h-l.rollingPageY[0],p=[...this.targets].filter(_=>l.initialTarget instanceof Node&&_.contains(l.initialTarget));this.inertia(e,p,i,Math.abs(g)/u,g>0?1:-1,d,Math.abs(f)/u,f>0?1:-1,h)}this.dispatchEvent(this.newGestureEvent(qt.End,l.initialTarget)),delete this.activeTouches[a.identifier]}this.dispatched&&(t.preventDefault(),t.stopPropagation(),this.dispatched=!1)}newGestureEvent(e,t){const i=document.createEvent("CustomEvent");return i.initEvent(e,!1,!0),i.initialTarget=t,i.tapCount=0,i}dispatchEvent(e){if(e.type===qt.Tap){const t=new Date().getTime();let i=0;t-this._lastSetTapCountTime>Kn.CLEAR_TAP_COUNT_TIME?i=1:i=2,this._lastSetTapCountTime=t,e.tapCount=i}else(e.type===qt.Change||e.type===qt.Contextmenu)&&(this._lastSetTapCountTime=0);if(e.initialTarget instanceof Node){for(const i of this.ignoreTargets)if(i.contains(e.initialTarget))return;const t=[];for(const i of this.targets)if(i.contains(e.initialTarget)){let n=0,o=e.initialTarget;for(;o&&o!==i;)n++,o=o.parentElement;t.push([n,i])}t.sort((i,n)=>i[0]-n[0]);for(const[i,n]of t)n.dispatchEvent(e),this.dispatched=!0}}inertia(e,t,i,n,o,r,a,l,c){this.handle=Co(e,()=>{const d=Date.now(),h=d-i;let u=0,g=0,f=!0;n+=Kn.SCROLL_FRICTION*h,a+=Kn.SCROLL_FRICTION*h,n>0&&(f=!1,u=o*n*h),a>0&&(f=!1,g=l*a*h);const p=this.newGestureEvent(qt.Change);p.translationX=u,p.translationY=g,t.forEach(_=>_.dispatchEvent(p)),f||this.inertia(e,t,d,n,o,r+u,a,l,c+g)})}onTouchMove(e){const t=Date.now();for(let i=0,n=e.changedTouches.length;i<n;i++){const o=e.changedTouches.item(i);if(!this.activeTouches.hasOwnProperty(String(o.identifier))){console.warn("end of an UNKNOWN touch",o);continue}const r=this.activeTouches[o.identifier],a=this.newGestureEvent(qt.Change,r.initialTarget);a.translationX=o.pageX-r.rollingPageX.at(-1),a.translationY=o.pageY-r.rollingPageY.at(-1),a.pageX=o.pageX,a.pageY=o.pageY,this.dispatchEvent(a),r.rollingPageX.length>3&&(r.rollingPageX.shift(),r.rollingPageY.shift(),r.rollingTimestamps.shift()),r.rollingPageX.push(o.pageX),r.rollingPageY.push(o.pageY),r.rollingTimestamps.push(t)}this.dispatched&&(e.preventDefault(),e.stopPropagation(),this.dispatched=!1)}};Kn.SCROLL_FRICTION=-.005,Kn.HOLD_DELAY=700,Kn.CLEAR_TAP_COUNT_TIME=400;let _s=Kn;dbe([Ai],_s,"isTouchDevice",null);let Xo=class extends F{onclick(e,t){this._register(W(e,se.CLICK,i=>t(new is(be(e),i))))}onmousedown(e,t){this._register(W(e,se.MOUSE_DOWN,i=>t(new is(be(e),i))))}onmouseover(e,t){this._register(W(e,se.MOUSE_OVER,i=>t(new is(be(e),i))))}onmouseleave(e,t){this._register(W(e,se.MOUSE_LEAVE,i=>t(new is(be(e),i))))}onkeydown(e,t){this._register(W(e,se.KEY_DOWN,i=>t(new Pt(i))))}onkeyup(e,t){this._register(W(e,se.KEY_UP,i=>t(new Pt(i))))}oninput(e,t){this._register(W(e,se.INPUT,t))}onblur(e,t){this._register(W(e,se.BLUR,t))}onfocus(e,t){this._register(W(e,se.FOCUS,t))}ignoreGesture(e){return _s.ignoreTarget(e)}};const uv=11;class hbe extends Xo{constructor(e){super(),this._onActivate=e.onActivate,this.bgDomNode=document.createElement("div"),this.bgDomNode.className="arrow-background",this.bgDomNode.style.position="absolute",this.bgDomNode.style.width=e.bgWidth+"px",this.bgDomNode.style.height=e.bgHeight+"px",typeof e.top<"u"&&(this.bgDomNode.style.top="0px"),typeof e.left<"u"&&(this.bgDomNode.style.left="0px"),typeof e.bottom<"u"&&(this.bgDomNode.style.bottom="0px"),typeof e.right<"u"&&(this.bgDomNode.style.right="0px"),this.domNode=document.createElement("div"),this.domNode.className=e.className,this.domNode.classList.add(...xe.asClassNameArray(e.icon)),this.domNode.style.position="absolute",this.domNode.style.width=uv+"px",this.domNode.style.height=uv+"px",typeof e.top<"u"&&(this.domNode.style.top=e.top+"px"),typeof e.left<"u"&&(this.domNode.style.left=e.left+"px"),typeof e.bottom<"u"&&(this.domNode.style.bottom=e.bottom+"px"),typeof e.right<"u"&&(this.domNode.style.right=e.right+"px"),this._pointerMoveMonitor=this._register(new eC),this._register(Fi(this.bgDomNode,se.POINTER_DOWN,t=>this._arrowPointerDown(t))),this._register(Fi(this.domNode,se.POINTER_DOWN,t=>this._arrowPointerDown(t))),this._pointerdownRepeatTimer=this._register(new cW),this._pointerdownScheduleRepeatTimer=this._register(new $o)}_arrowPointerDown(e){if(!e.target||!(e.target instanceof Element))return;const t=()=>{this._pointerdownRepeatTimer.cancelAndSet(()=>this._onActivate(),1e3/24,be(e))};this._onActivate(),this._pointerdownRepeatTimer.cancel(),this._pointerdownScheduleRepeatTimer.cancelAndSet(t,200),this._pointerMoveMonitor.startMonitoring(e.target,e.pointerId,e.buttons,i=>{},()=>{this._pointerdownRepeatTimer.cancel(),this._pointerdownScheduleRepeatTimer.cancel()}),e.preventDefault()}}class ube extends F{constructor(e,t,i){super(),this._visibility=e,this._visibleClassName=t,this._invisibleClassName=i,this._domNode=null,this._isVisible=!1,this._isNeeded=!1,this._rawShouldBeVisible=!1,this._shouldBeVisible=!1,this._revealTimer=this._register(new $o)}setVisibility(e){this._visibility!==e&&(this._visibility=e,this._updateShouldBeVisible())}setShouldBeVisible(e){this._rawShouldBeVisible=e,this._updateShouldBeVisible()}_applyVisibilitySetting(){return this._visibility===2?!1:this._visibility===3?!0:this._rawShouldBeVisible}_updateShouldBeVisible(){const e=this._applyVisibilitySetting();this._shouldBeVisible!==e&&(this._shouldBeVisible=e,this.ensureVisibility())}setIsNeeded(e){this._isNeeded!==e&&(this._isNeeded=e,this.ensureVisibility())}setDomNode(e){this._domNode=e,this._domNode.setClassName(this._invisibleClassName),this.setShouldBeVisible(!1)}ensureVisibility(){if(!this._isNeeded){this._hide(!1);return}this._shouldBeVisible?this._reveal():this._hide(!0)}_reveal(){this._isVisible||(this._isVisible=!0,this._revealTimer.setIfNotSet(()=>{this._domNode?.setClassName(this._visibleClassName)},0))}_hide(e){this._revealTimer.cancel(),this._isVisible&&(this._isVisible=!1,this._domNode?.setClassName(this._invisibleClassName+(e?" fade":"")))}}const gbe=140;class eee extends Xo{constructor(e){super(),this._lazyRender=e.lazyRender,this._host=e.host,this._scrollable=e.scrollable,this._scrollByPage=e.scrollByPage,this._scrollbarState=e.scrollbarState,this._visibilityController=this._register(new ube(e.visibility,"visible scrollbar "+e.extraScrollbarClassName,"invisible scrollbar "+e.extraScrollbarClassName)),this._visibilityController.setIsNeeded(this._scrollbarState.isNeeded()),this._pointerMoveMonitor=this._register(new eC),this._shouldRender=!0,this.domNode=wt(document.createElement("div")),this.domNode.setAttribute("role","presentation"),this.domNode.setAttribute("aria-hidden","true"),this._visibilityController.setDomNode(this.domNode),this.domNode.setPosition("absolute"),this._register(W(this.domNode.domNode,se.POINTER_DOWN,t=>this._domNodePointerDown(t)))}_createArrow(e){const t=this._register(new hbe(e));this.domNode.domNode.appendChild(t.bgDomNode),this.domNode.domNode.appendChild(t.domNode)}_createSlider(e,t,i,n){this.slider=wt(document.createElement("div")),this.slider.setClassName("slider"),this.slider.setPosition("absolute"),this.slider.setTop(e),this.slider.setLeft(t),typeof i=="number"&&this.slider.setWidth(i),typeof n=="number"&&this.slider.setHeight(n),this.slider.setLayerHinting(!0),this.slider.setContain("strict"),this.domNode.domNode.appendChild(this.slider.domNode),this._register(W(this.slider.domNode,se.POINTER_DOWN,o=>{o.button===0&&(o.preventDefault(),this._sliderPointerDown(o))})),this.onclick(this.slider.domNode,o=>{o.leftButton&&o.stopPropagation()})}_onElementSize(e){return this._scrollbarState.setVisibleSize(e)&&(this._visibilityController.setIsNeeded(this._scrollbarState.isNeeded()),this._shouldRender=!0,this._lazyRender||this.render()),this._shouldRender}_onElementScrollSize(e){return this._scrollbarState.setScrollSize(e)&&(this._visibilityController.setIsNeeded(this._scrollbarState.isNeeded()),this._shouldRender=!0,this._lazyRender||this.render()),this._shouldRender}_onElementScrollPosition(e){return this._scrollbarState.setScrollPosition(e)&&(this._visibilityController.setIsNeeded(this._scrollbarState.isNeeded()),this._shouldRender=!0,this._lazyRender||this.render()),this._shouldRender}beginReveal(){this._visibilityController.setShouldBeVisible(!0)}beginHide(){this._visibilityController.setShouldBeVisible(!1)}render(){this._shouldRender&&(this._shouldRender=!1,this._renderDomNode(this._scrollbarState.getRectangleLargeSize(),this._scrollbarState.getRectangleSmallSize()),this._updateSlider(this._scrollbarState.getSliderSize(),this._scrollbarState.getArrowSize()+this._scrollbarState.getSliderPosition()))}_domNodePointerDown(e){e.target===this.domNode.domNode&&this._onPointerDown(e)}delegatePointerDown(e){const t=this.domNode.domNode.getClientRects()[0].top,i=t+this._scrollbarState.getSliderPosition(),n=t+this._scrollbarState.getSliderPosition()+this._scrollbarState.getSliderSize(),o=this._sliderPointerPosition(e);i<=o&&o<=n?e.button===0&&(e.preventDefault(),this._sliderPointerDown(e)):this._onPointerDown(e)}_onPointerDown(e){let t,i;if(e.target===this.domNode.domNode&&typeof e.offsetX=="number"&&typeof e.offsetY=="number")t=e.offsetX,i=e.offsetY;else{const r=Li(this.domNode.domNode);t=e.pageX-r.left,i=e.pageY-r.top}const n=e.pointerType==="mouse",o=e.button===0;if(o||!n){const r=this._pointerDownRelativePosition(t,i);this._setDesiredScrollPositionNow(this._scrollByPage?this._scrollbarState.getDesiredScrollPositionFromOffsetPaged(r):this._scrollbarState.getDesiredScrollPositionFromOffset(r))}o&&(e.preventDefault(),this._sliderPointerDown(e))}_sliderPointerDown(e){if(!e.target||!(e.target instanceof Element))return;const t=this._sliderPointerPosition(e),i=this._sliderOrthogonalPointerPosition(e),n=this._scrollbarState.clone();this.slider.toggleClassName("active",!0),this._pointerMoveMonitor.startMonitoring(e.target,e.pointerId,e.buttons,o=>{const r=this._sliderOrthogonalPointerPosition(o),a=Math.abs(r-i);if(Bn&&a>gbe){this._setDesiredScrollPositionNow(n.getScrollPosition());return}const c=this._sliderPointerPosition(o)-t;this._setDesiredScrollPositionNow(n.getDesiredScrollPositionFromDelta(c))},()=>{this.slider.toggleClassName("active",!1),this._host.onDragEnd()}),this._host.onDragStart()}_setDesiredScrollPositionNow(e){const t={};this.writeScrollPosition(t,e),this._scrollable.setScrollPositionNow(t)}updateScrollbarSize(e){this._updateScrollbarSize(e),this._scrollbarState.setScrollbarSize(e),this._shouldRender=!0,this._lazyRender||this.render()}isNeeded(){return this._scrollbarState.isNeeded()}}const fbe=20;class gv{constructor(e,t,i,n,o,r){this._scrollbarSize=Math.round(t),this._oppositeScrollbarSize=Math.round(i),this._arrowSize=Math.round(e),this._visibleSize=n,this._scrollSize=o,this._scrollPosition=r,this._computedAvailableSize=0,this._computedIsNeeded=!1,this._computedSliderSize=0,this._computedSliderRatio=0,this._computedSliderPosition=0,this._refreshComputedValues()}clone(){return new gv(this._arrowSize,this._scrollbarSize,this._oppositeScrollbarSize,this._visibleSize,this._scrollSize,this._scrollPosition)}setVisibleSize(e){const t=Math.round(e);return this._visibleSize!==t?(this._visibleSize=t,this._refreshComputedValues(),!0):!1}setScrollSize(e){const t=Math.round(e);return this._scrollSize!==t?(this._scrollSize=t,this._refreshComputedValues(),!0):!1}setScrollPosition(e){const t=Math.round(e);return this._scrollPosition!==t?(this._scrollPosition=t,this._refreshComputedValues(),!0):!1}setScrollbarSize(e){this._scrollbarSize=Math.round(e)}setOppositeScrollbarSize(e){this._oppositeScrollbarSize=Math.round(e)}static _computeValues(e,t,i,n,o){const r=Math.max(0,i-e),a=Math.max(0,r-2*t),l=n>0&&n>i;if(!l)return{computedAvailableSize:Math.round(r),computedIsNeeded:l,computedSliderSize:Math.round(a),computedSliderRatio:0,computedSliderPosition:0};const c=Math.round(Math.max(fbe,Math.floor(i*a/n))),d=(a-c)/(n-i),h=o*d;return{computedAvailableSize:Math.round(r),computedIsNeeded:l,computedSliderSize:Math.round(c),computedSliderRatio:d,computedSliderPosition:Math.round(h)}}_refreshComputedValues(){const e=gv._computeValues(this._oppositeScrollbarSize,this._arrowSize,this._visibleSize,this._scrollSize,this._scrollPosition);this._computedAvailableSize=e.computedAvailableSize,this._computedIsNeeded=e.computedIsNeeded,this._computedSliderSize=e.computedSliderSize,this._computedSliderRatio=e.computedSliderRatio,this._computedSliderPosition=e.computedSliderPosition}getArrowSize(){return this._arrowSize}getScrollPosition(){return this._scrollPosition}getRectangleLargeSize(){return this._computedAvailableSize}getRectangleSmallSize(){return this._scrollbarSize}isNeeded(){return this._computedIsNeeded}getSliderSize(){return this._computedSliderSize}getSliderPosition(){return this._computedSliderPosition}getDesiredScrollPositionFromOffset(e){if(!this._computedIsNeeded)return 0;const t=e-this._arrowSize-this._computedSliderSize/2;return Math.round(t/this._computedSliderRatio)}getDesiredScrollPositionFromOffsetPaged(e){if(!this._computedIsNeeded)return 0;const t=e-this._arrowSize;let i=this._scrollPosition;return t<this._computedSliderPosition?i-=this._visibleSize:i+=this._visibleSize,i}getDesiredScrollPositionFromDelta(e){if(!this._computedIsNeeded)return 0;const t=this._computedSliderPosition+e;return Math.round(t/this._computedSliderRatio)}}class pbe extends eee{constructor(e,t,i){const n=e.getScrollDimensions(),o=e.getCurrentScrollPosition();if(super({lazyRender:t.lazyRender,host:i,scrollbarState:new gv(t.horizontalHasArrows?t.arrowSize:0,t.horizontal===2?0:t.horizontalScrollbarSize,t.vertical===2?0:t.verticalScrollbarSize,n.width,n.scrollWidth,o.scrollLeft),visibility:t.horizontal,extraScrollbarClassName:"horizontal",scrollable:e,scrollByPage:t.scrollByPage}),t.horizontalHasArrows){const r=(t.arrowSize-uv)/2,a=(t.horizontalScrollbarSize-uv)/2;this._createArrow({className:"scra",icon:Z.scrollbarButtonLeft,top:a,left:r,bottom:void 0,right:void 0,bgWidth:t.arrowSize,bgHeight:t.horizontalScrollbarSize,onActivate:()=>this._host.onMouseWheel(new Cm(null,1,0))}),this._createArrow({className:"scra",icon:Z.scrollbarButtonRight,top:a,left:void 0,bottom:void 0,right:r,bgWidth:t.arrowSize,bgHeight:t.horizontalScrollbarSize,onActivate:()=>this._host.onMouseWheel(new Cm(null,-1,0))})}this._createSlider(Math.floor((t.horizontalScrollbarSize-t.horizontalSliderSize)/2),0,void 0,t.horizontalSliderSize)}_updateSlider(e,t){this.slider.setWidth(e),this.slider.setLeft(t)}_renderDomNode(e,t){this.domNode.setWidth(e),this.domNode.setHeight(t),this.domNode.setLeft(0),this.domNode.setBottom(0)}onDidScroll(e){return this._shouldRender=this._onElementScrollSize(e.scrollWidth)||this._shouldRender,this._shouldRender=this._onElementScrollPosition(e.scrollLeft)||this._shouldRender,this._shouldRender=this._onElementSize(e.width)||this._shouldRender,this._shouldRender}_pointerDownRelativePosition(e,t){return e}_sliderPointerPosition(e){return e.pageX}_sliderOrthogonalPointerPosition(e){return e.pageY}_updateScrollbarSize(e){this.slider.setHeight(e)}writeScrollPosition(e,t){e.scrollLeft=t}updateOptions(e){this.updateScrollbarSize(e.horizontal===2?0:e.horizontalScrollbarSize),this._scrollbarState.setOppositeScrollbarSize(e.vertical===2?0:e.verticalScrollbarSize),this._visibilityController.setVisibility(e.horizontal),this._scrollByPage=e.scrollByPage}}class mbe extends eee{constructor(e,t,i){const n=e.getScrollDimensions(),o=e.getCurrentScrollPosition();if(super({lazyRender:t.lazyRender,host:i,scrollbarState:new gv(t.verticalHasArrows?t.arrowSize:0,t.vertical===2?0:t.verticalScrollbarSize,0,n.height,n.scrollHeight,o.scrollTop),visibility:t.vertical,extraScrollbarClassName:"vertical",scrollable:e,scrollByPage:t.scrollByPage}),t.verticalHasArrows){const r=(t.arrowSize-uv)/2,a=(t.verticalScrollbarSize-uv)/2;this._createArrow({className:"scra",icon:Z.scrollbarButtonUp,top:r,left:a,bottom:void 0,right:void 0,bgWidth:t.verticalScrollbarSize,bgHeight:t.arrowSize,onActivate:()=>this._host.onMouseWheel(new Cm(null,0,1))}),this._createArrow({className:"scra",icon:Z.scrollbarButtonDown,top:void 0,left:a,bottom:r,right:void 0,bgWidth:t.verticalScrollbarSize,bgHeight:t.arrowSize,onActivate:()=>this._host.onMouseWheel(new Cm(null,0,-1))})}this._createSlider(0,Math.floor((t.verticalScrollbarSize-t.verticalSliderSize)/2),t.verticalSliderSize,void 0)}_updateSlider(e,t){this.slider.setHeight(e),this.slider.setTop(t)}_renderDomNode(e,t){this.domNode.setWidth(t),this.domNode.setHeight(e),this.domNode.setRight(0),this.domNode.setTop(0)}onDidScroll(e){return this._shouldRender=this._onElementScrollSize(e.scrollHeight)||this._shouldRender,this._shouldRender=this._onElementScrollPosition(e.scrollTop)||this._shouldRender,this._shouldRender=this._onElementSize(e.height)||this._shouldRender,this._shouldRender}_pointerDownRelativePosition(e,t){return t}_sliderPointerPosition(e){return e.pageY}_sliderOrthogonalPointerPosition(e){return e.pageX}_updateScrollbarSize(e){this.slider.setWidth(e)}writeScrollPosition(e,t){e.scrollTop=t}updateOptions(e){this.updateScrollbarSize(e.vertical===2?0:e.verticalScrollbarSize),this._scrollbarState.setOppositeScrollbarSize(0),this._visibilityController.setVisibility(e.vertical),this._scrollByPage=e.scrollByPage}}class YE{constructor(e,t,i,n,o,r,a){this._forceIntegerValues=e,this._scrollStateBrand=void 0,this._forceIntegerValues&&(t=t|0,i=i|0,n=n|0,o=o|0,r=r|0,a=a|0),this.rawScrollLeft=n,this.rawScrollTop=a,t<0&&(t=0),n+t>i&&(n=i-t),n<0&&(n=0),o<0&&(o=0),a+o>r&&(a=r-o),a<0&&(a=0),this.width=t,this.scrollWidth=i,this.scrollLeft=n,this.height=o,this.scrollHeight=r,this.scrollTop=a}equals(e){return this.rawScrollLeft===e.rawScrollLeft&&this.rawScrollTop===e.rawScrollTop&&this.width===e.width&&this.scrollWidth===e.scrollWidth&&this.scrollLeft===e.scrollLeft&&this.height===e.height&&this.scrollHeight===e.scrollHeight&&this.scrollTop===e.scrollTop}withScrollDimensions(e,t){return new YE(this._forceIntegerValues,typeof e.width<"u"?e.width:this.width,typeof e.scrollWidth<"u"?e.scrollWidth:this.scrollWidth,t?this.rawScrollLeft:this.scrollLeft,typeof e.height<"u"?e.height:this.height,typeof e.scrollHeight<"u"?e.scrollHeight:this.scrollHeight,t?this.rawScrollTop:this.scrollTop)}withScrollPosition(e){return new YE(this._forceIntegerValues,this.width,this.scrollWidth,typeof e.scrollLeft<"u"?e.scrollLeft:this.rawScrollLeft,this.height,this.scrollHeight,typeof e.scrollTop<"u"?e.scrollTop:this.rawScrollTop)}createScrollEvent(e,t){const i=this.width!==e.width,n=this.scrollWidth!==e.scrollWidth,o=this.scrollLeft!==e.scrollLeft,r=this.height!==e.height,a=this.scrollHeight!==e.scrollHeight,l=this.scrollTop!==e.scrollTop;return{inSmoothScrolling:t,oldWidth:e.width,oldScrollWidth:e.scrollWidth,oldScrollLeft:e.scrollLeft,width:this.width,scrollWidth:this.scrollWidth,scrollLeft:this.scrollLeft,oldHeight:e.height,oldScrollHeight:e.scrollHeight,oldScrollTop:e.scrollTop,height:this.height,scrollHeight:this.scrollHeight,scrollTop:this.scrollTop,widthChanged:i,scrollWidthChanged:n,scrollLeftChanged:o,heightChanged:r,scrollHeightChanged:a,scrollTopChanged:l}}}class sC extends F{constructor(e){super(),this._scrollableBrand=void 0,this._onScroll=this._register(new M),this.onScroll=this._onScroll.event,this._smoothScrollDuration=e.smoothScrollDuration,this._scheduleAtNextAnimationFrame=e.scheduleAtNextAnimationFrame,this._state=new YE(e.forceIntegerValues,0,0,0,0,0,0),this._smoothScrolling=null}dispose(){this._smoothScrolling&&(this._smoothScrolling.dispose(),this._smoothScrolling=null),super.dispose()}setSmoothScrollDuration(e){this._smoothScrollDuration=e}validateScrollPosition(e){return this._state.withScrollPosition(e)}getScrollDimensions(){return this._state}setScrollDimensions(e,t){const i=this._state.withScrollDimensions(e,t);this._setState(i,!!this._smoothScrolling),this._smoothScrolling?.acceptScrollDimensions(this._state)}getFutureScrollPosition(){return this._smoothScrolling?this._smoothScrolling.to:this._state}getCurrentScrollPosition(){return this._state}setScrollPositionNow(e){const t=this._state.withScrollPosition(e);this._smoothScrolling&&(this._smoothScrolling.dispose(),this._smoothScrolling=null),this._setState(t,!1)}setScrollPositionSmooth(e,t){if(this._smoothScrollDuration===0)return this.setScrollPositionNow(e);if(this._smoothScrolling){e={scrollLeft:typeof e.scrollLeft>"u"?this._smoothScrolling.to.scrollLeft:e.scrollLeft,scrollTop:typeof e.scrollTop>"u"?this._smoothScrolling.to.scrollTop:e.scrollTop};const i=this._state.withScrollPosition(e);if(this._smoothScrolling.to.scrollLeft===i.scrollLeft&&this._smoothScrolling.to.scrollTop===i.scrollTop)return;let n;t?n=new uS(this._smoothScrolling.from,i,this._smoothScrolling.startTime,this._smoothScrolling.duration):n=this._smoothScrolling.combine(this._state,i,this._smoothScrollDuration),this._smoothScrolling.dispose(),this._smoothScrolling=n}else{const i=this._state.withScrollPosition(e);this._smoothScrolling=uS.start(this._state,i,this._smoothScrollDuration)}this._smoothScrolling.animationFrameDisposable=this._scheduleAtNextAnimationFrame(()=>{this._smoothScrolling&&(this._smoothScrolling.animationFrameDisposable=null,this._performSmoothScrolling())})}hasPendingScrollAnimation(){return!!this._smoothScrolling}_performSmoothScrolling(){if(!this._smoothScrolling)return;const e=this._smoothScrolling.tick(),t=this._state.withScrollPosition(e);if(this._setState(t,!0),!!this._smoothScrolling){if(e.isDone){this._smoothScrolling.dispose(),this._smoothScrolling=null;return}this._smoothScrolling.animationFrameDisposable=this._scheduleAtNextAnimationFrame(()=>{this._smoothScrolling&&(this._smoothScrolling.animationFrameDisposable=null,this._performSmoothScrolling())})}}_setState(e,t){const i=this._state;i.equals(e)||(this._state=e,this._onScroll.fire(this._state.createScrollEvent(i,t)))}}class V${constructor(e,t,i){this.scrollLeft=e,this.scrollTop=t,this.isDone=i}}function iM(s,e){const t=e-s;return function(i){return s+t*wbe(i)}}function _be(s,e,t){return function(i){return i<t?s(i/t):e((i-t)/(1-t))}}class uS{constructor(e,t,i,n){this.from=e,this.to=t,this.duration=n,this.startTime=i,this.animationFrameDisposable=null,this._initAnimations()}_initAnimations(){this.scrollLeft=this._initAnimation(this.from.scrollLeft,this.to.scrollLeft,this.to.width),this.scrollTop=this._initAnimation(this.from.scrollTop,this.to.scrollTop,this.to.height)}_initAnimation(e,t,i){if(Math.abs(e-t)>2.5*i){let o,r;return e<t?(o=e+.75*i,r=t-.75*i):(o=e-.75*i,r=t+.75*i),_be(iM(e,o),iM(r,t),.33)}return iM(e,t)}dispose(){this.animationFrameDisposable!==null&&(this.animationFrameDisposable.dispose(),this.animationFrameDisposable=null)}acceptScrollDimensions(e){this.to=e.withScrollPosition(this.to),this._initAnimations()}tick(){return this._tick(Date.now())}_tick(e){const t=(e-this.startTime)/this.duration;if(t<1){const i=this.scrollLeft(t),n=this.scrollTop(t);return new V$(i,n,!1)}return new V$(this.to.scrollLeft,this.to.scrollTop,!0)}combine(e,t,i){return uS.start(e,t,i)}static start(e,t,i){i=i+10;const n=Date.now()-10;return new uS(e,t,n,i)}}function bbe(s){return Math.pow(s,3)}function wbe(s){return 1-bbe(1-s)}const vbe=500,z$=50;class Cbe{constructor(e,t,i){this.timestamp=e,this.deltaX=t,this.deltaY=i,this.score=0}}const GT=class GT{constructor(){this._capacity=5,this._memory=[],this._front=-1,this._rear=-1}isPhysicalMouseWheel(){if(this._front===-1&&this._rear===-1)return!1;let e=1,t=0,i=1,n=this._rear;do{const o=n===this._front?e:Math.pow(2,-i);if(e-=o,t+=this._memory[n].score*o,n===this._front)break;n=(this._capacity+n-1)%this._capacity,i++}while(!0);return t<=.5}acceptStandardWheelEvent(e){if(B0){const t=be(e.browserEvent),i=JO(t);this.accept(Date.now(),e.deltaX*i,e.deltaY*i)}else this.accept(Date.now(),e.deltaX,e.deltaY)}accept(e,t,i){let n=null;const o=new Cbe(e,t,i);this._front===-1&&this._rear===-1?(this._memory[0]=o,this._front=0,this._rear=0):(n=this._memory[this._rear],this._rear=(this._rear+1)%this._capacity,this._rear===this._front&&(this._front=(this._front+1)%this._capacity),this._memory[this._rear]=o),o.score=this._computeScore(o,n)}_computeScore(e,t){if(Math.abs(e.deltaX)>0&&Math.abs(e.deltaY)>0)return 1;let i=.5;if((!this._isAlmostInt(e.deltaX)||!this._isAlmostInt(e.deltaY))&&(i+=.25),t){const n=Math.abs(e.deltaX),o=Math.abs(e.deltaY),r=Math.abs(t.deltaX),a=Math.abs(t.deltaY),l=Math.max(Math.min(n,r),1),c=Math.max(Math.min(o,a),1),d=Math.max(n,r),h=Math.max(o,a);d%l===0&&h%c===0&&(i-=.5)}return Math.min(Math.max(i,0),1)}_isAlmostInt(e){const t=Number.EPSILON*100;return Math.abs(Math.round(e)-e)<.01+t}};GT.INSTANCE=new GT;let ZE=GT;class zW extends Xo{get onScroll(){return this._onScroll.event}get options(){return this._options}constructor(e,t,i){super(),this._inertialTimeout=null,this._inertialSpeed={X:0,Y:0},this._onScroll=this._register(new M),this._onWillScroll=this._register(new M),e.style.overflow="hidden",this._options=ybe(t),this._scrollable=i,this._register(this._scrollable.onScroll(o=>{this._onWillScroll.fire(o),this._onDidScroll(o),this._onScroll.fire(o)}));const n={onMouseWheel:o=>this._onMouseWheel(o),onDragStart:()=>this._onDragStart(),onDragEnd:()=>this._onDragEnd()};this._verticalScrollbar=this._register(new mbe(this._scrollable,this._options,n)),this._horizontalScrollbar=this._register(new pbe(this._scrollable,this._options,n)),this._domNode=document.createElement("div"),this._domNode.className="monaco-scrollable-element "+this._options.className,this._domNode.setAttribute("role","presentation"),this._domNode.style.position="relative",this._domNode.style.overflow="hidden",this._domNode.appendChild(e),this._domNode.appendChild(this._horizontalScrollbar.domNode.domNode),this._domNode.appendChild(this._verticalScrollbar.domNode.domNode),this._options.useShadows?(this._leftShadowDomNode=wt(document.createElement("div")),this._leftShadowDomNode.setClassName("shadow"),this._domNode.appendChild(this._leftShadowDomNode.domNode),this._topShadowDomNode=wt(document.createElement("div")),this._topShadowDomNode.setClassName("shadow"),this._domNode.appendChild(this._topShadowDomNode.domNode),this._topLeftShadowDomNode=wt(document.createElement("div")),this._topLeftShadowDomNode.setClassName("shadow"),this._domNode.appendChild(this._topLeftShadowDomNode.domNode)):(this._leftShadowDomNode=null,this._topShadowDomNode=null,this._topLeftShadowDomNode=null),this._listenOnDomNode=this._options.listenOnDomNode||this._domNode,this._mouseWheelToDispose=[],this._setListeningToMouseWheel(this._options.handleMouseWheel),this.onmouseover(this._listenOnDomNode,o=>this._onMouseOver(o)),this.onmouseleave(this._listenOnDomNode,o=>this._onMouseLeave(o)),this._hideTimeout=this._register(new $o),this._isDragging=!1,this._mouseIsOver=!1,this._shouldRender=!0,this._revealOnScroll=!0}dispose(){this._mouseWheelToDispose=xt(this._mouseWheelToDispose),this._inertialTimeout&&(this._inertialTimeout.dispose(),this._inertialTimeout=null),super.dispose()}getDomNode(){return this._domNode}getOverviewRulerLayoutInfo(){return{parent:this._domNode,insertBefore:this._verticalScrollbar.domNode.domNode}}delegateVerticalScrollbarPointerDown(e){this._verticalScrollbar.delegatePointerDown(e)}getScrollDimensions(){return this._scrollable.getScrollDimensions()}setScrollDimensions(e){this._scrollable.setScrollDimensions(e,!1)}updateClassName(e){this._options.className=e,Ye&&(this._options.className+=" mac"),this._domNode.className="monaco-scrollable-element "+this._options.className}updateOptions(e){typeof e.handleMouseWheel<"u"&&(this._options.handleMouseWheel=e.handleMouseWheel,this._setListeningToMouseWheel(this._options.handleMouseWheel)),typeof e.mouseWheelScrollSensitivity<"u"&&(this._options.mouseWheelScrollSensitivity=e.mouseWheelScrollSensitivity),typeof e.fastScrollSensitivity<"u"&&(this._options.fastScrollSensitivity=e.fastScrollSensitivity),typeof e.scrollPredominantAxis<"u"&&(this._options.scrollPredominantAxis=e.scrollPredominantAxis),typeof e.horizontal<"u"&&(this._options.horizontal=e.horizontal),typeof e.vertical<"u"&&(this._options.vertical=e.vertical),typeof e.horizontalScrollbarSize<"u"&&(this._options.horizontalScrollbarSize=e.horizontalScrollbarSize),typeof e.verticalScrollbarSize<"u"&&(this._options.verticalScrollbarSize=e.verticalScrollbarSize),typeof e.scrollByPage<"u"&&(this._options.scrollByPage=e.scrollByPage),this._horizontalScrollbar.updateOptions(this._options),this._verticalScrollbar.updateOptions(this._options),this._options.lazyRender||this._render()}delegateScrollFromMouseWheelEvent(e){this._onMouseWheel(new Cm(e))}async _periodicSync(){let e=!1;(this._inertialSpeed.X!==0||this._inertialSpeed.Y!==0)&&(this._scrollable.setScrollPositionNow({scrollTop:this._scrollable.getCurrentScrollPosition().scrollTop-this._inertialSpeed.Y*100,scrollLeft:this._scrollable.getCurrentScrollPosition().scrollLeft-this._inertialSpeed.X*100}),this._inertialSpeed.X*=.9,this._inertialSpeed.Y*=.9,Math.abs(this._inertialSpeed.X)<.01&&(this._inertialSpeed.X=0),Math.abs(this._inertialSpeed.Y)<.01&&(this._inertialSpeed.Y=0),e=this._inertialSpeed.X!==0||this._inertialSpeed.Y!==0),e?(this._inertialTimeout||(this._inertialTimeout=new $o),this._inertialTimeout.cancelAndSet(()=>this._periodicSync(),1e3/60)):(this._inertialTimeout?.dispose(),this._inertialTimeout=null)}_setListeningToMouseWheel(e){if(this._mouseWheelToDispose.length>0!==e&&(this._mouseWheelToDispose=xt(this._mouseWheelToDispose),e)){const i=n=>{this._onMouseWheel(new Cm(n))};this._mouseWheelToDispose.push(W(this._listenOnDomNode,se.MOUSE_WHEEL,i,{passive:!1}))}}_onMouseWheel(e){if(e.browserEvent?.defaultPrevented)return;const t=ZE.INSTANCE;t.acceptStandardWheelEvent(e);let i=!1;if(e.deltaY||e.deltaX){let o=e.deltaY*this._options.mouseWheelScrollSensitivity,r=e.deltaX*this._options.mouseWheelScrollSensitivity;this._options.scrollPredominantAxis&&(this._options.scrollYToX&&r+o===0?r=o=0:Math.abs(o)>=Math.abs(r)?r=0:o=0),this._options.flipAxes&&([o,r]=[r,o]);const a=!Ye&&e.browserEvent&&e.browserEvent.shiftKey;(this._options.scrollYToX||a)&&!r&&(r=o,o=0),e.browserEvent&&e.browserEvent.altKey&&(r=r*this._options.fastScrollSensitivity,o=o*this._options.fastScrollSensitivity);const l=this._scrollable.getFutureScrollPosition();let c={};if(o){const d=z$*o,h=l.scrollTop-(d<0?Math.floor(d):Math.ceil(d));this._verticalScrollbar.writeScrollPosition(c,h)}if(r){const d=z$*r,h=l.scrollLeft-(d<0?Math.floor(d):Math.ceil(d));this._horizontalScrollbar.writeScrollPosition(c,h)}if(c=this._scrollable.validateScrollPosition(c),this._options.inertialScroll&&(r||o)&&!t.isPhysicalMouseWheel()){let d=!1;this._inertialSpeed.X===0&&this._inertialSpeed.Y===0&&(d=!0),this._inertialSpeed.Y=(o<0?-1:1)*Math.abs(o)**1.02,this._inertialSpeed.X=(r<0?-1:1)*Math.abs(r)**1.02,d&&this._periodicSync()}(l.scrollLeft!==c.scrollLeft||l.scrollTop!==c.scrollTop)&&(this._options.mouseWheelSmoothScroll&&t.isPhysicalMouseWheel()?this._scrollable.setScrollPositionSmooth(c):this._scrollable.setScrollPositionNow(c),i=!0)}let n=i;!n&&this._options.alwaysConsumeMouseWheel&&(n=!0),!n&&this._options.consumeMouseWheelIfScrollbarIsNeeded&&(this._verticalScrollbar.isNeeded()||this._horizontalScrollbar.isNeeded())&&(n=!0),n&&(e.preventDefault(),e.stopPropagation())}_onDidScroll(e){this._shouldRender=this._horizontalScrollbar.onDidScroll(e)||this._shouldRender,this._shouldRender=this._verticalScrollbar.onDidScroll(e)||this._shouldRender,this._options.useShadows&&(this._shouldRender=!0),this._revealOnScroll&&this._reveal(),this._options.lazyRender||this._render()}renderNow(){if(!this._options.lazyRender)throw new Error("Please use `lazyRender` together with `renderNow`!");this._render()}_render(){if(this._shouldRender&&(this._shouldRender=!1,this._horizontalScrollbar.render(),this._verticalScrollbar.render(),this._options.useShadows)){const e=this._scrollable.getCurrentScrollPosition(),t=e.scrollTop>0,i=e.scrollLeft>0,n=i?" left":"",o=t?" top":"",r=i||t?" top-left-corner":"";this._leftShadowDomNode.setClassName(`shadow${n}`),this._topShadowDomNode.setClassName(`shadow${o}`),this._topLeftShadowDomNode.setClassName(`shadow${r}${o}${n}`)}}_onDragStart(){this._isDragging=!0,this._reveal()}_onDragEnd(){this._isDragging=!1,this._hide()}_onMouseLeave(e){this._mouseIsOver=!1,this._hide()}_onMouseOver(e){this._mouseIsOver=!0,this._reveal()}_reveal(){this._verticalScrollbar.beginReveal(),this._horizontalScrollbar.beginReveal(),this._scheduleHide()}_hide(){!this._mouseIsOver&&!this._isDragging&&(this._verticalScrollbar.beginHide(),this._horizontalScrollbar.beginHide())}_scheduleHide(){!this._mouseIsOver&&!this._isDragging&&this._hideTimeout.cancelAndSet(()=>this._hide(),vbe)}}class tee extends zW{constructor(e,t){t=t||{},t.mouseWheelSmoothScroll=!1;const i=new sC({forceIntegerValues:!0,smoothScrollDuration:0,scheduleAtNextAnimationFrame:n=>Co(be(e),n)});super(e,t,i),this._register(i)}setScrollPosition(e){this._scrollable.setScrollPositionNow(e)}}class cA extends zW{constructor(e,t,i){super(e,t,i)}setScrollPosition(e){e.reuseAnimation?this._scrollable.setScrollPositionSmooth(e,e.reuseAnimation):this._scrollable.setScrollPositionNow(e)}getScrollPosition(){return this._scrollable.getCurrentScrollPosition()}}class tx extends zW{constructor(e,t){t=t||{},t.mouseWheelSmoothScroll=!1;const i=new sC({forceIntegerValues:!1,smoothScrollDuration:0,scheduleAtNextAnimationFrame:n=>Co(be(e),n)});super(e,t,i),this._register(i),this._element=e,this._register(this.onScroll(n=>{n.scrollTopChanged&&(this._element.scrollTop=n.scrollTop),n.scrollLeftChanged&&(this._element.scrollLeft=n.scrollLeft)})),this.scanDomNode()}setScrollPosition(e){this._scrollable.setScrollPositionNow(e)}getScrollPosition(){return this._scrollable.getCurrentScrollPosition()}scanDomNode(){this.setScrollDimensions({width:this._element.clientWidth,scrollWidth:this._element.scrollWidth,height:this._element.clientHeight,scrollHeight:this._element.scrollHeight}),this.setScrollPosition({scrollLeft:this._element.scrollLeft,scrollTop:this._element.scrollTop})}}function ybe(s){const e={lazyRender:typeof s.lazyRender<"u"?s.lazyRender:!1,className:typeof s.className<"u"?s.className:"",useShadows:typeof s.useShadows<"u"?s.useShadows:!0,handleMouseWheel:typeof s.handleMouseWheel<"u"?s.handleMouseWheel:!0,flipAxes:typeof s.flipAxes<"u"?s.flipAxes:!1,consumeMouseWheelIfScrollbarIsNeeded:typeof s.consumeMouseWheelIfScrollbarIsNeeded<"u"?s.consumeMouseWheelIfScrollbarIsNeeded:!1,alwaysConsumeMouseWheel:typeof s.alwaysConsumeMouseWheel<"u"?s.alwaysConsumeMouseWheel:!1,scrollYToX:typeof s.scrollYToX<"u"?s.scrollYToX:!1,mouseWheelScrollSensitivity:typeof s.mouseWheelScrollSensitivity<"u"?s.mouseWheelScrollSensitivity:1,fastScrollSensitivity:typeof s.fastScrollSensitivity<"u"?s.fastScrollSensitivity:5,scrollPredominantAxis:typeof s.scrollPredominantAxis<"u"?s.scrollPredominantAxis:!0,mouseWheelSmoothScroll:typeof s.mouseWheelSmoothScroll<"u"?s.mouseWheelSmoothScroll:!0,inertialScroll:typeof s.inertialScroll<"u"?s.inertialScroll:!1,arrowSize:typeof s.arrowSize<"u"?s.arrowSize:11,listenOnDomNode:typeof s.listenOnDomNode<"u"?s.listenOnDomNode:null,horizontal:typeof s.horizontal<"u"?s.horizontal:1,horizontalScrollbarSize:typeof s.horizontalScrollbarSize<"u"?s.horizontalScrollbarSize:10,horizontalSliderSize:typeof s.horizontalSliderSize<"u"?s.horizontalSliderSize:0,horizontalHasArrows:typeof s.horizontalHasArrows<"u"?s.horizontalHasArrows:!1,vertical:typeof s.vertical<"u"?s.vertical:1,verticalScrollbarSize:typeof s.verticalScrollbarSize<"u"?s.verticalScrollbarSize:10,verticalHasArrows:typeof s.verticalHasArrows<"u"?s.verticalHasArrows:!1,verticalSliderSize:typeof s.verticalSliderSize<"u"?s.verticalSliderSize:0,scrollByPage:typeof s.scrollByPage<"u"?s.scrollByPage:!1};return e.horizontalSliderSize=typeof s.horizontalSliderSize<"u"?s.horizontalSliderSize:e.horizontalScrollbarSize,e.verticalSliderSize=typeof s.verticalSliderSize<"u"?s.verticalSliderSize:e.verticalScrollbarSize,Ye&&(e.className+=" mac"),e}class iee extends F{constructor(e,t,i,n){super(),this._context=e,this._viewHelper=t,this._mouseTargetFactory=i,this._dispatchMouse=n,this._operation=null}dispose(){super.dispose(),this.stop()}start(e,t){this._operation?this._operation.setPosition(e,t):this._operation=this._createDragScrollingOperation(e,t)}stop(){this._operation&&(this._operation.dispose(),this._operation=null)}}class nee extends F{constructor(e,t,i,n,o,r){super(),this._context=e,this._viewHelper=t,this._mouseTargetFactory=i,this._dispatchMouse=n,this._position=o,this._mouseEvent=r,this._lastTime=Date.now(),this._animationFrameDisposable=Co(be(r.browserEvent),()=>this._execute())}dispose(){this._animationFrameDisposable.dispose(),super.dispose()}setPosition(e,t){this._position=e,this._mouseEvent=t}_tick(){const e=Date.now(),t=e-this._lastTime;return this._lastTime=e,t}}class Sbe extends iee{_createDragScrollingOperation(e,t){return new xbe(this._context,this._viewHelper,this._mouseTargetFactory,this._dispatchMouse,e,t)}}class xbe extends nee{_getScrollSpeed(){const e=this._context.configuration.options.get(75),t=this._context.configuration.options.get(165).height/e,i=this._position.outsideDistance/e;return i<=1.5?Math.max(30,t*(1+i)):i<=3?Math.max(60,t*(2+i)):Math.max(200,t*(7+i))}_execute(){const e=this._context.configuration.options.get(75),t=this._getScrollSpeed(),i=this._tick(),n=t*(i/1e3)*e,o=this._position.outsidePosition==="above"?-n:n;this._context.viewModel.viewLayout.deltaScrollNow(0,o),this._viewHelper.renderNow();const r=this._context.viewLayout.getLinesViewportData(),a=this._position.outsidePosition==="above"?r.startLineNumber:r.endLineNumber;let l;{const c=rA(this._viewHelper.viewDomNode),d=this._context.configuration.options.get(165).horizontalScrollbarHeight,h=new J0(this._mouseEvent.pos.x,c.y+c.height-d-.1),u=aA(this._viewHelper.viewDomNode,c,h);l=this._mouseTargetFactory.createMouseTarget(this._viewHelper.getLastRenderData(),c,h,u,null)}(!l.position||l.position.lineNumber!==a)&&(this._position.outsidePosition==="above"?l=Zn.createOutsideEditor(this._position.mouseColumn,new A(a,1),"above",this._position.outsideDistance):l=Zn.createOutsideEditor(this._position.mouseColumn,new A(a,this._context.viewModel.getLineMaxColumn(a)),"below",this._position.outsideDistance)),this._dispatchMouse(l,!0,2),this._animationFrameDisposable=Co(be(l.element),()=>this._execute())}}class Lbe extends iee{_createDragScrollingOperation(e,t){return new kbe(this._context,this._viewHelper,this._mouseTargetFactory,this._dispatchMouse,e,t)}}class kbe extends nee{_getScrollSpeed(){const e=this._context.configuration.options.get(59).typicalFullwidthCharacterWidth,t=this._context.configuration.options.get(165).contentWidth/e,i=this._position.outsideDistance/e;return i<=1.5?Math.max(30,t*(1+i)):i<=3?Math.max(60,t*(2+i)):Math.max(200,t*(7+i))}_execute(){const e=this._context.configuration.options.get(59).typicalFullwidthCharacterWidth,t=this._getScrollSpeed(),i=this._tick(),n=t*(i/1e3)*e*.5,o=this._position.outsidePosition==="left"?-n:n;if(this._context.viewModel.viewLayout.deltaScrollNow(o,0),this._viewHelper.renderNow(),!this._position.position)return;const r=this._position.position.lineNumber;let a;{const l=rA(this._viewHelper.viewDomNode),c=this._context.configuration.options.get(165).horizontalScrollbarHeight,d=new J0(this._mouseEvent.pos.x,l.y+l.height-c-.1),h=aA(this._viewHelper.viewDomNode,l,d);a=this._mouseTargetFactory.createMouseTarget(this._viewHelper.getLastRenderData(),l,d,h,null)}this._position.outsidePosition==="left"?a=Zn.createOutsideEditor(a.mouseColumn,new A(r,a.mouseColumn),"left",this._position.outsideDistance):a=Zn.createOutsideEditor(a.mouseColumn,new A(r,a.mouseColumn),"right",this._position.outsideDistance),this._dispatchMouse(a,!0,2),this._animationFrameDisposable=Co(be(a.element),()=>this._execute())}}class $W extends c_{constructor(e,t,i){super(),this._mouseLeaveMonitor=null,this._context=e,this.viewController=t,this.viewHelper=i,this.mouseTargetFactory=new Ls(this._context,i),this._mouseDownOperation=this._register(new Ebe(this._context,this.viewController,this.viewHelper,this.mouseTargetFactory,(r,a)=>this._createMouseTarget(r,a),r=>this._getMouseColumn(r))),this.lastMouseLeaveTime=-1,this._height=this._context.configuration.options.get(165).height;const n=new P_e(this.viewHelper.viewDomNode);this._register(n.onContextMenu(this.viewHelper.viewDomNode,r=>this._onContextMenu(r,!0))),this._register(n.onMouseMove(this.viewHelper.viewDomNode,r=>{this._onMouseMove(r),this._mouseLeaveMonitor||(this._mouseLeaveMonitor=W(this.viewHelper.viewDomNode.ownerDocument,"mousemove",a=>{this.viewHelper.viewDomNode.contains(a.target)||this._onMouseLeave(new bh(a,!1,this.viewHelper.viewDomNode))}))})),this._register(n.onMouseUp(this.viewHelper.viewDomNode,r=>this._onMouseUp(r))),this._register(n.onMouseLeave(this.viewHelper.viewDomNode,r=>this._onMouseLeave(r)));let o=0;this._register(n.onPointerDown(this.viewHelper.viewDomNode,(r,a)=>{o=a})),this._register(W(this.viewHelper.viewDomNode,se.POINTER_UP,r=>{this._mouseDownOperation.onPointerUp()})),this._register(n.onMouseDown(this.viewHelper.viewDomNode,r=>this._onMouseDown(r,o))),this._setupMouseWheelZoomListener(),this._context.addEventHandler(this)}_setupMouseWheelZoomListener(){const e=ZE.INSTANCE;let t=0,i=Gr.getZoomLevel(),n=!1,o=0;const r=l=>{if(this.viewController.emitMouseWheel(l),!this._context.configuration.options.get(84))return;const c=new Cm(l);if(e.acceptStandardWheelEvent(c),e.isPhysicalMouseWheel()){if(a(l)){const d=Gr.getZoomLevel(),h=c.deltaY>0?1:-1;Gr.setZoomLevel(d+h),c.preventDefault(),c.stopPropagation()}}else Date.now()-t>50&&(i=Gr.getZoomLevel(),n=a(l),o=0),t=Date.now(),o+=c.deltaY,n&&(Gr.setZoomLevel(i+o/5),c.preventDefault(),c.stopPropagation())};this._register(W(this.viewHelper.viewDomNode,se.MOUSE_WHEEL,r,{capture:!0,passive:!1}));function a(l){return Ye?(l.metaKey||l.ctrlKey)&&!l.shiftKey&&!l.altKey:l.ctrlKey&&!l.metaKey&&!l.shiftKey&&!l.altKey}}dispose(){this._context.removeEventHandler(this),this._mouseLeaveMonitor&&(this._mouseLeaveMonitor.dispose(),this._mouseLeaveMonitor=null),super.dispose()}onConfigurationChanged(e){if(e.hasChanged(165)){const t=this._context.configuration.options.get(165).height;this._height!==t&&(this._height=t,this._mouseDownOperation.onHeightChanged())}return!1}onCursorStateChanged(e){return this._mouseDownOperation.onCursorStateChanged(e),!1}onFocusChanged(e){return!1}getTargetAtClientPoint(e,t){const n=new jJ(e,t).toPageCoordinates(be(this.viewHelper.viewDomNode)),o=rA(this.viewHelper.viewDomNode);if(n.y<o.y||n.y>o.y+o.height||n.x<o.x||n.x>o.x+o.width)return null;const r=aA(this.viewHelper.viewDomNode,o,n);return this.mouseTargetFactory.createMouseTarget(this.viewHelper.getLastRenderData(),o,n,r,null)}_createMouseTarget(e,t){let i=e.target;if(!this.viewHelper.viewDomNode.contains(i)){const n=cg(this.viewHelper.viewDomNode);n&&(i=n.elementsFromPoint(e.posx,e.posy).find(r=>this.viewHelper.viewDomNode.contains(r))??null)}return this.mouseTargetFactory.createMouseTarget(this.viewHelper.getLastRenderData(),e.editorPos,e.pos,e.relativePos,t?i:null)}_getMouseColumn(e){return this.mouseTargetFactory.getMouseColumn(e.relativePos)}_onContextMenu(e,t){this.viewController.emitContextMenu({event:e,target:this._createMouseTarget(e,t)})}_onMouseMove(e){this.mouseTargetFactory.mouseTargetIsWidget(e)||e.preventDefault(),!(this._mouseDownOperation.isActive()||e.timestamp<this.lastMouseLeaveTime)&&this.viewController.emitMouseMove({event:e,target:this._createMouseTarget(e,!0)})}_onMouseLeave(e){this._mouseLeaveMonitor&&(this._mouseLeaveMonitor.dispose(),this._mouseLeaveMonitor=null),this.lastMouseLeaveTime=new Date().getTime(),this.viewController.emitMouseLeave({event:e,target:null})}_onMouseUp(e){this.viewController.emitMouseUp({event:e,target:this._createMouseTarget(e,!0)})}_onMouseDown(e,t){const i=this._createMouseTarget(e,!0),n=i.type===6||i.type===7,o=i.type===2||i.type===3||i.type===4,r=i.type===3,a=this._context.configuration.options.get(125),l=i.type===8||i.type===5,c=i.type===9;let d=e.leftButton||e.middleButton;Ye&&e.leftButton&&e.ctrlKey&&(d=!1);const h=()=>{e.preventDefault(),this.viewHelper.focusTextArea()};if(d&&(n||r&&a))h(),this._mouseDownOperation.start(i.type,e,t);else if(o)e.preventDefault();else if(l){const u=i.detail;d&&this.viewHelper.shouldSuppressMouseDownOnViewZone(u.viewZoneId)&&(h(),this._mouseDownOperation.start(i.type,e,t),e.preventDefault())}else c&&this.viewHelper.shouldSuppressMouseDownOnWidget(i.detail)&&(h(),e.preventDefault());this.viewController.emitMouseDown({event:e,target:i})}}class Ebe extends F{constructor(e,t,i,n,o,r){super(),this._context=e,this._viewController=t,this._viewHelper=i,this._mouseTargetFactory=n,this._createMouseTarget=o,this._getMouseColumn=r,this._mouseMoveMonitor=this._register(new B_e(this._viewHelper.viewDomNode)),this._topBottomDragScrolling=this._register(new Sbe(this._context,this._viewHelper,this._mouseTargetFactory,(a,l,c)=>this._dispatchMouse(a,l,c))),this._leftRightDragScrolling=this._register(new Lbe(this._context,this._viewHelper,this._mouseTargetFactory,(a,l,c)=>this._dispatchMouse(a,l,c))),this._mouseState=new iF,this._currentSelection=new he(1,1,1,1),this._isActive=!1,this._lastMouseEvent=null}dispose(){super.dispose()}isActive(){return this._isActive}_onMouseDownThenMove(e){this._lastMouseEvent=e,this._mouseState.setModifiers(e);const t=this._findMousePosition(e,!1);t&&(this._mouseState.isDragAndDrop?this._viewController.emitMouseDrag({event:e,target:t}):t.type===13?t.outsidePosition==="above"||t.outsidePosition==="below"?(this._topBottomDragScrolling.start(t,e),this._leftRightDragScrolling.stop()):(this._leftRightDragScrolling.start(t,e),this._topBottomDragScrolling.stop()):(this._topBottomDragScrolling.stop(),this._leftRightDragScrolling.stop(),this._dispatchMouse(t,!0,1)))}start(e,t,i){this._lastMouseEvent=t,this._mouseState.setStartedOnLineNumbers(e===3),this._mouseState.setStartButtons(t),this._mouseState.setModifiers(t);const n=this._findMousePosition(t,!0);if(!n||!n.position)return;this._mouseState.trySetCount(t.detail,n.position),t.detail=this._mouseState.count;const o=this._context.configuration.options;if(!o.get(104)&&o.get(42)&&!o.get(28)&&!this._mouseState.altKey&&t.detail<2&&!this._isActive&&!this._currentSelection.isEmpty()&&n.type===6&&n.position&&this._currentSelection.containsPosition(n.position)){this._mouseState.isDragAndDrop=!0,this._isActive=!0,this._mouseMoveMonitor.startMonitoring(this._viewHelper.viewLinesDomNode,i,t.buttons,r=>this._onMouseDownThenMove(r),r=>{const a=this._findMousePosition(this._lastMouseEvent,!1);Id(r)?this._viewController.emitMouseDropCanceled():this._viewController.emitMouseDrop({event:this._lastMouseEvent,target:a?this._createMouseTarget(this._lastMouseEvent,!0):null}),this._stop()});return}this._mouseState.isDragAndDrop=!1,this._dispatchMouse(n,t.shiftKey,1),this._isActive||(this._isActive=!0,this._mouseMoveMonitor.startMonitoring(this._viewHelper.viewLinesDomNode,i,t.buttons,r=>this._onMouseDownThenMove(r),()=>this._stop()))}_stop(){this._isActive=!1,this._topBottomDragScrolling.stop(),this._leftRightDragScrolling.stop()}onHeightChanged(){this._mouseMoveMonitor.stopMonitoring()}onPointerUp(){this._mouseMoveMonitor.stopMonitoring()}onCursorStateChanged(e){this._currentSelection=e.selections[0]}_getPositionOutsideEditor(e){const t=e.editorPos,i=this._context.viewModel,n=this._context.viewLayout,o=this._getMouseColumn(e);if(e.posy<t.y){const h=t.y-e.posy,u=Math.max(n.getCurrentScrollTop()-h,0),g=hv.getZoneAtCoord(this._context,u);if(g){const p=this._helpPositionJumpOverViewZone(g);if(p)return Zn.createOutsideEditor(o,p,"above",h)}const f=n.getLineNumberAtVerticalOffset(u);return Zn.createOutsideEditor(o,new A(f,1),"above",h)}if(e.posy>t.y+t.height){const h=e.posy-t.y-t.height,u=n.getCurrentScrollTop()+e.relativePos.y,g=hv.getZoneAtCoord(this._context,u);if(g){const p=this._helpPositionJumpOverViewZone(g);if(p)return Zn.createOutsideEditor(o,p,"below",h)}const f=n.getLineNumberAtVerticalOffset(u);return Zn.createOutsideEditor(o,new A(f,i.getLineMaxColumn(f)),"below",h)}const r=n.getLineNumberAtVerticalOffset(n.getCurrentScrollTop()+e.relativePos.y),a=this._context.configuration.options.get(165),l=a.contentLeft;if(e.relativePos.x<=l){const h=l-e.relativePos.x;return Zn.createOutsideEditor(o,new A(r,1),"left",h)}const d=a.minimap.minimapLeft===0?a.width-a.verticalScrollbarWidth:a.minimap.minimapLeft;if(e.relativePos.x>=d){const h=e.relativePos.x-d;return Zn.createOutsideEditor(o,new A(r,i.getLineMaxColumn(r)),"right",h)}return null}_findMousePosition(e,t){const i=this._getPositionOutsideEditor(e);if(i)return i;const n=this._createMouseTarget(e,t);if(!n.position)return null;if(n.type===8||n.type===5){const r=this._helpPositionJumpOverViewZone(n.detail);if(r)return Zn.createViewZone(n.type,n.element,n.mouseColumn,r,n.detail)}return n}_helpPositionJumpOverViewZone(e){const t=new A(this._currentSelection.selectionStartLineNumber,this._currentSelection.selectionStartColumn),i=e.positionBefore,n=e.positionAfter;return i&&n?i.isBefore(t)?i:n:null}_dispatchMouse(e,t,i){e.position&&this._viewController.dispatchMouse({position:e.position,mouseColumn:e.mouseColumn,startedOnLineNumbers:this._mouseState.startedOnLineNumbers,revealType:i,inSelectionMode:t,mouseDownCount:this._mouseState.count,altKey:this._mouseState.altKey,ctrlKey:this._mouseState.ctrlKey,metaKey:this._mouseState.metaKey,shiftKey:this._mouseState.shiftKey,leftButton:this._mouseState.leftButton,middleButton:this._mouseState.middleButton,onInjectedText:e.type===6&&e.detail.injectedText!==null})}}const KT=class KT{get altKey(){return this._altKey}get ctrlKey(){return this._ctrlKey}get metaKey(){return this._metaKey}get shiftKey(){return this._shiftKey}get leftButton(){return this._leftButton}get middleButton(){return this._middleButton}get startedOnLineNumbers(){return this._startedOnLineNumbers}constructor(){this._altKey=!1,this._ctrlKey=!1,this._metaKey=!1,this._shiftKey=!1,this._leftButton=!1,this._middleButton=!1,this._startedOnLineNumbers=!1,this._lastMouseDownPosition=null,this._lastMouseDownPositionEqualCount=0,this._lastMouseDownCount=0,this._lastSetMouseDownCountTime=0,this.isDragAndDrop=!1}get count(){return this._lastMouseDownCount}setModifiers(e){this._altKey=e.altKey,this._ctrlKey=e.ctrlKey,this._metaKey=e.metaKey,this._shiftKey=e.shiftKey}setStartButtons(e){this._leftButton=e.leftButton,this._middleButton=e.middleButton}setStartedOnLineNumbers(e){this._startedOnLineNumbers=e}trySetCount(e,t){const i=new Date().getTime();i-this._lastSetMouseDownCountTime>KT.CLEAR_MOUSE_DOWN_COUNT_TIME&&(e=1),this._lastSetMouseDownCountTime=i,e>this._lastMouseDownCount+1&&(e=this._lastMouseDownCount+1),this._lastMouseDownPosition&&this._lastMouseDownPosition.equals(t)?this._lastMouseDownPositionEqualCount++:this._lastMouseDownPositionEqualCount=1,this._lastMouseDownPosition=t,this._lastMouseDownCount=Math.min(e,this._lastMouseDownPositionEqualCount)}};KT.CLEAR_MOUSE_DOWN_COUNT_TIME=400;let iF=KT;class Lt{get event(){return this.emitter.event}constructor(e,t,i){const n=o=>this.emitter.fire(o);this.emitter=new M({onWillAddFirstListener:()=>e.addEventListener(t,n,i),onDidRemoveLastListener:()=>e.removeEventListener(t,n,i)})}dispose(){this.emitter.dispose()}}function see(s,e,t,i){const n=s.getPlainTextToCopy(e,t,Bn),o=s.model.getEOL(),r=t&&e.length===1&&e[0].isEmpty(),a=Array.isArray(n)?n:null,l=Array.isArray(n)?n.join(o):n;let c,d=null;if(nF.forceCopyWithSyntaxHighlighting||i&&l.length<65536){const u=s.getRichTextToCopy(e,t);u&&(c=u.html,d=u.mode)}return{isFromEmptySelection:r,multicursorText:a,text:l,html:c,mode:d}}const YT=class YT{constructor(){this._lastState=null}set(e,t){this._lastState={lastCopiedValue:e,data:t}}get(e){return this._lastState&&this._lastState.lastCopiedValue===e?this._lastState.data:(this._lastState=null,null)}};YT.INSTANCE=new YT;let xc=YT;const nF={forceCopyWithSyntaxHighlighting:!1},lp={getTextData(s){const e=s.getData(Ti.text);let t=null;const i=s.getData("vscode-editor-data");if(typeof i=="string")try{t=JSON.parse(i),t.version!==1&&(t=null)}catch{}return e.length===0&&t===null&&s.files.length>0?[Array.prototype.slice.call(s.files,0).map(o=>o.name).join(`
`),null]:[e,t]},setTextData(s,e,t,i){s.setData(Ti.text,e),typeof t=="string"&&s.setData("text/html",t),s.setData("vscode-editor-data",JSON.stringify(i))}},Af=class Af{constructor(e,t,i,n,o){this.value=e,this.selectionStart=t,this.selectionEnd=i,this.selection=n,this.newlineCountBeforeSelection=o}toString(){return`[ <${this.value}>, selectionStart: ${this.selectionStart}, selectionEnd: ${this.selectionEnd}]`}static readFromTextArea(e,t){const i=e.getValue(),n=e.getSelectionStart(),o=e.getSelectionEnd();let r;if(t){const a=i.substring(0,n),l=t.value.substring(0,t.selectionStart);a===l&&(r=t.newlineCountBeforeSelection)}return new Af(i,n,o,null,r)}collapseSelection(){return this.selectionStart===this.value.length?this:new Af(this.value,this.value.length,this.value.length,null,void 0)}isWrittenToTextArea(e,t){const i=this.value===e.getValue();return t?this.selectionStart===e.getSelectionStart()&&this.selectionEnd===e.getSelectionEnd()&&i:i}writeToTextArea(e,t,i){t.setValue(e,this.value),i&&t.setSelectionRange(e,this.selectionStart,this.selectionEnd)}deduceEditorPosition(e){if(e<=this.selectionStart){const n=this.value.substring(e,this.selectionStart);return this._finishDeduceEditorPosition(this.selection?.getStartPosition()??null,n,-1)}if(e>=this.selectionEnd){const n=this.value.substring(this.selectionEnd,e);return this._finishDeduceEditorPosition(this.selection?.getEndPosition()??null,n,1)}const t=this.value.substring(this.selectionStart,e);if(t.indexOf("…")===-1)return this._finishDeduceEditorPosition(this.selection?.getStartPosition()??null,t,1);const i=this.value.substring(e,this.selectionEnd);return this._finishDeduceEditorPosition(this.selection?.getEndPosition()??null,i,-1)}_finishDeduceEditorPosition(e,t,i){let n=0,o=-1;for(;(o=t.indexOf(`
`,o+1))!==-1;)n++;return[e,i*t.length,n]}static deduceInput(e,t,i){if(!e)return{text:"",replacePrevCharCnt:0,replaceNextCharCnt:0,positionDelta:0};const n=Math.min(Fa(e.value,t.value),e.selectionStart,t.selectionStart),o=Math.min(uh(e.value,t.value),e.value.length-e.selectionEnd,t.value.length-t.selectionEnd);e.value.substring(n,e.value.length-o);const r=t.value.substring(n,t.value.length-o),a=e.selectionStart-n,l=e.selectionEnd-n,c=t.selectionStart-n,d=t.selectionEnd-n;if(c===d){const u=e.selectionStart-n;return{text:r,replacePrevCharCnt:u,replaceNextCharCnt:0,positionDelta:0}}const h=l-a;return{text:r,replacePrevCharCnt:h,replaceNextCharCnt:0,positionDelta:0}}static deduceAndroidCompositionInput(e,t){if(!e)return{text:"",replacePrevCharCnt:0,replaceNextCharCnt:0,positionDelta:0};if(e.value===t.value)return{text:"",replacePrevCharCnt:0,replaceNextCharCnt:0,positionDelta:t.selectionEnd-e.selectionEnd};const i=Math.min(Fa(e.value,t.value),e.selectionEnd),n=Math.min(uh(e.value,t.value),e.value.length-e.selectionEnd),o=e.value.substring(i,e.value.length-n),r=t.value.substring(i,t.value.length-n);e.selectionStart-i;const a=e.selectionEnd-i;t.selectionStart-i;const l=t.selectionEnd-i;return{text:r,replacePrevCharCnt:a,replaceNextCharCnt:o.length-a,positionDelta:l-r.length}}static fromScreenReaderContentState(e){return new Af(e.value,e.selectionStart,e.selectionEnd,e.selection,e.newlineCountBeforeSelection)}};Af.EMPTY=new Af("",0,0,null,void 0);let ks=Af;const d_=(function(){if(typeof crypto.randomUUID=="function")return crypto.randomUUID.bind(crypto);const s=new Uint8Array(16),e=[];for(let t=0;t<256;t++)e.push(t.toString(16).padStart(2,"0"));return function(){crypto.getRandomValues(s),s[6]=s[6]&15|64,s[8]=s[8]&63|128;let i=0,n="";return n+=e[s[i++]],n+=e[s[i++]],n+=e[s[i++]],n+=e[s[i++]],n+="-",n+=e[s[i++]],n+=e[s[i++]],n+="-",n+=e[s[i++]],n+=e[s[i++]],n+="-",n+=e[s[i++]],n+=e[s[i++]],n+="-",n+=e[s[i++]],n+=e[s[i++]],n+=e[s[i++]],n+=e[s[i++]],n+=e[s[i++]],n+=e[s[i++]],n}})();function oee(s){return`${s}-${d_()}`}var Ibe=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},$$=function(s,e){return function(t,i){e(t,i,s)}},XE;(function(s){s.Tap="-monaco-textarea-synthetic-tap"})(XE||(XE={}));class Dbe{constructor(){this._lastTypeTextLength=0}handleCompositionUpdate(e){e=e||"";const t={text:e,replacePrevCharCnt:this._lastTypeTextLength,replaceNextCharCnt:0,positionDelta:0};return this._lastTypeTextLength=e.length,t}}let sF=class extends F{get textAreaState(){return this._textAreaState}constructor(e,t,i,n,o,r){super(),this._host=e,this._textArea=t,this._OS=i,this._browser=n,this._accessibilityService=o,this._logService=r,this._onFocus=this._register(new M),this.onFocus=this._onFocus.event,this._onBlur=this._register(new M),this.onBlur=this._onBlur.event,this._onKeyDown=this._register(new M),this.onKeyDown=this._onKeyDown.event,this._onKeyUp=this._register(new M),this.onKeyUp=this._onKeyUp.event,this._onCut=this._register(new M),this.onCut=this._onCut.event,this._onPaste=this._register(new M),this.onPaste=this._onPaste.event,this._onType=this._register(new M),this.onType=this._onType.event,this._onCompositionStart=this._register(new M),this.onCompositionStart=this._onCompositionStart.event,this._onCompositionUpdate=this._register(new M),this.onCompositionUpdate=this._onCompositionUpdate.event,this._onCompositionEnd=this._register(new M),this.onCompositionEnd=this._onCompositionEnd.event,this._onSelectionChangeRequest=this._register(new M),this.onSelectionChangeRequest=this._onSelectionChangeRequest.event,this._asyncFocusGainWriteScreenReaderContent=this._register(new vt),this._asyncTriggerCut=this._register(new Rt(()=>this._onCut.fire(),0)),this._textAreaState=ks.EMPTY,this._selectionChangeListener=null,this._accessibilityService.isScreenReaderOptimized()&&this.writeNativeTextAreaContent("ctor"),this._register(oe.runAndSubscribe(this._accessibilityService.onDidChangeScreenReaderOptimized,()=>{this._accessibilityService.isScreenReaderOptimized()&&!this._asyncFocusGainWriteScreenReaderContent.value?this._asyncFocusGainWriteScreenReaderContent.value=this._register(new Rt(()=>this.writeNativeTextAreaContent("asyncFocusGain"),0)):this._asyncFocusGainWriteScreenReaderContent.clear()})),this._hasFocus=!1,this._currentComposition=null;let a=null;this._register(this._textArea.onKeyDown(l=>{const c=new Pt(l);(c.keyCode===114||this._currentComposition&&c.keyCode===1)&&c.stopPropagation(),c.equals(9)&&c.preventDefault(),a=c,this._onKeyDown.fire(c)})),this._register(this._textArea.onKeyUp(l=>{const c=new Pt(l);this._onKeyUp.fire(c)})),this._register(this._textArea.onCompositionStart(l=>{const c=new Dbe;if(this._currentComposition){this._currentComposition=c;return}if(this._currentComposition=c,this._OS===2&&a&&a.equals(114)&&this._textAreaState.selectionStart===this._textAreaState.selectionEnd&&this._textAreaState.selectionStart>0&&this._textAreaState.value.substr(this._textAreaState.selectionStart-1,1)===l.data&&(a.code==="ArrowRight"||a.code==="ArrowLeft")){c.handleCompositionUpdate("x"),this._onCompositionStart.fire({data:l.data});return}if(this._browser.isAndroid){this._onCompositionStart.fire({data:l.data});return}this._onCompositionStart.fire({data:l.data})})),this._register(this._textArea.onCompositionUpdate(l=>{const c=this._currentComposition;if(!c)return;if(this._browser.isAndroid){const h=ks.readFromTextArea(this._textArea,this._textAreaState),u=ks.deduceAndroidCompositionInput(this._textAreaState,h);this._textAreaState=h,this._onType.fire(u),this._onCompositionUpdate.fire(l);return}const d=c.handleCompositionUpdate(l.data);this._textAreaState=ks.readFromTextArea(this._textArea,this._textAreaState),this._onType.fire(d),this._onCompositionUpdate.fire(l)})),this._register(this._textArea.onCompositionEnd(l=>{const c=this._currentComposition;if(!c)return;if(this._currentComposition=null,this._browser.isAndroid){const h=ks.readFromTextArea(this._textArea,this._textAreaState),u=ks.deduceAndroidCompositionInput(this._textAreaState,h);this._textAreaState=h,this._onType.fire(u),this._onCompositionEnd.fire();return}const d=c.handleCompositionUpdate(l.data);this._textAreaState=ks.readFromTextArea(this._textArea,this._textAreaState),this._onType.fire(d),this._onCompositionEnd.fire()})),this._register(this._textArea.onInput(l=>{if(this._textArea.setIgnoreSelectionChangeTime("received input event"),this._currentComposition)return;const c=ks.readFromTextArea(this._textArea,this._textAreaState),d=ks.deduceInput(this._textAreaState,c,this._OS===2);d.replacePrevCharCnt===0&&d.text.length===1&&(sn(d.text.charCodeAt(0))||d.text.charCodeAt(0)===127)||(this._textAreaState=c,(d.text!==""||d.replacePrevCharCnt!==0||d.replaceNextCharCnt!==0||d.positionDelta!==0)&&(l.inputType==="insertFromPaste"?this._onPaste.fire({text:d.text,metadata:xc.INSTANCE.get(d.text)}):this._onType.fire(d)))})),this._register(this._textArea.onCut(l=>{this._logService.trace("TextAreaInput#onCut",l),this._textArea.setIgnoreSelectionChangeTime("received cut event"),this._ensureClipboardGetsEditorSelection(l),this._asyncTriggerCut.schedule()})),this._register(this._textArea.onCopy(l=>{this._logService.trace("TextAreaInput#onCopy",l),this._ensureClipboardGetsEditorSelection(l)})),this._register(this._textArea.onPaste(l=>{if(this._logService.trace("TextAreaInput#onPaste",l),this._textArea.setIgnoreSelectionChangeTime("received paste event"),l.preventDefault(),!l.clipboardData)return;let[c,d]=lp.getTextData(l.clipboardData);this._logService.trace("TextAreaInput#onPaste with id : ",d?.id," with text.length: ",c.length),c&&(d=d||xc.INSTANCE.get(c),this._logService.trace("TextAreaInput#onPaste (before onPaste)"),this._onPaste.fire({text:c,metadata:d}))})),this._register(this._textArea.onFocus(()=>{const l=this._hasFocus;this._setHasFocus(!0),this._accessibilityService.isScreenReaderOptimized()&&this._browser.isSafari&&!l&&this._hasFocus&&(this._asyncFocusGainWriteScreenReaderContent.value||(this._asyncFocusGainWriteScreenReaderContent.value=new Rt(()=>this.writeNativeTextAreaContent("asyncFocusGain"),0)),this._asyncFocusGainWriteScreenReaderContent.value.schedule())})),this._register(this._textArea.onBlur(()=>{this._currentComposition&&(this._currentComposition=null,this.writeNativeTextAreaContent("blurWithoutCompositionEnd"),this._onCompositionEnd.fire()),this._setHasFocus(!1)})),this._register(this._textArea.onSyntheticTap(()=>{this._browser.isAndroid&&this._currentComposition&&(this._currentComposition=null,this.writeNativeTextAreaContent("tapWithoutCompositionEnd"),this._onCompositionEnd.fire())}))}_installSelectionChangeListener(){let e=0;return W(this._textArea.ownerDocument,"selectionchange",t=>{if(pu.onSelectionChange(),!this._hasFocus||this._currentComposition||!this._browser.isChrome)return;const i=Date.now(),n=i-e;if(e=i,n<5)return;const o=i-this._textArea.getIgnoreSelectionChangeTime();if(this._textArea.resetSelectionChangeTime(),o<100||!this._textAreaState.selection)return;const r=this._textArea.getValue();if(this._textAreaState.value!==r)return;const a=this._textArea.getSelectionStart(),l=this._textArea.getSelectionEnd();if(this._textAreaState.selectionStart===a&&this._textAreaState.selectionEnd===l)return;const c=this._textAreaState.deduceEditorPosition(a),d=this._host.deduceModelPosition(c[0],c[1],c[2]),h=this._textAreaState.deduceEditorPosition(l),u=this._host.deduceModelPosition(h[0],h[1],h[2]),g=new he(d.lineNumber,d.column,u.lineNumber,u.column);this._onSelectionChangeRequest.fire(g)})}dispose(){super.dispose(),this._selectionChangeListener&&(this._selectionChangeListener.dispose(),this._selectionChangeListener=null)}focusTextArea(){this._setHasFocus(!0),this.refreshFocusState()}isFocused(){return this._hasFocus}refreshFocusState(){this._setHasFocus(this._textArea.hasFocus())}_setHasFocus(e){this._hasFocus!==e&&(this._hasFocus=e,this._selectionChangeListener&&(this._selectionChangeListener.dispose(),this._selectionChangeListener=null),this._hasFocus&&(this._selectionChangeListener=this._installSelectionChangeListener()),this._hasFocus&&this.writeNativeTextAreaContent("focusgain"),this._hasFocus?this._onFocus.fire():this._onBlur.fire())}_setAndWriteTextAreaState(e,t){this._hasFocus||(t=t.collapseSelection()),t.isWrittenToTextArea(this._textArea,this._hasFocus)||this._logService.trace(`writeTextAreaState(reason: ${e})`),t.writeToTextArea(e,this._textArea,this._hasFocus),this._textAreaState=t}writeNativeTextAreaContent(e){!this._accessibilityService.isScreenReaderOptimized()&&e==="render"||this._currentComposition||this._setAndWriteTextAreaState(e,this._host.getScreenReaderContent())}_ensureClipboardGetsEditorSelection(e){const t=this._host.getDataToCopy();let i;this._logService.getLevel()===pn.Trace&&(i=d_());const n={version:1,id:i,isFromEmptySelection:t.isFromEmptySelection,multicursorText:t.multicursorText,mode:t.mode};xc.INSTANCE.set(this._browser.isFirefox?t.text.replace(/\r\n/g,`
`):t.text,n),e.preventDefault(),e.clipboardData&&lp.setTextData(e.clipboardData,t.text,t.html,n),this._logService.trace("TextAreaEditContextInput#_ensureClipboardGetsEditorSelection with id : ",i," with text.length: ",t.text.length)}};sF=Ibe([$$(4,Wn),$$(5,jt)],sF);class Tbe extends F{get ownerDocument(){return this._actual.ownerDocument}constructor(e){super(),this._actual=e,this._onSyntheticTap=this._register(new M),this.onSyntheticTap=this._onSyntheticTap.event,this._ignoreSelectionChangeTime=0,this.onKeyDown=this._register(new Lt(this._actual,"keydown")).event,this.onKeyPress=this._register(new Lt(this._actual,"keypress")).event,this.onKeyUp=this._register(new Lt(this._actual,"keyup")).event,this.onCompositionStart=this._register(new Lt(this._actual,"compositionstart")).event,this.onCompositionUpdate=this._register(new Lt(this._actual,"compositionupdate")).event,this.onCompositionEnd=this._register(new Lt(this._actual,"compositionend")).event,this.onBeforeInput=this._register(new Lt(this._actual,"beforeinput")).event,this.onInput=this._register(new Lt(this._actual,"input")).event,this.onCut=this._register(new Lt(this._actual,"cut")).event,this.onCopy=this._register(new Lt(this._actual,"copy")).event,this.onPaste=this._register(new Lt(this._actual,"paste")).event,this.onFocus=this._register(new Lt(this._actual,"focus")).event,this.onBlur=this._register(new Lt(this._actual,"blur")).event,this._register(this.onKeyDown(()=>pu.onKeyDown())),this._register(this.onBeforeInput(()=>pu.onBeforeInput())),this._register(this.onInput(()=>pu.onInput())),this._register(this.onKeyUp(()=>pu.onKeyUp())),this._register(W(this._actual,XE.Tap,()=>this._onSyntheticTap.fire()))}hasFocus(){const e=cg(this._actual);return e?e.activeElement===this._actual:this._actual.isConnected?ln()===this._actual:!1}setIgnoreSelectionChangeTime(e){this._ignoreSelectionChangeTime=Date.now()}getIgnoreSelectionChangeTime(){return this._ignoreSelectionChangeTime}resetSelectionChangeTime(){this._ignoreSelectionChangeTime=0}getValue(){return this._actual.value}setValue(e,t){const i=this._actual;i.value!==t&&(this.setIgnoreSelectionChangeTime("setValue"),i.value=t)}getSelectionStart(){return this._actual.selectionDirection==="backward"?this._actual.selectionEnd:this._actual.selectionStart}getSelectionEnd(){return this._actual.selectionDirection==="backward"?this._actual.selectionStart:this._actual.selectionEnd}setSelectionRange(e,t,i){const n=this._actual;let o=null;const r=cg(n);r?o=r.activeElement:o=ln();const a=be(o),l=o===n,c=n.selectionStart,d=n.selectionEnd;if(l&&c===t&&d===i){wo&&a.parent!==a&&n.focus();return}if(l){this.setIgnoreSelectionChangeTime("setSelectionRange"),n.setSelectionRange(t,i),wo&&a.parent!==a&&n.focus();return}try{const h=cue(n);this.setIgnoreSelectionChangeTime("setSelectionRange"),n.focus(),n.setSelectionRange(t,i),due(n,h)}catch{}}}class Nbe extends $W{constructor(e,t,i){super(e,t,i),this._register(_s.addTarget(this.viewHelper.linesContentDomNode)),this._register(W(this.viewHelper.linesContentDomNode,qt.Tap,o=>this.onTap(o))),this._register(W(this.viewHelper.linesContentDomNode,qt.Change,o=>this.onChange(o))),this._register(W(this.viewHelper.linesContentDomNode,qt.Contextmenu,o=>this._onContextMenu(new bh(o,!1,this.viewHelper.viewDomNode),!1))),this._lastPointerType="mouse",this._register(W(this.viewHelper.linesContentDomNode,"pointerdown",o=>{const r=o.pointerType;if(r==="mouse"){this._lastPointerType="mouse";return}else r==="touch"?this._lastPointerType="touch":this._lastPointerType="pen"}));const n=new F_e(this.viewHelper.viewDomNode);this._register(n.onPointerMove(this.viewHelper.viewDomNode,o=>this._onMouseMove(o))),this._register(n.onPointerUp(this.viewHelper.viewDomNode,o=>this._onMouseUp(o))),this._register(n.onPointerLeave(this.viewHelper.viewDomNode,o=>this._onMouseLeave(o))),this._register(n.onPointerDown(this.viewHelper.viewDomNode,(o,r)=>this._onMouseDown(o,r)))}onTap(e){!e.initialTarget||!this.viewHelper.linesContentDomNode.contains(e.initialTarget)||(e.preventDefault(),this.viewHelper.focusTextArea(),this._dispatchGesture(e,!1))}onChange(e){this._lastPointerType==="touch"&&this._context.viewModel.viewLayout.deltaScrollNow(-e.translationX,-e.translationY),this._lastPointerType==="pen"&&this._dispatchGesture(e,!0)}_dispatchGesture(e,t){const i=this._createMouseTarget(new bh(e,!1,this.viewHelper.viewDomNode),!1);i.position&&this.viewController.dispatchMouse({position:i.position,mouseColumn:i.position.column,startedOnLineNumbers:!1,revealType:1,mouseDownCount:e.tapCount,inSelectionMode:t,altKey:!1,ctrlKey:!1,metaKey:!1,shiftKey:!1,leftButton:!1,middleButton:!1,onInjectedText:i.type===6&&i.detail.injectedText!==null})}_onMouseDown(e,t){e.browserEvent.pointerType!=="touch"&&super._onMouseDown(e,t)}}class Rbe extends $W{constructor(e,t,i){super(e,t,i),this._register(_s.addTarget(this.viewHelper.linesContentDomNode)),this._register(W(this.viewHelper.linesContentDomNode,qt.Tap,n=>this.onTap(n))),this._register(W(this.viewHelper.linesContentDomNode,qt.Change,n=>this.onChange(n))),this._register(W(this.viewHelper.linesContentDomNode,qt.Contextmenu,n=>this._onContextMenu(new bh(n,!1,this.viewHelper.viewDomNode),!1)))}onTap(e){e.preventDefault(),this.viewHelper.focusTextArea();const t=this._createMouseTarget(new bh(e,!1,this.viewHelper.viewDomNode),!1);if(t.position){const i=document.createEvent("CustomEvent");i.initEvent(XE.Tap,!1,!0),this.viewHelper.dispatchTextAreaEvent(i),this.viewController.moveTo(t.position,1)}}onChange(e){this._context.viewModel.viewLayout.deltaScrollNow(-e.translationX,-e.translationY)}}class Abe extends F{constructor(e,t,i){super(),(na||ede&&HX)&&z0.pointerEvents?this.handler=this._register(new Nbe(e,t,i)):Nt.TouchEvent?this.handler=this._register(new Rbe(e,t,i)):this.handler=this._register(new $W(e,t,i))}getTargetAtClientPoint(e,t){return this.handler.getTargetAtClientPoint(e,t)}}class Mbe{constructor(e,t,i,n){this.configuration=e,this.viewModel=t,this.userInputEvents=i,this.commandDelegate=n}paste(e,t,i,n){this.commandDelegate.paste(e,t,i,n)}type(e){this.commandDelegate.type(e)}compositionType(e,t,i,n){this.commandDelegate.compositionType(e,t,i,n)}compositionStart(){this.commandDelegate.startComposition()}compositionEnd(){this.commandDelegate.endComposition()}cut(){this.commandDelegate.cut()}setSelection(e){yn.SetSelection.runCoreEditorCommand(this.viewModel,{source:"keyboard",selection:e})}_validateViewColumn(e){const t=this.viewModel.getLineMinColumn(e.lineNumber);return e.column<t?new A(e.lineNumber,t):e}_hasMulticursorModifier(e){switch(this.configuration.options.get(86)){case"altKey":return e.altKey;case"ctrlKey":return e.ctrlKey;case"metaKey":return e.metaKey;default:return!1}}_hasNonMulticursorModifier(e){switch(this.configuration.options.get(86)){case"altKey":return e.ctrlKey||e.metaKey;case"ctrlKey":return e.altKey||e.metaKey;case"metaKey":return e.ctrlKey||e.altKey;default:return!1}}dispatchMouse(e){const t=this.configuration.options,i=vo&&t.get(121),n=t.get(28),o=t.get(171);e.middleButton&&!i?o||this._columnSelect(e.position,e.mouseColumn,e.inSelectionMode):e.startedOnLineNumbers?this._hasMulticursorModifier(e)?e.inSelectionMode?this._lastCursorLineSelect(e.position,e.revealType):this._createCursor(e.position,!0):e.inSelectionMode?this._lineSelectDrag(e.position,e.revealType):this._lineSelect(e.position,e.revealType):e.mouseDownCount>=4?this._selectAll():e.mouseDownCount===3?this._hasMulticursorModifier(e)?e.inSelectionMode?this._lastCursorLineSelectDrag(e.position,e.revealType):this._lastCursorLineSelect(e.position,e.revealType):e.inSelectionMode?this._lineSelectDrag(e.position,e.revealType):this._lineSelect(e.position,e.revealType):e.mouseDownCount===2?e.onInjectedText||(this._hasMulticursorModifier(e)?this._lastCursorWordSelect(e.position,e.revealType):e.inSelectionMode?this._wordSelectDrag(e.position,e.revealType):this._wordSelect(e.position,e.revealType)):this._hasMulticursorModifier(e)?this._hasNonMulticursorModifier(e)||(e.shiftKey?this._columnSelect(e.position,e.mouseColumn,!0):e.inSelectionMode?this._lastCursorMoveToSelect(e.position,e.revealType):this._createCursor(e.position,!1)):e.inSelectionMode?e.altKey?this._columnSelect(e.position,e.mouseColumn,!0):n?this._columnSelect(e.position,e.mouseColumn,!0):this._moveToSelect(e.position,e.revealType):this.moveTo(e.position,e.revealType)}_usualArgs(e,t){return e=this._validateViewColumn(e),{source:"mouse",position:this._convertViewToModelPosition(e),viewPosition:e,revealType:t}}moveTo(e,t){yn.MoveTo.runCoreEditorCommand(this.viewModel,this._usualArgs(e,t))}_moveToSelect(e,t){yn.MoveToSelect.runCoreEditorCommand(this.viewModel,this._usualArgs(e,t))}_columnSelect(e,t,i){e=this._validateViewColumn(e),yn.ColumnSelect.runCoreEditorCommand(this.viewModel,{source:"mouse",position:this._convertViewToModelPosition(e),viewPosition:e,mouseColumn:t,doColumnSelect:i})}_createCursor(e,t){e=this._validateViewColumn(e),yn.CreateCursor.runCoreEditorCommand(this.viewModel,{source:"mouse",position:this._convertViewToModelPosition(e),viewPosition:e,wholeLine:t})}_lastCursorMoveToSelect(e,t){yn.LastCursorMoveToSelect.runCoreEditorCommand(this.viewModel,this._usualArgs(e,t))}_wordSelect(e,t){yn.WordSelect.runCoreEditorCommand(this.viewModel,this._usualArgs(e,t))}_wordSelectDrag(e,t){yn.WordSelectDrag.runCoreEditorCommand(this.viewModel,this._usualArgs(e,t))}_lastCursorWordSelect(e,t){yn.LastCursorWordSelect.runCoreEditorCommand(this.viewModel,this._usualArgs(e,t))}_lineSelect(e,t){yn.LineSelect.runCoreEditorCommand(this.viewModel,this._usualArgs(e,t))}_lineSelectDrag(e,t){yn.LineSelectDrag.runCoreEditorCommand(this.viewModel,this._usualArgs(e,t))}_lastCursorLineSelect(e,t){yn.LastCursorLineSelect.runCoreEditorCommand(this.viewModel,this._usualArgs(e,t))}_lastCursorLineSelectDrag(e,t){yn.LastCursorLineSelectDrag.runCoreEditorCommand(this.viewModel,this._usualArgs(e,t))}_selectAll(){yn.SelectAll.runCoreEditorCommand(this.viewModel,{source:"mouse"})}_convertViewToModelPosition(e){return this.viewModel.coordinatesConverter.convertViewPositionToModelPosition(e)}emitKeyDown(e){this.userInputEvents.emitKeyDown(e)}emitKeyUp(e){this.userInputEvents.emitKeyUp(e)}emitContextMenu(e){this.userInputEvents.emitContextMenu(e)}emitMouseMove(e){this.userInputEvents.emitMouseMove(e)}emitMouseLeave(e){this.userInputEvents.emitMouseLeave(e)}emitMouseUp(e){this.userInputEvents.emitMouseUp(e)}emitMouseDown(e){this.userInputEvents.emitMouseDown(e)}emitMouseDrag(e){this.userInputEvents.emitMouseDrag(e)}emitMouseDrop(e){this.userInputEvents.emitMouseDrop(e)}emitMouseDropCanceled(){this.userInputEvents.emitMouseDropCanceled()}emitMouseWheel(e){this.userInputEvents.emitMouseWheel(e)}}function $c(s,e){const t=M6();if(t?.createTrustedTypesPolicy)try{return t.createTrustedTypesPolicy(s,e)}catch(i){Oe(i);return}try{return globalThis.trustedTypes?.createPolicy(s,e)}catch(i){Oe(i);return}}class ree{constructor(e){this._lineFactory=e,this._set(1,[])}flush(){this._set(1,[])}_set(e,t){this._lines=t,this._rendLineNumberStart=e}_get(){return{rendLineNumberStart:this._rendLineNumberStart,lines:this._lines}}getStartLineNumber(){return this._rendLineNumberStart}getEndLineNumber(){return this._rendLineNumberStart+this._lines.length-1}getCount(){return this._lines.length}getLine(e){const t=e-this._rendLineNumberStart;if(t<0||t>=this._lines.length)throw new Ce("Illegal value for lineNumber");return this._lines[t]}onLinesDeleted(e,t){if(this.getCount()===0)return null;const i=this.getStartLineNumber(),n=this.getEndLineNumber();if(t<i){const l=t-e+1;return this._rendLineNumberStart-=l,null}if(e>n)return null;let o=0,r=0;for(let l=i;l<=n;l++){const c=l-this._rendLineNumberStart;e<=l&&l<=t&&(r===0?(o=c,r=1):r++)}if(e<i){let l=0;t<i?l=t-e+1:l=i-e,this._rendLineNumberStart-=l}return this._lines.splice(o,r)}onLinesChanged(e,t){const i=e+t-1;if(this.getCount()===0)return!1;const n=this.getStartLineNumber(),o=this.getEndLineNumber();let r=!1;for(let a=e;a<=i;a++)a>=n&&a<=o&&(this._lines[a-this._rendLineNumberStart].onContentChanged(),r=!0);return r}onLinesInserted(e,t){if(this.getCount()===0)return null;const i=t-e+1,n=this.getStartLineNumber(),o=this.getEndLineNumber();if(e<=n)return this._rendLineNumberStart+=i,null;if(e>o)return null;if(i+e>o)return this._lines.splice(e-this._rendLineNumberStart,o-e+1);const r=[];for(let h=0;h<i;h++)r[h]=this._lineFactory.createLine();const a=e-this._rendLineNumberStart,l=this._lines.slice(0,a),c=this._lines.slice(a,this._lines.length-i),d=this._lines.slice(this._lines.length-i,this._lines.length);return this._lines=l.concat(r).concat(c),d}onTokensChanged(e){if(this.getCount()===0)return!1;const t=this.getStartLineNumber(),i=this.getEndLineNumber();let n=!1;for(let o=0,r=e.length;o<r;o++){const a=e[o];if(a.toLineNumber<t||a.fromLineNumber>i)continue;const l=Math.max(t,a.fromLineNumber),c=Math.min(i,a.toLineNumber);for(let d=l;d<=c;d++){const h=d-this._rendLineNumberStart;this._lines[h].onTokensChanged(),n=!0}}return n}}class aee{constructor(e,t){this._viewContext=e,this._lineFactory=t,this.domNode=this._createDomNode(),this._linesCollection=new ree(this._lineFactory)}_createDomNode(){const e=wt(document.createElement("div"));return e.setClassName("view-layer"),e.setPosition("absolute"),e.domNode.setAttribute("role","presentation"),e.domNode.setAttribute("aria-hidden","true"),e}onConfigurationChanged(e){return!!e.hasChanged(165)}onFlushed(e,t){if(t){const i=this._linesCollection.getStartLineNumber(),n=this._linesCollection.getEndLineNumber();for(let o=i;o<=n;o++)this._linesCollection.getLine(o).getDomNode()?.remove()}return this._linesCollection.flush(),!0}onLinesChanged(e){return this._linesCollection.onLinesChanged(e.fromLineNumber,e.count)}onLinesDeleted(e){const t=this._linesCollection.onLinesDeleted(e.fromLineNumber,e.toLineNumber);if(t)for(let i=0,n=t.length;i<n;i++)t[i].getDomNode()?.remove();return!0}onLinesInserted(e){const t=this._linesCollection.onLinesInserted(e.fromLineNumber,e.toLineNumber);if(t)for(let i=0,n=t.length;i<n;i++)t[i].getDomNode()?.remove();return!0}onScrollChanged(e){return e.scrollTopChanged}onTokensChanged(e){return this._linesCollection.onTokensChanged(e.ranges)}onZonesChanged(e){return!0}getStartLineNumber(){return this._linesCollection.getStartLineNumber()}getEndLineNumber(){return this._linesCollection.getEndLineNumber()}getVisibleLine(e){return this._linesCollection.getLine(e)}renderLines(e){const t=this._linesCollection._get(),i=new oF(this.domNode.domNode,this._lineFactory,e,this._viewContext),n={rendLineNumberStart:t.rendLineNumberStart,lines:t.lines,linesLength:t.lines.length},o=i.render(n,e.startLineNumber,e.endLineNumber,e.relativeVerticalOffset);this._linesCollection._set(o.rendLineNumberStart,o.lines)}}const kd=class kd{constructor(e,t,i,n){this._domNode=e,this._lineFactory=t,this._viewportData=i,this._viewContext=n}render(e,t,i,n){const o={rendLineNumberStart:e.rendLineNumberStart,lines:e.lines.slice(0),linesLength:e.linesLength};if(o.rendLineNumberStart+o.linesLength-1<t||i<o.rendLineNumberStart){o.rendLineNumberStart=t,o.linesLength=i-t+1,o.lines=[];for(let r=t;r<=i;r++)o.lines[r-t]=this._lineFactory.createLine();return this._finishRendering(o,!0,n),o}if(this._renderUntouchedLines(o,Math.max(t-o.rendLineNumberStart,0),Math.min(i-o.rendLineNumberStart,o.linesLength-1),n,t),o.rendLineNumberStart>t){const r=t,a=Math.min(i,o.rendLineNumberStart-1);r<=a&&(this._insertLinesBefore(o,r,a,n,t),o.linesLength+=a-r+1)}else if(o.rendLineNumberStart<t){const r=Math.min(o.linesLength,t-o.rendLineNumberStart);r>0&&(this._removeLinesBefore(o,r),o.linesLength-=r)}if(o.rendLineNumberStart=t,o.rendLineNumberStart+o.linesLength-1<i){const r=o.rendLineNumberStart+o.linesLength,a=i;r<=a&&(this._insertLinesAfter(o,r,a,n,t),o.linesLength+=a-r+1)}else if(o.rendLineNumberStart+o.linesLength-1>i){const r=Math.max(0,i-o.rendLineNumberStart+1),l=o.linesLength-1-r+1;l>0&&(this._removeLinesAfter(o,l),o.linesLength-=l)}return this._finishRendering(o,!1,n),o}_renderUntouchedLines(e,t,i,n,o){const r=e.rendLineNumberStart,a=e.lines;for(let l=t;l<=i;l++){const c=r+l;a[l].layoutLine(c,n[c-o],this._lineHeightForLineNumber(c))}}_insertLinesBefore(e,t,i,n,o){const r=[];let a=0;for(let l=t;l<=i;l++)r[a++]=this._lineFactory.createLine();e.lines=r.concat(e.lines)}_removeLinesBefore(e,t){for(let i=0;i<t;i++)e.lines[i].getDomNode()?.remove();e.lines.splice(0,t)}_insertLinesAfter(e,t,i,n,o){const r=[];let a=0;for(let l=t;l<=i;l++)r[a++]=this._lineFactory.createLine();e.lines=e.lines.concat(r)}_removeLinesAfter(e,t){const i=e.linesLength-t;for(let n=0;n<t;n++)e.lines[i+n].getDomNode()?.remove();e.lines.splice(i,t)}_finishRenderingNewLines(e,t,i,n){kd._ttPolicy&&(i=kd._ttPolicy.createHTML(i));const o=this._domNode.lastChild;t||!o?this._domNode.innerHTML=i:o.insertAdjacentHTML("afterend",i);let r=this._domNode.lastChild;for(let a=e.linesLength-1;a>=0;a--){const l=e.lines[a];n[a]&&(l.setDomNode(r),r=r.previousSibling)}}_finishRenderingInvalidLines(e,t,i){const n=document.createElement("div");kd._ttPolicy&&(t=kd._ttPolicy.createHTML(t)),n.innerHTML=t;for(let o=0;o<e.linesLength;o++){const r=e.lines[o];if(i[o]){const a=n.firstChild;r.getDomNode().replaceWith(a),r.setDomNode(a)}}}_finishRendering(e,t,i){const n=kd._sb,o=e.linesLength,r=e.lines,a=e.rendLineNumberStart,l=[];{n.reset();let c=!1;for(let d=0;d<o;d++){const h=r[d];if(l[d]=!1,h.getDomNode())continue;const g=d+a;h.renderLine(g,i[d],this._lineHeightForLineNumber(g),this._viewportData,n)&&(l[d]=!0,c=!0)}c&&this._finishRenderingNewLines(e,t,n.build(),l)}{n.reset();let c=!1;const d=[];for(let h=0;h<o;h++){const u=r[h];if(d[h]=!1,l[h])continue;const g=h+a;u.renderLine(g,i[h],this._lineHeightForLineNumber(g),this._viewportData,n)&&(d[h]=!0,c=!0)}c&&this._finishRenderingInvalidLines(e,n.build(),d)}}_lineHeightForLineNumber(e){return this._viewContext.viewLayout.getLineHeightForLineNumber(e)}};kd._ttPolicy=$c("editorViewLayer",{createHTML:e=>e}),kd._sb=new Eg(1e5);let oF=kd;class lee extends no{constructor(e){super(e),this._dynamicOverlays=[],this._isFocused=!1,this._visibleLines=new aee(this._context,{createLine:()=>new Obe(this._dynamicOverlays)}),this.domNode=this._visibleLines.domNode;const i=this._context.configuration.options.get(59);In(this.domNode,i),this.domNode.setClassName("view-overlays")}shouldRender(){if(super.shouldRender())return!0;for(let e=0,t=this._dynamicOverlays.length;e<t;e++)if(this._dynamicOverlays[e].shouldRender())return!0;return!1}dispose(){super.dispose();for(let e=0,t=this._dynamicOverlays.length;e<t;e++)this._dynamicOverlays[e].dispose();this._dynamicOverlays=[]}getDomNode(){return this.domNode}addDynamicOverlay(e){this._dynamicOverlays.push(e)}onConfigurationChanged(e){this._visibleLines.onConfigurationChanged(e);const i=this._context.configuration.options.get(59);return In(this.domNode,i),!0}onFlushed(e){return this._visibleLines.onFlushed(e)}onFocusChanged(e){return this._isFocused=e.isFocused,!0}onLinesChanged(e){return this._visibleLines.onLinesChanged(e)}onLinesDeleted(e){return this._visibleLines.onLinesDeleted(e)}onLinesInserted(e){return this._visibleLines.onLinesInserted(e)}onScrollChanged(e){return this._visibleLines.onScrollChanged(e)||!0}onTokensChanged(e){return this._visibleLines.onTokensChanged(e)}onZonesChanged(e){return this._visibleLines.onZonesChanged(e)}prepareRender(e){const t=this._dynamicOverlays.filter(i=>i.shouldRender());for(let i=0,n=t.length;i<n;i++){const o=t[i];o.prepareRender(e),o.onDidRender()}}render(e){this._viewOverlaysRender(e),this.domNode.toggleClassName("focused",this._isFocused)}_viewOverlaysRender(e){this._visibleLines.renderLines(e.viewportData)}}class Obe{constructor(e){this._dynamicOverlays=e,this._domNode=null,this._renderedContent=null}getDomNode(){return this._domNode?this._domNode.domNode:null}setDomNode(e){this._domNode=wt(e)}onContentChanged(){}onTokensChanged(){}renderLine(e,t,i,n,o){let r="";for(let a=0,l=this._dynamicOverlays.length;a<l;a++){const c=this._dynamicOverlays[a];r+=c.render(n.startLineNumber,e)}return this._renderedContent===r?!1:(this._renderedContent=r,o.appendString('<div style="top:'),o.appendString(String(t)),o.appendString("px;height:"),o.appendString(String(i)),o.appendString("px;line-height:"),o.appendString(String(i)),o.appendString('px;">'),o.appendString(r),o.appendString("</div>"),!0)}layoutLine(e,t,i){this._domNode&&(this._domNode.setTop(t),this._domNode.setHeight(i),this._domNode.setLineHeight(i))}}class Pbe extends lee{constructor(e){super(e);const i=this._context.configuration.options.get(165);this._contentWidth=i.contentWidth,this.domNode.setHeight(0)}onConfigurationChanged(e){const i=this._context.configuration.options.get(165);return this._contentWidth=i.contentWidth,super.onConfigurationChanged(e)||!0}onScrollChanged(e){return super.onScrollChanged(e)||e.scrollWidthChanged}_viewOverlaysRender(e){super._viewOverlaysRender(e),this.domNode.setWidth(Math.max(e.scrollWidth,this._contentWidth))}}class Fbe extends lee{constructor(e){super(e);const t=this._context.configuration.options,i=t.get(165);this._contentLeft=i.contentLeft,this.domNode.setClassName("margin-view-overlays"),this.domNode.setWidth(1),In(this.domNode,t.get(59))}onConfigurationChanged(e){const t=this._context.configuration.options;In(this.domNode,t.get(59));const i=t.get(165);return this._contentLeft=i.contentLeft,super.onConfigurationChanged(e)||!0}onScrollChanged(e){return super.onScrollChanged(e)||e.scrollHeightChanged}_viewOverlaysRender(e){super._viewOverlaysRender(e);const t=Math.min(e.scrollHeight,1e6);this.domNode.setHeight(t),this.domNode.setWidth(this._contentLeft)}}class dA{constructor(e){this.onKeyDown=null,this.onKeyUp=null,this.onContextMenu=null,this.onMouseMove=null,this.onMouseLeave=null,this.onMouseDown=null,this.onMouseUp=null,this.onMouseDrag=null,this.onMouseDrop=null,this.onMouseDropCanceled=null,this.onMouseWheel=null,this._coordinatesConverter=e}emitKeyDown(e){this.onKeyDown?.(e)}emitKeyUp(e){this.onKeyUp?.(e)}emitContextMenu(e){this.onContextMenu?.(this._convertViewToModelMouseEvent(e))}emitMouseMove(e){this.onMouseMove?.(this._convertViewToModelMouseEvent(e))}emitMouseLeave(e){this.onMouseLeave?.(this._convertViewToModelMouseEvent(e))}emitMouseDown(e){this.onMouseDown?.(this._convertViewToModelMouseEvent(e))}emitMouseUp(e){this.onMouseUp?.(this._convertViewToModelMouseEvent(e))}emitMouseDrag(e){this.onMouseDrag?.(this._convertViewToModelMouseEvent(e))}emitMouseDrop(e){this.onMouseDrop?.(this._convertViewToModelMouseEvent(e))}emitMouseDropCanceled(){this.onMouseDropCanceled?.()}emitMouseWheel(e){this.onMouseWheel?.(e)}_convertViewToModelMouseEvent(e){return e.target?{event:e.event,target:this._convertViewToModelMouseTarget(e.target)}:e}_convertViewToModelMouseTarget(e){return dA.convertViewToModelMouseTarget(e,this._coordinatesConverter)}static convertViewToModelMouseTarget(e,t){const i={...e};return i.position&&(i.position=t.convertViewPositionToModelPosition(i.position)),i.range&&(i.range=t.convertViewRangeToModelRange(i.range)),(i.type===5||i.type===8)&&(i.detail=this.convertViewToModelViewZoneData(i.detail,t)),i}static convertViewToModelViewZoneData(e,t){return{viewZoneId:e.viewZoneId,positionBefore:e.positionBefore?t.convertViewPositionToModelPosition(e.positionBefore):e.positionBefore,positionAfter:e.positionAfter?t.convertViewPositionToModelPosition(e.positionAfter):e.positionAfter,position:t.convertViewPositionToModelPosition(e.position),afterLineNumber:t.convertViewPositionToModelPosition(new A(e.afterLineNumber,1)).lineNumber}}}class Bbe extends no{constructor(e){super(e),this.blocks=[],this.contentWidth=-1,this.contentLeft=0,this.domNode=wt(document.createElement("div")),this.domNode.setAttribute("role","presentation"),this.domNode.setAttribute("aria-hidden","true"),this.domNode.setClassName("blockDecorations-container"),this.update()}update(){let e=!1;const i=this._context.configuration.options.get(165),n=i.contentWidth-i.verticalScrollbarWidth;this.contentWidth!==n&&(this.contentWidth=n,e=!0);const o=i.contentLeft;return this.contentLeft!==o&&(this.contentLeft=o,e=!0),e}dispose(){super.dispose()}onConfigurationChanged(e){return this.update()}onScrollChanged(e){return e.scrollTopChanged||e.scrollLeftChanged}onDecorationsChanged(e){return!0}onZonesChanged(e){return!0}prepareRender(e){}render(e){let t=0;const i=e.getDecorationsInViewport();for(const n of i){if(!n.options.blockClassName)continue;let o=this.blocks[t];o||(o=this.blocks[t]=wt(document.createElement("div")),this.domNode.appendChild(o));let r,a;n.options.blockIsAfterEnd?(r=e.getVerticalOffsetAfterLineNumber(n.range.endLineNumber,!1),a=e.getVerticalOffsetAfterLineNumber(n.range.endLineNumber,!0)):(r=e.getVerticalOffsetForLineNumber(n.range.startLineNumber,!0),a=n.range.isEmpty()&&!n.options.blockDoesNotCollapse?e.getVerticalOffsetForLineNumber(n.range.startLineNumber,!1):e.getVerticalOffsetAfterLineNumber(n.range.endLineNumber,!0));const[l,c,d,h]=n.options.blockPadding??[0,0,0,0];o.setClassName("blockDecorations-block "+n.options.blockClassName),o.setLeft(this.contentLeft-h),o.setWidth(this.contentWidth+h+c),o.setTop(r-e.scrollTop-l),o.setHeight(a-r+l+d),t++}for(let n=t;n<this.blocks.length;n++)this.blocks[n].domNode.remove();this.blocks.length=t}}class Wbe extends no{constructor(e,t){super(e),this._viewDomNode=t,this._widgets={},this.domNode=wt(document.createElement("div")),kl.write(this.domNode,1),this.domNode.setClassName("contentWidgets"),this.domNode.setPosition("absolute"),this.domNode.setTop(0),this.overflowingContentWidgetsDomNode=wt(document.createElement("div")),kl.write(this.overflowingContentWidgetsDomNode,2),this.overflowingContentWidgetsDomNode.setClassName("overflowingContentWidgets")}dispose(){super.dispose(),this._widgets={}}onConfigurationChanged(e){const t=Object.keys(this._widgets);for(const i of t)this._widgets[i].onConfigurationChanged(e);return!0}onDecorationsChanged(e){return!0}onFlushed(e){return!0}onLineMappingChanged(e){return this._updateAnchorsViewPositions(),!0}onLinesChanged(e){return this._updateAnchorsViewPositions(),!0}onLinesDeleted(e){return this._updateAnchorsViewPositions(),!0}onLinesInserted(e){return this._updateAnchorsViewPositions(),!0}onScrollChanged(e){return!0}onZonesChanged(e){return!0}_updateAnchorsViewPositions(){const e=Object.keys(this._widgets);for(const t of e)this._widgets[t].updateAnchorViewPosition()}addWidget(e){const t=new Hbe(this._context,this._viewDomNode,e);this._widgets[t.id]=t,t.allowEditorOverflow?this.overflowingContentWidgetsDomNode.appendChild(t.domNode):this.domNode.appendChild(t.domNode),this.setShouldRender()}setWidgetPosition(e,t,i,n,o){this._widgets[e.getId()].setPosition(t,i,n,o),this.setShouldRender()}removeWidget(e){const t=e.getId();if(this._widgets.hasOwnProperty(t)){const i=this._widgets[t];delete this._widgets[t];const n=i.domNode.domNode;n.remove(),n.removeAttribute("monaco-visible-content-widget"),this.setShouldRender()}}shouldSuppressMouseDownOnWidget(e){return this._widgets.hasOwnProperty(e)?this._widgets[e].suppressMouseDown:!1}onBeforeRender(e){const t=Object.keys(this._widgets);for(const i of t)this._widgets[i].onBeforeRender(e)}prepareRender(e){const t=Object.keys(this._widgets);for(const i of t)this._widgets[i].prepareRender(e)}render(e){const t=Object.keys(this._widgets);for(const i of t)this._widgets[i].render(e)}}class Hbe{constructor(e,t,i){this._primaryAnchor=new EC(null,null),this._secondaryAnchor=new EC(null,null),this._context=e,this._viewDomNode=t,this._actual=i;const n=this._context.configuration.options,o=n.get(165),r=n.get(4);this.domNode=wt(this._actual.getDomNode()),this.id=this._actual.getId(),this.allowEditorOverflow=(this._actual.allowEditorOverflow||!1)&&r,this.suppressMouseDown=this._actual.suppressMouseDown||!1,this._fixedOverflowWidgets=n.get(51),this._contentWidth=o.contentWidth,this._contentLeft=o.contentLeft,this._affinity=null,this._preference=[],this._cachedDomNodeOffsetWidth=-1,this._cachedDomNodeOffsetHeight=-1,this._maxWidth=this._getMaxWidth(),this._isVisible=!1,this._renderData=null,this.domNode.setPosition(this._fixedOverflowWidgets&&this.allowEditorOverflow?"fixed":"absolute"),this.domNode.setDisplay("none"),this.domNode.setVisibility("hidden"),this.domNode.setAttribute("widgetId",this.id),this.domNode.setMaxWidth(this._maxWidth)}onConfigurationChanged(e){const t=this._context.configuration.options;if(e.hasChanged(165)){const i=t.get(165);this._contentLeft=i.contentLeft,this._contentWidth=i.contentWidth,this._maxWidth=this._getMaxWidth()}}updateAnchorViewPosition(){this._setPosition(this._affinity,this._primaryAnchor.modelPosition,this._secondaryAnchor.modelPosition)}_setPosition(e,t,i){this._affinity=e,this._primaryAnchor=n(t,this._context.viewModel,this._affinity),this._secondaryAnchor=n(i,this._context.viewModel,this._affinity);function n(o,r,a){if(!o)return new EC(null,null);const l=r.model.validatePosition(o);if(r.coordinatesConverter.modelPositionIsVisible(l)){const c=r.coordinatesConverter.convertModelPositionToViewPosition(l,a??void 0);return new EC(o,c)}return new EC(o,null)}}_getMaxWidth(){const e=this.domNode.domNode.ownerDocument,t=e.defaultView;return this.allowEditorOverflow?t?.innerWidth||e.documentElement.offsetWidth||e.body.offsetWidth:this._contentWidth}setPosition(e,t,i,n){this._setPosition(n,e,t),this._preference=i,this._primaryAnchor.viewPosition&&this._preference&&this._preference.length>0?this.domNode.setDisplay("block"):this.domNode.setDisplay("none"),this._cachedDomNodeOffsetWidth=-1,this._cachedDomNodeOffsetHeight=-1}_layoutBoxInViewport(e,t,i,n){const o=e.top,r=o,a=e.top+e.height,l=n.viewportHeight-a,c=o-i,d=r>=i,h=a,u=l>=i;let g=e.left;return g+t>n.scrollLeft+n.viewportWidth&&(g=n.scrollLeft+n.viewportWidth-t),g<n.scrollLeft&&(g=n.scrollLeft),{fitsAbove:d,aboveTop:c,fitsBelow:u,belowTop:h,left:g}}_layoutHorizontalSegmentInPage(e,t,i,n){const a=Math.max(15,t.left-n),l=Math.min(t.left+t.width+n,e.width-15),d=this._viewDomNode.domNode.ownerDocument.defaultView;let h=t.left+i-(d?.scrollX??0);if(h+n>l){const u=h-(l-n);h-=u,i-=u}if(h<a){const u=h-a;h-=u,i-=u}return[i,h]}_layoutBoxInPage(e,t,i,n){const o=e.top-i,r=e.top+e.height,a=Li(this._viewDomNode.domNode),l=this._viewDomNode.domNode.ownerDocument,c=l.defaultView,d=a.top+o-(c?.scrollY??0),h=a.top+r-(c?.scrollY??0),u=lg(l.body),[g,f]=this._layoutHorizontalSegmentInPage(u,a,e.left-n.scrollLeft+this._contentLeft,t),p=22,_=22,b=d>=p,w=h+i<=u.height-_;return this._fixedOverflowWidgets?{fitsAbove:b,aboveTop:Math.max(d,p),fitsBelow:w,belowTop:h,left:f}:{fitsAbove:b,aboveTop:o,fitsBelow:w,belowTop:r,left:g}}_prepareRenderWidgetAtExactPositionOverflowing(e){return new IC(e.top,e.left+this._contentLeft)}_getAnchorsCoordinates(e){const t=o(this._primaryAnchor.viewPosition,this._affinity),i=this._secondaryAnchor.viewPosition?.lineNumber===this._primaryAnchor.viewPosition?.lineNumber?this._secondaryAnchor.viewPosition:null,n=o(i,this._affinity);return{primary:t,secondary:n};function o(r,a){if(!r)return null;const l=e.visibleRangeForPosition(r);if(!l)return null;const c=r.column===1&&a===3?0:l.left,d=e.getVerticalOffsetForLineNumber(r.lineNumber)-e.scrollTop,h=e.getLineHeightForLineNumber(r.lineNumber);return new U$(d,c,h)}}_reduceAnchorCoordinates(e,t,i){if(!t)return e;const n=this._context.configuration.options.get(59);let o=t.left;return o<e.left?o=Math.max(o,e.left-i+n.typicalFullwidthCharacterWidth):o=Math.min(o,e.left+i-n.typicalFullwidthCharacterWidth),new U$(e.top,o,e.height)}_prepareRenderWidget(e){if(!this._preference||this._preference.length===0)return null;const{primary:t,secondary:i}=this._getAnchorsCoordinates(e);if(!t)return{kind:"offViewport",preserveFocus:this.domNode.domNode.contains(this.domNode.domNode.ownerDocument.activeElement)};if(this._cachedDomNodeOffsetWidth===-1||this._cachedDomNodeOffsetHeight===-1){let r=null;if(typeof this._actual.beforeRender=="function"&&(r=nM(this._actual.beforeRender,this._actual)),r)this._cachedDomNodeOffsetWidth=r.width,this._cachedDomNodeOffsetHeight=r.height;else{const l=this.domNode.domNode.getBoundingClientRect();this._cachedDomNodeOffsetWidth=Math.round(l.width),this._cachedDomNodeOffsetHeight=Math.round(l.height)}}const n=this._reduceAnchorCoordinates(t,i,this._cachedDomNodeOffsetWidth);let o;this.allowEditorOverflow?o=this._layoutBoxInPage(n,this._cachedDomNodeOffsetWidth,this._cachedDomNodeOffsetHeight,e):o=this._layoutBoxInViewport(n,this._cachedDomNodeOffsetWidth,this._cachedDomNodeOffsetHeight,e);for(let r=1;r<=2;r++)for(const a of this._preference)if(a===1){if(!o)return null;if(r===2||o.fitsAbove)return{kind:"inViewport",coordinate:new IC(o.aboveTop,o.left),position:1}}else if(a===2){if(!o)return null;if(r===2||o.fitsBelow)return{kind:"inViewport",coordinate:new IC(o.belowTop,o.left),position:2}}else return this.allowEditorOverflow?{kind:"inViewport",coordinate:this._prepareRenderWidgetAtExactPositionOverflowing(new IC(n.top,n.left)),position:0}:{kind:"inViewport",coordinate:new IC(n.top,n.left),position:0};return null}onBeforeRender(e){!this._primaryAnchor.viewPosition||!this._preference||this._primaryAnchor.viewPosition.lineNumber<e.startLineNumber||this._primaryAnchor.viewPosition.lineNumber>e.endLineNumber||this.domNode.setMaxWidth(this._maxWidth)}prepareRender(e){this._renderData=this._prepareRenderWidget(e)}render(e){if(!this._renderData||this._renderData.kind==="offViewport"){this._isVisible&&(this.domNode.removeAttribute("monaco-visible-content-widget"),this._isVisible=!1,this._renderData?.kind==="offViewport"&&this._renderData.preserveFocus?this.domNode.setTop(-1e3):this.domNode.setVisibility("hidden")),typeof this._actual.afterRender=="function"&&nM(this._actual.afterRender,this._actual,null,null);return}this.allowEditorOverflow?(this.domNode.setTop(this._renderData.coordinate.top),this.domNode.setLeft(this._renderData.coordinate.left)):(this.domNode.setTop(this._renderData.coordinate.top+e.scrollTop-e.bigNumbersDelta),this.domNode.setLeft(this._renderData.coordinate.left)),this._isVisible||(this.domNode.setVisibility("inherit"),this.domNode.setAttribute("monaco-visible-content-widget","true"),this._isVisible=!0),typeof this._actual.afterRender=="function"&&nM(this._actual.afterRender,this._actual,this._renderData.position,this._renderData.coordinate)}}class EC{constructor(e,t){this.modelPosition=e,this.viewPosition=t}}class IC{constructor(e,t){this.top=e,this.left=t,this._coordinateBrand=void 0}}class U${constructor(e,t,i){this.top=e,this.left=t,this.height=i,this._anchorCoordinateBrand=void 0}}function nM(s,e,...t){try{return s.call(e,...t)}catch{return null}}class Dg extends c_{}const wi=qe("themeService");function Ci(s){return{id:s}}function rF(s){switch(s){case Xr.DARK:return kb.VS_DARK;case Xr.HIGH_CONTRAST_DARK:return kb.HC_BLACK;case Xr.HIGH_CONTRAST_LIGHT:return kb.HC_LIGHT;default:return kb.VS}}const cee={ThemingContribution:"base.contributions.theming"};class Vbe extends F{constructor(){super(),this.themingParticipants=[],this.themingParticipants=[],this.onThemingParticipantAddedEmitter=this._register(new M)}onColorThemeChange(e){return this.themingParticipants.push(e),this.onThemingParticipantAddedEmitter.fire(e),pe(()=>{const t=this.themingParticipants.indexOf(e);this.themingParticipants.splice(t,1)})}getThemingParticipants(){return this.themingParticipants}}const dee=new Vbe;bi.add(cee.ThemingContribution,dee);function ua(s){return dee.onColorThemeChange(s)}class zbe extends F{constructor(e){super(),this.themeService=e,this.theme=e.getColorTheme(),this._register(this.themeService.onDidColorThemeChange(t=>this.onThemeChange(t)))}onThemeChange(e){this.theme=e,this.updateStyles()}updateStyles(){}}const hee=T("editor.lineHighlightBackground",null,m(610,"Background color for the highlight of line at the cursor position.")),q$=T("editor.lineHighlightBorder",{dark:"#282828",light:"#eeeeee",hcDark:"#f38518",hcLight:rt},m(611,"Background color for the border around the line at the cursor position."));T("editor.rangeHighlightBackground",{dark:"#ffffff0b",light:"#fdff0033",hcDark:null,hcLight:null},m(612,"Background color of highlighted ranges, like by quick open and find features. The color must not be opaque so as not to hide underlying decorations."),!0);T("editor.rangeHighlightBorder",{dark:null,light:null,hcDark:ri,hcLight:ri},m(613,"Background color of the border around highlighted ranges."));T("editor.symbolHighlightBackground",{dark:zd,light:zd,hcDark:null,hcLight:null},m(614,"Background color of highlighted symbol, like for go to definition or go next/previous symbol. The color must not be opaque so as not to hide underlying decorations."),!0);T("editor.symbolHighlightBorder",{dark:null,light:null,hcDark:ri,hcLight:ri},m(615,"Background color of the border around highlighted symbols."));const hA=T("editorCursor.foreground",{dark:"#AEAFAD",light:q.black,hcDark:q.white,hcLight:"#0F4A85"},m(616,"Color of the editor cursor.")),UW=T("editorCursor.background",null,m(617,"The background color of the editor cursor. Allows customizing the color of a character overlapped by a block cursor.")),uee=T("editorMultiCursor.primary.foreground",hA,m(618,"Color of the primary editor cursor when multiple cursors are present.")),$be=T("editorMultiCursor.primary.background",UW,m(619,"The background color of the primary editor cursor when multiple cursors are present. Allows customizing the color of a character overlapped by a block cursor.")),gee=T("editorMultiCursor.secondary.foreground",hA,m(620,"Color of secondary editor cursors when multiple cursors are present.")),Ube=T("editorMultiCursor.secondary.background",UW,m(621,"The background color of secondary editor cursors when multiple cursors are present. Allows customizing the color of a character overlapped by a block cursor.")),qW=T("editorWhitespace.foreground",{dark:"#e3e4e229",light:"#33333333",hcDark:"#e3e4e229",hcLight:"#CCCCCC"},m(622,"Color of whitespace characters in the editor.")),qbe=T("editorLineNumber.foreground",{dark:"#858585",light:"#237893",hcDark:q.white,hcLight:"#292929"},m(623,"Color of editor line numbers.")),jbe=T("editorIndentGuide.background",qW,m(624,"Color of the editor indentation guides."),!1,m(625,"'editorIndentGuide.background' is deprecated. Use 'editorIndentGuide.background1' instead.")),Gbe=T("editorIndentGuide.activeBackground",qW,m(626,"Color of the active editor indentation guides."),!1,m(627,"'editorIndentGuide.activeBackground' is deprecated. Use 'editorIndentGuide.activeBackground1' instead.")),ix=T("editorIndentGuide.background1",jbe,m(628,"Color of the editor indentation guides (1).")),Kbe=T("editorIndentGuide.background2","#00000000",m(629,"Color of the editor indentation guides (2).")),Ybe=T("editorIndentGuide.background3","#00000000",m(630,"Color of the editor indentation guides (3).")),Zbe=T("editorIndentGuide.background4","#00000000",m(631,"Color of the editor indentation guides (4).")),Xbe=T("editorIndentGuide.background5","#00000000",m(632,"Color of the editor indentation guides (5).")),Qbe=T("editorIndentGuide.background6","#00000000",m(633,"Color of the editor indentation guides (6).")),nx=T("editorIndentGuide.activeBackground1",Gbe,m(634,"Color of the active editor indentation guides (1).")),Jbe=T("editorIndentGuide.activeBackground2","#00000000",m(635,"Color of the active editor indentation guides (2).")),ewe=T("editorIndentGuide.activeBackground3","#00000000",m(636,"Color of the active editor indentation guides (3).")),twe=T("editorIndentGuide.activeBackground4","#00000000",m(637,"Color of the active editor indentation guides (4).")),iwe=T("editorIndentGuide.activeBackground5","#00000000",m(638,"Color of the active editor indentation guides (5).")),nwe=T("editorIndentGuide.activeBackground6","#00000000",m(639,"Color of the active editor indentation guides (6).")),swe=T("editorActiveLineNumber.foreground",{dark:"#c6c6c6",light:"#0B216F",hcDark:ri,hcLight:ri},m(640,"Color of editor active line number"),!1,m(641,"Id is deprecated. Use 'editorLineNumber.activeForeground' instead."));T("editorLineNumber.activeForeground",swe,m(642,"Color of editor active line number"));const owe=T("editorLineNumber.dimmedForeground",null,m(643,"Color of the final editor line when editor.renderFinalNewline is set to dimmed.")),rwe=T("editorRuler.foreground",{dark:"#5A5A5A",light:q.lightgrey,hcDark:q.white,hcLight:"#292929"},m(644,"Color of the editor rulers."));T("editorCodeLens.foreground",{dark:"#999999",light:"#919191",hcDark:"#999999",hcLight:"#292929"},m(645,"Foreground color of editor CodeLens"));T("editorBracketMatch.background",{dark:"#0064001a",light:"#0064001a",hcDark:"#0064001a",hcLight:"#0000"},m(646,"Background color behind matching brackets"));T("editorBracketMatch.border",{dark:"#888",light:"#B9B9B9",hcDark:rt,hcLight:rt},m(647,"Color for matching brackets boxes"));const awe=T("editorOverviewRuler.border",{dark:"#7f7f7f4d",light:"#7f7f7f4d",hcDark:"#7f7f7f4d",hcLight:"#666666"},m(648,"Color of the overview ruler border.")),lwe=T("editorOverviewRuler.background",null,m(649,"Background color of the editor overview ruler."));T("editorGutter.background",Bi,m(650,"Background color of the editor gutter. The gutter contains the glyph margins and the line numbers."));T("editorUnnecessaryCode.border",{dark:null,light:null,hcDark:q.fromHex("#fff").transparent(.8),hcLight:rt},m(651,"Border color of unnecessary (unused) source code in the editor."));const cwe=T("editorUnnecessaryCode.opacity",{dark:q.fromHex("#000a"),light:q.fromHex("#0007"),hcDark:null,hcLight:null},m(652,`Opacity of unnecessary (unused) source code in the editor. For example, "#000000c0" will render the code with 75% opacity. For high contrast themes, use the  'editorUnnecessaryCode.border' theme color to underline unnecessary code instead of fading it out.`));T("editorGhostText.border",{dark:null,light:null,hcDark:q.fromHex("#fff").transparent(.8),hcLight:q.fromHex("#292929").transparent(.8)},m(653,"Border color of ghost text in the editor."));const dwe=T("editorGhostText.foreground",{dark:q.fromHex("#ffffff56"),light:q.fromHex("#0007"),hcDark:null,hcLight:null},m(654,"Foreground color of the ghost text in the editor."));T("editorGhostText.background",null,m(655,"Background color of the ghost text in the editor."));const hwe=new q(new G(0,122,204,.6)),fee=T("editorOverviewRuler.rangeHighlightForeground",hwe,m(656,"Overview ruler marker color for range highlights. The color must not be opaque so as not to hide underlying decorations."),!0),uwe=T("editorOverviewRuler.errorForeground",{dark:new q(new G(255,18,18,.7)),light:new q(new G(255,18,18,.7)),hcDark:new q(new G(255,50,50,1)),hcLight:"#B5200D"},m(657,"Overview ruler marker color for errors.")),gwe=T("editorOverviewRuler.warningForeground",{dark:_h,light:_h,hcDark:aS,hcLight:aS},m(658,"Overview ruler marker color for warnings.")),fwe=T("editorOverviewRuler.infoForeground",{dark:Rc,light:Rc,hcDark:lS,hcLight:lS},m(659,"Overview ruler marker color for infos.")),pee=T("editorBracketHighlight.foreground1",{dark:"#FFD700",light:"#0431FAFF",hcDark:"#FFD700",hcLight:"#0431FAFF"},m(660,"Foreground color of brackets (1). Requires enabling bracket pair colorization.")),mee=T("editorBracketHighlight.foreground2",{dark:"#DA70D6",light:"#319331FF",hcDark:"#DA70D6",hcLight:"#319331FF"},m(661,"Foreground color of brackets (2). Requires enabling bracket pair colorization.")),_ee=T("editorBracketHighlight.foreground3",{dark:"#179FFF",light:"#7B3814FF",hcDark:"#87CEFA",hcLight:"#7B3814FF"},m(662,"Foreground color of brackets (3). Requires enabling bracket pair colorization.")),bee=T("editorBracketHighlight.foreground4","#00000000",m(663,"Foreground color of brackets (4). Requires enabling bracket pair colorization.")),wee=T("editorBracketHighlight.foreground5","#00000000",m(664,"Foreground color of brackets (5). Requires enabling bracket pair colorization.")),vee=T("editorBracketHighlight.foreground6","#00000000",m(665,"Foreground color of brackets (6). Requires enabling bracket pair colorization.")),pwe=T("editorBracketHighlight.unexpectedBracket.foreground",{dark:new q(new G(255,18,18,.8)),light:new q(new G(255,18,18,.8)),hcDark:new q(new G(255,50,50,1)),hcLight:"#B5200D"},m(666,"Foreground color of unexpected brackets.")),mwe=T("editorBracketPairGuide.background1","#00000000",m(667,"Background color of inactive bracket pair guides (1). Requires enabling bracket pair guides.")),_we=T("editorBracketPairGuide.background2","#00000000",m(668,"Background color of inactive bracket pair guides (2). Requires enabling bracket pair guides.")),bwe=T("editorBracketPairGuide.background3","#00000000",m(669,"Background color of inactive bracket pair guides (3). Requires enabling bracket pair guides.")),wwe=T("editorBracketPairGuide.background4","#00000000",m(670,"Background color of inactive bracket pair guides (4). Requires enabling bracket pair guides.")),vwe=T("editorBracketPairGuide.background5","#00000000",m(671,"Background color of inactive bracket pair guides (5). Requires enabling bracket pair guides.")),Cwe=T("editorBracketPairGuide.background6","#00000000",m(672,"Background color of inactive bracket pair guides (6). Requires enabling bracket pair guides.")),ywe=T("editorBracketPairGuide.activeBackground1","#00000000",m(673,"Background color of active bracket pair guides (1). Requires enabling bracket pair guides.")),Swe=T("editorBracketPairGuide.activeBackground2","#00000000",m(674,"Background color of active bracket pair guides (2). Requires enabling bracket pair guides.")),xwe=T("editorBracketPairGuide.activeBackground3","#00000000",m(675,"Background color of active bracket pair guides (3). Requires enabling bracket pair guides.")),Lwe=T("editorBracketPairGuide.activeBackground4","#00000000",m(676,"Background color of active bracket pair guides (4). Requires enabling bracket pair guides.")),kwe=T("editorBracketPairGuide.activeBackground5","#00000000",m(677,"Background color of active bracket pair guides (5). Requires enabling bracket pair guides.")),Ewe=T("editorBracketPairGuide.activeBackground6","#00000000",m(678,"Background color of active bracket pair guides (6). Requires enabling bracket pair guides."));T("editorUnicodeHighlight.border",_h,m(679,"Border color used to highlight unicode characters."));T("editorUnicodeHighlight.background",ome,m(680,"Background color used to highlight unicode characters."));ua((s,e)=>{const t=s.getColor(Bi),i=s.getColor(hee),n=i&&!i.isTransparent()?i:t;n&&e.addRule(`.monaco-editor .inputarea.ime-input { background-color: ${n}; }`)});class Cee extends Dg{constructor(e){super(),this._context=e;const t=this._context.configuration.options,i=t.get(165);this._renderLineHighlight=t.get(110),this._renderLineHighlightOnlyWhenFocus=t.get(111),this._wordWrap=i.isViewportWrapping,this._contentLeft=i.contentLeft,this._contentWidth=i.contentWidth,this._selectionIsEmpty=!0,this._focused=!1,this._cursorLineNumbers=[1],this._selections=[new he(1,1,1,1)],this._renderData=null,this._context.addEventHandler(this)}dispose(){this._context.removeEventHandler(this),super.dispose()}_readFromSelections(){let e=!1;const t=new Set;for(const o of this._selections)t.add(o.positionLineNumber);const i=Array.from(t);i.sort((o,r)=>o-r),ni(this._cursorLineNumbers,i)||(this._cursorLineNumbers=i,e=!0);const n=this._selections.every(o=>o.isEmpty());return this._selectionIsEmpty!==n&&(this._selectionIsEmpty=n,e=!0),e}onThemeChanged(e){return this._readFromSelections()}onConfigurationChanged(e){const t=this._context.configuration.options,i=t.get(165);return this._renderLineHighlight=t.get(110),this._renderLineHighlightOnlyWhenFocus=t.get(111),this._wordWrap=i.isViewportWrapping,this._contentLeft=i.contentLeft,this._contentWidth=i.contentWidth,!0}onCursorStateChanged(e){return this._selections=e.selections,this._readFromSelections()}onFlushed(e){return!0}onLinesDeleted(e){return!0}onLinesInserted(e){return!0}onScrollChanged(e){return e.scrollWidthChanged||e.scrollTopChanged}onZonesChanged(e){return!0}onFocusChanged(e){return this._renderLineHighlightOnlyWhenFocus?(this._focused=e.isFocused,!0):!1}prepareRender(e){if(!this._shouldRenderThis()){this._renderData=null;return}const t=e.visibleRange.startLineNumber,i=e.visibleRange.endLineNumber,n=[];for(let r=t;r<=i;r++){const a=r-t;n[a]=""}if(this._wordWrap){const r=this._renderOne(e,!1);for(const a of this._cursorLineNumbers){const l=this._context.viewModel.coordinatesConverter,c=l.convertViewPositionToModelPosition(new A(a,1)).lineNumber,d=l.convertModelPositionToViewPosition(new A(c,1)).lineNumber,h=l.convertModelPositionToViewPosition(new A(c,this._context.viewModel.model.getLineMaxColumn(c))).lineNumber,u=Math.max(d,t),g=Math.min(h,i);for(let f=u;f<=g;f++){const p=f-t;n[p]=r}}}const o=this._renderOne(e,!0);for(const r of this._cursorLineNumbers){if(r<t||r>i)continue;const a=r-t;n[a]=o}this._renderData=n}render(e,t){if(!this._renderData)return"";const i=t-e;return i>=this._renderData.length?"":this._renderData[i]}_shouldRenderInMargin(){return(this._renderLineHighlight==="gutter"||this._renderLineHighlight==="all")&&(!this._renderLineHighlightOnlyWhenFocus||this._focused)}_shouldRenderInContent(){return(this._renderLineHighlight==="line"||this._renderLineHighlight==="all")&&this._selectionIsEmpty&&(!this._renderLineHighlightOnlyWhenFocus||this._focused)}}class Iwe extends Cee{_renderOne(e,t){return`<div class="${"current-line"+(this._shouldRenderInMargin()?" current-line-both":"")+(t?" current-line-exact":"")}" style="width:${Math.max(e.scrollWidth,this._contentWidth)}px;"></div>`}_shouldRenderThis(){return this._shouldRenderInContent()}_shouldRenderOther(){return this._shouldRenderInMargin()}}class Dwe extends Cee{_renderOne(e,t){return`<div class="${"current-line"+(this._shouldRenderInMargin()?" current-line-margin":"")+(this._shouldRenderOther()?" current-line-margin-both":"")+(this._shouldRenderInMargin()&&t?" current-line-exact-margin":"")}" style="width:${this._contentLeft}px"></div>`}_shouldRenderThis(){return!0}_shouldRenderOther(){return this._shouldRenderInContent()}}ua((s,e)=>{const t=s.getColor(hee);if(t&&(e.addRule(`.monaco-editor .view-overlays .current-line { background-color: ${t}; }`),e.addRule(`.monaco-editor .margin-view-overlays .current-line-margin { background-color: ${t}; border: none; }`)),!t||t.isTransparent()||s.defines(q$)){const i=s.getColor(q$);i&&(e.addRule(`.monaco-editor .view-overlays .current-line-exact { border: 2px solid ${i}; }`),e.addRule(`.monaco-editor .margin-view-overlays .current-line-exact-margin { border: 2px solid ${i}; }`),Cl(s.type)&&(e.addRule(".monaco-editor .view-overlays .current-line-exact { border-width: 1px; }"),e.addRule(".monaco-editor .margin-view-overlays .current-line-exact-margin { border-width: 1px; }")))}});class Twe extends Dg{constructor(e){super(),this._context=e;const t=this._context.configuration.options;this._typicalHalfwidthCharacterWidth=t.get(59).typicalHalfwidthCharacterWidth,this._renderResult=null,this._context.addEventHandler(this)}dispose(){this._context.removeEventHandler(this),this._renderResult=null,super.dispose()}onConfigurationChanged(e){const t=this._context.configuration.options;return this._typicalHalfwidthCharacterWidth=t.get(59).typicalHalfwidthCharacterWidth,!0}onDecorationsChanged(e){return!0}onFlushed(e){return!0}onLinesChanged(e){return!0}onLinesDeleted(e){return!0}onLinesInserted(e){return!0}onScrollChanged(e){return e.scrollTopChanged||e.scrollWidthChanged}onZonesChanged(e){return!0}prepareRender(e){const t=e.getDecorationsInViewport();let i=[],n=0;for(let l=0,c=t.length;l<c;l++){const d=t[l];d.options.className&&(i[n++]=d)}i=i.sort((l,c)=>{if(l.options.zIndex<c.options.zIndex)return-1;if(l.options.zIndex>c.options.zIndex)return 1;const d=l.options.className,h=c.options.className;return d<h?-1:d>h?1:k.compareRangesUsingStarts(l.range,c.range)});const o=e.visibleRange.startLineNumber,r=e.visibleRange.endLineNumber,a=[];for(let l=o;l<=r;l++){const c=l-o;a[c]=""}this._renderWholeLineDecorations(e,i,a),this._renderNormalDecorations(e,i,a),this._renderResult=a}_renderWholeLineDecorations(e,t,i){const n=e.visibleRange.startLineNumber,o=e.visibleRange.endLineNumber;for(let r=0,a=t.length;r<a;r++){const l=t[r];if(!l.options.isWholeLine)continue;const c='<div class="cdr '+l.options.className+'" style="left:0;width:100%;"></div>',d=Math.max(l.range.startLineNumber,n),h=Math.min(l.range.endLineNumber,o);for(let u=d;u<=h;u++){const g=u-n;i[g]+=c}}}_renderNormalDecorations(e,t,i){const n=e.visibleRange.startLineNumber;let o=null,r=!1,a=null,l=!1;for(let c=0,d=t.length;c<d;c++){const h=t[c];if(h.options.isWholeLine)continue;const u=h.options.className,g=!!h.options.showIfCollapsed;let f=h.range;if(g&&f.endColumn===1&&f.endLineNumber!==f.startLineNumber&&(f=new k(f.startLineNumber,f.startColumn,f.endLineNumber-1,this._context.viewModel.getLineMaxColumn(f.endLineNumber-1))),o===u&&r===g&&k.areIntersectingOrTouching(a,f)){a=k.plusRange(a,f);continue}o!==null&&this._renderNormalDecoration(e,a,o,l,r,n,i),o=u,r=g,a=f,l=h.options.shouldFillLineOnLineBreak??!1}o!==null&&this._renderNormalDecoration(e,a,o,l,r,n,i)}_renderNormalDecoration(e,t,i,n,o,r,a){const l=e.linesVisibleRangesForRange(t,i==="findMatch");if(l)for(let c=0,d=l.length;c<d;c++){const h=l[c];if(h.outsideRenderedLine)continue;const u=h.lineNumber-r;if(o&&h.ranges.length===1){const g=h.ranges[0];if(g.width<this._typicalHalfwidthCharacterWidth){const f=Math.round(g.left+g.width/2),p=Math.max(0,Math.round(f-this._typicalHalfwidthCharacterWidth/2));h.ranges[0]=new ex(p,this._typicalHalfwidthCharacterWidth)}}for(let g=0,f=h.ranges.length;g<f;g++){const p=n&&h.continuesOnNextLine&&f===1,_=h.ranges[g],b='<div class="cdr '+i+'" style="left:'+String(_.left)+"px;width:"+(p?"100%;":String(_.width)+"px;")+'"></div>';a[u]+=b}}}render(e,t){if(!this._renderResult)return"";const i=t-e;return i<0||i>=this._renderResult.length?"":this._renderResult[i]}}class Nwe extends no{constructor(e,t,i,n){super(e);const o=this._context.configuration.options,r=o.get(117),a=o.get(83),l=o.get(49),c=o.get(120),d=o.get(158),h={listenOnDomNode:i.domNode,className:"editor-scrollable "+rF(e.theme.type),useShadows:!1,lazyRender:!0,vertical:r.vertical,horizontal:r.horizontal,verticalHasArrows:r.verticalHasArrows,horizontalHasArrows:r.horizontalHasArrows,verticalScrollbarSize:r.verticalScrollbarSize,verticalSliderSize:r.verticalSliderSize,horizontalScrollbarSize:r.horizontalScrollbarSize,horizontalSliderSize:r.horizontalSliderSize,handleMouseWheel:r.handleMouseWheel,alwaysConsumeMouseWheel:r.alwaysConsumeMouseWheel,arrowSize:r.arrowSize,mouseWheelScrollSensitivity:a,fastScrollSensitivity:l,scrollPredominantAxis:c,scrollByPage:r.scrollByPage,inertialScroll:d};this.scrollbar=this._register(new cA(t.domNode,h,this._context.viewLayout.getScrollable())),kl.write(this.scrollbar.getDomNode(),6),this.scrollbarDomNode=wt(this.scrollbar.getDomNode()),this.scrollbarDomNode.setPosition("absolute"),this._setLayout();const u=(g,f,p)=>{const _={};{const b=g.scrollTop;b&&(_.scrollTop=this._context.viewLayout.getCurrentScrollTop()+b,g.scrollTop=0)}if(p){const b=g.scrollLeft;b&&(_.scrollLeft=this._context.viewLayout.getCurrentScrollLeft()+b,g.scrollLeft=0)}this._context.viewModel.viewLayout.setScrollPosition(_,1)};this._register(W(i.domNode,"scroll",g=>u(i.domNode,!0,!0))),this._register(W(t.domNode,"scroll",g=>u(t.domNode,!0,!1))),this._register(W(n.domNode,"scroll",g=>u(n.domNode,!0,!1))),this._register(W(this.scrollbarDomNode.domNode,"scroll",g=>u(this.scrollbarDomNode.domNode,!0,!1)))}dispose(){super.dispose()}_setLayout(){const e=this._context.configuration.options,t=e.get(165);this.scrollbarDomNode.setLeft(t.contentLeft),e.get(81).side==="right"?this.scrollbarDomNode.setWidth(t.contentWidth+t.minimap.minimapWidth):this.scrollbarDomNode.setWidth(t.contentWidth),this.scrollbarDomNode.setHeight(t.height)}getOverviewRulerLayoutInfo(){return this.scrollbar.getOverviewRulerLayoutInfo()}getDomNode(){return this.scrollbarDomNode}delegateVerticalScrollbarPointerDown(e){this.scrollbar.delegateVerticalScrollbarPointerDown(e)}delegateScrollFromMouseWheelEvent(e){this.scrollbar.delegateScrollFromMouseWheelEvent(e)}onConfigurationChanged(e){if(e.hasChanged(117)||e.hasChanged(83)||e.hasChanged(49)){const t=this._context.configuration.options,i=t.get(117),n=t.get(83),o=t.get(49),r=t.get(120),a={vertical:i.vertical,horizontal:i.horizontal,verticalScrollbarSize:i.verticalScrollbarSize,horizontalScrollbarSize:i.horizontalScrollbarSize,scrollByPage:i.scrollByPage,handleMouseWheel:i.handleMouseWheel,mouseWheelScrollSensitivity:n,fastScrollSensitivity:o,scrollPredominantAxis:r};this.scrollbar.updateOptions(a)}return e.hasChanged(165)&&this._setLayout(),!0}onScrollChanged(e){return!0}onThemeChanged(e){return this.scrollbar.updateClassName("editor-scrollable "+rF(this._context.theme.type)),!0}prepareRender(e){}render(e){this.scrollbar.renderNow()}}class aF{constructor(e,t,i,n,o){this.startLineNumber=e,this.endLineNumber=t,this.className=i,this.tooltip=n,this._decorationToRenderBrand=void 0,this.zIndex=o??0}}class Rwe{constructor(e,t,i){this.className=e,this.zIndex=t,this.tooltip=i}}class Awe{constructor(){this.decorations=[]}add(e){this.decorations.push(e)}getDecorations(){return this.decorations}}class yee extends Dg{_render(e,t,i){const n=[];for(let a=e;a<=t;a++){const l=a-e;n[l]=new Awe}if(i.length===0)return n;i.sort((a,l)=>a.className===l.className?a.startLineNumber===l.startLineNumber?a.endLineNumber-l.endLineNumber:a.startLineNumber-l.startLineNumber:a.className<l.className?-1:1);let o=null,r=0;for(let a=0,l=i.length;a<l;a++){const c=i[a],d=c.className,h=c.zIndex;let u=Math.max(c.startLineNumber,e)-e;const g=Math.min(c.endLineNumber,t)-e;o===d?(u=Math.max(r+1,u),r=Math.max(r,g)):(o=d,r=g);for(let f=u;f<=r;f++)n[f].add(new Rwe(d,h,c.tooltip))}return n}}class Mwe extends no{constructor(e){super(e),this._widgets={},this._context=e;const t=this._context.configuration.options,i=t.get(165);this.domNode=wt(document.createElement("div")),this.domNode.setClassName("glyph-margin-widgets"),this.domNode.setPosition("absolute"),this.domNode.setTop(0),this._lineHeight=t.get(75),this._glyphMargin=t.get(66),this._glyphMarginLeft=i.glyphMarginLeft,this._glyphMarginWidth=i.glyphMarginWidth,this._glyphMarginDecorationLaneCount=i.glyphMarginDecorationLaneCount,this._managedDomNodes=[],this._decorationGlyphsToRender=[]}dispose(){this._managedDomNodes=[],this._decorationGlyphsToRender=[],this._widgets={},super.dispose()}getWidgets(){return Object.values(this._widgets)}onConfigurationChanged(e){const t=this._context.configuration.options,i=t.get(165);return this._lineHeight=t.get(75),this._glyphMargin=t.get(66),this._glyphMarginLeft=i.glyphMarginLeft,this._glyphMarginWidth=i.glyphMarginWidth,this._glyphMarginDecorationLaneCount=i.glyphMarginDecorationLaneCount,!0}onDecorationsChanged(e){return!0}onFlushed(e){return!0}onLinesChanged(e){return!0}onLinesDeleted(e){return!0}onLinesInserted(e){return!0}onScrollChanged(e){return e.scrollTopChanged}onZonesChanged(e){return!0}addWidget(e){const t=wt(e.getDomNode());this._widgets[e.getId()]={widget:e,preference:e.getPosition(),domNode:t,renderInfo:null},t.setPosition("absolute"),t.setDisplay("none"),t.setAttribute("widgetId",e.getId()),this.domNode.appendChild(t),this.setShouldRender()}setWidgetPosition(e,t){const i=this._widgets[e.getId()];return i.preference.lane===t.lane&&i.preference.zIndex===t.zIndex&&k.equalsRange(i.preference.range,t.range)?!1:(i.preference=t,this.setShouldRender(),!0)}removeWidget(e){const t=e.getId();if(this._widgets[t]){const n=this._widgets[t].domNode.domNode;delete this._widgets[t],n.remove(),this.setShouldRender()}}_collectDecorationBasedGlyphRenderRequest(e,t){const i=e.visibleRange.startLineNumber,n=e.visibleRange.endLineNumber,o=e.getDecorationsInViewport();for(const r of o){const a=r.options.glyphMarginClassName;if(!a)continue;const l=Math.max(r.range.startLineNumber,i),c=Math.min(r.range.endLineNumber,n),d=r.options.glyphMargin?.position??Ll.Center,h=r.options.zIndex??0;for(let u=l;u<=c;u++){const g=this._context.viewModel.coordinatesConverter.convertViewPositionToModelPosition(new A(u,0)),f=this._context.viewModel.glyphLanes.getLanesAtLine(g.lineNumber).indexOf(d);t.push(new Owe(u,f,h,a))}}}_collectWidgetBasedGlyphRenderRequest(e,t){const i=e.visibleRange.startLineNumber,n=e.visibleRange.endLineNumber;for(const o of Object.values(this._widgets)){const r=o.preference.range,{startLineNumber:a,endLineNumber:l}=this._context.viewModel.coordinatesConverter.convertModelRangeToViewRange(k.lift(r));if(!a||!l||l<i||a>n)continue;const c=Math.max(a,i),d=this._context.viewModel.coordinatesConverter.convertViewPositionToModelPosition(new A(c,0)),h=this._context.viewModel.glyphLanes.getLanesAtLine(d.lineNumber).indexOf(o.preference.lane);t.push(new Pwe(c,h,o.preference.zIndex,o))}}_collectSortedGlyphRenderRequests(e){const t=[];return this._collectDecorationBasedGlyphRenderRequest(e,t),this._collectWidgetBasedGlyphRenderRequest(e,t),t.sort((i,n)=>i.lineNumber===n.lineNumber?i.laneIndex===n.laneIndex?i.zIndex===n.zIndex?n.type===i.type?i.type===0&&n.type===0?i.className<n.className?-1:1:0:n.type-i.type:n.zIndex-i.zIndex:i.laneIndex-n.laneIndex:i.lineNumber-n.lineNumber),t}prepareRender(e){if(!this._glyphMargin){this._decorationGlyphsToRender=[];return}for(const n of Object.values(this._widgets))n.renderInfo=null;const t=new ph(this._collectSortedGlyphRenderRequests(e)),i=[];for(;t.length>0;){const n=t.peek();if(!n)break;const o=t.takeWhile(a=>a.lineNumber===n.lineNumber&&a.laneIndex===n.laneIndex);if(!o||o.length===0)break;const r=o[0];if(r.type===0){const a=[];for(const l of o){if(l.zIndex!==r.zIndex||l.type!==r.type)break;(a.length===0||a[a.length-1]!==l.className)&&a.push(l.className)}i.push(r.accept(a.join(" ")))}else r.widget.renderInfo={lineNumber:r.lineNumber,laneIndex:r.laneIndex}}this._decorationGlyphsToRender=i}render(e){if(!this._glyphMargin){for(const i of Object.values(this._widgets))i.domNode.setDisplay("none");for(;this._managedDomNodes.length>0;)this._managedDomNodes.pop()?.domNode.remove();return}const t=Math.round(this._glyphMarginWidth/this._glyphMarginDecorationLaneCount);for(const i of Object.values(this._widgets))if(!i.renderInfo)i.domNode.setDisplay("none");else{const n=e.viewportData.relativeVerticalOffset[i.renderInfo.lineNumber-e.viewportData.startLineNumber],o=this._glyphMarginLeft+i.renderInfo.laneIndex*this._lineHeight;i.domNode.setDisplay("block"),i.domNode.setTop(n),i.domNode.setLeft(o),i.domNode.setWidth(t),i.domNode.setHeight(this._lineHeight)}for(let i=0;i<this._decorationGlyphsToRender.length;i++){const n=this._decorationGlyphsToRender[i],o=n.lineNumber,r=e.viewportData.relativeVerticalOffset[o-e.viewportData.startLineNumber],a=this._glyphMarginLeft+n.laneIndex*this._lineHeight;let l;i<this._managedDomNodes.length?l=this._managedDomNodes[i]:(l=wt(document.createElement("div")),this._managedDomNodes.push(l),this.domNode.appendChild(l));const c=this._context.viewLayout.getLineHeightForLineNumber(o);l.setClassName("cgmr codicon "+n.combinedClassName),l.setPosition("absolute"),l.setTop(r),l.setLeft(a),l.setWidth(t),l.setHeight(c)}for(;this._managedDomNodes.length>this._decorationGlyphsToRender.length;)this._managedDomNodes.pop()?.domNode.remove()}}class Owe{constructor(e,t,i,n){this.lineNumber=e,this.laneIndex=t,this.zIndex=i,this.className=n,this.type=0}accept(e){return new Fwe(this.lineNumber,this.laneIndex,e)}}class Pwe{constructor(e,t,i,n){this.lineNumber=e,this.laneIndex=t,this.zIndex=i,this.widget=n,this.type=1}}class Fwe{constructor(e,t,i){this.lineNumber=e,this.laneIndex=t,this.combinedClassName=i}}function gS(s,e,t=s.length-1){const i=Bwe(s,e,t);if(i!==-1)return s[i]}function Bwe(s,e,t=s.length-1){for(let i=t;i>=0;i--){const n=s[i];if(e(n,i))return i}return-1}function fv(s,e){const t=pv(s,e);return t===-1?void 0:s[t]}function pv(s,e,t=0,i=s.length){let n=t,o=i;for(;n<o;){const r=Math.floor((n+o)/2);e(s[r])?n=r+1:o=r}return n-1}function Wwe(s,e){const t=fS(s,e);return t===s.length?void 0:s[t]}function fS(s,e,t=0,i=s.length){let n=t,o=i;for(;n<o;){const r=Math.floor((n+o)/2);e(s[r])?o=r:n=r+1}return n}const ZT=class ZT{constructor(e){this._array=e,this._findLastMonotonousLastIdx=0}findLastMonotonous(e){if(ZT.assertInvariants){if(this._prevFindLastPredicate){for(const i of this._array)if(this._prevFindLastPredicate(i)&&!e(i))throw new Error("MonotonousArray: current predicate must be weaker than (or equal to) the previous predicate.")}this._prevFindLastPredicate=e}const t=pv(this._array,e,this._findLastMonotonousLastIdx);return this._findLastMonotonousLastIdx=t+1,t===-1?void 0:this._array[t]}};ZT.assertInvariants=!1;let QE=ZT;function jW(s,e){if(s.length===0)return;let t=s[0];for(let i=1;i<s.length;i++){const n=s[i];e(n,t)>0&&(t=n)}return t}function See(s,e){if(s.length===0)return;let t=s[0];for(let i=1;i<s.length;i++){const n=s[i];e(n,t)>=0&&(t=n)}return t}function xee(s,e){return jW(s,(t,i)=>-e(t,i))}function Hwe(s,e){if(s.length===0)return-1;let t=0;for(let i=1;i<s.length;i++){const n=s[i];e(n,s[t])>0&&(t=i)}return t}function Vwe(s,e){for(const t of s){const i=e(t);if(i!==void 0)return i}}class Lee extends F{constructor(){super(...arguments),this._isDisposed=!1}dispose(){super.dispose(),this._isDisposed=!0}assertNotDisposed(){if(this._isDisposed)throw new Error("TextModelPart is disposed!")}}function uA(s,e){let t=0,i=0;const n=s.length;for(;i<n;){const o=s.charCodeAt(i);if(o===32)t++;else if(o===9)t=t-t%e+e;else break;i++}return i===n?-1:t}var cp;(function(s){s[s.Disabled=0]="Disabled",s[s.EnabledForActive=1]="EnabledForActive",s[s.Enabled=2]="Enabled"})(cp||(cp={}));class Gf{constructor(e,t,i,n,o,r){if(this.visibleColumn=e,this.column=t,this.className=i,this.horizontalLine=n,this.forWrappedLinesAfterColumn=o,this.forWrappedLinesBeforeOrAtColumn=r,e!==-1==(t!==-1))throw new Error}}class N1{constructor(e,t){this.top=e,this.endColumn=t}}class zwe extends Lee{constructor(e,t){super(),this.textModel=e,this.languageConfigurationService=t}getLanguageConfiguration(e){return this.languageConfigurationService.getLanguageConfiguration(e)}_computeIndentLevel(e){return uA(this.textModel.getLineContent(e+1),this.textModel.getOptions().tabSize)}getActiveIndentGuide(e,t,i){this.assertNotDisposed();const n=this.textModel.getLineCount();if(e<1||e>n)throw new Ce("Illegal value for lineNumber");const o=this.getLanguageConfiguration(this.textModel.getLanguageId()).foldingRules,r=!!(o&&o.offSide);let a=-2,l=-1,c=-2,d=-1;const h=L=>{if(a!==-1&&(a===-2||a>L-1)){a=-1,l=-1;for(let E=L-2;E>=0;E--){const I=this._computeIndentLevel(E);if(I>=0){a=E,l=I;break}}}if(c===-2){c=-1,d=-1;for(let E=L;E<n;E++){const I=this._computeIndentLevel(E);if(I>=0){c=E,d=I;break}}}};let u=-2,g=-1,f=-2,p=-1;const _=L=>{if(u===-2){u=-1,g=-1;for(let E=L-2;E>=0;E--){const I=this._computeIndentLevel(E);if(I>=0){u=E,g=I;break}}}if(f!==-1&&(f===-2||f<L-1)){f=-1,p=-1;for(let E=L;E<n;E++){const I=this._computeIndentLevel(E);if(I>=0){f=E,p=I;break}}}};let b=0,w=!0,C=0,v=!0,y=0,x=0;for(let L=0;w||v;L++){const E=e-L,I=e+L;L>1&&(E<1||E<t)&&(w=!1),L>1&&(I>n||I>i)&&(v=!1),L>5e4&&(w=!1,v=!1);let R=-1;if(w&&E>=1){const P=this._computeIndentLevel(E-1);P>=0?(c=E-1,d=P,R=Math.ceil(P/this.textModel.getOptions().indentSize)):(h(E),R=this._getIndentLevelForWhitespaceLine(r,l,d))}let O=-1;if(v&&I<=n){const P=this._computeIndentLevel(I-1);P>=0?(u=I-1,g=P,O=Math.ceil(P/this.textModel.getOptions().indentSize)):(_(I),O=this._getIndentLevelForWhitespaceLine(r,g,p))}if(L===0){x=R;continue}if(L===1){if(I<=n&&O>=0&&x+1===O){w=!1,b=I,C=I,y=O;continue}if(E>=1&&R>=0&&R-1===x){v=!1,b=E,C=E,y=R;continue}if(b=e,C=e,y=x,y===0)return{startLineNumber:b,endLineNumber:C,indent:y}}w&&(R>=y?b=E:w=!1),v&&(O>=y?C=I:v=!1)}return{startLineNumber:b,endLineNumber:C,indent:y}}getLinesBracketGuides(e,t,i,n){const o=[];for(let h=e;h<=t;h++)o.push([]);const r=!0,a=this.textModel.bracketPairs.getBracketPairsInRangeWithMinIndentation(new k(e,1,t,this.textModel.getLineMaxColumn(t))).toArray();let l;if(i&&a.length>0){const h=(e<=i.lineNumber&&i.lineNumber<=t?a:this.textModel.bracketPairs.getBracketPairsInRange(k.fromPositions(i)).toArray()).filter(u=>k.strictContainsPosition(u.range,i));l=gS(h,u=>r)?.range}const c=this.textModel.getOptions().bracketPairColorizationOptions.independentColorPoolPerBracketType,d=new kee;for(const h of a){if(!h.closingBracketRange)continue;const u=l&&h.range.equalsRange(l);if(!u&&!n.includeInactive)continue;const g=d.getInlineClassName(h.nestingLevel,h.nestingLevelOfEqualBracketType,c)+(n.highlightActive&&u?" "+d.activeClassName:""),f=h.openingBracketRange.getStartPosition(),p=h.closingBracketRange.getStartPosition(),_=n.horizontalGuides===cp.Enabled||n.horizontalGuides===cp.EnabledForActive&&u;if(h.range.startLineNumber===h.range.endLineNumber){_&&o[h.range.startLineNumber-e].push(new Gf(-1,h.openingBracketRange.getEndPosition().column,g,new N1(!1,p.column),-1,-1));continue}const b=this.getVisibleColumnFromPosition(p),w=this.getVisibleColumnFromPosition(h.openingBracketRange.getStartPosition()),C=Math.min(w,b,h.minVisibleColumnIndentation+1);let v=!1;As(this.textModel.getLineContent(h.closingBracketRange.startLineNumber))<h.closingBracketRange.startColumn-1&&(v=!0);const L=Math.max(f.lineNumber,e),E=Math.min(p.lineNumber,t),I=v?1:0;for(let R=L;R<E+I;R++)o[R-e].push(new Gf(C,-1,g,null,R===f.lineNumber?f.column:-1,R===p.lineNumber?p.column:-1));_&&(f.lineNumber>=e&&w>C&&o[f.lineNumber-e].push(new Gf(C,-1,g,new N1(!1,f.column),-1,-1)),p.lineNumber<=t&&b>C&&o[p.lineNumber-e].push(new Gf(C,-1,g,new N1(!v,p.column),-1,-1)))}for(const h of o)h.sort((u,g)=>u.visibleColumn-g.visibleColumn);return o}getVisibleColumnFromPosition(e){return pi.visibleColumnFromColumn(this.textModel.getLineContent(e.lineNumber),e.column,this.textModel.getOptions().tabSize)+1}getLinesIndentGuides(e,t){this.assertNotDisposed();const i=this.textModel.getLineCount();if(e<1||e>i)throw new Error("Illegal value for startLineNumber");if(t<1||t>i)throw new Error("Illegal value for endLineNumber");const n=this.textModel.getOptions(),o=this.getLanguageConfiguration(this.textModel.getLanguageId()).foldingRules,r=!!(o&&o.offSide),a=new Array(t-e+1);let l=-2,c=-1,d=-2,h=-1;for(let u=e;u<=t;u++){const g=u-e,f=this._computeIndentLevel(u-1);if(f>=0){l=u-1,c=f,a[g]=Math.ceil(f/n.indentSize);continue}if(l===-2){l=-1,c=-1;for(let p=u-2;p>=0;p--){const _=this._computeIndentLevel(p);if(_>=0){l=p,c=_;break}}}if(d!==-1&&(d===-2||d<u-1)){d=-1,h=-1;for(let p=u;p<i;p++){const _=this._computeIndentLevel(p);if(_>=0){d=p,h=_;break}}}a[g]=this._getIndentLevelForWhitespaceLine(r,c,h)}return a}_getIndentLevelForWhitespaceLine(e,t,i){const n=this.textModel.getOptions();return t===-1||i===-1?0:t<i?1+Math.floor(t/n.indentSize):t===i||e?Math.ceil(i/n.indentSize):1+Math.floor(i/n.indentSize)}}class kee{constructor(){this.activeClassName="indent-active"}getInlineClassName(e,t,i){return this.getInlineClassNameOfLevel(i?t:e)}getInlineClassNameOfLevel(e){return`bracket-indent-guide lvl-${e%30}`}}class $we extends Dg{constructor(e){super(),this._context=e,this._primaryPosition=null;const t=this._context.configuration.options,i=t.get(166),n=t.get(59);this._spaceWidth=n.spaceWidth,this._maxIndentLeft=i.wrappingColumn===-1?-1:i.wrappingColumn*n.typicalHalfwidthCharacterWidth,this._bracketPairGuideOptions=t.get(22),this._renderResult=null,this._context.addEventHandler(this)}dispose(){this._context.removeEventHandler(this),this._renderResult=null,super.dispose()}onConfigurationChanged(e){const t=this._context.configuration.options,i=t.get(166),n=t.get(59);return this._spaceWidth=n.spaceWidth,this._maxIndentLeft=i.wrappingColumn===-1?-1:i.wrappingColumn*n.typicalHalfwidthCharacterWidth,this._bracketPairGuideOptions=t.get(22),!0}onCursorStateChanged(e){const i=e.selections[0].getPosition();return this._primaryPosition?.equals(i)?!1:(this._primaryPosition=i,!0)}onDecorationsChanged(e){return!0}onFlushed(e){return!0}onLinesChanged(e){return!0}onLinesDeleted(e){return!0}onLinesInserted(e){return!0}onScrollChanged(e){return e.scrollTopChanged}onZonesChanged(e){return!0}onLanguageConfigurationChanged(e){return!0}prepareRender(e){if(!this._bracketPairGuideOptions.indentation&&this._bracketPairGuideOptions.bracketPairs===!1){this._renderResult=null;return}const t=e.visibleRange.startLineNumber,i=e.visibleRange.endLineNumber,n=e.scrollWidth,o=this._primaryPosition,r=this.getGuidesByLine(t,Math.min(i+1,this._context.viewModel.getLineCount()),o),a=[];for(let l=t;l<=i;l++){const c=l-t,d=r[c];let h="";const u=e.visibleRangeForPosition(new A(l,1))?.left??0;for(const g of d){const f=g.column===-1?u+(g.visibleColumn-1)*this._spaceWidth:e.visibleRangeForPosition(new A(l,g.column)).left;if(f>n||this._maxIndentLeft>0&&f>this._maxIndentLeft)break;const p=g.horizontalLine?g.horizontalLine.top?"horizontal-top":"horizontal-bottom":"vertical",_=g.horizontalLine?(e.visibleRangeForPosition(new A(l,g.horizontalLine.endColumn))?.left??f+this._spaceWidth)-f:this._spaceWidth;h+=`<div class="core-guide ${g.className} ${p}" style="left:${f}px;width:${_}px"></div>`}a[c]=h}this._renderResult=a}getGuidesByLine(e,t,i){const n=this._bracketPairGuideOptions.bracketPairs!==!1?this._context.viewModel.getBracketGuidesInRangeByLine(e,t,i,{highlightActive:this._bracketPairGuideOptions.highlightActiveBracketPair,horizontalGuides:this._bracketPairGuideOptions.bracketPairsHorizontal===!0?cp.Enabled:this._bracketPairGuideOptions.bracketPairsHorizontal==="active"?cp.EnabledForActive:cp.Disabled,includeInactive:this._bracketPairGuideOptions.bracketPairs===!0}):null,o=this._bracketPairGuideOptions.indentation?this._context.viewModel.getLinesIndentGuides(e,t):null;let r=0,a=0,l=0;if(this._bracketPairGuideOptions.highlightActiveIndentation!==!1&&i){const h=this._context.viewModel.getActiveIndentGuide(i.lineNumber,e,t);r=h.startLineNumber,a=h.endLineNumber,l=h.indent}const{indentSize:c}=this._context.viewModel.model.getOptions(),d=[];for(let h=e;h<=t;h++){const u=new Array;d.push(u);const g=n?n[h-e]:[],f=new ph(g),p=o?o[h-e]:0;for(let _=1;_<=p;_++){const b=(_-1)*c+1,w=(this._bracketPairGuideOptions.highlightActiveIndentation==="always"||g.length===0)&&r<=h&&h<=a&&_===l;u.push(...f.takeWhile(v=>v.visibleColumn<b)||[]);const C=f.peek();(!C||C.visibleColumn!==b||C.horizontalLine)&&u.push(new Gf(b,-1,`core-guide-indent lvl-${(_-1)%30}`+(w?" indent-active":""),null,-1,-1))}u.push(...f.takeWhile(_=>!0)||[])}return d}render(e,t){if(!this._renderResult)return"";const i=t-e;return i<0||i>=this._renderResult.length?"":this._renderResult[i]}}function H_(s){if(!(s&&s.isTransparent()))return s}ua((s,e)=>{const t=[{bracketColor:pee,guideColor:mwe,guideColorActive:ywe},{bracketColor:mee,guideColor:_we,guideColorActive:Swe},{bracketColor:_ee,guideColor:bwe,guideColorActive:xwe},{bracketColor:bee,guideColor:wwe,guideColorActive:Lwe},{bracketColor:wee,guideColor:vwe,guideColorActive:kwe},{bracketColor:vee,guideColor:Cwe,guideColorActive:Ewe}],i=new kee,n=[{indentColor:ix,indentColorActive:nx},{indentColor:Kbe,indentColorActive:Jbe},{indentColor:Ybe,indentColorActive:ewe},{indentColor:Zbe,indentColorActive:twe},{indentColor:Xbe,indentColorActive:iwe},{indentColor:Qbe,indentColorActive:nwe}],o=t.map(a=>{const l=s.getColor(a.bracketColor),c=s.getColor(a.guideColor),d=s.getColor(a.guideColorActive),h=H_(H_(c)??l?.transparent(.3)),u=H_(H_(d)??l);if(!(!h||!u))return{guideColor:h,guideColorActive:u}}).filter(Ln),r=n.map(a=>{const l=s.getColor(a.indentColor),c=s.getColor(a.indentColorActive),d=H_(l),h=H_(c);if(!(!d||!h))return{indentColor:d,indentColorActive:h}}).filter(Ln);if(o.length>0){for(let a=0;a<30;a++){const l=o[a%o.length];e.addRule(`.monaco-editor .${i.getInlineClassNameOfLevel(a).replace(/ /g,".")} { --guide-color: ${l.guideColor}; --guide-color-active: ${l.guideColorActive}; }`)}e.addRule(".monaco-editor .vertical { box-shadow: 1px 0 0 0 var(--guide-color) inset; }"),e.addRule(".monaco-editor .horizontal-top { border-top: 1px solid var(--guide-color); }"),e.addRule(".monaco-editor .horizontal-bottom { border-bottom: 1px solid var(--guide-color); }"),e.addRule(`.monaco-editor .vertical.${i.activeClassName} { box-shadow: 1px 0 0 0 var(--guide-color-active) inset; }`),e.addRule(`.monaco-editor .horizontal-top.${i.activeClassName} { border-top: 1px solid var(--guide-color-active); }`),e.addRule(`.monaco-editor .horizontal-bottom.${i.activeClassName} { border-bottom: 1px solid var(--guide-color-active); }`)}if(r.length>0){for(let a=0;a<30;a++){const l=r[a%r.length];e.addRule(`.monaco-editor .lines-content .core-guide-indent.lvl-${a} { --indent-color: ${l.indentColor}; --indent-color-active: ${l.indentColorActive}; }`)}e.addRule(".monaco-editor .lines-content .core-guide-indent { box-shadow: 1px 0 0 0 var(--indent-color) inset; }"),e.addRule(".monaco-editor .lines-content .core-guide-indent.indent-active { box-shadow: 1px 0 0 0 var(--indent-color-active) inset; }")}});const XT=class XT extends Dg{constructor(e){super(),this._context=e,this._readConfig(),this._lastCursorModelPosition=new A(1,1),this._renderResult=null,this._activeModelLineNumber=1,this._context.addEventHandler(this)}_readConfig(){const e=this._context.configuration.options;this._lineHeight=e.get(75);const t=e.get(76);this._renderLineNumbers=t.renderType,this._renderCustomLineNumbers=t.renderFn,this._renderFinalNewline=e.get(109);const i=e.get(165);this._lineNumbersLeft=i.lineNumbersLeft,this._lineNumbersWidth=i.lineNumbersWidth}dispose(){this._context.removeEventHandler(this),this._renderResult=null,super.dispose()}onConfigurationChanged(e){return this._readConfig(),!0}onCursorStateChanged(e){const t=e.selections[0].getPosition();this._lastCursorModelPosition=this._context.viewModel.coordinatesConverter.convertViewPositionToModelPosition(t);let i=!1;return this._activeModelLineNumber!==this._lastCursorModelPosition.lineNumber&&(this._activeModelLineNumber=this._lastCursorModelPosition.lineNumber,i=!0),(this._renderLineNumbers===2||this._renderLineNumbers===3)&&(i=!0),i}onFlushed(e){return!0}onLinesChanged(e){return!0}onLinesDeleted(e){return!0}onLinesInserted(e){return!0}onScrollChanged(e){return e.scrollTopChanged}onZonesChanged(e){return!0}onDecorationsChanged(e){return e.affectsLineNumber}_getLineRenderLineNumber(e){const t=this._context.viewModel.coordinatesConverter.convertViewPositionToModelPosition(new A(e,1));if(t.column!==1)return"";const i=t.lineNumber;if(this._renderCustomLineNumbers)return this._renderCustomLineNumbers(i);if(this._renderLineNumbers===2){const n=Math.abs(this._lastCursorModelPosition.lineNumber-i);return n===0?'<span class="relative-current-line-number">'+i+"</span>":String(n)}if(this._renderLineNumbers===3){if(this._lastCursorModelPosition.lineNumber===i||i%10===0)return String(i);const n=this._context.viewModel.getLineCount();return i===n?String(i):""}return String(i)}prepareRender(e){if(this._renderLineNumbers===0){this._renderResult=null;return}const t=vo?this._lineHeight%2===0?" lh-even":" lh-odd":"",i=e.visibleRange.startLineNumber,n=e.visibleRange.endLineNumber,o=this._context.viewModel.getDecorationsInViewport(e.visibleRange).filter(c=>!!c.options.lineNumberClassName);o.sort((c,d)=>k.compareRangesUsingEnds(c.range,d.range));let r=0;const a=this._context.viewModel.getLineCount(),l=[];for(let c=i;c<=n;c++){const d=c-i,h=this._context.viewModel.coordinatesConverter.convertViewPositionToModelPosition(new A(c,1)).lineNumber;let u=this._getLineRenderLineNumber(c),g="";for(;r<o.length&&o[r].range.endLineNumber<c;)r++;for(let f=r;f<o.length;f++){const{range:p,options:_}=o[f];p.startLineNumber<=c&&(g+=" "+_.lineNumberClassName)}if(!u&&!g){l[d]="";continue}c===a&&this._context.viewModel.getLineLength(c)===0&&(this._renderFinalNewline==="off"&&(u=""),this._renderFinalNewline==="dimmed"&&(g+=" dimmed-line-number")),h===this._activeModelLineNumber&&(g+=" active-line-number"),l[d]=`<div class="${XT.CLASS_NAME}${t}${g}" style="left:${this._lineNumbersLeft}px;width:${this._lineNumbersWidth}px;">${u}</div>`}this._renderResult=l}render(e,t){if(!this._renderResult)return"";const i=t-e;return i<0||i>=this._renderResult.length?"":this._renderResult[i]}};XT.CLASS_NAME="line-numbers";let JE=XT;ua((s,e)=>{const t=s.getColor(qbe),i=s.getColor(owe);i?e.addRule(`.monaco-editor .line-numbers.dimmed-line-number { color: ${i}; }`):t&&e.addRule(`.monaco-editor .line-numbers.dimmed-line-number { color: ${t.transparent(.4)}; }`)});const pw="monaco-mouse-cursor-text";class sM{get didDomLayout(){return this._didDomLayout}readClientRect(){if(!this._clientRectRead){this._clientRectRead=!0;const e=this._domNode.getBoundingClientRect();this.markDidDomLayout(),this._clientRectDeltaLeft=e.left,this._clientRectScale=e.width/this._domNode.offsetWidth}}get clientRectDeltaLeft(){return this._clientRectRead||this.readClientRect(),this._clientRectDeltaLeft}get clientRectScale(){return this._clientRectRead||this.readClientRect(),this._clientRectScale}constructor(e,t){this._domNode=e,this.endNode=t,this._didDomLayout=!1,this._clientRectDeltaLeft=0,this._clientRectScale=1,this._clientRectRead=!1}markDidDomLayout(){this._didDomLayout=!0}}class eI{constructor(e,t){this.themeType=t;const i=e.options,n=i.get(59);this.renderWhitespace=i.get(113),this.experimentalWhitespaceRendering=i.get(47),this.renderControlCharacters=i.get(108),this.spaceWidth=n.spaceWidth,this.middotWidth=n.middotWidth,this.wsmiddotWidth=n.wsmiddotWidth,this.useMonospaceOptimizations=n.isMonospace&&!i.get(40),this.canUseHalfwidthRightwardsArrow=n.canUseHalfwidthRightwardsArrow,this.lineHeight=i.get(75),this.stopRenderingLineAfter=i.get(133),this.fontLigatures=i.get(60),this.verticalScrollbarSize=i.get(117).verticalScrollbarSize,this.useGpu=i.get(46)==="on"}equals(e){return this.themeType===e.themeType&&this.renderWhitespace===e.renderWhitespace&&this.experimentalWhitespaceRendering===e.experimentalWhitespaceRendering&&this.renderControlCharacters===e.renderControlCharacters&&this.spaceWidth===e.spaceWidth&&this.middotWidth===e.middotWidth&&this.wsmiddotWidth===e.wsmiddotWidth&&this.useMonospaceOptimizations===e.useMonospaceOptimizations&&this.canUseHalfwidthRightwardsArrow===e.canUseHalfwidthRightwardsArrow&&this.lineHeight===e.lineHeight&&this.stopRenderingLineAfter===e.stopRenderingLineAfter&&this.fontLigatures===e.fontLigatures&&this.verticalScrollbarSize===e.verticalScrollbarSize&&this.useGpu===e.useGpu}}class Uwe{constructor(){this._currentVisibleRange=new k(1,1,1,1)}getCurrentVisibleRange(){return this._currentVisibleRange}setCurrentVisibleRange(e){this._currentVisibleRange=e}}class qwe{constructor(e,t,i,n,o,r,a){this.minimalReveal=e,this.lineNumber=t,this.startColumn=i,this.endColumn=n,this.startScrollTop=o,this.stopScrollTop=r,this.scrollType=a,this.type="range",this.minLineNumber=t,this.maxLineNumber=t}}class jwe{constructor(e,t,i,n,o){this.minimalReveal=e,this.selections=t,this.startScrollTop=i,this.stopScrollTop=n,this.scrollType=o,this.type="selections";let r=t[0].startLineNumber,a=t[0].endLineNumber;for(let l=1,c=t.length;l<c;l++){const d=t[l];r=Math.min(r,d.startLineNumber),a=Math.max(a,d.endLineNumber)}this.minLineNumber=r,this.maxLineNumber=a}}const QT=class QT extends no{constructor(e,t,i){super(e);const n=this._context.configuration,o=this._context.configuration.options,r=o.get(59),a=o.get(166);this._lineHeight=o.get(75),this._typicalHalfwidthCharacterWidth=r.typicalHalfwidthCharacterWidth,this._isViewportWrapping=a.isViewportWrapping,this._revealHorizontalRightPadding=o.get(114),this._cursorSurroundingLines=o.get(35),this._cursorSurroundingLinesStyle=o.get(36),this._canUseLayerHinting=!o.get(39),this._viewLineOptions=new eI(n,this._context.theme.type),this._linesContent=i,this._textRangeRestingSpot=document.createElement("div"),this._visibleLines=new aee(this._context,{createLine:()=>new Od(t,this._viewLineOptions)}),this.domNode=this._visibleLines.domNode,kl.write(this.domNode,8),this.domNode.setClassName(`view-lines ${pw}`),In(this.domNode,r),this._maxLineWidth=0,this._asyncUpdateLineWidths=new Rt(()=>{this._updateLineWidthsSlow()},200),this._asyncCheckMonospaceFontAssumptions=new Rt(()=>{this._checkMonospaceFontAssumptions()},2e3),this._lastRenderedData=new Uwe,this._horizontalRevealRequest=null,this._stickyScrollEnabled=o.get(131).enabled,this._maxNumberStickyLines=o.get(131).maxLineCount}dispose(){this._asyncUpdateLineWidths.dispose(),this._asyncCheckMonospaceFontAssumptions.dispose(),super.dispose()}getDomNode(){return this.domNode}onConfigurationChanged(e){this._visibleLines.onConfigurationChanged(e),e.hasChanged(166)&&(this._maxLineWidth=0);const t=this._context.configuration.options,i=t.get(59),n=t.get(166);return this._lineHeight=t.get(75),this._typicalHalfwidthCharacterWidth=i.typicalHalfwidthCharacterWidth,this._isViewportWrapping=n.isViewportWrapping,this._revealHorizontalRightPadding=t.get(114),this._cursorSurroundingLines=t.get(35),this._cursorSurroundingLinesStyle=t.get(36),this._canUseLayerHinting=!t.get(39),this._stickyScrollEnabled=t.get(131).enabled,this._maxNumberStickyLines=t.get(131).maxLineCount,In(this.domNode,i),this._onOptionsMaybeChanged(),e.hasChanged(165)&&(this._maxLineWidth=0),!0}_onOptionsMaybeChanged(){const e=this._context.configuration,t=new eI(e,this._context.theme.type);if(!this._viewLineOptions.equals(t)){this._viewLineOptions=t;const i=this._visibleLines.getStartLineNumber(),n=this._visibleLines.getEndLineNumber();for(let o=i;o<=n;o++)this._visibleLines.getVisibleLine(o).onOptionsChanged(this._viewLineOptions);return!0}return!1}onCursorStateChanged(e){const t=this._visibleLines.getStartLineNumber(),i=this._visibleLines.getEndLineNumber();let n=!1;for(let o=t;o<=i;o++)n=this._visibleLines.getVisibleLine(o).onSelectionChanged()||n;return n}onDecorationsChanged(e){{const t=this._visibleLines.getStartLineNumber(),i=this._visibleLines.getEndLineNumber();for(let n=t;n<=i;n++)this._visibleLines.getVisibleLine(n).onDecorationsChanged()}return!0}onFlushed(e){const t=this._visibleLines.onFlushed(e,this._viewLineOptions.useGpu);return this._maxLineWidth=0,t}onLinesChanged(e){return this._visibleLines.onLinesChanged(e)}onLinesDeleted(e){return this._visibleLines.onLinesDeleted(e)}onLinesInserted(e){return this._visibleLines.onLinesInserted(e)}onRevealRangeRequest(e){const t=this._computeScrollTopToRevealRange(this._context.viewLayout.getFutureViewport(),e.source,e.minimalReveal,e.range,e.selections,e.verticalType);if(t===-1)return!1;let i=this._context.viewLayout.validateScrollPosition({scrollTop:t});e.revealHorizontal?e.range&&e.range.startLineNumber!==e.range.endLineNumber?i={scrollTop:i.scrollTop,scrollLeft:0}:e.range?this._horizontalRevealRequest=new qwe(e.minimalReveal,e.range.startLineNumber,e.range.startColumn,e.range.endColumn,this._context.viewLayout.getCurrentScrollTop(),i.scrollTop,e.scrollType):e.selections&&e.selections.length>0&&(this._horizontalRevealRequest=new jwe(e.minimalReveal,e.selections,this._context.viewLayout.getCurrentScrollTop(),i.scrollTop,e.scrollType)):this._horizontalRevealRequest=null;const o=Math.abs(this._context.viewLayout.getCurrentScrollTop()-i.scrollTop)<=this._lineHeight?1:e.scrollType;return this._context.viewModel.viewLayout.setScrollPosition(i,o),!0}onScrollChanged(e){if(this._horizontalRevealRequest&&e.scrollLeftChanged&&(this._horizontalRevealRequest=null),this._horizontalRevealRequest&&e.scrollTopChanged){const t=Math.min(this._horizontalRevealRequest.startScrollTop,this._horizontalRevealRequest.stopScrollTop),i=Math.max(this._horizontalRevealRequest.startScrollTop,this._horizontalRevealRequest.stopScrollTop);(e.scrollTop<t||e.scrollTop>i)&&(this._horizontalRevealRequest=null)}return this.domNode.setWidth(e.scrollWidth),this._visibleLines.onScrollChanged(e)||!0}onTokensChanged(e){return this._visibleLines.onTokensChanged(e)}onZonesChanged(e){return this._context.viewModel.viewLayout.setMaxLineWidth(this._maxLineWidth),this._visibleLines.onZonesChanged(e)}onThemeChanged(e){return this._onOptionsMaybeChanged()}getPositionFromDOMInfo(e,t){const i=this._getViewLineDomNode(e);if(i===null)return null;const n=this._getLineNumberFor(i);if(n===-1||n<1||n>this._context.viewModel.getLineCount())return null;if(this._context.viewModel.getLineMaxColumn(n)===1)return new A(n,1);const o=this._visibleLines.getStartLineNumber(),r=this._visibleLines.getEndLineNumber();if(n<o||n>r)return null;let a=this._visibleLines.getVisibleLine(n).getColumnOfNodeOffset(e,t);const l=this._context.viewModel.getLineMinColumn(n);return a<l&&(a=l),new A(n,a)}_getViewLineDomNode(e){for(;e&&e.nodeType===1;){if(e.className===Od.CLASS_NAME)return e;e=e.parentElement}return null}_getLineNumberFor(e){const t=this._visibleLines.getStartLineNumber(),i=this._visibleLines.getEndLineNumber();for(let n=t;n<=i;n++){const o=this._visibleLines.getVisibleLine(n);if(e===o.getDomNode())return n}return-1}getLineWidth(e){const t=this._visibleLines.getStartLineNumber(),i=this._visibleLines.getEndLineNumber();if(e<t||e>i)return-1;const n=new sM(this.domNode.domNode,this._textRangeRestingSpot),o=this._visibleLines.getVisibleLine(e).getWidth(n);return this._updateLineWidthsSlowIfDomDidLayout(n),o}linesVisibleRangesForRange(e,t){if(this.shouldRender())return null;const i=e.endLineNumber,n=k.intersectRanges(e,this._lastRenderedData.getCurrentVisibleRange());if(!n)return null;const o=[];let r=0;const a=new sM(this.domNode.domNode,this._textRangeRestingSpot);let l=0;t&&(l=this._context.viewModel.coordinatesConverter.convertViewPositionToModelPosition(new A(n.startLineNumber,1)).lineNumber);const c=this._visibleLines.getStartLineNumber(),d=this._visibleLines.getEndLineNumber();for(let h=n.startLineNumber;h<=n.endLineNumber;h++){if(h<c||h>d)continue;const u=h===n.startLineNumber?n.startColumn:1,g=h!==i,f=g?this._context.viewModel.getLineMaxColumn(h):n.endColumn,_=this._visibleLines.getVisibleLine(h).getVisibleRangesForRange(h,u,f,a);if(_){if(t&&h<i){const b=l;if(l=this._context.viewModel.coordinatesConverter.convertViewPositionToModelPosition(new A(h+1,1)).lineNumber,b!==l){const w=_.ranges[_.ranges.length-1];w.width+=this._typicalHalfwidthCharacterWidth,this._context.viewModel.getTextDirection(b)===br.RTL&&(w.left-=this._typicalHalfwidthCharacterWidth)}}o[r++]=new GJ(_.outsideRenderedLine,h,ex.from(_.ranges),g)}}return this._updateLineWidthsSlowIfDomDidLayout(a),r===0?null:o}_visibleRangesForLineRange(e,t,i){if(this.shouldRender()||e<this._visibleLines.getStartLineNumber()||e>this._visibleLines.getEndLineNumber())return null;const n=new sM(this.domNode.domNode,this._textRangeRestingSpot),o=this._visibleLines.getVisibleLine(e).getVisibleRangesForRange(e,t,i,n);return this._updateLineWidthsSlowIfDomDidLayout(n),o}_lineIsRenderedRTL(e){return e<this._visibleLines.getStartLineNumber()||e>this._visibleLines.getEndLineNumber()?!1:this._visibleLines.getVisibleLine(e).isRenderedRTL()}visibleRangeForPosition(e){const t=this._visibleRangesForLineRange(e.lineNumber,e.column,e.column);return t?new KJ(t.outsideRenderedLine,t.ranges[0].left):null}_updateLineWidthsFast(){return this._updateLineWidths(!0)}_updateLineWidthsSlow(){this._updateLineWidths(!1)}_updateLineWidthsSlowIfDomDidLayout(e){e.didDomLayout&&(this._asyncUpdateLineWidths.isScheduled()||(this._asyncUpdateLineWidths.cancel(),this._updateLineWidthsSlow()))}_updateLineWidths(e){const t=this._visibleLines.getStartLineNumber(),i=this._visibleLines.getEndLineNumber();let n=1,o=!0;for(let r=t;r<=i;r++){const a=this._visibleLines.getVisibleLine(r);if(e&&!a.getWidthIsFast()){o=!1;continue}n=Math.max(n,a.getWidth(null))}return o&&t===1&&i===this._context.viewModel.getLineCount()&&(this._maxLineWidth=0),this._ensureMaxLineWidth(n),o}_checkMonospaceFontAssumptions(){let e=-1,t=-1;const i=this._visibleLines.getStartLineNumber(),n=this._visibleLines.getEndLineNumber();for(let o=i;o<=n;o++){const r=this._visibleLines.getVisibleLine(o);if(r.needsMonospaceFontCheck()){const a=r.getWidth(null);a>t&&(t=a,e=o)}}if(e!==-1&&!this._visibleLines.getVisibleLine(e).monospaceAssumptionsAreValid())for(let o=i;o<=n;o++)this._visibleLines.getVisibleLine(o).onMonospaceAssumptionsInvalidated()}prepareRender(){throw new Error("Not supported")}render(){throw new Error("Not supported")}renderText(e){if(this._visibleLines.renderLines(e),this._lastRenderedData.setCurrentVisibleRange(e.visibleRange),this.domNode.setWidth(this._context.viewLayout.getScrollWidth()),this.domNode.setHeight(Math.min(this._context.viewLayout.getScrollHeight(),1e6)),this._horizontalRevealRequest){const i=this._horizontalRevealRequest;if(e.startLineNumber<=i.minLineNumber&&i.maxLineNumber<=e.endLineNumber){this._horizontalRevealRequest=null,this.onDidRender();const n=this._computeScrollLeftToReveal(i);n&&(!this._isViewportWrapping&&!n.hasRTL&&this._ensureMaxLineWidth(n.maxHorizontalOffset),this._context.viewModel.viewLayout.setScrollPosition({scrollLeft:n.scrollLeft},i.scrollType))}}if(this._updateLineWidthsFast()?this._asyncUpdateLineWidths.cancel():this._asyncUpdateLineWidths.schedule(),vo&&!this._asyncCheckMonospaceFontAssumptions.isScheduled()){const i=this._visibleLines.getStartLineNumber(),n=this._visibleLines.getEndLineNumber();for(let o=i;o<=n;o++)if(this._visibleLines.getVisibleLine(o).needsMonospaceFontCheck()){this._asyncCheckMonospaceFontAssumptions.schedule();break}}this._linesContent.setLayerHinting(this._canUseLayerHinting),this._linesContent.setContain("strict");const t=this._context.viewLayout.getCurrentScrollTop()-e.bigNumbersDelta;this._linesContent.setTop(-t),this._linesContent.setLeft(-this._context.viewLayout.getCurrentScrollLeft())}_ensureMaxLineWidth(e){const t=Math.ceil(e);this._maxLineWidth<t&&(this._maxLineWidth=t,this._context.viewModel.viewLayout.setMaxLineWidth(this._maxLineWidth))}_computeScrollTopToRevealRange(e,t,i,n,o,r){const a=e.top,l=e.height,c=a+l;let d,h,u;if(o&&o.length>0){let b=o[0].startLineNumber,w=o[0].endLineNumber;for(let C=1,v=o.length;C<v;C++){const y=o[C];b=Math.min(b,y.startLineNumber),w=Math.max(w,y.endLineNumber)}d=!1,h=this._context.viewLayout.getVerticalOffsetForLineNumber(b),u=this._context.viewLayout.getVerticalOffsetForLineNumber(w)+this._lineHeight}else if(n)d=!0,h=this._context.viewLayout.getVerticalOffsetForLineNumber(n.startLineNumber),u=this._context.viewLayout.getVerticalOffsetForLineNumber(n.endLineNumber)+this._lineHeight;else return-1;const g=(t==="mouse"||i)&&this._cursorSurroundingLinesStyle==="default";let f=0,p=0;if(g)i||(f=this._lineHeight);else{const b=l/this._lineHeight,w=Math.max(this._cursorSurroundingLines,this._stickyScrollEnabled?this._maxNumberStickyLines:0),C=Math.min(b/2,w);f=C*this._lineHeight,p=Math.max(0,C-1)*this._lineHeight}i||(r===0||r===4)&&(p+=this._lineHeight),h-=f,u+=p;let _;if(u-h>l){if(!d)return-1;_=h}else if(r===5||r===6)if(r===6&&a<=h&&u<=c)_=a;else{const b=Math.max(5*this._lineHeight,l*.2),w=h-b,C=u-l;_=Math.max(C,w)}else if(r===1||r===2)if(r===2&&a<=h&&u<=c)_=a;else{const b=(h+u)/2;_=Math.max(0,b-l/2)}else _=this._computeMinimumScrolling(a,c,h,u,r===3,r===4);return _}_computeScrollLeftToReveal(e){const t=this._context.viewLayout.getCurrentViewport(),i=this._context.configuration.options.get(165),n=t.left,o=n+t.width-i.verticalScrollbarWidth;let r=1073741824,a=0,l=!1;if(e.type==="range"){l=this._lineIsRenderedRTL(e.lineNumber);const d=this._visibleRangesForLineRange(e.lineNumber,e.startColumn,e.endColumn);if(!d)return null;for(const h of d.ranges)r=Math.min(r,Math.round(h.left)),a=Math.max(a,Math.round(h.left+h.width))}else for(const d of e.selections){if(d.startLineNumber!==d.endLineNumber)return null;const h=this._visibleRangesForLineRange(d.startLineNumber,d.startColumn,d.endColumn);if(l||=this._lineIsRenderedRTL(d.startLineNumber),!h)return null;for(const u of h.ranges)r=Math.min(r,Math.round(u.left)),a=Math.max(a,Math.round(u.left+u.width))}return e.minimalReveal||(r=Math.max(0,r-QT.HORIZONTAL_EXTRA_PX),a+=this._revealHorizontalRightPadding),e.type==="selections"&&a-r>t.width?null:{scrollLeft:this._computeMinimumScrolling(n,o,r,a),maxHorizontalOffset:a,hasRTL:l}}_computeMinimumScrolling(e,t,i,n,o,r){e=e|0,t=t|0,i=i|0,n=n|0,o=!!o,r=!!r;const a=t-e;if(n-i<a){if(o)return i;if(r)return Math.max(0,n-a);if(i<e)return i;if(n>t)return Math.max(0,n-a)}else return i;return e}};QT.HORIZONTAL_EXTRA_PX=30;let lF=QT;class Gwe extends yee{constructor(e){super(),this._context=e;const i=this._context.configuration.options.get(165);this._decorationsLeft=i.decorationsLeft,this._decorationsWidth=i.decorationsWidth,this._renderResult=null,this._context.addEventHandler(this)}dispose(){this._context.removeEventHandler(this),this._renderResult=null,super.dispose()}onConfigurationChanged(e){const i=this._context.configuration.options.get(165);return this._decorationsLeft=i.decorationsLeft,this._decorationsWidth=i.decorationsWidth,!0}onDecorationsChanged(e){return!0}onFlushed(e){return!0}onLinesChanged(e){return!0}onLinesDeleted(e){return!0}onLinesInserted(e){return!0}onScrollChanged(e){return e.scrollTopChanged}onZonesChanged(e){return!0}_getDecorations(e){const t=e.getDecorationsInViewport(),i=[];let n=0;for(let o=0,r=t.length;o<r;o++){const a=t[o],l=a.options.linesDecorationsClassName,c=a.options.zIndex;l&&(i[n++]=new aF(a.range.startLineNumber,a.range.endLineNumber,l,a.options.linesDecorationsTooltip??null,c));const d=a.options.firstLineDecorationClassName;d&&(i[n++]=new aF(a.range.startLineNumber,a.range.startLineNumber,d,a.options.linesDecorationsTooltip??null,c))}return i}prepareRender(e){const t=e.visibleRange.startLineNumber,i=e.visibleRange.endLineNumber,n=this._render(t,i,this._getDecorations(e)),o=this._decorationsLeft.toString(),r=this._decorationsWidth.toString(),a='" style="left:'+o+"px;width:"+r+'px;"></div>',l=[];for(let c=t;c<=i;c++){const d=c-t,h=n[d].getDecorations();let u="";for(const g of h){let f='<div class="cldr '+g.className;g.tooltip!==null&&(f+='" title="'+g.tooltip),f+=a,u+=f}l[d]=u}this._renderResult=l}render(e,t){return this._renderResult?this._renderResult[t-e]:""}}const Zb=class Zb extends no{constructor(e){super(e);const t=this._context.configuration.options,i=t.get(165);this._canUseLayerHinting=!t.get(39),this._contentLeft=i.contentLeft,this._glyphMarginLeft=i.glyphMarginLeft,this._glyphMarginWidth=i.glyphMarginWidth,this._domNode=wt(document.createElement("div")),this._domNode.setClassName(Zb.OUTER_CLASS_NAME),this._domNode.setPosition("absolute"),this._domNode.setAttribute("role","presentation"),this._domNode.setAttribute("aria-hidden","true"),this._glyphMarginBackgroundDomNode=wt(document.createElement("div")),this._glyphMarginBackgroundDomNode.setClassName(Zb.CLASS_NAME),this._domNode.appendChild(this._glyphMarginBackgroundDomNode)}dispose(){super.dispose()}getDomNode(){return this._domNode}onConfigurationChanged(e){const t=this._context.configuration.options,i=t.get(165);return this._canUseLayerHinting=!t.get(39),this._contentLeft=i.contentLeft,this._glyphMarginLeft=i.glyphMarginLeft,this._glyphMarginWidth=i.glyphMarginWidth,!0}onScrollChanged(e){return super.onScrollChanged(e)||e.scrollTopChanged}prepareRender(e){}render(e){this._domNode.setLayerHinting(this._canUseLayerHinting),this._domNode.setContain("strict");const t=e.scrollTop-e.bigNumbersDelta;this._domNode.setTop(-t);const i=Math.min(e.scrollHeight,1e6);this._domNode.setHeight(i),this._domNode.setWidth(this._contentLeft),this._glyphMarginBackgroundDomNode.setLeft(this._glyphMarginLeft),this._glyphMarginBackgroundDomNode.setWidth(this._glyphMarginWidth),this._glyphMarginBackgroundDomNode.setHeight(i)}};Zb.CLASS_NAME="glyph-margin",Zb.OUTER_CLASS_NAME="margin";let tI=Zb;class Kwe extends yee{constructor(e){super(),this._context=e,this._renderResult=null,this._context.addEventHandler(this)}dispose(){this._context.removeEventHandler(this),this._renderResult=null,super.dispose()}onConfigurationChanged(e){return!0}onDecorationsChanged(e){return!0}onFlushed(e){return!0}onLinesChanged(e){return!0}onLinesDeleted(e){return!0}onLinesInserted(e){return!0}onScrollChanged(e){return e.scrollTopChanged}onZonesChanged(e){return!0}_getDecorations(e){const t=e.getDecorationsInViewport(),i=[];let n=0;for(let o=0,r=t.length;o<r;o++){const a=t[o],l=a.options.marginClassName,c=a.options.zIndex;l&&(i[n++]=new aF(a.range.startLineNumber,a.range.endLineNumber,l,null,c))}return i}prepareRender(e){const t=e.visibleRange.startLineNumber,i=e.visibleRange.endLineNumber,n=this._render(t,i,this._getDecorations(e)),o=[];for(let r=t;r<=i;r++){const a=r-t,l=n[a].getDecorations();let c="";for(const d of l)c+='<div class="cmdr '+d.className+'" style=""></div>';o[a]=c}this._renderResult=o}render(e,t){return this._renderResult?this._renderResult[t-e]:""}}const Xh=class Xh{constructor(e,t,i,n){this._rgba8Brand=void 0,this.r=Xh._clamp(e),this.g=Xh._clamp(t),this.b=Xh._clamp(i),this.a=Xh._clamp(n)}equals(e){return this.r===e.r&&this.g===e.g&&this.b===e.b&&this.a===e.a}static _clamp(e){return e<0?0:e>255?255:e|0}};Xh.Empty=new Xh(0,0,0,0);let nh=Xh,Ywe=class{constructor(){this._tokenizationSupports=new Map,this._factories=new Map,this._onDidChange=new M,this.onDidChange=this._onDidChange.event,this._colorMap=null}handleChange(e){this._onDidChange.fire({changedLanguages:e,changedColorMap:!1})}register(e,t){return this._tokenizationSupports.set(e,t),this.handleChange([e]),pe(()=>{this._tokenizationSupports.get(e)===t&&(this._tokenizationSupports.delete(e),this.handleChange([e]))})}get(e){return this._tokenizationSupports.get(e)||null}registerFactory(e,t){this._factories.get(e)?.dispose();const i=new Zwe(this,e,t);return this._factories.set(e,i),pe(()=>{const n=this._factories.get(e);!n||n!==i||(this._factories.delete(e),n.dispose())})}async getOrCreate(e){const t=this.get(e);if(t)return t;const i=this._factories.get(e);return!i||i.isResolved?null:(await i.resolve(),this.get(e))}isResolved(e){if(this.get(e))return!0;const i=this._factories.get(e);return!!(!i||i.isResolved)}setColorMap(e){this._colorMap=e,this._onDidChange.fire({changedLanguages:Array.from(this._tokenizationSupports.keys()),changedColorMap:!0})}getColorMap(){return this._colorMap}getDefaultBackground(){return this._colorMap&&this._colorMap.length>2?this._colorMap[2]:null}};class Zwe extends F{get isResolved(){return this._isResolved}constructor(e,t,i){super(),this._registry=e,this._languageId=t,this._factory=i,this._isDisposed=!1,this._resolvePromise=null,this._isResolved=!1}dispose(){this._isDisposed=!0,super.dispose()}async resolve(){return this._resolvePromise||(this._resolvePromise=this._create()),this._resolvePromise}async _create(){const e=await this._factory.tokenizationSupport;this._isResolved=!0,e&&!this._isDisposed&&this._register(this._registry.register(this._languageId,e))}}let pS=class{constructor(e,t,i){this.offset=e,this.type=t,this.language=i,this._tokenBrand=void 0}toString(){return"("+this.offset+", "+this.type+")"}};class GW{constructor(e,t){this.tokens=e,this.endState=t,this._tokenizationResultBrand=void 0}}class gA{constructor(e,t){this.tokens=e,this.endState=t,this._encodedTokenizationResultBrand=void 0}}var Ao;(function(s){s[s.Increase=0]="Increase",s[s.Decrease=1]="Decrease"})(Ao||(Ao={}));var mv;(function(s){const e=new Map;e.set(0,Z.symbolMethod),e.set(1,Z.symbolFunction),e.set(2,Z.symbolConstructor),e.set(3,Z.symbolField),e.set(4,Z.symbolVariable),e.set(5,Z.symbolClass),e.set(6,Z.symbolStruct),e.set(7,Z.symbolInterface),e.set(8,Z.symbolModule),e.set(9,Z.symbolProperty),e.set(10,Z.symbolEvent),e.set(11,Z.symbolOperator),e.set(12,Z.symbolUnit),e.set(13,Z.symbolValue),e.set(15,Z.symbolEnum),e.set(14,Z.symbolConstant),e.set(15,Z.symbolEnum),e.set(16,Z.symbolEnumMember),e.set(17,Z.symbolKeyword),e.set(28,Z.symbolSnippet),e.set(18,Z.symbolText),e.set(19,Z.symbolColor),e.set(20,Z.symbolFile),e.set(21,Z.symbolReference),e.set(22,Z.symbolCustomColor),e.set(23,Z.symbolFolder),e.set(24,Z.symbolTypeParameter),e.set(25,Z.account),e.set(26,Z.issues),e.set(27,Z.tools);function t(r){let a=e.get(r);return a||(console.info("No codicon found for CompletionItemKind "+r),a=Z.symbolProperty),a}s.toIcon=t;function i(r){switch(r){case 0:return m(728,"Method");case 1:return m(729,"Function");case 2:return m(730,"Constructor");case 3:return m(731,"Field");case 4:return m(732,"Variable");case 5:return m(733,"Class");case 6:return m(734,"Struct");case 7:return m(735,"Interface");case 8:return m(736,"Module");case 9:return m(737,"Property");case 10:return m(738,"Event");case 11:return m(739,"Operator");case 12:return m(740,"Unit");case 13:return m(741,"Value");case 14:return m(742,"Constant");case 15:return m(743,"Enum");case 16:return m(744,"Enum Member");case 17:return m(745,"Keyword");case 18:return m(746,"Text");case 19:return m(747,"Color");case 20:return m(748,"File");case 21:return m(749,"Reference");case 22:return m(750,"Custom Color");case 23:return m(751,"Folder");case 24:return m(752,"Type Parameter");case 25:return m(753,"User");case 26:return m(754,"Issue");case 27:return m(755,"Tool");case 28:return m(756,"Snippet");default:return""}}s.toLabel=i;const n=new Map;n.set("method",0),n.set("function",1),n.set("constructor",2),n.set("field",3),n.set("variable",4),n.set("class",5),n.set("struct",6),n.set("interface",7),n.set("module",8),n.set("property",9),n.set("event",10),n.set("operator",11),n.set("unit",12),n.set("value",13),n.set("constant",14),n.set("enum",15),n.set("enum-member",16),n.set("enumMember",16),n.set("keyword",17),n.set("snippet",28),n.set("text",18),n.set("color",19),n.set("file",20),n.set("reference",21),n.set("customcolor",22),n.set("folder",23),n.set("type-parameter",24),n.set("typeParameter",24),n.set("account",25),n.set("issue",26),n.set("tool",27);function o(r,a){let l=n.get(r);return typeof l>"u"&&!a&&(l=9),l}s.fromString=o})(mv||(mv={}));var ho;(function(s){s[s.Automatic=0]="Automatic",s[s.Explicit=1]="Explicit"})(ho||(ho={}));class Eee{constructor(e,t,i,n){this.range=e,this.text=t,this.completionKind=i,this.isSnippetText=n}equals(e){return k.lift(this.range).equalsRange(e.range)&&this.text===e.text&&this.completionKind===e.completionKind&&this.isSnippetText===e.isSnippetText}}var cF;(function(s){s[s.Code=1]="Code",s[s.Label=2]="Label"})(cF||(cF={}));class fA{static fromExtensionId(e){return new fA(e,void 0,void 0)}constructor(e,t,i){this.extensionId=e,this.extensionVersion=t,this.providerId=i}toString(){let e="";return this.extensionId&&(e+=this.extensionId),this.extensionVersion&&(e+=`@${this.extensionVersion}`),this.providerId&&(e+=`:${this.providerId}`),e.length===0&&(e="unknown"),e}toStringWithoutVersion(){let e="";return this.extensionId&&(e+=this.extensionId),this.providerId&&(e+=`:${this.providerId}`),e}}var Dm;(function(s){s[s.Accepted=0]="Accepted",s[s.Rejected=1]="Rejected",s[s.Ignored=2]="Ignored"})(Dm||(Dm={}));var mS;(function(s){s[s.Automatic=0]="Automatic",s[s.PasteAs=1]="PasteAs"})(mS||(mS={}));var Lc;(function(s){s[s.Invoke=1]="Invoke",s[s.TriggerCharacter=2]="TriggerCharacter",s[s.ContentChange=3]="ContentChange"})(Lc||(Lc={}));var _v;(function(s){s[s.Text=0]="Text",s[s.Read=1]="Read",s[s.Write=2]="Write"})(_v||(_v={}));function Xwe(s){return!!s&&Le.isUri(s.uri)&&k.isIRange(s.range)&&(k.isIRange(s.originSelectionRange)||k.isIRange(s.targetSelectionRange))}const Qwe={17:m(757,"array"),16:m(758,"boolean"),4:m(759,"class"),13:m(760,"constant"),8:m(761,"constructor"),9:m(762,"enumeration"),21:m(763,"enumeration member"),23:m(764,"event"),7:m(765,"field"),0:m(766,"file"),11:m(767,"function"),10:m(768,"interface"),19:m(769,"key"),5:m(770,"method"),1:m(771,"module"),2:m(772,"namespace"),20:m(773,"null"),15:m(774,"number"),18:m(775,"object"),24:m(776,"operator"),3:m(777,"package"),6:m(778,"property"),14:m(779,"string"),22:m(780,"struct"),25:m(781,"type parameter"),12:m(782,"variable")};function Jwe(s,e){return m(783,"{0} ({1})",s,Qwe[e])}var iI;(function(s){const e=new Map;e.set(0,Z.symbolFile),e.set(1,Z.symbolModule),e.set(2,Z.symbolNamespace),e.set(3,Z.symbolPackage),e.set(4,Z.symbolClass),e.set(5,Z.symbolMethod),e.set(6,Z.symbolProperty),e.set(7,Z.symbolField),e.set(8,Z.symbolConstructor),e.set(9,Z.symbolEnum),e.set(10,Z.symbolInterface),e.set(11,Z.symbolFunction),e.set(12,Z.symbolVariable),e.set(13,Z.symbolConstant),e.set(14,Z.symbolString),e.set(15,Z.symbolNumber),e.set(16,Z.symbolBoolean),e.set(17,Z.symbolArray),e.set(18,Z.symbolObject),e.set(19,Z.symbolKey),e.set(20,Z.symbolNull),e.set(21,Z.symbolEnumMember),e.set(22,Z.symbolStruct),e.set(23,Z.symbolEvent),e.set(24,Z.symbolOperator),e.set(25,Z.symbolTypeParameter);function t(o){let r=e.get(o);return r||(console.info("No codicon found for SymbolKind "+o),r=Z.symbolProperty),r}s.toIcon=t;const i=new Map;i.set(0,20),i.set(1,8),i.set(2,8),i.set(3,8),i.set(4,5),i.set(5,0),i.set(6,9),i.set(7,3),i.set(8,2),i.set(9,15),i.set(10,7),i.set(11,1),i.set(12,4),i.set(13,14),i.set(14,18),i.set(15,13),i.set(16,13),i.set(17,13),i.set(18,13),i.set(19,17),i.set(20,13),i.set(21,16),i.set(22,6),i.set(23,10),i.set(24,11),i.set(25,24);function n(o){let r=i.get(o);return r===void 0&&(console.info("No completion kind found for SymbolKind "+o),r=20),r}s.toCompletionKind=n})(iI||(iI={}));var No;let h_=(No=class{static fromValue(e){switch(e){case"comment":return No.Comment;case"imports":return No.Imports;case"region":return No.Region}return new No(e)}constructor(e){this.value=e}},No.Comment=new No("comment"),No.Imports=new No("imports"),No.Region=new No("region"),No);var dF;(function(s){s[s.AIGenerated=1]="AIGenerated"})(dF||(dF={}));var _S;(function(s){s[s.Invoke=0]="Invoke",s[s.Automatic=1]="Automatic"})(_S||(_S={}));var hF;(function(s){function e(t){return!t||typeof t!="object"?!1:typeof t.id=="string"&&typeof t.title=="string"}s.is=e})(hF||(hF={}));var nI;(function(s){s[s.Type=1]="Type",s[s.Parameter=2]="Parameter"})(nI||(nI={}));class eve{constructor(e){this.createSupport=e,this._tokenizationSupport=null}dispose(){this._tokenizationSupport&&this._tokenizationSupport.then(e=>{e&&e.dispose()})}get tokenizationSupport(){return this._tokenizationSupport||(this._tokenizationSupport=this.createSupport()),this._tokenizationSupport}}const yi=new Ywe,JT=class JT extends F{static getInstance(){return this._INSTANCE||(this._INSTANCE=new JT),this._INSTANCE}constructor(){super(),this._onDidChange=new M,this.onDidChange=this._onDidChange.event,this._updateColorMap(),this._register(yi.onDidChange(e=>{e.changedColorMap&&this._updateColorMap()}))}_updateColorMap(){const e=yi.getColorMap();if(!e){this._colors=[nh.Empty],this._backgroundIsLight=!0;return}this._colors=[nh.Empty];for(let i=1;i<e.length;i++){const n=e[i].rgba;this._colors[i]=new nh(n.r,n.g,n.b,Math.round(n.a*255))}const t=e[2].getRelativeLuminance();this._backgroundIsLight=t>=.5,this._onDidChange.fire(void 0)}getColor(e){return(e<1||e>=this._colors.length)&&(e=2),this._colors[e]}backgroundIsLight(){return this._backgroundIsLight}};JT._INSTANCE=null;let sI=JT;const tve=(()=>{const s=[];for(let e=32;e<=126;e++)s.push(e);return s.push(65533),s})(),ive=(s,e)=>(s-=32,s<0||s>96?e<=2?(s+96)%96:95:s);class bS{constructor(e,t){this.scale=t,this._minimapCharRendererBrand=void 0,this.charDataNormal=bS.soften(e,12/15),this.charDataLight=bS.soften(e,50/60)}static soften(e,t){const i=new Uint8ClampedArray(e.length);for(let n=0,o=e.length;n<o;n++)i[n]=DE(e[n]*t);return i}renderChar(e,t,i,n,o,r,a,l,c,d,h){const u=1*this.scale,g=2*this.scale,f=h?1:g;if(t+u>e.width||i+f>e.height){console.warn("bad render request outside image data");return}const p=d?this.charDataLight:this.charDataNormal,_=ive(n,c),b=e.width*4,w=a.r,C=a.g,v=a.b,y=o.r-w,x=o.g-C,L=o.b-v,E=Math.max(r,l),I=e.data;let R=_*u*g,O=i*b+t*4;for(let P=0;P<f;P++){let B=O;for(let $=0;$<u;$++){const z=p[R++]/255*(r/255);I[B++]=w+y*z,I[B++]=C+x*z,I[B++]=v+L*z,I[B++]=E}O+=b}}blockRenderChar(e,t,i,n,o,r,a,l){const c=1*this.scale,d=2*this.scale,h=l?1:d;if(t+c>e.width||i+h>e.height){console.warn("bad render request outside image data");return}const u=e.width*4,g=.5*(o/255),f=r.r,p=r.g,_=r.b,b=n.r-f,w=n.g-p,C=n.b-_,v=f+b*g,y=p+w*g,x=_+C*g,L=Math.max(o,a),E=e.data;let I=i*u+t*4;for(let R=0;R<h;R++){let O=I;for(let P=0;P<c;P++)E[O++]=v,E[O++]=y,E[O++]=x,E[O++]=L;I+=u}}}const j$={0:0,1:1,2:2,3:3,4:4,5:5,6:6,7:7,8:8,9:9,A:10,B:11,C:12,D:13,E:14,F:15},G$=s=>{const e=new Uint8ClampedArray(s.length/2);for(let t=0;t<s.length;t+=2)e[t>>1]=j$[s[t]]<<4|j$[s[t+1]]&15;return e},K$={1:ym(()=>G$("0000511D6300CF609C709645A78432005642574171487021003C451900274D35D762755E8B629C5BA856AF57BA649530C167D1512A272A3F6038604460398526BCA2A968DB6F8957C768BE5FBE2FB467CF5D8D5B795DC7625B5DFF50DE64C466DB2FC47CD860A65E9A2EB96CB54CE06DA763AB2EA26860524D3763536601005116008177A8705E53AB738E6A982F88BAA35B5F5B626D9C636B449B737E5B7B678598869A662F6B5B8542706C704C80736A607578685B70594A49715A4522E792")),2:ym(()=>G$("000000000000000055394F383D2800008B8B1F210002000081B1CBCBCC820000847AAF6B9AAF2119BE08B8881AD60000A44FD07DCCF107015338130C00000000385972265F390B406E2437634B4B48031B12B8A0847000001E15B29A402F0000000000004B33460B00007A752C2A0000000000004D3900000084394B82013400ABA5CFC7AD9C0302A45A3E5A98AB000089A43382D97900008BA54AA087A70A0248A6A7AE6DBE0000BF6F94987EA40A01A06DCFA7A7A9030496C32F77891D0000A99FB1A0AFA80603B29AB9CA75930D010C0948354D3900000C0948354F37460D0028BE673D8400000000AF9D7B6E00002B007AA8933400007AA642675C2700007984CFB9C3985B768772A8A6B7B20000CAAECAAFC4B700009F94A6009F840009D09F9BA4CA9C0000CC8FC76DC87F0000C991C472A2000000A894A48CA7B501079BA2C9C69BA20000B19A5D3FA89000005CA6009DA2960901B0A7F0669FB200009D009E00B7890000DAD0F5D092820000D294D4C48BD10000B5A7A4A3B1A50402CAB6CBA6A2000000B5A7A4A3B1A8044FCDADD19D9CB00000B7778F7B8AAE0803C9AB5D3F5D3F00009EA09EA0BAB006039EA0989A8C7900009B9EF4D6B7C00000A9A7816CACA80000ABAC84705D3F000096DA635CDC8C00006F486F266F263D4784006124097B00374F6D2D6D2D6D4A3A95872322000000030000000000008D8939130000000000002E22A5C9CBC70600AB25C0B5C9B400061A2DB04CA67001082AA6BEBEBFC606002321DACBC19E03087AA08B6768380000282FBAC0B8CA7A88AD25BBA5A29900004C396C5894A6000040485A6E356E9442A32CD17EADA70000B4237923628600003E2DE9C1D7B500002F25BBA5A2990000231DB6AFB4A804023025C0B5CAB588062B2CBDBEC0C706882435A75CA20000002326BD6A82A908048B4B9A5A668000002423A09CB4BB060025259C9D8A7900001C1FCAB2C7C700002A2A9387ABA200002626A4A47D6E9D14333163A0C87500004B6F9C2D643A257049364936493647358A34438355497F1A0000A24C1D590000D38DFFBDD4CD3126"))};class R1{static create(e,t){if(this.lastCreated&&e===this.lastCreated.scale&&t===this.lastFontFamily)return this.lastCreated;let i;return K$[e]?i=new bS(K$[e](),e):i=R1.createFromSampleData(R1.createSampleData(t).data,e),this.lastFontFamily=t,this.lastCreated=i,i}static createSampleData(e){const t=document.createElement("canvas"),i=t.getContext("2d");t.style.height="16px",t.height=16,t.width=960,t.style.width="960px",i.fillStyle="#ffffff",i.font=`bold 16px ${e}`,i.textBaseline="middle";let n=0;for(const o of tve)i.fillText(String.fromCharCode(o),n,16/2),n+=10;return i.getImageData(0,0,960,16)}static createFromSampleData(e,t){if(e.length!==61440)throw new Error("Unexpected source in MinimapCharRenderer");const n=R1._downsample(e,t);return new bS(n,t)}static _downsampleChar(e,t,i,n,o){const r=1*o,a=2*o;let l=n,c=0;for(let d=0;d<a;d++){const h=d/a*16,u=(d+1)/a*16;for(let g=0;g<r;g++){const f=g/r*10,p=(g+1)/r*10;let _=0,b=0;for(let C=h;C<u;C++){const v=t+Math.floor(C)*3840,y=1-(C-Math.floor(C));for(let x=f;x<p;x++){const L=1-(x-Math.floor(x)),E=v+Math.floor(x)*4,I=L*y;b+=I,_+=e[E]*e[E+3]/255*I}}const w=_/b;c=Math.max(c,w),i[l++]=DE(w)}}return c}static _downsample(e,t){const i=2*t*1*t,n=i*96,o=new Uint8ClampedArray(n);let r=0,a=0,l=0;for(let c=0;c<96;c++)l=Math.max(l,this._downsampleChar(e,a,o,r,t)),r+=i,a+=40;if(l>0){const c=255/l;for(let d=0;d<n;d++)o[d]*=c}return o}}const nve=Bn?'"Segoe WPC", "Segoe UI", sans-serif':Ye?"-apple-system, BlinkMacSystemFont, sans-serif":'system-ui, "Ubuntu", "Droid Sans", sans-serif';class Iee{constructor(e,t){this._viewModelDecorationBrand=void 0,this.range=e,this.options=t}}function KW(s,e){return!(e.options.hideInCommentTokens&&YW(s,e)||e.options.hideInStringTokens&&ZW(s,e))}function YW(s,e){return Dee(s,e.range,t=>t===1)}function ZW(s,e){return Dee(s,e.range,t=>t===2)}function Dee(s,e,t){for(let i=e.startLineNumber;i<=e.endLineNumber;i++){const n=s.tokenization.getLineTokens(i),o=i===e.startLineNumber,r=i===e.endLineNumber;let a=o?n.findTokenIndexAtOffset(e.startColumn-1):0;for(;a<n.getCount()&&!(r&&n.getStartOffset(a)>e.endColumn-1);){if(!t(n.getStandardTokenType(a)))return!1;a++}}return!0}const sve=140,ove=2;class mw{constructor(e,t,i){const n=e.options,o=n.get(163),r=n.get(165),a=r.minimap,l=n.get(59),c=n.get(81);this.renderMinimap=a.renderMinimap,this.size=c.size,this.minimapHeightIsEditorHeight=a.minimapHeightIsEditorHeight,this.scrollBeyondLastLine=n.get(119),this.paddingTop=n.get(96).top,this.paddingBottom=n.get(96).bottom,this.showSlider=c.showSlider,this.autohide=c.autohide,this.pixelRatio=o,this.typicalHalfwidthCharacterWidth=l.typicalHalfwidthCharacterWidth,this.lineHeight=n.get(75),this.minimapLeft=a.minimapLeft,this.minimapWidth=a.minimapWidth,this.minimapHeight=r.height,this.canvasInnerWidth=a.minimapCanvasInnerWidth,this.canvasInnerHeight=a.minimapCanvasInnerHeight,this.canvasOuterWidth=a.minimapCanvasOuterWidth,this.canvasOuterHeight=a.minimapCanvasOuterHeight,this.isSampling=a.minimapIsSampling,this.editorHeight=r.height,this.fontScale=a.minimapScale,this.minimapLineHeight=a.minimapLineHeight,this.minimapCharWidth=1*this.fontScale,this.sectionHeaderFontFamily=nve,this.sectionHeaderFontSize=c.sectionHeaderFontSize*o,this.sectionHeaderLetterSpacing=c.sectionHeaderLetterSpacing,this.sectionHeaderFontColor=mw._getSectionHeaderColor(t,i.getColor(1)),this.charRenderer=ym(()=>R1.create(this.fontScale,l.fontFamily)),this.defaultBackgroundColor=i.getColor(2),this.backgroundColor=mw._getMinimapBackground(t,this.defaultBackgroundColor),this.foregroundAlpha=mw._getMinimapForegroundOpacity(t)}static _getMinimapBackground(e,t){const i=e.getColor(Ime);return i?new nh(i.rgba.r,i.rgba.g,i.rgba.b,Math.round(255*i.rgba.a)):t}static _getMinimapForegroundOpacity(e){const t=e.getColor(Dme);return t?nh._clamp(Math.round(255*t.rgba.a)):255}static _getSectionHeaderColor(e,t){const i=e.getColor(zc);return i?new nh(i.rgba.r,i.rgba.g,i.rgba.b,Math.round(255*i.rgba.a)):t}equals(e){return this.renderMinimap===e.renderMinimap&&this.size===e.size&&this.minimapHeightIsEditorHeight===e.minimapHeightIsEditorHeight&&this.scrollBeyondLastLine===e.scrollBeyondLastLine&&this.paddingTop===e.paddingTop&&this.paddingBottom===e.paddingBottom&&this.showSlider===e.showSlider&&this.autohide===e.autohide&&this.pixelRatio===e.pixelRatio&&this.typicalHalfwidthCharacterWidth===e.typicalHalfwidthCharacterWidth&&this.lineHeight===e.lineHeight&&this.minimapLeft===e.minimapLeft&&this.minimapWidth===e.minimapWidth&&this.minimapHeight===e.minimapHeight&&this.canvasInnerWidth===e.canvasInnerWidth&&this.canvasInnerHeight===e.canvasInnerHeight&&this.canvasOuterWidth===e.canvasOuterWidth&&this.canvasOuterHeight===e.canvasOuterHeight&&this.isSampling===e.isSampling&&this.editorHeight===e.editorHeight&&this.fontScale===e.fontScale&&this.minimapLineHeight===e.minimapLineHeight&&this.minimapCharWidth===e.minimapCharWidth&&this.sectionHeaderFontSize===e.sectionHeaderFontSize&&this.sectionHeaderLetterSpacing===e.sectionHeaderLetterSpacing&&this.defaultBackgroundColor&&this.defaultBackgroundColor.equals(e.defaultBackgroundColor)&&this.backgroundColor&&this.backgroundColor.equals(e.backgroundColor)&&this.foregroundAlpha===e.foregroundAlpha}}class A1{constructor(e,t,i,n,o,r,a,l,c){this.scrollTop=e,this.scrollHeight=t,this.sliderNeeded=i,this._computedSliderRatio=n,this.sliderTop=o,this.sliderHeight=r,this.topPaddingLineCount=a,this.startLineNumber=l,this.endLineNumber=c}getDesiredScrollTopFromDelta(e){return Math.round(this.scrollTop+e/this._computedSliderRatio)}getDesiredScrollTopFromTouchLocation(e){return Math.round((e-this.sliderHeight/2)/this._computedSliderRatio)}intersectWithViewport(e){const t=Math.max(this.startLineNumber,e.startLineNumber),i=Math.min(this.endLineNumber,e.endLineNumber);return t>i?null:[t,i]}getYForLineNumber(e,t){return+(e-this.startLineNumber+this.topPaddingLineCount)*t}static create(e,t,i,n,o,r,a,l,c,d,h){const u=e.pixelRatio,g=e.minimapLineHeight,f=Math.floor(e.canvasInnerHeight/g),p=e.lineHeight;if(e.minimapHeightIsEditorHeight){let x=l*e.lineHeight+e.paddingTop+e.paddingBottom;e.scrollBeyondLastLine&&(x+=Math.max(0,o-e.lineHeight-e.paddingBottom));const L=Math.max(1,Math.floor(o*o/x)),E=Math.max(0,e.minimapHeight-L),I=E/(d-o),R=c*I,O=E>0,P=Math.floor(e.canvasInnerHeight/e.minimapLineHeight),B=Math.floor(e.paddingTop/e.lineHeight);return new A1(c,d,O,I,R,L,B,1,Math.min(a,P))}let _;if(r&&i!==a){const x=i-t+1;_=Math.floor(x*g/u)}else{const x=o/p;_=Math.floor(x*g/u)}const b=Math.floor(e.paddingTop/p);let w=Math.floor(e.paddingBottom/p);if(e.scrollBeyondLastLine){const x=o/p;w=Math.max(w,x-1)}let C;if(w>0){const x=o/p;C=(b+a+w-x-1)*g/u}else C=Math.max(0,(b+a)*g/u-_);C=Math.min(e.minimapHeight-_,C);const v=C/(d-o),y=c*v;if(f>=b+a+w){const x=C>0;return new A1(c,d,x,v,y,_,b,1,a)}else{let x;t>1?x=t+b:x=Math.max(1,c/p);let L,E=Math.max(1,Math.floor(x-y*u/g));E<b?(L=b-E+1,E=1):(L=0,E=Math.max(1,E-b)),h&&h.scrollHeight===d&&(h.scrollTop>c&&(E=Math.min(E,h.startLineNumber),L=Math.max(L,h.topPaddingLineCount)),h.scrollTop<c&&(E=Math.max(E,h.startLineNumber),L=Math.min(L,h.topPaddingLineCount)));const I=Math.min(a,E-L+f-1),R=(c-n)/p;let O;return c>=e.paddingTop?O=(t-E+L+R)*g/u:O=c/e.paddingTop*(L+R)*g/u,new A1(c,d,!0,v,O,_,L,E,I)}}}const eN=class eN{constructor(e){this.dy=e}onContentChanged(){this.dy=-1}onTokensChanged(){this.dy=-1}};eN.INVALID=new eN(-1);let oI=eN;class Y${constructor(e,t,i){this.renderedLayout=e,this._imageData=t,this._renderedLines=new ree({createLine:()=>oI.INVALID}),this._renderedLines._set(e.startLineNumber,i)}linesEquals(e){if(!this.scrollEquals(e))return!1;const i=this._renderedLines._get().lines;for(let n=0,o=i.length;n<o;n++)if(i[n].dy===-1)return!1;return!0}scrollEquals(e){return this.renderedLayout.startLineNumber===e.startLineNumber&&this.renderedLayout.endLineNumber===e.endLineNumber}_get(){const e=this._renderedLines._get();return{imageData:this._imageData,rendLineNumberStart:e.rendLineNumberStart,lines:e.lines}}onLinesChanged(e,t){return this._renderedLines.onLinesChanged(e,t)}onLinesDeleted(e,t){this._renderedLines.onLinesDeleted(e,t)}onLinesInserted(e,t){this._renderedLines.onLinesInserted(e,t)}onTokensChanged(e){return this._renderedLines.onTokensChanged(e)}}class XW{constructor(e,t,i,n){this._backgroundFillData=XW._createBackgroundFillData(t,i,n),this._buffers=[e.createImageData(t,i),e.createImageData(t,i)],this._lastUsedBuffer=0}getBuffer(){this._lastUsedBuffer=1-this._lastUsedBuffer;const e=this._buffers[this._lastUsedBuffer];return e.data.set(this._backgroundFillData),e}static _createBackgroundFillData(e,t,i){const n=i.r,o=i.g,r=i.b,a=i.a,l=new Uint8ClampedArray(e*t*4);let c=0;for(let d=0;d<t;d++)for(let h=0;h<e;h++)l[c]=n,l[c+1]=o,l[c+2]=r,l[c+3]=a,c+=4;return l}}class wS{static compute(e,t,i){if(e.renderMinimap===0||!e.isSampling)return[null,[]];const{minimapLineCount:n}=uw.computeContainedMinimapLineCount({viewLineCount:t,scrollBeyondLastLine:e.scrollBeyondLastLine,paddingTop:e.paddingTop,paddingBottom:e.paddingBottom,height:e.editorHeight,lineHeight:e.lineHeight,pixelRatio:e.pixelRatio}),o=t/n,r=o/2;if(!i||i.minimapLines.length===0){const _=[];if(_[0]=1,n>1){for(let b=0,w=n-1;b<w;b++)_[b]=Math.round(b*o+r);_[n-1]=t}return[new wS(o,_),[]]}const a=i.minimapLines,l=a.length,c=[];let d=0,h=0,u=1;const g=10;let f=[],p=null;for(let _=0;_<n;_++){const b=Math.max(u,Math.round(_*o)),w=Math.max(b,Math.round((_+1)*o));for(;d<l&&a[d]<b;){if(f.length<g){const v=d+1+h;p&&p.type==="deleted"&&p._oldIndex===d-1?p.deleteToLineNumber++:(p={type:"deleted",_oldIndex:d,deleteFromLineNumber:v,deleteToLineNumber:v},f.push(p)),h--}d++}let C;if(d<l&&a[d]<=w)C=a[d],d++;else if(_===0?C=1:_+1===n?C=t:C=Math.round(_*o+r),f.length<g){const v=d+1+h;p&&p.type==="inserted"&&p._i===_-1?p.insertToLineNumber++:(p={type:"inserted",_i:_,insertFromLineNumber:v,insertToLineNumber:v},f.push(p)),h++}c[_]=C,u=C}if(f.length<g)for(;d<l;){const _=d+1+h;p&&p.type==="deleted"&&p._oldIndex===d-1?p.deleteToLineNumber++:(p={type:"deleted",_oldIndex:d,deleteFromLineNumber:_,deleteToLineNumber:_},f.push(p)),h--,d++}else f=[{type:"flush"}];return[new wS(o,c),f]}constructor(e,t){this.samplingRatio=e,this.minimapLines=t}modelLineToMinimapLine(e){return Math.min(this.minimapLines.length,Math.max(1,Math.round(e/this.samplingRatio)))}modelLineRangeToMinimapLineRange(e,t){let i=this.modelLineToMinimapLine(e)-1;for(;i>0&&this.minimapLines[i-1]>=e;)i--;let n=this.modelLineToMinimapLine(t)-1;for(;n+1<this.minimapLines.length&&this.minimapLines[n+1]<=t;)n++;if(i===n){const o=this.minimapLines[i];if(o<e||o>t)return null}return[i+1,n+1]}decorationLineRangeToMinimapLineRange(e,t){let i=this.modelLineToMinimapLine(e),n=this.modelLineToMinimapLine(t);return e!==t&&n===i&&(n===this.minimapLines.length?i>1&&i--:n++),[i,n]}onLinesDeleted(e){const t=e.toLineNumber-e.fromLineNumber+1;let i=this.minimapLines.length,n=0;for(let o=this.minimapLines.length-1;o>=0&&!(this.minimapLines[o]<e.fromLineNumber);o--)this.minimapLines[o]<=e.toLineNumber?(this.minimapLines[o]=Math.max(1,e.fromLineNumber-1),i=Math.min(i,o),n=Math.max(n,o)):this.minimapLines[o]-=t;return[i,n]}onLinesInserted(e){const t=e.toLineNumber-e.fromLineNumber+1;for(let i=this.minimapLines.length-1;i>=0&&!(this.minimapLines[i]<e.fromLineNumber);i--)this.minimapLines[i]+=t}}class rve extends no{constructor(e){super(e),this._sectionHeaderCache=new Va(10,1.5),this.tokensColorTracker=sI.getInstance(),this._selections=[],this._minimapSelections=null,this.options=new mw(this._context.configuration,this._context.theme,this.tokensColorTracker);const[t]=wS.compute(this.options,this._context.viewModel.getLineCount(),null);this._samplingState=t,this._shouldCheckSampling=!1,this._actual=new Eb(e.theme,this)}dispose(){this._actual.dispose(),super.dispose()}getDomNode(){return this._actual.getDomNode()}_onOptionsMaybeChanged(){const e=new mw(this._context.configuration,this._context.theme,this.tokensColorTracker);return this.options.equals(e)?!1:(this.options=e,this._recreateLineSampling(),this._actual.onDidChangeOptions(),!0)}onConfigurationChanged(e){return this._onOptionsMaybeChanged()}onCursorStateChanged(e){return this._selections=e.selections,this._minimapSelections=null,this._actual.onSelectionChanged()}onDecorationsChanged(e){return e.affectsMinimap?this._actual.onDecorationsChanged():!1}onFlushed(e){return this._samplingState&&(this._shouldCheckSampling=!0),this._actual.onFlushed()}onLinesChanged(e){if(this._samplingState){const t=this._samplingState.modelLineRangeToMinimapLineRange(e.fromLineNumber,e.fromLineNumber+e.count-1);return t?this._actual.onLinesChanged(t[0],t[1]-t[0]+1):!1}else return this._actual.onLinesChanged(e.fromLineNumber,e.count)}onLinesDeleted(e){if(this._samplingState){const[t,i]=this._samplingState.onLinesDeleted(e);return t<=i&&this._actual.onLinesChanged(t+1,i-t+1),this._shouldCheckSampling=!0,!0}else return this._actual.onLinesDeleted(e.fromLineNumber,e.toLineNumber)}onLinesInserted(e){return this._samplingState?(this._samplingState.onLinesInserted(e),this._shouldCheckSampling=!0,!0):this._actual.onLinesInserted(e.fromLineNumber,e.toLineNumber)}onScrollChanged(e){return this._actual.onScrollChanged(e)}onThemeChanged(e){return this._actual.onThemeChanged(),this._onOptionsMaybeChanged(),!0}onTokensChanged(e){if(this._samplingState){const t=[];for(const i of e.ranges){const n=this._samplingState.modelLineRangeToMinimapLineRange(i.fromLineNumber,i.toLineNumber);n&&t.push({fromLineNumber:n[0],toLineNumber:n[1]})}return t.length?this._actual.onTokensChanged(t):!1}else return this._actual.onTokensChanged(e.ranges)}onTokensColorsChanged(e){return this._onOptionsMaybeChanged(),this._actual.onTokensColorsChanged()}onZonesChanged(e){return this._actual.onZonesChanged()}prepareRender(e){this._shouldCheckSampling&&(this._shouldCheckSampling=!1,this._recreateLineSampling())}render(e){let t=e.visibleRange.startLineNumber,i=e.visibleRange.endLineNumber;this._samplingState&&(t=this._samplingState.modelLineToMinimapLine(t),i=this._samplingState.modelLineToMinimapLine(i));const n={viewportContainsWhitespaceGaps:e.viewportData.whitespaceViewportData.length>0,scrollWidth:e.scrollWidth,scrollHeight:e.scrollHeight,viewportStartLineNumber:t,viewportEndLineNumber:i,viewportStartLineNumberVerticalOffset:e.getVerticalOffsetForLineNumber(t),scrollTop:e.scrollTop,scrollLeft:e.scrollLeft,viewportWidth:e.viewportWidth,viewportHeight:e.viewportHeight};this._actual.render(n)}_recreateLineSampling(){this._minimapSelections=null;const e=!!this._samplingState,[t,i]=wS.compute(this.options,this._context.viewModel.getLineCount(),this._samplingState);if(this._samplingState=t,e&&this._samplingState)for(const n of i)switch(n.type){case"deleted":this._actual.onLinesDeleted(n.deleteFromLineNumber,n.deleteToLineNumber);break;case"inserted":this._actual.onLinesInserted(n.insertFromLineNumber,n.insertToLineNumber);break;case"flush":this._actual.onFlushed();break}}getLineCount(){return this._samplingState?this._samplingState.minimapLines.length:this._context.viewModel.getLineCount()}getRealLineCount(){return this._context.viewModel.getLineCount()}getLineContent(e){return this._samplingState?this._context.viewModel.getLineContent(this._samplingState.minimapLines[e-1]):this._context.viewModel.getLineContent(e)}getLineMaxColumn(e){return this._samplingState?this._context.viewModel.getLineMaxColumn(this._samplingState.minimapLines[e-1]):this._context.viewModel.getLineMaxColumn(e)}getMinimapLinesRenderingData(e,t,i){if(this._samplingState){const n=[];for(let o=0,r=t-e+1;o<r;o++)i[o]?n[o]=this._context.viewModel.getViewLineData(this._samplingState.minimapLines[e+o-1]):n[o]=null;return n}return this._context.viewModel.getMinimapLinesRenderingData(e,t,i).data}getSelections(){if(this._minimapSelections===null)if(this._samplingState){this._minimapSelections=[];for(const e of this._selections){const[t,i]=this._samplingState.decorationLineRangeToMinimapLineRange(e.startLineNumber,e.endLineNumber);this._minimapSelections.push(new he(t,e.startColumn,i,e.endColumn))}}else this._minimapSelections=this._selections;return this._minimapSelections}getMinimapDecorationsInViewport(e,t){return this._getMinimapDecorationsInViewport(e,t).filter(i=>!i.options.minimap?.sectionHeaderStyle)}getSectionHeaderDecorationsInViewport(e,t){const i=this.options.sectionHeaderFontSize/this.options.minimapLineHeight;return e=Math.floor(Math.max(1,e-i)),this._getMinimapDecorationsInViewport(e,t).filter(n=>!!n.options.minimap?.sectionHeaderStyle)}_getMinimapDecorationsInViewport(e,t){let i;if(this._samplingState){const o=this._samplingState.minimapLines[e-1],r=this._samplingState.minimapLines[t-1];i=new k(o,1,r,this._context.viewModel.getLineMaxColumn(r))}else i=new k(e,1,t,this._context.viewModel.getLineMaxColumn(t));const n=this._context.viewModel.getMinimapDecorationsInRange(i);if(this._samplingState){const o=[];for(const r of n){if(!r.options.minimap)continue;const a=r.range,l=this._samplingState.modelLineToMinimapLine(a.startLineNumber),c=this._samplingState.modelLineToMinimapLine(a.endLineNumber);o.push(new Iee(new k(l,a.startColumn,c,a.endColumn),r.options))}return o}return n}getSectionHeaderText(e,t){const i=e.options.minimap?.sectionHeaderText;if(!i)return null;const n=this._sectionHeaderCache.get(i);if(n)return n;const o=t(i);return this._sectionHeaderCache.set(i,o),o}getOptions(){return this._context.viewModel.model.getOptions()}revealLineNumber(e){this._samplingState&&(e=this._samplingState.minimapLines[e-1]),this._context.viewModel.revealRange("mouse",!1,new k(e,1,e,1),1,0)}setScrollTop(e){this._context.viewModel.viewLayout.setScrollPosition({scrollTop:e},1)}}class Eb extends F{constructor(e,t){super(),this._renderDecorations=!1,this._gestureInProgress=!1,this._isMouseOverMinimap=!1,this._theme=e,this._model=t,this._lastRenderData=null,this._buffers=null,this._selectionColor=this._theme.getColor(F$),this._domNode=wt(document.createElement("div")),kl.write(this._domNode,9),this._domNode.setClassName(this._getMinimapDomNodeClassName()),this._domNode.setPosition("absolute"),this._domNode.setAttribute("role","presentation"),this._domNode.setAttribute("aria-hidden","true"),this._shadow=wt(document.createElement("div")),this._shadow.setClassName("minimap-shadow-hidden"),this._domNode.appendChild(this._shadow),this._canvas=wt(document.createElement("canvas")),this._canvas.setPosition("absolute"),this._canvas.setLeft(0),this._domNode.appendChild(this._canvas),this._decorationsCanvas=wt(document.createElement("canvas")),this._decorationsCanvas.setPosition("absolute"),this._decorationsCanvas.setClassName("minimap-decorations-layer"),this._decorationsCanvas.setLeft(0),this._domNode.appendChild(this._decorationsCanvas),this._slider=wt(document.createElement("div")),this._slider.setPosition("absolute"),this._slider.setClassName("minimap-slider"),this._slider.setLayerHinting(!0),this._slider.setContain("strict"),this._domNode.appendChild(this._slider),this._sliderHorizontal=wt(document.createElement("div")),this._sliderHorizontal.setPosition("absolute"),this._sliderHorizontal.setClassName("minimap-slider-horizontal"),this._slider.appendChild(this._sliderHorizontal),this._applyLayout(),this._hideDelayedScheduler=this._register(new Rt(()=>this._hideImmediatelyIfMouseIsOutside(),500)),this._register(Fi(this._domNode.domNode,se.MOUSE_OVER,()=>{this._isMouseOverMinimap=!0})),this._register(Fi(this._domNode.domNode,se.MOUSE_LEAVE,()=>{this._isMouseOverMinimap=!1})),this._pointerDownListener=Fi(this._domNode.domNode,se.POINTER_DOWN,i=>{i.preventDefault();const n=i.pointerType==="mouse",o=i.button===0;if(this._model.options.renderMinimap!==0&&this._lastRenderData){if(this._model.options.size!=="proportional"){if(o&&this._lastRenderData){const a=Li(this._slider.domNode),l=a.top+a.height/2;this._startSliderDragging(i,l,this._lastRenderData.renderedLayout)}return}if(o||!n){const a=this._model.options.minimapLineHeight,l=this._model.options.canvasInnerHeight/this._model.options.canvasOuterHeight*i.offsetY;let d=Math.floor(l/a)+this._lastRenderData.renderedLayout.startLineNumber-this._lastRenderData.renderedLayout.topPaddingLineCount;d=Math.min(d,this._model.getLineCount()),this._model.revealLineNumber(d)}}}),this._sliderPointerMoveMonitor=new eC,this._sliderPointerDownListener=Fi(this._slider.domNode,se.POINTER_DOWN,i=>{i.preventDefault(),i.stopPropagation(),i.button===0&&this._lastRenderData&&this._startSliderDragging(i,i.pageY,this._lastRenderData.renderedLayout)}),this._gestureDisposable=_s.addTarget(this._domNode.domNode),this._sliderTouchStartListener=W(this._domNode.domNode,qt.Start,i=>{i.preventDefault(),i.stopPropagation(),this._lastRenderData&&(this._slider.toggleClassName("active",!0),this._gestureInProgress=!0,this.scrollDueToTouchEvent(i))},{passive:!1}),this._sliderTouchMoveListener=W(this._domNode.domNode,qt.Change,i=>{i.preventDefault(),i.stopPropagation(),this._lastRenderData&&this._gestureInProgress&&this.scrollDueToTouchEvent(i)},{passive:!1}),this._sliderTouchEndListener=Fi(this._domNode.domNode,qt.End,i=>{i.preventDefault(),i.stopPropagation(),this._gestureInProgress=!1,this._slider.toggleClassName("active",!1)})}_hideSoon(){this._hideDelayedScheduler.cancel(),this._hideDelayedScheduler.schedule()}_hideImmediatelyIfMouseIsOutside(){if(this._isMouseOverMinimap){this._hideSoon();return}this._domNode.toggleClassName("active",!1)}_startSliderDragging(e,t,i){if(!e.target||!(e.target instanceof Element))return;const n=e.pageX;this._slider.toggleClassName("active",!0);const o=(r,a)=>{const l=Li(this._domNode.domNode),c=Math.min(Math.abs(a-n),Math.abs(a-l.left),Math.abs(a-l.left-l.width));if(Bn&&c>sve){this._model.setScrollTop(i.scrollTop);return}const d=r-t;this._model.setScrollTop(i.getDesiredScrollTopFromDelta(d))};e.pageY!==t&&o(e.pageY,n),this._sliderPointerMoveMonitor.startMonitoring(e.target,e.pointerId,e.buttons,r=>o(r.pageY,r.pageX),()=>{this._slider.toggleClassName("active",!1)})}scrollDueToTouchEvent(e){const t=this._domNode.domNode.getBoundingClientRect().top,i=this._lastRenderData.renderedLayout.getDesiredScrollTopFromTouchLocation(e.pageY-t);this._model.setScrollTop(i)}dispose(){this._pointerDownListener.dispose(),this._sliderPointerMoveMonitor.dispose(),this._sliderPointerDownListener.dispose(),this._gestureDisposable.dispose(),this._sliderTouchStartListener.dispose(),this._sliderTouchMoveListener.dispose(),this._sliderTouchEndListener.dispose(),super.dispose()}_getMinimapDomNodeClassName(){const e=["minimap"];return this._model.options.showSlider==="always"?e.push("slider-always"):e.push("slider-mouseover"),this._model.options.autohide==="mouseover"?e.push("minimap-autohide-mouseover"):this._model.options.autohide==="scroll"&&e.push("minimap-autohide-scroll"),e.join(" ")}getDomNode(){return this._domNode}_applyLayout(){this._domNode.setLeft(this._model.options.minimapLeft),this._domNode.setWidth(this._model.options.minimapWidth),this._domNode.setHeight(this._model.options.minimapHeight),this._shadow.setHeight(this._model.options.minimapHeight),this._canvas.setWidth(this._model.options.canvasOuterWidth),this._canvas.setHeight(this._model.options.canvasOuterHeight),this._canvas.domNode.width=this._model.options.canvasInnerWidth,this._canvas.domNode.height=this._model.options.canvasInnerHeight,this._decorationsCanvas.setWidth(this._model.options.canvasOuterWidth),this._decorationsCanvas.setHeight(this._model.options.canvasOuterHeight),this._decorationsCanvas.domNode.width=this._model.options.canvasInnerWidth,this._decorationsCanvas.domNode.height=this._model.options.canvasInnerHeight,this._slider.setWidth(this._model.options.minimapWidth)}_getBuffer(){return this._buffers||this._model.options.canvasInnerWidth>0&&this._model.options.canvasInnerHeight>0&&(this._buffers=new XW(this._canvas.domNode.getContext("2d"),this._model.options.canvasInnerWidth,this._model.options.canvasInnerHeight,this._model.options.backgroundColor)),this._buffers?this._buffers.getBuffer():null}onDidChangeOptions(){this._lastRenderData=null,this._buffers=null,this._applyLayout(),this._domNode.setClassName(this._getMinimapDomNodeClassName())}onSelectionChanged(){return this._renderDecorations=!0,!0}onDecorationsChanged(){return this._renderDecorations=!0,!0}onFlushed(){return this._lastRenderData=null,!0}onLinesChanged(e,t){return this._lastRenderData?this._lastRenderData.onLinesChanged(e,t):!1}onLinesDeleted(e,t){return this._lastRenderData?.onLinesDeleted(e,t),!0}onLinesInserted(e,t){return this._lastRenderData?.onLinesInserted(e,t),!0}onScrollChanged(e){return this._model.options.autohide==="scroll"&&(e.scrollTopChanged||e.scrollHeightChanged)&&(this._domNode.toggleClassName("active",!0),this._hideSoon()),this._renderDecorations=!0,!0}onThemeChanged(){return this._selectionColor=this._theme.getColor(F$),this._renderDecorations=!0,!0}onTokensChanged(e){return this._lastRenderData?this._lastRenderData.onTokensChanged(e):!1}onTokensColorsChanged(){return this._lastRenderData=null,this._buffers=null,!0}onZonesChanged(){return this._lastRenderData=null,!0}render(e){if(this._model.options.renderMinimap===0){this._shadow.setClassName("minimap-shadow-hidden"),this._sliderHorizontal.setWidth(0),this._sliderHorizontal.setHeight(0);return}e.scrollLeft+e.viewportWidth>=e.scrollWidth?this._shadow.setClassName("minimap-shadow-hidden"):this._shadow.setClassName("minimap-shadow-visible");const i=A1.create(this._model.options,e.viewportStartLineNumber,e.viewportEndLineNumber,e.viewportStartLineNumberVerticalOffset,e.viewportHeight,e.viewportContainsWhitespaceGaps,this._model.getLineCount(),this._model.getRealLineCount(),e.scrollTop,e.scrollHeight,this._lastRenderData?this._lastRenderData.renderedLayout:null);this._slider.setDisplay(i.sliderNeeded?"block":"none"),this._slider.setTop(i.sliderTop),this._slider.setHeight(i.sliderHeight),this._sliderHorizontal.setLeft(0),this._sliderHorizontal.setWidth(this._model.options.minimapWidth),this._sliderHorizontal.setTop(0),this._sliderHorizontal.setHeight(i.sliderHeight),this.renderDecorations(i),this._lastRenderData=this.renderLines(i)}renderDecorations(e){if(this._renderDecorations){this._renderDecorations=!1;const t=this._model.getSelections();t.sort(k.compareRangesUsingStarts);const i=this._model.getMinimapDecorationsInViewport(e.startLineNumber,e.endLineNumber);i.sort((u,g)=>(u.options.zIndex||0)-(g.options.zIndex||0));const{canvasInnerWidth:n,canvasInnerHeight:o}=this._model.options,r=this._model.options.minimapLineHeight,a=this._model.options.minimapCharWidth,l=this._model.getOptions().tabSize,c=this._decorationsCanvas.domNode.getContext("2d");c.clearRect(0,0,n,o);const d=new Z$(e.startLineNumber,e.endLineNumber,!1);this._renderSelectionLineHighlights(c,t,d,e,r),this._renderDecorationsLineHighlights(c,i,d,e,r);const h=new Z$(e.startLineNumber,e.endLineNumber,null);this._renderSelectionsHighlights(c,t,h,e,r,l,a,n),this._renderDecorationsHighlights(c,i,h,e,r,l,a,n),this._renderSectionHeaders(e)}}_renderSelectionLineHighlights(e,t,i,n,o){if(!this._selectionColor||this._selectionColor.isTransparent())return;e.fillStyle=this._selectionColor.transparent(.5).toString();let r=0,a=0;for(const l of t){const c=n.intersectWithViewport(l);if(!c)continue;const[d,h]=c;for(let f=d;f<=h;f++)i.set(f,!0);const u=n.getYForLineNumber(d,o),g=n.getYForLineNumber(h,o);a>=u||(a>r&&e.fillRect(ql,r,e.canvas.width,a-r),r=u),a=g}a>r&&e.fillRect(ql,r,e.canvas.width,a-r)}_renderDecorationsLineHighlights(e,t,i,n,o){const r=new Map;for(let a=t.length-1;a>=0;a--){const l=t[a],c=l.options.minimap;if(!c||c.position!==1)continue;const d=n.intersectWithViewport(l.range);if(!d)continue;const[h,u]=d,g=c.getColor(this._theme.value);if(!g||g.isTransparent())continue;let f=r.get(g.toString());f||(f=g.transparent(.5).toString(),r.set(g.toString(),f)),e.fillStyle=f;for(let p=h;p<=u;p++){if(i.has(p))continue;i.set(p,!0);const _=n.getYForLineNumber(h,o);e.fillRect(ql,_,e.canvas.width,o)}}}_renderSelectionsHighlights(e,t,i,n,o,r,a,l){if(!(!this._selectionColor||this._selectionColor.isTransparent()))for(const c of t){const d=n.intersectWithViewport(c);if(!d)continue;const[h,u]=d;for(let g=h;g<=u;g++)this.renderDecorationOnLine(e,i,c,this._selectionColor,n,g,o,o,r,a,l)}}_renderDecorationsHighlights(e,t,i,n,o,r,a,l){for(const c of t){const d=c.options.minimap;if(!d)continue;const h=n.intersectWithViewport(c.range);if(!h)continue;const[u,g]=h,f=d.getColor(this._theme.value);if(!(!f||f.isTransparent()))for(let p=u;p<=g;p++)switch(d.position){case 1:this.renderDecorationOnLine(e,i,c.range,f,n,p,o,o,r,a,l);continue;case 2:{const _=n.getYForLineNumber(p,o);this.renderDecoration(e,f,2,_,ove,o);continue}}}}renderDecorationOnLine(e,t,i,n,o,r,a,l,c,d,h){const u=o.getYForLineNumber(r,l);if(u+a<0||u>this._model.options.canvasInnerHeight)return;const{startLineNumber:g,endLineNumber:f}=i,p=g===r?i.startColumn:1,_=f===r?i.endColumn:this._model.getLineMaxColumn(r),b=this.getXOffsetForPosition(t,r,p,c,d,h),w=this.getXOffsetForPosition(t,r,_,c,d,h);this.renderDecoration(e,n,b,u,w-b,a)}getXOffsetForPosition(e,t,i,n,o,r){if(i===1)return ql;if((i-1)*o>=r)return r;let l=e.get(t);if(!l){const c=this._model.getLineContent(t);l=[ql];let d=ql;for(let h=1;h<c.length+1;h++){const u=c.charCodeAt(h-1),g=u===9?n*o:gh(u)?2*o:o,f=d+g;if(f>=r){l[h]=r;break}l[h]=f,d=f}e.set(t,l)}return i-1<l.length?l[i-1]:r}renderDecoration(e,t,i,n,o,r){e.fillStyle=t&&t.toString()||"",e.fillRect(i,n,o,r)}_renderSectionHeaders(e){const t=this._model.options.minimapLineHeight,i=this._model.options.sectionHeaderFontSize,n=this._model.options.sectionHeaderLetterSpacing,o=i*1.5,{canvasInnerWidth:r}=this._model.options,a=this._model.options.backgroundColor,l=`rgb(${a.r} ${a.g} ${a.b} / .7)`,c=this._model.options.sectionHeaderFontColor,d=`rgb(${c.r} ${c.g} ${c.b})`,h=d,u=this._decorationsCanvas.domNode.getContext("2d");u.letterSpacing=n+"px",u.font="500 "+i+"px "+this._model.options.sectionHeaderFontFamily,u.strokeStyle=h,u.lineWidth=.4;const g=this._model.getSectionHeaderDecorationsInViewport(e.startLineNumber,e.endLineNumber);g.sort((p,_)=>p.range.startLineNumber-_.range.startLineNumber);const f=Eb._fitSectionHeader.bind(null,u,r-ql);for(const p of g){const _=e.getYForLineNumber(p.range.startLineNumber,t)+i,b=_-i,w=b+2,C=this._model.getSectionHeaderText(p,f);Eb._renderSectionLabel(u,C,p.options.minimap?.sectionHeaderStyle===2,l,d,r,b,o,_,w)}}static _fitSectionHeader(e,t,i){if(!i)return i;const n="…",o=e.measureText(i).width,r=e.measureText(n).width;if(o<=t||o<=r)return i;const a=i.length,l=o/i.length,c=Math.floor((t-r)/l)-1;let d=Math.ceil(c/2);for(;d>0&&/\s/.test(i[d-1]);)--d;return i.substring(0,d)+n+i.substring(a-(c-d))}static _renderSectionLabel(e,t,i,n,o,r,a,l,c,d){t&&(e.fillStyle=n,e.fillRect(0,a,r,l),e.fillStyle=o,e.fillText(t,ql,c)),i&&(e.beginPath(),e.moveTo(0,d),e.lineTo(r,d),e.closePath(),e.stroke())}renderLines(e){const t=e.startLineNumber,i=e.endLineNumber,n=this._model.options.minimapLineHeight;if(this._lastRenderData&&this._lastRenderData.linesEquals(e)){const z=this._lastRenderData._get();return new Y$(e,z.imageData,z.lines)}const o=this._getBuffer();if(!o)return null;const[r,a,l]=Eb._renderUntouchedLines(o,e.topPaddingLineCount,t,i,n,this._lastRenderData),c=this._model.getMinimapLinesRenderingData(t,i,l),d=this._model.getOptions().tabSize,h=this._model.options.defaultBackgroundColor,u=this._model.options.backgroundColor,g=this._model.options.foregroundAlpha,f=this._model.tokensColorTracker,p=f.backgroundIsLight(),_=this._model.options.renderMinimap,b=this._model.options.charRenderer(),w=this._model.options.fontScale,C=this._model.options.minimapCharWidth,y=(_===1?2:3)*w,x=n>y?Math.floor((n-y)/2):0,L=u.a/255,E=new nh(Math.round((u.r-h.r)*L+h.r),Math.round((u.g-h.g)*L+h.g),Math.round((u.b-h.b)*L+h.b),255);let I=e.topPaddingLineCount*n;const R=[];for(let z=0,Q=i-t+1;z<Q;z++)l[z]&&Eb._renderLine(o,E,u.a,p,_,C,f,g,b,I,x,d,c[z],w,n),R[z]=new oI(I),I+=n;const O=r===-1?0:r,B=(a===-1?o.height:a)-O;return this._canvas.domNode.getContext("2d").putImageData(o,0,0,0,O,o.width,B),new Y$(e,o,R)}static _renderUntouchedLines(e,t,i,n,o,r){const a=[];if(!r){for(let I=0,R=n-i+1;I<R;I++)a[I]=!0;return[-1,-1,a]}const l=r._get(),c=l.imageData.data,d=l.rendLineNumberStart,h=l.lines,u=h.length,g=e.width,f=e.data,p=(n-i+1)*o*g*4;let _=-1,b=-1,w=-1,C=-1,v=-1,y=-1,x=t*o;for(let I=i;I<=n;I++){const R=I-i,O=I-d,P=O>=0&&O<u?h[O].dy:-1;if(P===-1){a[R]=!0,x+=o;continue}const B=P*g*4,$=(P+o)*g*4,z=x*g*4,Q=(x+o)*g*4;C===B&&y===z?(C=$,y=Q):(w!==-1&&(f.set(c.subarray(w,C),v),_===-1&&w===0&&w===v&&(_=C),b===-1&&C===p&&w===v&&(b=w)),w=B,C=$,v=z,y=Q),a[R]=!1,x+=o}w!==-1&&(f.set(c.subarray(w,C),v),_===-1&&w===0&&w===v&&(_=C),b===-1&&C===p&&w===v&&(b=w));const L=_===-1?-1:_/(g*4),E=b===-1?-1:b/(g*4);return[L,E,a]}static _renderLine(e,t,i,n,o,r,a,l,c,d,h,u,g,f,p){const _=g.content,b=g.tokens,w=e.width-r,C=p===1;let v=ql,y=0,x=0;for(let L=0,E=b.getCount();L<E;L++){const I=b.getEndOffset(L),R=b.getForeground(L),O=a.getColor(R);for(;y<I;y++){if(v>w)return;const P=_.charCodeAt(y);if(P===9){const B=u-(y+x)%u;x+=B-1,v+=B*r}else if(P===32)v+=r;else{const B=gh(P)?2:1;for(let $=0;$<B;$++)if(o===2?c.blockRenderChar(e,v,d+h,O,l,t,i,C):c.renderChar(e,v,d+h,P,O,l,t,i,f,n,C),v+=r,v>w)return}}}}}class Z${constructor(e,t,i){this._startLineNumber=e,this._endLineNumber=t,this._defaultValue=i,this._values=[];for(let n=0,o=this._endLineNumber-this._startLineNumber+1;n<o;n++)this._values[n]=i}has(e){return this.get(e)!==this._defaultValue}set(e,t){e<this._startLineNumber||e>this._endLineNumber||(this._values[e-this._startLineNumber]=t)}get(e){return e<this._startLineNumber||e>this._endLineNumber?this._defaultValue:this._values[e-this._startLineNumber]}}class ave extends no{constructor(e,t){super(e),this._viewDomNode=t;const n=this._context.configuration.options.get(165);this._widgets={},this._verticalScrollbarWidth=n.verticalScrollbarWidth,this._minimapWidth=n.minimap.minimapWidth,this._horizontalScrollbarHeight=n.horizontalScrollbarHeight,this._editorHeight=n.height,this._editorWidth=n.width,this._viewDomNodeRect={top:0,left:0,width:0,height:0},this._domNode=wt(document.createElement("div")),kl.write(this._domNode,4),this._domNode.setClassName("overlayWidgets"),this.overflowingOverlayWidgetsDomNode=wt(document.createElement("div")),kl.write(this.overflowingOverlayWidgetsDomNode,5),this.overflowingOverlayWidgetsDomNode.setClassName("overflowingOverlayWidgets")}dispose(){super.dispose(),this._widgets={}}getDomNode(){return this._domNode}onConfigurationChanged(e){const i=this._context.configuration.options.get(165);return this._verticalScrollbarWidth=i.verticalScrollbarWidth,this._minimapWidth=i.minimap.minimapWidth,this._horizontalScrollbarHeight=i.horizontalScrollbarHeight,this._editorHeight=i.height,this._editorWidth=i.width,!0}_widgetCanOverflow(e){const i=this._context.configuration.options.get(4);return(e.allowEditorOverflow||!1)&&i}addWidget(e){const t=wt(e.getDomNode());this._widgets[e.getId()]={widget:e,preference:null,domNode:t},t.setPosition("absolute"),t.setAttribute("widgetId",e.getId()),this._widgetCanOverflow(e)?this.overflowingOverlayWidgetsDomNode.appendChild(t):this._domNode.appendChild(t),this.setShouldRender(),this._updateMaxMinWidth()}setWidgetPosition(e,t){const i=this._widgets[e.getId()],n=t?t.preference:null,o=t?.stackOridinal;return i.preference===n&&i.stack===o?(this._updateMaxMinWidth(),!1):(i.preference=n,i.stack=o,this.setShouldRender(),this._updateMaxMinWidth(),!0)}removeWidget(e){const t=e.getId();if(this._widgets.hasOwnProperty(t)){const n=this._widgets[t].domNode.domNode;delete this._widgets[t],n.remove(),this.setShouldRender(),this._updateMaxMinWidth()}}_updateMaxMinWidth(){let e=0;const t=Object.keys(this._widgets);for(let i=0,n=t.length;i<n;i++){const o=t[i],a=this._widgets[o].widget.getMinContentWidthInPx?.();typeof a<"u"&&(e=Math.max(e,a))}this._context.viewLayout.setOverlayWidgetsMinWidth(e)}_renderWidget(e,t){const i=e.domNode;if(e.preference===null){i.setTop("");return}const n=2*this._verticalScrollbarWidth+this._minimapWidth;if(e.preference===0||e.preference===1){if(e.preference===1){const o=i.domNode.clientHeight;i.setTop(this._editorHeight-o-2*this._horizontalScrollbarHeight)}else i.setTop(0);e.stack!==void 0?(i.setTop(t[e.preference]),t[e.preference]+=i.domNode.clientWidth):i.setRight(n)}else if(e.preference===2)i.domNode.style.right="50%",e.stack!==void 0?(i.setTop(t[2]),t[2]+=i.domNode.clientHeight):i.setTop(0);else{const{top:o,left:r}=e.preference;if(this._context.configuration.options.get(51)&&this._widgetCanOverflow(e.widget)){const l=this._viewDomNodeRect;i.setTop(o+l.top),i.setLeft(r+l.left),i.setPosition("fixed")}else i.setTop(o),i.setLeft(r),i.setPosition("absolute")}}prepareRender(e){this._viewDomNodeRect=Li(this._viewDomNode.domNode)}render(e){this._domNode.setWidth(this._editorWidth);const t=Object.keys(this._widgets),i=Array.from({length:3},()=>0);t.sort((n,o)=>(this._widgets[n].stack||0)-(this._widgets[o].stack||0));for(let n=0,o=t.length;n<o;n++){const r=t[n];this._renderWidget(this._widgets[r],i)}}}class X${constructor(e,t,i,n){this._viewportBrand=void 0,this.top=e|0,this.left=t|0,this.width=i|0,this.height=n|0}}class lve{constructor(e,t){this.tabSize=e,this.data=t}}class QW{constructor(e,t,i,n,o,r,a){this._viewLineDataBrand=void 0,this.content=e,this.continuesWithWrappedLine=t,this.minColumn=i,this.maxColumn=n,this.startVisibleColumn=o,this.tokens=r,this.inlineDecorations=a}}class Er{constructor(e,t,i,n,o,r,a,l,c,d,h,u){this.minColumn=e,this.maxColumn=t,this.content=i,this.continuesWithWrappedLine=n,this.isBasicASCII=Er.isBasicASCII(i,r),this.containsRTL=Er.containsRTL(i,this.isBasicASCII,o),this.tokens=a,this.inlineDecorations=l,this.tabSize=c,this.startVisibleColumn=d,this.textDirection=h,this.hasVariableFonts=u}static isBasicASCII(e,t){return t?q0(e):!0}static containsRTL(e,t,i){return!t&&i?iv(e):!1}}class vS{constructor(e,t,i){this.color=e,this.zIndex=t,this.data=i}static compareByRenderingProps(e,t){return e.zIndex===t.zIndex?e.color<t.color?-1:e.color>t.color?1:0:e.zIndex-t.zIndex}static equals(e,t){return e.color===t.color&&e.zIndex===t.zIndex&&ni(e.data,t.data)}static equalsArr(e,t){return ni(e,t,vS.equals)}}class cve{constructor(e,t){const i=e.options;this.lineHeight=i.get(75),this.pixelRatio=i.get(163),this.overviewRulerLanes=i.get(95),this.renderBorder=i.get(94);const n=t.getColor(awe);this.borderColor=n?n.toString():null,this.hideCursor=i.get(68);const o=t.getColor(hA);this.cursorColorSingle=o?o.transparent(.7).toString():null;const r=t.getColor(uee);this.cursorColorPrimary=r?r.transparent(.7).toString():null;const a=t.getColor(gee);this.cursorColorSecondary=a?a.transparent(.7).toString():null,this.themeType=t.type;const l=i.get(81),c=l.enabled,d=l.side,h=t.getColor(lwe),u=yi.getDefaultBackground();h?this.backgroundColor=h:c&&d==="right"?this.backgroundColor=u:this.backgroundColor=null;const f=i.get(165).overviewRuler;this.top=f.top,this.right=f.right,this.domWidth=f.width,this.domHeight=f.height,this.overviewRulerLanes===0?(this.canvasWidth=0,this.canvasHeight=0):(this.canvasWidth=this.domWidth*this.pixelRatio|0,this.canvasHeight=this.domHeight*this.pixelRatio|0);const[p,_]=this._initLanes(1,this.canvasWidth,this.overviewRulerLanes);this.x=p,this.w=_}_initLanes(e,t,i){const n=t-e;if(i>=3){const o=Math.floor(n/3),r=Math.floor(n/3),a=n-o-r,l=e,c=l+o,d=l+o+a;return[[0,l,c,l,d,l,c,l],[0,o,a,o+a,r,o+a+r,a+r,o+a+r]]}else if(i===2){const o=Math.floor(n/2),r=n-o,a=e,l=a+o;return[[0,a,a,a,l,a,a,a],[0,o,o,o,r,o+r,o+r,o+r]]}else{const o=e,r=n;return[[0,o,o,o,o,o,o,o],[0,r,r,r,r,r,r,r]]}}equals(e){return this.lineHeight===e.lineHeight&&this.pixelRatio===e.pixelRatio&&this.overviewRulerLanes===e.overviewRulerLanes&&this.renderBorder===e.renderBorder&&this.borderColor===e.borderColor&&this.hideCursor===e.hideCursor&&this.cursorColorSingle===e.cursorColorSingle&&this.cursorColorPrimary===e.cursorColorPrimary&&this.cursorColorSecondary===e.cursorColorSecondary&&this.themeType===e.themeType&&q.equals(this.backgroundColor,e.backgroundColor)&&this.top===e.top&&this.right===e.right&&this.domWidth===e.domWidth&&this.domHeight===e.domHeight&&this.canvasWidth===e.canvasWidth&&this.canvasHeight===e.canvasHeight}}class dve extends no{constructor(e){super(e),this._actualShouldRender=0,this._renderedDecorations=[],this._renderedCursorPositions=[],this._domNode=wt(document.createElement("canvas")),this._domNode.setClassName("decorationsOverviewRuler"),this._domNode.setPosition("absolute"),this._domNode.setLayerHinting(!0),this._domNode.setContain("strict"),this._domNode.setAttribute("aria-hidden","true"),this._updateSettings(!1),this._tokensColorTrackerListener=yi.onDidChange(t=>{t.changedColorMap&&this._updateSettings(!0)}),this._cursorPositions=[{position:new A(1,1),color:this._settings.cursorColorSingle}]}dispose(){super.dispose(),this._tokensColorTrackerListener.dispose()}_updateSettings(e){const t=new cve(this._context.configuration,this._context.theme);return this._settings&&this._settings.equals(t)?!1:(this._settings=t,this._domNode.setTop(this._settings.top),this._domNode.setRight(this._settings.right),this._domNode.setWidth(this._settings.domWidth),this._domNode.setHeight(this._settings.domHeight),this._domNode.domNode.width=this._settings.canvasWidth,this._domNode.domNode.height=this._settings.canvasHeight,e&&this._render(),!0)}_markRenderingIsNeeded(){return this._actualShouldRender=2,!0}_markRenderingIsMaybeNeeded(){return this._actualShouldRender=1,!0}onConfigurationChanged(e){return this._updateSettings(!1)?this._markRenderingIsNeeded():!1}onCursorStateChanged(e){this._cursorPositions=[];for(let t=0,i=e.selections.length;t<i;t++){let n=this._settings.cursorColorSingle;i>1&&(n=t===0?this._settings.cursorColorPrimary:this._settings.cursorColorSecondary),this._cursorPositions.push({position:e.selections[t].getPosition(),color:n})}return this._cursorPositions.sort((t,i)=>A.compare(t.position,i.position)),this._markRenderingIsMaybeNeeded()}onDecorationsChanged(e){return e.affectsOverviewRuler?this._markRenderingIsMaybeNeeded():!1}onFlushed(e){return this._markRenderingIsNeeded()}onScrollChanged(e){return e.scrollHeightChanged?this._markRenderingIsNeeded():!1}onZonesChanged(e){return this._markRenderingIsNeeded()}onThemeChanged(e){return this._updateSettings(!1)?this._markRenderingIsNeeded():!1}getDomNode(){return this._domNode.domNode}prepareRender(e){}render(e){this._render(),this._actualShouldRender=0}_render(){const e=this._settings.backgroundColor;if(this._settings.overviewRulerLanes===0){this._domNode.setBackgroundColor(e?q.Format.CSS.formatHexA(e):""),this._domNode.setDisplay("none");return}const t=this._context.viewModel.getAllOverviewRulerDecorations(this._context.theme);if(t.sort(vS.compareByRenderingProps),this._actualShouldRender===1&&!vS.equalsArr(this._renderedDecorations,t)&&(this._actualShouldRender=2),this._actualShouldRender===1&&!ni(this._renderedCursorPositions,this._cursorPositions,(f,p)=>f.position.lineNumber===p.position.lineNumber&&f.color===p.color)&&(this._actualShouldRender=2),this._actualShouldRender===1)return;this._renderedDecorations=t,this._renderedCursorPositions=this._cursorPositions,this._domNode.setDisplay("block");const i=this._settings.canvasWidth,n=this._settings.canvasHeight,o=this._settings.lineHeight,r=this._context.viewLayout,a=this._context.viewLayout.getScrollHeight(),l=n/a,c=6*this._settings.pixelRatio|0,d=c/2|0,h=this._domNode.domNode.getContext("2d");e?e.isOpaque()?(h.fillStyle=q.Format.CSS.formatHexA(e),h.fillRect(0,0,i,n)):(h.clearRect(0,0,i,n),h.fillStyle=q.Format.CSS.formatHexA(e),h.fillRect(0,0,i,n)):h.clearRect(0,0,i,n);const u=this._settings.x,g=this._settings.w;for(const f of t){const p=f.color,_=f.data;h.fillStyle=p;let b=0,w=0,C=0;for(let v=0,y=_.length/3;v<y;v++){const x=_[3*v],L=_[3*v+1],E=_[3*v+2];let I=r.getVerticalOffsetForLineNumber(L)*l|0,R=(r.getVerticalOffsetForLineNumber(E)+o)*l|0;if(R-I<c){let P=(I+R)/2|0;P<d?P=d:P+d>n&&(P=n-d),I=P-d,R=P+d}I>C+1||x!==b?(v!==0&&h.fillRect(u[b],w,g[b],C-w),b=x,w=I,C=R):R>C&&(C=R)}h.fillRect(u[b],w,g[b],C-w)}if(!this._settings.hideCursor){const f=2*this._settings.pixelRatio|0,p=f/2|0,_=this._settings.x[7],b=this._settings.w[7];let w=-100,C=-100,v=null;for(let y=0,x=this._cursorPositions.length;y<x;y++){const L=this._cursorPositions[y].color;if(!L)continue;const E=this._cursorPositions[y].position;let I=r.getVerticalOffsetForLineNumber(E.lineNumber)*l|0;I<p?I=p:I+p>n&&(I=n-p);const R=I-p,O=R+f;R>C+1||L!==v?(y!==0&&v&&h.fillRect(_,w,b,C-w),w=R,C=O):O>C&&(C=O),v=L,h.fillStyle=L}v&&h.fillRect(_,w,b,C-w)}this._settings.renderBorder&&this._settings.borderColor&&this._settings.overviewRulerLanes>0&&(h.beginPath(),h.lineWidth=1,h.strokeStyle=this._settings.borderColor,h.moveTo(0,0),h.lineTo(0,n),h.moveTo(1,0),h.lineTo(i,0),h.stroke())}}class Q${constructor(e,t,i){this._colorZoneBrand=void 0,this.from=e|0,this.to=t|0,this.colorId=i|0}static compare(e,t){return e.colorId===t.colorId?e.from===t.from?e.to-t.to:e.from-t.from:e.colorId-t.colorId}}class Tee{constructor(e,t,i,n){this._overviewRulerZoneBrand=void 0,this.startLineNumber=e,this.endLineNumber=t,this.heightInLines=i,this.color=n,this._colorZone=null}static compare(e,t){return e.color===t.color?e.startLineNumber===t.startLineNumber?e.heightInLines===t.heightInLines?e.endLineNumber-t.endLineNumber:e.heightInLines-t.heightInLines:e.startLineNumber-t.startLineNumber:e.color<t.color?-1:1}setColorZone(e){this._colorZone=e}getColorZones(){return this._colorZone}}class hve{constructor(e){this._getVerticalOffsetForLine=e,this._zones=[],this._colorZonesInvalid=!1,this._lineHeight=0,this._domWidth=0,this._domHeight=0,this._outerHeight=0,this._pixelRatio=1,this._lastAssignedId=0,this._color2Id=Object.create(null),this._id2Color=[]}getId2Color(){return this._id2Color}setZones(e){this._zones=e,this._zones.sort(Tee.compare)}setLineHeight(e){return this._lineHeight===e?!1:(this._lineHeight=e,this._colorZonesInvalid=!0,!0)}setPixelRatio(e){this._pixelRatio=e,this._colorZonesInvalid=!0}getDOMWidth(){return this._domWidth}getCanvasWidth(){return this._domWidth*this._pixelRatio}setDOMWidth(e){return this._domWidth===e?!1:(this._domWidth=e,this._colorZonesInvalid=!0,!0)}getDOMHeight(){return this._domHeight}getCanvasHeight(){return this._domHeight*this._pixelRatio}setDOMHeight(e){return this._domHeight===e?!1:(this._domHeight=e,this._colorZonesInvalid=!0,!0)}getOuterHeight(){return this._outerHeight}setOuterHeight(e){return this._outerHeight===e?!1:(this._outerHeight=e,this._colorZonesInvalid=!0,!0)}resolveColorZones(){const e=this._colorZonesInvalid,t=Math.floor(this._lineHeight),i=Math.floor(this.getCanvasHeight()),n=Math.floor(this._outerHeight),o=i/n,r=Math.floor(4*this._pixelRatio/2),a=[];for(let l=0,c=this._zones.length;l<c;l++){const d=this._zones[l];if(!e){const v=d.getColorZones();if(v){a.push(v);continue}}const h=this._getVerticalOffsetForLine(d.startLineNumber),u=d.heightInLines===0?this._getVerticalOffsetForLine(d.endLineNumber)+t:h+d.heightInLines*t,g=Math.floor(o*h),f=Math.floor(o*u);let p=Math.floor((g+f)/2),_=f-p;_<r&&(_=r),p-_<0&&(p=_),p+_>i&&(p=i-_);const b=d.color;let w=this._color2Id[b];w||(w=++this._lastAssignedId,this._color2Id[b]=w,this._id2Color[w]=b);const C=new Q$(p-_,p+_,w);d.setColorZone(C),a.push(C)}return this._colorZonesInvalid=!1,a.sort(Q$.compare),a}}class uve extends c_{constructor(e,t){super(),this._context=e;const i=this._context.configuration.options;this._domNode=wt(document.createElement("canvas")),this._domNode.setClassName(t),this._domNode.setPosition("absolute"),this._domNode.setLayerHinting(!0),this._domNode.setContain("strict"),this._zoneManager=new hve(n=>this._context.viewLayout.getVerticalOffsetForLineNumber(n)),this._zoneManager.setDOMWidth(0),this._zoneManager.setDOMHeight(0),this._zoneManager.setOuterHeight(this._context.viewLayout.getScrollHeight()),this._zoneManager.setLineHeight(i.get(75)),this._zoneManager.setPixelRatio(i.get(163)),this._context.addEventHandler(this)}dispose(){this._context.removeEventHandler(this),super.dispose()}onConfigurationChanged(e){const t=this._context.configuration.options;return e.hasChanged(75)&&(this._zoneManager.setLineHeight(t.get(75)),this._render()),e.hasChanged(163)&&(this._zoneManager.setPixelRatio(t.get(163)),this._domNode.setWidth(this._zoneManager.getDOMWidth()),this._domNode.setHeight(this._zoneManager.getDOMHeight()),this._domNode.domNode.width=this._zoneManager.getCanvasWidth(),this._domNode.domNode.height=this._zoneManager.getCanvasHeight(),this._render()),!0}onFlushed(e){return this._render(),!0}onScrollChanged(e){return e.scrollHeightChanged&&(this._zoneManager.setOuterHeight(e.scrollHeight),this._render()),!0}onZonesChanged(e){return this._render(),!0}getDomNode(){return this._domNode.domNode}setLayout(e){this._domNode.setTop(e.top),this._domNode.setRight(e.right);let t=!1;t=this._zoneManager.setDOMWidth(e.width)||t,t=this._zoneManager.setDOMHeight(e.height)||t,t&&(this._domNode.setWidth(this._zoneManager.getDOMWidth()),this._domNode.setHeight(this._zoneManager.getDOMHeight()),this._domNode.domNode.width=this._zoneManager.getCanvasWidth(),this._domNode.domNode.height=this._zoneManager.getCanvasHeight(),this._render())}setZones(e){this._zoneManager.setZones(e),this._render()}_render(){if(this._zoneManager.getOuterHeight()===0)return!1;const e=this._zoneManager.getCanvasWidth(),t=this._zoneManager.getCanvasHeight(),i=this._zoneManager.resolveColorZones(),n=this._zoneManager.getId2Color(),o=this._domNode.domNode.getContext("2d");return o.clearRect(0,0,e,t),i.length>0&&this._renderOneLane(o,i,n,e),!0}_renderOneLane(e,t,i,n){let o=0,r=0,a=0;for(const l of t){const c=l.colorId,d=l.from,h=l.to;c!==o?(e.fillRect(0,r,n,a-r),o=c,e.fillStyle=i[o],r=d,a=h):a>=d?a=Math.max(a,h):(e.fillRect(0,r,n,a-r),r=d,a=h)}e.fillRect(0,r,n,a-r)}}class gve extends no{constructor(e){super(e),this.domNode=wt(document.createElement("div")),this.domNode.setAttribute("role","presentation"),this.domNode.setAttribute("aria-hidden","true"),this.domNode.setClassName("view-rulers"),this._renderedRulers=[];const t=this._context.configuration.options;this._rulers=t.get(116),this._typicalHalfwidthCharacterWidth=t.get(59).typicalHalfwidthCharacterWidth}dispose(){super.dispose()}onConfigurationChanged(e){const t=this._context.configuration.options;return this._rulers=t.get(116),this._typicalHalfwidthCharacterWidth=t.get(59).typicalHalfwidthCharacterWidth,!0}onScrollChanged(e){return e.scrollHeightChanged}prepareRender(e){}_ensureRulersCount(){const e=this._renderedRulers.length,t=this._rulers.length;if(e===t)return;if(e<t){const{tabSize:n}=this._context.viewModel.model.getOptions(),o=n;let r=t-e;for(;r>0;){const a=wt(document.createElement("div"));a.setClassName("view-ruler"),a.setWidth(o),this.domNode.appendChild(a),this._renderedRulers.push(a),r--}return}let i=e-t;for(;i>0;){const n=this._renderedRulers.pop();this.domNode.removeChild(n),i--}}render(e){this._ensureRulersCount();for(let t=0,i=this._rulers.length;t<i;t++){const n=this._renderedRulers[t],o=this._rulers[t];n.setBoxShadow(o.color?`1px 0 0 0 ${o.color} inset`:""),n.setHeight(Math.min(e.scrollHeight,1e6)),n.setLeft(o.column*this._typicalHalfwidthCharacterWidth)}}}class fve extends no{constructor(e){super(e),this._scrollTop=0,this._width=0,this._updateWidth(),this._shouldShow=!1;const i=this._context.configuration.options.get(117);this._useShadows=i.useShadows,this._domNode=wt(document.createElement("div")),this._domNode.setAttribute("role","presentation"),this._domNode.setAttribute("aria-hidden","true")}dispose(){super.dispose()}_updateShouldShow(){const e=this._useShadows&&this._scrollTop>0;return this._shouldShow!==e?(this._shouldShow=e,!0):!1}getDomNode(){return this._domNode}_updateWidth(){const t=this._context.configuration.options.get(165);t.minimap.renderMinimap===0||t.minimap.minimapWidth>0&&t.minimap.minimapLeft===0?this._width=t.width:this._width=t.width-t.verticalScrollbarWidth}onConfigurationChanged(e){const i=this._context.configuration.options.get(117);return this._useShadows=i.useShadows,this._updateWidth(),this._updateShouldShow(),!0}onScrollChanged(e){return this._scrollTop=e.scrollTop,this._updateShouldShow()}prepareRender(e){}render(e){this._domNode.setWidth(this._width),this._domNode.setClassName(this._shouldShow?"scroll-decoration":"")}}class pve{constructor(e){this.left=e.left,this.width=e.width,this.startStyle=null,this.endStyle=null}}class mve{constructor(e,t){this.lineNumber=e,this.ranges=t}}function _ve(s){return new pve(s)}function bve(s){return new mve(s.lineNumber,s.ranges.map(_ve))}const Oi=class Oi extends Dg{constructor(e){super(),this._previousFrameVisibleRangesWithStyle=[],this._context=e;const t=this._context.configuration.options;this._roundedSelection=t.get(115),this._typicalHalfwidthCharacterWidth=t.get(59).typicalHalfwidthCharacterWidth,this._selections=[],this._renderResult=null,this._context.addEventHandler(this)}dispose(){this._context.removeEventHandler(this),this._renderResult=null,super.dispose()}onConfigurationChanged(e){const t=this._context.configuration.options;return this._roundedSelection=t.get(115),this._typicalHalfwidthCharacterWidth=t.get(59).typicalHalfwidthCharacterWidth,!0}onCursorStateChanged(e){return this._selections=e.selections.slice(0),!0}onDecorationsChanged(e){return!0}onFlushed(e){return!0}onLinesChanged(e){return!0}onLinesDeleted(e){return!0}onLinesInserted(e){return!0}onScrollChanged(e){return e.scrollTopChanged}onZonesChanged(e){return!0}_visibleRangesHaveGaps(e){for(let t=0,i=e.length;t<i;t++)if(e[t].ranges.length>1)return!0;return!1}_enrichVisibleRangesWithStyle(e,t,i){const n=this._typicalHalfwidthCharacterWidth/4;let o=null,r=null;if(i&&i.length>0&&t.length>0){const a=t[0].lineNumber;if(a===e.startLineNumber)for(let c=0;!o&&c<i.length;c++)i[c].lineNumber===a&&(o=i[c].ranges[0]);const l=t[t.length-1].lineNumber;if(l===e.endLineNumber)for(let c=i.length-1;!r&&c>=0;c--)i[c].lineNumber===l&&(r=i[c].ranges[0]);o&&!o.startStyle&&(o=null),r&&!r.startStyle&&(r=null)}for(let a=0,l=t.length;a<l;a++){const c=t[a].ranges[0],d=c.left,h=c.left+c.width,u={top:0,bottom:0},g={top:0,bottom:0};if(a>0){const f=t[a-1].ranges[0].left,p=t[a-1].ranges[0].left+t[a-1].ranges[0].width;Jx(d-f)<n?u.top=2:d>f&&(u.top=1),Jx(h-p)<n?g.top=2:f<h&&h<p&&(g.top=1)}else o&&(u.top=o.startStyle.top,g.top=o.endStyle.top);if(a+1<l){const f=t[a+1].ranges[0].left,p=t[a+1].ranges[0].left+t[a+1].ranges[0].width;Jx(d-f)<n?u.bottom=2:f<d&&d<p&&(u.bottom=1),Jx(h-p)<n?g.bottom=2:h<p&&(g.bottom=1)}else r&&(u.bottom=r.startStyle.bottom,g.bottom=r.endStyle.bottom);c.startStyle=u,c.endStyle=g}}_getVisibleRangesWithStyle(e,t,i){const o=(t.linesVisibleRangesForRange(e,!0)||[]).map(bve);return!this._visibleRangesHaveGaps(o)&&this._roundedSelection&&this._enrichVisibleRangesWithStyle(t.visibleRange,o,i),o}_createSelectionPiece(e,t,i,n,o){return'<div class="cslr '+i+'" style="top:'+e.toString()+"px;bottom:"+t.toString()+"px;left:"+n.toString()+"px;width:"+o.toString()+'px;"></div>'}_actualRenderOneSelection(e,t,i,n){if(n.length===0)return;const o=!!n[0].ranges[0].startStyle,r=n[0].lineNumber,a=n[n.length-1].lineNumber;for(let l=0,c=n.length;l<c;l++){const d=n[l],h=d.lineNumber,u=h-t,g=i&&h===r?1:0,f=i&&h!==r&&h===a?1:0;let p="",_="";for(let b=0,w=d.ranges.length;b<w;b++){const C=d.ranges[b];if(o){const y=C.startStyle,x=C.endStyle;if(y.top===1||y.bottom===1){p+=this._createSelectionPiece(g,f,Oi.SELECTION_CLASS_NAME,C.left-Oi.ROUNDED_PIECE_WIDTH,Oi.ROUNDED_PIECE_WIDTH);let L=Oi.EDITOR_BACKGROUND_CLASS_NAME;y.top===1&&(L+=" "+Oi.SELECTION_TOP_RIGHT),y.bottom===1&&(L+=" "+Oi.SELECTION_BOTTOM_RIGHT),p+=this._createSelectionPiece(g,f,L,C.left-Oi.ROUNDED_PIECE_WIDTH,Oi.ROUNDED_PIECE_WIDTH)}if(x.top===1||x.bottom===1){p+=this._createSelectionPiece(g,f,Oi.SELECTION_CLASS_NAME,C.left+C.width,Oi.ROUNDED_PIECE_WIDTH);let L=Oi.EDITOR_BACKGROUND_CLASS_NAME;x.top===1&&(L+=" "+Oi.SELECTION_TOP_LEFT),x.bottom===1&&(L+=" "+Oi.SELECTION_BOTTOM_LEFT),p+=this._createSelectionPiece(g,f,L,C.left+C.width,Oi.ROUNDED_PIECE_WIDTH)}}let v=Oi.SELECTION_CLASS_NAME;if(o){const y=C.startStyle,x=C.endStyle;y.top===0&&(v+=" "+Oi.SELECTION_TOP_LEFT),y.bottom===0&&(v+=" "+Oi.SELECTION_BOTTOM_LEFT),x.top===0&&(v+=" "+Oi.SELECTION_TOP_RIGHT),x.bottom===0&&(v+=" "+Oi.SELECTION_BOTTOM_RIGHT)}_+=this._createSelectionPiece(g,f,v,C.left,C.width)}e[u][0]+=p,e[u][1]+=_}}prepareRender(e){const t=[],i=e.visibleRange.startLineNumber,n=e.visibleRange.endLineNumber;for(let r=i;r<=n;r++){const a=r-i;t[a]=["",""]}const o=[];for(let r=0,a=this._selections.length;r<a;r++){const l=this._selections[r];if(l.isEmpty()){o[r]=null;continue}const c=this._getVisibleRangesWithStyle(l,e,this._previousFrameVisibleRangesWithStyle[r]);o[r]=c,this._actualRenderOneSelection(t,i,this._selections.length>1,c)}this._previousFrameVisibleRangesWithStyle=o,this._renderResult=t.map(([r,a])=>r+a)}render(e,t){if(!this._renderResult)return"";const i=t-e;return i<0||i>=this._renderResult.length?"":this._renderResult[i]}};Oi.SELECTION_CLASS_NAME="selected-text",Oi.SELECTION_TOP_LEFT="top-left-radius",Oi.SELECTION_BOTTOM_LEFT="bottom-left-radius",Oi.SELECTION_TOP_RIGHT="top-right-radius",Oi.SELECTION_BOTTOM_RIGHT="bottom-right-radius",Oi.EDITOR_BACKGROUND_CLASS_NAME="monaco-editor-background",Oi.ROUNDED_PIECE_WIDTH=10;let uF=Oi;ua((s,e)=>{const t=s.getColor(lme);t&&!t.isTransparent()&&e.addRule(`.monaco-editor .view-line span.inline-selected-text { color: ${t}; }`)});function Jx(s){return s<0?-s:s}class J${constructor(e,t,i,n,o,r,a){this.top=e,this.left=t,this.paddingLeft=i,this.width=n,this.height=o,this.textContent=r,this.textContentClassName=a}}var $d;(function(s){s[s.Single=0]="Single",s[s.MultiPrimary=1]="MultiPrimary",s[s.MultiSecondary=2]="MultiSecondary"})($d||($d={}));class eU{constructor(e,t){this._context=e;const i=this._context.configuration.options,n=i.get(59);this._cursorStyle=i.get(161),this._typicalHalfwidthCharacterWidth=n.typicalHalfwidthCharacterWidth,this._lineCursorWidth=Math.min(i.get(37),this._typicalHalfwidthCharacterWidth),this._lineCursorHeight=i.get(38),this._isVisible=!0,this._domNode=wt(document.createElement("div")),this._domNode.setClassName(`cursor ${pw}`),this._domNode.setHeight(this._context.viewLayout.getLineHeightForLineNumber(1)),this._domNode.setTop(0),this._domNode.setLeft(0),In(this._domNode,n),this._domNode.setDisplay("none"),this._position=new A(1,1),this._pluralityClass="",this.setPlurality(t),this._lastRenderedContent="",this._renderData=null}getDomNode(){return this._domNode}getPosition(){return this._position}setPlurality(e){switch(e){default:case $d.Single:this._pluralityClass="";break;case $d.MultiPrimary:this._pluralityClass="cursor-primary";break;case $d.MultiSecondary:this._pluralityClass="cursor-secondary";break}}show(){this._isVisible||(this._domNode.setVisibility("inherit"),this._isVisible=!0)}hide(){this._isVisible&&(this._domNode.setVisibility("hidden"),this._isVisible=!1)}onConfigurationChanged(e){const t=this._context.configuration.options,i=t.get(59);return this._cursorStyle=t.get(161),this._typicalHalfwidthCharacterWidth=i.typicalHalfwidthCharacterWidth,this._lineCursorWidth=Math.min(t.get(37),this._typicalHalfwidthCharacterWidth),this._lineCursorHeight=t.get(38),In(this._domNode,i),!0}onCursorPositionChanged(e,t){return t?this._domNode.domNode.style.transitionProperty="none":this._domNode.domNode.style.transitionProperty="",this._position=e,!0}_getGraphemeAwarePosition(){const{lineNumber:e,column:t}=this._position,i=this._context.viewModel.getLineContent(e),[n,o]=Wde(i,t-1);return[new A(e,n+1),i.substring(n,o)]}_prepareRender(e){let t="",i="";const[n,o]=this._getGraphemeAwarePosition(),r=this._context.viewLayout.getLineHeightForLineNumber(n.lineNumber),a=this._lineCursorHeight===0?r:Math.min(r,this._lineCursorHeight),l=(r-a)/2;if(this._cursorStyle===mn.Line||this._cursorStyle===mn.LineThin){const p=e.visibleRangeForPosition(n);if(!p||p.outsideRenderedLine)return null;const _=be(this._domNode.domNode);let b;this._cursorStyle===mn.Line?(b=Kz(_,this._lineCursorWidth>0?this._lineCursorWidth:2),b>2&&(t=o,i=this._getTokenClassName(n))):b=Kz(_,1);let w=p.left,C=0;b>=2&&w>=1&&(C=1,w-=C);const v=e.getVerticalOffsetForLineNumber(n.lineNumber)-e.bigNumbersDelta+l;return new J$(v,w,C,b,a,t,i)}const c=e.linesVisibleRangesForRange(new k(n.lineNumber,n.column,n.lineNumber,n.column+o.length),!1);if(!c||c.length===0)return null;const d=c[0];if(d.outsideRenderedLine||d.ranges.length===0)return null;const h=d.ranges[0],u=o==="	"?this._typicalHalfwidthCharacterWidth:h.width<1?this._typicalHalfwidthCharacterWidth:h.width;this._cursorStyle===mn.Block&&(t=o,i=this._getTokenClassName(n));let g=e.getVerticalOffsetForLineNumber(n.lineNumber)-e.bigNumbersDelta,f=r;return(this._cursorStyle===mn.Underline||this._cursorStyle===mn.UnderlineThin)&&(g+=r-2,f=2),new J$(g,h.left,0,u,f,t,i)}_getTokenClassName(e){const t=this._context.viewModel.getViewLineData(e.lineNumber),i=t.tokens.findTokenIndexAtOffset(e.column-1);return t.tokens.getClassName(i)}prepareRender(e){this._renderData=this._prepareRender(e)}render(e){return this._renderData?(this._lastRenderedContent!==this._renderData.textContent&&(this._lastRenderedContent=this._renderData.textContent,this._domNode.domNode.textContent=this._lastRenderedContent),this._domNode.setClassName(`cursor ${this._pluralityClass} ${pw} ${this._renderData.textContentClassName}`),this._domNode.setDisplay("block"),this._domNode.setTop(this._renderData.top),this._domNode.setLeft(this._renderData.left),this._domNode.setPaddingLeft(this._renderData.paddingLeft),this._domNode.setWidth(this._renderData.width),this._domNode.setLineHeight(this._renderData.height),this._domNode.setHeight(this._renderData.height),{domNode:this._domNode.domNode,position:this._position,contentLeft:this._renderData.left,height:this._renderData.height,width:2}):(this._domNode.setDisplay("none"),null)}}const Ly=class Ly extends no{constructor(e){super(e);const t=this._context.configuration.options;this._readOnly=t.get(104),this._cursorBlinking=t.get(32),this._cursorStyle=t.get(161),this._cursorSmoothCaretAnimation=t.get(33),this._editContextEnabled=t.get(170),this._selectionIsEmpty=!0,this._isComposingInput=!1,this._isVisible=!1,this._primaryCursor=new eU(this._context,$d.Single),this._secondaryCursors=[],this._renderData=[],this._domNode=wt(document.createElement("div")),this._domNode.setAttribute("role","presentation"),this._domNode.setAttribute("aria-hidden","true"),this._updateDomClassName(),this._domNode.appendChild(this._primaryCursor.getDomNode()),this._startCursorBlinkAnimation=new $o,this._cursorFlatBlinkInterval=new cW,this._blinkingEnabled=!1,this._editorHasFocus=!1,this._updateBlinking()}dispose(){super.dispose(),this._startCursorBlinkAnimation.dispose(),this._cursorFlatBlinkInterval.dispose()}getDomNode(){return this._domNode}onCompositionStart(e){return this._isComposingInput=!0,this._updateBlinking(),!0}onCompositionEnd(e){return this._isComposingInput=!1,this._updateBlinking(),!0}onConfigurationChanged(e){const t=this._context.configuration.options;this._readOnly=t.get(104),this._cursorBlinking=t.get(32),this._cursorStyle=t.get(161),this._cursorSmoothCaretAnimation=t.get(33),this._editContextEnabled=t.get(170),this._updateBlinking(),this._updateDomClassName(),this._primaryCursor.onConfigurationChanged(e);for(let i=0,n=this._secondaryCursors.length;i<n;i++)this._secondaryCursors[i].onConfigurationChanged(e);return!0}_onCursorPositionChanged(e,t,i){const n=this._secondaryCursors.length!==t.length||this._cursorSmoothCaretAnimation==="explicit"&&i!==3;if(this._primaryCursor.setPlurality(t.length?$d.MultiPrimary:$d.Single),this._primaryCursor.onCursorPositionChanged(e,n),this._updateBlinking(),this._secondaryCursors.length<t.length){const o=t.length-this._secondaryCursors.length;for(let r=0;r<o;r++){const a=new eU(this._context,$d.MultiSecondary);this._domNode.domNode.insertBefore(a.getDomNode().domNode,this._primaryCursor.getDomNode().domNode.nextSibling),this._secondaryCursors.push(a)}}else if(this._secondaryCursors.length>t.length){const o=this._secondaryCursors.length-t.length;for(let r=0;r<o;r++)this._domNode.removeChild(this._secondaryCursors[0].getDomNode()),this._secondaryCursors.splice(0,1)}for(let o=0;o<t.length;o++)this._secondaryCursors[o].onCursorPositionChanged(t[o],n)}onCursorStateChanged(e){const t=[];for(let n=0,o=e.selections.length;n<o;n++)t[n]=e.selections[n].getPosition();this._onCursorPositionChanged(t[0],t.slice(1),e.reason);const i=e.selections[0].isEmpty();return this._selectionIsEmpty!==i&&(this._selectionIsEmpty=i,this._updateDomClassName()),!0}onDecorationsChanged(e){return!0}onFlushed(e){return!0}onFocusChanged(e){return this._editorHasFocus=e.isFocused,this._updateBlinking(),!1}onLinesChanged(e){return!0}onLinesDeleted(e){return!0}onLinesInserted(e){return!0}onScrollChanged(e){return!0}onTokensChanged(e){const t=i=>{for(let n=0,o=e.ranges.length;n<o;n++)if(e.ranges[n].fromLineNumber<=i.lineNumber&&i.lineNumber<=e.ranges[n].toLineNumber)return!0;return!1};if(t(this._primaryCursor.getPosition()))return!0;for(const i of this._secondaryCursors)if(t(i.getPosition()))return!0;return!1}onZonesChanged(e){return!0}_getCursorBlinking(){return this._isComposingInput&&!this._editContextEnabled||!this._editorHasFocus?0:this._readOnly?5:this._cursorBlinking}_updateBlinking(){this._startCursorBlinkAnimation.cancel(),this._cursorFlatBlinkInterval.cancel();const e=this._getCursorBlinking(),t=e===0,i=e===5;t?this._hide():this._show(),this._blinkingEnabled=!1,this._updateDomClassName(),!t&&!i&&(e===1?this._cursorFlatBlinkInterval.cancelAndSet(()=>{this._isVisible?this._hide():this._show()},Ly.BLINK_INTERVAL,be(this._domNode.domNode)):this._startCursorBlinkAnimation.setIfNotSet(()=>{this._blinkingEnabled=!0,this._updateDomClassName()},Ly.BLINK_INTERVAL))}_updateDomClassName(){this._domNode.setClassName(this._getClassName())}_getClassName(){let e="cursors-layer";switch(this._selectionIsEmpty||(e+=" has-selection"),this._cursorStyle){case mn.Line:e+=" cursor-line-style";break;case mn.Block:e+=" cursor-block-style";break;case mn.Underline:e+=" cursor-underline-style";break;case mn.LineThin:e+=" cursor-line-thin-style";break;case mn.BlockOutline:e+=" cursor-block-outline-style";break;case mn.UnderlineThin:e+=" cursor-underline-thin-style";break;default:e+=" cursor-line-style"}if(this._blinkingEnabled)switch(this._getCursorBlinking()){case 1:e+=" cursor-blink";break;case 2:e+=" cursor-smooth";break;case 3:e+=" cursor-phase";break;case 4:e+=" cursor-expand";break;case 5:e+=" cursor-solid";break;default:e+=" cursor-solid"}else e+=" cursor-solid";return(this._cursorSmoothCaretAnimation==="on"||this._cursorSmoothCaretAnimation==="explicit")&&(e+=" cursor-smooth-caret-animation"),e}_show(){this._primaryCursor.show();for(let e=0,t=this._secondaryCursors.length;e<t;e++)this._secondaryCursors[e].show();this._isVisible=!0}_hide(){this._primaryCursor.hide();for(let e=0,t=this._secondaryCursors.length;e<t;e++)this._secondaryCursors[e].hide();this._isVisible=!1}prepareRender(e){this._primaryCursor.prepareRender(e);for(let t=0,i=this._secondaryCursors.length;t<i;t++)this._secondaryCursors[t].prepareRender(e)}render(e){const t=[];let i=0;const n=this._primaryCursor.render(e);n&&(t[i++]=n);for(let o=0,r=this._secondaryCursors.length;o<r;o++){const a=this._secondaryCursors[o].render(e);a&&(t[i++]=a)}this._renderData=t}getLastRenderData(){return this._renderData}};Ly.BLINK_INTERVAL=500;let gF=Ly;ua((s,e)=>{const t=[{class:".cursor",foreground:hA,background:UW},{class:".cursor-primary",foreground:uee,background:$be},{class:".cursor-secondary",foreground:gee,background:Ube}];for(const i of t){const n=s.getColor(i.foreground);if(n){let o=s.getColor(i.background);o||(o=n.opposite()),e.addRule(`.monaco-editor .cursors-layer ${i.class} { background-color: ${n}; border-color: ${n}; color: ${o}; }`),Cl(s.type)&&e.addRule(`.monaco-editor .cursors-layer.has-selection ${i.class} { border-left: 1px solid ${o}; border-right: 1px solid ${o}; }`)}}});const oM=()=>{throw new Error("Invalid change accessor")};class wve extends no{constructor(e){super(e);const t=this._context.configuration.options,i=t.get(165);this._lineHeight=t.get(75),this._contentWidth=i.contentWidth,this._contentLeft=i.contentLeft,this.domNode=wt(document.createElement("div")),this.domNode.setClassName("view-zones"),this.domNode.setPosition("absolute"),this.domNode.setAttribute("role","presentation"),this.domNode.setAttribute("aria-hidden","true"),this.marginDomNode=wt(document.createElement("div")),this.marginDomNode.setClassName("margin-view-zones"),this.marginDomNode.setPosition("absolute"),this.marginDomNode.setAttribute("role","presentation"),this.marginDomNode.setAttribute("aria-hidden","true"),this._zones={}}dispose(){super.dispose(),this._zones={}}_recomputeWhitespacesProps(){const e=this._context.viewLayout.getWhitespaces(),t=new Map;for(const n of e)t.set(n.id,n);let i=!1;return this._context.viewModel.changeWhitespace(n=>{const o=Object.keys(this._zones);for(let r=0,a=o.length;r<a;r++){const l=o[r],c=this._zones[l],d=this._computeWhitespaceProps(c.delegate);c.isInHiddenArea=d.isInHiddenArea;const h=t.get(l);h&&(h.afterLineNumber!==d.afterViewLineNumber||h.height!==d.heightInPx)&&(n.changeOneWhitespace(l,d.afterViewLineNumber,d.heightInPx),this._safeCallOnComputedHeight(c.delegate,d.heightInPx),i=!0)}}),i}onConfigurationChanged(e){const t=this._context.configuration.options,i=t.get(165);return this._lineHeight=t.get(75),this._contentWidth=i.contentWidth,this._contentLeft=i.contentLeft,e.hasChanged(75)&&this._recomputeWhitespacesProps(),!0}onLineMappingChanged(e){return this._recomputeWhitespacesProps()}onLinesDeleted(e){return!0}onScrollChanged(e){return e.scrollTopChanged||e.scrollWidthChanged}onZonesChanged(e){return!0}onLinesInserted(e){return!0}_getZoneOrdinal(e){return e.ordinal??e.afterColumn??1e4}_computeWhitespaceProps(e){if(e.afterLineNumber===0)return{isInHiddenArea:!1,afterViewLineNumber:0,heightInPx:this._heightInPixels(e),minWidthInPx:this._minWidthInPixels(e)};let t;if(typeof e.afterColumn<"u")t=this._context.viewModel.model.validatePosition({lineNumber:e.afterLineNumber,column:e.afterColumn});else{const r=this._context.viewModel.model.validatePosition({lineNumber:e.afterLineNumber,column:1}).lineNumber;t=new A(r,this._context.viewModel.model.getLineMaxColumn(r))}let i;t.column===this._context.viewModel.model.getLineMaxColumn(t.lineNumber)?i=this._context.viewModel.model.validatePosition({lineNumber:t.lineNumber+1,column:1}):i=this._context.viewModel.model.validatePosition({lineNumber:t.lineNumber,column:t.column+1});const n=this._context.viewModel.coordinatesConverter.convertModelPositionToViewPosition(t,e.afterColumnAffinity,!0),o=e.showInHiddenAreas||this._context.viewModel.coordinatesConverter.modelPositionIsVisible(i);return{isInHiddenArea:!o,afterViewLineNumber:n.lineNumber,heightInPx:o?this._heightInPixels(e):0,minWidthInPx:this._minWidthInPixels(e)}}changeViewZones(e){let t=!1;return this._context.viewModel.changeWhitespace(i=>{const n={addZone:o=>(t=!0,this._addZone(i,o)),removeZone:o=>{o&&(t=this._removeZone(i,o)||t)},layoutZone:o=>{o&&(t=this._layoutZone(i,o)||t)}};vve(e,n),n.addZone=oM,n.removeZone=oM,n.layoutZone=oM}),t}_addZone(e,t){const i=this._computeWhitespaceProps(t),o={whitespaceId:e.insertWhitespace(i.afterViewLineNumber,this._getZoneOrdinal(t),i.heightInPx,i.minWidthInPx),delegate:t,isInHiddenArea:i.isInHiddenArea,isVisible:!1,domNode:wt(t.domNode),marginDomNode:t.marginDomNode?wt(t.marginDomNode):null};return this._safeCallOnComputedHeight(o.delegate,i.heightInPx),o.domNode.setPosition("absolute"),o.domNode.domNode.style.width="100%",o.domNode.setDisplay("none"),o.domNode.setAttribute("monaco-view-zone",o.whitespaceId),this.domNode.appendChild(o.domNode),o.marginDomNode&&(o.marginDomNode.setPosition("absolute"),o.marginDomNode.domNode.style.width="100%",o.marginDomNode.setDisplay("none"),o.marginDomNode.setAttribute("monaco-view-zone",o.whitespaceId),this.marginDomNode.appendChild(o.marginDomNode)),this._zones[o.whitespaceId]=o,this.setShouldRender(),o.whitespaceId}_removeZone(e,t){if(this._zones.hasOwnProperty(t)){const i=this._zones[t];return delete this._zones[t],e.removeWhitespace(i.whitespaceId),i.domNode.removeAttribute("monaco-visible-view-zone"),i.domNode.removeAttribute("monaco-view-zone"),i.domNode.domNode.remove(),i.marginDomNode&&(i.marginDomNode.removeAttribute("monaco-visible-view-zone"),i.marginDomNode.removeAttribute("monaco-view-zone"),i.marginDomNode.domNode.remove()),this.setShouldRender(),!0}return!1}_layoutZone(e,t){if(this._zones.hasOwnProperty(t)){const i=this._zones[t],n=this._computeWhitespaceProps(i.delegate);return i.isInHiddenArea=n.isInHiddenArea,e.changeOneWhitespace(i.whitespaceId,n.afterViewLineNumber,n.heightInPx),this._safeCallOnComputedHeight(i.delegate,n.heightInPx),this.setShouldRender(),!0}return!1}shouldSuppressMouseDownOnViewZone(e){return this._zones.hasOwnProperty(e)?!!this._zones[e].delegate.suppressMouseDown:!1}_heightInPixels(e){return typeof e.heightInPx=="number"?e.heightInPx:typeof e.heightInLines=="number"?this._lineHeight*e.heightInLines:this._lineHeight}_minWidthInPixels(e){return typeof e.minWidthInPx=="number"?e.minWidthInPx:0}_safeCallOnComputedHeight(e,t){if(typeof e.onComputedHeight=="function")try{e.onComputedHeight(t)}catch(i){Oe(i)}}_safeCallOnDomNodeTop(e,t){if(typeof e.onDomNodeTop=="function")try{e.onDomNodeTop(t)}catch(i){Oe(i)}}prepareRender(e){}render(e){const t=e.viewportData.whitespaceViewportData,i={};let n=!1;for(const r of t)this._zones[r.id].isInHiddenArea||(i[r.id]=r,n=!0);const o=Object.keys(this._zones);for(let r=0,a=o.length;r<a;r++){const l=o[r],c=this._zones[l];let d=0,h=0,u="none";i.hasOwnProperty(l)?(d=i[l].verticalOffset-e.bigNumbersDelta,h=i[l].height,u="block",c.isVisible||(c.domNode.setAttribute("monaco-visible-view-zone","true"),c.isVisible=!0),this._safeCallOnDomNodeTop(c.delegate,e.getScrolledTopFromAbsoluteTop(i[l].verticalOffset))):(c.isVisible&&(c.domNode.removeAttribute("monaco-visible-view-zone"),c.isVisible=!1),this._safeCallOnDomNodeTop(c.delegate,e.getScrolledTopFromAbsoluteTop(-1e6))),c.domNode.setTop(d),c.domNode.setHeight(h),c.domNode.setDisplay(u),c.marginDomNode&&(c.marginDomNode.setTop(d),c.marginDomNode.setHeight(h),c.marginDomNode.setDisplay(u))}n&&(this.domNode.setWidth(Math.max(e.scrollWidth,this._contentWidth)),this.marginDomNode.setWidth(this._contentLeft))}}function vve(s,e){try{return s(e)}catch(t){Oe(t);return}}class Cve extends Dg{constructor(e){super(),this._context=e,this._options=new tU(this._context.configuration),this._selection=[],this._renderResult=null,this._context.addEventHandler(this)}dispose(){this._context.removeEventHandler(this),this._renderResult=null,super.dispose()}onConfigurationChanged(e){const t=new tU(this._context.configuration);return this._options.equals(t)?e.hasChanged(165):(this._options=t,!0)}onCursorStateChanged(e){return this._selection=e.selections,this._options.renderWhitespace==="selection"}onDecorationsChanged(e){return!0}onFlushed(e){return!0}onLinesChanged(e){return!0}onLinesDeleted(e){return!0}onLinesInserted(e){return!0}onScrollChanged(e){return e.scrollTopChanged}onZonesChanged(e){return!0}prepareRender(e){if(this._options.renderWhitespace==="none"){this._renderResult=null;return}const t=e.visibleRange.startLineNumber,n=e.visibleRange.endLineNumber-t+1,o=new Array(n);for(let r=0;r<n;r++)o[r]=!0;this._renderResult=[];for(let r=e.viewportData.startLineNumber;r<=e.viewportData.endLineNumber;r++){const a=r-e.viewportData.startLineNumber,l=this._context.viewModel.getViewLineRenderingData(r);let c=null;if(this._options.renderWhitespace==="selection"){const d=this._selection;for(const h of d){if(h.endLineNumber<r||h.startLineNumber>r)continue;const u=h.startLineNumber===r?h.startColumn:l.minColumn,g=h.endLineNumber===r?h.endColumn:l.maxColumn;u<g&&(c||(c=[]),c.push(new me(u-1,g-1)))}}this._renderResult[a]=this._applyRenderWhitespace(e,r,c,l)}}_applyRenderWhitespace(e,t,i,n){if(n.hasVariableFonts||this._options.renderWhitespace==="selection"&&!i||this._options.renderWhitespace==="trailing"&&n.continuesWithWrappedLine)return"";const o=this._context.theme.getColor(qW),r=this._options.renderWithSVG,a=n.content,l=this._options.stopRenderingLineAfter===-1?a.length:Math.min(this._options.stopRenderingLineAfter,a.length),c=n.continuesWithWrappedLine,d=n.minColumn-1,h=this._options.renderWhitespace==="boundary",u=this._options.renderWhitespace==="trailing",g=e.getLineHeightForLineNumber(t),f=this._options.middotWidth,p=this._options.wsmiddotWidth,_=this._options.spaceWidth,b=Math.abs(p-_),w=Math.abs(f-_),C=b<w?11825:183,v=this._options.canUseHalfwidthRightwardsArrow;let y="",x=!1,L=As(a),E;L===-1?(x=!0,L=l,E=l):E=Ma(a);let I=0,R=i&&i[I],O=0;for(let P=d;P<l;P++){const B=a.charCodeAt(P);if(R&&R.endExclusive<=P&&(I++,R=i&&i[I]),B!==9&&B!==32||u&&!x&&P<=E)continue;if(h&&P>=L&&P<=E&&B===32){const z=P-1>=0?a.charCodeAt(P-1):0,Q=P+1<l?a.charCodeAt(P+1):0;if(z!==32&&Q!==32)continue}if(h&&c&&P===l-1){const z=P-1>=0?a.charCodeAt(P-1):0;if(B===32&&z!==32&&z!==9)continue}if(i&&!(R&&R.start<=P&&P<R.endExclusive))continue;const $=e.visibleRangeForPosition(new A(t,P+1));$&&(r?(O=Math.max(O,$.left),B===9?y+=this._renderArrow(g,_,$.left):y+=`<circle cx="${($.left+_/2).toFixed(2)}" cy="${(g/2).toFixed(2)}" r="${(_/7).toFixed(2)}" />`):B===9?y+=`<div class="mwh" style="left:${$.left}px;height:${g}px;">${v?"￫":"→"}</div>`:y+=`<div class="mwh" style="left:${$.left}px;height:${g}px;">${String.fromCharCode(C)}</div>`)}return r?(O=Math.round(O+_),`<svg style="bottom:0;position:absolute;width:${O}px;height:${g}px" viewBox="0 0 ${O} ${g}" xmlns="http://www.w3.org/2000/svg" fill="${o}">`+y+"</svg>"):y}_renderArrow(e,t,i){const n=t/7,o=t,r=e/2,a=i,l={x:0,y:n/2},c={x:100/125*o,y:l.y},d={x:c.x-.2*c.x,y:c.y+.2*c.x},h={x:d.x+.1*c.x,y:d.y+.1*c.x},u={x:h.x+.35*c.x,y:h.y-.35*c.x},g={x:u.x,y:-u.y},f={x:h.x,y:-h.y},p={x:d.x,y:-d.y},_={x:c.x,y:-c.y},b={x:l.x,y:-l.y};return`<path d="M ${[l,c,d,h,u,g,f,p,_,b].map(v=>`${(a+v.x).toFixed(2)} ${(r+v.y).toFixed(2)}`).join(" L ")}" />`}render(e,t){if(!this._renderResult)return"";const i=t-e;return i<0||i>=this._renderResult.length?"":this._renderResult[i]}}class tU{constructor(e){const t=e.options,i=t.get(59),n=t.get(47);n==="off"?(this.renderWhitespace="none",this.renderWithSVG=!1):n==="svg"?(this.renderWhitespace=t.get(113),this.renderWithSVG=!0):(this.renderWhitespace=t.get(113),this.renderWithSVG=!1),this.spaceWidth=i.spaceWidth,this.middotWidth=i.middotWidth,this.wsmiddotWidth=i.wsmiddotWidth,this.canUseHalfwidthRightwardsArrow=i.canUseHalfwidthRightwardsArrow,this.lineHeight=t.get(75),this.stopRenderingLineAfter=t.get(133)}equals(e){return this.renderWhitespace===e.renderWhitespace&&this.renderWithSVG===e.renderWithSVG&&this.spaceWidth===e.spaceWidth&&this.middotWidth===e.middotWidth&&this.wsmiddotWidth===e.wsmiddotWidth&&this.canUseHalfwidthRightwardsArrow===e.canUseHalfwidthRightwardsArrow&&this.lineHeight===e.lineHeight&&this.stopRenderingLineAfter===e.stopRenderingLineAfter}}class yve{constructor(e,t,i,n){this.selections=e,this.startLineNumber=t.startLineNumber|0,this.endLineNumber=t.endLineNumber|0,this.relativeVerticalOffset=t.relativeVerticalOffset,this.bigNumbersDelta=t.bigNumbersDelta|0,this.lineHeight=t.lineHeight|0,this.whitespaceViewportData=i,this._model=n,this.visibleRange=new k(t.startLineNumber,this._model.getLineMinColumn(t.startLineNumber),t.endLineNumber,this._model.getLineMaxColumn(t.endLineNumber))}getViewLineRenderingData(e){return this._model.getViewportViewLineRenderingData(this.visibleRange,e)}getDecorationsInViewport(){return this._model.getDecorationsInViewport(this.visibleRange)}}class Sve{get type(){return this._theme.type}get value(){return this._theme}constructor(e){this._theme=e}update(e){this._theme=e}getColor(e){return this._theme.getColor(e)}}class xve{constructor(e,t,i){this.configuration=e,this.theme=new Sve(t),this.viewModel=i,this.viewLayout=i.viewLayout}addEventHandler(e){this.viewModel.addViewEventHandler(e)}removeEventHandler(e){this.viewModel.removeViewEventHandler(e)}}const dp=new Float32Array([1,0,1,1,0,1,0,0,0,1,1,0]);function Tm(s){if(!s)throw new Error(`Value "${s}" cannot be null`);return s}function Nee(s,e,t){let i=new e.ResizeObserver(n=>{const o=n.find(l=>l.target===s);if(!o)return;if(!("devicePixelContentBoxSize"in o)){i?.disconnect(),i=void 0;return}const r=o.devicePixelContentBoxSize[0].inlineSize,a=o.devicePixelContentBoxSize[0].blockSize;r>0&&a>0&&t(r,a)});try{i.observe(s,{box:["device-pixel-content-box"]})}catch{throw i.disconnect(),i=void 0,new Ce("Could not observe device pixel dimensions")}return pe(()=>i?.disconnect())}var Lve=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o};let kve=0;class CS extends F{get cacheKey(){return`${this.fontFamily}_${this.fontSize}px`}constructor(e,t,i,n){super(),this.fontSize=e,this.fontFamily=t,this.devicePixelRatio=i,this._decorationStyleCache=n,this.id=kve++,this._workGlyph={source:null,boundingBox:{left:0,bottom:0,right:0,top:0},originOffset:{x:0,y:0},fontBoundingBoxAscent:0,fontBoundingBoxDescent:0},this._workGlyphConfig={chars:void 0,tokenMetadata:0,decorationStyleSetId:0},this._antiAliasing=Ye?"greyscale":"subpixel";const o=Math.ceil(this.fontSize*i);this._canvas=new OffscreenCanvas(o*3,o*3),this._ctx=Tm(this._canvas.getContext("2d",{willReadFrequently:!0,alpha:this._antiAliasing==="greyscale"})),this._ctx.textBaseline="top",this._ctx.fillStyle="#FFFFFF",this._ctx.font=`${o}px ${this.fontFamily}`,this._textMetrics=this._ctx.measureText("A")}rasterizeGlyph(e,t,i,n){return e===""?{source:this._canvas,boundingBox:{top:0,left:0,bottom:-1,right:-1},originOffset:{x:0,y:0},fontBoundingBoxAscent:0,fontBoundingBoxDescent:0}:this._workGlyphConfig.chars===e&&this._workGlyphConfig.tokenMetadata===t&&this._workGlyphConfig.decorationStyleSetId===i?this._workGlyph:(this._workGlyphConfig.chars=e,this._workGlyphConfig.tokenMetadata=t,this._workGlyphConfig.decorationStyleSetId=i,this._rasterizeGlyph(e,t,i,n))}_rasterizeGlyph(e,t,i,n){const o=Math.ceil(this.fontSize*this.devicePixelRatio),r=o*3;this._canvas.width!==r&&(this._canvas.width=r,this._canvas.height=r),this._ctx.save();const a=(t&15)/10,l=ds.getBackground(t),c=n[l],d=this._decorationStyleCache.getStyleSet(i);this._antiAliasing==="subpixel"?(this._ctx.fillStyle=c,this._ctx.fillRect(0,0,this._canvas.width,this._canvas.height)):this._ctx.clearRect(0,0,this._canvas.width,this._canvas.height);const h=new Eg(200),u=ds.getFontStyle(t);u&1&&h.appendString("italic "),d?.bold!==void 0?d.bold&&h.appendString("bold "):u&2&&h.appendString("bold "),h.appendString(`${o}px ${this.fontFamily}`),this._ctx.font=h.build();const g=o,f=o;d?.color!==void 0?this._ctx.fillStyle=`#${d.color.toString(16).padStart(8,"0")}`:this._ctx.fillStyle=n[ds.getForeground(t)],this._ctx.textBaseline="top",d?.opacity!==void 0&&(this._ctx.globalAlpha=d.opacity),this._ctx.fillText(e,g+a,f),this._ctx.restore();const p=this._ctx.getImageData(0,0,this._canvas.width,this._canvas.height);if(this._antiAliasing==="subpixel"){const _=parseInt(c.substring(1,3),16),b=parseInt(c.substring(3,5),16),w=parseInt(c.substring(5,7),16);this._clearColor(p,_,b,w),this._ctx.putImageData(p,0,0)}return this._findGlyphBoundingBox(p,this._workGlyph.boundingBox),this._workGlyph.source=this._canvas,this._workGlyph.originOffset.x=this._workGlyph.boundingBox.left-g,this._workGlyph.originOffset.y=this._workGlyph.boundingBox.top-f,this._workGlyph.fontBoundingBoxAscent=this._textMetrics.fontBoundingBoxAscent,this._workGlyph.fontBoundingBoxDescent=this._textMetrics.fontBoundingBoxDescent,this._workGlyph}_clearColor(e,t,i,n){for(let o=0;o<e.data.length;o+=4)e.data[o]===t&&e.data[o+1]===i&&e.data[o+2]===n&&(e.data[o+3]=0)}_findGlyphBoundingBox(e,t){const i=this._canvas.height,n=this._canvas.width;let o=!1;for(let r=0;r<i;r++){for(let a=0;a<n;a++){const l=r*n*4+a*4+3;if(e.data[l]!==0){t.top=r,o=!0;break}}if(o)break}t.left=0,o=!1;for(let r=0;r<n;r++){for(let a=0;a<i;a++){const l=a*n*4+r*4+3;if(e.data[l]!==0){t.left=r,o=!0;break}}if(o)break}t.right=n,o=!1;for(let r=n-1;r>=t.left;r--){for(let a=0;a<i;a++){const l=a*n*4+r*4+3;if(e.data[l]!==0){t.right=r,o=!0;break}}if(o)break}t.bottom=t.top,o=!1;for(let r=i-1;r>=0;r--){for(let a=0;a<n;a++){const l=r*n*4+a*4+3;if(e.data[l]!==0){t.bottom=r,o=!0;break}}if(o)break}}getTextMetrics(e){return this._ctx.measureText(e)}}Lve([Ai],CS.prototype,"cacheKey",null);var Eve=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},Ive=function(s,e){return function(t,i){e(t,i,s)}};let rI=class extends F{constructor(e){super(),this._logService=e,this._tasks=[],this._i=0,this._register(pe(()=>this.clear()))}enqueue(e){this._tasks.push(e),this._start()}clear(){this._idleCallback&&(this._cancelCallback(this._idleCallback),this._idleCallback=void 0),this._i=0,this._tasks.length=0}_start(){this._idleCallback||(this._idleCallback=this._requestCallback(this._process.bind(this)))}_process(e){this._idleCallback=void 0;let t=0,i=0,n=e.timeRemaining(),o=0;for(;this._i<this._tasks.length;){if(t=Date.now(),this._tasks[this._i]()||this._i++,t=Math.max(1,Date.now()-t),i=Math.max(t,i),o=e.timeRemaining(),i*1.5>o){n-t<-20&&this._logService.warn(`task queue exceeded allotted deadline by ${Math.abs(Math.round(n-t))}ms`),this._start();return}n=o}this.clear()}};rI=Eve([Ive(0,jt)],rI);class Dve extends rI{_requestCallback(e){return kt().setTimeout(()=>e(this._createDeadline(16)))}_cancelCallback(e){kt().clearTimeout(e)}_createDeadline(e){const t=Date.now()+e;return{timeRemaining:()=>Math.max(0,t-Date.now())}}}class Tve extends rI{_requestCallback(e){return kt().requestIdleCallback(e)}_cancelCallback(e){kt().cancelIdleCallback(e)}}const Nve="requestIdleCallback"in kt()?Tve:Dve;class Rve{constructor(e,t){this._canvas=e,this._textureIndex=t,this._currentRow={x:0,y:0,h:0},this._allocatedGlyphs=new Set,this._nextIndex=0,this._ctx=Tm(this._canvas.getContext("2d",{willReadFrequently:!0}))}allocate(e){const t=e.boundingBox.right-e.boundingBox.left+1,i=e.boundingBox.bottom-e.boundingBox.top+1;if(t>this._canvas.width||i>this._canvas.height)throw new Ce("Glyph is too large for the atlas page");if(e.boundingBox.right-e.boundingBox.left+1>this._canvas.width-this._currentRow.x&&(this._currentRow.x=0,this._currentRow.y+=this._currentRow.h,this._currentRow.h=1),this._currentRow.y+e.boundingBox.bottom-e.boundingBox.top+1>this._canvas.height)return;this._ctx.drawImage(e.source,e.boundingBox.left,e.boundingBox.top,t,i,this._currentRow.x,this._currentRow.y,t,i);const n={pageIndex:this._textureIndex,glyphIndex:this._nextIndex++,x:this._currentRow.x,y:this._currentRow.y,w:t,h:i,originOffsetX:e.originOffset.x,originOffsetY:e.originOffset.y,fontBoundingBoxAscent:e.fontBoundingBoxAscent,fontBoundingBoxDescent:e.fontBoundingBoxDescent};return this._currentRow.x+=t,this._currentRow.h=Math.max(this._currentRow.h,i),this._allocatedGlyphs.add(n),n}getUsagePreview(){const e=this._canvas.width,t=this._canvas.height,i=new OffscreenCanvas(e,t),n=Tm(i.getContext("2d"));n.fillStyle="#808080",n.fillRect(0,0,e,t);const o=new Map,r=new Map;for(const a of this._allocatedGlyphs)o.set(a.y,Math.max(o.get(a.y)??0,a.h)),r.set(a.y,Math.max(r.get(a.y)??0,a.x+a.w));for(const a of this._allocatedGlyphs)n.fillStyle="#4040FF",n.fillRect(a.x,a.y,a.w,a.h),n.fillStyle="#FF0000",n.fillRect(a.x,a.y+a.h,a.w,o.get(a.y)-a.h);for(const[a,l]of r.entries())a!==this._currentRow.y&&(n.fillStyle="#FF0000",n.fillRect(l,a,e-l,o.get(a)));return i.convertToBlob()}getStats(){const e=this._canvas.width,t=this._canvas.height;let i=0,n=0;const o=e*t,r=new Map,a=new Map;for(const l of this._allocatedGlyphs)r.set(l.y,Math.max(r.get(l.y)??0,l.h)),a.set(l.y,Math.max(a.get(l.y)??0,l.x+l.w));for(const l of this._allocatedGlyphs)i+=l.w*l.h,n+=l.w*(r.get(l.y)-l.h);for(const[l,c]of a.entries())l!==this._currentRow.y&&(n+=(e-c)*r.get(l));return[`page${this._textureIndex}:`,`     Total: ${o} (${e}x${t})`,`      Used: ${i} (${(i/o*100).toPrecision(2)}%)`,`    Wasted: ${n} (${(n/o*100).toPrecision(2)}%)`,`Efficiency: ${(i/(i+n)*100).toPrecision(2)}%`].join(`
`)}}class Ave{constructor(e,t,i){this._canvas=e,this._textureIndex=t,this._slabs=[],this._activeSlabsByDims=new VR,this._unusedRects=[],this._openRegionsByHeight=new Map,this._openRegionsByWidth=new Map,this._allocatedGlyphs=new Set,this._nextIndex=0,this._ctx=Tm(this._canvas.getContext("2d",{willReadFrequently:!0})),this._slabW=Math.min(i?.slabW??64<<Math.max(Math.floor(kt().devicePixelRatio)-1,0),this._canvas.width),this._slabH=Math.min(i?.slabH??this._slabW,this._canvas.height),this._slabsPerRow=Math.floor(this._canvas.width/this._slabW),this._slabsPerColumn=Math.floor(this._canvas.height/this._slabH)}allocate(e){const t=e.boundingBox.right-e.boundingBox.left+1,i=e.boundingBox.bottom-e.boundingBox.top+1;if(t>this._canvas.width||i>this._canvas.height)throw new Ce("Glyph is too large for the atlas page");if(t>this._slabW||i>this._slabH){if(this._allocatedGlyphs.size>0)return;let c=this._canvas.width;for(;t<c/2&&i<c/2;)c/=2;this._slabW=c,this._slabH=c,this._slabsPerRow=Math.floor(this._canvas.width/this._slabW),this._slabsPerColumn=Math.floor(this._canvas.height/this._slabH)}const n={w:t,h:i};let o=this._activeSlabsByDims.get(n.w,n.h);if(o){const c=Math.floor(this._slabW/o.entryW)*Math.floor(this._slabH/o.entryH);o.count>=c&&(o=void 0)}let r,a;if(!o)if(t<i){const c=this._openRegionsByWidth.get(t);if(c?.length)for(let d=c.length-1;d>=0;d--){const h=c[d];if(h.w>=t&&h.h>=i){r=h.x,a=h.y,t<h.w&&this._unusedRects.push({x:h.x+t,y:h.y,w:h.w-t,h:i}),h.y+=i,h.h-=i,h.h===0&&(d===c.length-1?c.pop():this._unusedRects.splice(d,1));break}}}else{const c=this._openRegionsByHeight.get(i);if(c?.length)for(let d=c.length-1;d>=0;d--){const h=c[d];if(h.w>=t&&h.h>=i){r=h.x,a=h.y,i<h.h&&this._unusedRects.push({x:h.x,y:h.y+i,w:t,h:h.h-i}),h.x+=t,h.w-=t,h.h===0&&(d===c.length-1?c.pop():this._unusedRects.splice(d,1));break}}}if(r===void 0||a===void 0){if(!o){if(this._slabs.length>=this._slabsPerRow*this._slabsPerColumn)return;o={x:Math.floor(this._slabs.length%this._slabsPerRow)*this._slabW,y:Math.floor(this._slabs.length/this._slabsPerRow)*this._slabH,entryW:n.w,entryH:n.h,count:0};const d=this._slabW%o.entryW,h=this._slabH%o.entryH;d&&iU(this._openRegionsByWidth,d,{x:o.x+this._slabW-d,w:d,y:o.y,h:this._slabH-(h??0)}),h&&iU(this._openRegionsByHeight,h,{x:o.x,w:this._slabW,y:o.y+this._slabH-h,h}),this._slabs.push(o),this._activeSlabsByDims.set(o,n.w,n.h)}const c=Math.floor(this._slabW/o.entryW);r=o.x+Math.floor(o.count%c)*o.entryW,a=o.y+Math.floor(o.count/c)*o.entryH,o.count++}this._ctx.drawImage(e.source,e.boundingBox.left,e.boundingBox.top,t,i,r,a,t,i);const l={pageIndex:this._textureIndex,glyphIndex:this._nextIndex++,x:r,y:a,w:t,h:i,originOffsetX:e.originOffset.x,originOffsetY:e.originOffset.y,fontBoundingBoxAscent:e.fontBoundingBoxAscent,fontBoundingBoxDescent:e.fontBoundingBoxDescent};return this._allocatedGlyphs.add(l),l}getUsagePreview(){const e=this._canvas.width,t=this._canvas.height,i=new OffscreenCanvas(e,t),n=Tm(i.getContext("2d"));n.fillStyle="#808080",n.fillRect(0,0,e,t);let o=0,r=0,a=0;const l=64<<Math.floor(kt().devicePixelRatio)-1,c=l;for(const h of this._slabs){let u=0,g=0;for(let _=0;_<h.count;_++)u+h.entryW>l&&(u=0,g+=h.entryH),n.fillStyle="#FF0000",n.fillRect(h.x+u,h.y+g,h.entryW,h.entryH),o+=h.entryW*h.entryH,u+=h.entryW;const f=Math.floor(l/h.entryW),p=Math.floor(c/h.entryH);h.entryW*f*h.entryH*p}for(const h of this._allocatedGlyphs)r+=h.w*h.h,n.fillStyle="#4040FF",n.fillRect(h.x,h.y,h.w,h.h);const d=Array.from(this._openRegionsByWidth.values()).flat().concat(Array.from(this._openRegionsByHeight.values()).flat());for(const h of d)n.fillStyle="#FF000088",n.fillRect(h.x,h.y,h.w,h.h),a+=h.w*h.h;return n.globalAlpha=.5,n.drawImage(this._canvas,0,0),n.globalAlpha=1,i.convertToBlob()}getStats(){const e=this._canvas.width,t=this._canvas.height;let i=0,n=0,o=0,r=0,a=0;const l=e*t,c=64<<Math.floor(kt().devicePixelRatio)-1,d=c;for(const f of this._slabs){let p=0,_=0;for(let v=0;v<f.count;v++)p+f.entryW>c&&(p=0,_+=f.entryH),i+=f.entryW*f.entryH,p+=f.entryW;const b=Math.floor(c/f.entryW),w=Math.floor(d/f.entryH),C=f.entryW*b*f.entryH*w;o+=c*d-C}for(const f of this._allocatedGlyphs)n+=f.w*f.h;const h=Array.from(this._openRegionsByWidth.values()).flat().concat(Array.from(this._openRegionsByHeight.values()).flat());for(const f of h)a+=f.w*f.h;const u=o-a;r=i-(n-u);const g=n/(n+r+a);return[`page[${this._textureIndex}]:`,`     Total: ${l}px (${e}x${t})`,`      Used: ${n}px (${(n/l*100).toFixed(2)}%)`,`    Wasted: ${r}px (${(r/l*100).toFixed(2)}%)`,`Restricted: ${a}px (${(a/l*100).toFixed(2)}%) (hard to allocate)`,`Efficiency: ${g===1?"100":(g*100).toFixed(2)}%`,`     Slabs: ${this._slabs.length} of ${Math.floor(this._canvas.width/c)*Math.floor(this._canvas.height/d)}`].join(`
`)}}function iU(s,e,t){let i=s.get(e);i||(i=[],s.set(e,i)),i.push(t)}var Mve=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},nU=function(s,e){return function(t,i){e(t,i,s)}},fF,yp;let al=(yp=class extends F{get version(){return this._version}get usedArea(){return this._usedArea}get source(){return this._canvas}get glyphs(){return this._glyphInOrderSet.values()}constructor(e,t,i,n,o){switch(super(),this._logService=n,this._version=0,this._usedArea={left:0,top:0,right:0,bottom:0},this._glyphMap=new VR,this._glyphInOrderSet=new Set,this._canvas=new OffscreenCanvas(t,t),this._colorMap=o.getColorTheme().tokenColorMap,i){case"shelf":this._allocator=new Rve(this._canvas,e);break;case"slab":this._allocator=new Ave(this._canvas,e);break;default:this._allocator=i(this._canvas,e);break}this._register(pe(()=>{this._canvas.width=1,this._canvas.height=1}))}getGlyph(e,t,i,n){return this._glyphMap.get(t,i,n,e.cacheKey)??this._createGlyph(e,t,i,n)}_createGlyph(e,t,i,n){if(this._glyphInOrderSet.size>=fF.maximumGlyphCount)return;const o=e.rasterizeGlyph(t,i,n,this._colorMap),r=this._allocator.allocate(o);if(r!==void 0)return this._glyphMap.set(r,t,i,n,e.cacheKey),this._glyphInOrderSet.add(r),this._version++,this._usedArea.right=Math.max(this._usedArea.right,r.x+r.w-1),this._usedArea.bottom=Math.max(this._usedArea.bottom,r.y+r.h-1),this._logService.getLevel()===pn.Trace&&this._logService.trace("New glyph",{chars:t,tokenMetadata:i,decorationStyleSetId:n,rasterizedGlyph:o,glyph:r}),r}getUsagePreview(){return this._allocator.getUsagePreview()}getStats(){return this._allocator.getStats()}},fF=yp,yp.maximumGlyphCount=5e3,yp);al=fF=Mve([nU(3,jt),nU(4,wi)],al);var Ove=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},sU=function(s,e){return function(t,i){e(t,i,s)}},Rk,Sp;let wu=(Sp=class extends F{get pages(){return this._pages}constructor(e,t,i,n,o){super(),this._maxTextureSize=e,this._decorationStyleCache=i,this._themeService=n,this._instantiationService=o,this._warmUpTask=this._register(new vt),this._warmedUpRasterizers=new Set,this._pages=[],this._glyphPageIndex=new VR,this._onDidDeleteGlyphs=this._register(new M),this.onDidDeleteGlyphs=this._onDidDeleteGlyphs.event,this._allocatorType=t?.allocatorType??"slab",this._register(oe.runAndSubscribe(this._themeService.onDidColorThemeChange,()=>{this._colorMap&&this.clear(),this._colorMap=this._themeService.getColorTheme().tokenColorMap}));const r=Math.max(1,Math.floor(kt().devicePixelRatio));this.pageSize=Math.min(1024*r,this._maxTextureSize),this._initFirstPage(),this._register(pe(()=>xt(this._pages)))}_initFirstPage(){const e=this._instantiationService.createInstance(al,0,this.pageSize,this._allocatorType);this._pages.push(e);const t=new CS(1,"",1,this._decorationStyleCache);e.getGlyph(t,"",0,0),t.dispose()}clear(){for(const e of this._pages)e.dispose();this._pages.length=0,this._glyphPageIndex.clear(),this._warmedUpRasterizers.clear(),this._warmUpTask.clear(),this._initFirstPage(),this._onDidDeleteGlyphs.fire()}getGlyph(e,t,i,n,o){return i&=-2048,i|=Math.floor(o%1*10),this._warmedUpRasterizers.has(e.id)||(this._warmUpAtlas(e),this._warmedUpRasterizers.add(e.id)),this._tryGetGlyph(this._glyphPageIndex.get(t,i,n,e.cacheKey)??0,e,t,i,n)}_tryGetGlyph(e,t,i,n,o){return this._glyphPageIndex.set(e,i,n,o,t.cacheKey),this._pages[e].getGlyph(t,i,n,o)??(e+1<this._pages.length?this._tryGetGlyph(e+1,t,i,n,o):void 0)??this._getGlyphFromNewPage(t,i,n,o)}_getGlyphFromNewPage(e,t,i,n){if(this._pages.length>=Rk.maximumPageCount)throw new Error(`Attempt to create a texture atlas page past the limit ${Rk.maximumPageCount}`);return this._pages.push(this._instantiationService.createInstance(al,this._pages.length,this.pageSize,this._allocatorType)),this._glyphPageIndex.set(this._pages.length-1,t,i,n,e.cacheKey),this._pages[this._pages.length-1].getGlyph(e,t,i,n)}getStats(){return this._pages.map(e=>e.getStats())}_warmUpAtlas(e){const t=this._colorMap;if(!t)throw new Ce("Cannot warm atlas without color map");this._warmUpTask.value?.clear();const i=this._warmUpTask.value=this._instantiationService.createInstance(Nve);for(let n=65;n<=90;n++)for(const o of t.keys())i.enqueue(()=>{for(let r=0;r<1;r+=.1)this.getGlyph(e,String.fromCharCode(n),o<<15&16744448,0,r)});for(let n=97;n<=122;n++)for(const o of t.keys())i.enqueue(()=>{for(let r=0;r<1;r+=.1)this.getGlyph(e,String.fromCharCode(n),o<<15&16744448,0,r)});for(let n=33;n<=126;n++)for(const o of t.keys())i.enqueue(()=>{for(let r=0;r<1;r+=.1)this.getGlyph(e,String.fromCharCode(n),o<<15&16744448,0,r)})}},Rk=Sp,Sp.maximumPageCount=16,Sp);wu=Rk=Ove([sU(3,wi),sU(4,_e)],wu);var yS;(function(s){s[s.Ignore=0]="Ignore",s[s.Info=1]="Info",s[s.Warning=2]="Warning",s[s.Error=3]="Error"})(yS||(yS={}));(function(s){const e="error",t="warning",i="warn",n="info",o="ignore";function r(l){return l?Hf(e,l)?s.Error:Hf(t,l)||Hf(i,l)?s.Warning:Hf(n,l)?s.Info:s.Ignore:s.Ignore}s.fromValue=r;function a(l){switch(l){case s.Error:return e;case s.Warning:return t;case s.Info:return n;default:return o}}s.toString=a})(yS||(yS={}));var fi=yS,oC=fi;const Ii=qe("notificationService");class Pve{}var fo;(function(s){async function e(n){try{if(!navigator.gpu)throw new Error("This browser does not support WebGPU");const o=await navigator.gpu.requestAdapter();if(!o)throw new Error("This browser supports WebGPU but it appears to be disabled");return rM(await o.requestDevice())}catch(o){throw n&&n(o.message),o}}s.requestDevice=e;function t(n,o,r){const a=n.createBuffer(o);return r&&n.queue.writeBuffer(a,0,vm(r)?r():r),rM(a)}s.createBuffer=t;function i(n,o){return rM(n.createTexture(o))}s.createTexture=i})(fo||(fo={}));function rM(s){return{object:s,dispose:()=>s.destroy()}}class Fve{get dataOffset(){return this._startIndex}get dirtySize(){if(!(this._startIndex===void 0||this._endIndex===void 0))return this._endIndex-this._startIndex+1}get isDirty(){return this._startIndex!==void 0}flag(e,t=1){return this._flag(e),t>1&&this._flag(e+t-1),e}_flag(e){(this._startIndex===void 0||e<this._startIndex)&&(this._startIndex=e),(this._endIndex===void 0||e>this._endIndex)&&(this._endIndex=e)}clear(){this._startIndex=void 0,this._endIndex=void 0}}function Bve(s,e){return new Wve(s,e)}class Wve extends F{get entryCount(){return this._entries.size}get dirtyTracker(){return this._dirtyTracker}constructor(e,t){super(),this.propertySpecs=e,this.capacity=t,this._dirtyTracker=new Fve,this._propertySpecsMap=new Map,this._entries=new Ns,this._onDidChange=this._register(new M),this._onDidChangeBuffer=this._register(new M),this.onDidChangeBuffer=this._onDidChangeBuffer.event,this.view=new Float32Array(t*e.length),this.buffer=this.view.buffer,this._entrySize=e.length;for(let i=0;i<e.length;i++){const n={offset:i,...e[i]};this._propertySpecsMap.set(n.name,n)}this._register(pe(()=>xt(this._entries)))}createEntry(e){this._entries.size===this.capacity&&(this._expandBuffer(),this._onDidChangeBuffer.fire());const t=new Hve(this.view,this._propertySpecsMap,this._dirtyTracker,this._entries.size,e),i=this._entries.push(t),n=[];return n.push(oe.forward(t.onDidChange,this._onDidChange)),n.push(t.onWillDispose(()=>{const o=t.i;i(),this.view.set(this.view.subarray(o*this._entrySize+2,this._entries.size*this._entrySize+2),o*this._entrySize);for(const r of this._entries)r.i>o&&r.i--;this._dirtyTracker.flag(o,(this._entries.size-o)*this._entrySize),xt(n)})),t}_expandBuffer(){this.capacity*=2;const e=new Float32Array(this.capacity*this._entrySize);e.set(this.view),this.view=e,this.buffer=this.view.buffer}}class Hve extends F{constructor(e,t,i,n,o){super(),this._view=e,this._propertySpecsMap=t,this._dirtyTracker=i,this.i=n,this._onDidChange=this._register(new M),this.onDidChange=this._onDidChange.event,this._onWillDispose=this._register(new M),this.onWillDispose=this._onWillDispose.event;for(const r of this._propertySpecsMap.values())this._view[this.i*this._propertySpecsMap.size+r.offset]=o[r.name];this._dirtyTracker.flag(this.i*this._propertySpecsMap.size,this._propertySpecsMap.size)}dispose(){this._onWillDispose.fire(),super.dispose()}setRaw(e){if(e.length!==this._propertySpecsMap.size)throw new Error(`Data length ${e.length} does not match the number of properties in the collection (${this._propertySpecsMap.size})`);this._view.set(e,this.i*this._propertySpecsMap.size),this._dirtyTracker.flag(this.i*this._propertySpecsMap.size,this._propertySpecsMap.size)}}const Vve=`

struct Vertex {
	@location(0) position: vec2f,
};

struct LayoutInfo {
	canvasDims: vec2f,
	viewportOffset: vec2f,
	viewportDims: vec2f,
}

struct ScrollOffset {
	offset: vec2f,
}

struct Shape {
	position: vec2f,
	size: vec2f,
	color: vec4f,
};

struct VSOutput {
	@builtin(position) position: vec4f,
	@location(1)       color:    vec4f,
};

// Uniforms
@group(0) @binding(1) var<uniform>       layoutInfo:      LayoutInfo;

// Storage buffers
@group(0) @binding(0)            var<storage, read> shapes:          array<Shape>;
@group(0) @binding(2)      var<uniform>       scrollOffset:    ScrollOffset;

@vertex fn vs(
	vert: Vertex,
	@builtin(instance_index) instanceIndex: u32,
	@builtin(vertex_index) vertexIndex : u32
) -> VSOutput {
	let shape = shapes[instanceIndex];

	var vsOut: VSOutput;
	vsOut.position = vec4f(
		(
			// Top left corner
			vec2f(-1,  1) +
			// Convert pixel position to clipspace
			vec2f( 2, -2) / layoutInfo.canvasDims *
			// Shape position and size
			(layoutInfo.viewportOffset - scrollOffset.offset + shape.position + vert.position * shape.size)
		),
		0.0,
		1.0
	);
	vsOut.color = shape.color;
	return vsOut;
}

@fragment fn fs(vsOut: VSOutput) -> @location(0) vec4f {
	return vsOut.color;
}
`;class zve extends c_{constructor(e,t,i,n,o,r){super(),this._context=e,this._contentLeft=t,this._devicePixelRatio=i,this._canvas=n,this._ctx=o,this._shapeBindBuffer=this._register(new vt),this._initialized=!1,this._shapeCollection=this._register(Bve([{name:"x"},{name:"y"},{name:"width"},{name:"height"},{name:"red"},{name:"green"},{name:"blue"},{name:"alpha"}],32)),this._context.addEventHandler(this),this._initWebgpu(r)}async _initWebgpu(e){if(this._device=await e,this._store.isDisposed)return;const t=navigator.gpu.getPreferredCanvasFormat();this._ctx.configure({device:this._device,format:t,alphaMode:"premultiplied"}),this._renderPassColorAttachment={view:null,loadOp:"load",storeOp:"store"},this._renderPassDescriptor={label:"Monaco rectangle renderer render pass",colorAttachments:[this._renderPassColorAttachment]};let i;{const a=new Float32Array(6),l=(c=this._canvas.width,d=this._canvas.height)=>(a[0]=c,a[1]=d,a[2]=Math.ceil(this._context.configuration.options.get(165).contentLeft*kt().devicePixelRatio),a[3]=0,a[4]=a[0]-a[2],a[5]=a[1]-a[3],a);i=this._register(fo.createBuffer(this._device,{label:"Monaco rectangle renderer uniform buffer",size:24,usage:GPUBufferUsage.UNIFORM|GPUBufferUsage.COPY_DST},()=>l())).object,this._register(Nee(this._canvas,kt(),(c,d)=>{this._device.queue.writeBuffer(i,0,l(c,d))}))}const n=2;this._scrollOffsetBindBuffer=this._register(fo.createBuffer(this._device,{label:"Monaco rectangle renderer scroll offset buffer",size:n*Float32Array.BYTES_PER_ELEMENT,usage:GPUBufferUsage.UNIFORM|GPUBufferUsage.COPY_DST})).object,this._scrollOffsetValueBuffer=new Float32Array(n);const o=()=>fo.createBuffer(this._device,{label:"Monaco rectangle renderer shape buffer",size:this._shapeCollection.buffer.byteLength,usage:GPUBufferUsage.STORAGE|GPUBufferUsage.COPY_DST});this._shapeBindBuffer.value=o(),this._register(oe.runAndSubscribe(this._shapeCollection.onDidChangeBuffer,()=>{this._shapeBindBuffer.value=o(),this._pipeline&&this._updateBindGroup(this._pipeline,i)})),this._vertexBuffer=this._register(fo.createBuffer(this._device,{label:"Monaco rectangle renderer vertex buffer",size:dp.byteLength,usage:GPUBufferUsage.VERTEX|GPUBufferUsage.COPY_DST},dp)).object;const r=this._device.createShaderModule({label:"Monaco rectangle renderer shader module",code:Vve});this._pipeline=this._device.createRenderPipeline({label:"Monaco rectangle renderer render pipeline",layout:"auto",vertex:{module:r,buffers:[{arrayStride:2*Float32Array.BYTES_PER_ELEMENT,attributes:[{shaderLocation:0,offset:0,format:"float32x2"}]}]},fragment:{module:r,targets:[{format:t,blend:{color:{srcFactor:"src-alpha",dstFactor:"one-minus-src-alpha"},alpha:{srcFactor:"src-alpha",dstFactor:"one-minus-src-alpha"}}}]}}),this._updateBindGroup(this._pipeline,i),this._initialized=!0}_updateBindGroup(e,t){this._bindGroup=this._device.createBindGroup({label:"Monaco rectangle renderer bind group",layout:e.getBindGroupLayout(0),entries:[{binding:0,resource:{buffer:this._shapeBindBuffer.value.object}},{binding:1,resource:{buffer:t}},{binding:2,resource:{buffer:this._scrollOffsetBindBuffer}}]})}register(e,t,i,n,o,r,a,l){return this._shapeCollection.createEntry({x:e,y:t,width:i,height:n,red:o,green:r,blue:a,alpha:l})}onScrollChanged(e){if(this._device){const t=kt().devicePixelRatio;this._scrollOffsetValueBuffer[0]=this._context.viewLayout.getCurrentScrollLeft()*t,this._scrollOffsetValueBuffer[1]=this._context.viewLayout.getCurrentScrollTop()*t,this._device.queue.writeBuffer(this._scrollOffsetBindBuffer,0,this._scrollOffsetValueBuffer)}return!0}_update(){if(!this._device)return;const e=this._shapeCollection;e.dirtyTracker.isDirty&&(this._device.queue.writeBuffer(this._shapeBindBuffer.value.object,0,e.buffer,e.dirtyTracker.dataOffset,e.dirtyTracker.dirtySize*e.view.BYTES_PER_ELEMENT),e.dirtyTracker.clear())}draw(e){if(!this._initialized)return;this._update();const t=this._device.createCommandEncoder({label:"Monaco rectangle renderer command encoder"});this._renderPassColorAttachment.view=this._ctx.getCurrentTexture().createView();const i=t.beginRenderPass(this._renderPassDescriptor);i.setPipeline(this._pipeline),i.setVertexBuffer(0,this._vertexBuffer),i.setBindGroup(0,this._bindGroup);const n=Math.ceil(this._contentLeft.get()*this._devicePixelRatio.get());i.setScissorRect(n,0,this._canvas.width-n,this._canvas.height),i.draw(dp.length/2,this._shapeCollection.entryCount),i.end();const o=t.finish();this._device.queue.submit([o])}}class $ve extends F{constructor(){super(),this._ruleCache=new Map,this._container=ie("div.monaco-decoration-css-rule-extractor"),this._dummyElement=ie("span"),this._container.appendChild(this._dummyElement),this._register(pe(()=>this._container.remove()))}getStyleRules(e,t){const i=this._ruleCache.get(t);if(i)return i;this._dummyElement.className=t,e.appendChild(this._container);const n=this._getStyleRules(t);return this._ruleCache.set(t,n),e.removeChild(this._container),n}_getStyleRules(e){const t=[],n=[...K0().styleSheets];for(let o=0;o<n.length;o++){const r=n[o];for(const a of r.cssRules)if(a instanceof CSSImportRule)a.styleSheet&&n.push(a.styleSheet);else if(a instanceof CSSStyleRule){const l=`.${e}`,c=a.selectorText.indexOf(l);if(c!==-1){const d=c+l.length;(a.selectorText.length===d||a.selectorText.substring(d,d+1).match(/[ :]/))&&t.push(a)}}}return t}}class Uve{constructor(){this._nextId=1,this._cacheById=new Map,this._cacheByStyle=new VR}getOrCreateEntry(e,t,i){if(e===void 0&&t===void 0&&i===void 0)return 0;const n=this._cacheByStyle.get(e??0,t?1:0,i===void 0?"":i.toFixed(2));if(n)return n.id;const o=this._nextId++,r={id:o,color:e,bold:t,opacity:i};return this._cacheById.set(o,r),this._cacheByStyle.set(r,e??0,t?1:0,i===void 0?"":i.toFixed(2)),o}getStyleSet(e){if(e!==0)return this._cacheById.get(e)}}function yl(s,e){let t,i=!0;return Hhe({changeTracker:{createChangeSummary:()=>({deltas:[],didChange:!1}),handleChange:(n,o)=>{if(n.didChange(s)){const r=n.change;r!==void 0&&o.deltas.push(r),o.didChange=!0}return!0}}},(n,o)=>{const r=s.read(n),a=t;o.didChange&&(t=r,e(r,a,o.deltas)),i&&(i=!1,t=r)})}function Ree(s,e){const t=new U,i=yl(s,(n,o,r)=>{t.clear(),e(n,o,r,t)});return{dispose(){i.dispose(),t.dispose()}}}function qve(s,e){return Ree(s,(t,i,n,o)=>{e(t,i,n,lP(o))})}var jve=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},aM=function(s,e){return function(t,i){e(t,i,s)}},zs,Vu;let Ds=(Vu=class extends F{static get decorationCssRuleExtractor(){return zs._decorationCssRuleExtractor}static get decorationStyleCache(){return zs._decorationStyleCache}static get atlas(){if(!zs._atlas)throw new Ce("Cannot call ViewGpuContext.textureAtlas before device is resolved");return zs._atlas}get atlas(){return zs.atlas}constructor(e,t,i,n){super(),this._instantiationService=t,this._notificationService=i,this.configurationService=n,this.maxGpuCols=2e3,this.canvas=wt(document.createElement("canvas")),this.canvas.setClassName("editorCanvas"),this._register(oe.runAndSubscribe(n.onDidChangeConfiguration,l=>{if(!l||l.affectsConfiguration("editor.scrollbar.verticalScrollbarSize")){const c=n.getValue("editor").scrollbar?.verticalScrollbarSize??14;this.canvas.domNode.style.boxSizing="border-box",this.canvas.domNode.style.paddingRight=`${c}px`}})),this.ctx=Tm(this.canvas.domNode.getContext("webgpu")),zs.device||(zs.device=fo.requestDevice(l=>{const c=[{label:m(78,"Use DOM-based rendering"),run:()=>this.configurationService.updateValue("editor.experimentalGpuAcceleration","off")}];this._notificationService.prompt(oC.Warning,l,c)}).then(l=>(zs.deviceSync=l.object,zs._atlas||(zs._atlas=this._instantiationService.createInstance(wu,l.object.limits.maxTextureDimension2D,void 0,zs.decorationStyleCache)),l.object)));const o=Re(this,kt().devicePixelRatio);this._register(W(kt(),"resize",()=>{o.set(kt().devicePixelRatio,void 0)})),this.devicePixelRatio=o,this._register(yl(this.devicePixelRatio,()=>zs.atlas?.clear()));const r=Re(this,{width:this.canvas.domNode.width,height:this.canvas.domNode.height});this._register(Nee(this.canvas.domNode,kt(),(l,c)=>{this.canvas.domNode.width=l,this.canvas.domNode.height=c,r.set({width:l,height:c},void 0)})),this.canvasDevicePixelDimensions=r;const a=Re(this,0);this._register(this.configurationService.onDidChangeConfiguration(l=>{a.set(e.configuration.options.get(165).contentLeft,void 0)})),this.contentLeft=a,this.rectangleRenderer=this._instantiationService.createInstance(zve,e,this.contentLeft,this.devicePixelRatio,this.canvas.domNode,this.ctx,zs.device)}canRender(e,t,i){const n=t.getViewLineRenderingData(i);if(n.containsRTL||n.maxColumn>this.maxGpuCols)return!1;if(n.inlineDecorations.length>0){let o=!0;for(const r of n.inlineDecorations){if(r.type!==0){o=!1;break}const a=zs._decorationCssRuleExtractor.getStyleRules(this.canvas.domNode,r.inlineClassName);if(o&&=a.every(l=>{if(l.selectorText.includes(":"))return!1;for(const c of l.style)if(!oU(c,l.style))return!1;return!0}),!o)break}return o}return!0}canRenderDetailed(e,t,i){const n=t.getViewLineRenderingData(i),o=[];if(n.containsRTL&&o.push("containsRTL"),n.maxColumn>this.maxGpuCols&&o.push("maxColumn > maxGpuCols"),n.inlineDecorations.length>0){let r=!0;const a=[],l=[],c=[];for(const d of n.inlineDecorations){if(d.type!==0){a.push(d.type),r=!1;continue}const h=zs._decorationCssRuleExtractor.getStyleRules(this.canvas.domNode,d.inlineClassName);r&&=h.every(u=>{if(u.selectorText.includes(":"))return l.push(u.selectorText),!1;for(const g of u.style)if(!oU(g,u.style))return c.push(`${g}: ${u.style[g]}`),!1;return!0})}a.length>0&&o.push(`inlineDecorations with unsupported types (${a.map(d=>`\`${d}\``).join(", ")})`),c.length>0&&o.push(`inlineDecorations with unsupported CSS rules (${c.map(d=>`\`${d}\``).join(", ")})`),l.length>0&&o.push(`inlineDecorations with unsupported CSS selectors (${l.map(d=>`\`${d}\``).join(", ")})`)}return o}},zs=Vu,Vu._decorationCssRuleExtractor=new $ve,Vu._decorationStyleCache=new Uve,Vu);Ds=zs=jve([aM(1,_e),aM(2,Ii),aM(3,He)],Ds);const Gve=["color","font-weight","opacity"];function oU(s,e){return!!Gve.includes(s)}function aI(s,e){return s.isBasicASCII&&e.useMonospaceOptimizations?new Kve(s):new Yve(s)}class Kve{constructor(e){this._content=e.content}getSegmentAtIndex(e){return this._content[e]}}class Yve{constructor(e){this._segments=[];const t=e.content,i=a_.Segmenter(void 0,{granularity:"grapheme"}).value,n=Array.from(i.segment(t));let o=0;for(let r=0;r<t.length;r++){const a=n[o];if(!a)break;if(a.index!==r){this._segments.push(void 0);continue}o++,this._segments.push(a)}}getSegmentAtIndex(e){return this._segments[e]?.segment}}class Aee extends c_{get glyphRasterizer(){return this._glyphRasterizer.value}constructor(e,t,i,n){super(),this._context=e,this._viewGpuContext=t,this._device=i,this._glyphRasterizer=n,this._context.addEventHandler(this)}}const Mee=`
struct GlyphInfo {
	position: vec2f,
	size: vec2f,
	origin: vec2f,
};

struct Vertex {
	@location(0) position: vec2f,
};

struct Cell {
	position: vec2f,
	unused1: vec2f,
	glyphIndex: f32,
	textureIndex: f32
};

struct LayoutInfo {
	canvasDims: vec2f,
	viewportOffset: vec2f,
	viewportDims: vec2f,
}

struct ScrollOffset {
	offset: vec2f
}

struct VSOutput {
	@builtin(position) position:   vec4f,
	@location(1)       layerIndex: f32,
	@location(0)       texcoord:   vec2f,
};

// Uniforms
@group(0) @binding(4)       var<uniform>       layoutInfo:      LayoutInfo;
@group(0) @binding(5)  var<uniform>       atlasDims:       vec2f;
@group(0) @binding(6)            var<uniform>       scrollOffset:    ScrollOffset;

// Storage buffers
@group(0) @binding(0)               var<storage, read> glyphInfo:       array<array<GlyphInfo, ${al.maximumGlyphCount}>, ${wu.maximumPageCount}>;
@group(0) @binding(1)                   var<storage, read> cells:           array<Cell>;

@vertex fn vs(
	vert: Vertex,
	@builtin(instance_index) instanceIndex: u32,
	@builtin(vertex_index) vertexIndex : u32
) -> VSOutput {
	let cell = cells[instanceIndex];
	var glyph = glyphInfo[u32(cell.textureIndex)][u32(cell.glyphIndex)];

	var vsOut: VSOutput;
	// Multiple vert.position by 2,-2 to get it into clipspace which ranged from -1 to 1
	vsOut.position = vec4f(
		// Make everything relative to top left instead of center
		vec2f(-1, 1) +
		((vert.position * vec2f(2, -2)) / layoutInfo.canvasDims) * glyph.size +
		((cell.position * vec2f(2, -2)) / layoutInfo.canvasDims) +
		((glyph.origin * vec2f(2, -2)) / layoutInfo.canvasDims) +
		(((layoutInfo.viewportOffset - scrollOffset.offset * vec2(1, -1)) * 2) / layoutInfo.canvasDims),
		0.0,
		1.0
	);

	vsOut.layerIndex = cell.textureIndex;
	// Textures are flipped from natural direction on the y-axis, so flip it back
	vsOut.texcoord = vert.position;
	vsOut.texcoord = (
		// Glyph offset (0-1)
		(glyph.position / atlasDims) +
		// Glyph coordinate (0-1)
		(vsOut.texcoord * (glyph.size / atlasDims))
	);

	return vsOut;
}

@group(0) @binding(2) var ourSampler: sampler;
@group(0) @binding(3)        var ourTexture: texture_2d_array<f32>;

@fragment fn fs(vsOut: VSOutput) -> @location(0) vec4f {
	return textureSample(ourTexture, ourSampler, vsOut.texcoord, u32(vsOut.layerIndex));
}
`,Ql=class Ql extends Aee{get bindGroupEntries(){return[{binding:1,resource:{buffer:this._cellBindBuffer}},{binding:6,resource:{buffer:this._scrollOffsetBindBuffer}}]}constructor(e,t,i,n){super(e,t,i,n),this.type="viewport",this.wgsl=Mee,this._cellBindBufferLineCapacity=63,this._activeDoubleBufferIndex=0,this._visibleObjectCount=0,this._scrollInitialized=!1,this._onDidChangeBindGroupEntries=this._register(new M),this.onDidChangeBindGroupEntries=this._onDidChangeBindGroupEntries.event,this._rebuildCellBuffer(this._cellBindBufferLineCapacity);const o=2;this._scrollOffsetBindBuffer=this._register(fo.createBuffer(this._device,{label:"Monaco scroll offset buffer",size:o*Float32Array.BYTES_PER_ELEMENT,usage:GPUBufferUsage.UNIFORM|GPUBufferUsage.COPY_DST})).object,this._scrollOffsetValueBuffer=new Float32Array(o)}_rebuildCellBuffer(e){this._cellBindBuffer?.destroy();const t=(Math.floor(e/32)+1)*32,i=t*Ql.maxSupportedColumns*6*Float32Array.BYTES_PER_ELEMENT;this._cellBindBuffer=this._register(fo.createBuffer(this._device,{label:"Monaco full file cell buffer",size:i,usage:GPUBufferUsage.STORAGE|GPUBufferUsage.COPY_DST})).object,this._cellValueBuffers=[new ArrayBuffer(i),new ArrayBuffer(i)],this._cellBindBufferLineCapacity=t,this._onDidChangeBindGroupEntries.fire()}onConfigurationChanged(e){return!0}onDecorationsChanged(e){return!0}onTokensChanged(e){return!0}onLinesDeleted(e){return!0}onLinesInserted(e){return!0}onLinesChanged(e){return!0}onScrollChanged(e){const t=kt().devicePixelRatio;return this._scrollOffsetValueBuffer[0]=(e?.scrollLeft??this._context.viewLayout.getCurrentScrollLeft())*t,this._scrollOffsetValueBuffer[1]=(e?.scrollTop??this._context.viewLayout.getCurrentScrollTop())*t,this._device.queue.writeBuffer(this._scrollOffsetBindBuffer,0,this._scrollOffsetValueBuffer),!0}onThemeChanged(e){return!0}onLineMappingChanged(e){return!0}onZonesChanged(e){return!0}reset(){for(const e of[0,1]){const t=new Float32Array(this._cellValueBuffers[e]);t.fill(0,0,t.length),this._device.queue.writeBuffer(this._cellBindBuffer,0,t.buffer,0,t.byteLength)}}update(e,t){let i="",n,o=0,r=0,a=0,l=0,c=0,d=0,h,u=0,g=0,f=0,p=0,_,b,w,C,v,y=0,x=0,L;const E=kt().devicePixelRatio;let I;this._scrollInitialized||(this.onScrollChanged(),this._scrollInitialized=!0),this._cellBindBufferLineCapacity<e.endLineNumber-e.startLineNumber+1&&this._rebuildCellBuffer(e.endLineNumber-e.startLineNumber+1);const R=new Float32Array(this._cellValueBuffers[this._activeDoubleBufferIndex]);R.fill(0);const O=Ql.maxSupportedColumns*6;for(r=e.startLineNumber;r<=e.endLineNumber;r++)if(this._viewGpuContext.canRender(t,e,r)){C=e.getViewLineRenderingData(r),d=0,I=aI(C,t),o=t.spaceWidth*E,l=0,L=C.tokens,g=C.minColumn-1,f=0;for(let B=0,$=L.getCount();B<$;B++)if(f=L.getEndOffset(B),!(f<=g)){for(p=L.getMetadata(B),a=g;a<f&&!(a>Ql.maxSupportedColumns);a++){if(n=I.getSegmentAtIndex(a),n===void 0)continue;i=n,C.isBasicASCII&&t.useMonospaceOptimizations||(o=this.glyphRasterizer.getTextMetrics(i).width),b=void 0,_=void 0,w=void 0;for(v of C.inlineDecorations){if(r<v.range.startLineNumber||r>v.range.endLineNumber||r===v.range.startLineNumber&&a<v.range.startColumn-1||r===v.range.endLineNumber&&a>=v.range.endColumn-1)continue;const Q=Ds.decorationCssRuleExtractor.getStyleRules(this._viewGpuContext.canvas.domNode,v.inlineClassName);for(const J of Q)for(const ne of J.style){const de=J.styleMap.get(ne)?.toString()??"";switch(ne){case"color":{const ke=q.Format.CSS.parse(de);if(!ke)throw new Ce("Invalid color format "+de);b=ke.toNumber32Bit();break}case"font-weight":{Zve(de)>=400?_=!0:_=!1;break}case"opacity":{w=Xve(de);break}default:throw new Ce("Unexpected inline decoration style")}}}if(i===" "||i==="	"){if(u=((r-1)*Ql.maxSupportedColumns+a)*6,R.fill(0,u,u+6),i==="	"){const Q=a+d;d=pi.nextRenderTabStop(a+d,C.tabSize),l+=o*(d-Q),d-=a+1}else l+=o;continue}const z=Ds.decorationStyleCache.getOrCreateEntry(b,_,w);h=this._viewGpuContext.atlas.getGlyph(this.glyphRasterizer,i,p,z,l),c=Math.round(e.relativeVerticalOffset[r-e.startLineNumber]*E+Math.floor((e.lineHeight*E-(h.fontBoundingBoxAscent+h.fontBoundingBoxDescent))/2)+h.fontBoundingBoxAscent),u=((r-e.startLineNumber)*Ql.maxSupportedColumns+a)*6,R[u+0]=Math.floor(l),R[u+1]=c,R[u+4]=h.glyphIndex,R[u+5]=h.pageIndex,l+=o}g=f}y=((r-e.startLineNumber)*Ql.maxSupportedColumns+f)*6,x=(r-e.startLineNumber)*Ql.maxSupportedColumns*6,R.fill(0,y,x)}const P=(e.endLineNumber-e.startLineNumber+1)*O;return this._device.queue.writeBuffer(this._cellBindBuffer,0,R.buffer,0,(e.endLineNumber-e.startLineNumber)*O*Float32Array.BYTES_PER_ELEMENT),this._activeDoubleBufferIndex=this._activeDoubleBufferIndex?0:1,this._visibleObjectCount=P,P}draw(e,t){if(this._visibleObjectCount<=0)throw new Ce("Attempt to draw 0 objects");e.draw(dp.length/2,this._visibleObjectCount)}};Ql.maxSupportedColumns=2e3;let pF=Ql;function Zve(s){switch(s){case"lighter":case"normal":return 400;case"bolder":case"bold":return 700}return parseInt(s)}function Xve(s){return s.endsWith("%")?parseFloat(s.substring(0,s.length-1))/100:s.match(/^\d+(?:\.\d*)/)?parseFloat(s):1}const Yn=class Yn extends Aee{get bindGroupEntries(){return[{binding:1,resource:{buffer:this._cellBindBuffer}},{binding:6,resource:{buffer:this._scrollOffsetBindBuffer}}]}constructor(e,t,i,n){super(e,t,i,n),this.type="fullfile",this.wgsl=Mee,this._activeDoubleBufferIndex=0,this._upToDateLines=[new Set,new Set],this._visibleObjectCount=0,this._finalRenderedLine=0,this._scrollInitialized=!1,this._queuedBufferUpdates=[[],[]];const o=Yn.maxSupportedLines*Yn.maxSupportedColumns*6*Float32Array.BYTES_PER_ELEMENT;this._cellBindBuffer=this._register(fo.createBuffer(this._device,{label:"Monaco full file cell buffer",size:o,usage:GPUBufferUsage.STORAGE|GPUBufferUsage.COPY_DST})).object,this._cellValueBuffers=[new ArrayBuffer(o),new ArrayBuffer(o)];const r=2;this._scrollOffsetBindBuffer=this._register(fo.createBuffer(this._device,{label:"Monaco scroll offset buffer",size:r*Float32Array.BYTES_PER_ELEMENT,usage:GPUBufferUsage.UNIFORM|GPUBufferUsage.COPY_DST})).object,this._scrollOffsetValueBuffer=new Float32Array(r)}onConfigurationChanged(e){return this._invalidateAllLines(),this._queueBufferUpdate(e),!0}onDecorationsChanged(e){return this._invalidateAllLines(),!0}onTokensChanged(e){for(const t of e.ranges)this._invalidateLineRange(t.fromLineNumber,t.toLineNumber);return!0}onLinesDeleted(e){return this._invalidateLinesFrom(e.fromLineNumber),this._queueBufferUpdate(e),!0}onLinesInserted(e){return this._invalidateLinesFrom(e.fromLineNumber),!0}onLinesChanged(e){return this._invalidateLineRange(e.fromLineNumber,e.fromLineNumber+e.count),!0}onScrollChanged(e){const t=kt().devicePixelRatio;return this._scrollOffsetValueBuffer[0]=(e?.scrollLeft??this._context.viewLayout.getCurrentScrollLeft())*t,this._scrollOffsetValueBuffer[1]=(e?.scrollTop??this._context.viewLayout.getCurrentScrollTop())*t,this._device.queue.writeBuffer(this._scrollOffsetBindBuffer,0,this._scrollOffsetValueBuffer),!0}onThemeChanged(e){return this._invalidateAllLines(),!0}onLineMappingChanged(e){return this._invalidateAllLines(),this._queueBufferUpdate(e),!0}onZonesChanged(e){return this._invalidateAllLines(),this._queueBufferUpdate(e),!0}_invalidateAllLines(){this._upToDateLines[0].clear(),this._upToDateLines[1].clear()}_invalidateLinesFrom(e){for(const t of[0,1]){const i=this._upToDateLines[t];for(const n of i)n>=e&&i.delete(n)}}_invalidateLineRange(e,t){for(let i=e;i<=t;i++)this._upToDateLines[0].delete(i),this._upToDateLines[1].delete(i)}reset(){this._invalidateAllLines();for(const e of[0,1]){const t=new Float32Array(this._cellValueBuffers[e]);t.fill(0,0,t.length),this._device.queue.writeBuffer(this._cellBindBuffer,0,t.buffer,0,t.byteLength)}this._finalRenderedLine=0}update(e,t){let i="",n,o=0,r=0,a=0,l=0,c=0,d=0,h,u=0,g=0,f=0,p=0,_,b,w,C,v,y=0,x=0,L;const E=kt().devicePixelRatio;let I;this._scrollInitialized||(this.onScrollChanged(),this._scrollInitialized=!0);const R=new Float32Array(this._cellValueBuffers[this._activeDoubleBufferIndex]),O=Yn.maxSupportedColumns*6,P=this._upToDateLines[this._activeDoubleBufferIndex];let B=3e3,$=0;const z=this._queuedBufferUpdates[this._activeDoubleBufferIndex];for(;z.length;){const J=z.shift();switch(J.type){case 2:case 8:case 17:{R.fill(0),B=1,$=Math.max($,this._finalRenderedLine),this._finalRenderedLine=0;break}case 10:{const ne=(J.fromLineNumber-1)*Yn.maxSupportedColumns*6,de=J.toLineNumber*Yn.maxSupportedColumns*6,ke=(this._finalRenderedLine-(J.toLineNumber-J.fromLineNumber+1))*Yn.maxSupportedColumns*6;R.set(R.subarray(de),ne),R.fill(0,ke),B=Math.min(B,J.fromLineNumber),$=Math.max($,this._finalRenderedLine),this._finalRenderedLine-=J.toLineNumber-J.fromLineNumber+1;break}}}for(r=e.startLineNumber;r<=e.endLineNumber;r++){if(!this._viewGpuContext.canRender(t,e,r)){y=(r-1)*Yn.maxSupportedColumns*6,x=r*Yn.maxSupportedColumns*6,R.fill(0,y,x),B=Math.min(B,r),$=Math.max($,r);continue}if(!P.has(r)){B=Math.min(B,r),$=Math.max($,r),C=e.getViewLineRenderingData(r),d=0,I=aI(C,t),o=t.spaceWidth*E,l=0,L=C.tokens,g=C.minColumn-1,f=0;for(let J=0,ne=L.getCount();J<ne;J++)if(f=L.getEndOffset(J),!(f<=g)){for(p=L.getMetadata(J),a=g;a<f&&!(a>Yn.maxSupportedColumns);a++){if(n=I.getSegmentAtIndex(a),n===void 0)continue;i=n,C.isBasicASCII&&t.useMonospaceOptimizations||(o=this.glyphRasterizer.getTextMetrics(i).width),b=void 0,_=void 0,w=void 0;for(v of C.inlineDecorations){if(r<v.range.startLineNumber||r>v.range.endLineNumber||r===v.range.startLineNumber&&a<v.range.startColumn-1||r===v.range.endLineNumber&&a>=v.range.endColumn-1)continue;const ke=Ds.decorationCssRuleExtractor.getStyleRules(this._viewGpuContext.canvas.domNode,v.inlineClassName);for(const Me of ke)for(const Ze of Me.style){const Xe=Me.styleMap.get(Ze)?.toString()??"";switch(Ze){case"color":{const At=q.Format.CSS.parse(Xe);if(!At)throw new Ce("Invalid color format "+Xe);b=At.toNumber32Bit();break}case"font-weight":{Qve(Xe)>=400?_=!0:_=!1;break}case"opacity":{w=Jve(Xe);break}default:throw new Ce("Unexpected inline decoration style")}}}if(i===" "||i==="	"){if(u=((r-1)*Yn.maxSupportedColumns+a)*6,R.fill(0,u,u+6),i==="	"){const ke=a+d;d=pi.nextRenderTabStop(a+d,C.tabSize),l+=o*(d-ke),d-=a+1}else l+=o;continue}const de=Ds.decorationStyleCache.getOrCreateEntry(b,_,w);h=this._viewGpuContext.atlas.getGlyph(this.glyphRasterizer,i,p,de,l),c=Math.round(e.relativeVerticalOffset[r-e.startLineNumber]*E+Math.floor((e.lineHeight*E-(h.fontBoundingBoxAscent+h.fontBoundingBoxDescent))/2)+h.fontBoundingBoxAscent),u=((r-1)*Yn.maxSupportedColumns+a)*6,R[u+0]=Math.floor(l),R[u+1]=c,R[u+4]=h.glyphIndex,R[u+5]=h.pageIndex,l+=o}g=f}y=((r-1)*Yn.maxSupportedColumns+f)*6,x=r*Yn.maxSupportedColumns*6,R.fill(0,y,x),P.add(r)}}const Q=(e.endLineNumber-e.startLineNumber+1)*O;return B=Math.min(B,Yn.maxSupportedLines),$=Math.min($,Yn.maxSupportedLines),B<=$&&this._device.queue.writeBuffer(this._cellBindBuffer,(B-1)*O*Float32Array.BYTES_PER_ELEMENT,R.buffer,(B-1)*O*Float32Array.BYTES_PER_ELEMENT,($-B+1)*O*Float32Array.BYTES_PER_ELEMENT),this._finalRenderedLine=Math.max(this._finalRenderedLine,$),this._activeDoubleBufferIndex=this._activeDoubleBufferIndex?0:1,this._visibleObjectCount=Q,Q}draw(e,t){if(this._visibleObjectCount<=0)throw new Ce("Attempt to draw 0 objects");e.draw(dp.length/2,this._visibleObjectCount,void 0,(t.startLineNumber-1)*Yn.maxSupportedColumns)}_queueBufferUpdate(e){this._queuedBufferUpdates[0].push(e),this._queuedBufferUpdates[1].push(e)}};Yn.maxSupportedLines=3e3,Yn.maxSupportedColumns=200;let mf=Yn;function Qve(s){switch(s){case"lighter":case"normal":return 400;case"bolder":case"bold":return 700}return parseInt(s)}function Jve(s){return s.endsWith("%")?parseFloat(s.substring(0,s.length-1))/100:s.match(/^\d+(?:\.\d*)/)?parseFloat(s):1}var eCe=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},rU=function(s,e){return function(t,i){e(t,i,s)}};let mF=class extends no{constructor(e,t,i,n){super(e),this._viewGpuContext=t,this._instantiationService=i,this._logService=n,this._atlasGpuTextureVersions=[],this._initialized=!1,this._glyphRasterizer=this._register(new vt),this._renderStrategy=this._register(new vt),this.canvas=this._viewGpuContext.canvas.domNode,this._register(De(o=>{this._viewGpuContext.canvasDevicePixelDimensions.read(o);const r=this._lastViewportData;r&&setTimeout(()=>{r===this._lastViewportData&&this.renderText(r)})})),this.initWebgpu()}async initWebgpu(){if(this._device=Ds.deviceSync||await Ds.device,this._store.isDisposed)return;const e=Ds.atlas;this._register(e.onDidDeleteGlyphs(()=>{this._atlasGpuTextureVersions.length=0,this._atlasGpuTextureVersions[0]=0,this._atlasGpuTextureVersions[1]=0,this._renderStrategy.value.reset()}));const t=navigator.gpu.getPreferredCanvasFormat();this._viewGpuContext.ctx.configure({device:this._device,format:t,alphaMode:"premultiplied"}),this._renderPassColorAttachment={view:null,loadOp:"load",storeOp:"store"},this._renderPassDescriptor={label:"Monaco render pass",colorAttachments:[this._renderPassColorAttachment]};let i;{const l=new Float32Array(6),c=(d=this.canvas.width,h=this.canvas.height)=>(l[0]=d,l[1]=h,l[2]=Math.ceil(this._context.configuration.options.get(165).contentLeft*kt().devicePixelRatio),l[3]=0,l[4]=l[0]-l[2],l[5]=l[1]-l[3],l);i=this._register(fo.createBuffer(this._device,{label:"Monaco uniform buffer",size:24,usage:GPUBufferUsage.UNIFORM|GPUBufferUsage.COPY_DST},()=>c())).object,this._register(yl(this._viewGpuContext.canvasDevicePixelDimensions,({width:d,height:h})=>{this._device.queue.writeBuffer(i,0,c(d,h))})),this._register(yl(this._viewGpuContext.contentLeft,()=>{this._device.queue.writeBuffer(i,0,c())}))}let n;n=this._register(fo.createBuffer(this._device,{label:"Monaco atlas info uniform buffer",size:8,usage:GPUBufferUsage.UNIFORM|GPUBufferUsage.COPY_DST},()=>{const l=new Float32Array(2);return l[0]=e.pageSize,l[1]=e.pageSize,l})).object;const o=this._context.configuration.options.get(58),r=this._context.configuration.options.get(61);this._glyphRasterizer.value=this._register(new CS(r,o,this._viewGpuContext.devicePixelRatio.get(),Ds.decorationStyleCache)),this._register(yl(this._viewGpuContext.devicePixelRatio,()=>{this._refreshGlyphRasterizer()})),this._renderStrategy.value=this._instantiationService.createInstance(mf,this._context,this._viewGpuContext,this._device,this._glyphRasterizer),this._glyphStorageBuffer=this._register(fo.createBuffer(this._device,{label:"Monaco glyph storage buffer",size:wu.maximumPageCount*(al.maximumGlyphCount*24),usage:GPUBufferUsage.STORAGE|GPUBufferUsage.COPY_DST})).object,this._atlasGpuTextureVersions[0]=0,this._atlasGpuTextureVersions[1]=0,this._atlasGpuTexture=this._register(fo.createTexture(this._device,{label:"Monaco atlas texture",format:"rgba8unorm",size:{width:e.pageSize,height:e.pageSize,depthOrArrayLayers:wu.maximumPageCount},dimension:"2d",usage:GPUTextureUsage.TEXTURE_BINDING|GPUTextureUsage.COPY_DST|GPUTextureUsage.RENDER_ATTACHMENT})).object,this._updateAtlasStorageBufferAndTexture(),this._vertexBuffer=this._register(fo.createBuffer(this._device,{label:"Monaco vertex buffer",size:dp.byteLength,usage:GPUBufferUsage.VERTEX|GPUBufferUsage.COPY_DST},dp)).object;const a=this._device.createShaderModule({label:"Monaco shader module",code:this._renderStrategy.value.wgsl});if(this._pipeline=this._device.createRenderPipeline({label:"Monaco render pipeline",layout:"auto",vertex:{module:a,buffers:[{arrayStride:2*Float32Array.BYTES_PER_ELEMENT,attributes:[{shaderLocation:0,offset:0,format:"float32x2"}]}]},fragment:{module:a,targets:[{format:t,blend:{color:{srcFactor:"src-alpha",dstFactor:"one-minus-src-alpha"},alpha:{srcFactor:"src-alpha",dstFactor:"one-minus-src-alpha"}}}]}}),this._rebuildBindGroup=()=>{this._bindGroup=this._device.createBindGroup({label:"Monaco bind group",layout:this._pipeline.getBindGroupLayout(0),entries:[{binding:0,resource:{buffer:this._glyphStorageBuffer}},{binding:2,resource:this._device.createSampler({label:"Monaco atlas sampler",magFilter:"nearest",minFilter:"nearest"})},{binding:3,resource:this._atlasGpuTexture.createView()},{binding:4,resource:{buffer:i}},{binding:5,resource:{buffer:n}},...this._renderStrategy.value.bindGroupEntries]})},this._rebuildBindGroup(),this._initialized=!0,this._initViewportData){for(const l of this._initViewportData)this.renderText(l);this._initViewportData=void 0}}_refreshRenderStrategy(e){if(this._renderStrategy.value?.type==="viewport"||e.endLineNumber<mf.maxSupportedLines&&this._viewportMaxColumn(e)<mf.maxSupportedColumns)return;this._logService.trace(`File is larger than ${mf.maxSupportedLines} lines or ${mf.maxSupportedColumns} columns, switching to viewport render strategy`);const t=this._instantiationService.createInstance(pF,this._context,this._viewGpuContext,this._device,this._glyphRasterizer);this._renderStrategy.value=t,this._register(t.onDidChangeBindGroupEntries(()=>this._rebuildBindGroup?.())),this._rebuildBindGroup?.()}_viewportMaxColumn(e){let t=0,i;for(let n=e.startLineNumber;n<=e.endLineNumber;n++)i=e.getViewLineRenderingData(n),t=Math.max(t,i.maxColumn);return t}_updateAtlasStorageBufferAndTexture(){for(const[e,t]of Ds.atlas.pages.entries()){if(e>=wu.maximumPageCount){console.log(`Attempt to upload atlas page [${e}], only ${wu.maximumPageCount} are supported currently`);continue}if(t.version===this._atlasGpuTextureVersions[e])continue;this._logService.trace("Updating atlas page[",e,"] from version ",this._atlasGpuTextureVersions[e]," to version ",t.version);const i=6*al.maximumGlyphCount,n=new Float32Array(i);let o=0;for(const r of t.glyphs)n[o+0]=r.x,n[o+0+1]=r.y,n[o+2]=r.w,n[o+2+1]=r.h,n[o+4]=r.originOffsetX,n[o+4+1]=r.originOffsetY,o+=6;if(o/6>al.maximumGlyphCount)throw new Error(`Attempting to write more glyphs (${o/6}) than the GPUBuffer can hold (${al.maximumGlyphCount})`);this._device.queue.writeBuffer(this._glyphStorageBuffer,e*6*al.maximumGlyphCount*Float32Array.BYTES_PER_ELEMENT,n,0,6*al.maximumGlyphCount),t.usedArea.right-t.usedArea.left>0&&t.usedArea.bottom-t.usedArea.top>0&&this._device.queue.copyExternalImageToTexture({source:t.source},{texture:this._atlasGpuTexture,origin:{x:t.usedArea.left,y:t.usedArea.top,z:e}},{width:t.usedArea.right-t.usedArea.left+1,height:t.usedArea.bottom-t.usedArea.top+1}),this._atlasGpuTextureVersions[e]=t.version}}prepareRender(e){throw new Ce("Should not be called")}render(e){throw new Ce("Should not be called")}onConfigurationChanged(e){return this._refreshGlyphRasterizer(),!0}onCursorStateChanged(e){return!0}onDecorationsChanged(e){return!0}onFlushed(e){return!0}onLinesChanged(e){return!0}onLinesDeleted(e){return!0}onLinesInserted(e){return!0}onLineMappingChanged(e){return!0}onRevealRangeRequest(e){return!0}onScrollChanged(e){return!0}onThemeChanged(e){return!0}onZonesChanged(e){return!0}_refreshGlyphRasterizer(){const e=this._glyphRasterizer.value;if(!e)return;const t=this._context.configuration.options.get(58),i=this._context.configuration.options.get(61),n=this._viewGpuContext.devicePixelRatio.get();(e.fontFamily!==t||e.fontSize!==i||e.devicePixelRatio!==n)&&(this._glyphRasterizer.value=new CS(i,t,n,Ds.decorationStyleCache))}renderText(e){if(this._initialized)return this._refreshRenderStrategy(e),this._renderText(e);this._initViewportData=this._initViewportData??[],this._initViewportData.push(e)}_renderText(e){this._viewGpuContext.rectangleRenderer.draw(e);const t=new eI(this._context.configuration,this._context.theme.type);this._renderStrategy.value.update(e,t),this._updateAtlasStorageBufferAndTexture();const i=this._device.createCommandEncoder({label:"Monaco command encoder"});this._renderPassColorAttachment.view=this._viewGpuContext.ctx.getCurrentTexture().createView({label:"Monaco canvas texture view"});const n=i.beginRenderPass(this._renderPassDescriptor);n.setPipeline(this._pipeline),n.setVertexBuffer(0,this._vertexBuffer);const o=Math.ceil(this._viewGpuContext.contentLeft.get()*this._viewGpuContext.devicePixelRatio.get());n.setScissorRect(o,0,this.canvas.width-o,this.canvas.height),n.setBindGroup(0,this._bindGroup),this._renderStrategy.value.draw(n,e),n.end();const r=i.finish();this._device.queue.submit([r]),this._lastViewportData=e,this._lastViewLineOptions=t}linesVisibleRangesForRange(e,t){if(!this._lastViewportData)return null;const i=e.endLineNumber,n=k.intersectRanges(e,this._lastViewportData.visibleRange);if(!n)return null;const o=this._lastViewportData.startLineNumber,r=this._lastViewportData.endLineNumber,a=this._lastViewportData,l=this._lastViewLineOptions;if(!a||!l)return null;const c=[];let d=0;t&&(d=this._context.viewModel.coordinatesConverter.convertViewPositionToModelPosition(new A(n.startLineNumber,1)).lineNumber);for(let h=n.startLineNumber;h<=n.endLineNumber;h++){if(h<o||h>r)continue;const u=h===n.startLineNumber?n.startColumn:1,g=h!==i,f=g?this._context.viewModel.getLineMaxColumn(h):n.endColumn,p=this._visibleRangesForLineRange(h,u,f);if(p){if(t&&h<i){const _=d;d=this._context.viewModel.coordinatesConverter.convertViewPositionToModelPosition(new A(h+1,1)).lineNumber,_!==d&&(p.ranges[p.ranges.length-1].width+=l.spaceWidth)}c.push(new GJ(p.outsideRenderedLine,h,ex.from(p.ranges),g))}}return c.length===0?null:c}_visibleRangesForLineRange(e,t,i){if(this.shouldRender())return null;const n=this._lastViewportData,o=this._lastViewLineOptions;if(!n||!o||e<n.startLineNumber||e>n.endLineNumber)return null;const r=n.getViewLineRenderingData(e),a=r.content;let l;r.isBasicASCII&&o.useMonospaceOptimizations||(l=aI(r,o));let c="",d=0,h=0;for(let p=0;p<t-1;p++){if(r.isBasicASCII&&o.useMonospaceOptimizations)c=a.charAt(p);else{if(c=l.getSegmentAtIndex(p),c===void 0)continue;h+=this._renderStrategy.value.glyphRasterizer.getTextMetrics(c).width/kt().devicePixelRatio-o.spaceWidth}c==="	"?d=pi.nextRenderTabStop(d,r.tabSize):d++}let u=d,g=0;for(let p=t-1;p<i-1;p++){if(r.isBasicASCII&&o.useMonospaceOptimizations)c=a.charAt(p);else{if(c=l.getSegmentAtIndex(p),c===void 0)continue;g+=this._renderStrategy.value.glyphRasterizer.getTextMetrics(c).width/kt().devicePixelRatio-o.spaceWidth}c==="	"?u=pi.nextRenderTabStop(u,r.tabSize):u++}return new eF(!1,[new Ou(d*o.spaceWidth+h,(u-d)*o.spaceWidth+g)])}visibleRangeForPosition(e){const t=this._visibleRangesForLineRange(e.lineNumber,e.column,e.column);return t?new KJ(t.outsideRenderedLine,t.ranges[0].left):null}getLineWidth(e){if(!this._lastViewportData||!this._lastViewLineOptions||!this._viewGpuContext.canRender(this._lastViewLineOptions,this._lastViewportData,e))return;const t=this._lastViewportData.getViewLineRenderingData(e),n=this._visibleRangesForLineRange(e,1,t.maxColumn)?.ranges.at(-1);if(n)return n.width}getPositionAtCoordinate(e,t){if(!this._lastViewportData||!this._lastViewLineOptions||!this._viewGpuContext.canRender(this._lastViewLineOptions,this._lastViewportData,e))return;const i=this._lastViewportData.getViewLineRenderingData(e),n=i.content,o=kt().devicePixelRatio,r=t*o,a=this._lastViewLineOptions.spaceWidth*o,l=aI(i,this._lastViewLineOptions);let c=0,d=0,h=0,u=0;for(let g=0;g<n.length;g++){const f=l.getSegmentAtIndex(g);if(f===void 0){u++;continue}if(f==="	"){const p=g+h;h=pi.nextRenderTabStop(g+h,i.tabSize),d=a*(h-p),h-=g+1}else i.isBasicASCII&&this._lastViewLineOptions.useMonospaceOptimizations?d=a:d=this._renderStrategy.value.glyphRasterizer.getTextMetrics(f).width;if(r<c+d/2)break;c+=d,u++}return new A(e,u+1)}};mF=eCe([rU(2,_e),rU(3,jt)],mF);class tCe{constructor(){this._onDidChange=new M,this.onDidChange=this._onDidChange.event,this._enabled=!0}get enabled(){return this._enabled}enable(){this._enabled=!0,this._onDidChange.fire()}disable(){this._enabled=!1,this._onDidChange.fire()}}const kc=new tCe,ft=qe("keybindingService");class Oee extends no{}class Pee{_getPageOfLine(e,t){return Math.floor((e-1)/t)}_getRangeForPage(e,t){const i=e*t,n=i+1,o=i+t;return new k(n,1,o+1,1)}fromEditorSelection(e,t,i,n){const r=this._getPageOfLine(t.startLineNumber,i),a=this._getRangeForPage(r,i),l=this._getPageOfLine(t.endLineNumber,i),c=this._getRangeForPage(l,i);let d=a.intersectRanges(new k(1,1,t.startLineNumber,t.startColumn));if(n&&e.getValueLengthInRange(d,1)>500){const C=e.modifyPosition(d.getEndPosition(),-500);d=k.fromPositions(C,d.getEndPosition())}const h=e.getValueInRange(d,1),u=e.getLineCount(),g=e.getLineMaxColumn(u);let f=c.intersectRanges(new k(t.endLineNumber,t.endColumn,u,g));if(n&&e.getValueLengthInRange(f,1)>500){const C=e.modifyPosition(f.getStartPosition(),500);f=k.fromPositions(f.getStartPosition(),C)}const p=e.getValueInRange(f,1);let _;if(r===l||r+1===l)_=e.getValueInRange(t,1);else{const C=a.intersectRanges(t),v=c.intersectRanges(t);_=e.getValueInRange(C,1)+"…"+e.getValueInRange(v,1)}n&&_.length>2*500&&(_=_.substring(0,500)+"…"+_.substring(_.length-500,_.length));let b,w;return t.getDirection()===0?(b=h.length,w=h.length+_.length):(w=h.length,b=h.length+_.length),{value:h+_+p,selection:t,selectionStart:b,selectionEnd:w,startPositionWithinEditor:d.getStartPosition(),newlineCountBeforeSelection:d.endLineNumber-d.startLineNumber}}}function _F(s,e){if(s.get(2)===1){const i=e.lookupKeybinding("editor.action.toggleScreenReaderAccessibilityMode")?.getAriaLabel(),n=e.lookupKeybinding("workbench.action.showCommands")?.getAriaLabel(),o=e.lookupKeybinding("workbench.action.openGlobalKeybindings")?.getAriaLabel(),r=m(61,"The editor is not accessible at this time.");return i?m(62,"{0} To enable screen reader optimized mode, use {1}",r,i):n?m(63,"{0} To enable screen reader optimized mode, open the quick pick with {1} and run the command Toggle Screen Reader Accessibility Mode, which is currently not triggerable via keyboard.",r,n):o?m(64,"{0} Please assign a keybinding for the command Toggle Screen Reader Accessibility Mode by accessing the keybindings editor with {1} and run it.",r,o):r}return s.get(8)}function aU(s){let e=0,t=-1;do{if(t=s.indexOf(`
`,t+1),t===-1)break;e++}while(!0);return e}var iCe=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},lU=function(s,e){return function(t,i){e(t,i,s)}};class nCe{constructor(e,t,i,n,o){this._context=e,this.modelLineNumber=t,this.distanceToModelLineStart=i,this.widthOfHiddenLineTextBefore=n,this.distanceToModelLineEnd=o,this._visibleTextAreaBrand=void 0,this.startPosition=null,this.endPosition=null,this.visibleTextareaStart=null,this.visibleTextareaEnd=null,this._previousPresentation=null}prepareRender(e){const t=new A(this.modelLineNumber,this.distanceToModelLineStart+1),i=new A(this.modelLineNumber,this._context.viewModel.model.getLineMaxColumn(this.modelLineNumber)-this.distanceToModelLineEnd);this.startPosition=this._context.viewModel.coordinatesConverter.convertModelPositionToViewPosition(t),this.endPosition=this._context.viewModel.coordinatesConverter.convertModelPositionToViewPosition(i),this.startPosition.lineNumber===this.endPosition.lineNumber?(this.visibleTextareaStart=e.visibleRangeForPosition(this.startPosition),this.visibleTextareaEnd=e.visibleRangeForPosition(this.endPosition)):(this.visibleTextareaStart=null,this.visibleTextareaEnd=null)}definePresentation(e){return this._previousPresentation||(e?this._previousPresentation=e:this._previousPresentation={foreground:1,italic:!1,bold:!1,underline:!1,strikethrough:!1}),this._previousPresentation}}const lM=wo;let bF=class extends Oee{constructor(e,t,i,n,o,r){super(e),this._keybindingService=o,this._instantiationService=r,this._primaryCursorPosition=new A(1,1),this._primaryCursorVisibleRange=null,this._viewController=i,this._visibleRangeProvider=n,this._scrollLeft=0,this._scrollTop=0;const a=this._context.configuration.options,l=a.get(165);this._setAccessibilityOptions(a),this._contentLeft=l.contentLeft,this._contentWidth=l.contentWidth,this._contentHeight=l.height,this._fontInfo=a.get(59),this._emptySelectionClipboard=a.get(45),this._copyWithSyntaxHighlighting=a.get(31),this._visibleTextArea=null,this._selections=[new he(1,1,1,1)],this._modelSelections=[new he(1,1,1,1)],this._lastRenderPosition=null,this.textArea=wt(document.createElement("textarea")),kl.write(this.textArea,7),this.textArea.setClassName(`inputarea ${pw}`),this.textArea.setAttribute("wrap",this._textAreaWrapping&&!this._visibleTextArea?"on":"off");const{tabSize:c}=this._context.viewModel.model.getOptions();this.textArea.domNode.style.tabSize=`${c*this._fontInfo.spaceWidth}px`,this.textArea.setAttribute("autocorrect","off"),this.textArea.setAttribute("autocapitalize","off"),this.textArea.setAttribute("autocomplete","off"),this.textArea.setAttribute("spellcheck","false"),this.textArea.setAttribute("aria-label",_F(a,this._keybindingService)),this.textArea.setAttribute("aria-required",a.get(9)?"true":"false"),this.textArea.setAttribute("tabindex",String(a.get(140))),this.textArea.setAttribute("role","textbox"),this.textArea.setAttribute("aria-roledescription",m(65,"editor")),this.textArea.setAttribute("aria-multiline","true"),this.textArea.setAttribute("aria-autocomplete",a.get(104)?"none":"both"),this._ensureReadOnlyAttribute(),this.textAreaCover=wt(document.createElement("div")),this.textAreaCover.setPosition("absolute"),t.appendChild(this.textArea),t.appendChild(this.textAreaCover);const d=new Pee,h={getDataToCopy:()=>see(this._context.viewModel,this._modelSelections,this._emptySelectionClipboard,this._copyWithSyntaxHighlighting),getScreenReaderContent:()=>{if(this._accessibilitySupport===1){const f=this._selections[0];if(Ye&&f.isEmpty()){const _=f.getStartPosition();let b=this._getWordBeforePosition(_);if(b.length===0&&(b=this._getCharacterBeforePosition(_)),b.length>0)return new ks(b,b.length,b.length,k.fromPositions(_),0)}if(Ye&&!f.isEmpty()&&this._context.viewModel.getValueLengthInRange(f,0)<500){const _=this._context.viewModel.getValueInRange(f,0);return new ks(_,0,_.length,f,0)}if(wm&&!f.isEmpty()){const _="vscode-placeholder";return new ks(_,0,_.length,null,void 0)}return ks.EMPTY}if(Cz){const f=this._selections[0];if(f.isEmpty()){const p=f.getStartPosition(),[_,b]=this._getAndroidWordAtPosition(p);if(_.length>0)return new ks(_,b,b,k.fromPositions(p),0)}return ks.EMPTY}const g=d.fromEditorSelection(this._context.viewModel,this._selections[0],this._accessibilityPageSize,this._accessibilitySupport===0);return ks.fromScreenReaderContentState(g)},deduceModelPosition:(g,f,p)=>this._context.viewModel.deduceModelPositionRelativeToViewPosition(g,f,p)},u=this._register(new Tbe(this.textArea.domNode));this._textAreaInput=this._register(this._instantiationService.createInstance(sF,h,u,Fo,{isAndroid:Cz,isChrome:B0,isFirefox:wo,isSafari:wm})),this._register(this._textAreaInput.onKeyDown(g=>{this._viewController.emitKeyDown(g)})),this._register(this._textAreaInput.onKeyUp(g=>{this._viewController.emitKeyUp(g)})),this._register(this._textAreaInput.onPaste(g=>{let f=!1,p=null,_=null;g.metadata&&(f=this._emptySelectionClipboard&&!!g.metadata.isFromEmptySelection,p=typeof g.metadata.multicursorText<"u"?g.metadata.multicursorText:null,_=g.metadata.mode),this._viewController.paste(g.text,f,p,_)})),this._register(this._textAreaInput.onCut(()=>{this._viewController.cut()})),this._register(this._textAreaInput.onType(g=>{g.replacePrevCharCnt||g.replaceNextCharCnt||g.positionDelta?this._viewController.compositionType(g.text,g.replacePrevCharCnt,g.replaceNextCharCnt,g.positionDelta):this._viewController.type(g.text)})),this._register(this._textAreaInput.onSelectionChangeRequest(g=>{this._viewController.setSelection(g)})),this._register(this._textAreaInput.onCompositionStart(g=>{const f=this.textArea.domNode,p=this._modelSelections[0],{distanceToModelLineStart:_,widthOfHiddenTextBefore:b}=(()=>{const C=f.value.substring(0,Math.min(f.selectionStart,f.selectionEnd)),v=C.lastIndexOf(`
`),y=C.substring(v+1),x=y.lastIndexOf("	"),L=y.length-x-1,E=p.getStartPosition(),I=Math.min(E.column-1,L),R=E.column-1-I,O=y.substring(0,y.length-I),{tabSize:P}=this._context.viewModel.model.getOptions(),B=sCe(this.textArea.domNode.ownerDocument,O,this._fontInfo,P);return{distanceToModelLineStart:R,widthOfHiddenTextBefore:B}})(),{distanceToModelLineEnd:w}=(()=>{const C=f.value.substring(Math.max(f.selectionStart,f.selectionEnd)),v=C.indexOf(`
`),y=v===-1?C:C.substring(0,v),x=y.indexOf("	"),L=x===-1?y.length:y.length-x-1,E=p.getEndPosition(),I=Math.min(this._context.viewModel.model.getLineMaxColumn(E.lineNumber)-E.column,L);return{distanceToModelLineEnd:this._context.viewModel.model.getLineMaxColumn(E.lineNumber)-E.column-I}})();this._context.viewModel.revealRange("keyboard",!0,k.fromPositions(this._selections[0].getStartPosition()),0,1),this._visibleTextArea=new nCe(this._context,p.startLineNumber,_,b,w),this.textArea.setAttribute("wrap",this._textAreaWrapping&&!this._visibleTextArea?"on":"off"),this._visibleTextArea.prepareRender(this._visibleRangeProvider),this._render(),this.textArea.setClassName(`inputarea ${pw} ime-input`),this._viewController.compositionStart(),this._context.viewModel.onCompositionStart()})),this._register(this._textAreaInput.onCompositionUpdate(g=>{this._visibleTextArea&&(this._visibleTextArea.prepareRender(this._visibleRangeProvider),this._render())})),this._register(this._textAreaInput.onCompositionEnd(()=>{this._visibleTextArea=null,this.textArea.setAttribute("wrap",this._textAreaWrapping&&!this._visibleTextArea?"on":"off"),this._render(),this.textArea.setClassName(`inputarea ${pw}`),this._viewController.compositionEnd(),this._context.viewModel.onCompositionEnd()})),this._register(this._textAreaInput.onFocus(()=>{this._context.viewModel.setHasFocus(!0)})),this._register(this._textAreaInput.onBlur(()=>{this._context.viewModel.setHasFocus(!1)})),this._register(kc.onDidChange(()=>{this._ensureReadOnlyAttribute()}))}get domNode(){return this.textArea}writeScreenReaderContent(e){this._textAreaInput.writeNativeTextAreaContent(e)}dispose(){super.dispose(),this.textArea.domNode.remove(),this.textAreaCover.domNode.remove()}_getAndroidWordAtPosition(e){const t='`~!@#$%^&*()-=+[{]}\\|;:",.<>/?',i=this._context.viewModel.getLineContent(e.lineNumber),n=aa(t,[]);let o=!0,r=e.column,a=!0,l=e.column,c=0;for(;c<50&&(o||a);){if(o&&r<=1&&(o=!1),o){const d=i.charCodeAt(r-2);n.get(d)!==0?o=!1:r--}if(a&&l>i.length&&(a=!1),a){const d=i.charCodeAt(l-1);n.get(d)!==0?a=!1:l++}c++}return[i.substring(r-1,l-1),e.column-r]}_getWordBeforePosition(e){const t=this._context.viewModel.getLineContent(e.lineNumber),i=aa(this._context.configuration.options.get(148),[]);let n=e.column,o=0;for(;n>1;){const r=t.charCodeAt(n-2);if(i.get(r)!==0||o>50)return t.substring(n-1,e.column-1);o++,n--}return t.substring(0,e.column-1)}_getCharacterBeforePosition(e){if(e.column>1){const i=this._context.viewModel.getLineContent(e.lineNumber).charAt(e.column-2);if(!sn(i.charCodeAt(0)))return i}return""}_setAccessibilityOptions(e){this._accessibilitySupport=e.get(2);const t=e.get(3);this._accessibilitySupport===2&&t===Is.accessibilityPageSize.defaultValue?this._accessibilityPageSize=500:this._accessibilityPageSize=t;const n=e.get(165).wrappingColumn;if(n!==-1&&this._accessibilitySupport!==1){const o=e.get(59);this._textAreaWrapping=!0,this._textAreaWidth=Math.round(n*o.typicalHalfwidthCharacterWidth)}else this._textAreaWrapping=!1,this._textAreaWidth=lM?0:1}onConfigurationChanged(e){const t=this._context.configuration.options,i=t.get(165);this._setAccessibilityOptions(t),this._contentLeft=i.contentLeft,this._contentWidth=i.contentWidth,this._contentHeight=i.height,this._fontInfo=t.get(59),this._emptySelectionClipboard=t.get(45),this._copyWithSyntaxHighlighting=t.get(31),this.textArea.setAttribute("wrap",this._textAreaWrapping&&!this._visibleTextArea?"on":"off");const{tabSize:n}=this._context.viewModel.model.getOptions();return this.textArea.domNode.style.tabSize=`${n*this._fontInfo.spaceWidth}px`,this.textArea.setAttribute("aria-label",_F(t,this._keybindingService)),this.textArea.setAttribute("aria-required",t.get(9)?"true":"false"),this.textArea.setAttribute("tabindex",String(t.get(140))),(e.hasChanged(41)||e.hasChanged(104))&&this._ensureReadOnlyAttribute(),e.hasChanged(2)&&this._textAreaInput.writeNativeTextAreaContent("strategy changed"),!0}onCursorStateChanged(e){return this._selections=e.selections.slice(0),this._modelSelections=e.modelSelections.slice(0),this._textAreaInput.writeNativeTextAreaContent("selection changed"),!0}onDecorationsChanged(e){return!0}onFlushed(e){return!0}onLinesChanged(e){return!0}onLinesDeleted(e){return!0}onLinesInserted(e){return!0}onScrollChanged(e){return this._scrollLeft=e.scrollLeft,this._scrollTop=e.scrollTop,!0}onZonesChanged(e){return!0}isFocused(){return this._textAreaInput.isFocused()}focus(){this._textAreaInput.focusTextArea()}refreshFocusState(){this._textAreaInput.refreshFocusState()}getLastRenderData(){return this._lastRenderPosition}setAriaOptions(e){e.activeDescendant?(this.textArea.setAttribute("aria-haspopup","true"),this.textArea.setAttribute("aria-autocomplete","list"),this.textArea.setAttribute("aria-activedescendant",e.activeDescendant)):(this.textArea.setAttribute("aria-haspopup","false"),this.textArea.setAttribute("aria-autocomplete","both"),this.textArea.removeAttribute("aria-activedescendant")),e.role&&this.textArea.setAttribute("role",e.role)}_ensureReadOnlyAttribute(){const e=this._context.configuration.options;!kc.enabled||e.get(41)&&e.get(104)?this.textArea.setAttribute("readonly","true"):this.textArea.removeAttribute("readonly")}prepareRender(e){this._primaryCursorPosition=new A(this._selections[0].positionLineNumber,this._selections[0].positionColumn),this._primaryCursorVisibleRange=e.visibleRangeForPosition(this._primaryCursorPosition),this._visibleTextArea?.prepareRender(e)}render(e){this._textAreaInput.writeNativeTextAreaContent("render"),this._render()}_render(){if(this._visibleTextArea){const i=this._visibleTextArea.visibleTextareaStart,n=this._visibleTextArea.visibleTextareaEnd,o=this._visibleTextArea.startPosition,r=this._visibleTextArea.endPosition;if(o&&r&&i&&n&&n.left>=this._scrollLeft&&i.left<=this._scrollLeft+this._contentWidth){const a=this._context.viewLayout.getVerticalOffsetForLineNumber(this._primaryCursorPosition.lineNumber)-this._scrollTop,l=aU(this.textArea.domNode.value.substr(0,this.textArea.domNode.selectionStart));let c=this._visibleTextArea.widthOfHiddenLineTextBefore,d=this._contentLeft+i.left-this._scrollLeft,h=n.left-i.left+1;if(d<this._contentLeft){const C=this._contentLeft-d;d+=C,c+=C,h-=C}h>this._contentWidth&&(h=this._contentWidth);const u=this._context.viewLayout.getLineHeightForLineNumber(o.lineNumber),g=this._context.viewModel.getFontSizeAtPosition(this._primaryCursorPosition),f=this._context.viewModel.getViewLineData(o.lineNumber),p=f.tokens.findTokenIndexAtOffset(o.column-1),_=f.tokens.findTokenIndexAtOffset(r.column-1),b=p===_,w=this._visibleTextArea.definePresentation(b?f.tokens.getPresentation(p):null);this.textArea.domNode.scrollTop=l*u,this.textArea.domNode.scrollLeft=c,this._doRender({lastRenderPosition:null,top:a,left:d,width:h,height:u,useCover:!1,color:(yi.getColorMap()||[])[w.foreground],italic:w.italic,bold:w.bold,underline:w.underline,strikethrough:w.strikethrough,fontSize:g})}return}if(!this._primaryCursorVisibleRange){this._renderAtTopLeft();return}const e=this._contentLeft+this._primaryCursorVisibleRange.left-this._scrollLeft;if(e<this._contentLeft||e>this._contentLeft+this._contentWidth){this._renderAtTopLeft();return}const t=this._context.viewLayout.getVerticalOffsetForLineNumber(this._selections[0].positionLineNumber)-this._scrollTop;if(t<0||t>this._contentHeight){this._renderAtTopLeft();return}if(Ye||this._accessibilitySupport===2){const i=this._primaryCursorPosition.lineNumber,n=this._context.viewLayout.getLineHeightForLineNumber(i);this._doRender({lastRenderPosition:this._primaryCursorPosition,top:t,left:this._textAreaWrapping?this._contentLeft:e,width:this._textAreaWidth,height:n,useCover:!1}),this.textArea.domNode.scrollLeft=this._primaryCursorVisibleRange.left;const o=this._textAreaInput.textAreaState.newlineCountBeforeSelection??aU(this.textArea.domNode.value.substring(0,this.textArea.domNode.selectionStart));this.textArea.domNode.scrollTop=o*n;return}this._doRender({lastRenderPosition:this._primaryCursorPosition,top:t,left:this._textAreaWrapping?this._contentLeft:e,width:this._textAreaWidth,height:lM?0:1,useCover:!1})}_renderAtTopLeft(){this._doRender({lastRenderPosition:null,top:0,left:0,width:this._textAreaWidth,height:lM?0:1,useCover:!0})}_doRender(e){this._lastRenderPosition=e.lastRenderPosition;const t=this.textArea,i=this.textAreaCover;In(t,this._fontInfo),t.setTop(e.top),t.setLeft(e.left),t.setWidth(e.width),t.setHeight(e.height),t.setLineHeight(e.height),t.setFontSize(e.fontSize??this._fontInfo.fontSize),t.setColor(e.color?q.Format.CSS.formatHex(e.color):""),t.setFontStyle(e.italic?"italic":""),e.bold&&t.setFontWeight("bold"),t.setTextDecoration(`${e.underline?" underline":""}${e.strikethrough?" line-through":""}`),i.setTop(e.useCover?e.top:0),i.setLeft(e.useCover?e.left:0),i.setWidth(e.useCover?e.width:0),i.setHeight(e.useCover?e.height:0);const n=this._context.configuration.options;n.get(66)?i.setClassName("monaco-editor-background textAreaCover "+tI.OUTER_CLASS_NAME):n.get(76).renderType!==0?i.setClassName("monaco-editor-background textAreaCover "+JE.CLASS_NAME):i.setClassName("monaco-editor-background textAreaCover")}};bF=iCe([lU(4,ft),lU(5,_e)],bF);function sCe(s,e,t,i){if(e.length===0)return 0;const n=s.createElement("div");n.style.position="absolute",n.style.top="-50000px",n.style.width="50000px";const o=s.createElement("span");In(o,t),o.style.whiteSpace="pre",o.style.tabSize=`${i*t.spaceWidth}px`,o.append(e),n.appendChild(o),s.body.appendChild(n);const r=o.offsetWidth;return n.remove(),r}var oCe=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},rCe=function(s,e){return function(t,i){e(t,i,s)}};let wF=class extends F{constructor(e,t,i){super(),this._domNode=t,this._onFocusChange=i,this._isFocused=!1,this._isPaused=!1,this._register(W(this._domNode,"focus",()=>{e.trace("NativeEditContext.focus"),!this._isPaused&&this.refreshFocusState()})),this._register(W(this._domNode,"blur",()=>{e.trace("NativeEditContext.blur"),!this._isPaused&&this._handleFocusedChanged(!1)}))}pause(){this._isPaused=!0}resume(){this._isPaused=!1,this.refreshFocusState()}_handleFocusedChanged(e){this._isFocused!==e&&(this._isFocused=e,this._onFocusChange(this._isFocused))}focus(){this._domNode.focus(),this.refreshFocusState()}refreshFocusState(){const e=cg(this._domNode),t=e?e.activeElement:ln(),i=this._domNode===t;this._handleFocusedChanged(i)}get isFocused(){return this._isFocused}};wF=oCe([rCe(0,jt)],wF);function DC(s,e,t,i){return s.addEventListener(e,t,i),{dispose(){s.removeEventListener(e,t)}}}var aCe=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},lCe=function(s,e){return function(t,i){e(t,i,s)}};const cCe=$c("richScreenReaderContent",{createHTML:s=>s}),cM="data-line-number";let vF=class extends F{constructor(e,t,i,n){super(),this._domNode=e,this._context=t,this._viewController=i,this._accessibilityService=n,this._selectionChangeListener=this._register(new vt),this._accessibilityPageSize=1,this._ignoreSelectionChangeTime=0,this._state=SS.NULL,this._strategy=new uCe,this._renderedLines=new Map,this._renderedSelection=new he(1,1,1,1),this.onConfigurationChanged(this._context.configuration.options)}updateScreenReaderContent(e){const t=kt().document.activeElement;if(!t||t!==this._domNode.domNode)return;if(this._accessibilityService.isScreenReaderOptimized()){const n=this._getScreenReaderContentLineIntervals(e);this._state.equals(n)||(this._state=n,this._renderedLines=this._renderScreenReaderContent(n)),this._renderedSelection.equalsSelection(e)||(this._renderedSelection=e,this._setSelectionOnScreenReaderContent(this._context,this._renderedLines,e))}else this._state=SS.NULL,this._setIgnoreSelectionChangeTime("setValue"),this._domNode.domNode.textContent=""}updateScrollTop(e){const t=this._state.intervals;if(!t.length)return;const i=this._context.viewModel.viewLayout,n=t[0].startLine,o=i.getVerticalOffsetForLineNumber(n),r=i.getVerticalOffsetForLineNumber(e.positionLineNumber);this._domNode.domNode.scrollTop=r-o}onFocusChange(e){e?this._selectionChangeListener.value=this._setSelectionChangeListener():this._selectionChangeListener.value=void 0}onConfigurationChanged(e){this._accessibilityPageSize=e.get(3)}onWillCut(){this._setIgnoreSelectionChangeTime("onCut")}onWillPaste(){this._setIgnoreSelectionChangeTime("onWillPaste")}_setIgnoreSelectionChangeTime(e){this._ignoreSelectionChangeTime=Date.now()}_setSelectionChangeListener(){let e=0;return W(this._domNode.domNode.ownerDocument,"selectionchange",()=>{if(!(kt().document.activeElement===this._domNode.domNode)||!this._accessibilityService.isScreenReaderOptimized()||!kc.enabled)return;const o=Date.now(),r=o-e;if(e=o,r<5)return;const a=o-this._ignoreSelectionChangeTime;if(this._ignoreSelectionChangeTime=0,a<100)return;const l=this._getEditorSelectionFromDomRange();l&&this._viewController.setSelection(l)})}_renderScreenReaderContent(e){const t=[],i=new Map;for(const n of e.intervals)for(let o=n.startLine;o<=n.endLine;o++){const r=this._renderLine(o);i.set(o,r),t.push(r.domNode)}return this._setIgnoreSelectionChangeTime("setValue"),this._domNode.domNode.replaceChildren(...t),i}_renderLine(e){const t=this._context.viewModel,i=t.getViewLineRenderingData(e),n=this._context.configuration.options,o=n.get(59),r=n.get(133),a=n.get(108),l=n.get(60),c=n.get(40),d=Rs.filter(i.inlineDecorations,e,i.minColumn,i.maxColumn),h=o.isMonospace&&!c,u=l!==mh.OFF;let g;n.get(47)==="off"?g=n.get(113):g="none";const p=new Dh(h,o.canUseHalfwidthRightwardsArrow,i.content,i.continuesWithWrappedLine,i.isBasicASCII,i.containsRTL,i.minColumn-1,i.tokens,d,i.tabSize,i.startVisibleColumn,o.spaceWidth,o.middotWidth,o.wsmiddotWidth,r,g,a,u,null,null,0,!0),_=new Eg(1e4),b=nC(p,_),w=_.build(),C=cCe?.createHTML(w)??w,v=t.viewLayout.getLineHeightForLineNumber(e)+"px",y=document.createElement("div");return y.innerHTML=C,y.style.lineHeight=v,y.style.height=v,y.setAttribute(cM,e.toString()),new dCe(y,b.characterMapping)}_setSelectionOnScreenReaderContent(e,t,i){const o=kt().document.getSelection();if(!o)return;const r=i.startLineNumber,a=i.endLineNumber,l=t.get(r),c=t.get(a);if(!l||!c)return;const d=e.viewModel,h=d.model,u=d.coordinatesConverter,g=new k(r,1,r,i.selectionStartColumn),f=u.convertViewRangeToModelRange(g),p=h.getCharacterCountInRange(f),_=new k(a,1,a,i.positionColumn),b=u.convertViewRangeToModelRange(_),w=h.getCharacterCountInRange(b),C=l.characterMapping.getDomPosition(p),v=c.characterMapping.getDomPosition(w),y=l.domNode.firstChild,x=c.domNode.firstChild,L=y.childNodes,E=x.childNodes,I=L.item(C.partIndex),R=E.item(v.partIndex);!I.firstChild||!R.firstChild||(this._setIgnoreSelectionChangeTime("setRange"),o.setBaseAndExtent(I.firstChild,i.startColumn===1?0:C.charIndex+1,R.firstChild,i.endColumn===1?0:v.charIndex+1))}_getScreenReaderContentLineIntervals(e){return this._strategy.fromEditorSelection(this._context.viewModel,e,this._accessibilityPageSize)}_getEditorSelectionFromDomRange(){if(!this._renderedLines)return;const e=kt().document.getSelection();if(!e||e.rangeCount===0)return;const i=e.getRangeAt(0),n=i.startContainer,o=i.endContainer,r=n.parentElement,a=o.parentElement;if(!r||!ki(r)||!a||!ki(a))return;const l=r.parentElement?.parentElement,c=a.parentElement?.parentElement;if(!l||!c)return;const d=l.getAttribute(cM),h=c.getAttribute(cM);if(!d||!h)return;const u=parseInt(d),g=parseInt(h),f=this._renderedLines.get(u)?.characterMapping,p=this._renderedLines.get(g)?.characterMapping;if(!f||!p)return;const _=dv(f,r,i.startOffset),b=dv(p,a,i.endOffset);return e.direction==="forward"?new he(u,_,g,b):new he(g,b,u,_)}};vF=aCe([lCe(3,Wn)],vF);class dCe{constructor(e,t){this.domNode=e,this.characterMapping=t}}class hCe{constructor(e,t){this.startLine=e,this.endLine=t}}class SS{constructor(e,t){this.intervals=t;let i="";for(const n of t)for(let o=n.startLine;o<=n.endLine;o++)i+=e.getLineContent(o)+`
`;this.value=i}equals(e){return this.value===e.value}static get NULL(){const e={getLineContent:()=>"",getLineCount:()=>1,getLineMaxColumn:()=>1,getValueInRange:()=>"",getValueLengthInRange:()=>0,modifyPosition:(t,i)=>t};return new SS(e,[])}}class uCe{constructor(){}_getPageOfLine(e,t){return Math.floor((e-1)/t)}_getRangeForPage(e,t,i){const n=t*i,o=n+1,r=Math.min(n+i,e.getLineCount());return new hCe(o,r)}fromEditorSelection(e,t,i){const n=this._getPageOfLine(t.startLineNumber,i),o=this._getRangeForPage(e,n,i),r=this._getPageOfLine(t.endLineNumber,i),a=this._getRangeForPage(e,r,i),l=[{startLine:o.startLine,endLine:o.endLine}];return n+1<r&&l.push({startLine:a.startLine,endLine:a.endLine}),new SS(e,l)}}class gCe{constructor(e){this.replacements=e;let t=-1;for(const i of e){if(!(i.replaceRange.start>=t))throw new Ce(`Edits must be disjoint and sorted. Found ${i} after ${t}`);t=i.replaceRange.endExclusive}}toString(){return`[${this.replacements.map(t=>t.toString()).join(", ")}]`}normalize(){const e=[];let t;for(const i of this.replacements)if(!(i.getNewLength()===0&&i.replaceRange.length===0)){if(t&&t.replaceRange.endExclusive===i.replaceRange.start){const n=t.tryJoinTouching(i);if(n){t=n;continue}}t&&e.push(t),t=i}return t&&e.push(t),this._createNew(e)}compose(e){const t=this.normalize(),i=e.normalize();if(t.isEmpty())return i;if(i.isEmpty())return t;const n=[...t.replacements],o=[];let r=0;for(const a of i.replacements){for(;;){const h=n[0];if(!h||h.replaceRange.start+r+h.getNewLength()>=a.replaceRange.start)break;n.shift(),o.push(h),r+=h.getNewLength()-h.replaceRange.length}const l=r;let c,d;for(;;){const h=n[0];if(!h||h.replaceRange.start+r>a.replaceRange.endExclusive)break;c||(c=h),d=h,n.shift(),r+=h.getNewLength()-h.replaceRange.length}if(!c)o.push(a.delta(-r));else{const h=Math.min(c.replaceRange.start,a.replaceRange.start-l),u=a.replaceRange.start-(c.replaceRange.start+l);if(u>0){const _=c.slice(me.emptyAt(h),new me(0,u));o.push(_)}if(!d)throw new Ce("Invariant violation: lastIntersecting is undefined");const g=d.replaceRange.endExclusive+r-a.replaceRange.endExclusive;if(g>0){const _=d.slice(me.ofStartAndLength(d.replaceRange.endExclusive,0),new me(d.getNewLength()-g,d.getNewLength()));n.unshift(_),r-=_.getNewLength()-_.replaceRange.length}const f=new me(h,a.replaceRange.endExclusive-r),p=a.slice(f,new me(0,a.getNewLength()));o.push(p)}}for(;;){const a=n.shift();if(!a)break;o.push(a)}return this._createNew(o).normalize()}getNewRanges(){const e=[];let t=0;for(const i of this.replacements)e.push(me.ofStartAndLength(i.replaceRange.start+t,i.getNewLength())),t+=i.getLengthDelta();return e}isEmpty(){return this.replacements.length===0}}class fCe{constructor(e){this.replaceRange=e}delta(e){return this.slice(this.replaceRange.delta(e),new me(0,this.getNewLength()))}getLengthDelta(){return this.getNewLength()-this.replaceRange.length}toString(){return`{ ${this.replaceRange.toString()} -> ${this.getNewLength()} }`}get isEmpty(){return this.getNewLength()===0&&this.replaceRange.length===0}getRangeAfterReplace(){return new me(this.replaceRange.start,this.replaceRange.start+this.getNewLength())}}class pCe extends gCe{apply(e){const t=[];let i=0;for(const n of this.replacements)t.push(e.substring(i,n.replaceRange.start)),t.push(n.newText),i=n.replaceRange.endExclusive;return t.push(e.substring(i)),t.join("")}}class mCe extends fCe{constructor(e,t){super(e),this.newText=t}getNewLength(){return this.newText.length}toString(){return`${this.replaceRange} -> ${JSON.stringify(this.newText)}`}replace(e){return e.substring(0,this.replaceRange.start)+this.newText+e.substring(this.replaceRange.endExclusive)}removeCommonSuffixPrefix(e){const t=e.substring(this.replaceRange.start,this.replaceRange.endExclusive),i=Fa(t,this.newText),n=Math.min(t.length-i,this.newText.length-i,uh(t,this.newText)),o=new me(this.replaceRange.start+i,this.replaceRange.endExclusive-n),r=this.newText.substring(i,this.newText.length-n);return new On(o,r)}removeCommonSuffixAndPrefix(e){return this.removeCommonSuffix(e).removeCommonPrefix(e)}removeCommonPrefix(e){const t=this.replaceRange.substring(e),i=Fa(t,this.newText);return i===0?this:this.slice(this.replaceRange.deltaStart(i),new me(i,this.newText.length))}removeCommonSuffix(e){const t=this.replaceRange.substring(e),i=uh(t,this.newText);return i===0?this:this.slice(this.replaceRange.deltaEnd(-i),new me(0,this.newText.length-i))}toJson(){return{txt:this.newText,pos:this.replaceRange.start,len:this.replaceRange.length}}}const Xb=class Xb extends pCe{static compose(e){if(e.length===0)return Xb.empty;let t=e[0];for(let i=1;i<e.length;i++)t=t.compose(e[i]);return t}constructor(e){super(e)}_createNew(e){return new Xb(e)}};Xb.empty=new Xb([]);let vh=Xb;class On extends mCe{static insert(e,t){return new On(me.emptyAt(e),t)}static replace(e,t){return new On(e,t)}equals(e){return this.replaceRange.equals(e.replaceRange)&&this.newText===e.newText}tryJoinTouching(e){return new On(this.replaceRange.joinRightTouching(e.replaceRange),this.newText+e.newText)}slice(e,t){return new On(e,t?t.substring(this.newText):this.newText)}}function Fee(s,e){s=s.slice();const t=[];let i=0;for(const n of e.replacements){for(;;){const r=s[0];if(!r||r.endExclusive>=n.replaceRange.start)break;s.shift(),t.push(r.delta(i))}const o=[];for(;;){const r=s[0];if(!r||!r.intersectsOrTouches(n.replaceRange))break;s.shift(),o.push(r)}for(let r=o.length-1;r>=0;r--){let a=o[r];const l=a.intersect(n.replaceRange).length;a=a.deltaEnd(-l+(r===0?n.newText.length:0));const c=a.start-n.replaceRange.start;c>0&&(a=a.delta(-c)),r!==0&&(a=a.delta(n.newText.length)),a=a.delta(-(n.newText.length-n.replaceRange.length)),s.unshift(a)}i+=n.newText.length-n.replaceRange.length}for(;;){const n=s[0];if(!n)break;s.shift(),t.push(n.delta(i))}return t}const Mr=class Mr{static ofLength(e,t){return new Mr(e,e+t)}static fromRange(e){return new Mr(e.startLineNumber,e.endLineNumber)}static fromRangeInclusive(e){return new Mr(e.startLineNumber,e.endLineNumber+1)}static joinMany(e){if(e.length===0)return[];let t=new Kr(e[0].slice());for(let i=1;i<e.length;i++)t=t.getUnion(new Kr(e[i].slice()));return t.ranges}static join(e){if(e.length===0)throw new Ce("lineRanges cannot be empty");let t=e[0].startLineNumber,i=e[0].endLineNumberExclusive;for(let n=1;n<e.length;n++)t=Math.min(t,e[n].startLineNumber),i=Math.max(i,e[n].endLineNumberExclusive);return new Mr(t,i)}static deserialize(e){return new Mr(e[0],e[1])}constructor(e,t){if(e>t)throw new Ce(`startLineNumber ${e} cannot be after endLineNumberExclusive ${t}`);this.startLineNumber=e,this.endLineNumberExclusive=t}contains(e){return this.startLineNumber<=e&&e<this.endLineNumberExclusive}get isEmpty(){return this.startLineNumber===this.endLineNumberExclusive}delta(e){return new Mr(this.startLineNumber+e,this.endLineNumberExclusive+e)}deltaLength(e){return new Mr(this.startLineNumber,this.endLineNumberExclusive+e)}get length(){return this.endLineNumberExclusive-this.startLineNumber}join(e){return new Mr(Math.min(this.startLineNumber,e.startLineNumber),Math.max(this.endLineNumberExclusive,e.endLineNumberExclusive))}toString(){return`[${this.startLineNumber},${this.endLineNumberExclusive})`}intersect(e){const t=Math.max(this.startLineNumber,e.startLineNumber),i=Math.min(this.endLineNumberExclusive,e.endLineNumberExclusive);if(t<=i)return new Mr(t,i)}intersectsStrict(e){return this.startLineNumber<e.endLineNumberExclusive&&e.startLineNumber<this.endLineNumberExclusive}intersectsOrTouches(e){return this.startLineNumber<=e.endLineNumberExclusive&&e.startLineNumber<=this.endLineNumberExclusive}equals(e){return this.startLineNumber===e.startLineNumber&&this.endLineNumberExclusive===e.endLineNumberExclusive}toInclusiveRange(){return this.isEmpty?null:new k(this.startLineNumber,1,this.endLineNumberExclusive-1,Number.MAX_SAFE_INTEGER)}toExclusiveRange(){return new k(this.startLineNumber,1,this.endLineNumberExclusive,1)}mapToLineArray(e){const t=[];for(let i=this.startLineNumber;i<this.endLineNumberExclusive;i++)t.push(e(i));return t}forEach(e){for(let t=this.startLineNumber;t<this.endLineNumberExclusive;t++)e(t)}serialize(){return[this.startLineNumber,this.endLineNumberExclusive]}toOffsetRange(){return new me(this.startLineNumber-1,this.endLineNumberExclusive-1)}addMargin(e,t){return new Mr(this.startLineNumber-e,this.endLineNumberExclusive+t)}};Mr.compareByStart=ns(e=>e.startLineNumber,Wo);let Ne=Mr;class Kr{constructor(e=[]){this._normalizedRanges=e}get ranges(){return this._normalizedRanges}addRange(e){if(e.length===0)return;const t=fS(this._normalizedRanges,n=>n.endLineNumberExclusive>=e.startLineNumber),i=pv(this._normalizedRanges,n=>n.startLineNumber<=e.endLineNumberExclusive)+1;if(t===i)this._normalizedRanges.splice(t,0,e);else if(t===i-1){const n=this._normalizedRanges[t];this._normalizedRanges[t]=n.join(e)}else{const n=this._normalizedRanges[t].join(this._normalizedRanges[i-1]).join(e);this._normalizedRanges.splice(t,i-t,n)}}contains(e){const t=fv(this._normalizedRanges,i=>i.startLineNumber<=e);return!!t&&t.endLineNumberExclusive>e}intersects(e){const t=fv(this._normalizedRanges,i=>i.startLineNumber<e.endLineNumberExclusive);return!!t&&t.endLineNumberExclusive>e.startLineNumber}getUnion(e){if(this._normalizedRanges.length===0)return e;if(e._normalizedRanges.length===0)return this;const t=[];let i=0,n=0,o=null;for(;i<this._normalizedRanges.length||n<e._normalizedRanges.length;){let r=null;if(i<this._normalizedRanges.length&&n<e._normalizedRanges.length){const a=this._normalizedRanges[i],l=e._normalizedRanges[n];a.startLineNumber<l.startLineNumber?(r=a,i++):(r=l,n++)}else i<this._normalizedRanges.length?(r=this._normalizedRanges[i],i++):(r=e._normalizedRanges[n],n++);o===null?o=r:o.endLineNumberExclusive>=r.startLineNumber?o=new Ne(o.startLineNumber,Math.max(o.endLineNumberExclusive,r.endLineNumberExclusive)):(t.push(o),o=r)}return o!==null&&t.push(o),new Kr(t)}subtractFrom(e){const t=fS(this._normalizedRanges,r=>r.endLineNumberExclusive>=e.startLineNumber),i=pv(this._normalizedRanges,r=>r.startLineNumber<=e.endLineNumberExclusive)+1;if(t===i)return new Kr([e]);const n=[];let o=e.startLineNumber;for(let r=t;r<i;r++){const a=this._normalizedRanges[r];a.startLineNumber>o&&n.push(new Ne(o,a.startLineNumber)),o=a.endLineNumberExclusive}return o<e.endLineNumberExclusive&&n.push(new Ne(o,e.endLineNumberExclusive)),new Kr(n)}toString(){return this._normalizedRanges.map(e=>e.toString()).join(", ")}getIntersection(e){const t=[];let i=0,n=0;for(;i<this._normalizedRanges.length&&n<e._normalizedRanges.length;){const o=this._normalizedRanges[i],r=e._normalizedRanges[n],a=o.intersect(r);a&&!a.isEmpty&&t.push(a),o.endLineNumberExclusive<r.endLineNumberExclusive?i++:n++}return new Kr(t)}getWithDelta(e){return new Kr(this._normalizedRanges.map(t=>t.delta(e)))}}const el=class el{static betweenPositions(e,t){return e.lineNumber===t.lineNumber?new el(0,t.column-e.column):new el(t.lineNumber-e.lineNumber,t.column-1)}static fromPosition(e){return new el(e.lineNumber-1,e.column-1)}static ofRange(e){return el.betweenPositions(e.getStartPosition(),e.getEndPosition())}static ofText(e){let t=0,i=0;for(const n of e)n===`
`?(t++,i=0):i++;return new el(t,i)}constructor(e,t){this.lineCount=e,this.columnCount=t}isGreaterThanOrEqualTo(e){return this.lineCount!==e.lineCount?this.lineCount>e.lineCount:this.columnCount>=e.columnCount}add(e){return e.lineCount===0?new el(this.lineCount,this.columnCount+e.columnCount):new el(this.lineCount+e.lineCount,e.columnCount)}createRange(e){return this.lineCount===0?new k(e.lineNumber,e.column,e.lineNumber,e.column+this.columnCount):new k(e.lineNumber,e.column,e.lineNumber+this.lineCount,this.columnCount+1)}toRange(){return new k(1,1,this.lineCount+1,this.columnCount+1)}toLineRange(){return Ne.ofLength(1,this.lineCount+1)}addToPosition(e){return this.lineCount===0?new A(e.lineNumber,e.column+this.columnCount):new A(e.lineNumber+this.lineCount,this.columnCount+1)}toString(){return`${this.lineCount},${this.columnCount}`}};el.zero=new el(0,0);let cn=el;class Bee{getOffsetRange(e){return new me(this.getOffset(e.getStartPosition()),this.getOffset(e.getEndPosition()))}getRange(e){return k.fromPositions(this.getPosition(e.start),this.getPosition(e.endExclusive))}getStringReplacement(e){return new hp.deps.StringReplacement(this.getOffsetRange(e.range),e.text)}getTextReplacement(e){return new hp.deps.TextReplacement(this.getRange(e.replaceRange),e.newText)}getTextEdit(e){const t=e.replacements.map(i=>this.getTextReplacement(i));return new hp.deps.TextEdit(t)}}const BV=class BV{static get deps(){if(!this._deps)throw new Error("Dependencies not set. Call _setDependencies first.");return this._deps}};BV._deps=void 0;let hp=BV;function _Ce(s){hp._deps=s}class pA extends Bee{constructor(e){super(),this.text=e,this.lineStartOffsetByLineIdx=[],this.lineEndOffsetByLineIdx=[],this.lineStartOffsetByLineIdx.push(0);for(let t=0;t<e.length;t++)e.charAt(t)===`
`&&(this.lineStartOffsetByLineIdx.push(t+1),t>0&&e.charAt(t-1)==="\r"?this.lineEndOffsetByLineIdx.push(t-1):this.lineEndOffsetByLineIdx.push(t));this.lineEndOffsetByLineIdx.push(e.length)}getOffset(e){const t=this._validatePosition(e);return this.lineStartOffsetByLineIdx[t.lineNumber-1]+t.column-1}_validatePosition(e){if(e.lineNumber<1)return new A(1,1);const t=this.textLength.lineCount+1;if(e.lineNumber>t){const n=this.getLineLength(t);return new A(t,n+1)}if(e.column<1)return new A(e.lineNumber,1);const i=this.getLineLength(e.lineNumber);return e.column-1>i?new A(e.lineNumber,i+1):e}getPosition(e){const t=pv(this.lineStartOffsetByLineIdx,o=>o<=e),i=t+1,n=e-this.lineStartOffsetByLineIdx[t]+1;return new A(i,n)}get textLength(){const e=this.lineStartOffsetByLineIdx.length-1;return new hp.deps.TextLength(e,this.text.length-this.lineStartOffsetByLineIdx[e])}getLineLength(e){return this.lineEndOffsetByLineIdx[e-1]-this.lineStartOffsetByLineIdx[e-1]}}class JW{constructor(){this._transformer=void 0}get endPositionExclusive(){return this.length.addToPosition(new A(1,1))}get lineRange(){return this.length.toLineRange()}getValue(){return this.getValueOfRange(this.length.toRange())}getValueOfOffsetRange(e){return this.getValueOfRange(this.getTransformer().getRange(e))}getLineLength(e){return this.getValueOfRange(new k(e,1,e,Number.MAX_SAFE_INTEGER)).length}getTransformer(){return this._transformer||(this._transformer=new pA(this.getValue())),this._transformer}getLineAt(e){return this.getValueOfRange(new k(e,1,e,Number.MAX_SAFE_INTEGER))}}class bCe extends JW{constructor(e,t){Vy(t>=1),super(),this._getLineContent=e,this._lineCount=t}getValueOfRange(e){if(e.startLineNumber===e.endLineNumber)return this._getLineContent(e.startLineNumber).substring(e.startColumn-1,e.endColumn-1);let t=this._getLineContent(e.startLineNumber).substring(e.startColumn-1);for(let i=e.startLineNumber+1;i<e.endLineNumber;i++)t+=`
`+this._getLineContent(i);return t+=`
`+this._getLineContent(e.endLineNumber).substring(0,e.endColumn-1),t}getLineLength(e){return this._getLineContent(e).length}get length(){const e=this._getLineContent(this._lineCount);return new cn(this._lineCount-1,e.length)}}class eL extends bCe{constructor(e){super(t=>e[t-1],e.length)}}class vu extends JW{constructor(e){super(),this.value=e,this._t=new pA(this.value)}getValueOfRange(e){return this._t.getOffsetRange(e).substring(this.value)}get length(){return this._t.textLength}getTransformer(){return this._t}}let Na=class CF{static fromStringEdit(e,t){const i=e.replacements.map(n=>Ui.fromStringReplacement(n,t));return new CF(i)}static fromParallelReplacementsUnsorted(e){const t=e.slice().sort(ns(i=>i.range,k.compareRangesUsingStarts));return new CF(t)}constructor(e){this.replacements=e,rg(()=>H0(e,(t,i)=>t.range.getEndPosition().isBeforeOrEqual(i.range.getStartPosition())))}mapPosition(e){let t=0,i=0,n=0;for(const o of this.replacements){const r=o.range.getStartPosition();if(e.isBeforeOrEqual(r))break;const a=o.range.getEndPosition(),l=cn.ofText(o.text);if(e.isBefore(a)){const c=new A(r.lineNumber+t,r.column+(r.lineNumber+t===i?n:0)),d=l.addToPosition(c);return tL(c,d)}r.lineNumber+t!==i&&(n=0),t+=l.lineCount-(o.range.endLineNumber-o.range.startLineNumber),l.lineCount===0?a.lineNumber!==r.lineNumber?n+=l.columnCount-(a.column-1):n+=l.columnCount-(a.column-r.column):n=l.columnCount,i=a.lineNumber+t}return new A(e.lineNumber+t,e.column+(e.lineNumber+t===i?n:0))}mapRange(e){function t(r){return r instanceof A?r:r.getStartPosition()}function i(r){return r instanceof A?r:r.getEndPosition()}const n=t(this.mapPosition(e.getStartPosition())),o=i(this.mapPosition(e.getEndPosition()));return tL(n,o)}apply(e){let t="",i=new A(1,1);for(const o of this.replacements){const r=o.range,a=r.getStartPosition(),l=r.getEndPosition(),c=tL(i,a);c.isEmpty()||(t+=e.getValueOfRange(c)),t+=o.text,i=l}const n=tL(i,e.endPositionExclusive);return n.isEmpty()||(t+=e.getValueOfRange(n)),t}applyToString(e){const t=new vu(e);return this.apply(t)}getNewRanges(){const e=[];let t=0,i=0,n=0;for(const o of this.replacements){const r=cn.ofText(o.text),a=A.lift({lineNumber:o.range.startLineNumber+i,column:o.range.startColumn+(o.range.startLineNumber===t?n:0)}),l=r.createRange(a);e.push(l),i=l.endLineNumber-o.range.endLineNumber,n=l.endColumn-o.range.endColumn,t=o.range.endLineNumber}return e}toReplacement(e){if(this.replacements.length===0)throw new Ce;if(this.replacements.length===1)return this.replacements[0];const t=this.replacements[0].range.getStartPosition(),i=this.replacements[this.replacements.length-1].range.getEndPosition();let n="";for(let o=0;o<this.replacements.length;o++){const r=this.replacements[o];if(n+=r.text,o<this.replacements.length-1){const a=this.replacements[o+1],l=k.fromPositions(r.range.getEndPosition(),a.range.getStartPosition()),c=e.getValueOfRange(l);n+=c}}return new Ui(k.fromPositions(t,i),n)}toString(e){return e===void 0?this.replacements.map(t=>t.toString()).join(`
`):typeof e=="string"?this.toString(new vu(e)):this.replacements.length===0?"":this.replacements.map(t=>{const n=e.getValueOfRange(t.range),o=k.fromPositions(new A(Math.max(1,t.range.startLineNumber-1),1),t.range.getStartPosition());let r=e.getValueOfRange(o);r.length>10&&(r="..."+r.substring(r.length-10));const a=k.fromPositions(t.range.getEndPosition(),new A(t.range.endLineNumber+1,1));let l=e.getValueOfRange(a);l.length>10&&(l=l.substring(0,10)+"...");let c=n;if(c.length>10){const h=Math.floor(5);c=c.substring(0,h)+"..."+c.substring(c.length-h)}let d=t.text;if(d.length>10){const h=Math.floor(5);d=d.substring(0,h)+"..."+d.substring(d.length-h)}return c.length===0?`${r}❰${d}❱${l}`:`${r}❰${c}↦${d}❱${l}`}).join(`
`)}};class Ui{static joinReplacements(e,t){if(e.length===0)throw new Ce;if(e.length===1)return e[0];const i=e[0].range.getStartPosition(),n=e[e.length-1].range.getEndPosition();let o="";for(let r=0;r<e.length;r++){const a=e[r];if(o+=a.text,r<e.length-1){const l=e[r+1],c=k.fromPositions(a.range.getEndPosition(),l.range.getStartPosition()),d=t.getValueOfRange(c);o+=d}}return new Ui(k.fromPositions(i,n),o)}static fromStringReplacement(e,t){return new Ui(t.getTransformer().getRange(e.replaceRange),e.newText)}static delete(e){return new Ui(e,"")}constructor(e,t){this.range=e,this.text=t}get isEmpty(){return this.range.isEmpty()&&this.text.length===0}static equals(e,t){return e.range.equalsRange(t.range)&&e.text===t.text}equals(e){return Ui.equals(this,e)}removeCommonPrefixAndSuffix(e){return this.removeCommonPrefix(e).removeCommonSuffix(e)}removeCommonPrefix(e){const t=e.getValueOfRange(this.range).replaceAll(`\r
`,`
`),i=this.text.replaceAll(`\r
`,`
`),n=Fa(t,i),o=cn.ofText(t.substring(0,n)).addToPosition(this.range.getStartPosition()),r=i.substring(n),a=k.fromPositions(o,this.range.getEndPosition());return new Ui(a,r)}removeCommonSuffix(e){const t=e.getValueOfRange(this.range).replaceAll(`\r
`,`
`),i=this.text.replaceAll(`\r
`,`
`),n=uh(t,i),o=cn.ofText(t.substring(0,t.length-n)).addToPosition(this.range.getStartPosition()),r=i.substring(0,i.length-n),a=k.fromPositions(this.range.getStartPosition(),o);return new Ui(a,r)}toString(){const e=this.range.getStartPosition(),t=this.range.getEndPosition();return`(${e.lineNumber},${e.column} -> ${t.lineNumber},${t.column}): "${this.text}"`}}function tL(s,e){if(s.lineNumber===e.lineNumber&&s.column===Number.MAX_SAFE_INTEGER)return k.fromPositions(e,e);if(!s.isBeforeOrEqual(e))throw new Ce("start must be before end");return new k(s.lineNumber,s.column,e.lineNumber,e.column)}_Ce({StringEdit:vh,StringReplacement:On,TextReplacement:Ui,TextEdit:Na,TextLength:cn});var wCe=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},vCe=function(s,e){return function(t,i){e(t,i,s)}};let yF=class extends F{constructor(e,t,i,n){super(),this._domNode=e,this._context=t,this._viewController=i,this._accessibilityService=n,this._selectionChangeListener=this._register(new vt),this._accessibilityPageSize=1,this._ignoreSelectionChangeTime=0,this._strategy=new Pee,this.onConfigurationChanged(this._context.configuration.options)}updateScreenReaderContent(e){const t=this._domNode.domNode,i=kt().document.activeElement;if(!i||i!==t)return;if(this._accessibilityService.isScreenReaderOptimized()){this._state=this._getScreenReaderContentState(e),t.textContent!==this._state.value&&(this._setIgnoreSelectionChangeTime("setValue"),t.textContent=this._state.value);const o=kt().document.getSelection();if(!o)return;const r=this._getScreenReaderRange(this._state.selectionStart,this._state.selectionEnd);if(!r)return;this._setIgnoreSelectionChangeTime("setRange"),o.setBaseAndExtent(r.anchorNode,r.anchorOffset,r.focusNode,r.focusOffset)}else this._state=void 0,this._setIgnoreSelectionChangeTime("setValue"),this._domNode.domNode.textContent=""}updateScrollTop(e){if(!this._state)return;const t=this._context.viewModel.viewLayout,i=this._state.startPositionWithinEditor.lineNumber,n=t.getVerticalOffsetForLineNumber(i),o=t.getVerticalOffsetForLineNumber(e.positionLineNumber);this._domNode.domNode.scrollTop=o-n}onFocusChange(e){e?this._selectionChangeListener.value=this._setSelectionChangeListener():this._selectionChangeListener.value=void 0}onConfigurationChanged(e){this._accessibilityPageSize=e.get(3)}onWillCut(){this._setIgnoreSelectionChangeTime("onCut")}onWillPaste(){this._setIgnoreSelectionChangeTime("onWillPaste")}_setIgnoreSelectionChangeTime(e){this._ignoreSelectionChangeTime=Date.now()}_setSelectionChangeListener(){let e=0;return W(this._domNode.domNode.ownerDocument,"selectionchange",()=>{const t=this._accessibilityService.isScreenReaderOptimized();if(!this._state||!t||!kc.enabled||!(kt().document.activeElement===this._domNode.domNode))return;const o=kt().document.getSelection();if(!o||o.rangeCount===0)return;const a=o.getRangeAt(0),l=Date.now(),c=l-e;if(e=l,c<5)return;const d=l-this._ignoreSelectionChangeTime;this._ignoreSelectionChangeTime=0,!(d<100)&&this._viewController.setSelection(this._getEditorSelectionFromDomRange(this._context,this._state,o.direction,a))})}_getScreenReaderContentState(e){const t=this._strategy.fromEditorSelection(this._context.viewModel,e,this._accessibilityPageSize,this._accessibilityService.getAccessibilitySupport()===0),i=this._context.viewModel.model.getPositionAt(1/0);let n=t.value;return i.column===1&&e.getEndPosition().equals(i)&&(n+=`
`),t.value=n,t}_getScreenReaderRange(e,t){const i=this._domNode.domNode.firstChild;if(!i)return;const n=new globalThis.Range;return n.setStart(i,e),n.setEnd(i,t),{anchorNode:i,anchorOffset:e,focusNode:i,focusOffset:t}}_getEditorSelectionFromDomRange(e,t,i,n){const o=e.viewModel,r=o.model,l=o.coordinatesConverter.convertViewPositionToModelPosition(t.startPositionWithinEditor),c=r.getOffsetAt(l);let d=n.startOffset+c,h=n.endOffset+c;if(r.getEndOfLineSequence()===1){const b=t.value,w=new pA(b),C=w.getPosition(n.startOffset),v=w.getPosition(n.endOffset);d+=C.lineNumber-1,h+=v.lineNumber-1}const g=r.getPositionAt(d),f=r.getPositionAt(h),p=i==="forward"?g:f,_=i==="forward"?f:g;return he.fromPositions(p,_)}};yF=wCe([vCe(3,Wn)],yF);var CCe=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},cU=function(s,e){return function(t,i){e(t,i,s)}};let SF=class extends F{constructor(e,t,i,n,o){super(),this._domNode=e,this._context=t,this._viewController=i,this._keybindingService=n,this._accessibilityService=o,this._contentLeft=1,this._contentWidth=1,this._contentHeight=1,this._divWidth=1,this._primarySelection=new he(1,1,1,1),this._primaryCursorVisibleRange=null,this._state=this._register(new vt),this._instantiateScreenReaderContent(),this._updateConfigurationSettings(),this._updateDomAttributes()}onWillPaste(){this._state.value?.onWillPaste()}onWillCut(){this._state.value?.onWillCut()}handleFocusChange(e){this._state.value?.onFocusChange(e),this.writeScreenReaderContent()}onConfigurationChanged(e){this._instantiateScreenReaderContent(),this._updateConfigurationSettings(),this._updateDomAttributes(),e.hasChanged(2)&&this.writeScreenReaderContent()}_instantiateScreenReaderContent(){const e=this._context.configuration.options.get(107);this._renderRichContent!==e&&(this._renderRichContent=e,this._state.value=this._createScreenReaderContent(e))}_createScreenReaderContent(e){return e?new vF(this._domNode,this._context,this._viewController,this._accessibilityService):new yF(this._domNode,this._context,this._viewController,this._accessibilityService)}_updateConfigurationSettings(){const e=this._context.configuration.options,t=e.get(165),i=t.wrappingColumn;this._contentLeft=t.contentLeft,this._contentWidth=t.contentWidth,this._contentHeight=t.height,this._fontInfo=e.get(59),this._divWidth=Math.round(i*this._fontInfo.typicalHalfwidthCharacterWidth),this._state.value?.onConfigurationChanged(e)}_updateDomAttributes(){const e=this._context.configuration.options;this._domNode.domNode.setAttribute("role","textbox"),this._domNode.domNode.setAttribute("aria-required",e.get(9)?"true":"false"),this._domNode.domNode.setAttribute("aria-multiline","true"),this._domNode.domNode.setAttribute("aria-autocomplete",e.get(104)?"none":"both"),this._domNode.domNode.setAttribute("aria-roledescription",m(60,"editor")),this._domNode.domNode.setAttribute("aria-label",_F(e,this._keybindingService));const t=this._context.viewModel.model.getOptions().tabSize,i=e.get(59).spaceWidth;this._domNode.domNode.style.tabSize=`${t*i}px`;const n=e.get(154),o=n==="inherit"?e.get(153):n,r=o==="inherit"?e.get(149):o;this._domNode.domNode.style.textWrap=r==="off"?"nowrap":"wrap"}onCursorStateChanged(e){this._primarySelection=e.selections[0]??new he(1,1,1,1)}prepareRender(e){this.writeScreenReaderContent(),this._primaryCursorVisibleRange=e.visibleRangeForPosition(this._primarySelection.getPosition())}render(e){if(!this._primaryCursorVisibleRange){this._renderAtTopLeft();return}const t=this._context.viewLayout.getCurrentScrollLeft(),i=this._contentLeft+this._primaryCursorVisibleRange.left-t;if(i<this._contentLeft||i>this._contentLeft+this._contentWidth){this._renderAtTopLeft();return}const n=this._context.viewLayout.getCurrentScrollTop(),o=this._primarySelection.positionLineNumber,r=this._context.viewLayout.getVerticalOffsetForLineNumber(o)-n;if(r<0||r>this._contentHeight){this._renderAtTopLeft();return}const a=this._context.viewLayout.getLineHeightForLineNumber(o);this._doRender(r,this._contentLeft,this._divWidth,a),this._state.value?.updateScrollTop(this._primarySelection)}_renderAtTopLeft(){this._doRender(0,0,this._contentWidth,1)}_doRender(e,t,i,n){In(this._domNode,this._fontInfo),this._domNode.setTop(e),this._domNode.setLeft(t),this._domNode.setWidth(i),this._domNode.setHeight(n),this._domNode.setLineHeight(n)}setAriaOptions(e){e.activeDescendant?(this._domNode.setAttribute("aria-haspopup","true"),this._domNode.setAttribute("aria-autocomplete","list"),this._domNode.setAttribute("aria-activedescendant",e.activeDescendant)):(this._domNode.setAttribute("aria-haspopup","false"),this._domNode.setAttribute("aria-autocomplete","both"),this._domNode.removeAttribute("aria-activedescendant")),e.role&&this._domNode.setAttribute("role",e.role)}writeScreenReaderContent(){this._state.value?.updateScreenReaderContent(this._primarySelection)}};SF=CCe([cU(3,ft),cU(4,Wn)],SF);var xF;(function(s){function e(t,i){return new t.EditContext(i)}s.create=e})(xF||(xF={}));class yCe{constructor(){this._nativeEditContextMapping=new Map}register(e,t){return this._nativeEditContextMapping.set(e,t),{dispose:()=>{this._nativeEditContextMapping.delete(e)}}}get(e){return this._nativeEditContextMapping.get(e)}}const e7=new yCe;var SCe=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},dU=function(s,e){return function(t,i){e(t,i,s)}},M1;(function(s){s.NONE="edit-context-composition-none",s.SECONDARY="edit-context-composition-secondary",s.PRIMARY="edit-context-composition-primary"})(M1||(M1={}));let lI=class extends Oee{constructor(e,t,i,n,o,r,a){super(t),this._viewController=n,this._visibleRangeProvider=o,this.logService=a,this._previousEditContextSelection=new me(0,0),this._editContextPrimarySelection=new he(1,1,1,1),this._decorations=[],this._primarySelection=new he(1,1,1,1),this._targetWindowId=-1,this._scrollTop=0,this._scrollLeft=0,this._linesVisibleRanges=null,this.domNode=new FE(document.createElement("div")),this.domNode.setClassName("native-edit-context"),this._imeTextArea=new FE(document.createElement("textarea")),this._imeTextArea.setClassName("ime-text-area"),this._imeTextArea.setAttribute("readonly","true"),this._imeTextArea.setAttribute("tabindex","-1"),this._imeTextArea.setAttribute("aria-hidden","true"),this.domNode.setAttribute("autocorrect","off"),this.domNode.setAttribute("autocapitalize","off"),this.domNode.setAttribute("autocomplete","off"),this.domNode.setAttribute("spellcheck","false"),this._updateDomAttributes(),i.appendChild(this.domNode),i.appendChild(this._imeTextArea),this._parent=i.domNode,this._focusTracker=this._register(new wF(a,this.domNode.domNode,h=>{a.trace("NativeEditContext#handleFocusChange : ",h),this._screenReaderSupport.handleFocusChange(h),this._context.viewModel.setHasFocus(h)}));const l=be(this.domNode.domNode);this._editContext=xF.create(l),this.setEditContextOnDomNode(),this._screenReaderSupport=this._register(r.createInstance(SF,this.domNode,t,this._viewController)),this._register(W(this.domNode.domNode,"copy",h=>{this.logService.trace("NativeEditContext#copy"),this._ensureClipboardGetsEditorSelection(h)})),this._register(W(this.domNode.domNode,"cut",h=>{this.logService.trace("NativeEditContext#cut"),this._screenReaderSupport.onWillCut(),this._ensureClipboardGetsEditorSelection(h),this.logService.trace("NativeEditContext#cut (before viewController.cut)"),this._viewController.cut()})),this._register(W(this.domNode.domNode,"keyup",h=>this._onKeyUp(h))),this._register(W(this.domNode.domNode,"keydown",async h=>this._onKeyDown(h))),this._register(W(this._imeTextArea.domNode,"keyup",h=>this._onKeyUp(h))),this._register(W(this._imeTextArea.domNode,"keydown",async h=>this._onKeyDown(h))),this._register(W(this.domNode.domNode,"beforeinput",async h=>{(h.inputType==="insertParagraph"||h.inputType==="insertLineBreak")&&this._onType(this._viewController,{text:`
`,replacePrevCharCnt:0,replaceNextCharCnt:0,positionDelta:0})})),this._register(W(this.domNode.domNode,"paste",h=>{if(this.logService.trace("NativeEditContext#paste"),h.preventDefault(),!h.clipboardData)return;let[u,g]=lp.getTextData(h.clipboardData);if(this.logService.trace("NativeEditContext#paste with id : ",g?.id," with text.length: ",u.length),!u)return;g=g||xc.INSTANCE.get(u);let f=!1,p=null,_=null;g&&(f=this._context.configuration.options.get(45)&&!!g.isFromEmptySelection,p=typeof g.multicursorText<"u"?g.multicursorText:null,_=g.mode),this.logService.trace("NativeEditContext#paste (before viewController.paste)"),this._viewController.paste(u,f,p,_)})),this._register(DC(this._editContext,"textformatupdate",h=>this._handleTextFormatUpdate(h))),this._register(DC(this._editContext,"characterboundsupdate",h=>this._updateCharacterBounds(h)));let c;this._register(DC(this._editContext,"textupdate",h=>{const u=h.text;if(u.length===1){const g=u.charCodeAt(0);if(sn(g)){c=u;return}if(ag(g)&&c){const f={text:c+u,selectionEnd:h.selectionEnd,selectionStart:h.selectionStart,updateRangeStart:h.updateRangeStart-1,updateRangeEnd:h.updateRangeEnd-1};c=void 0,this._emitTypeEvent(this._viewController,f);return}}this._emitTypeEvent(this._viewController,h)})),this._register(DC(this._editContext,"compositionstart",h=>{this._updateEditContext(),this._viewController.compositionStart(),this._context.viewModel.onCompositionStart()})),this._register(DC(this._editContext,"compositionend",h=>{this._updateEditContext(),this._viewController.compositionEnd(),this._context.viewModel.onCompositionEnd()}));let d=!1;this._register(kc.onDidChange(()=>{kc.enabled&&d&&(this._focusTracker.resume(),this.domNode.focus(),d=!1),!kc.enabled&&this.isFocused()&&(this._focusTracker.pause(),this._imeTextArea.focus(),d=!0)})),this._register(e7.register(e,this))}dispose(){this.domNode.domNode.editContext=void 0,this.domNode.domNode.blur(),this.domNode.domNode.remove(),this._imeTextArea.domNode.remove(),super.dispose()}setAriaOptions(e){this._screenReaderSupport.setAriaOptions(e)}getLastRenderData(){return this._primarySelection.getPosition()}prepareRender(e){this._screenReaderSupport.prepareRender(e),this._updateSelectionAndControlBoundsData(e)}onDidRender(){this._updateSelectionAndControlBoundsAfterRender()}render(e){this._screenReaderSupport.render(e)}onCursorStateChanged(e){return this._primarySelection=e.modelSelections[0]??new he(1,1,1,1),this._screenReaderSupport.onCursorStateChanged(e),this._updateEditContext(),!0}onConfigurationChanged(e){return this._screenReaderSupport.onConfigurationChanged(e),this._updateDomAttributes(),!0}onDecorationsChanged(e){return!0}onFlushed(e){return!0}onLinesChanged(e){return this._updateEditContextOnLineChange(e.fromLineNumber,e.fromLineNumber+e.count-1),!0}onLinesDeleted(e){return this._updateEditContextOnLineChange(e.fromLineNumber,e.toLineNumber),!0}onLinesInserted(e){return this._updateEditContextOnLineChange(e.fromLineNumber,e.toLineNumber),!0}_updateEditContextOnLineChange(e,t){this._editContextPrimarySelection.endLineNumber<e||this._editContextPrimarySelection.startLineNumber>t||this._updateEditContext()}onScrollChanged(e){return this._scrollLeft=e.scrollLeft,this._scrollTop=e.scrollTop,!0}onZonesChanged(e){return!0}onWillPaste(){this.logService.trace("NativeEditContext#onWillPaste"),this._onWillPaste()}_onWillPaste(){this._screenReaderSupport.onWillPaste()}onWillCopy(){this.logService.trace("NativeEditContext#onWillCopy"),this.logService.trace("NativeEditContext#isFocused : ",this.domNode.domNode===ln())}writeScreenReaderContent(){this._screenReaderSupport.writeScreenReaderContent()}isFocused(){return this._focusTracker.isFocused}focus(){this._focusTracker.focus(),this.refreshFocusState()}refreshFocusState(){this._focusTracker.refreshFocusState()}setEditContextOnDomNode(){const e=be(this.domNode.domNode),t=Zy(e);this._targetWindowId!==t&&(this.domNode.domNode.editContext=this._editContext,this._targetWindowId=t)}_onKeyUp(e){this._viewController.emitKeyUp(new Pt(e))}_onKeyDown(e){const t=new Pt(e);t.keyCode===114&&t.stopPropagation(),this._viewController.emitKeyDown(t)}_updateDomAttributes(){const e=this._context.configuration.options;this.domNode.domNode.setAttribute("tabindex",String(e.get(140)))}_updateEditContext(){const e=this._getNewEditContextState();e&&(this._editContext.updateText(0,Number.MAX_SAFE_INTEGER,e.text??" "),this._editContext.updateSelection(e.selectionStartOffset,e.selectionEndOffset),this._editContextPrimarySelection=e.editContextPrimarySelection,this._previousEditContextSelection=new me(e.selectionStartOffset,e.selectionEndOffset))}_emitTypeEvent(e,t){if(!this._editContext)return;const i=this._previousEditContextSelection.endExclusive,n=this._previousEditContextSelection.start;this._previousEditContextSelection=new me(t.selectionStart,t.selectionEnd);let o=0,r=0;t.updateRangeEnd>i&&(o=t.updateRangeEnd-i),t.updateRangeStart<n&&(r=n-t.updateRangeStart);let a="";n<t.updateRangeStart&&(a+=this._editContext.text.substring(n,t.updateRangeStart)),a+=t.text,i>t.updateRangeEnd&&(a+=this._editContext.text.substring(t.updateRangeEnd,i));let l=0;t.selectionStart===t.selectionEnd&&n===i&&(l=t.selectionStart-(t.updateRangeStart+t.text.length));const c={text:a,replacePrevCharCnt:r,replaceNextCharCnt:o,positionDelta:l};this._onType(e,c)}_onType(e,t){t.replacePrevCharCnt||t.replaceNextCharCnt||t.positionDelta?e.compositionType(t.text,t.replacePrevCharCnt,t.replaceNextCharCnt,t.positionDelta):e.type(t.text)}_getNewEditContextState(){const e=this._primarySelection,t=this._context.viewModel.model;if(!t.isValidRange(e))return;const i=e.startLineNumber,n=e.endLineNumber,o=t.getLineMaxColumn(n),r=new k(i,1,n,o),a=t.getValueInRange(r,0),l=e.startColumn-1,c=a.length+e.endColumn-o;return{text:a,selectionStartOffset:l,selectionEndOffset:c,editContextPrimarySelection:e}}_editContextStartPosition(){return new A(this._editContextPrimarySelection.startLineNumber,1)}_handleTextFormatUpdate(e){if(!this._editContext)return;const t=e.getTextFormats(),i=this._editContextStartPosition(),n=[];t.forEach(o=>{const r=this._context.viewModel.model,a=r.getOffsetAt(i),l=r.getPositionAt(a+o.rangeStart),c=r.getPositionAt(a+o.rangeEnd),d=k.fromPositions(l,c),h=o.underlineThickness.toLowerCase();let u=M1.NONE;switch(h){case"thin":u=M1.SECONDARY;break;case"thick":u=M1.PRIMARY;break}n.push({range:d,options:{description:"textFormatDecoration",inlineClassName:u}})}),this._decorations=this._context.viewModel.model.deltaDecorations(this._decorations,n)}_updateSelectionAndControlBoundsData(e){const t=this._context.viewModel.coordinatesConverter.convertModelRangeToViewRange(this._primarySelection);if(this._primarySelection.isEmpty()){const i=e.visibleRangeForPosition(t.getStartPosition());this._linesVisibleRanges=i}else this._linesVisibleRanges=null}_updateSelectionAndControlBoundsAfterRender(){const t=this._context.configuration.options.get(165).contentLeft,i=this._context.viewModel.coordinatesConverter.convertModelRangeToViewRange(this._primarySelection),n=this._context.viewLayout.getVerticalOffsetForLineNumber(i.startLineNumber),o=this._context.viewLayout.getVerticalOffsetAfterLineNumber(i.endLineNumber),r=this._parent.getBoundingClientRect(),a=r.top+n-this._scrollTop,l=o-n;let c=r.left+t-this._scrollLeft,d;this._primarySelection.isEmpty()?(this._linesVisibleRanges&&(c+=this._linesVisibleRanges.left),d=0):d=r.width-t;const h=new DOMRect(c,a,d,l);this._editContext.updateSelectionBounds(h),this._editContext.updateControlBounds(h)}_updateCharacterBounds(e){const t=this._context.configuration.options,i=t.get(59).typicalHalfwidthCharacterWidth,n=t.get(165).contentLeft,o=this._parent.getBoundingClientRect(),r=[],a=new pA(this._editContext.text);for(let l=e.rangeStart;l<e.rangeEnd;l++){const c=a.getPosition(l),d=this._editContextPrimarySelection.startLineNumber-1,h=new A(d+c.lineNumber,c.column),u=h.delta(0,1),g=k.fromPositions(h,u),f=this._context.viewModel.coordinatesConverter.convertModelRangeToViewRange(g),p=this._visibleRangeProvider.linesVisibleRangesForRange(f,!0)??[],_=f.startLineNumber,b=this._context.viewLayout.getVerticalOffsetForLineNumber(_),w=o.top+b-this._scrollTop;let C=0,v=i;if(p.length>0)for(const x of p[0].ranges){C=x.left,v=x.width;break}const y=this._context.viewLayout.getLineHeightForLineNumber(_);r.push(new DOMRect(o.left+n+C-this._scrollLeft,w,v,y))}this._editContext.updateCharacterBounds(e.rangeStart,r)}_ensureClipboardGetsEditorSelection(e){const t=this._context.configuration.options,i=t.get(45),n=t.get(31),o=this._context.viewModel.getCursorStates().map(c=>c.modelState.selection),r=see(this._context.viewModel,o,i,n);let a;this.logService.getLevel()===pn.Trace&&(a=d_());const l={version:1,id:a,isFromEmptySelection:r.isFromEmptySelection,multicursorText:r.multicursorText,mode:r.mode};xc.INSTANCE.set(wo?r.text.replace(/\r\n/g,`
`):r.text,l),e.preventDefault(),e.clipboardData&&lp.setTextData(e.clipboardData,r.text,r.html,l),this.logService.trace("NativeEditContext#_ensureClipboardGetsEditorSelectios with id : ",a," with text.length: ",r.text.length)}};lI=SCe([dU(5,_e),dU(6,jt)],lI);class xCe extends no{constructor(e,t){super(e),this._viewGpuContext=t,this._gpuShapes=[],this._register(De(i=>this._updateEntries(i)))}onConfigurationChanged(e){return this._updateEntries(void 0),!0}prepareRender(e){}render(e){}_updateEntries(e){const t=this._context.configuration.options,i=t.get(116),n=t.get(59).typicalHalfwidthCharacterWidth,o=this._viewGpuContext.devicePixelRatio.read(e);for(let r=0,a=i.length;r<a;r++){const l=i[r],c=this._gpuShapes[r],d=l.color?q.fromHex(l.color):this._context.theme.getColor(rwe)??q.white,h=[l.column*n*o,0,Math.max(1,Math.ceil(o)),Number.MAX_SAFE_INTEGER,d.rgba.r/255,d.rgba.g/255,d.rgba.b/255,d.rgba.a];c?c.setRaw(h):this._gpuShapes[r]=this._viewGpuContext.rectangleRenderer.register(...h)}for(;this._gpuShapes.length>i.length;)this._gpuShapes.splice(-1,1)[0].dispose()}}const tN=class tN extends Dg{constructor(e,t){super(),this._viewGpuContext=t,this._context=e,this._renderResult=null,this._context.addEventHandler(this)}dispose(){this._context.removeEventHandler(this),this._renderResult=null,super.dispose()}onConfigurationChanged(e){return!0}onCursorStateChanged(e){return!0}onFlushed(e){return!0}onLinesChanged(e){return!0}onLinesDeleted(e){return!0}onLinesInserted(e){return!0}onScrollChanged(e){return e.scrollTopChanged}onZonesChanged(e){return!0}onDecorationsChanged(e){return!0}prepareRender(e){const t=e.visibleRange.startLineNumber,i=e.visibleRange.endLineNumber,n=e.viewportData,o=new eI(this._context.configuration,this._context.theme.type),r=[];for(let a=t;a<=i;a++){const l=a-t,c=this._viewGpuContext.canRenderDetailed(o,n,a);r[l]=c.length?`<div class="${tN.CLASS_NAME}" title="Cannot render on GPU: ${c.join(", ")}"></div>`:""}this._renderResult=r}render(e,t){if(!this._renderResult)return"";const i=t-e;return i<0||i>=this._renderResult.length?"":this._renderResult[i]}};tN.CLASS_NAME="gpu-mark";let LF=tN;var LCe=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},kCe=function(s,e){return function(t,i){e(t,i,s)}};let kF=class extends c_{constructor(e,t,i,n,o,r,a,l,c){super(),this._instantiationService=c,this._shouldRecomputeGlyphMarginLanes=!1,this._ownerID=t,this._widgetFocusTracker=this._register(new ECe(e,l)),this._register(this._widgetFocusTracker.onChange(()=>{this._context.viewModel.setHasWidgetFocus(this._widgetFocusTracker.hasFocus())})),this._selections=[new he(1,1,1,1)],this._renderAnimationFrame=null,this._overflowGuardContainer=wt(document.createElement("div")),kl.write(this._overflowGuardContainer,3),this._overflowGuardContainer.setClassName("overflow-guard"),this._viewController=new Mbe(n,r,a,i),this._context=new xve(n,o,r),this._context.addEventHandler(this),this._viewParts=[],this._editContextEnabled=this._context.configuration.options.get(170),this._accessibilitySupport=this._context.configuration.options.get(2),this._editContext=this._instantiateEditContext(),this._viewParts.push(this._editContext),this._linesContent=wt(document.createElement("div")),this._linesContent.setClassName("lines-content monaco-editor-background"),this._linesContent.setPosition("absolute"),this.domNode=wt(document.createElement("div")),this.domNode.setClassName(this._getEditorClassName()),this.domNode.setAttribute("role","code"),this._context.configuration.options.get(46)==="on"&&(this._viewGpuContext=this._instantiationService.createInstance(Ds,this._context)),this._scrollbar=new Nwe(this._context,this._linesContent,this.domNode,this._overflowGuardContainer),this._viewParts.push(this._scrollbar),this._viewLines=new lF(this._context,this._viewGpuContext,this._linesContent),this._viewGpuContext&&(this._viewLinesGpu=this._instantiationService.createInstance(mF,this._context,this._viewGpuContext)),this._viewZones=new wve(this._context),this._viewParts.push(this._viewZones);const d=new dve(this._context);this._viewParts.push(d);const h=new fve(this._context);this._viewParts.push(h);const u=new Pbe(this._context);this._viewParts.push(u),u.addDynamicOverlay(new Iwe(this._context)),u.addDynamicOverlay(new uF(this._context)),u.addDynamicOverlay(new $we(this._context)),u.addDynamicOverlay(new Twe(this._context)),u.addDynamicOverlay(new Cve(this._context));const g=new Fbe(this._context);this._viewParts.push(g),g.addDynamicOverlay(new Dwe(this._context)),g.addDynamicOverlay(new Kwe(this._context)),g.addDynamicOverlay(new Gwe(this._context)),g.addDynamicOverlay(new JE(this._context)),this._viewGpuContext&&g.addDynamicOverlay(new LF(this._context,this._viewGpuContext)),this._glyphMarginWidgets=new Mwe(this._context),this._viewParts.push(this._glyphMarginWidgets);const f=new tI(this._context);f.getDomNode().appendChild(this._viewZones.marginDomNode),f.getDomNode().appendChild(g.getDomNode()),f.getDomNode().appendChild(this._glyphMarginWidgets.domNode),this._viewParts.push(f),this._contentWidgets=new Wbe(this._context,this.domNode),this._viewParts.push(this._contentWidgets),this._viewCursors=new gF(this._context),this._viewParts.push(this._viewCursors),this._overlayWidgets=new ave(this._context,this.domNode),this._viewParts.push(this._overlayWidgets);const p=this._viewGpuContext?new xCe(this._context,this._viewGpuContext):new gve(this._context);this._viewParts.push(p);const _=new Bbe(this._context);this._viewParts.push(_);const b=new rve(this._context);if(this._viewParts.push(b),d){const w=this._scrollbar.getOverviewRulerLayoutInfo();w.parent.insertBefore(d.getDomNode(),w.insertBefore)}this._linesContent.appendChild(u.getDomNode()),"domNode"in p&&this._linesContent.appendChild(p.domNode),this._linesContent.appendChild(this._viewZones.domNode),this._linesContent.appendChild(this._viewLines.getDomNode()),this._linesContent.appendChild(this._contentWidgets.domNode),this._linesContent.appendChild(this._viewCursors.getDomNode()),this._overflowGuardContainer.appendChild(f.getDomNode()),this._overflowGuardContainer.appendChild(this._scrollbar.getDomNode()),this._viewGpuContext&&this._overflowGuardContainer.appendChild(this._viewGpuContext.canvas),this._overflowGuardContainer.appendChild(h.getDomNode()),this._overflowGuardContainer.appendChild(this._overlayWidgets.getDomNode()),this._overflowGuardContainer.appendChild(b.getDomNode()),this._overflowGuardContainer.appendChild(_.domNode),this.domNode.appendChild(this._overflowGuardContainer),l?(l.appendChild(this._contentWidgets.overflowingContentWidgetsDomNode.domNode),l.appendChild(this._overlayWidgets.overflowingOverlayWidgetsDomNode.domNode)):(this.domNode.appendChild(this._contentWidgets.overflowingContentWidgetsDomNode),this.domNode.appendChild(this._overlayWidgets.overflowingOverlayWidgetsDomNode)),this._applyLayout(),this._pointerHandler=this._register(new Abe(this._context,this._viewController,this._createPointerHandlerHelper()))}_instantiateEditContext(){return this._context.configuration.options.get(170)?this._instantiationService.createInstance(lI,this._ownerID,this._context,this._overflowGuardContainer,this._viewController,this._createTextAreaHandlerHelper()):this._instantiationService.createInstance(bF,this._context,this._overflowGuardContainer,this._viewController,this._createTextAreaHandlerHelper())}_updateEditContext(){const e=this._context.configuration.options.get(170),t=this._context.configuration.options.get(2);if(this._editContextEnabled===e&&this._accessibilitySupport===t)return;this._editContextEnabled=e,this._accessibilitySupport=t;const i=this._editContext.isFocused(),n=this._viewParts.indexOf(this._editContext);this._editContext.dispose(),this._editContext=this._instantiateEditContext(),i&&this._editContext.focus(),n!==-1&&this._viewParts.splice(n,1,this._editContext)}_computeGlyphMarginLanes(){const e=this._context.viewModel.model,t=this._context.viewModel.glyphLanes;let i=[],n=0;i=i.concat(e.getAllMarginDecorations().map(o=>{const r=o.options.glyphMargin?.position??Ll.Center;return n=Math.max(n,o.range.endLineNumber),{range:o.range,lane:r,persist:o.options.glyphMargin?.persistLane}})),i=i.concat(this._glyphMarginWidgets.getWidgets().map(o=>{const r=e.validateRange(o.preference.range);return n=Math.max(n,r.endLineNumber),{range:r,lane:o.preference.lane}})),i.sort((o,r)=>k.compareRangesUsingStarts(o.range,r.range)),t.reset(n);for(const o of i)t.push(o.lane,o.range,o.persist);return t}_createPointerHandlerHelper(){return{viewDomNode:this.domNode.domNode,linesContentDomNode:this._linesContent.domNode,viewLinesDomNode:this._viewLines.getDomNode().domNode,viewLinesGpu:this._viewLinesGpu,focusTextArea:()=>{this.focus()},dispatchTextAreaEvent:e=>{this._editContext.domNode.domNode.dispatchEvent(e)},getLastRenderData:()=>{const e=this._viewCursors.getLastRenderData()||[],t=this._editContext.getLastRenderData();return new rbe(e,t)},renderNow:()=>{this.render(!0,!1)},shouldSuppressMouseDownOnViewZone:e=>this._viewZones.shouldSuppressMouseDownOnViewZone(e),shouldSuppressMouseDownOnWidget:e=>this._contentWidgets.shouldSuppressMouseDownOnWidget(e),getPositionFromDOMInfo:(e,t)=>(this._flushAccumulatedAndRenderNow(),this._viewLines.getPositionFromDOMInfo(e,t)),visibleRangeForPosition:(e,t)=>{this._flushAccumulatedAndRenderNow();const i=new A(e,t);return this._viewLines.visibleRangeForPosition(i)??this._viewLinesGpu?.visibleRangeForPosition(i)??null},getLineWidth:e=>{if(this._flushAccumulatedAndRenderNow(),this._viewLinesGpu){const t=this._viewLinesGpu.getLineWidth(e);if(t!==void 0)return t}return this._viewLines.getLineWidth(e)}}}_createTextAreaHandlerHelper(){return{visibleRangeForPosition:e=>(this._flushAccumulatedAndRenderNow(),this._viewLines.visibleRangeForPosition(e)),linesVisibleRangesForRange:(e,t)=>(this._flushAccumulatedAndRenderNow(),this._viewLines.linesVisibleRangesForRange(e,t))}}_applyLayout(){const t=this._context.configuration.options.get(165);this.domNode.setWidth(t.width),this.domNode.setHeight(t.height),this._overflowGuardContainer.setWidth(t.width),this._overflowGuardContainer.setHeight(t.height),this._linesContent.setWidth(16777216),this._linesContent.setHeight(16777216)}_getEditorClassName(){const e=this._editContext.isFocused()?" focused":"";return this._context.configuration.options.get(162)+" "+rF(this._context.theme.type)+e}handleEvents(e){super.handleEvents(e),this._scheduleRender()}onConfigurationChanged(e){return this.domNode.setClassName(this._getEditorClassName()),this._updateEditContext(),this._applyLayout(),!1}onCursorStateChanged(e){return this._selections=e.selections,!1}onDecorationsChanged(e){return e.affectsGlyphMargin&&(this._shouldRecomputeGlyphMarginLanes=!0),!1}onFocusChanged(e){return this.domNode.setClassName(this._getEditorClassName()),!1}onThemeChanged(e){return this._context.theme.update(e.theme),this.domNode.setClassName(this._getEditorClassName()),!1}dispose(){this._renderAnimationFrame!==null&&(this._renderAnimationFrame.dispose(),this._renderAnimationFrame=null),this._contentWidgets.overflowingContentWidgetsDomNode.domNode.remove(),this._overlayWidgets.overflowingOverlayWidgetsDomNode.domNode.remove(),this._context.removeEventHandler(this),this._viewGpuContext?.dispose(),this._viewLines.dispose(),this._viewLinesGpu?.dispose();for(const e of this._viewParts)e.dispose();super.dispose()}_scheduleRender(){if(this._store.isDisposed)throw new Ce;if(this._renderAnimationFrame===null){this._editContext instanceof lI&&this._editContext.setEditContextOnDomNode();const e=this._createCoordinatedRendering();this._renderAnimationFrame=EF.INSTANCE.scheduleCoordinatedRendering({window:be(this.domNode?.domNode),prepareRenderText:()=>{if(this._store.isDisposed)throw new Ce;try{return e.prepareRenderText()}finally{this._renderAnimationFrame=null}},renderText:()=>{if(this._store.isDisposed)throw new Ce;return e.renderText()},prepareRender:(t,i)=>{if(this._store.isDisposed)throw new Ce;return e.prepareRender(t,i)},render:(t,i)=>{if(this._store.isDisposed)throw new Ce;return e.render(t,i)}})}}_flushAccumulatedAndRenderNow(){const e=this._createCoordinatedRendering();Cu(()=>e.prepareRenderText());const t=Cu(()=>e.renderText());if(t){const[i,n]=t;Cu(()=>e.prepareRender(i,n)),Cu(()=>e.render(i,n))}}_getViewPartsToRender(){const e=[];let t=0;for(const i of this._viewParts)i.shouldRender()&&(e[t++]=i);return e}_createCoordinatedRendering(){return{prepareRenderText:()=>{if(this._shouldRecomputeGlyphMarginLanes){this._shouldRecomputeGlyphMarginLanes=!1;const e=this._computeGlyphMarginLanes();this._context.configuration.setGlyphMarginDecorationLaneCount(e.requiredLanes)}pu.onRenderStart()},renderText:()=>{if(!this.domNode.domNode.isConnected)return null;let e=this._getViewPartsToRender();if(!this._viewLines.shouldRender()&&e.length===0)return null;const t=this._context.viewLayout.getLinesViewportData();this._context.viewModel.setViewport(t.startLineNumber,t.endLineNumber,t.centeredLineNumber);const i=new yve(this._selections,t,this._context.viewLayout.getWhitespaceViewportData(),this._context.viewModel);return this._contentWidgets.shouldRender()&&this._contentWidgets.onBeforeRender(i),this._viewLines.shouldRender()&&(this._viewLines.renderText(i),this._viewLines.onDidRender(),e=this._getViewPartsToRender()),this._viewLinesGpu?.shouldRender()&&(this._viewLinesGpu.renderText(i),this._viewLinesGpu.onDidRender()),[e,new z_e(this._context.viewLayout,i,this._viewLines,this._viewLinesGpu)]},prepareRender:(e,t)=>{for(const i of e)i.prepareRender(t)},render:(e,t)=>{for(const i of e)i.render(t),i.onDidRender()}}}delegateVerticalScrollbarPointerDown(e){this._scrollbar.delegateVerticalScrollbarPointerDown(e)}delegateScrollFromMouseWheelEvent(e){this._scrollbar.delegateScrollFromMouseWheelEvent(e)}restoreState(e){this._context.viewModel.viewLayout.setScrollPosition({scrollTop:e.scrollTop,scrollLeft:e.scrollLeft},1),this._context.viewModel.visibleLinesStabilized()}getOffsetForColumn(e,t){const i=this._context.viewModel.model.validatePosition({lineNumber:e,column:t}),n=this._context.viewModel.coordinatesConverter.convertModelPositionToViewPosition(i);this._flushAccumulatedAndRenderNow();const o=this._viewLines.visibleRangeForPosition(new A(n.lineNumber,n.column));return o?o.left:-1}getTargetAtClientPoint(e,t){const i=this._pointerHandler.getTargetAtClientPoint(e,t);return i?dA.convertViewToModelMouseTarget(i,this._context.viewModel.coordinatesConverter):null}createOverviewRuler(e){return new uve(this._context,e)}change(e){this._viewZones.changeViewZones(e),this._scheduleRender()}render(e,t){if(t){this._viewLines.forceShouldRender();for(const i of this._viewParts)i.forceShouldRender()}e?this._flushAccumulatedAndRenderNow():this._scheduleRender()}writeScreenReaderContent(e){this._editContext.writeScreenReaderContent(e)}focus(){this._editContext.focus()}isFocused(){return this._editContext.isFocused()}isWidgetFocused(){return this._widgetFocusTracker.hasFocus()}setAriaOptions(e){this._editContext.setAriaOptions(e)}addContentWidget(e){this._contentWidgets.addWidget(e.widget),this.layoutContentWidget(e),this._scheduleRender()}layoutContentWidget(e){this._contentWidgets.setWidgetPosition(e.widget,e.position?.position??null,e.position?.secondaryPosition??null,e.position?.preference??null,e.position?.positionAffinity??null),this._scheduleRender()}removeContentWidget(e){this._contentWidgets.removeWidget(e.widget),this._scheduleRender()}addOverlayWidget(e){this._overlayWidgets.addWidget(e.widget),this.layoutOverlayWidget(e),this._scheduleRender()}layoutOverlayWidget(e){this._overlayWidgets.setWidgetPosition(e.widget,e.position)&&this._scheduleRender()}removeOverlayWidget(e){this._overlayWidgets.removeWidget(e.widget),this._scheduleRender()}addGlyphMarginWidget(e){this._glyphMarginWidgets.addWidget(e.widget),this._shouldRecomputeGlyphMarginLanes=!0,this._scheduleRender()}layoutGlyphMarginWidget(e){const t=e.position;this._glyphMarginWidgets.setWidgetPosition(e.widget,t)&&(this._shouldRecomputeGlyphMarginLanes=!0,this._scheduleRender())}removeGlyphMarginWidget(e){this._glyphMarginWidgets.removeWidget(e.widget),this._shouldRecomputeGlyphMarginLanes=!0,this._scheduleRender()}};kF=LCe([kCe(8,_e)],kF);function Cu(s){try{return s()}catch(e){return Oe(e),null}}const iN=class iN{constructor(){this._coordinatedRenderings=[],this._animationFrameRunners=new Map}scheduleCoordinatedRendering(e){return this._coordinatedRenderings.push(e),this._scheduleRender(e.window),{dispose:()=>{const t=this._coordinatedRenderings.indexOf(e);if(t!==-1&&(this._coordinatedRenderings.splice(t,1),this._coordinatedRenderings.length===0)){for(const[i,n]of this._animationFrameRunners)n.dispose();this._animationFrameRunners.clear()}}}}_scheduleRender(e){if(!this._animationFrameRunners.has(e)){const t=()=>{this._animationFrameRunners.delete(e),this._onRenderScheduled()};this._animationFrameRunners.set(e,xE(e,t,100))}}_onRenderScheduled(){const e=this._coordinatedRenderings.slice(0);this._coordinatedRenderings=[];for(const i of e)Cu(()=>i.prepareRenderText());const t=[];for(let i=0,n=e.length;i<n;i++){const o=e[i];t[i]=Cu(()=>o.renderText())}for(let i=0,n=e.length;i<n;i++){const o=e[i],r=t[i];if(!r)continue;const[a,l]=r;Cu(()=>o.prepareRender(a,l))}for(let i=0,n=e.length;i<n;i++){const o=e[i],r=t[i];if(!r)continue;const[a,l]=r;Cu(()=>o.render(a,l))}}};iN.INSTANCE=new iN;let EF=iN;class ECe extends F{constructor(e,t){super(),this._onChange=this._register(new M),this.onChange=this._onChange.event,this._hadFocus=void 0,this._hasDomElementFocus=!1,this._domFocusTracker=this._register(oa(e)),this._overflowWidgetsDomNodeHasFocus=!1,this._register(this._domFocusTracker.onDidFocus(()=>{this._hasDomElementFocus=!0,this._update()})),this._register(this._domFocusTracker.onDidBlur(()=>{this._hasDomElementFocus=!1,this._update()})),t&&(this._overflowWidgetsDomNode=this._register(oa(t)),this._register(this._overflowWidgetsDomNode.onDidFocus(()=>{this._overflowWidgetsDomNodeHasFocus=!0,this._update()})),this._register(this._overflowWidgetsDomNode.onDidBlur(()=>{this._overflowWidgetsDomNodeHasFocus=!1,this._update()})))}_update(){const e=this._hasDomElementFocus||this._overflowWidgetsDomNodeHasFocus;this._hadFocus!==e&&(this._hadFocus=e,this._onChange.fire(void 0))}hasFocus(){return this._hadFocus??!1}}class O1{constructor(e,t,i,n,o){this.injectionOffsets=e,this.injectionOptions=t,this.breakOffsets=i,this.breakOffsetsVisibleColumn=n,this.wrappedTextIndentLength=o}getOutputLineCount(){return this.breakOffsets.length}getMinOutputOffset(e){return e>0?this.wrappedTextIndentLength:0}getLineLength(e){const t=e>0?this.breakOffsets[e-1]:0;let n=this.breakOffsets[e]-t;return e>0&&(n+=this.wrappedTextIndentLength),n}getMaxOutputOffset(e){return this.getLineLength(e)}translateToInputOffset(e,t){e>0&&(t=Math.max(0,t-this.wrappedTextIndentLength));let n=e===0?t:this.breakOffsets[e-1]+t;if(this.injectionOffsets!==null)for(let o=0;o<this.injectionOffsets.length&&n>this.injectionOffsets[o];o++)n<this.injectionOffsets[o]+this.injectionOptions[o].content.length?n=this.injectionOffsets[o]:n-=this.injectionOptions[o].content.length;return n}translateToOutputPosition(e,t=2){let i=e;if(this.injectionOffsets!==null)for(let n=0;n<this.injectionOffsets.length&&!(e<this.injectionOffsets[n]||t!==1&&e===this.injectionOffsets[n]);n++)i+=this.injectionOptions[n].content.length;return this.offsetInInputWithInjectionsToOutputPosition(i,t)}offsetInInputWithInjectionsToOutputPosition(e,t=2){let i=0,n=this.breakOffsets.length-1,o=0,r=0;for(;i<=n;){o=i+(n-i)/2|0;const l=this.breakOffsets[o];if(r=o>0?this.breakOffsets[o-1]:0,t===0)if(e<=r)n=o-1;else if(e>l)i=o+1;else break;else if(e<r)n=o-1;else if(e>=l)i=o+1;else break}let a=e-r;return o>0&&(a+=this.wrappedTextIndentLength),new iL(o,a)}normalizeOutputPosition(e,t,i){if(this.injectionOffsets!==null){const n=this.outputPositionToOffsetInInputWithInjections(e,t),o=this.normalizeOffsetInInputWithInjectionsAroundInjections(n,i);if(o!==n)return this.offsetInInputWithInjectionsToOutputPosition(o,i)}if(i===0){if(e>0&&t===this.getMinOutputOffset(e))return new iL(e-1,this.getMaxOutputOffset(e-1))}else if(i===1){const n=this.getOutputLineCount()-1;if(e<n&&t===this.getMaxOutputOffset(e))return new iL(e+1,this.getMinOutputOffset(e+1))}return new iL(e,t)}outputPositionToOffsetInInputWithInjections(e,t){return e>0&&(t=Math.max(0,t-this.wrappedTextIndentLength)),(e>0?this.breakOffsets[e-1]:0)+t}normalizeOffsetInInputWithInjectionsAroundInjections(e,t){const i=this.getInjectedTextAtOffset(e);if(!i)return e;if(t===2){if(e===i.offsetInInputWithInjections+i.length&&hU(this.injectionOptions[i.injectedTextIndex].cursorStops))return i.offsetInInputWithInjections+i.length;{let n=i.offsetInInputWithInjections;if(uU(this.injectionOptions[i.injectedTextIndex].cursorStops))return n;let o=i.injectedTextIndex-1;for(;o>=0&&this.injectionOffsets[o]===this.injectionOffsets[i.injectedTextIndex]&&!(hU(this.injectionOptions[o].cursorStops)||(n-=this.injectionOptions[o].content.length,uU(this.injectionOptions[o].cursorStops)));)o--;return n}}else if(t===1||t===4){let n=i.offsetInInputWithInjections+i.length,o=i.injectedTextIndex;for(;o+1<this.injectionOffsets.length&&this.injectionOffsets[o+1]===this.injectionOffsets[o];)n+=this.injectionOptions[o+1].content.length,o++;return n}else if(t===0||t===3){let n=i.offsetInInputWithInjections,o=i.injectedTextIndex;for(;o-1>=0&&this.injectionOffsets[o-1]===this.injectionOffsets[o];)n-=this.injectionOptions[o-1].content.length,o--;return n}W0()}getInjectedText(e,t){const i=this.outputPositionToOffsetInInputWithInjections(e,t),n=this.getInjectedTextAtOffset(i);return n?{options:this.injectionOptions[n.injectedTextIndex]}:null}getInjectedTextAtOffset(e){const t=this.injectionOffsets,i=this.injectionOptions;if(t!==null){let n=0;for(let o=0;o<t.length;o++){const r=i[o].content.length,a=t[o]+n,l=t[o]+n+r;if(a>e)break;if(e<=l)return{injectedTextIndex:o,offsetInInputWithInjections:a,length:r};n+=r}}}}function hU(s){return s==null?!0:s===jr.Right||s===jr.Both}function uU(s){return s==null?!0:s===jr.Left||s===jr.Both}class iL{constructor(e,t){this.outputLineIndex=e,this.outputOffset=t}toString(){return`${this.outputLineIndex}:${this.outputOffset}`}toPosition(e){return new A(e+this.outputLineIndex,this.outputOffset+1)}}class ICe{constructor(){this.changeType=1}}class El{static applyInjectedText(e,t){if(!t||t.length===0)return e;let i="",n=0;for(const o of t)i+=e.substring(n,o.column-1),n=o.column-1,i+=o.options.content;return i+=e.substring(n),i}static fromDecorations(e){const t=[];for(const i of e)i.options.before&&i.options.before.content.length>0&&t.push(new El(i.ownerId,i.range.startLineNumber,i.range.startColumn,i.options.before,0)),i.options.after&&i.options.after.content.length>0&&t.push(new El(i.ownerId,i.range.endLineNumber,i.range.endColumn,i.options.after,1));return t.sort((i,n)=>i.lineNumber===n.lineNumber?i.column===n.column?i.order-n.order:i.column-n.column:i.lineNumber-n.lineNumber),t}constructor(e,t,i,n,o){this.ownerId=e,this.lineNumber=t,this.column=i,this.options=n,this.order=o}}class gU{constructor(e,t,i){this.changeType=2,this.lineNumber=e,this.detail=t,this.injectedText=i}}class DCe{constructor(e,t,i,n){this.ownerId=e,this.decorationId=t,this.lineNumber=i,this.lineHeight=n}}class TCe{constructor(e,t){this.ownerId=e,this.lineNumber=t}}class NCe{constructor(e,t){this.changeType=3,this.fromLineNumber=e,this.toLineNumber=t}}class RCe{constructor(e,t,i,n){this.changeType=4,this.injectedTexts=n,this.fromLineNumber=e,this.toLineNumber=t,this.detail=i}}class ACe{constructor(){this.changeType=5}}class _w{constructor(e,t,i,n){this.changes=e,this.versionId=t,this.isUndoing=i,this.isRedoing=n,this.resultingSelection=null}containsEvent(e){for(let t=0,i=this.changes.length;t<i;t++)if(this.changes[t].changeType===e)return!0;return!1}static merge(e,t){const i=[].concat(e.changes).concat(t.changes),n=t.versionId,o=e.isUndoing||t.isUndoing,r=e.isRedoing||t.isRedoing;return new _w(i,n,o,r)}}class Wee{constructor(e){this.changes=e}}let Hee=class{constructor(e){this.changes=e}},Vee=class{constructor(e){this.changes=e}};class up{constructor(e,t){this.rawContentChangedEvent=e,this.contentChangedEvent=t}merge(e){const t=_w.merge(this.rawContentChangedEvent,e.rawContentChangedEvent),i=up._mergeChangeEvents(this.contentChangedEvent,e.contentChangedEvent);return new up(t,i)}static _mergeChangeEvents(e,t){const i=[].concat(e.changes).concat(t.changes),n=t.eol,o=t.versionId,r=e.isUndoing||t.isUndoing,a=e.isRedoing||t.isRedoing,l=e.isFlush||t.isFlush,c=e.isEolChange&&t.isEolChange;return{changes:i,eol:n,isEolChange:c,versionId:o,isUndoing:r,isRedoing:a,isFlush:l,detailedReasons:e.detailedReasons.concat(t.detailedReasons),detailedReasonsChangeLengths:e.detailedReasonsChangeLengths.concat(t.detailedReasonsChangeLengths)}}}const MCe=$c("domLineBreaksComputer",{createHTML:s=>s});class t7{static create(e){return new t7(new WeakRef(e))}constructor(e){this.targetWindow=e}createLineBreaksComputer(e,t,i,n,o,r){const a=[],l=[];return{addRequest:(c,d,h)=>{a.push(c),l.push(d)},finalize:()=>OCe(gu(this.targetWindow.deref()),a,e,t,i,n,o,l)}}}function OCe(s,e,t,i,n,o,r,a){function l(I){const R=a[I];if(R){const O=El.applyInjectedText(e[I],R),P=R.map($=>$.options),B=R.map($=>$.column-1);return new O1(B,P,[O.length],[],0)}else return null}if(n===-1){const I=[];for(let R=0,O=e.length;R<O;R++)I[R]=l(R);return I}const c=Math.round(n*t.typicalHalfwidthCharacterWidth),h=Math.round(i*(o===3?2:o===2?1:0)),u=Math.ceil(t.spaceWidth*h),g=document.createElement("div");In(g,t);const f=new Eg(1e4),p=[],_=[],b=[],w=[],C=[];for(let I=0;I<e.length;I++){const R=El.applyInjectedText(e[I],a[I]);let O=0,P=0,B=c;if(o!==0)if(O=As(R),O===-1)O=0;else{for(let J=0;J<O;J++){const ne=R.charCodeAt(J)===9?i-P%i:1;P+=ne}const Q=Math.ceil(t.spaceWidth*P);Q+t.typicalFullwidthCharacterWidth>c?(O=0,P=0):B=c-Q}const $=R.substr(O),z=PCe($,P,i,B,f,u);p[I]=O,_[I]=P,b[I]=$,w[I]=z[0],C[I]=z[1]}const v=f.build(),y=MCe?.createHTML(v)??v;g.innerHTML=y,g.style.position="absolute",g.style.top="10000",r==="keepAll"?(g.style.wordBreak="keep-all",g.style.overflowWrap="anywhere"):(g.style.wordBreak="inherit",g.style.overflowWrap="break-word"),s.document.body.appendChild(g);const x=document.createRange(),L=Array.prototype.slice.call(g.children,0),E=[];for(let I=0;I<e.length;I++){const R=L[I],O=FCe(x,R,b[I],w[I]);if(O===null){E[I]=l(I);continue}const P=p[I],B=_[I]+h,$=C[I],z=[];for(let de=0,ke=O.length;de<ke;de++)z[de]=$[O[de]];if(P!==0)for(let de=0,ke=O.length;de<ke;de++)O[de]+=P;let Q,J;const ne=a[I];ne?(Q=ne.map(de=>de.options),J=ne.map(de=>de.column-1)):(Q=null,J=null),E[I]=new O1(J,Q,O,z,B)}return g.remove(),E}function PCe(s,e,t,i,n,o){if(o!==0){const u=String(o);n.appendString('<div style="text-indent: -'),n.appendString(u),n.appendString("px; padding-left: "),n.appendString(u),n.appendString("px; box-sizing: border-box; width:")}else n.appendString('<div style="width:');n.appendString(String(i)),n.appendString('px;">');const r=s.length;let a=e,l=0;const c=[],d=[];let h=0<r?s.charCodeAt(0):0;n.appendString("<span>");for(let u=0;u<r;u++){u!==0&&u%16384===0&&n.appendString("</span><span>"),c[u]=l,d[u]=a;const g=h;h=u+1<r?s.charCodeAt(u+1):0;let f=1,p=1;switch(g){case 9:f=t-a%t,p=f;for(let _=1;_<=f;_++)_<f?n.appendCharCode(160):n.appendASCIICharCode(32);break;case 32:h===32?n.appendCharCode(160):n.appendASCIICharCode(32);break;case 60:n.appendString("&lt;");break;case 62:n.appendString("&gt;");break;case 38:n.appendString("&amp;");break;case 0:n.appendString("&#00;");break;case 65279:case 8232:case 8233:case 133:n.appendCharCode(65533);break;default:gh(g)&&p++,g<32?n.appendCharCode(9216+g):n.appendCharCode(g)}l+=f,a+=p}return n.appendString("</span>"),c[s.length]=l,d[s.length]=a,n.appendString("</div>"),[c,d]}function FCe(s,e,t,i){if(t.length<=1)return null;const n=Array.prototype.slice.call(e.children,0),o=[];try{IF(s,n,i,0,null,t.length-1,null,o)}catch(r){return console.log(r),null}return o.length===0?null:(o.push(t.length),o)}function IF(s,e,t,i,n,o,r,a){if(i===o||(n=n||dM(s,e,t[i],t[i+1]),r=r||dM(s,e,t[o],t[o+1]),Math.abs(n[0].top-r[0].top)<=.1))return;if(i+1===o){a.push(o);return}const l=i+(o-i)/2|0,c=dM(s,e,t[l],t[l+1]);IF(s,e,t,i,n,l,c,a),IF(s,e,t,l,c,o,r,a)}function dM(s,e,t,i){return s.setStart(e[t/16384|0].firstChild,t%16384),s.setEnd(e[i/16384|0].firstChild,i%16384),s.getClientRects()}class BCe extends F{constructor(){super(),this._editor=null,this._instantiationService=null,this._instances=this._register(new LR),this._pending=new Map,this._finishedInstantiation=[],this._finishedInstantiation[0]=!1,this._finishedInstantiation[1]=!1,this._finishedInstantiation[2]=!1,this._finishedInstantiation[3]=!1}initialize(e,t,i){this._editor=e,this._instantiationService=i;for(const n of t){if(this._pending.has(n.id)){Oe(new Error(`Cannot have two contributions with the same id ${n.id}`));continue}this._pending.set(n.id,n)}this._instantiateSome(0),this._register(r1(be(this._editor.getDomNode()),()=>{this._instantiateSome(1)})),this._register(r1(be(this._editor.getDomNode()),()=>{this._instantiateSome(2)})),this._register(r1(be(this._editor.getDomNode()),()=>{this._instantiateSome(3)},5e3))}saveViewState(){const e={};for(const[t,i]of this._instances)typeof i.saveViewState=="function"&&(e[t]=i.saveViewState());return e}restoreViewState(e){for(const[t,i]of this._instances)typeof i.restoreViewState=="function"&&i.restoreViewState(e[t])}get(e){return this._instantiateById(e),this._instances.get(e)||null}onBeforeInteractionEvent(){this._instantiateSome(2)}onAfterModelAttached(){return r1(be(this._editor?.getDomNode()),()=>{this._instantiateSome(1)},50)}_instantiateSome(e){if(this._finishedInstantiation[e])return;this._finishedInstantiation[e]=!0;const t=this._findPendingContributionsByInstantiation(e);for(const i of t)this._instantiateById(i.id)}_findPendingContributionsByInstantiation(e){const t=[];for(const[,i]of this._pending)i.instantiation===e&&t.push(i);return t}_instantiateById(e){const t=this._pending.get(e);if(t){if(this._pending.delete(e),!this._instantiationService||!this._editor)throw new Error("Cannot instantiate contributions before being initialized!");try{const i=this._instantiationService.createInstance(t.ctor,this._editor);this._instances.set(t.id,i),typeof i.restoreViewState=="function"&&t.instantiation!==0&&console.warn(`Editor contribution '${t.id}' should be eager instantiated because it uses saveViewState / restoreViewState.`)}catch(i){Oe(i)}}}}class zee{constructor(e,t,i,n,o,r,a){this.id=e,this.label=t,this.alias=i,this.metadata=n,this._precondition=o,this._run=r,this._contextKeyService=a}isSupported(){return this._contextKeyService.contextMatchesRules(this._precondition)}run(e){return this.isSupported()?this._run(e):Promise.resolve(void 0)}}const sx={ICodeEditor:"vs.editor.ICodeEditor",IDiffEditor:"vs.editor.IDiffEditor"};function fg(s){let e=0,t=0,i=0,n=0;for(let o=0,r=s.length;o<r;o++){const a=s.charCodeAt(o);a===13?(e===0&&(t=o),e++,o+1<r&&s.charCodeAt(o+1)===10?(n|=2,o++):n|=3,i=o+1):a===10&&(n|=1,e===0&&(t=o),e++,i=o+1)}return e===0&&(t=s.length),[e,t,s.length-i,n]}class fU{constructor(e,t,i,n){this.range=e,this.nestingLevel=t,this.nestingLevelOfEqualBracketType=i,this.isInvalid=n}}class WCe{constructor(e,t,i,n,o,r){this.range=e,this.openingBracketRange=t,this.closingBracketRange=i,this.nestingLevel=n,this.nestingLevelOfEqualBracketType=o,this.bracketPairNode=r}get openingBracketInfo(){return this.bracketPairNode.openingBracket.bracketInfo}}class HCe extends WCe{constructor(e,t,i,n,o,r,a){super(e,t,i,n,o,r),this.minVisibleColumnIndentation=a}}function VCe(s,e,t,i){return s!==t?$i(t-s,i):$i(0,i-e)}const Qs=0;function cI(s){return s===0}const vr=2**26;function $i(s,e){return s*vr+e}function ea(s){const e=s,t=Math.floor(e/vr),i=e-t*vr;return new cn(t,i)}function zCe(s){return Math.floor(s/vr)}function xi(s,e){let t=s+e;return e>=vr&&(t=t-s%vr),t}function $Ce(s,e){return s.reduce((t,i)=>xi(t,e(i)),Qs)}function $ee(s,e){return s===e}function xS(s,e){const t=s,i=e;if(i-t<=0)return Qs;const o=Math.floor(t/vr),r=Math.floor(i/vr),a=i-r*vr;if(o===r){const l=t-o*vr;return $i(0,a-l)}else return $i(r-o,a)}function bw(s,e){return s<e}function ww(s,e){return s<=e}function l1(s,e){return s>=e}function Ib(s){return $i(s.lineNumber-1,s.column-1)}function gp(s,e){const t=s,i=Math.floor(t/vr),n=t-i*vr,o=e,r=Math.floor(o/vr),a=o-r*vr;return new k(i+1,n+1,r+1,a+1)}function UCe(s){const e=qo(s);return $i(e.length-1,e[e.length-1].length)}class Ud{static fromModelContentChanges(e){return e.map(i=>{const n=k.lift(i.range);return new Ud(Ib(n.getStartPosition()),Ib(n.getEndPosition()),UCe(i.text))}).reverse()}constructor(e,t,i){this.startOffset=e,this.endOffset=t,this.newLength=i}toString(){return`[${ea(this.startOffset)}...${ea(this.endOffset)}) -> ${ea(this.newLength)}`}}class qCe{constructor(e){this.nextEditIdx=0,this.deltaOldToNewLineCount=0,this.deltaOldToNewColumnCount=0,this.deltaLineIdxInOld=-1,this.edits=e.map(t=>i7.from(t))}getOffsetBeforeChange(e){return this.adjustNextEdit(e),this.translateCurToOld(e)}getDistanceToNextChange(e){this.adjustNextEdit(e);const t=this.edits[this.nextEditIdx],i=t?this.translateOldToCur(t.offsetObj):null;return i===null?null:xS(e,i)}translateOldToCur(e){return e.lineCount===this.deltaLineIdxInOld?$i(e.lineCount+this.deltaOldToNewLineCount,e.columnCount+this.deltaOldToNewColumnCount):$i(e.lineCount+this.deltaOldToNewLineCount,e.columnCount)}translateCurToOld(e){const t=ea(e);return t.lineCount-this.deltaOldToNewLineCount===this.deltaLineIdxInOld?$i(t.lineCount-this.deltaOldToNewLineCount,t.columnCount-this.deltaOldToNewColumnCount):$i(t.lineCount-this.deltaOldToNewLineCount,t.columnCount)}adjustNextEdit(e){for(;this.nextEditIdx<this.edits.length;){const t=this.edits[this.nextEditIdx],i=this.translateOldToCur(t.endOffsetAfterObj);if(ww(i,e)){this.nextEditIdx++;const n=ea(i),o=ea(this.translateOldToCur(t.endOffsetBeforeObj)),r=n.lineCount-o.lineCount;this.deltaOldToNewLineCount+=r;const a=this.deltaLineIdxInOld===t.endOffsetBeforeObj.lineCount?this.deltaOldToNewColumnCount:0,l=n.columnCount-o.columnCount;this.deltaOldToNewColumnCount=a+l,this.deltaLineIdxInOld=t.endOffsetBeforeObj.lineCount}else break}}}class i7{static from(e){return new i7(e.startOffset,e.endOffset,e.newLength)}constructor(e,t,i){this.endOffsetBeforeObj=ea(t),this.endOffsetAfterObj=ea(xi(e,i)),this.offsetObj=ea(e)}}const nL=[],Or=class Or{static create(e,t){if(e<=128&&t.length===0){let i=Or.cache[e];return i||(i=new Or(e,t),Or.cache[e]=i),i}return new Or(e,t)}static getEmpty(){return this.empty}constructor(e,t){this.items=e,this.additionalItems=t}add(e,t){const i=t.getKey(e);let n=i>>5;if(n===0){const r=1<<i|this.items;return r===this.items?this:Or.create(r,this.additionalItems)}n--;const o=this.additionalItems.slice(0);for(;o.length<n;)o.push(0);return o[n]|=1<<(i&31),Or.create(this.items,o)}merge(e){const t=this.items|e.items;if(this.additionalItems===nL&&e.additionalItems===nL)return t===this.items?this:t===e.items?e:Or.create(t,nL);const i=[];for(let n=0;n<Math.max(this.additionalItems.length,e.additionalItems.length);n++){const o=this.additionalItems[n]||0,r=e.additionalItems[n]||0;i.push(o|r)}return Or.create(t,i)}intersects(e){if((this.items&e.items)!==0)return!0;for(let t=0;t<Math.min(this.additionalItems.length,e.additionalItems.length);t++)if((this.additionalItems[t]&e.additionalItems[t])!==0)return!0;return!1}};Or.cache=new Array(129),Or.empty=Or.create(0,nL);let Oo=Or;const pU={getKey(s){return s}};class Uee{constructor(){this.items=new Map}getKey(e){let t=this.items.get(e);return t===void 0&&(t=this.items.size,this.items.set(e,t)),t}}class n7{get length(){return this._length}constructor(e){this._length=e}}class LS extends n7{static create(e,t,i){let n=e.length;return t&&(n=xi(n,t.length)),i&&(n=xi(n,i.length)),new LS(n,e,t,i,t?t.missingOpeningBracketIds:Oo.getEmpty())}get kind(){return 2}get listHeight(){return 0}get childrenLength(){return 3}getChild(e){switch(e){case 0:return this.openingBracket;case 1:return this.child;case 2:return this.closingBracket}throw new Error("Invalid child index")}get children(){const e=[];return e.push(this.openingBracket),this.child&&e.push(this.child),this.closingBracket&&e.push(this.closingBracket),e}constructor(e,t,i,n,o){super(e),this.openingBracket=t,this.child=i,this.closingBracket=n,this.missingOpeningBracketIds=o}canBeReused(e){return!(this.closingBracket===null||e.intersects(this.missingOpeningBracketIds))}deepClone(){return new LS(this.length,this.openingBracket.deepClone(),this.child&&this.child.deepClone(),this.closingBracket&&this.closingBracket.deepClone(),this.missingOpeningBracketIds)}computeMinIndentation(e,t){return this.child?this.child.computeMinIndentation(xi(e,this.openingBracket.length),t):Number.MAX_SAFE_INTEGER}}class Ac extends n7{static create23(e,t,i,n=!1){let o=e.length,r=e.missingOpeningBracketIds;if(e.listHeight!==t.listHeight)throw new Error("Invalid list heights");if(o=xi(o,t.length),r=r.merge(t.missingOpeningBracketIds),i){if(e.listHeight!==i.listHeight)throw new Error("Invalid list heights");o=xi(o,i.length),r=r.merge(i.missingOpeningBracketIds)}return n?new jCe(o,e.listHeight+1,e,t,i,r):new kS(o,e.listHeight+1,e,t,i,r)}static getEmpty(){return new GCe(Qs,0,[],Oo.getEmpty())}get kind(){return 4}get missingOpeningBracketIds(){return this._missingOpeningBracketIds}constructor(e,t,i){super(e),this.listHeight=t,this._missingOpeningBracketIds=i,this.cachedMinIndentation=-1}throwIfImmutable(){}makeLastElementMutable(){this.throwIfImmutable();const e=this.childrenLength;if(e===0)return;const t=this.getChild(e-1),i=t.kind===4?t.toMutable():t;return t!==i&&this.setChild(e-1,i),i}makeFirstElementMutable(){if(this.throwIfImmutable(),this.childrenLength===0)return;const t=this.getChild(0),i=t.kind===4?t.toMutable():t;return t!==i&&this.setChild(0,i),i}canBeReused(e){if(e.intersects(this.missingOpeningBracketIds)||this.childrenLength===0)return!1;let t=this;for(;t.kind===4;){const i=t.childrenLength;if(i===0)throw new Ce;t=t.getChild(i-1)}return t.canBeReused(e)}handleChildrenChanged(){this.throwIfImmutable();const e=this.childrenLength;let t=this.getChild(0).length,i=this.getChild(0).missingOpeningBracketIds;for(let n=1;n<e;n++){const o=this.getChild(n);t=xi(t,o.length),i=i.merge(o.missingOpeningBracketIds)}this._length=t,this._missingOpeningBracketIds=i,this.cachedMinIndentation=-1}computeMinIndentation(e,t){if(this.cachedMinIndentation!==-1)return this.cachedMinIndentation;let i=Number.MAX_SAFE_INTEGER,n=e;for(let o=0;o<this.childrenLength;o++){const r=this.getChild(o);r&&(i=Math.min(i,r.computeMinIndentation(n,t)),n=xi(n,r.length))}return this.cachedMinIndentation=i,i}}class kS extends Ac{get childrenLength(){return this._item3!==null?3:2}getChild(e){switch(e){case 0:return this._item1;case 1:return this._item2;case 2:return this._item3}throw new Error("Invalid child index")}setChild(e,t){switch(e){case 0:this._item1=t;return;case 1:this._item2=t;return;case 2:this._item3=t;return}throw new Error("Invalid child index")}get children(){return this._item3?[this._item1,this._item2,this._item3]:[this._item1,this._item2]}get item1(){return this._item1}get item2(){return this._item2}get item3(){return this._item3}constructor(e,t,i,n,o,r){super(e,t,r),this._item1=i,this._item2=n,this._item3=o}deepClone(){return new kS(this.length,this.listHeight,this._item1.deepClone(),this._item2.deepClone(),this._item3?this._item3.deepClone():null,this.missingOpeningBracketIds)}appendChildOfSameHeight(e){if(this._item3)throw new Error("Cannot append to a full (2,3) tree node");this.throwIfImmutable(),this._item3=e,this.handleChildrenChanged()}unappendChild(){if(!this._item3)throw new Error("Cannot remove from a non-full (2,3) tree node");this.throwIfImmutable();const e=this._item3;return this._item3=null,this.handleChildrenChanged(),e}prependChildOfSameHeight(e){if(this._item3)throw new Error("Cannot prepend to a full (2,3) tree node");this.throwIfImmutable(),this._item3=this._item2,this._item2=this._item1,this._item1=e,this.handleChildrenChanged()}unprependChild(){if(!this._item3)throw new Error("Cannot remove from a non-full (2,3) tree node");this.throwIfImmutable();const e=this._item1;return this._item1=this._item2,this._item2=this._item3,this._item3=null,this.handleChildrenChanged(),e}toMutable(){return this}}class jCe extends kS{toMutable(){return new kS(this.length,this.listHeight,this.item1,this.item2,this.item3,this.missingOpeningBracketIds)}throwIfImmutable(){throw new Error("this instance is immutable")}}class dI extends Ac{get childrenLength(){return this._children.length}getChild(e){return this._children[e]}setChild(e,t){this._children[e]=t}get children(){return this._children}constructor(e,t,i,n){super(e,t,n),this._children=i}deepClone(){const e=new Array(this._children.length);for(let t=0;t<this._children.length;t++)e[t]=this._children[t].deepClone();return new dI(this.length,this.listHeight,e,this.missingOpeningBracketIds)}appendChildOfSameHeight(e){this.throwIfImmutable(),this._children.push(e),this.handleChildrenChanged()}unappendChild(){this.throwIfImmutable();const e=this._children.pop();return this.handleChildrenChanged(),e}prependChildOfSameHeight(e){this.throwIfImmutable(),this._children.unshift(e),this.handleChildrenChanged()}unprependChild(){this.throwIfImmutable();const e=this._children.shift();return this.handleChildrenChanged(),e}toMutable(){return this}}class GCe extends dI{toMutable(){return new dI(this.length,this.listHeight,[...this.children],this.missingOpeningBracketIds)}throwIfImmutable(){throw new Error("this instance is immutable")}}const KCe=[];class s7 extends n7{get listHeight(){return 0}get childrenLength(){return 0}getChild(e){return null}get children(){return KCe}deepClone(){return this}}class _f extends s7{get kind(){return 0}get missingOpeningBracketIds(){return Oo.getEmpty()}canBeReused(e){return!0}computeMinIndentation(e,t){const i=ea(e),n=(i.columnCount===0?i.lineCount:i.lineCount+1)+1,o=zCe(xi(e,this.length))+1;let r=Number.MAX_SAFE_INTEGER;for(let a=n;a<=o;a++){const l=t.getLineFirstNonWhitespaceColumn(a),c=t.getLineContent(a);if(l===0)continue;const d=pi.visibleColumnFromColumn(c,l,t.getOptions().tabSize);r=Math.min(r,d)}return r}}class hI extends s7{static create(e,t,i){return new hI(e,t,i)}get kind(){return 1}get missingOpeningBracketIds(){return Oo.getEmpty()}constructor(e,t,i){super(e),this.bracketInfo=t,this.bracketIds=i}get text(){return this.bracketInfo.bracketText}get languageId(){return this.bracketInfo.languageId}canBeReused(e){return!1}computeMinIndentation(e,t){return Number.MAX_SAFE_INTEGER}}class YCe extends s7{get kind(){return 3}constructor(e,t){super(t),this.missingOpeningBracketIds=e}canBeReused(e){return!e.intersects(this.missingOpeningBracketIds)}computeMinIndentation(e,t){return Number.MAX_SAFE_INTEGER}}let au=class{constructor(e,t,i,n,o){this.length=e,this.kind=t,this.bracketId=i,this.bracketIds=n,this.astNode=o}};class qee{constructor(e,t){this.textModel=e,this.bracketTokens=t,this.reader=new ZCe(this.textModel,this.bracketTokens),this._offset=Qs,this.didPeek=!1,this.peeked=null,this.textBufferLineCount=e.getLineCount(),this.textBufferLastLineLength=e.getLineLength(this.textBufferLineCount)}get offset(){return this._offset}get length(){return $i(this.textBufferLineCount-1,this.textBufferLastLineLength)}skip(e){this.didPeek=!1,this._offset=xi(this._offset,e);const t=ea(this._offset);this.reader.setPosition(t.lineCount,t.columnCount)}read(){let e;return this.peeked?(this.didPeek=!1,e=this.peeked):e=this.reader.read(),e&&(this._offset=xi(this._offset,e.length)),e}peek(){return this.didPeek||(this.peeked=this.reader.read(),this.didPeek=!0),this.peeked}}class ZCe{constructor(e,t){this.textModel=e,this.bracketTokens=t,this.lineIdx=0,this.line=null,this.lineCharOffset=0,this.lineTokens=null,this.lineTokenOffset=0,this.peekedToken=null,this.textBufferLineCount=e.getLineCount(),this.textBufferLastLineLength=e.getLineLength(this.textBufferLineCount)}setPosition(e,t){e===this.lineIdx?(this.lineCharOffset=t,this.line!==null&&(this.lineTokenOffset=this.lineCharOffset===0?0:this.lineTokens.findTokenIndexAtOffset(this.lineCharOffset))):(this.lineIdx=e,this.lineCharOffset=t,this.line=null),this.peekedToken=null}read(){if(this.peekedToken){const o=this.peekedToken;return this.peekedToken=null,this.lineCharOffset+=o.length,o}if(this.lineIdx>this.textBufferLineCount-1||this.lineIdx===this.textBufferLineCount-1&&this.lineCharOffset>=this.textBufferLastLineLength)return null;this.line===null&&(this.lineTokens=this.textModel.tokenization.getLineTokens(this.lineIdx+1),this.line=this.lineTokens.getLineContent(),this.lineTokenOffset=this.lineCharOffset===0?0:this.lineTokens.findTokenIndexAtOffset(this.lineCharOffset));const e=this.lineIdx,t=this.lineCharOffset;let i=0;for(;;){const o=this.lineTokens,r=o.getCount();let a=null;if(this.lineTokenOffset<r){const l=o.getMetadata(this.lineTokenOffset);for(;this.lineTokenOffset+1<r&&l===o.getMetadata(this.lineTokenOffset+1);)this.lineTokenOffset++;const c=ds.getTokenType(l)===0,d=ds.containsBalancedBrackets(l),h=o.getEndOffset(this.lineTokenOffset);if(d&&c&&this.lineCharOffset<h){const u=o.getLanguageId(this.lineTokenOffset),g=this.line.substring(this.lineCharOffset,h),f=this.bracketTokens.getSingleLanguageBracketTokens(u),p=f.regExpGlobal;if(p){p.lastIndex=0;const _=p.exec(g);_&&(a=f.getToken(_[0]),a&&(this.lineCharOffset+=_.index))}}if(i+=h-this.lineCharOffset,a)if(e!==this.lineIdx||t!==this.lineCharOffset){this.peekedToken=a;break}else return this.lineCharOffset+=a.length,a;else this.lineTokenOffset++,this.lineCharOffset=h}else if(this.lineIdx===this.textBufferLineCount-1||(this.lineIdx++,this.lineTokens=this.textModel.tokenization.getLineTokens(this.lineIdx+1),this.lineTokenOffset=0,this.line=this.lineTokens.getLineContent(),this.lineCharOffset=0,i+=33,i>1e3))break;if(i>1500)break}const n=VCe(e,t,this.lineIdx,this.lineCharOffset);return new au(n,0,-1,Oo.getEmpty(),new _f(n))}}class XCe{constructor(e,t){this.text=e,this._offset=Qs,this.idx=0;const i=t.getRegExpStr(),n=i?new RegExp(i+`|
`,"gi"):null,o=[];let r,a=0,l=0,c=0,d=0;const h=[];for(let f=0;f<60;f++)h.push(new au($i(0,f),0,-1,Oo.getEmpty(),new _f($i(0,f))));const u=[];for(let f=0;f<60;f++)u.push(new au($i(1,f),0,-1,Oo.getEmpty(),new _f($i(1,f))));if(n)for(n.lastIndex=0;(r=n.exec(e))!==null;){const f=r.index,p=r[0];if(p===`
`)a++,l=f+1;else{if(c!==f){let _;if(d===a){const b=f-c;if(b<h.length)_=h[b];else{const w=$i(0,b);_=new au(w,0,-1,Oo.getEmpty(),new _f(w))}}else{const b=a-d,w=f-l;if(b===1&&w<u.length)_=u[w];else{const C=$i(b,w);_=new au(C,0,-1,Oo.getEmpty(),new _f(C))}}o.push(_)}o.push(t.getToken(p)),c=f+p.length,d=a}}const g=e.length;if(c!==g){const f=d===a?$i(0,g-c):$i(a-d,g-l);o.push(new au(f,0,-1,Oo.getEmpty(),new _f(f)))}this.length=$i(a,g-l),this.tokens=o}get offset(){return this._offset}read(){return this.tokens[this.idx++]||null}peek(){return this.tokens[this.idx]||null}skip(e){throw new Hce}}class o7{static createFromLanguage(e,t){function i(o){return t.getKey(`${o.languageId}:::${o.bracketText}`)}const n=new Map;for(const o of e.bracketsNew.openingBrackets){const r=$i(0,o.bracketText.length),a=i(o),l=Oo.getEmpty().add(a,pU);n.set(o.bracketText,new au(r,1,a,l,hI.create(r,o,l)))}for(const o of e.bracketsNew.closingBrackets){const r=$i(0,o.bracketText.length);let a=Oo.getEmpty();const l=o.getOpeningBrackets();for(const c of l)a=a.add(i(c),pU);n.set(o.bracketText,new au(r,2,i(l[0]),a,hI.create(r,o,a)))}return new o7(n)}constructor(e){this.map=e,this.hasRegExp=!1,this._regExpGlobal=null}getRegExpStr(){if(this.isEmpty)return null;{const e=[...this.map.keys()];return e.sort(),e.reverse(),e.map(t=>QCe(t)).join("|")}}get regExpGlobal(){if(!this.hasRegExp){const e=this.getRegExpStr();this._regExpGlobal=e?new RegExp(e,"gi"):null,this.hasRegExp=!0}return this._regExpGlobal}getToken(e){return this.map.get(e.toLowerCase())}findClosingTokenText(e){for(const[t,i]of this.map)if(i.kind===2&&i.bracketIds.intersects(e))return t}get isEmpty(){return this.map.size===0}}function QCe(s){let e=Uo(s);return/^[\w ]+/.test(s)&&(e=`\\b${e}`),/[\w ]+$/.test(s)&&(e=`${e}\\b`),e}class jee{constructor(e,t){this.denseKeyProvider=e,this.getLanguageConfiguration=t,this.languageIdToBracketTokens=new Map}didLanguageChange(e){return this.languageIdToBracketTokens.has(e)}getSingleLanguageBracketTokens(e){let t=this.languageIdToBracketTokens.get(e);return t||(t=o7.createFromLanguage(this.getLanguageConfiguration(e),this.denseKeyProvider),this.languageIdToBracketTokens.set(e,t)),t}}function JCe(s){if(s.length===0)return null;if(s.length===1)return s[0];let e=0;function t(){if(e>=s.length)return null;const r=e,a=s[r].listHeight;for(e++;e<s.length&&s[e].listHeight===a;)e++;return e-r>=2?Gee(r===0&&e===s.length?s:s.slice(r,e),!1):s[r]}let i=t(),n=t();if(!n)return i;for(let r=t();r;r=t())mU(i,n)<=mU(n,r)?(i=hM(i,n),n=r):n=hM(n,r);return hM(i,n)}function Gee(s,e=!1){if(s.length===0)return null;if(s.length===1)return s[0];let t=s.length;for(;t>3;){const i=t>>1;for(let n=0;n<i;n++){const o=n<<1;s[n]=Ac.create23(s[o],s[o+1],o+3===t?s[o+2]:null,e)}t=i}return Ac.create23(s[0],s[1],t>=3?s[2]:null,e)}function mU(s,e){return Math.abs(s.listHeight-e.listHeight)}function hM(s,e){return s.listHeight===e.listHeight?Ac.create23(s,e,null,!1):s.listHeight>e.listHeight?e1e(s,e):t1e(e,s)}function e1e(s,e){s=s.toMutable();let t=s;const i=[];let n;for(;;){if(e.listHeight===t.listHeight){n=e;break}if(t.kind!==4)throw new Error("unexpected");i.push(t),t=t.makeLastElementMutable()}for(let o=i.length-1;o>=0;o--){const r=i[o];n?r.childrenLength>=3?n=Ac.create23(r.unappendChild(),n,null,!1):(r.appendChildOfSameHeight(n),n=void 0):r.handleChildrenChanged()}return n?Ac.create23(s,n,null,!1):s}function t1e(s,e){s=s.toMutable();let t=s;const i=[];for(;e.listHeight!==t.listHeight;){if(t.kind!==4)throw new Error("unexpected");i.push(t),t=t.makeFirstElementMutable()}let n=e;for(let o=i.length-1;o>=0;o--){const r=i[o];n?r.childrenLength>=3?n=Ac.create23(n,r.unprependChild(),null,!1):(r.prependChildOfSameHeight(n),n=void 0):r.handleChildrenChanged()}return n?Ac.create23(n,s,null,!1):s}class i1e{constructor(e){this.lastOffset=Qs,this.nextNodes=[e],this.offsets=[Qs],this.idxs=[]}readLongestNodeAt(e,t){if(bw(e,this.lastOffset))throw new Error("Invalid offset");for(this.lastOffset=e;;){const i=TC(this.nextNodes);if(!i)return;const n=TC(this.offsets);if(bw(e,n))return;if(bw(n,e))if(xi(n,i.length)<=e)this.nextNodeAfterCurrent();else{const o=uM(i);o!==-1?(this.nextNodes.push(i.getChild(o)),this.offsets.push(n),this.idxs.push(o)):this.nextNodeAfterCurrent()}else{if(t(i))return this.nextNodeAfterCurrent(),i;{const o=uM(i);if(o===-1){this.nextNodeAfterCurrent();return}else this.nextNodes.push(i.getChild(o)),this.offsets.push(n),this.idxs.push(o)}}}}nextNodeAfterCurrent(){for(;;){const e=TC(this.offsets),t=TC(this.nextNodes);if(this.nextNodes.pop(),this.offsets.pop(),this.idxs.length===0)break;const i=TC(this.nextNodes),n=uM(i,this.idxs[this.idxs.length-1]);if(n!==-1){this.nextNodes.push(i.getChild(n)),this.offsets.push(xi(e,t.length)),this.idxs[this.idxs.length-1]=n;break}else this.idxs.pop()}}}function uM(s,e=-1){for(;;){if(e++,e>=s.childrenLength)return-1;if(s.getChild(e))return e}}function TC(s){return s.length>0?s[s.length-1]:void 0}function DF(s,e,t,i){return new n1e(s,e,t,i).parseDocument()}class n1e{constructor(e,t,i,n){if(this.tokenizer=e,this.createImmutableLists=n,this._itemsConstructed=0,this._itemsFromCache=0,i&&n)throw new Error("Not supported");this.oldNodeReader=i?new i1e(i):void 0,this.positionMapper=new qCe(t)}parseDocument(){this._itemsConstructed=0,this._itemsFromCache=0;let e=this.parseList(Oo.getEmpty(),0);return e||(e=Ac.getEmpty()),e}parseList(e,t){const i=[];for(;;){let o=this.tryReadChildFromCache(e);if(!o){const r=this.tokenizer.peek();if(!r||r.kind===2&&r.bracketIds.intersects(e))break;o=this.parseChild(e,t+1)}o.kind===4&&o.childrenLength===0||i.push(o)}return this.oldNodeReader?JCe(i):Gee(i,this.createImmutableLists)}tryReadChildFromCache(e){if(this.oldNodeReader){const t=this.positionMapper.getDistanceToNextChange(this.tokenizer.offset);if(t===null||!cI(t)){const i=this.oldNodeReader.readLongestNodeAt(this.positionMapper.getOffsetBeforeChange(this.tokenizer.offset),n=>t!==null&&!bw(n.length,t)?!1:n.canBeReused(e));if(i)return this._itemsFromCache++,this.tokenizer.skip(i.length),i}}}parseChild(e,t){this._itemsConstructed++;const i=this.tokenizer.read();switch(i.kind){case 2:return new YCe(i.bracketIds,i.length);case 0:return i.astNode;case 1:{if(t>300)return new _f(i.length);const n=e.merge(i.bracketIds),o=this.parseList(n,t+1),r=this.tokenizer.peek();return r&&r.kind===2&&(r.bracketId===i.bracketId||r.bracketIds.intersects(i.bracketIds))?(this.tokenizer.read(),LS.create(i.astNode,o,r.astNode)):LS.create(i.astNode,o,null)}default:throw new Error("unexpected")}}}function uI(s,e){if(s.length===0)return e;if(e.length===0)return s;const t=new ph(_U(s)),i=_U(e);i.push({modified:!1,lengthBefore:void 0,lengthAfter:void 0});let n=t.dequeue();function o(c){if(c===void 0){const h=t.takeWhile(u=>!0)||[];return n&&h.unshift(n),h}const d=[];for(;n&&!cI(c);){const[h,u]=n.splitAt(c);d.push(h),c=xS(h.lengthAfter,c),n=u??t.dequeue()}return cI(c)||d.push(new yu(!1,c,c)),d}const r=[];function a(c,d,h){if(r.length>0&&$ee(r[r.length-1].endOffset,c)){const u=r[r.length-1];r[r.length-1]=new Ud(u.startOffset,d,xi(u.newLength,h))}else r.push({startOffset:c,endOffset:d,newLength:h})}let l=Qs;for(const c of i){const d=o(c.lengthBefore);if(c.modified){const h=$Ce(d,g=>g.lengthBefore),u=xi(l,h);a(l,u,c.lengthAfter),l=u}else for(const h of d){const u=l;l=xi(l,h.lengthBefore),h.modified&&a(u,l,h.lengthAfter)}}return r}class yu{constructor(e,t,i){this.modified=e,this.lengthBefore=t,this.lengthAfter=i}splitAt(e){const t=xS(e,this.lengthAfter);return $ee(t,Qs)?[this,void 0]:this.modified?[new yu(this.modified,this.lengthBefore,e),new yu(this.modified,Qs,t)]:[new yu(this.modified,e,e),new yu(this.modified,t,t)]}toString(){return`${this.modified?"M":"U"}:${ea(this.lengthBefore)} -> ${ea(this.lengthAfter)}`}}function _U(s){const e=[];let t=Qs;for(const i of s){const n=xS(t,i.startOffset);cI(n)||e.push(new yu(!1,n,n));const o=xS(i.startOffset,i.endOffset);e.push(new yu(!0,o,i.newLength)),t=i.endOffset}return e}class s1e extends F{didLanguageChange(e){return this.brackets.didLanguageChange(e)}constructor(e,t){if(super(),this.textModel=e,this.getLanguageConfiguration=t,this.didChangeEmitter=new M,this.denseKeyProvider=new Uee,this.brackets=new jee(this.denseKeyProvider,this.getLanguageConfiguration),this.onDidChange=this.didChangeEmitter.event,this.queuedTextEditsForInitialAstWithoutTokens=[],this.queuedTextEdits=[],e.tokenization.hasTokens)e.tokenization.backgroundTokenizationState===2?(this.initialAstWithoutTokens=void 0,this.astWithTokens=this.parseDocumentFromTextBuffer([],void 0,!1)):(this.initialAstWithoutTokens=this.parseDocumentFromTextBuffer([],void 0,!0),this.astWithTokens=this.initialAstWithoutTokens);else{const i=this.brackets.getSingleLanguageBracketTokens(this.textModel.getLanguageId()),n=new XCe(this.textModel.getValue(),i);this.initialAstWithoutTokens=DF(n,[],void 0,!0),this.astWithTokens=this.initialAstWithoutTokens}}handleDidChangeBackgroundTokenizationState(){if(this.textModel.tokenization.backgroundTokenizationState===2){const e=this.initialAstWithoutTokens===void 0;this.initialAstWithoutTokens=void 0,e||this.didChangeEmitter.fire()}}handleDidChangeTokens({ranges:e}){const t=e.map(i=>new Ud($i(i.fromLineNumber-1,0),$i(i.toLineNumber,0),$i(i.toLineNumber-i.fromLineNumber+1,0)));this.handleEdits(t,!0),this.initialAstWithoutTokens||this.didChangeEmitter.fire()}handleContentChanged(e){const t=Ud.fromModelContentChanges(e.changes);this.handleEdits(t,!1)}handleEdits(e,t){const i=uI(this.queuedTextEdits,e);this.queuedTextEdits=i,this.initialAstWithoutTokens&&!t&&(this.queuedTextEditsForInitialAstWithoutTokens=uI(this.queuedTextEditsForInitialAstWithoutTokens,e))}flushQueue(){this.queuedTextEdits.length>0&&(this.astWithTokens=this.parseDocumentFromTextBuffer(this.queuedTextEdits,this.astWithTokens,!1),this.queuedTextEdits=[]),this.queuedTextEditsForInitialAstWithoutTokens.length>0&&(this.initialAstWithoutTokens&&(this.initialAstWithoutTokens=this.parseDocumentFromTextBuffer(this.queuedTextEditsForInitialAstWithoutTokens,this.initialAstWithoutTokens,!1)),this.queuedTextEditsForInitialAstWithoutTokens=[])}parseDocumentFromTextBuffer(e,t,i){const n=t,o=new qee(this.textModel,this.brackets);return DF(o,e,n,i)}getBracketsInRange(e,t){this.flushQueue();const i=$i(e.startLineNumber-1,e.startColumn-1),n=$i(e.endLineNumber-1,e.endColumn-1);return new op(o=>{const r=this.initialAstWithoutTokens||this.astWithTokens;TF(r,Qs,r.length,i,n,o,0,0,new Map,t)})}getBracketPairsInRange(e,t){this.flushQueue();const i=Ib(e.getStartPosition()),n=Ib(e.getEndPosition());return new op(o=>{const r=this.initialAstWithoutTokens||this.astWithTokens,a=new o1e(o,t,this.textModel);NF(r,Qs,r.length,i,n,a,0,new Map)})}getFirstBracketAfter(e){this.flushQueue();const t=this.initialAstWithoutTokens||this.astWithTokens;return Yee(t,Qs,t.length,Ib(e))}getFirstBracketBefore(e){this.flushQueue();const t=this.initialAstWithoutTokens||this.astWithTokens;return Kee(t,Qs,t.length,Ib(e))}}function Kee(s,e,t,i){if(s.kind===4||s.kind===2){const n=[];for(const o of s.children)t=xi(e,o.length),n.push({nodeOffsetStart:e,nodeOffsetEnd:t}),e=t;for(let o=n.length-1;o>=0;o--){const{nodeOffsetStart:r,nodeOffsetEnd:a}=n[o];if(bw(r,i)){const l=Kee(s.children[o],r,a,i);if(l)return l}}return null}else{if(s.kind===3)return null;if(s.kind===1){const n=gp(e,t);return{bracketInfo:s.bracketInfo,range:n}}}return null}function Yee(s,e,t,i){if(s.kind===4||s.kind===2){for(const n of s.children){if(t=xi(e,n.length),bw(i,t)){const o=Yee(n,e,t,i);if(o)return o}e=t}return null}else{if(s.kind===3)return null;if(s.kind===1){const n=gp(e,t);return{bracketInfo:s.bracketInfo,range:n}}}return null}function TF(s,e,t,i,n,o,r,a,l,c,d=!1){if(r>200)return!0;e:for(;;)switch(s.kind){case 4:{const h=s.childrenLength;for(let u=0;u<h;u++){const g=s.getChild(u);if(g){if(t=xi(e,g.length),ww(e,n)&&l1(t,i)){if(l1(t,n)){s=g;continue e}if(!TF(g,e,t,i,n,o,r,0,l,c))return!1}e=t}}return!0}case 2:{const h=!c||!s.closingBracket||s.closingBracket.bracketInfo.closesColorized(s.openingBracket.bracketInfo);let u=0;if(l){let f=l.get(s.openingBracket.text);f===void 0&&(f=0),u=f,h&&(f++,l.set(s.openingBracket.text,f))}const g=s.childrenLength;for(let f=0;f<g;f++){const p=s.getChild(f);if(p){if(t=xi(e,p.length),ww(e,n)&&l1(t,i)){if(l1(t,n)&&p.kind!==1){s=p,h?(r++,a=u+1):a=u;continue e}if((h||p.kind!==1||!s.closingBracket)&&!TF(p,e,t,i,n,o,h?r+1:r,h?u+1:u,l,c,!s.closingBracket))return!1}e=t}}return l?.set(s.openingBracket.text,u),!0}case 3:{const h=gp(e,t);return o(new fU(h,r-1,0,!0))}case 1:{const h=gp(e,t);return o(new fU(h,r-1,a-1,d))}case 0:return!0}}class o1e{constructor(e,t,i){this.push=e,this.includeMinIndentation=t,this.textModel=i}}function NF(s,e,t,i,n,o,r,a){if(r>200)return!0;let l=!0;if(s.kind===2){let c=0;if(a){let u=a.get(s.openingBracket.text);u===void 0&&(u=0),c=u,u++,a.set(s.openingBracket.text,u)}const d=xi(e,s.openingBracket.length);let h=-1;if(o.includeMinIndentation&&(h=s.computeMinIndentation(e,o.textModel)),l=o.push(new HCe(gp(e,t),gp(e,d),s.closingBracket?gp(xi(d,s.child?.length||Qs),t):void 0,r,c,s,h)),e=d,l&&s.child){const u=s.child;if(t=xi(e,u.length),ww(e,n)&&l1(t,i)&&(l=NF(u,e,t,i,n,o,r+1,a),!l))return!1}a?.set(s.openingBracket.text,c)}else{let c=e;for(const d of s.children){const h=c;if(c=xi(c,d.length),ww(h,n)&&ww(i,c)&&(l=NF(d,h,c,i,n,o,r,a),!l))return!1}}return l}class r1e extends F{get canBuildAST(){return this.textModel.getValueLength()<=5e6}constructor(e,t){super(),this.textModel=e,this.languageConfigurationService=t,this.bracketPairsTree=this._register(new vt),this.onDidChangeEmitter=new M,this.onDidChange=this.onDidChangeEmitter.event,this.bracketsRequested=!1}handleLanguageConfigurationServiceChange(e){(!e.languageId||this.bracketPairsTree.value?.object.didLanguageChange(e.languageId))&&(this.bracketPairsTree.clear(),this.updateBracketPairsTree())}handleDidChangeOptions(e){this.bracketPairsTree.clear(),this.updateBracketPairsTree()}handleDidChangeLanguage(e){this.bracketPairsTree.clear(),this.updateBracketPairsTree()}handleDidChangeContent(e){this.bracketPairsTree.value?.object.handleContentChanged(e)}handleDidChangeBackgroundTokenizationState(){this.bracketPairsTree.value?.object.handleDidChangeBackgroundTokenizationState()}handleDidChangeTokens(e){this.bracketPairsTree.value?.object.handleDidChangeTokens(e)}updateBracketPairsTree(){if(this.bracketsRequested&&this.canBuildAST){if(!this.bracketPairsTree.value){const e=new U;this.bracketPairsTree.value=a1e(e.add(new s1e(this.textModel,t=>this.languageConfigurationService.getLanguageConfiguration(t))),e),e.add(this.bracketPairsTree.value.object.onDidChange(t=>this.onDidChangeEmitter.fire(t))),this.onDidChangeEmitter.fire()}}else this.bracketPairsTree.value&&(this.bracketPairsTree.clear(),this.onDidChangeEmitter.fire())}getBracketPairsInRange(e){return this.bracketsRequested=!0,this.updateBracketPairsTree(),this.bracketPairsTree.value?.object.getBracketPairsInRange(e,!1)||op.empty}getBracketPairsInRangeWithMinIndentation(e){return this.bracketsRequested=!0,this.updateBracketPairsTree(),this.bracketPairsTree.value?.object.getBracketPairsInRange(e,!0)||op.empty}getBracketsInRange(e,t=!1){return this.bracketsRequested=!0,this.updateBracketPairsTree(),this.bracketPairsTree.value?.object.getBracketsInRange(e,t)||op.empty}findMatchingBracketUp(e,t,i){const n=this.textModel.validatePosition(t),o=this.textModel.getLanguageIdAtPosition(n.lineNumber,n.column);if(this.canBuildAST){const r=this.languageConfigurationService.getLanguageConfiguration(o).bracketsNew.getClosingBracketInfo(e);if(!r)return null;const a=this.getBracketPairsInRange(k.fromPositions(t,t)).findLast(l=>r.closes(l.openingBracketInfo));return a?a.openingBracketRange:null}else{const r=e.toLowerCase(),a=this.languageConfigurationService.getLanguageConfiguration(o).brackets;if(!a)return null;const l=a.textIsBracket[r];return l?sL(this._findMatchingBracketUp(l,n,gM(i))):null}}matchBracket(e,t){if(this.canBuildAST){const i=this.getBracketPairsInRange(k.fromPositions(e,e)).filter(n=>n.closingBracketRange!==void 0&&(n.openingBracketRange.containsPosition(e)||n.closingBracketRange.containsPosition(e))).findLastMaxBy(ns(n=>n.openingBracketRange.containsPosition(e)?n.openingBracketRange:n.closingBracketRange,k.compareRangesUsingStarts));return i?[i.openingBracketRange,i.closingBracketRange]:null}else{const i=gM(t);return this._matchBracket(this.textModel.validatePosition(e),i)}}_establishBracketSearchOffsets(e,t,i,n){const o=t.getCount(),r=t.getLanguageId(n);let a=Math.max(0,e.column-1-i.maxBracketLength);for(let c=n-1;c>=0;c--){const d=t.getEndOffset(c);if(d<=a)break;if(Ul(t.getStandardTokenType(c))||t.getLanguageId(c)!==r){a=d;break}}let l=Math.min(t.getLineContent().length,e.column-1+i.maxBracketLength);for(let c=n+1;c<o;c++){const d=t.getStartOffset(c);if(d>=l)break;if(Ul(t.getStandardTokenType(c))||t.getLanguageId(c)!==r){l=d;break}}return{searchStartOffset:a,searchEndOffset:l}}_matchBracket(e,t){const i=e.lineNumber,n=this.textModel.tokenization.getLineTokens(i),o=this.textModel.getLineContent(i),r=n.findTokenIndexAtOffset(e.column-1);if(r<0)return null;const a=this.languageConfigurationService.getLanguageConfiguration(n.getLanguageId(r)).brackets;if(a&&!Ul(n.getStandardTokenType(r))){let{searchStartOffset:l,searchEndOffset:c}=this._establishBracketSearchOffsets(e,n,a,r),d=null;for(;;){const h=_a.findNextBracketInRange(a.forwardRegex,i,o,l,c);if(!h)break;if(h.startColumn<=e.column&&e.column<=h.endColumn){const u=o.substring(h.startColumn-1,h.endColumn-1).toLowerCase(),g=this._matchFoundBracket(h,a.textIsBracket[u],a.textIsOpenBracket[u],t);if(g){if(g instanceof Td)return null;d=g}}l=h.endColumn-1}if(d)return d}if(r>0&&n.getStartOffset(r)===e.column-1){const l=r-1,c=this.languageConfigurationService.getLanguageConfiguration(n.getLanguageId(l)).brackets;if(c&&!Ul(n.getStandardTokenType(l))){const{searchStartOffset:d,searchEndOffset:h}=this._establishBracketSearchOffsets(e,n,c,l),u=_a.findPrevBracketInRange(c.reversedRegex,i,o,d,h);if(u&&u.startColumn<=e.column&&e.column<=u.endColumn){const g=o.substring(u.startColumn-1,u.endColumn-1).toLowerCase(),f=this._matchFoundBracket(u,c.textIsBracket[g],c.textIsOpenBracket[g],t);if(f)return f instanceof Td?null:f}}}return null}_matchFoundBracket(e,t,i,n){if(!t)return null;const o=i?this._findMatchingBracketDown(t,e.getEndPosition(),n):this._findMatchingBracketUp(t,e.getStartPosition(),n);return o?o instanceof Td?o:[e,o]:null}_findMatchingBracketUp(e,t,i){const n=e.languageId,o=e.reversedRegex;let r=-1,a=0;const l=(c,d,h,u)=>{for(;;){if(i&&++a%100===0&&!i())return Td.INSTANCE;const g=_a.findPrevBracketInRange(o,c,d,h,u);if(!g)break;const f=d.substring(g.startColumn-1,g.endColumn-1).toLowerCase();if(e.isOpen(f)?r++:e.isClose(f)&&r--,r===0)return g;u=g.startColumn-1}return null};for(let c=t.lineNumber;c>=1;c--){const d=this.textModel.tokenization.getLineTokens(c),h=d.getCount(),u=this.textModel.getLineContent(c);let g=h-1,f=u.length,p=u.length;c===t.lineNumber&&(g=d.findTokenIndexAtOffset(t.column-1),f=t.column-1,p=t.column-1);let _=!0;for(;g>=0;g--){const b=d.getLanguageId(g)===n&&!Ul(d.getStandardTokenType(g));if(b)_?f=d.getStartOffset(g):(f=d.getStartOffset(g),p=d.getEndOffset(g));else if(_&&f!==p){const w=l(c,u,f,p);if(w)return w}_=b}if(_&&f!==p){const b=l(c,u,f,p);if(b)return b}}return null}_findMatchingBracketDown(e,t,i){const n=e.languageId,o=e.forwardRegex;let r=1,a=0;const l=(d,h,u,g)=>{for(;;){if(i&&++a%100===0&&!i())return Td.INSTANCE;const f=_a.findNextBracketInRange(o,d,h,u,g);if(!f)break;const p=h.substring(f.startColumn-1,f.endColumn-1).toLowerCase();if(e.isOpen(p)?r++:e.isClose(p)&&r--,r===0)return f;u=f.endColumn-1}return null},c=this.textModel.getLineCount();for(let d=t.lineNumber;d<=c;d++){const h=this.textModel.tokenization.getLineTokens(d),u=h.getCount(),g=this.textModel.getLineContent(d);let f=0,p=0,_=0;d===t.lineNumber&&(f=h.findTokenIndexAtOffset(t.column-1),p=t.column-1,_=t.column-1);let b=!0;for(;f<u;f++){const w=h.getLanguageId(f)===n&&!Ul(h.getStandardTokenType(f));if(w)b||(p=h.getStartOffset(f)),_=h.getEndOffset(f);else if(b&&p!==_){const C=l(d,g,p,_);if(C)return C}b=w}if(b&&p!==_){const w=l(d,g,p,_);if(w)return w}}return null}findPrevBracket(e){const t=this.textModel.validatePosition(e);if(this.canBuildAST)return this.bracketsRequested=!0,this.updateBracketPairsTree(),this.bracketPairsTree.value?.object.getFirstBracketBefore(t)||null;let i=null,n=null,o=null;for(let r=t.lineNumber;r>=1;r--){const a=this.textModel.tokenization.getLineTokens(r),l=a.getCount(),c=this.textModel.getLineContent(r);let d=l-1,h=c.length,u=c.length;if(r===t.lineNumber){d=a.findTokenIndexAtOffset(t.column-1),h=t.column-1,u=t.column-1;const f=a.getLanguageId(d);i!==f&&(i=f,n=this.languageConfigurationService.getLanguageConfiguration(i).brackets,o=this.languageConfigurationService.getLanguageConfiguration(i).bracketsNew)}let g=!0;for(;d>=0;d--){const f=a.getLanguageId(d);if(i!==f){if(n&&o&&g&&h!==u){const _=_a.findPrevBracketInRange(n.reversedRegex,r,c,h,u);if(_)return this._toFoundBracket(o,_);g=!1}i=f,n=this.languageConfigurationService.getLanguageConfiguration(i).brackets,o=this.languageConfigurationService.getLanguageConfiguration(i).bracketsNew}const p=!!n&&!Ul(a.getStandardTokenType(d));if(p)g?h=a.getStartOffset(d):(h=a.getStartOffset(d),u=a.getEndOffset(d));else if(o&&n&&g&&h!==u){const _=_a.findPrevBracketInRange(n.reversedRegex,r,c,h,u);if(_)return this._toFoundBracket(o,_)}g=p}if(o&&n&&g&&h!==u){const f=_a.findPrevBracketInRange(n.reversedRegex,r,c,h,u);if(f)return this._toFoundBracket(o,f)}}return null}findNextBracket(e){const t=this.textModel.validatePosition(e);if(this.canBuildAST)return this.bracketsRequested=!0,this.updateBracketPairsTree(),this.bracketPairsTree.value?.object.getFirstBracketAfter(t)||null;const i=this.textModel.getLineCount();let n=null,o=null,r=null;for(let a=t.lineNumber;a<=i;a++){const l=this.textModel.tokenization.getLineTokens(a),c=l.getCount(),d=this.textModel.getLineContent(a);let h=0,u=0,g=0;if(a===t.lineNumber){h=l.findTokenIndexAtOffset(t.column-1),u=t.column-1,g=t.column-1;const p=l.getLanguageId(h);n!==p&&(n=p,o=this.languageConfigurationService.getLanguageConfiguration(n).brackets,r=this.languageConfigurationService.getLanguageConfiguration(n).bracketsNew)}let f=!0;for(;h<c;h++){const p=l.getLanguageId(h);if(n!==p){if(r&&o&&f&&u!==g){const b=_a.findNextBracketInRange(o.forwardRegex,a,d,u,g);if(b)return this._toFoundBracket(r,b);f=!1}n=p,o=this.languageConfigurationService.getLanguageConfiguration(n).brackets,r=this.languageConfigurationService.getLanguageConfiguration(n).bracketsNew}const _=!!o&&!Ul(l.getStandardTokenType(h));if(_)f||(u=l.getStartOffset(h)),g=l.getEndOffset(h);else if(r&&o&&f&&u!==g){const b=_a.findNextBracketInRange(o.forwardRegex,a,d,u,g);if(b)return this._toFoundBracket(r,b)}f=_}if(r&&o&&f&&u!==g){const p=_a.findNextBracketInRange(o.forwardRegex,a,d,u,g);if(p)return this._toFoundBracket(r,p)}}return null}findEnclosingBrackets(e,t){const i=this.textModel.validatePosition(e);if(this.canBuildAST){const g=k.fromPositions(i),f=this.getBracketPairsInRange(k.fromPositions(i,i)).findLast(p=>p.closingBracketRange!==void 0&&p.range.strictContainsRange(g));return f?[f.openingBracketRange,f.closingBracketRange]:null}const n=gM(t),o=this.textModel.getLineCount(),r=new Map;let a=[];const l=(g,f)=>{if(!r.has(g)){const p=[];for(let _=0,b=f?f.brackets.length:0;_<b;_++)p[_]=0;r.set(g,p)}a=r.get(g)};let c=0;const d=(g,f,p,_,b)=>{for(;;){if(n&&++c%100===0&&!n())return Td.INSTANCE;const w=_a.findNextBracketInRange(g.forwardRegex,f,p,_,b);if(!w)break;const C=p.substring(w.startColumn-1,w.endColumn-1).toLowerCase(),v=g.textIsBracket[C];if(v&&(v.isOpen(C)?a[v.index]++:v.isClose(C)&&a[v.index]--,a[v.index]===-1))return this._matchFoundBracket(w,v,!1,n);_=w.endColumn-1}return null};let h=null,u=null;for(let g=i.lineNumber;g<=o;g++){const f=this.textModel.tokenization.getLineTokens(g),p=f.getCount(),_=this.textModel.getLineContent(g);let b=0,w=0,C=0;if(g===i.lineNumber){b=f.findTokenIndexAtOffset(i.column-1),w=i.column-1,C=i.column-1;const y=f.getLanguageId(b);h!==y&&(h=y,u=this.languageConfigurationService.getLanguageConfiguration(h).brackets,l(h,u))}let v=!0;for(;b<p;b++){const y=f.getLanguageId(b);if(h!==y){if(u&&v&&w!==C){const L=d(u,g,_,w,C);if(L)return sL(L);v=!1}h=y,u=this.languageConfigurationService.getLanguageConfiguration(h).brackets,l(h,u)}const x=!!u&&!Ul(f.getStandardTokenType(b));if(x)v||(w=f.getStartOffset(b)),C=f.getEndOffset(b);else if(u&&v&&w!==C){const L=d(u,g,_,w,C);if(L)return sL(L)}v=x}if(u&&v&&w!==C){const y=d(u,g,_,w,C);if(y)return sL(y)}}return null}_toFoundBracket(e,t){if(!t)return null;let i=this.textModel.getValueInRange(t);i=i.toLowerCase();const n=e.getBracketInfo(i);return n?{range:t,bracketInfo:n}:null}}function a1e(s,e){return{object:s,dispose:()=>e?.dispose()}}function gM(s){if(typeof s>"u")return()=>!0;{const e=Date.now();return()=>Date.now()-e<=s}}const nN=class nN{constructor(){this._searchCanceledBrand=void 0}};nN.INSTANCE=new nN;let Td=nN;function sL(s){return s instanceof Td?null:s}class l1e extends F{constructor(e){super(),this.textModel=e,this.colorProvider=new Zee,this.onDidChangeEmitter=new M,this.onDidChange=this.onDidChangeEmitter.event,this.colorizationOptions=e.getOptions().bracketPairColorizationOptions,this._register(e.bracketPairs.onDidChange(t=>{this.onDidChangeEmitter.fire()}))}handleDidChangeOptions(e){this.colorizationOptions=this.textModel.getOptions().bracketPairColorizationOptions}getDecorationsInRange(e,t,i,n){return n?[]:t===void 0?[]:this.colorizationOptions.enabled?this.textModel.bracketPairs.getBracketsInRange(e,!0).map(r=>({id:`bracket${r.range.toString()}-${r.nestingLevel}`,options:{description:"BracketPairColorization",inlineClassName:this.colorProvider.getInlineClassName(r,this.colorizationOptions.independentColorPoolPerBracketType)},ownerId:0,range:r.range})).toArray():[]}getAllDecorations(e,t){return e===void 0?[]:this.colorizationOptions.enabled?this.getDecorationsInRange(new k(1,1,this.textModel.getLineCount(),1),e,t):[]}}class Zee{constructor(){this.unexpectedClosingBracketClassName="unexpected-closing-bracket"}getInlineClassName(e,t){return e.isInvalid?this.unexpectedClosingBracketClassName:this.getInlineClassNameOfLevel(t?e.nestingLevelOfEqualBracketType:e.nestingLevel)}getInlineClassNameOfLevel(e){return`bracket-highlighting-${e%30}`}}ua((s,e)=>{const t=[pee,mee,_ee,bee,wee,vee],i=new Zee;e.addRule(`.monaco-editor .${i.unexpectedClosingBracketClassName} { color: ${s.getColor(pwe)}; }`);const n=t.map(o=>s.getColor(o)).filter(o=>!!o).filter(o=>!o.isTransparent());for(let o=0;o<30;o++){const r=n[o%n.length];e.addRule(`.monaco-editor .${i.getInlineClassNameOfLevel(o)} { color: ${r}; }`)}});function oL(s){return s.replace(/\n/g,"\\n").replace(/\r/g,"\\r")}class Xn{get oldLength(){return this.oldText.length}get oldEnd(){return this.oldPosition+this.oldText.length}get newLength(){return this.newText.length}get newEnd(){return this.newPosition+this.newText.length}constructor(e,t,i,n){this.oldPosition=e,this.oldText=t,this.newPosition=i,this.newText=n}toString(){return this.oldText.length===0?`(insert@${this.oldPosition} "${oL(this.newText)}")`:this.newText.length===0?`(delete@${this.oldPosition} "${oL(this.oldText)}")`:`(replace@${this.oldPosition} "${oL(this.oldText)}" with "${oL(this.newText)}")`}static _writeStringSize(e){return 4+2*e.length}static _writeString(e,t,i){const n=t.length;il(e,n,i),i+=4;for(let o=0;o<n;o++)bhe(e,t.charCodeAt(o),i),i+=2;return i}static _readString(e,t){const i=tl(e,t);return t+=4,Dge(e,t,i)}writeSize(){return 8+Xn._writeStringSize(this.oldText)+Xn._writeStringSize(this.newText)}write(e,t){return il(e,this.oldPosition,t),t+=4,il(e,this.newPosition,t),t+=4,t=Xn._writeString(e,this.oldText,t),t=Xn._writeString(e,this.newText,t),t}static read(e,t,i){const n=tl(e,t);t+=4;const o=tl(e,t);t+=4;const r=Xn._readString(e,t);t+=Xn._writeStringSize(r);const a=Xn._readString(e,t);return t+=Xn._writeStringSize(a),i.push(new Xn(n,r,o,a)),t}}function c1e(s,e){return s===null||s.length===0?e:new Yl(s,e).compress()}class Yl{constructor(e,t){this._prevEdits=e,this._currEdits=t,this._result=[],this._resultLen=0,this._prevLen=this._prevEdits.length,this._prevDeltaOffset=0,this._currLen=this._currEdits.length,this._currDeltaOffset=0}compress(){let e=0,t=0,i=this._getPrev(e),n=this._getCurr(t);for(;e<this._prevLen||t<this._currLen;){if(i===null){this._acceptCurr(n),n=this._getCurr(++t);continue}if(n===null){this._acceptPrev(i),i=this._getPrev(++e);continue}if(n.oldEnd<=i.newPosition){this._acceptCurr(n),n=this._getCurr(++t);continue}if(i.newEnd<=n.oldPosition){this._acceptPrev(i),i=this._getPrev(++e);continue}if(n.oldPosition<i.newPosition){const[c,d]=Yl._splitCurr(n,i.newPosition-n.oldPosition);this._acceptCurr(c),n=d;continue}if(i.newPosition<n.oldPosition){const[c,d]=Yl._splitPrev(i,n.oldPosition-i.newPosition);this._acceptPrev(c),i=d;continue}let a,l;if(n.oldEnd===i.newEnd)a=i,l=n,i=this._getPrev(++e),n=this._getCurr(++t);else if(n.oldEnd<i.newEnd){const[c,d]=Yl._splitPrev(i,n.oldLength);a=c,l=n,i=d,n=this._getCurr(++t)}else{const[c,d]=Yl._splitCurr(n,i.newLength);a=i,l=c,i=this._getPrev(++e),n=d}this._result[this._resultLen++]=new Xn(a.oldPosition,a.oldText,l.newPosition,l.newText),this._prevDeltaOffset+=a.newLength-a.oldLength,this._currDeltaOffset+=l.newLength-l.oldLength}const o=Yl._merge(this._result);return Yl._removeNoOps(o)}_acceptCurr(e){this._result[this._resultLen++]=Yl._rebaseCurr(this._prevDeltaOffset,e),this._currDeltaOffset+=e.newLength-e.oldLength}_getCurr(e){return e<this._currLen?this._currEdits[e]:null}_acceptPrev(e){this._result[this._resultLen++]=Yl._rebasePrev(this._currDeltaOffset,e),this._prevDeltaOffset+=e.newLength-e.oldLength}_getPrev(e){return e<this._prevLen?this._prevEdits[e]:null}static _rebaseCurr(e,t){return new Xn(t.oldPosition-e,t.oldText,t.newPosition,t.newText)}static _rebasePrev(e,t){return new Xn(t.oldPosition,t.oldText,t.newPosition+e,t.newText)}static _splitPrev(e,t){const i=e.newText.substr(0,t),n=e.newText.substr(t);return[new Xn(e.oldPosition,e.oldText,e.newPosition,i),new Xn(e.oldEnd,"",e.newPosition+t,n)]}static _splitCurr(e,t){const i=e.oldText.substr(0,t),n=e.oldText.substr(t);return[new Xn(e.oldPosition,i,e.newPosition,e.newText),new Xn(e.oldPosition+t,n,e.newEnd,"")]}static _merge(e){if(e.length===0)return e;const t=[];let i=0,n=e[0];for(let o=1;o<e.length;o++){const r=e[o];n.oldEnd===r.oldPosition?n=new Xn(n.oldPosition,n.oldText+r.oldText,n.newPosition,n.newText+r.newText):(t[i++]=n,n=r)}return t[i++]=n,t}static _removeNoOps(e){if(e.length===0)return e;const t=[];let i=0;for(let n=0;n<e.length;n++){const o=e[n];o.oldText!==o.newText&&(t[i++]=o)}return t}}const d1e=Symbol("TextModelEditSource");class Xee{constructor(e,t){this.metadata=e}toString(){return`${this.metadata.source}`}getType(){const e=this.metadata;switch(e.source){case"cursor":return e.kind;case"inlineCompletionAccept":return e.source+(e.$nes?":nes":"");case"unknown":return e.name||"unknown";default:return e.source}}toKey(e,t={}){const i=this.metadata;return Object.entries(i).filter(([o,r])=>{const a=t[o];return a!==void 0?a:(o.match(/\$/g)||[]).length<=e&&r!==void 0&&r!==null&&r!==""}).map(([o,r])=>`${o}:${r}`).join("-")}}function lo(s){return new Xee(s,d1e)}const cs={unknown(s){return lo({source:"unknown",name:s.name})},rename:()=>lo({source:"rename"}),chatApplyEdits(s){return lo({source:"Chat.applyEdits",$modelId:bU(s.modelId),$extensionId:s.extensionId?.extensionId,$extensionVersion:s.extensionId?.version,$$languageId:s.languageId,$$sessionId:s.sessionId,$$requestId:s.requestId,$$mode:s.mode,$$codeBlockSuggestionId:s.codeBlockSuggestionId})},chatUndoEdits:()=>lo({source:"Chat.undoEdits"}),chatReset:()=>lo({source:"Chat.reset"}),inlineCompletionAccept(s){return lo({source:"inlineCompletionAccept",$nes:s.nes,...rL(s.providerId),$$requestUuid:s.requestUuid,$$languageId:s.languageId})},inlineCompletionPartialAccept(s){return lo({source:"inlineCompletionPartialAccept",type:s.type,$nes:s.nes,...rL(s.providerId),$$requestUuid:s.requestUuid,$$languageId:s.languageId})},inlineChatApplyEdit(s){return lo({source:"inlineChat.applyEdits",$modelId:bU(s.modelId),$extensionId:s.extensionId?.extensionId,$extensionVersion:s.extensionId?.version,$$sessionId:s.sessionId,$$requestId:s.requestId,$$languageId:s.languageId})},reloadFromDisk:()=>lo({source:"reloadFromDisk"}),cursor(s){return lo({source:"cursor",kind:s.kind,detailedSource:s.detailedSource})},setValue:()=>lo({source:"setValue"}),eolChange:()=>lo({source:"eolChange"}),applyEdits:()=>lo({source:"applyEdits"}),snippet:()=>lo({source:"snippet"}),suggest:s=>lo({source:"suggest",...rL(s.providerId)}),codeAction:s=>lo({source:"codeAction",$kind:s.kind,...rL(s.providerId)})};function rL(s){return s?{$extensionId:s.extensionId,$extensionVersion:s.extensionVersion,$providerId:s.providerId}:{}}function bU(s){if(s!==void 0)return s.replaceAll("/","|")}class ES{static fromText(e){const t=cn.ofText(e).lineCount,i=e.length;return new ES(t,0,i,0)}static tryCreate(e,t,i,n){if(!(e===void 0||t===void 0||i===void 0||n===void 0))return new ES(e,t,i,n)}constructor(e,t,i,n){this.linesAdded=e,this.linesRemoved=t,this.charsAdded=i,this.charsRemoved=n}}var wU;(function(s){function e(t){const i=t?t("sgt"):oee("sgt");return i}s.newId=e})(wU||(wU={}));function V_(s){return s.toString()}class Sn{static create(e,t){const i=e.getAlternativeVersionId(),n=RF(e);return new Sn(i,i,n,n,t,t,[])}constructor(e,t,i,n,o,r,a){this.beforeVersionId=e,this.afterVersionId=t,this.beforeEOL=i,this.afterEOL=n,this.beforeCursorState=o,this.afterCursorState=r,this.changes=a}append(e,t,i,n,o){t.length>0&&(this.changes=c1e(this.changes,t)),this.afterEOL=i,this.afterVersionId=n,this.afterCursorState=o}static _writeSelectionsSize(e){return 4+16*(e?e.length:0)}static _writeSelections(e,t,i){if(il(e,t?t.length:0,i),i+=4,t)for(const n of t)il(e,n.selectionStartLineNumber,i),i+=4,il(e,n.selectionStartColumn,i),i+=4,il(e,n.positionLineNumber,i),i+=4,il(e,n.positionColumn,i),i+=4;return i}static _readSelections(e,t,i){const n=tl(e,t);t+=4;for(let o=0;o<n;o++){const r=tl(e,t);t+=4;const a=tl(e,t);t+=4;const l=tl(e,t);t+=4;const c=tl(e,t);t+=4,i.push(new he(r,a,l,c))}return t}serialize(){let e=10+Sn._writeSelectionsSize(this.beforeCursorState)+Sn._writeSelectionsSize(this.afterCursorState)+4;for(const n of this.changes)e+=n.writeSize();const t=new Uint8Array(e);let i=0;il(t,this.beforeVersionId,i),i+=4,il(t,this.afterVersionId,i),i+=4,Az(t,this.beforeEOL,i),i+=1,Az(t,this.afterEOL,i),i+=1,i=Sn._writeSelections(t,this.beforeCursorState,i),i=Sn._writeSelections(t,this.afterCursorState,i),il(t,this.changes.length,i),i+=4;for(const n of this.changes)i=n.write(t,i);return t.buffer}static deserialize(e){const t=new Uint8Array(e);let i=0;const n=tl(t,i);i+=4;const o=tl(t,i);i+=4;const r=Rz(t,i);i+=1;const a=Rz(t,i);i+=1;const l=[];i=Sn._readSelections(t,i,l);const c=[];i=Sn._readSelections(t,i,c);const d=tl(t,i);i+=4;const h=[];for(let u=0;u<d;u++)i=Xn.read(t,i,h);return new Sn(n,o,r,a,l,c,h)}}class Qee{get type(){return 0}get resource(){return Le.isUri(this.model)?this.model:this.model.uri}constructor(e,t,i,n){this.label=e,this.code=t,this.model=i,this._data=Sn.create(i,n)}toString(){return(this._data instanceof Sn?this._data:Sn.deserialize(this._data)).changes.map(t=>t.toString()).join(", ")}matchesResource(e){return(Le.isUri(this.model)?this.model:this.model.uri).toString()===e.toString()}setModel(e){this.model=e}canAppend(e){return this.model===e&&this._data instanceof Sn}append(e,t,i,n,o){this._data instanceof Sn&&this._data.append(e,t,i,n,o)}close(){this._data instanceof Sn&&(this._data=this._data.serialize())}open(){this._data instanceof Sn||(this._data=Sn.deserialize(this._data))}undo(){if(Le.isUri(this.model))throw new Error("Invalid SingleModelEditStackElement");this._data instanceof Sn&&(this._data=this._data.serialize());const e=Sn.deserialize(this._data);this.model._applyUndo(e.changes,e.beforeEOL,e.beforeVersionId,e.beforeCursorState)}redo(){if(Le.isUri(this.model))throw new Error("Invalid SingleModelEditStackElement");this._data instanceof Sn&&(this._data=this._data.serialize());const e=Sn.deserialize(this._data);this.model._applyRedo(e.changes,e.afterEOL,e.afterVersionId,e.afterCursorState)}heapSize(){return this._data instanceof Sn&&(this._data=this._data.serialize()),this._data.byteLength+168}}class h1e{get resources(){return this._editStackElementsArr.map(e=>e.resource)}constructor(e,t,i){this.label=e,this.code=t,this.type=1,this._isOpen=!0,this._editStackElementsArr=i.slice(0),this._editStackElementsMap=new Map;for(const n of this._editStackElementsArr){const o=V_(n.resource);this._editStackElementsMap.set(o,n)}this._delegate=null}prepareUndoRedo(){if(this._delegate)return this._delegate.prepareUndoRedo(this)}matchesResource(e){const t=V_(e);return this._editStackElementsMap.has(t)}setModel(e){const t=V_(Le.isUri(e)?e:e.uri);this._editStackElementsMap.has(t)&&this._editStackElementsMap.get(t).setModel(e)}canAppend(e){if(!this._isOpen)return!1;const t=V_(e.uri);return this._editStackElementsMap.has(t)?this._editStackElementsMap.get(t).canAppend(e):!1}append(e,t,i,n,o){const r=V_(e.uri);this._editStackElementsMap.get(r).append(e,t,i,n,o)}close(){this._isOpen=!1}open(){}undo(){this._isOpen=!1;for(const e of this._editStackElementsArr)e.undo()}redo(){for(const e of this._editStackElementsArr)e.redo()}heapSize(e){const t=V_(e);return this._editStackElementsMap.has(t)?this._editStackElementsMap.get(t).heapSize():0}split(){return this._editStackElementsArr}toString(){const e=[];for(const t of this._editStackElementsArr)e.push(`${ra(t.resource)}: ${t}`);return`{${e.join(", ")}}`}}function RF(s){return s.getEOL()===`
`?0:1}function Nd(s){return s?s instanceof Qee||s instanceof h1e:!1}class r7{constructor(e,t){this._model=e,this._undoRedoService=t}pushStackElement(){const e=this._undoRedoService.getLastElement(this._model.uri);Nd(e)&&e.close()}popStackElement(){const e=this._undoRedoService.getLastElement(this._model.uri);Nd(e)&&e.open()}clear(){this._undoRedoService.removeElements(this._model.uri)}_getOrCreateEditStackElement(e,t){const i=this._undoRedoService.getLastElement(this._model.uri);if(Nd(i)&&i.canAppend(this._model))return i;const n=new Qee(m(785,"Typing"),"undoredo.textBufferEdit",this._model,e);return this._undoRedoService.pushElement(n,t),n}pushEOL(e){const t=this._getOrCreateEditStackElement(null,void 0);this._model.setEOL(e),t.append(this._model,[],RF(this._model),this._model.getAlternativeVersionId(),null)}pushEditOperation(e,t,i,n,o=cs.unknown({name:"pushEditOperation"})){const r=this._getOrCreateEditStackElement(e,n),a=this._model.applyEdits(t,!0,o),l=r7._computeCursorState(i,a),c=a.map((d,h)=>({index:h,textChange:d.textChange}));return c.sort((d,h)=>d.textChange.oldPosition===h.textChange.oldPosition?d.index-h.index:d.textChange.oldPosition-h.textChange.oldPosition),r.append(this._model,c.map(d=>d.textChange),RF(this._model),this._model.getAlternativeVersionId(),l),l}static _computeCursorState(e,t){try{return e?e(t):null}catch(i){return Oe(i),null}}}class u1e{constructor(){this.spacesDiff=0,this.looksLikeAlignment=!1}}function g1e(s,e,t,i,n){n.spacesDiff=0,n.looksLikeAlignment=!1;let o;for(o=0;o<e&&o<i;o++){const u=s.charCodeAt(o),g=t.charCodeAt(o);if(u!==g)break}let r=0,a=0;for(let u=o;u<e;u++)s.charCodeAt(u)===32?r++:a++;let l=0,c=0;for(let u=o;u<i;u++)t.charCodeAt(u)===32?l++:c++;if(r>0&&a>0||l>0&&c>0)return;const d=Math.abs(a-c),h=Math.abs(r-l);if(d===0){n.spacesDiff=h,h>0&&0<=l-1&&l-1<s.length&&l<t.length&&t.charCodeAt(l)!==32&&s.charCodeAt(l-1)===32&&s.charCodeAt(s.length-1)===44&&(n.looksLikeAlignment=!0);return}if(h%d===0){n.spacesDiff=h/d;return}}function vU(s,e,t){const i=Math.min(s.getLineCount(),1e4);let n=0,o=0,r="",a=0;const l=[2,4,6,8,3,5,7],c=8,d=[0,0,0,0,0,0,0,0,0],h=new u1e;for(let f=1;f<=i;f++){const p=s.getLineLength(f),_=s.getLineContent(f),b=p<=65536;let w=!1,C=0,v=0,y=0;for(let L=0,E=p;L<E;L++){const I=b?_.charCodeAt(L):s.getLineCharCode(f,L);if(I===9)y++;else if(I===32)v++;else{w=!0,C=L;break}}if(!w||(y>0?n++:v>1&&o++,g1e(r,a,_,C,h),h.looksLikeAlignment&&!(t&&e===h.spacesDiff)))continue;const x=h.spacesDiff;x<=c&&d[x]++,r=_,a=C}let u=t;n!==o&&(u=n<o);let g=e;if(u){let f=u?0:.1*i;l.forEach(p=>{const _=d[p];_>f&&(f=_,g=p)}),g===4&&d[4]>0&&d[2]>0&&d[2]>=d[4]/2&&(g=2)}return{insertSpaces:u,tabSize:g}}function co(s){return(s.metadata&1)>>>0}function Jt(s,e){s.metadata=s.metadata&254|e<<0}function ss(s){return(s.metadata&2)>>>1===1}function Zt(s,e){s.metadata=s.metadata&253|(e?1:0)<<1}function Jee(s){return(s.metadata&4)>>>2===1}function CU(s,e){s.metadata=s.metadata&251|(e?1:0)<<2}function ete(s){return(s.metadata&64)>>>6===1}function yU(s,e){s.metadata=s.metadata&191|(e?1:0)<<6}function tte(s){return(s.metadata&128)>>>7===1}function SU(s,e){s.metadata=s.metadata&127|(e?1:0)<<7}function f1e(s){return(s.metadata&24)>>>3}function xU(s,e){s.metadata=s.metadata&231|e<<3}function p1e(s){return(s.metadata&32)>>>5===1}function LU(s,e){s.metadata=s.metadata&223|(e?1:0)<<5}class ite{constructor(e,t,i){this.metadata=0,this.parent=this,this.left=this,this.right=this,Jt(this,1),this.start=t,this.end=i,this.delta=0,this.maxEnd=i,this.id=e,this.ownerId=0,this.options=null,CU(this,!1),yU(this,!1),xU(this,1),LU(this,!1),SU(this,!1),this.cachedVersionId=0,this.cachedAbsoluteStart=t,this.cachedAbsoluteEnd=i,this.range=null,Zt(this,!1)}reset(e,t,i,n){this.start=t,this.end=i,this.maxEnd=i,this.cachedVersionId=e,this.cachedAbsoluteStart=t,this.cachedAbsoluteEnd=i,this.range=n}setOptions(e){this.options=e;const t=this.options.className;CU(this,t==="squiggly-error"||t==="squiggly-warning"||t==="squiggly-info"),yU(this,this.options.glyphMarginClassName!==null),xU(this,this.options.stickiness),LU(this,this.options.collapseOnReplaceEdit),SU(this,this.options.affectsFont??!1)}setCachedOffsets(e,t,i){this.cachedVersionId!==i&&(this.range=null),this.cachedVersionId=i,this.cachedAbsoluteStart=e,this.cachedAbsoluteEnd=t}detach(){this.parent=null,this.left=null,this.right=null}}const it=new ite(null,0,0);it.parent=it;it.left=it;it.right=it;Jt(it,0);class fM{constructor(){this.root=it,this.requestNormalizeDelta=!1}intervalSearch(e,t,i,n,o,r,a){return this.root===it?[]:S1e(this,e,t,i,n,o,r,a)}search(e,t,i,n,o){return this.root===it?[]:y1e(this,e,t,i,n,o)}collectNodesFromOwner(e){return v1e(this,e)}collectNodesPostOrder(){return C1e(this)}insert(e){kU(this,e),this._normalizeDeltaIfNecessary()}delete(e){EU(this,e),this._normalizeDeltaIfNecessary()}resolveNode(e,t){const i=e;let n=0;for(;e!==this.root;)e===e.parent.right&&(n+=e.parent.delta),e=e.parent;const o=i.start+n,r=i.end+n;i.setCachedOffsets(o,r,t)}acceptReplace(e,t,i,n){const o=b1e(this,e,e+t);for(let r=0,a=o.length;r<a;r++){const l=o[r];EU(this,l)}this._normalizeDeltaIfNecessary(),w1e(this,e,e+t,i),this._normalizeDeltaIfNecessary();for(let r=0,a=o.length;r<a;r++){const l=o[r];l.start=l.cachedAbsoluteStart,l.end=l.cachedAbsoluteEnd,_1e(l,e,e+t,i,n),l.maxEnd=l.end,kU(this,l)}this._normalizeDeltaIfNecessary()}_normalizeDeltaIfNecessary(){this.requestNormalizeDelta&&(this.requestNormalizeDelta=!1,m1e(this))}}function m1e(s){let e=s.root,t=0;for(;e!==it;){if(e.left!==it&&!ss(e.left)){e=e.left;continue}if(e.right!==it&&!ss(e.right)){t+=e.delta,e=e.right;continue}e.start=t+e.start,e.end=t+e.end,e.delta=0,pg(e),Zt(e,!0),Zt(e.left,!1),Zt(e.right,!1),e===e.parent.right&&(t-=e.parent.delta),e=e.parent}Zt(s.root,!1)}function z_(s,e,t,i){return s<t?!0:s>t||i===1?!1:i===2?!0:e}function _1e(s,e,t,i,n){const o=f1e(s),r=o===0||o===2,a=o===1||o===2,l=t-e,c=i,d=Math.min(l,c),h=s.start;let u=!1;const g=s.end;let f=!1;e<=h&&g<=t&&p1e(s)&&(s.start=e,u=!0,s.end=e,f=!0);{const _=n?1:l>0?2:0;!u&&z_(h,r,e,_)&&(u=!0),!f&&z_(g,a,e,_)&&(f=!0)}if(d>0&&!n){const _=l>c?2:0;!u&&z_(h,r,e+d,_)&&(u=!0),!f&&z_(g,a,e+d,_)&&(f=!0)}{const _=n?1:0;!u&&z_(h,r,t,_)&&(s.start=e+c,u=!0),!f&&z_(g,a,t,_)&&(s.end=e+c,f=!0)}const p=c-l;u||(s.start=Math.max(0,h+p)),f||(s.end=Math.max(0,g+p)),s.start>s.end&&(s.end=s.start)}function b1e(s,e,t){let i=s.root,n=0,o=0,r=0,a=0;const l=[];let c=0;for(;i!==it;){if(ss(i)){Zt(i.left,!1),Zt(i.right,!1),i===i.parent.right&&(n-=i.parent.delta),i=i.parent;continue}if(!ss(i.left)){if(o=n+i.maxEnd,o<e){Zt(i,!0);continue}if(i.left!==it){i=i.left;continue}}if(r=n+i.start,r>t){Zt(i,!0);continue}if(a=n+i.end,a>=e&&(i.setCachedOffsets(r,a,0),l[c++]=i),Zt(i,!0),i.right!==it&&!ss(i.right)){n+=i.delta,i=i.right;continue}}return Zt(s.root,!1),l}function w1e(s,e,t,i){let n=s.root,o=0,r=0,a=0;const l=i-(t-e);for(;n!==it;){if(ss(n)){Zt(n.left,!1),Zt(n.right,!1),n===n.parent.right&&(o-=n.parent.delta),pg(n),n=n.parent;continue}if(!ss(n.left)){if(r=o+n.maxEnd,r<e){Zt(n,!0);continue}if(n.left!==it){n=n.left;continue}}if(a=o+n.start,a>t){n.start+=l,n.end+=l,n.delta+=l,(n.delta<-1073741824||n.delta>1073741824)&&(s.requestNormalizeDelta=!0),Zt(n,!0);continue}if(Zt(n,!0),n.right!==it&&!ss(n.right)){o+=n.delta,n=n.right;continue}}Zt(s.root,!1)}function v1e(s,e){let t=s.root;const i=[];let n=0;for(;t!==it;){if(ss(t)){Zt(t.left,!1),Zt(t.right,!1),t=t.parent;continue}if(t.left!==it&&!ss(t.left)){t=t.left;continue}if(t.ownerId===e&&(i[n++]=t),Zt(t,!0),t.right!==it&&!ss(t.right)){t=t.right;continue}}return Zt(s.root,!1),i}function C1e(s){let e=s.root;const t=[];let i=0;for(;e!==it;){if(ss(e)){Zt(e.left,!1),Zt(e.right,!1),e=e.parent;continue}if(e.left!==it&&!ss(e.left)){e=e.left;continue}if(e.right!==it&&!ss(e.right)){e=e.right;continue}t[i++]=e,Zt(e,!0)}return Zt(s.root,!1),t}function y1e(s,e,t,i,n,o){let r=s.root,a=0,l=0,c=0;const d=[];let h=0;for(;r!==it;){if(ss(r)){Zt(r.left,!1),Zt(r.right,!1),r===r.parent.right&&(a-=r.parent.delta),r=r.parent;continue}if(r.left!==it&&!ss(r.left)){r=r.left;continue}l=a+r.start,c=a+r.end,r.setCachedOffsets(l,c,n);let u=!0;if(e&&r.ownerId&&r.ownerId!==e&&(u=!1),t&&Jee(r)&&(u=!1),i&&tte(r)&&(u=!1),o&&!ete(r)&&(u=!1),u&&(d[h++]=r),Zt(r,!0),r.right!==it&&!ss(r.right)){a+=r.delta,r=r.right;continue}}return Zt(s.root,!1),d}function S1e(s,e,t,i,n,o,r,a){let l=s.root,c=0,d=0,h=0,u=0;const g=[];let f=0;for(;l!==it;){if(ss(l)){Zt(l.left,!1),Zt(l.right,!1),l===l.parent.right&&(c-=l.parent.delta),l=l.parent;continue}if(!ss(l.left)){if(d=c+l.maxEnd,d<e){Zt(l,!0);continue}if(l.left!==it){l=l.left;continue}}if(h=c+l.start,h>t){Zt(l,!0);continue}if(u=c+l.end,u>=e){l.setCachedOffsets(h,u,r);let p=!0;i&&l.ownerId&&l.ownerId!==i&&(p=!1),n&&Jee(l)&&(p=!1),o&&tte(l)&&(p=!1),a&&!ete(l)&&(p=!1),p&&(g[f++]=l)}if(Zt(l,!0),l.right!==it&&!ss(l.right)){c+=l.delta,l=l.right;continue}}return Zt(s.root,!1),g}function kU(s,e){if(s.root===it)return e.parent=it,e.left=it,e.right=it,Jt(e,0),s.root=e,s.root;x1e(s,e),Uh(e.parent);let t=e;for(;t!==s.root&&co(t.parent)===1;)if(t.parent===t.parent.parent.left){const i=t.parent.parent.right;co(i)===1?(Jt(t.parent,0),Jt(i,0),Jt(t.parent.parent,1),t=t.parent.parent):(t===t.parent.right&&(t=t.parent,P1(s,t)),Jt(t.parent,0),Jt(t.parent.parent,1),F1(s,t.parent.parent))}else{const i=t.parent.parent.left;co(i)===1?(Jt(t.parent,0),Jt(i,0),Jt(t.parent.parent,1),t=t.parent.parent):(t===t.parent.left&&(t=t.parent,F1(s,t)),Jt(t.parent,0),Jt(t.parent.parent,1),P1(s,t.parent.parent))}return Jt(s.root,0),e}function x1e(s,e){let t=0,i=s.root;const n=e.start,o=e.end;for(;;)if(k1e(n,o,i.start+t,i.end+t)<0)if(i.left===it){e.start-=t,e.end-=t,e.maxEnd-=t,i.left=e;break}else i=i.left;else if(i.right===it){e.start-=t+i.delta,e.end-=t+i.delta,e.maxEnd-=t+i.delta,i.right=e;break}else t+=i.delta,i=i.right;e.parent=i,e.left=it,e.right=it,Jt(e,1)}function EU(s,e){let t,i;if(e.left===it?(t=e.right,i=e,t.delta+=e.delta,(t.delta<-1073741824||t.delta>1073741824)&&(s.requestNormalizeDelta=!0),t.start+=e.delta,t.end+=e.delta):e.right===it?(t=e.left,i=e):(i=L1e(e.right),t=i.right,t.start+=i.delta,t.end+=i.delta,t.delta+=i.delta,(t.delta<-1073741824||t.delta>1073741824)&&(s.requestNormalizeDelta=!0),i.start+=e.delta,i.end+=e.delta,i.delta=e.delta,(i.delta<-1073741824||i.delta>1073741824)&&(s.requestNormalizeDelta=!0)),i===s.root){s.root=t,Jt(t,0),e.detach(),pM(),pg(t),s.root.parent=it;return}const n=co(i)===1;if(i===i.parent.left?i.parent.left=t:i.parent.right=t,i===e?t.parent=i.parent:(i.parent===e?t.parent=i:t.parent=i.parent,i.left=e.left,i.right=e.right,i.parent=e.parent,Jt(i,co(e)),e===s.root?s.root=i:e===e.parent.left?e.parent.left=i:e.parent.right=i,i.left!==it&&(i.left.parent=i),i.right!==it&&(i.right.parent=i)),e.detach(),n){Uh(t.parent),i!==e&&(Uh(i),Uh(i.parent)),pM();return}Uh(t),Uh(t.parent),i!==e&&(Uh(i),Uh(i.parent));let o;for(;t!==s.root&&co(t)===0;)t===t.parent.left?(o=t.parent.right,co(o)===1&&(Jt(o,0),Jt(t.parent,1),P1(s,t.parent),o=t.parent.right),co(o.left)===0&&co(o.right)===0?(Jt(o,1),t=t.parent):(co(o.right)===0&&(Jt(o.left,0),Jt(o,1),F1(s,o),o=t.parent.right),Jt(o,co(t.parent)),Jt(t.parent,0),Jt(o.right,0),P1(s,t.parent),t=s.root)):(o=t.parent.left,co(o)===1&&(Jt(o,0),Jt(t.parent,1),F1(s,t.parent),o=t.parent.left),co(o.left)===0&&co(o.right)===0?(Jt(o,1),t=t.parent):(co(o.left)===0&&(Jt(o.right,0),Jt(o,1),P1(s,o),o=t.parent.left),Jt(o,co(t.parent)),Jt(t.parent,0),Jt(o.left,0),F1(s,t.parent),t=s.root));Jt(t,0),pM()}function L1e(s){for(;s.left!==it;)s=s.left;return s}function pM(){it.parent=it,it.delta=0,it.start=0,it.end=0}function P1(s,e){const t=e.right;t.delta+=e.delta,(t.delta<-1073741824||t.delta>1073741824)&&(s.requestNormalizeDelta=!0),t.start+=e.delta,t.end+=e.delta,e.right=t.left,t.left!==it&&(t.left.parent=e),t.parent=e.parent,e.parent===it?s.root=t:e===e.parent.left?e.parent.left=t:e.parent.right=t,t.left=e,e.parent=t,pg(e),pg(t)}function F1(s,e){const t=e.left;e.delta-=t.delta,(e.delta<-1073741824||e.delta>1073741824)&&(s.requestNormalizeDelta=!0),e.start-=t.delta,e.end-=t.delta,e.left=t.right,t.right!==it&&(t.right.parent=e),t.parent=e.parent,e.parent===it?s.root=t:e===e.parent.right?e.parent.right=t:e.parent.left=t,t.right=e,e.parent=t,pg(e),pg(t)}function nte(s){let e=s.end;if(s.left!==it){const t=s.left.maxEnd;t>e&&(e=t)}if(s.right!==it){const t=s.right.maxEnd+s.delta;t>e&&(e=t)}return e}function pg(s){s.maxEnd=nte(s)}function Uh(s){for(;s!==it;){const e=nte(s);if(s.maxEnd===e)return;s.maxEnd=e,s=s.parent}}function k1e(s,e,t,i){return s===t?e-i:s-t}class AF{constructor(e,t){this.piece=e,this.color=t,this.size_left=0,this.lf_left=0,this.parent=this,this.left=this,this.right=this}next(){if(this.right!==$e)return a7(this.right);let e=this;for(;e.parent!==$e&&e.parent.left!==e;)e=e.parent;return e.parent===$e?$e:e.parent}prev(){if(this.left!==$e)return ste(this.left);let e=this;for(;e.parent!==$e&&e.parent.right!==e;)e=e.parent;return e.parent===$e?$e:e.parent}detach(){this.parent=null,this.left=null,this.right=null}}const $e=new AF(null,0);$e.parent=$e;$e.left=$e;$e.right=$e;$e.color=0;function a7(s){for(;s.left!==$e;)s=s.left;return s}function ste(s){for(;s.right!==$e;)s=s.right;return s}function l7(s){return s===$e?0:s.size_left+s.piece.length+l7(s.right)}function c7(s){return s===$e?0:s.lf_left+s.piece.lineFeedCnt+c7(s.right)}function mM(){$e.parent=$e}function B1(s,e){const t=e.right;t.size_left+=e.size_left+(e.piece?e.piece.length:0),t.lf_left+=e.lf_left+(e.piece?e.piece.lineFeedCnt:0),e.right=t.left,t.left!==$e&&(t.left.parent=e),t.parent=e.parent,e.parent===$e?s.root=t:e.parent.left===e?e.parent.left=t:e.parent.right=t,t.left=e,e.parent=t}function W1(s,e){const t=e.left;e.left=t.right,t.right!==$e&&(t.right.parent=e),t.parent=e.parent,e.size_left-=t.size_left+(t.piece?t.piece.length:0),e.lf_left-=t.lf_left+(t.piece?t.piece.lineFeedCnt:0),e.parent===$e?s.root=t:e===e.parent.right?e.parent.right=t:e.parent.left=t,t.right=e,e.parent=t}function aL(s,e){let t,i;if(e.left===$e?(i=e,t=i.right):e.right===$e?(i=e,t=i.left):(i=a7(e.right),t=i.right),i===s.root){s.root=t,t.color=0,e.detach(),mM(),s.root.parent=$e;return}const n=i.color===1;if(i===i.parent.left?i.parent.left=t:i.parent.right=t,i===e?(t.parent=i.parent,c1(s,t)):(i.parent===e?t.parent=i:t.parent=i.parent,c1(s,t),i.left=e.left,i.right=e.right,i.parent=e.parent,i.color=e.color,e===s.root?s.root=i:e===e.parent.left?e.parent.left=i:e.parent.right=i,i.left!==$e&&(i.left.parent=i),i.right!==$e&&(i.right.parent=i),i.size_left=e.size_left,i.lf_left=e.lf_left,c1(s,i)),e.detach(),t.parent.left===t){const r=l7(t),a=c7(t);if(r!==t.parent.size_left||a!==t.parent.lf_left){const l=r-t.parent.size_left,c=a-t.parent.lf_left;t.parent.size_left=r,t.parent.lf_left=a,gd(s,t.parent,l,c)}}if(c1(s,t.parent),n){mM();return}let o;for(;t!==s.root&&t.color===0;)t===t.parent.left?(o=t.parent.right,o.color===1&&(o.color=0,t.parent.color=1,B1(s,t.parent),o=t.parent.right),o.left.color===0&&o.right.color===0?(o.color=1,t=t.parent):(o.right.color===0&&(o.left.color=0,o.color=1,W1(s,o),o=t.parent.right),o.color=t.parent.color,t.parent.color=0,o.right.color=0,B1(s,t.parent),t=s.root)):(o=t.parent.left,o.color===1&&(o.color=0,t.parent.color=1,W1(s,t.parent),o=t.parent.left),o.left.color===0&&o.right.color===0?(o.color=1,t=t.parent):(o.left.color===0&&(o.right.color=0,o.color=1,B1(s,o),o=t.parent.left),o.color=t.parent.color,t.parent.color=0,o.left.color=0,W1(s,t.parent),t=s.root));t.color=0,mM()}function IU(s,e){for(c1(s,e);e!==s.root&&e.parent.color===1;)if(e.parent===e.parent.parent.left){const t=e.parent.parent.right;t.color===1?(e.parent.color=0,t.color=0,e.parent.parent.color=1,e=e.parent.parent):(e===e.parent.right&&(e=e.parent,B1(s,e)),e.parent.color=0,e.parent.parent.color=1,W1(s,e.parent.parent))}else{const t=e.parent.parent.left;t.color===1?(e.parent.color=0,t.color=0,e.parent.parent.color=1,e=e.parent.parent):(e===e.parent.left&&(e=e.parent,W1(s,e)),e.parent.color=0,e.parent.parent.color=1,B1(s,e.parent.parent))}s.root.color=0}function gd(s,e,t,i){for(;e!==s.root&&e!==$e;)e.parent.left===e&&(e.parent.size_left+=t,e.parent.lf_left+=i),e=e.parent}function c1(s,e){let t=0,i=0;if(e!==s.root){for(;e!==s.root&&e===e.parent.right;)e=e.parent;if(e!==s.root)for(e=e.parent,t=l7(e.left)-e.size_left,i=c7(e.left)-e.lf_left,e.size_left+=t,e.lf_left+=i;e!==s.root&&(t!==0||i!==0);)e.parent.left===e&&(e.parent.size_left+=t,e.parent.lf_left+=i),e=e.parent}}const E1e=999;class nf{constructor(e,t,i,n){this.searchString=e,this.isRegex=t,this.matchCase=i,this.wordSeparators=n}parseSearchRequest(){if(this.searchString==="")return null;let e;this.isRegex?e=ote(this.searchString):e=this.searchString.indexOf(`
`)>=0;let t=null;try{t=JX(this.searchString,this.isRegex,{matchCase:this.matchCase,wholeWord:!1,multiline:e,global:!0,unicode:!0})}catch{return null}if(!t)return null;let i=!this.isRegex&&!e;return i&&this.searchString.toLowerCase()!==this.searchString.toUpperCase()&&(i=this.matchCase),new Sge(t,this.wordSeparators?aa(this.wordSeparators,[]):null,i?this.searchString:null)}}function ote(s){if(!s||s.length===0)return!1;for(let e=0,t=s.length;e<t;e++){const i=s.charCodeAt(e);if(i===10)return!0;if(i===92){if(e++,e>=t)break;const n=s.charCodeAt(e);if(n===110||n===114||n===87)return!0}}return!1}function bf(s,e,t){if(!t)return new tS(s,null);const i=[];for(let n=0,o=e.length;n<o;n++)i[n]=e[n];return new tS(s,i)}class DU{constructor(e){const t=[];let i=0;for(let n=0,o=e.length;n<o;n++)e.charCodeAt(n)===10&&(t[i++]=n);this._lineFeedsOffsets=t}findLineFeedCountBeforeOffset(e){const t=this._lineFeedsOffsets;let i=0,n=t.length-1;if(n===-1||e<=t[0])return 0;for(;i<n;){const o=i+((n-i)/2>>0);t[o]>=e?n=o-1:t[o+1]>=e?(i=o,n=o):i=o+1}return i+1}}class lL{static findMatches(e,t,i,n,o){const r=t.parseSearchRequest();return r?r.regex.multiline?this._doFindMatchesMultiline(e,i,new Db(r.wordSeparators,r.regex),n,o):this._doFindMatchesLineByLine(e,i,r,n,o):[]}static _getMultilineMatchRange(e,t,i,n,o,r){let a,l=0;n?(l=n.findLineFeedCountBeforeOffset(o),a=t+o+l):a=t+o;let c;if(n){const g=n.findLineFeedCountBeforeOffset(o+r.length)-l;c=a+r.length+g}else c=a+r.length;const d=e.getPositionAt(a),h=e.getPositionAt(c);return new k(d.lineNumber,d.column,h.lineNumber,h.column)}static _doFindMatchesMultiline(e,t,i,n,o){const r=e.getOffsetAt(t.getStartPosition()),a=e.getValueInRange(t,1),l=e.getEOL()===`\r
`?new DU(a):null,c=[];let d=0,h;for(i.reset(0);h=i.next(a);)if(c[d++]=bf(this._getMultilineMatchRange(e,r,a,l,h.index,h[0]),h,n),d>=o)return c;return c}static _doFindMatchesLineByLine(e,t,i,n,o){const r=[];let a=0;if(t.startLineNumber===t.endLineNumber){const c=e.getLineContent(t.startLineNumber).substring(t.startColumn-1,t.endColumn-1);return a=this._findMatchesInLine(i,c,t.startLineNumber,t.startColumn-1,a,r,n,o),r}const l=e.getLineContent(t.startLineNumber).substring(t.startColumn-1);a=this._findMatchesInLine(i,l,t.startLineNumber,t.startColumn-1,a,r,n,o);for(let c=t.startLineNumber+1;c<t.endLineNumber&&a<o;c++)a=this._findMatchesInLine(i,e.getLineContent(c),c,0,a,r,n,o);if(a<o){const c=e.getLineContent(t.endLineNumber).substring(0,t.endColumn-1);a=this._findMatchesInLine(i,c,t.endLineNumber,0,a,r,n,o)}return r}static _findMatchesInLine(e,t,i,n,o,r,a,l){const c=e.wordSeparators;if(!a&&e.simpleSearch){const u=e.simpleSearch,g=u.length,f=t.length;let p=-g;for(;(p=t.indexOf(u,p+g))!==-1;)if((!c||d7(c,t,f,p,g))&&(r[o++]=new tS(new k(i,p+1+n,i,p+1+g+n),null),o>=l))return o;return o}const d=new Db(e.wordSeparators,e.regex);let h;d.reset(0);do if(h=d.next(t),h&&(r[o++]=bf(new k(i,h.index+1+n,i,h.index+1+h[0].length+n),h,a),o>=l))return o;while(h);return o}static findNextMatch(e,t,i,n){const o=t.parseSearchRequest();if(!o)return null;const r=new Db(o.wordSeparators,o.regex);return o.regex.multiline?this._doFindNextMatchMultiline(e,i,r,n):this._doFindNextMatchLineByLine(e,i,r,n)}static _doFindNextMatchMultiline(e,t,i,n){const o=new A(t.lineNumber,1),r=e.getOffsetAt(o),a=e.getLineCount(),l=e.getValueInRange(new k(o.lineNumber,o.column,a,e.getLineMaxColumn(a)),1),c=e.getEOL()===`\r
`?new DU(l):null;i.reset(t.column-1);const d=i.next(l);return d?bf(this._getMultilineMatchRange(e,r,l,c,d.index,d[0]),d,n):t.lineNumber!==1||t.column!==1?this._doFindNextMatchMultiline(e,new A(1,1),i,n):null}static _doFindNextMatchLineByLine(e,t,i,n){const o=e.getLineCount(),r=t.lineNumber,a=e.getLineContent(r),l=this._findFirstMatchInLine(i,a,r,t.column,n);if(l)return l;for(let c=1;c<=o;c++){const d=(r+c-1)%o,h=e.getLineContent(d+1),u=this._findFirstMatchInLine(i,h,d+1,1,n);if(u)return u}return null}static _findFirstMatchInLine(e,t,i,n,o){e.reset(n-1);const r=e.next(t);return r?bf(new k(i,r.index+1,i,r.index+1+r[0].length),r,o):null}static findPreviousMatch(e,t,i,n){const o=t.parseSearchRequest();if(!o)return null;const r=new Db(o.wordSeparators,o.regex);return o.regex.multiline?this._doFindPreviousMatchMultiline(e,i,r,n):this._doFindPreviousMatchLineByLine(e,i,r,n)}static _doFindPreviousMatchMultiline(e,t,i,n){const o=this._doFindMatchesMultiline(e,new k(1,1,t.lineNumber,t.column),i,n,10*E1e);if(o.length>0)return o[o.length-1];const r=e.getLineCount();return t.lineNumber!==r||t.column!==e.getLineMaxColumn(r)?this._doFindPreviousMatchMultiline(e,new A(r,e.getLineMaxColumn(r)),i,n):null}static _doFindPreviousMatchLineByLine(e,t,i,n){const o=e.getLineCount(),r=t.lineNumber,a=e.getLineContent(r).substring(0,t.column-1),l=this._findLastMatchInLine(i,a,r,n);if(l)return l;for(let c=1;c<=o;c++){const d=(o+r-c-1)%o,h=e.getLineContent(d+1),u=this._findLastMatchInLine(i,h,d+1,n);if(u)return u}return null}static _findLastMatchInLine(e,t,i,n){let o=null,r;for(e.reset(0);r=e.next(t);)o=bf(new k(i,r.index+1,i,r.index+1+r[0].length),r,n);return o}}function I1e(s,e,t,i,n){if(i===0)return!0;const o=e.charCodeAt(i-1);if(s.get(o)!==0||o===13||o===10)return!0;if(n>0){const r=e.charCodeAt(i);if(s.get(r)!==0)return!0}return!1}function D1e(s,e,t,i,n){if(i+n===t)return!0;const o=e.charCodeAt(i+n);if(s.get(o)!==0||o===13||o===10)return!0;if(n>0){const r=e.charCodeAt(i+n-1);if(s.get(r)!==0)return!0}return!1}function d7(s,e,t,i,n){return I1e(s,e,t,i,n)&&D1e(s,e,t,i,n)}class Db{constructor(e,t){this._wordSeparators=e,this._searchRegex=t,this._prevMatchStartIndex=-1,this._prevMatchLength=0}reset(e){this._searchRegex.lastIndex=e,this._prevMatchStartIndex=-1,this._prevMatchLength=0}next(e){const t=e.length;let i;do{if(this._prevMatchStartIndex+this._prevMatchLength===t||(i=this._searchRegex.exec(e),!i))return null;const n=i.index,o=i[0].length;if(n===this._prevMatchStartIndex&&o===this._prevMatchLength){if(o===0){gE(e,t,this._searchRegex.lastIndex)>65535?this._searchRegex.lastIndex+=2:this._searchRegex.lastIndex+=1;continue}return null}if(this._prevMatchStartIndex=n,this._prevMatchLength=o,!this._wordSeparators||d7(this._wordSeparators,e,t,n,o))return i}while(i);return null}}const id=65535;function rte(s){let e;return s[s.length-1]<65536?e=new Uint16Array(s.length):e=new Uint32Array(s.length),e.set(s,0),e}class T1e{constructor(e,t,i,n,o){this.lineStarts=e,this.cr=t,this.lf=i,this.crlf=n,this.isBasicASCII=o}}function _d(s,e=!0){const t=[0];let i=1;for(let n=0,o=s.length;n<o;n++){const r=s.charCodeAt(n);r===13?n+1<o&&s.charCodeAt(n+1)===10?(t[i++]=n+2,n++):t[i++]=n+1:r===10&&(t[i++]=n+1)}return e?rte(t):t}function N1e(s,e){s.length=0,s[0]=0;let t=1,i=0,n=0,o=0,r=!0;for(let l=0,c=e.length;l<c;l++){const d=e.charCodeAt(l);d===13?l+1<c&&e.charCodeAt(l+1)===10?(o++,s[t++]=l+2,l++):(i++,s[t++]=l+1):d===10?(n++,s[t++]=l+1):r&&d!==9&&(d<32||d>126)&&(r=!1)}const a=new T1e(rte(s),i,n,o,r);return s.length=0,a}class Eo{constructor(e,t,i,n,o){this.bufferIndex=e,this.start=t,this.end=i,this.lineFeedCnt=n,this.length=o}}class wf{constructor(e,t){this.buffer=e,this.lineStarts=t}}class R1e{constructor(e,t){this._pieces=[],this._tree=e,this._BOM=t,this._index=0,e.root!==$e&&e.iterate(e.root,i=>(i!==$e&&this._pieces.push(i.piece),!0))}read(){return this._pieces.length===0?this._index===0?(this._index++,this._BOM):null:this._index>this._pieces.length-1?null:this._index===0?this._BOM+this._tree.getPieceContent(this._pieces[this._index++]):this._tree.getPieceContent(this._pieces[this._index++])}}class A1e{constructor(e){this._limit=e,this._cache=[]}get(e){for(let t=this._cache.length-1;t>=0;t--){const i=this._cache[t];if(i.nodeStartOffset<=e&&i.nodeStartOffset+i.node.piece.length>=e)return i}return null}get2(e){for(let t=this._cache.length-1;t>=0;t--){const i=this._cache[t];if(i.nodeStartLineNumber&&i.nodeStartLineNumber<e&&i.nodeStartLineNumber+i.node.piece.lineFeedCnt>=e)return i}return null}set(e){this._cache.length>=this._limit&&this._cache.shift(),this._cache.push(e)}validate(e){let t=!1;const i=this._cache;for(let n=0;n<i.length;n++){const o=i[n];if(o.node.parent===null||o.nodeStartOffset>=e){i[n]=null,t=!0;continue}}if(t){const n=[];for(const o of i)o!==null&&n.push(o);this._cache=n}}}class M1e{constructor(e,t,i){this.create(e,t,i)}create(e,t,i){this._buffers=[new wf("",[0])],this._lastChangeBufferPos={line:0,column:0},this.root=$e,this._lineCnt=1,this._length=0,this._EOL=t,this._EOLLength=t.length,this._EOLNormalized=i;let n=null;for(let o=0,r=e.length;o<r;o++)if(e[o].buffer.length>0){e[o].lineStarts||(e[o].lineStarts=_d(e[o].buffer));const a=new Eo(o+1,{line:0,column:0},{line:e[o].lineStarts.length-1,column:e[o].buffer.length-e[o].lineStarts[e[o].lineStarts.length-1]},e[o].lineStarts.length-1,e[o].buffer.length);this._buffers.push(e[o]),n=this.rbInsertRight(n,a)}this._searchCache=new A1e(1),this._lastVisitedLine={lineNumber:0,value:""},this.computeBufferMetadata()}normalizeEOL(e){const t=id,i=t-Math.floor(t/3),n=i*2;let o="",r=0;const a=[];if(this.iterate(this.root,l=>{const c=this.getNodeContent(l),d=c.length;if(r<=i||r+d<n)return o+=c,r+=d,!0;const h=o.replace(/\r\n|\r|\n/g,e);return a.push(new wf(h,_d(h))),o=c,r=d,!0}),r>0){const l=o.replace(/\r\n|\r|\n/g,e);a.push(new wf(l,_d(l)))}this.create(a,e,!0)}getEOL(){return this._EOL}setEOL(e){this._EOL=e,this._EOLLength=this._EOL.length,this.normalizeEOL(e)}createSnapshot(e){return new R1e(this,e)}getOffsetAt(e,t){let i=0,n=this.root;for(;n!==$e;)if(n.left!==$e&&n.lf_left+1>=e)n=n.left;else if(n.lf_left+n.piece.lineFeedCnt+1>=e){i+=n.size_left;const o=this.getAccumulatedValue(n,e-n.lf_left-2);return i+=o+t-1}else e-=n.lf_left+n.piece.lineFeedCnt,i+=n.size_left+n.piece.length,n=n.right;return i}getPositionAt(e){e=Math.floor(e),e=Math.max(0,e);let t=this.root,i=0;const n=e;for(;t!==$e;)if(t.size_left!==0&&t.size_left>=e)t=t.left;else if(t.size_left+t.piece.length>=e){const o=this.getIndexOf(t,e-t.size_left);if(i+=t.lf_left+o.index,o.index===0){const r=this.getOffsetAt(i+1,1),a=n-r;return new A(i+1,a+1)}return new A(i+1,o.remainder+1)}else if(e-=t.size_left+t.piece.length,i+=t.lf_left+t.piece.lineFeedCnt,t.right===$e){const o=this.getOffsetAt(i+1,1),r=n-e-o;return new A(i+1,r+1)}else t=t.right;return new A(1,1)}getValueInRange(e,t){if(e.startLineNumber===e.endLineNumber&&e.startColumn===e.endColumn)return"";const i=this.nodeAt2(e.startLineNumber,e.startColumn),n=this.nodeAt2(e.endLineNumber,e.endColumn),o=this.getValueInRange2(i,n);return t?t!==this._EOL||!this._EOLNormalized?o.replace(/\r\n|\r|\n/g,t):t===this.getEOL()&&this._EOLNormalized?o:o.replace(/\r\n|\r|\n/g,t):o}getValueInRange2(e,t){if(e.node===t.node){const a=e.node,l=this._buffers[a.piece.bufferIndex].buffer,c=this.offsetInBuffer(a.piece.bufferIndex,a.piece.start);return l.substring(c+e.remainder,c+t.remainder)}let i=e.node;const n=this._buffers[i.piece.bufferIndex].buffer,o=this.offsetInBuffer(i.piece.bufferIndex,i.piece.start);let r=n.substring(o+e.remainder,o+i.piece.length);for(i=i.next();i!==$e;){const a=this._buffers[i.piece.bufferIndex].buffer,l=this.offsetInBuffer(i.piece.bufferIndex,i.piece.start);if(i===t.node){r+=a.substring(l,l+t.remainder);break}else r+=a.substr(l,i.piece.length);i=i.next()}return r}getLinesContent(){const e=[];let t=0,i="",n=!1;return this.iterate(this.root,o=>{if(o===$e)return!0;const r=o.piece;let a=r.length;if(a===0)return!0;const l=this._buffers[r.bufferIndex].buffer,c=this._buffers[r.bufferIndex].lineStarts,d=r.start.line,h=r.end.line;let u=c[d]+r.start.column;if(n&&(l.charCodeAt(u)===10&&(u++,a--),e[t++]=i,i="",n=!1,a===0))return!0;if(d===h)return!this._EOLNormalized&&l.charCodeAt(u+a-1)===13?(n=!0,i+=l.substr(u,a-1)):i+=l.substr(u,a),!0;i+=this._EOLNormalized?l.substring(u,Math.max(u,c[d+1]-this._EOLLength)):l.substring(u,c[d+1]).replace(/(\r\n|\r|\n)$/,""),e[t++]=i;for(let g=d+1;g<h;g++)i=this._EOLNormalized?l.substring(c[g],c[g+1]-this._EOLLength):l.substring(c[g],c[g+1]).replace(/(\r\n|\r|\n)$/,""),e[t++]=i;return!this._EOLNormalized&&l.charCodeAt(c[h]+r.end.column-1)===13?(n=!0,r.end.column===0?t--:i=l.substr(c[h],r.end.column-1)):i=l.substr(c[h],r.end.column),!0}),n&&(e[t++]=i,i=""),e[t++]=i,e}getLength(){return this._length}getLineCount(){return this._lineCnt}getLineContent(e){return this._lastVisitedLine.lineNumber===e?this._lastVisitedLine.value:(this._lastVisitedLine.lineNumber=e,e===this._lineCnt?this._lastVisitedLine.value=this.getLineRawContent(e):this._EOLNormalized?this._lastVisitedLine.value=this.getLineRawContent(e,this._EOLLength):this._lastVisitedLine.value=this.getLineRawContent(e).replace(/(\r\n|\r|\n)$/,""),this._lastVisitedLine.value)}_getCharCode(e){if(e.remainder===e.node.piece.length){const t=e.node.next();if(!t)return 0;const i=this._buffers[t.piece.bufferIndex],n=this.offsetInBuffer(t.piece.bufferIndex,t.piece.start);return i.buffer.charCodeAt(n)}else{const t=this._buffers[e.node.piece.bufferIndex],n=this.offsetInBuffer(e.node.piece.bufferIndex,e.node.piece.start)+e.remainder;return t.buffer.charCodeAt(n)}}getLineCharCode(e,t){const i=this.nodeAt2(e,t+1);return this._getCharCode(i)}getLineLength(e){if(e===this.getLineCount()){const t=this.getOffsetAt(e,1);return this.getLength()-t}return this.getOffsetAt(e+1,1)-this.getOffsetAt(e,1)-this._EOLLength}getNearestChunk(e){const t=this.nodeAt(e);if(t.remainder===t.node.piece.length){const i=t.node.next();if(!i||i===$e)return"";const n=this._buffers[i.piece.bufferIndex],o=this.offsetInBuffer(i.piece.bufferIndex,i.piece.start);return n.buffer.substring(o,o+i.piece.length)}else{const i=this._buffers[t.node.piece.bufferIndex],n=this.offsetInBuffer(t.node.piece.bufferIndex,t.node.piece.start),o=n+t.remainder,r=n+t.node.piece.length;return i.buffer.substring(o,r)}}findMatchesInNode(e,t,i,n,o,r,a,l,c,d,h){const u=this._buffers[e.piece.bufferIndex],g=this.offsetInBuffer(e.piece.bufferIndex,e.piece.start),f=this.offsetInBuffer(e.piece.bufferIndex,o),p=this.offsetInBuffer(e.piece.bufferIndex,r);let _;const b={line:0,column:0};let w,C;t._wordSeparators?(w=u.buffer.substring(f,p),C=v=>v+f,t.reset(0)):(w=u.buffer,C=v=>v,t.reset(f));do if(_=t.next(w),_){if(C(_.index)>=p)return d;this.positionInBuffer(e,C(_.index)-g,b);const v=this.getLineFeedCnt(e.piece.bufferIndex,o,b),y=b.line===o.line?b.column-o.column+n:b.column+1,x=y+_[0].length;if(h[d++]=bf(new k(i+v,y,i+v,x),_,l),C(_.index)+_[0].length>=p||d>=c)return d}while(_);return d}findMatchesLineByLine(e,t,i,n){const o=[];let r=0;const a=new Db(t.wordSeparators,t.regex);let l=this.nodeAt2(e.startLineNumber,e.startColumn);if(l===null)return[];const c=this.nodeAt2(e.endLineNumber,e.endColumn);if(c===null)return[];let d=this.positionInBuffer(l.node,l.remainder);const h=this.positionInBuffer(c.node,c.remainder);if(l.node===c.node)return this.findMatchesInNode(l.node,a,e.startLineNumber,e.startColumn,d,h,t,i,n,r,o),o;let u=e.startLineNumber,g=l.node;for(;g!==c.node;){const p=this.getLineFeedCnt(g.piece.bufferIndex,d,g.piece.end);if(p>=1){const b=this._buffers[g.piece.bufferIndex].lineStarts,w=this.offsetInBuffer(g.piece.bufferIndex,g.piece.start),C=b[d.line+p],v=u===e.startLineNumber?e.startColumn:1;if(r=this.findMatchesInNode(g,a,u,v,d,this.positionInBuffer(g,C-w),t,i,n,r,o),r>=n)return o;u+=p}const _=u===e.startLineNumber?e.startColumn-1:0;if(u===e.endLineNumber){const b=this.getLineContent(u).substring(_,e.endColumn-1);return r=this._findMatchesInLine(t,a,b,e.endLineNumber,_,r,o,i,n),o}if(r=this._findMatchesInLine(t,a,this.getLineContent(u).substr(_),u,_,r,o,i,n),r>=n)return o;u++,l=this.nodeAt2(u,1),g=l.node,d=this.positionInBuffer(l.node,l.remainder)}if(u===e.endLineNumber){const p=u===e.startLineNumber?e.startColumn-1:0,_=this.getLineContent(u).substring(p,e.endColumn-1);return r=this._findMatchesInLine(t,a,_,e.endLineNumber,p,r,o,i,n),o}const f=u===e.startLineNumber?e.startColumn:1;return r=this.findMatchesInNode(c.node,a,u,f,d,h,t,i,n,r,o),o}_findMatchesInLine(e,t,i,n,o,r,a,l,c){const d=e.wordSeparators;if(!l&&e.simpleSearch){const u=e.simpleSearch,g=u.length,f=i.length;let p=-g;for(;(p=i.indexOf(u,p+g))!==-1;)if((!d||d7(d,i,f,p,g))&&(a[r++]=new tS(new k(n,p+1+o,n,p+1+g+o),null),r>=c))return r;return r}let h;t.reset(0);do if(h=t.next(i),h&&(a[r++]=bf(new k(n,h.index+1+o,n,h.index+1+h[0].length+o),h,l),r>=c))return r;while(h);return r}insert(e,t,i=!1){if(this._EOLNormalized=this._EOLNormalized&&i,this._lastVisitedLine.lineNumber=0,this._lastVisitedLine.value="",this.root!==$e){const{node:n,remainder:o,nodeStartOffset:r}=this.nodeAt(e),a=n.piece,l=a.bufferIndex,c=this.positionInBuffer(n,o);if(n.piece.bufferIndex===0&&a.end.line===this._lastChangeBufferPos.line&&a.end.column===this._lastChangeBufferPos.column&&r+a.length===e&&t.length<id){this.appendToNode(n,t),this.computeBufferMetadata();return}if(r===e)this.insertContentToNodeLeft(t,n),this._searchCache.validate(e);else if(r+n.piece.length>e){const d=[];let h=new Eo(a.bufferIndex,c,a.end,this.getLineFeedCnt(a.bufferIndex,c,a.end),this.offsetInBuffer(l,a.end)-this.offsetInBuffer(l,c));if(this.shouldCheckCRLF()&&this.endWithCR(t)&&this.nodeCharCodeAt(n,o)===10){const p={line:h.start.line+1,column:0};h=new Eo(h.bufferIndex,p,h.end,this.getLineFeedCnt(h.bufferIndex,p,h.end),h.length-1),t+=`
`}if(this.shouldCheckCRLF()&&this.startWithLF(t))if(this.nodeCharCodeAt(n,o-1)===13){const p=this.positionInBuffer(n,o-1);this.deleteNodeTail(n,p),t="\r"+t,n.piece.length===0&&d.push(n)}else this.deleteNodeTail(n,c);else this.deleteNodeTail(n,c);const u=this.createNewPieces(t);h.length>0&&this.rbInsertRight(n,h);let g=n;for(let f=0;f<u.length;f++)g=this.rbInsertRight(g,u[f]);this.deleteNodes(d)}else this.insertContentToNodeRight(t,n)}else{const n=this.createNewPieces(t);let o=this.rbInsertLeft(null,n[0]);for(let r=1;r<n.length;r++)o=this.rbInsertRight(o,n[r])}this.computeBufferMetadata()}delete(e,t){if(this._lastVisitedLine.lineNumber=0,this._lastVisitedLine.value="",t<=0||this.root===$e)return;const i=this.nodeAt(e),n=this.nodeAt(e+t),o=i.node,r=n.node;if(o===r){const u=this.positionInBuffer(o,i.remainder),g=this.positionInBuffer(o,n.remainder);if(i.nodeStartOffset===e){if(t===o.piece.length){const f=o.next();aL(this,o),this.validateCRLFWithPrevNode(f),this.computeBufferMetadata();return}this.deleteNodeHead(o,g),this._searchCache.validate(e),this.validateCRLFWithPrevNode(o),this.computeBufferMetadata();return}if(i.nodeStartOffset+o.piece.length===e+t){this.deleteNodeTail(o,u),this.validateCRLFWithNextNode(o),this.computeBufferMetadata();return}this.shrinkNode(o,u,g),this.computeBufferMetadata();return}const a=[],l=this.positionInBuffer(o,i.remainder);this.deleteNodeTail(o,l),this._searchCache.validate(e),o.piece.length===0&&a.push(o);const c=this.positionInBuffer(r,n.remainder);this.deleteNodeHead(r,c),r.piece.length===0&&a.push(r);const d=o.next();for(let u=d;u!==$e&&u!==r;u=u.next())a.push(u);const h=o.piece.length===0?o.prev():o;this.deleteNodes(a),this.validateCRLFWithNextNode(h),this.computeBufferMetadata()}insertContentToNodeLeft(e,t){const i=[];if(this.shouldCheckCRLF()&&this.endWithCR(e)&&this.startWithLF(t)){const r=t.piece,a={line:r.start.line+1,column:0},l=new Eo(r.bufferIndex,a,r.end,this.getLineFeedCnt(r.bufferIndex,a,r.end),r.length-1);t.piece=l,e+=`
`,gd(this,t,-1,-1),t.piece.length===0&&i.push(t)}const n=this.createNewPieces(e);let o=this.rbInsertLeft(t,n[n.length-1]);for(let r=n.length-2;r>=0;r--)o=this.rbInsertLeft(o,n[r]);this.validateCRLFWithPrevNode(o),this.deleteNodes(i)}insertContentToNodeRight(e,t){this.adjustCarriageReturnFromNext(e,t)&&(e+=`
`);const i=this.createNewPieces(e),n=this.rbInsertRight(t,i[0]);let o=n;for(let r=1;r<i.length;r++)o=this.rbInsertRight(o,i[r]);this.validateCRLFWithPrevNode(n)}positionInBuffer(e,t,i){const n=e.piece,o=e.piece.bufferIndex,r=this._buffers[o].lineStarts,l=r[n.start.line]+n.start.column+t;let c=n.start.line,d=n.end.line,h=0,u=0,g=0;for(;c<=d&&(h=c+(d-c)/2|0,g=r[h],h!==d);)if(u=r[h+1],l<g)d=h-1;else if(l>=u)c=h+1;else break;return i?(i.line=h,i.column=l-g,null):{line:h,column:l-g}}getLineFeedCnt(e,t,i){if(i.column===0)return i.line-t.line;const n=this._buffers[e].lineStarts;if(i.line===n.length-1)return i.line-t.line;const o=n[i.line+1],r=n[i.line]+i.column;if(o>r+1)return i.line-t.line;const a=r-1;return this._buffers[e].buffer.charCodeAt(a)===13?i.line-t.line+1:i.line-t.line}offsetInBuffer(e,t){return this._buffers[e].lineStarts[t.line]+t.column}deleteNodes(e){for(let t=0;t<e.length;t++)aL(this,e[t])}createNewPieces(e){if(e.length>id){const d=[];for(;e.length>id;){const u=e.charCodeAt(id-1);let g;u===13||u>=55296&&u<=56319?(g=e.substring(0,id-1),e=e.substring(id-1)):(g=e.substring(0,id),e=e.substring(id));const f=_d(g);d.push(new Eo(this._buffers.length,{line:0,column:0},{line:f.length-1,column:g.length-f[f.length-1]},f.length-1,g.length)),this._buffers.push(new wf(g,f))}const h=_d(e);return d.push(new Eo(this._buffers.length,{line:0,column:0},{line:h.length-1,column:e.length-h[h.length-1]},h.length-1,e.length)),this._buffers.push(new wf(e,h)),d}let t=this._buffers[0].buffer.length;const i=_d(e,!1);let n=this._lastChangeBufferPos;if(this._buffers[0].lineStarts[this._buffers[0].lineStarts.length-1]===t&&t!==0&&this.startWithLF(e)&&this.endWithCR(this._buffers[0].buffer)){this._lastChangeBufferPos={line:this._lastChangeBufferPos.line,column:this._lastChangeBufferPos.column+1},n=this._lastChangeBufferPos;for(let d=0;d<i.length;d++)i[d]+=t+1;this._buffers[0].lineStarts=this._buffers[0].lineStarts.concat(i.slice(1)),this._buffers[0].buffer+="_"+e,t+=1}else{if(t!==0)for(let d=0;d<i.length;d++)i[d]+=t;this._buffers[0].lineStarts=this._buffers[0].lineStarts.concat(i.slice(1)),this._buffers[0].buffer+=e}const o=this._buffers[0].buffer.length,r=this._buffers[0].lineStarts.length-1,a=o-this._buffers[0].lineStarts[r],l={line:r,column:a},c=new Eo(0,n,l,this.getLineFeedCnt(0,n,l),o-t);return this._lastChangeBufferPos=l,[c]}getLineRawContent(e,t=0){let i=this.root,n="";const o=this._searchCache.get2(e);if(o){i=o.node;const r=this.getAccumulatedValue(i,e-o.nodeStartLineNumber-1),a=this._buffers[i.piece.bufferIndex].buffer,l=this.offsetInBuffer(i.piece.bufferIndex,i.piece.start);if(o.nodeStartLineNumber+i.piece.lineFeedCnt===e)n=a.substring(l+r,l+i.piece.length);else{const c=this.getAccumulatedValue(i,e-o.nodeStartLineNumber);return a.substring(l+r,l+c-t)}}else{let r=0;const a=e;for(;i!==$e;)if(i.left!==$e&&i.lf_left>=e-1)i=i.left;else if(i.lf_left+i.piece.lineFeedCnt>e-1){const l=this.getAccumulatedValue(i,e-i.lf_left-2),c=this.getAccumulatedValue(i,e-i.lf_left-1),d=this._buffers[i.piece.bufferIndex].buffer,h=this.offsetInBuffer(i.piece.bufferIndex,i.piece.start);return r+=i.size_left,this._searchCache.set({node:i,nodeStartOffset:r,nodeStartLineNumber:a-(e-1-i.lf_left)}),d.substring(h+l,h+c-t)}else if(i.lf_left+i.piece.lineFeedCnt===e-1){const l=this.getAccumulatedValue(i,e-i.lf_left-2),c=this._buffers[i.piece.bufferIndex].buffer,d=this.offsetInBuffer(i.piece.bufferIndex,i.piece.start);n=c.substring(d+l,d+i.piece.length);break}else e-=i.lf_left+i.piece.lineFeedCnt,r+=i.size_left+i.piece.length,i=i.right}for(i=i.next();i!==$e;){const r=this._buffers[i.piece.bufferIndex].buffer;if(i.piece.lineFeedCnt>0){const a=this.getAccumulatedValue(i,0),l=this.offsetInBuffer(i.piece.bufferIndex,i.piece.start);return n+=r.substring(l,l+a-t),n}else{const a=this.offsetInBuffer(i.piece.bufferIndex,i.piece.start);n+=r.substr(a,i.piece.length)}i=i.next()}return n}computeBufferMetadata(){let e=this.root,t=1,i=0;for(;e!==$e;)t+=e.lf_left+e.piece.lineFeedCnt,i+=e.size_left+e.piece.length,e=e.right;this._lineCnt=t,this._length=i,this._searchCache.validate(this._length)}getIndexOf(e,t){const i=e.piece,n=this.positionInBuffer(e,t),o=n.line-i.start.line;if(this.offsetInBuffer(i.bufferIndex,i.end)-this.offsetInBuffer(i.bufferIndex,i.start)===t){const r=this.getLineFeedCnt(e.piece.bufferIndex,i.start,n);if(r!==o)return{index:r,remainder:0}}return{index:o,remainder:n.column}}getAccumulatedValue(e,t){if(t<0)return 0;const i=e.piece,n=this._buffers[i.bufferIndex].lineStarts,o=i.start.line+t+1;return o>i.end.line?n[i.end.line]+i.end.column-n[i.start.line]-i.start.column:n[o]-n[i.start.line]-i.start.column}deleteNodeTail(e,t){const i=e.piece,n=i.lineFeedCnt,o=this.offsetInBuffer(i.bufferIndex,i.end),r=t,a=this.offsetInBuffer(i.bufferIndex,r),l=this.getLineFeedCnt(i.bufferIndex,i.start,r),c=l-n,d=a-o,h=i.length+d;e.piece=new Eo(i.bufferIndex,i.start,r,l,h),gd(this,e,d,c)}deleteNodeHead(e,t){const i=e.piece,n=i.lineFeedCnt,o=this.offsetInBuffer(i.bufferIndex,i.start),r=t,a=this.getLineFeedCnt(i.bufferIndex,r,i.end),l=this.offsetInBuffer(i.bufferIndex,r),c=a-n,d=o-l,h=i.length+d;e.piece=new Eo(i.bufferIndex,r,i.end,a,h),gd(this,e,d,c)}shrinkNode(e,t,i){const n=e.piece,o=n.start,r=n.end,a=n.length,l=n.lineFeedCnt,c=t,d=this.getLineFeedCnt(n.bufferIndex,n.start,c),h=this.offsetInBuffer(n.bufferIndex,t)-this.offsetInBuffer(n.bufferIndex,o);e.piece=new Eo(n.bufferIndex,n.start,c,d,h),gd(this,e,h-a,d-l);const u=new Eo(n.bufferIndex,i,r,this.getLineFeedCnt(n.bufferIndex,i,r),this.offsetInBuffer(n.bufferIndex,r)-this.offsetInBuffer(n.bufferIndex,i)),g=this.rbInsertRight(e,u);this.validateCRLFWithPrevNode(g)}appendToNode(e,t){this.adjustCarriageReturnFromNext(t,e)&&(t+=`
`);const i=this.shouldCheckCRLF()&&this.startWithLF(t)&&this.endWithCR(e),n=this._buffers[0].buffer.length;this._buffers[0].buffer+=t;const o=_d(t,!1);for(let g=0;g<o.length;g++)o[g]+=n;if(i){const g=this._buffers[0].lineStarts[this._buffers[0].lineStarts.length-2];this._buffers[0].lineStarts.pop(),this._lastChangeBufferPos={line:this._lastChangeBufferPos.line-1,column:n-g}}this._buffers[0].lineStarts=this._buffers[0].lineStarts.concat(o.slice(1));const r=this._buffers[0].lineStarts.length-1,a=this._buffers[0].buffer.length-this._buffers[0].lineStarts[r],l={line:r,column:a},c=e.piece.length+t.length,d=e.piece.lineFeedCnt,h=this.getLineFeedCnt(0,e.piece.start,l),u=h-d;e.piece=new Eo(e.piece.bufferIndex,e.piece.start,l,h,c),this._lastChangeBufferPos=l,gd(this,e,t.length,u)}nodeAt(e){let t=this.root;const i=this._searchCache.get(e);if(i)return{node:i.node,nodeStartOffset:i.nodeStartOffset,remainder:e-i.nodeStartOffset};let n=0;for(;t!==$e;)if(t.size_left>e)t=t.left;else if(t.size_left+t.piece.length>=e){n+=t.size_left;const o={node:t,remainder:e-t.size_left,nodeStartOffset:n};return this._searchCache.set(o),o}else e-=t.size_left+t.piece.length,n+=t.size_left+t.piece.length,t=t.right;return null}nodeAt2(e,t){let i=this.root,n=0;for(;i!==$e;)if(i.left!==$e&&i.lf_left>=e-1)i=i.left;else if(i.lf_left+i.piece.lineFeedCnt>e-1){const o=this.getAccumulatedValue(i,e-i.lf_left-2),r=this.getAccumulatedValue(i,e-i.lf_left-1);return n+=i.size_left,{node:i,remainder:Math.min(o+t-1,r),nodeStartOffset:n}}else if(i.lf_left+i.piece.lineFeedCnt===e-1){const o=this.getAccumulatedValue(i,e-i.lf_left-2);if(o+t-1<=i.piece.length)return{node:i,remainder:o+t-1,nodeStartOffset:n};t-=i.piece.length-o;break}else e-=i.lf_left+i.piece.lineFeedCnt,n+=i.size_left+i.piece.length,i=i.right;for(i=i.next();i!==$e;){if(i.piece.lineFeedCnt>0){const o=this.getAccumulatedValue(i,0),r=this.offsetOfNode(i);return{node:i,remainder:Math.min(t-1,o),nodeStartOffset:r}}else if(i.piece.length>=t-1){const o=this.offsetOfNode(i);return{node:i,remainder:t-1,nodeStartOffset:o}}else t-=i.piece.length;i=i.next()}return null}nodeCharCodeAt(e,t){if(e.piece.lineFeedCnt<1)return-1;const i=this._buffers[e.piece.bufferIndex],n=this.offsetInBuffer(e.piece.bufferIndex,e.piece.start)+t;return i.buffer.charCodeAt(n)}offsetOfNode(e){if(!e)return 0;let t=e.size_left;for(;e!==this.root;)e.parent.right===e&&(t+=e.parent.size_left+e.parent.piece.length),e=e.parent;return t}shouldCheckCRLF(){return!(this._EOLNormalized&&this._EOL===`
`)}startWithLF(e){if(typeof e=="string")return e.charCodeAt(0)===10;if(e===$e||e.piece.lineFeedCnt===0)return!1;const t=e.piece,i=this._buffers[t.bufferIndex].lineStarts,n=t.start.line,o=i[n]+t.start.column;return n===i.length-1||i[n+1]>o+1?!1:this._buffers[t.bufferIndex].buffer.charCodeAt(o)===10}endWithCR(e){return typeof e=="string"?e.charCodeAt(e.length-1)===13:e===$e||e.piece.lineFeedCnt===0?!1:this.nodeCharCodeAt(e,e.piece.length-1)===13}validateCRLFWithPrevNode(e){if(this.shouldCheckCRLF()&&this.startWithLF(e)){const t=e.prev();this.endWithCR(t)&&this.fixCRLF(t,e)}}validateCRLFWithNextNode(e){if(this.shouldCheckCRLF()&&this.endWithCR(e)){const t=e.next();this.startWithLF(t)&&this.fixCRLF(e,t)}}fixCRLF(e,t){const i=[],n=this._buffers[e.piece.bufferIndex].lineStarts;let o;e.piece.end.column===0?o={line:e.piece.end.line-1,column:n[e.piece.end.line]-n[e.piece.end.line-1]-1}:o={line:e.piece.end.line,column:e.piece.end.column-1};const r=e.piece.length-1,a=e.piece.lineFeedCnt-1;e.piece=new Eo(e.piece.bufferIndex,e.piece.start,o,a,r),gd(this,e,-1,-1),e.piece.length===0&&i.push(e);const l={line:t.piece.start.line+1,column:0},c=t.piece.length-1,d=this.getLineFeedCnt(t.piece.bufferIndex,l,t.piece.end);t.piece=new Eo(t.piece.bufferIndex,l,t.piece.end,d,c),gd(this,t,-1,-1),t.piece.length===0&&i.push(t);const h=this.createNewPieces(`\r
`);this.rbInsertRight(e,h[0]);for(let u=0;u<i.length;u++)aL(this,i[u])}adjustCarriageReturnFromNext(e,t){if(this.shouldCheckCRLF()&&this.endWithCR(e)){const i=t.next();if(this.startWithLF(i)){if(e+=`
`,i.piece.length===1)aL(this,i);else{const n=i.piece,o={line:n.start.line+1,column:0},r=n.length-1,a=this.getLineFeedCnt(n.bufferIndex,o,n.end);i.piece=new Eo(n.bufferIndex,o,n.end,a,r),gd(this,i,-1,-1)}return!0}}return!1}iterate(e,t){if(e===$e)return t($e);const i=this.iterate(e.left,t);return i&&t(e)&&this.iterate(e.right,t)}getNodeContent(e){if(e===$e)return"";const t=this._buffers[e.piece.bufferIndex],i=e.piece,n=this.offsetInBuffer(i.bufferIndex,i.start),o=this.offsetInBuffer(i.bufferIndex,i.end);return t.buffer.substring(n,o)}getPieceContent(e){const t=this._buffers[e.bufferIndex],i=this.offsetInBuffer(e.bufferIndex,e.start),n=this.offsetInBuffer(e.bufferIndex,e.end);return t.buffer.substring(i,n)}rbInsertRight(e,t){const i=new AF(t,1);if(i.left=$e,i.right=$e,i.parent=$e,i.size_left=0,i.lf_left=0,this.root===$e)this.root=i,i.color=0;else if(e.right===$e)e.right=i,i.parent=e;else{const o=a7(e.right);o.left=i,i.parent=o}return IU(this,i),i}rbInsertLeft(e,t){const i=new AF(t,1);if(i.left=$e,i.right=$e,i.parent=$e,i.size_left=0,i.lf_left=0,this.root===$e)this.root=i,i.color=0;else if(e.left===$e)e.left=i,i.parent=e;else{const n=ste(e.left);n.right=i,i.parent=n}return IU(this,i),i}}class vw extends F{constructor(e,t,i,n,o,r,a){super(),this._onDidChangeContent=this._register(new M),this._BOM=t,this._mightContainNonBasicASCII=!r,this._mightContainRTL=n,this._mightContainUnusualLineTerminators=o,this._pieceTree=new M1e(e,i,a)}mightContainRTL(){return this._mightContainRTL}mightContainUnusualLineTerminators(){return this._mightContainUnusualLineTerminators}resetMightContainUnusualLineTerminators(){this._mightContainUnusualLineTerminators=!1}mightContainNonBasicASCII(){return this._mightContainNonBasicASCII}getBOM(){return this._BOM}getEOL(){return this._pieceTree.getEOL()}createSnapshot(e){return this._pieceTree.createSnapshot(e?this._BOM:"")}getOffsetAt(e,t){return this._pieceTree.getOffsetAt(e,t)}getPositionAt(e){return this._pieceTree.getPositionAt(e)}getRangeAt(e,t){const i=e+t,n=this.getPositionAt(e),o=this.getPositionAt(i);return new k(n.lineNumber,n.column,o.lineNumber,o.column)}getValueInRange(e,t=0){if(e.isEmpty())return"";const i=this._getEndOfLine(t);return this._pieceTree.getValueInRange(e,i)}getValueLengthInRange(e,t=0){if(e.isEmpty())return 0;if(e.startLineNumber===e.endLineNumber)return e.endColumn-e.startColumn;const i=this.getOffsetAt(e.startLineNumber,e.startColumn),n=this.getOffsetAt(e.endLineNumber,e.endColumn);let o=0;const r=this._getEndOfLine(t),a=this.getEOL();if(r.length!==a.length){const l=r.length-a.length,c=e.endLineNumber-e.startLineNumber;o=l*c}return n-i+o}getCharacterCountInRange(e,t=0){if(this._mightContainNonBasicASCII){let i=0;const n=e.startLineNumber,o=e.endLineNumber;for(let r=n;r<=o;r++){const a=this.getLineContent(r),l=r===n?e.startColumn-1:0,c=r===o?e.endColumn-1:a.length;for(let d=l;d<c;d++)sn(a.charCodeAt(d))?(i=i+1,d=d+1):i=i+1}return i+=this._getEndOfLine(t).length*(o-n),i}return this.getValueLengthInRange(e,t)}getNearestChunk(e){return this._pieceTree.getNearestChunk(e)}getLength(){return this._pieceTree.getLength()}getLineCount(){return this._pieceTree.getLineCount()}getLinesContent(){return this._pieceTree.getLinesContent()}getLineContent(e){return this._pieceTree.getLineContent(e)}getLineCharCode(e,t){return this._pieceTree.getLineCharCode(e,t)}getLineLength(e){return this._pieceTree.getLineLength(e)}getLineFirstNonWhitespaceColumn(e){const t=As(this.getLineContent(e));return t===-1?0:t+1}getLineLastNonWhitespaceColumn(e){const t=Ma(this.getLineContent(e));return t===-1?0:t+2}_getEndOfLine(e){switch(e){case 1:return`
`;case 2:return`\r
`;case 0:return this.getEOL();default:throw new Error("Unknown EOL preference")}}setEOL(e){this._pieceTree.setEOL(e)}applyEdits(e,t,i){let n=this._mightContainRTL,o=this._mightContainUnusualLineTerminators,r=this._mightContainNonBasicASCII,a=!0,l=[];for(let p=0;p<e.length;p++){const _=e[p];a&&_._isTracked&&(a=!1);const b=_.range;if(_.text){let x=!0;r||(x=!q0(_.text),r=x),!n&&x&&(n=iv(_.text)),!o&&x&&(o=nQ(_.text))}let w="",C=0,v=0,y=0;if(_.text){let x;[C,v,y,x]=fg(_.text);const L=this.getEOL();x===0||x===(L===`\r
`?2:1)?w=_.text:w=_.text.replace(/\r\n|\r|\n/g,L)}l[p]={sortIndex:p,identifier:_.identifier||null,range:b,rangeOffset:this.getOffsetAt(b.startLineNumber,b.startColumn),rangeLength:this.getValueLengthInRange(b),text:w,eolCount:C,firstLineLength:v,lastLineLength:y,forceMoveMarkers:!!_.forceMoveMarkers,isAutoWhitespaceEdit:_.isAutoWhitespaceEdit||!1}}l.sort(vw._sortOpsAscending);let c=!1;for(let p=0,_=l.length-1;p<_;p++){const b=l[p].range.getEndPosition(),w=l[p+1].range.getStartPosition();if(w.isBeforeOrEqual(b)){if(w.isBefore(b))throw new Error("Overlapping ranges are not allowed!");c=!0}}a&&(l=this._reduceOperations(l));const d=i||t?vw._getInverseEditRanges(l):[],h=[];if(t)for(let p=0;p<l.length;p++){const _=l[p],b=d[p];if(_.isAutoWhitespaceEdit&&_.range.isEmpty())for(let w=b.startLineNumber;w<=b.endLineNumber;w++){let C="";w===b.startLineNumber&&(C=this.getLineContent(_.range.startLineNumber),As(C)!==-1)||h.push({lineNumber:w,oldContent:C})}}let u=null;if(i){let p=0;u=[];for(let _=0;_<l.length;_++){const b=l[_],w=d[_],C=this.getValueInRange(b.range),v=b.rangeOffset+p;p+=b.text.length-C.length,u[_]={sortIndex:b.sortIndex,identifier:b.identifier,range:w,text:C,textChange:new Xn(b.rangeOffset,C,v,b.text)}}c||u.sort((_,b)=>_.sortIndex-b.sortIndex)}this._mightContainRTL=n,this._mightContainUnusualLineTerminators=o,this._mightContainNonBasicASCII=r;const g=this._doApplyEdits(l);let f=null;if(t&&h.length>0){h.sort((p,_)=>_.lineNumber-p.lineNumber),f=[];for(let p=0,_=h.length;p<_;p++){const b=h[p].lineNumber;if(p>0&&h[p-1].lineNumber===b)continue;const w=h[p].oldContent,C=this.getLineContent(b);C.length===0||C===w||As(C)!==-1||f.push(b)}}return this._onDidChangeContent.fire(),new xge(u,g,f)}_reduceOperations(e){return e.length<1e3?e:[this._toSingleEditOperation(e)]}_toSingleEditOperation(e){let t=!1;const i=e[0].range,n=e[e.length-1].range,o=new k(i.startLineNumber,i.startColumn,n.endLineNumber,n.endColumn);let r=i.startLineNumber,a=i.startColumn;const l=[];for(let g=0,f=e.length;g<f;g++){const p=e[g],_=p.range;t=t||p.forceMoveMarkers,l.push(this.getValueInRange(new k(r,a,_.startLineNumber,_.startColumn))),p.text.length>0&&l.push(p.text),r=_.endLineNumber,a=_.endColumn}const c=l.join(""),[d,h,u]=fg(c);return{sortIndex:0,identifier:e[0].identifier,range:o,rangeOffset:this.getOffsetAt(o.startLineNumber,o.startColumn),rangeLength:this.getValueLengthInRange(o,0),text:c,eolCount:d,firstLineLength:h,lastLineLength:u,forceMoveMarkers:t,isAutoWhitespaceEdit:!1}}_doApplyEdits(e){e.sort(vw._sortOpsDescending);const t=[];for(let i=0;i<e.length;i++){const n=e[i],o=n.range.startLineNumber,r=n.range.startColumn,a=n.range.endLineNumber,l=n.range.endColumn;if(o===a&&r===l&&n.text.length===0)continue;n.text?(this._pieceTree.delete(n.rangeOffset,n.rangeLength),this._pieceTree.insert(n.rangeOffset,n.text,!0)):this._pieceTree.delete(n.rangeOffset,n.rangeLength);const c=new k(o,r,a,l);t.push({range:c,rangeLength:n.rangeLength,text:n.text,rangeOffset:n.rangeOffset,forceMoveMarkers:n.forceMoveMarkers})}return t}findMatchesLineByLine(e,t,i,n){return this._pieceTree.findMatchesLineByLine(e,t,i,n)}static _getInverseEditRanges(e){const t=[];let i=0,n=0,o=null;for(let r=0,a=e.length;r<a;r++){const l=e[r];let c,d;o?o.range.endLineNumber===l.range.startLineNumber?(c=i,d=n+(l.range.startColumn-o.range.endColumn)):(c=i+(l.range.startLineNumber-o.range.endLineNumber),d=l.range.startColumn):(c=l.range.startLineNumber,d=l.range.startColumn);let h;if(l.text.length>0){const u=l.eolCount+1;u===1?h=new k(c,d,c,d+l.firstLineLength):h=new k(c,d,c+u-1,l.lastLineLength+1)}else h=new k(c,d,c,d);i=h.endLineNumber,n=h.endColumn,t.push(h),o=l}return t}static _sortOpsAscending(e,t){const i=k.compareRangesUsingEnds(e.range,t.range);return i===0?e.sortIndex-t.sortIndex:i}static _sortOpsDescending(e,t){const i=k.compareRangesUsingEnds(e.range,t.range);return i===0?t.sortIndex-e.sortIndex:-i}}class O1e{constructor(e,t,i,n,o,r,a,l,c){this._chunks=e,this._bom=t,this._cr=i,this._lf=n,this._crlf=o,this._containsRTL=r,this._containsUnusualLineTerminators=a,this._isBasicASCII=l,this._normalizeEOL=c}_getEOL(e){const t=this._cr+this._lf+this._crlf,i=this._cr+this._crlf;return t===0?e===1?`
`:`\r
`:i>t/2?`\r
`:`
`}create(e){const t=this._getEOL(e),i=this._chunks;if(this._normalizeEOL&&(t===`\r
`&&(this._cr>0||this._lf>0)||t===`
`&&(this._cr>0||this._crlf>0)))for(let o=0,r=i.length;o<r;o++){const a=i[o].buffer.replace(/\r\n|\r|\n/g,t),l=_d(a);i[o]=new wf(a,l)}const n=new vw(i,this._bom,t,this._containsRTL,this._containsUnusualLineTerminators,this._isBasicASCII,this._normalizeEOL);return{textBuffer:n,disposable:n}}}class ate{constructor(){this.chunks=[],this.BOM="",this._hasPreviousChar=!1,this._previousChar=0,this._tmpLineStarts=[],this.cr=0,this.lf=0,this.crlf=0,this.containsRTL=!1,this.containsUnusualLineTerminators=!1,this.isBasicASCII=!0}acceptChunk(e){if(e.length===0)return;this.chunks.length===0&&X6(e)&&(this.BOM=zde,e=e.substr(1));const t=e.charCodeAt(e.length-1);t===13||t>=55296&&t<=56319?(this._acceptChunk1(e.substr(0,e.length-1),!1),this._hasPreviousChar=!0,this._previousChar=t):(this._acceptChunk1(e,!1),this._hasPreviousChar=!1,this._previousChar=t)}_acceptChunk1(e,t){!t&&e.length===0||(this._hasPreviousChar?this._acceptChunk2(String.fromCharCode(this._previousChar)+e):this._acceptChunk2(e))}_acceptChunk2(e){const t=N1e(this._tmpLineStarts,e);this.chunks.push(new wf(e,t.lineStarts)),this.cr+=t.cr,this.lf+=t.lf,this.crlf+=t.crlf,t.isBasicASCII||(this.isBasicASCII=!1,this.containsRTL||(this.containsRTL=iv(e)),this.containsUnusualLineTerminators||(this.containsUnusualLineTerminators=nQ(e)))}finish(e=!0){return this._finish(),new O1e(this.chunks,this.BOM,this.cr,this.lf,this.crlf,this.containsRTL,this.containsUnusualLineTerminators,this.isBasicASCII,e)}_finish(){if(this.chunks.length===0&&this._acceptChunk1("",!0),this._hasPreviousChar){this._hasPreviousChar=!1;const e=this.chunks[this.chunks.length-1];e.buffer+=String.fromCharCode(this._previousChar);const t=_d(e.buffer);e.lineStarts=t,this._previousChar===13&&this.cr++}}}function Yr(s,e=dn.ofCaller()){return typeof s=="string"?new TU(s,void 0,e):new TU(void 0,s,e)}class TU extends qv{get debugName(){return new os(this._owner,this._debugName,void 0).getDebugName(this)??"Observable Signal"}toString(){return this.debugName}constructor(e,t,i){super(i),this._debugName=e,this._owner=t}trigger(e,t){if(!e){Vt(i=>{this.trigger(i,t)},()=>`Trigger signal ${this.debugName}`);return}for(const i of this._observers)e.updateObserver(i,this),i.handleChange(this,t)}get(){}}class P1e extends qv{get debugName(){return this._debugNameData.getDebugName(this)??"LazyObservableValue"}constructor(e,t,i,n){super(n),this._debugNameData=e,this._equalityComparator=i,this._isUpToDate=!0,this._deltas=[],this._updateCounter=0,this._value=t}get(){return this._update(),this._value}_update(){if(!this._isUpToDate)if(this._isUpToDate=!0,this._deltas.length>0){for(const e of this._deltas){gs()?.handleObservableUpdated(this,{change:e,didChange:!0,oldValue:"(unknown)",newValue:this._value,hadValue:!0});for(const t of this._observers)t.handleChange(this,e)}this._deltas.length=0}else{gs()?.handleObservableUpdated(this,{change:void 0,didChange:!0,oldValue:"(unknown)",newValue:this._value,hadValue:!0});for(const e of this._observers)e.handleChange(this,void 0)}}_beginUpdate(){if(this._updateCounter++,this._updateCounter===1)for(const e of this._observers)e.beginUpdate(this)}_endUpdate(){if(this._updateCounter--,this._updateCounter===0){this._update();const e=[...this._observers];for(const t of e)t.endUpdate(this)}}addObserver(e){const t=!this._observers.has(e)&&this._updateCounter>0;super.addObserver(e),t&&e.beginUpdate(this)}removeObserver(e){const t=this._observers.has(e)&&this._updateCounter>0;super.removeObserver(e),t&&e.endUpdate(this)}set(e,t,i){if(i===void 0&&this._equalityComparator(this._value,e))return;let n;t||(t=n=new Gv(()=>{},()=>`Setting ${this.debugName}`));try{if(this._isUpToDate=!1,this._setValue(e),i!==void 0&&this._deltas.push(i),t.updateObserver({beginUpdate:()=>this._beginUpdate(),endUpdate:()=>this._endUpdate(),handleChange:(o,r)=>{},handlePossibleChange:o=>{}},this),this._updateCounter>1)for(const o of this._observers)o.handlePossibleChange(this)}finally{n&&n.finish()}}toString(){return`${this.debugName}: ${this._value}`}_setValue(e){this._value=e}}function H1(s,e,t=dn.ofCaller()){return s.lazy?new P1e(new os(s.owner,s.debugName,void 0),e,s.equalsFn??kr,t):new Yy(new os(s.owner,s.debugName,void 0),e,s.equalsFn??kr,t)}class F1e{constructor(){this._onDidChangeVisibleRanges=new M,this.onDidChangeVisibleRanges=this._onDidChangeVisibleRanges.event,this._views=new Set,this._viewsChanged=Yr(this),this.visibleLineRanges=Jn({owner:this,equalsFn:Gy(nW())},e=>(this._viewsChanged.read(e),Ne.joinMany([...this._views].map(i=>i.state.read(e)?.visibleLineRanges??[]))))}attachView(){const e=new W1e(t=>{this._onDidChangeVisibleRanges.fire({view:e,state:t})});return this._views.add(e),this._viewsChanged.trigger(void 0),e}detachView(e){this._views.delete(e),this._onDidChangeVisibleRanges.fire({view:e,state:void 0}),this._viewsChanged.trigger(void 0)}}class B1e{constructor(e,t){this.visibleLineRanges=e,this.stabilized=t}equals(e){return this===e?!0:!(!ni(this.visibleLineRanges,e.visibleLineRanges,(t,i)=>t.equals(i))||this.stabilized!==e.stabilized)}}class W1e{get state(){return this._state}constructor(e){this.handleStateChange=e,this._state=H1({owner:this,equalsFn:Vf((t,i)=>t.equals(i))},void 0)}setVisibleLines(e,t){const i=e.map(o=>new Ne(o.startLineNumber,o.endLineNumber+1)),n=new B1e(i,t);this._state.set(n,void 0,void 0),this.handleStateChange(n)}}class H1e extends F{get lineRanges(){return this._lineRanges}constructor(e){super(),this._refreshTokens=e,this.runner=this._register(new Rt(()=>this.update(),50)),this._computedLineRanges=[],this._lineRanges=[]}update(){ni(this._computedLineRanges,this._lineRanges,(e,t)=>e.equals(t))||(this._computedLineRanges=this._lineRanges,this._refreshTokens())}handleStateChange(e){this._lineRanges=e.visibleLineRanges,e.stabilized?(this.runner.cancel(),this.update()):this.runner.schedule()}}class lte extends F{get backgroundTokenizationState(){return this._backgroundTokenizationState}constructor(e,t){super(),this._languageIdCodec=e,this._textModel=t,this._onDidChangeTokens=this._register(new M),this.onDidChangeTokens=this._onDidChangeTokens.event}tokenizeIfCheap(e){this.isCheapToTokenize(e)&&this.forceTokenization(e)}}function V1e(s,e){const t=s.gotoNextSibling(),i=e.gotoNextSibling();if(t!==i)throw new Error("Trees are out of sync");return t&&i}function cte(s,e){const t=s.gotoParent(),i=e.gotoParent();if(t!==i)throw new Error("Trees are out of sync");return t&&i}function z1e(s,e,t){const i=s.gotoFirstChild(),n=e.gotoFirstChild();if(i!==n)throw new Error("Trees are out of sync");if(t===0)return i&&n;for(let o=1;o<=t;o++){const r=s.gotoNextSibling(),a=e.gotoNextSibling();if(r!==a)throw new Error("Trees are out of sync");if(!r||!a)return!1}return i&&n}function NU(s,e){do{if(s.currentNode.nextSibling)return V1e(s,e);s.currentNode.parent&&cte(s,e)}while(s.currentNode.nextSibling||s.currentNode.parent);return!1}function $1e(s,e){const t=e.walk();t.resetTo(s);const i=s.currentNode;do if(t.currentNode.previousSibling&&t.currentNode.endIndex-t.currentNode.startIndex!==0)t.gotoPreviousSibling();else{for(;!t.currentNode.previousSibling&&t.currentNode.parent;)t.gotoParent();t.gotoPreviousSibling()}while(t.currentNode.endIndex>i.startIndex&&(t.currentNode.parent||t.currentNode.previousSibling)&&t.currentNode.id!==i.id);if(t.currentNode.id!==i.id&&t.currentNode.endIndex<=i.startIndex)return t.currentNode}var U1e=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},RU=function(s,e){return function(t,i){e(t,i,s)}};let MF=class extends F{constructor(e,t,i,n,o,r,a){super(),this.languageId=e,this._ranges=t,this._parser=i,this._parserClass=n,this.textModel=o,this._logService=r,this._telemetryService=a,this._tree=Re(this,void 0),this.tree=this._tree,this._treeLastParsedVersion=Re(this,-1),this.treeLastParsedVersion=this._treeLastParsedVersion,this._onDidChangeContentQueue=new Ede,this._tree=Re(this,void 0),this.tree=this._tree,this._register(pe(()=>{this._tree.get()?.delete(),this._lastFullyParsed?.delete(),this._lastFullyParsedWithEdits?.delete(),this._parser.delete()})),this.handleContentChange(void 0,this._ranges)}handleContentChange(e,t){const i=this.textModel.getVersionId();let n=[];t&&(n=this._setRanges(t)),e&&this._applyEdits(e.changes),this._onDidChangeContentQueue.clearPending(),this._onDidChangeContentQueue.schedule(async()=>{if(this._store.isDisposed)return;const o=this._lastFullyParsed;let r;this._lastFullyParsedWithEdits&&this._lastFullyParsed&&(r=this._findChangedNodes(this._lastFullyParsedWithEdits,this._lastFullyParsed));const a=await this._parseAndUpdateTree(i);if(a){let l;r?o&&r&&(l=this._findTreeChanges(a,r,n)):this._ranges&&(l=this._ranges.map(d=>({newRange:new k(d.startPosition.row+1,d.startPosition.column+1,d.endPosition.row+1,d.endPosition.column+1),oldRangeLength:d.endIndex-d.startIndex,newRangeStartOffset:d.startIndex,newRangeEndOffset:d.endIndex}))),l||(l=[{newRange:this.textModel.getFullModelRange(),newRangeStartOffset:0,newRangeEndOffset:this.textModel.getValueLength()}]);const c=this._tree.get();Vt(d=>{this._tree.set(a,d,{ranges:l,versionId:i}),this._treeLastParsedVersion.set(i,d)}),c?.delete()}})}_applyEdits(e){for(const t of e){const i=cn.ofRange(k.lift(t.range)),n=cn.ofText(t.text),o=t.text.length===0?n:i.add(n),r={startIndex:t.rangeOffset,oldEndIndex:t.rangeOffset+t.rangeLength,newEndIndex:t.rangeOffset+t.text.length,startPosition:{row:t.range.startLineNumber-1,column:t.range.startColumn-1},oldEndPosition:{row:t.range.endLineNumber-1,column:t.range.endColumn-1},newEndPosition:{row:t.range.startLineNumber+o.lineCount-1,column:o.lineCount?o.columnCount:t.range.endColumn+o.columnCount}};this._tree.get()?.edit(r),this._lastFullyParsedWithEdits?.edit(r)}}_findChangedNodes(e,t){if(this._ranges&&this._ranges.every(a=>a.startPosition.row!==e.rootNode.startPosition.row)||e.rootNode.startPosition.row!==0)return[];const i=e.walk(),n=t.walk(),o=[];let r=!0;do if(i.currentNode.hasChanges){const a=i.currentNode.children,l=[],c=a.filter((d,h)=>d?.hasChanges||n.currentNode.children.length<=h?(l.push(h),!0):!1);if(c.length===0||i.currentNode.hasError!==n.currentNode.hasError){for(;i.currentNode.parent&&r&&!i.currentNode.isNamed;)r=cte(i,n);const d=i.currentNode,h=$1e(i,e)??d;o.push({startIndex:h.startIndex,endIndex:d.endIndex,startPosition:h.startPosition,endPosition:d.endPosition}),r=NU(i,n)}else c.length>=1&&(r=z1e(i,n,l[0]))}else r=NU(i,n);while(r);return i.delete(),n.delete(),o}_findTreeChanges(e,t,i){let n=0;const o=[];for(let r=0;r<t.length;r++){const a=t[r];if(o.length>0&&a.startIndex>=o[o.length-1].newRangeStartOffset&&a.endIndex<=o[o.length-1].newRangeEndOffset)continue;const l=e.walk(),c=()=>l.startIndex<a.startIndex&&l.endIndex>a.endIndex;for(;c();){let p=l.gotoFirstChild(),_=!1;for(;p;)if(c()&&l.currentNode.isNamed){_=!0;break}else p=l.gotoNextSibling();if(!_){l.gotoParent();break}if(l.currentNode.childCount===0)break}const d=l.currentNode.startPosition,h=l.currentNode.endPosition,u=l.currentNode.startIndex,g=l.currentNode.endIndex,f={newRange:new k(d.row+1,d.column+1,h.row+1,h.column+1),newRangeStartOffset:u,newRangeEndOffset:g};n<i.length&&AU(i[n],{startIndex:u,endIndex:g})?(i[n].startIndex<f.newRangeStartOffset&&(f.newRange=f.newRange.setStartPosition(i[n].startPosition.row+1,i[n].startPosition.column+1),f.newRangeStartOffset=i[n].startIndex),i[n].endIndex>f.newRangeEndOffset&&(f.newRange=f.newRange.setEndPosition(i[n].endPosition.row+1,i[n].endPosition.column+1),f.newRangeEndOffset=i[n].endIndex),n++):n<i.length&&i[n].endIndex<f.newRangeStartOffset&&o.push({newRange:new k(i[n].startPosition.row+1,i[n].startPosition.column+1,i[n].endPosition.row+1,i[n].endPosition.column+1),newRangeStartOffset:i[n].startIndex,newRangeEndOffset:i[n].endIndex}),o.length>0&&o[o.length-1].newRangeEndOffset>=f.newRangeStartOffset?(o[o.length-1].newRange=k.fromPositions(o[o.length-1].newRange.getStartPosition(),f.newRange.getEndPosition()),o[o.length-1].newRangeEndOffset=f.newRangeEndOffset):o.push(f)}return this._constrainRanges(o)}_constrainRanges(e){if(!this._ranges)return e;const t=[];let i=0,n=0;for(;i<e.length&&n<this._ranges.length;){const o=e[i],r=this._ranges[n];if(o.newRangeEndOffset<r.startIndex)i++;else if(o.newRangeStartOffset>r.endIndex)n++;else{const a=Math.max(o.newRangeStartOffset,r.startIndex),l=Math.min(o.newRangeEndOffset,r.endIndex),c=o.newRange.intersectRanges(new k(r.startPosition.row+1,r.startPosition.column+1,r.endPosition.row+1,r.endPosition.column+1));t.push({newRange:c,newRangeEndOffset:l,newRangeStartOffset:a}),l<o.newRangeEndOffset?(o.newRange=k.fromPositions(c.getEndPosition(),o.newRange.getEndPosition()),o.newRangeStartOffset=l+1):i++}}return t}async _parseAndUpdateTree(e){const t=await this._parse();if(t)return this._lastFullyParsed?.delete(),this._lastFullyParsed=t.copy(),this._lastFullyParsedWithEdits?.delete(),this._lastFullyParsedWithEdits=t.copy(),t;this._tree.get()||this._parser.reset()}_parse(){let e="fullParse";return this._tree.get()&&(e="incrementalParse"),this._parseAndYield(e)}async _parseAndYield(e){let t=0,i=0;const n=this.textModel.getVersionId();let o;const r=q1e();do{const a=performance.now();o=this._parser.parse((l,c)=>this._parseCallback(l),this._tree.get(),{progressCallback:r,includedRanges:this._ranges}),t+=performance.now()-a,i++,await new Promise(l=>V0(l))}while(!this._store.isDisposed&&!o&&n===this.textModel.getVersionId());return this._sendParseTimeTelemetry(e,t,i),o&&n===this.textModel.getVersionId()?o:void 0}_parseCallback(e){try{return this.textModel.getTextBuffer().getNearestChunk(e)}catch(t){this._logService.debug("Error getting chunk for tree-sitter parsing",t)}}_setRanges(e){const t=[];if(this._ranges)for(const i of e){let n=!1;for(let o=0;o<this._ranges.length;o++){const r=this._ranges[o];if(j1e(r,i)||AU(r,i)){n=!0;break}}n||t.push(i)}else t.push(...e);return this._ranges=e,t}_sendParseTimeTelemetry(e,t,i){this._logService.debug(`Tree parsing (${e}) took ${t} ms and ${i} passes.`),e==="fullParse"?this._telemetryService.publicLog2("treeSitter.fullParse",{languageId:this.languageId,time:t,passes:i}):this._telemetryService.publicLog2("treeSitter.incrementalParse",{languageId:this.languageId,time:t,passes:i})}createParsedTreeSync(e){const t=new this._parserClass;t.setLanguage(this._parser.language);const i=t.parse(e);return t.delete(),i??void 0}};MF=U1e([RU(5,jt),RU(6,vs)],MF);function q1e(){let s=performance.now();return function(t){const i=performance.now();return i-s>50?(s=i,!0):!1}}function j1e(s,e){return s.startPosition.row===e.startPosition.row&&s.startPosition.column===e.startPosition.column&&s.endPosition.row===e.endPosition.row&&s.endPosition.column===e.endPosition.column&&s.startIndex===e.startIndex&&s.endIndex===e.endIndex}function AU(s,e){return s.startIndex<=e.startIndex&&s.endIndex>=e.startIndex||e.startIndex<=s.startIndex&&e.endIndex>=s.startIndex}const bv=new class{clone(){return this}equals(s){return this===s}};function h7(s,e){return new GW([new pS(0,"",s)],e)}function mA(s,e){const t=new Uint32Array(2);return t[0]=0,t[1]=(s<<0|0|0|32768|2<<24)>>>0,new gA(t,e===null?bv:e)}class G1e{constructor(e){this._default=e,this._store=[]}get(e){return e<this._store.length?this._store[e]:this._default}set(e,t){for(;e>=this._store.length;)this._store[this._store.length]=this._default;this._store[e]=t}replace(e,t,i){if(e>=this._store.length)return;if(t===0){this.insert(e,i);return}else if(i===0){this.delete(e,t);return}const n=this._store.slice(0,e),o=this._store.slice(e+t),r=K1e(i,this._default);this._store=n.concat(r,o)}delete(e,t){t===0||e>=this._store.length||this._store.splice(e,t)}insert(e,t){if(t===0||e>=this._store.length)return;const i=[];for(let n=0;n<t;n++)i[n]=this._default;this._store=IR(this._store,e,i)}}function K1e(s,e){const t=[];for(let i=0;i<s;i++)t[i]=e;return t}class Y1e{get startLineNumber(){return this._startLineNumber}get endLineNumber(){return this._startLineNumber+this._tokens.length-1}constructor(e,t){this._startLineNumber=e,this._tokens=t}getLineTokens(e){return this._tokens[e-this._startLineNumber]}appendLineTokens(e){this._tokens.push(e)}}class OF{constructor(){this._tokens=[]}add(e,t){if(this._tokens.length>0){const i=this._tokens[this._tokens.length-1];if(i.endLineNumber+1===e){i.appendLineTokens(t);return}}this._tokens.push(new Y1e(e,[t]))}finalize(){return this._tokens}}class Z1e{constructor(e,t){this.tokenizationSupport=t,this.initialState=this.tokenizationSupport.getInitialState(),this.store=new PF(e)}getStartState(e){return this.store.getStartState(e,this.initialState)}getFirstInvalidLine(){return this.store.getFirstInvalidLine(this.initialState)}}class X1e extends Z1e{constructor(e,t,i,n){super(e,t),this._textModel=i,this._languageIdCodec=n}updateTokensUntilLine(e,t){const i=this._textModel.getLanguageId();for(;;){const n=this.getFirstInvalidLine();if(!n||n.lineNumber>t)break;const o=this._textModel.getLineContent(n.lineNumber),r=NC(this._languageIdCodec,i,this.tokenizationSupport,o,!0,n.startState);e.add(n.lineNumber,r.tokens),this.store.setEndState(n.lineNumber,r.endState)}}getTokenTypeIfInsertingCharacter(e,t){const i=this.getStartState(e.lineNumber);if(!i)return 0;const n=this._textModel.getLanguageId(),o=this._textModel.getLineContent(e.lineNumber),r=o.substring(0,e.column-1)+t+o.substring(e.column-1),a=NC(this._languageIdCodec,n,this.tokenizationSupport,r,!0,i),l=new Ri(a.tokens,r,this._languageIdCodec);if(l.getCount()===0)return 0;const c=l.findTokenIndexAtOffset(e.column-1);return l.getStandardTokenType(c)}tokenizeLinesAt(e,t){const i=this.getStartState(e);if(!i)return null;const n=this._textModel.getLanguageId(),o=[];let r=i;for(const a of t){const l=NC(this._languageIdCodec,n,this.tokenizationSupport,a,!0,r);o.push(new Ri(l.tokens,a,this._languageIdCodec)),r=l.endState}return o}hasAccurateTokensForLine(e){const t=this.store.getFirstInvalidEndStateLineNumberOrMax();return e<t}isCheapToTokenize(e){const t=this.store.getFirstInvalidEndStateLineNumberOrMax();return e<t||e===t&&this._textModel.getLineLength(e)<2048}tokenizeHeuristically(e,t,i){if(i<=this.store.getFirstInvalidEndStateLineNumberOrMax())return{heuristicTokens:!1};if(t<=this.store.getFirstInvalidEndStateLineNumberOrMax())return this.updateTokensUntilLine(e,i),{heuristicTokens:!1};let n=this.guessStartState(t);const o=this._textModel.getLanguageId();for(let r=t;r<=i;r++){const a=this._textModel.getLineContent(r),l=NC(this._languageIdCodec,o,this.tokenizationSupport,a,!0,n);e.add(r,l.tokens),n=l.endState}return{heuristicTokens:!0}}guessStartState(e){let{likelyRelevantLines:t,initialState:i}=dte(this._textModel,e,this);i||(i=this.tokenizationSupport.getInitialState());const n=this._textModel.getLanguageId();let o=i;for(const r of t)o=NC(this._languageIdCodec,n,this.tokenizationSupport,r,!1,o).endState;return o}}function dte(s,e,t){let i=s.getLineFirstNonWhitespaceColumn(e);const n=[];let o=null;for(let r=e-1;i>1&&r>=1;r--){const a=s.getLineFirstNonWhitespaceColumn(r);if(a!==0&&a<i&&(n.push(s.getLineContent(r)),i=a,o=t?.getStartState(r),o))break}return n.reverse(),{likelyRelevantLines:n,initialState:o??void 0}}class PF{constructor(e){this.lineCount=e,this._tokenizationStateStore=new Q1e,this._invalidEndStatesLineNumbers=new J1e,this._invalidEndStatesLineNumbers.addRange(new me(1,e+1))}getEndState(e){return this._tokenizationStateStore.getEndState(e)}setEndState(e,t){if(!t)throw new Ce("Cannot set null/undefined state");this._invalidEndStatesLineNumbers.delete(e);const i=this._tokenizationStateStore.setEndState(e,t);return i&&e<this.lineCount&&this._invalidEndStatesLineNumbers.addRange(new me(e+1,e+2)),i}acceptChange(e,t){this.lineCount+=t-e.length,this._tokenizationStateStore.acceptChange(e,t),this._invalidEndStatesLineNumbers.addRangeAndResize(new me(e.startLineNumber,e.endLineNumberExclusive),t)}acceptChanges(e){for(const t of e){const[i]=fg(t.text);this.acceptChange(new Ne(t.range.startLineNumber,t.range.endLineNumber+1),i+1)}}invalidateEndStateRange(e){this._invalidEndStatesLineNumbers.addRange(new me(e.startLineNumber,e.endLineNumberExclusive))}getFirstInvalidEndStateLineNumber(){return this._invalidEndStatesLineNumbers.min}getFirstInvalidEndStateLineNumberOrMax(){return this.getFirstInvalidEndStateLineNumber()||Number.MAX_SAFE_INTEGER}allStatesValid(){return this._invalidEndStatesLineNumbers.min===null}getStartState(e,t){return e===1?t:this.getEndState(e-1)}getFirstInvalidLine(e){const t=this.getFirstInvalidEndStateLineNumber();if(t===null)return null;const i=this.getStartState(t,e);if(!i)throw new Ce("Start state must be defined");return{lineNumber:t,startState:i}}}class Q1e{constructor(){this._lineEndStates=new G1e(null)}getEndState(e){return this._lineEndStates.get(e)}setEndState(e,t){const i=this._lineEndStates.get(e);return i&&i.equals(t)?!1:(this._lineEndStates.set(e,t),!0)}acceptChange(e,t){let i=e.length;t>0&&i>0&&(i--,t--),this._lineEndStates.replace(e.startLineNumber,i,t)}}class J1e{constructor(){this._ranges=[]}get min(){return this._ranges.length===0?null:this._ranges[0].start}delete(e){const t=this._ranges.findIndex(i=>i.contains(e));if(t!==-1){const i=this._ranges[t];i.start===e?i.endExclusive===e+1?this._ranges.splice(t,1):this._ranges[t]=new me(e+1,i.endExclusive):i.endExclusive===e+1?this._ranges[t]=new me(i.start,e):this._ranges.splice(t,1,new me(i.start,e),new me(e+1,i.endExclusive))}}addRange(e){me.addRange(e,this._ranges)}addRangeAndResize(e,t){let i=0;for(;!(i>=this._ranges.length||e.start<=this._ranges[i].endExclusive);)i++;let n=i;for(;!(n>=this._ranges.length||e.endExclusive<this._ranges[n].start);)n++;const o=t-e.length;for(let r=n;r<this._ranges.length;r++)this._ranges[r]=this._ranges[r].delta(o);if(i===n){const r=new me(e.start,e.start+t);r.isEmpty||this._ranges.splice(i,0,r)}else{const r=Math.min(e.start,this._ranges[i].start),a=Math.max(e.endExclusive,this._ranges[n-1].endExclusive),l=new me(r,a+o);l.isEmpty?this._ranges.splice(i,n-i):this._ranges.splice(i,n-i,l)}}toString(){return this._ranges.map(e=>e.toString()).join(" + ")}}function NC(s,e,t,i,n,o){let r=null;if(t)try{r=t.tokenizeEncoded(i,n,o.clone())}catch(a){Oe(a)}return r||(r=mA(s.encodeLanguageId(e),o)),Ri.convertToEndOffset(r.tokens,i.length),r}class eye{constructor(e,t){this._tokenizerWithStateStore=e,this._backgroundTokenStore=t,this._isDisposed=!1,this._isScheduled=!1}dispose(){this._isDisposed=!0}handleChanges(){this._beginBackgroundTokenization()}_beginBackgroundTokenization(){this._isScheduled||!this._tokenizerWithStateStore._textModel.isAttachedToEditor()||!this._hasLinesToTokenize()||(this._isScheduled=!0,U6(e=>{this._isScheduled=!1,this._backgroundTokenizeWithDeadline(e)}))}_backgroundTokenizeWithDeadline(e){const t=Date.now()+e.timeRemaining(),i=()=>{this._isDisposed||!this._tokenizerWithStateStore._textModel.isAttachedToEditor()||!this._hasLinesToTokenize()||(this._backgroundTokenizeForAtLeast1ms(),Date.now()<t?V0(i):this._beginBackgroundTokenization())};i()}_backgroundTokenizeForAtLeast1ms(){const e=this._tokenizerWithStateStore._textModel.getLineCount(),t=new OF,i=Cn.create(!1);do if(i.elapsed()>1||this._tokenizeOneInvalidLine(t)>=e)break;while(this._hasLinesToTokenize());this._backgroundTokenStore.setTokens(t.finalize()),this.checkFinished()}_hasLinesToTokenize(){return this._tokenizerWithStateStore?!this._tokenizerWithStateStore.store.allStatesValid():!1}_tokenizeOneInvalidLine(e){const t=this._tokenizerWithStateStore?.getFirstInvalidLine();return t?(this._tokenizerWithStateStore.updateTokensUntilLine(e,t.lineNumber),t.lineNumber):this._tokenizerWithStateStore._textModel.getLineCount()+1}checkFinished(){this._isDisposed||this._tokenizerWithStateStore.store.allStatesValid()&&this._backgroundTokenStore.backgroundTokenizationFinished()}requestTokens(e,t){this._tokenizerWithStateStore.store.invalidateEndStateRange(new Ne(e,t))}}class Nm{get children(){return this._children}get length(){return this._length}constructor(e){this.height=e,this._children=[],this._length=0}static create(e,t){const i=new Nm(e.height+1);return i.appendChild(e),i.appendChild(t),i}canAppendChild(){return this._children.length<3}appendChild(e){if(!this.canAppendChild())throw new Error("Cannot insert more than 3 children in a ListNode");this._children.push(e),this._length+=e.length,this._updateParentLength(e.length),$s(e)||(e.parent=this)}_updateParentLength(e){let t=this.parent;for(;t;)t._length+=e,t=t.parent}unappendChild(){const e=this._children.pop();return this._length-=e.length,this._updateParentLength(-e.length),e}prependChild(e){if(this._children.length>=3)throw new Error("Cannot prepend more than 3 children in a ListNode");this._children.unshift(e),this._length+=e.length,this._updateParentLength(e.length),$s(e)||(e.parent=this)}unprependChild(){const e=this._children.shift();return this._length-=e.length,this._updateParentLength(-e.length),e}lastChild(){return this._children[this._children.length-1]}dispose(){this._children.splice(0,this._children.length)}}var fr;(function(s){s[s.None=0]="None",s[s.ViewportGuess=1]="ViewportGuess",s[s.EditGuess=2]="EditGuess",s[s.Accurate=3]="Accurate"})(fr||(fr={}));function $s(s){return s.token!==void 0}function hte(s,e){let t=s;const i=[];let n;for(;;){if(e.height===t.height){n=e;break}if($s(t))throw new Error("unexpected");i.push(t),t=t.lastChild()}for(let o=i.length-1;o>=0;o--){const r=i[o];n&&(r.children.length>=3?n=Nm.create(r.unappendChild(),n):(r.appendChild(n),n=void 0))}if(n){const o=new Nm(n.height+1);return o.appendChild(s),o.appendChild(n),o}else return s}function tye(s,e){let t=s;const i=[];for(;e.height!==t.height;){if($s(t))throw new Error("unexpected");i.push(t),t=t.children[0]}let n=e;for(let o=i.length-1;o>=0;o--){const r=i[o];n&&(r.children.length>=3?n=Nm.create(n,r.unprependChild()):(r.prependChild(n),n=void 0))}return n?Nm.create(n,s):s}function iye(s,e){return s.height===e.height?Nm.create(s,e):s.height>e.height?hte(s,e):tye(e,s)}class nye{constructor(e){this._textModel=e,this._root=this.createEmptyRoot()}createEmptyRoot(){return{length:this._textModel.getValueLength(),token:0,height:0,tokenQuality:fr.None}}buildStore(e,t){this._root=this.createFromUpdates(e,t)}createFromUpdates(e,t){if(e.length===0)return this.createEmptyRoot();let i={length:e[0].length,token:e[0].token,height:0,tokenQuality:t};for(let n=1;n<e.length;n++)i=hte(i,{length:e[n].length,token:e[n].token,height:0,tokenQuality:t});return i}update(e,t,i){t.length!==0&&this.replace(e,t[0].startOffsetInclusive,t,i)}delete(e,t){this.replace(e,t,[],fr.EditGuess)}replace(e,t,i,n){const o=t+e,r=[],a=[],l=[{node:this._root,offset:0}];for(;l.length>0;){const h=l.pop(),u=h.offset;if(u<t&&u+h.node.length<=t){$s(h.node)||(h.node.parent=void 0),r.push(h.node);continue}else $s(h.node)&&u<t&&r.push({length:t-u,token:h.node.token,height:0,tokenQuality:h.node.tokenQuality});if(!(t<=u&&u+h.node.length<=o)){if(u>=o){$s(h.node)||(h.node.parent=void 0),a.push(h.node);continue}else if($s(h.node)&&u+h.node.length>o){a.push({length:u+h.node.length-o,token:h.node.token,height:0,tokenQuality:h.node.tokenQuality});continue}if(!$s(h.node)){let g=u+h.node.length;for(let f=h.node.children.length-1;f>=0;f--)g-=h.node.children[f].length,l.push({node:h.node.children[f],offset:g})}}}let c;i.length>0?c=r.concat(this.createFromUpdates(i,n),a):c=r.concat(a);let d=c[0];for(let h=1;h<c.length;h++)d=iye(d,c[h]);this._root=d??this.createEmptyRoot()}traverseInOrderInRange(e,t,i){const n=[{node:this._root,offset:0}];for(;n.length>0;){const{node:o,offset:r}=n.pop();if(!(r+o.length<=e||r>=t)){if(i(o,r))return;if(!$s(o)){let l=r+o.length;for(let c=o.children.length-1;c>=0;c--)l-=o.children[c].length,n.push({node:o.children[c],offset:l})}}}}getTokenAt(e){let t;return this.traverseInOrderInRange(e,this._root.length,(i,n)=>$s(i)?(t={token:i.token,startOffsetInclusive:n,length:i.length},!0):!1),t}getTokensInRange(e,t){const i=[];return this.traverseInOrderInRange(e,t,(n,o)=>{if($s(n)){let r=n.length,a=o;o<e&&o+n.length>t?(a=e,r=t-e):o<e?(r-=e-o,a=e):o+n.length>t&&(r-=o+n.length-t),i.push({token:n.token,startOffsetInclusive:a,length:r})}return!1}),i}markForRefresh(e,t){this.traverseInOrderInRange(e,t,i=>($s(i)&&(i.tokenQuality=fr.None),!1))}rangeHasTokens(e,t,i){let n=!0;return this.traverseInOrderInRange(e,t,o=>($s(o)&&o.tokenQuality<i&&(n=!1),!1)),n}rangeNeedsRefresh(e,t){let i=!1;return this.traverseInOrderInRange(e,t,n=>($s(n)&&n.tokenQuality!==fr.Accurate&&(i=!0),!1)),i}getNeedsRefresh(){const e=[];return this.traverseInOrderInRange(0,this._textModel.getValueLength(),(t,i)=>($s(t)&&t.tokenQuality!==fr.Accurate&&(e.length>0&&e[e.length-1].endOffset===i?e[e.length-1].endOffset+=t.length:e.push({startOffset:i,endOffset:i+t.length})),!1)),e}dispose(){const e=[[this._root,!1]];for(;e.length>0;){const[t,i]=e.pop();if(!$s(t))if(i)t.dispose(),t.parent=void 0;else{e.push([t,!0]);for(let n=t.children.length-1;n>=0;n--)e.push([t.children[n],!1])}}this._root=void 0}}const sye=qe("treeSitterThemeService");function oye(s){return{createChangeSummary:e=>({changes:[]}),handleChange(e,t){for(const i in s)e.didChange(s[i])&&t.changes.push({key:i,change:e.change});return!0},beforeUpdate(e,t){for(const i in s){if(i==="changes")throw new Ce('property name "changes" is reserved for change tracking');t[i]=s[i].read(e)}}}}function rye(s){let e;return{createChangeSummary:t=>({changes:[]}),handleChange(t,i){e||(e=s());for(const n in e)t.didChange(e[n])&&i.changes.push({key:n,change:t.change});return!0},beforeUpdate(t,i){e||(e=s());for(const n in e){if(n==="changes")throw new Ce('property name "changes" is reserved for change tracking');i[n]=e[n].read(t)}}}}var aye=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},lye=function(s,e){return function(t,i){e(t,i,s)}};let FF=class extends F{get _textModel(){return this._tree.textModel}constructor(e,t,i,n,o){super(),this._tree=e,this._highlightingQueries=t,this._languageIdCodec=i,this._visibleLineRanges=n,this._treeSitterThemeService=o,this._onDidChangeTokens=this._register(new M),this.onDidChangeTokens=this._onDidChangeTokens.event,this._onDidCompleteBackgroundTokenization=this._register(new M),this.onDidChangeBackgroundTokenization=this._onDidCompleteBackgroundTokenization.event,this._encodedLanguageId=this._languageIdCodec.encodeLanguageId(this._tree.languageId),this._register(yl(this._treeSitterThemeService.onChange,()=>{this._updateTheme()})),this._tokenStore=this._register(new nye(this._textModel)),this._accurateVersion=this._textModel.getVersionId(),this._guessVersion=this._textModel.getVersionId(),this._tokenStore.buildStore(this._createEmptyTokens(),fr.None),this._register(De(r=>{const a=this._visibleLineRanges.read(r);this._parseAndTokenizeViewPort(a)})),this._register(jv({owner:this,changeTracker:oye({tree:this._tree.tree})},(r,a)=>{const l=a.changes.at(0)?.change;if(a.changes.length>1)throw new Ce("The tree changed twice in one transaction. This is currently not supported and should not happen.");if(!l)a.tree&&this._firstTreeUpdate(this._tree.treeLastParsedVersion.read(r));else{if(this.hasTokens())for(const c of l.ranges)this._markForRefresh(c.newRange);this.hasTokens()?this._handleTreeUpdate(l.ranges,l.versionId):this._firstTreeUpdate(l.versionId)}}))}handleContentChanged(e){this._guessVersion=e.versionId;for(const t of e.changes)if(t.text.length>t.rangeLength){const i=t.rangeOffset>0?t.rangeOffset-1:t.rangeOffset,n=this._tokenStore.getTokenAt(i);let o;n?(o={startOffsetInclusive:n.startOffsetInclusive,length:n.length+t.text.length-t.rangeLength,token:n.token},this._tokenStore.markForRefresh(i,t.rangeOffset+(t.text.length>t.rangeLength?t.text.length:t.rangeLength))):o={startOffsetInclusive:i,length:t.text.length,token:0},this._tokenStore.update(n?.length??0,[o],fr.EditGuess)}else if(t.text.length<t.rangeLength){const i=t.rangeLength-t.text.length;this._tokenStore.delete(i,t.rangeOffset)}}getLineTokens(e){const t=this._textModel.getLineContent(e),i=this.getTokens(e);return new Ri(i,t,this._languageIdCodec)}_createEmptyTokens(){const e=this._emptyToken(),t=this._textModel.getValueLength();return[this._emptyTokensForOffsetAndLength(0,t,e)]}_emptyToken(){return this._treeSitterThemeService.findMetadata([],this._encodedLanguageId,!1,void 0)}_emptyTokensForOffsetAndLength(e,t,i){return{token:i,length:e+t,startOffsetInclusive:0}}hasAccurateTokensForLine(e){return this.hasTokens(new k(e,1,e,this._textModel.getLineMaxColumn(e)))}tokenizeLinesAt(e,t){const i=this._guessTokensForLinesContent(e,t),n=[];if(!i)return null;for(let o=0;o<i.length;o++)n.push(new Ri(i[o],t[o],this._languageIdCodec));return n}_rangeHasTokens(e,t){return this._tokenStore.rangeHasTokens(this._textModel.getOffsetAt(e.getStartPosition()),this._textModel.getOffsetAt(e.getEndPosition()),t)}hasTokens(e){return!e||this._guessVersion===this._accurateVersion?!0:!this._tokenStore.rangeNeedsRefresh(this._textModel.getOffsetAt(e.getStartPosition()),this._textModel.getOffsetAt(e.getEndPosition()))}getTokens(e){const t=this._textModel.getOffsetAt({lineNumber:e,column:1}),i=this._textModel.getOffsetAt({lineNumber:e,column:this._textModel.getLineLength(e)+1}),n=this._tokenStore.getTokensInRange(t,i),o=new Uint32Array(n.length*2);for(let r=0;r<n.length;r++)o[r*2]=n[r].startOffsetInclusive-t+n[r].length,o[r*2+1]=n[r].token;return o}getTokensInRange(e,t,i,n){const o=n?this._tokenizeCapturesWithMetadata(n,t,i):this._tokenize(e,t,i);if(o?.endOffsetsAndMetadata)return this._rangeTokensAsUpdates(t,o.endOffsetsAndMetadata)}_updateTokensInStore(e,t,i){this._accurateVersion=e;for(const n of t){const o=n.newTokens.length>0?n.newTokens[n.newTokens.length-1]:void 0;let r;o&&this._guessVersion>=e?r=o.startOffsetInclusive+o.length-n.newTokens[0].startOffsetInclusive:n.oldRangeLength?r=n.oldRangeLength:r=0,this._tokenStore.update(r,n.newTokens,i)}}_markForRefresh(e){this._tokenStore.markForRefresh(this._textModel.getOffsetAt(e.getStartPosition()),this._textModel.getOffsetAt(e.getEndPosition()))}_getNeedsRefresh(){const e=this._tokenStore.getNeedsRefresh();return e?e.map(t=>({range:k.fromPositions(this._textModel.getPositionAt(t.startOffset),this._textModel.getPositionAt(t.endOffset)),startOffset:t.startOffset,endOffset:t.endOffset})):[]}_parseAndTokenizeViewPort(e){const t=e.map(i=>i.toInclusiveRange()).filter(Ln);for(const i of t){const n=this._textModel.getOffsetAt(i.getStartPosition()),o=this._textModel.getOffsetAt(i.getEndPosition()),r=this._textModel.getVersionId();if(this._rangeHasTokens(i,fr.ViewportGuess))continue;const a=this._textModel.getValueInRange(i),l=this._forceParseAndTokenizeContent(i,n,o,a,!0);if(!l||this._rangeHasTokens(i,fr.ViewportGuess)||l.length===0)continue;const c=l[l.length-1],d=c.startOffsetInclusive+c.length-l[0].startOffsetInclusive;this._updateTokensInStore(r,[{newTokens:l,oldRangeLength:d}],fr.ViewportGuess),this._onDidChangeTokens.fire({changes:{semanticTokensApplied:!1,ranges:[{fromLineNumber:i.startLineNumber,toLineNumber:i.endLineNumber}]}})}}_guessTokensForLinesContent(e,t){if(t.length===0)return;const i=t.join(this._textModel.getEOL()),n=new k(1,1,e+t.length,t[t.length-1].length+1),o=this._textModel.getOffsetAt({lineNumber:e,column:1}),r=this._forceParseAndTokenizeContent(n,o,o+i.length,i,!1);if(!r)return;const a=new Array(t.length);let l=0,c=0,d=0;for(let h=0;h<t.length;h++){const u=[];let g=!1;for(let f=l;!g&&f<r.length;f++){const p=r[f],_=p.endOffset-d,b=c-d;if(_<=t[h].length)u.push({endOffset:_,metadata:p.metadata}),l++;else if(b<t[h].length){const w={endOffset:t[h].length,metadata:p.metadata};u.push(w),g=!0}else g=!0;c=p.endOffset}a[h]=this._endOffsetTokensToUint32Array(u),d+=t[h].length+this._textModel.getEOL().length}return a}_forceParseAndTokenizeContent(e,t,i,n,o){const r=dte(this._textModel,e.startLineNumber).likelyRelevantLines,a=r.join(this._textModel.getEOL()),l=this._tree.createParsedTreeSync(`${a}${n}`);if(!l)return;const c=new k(1,1,e.endLineNumber-e.startLineNumber+1+r.length,e.endColumn),d=this.captureAtRange(c),h=this._tokenizeCapturesWithMetadata(d,a.length,i-t+a.length);if(l.delete(),!!h)return o?this._rangeTokensAsUpdates(t,h.endOffsetsAndMetadata,a.length):h.endOffsetsAndMetadata}_firstTreeUpdate(e){return this._setViewPortTokens(e)}_setViewPortTokens(e){const t=this._visibleLineRanges.get().map(i=>{const n=i.toInclusiveRange();if(!n)return;const o=this._textModel.getOffsetAt(n.getStartPosition()),r=this._textModel.getOffsetAt(n.getEndPosition());return{newRange:n,newRangeEndOffset:r,newRangeStartOffset:o}}).filter(Ln);return this._handleTreeUpdate(t,e)}_handleTreeUpdate(e,t){const i=[];for(let r=0;r<e.length;r++)if(e[r].newRange.endLineNumber-e[r].newRange.startLineNumber>1e3){const l=e[r].newRange.endLineNumber;let c=e[r].newRange.startLineNumber,d=e[r].newRange.startColumn,h=c+1e3;do{const u=new A(c,d),g=h===e[r].newRange.endLineNumber?e[r].newRange.endColumn:this._textModel.getLineMaxColumn(h),f=new A(h,g),p=k.fromPositions(u,f);i.push({range:p,startOffset:this._textModel.getOffsetAt(p.getStartPosition()),endOffset:this._textModel.getOffsetAt(p.getEndPosition())}),c=h+1,d=1,h<l&&h+1e3>l?h=l:h=h+1e3}while(h<=l)}else if(r===0||i[r-1].endOffset<e[r].newRangeStartOffset)i.push({range:e[r].newRange,startOffset:e[r].newRangeStartOffset,endOffset:e[r].newRangeEndOffset});else if(i[r-1].endOffset<e[r].newRangeEndOffset){const l=this._textModel.getPositionAt(i[r-1].endOffset+1),c=new k(l.lineNumber,l.column,e[r].newRange.endLineNumber,e[r].newRange.endColumn);i.push({range:c,startOffset:i[r-1].endOffset+1,endOffset:e[r].newRangeEndOffset})}const o=i.map(r=>this._getCaptures(r.range));return this._updateTreeForRanges(i,t,o).then(()=>{!this._textModel.isDisposed()&&this._tree.treeLastParsedVersion.get()===this._textModel.getVersionId()&&this._refreshNeedsRefresh(t)})}async _updateTreeForRanges(e,t,i){let n;for(let o=0;o<e.length&&!(!this._textModel.isDisposed()&&t!==this._textModel.getVersionId());o++){const r=i[o],a=e[o],l=this.getTokensInRange(a.range,a.startOffset,a.endOffset,r);l?n={newTokens:l}:n={newTokens:[]},this._updateTokensInStore(t,[n],fr.Accurate),this._onDidChangeTokens.fire({changes:{semanticTokensApplied:!1,ranges:[{fromLineNumber:a.range.getStartPosition().lineNumber,toLineNumber:a.range.getEndPosition().lineNumber}]}}),await new Promise(c=>V0(c))}this._onDidCompleteBackgroundTokenization.fire()}_refreshNeedsRefresh(e){const t=this._getNeedsRefresh();if(t.length===0)return;const i=new Array(t.length);for(let n=0;n<t.length;n++){const o=t[n];i[n]={newRange:o.range,newRangeStartOffset:o.startOffset,newRangeEndOffset:o.endOffset}}this._handleTreeUpdate(i,e)}_rangeTokensAsUpdates(e,t,i){const n=[];let o=0;for(const r of t){if(r.endOffset<=o||i&&r.endOffset<i)continue;let a;i&&o<i?a={startOffsetInclusive:e+i,length:r.endOffset-i,token:r.metadata}:a={startOffsetInclusive:e+o,length:r.endOffset-o,token:r.metadata},n.push(a),o=r.endOffset}return n}_updateTheme(){const e=this._textModel.getFullModelRange();this._markForRefresh(e),this._parseAndTokenizeViewPort(this._visibleLineRanges.get())}captureAtRange(e){const t=this._tree.tree.get();return t?this._highlightingQueries.captures(t.rootNode,{startPosition:{row:e.startLineNumber-1,column:e.startColumn-1},endPosition:{row:e.endLineNumber-1,column:e.endColumn-1}}).map(i=>({name:i.name,text:i.node.text,node:{startIndex:i.node.startIndex,endIndex:i.node.endIndex,startPosition:{lineNumber:i.node.startPosition.row+1,column:i.node.startPosition.column+1},endPosition:{lineNumber:i.node.endPosition.row+1,column:i.node.endPosition.column+1}},encodedLanguageId:this._encodedLanguageId})):[]}captureAtRangeWithInjections(e){const t=this.captureAtRange(e);for(let i=0;i<t.length;i++){const n=t[i],o=n.node.startPosition.lineNumber,r=n.node.endPosition.lineNumber,a=n.node.startPosition.column,l=n.node.endPosition.column,c=o>e.startLineNumber&&o<e.endLineNumber?o:e.startLineNumber,d=r>e.startLineNumber&&r<e.endLineNumber?r:e.endLineNumber,h=o===e.startLineNumber?a<e.startColumn?e.startColumn:a:o<e.startLineNumber?e.startColumn:a,u=r===e.endLineNumber?l>e.endColumn?e.endColumn:l:r>e.endLineNumber?e.endColumn:l,g=new k(c,h,d,u),f=this._getInjectionCaptures(n,g);f&&f.length>0&&(t.splice(i+1,0,...f),i+=f.length)}return t}tokenizeEncoded(e){const t=this._tokenizeEncoded(e);if(!t)return;const i=this._rangeTokensAsUpdates(this._textModel.getOffsetAt({lineNumber:e,column:1}),t.result);t.versionId===this._textModel.getVersionId()&&this._updateTokensInStore(t.versionId,[{newTokens:i,oldRangeLength:this._textModel.getLineLength(e)}],fr.Accurate)}_getCaptures(e){return this.captureAtRangeWithInjections(e)}_tokenize(e,t,i){const n=this._getCaptures(e),o=this._tokenizeCapturesWithMetadata(n,t,i);if(o)return{...o,versionId:this._tree.treeLastParsedVersion.get()}}_createTokensFromCaptures(e,t,i){const n=this._tree.tree.get(),o=Cn.create(),r=i-t,a=this._languageIdCodec.encodeLanguageId(this._tree.languageId),l=cye[this._tree.languageId]||"source";if(e.length===0)return n?(o.stop(),{endOffsets:[{endOffset:r,scopes:[],encodedLanguageId:a}],captureTime:o.elapsed()}):void 0;const c=Array(e.length);c.fill({endOffset:0,scopes:[l],encodedLanguageId:a});let d=0;const h=()=>{c.push({endOffset:0,scopes:[l],encodedLanguageId:a})},u=(p,_)=>p.name.includes("punctuation")&&p.text?Array.from(p.text.matchAll(dye)).map(b=>_+b.index):void 0,g=(p,_,b,w)=>{if(w!==void 0){const C=c[w].scopes;let v=c[w].bracket;if((w>0?c[w-1].endOffset:0)!==_){let x;if(v&&v.length>0){x=[];const L=[];for(let E=0;E<v.length;E++){const I=v[E];I<_?x.push(I):I>b&&L.push(I)}x.length===0&&(x=void 0),L.length===0?v=void 0:v=L}c.splice(w,0,{endOffset:_,scopes:[...C],bracket:x,encodedLanguageId:p.encodedLanguageId}),w++,h(),d++}c.splice(w,0,{endOffset:b,scopes:[...C,p.name],bracket:u(p,_),encodedLanguageId:p.encodedLanguageId}),c[d].bracket=v}else c[d]={endOffset:b,scopes:[l,p.name],bracket:u(p,_),encodedLanguageId:p.encodedLanguageId};d++};for(let p=0;p<e.length;p++){const _=e[p],b=_.node.endIndex<i?_.node.endIndex<t?t:_.node.endIndex:i,w=_.node.startIndex<t?t:_.node.startIndex,C=b-t;let v;const y=b-w;p>0?v=c[d-1].endOffset:v=w-t-1;const x=C-y;if(v>=0&&v<x&&(c[d]={endOffset:x,scopes:[l],encodedLanguageId:this._encodedLanguageId},d++,h()),!(y<0))if(v>=C){let L=d-1,E=c[L].endOffset,I=L>=2?c[L-1].endOffset:0;do{if(I+y===E){if(I===x){c[L].scopes.push(_.name);const R=c[L].bracket;c[L].bracket=R&&R.length>0?R:u(_,x)}}else if(I<=x){g(_,x,C,L);break}L--,I=L>=1?c[L-1].endOffset:0,E=L>=0?c[L].endOffset:0}while(E>x)}else g(_,x,C)}c[d-1].endOffset<r&&r-c[d-1].endOffset>0&&(h(),c[d]={endOffset:r,scopes:c[d].scopes,encodedLanguageId:this._encodedLanguageId},d++);for(let p=0;p<c.length;p++)if(c[p].endOffset===0&&p!==0){c.splice(p,c.length-p);break}const f=o.elapsed();return{endOffsets:c,captureTime:f}}_getInjectionCaptures(e,t){return[]}_tokenizeCapturesWithMetadata(e,t,i){const n=Cn.create(),o=this._createTokensFromCaptures(e,t,i);if(!o)return;const r=o.endOffsets;for(let l=0;l<r.length;l++){const c=r[l];c.metadata=this._treeSitterThemeService.findMetadata(c.scopes,c.encodedLanguageId,!!c.bracket&&c.bracket.length>0,void 0)}const a=n.elapsed();return{endOffsetsAndMetadata:r,captureTime:o.captureTime,metadataTime:a}}_tokenizeEncoded(e){const t=this._textModel.getOffsetAt({lineNumber:e,column:1}),i=this._textModel.getLineCount(),n=e+1<=i?this._textModel.getOffsetAt({lineNumber:e+1,column:1}):this._textModel.getValueLength(),o=n-t,r=this._tokenize(new k(e,1,e,o+1),t,n);if(r)return{result:r.endOffsetsAndMetadata,captureTime:r.captureTime,metadataTime:r.metadataTime,versionId:r.versionId}}_endOffsetTokensToUint32Array(e){const t=new Uint32Array(e.length*2);for(let i=0;i<e.length;i++)t[i*2]=e[i].endOffset,t[i*2+1]=e[i].metadata;return t}};FF=aye([lye(4,sye)],FF);const cye={css:"source.css",typescript:"source.ts",ini:"source.ini",regex:"source.regex"},dye=/[\{\}\[\]\<\>\(\)]/g,u7=qe("treeSitterLibraryService");class hye{constructor(e){this._value=Re(this,void 0),this.promiseResult=this._value,this.promise=e.then(t=>(Vt(i=>{this._value.set(new MU(t,void 0),i)}),t),t=>{throw Vt(i=>{this._value.set(new MU(void 0,t),i)}),t})}}class MU{constructor(e,t){this.data=e,this.error=t}getDataOrThrow(){if(this.error)throw this.error;return this.data}}var uye=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},OU=function(s,e){return function(t,i){e(t,i,s)}};let BF=class extends lte{constructor(e,t,i,n,o,r){super(t,i),this._languageIdObs=e,this._treeSitterLibraryService=o,this._instantiationService=r,this._backgroundTokenizationState=1,this._onDidChangeBackgroundTokenizationState=this._register(new M),this.onDidChangeBackgroundTokenizationState=this._onDidChangeBackgroundTokenizationState.event;const a=new hye(this._treeSitterLibraryService.getParserClass()),l=j(this,c=>a.promiseResult?.read(c)?.getDataOrThrow());this._tree=j(this,c=>{const d=l.read(c);if(!d)return;const h=this._languageIdObs.read(c),u=this._treeSitterLibraryService.getLanguage(h,!1,c);if(!u)return;const g=new d;if(c.store.add(pe(()=>{g.delete()})),g.setLanguage(u),this._treeSitterLibraryService.getInjectionQueries(h,c)!==void 0)return c.store.add(this._instantiationService.createInstance(MF,h,void 0,g,d,this._textModel))}),this._tokenizationImpl=j(this,c=>{const d=this._tree.read(c);if(!d)return;const h=this._treeSitterLibraryService.getHighlightingQueries(d.languageId,c);if(h)return c.store.add(this._instantiationService.createInstance(FF,d,h,this._languageIdCodec,n))}),this._register(De(c=>{const d=this._tokenizationImpl.read(c);d&&(c.store.add(d.onDidChangeTokens(h=>{this._onDidChangeTokens.fire(h.changes)})),c.store.add(d.onDidChangeBackgroundTokenization(h=>{this._backgroundTokenizationState=2,this._onDidChangeBackgroundTokenizationState.fire()})))}))}getLineTokens(e){const t=this._tokenizationImpl.get();if(!t){const i=this._textModel.getLineContent(e);return Ri.createEmpty(i,this._languageIdCodec)}return t.getLineTokens(e)}todo_resetTokenization(e=!0){e&&this._onDidChangeTokens.fire({semanticTokensApplied:!1,ranges:[{fromLineNumber:1,toLineNumber:this._textModel.getLineCount()}]})}handleDidChangeAttached(){}handleDidChangeContent(e){e.isFlush?this.todo_resetTokenization(!1):this._tokenizationImpl.get()?.handleContentChanged(e),this._tree.get()?.handleContentChange(e)}forceTokenization(e){const t=this._tokenizationImpl.get();t&&(t.hasAccurateTokensForLine(e)||t.tokenizeEncoded(e))}hasAccurateTokensForLine(e){const t=this._tokenizationImpl.get();return t?t.hasAccurateTokensForLine(e):!1}isCheapToTokenize(e){return!0}getTokenTypeIfInsertingCharacter(e,t,i){return 0}tokenizeLinesAt(e,t){const i=this._tokenizationImpl.get();return i?i.tokenizeLinesAt(e,t):null}get hasTokens(){const e=this._tokenizationImpl.get();return e?e.hasTokens():!1}};BF=uye([OU(4,u7),OU(5,_e)],BF);class g7{constructor(e){this._pieces=[],this._isComplete=!1,this._languageIdCodec=e}flush(){this._pieces=[],this._isComplete=!1}isEmpty(){return this._pieces.length===0}set(e,t,i=void 0){if(this._pieces=e||[],this._isComplete=t,i)for(const n of this._pieces)n.reportIfInvalid(i)}setPartial(e,t){let i=e;if(t.length>0){const o=t[0].getRange(),r=t[t.length-1].getRange();if(!o||!r)return e;i=e.plusRange(o).plusRange(r)}let n=null;for(let o=0,r=this._pieces.length;o<r;o++){const a=this._pieces[o];if(a.endLineNumber<i.startLineNumber)continue;if(a.startLineNumber>i.endLineNumber){n=n||{index:o};break}if(a.removeTokens(i),a.isEmpty()){this._pieces.splice(o,1),o--,r--;continue}if(a.endLineNumber<i.startLineNumber)continue;if(a.startLineNumber>i.endLineNumber){n=n||{index:o};continue}const[l,c]=a.split(i);if(l.isEmpty()){n=n||{index:o};continue}c.isEmpty()||(this._pieces.splice(o,1,l,c),o++,r++,n=n||{index:o})}return n=n||{index:this._pieces.length},t.length>0&&(this._pieces=IR(this._pieces,n.index,t)),i}isComplete(){return this._isComplete}addSparseTokens(e,t){if(t.getTextLength()===0)return t;const i=this._pieces;if(i.length===0)return t;const n=g7._findFirstPieceWithLine(i,e),o=i[n].getLineTokens(e);if(!o)return t;const r=t.getCount(),a=o.getCount();let l=0;const c=[];let d=0,h=0;const u=(g,f)=>{g!==h&&(h=g,c[d++]=g,c[d++]=f)};for(let g=0;g<a;g++){const f=Math.min(o.getStartCharacter(g),t.getTextLength()),p=Math.min(o.getEndCharacter(g),t.getTextLength()),_=o.getMetadata(g),b=((_&1?2048:0)|(_&2?4096:0)|(_&4?8192:0)|(_&8?16384:0)|(_&16?16744448:0)|(_&32?4278190080:0))>>>0,w=~b>>>0;for(;l<r&&t.getEndOffset(l)<=f;)u(t.getEndOffset(l),t.getMetadata(l)),l++;for(l<r&&t.getStartOffset(l)<f&&u(f,t.getMetadata(l));l<r&&t.getEndOffset(l)<p;)u(t.getEndOffset(l),t.getMetadata(l)&w|_&b),l++;if(l<r)u(p,t.getMetadata(l)&w|_&b),t.getEndOffset(l)===p&&l++;else{const C=Math.min(Math.max(0,l-1),r-1);u(p,t.getMetadata(C)&w|_&b)}}for(;l<r;)u(t.getEndOffset(l),t.getMetadata(l)),l++;return new Ri(new Uint32Array(c),t.getLineContent(),this._languageIdCodec)}static _findFirstPieceWithLine(e,t){let i=0,n=e.length-1;for(;i<n;){let o=i+Math.floor((n-i)/2);if(e[o].endLineNumber<t)i=o+1;else if(e[o].startLineNumber>t)n=o-1;else{for(;o>i&&e[o-1].startLineNumber<=t&&t<=e[o-1].endLineNumber;)o--;return o}}return i}acceptEdit(e,t,i,n,o){for(let r=0;r<this._pieces.length;r++){const a=this._pieces[r];a.acceptEdit(e,t,i,n,o),a.isEmpty()&&(this._pieces.splice(r,1),r--)}}}const bd=new Uint32Array(0).buffer;class nc{static deleteBeginning(e,t){return e===null||e===bd?e:nc.delete(e,0,t)}static deleteEnding(e,t){if(e===null||e===bd)return e;const i=Pd(e),n=i[i.length-2];return nc.delete(e,t,n)}static delete(e,t,i){if(e===null||e===bd||t===i)return e;const n=Pd(e),o=n.length>>>1;if(t===0&&n[n.length-2]===i)return bd;const r=Ri.findIndexInTokensArray(n,t),a=r>0?n[r-1<<1]:0,l=n[r<<1];if(i<l){const g=i-t;for(let f=r;f<o;f++)n[f<<1]-=g;return e}let c,d;a!==t?(n[r<<1]=t,c=r+1<<1,d=t):(c=r<<1,d=a);const h=i-t;for(let g=r+1;g<o;g++){const f=n[g<<1]-h;f>d&&(n[c++]=f,n[c++]=n[(g<<1)+1],d=f)}if(c===n.length)return e;const u=new Uint32Array(c);return u.set(n.subarray(0,c),0),u.buffer}static append(e,t){if(t===bd)return e;if(e===bd)return t;if(e===null)return e;if(t===null)return null;const i=Pd(e),n=Pd(t),o=n.length>>>1,r=new Uint32Array(i.length+n.length);r.set(i,0);let a=i.length;const l=i[i.length-2];for(let c=0;c<o;c++)r[a++]=n[c<<1]+l,r[a++]=n[(c<<1)+1];return r.buffer}static insert(e,t,i){if(e===null||e===bd)return e;const n=Pd(e),o=n.length>>>1;let r=Ri.findIndexInTokensArray(n,t);r>0&&n[r-1<<1]===t&&r--;for(let a=r;a<o;a++)n[a<<1]+=i;return e}}function Pd(s){return s instanceof Uint32Array?s:new Uint32Array(s)}class IS{constructor(e){this._lineTokens=[],this._len=0,this._languageIdCodec=e}flush(){this._lineTokens=[],this._len=0}get hasTokens(){return this._lineTokens.length>0}getTokens(e,t,i){let n=null;if(t<this._len&&(n=this._lineTokens[t]),n!==null&&n!==bd)return new Ri(Pd(n),i,this._languageIdCodec);const o=new Uint32Array(2);return o[0]=i.length,o[1]=PU(this._languageIdCodec.encodeLanguageId(e)),new Ri(o,i,this._languageIdCodec)}static _massageTokens(e,t,i){const n=i?Pd(i):null;if(t===0){let o=!1;if(n&&n.length>1&&(o=ds.getLanguageId(n[1])!==e),!o)return bd}if(!n||n.length===0){const o=new Uint32Array(2);return o[0]=t,o[1]=PU(e),o.buffer}return n[n.length-2]=t,n.byteOffset===0&&n.byteLength===n.buffer.byteLength?n.buffer:n}_ensureLine(e){for(;e>=this._len;)this._lineTokens[this._len]=null,this._len++}_deleteLines(e,t){t!==0&&(e+t>this._len&&(t=this._len-e),this._lineTokens.splice(e,t),this._len-=t)}_insertLines(e,t){if(t===0)return;const i=[];for(let n=0;n<t;n++)i[n]=null;this._lineTokens=IR(this._lineTokens,e,i),this._len+=t}setTokens(e,t,i,n,o){const r=IS._massageTokens(this._languageIdCodec.encodeLanguageId(e),i,n);this._ensureLine(t);const a=this._lineTokens[t];return this._lineTokens[t]=r,o?!IS._equals(a,r):!1}static _equals(e,t){if(!e||!t)return!e&&!t;const i=Pd(e),n=Pd(t);if(i.length!==n.length)return!1;for(let o=0,r=i.length;o<r;o++)if(i[o]!==n[o])return!1;return!0}acceptEdit(e,t,i){this._acceptDeleteRange(e),this._acceptInsertText(new A(e.startLineNumber,e.startColumn),t,i)}_acceptDeleteRange(e){const t=e.startLineNumber-1;if(t>=this._len)return;if(e.startLineNumber===e.endLineNumber){if(e.startColumn===e.endColumn)return;this._lineTokens[t]=nc.delete(this._lineTokens[t],e.startColumn-1,e.endColumn-1);return}this._lineTokens[t]=nc.deleteEnding(this._lineTokens[t],e.startColumn-1);const i=e.endLineNumber-1;let n=null;i<this._len&&(n=nc.deleteBeginning(this._lineTokens[i],e.endColumn-1)),this._lineTokens[t]=nc.append(this._lineTokens[t],n),this._deleteLines(e.startLineNumber,e.endLineNumber-e.startLineNumber)}_acceptInsertText(e,t,i){if(t===0&&i===0)return;const n=e.lineNumber-1;if(!(n>=this._len)){if(t===0){this._lineTokens[n]=nc.insert(this._lineTokens[n],e.column-1,i);return}this._lineTokens[n]=nc.deleteEnding(this._lineTokens[n],e.column-1),this._lineTokens[n]=nc.insert(this._lineTokens[n],e.column-1,i),this._insertLines(e.lineNumber,t)}}setMultilineTokens(e,t){if(e.length===0)return{changes:[]};const i=[];for(let n=0,o=e.length;n<o;n++){const r=e[n];let a=0,l=0,c=!1;for(let d=r.startLineNumber;d<=r.endLineNumber;d++)c?(this.setTokens(t.getLanguageId(),d-1,t.getLineLength(d),r.getLineTokens(d),!1),l=d):this.setTokens(t.getLanguageId(),d-1,t.getLineLength(d),r.getLineTokens(d),!0)&&(c=!0,a=d,l=d);c&&i.push({fromLineNumber:a,toLineNumber:l})}return{changes:i}}}function PU(s){return(s<<0|0|0|32768|2<<24|1024)>>>0}class gye extends lte{constructor(e,t,i,n){super(e,t),this.getLanguageId=i,this._tokenizer=null,this._backgroundTokenizationState=1,this._onDidChangeBackgroundTokenizationState=this._register(new M),this.onDidChangeBackgroundTokenizationState=this._onDidChangeBackgroundTokenizationState.event,this._defaultBackgroundTokenizer=null,this._backgroundTokenizer=this._register(new vt),this._tokens=new IS(this._languageIdCodec),this._debugBackgroundTokenizer=this._register(new vt),this._attachedViewStates=this._register(new LR),this._register(yi.onDidChange(o=>{const r=this.getLanguageId();o.changedLanguages.indexOf(r)!==-1&&this.todo_resetTokenization()})),this.todo_resetTokenization(),this._register(n.onDidChangeVisibleRanges(({view:o,state:r})=>{if(r){let a=this._attachedViewStates.get(o);a||(a=new H1e(()=>this.refreshRanges(a.lineRanges)),this._attachedViewStates.set(o,a)),a.handleStateChange(r)}else this._attachedViewStates.deleteAndDispose(o)}))}todo_resetTokenization(e=!0){this._tokens.flush(),this._debugBackgroundTokens?.flush(),this._debugBackgroundStates&&(this._debugBackgroundStates=new PF(this._textModel.getLineCount())),e&&this._onDidChangeTokens.fire({semanticTokensApplied:!1,ranges:[{fromLineNumber:1,toLineNumber:this._textModel.getLineCount()}]});const t=()=>{if(this._textModel.isTooLargeForTokenization())return[null,null];const o=yi.get(this.getLanguageId());if(!o)return[null,null];let r;try{r=o.getInitialState()}catch(a){return Oe(a),[null,null]}return[o,r]},[i,n]=t();if(i&&n?this._tokenizer=new X1e(this._textModel.getLineCount(),i,this._textModel,this._languageIdCodec):this._tokenizer=null,this._backgroundTokenizer.clear(),this._defaultBackgroundTokenizer=null,this._tokenizer){const o={setTokens:r=>{this.setTokens(r)},backgroundTokenizationFinished:()=>{if(this._backgroundTokenizationState===2)return;const r=2;this._backgroundTokenizationState=r,this._onDidChangeBackgroundTokenizationState.fire()},setEndState:(r,a)=>{if(!this._tokenizer)return;const l=this._tokenizer.store.getFirstInvalidEndStateLineNumber();l!==null&&r>=l&&this._tokenizer?.store.setEndState(r,a)}};i&&i.createBackgroundTokenizer&&!i.backgroundTokenizerShouldOnlyVerifyTokens&&(this._backgroundTokenizer.value=i.createBackgroundTokenizer(this._textModel,o)),!this._backgroundTokenizer.value&&!this._textModel.isTooLargeForTokenization()&&(this._backgroundTokenizer.value=this._defaultBackgroundTokenizer=new eye(this._tokenizer,o),this._defaultBackgroundTokenizer.handleChanges()),i?.backgroundTokenizerShouldOnlyVerifyTokens&&i.createBackgroundTokenizer?(this._debugBackgroundTokens=new IS(this._languageIdCodec),this._debugBackgroundStates=new PF(this._textModel.getLineCount()),this._debugBackgroundTokenizer.clear(),this._debugBackgroundTokenizer.value=i.createBackgroundTokenizer(this._textModel,{setTokens:r=>{this._debugBackgroundTokens?.setMultilineTokens(r,this._textModel)},backgroundTokenizationFinished(){},setEndState:(r,a)=>{this._debugBackgroundStates?.setEndState(r,a)}})):(this._debugBackgroundTokens=void 0,this._debugBackgroundStates=void 0,this._debugBackgroundTokenizer.value=void 0)}this.refreshAllVisibleLineTokens()}handleDidChangeAttached(){this._defaultBackgroundTokenizer?.handleChanges()}handleDidChangeContent(e){if(e.isFlush)this.todo_resetTokenization(!1);else if(!e.isEolChange){for(const t of e.changes){const[i,n]=fg(t.text);this._tokens.acceptEdit(t.range,i,n),this._debugBackgroundTokens?.acceptEdit(t.range,i,n)}this._debugBackgroundStates?.acceptChanges(e.changes),this._tokenizer&&this._tokenizer.store.acceptChanges(e.changes),this._defaultBackgroundTokenizer?.handleChanges()}}setTokens(e){const{changes:t}=this._tokens.setMultilineTokens(e,this._textModel);return t.length>0&&this._onDidChangeTokens.fire({semanticTokensApplied:!1,ranges:t}),{changes:t}}refreshAllVisibleLineTokens(){const e=Ne.joinMany([...this._attachedViewStates].map(([t,i])=>i.lineRanges));this.refreshRanges(e)}refreshRanges(e){for(const t of e)this.refreshRange(t.startLineNumber,t.endLineNumberExclusive-1)}refreshRange(e,t){if(!this._tokenizer)return;e=Math.max(1,Math.min(this._textModel.getLineCount(),e)),t=Math.min(this._textModel.getLineCount(),t);const i=new OF,{heuristicTokens:n}=this._tokenizer.tokenizeHeuristically(i,e,t),o=this.setTokens(i.finalize());if(n)for(const r of o.changes)this._backgroundTokenizer.value?.requestTokens(r.fromLineNumber,r.toLineNumber+1);this._defaultBackgroundTokenizer?.checkFinished()}forceTokenization(e){const t=new OF;this._tokenizer?.updateTokensUntilLine(t,e),this.setTokens(t.finalize()),this._defaultBackgroundTokenizer?.checkFinished()}hasAccurateTokensForLine(e){return this._tokenizer?this._tokenizer.hasAccurateTokensForLine(e):!0}isCheapToTokenize(e){return this._tokenizer?this._tokenizer.isCheapToTokenize(e):!0}getLineTokens(e){const t=this._textModel.getLineContent(e),i=this._tokens.getTokens(this._textModel.getLanguageId(),e-1,t);if(this._debugBackgroundTokens&&this._debugBackgroundStates&&this._tokenizer&&this._debugBackgroundStates.getFirstInvalidEndStateLineNumberOrMax()>e&&this._tokenizer.store.getFirstInvalidEndStateLineNumberOrMax()>e){const n=this._debugBackgroundTokens.getTokens(this._textModel.getLanguageId(),e-1,t);!i.equals(n)&&this._debugBackgroundTokenizer.value?.reportMismatchingTokens&&this._debugBackgroundTokenizer.value.reportMismatchingTokens(e)}return i}getTokenTypeIfInsertingCharacter(e,t,i){if(!this._tokenizer)return 0;const n=this._textModel.validatePosition(new A(e,t));return this.forceTokenization(n.lineNumber),this._tokenizer.getTokenTypeIfInsertingCharacter(n,i)}tokenizeLinesAt(e,t){return this._tokenizer?(this.forceTokenization(e),this._tokenizer.tokenizeLinesAt(e,t)):null}get hasTokens(){return this._tokens.hasTokens}}var fye=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},cL=function(s,e){return function(t,i){e(t,i,s)}},Ak;let WF=Ak=class extends Lee{constructor(e,t,i,n,o,r,a,l){super(),this._textModel=e,this._bracketPairsTextModelPart=t,this._languageId=i,this._attachedViews=n,this._languageService=o,this._languageConfigurationService=r,this._instantiationService=a,this._treeSitterLibraryService=l,this._languageIdObs=Re(this,this._languageId),this._useTreeSitter=j(this,d=>{const h=this._languageIdObs.read(d);return this._treeSitterLibraryService.supportsLanguage(h,d)}),this.tokens=j(this,d=>{let h;return this._useTreeSitter.read(d)?h=d.store.add(this._instantiationService.createInstance(BF,this._languageIdObs,this._languageService.languageIdCodec,this._textModel,this._attachedViews.visibleLineRanges)):h=d.store.add(new gye(this._languageService.languageIdCodec,this._textModel,()=>this._languageId,this._attachedViews)),d.store.add(h.onDidChangeTokens(u=>{this._emitModelTokensChangedEvent(u)})),d.store.add(h.onDidChangeBackgroundTokenizationState(u=>{this._bracketPairsTextModelPart.handleDidChangeBackgroundTokenizationState()})),h});let c=!1;this.tokens.recomputeInitiallyAndOnChange(this._store,d=>{c&&d.todo_resetTokenization(),c=!0}),this._semanticTokens=new g7(this._languageService.languageIdCodec),this._onDidChangeLanguage=this._register(new M),this.onDidChangeLanguage=this._onDidChangeLanguage.event,this._onDidChangeLanguageConfiguration=this._register(new M),this.onDidChangeLanguageConfiguration=this._onDidChangeLanguageConfiguration.event,this._onDidChangeTokens=this._register(new M),this.onDidChangeTokens=this._onDidChangeTokens.event}handleLanguageConfigurationServiceChange(e){e.affects(this._languageId)&&this._onDidChangeLanguageConfiguration.fire({})}handleDidChangeContent(e){if(e.isFlush)this._semanticTokens.flush();else if(!e.isEolChange)for(const t of e.changes){const[i,n,o]=fg(t.text);this._semanticTokens.acceptEdit(t.range,i,n,o,t.text.length>0?t.text.charCodeAt(0):0)}this.tokens.get().handleDidChangeContent(e)}handleDidChangeAttached(){this.tokens.get().handleDidChangeAttached()}getLineTokens(e){this.validateLineNumber(e);const t=this.tokens.get().getLineTokens(e);return this._semanticTokens.addSparseTokens(e,t)}_emitModelTokensChangedEvent(e){this._textModel._isDisposing()||(this._bracketPairsTextModelPart.handleDidChangeTokens(e),this._onDidChangeTokens.fire(e))}validateLineNumber(e){if(e<1||e>this._textModel.getLineCount())throw new Ce("Illegal value for lineNumber")}get hasTokens(){return this.tokens.get().hasTokens}resetTokenization(){this.tokens.get().todo_resetTokenization()}get backgroundTokenizationState(){return this.tokens.get().backgroundTokenizationState}forceTokenization(e){this.validateLineNumber(e),this.tokens.get().forceTokenization(e)}hasAccurateTokensForLine(e){return this.validateLineNumber(e),this.tokens.get().hasAccurateTokensForLine(e)}isCheapToTokenize(e){return this.validateLineNumber(e),this.tokens.get().isCheapToTokenize(e)}tokenizeIfCheap(e){this.validateLineNumber(e),this.tokens.get().tokenizeIfCheap(e)}getTokenTypeIfInsertingCharacter(e,t,i){return this.tokens.get().getTokenTypeIfInsertingCharacter(e,t,i)}tokenizeLinesAt(e,t){return this.tokens.get().tokenizeLinesAt(e,t)}setSemanticTokens(e,t){this._semanticTokens.set(e,t,this._textModel),this._emitModelTokensChangedEvent({semanticTokensApplied:e!==null,ranges:[{fromLineNumber:1,toLineNumber:this._textModel.getLineCount()}]})}hasCompleteSemanticTokens(){return this._semanticTokens.isComplete()}hasSomeSemanticTokens(){return!this._semanticTokens.isEmpty()}setPartialSemanticTokens(e,t){if(this.hasCompleteSemanticTokens())return;const i=this._textModel.validateRange(this._semanticTokens.setPartial(e,t));this._emitModelTokensChangedEvent({semanticTokensApplied:!0,ranges:[{fromLineNumber:i.startLineNumber,toLineNumber:i.endLineNumber}]})}getWordAtPosition(e){this.assertNotDisposed();const t=this._textModel.validatePosition(e),i=this._textModel.getLineContent(t.lineNumber),n=this.getLineTokens(t.lineNumber),o=n.findTokenIndexAtOffset(t.column-1),[r,a]=Ak._findLanguageBoundaries(n,o),l=iS(t.column,this.getLanguageConfiguration(n.getLanguageId(o)).getWordDefinition(),i.substring(r,a),r);if(l&&l.startColumn<=e.column&&e.column<=l.endColumn)return l;if(o>0&&r===t.column-1){const[c,d]=Ak._findLanguageBoundaries(n,o-1),h=iS(t.column,this.getLanguageConfiguration(n.getLanguageId(o-1)).getWordDefinition(),i.substring(c,d),c);if(h&&h.startColumn<=e.column&&e.column<=h.endColumn)return h}return null}getLanguageConfiguration(e){return this._languageConfigurationService.getLanguageConfiguration(e)}static _findLanguageBoundaries(e,t){const i=e.getLanguageId(t);let n=0;for(let r=t;r>=0&&e.getLanguageId(r)===i;r--)n=e.getStartOffset(r);let o=e.getLineContent().length;for(let r=t,a=e.getCount();r<a&&e.getLanguageId(r)===i;r++)o=e.getEndOffset(r);return[n,o]}getWordUntilPosition(e){const t=this.getWordAtPosition(e);return t?{word:t.word.substr(0,e.column-t.startColumn),startColumn:t.startColumn,endColumn:e.column}:{word:"",startColumn:e.column,endColumn:e.column}}getLanguageId(){return this._languageId}getLanguageIdAtPosition(e,t){const i=this._textModel.validatePosition(new A(e,t)),n=this.getLineTokens(i.lineNumber);return n.getLanguageId(n.findTokenIndexAtOffset(i.column-1))}setLanguageId(e,t="api"){if(this._languageId===e)return;const i={oldLanguage:this._languageId,newLanguage:e,source:t};this._languageId=e,this._languageIdObs.set(e,void 0),this._bracketPairsTextModelPart.handleDidChangeLanguage(i),this._onDidChangeLanguage.fire(i),this._onDidChangeLanguageConfiguration.fire({})}};WF=Ak=fye([cL(4,Ei),cL(5,ui),cL(6,_e),cL(7,u7)],WF);const f7=qe("undoRedoService");class ute{constructor(e,t){this.resource=e,this.elements=t}}const Qb=class Qb{constructor(){this.id=Qb._ID++,this.order=1}nextOrder(){return this.id===0?0:this.order++}};Qb._ID=0,Qb.None=new Qb;let HF=Qb;const Jb=class Jb{constructor(){this.id=Jb._ID++,this.order=1}nextOrder(){return this.id===0?0:this.order++}};Jb._ID=0,Jb.None=new Jb;let vf=Jb;var FU;function pye(s,e){const t=new Map;for(const i of s){const n=e(i);let o=t.get(n);o||(o=[],t.set(n,o)),o.push(i)}return t}function mye(s,e){const t=[],i=[];for(const n of s)e.has(n)||t.push(n);for(const n of e)s.has(n)||i.push(n);return{removed:t,added:i}}function gte(s,e){const t=new Set;for(const i of e)s.has(i)&&t.add(i);return t}class BU{constructor(e,t){this.toKey=t,this._map=new Map,this[FU]="SetWithKey";for(const i of e)this.add(i)}get size(){return this._map.size}add(e){const t=this.toKey(e);return this._map.set(t,e),this}delete(e){return this._map.delete(this.toKey(e))}has(e){return this._map.has(this.toKey(e))}*entries(){for(const e of this._map.values())yield[e,e]}keys(){return this.values()}*values(){for(const e of this._map.values())yield e}clear(){this._map.clear()}forEach(e,t){this._map.forEach(i=>e.call(t,i,i,this))}[Symbol.iterator](){return this.values()}}FU=Symbol.toStringTag;var _ye=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},dL=function(s,e){return function(t,i){e(t,i,s)}},sf;function bye(s){const e=new ate;return e.acceptChunk(s),e.finish()}function wye(s){const e=new ate;let t;for(;typeof(t=s.read())=="string";)e.acceptChunk(t);return e.finish()}function WU(s,e){let t;return typeof s=="string"?t=bye(s):yge(s)?t=wye(s):t=s,t.create(e)}let hL=0;const vye=999,Cye=1e4,yye=300;class Sye{constructor(e){this._source=e,this._eos=!1}read(){if(this._eos)return null;const e=[];let t=0,i=0;do{const n=this._source.read();if(n===null)return this._eos=!0,t===0?null:e.join("");if(n.length>0&&(e[t++]=n,i+=n.length),i>=64*1024)return e.join("")}while(!0)}}const RC=()=>{throw new Error("Invalid change accessor")};var dl;let Rm=(dl=class extends F{static resolveOptions(e,t){if(t.detectIndentation){const i=vU(e,t.tabSize,t.insertSpaces);return new kk({tabSize:i.tabSize,indentSize:"tabSize",insertSpaces:i.insertSpaces,trimAutoWhitespace:t.trimAutoWhitespace,defaultEOL:t.defaultEOL,bracketPairColorizationOptions:t.bracketPairColorizationOptions})}return new kk(t)}get onDidChangeLanguage(){return this._tokenizationTextModelPart.onDidChangeLanguage}get onDidChangeLanguageConfiguration(){return this._tokenizationTextModelPart.onDidChangeLanguageConfiguration}get onDidChangeTokens(){return this._tokenizationTextModelPart.onDidChangeTokens}get onDidChangeOptions(){return this._onDidChangeOptions.event}get onDidChangeAttached(){return this._onDidChangeAttached.event}get onDidChangeLineHeight(){return this._onDidChangeLineHeight.event}get onDidChangeFont(){return this._onDidChangeFont.event}onDidChangeContent(e){return this._eventEmitter.slowEvent(t=>e(t.contentChangedEvent))}onDidChangeContentOrInjectedText(e){return Ra(this._eventEmitter.fastEvent(t=>e(t)),this._onDidChangeInjectedText.event(t=>e(t)))}_isDisposing(){return this.__isDisposing}get tokenization(){return this._tokenizationTextModelPart}get bracketPairs(){return this._bracketPairs}get guides(){return this._guidesTextModelPart}constructor(e,t,i,n=null,o,r,a,l){super(),this._undoRedoService=o,this._languageService=r,this._languageConfigurationService=a,this.instantiationService=l,this._onWillDispose=this._register(new M),this.onWillDispose=this._onWillDispose.event,this._onDidChangeDecorations=this._register(new Iye((f,p,_)=>this.handleBeforeFireDecorationsChangedEvent(f,p,_))),this.onDidChangeDecorations=this._onDidChangeDecorations.event,this._onDidChangeOptions=this._register(new M),this._onDidChangeAttached=this._register(new M),this._onDidChangeInjectedText=this._register(new M),this._onDidChangeLineHeight=this._register(new M),this._onDidChangeFont=this._register(new M),this._eventEmitter=this._register(new Dye),this._languageSelectionListener=this._register(new vt),this._deltaDecorationCallCnt=0,this._attachedViews=new F1e,hL++,this.id="$model"+hL,this.isForSimpleWidget=i.isForSimpleWidget,typeof n>"u"||n===null?this._associatedResource=Le.parse("inmemory://model/"+hL):this._associatedResource=n,this._attachedEditorCount=0;const{textBuffer:c,disposable:d}=WU(e,i.defaultEOL);this._buffer=c,this._bufferDisposable=d;const h=this._buffer.getLineCount(),u=this._buffer.getValueLengthInRange(new k(1,1,h,this._buffer.getLineLength(h)+1),0);i.largeFileOptimizations?(this._isTooLargeForTokenization=u>sf.LARGE_FILE_SIZE_THRESHOLD||h>sf.LARGE_FILE_LINE_COUNT_THRESHOLD,this._isTooLargeForHeapOperation=u>sf.LARGE_FILE_HEAP_OPERATION_THRESHOLD):(this._isTooLargeForTokenization=!1,this._isTooLargeForHeapOperation=!1),this._options=sf.resolveOptions(this._buffer,i);const g=typeof t=="string"?t:t.languageId;typeof t!="string"&&(this._languageSelectionListener.value=t.onDidChange(()=>this._setLanguage(t.languageId))),this._bracketPairs=this._register(new r1e(this,this._languageConfigurationService)),this._guidesTextModelPart=this._register(new zwe(this,this._languageConfigurationService)),this._decorationProvider=this._register(new l1e(this)),this._tokenizationTextModelPart=this.instantiationService.createInstance(WF,this,this._bracketPairs,g,this._attachedViews),this._isTooLargeForSyncing=u>sf._MODEL_SYNC_LIMIT,this._versionId=1,this._alternativeVersionId=1,this._initialUndoRedoSnapshot=null,this._isDisposed=!1,this.__isDisposing=!1,this._instanceId=sQ(hL),this._lastDecorationId=0,this._decorations=Object.create(null),this._decorationsTree=new HU,this._commandManager=new r7(this,this._undoRedoService),this._isUndoing=!1,this._isRedoing=!1,this._trimAutoWhitespaceLines=null,this._register(this._decorationProvider.onDidChange(()=>{this._onDidChangeDecorations.beginDeferredEmit(),this._onDidChangeDecorations.fire(),this._onDidChangeDecorations.endDeferredEmit()})),this._languageService.requestRichLanguageFeatures(g),this._register(this._languageConfigurationService.onDidChange(f=>{this._bracketPairs.handleLanguageConfigurationServiceChange(f),this._tokenizationTextModelPart.handleLanguageConfigurationServiceChange(f)}))}dispose(){this.__isDisposing=!0,this._onWillDispose.fire(),this._tokenizationTextModelPart.dispose(),this._isDisposed=!0,super.dispose(),this._bufferDisposable.dispose(),this.__isDisposing=!1;const e=new vw([],"",`
`,!1,!1,!0,!0);e.dispose(),this._buffer=e,this._bufferDisposable=F.None}_assertNotDisposed(){if(this._isDisposed)throw new Ce("Model is disposed!")}getTextBuffer(){return this._assertNotDisposed(),this._buffer}_emitContentChangedEvent(e,t){this.__isDisposing||(this._tokenizationTextModelPart.handleDidChangeContent(t),this._bracketPairs.handleDidChangeContent(t),this._eventEmitter.fire(new up(e,t)))}setValue(e,t=cs.setValue()){if(this._assertNotDisposed(),e==null)throw Qr();const{textBuffer:i,disposable:n}=WU(e,this._options.defaultEOL);this._setValueFromTextBuffer(i,n,t)}_createContentChanged2(e,t,i,n,o,r,a,l,c,d){return{changes:[{range:e,rangeOffset:t,rangeLength:i,text:o}],eol:this._buffer.getEOL(),isEolChange:c,versionId:this.getVersionId(),isUndoing:r,isRedoing:a,isFlush:l,detailedReasons:[d],detailedReasonsChangeLengths:[1]}}_setValueFromTextBuffer(e,t,i){this._assertNotDisposed();const n=this.getFullModelRange(),o=this.getValueLengthInRange(n),r=this.getLineCount(),a=this.getLineMaxColumn(r);this._buffer=e,this._bufferDisposable.dispose(),this._bufferDisposable=t,this._increaseVersionId(),this._decorations=Object.create(null),this._decorationsTree=new HU,this._commandManager.clear(),this._trimAutoWhitespaceLines=null,this._emitContentChangedEvent(new _w([new ICe],this._versionId,!1,!1),this._createContentChanged2(new k(1,1,r,a),0,o,new A(r,a),this.getValue(),!1,!1,!0,!1,i))}setEOL(e){this._assertNotDisposed();const t=e===1?`\r
`:`
`;if(this._buffer.getEOL()===t)return;const i=this.getFullModelRange(),n=this.getValueLengthInRange(i),o=this.getLineCount(),r=this.getLineMaxColumn(o);this._onBeforeEOLChange(),this._buffer.setEOL(t),this._increaseVersionId(),this._onAfterEOLChange(),this._emitContentChangedEvent(new _w([new ACe],this._versionId,!1,!1),this._createContentChanged2(new k(1,1,o,r),0,n,new A(o,r),this.getValue(),!1,!1,!1,!0,cs.eolChange()))}_onBeforeEOLChange(){this._decorationsTree.ensureAllNodesHaveRanges(this)}_onAfterEOLChange(){const e=this.getVersionId(),t=this._decorationsTree.collectNodesPostOrder();for(let i=0,n=t.length;i<n;i++){const o=t[i],r=o.range,a=o.cachedAbsoluteStart-o.start,l=this._buffer.getOffsetAt(r.startLineNumber,r.startColumn),c=this._buffer.getOffsetAt(r.endLineNumber,r.endColumn);o.cachedAbsoluteStart=l,o.cachedAbsoluteEnd=c,o.cachedVersionId=e,o.start=l-a,o.end=c-a,pg(o)}}onBeforeAttached(){return this._attachedEditorCount++,this._attachedEditorCount===1&&(this._tokenizationTextModelPart.handleDidChangeAttached(),this._onDidChangeAttached.fire(void 0)),this._attachedViews.attachView()}onBeforeDetached(e){this._attachedEditorCount--,this._attachedEditorCount===0&&(this._tokenizationTextModelPart.handleDidChangeAttached(),this._onDidChangeAttached.fire(void 0)),this._attachedViews.detachView(e)}isAttachedToEditor(){return this._attachedEditorCount>0}getAttachedEditorCount(){return this._attachedEditorCount}isTooLargeForSyncing(){return this._isTooLargeForSyncing}isTooLargeForTokenization(){return this._isTooLargeForTokenization}isTooLargeForHeapOperation(){return this._isTooLargeForHeapOperation}isDisposed(){return this._isDisposed}isDominatedByLongLines(){if(this._assertNotDisposed(),this.isTooLargeForTokenization())return!1;let e=0,t=0;const i=this._buffer.getLineCount();for(let n=1;n<=i;n++){const o=this._buffer.getLineLength(n);o>=Cye?t+=o:e+=o}return t>e}get uri(){return this._associatedResource}getOptions(){return this._assertNotDisposed(),this._options}getFormattingOptions(){return{tabSize:this._options.indentSize,insertSpaces:this._options.insertSpaces}}updateOptions(e){this._assertNotDisposed();const t=typeof e.tabSize<"u"?e.tabSize:this._options.tabSize,i=typeof e.indentSize<"u"?e.indentSize:this._options.originalIndentSize,n=typeof e.insertSpaces<"u"?e.insertSpaces:this._options.insertSpaces,o=typeof e.trimAutoWhitespace<"u"?e.trimAutoWhitespace:this._options.trimAutoWhitespace,r=typeof e.bracketColorizationOptions<"u"?e.bracketColorizationOptions:this._options.bracketPairColorizationOptions,a=new kk({tabSize:t,indentSize:i,insertSpaces:n,defaultEOL:this._options.defaultEOL,trimAutoWhitespace:o,bracketPairColorizationOptions:r});if(this._options.equals(a))return;const l=this._options.createChangeEvent(a);this._options=a,this._bracketPairs.handleDidChangeOptions(l),this._decorationProvider.handleDidChangeOptions(l),this._onDidChangeOptions.fire(l)}detectIndentation(e,t){this._assertNotDisposed();const i=vU(this._buffer,t,e);this.updateOptions({insertSpaces:i.insertSpaces,tabSize:i.tabSize,indentSize:i.tabSize})}normalizeIndentation(e){return this._assertNotDisposed(),fW(e,this._options.indentSize,this._options.insertSpaces)}getVersionId(){return this._assertNotDisposed(),this._versionId}mightContainRTL(){return this._buffer.mightContainRTL()}mightContainUnusualLineTerminators(){return this._buffer.mightContainUnusualLineTerminators()}removeUnusualLineTerminators(e=null){const t=this.findMatches(iQ.source,!1,!0,!1,null,!1,1073741824);this._buffer.resetMightContainUnusualLineTerminators(),this.pushEditOperations(e,t.map(i=>({range:i.range,text:null})),()=>null)}mightContainNonBasicASCII(){return this._buffer.mightContainNonBasicASCII()}getAlternativeVersionId(){return this._assertNotDisposed(),this._alternativeVersionId}getInitialUndoRedoSnapshot(){return this._assertNotDisposed(),this._initialUndoRedoSnapshot}getOffsetAt(e){this._assertNotDisposed();const t=this._validatePosition(e.lineNumber,e.column,0);return this._buffer.getOffsetAt(t.lineNumber,t.column)}getPositionAt(e){this._assertNotDisposed();const t=Math.min(this._buffer.getLength(),Math.max(0,e));return this._buffer.getPositionAt(t)}_increaseVersionId(){this._versionId=this._versionId+1,this._alternativeVersionId=this._versionId}_overwriteVersionId(e){this._versionId=e}_overwriteAlternativeVersionId(e){this._alternativeVersionId=e}_overwriteInitialUndoRedoSnapshot(e){this._initialUndoRedoSnapshot=e}getValue(e,t=!1){if(this._assertNotDisposed(),this.isTooLargeForHeapOperation())throw new Ce("Operation would exceed heap memory limits");const i=this.getFullModelRange(),n=this.getValueInRange(i,e);return t?this._buffer.getBOM()+n:n}createSnapshot(e=!1){return new Sye(this._buffer.createSnapshot(e))}getValueLength(e,t=!1){this._assertNotDisposed();const i=this.getFullModelRange(),n=this.getValueLengthInRange(i,e);return t?this._buffer.getBOM().length+n:n}getValueInRange(e,t=0){return this._assertNotDisposed(),this._buffer.getValueInRange(this.validateRange(e),t)}getValueLengthInRange(e,t=0){return this._assertNotDisposed(),this._buffer.getValueLengthInRange(this.validateRange(e),t)}getCharacterCountInRange(e,t=0){return this._assertNotDisposed(),this._buffer.getCharacterCountInRange(this.validateRange(e),t)}getLineCount(){return this._assertNotDisposed(),this._buffer.getLineCount()}getLineContent(e){if(this._assertNotDisposed(),e<1||e>this.getLineCount())throw new Ce("Illegal value for lineNumber");return this._buffer.getLineContent(e)}getLineLength(e){if(this._assertNotDisposed(),e<1||e>this.getLineCount())throw new Ce("Illegal value for lineNumber");return this._buffer.getLineLength(e)}getLinesContent(){if(this._assertNotDisposed(),this.isTooLargeForHeapOperation())throw new Ce("Operation would exceed heap memory limits");return this._buffer.getLinesContent()}getEOL(){return this._assertNotDisposed(),this._buffer.getEOL()}getEndOfLineSequence(){return this._assertNotDisposed(),this._buffer.getEOL()===`
`?0:1}getLineMinColumn(e){return this._assertNotDisposed(),1}getLineMaxColumn(e){if(this._assertNotDisposed(),e<1||e>this.getLineCount())throw new Ce("Illegal value for lineNumber");return this._buffer.getLineLength(e)+1}getLineFirstNonWhitespaceColumn(e){if(this._assertNotDisposed(),e<1||e>this.getLineCount())throw new Ce("Illegal value for lineNumber");return this._buffer.getLineFirstNonWhitespaceColumn(e)}getLineLastNonWhitespaceColumn(e){if(this._assertNotDisposed(),e<1||e>this.getLineCount())throw new Ce("Illegal value for lineNumber");return this._buffer.getLineLastNonWhitespaceColumn(e)}_validateRangeRelaxedNoAllocations(e){const t=this._buffer.getLineCount(),i=e.startLineNumber,n=e.startColumn;let o=Math.floor(typeof i=="number"&&!isNaN(i)?i:1),r=Math.floor(typeof n=="number"&&!isNaN(n)?n:1);if(o<1)o=1,r=1;else if(o>t)o=t,r=this.getLineMaxColumn(o);else if(r<=1)r=1;else{const h=this.getLineMaxColumn(o);r>=h&&(r=h)}const a=e.endLineNumber,l=e.endColumn;let c=Math.floor(typeof a=="number"&&!isNaN(a)?a:1),d=Math.floor(typeof l=="number"&&!isNaN(l)?l:1);if(c<1)c=1,d=1;else if(c>t)c=t,d=this.getLineMaxColumn(c);else if(d<=1)d=1;else{const h=this.getLineMaxColumn(c);d>=h&&(d=h)}return i===o&&n===r&&a===c&&l===d&&e instanceof k&&!(e instanceof he)?e:new k(o,r,c,d)}_isValidPosition(e,t,i){if(typeof e!="number"||typeof t!="number"||isNaN(e)||isNaN(t)||e<1||t<1||(e|0)!==e||(t|0)!==t)return!1;const n=this._buffer.getLineCount();if(e>n)return!1;if(t===1)return!0;const o=this.getLineMaxColumn(e);if(t>o)return!1;if(i===1){const r=this._buffer.getLineCharCode(e,t-2);if(sn(r))return!1}return!0}_validatePosition(e,t,i){const n=Math.floor(typeof e=="number"&&!isNaN(e)?e:1),o=Math.floor(typeof t=="number"&&!isNaN(t)?t:1),r=this._buffer.getLineCount();if(n<1)return new A(1,1);if(n>r)return new A(r,this.getLineMaxColumn(r));if(o<=1)return new A(n,1);const a=this.getLineMaxColumn(n);if(o>=a)return new A(n,a);if(i===1){const l=this._buffer.getLineCharCode(n,o-2);if(sn(l))return new A(n,o-1)}return new A(n,o)}validatePosition(e){return this._assertNotDisposed(),e instanceof A&&this._isValidPosition(e.lineNumber,e.column,1)?e:this._validatePosition(e.lineNumber,e.column,1)}isValidRange(e){return this._isValidRange(e,1)}_isValidRange(e,t){const i=e.startLineNumber,n=e.startColumn,o=e.endLineNumber,r=e.endColumn;if(!this._isValidPosition(i,n,0)||!this._isValidPosition(o,r,0))return!1;if(t===1){const a=n>1?this._buffer.getLineCharCode(i,n-2):0,l=r>1&&r<=this._buffer.getLineLength(o)?this._buffer.getLineCharCode(o,r-2):0,c=sn(a),d=sn(l);return!c&&!d}return!0}validateRange(e){if(this._assertNotDisposed(),e instanceof k&&!(e instanceof he)&&this._isValidRange(e,1))return e;const i=this._validatePosition(e.startLineNumber,e.startColumn,0),n=this._validatePosition(e.endLineNumber,e.endColumn,0),o=i.lineNumber,r=i.column,a=n.lineNumber,l=n.column;{const c=r>1?this._buffer.getLineCharCode(o,r-2):0,d=l>1&&l<=this._buffer.getLineLength(a)?this._buffer.getLineCharCode(a,l-2):0,h=sn(c),u=sn(d);return!h&&!u?new k(o,r,a,l):o===a&&r===l?new k(o,r-1,a,l-1):h&&u?new k(o,r-1,a,l+1):h?new k(o,r-1,a,l):new k(o,r,a,l+1)}}modifyPosition(e,t){this._assertNotDisposed();const i=this.getOffsetAt(e)+t;return this.getPositionAt(Math.min(this._buffer.getLength(),Math.max(0,i)))}getFullModelRange(){this._assertNotDisposed();const e=this.getLineCount();return new k(1,1,e,this.getLineMaxColumn(e))}findMatchesLineByLine(e,t,i,n){return this._buffer.findMatchesLineByLine(e,t,i,n)}findMatches(e,t,i,n,o,r,a=vye){this._assertNotDisposed();let l=null;t!==null&&typeof t!="boolean"&&(Array.isArray(t)||(t=[t]),t.every(h=>k.isIRange(h))&&(l=t.map(h=>this.validateRange(h)))),l===null&&(l=[this.getFullModelRange()]),l=l.sort((h,u)=>h.startLineNumber-u.startLineNumber||h.startColumn-u.startColumn);const c=[];c.push(l.reduce((h,u)=>k.areIntersecting(h,u)?h.plusRange(u):(c.push(h),u)));let d;if(!i&&e.indexOf(`
`)<0){const u=new nf(e,i,n,o).parseSearchRequest();if(!u)return[];d=g=>this.findMatchesLineByLine(g,u,r,a)}else d=h=>lL.findMatches(this,new nf(e,i,n,o),h,r,a);return c.map(d).reduce((h,u)=>h.concat(u),[])}findNextMatch(e,t,i,n,o,r){this._assertNotDisposed();const a=this.validatePosition(t);if(!i&&e.indexOf(`
`)<0){const c=new nf(e,i,n,o).parseSearchRequest();if(!c)return null;const d=this.getLineCount();let h=new k(a.lineNumber,a.column,d,this.getLineMaxColumn(d)),u=this.findMatchesLineByLine(h,c,r,1);return lL.findNextMatch(this,new nf(e,i,n,o),a,r),u.length>0||(h=new k(1,1,a.lineNumber,this.getLineMaxColumn(a.lineNumber)),u=this.findMatchesLineByLine(h,c,r,1),u.length>0)?u[0]:null}return lL.findNextMatch(this,new nf(e,i,n,o),a,r)}findPreviousMatch(e,t,i,n,o,r){this._assertNotDisposed();const a=this.validatePosition(t);return lL.findPreviousMatch(this,new nf(e,i,n,o),a,r)}pushStackElement(){this._commandManager.pushStackElement()}popStackElement(){this._commandManager.popStackElement()}pushEOL(e){if((this.getEOL()===`
`?0:1)!==e)try{this._onDidChangeDecorations.beginDeferredEmit(),this._eventEmitter.beginDeferredEmit(),this._initialUndoRedoSnapshot===null&&(this._initialUndoRedoSnapshot=this._undoRedoService.createSnapshot(this.uri)),this._commandManager.pushEOL(e)}finally{this._eventEmitter.endDeferredEmit(),this._onDidChangeDecorations.endDeferredEmit()}}_validateEditOperation(e){return e instanceof U2?e:new U2(e.identifier||null,this.validateRange(e.range),e.text,e.forceMoveMarkers||!1,e.isAutoWhitespaceEdit||!1,e._isTracked||!1)}_validateEditOperations(e){const t=[];for(let i=0,n=e.length;i<n;i++)t[i]=this._validateEditOperation(e[i]);return t}pushEditOperations(e,t,i,n,o){try{return this._onDidChangeDecorations.beginDeferredEmit(),this._eventEmitter.beginDeferredEmit(),this._pushEditOperations(e,this._validateEditOperations(t),i,n,o)}finally{this._eventEmitter.endDeferredEmit(),this._onDidChangeDecorations.endDeferredEmit()}}_pushEditOperations(e,t,i,n,o){if(this._options.trimAutoWhitespace&&this._trimAutoWhitespaceLines){const r=t.map(l=>({range:this.validateRange(l.range),text:l.text}));let a=!0;if(e)for(let l=0,c=e.length;l<c;l++){const d=e[l];let h=!1;for(let u=0,g=r.length;u<g;u++){const f=r[u].range,p=f.startLineNumber>d.endLineNumber,_=d.startLineNumber>f.endLineNumber;if(!p&&!_){h=!0;break}}if(!h){a=!1;break}}if(a)for(let l=0,c=this._trimAutoWhitespaceLines.length;l<c;l++){const d=this._trimAutoWhitespaceLines[l],h=this.getLineMaxColumn(d);let u=!0;for(let g=0,f=r.length;g<f;g++){const p=r[g].range,_=r[g].text;if(!(d<p.startLineNumber||d>p.endLineNumber)&&!(d===p.startLineNumber&&p.startColumn===h&&p.isEmpty()&&_&&_.length>0&&_.charAt(0)===`
`)&&!(d===p.startLineNumber&&p.startColumn===1&&p.isEmpty()&&_&&_.length>0&&_.charAt(_.length-1)===`
`)){u=!1;break}}if(u){const g=new k(d,1,d,h);t.push(new U2(null,g,null,!1,!1,!1))}}this._trimAutoWhitespaceLines=null}return this._initialUndoRedoSnapshot===null&&(this._initialUndoRedoSnapshot=this._undoRedoService.createSnapshot(this.uri)),this._commandManager.pushEditOperation(e,t,i,n,o)}_applyUndo(e,t,i,n){const o=e.map(r=>{const a=this.getPositionAt(r.newPosition),l=this.getPositionAt(r.newEnd);return{range:new k(a.lineNumber,a.column,l.lineNumber,l.column),text:r.oldText}});this._applyUndoRedoEdits(o,t,!0,!1,i,n)}_applyRedo(e,t,i,n){const o=e.map(r=>{const a=this.getPositionAt(r.oldPosition),l=this.getPositionAt(r.oldEnd);return{range:new k(a.lineNumber,a.column,l.lineNumber,l.column),text:r.newText}});this._applyUndoRedoEdits(o,t,!1,!0,i,n)}_applyUndoRedoEdits(e,t,i,n,o,r){try{this._onDidChangeDecorations.beginDeferredEmit(),this._eventEmitter.beginDeferredEmit(),this._isUndoing=i,this._isRedoing=n,this.applyEdits(e,!1),this.setEOL(t),this._overwriteAlternativeVersionId(o)}finally{this._isUndoing=!1,this._isRedoing=!1,this._eventEmitter.endDeferredEmit(r),this._onDidChangeDecorations.endDeferredEmit()}}applyEdits(e,t,i){try{this._onDidChangeDecorations.beginDeferredEmit(),this._eventEmitter.beginDeferredEmit();const n=this._validateEditOperations(e);return this._doApplyEdits(n,t??!1,i??cs.applyEdits())}finally{this._eventEmitter.endDeferredEmit(),this._onDidChangeDecorations.endDeferredEmit()}}_doApplyEdits(e,t,i){const n=this._buffer.getLineCount(),o=this._buffer.applyEdits(e,this._options.trimAutoWhitespace,t),r=this._buffer.getLineCount(),a=o.changes;if(this._trimAutoWhitespaceLines=o.trimAutoWhitespaceLineNumbers,a.length!==0){for(let d=0,h=a.length;d<h;d++){const u=a[d];this._decorationsTree.acceptReplace(u.rangeOffset,u.rangeLength,u.text.length,u.forceMoveMarkers)}const l=[];this._increaseVersionId();let c=n;for(let d=0,h=a.length;d<h;d++){const u=a[d],[g]=fg(u.text);this._onDidChangeDecorations.fire();const f=u.range.startLineNumber,p=u.range.endLineNumber,_=p-f,b=g,w=Math.min(_,b),C=b-_,v=r-c-C+f,y=v,x=v+b,L=this._decorationsTree.getInjectedTextInInterval(this,this.getOffsetAt(new A(y,1)),this.getOffsetAt(new A(x,this.getLineMaxColumn(x))),0),E=El.fromDecorations(L),I=new ph(E);for(let R=w;R>=0;R--){const O=f+R,P=v+R;I.takeFromEndWhile($=>$.lineNumber>P);const B=I.takeFromEndWhile($=>$.lineNumber===P);l.push(new gU(O,this.getLineContent(P),B))}if(w<_){const R=f+w;l.push(new NCe(R+1,p))}if(w<b){const R=new ph(E),O=f+w,P=b-w,B=r-c-P+O+1,$=[],z=[];for(let Q=0;Q<P;Q++){const J=B+Q;z[Q]=this.getLineContent(J),R.takeWhile(ne=>ne.lineNumber<J),$[Q]=R.takeWhile(ne=>ne.lineNumber===J)}l.push(new RCe(O+1,f+b,z,$))}c+=C}this._emitContentChangedEvent(new _w(l,this.getVersionId(),this._isUndoing,this._isRedoing),{changes:a,eol:this._buffer.getEOL(),isEolChange:!1,versionId:this.getVersionId(),isUndoing:this._isUndoing,isRedoing:this._isRedoing,isFlush:!1,detailedReasons:[i],detailedReasonsChangeLengths:[a.length]})}return o.reverseEdits===null?void 0:o.reverseEdits}undo(){return this._undoRedoService.undo(this.uri)}canUndo(){return this._undoRedoService.canUndo(this.uri)}redo(){return this._undoRedoService.redo(this.uri)}canRedo(){return this._undoRedoService.canRedo(this.uri)}handleBeforeFireDecorationsChangedEvent(e,t,i){if(e&&e.size>0){const o=Array.from(e).map(r=>new gU(r,this.getLineContent(r),this._getInjectedTextInLine(r)));this._onDidChangeInjectedText.fire(new Wee(o))}if(t&&t.size>0){const o=Array.from(t).map(r=>new DCe(r.ownerId,r.decorationId,r.lineNumber,r.lineHeight));this._onDidChangeLineHeight.fire(new Hee(o))}if(i&&i.size>0){const o=Array.from(i).map(r=>new TCe(r.ownerId,r.lineNumber));this._onDidChangeFont.fire(new Vee(o))}}changeDecorations(e,t=0){this._assertNotDisposed();try{return this._onDidChangeDecorations.beginDeferredEmit(),this._changeDecorations(t,e)}finally{this._onDidChangeDecorations.endDeferredEmit()}}_changeDecorations(e,t){const i={addDecoration:(o,r)=>this._deltaDecorationsImpl(e,[],[{range:o,options:r}])[0],changeDecoration:(o,r)=>{this._changeDecorationImpl(e,o,r)},changeDecorationOptions:(o,r)=>{this._changeDecorationOptionsImpl(e,o,zU(r))},removeDecoration:o=>{this._deltaDecorationsImpl(e,[o],[])},deltaDecorations:(o,r)=>o.length===0&&r.length===0?[]:this._deltaDecorationsImpl(e,o,r)};let n=null;try{n=t(i)}catch(o){Oe(o)}return i.addDecoration=RC,i.changeDecoration=RC,i.changeDecorationOptions=RC,i.removeDecoration=RC,i.deltaDecorations=RC,n}deltaDecorations(e,t,i=0){if(this._assertNotDisposed(),e||(e=[]),e.length===0&&t.length===0)return[];try{return this._deltaDecorationCallCnt++,this._deltaDecorationCallCnt>1&&(console.warn("Invoking deltaDecorations recursively could lead to leaking decorations."),Oe(new Error("Invoking deltaDecorations recursively could lead to leaking decorations."))),this._onDidChangeDecorations.beginDeferredEmit(),this._deltaDecorationsImpl(i,e,t)}finally{this._onDidChangeDecorations.endDeferredEmit(),this._deltaDecorationCallCnt--}}_getTrackedRange(e){return this.getDecorationRange(e)}_setTrackedRange(e,t,i){const n=e?this._decorations[e]:null;if(!n)return t?this._deltaDecorationsImpl(0,[],[{range:t,options:VU[i]}],!0)[0]:null;if(!t)return this._decorationsTree.delete(n),delete this._decorations[n.id],null;const o=this._validateRangeRelaxedNoAllocations(t),r=this._buffer.getOffsetAt(o.startLineNumber,o.startColumn),a=this._buffer.getOffsetAt(o.endLineNumber,o.endColumn);return this._decorationsTree.delete(n),n.reset(this.getVersionId(),r,a,o),n.setOptions(VU[i]),this._decorationsTree.insert(n),n.id}removeAllDecorationsWithOwnerId(e){if(this._isDisposed)return;const t=this._decorationsTree.collectNodesFromOwner(e);for(let i=0,n=t.length;i<n;i++){const o=t[i];this._decorationsTree.delete(o),delete this._decorations[o.id]}}getDecorationOptions(e){const t=this._decorations[e];return t?t.options:null}getDecorationRange(e){const t=this._decorations[e];return t?this._decorationsTree.getNodeRange(this,t):null}getLineDecorations(e,t=0,i=!1,n=!1){return e<1||e>this.getLineCount()?[]:this.getLinesDecorations(e,e,t,i,n)}getLinesDecorations(e,t,i=0,n=!1,o=!1,r=!1){const a=this.getLineCount(),l=Math.min(a,Math.max(1,e)),c=Math.min(a,Math.max(1,t)),d=this.getLineMaxColumn(c),h=new k(l,1,c,d),u=this._getDecorationsInRange(h,i,n,o,r);return wP(u,this._decorationProvider.getDecorationsInRange(h,i,n)),u}getDecorationsInRange(e,t=0,i=!1,n=!1,o=!1,r=!1){const a=this.validateRange(e),l=this._getDecorationsInRange(a,t,i,n,r);return wP(l,this._decorationProvider.getDecorationsInRange(a,t,i,o)),l}getOverviewRulerDecorations(e=0,t=!1,i=!1){return this._decorationsTree.getAll(this,e,t,i,!0,!1)}getInjectedTextDecorations(e=0){return this._decorationsTree.getAllInjectedText(this,e)}getCustomLineHeightsDecorations(e=0){return this._decorationsTree.getAllCustomLineHeights(this,e)}_getInjectedTextInLine(e){const t=this._buffer.getOffsetAt(e,1),i=t+this._buffer.getLineLength(e),n=this._decorationsTree.getInjectedTextInInterval(this,t,i,0);return El.fromDecorations(n).filter(o=>o.lineNumber===e)}getFontDecorationsInRange(e,t=0){const i=this._buffer.getOffsetAt(e.startLineNumber,e.startColumn),n=this._buffer.getOffsetAt(e.endLineNumber,e.endColumn);return this._decorationsTree.getFontDecorationsInInterval(this,i,n,t)}getAllDecorations(e=0,t=!1,i=!1){let n=this._decorationsTree.getAll(this,e,t,i,!1,!1);return n=n.concat(this._decorationProvider.getAllDecorations(e,t)),n}getAllMarginDecorations(e=0){return this._decorationsTree.getAll(this,e,!1,!1,!1,!0)}_getDecorationsInRange(e,t,i,n,o){const r=this._buffer.getOffsetAt(e.startLineNumber,e.startColumn),a=this._buffer.getOffsetAt(e.endLineNumber,e.endColumn);return this._decorationsTree.getAllInInterval(this,r,a,t,i,n,o)}getRangeAt(e,t){return this._buffer.getRangeAt(e,t-e)}_changeDecorationImpl(e,t,i){const n=this._decorations[t];if(!n)return;if(n.options.after){const l=this.getDecorationRange(t);this._onDidChangeDecorations.recordLineAffectedByInjectedText(l.endLineNumber)}if(n.options.before){const l=this.getDecorationRange(t);this._onDidChangeDecorations.recordLineAffectedByInjectedText(l.startLineNumber)}if(n.options.lineHeight!==null){const l=this.getDecorationRange(t);this._onDidChangeDecorations.recordLineAffectedByLineHeightChange(e,t,l.startLineNumber,null)}if(n.options.affectsFont){const l=this.getDecorationRange(t);this._onDidChangeDecorations.recordLineAffectedByFontChange(e,n.id,l.startLineNumber)}const o=this._validateRangeRelaxedNoAllocations(i),r=this._buffer.getOffsetAt(o.startLineNumber,o.startColumn),a=this._buffer.getOffsetAt(o.endLineNumber,o.endColumn);this._decorationsTree.delete(n),n.reset(this.getVersionId(),r,a,o),this._decorationsTree.insert(n),this._onDidChangeDecorations.checkAffectedAndFire(n.options),n.options.after&&this._onDidChangeDecorations.recordLineAffectedByInjectedText(o.endLineNumber),n.options.before&&this._onDidChangeDecorations.recordLineAffectedByInjectedText(o.startLineNumber),n.options.lineHeight!==null&&this._onDidChangeDecorations.recordLineAffectedByLineHeightChange(e,t,o.startLineNumber,n.options.lineHeight),n.options.affectsFont&&this._onDidChangeDecorations.recordLineAffectedByFontChange(e,n.id,o.startLineNumber)}_changeDecorationOptionsImpl(e,t,i){const n=this._decorations[t];if(!n)return;const o=!!(n.options.overviewRuler&&n.options.overviewRuler.color),r=!!(i.overviewRuler&&i.overviewRuler.color);if(this._onDidChangeDecorations.checkAffectedAndFire(n.options),this._onDidChangeDecorations.checkAffectedAndFire(i),n.options.after||i.after){const c=this._decorationsTree.getNodeRange(this,n);this._onDidChangeDecorations.recordLineAffectedByInjectedText(c.endLineNumber)}if(n.options.before||i.before){const c=this._decorationsTree.getNodeRange(this,n);this._onDidChangeDecorations.recordLineAffectedByInjectedText(c.startLineNumber)}if(n.options.lineHeight!==null||i.lineHeight!==null){const c=this._decorationsTree.getNodeRange(this,n);this._onDidChangeDecorations.recordLineAffectedByLineHeightChange(e,t,c.startLineNumber,i.lineHeight)}if(n.options.affectsFont||i.affectsFont){const c=this._decorationsTree.getNodeRange(this,n);this._onDidChangeDecorations.recordLineAffectedByFontChange(e,t,c.startLineNumber)}const a=o!==r,l=xye(i)!==Mk(n);a||l?(this._decorationsTree.delete(n),n.setOptions(i),this._decorationsTree.insert(n)):n.setOptions(i)}_deltaDecorationsImpl(e,t,i,n=!1){const o=this.getVersionId(),r=t.length;let a=0;const l=i.length;let c=0;this._onDidChangeDecorations.beginDeferredEmit();try{const d=new Array(l);for(;a<r||c<l;){let h=null;if(a<r){let u;do u=t[a++],h=this._decorations[u];while(!h&&a<r);if(h){if(h.options.after){const g=this._decorationsTree.getNodeRange(this,h);this._onDidChangeDecorations.recordLineAffectedByInjectedText(g.endLineNumber)}if(h.options.before){const g=this._decorationsTree.getNodeRange(this,h);this._onDidChangeDecorations.recordLineAffectedByInjectedText(g.startLineNumber)}if(h.options.lineHeight!==null){const g=this._decorationsTree.getNodeRange(this,h);this._onDidChangeDecorations.recordLineAffectedByLineHeightChange(e,u,g.startLineNumber,null)}if(h.options.affectsFont){const g=this._decorationsTree.getNodeRange(this,h);this._onDidChangeDecorations.recordLineAffectedByFontChange(e,u,g.startLineNumber)}this._decorationsTree.delete(h),n||this._onDidChangeDecorations.checkAffectedAndFire(h.options)}}if(c<l){if(!h){const b=++this._lastDecorationId,w=`${this._instanceId};${b}`;h=new ite(w,0,0),this._decorations[w]=h}const u=i[c],g=this._validateRangeRelaxedNoAllocations(u.range),f=zU(u.options),p=this._buffer.getOffsetAt(g.startLineNumber,g.startColumn),_=this._buffer.getOffsetAt(g.endLineNumber,g.endColumn);h.ownerId=e,h.reset(o,p,_,g),h.setOptions(f),h.options.after&&this._onDidChangeDecorations.recordLineAffectedByInjectedText(g.endLineNumber),h.options.before&&this._onDidChangeDecorations.recordLineAffectedByInjectedText(g.startLineNumber),h.options.lineHeight!==null&&this._onDidChangeDecorations.recordLineAffectedByLineHeightChange(e,h.id,g.startLineNumber,h.options.lineHeight),h.options.affectsFont&&this._onDidChangeDecorations.recordLineAffectedByFontChange(e,h.id,g.startLineNumber),n||this._onDidChangeDecorations.checkAffectedAndFire(f),this._decorationsTree.insert(h),d[c]=h.id,c++}else h&&delete this._decorations[h.id]}return d}finally{this._onDidChangeDecorations.endDeferredEmit()}}getLanguageId(){return this.tokenization.getLanguageId()}setLanguage(e,t){typeof e=="string"?(this._languageSelectionListener.clear(),this._setLanguage(e,t)):(this._languageSelectionListener.value=e.onDidChange(()=>this._setLanguage(e.languageId,t)),this._setLanguage(e.languageId,t))}_setLanguage(e,t){this.tokenization.setLanguageId(e,t),this._languageService.requestRichLanguageFeatures(e)}getLanguageIdAtPosition(e,t){return this.tokenization.getLanguageIdAtPosition(e,t)}getWordAtPosition(e){return this._tokenizationTextModelPart.getWordAtPosition(e)}getWordUntilPosition(e){return this._tokenizationTextModelPart.getWordUntilPosition(e)}normalizePosition(e,t){return e}getLineIndentColumn(e){return VF(this.getLineContent(e))+1}toString(){return`TextModel(${this.uri.toString()})`}},sf=dl,dl._MODEL_SYNC_LIMIT=50*1024*1024,dl.LARGE_FILE_SIZE_THRESHOLD=20*1024*1024,dl.LARGE_FILE_LINE_COUNT_THRESHOLD=300*1e3,dl.LARGE_FILE_HEAP_OPERATION_THRESHOLD=256*1024*1024,dl.DEFAULT_CREATION_OPTIONS={isForSimpleWidget:!1,tabSize:Qn.tabSize,indentSize:Qn.indentSize,insertSpaces:Qn.insertSpaces,detectIndentation:!1,defaultEOL:1,trimAutoWhitespace:Qn.trimAutoWhitespace,largeFileOptimizations:Qn.largeFileOptimizations,bracketPairColorizationOptions:Qn.bracketPairColorizationOptions},dl);Rm=sf=_ye([dL(4,f7),dL(5,Ei),dL(6,ui),dL(7,_e)],Rm);function VF(s){let e=0;for(const t of s)if(t===" "||t==="	")e++;else break;return e}function _M(s){return!!(s.options.overviewRuler&&s.options.overviewRuler.color)}function xye(s){return!!s.after||!!s.before}function Mk(s){return!!s.options.after||!!s.options.before}class HU{constructor(){this._decorationsTree0=new fM,this._decorationsTree1=new fM,this._injectedTextDecorationsTree=new fM}ensureAllNodesHaveRanges(e){this.getAll(e,0,!1,!1,!1,!1)}_ensureNodesHaveRanges(e,t){for(const i of t)i.range===null&&(i.range=e.getRangeAt(i.cachedAbsoluteStart,i.cachedAbsoluteEnd));return t}getAllInInterval(e,t,i,n,o,r,a){const l=e.getVersionId(),c=this._intervalSearch(t,i,n,o,r,l,a);return this._ensureNodesHaveRanges(e,c)}_intervalSearch(e,t,i,n,o,r,a){const l=this._decorationsTree0.intervalSearch(e,t,i,n,o,r,a),c=this._decorationsTree1.intervalSearch(e,t,i,n,o,r,a),d=this._injectedTextDecorationsTree.intervalSearch(e,t,i,n,o,r,a);return l.concat(c).concat(d)}getInjectedTextInInterval(e,t,i,n){const o=e.getVersionId(),r=this._injectedTextDecorationsTree.intervalSearch(t,i,n,!1,!1,o,!1);return this._ensureNodesHaveRanges(e,r).filter(a=>a.options.showIfCollapsed||!a.range.isEmpty())}getFontDecorationsInInterval(e,t,i,n){const o=e.getVersionId(),r=this._decorationsTree0.intervalSearch(t,i,n,!1,!1,o,!1);return this._ensureNodesHaveRanges(e,r).filter(a=>a.options.affectsFont)}getAllInjectedText(e,t){const i=e.getVersionId(),n=this._injectedTextDecorationsTree.search(t,!1,!1,i,!1);return this._ensureNodesHaveRanges(e,n).filter(o=>o.options.showIfCollapsed||!o.range.isEmpty())}getAllCustomLineHeights(e,t){const i=e.getVersionId(),n=this._search(t,!1,!1,!1,i,!1);return this._ensureNodesHaveRanges(e,n).filter(o=>typeof o.options.lineHeight=="number")}getAll(e,t,i,n,o,r){const a=e.getVersionId(),l=this._search(t,i,n,o,a,r);return this._ensureNodesHaveRanges(e,l)}_search(e,t,i,n,o,r){if(n)return this._decorationsTree1.search(e,t,i,o,r);{const a=this._decorationsTree0.search(e,t,i,o,r),l=this._decorationsTree1.search(e,t,i,o,r),c=this._injectedTextDecorationsTree.search(e,t,i,o,r);return a.concat(l).concat(c)}}collectNodesFromOwner(e){const t=this._decorationsTree0.collectNodesFromOwner(e),i=this._decorationsTree1.collectNodesFromOwner(e),n=this._injectedTextDecorationsTree.collectNodesFromOwner(e);return t.concat(i).concat(n)}collectNodesPostOrder(){const e=this._decorationsTree0.collectNodesPostOrder(),t=this._decorationsTree1.collectNodesPostOrder(),i=this._injectedTextDecorationsTree.collectNodesPostOrder();return e.concat(t).concat(i)}insert(e){Mk(e)?this._injectedTextDecorationsTree.insert(e):_M(e)?this._decorationsTree1.insert(e):this._decorationsTree0.insert(e)}delete(e){Mk(e)?this._injectedTextDecorationsTree.delete(e):_M(e)?this._decorationsTree1.delete(e):this._decorationsTree0.delete(e)}getNodeRange(e,t){const i=e.getVersionId();return t.cachedVersionId!==i&&this._resolveNode(t,i),t.range===null&&(t.range=e.getRangeAt(t.cachedAbsoluteStart,t.cachedAbsoluteEnd)),t.range}_resolveNode(e,t){Mk(e)?this._injectedTextDecorationsTree.resolveNode(e,t):_M(e)?this._decorationsTree1.resolveNode(e,t):this._decorationsTree0.resolveNode(e,t)}acceptReplace(e,t,i,n){this._decorationsTree0.acceptReplace(e,t,i,n),this._decorationsTree1.acceptReplace(e,t,i,n),this._injectedTextDecorationsTree.acceptReplace(e,t,i,n)}}function Fl(s){return s.replace(/[^a-z0-9\-_]/gi," ")}class fte{constructor(e){this.color=e.color||"",this.darkColor=e.darkColor||""}}class Lye extends fte{constructor(e){super(e),this._resolvedColor=null,this.position=typeof e.position=="number"?e.position:Sr.Center}getColor(e){return this._resolvedColor||(wh(e.type)&&this.darkColor?this._resolvedColor=this._resolveColor(this.darkColor,e):this._resolvedColor=this._resolveColor(this.color,e)),this._resolvedColor}invalidateCachedColor(){this._resolvedColor=null}_resolveColor(e,t){if(typeof e=="string")return e;const i=e?t.getColor(e.id):null;return i?i.toString():""}}class kye{constructor(e){this.position=e?.position??Ll.Center,this.persistLane=e?.persistLane}}class Eye extends fte{constructor(e){super(e),this.position=e.position,this.sectionHeaderStyle=e.sectionHeaderStyle??null,this.sectionHeaderText=e.sectionHeaderText??null}getColor(e){return this._resolvedColor||(wh(e.type)&&this.darkColor?this._resolvedColor=this._resolveColor(this.darkColor,e):this._resolvedColor=this._resolveColor(this.color,e)),this._resolvedColor}invalidateCachedColor(){this._resolvedColor=void 0}_resolveColor(e,t){return typeof e=="string"?q.fromHex(e):t.getColor(e.id)}}class mg{static from(e){return e instanceof mg?e:new mg(e)}constructor(e){this.content=e.content||"",this.tokens=e.tokens??null,this.inlineClassName=e.inlineClassName||null,this.inlineClassNameAffectsLetterSpacing=e.inlineClassNameAffectsLetterSpacing||!1,this.attachedData=e.attachedData||null,this.cursorStops=e.cursorStops||null}}class Fe{static register(e){return new Fe(e)}static createDynamic(e){return new Fe(e)}constructor(e){this.description=e.description,this.blockClassName=e.blockClassName?Fl(e.blockClassName):null,this.blockDoesNotCollapse=e.blockDoesNotCollapse??null,this.blockIsAfterEnd=e.blockIsAfterEnd??null,this.blockPadding=e.blockPadding??null,this.stickiness=e.stickiness||0,this.zIndex=e.zIndex||0,this.className=e.className?Fl(e.className):null,this.shouldFillLineOnLineBreak=e.shouldFillLineOnLineBreak??null,this.hoverMessage=e.hoverMessage||null,this.glyphMarginHoverMessage=e.glyphMarginHoverMessage||null,this.lineNumberHoverMessage=e.lineNumberHoverMessage||null,this.isWholeLine=e.isWholeLine||!1,this.lineHeight=e.lineHeight?Math.min(e.lineHeight,yye):null,this.fontSize=e.fontSize||null,this.affectsFont=!!e.fontSize||!!e.fontFamily||!!e.fontWeight||!!e.fontStyle,this.showIfCollapsed=e.showIfCollapsed||!1,this.collapseOnReplaceEdit=e.collapseOnReplaceEdit||!1,this.overviewRuler=e.overviewRuler?new Lye(e.overviewRuler):null,this.minimap=e.minimap?new Eye(e.minimap):null,this.glyphMargin=e.glyphMarginClassName?new kye(e.glyphMargin):null,this.glyphMarginClassName=e.glyphMarginClassName?Fl(e.glyphMarginClassName):null,this.linesDecorationsClassName=e.linesDecorationsClassName?Fl(e.linesDecorationsClassName):null,this.lineNumberClassName=e.lineNumberClassName?Fl(e.lineNumberClassName):null,this.linesDecorationsTooltip=e.linesDecorationsTooltip?Mde(e.linesDecorationsTooltip):null,this.firstLineDecorationClassName=e.firstLineDecorationClassName?Fl(e.firstLineDecorationClassName):null,this.marginClassName=e.marginClassName?Fl(e.marginClassName):null,this.inlineClassName=e.inlineClassName?Fl(e.inlineClassName):null,this.inlineClassNameAffectsLetterSpacing=e.inlineClassNameAffectsLetterSpacing||!1,this.beforeContentClassName=e.beforeContentClassName?Fl(e.beforeContentClassName):null,this.afterContentClassName=e.afterContentClassName?Fl(e.afterContentClassName):null,this.after=e.after?mg.from(e.after):null,this.before=e.before?mg.from(e.before):null,this.hideInCommentTokens=e.hideInCommentTokens??!1,this.hideInStringTokens=e.hideInStringTokens??!1,this.textDirection=e.textDirection??null}}Fe.EMPTY=Fe.register({description:"empty"});const VU=[Fe.register({description:"tracked-range-always-grows-when-typing-at-edges",stickiness:0}),Fe.register({description:"tracked-range-never-grows-when-typing-at-edges",stickiness:1}),Fe.register({description:"tracked-range-grows-only-when-typing-before",stickiness:2}),Fe.register({description:"tracked-range-grows-only-when-typing-after",stickiness:3})];function zU(s){return s instanceof Fe?s:Fe.createDynamic(s)}class $U{static toKey(e){return`${e.ownerId};${e.decorationId};${e.lineNumber}`}constructor(e,t,i,n){this.ownerId=e,this.decorationId=t,this.lineNumber=i,this.lineHeight=n}}class UU{static toKey(e){return`${e.ownerId};${e.decorationId};${e.lineNumber}`}constructor(e,t,i){this.ownerId=e,this.decorationId=t,this.lineNumber=i}}class Iye extends F{constructor(e){super(),this.handleBeforeFire=e,this._actual=this._register(new M),this.event=this._actual.event,this._affectedInjectedTextLines=null,this._affectedLineHeights=null,this._affectedFontLines=null,this._deferredCnt=0,this._shouldFireDeferred=!1,this._affectsMinimap=!1,this._affectsOverviewRuler=!1,this._affectsGlyphMargin=!1,this._affectsLineNumber=!1}beginDeferredEmit(){this._deferredCnt++}endDeferredEmit(){this._deferredCnt--,this._deferredCnt===0&&(this._shouldFireDeferred&&this.doFire(),this._affectedInjectedTextLines?.clear(),this._affectedInjectedTextLines=null,this._affectedLineHeights?.clear(),this._affectedLineHeights=null,this._affectedFontLines?.clear(),this._affectedFontLines=null)}recordLineAffectedByInjectedText(e){this._affectedInjectedTextLines||(this._affectedInjectedTextLines=new Set),this._affectedInjectedTextLines.add(e)}recordLineAffectedByLineHeightChange(e,t,i,n){this._affectedLineHeights||(this._affectedLineHeights=new BU([],$U.toKey)),this._affectedLineHeights.add(new $U(e,t,i,n))}recordLineAffectedByFontChange(e,t,i){this._affectedFontLines||(this._affectedFontLines=new BU([],UU.toKey)),this._affectedFontLines.add(new UU(e,t,i))}checkAffectedAndFire(e){this._affectsMinimap||=!!e.minimap?.position,this._affectsOverviewRuler||=!!e.overviewRuler?.color,this._affectsGlyphMargin||=!!e.glyphMarginClassName,this._affectsLineNumber||=!!e.lineNumberClassName,this.tryFire()}fire(){this._affectsMinimap=!0,this._affectsOverviewRuler=!0,this._affectsGlyphMargin=!0,this.tryFire()}tryFire(){this._deferredCnt===0?this.doFire():this._shouldFireDeferred=!0}doFire(){this.handleBeforeFire(this._affectedInjectedTextLines,this._affectedLineHeights,this._affectedFontLines);const e={affectsMinimap:this._affectsMinimap,affectsOverviewRuler:this._affectsOverviewRuler,affectsGlyphMargin:this._affectsGlyphMargin,affectsLineNumber:this._affectsLineNumber};this._shouldFireDeferred=!1,this._affectsMinimap=!1,this._affectsOverviewRuler=!1,this._affectsGlyphMargin=!1,this._actual.fire(e)}}class Dye extends F{constructor(){super(),this._fastEmitter=this._register(new M),this.fastEvent=this._fastEmitter.event,this._slowEmitter=this._register(new M),this.slowEvent=this._slowEmitter.event,this._deferredCnt=0,this._deferredEvent=null}beginDeferredEmit(){this._deferredCnt++}endDeferredEmit(e=null){if(this._deferredCnt--,this._deferredCnt===0&&this._deferredEvent!==null){this._deferredEvent.rawContentChangedEvent.resultingSelection=e;const t=this._deferredEvent;this._deferredEvent=null,this._fastEmitter.fire(t),this._slowEmitter.fire(t)}}fire(e){if(this._deferredCnt>0){this._deferredEvent?this._deferredEvent=this._deferredEvent.merge(e):this._deferredEvent=e;return}this._fastEmitter.fire(e),this._slowEmitter.fire(e)}}const ge=qe("ILanguageFeaturesService");class p7{static create(e){return new p7(e.get(151),e.get(150))}constructor(e,t){this.classifier=new Tye(e,t)}createLineBreaksComputer(e,t,i,n,o,r){const a=[],l=[],c=[];return{addRequest:(d,h,u)=>{a.push(d),l.push(h),c.push(u)},finalize:()=>{const d=e.typicalFullwidthCharacterWidth/e.typicalHalfwidthCharacterWidth,h=[];for(let u=0,g=a.length;u<g;u++){const f=l[u],p=c[u],_=a[u],b=r&&_.includes('"')&&_.includes("\\n");p&&!p.injectionOptions&&!f&&!b?h[u]=Nye(this.classifier,p,_,t,i,d,n,o):h[u]=Rye(this.classifier,_,f,t,i,d,n,o,b)}return zF.length=0,$F.length=0,h}}}}class Tye extends Zv{constructor(e,t){super(0);for(let i=0;i<e.length;i++)this.set(e.charCodeAt(i),1);for(let i=0;i<t.length;i++)this.set(t.charCodeAt(i),2)}get(e){return e>=0&&e<256?this._asciiMap[e]:e>=12352&&e<=12543||e>=13312&&e<=19903||e>=19968&&e<=40959?3:this._map.get(e)||this._defaultValue}}let zF=[],$F=[];function Nye(s,e,t,i,n,o,r,a){if(n===-1)return null;const l=t.length;if(l<=1)return null;const c=a==="keepAll",d=e.breakOffsets,h=e.breakOffsetsVisibleColumn,u=pte(t,i,n,o,r),g=n-u,f=zF,p=$F;let _=0,b=0,w=0,C=n;const v=d.length;let y=0;if(y>=0){let x=Math.abs(h[y]-C);for(;y+1<v;){const L=Math.abs(h[y+1]-C);if(L>=x)break;x=L,y++}}for(;y<v;){let x=y<0?0:d[y],L=y<0?0:h[y];b>x&&(x=b,L=w);let E=0,I=0,R=0,O=0;if(L<=C){let B=L,$=x===0?0:t.charCodeAt(x-1),z=x===0?0:s.get($),Q=!0;for(let J=x;J<l;J++){const ne=J,de=t.charCodeAt(J);let ke,Me;if(sn(de)?(J++,ke=0,Me=2):(ke=s.get(de),Me=V1(de,B,i,o)),ne>b&&UF($,z,de,ke,c)&&(E=ne,I=B),B+=Me,B>C){ne>b?(R=ne,O=B-Me):(R=J+1,O=B),B-I>g&&(E=0),Q=!1;break}$=de,z=ke}if(Q){_>0&&(f[_]=d[d.length-1],p[_]=h[d.length-1],_++);break}}if(E===0){let B=L,$=t.charCodeAt(x),z=s.get($),Q=!1;for(let J=x-1;J>=b;J--){const ne=J+1,de=t.charCodeAt(J);if(de===9){Q=!0;break}let ke,Me;if(ag(de)?(J--,ke=0,Me=2):(ke=s.get(de),Me=gh(de)?o:1),B<=C){if(R===0&&(R=ne,O=B),B<=C-g)break;if(UF(de,ke,$,z,c)){E=ne,I=B;break}}B-=Me,$=de,z=ke}if(E!==0){const J=g-(O-I);if(J<=i){const ne=t.charCodeAt(R);let de;sn(ne)?de=2:de=V1(ne,O,i,o),J-de<0&&(E=0)}}if(Q){y--;continue}}if(E===0&&(E=R,I=O),E<=b){const B=t.charCodeAt(b);sn(B)?(E=b+2,I=w+2):(E=b+1,I=w+V1(B,w,i,o))}for(b=E,f[_]=E,w=I,p[_]=I,_++,C=I+g;y<0||y<v&&h[y]<I;)y++;let P=Math.abs(h[y]-C);for(;y+1<v;){const B=Math.abs(h[y+1]-C);if(B>=P)break;P=B,y++}}return _===0?null:(f.length=_,p.length=_,zF=e.breakOffsets,$F=e.breakOffsetsVisibleColumn,e.breakOffsets=f,e.breakOffsetsVisibleColumn=p,e.wrappedTextIndentLength=u,e)}function Rye(s,e,t,i,n,o,r,a,l){const c=El.applyInjectedText(e,t);let d,h;if(t&&t.length>0?(d=t.map(R=>R.options),h=t.map(R=>R.column-1)):(d=null,h=null),n===-1)return d?new O1(h,d,[c.length],[],0):null;const u=c.length;if(u<=1)return d?new O1(h,d,[c.length],[],0):null;const g=a==="keepAll",f=pte(c,i,n,o,r),p=n-f,_=[],b=[];let w=0,C=0,v=0,y=n,x=c.charCodeAt(0),L=s.get(x),E=V1(x,0,i,o),I=1;sn(x)&&(E+=1,x=c.charCodeAt(1),L=s.get(x),I++);for(let R=I;R<u;R++){const O=R,P=c.charCodeAt(R);let B,$,z=!1;sn(P)?(R++,B=0,$=2):(B=s.get(P),$=V1(P,E,i,o)),l&&Aye(c,R)?(C=O,v=E,z=!0):UF(x,L,P,B,g)&&(C=O,v=E),E+=$,(E>y||z)&&((C===0||E-v>p)&&(C=O,v=E-$),_[w]=C,b[w]=v,w++,y=v+p,C=0),x=P,L=B}return w===0&&(!t||t.length===0)?null:(_[w]=u,b[w]=E,new O1(h,d,_,b,f))}function V1(s,e,t,i){return s===9?t-e%t:gh(s)||s<32?i:1}function qU(s,e){return e-s%e}function Aye(s,e){if(e>=2&&s.charAt(e-1)==="n"){let t=0;for(let i=e-2;i>=0;i--)if(s.charAt(i)==="\\")t++;else return t%2===1}return!1}function UF(s,e,t,i,n){return t!==32&&(e===2&&i!==2||e!==1&&i===1||!n&&e===3&&i!==2||!n&&i===3&&e!==1)}function pte(s,e,t,i,n){let o=0;if(n!==0){const r=As(s);if(r!==-1){for(let l=0;l<r;l++){const c=s.charCodeAt(l)===9?qU(o,e):1;o+=c}const a=n===3?2:n===2?1:0;for(let l=0;l<a;l++){const c=qU(o,e);o+=c}o+i>t&&(o=0)}}return o}class gI{constructor(e){this._selTrackedRange=null,this._trackSelection=!0,this._setState(e,new An(new k(1,1,1,1),0,0,new A(1,1),0),new An(new k(1,1,1,1),0,0,new A(1,1),0))}dispose(e){this._removeTrackedRange(e)}startTrackingSelection(e){this._trackSelection=!0,this._updateTrackedRange(e)}stopTrackingSelection(e){this._trackSelection=!1,this._removeTrackedRange(e)}_updateTrackedRange(e){this._trackSelection&&(this._selTrackedRange=e.model._setTrackedRange(this._selTrackedRange,this.modelState.selection,0))}_removeTrackedRange(e){this._selTrackedRange=e.model._setTrackedRange(this._selTrackedRange,null,0)}asCursorState(){return new ut(this.modelState,this.viewState)}readSelectionFromMarkers(e){const t=e.model._getTrackedRange(this._selTrackedRange);return this.modelState.selection.isEmpty()&&!t.isEmpty()?he.fromRange(t.collapseToEnd(),this.modelState.selection.getDirection()):he.fromRange(t,this.modelState.selection.getDirection())}ensureValidState(e){this._setState(e,this.modelState,this.viewState)}setState(e,t,i){this._setState(e,t,i)}static _validatePositionWithCache(e,t,i,n){return t.equals(i)?n:e.normalizePosition(t,2)}static _validateViewState(e,t){const i=t.position,n=t.selectionStart.getStartPosition(),o=t.selectionStart.getEndPosition(),r=e.normalizePosition(i,2),a=this._validatePositionWithCache(e,n,i,r),l=this._validatePositionWithCache(e,o,n,a);return i.equals(r)&&n.equals(a)&&o.equals(l)?t:new An(k.fromPositions(a,l),t.selectionStartKind,t.selectionStartLeftoverVisibleColumns+n.column-a.column,r,t.leftoverVisibleColumns+i.column-r.column)}_setState(e,t,i){if(i&&(i=gI._validateViewState(e.viewModel,i)),t){const n=e.model.validateRange(t.selectionStart),o=t.selectionStart.equalsRange(n)?t.selectionStartLeftoverVisibleColumns:0,r=e.model.validatePosition(t.position),a=t.position.equals(r)?t.leftoverVisibleColumns:0;t=new An(n,t.selectionStartKind,o,r,a)}else{if(!i)return;const n=e.model.validateRange(e.coordinatesConverter.convertViewRangeToModelRange(i.selectionStart)),o=e.model.validatePosition(e.coordinatesConverter.convertViewPositionToModelPosition(i.position));t=new An(n,i.selectionStartKind,i.selectionStartLeftoverVisibleColumns,o,i.leftoverVisibleColumns)}if(i){const n=e.coordinatesConverter.validateViewRange(i.selectionStart,t.selectionStart),o=e.coordinatesConverter.validateViewPosition(i.position,t.position);i=new An(n,t.selectionStartKind,t.selectionStartLeftoverVisibleColumns,o,t.leftoverVisibleColumns)}else{const n=e.coordinatesConverter.convertModelPositionToViewPosition(new A(t.selectionStart.startLineNumber,t.selectionStart.startColumn)),o=e.coordinatesConverter.convertModelPositionToViewPosition(new A(t.selectionStart.endLineNumber,t.selectionStart.endColumn)),r=new k(n.lineNumber,n.column,o.lineNumber,o.column),a=e.coordinatesConverter.convertModelPositionToViewPosition(t.position);i=new An(r,t.selectionStartKind,t.selectionStartLeftoverVisibleColumns,a,t.leftoverVisibleColumns)}this.modelState=t,this.viewState=i,this._updateTrackedRange(e)}}class jU{constructor(e){this.context=e,this.cursors=[new gI(e)],this.lastAddedCursorIndex=0}dispose(){for(const e of this.cursors)e.dispose(this.context)}startTrackingSelections(){for(const e of this.cursors)e.startTrackingSelection(this.context)}stopTrackingSelections(){for(const e of this.cursors)e.stopTrackingSelection(this.context)}updateContext(e){this.context=e}ensureValidState(){for(const e of this.cursors)e.ensureValidState(this.context)}readSelectionFromMarkers(){return this.cursors.map(e=>e.readSelectionFromMarkers(this.context))}getAll(){return this.cursors.map(e=>e.asCursorState())}getViewPositions(){return this.cursors.map(e=>e.viewState.position)}getTopMostViewPosition(){return xee(this.cursors,ns(e=>e.viewState.position,A.compare)).viewState.position}getBottomMostViewPosition(){return See(this.cursors,ns(e=>e.viewState.position,A.compare)).viewState.position}getSelections(){return this.cursors.map(e=>e.modelState.selection)}getViewSelections(){return this.cursors.map(e=>e.viewState.selection)}setSelections(e){this.setStates(ut.fromModelSelections(e))}getPrimaryCursor(){return this.cursors[0].asCursorState()}setStates(e){e!==null&&(this.cursors[0].setState(this.context,e[0].modelState,e[0].viewState),this._setSecondaryStates(e.slice(1)))}_setSecondaryStates(e){const t=this.cursors.length-1,i=e.length;if(t<i){const n=i-t;for(let o=0;o<n;o++)this._addSecondaryCursor()}else if(t>i){const n=t-i;for(let o=0;o<n;o++)this._removeSecondaryCursor(this.cursors.length-2)}for(let n=0;n<i;n++)this.cursors[n+1].setState(this.context,e[n].modelState,e[n].viewState)}killSecondaryCursors(){this._setSecondaryStates([])}_addSecondaryCursor(){this.cursors.push(new gI(this.context)),this.lastAddedCursorIndex=this.cursors.length-1}getLastAddedCursorIndex(){return this.cursors.length===1||this.lastAddedCursorIndex===0?0:this.lastAddedCursorIndex}_removeSecondaryCursor(e){this.lastAddedCursorIndex>=e+1&&this.lastAddedCursorIndex--,this.cursors[e+1].dispose(this.context),this.cursors.splice(e+1,1)}normalize(){if(this.cursors.length===1)return;const e=this.cursors.slice(0),t=[];for(let i=0,n=e.length;i<n;i++)t.push({index:i,selection:e[i].modelState.selection});t.sort(ns(i=>i.selection,k.compareRangesUsingStarts));for(let i=0;i<t.length-1;i++){const n=t[i],o=t[i+1],r=n.selection,a=o.selection;if(!this.context.cursorConfig.multiCursorMergeOverlapping)continue;let l;if(a.isEmpty()||r.isEmpty()?l=a.getStartPosition().isBeforeOrEqual(r.getEndPosition()):l=a.getStartPosition().isBefore(r.getEndPosition()),l){const c=n.index<o.index?i:i+1,d=n.index<o.index?i+1:i,h=t[d].index,u=t[c].index,g=t[d].selection,f=t[c].selection;if(!g.equalsSelection(f)){const p=g.plusRange(f),_=g.selectionStartLineNumber===g.startLineNumber&&g.selectionStartColumn===g.startColumn,b=f.selectionStartLineNumber===f.startLineNumber&&f.selectionStartColumn===f.startColumn;let w;h===this.lastAddedCursorIndex?(w=_,this.lastAddedCursorIndex=u):w=b;let C;w?C=new he(p.startLineNumber,p.startColumn,p.endLineNumber,p.endColumn):C=new he(p.endLineNumber,p.endColumn,p.startLineNumber,p.startColumn),t[c].selection=C;const v=ut.fromModelSelection(C);e[u].setState(this.context,v.modelState,v.viewState)}for(const p of t)p.index>h&&p.index--;e.splice(h,1),t.splice(d,1),this._removeSecondaryCursor(h-1),i--}}}}class GU{constructor(e,t,i,n){this._cursorContextBrand=void 0,this.model=e,this.viewModel=t,this.coordinatesConverter=i,this.cursorConfig=n}}class Mye{constructor(){this.type=0}}class Oye{constructor(){this.type=1}}class Pye{constructor(e){this.type=2,this._source=e}hasChanged(e){return this._source.hasChanged(e)}}class Fye{constructor(e,t,i){this.selections=e,this.modelSelections=t,this.reason=i,this.type=3}}class Hg{constructor(e){this.type=4,e?(this.affectsMinimap=e.affectsMinimap,this.affectsOverviewRuler=e.affectsOverviewRuler,this.affectsGlyphMargin=e.affectsGlyphMargin,this.affectsLineNumber=e.affectsLineNumber):(this.affectsMinimap=!0,this.affectsOverviewRuler=!0,this.affectsGlyphMargin=!0,this.affectsLineNumber=!0)}}class uL{constructor(){this.type=5}}class Bye{constructor(e){this.type=6,this.isFocused=e}}class Wye{constructor(){this.type=7}}class gL{constructor(){this.type=8}}class mte{constructor(e,t){this.fromLineNumber=e,this.count=t,this.type=9}}class qF{constructor(e,t){this.type=10,this.fromLineNumber=e,this.toLineNumber=t}}class jF{constructor(e,t){this.type=11,this.fromLineNumber=e,this.toLineNumber=t}}class z1{constructor(e,t,i,n,o,r,a){this.source=e,this.minimalReveal=t,this.range=i,this.selections=n,this.verticalType=o,this.revealHorizontal=r,this.scrollType=a,this.type=12}}class Hye{constructor(e){this.type=13,this.scrollWidth=e.scrollWidth,this.scrollLeft=e.scrollLeft,this.scrollHeight=e.scrollHeight,this.scrollTop=e.scrollTop,this.scrollWidthChanged=e.scrollWidthChanged,this.scrollLeftChanged=e.scrollLeftChanged,this.scrollHeightChanged=e.scrollHeightChanged,this.scrollTopChanged=e.scrollTopChanged}}class Vye{constructor(e){this.theme=e,this.type=14}}class zye{constructor(e){this.type=15,this.ranges=e}}class $ye{constructor(){this.type=16}}let Uye=class{constructor(){this.type=17}};class qye extends F{constructor(){super(),this._onEvent=this._register(new M),this.onEvent=this._onEvent.event,this._eventHandlers=[],this._viewEventQueue=null,this._isConsumingViewEventQueue=!1,this._collector=null,this._collectorCnt=0,this._outgoingEvents=[]}emitOutgoingEvent(e){this._addOutgoingEvent(e),this._emitOutgoingEvents()}_addOutgoingEvent(e){for(let t=0,i=this._outgoingEvents.length;t<i;t++){const n=this._outgoingEvents[t].kind===e.kind?this._outgoingEvents[t].attemptToMerge(e):null;if(n){this._outgoingEvents[t]=n;return}}this._outgoingEvents.push(e)}_emitOutgoingEvents(){for(;this._outgoingEvents.length>0;){if(this._collector||this._isConsumingViewEventQueue)return;const e=this._outgoingEvents.shift();e.isNoOp()||this._onEvent.fire(e)}}addViewEventHandler(e){for(let t=0,i=this._eventHandlers.length;t<i;t++)this._eventHandlers[t]===e&&console.warn("Detected duplicate listener in ViewEventDispatcher",e);this._eventHandlers.push(e)}removeViewEventHandler(e){for(let t=0;t<this._eventHandlers.length;t++)if(this._eventHandlers[t]===e){this._eventHandlers.splice(t,1);break}}beginEmitViewEvents(){return this._collectorCnt++,this._collectorCnt===1&&(this._collector=new jye),this._collector}endEmitViewEvents(){if(this._collectorCnt--,this._collectorCnt===0){const e=this._collector.outgoingEvents,t=this._collector.viewEvents;this._collector=null;for(const i of e)this._addOutgoingEvent(i);t.length>0&&this._emitMany(t)}this._emitOutgoingEvents()}emitSingleViewEvent(e){try{this.beginEmitViewEvents().emitViewEvent(e)}finally{this.endEmitViewEvents()}}_emitMany(e){this._viewEventQueue?this._viewEventQueue=this._viewEventQueue.concat(e):this._viewEventQueue=e,this._isConsumingViewEventQueue||this._consumeViewEventQueue()}_consumeViewEventQueue(){try{this._isConsumingViewEventQueue=!0,this._doConsumeQueue()}finally{this._isConsumingViewEventQueue=!1}}_doConsumeQueue(){for(;this._viewEventQueue;){const e=this._viewEventQueue;this._viewEventQueue=null;const t=this._eventHandlers.slice(0);for(const i of t)i.handleEvents(e)}}}class jye{constructor(){this.viewEvents=[],this.outgoingEvents=[]}emitViewEvent(e){this.viewEvents.push(e)}emitOutgoingEvent(e){this.outgoingEvents.push(e)}}class m7{constructor(e,t,i,n){this.kind=0,this._oldContentWidth=e,this._oldContentHeight=t,this.contentWidth=i,this.contentHeight=n,this.contentWidthChanged=this._oldContentWidth!==this.contentWidth,this.contentHeightChanged=this._oldContentHeight!==this.contentHeight}isNoOp(){return!this.contentWidthChanged&&!this.contentHeightChanged}attemptToMerge(e){return e.kind!==this.kind?null:new m7(this._oldContentWidth,this._oldContentHeight,e.contentWidth,e.contentHeight)}}class _A{constructor(e,t){this.kind=1,this.oldHasFocus=e,this.hasFocus=t}isNoOp(){return this.oldHasFocus===this.hasFocus}attemptToMerge(e){return e.kind!==this.kind?null:new _A(this.oldHasFocus,e.hasFocus)}}class Gye{constructor(e,t){this.kind=2,this.oldHasFocus=e,this.hasFocus=t}isNoOp(){return this.oldHasFocus===this.hasFocus}attemptToMerge(e){return e.kind!==this.kind?null:new _A(this.oldHasFocus,e.hasFocus)}}class _7{constructor(e,t,i,n,o,r,a,l){this.kind=3,this._oldScrollWidth=e,this._oldScrollLeft=t,this._oldScrollHeight=i,this._oldScrollTop=n,this.scrollWidth=o,this.scrollLeft=r,this.scrollHeight=a,this.scrollTop=l,this.scrollWidthChanged=this._oldScrollWidth!==this.scrollWidth,this.scrollLeftChanged=this._oldScrollLeft!==this.scrollLeft,this.scrollHeightChanged=this._oldScrollHeight!==this.scrollHeight,this.scrollTopChanged=this._oldScrollTop!==this.scrollTop}isNoOp(){return!this.scrollWidthChanged&&!this.scrollLeftChanged&&!this.scrollHeightChanged&&!this.scrollTopChanged}attemptToMerge(e){return e.kind!==this.kind?null:new _7(this._oldScrollWidth,this._oldScrollLeft,this._oldScrollHeight,this._oldScrollTop,e.scrollWidth,e.scrollLeft,e.scrollHeight,e.scrollTop)}}class Kye{constructor(){this.kind=4}isNoOp(){return!1}attemptToMerge(e){return e.kind!==this.kind?null:this}}class Yye{constructor(){this.kind=5}isNoOp(){return!1}attemptToMerge(e){return e.kind!==this.kind?null:this}}class fI{constructor(e,t,i,n,o,r,a){this.kind=7,this.oldSelections=e,this.selections=t,this.oldModelVersionId=i,this.modelVersionId=n,this.source=o,this.reason=r,this.reachedMaxCursorCount=a}static _selectionsAreEqual(e,t){if(!e&&!t)return!0;if(!e||!t)return!1;const i=e.length,n=t.length;if(i!==n)return!1;for(let o=0;o<i;o++)if(!e[o].equalsSelection(t[o]))return!1;return!0}isNoOp(){return fI._selectionsAreEqual(this.oldSelections,this.selections)&&this.oldModelVersionId===this.modelVersionId}attemptToMerge(e){return e.kind!==this.kind?null:new fI(this.oldSelections,e.selections,this.oldModelVersionId,e.modelVersionId,e.source,e.reason,this.reachedMaxCursorCount||e.reachedMaxCursorCount)}}class Zye{constructor(){this.kind=6}isNoOp(){return!1}attemptToMerge(e){return e.kind!==this.kind?null:this}}class Xye{constructor(e){this.event=e,this.kind=8}isNoOp(){return!1}attemptToMerge(e){return null}}class Qye{constructor(e){this.event=e,this.kind=9}isNoOp(){return!1}attemptToMerge(e){return null}}class Jye{constructor(e){this.event=e,this.kind=10}isNoOp(){return!1}attemptToMerge(e){return null}}class eSe{constructor(e){this.event=e,this.kind=11}isNoOp(){return!1}attemptToMerge(e){return null}}class tSe{constructor(e){this.event=e,this.kind=12}isNoOp(){return!1}attemptToMerge(e){return null}}class iSe{constructor(e){this.event=e,this.kind=13}isNoOp(){return!1}attemptToMerge(e){return null}}class nSe{constructor(e){this.event=e,this.kind=14}isNoOp(){return!1}attemptToMerge(e){return null}}class sSe{constructor(e){this.event=e,this.kind=15}isNoOp(){return!1}attemptToMerge(e){return null}}class oSe extends F{constructor(e,t,i,n){super(),this._model=e,this._knownModelVersionId=this._model.getVersionId(),this._viewModel=t,this._coordinatesConverter=i,this.context=new GU(this._model,this._viewModel,this._coordinatesConverter,n),this._cursors=new jU(this.context),this._hasFocus=!1,this._isHandling=!1,this._compositionState=null,this._columnSelectData=null,this._autoClosedActions=[],this._prevEditOperationType=0}dispose(){this._cursors.dispose(),this._autoClosedActions=xt(this._autoClosedActions),super.dispose()}updateConfiguration(e){this.context=new GU(this._model,this._viewModel,this._coordinatesConverter,e),this._cursors.updateContext(this.context)}onLineMappingChanged(e){this._knownModelVersionId===this._model.getVersionId()&&this.setStates(e,"viewModel",0,this.getCursorStates())}setHasFocus(e){this._hasFocus=e}_validateAutoClosedActions(){if(this._autoClosedActions.length>0){const e=this._cursors.getSelections();for(let t=0;t<this._autoClosedActions.length;t++){const i=this._autoClosedActions[t];i.isValid(e)||(i.dispose(),this._autoClosedActions.splice(t,1),t--)}}}getPrimaryCursorState(){return this._cursors.getPrimaryCursor()}getLastAddedCursorIndex(){return this._cursors.getLastAddedCursorIndex()}getCursorStates(){return this._cursors.getAll()}setStates(e,t,i,n){let o=!1;const r=this.context.cursorConfig.multiCursorLimit;n!==null&&n.length>r&&(n=n.slice(0,r),o=!0);const a=$1.from(this._model,this);return this._cursors.setStates(n),this._cursors.normalize(),this._columnSelectData=null,this._validateAutoClosedActions(),this._emitStateChangedIfNecessary(e,t,i,a,o)}setCursorColumnSelectData(e){this._columnSelectData=e}revealAll(e,t,i,n,o,r){const a=this._cursors.getViewPositions();let l=null,c=null;a.length>1?c=this._cursors.getViewSelections():l=k.fromPositions(a[0],a[0]),e.emitViewEvent(new z1(t,i,l,c,n,o,r))}revealPrimary(e,t,i,n,o,r){const l=[this._cursors.getPrimaryCursor().viewState.selection];e.emitViewEvent(new z1(t,i,null,l,n,o,r))}saveState(){const e=[],t=this._cursors.getSelections();for(let i=0,n=t.length;i<n;i++){const o=t[i];e.push({inSelectionMode:!o.isEmpty(),selectionStart:{lineNumber:o.selectionStartLineNumber,column:o.selectionStartColumn},position:{lineNumber:o.positionLineNumber,column:o.positionColumn}})}return e}restoreState(e,t){const i=[];for(let n=0,o=t.length;n<o;n++){const r=t[n];let a=1,l=1;r.position&&r.position.lineNumber&&(a=r.position.lineNumber),r.position&&r.position.column&&(l=r.position.column);let c=a,d=l;r.selectionStart&&r.selectionStart.lineNumber&&(c=r.selectionStart.lineNumber),r.selectionStart&&r.selectionStart.column&&(d=r.selectionStart.column),i.push({selectionStartLineNumber:c,selectionStartColumn:d,positionLineNumber:a,positionColumn:l})}this.setStates(e,"restoreState",0,ut.fromModelSelections(i)),this.revealAll(e,"restoreState",!1,0,!0,1)}onModelContentChanged(e,t){if(t instanceof Wee){if(this._isHandling)return;this._isHandling=!0;try{this.setStates(e,"modelChange",0,this.getCursorStates())}finally{this._isHandling=!1}}else{const i=t.rawContentChangedEvent;if(this._knownModelVersionId=i.versionId,this._isHandling)return;const n=i.containsEvent(1);if(this._prevEditOperationType=0,n)this._cursors.dispose(),this._cursors=new jU(this.context),this._validateAutoClosedActions(),this._emitStateChangedIfNecessary(e,"model",1,null,!1);else if(this._hasFocus&&i.resultingSelection&&i.resultingSelection.length>0){const o=ut.fromModelSelections(i.resultingSelection);this.setStates(e,"modelChange",i.isUndoing?5:i.isRedoing?6:2,o)&&this.revealAll(e,"modelChange",!1,0,!0,0)}else{const o=this._cursors.readSelectionFromMarkers();this.setStates(e,"modelChange",2,ut.fromModelSelections(o))}}}getSelection(){return this._cursors.getPrimaryCursor().modelState.selection}getTopMostViewPosition(){return this._cursors.getTopMostViewPosition()}getBottomMostViewPosition(){return this._cursors.getBottomMostViewPosition()}getCursorColumnSelectData(){if(this._columnSelectData)return this._columnSelectData;const e=this._cursors.getPrimaryCursor(),t=e.viewState.selectionStart.getStartPosition(),i=e.viewState.position;return{isReal:!1,fromViewLineNumber:t.lineNumber,fromViewVisualColumn:this.context.cursorConfig.visibleColumnFromColumn(this._viewModel,t),toViewLineNumber:i.lineNumber,toViewVisualColumn:this.context.cursorConfig.visibleColumnFromColumn(this._viewModel,i)}}getSelections(){return this._cursors.getSelections()}setSelections(e,t,i,n){this.setStates(e,t,n,ut.fromModelSelections(i))}getPrevEditOperationType(){return this._prevEditOperationType}setPrevEditOperationType(e){this._prevEditOperationType=e}_pushAutoClosedAction(e,t){const i=[],n=[];for(let a=0,l=e.length;a<l;a++)i.push({range:e[a],options:{description:"auto-closed-character",inlineClassName:"auto-closed-character",stickiness:1}}),n.push({range:t[a],options:{description:"auto-closed-enclosing",stickiness:1}});const o=this._model.deltaDecorations([],i),r=this._model.deltaDecorations([],n);this._autoClosedActions.push(new KU(this._model,o,r))}_executeEditOperation(e,t){if(!e)return;e.shouldPushStackElementBefore&&this._model.pushStackElement();const i=rSe.executeCommands(this._model,this._cursors.getSelections(),e.commands,t);if(i){this._interpretCommandResult(i);const n=[],o=[];for(let r=0;r<e.commands.length;r++){const a=e.commands[r];a instanceof xW&&a.enclosingRange&&a.closeCharacterRange&&(n.push(a.closeCharacterRange),o.push(a.enclosingRange))}n.length>0&&this._pushAutoClosedAction(n,o),this._prevEditOperationType=e.type}e.shouldPushStackElementAfter&&this._model.pushStackElement()}_interpretCommandResult(e){(!e||e.length===0)&&(e=this._cursors.readSelectionFromMarkers()),this._columnSelectData=null,this._cursors.setSelections(e),this._cursors.normalize()}_emitStateChangedIfNecessary(e,t,i,n,o){const r=$1.from(this._model,this);if(r.equals(n))return!1;const a=this._cursors.getSelections(),l=this._cursors.getViewSelections();if(e.emitViewEvent(new Fye(l,a,i)),!n||n.cursorState.length!==r.cursorState.length||r.cursorState.some((c,d)=>!c.modelState.equals(n.cursorState[d].modelState))){const c=n?n.cursorState.map(h=>h.modelState.selection):null,d=n?n.modelVersionId:0;e.emitOutgoingEvent(new fI(c,a,d,r.modelVersionId,t||"keyboard",i,o))}return!0}_findAutoClosingPairs(e){if(!e.length)return null;const t=[];for(let i=0,n=e.length;i<n;i++){const o=e[i];if(!o.text||o.text.indexOf(`
`)>=0)return null;const r=o.text.match(/([)\]}>'"`])([^)\]}>'"`]*)$/);if(!r)return null;const a=r[1],l=this.context.cursorConfig.autoClosingPairs.autoClosingPairsCloseSingleChar.get(a);if(!l||l.length!==1)return null;const c=l[0].open,d=o.text.length-r[2].length-1,h=o.text.lastIndexOf(c,d-1);if(h===-1)return null;t.push([h,d])}return t}executeEdits(e,t,i,n,o){let r=null;t==="snippet"&&(r=this._findAutoClosingPairs(i)),r&&(i[0]._isTracked=!0);const a=[],l=[],c=this._model.pushEditOperations(this.getSelections(),i,d=>{if(r)for(let u=0,g=r.length;u<g;u++){const[f,p]=r[u],_=d[u],b=_.range.startLineNumber,w=_.range.startColumn-1+f,C=_.range.startColumn-1+p;a.push(new k(b,C+1,b,C+2)),l.push(new k(b,w+1,b,C+2))}const h=n(d);return h&&(this._isHandling=!0),h},void 0,o);c&&(this._isHandling=!1,this.setSelections(e,t,c,0)),a.length>0&&this._pushAutoClosedAction(a,l)}_executeEdit(e,t,i,n=0){if(this.context.cursorConfig.readOnly)return;const o=$1.from(this._model,this);this._cursors.stopTrackingSelections(),this._isHandling=!0;try{this._cursors.ensureValidState(),e()}catch(r){Oe(r)}this._isHandling=!1,this._cursors.startTrackingSelections(),this._validateAutoClosedActions(),this._emitStateChangedIfNecessary(t,i,n,o,!1)&&this.revealAll(t,i,!1,0,!0,0)}getAutoClosedCharacters(){return KU.getAllAutoClosedCharacters(this._autoClosedActions)}startComposition(e){this._compositionState=new U1(this._model,this.getSelections())}endComposition(e,t){const i=cs.cursor({kind:"compositionEnd",detailedSource:t}),n=this._compositionState?this._compositionState.deduceOutcome(this._model,this.getSelections()):null;this._compositionState=null,this._executeEdit(()=>{t==="keyboard"&&this._executeEditOperation(ou.compositionEndWithInterceptors(this._prevEditOperationType,this.context.cursorConfig,this._model,n,this.getSelections(),this.getAutoClosedCharacters()),i)},e,t)}type(e,t,i){const n=cs.cursor({kind:"type",detailedSource:i});this._executeEdit(()=>{if(i==="keyboard"){const o=t.length;let r=0;for(;r<o;){const a=Y6(t,r),l=t.substr(r,a);this._executeEditOperation(ou.typeWithInterceptors(!!this._compositionState,this._prevEditOperationType,this.context.cursorConfig,this._model,this.getSelections(),this.getAutoClosedCharacters(),l),n),r+=a}}else this._executeEditOperation(ou.typeWithoutInterceptors(this._prevEditOperationType,this.context.cursorConfig,this._model,this.getSelections(),t),n)},e,i)}compositionType(e,t,i,n,o,r){const a=cs.cursor({kind:"compositionType",detailedSource:r});if(t.length===0&&i===0&&n===0){if(o!==0){const l=this.getSelections().map(c=>{const d=c.getPosition();return new he(d.lineNumber,d.column+o,d.lineNumber,d.column+o)});this.setSelections(e,r,l,0)}return}this._executeEdit(()=>{this._executeEditOperation(ou.compositionType(this._prevEditOperationType,this.context.cursorConfig,this._model,this.getSelections(),t,i,n,o),a)},e,r)}paste(e,t,i,n,o){const r=cs.cursor({kind:"paste",detailedSource:o});this._executeEdit(()=>{this._executeEditOperation(ou.paste(this.context.cursorConfig,this._model,this.getSelections(),t,i,n||[]),r)},e,o,4)}cut(e,t){const i=cs.cursor({kind:"cut",detailedSource:t});this._executeEdit(()=>{this._executeEditOperation(km.cut(this.context.cursorConfig,this._model,this.getSelections()),i)},e,t)}executeCommand(e,t,i){const n=cs.cursor({kind:"executeCommand",detailedSource:i});this._executeEdit(()=>{this._cursors.killSecondaryCursors(),this._executeEditOperation(new to(0,[t],{shouldPushStackElementBefore:!1,shouldPushStackElementAfter:!1}),n)},e,i)}executeCommands(e,t,i){const n=cs.cursor({kind:"executeCommands",detailedSource:i});this._executeEdit(()=>{this._executeEditOperation(new to(0,t,{shouldPushStackElementBefore:!1,shouldPushStackElementAfter:!1}),n)},e,i)}}class $1{static from(e,t){return new $1(e.getVersionId(),t.getCursorStates())}constructor(e,t){this.modelVersionId=e,this.cursorState=t}equals(e){if(!e||this.modelVersionId!==e.modelVersionId||this.cursorState.length!==e.cursorState.length)return!1;for(let t=0,i=this.cursorState.length;t<i;t++)if(!this.cursorState[t].equals(e.cursorState[t]))return!1;return!0}}class KU{static getAllAutoClosedCharacters(e){let t=[];for(const i of e)t=t.concat(i.getAutoClosedCharactersRanges());return t}constructor(e,t,i){this._model=e,this._autoClosedCharactersDecorations=t,this._autoClosedEnclosingDecorations=i}dispose(){this._autoClosedCharactersDecorations=this._model.deltaDecorations(this._autoClosedCharactersDecorations,[]),this._autoClosedEnclosingDecorations=this._model.deltaDecorations(this._autoClosedEnclosingDecorations,[])}getAutoClosedCharactersRanges(){const e=[];for(let t=0;t<this._autoClosedCharactersDecorations.length;t++){const i=this._model.getDecorationRange(this._autoClosedCharactersDecorations[t]);i&&e.push(i)}return e}isValid(e){const t=[];for(let i=0;i<this._autoClosedEnclosingDecorations.length;i++){const n=this._model.getDecorationRange(this._autoClosedEnclosingDecorations[i]);if(n&&(t.push(n),n.startLineNumber!==n.endLineNumber))return!1}t.sort(k.compareRangesUsingStarts),e.sort(k.compareRangesUsingStarts);for(let i=0;i<e.length;i++)if(i>=t.length||!t[i].strictContainsRange(e[i]))return!1;return!0}}class rSe{static executeCommands(e,t,i,n=cs.unknown({name:"executeCommands"})){const o={model:e,selectionsBefore:t,trackedRanges:[],trackedRangesDirection:[]},r=this._innerExecuteCommands(o,i,n);for(let a=0,l=o.trackedRanges.length;a<l;a++)o.model._setTrackedRange(o.trackedRanges[a],null,0);return r}static _innerExecuteCommands(e,t,i){if(this._arrayIsEmpty(t))return null;const n=this._getEditOperations(e,t);if(n.operations.length===0)return null;const o=n.operations,r=this._getLoserCursorMap(o);if(r.hasOwnProperty("0"))return console.warn("Ignoring commands"),null;const a=[];for(let d=0,h=o.length;d<h;d++)r.hasOwnProperty(o[d].identifier.major.toString())||a.push(o[d]);n.hadTrackedEditOperation&&a.length>0&&(a[0]._isTracked=!0);let l=e.model.pushEditOperations(e.selectionsBefore,a,d=>{const h=[];for(let f=0;f<e.selectionsBefore.length;f++)h[f]=[];for(const f of d)f.identifier&&h[f.identifier.major].push(f);const u=(f,p)=>f.identifier.minor-p.identifier.minor,g=[];for(let f=0;f<e.selectionsBefore.length;f++)h[f].length>0?(h[f].sort(u),g[f]=t[f].computeCursorState(e.model,{getInverseEditOperations:()=>h[f],getTrackedSelection:p=>{const _=parseInt(p,10),b=e.model._getTrackedRange(e.trackedRanges[_]);return e.trackedRangesDirection[_]===0?new he(b.startLineNumber,b.startColumn,b.endLineNumber,b.endColumn):new he(b.endLineNumber,b.endColumn,b.startLineNumber,b.startColumn)}})):g[f]=e.selectionsBefore[f];return g},void 0,i);l||(l=e.selectionsBefore);const c=[];for(const d in r)r.hasOwnProperty(d)&&c.push(parseInt(d,10));c.sort((d,h)=>h-d);for(const d of c)l.splice(d,1);return l}static _arrayIsEmpty(e){for(let t=0,i=e.length;t<i;t++)if(e[t])return!1;return!0}static _getEditOperations(e,t){let i=[],n=!1;for(let o=0,r=t.length;o<r;o++){const a=t[o];if(a){const l=this._getEditOperationsFromCommand(e,o,a);i=i.concat(l.operations),n=n||l.hadTrackedEditOperation}}return{operations:i,hadTrackedEditOperation:n}}static _getEditOperationsFromCommand(e,t,i){const n=[];let o=0;const r=(h,u,g=!1)=>{k.isEmpty(h)&&u===""||n.push({identifier:{major:t,minor:o++},range:h,text:u,forceMoveMarkers:g,isAutoWhitespaceEdit:i.insertsAutoWhitespace})};let a=!1;const d={addEditOperation:r,addTrackedEditOperation:(h,u,g)=>{a=!0,r(h,u,g)},trackSelection:(h,u)=>{const g=he.liftSelection(h);let f;if(g.isEmpty())if(typeof u=="boolean")u?f=2:f=3;else{const b=e.model.getLineMaxColumn(g.startLineNumber);g.startColumn===b?f=2:f=3}else f=1;const p=e.trackedRanges.length,_=e.model._setTrackedRange(null,g,f);return e.trackedRanges[p]=_,e.trackedRangesDirection[p]=g.getDirection(),p.toString()}};try{i.getEditOperations(e.model,d)}catch(h){return Oe(h),{operations:[],hadTrackedEditOperation:!1}}return{operations:n,hadTrackedEditOperation:a}}static _getLoserCursorMap(e){e=e.slice(0),e.sort((i,n)=>-k.compareRangesUsingEnds(i.range,n.range));const t={};for(let i=1;i<e.length;i++){const n=e[i-1],o=e[i];if(k.getStartPosition(n.range).isBefore(k.getEndPosition(o.range))){let r;n.identifier.major>o.identifier.major?r=n.identifier.major:r=o.identifier.major,t[r.toString()]=!0;for(let a=0;a<e.length;a++)e[a].identifier.major===r&&(e.splice(a,1),a<i&&i--,a--);i>0&&i--}}return t}}class aSe{constructor(e,t,i,n){this.text=e,this.lineNumber=t,this.startSelectionOffset=i,this.endSelectionOffset=n}}class U1{static _capture(e,t){const i=[];for(const n of t){if(n.startLineNumber!==n.endLineNumber)return null;const o=n.startLineNumber;i.push(new aSe(e.getLineContent(o),o,n.startColumn-1,n.endColumn-1))}return i}constructor(e,t){this._original=U1._capture(e,t)}deduceOutcome(e,t){if(!this._original)return null;const i=U1._capture(e,t);if(!i||this._original.length!==i.length)return null;const n=[];for(let o=0,r=this._original.length;o<r;o++)n.push(U1._deduceOutcome(this._original[o],i[o]));return n}static _deduceOutcome(e,t){const i=Math.min(e.startSelectionOffset,t.startSelectionOffset,Fa(e.text,t.text)),n=Math.min(e.text.length-e.endSelectionOffset,t.text.length-t.endSelectionOffset,uh(e.text,t.text)),o=e.text.substring(i,e.text.length-n),r=i,a=t.text.length-n,l=t.text.substring(r,a),c=new k(t.lineNumber,r+1,t.lineNumber,a+1);return new Dfe(o,e.startSelectionOffset-i,e.endSelectionOffset-i,l,t.startSelectionOffset-i,t.endSelectionOffset-i,c)}}const YU={getInitialState:()=>bv,tokenizeEncoded:(s,e,t)=>mA(0,t)};async function lSe(s,e,t){if(!t)return ZU(e,s.languageIdCodec,YU);const i=await yi.getOrCreate(t);return ZU(e,s.languageIdCodec,i||YU)}function cSe(s,e,t,i,n,o,r){let a="<div>",l=0,c=0,d=!0;for(let h=0,u=e.getCount();h<u;h++){const g=e.getEndOffset(h);let f="";for(;l<g&&l<n;l++){const p=s.charCodeAt(l),_=p===9;if(c+=gh(p)?2:_?0:1,l<i){if(_){const b=c%o;c+=b===0?o:o-b}continue}switch(p){case 9:{const b=c%o,w=b===0?o:o-b;c+=w;let C=w;for(;C>0;)r&&d?(f+="&#160;",d=!1):(f+=" ",d=!0),C--;break}case 60:f+="&lt;",d=!1;break;case 62:f+="&gt;",d=!1;break;case 38:f+="&amp;",d=!1;break;case 0:f+="&#00;",d=!1;break;case 65279:case 8232:case 8233:case 133:f+="�",d=!1;break;case 13:f+="&#8203",d=!1;break;case 32:r&&d?(f+="&#160;",d=!1):(f+=" ",d=!0);break;default:f+=String.fromCharCode(p),d=!1}}if(!(g<=i)&&(a+=`<span style="${e.getInlineStyle(h,t)}">${f}</span>`,g>n||l>=n||i>=n))break}return a+="</div>",a}function ZU(s,e,t){let i='<div class="monaco-tokenized-source">';const n=qo(s);let o=t.getInitialState();for(let r=0,a=n.length;r<a;r++){const l=n[r];r>0&&(i+="<br/>");const c=t.tokenizeEncoded(l,!0,o);Ri.convertToEndOffset(c.tokens,l.length);const h=new Ri(c.tokens,l,e).inflate();let u=0;for(let g=0,f=h.getCount();g<f;g++){const p=h.getClassName(g),_=h.getEndOffset(g);i+=`<span class="${p}">${Aa(l.substring(u,_))}</span>`,u=_}o=c.endState}return i+="</div>",i}class dSe{constructor(e,t,i,n,o){this.decorationId=e,this.index=t,this.lineNumber=i,this.specialHeight=n,this.prefixSum=o,this.maximumSpecialHeight=n,this.deleted=!1}}class XU{constructor(e,t){if(this._decorationIDToCustomLine=new bM,this._orderedCustomLines=[],this._pendingSpecialLinesToInsert=[],this._invalidIndex=0,this._hasPending=!1,this._defaultLineHeight=e,t.length>0){for(const i of t)this.insertOrChangeCustomLineHeight(i.decorationId,i.startLineNumber,i.endLineNumber,i.lineHeight);this.commit()}}set defaultLineHeight(e){this._defaultLineHeight=e}get defaultLineHeight(){return this._defaultLineHeight}removeCustomLineHeight(e){const t=this._decorationIDToCustomLine.get(e);if(t){this._decorationIDToCustomLine.delete(e);for(const i of t)i.deleted=!0,this._invalidIndex=Math.min(this._invalidIndex,i.index);this._hasPending=!0}}insertOrChangeCustomLineHeight(e,t,i,n){this.removeCustomLineHeight(e);for(let o=t;o<=i;o++){const r=new dSe(e,-1,o,n,0);this._pendingSpecialLinesToInsert.push(r)}this._hasPending=!0}heightForLineNumber(e){const t=this._binarySearchOverOrderedCustomLinesArray(e);return t>=0?this._orderedCustomLines[t].maximumSpecialHeight:this._defaultLineHeight}getAccumulatedLineHeightsIncludingLineNumber(e){const t=this._binarySearchOverOrderedCustomLinesArray(e);if(t>=0)return this._orderedCustomLines[t].prefixSum+this._orderedCustomLines[t].maximumSpecialHeight;if(t===-1)return this._defaultLineHeight*e;const i=-(t+1),n=this._orderedCustomLines[i-1];return n.prefixSum+n.maximumSpecialHeight+this._defaultLineHeight*(e-n.lineNumber)}onLinesDeleted(e,t){const i=t-e+1,n=this._orderedCustomLines.length,o=this._binarySearchOverOrderedCustomLinesArray(e);let r;if(o>=0){r=o;for(let h=o-1;h>=0&&this._orderedCustomLines[h].lineNumber===e;h--)r--}else r=o===-(n+1)&&o!==-1?n-1:-(o+1);const a=this._binarySearchOverOrderedCustomLinesArray(t);let l;if(a>=0){l=a;for(let h=a+1;h<n&&this._orderedCustomLines[h].lineNumber===t;h++)l++}else l=a===-(n+1)&&a!==-1?n-1:-(a+1);const c=l>r,d=l===r&&this._orderedCustomLines[r]&&this._orderedCustomLines[r].lineNumber>=e&&this._orderedCustomLines[r].lineNumber<=t;if(c||d){let h=0;for(let x=r;x<=l;x++)h=Math.max(h,this._orderedCustomLines[x].maximumSpecialHeight);let u=0;if(r>0){const x=this._orderedCustomLines[r-1];u=x.prefixSum+x.maximumSpecialHeight+this._defaultLineHeight*(e-x.lineNumber-1)}else u=e>0?(e-1)*this._defaultLineHeight:0;const g=this._orderedCustomLines[r],f=this._orderedCustomLines[l],p=this._orderedCustomLines[l+1],_=p&&p.lineNumber===t+1?p.maximumSpecialHeight:this._defaultLineHeight,b=f.prefixSum+f.maximumSpecialHeight-g.prefixSum+this._defaultLineHeight*(t-f.lineNumber)+this._defaultLineHeight*(g.lineNumber-e)+_-h,w=new Set,C=[],v=new bM;let y=0;for(let x=0;x<this._orderedCustomLines.length;x++){const L=this._orderedCustomLines[x];if(x<r)C.push(L),v.add(L.decorationId,L);else if(x>=r&&x<=l){const E=L.decorationId;w.has(E)?y++:(L.index-=y,L.lineNumber=e,L.prefixSum=u,L.maximumSpecialHeight=h,C.push(L),v.add(L.decorationId,L))}else x>l&&(L.index-=y,L.lineNumber-=i,L.prefixSum-=b,C.push(L),v.add(L.decorationId,L));w.add(L.decorationId)}this._orderedCustomLines=C,this._decorationIDToCustomLine=v}else{const h=i*this._defaultLineHeight;for(let u=l;u<this._orderedCustomLines.length;u++){const g=this._orderedCustomLines[u];g.lineNumber>t&&(g.lineNumber-=i,g.prefixSum-=h)}}}onLinesInserted(e,t){const i=t-e+1,n=this._binarySearchOverOrderedCustomLinesArray(e);let o;if(n>=0){o=n;for(let d=n-1;d>=0&&this._orderedCustomLines[d].lineNumber===e;d--)o--}else o=-(n+1);const r=[],a=new Set;for(let d=o;d<this._orderedCustomLines.length;d++)this._orderedCustomLines[d].lineNumber===e&&a.add(this._orderedCustomLines[d].decorationId);const l=new Set;for(let d=o-1;d>=0;d--)this._orderedCustomLines[d].lineNumber===e-1&&l.add(this._orderedCustomLines[d].decorationId);const c=gte(l,a);for(let d=o;d<this._orderedCustomLines.length;d++)this._orderedCustomLines[d].lineNumber+=i,this._orderedCustomLines[d].prefixSum+=this._defaultLineHeight*i;if(c.size>0){for(const d of c){const h=this._decorationIDToCustomLine.get(d);if(h){const u=h.reduce((p,_)=>Math.min(p,_.lineNumber),e),g=h.reduce((p,_)=>Math.max(p,_.lineNumber),e),f=h.reduce((p,_)=>Math.max(p,_.specialHeight),0);r.push({decorationId:d,startLineNumber:u,endLineNumber:g,lineHeight:f})}}for(const d of r)this.insertOrChangeCustomLineHeight(d.decorationId,d.startLineNumber,d.endLineNumber,d.lineHeight);this.commit()}}commit(){if(!this._hasPending)return;for(const o of this._pendingSpecialLinesToInsert){const r=this._binarySearchOverOrderedCustomLinesArray(o.lineNumber),a=r>=0?r:-(r+1);this._orderedCustomLines.splice(a,0,o),this._invalidIndex=Math.min(this._invalidIndex,a)}this._pendingSpecialLinesToInsert=[];const e=new bM,t=[];for(let o=0;o<this._invalidIndex;o++){const r=this._orderedCustomLines[o];t.push(r),e.add(r.decorationId,r)}let i=0,n=this._invalidIndex>0?t[this._invalidIndex-1]:void 0;for(let o=this._invalidIndex;o<this._orderedCustomLines.length;o++){const r=this._orderedCustomLines[o];if(r.deleted){i++;continue}if(r.index=o-i,n&&n.lineNumber===r.lineNumber)r.maximumSpecialHeight=n.maximumSpecialHeight,r.prefixSum=n.prefixSum;else{let a=r.specialHeight;for(let c=o;c<this._orderedCustomLines.length;c++){const d=this._orderedCustomLines[c];if(!d.deleted){if(d.lineNumber!==r.lineNumber)break;a=Math.max(a,d.specialHeight)}}r.maximumSpecialHeight=a;let l;n?l=n.prefixSum+n.maximumSpecialHeight+this._defaultLineHeight*(r.lineNumber-n.lineNumber-1):l=this._defaultLineHeight*(r.lineNumber-1),r.prefixSum=l}n=r,t.push(r),e.add(r.decorationId,r)}this._orderedCustomLines=t,this._decorationIDToCustomLine=e,this._invalidIndex=1/0,this._hasPending=!1}_binarySearchOverOrderedCustomLinesArray(e){return J6(this._orderedCustomLines.length,t=>{const i=this._orderedCustomLines[t];return i.lineNumber===e?0:i.lineNumber<e?-1:1})}}class bM{constructor(){this._map=new Map}add(e,t){const i=this._map.get(e);i?i.push(t):this._map.set(e,[t])}get(e){return this._map.get(e)}delete(e){this._map.delete(e)}}class hSe{constructor(){this._hasPending=!1,this._inserts=[],this._changes=[],this._removes=[]}insert(e){this._hasPending=!0,this._inserts.push(e)}change(e){this._hasPending=!0,this._changes.push(e)}remove(e){this._hasPending=!0,this._removes.push(e)}commit(e){if(!this._hasPending)return;const t=this._inserts,i=this._changes,n=this._removes;this._hasPending=!1,this._inserts=[],this._changes=[],this._removes=[],e._commitPendingChanges(t,i,n)}}class uSe{constructor(e,t,i,n,o){this.id=e,this.afterLineNumber=t,this.ordinal=i,this.height=n,this.minWidth=o,this.prefixSum=0}}var zu;let gSe=(zu=class{constructor(e,t,i,n,o){this._instanceId=sQ(++zu.INSTANCE_COUNT),this._pendingChanges=new hSe,this._lastWhitespaceId=0,this._arr=[],this._prefixSumValidIndex=-1,this._minWidth=-1,this._lineCount=e,this._paddingTop=i,this._paddingBottom=n,this._lineHeightsManager=new XU(t,o)}static findInsertionIndex(e,t,i){let n=0,o=e.length;for(;n<o;){const r=n+o>>>1;t===e[r].afterLineNumber?i<e[r].ordinal?o=r:n=r+1:t<e[r].afterLineNumber?o=r:n=r+1}return n}setDefaultLineHeight(e){this._lineHeightsManager.defaultLineHeight=e}setPadding(e,t){this._paddingTop=e,this._paddingBottom=t}onFlushed(e,t){this._lineCount=e,this._lineHeightsManager=new XU(this._lineHeightsManager.defaultLineHeight,t)}changeLineHeights(e){let t=!1;try{e({insertOrChangeCustomLineHeight:(n,o,r,a)=>{t=!0,this._lineHeightsManager.insertOrChangeCustomLineHeight(n,o,r,a)},removeCustomLineHeight:n=>{t=!0,this._lineHeightsManager.removeCustomLineHeight(n)}})}finally{this._lineHeightsManager.commit()}return t}changeWhitespace(e){let t=!1;try{e({insertWhitespace:(n,o,r,a)=>{t=!0,n=n|0,o=o|0,r=r|0,a=a|0;const l=this._instanceId+ ++this._lastWhitespaceId;return this._pendingChanges.insert(new uSe(l,n,o,r,a)),l},changeOneWhitespace:(n,o,r)=>{t=!0,o=o|0,r=r|0,this._pendingChanges.change({id:n,newAfterLineNumber:o,newHeight:r})},removeWhitespace:n=>{t=!0,this._pendingChanges.remove({id:n})}})}finally{this._pendingChanges.commit(this)}return t}_commitPendingChanges(e,t,i){if((e.length>0||i.length>0)&&(this._minWidth=-1),e.length+t.length+i.length<=1){for(const l of e)this._insertWhitespace(l);for(const l of t)this._changeOneWhitespace(l.id,l.newAfterLineNumber,l.newHeight);for(const l of i){const c=this._findWhitespaceIndex(l.id);c!==-1&&this._removeWhitespace(c)}return}const n=new Set;for(const l of i)n.add(l.id);const o=new Map;for(const l of t)o.set(l.id,l);const r=l=>{const c=[];for(const d of l)if(!n.has(d.id)){if(o.has(d.id)){const h=o.get(d.id);d.afterLineNumber=h.newAfterLineNumber,d.height=h.newHeight}c.push(d)}return c},a=r(this._arr).concat(r(e));a.sort((l,c)=>l.afterLineNumber===c.afterLineNumber?l.ordinal-c.ordinal:l.afterLineNumber-c.afterLineNumber),this._arr=a,this._prefixSumValidIndex=-1}_insertWhitespace(e){const t=zu.findInsertionIndex(this._arr,e.afterLineNumber,e.ordinal);this._arr.splice(t,0,e),this._prefixSumValidIndex=Math.min(this._prefixSumValidIndex,t-1)}_findWhitespaceIndex(e){const t=this._arr;for(let i=0,n=t.length;i<n;i++)if(t[i].id===e)return i;return-1}_changeOneWhitespace(e,t,i){const n=this._findWhitespaceIndex(e);if(n!==-1&&(this._arr[n].height!==i&&(this._arr[n].height=i,this._prefixSumValidIndex=Math.min(this._prefixSumValidIndex,n-1)),this._arr[n].afterLineNumber!==t)){const o=this._arr[n];this._removeWhitespace(n),o.afterLineNumber=t,this._insertWhitespace(o)}}_removeWhitespace(e){this._arr.splice(e,1),this._prefixSumValidIndex=Math.min(this._prefixSumValidIndex,e-1)}onLinesDeleted(e,t){e=e|0,t=t|0,this._lineCount-=t-e+1;for(let i=0,n=this._arr.length;i<n;i++){const o=this._arr[i].afterLineNumber;e<=o&&o<=t?this._arr[i].afterLineNumber=e-1:o>t&&(this._arr[i].afterLineNumber-=t-e+1)}this._lineHeightsManager.onLinesDeleted(e,t)}onLinesInserted(e,t){e=e|0,t=t|0,this._lineCount+=t-e+1;for(let i=0,n=this._arr.length;i<n;i++){const o=this._arr[i].afterLineNumber;e<=o&&(this._arr[i].afterLineNumber+=t-e+1)}this._lineHeightsManager.onLinesInserted(e,t)}getWhitespacesTotalHeight(){return this._arr.length===0?0:this.getWhitespacesAccumulatedHeight(this._arr.length-1)}getWhitespacesAccumulatedHeight(e){e=e|0;let t=Math.max(0,this._prefixSumValidIndex+1);t===0&&(this._arr[0].prefixSum=this._arr[0].height,t++);for(let i=t;i<=e;i++)this._arr[i].prefixSum=this._arr[i-1].prefixSum+this._arr[i].height;return this._prefixSumValidIndex=Math.max(this._prefixSumValidIndex,e),this._arr[e].prefixSum}getLinesTotalHeight(){const e=this._lineHeightsManager.getAccumulatedLineHeightsIncludingLineNumber(this._lineCount),t=this.getWhitespacesTotalHeight();return e+t+this._paddingTop+this._paddingBottom}getWhitespaceAccumulatedHeightBeforeLineNumber(e){e=e|0;const t=this._findLastWhitespaceBeforeLineNumber(e);return t===-1?0:this.getWhitespacesAccumulatedHeight(t)}_findLastWhitespaceBeforeLineNumber(e){e=e|0;const t=this._arr;let i=0,n=t.length-1;for(;i<=n;){const r=(n-i|0)/2|0,a=i+r|0;if(t[a].afterLineNumber<e){if(a+1>=t.length||t[a+1].afterLineNumber>=e)return a;i=a+1|0}else n=a-1|0}return-1}_findFirstWhitespaceAfterLineNumber(e){e=e|0;const i=this._findLastWhitespaceBeforeLineNumber(e)+1;return i<this._arr.length?i:-1}getFirstWhitespaceIndexAfterLineNumber(e){return e=e|0,this._findFirstWhitespaceAfterLineNumber(e)}getVerticalOffsetForLineNumber(e,t=!1){e=e|0;let i;e>1?i=this._lineHeightsManager.getAccumulatedLineHeightsIncludingLineNumber(e-1):i=0;const n=this.getWhitespaceAccumulatedHeightBeforeLineNumber(e-(t?1:0));return i+n+this._paddingTop}getLineHeightForLineNumber(e){return this._lineHeightsManager.heightForLineNumber(e)}getVerticalOffsetAfterLineNumber(e,t=!1){e=e|0;const i=this._lineHeightsManager.getAccumulatedLineHeightsIncludingLineNumber(e),n=this.getWhitespaceAccumulatedHeightBeforeLineNumber(e+(t?1:0));return i+n+this._paddingTop}getWhitespaceMinWidth(){if(this._minWidth===-1){let e=0;for(let t=0,i=this._arr.length;t<i;t++)e=Math.max(e,this._arr[t].minWidth);this._minWidth=e}return this._minWidth}isAfterLines(e){const t=this.getLinesTotalHeight();return e>t}isInTopPadding(e){return this._paddingTop===0?!1:e<this._paddingTop}isInBottomPadding(e){if(this._paddingBottom===0)return!1;const t=this.getLinesTotalHeight();return e>=t-this._paddingBottom}getLineNumberAtOrAfterVerticalOffset(e){if(e=e|0,e<0)return 1;const t=this._lineCount|0;let i=1,n=t;for(;i<n;){const o=(i+n)/2|0,r=this.getLineHeightForLineNumber(o),a=this.getVerticalOffsetForLineNumber(o)|0;if(e>=a+r)i=o+1;else{if(e>=a)return o;n=o}}return i>t?t:i}getLinesViewportData(e,t){e=e|0,t=t|0;const i=this.getLineNumberAtOrAfterVerticalOffset(e)|0,n=this.getVerticalOffsetForLineNumber(i)|0;let o=this._lineCount|0,r=this.getFirstWhitespaceIndexAfterLineNumber(i)|0;const a=this.getWhitespacesCount()|0;let l,c;r===-1?(r=a,c=o+1,l=0):(c=this.getAfterLineNumberForWhitespaceIndex(r)|0,l=this.getHeightForWhitespaceIndex(r)|0);let d=n,h=d;const u=5e5;let g=0;n>=u&&(g=Math.floor(n/u)*u,g=Math.floor(g/this._lineHeightsManager.defaultLineHeight)*this._lineHeightsManager.defaultLineHeight,h-=g);const f=[],p=e+(t-e)/2;let _=-1;for(let v=i;v<=o;v++){const y=this.getLineHeightForLineNumber(v);if(_===-1){const x=d,L=d+y;(x<=p&&p<L||x>p)&&(_=v)}for(d+=y,f[v-i]=h,h+=y;c===v;)h+=l,d+=l,r++,r>=a?c=o+1:(c=this.getAfterLineNumberForWhitespaceIndex(r)|0,l=this.getHeightForWhitespaceIndex(r)|0);if(d>=t){o=v;break}}_===-1&&(_=o);const b=this.getVerticalOffsetForLineNumber(o)|0;let w=i,C=o;if(w<C&&n<e&&w++,w<C){const v=this.getLineHeightForLineNumber(o);b+v>t&&C--}return{bigNumbersDelta:g,startLineNumber:i,endLineNumber:o,relativeVerticalOffset:f,centeredLineNumber:_,completelyVisibleStartLineNumber:w,completelyVisibleEndLineNumber:C,lineHeight:this._lineHeightsManager.defaultLineHeight}}getVerticalOffsetForWhitespaceIndex(e){e=e|0;const t=this.getAfterLineNumberForWhitespaceIndex(e);let i;t>=1?i=this._lineHeightsManager.getAccumulatedLineHeightsIncludingLineNumber(t):i=0;let n;return e>0?n=this.getWhitespacesAccumulatedHeight(e-1):n=0,i+n+this._paddingTop}getWhitespaceIndexAtOrAfterVerticallOffset(e){e=e|0;let t=0,i=this.getWhitespacesCount()-1;if(i<0)return-1;const n=this.getVerticalOffsetForWhitespaceIndex(i),o=this.getHeightForWhitespaceIndex(i);if(e>=n+o)return-1;for(;t<i;){const r=Math.floor((t+i)/2),a=this.getVerticalOffsetForWhitespaceIndex(r),l=this.getHeightForWhitespaceIndex(r);if(e>=a+l)t=r+1;else{if(e>=a)return r;i=r}}return t}getWhitespaceAtVerticalOffset(e){e=e|0;const t=this.getWhitespaceIndexAtOrAfterVerticallOffset(e);if(t<0||t>=this.getWhitespacesCount())return null;const i=this.getVerticalOffsetForWhitespaceIndex(t);if(i>e)return null;const n=this.getHeightForWhitespaceIndex(t),o=this.getIdForWhitespaceIndex(t),r=this.getAfterLineNumberForWhitespaceIndex(t);return{id:o,afterLineNumber:r,verticalOffset:i,height:n}}getWhitespaceViewportData(e,t){e=e|0,t=t|0;const i=this.getWhitespaceIndexAtOrAfterVerticallOffset(e),n=this.getWhitespacesCount()-1;if(i<0)return[];const o=[];for(let r=i;r<=n;r++){const a=this.getVerticalOffsetForWhitespaceIndex(r),l=this.getHeightForWhitespaceIndex(r);if(a>=t)break;o.push({id:this.getIdForWhitespaceIndex(r),afterLineNumber:this.getAfterLineNumberForWhitespaceIndex(r),verticalOffset:a,height:l})}return o}getWhitespaces(){return this._arr.slice(0)}getWhitespacesCount(){return this._arr.length}getIdForWhitespaceIndex(e){return e=e|0,this._arr[e].id}getAfterLineNumberForWhitespaceIndex(e){return e=e|0,this._arr[e].afterLineNumber}getHeightForWhitespaceIndex(e){return e=e|0,this._arr[e].height}},zu.INSTANCE_COUNT=0,zu);const fSe=125;class d1{constructor(e,t,i,n){e=e|0,t=t|0,i=i|0,n=n|0,e<0&&(e=0),t<0&&(t=0),i<0&&(i=0),n<0&&(n=0),this.width=e,this.contentWidth=t,this.scrollWidth=Math.max(e,t),this.height=i,this.contentHeight=n,this.scrollHeight=Math.max(i,n)}equals(e){return this.width===e.width&&this.contentWidth===e.contentWidth&&this.height===e.height&&this.contentHeight===e.contentHeight}}class pSe extends F{constructor(e,t){super(),this._onDidContentSizeChange=this._register(new M),this.onDidContentSizeChange=this._onDidContentSizeChange.event,this._dimensions=new d1(0,0,0,0),this._scrollable=this._register(new sC({forceIntegerValues:!0,smoothScrollDuration:e,scheduleAtNextAnimationFrame:t})),this.onDidScroll=this._scrollable.onScroll}getScrollable(){return this._scrollable}setSmoothScrollDuration(e){this._scrollable.setSmoothScrollDuration(e)}validateScrollPosition(e){return this._scrollable.validateScrollPosition(e)}getScrollDimensions(){return this._dimensions}setScrollDimensions(e){if(this._dimensions.equals(e))return;const t=this._dimensions;this._dimensions=e,this._scrollable.setScrollDimensions({width:e.width,scrollWidth:e.scrollWidth,height:e.height,scrollHeight:e.scrollHeight},!0);const i=t.contentWidth!==e.contentWidth,n=t.contentHeight!==e.contentHeight;(i||n)&&this._onDidContentSizeChange.fire(new m7(t.contentWidth,t.contentHeight,e.contentWidth,e.contentHeight))}getFutureScrollPosition(){return this._scrollable.getFutureScrollPosition()}getCurrentScrollPosition(){return this._scrollable.getCurrentScrollPosition()}setScrollPositionNow(e){this._scrollable.setScrollPositionNow(e)}setScrollPositionSmooth(e){this._scrollable.setScrollPositionSmooth(e)}hasPendingScrollAnimation(){return this._scrollable.hasPendingScrollAnimation()}}class mSe extends F{constructor(e,t,i,n){super(),this._configuration=e;const o=this._configuration.options,r=o.get(165),a=o.get(96);this._linesLayout=new gSe(t,o.get(75),a.top,a.bottom,i),this._maxLineWidth=0,this._overlayWidgetsMinWidth=0,this._scrollable=this._register(new pSe(0,n)),this._configureSmoothScrollDuration(),this._scrollable.setScrollDimensions(new d1(r.contentWidth,0,r.height,0)),this.onDidScroll=this._scrollable.onDidScroll,this.onDidContentSizeChange=this._scrollable.onDidContentSizeChange,this._updateHeight()}dispose(){super.dispose()}getScrollable(){return this._scrollable.getScrollable()}onHeightMaybeChanged(){this._updateHeight()}_configureSmoothScrollDuration(){this._scrollable.setSmoothScrollDuration(this._configuration.options.get(130)?fSe:0)}onConfigurationChanged(e){const t=this._configuration.options;if(e.hasChanged(75)&&this._linesLayout.setDefaultLineHeight(t.get(75)),e.hasChanged(96)){const i=t.get(96);this._linesLayout.setPadding(i.top,i.bottom)}if(e.hasChanged(165)){const i=t.get(165),n=i.contentWidth,o=i.height,r=this._scrollable.getScrollDimensions(),a=r.contentWidth;this._scrollable.setScrollDimensions(new d1(n,r.contentWidth,o,this._getContentHeight(n,o,a)))}else this._updateHeight();e.hasChanged(130)&&this._configureSmoothScrollDuration()}onFlushed(e,t){this._linesLayout.onFlushed(e,t)}onLinesDeleted(e,t){this._linesLayout.onLinesDeleted(e,t)}onLinesInserted(e,t){this._linesLayout.onLinesInserted(e,t)}_getHorizontalScrollbarHeight(e,t){const n=this._configuration.options.get(117);return n.horizontal===2||e>=t?0:n.horizontalScrollbarSize}_getContentHeight(e,t,i){const n=this._configuration.options;let o=this._linesLayout.getLinesTotalHeight();return n.get(119)?o+=Math.max(0,t-n.get(75)-n.get(96).bottom):n.get(117).ignoreHorizontalScrollbarInContentHeight||(o+=this._getHorizontalScrollbarHeight(e,i)),o}_updateHeight(){const e=this._scrollable.getScrollDimensions(),t=e.width,i=e.height,n=e.contentWidth;this._scrollable.setScrollDimensions(new d1(t,e.contentWidth,i,this._getContentHeight(t,i,n)))}getCurrentViewport(){const e=this._scrollable.getScrollDimensions(),t=this._scrollable.getCurrentScrollPosition();return new X$(t.scrollTop,t.scrollLeft,e.width,e.height)}getFutureViewport(){const e=this._scrollable.getScrollDimensions(),t=this._scrollable.getFutureScrollPosition();return new X$(t.scrollTop,t.scrollLeft,e.width,e.height)}_computeContentWidth(){const e=this._configuration.options,t=this._maxLineWidth,i=e.get(166),n=e.get(59),o=e.get(165);if(i.isViewportWrapping){const r=e.get(81);return t>o.contentWidth+n.typicalHalfwidthCharacterWidth&&r.enabled&&r.side==="right"?t+o.verticalScrollbarWidth:t}else{const r=e.get(118)*n.typicalHalfwidthCharacterWidth,a=this._linesLayout.getWhitespaceMinWidth();return Math.max(t+r+o.verticalScrollbarWidth,a,this._overlayWidgetsMinWidth)}}setMaxLineWidth(e){this._maxLineWidth=e,this._updateContentWidth()}setOverlayWidgetsMinWidth(e){this._overlayWidgetsMinWidth=e,this._updateContentWidth()}_updateContentWidth(){const e=this._scrollable.getScrollDimensions();this._scrollable.setScrollDimensions(new d1(e.width,this._computeContentWidth(),e.height,e.contentHeight)),this._updateHeight()}saveState(){const e=this._scrollable.getFutureScrollPosition(),t=e.scrollTop,i=this._linesLayout.getLineNumberAtOrAfterVerticalOffset(t),n=this._linesLayout.getWhitespaceAccumulatedHeightBeforeLineNumber(i);return{scrollTop:t,scrollTopWithoutViewZones:t-n,scrollLeft:e.scrollLeft}}changeWhitespace(e){const t=this._linesLayout.changeWhitespace(e);return t&&this.onHeightMaybeChanged(),t}changeSpecialLineHeights(e){const t=this._linesLayout.changeLineHeights(e);return t&&this.onHeightMaybeChanged(),t}getVerticalOffsetForLineNumber(e,t=!1){return this._linesLayout.getVerticalOffsetForLineNumber(e,t)}getVerticalOffsetAfterLineNumber(e,t=!1){return this._linesLayout.getVerticalOffsetAfterLineNumber(e,t)}getLineHeightForLineNumber(e){return this._linesLayout.getLineHeightForLineNumber(e)}isAfterLines(e){return this._linesLayout.isAfterLines(e)}isInTopPadding(e){return this._linesLayout.isInTopPadding(e)}isInBottomPadding(e){return this._linesLayout.isInBottomPadding(e)}getLineNumberAtVerticalOffset(e){return this._linesLayout.getLineNumberAtOrAfterVerticalOffset(e)}getWhitespaceAtVerticalOffset(e){return this._linesLayout.getWhitespaceAtVerticalOffset(e)}getLinesViewportData(){const e=this.getCurrentViewport();return this._linesLayout.getLinesViewportData(e.top,e.top+e.height)}getLinesViewportDataAtScrollTop(e){const t=this._scrollable.getScrollDimensions();return e+t.height>t.scrollHeight&&(e=t.scrollHeight-t.height),e<0&&(e=0),this._linesLayout.getLinesViewportData(e,e+t.height)}getWhitespaceViewportData(){const e=this.getCurrentViewport();return this._linesLayout.getWhitespaceViewportData(e.top,e.top+e.height)}getWhitespaces(){return this._linesLayout.getWhitespaces()}getContentWidth(){return this._scrollable.getScrollDimensions().contentWidth}getScrollWidth(){return this._scrollable.getScrollDimensions().scrollWidth}getContentHeight(){return this._scrollable.getScrollDimensions().contentHeight}getScrollHeight(){return this._scrollable.getScrollDimensions().scrollHeight}getCurrentScrollLeft(){return this._scrollable.getCurrentScrollPosition().scrollLeft}getCurrentScrollTop(){return this._scrollable.getCurrentScrollPosition().scrollTop}validateScrollPosition(e){return this._scrollable.validateScrollPosition(e)}setScrollPosition(e,t){t===1?this._scrollable.setScrollPositionNow(e):this._scrollable.setScrollPositionSmooth(e)}hasPendingScrollAnimation(){return this._scrollable.hasPendingScrollAnimation()}deltaScrollNow(e,t){const i=this._scrollable.getCurrentScrollPosition();this._scrollable.setScrollPositionNow({scrollLeft:i.scrollLeft+e,scrollTop:i.scrollTop+t})}}class fp{constructor(e,t,i){this.range=e,this.inlineClassName=t,this.type=i}}class _Se{constructor(e,t,i,n){this.startOffset=e,this.endOffset=t,this.inlineClassName=i,this.inlineClassNameAffectsLetterSpacing=n}toInlineDecoration(e){return new fp(new k(e,this.startOffset+1,e,this.endOffset+1),this.inlineClassName,this.inlineClassNameAffectsLetterSpacing?3:0)}}class bSe{constructor(e,t,i,n,o){this.editorId=e,this.model=t,this.configuration=i,this._linesCollection=n,this._coordinatesConverter=o,this._decorationsCache=Object.create(null),this._cachedModelDecorationsResolver=null,this._cachedModelDecorationsResolverViewRange=null}_clearCachedModelDecorationsResolver(){this._cachedModelDecorationsResolver=null,this._cachedModelDecorationsResolverViewRange=null}dispose(){this._decorationsCache=Object.create(null),this._clearCachedModelDecorationsResolver()}reset(){this._decorationsCache=Object.create(null),this._clearCachedModelDecorationsResolver()}onModelDecorationsChanged(){this._decorationsCache=Object.create(null),this._clearCachedModelDecorationsResolver()}onLineMappingChanged(){this._decorationsCache=Object.create(null),this._clearCachedModelDecorationsResolver()}_getOrCreateViewModelDecoration(e){const t=e.id;let i=this._decorationsCache[t];if(!i){const n=e.range,o=e.options;let r;if(o.isWholeLine){const a=this._coordinatesConverter.convertModelPositionToViewPosition(new A(n.startLineNumber,1),0,!1,!0),l=this._coordinatesConverter.convertModelPositionToViewPosition(new A(n.endLineNumber,this.model.getLineMaxColumn(n.endLineNumber)),1);r=new k(a.lineNumber,a.column,l.lineNumber,l.column)}else r=this._coordinatesConverter.convertModelRangeToViewRange(n,1);i=new Iee(r,o),this._decorationsCache[t]=i}return i}getMinimapDecorationsInRange(e){return this._getDecorationsInRange(e,!0,!1).decorations}getDecorationsViewportData(e){let t=this._cachedModelDecorationsResolver!==null;return t=t&&e.equalsRange(this._cachedModelDecorationsResolverViewRange),t||(this._cachedModelDecorationsResolver=this._getDecorationsInRange(e,!1,!1),this._cachedModelDecorationsResolverViewRange=e),this._cachedModelDecorationsResolver}getDecorationsOnLine(e,t=!1,i=!1){const n=new k(e,this._linesCollection.getViewLineMinColumn(e),e,this._linesCollection.getViewLineMaxColumn(e));return this._getDecorationsInRange(n,t,i)}_getDecorationsInRange(e,t,i){const n=this._linesCollection.getDecorationsInRange(e,this.editorId,WE(this.configuration.options),HE(this.configuration.options),t,i),o=e.startLineNumber,r=e.endLineNumber,a=[];let l=0;const c=[];for(let h=o;h<=r;h++)c[h-o]=[];let d=!1;for(let h=0,u=n.length;h<u;h++){const g=n[h],f=g.options;if(!KW(this.model,g))continue;const p=this._getOrCreateViewModelDecoration(g),_=p.range;if(a[l++]=p,f.inlineClassName){const b=new fp(_,f.inlineClassName,f.inlineClassNameAffectsLetterSpacing?3:0),w=Math.max(o,_.startLineNumber),C=Math.min(r,_.endLineNumber);for(let v=w;v<=C;v++)c[v-o].push(b)}if(f.beforeContentClassName&&o<=_.startLineNumber&&_.startLineNumber<=r){const b=new fp(new k(_.startLineNumber,_.startColumn,_.startLineNumber,_.startColumn),f.beforeContentClassName,1);c[_.startLineNumber-o].push(b)}if(f.afterContentClassName&&o<=_.endLineNumber&&_.endLineNumber<=r){const b=new fp(new k(_.endLineNumber,_.endColumn,_.endLineNumber,_.endColumn),f.afterContentClassName,2);c[_.endLineNumber-o].push(b)}f.affectsFont&&(d=!0)}return{decorations:a,inlineDecorations:c,hasVariableFonts:d}}}function wM(s,e){return s===null?e?pI.INSTANCE:mI.INSTANCE:new wSe(s,e)}class wSe{constructor(e,t){this._projectionData=e,this._isVisible=t}isVisible(){return this._isVisible}setVisible(e){return this._isVisible=e,this}getProjectionData(){return this._projectionData}getViewLineCount(){return this._isVisible?this._projectionData.getOutputLineCount():0}getViewLineContent(e,t,i){this._assertVisible();const n=i>0?this._projectionData.breakOffsets[i-1]:0,o=this._projectionData.breakOffsets[i];let r;if(this._projectionData.injectionOffsets!==null){const a=this._projectionData.injectionOffsets.map((c,d)=>new El(0,0,c+1,this._projectionData.injectionOptions[d],0));r=El.applyInjectedText(e.getLineContent(t),a).substring(n,o)}else r=e.getValueInRange({startLineNumber:t,startColumn:n+1,endLineNumber:t,endColumn:o+1});return i>0&&(r=QU(this._projectionData.wrappedTextIndentLength)+r),r}getViewLineLength(e,t,i){return this._assertVisible(),this._projectionData.getLineLength(i)}getViewLineMinColumn(e,t,i){return this._assertVisible(),this._projectionData.getMinOutputOffset(i)+1}getViewLineMaxColumn(e,t,i){return this._assertVisible(),this._projectionData.getMaxOutputOffset(i)+1}getViewLineData(e,t,i){const n=new Array;return this.getViewLinesData(e,t,i,1,0,[!0],n),n[0]}getViewLinesData(e,t,i,n,o,r,a){this._assertVisible();const l=this._projectionData,c=l.injectionOffsets,d=l.injectionOptions;let h=null;if(c){h=[];let g=0,f=0;for(let p=0;p<l.getOutputLineCount();p++){const _=new Array;h[p]=_;const b=p>0?l.breakOffsets[p-1]:0,w=l.breakOffsets[p];for(;f<c.length;){const C=d[f].content.length,v=c[f]+g,y=v+C;if(v>w)break;if(b<y){const x=d[f];if(x.inlineClassName){const L=p>0?l.wrappedTextIndentLength:0,E=L+Math.max(v-b,0),I=L+Math.min(y-b,w-b);E!==I&&_.push(new _Se(E,I,x.inlineClassName,x.inlineClassNameAffectsLetterSpacing))}}if(y<=w)g+=C,f++;else break}}}let u;if(c){const g=[];for(let f=0;f<c.length;f++){const p=c[f],_=d[f].tokens;_?_.forEach((b,w)=>{g.push({offset:p,text:b.substring(d[f].content),tokenMetadata:w.metadata})}):g.push({offset:p,text:d[f].content,tokenMetadata:Ri.defaultTokenMetadata})}u=e.tokenization.getLineTokens(t).withInserted(g)}else u=e.tokenization.getLineTokens(t);for(let g=i;g<i+n;g++){const f=o+g-i;if(!r[f]){a[f]=null;continue}a[f]=this._getViewLineData(u,h?h[g]:null,g)}}_getViewLineData(e,t,i){this._assertVisible();const n=this._projectionData,o=i>0?n.wrappedTextIndentLength:0,r=i>0?n.breakOffsets[i-1]:0,a=n.breakOffsets[i],l=e.sliceAndInflate(r,a,o);let c=l.getLineContent();i>0&&(c=QU(n.wrappedTextIndentLength)+c);const d=this._projectionData.getMinOutputOffset(i)+1,h=c.length+1,u=i+1<this.getViewLineCount(),g=i===0?0:n.breakOffsetsVisibleColumn[i-1];return new QW(c,u,d,h,g,l,t)}getModelColumnOfViewPosition(e,t){return this._assertVisible(),this._projectionData.translateToInputOffset(e,t-1)+1}getViewPositionOfModelPosition(e,t,i=2){return this._assertVisible(),this._projectionData.translateToOutputPosition(t-1,i).toPosition(e)}getViewLineNumberOfModelPosition(e,t){this._assertVisible();const i=this._projectionData.translateToOutputPosition(t-1);return e+i.outputLineIndex}normalizePosition(e,t,i){const n=t.lineNumber-e;return this._projectionData.normalizeOutputPosition(e,t.column-1,i).toPosition(n)}getInjectedTextAt(e,t){return this._projectionData.getInjectedText(e,t-1)}_assertVisible(){if(!this._isVisible)throw new Error("Not supported")}}const sN=class sN{constructor(){}isVisible(){return!0}setVisible(e){return e?this:mI.INSTANCE}getProjectionData(){return null}getViewLineCount(){return 1}getViewLineContent(e,t,i){return e.getLineContent(t)}getViewLineLength(e,t,i){return e.getLineLength(t)}getViewLineMinColumn(e,t,i){return e.getLineMinColumn(t)}getViewLineMaxColumn(e,t,i){return e.getLineMaxColumn(t)}getViewLineData(e,t,i){const n=e.tokenization.getLineTokens(t),o=n.getLineContent();return new QW(o,!1,1,o.length+1,0,n.inflate(),null)}getViewLinesData(e,t,i,n,o,r,a){if(!r[o]){a[o]=null;return}a[o]=this.getViewLineData(e,t,0)}getModelColumnOfViewPosition(e,t){return t}getViewPositionOfModelPosition(e,t){return new A(e,t)}getViewLineNumberOfModelPosition(e,t){return e}normalizePosition(e,t,i){return t}getInjectedTextAt(e,t){return null}};sN.INSTANCE=new sN;let pI=sN;const oN=class oN{constructor(){}isVisible(){return!1}setVisible(e){return e?pI.INSTANCE:this}getProjectionData(){return null}getViewLineCount(){return 0}getViewLineContent(e,t,i){throw new Error("Not supported")}getViewLineLength(e,t,i){throw new Error("Not supported")}getViewLineMinColumn(e,t,i){throw new Error("Not supported")}getViewLineMaxColumn(e,t,i){throw new Error("Not supported")}getViewLineData(e,t,i){throw new Error("Not supported")}getViewLinesData(e,t,i,n,o,r,a){throw new Error("Not supported")}getModelColumnOfViewPosition(e,t){throw new Error("Not supported")}getViewPositionOfModelPosition(e,t){throw new Error("Not supported")}getViewLineNumberOfModelPosition(e,t){throw new Error("Not supported")}normalizePosition(e,t,i){throw new Error("Not supported")}getInjectedTextAt(e,t){throw new Error("Not supported")}};oN.INSTANCE=new oN;let mI=oN;const vM=[""];function QU(s){if(s>=vM.length)for(let e=1;e<=s;e++)vM[e]=vSe(e);return vM[s]}function vSe(s){return new Array(s+1).join(" ")}class CSe{constructor(e){this.values=e,this.prefixSum=new Uint32Array(e.length),this.prefixSumValidIndex=new Int32Array(1),this.prefixSumValidIndex[0]=-1}insertValues(e,t){e=W_(e);const i=this.values,n=this.prefixSum,o=t.length;return o===0?!1:(this.values=new Uint32Array(i.length+o),this.values.set(i.subarray(0,e),0),this.values.set(i.subarray(e),e+o),this.values.set(t,e),e-1<this.prefixSumValidIndex[0]&&(this.prefixSumValidIndex[0]=e-1),this.prefixSum=new Uint32Array(this.values.length),this.prefixSumValidIndex[0]>=0&&this.prefixSum.set(n.subarray(0,this.prefixSumValidIndex[0]+1)),!0)}setValue(e,t){return e=W_(e),t=W_(t),this.values[e]===t?!1:(this.values[e]=t,e-1<this.prefixSumValidIndex[0]&&(this.prefixSumValidIndex[0]=e-1),!0)}removeValues(e,t){e=W_(e),t=W_(t);const i=this.values,n=this.prefixSum;if(e>=i.length)return!1;const o=i.length-e;return t>=o&&(t=o),t===0?!1:(this.values=new Uint32Array(i.length-t),this.values.set(i.subarray(0,e),0),this.values.set(i.subarray(e+t),e),this.prefixSum=new Uint32Array(this.values.length),e-1<this.prefixSumValidIndex[0]&&(this.prefixSumValidIndex[0]=e-1),this.prefixSumValidIndex[0]>=0&&this.prefixSum.set(n.subarray(0,this.prefixSumValidIndex[0]+1)),!0)}getTotalSum(){return this.values.length===0?0:this._getPrefixSum(this.values.length-1)}getPrefixSum(e){return e<0?0:(e=W_(e),this._getPrefixSum(e))}_getPrefixSum(e){if(e<=this.prefixSumValidIndex[0])return this.prefixSum[e];let t=this.prefixSumValidIndex[0]+1;t===0&&(this.prefixSum[0]=this.values[0],t++),e>=this.values.length&&(e=this.values.length-1);for(let i=t;i<=e;i++)this.prefixSum[i]=this.prefixSum[i-1]+this.values[i];return this.prefixSumValidIndex[0]=Math.max(this.prefixSumValidIndex[0],e),this.prefixSum[e]}getIndexOf(e){e=Math.floor(e),this.getTotalSum();let t=0,i=this.values.length-1,n=0,o=0,r=0;for(;t<=i;)if(n=t+(i-t)/2|0,o=this.prefixSum[n],r=o-this.values[n],e<r)i=n-1;else if(e>=o)t=n+1;else break;return new _te(n,e-r)}}class ySe{constructor(e){this._values=e,this._isValid=!1,this._validEndIndex=-1,this._prefixSum=[],this._indexBySum=[]}getTotalSum(){return this._ensureValid(),this._indexBySum.length}getPrefixSum(e){return this._ensureValid(),e===0?0:this._prefixSum[e-1]}getIndexOf(e){this._ensureValid();const t=this._indexBySum[e],i=t>0?this._prefixSum[t-1]:0;return new _te(t,e-i)}removeValues(e,t){this._values.splice(e,t),this._invalidate(e)}insertValues(e,t){this._values=IR(this._values,e,t),this._invalidate(e)}_invalidate(e){this._isValid=!1,this._validEndIndex=Math.min(this._validEndIndex,e-1)}_ensureValid(){if(!this._isValid){for(let e=this._validEndIndex+1,t=this._values.length;e<t;e++){const i=this._values[e],n=e>0?this._prefixSum[e-1]:0;this._prefixSum[e]=n+i;for(let o=0;o<i;o++)this._indexBySum[n+o]=e}this._prefixSum.length=this._values.length,this._indexBySum.length=this._prefixSum[this._prefixSum.length-1],this._isValid=!0,this._validEndIndex=this._values.length-1}}setValue(e,t){this._values[e]!==t&&(this._values[e]=t,this._invalidate(e))}}class _te{constructor(e,t){this.index=e,this.remainder=t,this._prefixSumIndexOfResultBrand=void 0,this.index=e,this.remainder=t}}class SSe{constructor(e){this._model=e}_validPosition(e){return this._model.validatePosition(e)}_validRange(e){return this._model.validateRange(e)}convertViewPositionToModelPosition(e){return this._validPosition(e)}convertViewRangeToModelRange(e){return this._validRange(e)}validateViewPosition(e,t){return this._validPosition(t)}validateViewRange(e,t){return this._validRange(t)}convertModelPositionToViewPosition(e){return this._validPosition(e)}convertModelRangeToViewRange(e){return this._validRange(e)}modelPositionIsVisible(e){const t=this._model.getLineCount();return!(e.lineNumber<1||e.lineNumber>t)}getModelLineViewLineCount(e){return 1}getViewLineNumberOfModelPosition(e,t){return e}}class xSe{constructor(e,t,i,n,o,r,a,l,c,d,h){this._editorId=e,this.model=t,this._validModelVersionId=-1,this._domLineBreaksComputerFactory=i,this._monospaceLineBreaksComputerFactory=n,this.fontInfo=o,this.tabSize=r,this.wrappingStrategy=a,this.wrappingColumn=l,this.wrappingIndent=c,this.wordBreak=d,this.wrapOnEscapedLineFeeds=h,this._constructLines(!0,null)}dispose(){this.hiddenAreasDecorationIds=this.model.deltaDecorations(this.hiddenAreasDecorationIds,[])}createCoordinatesConverter(){return new kSe(this)}_constructLines(e,t){this.modelLineProjections=[],e&&(this.hiddenAreasDecorationIds=this.model.deltaDecorations(this.hiddenAreasDecorationIds,[]));const i=this.model.getLinesContent(),n=this.model.getInjectedTextDecorations(this._editorId),o=i.length,r=this.createLineBreaksComputer(),a=new ph(El.fromDecorations(n));for(let p=0;p<o;p++){const _=a.takeWhile(b=>b.lineNumber===p+1);r.addRequest(i[p],_,t?t[p]:null)}const l=r.finalize(),c=[],d=this.hiddenAreasDecorationIds.map(p=>this.model.getDecorationRange(p)).sort(k.compareRangesUsingStarts);let h=1,u=0,g=-1,f=g+1<d.length?u+1:o+2;for(let p=0;p<o;p++){const _=p+1;_===f&&(g++,h=d[g].startLineNumber,u=d[g].endLineNumber,f=g+1<d.length?u+1:o+2);const b=_>=h&&_<=u,w=wM(l[p],!b);c[p]=w.getViewLineCount(),this.modelLineProjections[p]=w}this._validModelVersionId=this.model.getVersionId(),this.projectedModelLineLineCounts=new ySe(c)}getHiddenAreas(){return this.hiddenAreasDecorationIds.map(e=>this.model.getDecorationRange(e))}setHiddenAreas(e){const t=e.map(u=>this.model.validateRange(u)),i=LSe(t),n=this.hiddenAreasDecorationIds.map(u=>this.model.getDecorationRange(u)).sort(k.compareRangesUsingStarts);if(i.length===n.length){let u=!1;for(let g=0;g<i.length;g++)if(!i[g].equalsRange(n[g])){u=!0;break}if(!u)return!1}const o=i.map(u=>({range:u,options:Fe.EMPTY}));this.hiddenAreasDecorationIds=this.model.deltaDecorations(this.hiddenAreasDecorationIds,o);const r=i;let a=1,l=0,c=-1,d=c+1<r.length?l+1:this.modelLineProjections.length+2,h=!1;for(let u=0;u<this.modelLineProjections.length;u++){const g=u+1;g===d&&(c++,a=r[c].startLineNumber,l=r[c].endLineNumber,d=c+1<r.length?l+1:this.modelLineProjections.length+2);let f=!1;if(g>=a&&g<=l?this.modelLineProjections[u].isVisible()&&(this.modelLineProjections[u]=this.modelLineProjections[u].setVisible(!1),f=!0):(h=!0,this.modelLineProjections[u].isVisible()||(this.modelLineProjections[u]=this.modelLineProjections[u].setVisible(!0),f=!0)),f){const p=this.modelLineProjections[u].getViewLineCount();this.projectedModelLineLineCounts.setValue(u,p)}}return h||this.setHiddenAreas([]),!0}modelPositionIsVisible(e,t){return e<1||e>this.modelLineProjections.length?!1:this.modelLineProjections[e-1].isVisible()}getModelLineViewLineCount(e){return e<1||e>this.modelLineProjections.length?1:this.modelLineProjections[e-1].getViewLineCount()}setTabSize(e){return this.tabSize===e?!1:(this.tabSize=e,this._constructLines(!1,null),!0)}setWrappingSettings(e,t,i,n,o){const r=this.fontInfo.equals(e),a=this.wrappingStrategy===t,l=this.wrappingColumn===i,c=this.wrappingIndent===n,d=this.wordBreak===o;if(r&&a&&l&&c&&d)return!1;const h=r&&a&&!l&&c&&d;this.fontInfo=e,this.wrappingStrategy=t,this.wrappingColumn=i,this.wrappingIndent=n,this.wordBreak=o;let u=null;if(h){u=[];for(let g=0,f=this.modelLineProjections.length;g<f;g++)u[g]=this.modelLineProjections[g].getProjectionData()}return this._constructLines(!1,u),!0}createLineBreaksComputer(){return(this.wrappingStrategy==="advanced"?this._domLineBreaksComputerFactory:this._monospaceLineBreaksComputerFactory).createLineBreaksComputer(this.fontInfo,this.tabSize,this.wrappingColumn,this.wrappingIndent,this.wordBreak,this.wrapOnEscapedLineFeeds)}onModelFlushed(){this._constructLines(!0,null)}onModelLinesDeleted(e,t,i){if(!e||e<=this._validModelVersionId)return null;const n=t===1?1:this.projectedModelLineLineCounts.getPrefixSum(t-1)+1,o=this.projectedModelLineLineCounts.getPrefixSum(i);return this.modelLineProjections.splice(t-1,i-t+1),this.projectedModelLineLineCounts.removeValues(t-1,i-t+1),new qF(n,o)}onModelLinesInserted(e,t,i,n){if(!e||e<=this._validModelVersionId)return null;const o=t>2&&!this.modelLineProjections[t-2].isVisible(),r=t===1?1:this.projectedModelLineLineCounts.getPrefixSum(t-1)+1;let a=0;const l=[],c=[];for(let d=0,h=n.length;d<h;d++){const u=wM(n[d],!o);l.push(u);const g=u.getViewLineCount();a+=g,c[d]=g}return this.modelLineProjections=this.modelLineProjections.slice(0,t-1).concat(l).concat(this.modelLineProjections.slice(t-1)),this.projectedModelLineLineCounts.insertValues(t-1,c),new jF(r,r+a-1)}onModelLineChanged(e,t,i){if(e!==null&&e<=this._validModelVersionId)return[!1,null,null,null];const n=t-1,o=this.modelLineProjections[n].getViewLineCount(),r=this.modelLineProjections[n].isVisible(),a=wM(i,r);this.modelLineProjections[n]=a;const l=this.modelLineProjections[n].getViewLineCount();let c=!1,d=0,h=-1,u=0,g=-1,f=0,p=-1;o>l?(d=this.projectedModelLineLineCounts.getPrefixSum(t-1)+1,h=d+l-1,f=h+1,p=f+(o-l)-1,c=!0):o<l?(d=this.projectedModelLineLineCounts.getPrefixSum(t-1)+1,h=d+o-1,u=h+1,g=u+(l-o)-1,c=!0):(d=this.projectedModelLineLineCounts.getPrefixSum(t-1)+1,h=d+l-1),this.projectedModelLineLineCounts.setValue(n,l);const _=d<=h?new mte(d,h-d+1):null,b=u<=g?new jF(u,g):null,w=f<=p?new qF(f,p):null;return[c,_,b,w]}acceptVersionId(e){this._validModelVersionId=e,this.modelLineProjections.length===1&&!this.modelLineProjections[0].isVisible()&&this.setHiddenAreas([])}getViewLineCount(){return this.projectedModelLineLineCounts.getTotalSum()}_toValidViewLineNumber(e){if(e<1)return 1;const t=this.getViewLineCount();return e>t?t:e|0}getActiveIndentGuide(e,t,i){e=this._toValidViewLineNumber(e),t=this._toValidViewLineNumber(t),i=this._toValidViewLineNumber(i);const n=this.convertViewPositionToModelPosition(e,this.getViewLineMinColumn(e)),o=this.convertViewPositionToModelPosition(t,this.getViewLineMinColumn(t)),r=this.convertViewPositionToModelPosition(i,this.getViewLineMinColumn(i)),a=this.model.guides.getActiveIndentGuide(n.lineNumber,o.lineNumber,r.lineNumber),l=this.convertModelPositionToViewPosition(a.startLineNumber,1),c=this.convertModelPositionToViewPosition(a.endLineNumber,this.model.getLineMaxColumn(a.endLineNumber));return{startLineNumber:l.lineNumber,endLineNumber:c.lineNumber,indent:a.indent}}getViewLineInfo(e){e=this._toValidViewLineNumber(e);const t=this.projectedModelLineLineCounts.getIndexOf(e-1),i=t.index,n=t.remainder;return new JU(i+1,n)}getMinColumnOfViewLine(e){return this.modelLineProjections[e.modelLineNumber-1].getViewLineMinColumn(this.model,e.modelLineNumber,e.modelLineWrappedLineIdx)}getMaxColumnOfViewLine(e){return this.modelLineProjections[e.modelLineNumber-1].getViewLineMaxColumn(this.model,e.modelLineNumber,e.modelLineWrappedLineIdx)}getModelStartPositionOfViewLine(e){const t=this.modelLineProjections[e.modelLineNumber-1],i=t.getViewLineMinColumn(this.model,e.modelLineNumber,e.modelLineWrappedLineIdx),n=t.getModelColumnOfViewPosition(e.modelLineWrappedLineIdx,i);return new A(e.modelLineNumber,n)}getModelEndPositionOfViewLine(e){const t=this.modelLineProjections[e.modelLineNumber-1],i=t.getViewLineMaxColumn(this.model,e.modelLineNumber,e.modelLineWrappedLineIdx),n=t.getModelColumnOfViewPosition(e.modelLineWrappedLineIdx,i);return new A(e.modelLineNumber,n)}getViewLineInfosGroupedByModelRanges(e,t){const i=this.getViewLineInfo(e),n=this.getViewLineInfo(t),o=new Array;let r=this.getModelStartPositionOfViewLine(i),a=new Array;for(let l=i.modelLineNumber;l<=n.modelLineNumber;l++){const c=this.modelLineProjections[l-1];if(c.isVisible()){const d=l===i.modelLineNumber?i.modelLineWrappedLineIdx:0,h=l===n.modelLineNumber?n.modelLineWrappedLineIdx+1:c.getViewLineCount();for(let u=d;u<h;u++)a.push(new JU(l,u))}if(!c.isVisible()&&r){const d=new A(l-1,this.model.getLineMaxColumn(l-1)+1),h=k.fromPositions(r,d);o.push(new eq(h,a)),a=[],r=null}else c.isVisible()&&!r&&(r=new A(l,1))}if(r){const l=k.fromPositions(r,this.getModelEndPositionOfViewLine(n));o.push(new eq(l,a))}return o}getViewLinesBracketGuides(e,t,i,n){const o=i?this.convertViewPositionToModelPosition(i.lineNumber,i.column):null,r=[];for(const a of this.getViewLineInfosGroupedByModelRanges(e,t)){const l=a.modelRange.startLineNumber,c=this.model.guides.getLinesBracketGuides(l,a.modelRange.endLineNumber,o,n);for(const d of a.viewLines){const u=c[d.modelLineNumber-l].map(g=>{if(g.forWrappedLinesAfterColumn!==-1&&this.modelLineProjections[d.modelLineNumber-1].getViewPositionOfModelPosition(0,g.forWrappedLinesAfterColumn).lineNumber>=d.modelLineWrappedLineIdx||g.forWrappedLinesBeforeOrAtColumn!==-1&&this.modelLineProjections[d.modelLineNumber-1].getViewPositionOfModelPosition(0,g.forWrappedLinesBeforeOrAtColumn).lineNumber<d.modelLineWrappedLineIdx)return;if(!g.horizontalLine)return g;let f=-1;if(g.column!==-1){const b=this.modelLineProjections[d.modelLineNumber-1].getViewPositionOfModelPosition(0,g.column);if(b.lineNumber===d.modelLineWrappedLineIdx)f=b.column;else if(b.lineNumber<d.modelLineWrappedLineIdx)f=this.getMinColumnOfViewLine(d);else if(b.lineNumber>d.modelLineWrappedLineIdx)return}const p=this.convertModelPositionToViewPosition(d.modelLineNumber,g.horizontalLine.endColumn),_=this.modelLineProjections[d.modelLineNumber-1].getViewPositionOfModelPosition(0,g.horizontalLine.endColumn);return _.lineNumber===d.modelLineWrappedLineIdx?new Gf(g.visibleColumn,f,g.className,new N1(g.horizontalLine.top,p.column),-1,-1):_.lineNumber<d.modelLineWrappedLineIdx||g.visibleColumn!==-1?void 0:new Gf(g.visibleColumn,f,g.className,new N1(g.horizontalLine.top,this.getMaxColumnOfViewLine(d)),-1,-1)});r.push(u.filter(g=>!!g))}}return r}getViewLinesIndentGuides(e,t){e=this._toValidViewLineNumber(e),t=this._toValidViewLineNumber(t);const i=this.convertViewPositionToModelPosition(e,this.getViewLineMinColumn(e)),n=this.convertViewPositionToModelPosition(t,this.getViewLineMaxColumn(t));let o=[];const r=[],a=[],l=i.lineNumber-1,c=n.lineNumber-1;let d=null;for(let f=l;f<=c;f++){const p=this.modelLineProjections[f];if(p.isVisible()){const _=p.getViewLineNumberOfModelPosition(0,f===l?i.column:1),b=p.getViewLineNumberOfModelPosition(0,this.model.getLineMaxColumn(f+1)),w=b-_+1;let C=0;w>1&&p.getViewLineMinColumn(this.model,f+1,b)===1&&(C=_===0?1:2),r.push(w),a.push(C),d===null&&(d=new A(f+1,0))}else d!==null&&(o=o.concat(this.model.guides.getLinesIndentGuides(d.lineNumber,f)),d=null)}d!==null&&(o=o.concat(this.model.guides.getLinesIndentGuides(d.lineNumber,n.lineNumber)),d=null);const h=t-e+1,u=new Array(h);let g=0;for(let f=0,p=o.length;f<p;f++){let _=o[f];const b=Math.min(h-g,r[f]),w=a[f];let C;w===2?C=0:w===1?C=1:C=b;for(let v=0;v<b;v++)v===C&&(_=0),u[g++]=_}return u}getViewLineContent(e){const t=this.getViewLineInfo(e);return this.modelLineProjections[t.modelLineNumber-1].getViewLineContent(this.model,t.modelLineNumber,t.modelLineWrappedLineIdx)}getViewLineLength(e){const t=this.getViewLineInfo(e);return this.modelLineProjections[t.modelLineNumber-1].getViewLineLength(this.model,t.modelLineNumber,t.modelLineWrappedLineIdx)}getViewLineMinColumn(e){const t=this.getViewLineInfo(e);return this.modelLineProjections[t.modelLineNumber-1].getViewLineMinColumn(this.model,t.modelLineNumber,t.modelLineWrappedLineIdx)}getViewLineMaxColumn(e){const t=this.getViewLineInfo(e);return this.modelLineProjections[t.modelLineNumber-1].getViewLineMaxColumn(this.model,t.modelLineNumber,t.modelLineWrappedLineIdx)}getViewLineData(e){const t=this.getViewLineInfo(e);return this.modelLineProjections[t.modelLineNumber-1].getViewLineData(this.model,t.modelLineNumber,t.modelLineWrappedLineIdx)}getViewLinesData(e,t,i){e=this._toValidViewLineNumber(e),t=this._toValidViewLineNumber(t);const n=this.projectedModelLineLineCounts.getIndexOf(e-1);let o=e;const r=n.index,a=n.remainder,l=[];for(let c=r,d=this.model.getLineCount();c<d;c++){const h=this.modelLineProjections[c];if(!h.isVisible())continue;const u=c===r?a:0;let g=h.getViewLineCount()-u,f=!1;if(o+g>t&&(f=!0,g=t-o+1),h.getViewLinesData(this.model,c+1,u,g,o-e,i,l),o+=g,f)break}return l}validateViewPosition(e,t,i){e=this._toValidViewLineNumber(e);const n=this.projectedModelLineLineCounts.getIndexOf(e-1),o=n.index,r=n.remainder,a=this.modelLineProjections[o],l=a.getViewLineMinColumn(this.model,o+1,r),c=a.getViewLineMaxColumn(this.model,o+1,r);t<l&&(t=l),t>c&&(t=c);const d=a.getModelColumnOfViewPosition(r,t);return this.model.validatePosition(new A(o+1,d)).equals(i)?new A(e,t):this.convertModelPositionToViewPosition(i.lineNumber,i.column)}validateViewRange(e,t){const i=this.validateViewPosition(e.startLineNumber,e.startColumn,t.getStartPosition()),n=this.validateViewPosition(e.endLineNumber,e.endColumn,t.getEndPosition());return new k(i.lineNumber,i.column,n.lineNumber,n.column)}convertViewPositionToModelPosition(e,t){const i=this.getViewLineInfo(e),n=this.modelLineProjections[i.modelLineNumber-1].getModelColumnOfViewPosition(i.modelLineWrappedLineIdx,t);return this.model.validatePosition(new A(i.modelLineNumber,n))}convertViewRangeToModelRange(e){const t=this.convertViewPositionToModelPosition(e.startLineNumber,e.startColumn),i=this.convertViewPositionToModelPosition(e.endLineNumber,e.endColumn);return new k(t.lineNumber,t.column,i.lineNumber,i.column)}convertModelPositionToViewPosition(e,t,i=2,n=!1,o=!1){const r=this.model.validatePosition(new A(e,t)),a=r.lineNumber,l=r.column;let c=a-1,d=!1;if(o)for(;c<this.modelLineProjections.length&&!this.modelLineProjections[c].isVisible();)c++,d=!0;else for(;c>0&&!this.modelLineProjections[c].isVisible();)c--,d=!0;if(c===0&&!this.modelLineProjections[c].isVisible())return new A(n?0:1,1);const h=1+this.projectedModelLineLineCounts.getPrefixSum(c);let u;return d?o?u=this.modelLineProjections[c].getViewPositionOfModelPosition(h,1,i):u=this.modelLineProjections[c].getViewPositionOfModelPosition(h,this.model.getLineMaxColumn(c+1),i):u=this.modelLineProjections[a-1].getViewPositionOfModelPosition(h,l,i),u}convertModelRangeToViewRange(e,t=0){if(e.isEmpty()){const i=this.convertModelPositionToViewPosition(e.startLineNumber,e.startColumn,t);return k.fromPositions(i)}else{const i=this.convertModelPositionToViewPosition(e.startLineNumber,e.startColumn,1),n=this.convertModelPositionToViewPosition(e.endLineNumber,e.endColumn,0);return new k(i.lineNumber,i.column,n.lineNumber,n.column)}}getViewLineNumberOfModelPosition(e,t){let i=e-1;if(this.modelLineProjections[i].isVisible()){const o=1+this.projectedModelLineLineCounts.getPrefixSum(i);return this.modelLineProjections[i].getViewLineNumberOfModelPosition(o,t)}for(;i>0&&!this.modelLineProjections[i].isVisible();)i--;if(i===0&&!this.modelLineProjections[i].isVisible())return 1;const n=1+this.projectedModelLineLineCounts.getPrefixSum(i);return this.modelLineProjections[i].getViewLineNumberOfModelPosition(n,this.model.getLineMaxColumn(i+1))}getDecorationsInRange(e,t,i,n,o,r){const a=this.convertViewPositionToModelPosition(e.startLineNumber,e.startColumn),l=this.convertViewPositionToModelPosition(e.endLineNumber,e.endColumn);if(l.lineNumber-a.lineNumber<=e.endLineNumber-e.startLineNumber)return this.model.getDecorationsInRange(new k(a.lineNumber,1,l.lineNumber,l.column),t,i,n,o,r);let c=[];const d=a.lineNumber-1,h=l.lineNumber-1;let u=null;for(let _=d;_<=h;_++)if(this.modelLineProjections[_].isVisible())u===null&&(u=new A(_+1,_===d?a.column:1));else if(u!==null){const w=this.model.getLineMaxColumn(_);c=c.concat(this.model.getDecorationsInRange(new k(u.lineNumber,u.column,_,w),t,i,n,o)),u=null}u!==null&&(c=c.concat(this.model.getDecorationsInRange(new k(u.lineNumber,u.column,l.lineNumber,l.column),t,i,n,o)),u=null),c.sort((_,b)=>{const w=k.compareRangesUsingStarts(_.range,b.range);return w===0?_.id<b.id?-1:_.id>b.id?1:0:w});const g=[];let f=0,p=null;for(const _ of c){const b=_.id;p!==b&&(p=b,g[f++]=_)}return g}getInjectedTextAt(e){const t=this.getViewLineInfo(e.lineNumber);return this.modelLineProjections[t.modelLineNumber-1].getInjectedTextAt(t.modelLineWrappedLineIdx,e.column)}normalizePosition(e,t){const i=this.getViewLineInfo(e.lineNumber);return this.modelLineProjections[i.modelLineNumber-1].normalizePosition(i.modelLineWrappedLineIdx,e,t)}getLineIndentColumn(e){const t=this.getViewLineInfo(e);return t.modelLineWrappedLineIdx===0?this.model.getLineIndentColumn(t.modelLineNumber):0}}function LSe(s){if(s.length===0)return[];const e=s.slice();e.sort(k.compareRangesUsingStarts);const t=[];let i=e[0].startLineNumber,n=e[0].endLineNumber;for(let o=1,r=e.length;o<r;o++){const a=e[o];a.startLineNumber>n+1?(t.push(new k(i,1,n,1)),i=a.startLineNumber,n=a.endLineNumber):a.endLineNumber>n&&(n=a.endLineNumber)}return t.push(new k(i,1,n,1)),t}class JU{constructor(e,t){this.modelLineNumber=e,this.modelLineWrappedLineIdx=t}}class eq{constructor(e,t){this.modelRange=e,this.viewLines=t}}class kSe{constructor(e){this._lines=e}convertViewPositionToModelPosition(e){return this._lines.convertViewPositionToModelPosition(e.lineNumber,e.column)}convertViewRangeToModelRange(e){return this._lines.convertViewRangeToModelRange(e)}validateViewPosition(e,t){return this._lines.validateViewPosition(e.lineNumber,e.column,t)}validateViewRange(e,t){return this._lines.validateViewRange(e,t)}convertModelPositionToViewPosition(e,t,i,n){return this._lines.convertModelPositionToViewPosition(e.lineNumber,e.column,t,i,n)}convertModelRangeToViewRange(e,t){return this._lines.convertModelRangeToViewRange(e,t)}modelPositionIsVisible(e){return this._lines.modelPositionIsVisible(e.lineNumber,e.column)}getModelLineViewLineCount(e){return this._lines.getModelLineViewLineCount(e)}getViewLineNumberOfModelPosition(e,t){return this._lines.getViewLineNumberOfModelPosition(e,t)}}class ESe{constructor(e){this.model=e}dispose(){}createCoordinatesConverter(){return new SSe(this.model)}getHiddenAreas(){return[]}setHiddenAreas(e){return!1}setTabSize(e){return!1}setWrappingSettings(e,t,i,n){return!1}createLineBreaksComputer(){const e=[];return{addRequest:(t,i,n)=>{e.push(null)},finalize:()=>e}}onModelFlushed(){}onModelLinesDeleted(e,t,i){return new qF(t,i)}onModelLinesInserted(e,t,i,n){return new jF(t,i)}onModelLineChanged(e,t,i){return[!1,new mte(t,1),null,null]}acceptVersionId(e){}getViewLineCount(){return this.model.getLineCount()}getActiveIndentGuide(e,t,i){return{startLineNumber:e,endLineNumber:e,indent:0}}getViewLinesBracketGuides(e,t,i){return new Array(t-e+1).fill([])}getViewLinesIndentGuides(e,t){const i=t-e+1,n=new Array(i);for(let o=0;o<i;o++)n[o]=0;return n}getViewLineContent(e){return this.model.getLineContent(e)}getViewLineLength(e){return this.model.getLineLength(e)}getViewLineMinColumn(e){return this.model.getLineMinColumn(e)}getViewLineMaxColumn(e){return this.model.getLineMaxColumn(e)}getViewLineData(e){const t=this.model.tokenization.getLineTokens(e),i=t.getLineContent();return new QW(i,!1,1,i.length+1,0,t.inflate(),null)}getViewLinesData(e,t,i){const n=this.model.getLineCount();e=Math.min(Math.max(1,e),n),t=Math.min(Math.max(1,t),n);const o=[];for(let r=e;r<=t;r++){const a=r-e;o[a]=i[a]?this.getViewLineData(r):null}return o}getDecorationsInRange(e,t,i,n,o,r){return this.model.getDecorationsInRange(e,t,i,n,o,r)}normalizePosition(e,t){return this.model.normalizePosition(e,t)}getLineIndentColumn(e){return this.model.getLineIndentColumn(e)}getInjectedTextAt(e){return null}}const Vg=Ll.Right;class ISe{constructor(e){this.persist=0,this._requiredLanes=1,this.lanes=new Uint8Array(Math.ceil((e+1)*Vg/8))}reset(e){const t=Math.ceil((e+1)*Vg/8);this.lanes.length<t?this.lanes=new Uint8Array(t):this.lanes.fill(0),this._requiredLanes=1}get requiredLanes(){return this._requiredLanes}push(e,t,i){i&&(this.persist|=1<<e-1);for(let n=t.startLineNumber;n<=t.endLineNumber;n++){const o=Vg*n+(e-1);this.lanes[o>>>3]|=1<<o%8,this._requiredLanes=Math.max(this._requiredLanes,this.countAtLine(n))}}getLanesAtLine(e){const t=[];let i=Vg*e;for(let n=0;n<Vg;n++)(this.persist&1<<n||this.lanes[i>>>3]&1<<i%8)&&t.push(n+1),i++;return t.length?t:[Ll.Center]}countAtLine(e){let t=Vg*e,i=0;for(let n=0;n<Vg;n++)(this.persist&1<<n||this.lanes[t>>>3]&1<<t%8)&&i++,t++;return i}}let DSe=class extends F{constructor(e,t,i,n,o,r,a,l,c,d){if(super(),this.languageConfigurationService=a,this._themeService=l,this._attachedView=c,this._transactionalTarget=d,this.hiddenAreasModel=new NSe,this.previousHiddenAreas=[],this._editorId=e,this._configuration=t,this.model=i,this._eventDispatcher=new qye,this.onEvent=this._eventDispatcher.onEvent,this.cursorConfig=new B_(this.model.getLanguageId(),this.model.getOptions(),this._configuration,this.languageConfigurationService),this._updateConfigurationViewLineCount=this._register(new Rt(()=>this._updateConfigurationViewLineCountNow(),0)),this._hasFocus=!1,this._viewportStart=b7.create(this.model),this.glyphLanes=new ISe(0),this.model.isTooLargeForTokenization())this._lines=new ESe(this.model);else{const h=this._configuration.options,u=h.get(59),g=h.get(156),f=h.get(166),p=h.get(155),_=h.get(146),b=h.get(160);this._lines=new xSe(this._editorId,this.model,n,o,u,this.model.getOptions().tabSize,g,f.wrappingColumn,p,_,b)}this.coordinatesConverter=this._lines.createCoordinatesConverter(),this._cursor=this._register(new oSe(i,this,this.coordinatesConverter,this.cursorConfig)),this.viewLayout=this._register(new mSe(this._configuration,this.getLineCount(),this._getCustomLineHeights(),r)),this._register(this.viewLayout.onDidScroll(h=>{h.scrollTopChanged&&this._handleVisibleLinesChanged(),h.scrollTopChanged&&this._viewportStart.invalidate(),this._eventDispatcher.emitSingleViewEvent(new Hye(h)),this._eventDispatcher.emitOutgoingEvent(new _7(h.oldScrollWidth,h.oldScrollLeft,h.oldScrollHeight,h.oldScrollTop,h.scrollWidth,h.scrollLeft,h.scrollHeight,h.scrollTop))})),this._register(this.viewLayout.onDidContentSizeChange(h=>{this._eventDispatcher.emitOutgoingEvent(h)})),this._decorations=new bSe(this._editorId,this.model,this._configuration,this._lines,this.coordinatesConverter),this._registerModelEvents(),this._register(this._configuration.onDidChangeFast(h=>{try{const u=this._eventDispatcher.beginEmitViewEvents();this._onConfigurationChanged(u,h)}finally{this._eventDispatcher.endEmitViewEvents()}})),this._register(sI.getInstance().onDidChange(()=>{this._eventDispatcher.emitSingleViewEvent(new $ye)})),this._register(this._themeService.onDidColorThemeChange(h=>{this._invalidateDecorationsColorCache(),this._eventDispatcher.emitSingleViewEvent(new Vye(h))})),this._updateConfigurationViewLineCountNow()}dispose(){super.dispose(),this._decorations.dispose(),this._lines.dispose(),this._viewportStart.dispose(),this._eventDispatcher.dispose()}createLineBreaksComputer(){return this._lines.createLineBreaksComputer()}addViewEventHandler(e){this._eventDispatcher.addViewEventHandler(e)}removeViewEventHandler(e){this._eventDispatcher.removeViewEventHandler(e)}_getCustomLineHeights(){return this._configuration.options.get(5)?this.model.getCustomLineHeightsDecorations(this._editorId).map(i=>{const n=i.range.startLineNumber,o=this.coordinatesConverter.convertModelRangeToViewRange(new k(n,1,n,this.model.getLineMaxColumn(n)));return{decorationId:i.id,startLineNumber:o.startLineNumber,endLineNumber:o.endLineNumber,lineHeight:i.options.lineHeight||0}}):[]}_updateConfigurationViewLineCountNow(){this._configuration.setViewLineCount(this._lines.getViewLineCount())}getModelVisibleRanges(){const e=this.viewLayout.getLinesViewportData(),t=new k(e.startLineNumber,this.getLineMinColumn(e.startLineNumber),e.endLineNumber,this.getLineMaxColumn(e.endLineNumber));return this._toModelVisibleRanges(t)}visibleLinesStabilized(){const e=this.getModelVisibleRanges();this._attachedView.setVisibleLines(e,!0)}_handleVisibleLinesChanged(){const e=this.getModelVisibleRanges();this._attachedView.setVisibleLines(e,!1)}setHasFocus(e){this._hasFocus=e,this._cursor.setHasFocus(e),this._eventDispatcher.emitSingleViewEvent(new Bye(e)),this._eventDispatcher.emitOutgoingEvent(new _A(!e,e))}setHasWidgetFocus(e){this._eventDispatcher.emitOutgoingEvent(new Gye(!e,e))}onCompositionStart(){this._eventDispatcher.emitSingleViewEvent(new Mye)}onCompositionEnd(){this._eventDispatcher.emitSingleViewEvent(new Oye)}_captureStableViewport(){if(this._viewportStart.isValid&&this.viewLayout.getCurrentScrollTop()>0){const e=new A(this._viewportStart.viewLineNumber,this.getLineMinColumn(this._viewportStart.viewLineNumber)),t=this.coordinatesConverter.convertViewPositionToModelPosition(e);return new iq(t,this._viewportStart.startLineDelta)}return new iq(null,0)}_onConfigurationChanged(e,t){const i=this._captureStableViewport(),n=this._configuration.options,o=n.get(59),r=n.get(156),a=n.get(166),l=n.get(155),c=n.get(146);this._lines.setWrappingSettings(o,r,a.wrappingColumn,l,c)&&(e.emitViewEvent(new uL),e.emitViewEvent(new gL),e.emitViewEvent(new Hg(null)),this._cursor.onLineMappingChanged(e),this._decorations.onLineMappingChanged(),this.viewLayout.onFlushed(this.getLineCount(),this._getCustomLineHeights()),this._updateConfigurationViewLineCount.schedule()),t.hasChanged(104)&&(this._decorations.reset(),e.emitViewEvent(new Hg(null))),t.hasChanged(112)&&(this._decorations.reset(),e.emitViewEvent(new Hg(null))),e.emitViewEvent(new Pye(t)),this.viewLayout.onConfigurationChanged(t),i.recoverViewportStart(this.coordinatesConverter,this.viewLayout),B_.shouldRecreate(t)&&(this.cursorConfig=new B_(this.model.getLanguageId(),this.model.getOptions(),this._configuration,this.languageConfigurationService),this._cursor.updateConfiguration(this.cursorConfig))}_registerModelEvents(){this._register(this.model.onDidChangeContentOrInjectedText(i=>{try{const o=this._eventDispatcher.beginEmitViewEvents();let r=!1,a=!1;const l=i instanceof up?i.rawContentChangedEvent.changes:i.changes,c=i instanceof up?i.rawContentChangedEvent.versionId:null,d=this._lines.createLineBreaksComputer();for(const g of l)switch(g.changeType){case 4:{for(let f=0;f<g.detail.length;f++){const p=g.detail[f];let _=g.injectedTexts[f];_&&(_=_.filter(b=>!b.ownerId||b.ownerId===this._editorId)),d.addRequest(p,_,null)}break}case 2:{let f=null;g.injectedText&&(f=g.injectedText.filter(p=>!p.ownerId||p.ownerId===this._editorId)),d.addRequest(g.detail,f,null);break}}const h=d.finalize(),u=new ph(h);for(const g of l)switch(g.changeType){case 1:{this._lines.onModelFlushed(),o.emitViewEvent(new uL),this._decorations.reset(),this.viewLayout.onFlushed(this.getLineCount(),this._getCustomLineHeights()),r=!0;break}case 3:{const f=this._lines.onModelLinesDeleted(c,g.fromLineNumber,g.toLineNumber);f!==null&&(o.emitViewEvent(f),this.viewLayout.onLinesDeleted(f.fromLineNumber,f.toLineNumber)),r=!0;break}case 4:{const f=u.takeCount(g.detail.length),p=this._lines.onModelLinesInserted(c,g.fromLineNumber,g.toLineNumber,f);p!==null&&(o.emitViewEvent(p),this.viewLayout.onLinesInserted(p.fromLineNumber,p.toLineNumber)),r=!0;break}case 2:{const f=u.dequeue(),[p,_,b,w]=this._lines.onModelLineChanged(c,g.lineNumber,f);a=p,_&&o.emitViewEvent(_),b&&(o.emitViewEvent(b),this.viewLayout.onLinesInserted(b.fromLineNumber,b.toLineNumber)),w&&(o.emitViewEvent(w),this.viewLayout.onLinesDeleted(w.fromLineNumber,w.toLineNumber));break}case 5:break}c!==null&&this._lines.acceptVersionId(c),this.viewLayout.onHeightMaybeChanged(),!r&&a&&(o.emitViewEvent(new gL),o.emitViewEvent(new Hg(null)),this._cursor.onLineMappingChanged(o),this._decorations.onLineMappingChanged())}finally{this._eventDispatcher.endEmitViewEvents()}const n=this._viewportStart.isValid;if(this._viewportStart.invalidate(),this._configuration.setModelLineCount(this.model.getLineCount()),this._updateConfigurationViewLineCountNow(),!this._hasFocus&&this.model.getAttachedEditorCount()>=2&&n){const o=this.model._getTrackedRange(this._viewportStart.modelTrackedRange);if(o){const r=this.coordinatesConverter.convertModelPositionToViewPosition(o.getStartPosition()),a=this.viewLayout.getVerticalOffsetForLineNumber(r.lineNumber);this.viewLayout.setScrollPosition({scrollTop:a+this._viewportStart.startLineDelta},1)}}try{const o=this._eventDispatcher.beginEmitViewEvents();i instanceof up&&o.emitOutgoingEvent(new eSe(i.contentChangedEvent)),this._cursor.onModelContentChanged(o,i)}finally{this._eventDispatcher.endEmitViewEvents()}this._handleVisibleLinesChanged()})),this._configuration.options.get(5)&&this._register(this.model.onDidChangeLineHeight(i=>{const n=i.changes.filter(o=>o.ownerId===this._editorId||o.ownerId===0);if(this.viewLayout.changeSpecialLineHeights(o=>{for(const r of n){const{decorationId:a,lineNumber:l,lineHeight:c}=r,d=this.coordinatesConverter.convertModelRangeToViewRange(new k(l,1,l,this.model.getLineMaxColumn(l)));c!==null?o.insertOrChangeCustomLineHeight(a,d.startLineNumber,d.endLineNumber,c):o.removeCustomLineHeight(a)}}),n.length>0){const o=new Hee(n);this._eventDispatcher.emitOutgoingEvent(new nSe(o))}})),this._configuration.options.get(172)&&this._register(this.model.onDidChangeFont(i=>{const n=i.changes.filter(o=>o.ownerId===this._editorId||o.ownerId===0);if(n.length>0){const o=new Vee(n);this._eventDispatcher.emitOutgoingEvent(new sSe(o))}})),this._register(this.model.onDidChangeTokens(i=>{const n=[];for(let o=0,r=i.ranges.length;o<r;o++){const a=i.ranges[o],l=this.coordinatesConverter.convertModelPositionToViewPosition(new A(a.fromLineNumber,1)).lineNumber,c=this.coordinatesConverter.convertModelPositionToViewPosition(new A(a.toLineNumber,this.model.getLineMaxColumn(a.toLineNumber))).lineNumber;n[o]={fromLineNumber:l,toLineNumber:c}}this._eventDispatcher.emitSingleViewEvent(new zye(n)),this._eventDispatcher.emitOutgoingEvent(new iSe(i))})),this._register(this.model.onDidChangeLanguageConfiguration(i=>{this._eventDispatcher.emitSingleViewEvent(new Wye),this.cursorConfig=new B_(this.model.getLanguageId(),this.model.getOptions(),this._configuration,this.languageConfigurationService),this._cursor.updateConfiguration(this.cursorConfig),this._eventDispatcher.emitOutgoingEvent(new Jye(i))})),this._register(this.model.onDidChangeLanguage(i=>{this.cursorConfig=new B_(this.model.getLanguageId(),this.model.getOptions(),this._configuration,this.languageConfigurationService),this._cursor.updateConfiguration(this.cursorConfig),this._eventDispatcher.emitOutgoingEvent(new Qye(i))})),this._register(this.model.onDidChangeOptions(i=>{if(this._lines.setTabSize(this.model.getOptions().tabSize)){try{const n=this._eventDispatcher.beginEmitViewEvents();n.emitViewEvent(new uL),n.emitViewEvent(new gL),n.emitViewEvent(new Hg(null)),this._cursor.onLineMappingChanged(n),this._decorations.onLineMappingChanged(),this.viewLayout.onFlushed(this.getLineCount(),this._getCustomLineHeights())}finally{this._eventDispatcher.endEmitViewEvents()}this._updateConfigurationViewLineCount.schedule()}this.cursorConfig=new B_(this.model.getLanguageId(),this.model.getOptions(),this._configuration,this.languageConfigurationService),this._cursor.updateConfiguration(this.cursorConfig),this._eventDispatcher.emitOutgoingEvent(new tSe(i))})),this._register(this.model.onDidChangeDecorations(i=>{this._decorations.onModelDecorationsChanged(),this._eventDispatcher.emitSingleViewEvent(new Hg(i)),this._eventDispatcher.emitOutgoingEvent(new Xye(i))}))}getFontSizeAtPosition(e){if(!this._configuration.options.get(172))return null;const i=this.model.getFontDecorationsInRange(k.fromPositions(e),this._editorId);let n=this._configuration.options.get(59).fontSize+"px";for(const o of i)if(o.options.fontSize){n=o.options.fontSize;break}return n}setHiddenAreas(e,t,i){this.hiddenAreasModel.setHiddenAreas(t,e);const n=this.hiddenAreasModel.getMergedRanges();if(n===this.previousHiddenAreas&&!i)return;this.previousHiddenAreas=n;const o=this._captureStableViewport();let r=!1;try{const a=this._eventDispatcher.beginEmitViewEvents();r=this._lines.setHiddenAreas(n),r&&(a.emitViewEvent(new uL),a.emitViewEvent(new gL),a.emitViewEvent(new Hg(null)),this._cursor.onLineMappingChanged(a),this._decorations.onLineMappingChanged(),this.viewLayout.onFlushed(this.getLineCount(),this._getCustomLineHeights()),this.viewLayout.onHeightMaybeChanged());const l=o.viewportStartModelPosition?.lineNumber;l&&n.some(d=>d.startLineNumber<=l&&l<=d.endLineNumber)||o.recoverViewportStart(this.coordinatesConverter,this.viewLayout)}finally{this._eventDispatcher.endEmitViewEvents()}this._updateConfigurationViewLineCount.schedule(),r&&this._eventDispatcher.emitOutgoingEvent(new Yye)}getVisibleRangesPlusViewportAboveBelow(){const e=this._configuration.options.get(165),t=this._configuration.options.get(75),i=Math.max(20,Math.round(e.height/t)),n=this.viewLayout.getLinesViewportData(),o=Math.max(1,n.completelyVisibleStartLineNumber-i),r=Math.min(this.getLineCount(),n.completelyVisibleEndLineNumber+i);return this._toModelVisibleRanges(new k(o,this.getLineMinColumn(o),r,this.getLineMaxColumn(r)))}getVisibleRanges(){const e=this.getCompletelyVisibleViewRange();return this._toModelVisibleRanges(e)}getHiddenAreas(){return this._lines.getHiddenAreas()}_toModelVisibleRanges(e){const t=this.coordinatesConverter.convertViewRangeToModelRange(e),i=this._lines.getHiddenAreas();if(i.length===0)return[t];const n=[];let o=0,r=t.startLineNumber,a=t.startColumn;const l=t.endLineNumber,c=t.endColumn;for(let d=0,h=i.length;d<h;d++){const u=i[d].startLineNumber,g=i[d].endLineNumber;g<r||u>l||(r<u&&(n[o++]=new k(r,a,u-1,this.model.getLineMaxColumn(u-1))),r=g+1,a=1)}return(r<l||r===l&&a<c)&&(n[o++]=new k(r,a,l,c)),n}getCompletelyVisibleViewRange(){const e=this.viewLayout.getLinesViewportData(),t=e.completelyVisibleStartLineNumber,i=e.completelyVisibleEndLineNumber;return new k(t,this.getLineMinColumn(t),i,this.getLineMaxColumn(i))}getCompletelyVisibleViewRangeAtScrollTop(e){const t=this.viewLayout.getLinesViewportDataAtScrollTop(e),i=t.completelyVisibleStartLineNumber,n=t.completelyVisibleEndLineNumber;return new k(i,this.getLineMinColumn(i),n,this.getLineMaxColumn(n))}saveState(){const e=this.viewLayout.saveState(),t=e.scrollTop,i=this.viewLayout.getLineNumberAtVerticalOffset(t),n=this.coordinatesConverter.convertViewPositionToModelPosition(new A(i,this.getLineMinColumn(i))),o=this.viewLayout.getVerticalOffsetForLineNumber(i)-t;return{scrollLeft:e.scrollLeft,firstPosition:n,firstPositionDeltaTop:o}}reduceRestoreState(e){if(typeof e.firstPosition>"u")return this._reduceRestoreStateCompatibility(e);const t=this.model.validatePosition(e.firstPosition),i=this.coordinatesConverter.convertModelPositionToViewPosition(t),n=this.viewLayout.getVerticalOffsetForLineNumber(i.lineNumber)-e.firstPositionDeltaTop;return{scrollLeft:e.scrollLeft,scrollTop:n}}_reduceRestoreStateCompatibility(e){return{scrollLeft:e.scrollLeft,scrollTop:e.scrollTopWithoutViewZones}}getTabSize(){return this.model.getOptions().tabSize}getLineCount(){return this._lines.getViewLineCount()}setViewport(e,t,i){this._viewportStart.update(this,e)}getActiveIndentGuide(e,t,i){return this._lines.getActiveIndentGuide(e,t,i)}getLinesIndentGuides(e,t){return this._lines.getViewLinesIndentGuides(e,t)}getBracketGuidesInRangeByLine(e,t,i,n){return this._lines.getViewLinesBracketGuides(e,t,i,n)}getLineContent(e){return this._lines.getViewLineContent(e)}getLineLength(e){return this._lines.getViewLineLength(e)}getLineMinColumn(e){return this._lines.getViewLineMinColumn(e)}getLineMaxColumn(e){return this._lines.getViewLineMaxColumn(e)}getLineFirstNonWhitespaceColumn(e){const t=As(this.getLineContent(e));return t===-1?0:t+1}getLineLastNonWhitespaceColumn(e){const t=Ma(this.getLineContent(e));return t===-1?0:t+2}getMinimapDecorationsInRange(e){return this._decorations.getMinimapDecorationsInRange(e)}getDecorationsInViewport(e){return this._decorations.getDecorationsViewportData(e).decorations}getInjectedTextAt(e){return this._lines.getInjectedTextAt(e)}_getTextDirection(e,t){let i=0;for(const n of t){const o=n.range;if(o.startLineNumber>e||o.endLineNumber<e)continue;const r=n.options.textDirection;r===br.RTL?i++:r===br.LTR&&i--}return i>0?br.RTL:br.LTR}getTextDirection(e){const t=this._decorations.getDecorationsOnLine(e);return this._getTextDirection(e,t.decorations)}getViewportViewLineRenderingData(e,t){const i=this._decorations.getDecorationsViewportData(e),n=i.inlineDecorations[t-e.startLineNumber];return this._getViewLineRenderingData(t,n,i.hasVariableFonts,i.decorations)}getViewLineRenderingData(e){const t=this._decorations.getDecorationsOnLine(e);return this._getViewLineRenderingData(e,t.inlineDecorations[0],t.hasVariableFonts,t.decorations)}_getViewLineRenderingData(e,t,i,n){const o=this.model.mightContainRTL(),r=this.model.mightContainNonBasicASCII(),a=this.getTabSize(),l=this._lines.getViewLineData(e);return l.inlineDecorations&&(t=[...t,...l.inlineDecorations.map(c=>c.toInlineDecoration(e))]),new Er(l.minColumn,l.maxColumn,l.content,l.continuesWithWrappedLine,o,r,l.tokens,t,a,l.startVisibleColumn,this._getTextDirection(e,n),i)}getViewLineData(e){return this._lines.getViewLineData(e)}getMinimapLinesRenderingData(e,t,i){const n=this._lines.getViewLinesData(e,t,i);return new lve(this.getTabSize(),n)}getAllOverviewRulerDecorations(e){const t=this.model.getOverviewRulerDecorations(this._editorId,WE(this._configuration.options),HE(this._configuration.options)),i=new TSe;for(const n of t){const o=n.options,r=o.overviewRuler;if(!r)continue;const a=r.position;if(a===0)continue;const l=r.getColor(e.value),c=this.coordinatesConverter.getViewLineNumberOfModelPosition(n.range.startLineNumber,n.range.startColumn),d=this.coordinatesConverter.getViewLineNumberOfModelPosition(n.range.endLineNumber,n.range.endColumn);i.accept(l,o.zIndex,c,d,a)}return i.asArray}_invalidateDecorationsColorCache(){const e=this.model.getOverviewRulerDecorations();for(const t of e)t.options.overviewRuler?.invalidateCachedColor(),t.options.minimap?.invalidateCachedColor()}getValueInRange(e,t){const i=this.coordinatesConverter.convertViewRangeToModelRange(e);return this.model.getValueInRange(i,t)}getValueLengthInRange(e,t){const i=this.coordinatesConverter.convertViewRangeToModelRange(e);return this.model.getValueLengthInRange(i,t)}modifyPosition(e,t){const i=this.coordinatesConverter.convertViewPositionToModelPosition(e),n=this.model.modifyPosition(i,t);return this.coordinatesConverter.convertModelPositionToViewPosition(n)}deduceModelPositionRelativeToViewPosition(e,t,i){const n=this.coordinatesConverter.convertViewPositionToModelPosition(e);this.model.getEOL().length===2&&(t<0?t-=i:t+=i);const r=this.model.getOffsetAt(n)+t;return this.model.getPositionAt(r)}getPlainTextToCopy(e,t,i){const n=i?`\r
`:this.model.getEOL();e=e.slice(0),e.sort(k.compareRangesUsingStarts);let o=!1,r=!1;for(const l of e)l.isEmpty()?o=!0:r=!0;if(!r){if(!t)return"";const l=e.map(d=>d.startLineNumber);let c="";for(let d=0;d<l.length;d++)d>0&&l[d-1]===l[d]||(c+=this.model.getLineContent(l[d])+n);return c}if(o&&t){const l=[];let c=0;for(const d of e){const h=d.startLineNumber;d.isEmpty()?h!==c&&l.push(this.model.getLineContent(h)):l.push(this.model.getValueInRange(d,i?2:0)),c=h}return l.length===1?l[0]:l}const a=[];for(const l of e)l.isEmpty()||a.push(this.model.getValueInRange(l,i?2:0));return a.length===1?a[0]:a}getRichTextToCopy(e,t){const i=this.model.getLanguageId();if(i===xr||e.length!==1)return null;let n=e[0];if(n.isEmpty()){if(!t)return null;const d=n.startLineNumber;n=new k(d,this.model.getLineMinColumn(d),d,this.model.getLineMaxColumn(d))}const o=this._configuration.options.get(59),r=this._getColorMap(),l=/[:;\\\/<>]/.test(o.fontFamily)||o.fontFamily===_o.fontFamily;let c;return l?c=_o.fontFamily:(c=o.fontFamily,c=c.replace(/"/g,"'"),/[,']/.test(c)||/[+ ]/.test(c)&&(c=`'${c}'`),c=`${c}, ${_o.fontFamily}`),{mode:i,html:`<div style="color: ${r[1]};background-color: ${r[2]};font-family: ${c};font-weight: ${o.fontWeight};font-size: ${o.fontSize}px;line-height: ${o.lineHeight}px;white-space: pre;">`+this._getHTMLToCopy(n,r)+"</div>"}}_getHTMLToCopy(e,t){const i=e.startLineNumber,n=e.startColumn,o=e.endLineNumber,r=e.endColumn,a=this.getTabSize();let l="";for(let c=i;c<=o;c++){const d=this.model.tokenization.getLineTokens(c),h=d.getLineContent(),u=c===i?n-1:0,g=c===o?r-1:h.length;h===""?l+="<br>":l+=cSe(h,d.inflate(),t,u,g,a,Bn)}return l}_getColorMap(){const e=yi.getColorMap(),t=["#000000"];if(e)for(let i=1,n=e.length;i<n;i++)t[i]=q.Format.CSS.formatHex(e[i]);return t}getPrimaryCursorState(){return this._cursor.getPrimaryCursorState()}getLastAddedCursorIndex(){return this._cursor.getLastAddedCursorIndex()}getCursorStates(){return this._cursor.getCursorStates()}setCursorStates(e,t,i){return this._withViewEventsCollector(n=>this._cursor.setStates(n,e,t,i))}getCursorColumnSelectData(){return this._cursor.getCursorColumnSelectData()}getCursorAutoClosedCharacters(){return this._cursor.getAutoClosedCharacters()}setCursorColumnSelectData(e){this._cursor.setCursorColumnSelectData(e)}getPrevEditOperationType(){return this._cursor.getPrevEditOperationType()}setPrevEditOperationType(e){this._cursor.setPrevEditOperationType(e)}getSelection(){return this._cursor.getSelection()}getSelections(){return this._cursor.getSelections()}getPosition(){return this._cursor.getPrimaryCursorState().modelState.position}setSelections(e,t,i=0){this._withViewEventsCollector(n=>this._cursor.setSelections(n,e,t,i))}saveCursorState(){return this._cursor.saveState()}restoreCursorState(e){this._withViewEventsCollector(t=>this._cursor.restoreState(t,e))}_executeCursorEdit(e){if(this._cursor.context.cursorConfig.readOnly){this._eventDispatcher.emitOutgoingEvent(new Zye);return}this._withViewEventsCollector(e)}executeEdits(e,t,i,n){this._executeCursorEdit(o=>this._cursor.executeEdits(o,e,t,i,n))}startComposition(){this._executeCursorEdit(e=>this._cursor.startComposition(e))}endComposition(e){this._executeCursorEdit(t=>this._cursor.endComposition(t,e))}type(e,t){this._executeCursorEdit(i=>this._cursor.type(i,e,t))}compositionType(e,t,i,n,o){this._executeCursorEdit(r=>this._cursor.compositionType(r,e,t,i,n,o))}paste(e,t,i,n){this._executeCursorEdit(o=>this._cursor.paste(o,e,t,i,n))}cut(e){this._executeCursorEdit(t=>this._cursor.cut(t,e))}executeCommand(e,t){this._executeCursorEdit(i=>this._cursor.executeCommand(i,e,t))}executeCommands(e,t){this._executeCursorEdit(i=>this._cursor.executeCommands(i,e,t))}revealAllCursors(e,t,i=!1){this._withViewEventsCollector(n=>this._cursor.revealAll(n,e,i,0,t,0))}revealPrimaryCursor(e,t,i=!1){this._withViewEventsCollector(n=>this._cursor.revealPrimary(n,e,i,0,t,0))}revealTopMostCursor(e){const t=this._cursor.getTopMostViewPosition(),i=new k(t.lineNumber,t.column,t.lineNumber,t.column);this._withViewEventsCollector(n=>n.emitViewEvent(new z1(e,!1,i,null,0,!0,0)))}revealBottomMostCursor(e){const t=this._cursor.getBottomMostViewPosition(),i=new k(t.lineNumber,t.column,t.lineNumber,t.column);this._withViewEventsCollector(n=>n.emitViewEvent(new z1(e,!1,i,null,0,!0,0)))}revealRange(e,t,i,n,o){this._withViewEventsCollector(r=>r.emitViewEvent(new z1(e,!1,i,null,n,t,o)))}changeWhitespace(e){this.viewLayout.changeWhitespace(e)&&(this._eventDispatcher.emitSingleViewEvent(new Uye),this._eventDispatcher.emitOutgoingEvent(new Kye))}_withViewEventsCollector(e){return this._transactionalTarget.batchChanges(()=>{try{const t=this._eventDispatcher.beginEmitViewEvents();return e(t)}finally{this._eventDispatcher.endEmitViewEvents()}})}batchEvents(e){this._withViewEventsCollector(()=>{e()})}normalizePosition(e,t){return this._lines.normalizePosition(e,t)}getLineIndentColumn(e){return this._lines.getLineIndentColumn(e)}};class b7{static create(e){const t=e._setTrackedRange(null,new k(1,1,1,1),1);return new b7(e,1,!1,t,0)}get viewLineNumber(){return this._viewLineNumber}get isValid(){return this._isValid}get modelTrackedRange(){return this._modelTrackedRange}get startLineDelta(){return this._startLineDelta}constructor(e,t,i,n,o){this._model=e,this._viewLineNumber=t,this._isValid=i,this._modelTrackedRange=n,this._startLineDelta=o}dispose(){this._model._setTrackedRange(this._modelTrackedRange,null,1)}update(e,t){const i=e.coordinatesConverter.convertViewPositionToModelPosition(new A(t,e.getLineMinColumn(t))),n=e.model._setTrackedRange(this._modelTrackedRange,new k(i.lineNumber,i.column,i.lineNumber,i.column),1),o=e.viewLayout.getVerticalOffsetForLineNumber(t),r=e.viewLayout.getCurrentScrollTop();this._viewLineNumber=t,this._isValid=!0,this._modelTrackedRange=n,this._startLineDelta=r-o}invalidate(){this._isValid=!1}}class TSe{constructor(){this._asMap=Object.create(null),this.asArray=[]}accept(e,t,i,n,o){const r=this._asMap[e];if(r){const a=r.data,l=a[a.length-3],c=a[a.length-1];if(l===o&&c+1>=i){n>c&&(a[a.length-1]=n);return}a.push(o,i,n)}else{const a=new vS(e,t,[o,i,n]);this._asMap[e]=a,this.asArray.push(a)}}}class NSe{constructor(){this.hiddenAreas=new Map,this.shouldRecompute=!1,this.ranges=[]}setHiddenAreas(e,t){const i=this.hiddenAreas.get(e);i&&tq(i,t)||(this.hiddenAreas.set(e,t),this.shouldRecompute=!0)}getMergedRanges(){if(!this.shouldRecompute)return this.ranges;this.shouldRecompute=!1;const e=Array.from(this.hiddenAreas.values()).reduce((t,i)=>RSe(t,i),[]);return tq(this.ranges,e)?this.ranges:(this.ranges=e,this.ranges)}}function RSe(s,e){const t=[];let i=0,n=0;for(;i<s.length&&n<e.length;){const o=s[i],r=e[n];if(o.endLineNumber<r.startLineNumber-1)t.push(s[i++]);else if(r.endLineNumber<o.startLineNumber-1)t.push(e[n++]);else{const a=Math.min(o.startLineNumber,r.startLineNumber),l=Math.max(o.endLineNumber,r.endLineNumber);t.push(new k(a,1,l,1)),i++,n++}}for(;i<s.length;)t.push(s[i++]);for(;n<e.length;)t.push(e[n++]);return t}function tq(s,e){if(s.length!==e.length)return!1;for(let t=0;t<s.length;t++)if(!s[t].equalsRange(e[t]))return!1;return!0}class iq{constructor(e,t){this.viewportStartModelPosition=e,this.startLineDelta=t}recoverViewportStart(e,t){if(!this.viewportStartModelPosition)return;const i=e.convertModelPositionToViewPosition(this.viewportStartModelPosition),n=t.getVerticalOffsetForLineNumber(i.lineNumber);t.setScrollPosition({scrollTop:n+this.startLineDelta},1)}}class rC{constructor(...e){this._entries=new Map;for(const[t,i]of e)this.set(t,i)}set(e,t){const i=this._entries.get(e);return this._entries.set(e,t),i}get(e){return this._entries.get(e)}}var ASe=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},nd=function(s,e){return function(t,i){e(t,i,s)}},of,xp;let Am=(xp=class extends F{get isSimpleWidget(){return this._configuration.isSimpleWidget}get contextMenuId(){return this._configuration.contextMenuId}get contextKeyService(){return this._contextKeyService}constructor(e,t,i,n,o,r,a,l,c,d,h,u){super(),this.languageConfigurationService=h,this._deliveryQueue=vde(),this._contributions=this._register(new BCe),this._onDidDispose=this._register(new M),this.onDidDispose=this._onDidDispose.event,this._onDidChangeModelContent=this._register(new M({deliveryQueue:this._deliveryQueue})),this.onDidChangeModelContent=this._onDidChangeModelContent.event,this._onDidChangeModelLanguage=this._register(new M({deliveryQueue:this._deliveryQueue})),this.onDidChangeModelLanguage=this._onDidChangeModelLanguage.event,this._onDidChangeModelLanguageConfiguration=this._register(new M({deliveryQueue:this._deliveryQueue})),this.onDidChangeModelLanguageConfiguration=this._onDidChangeModelLanguageConfiguration.event,this._onDidChangeModelOptions=this._register(new M({deliveryQueue:this._deliveryQueue})),this.onDidChangeModelOptions=this._onDidChangeModelOptions.event,this._onDidChangeModelDecorations=this._register(new M({deliveryQueue:this._deliveryQueue})),this.onDidChangeModelDecorations=this._onDidChangeModelDecorations.event,this._onDidChangeLineHeight=this._register(new M({deliveryQueue:this._deliveryQueue})),this.onDidChangeLineHeight=this._onDidChangeLineHeight.event,this._onDidChangeFont=this._register(new M({deliveryQueue:this._deliveryQueue})),this.onDidChangeFont=this._onDidChangeFont.event,this._onDidChangeModelTokens=this._register(new M({deliveryQueue:this._deliveryQueue})),this.onDidChangeModelTokens=this._onDidChangeModelTokens.event,this._onDidChangeConfiguration=this._register(new M({deliveryQueue:this._deliveryQueue})),this.onDidChangeConfiguration=this._onDidChangeConfiguration.event,this._onWillChangeModel=this._register(new M({deliveryQueue:this._deliveryQueue})),this.onWillChangeModel=this._onWillChangeModel.event,this._onDidChangeModel=this._register(new M({deliveryQueue:this._deliveryQueue})),this.onDidChangeModel=this._onDidChangeModel.event,this._onDidChangeCursorPosition=this._register(new M({deliveryQueue:this._deliveryQueue})),this.onDidChangeCursorPosition=this._onDidChangeCursorPosition.event,this._onDidChangeCursorSelection=this._register(new M({deliveryQueue:this._deliveryQueue})),this.onDidChangeCursorSelection=this._onDidChangeCursorSelection.event,this._onDidAttemptReadOnlyEdit=this._register(new ys(this._contributions,this._deliveryQueue)),this.onDidAttemptReadOnlyEdit=this._onDidAttemptReadOnlyEdit.event,this._onDidLayoutChange=this._register(new M({deliveryQueue:this._deliveryQueue})),this.onDidLayoutChange=this._onDidLayoutChange.event,this._editorTextFocus=this._register(new nq({deliveryQueue:this._deliveryQueue})),this.onDidFocusEditorText=this._editorTextFocus.onDidChangeToTrue,this.onDidBlurEditorText=this._editorTextFocus.onDidChangeToFalse,this._editorWidgetFocus=this._register(new nq({deliveryQueue:this._deliveryQueue})),this.onDidFocusEditorWidget=this._editorWidgetFocus.onDidChangeToTrue,this.onDidBlurEditorWidget=this._editorWidgetFocus.onDidChangeToFalse,this._onWillType=this._register(new ys(this._contributions,this._deliveryQueue)),this.onWillType=this._onWillType.event,this._onDidType=this._register(new ys(this._contributions,this._deliveryQueue)),this.onDidType=this._onDidType.event,this._onDidCompositionStart=this._register(new ys(this._contributions,this._deliveryQueue)),this.onDidCompositionStart=this._onDidCompositionStart.event,this._onDidCompositionEnd=this._register(new ys(this._contributions,this._deliveryQueue)),this.onDidCompositionEnd=this._onDidCompositionEnd.event,this._onDidPaste=this._register(new ys(this._contributions,this._deliveryQueue)),this.onDidPaste=this._onDidPaste.event,this._onMouseUp=this._register(new ys(this._contributions,this._deliveryQueue)),this.onMouseUp=this._onMouseUp.event,this._onMouseDown=this._register(new ys(this._contributions,this._deliveryQueue)),this.onMouseDown=this._onMouseDown.event,this._onMouseDrag=this._register(new ys(this._contributions,this._deliveryQueue)),this.onMouseDrag=this._onMouseDrag.event,this._onMouseDrop=this._register(new ys(this._contributions,this._deliveryQueue)),this.onMouseDrop=this._onMouseDrop.event,this._onMouseDropCanceled=this._register(new ys(this._contributions,this._deliveryQueue)),this.onMouseDropCanceled=this._onMouseDropCanceled.event,this._onDropIntoEditor=this._register(new ys(this._contributions,this._deliveryQueue)),this.onDropIntoEditor=this._onDropIntoEditor.event,this._onContextMenu=this._register(new ys(this._contributions,this._deliveryQueue)),this.onContextMenu=this._onContextMenu.event,this._onMouseMove=this._register(new ys(this._contributions,this._deliveryQueue)),this.onMouseMove=this._onMouseMove.event,this._onMouseLeave=this._register(new ys(this._contributions,this._deliveryQueue)),this.onMouseLeave=this._onMouseLeave.event,this._onMouseWheel=this._register(new ys(this._contributions,this._deliveryQueue)),this.onMouseWheel=this._onMouseWheel.event,this._onKeyUp=this._register(new ys(this._contributions,this._deliveryQueue)),this.onKeyUp=this._onKeyUp.event,this._onKeyDown=this._register(new ys(this._contributions,this._deliveryQueue)),this.onKeyDown=this._onKeyDown.event,this._onDidContentSizeChange=this._register(new M({deliveryQueue:this._deliveryQueue})),this.onDidContentSizeChange=this._onDidContentSizeChange.event,this._onDidScrollChange=this._register(new M({deliveryQueue:this._deliveryQueue})),this.onDidScrollChange=this._onDidScrollChange.event,this._onDidChangeViewZones=this._register(new M({deliveryQueue:this._deliveryQueue})),this.onDidChangeViewZones=this._onDidChangeViewZones.event,this._onDidChangeHiddenAreas=this._register(new M({deliveryQueue:this._deliveryQueue})),this.onDidChangeHiddenAreas=this._onDidChangeHiddenAreas.event,this._updateCounter=0,this._onWillTriggerEditorOperationEvent=this._register(new M),this._onBeginUpdate=this._register(new M),this.onBeginUpdate=this._onBeginUpdate.event,this._onEndUpdate=this._register(new M),this.onEndUpdate=this._onEndUpdate.event,this._onBeforeExecuteEdit=this._register(new M),this.onBeforeExecuteEdit=this._onBeforeExecuteEdit.event,this._actions=new Map,this._bannerDomNode=null,this._dropIntoEditorDecorations=this.createDecorationsCollection(),this.inComposition=!1,o.willCreateCodeEditor();const g={...t};if(this._domElement=e,this._overflowWidgetsDomNode=g.overflowWidgetsDomNode,delete g.overflowWidgetsDomNode,this._id=++MSe,this._decorationTypeKeysToIds={},this._decorationTypeSubtypes={},this._telemetryData=i.telemetryData,this._configuration=this._register(this._createConfiguration(i.isSimpleWidget||!1,i.contextMenuId??(i.isSimpleWidget?fe.SimpleEditorContext:fe.EditorContext),g,d)),this._register(this._configuration.onDidChange(_=>{this._onDidChangeConfiguration.fire(_);const b=this._configuration.options;if(_.hasChanged(165)){const w=b.get(165);this._onDidLayoutChange.fire(w)}})),this._contextKeyService=this._register(a.createScoped(this._domElement)),i.contextKeyValues)for(const[_,b]of Object.entries(i.contextKeyValues))this._contextKeyService.createKey(_,b);this._notificationService=c,this._codeEditorService=o,this._commandService=r,this._themeService=l,this._register(new PSe(this,this._contextKeyService)),this._register(new FSe(this,this._contextKeyService,u)),this._instantiationService=this._register(n.createChild(new rC([Ae,this._contextKeyService]))),this._modelData=null,this._contentWidgets={},this._overlayWidgets={},this._glyphMarginWidgets={};let f;Array.isArray(i.contributions)?f=i.contributions:f=lw.getEditorContributions(),this._contributions.initialize(this,f,this._instantiationService);for(const _ of lw.getEditorActions()){if(this._actions.has(_.id)){Oe(new Error(`Cannot have two actions with the same id ${_.id}`));continue}const b=new zee(_.id,_.label,_.alias,_.metadata,_.precondition??void 0,w=>this._instantiationService.invokeFunction(C=>Promise.resolve(_.runEditorCommand(C,this,w))),this._contextKeyService);this._actions.set(b.id,b)}const p=()=>!this._configuration.options.get(104)&&this._configuration.options.get(43).enabled;this._register(new mue(this._domElement,{onDragOver:_=>{if(!p())return;const b=this.getTargetAtClientPoint(_.clientX,_.clientY);b?.position&&this.showDropIndicatorAt(b.position)},onDrop:async _=>{if(!p()||(this.removeDropIndicator(),!_.dataTransfer))return;const b=this.getTargetAtClientPoint(_.clientX,_.clientY);b?.position&&this._onDropIntoEditor.fire({position:b.position,event:_})},onDragLeave:()=>{this.removeDropIndicator()},onDragEnd:()=>{this.removeDropIndicator()}})),this._codeEditorService.addCodeEditor(this)}writeScreenReaderContent(e){this._modelData?.view.writeScreenReaderContent(e)}_createConfiguration(e,t,i,n){return new qP(e,t,i,this._domElement,n)}getId(){return this.getEditorType()+":"+this._id}getEditorType(){return sx.ICodeEditor}dispose(){this._codeEditorService.removeCodeEditor(this),this._actions.clear(),this._contentWidgets={},this._overlayWidgets={},this._removeDecorationTypes(),this._postDetachModelCleanup(this._detachModel()),this._onDidDispose.fire(),super.dispose()}invokeWithinContext(e){return this._instantiationService.invokeFunction(e)}updateOptions(e){this._configuration.updateOptions(e||{})}getOptions(){return this._configuration.options}getOption(e){return this._configuration.options.get(e)}getRawOptions(){return this._configuration.getRawOptions()}getOverflowWidgetsDomNode(){return this._overflowWidgetsDomNode}getConfiguredWordAtPosition(e){return this._modelData?yt.getWordAtPosition(this._modelData.model,this._configuration.options.get(148),this._configuration.options.get(147),e):null}getValue(e=null){if(!this._modelData)return"";const t=!!(e&&e.preserveBOM);let i=0;return e&&e.lineEnding&&e.lineEnding===`
`?i=1:e&&e.lineEnding&&e.lineEnding===`\r
`&&(i=2),this._modelData.model.getValue(i,t)}setValue(e){try{if(this._beginUpdate(),!this._modelData)return;this._modelData.model.setValue(e)}finally{this._endUpdate()}}getModel(){return this._modelData?this._modelData.model:null}setModel(e=null){try{this._beginUpdate();const t=e;if(this._modelData===null&&t===null||this._modelData&&this._modelData.model===t)return;const i={oldModelUrl:this._modelData?.model.uri||null,newModelUrl:t?.uri||null};this._onWillChangeModel.fire(i);const n=this.hasTextFocus(),o=this._detachModel();this._attachModel(t),this.hasModel()?n&&this.focus():(this._editorTextFocus.setValue(!1),this._editorWidgetFocus.setValue(!1)),this._removeDecorationTypes(),this._onDidChangeModel.fire(i),this._postDetachModelCleanup(o),this._contributionsDisposable=this._contributions.onAfterModelAttached()}finally{this._endUpdate()}}_removeDecorationTypes(){if(this._decorationTypeKeysToIds={},this._decorationTypeSubtypes){for(const e in this._decorationTypeSubtypes){const t=this._decorationTypeSubtypes[e];for(const i in t)this._removeDecorationType(e+"-"+i)}this._decorationTypeSubtypes={}}}getVisibleRanges(){return this._modelData?this._modelData.viewModel.getVisibleRanges():[]}getVisibleRangesPlusViewportAboveBelow(){return this._modelData?this._modelData.viewModel.getVisibleRangesPlusViewportAboveBelow():[]}getWhitespaces(){return this._modelData?this._modelData.viewModel.viewLayout.getWhitespaces():[]}static _getVerticalOffsetAfterPosition(e,t,i,n){const o=e.model.validatePosition({lineNumber:t,column:i}),r=e.viewModel.coordinatesConverter.convertModelPositionToViewPosition(o);return e.viewModel.viewLayout.getVerticalOffsetAfterLineNumber(r.lineNumber,n)}getTopForLineNumber(e,t=!1){return this._modelData?of._getVerticalOffsetForPosition(this._modelData,e,1,t):-1}getTopForPosition(e,t){return this._modelData?of._getVerticalOffsetForPosition(this._modelData,e,t,!1):-1}static _getVerticalOffsetForPosition(e,t,i,n=!1){const o=e.model.validatePosition({lineNumber:t,column:i}),r=e.viewModel.coordinatesConverter.convertModelPositionToViewPosition(o);return e.viewModel.viewLayout.getVerticalOffsetForLineNumber(r.lineNumber,n)}getBottomForLineNumber(e,t=!1){if(!this._modelData)return-1;const i=this._modelData.model.getLineMaxColumn(e);return of._getVerticalOffsetAfterPosition(this._modelData,e,i,t)}getLineHeightForPosition(e){if(!this._modelData)return-1;const t=this._modelData.viewModel,i=t.coordinatesConverter,n=A.lift(e);if(i.modelPositionIsVisible(n)){const o=i.convertModelPositionToViewPosition(n);return t.viewLayout.getLineHeightForLineNumber(o.lineNumber)}return 0}setHiddenAreas(e,t,i){this._modelData?.viewModel.setHiddenAreas(e.map(n=>k.lift(n)),t,i)}getVisibleColumnFromPosition(e){if(!this._modelData)return e.column;const t=this._modelData.model.validatePosition(e),i=this._modelData.model.getOptions().tabSize;return pi.visibleColumnFromColumn(this._modelData.model.getLineContent(t.lineNumber),t.column,i)+1}getPosition(){return this._modelData?this._modelData.viewModel.getPosition():null}setPosition(e,t="api"){if(this._modelData){if(!A.isIPosition(e))throw new Error("Invalid arguments");this._modelData.viewModel.setSelections(t,[{selectionStartLineNumber:e.lineNumber,selectionStartColumn:e.column,positionLineNumber:e.lineNumber,positionColumn:e.column}])}}_sendRevealRange(e,t,i,n){if(!this._modelData)return;if(!k.isIRange(e))throw new Error("Invalid arguments");const o=this._modelData.model.validateRange(e),r=this._modelData.viewModel.coordinatesConverter.convertModelRangeToViewRange(o);this._modelData.viewModel.revealRange("api",i,r,t,n)}revealLine(e,t=0){this._revealLine(e,0,t)}revealLineInCenter(e,t=0){this._revealLine(e,1,t)}revealLineInCenterIfOutsideViewport(e,t=0){this._revealLine(e,2,t)}revealLineNearTop(e,t=0){this._revealLine(e,5,t)}_revealLine(e,t,i){if(typeof e!="number")throw new Error("Invalid arguments");this._sendRevealRange(new k(e,1,e,1),t,!1,i)}revealPosition(e,t=0){this._revealPosition(e,0,!0,t)}revealPositionInCenter(e,t=0){this._revealPosition(e,1,!0,t)}revealPositionInCenterIfOutsideViewport(e,t=0){this._revealPosition(e,2,!0,t)}revealPositionNearTop(e,t=0){this._revealPosition(e,5,!0,t)}_revealPosition(e,t,i,n){if(!A.isIPosition(e))throw new Error("Invalid arguments");this._sendRevealRange(new k(e.lineNumber,e.column,e.lineNumber,e.column),t,i,n)}getSelection(){return this._modelData?this._modelData.viewModel.getSelection():null}getSelections(){return this._modelData?this._modelData.viewModel.getSelections():null}setSelection(e,t="api"){const i=he.isISelection(e),n=k.isIRange(e);if(!i&&!n)throw new Error("Invalid arguments");if(i)this._setSelectionImpl(e,t);else if(n){const o={selectionStartLineNumber:e.startLineNumber,selectionStartColumn:e.startColumn,positionLineNumber:e.endLineNumber,positionColumn:e.endColumn};this._setSelectionImpl(o,t)}}_setSelectionImpl(e,t){if(!this._modelData)return;const i=new he(e.selectionStartLineNumber,e.selectionStartColumn,e.positionLineNumber,e.positionColumn);this._modelData.viewModel.setSelections(t,[i])}revealLines(e,t,i=0){this._revealLines(e,t,0,i)}revealLinesInCenter(e,t,i=0){this._revealLines(e,t,1,i)}revealLinesInCenterIfOutsideViewport(e,t,i=0){this._revealLines(e,t,2,i)}revealLinesNearTop(e,t,i=0){this._revealLines(e,t,5,i)}_revealLines(e,t,i,n){if(typeof e!="number"||typeof t!="number")throw new Error("Invalid arguments");this._sendRevealRange(new k(e,1,t,1),i,!1,n)}revealRange(e,t=0,i=!1,n=!0){this._revealRange(e,i?1:0,n,t)}revealRangeInCenter(e,t=0){this._revealRange(e,1,!0,t)}revealRangeInCenterIfOutsideViewport(e,t=0){this._revealRange(e,2,!0,t)}revealRangeNearTop(e,t=0){this._revealRange(e,5,!0,t)}revealRangeNearTopIfOutsideViewport(e,t=0){this._revealRange(e,6,!0,t)}revealRangeAtTop(e,t=0){this._revealRange(e,3,!0,t)}_revealRange(e,t,i,n){if(!k.isIRange(e))throw new Error("Invalid arguments");this._sendRevealRange(k.lift(e),t,i,n)}setSelections(e,t="api",i=0){if(this._modelData){if(!e||e.length===0)throw new Error("Invalid arguments");for(let n=0,o=e.length;n<o;n++)if(!he.isISelection(e[n]))throw new Error("Invalid arguments");this._modelData.viewModel.setSelections(t,e,i)}}getContentWidth(){return this._modelData?this._modelData.viewModel.viewLayout.getContentWidth():-1}getScrollWidth(){return this._modelData?this._modelData.viewModel.viewLayout.getScrollWidth():-1}getScrollLeft(){return this._modelData?this._modelData.viewModel.viewLayout.getCurrentScrollLeft():-1}getContentHeight(){return this._modelData?this._modelData.viewModel.viewLayout.getContentHeight():-1}getScrollHeight(){return this._modelData?this._modelData.viewModel.viewLayout.getScrollHeight():-1}getScrollTop(){return this._modelData?this._modelData.viewModel.viewLayout.getCurrentScrollTop():-1}setScrollLeft(e,t=1){if(this._modelData){if(typeof e!="number")throw new Error("Invalid arguments");this._modelData.viewModel.viewLayout.setScrollPosition({scrollLeft:e},t)}}setScrollTop(e,t=1){if(this._modelData){if(typeof e!="number")throw new Error("Invalid arguments");this._modelData.viewModel.viewLayout.setScrollPosition({scrollTop:e},t)}}setScrollPosition(e,t=1){this._modelData&&this._modelData.viewModel.viewLayout.setScrollPosition(e,t)}hasPendingScrollAnimation(){return this._modelData?this._modelData.viewModel.viewLayout.hasPendingScrollAnimation():!1}saveViewState(){if(!this._modelData)return null;const e=this._contributions.saveViewState(),t=this._modelData.viewModel.saveCursorState(),i=this._modelData.viewModel.saveState();return{cursorState:t,viewState:i,contributionsState:e}}restoreViewState(e){if(!this._modelData||!this._modelData.hasRealView)return;const t=e;if(t&&t.cursorState&&t.viewState){const i=t.cursorState;Array.isArray(i)?i.length>0&&this._modelData.viewModel.restoreCursorState(i):this._modelData.viewModel.restoreCursorState([i]),this._contributions.restoreViewState(t.contributionsState||{});const n=this._modelData.viewModel.reduceRestoreState(t.viewState);this._modelData.view.restoreState(n)}}handleInitialized(){this._getViewModel()?.visibleLinesStabilized()}getContribution(e){return this._contributions.get(e)}getActions(){return Array.from(this._actions.values())}getSupportedActions(){let e=this.getActions();return e=e.filter(t=>t.isSupported()),e}getAction(e){return this._actions.get(e)||null}trigger(e,t,i){i=i||{};try{switch(this._onWillTriggerEditorOperationEvent.fire({source:e,handlerId:t,payload:i}),this._beginUpdate(),t){case"compositionStart":this._startComposition();return;case"compositionEnd":this._endComposition(e);return;case"type":{const o=i;this._type(e,o.text||"");return}case"replacePreviousChar":{const o=i;this._compositionType(e,o.text||"",o.replaceCharCnt||0,0,0);return}case"compositionType":{const o=i;this._compositionType(e,o.text||"",o.replacePrevCharCnt||0,o.replaceNextCharCnt||0,o.positionDelta||0);return}case"paste":{const o=i;this._paste(e,o.text||"",o.pasteOnNewLine||!1,o.multicursorText||null,o.mode||null,o.clipboardEvent);return}case"cut":this._cut(e);return}const n=this.getAction(t);if(n){Promise.resolve(n.run(i)).then(void 0,Oe);return}if(!this._modelData||this._triggerEditorCommand(e,t,i))return;this._triggerCommand(t,i)}finally{this._endUpdate()}}_triggerCommand(e,t){this._commandService.executeCommand(e,t)}_startComposition(){this._modelData&&(this.inComposition=!0,this._modelData.viewModel.startComposition(),this._onDidCompositionStart.fire())}_endComposition(e){this._modelData&&(this.inComposition=!1,this._modelData.viewModel.endComposition(e),this._onDidCompositionEnd.fire())}_type(e,t){!this._modelData||t.length===0||(e==="keyboard"&&this._onWillType.fire(t),this._modelData.viewModel.type(t,e),e==="keyboard"&&this._onDidType.fire(t))}_compositionType(e,t,i,n,o){this._modelData&&this._modelData.viewModel.compositionType(t,i,n,o,e)}_paste(e,t,i,n,o,r){if(!this._modelData)return;const a=this._modelData.viewModel,l=a.getSelection().getStartPosition();a.paste(t,i,n,e);const c=a.getSelection().getStartPosition();e==="keyboard"&&this._onDidPaste.fire({clipboardEvent:r,range:new k(l.lineNumber,l.column,c.lineNumber,c.column),languageId:o})}_cut(e){this._modelData&&this._modelData.viewModel.cut(e)}_triggerEditorCommand(e,t,i){const n=lw.getEditorCommand(t);return n?(i=i||{},rn(i)&&(i.source=e),this._instantiationService.invokeFunction(o=>{Promise.resolve(n.runEditorCommand(o,this,i)).then(void 0,Oe)}),!0):!1}_getViewModel(){return this._modelData?this._modelData.viewModel:null}pushUndoStop(){return!this._modelData||this._configuration.options.get(104)?!1:(this._modelData.model.pushStackElement(),!0)}popUndoStop(){return!this._modelData||this._configuration.options.get(104)?!1:(this._modelData.model.popStackElement(),!0)}edit(e,t){return this.executeEdits(t,e.replacements.map(i=>({range:i.range,text:i.text})),void 0)}executeEdits(e,t,i){if(!this._modelData||this._configuration.options.get(104))return!1;let n;i?Array.isArray(i)?n=()=>i:n=i:n=()=>null;let o,r;return e instanceof Xee?(r=e,o=e.metadata.source):(r=cs.unknown({name:o}),o=e),this._onBeforeExecuteEdit.fire({source:o??void 0}),this._modelData.viewModel.executeEdits(o,t,n,r),!0}executeCommand(e,t){this._modelData&&this._modelData.viewModel.executeCommand(t,e)}executeCommands(e,t){this._modelData&&this._modelData.viewModel.executeCommands(t,e)}createDecorationsCollection(e){return new BSe(this,e)}changeDecorations(e){return this._modelData?this._modelData.model.changeDecorations(e,this._id):null}getLineDecorations(e){if(!this._modelData)return null;const t=this._configuration.options;return this._modelData.model.getLineDecorations(e,this._id,WE(t),HE(t))}getDecorationsInRange(e){if(!this._modelData)return null;const t=this._configuration.options;return this._modelData.model.getDecorationsInRange(e,this._id,WE(t),HE(t))}getFontSizeAtPosition(e){return this._modelData?this._modelData.viewModel.getFontSizeAtPosition(e):null}deltaDecorations(e,t){return this._modelData?e.length===0&&t.length===0?e:this._modelData.model.deltaDecorations(e,t,this._id):[]}removeDecorations(e){!this._modelData||e.length===0||this._modelData.model.changeDecorations(t=>{t.deltaDecorations(e,[])})}removeDecorationsByType(e){const t=this._decorationTypeKeysToIds[e];t&&this.changeDecorations(i=>i.deltaDecorations(t,[])),this._decorationTypeKeysToIds.hasOwnProperty(e)&&delete this._decorationTypeKeysToIds[e],this._decorationTypeSubtypes.hasOwnProperty(e)&&delete this._decorationTypeSubtypes[e]}getLayoutInfo(){return this._configuration.options.get(165)}createOverviewRuler(e){return!this._modelData||!this._modelData.hasRealView?null:this._modelData.view.createOverviewRuler(e)}getContainerDomNode(){return this._domElement}getDomNode(){return!this._modelData||!this._modelData.hasRealView?null:this._modelData.view.domNode.domNode}delegateVerticalScrollbarPointerDown(e){!this._modelData||!this._modelData.hasRealView||this._modelData.view.delegateVerticalScrollbarPointerDown(e)}delegateScrollFromMouseWheelEvent(e){!this._modelData||!this._modelData.hasRealView||this._modelData.view.delegateScrollFromMouseWheelEvent(e)}layout(e,t=!1){this._configuration.observeContainer(e),t||this.render()}focus(){!this._modelData||!this._modelData.hasRealView||this._modelData.view.focus()}hasTextFocus(){return!this._modelData||!this._modelData.hasRealView?!1:this._modelData.view.isFocused()}hasWidgetFocus(){return!this._modelData||!this._modelData.hasRealView?!1:this._modelData.view.isWidgetFocused()}addContentWidget(e){const t={widget:e,position:e.getPosition()};this._contentWidgets.hasOwnProperty(e.getId())&&console.warn("Overwriting a content widget with the same id:"+e.getId()),this._contentWidgets[e.getId()]=t,this._modelData&&this._modelData.hasRealView&&this._modelData.view.addContentWidget(t)}layoutContentWidget(e){const t=e.getId();if(this._contentWidgets.hasOwnProperty(t)){const i=this._contentWidgets[t];i.position=e.getPosition(),this._modelData&&this._modelData.hasRealView&&this._modelData.view.layoutContentWidget(i)}}removeContentWidget(e){const t=e.getId();if(this._contentWidgets.hasOwnProperty(t)){const i=this._contentWidgets[t];delete this._contentWidgets[t],this._modelData&&this._modelData.hasRealView&&this._modelData.view.removeContentWidget(i)}}addOverlayWidget(e){const t={widget:e,position:e.getPosition()};this._overlayWidgets.hasOwnProperty(e.getId())&&console.warn("Overwriting an overlay widget with the same id."),this._overlayWidgets[e.getId()]=t,this._modelData&&this._modelData.hasRealView&&this._modelData.view.addOverlayWidget(t)}layoutOverlayWidget(e){const t=e.getId();if(this._overlayWidgets.hasOwnProperty(t)){const i=this._overlayWidgets[t];i.position=e.getPosition(),this._modelData&&this._modelData.hasRealView&&this._modelData.view.layoutOverlayWidget(i)}}removeOverlayWidget(e){const t=e.getId();if(this._overlayWidgets.hasOwnProperty(t)){const i=this._overlayWidgets[t];delete this._overlayWidgets[t],this._modelData&&this._modelData.hasRealView&&this._modelData.view.removeOverlayWidget(i)}}addGlyphMarginWidget(e){const t={widget:e,position:e.getPosition()};this._glyphMarginWidgets.hasOwnProperty(e.getId())&&console.warn("Overwriting a glyph margin widget with the same id."),this._glyphMarginWidgets[e.getId()]=t,this._modelData&&this._modelData.hasRealView&&this._modelData.view.addGlyphMarginWidget(t)}layoutGlyphMarginWidget(e){const t=e.getId();if(this._glyphMarginWidgets.hasOwnProperty(t)){const i=this._glyphMarginWidgets[t];i.position=e.getPosition(),this._modelData&&this._modelData.hasRealView&&this._modelData.view.layoutGlyphMarginWidget(i)}}removeGlyphMarginWidget(e){const t=e.getId();if(this._glyphMarginWidgets.hasOwnProperty(t)){const i=this._glyphMarginWidgets[t];delete this._glyphMarginWidgets[t],this._modelData&&this._modelData.hasRealView&&this._modelData.view.removeGlyphMarginWidget(i)}}changeViewZones(e){!this._modelData||!this._modelData.hasRealView||this._modelData.view.change(e)}getTargetAtClientPoint(e,t){return!this._modelData||!this._modelData.hasRealView?null:this._modelData.view.getTargetAtClientPoint(e,t)}getScrolledVisiblePosition(e){if(!this._modelData||!this._modelData.hasRealView)return null;const t=this._modelData.model.validatePosition(e),n=this._configuration.options.get(165),o=of._getVerticalOffsetForPosition(this._modelData,t.lineNumber,t.column)-this.getScrollTop(),r=this._modelData.view.getOffsetForColumn(t.lineNumber,t.column)+n.glyphMarginWidth+n.lineNumbersWidth+n.decorationsWidth-this.getScrollLeft(),a=this.getLineHeightForPosition(t);return{top:o,left:r,height:a}}getOffsetForColumn(e,t){return!this._modelData||!this._modelData.hasRealView?-1:this._modelData.view.getOffsetForColumn(e,t)}render(e=!1){!this._modelData||!this._modelData.hasRealView||this._modelData.viewModel.batchEvents(()=>{this._modelData.view.render(!0,e)})}setAriaOptions(e){!this._modelData||!this._modelData.hasRealView||this._modelData.view.setAriaOptions(e)}applyFontInfo(e){In(e,this._configuration.options.get(59))}setBanner(e,t){this._bannerDomNode&&this._domElement.contains(this._bannerDomNode)&&this._bannerDomNode.remove(),this._bannerDomNode=e,this._configuration.setReservedHeight(e?t:0),this._bannerDomNode&&this._domElement.prepend(this._bannerDomNode)}_attachModel(e){if(!e){this._modelData=null;return}const t=[];this._domElement.setAttribute("data-mode-id",e.getLanguageId()),this._configuration.setIsDominatedByLongLines(e.isDominatedByLongLines()),this._configuration.setModelLineCount(e.getLineCount());const i=e.onBeforeAttached(),n=new DSe(this._id,this._configuration,e,t7.create(be(this._domElement)),p7.create(this._configuration.options),a=>Co(be(this._domElement),a),this.languageConfigurationService,this._themeService,i,{batchChanges:a=>{try{return this._beginUpdate(),a()}finally{this._endUpdate()}}});t.push(e.onWillDispose(()=>this.setModel(null))),t.push(n.onEvent(a=>{switch(a.kind){case 0:this._onDidContentSizeChange.fire(a);break;case 1:this._editorTextFocus.setValue(a.hasFocus);break;case 2:this._editorWidgetFocus.setValue(a.hasFocus);break;case 3:this._onDidScrollChange.fire(a);break;case 4:this._onDidChangeViewZones.fire();break;case 5:this._onDidChangeHiddenAreas.fire();break;case 6:this._onDidAttemptReadOnlyEdit.fire();break;case 7:{if(a.reachedMaxCursorCount){const h=this.getOption(89),u=m(83,"The number of cursors has been limited to {0}. Consider using [find and replace](https://code.visualstudio.com/docs/editor/codebasics#_find-and-replace) for larger changes or increase the editor multi cursor limit setting.",h);this._notificationService.prompt(oC.Warning,u,[{label:"Find and Replace",run:()=>{this._commandService.executeCommand("editor.action.startFindReplaceAction")}},{label:m(84,"Increase Multi Cursor Limit"),run:()=>{this._commandService.executeCommand("workbench.action.openSettings2",{query:"editor.multiCursorLimit"})}}])}const l=[];for(let h=0,u=a.selections.length;h<u;h++)l[h]=a.selections[h].getPosition();const c={position:l[0],secondaryPositions:l.slice(1),reason:a.reason,source:a.source};this._onDidChangeCursorPosition.fire(c);const d={selection:a.selections[0],secondarySelections:a.selections.slice(1),modelVersionId:a.modelVersionId,oldSelections:a.oldSelections,oldModelVersionId:a.oldModelVersionId,source:a.source,reason:a.reason};this._onDidChangeCursorSelection.fire(d);break}case 8:this._onDidChangeModelDecorations.fire(a.event);break;case 9:this._domElement.setAttribute("data-mode-id",e.getLanguageId()),this._onDidChangeModelLanguage.fire(a.event);break;case 10:this._onDidChangeModelLanguageConfiguration.fire(a.event);break;case 11:this._onDidChangeModelContent.fire(a.event);break;case 12:this._onDidChangeModelOptions.fire(a.event);break;case 13:this._onDidChangeModelTokens.fire(a.event);break;case 14:this._onDidChangeLineHeight.fire(a.event);break;case 15:this._onDidChangeFont.fire(a.event);break}}));const[o,r]=this._createView(n);if(r){this._domElement.appendChild(o.domNode.domNode);let a=Object.keys(this._contentWidgets);for(let l=0,c=a.length;l<c;l++){const d=a[l];o.addContentWidget(this._contentWidgets[d])}a=Object.keys(this._overlayWidgets);for(let l=0,c=a.length;l<c;l++){const d=a[l];o.addOverlayWidget(this._overlayWidgets[d])}a=Object.keys(this._glyphMarginWidgets);for(let l=0,c=a.length;l<c;l++){const d=a[l];o.addGlyphMarginWidget(this._glyphMarginWidgets[d])}o.render(!1,!0),o.domNode.domNode.setAttribute("data-uri",e.uri.toString())}this._modelData=new OSe(e,n,o,r,t,i)}_createView(e){let t;this.isSimpleWidget?t={paste:(o,r,a,l)=>{this._paste("keyboard",o,r,a,l)},type:o=>{this._type("keyboard",o)},compositionType:(o,r,a,l)=>{this._compositionType("keyboard",o,r,a,l)},startComposition:()=>{this._startComposition()},endComposition:()=>{this._endComposition("keyboard")},cut:()=>{this._cut("keyboard")}}:t={paste:(o,r,a,l)=>{const c={text:o,pasteOnNewLine:r,multicursorText:a,mode:l};this._commandService.executeCommand("paste",c)},type:o=>{const r={text:o};this._commandService.executeCommand("type",r)},compositionType:(o,r,a,l)=>{if(a||l){const c={text:o,replacePrevCharCnt:r,replaceNextCharCnt:a,positionDelta:l};this._commandService.executeCommand("compositionType",c)}else{const c={text:o,replaceCharCnt:r};this._commandService.executeCommand("replacePreviousChar",c)}},startComposition:()=>{this._commandService.executeCommand("compositionStart",{})},endComposition:()=>{this._commandService.executeCommand("compositionEnd",{})},cut:()=>{this._commandService.executeCommand("cut",{})}};const i=new dA(e.coordinatesConverter);return i.onKeyDown=o=>this._onKeyDown.fire(o),i.onKeyUp=o=>this._onKeyUp.fire(o),i.onContextMenu=o=>this._onContextMenu.fire(o),i.onMouseMove=o=>this._onMouseMove.fire(o),i.onMouseLeave=o=>this._onMouseLeave.fire(o),i.onMouseDown=o=>this._onMouseDown.fire(o),i.onMouseUp=o=>this._onMouseUp.fire(o),i.onMouseDrag=o=>this._onMouseDrag.fire(o),i.onMouseDrop=o=>this._onMouseDrop.fire(o),i.onMouseDropCanceled=o=>this._onMouseDropCanceled.fire(o),i.onMouseWheel=o=>this._onMouseWheel.fire(o),[new kF(this._domElement,this.getId(),t,this._configuration,this._themeService.getColorTheme(),e,i,this._overflowWidgetsDomNode,this._instantiationService),!0]}_postDetachModelCleanup(e){e?.removeAllDecorationsWithOwnerId(this._id)}_detachModel(){if(this._contributionsDisposable?.dispose(),this._contributionsDisposable=void 0,!this._modelData)return null;const e=this._modelData.model,t=this._modelData.hasRealView?this._modelData.view.domNode.domNode:null;return this._modelData.dispose(),this._modelData=null,this._domElement.removeAttribute("data-mode-id"),t&&this._domElement.contains(t)&&t.remove(),this._bannerDomNode&&this._domElement.contains(this._bannerDomNode)&&this._bannerDomNode.remove(),e}_removeDecorationType(e){this._codeEditorService.removeDecorationType(e)}hasModel(){return this._modelData!==null}showDropIndicatorAt(e){const t=[{range:new k(e.lineNumber,e.column,e.lineNumber,e.column),options:of.dropIntoEditorDecorationOptions}];this._dropIntoEditorDecorations.set(t),this.revealPosition(e,1)}removeDropIndicator(){this._dropIntoEditorDecorations.clear()}_beginUpdate(){this._updateCounter++,this._updateCounter===1&&this._onBeginUpdate.fire()}_endUpdate(){this._updateCounter--,this._updateCounter===0&&this._onEndUpdate.fire()}},of=xp,xp.dropIntoEditorDecorationOptions=Fe.register({description:"workbench-dnd-target",className:"dnd-target"}),xp);Am=of=ASe([nd(3,_e),nd(4,ht),nd(5,Gt),nd(6,Ae),nd(7,wi),nd(8,Ii),nd(9,Wn),nd(10,ui),nd(11,ge)],Am);let MSe=0,OSe=class{constructor(e,t,i,n,o,r){this.model=e,this.viewModel=t,this.view=i,this.hasRealView=n,this.listenersToRemove=o,this.attachedView=r}dispose(){xt(this.listenersToRemove),this.model.onBeforeDetached(this.attachedView),this.hasRealView&&this.view.dispose(),this.viewModel.dispose()}};class nq extends F{constructor(e){super(),this._emitterOptions=e,this._onDidChangeToTrue=this._register(new M(this._emitterOptions)),this.onDidChangeToTrue=this._onDidChangeToTrue.event,this._onDidChangeToFalse=this._register(new M(this._emitterOptions)),this.onDidChangeToFalse=this._onDidChangeToFalse.event,this._value=0}setValue(e){const t=e?2:1;this._value!==t&&(this._value=t,this._value===2?this._onDidChangeToTrue.fire():this._value===1&&this._onDidChangeToFalse.fire())}}class ys extends M{constructor(e,t){super({deliveryQueue:t}),this._contributions=e}fire(e){this._contributions.onBeforeInteractionEvent(),super.fire(e)}}class PSe extends F{constructor(e,t){super(),this._editor=e,t.createKey("editorId",e.getId()),this._editorSimpleInput=N.editorSimpleInput.bindTo(t),this._editorFocus=N.focus.bindTo(t),this._textInputFocus=N.textInputFocus.bindTo(t),this._editorTextFocus=N.editorTextFocus.bindTo(t),this._tabMovesFocus=N.tabMovesFocus.bindTo(t),this._editorReadonly=N.readOnly.bindTo(t),this._inDiffEditor=N.inDiffEditor.bindTo(t),this._editorColumnSelection=N.columnSelection.bindTo(t),this._hasMultipleSelections=N.hasMultipleSelections.bindTo(t),this._hasNonEmptySelection=N.hasNonEmptySelection.bindTo(t),this._canUndo=N.canUndo.bindTo(t),this._canRedo=N.canRedo.bindTo(t),this._register(this._editor.onDidChangeConfiguration(()=>this._updateFromConfig())),this._register(this._editor.onDidChangeCursorSelection(()=>this._updateFromSelection())),this._register(this._editor.onDidFocusEditorWidget(()=>this._updateFromFocus())),this._register(this._editor.onDidBlurEditorWidget(()=>this._updateFromFocus())),this._register(this._editor.onDidFocusEditorText(()=>this._updateFromFocus())),this._register(this._editor.onDidBlurEditorText(()=>this._updateFromFocus())),this._register(this._editor.onDidChangeModel(()=>this._updateFromModel())),this._register(this._editor.onDidChangeConfiguration(()=>this._updateFromModel())),this._register(av.onDidChangeTabFocus(i=>this._tabMovesFocus.set(i))),this._updateFromConfig(),this._updateFromSelection(),this._updateFromFocus(),this._updateFromModel(),this._editorSimpleInput.set(this._editor.isSimpleWidget)}_updateFromConfig(){const e=this._editor.getOptions();this._tabMovesFocus.set(e.get(164)||av.getTabFocusMode()),this._editorReadonly.set(e.get(104)),this._inDiffEditor.set(e.get(70)),this._editorColumnSelection.set(e.get(28))}_updateFromSelection(){const e=this._editor.getSelections();e?(this._hasMultipleSelections.set(e.length>1),this._hasNonEmptySelection.set(e.some(t=>!t.isEmpty()))):(this._hasMultipleSelections.reset(),this._hasNonEmptySelection.reset())}_updateFromFocus(){this._editorFocus.set(this._editor.hasWidgetFocus()&&!this._editor.isSimpleWidget),this._editorTextFocus.set(this._editor.hasTextFocus()&&!this._editor.isSimpleWidget),this._textInputFocus.set(this._editor.hasTextFocus())}_updateFromModel(){const e=this._editor.getModel();this._canUndo.set(!!(e&&e.canUndo())),this._canRedo.set(!!(e&&e.canRedo()))}}class FSe extends F{constructor(e,t,i){super(),this._editor=e,this._contextKeyService=t,this._languageFeaturesService=i,this._langId=N.languageId.bindTo(t),this._hasCompletionItemProvider=N.hasCompletionItemProvider.bindTo(t),this._hasCodeActionsProvider=N.hasCodeActionsProvider.bindTo(t),this._hasCodeLensProvider=N.hasCodeLensProvider.bindTo(t),this._hasDefinitionProvider=N.hasDefinitionProvider.bindTo(t),this._hasDeclarationProvider=N.hasDeclarationProvider.bindTo(t),this._hasImplementationProvider=N.hasImplementationProvider.bindTo(t),this._hasTypeDefinitionProvider=N.hasTypeDefinitionProvider.bindTo(t),this._hasHoverProvider=N.hasHoverProvider.bindTo(t),this._hasDocumentHighlightProvider=N.hasDocumentHighlightProvider.bindTo(t),this._hasDocumentSymbolProvider=N.hasDocumentSymbolProvider.bindTo(t),this._hasReferenceProvider=N.hasReferenceProvider.bindTo(t),this._hasRenameProvider=N.hasRenameProvider.bindTo(t),this._hasSignatureHelpProvider=N.hasSignatureHelpProvider.bindTo(t),this._hasInlayHintsProvider=N.hasInlayHintsProvider.bindTo(t),this._hasDocumentFormattingProvider=N.hasDocumentFormattingProvider.bindTo(t),this._hasDocumentSelectionFormattingProvider=N.hasDocumentSelectionFormattingProvider.bindTo(t),this._hasMultipleDocumentFormattingProvider=N.hasMultipleDocumentFormattingProvider.bindTo(t),this._hasMultipleDocumentSelectionFormattingProvider=N.hasMultipleDocumentSelectionFormattingProvider.bindTo(t),this._isInEmbeddedEditor=N.isInEmbeddedEditor.bindTo(t);const n=()=>this._update();this._register(e.onDidChangeModel(n)),this._register(e.onDidChangeModelLanguage(n)),this._register(i.completionProvider.onDidChange(n)),this._register(i.codeActionProvider.onDidChange(n)),this._register(i.codeLensProvider.onDidChange(n)),this._register(i.definitionProvider.onDidChange(n)),this._register(i.declarationProvider.onDidChange(n)),this._register(i.implementationProvider.onDidChange(n)),this._register(i.typeDefinitionProvider.onDidChange(n)),this._register(i.hoverProvider.onDidChange(n)),this._register(i.documentHighlightProvider.onDidChange(n)),this._register(i.documentSymbolProvider.onDidChange(n)),this._register(i.referenceProvider.onDidChange(n)),this._register(i.renameProvider.onDidChange(n)),this._register(i.documentFormattingEditProvider.onDidChange(n)),this._register(i.documentRangeFormattingEditProvider.onDidChange(n)),this._register(i.signatureHelpProvider.onDidChange(n)),this._register(i.inlayHintsProvider.onDidChange(n)),n()}dispose(){super.dispose()}reset(){this._contextKeyService.bufferChangeEvents(()=>{this._langId.reset(),this._hasCompletionItemProvider.reset(),this._hasCodeActionsProvider.reset(),this._hasCodeLensProvider.reset(),this._hasDefinitionProvider.reset(),this._hasDeclarationProvider.reset(),this._hasImplementationProvider.reset(),this._hasTypeDefinitionProvider.reset(),this._hasHoverProvider.reset(),this._hasDocumentHighlightProvider.reset(),this._hasDocumentSymbolProvider.reset(),this._hasReferenceProvider.reset(),this._hasRenameProvider.reset(),this._hasDocumentFormattingProvider.reset(),this._hasDocumentSelectionFormattingProvider.reset(),this._hasSignatureHelpProvider.reset(),this._isInEmbeddedEditor.reset()})}_update(){const e=this._editor.getModel();if(!e){this.reset();return}this._contextKeyService.bufferChangeEvents(()=>{this._langId.set(e.getLanguageId()),this._hasCompletionItemProvider.set(this._languageFeaturesService.completionProvider.has(e)),this._hasCodeActionsProvider.set(this._languageFeaturesService.codeActionProvider.has(e)),this._hasCodeLensProvider.set(this._languageFeaturesService.codeLensProvider.has(e)),this._hasDefinitionProvider.set(this._languageFeaturesService.definitionProvider.has(e)),this._hasDeclarationProvider.set(this._languageFeaturesService.declarationProvider.has(e)),this._hasImplementationProvider.set(this._languageFeaturesService.implementationProvider.has(e)),this._hasTypeDefinitionProvider.set(this._languageFeaturesService.typeDefinitionProvider.has(e)),this._hasHoverProvider.set(this._languageFeaturesService.hoverProvider.has(e)),this._hasDocumentHighlightProvider.set(this._languageFeaturesService.documentHighlightProvider.has(e)),this._hasDocumentSymbolProvider.set(this._languageFeaturesService.documentSymbolProvider.has(e)),this._hasReferenceProvider.set(this._languageFeaturesService.referenceProvider.has(e)),this._hasRenameProvider.set(this._languageFeaturesService.renameProvider.has(e)),this._hasSignatureHelpProvider.set(this._languageFeaturesService.signatureHelpProvider.has(e)),this._hasInlayHintsProvider.set(this._languageFeaturesService.inlayHintsProvider.has(e)),this._hasDocumentFormattingProvider.set(this._languageFeaturesService.documentFormattingEditProvider.has(e)||this._languageFeaturesService.documentRangeFormattingEditProvider.has(e)),this._hasDocumentSelectionFormattingProvider.set(this._languageFeaturesService.documentRangeFormattingEditProvider.has(e)),this._hasMultipleDocumentFormattingProvider.set(this._languageFeaturesService.documentFormattingEditProvider.all(e).length+this._languageFeaturesService.documentRangeFormattingEditProvider.all(e).length>1),this._hasMultipleDocumentSelectionFormattingProvider.set(this._languageFeaturesService.documentRangeFormattingEditProvider.all(e).length>1),this._isInEmbeddedEditor.set(e.uri.scheme===Te.walkThroughSnippet||e.uri.scheme===Te.vscodeChatCodeBlock)})}}class BSe{get length(){return this._decorationIds.length}constructor(e,t){this._editor=e,this._decorationIds=[],this._isChangingDecorations=!1,Array.isArray(t)&&t.length>0&&this.set(t)}onDidChange(e,t,i){return this._editor.onDidChangeModelDecorations(n=>{this._isChangingDecorations||e.call(t,n)},i)}getRange(e){return!this._editor.hasModel()||e>=this._decorationIds.length?null:this._editor.getModel().getDecorationRange(this._decorationIds[e])}getRanges(){if(!this._editor.hasModel())return[];const e=this._editor.getModel(),t=[];for(const i of this._decorationIds){const n=e.getDecorationRange(i);n&&t.push(n)}return t}has(e){return this._decorationIds.includes(e.id)}clear(){this._decorationIds.length!==0&&this.set([])}set(e){try{this._isChangingDecorations=!0,this._editor.changeDecorations(t=>{this._decorationIds=t.deltaDecorations(this._decorationIds,e)})}finally{this._isChangingDecorations=!1}return this._decorationIds}append(e){let t=[];try{this._isChangingDecorations=!0,this._editor.changeDecorations(i=>{t=i.deltaDecorations([],e),this._decorationIds=this._decorationIds.concat(t)})}finally{this._isChangingDecorations=!1}return t}}const WSe=encodeURIComponent("<svg xmlns='http://www.w3.org/2000/svg' viewBox='0 0 6 3' enable-background='new 0 0 6 3' height='3' width='6'><g fill='"),HSe=encodeURIComponent("'><polygon points='5.5,0 2.5,3 1.1,3 4.1,0'/><polygon points='4,0 6,2 6,0.6 5.4,0'/><polygon points='0,2 1,3 2.4,3 0,0.6'/></g></svg>");function $_(s){return WSe+encodeURIComponent(s.toString())+HSe}const VSe=encodeURIComponent('<svg xmlns="http://www.w3.org/2000/svg" height="3" width="12"><g fill="'),zSe=encodeURIComponent('"><circle cx="1" cy="1" r="1"/><circle cx="5" cy="1" r="1"/><circle cx="9" cy="1" r="1"/></g></svg>');function sq(s){return VSe+encodeURIComponent(s.toString())+zSe}ua((s,e)=>{const t=s.getColor(iA);t&&(e.addRule(`.monaco-editor .squiggly-error { background: url("data:image/svg+xml,${$_(t)}") repeat-x bottom left; }`),e.addRule(`:root { --monaco-editor-error-decoration: url("data:image/svg+xml,${$_(t)}"); }`));const i=s.getColor(_h);i&&(e.addRule(`.monaco-editor .squiggly-warning { background: url("data:image/svg+xml,${$_(i)}") repeat-x bottom left; }`),e.addRule(`:root { --monaco-editor-warning-decoration: url("data:image/svg+xml,${$_(i)}"); }`));const n=s.getColor(Rc);n&&(e.addRule(`.monaco-editor .squiggly-info { background: url("data:image/svg+xml,${$_(n)}") repeat-x bottom left; }`),e.addRule(`:root { --monaco-editor-info-decoration: url("data:image/svg+xml,${$_(n)}"); }`));const o=s.getColor(rme);o&&(e.addRule(`.monaco-editor .squiggly-hint { background: url("data:image/svg+xml,${sq(o)}") no-repeat bottom left; }`),e.addRule(`:root { --monaco-editor-hint-decoration: url("data:image/svg+xml,${sq(o)}"); }`));const r=s.getColor(cwe);r&&(e.addRule(`.monaco-editor.showUnused .squiggly-inline-unnecessary { opacity: ${r.rgba.a}; }`),e.addRule(`:root { --monaco-editor-unnecessary-decoration-opacity: ${r.rgba.a}; }`))});function zt(s){return new $Se(s)}class $Se extends CQ{constructor(e){super(),this.value=e}get debugName(){return this.toString()}get(){return this.value}addObserver(e){}removeObserver(e){}toString(){return`Const: ${this.value}`}}function fd(s,e){return s}function bA(s){return zt(s)}const Th=qe("accessibilitySignalService"),Ue=class Ue{static register(e){return new Ue(e.fileName)}constructor(e){this.fileName=e}};Ue.error=Ue.register({fileName:"error.mp3"}),Ue.warning=Ue.register({fileName:"warning.mp3"}),Ue.success=Ue.register({fileName:"success.mp3"}),Ue.foldedArea=Ue.register({fileName:"foldedAreas.mp3"}),Ue.break=Ue.register({fileName:"break.mp3"}),Ue.quickFixes=Ue.register({fileName:"quickFixes.mp3"}),Ue.taskCompleted=Ue.register({fileName:"taskCompleted.mp3"}),Ue.taskFailed=Ue.register({fileName:"taskFailed.mp3"}),Ue.terminalBell=Ue.register({fileName:"terminalBell.mp3"}),Ue.diffLineInserted=Ue.register({fileName:"diffLineInserted.mp3"}),Ue.diffLineDeleted=Ue.register({fileName:"diffLineDeleted.mp3"}),Ue.diffLineModified=Ue.register({fileName:"diffLineModified.mp3"}),Ue.requestSent=Ue.register({fileName:"requestSent.mp3"}),Ue.responseReceived1=Ue.register({fileName:"responseReceived1.mp3"}),Ue.responseReceived2=Ue.register({fileName:"responseReceived2.mp3"}),Ue.responseReceived3=Ue.register({fileName:"responseReceived3.mp3"}),Ue.responseReceived4=Ue.register({fileName:"responseReceived4.mp3"}),Ue.clear=Ue.register({fileName:"clear.mp3"}),Ue.save=Ue.register({fileName:"save.mp3"}),Ue.format=Ue.register({fileName:"format.mp3"}),Ue.voiceRecordingStarted=Ue.register({fileName:"voiceRecordingStarted.mp3"}),Ue.voiceRecordingStopped=Ue.register({fileName:"voiceRecordingStopped.mp3"}),Ue.progress=Ue.register({fileName:"progress.mp3"}),Ue.chatEditModifiedFile=Ue.register({fileName:"chatEditModifiedFile.mp3"}),Ue.editsKept=Ue.register({fileName:"editsKept.mp3"}),Ue.editsUndone=Ue.register({fileName:"editsUndone.mp3"}),Ue.nextEditSuggestion=Ue.register({fileName:"nextEditSuggestion.mp3"}),Ue.terminalCommandSucceeded=Ue.register({fileName:"terminalCommandSucceeded.mp3"}),Ue.chatUserActionRequired=Ue.register({fileName:"chatUserActionRequired.mp3"}),Ue.codeActionTriggered=Ue.register({fileName:"codeActionTriggered.mp3"}),Ue.codeActionApplied=Ue.register({fileName:"codeActionApplied.mp3"});let Mt=Ue;class USe{constructor(e){this.randomOneOf=e}}const We=class We{constructor(e,t,i,n,o,r,a=!1){this.sound=e,this.name=t,this.legacySoundSettingsKey=i,this.settingsKey=n,this.legacyAnnouncementSettingsKey=o,this.announcementMessage=r,this.managesOwnEnablement=a}static register(e){const t=new USe("randomOneOf"in e.sound?e.sound.randomOneOf:[e.sound]),i=new We(t,e.name,e.legacySoundSettingsKey,e.settingsKey,e.legacyAnnouncementSettingsKey,e.announcementMessage,e.managesOwnEnablement);return We._signals.add(i),i}};We._signals=new Set,We.errorAtPosition=We.register({name:m(1576,"Error at Position"),sound:Mt.error,announcementMessage:m(1577,"Error"),settingsKey:"accessibility.signals.positionHasError",delaySettingsKey:"accessibility.signalOptions.delays.errorAtPosition"}),We.warningAtPosition=We.register({name:m(1578,"Warning at Position"),sound:Mt.warning,announcementMessage:m(1579,"Warning"),settingsKey:"accessibility.signals.positionHasWarning",delaySettingsKey:"accessibility.signalOptions.delays.warningAtPosition"}),We.errorOnLine=We.register({name:m(1580,"Error on Line"),sound:Mt.error,legacySoundSettingsKey:"audioCues.lineHasError",legacyAnnouncementSettingsKey:"accessibility.alert.error",announcementMessage:m(1581,"Error on Line"),settingsKey:"accessibility.signals.lineHasError"}),We.warningOnLine=We.register({name:m(1582,"Warning on Line"),sound:Mt.warning,legacySoundSettingsKey:"audioCues.lineHasWarning",legacyAnnouncementSettingsKey:"accessibility.alert.warning",announcementMessage:m(1583,"Warning on Line"),settingsKey:"accessibility.signals.lineHasWarning"}),We.foldedArea=We.register({name:m(1584,"Folded Area on Line"),sound:Mt.foldedArea,legacySoundSettingsKey:"audioCues.lineHasFoldedArea",legacyAnnouncementSettingsKey:"accessibility.alert.foldedArea",announcementMessage:m(1585,"Folded"),settingsKey:"accessibility.signals.lineHasFoldedArea"}),We.break=We.register({name:m(1586,"Breakpoint on Line"),sound:Mt.break,legacySoundSettingsKey:"audioCues.lineHasBreakpoint",legacyAnnouncementSettingsKey:"accessibility.alert.breakpoint",announcementMessage:m(1587,"Breakpoint"),settingsKey:"accessibility.signals.lineHasBreakpoint"}),We.inlineSuggestion=We.register({name:m(1588,"Inline Suggestion on Line"),sound:Mt.quickFixes,legacySoundSettingsKey:"audioCues.lineHasInlineSuggestion",settingsKey:"accessibility.signals.lineHasInlineSuggestion"}),We.nextEditSuggestion=We.register({name:m(1589,"Next Edit Suggestion on Line"),sound:Mt.nextEditSuggestion,legacySoundSettingsKey:"audioCues.nextEditSuggestion",settingsKey:"accessibility.signals.nextEditSuggestion",announcementMessage:m(1590,"Next Edit Suggestion")}),We.terminalQuickFix=We.register({name:m(1591,"Terminal Quick Fix"),sound:Mt.quickFixes,legacySoundSettingsKey:"audioCues.terminalQuickFix",legacyAnnouncementSettingsKey:"accessibility.alert.terminalQuickFix",announcementMessage:m(1592,"Quick Fix"),settingsKey:"accessibility.signals.terminalQuickFix"}),We.onDebugBreak=We.register({name:m(1593,"Debugger Stopped on Breakpoint"),sound:Mt.break,legacySoundSettingsKey:"audioCues.onDebugBreak",legacyAnnouncementSettingsKey:"accessibility.alert.onDebugBreak",announcementMessage:m(1594,"Breakpoint"),settingsKey:"accessibility.signals.onDebugBreak"}),We.noInlayHints=We.register({name:m(1595,"No Inlay Hints on Line"),sound:Mt.error,legacySoundSettingsKey:"audioCues.noInlayHints",legacyAnnouncementSettingsKey:"accessibility.alert.noInlayHints",announcementMessage:m(1596,"No Inlay Hints"),settingsKey:"accessibility.signals.noInlayHints"}),We.taskCompleted=We.register({name:m(1597,"Task Completed"),sound:Mt.taskCompleted,legacySoundSettingsKey:"audioCues.taskCompleted",legacyAnnouncementSettingsKey:"accessibility.alert.taskCompleted",announcementMessage:m(1598,"Task Completed"),settingsKey:"accessibility.signals.taskCompleted"}),We.taskFailed=We.register({name:m(1599,"Task Failed"),sound:Mt.taskFailed,legacySoundSettingsKey:"audioCues.taskFailed",legacyAnnouncementSettingsKey:"accessibility.alert.taskFailed",announcementMessage:m(1600,"Task Failed"),settingsKey:"accessibility.signals.taskFailed"}),We.terminalCommandFailed=We.register({name:m(1601,"Terminal Command Failed"),sound:Mt.error,legacySoundSettingsKey:"audioCues.terminalCommandFailed",legacyAnnouncementSettingsKey:"accessibility.alert.terminalCommandFailed",announcementMessage:m(1602,"Command Failed"),settingsKey:"accessibility.signals.terminalCommandFailed"}),We.terminalCommandSucceeded=We.register({name:m(1603,"Terminal Command Succeeded"),sound:Mt.terminalCommandSucceeded,announcementMessage:m(1604,"Command Succeeded"),settingsKey:"accessibility.signals.terminalCommandSucceeded"}),We.terminalBell=We.register({name:m(1605,"Terminal Bell"),sound:Mt.terminalBell,legacySoundSettingsKey:"audioCues.terminalBell",legacyAnnouncementSettingsKey:"accessibility.alert.terminalBell",announcementMessage:m(1606,"Terminal Bell"),settingsKey:"accessibility.signals.terminalBell"}),We.notebookCellCompleted=We.register({name:m(1607,"Notebook Cell Completed"),sound:Mt.taskCompleted,legacySoundSettingsKey:"audioCues.notebookCellCompleted",legacyAnnouncementSettingsKey:"accessibility.alert.notebookCellCompleted",announcementMessage:m(1608,"Notebook Cell Completed"),settingsKey:"accessibility.signals.notebookCellCompleted"}),We.notebookCellFailed=We.register({name:m(1609,"Notebook Cell Failed"),sound:Mt.taskFailed,legacySoundSettingsKey:"audioCues.notebookCellFailed",legacyAnnouncementSettingsKey:"accessibility.alert.notebookCellFailed",announcementMessage:m(1610,"Notebook Cell Failed"),settingsKey:"accessibility.signals.notebookCellFailed"}),We.diffLineInserted=We.register({name:m(1611,"Diff Line Inserted"),sound:Mt.diffLineInserted,legacySoundSettingsKey:"audioCues.diffLineInserted",settingsKey:"accessibility.signals.diffLineInserted"}),We.diffLineDeleted=We.register({name:m(1612,"Diff Line Deleted"),sound:Mt.diffLineDeleted,legacySoundSettingsKey:"audioCues.diffLineDeleted",settingsKey:"accessibility.signals.diffLineDeleted"}),We.diffLineModified=We.register({name:m(1613,"Diff Line Modified"),sound:Mt.diffLineModified,legacySoundSettingsKey:"audioCues.diffLineModified",settingsKey:"accessibility.signals.diffLineModified"}),We.chatEditModifiedFile=We.register({name:m(1614,"Chat Edit Modified File"),sound:Mt.chatEditModifiedFile,announcementMessage:m(1615,"File Modified from Chat Edits"),settingsKey:"accessibility.signals.chatEditModifiedFile"}),We.chatRequestSent=We.register({name:m(1616,"Chat Request Sent"),sound:Mt.requestSent,legacySoundSettingsKey:"audioCues.chatRequestSent",legacyAnnouncementSettingsKey:"accessibility.alert.chatRequestSent",announcementMessage:m(1617,"Chat Request Sent"),settingsKey:"accessibility.signals.chatRequestSent"}),We.chatResponseReceived=We.register({name:m(1618,"Chat Response Received"),legacySoundSettingsKey:"audioCues.chatResponseReceived",sound:{randomOneOf:[Mt.responseReceived1,Mt.responseReceived2,Mt.responseReceived3,Mt.responseReceived4]},settingsKey:"accessibility.signals.chatResponseReceived"}),We.codeActionTriggered=We.register({name:m(1619,"Code Action Request Triggered"),sound:Mt.codeActionTriggered,legacySoundSettingsKey:"audioCues.codeActionRequestTriggered",legacyAnnouncementSettingsKey:"accessibility.alert.codeActionRequestTriggered",announcementMessage:m(1620,"Code Action Request Triggered"),settingsKey:"accessibility.signals.codeActionTriggered"}),We.codeActionApplied=We.register({name:m(1621,"Code Action Applied"),legacySoundSettingsKey:"audioCues.codeActionApplied",sound:Mt.codeActionApplied,settingsKey:"accessibility.signals.codeActionApplied"}),We.progress=We.register({name:m(1622,"Progress"),sound:Mt.progress,legacySoundSettingsKey:"audioCues.chatResponsePending",legacyAnnouncementSettingsKey:"accessibility.alert.progress",announcementMessage:m(1623,"Progress"),settingsKey:"accessibility.signals.progress"}),We.clear=We.register({name:m(1624,"Clear"),sound:Mt.clear,legacySoundSettingsKey:"audioCues.clear",legacyAnnouncementSettingsKey:"accessibility.alert.clear",announcementMessage:m(1625,"Clear"),settingsKey:"accessibility.signals.clear"}),We.save=We.register({name:m(1626,"Save"),sound:Mt.save,legacySoundSettingsKey:"audioCues.save",legacyAnnouncementSettingsKey:"accessibility.alert.save",announcementMessage:m(1627,"Save"),settingsKey:"accessibility.signals.save"}),We.format=We.register({name:m(1628,"Format"),sound:Mt.format,legacySoundSettingsKey:"audioCues.format",legacyAnnouncementSettingsKey:"accessibility.alert.format",announcementMessage:m(1629,"Format"),settingsKey:"accessibility.signals.format"}),We.voiceRecordingStarted=We.register({name:m(1630,"Voice Recording Started"),sound:Mt.voiceRecordingStarted,legacySoundSettingsKey:"audioCues.voiceRecordingStarted",settingsKey:"accessibility.signals.voiceRecordingStarted"}),We.voiceRecordingStopped=We.register({name:m(1631,"Voice Recording Stopped"),sound:Mt.voiceRecordingStopped,legacySoundSettingsKey:"audioCues.voiceRecordingStopped",settingsKey:"accessibility.signals.voiceRecordingStopped"}),We.editsKept=We.register({name:m(1632,"Edits Kept"),sound:Mt.editsKept,announcementMessage:m(1633,"Edits Kept"),settingsKey:"accessibility.signals.editsKept"}),We.editsUndone=We.register({name:m(1634,"Undo Edits"),sound:Mt.editsUndone,announcementMessage:m(1635,"Edits Undone"),settingsKey:"accessibility.signals.editsUndone"}),We.chatUserActionRequired=We.register({name:m(1636,"Chat User Action Required"),sound:Mt.chatUserActionRequired,announcementMessage:m(1637,"Chat User Action Required"),settingsKey:"accessibility.signals.chatUserActionRequired",managesOwnEnablement:!0});let Gs=We;function oq(s,e,t,i=dn.ofCaller()){return SQ({debugName:()=>`Configuration Key "${s}"`},n=>t.onDidChangeConfiguration(o=>{o.affectsConfiguration(s)&&n(o)}),()=>t.getValue(s)??e,i)}function jl(s,e,t,i=dn.ofCaller()){const n=s.bindTo(e),o=new U;return Jn({debugName:()=>`Set Context Key "${s.key}"`},r=>{const a=t(r);return n.set(a),a},i).recomputeInitiallyAndOnChange(o),o}const bte=qe("progressService"),WV=class WV{constructor(e){this.callback=e}report(e){this._value=e,this.callback(this._value)}};WV.None=Object.freeze({report(){}});let ml=WV;const Ch=qe("editorProgressService");class Il{static capture(e){if(e.getScrollTop()===0||e.hasPendingScrollAnimation())return new Il(e.getScrollTop(),e.getContentHeight(),null,0,null);let t=null,i=0;const n=e.getVisibleRanges();if(n.length>0){t=n[0].getStartPosition();const o=e.getTopForPosition(t.lineNumber,t.column);i=e.getScrollTop()-o}return new Il(e.getScrollTop(),e.getContentHeight(),t,i,e.getPosition())}constructor(e,t,i,n,o){this._initialScrollTop=e,this._initialContentHeight=t,this._visiblePosition=i,this._visiblePositionScrollDelta=n,this._cursorPosition=o}restore(e){if(!(this._initialContentHeight===e.getContentHeight()&&this._initialScrollTop===e.getScrollTop())&&this._visiblePosition){const t=e.getTopForPosition(this._visiblePosition.lineNumber,this._visiblePosition.column);e.setScrollTop(t+this._visiblePositionScrollDelta)}}restoreRelativeVerticalPositionOfCursor(e){if(this._initialContentHeight===e.getContentHeight()&&this._initialScrollTop===e.getScrollTop())return;const t=e.getPosition();if(!this._cursorPosition||!t)return;const i=e.getTopForLineNumber(t.lineNumber)-e.getTopForLineNumber(this._cursorPosition.lineNumber);e.setScrollTop(e.getScrollTop()+i,1)}}const Mm={RESOURCES:"ResourceURLs",TEXT:Ti.text,INTERNAL_URI_LIST:"application/vnd.code.uri-list"},qSe=()=>({get delay(){return-1},dispose:()=>{},showHover:()=>{}});let wA=qSe;const jSe=new es(()=>wA("mouse",!1)),GSe=new es(()=>wA("element",!1));function KSe(s){wA=s}function Uc(s){return s==="element"?GSe.value:jSe.value}function wte(){return wA("element",!0)}function sh(s,e){if(s!==void 0){const t=s.match(/^\s*var\((.+)\)$/);if(t){const i=t[1].split(",",2);return i.length===2&&(e=sh(i[1].trim(),e)),`var(${i[0]}, ${e})`}return s}return e}function rq(s){const e=s.replaceAll(/[^_\-a-z0-9]/gi,"");return e!==s&&console.warn(`CSS ident value ${s} modified to ${e} to be safe for CSS`),e}function Hh(s){return`'${s.replaceAll(/'/g,"\\000027")}'`}function Mc(s){return s?Do`url('${CSS.escape(gQ.uriToBrowserUri(s).toString(!0))}')`:"url('')"}function fL(s,e=!1){const t=CSS.escape(s);return!e&&t!==s&&console.warn(`CSS class name ${s} modified to ${t} to be safe for CSS`),t}function Do(s,...e){return s.reduce((t,i,n)=>{const o=e[n]||"";return t+i+o},"")}class CM{constructor(){this._parts=[]}push(...e){this._parts.push(...e)}join(e=`
`){return this._parts.join(e)}}function YSe(s){const e=ZSe(s);if(e&&e.length>0)return new Uint32Array(e)}let cr=0;const lu=new Uint32Array(10);function ZSe(s){if(cr=0,Bl(s,yM,4352),cr>0||(Bl(s,SM,4449),cr>0)||(Bl(s,xM,4520),cr>0)||(Bl(s,zg,12593),cr))return lu.subarray(0,cr);if(s>=44032&&s<=55203){const e=s-44032,t=e%588,i=Math.floor(e/588),n=Math.floor(t/28),o=t%28-1;if(i<yM.length?Bl(i,yM,0):4352+i-12593<zg.length&&Bl(4352+i,zg,12593),n<SM.length?Bl(n,SM,0):4449+n-12593<zg.length&&Bl(4449+n-12593,zg,12593),o>=0&&(o<xM.length?Bl(o,xM,0):4520+o-12593<zg.length&&Bl(4520+o-12593,zg,12593)),cr>0)return lu.subarray(0,cr)}}function Bl(s,e,t){s>=t&&s<t+e.length&&XSe(e[s-t])}function XSe(s){s!==0&&(lu[cr++]=s&255,s>>8&&(lu[cr++]=s>>8&255),s>>16&&(lu[cr++]=s>>16&255))}const yM=new Uint8Array([114,82,115,101,69,102,97,113,81,116,84,100,119,87,99,122,120,118,103]),SM=new Uint16Array([107,111,105,79,106,112,117,80,104,27496,28520,27752,121,110,27246,28782,27758,98,109,27757,108]),xM=new Uint16Array([114,82,29810,115,30579,26483,101,102,29286,24934,29030,29798,30822,30310,26470,97,113,29809,116,84,100,119,99,122,120,118,103]),zg=new Uint16Array([114,82,29810,115,30579,26483,101,69,102,29286,24934,29030,29798,30822,30310,26470,97,113,81,29809,116,84,100,119,87,99,122,120,118,103,107,111,105,79,106,112,117,80,104,27496,28520,27752,121,110,27246,28782,27758,98,109,27757,108]);function w7(...s){return function(e,t){for(let i=0,n=s.length;i<n;i++){const o=s[i](e,t);if(o)return o}return null}}vte.bind(void 0,!1);const DS=vte.bind(void 0,!0);function vte(s,e,t){if(!t||t.length<e.length)return null;let i;return s?i=j6(t,e):i=t.indexOf(e)===0,i?e.length>0?[{start:0,end:e.length}]:[]:null}function Cte(s,e){const t=e.toLowerCase().indexOf(s.toLowerCase());return t===-1?null:[{start:t,end:t+s.length}]}function yte(s,e){return GF(s.toLowerCase(),e.toLowerCase(),0,0)}function GF(s,e,t,i){if(t===s.length)return[];if(i===e.length)return null;if(s[t]===e[i]){let n=null;return(n=GF(s,e,t+1,i+1))?y7({start:i,end:i+1},n):null}return GF(s,e,t,i+1)}function v7(s){return 97<=s&&s<=122}function vA(s){return 65<=s&&s<=90}function C7(s){return 48<=s&&s<=57}function Ste(s){return s===32||s===9||s===10||s===13}const xte=new Set;"()[]{}<>`'\"-/;:,.?!".split("").forEach(s=>xte.add(s.charCodeAt(0)));function _I(s){return Ste(s)||xte.has(s)}function aq(s,e){return s===e||_I(s)&&_I(e)}const LM=new Map;function lq(s){if(LM.has(s))return LM.get(s);let e;const t=YSe(s);return t&&(e=t),LM.set(s,e),e}function Lte(s){return v7(s)||vA(s)||C7(s)}function y7(s,e){return e.length===0?e=[s]:s.end===e[0].start?e[0].start=s.start:e.unshift(s),e}function kte(s,e){for(let t=e;t<s.length;t++){const i=s.charCodeAt(t);if(vA(i)||C7(i)||t>0&&!Lte(s.charCodeAt(t-1)))return t}return s.length}function KF(s,e,t,i){if(t===s.length)return[];if(i===e.length)return null;if(s[t]!==e[i].toLowerCase())return null;{let n=null,o=i+1;for(n=KF(s,e,t+1,i+1);!n&&(o=kte(e,o))<e.length;)n=KF(s,e,t+1,o),o++;return n===null?null:y7({start:i,end:i+1},n)}}function QSe(s){let e=0,t=0,i=0,n=0,o=0;for(let d=0;d<s.length;d++)o=s.charCodeAt(d),vA(o)&&e++,v7(o)&&t++,Lte(o)&&i++,C7(o)&&n++;const r=e/s.length,a=t/s.length,l=i/s.length,c=n/s.length;return{upperPercent:r,lowerPercent:a,alphaPercent:l,numericPercent:c}}function JSe(s){const{upperPercent:e,lowerPercent:t}=s;return t===0&&e>.6}function e0e(s){const{upperPercent:e,lowerPercent:t,alphaPercent:i,numericPercent:n}=s;return t>.2&&e<.8&&i>.6&&n<.2}function t0e(s){let e=0,t=0,i=0,n=0;for(let o=0;o<s.length;o++)i=s.charCodeAt(o),vA(i)&&e++,v7(i)&&t++,Ste(i)&&n++;return(e===0||t===0)&&n===0?s.length<=30:e<=5}function Ete(s,e){if(!e||(e=e.trim(),e.length===0)||!t0e(s))return null;e.length>60&&(e=e.substring(0,60));const t=QSe(e);if(!e0e(t)){if(!JSe(t))return null;e=e.toLowerCase()}let i=null,n=0;for(s=s.toLowerCase();n<e.length&&(i=KF(s,e,0,n))===null;)n=kte(e,n+1);return i}function i0e(s,e,t=!1){if(!e||e.length===0)return null;let i=null,n=0;for(s=s.toLowerCase(),e=e.toLowerCase();n<e.length&&(i=YF(s,e,0,n,t),i===null);)n=Ite(e,n+1);return i}function YF(s,e,t,i,n){let o=0;if(t===s.length)return[];if(i===e.length)return null;if(!aq(s.charCodeAt(t),e.charCodeAt(i))){const l=lq(s.charCodeAt(t));if(!l)return null;for(let c=0;c<l.length;c++)if(!aq(l[c],e.charCodeAt(i+c)))return null;o+=l.length-1}let r=null,a=i+o+1;if(r=YF(s,e,t+1,a,n),!n)for(;!r&&(a=Ite(e,a))<e.length;)r=YF(s,e,t+1,a,n),a++;if(!r)return null;if(s.charCodeAt(t)!==e.charCodeAt(i)){const l=lq(s.charCodeAt(t));if(!l)return r;for(let c=0;c<l.length;c++)if(l[c]!==e.charCodeAt(i+c))return r}return y7({start:i,end:i+o+1},r)}function Ite(s,e){for(let t=e;t<s.length;t++)if(_I(s.charCodeAt(t))||t>0&&_I(s.charCodeAt(t-1)))return t;return s.length}const n0e=w7(DS,Ete,Cte),s0e=w7(DS,Ete,yte),cq=new Va(1e4);function dq(s,e,t=!1){if(typeof s!="string"||typeof e!="string")return null;let i=cq.get(s);i||(i=new RegExp(Pde(s),"i"),cq.set(s,i));const n=i.exec(e);return n?[{start:n.index,end:n.index+n[0].length}]:t?s0e(s,e):n0e(s,e)}function o0e(s,e){const t=Om(s,s.toLowerCase(),0,e,e.toLowerCase(),0,{firstMatchCanBeWeak:!0,boostFullMatch:!0});return t?ox(t):null}function r0e(s,e,t,i,n,o){const r=Math.min(13,s.length);for(;t<r;t++){const a=Om(s,e,t,i,n,o,{firstMatchCanBeWeak:!0,boostFullMatch:!0});if(a)return a}return[0,o]}function ox(s){if(typeof s>"u")return[];const e=[],t=s[1];for(let i=s.length-1;i>1;i--){const n=s[i]+t,o=e[e.length-1];o&&o.end===n?o.end=n+1:e.push({start:n,end:n+1})}return e}const Su=128;function S7(){const s=[],e=[];for(let t=0;t<=Su;t++)e[t]=0;for(let t=0;t<=Su;t++)s.push(e.slice(0));return s}function Dte(s){const e=[];for(let t=0;t<=s;t++)e[t]=0;return e}const Tte=Dte(2*Su),ZF=Dte(2*Su),sd=S7(),$g=S7(),pL=S7();function mL(s,e){if(e<0||e>=s.length)return!1;const t=s.codePointAt(e);switch(t){case 95:case 45:case 46:case 32:case 47:case 92:case 39:case 34:case 58:case 36:case 60:case 62:case 40:case 41:case 91:case 93:case 123:case 125:return!0;case void 0:return!1;default:return!!Z6(t)}}function hq(s,e){if(e<0||e>=s.length)return!1;switch(s.charCodeAt(e)){case 32:case 9:return!0;default:return!1}}function Ok(s,e,t){return e[s]!==t[s]}function a0e(s,e,t,i,n,o,r=!1){for(;e<t&&n<o;)s[e]===i[n]&&(r&&(Tte[e]=n),e+=1),n+=1;return e===t}var Oa;(function(s){s.Default=[-100,0];function e(t){return!t||t.length===2&&t[0]===-100&&t[1]===0}s.isDefault=e})(Oa||(Oa={}));const HV=class HV{constructor(e,t){this.firstMatchCanBeWeak=e,this.boostFullMatch=t}};HV.default={boostFullMatch:!0,firstMatchCanBeWeak:!1};let TS=HV;function Om(s,e,t,i,n,o,r=TS.default){const a=s.length>Su?Su:s.length,l=i.length>Su?Su:i.length;if(t>=a||o>=l||a-t>l-o||!a0e(e,t,a,n,o,l,!0))return;l0e(a,l,t,o,e,n);let c=1,d=1,h=t,u=o;const g=[!1];for(c=1,h=t;h<a;c++,h++){const w=Tte[h],C=ZF[h],v=h+1<a?ZF[h+1]:l;for(d=w-o+1,u=w;u<v;d++,u++){let y=Number.MIN_SAFE_INTEGER,x=!1;u<=C&&(y=c0e(s,e,h,t,i,n,u,l,o,sd[c-1][d-1]===0,g));let L=0;y!==Number.MIN_SAFE_INTEGER&&(x=!0,L=y+$g[c-1][d-1]);const E=u>w,I=E?$g[c][d-1]+(sd[c][d-1]>0?-5:0):0,R=u>w+1&&sd[c][d-1]>0,O=R?$g[c][d-2]+(sd[c][d-2]>0?-5:0):0;if(R&&(!E||O>=I)&&(!x||O>=L))$g[c][d]=O,pL[c][d]=3,sd[c][d]=0;else if(E&&(!x||I>=L))$g[c][d]=I,pL[c][d]=2,sd[c][d]=0;else if(x)$g[c][d]=L,pL[c][d]=1,sd[c][d]=sd[c-1][d-1]+1;else throw new Error("not possible")}}if(!g[0]&&!r.firstMatchCanBeWeak)return;c--,d--;const f=[$g[c][d],o];let p=0,_=0;for(;c>=1;){let w=d;do{const C=pL[c][w];if(C===3)w=w-2;else if(C===2)w=w-1;else break}while(w>=1);p>1&&e[t+c-1]===n[o+d-1]&&!Ok(w+o-1,i,n)&&p+1>sd[c][w]&&(w=d),w===d?p++:p=1,_||(_=w),c--,d=w-1,f.push(d)}l-o===a&&r.boostFullMatch&&(f[0]+=2);const b=_-a;return f[0]-=b,f}function l0e(s,e,t,i,n,o){let r=s-1,a=e-1;for(;r>=t&&a>=i;)n[r]===o[a]&&(ZF[r]=a,r--),a--}function c0e(s,e,t,i,n,o,r,a,l,c,d){if(e[t]!==o[r])return Number.MIN_SAFE_INTEGER;let h=1,u=!1;return r===t-i?h=s[t]===n[r]?7:5:Ok(r,n,o)&&(r===0||!Ok(r-1,n,o))?(h=s[t]===n[r]?7:5,u=!0):mL(o,r)&&(r===0||!mL(o,r-1))?h=5:(mL(o,r-1)||hq(o,r-1))&&(h=5,u=!0),h>1&&t===i&&(d[0]=!0),u||(u=Ok(r,n,o)||mL(o,r-1)||hq(o,r-1)),t===i?r>l&&(h-=u?3:5):c?h+=u?2:0:h+=u?0:1,r+1===a&&(h-=u?3:5),h}function d0e(s,e,t,i,n,o,r){return h0e(s,e,t,i,n,o,!0,r)}function h0e(s,e,t,i,n,o,r,a){let l=Om(s,e,t,i,n,o,a);if(s.length>=3){const c=Math.min(7,s.length-1);for(let d=t+1;d<c;d++){const h=u0e(s,d);if(h){const u=Om(h,h.toLowerCase(),t,i,n,o,a);u&&(u[0]-=3,(!l||u[0]>l[0])&&(l=u))}}}return l}function u0e(s,e){if(e+1>=s.length)return;const t=s[e],i=s[e+1];if(t!==i)return s.slice(0,e)+i+t+s.slice(e+2)}const g0e="$(",x7=new RegExp(`\\$\\(${xe.iconNameExpression}(?:${xe.iconModifierExpression})?\\)`,"g"),f0e=new RegExp(`(\\\\)?${x7.source}`,"g");function p0e(s){return s.replace(f0e,(e,t)=>t?e:`\\${e}`)}const m0e=new RegExp(`\\\\${x7.source}`,"g");function _0e(s){return s.replace(m0e,e=>`\\${e}`)}const b0e=new RegExp(`(\\s)?(\\\\)?${x7.source}(\\s)?`,"g");function L7(s){return s.indexOf(g0e)===-1?s:s.replace(b0e,(e,t,i,n)=>i?e:t||n||"")}function Nte(s){return s?s.replace(/\$\((.*?)\)/g,(e,t)=>` ${t} `).trim():""}const kM=new RegExp(`\\$\\(${xe.iconNameCharacter}+\\)`,"g");function Kf(s){kM.lastIndex=0;let e="";const t=[];let i=0;for(;;){const n=kM.lastIndex,o=kM.exec(s),r=s.substring(n,o?.index);if(r.length>0){e+=r;for(let a=0;a<r.length;a++)t.push(i)}if(!o)break;i+=o[0].length}return{text:e,iconOffsets:t}}function q1(s,e,t=!1){const{text:i,iconOffsets:n}=e;if(!n||n.length===0)return dq(s,i,t);const o=U0(i," "),r=i.length-o.length,a=dq(s,o,t);if(a)for(const l of a){const c=n[l.start+r]+r;l.start+=c,l.end+=c}return a}class hs{constructor(e="",t=!1){if(this.value=e,typeof this.value!="string")throw Qr("value");typeof t=="boolean"?(this.isTrusted=t,this.supportThemeIcons=!1,this.supportHtml=!1,this.supportAlertSyntax=!1):(this.isTrusted=t.isTrusted??void 0,this.supportThemeIcons=t.supportThemeIcons??!1,this.supportHtml=t.supportHtml??!1,this.supportAlertSyntax=t.supportAlertSyntax??!1)}appendText(e,t=0){return this.value+=v0e(this.supportThemeIcons?p0e(e):e).replace(/([ \t]+)/g,(i,n)=>"&nbsp;".repeat(n.length)).replace(/\>/gm,"\\>").replace(/\n/g,t===1?`\\
`:`

`),this}appendMarkdown(e){return this.value+=e,this}appendCodeblock(e,t){return this.value+=`
${C0e(t,e)}
`,this}appendLink(e,t,i){return this.value+="[",this.value+=this._escape(t,"]"),this.value+="](",this.value+=this._escape(String(e),")"),i&&(this.value+=` "${this._escape(this._escape(i,'"'),")")}"`),this.value+=")",this}_escape(e,t){const i=new RegExp(Uo(t),"g");return e.replace(i,(n,o)=>e.charAt(o-1)!=="\\"?`\\${n}`:n)}}function wv(s){return oh(s)?!s.value:Array.isArray(s)?s.every(wv):!0}function oh(s){return s instanceof hs?!0:s&&typeof s=="object"?typeof s.value=="string"&&(typeof s.isTrusted=="boolean"||typeof s.isTrusted=="object"||s.isTrusted===void 0)&&(typeof s.supportThemeIcons=="boolean"||s.supportThemeIcons===void 0)&&(typeof s.supportAlertSyntax=="boolean"||s.supportAlertSyntax===void 0):!1}function w0e(s,e){return s===e?!0:!s||!e?!1:s.value===e.value&&s.isTrusted===e.isTrusted&&s.supportThemeIcons===e.supportThemeIcons&&s.supportHtml===e.supportHtml&&s.supportAlertSyntax===e.supportAlertSyntax&&(s.baseUri===e.baseUri||!!s.baseUri&&!!e.baseUri&&dg(Le.from(s.baseUri),Le.from(e.baseUri)))}function v0e(s){return s.replace(/[\\`*_{}[\]()#+\-!~]/g,"\\$&")}function C0e(s,e){const t=s.match(/^`+/gm)?.reduce((n,o)=>n.length>o.length?n:o).length??0,i=t>=3?t+1:3;return[`${"`".repeat(i)}${e}`,s,`${"`".repeat(i)}`].join(`
`)}function _L(s){return s.replace(/"/g,"&quot;")}function EM(s){return s&&s.replace(/\\([\\`*_{}[\]()#+\-.!~])/g,"$1")}function y0e(s){const e=[],t=s.split("|").map(n=>n.trim());s=t[0];const i=t[1];if(i){const n=/height=(\d+)/.exec(i),o=/width=(\d+)/.exec(i),r=n?n[1]:"",a=o?o[1]:"",l=isFinite(parseInt(a)),c=isFinite(parseInt(r));l&&e.push(`width="${a}"`),c&&e.push(`height="${r}"`)}return{href:s,dimensions:e}}function Rte(s,...e){return Le.from({scheme:Te.command,path:s,query:e.length?encodeURIComponent(JSON.stringify(e)):void 0})}class k7{constructor(e){this._prefix=e,this._lastId=0}nextId(){return this._prefix+ ++this._lastId}}const XF=new k7("id#");function E7(){return{async:!1,breaks:!1,extensions:null,gfm:!0,hooks:null,pedantic:!1,renderer:null,silent:!1,tokenizer:null,walkTokens:null}}let u_=E7();function Ate(s){u_=s}const Mte=/[&<>"']/,S0e=new RegExp(Mte.source,"g"),Ote=/[<>"']|&(?!(#\d{1,7}|#[Xx][a-fA-F0-9]{1,6}|\w+);)/,x0e=new RegExp(Ote.source,"g"),L0e={"&":"&amp;","<":"&lt;",">":"&gt;",'"':"&quot;","'":"&#39;"},uq=s=>L0e[s];function zr(s,e){if(e){if(Mte.test(s))return s.replace(S0e,uq)}else if(Ote.test(s))return s.replace(x0e,uq);return s}const k0e=/(^|[^\[])\^/g;function Hi(s,e){let t=typeof s=="string"?s:s.source;e=e||"";const i={replace:(n,o)=>{let r=typeof o=="string"?o:o.source;return r=r.replace(k0e,"$1"),t=t.replace(n,r),i},getRegex:()=>new RegExp(t,e)};return i}function gq(s){try{s=encodeURI(s).replace(/%25/g,"%")}catch{return null}return s}const j1={exec:()=>null};function fq(s,e){const t=s.replace(/\|/g,(o,r,a)=>{let l=!1,c=r;for(;--c>=0&&a[c]==="\\";)l=!l;return l?"|":" |"}),i=t.split(/ \|/);let n=0;if(i[0].trim()||i.shift(),i.length>0&&!i[i.length-1].trim()&&i.pop(),e)if(i.length>e)i.splice(e);else for(;i.length<e;)i.push("");for(;n<i.length;n++)i[n]=i[n].trim().replace(/\\\|/g,"|");return i}function AC(s,e,t){const i=s.length;if(i===0)return"";let n=0;for(;n<i&&s.charAt(i-n-1)===e;)n++;return s.slice(0,i-n)}function E0e(s,e){if(s.indexOf(e[1])===-1)return-1;let t=0;for(let i=0;i<s.length;i++)if(s[i]==="\\")i++;else if(s[i]===e[0])t++;else if(s[i]===e[1]&&(t--,t<0))return i;return-1}function pq(s,e,t,i){const n=e.href,o=e.title?zr(e.title):null,r=s[1].replace(/\\([\[\]])/g,"$1");if(s[0].charAt(0)!=="!"){i.state.inLink=!0;const a={type:"link",raw:t,href:n,title:o,text:r,tokens:i.inlineTokens(r)};return i.state.inLink=!1,a}return{type:"image",raw:t,href:n,title:o,text:zr(r)}}function I0e(s,e){const t=s.match(/^(\s+)(?:```)/);if(t===null)return e;const i=t[1];return e.split(`
`).map(n=>{const o=n.match(/^\s+/);if(o===null)return n;const[r]=o;return r.length>=i.length?n.slice(i.length):n}).join(`
`)}class bI{options;rules;lexer;constructor(e){this.options=e||u_}space(e){const t=this.rules.block.newline.exec(e);if(t&&t[0].length>0)return{type:"space",raw:t[0]}}code(e){const t=this.rules.block.code.exec(e);if(t){const i=t[0].replace(/^ {1,4}/gm,"");return{type:"code",raw:t[0],codeBlockStyle:"indented",text:this.options.pedantic?i:AC(i,`
`)}}}fences(e){const t=this.rules.block.fences.exec(e);if(t){const i=t[0],n=I0e(i,t[3]||"");return{type:"code",raw:i,lang:t[2]?t[2].trim().replace(this.rules.inline.anyPunctuation,"$1"):t[2],text:n}}}heading(e){const t=this.rules.block.heading.exec(e);if(t){let i=t[2].trim();if(/#$/.test(i)){const n=AC(i,"#");(this.options.pedantic||!n||/ $/.test(n))&&(i=n.trim())}return{type:"heading",raw:t[0],depth:t[1].length,text:i,tokens:this.lexer.inline(i)}}}hr(e){const t=this.rules.block.hr.exec(e);if(t)return{type:"hr",raw:AC(t[0],`
`)}}blockquote(e){const t=this.rules.block.blockquote.exec(e);if(t){let i=AC(t[0],`
`).split(`
`),n="",o="";const r=[];for(;i.length>0;){let a=!1;const l=[];let c;for(c=0;c<i.length;c++)if(/^ {0,3}>/.test(i[c]))l.push(i[c]),a=!0;else if(!a)l.push(i[c]);else break;i=i.slice(c);const d=l.join(`
`),h=d.replace(/\n {0,3}((?:=+|-+) *)(?=\n|$)/g,`
    $1`).replace(/^ {0,3}>[ \t]?/gm,"");n=n?`${n}
${d}`:d,o=o?`${o}
${h}`:h;const u=this.lexer.state.top;if(this.lexer.state.top=!0,this.lexer.blockTokens(h,r,!0),this.lexer.state.top=u,i.length===0)break;const g=r[r.length-1];if(g?.type==="code")break;if(g?.type==="blockquote"){const f=g,p=f.raw+`
`+i.join(`
`),_=this.blockquote(p);r[r.length-1]=_,n=n.substring(0,n.length-f.raw.length)+_.raw,o=o.substring(0,o.length-f.text.length)+_.text;break}else if(g?.type==="list"){const f=g,p=f.raw+`
`+i.join(`
`),_=this.list(p);r[r.length-1]=_,n=n.substring(0,n.length-g.raw.length)+_.raw,o=o.substring(0,o.length-f.raw.length)+_.raw,i=p.substring(r[r.length-1].raw.length).split(`
`);continue}}return{type:"blockquote",raw:n,tokens:r,text:o}}}list(e){let t=this.rules.block.list.exec(e);if(t){let i=t[1].trim();const n=i.length>1,o={type:"list",raw:"",ordered:n,start:n?+i.slice(0,-1):"",loose:!1,items:[]};i=n?`\\d{1,9}\\${i.slice(-1)}`:`\\${i}`,this.options.pedantic&&(i=n?i:"[*+-]");const r=new RegExp(`^( {0,3}${i})((?:[	 ][^\\n]*)?(?:\\n|$))`);let a=!1;for(;e;){let l=!1,c="",d="";if(!(t=r.exec(e))||this.rules.block.hr.test(e))break;c=t[0],e=e.substring(c.length);let h=t[2].split(`
`,1)[0].replace(/^\t+/,b=>" ".repeat(3*b.length)),u=e.split(`
`,1)[0],g=!h.trim(),f=0;if(this.options.pedantic?(f=2,d=h.trimStart()):g?f=t[1].length+1:(f=t[2].search(/[^ ]/),f=f>4?1:f,d=h.slice(f),f+=t[1].length),g&&/^ *$/.test(u)&&(c+=u+`
`,e=e.substring(u.length+1),l=!0),!l){const b=new RegExp(`^ {0,${Math.min(3,f-1)}}(?:[*+-]|\\d{1,9}[.)])((?:[ 	][^\\n]*)?(?:\\n|$))`),w=new RegExp(`^ {0,${Math.min(3,f-1)}}((?:- *){3,}|(?:_ *){3,}|(?:\\* *){3,})(?:\\n+|$)`),C=new RegExp(`^ {0,${Math.min(3,f-1)}}(?:\`\`\`|~~~)`),v=new RegExp(`^ {0,${Math.min(3,f-1)}}#`);for(;e;){const y=e.split(`
`,1)[0];if(u=y,this.options.pedantic&&(u=u.replace(/^ {1,4}(?=( {4})*[^ ])/g,"  ")),C.test(u)||v.test(u)||b.test(u)||w.test(e))break;if(u.search(/[^ ]/)>=f||!u.trim())d+=`
`+u.slice(f);else{if(g||h.search(/[^ ]/)>=4||C.test(h)||v.test(h)||w.test(h))break;d+=`
`+u}!g&&!u.trim()&&(g=!0),c+=y+`
`,e=e.substring(y.length+1),h=u.slice(f)}}o.loose||(a?o.loose=!0:/\n *\n *$/.test(c)&&(a=!0));let p=null,_;this.options.gfm&&(p=/^\[[ xX]\] /.exec(d),p&&(_=p[0]!=="[ ] ",d=d.replace(/^\[[ xX]\] +/,""))),o.items.push({type:"list_item",raw:c,task:!!p,checked:_,loose:!1,text:d,tokens:[]}),o.raw+=c}o.items[o.items.length-1].raw=o.items[o.items.length-1].raw.trimEnd(),o.items[o.items.length-1].text=o.items[o.items.length-1].text.trimEnd(),o.raw=o.raw.trimEnd();for(let l=0;l<o.items.length;l++)if(this.lexer.state.top=!1,o.items[l].tokens=this.lexer.blockTokens(o.items[l].text,[]),!o.loose){const c=o.items[l].tokens.filter(h=>h.type==="space"),d=c.length>0&&c.some(h=>/\n.*\n/.test(h.raw));o.loose=d}if(o.loose)for(let l=0;l<o.items.length;l++)o.items[l].loose=!0;return o}}html(e){const t=this.rules.block.html.exec(e);if(t)return{type:"html",block:!0,raw:t[0],pre:t[1]==="pre"||t[1]==="script"||t[1]==="style",text:t[0]}}def(e){const t=this.rules.block.def.exec(e);if(t){const i=t[1].toLowerCase().replace(/\s+/g," "),n=t[2]?t[2].replace(/^<(.*)>$/,"$1").replace(this.rules.inline.anyPunctuation,"$1"):"",o=t[3]?t[3].substring(1,t[3].length-1).replace(this.rules.inline.anyPunctuation,"$1"):t[3];return{type:"def",tag:i,raw:t[0],href:n,title:o}}}table(e){const t=this.rules.block.table.exec(e);if(!t||!/[:|]/.test(t[2]))return;const i=fq(t[1]),n=t[2].replace(/^\||\| *$/g,"").split("|"),o=t[3]&&t[3].trim()?t[3].replace(/\n[ \t]*$/,"").split(`
`):[],r={type:"table",raw:t[0],header:[],align:[],rows:[]};if(i.length===n.length){for(const a of n)/^ *-+: *$/.test(a)?r.align.push("right"):/^ *:-+: *$/.test(a)?r.align.push("center"):/^ *:-+ *$/.test(a)?r.align.push("left"):r.align.push(null);for(let a=0;a<i.length;a++)r.header.push({text:i[a],tokens:this.lexer.inline(i[a]),header:!0,align:r.align[a]});for(const a of o)r.rows.push(fq(a,r.header.length).map((l,c)=>({text:l,tokens:this.lexer.inline(l),header:!1,align:r.align[c]})));return r}}lheading(e){const t=this.rules.block.lheading.exec(e);if(t)return{type:"heading",raw:t[0],depth:t[2].charAt(0)==="="?1:2,text:t[1],tokens:this.lexer.inline(t[1])}}paragraph(e){const t=this.rules.block.paragraph.exec(e);if(t){const i=t[1].charAt(t[1].length-1)===`
`?t[1].slice(0,-1):t[1];return{type:"paragraph",raw:t[0],text:i,tokens:this.lexer.inline(i)}}}text(e){const t=this.rules.block.text.exec(e);if(t)return{type:"text",raw:t[0],text:t[0],tokens:this.lexer.inline(t[0])}}escape(e){const t=this.rules.inline.escape.exec(e);if(t)return{type:"escape",raw:t[0],text:zr(t[1])}}tag(e){const t=this.rules.inline.tag.exec(e);if(t)return!this.lexer.state.inLink&&/^<a /i.test(t[0])?this.lexer.state.inLink=!0:this.lexer.state.inLink&&/^<\/a>/i.test(t[0])&&(this.lexer.state.inLink=!1),!this.lexer.state.inRawBlock&&/^<(pre|code|kbd|script)(\s|>)/i.test(t[0])?this.lexer.state.inRawBlock=!0:this.lexer.state.inRawBlock&&/^<\/(pre|code|kbd|script)(\s|>)/i.test(t[0])&&(this.lexer.state.inRawBlock=!1),{type:"html",raw:t[0],inLink:this.lexer.state.inLink,inRawBlock:this.lexer.state.inRawBlock,block:!1,text:t[0]}}link(e){const t=this.rules.inline.link.exec(e);if(t){const i=t[2].trim();if(!this.options.pedantic&&/^</.test(i)){if(!/>$/.test(i))return;const r=AC(i.slice(0,-1),"\\");if((i.length-r.length)%2===0)return}else{const r=E0e(t[2],"()");if(r>-1){const l=(t[0].indexOf("!")===0?5:4)+t[1].length+r;t[2]=t[2].substring(0,r),t[0]=t[0].substring(0,l).trim(),t[3]=""}}let n=t[2],o="";if(this.options.pedantic){const r=/^([^'"]*[^\s])\s+(['"])(.*)\2/.exec(n);r&&(n=r[1],o=r[3])}else o=t[3]?t[3].slice(1,-1):"";return n=n.trim(),/^</.test(n)&&(this.options.pedantic&&!/>$/.test(i)?n=n.slice(1):n=n.slice(1,-1)),pq(t,{href:n&&n.replace(this.rules.inline.anyPunctuation,"$1"),title:o&&o.replace(this.rules.inline.anyPunctuation,"$1")},t[0],this.lexer)}}reflink(e,t){let i;if((i=this.rules.inline.reflink.exec(e))||(i=this.rules.inline.nolink.exec(e))){const n=(i[2]||i[1]).replace(/\s+/g," "),o=t[n.toLowerCase()];if(!o){const r=i[0].charAt(0);return{type:"text",raw:r,text:r}}return pq(i,o,i[0],this.lexer)}}emStrong(e,t,i=""){let n=this.rules.inline.emStrongLDelim.exec(e);if(!n||n[3]&&i.match(/[\p{L}\p{N}]/u))return;if(!(n[1]||n[2]||"")||!i||this.rules.inline.punctuation.exec(i)){const r=[...n[0]].length-1;let a,l,c=r,d=0;const h=n[0][0]==="*"?this.rules.inline.emStrongRDelimAst:this.rules.inline.emStrongRDelimUnd;for(h.lastIndex=0,t=t.slice(-1*e.length+r);(n=h.exec(t))!=null;){if(a=n[1]||n[2]||n[3]||n[4]||n[5]||n[6],!a)continue;if(l=[...a].length,n[3]||n[4]){c+=l;continue}else if((n[5]||n[6])&&r%3&&!((r+l)%3)){d+=l;continue}if(c-=l,c>0)continue;l=Math.min(l,l+c+d);const u=[...n[0]][0].length,g=e.slice(0,r+n.index+u+l);if(Math.min(r,l)%2){const p=g.slice(1,-1);return{type:"em",raw:g,text:p,tokens:this.lexer.inlineTokens(p)}}const f=g.slice(2,-2);return{type:"strong",raw:g,text:f,tokens:this.lexer.inlineTokens(f)}}}}codespan(e){const t=this.rules.inline.code.exec(e);if(t){let i=t[2].replace(/\n/g," ");const n=/[^ ]/.test(i),o=/^ /.test(i)&&/ $/.test(i);return n&&o&&(i=i.substring(1,i.length-1)),i=zr(i,!0),{type:"codespan",raw:t[0],text:i}}}br(e){const t=this.rules.inline.br.exec(e);if(t)return{type:"br",raw:t[0]}}del(e){const t=this.rules.inline.del.exec(e);if(t)return{type:"del",raw:t[0],text:t[2],tokens:this.lexer.inlineTokens(t[2])}}autolink(e){const t=this.rules.inline.autolink.exec(e);if(t){let i,n;return t[2]==="@"?(i=zr(t[1]),n="mailto:"+i):(i=zr(t[1]),n=i),{type:"link",raw:t[0],text:i,href:n,tokens:[{type:"text",raw:i,text:i}]}}}url(e){let t;if(t=this.rules.inline.url.exec(e)){let i,n;if(t[2]==="@")i=zr(t[0]),n="mailto:"+i;else{let o;do o=t[0],t[0]=this.rules.inline._backpedal.exec(t[0])?.[0]??"";while(o!==t[0]);i=zr(t[0]),t[1]==="www."?n="http://"+t[0]:n=t[0]}return{type:"link",raw:t[0],text:i,href:n,tokens:[{type:"text",raw:i,text:i}]}}}inlineText(e){const t=this.rules.inline.text.exec(e);if(t){let i;return this.lexer.state.inRawBlock?i=t[0]:i=zr(t[0]),{type:"text",raw:t[0],text:i}}}}const D0e=/^(?: *(?:\n|$))+/,T0e=/^( {4}[^\n]+(?:\n(?: *(?:\n|$))*)?)+/,N0e=/^ {0,3}(`{3,}(?=[^`\n]*(?:\n|$))|~{3,})([^\n]*)(?:\n|$)(?:|([\s\S]*?)(?:\n|$))(?: {0,3}\1[~`]* *(?=\n|$)|$)/,rx=/^ {0,3}((?:-[\t ]*){3,}|(?:_[ \t]*){3,}|(?:\*[ \t]*){3,})(?:\n+|$)/,R0e=/^ {0,3}(#{1,6})(?=\s|$)(.*)(?:\n+|$)/,Pte=/(?:[*+-]|\d{1,9}[.)])/,Fte=Hi(/^(?!bull |blockCode|fences|blockquote|heading|html)((?:.|\n(?!\s*?\n|bull |blockCode|fences|blockquote|heading|html))+?)\n {0,3}(=+|-+) *(?:\n+|$)/).replace(/bull/g,Pte).replace(/blockCode/g,/ {4}/).replace(/fences/g,/ {0,3}(?:`{3,}|~{3,})/).replace(/blockquote/g,/ {0,3}>/).replace(/heading/g,/ {0,3}#{1,6}/).replace(/html/g,/ {0,3}<[^\n>]+>\n/).getRegex(),I7=/^([^\n]+(?:\n(?!hr|heading|lheading|blockquote|fences|list|html|table| +\n)[^\n]+)*)/,A0e=/^[^\n]+/,D7=/(?!\s*\])(?:\\.|[^\[\]\\])+/,M0e=Hi(/^ {0,3}\[(label)\]: *(?:\n *)?([^<\s][^\s]*|<.*?>)(?:(?: +(?:\n *)?| *\n *)(title))? *(?:\n+|$)/).replace("label",D7).replace("title",/(?:"(?:\\"?|[^"\\])*"|'[^'\n]*(?:\n[^'\n]+)*\n?'|\([^()]*\))/).getRegex(),O0e=Hi(/^( {0,3}bull)([ \t][^\n]+?)?(?:\n|$)/).replace(/bull/g,Pte).getRegex(),CA="address|article|aside|base|basefont|blockquote|body|caption|center|col|colgroup|dd|details|dialog|dir|div|dl|dt|fieldset|figcaption|figure|footer|form|frame|frameset|h[1-6]|head|header|hr|html|iframe|legend|li|link|main|menu|menuitem|meta|nav|noframes|ol|optgroup|option|p|param|search|section|summary|table|tbody|td|tfoot|th|thead|title|tr|track|ul",T7=/<!--(?:-?>|[\s\S]*?(?:-->|$))/,P0e=Hi("^ {0,3}(?:<(script|pre|style|textarea)[\\s>][\\s\\S]*?(?:</\\1>[^\\n]*\\n+|$)|comment[^\\n]*(\\n+|$)|<\\?[\\s\\S]*?(?:\\?>\\n*|$)|<![A-Z][\\s\\S]*?(?:>\\n*|$)|<!\\[CDATA\\[[\\s\\S]*?(?:\\]\\]>\\n*|$)|</?(tag)(?: +|\\n|/?>)[\\s\\S]*?(?:(?:\\n *)+\\n|$)|<(?!script|pre|style|textarea)([a-z][\\w-]*)(?:attribute)*? */?>(?=[ \\t]*(?:\\n|$))[\\s\\S]*?(?:(?:\\n *)+\\n|$)|</(?!script|pre|style|textarea)[a-z][\\w-]*\\s*>(?=[ \\t]*(?:\\n|$))[\\s\\S]*?(?:(?:\\n *)+\\n|$))","i").replace("comment",T7).replace("tag",CA).replace("attribute",/ +[a-zA-Z:_][\w.:-]*(?: *= *"[^"\n]*"| *= *'[^'\n]*'| *= *[^\s"'=<>`]+)?/).getRegex(),Bte=Hi(I7).replace("hr",rx).replace("heading"," {0,3}#{1,6}(?:\\s|$)").replace("|lheading","").replace("|table","").replace("blockquote"," {0,3}>").replace("fences"," {0,3}(?:`{3,}(?=[^`\\n]*\\n)|~{3,})[^\\n]*\\n").replace("list"," {0,3}(?:[*+-]|1[.)]) ").replace("html","</?(?:tag)(?: +|\\n|/?>)|<(?:script|pre|style|textarea|!--)").replace("tag",CA).getRegex(),F0e=Hi(/^( {0,3}> ?(paragraph|[^\n]*)(?:\n|$))+/).replace("paragraph",Bte).getRegex(),N7={blockquote:F0e,code:T0e,def:M0e,fences:N0e,heading:R0e,hr:rx,html:P0e,lheading:Fte,list:O0e,newline:D0e,paragraph:Bte,table:j1,text:A0e},mq=Hi("^ *([^\\n ].*)\\n {0,3}((?:\\| *)?:?-+:? *(?:\\| *:?-+:? *)*(?:\\| *)?)(?:\\n((?:(?! *\\n|hr|heading|blockquote|code|fences|list|html).*(?:\\n|$))*)\\n*|$)").replace("hr",rx).replace("heading"," {0,3}#{1,6}(?:\\s|$)").replace("blockquote"," {0,3}>").replace("code"," {4}[^\\n]").replace("fences"," {0,3}(?:`{3,}(?=[^`\\n]*\\n)|~{3,})[^\\n]*\\n").replace("list"," {0,3}(?:[*+-]|1[.)]) ").replace("html","</?(?:tag)(?: +|\\n|/?>)|<(?:script|pre|style|textarea|!--)").replace("tag",CA).getRegex(),B0e={...N7,table:mq,paragraph:Hi(I7).replace("hr",rx).replace("heading"," {0,3}#{1,6}(?:\\s|$)").replace("|lheading","").replace("table",mq).replace("blockquote"," {0,3}>").replace("fences"," {0,3}(?:`{3,}(?=[^`\\n]*\\n)|~{3,})[^\\n]*\\n").replace("list"," {0,3}(?:[*+-]|1[.)]) ").replace("html","</?(?:tag)(?: +|\\n|/?>)|<(?:script|pre|style|textarea|!--)").replace("tag",CA).getRegex()},W0e={...N7,html:Hi(`^ *(?:comment *(?:\\n|\\s*$)|<(tag)[\\s\\S]+?</\\1> *(?:\\n{2,}|\\s*$)|<tag(?:"[^"]*"|'[^']*'|\\s[^'"/>\\s]*)*?/?> *(?:\\n{2,}|\\s*$))`).replace("comment",T7).replace(/tag/g,"(?!(?:a|em|strong|small|s|cite|q|dfn|abbr|data|time|code|var|samp|kbd|sub|sup|i|b|u|mark|ruby|rt|rp|bdi|bdo|span|br|wbr|ins|del|img)\\b)\\w+(?!:|[^\\w\\s@]*@)\\b").getRegex(),def:/^ *\[([^\]]+)\]: *<?([^\s>]+)>?(?: +(["(][^\n]+[")]))? *(?:\n+|$)/,heading:/^(#{1,6})(.*)(?:\n+|$)/,fences:j1,lheading:/^(.+?)\n {0,3}(=+|-+) *(?:\n+|$)/,paragraph:Hi(I7).replace("hr",rx).replace("heading",` *#{1,6} *[^
]`).replace("lheading",Fte).replace("|table","").replace("blockquote"," {0,3}>").replace("|fences","").replace("|list","").replace("|html","").replace("|tag","").getRegex()},Wte=/^\\([!"#$%&'()*+,\-./:;<=>?@\[\]\\^_`{|}~])/,H0e=/^(`+)([^`]|[^`][\s\S]*?[^`])\1(?!`)/,Hte=/^( {2,}|\\)\n(?!\s*$)/,V0e=/^(`+|[^`])(?:(?= {2,}\n)|[\s\S]*?(?:(?=[\\<!\[`*_]|\b_|$)|[^ ](?= {2,}\n)))/,ax="\\p{P}\\p{S}",z0e=Hi(/^((?![*_])[\spunctuation])/,"u").replace(/punctuation/g,ax).getRegex(),$0e=/\[[^[\]]*?\]\([^\(\)]*?\)|`[^`]*?`|<[^<>]*?>/g,U0e=Hi(/^(?:\*+(?:((?!\*)[punct])|[^\s*]))|^_+(?:((?!_)[punct])|([^\s_]))/,"u").replace(/punct/g,ax).getRegex(),q0e=Hi("^[^_*]*?__[^_*]*?\\*[^_*]*?(?=__)|[^*]+(?=[^*])|(?!\\*)[punct](\\*+)(?=[\\s]|$)|[^punct\\s](\\*+)(?!\\*)(?=[punct\\s]|$)|(?!\\*)[punct\\s](\\*+)(?=[^punct\\s])|[\\s](\\*+)(?!\\*)(?=[punct])|(?!\\*)[punct](\\*+)(?!\\*)(?=[punct])|[^punct\\s](\\*+)(?=[^punct\\s])","gu").replace(/punct/g,ax).getRegex(),j0e=Hi("^[^_*]*?\\*\\*[^_*]*?_[^_*]*?(?=\\*\\*)|[^_]+(?=[^_])|(?!_)[punct](_+)(?=[\\s]|$)|[^punct\\s](_+)(?!_)(?=[punct\\s]|$)|(?!_)[punct\\s](_+)(?=[^punct\\s])|[\\s](_+)(?!_)(?=[punct])|(?!_)[punct](_+)(?!_)(?=[punct])","gu").replace(/punct/g,ax).getRegex(),G0e=Hi(/\\([punct])/,"gu").replace(/punct/g,ax).getRegex(),K0e=Hi(/^<(scheme:[^\s\x00-\x1f<>]*|email)>/).replace("scheme",/[a-zA-Z][a-zA-Z0-9+.-]{1,31}/).replace("email",/[a-zA-Z0-9.!#$%&'*+/=?^_`{|}~-]+(@)[a-zA-Z0-9](?:[a-zA-Z0-9-]{0,61}[a-zA-Z0-9])?(?:\.[a-zA-Z0-9](?:[a-zA-Z0-9-]{0,61}[a-zA-Z0-9])?)+(?![-_])/).getRegex(),Y0e=Hi(T7).replace("(?:-->|$)","-->").getRegex(),Z0e=Hi("^comment|^</[a-zA-Z][\\w:-]*\\s*>|^<[a-zA-Z][\\w-]*(?:attribute)*?\\s*/?>|^<\\?[\\s\\S]*?\\?>|^<![a-zA-Z]+\\s[\\s\\S]*?>|^<!\\[CDATA\\[[\\s\\S]*?\\]\\]>").replace("comment",Y0e).replace("attribute",/\s+[a-zA-Z:_][\w.:-]*(?:\s*=\s*"[^"]*"|\s*=\s*'[^']*'|\s*=\s*[^\s"'=<>`]+)?/).getRegex(),wI=/(?:\[(?:\\.|[^\[\]\\])*\]|\\.|`[^`]*`|[^\[\]\\`])*?/,X0e=Hi(/^!?\[(label)\]\(\s*(href)(?:\s+(title))?\s*\)/).replace("label",wI).replace("href",/<(?:\\.|[^\n<>\\])+>|[^\s\x00-\x1f]*/).replace("title",/"(?:\\"?|[^"\\])*"|'(?:\\'?|[^'\\])*'|\((?:\\\)?|[^)\\])*\)/).getRegex(),Vte=Hi(/^!?\[(label)\]\[(ref)\]/).replace("label",wI).replace("ref",D7).getRegex(),zte=Hi(/^!?\[(ref)\](?:\[\])?/).replace("ref",D7).getRegex(),Q0e=Hi("reflink|nolink(?!\\()","g").replace("reflink",Vte).replace("nolink",zte).getRegex(),R7={_backpedal:j1,anyPunctuation:G0e,autolink:K0e,blockSkip:$0e,br:Hte,code:H0e,del:j1,emStrongLDelim:U0e,emStrongRDelimAst:q0e,emStrongRDelimUnd:j0e,escape:Wte,link:X0e,nolink:zte,punctuation:z0e,reflink:Vte,reflinkSearch:Q0e,tag:Z0e,text:V0e,url:j1},J0e={...R7,link:Hi(/^!?\[(label)\]\((.*?)\)/).replace("label",wI).getRegex(),reflink:Hi(/^!?\[(label)\]\s*\[([^\]]*)\]/).replace("label",wI).getRegex()},QF={...R7,escape:Hi(Wte).replace("])","~|])").getRegex(),url:Hi(/^((?:ftp|https?):\/\/|www\.)(?:[a-zA-Z0-9\-]+\.?)+[^\s<]*|^email/,"i").replace("email",/[A-Za-z0-9._+-]+(@)[a-zA-Z0-9-_]+(?:\.[a-zA-Z0-9-_]*[a-zA-Z0-9])+(?![-_])/).getRegex(),_backpedal:/(?:[^?!.,:;*_'"~()&]+|\([^)]*\)|&(?![a-zA-Z0-9]+;$)|[?!.,:;*_'"~)]+(?!$))+/,del:/^(~~?)(?=[^\s~])([\s\S]*?[^\s~])\1(?=[^~]|$)/,text:/^([`~]+|[^`~])(?:(?= {2,}\n)|(?=[a-zA-Z0-9.!#$%&'*+\/=?_`{\|}~-]+@)|[\s\S]*?(?:(?=[\\<!\[`*~_]|\b_|https?:\/\/|ftp:\/\/|www\.|$)|[^ ](?= {2,}\n)|[^a-zA-Z0-9.!#$%&'*+\/=?_`{\|}~-](?=[a-zA-Z0-9.!#$%&'*+\/=?_`{\|}~-]+@)))/},exe={...QF,br:Hi(Hte).replace("{2,}","*").getRegex(),text:Hi(QF.text).replace("\\b_","\\b_| {2,}\\n").replace(/\{2,\}/g,"*").getRegex()},bL={normal:N7,gfm:B0e,pedantic:W0e},MC={normal:R7,gfm:QF,breaks:exe,pedantic:J0e};class Cc{tokens;options;state;tokenizer;inlineQueue;constructor(e){this.tokens=[],this.tokens.links=Object.create(null),this.options=e||u_,this.options.tokenizer=this.options.tokenizer||new bI,this.tokenizer=this.options.tokenizer,this.tokenizer.options=this.options,this.tokenizer.lexer=this,this.inlineQueue=[],this.state={inLink:!1,inRawBlock:!1,top:!0};const t={block:bL.normal,inline:MC.normal};this.options.pedantic?(t.block=bL.pedantic,t.inline=MC.pedantic):this.options.gfm&&(t.block=bL.gfm,this.options.breaks?t.inline=MC.breaks:t.inline=MC.gfm),this.tokenizer.rules=t}static get rules(){return{block:bL,inline:MC}}static lex(e,t){return new Cc(t).lex(e)}static lexInline(e,t){return new Cc(t).inlineTokens(e)}lex(e){e=e.replace(/\r\n|\r/g,`
`),this.blockTokens(e,this.tokens);for(let t=0;t<this.inlineQueue.length;t++){const i=this.inlineQueue[t];this.inlineTokens(i.src,i.tokens)}return this.inlineQueue=[],this.tokens}blockTokens(e,t=[],i=!1){this.options.pedantic?e=e.replace(/\t/g,"    ").replace(/^ +$/gm,""):e=e.replace(/^( *)(\t+)/gm,(a,l,c)=>l+"    ".repeat(c.length));let n,o,r;for(;e;)if(!(this.options.extensions&&this.options.extensions.block&&this.options.extensions.block.some(a=>(n=a.call({lexer:this},e,t))?(e=e.substring(n.raw.length),t.push(n),!0):!1))){if(n=this.tokenizer.space(e)){e=e.substring(n.raw.length),n.raw.length===1&&t.length>0?t[t.length-1].raw+=`
`:t.push(n);continue}if(n=this.tokenizer.code(e)){e=e.substring(n.raw.length),o=t[t.length-1],o&&(o.type==="paragraph"||o.type==="text")?(o.raw+=`
`+n.raw,o.text+=`
`+n.text,this.inlineQueue[this.inlineQueue.length-1].src=o.text):t.push(n);continue}if(n=this.tokenizer.fences(e)){e=e.substring(n.raw.length),t.push(n);continue}if(n=this.tokenizer.heading(e)){e=e.substring(n.raw.length),t.push(n);continue}if(n=this.tokenizer.hr(e)){e=e.substring(n.raw.length),t.push(n);continue}if(n=this.tokenizer.blockquote(e)){e=e.substring(n.raw.length),t.push(n);continue}if(n=this.tokenizer.list(e)){e=e.substring(n.raw.length),t.push(n);continue}if(n=this.tokenizer.html(e)){e=e.substring(n.raw.length),t.push(n);continue}if(n=this.tokenizer.def(e)){e=e.substring(n.raw.length),o=t[t.length-1],o&&(o.type==="paragraph"||o.type==="text")?(o.raw+=`
`+n.raw,o.text+=`
`+n.raw,this.inlineQueue[this.inlineQueue.length-1].src=o.text):this.tokens.links[n.tag]||(this.tokens.links[n.tag]={href:n.href,title:n.title});continue}if(n=this.tokenizer.table(e)){e=e.substring(n.raw.length),t.push(n);continue}if(n=this.tokenizer.lheading(e)){e=e.substring(n.raw.length),t.push(n);continue}if(r=e,this.options.extensions&&this.options.extensions.startBlock){let a=1/0;const l=e.slice(1);let c;this.options.extensions.startBlock.forEach(d=>{c=d.call({lexer:this},l),typeof c=="number"&&c>=0&&(a=Math.min(a,c))}),a<1/0&&a>=0&&(r=e.substring(0,a+1))}if(this.state.top&&(n=this.tokenizer.paragraph(r))){o=t[t.length-1],i&&o?.type==="paragraph"?(o.raw+=`
`+n.raw,o.text+=`
`+n.text,this.inlineQueue.pop(),this.inlineQueue[this.inlineQueue.length-1].src=o.text):t.push(n),i=r.length!==e.length,e=e.substring(n.raw.length);continue}if(n=this.tokenizer.text(e)){e=e.substring(n.raw.length),o=t[t.length-1],o&&o.type==="text"?(o.raw+=`
`+n.raw,o.text+=`
`+n.text,this.inlineQueue.pop(),this.inlineQueue[this.inlineQueue.length-1].src=o.text):t.push(n);continue}if(e){const a="Infinite loop on byte: "+e.charCodeAt(0);if(this.options.silent){console.error(a);break}else throw new Error(a)}}return this.state.top=!0,t}inline(e,t=[]){return this.inlineQueue.push({src:e,tokens:t}),t}inlineTokens(e,t=[]){let i,n,o,r=e,a,l,c;if(this.tokens.links){const d=Object.keys(this.tokens.links);if(d.length>0)for(;(a=this.tokenizer.rules.inline.reflinkSearch.exec(r))!=null;)d.includes(a[0].slice(a[0].lastIndexOf("[")+1,-1))&&(r=r.slice(0,a.index)+"["+"a".repeat(a[0].length-2)+"]"+r.slice(this.tokenizer.rules.inline.reflinkSearch.lastIndex))}for(;(a=this.tokenizer.rules.inline.blockSkip.exec(r))!=null;)r=r.slice(0,a.index)+"["+"a".repeat(a[0].length-2)+"]"+r.slice(this.tokenizer.rules.inline.blockSkip.lastIndex);for(;(a=this.tokenizer.rules.inline.anyPunctuation.exec(r))!=null;)r=r.slice(0,a.index)+"++"+r.slice(this.tokenizer.rules.inline.anyPunctuation.lastIndex);for(;e;)if(l||(c=""),l=!1,!(this.options.extensions&&this.options.extensions.inline&&this.options.extensions.inline.some(d=>(i=d.call({lexer:this},e,t))?(e=e.substring(i.raw.length),t.push(i),!0):!1))){if(i=this.tokenizer.escape(e)){e=e.substring(i.raw.length),t.push(i);continue}if(i=this.tokenizer.tag(e)){e=e.substring(i.raw.length),n=t[t.length-1],n&&i.type==="text"&&n.type==="text"?(n.raw+=i.raw,n.text+=i.text):t.push(i);continue}if(i=this.tokenizer.link(e)){e=e.substring(i.raw.length),t.push(i);continue}if(i=this.tokenizer.reflink(e,this.tokens.links)){e=e.substring(i.raw.length),n=t[t.length-1],n&&i.type==="text"&&n.type==="text"?(n.raw+=i.raw,n.text+=i.text):t.push(i);continue}if(i=this.tokenizer.emStrong(e,r,c)){e=e.substring(i.raw.length),t.push(i);continue}if(i=this.tokenizer.codespan(e)){e=e.substring(i.raw.length),t.push(i);continue}if(i=this.tokenizer.br(e)){e=e.substring(i.raw.length),t.push(i);continue}if(i=this.tokenizer.del(e)){e=e.substring(i.raw.length),t.push(i);continue}if(i=this.tokenizer.autolink(e)){e=e.substring(i.raw.length),t.push(i);continue}if(!this.state.inLink&&(i=this.tokenizer.url(e))){e=e.substring(i.raw.length),t.push(i);continue}if(o=e,this.options.extensions&&this.options.extensions.startInline){let d=1/0;const h=e.slice(1);let u;this.options.extensions.startInline.forEach(g=>{u=g.call({lexer:this},h),typeof u=="number"&&u>=0&&(d=Math.min(d,u))}),d<1/0&&d>=0&&(o=e.substring(0,d+1))}if(i=this.tokenizer.inlineText(o)){e=e.substring(i.raw.length),i.raw.slice(-1)!=="_"&&(c=i.raw.slice(-1)),l=!0,n=t[t.length-1],n&&n.type==="text"?(n.raw+=i.raw,n.text+=i.text):t.push(i);continue}if(e){const d="Infinite loop on byte: "+e.charCodeAt(0);if(this.options.silent){console.error(d);break}else throw new Error(d)}}return t}}class NS{options;parser;constructor(e){this.options=e||u_}space(e){return""}code({text:e,lang:t,escaped:i}){const n=(t||"").match(/^\S*/)?.[0],o=e.replace(/\n$/,"")+`
`;return n?'<pre><code class="language-'+zr(n)+'">'+(i?o:zr(o,!0))+`</code></pre>
`:"<pre><code>"+(i?o:zr(o,!0))+`</code></pre>
`}blockquote({tokens:e}){return`<blockquote>
${this.parser.parse(e)}</blockquote>
`}html({text:e}){return e}heading({tokens:e,depth:t}){return`<h${t}>${this.parser.parseInline(e)}</h${t}>
`}hr(e){return`<hr>
`}list(e){const t=e.ordered,i=e.start;let n="";for(let a=0;a<e.items.length;a++){const l=e.items[a];n+=this.listitem(l)}const o=t?"ol":"ul",r=t&&i!==1?' start="'+i+'"':"";return"<"+o+r+`>
`+n+"</"+o+`>
`}listitem(e){let t="";if(e.task){const i=this.checkbox({checked:!!e.checked});e.loose?e.tokens.length>0&&e.tokens[0].type==="paragraph"?(e.tokens[0].text=i+" "+e.tokens[0].text,e.tokens[0].tokens&&e.tokens[0].tokens.length>0&&e.tokens[0].tokens[0].type==="text"&&(e.tokens[0].tokens[0].text=i+" "+e.tokens[0].tokens[0].text)):e.tokens.unshift({type:"text",raw:i+" ",text:i+" "}):t+=i+" "}return t+=this.parser.parse(e.tokens,!!e.loose),`<li>${t}</li>
`}checkbox({checked:e}){return"<input "+(e?'checked="" ':"")+'disabled="" type="checkbox">'}paragraph({tokens:e}){return`<p>${this.parser.parseInline(e)}</p>
`}table(e){let t="",i="";for(let o=0;o<e.header.length;o++)i+=this.tablecell(e.header[o]);t+=this.tablerow({text:i});let n="";for(let o=0;o<e.rows.length;o++){const r=e.rows[o];i="";for(let a=0;a<r.length;a++)i+=this.tablecell(r[a]);n+=this.tablerow({text:i})}return n&&(n=`<tbody>${n}</tbody>`),`<table>
<thead>
`+t+`</thead>
`+n+`</table>
`}tablerow({text:e}){return`<tr>
${e}</tr>
`}tablecell(e){const t=this.parser.parseInline(e.tokens),i=e.header?"th":"td";return(e.align?`<${i} align="${e.align}">`:`<${i}>`)+t+`</${i}>
`}strong({tokens:e}){return`<strong>${this.parser.parseInline(e)}</strong>`}em({tokens:e}){return`<em>${this.parser.parseInline(e)}</em>`}codespan({text:e}){return`<code>${e}</code>`}br(e){return"<br>"}del({tokens:e}){return`<del>${this.parser.parseInline(e)}</del>`}link({href:e,title:t,tokens:i}){const n=this.parser.parseInline(i),o=gq(e);if(o===null)return n;e=o;let r='<a href="'+e+'"';return t&&(r+=' title="'+t+'"'),r+=">"+n+"</a>",r}image({href:e,title:t,text:i}){const n=gq(e);if(n===null)return i;e=n;let o=`<img src="${e}" alt="${i}"`;return t&&(o+=` title="${t}"`),o+=">",o}text(e){return"tokens"in e&&e.tokens?this.parser.parseInline(e.tokens):e.text}}class A7{strong({text:e}){return e}em({text:e}){return e}codespan({text:e}){return e}del({text:e}){return e}html({text:e}){return e}text({text:e}){return e}link({text:e}){return""+e}image({text:e}){return""+e}br(){return""}}class yc{options;renderer;textRenderer;constructor(e){this.options=e||u_,this.options.renderer=this.options.renderer||new NS,this.renderer=this.options.renderer,this.renderer.options=this.options,this.renderer.parser=this,this.textRenderer=new A7}static parse(e,t){return new yc(t).parse(e)}static parseInline(e,t){return new yc(t).parseInline(e)}parse(e,t=!0){let i="";for(let n=0;n<e.length;n++){const o=e[n];if(this.options.extensions&&this.options.extensions.renderers&&this.options.extensions.renderers[o.type]){const a=o,l=this.options.extensions.renderers[a.type].call({parser:this},a);if(l!==!1||!["space","hr","heading","code","table","blockquote","list","html","paragraph","text"].includes(a.type)){i+=l||"";continue}}const r=o;switch(r.type){case"space":{i+=this.renderer.space(r);continue}case"hr":{i+=this.renderer.hr(r);continue}case"heading":{i+=this.renderer.heading(r);continue}case"code":{i+=this.renderer.code(r);continue}case"table":{i+=this.renderer.table(r);continue}case"blockquote":{i+=this.renderer.blockquote(r);continue}case"list":{i+=this.renderer.list(r);continue}case"html":{i+=this.renderer.html(r);continue}case"paragraph":{i+=this.renderer.paragraph(r);continue}case"text":{let a=r,l=this.renderer.text(a);for(;n+1<e.length&&e[n+1].type==="text";)a=e[++n],l+=`
`+this.renderer.text(a);t?i+=this.renderer.paragraph({type:"paragraph",raw:l,text:l,tokens:[{type:"text",raw:l,text:l}]}):i+=l;continue}default:{const a='Token with "'+r.type+'" type was not found.';if(this.options.silent)return console.error(a),"";throw new Error(a)}}}return i}parseInline(e,t){t=t||this.renderer;let i="";for(let n=0;n<e.length;n++){const o=e[n];if(this.options.extensions&&this.options.extensions.renderers&&this.options.extensions.renderers[o.type]){const a=this.options.extensions.renderers[o.type].call({parser:this},o);if(a!==!1||!["escape","html","link","image","strong","em","codespan","br","del","text"].includes(o.type)){i+=a||"";continue}}const r=o;switch(r.type){case"escape":{i+=t.text(r);break}case"html":{i+=t.html(r);break}case"link":{i+=t.link(r);break}case"image":{i+=t.image(r);break}case"strong":{i+=t.strong(r);break}case"em":{i+=t.em(r);break}case"codespan":{i+=t.codespan(r);break}case"br":{i+=t.br(r);break}case"del":{i+=t.del(r);break}case"text":{i+=t.text(r);break}default:{const a='Token with "'+r.type+'" type was not found.';if(this.options.silent)return console.error(a),"";throw new Error(a)}}}return i}}class Pk{options;constructor(e){this.options=e||u_}static passThroughHooks=new Set(["preprocess","postprocess","processAllTokens"]);preprocess(e){return e}postprocess(e){return e}processAllTokens(e){return e}}class $te{defaults=E7();options=this.setOptions;parse=this.parseMarkdown(Cc.lex,yc.parse);parseInline=this.parseMarkdown(Cc.lexInline,yc.parseInline);Parser=yc;Renderer=NS;TextRenderer=A7;Lexer=Cc;Tokenizer=bI;Hooks=Pk;constructor(...e){this.use(...e)}walkTokens(e,t){let i=[];for(const n of e)switch(i=i.concat(t.call(this,n)),n.type){case"table":{const o=n;for(const r of o.header)i=i.concat(this.walkTokens(r.tokens,t));for(const r of o.rows)for(const a of r)i=i.concat(this.walkTokens(a.tokens,t));break}case"list":{const o=n;i=i.concat(this.walkTokens(o.items,t));break}default:{const o=n;this.defaults.extensions?.childTokens?.[o.type]?this.defaults.extensions.childTokens[o.type].forEach(r=>{const a=o[r].flat(1/0);i=i.concat(this.walkTokens(a,t))}):o.tokens&&(i=i.concat(this.walkTokens(o.tokens,t)))}}return i}use(...e){const t=this.defaults.extensions||{renderers:{},childTokens:{}};return e.forEach(i=>{const n={...i};if(n.async=this.defaults.async||n.async||!1,i.extensions&&(i.extensions.forEach(o=>{if(!o.name)throw new Error("extension name required");if("renderer"in o){const r=t.renderers[o.name];r?t.renderers[o.name]=function(...a){let l=o.renderer.apply(this,a);return l===!1&&(l=r.apply(this,a)),l}:t.renderers[o.name]=o.renderer}if("tokenizer"in o){if(!o.level||o.level!=="block"&&o.level!=="inline")throw new Error("extension level must be 'block' or 'inline'");const r=t[o.level];r?r.unshift(o.tokenizer):t[o.level]=[o.tokenizer],o.start&&(o.level==="block"?t.startBlock?t.startBlock.push(o.start):t.startBlock=[o.start]:o.level==="inline"&&(t.startInline?t.startInline.push(o.start):t.startInline=[o.start]))}"childTokens"in o&&o.childTokens&&(t.childTokens[o.name]=o.childTokens)}),n.extensions=t),i.renderer){const o=this.defaults.renderer||new NS(this.defaults);for(const r in i.renderer){if(!(r in o))throw new Error(`renderer '${r}' does not exist`);if(["options","parser"].includes(r))continue;const a=r,l=i.renderer[a],c=o[a];o[a]=(...d)=>{let h=l.apply(o,d);return h===!1&&(h=c.apply(o,d)),h||""}}n.renderer=o}if(i.tokenizer){const o=this.defaults.tokenizer||new bI(this.defaults);for(const r in i.tokenizer){if(!(r in o))throw new Error(`tokenizer '${r}' does not exist`);if(["options","rules","lexer"].includes(r))continue;const a=r,l=i.tokenizer[a],c=o[a];o[a]=(...d)=>{let h=l.apply(o,d);return h===!1&&(h=c.apply(o,d)),h}}n.tokenizer=o}if(i.hooks){const o=this.defaults.hooks||new Pk;for(const r in i.hooks){if(!(r in o))throw new Error(`hook '${r}' does not exist`);if(r==="options")continue;const a=r,l=i.hooks[a],c=o[a];Pk.passThroughHooks.has(r)?o[a]=d=>{if(this.defaults.async)return Promise.resolve(l.call(o,d)).then(u=>c.call(o,u));const h=l.call(o,d);return c.call(o,h)}:o[a]=(...d)=>{let h=l.apply(o,d);return h===!1&&(h=c.apply(o,d)),h}}n.hooks=o}if(i.walkTokens){const o=this.defaults.walkTokens,r=i.walkTokens;n.walkTokens=function(a){let l=[];return l.push(r.call(this,a)),o&&(l=l.concat(o.call(this,a))),l}}this.defaults={...this.defaults,...n}}),this}setOptions(e){return this.defaults={...this.defaults,...e},this}lexer(e,t){return Cc.lex(e,t??this.defaults)}parser(e,t){return yc.parse(e,t??this.defaults)}parseMarkdown(e,t){return(n,o)=>{const r={...o},a={...this.defaults,...r},l=this.onError(!!a.silent,!!a.async);if(this.defaults.async===!0&&r.async===!1)return l(new Error("marked(): The async option was set to true by an extension. Remove async: false from the parse options object to return a Promise."));if(typeof n>"u"||n===null)return l(new Error("marked(): input parameter is undefined or null"));if(typeof n!="string")return l(new Error("marked(): input parameter is of type "+Object.prototype.toString.call(n)+", string expected"));if(a.hooks&&(a.hooks.options=a),a.async)return Promise.resolve(a.hooks?a.hooks.preprocess(n):n).then(c=>e(c,a)).then(c=>a.hooks?a.hooks.processAllTokens(c):c).then(c=>a.walkTokens?Promise.all(this.walkTokens(c,a.walkTokens)).then(()=>c):c).then(c=>t(c,a)).then(c=>a.hooks?a.hooks.postprocess(c):c).catch(l);try{a.hooks&&(n=a.hooks.preprocess(n));let c=e(n,a);a.hooks&&(c=a.hooks.processAllTokens(c)),a.walkTokens&&this.walkTokens(c,a.walkTokens);let d=t(c,a);return a.hooks&&(d=a.hooks.postprocess(d)),d}catch(c){return l(c)}}}onError(e,t){return i=>{if(i.message+=`
Please report this to https://github.com/markedjs/marked.`,e){const n="<p>An error occurred:</p><pre>"+zr(i.message+"",!0)+"</pre>";return t?Promise.resolve(n):n}if(t)return Promise.reject(i);throw i}}}const Pm=new $te;function Ni(s,e){return Pm.parse(s,e)}Ni.options=Ni.setOptions=function(s){return Pm.setOptions(s),Ni.defaults=Pm.defaults,Ate(Ni.defaults),Ni};Ni.getDefaults=E7;Ni.defaults=u_;Ni.use=function(...s){return Pm.use(...s),Ni.defaults=Pm.defaults,Ate(Ni.defaults),Ni};Ni.walkTokens=function(s,e){return Pm.walkTokens(s,e)};Ni.parseInline=Pm.parseInline;Ni.Parser=yc;Ni.parser=yc.parse;Ni.Renderer=NS;Ni.TextRenderer=A7;Ni.Lexer=Cc;Ni.lexer=Cc.lex;Ni.Tokenizer=bI;Ni.Hooks=Pk;Ni.parse=Ni;Ni.options;Ni.setOptions;Ni.use;Ni.walkTokens;Ni.parseInline;const txe=Ni;yc.parse;const RS=Cc.lex;function ixe(s){return JSON.stringify(s,nxe)}function JF(s){let e=JSON.parse(s);return e=e4(e),e}function nxe(s,e){return e instanceof RegExp?{$mid:2,source:e.source,flags:e.flags}:e}function e4(s,e=0){if(!s||e>200)return s;if(typeof s=="object"){switch(s.$mid){case 1:return Le.revive(s);case 2:return new RegExp(s.source,s.flags);case 17:return new Date(s.source)}if(s instanceof Ru||s instanceof Uint8Array)return s;if(Array.isArray(s))for(let t=0;t<s.length;++t)s[t]=e4(s[t],e+1);else for(const t in s)Object.hasOwnProperty.call(s,t)&&(s[t]=e4(s[t],e+1))}return s}const{entries:Ute,setPrototypeOf:_q,isFrozen:sxe,getPrototypeOf:oxe,getOwnPropertyDescriptor:rxe}=Object;let{freeze:jo,seal:Ba,create:qte}=Object,{apply:t4,construct:i4}=typeof Reflect<"u"&&Reflect;jo||(jo=function(e){return e});Ba||(Ba=function(e){return e});t4||(t4=function(e,t){for(var i=arguments.length,n=new Array(i>2?i-2:0),o=2;o<i;o++)n[o-2]=arguments[o];return e.apply(t,n)});i4||(i4=function(e){for(var t=arguments.length,i=new Array(t>1?t-1:0),n=1;n<t;n++)i[n-1]=arguments[n];return new e(...i)});const wL=Go(Array.prototype.forEach),axe=Go(Array.prototype.lastIndexOf),bq=Go(Array.prototype.pop),OC=Go(Array.prototype.push),lxe=Go(Array.prototype.splice),Fk=Go(String.prototype.toLowerCase),IM=Go(String.prototype.toString),DM=Go(String.prototype.match),PC=Go(String.prototype.replace),cxe=Go(String.prototype.indexOf),dxe=Go(String.prototype.trim),Xa=Go(Object.prototype.hasOwnProperty),Io=Go(RegExp.prototype.test),FC=hxe(TypeError);function Go(s){return function(e){e instanceof RegExp&&(e.lastIndex=0);for(var t=arguments.length,i=new Array(t>1?t-1:0),n=1;n<t;n++)i[n-1]=arguments[n];return t4(s,e,i)}}function hxe(s){return function(){for(var e=arguments.length,t=new Array(e),i=0;i<e;i++)t[i]=arguments[i];return i4(s,t)}}function Wt(s,e){let t=arguments.length>2&&arguments[2]!==void 0?arguments[2]:Fk;_q&&_q(s,null);let i=e.length;for(;i--;){let n=e[i];if(typeof n=="string"){const o=t(n);o!==n&&(sxe(e)||(e[i]=o),n=o)}s[n]=!0}return s}function uxe(s){for(let e=0;e<s.length;e++)Xa(s,e)||(s[e]=null);return s}function pd(s){const e=qte(null);for(const[t,i]of Ute(s))Xa(s,t)&&(Array.isArray(i)?e[t]=uxe(i):i&&typeof i=="object"&&i.constructor===Object?e[t]=pd(i):e[t]=i);return e}function BC(s,e){for(;s!==null;){const i=rxe(s,e);if(i){if(i.get)return Go(i.get);if(typeof i.value=="function")return Go(i.value)}s=oxe(s)}function t(){return null}return t}const wq=jo(["a","abbr","acronym","address","area","article","aside","audio","b","bdi","bdo","big","blink","blockquote","body","br","button","canvas","caption","center","cite","code","col","colgroup","content","data","datalist","dd","decorator","del","details","dfn","dialog","dir","div","dl","dt","element","em","fieldset","figcaption","figure","font","footer","form","h1","h2","h3","h4","h5","h6","head","header","hgroup","hr","html","i","img","input","ins","kbd","label","legend","li","main","map","mark","marquee","menu","menuitem","meter","nav","nobr","ol","optgroup","option","output","p","picture","pre","progress","q","rp","rt","ruby","s","samp","search","section","select","shadow","slot","small","source","spacer","span","strike","strong","style","sub","summary","sup","table","tbody","td","template","textarea","tfoot","th","thead","time","tr","track","tt","u","ul","var","video","wbr"]),TM=jo(["svg","a","altglyph","altglyphdef","altglyphitem","animatecolor","animatemotion","animatetransform","circle","clippath","defs","desc","ellipse","enterkeyhint","exportparts","filter","font","g","glyph","glyphref","hkern","image","inputmode","line","lineargradient","marker","mask","metadata","mpath","part","path","pattern","polygon","polyline","radialgradient","rect","slot","stop","style","switch","symbol","text","textpath","title","tref","tspan","view","vkern"]),NM=jo(["feBlend","feColorMatrix","feComponentTransfer","feComposite","feConvolveMatrix","feDiffuseLighting","feDisplacementMap","feDistantLight","feDropShadow","feFlood","feFuncA","feFuncB","feFuncG","feFuncR","feGaussianBlur","feImage","feMerge","feMergeNode","feMorphology","feOffset","fePointLight","feSpecularLighting","feSpotLight","feTile","feTurbulence"]),gxe=jo(["animate","color-profile","cursor","discard","font-face","font-face-format","font-face-name","font-face-src","font-face-uri","foreignobject","hatch","hatchpath","mesh","meshgradient","meshpatch","meshrow","missing-glyph","script","set","solidcolor","unknown","use"]),RM=jo(["math","menclose","merror","mfenced","mfrac","mglyph","mi","mlabeledtr","mmultiscripts","mn","mo","mover","mpadded","mphantom","mroot","mrow","ms","mspace","msqrt","mstyle","msub","msup","msubsup","mtable","mtd","mtext","mtr","munder","munderover","mprescripts"]),fxe=jo(["maction","maligngroup","malignmark","mlongdiv","mscarries","mscarry","msgroup","mstack","msline","msrow","semantics","annotation","annotation-xml","mprescripts","none"]),vq=jo(["#text"]),Cq=jo(["accept","action","align","alt","autocapitalize","autocomplete","autopictureinpicture","autoplay","background","bgcolor","border","capture","cellpadding","cellspacing","checked","cite","class","clear","color","cols","colspan","controls","controlslist","coords","crossorigin","datetime","decoding","default","dir","disabled","disablepictureinpicture","disableremoteplayback","download","draggable","enctype","enterkeyhint","exportparts","face","for","headers","height","hidden","high","href","hreflang","id","inert","inputmode","integrity","ismap","kind","label","lang","list","loading","loop","low","max","maxlength","media","method","min","minlength","multiple","muted","name","nonce","noshade","novalidate","nowrap","open","optimum","part","pattern","placeholder","playsinline","popover","popovertarget","popovertargetaction","poster","preload","pubdate","radiogroup","readonly","rel","required","rev","reversed","role","rows","rowspan","spellcheck","scope","selected","shape","size","sizes","slot","span","srclang","start","src","srcset","step","style","summary","tabindex","title","translate","type","usemap","valign","value","width","wrap","xmlns","slot"]),AM=jo(["accent-height","accumulate","additive","alignment-baseline","amplitude","ascent","attributename","attributetype","azimuth","basefrequency","baseline-shift","begin","bias","by","class","clip","clippathunits","clip-path","clip-rule","color","color-interpolation","color-interpolation-filters","color-profile","color-rendering","cx","cy","d","dx","dy","diffuseconstant","direction","display","divisor","dur","edgemode","elevation","end","exponent","fill","fill-opacity","fill-rule","filter","filterunits","flood-color","flood-opacity","font-family","font-size","font-size-adjust","font-stretch","font-style","font-variant","font-weight","fx","fy","g1","g2","glyph-name","glyphref","gradientunits","gradienttransform","height","href","id","image-rendering","in","in2","intercept","k","k1","k2","k3","k4","kerning","keypoints","keysplines","keytimes","lang","lengthadjust","letter-spacing","kernelmatrix","kernelunitlength","lighting-color","local","marker-end","marker-mid","marker-start","markerheight","markerunits","markerwidth","maskcontentunits","maskunits","max","mask","media","method","mode","min","name","numoctaves","offset","operator","opacity","order","orient","orientation","origin","overflow","paint-order","path","pathlength","patterncontentunits","patterntransform","patternunits","points","preservealpha","preserveaspectratio","primitiveunits","r","rx","ry","radius","refx","refy","repeatcount","repeatdur","restart","result","rotate","scale","seed","shape-rendering","slope","specularconstant","specularexponent","spreadmethod","startoffset","stddeviation","stitchtiles","stop-color","stop-opacity","stroke-dasharray","stroke-dashoffset","stroke-linecap","stroke-linejoin","stroke-miterlimit","stroke-opacity","stroke","stroke-width","style","surfacescale","systemlanguage","tabindex","tablevalues","targetx","targety","transform","transform-origin","text-anchor","text-decoration","text-rendering","textlength","type","u1","u2","unicode","values","viewbox","visibility","version","vert-adv-y","vert-origin-x","vert-origin-y","width","word-spacing","wrap","writing-mode","xchannelselector","ychannelselector","x","x1","x2","xmlns","y","y1","y2","z","zoomandpan"]),yq=jo(["accent","accentunder","align","bevelled","close","columnsalign","columnlines","columnspan","denomalign","depth","dir","display","displaystyle","encoding","fence","frame","height","href","id","largeop","length","linethickness","lspace","lquote","mathbackground","mathcolor","mathsize","mathvariant","maxsize","minsize","movablelimits","notation","numalign","open","rowalign","rowlines","rowspacing","rowspan","rspace","rquote","scriptlevel","scriptminsize","scriptsizemultiplier","selection","separator","separators","stretchy","subscriptshift","supscriptshift","symmetric","voffset","width","xmlns"]),vL=jo(["xlink:href","xml:id","xlink:title","xml:space","xmlns:xlink"]),pxe=Ba(/\{\{[\w\W]*|[\w\W]*\}\}/gm),mxe=Ba(/<%[\w\W]*|[\w\W]*%>/gm),_xe=Ba(/\$\{[\w\W]*/gm),bxe=Ba(/^data-[\-\w.\u00B7-\uFFFF]+$/),wxe=Ba(/^aria-[\-\w]+$/),jte=Ba(/^(?:(?:(?:f|ht)tps?|mailto|tel|callto|sms|cid|xmpp|matrix):|[^a-z]|[a-z+.\-]+(?:[^a-z+.\-:]|$))/i),vxe=Ba(/^(?:\w+script|data):/i),Cxe=Ba(/[\u0000-\u0020\u00A0\u1680\u180E\u2000-\u2029\u205F\u3000]/g),Gte=Ba(/^html$/i),yxe=Ba(/^[a-z][.\w]*(-[.\w]+)+$/i);var Sq=Object.freeze({__proto__:null,ARIA_ATTR:wxe,ATTR_WHITESPACE:Cxe,CUSTOM_ELEMENT:yxe,DATA_ATTR:bxe,DOCTYPE_NAME:Gte,ERB_EXPR:mxe,IS_ALLOWED_URI:jte,IS_SCRIPT_OR_DATA:vxe,MUSTACHE_EXPR:pxe,TMPLIT_EXPR:_xe});const WC={element:1,text:3,progressingInstruction:7,comment:8,document:9},Sxe=function(){return typeof window>"u"?null:window},xxe=function(e,t){if(typeof e!="object"||typeof e.createPolicy!="function")return null;let i=null;const n="data-tt-policy-suffix";t&&t.hasAttribute(n)&&(i=t.getAttribute(n));const o="dompurify"+(i?"#"+i:"");try{return e.createPolicy(o,{createHTML(r){return r},createScriptURL(r){return r}})}catch{return console.warn("TrustedTypes policy "+o+" could not be created."),null}},xq=function(){return{afterSanitizeAttributes:[],afterSanitizeElements:[],afterSanitizeShadowDOM:[],beforeSanitizeAttributes:[],beforeSanitizeElements:[],beforeSanitizeShadowDOM:[],uponSanitizeAttribute:[],uponSanitizeElement:[],uponSanitizeShadowNode:[]}};function Kte(){let s=arguments.length>0&&arguments[0]!==void 0?arguments[0]:Sxe();const e=st=>Kte(st);if(e.version="3.2.7",e.removed=[],!s||!s.document||s.document.nodeType!==WC.document||!s.Element)return e.isSupported=!1,e;let{document:t}=s;const i=t,n=i.currentScript,{DocumentFragment:o,HTMLTemplateElement:r,Node:a,Element:l,NodeFilter:c,NamedNodeMap:d=s.NamedNodeMap||s.MozNamedAttrMap,HTMLFormElement:h,DOMParser:u,trustedTypes:g}=s,f=l.prototype,p=BC(f,"cloneNode"),_=BC(f,"remove"),b=BC(f,"nextSibling"),w=BC(f,"childNodes"),C=BC(f,"parentNode");if(typeof r=="function"){const st=t.createElement("template");st.content&&st.content.ownerDocument&&(t=st.content.ownerDocument)}let v,y="";const{implementation:x,createNodeIterator:L,createDocumentFragment:E,getElementsByTagName:I}=t,{importNode:R}=i;let O=xq();e.isSupported=typeof Ute=="function"&&typeof C=="function"&&x&&x.createHTMLDocument!==void 0;const{MUSTACHE_EXPR:P,ERB_EXPR:B,TMPLIT_EXPR:$,DATA_ATTR:z,ARIA_ATTR:Q,IS_SCRIPT_OR_DATA:J,ATTR_WHITESPACE:ne,CUSTOM_ELEMENT:de}=Sq;let{IS_ALLOWED_URI:ke}=Sq,Me=null;const Ze=Wt({},[...wq,...TM,...NM,...RM,...vq]);let Xe=null;const At=Wt({},[...Cq,...AM,...yq,...vL]);let Et=Object.seal(qte(null,{tagNameCheck:{writable:!0,configurable:!1,enumerable:!0,value:null},attributeNameCheck:{writable:!0,configurable:!1,enumerable:!0,value:null},allowCustomizedBuiltInElements:{writable:!0,configurable:!1,enumerable:!0,value:!1}})),un=null,Vi=null,Ws=!0,wC=!0,Rx=!1,YV=!0,T_=!1,Ax=!0,Pg=!1,m2=!1,_2=!1,N_=!1,Mx=!1,Ox=!1,ZV=!0,XV=!1;const Cle="user-content-";let b2=!0,vC=!1,R_={},A_=null;const QV=Wt({},["annotation-xml","audio","colgroup","desc","foreignobject","head","iframe","math","mi","mn","mo","ms","mtext","noembed","noframes","noscript","plaintext","script","style","svg","template","thead","title","video","xmp"]);let JV=null;const ez=Wt({},["audio","video","img","source","image","track"]);let w2=null;const tz=Wt({},["alt","class","for","id","label","name","pattern","placeholder","role","summary","title","value","style","xmlns"]),Px="http://www.w3.org/1998/Math/MathML",Fx="http://www.w3.org/2000/svg",Xc="http://www.w3.org/1999/xhtml";let M_=Xc,v2=!1,C2=null;const yle=Wt({},[Px,Fx,Xc],IM);let Bx=Wt({},["mi","mo","mn","ms","mtext"]),Wx=Wt({},["annotation-xml"]);const Sle=Wt({},["title","style","font","a","script"]);let CC=null;const xle=["application/xhtml+xml","text/html"],Lle="text/html";let Cs=null,O_=null;const kle=t.createElement("form"),iz=function(Y){return Y instanceof RegExp||Y instanceof Function},y2=function(){let Y=arguments.length>0&&arguments[0]!==void 0?arguments[0]:{};if(!(O_&&O_===Y)){if((!Y||typeof Y!="object")&&(Y={}),Y=pd(Y),CC=xle.indexOf(Y.PARSER_MEDIA_TYPE)===-1?Lle:Y.PARSER_MEDIA_TYPE,Cs=CC==="application/xhtml+xml"?IM:Fk,Me=Xa(Y,"ALLOWED_TAGS")?Wt({},Y.ALLOWED_TAGS,Cs):Ze,Xe=Xa(Y,"ALLOWED_ATTR")?Wt({},Y.ALLOWED_ATTR,Cs):At,C2=Xa(Y,"ALLOWED_NAMESPACES")?Wt({},Y.ALLOWED_NAMESPACES,IM):yle,w2=Xa(Y,"ADD_URI_SAFE_ATTR")?Wt(pd(tz),Y.ADD_URI_SAFE_ATTR,Cs):tz,JV=Xa(Y,"ADD_DATA_URI_TAGS")?Wt(pd(ez),Y.ADD_DATA_URI_TAGS,Cs):ez,A_=Xa(Y,"FORBID_CONTENTS")?Wt({},Y.FORBID_CONTENTS,Cs):QV,un=Xa(Y,"FORBID_TAGS")?Wt({},Y.FORBID_TAGS,Cs):pd({}),Vi=Xa(Y,"FORBID_ATTR")?Wt({},Y.FORBID_ATTR,Cs):pd({}),R_=Xa(Y,"USE_PROFILES")?Y.USE_PROFILES:!1,Ws=Y.ALLOW_ARIA_ATTR!==!1,wC=Y.ALLOW_DATA_ATTR!==!1,Rx=Y.ALLOW_UNKNOWN_PROTOCOLS||!1,YV=Y.ALLOW_SELF_CLOSE_IN_ATTR!==!1,T_=Y.SAFE_FOR_TEMPLATES||!1,Ax=Y.SAFE_FOR_XML!==!1,Pg=Y.WHOLE_DOCUMENT||!1,N_=Y.RETURN_DOM||!1,Mx=Y.RETURN_DOM_FRAGMENT||!1,Ox=Y.RETURN_TRUSTED_TYPE||!1,_2=Y.FORCE_BODY||!1,ZV=Y.SANITIZE_DOM!==!1,XV=Y.SANITIZE_NAMED_PROPS||!1,b2=Y.KEEP_CONTENT!==!1,vC=Y.IN_PLACE||!1,ke=Y.ALLOWED_URI_REGEXP||jte,M_=Y.NAMESPACE||Xc,Bx=Y.MATHML_TEXT_INTEGRATION_POINTS||Bx,Wx=Y.HTML_INTEGRATION_POINTS||Wx,Et=Y.CUSTOM_ELEMENT_HANDLING||{},Y.CUSTOM_ELEMENT_HANDLING&&iz(Y.CUSTOM_ELEMENT_HANDLING.tagNameCheck)&&(Et.tagNameCheck=Y.CUSTOM_ELEMENT_HANDLING.tagNameCheck),Y.CUSTOM_ELEMENT_HANDLING&&iz(Y.CUSTOM_ELEMENT_HANDLING.attributeNameCheck)&&(Et.attributeNameCheck=Y.CUSTOM_ELEMENT_HANDLING.attributeNameCheck),Y.CUSTOM_ELEMENT_HANDLING&&typeof Y.CUSTOM_ELEMENT_HANDLING.allowCustomizedBuiltInElements=="boolean"&&(Et.allowCustomizedBuiltInElements=Y.CUSTOM_ELEMENT_HANDLING.allowCustomizedBuiltInElements),T_&&(wC=!1),Mx&&(N_=!0),R_&&(Me=Wt({},vq),Xe=[],R_.html===!0&&(Wt(Me,wq),Wt(Xe,Cq)),R_.svg===!0&&(Wt(Me,TM),Wt(Xe,AM),Wt(Xe,vL)),R_.svgFilters===!0&&(Wt(Me,NM),Wt(Xe,AM),Wt(Xe,vL)),R_.mathMl===!0&&(Wt(Me,RM),Wt(Xe,yq),Wt(Xe,vL))),Y.ADD_TAGS&&(Me===Ze&&(Me=pd(Me)),Wt(Me,Y.ADD_TAGS,Cs)),Y.ADD_ATTR&&(Xe===At&&(Xe=pd(Xe)),Wt(Xe,Y.ADD_ATTR,Cs)),Y.ADD_URI_SAFE_ATTR&&Wt(w2,Y.ADD_URI_SAFE_ATTR,Cs),Y.FORBID_CONTENTS&&(A_===QV&&(A_=pd(A_)),Wt(A_,Y.FORBID_CONTENTS,Cs)),b2&&(Me["#text"]=!0),Pg&&Wt(Me,["html","head","body"]),Me.table&&(Wt(Me,["tbody"]),delete un.tbody),Y.TRUSTED_TYPES_POLICY){if(typeof Y.TRUSTED_TYPES_POLICY.createHTML!="function")throw FC('TRUSTED_TYPES_POLICY configuration option must provide a "createHTML" hook.');if(typeof Y.TRUSTED_TYPES_POLICY.createScriptURL!="function")throw FC('TRUSTED_TYPES_POLICY configuration option must provide a "createScriptURL" hook.');v=Y.TRUSTED_TYPES_POLICY,y=v.createHTML("")}else v===void 0&&(v=xxe(g,n)),v!==null&&typeof y=="string"&&(y=v.createHTML(""));jo&&jo(Y),O_=Y}},nz=Wt({},[...TM,...NM,...gxe]),sz=Wt({},[...RM,...fxe]),Ele=function(Y){let Ee=C(Y);(!Ee||!Ee.tagName)&&(Ee={namespaceURI:M_,tagName:"template"});const Qe=Fk(Y.tagName),nn=Fk(Ee.tagName);return C2[Y.namespaceURI]?Y.namespaceURI===Fx?Ee.namespaceURI===Xc?Qe==="svg":Ee.namespaceURI===Px?Qe==="svg"&&(nn==="annotation-xml"||Bx[nn]):!!nz[Qe]:Y.namespaceURI===Px?Ee.namespaceURI===Xc?Qe==="math":Ee.namespaceURI===Fx?Qe==="math"&&Wx[nn]:!!sz[Qe]:Y.namespaceURI===Xc?Ee.namespaceURI===Fx&&!Wx[nn]||Ee.namespaceURI===Px&&!Bx[nn]?!1:!sz[Qe]&&(Sle[Qe]||!nz[Qe]):!!(CC==="application/xhtml+xml"&&C2[Y.namespaceURI]):!1},Pl=function(Y){OC(e.removed,{element:Y});try{C(Y).removeChild(Y)}catch{_(Y)}},Fg=function(Y,Ee){try{OC(e.removed,{attribute:Ee.getAttributeNode(Y),from:Ee})}catch{OC(e.removed,{attribute:null,from:Ee})}if(Ee.removeAttribute(Y),Y==="is")if(N_||Mx)try{Pl(Ee)}catch{}else try{Ee.setAttribute(Y,"")}catch{}},oz=function(Y){let Ee=null,Qe=null;if(_2)Y="<remove></remove>"+Y;else{const zn=DM(Y,/^[\r\n\t ]+/);Qe=zn&&zn[0]}CC==="application/xhtml+xml"&&M_===Xc&&(Y='<html xmlns="http://www.w3.org/1999/xhtml"><head></head><body>'+Y+"</body></html>");const nn=v?v.createHTML(Y):Y;if(M_===Xc)try{Ee=new u().parseFromString(nn,CC)}catch{}if(!Ee||!Ee.documentElement){Ee=x.createDocument(M_,"template",null);try{Ee.documentElement.innerHTML=v2?y:nn}catch{}}const ao=Ee.body||Ee.documentElement;return Y&&Qe&&ao.insertBefore(t.createTextNode(Qe),ao.childNodes[0]||null),M_===Xc?I.call(Ee,Pg?"html":"body")[0]:Pg?Ee.documentElement:ao},rz=function(Y){return L.call(Y.ownerDocument||Y,Y,c.SHOW_ELEMENT|c.SHOW_COMMENT|c.SHOW_TEXT|c.SHOW_PROCESSING_INSTRUCTION|c.SHOW_CDATA_SECTION,null)},S2=function(Y){return Y instanceof h&&(typeof Y.nodeName!="string"||typeof Y.textContent!="string"||typeof Y.removeChild!="function"||!(Y.attributes instanceof d)||typeof Y.removeAttribute!="function"||typeof Y.setAttribute!="function"||typeof Y.namespaceURI!="string"||typeof Y.insertBefore!="function"||typeof Y.hasChildNodes!="function")},az=function(Y){return typeof a=="function"&&Y instanceof a};function Qc(st,Y,Ee){wL(st,Qe=>{Qe.call(e,Y,Ee,O_)})}const lz=function(Y){let Ee=null;if(Qc(O.beforeSanitizeElements,Y,null),S2(Y))return Pl(Y),!0;const Qe=Cs(Y.nodeName);if(Qc(O.uponSanitizeElement,Y,{tagName:Qe,allowedTags:Me}),Ax&&Y.hasChildNodes()&&!az(Y.firstElementChild)&&Io(/<[/\w!]/g,Y.innerHTML)&&Io(/<[/\w!]/g,Y.textContent)||Y.nodeType===WC.progressingInstruction||Ax&&Y.nodeType===WC.comment&&Io(/<[/\w]/g,Y.data))return Pl(Y),!0;if(!Me[Qe]||un[Qe]){if(!un[Qe]&&dz(Qe)&&(Et.tagNameCheck instanceof RegExp&&Io(Et.tagNameCheck,Qe)||Et.tagNameCheck instanceof Function&&Et.tagNameCheck(Qe)))return!1;if(b2&&!A_[Qe]){const nn=C(Y)||Y.parentNode,ao=w(Y)||Y.childNodes;if(ao&&nn){const zn=ao.length;for(let Jo=zn-1;Jo>=0;--Jo){const Jc=p(ao[Jo],!0);Jc.__removalCount=(Y.__removalCount||0)+1,nn.insertBefore(Jc,b(Y))}}}return Pl(Y),!0}return Y instanceof l&&!Ele(Y)||(Qe==="noscript"||Qe==="noembed"||Qe==="noframes")&&Io(/<\/no(script|embed|frames)/i,Y.innerHTML)?(Pl(Y),!0):(T_&&Y.nodeType===WC.text&&(Ee=Y.textContent,wL([P,B,$],nn=>{Ee=PC(Ee,nn," ")}),Y.textContent!==Ee&&(OC(e.removed,{element:Y.cloneNode()}),Y.textContent=Ee)),Qc(O.afterSanitizeElements,Y,null),!1)},cz=function(Y,Ee,Qe){if(ZV&&(Ee==="id"||Ee==="name")&&(Qe in t||Qe in kle))return!1;if(!(wC&&!Vi[Ee]&&Io(z,Ee))){if(!(Ws&&Io(Q,Ee))){if(!Xe[Ee]||Vi[Ee]){if(!(dz(Y)&&(Et.tagNameCheck instanceof RegExp&&Io(Et.tagNameCheck,Y)||Et.tagNameCheck instanceof Function&&Et.tagNameCheck(Y))&&(Et.attributeNameCheck instanceof RegExp&&Io(Et.attributeNameCheck,Ee)||Et.attributeNameCheck instanceof Function&&Et.attributeNameCheck(Ee,Y))||Ee==="is"&&Et.allowCustomizedBuiltInElements&&(Et.tagNameCheck instanceof RegExp&&Io(Et.tagNameCheck,Qe)||Et.tagNameCheck instanceof Function&&Et.tagNameCheck(Qe))))return!1}else if(!w2[Ee]){if(!Io(ke,PC(Qe,ne,""))){if(!((Ee==="src"||Ee==="xlink:href"||Ee==="href")&&Y!=="script"&&cxe(Qe,"data:")===0&&JV[Y])){if(!(Rx&&!Io(J,PC(Qe,ne,"")))){if(Qe)return!1}}}}}}return!0},dz=function(Y){return Y!=="annotation-xml"&&DM(Y,de)},hz=function(Y){Qc(O.beforeSanitizeAttributes,Y,null);const{attributes:Ee}=Y;if(!Ee||S2(Y))return;const Qe={attrName:"",attrValue:"",keepAttr:!0,allowedAttributes:Xe,forceKeepAttr:void 0};let nn=Ee.length;for(;nn--;){const ao=Ee[nn],{name:zn,namespaceURI:Jo,value:Jc}=ao,P_=Cs(zn),x2=Jc;let Hs=zn==="value"?x2:dxe(x2);if(Qe.attrName=P_,Qe.attrValue=Hs,Qe.keepAttr=!0,Qe.forceKeepAttr=void 0,Qc(O.uponSanitizeAttribute,Y,Qe),Hs=Qe.attrValue,XV&&(P_==="id"||P_==="name")&&(Fg(zn,Y),Hs=Cle+Hs),Ax&&Io(/((--!?|])>)|<\/(style|title|textarea)/i,Hs)){Fg(zn,Y);continue}if(P_==="attributename"&&DM(Hs,"href")){Fg(zn,Y);continue}if(Qe.forceKeepAttr)continue;if(!Qe.keepAttr){Fg(zn,Y);continue}if(!YV&&Io(/\/>/i,Hs)){Fg(zn,Y);continue}T_&&wL([P,B,$],gz=>{Hs=PC(Hs,gz," ")});const uz=Cs(Y.nodeName);if(!cz(uz,P_,Hs)){Fg(zn,Y);continue}if(v&&typeof g=="object"&&typeof g.getAttributeType=="function"&&!Jo)switch(g.getAttributeType(uz,P_)){case"TrustedHTML":{Hs=v.createHTML(Hs);break}case"TrustedScriptURL":{Hs=v.createScriptURL(Hs);break}}if(Hs!==x2)try{Jo?Y.setAttributeNS(Jo,zn,Hs):Y.setAttribute(zn,Hs),S2(Y)?Pl(Y):bq(e.removed)}catch{Fg(zn,Y)}}Qc(O.afterSanitizeAttributes,Y,null)},Ile=function st(Y){let Ee=null;const Qe=rz(Y);for(Qc(O.beforeSanitizeShadowDOM,Y,null);Ee=Qe.nextNode();)Qc(O.uponSanitizeShadowNode,Ee,null),lz(Ee),hz(Ee),Ee.content instanceof o&&st(Ee.content);Qc(O.afterSanitizeShadowDOM,Y,null)};return e.sanitize=function(st){let Y=arguments.length>1&&arguments[1]!==void 0?arguments[1]:{},Ee=null,Qe=null,nn=null,ao=null;if(v2=!st,v2&&(st="<!-->"),typeof st!="string"&&!az(st))if(typeof st.toString=="function"){if(st=st.toString(),typeof st!="string")throw FC("dirty is not a string, aborting")}else throw FC("toString is not a function");if(!e.isSupported)return st;if(m2||y2(Y),e.removed=[],typeof st=="string"&&(vC=!1),vC){if(st.nodeName){const Jc=Cs(st.nodeName);if(!Me[Jc]||un[Jc])throw FC("root node is forbidden and cannot be sanitized in-place")}}else if(st instanceof a)Ee=oz("<!---->"),Qe=Ee.ownerDocument.importNode(st,!0),Qe.nodeType===WC.element&&Qe.nodeName==="BODY"||Qe.nodeName==="HTML"?Ee=Qe:Ee.appendChild(Qe);else{if(!N_&&!T_&&!Pg&&st.indexOf("<")===-1)return v&&Ox?v.createHTML(st):st;if(Ee=oz(st),!Ee)return N_?null:Ox?y:""}Ee&&_2&&Pl(Ee.firstChild);const zn=rz(vC?st:Ee);for(;nn=zn.nextNode();)lz(nn),hz(nn),nn.content instanceof o&&Ile(nn.content);if(vC)return st;if(N_){if(Mx)for(ao=E.call(Ee.ownerDocument);Ee.firstChild;)ao.appendChild(Ee.firstChild);else ao=Ee;return(Xe.shadowroot||Xe.shadowrootmode)&&(ao=R.call(i,ao,!0)),ao}let Jo=Pg?Ee.outerHTML:Ee.innerHTML;return Pg&&Me["!doctype"]&&Ee.ownerDocument&&Ee.ownerDocument.doctype&&Ee.ownerDocument.doctype.name&&Io(Gte,Ee.ownerDocument.doctype.name)&&(Jo="<!DOCTYPE "+Ee.ownerDocument.doctype.name+`>
`+Jo),T_&&wL([P,B,$],Jc=>{Jo=PC(Jo,Jc," ")}),v&&Ox?v.createHTML(Jo):Jo},e.setConfig=function(){let st=arguments.length>0&&arguments[0]!==void 0?arguments[0]:{};y2(st),m2=!0},e.clearConfig=function(){O_=null,m2=!1},e.isValidAttribute=function(st,Y,Ee){O_||y2({});const Qe=Cs(st),nn=Cs(Y);return cz(Qe,nn,Ee)},e.addHook=function(st,Y){typeof Y=="function"&&OC(O[st],Y)},e.removeHook=function(st,Y){if(Y!==void 0){const Ee=axe(O[st],Y);return Ee===-1?void 0:lxe(O[st],Ee,1)[0]}return bq(O[st])},e.removeHooks=function(st){O[st]=[]},e.removeAllHooks=function(){O=xq()},e}var hb=Kte();const Yte=Object.freeze(["a","abbr","b","bdo","blockquote","br","caption","cite","code","col","colgroup","dd","del","details","dfn","div","dl","dt","em","figcaption","figure","h1","h2","h3","h4","h5","h6","hr","i","img","ins","kbd","label","li","mark","ol","p","pre","q","rp","rt","ruby","s","samp","small","small","source","span","strike","strong","sub","summary","sup","table","tbody","td","tfoot","th","thead","time","tr","tt","u","ul","var","video","wbr"]),Zte=Object.freeze(["href","target","src","alt","title","for","name","role","tabindex","x-dispatch","required","checked","placeholder","type","start","width","height","align"]),MM="vscode-relative-path";function Lq(s,e){if(e.override==="*")return!0;try{const t=new URL(s,MM+"://");return!!(e.override.includes(t.protocol.replace(/:$/,""))||e.allowRelativePaths&&t.protocol===MM+":"&&!s.trim().toLowerCase().startsWith(MM))}catch{return!1}}function Lxe(s,e){hb.addHook("afterSanitizeAttributes",t=>{for(const i of["href","src"])if(t.hasAttribute(i)){const n=t.getAttribute(i);i==="href"?!n.startsWith("#")&&!Lq(n,s)&&t.removeAttribute(i):Lq(n,e)||t.removeAttribute(i)}})}const kxe=Object.freeze({ALLOWED_TAGS:[...Yte],ALLOWED_ATTR:[...Zte],ALLOW_UNKNOWN_PROTOCOLS:!0});function Exe(s,e){return Xte(s,e,"trusted")}function Xte(s,e,t){try{const i={...kxe};e?.allowedTags&&(e.allowedTags.override&&(i.ALLOWED_TAGS=[...e.allowedTags.override]),e.allowedTags.augment&&(i.ALLOWED_TAGS=[...i.ALLOWED_TAGS??[],...e.allowedTags.augment]));let n=[...Zte];e?.allowedAttributes&&(e.allowedAttributes.override&&(n=[...e.allowedAttributes.override]),e.allowedAttributes.augment&&(n=[...n,...e.allowedAttributes.augment])),n=n.map(a=>typeof a=="string"?a.toLowerCase():{attributeName:a.attributeName.toLowerCase(),shouldKeep:a.shouldKeep});const o=new Set(n.map(a=>typeof a=="string"?a:a.attributeName)),r=new Map;for(const a of n)typeof a=="string"?r.delete(a):r.set(a.attributeName,a);return i.ALLOWED_ATTR=Array.from(o),Lxe({override:e?.allowedLinkProtocols?.override??[Te.http,Te.https],allowRelativePaths:e?.allowRelativeLinkPaths??!1},{override:e?.allowedMediaProtocols?.override??[Te.http,Te.https],allowRelativePaths:e?.allowRelativeMediaPaths??!1}),e?.replaceWithPlaintext&&hb.addHook("uponSanitizeElement",Dxe),r.size&&hb.addHook("uponSanitizeAttribute",(a,l)=>{const c=r.get(l.attrName);if(c){const d=c.shouldKeep(a,l);typeof d=="string"?(l.keepAttr=!0,l.attrValue=d):l.keepAttr=d}else l.keepAttr=o.has(l.attrName)}),t==="dom"?hb.sanitize(s,{...i,RETURN_DOM_FRAGMENT:!0}):hb.sanitize(s,{...i,RETURN_TRUSTED_TYPE:!0})}finally{hb.removeAllHooks()}}const Ixe=["area","base","br","col","command","embed","hr","img","input","keygen","link","meta","param","source","track","wbr"],Dxe=(s,e,t)=>{if(!e.allowedTags[e.tagName]&&e.tagName!=="body"){const i=Qte(s);i&&(s.nodeType===Node.COMMENT_NODE?s.parentElement?.insertBefore(i,s):s.parentElement?.replaceChild(i,s))}};function Qte(s){if(!s.ownerDocument)return;let e,t;if(s.nodeType===Node.COMMENT_NODE)e=`<!--${s.textContent}-->`;else if(s instanceof Element){const r=s.tagName.toLowerCase(),a=Ixe.includes(r),l=s.attributes.length?" "+Array.from(s.attributes).map(c=>`${c.name}="${c.value}"`).join(" "):"";e=`<${r}${l}>`,a||(t=`</${r}>`)}else return;const i=document.createDocumentFragment(),n=s.ownerDocument.createTextNode(e);for(i.appendChild(n);s.firstChild;)i.appendChild(s.firstChild);const o=t?s.ownerDocument.createTextNode(t):void 0;return o&&i.appendChild(o),i}function Jte(s,e,t){const i=Xte(e,t,"dom");vn(s,i)}const Txe=new RegExp(`(\\\\)?\\$\\((${xe.iconNameExpression}(?:${xe.iconModifierExpression})?)\\)`,"g");function Pu(s){const e=new Array;let t,i=0,n=0;for(;(t=Txe.exec(s))!==null;){n=t.index||0,i<n&&e.push(s.substring(i,n)),i=(t.index||0)+t[0].length;const[,o,r]=t;e.push(o?`$(${r})`:Dl({id:r}))}return i<s.length&&e.push(s.substring(i)),e}function Dl(s){const e=ie("span");return e.classList.add(...xe.asClassNameArray(s)),e}const OM=Object.freeze({image:({href:s,title:e,text:t})=>{let i=[],n=[];return s&&({href:s,dimensions:i}=y0e(s),n.push(`src="${_L(s)}"`)),t&&n.push(`alt="${_L(t)}"`),e&&n.push(`title="${_L(e)}"`),i.length&&(n=n.concat(i)),"<img "+n.join(" ")+">"},paragraph({tokens:s}){return`<p>${this.parser.parseInline(s)}</p>`},link({href:s,title:e,tokens:t}){let i=this.parser.parseInline(t);return typeof s!="string"?"":(s===i&&(i=EM(i)),e=typeof e=="string"?_L(EM(e)):"",s=EM(s),s=s.replace(/&/g,"&amp;").replace(/</g,"&lt;").replace(/>/g,"&gt;").replace(/"/g,"&quot;").replace(/'/g,"&#39;"),`<a href="${s}" title="${e||s}" draggable="false">${i}</a>`)}});function Nxe(s){return function(e){const{tokens:t}=e,i=t[0];if(i?.type!=="paragraph")return s.call(this,e);const n=i.tokens;if(!n||n.length===0)return s.call(this,e);const o=n[0];if(o?.type!=="text")return s.call(this,e);const r=/^\s*\[!(NOTE|TIP|IMPORTANT|WARNING|CAUTION)\]\s*?\n*/i,a=o.raw.match(r);if(!a)return s.call(this,e);o.raw=o.raw.replace(r,""),o.text=o.text.replace(r,"");const l={note:"info",tip:"light-bulb",important:"comment",warning:"alert",caution:"stop"},c=a[1],d=c.charAt(0).toUpperCase()+c.slice(1).toLowerCase(),h=c.toLowerCase(),u=Dl({id:l[h]}).outerHTML,g=this.parser.parse(t);return`<blockquote data-severity="${h}"><p><span>${u}${d}</span>${g.substring(3)}</blockquote>
`}}function lx(s,e={},t){const i=new U;let n=!1;const o=new $te(...e.markedExtensions??[]),{renderer:r,codeBlocks:a,syncCodeBlocks:l}=Axe(o,e,s),c=Mxe(s);let d;if(e.fillInIncompleteTokens){const f={...o.defaults,...e.markedOptions,renderer:r},p=o.lexer(c,f),_=Gxe(p);d=o.parser(_,f)}else d=o.parse(c,{...e?.markedOptions,renderer:r,async:!1});s.supportThemeIcons&&(d=Pu(d).map(p=>typeof p=="string"?p:p.outerHTML).join(""));const h=document.createElement("div"),u=eie(s,e.sanitizerConfig??{});Jte(h,d,u),Rxe(s,e,h);let g;if(t?(g=t,vn(t,...h.children)):g=h,a.length>0)Promise.all(a).then(f=>{if(n)return;const p=new Map(f),_=g.querySelectorAll("div[data-code]");for(const b of _){const w=p.get(b.dataset.code??"");w&&vn(b,w)}e.asyncRenderCallback?.()});else if(l.length>0){const f=new Map(l),p=g.querySelectorAll("div[data-code]");for(const _ of p){const b=f.get(_.dataset.code??"");b&&vn(_,b)}}if(e.asyncRenderCallback)for(const f of g.getElementsByTagName("img")){const p=i.add(W(f,"load",()=>{p.dispose(),e.asyncRenderCallback()}))}if(e.actionHandler){const f=p=>{const _=new is(be(g),p);!_.leftButton&&!_.middleButton||kq(s,e,_)};i.add(W(g,"click",f)),i.add(W(g,"auxclick",f)),i.add(W(g,"keydown",p=>{const _=new Pt(p);!_.equals(10)&&!_.equals(3)||kq(s,e,_)}))}for(const f of[...g.getElementsByTagName("input")])if(f.attributes.getNamedItem("type")?.value==="checkbox")f.setAttribute("disabled","");else if(e.sanitizerConfig?.replaceWithPlaintext){const p=Qte(f);p?f.parentElement?.replaceChild(p,f):f.remove()}else f.remove();return{element:g,dispose:()=>{n=!0,i.dispose()}}}function Rxe(s,e,t){for(const i of t.querySelectorAll("img, audio, video, source")){const n=i.getAttribute("src");if(n){let o=n;try{s.baseUri&&(o=n4(Le.from(s.baseUri),o))}catch{}if(i.setAttribute("src",Eq(s,o,!0)),e.sanitizerConfig?.remoteImageIsAllowed){const r=Le.parse(o);r.scheme!==Te.file&&r.scheme!==Te.data&&!e.sanitizerConfig.remoteImageIsAllowed(r)&&i.replaceWith(ie("",void 0,i.outerHTML))}}}for(const i of t.querySelectorAll("a")){const n=i.getAttribute("href");if(i.setAttribute("href",""),!n||/^data:|javascript:/i.test(n)||/^command:/i.test(n)&&!s.isTrusted||/^command:(\/\/\/)?_workbench\.downloadResource/i.test(n))i.replaceWith(...i.childNodes);else{let o=Eq(s,n,!1);s.baseUri&&(o=n4(Le.from(s.baseUri),n)),i.dataset.href=o}}}function Axe(s,e,t){const i=new s.Renderer(e.markedOptions);i.image=OM.image,i.link=OM.link,i.paragraph=OM.paragraph,t.supportAlertSyntax&&(i.blockquote=Nxe(i.blockquote));const n=[],o=[];return e.codeBlockRendererSync?i.code=({text:r,lang:a,raw:l})=>{const c=XF.nextId(),d=e.codeBlockRendererSync(Iq(a),r,l);return o.push([c,d]),`<div class="code" data-code="${c}">${Aa(r)}</div>`}:e.codeBlockRenderer&&(i.code=({text:r,lang:a})=>{const l=XF.nextId(),c=e.codeBlockRenderer(Iq(a),r);return n.push(c.then(d=>[l,d])),`<div class="code" data-code="${l}">${Aa(r)}</div>`}),t.supportHtml||(i.html=({text:r})=>e.sanitizerConfig?.replaceWithPlaintext?Aa(r):(t.isTrusted?r.match(/^(<span[^>]+>)|(<\/\s*span>)$/):void 0)?r:""),{renderer:i,codeBlocks:n,syncCodeBlocks:o}}function Mxe(s){let e=s.value;return e.length>1e5&&(e=`${e.substr(0,1e5)}…`),s.supportThemeIcons&&(e=_0e(e)),e}function kq(s,e,t){const i=t.target.closest("a[data-href]");if(ki(i))try{let n=i.dataset.href;n&&(s.baseUri&&(n=n4(Le.from(s.baseUri),n)),e.actionHandler?.(n,s))}catch(n){Oe(n)}finally{t.preventDefault()}}function Oxe(s,e){let t;try{t=JF(decodeURIComponent(e))}catch{}return t?(t=iJ(t,i=>{if(s.uris&&s.uris[i])return Le.revive(s.uris[i])}),encodeURIComponent(JSON.stringify(t))):e}function Eq(s,e,t){const i=s.uris&&s.uris[e];let n=Le.revive(i);return t?e.startsWith(Te.data+":")?e:(n||(n=Le.parse(e)),gQ.uriToBrowserUri(n).toString(!0)):!n||Le.parse(e).toString()===n.toString()?e:(n.query&&(n=n.with({query:Oxe(s,n.query)})),n.toString())}function Iq(s){if(!s)return"";const e=s.split(/[\s+|:|,|\{|\?]/,1);return e.length?e[0]:s}function n4(s,e){return/^\w[\w\d+.-]*:/.test(e)?e:s.path.endsWith("/")?l$(s,e).toString():l$(zR(s),e).toString()}function Pxe(s,e,t={}){const i=eie(e,t);return Exe(s,i)}const Fxe=Object.freeze([...Yte,"input"]),Bxe=Object.freeze(["align","autoplay","alt","colspan","controls","draggable","height","href","loop","muted","playsinline","poster","rowspan","src","target","title","type","width","start","checked","disabled","value","data-code","data-href","data-severity",{attributeName:"style",shouldKeep:(s,e)=>s.tagName==="SPAN"&&e.attrName==="style"?/^(color\:(#[0-9a-fA-F]+|var\(--vscode(-[a-zA-Z0-9]+)+\));)?(background-color\:(#[0-9a-fA-F]+|var\(--vscode(-[a-zA-Z0-9]+)+\));)?(border-radius:[0-9]+px;)?$/.test(e.attrValue):!1},{attributeName:"class",shouldKeep:(s,e)=>s.tagName==="SPAN"&&e.attrName==="class"?/^codicon codicon-[a-z\-]+( codicon-modifier-[a-z\-]+)?$/.test(e.attrValue):!1}]);function eie(s,e){const t=s.isTrusted??!1,i=[Te.http,Te.https,Te.mailto,Te.file,Te.vscodeFileResource,Te.vscodeRemote,Te.vscodeRemoteResource,Te.vscodeNotebookCell];return t&&i.push(Te.command),e.allowedLinkSchemes?.augment&&i.push(...e.allowedLinkSchemes.augment),{allowedTags:{override:e.allowedTags?.override??Fxe},allowedAttributes:{override:e.allowedAttributes?.override??Bxe},allowedLinkProtocols:{override:i},allowRelativeLinkPaths:!!s.baseUri,allowedMediaProtocols:{override:[Te.http,Te.https,Te.data,Te.file,Te.vscodeFileResource,Te.vscodeRemote,Te.vscodeRemoteResource]},allowRelativeMediaPaths:!!s.baseUri,replaceWithPlaintext:e.replaceWithPlaintext}}function Wxe(s,e){if(typeof s=="string")return s;let t=s.value??"";t.length>1e5&&(t=`${t.substr(0,1e5)}…`);const i=txe(t,{async:!1,renderer:Vxe.value});return Pxe(i,{isTrusted:!1},{}).toString().replace(/&(#\d+|[a-zA-Z]+);/g,n=>Hxe.get(n)??n).trim()}const Hxe=new Map([["&quot;",'"'],["&nbsp;"," "],["&amp;","&"],["&#39;","'"],["&lt;","<"],["&gt;",">"]]);function tie(){const s=new NS;return s.code=({text:e})=>Aa(e),s.blockquote=({text:e})=>e+`
`,s.html=e=>"",s.heading=function({tokens:e}){return this.parser.parseInline(e)+`
`},s.hr=()=>"",s.list=function({items:e}){return e.map(t=>this.listitem(t)).join(`
`)+`
`},s.listitem=({text:e})=>e+`
`,s.paragraph=function({tokens:e}){return this.parser.parseInline(e)+`
`},s.table=function({header:e,rows:t}){return e.map(i=>this.tablecell(i)).join(" ")+`
`+t.map(i=>i.map(n=>this.tablecell(n)).join(" ")).join(`
`)+`
`},s.tablerow=({text:e})=>e,s.tablecell=function({tokens:e}){return this.parser.parseInline(e)},s.strong=({text:e})=>e,s.em=({text:e})=>e,s.codespan=({text:e})=>Aa(e),s.br=e=>`
`,s.del=({text:e})=>e,s.image=e=>"",s.text=({text:e})=>e,s.link=({text:e})=>e,s}const Vxe=new es(tie);new es(()=>{const s=tie();return s.code=({text:e})=>`
\`\`\`
${Aa(e)}
\`\`\`
`,s});function AS(s){let e="";return s.forEach(t=>{e+=t.raw}),e}function iie(s){if(s.tokens)for(let e=s.tokens.length-1;e>=0;e--){const t=s.tokens[e];if(t.type==="text"){const i=t.raw.split(`
`),n=i[i.length-1];if(n.includes("`"))return Yxe(s);if(n.includes("**"))return tLe(s);if(n.match(/\*\w/))return Zxe(s);if(n.match(/(^|\s)__\w/))return iLe(s);if(n.match(/(^|\s)_\w/))return Xxe(s);if(zxe(n)||$xe(n)&&s.tokens.slice(0,e).some(o=>o.type==="text"&&o.raw.match(/\[[^\]]*$/))){const o=s.tokens.slice(e+1);return o[0]?.type==="link"&&o[1]?.type==="text"&&o[1].raw.match(/^ *"[^"]*$/)||n.match(/^[^"]* +"[^"]*$/)?Jxe(s):Qxe(s)}else if(n.match(/(^|\s)\[\w*[^\]]*$/))return eLe(s)}}}function zxe(s){return!!s.match(/(^|\s)\[.*\]\(\w*/)}function $xe(s){return!!s.match(/^[^\[]*\]\([^\)]*$/)}function Uxe(s){const e=s.items[s.items.length-1],t=e.tokens?e.tokens[e.tokens.length-1]:void 0,i=c=>{const h=c.items.at(-1)?.tokens.at(-1);return h?.type==="heading"||h?.type==="list"&&i(h)};let n;if(t?.type==="text"&&!("inRawBlock"in e))n=iie(t);else if(i(s)){const c=RS(s.raw.trim()+" &nbsp;")[0];return c.type!=="list"?void 0:c}if(!n||n.type!=="paragraph")return;const o=AS(s.items.slice(0,-1)),r=e.raw.match(/^(\s*(-|\d+\.|\*) +)/)?.[0];if(!r)return;const a=r+AS(e.tokens.slice(0,-1))+n.raw,l=RS(o+a)[0];if(l.type==="list")return l}function qxe(s,e){if(s.raw.match(/-\s*$/))return RS(e+" &nbsp;")}const jxe=3;function Gxe(s){for(let e=0;e<jxe;e++){const t=Kxe(s);if(t)s=t;else break}return s}function Kxe(s){let e,t;for(e=0;e<s.length;e++){const n=s[e];if(n.type==="paragraph"&&n.raw.match(/(\n|^)\|/)){t=nLe(s.slice(e));break}}const i=s.at(-1);if(!t&&i?.type==="list"){const n=Uxe(i);n&&(t=[n],e=s.length-1)}if(!t&&i?.type==="paragraph"){const n=iie(i);n&&(t=[n],e=s.length-1)}if(t){const n=[...s.slice(0,e),...t];return n.links=s.links,n}if(i?.type==="heading"){const n=qxe(i,AS(s));if(n)return n}return null}function Yxe(s){return Tg(s,"`")}function Zxe(s){return Tg(s,"*")}function Xxe(s){return Tg(s,"_")}function Qxe(s){return Tg(s,")",!1)}function Jxe(s){return Tg(s,'")',!1)}function eLe(s){return Tg(s,"](https://microsoft.com)",!1)}function tLe(s){return Tg(s,"**")}function iLe(s){return Tg(s,"__")}function Tg(s,e,t=!0){const i=AS(Array.isArray(s)?s:[s]),n=t?i.trimEnd():i;return RS(n+e)[0]}function nLe(s){const e=AS(s),t=e.split(`
`);let i,n=!1;for(let o=0;o<t.length;o++){const r=t[o].trim();if(typeof i>"u"&&r.match(/^\s*\|/)){const a=r.match(/(\|[^\|]+)(?=\||$)/g);a&&(i=a.length)}else if(typeof i=="number")if(r.match(/^\s*\|/)){if(o!==t.length-1)return;n=!0}else return}if(typeof i=="number"&&i>0){const o=n?t.slice(0,-1).join(`
`):e,r=!!o.match(/\|\s*$/),a=o+(r?"":"|")+`
|${" --- |".repeat(i)}`;return RS(a)}}let nie={showInstantHover:()=>{},showDelayedHover:()=>{},setupDelayedHover:()=>F.None,setupDelayedHoverAtMouse:()=>F.None,hideHover:()=>{},showAndFocusLastHover:()=>{},setupManagedHover:()=>({dispose:()=>{},show:()=>{},hide:()=>{},update:()=>{}}),showManagedHover:()=>{}};function sLe(s){nie=s}function $a(){return nie}class oLe{constructor(e){this.spliceables=e}splice(e,t,i){this.spliceables.forEach(n=>n.splice(e,t,i))}}function qs(s,e,t){return Math.min(Math.max(s,e),t)}class sie{constructor(){this._n=1,this._val=0}update(e){return this._val=this._val+(e-this._val)/this._n,this._n+=1,this._val}get value(){return this._val}}class rLe{constructor(e){this._n=0,this._val=0,this._values=[],this._index=0,this._sum=0,this._values=new Array(e),this._values.fill(0,0,e)}update(e){const t=this._values[this._index];return this._values[this._index]=e,this._index=(this._index+1)%this._values.length,this._sum-=t,this._sum+=e,this._n<this._values.length&&(this._n+=1),this._val=this._sum/this._n,this._val}get value(){return this._val}}class Ug extends Error{constructor(e,t){super(`ListError [${e}] ${t}`)}}var as;(function(s){function e(o,r){if(o.start>=r.end||r.start>=o.end)return{start:0,end:0};const a=Math.max(o.start,r.start),l=Math.min(o.end,r.end);return l-a<=0?{start:0,end:0}:{start:a,end:l}}s.intersect=e;function t(o){return o.end-o.start<=0}s.isEmpty=t;function i(o,r){return!t(e(o,r))}s.intersects=i;function n(o,r){const a=[],l={start:o.start,end:Math.min(r.start,o.end)},c={start:Math.max(r.end,o.start),end:o.end};return t(l)||a.push(l),t(c)||a.push(c),a}s.relativeComplement=n})(as||(as={}));function Dq(s,e){const t=[];for(const i of e){if(s.start>=i.range.end)continue;if(s.end<i.range.start)break;const n=as.intersect(s,i.range);as.isEmpty(n)||t.push({range:n,size:i.size})}return t}function s4({start:s,end:e},t){return{start:s+t,end:e+t}}function aLe(s){const e=[];let t=null;for(const i of s){const n=i.range.start,o=i.range.end,r=i.size;if(t&&r===t.size){t.range.end=o;continue}t={range:{start:n,end:o},size:r},e.push(t)}return e}function lLe(...s){return aLe(s.reduce((e,t)=>e.concat(t),[]))}class cLe{get paddingTop(){return this._paddingTop}set paddingTop(e){this._size=this._size+e-this._paddingTop,this._paddingTop=e}constructor(e){this.groups=[],this._size=0,this._paddingTop=0,this._paddingTop=e??0,this._size=this._paddingTop}splice(e,t,i=[]){const n=i.length-t,o=Dq({start:0,end:e},this.groups),r=Dq({start:e+t,end:Number.POSITIVE_INFINITY},this.groups).map(l=>({range:s4(l.range,n),size:l.size})),a=i.map((l,c)=>({range:{start:e+c,end:e+c+1},size:l.size}));this.groups=lLe(o,a,r),this._size=this._paddingTop+this.groups.reduce((l,c)=>l+c.size*(c.range.end-c.range.start),0)}get count(){const e=this.groups.length;return e?this.groups[e-1].range.end:0}get size(){return this._size}indexAt(e){if(e<0)return-1;if(e<this._paddingTop)return 0;let t=0,i=this._paddingTop;for(const n of this.groups){const o=n.range.end-n.range.start,r=i+o*n.size;if(e<r)return t+Math.floor((e-i)/n.size);t+=o,i=r}return t}indexAfter(e){return Math.min(this.indexAt(e)+1,this.count)}positionAt(e){if(e<0)return-1;let t=0,i=0;for(const n of this.groups){const o=n.range.end-n.range.start,r=i+o;if(e<r)return this._paddingTop+t+(e-i)*n.size;t+=o*n.size,i=r}return-1}}class dLe{constructor(e){this.renderers=e,this.cache=new Map,this.transactionNodesPendingRemoval=new Set,this.inTransaction=!1}alloc(e){let t=this.getTemplateCache(e).pop(),i=!1;if(t)i=this.transactionNodesPendingRemoval.has(t.domNode),i&&this.transactionNodesPendingRemoval.delete(t.domNode);else{const n=ie(".monaco-list-row"),r=this.getRenderer(e).renderTemplate(n);t={domNode:n,templateId:e,templateData:r}}return{row:t,isReusingConnectedDomNode:i}}release(e){e&&this.releaseRow(e)}transact(e){if(this.inTransaction)throw new Error("Already in transaction");this.inTransaction=!0;try{e()}finally{for(const t of this.transactionNodesPendingRemoval)this.doRemoveNode(t);this.transactionNodesPendingRemoval.clear(),this.inTransaction=!1}}releaseRow(e){const{domNode:t,templateId:i}=e;t&&(this.inTransaction?this.transactionNodesPendingRemoval.add(t):this.doRemoveNode(t)),this.getTemplateCache(i).push(e)}doRemoveNode(e){e.classList.remove("scrolling"),e.remove()}getTemplateCache(e){let t=this.cache.get(e);return t||(t=[],this.cache.set(e,t)),t}dispose(){this.cache.forEach((e,t)=>{for(const i of e)this.getRenderer(t).disposeTemplate(i.templateData),i.templateData=null}),this.cache.clear(),this.transactionNodesPendingRemoval.clear()}getRenderer(e){const t=this.renderers.get(e);if(!t)throw new Error(`No renderer found for ${e}`);return t}}function hLe(s,e,t,i=[]){if(!s.dataTransfer)return;const n=ie(".monaco-drag-image");n.textContent=t,n.classList.add(...i),(a=>{for(;a&&!a.classList.contains("monaco-workbench");)a=a.parentElement;return a||e.ownerDocument.body})(e).appendChild(n),s.dataTransfer.setDragImage(n,-10,-10),setTimeout(()=>n.remove(),0)}var Nh=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o};const qg={CurrentDragAndDropData:void 0},Wl={useShadows:!0,verticalScrollMode:1,setRowLineHeight:!0,setRowHeight:!0,supportDynamicHeights:!1,dnd:{getDragElements(s){return[s]},getDragURI(){return null},onDragStart(){},onDragOver(){return!1},drop(){},dispose(){}},horizontalScrolling:!1,transformOptimization:!0,alwaysConsumeMouseWheel:!0};class cx{constructor(e){this.elements=e}update(){}getData(){return this.elements}}class uLe{constructor(e){this.elements=e}update(){}getData(){return this.elements}}class gLe{constructor(){this.types=[],this.files=[]}update(e){if(e.types&&this.types.splice(0,this.types.length,...e.types),e.files){this.files.splice(0,this.files.length);for(let t=0;t<e.files.length;t++){const i=e.files.item(t);i&&(i.size||i.type)&&this.files.push(i)}}}getData(){return{types:this.types,files:this.files}}}function fLe(s,e){return Array.isArray(s)&&Array.isArray(e)?ni(s,e):s===e}class pLe{constructor(e){e?.getSetSize?this.getSetSize=e.getSetSize.bind(e):this.getSetSize=(t,i,n)=>n,e?.getPosInSet?this.getPosInSet=e.getPosInSet.bind(e):this.getPosInSet=(t,i)=>i+1,e?.getRole?this.getRole=e.getRole.bind(e):this.getRole=t=>"listitem",e?.isChecked?this.isChecked=e.isChecked.bind(e):this.isChecked=t=>{}}}const rN=class rN{get contentHeight(){return this.rangeMap.size}get onDidScroll(){return this.scrollableElement.onScroll}get scrollableElementDomNode(){return this.scrollableElement.getDomNode()}get horizontalScrolling(){return this._horizontalScrolling}set horizontalScrolling(e){if(e!==this._horizontalScrolling){if(e&&this.supportDynamicHeights)throw new Error("Horizontal scrolling and dynamic heights not supported simultaneously");if(this._horizontalScrolling=e,this.domNode.classList.toggle("horizontal-scrolling",this._horizontalScrolling),this._horizontalScrolling){for(const t of this.items)this.measureItemWidth(t);this.updateScrollWidth(),this.scrollableElement.setScrollDimensions({width:H2(this.domNode)}),this.rowsContainer.style.width=`${Math.max(this.scrollWidth||0,this.renderWidth)}px`}else this.scrollableElementWidthDelayer.cancel(),this.scrollableElement.setScrollDimensions({width:this.renderWidth,scrollWidth:this.renderWidth}),this.rowsContainer.style.width=""}}constructor(e,t,i,n=Wl){if(this.virtualDelegate=t,this.domId=`list_id_${++rN.InstanceCount}`,this.renderers=new Map,this.renderWidth=0,this._scrollHeight=0,this.scrollableElementUpdateDisposable=null,this.scrollableElementWidthDelayer=new sa(50),this.splicing=!1,this.dragOverAnimationStopDisposable=F.None,this.dragOverMouseY=0,this.canDrop=!1,this.currentDragFeedbackDisposable=F.None,this.onDragLeaveTimeout=F.None,this.currentSelectionDisposable=F.None,this.disposables=new U,this._onDidChangeContentHeight=new M,this._onDidChangeContentWidth=new M,this.onDidChangeContentHeight=oe.latch(this._onDidChangeContentHeight.event,void 0,this.disposables),this._horizontalScrolling=!1,n.horizontalScrolling&&n.supportDynamicHeights)throw new Error("Horizontal scrolling and dynamic heights not supported simultaneously");this.items=[],this.itemId=0,this.rangeMap=this.createRangeMap(n.paddingTop??0);for(const r of i)this.renderers.set(r.templateId,r);if(this.cache=this.disposables.add(new dLe(this.renderers)),this.lastRenderTop=0,this.lastRenderHeight=0,this.domNode=document.createElement("div"),this.domNode.className="monaco-list",this.domNode.classList.add(this.domId),this.domNode.tabIndex=0,this.domNode.classList.toggle("mouse-support",typeof n.mouseSupport=="boolean"?n.mouseSupport:!0),this._horizontalScrolling=n.horizontalScrolling??Wl.horizontalScrolling,this.domNode.classList.toggle("horizontal-scrolling",this._horizontalScrolling),this.paddingBottom=typeof n.paddingBottom>"u"?0:n.paddingBottom,this.accessibilityProvider=new pLe(n.accessibilityProvider),this.rowsContainer=document.createElement("div"),this.rowsContainer.className="monaco-list-rows",(n.transformOptimization??Wl.transformOptimization)&&(this.rowsContainer.style.transform="translate3d(0px, 0px, 0px)",this.rowsContainer.style.overflow="hidden",this.rowsContainer.style.contain="strict"),this.disposables.add(_s.addTarget(this.rowsContainer)),this.scrollable=this.disposables.add(new sC({forceIntegerValues:!0,smoothScrollDuration:n.smoothScrolling??!1?125:0,scheduleAtNextAnimationFrame:r=>Co(be(this.domNode),r)})),this.scrollableElement=this.disposables.add(new cA(this.rowsContainer,{alwaysConsumeMouseWheel:n.alwaysConsumeMouseWheel??Wl.alwaysConsumeMouseWheel,horizontal:1,vertical:n.verticalScrollMode??Wl.verticalScrollMode,useShadows:n.useShadows??Wl.useShadows,mouseWheelScrollSensitivity:n.mouseWheelScrollSensitivity,fastScrollSensitivity:n.fastScrollSensitivity,scrollByPage:n.scrollByPage},this.scrollable)),this.domNode.appendChild(this.scrollableElement.getDomNode()),e.appendChild(this.domNode),this.scrollableElement.onScroll(this.onScroll,this,this.disposables),this.disposables.add(W(this.rowsContainer,qt.Change,r=>this.onTouchChange(r))),this.disposables.add(W(this.scrollableElement.getDomNode(),"scroll",r=>{const a=r.target,l=a.scrollTop;a.scrollTop=0,n.scrollToActiveElement&&this.setScrollTop(this.scrollTop+l)})),this.disposables.add(W(this.domNode,"dragover",r=>this.onDragOver(this.toDragEvent(r)))),this.disposables.add(W(this.domNode,"drop",r=>this.onDrop(this.toDragEvent(r)))),this.disposables.add(W(this.domNode,"dragleave",r=>this.onDragLeave(this.toDragEvent(r)))),this.disposables.add(W(this.domNode,"dragend",r=>this.onDragEnd(r))),n.userSelection){if(n.dnd)throw new Error("DND and user selection cannot be used simultaneously");this.disposables.add(W(this.domNode,"mousedown",r=>this.onPotentialSelectionStart(r)))}this.setRowLineHeight=n.setRowLineHeight??Wl.setRowLineHeight,this.setRowHeight=n.setRowHeight??Wl.setRowHeight,this.supportDynamicHeights=n.supportDynamicHeights??Wl.supportDynamicHeights,this.dnd=n.dnd??this.disposables.add(Wl.dnd),this.layout(n.initialSize?.height,n.initialSize?.width),n.scrollToActiveElement&&this._setupFocusObserver(e)}_setupFocusObserver(e){this.disposables.add(W(e,"focus",()=>{const t=ln();this.activeElement!==t&&t!==null&&(this.activeElement=t,this._scrollToActiveElement(this.activeElement,e))},!0))}_scrollToActiveElement(e,t){const i=t.getBoundingClientRect(),o=e.getBoundingClientRect().top-i.top;o<0&&this.setScrollTop(this.scrollTop+o)}updateOptions(e){e.paddingBottom!==void 0&&(this.paddingBottom=e.paddingBottom,this.scrollableElement.setScrollDimensions({scrollHeight:this.scrollHeight})),e.smoothScrolling!==void 0&&this.scrollable.setSmoothScrollDuration(e.smoothScrolling?125:0),e.horizontalScrolling!==void 0&&(this.horizontalScrolling=e.horizontalScrolling);let t;if(e.scrollByPage!==void 0&&(t={...t??{},scrollByPage:e.scrollByPage}),e.mouseWheelScrollSensitivity!==void 0&&(t={...t??{},mouseWheelScrollSensitivity:e.mouseWheelScrollSensitivity}),e.fastScrollSensitivity!==void 0&&(t={...t??{},fastScrollSensitivity:e.fastScrollSensitivity}),t&&this.scrollableElement.updateOptions(t),e.paddingTop!==void 0&&e.paddingTop!==this.rangeMap.paddingTop){const i=this.getRenderRange(this.lastRenderTop,this.lastRenderHeight),n=e.paddingTop-this.rangeMap.paddingTop;this.rangeMap.paddingTop=e.paddingTop,this.render(i,Math.max(0,this.lastRenderTop+n),this.lastRenderHeight,void 0,void 0,!0),this.setScrollTop(this.lastRenderTop),this.eventuallyUpdateScrollDimensions(),this.supportDynamicHeights&&this._rerender(this.lastRenderTop,this.lastRenderHeight)}}createRangeMap(e){return new cLe(e)}splice(e,t,i=[]){if(this.splicing)throw new Error("Can't run recursive splices.");this.splicing=!0;try{return this._splice(e,t,i)}finally{this.splicing=!1,this._onDidChangeContentHeight.fire(this.contentHeight)}}_splice(e,t,i=[]){const n=this.getRenderRange(this.lastRenderTop,this.lastRenderHeight),o={start:e,end:e+t},r=as.intersect(n,o),a=new Map;for(let y=r.end-1;y>=r.start;y--){const x=this.items[y];if(x.dragStartDisposable.dispose(),x.checkedDisposable.dispose(),x.row){let L=a.get(x.templateId);L||(L=[],a.set(x.templateId,L));const E=this.renderers.get(x.templateId);E&&E.disposeElement&&E.disposeElement(x.element,y,x.row.templateData,{height:x.size}),L.unshift(x.row)}x.row=null,x.stale=!0}const l={start:e+t,end:this.items.length},c=as.intersect(l,n),d=as.relativeComplement(l,n),h=i.map(y=>({id:String(this.itemId++),element:y,templateId:this.virtualDelegate.getTemplateId(y),size:this.virtualDelegate.getHeight(y),width:void 0,hasDynamicHeight:!!this.virtualDelegate.hasDynamicHeight&&this.virtualDelegate.hasDynamicHeight(y),lastDynamicHeightWidth:void 0,row:null,uri:void 0,dropTarget:!1,dragStartDisposable:F.None,checkedDisposable:F.None,stale:!1}));let u;e===0&&t>=this.items.length?(this.rangeMap=this.createRangeMap(this.rangeMap.paddingTop),this.rangeMap.splice(0,0,h),u=this.items,this.items=h):(this.rangeMap.splice(e,t,h),u=CE(this.items,e,t,h));const g=i.length-t,f=this.getRenderRange(this.lastRenderTop,this.lastRenderHeight),p=s4(c,g),_=as.intersect(f,p);for(let y=_.start;y<_.end;y++)this.updateItemInDOM(this.items[y],y);const b=as.relativeComplement(p,f);for(const y of b)for(let x=y.start;x<y.end;x++)this.removeItemFromDOM(x);const w=d.map(y=>s4(y,g)),v=[{start:e,end:e+i.length},...w].map(y=>as.intersect(f,y)).reverse();for(const y of v)for(let x=y.end-1;x>=y.start;x--){const L=this.items[x],I=a.get(L.templateId)?.pop();this.insertItemInDOM(x,I)}for(const y of a.values())for(const x of y)this.cache.release(x);return this.eventuallyUpdateScrollDimensions(),this.supportDynamicHeights&&this._rerender(this.scrollTop,this.renderHeight),u.map(y=>y.element)}eventuallyUpdateScrollDimensions(){this._scrollHeight=this.contentHeight,this.rowsContainer.style.height=`${this._scrollHeight}px`,this.scrollableElementUpdateDisposable||(this.scrollableElementUpdateDisposable=Co(be(this.domNode),()=>{this.scrollableElement.setScrollDimensions({scrollHeight:this.scrollHeight}),this.updateScrollWidth(),this.scrollableElementUpdateDisposable=null}))}eventuallyUpdateScrollWidth(){if(!this.horizontalScrolling){this.scrollableElementWidthDelayer.cancel();return}this.scrollableElementWidthDelayer.trigger(()=>this.updateScrollWidth())}updateScrollWidth(){if(!this.horizontalScrolling)return;let e=0;for(const t of this.items)typeof t.width<"u"&&(e=Math.max(e,t.width));this.scrollWidth=e,this.scrollableElement.setScrollDimensions({scrollWidth:e===0?0:e+10}),this._onDidChangeContentWidth.fire(this.scrollWidth)}rerender(){if(this.supportDynamicHeights){for(const e of this.items)e.lastDynamicHeightWidth=void 0;this._rerender(this.lastRenderTop,this.lastRenderHeight)}}get length(){return this.items.length}get renderHeight(){return this.scrollableElement.getScrollDimensions().height}get firstVisibleIndex(){return this.getVisibleRange(this.lastRenderTop,this.lastRenderHeight).start}element(e){return this.items[e].element}indexOf(e){return this.items.findIndex(t=>t.element===e)}domElement(e){const t=this.items[e].row;return t&&t.domNode}elementHeight(e){return this.items[e].size}elementTop(e){return this.rangeMap.positionAt(e)}indexAt(e){return this.rangeMap.indexAt(e)}indexAfter(e){return this.rangeMap.indexAfter(e)}layout(e,t){const i={height:typeof e=="number"?e:oue(this.domNode)};this.scrollableElementUpdateDisposable&&(this.scrollableElementUpdateDisposable.dispose(),this.scrollableElementUpdateDisposable=null,i.scrollHeight=this.scrollHeight),this.scrollableElement.setScrollDimensions(i),typeof t<"u"&&(this.renderWidth=t,this.supportDynamicHeights&&this._rerender(this.scrollTop,this.renderHeight)),this.horizontalScrolling&&this.scrollableElement.setScrollDimensions({width:typeof t=="number"?t:H2(this.domNode)})}render(e,t,i,n,o,r=!1,a=!1){const l=this.getRenderRange(t,i),c=as.relativeComplement(l,e).reverse(),d=as.relativeComplement(e,l);if(r){const h=as.intersect(e,l);for(let u=h.start;u<h.end;u++)this.updateItemInDOM(this.items[u],u)}this.cache.transact(()=>{for(const h of d)for(let u=h.start;u<h.end;u++)this.removeItemFromDOM(u,a);for(const h of c)for(let u=h.end-1;u>=h.start;u--)this.insertItemInDOM(u)}),n!==void 0&&(this.rowsContainer.style.left=`-${n}px`),this.rowsContainer.style.top=`-${t}px`,this.horizontalScrolling&&o!==void 0&&(this.rowsContainer.style.width=`${Math.max(o,this.renderWidth)}px`),this.lastRenderTop=t,this.lastRenderHeight=i}insertItemInDOM(e,t){const i=this.items[e];if(!i.row)if(t)i.row=t,i.stale=!0;else{const c=this.cache.alloc(i.templateId);i.row=c.row,i.stale||=c.isReusingConnectedDomNode}const n=this.accessibilityProvider.getRole(i.element)||"listitem";i.row.domNode.setAttribute("role",n);const o=this.accessibilityProvider.isChecked(i.element),r=c=>c==="mixed"?"mixed":String(!!c);if(typeof o=="boolean"||o==="mixed")i.row.domNode.setAttribute("aria-checked",r(o));else if(o){const c=d=>i.row.domNode.setAttribute("aria-checked",r(d));c(o.value),i.checkedDisposable=o.onDidChange(()=>c(o.value))}if(i.stale||!i.row.domNode.parentElement){const c=this.items.at(e+1)?.row?.domNode??null;(i.row.domNode.parentElement!==this.rowsContainer||i.row.domNode.nextElementSibling!==c)&&this.rowsContainer.insertBefore(i.row.domNode,c),i.stale=!1}this.updateItemInDOM(i,e);const a=this.renderers.get(i.templateId);if(!a)throw new Error(`No renderer found for template id ${i.templateId}`);a?.renderElement(i.element,e,i.row.templateData,{height:i.size});const l=this.dnd.getDragURI(i.element);i.dragStartDisposable.dispose(),i.row.domNode.draggable=!!l,l&&(i.dragStartDisposable=W(i.row.domNode,"dragstart",c=>this.onDragStart(i.element,l,c))),this.horizontalScrolling&&(this.measureItemWidth(i),this.eventuallyUpdateScrollWidth())}measureItemWidth(e){if(!e.row||!e.row.domNode)return;e.row.domNode.style.width="fit-content",e.width=H2(e.row.domNode);const t=be(e.row.domNode).getComputedStyle(e.row.domNode);t.paddingLeft&&(e.width+=parseFloat(t.paddingLeft)),t.paddingRight&&(e.width+=parseFloat(t.paddingRight)),e.row.domNode.style.width=""}updateItemInDOM(e,t){e.row.domNode.style.top=`${this.elementTop(t)}px`,this.setRowHeight&&(e.row.domNode.style.height=`${e.size}px`),this.setRowLineHeight&&(e.row.domNode.style.lineHeight=`${e.size}px`),e.row.domNode.setAttribute("data-index",`${t}`),e.row.domNode.setAttribute("data-last-element",t===this.length-1?"true":"false"),e.row.domNode.setAttribute("data-parity",t%2===0?"even":"odd"),e.row.domNode.setAttribute("aria-setsize",String(this.accessibilityProvider.getSetSize(e.element,t,this.length))),e.row.domNode.setAttribute("aria-posinset",String(this.accessibilityProvider.getPosInSet(e.element,t))),e.row.domNode.setAttribute("id",this.getElementDomId(t)),e.row.domNode.classList.toggle("drop-target",e.dropTarget)}removeItemFromDOM(e,t){const i=this.items[e];if(i.dragStartDisposable.dispose(),i.checkedDisposable.dispose(),i.row){const n=this.renderers.get(i.templateId);n&&n.disposeElement&&n.disposeElement(i.element,e,i.row.templateData,{height:i.size,onScroll:t}),this.cache.release(i.row),i.row=null}this.horizontalScrolling&&this.eventuallyUpdateScrollWidth()}getScrollTop(){return this.scrollableElement.getScrollPosition().scrollTop}setScrollTop(e,t){this.scrollableElementUpdateDisposable&&(this.scrollableElementUpdateDisposable.dispose(),this.scrollableElementUpdateDisposable=null,this.scrollableElement.setScrollDimensions({scrollHeight:this.scrollHeight})),this.scrollableElement.setScrollPosition({scrollTop:e,reuseAnimation:t})}get scrollTop(){return this.getScrollTop()}set scrollTop(e){this.setScrollTop(e)}get scrollHeight(){return this._scrollHeight+(this.horizontalScrolling?10:0)+this.paddingBottom}get onMouseClick(){return oe.map(this.disposables.add(new Lt(this.domNode,"click")).event,e=>this.toMouseEvent(e),this.disposables)}get onMouseDblClick(){return oe.map(this.disposables.add(new Lt(this.domNode,"dblclick")).event,e=>this.toMouseEvent(e),this.disposables)}get onMouseMiddleClick(){return oe.filter(oe.map(this.disposables.add(new Lt(this.domNode,"auxclick")).event,e=>this.toMouseEvent(e),this.disposables),e=>e.browserEvent.button===1,this.disposables)}get onMouseDown(){return oe.map(this.disposables.add(new Lt(this.domNode,"mousedown")).event,e=>this.toMouseEvent(e),this.disposables)}get onMouseOver(){return oe.map(this.disposables.add(new Lt(this.domNode,"mouseover")).event,e=>this.toMouseEvent(e),this.disposables)}get onMouseOut(){return oe.map(this.disposables.add(new Lt(this.domNode,"mouseout")).event,e=>this.toMouseEvent(e),this.disposables)}get onContextMenu(){return oe.any(oe.map(this.disposables.add(new Lt(this.domNode,"contextmenu")).event,e=>this.toMouseEvent(e),this.disposables),oe.map(this.disposables.add(new Lt(this.domNode,qt.Contextmenu)).event,e=>this.toGestureEvent(e),this.disposables))}get onTouchStart(){return oe.map(this.disposables.add(new Lt(this.domNode,"touchstart")).event,e=>this.toTouchEvent(e),this.disposables)}get onTap(){return oe.map(this.disposables.add(new Lt(this.rowsContainer,qt.Tap)).event,e=>this.toGestureEvent(e),this.disposables)}toMouseEvent(e){const t=this.getItemIndexFromEventTarget(e.target||null),i=typeof t>"u"?void 0:this.items[t],n=i&&i.element;return{browserEvent:e,index:t,element:n}}toTouchEvent(e){const t=this.getItemIndexFromEventTarget(e.target||null),i=typeof t>"u"?void 0:this.items[t],n=i&&i.element;return{browserEvent:e,index:t,element:n}}toGestureEvent(e){const t=this.getItemIndexFromEventTarget(e.initialTarget||null),i=typeof t>"u"?void 0:this.items[t],n=i&&i.element;return{browserEvent:e,index:t,element:n}}toDragEvent(e){const t=this.getItemIndexFromEventTarget(e.target||null),i=typeof t>"u"?void 0:this.items[t],n=i&&i.element,o=this.getTargetSector(e,t);return{browserEvent:e,index:t,element:n,sector:o}}onScroll(e){try{const t=this.getRenderRange(this.lastRenderTop,this.lastRenderHeight);this.render(t,e.scrollTop,e.height,e.scrollLeft,e.scrollWidth,void 0,!0),this.supportDynamicHeights&&this._rerender(e.scrollTop,e.height,e.inSmoothScrolling)}catch(t){throw console.error("Got bad scroll event:",e),t}}onTouchChange(e){e.preventDefault(),e.stopPropagation(),this.scrollTop-=e.translationY}onDragStart(e,t,i){if(!i.dataTransfer)return;const n=this.dnd.getDragElements(e);i.dataTransfer.effectAllowed="copyMove",i.dataTransfer.setData(Mm.TEXT,t);let o;this.dnd.getDragLabel&&(o=this.dnd.getDragLabel(n,i)),typeof o>"u"&&(o=String(n.length)),hLe(i,this.domNode,o,[this.domId]),this.domNode.classList.add("dragging"),this.currentDragData=new cx(n),qg.CurrentDragAndDropData=new uLe(n),this.dnd.onDragStart?.(this.currentDragData,i)}onPotentialSelectionStart(e){this.currentSelectionDisposable.dispose();const t=Zhe(this.domNode),i=this.currentSelectionDisposable=new U,n=i.add(new U);n.add(W(this.domNode,"selectstart",()=>{n.add(W(t,"mousemove",o=>{t.getSelection()?.isCollapsed===!1&&this.setupDragAndDropScrollTopAnimation(o)})),i.add(pe(()=>{const o=this.getRenderRange(this.lastRenderTop,this.lastRenderHeight);this.currentSelectionBounds=void 0,this.render(o,this.lastRenderTop,this.lastRenderHeight,void 0,void 0)})),i.add(W(t,"selectionchange",()=>{const o=t.getSelection();if(!o||o.isCollapsed){n.isDisposed&&i.dispose();return}let r=this.getIndexOfListElement(o.anchorNode),a=this.getIndexOfListElement(o.focusNode);r!==void 0&&a!==void 0&&(a<r&&([r,a]=[a,r]),this.currentSelectionBounds={start:r,end:a})}))})),n.add(W(t,"mouseup",()=>{n.dispose(),this.teardownDragAndDropScrollTopAnimation(),t.getSelection()?.isCollapsed!==!1&&i.dispose()}))}getIndexOfListElement(e){if(!(!e||!this.domNode.contains(e)))for(;e&&e!==this.domNode;){if(e.dataset?.index)return Number(e.dataset.index);e=e.parentElement}}onDragOver(e){if(e.browserEvent.preventDefault(),this.onDragLeaveTimeout.dispose(),qg.CurrentDragAndDropData&&qg.CurrentDragAndDropData.getData()==="vscode-ui"||(this.setupDragAndDropScrollTopAnimation(e.browserEvent),!e.browserEvent.dataTransfer))return!1;if(!this.currentDragData)if(qg.CurrentDragAndDropData)this.currentDragData=qg.CurrentDragAndDropData;else{if(!e.browserEvent.dataTransfer.types)return!1;this.currentDragData=new gLe}const t=this.dnd.onDragOver(this.currentDragData,e.element,e.index,e.sector,e.browserEvent);if(this.canDrop=typeof t=="boolean"?t:t.accept,!this.canDrop)return this.currentDragFeedback=void 0,this.currentDragFeedbackDisposable.dispose(),!1;e.browserEvent.dataTransfer.dropEffect=typeof t!="boolean"&&t.effect?.type===0?"copy":"move";let i;typeof t!="boolean"&&t.feedback?i=t.feedback:typeof e.index>"u"?i=[-1]:i=[e.index],i=fh(i).filter(o=>o>=-1&&o<this.length).sort((o,r)=>o-r),i=i[0]===-1?[-1]:i;let n=typeof t!="boolean"&&t.effect&&t.effect.position?t.effect.position:"drop-target";if(fLe(this.currentDragFeedback,i)&&this.currentDragFeedbackPosition===n)return!0;if(this.currentDragFeedback=i,this.currentDragFeedbackPosition=n,this.currentDragFeedbackDisposable.dispose(),i[0]===-1)this.domNode.classList.add(n),this.rowsContainer.classList.add(n),this.currentDragFeedbackDisposable=pe(()=>{this.domNode.classList.remove(n),this.rowsContainer.classList.remove(n)});else{if(i.length>1&&n!=="drop-target")throw new Error("Can't use multiple feedbacks with position different than 'over'");n==="drop-target-after"&&i[0]<this.length-1&&(i[0]+=1,n="drop-target-before");for(const o of i){const r=this.items[o];r.dropTarget=!0,r.row?.domNode.classList.add(n)}this.currentDragFeedbackDisposable=pe(()=>{for(const o of i){const r=this.items[o];r.dropTarget=!1,r.row?.domNode.classList.remove(n)}})}return!0}onDragLeave(e){this.onDragLeaveTimeout.dispose(),this.onDragLeaveTimeout=hh(()=>this.clearDragOverFeedback(),100,this.disposables),this.currentDragData&&this.dnd.onDragLeave?.(this.currentDragData,e.element,e.index,e.browserEvent)}onDrop(e){if(!this.canDrop)return;const t=this.currentDragData;this.teardownDragAndDropScrollTopAnimation(),this.clearDragOverFeedback(),this.domNode.classList.remove("dragging"),this.currentDragData=void 0,qg.CurrentDragAndDropData=void 0,!(!t||!e.browserEvent.dataTransfer)&&(e.browserEvent.preventDefault(),t.update(e.browserEvent.dataTransfer),this.dnd.drop(t,e.element,e.index,e.sector,e.browserEvent))}onDragEnd(e){this.canDrop=!1,this.teardownDragAndDropScrollTopAnimation(),this.clearDragOverFeedback(),this.domNode.classList.remove("dragging"),this.currentDragData=void 0,qg.CurrentDragAndDropData=void 0,this.dnd.onDragEnd?.(e)}clearDragOverFeedback(){this.currentDragFeedback=void 0,this.currentDragFeedbackPosition=void 0,this.currentDragFeedbackDisposable.dispose(),this.currentDragFeedbackDisposable=F.None}setupDragAndDropScrollTopAnimation(e){if(!this.dragOverAnimationDisposable){const t=DQ(this.domNode).top;this.dragOverAnimationDisposable=pue(be(this.domNode),this.animateDragAndDropScrollTop.bind(this,t))}this.dragOverAnimationStopDisposable.dispose(),this.dragOverAnimationStopDisposable=hh(()=>{this.dragOverAnimationDisposable&&(this.dragOverAnimationDisposable.dispose(),this.dragOverAnimationDisposable=void 0)},1e3,this.disposables),this.dragOverMouseY=e.pageY}animateDragAndDropScrollTop(e){if(this.dragOverMouseY===void 0)return;const t=this.dragOverMouseY-e,i=this.renderHeight-35;t<35?this.scrollTop+=Math.max(-14,Math.floor(.3*(t-35))):t>i&&(this.scrollTop+=Math.min(14,Math.floor(.3*(t-i))))}teardownDragAndDropScrollTopAnimation(){this.dragOverAnimationStopDisposable.dispose(),this.dragOverAnimationDisposable&&(this.dragOverAnimationDisposable.dispose(),this.dragOverAnimationDisposable=void 0)}getTargetSector(e,t){if(t===void 0)return;const i=e.offsetY/this.items[t].size,n=Math.floor(i/.25);return qs(n,0,3)}getItemIndexFromEventTarget(e){const t=this.scrollableElement.getDomNode();let i=e;for(;(ki(i)||AQ(i))&&i!==this.rowsContainer&&t.contains(i);){const n=i.getAttribute("data-index");if(n){const o=Number(n);if(!isNaN(o))return o}i=i.parentElement}}getVisibleRange(e,t){return{start:this.rangeMap.indexAt(e),end:this.rangeMap.indexAfter(e+t-1)}}getRenderRange(e,t){const i=this.getVisibleRange(e,t);if(this.currentSelectionBounds){const n=this.rangeMap.count;i.start=Math.min(i.start,this.currentSelectionBounds.start,n),i.end=Math.min(Math.max(i.end,this.currentSelectionBounds.end+1),n)}return i}_rerender(e,t,i){const n=this.getRenderRange(e,t);let o,r;e===this.elementTop(n.start)?(o=n.start,r=0):n.end-n.start>1&&(o=n.start+1,r=this.elementTop(o)-e);let a=0;for(;;){const l=this.getRenderRange(e,t);let c=!1;for(let d=l.start;d<l.end;d++){const h=this.probeDynamicHeight(d);h!==0&&this.rangeMap.splice(d,1,[this.items[d]]),a+=h,c=c||h!==0}if(!c){a!==0&&this.eventuallyUpdateScrollDimensions();const d=as.relativeComplement(n,l);for(const u of d)for(let g=u.start;g<u.end;g++)this.items[g].row&&this.removeItemFromDOM(g);const h=as.relativeComplement(l,n).reverse();for(const u of h)for(let g=u.end-1;g>=u.start;g--)this.insertItemInDOM(g);for(let u=l.start;u<l.end;u++)this.items[u].row&&this.updateItemInDOM(this.items[u],u);if(typeof o=="number"){const u=this.scrollable.getFutureScrollPosition().scrollTop-e,g=this.elementTop(o)-r+u;this.setScrollTop(g,i)}this._onDidChangeContentHeight.fire(this.contentHeight);return}}}probeDynamicHeight(e){const t=this.items[e];if(this.virtualDelegate.getDynamicHeight){const r=this.virtualDelegate.getDynamicHeight(t.element);if(r!==null){const a=t.size;return t.size=r,t.lastDynamicHeightWidth=this.renderWidth,r-a}}if(!t.hasDynamicHeight||t.lastDynamicHeightWidth===this.renderWidth||this.virtualDelegate.hasDynamicHeight&&!this.virtualDelegate.hasDynamicHeight(t.element))return 0;const i=t.size;if(t.row)return t.row.domNode.style.height="",t.size=t.row.domNode.offsetHeight,t.size===0&&!wn(t.row.domNode,be(t.row.domNode).document.body)&&console.warn("Measuring item node that is not in DOM! Add ListView to the DOM before measuring row height!",new Error().stack),t.lastDynamicHeightWidth=this.renderWidth,t.size-i;const{row:n}=this.cache.alloc(t.templateId);n.domNode.style.height="",this.rowsContainer.appendChild(n.domNode);const o=this.renderers.get(t.templateId);if(!o)throw new Ce("Missing renderer for templateId: "+t.templateId);return o.renderElement(t.element,e,n.templateData),t.size=n.domNode.offsetHeight,o.disposeElement?.(t.element,e,n.templateData),this.virtualDelegate.setDynamicHeight?.(t.element,t.size),t.lastDynamicHeightWidth=this.renderWidth,n.domNode.remove(),this.cache.release(n),t.size-i}getElementDomId(e){return`${this.domId}_${e}`}dispose(){for(const e of this.items)if(e.dragStartDisposable.dispose(),e.checkedDisposable.dispose(),e.row){const t=this.renderers.get(e.row.templateId);t&&(t.disposeElement?.(e.element,-1,e.row.templateData,void 0),t.disposeTemplate(e.row.templateData))}this.items=[],this.domNode?.remove(),this.dragOverAnimationDisposable?.dispose(),this.disposables.dispose()}};rN.InstanceCount=0;let Wa=rN;Nh([Ai],Wa.prototype,"onMouseClick",null);Nh([Ai],Wa.prototype,"onMouseDblClick",null);Nh([Ai],Wa.prototype,"onMouseMiddleClick",null);Nh([Ai],Wa.prototype,"onMouseDown",null);Nh([Ai],Wa.prototype,"onMouseOver",null);Nh([Ai],Wa.prototype,"onMouseOut",null);Nh([Ai],Wa.prototype,"onContextMenu",null);Nh([Ai],Wa.prototype,"onTouchStart",null);Nh([Ai],Wa.prototype,"onTap",null);var Ng=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o};class mLe{constructor(e){this.trait=e,this.renderedElements=[]}get templateId(){return`template:${this.trait.name}`}renderTemplate(e){return e}renderElement(e,t,i){const n=this.renderedElements.findIndex(o=>o.templateData===i);if(n>=0){const o=this.renderedElements[n];this.trait.unrender(i),o.index=t}else{const o={index:t,templateData:i};this.renderedElements.push(o)}this.trait.renderIndex(t,i)}splice(e,t,i){const n=[];for(const o of this.renderedElements)o.index<e?n.push(o):o.index>=e+t&&n.push({index:o.index+i-t,templateData:o.templateData});this.renderedElements=n}renderIndexes(e){for(const{index:t,templateData:i}of this.renderedElements)e.indexOf(t)>-1&&this.trait.renderIndex(t,i)}disposeTemplate(e){const t=this.renderedElements.findIndex(i=>i.templateData===e);t<0||this.renderedElements.splice(t,1)}}let vI=class{get onChange(){return this._onChange.event}get name(){return this._trait}get renderer(){return new mLe(this)}constructor(e){this._trait=e,this.indexes=[],this.sortedIndexes=[],this._onChange=new M}splice(e,t,i){const n=i.length-t,o=e+t,r=[];let a=0;for(;a<this.sortedIndexes.length&&this.sortedIndexes[a]<e;)r.push(this.sortedIndexes[a++]);for(let l=0;l<i.length;l++)i[l]&&r.push(l+e);for(;a<this.sortedIndexes.length&&this.sortedIndexes[a]>=o;)r.push(this.sortedIndexes[a++]+n);this.renderer.splice(e,t,i.length),this._set(r,r)}renderIndex(e,t){t.classList.toggle(this._trait,this.contains(e))}unrender(e){e.classList.remove(this._trait)}set(e,t){return this._set(e,[...e].sort(Nq),t)}_set(e,t,i){const n=this.indexes,o=this.sortedIndexes;this.indexes=e,this.sortedIndexes=t;const r=o4(o,e);return this.renderer.renderIndexes(r),this._onChange.fire({indexes:e,browserEvent:i}),n}get(){return this.indexes}contains(e){return vE(this.sortedIndexes,e,Nq)>=0}dispose(){xt(this._onChange)}};Ng([Ai],vI.prototype,"renderer",null);class _Le extends vI{constructor(e){super("selected"),this.setAriaSelected=e}renderIndex(e,t){super.renderIndex(e,t),this.setAriaSelected&&(this.contains(e)?t.setAttribute("aria-selected","true"):t.setAttribute("aria-selected","false"))}}class PM{constructor(e,t,i){this.trait=e,this.view=t,this.identityProvider=i}splice(e,t,i){if(!this.identityProvider)return this.trait.splice(e,t,new Array(i.length).fill(!1));const n=this.trait.get().map(a=>this.identityProvider.getId(this.view.element(a)).toString());if(n.length===0)return this.trait.splice(e,t,new Array(i.length).fill(!1));const o=new Set(n),r=i.map(a=>o.has(this.identityProvider.getId(a).toString()));this.trait.splice(e,t,r)}}function dx(s,e){return s.classList.contains(e)?!0:s.classList.contains("monaco-list")||!s.parentElement?!1:dx(s.parentElement,e)}function h1(s){return dx(s,"monaco-editor")}function bLe(s){return dx(s,"monaco-custom-toggle")}function wLe(s){return dx(s,"action-item")}function G1(s){return dx(s,"monaco-tree-sticky-row")}function MS(s){return s.classList.contains("monaco-tree-sticky-container")}function oie(s){return s.tagName==="A"&&s.classList.contains("monaco-button")||s.tagName==="DIV"&&s.classList.contains("monaco-button-dropdown")?!0:s.classList.contains("monaco-list")||!s.parentElement?!1:oie(s.parentElement)}class rie{get onKeyDown(){return oe.chain(this.disposables.add(new Lt(this.view.domNode,"keydown")).event,e=>e.filter(t=>!wl(t.target)).map(t=>new Pt(t)))}constructor(e,t,i){this.list=e,this.view=t,this.disposables=new U,this.multipleSelectionDisposables=new U,this.multipleSelectionSupport=i.multipleSelectionSupport,this.disposables.add(this.onKeyDown(n=>{switch(n.keyCode){case 3:return this.onEnter(n);case 16:return this.onUpArrow(n);case 18:return this.onDownArrow(n);case 11:return this.onPageUpArrow(n);case 12:return this.onPageDownArrow(n);case 9:return this.onEscape(n);case 31:this.multipleSelectionSupport&&(Ye?n.metaKey:n.ctrlKey)&&this.onCtrlA(n)}}))}updateOptions(e){e.multipleSelectionSupport!==void 0&&(this.multipleSelectionSupport=e.multipleSelectionSupport)}onEnter(e){e.preventDefault(),e.stopPropagation(),this.list.setSelection(this.list.getFocus(),e.browserEvent)}onUpArrow(e){e.preventDefault(),e.stopPropagation(),this.list.focusPrevious(1,!1,e.browserEvent);const t=this.list.getFocus()[0];this.list.setAnchor(t),this.list.reveal(t),this.view.domNode.focus()}onDownArrow(e){e.preventDefault(),e.stopPropagation(),this.list.focusNext(1,!1,e.browserEvent);const t=this.list.getFocus()[0];this.list.setAnchor(t),this.list.reveal(t),this.view.domNode.focus()}onPageUpArrow(e){e.preventDefault(),e.stopPropagation(),this.list.focusPreviousPage(e.browserEvent);const t=this.list.getFocus()[0];this.list.setAnchor(t),this.list.reveal(t),this.view.domNode.focus()}onPageDownArrow(e){e.preventDefault(),e.stopPropagation(),this.list.focusNextPage(e.browserEvent);const t=this.list.getFocus()[0];this.list.setAnchor(t),this.list.reveal(t),this.view.domNode.focus()}onCtrlA(e){e.preventDefault(),e.stopPropagation(),this.list.setSelection(Us(this.list.length),e.browserEvent),this.list.setAnchor(void 0),this.view.domNode.focus()}onEscape(e){this.list.getSelection().length&&(e.preventDefault(),e.stopPropagation(),this.list.setSelection([],e.browserEvent),this.list.setAnchor(void 0),this.view.domNode.focus())}dispose(){this.disposables.dispose(),this.multipleSelectionDisposables.dispose()}}Ng([Ai],rie.prototype,"onKeyDown",null);var uc;(function(s){s[s.Automatic=0]="Automatic",s[s.Trigger=1]="Trigger"})(uc||(uc={}));var Tb;(function(s){s[s.Idle=0]="Idle",s[s.Typing=1]="Typing"})(Tb||(Tb={}));const vLe=new class{mightProducePrintableCharacter(s){return s.ctrlKey||s.metaKey||s.altKey?!1:s.keyCode>=31&&s.keyCode<=56||s.keyCode>=21&&s.keyCode<=30||s.keyCode>=98&&s.keyCode<=107||s.keyCode>=85&&s.keyCode<=95}};class CLe{constructor(e,t,i,n,o){this.list=e,this.view=t,this.keyboardNavigationLabelProvider=i,this.keyboardNavigationEventFilter=n,this.delegate=o,this.enabled=!1,this.state=Tb.Idle,this.mode=uc.Automatic,this.triggered=!1,this.previouslyFocused=-1,this.enabledDisposables=new U,this.disposables=new U,this.updateOptions(e.options)}updateOptions(e){e.typeNavigationEnabled??!0?this.enable():this.disable(),this.mode=e.typeNavigationMode??uc.Automatic}enable(){if(this.enabled)return;let e=!1;const t=oe.chain(this.enabledDisposables.add(new Lt(this.view.domNode,"keydown")).event,o=>o.filter(r=>!wl(r.target)).filter(()=>this.mode===uc.Automatic||this.triggered).map(r=>new Pt(r)).filter(r=>e||this.keyboardNavigationEventFilter(r)).filter(r=>this.delegate.mightProducePrintableCharacter(r)).forEach(r=>gt.stop(r,!0)).map(r=>r.browserEvent.key)),i=oe.debounce(t,()=>null,800,void 0,void 0,void 0,this.enabledDisposables);oe.reduce(oe.any(t,i),(o,r)=>r===null?null:(o||"")+r,void 0,this.enabledDisposables)(this.onInput,this,this.enabledDisposables),i(this.onClear,this,this.enabledDisposables),t(()=>e=!0,void 0,this.enabledDisposables),i(()=>e=!1,void 0,this.enabledDisposables),this.enabled=!0,this.triggered=!1}disable(){this.enabled&&(this.enabledDisposables.clear(),this.enabled=!1,this.triggered=!1)}onClear(){const e=this.list.getFocus();if(e.length>0&&e[0]===this.previouslyFocused){const t=this.list.options.accessibilityProvider?.getAriaLabel(this.list.element(e[0]));typeof t=="string"?Js(t):t&&Js(t.get())}this.previouslyFocused=-1}onInput(e){if(!e){this.state=Tb.Idle,this.triggered=!1;return}const t=this.list.getFocus(),i=t.length>0?t[0]:0,n=this.state===Tb.Idle?1:0;this.state=Tb.Typing;for(let o=0;o<this.list.length;o++){const r=(i+o+n)%this.list.length,a=this.keyboardNavigationLabelProvider.getKeyboardNavigationLabel(this.view.element(r)),l=a&&a.toString();if(this.list.options.typeNavigationEnabled){if(typeof l<"u"){if(DS(e,l)){this.previouslyFocused=i,this.list.setFocus([r]),this.list.reveal(r);return}const c=o0e(e,l);if(c&&c[0].end-c[0].start>1&&c.length===1){this.previouslyFocused=i,this.list.setFocus([r]),this.list.reveal(r);return}}}else if(typeof l>"u"||DS(e,l)){this.previouslyFocused=i,this.list.setFocus([r]),this.list.reveal(r);return}}}dispose(){this.disable(),this.enabledDisposables.dispose(),this.disposables.dispose()}}class yLe{constructor(e,t){this.list=e,this.view=t,this.disposables=new U;const i=oe.chain(this.disposables.add(new Lt(t.domNode,"keydown")).event,o=>o.filter(r=>!wl(r.target)).map(r=>new Pt(r)));oe.chain(i,o=>o.filter(r=>r.keyCode===2&&!r.ctrlKey&&!r.metaKey&&!r.shiftKey&&!r.altKey))(this.onTab,this,this.disposables)}onTab(e){if(e.target!==this.view.domNode)return;const t=this.list.getFocus();if(t.length===0)return;const i=this.view.domElement(t[0]);if(!i)return;const n=i.querySelector("[tabIndex]");if(!n||!ki(n)||n.tabIndex===-1)return;const o=be(n).getComputedStyle(n);o.visibility==="hidden"||o.display==="none"||(e.preventDefault(),e.stopPropagation(),n.focus())}dispose(){this.disposables.dispose()}}function aie(s){return Ye?s.browserEvent.metaKey:s.browserEvent.ctrlKey}function lie(s){return s.browserEvent.shiftKey}function SLe(s){return dW(s)&&s.button===2}const Tq={isSelectionSingleChangeEvent:aie,isSelectionRangeChangeEvent:lie};class cie{get onPointer(){return this._onPointer.event}constructor(e){this.list=e,this.disposables=new U,this._onPointer=this.disposables.add(new M),e.options.multipleSelectionSupport!==!1&&(this.multipleSelectionController=this.list.options.multipleSelectionController||Tq),this.mouseSupport=typeof e.options.mouseSupport>"u"||!!e.options.mouseSupport,this.mouseSupport&&(e.onMouseDown(this.onMouseDown,this,this.disposables),e.onContextMenu(this.onContextMenu,this,this.disposables),e.onMouseDblClick(this.onDoubleClick,this,this.disposables),e.onTouchStart(this.onMouseDown,this,this.disposables),this.disposables.add(_s.addTarget(e.getHTMLElement()))),oe.any(e.onMouseClick,e.onMouseMiddleClick,e.onTap)(this.onViewPointer,this,this.disposables)}updateOptions(e){e.multipleSelectionSupport!==void 0&&(this.multipleSelectionController=void 0,e.multipleSelectionSupport&&(this.multipleSelectionController=this.list.options.multipleSelectionController||Tq))}isSelectionSingleChangeEvent(e){return this.multipleSelectionController?this.multipleSelectionController.isSelectionSingleChangeEvent(e):!1}isSelectionRangeChangeEvent(e){return this.multipleSelectionController?this.multipleSelectionController.isSelectionRangeChangeEvent(e):!1}isSelectionChangeEvent(e){return this.isSelectionSingleChangeEvent(e)||this.isSelectionRangeChangeEvent(e)}onMouseDown(e){h1(e.browserEvent.target)||ln()!==e.browserEvent.target&&this.list.domFocus()}onContextMenu(e){if(wl(e.browserEvent.target)||h1(e.browserEvent.target))return;const t=typeof e.index>"u"?[]:[e.index];this.list.setFocus(t,e.browserEvent)}onViewPointer(e){if(!this.mouseSupport||wl(e.browserEvent.target)||h1(e.browserEvent.target)||e.browserEvent.isHandledByList)return;e.browserEvent.isHandledByList=!0;const t=e.index;if(typeof t>"u"){this.list.setFocus([],e.browserEvent),this.list.setSelection([],e.browserEvent),this.list.setAnchor(void 0);return}if(this.isSelectionChangeEvent(e))return this.changeSelection(e);this.list.setFocus([t],e.browserEvent),this.list.setAnchor(t),SLe(e.browserEvent)||this.list.setSelection([t],e.browserEvent),this._onPointer.fire(e)}onDoubleClick(e){if(wl(e.browserEvent.target)||h1(e.browserEvent.target)||this.isSelectionChangeEvent(e)||e.browserEvent.isHandledByList)return;e.browserEvent.isHandledByList=!0;const t=this.list.getFocus();this.list.setSelection(t,e.browserEvent)}changeSelection(e){const t=e.index;let i=this.list.getAnchor();if(this.isSelectionRangeChangeEvent(e)){typeof i>"u"&&(i=this.list.getFocus()[0]??t,this.list.setAnchor(i));const n=Math.min(i,t),o=Math.max(i,t),r=Us(n,o+1),a=this.list.getSelection(),l=ELe(o4(a,[i]),i);if(l.length===0)return;const c=o4(r,ILe(a,l));this.list.setSelection(c,e.browserEvent),this.list.setFocus([t],e.browserEvent)}else if(this.isSelectionSingleChangeEvent(e)){const n=this.list.getSelection(),o=n.filter(r=>r!==t);this.list.setFocus([t]),this.list.setAnchor(t),n.length===o.length?this.list.setSelection([...o,t],e.browserEvent):this.list.setSelection(o,e.browserEvent)}}dispose(){this.disposables.dispose()}}class xLe{constructor(e,t){this.styleElement=e,this.selectorSuffix=t}style(e){const t=this.selectorSuffix&&`.${this.selectorSuffix}`,i=[];e.listBackground&&i.push(`.monaco-list${t} .monaco-list-rows { background: ${e.listBackground}; }`),e.listFocusBackground&&(i.push(`.monaco-list${t}:focus .monaco-list-row.focused { background-color: ${e.listFocusBackground}; }`),i.push(`.monaco-list${t}:focus .monaco-list-row.focused:hover { background-color: ${e.listFocusBackground}; }`)),e.listFocusForeground&&i.push(`.monaco-list${t}:focus .monaco-list-row.focused { color: ${e.listFocusForeground}; }`),e.listActiveSelectionBackground&&(i.push(`.monaco-list${t}:focus .monaco-list-row.selected { background-color: ${e.listActiveSelectionBackground}; }`),i.push(`.monaco-list${t}:focus .monaco-list-row.selected:hover { background-color: ${e.listActiveSelectionBackground}; }`)),e.listActiveSelectionForeground&&i.push(`.monaco-list${t}:focus .monaco-list-row.selected { color: ${e.listActiveSelectionForeground}; }`),e.listActiveSelectionIconForeground&&i.push(`.monaco-list${t}:focus .monaco-list-row.selected .codicon { color: ${e.listActiveSelectionIconForeground}; }`),e.listFocusAndSelectionBackground&&i.push(`
				.monaco-drag-image${t},
				.monaco-list${t}:focus .monaco-list-row.selected.focused { background-color: ${e.listFocusAndSelectionBackground}; }
			`),e.listFocusAndSelectionForeground&&i.push(`
				.monaco-drag-image${t},
				.monaco-list${t}:focus .monaco-list-row.selected.focused { color: ${e.listFocusAndSelectionForeground}; }
			`),e.listInactiveFocusForeground&&(i.push(`.monaco-list${t} .monaco-list-row.focused { color:  ${e.listInactiveFocusForeground}; }`),i.push(`.monaco-list${t} .monaco-list-row.focused:hover { color:  ${e.listInactiveFocusForeground}; }`)),e.listInactiveSelectionIconForeground&&i.push(`.monaco-list${t} .monaco-list-row.focused .codicon { color:  ${e.listInactiveSelectionIconForeground}; }`),e.listInactiveFocusBackground&&(i.push(`.monaco-list${t} .monaco-list-row.focused { background-color:  ${e.listInactiveFocusBackground}; }`),i.push(`.monaco-list${t} .monaco-list-row.focused:hover { background-color:  ${e.listInactiveFocusBackground}; }`)),e.listInactiveSelectionBackground&&(i.push(`.monaco-list${t} .monaco-list-row.selected { background-color:  ${e.listInactiveSelectionBackground}; }`),i.push(`.monaco-list${t} .monaco-list-row.selected:hover { background-color:  ${e.listInactiveSelectionBackground}; }`)),e.listInactiveSelectionForeground&&i.push(`.monaco-list${t} .monaco-list-row.selected { color: ${e.listInactiveSelectionForeground}; }`),e.listHoverBackground&&i.push(`.monaco-list${t}:not(.drop-target):not(.dragging) .monaco-list-row:hover:not(.selected):not(.focused) { background-color: ${e.listHoverBackground}; }`),e.listHoverForeground&&i.push(`.monaco-list${t}:not(.drop-target):not(.dragging) .monaco-list-row:hover:not(.selected):not(.focused) { color:  ${e.listHoverForeground}; }`);const n=sh(e.listFocusAndSelectionOutline,sh(e.listSelectionOutline,e.listFocusOutline??""));n&&i.push(`.monaco-list${t}:focus .monaco-list-row.focused.selected { outline: 1px solid ${n}; outline-offset: -1px;}`),e.listFocusOutline&&i.push(`
				.monaco-drag-image${t},
				.monaco-list${t}:focus .monaco-list-row.focused,
				.context-menu-visible .monaco-list${t}.last-focused .monaco-list-row.focused { outline: 1px solid ${e.listFocusOutline}; outline-offset: -1px; }
			`);const o=sh(e.listSelectionOutline,e.listInactiveFocusOutline??"");o&&i.push(`.monaco-list${t} .monaco-list-row.focused.selected { outline: 1px dotted ${o}; outline-offset: -1px; }`),e.listSelectionOutline&&i.push(`.monaco-list${t} .monaco-list-row.selected { outline: 1px dotted ${e.listSelectionOutline}; outline-offset: -1px; }`),e.listInactiveFocusOutline&&i.push(`.monaco-list${t} .monaco-list-row.focused { outline: 1px dotted ${e.listInactiveFocusOutline}; outline-offset: -1px; }`),e.listHoverOutline&&i.push(`.monaco-list${t} .monaco-list-row:hover { outline: 1px dashed ${e.listHoverOutline}; outline-offset: -1px; }`),e.listDropOverBackground&&i.push(`
				.monaco-list${t}.drop-target,
				.monaco-list${t} .monaco-list-rows.drop-target,
				.monaco-list${t} .monaco-list-row.drop-target { background-color: ${e.listDropOverBackground} !important; color: inherit !important; }
			`),e.listDropBetweenBackground&&(i.push(`
			.monaco-list${t} .monaco-list-rows.drop-target-before .monaco-list-row:first-child::before,
			.monaco-list${t} .monaco-list-row.drop-target-before::before {
				content: ""; position: absolute; top: 0px; left: 0px; width: 100%; height: 1px;
				background-color: ${e.listDropBetweenBackground};
			}`),i.push(`
			.monaco-list${t} .monaco-list-rows.drop-target-after .monaco-list-row:last-child::after,
			.monaco-list${t} .monaco-list-row.drop-target-after::after {
				content: ""; position: absolute; bottom: 0px; left: 0px; width: 100%; height: 1px;
				background-color: ${e.listDropBetweenBackground};
			}`)),e.tableColumnsBorder&&i.push(`
				.monaco-table > .monaco-split-view2,
				.monaco-table > .monaco-split-view2 .monaco-sash.vertical::before,
				.monaco-enable-motion .monaco-table:hover > .monaco-split-view2,
				.monaco-enable-motion .monaco-table:hover > .monaco-split-view2 .monaco-sash.vertical::before {
					border-color: ${e.tableColumnsBorder};
				}

				.monaco-enable-motion .monaco-table > .monaco-split-view2,
				.monaco-enable-motion .monaco-table > .monaco-split-view2 .monaco-sash.vertical::before {
					border-color: transparent;
				}
			`),e.tableOddRowsBackgroundColor&&i.push(`
				.monaco-table .monaco-list-row[data-parity=odd]:not(.focused):not(.selected):not(:hover) .monaco-table-tr,
				.monaco-table .monaco-list:not(:focus) .monaco-list-row[data-parity=odd].focused:not(.selected):not(:hover) .monaco-table-tr,
				.monaco-table .monaco-list:not(.focused) .monaco-list-row[data-parity=odd].focused:not(.selected):not(:hover) .monaco-table-tr {
					background-color: ${e.tableOddRowsBackgroundColor};
				}
			`),this.styleElement.textContent=i.join(`
`)}}const LLe={listFocusBackground:"#7FB0D0",listActiveSelectionBackground:"#0E639C",listActiveSelectionForeground:"#FFFFFF",listActiveSelectionIconForeground:"#FFFFFF",listFocusAndSelectionOutline:"#90C2F9",listFocusAndSelectionBackground:"#094771",listFocusAndSelectionForeground:"#FFFFFF",listInactiveSelectionBackground:"#3F3F46",listInactiveSelectionIconForeground:"#FFFFFF",listHoverBackground:"#2A2D2E",listDropOverBackground:"#383B3D",listDropBetweenBackground:"#EEEEEE",treeIndentGuidesStroke:"#a9a9a9",treeInactiveIndentGuidesStroke:q.fromHex("#a9a9a9").transparent(.4).toString(),tableColumnsBorder:q.fromHex("#cccccc").transparent(.2).toString(),tableOddRowsBackgroundColor:q.fromHex("#cccccc").transparent(.04).toString(),listBackground:void 0,listFocusForeground:void 0,listInactiveSelectionForeground:void 0,listInactiveFocusForeground:void 0,listInactiveFocusBackground:void 0,listHoverForeground:void 0,listFocusOutline:void 0,listInactiveFocusOutline:void 0,listSelectionOutline:void 0,listHoverOutline:void 0,treeStickyScrollBackground:void 0,treeStickyScrollBorder:void 0,treeStickyScrollShadow:void 0},kLe={keyboardSupport:!0,mouseSupport:!0,multipleSelectionSupport:!0,dnd:{getDragURI(){return null},onDragStart(){},onDragOver(){return!1},drop(){},dispose(){}}};function ELe(s,e){const t=s.indexOf(e);if(t===-1)return[];const i=[];let n=t-1;for(;n>=0&&s[n]===e-(t-n);)i.push(s[n--]);for(i.reverse(),n=t;n<s.length&&s[n]===e+(n-t);)i.push(s[n++]);return i}function o4(s,e){const t=[];let i=0,n=0;for(;i<s.length||n<e.length;)if(i>=s.length)t.push(e[n++]);else if(n>=e.length)t.push(s[i++]);else if(s[i]===e[n]){t.push(s[i]),i++,n++;continue}else s[i]<e[n]?t.push(s[i++]):t.push(e[n++]);return t}function ILe(s,e){const t=[];let i=0,n=0;for(;i<s.length||n<e.length;)if(i>=s.length)t.push(e[n++]);else if(n>=e.length)t.push(s[i++]);else if(s[i]===e[n]){i++,n++;continue}else s[i]<e[n]?t.push(s[i++]):n++;return t}const Nq=(s,e)=>s-e;class DLe{constructor(e,t){this._templateId=e,this.renderers=t}get templateId(){return this._templateId}renderTemplate(e){return this.renderers.map(t=>t.renderTemplate(e))}renderElement(e,t,i,n){let o=0;for(const r of this.renderers)r.renderElement(e,t,i[o++],n)}disposeElement(e,t,i,n){let o=0;for(const r of this.renderers)r.disposeElement?.(e,t,i[o],n),o+=1}disposeTemplate(e){let t=0;for(const i of this.renderers)i.disposeTemplate(e[t++])}}class TLe{constructor(e){this.accessibilityProvider=e,this.templateId="a18n"}renderTemplate(e){return{container:e,disposables:new U}}renderElement(e,t,i){const n=this.accessibilityProvider.getAriaLabel(e),o=n&&typeof n!="string"?n:zt(n);i.disposables.add(De(a=>{this.setAriaLabel(a.readObservable(o),i.container)}));const r=this.accessibilityProvider.getAriaLevel&&this.accessibilityProvider.getAriaLevel(e);typeof r=="number"?i.container.setAttribute("aria-level",`${r}`):i.container.removeAttribute("aria-level")}setAriaLabel(e,t){e?t.setAttribute("aria-label",e):t.removeAttribute("aria-label")}disposeElement(e,t,i){i.disposables.clear()}disposeTemplate(e){e.disposables.dispose()}}class NLe{constructor(e,t){this.list=e,this.dnd=t}getDragElements(e){const t=this.list.getSelectedElements();return t.indexOf(e)>-1?t:[e]}getDragURI(e){return this.dnd.getDragURI(e)}getDragLabel(e,t){if(this.dnd.getDragLabel)return this.dnd.getDragLabel(e,t)}onDragStart(e,t){this.dnd.onDragStart?.(e,t)}onDragOver(e,t,i,n,o){return this.dnd.onDragOver(e,t,i,n,o)}onDragLeave(e,t,i,n){this.dnd.onDragLeave?.(e,t,i,n)}onDragEnd(e){this.dnd.onDragEnd?.(e)}drop(e,t,i,n,o){this.dnd.drop(e,t,i,n,o)}dispose(){this.dnd.dispose()}}class Tr{get onDidChangeFocus(){return oe.map(this.eventBufferer.wrapEvent(this.focus.onChange),e=>this.toListEvent(e),this.disposables)}get onDidChangeSelection(){return oe.map(this.eventBufferer.wrapEvent(this.selection.onChange),e=>this.toListEvent(e),this.disposables)}get domId(){return this.view.domId}get onDidScroll(){return this.view.onDidScroll}get onMouseClick(){return this.view.onMouseClick}get onMouseDblClick(){return this.view.onMouseDblClick}get onMouseMiddleClick(){return this.view.onMouseMiddleClick}get onPointer(){return this.mouseController.onPointer}get onMouseDown(){return this.view.onMouseDown}get onMouseOver(){return this.view.onMouseOver}get onMouseOut(){return this.view.onMouseOut}get onTouchStart(){return this.view.onTouchStart}get onTap(){return this.view.onTap}get onContextMenu(){let e=!1;const t=oe.chain(this.disposables.add(new Lt(this.view.domNode,"keydown")).event,o=>o.map(r=>new Pt(r)).filter(r=>e=r.keyCode===58||r.shiftKey&&r.keyCode===68).map(r=>gt.stop(r,!0)).filter(()=>!1)),i=oe.chain(this.disposables.add(new Lt(this.view.domNode,"keyup")).event,o=>o.forEach(()=>e=!1).map(r=>new Pt(r)).filter(r=>r.keyCode===58||r.shiftKey&&r.keyCode===68).map(r=>gt.stop(r,!0)).map(({browserEvent:r})=>{const a=this.getFocus(),l=a.length?a[0]:void 0,c=typeof l<"u"?this.view.element(l):void 0,d=typeof l<"u"?this.view.domElement(l):this.view.domNode;return{index:l,element:c,anchor:d,browserEvent:r}})),n=oe.chain(this.view.onContextMenu,o=>o.filter(r=>!e).map(({element:r,index:a,browserEvent:l})=>({element:r,index:a,anchor:new is(be(this.view.domNode),l),browserEvent:l})));return oe.any(t,i,n)}get onKeyDown(){return this.disposables.add(new Lt(this.view.domNode,"keydown")).event}get onDidFocus(){return oe.signal(this.disposables.add(new Lt(this.view.domNode,"focus",!0)).event)}get onDidBlur(){return oe.signal(this.disposables.add(new Lt(this.view.domNode,"blur",!0)).event)}constructor(e,t,i,n,o=kLe){this.user=e,this._options=o,this.focus=new vI("focused"),this.anchor=new vI("anchor"),this.eventBufferer=new $0,this._ariaLabel="",this.disposables=new U,this._onDidDispose=new M,this.onDidDispose=this._onDidDispose.event;const r=this._options.accessibilityProvider&&this._options.accessibilityProvider.getWidgetRole?this._options.accessibilityProvider?.getWidgetRole():"list";this.selection=new _Le(r!=="listbox");const a=[this.focus.renderer,this.selection.renderer];this.accessibilityProvider=o.accessibilityProvider,this.accessibilityProvider&&(a.push(new TLe(this.accessibilityProvider)),this.accessibilityProvider.onDidChangeActiveDescendant?.(this.onDidChangeActiveDescendant,this,this.disposables)),n=n.map(c=>new DLe(c.templateId,[...a,c]));const l={...o,dnd:o.dnd&&new NLe(this,o.dnd)};if(this.view=this.createListView(t,i,n,l),this.view.domNode.setAttribute("role",r),o.styleController)this.styleController=o.styleController(this.view.domId);else{const c=ca(this.view.domNode);this.styleController=new xLe(c,this.view.domId)}if(this.spliceable=new oLe([new PM(this.focus,this.view,o.identityProvider),new PM(this.selection,this.view,o.identityProvider),new PM(this.anchor,this.view,o.identityProvider),this.view]),this.disposables.add(this.focus),this.disposables.add(this.selection),this.disposables.add(this.anchor),this.disposables.add(this.view),this.disposables.add(this._onDidDispose),this.disposables.add(new yLe(this,this.view)),(typeof o.keyboardSupport!="boolean"||o.keyboardSupport)&&(this.keyboardController=new rie(this,this.view,o),this.disposables.add(this.keyboardController)),o.keyboardNavigationLabelProvider){const c=o.keyboardNavigationDelegate||vLe;this.typeNavigationController=new CLe(this,this.view,o.keyboardNavigationLabelProvider,o.keyboardNavigationEventFilter??(()=>!0),c),this.disposables.add(this.typeNavigationController)}if(this.mouseController=this.createMouseController(o),this.disposables.add(this.mouseController),this.onDidChangeFocus(this._onFocusChange,this,this.disposables),this.onDidChangeSelection(this._onSelectionChange,this,this.disposables),this.accessibilityProvider){const c=this.accessibilityProvider.getWidgetAriaLabel(),d=c&&typeof c!="string"?c:zt(c);this.disposables.add(De(h=>{this.ariaLabel=h.readObservable(d)}))}this._options.multipleSelectionSupport!==!1&&this.view.domNode.setAttribute("aria-multiselectable","true")}createListView(e,t,i,n){return new Wa(e,t,i,n)}createMouseController(e){return new cie(this)}updateOptions(e={}){this._options={...this._options,...e},this.typeNavigationController?.updateOptions(this._options),this._options.multipleSelectionController!==void 0&&(this._options.multipleSelectionSupport?this.view.domNode.setAttribute("aria-multiselectable","true"):this.view.domNode.removeAttribute("aria-multiselectable")),this.mouseController.updateOptions(e),this.keyboardController?.updateOptions(e),this.view.updateOptions(e)}get options(){return this._options}splice(e,t,i=[]){if(e<0||e>this.view.length)throw new Ug(this.user,`Invalid start index: ${e}`);if(t<0)throw new Ug(this.user,`Invalid delete count: ${t}`);t===0&&i.length===0||this.eventBufferer.bufferEvents(()=>this.spliceable.splice(e,t,i))}rerender(){this.view.rerender()}element(e){return this.view.element(e)}indexOf(e){return this.view.indexOf(e)}indexAt(e){return this.view.indexAt(e)}get length(){return this.view.length}get contentHeight(){return this.view.contentHeight}get onDidChangeContentHeight(){return this.view.onDidChangeContentHeight}get scrollTop(){return this.view.getScrollTop()}set scrollTop(e){this.view.setScrollTop(e)}get scrollHeight(){return this.view.scrollHeight}get renderHeight(){return this.view.renderHeight}get firstVisibleIndex(){return this.view.firstVisibleIndex}get ariaLabel(){return this._ariaLabel}set ariaLabel(e){this._ariaLabel=e,this.view.domNode.setAttribute("aria-label",e)}domFocus(){this.view.domNode.focus({preventScroll:!0})}layout(e,t){this.view.layout(e,t)}setSelection(e,t){for(const i of e)if(i<0||i>=this.length)throw new Ug(this.user,`Invalid index ${i}`);this.selection.set(e,t)}getSelection(){return this.selection.get()}getSelectedElements(){return this.getSelection().map(e=>this.view.element(e))}setAnchor(e){if(typeof e>"u"){this.anchor.set([]);return}if(e<0||e>=this.length)throw new Ug(this.user,`Invalid index ${e}`);this.anchor.set([e])}getAnchor(){return this.anchor.get().at(0)}getAnchorElement(){const e=this.getAnchor();return typeof e>"u"?void 0:this.element(e)}setFocus(e,t){for(const i of e)if(i<0||i>=this.length)throw new Ug(this.user,`Invalid index ${i}`);this.focus.set(e,t)}focusNext(e=1,t=!1,i,n){if(this.length===0)return;const o=this.focus.get(),r=this.findNextIndex(o.length>0?o[0]+e:0,t,n);r>-1&&this.setFocus([r],i)}focusPrevious(e=1,t=!1,i,n){if(this.length===0)return;const o=this.focus.get(),r=this.findPreviousIndex(o.length>0?o[0]-e:0,t,n);r>-1&&this.setFocus([r],i)}async focusNextPage(e,t){let i=this.view.indexAt(this.view.getScrollTop()+this.view.renderHeight);i=i===0?0:i-1;const n=this.getFocus()[0];if(n!==i&&(n===void 0||i>n)){const o=this.findPreviousIndex(i,!1,t);o>-1&&n!==o?this.setFocus([o],e):this.setFocus([i],e)}else{const o=this.view.getScrollTop();let r=o+this.view.renderHeight;i>n&&(r-=this.view.elementHeight(i)),this.view.setScrollTop(r),this.view.getScrollTop()!==o&&(this.setFocus([]),await Tc(0),await this.focusNextPage(e,t))}}async focusPreviousPage(e,t,i=()=>0){let n;const o=i(),r=this.view.getScrollTop()+o;r===0?n=this.view.indexAt(r):n=this.view.indexAfter(r-1);const a=this.getFocus()[0];if(a!==n&&(a===void 0||a>=n)){const l=this.findNextIndex(n,!1,t);l>-1&&a!==l?this.setFocus([l],e):this.setFocus([n],e)}else{const l=r;this.view.setScrollTop(r-this.view.renderHeight-o),this.view.getScrollTop()+i()!==l&&(this.setFocus([]),await Tc(0),await this.focusPreviousPage(e,t,i))}}focusLast(e,t){if(this.length===0)return;const i=this.findPreviousIndex(this.length-1,!1,t);i>-1&&this.setFocus([i],e)}focusFirst(e,t){this.focusNth(0,e,t)}focusNth(e,t,i){if(this.length===0)return;const n=this.findNextIndex(e,!1,i);n>-1&&this.setFocus([n],t)}findNextIndex(e,t=!1,i){for(let n=0;n<this.length;n++){if(e>=this.length&&!t)return-1;if(e=e%this.length,!i||i(this.element(e)))return e;e++}return-1}findPreviousIndex(e,t=!1,i){for(let n=0;n<this.length;n++){if(e<0&&!t)return-1;if(e=(this.length+e%this.length)%this.length,!i||i(this.element(e)))return e;e--}return-1}getFocus(){return this.focus.get()}getFocusedElements(){return this.getFocus().map(e=>this.view.element(e))}reveal(e,t,i=0){if(e<0||e>=this.length)throw new Ug(this.user,`Invalid index ${e}`);const n=this.view.getScrollTop(),o=this.view.elementTop(e),r=this.view.elementHeight(e);if(ch(t)){const a=r-this.view.renderHeight+i;this.view.setScrollTop(a*qs(t,0,1)+o-i)}else{const a=o+r,l=n+this.view.renderHeight;o<n+i&&a>=l||(o<n+i||a>=l&&r>=this.view.renderHeight?this.view.setScrollTop(o-i):a>=l&&this.view.setScrollTop(a-this.view.renderHeight))}}getRelativeTop(e,t=0){if(e<0||e>=this.length)throw new Ug(this.user,`Invalid index ${e}`);const i=this.view.getScrollTop(),n=this.view.elementTop(e),o=this.view.elementHeight(e);if(n<i+t||n+o>i+this.view.renderHeight)return null;const r=o-this.view.renderHeight+t;return Math.abs((i+t-n)/r)}getHTMLElement(){return this.view.domNode}getScrollableElement(){return this.view.scrollableElementDomNode}getElementID(e){return this.view.getElementDomId(e)}getElementTop(e){return this.view.elementTop(e)}style(e){this.styleController.style(e)}toListEvent({indexes:e,browserEvent:t}){return{indexes:e,elements:e.map(i=>this.view.element(i)),browserEvent:t}}_onFocusChange(){const e=this.focus.get();this.view.domNode.classList.toggle("element-focused",e.length>0),this.onDidChangeActiveDescendant()}onDidChangeActiveDescendant(){const e=this.focus.get();if(e.length>0){let t;this.accessibilityProvider?.getActiveDescendantId&&(t=this.accessibilityProvider.getActiveDescendantId(this.view.element(e[0]))),this.view.domNode.setAttribute("aria-activedescendant",t||this.view.getElementDomId(e[0]))}else this.view.domNode.removeAttribute("aria-activedescendant")}_onSelectionChange(){const e=this.selection.get();this.view.domNode.classList.toggle("selection-none",e.length===0),this.view.domNode.classList.toggle("selection-single",e.length===1),this.view.domNode.classList.toggle("selection-multiple",e.length>1)}dispose(){this._onDidDispose.fire(),this.disposables.dispose(),this._onDidDispose.dispose()}}Ng([Ai],Tr.prototype,"onDidChangeFocus",null);Ng([Ai],Tr.prototype,"onDidChangeSelection",null);Ng([Ai],Tr.prototype,"onContextMenu",null);Ng([Ai],Tr.prototype,"onKeyDown",null);Ng([Ai],Tr.prototype,"onDidFocus",null);Ng([Ai],Tr.prototype,"onDidBlur",null);const Yf=ie,die="selectOption.entry.template";class RLe{get templateId(){return die}renderTemplate(e){const t=Object.create(null);return t.root=e,t.text=X(e,Yf(".option-text")),t.detail=X(e,Yf(".option-detail")),t.decoratorRight=X(e,Yf(".option-decorator-right")),t}renderElement(e,t,i){const n=i,o=e.text,r=e.detail,a=e.decoratorRight,l=e.isDisabled;n.text.textContent=o,n.detail.textContent=r||"",n.decoratorRight.textContent=a||"",l?n.root.classList.add("option-disabled"):n.root.classList.remove("option-disabled")}disposeTemplate(e){}}const Jl=class Jl extends F{constructor(e,t,i,n,o){super(),this.options=[],this._currentSelection=0,this._hasDetails=!1,this._selectionDetailsDisposables=this._register(new U),this._skipLayout=!1,this._sticky=!1,this._isVisible=!1,this.styles=n,this.selectBoxOptions=o||Object.create(null),typeof this.selectBoxOptions.minBottomMargin!="number"?this.selectBoxOptions.minBottomMargin=Jl.DEFAULT_DROPDOWN_MINIMUM_BOTTOM_MARGIN:this.selectBoxOptions.minBottomMargin<0&&(this.selectBoxOptions.minBottomMargin=0),this.selectElement=document.createElement("select"),this.selectElement.className="monaco-select-box monaco-select-box-dropdown-padding",typeof this.selectBoxOptions.ariaLabel=="string"&&this.selectElement.setAttribute("aria-label",this.selectBoxOptions.ariaLabel),typeof this.selectBoxOptions.ariaDescription=="string"&&this.selectElement.setAttribute("aria-description",this.selectBoxOptions.ariaDescription),this._onDidSelect=new M,this._register(this._onDidSelect),this.registerListeners(),this.constructSelectDropDown(i),this.selected=t||0,e&&this.setOptions(e,t),this.initStyleSheet()}setTitle(e){!this._hover&&e?this._hover=this._register($a().setupManagedHover(Uc("mouse"),this.selectElement,e)):this._hover&&this._hover.update(e)}getHeight(){return 22}getTemplateId(){return die}constructSelectDropDown(e){this.contextViewProvider=e,this.selectDropDownContainer=ie(".monaco-select-box-dropdown-container"),this.selectDropDownContainer.classList.add("monaco-select-box-dropdown-padding"),this.selectionDetailsPane=X(this.selectDropDownContainer,Yf(".select-box-details-pane"));const t=X(this.selectDropDownContainer,Yf(".select-box-dropdown-container-width-control")),i=X(t,Yf(".width-control-div"));this.widthControlElement=document.createElement("span"),this.widthControlElement.className="option-text-width-control",X(i,this.widthControlElement),this._dropDownPosition=0,this.styleElement=ca(this.selectDropDownContainer),this.selectDropDownContainer.setAttribute("draggable","true"),this._register(W(this.selectDropDownContainer,se.DRAG_START,n=>{gt.stop(n,!0)}))}registerListeners(){this._register(Fi(this.selectElement,"change",t=>{this.selected=t.target.selectedIndex,this._onDidSelect.fire({index:t.target.selectedIndex,selected:t.target.value}),this.options[this.selected]&&this.options[this.selected].text&&this.setTitle(this.options[this.selected].text)})),this._register(W(this.selectElement,se.CLICK,t=>{gt.stop(t),this._isVisible?this.hideSelectDropDown(!0):this.showSelectDropDown()})),this._register(W(this.selectElement,se.MOUSE_DOWN,t=>{gt.stop(t)}));let e;this._register(W(this.selectElement,"touchstart",t=>{e=this._isVisible})),this._register(W(this.selectElement,"touchend",t=>{gt.stop(t),e?this.hideSelectDropDown(!0):this.showSelectDropDown()})),this._register(W(this.selectElement,se.KEY_DOWN,t=>{const i=new Pt(t);let n=!1;Ye?(i.keyCode===18||i.keyCode===16||i.keyCode===10||i.keyCode===3)&&(n=!0):(i.keyCode===18&&i.altKey||i.keyCode===16&&i.altKey||i.keyCode===10||i.keyCode===3)&&(n=!0),n&&(this.showSelectDropDown(),gt.stop(t,!0))}))}get onDidSelect(){return this._onDidSelect.event}setOptions(e,t){ni(this.options,e)||(this.options=e,this.selectElement.options.length=0,this._hasDetails=!1,this._cachedMaxDetailsHeight=void 0,this.options.forEach((i,n)=>{this.selectElement.add(this.createOption(i.text,n,i.isDisabled)),typeof i.description=="string"&&(this._hasDetails=!0)})),t!==void 0&&(this.select(t),this._currentSelection=this.selected)}setOptionsList(){this.selectList?.splice(0,this.selectList.length,this.options)}select(e){e>=0&&e<this.options.length?this.selected=e:e>this.options.length-1?this.select(this.options.length-1):this.selected<0&&(this.selected=0),this.selectElement.selectedIndex=this.selected,this.options[this.selected]&&this.options[this.selected].text&&this.setTitle(this.options[this.selected].text)}focus(){this.selectElement&&(this.selectElement.tabIndex=0,this.selectElement.focus())}blur(){this.selectElement&&(this.selectElement.tabIndex=-1,this.selectElement.blur())}setFocusable(e){this.selectElement.tabIndex=e?0:-1}render(e){this.container=e,e.classList.add("select-container"),e.appendChild(this.selectElement),this.styleSelectElement()}initStyleSheet(){const e=[];this.styles.listFocusBackground&&e.push(`.monaco-select-box-dropdown-container > .select-box-dropdown-list-container .monaco-list .monaco-list-row.focused { background-color: ${this.styles.listFocusBackground} !important; }`),this.styles.listFocusForeground&&e.push(`.monaco-select-box-dropdown-container > .select-box-dropdown-list-container .monaco-list .monaco-list-row.focused { color: ${this.styles.listFocusForeground} !important; }`),this.styles.decoratorRightForeground&&e.push(`.monaco-select-box-dropdown-container > .select-box-dropdown-list-container .monaco-list .monaco-list-row:not(.focused) .option-decorator-right { color: ${this.styles.decoratorRightForeground}; }`),this.styles.selectBackground&&this.styles.selectBorder&&this.styles.selectBorder!==this.styles.selectBackground?(e.push(`.monaco-select-box-dropdown-container { border: 1px solid ${this.styles.selectBorder} } `),e.push(`.monaco-select-box-dropdown-container > .select-box-details-pane.border-top { border-top: 1px solid ${this.styles.selectBorder} } `),e.push(`.monaco-select-box-dropdown-container > .select-box-details-pane.border-bottom { border-bottom: 1px solid ${this.styles.selectBorder} } `)):this.styles.selectListBorder&&(e.push(`.monaco-select-box-dropdown-container > .select-box-details-pane.border-top { border-top: 1px solid ${this.styles.selectListBorder} } `),e.push(`.monaco-select-box-dropdown-container > .select-box-details-pane.border-bottom { border-bottom: 1px solid ${this.styles.selectListBorder} } `)),this.styles.listHoverForeground&&e.push(`.monaco-select-box-dropdown-container > .select-box-dropdown-list-container .monaco-list .monaco-list-row:not(.option-disabled):not(.focused):hover { color: ${this.styles.listHoverForeground} !important; }`),this.styles.listHoverBackground&&e.push(`.monaco-select-box-dropdown-container > .select-box-dropdown-list-container .monaco-list .monaco-list-row:not(.option-disabled):not(.focused):hover { background-color: ${this.styles.listHoverBackground} !important; }`),this.styles.listFocusOutline&&e.push(`.monaco-select-box-dropdown-container > .select-box-dropdown-list-container .monaco-list .monaco-list-row.focused { outline: 1.6px dotted ${this.styles.listFocusOutline} !important; outline-offset: -1.6px !important; }`),this.styles.listHoverOutline&&e.push(`.monaco-select-box-dropdown-container > .select-box-dropdown-list-container .monaco-list .monaco-list-row:not(.option-disabled):not(.focused):hover { outline: 1.6px dashed ${this.styles.listHoverOutline} !important; outline-offset: -1.6px !important; }`),e.push(".monaco-select-box-dropdown-container > .select-box-dropdown-list-container .monaco-list .monaco-list-row.option-disabled.focused { background-color: transparent !important; color: inherit !important; outline: none !important; }"),e.push(".monaco-select-box-dropdown-container > .select-box-dropdown-list-container .monaco-list .monaco-list-row.option-disabled:hover { background-color: transparent !important; color: inherit !important; outline: none !important; }"),this.styleElement.textContent=e.join(`
`)}styleSelectElement(){const e=this.styles.selectBackground??"",t=this.styles.selectForeground??"",i=this.styles.selectBorder??"";this.selectElement.style.backgroundColor=e,this.selectElement.style.color=t,this.selectElement.style.borderColor=i}styleList(){const e=this.styles.selectBackground??"",t=sh(this.styles.selectListBackground,e);this.selectDropDownListContainer.style.backgroundColor=t,this.selectionDetailsPane.style.backgroundColor=t;const i=this.styles.focusBorder??"";this.selectDropDownContainer.style.outlineColor=i,this.selectDropDownContainer.style.outlineOffset="-1px",this.selectList.style(this.styles)}createOption(e,t,i){const n=document.createElement("option");return n.value=e,n.text=e,n.disabled=!!i,n}showSelectDropDown(){this.selectionDetailsPane.textContent="",!(!this.contextViewProvider||this._isVisible)&&(this.createSelectList(this.selectDropDownContainer),this.setOptionsList(),this.contextViewProvider.showContextView({getAnchor:()=>this.selectElement,render:e=>this.renderSelectDropDown(e,!0),layout:()=>{this.layoutSelectDropDown()},onHide:()=>{this.selectDropDownContainer.classList.remove("visible"),this.selectElement.classList.remove("synthetic-focus")},anchorPosition:this._dropDownPosition},this.selectBoxOptions.optionsAsChildren?this.container:void 0),this._isVisible=!0,this.hideSelectDropDown(!1),this.contextViewProvider.showContextView({getAnchor:()=>this.selectElement,render:e=>this.renderSelectDropDown(e),layout:()=>this.layoutSelectDropDown(),onHide:()=>{this.selectDropDownContainer.classList.remove("visible"),this.selectElement.classList.remove("synthetic-focus")},anchorPosition:this._dropDownPosition},this.selectBoxOptions.optionsAsChildren?this.container:void 0),this._currentSelection=this.selected,this._isVisible=!0,this.selectElement.setAttribute("aria-expanded","true"))}hideSelectDropDown(e){!this.contextViewProvider||!this._isVisible||(this._isVisible=!1,this.selectElement.setAttribute("aria-expanded","false"),e&&this.selectElement.focus(),this.contextViewProvider.hideContextView())}renderSelectDropDown(e,t){return e.appendChild(this.selectDropDownContainer),this.layoutSelectDropDown(t),{dispose:()=>{this.selectDropDownContainer.remove()}}}measureMaxDetailsHeight(){let e=0;return this.options.forEach((t,i)=>{this.updateDetail(i),this.selectionDetailsPane.offsetHeight>e&&(e=this.selectionDetailsPane.offsetHeight)}),e}layoutSelectDropDown(e){if(this._skipLayout)return!1;if(this.selectList){this.selectDropDownContainer.classList.add("visible");const t=be(this.selectElement),i=Li(this.selectElement),n=be(this.selectElement).getComputedStyle(this.selectElement),o=parseFloat(n.getPropertyValue("--dropdown-padding-top"))+parseFloat(n.getPropertyValue("--dropdown-padding-bottom")),r=t.innerHeight-i.top-i.height-(this.selectBoxOptions.minBottomMargin||0),a=i.top-Jl.DEFAULT_DROPDOWN_MINIMUM_TOP_MARGIN,l=this.selectElement.offsetWidth,c=this.setWidthControlElement(this.widthControlElement),d=Math.max(c,Math.round(l)).toString()+"px";this.selectDropDownContainer.style.width=d,this.selectList.getHTMLElement().style.height="",this.selectList.layout();let h=this.selectList.contentHeight;this._hasDetails&&this._cachedMaxDetailsHeight===void 0&&(this._cachedMaxDetailsHeight=this.measureMaxDetailsHeight());const u=this._hasDetails?this._cachedMaxDetailsHeight:0,g=h+o+u,f=Math.floor((r-o-u)/this.getHeight()),p=Math.floor((a-o-u)/this.getHeight());if(e)return i.top+i.height>t.innerHeight-22||i.top<Jl.DEFAULT_DROPDOWN_MINIMUM_TOP_MARGIN||f<1&&p<1?!1:(f<Jl.DEFAULT_MINIMUM_VISIBLE_OPTIONS&&p>f&&this.options.length>f?(this._dropDownPosition=1,this.selectDropDownListContainer.remove(),this.selectionDetailsPane.remove(),this.selectDropDownContainer.appendChild(this.selectionDetailsPane),this.selectDropDownContainer.appendChild(this.selectDropDownListContainer),this.selectionDetailsPane.classList.remove("border-top"),this.selectionDetailsPane.classList.add("border-bottom")):(this._dropDownPosition=0,this.selectDropDownListContainer.remove(),this.selectionDetailsPane.remove(),this.selectDropDownContainer.appendChild(this.selectDropDownListContainer),this.selectDropDownContainer.appendChild(this.selectionDetailsPane),this.selectionDetailsPane.classList.remove("border-bottom"),this.selectionDetailsPane.classList.add("border-top")),!0);if(i.top+i.height>t.innerHeight-22||i.top<Jl.DEFAULT_DROPDOWN_MINIMUM_TOP_MARGIN||this._dropDownPosition===0&&f<1||this._dropDownPosition===1&&p<1)return this.hideSelectDropDown(!0),!1;if(this._dropDownPosition===0){if(this._isVisible&&f+p<1)return this.hideSelectDropDown(!0),!1;g>r&&(h=f*this.getHeight())}else g>a&&(h=p*this.getHeight());return this.selectList.layout(h),this.selectList.domFocus(),this.selectList.length>0&&(this.selectList.setFocus([this.selected||0]),this.selectList.reveal(this.selectList.getFocus()[0]||0)),this._hasDetails?(this.selectList.getHTMLElement().style.height=h+o+"px",this.selectDropDownContainer.style.height=""):this.selectDropDownContainer.style.height=h+o+"px",this.updateDetail(this.selected),this.selectDropDownContainer.style.width=d,this.selectDropDownListContainer.setAttribute("tabindex","0"),this.selectElement.classList.add("synthetic-focus"),this.selectDropDownContainer.classList.add("synthetic-focus"),!0}else return!1}setWidthControlElement(e){let t=0;if(e){let i=0,n=0;this.options.forEach((o,r)=>{const a=o.detail?o.detail.length:0,l=o.decoratorRight?o.decoratorRight.length:0,c=o.text.length+a+l;c>n&&(i=r,n=c)}),e.textContent=this.options[i].text+(this.options[i].decoratorRight?this.options[i].decoratorRight+" ":""),t=Ro(e)}return t}createSelectList(e){if(this.selectList)return;this.selectDropDownListContainer=X(e,Yf(".select-box-dropdown-list-container")),this.listRenderer=new RLe,this.selectList=this._register(new Tr("SelectBoxCustom",this.selectDropDownListContainer,this,[this.listRenderer],{useShadows:!1,verticalScrollMode:3,keyboardSupport:!1,mouseSupport:!1,accessibilityProvider:{getAriaLabel:n=>{let o=n.text;return n.detail&&(o+=`. ${n.detail}`),n.decoratorRight&&(o+=`. ${n.decoratorRight}`),n.description&&(o+=`. ${n.description}`),o},getWidgetAriaLabel:()=>m(16,"Select Box"),getRole:()=>Ye?"":"option",getWidgetRole:()=>"listbox"}})),this.selectBoxOptions.ariaLabel&&(this.selectList.ariaLabel=this.selectBoxOptions.ariaLabel);const t=this._register(new Lt(this.selectDropDownListContainer,"keydown")),i=oe.chain(t.event,n=>n.filter(()=>this.selectList.length>0).map(o=>new Pt(o)));this._register(oe.chain(i,n=>n.filter(o=>o.keyCode===3))(this.onEnter,this)),this._register(oe.chain(i,n=>n.filter(o=>o.keyCode===2))(this.onEnter,this)),this._register(oe.chain(i,n=>n.filter(o=>o.keyCode===9))(this.onEscape,this)),this._register(oe.chain(i,n=>n.filter(o=>o.keyCode===16))(this.onUpArrow,this)),this._register(oe.chain(i,n=>n.filter(o=>o.keyCode===18))(this.onDownArrow,this)),this._register(oe.chain(i,n=>n.filter(o=>o.keyCode===12))(this.onPageDown,this)),this._register(oe.chain(i,n=>n.filter(o=>o.keyCode===11))(this.onPageUp,this)),this._register(oe.chain(i,n=>n.filter(o=>o.keyCode===14))(this.onHome,this)),this._register(oe.chain(i,n=>n.filter(o=>o.keyCode===13))(this.onEnd,this)),this._register(oe.chain(i,n=>n.filter(o=>o.keyCode>=21&&o.keyCode<=56||o.keyCode>=85&&o.keyCode<=113))(this.onCharacter,this)),this._register(W(this.selectList.getHTMLElement(),se.POINTER_UP,n=>this.onPointerUp(n))),this._register(this.selectList.onMouseOver(n=>typeof n.index<"u"&&this.selectList.setFocus([n.index]))),this._register(this.selectList.onDidChangeFocus(n=>this.onListFocus(n))),this._register(W(this.selectDropDownContainer,se.FOCUS_OUT,n=>{!this._isVisible||wn(n.relatedTarget,this.selectDropDownContainer)||this.onListBlur()})),this.selectList.getHTMLElement().setAttribute("aria-label",this.selectBoxOptions.ariaLabel||""),this.selectList.getHTMLElement().setAttribute("aria-expanded","true"),this.styleList()}onPointerUp(e){if(!this.selectList.length)return;gt.stop(e);const t=e.target;if(!t||t.classList.contains("slider"))return;const i=t.closest(".monaco-list-row");if(!i)return;const n=Number(i.getAttribute("data-index")),o=i.classList.contains("option-disabled");n>=0&&n<this.options.length&&!o&&(this.selected=n,this.select(this.selected),this.selectList.setFocus([this.selected]),this.selectList.reveal(this.selectList.getFocus()[0]),this.selected!==this._currentSelection&&(this._currentSelection=this.selected,this._onDidSelect.fire({index:this.selectElement.selectedIndex,selected:this.options[this.selected].text}),this.options[this.selected]&&this.options[this.selected].text&&this.setTitle(this.options[this.selected].text)),this.hideSelectDropDown(!0))}onListBlur(){this._sticky||(this.selected!==this._currentSelection&&this.select(this._currentSelection),this.hideSelectDropDown(!1))}renderDescriptionMarkdown(e,t){const i=o=>{for(let r=0;r<o.childNodes.length;r++){const a=o.childNodes.item(r);(a.tagName&&a.tagName.toLowerCase())==="img"?a.remove():i(a)}},n=lx({value:e,supportThemeIcons:!0},{actionHandler:t});return n.element.classList.add("select-box-description-markdown"),i(n.element),n}onListFocus(e){!this._isVisible||!this._hasDetails||this.updateDetail(e.indexes[0])}updateDetail(e){this._selectionDetailsDisposables.clear(),this.selectionDetailsPane.textContent="";const t=this.options[e],i=t?.description??"",n=t?.descriptionIsMarkdown??!1;if(i){if(n){const o=t.descriptionMarkdownActionHandler,r=this._selectionDetailsDisposables.add(this.renderDescriptionMarkdown(i,o));this.selectionDetailsPane.appendChild(r.element)}else this.selectionDetailsPane.textContent=i;this.selectionDetailsPane.style.display="block"}else this.selectionDetailsPane.style.display="none";this._skipLayout=!0,this.contextViewProvider.layout(),this._skipLayout=!1}onEscape(e){gt.stop(e),this.select(this._currentSelection),this.hideSelectDropDown(!0)}onEnter(e){gt.stop(e),this.selected!==this._currentSelection&&(this._currentSelection=this.selected,this._onDidSelect.fire({index:this.selectElement.selectedIndex,selected:this.options[this.selected].text}),this.options[this.selected]&&this.options[this.selected].text&&this.setTitle(this.options[this.selected].text)),this.hideSelectDropDown(!0)}onDownArrow(e){if(this.selected<this.options.length-1){gt.stop(e,!0);const t=this.options[this.selected+1].isDisabled;if(t&&this.options.length>this.selected+2)this.selected+=2;else{if(t)return;this.selected++}this.select(this.selected),this.selectList.setFocus([this.selected]),this.selectList.reveal(this.selectList.getFocus()[0])}}onUpArrow(e){this.selected>0&&(gt.stop(e,!0),this.options[this.selected-1].isDisabled&&this.selected>1?this.selected-=2:this.selected--,this.select(this.selected),this.selectList.setFocus([this.selected]),this.selectList.reveal(this.selectList.getFocus()[0]))}onPageUp(e){gt.stop(e),this.selectList.focusPreviousPage(),setTimeout(()=>{this.selected=this.selectList.getFocus()[0],this.options[this.selected].isDisabled&&this.selected<this.options.length-1&&(this.selected++,this.selectList.setFocus([this.selected])),this.selectList.reveal(this.selected),this.select(this.selected)},1)}onPageDown(e){gt.stop(e),this.selectList.focusNextPage(),setTimeout(()=>{this.selected=this.selectList.getFocus()[0],this.options[this.selected].isDisabled&&this.selected>0&&(this.selected--,this.selectList.setFocus([this.selected])),this.selectList.reveal(this.selected),this.select(this.selected)},1)}onHome(e){gt.stop(e),!(this.options.length<2)&&(this.selected=0,this.options[this.selected].isDisabled&&this.selected>1&&this.selected++,this.selectList.setFocus([this.selected]),this.selectList.reveal(this.selected),this.select(this.selected))}onEnd(e){gt.stop(e),!(this.options.length<2)&&(this.selected=this.options.length-1,this.options[this.selected].isDisabled&&this.selected>1&&this.selected--,this.selectList.setFocus([this.selected]),this.selectList.reveal(this.selected),this.select(this.selected))}onCharacter(e){const t=Ad.toString(e.keyCode);let i=-1;for(let n=0;n<this.options.length-1;n++)if(i=(n+this.selected+1)%this.options.length,this.options[i].text.charAt(0).toUpperCase()===t&&!this.options[i].isDisabled){this.select(i),this.selectList.setFocus([i]),this.selectList.reveal(this.selectList.getFocus()[0]),gt.stop(e);break}}dispose(){this.hideSelectDropDown(!1),super.dispose()}};Jl.DEFAULT_DROPDOWN_MINIMUM_BOTTOM_MARGIN=32,Jl.DEFAULT_DROPDOWN_MINIMUM_TOP_MARGIN=2,Jl.DEFAULT_MINIMUM_VISIBLE_OPTIONS=3;let r4=Jl;class ALe extends F{constructor(e,t,i,n){super(),this.selected=0,this.selectBoxOptions=n||Object.create(null),this.options=[],this.selectElement=document.createElement("select"),this.selectElement.className="monaco-select-box",typeof this.selectBoxOptions.ariaLabel=="string"&&this.selectElement.setAttribute("aria-label",this.selectBoxOptions.ariaLabel),typeof this.selectBoxOptions.ariaDescription=="string"&&this.selectElement.setAttribute("aria-description",this.selectBoxOptions.ariaDescription),this._onDidSelect=this._register(new M),this.styles=i,this.registerListeners(),this.setOptions(e,t)}registerListeners(){this._register(_s.addTarget(this.selectElement)),[qt.Tap].forEach(e=>{this._register(W(this.selectElement,e,t=>{this.selectElement.focus()}))}),this._register(Fi(this.selectElement,"click",e=>{gt.stop(e,!0)})),this._register(Fi(this.selectElement,"change",e=>{this.selectElement.title=e.target.value,this._onDidSelect.fire({index:e.target.selectedIndex,selected:e.target.value})})),this._register(Fi(this.selectElement,"keydown",e=>{let t=!1;Ye?(e.keyCode===18||e.keyCode===16||e.keyCode===10)&&(t=!0):(e.keyCode===18&&e.altKey||e.keyCode===10||e.keyCode===3)&&(t=!0),t&&e.stopPropagation()}))}get onDidSelect(){return this._onDidSelect.event}setOptions(e,t){(!this.options||!ni(this.options,e))&&(this.options=e,this.selectElement.options.length=0,this.options.forEach((i,n)=>{this.selectElement.add(this.createOption(i.text,n,i.isDisabled))})),t!==void 0&&this.select(t)}select(e){this.options.length===0?this.selected=0:e>=0&&e<this.options.length?this.selected=e:e>this.options.length-1?this.select(this.options.length-1):this.selected<0&&(this.selected=0),this.selectElement.selectedIndex=this.selected,this.selected<this.options.length&&typeof this.options[this.selected].text=="string"?this.selectElement.title=this.options[this.selected].text:this.selectElement.title=""}focus(){this.selectElement&&(this.selectElement.tabIndex=0,this.selectElement.focus())}blur(){this.selectElement&&(this.selectElement.tabIndex=-1,this.selectElement.blur())}setFocusable(e){this.selectElement.tabIndex=e?0:-1}render(e){e.classList.add("select-container"),e.appendChild(this.selectElement),this.setOptions(this.options,this.selected),this.applyStyles()}applyStyles(){this.selectElement&&(this.selectElement.style.backgroundColor=this.styles.selectBackground??"",this.selectElement.style.color=this.styles.selectForeground??"",this.selectElement.style.borderColor=this.styles.selectBorder??"")}createOption(e,t,i){const n=document.createElement("option");return n.value=e,n.text=e,n.disabled=!!i,n}}class MLe extends Xo{constructor(e,t,i,n,o){super(),Ye&&!o?.useCustomDrawn?this.selectBoxDelegate=new ALe(e,t,n,o):this.selectBoxDelegate=new r4(e,t,i,n,o),this._register(this.selectBoxDelegate)}get onDidSelect(){return this.selectBoxDelegate.onDidSelect}setOptions(e,t){this.selectBoxDelegate.setOptions(e,t)}select(e){this.selectBoxDelegate.select(e)}focus(){this.selectBoxDelegate.focus()}blur(){this.selectBoxDelegate.blur()}setFocusable(e){this.selectBoxDelegate.setFocusable(e)}render(e){this.selectBoxDelegate.render(e)}}class ll extends F{get action(){return this._action}constructor(e,t,i={}){super(),this.options=i,this._context=e||this,this._action=t,t instanceof Cr&&this._register(t.onDidChange(n=>{this.element&&this.handleActionChangeEvent(n)}))}handleActionChangeEvent(e){e.enabled!==void 0&&this.updateEnabled(),e.checked!==void 0&&this.updateChecked(),e.class!==void 0&&this.updateClass(),e.label!==void 0&&(this.updateLabel(),this.updateTooltip()),e.tooltip!==void 0&&this.updateTooltip()}get actionRunner(){return this._actionRunner||(this._actionRunner=this._register(new Lm)),this._actionRunner}set actionRunner(e){this._actionRunner=e}isEnabled(){return this._action.enabled}setActionContext(e){this._context=e}render(e){const t=this.element=e;this._register(_s.addTarget(e));const i=this.options&&this.options.draggable;i&&(e.draggable=!0,wo&&this._register(W(e,se.DRAG_START,n=>n.dataTransfer?.setData(Mm.TEXT,this._action.label)))),this._register(W(t,qt.Tap,n=>this.onClick(n,!0))),this._register(W(t,se.MOUSE_DOWN,n=>{i||gt.stop(n,!0),this._action.enabled&&n.button===0&&t.classList.add("active")})),Ye&&this._register(W(t,se.CONTEXT_MENU,n=>{n.button===0&&n.ctrlKey===!0&&this.onClick(n)})),this._register(W(t,se.CLICK,n=>{gt.stop(n,!0),this.options&&this.options.isMenu||this.onClick(n)})),this._register(W(t,se.DBLCLICK,n=>{gt.stop(n,!0)})),[se.MOUSE_UP,se.MOUSE_OUT].forEach(n=>{this._register(W(t,n,o=>{gt.stop(o),t.classList.remove("active")}))})}onClick(e,t=!1){gt.stop(e,!0);const i=Ur(this._context)?this.options?.useEventAsContext?e:{preserveFocus:t}:this._context;this.actionRunner.run(this._action,i)}focus(){this.element&&(this.element.tabIndex=0,this.element.focus(),this.element.classList.add("focused"))}blur(){this.element&&(this.element.blur(),this.element.tabIndex=-1,this.element.classList.remove("focused"))}setFocusable(e){this.element&&(this.element.tabIndex=e?0:-1)}get trapsArrowNavigation(){return!1}updateEnabled(){}updateLabel(){}getClass(){return this.action.class}getTooltip(){return this.action.tooltip}getHoverContents(){return this.getTooltip()}updateTooltip(){if(!this.element)return;const e=this.getHoverContents()??"";if(this.updateAriaLabel(),!this.customHover&&e!==""){const t=this.options.hoverDelegate??Uc("element");this.customHover=this._store.add($a().setupManagedHover(t,this.element,e))}else this.customHover&&this.customHover.update(e)}updateAriaLabel(){if(this.element){const e=this.getTooltip()??"";this.element.setAttribute("aria-label",e)}}updateClass(){}updateChecked(){}dispose(){this.element&&(this.element.remove(),this.element=void 0),this._context=void 0,super.dispose()}}class vv extends ll{constructor(e,t,i){i={...i,icon:i.icon!==void 0?i.icon:!1,label:i.label!==void 0?i.label:!0},super(e,t,i),this.options=i,this.cssClass=""}render(e){super.render(e),dt(this.element);const t=document.createElement("a");if(t.classList.add("action-label"),t.setAttribute("role",this.getDefaultAriaRole()),this.label=t,this.element.appendChild(t),this.options.label&&this.options.keybinding&&!this.options.keybindingNotRenderedWithLabel){const i=document.createElement("span");i.classList.add("keybinding"),i.textContent=this.options.keybinding,this.element.appendChild(i)}this.updateClass(),this.updateLabel(),this.updateTooltip(),this.updateEnabled(),this.updateChecked()}getDefaultAriaRole(){return this._action.id===tn.ID?"presentation":this.options.isMenu?"menuitem":this.options.isTabList?"tab":"button"}focus(){this.label&&(this.label.tabIndex=0,this.label.focus())}blur(){this.label&&(this.label.tabIndex=-1)}setFocusable(e){this.label&&(this.label.tabIndex=e?0:-1)}updateLabel(){this.options.label&&this.label&&(this.label.textContent=this.action.label)}getTooltip(){let e=null;return this.action.tooltip?e=this.action.tooltip:this.action.label&&(e=this.action.label,this.options.keybinding&&(e=m(0,"{0} ({1})",e,this.options.keybinding))),e??void 0}updateClass(){this.cssClass&&this.label&&this.label.classList.remove(...this.cssClass.split(" ")),this.options.icon?(this.cssClass=this.getClass(),this.label&&(this.label.classList.add("codicon"),this.cssClass&&this.label.classList.add(...this.cssClass.split(" "))),this.updateEnabled()):this.label?.classList.remove("codicon")}updateEnabled(){this.action.enabled?(this.label&&(this.label.removeAttribute("aria-disabled"),this.label.classList.remove("disabled")),this.element?.classList.remove("disabled")):(this.label&&(this.label.setAttribute("aria-disabled","true"),this.label.classList.add("disabled")),this.element?.classList.add("disabled"))}updateAriaLabel(){if(this.label){const e=this.getTooltip()??"";this.label.setAttribute("aria-label",e)}}updateChecked(){this.label&&(this.action.checked!==void 0?(this.label.classList.toggle("checked",this.action.checked),this.options.isTabList?this.label.setAttribute("aria-selected",this.action.checked?"true":"false"):(this.label.setAttribute("aria-checked",this.action.checked?"true":"false"),this.label.setAttribute("role","checkbox"))):(this.label.classList.remove("checked"),this.label.removeAttribute(this.options.isTabList?"aria-selected":"aria-checked"),this.label.setAttribute("role",this.getDefaultAriaRole())))}}class OLe extends ll{constructor(e,t,i,n,o,r,a){super(e,t),this.selectBox=new MLe(i,n,o,r,a),this.selectBox.setFocusable(!1),this._register(this.selectBox),this.registerListeners()}select(e){this.selectBox.select(e)}registerListeners(){this._register(this.selectBox.onDidSelect(e=>this.runAction(e.selected,e.index)))}runAction(e,t){this.actionRunner.run(this._action,this.getActionContext(e,t))}getActionContext(e,t){return e}setFocusable(e){this.selectBox.setFocusable(e)}focus(){this.selectBox?.focus()}blur(){this.selectBox?.blur()}render(e){this.selectBox.render(e)}}class bo extends F{get onDidBlur(){return this._onDidBlur.event}get onDidCancel(){return this._onDidCancel.event}get onDidRun(){return this._onDidRun.event}get onWillRun(){return this._onWillRun.event}constructor(e,t={}){super(),this._actionRunnerDisposables=this._register(new U),this.viewItemDisposables=this._register(new LR),this.triggerKeyDown=!1,this.focusable=!0,this._onDidBlur=this._register(new M),this._onDidCancel=this._register(new M({onWillAddFirstListener:()=>this.cancelHasListener=!0})),this.cancelHasListener=!1,this._onDidRun=this._register(new M),this._onWillRun=this._register(new M),this.options=t,this._context=t.context??null,this._orientation=this.options.orientation??0,this._triggerKeys={keyDown:this.options.triggerKeys?.keyDown??!1,keys:this.options.triggerKeys?.keys??[3,10]},this._hoverDelegate=t.hoverDelegate??this._register(wte()),this.options.actionRunner?this._actionRunner=this.options.actionRunner:(this._actionRunner=new Lm,this._actionRunnerDisposables.add(this._actionRunner)),this._actionRunnerDisposables.add(this._actionRunner.onDidRun(o=>this._onDidRun.fire(o))),this._actionRunnerDisposables.add(this._actionRunner.onWillRun(o=>this._onWillRun.fire(o))),this.viewItems=[],this.focusedItem=void 0,this.domNode=document.createElement("div"),this.domNode.className="monaco-action-bar";let i,n;switch(this._orientation){case 0:i=[15],n=[17];break;case 1:i=[16],n=[18],this.domNode.className+=" vertical";break}this._register(W(this.domNode,se.KEY_DOWN,o=>{const r=new Pt(o);let a=!0;const l=typeof this.focusedItem=="number"?this.viewItems[this.focusedItem]:void 0;i&&(r.equals(i[0])||r.equals(i[1]))?a=this.focusPrevious():n&&(r.equals(n[0])||r.equals(n[1]))?a=this.focusNext():r.equals(9)&&this.cancelHasListener?this._onDidCancel.fire():r.equals(14)?a=this.focusFirst():r.equals(13)?a=this.focusLast():r.equals(2)&&l instanceof ll&&l.trapsArrowNavigation?a=this.focusNext(void 0,!0):this.isTriggerKeyEvent(r)?this._triggerKeys.keyDown?this.doTrigger(r):this.triggerKeyDown=!0:a=!1,a&&(r.preventDefault(),r.stopPropagation())})),this._register(W(this.domNode,se.KEY_UP,o=>{const r=new Pt(o);this.isTriggerKeyEvent(r)?(!this._triggerKeys.keyDown&&this.triggerKeyDown&&(this.triggerKeyDown=!1,this.doTrigger(r)),r.preventDefault(),r.stopPropagation()):(r.equals(2)||r.equals(1026)||r.equals(16)||r.equals(18)||r.equals(15)||r.equals(17))&&this.updateFocusedItem()})),this.focusTracker=this._register(oa(this.domNode)),this._register(this.focusTracker.onDidBlur(()=>{(ln()===this.domNode||!wn(ln(),this.domNode))&&(this._onDidBlur.fire(),this.previouslyFocusedItem=this.focusedItem,this.focusedItem=void 0,this.triggerKeyDown=!1)})),this._register(this.focusTracker.onDidFocus(()=>this.updateFocusedItem())),this.actionsList=document.createElement("ul"),this.actionsList.className="actions-container",this.options.highlightToggledItems&&this.actionsList.classList.add("highlight-toggled"),this.actionsList.setAttribute("role",this.options.ariaRole||"toolbar"),this.options.ariaLabel&&this.actionsList.setAttribute("aria-label",this.options.ariaLabel),this.domNode.appendChild(this.actionsList),e.appendChild(this.domNode)}refreshRole(){this.length()>=1?this.actionsList.setAttribute("role",this.options.ariaRole||"toolbar"):this.actionsList.setAttribute("role","presentation")}setFocusable(e){if(this.focusable=e,this.focusable){const t=this.viewItems.find(i=>i instanceof ll&&i.isEnabled());t instanceof ll&&t.setFocusable(!0)}else this.viewItems.forEach(t=>{t instanceof ll&&t.setFocusable(!1)})}isTriggerKeyEvent(e){let t=!1;return this._triggerKeys.keys.forEach(i=>{t=t||e.equals(i)}),t}updateFocusedItem(){for(let e=0;e<this.actionsList.children.length;e++){const t=this.actionsList.children[e];if(wn(ln(),t)){this.focusedItem=e,this.viewItems[this.focusedItem]?.showHover?.();break}}}get context(){return this._context}set context(e){this._context=e,this.viewItems.forEach(t=>t.setActionContext(e))}get actionRunner(){return this._actionRunner}set actionRunner(e){this._actionRunner=e,this._actionRunnerDisposables.clear(),this._actionRunnerDisposables.add(this._actionRunner.onDidRun(t=>this._onDidRun.fire(t))),this._actionRunnerDisposables.add(this._actionRunner.onWillRun(t=>this._onWillRun.fire(t))),this.viewItems.forEach(t=>t.actionRunner=e)}getContainer(){return this.domNode}getAction(e){if(typeof e=="number")return this.viewItems[e]?.action;if(ki(e)){for(;e.parentElement!==this.actionsList;){if(!e.parentElement)return;e=e.parentElement}for(let t=0;t<this.actionsList.childNodes.length;t++)if(this.actionsList.childNodes[t]===e)return this.viewItems[t].action}}push(e,t={}){const i=Array.isArray(e)?e:[e];let n=ch(t.index)?t.index:null;if(i.forEach(o=>{const r=document.createElement("li");r.className="action-item",r.setAttribute("role","presentation");let a;const l={hoverDelegate:this._hoverDelegate,...t,isTabList:this.options.ariaRole==="tablist"};this.options.actionViewItemProvider&&(a=this.options.actionViewItemProvider(o,l)),a||(a=new vv(this.context,o,l)),this.options.allowContextMenu||this.viewItemDisposables.set(a,W(r,se.CONTEXT_MENU,c=>{gt.stop(c,!0)})),a.actionRunner=this._actionRunner,a.setActionContext(this.context),a.render(r),n===null||n<0||n>=this.actionsList.children.length?(this.actionsList.appendChild(r),this.viewItems.push(a)):(this.actionsList.insertBefore(r,this.actionsList.children[n]),this.viewItems.splice(n,0,a),n++)}),this.focusable){let o=!1;for(const r of this.viewItems){if(!(r instanceof ll))continue;let a;o||r.action.id===tn.ID||!r.isEnabled()&&this.options.focusOnlyEnabledItems?a=!1:a=!0,a?(r.setFocusable(!0),o=!0):r.setFocusable(!1)}}typeof this.focusedItem=="number"&&this.focus(this.focusedItem),this.refreshRole()}getWidth(e){if(e>=0&&e<this.actionsList.children.length){const t=this.actionsList.children.item(e);if(t)return t.clientWidth}return 0}pull(e){e>=0&&e<this.viewItems.length&&(this.actionsList.childNodes[e].remove(),this.viewItemDisposables.deleteAndDispose(this.viewItems[e]),xt(this.viewItems.splice(e,1)),this.refreshRole())}clear(){this.isEmpty()||(this.viewItems=xt(this.viewItems),this.viewItemDisposables.clearAndDisposeAll(),Pn(this.actionsList),this.refreshRole())}length(){return this.viewItems.length}isEmpty(){return this.viewItems.length===0}focus(e){let t=!1,i;if(e===void 0?t=!0:typeof e=="number"?i=e:typeof e=="boolean"&&(t=e),t&&typeof this.focusedItem>"u"){const n=this.viewItems.findIndex(o=>o.isEnabled());this.focusedItem=n===-1?void 0:n,this.updateFocus(void 0,void 0,!0)}else i!==void 0&&(this.focusedItem=i),this.updateFocus(void 0,void 0,!0)}focusFirst(){return this.focusedItem=this.length()-1,this.focusNext(!0)}focusLast(){return this.focusedItem=0,this.focusPrevious(!0)}focusNext(e,t){if(typeof this.focusedItem>"u")this.focusedItem=this.viewItems.length-1;else if(this.viewItems.length<=1)return!1;const i=this.focusedItem;let n;do{if(!e&&this.options.preventLoopNavigation&&this.focusedItem+1>=this.viewItems.length)return this.focusedItem=i,!1;this.focusedItem=(this.focusedItem+1)%this.viewItems.length,n=this.viewItems[this.focusedItem]}while(this.focusedItem!==i&&(this.options.focusOnlyEnabledItems&&!n.isEnabled()||n.action.id===tn.ID));return this.updateFocus(void 0,void 0,t),!0}focusPrevious(e){if(typeof this.focusedItem>"u")this.focusedItem=0;else if(this.viewItems.length<=1)return!1;const t=this.focusedItem;let i;do{if(this.focusedItem=this.focusedItem-1,this.focusedItem<0){if(!e&&this.options.preventLoopNavigation)return this.focusedItem=t,!1;this.focusedItem=this.viewItems.length-1}i=this.viewItems[this.focusedItem]}while(this.focusedItem!==t&&(this.options.focusOnlyEnabledItems&&!i.isEnabled()||i.action.id===tn.ID));return this.updateFocus(!0),!0}updateFocus(e,t,i=!1){typeof this.focusedItem>"u"&&this.actionsList.focus({preventScroll:t}),this.previouslyFocusedItem!==void 0&&this.previouslyFocusedItem!==this.focusedItem&&this.viewItems[this.previouslyFocusedItem]?.blur();const n=this.focusedItem!==void 0?this.viewItems[this.focusedItem]:void 0;if(n){let o=!0;vm(n.focus)||(o=!1),this.options.focusOnlyEnabledItems&&vm(n.isEnabled)&&!n.isEnabled()&&(o=!1),n.action.id===tn.ID&&(o=!1),o?(i||this.previouslyFocusedItem!==this.focusedItem)&&(n.focus(e),this.previouslyFocusedItem=this.focusedItem):(this.actionsList.focus({preventScroll:t}),this.previouslyFocusedItem=void 0),o&&n.showHover?.()}}doTrigger(e){if(typeof this.focusedItem>"u")return;const t=this.viewItems[this.focusedItem];if(t instanceof ll){const i=t._context===null||t._context===void 0?e:t._context;this.run(t._action,i)}}async run(e,t){await this._actionRunner.run(e,t)}dispose(){this._context=void 0,this.viewItems=xt(this.viewItems),this.getContainer().remove(),super.dispose()}}function PLe(s,e,t,i){if(s.length===0)return e;if(e.length===0)return s;const n=[];let o=0,r=0;for(;o<s.length&&r<e.length;){const a=s[o],l=e[r],c=t(a),d=t(l);c<d?(n.push(a),o++):c>d?(n.push(l),r++):(n.push(i(a,l)),o++,r++)}for(;o<s.length;)n.push(s[o]),o++;for(;r<e.length;)n.push(e[r]),r++;return n}function CI(s,e){const t=new U,i=s.createDecorationsCollection();return t.add(DR({debugName:()=>`Apply decorations from ${e.debugName}`},n=>{const o=e.read(n);i.set(o)})),t.add({dispose:()=>{i.clear()}}),t}function Nb(s,e){return s.appendChild(e),pe(()=>{e.remove()})}function FLe(s,e){return s.prepend(e),pe(()=>{e.remove()})}class hie extends F{get width(){return this._width}get height(){return this._height}get automaticLayout(){return this._automaticLayout}constructor(e,t){super(),this._automaticLayout=!1,this.elementSizeObserver=this._register(new vJ(e,t)),this._width=Re(this,this.elementSizeObserver.getWidth()),this._height=Re(this,this.elementSizeObserver.getHeight()),this._register(this.elementSizeObserver.onDidChange(i=>Vt(n=>{this._width.set(this.elementSizeObserver.getWidth(),n),this._height.set(this.elementSizeObserver.getHeight(),n)})))}observe(e){this.elementSizeObserver.observe(e)}setAutomaticLayout(e){this._automaticLayout=e,e?this.elementSizeObserver.startObserving():this.elementSizeObserver.stopObserving()}}function Rq(s,e,t){let i=e.get(),n=i,o=i;const r=Re("animatedValue",i);let a=-1;const l=300;let c;t.add(jv({changeTracker:{createChangeSummary:()=>({animate:!1}),handleChange:(h,u)=>(h.didChange(e)&&(u.animate=u.animate||h.change),!0)}},(h,u)=>{c!==void 0&&(s.cancelAnimationFrame(c),c=void 0),n=o,i=e.read(h),a=Date.now()-(u.animate?0:l),d()}));function d(){const h=Date.now()-a;o=Math.floor(BLe(h,n,i-n,l)),h<l?c=s.requestAnimationFrame(d):o=i,r.set(o,void 0)}return r}function BLe(s,e,t,i){return s===i?e+t:t*(-Math.pow(2,-10*s/i)+1)+e}class M7 extends F{constructor(e,t,i){super(),this._register(new a4(e,i)),this._register(_g(i,{height:t.actualHeight,top:t.actualTop}))}}class Rb{get afterLineNumber(){return this._afterLineNumber.get()}constructor(e,t){this._afterLineNumber=e,this.heightInPx=t,this.domNode=document.createElement("div"),this._actualTop=Re(this,void 0),this._actualHeight=Re(this,void 0),this.actualTop=this._actualTop,this.actualHeight=this._actualHeight,this.showInHiddenAreas=!0,this.onChange=this._afterLineNumber,this.onDomNodeTop=i=>{this._actualTop.set(i,void 0)},this.onComputedHeight=i=>{this._actualHeight.set(i,void 0)}}}const aN=class aN{constructor(e,t){this._editor=e,this._domElement=t,this._overlayWidgetId=`managedOverlayWidget-${aN._counter++}`,this._overlayWidget={getId:()=>this._overlayWidgetId,getDomNode:()=>this._domElement,getPosition:()=>null},this._editor.addOverlayWidget(this._overlayWidget)}dispose(){this._editor.removeOverlayWidget(this._overlayWidget)}};aN._counter=0;let a4=aN;function _g(s,e){return De(t=>{for(let[i,n]of Object.entries(e))n&&typeof n=="object"&&"read"in n&&(n=n.read(t)),typeof n=="number"&&(n=`${n}px`),i=i.replace(/[A-Z]/g,o=>"-"+o.toLowerCase()),s.style[i]=n})}function yI(s,e,t,i){const n=new U,o=[];return n.add(ps((r,a)=>{const l=e.read(r),c=new Map,d=new Map;t&&t(!0),s.changeViewZones(h=>{for(const u of o)h.removeZone(u),i?.delete(u);o.length=0;for(const u of l){const g=h.addZone(u);u.setZoneId&&u.setZoneId(g),o.push(g),i?.add(g),c.set(u,g)}}),t&&t(!1),a.add(jv({changeTracker:{createChangeSummary(){return{zoneIds:[]}},handleChange(h,u){const g=d.get(h.changedObservable);return g!==void 0&&u.zoneIds.push(g),!0}}},(h,u)=>{for(const g of l)g.onChange&&(d.set(g.onChange,c.get(g)),g.onChange.read(h));t&&t(!0),s.changeViewZones(g=>{for(const f of u.zoneIds)g.layoutZone(f)}),t&&t(!1)}))})),n.add({dispose(){t&&t(!0),s.changeViewZones(r=>{for(const a of o)r.removeZone(a)}),i?.clear(),t&&t(!1)}}),n}class WLe extends si{dispose(){super.dispose(!0)}}function Aq(s,e){const t=gS(e,n=>n.original.startLineNumber<=s.lineNumber);if(!t)return k.fromPositions(s);if(t.original.endLineNumberExclusive<=s.lineNumber){const n=s.lineNumber-t.original.endLineNumberExclusive+t.modified.endLineNumberExclusive;return k.fromPositions(new A(n,s.column))}if(!t.innerChanges)return k.fromPositions(new A(t.modified.startLineNumber,1));const i=gS(t.innerChanges,n=>n.originalRange.getStartPosition().isBeforeOrEqual(s));if(!i){const n=s.lineNumber-t.original.startLineNumber+t.modified.startLineNumber;return k.fromPositions(new A(n,s.column))}if(i.originalRange.containsPosition(s))return i.modifiedRange;{const n=HLe(i.originalRange.getEndPosition(),s);return k.fromPositions(n.addToPosition(i.modifiedRange.getEndPosition()))}}function HLe(s,e){return s.lineNumber===e.lineNumber?new cn(0,e.column-s.column):new cn(e.lineNumber-s.lineNumber,e.column-1)}function VLe(s,e){let t;return s.filter(i=>{const n=e(i,t);return t=i,n})}class SI{static create(e,t=void 0){return new Mq(e,e,t)}static createWithDisposable(e,t,i=void 0){const n=new U;return n.add(t),n.add(e),new Mq(e,n,i)}}class Mq extends SI{constructor(e,t,i){super(),this.object=e,this._disposable=t,this._debugOwner=i,this._refCount=1,this._isDisposed=!1,this._owners=[],i&&this._addOwner(i)}_addOwner(e){e&&this._owners.push(e)}createNewRef(e){return this._refCount++,e&&this._addOwner(e),new zLe(this,e)}dispose(){this._isDisposed||(this._isDisposed=!0,this._decreaseRefCount(this._debugOwner))}_decreaseRefCount(e){if(this._refCount--,this._refCount===0&&this._disposable.dispose(),e){const t=this._owners.indexOf(e);t!==-1&&this._owners.splice(t,1)}}}class zLe extends SI{constructor(e,t){super(),this._base=e,this._debugOwner=t,this._isDisposed=!1}get object(){return this._base.object}createNewRef(e){return this._base.createNewRef(e)}dispose(){this._isDisposed||(this._isDisposed=!0,this._base._decreaseRefCount(this._debugOwner))}}class Ts{static inverse(e,t,i){const n=[];let o=1,r=1;for(const l of e){const c=new Ts(new Ne(o,l.original.startLineNumber),new Ne(r,l.modified.startLineNumber));c.modified.isEmpty||n.push(c),o=l.original.endLineNumberExclusive,r=l.modified.endLineNumberExclusive}const a=new Ts(new Ne(o,t+1),new Ne(r,i+1));return a.modified.isEmpty||n.push(a),n}static clip(e,t,i){const n=[];for(const o of e){const r=o.original.intersect(t),a=o.modified.intersect(i);r&&!r.isEmpty&&a&&!a.isEmpty&&n.push(new Ts(r,a))}return n}constructor(e,t){this.original=e,this.modified=t}toString(){return`{${this.original.toString()}->${this.modified.toString()}}`}flip(){return new Ts(this.modified,this.original)}join(e){return new Ts(this.original.join(e.original),this.modified.join(e.modified))}toRangeMapping(){const e=this.original.toInclusiveRange(),t=this.modified.toInclusiveRange();if(e&&t)return new Ks(e,t);if(this.original.startLineNumber===1||this.modified.startLineNumber===1){if(!(this.modified.startLineNumber===1&&this.original.startLineNumber===1))throw new Ce("not a valid diff");return new Ks(new k(this.original.startLineNumber,1,this.original.endLineNumberExclusive,1),new k(this.modified.startLineNumber,1,this.modified.endLineNumberExclusive,1))}else return new Ks(new k(this.original.startLineNumber-1,Number.MAX_SAFE_INTEGER,this.original.endLineNumberExclusive-1,Number.MAX_SAFE_INTEGER),new k(this.modified.startLineNumber-1,Number.MAX_SAFE_INTEGER,this.modified.endLineNumberExclusive-1,Number.MAX_SAFE_INTEGER))}toRangeMapping2(e,t){if(Oq(this.original.endLineNumberExclusive,e)&&Oq(this.modified.endLineNumberExclusive,t))return new Ks(new k(this.original.startLineNumber,1,this.original.endLineNumberExclusive,1),new k(this.modified.startLineNumber,1,this.modified.endLineNumberExclusive,1));if(!this.original.isEmpty&&!this.modified.isEmpty)return new Ks(k.fromPositions(new A(this.original.startLineNumber,1),U_(new A(this.original.endLineNumberExclusive-1,Number.MAX_SAFE_INTEGER),e)),k.fromPositions(new A(this.modified.startLineNumber,1),U_(new A(this.modified.endLineNumberExclusive-1,Number.MAX_SAFE_INTEGER),t)));if(this.original.startLineNumber>1&&this.modified.startLineNumber>1)return new Ks(k.fromPositions(U_(new A(this.original.startLineNumber-1,Number.MAX_SAFE_INTEGER),e),U_(new A(this.original.endLineNumberExclusive-1,Number.MAX_SAFE_INTEGER),e)),k.fromPositions(U_(new A(this.modified.startLineNumber-1,Number.MAX_SAFE_INTEGER),t),U_(new A(this.modified.endLineNumberExclusive-1,Number.MAX_SAFE_INTEGER),t)));throw new Ce}}function U_(s,e){if(s.lineNumber<1)return new A(1,1);if(s.lineNumber>e.length)return new A(e.length,e[e.length-1].length+1);const t=e[s.lineNumber-1];return s.column>t.length+1?new A(s.lineNumber,t.length+1):s}function Oq(s,e){return s>=1&&s<=e.length}class Lr extends Ts{static fromRangeMappings(e){const t=Ne.join(e.map(n=>Ne.fromRangeInclusive(n.originalRange))),i=Ne.join(e.map(n=>Ne.fromRangeInclusive(n.modifiedRange)));return new Lr(t,i,e)}constructor(e,t,i){super(e,t),this.innerChanges=i}flip(){return new Lr(this.modified,this.original,this.innerChanges?.map(e=>e.flip()))}withInnerChangesFromLineRanges(){return new Lr(this.original,this.modified,[this.toRangeMapping()])}}class Ks{static fromEdit(e){const t=e.getNewRanges();return e.replacements.map((n,o)=>new Ks(n.range,t[o]))}static assertSorted(e){for(let t=1;t<e.length;t++){const i=e[t-1],n=e[t];if(!(i.originalRange.getEndPosition().isBeforeOrEqual(n.originalRange.getStartPosition())&&i.modifiedRange.getEndPosition().isBeforeOrEqual(n.modifiedRange.getStartPosition())))throw new Ce("Range mappings must be sorted")}}constructor(e,t){this.originalRange=e,this.modifiedRange=t}toString(){return`{${this.originalRange.toString()}->${this.modifiedRange.toString()}}`}flip(){return new Ks(this.modifiedRange,this.originalRange)}toTextEdit(e){const t=e.getValueOfRange(this.modifiedRange);return new Ui(this.originalRange,t)}}function xI(s,e,t,i=!1){const n=[];for(const o of eW(s.map(r=>$Le(r,e,t)),(r,a)=>r.original.intersectsOrTouches(a.original)||r.modified.intersectsOrTouches(a.modified))){const r=o[0],a=o[o.length-1];n.push(new Lr(r.original.join(a.original),r.modified.join(a.modified),o.map(l=>l.innerChanges[0])))}return rg(()=>!i&&n.length>0&&(n[0].modified.startLineNumber!==n[0].original.startLineNumber||t.length.lineCount-n[n.length-1].modified.endLineNumberExclusive!==e.length.lineCount-n[n.length-1].original.endLineNumberExclusive)?!1:H0(n,(o,r)=>r.original.startLineNumber-o.original.endLineNumberExclusive===r.modified.startLineNumber-o.modified.endLineNumberExclusive&&o.original.endLineNumberExclusive<r.original.startLineNumber&&o.modified.endLineNumberExclusive<r.modified.startLineNumber)),n}function $Le(s,e,t){let i=0,n=0;s.modifiedRange.endColumn===1&&s.originalRange.endColumn===1&&s.originalRange.startLineNumber+i<=s.originalRange.endLineNumber&&s.modifiedRange.startLineNumber+i<=s.modifiedRange.endLineNumber&&(n=-1),s.modifiedRange.startColumn-1>=t.getLineLength(s.modifiedRange.startLineNumber)&&s.originalRange.startColumn-1>=e.getLineLength(s.originalRange.startLineNumber)&&s.originalRange.startLineNumber<=s.originalRange.endLineNumber+n&&s.modifiedRange.startLineNumber<=s.modifiedRange.endLineNumber+n&&(i=1);const o=new Ne(s.originalRange.startLineNumber+i,s.originalRange.endLineNumber+1+n),r=new Ne(s.modifiedRange.startLineNumber+i,s.modifiedRange.endLineNumber+1+n);return new Lr(o,r,[s])}const ULe={IconContribution:"base.contributions.icons"};var Pq;(function(s){function e(t,i){let n=t.defaults;for(;xe.isThemeIcon(n);){const o=g_.getIcon(n.id);if(!o)return;n=o.defaults}return n}s.getDefinition=e})(Pq||(Pq={}));var Fq;(function(s){function e(i){return{weight:i.weight,style:i.style,src:i.src.map(n=>({format:n.format,location:n.location.toString()}))}}s.toJSONObject=e;function t(i){const n=o=>bn(o)?o:void 0;if(i&&Array.isArray(i.src)&&i.src.every(o=>bn(o.format)&&bn(o.location)))return{weight:n(i.weight),style:n(i.style),src:i.src.map(o=>({format:o.format,location:Le.parse(o.location)}))}}s.fromJSONObject=t})(Fq||(Fq={}));const qLe=/^([\w_-]+)$/,jLe=m(2024,"The font ID must only contain letters, numbers, underscores and dashes.");class GLe extends F{constructor(){super(),this._onDidChange=this._register(new M),this.onDidChange=this._onDidChange.event,this.iconSchema={definitions:{icons:{type:"object",properties:{fontId:{type:"string",description:m(2025,"The id of the font to use. If not set, the font that is defined first is used."),pattern:qLe.source,patternErrorMessage:jLe},fontCharacter:{type:"string",description:m(2026,"The font character associated with the icon definition.")}},additionalProperties:!1,defaultSnippets:[{body:{fontCharacter:"\\\\e030"}}]}},type:"object",properties:{}},this.iconReferenceSchema={type:"string",pattern:`^${xe.iconNameExpression}$`,enum:[],enumDescriptions:[]},this.iconsById={},this.iconFontsById={}}registerIcon(e,t,i,n){const o=this.iconsById[e];if(o){if(i&&!o.description){o.description=i,this.iconSchema.properties[e].markdownDescription=`${i} $(${e})`;const l=this.iconReferenceSchema.enum.indexOf(e);l!==-1&&(this.iconReferenceSchema.enumDescriptions[l]=i),this._onDidChange.fire()}return o}const r={id:e,description:i,defaults:t,deprecationMessage:n};this.iconsById[e]=r;const a={$ref:"#/definitions/icons"};return n&&(a.deprecationMessage=n),i&&(a.markdownDescription=`${i}: $(${e})`),this.iconSchema.properties[e]=a,this.iconReferenceSchema.enum.push(e),this.iconReferenceSchema.enumDescriptions.push(i||""),this._onDidChange.fire(),{id:e}}getIcons(){return Object.keys(this.iconsById).map(e=>this.iconsById[e])}getIcon(e){return this.iconsById[e]}getIconSchema(){return this.iconSchema}toString(){const e=(o,r)=>o.id.localeCompare(r.id),t=o=>{for(;xe.isThemeIcon(o.defaults);)o=this.iconsById[o.defaults.id];return`codicon codicon-${o?o.id:""}`},i=[];i.push("| preview     | identifier                        | default codicon ID                | description"),i.push("| ----------- | --------------------------------- | --------------------------------- | --------------------------------- |");const n=Object.keys(this.iconsById).map(o=>this.iconsById[o]);for(const o of n.filter(r=>!!r.description).sort(e))i.push(`|<i class="${t(o)}"></i>|${o.id}|${xe.isThemeIcon(o.defaults)?o.defaults.id:o.id}|${o.description||""}|`);i.push("| preview     | identifier                        "),i.push("| ----------- | --------------------------------- |");for(const o of n.filter(r=>!xe.isThemeIcon(r.defaults)).sort(e))i.push(`|<i class="${t(o)}"></i>|${o.id}|`);return i.join(`
`)}}const g_=new GLe;bi.add(ULe.IconContribution,g_);function Qt(s,e,t,i){return g_.registerIcon(s,e,t,i)}function uie(){return g_}function KLe(){const s=WQ();for(const e in s){const t="\\"+s[e].toString(16);g_.registerIcon(e,{fontCharacter:t})}}KLe();const gie="vscode://schemas/icons",fie=bi.as(YR.JSONContribution);fie.registerSchema(gie,g_.getIconSchema());const Bq=new Rt(()=>fie.notifySchemaChanged(gie),200);g_.onDidChange(()=>{Bq.isScheduled()||Bq.schedule()});const pie=Qt("widget-close",Z.close,m(2027,"Icon for the close action in widgets."));Qt("goto-previous-location",Z.arrowUp,m(2028,"Icon for goto previous editor location."));Qt("goto-next-location",Z.arrowDown,m(2029,"Icon for goto next editor location."));xe.modify(Z.sync,"spin");xe.modify(Z.loading,"spin");var O7=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},P7=function(s,e){return function(t,i){e(t,i,s)}};const YLe=Qt("diff-review-insert",Z.add,m(97,"Icon for 'Insert' in accessible diff viewer.")),ZLe=Qt("diff-review-remove",Z.remove,m(98,"Icon for 'Remove' in accessible diff viewer.")),XLe=Qt("diff-review-close",Z.close,m(99,"Icon for 'Close' in accessible diff viewer."));var Vw;let Zf=(Vw=class extends F{constructor(e,t,i,n,o,r,a,l,c){super(),this._parentNode=e,this._visible=t,this._setVisible=i,this._canClose=n,this._width=o,this._height=r,this._diffs=a,this._models=l,this._instantiationService=c,this._state=j(this,d=>{const h=this._visible.read(d);if(this._parentNode.style.visibility=h?"visible":"hidden",!h)return null;const u=d.store.add(this._instantiationService.createInstance(l4,this._diffs,this._models,this._setVisible,this._canClose)),g=d.store.add(this._instantiationService.createInstance(c4,this._parentNode,u,this._width,this._height,this._models));return{model:u,view:g}}).recomputeInitiallyAndOnChange(this._store)}next(){Vt(e=>{const t=this._visible.get();this._setVisible(!0,e),t&&this._state.get().model.nextGroup(e)})}prev(){Vt(e=>{this._setVisible(!0,e),this._state.get().model.previousGroup(e)})}close(){Vt(e=>{this._setVisible(!1,e)})}},Vw._ttPolicy=$c("diffReview",{createHTML:e=>e}),Vw);Zf=O7([P7(8,_e)],Zf);let l4=class extends F{constructor(e,t,i,n,o){super(),this._diffs=e,this._models=t,this._setVisible=i,this.canClose=n,this._accessibilitySignalService=o,this._groups=Re(this,[]),this._currentGroupIdx=Re(this,0),this._currentElementIdx=Re(this,0),this.groups=this._groups,this.currentGroup=this._currentGroupIdx.map((r,a)=>this._groups.read(a)[r]),this.currentGroupIndex=this._currentGroupIdx,this.currentElement=this._currentElementIdx.map((r,a)=>this.currentGroup.read(a)?.lines[r]),this._register(De(r=>{const a=this._diffs.read(r);if(!a){this._groups.set([],void 0);return}const l=QLe(a,this._models.getOriginalModel().getLineCount(),this._models.getModifiedModel().getLineCount());Vt(c=>{const d=this._models.getModifiedPosition();if(d){const h=l.findIndex(u=>d?.lineNumber<u.range.modified.endLineNumberExclusive);h!==-1&&this._currentGroupIdx.set(h,c)}this._groups.set(l,c)})})),this._register(De(r=>{const a=this.currentElement.read(r);a?.type===Ys.Deleted?this._accessibilitySignalService.playSignal(Gs.diffLineDeleted,{source:"accessibleDiffViewer.currentElementChanged"}):a?.type===Ys.Added&&this._accessibilitySignalService.playSignal(Gs.diffLineInserted,{source:"accessibleDiffViewer.currentElementChanged"})})),this._register(De(r=>{const a=this.currentElement.read(r);if(a&&a.type!==Ys.Header){const l=a.modifiedLineNumber??a.diff.modified.startLineNumber;this._models.modifiedSetSelection(k.fromPositions(new A(l,1)))}}))}_goToGroupDelta(e,t){const i=this.groups.get();!i||i.length<=1||nv(t,n=>{this._currentGroupIdx.set(me.ofLength(i.length).clipCyclic(this._currentGroupIdx.get()+e),n),this._currentElementIdx.set(0,n)})}nextGroup(e){this._goToGroupDelta(1,e)}previousGroup(e){this._goToGroupDelta(-1,e)}_goToLineDelta(e){const t=this.currentGroup.get();!t||t.lines.length<=1||Vt(i=>{this._currentElementIdx.set(me.ofLength(t.lines.length).clip(this._currentElementIdx.get()+e),i)})}goToNextLine(){this._goToLineDelta(1)}goToPreviousLine(){this._goToLineDelta(-1)}goToLine(e){const t=this.currentGroup.get();if(!t)return;const i=t.lines.indexOf(e);i!==-1&&Vt(n=>{this._currentElementIdx.set(i,n)})}revealCurrentElementInEditor(){if(!this.canClose.get())return;this._setVisible(!1,void 0);const e=this.currentElement.get();e&&(e.type===Ys.Deleted?this._models.originalReveal(k.fromPositions(new A(e.originalLineNumber,1))):this._models.modifiedReveal(e.type!==Ys.Header?k.fromPositions(new A(e.modifiedLineNumber,1)):void 0))}close(){this.canClose.get()&&(this._setVisible(!1,void 0),this._models.modifiedFocus())}};l4=O7([P7(4,Th)],l4);const HC=3;function QLe(s,e,t){const i=[];for(const n of eW(s,(o,r)=>r.modified.startLineNumber-o.modified.endLineNumberExclusive<2*HC)){const o=[];o.push(new eke);const r=new Ne(Math.max(1,n[0].original.startLineNumber-HC),Math.min(n[n.length-1].original.endLineNumberExclusive+HC,e+1)),a=new Ne(Math.max(1,n[0].modified.startLineNumber-HC),Math.min(n[n.length-1].modified.endLineNumberExclusive+HC,t+1));fQ(n,(d,h)=>{const u=new Ne(d?d.original.endLineNumberExclusive:r.startLineNumber,h?h.original.startLineNumber:r.endLineNumberExclusive),g=new Ne(d?d.modified.endLineNumberExclusive:a.startLineNumber,h?h.modified.startLineNumber:a.endLineNumberExclusive);u.forEach(f=>{o.push(new nke(f,g.startLineNumber+(f-u.startLineNumber)))}),h&&(h.original.forEach(f=>{o.push(new tke(h,f))}),h.modified.forEach(f=>{o.push(new ike(h,f))}))});const l=n[0].modified.join(n[n.length-1].modified),c=n[0].original.join(n[n.length-1].original);i.push(new JLe(new Ts(l,c),o))}return i}var Ys;(function(s){s[s.Header=0]="Header",s[s.Unchanged=1]="Unchanged",s[s.Deleted=2]="Deleted",s[s.Added=3]="Added"})(Ys||(Ys={}));class JLe{constructor(e,t){this.range=e,this.lines=t}}class eke{constructor(){this.type=Ys.Header}}class tke{constructor(e,t){this.diff=e,this.originalLineNumber=t,this.type=Ys.Deleted,this.modifiedLineNumber=void 0}}class ike{constructor(e,t){this.diff=e,this.modifiedLineNumber=t,this.type=Ys.Added,this.originalLineNumber=void 0}}class nke{constructor(e,t){this.originalLineNumber=e,this.modifiedLineNumber=t,this.type=Ys.Unchanged}}let c4=class extends F{constructor(e,t,i,n,o,r){super(),this._element=e,this._model=t,this._width=i,this._height=n,this._models=o,this._languageService=r,this.domNode=this._element,this.domNode.className="monaco-component diff-review monaco-editor-background";const a=document.createElement("div");a.className="diff-review-actions",this._actionBar=this._register(new bo(a)),this._register(De(l=>{this._actionBar.clear(),this._model.canClose.read(l)&&this._actionBar.push(rp({id:"diffreview.close",label:m(100,"Close"),class:"close-diff-review "+xe.asClassName(XLe),enabled:!0,run:async()=>t.close()}),{label:!1,icon:!0})})),this._content=document.createElement("div"),this._content.className="diff-review-content",this._content.setAttribute("role","code"),this._scrollbar=this._register(new tx(this._content,{})),vn(this.domNode,this._scrollbar.getDomNode(),a),this._register(De(l=>{this._height.read(l),this._width.read(l),this._scrollbar.scanDomNode()})),this._register(pe(()=>{vn(this.domNode)})),this._register(_g(this.domNode,{width:this._width,height:this._height})),this._register(_g(this._content,{width:this._width,height:this._height})),this._register(ps((l,c)=>{this._model.currentGroup.read(l),this._render(c)})),this._register(Fi(this.domNode,"keydown",l=>{(l.equals(18)||l.equals(2066)||l.equals(530))&&(l.preventDefault(),this._model.goToNextLine()),(l.equals(16)||l.equals(2064)||l.equals(528))&&(l.preventDefault(),this._model.goToPreviousLine()),(l.equals(9)||l.equals(2057)||l.equals(521)||l.equals(1033))&&(l.preventDefault(),this._model.close()),(l.equals(10)||l.equals(3))&&(l.preventDefault(),this._model.revealCurrentElementInEditor())}))}_render(e){const t=this._models.getOriginalOptions(),i=this._models.getModifiedOptions(),n=document.createElement("div");n.className="diff-review-table",n.setAttribute("role","list"),n.setAttribute("aria-label",m(101,"Accessible Diff Viewer. Use arrow up and down to navigate.")),In(n,i.get(59)),vn(this._content,n);const o=this._models.getOriginalModel(),r=this._models.getModifiedModel();if(!o||!r)return;const a=o.getOptions(),l=r.getOptions(),c=i.get(75),d=this._model.currentGroup.get();for(const h of d?.lines||[]){if(!d)break;let u;if(h.type===Ys.Header){const f=document.createElement("div");f.className="diff-review-row",f.setAttribute("role","listitem");const p=d.range,_=this._model.currentGroupIndex.get(),b=this._model.groups.get().length,w=x=>x===0?m(102,"no lines changed"):x===1?m(103,"1 line changed"):m(104,"{0} lines changed",x),C=w(p.original.length),v=w(p.modified.length);f.setAttribute("aria-label",m(105,"Difference {0} of {1}: original line {2}, {3}, modified line {4}, {5}",_+1,b,p.original.startLineNumber,C,p.modified.startLineNumber,v));const y=document.createElement("div");y.className="diff-review-cell diff-review-summary",y.appendChild(document.createTextNode(`${_+1}/${b}: @@ -${p.original.startLineNumber},${p.original.length} +${p.modified.startLineNumber},${p.modified.length} @@`)),f.appendChild(y),u=f}else u=this._createRow(h,c,this._width.get(),t,o,a,i,r,l);n.appendChild(u);const g=j(f=>this._model.currentElement.read(f)===h);e.add(De(f=>{const p=g.read(f);u.tabIndex=p?0:-1,p&&u.focus()})),e.add(W(u,"focus",()=>{this._model.goToLine(h)}))}this._scrollbar.scanDomNode()}_createRow(e,t,i,n,o,r,a,l,c){const d=n.get(165),h=d.glyphMarginWidth+d.lineNumbersWidth,u=a.get(165),g=10+u.glyphMarginWidth+u.lineNumbersWidth;let f="diff-review-row",p="";const _="diff-review-spacer";let b=null;switch(e.type){case Ys.Added:f="diff-review-row line-insert",p=" char-insert",b=YLe;break;case Ys.Deleted:f="diff-review-row line-delete",p=" char-delete",b=ZLe;break}const w=document.createElement("div");w.style.minWidth=i+"px",w.className=f,w.setAttribute("role","listitem"),w.ariaLevel="";const C=document.createElement("div");C.className="diff-review-cell",C.style.height=`${t}px`,w.appendChild(C);const v=document.createElement("span");v.style.width=h+"px",v.style.minWidth=h+"px",v.className="diff-review-line-number"+p,e.originalLineNumber!==void 0?v.appendChild(document.createTextNode(String(e.originalLineNumber))):v.innerText=" ",C.appendChild(v);const y=document.createElement("span");y.style.width=g+"px",y.style.minWidth=g+"px",y.style.paddingRight="10px",y.className="diff-review-line-number"+p,e.modifiedLineNumber!==void 0?y.appendChild(document.createTextNode(String(e.modifiedLineNumber))):y.innerText=" ",C.appendChild(y);const x=document.createElement("span");if(x.className=_,b){const I=document.createElement("span");I.className=xe.asClassName(b),I.innerText="  ",x.appendChild(I)}else x.innerText="  ";C.appendChild(x);let L;if(e.modifiedLineNumber!==void 0){let I=this._getLineHtml(l,a,c.tabSize,e.modifiedLineNumber,this._languageService.languageIdCodec);Zf._ttPolicy&&(I=Zf._ttPolicy.createHTML(I)),C.insertAdjacentHTML("beforeend",I),L=l.getLineContent(e.modifiedLineNumber)}else{let I=this._getLineHtml(o,n,r.tabSize,e.originalLineNumber,this._languageService.languageIdCodec);Zf._ttPolicy&&(I=Zf._ttPolicy.createHTML(I)),C.insertAdjacentHTML("beforeend",I),L=o.getLineContent(e.originalLineNumber)}L.length===0&&(L=m(106,"blank"));let E="";switch(e.type){case Ys.Unchanged:e.originalLineNumber===e.modifiedLineNumber?E=m(107,"{0} unchanged line {1}",L,e.originalLineNumber):E=m(108,"{0} original line {1} modified line {2}",L,e.originalLineNumber,e.modifiedLineNumber);break;case Ys.Added:E=m(109,"+ {0} modified line {1}",L,e.modifiedLineNumber);break;case Ys.Deleted:E=m(110,"- {0} original line {1}",L,e.originalLineNumber);break}return w.setAttribute("aria-label",E),w}_getLineHtml(e,t,i,n,o){const r=e.getLineContent(n),a=t.get(59),l=t.get(117).verticalScrollbarSize,c=Ri.createEmpty(r,o),d=Er.isBasicASCII(r,e.mightContainNonBasicASCII()),h=Er.containsRTL(r,d,e.mightContainRTL());return lA(new Dh(a.isMonospace&&!t.get(40),a.canUseHalfwidthRightwardsArrow,r,!1,d,h,0,c,[],i,0,a.spaceWidth,a.middotWidth,a.wsmiddotWidth,t.get(133),t.get(113),t.get(108),t.get(60)!==mh.OFF,null,null,l)).html}};c4=O7([P7(5,Ei)],c4);class ske{constructor(e){this.editors=e}getOriginalModel(){return this.editors.original.getModel()}getOriginalOptions(){return this.editors.original.getOptions()}originalReveal(e){this.editors.original.revealRange(e),this.editors.original.setSelection(e),this.editors.original.focus()}getModifiedModel(){return this.editors.modified.getModel()}getModifiedOptions(){return this.editors.modified.getOptions()}modifiedReveal(e){e&&(this.editors.modified.revealRange(e),this.editors.modified.setSelection(e)),this.editors.modified.focus()}modifiedSetSelection(e){this.editors.modified.setSelection(e)}modifiedFocus(){this.editors.modified.focus()}getModifiedPosition(){return this.editors.modified.getPosition()??void 0}}T("diffEditor.move.border","#8b8b8b9c",m(137,"The border color for text that got moved in the diff editor."));T("diffEditor.moveActive.border","#FFA500",m(138,"The active border color for text that got moved in the diff editor."));T("diffEditor.unchangedRegionShadow",{dark:"#000000",light:"#737373BF",hcDark:"#000000",hcLight:"#737373BF"},m(139,"The color of the shadow around unchanged region widgets."));const oke=Qt("diff-insert",Z.add,m(140,"Line decoration for inserts in the diff editor.")),mie=Qt("diff-remove",Z.remove,m(141,"Line decoration for removals in the diff editor.")),Wq=Fe.register({className:"line-insert",description:"line-insert",isWholeLine:!0,linesDecorationsClassName:"insert-sign "+xe.asClassName(oke),marginClassName:"gutter-insert"}),Hq=Fe.register({className:"line-delete",description:"line-delete",isWholeLine:!0,linesDecorationsClassName:"delete-sign "+xe.asClassName(mie),marginClassName:"gutter-delete"}),Vq=Fe.register({className:"line-insert",description:"line-insert",isWholeLine:!0,marginClassName:"gutter-insert"}),zq=Fe.register({className:"line-delete",description:"line-delete",isWholeLine:!0,marginClassName:"gutter-delete"}),$q=Fe.register({className:"char-insert",description:"char-insert",shouldFillLineOnLineBreak:!0}),rke=Fe.register({className:"char-insert",description:"char-insert",isWholeLine:!0}),ake=Fe.register({className:"char-insert diff-range-empty",description:"char-insert diff-range-empty"}),d4=Fe.register({className:"char-delete",description:"char-delete",shouldFillLineOnLineBreak:!0}),lke=Fe.register({className:"char-delete",description:"char-delete",isWholeLine:!0}),cke=Fe.register({className:"char-delete diff-range-empty",description:"char-delete diff-range-empty"}),ko=qe("editorWorkerService");var _ie=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},h4=function(s,e){return function(t,i){e(t,i,s)}},rf;const bie=qe("diffProviderFactoryService");let u4=class{constructor(e){this.instantiationService=e}createDiffProvider(e){return this.instantiationService.createInstance(g4,e)}};u4=_ie([h4(0,_e)],u4);tt(bie,u4,1);var Lp;let g4=(Lp=class{constructor(e,t,i){this.editorWorkerService=t,this.telemetryService=i,this.onDidChangeEventEmitter=new M,this.onDidChange=this.onDidChangeEventEmitter.event,this.diffAlgorithm="advanced",this.diffAlgorithmOnDidChangeSubscription=void 0,this.setOptions(e)}dispose(){this.diffAlgorithmOnDidChangeSubscription?.dispose()}async computeDiff(e,t,i,n){if(typeof this.diffAlgorithm!="string")return this.diffAlgorithm.computeDiff(e,t,i,n);if(e.isDisposed()||t.isDisposed())return{changes:[],identical:!0,quitEarly:!1,moves:[]};if(e.getLineCount()===1&&e.getLineMaxColumn(1)===1)return t.getLineCount()===1&&t.getLineMaxColumn(1)===1?{changes:[],identical:!0,quitEarly:!1,moves:[]}:{changes:[new Lr(new Ne(1,2),new Ne(1,t.getLineCount()+1),[new Ks(e.getFullModelRange(),t.getFullModelRange())])],identical:!1,quitEarly:!1,moves:[]};const o=JSON.stringify([e.uri.toString(),t.uri.toString()]),r=JSON.stringify([e.id,t.id,e.getAlternativeVersionId(),t.getAlternativeVersionId(),JSON.stringify(i)]),a=rf.diffCache.get(o);if(a&&a.context===r)return a.result;const l=Cn.create(),c=await this.editorWorkerService.computeDiff(e.uri,t.uri,i,this.diffAlgorithm),d=l.elapsed();if(this.telemetryService.publicLog2("diffEditor.computeDiff",{timeMs:d,timedOut:c?.quitEarly??!0,detectedMoves:i.computeMoves?c?.moves.length??0:-1}),n.isCancellationRequested)return{changes:[],identical:!1,quitEarly:!0,moves:[]};if(!c)throw new Error("no diff result available");return rf.diffCache.size>10&&rf.diffCache.delete(rf.diffCache.keys().next().value),rf.diffCache.set(o,{result:c,context:r}),c}setOptions(e){let t=!1;e.diffAlgorithm&&this.diffAlgorithm!==e.diffAlgorithm&&(this.diffAlgorithmOnDidChangeSubscription?.dispose(),this.diffAlgorithmOnDidChangeSubscription=void 0,this.diffAlgorithm=e.diffAlgorithm,typeof e.diffAlgorithm!="string"&&(this.diffAlgorithmOnDidChangeSubscription=e.diffAlgorithm.onDidChange(()=>this.onDidChangeEventEmitter.fire())),t=!0),t&&this.onDidChangeEventEmitter.fire()}},rf=Lp,Lp.diffCache=new Map,Lp);g4=rf=_ie([h4(1,ko),h4(2,vs)],g4);class rh{static trivial(e,t){return new rh([new fn(me.ofLength(e.length),me.ofLength(t.length))],!1)}static trivialTimedOut(e,t){return new rh([new fn(me.ofLength(e.length),me.ofLength(t.length))],!0)}constructor(e,t){this.diffs=e,this.hitTimeout=t}}class fn{static invert(e,t){const i=[];return fQ(e,(n,o)=>{i.push(fn.fromOffsetPairs(n?n.getEndExclusives():qd.zero,o?o.getStarts():new qd(t,(n?n.seq2Range.endExclusive-n.seq1Range.endExclusive:0)+t)))}),i}static fromOffsetPairs(e,t){return new fn(new me(e.offset1,t.offset1),new me(e.offset2,t.offset2))}static assertSorted(e){let t;for(const i of e){if(t&&!(t.seq1Range.endExclusive<=i.seq1Range.start&&t.seq2Range.endExclusive<=i.seq2Range.start))throw new Ce("Sequence diffs must be sorted");t=i}}constructor(e,t){this.seq1Range=e,this.seq2Range=t}swap(){return new fn(this.seq2Range,this.seq1Range)}toString(){return`${this.seq1Range} <-> ${this.seq2Range}`}join(e){return new fn(this.seq1Range.join(e.seq1Range),this.seq2Range.join(e.seq2Range))}delta(e){return e===0?this:new fn(this.seq1Range.delta(e),this.seq2Range.delta(e))}deltaStart(e){return e===0?this:new fn(this.seq1Range.deltaStart(e),this.seq2Range.deltaStart(e))}deltaEnd(e){return e===0?this:new fn(this.seq1Range.deltaEnd(e),this.seq2Range.deltaEnd(e))}intersect(e){const t=this.seq1Range.intersect(e.seq1Range),i=this.seq2Range.intersect(e.seq2Range);if(!(!t||!i))return new fn(t,i)}getStarts(){return new qd(this.seq1Range.start,this.seq2Range.start)}getEndExclusives(){return new qd(this.seq1Range.endExclusive,this.seq2Range.endExclusive)}}const Mf=class Mf{constructor(e,t){this.offset1=e,this.offset2=t}toString(){return`${this.offset1} <-> ${this.offset2}`}delta(e){return e===0?this:new Mf(this.offset1+e,this.offset2+e)}equals(e){return this.offset1===e.offset1&&this.offset2===e.offset2}};Mf.zero=new Mf(0,0),Mf.max=new Mf(Number.MAX_SAFE_INTEGER,Number.MAX_SAFE_INTEGER);let qd=Mf;const lN=class lN{isValid(){return!0}};lN.instance=new lN;let OS=lN;class dke{constructor(e){if(this.timeout=e,this.startTime=Date.now(),this.valid=!0,e<=0)throw new Ce("timeout must be positive")}isValid(){return!(Date.now()-this.startTime<this.timeout)&&this.valid&&(this.valid=!1),this.valid}}class FM{constructor(e,t){this.width=e,this.height=t,this.array=[],this.array=new Array(e*t)}get(e,t){return this.array[e+t*this.width]}set(e,t,i){this.array[e+t*this.width]=i}}function f4(s){return s===32||s===9}const ky=class ky{static getKey(e){let t=this.chrKeys.get(e);return t===void 0&&(t=this.chrKeys.size,this.chrKeys.set(e,t)),t}constructor(e,t,i){this.range=e,this.lines=t,this.source=i,this.histogram=[];let n=0;for(let o=e.startLineNumber-1;o<e.endLineNumberExclusive-1;o++){const r=t[o];for(let l=0;l<r.length;l++){n++;const c=r[l],d=ky.getKey(c);this.histogram[d]=(this.histogram[d]||0)+1}n++;const a=ky.getKey(`
`);this.histogram[a]=(this.histogram[a]||0)+1}this.totalCount=n}computeSimilarity(e){let t=0;const i=Math.max(this.histogram.length,e.histogram.length);for(let n=0;n<i;n++)t+=Math.abs((this.histogram[n]??0)-(e.histogram[n]??0));return 1-t/(this.totalCount+e.totalCount)}};ky.chrKeys=new Map;let LI=ky;function wie(s,e,t,i){return e||(e=n=>n!=null),new Promise((n,o)=>{let r=!0,a=!1;const l=s.map(d=>({isFinished:e(d),error:t?t(d):!1,state:d})),c=De(d=>{const{isFinished:h,error:u,state:g}=l.read(d);(h||u)&&(r?a=!0:c.dispose(),u?o(u===!0?g:u):n(g))});if(i){const d=i.onCancellationRequested(()=>{c.dispose(),d.dispose(),o(new ia)});if(i.isCancellationRequested){c.dispose(),d.dispose(),o(new ia);return}}r=!1,a&&c.dispose()})}function Po(s,e,t=dn.ofCaller()){return new hke(typeof s=="string"?s:new os(s,void 0,void 0),e,t)}class hke extends qv{constructor(e,t,i){super(i),this.event=t,this.handleEvent=()=>{Vt(n=>{for(const o of this._observers)n.updateObserver(o,this),o.handleChange(this,void 0)},()=>this.debugName)},this.debugName=typeof e=="string"?e:e.getDebugName(this)??"Observable Signal From Event"}onFirstObserverAdded(){this.subscription=this.event(this.handleEvent)}onLastObserverRemoved(){this.subscription.dispose(),this.subscription=void 0}get(){}}var uke=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},gke=function(s,e){return function(t,i){e(t,i,s)}};let p4=class extends F{setActiveMovedText(e){this._activeMovedText.set(e,void 0)}constructor(e,t,i){super(),this.model=e,this._options=t,this._diffProviderFactoryService=i,this._isDiffUpToDate=Re(this,!1),this.isDiffUpToDate=this._isDiffUpToDate,this._diff=Re(this,void 0),this.diff=this._diff,this._unchangedRegions=Re(this,void 0),this.unchangedRegions=j(this,a=>this._options.hideUnchangedRegions.read(a)?this._unchangedRegions.read(a)?.regions??[]:(Vt(l=>{for(const c of this._unchangedRegions.read(void 0)?.regions||[])c.collapseAll(l)}),[])),this.movedTextToCompare=Re(this,void 0),this._activeMovedText=Re(this,void 0),this._hoveredMovedText=Re(this,void 0),this.activeMovedText=j(this,a=>this.movedTextToCompare.read(a)??this._hoveredMovedText.read(a)??this._activeMovedText.read(a)),this._cancellationTokenSource=new si,this._diffProvider=j(this,a=>{const l=this._diffProviderFactoryService.createDiffProvider({diffAlgorithm:this._options.diffAlgorithm.read(a)}),c=Po("onDidChange",l.onDidChange);return{diffProvider:l,onChangeSignal:c}}),this._register(pe(()=>this._cancellationTokenSource.cancel()));const n=Yr("contentChangedSignal"),o=this._register(new Rt(()=>n.trigger(void 0),200));this._register(De(a=>{const l=this._unchangedRegions.read(a);if(!l||l.regions.some(f=>f.isDragged.read(a)))return;const c=l.originalDecorationIds.map(f=>e.original.getDecorationRange(f)).map(f=>f?Ne.fromRangeInclusive(f):void 0),d=l.modifiedDecorationIds.map(f=>e.modified.getDecorationRange(f)).map(f=>f?Ne.fromRangeInclusive(f):void 0),h=l.regions.map((f,p)=>!c[p]||!d[p]?void 0:new xu(c[p].startLineNumber,d[p].startLineNumber,c[p].length,f.visibleLineCountTop.read(a),f.visibleLineCountBottom.read(a))).filter(Ln),u=[];let g=!1;for(const f of eW(h,(p,_)=>p.getHiddenModifiedRange(a).endLineNumberExclusive===_.getHiddenModifiedRange(a).startLineNumber))if(f.length>1){g=!0;const p=f.reduce((b,w)=>b+w.lineCount,0),_=new xu(f[0].originalLineNumber,f[0].modifiedLineNumber,p,f[0].visibleLineCountTop.read(void 0),f[f.length-1].visibleLineCountBottom.read(void 0));u.push(_)}else u.push(f[0]);if(g){const f=e.original.deltaDecorations(l.originalDecorationIds,u.map(_=>({range:_.originalUnchangedRange.toInclusiveRange(),options:{description:"unchanged"}}))),p=e.modified.deltaDecorations(l.modifiedDecorationIds,u.map(_=>({range:_.modifiedUnchangedRange.toInclusiveRange(),options:{description:"unchanged"}})));Vt(_=>{this._unchangedRegions.set({regions:u,originalDecorationIds:f,modifiedDecorationIds:p},_)})}}));const r=(a,l,c)=>{const d=xu.fromDiffs(a.changes,e.original.getLineCount(),e.modified.getLineCount(),this._options.hideUnchangedRegionsMinimumLineCount.read(c),this._options.hideUnchangedRegionsContextLineCount.read(c));let h;const u=this._unchangedRegions.get();if(u){const _=u.originalDecorationIds.map(v=>e.original.getDecorationRange(v)).map(v=>v?Ne.fromRangeInclusive(v):void 0),b=u.modifiedDecorationIds.map(v=>e.modified.getDecorationRange(v)).map(v=>v?Ne.fromRangeInclusive(v):void 0);let C=VLe(u.regions.map((v,y)=>{if(!_[y]||!b[y])return;const x=_[y].length;return new xu(_[y].startLineNumber,b[y].startLineNumber,x,Math.min(v.visibleLineCountTop.get(),x),Math.min(v.visibleLineCountBottom.get(),x-v.visibleLineCountTop.get()))}).filter(Ln),(v,y)=>!y||v.modifiedLineNumber>=y.modifiedLineNumber+y.lineCount&&v.originalLineNumber>=y.originalLineNumber+y.lineCount).map(v=>new Ts(v.getHiddenOriginalRange(c),v.getHiddenModifiedRange(c)));C=Ts.clip(C,Ne.ofLength(1,e.original.getLineCount()),Ne.ofLength(1,e.modified.getLineCount())),h=Ts.inverse(C,e.original.getLineCount(),e.modified.getLineCount())}const g=[];if(h)for(const _ of d){const b=h.filter(w=>w.original.intersectsStrict(_.originalUnchangedRange)&&w.modified.intersectsStrict(_.modifiedUnchangedRange));g.push(..._.setVisibleRanges(b,l))}else g.push(...d);const f=e.original.deltaDecorations(u?.originalDecorationIds||[],g.map(_=>({range:_.originalUnchangedRange.toInclusiveRange(),options:{description:"unchanged"}}))),p=e.modified.deltaDecorations(u?.modifiedDecorationIds||[],g.map(_=>({range:_.modifiedUnchangedRange.toInclusiveRange(),options:{description:"unchanged"}})));this._unchangedRegions.set({regions:g,originalDecorationIds:f,modifiedDecorationIds:p},l)};this._register(e.modified.onDidChangeContent(a=>{if(this._diff.get()){const c=Ud.fromModelContentChanges(a.changes);this._lastDiff,e.original,e.modified}this._isDiffUpToDate.set(!1,void 0),o.schedule()})),this._register(e.original.onDidChangeContent(a=>{if(this._diff.get()){const c=Ud.fromModelContentChanges(a.changes);this._lastDiff,e.original,e.modified}this._isDiffUpToDate.set(!1,void 0),o.schedule()})),this._register(ps(async(a,l)=>{this._options.hideUnchangedRegionsMinimumLineCount.read(a),this._options.hideUnchangedRegionsContextLineCount.read(a),o.cancel(),n.read(a);const c=this._diffProvider.read(a);c.onChangeSignal.read(a),this._isDiffUpToDate.set(!1,void 0);let d=[];l.add(e.original.onDidChangeContent(g=>{const f=Ud.fromModelContentChanges(g.changes);d=uI(d,f)}));let h=[];l.add(e.modified.onDidChangeContent(g=>{const f=Ud.fromModelContentChanges(g.changes);h=uI(h,f)}));let u=await c.diffProvider.computeDiff(e.original,e.modified,{ignoreTrimWhitespace:this._options.ignoreTrimWhitespace.read(a),maxComputationTimeMs:this._options.maxComputationTimeMs.read(a),computeMoves:this._options.showMoves.read(a)},this._cancellationTokenSource.token);this._cancellationTokenSource.token.isCancellationRequested||e.original.isDisposed()||e.modified.isDisposed()||(u=fke(u,e.original,e.modified),u=(e.original,e.modified,void 0)??u,u=(e.original,e.modified,void 0)??u,Vt(g=>{r(u,g),this._lastDiff=u;const f=F7.fromDiffResult(u);this._diff.set(f,g),this._isDiffUpToDate.set(!0,g);const p=this.movedTextToCompare.read(void 0);this.movedTextToCompare.set(p?this._lastDiff.moves.find(_=>_.lineRangeMapping.modified.intersect(p.lineRangeMapping.modified)):void 0,g)}))}))}ensureModifiedLineIsVisible(e,t,i){if(this.diff.get()?.mappings.length===0)return;const n=this._unchangedRegions.get()?.regions||[];for(const o of n)if(o.getHiddenModifiedRange(void 0).contains(e)){o.showModifiedLine(e,t,i);return}}ensureOriginalLineIsVisible(e,t,i){if(this.diff.get()?.mappings.length===0)return;const n=this._unchangedRegions.get()?.regions||[];for(const o of n)if(o.getHiddenOriginalRange(void 0).contains(e)){o.showOriginalLine(e,t,i);return}}async waitForDiff(){await wie(this.isDiffUpToDate,e=>e)}serializeState(){return{collapsedRegions:this._unchangedRegions.get()?.regions.map(t=>({range:t.getHiddenModifiedRange(void 0).serialize()}))}}restoreSerializedState(e){const t=e.collapsedRegions?.map(n=>Ne.deserialize(n.range)),i=this._unchangedRegions.get();!i||!t||Vt(n=>{for(const o of i.regions)for(const r of t)if(o.modifiedUnchangedRange.intersect(r)){o.setHiddenModifiedRange(r,n);break}})}};p4=uke([gke(2,bie)],p4);function fke(s,e,t){return{changes:s.changes.map(i=>new Lr(i.original,i.modified,i.innerChanges?i.innerChanges.map(n=>pke(n,e,t)):void 0)),moves:s.moves,identical:s.identical,quitEarly:s.quitEarly}}function pke(s,e,t){let i=s.originalRange,n=s.modifiedRange;return i.startColumn===1&&n.startColumn===1&&(i.endColumn!==1||n.endColumn!==1)&&i.endColumn===e.getLineMaxColumn(i.endLineNumber)&&n.endColumn===t.getLineMaxColumn(n.endLineNumber)&&i.endLineNumber<e.getLineCount()&&n.endLineNumber<t.getLineCount()&&(i=i.setEndPosition(i.endLineNumber+1,1),n=n.setEndPosition(n.endLineNumber+1,1)),new Ks(i,n)}class F7{static fromDiffResult(e){return new F7(e.changes.map(t=>new vie(t)),e.moves||[],e.identical,e.quitEarly)}constructor(e,t,i,n){this.mappings=e,this.movedTexts=t,this.identical=i,this.quitEarly=n}}class vie{constructor(e){this.lineRangeMapping=e}}class xu{static fromDiffs(e,t,i,n,o){const r=Lr.inverse(e,t,i),a=[];for(const l of r){let c=l.original.startLineNumber,d=l.modified.startLineNumber,h=l.original.length;const u=c===1&&d===1,g=c+h===t+1&&d+h===i+1;(u||g)&&h>=o+n?(u&&!g&&(h-=o),g&&!u&&(c+=o,d+=o,h-=o),a.push(new xu(c,d,h,0,0))):h>=o*2+n&&(c+=o,d+=o,h-=o*2,a.push(new xu(c,d,h,0,0)))}return a}get originalUnchangedRange(){return Ne.ofLength(this.originalLineNumber,this.lineCount)}get modifiedUnchangedRange(){return Ne.ofLength(this.modifiedLineNumber,this.lineCount)}constructor(e,t,i,n,o){this.originalLineNumber=e,this.modifiedLineNumber=t,this.lineCount=i,this._visibleLineCountTop=Re(this,0),this.visibleLineCountTop=this._visibleLineCountTop,this._visibleLineCountBottom=Re(this,0),this.visibleLineCountBottom=this._visibleLineCountBottom,this._shouldHideControls=j(this,l=>this.visibleLineCountTop.read(l)+this.visibleLineCountBottom.read(l)===this.lineCount&&!this.isDragged.read(l)),this.isDragged=Re(this,void 0);const r=Math.max(Math.min(n,this.lineCount),0),a=Math.max(Math.min(o,this.lineCount-n),0);yz(n===r),yz(o===a),this._visibleLineCountTop.set(r,void 0),this._visibleLineCountBottom.set(a,void 0)}setVisibleRanges(e,t){const i=[],n=new Kr(e.map(l=>l.modified)).subtractFrom(this.modifiedUnchangedRange);let o=this.originalLineNumber,r=this.modifiedLineNumber;const a=this.modifiedLineNumber+this.lineCount;if(n.ranges.length===0)this.showAll(t),i.push(this);else{let l=0;for(const c of n.ranges){const d=l===n.ranges.length-1;l++;const h=(d?a:c.endLineNumberExclusive)-r,u=new xu(o,r,h,0,0);u.setHiddenModifiedRange(c,t),i.push(u),o=u.originalUnchangedRange.endLineNumberExclusive,r=u.modifiedUnchangedRange.endLineNumberExclusive}}return i}shouldHideControls(e){return this._shouldHideControls.read(e)}getHiddenOriginalRange(e){return Ne.ofLength(this.originalLineNumber+this._visibleLineCountTop.read(e),this.lineCount-this._visibleLineCountTop.read(e)-this._visibleLineCountBottom.read(e))}getHiddenModifiedRange(e){return Ne.ofLength(this.modifiedLineNumber+this._visibleLineCountTop.read(e),this.lineCount-this._visibleLineCountTop.read(e)-this._visibleLineCountBottom.read(e))}setHiddenModifiedRange(e,t){const i=e.startLineNumber-this.modifiedLineNumber,n=this.modifiedLineNumber+this.lineCount-e.endLineNumberExclusive;this.setState(i,n,t)}getMaxVisibleLineCountTop(){return this.lineCount-this._visibleLineCountBottom.get()}getMaxVisibleLineCountBottom(){return this.lineCount-this._visibleLineCountTop.get()}showMoreAbove(e=10,t){const i=this.getMaxVisibleLineCountTop();this._visibleLineCountTop.set(Math.min(this._visibleLineCountTop.get()+e,i),t)}showMoreBelow(e=10,t){const i=this.lineCount-this._visibleLineCountTop.get();this._visibleLineCountBottom.set(Math.min(this._visibleLineCountBottom.get()+e,i),t)}showAll(e){this._visibleLineCountBottom.set(this.lineCount-this._visibleLineCountTop.get(),e)}showModifiedLine(e,t,i){const n=e+1-(this.modifiedLineNumber+this._visibleLineCountTop.get()),o=this.modifiedLineNumber-this._visibleLineCountBottom.get()+this.lineCount-e;t===0&&n<o||t===1?this._visibleLineCountTop.set(this._visibleLineCountTop.get()+n,i):this._visibleLineCountBottom.set(this._visibleLineCountBottom.get()+o,i)}showOriginalLine(e,t,i){const n=e-this.originalLineNumber,o=this.originalLineNumber+this.lineCount-e;t===0&&n<o||t===1?this._visibleLineCountTop.set(Math.min(this._visibleLineCountTop.get()+o-n,this.getMaxVisibleLineCountTop()),i):this._visibleLineCountBottom.set(Math.min(this._visibleLineCountBottom.get()+n-o,this.getMaxVisibleLineCountBottom()),i)}collapseAll(e){this._visibleLineCountTop.set(0,e),this._visibleLineCountBottom.set(0,e)}setState(e,t,i){e=Math.max(Math.min(e,this.lineCount),0),t=Math.max(Math.min(t,this.lineCount-e),0),this._visibleLineCountTop.set(e,i),this._visibleLineCountBottom.set(t,i)}}function mke(s){const{domNode:e,renderLinesResult:t,diffEntry:i,originalModel:n,clipboardService:o}=s,r=new U;return r.add(W(e,"copy",a=>{a.preventDefault();const l=e.ownerDocument.getSelection();if(!l||l.rangeCount===0)return;const c=l.getRangeAt(0);if(!c||c.collapsed)return;const d=c.startContainer.nodeType===Node.TEXT_NODE?c.startContainer.parentElement:c.startContainer,h=c.endContainer.nodeType===Node.TEXT_NODE?c.endContainer.parentElement:c.endContainer;if(!d||!h)return;const u=t.getModelPositionAt(d,c.startOffset),g=t.getModelPositionAt(h,c.endOffset);if(!u||!g)return;const f=u.delta(i.original.startLineNumber-1),p=g.delta(i.original.startLineNumber-1),_=p.isBefore(f)?k.fromPositions(p,f):k.fromPositions(f,p),b=n.getValueInRange(_);o.writeText(b)})),r}class _ke extends F{get visibility(){return this._visibility}set visibility(e){this._visibility!==e&&(this._visibility=e,this._diffActions.style.visibility=e?"visible":"hidden")}constructor(e,t,i,n,o,r,a,l,c,d){super(),this._getViewZoneId=e,this._marginDomNode=t,this._deletedCodeDomNode=i,this._modifiedEditor=n,this._diff=o,this._editor=r,this._renderLinesResult=a,this._originalTextModel=l,this._contextMenuService=c,this._clipboardService=d,this._visibility=!1,this._marginDomNode.style.zIndex="10",this._diffActions=document.createElement("div"),this._diffActions.className=xe.asClassName(Z.lightBulb)+" lightbulb-glyph",this._diffActions.style.position="absolute";const h=this._modifiedEditor.getOption(75);this._diffActions.style.right="0px",this._diffActions.style.visibility="hidden",this._diffActions.style.height=`${h}px`,this._diffActions.style.lineHeight=`${h}px`,this._marginDomNode.appendChild(this._diffActions);let u=0;const g=n.getOption(144)&&!na,f=(p,_,b)=>{this._contextMenuService.showContextMenu({domForShadowRoot:g?n.getDomNode()??void 0:void 0,getAnchor:()=>p,onHide:b,getActions:()=>{const w=[],C=o.modified.isEmpty;return w.push(new Cr("diff.clipboard.copyDeletedContent",C?o.original.length>1?m(112,"Copy deleted lines"):m(113,"Copy deleted line"):o.original.length>1?m(114,"Copy changed lines"):m(115,"Copy changed line"),void 0,!0,async()=>{const y=this._originalTextModel.getValueInRange(o.original.toExclusiveRange());await this._clipboardService.writeText(y)})),o.original.length>1&&w.push(new Cr("diff.clipboard.copyDeletedLineContent",C?m(116,"Copy deleted line ({0})",o.original.startLineNumber+u):m(117,"Copy changed line ({0})",o.original.startLineNumber+u),void 0,!0,async()=>{let y=this._originalTextModel.getLineContent(o.original.startLineNumber+u);y===""&&(y=this._originalTextModel.getEndOfLineSequence()===0?`
`:`\r
`),await this._clipboardService.writeText(y)})),n.getOption(104)||w.push(new Cr("diff.inline.revertChange",m(118,"Revert this change"),void 0,!0,async()=>{this._editor.revert(this._diff)})),w},autoSelectFirstItem:!0})};this._register(Fi(this._diffActions,"mousedown",p=>{if(!p.leftButton)return;const{top:_,height:b}=Li(this._diffActions),w=Math.floor(h/3);p.preventDefault(),f({x:p.posx,y:_+b+w})})),this._register(n.onMouseMove(p=>{(p.target.type===8||p.target.type===5)&&p.target.detail.viewZoneId===this._getViewZoneId()?(u=this._updateLightBulbPosition(this._marginDomNode,p.event.browserEvent.y,h),this.visibility=!0):this.visibility=!1})),this._register(mke({domNode:this._deletedCodeDomNode,diffEntry:o,originalModel:this._originalTextModel,renderLinesResult:this._renderLinesResult,clipboardService:d}))}_updateLightBulbPosition(e,t,i){const{top:n}=Li(e),o=t-n,r=Math.floor(o/i),a=r*i;if(this._diffActions.style.top=`${a}px`,this._renderLinesResult.viewLineCounts){let l=0;for(let c=0;c<this._renderLinesResult.viewLineCounts.length;c++)if(l+=this._renderLinesResult.viewLineCounts[c],r<l)return c}return r}}const Uq=$c("diffEditorWidget",{createHTML:s=>s});function hx(s,e,t,i,n=!1){In(i,e.fontInfo);const o=t.length>0,r=new Eg(1e4);let a=0,l=0;const c=[],d=[];for(let f=0;f<s.lineTokens.length;f++){const p=f+1,_=s.lineTokens[f],b=s.lineBreakData[f],w=Rs.filter(t,p,1,Number.MAX_SAFE_INTEGER);if(b){let C=0;for(const v of b.breakOffsets){const y=_.sliceAndInflate(C,v,0),x=jq(l,y,Rs.extractWrapped(w,C,v),o,s.mightContainNonBasicASCII,s.mightContainRTL,e,r,n);a=Math.max(a,x.maxCharWidth),d.push(new qq(x.output.characterMapping,x.output.containsForeignElements,C)),l++,C=v}c.push(b.breakOffsets.length)}else{c.push(1);const C=jq(l,_,w,o,s.mightContainNonBasicASCII,s.mightContainRTL,e,r,n);a=Math.max(a,C.maxCharWidth),d.push(new qq(C.output.characterMapping,C.output.containsForeignElements,0)),l++}}a+=e.scrollBeyondLastColumn;const h=r.build(),u=Uq?Uq.createHTML(h):h;i.innerHTML=u;const g=a*e.typicalHalfwidthCharacterWidth;return new bke(l,g,c,d,s)}class ux{constructor(e,t=e.map(o=>null),i=!0,n=!0){this.lineTokens=e,this.lineBreakData=t,this.mightContainNonBasicASCII=i,this.mightContainRTL=n}}class ah{static fromEditor(e){const t=e.getOptions(),i=t.get(59),n=t.get(165);return new ah(e.getModel()?.getOptions().tabSize||0,i,t.get(40),i.typicalHalfwidthCharacterWidth,t.get(118),t.get(75),n.decorationsWidth,t.get(133),t.get(113),t.get(108),t.get(60),t.get(117).verticalScrollbarSize)}constructor(e,t,i,n,o,r,a,l,c,d,h,u,g=!0){this.tabSize=e,this.fontInfo=t,this.disableMonospaceOptimizations=i,this.typicalHalfwidthCharacterWidth=n,this.scrollBeyondLastColumn=o,this.lineHeight=r,this.lineDecorationsWidth=a,this.stopRenderingLineAfter=l,this.renderWhitespace=c,this.renderControlCharacters=d,this.fontLigatures=h,this.verticalScrollbarSize=u,this.setWidth=g}withSetWidth(e){return new ah(this.tabSize,this.fontInfo,this.disableMonospaceOptimizations,this.typicalHalfwidthCharacterWidth,this.scrollBeyondLastColumn,this.lineHeight,this.lineDecorationsWidth,this.stopRenderingLineAfter,this.renderWhitespace,this.renderControlCharacters,this.fontLigatures,this.verticalScrollbarSize,e)}withScrollBeyondLastColumn(e){return new ah(this.tabSize,this.fontInfo,this.disableMonospaceOptimizations,this.typicalHalfwidthCharacterWidth,e,this.lineHeight,this.lineDecorationsWidth,this.stopRenderingLineAfter,this.renderWhitespace,this.renderControlCharacters,this.fontLigatures,this.verticalScrollbarSize,this.setWidth)}}class bke{constructor(e,t,i,n,o){this.heightInLines=e,this.minWidthInPx=t,this.viewLineCounts=i,this._renderOutputs=n,this._source=o}getModelPositionAt(e,t){let i=e;for(;i&&!i.classList.contains("view-line");)i=i.parentElement;if(!i)return;const n=i.parentElement;if(!n)return;const o=n.querySelectorAll(".view-line");let r=-1;for(let h=0;h<o.length;h++)if(o[h]===i){r=h;break}if(r===-1||r>=this._renderOutputs.length)return;let a=1,l=r;for(let h=0;h<this.viewLineCounts.length;h++){if(l<this.viewLineCounts[h]){a=h+1;break}l-=this.viewLineCounts[h]}if(a>this._source.lineTokens.length)return;const c=this._renderOutputs[r];if(!c)return;const d=dv(c.characterMapping,e,t)+c.offset;return new A(a,d)}}class qq extends KE{constructor(e,t,i){super(e,t),this.offset=i}}function jq(s,e,t,i,n,o,r,a,l){a.appendString('<div class="view-line'),!l&&!i&&a.appendString(" char-delete"),a.appendString('" style="top:'),a.appendString(String(s*r.lineHeight)),r.setWidth?a.appendString('px;width:1000000px;">'):a.appendString('px;">');const c=e.getLineContent(),d=Er.isBasicASCII(c,n),h=Er.containsRTL(c,d,o),u=nC(new Dh(r.fontInfo.isMonospace&&!r.disableMonospaceOptimizations,r.fontInfo.canUseHalfwidthRightwardsArrow,c,!1,d,h,0,e,t,r.tabSize,0,r.fontInfo.spaceWidth,r.fontInfo.middotWidth,r.fontInfo.wsmiddotWidth,r.stopRenderingLineAfter,r.renderWhitespace,r.renderControlCharacters,r.fontLigatures!==mh.OFF,null,null,r.verticalScrollbarSize),a);a.appendString("</div>");const g=u.characterMapping.getHorizontalOffset(u.characterMapping.length);return{output:u,maxCharWidth:g}}const Ko=qe("clipboardService"),Rh=qe("contextViewService"),Nr=qe("contextMenuService");var wke=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},Gq=function(s,e){return function(t,i){e(t,i,s)}};let m4=class extends F{constructor(e,t,i,n,o,r,a,l,c,d){super(),this._targetWindow=e,this._editors=t,this._diffModel=i,this._options=n,this._diffEditorWidget=o,this._canIgnoreViewZoneUpdateEvent=r,this._origViewZonesToIgnore=a,this._modViewZonesToIgnore=l,this._clipboardService=c,this._contextMenuService=d,this._originalTopPadding=Re(this,0),this._originalScrollOffset=Re(this,0),this._originalScrollOffsetAnimated=Rq(this._targetWindow,this._originalScrollOffset,this._store),this._modifiedTopPadding=Re(this,0),this._modifiedScrollOffset=Re(this,0),this._modifiedScrollOffsetAnimated=Rq(this._targetWindow,this._modifiedScrollOffset,this._store);const h=Re("invalidateAlignmentsState",0),u=this._register(new Rt(()=>{h.set(h.get()+1,void 0)},0));this._register(this._editors.original.onDidChangeViewZones(C=>{this._canIgnoreViewZoneUpdateEvent()||u.schedule()})),this._register(this._editors.modified.onDidChangeViewZones(C=>{this._canIgnoreViewZoneUpdateEvent()||u.schedule()})),this._register(this._editors.original.onDidChangeConfiguration(C=>{(C.hasChanged(166)||C.hasChanged(75))&&u.schedule()})),this._register(this._editors.modified.onDidChangeConfiguration(C=>{(C.hasChanged(166)||C.hasChanged(75))&&u.schedule()}));const g=this._diffModel.map(C=>C?bt(this,C.model.original.onDidChangeTokens,()=>C.model.original.tokenization.backgroundTokenizationState===2):void 0).map((C,v)=>C?.read(v)),f=j(C=>{const v=this._diffModel.read(C),y=v?.diff.read(C);if(!v||!y)return null;h.read(C);const L=this._options.renderSideBySide.read(C);return Kq(this._editors.original,this._editors.modified,y.mappings,this._origViewZonesToIgnore,this._modViewZonesToIgnore,L)}),p=j(C=>{const v=this._diffModel.read(C)?.movedTextToCompare.read(C);if(!v)return null;h.read(C);const y=v.changes.map(x=>new vie(x));return Kq(this._editors.original,this._editors.modified,y,this._origViewZonesToIgnore,this._modViewZonesToIgnore,!0)});function _(){const C=document.createElement("div");return C.className="diagonal-fill",C}const b=this._register(new U);this.viewZones=j(this,C=>{b.clear();const v=f.read(C)||[],y=[],x=[],L=this._modifiedTopPadding.read(C);L>0&&x.push({afterLineNumber:0,domNode:document.createElement("div"),heightInPx:L,showInHiddenAreas:!0,suppressMouseDown:!0});const E=this._originalTopPadding.read(C);E>0&&y.push({afterLineNumber:0,domNode:document.createElement("div"),heightInPx:E,showInHiddenAreas:!0,suppressMouseDown:!0});const I=this._options.renderSideBySide.read(C),R=I?void 0:this._editors.modified._getViewModel()?.createLineBreaksComputer();if(R){const ne=this._editors.original.getModel();for(const de of v)if(de.diff)for(let ke=de.originalRange.startLineNumber;ke<de.originalRange.endLineNumberExclusive;ke++){if(ke>ne.getLineCount())return{orig:y,mod:x};R?.addRequest(ne.getLineContent(ke),null,null)}}const O=R?.finalize()??[];let P=0;const B=this._editors.modified.getOption(75),$=this._diffModel.read(C)?.movedTextToCompare.read(C),z=this._editors.original.getModel()?.mightContainNonBasicASCII()??!1,Q=this._editors.original.getModel()?.mightContainRTL()??!1,J=ah.fromEditor(this._editors.modified);for(const ne of v)if(ne.diff&&!I&&(!this._options.useTrueInlineDiffRendering.read(C)||!B7(ne.diff))){if(!ne.originalRange.isEmpty){g.read(C);const ke=document.createElement("div");ke.classList.add("view-lines","line-delete","line-delete-selectable","monaco-mouse-cursor-text");const Me=this._editors.original.getModel();if(ne.originalRange.endLineNumberExclusive-1>Me.getLineCount())return{orig:y,mod:x};const Ze=new ux(ne.originalRange.mapToLineArray(Vi=>Me.tokenization.getLineTokens(Vi)),ne.originalRange.mapToLineArray(Vi=>O[P++]),z,Q),Xe=[];for(const Vi of ne.diff.innerChanges||[])Xe.push(new fp(Vi.originalRange.delta(-(ne.diff.original.startLineNumber-1)),d4.className,0));const At=hx(Ze,J,Xe,ke),Et=document.createElement("div");if(Et.className="inline-deleted-margin-view-zone",In(Et,J.fontInfo),this._options.renderIndicators.read(C))for(let Vi=0;Vi<At.heightInLines;Vi++){const Ws=document.createElement("div");Ws.className=`delete-sign ${xe.asClassName(mie)}`,Ws.setAttribute("style",`position:absolute;top:${Vi*B}px;width:${J.lineDecorationsWidth}px;height:${B}px;right:0;`),Et.appendChild(Ws)}let un;b.add(new _ke(()=>gu(un),Et,ke,this._editors.modified,ne.diff,this._diffEditorWidget,At,this._editors.original.getModel(),this._contextMenuService,this._clipboardService));for(let Vi=0;Vi<At.viewLineCounts.length;Vi++){const Ws=At.viewLineCounts[Vi];Ws>1&&y.push({afterLineNumber:ne.originalRange.startLineNumber+Vi,domNode:_(),heightInPx:(Ws-1)*B,showInHiddenAreas:!0,suppressMouseDown:!0})}x.push({afterLineNumber:ne.modifiedRange.startLineNumber-1,domNode:ke,heightInPx:At.heightInLines*B,minWidthInPx:At.minWidthInPx,marginDomNode:Et,setZoneId(Vi){un=Vi},showInHiddenAreas:!0,suppressMouseDown:!1})}const de=document.createElement("div");de.className="gutter-delete",y.push({afterLineNumber:ne.originalRange.endLineNumberExclusive-1,domNode:_(),heightInPx:ne.modifiedHeightInPx,marginDomNode:de,showInHiddenAreas:!0,suppressMouseDown:!0})}else{const de=ne.modifiedHeightInPx-ne.originalHeightInPx;if(de>0){if($?.lineRangeMapping.original.delta(-1).deltaLength(2).contains(ne.originalRange.endLineNumberExclusive-1))continue;y.push({afterLineNumber:ne.originalRange.endLineNumberExclusive-1,domNode:_(),heightInPx:de,showInHiddenAreas:!0,suppressMouseDown:!0})}else{let ke=function(){const Ze=document.createElement("div");return Ze.className="arrow-revert-change "+xe.asClassName(Z.arrowRight),C.store.add(W(Ze,"mousedown",Xe=>Xe.stopPropagation())),C.store.add(W(Ze,"click",Xe=>{Xe.stopPropagation(),o.revert(ne.diff)})),ie("div",{},Ze)};if($?.lineRangeMapping.modified.delta(-1).deltaLength(2).contains(ne.modifiedRange.endLineNumberExclusive-1))continue;let Me;ne.diff&&ne.diff.modified.isEmpty&&this._options.shouldRenderOldRevertArrows.read(C)&&(Me=ke()),x.push({afterLineNumber:ne.modifiedRange.endLineNumberExclusive-1,domNode:_(),heightInPx:-de,marginDomNode:Me,showInHiddenAreas:!0,suppressMouseDown:!0})}}for(const ne of p.read(C)??[]){if(!$?.lineRangeMapping.original.intersect(ne.originalRange)||!$?.lineRangeMapping.modified.intersect(ne.modifiedRange))continue;const de=ne.modifiedHeightInPx-ne.originalHeightInPx;de>0?y.push({afterLineNumber:ne.originalRange.endLineNumberExclusive-1,domNode:_(),heightInPx:de,showInHiddenAreas:!0,suppressMouseDown:!0}):x.push({afterLineNumber:ne.modifiedRange.endLineNumberExclusive-1,domNode:_(),heightInPx:-de,showInHiddenAreas:!0,suppressMouseDown:!0})}return{orig:y,mod:x}});let w=!1;this._register(this._editors.original.onDidScrollChange(C=>{C.scrollLeftChanged&&!w&&(w=!0,this._editors.modified.setScrollLeft(C.scrollLeft),w=!1)})),this._register(this._editors.modified.onDidScrollChange(C=>{C.scrollLeftChanged&&!w&&(w=!0,this._editors.original.setScrollLeft(C.scrollLeft),w=!1)})),this._originalScrollTop=bt(this._editors.original.onDidScrollChange,()=>this._editors.original.getScrollTop()),this._modifiedScrollTop=bt(this._editors.modified.onDidScrollChange,()=>this._editors.modified.getScrollTop()),this._register(De(C=>{const v=this._originalScrollTop.read(C)-(this._originalScrollOffsetAnimated.read(void 0)-this._modifiedScrollOffsetAnimated.read(C))-(this._originalTopPadding.read(void 0)-this._modifiedTopPadding.read(C));v!==this._editors.modified.getScrollTop()&&this._editors.modified.setScrollTop(v,1)})),this._register(De(C=>{const v=this._modifiedScrollTop.read(C)-(this._modifiedScrollOffsetAnimated.read(void 0)-this._originalScrollOffsetAnimated.read(C))-(this._modifiedTopPadding.read(void 0)-this._originalTopPadding.read(C));v!==this._editors.original.getScrollTop()&&this._editors.original.setScrollTop(v,1)})),this._register(De(C=>{const v=this._diffModel.read(C)?.movedTextToCompare.read(C);let y=0;if(v){const x=this._editors.original.getTopForLineNumber(v.lineRangeMapping.original.startLineNumber,!0)-this._originalTopPadding.read(void 0);y=this._editors.modified.getTopForLineNumber(v.lineRangeMapping.modified.startLineNumber,!0)-this._modifiedTopPadding.read(void 0)-x}y>0?(this._modifiedTopPadding.set(0,void 0),this._originalTopPadding.set(y,void 0)):y<0?(this._modifiedTopPadding.set(-y,void 0),this._originalTopPadding.set(0,void 0)):setTimeout(()=>{this._modifiedTopPadding.set(0,void 0),this._originalTopPadding.set(0,void 0)},400),this._editors.modified.hasTextFocus()?this._originalScrollOffset.set(this._modifiedScrollOffset.read(void 0)-y,void 0,!0):this._modifiedScrollOffset.set(this._originalScrollOffset.read(void 0)+y,void 0,!0)}))}};m4=wke([Gq(8,Ko),Gq(9,Nr)],m4);function Kq(s,e,t,i,n,o){const r=new ph(Yq(s,i)),a=new ph(Yq(e,n)),l=s.getOption(75),c=e.getOption(75),d=[];let h=0,u=0;function g(f,p){for(;;){let _=r.peek(),b=a.peek();if(_&&_.lineNumber>=f&&(_=void 0),b&&b.lineNumber>=p&&(b=void 0),!_&&!b)break;const w=_?_.lineNumber-h:Number.MAX_VALUE,C=b?b.lineNumber-u:Number.MAX_VALUE;w<C?(r.dequeue(),b={lineNumber:_.lineNumber-h+u,heightInPx:0}):w>C?(a.dequeue(),_={lineNumber:b.lineNumber-u+h,heightInPx:0}):(r.dequeue(),a.dequeue()),d.push({originalRange:Ne.ofLength(_.lineNumber,1),modifiedRange:Ne.ofLength(b.lineNumber,1),originalHeightInPx:l+_.heightInPx,modifiedHeightInPx:c+b.heightInPx,diff:void 0})}}for(const f of t){let C=function(v,y,x=!1){if(v<w||y<b)return;if(_)_=!1;else if(!x&&(v===w||y===b))return;const L=new Ne(w,v),E=new Ne(b,y);if(L.isEmpty&&E.isEmpty)return;const I=r.takeWhile(O=>O.lineNumber<v)?.reduce((O,P)=>O+P.heightInPx,0)??0,R=a.takeWhile(O=>O.lineNumber<y)?.reduce((O,P)=>O+P.heightInPx,0)??0;d.push({originalRange:L,modifiedRange:E,originalHeightInPx:L.length*l+I,modifiedHeightInPx:E.length*c+R,diff:f.lineRangeMapping}),w=v,b=y};const p=f.lineRangeMapping;g(p.original.startLineNumber,p.modified.startLineNumber);let _=!0,b=p.modified.startLineNumber,w=p.original.startLineNumber;if(o)for(const v of p.innerChanges||[]){v.originalRange.startColumn>1&&v.modifiedRange.startColumn>1&&C(v.originalRange.startLineNumber,v.modifiedRange.startLineNumber);const y=s.getModel(),x=v.originalRange.endLineNumber<=y.getLineCount()?y.getLineMaxColumn(v.originalRange.endLineNumber):Number.MAX_SAFE_INTEGER;v.originalRange.endColumn<x&&C(v.originalRange.endLineNumber,v.modifiedRange.endLineNumber)}C(p.original.endLineNumberExclusive,p.modified.endLineNumberExclusive,!0),h=p.original.endLineNumberExclusive,u=p.modified.endLineNumberExclusive}return g(Number.MAX_VALUE,Number.MAX_VALUE),d}function Yq(s,e){const t=[],i=[],n=s.getOption(166).wrappingColumn!==-1,o=s._getViewModel().coordinatesConverter,r=s.getOption(75);if(n)for(let l=1;l<=s.getModel().getLineCount();l++){const c=o.getModelLineViewLineCount(l);c>1&&i.push({lineNumber:l,heightInPx:r*(c-1)})}for(const l of s.getWhitespaces()){if(e.has(l.id))continue;const c=l.afterLineNumber===0?0:o.convertViewPositionToModelPosition(new A(l.afterLineNumber,1)).lineNumber;t.push({lineNumber:c,heightInPx:l.height})}return PLe(t,i,l=>l.lineNumber,(l,c)=>({lineNumber:l.lineNumber,heightInPx:l.heightInPx+c.heightInPx}))}function B7(s){return s.innerChanges?s.innerChanges.every(e=>kI(e.modifiedRange)&&kI(e.originalRange)||e.originalRange.equalsRange(new k(1,1,1,1))):!1}function kI(s){return s.startLineNumber===s.endLineNumber}const Ey=class Ey extends F{constructor(e,t,i,n,o){super(),this._rootElement=e,this._diffModel=t,this._originalEditorLayoutInfo=i,this._modifiedEditorLayoutInfo=n,this._editors=o,this._originalScrollTop=bt(this,this._editors.original.onDidScrollChange,()=>this._editors.original.getScrollTop()),this._modifiedScrollTop=bt(this,this._editors.modified.onDidScrollChange,()=>this._editors.modified.getScrollTop()),this._viewZonesChanged=Po("onDidChangeViewZones",this._editors.modified.onDidChangeViewZones),this.width=Re(this,0),this._modifiedViewZonesChangedSignal=Po("modified.onDidChangeViewZones",this._editors.modified.onDidChangeViewZones),this._originalViewZonesChangedSignal=Po("original.onDidChangeViewZones",this._editors.original.onDidChangeViewZones),this._state=j(this,d=>{this._element.replaceChildren();const h=this._diffModel.read(d),u=h?.diff.read(d)?.movedTexts;if(!u||u.length===0){this.width.set(0,void 0);return}this._viewZonesChanged.read(d);const g=this._originalEditorLayoutInfo.read(d),f=this._modifiedEditorLayoutInfo.read(d);if(!g||!f){this.width.set(0,void 0);return}this._modifiedViewZonesChangedSignal.read(d),this._originalViewZonesChangedSignal.read(d);const p=u.map(x=>{function L(Q,J){const ne=J.getTopForLineNumber(Q.startLineNumber,!0),de=J.getTopForLineNumber(Q.endLineNumberExclusive,!0);return(ne+de)/2}const E=L(x.lineRangeMapping.original,this._editors.original),I=this._originalScrollTop.read(d),R=L(x.lineRangeMapping.modified,this._editors.modified),O=this._modifiedScrollTop.read(d),P=E-I,B=R-O,$=Math.min(E,R),z=Math.max(E,R);return{range:new me($,z),from:P,to:B,fromWithoutScroll:E,toWithoutScroll:R,move:x}});p.sort(Ihe(ns(x=>x.fromWithoutScroll>x.toWithoutScroll,_Q),ns(x=>x.fromWithoutScroll>x.toWithoutScroll?x.fromWithoutScroll:-x.toWithoutScroll,Wo)));const _=W7.compute(p.map(x=>x.range)),b=10,w=g.verticalScrollbarWidth,C=(_.getTrackCount()-1)*10+b*2,v=w+C+(f.contentLeft-Ey.movedCodeBlockPadding);let y=0;for(const x of p){const L=_.getTrack(y),E=w+b+L*10,I=15,R=15,O=v,P=f.glyphMarginWidth+f.lineNumbersWidth,B=18,$=document.createElementNS("http://www.w3.org/2000/svg","rect");$.classList.add("arrow-rectangle"),$.setAttribute("x",`${O-P}`),$.setAttribute("y",`${x.to-B/2}`),$.setAttribute("width",`${P}`),$.setAttribute("height",`${B}`),this._element.appendChild($);const z=document.createElementNS("http://www.w3.org/2000/svg","g"),Q=document.createElementNS("http://www.w3.org/2000/svg","path");Q.setAttribute("d",`M 0 ${x.from} L ${E} ${x.from} L ${E} ${x.to} L ${O-R} ${x.to}`),Q.setAttribute("fill","none"),z.appendChild(Q);const J=document.createElementNS("http://www.w3.org/2000/svg","polygon");J.classList.add("arrow"),d.store.add(De(ne=>{Q.classList.toggle("currentMove",x.move===h.activeMovedText.read(ne)),J.classList.toggle("currentMove",x.move===h.activeMovedText.read(ne))})),J.setAttribute("points",`${O-R},${x.to-I/2} ${O},${x.to} ${O-R},${x.to+I/2}`),z.appendChild(J),this._element.appendChild(z),y++}this.width.set(C,void 0)}),this._element=document.createElementNS("http://www.w3.org/2000/svg","svg"),this._element.setAttribute("class","moved-blocks-lines"),this._rootElement.appendChild(this._element),this._register(pe(()=>this._element.remove())),this._register(De(d=>{const h=this._originalEditorLayoutInfo.read(d),u=this._modifiedEditorLayoutInfo.read(d);!h||!u||(this._element.style.left=`${h.width-h.verticalScrollbarWidth}px`,this._element.style.height=`${h.height}px`,this._element.style.width=`${h.verticalScrollbarWidth+h.contentLeft-Ey.movedCodeBlockPadding+this.width.read(d)}px`)})),this._register(sv(this._state));const r=j(d=>{const u=this._diffModel.read(d)?.diff.read(d);return u?u.movedTexts.map(g=>({move:g,original:new Rb(zt(g.lineRangeMapping.original.startLineNumber-1),18),modified:new Rb(zt(g.lineRangeMapping.modified.startLineNumber-1),18)})):[]});this._register(yI(this._editors.original,r.map(d=>d.map(h=>h.original)))),this._register(yI(this._editors.modified,r.map(d=>d.map(h=>h.modified)))),this._register(ps((d,h)=>{const u=r.read(d);for(const g of u)h.add(new Zq(this._editors.original,g.original,g.move,"original",this._diffModel.get())),h.add(new Zq(this._editors.modified,g.modified,g.move,"modified",this._diffModel.get()))}));const a=Po("original.onDidFocusEditorWidget",d=>this._editors.original.onDidFocusEditorWidget(()=>setTimeout(()=>d(void 0),0))),l=Po("modified.onDidFocusEditorWidget",d=>this._editors.modified.onDidFocusEditorWidget(()=>setTimeout(()=>d(void 0),0)));let c="modified";this._register(jv({changeTracker:{createChangeSummary:()=>{},handleChange:(d,h)=>(d.didChange(a)&&(c="original"),d.didChange(l)&&(c="modified"),!0)}},d=>{a.read(d),l.read(d);const h=this._diffModel.read(d);if(!h)return;const u=h.diff.read(d);let g;if(u&&c==="original"){const f=this._editors.originalCursor.read(d);f&&(g=u.movedTexts.find(p=>p.lineRangeMapping.original.contains(f.lineNumber)))}if(u&&c==="modified"){const f=this._editors.modifiedCursor.read(d);f&&(g=u.movedTexts.find(p=>p.lineRangeMapping.modified.contains(f.lineNumber)))}g!==h.movedTextToCompare.read(void 0)&&h.movedTextToCompare.set(void 0,void 0),h.setActiveMovedText(g)}))}};Ey.movedCodeBlockPadding=4;let Cw=Ey;class W7{static compute(e){const t=[],i=[];for(const n of e){let o=t.findIndex(r=>!r.intersectsStrict(n));o===-1&&(t.length>=6?o=Hwe(t,ns(a=>a.intersectWithRangeLength(n),Wo)):(o=t.length,t.push(new wW))),t[o].addRange(n),i.push(o)}return new W7(t.length,i)}constructor(e,t){this._trackCount=e,this.trackPerLineIdx=t}getTrack(e){return this.trackPerLineIdx[e]}getTrackCount(){return this._trackCount}}class Zq extends M7{constructor(e,t,i,n,o){const r=ct("div.diff-hidden-lines-widget");super(e,t,r.root),this._editor=e,this._move=i,this._kind=n,this._diffModel=o,this._nodes=ct("div.diff-moved-code-block",{style:{marginRight:"4px"}},[ct("div.text-content@textContent"),ct("div.action-bar@actionBar")]),r.root.appendChild(this._nodes.root);const a=bt(this._editor.onDidLayoutChange,()=>this._editor.getLayoutInfo());this._register(_g(this._nodes.root,{paddingRight:a.map(u=>u.verticalScrollbarWidth)}));let l;i.changes.length>0?l=this._kind==="original"?m(131,"Code moved with changes to line {0}-{1}",this._move.lineRangeMapping.modified.startLineNumber,this._move.lineRangeMapping.modified.endLineNumberExclusive-1):m(132,"Code moved with changes from line {0}-{1}",this._move.lineRangeMapping.original.startLineNumber,this._move.lineRangeMapping.original.endLineNumberExclusive-1):l=this._kind==="original"?m(133,"Code moved to line {0}-{1}",this._move.lineRangeMapping.modified.startLineNumber,this._move.lineRangeMapping.modified.endLineNumberExclusive-1):m(134,"Code moved from line {0}-{1}",this._move.lineRangeMapping.original.startLineNumber,this._move.lineRangeMapping.original.endLineNumberExclusive-1);const c=this._register(new bo(this._nodes.actionBar,{highlightToggledItems:!0})),d=new Cr("",l,"",!1);c.push(d,{icon:!1,label:!0});const h=new Cr("","Compare",xe.asClassName(Z.compareChanges),!0,()=>{this._editor.focus(),this._diffModel.movedTextToCompare.set(this._diffModel.movedTextToCompare.get()===i?void 0:this._move,void 0)});this._register(De(u=>{const g=this._diffModel.movedTextToCompare.read(u)===i;h.checked=g})),c.push(h,{icon:!1,label:!0})}}class vke extends F{constructor(e,t,i,n){super(),this._editors=e,this._diffModel=t,this._options=i,this._decorations=j(this,o=>{const r=this._diffModel.read(o),a=r?.diff.read(o);if(!a)return null;const l=this._diffModel.read(o).movedTextToCompare.read(o),c=this._options.renderIndicators.read(o),d=this._options.showEmptyDecorations.read(o),h=[],u=[];if(!l)for(const f of a.mappings)if(f.lineRangeMapping.original.isEmpty||h.push({range:f.lineRangeMapping.original.toInclusiveRange(),options:c?Hq:zq}),f.lineRangeMapping.modified.isEmpty||u.push({range:f.lineRangeMapping.modified.toInclusiveRange(),options:c?Wq:Vq}),f.lineRangeMapping.modified.isEmpty||f.lineRangeMapping.original.isEmpty)f.lineRangeMapping.original.isEmpty||h.push({range:f.lineRangeMapping.original.toInclusiveRange(),options:lke}),f.lineRangeMapping.modified.isEmpty||u.push({range:f.lineRangeMapping.modified.toInclusiveRange(),options:rke});else{const p=this._options.useTrueInlineDiffRendering.read(o)&&B7(f.lineRangeMapping);for(const _ of f.lineRangeMapping.innerChanges||[])if(f.lineRangeMapping.original.contains(_.originalRange.startLineNumber)&&h.push({range:_.originalRange,options:_.originalRange.isEmpty()&&d?cke:d4}),f.lineRangeMapping.modified.contains(_.modifiedRange.startLineNumber)&&u.push({range:_.modifiedRange,options:_.modifiedRange.isEmpty()&&d&&!p?ake:$q}),p){const b=r.model.original.getValueInRange(_.originalRange);u.push({range:_.modifiedRange,options:{description:"deleted-text",before:{content:b,inlineClassName:"inline-deleted-text"},zIndex:1e5,showIfCollapsed:!0}})}}if(l)for(const f of l.changes){const p=f.original.toInclusiveRange();p&&h.push({range:p,options:c?Hq:zq});const _=f.modified.toInclusiveRange();_&&u.push({range:_,options:c?Wq:Vq});for(const b of f.innerChanges||[])h.push({range:b.originalRange,options:d4}),u.push({range:b.modifiedRange,options:$q})}const g=this._diffModel.read(o).activeMovedText.read(o);for(const f of a.movedTexts)h.push({range:f.lineRangeMapping.original.toInclusiveRange(),options:{description:"moved",blockClassName:"movedOriginal"+(f===g?" currentMove":""),blockPadding:[Cw.movedCodeBlockPadding,0,Cw.movedCodeBlockPadding,Cw.movedCodeBlockPadding]}}),u.push({range:f.lineRangeMapping.modified.toInclusiveRange(),options:{description:"moved",blockClassName:"movedModified"+(f===g?" currentMove":""),blockPadding:[4,0,4,4]}});return{originalDecorations:h,modifiedDecorations:u}}),this._register(CI(this._editors.original,this._decorations.map(o=>o?.originalDecorations||[]))),this._register(CI(this._editors.modified,this._decorations.map(o=>o?.modifiedDecorations||[])))}}class _n{static equals(e,t){return e.x===t.x&&e.y===t.y}constructor(e,t){this.x=e,this.y=t}add(e){return new _n(this.x+e.x,this.y+e.y)}deltaX(e){return new _n(this.x+e,this.y)}deltaY(e){return new _n(this.x,this.y+e)}toString(){return`(${this.x},${this.y})`}subtract(e){return new _n(this.x-e.x,this.y-e.y)}scale(e){return new _n(this.x*e,this.y*e)}mapComponents(e){return new _n(e(this.x),e(this.y))}isZero(){return this.x===0&&this.y===0}withThreshold(e){return this.mapComponents(t=>t>e?t-e:t<-e?t+e:0)}}function di(s){return _4.get(s)}const Qh=class Qh extends F{static get(e){let t=Qh._map.get(e);if(!t){t=new Qh(e),Qh._map.set(e,t);const i=e.onDidDispose(()=>{const n=Qh._map.get(e);n&&(Qh._map.delete(e),n.dispose(),i.dispose())})}return t}_beginUpdate(){this._updateCounter++,this._updateCounter===1&&(this._currentTransaction=new Gv(()=>{}))}_endUpdate(){if(this._updateCounter--,this._updateCounter===0){const e=this._currentTransaction;this._currentTransaction=void 0,e.finish()}}constructor(e){super(),this.editor=e,this._updateCounter=0,this._currentTransaction=void 0,this._model=Re(this,this.editor.getModel()),this.model=this._model,this.isReadonly=bt(this,this.editor.onDidChangeConfiguration,()=>this.editor.getOption(104)),this._versionId=H1({owner:this,lazy:!0},this.editor.getModel()?.getVersionId()??null),this.versionId=this._versionId,this._selections=H1({owner:this,equalsFn:Vf(Gy(he.selectionsEqual)),lazy:!0},this.editor.getSelections()??null),this.selections=this._selections,this.positions=Jn({owner:this,equalsFn:Vf(Gy(A.equals))},t=>this.selections.read(t)?.map(i=>i.getStartPosition())??null),this.isFocused=bt(this,t=>{const i=this.editor.onDidFocusEditorWidget(t),n=this.editor.onDidBlurEditorWidget(t);return{dispose(){i.dispose(),n.dispose()}}},()=>this.editor.hasWidgetFocus()),this.isTextFocused=bt(this,t=>{const i=this.editor.onDidFocusEditorText(t),n=this.editor.onDidBlurEditorText(t);return{dispose(){i.dispose(),n.dispose()}}},()=>this.editor.hasTextFocus()),this.inComposition=bt(this,t=>{const i=this.editor.onDidCompositionStart(()=>{t(void 0)}),n=this.editor.onDidCompositionEnd(()=>{t(void 0)});return{dispose(){i.dispose(),n.dispose()}}},()=>this.editor.inComposition),this.value=oW(this,t=>(this.versionId.read(t),this.model.read(t)?.getValue()??""),(t,i)=>{const n=this.model.get();n!==null&&t!==n.getValue()&&n.setValue(t)}),this.valueIsEmpty=j(this,t=>(this.versionId.read(t),this.editor.getModel()?.getValueLength()===0)),this.cursorSelection=Jn({owner:this,equalsFn:Vf(he.selectionsEqual)},t=>this.selections.read(t)?.[0]??null),this.cursorPosition=Jn({owner:this,equalsFn:A.equals},t=>this.selections.read(t)?.[0]?.getPosition()??null),this.cursorLineNumber=j(this,t=>this.cursorPosition.read(t)?.lineNumber??null),this.onDidType=Yr(this),this.onDidPaste=Yr(this),this.scrollTop=bt(this.editor.onDidScrollChange,()=>this.editor.getScrollTop()),this.scrollLeft=bt(this.editor.onDidScrollChange,()=>this.editor.getScrollLeft()),this.layoutInfo=bt(this.editor.onDidLayoutChange,()=>this.editor.getLayoutInfo()),this.layoutInfoContentLeft=this.layoutInfo.map(t=>t.contentLeft),this.layoutInfoDecorationsLeft=this.layoutInfo.map(t=>t.decorationsLeft),this.layoutInfoWidth=this.layoutInfo.map(t=>t.width),this.layoutInfoHeight=this.layoutInfo.map(t=>t.height),this.layoutInfoMinimap=this.layoutInfo.map(t=>t.minimap),this.layoutInfoVerticalScrollbarWidth=this.layoutInfo.map(t=>t.verticalScrollbarWidth),this.contentWidth=bt(this.editor.onDidContentSizeChange,()=>this.editor.getContentWidth()),this.contentHeight=bt(this.editor.onDidContentSizeChange,()=>this.editor.getContentHeight()),this._widgetCounter=0,this.openedPeekWidgets=Re(this,0),this._register(this.editor.onBeginUpdate(()=>this._beginUpdate())),this._register(this.editor.onEndUpdate(()=>this._endUpdate())),this._register(this.editor.onDidChangeModel(()=>{this._beginUpdate();try{this._model.set(this.editor.getModel(),this._currentTransaction),this._forceUpdate()}finally{this._endUpdate()}})),this._register(this.editor.onDidType(t=>{this._beginUpdate();try{this._forceUpdate(),this.onDidType.trigger(this._currentTransaction,t)}finally{this._endUpdate()}})),this._register(this.editor.onDidPaste(t=>{this._beginUpdate();try{this._forceUpdate(),this.onDidPaste.trigger(this._currentTransaction,t)}finally{this._endUpdate()}})),this._register(this.editor.onDidChangeModelContent(t=>{this._beginUpdate();try{this._versionId.set(this.editor.getModel()?.getVersionId()??null,this._currentTransaction,t),this._forceUpdate()}finally{this._endUpdate()}})),this._register(this.editor.onDidChangeCursorSelection(t=>{this._beginUpdate();try{this._selections.set(this.editor.getSelections(),this._currentTransaction,t),this._forceUpdate()}finally{this._endUpdate()}})),this.domNode=j(t=>(this.model.read(t),this.editor.getDomNode()))}forceUpdate(e){this._beginUpdate();try{return this._forceUpdate(),e?e(this._currentTransaction):void 0}finally{this._endUpdate()}}_forceUpdate(){this._beginUpdate();try{this._model.set(this.editor.getModel(),this._currentTransaction),this._versionId.set(this.editor.getModel()?.getVersionId()??null,this._currentTransaction,void 0),this._selections.set(this.editor.getSelections(),this._currentTransaction,void 0)}finally{this._endUpdate()}}getOption(e,t=dn.ofCaller()){return bt(this,i=>this.editor.onDidChangeConfiguration(n=>{n.hasChanged(e)&&i(void 0)}),()=>this.editor.getOption(e),t)}setDecorations(e){const t=new U,i=this.editor.createDecorationsCollection();return t.add(DR({owner:this,debugName:()=>`Apply decorations from ${e.debugName}`},n=>{const o=e.read(n);i.set(o)})),t.add({dispose:()=>{i.clear()}}),t}createOverlayWidget(e){const t="observableOverlayWidget"+this._widgetCounter++,i={getDomNode:()=>e.domNode,getPosition:()=>e.position.get(),getId:()=>t,allowEditorOverflow:e.allowEditorOverflow,getMinContentWidthInPx:()=>e.minContentWidthInPx.get()};this.editor.addOverlayWidget(i);const n=De(o=>{e.position.read(o),e.minContentWidthInPx.read(o),this.editor.layoutOverlayWidget(i)});return pe(()=>{n.dispose(),this.editor.removeOverlayWidget(i)})}createContentWidget(e){const t="observableContentWidget"+this._widgetCounter++,i={getDomNode:()=>e.domNode,getPosition:()=>e.position.get(),getId:()=>t,allowEditorOverflow:e.allowEditorOverflow};this.editor.addContentWidget(i);const n=De(o=>{e.position.read(o),this.editor.layoutContentWidget(i)});return pe(()=>{n.dispose(),this.editor.removeContentWidget(i)})}observeLineOffsetRange(e,t){const i=this.observePosition(e.map(o=>new A(o.startLineNumber,1)),t),n=this.observePosition(e.map(o=>new A(o.endLineNumberExclusive+1,1)),t);return j(o=>{i.read(o),n.read(o);const r=e.read(o),a=this.model.read(o)?.getLineCount(),l=(typeof a<"u"&&r.startLineNumber>a?this.editor.getBottomForLineNumber(a):this.editor.getTopForLineNumber(r.startLineNumber))-this.scrollTop.read(o),c=r.isEmpty?l:this.editor.getBottomForLineNumber(r.endLineNumberExclusive-1)-this.scrollTop.read(o);return new me(l,c)})}observePosition(e,t){let i=e.get();const n=H1({owner:this,debugName:()=>`topLeftOfPosition${i?.toString()}`,equalsFn:Vf(_n.equals)},new _n(0,0)),o="observablePositionWidget"+this._widgetCounter++,r=document.createElement("div"),a={getDomNode:()=>r,getPosition:()=>i?{preference:[0],position:e.get()}:null,getId:()=>o,allowEditorOverflow:!1,afterRender:(l,c)=>{const d=this._model.get();d&&i&&i.lineNumber>d.getLineCount()?n.set(new _n(0,this.editor.getBottomForLineNumber(d.getLineCount())-this.scrollTop.get()),void 0):n.set(c?new _n(c.left,c.top):null,void 0)}};return this.editor.addContentWidget(a),t.add(De(l=>{i=e.read(l),this.editor.layoutContentWidget(a)})),t.add(pe(()=>{this.editor.removeContentWidget(a)})),n}isTargetHovered(e,t){const i=Re("isInjectedTextHovered",!1);return t.add(this.editor.onMouseMove(n=>{const o=e(n);i.set(o,void 0)})),t.add(this.editor.onMouseLeave(n=>{i.set(!1,void 0)})),i}observeLineHeightForPosition(e){return j(t=>{const i=e instanceof A?e:e.read(t);return i===null?null:(this.getOption(75).read(t),this.editor.getLineHeightForPosition(i))})}observeLineHeightForLine(e){return typeof e=="number"?this.observeLineHeightForPosition(new A(e,1)):j(t=>{const i=e.read(t);return i===null?null:this.observeLineHeightForPosition(new A(i,1)).read(t)})}observeLineHeightsForLineRange(e){return j(t=>{const i=e instanceof Ne?e:e.read(t),n=[];for(let o=i.startLineNumber;o<i.endLineNumberExclusive;o++)n.push(this.observeLineHeightForLine(o).read(t));return n})}};Qh._map=new Map;let _4=Qh;var Cke=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},yke=function(s,e){return function(t,i){e(t,i,s)}},Ka,Zd;let PS=(Zd=class extends F{constructor(e,t,i,n,o,r,a){super(),this._editors=e,this._rootElement=t,this._diffModel=i,this._rootWidth=n,this._rootHeight=o,this._modifiedEditorLayoutInfo=r,this._themeService=a,this.width=Ka.ENTIRE_DIFF_OVERVIEW_WIDTH;const l=bt(this._themeService.onDidColorThemeChange,()=>this._themeService.getColorTheme()),c=j(u=>{const g=l.read(u),f=g.getColor(_me)||(g.getColor(qf)||GP).transparent(2),p=g.getColor(bme)||(g.getColor(bu)||KP).transparent(2);return{insertColor:f,removeColor:p}}),d=wt(document.createElement("div"));d.setClassName("diffViewport"),d.setPosition("absolute");const h=ct("div.diffOverview",{style:{position:"absolute",top:"0px",width:Ka.ENTIRE_DIFF_OVERVIEW_WIDTH+"px"}}).root;this._register(Nb(h,d.domNode)),this._register(Fi(h,se.POINTER_DOWN,u=>{this._editors.modified.delegateVerticalScrollbarPointerDown(u)})),this._register(W(h,se.MOUSE_WHEEL,u=>{this._editors.modified.delegateScrollFromMouseWheelEvent(u)},{passive:!1})),this._register(Nb(this._rootElement,h)),this._register(ps((u,g)=>{const f=this._diffModel.read(u),p=this._editors.original.createOverviewRuler("original diffOverviewRuler");p&&(g.add(p),g.add(Nb(h,p.getDomNode())));const _=this._editors.modified.createOverviewRuler("modified diffOverviewRuler");if(_&&(g.add(_),g.add(Nb(h,_.getDomNode()))),!p||!_)return;const b=Po("viewZoneChanged",this._editors.original.onDidChangeViewZones),w=Po("viewZoneChanged",this._editors.modified.onDidChangeViewZones),C=Po("hiddenRangesChanged",this._editors.original.onDidChangeHiddenAreas),v=Po("hiddenRangesChanged",this._editors.modified.onDidChangeHiddenAreas);g.add(De(y=>{b.read(y),w.read(y),C.read(y),v.read(y);const x=c.read(y),L=f?.diff.read(y)?.mappings;function E(O,P,B){const $=B._getViewModel();return $?O.filter(z=>z.length>0).map(z=>{const Q=$.coordinatesConverter.convertModelPositionToViewPosition(new A(z.startLineNumber,1)),J=$.coordinatesConverter.convertModelPositionToViewPosition(new A(z.endLineNumberExclusive,1)),ne=J.lineNumber-Q.lineNumber;return new Tee(Q.lineNumber,J.lineNumber,ne,P.toString())}):[]}const I=E((L||[]).map(O=>O.lineRangeMapping.original),x.removeColor,this._editors.original),R=E((L||[]).map(O=>O.lineRangeMapping.modified),x.insertColor,this._editors.modified);p?.setZones(I),_?.setZones(R)})),g.add(De(y=>{const x=this._rootHeight.read(y),L=this._rootWidth.read(y),E=this._modifiedEditorLayoutInfo.read(y);if(E){const I=Ka.ENTIRE_DIFF_OVERVIEW_WIDTH-2*Ka.ONE_OVERVIEW_WIDTH;p.setLayout({top:0,height:x,right:I+Ka.ONE_OVERVIEW_WIDTH,width:Ka.ONE_OVERVIEW_WIDTH}),_.setLayout({top:0,height:x,right:0,width:Ka.ONE_OVERVIEW_WIDTH});const R=this._editors.modifiedScrollTop.read(y),O=this._editors.modifiedScrollHeight.read(y),P=this._editors.modified.getOption(117),B=new gv(P.verticalHasArrows?P.arrowSize:0,P.verticalScrollbarSize,0,E.height,O,R);d.setTop(B.getSliderPosition()),d.setHeight(B.getSliderSize())}else d.setTop(0),d.setHeight(0);h.style.height=x+"px",h.style.left=L-Ka.ENTIRE_DIFF_OVERVIEW_WIDTH+"px",d.setWidth(Ka.ENTIRE_DIFF_OVERVIEW_WIDTH)}))}))}},Ka=Zd,Zd.ONE_OVERVIEW_WIDTH=15,Zd.ENTIRE_DIFF_OVERVIEW_WIDTH=Zd.ONE_OVERVIEW_WIDTH*2,Zd);PS=Ka=Cke([yke(6,wi)],PS);var Ske=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},BM=function(s,e){return function(t,i){e(t,i,s)}};let b4=class extends F{get onDidContentSizeChange(){return this._onDidContentSizeChange.event}constructor(e,t,i,n,o,r,a,l){super(),this.originalEditorElement=e,this.modifiedEditorElement=t,this._options=i,this._argCodeEditorWidgetOptions=n,this._createInnerEditor=o,this._contextKeyService=r,this._instantiationService=a,this._keybindingService=l,this.original=this._register(this._createLeftHandSideEditor(this._options.editorOptions.get(),this._argCodeEditorWidgetOptions.originalEditor||{})),this.modified=this._register(this._createRightHandSideEditor(this._options.editorOptions.get(),this._argCodeEditorWidgetOptions.modifiedEditor||{})),this._onDidContentSizeChange=this._register(new M),this.modifiedScrollTop=bt(this,this.modified.onDidScrollChange,()=>this.modified.getScrollTop()),this.modifiedScrollHeight=bt(this,this.modified.onDidScrollChange,()=>this.modified.getScrollHeight()),this.modifiedObs=di(this.modified),this.originalObs=di(this.original),this.modifiedModel=this.modifiedObs.model,this.modifiedSelections=bt(this,this.modified.onDidChangeCursorSelection,()=>this.modified.getSelections()??[]),this.modifiedCursor=Jn({owner:this,equalsFn:A.equals},c=>this.modifiedSelections.read(c)[0]?.getPosition()??new A(1,1)),this.originalCursor=bt(this,this.original.onDidChangeCursorPosition,()=>this.original.getPosition()??new A(1,1)),this.isOriginalFocused=di(this.original).isFocused,this.isModifiedFocused=di(this.modified).isFocused,this.isFocused=j(this,c=>this.isOriginalFocused.read(c)||this.isModifiedFocused.read(c)),this._argCodeEditorWidgetOptions=null,this._register(jv({changeTracker:{createChangeSummary:()=>({}),handleChange:(c,d)=>(c.didChange(i.editorOptions)&&Object.assign(d,c.change.changedOptions),!0)}},(c,d)=>{i.editorOptions.read(c),this._options.renderSideBySide.read(c),this.modified.updateOptions(this._adjustOptionsForRightHandSide(c,d)),this.original.updateOptions(this._adjustOptionsForLeftHandSide(c,d))}))}_createLeftHandSideEditor(e,t){const i=this._adjustOptionsForLeftHandSide(void 0,e),n=this._constructInnerEditor(this._instantiationService,this.originalEditorElement,i,t),o=this._contextKeyService.createKey("isInDiffLeftEditor",n.hasWidgetFocus());return this._register(n.onDidFocusEditorWidget(()=>o.set(!0))),this._register(n.onDidBlurEditorWidget(()=>o.set(!1))),n}_createRightHandSideEditor(e,t){const i=this._adjustOptionsForRightHandSide(void 0,e),n=this._constructInnerEditor(this._instantiationService,this.modifiedEditorElement,i,t),o=this._contextKeyService.createKey("isInDiffRightEditor",n.hasWidgetFocus());return this._register(n.onDidFocusEditorWidget(()=>o.set(!0))),this._register(n.onDidBlurEditorWidget(()=>o.set(!1))),n}_constructInnerEditor(e,t,i,n){const o=this._createInnerEditor(e,t,i,n);return this._register(o.onDidContentSizeChange(r=>{const a=this.original.getContentWidth()+this.modified.getContentWidth()+PS.ENTIRE_DIFF_OVERVIEW_WIDTH,l=Math.max(this.modified.getContentHeight(),this.original.getContentHeight());this._onDidContentSizeChange.fire({contentHeight:l,contentWidth:a,contentHeightChanged:r.contentHeightChanged,contentWidthChanged:r.contentWidthChanged})})),o}_adjustOptionsForLeftHandSide(e,t){const i=this._adjustOptionsForSubEditor(t);return this._options.renderSideBySide.get()?(i.unicodeHighlight=this._options.editorOptions.get().unicodeHighlight||{},i.wordWrapOverride1=this._options.diffWordWrap.get()):(i.wordWrapOverride1="off",i.wordWrapOverride2="off",i.stickyScroll={enabled:!1},i.unicodeHighlight={nonBasicASCII:!1,ambiguousCharacters:!1,invisibleCharacters:!1}),i.glyphMargin=this._options.renderSideBySide.get(),t.originalAriaLabel&&(i.ariaLabel=t.originalAriaLabel),i.ariaLabel=this._updateAriaLabel(i.ariaLabel),i.readOnly=!this._options.originalEditable.get(),i.dropIntoEditor={enabled:!i.readOnly},i.extraEditorClassName="original-in-monaco-diff-editor",i}_adjustOptionsForRightHandSide(e,t){const i=this._adjustOptionsForSubEditor(t);return t.modifiedAriaLabel&&(i.ariaLabel=t.modifiedAriaLabel),i.ariaLabel=this._updateAriaLabel(i.ariaLabel),i.wordWrapOverride1=this._options.diffWordWrap.get(),i.revealHorizontalRightPadding=Is.revealHorizontalRightPadding.defaultValue+PS.ENTIRE_DIFF_OVERVIEW_WIDTH,i.scrollbar.verticalHasArrows=!1,i.extraEditorClassName="modified-in-monaco-diff-editor",i}_adjustOptionsForSubEditor(e){const t={...e,dimension:{height:0,width:0}};return t.inDiffEditor=!0,t.automaticLayout=!1,t.allowVariableLineHeights=!1,t.allowVariableFonts=!1,t.allowVariableFontsInAccessibilityMode=!1,t.scrollbar={...t.scrollbar||{}},t.folding=!1,t.codeLens=this._options.diffCodeLens.get(),t.fixedOverflowWidgets=!0,t.minimap={...t.minimap||{}},t.minimap.enabled=!1,this._options.hideUnchangedRegions.get()?t.stickyScroll={enabled:!1}:t.stickyScroll=this._options.editorOptions.get().stickyScroll,t}_updateAriaLabel(e){e||(e="");const t=m(111," use {0} to open the accessibility help.",this._keybindingService.lookupKeybinding("editor.action.accessibilityHelp")?.getAriaLabel());return this._options.accessibilityVerbose.get()?e+t:e?e.replaceAll(t,""):""}};b4=Ske([BM(5,Ae),BM(6,_e),BM(7,ft)],b4);var aC=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o};const xke=!1;var EI;(function(s){s.North="north",s.South="south",s.East="east",s.West="west"})(EI||(EI={}));let Lke=4;const kke=new M;let Eke=300;const Ike=new M;class H7{constructor(e){this.el=e,this.disposables=new U}get onPointerMove(){return this.disposables.add(new Lt(be(this.el),"mousemove")).event}get onPointerUp(){return this.disposables.add(new Lt(be(this.el),"mouseup")).event}dispose(){this.disposables.dispose()}}aC([Ai],H7.prototype,"onPointerMove",null);aC([Ai],H7.prototype,"onPointerUp",null);class V7{get onPointerMove(){return this.disposables.add(new Lt(this.el,qt.Change)).event}get onPointerUp(){return this.disposables.add(new Lt(this.el,qt.End)).event}constructor(e){this.el=e,this.disposables=new U}dispose(){this.disposables.dispose()}}aC([Ai],V7.prototype,"onPointerMove",null);aC([Ai],V7.prototype,"onPointerUp",null);class II{get onPointerMove(){return this.factory.onPointerMove}get onPointerUp(){return this.factory.onPointerUp}constructor(e){this.factory=e}dispose(){}}aC([Ai],II.prototype,"onPointerMove",null);aC([Ai],II.prototype,"onPointerUp",null);const Xq="pointer-events-disabled";class ls extends F{get state(){return this._state}get orthogonalStartSash(){return this._orthogonalStartSash}get orthogonalEndSash(){return this._orthogonalEndSash}set state(e){this._state!==e&&(this.el.classList.toggle("disabled",e===0),this.el.classList.toggle("minimum",e===1),this.el.classList.toggle("maximum",e===2),this._state=e,this.onDidEnablementChange.fire(e))}get onDidStart(){return this._onDidStart.event}get onDidChange(){return this._onDidChange.event}get onDidReset(){return this._onDidReset.event}get onDidEnd(){return this._onDidEnd.event}set orthogonalStartSash(e){if(this._orthogonalStartSash!==e){if(this.orthogonalStartDragHandleDisposables.clear(),this.orthogonalStartSashDisposables.clear(),e){const t=i=>{this.orthogonalStartDragHandleDisposables.clear(),i!==0&&(this._orthogonalStartDragHandle=X(this.el,ie(".orthogonal-drag-handle.start")),this.orthogonalStartDragHandleDisposables.add(pe(()=>this._orthogonalStartDragHandle.remove())),this.orthogonalStartDragHandleDisposables.add(W(this._orthogonalStartDragHandle,"mouseenter",()=>ls.onMouseEnter(e))),this.orthogonalStartDragHandleDisposables.add(W(this._orthogonalStartDragHandle,"mouseleave",()=>ls.onMouseLeave(e))))};this.orthogonalStartSashDisposables.add(e.onDidEnablementChange.event(t,this)),t(e.state)}this._orthogonalStartSash=e}}set orthogonalEndSash(e){if(this._orthogonalEndSash!==e){if(this.orthogonalEndDragHandleDisposables.clear(),this.orthogonalEndSashDisposables.clear(),e){const t=i=>{this.orthogonalEndDragHandleDisposables.clear(),i!==0&&(this._orthogonalEndDragHandle=X(this.el,ie(".orthogonal-drag-handle.end")),this.orthogonalEndDragHandleDisposables.add(pe(()=>this._orthogonalEndDragHandle.remove())),this.orthogonalEndDragHandleDisposables.add(W(this._orthogonalEndDragHandle,"mouseenter",()=>ls.onMouseEnter(e))),this.orthogonalEndDragHandleDisposables.add(W(this._orthogonalEndDragHandle,"mouseleave",()=>ls.onMouseLeave(e))))};this.orthogonalEndSashDisposables.add(e.onDidEnablementChange.event(t,this)),t(e.state)}this._orthogonalEndSash=e}}constructor(e,t,i){super(),this.hoverDelay=Eke,this.hoverDelayer=this._register(new sa(this.hoverDelay)),this._state=3,this.onDidEnablementChange=this._register(new M),this._onDidStart=this._register(new M),this._onDidChange=this._register(new M),this._onDidReset=this._register(new M),this._onDidEnd=this._register(new M),this.orthogonalStartSashDisposables=this._register(new U),this.orthogonalStartDragHandleDisposables=this._register(new U),this.orthogonalEndSashDisposables=this._register(new U),this.orthogonalEndDragHandleDisposables=this._register(new U),this.linkedSash=void 0,this.el=X(e,ie(".monaco-sash")),i.orthogonalEdge&&this.el.classList.add(`orthogonal-edge-${i.orthogonalEdge}`),Ye&&this.el.classList.add("mac"),this._register(W(this.el,"mousedown",o=>this.onPointerStart(o,new H7(e)))),this._register(W(this.el,"dblclick",o=>this.onPointerDoublePress(o))),this._register(W(this.el,"mouseenter",()=>ls.onMouseEnter(this))),this._register(W(this.el,"mouseleave",()=>ls.onMouseLeave(this))),this._register(_s.addTarget(this.el)),this._register(W(this.el,qt.Start,o=>this.onPointerStart(o,new V7(this.el))));let n;this._register(W(this.el,qt.Tap,o=>{if(n){clearTimeout(n),n=void 0,this.onPointerDoublePress(o);return}clearTimeout(n),n=setTimeout(()=>n=void 0,250)})),typeof i.size=="number"?(this.size=i.size,i.orientation===0?this.el.style.width=`${this.size}px`:this.el.style.height=`${this.size}px`):(this.size=Lke,this._register(kke.event(o=>{this.size=o,this.layout()}))),this._register(Ike.event(o=>this.hoverDelay=o)),this.layoutProvider=t,this.orthogonalStartSash=i.orthogonalStartSash,this.orthogonalEndSash=i.orthogonalEndSash,this.orientation=i.orientation||0,this.orientation===1?(this.el.classList.add("horizontal"),this.el.classList.remove("vertical")):(this.el.classList.remove("horizontal"),this.el.classList.add("vertical")),this.el.classList.toggle("debug",xke),this.layout()}onPointerStart(e,t){gt.stop(e);let i=!1;if(!e.__orthogonalSashEvent){const f=this.getOrthogonalSash(e);f&&(i=!0,e.__orthogonalSashEvent=!0,f.onPointerStart(e,new II(t)))}if(this.linkedSash&&!e.__linkedSashEvent&&(e.__linkedSashEvent=!0,this.linkedSash.onPointerStart(e,new II(t))),!this.state)return;const n=this.el.ownerDocument.getElementsByTagName("iframe");for(const f of n)f.classList.add(Xq);const o=e.pageX,r=e.pageY,a=e.altKey,l={startX:o,currentX:o,startY:r,currentY:r,altKey:a};this.el.classList.add("active"),this._onDidStart.fire(l);const c=ca(this.el),d=()=>{let f="";i?f="all-scroll":this.orientation===1?this.state===1?f="s-resize":this.state===2?f="n-resize":f=Ye?"row-resize":"ns-resize":this.state===1?f="e-resize":this.state===2?f="w-resize":f=Ye?"col-resize":"ew-resize",c.textContent=`* { cursor: ${f} !important; }`},h=new U;d(),i||this.onDidEnablementChange.event(d,null,h);const u=f=>{gt.stop(f,!1);const p={startX:o,currentX:f.pageX,startY:r,currentY:f.pageY,altKey:a};this._onDidChange.fire(p)},g=f=>{gt.stop(f,!1),c.remove(),this.el.classList.remove("active"),this._onDidEnd.fire(),h.dispose();for(const p of n)p.classList.remove(Xq)};t.onPointerMove(u,null,h),t.onPointerUp(g,null,h),h.add(t)}onPointerDoublePress(e){const t=this.getOrthogonalSash(e);t&&t._onDidReset.fire(),this.linkedSash&&this.linkedSash._onDidReset.fire(),this._onDidReset.fire()}static onMouseEnter(e,t=!1){e.el.classList.contains("active")?(e.hoverDelayer.cancel(),e.el.classList.add("hover")):e.hoverDelayer.trigger(()=>e.el.classList.add("hover"),e.hoverDelay).then(void 0,()=>{}),!t&&e.linkedSash&&ls.onMouseEnter(e.linkedSash,!0)}static onMouseLeave(e,t=!1){e.hoverDelayer.cancel(),e.el.classList.remove("hover"),!t&&e.linkedSash&&ls.onMouseLeave(e.linkedSash,!0)}clearSashHoverState(){ls.onMouseLeave(this)}layout(){if(this.orientation===0){const e=this.layoutProvider;this.el.style.left=e.getVerticalSashLeft(this)-this.size/2+"px",e.getVerticalSashTop&&(this.el.style.top=e.getVerticalSashTop(this)+"px"),e.getVerticalSashHeight&&(this.el.style.height=e.getVerticalSashHeight(this)+"px")}else{const e=this.layoutProvider;this.el.style.top=e.getHorizontalSashTop(this)-this.size/2+"px",e.getHorizontalSashLeft&&(this.el.style.left=e.getHorizontalSashLeft(this)+"px"),e.getHorizontalSashWidth&&(this.el.style.width=e.getHorizontalSashWidth(this)+"px")}}getOrthogonalSash(e){const t=e.initialTarget??e.target;if(!(!t||!ki(t))&&t.classList.contains("orthogonal-drag-handle"))return t.classList.contains("start")?this.orthogonalStartSash:this.orthogonalEndSash}dispose(){super.dispose(),this.el.remove()}}class Dke{resetSash(){this._sashRatio.set(void 0,void 0)}constructor(e,t){this._options=e,this.dimensions=t,this.sashLeft=oW(this,i=>{const n=this._sashRatio.read(i)??this._options.splitViewDefaultRatio.read(i);return this._computeSashLeft(n,i)},(i,n)=>{const o=this.dimensions.width.get();this._sashRatio.set(i/o,n)}),this._sashRatio=Re(this,void 0)}_computeSashLeft(e,t){const i=this.dimensions.width.read(t),n=Math.floor(this._options.splitViewDefaultRatio.read(t)*i),o=this._options.enableSplitViewResizing.read(t)?Math.floor(e*i):n,r=100;return i<=r*2?n:o<r?r:o>i-r?i-r:o}}class Cie extends F{constructor(e,t,i,n,o,r){super(),this._domNode=e,this._dimensions=t,this._enabled=i,this._boundarySashes=n,this.sashLeft=o,this._resetSash=r,this._sash=this._register(new ls(this._domNode,{getVerticalSashTop:a=>0,getVerticalSashLeft:a=>this.sashLeft.get(),getVerticalSashHeight:a=>this._dimensions.height.get()},{orientation:0})),this._startSashPosition=void 0,this._register(this._sash.onDidStart(()=>{this._startSashPosition=this.sashLeft.get()})),this._register(this._sash.onDidChange(a=>{this.sashLeft.set(this._startSashPosition+(a.currentX-a.startX),void 0)})),this._register(this._sash.onDidEnd(()=>this._sash.layout())),this._register(this._sash.onDidReset(()=>this._resetSash())),this._register(De(a=>{const l=this._boundarySashes.read(a);l&&(this._sash.orthogonalEndSash=l.bottom)})),this._register(De(a=>{const l=this._enabled.read(a);this._sash.state=l?3:0,this.sashLeft.read(a),this._dimensions.height.read(a),this._sash.layout()}))}}const cN=class cN extends F{constructor(){super(...arguments),this._id=++cN.idCounter,this._onDidDispose=this._register(new M),this.onDidDispose=this._onDidDispose.event}getId(){return this.getEditorType()+":v2:"+this._id}getVisibleColumnFromPosition(e){return this._targetEditor.getVisibleColumnFromPosition(e)}getPosition(){return this._targetEditor.getPosition()}setPosition(e,t="api"){this._targetEditor.setPosition(e,t)}revealLine(e,t=0){this._targetEditor.revealLine(e,t)}revealLineInCenter(e,t=0){this._targetEditor.revealLineInCenter(e,t)}revealLineInCenterIfOutsideViewport(e,t=0){this._targetEditor.revealLineInCenterIfOutsideViewport(e,t)}revealLineNearTop(e,t=0){this._targetEditor.revealLineNearTop(e,t)}revealPosition(e,t=0){this._targetEditor.revealPosition(e,t)}revealPositionInCenter(e,t=0){this._targetEditor.revealPositionInCenter(e,t)}revealPositionInCenterIfOutsideViewport(e,t=0){this._targetEditor.revealPositionInCenterIfOutsideViewport(e,t)}revealPositionNearTop(e,t=0){this._targetEditor.revealPositionNearTop(e,t)}getSelection(){return this._targetEditor.getSelection()}getSelections(){return this._targetEditor.getSelections()}setSelection(e,t="api"){this._targetEditor.setSelection(e,t)}setSelections(e,t="api"){this._targetEditor.setSelections(e,t)}revealLines(e,t,i=0){this._targetEditor.revealLines(e,t,i)}revealLinesInCenter(e,t,i=0){this._targetEditor.revealLinesInCenter(e,t,i)}revealLinesInCenterIfOutsideViewport(e,t,i=0){this._targetEditor.revealLinesInCenterIfOutsideViewport(e,t,i)}revealLinesNearTop(e,t,i=0){this._targetEditor.revealLinesNearTop(e,t,i)}revealRange(e,t=0,i=!1,n=!0){this._targetEditor.revealRange(e,t,i,n)}revealRangeInCenter(e,t=0){this._targetEditor.revealRangeInCenter(e,t)}revealRangeInCenterIfOutsideViewport(e,t=0){this._targetEditor.revealRangeInCenterIfOutsideViewport(e,t)}revealRangeNearTop(e,t=0){this._targetEditor.revealRangeNearTop(e,t)}revealRangeNearTopIfOutsideViewport(e,t=0){this._targetEditor.revealRangeNearTopIfOutsideViewport(e,t)}revealRangeAtTop(e,t=0){this._targetEditor.revealRangeAtTop(e,t)}getSupportedActions(){return this._targetEditor.getSupportedActions()}focus(){this._targetEditor.focus()}trigger(e,t,i){this._targetEditor.trigger(e,t,i)}createDecorationsCollection(e){return this._targetEditor.createDecorationsCollection(e)}changeDecorations(e){return this._targetEditor.changeDecorations(e)}};cN.idCounter=0;let w4=cN;const $n={enableSplitViewResizing:!0,renderSideBySide:!0,renderMarginRevertIcon:!0,renderGutterMenu:!0,maxComputationTime:5e3,maxFileSize:50,ignoreTrimWhitespace:!0,renderIndicators:!0,originalEditable:!1,diffCodeLens:!1,renderOverviewRuler:!0,diffWordWrap:"inherit",diffAlgorithm:"advanced",accessibilityVerbose:!1,experimental:{showMoves:!1,showEmptyDecorations:!0,useTrueInlineView:!1},hideUnchangedRegions:{enabled:!1,contextLineCount:3,minimumLineCount:3,revealLineCount:20},isInEmbeddedEditor:!1,onlyShowAccessibleDiffViewer:!1,renderSideBySideInlineBreakpoint:900,useInlineViewWhenSpaceIsLimited:!0,compactMode:!1};function Tke(s,e){return Ih(s,(t,i)=>i??e(t))}var Nke=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},Rke=function(s,e){return function(t,i){e(t,i,s)}};let v4=class{get editorOptions(){return this._options}constructor(e,t){this._accessibilityService=t,this._diffEditorWidth=Re(this,0),this._screenReaderMode=bt(this,this._accessibilityService.onDidChangeScreenReaderOptimized,()=>this._accessibilityService.isScreenReaderOptimized()),this.couldShowInlineViewBecauseOfSize=j(this,n=>this._options.read(n).renderSideBySide&&this._diffEditorWidth.read(n)<=this._options.read(n).renderSideBySideInlineBreakpoint),this.renderOverviewRuler=j(this,n=>this._options.read(n).renderOverviewRuler),this.renderSideBySide=j(this,n=>this.compactMode.read(n)&&this.shouldRenderInlineViewInSmartMode.read(n)?!1:this._options.read(n).renderSideBySide&&!(this._options.read(n).useInlineViewWhenSpaceIsLimited&&this.couldShowInlineViewBecauseOfSize.read(n)&&!this._screenReaderMode.read(n))),this.readOnly=j(this,n=>this._options.read(n).readOnly),this.shouldRenderOldRevertArrows=j(this,n=>!(!this._options.read(n).renderMarginRevertIcon||!this.renderSideBySide.read(n)||this.readOnly.read(n)||this.shouldRenderGutterMenu.read(n))),this.shouldRenderGutterMenu=j(this,n=>this._options.read(n).renderGutterMenu),this.renderIndicators=j(this,n=>this._options.read(n).renderIndicators),this.enableSplitViewResizing=j(this,n=>this._options.read(n).enableSplitViewResizing),this.splitViewDefaultRatio=j(this,n=>this._options.read(n).splitViewDefaultRatio),this.ignoreTrimWhitespace=j(this,n=>this._options.read(n).ignoreTrimWhitespace),this.maxComputationTimeMs=j(this,n=>this._options.read(n).maxComputationTime),this.showMoves=j(this,n=>this._options.read(n).experimental.showMoves&&this.renderSideBySide.read(n)),this.isInEmbeddedEditor=j(this,n=>this._options.read(n).isInEmbeddedEditor),this.diffWordWrap=j(this,n=>this._options.read(n).diffWordWrap),this.originalEditable=j(this,n=>this._options.read(n).originalEditable),this.diffCodeLens=j(this,n=>this._options.read(n).diffCodeLens),this.accessibilityVerbose=j(this,n=>this._options.read(n).accessibilityVerbose),this.diffAlgorithm=j(this,n=>this._options.read(n).diffAlgorithm),this.showEmptyDecorations=j(this,n=>this._options.read(n).experimental.showEmptyDecorations),this.onlyShowAccessibleDiffViewer=j(this,n=>this._options.read(n).onlyShowAccessibleDiffViewer),this.compactMode=j(this,n=>this._options.read(n).compactMode),this.trueInlineDiffRenderingEnabled=j(this,n=>this._options.read(n).experimental.useTrueInlineView),this.useTrueInlineDiffRendering=j(this,n=>!this.renderSideBySide.read(n)&&this.trueInlineDiffRenderingEnabled.read(n)),this.hideUnchangedRegions=j(this,n=>this._options.read(n).hideUnchangedRegions.enabled),this.hideUnchangedRegionsRevealLineCount=j(this,n=>this._options.read(n).hideUnchangedRegions.revealLineCount),this.hideUnchangedRegionsContextLineCount=j(this,n=>this._options.read(n).hideUnchangedRegions.contextLineCount),this.hideUnchangedRegionsMinimumLineCount=j(this,n=>this._options.read(n).hideUnchangedRegions.minimumLineCount),this._model=Re(this,void 0),this.shouldRenderInlineViewInSmartMode=this._model.map(this,n=>Tke(this,o=>{const r=n?.diff.read(o);return r?Ake(r,this.trueInlineDiffRenderingEnabled.read(o)):void 0})).flatten().map(this,n=>!!n),this.inlineViewHideOriginalLineNumbers=this.compactMode;const i={...e,...Qq(e,$n)};this._options=Re(this,i)}updateOptions(e){const t=Qq(e,this._options.get()),i={...this._options.get(),...e,...t};this._options.set(i,void 0,{changedOptions:e})}setWidth(e){this._diffEditorWidth.set(e,void 0)}setModel(e){this._model.set(e,void 0)}};v4=Nke([Rke(1,Wn)],v4);function Ake(s,e){return s.mappings.every(t=>Mke(t.lineRangeMapping)||Oke(t.lineRangeMapping)||e&&B7(t.lineRangeMapping))}function Mke(s){return s.original.length===0}function Oke(s){return s.modified.length===0}function Qq(s,e){return{enableSplitViewResizing:we(s.enableSplitViewResizing,e.enableSplitViewResizing),splitViewDefaultRatio:$fe(s.splitViewDefaultRatio,.5,.1,.9),renderSideBySide:we(s.renderSideBySide,e.renderSideBySide),renderMarginRevertIcon:we(s.renderMarginRevertIcon,e.renderMarginRevertIcon),maxComputationTime:Dd(s.maxComputationTime,e.maxComputationTime,0,1073741824),maxFileSize:Dd(s.maxFileSize,e.maxFileSize,0,1073741824),ignoreTrimWhitespace:we(s.ignoreTrimWhitespace,e.ignoreTrimWhitespace),renderIndicators:we(s.renderIndicators,e.renderIndicators),originalEditable:we(s.originalEditable,e.originalEditable),diffCodeLens:we(s.diffCodeLens,e.diffCodeLens),renderOverviewRuler:we(s.renderOverviewRuler,e.renderOverviewRuler),diffWordWrap:Yt(s.diffWordWrap,e.diffWordWrap,["off","on","inherit"]),diffAlgorithm:Yt(s.diffAlgorithm,e.diffAlgorithm,["legacy","advanced"],{smart:"legacy",experimental:"advanced"}),accessibilityVerbose:we(s.accessibilityVerbose,e.accessibilityVerbose),experimental:{showMoves:we(s.experimental?.showMoves,e.experimental.showMoves),showEmptyDecorations:we(s.experimental?.showEmptyDecorations,e.experimental.showEmptyDecorations),useTrueInlineView:we(s.experimental?.useTrueInlineView,e.experimental.useTrueInlineView)},hideUnchangedRegions:{enabled:we(s.hideUnchangedRegions?.enabled??s.experimental?.collapseUnchangedRegions,e.hideUnchangedRegions.enabled),contextLineCount:Dd(s.hideUnchangedRegions?.contextLineCount,e.hideUnchangedRegions.contextLineCount,0,1073741824),minimumLineCount:Dd(s.hideUnchangedRegions?.minimumLineCount,e.hideUnchangedRegions.minimumLineCount,0,1073741824),revealLineCount:Dd(s.hideUnchangedRegions?.revealLineCount,e.hideUnchangedRegions.revealLineCount,0,1073741824)},isInEmbeddedEditor:we(s.isInEmbeddedEditor,e.isInEmbeddedEditor),onlyShowAccessibleDiffViewer:we(s.onlyShowAccessibleDiffViewer,e.onlyShowAccessibleDiffViewer),renderSideBySideInlineBreakpoint:Dd(s.renderSideBySideInlineBreakpoint,e.renderSideBySideInlineBreakpoint,0,1073741824),useInlineViewWhenSpaceIsLimited:we(s.useInlineViewWhenSpaceIsLimited,e.useInlineViewWhenSpaceIsLimited),renderGutterMenu:we(s.renderGutterMenu,e.renderGutterMenu),compactMode:we(s.compactMode,e.compactMode)}}class Pke extends Lm{constructor(e,t){super(),this._onDidChangeVisibility=this._register(new M),this.onDidChangeVisibility=this._onDidChangeVisibility.event,this._element=X(e,ie(".monaco-dropdown")),this._label=X(this._element,ie(".dropdown-label"));let i=t.labelRenderer;i||(i=o=>(o.textContent=t.label||"",null));for(const o of[se.CLICK,se.MOUSE_DOWN,qt.Tap])this._register(W(this.element,o,r=>gt.stop(r,!0)));for(const o of[se.MOUSE_DOWN,qt.Tap])this._register(W(this._label,o,r=>{dW(r)&&r.button!==0||(this.visible?this.hide():this.show())}));this._register(W(this._label,se.KEY_DOWN,o=>{const r=new Pt(o);(r.equals(3)||r.equals(10))&&(gt.stop(o,!0),this.visible?this.hide():this.show())}));const n=i(this._label);n&&this._register(n),this._register(_s.addTarget(this._label))}get element(){return this._element}show(){this.visible||(this.visible=!0,this._onDidChangeVisibility.fire(!0))}hide(){this.visible&&(this.visible=!1,this._onDidChangeVisibility.fire(!1))}dispose(){super.dispose(),this.hide(),this.boxContainer&&(this.boxContainer.remove(),this.boxContainer=void 0),this.contents&&(this.contents.remove(),this.contents=void 0),this._label&&(this._label.remove(),this._label=void 0)}}class Fke extends Pke{constructor(e,t){super(e,t),this._options=t,this._actions=[],this.actions=t.actions||[]}set menuOptions(e){this._menuOptions=e}get menuOptions(){return this._menuOptions}get actions(){return this._options.actionProvider?this._options.actionProvider.getActions():this._actions}set actions(e){this._actions=e}show(){super.show(),this.element.classList.add("active"),this._options.contextMenuProvider.showContextMenu({getAnchor:()=>this.element,getActions:()=>this.actions,getActionsContext:()=>this.menuOptions?this.menuOptions.context:null,getActionViewItem:(e,t)=>this.menuOptions&&this.menuOptions.actionViewItemProvider?this.menuOptions.actionViewItemProvider(e,t):void 0,getKeyBinding:e=>this.menuOptions&&this.menuOptions.getKeyBinding?this.menuOptions.getKeyBinding(e):void 0,getMenuClassName:()=>this._options.menuClassName||"",onHide:()=>this.onHide(),actionRunner:this.menuOptions?this.menuOptions.actionRunner:void 0,anchorAlignment:this.menuOptions?this.menuOptions.anchorAlignment:0,domForShadowRoot:this._options.menuAsChild?this.element:void 0,skipTelemetry:this._options.skipTelemetry})}hide(){super.hide()}onHide(){this.hide(),this.element.classList.remove("active")}}class DI extends ll{get onDidChangeVisibility(){return this._onDidChangeVisibility.event}constructor(e,t,i,n=Object.create(null)){super(null,e,n),this.actionItem=null,this._onDidChangeVisibility=this._register(new M),this.menuActionsOrProvider=t,this.contextMenuProvider=i,this.options=n,this.options.actionRunner&&(this.actionRunner=this.options.actionRunner)}render(e){this.actionItem=e;const t=o=>(this.element=X(o,ie("a.action-label")),this.renderLabel(this.element)),i=Array.isArray(this.menuActionsOrProvider),n={contextMenuProvider:this.contextMenuProvider,labelRenderer:t,menuAsChild:this.options.menuAsChild,actions:i?this.menuActionsOrProvider:void 0,actionProvider:i?void 0:this.menuActionsOrProvider,skipTelemetry:this.options.skipTelemetry};if(this.dropdownMenu=this._register(new Fke(e,n)),this._register(this.dropdownMenu.onDidChangeVisibility(o=>{this.element?.setAttribute("aria-expanded",`${o}`),this._onDidChangeVisibility.fire(o)})),this.dropdownMenu.menuOptions={actionViewItemProvider:this.options.actionViewItemProvider,actionRunner:this.actionRunner,getKeyBinding:this.options.keybindingProvider,context:this._context},this.options.anchorAlignmentProvider){const o=this;this.dropdownMenu.menuOptions={...this.dropdownMenu.menuOptions,get anchorAlignment(){return o.options.anchorAlignmentProvider()}}}this.updateTooltip(),this.updateEnabled()}renderLabel(e){let t=[];return typeof this.options.classNames=="string"?t=this.options.classNames.split(/\s+/g).filter(i=>!!i):this.options.classNames&&(t=this.options.classNames),t.find(i=>i==="icon")||t.push("codicon"),e.classList.add(...t),this._action.label&&this._register($a().setupManagedHover(this.options.hoverDelegate??Uc("mouse"),e,this._action.label)),null}getTooltip(){let e=null;return this.action.tooltip?e=this.action.tooltip:this.action.label&&(e=this.action.label),e??void 0}setActionContext(e){super.setActionContext(e),this.dropdownMenu&&(this.dropdownMenu.menuOptions?this.dropdownMenu.menuOptions.context=e:this.dropdownMenu.menuOptions={context:e})}show(){this.dropdownMenu?.show()}updateEnabled(){const e=!this.action.enabled;this.actionItem?.classList.toggle("disabled",e),this.element?.classList.toggle("disabled",e)}}const Jq=24;class Bke extends F{get onDidChangeDropdownVisibility(){return this._onDidChangeDropdownVisibility.event}constructor(e,t,i={orientation:0}){if(super(),this.submenuActionViewItems=[],this.hasSecondaryActions=!1,this._onDidChangeDropdownVisibility=this._register(new yde),this.originalPrimaryActions=[],this.originalSecondaryActions=[],this.hiddenActions=[],this.disposables=this._register(new U),i.hoverDelegate=i.hoverDelegate??this._register(wte()),this.options=i,this.toggleMenuAction=this._register(new FS(()=>this.toggleMenuActionViewItem?.show(),i.toggleMenuTitle)),this.element=document.createElement("div"),this.element.className="monaco-toolbar",e.appendChild(this.element),this.actionBar=this._register(new bo(this.element,{orientation:i.orientation,ariaLabel:i.ariaLabel,actionRunner:i.actionRunner,allowContextMenu:i.allowContextMenu,highlightToggledItems:i.highlightToggledItems,hoverDelegate:i.hoverDelegate,actionViewItemProvider:(n,o)=>{if(n.id===FS.ID)return this.toggleMenuActionViewItem=new DI(n,{getActions:()=>this.toggleMenuAction.menuActions},t,{actionViewItemProvider:this.options.actionViewItemProvider,actionRunner:this.actionRunner,keybindingProvider:this.options.getKeyBinding,classNames:xe.asClassNameArray(i.moreIcon??Z.toolBarMore),anchorAlignmentProvider:this.options.anchorAlignmentProvider,menuAsChild:!!this.options.renderDropdownAsChildElement,skipTelemetry:this.options.skipTelemetry,isMenu:!0,hoverDelegate:this.options.hoverDelegate}),this.toggleMenuActionViewItem.setActionContext(this.actionBar.context),this.disposables.add(this._onDidChangeDropdownVisibility.add(this.toggleMenuActionViewItem.onDidChangeVisibility)),this.toggleMenuActionViewItem;if(i.actionViewItemProvider){const r=i.actionViewItemProvider(n,o);if(r)return r}if(n instanceof ov){const r=new DI(n,n.actions,t,{actionViewItemProvider:this.options.actionViewItemProvider,actionRunner:this.actionRunner,keybindingProvider:this.options.getKeyBinding,classNames:n.class,anchorAlignmentProvider:this.options.anchorAlignmentProvider,menuAsChild:!!this.options.renderDropdownAsChildElement,skipTelemetry:this.options.skipTelemetry,hoverDelegate:this.options.hoverDelegate});return r.setActionContext(this.actionBar.context),this.submenuActionViewItems.push(r),this.disposables.add(this._onDidChangeDropdownVisibility.add(r.onDidChangeVisibility)),r}}})),this.options.responsive){this.element.classList.add("responsive");const n=new ResizeObserver(()=>{this.setToolbarMaxWidth(this.element.getBoundingClientRect().width)});n.observe(this.element),this._store.add(pe(()=>n.disconnect()))}}set actionRunner(e){this.actionBar.actionRunner=e}get actionRunner(){return this.actionBar.actionRunner}set context(e){this.actionBar.context=e,this.toggleMenuActionViewItem?.setActionContext(e);for(const t of this.submenuActionViewItems)t.setActionContext(e)}getElement(){return this.element}getItemAction(e){return this.actionBar.getAction(e)}getItemWidth(e){return this.actionBar.getWidth(e)}setActions(e,t){this.clear(),this.originalPrimaryActions=e?e.slice(0):[],this.originalSecondaryActions=t?t.slice(0):[];const i=e?e.slice(0):[];this.hasSecondaryActions=!!(t&&t.length>0),this.hasSecondaryActions&&t&&(this.toggleMenuAction.menuActions=t.slice(0),i.push(this.toggleMenuAction)),i.length>0&&this.options.trailingSeparator&&i.push(new tn),i.forEach(n=>{this.actionBar.push(n,{icon:this.options.icon??!0,label:this.options.label??!1,keybinding:this.getKeybindingLabel(n)})}),this.options.responsive&&(this.hiddenActions.length=0,this.setToolbarMaxWidth(this.element.getBoundingClientRect().width))}getKeybindingLabel(e){return this.options.getKeyBinding?.(e)?.getLabel()??void 0}getItemsWidthResponsive(){return this.actionBar.length()*Jq}setToolbarMaxWidth(e){if(this.actionBar.isEmpty()||this.getItemsWidthResponsive()<=e&&this.hiddenActions.length===0)return;if(this.getItemsWidthResponsive()>e)for(;this.getItemsWidthResponsive()>e&&this.actionBar.length()>0;){const i=this.originalPrimaryActions.length-this.hiddenActions.length-1;if(i<0)break;const n=Math.min(Jq,this.getItemWidth(i)),o=this.originalPrimaryActions[i];this.hiddenActions.unshift({action:o,size:n}),this.actionBar.pull(i),this.originalSecondaryActions.length===0&&this.hiddenActions.length===1&&this.actionBar.push(this.toggleMenuAction,{icon:this.options.icon??!0,label:this.options.label??!1,keybinding:this.getKeybindingLabel(this.toggleMenuAction)})}else for(;this.hiddenActions.length>0;){const i=this.hiddenActions.shift();if(this.getItemsWidthResponsive()+i.size>e){this.hiddenActions.unshift(i);break}this.actionBar.push(i.action,{icon:this.options.icon??!0,label:this.options.label??!1,keybinding:this.getKeybindingLabel(i.action),index:this.originalPrimaryActions.length-this.hiddenActions.length-1}),this.originalSecondaryActions.length===0&&this.hiddenActions.length===1&&(this.toggleMenuAction.menuActions=[],this.actionBar.pull(this.actionBar.length()-1))}const t=this.hiddenActions.map(i=>i.action);if(this.originalSecondaryActions.length>0||t.length>0){const i=this.originalSecondaryActions.slice(0);this.toggleMenuAction.menuActions=tn.join(t,i)}}clear(){this.submenuActionViewItems=[],this.disposables.clear(),this.actionBar.clear()}dispose(){this.clear(),this.disposables.dispose(),super.dispose()}}const dN=class dN extends Cr{constructor(e,t){t=t||m(17,"More Actions..."),super(dN.ID,t,void 0,!0),this._menuActions=[],this.toggleDropdownMenu=e}async run(){this.toggleDropdownMenu()}get menuActions(){return this._menuActions}set menuActions(e){this._menuActions=e}};dN.ID="toolbar.toggle.more";let FS=dN;class yA{constructor(e,t,i=t){this.modifierLabels=[null],this.modifierLabels[2]=e,this.modifierLabels[1]=t,this.modifierLabels[3]=i}toLabel(e,t,i){if(t.length===0)return null;const n=[];for(let o=0,r=t.length;o<r;o++){const a=t[o],l=i(a);if(l===null)return null;n[o]=zke(a,l,this.modifierLabels[e])}return n.join(" ")}}const z7=new yA({ctrlKey:"⌃",shiftKey:"⇧",altKey:"⌥",metaKey:"⌘",separator:""},{ctrlKey:m(35,"Ctrl"),shiftKey:m(36,"Shift"),altKey:m(37,"Alt"),metaKey:m(38,"Windows"),separator:"+"},{ctrlKey:m(39,"Ctrl"),shiftKey:m(40,"Shift"),altKey:m(41,"Alt"),metaKey:m(42,"Super"),separator:"+"}),Wke=new yA({ctrlKey:m(43,"Control"),shiftKey:m(44,"Shift"),altKey:m(45,"Option"),metaKey:m(46,"Command"),separator:"+"},{ctrlKey:m(47,"Control"),shiftKey:m(48,"Shift"),altKey:m(49,"Alt"),metaKey:m(50,"Windows"),separator:"+"},{ctrlKey:m(51,"Control"),shiftKey:m(52,"Shift"),altKey:m(53,"Alt"),metaKey:m(54,"Super"),separator:"+"}),Hke=new yA({ctrlKey:"Ctrl",shiftKey:"Shift",altKey:"Alt",metaKey:"Cmd",separator:"+"},{ctrlKey:"Ctrl",shiftKey:"Shift",altKey:"Alt",metaKey:"Super",separator:"+"}),Vke=new yA({ctrlKey:"ctrl",shiftKey:"shift",altKey:"alt",metaKey:"cmd",separator:"+"},{ctrlKey:"ctrl",shiftKey:"shift",altKey:"alt",metaKey:"win",separator:"+"},{ctrlKey:"ctrl",shiftKey:"shift",altKey:"alt",metaKey:"meta",separator:"+"});function zke(s,e,t){if(e===null)return"";const i=[];return s.ctrlKey&&i.push(t.ctrlKey),s.shiftKey&&i.push(t.shiftKey),s.altKey&&i.push(t.altKey),s.metaKey&&i.push(t.metaKey),e!==""&&i.push(e),i.join(t.separator)}function $ke(s){return!!s&&typeof s=="object"&&typeof s.original=="string"&&typeof s.value=="string"}function Uke(s){return s?s.condition!==void 0:!1}var yw;(function(s){s[s.STORAGE_DOES_NOT_EXIST=0]="STORAGE_DOES_NOT_EXIST",s[s.STORAGE_IN_MEMORY=1]="STORAGE_IN_MEMORY"})(yw||(yw={}));var Cf;(function(s){s[s.None=0]="None",s[s.Initialized=1]="Initialized",s[s.Closed=2]="Closed"})(Cf||(Cf={}));const hN=class hN extends F{constructor(e,t=Object.create(null)){super(),this.database=e,this.options=t,this._onDidChangeStorage=this._register(new Sm),this.onDidChangeStorage=this._onDidChangeStorage.event,this.state=Cf.None,this.cache=new Map,this.flushDelayer=this._register(new KX(hN.DEFAULT_FLUSH_DELAY)),this.pendingDeletes=new Set,this.pendingInserts=new Map,this.pendingClose=void 0,this.whenFlushedCallbacks=[],this.registerListeners()}registerListeners(){this._register(this.database.onDidChangeItemsExternal(e=>this.onDidChangeItemsExternal(e)))}onDidChangeItemsExternal(e){this._onDidChangeStorage.pause();try{e.changed?.forEach((t,i)=>this.acceptExternal(i,t)),e.deleted?.forEach(t=>this.acceptExternal(t,void 0))}finally{this._onDidChangeStorage.resume()}}acceptExternal(e,t){if(this.state===Cf.Closed)return;let i=!1;Ur(t)?i=this.cache.delete(e):this.cache.get(e)!==t&&(this.cache.set(e,t),i=!0),i&&this._onDidChangeStorage.fire({key:e,external:!0})}get(e,t){const i=this.cache.get(e);return Ur(i)?t:i}getBoolean(e,t){const i=this.get(e);return Ur(i)?t:i==="true"}getNumber(e,t){const i=this.get(e);return Ur(i)?t:parseInt(i,10)}async set(e,t,i=!1){if(this.state===Cf.Closed)return;if(Ur(t))return this.delete(e,i);const n=rn(t)||Array.isArray(t)?ixe(t):String(t);if(this.cache.get(e)!==n)return this.cache.set(e,n),this.pendingInserts.set(e,n),this.pendingDeletes.delete(e),this._onDidChangeStorage.fire({key:e,external:i}),this.doFlush()}async delete(e,t=!1){if(!(this.state===Cf.Closed||!this.cache.delete(e)))return this.pendingDeletes.has(e)||this.pendingDeletes.add(e),this.pendingInserts.delete(e),this._onDidChangeStorage.fire({key:e,external:t}),this.doFlush()}get hasPending(){return this.pendingInserts.size>0||this.pendingDeletes.size>0}async flushPending(){if(!this.hasPending)return;const e={insert:this.pendingInserts,delete:this.pendingDeletes};return this.pendingDeletes=new Set,this.pendingInserts=new Map,this.database.updateItems(e).finally(()=>{if(!this.hasPending)for(;this.whenFlushedCallbacks.length;)this.whenFlushedCallbacks.pop()?.()})}async flush(e){if(!(this.state===Cf.Closed||this.pendingClose))return this.doFlush(e)}async doFlush(e){return this.options.hint===yw.STORAGE_IN_MEMORY?this.flushPending():this.flushDelayer.trigger(()=>this.flushPending(),e)}async whenFlushed(){if(this.hasPending)return new Promise(e=>this.whenFlushedCallbacks.push(e))}};hN.DEFAULT_FLUSH_DELAY=100;let K1=hN;class WM{constructor(){this.onDidChangeItemsExternal=oe.None,this.items=new Map}async updateItems(e){e.insert?.forEach((t,i)=>this.items.set(i,t)),e.delete?.forEach(t=>this.items.delete(t))}}const Bk="__$__targetStorageMarker",Bs=qe("storageService");var Fu;(function(s){s[s.NONE=0]="NONE",s[s.SHUTDOWN=1]="SHUTDOWN"})(Fu||(Fu={}));function qke(s){const e=s.get(Bk);if(e)try{return JSON.parse(e)}catch{}return Object.create(null)}const uN=class uN extends F{constructor(e={flushInterval:uN.DEFAULT_FLUSH_INTERVAL}){super(),this._onDidChangeValue=this._register(new Sm),this._onDidChangeTarget=this._register(new Sm),this._onWillSaveState=this._register(new M),this.onWillSaveState=this._onWillSaveState.event,this.runFlushWhenIdle=this._register(new vt),this._workspaceKeyTargets=void 0,this._profileKeyTargets=void 0,this._applicationKeyTargets=void 0,this.flushWhenIdleScheduler=this._register(new Rt(()=>this.doFlushWhenIdle(),e.flushInterval))}onDidChangeValue(e,t,i){return oe.filter(this._onDidChangeValue.event,n=>n.scope===e&&(t===void 0||n.key===t),i)}doFlushWhenIdle(){this.runFlushWhenIdle.value=U6(()=>{this.shouldFlushWhenIdle()&&this.flush(),this.flushWhenIdleScheduler.schedule()})}shouldFlushWhenIdle(){return!0}emitDidChangeValue(e,t){const{key:i,external:n}=t;if(i===Bk){switch(e){case-1:this._applicationKeyTargets=void 0;break;case 0:this._profileKeyTargets=void 0;break;case 1:this._workspaceKeyTargets=void 0;break}this._onDidChangeTarget.fire({scope:e})}else this._onDidChangeValue.fire({scope:e,key:i,target:this.getKeyTargets(e)[i],external:n})}get(e,t,i){return this.getStorage(t)?.get(e,i)}getBoolean(e,t,i){return this.getStorage(t)?.getBoolean(e,i)}getNumber(e,t,i){return this.getStorage(t)?.getNumber(e,i)}store(e,t,i,n,o=!1){if(Ur(t)){this.remove(e,i,o);return}this.withPausedEmitters(()=>{this.updateKeyTarget(e,i,n),this.getStorage(i)?.set(e,t,o)})}remove(e,t,i=!1){this.withPausedEmitters(()=>{this.updateKeyTarget(e,t,void 0),this.getStorage(t)?.delete(e,i)})}withPausedEmitters(e){this._onDidChangeValue.pause(),this._onDidChangeTarget.pause();try{e()}finally{this._onDidChangeValue.resume(),this._onDidChangeTarget.resume()}}updateKeyTarget(e,t,i,n=!1){const o=this.getKeyTargets(t);typeof i=="number"?o[e]!==i&&(o[e]=i,this.getStorage(t)?.set(Bk,JSON.stringify(o),n)):typeof o[e]=="number"&&(delete o[e],this.getStorage(t)?.set(Bk,JSON.stringify(o),n))}get workspaceKeyTargets(){return this._workspaceKeyTargets||(this._workspaceKeyTargets=this.loadKeyTargets(1)),this._workspaceKeyTargets}get profileKeyTargets(){return this._profileKeyTargets||(this._profileKeyTargets=this.loadKeyTargets(0)),this._profileKeyTargets}get applicationKeyTargets(){return this._applicationKeyTargets||(this._applicationKeyTargets=this.loadKeyTargets(-1)),this._applicationKeyTargets}getKeyTargets(e){switch(e){case-1:return this.applicationKeyTargets;case 0:return this.profileKeyTargets;default:return this.workspaceKeyTargets}}loadKeyTargets(e){const t=this.getStorage(e);return t?qke(t):Object.create(null)}async flush(e=Fu.NONE){this._onWillSaveState.fire({reason:e});const t=this.getStorage(-1),i=this.getStorage(0),n=this.getStorage(1);switch(e){case Fu.NONE:await Uy.settled([t?.whenFlushed()??Promise.resolve(),i?.whenFlushed()??Promise.resolve(),n?.whenFlushed()??Promise.resolve()]);break;case Fu.SHUTDOWN:await Uy.settled([t?.flush(0)??Promise.resolve(),i?.flush(0)??Promise.resolve(),n?.flush(0)??Promise.resolve()]);break}}};uN.DEFAULT_FLUSH_INTERVAL=60*1e3;let C4=uN;class jke extends C4{constructor(){super(),this.applicationStorage=this._register(new K1(new WM,{hint:yw.STORAGE_IN_MEMORY})),this.profileStorage=this._register(new K1(new WM,{hint:yw.STORAGE_IN_MEMORY})),this.workspaceStorage=this._register(new K1(new WM,{hint:yw.STORAGE_IN_MEMORY})),this._register(this.workspaceStorage.onDidChangeStorage(e=>this.emitDidChangeValue(1,e))),this._register(this.profileStorage.onDidChangeStorage(e=>this.emitDidChangeValue(0,e))),this._register(this.applicationStorage.onDidChangeStorage(e=>this.emitDidChangeValue(-1,e)))}getStorage(e){switch(e){case-1:return this.applicationStorage;case 0:return this.profileStorage;default:return this.workspaceStorage}}shouldFlushWhenIdle(){return!1}}function Gke(s,e){const t={...e};for(const i in s){const n=s[i];t[i]=n!==void 0?te(n):void 0}return t}const yie={keybindingLabelBackground:te(e_e),keybindingLabelForeground:te(t_e),keybindingLabelBorder:te(i_e),keybindingLabelBottomBorder:te(n_e),keybindingLabelShadow:te(tC)},Kke={buttonForeground:te(oA),buttonSeparator:te(Vme),buttonBackground:te(jf),buttonHoverBackground:te(zme),buttonSecondaryForeground:te(HJ),buttonSecondaryBackground:te(qE),buttonSecondaryHoverBackground:te(Ume),buttonBorder:te($me)},Yke={progressBarBackground:te(nme)},TI={inputActiveOptionBorder:te(X0),inputActiveOptionForeground:te(Q0),inputActiveOptionBackground:te(iC)};te(a1),te(qme),te(jme),te(Gme),te(Kme),te(Yme),te(Zme);const $7={checkboxBackground:te(WW),checkboxBorder:te(Xme),checkboxForeground:te(HW),checkboxDisabledBackground:te(Qme),checkboxDisabledForeground:te(Jme)};te(wr),te(tA),te(tC),te(OW),te(yme),te(Sme),te(xme),te(tme);const NI={inputBackground:te(JP),inputForeground:te(BJ),inputBorder:te(WJ),inputValidationInfoBorder:te(Ame),inputValidationInfoBackground:te(Nme),inputValidationInfoForeground:te(Rme),inputValidationWarningBorder:te(Pme),inputValidationWarningBackground:te(Mme),inputValidationWarningForeground:te(Ome),inputValidationErrorBorder:te(Wme),inputValidationErrorBackground:te(Fme),inputValidationErrorForeground:te(Bme)},Zke={listFilterWidgetBackground:te(m_e),listFilterWidgetOutline:te(__e),listFilterWidgetNoMatchesOutline:te(b_e),listFilterWidgetShadow:te(w_e),inputBoxStyles:NI,toggleStyles:TI},Sie={badgeBackground:te(Tk),badgeForeground:te(ime),badgeBorder:te(rt)};te(vme),te(wme),te(A$),te(A$),te(Cme);const lC={listBackground:void 0,listInactiveFocusForeground:void 0,listFocusBackground:te(s_e),listFocusForeground:te(o_e),listFocusOutline:te(r_e),listActiveSelectionBackground:te(Im),listActiveSelectionForeground:te(cS),listActiveSelectionIconForeground:te(VJ),listFocusAndSelectionOutline:te(a_e),listFocusAndSelectionBackground:te(Im),listFocusAndSelectionForeground:te(cS),listInactiveSelectionBackground:te(l_e),listInactiveSelectionIconForeground:te(d_e),listInactiveSelectionForeground:te(c_e),listInactiveFocusBackground:te(h_e),listInactiveFocusOutline:te(u_e),listHoverBackground:te(zJ),listHoverForeground:te($J),listDropOverBackground:te(g_e),listDropBetweenBackground:te(f_e),listSelectionOutline:te(ri),listHoverOutline:te(ri),treeIndentGuidesStroke:te(UJ),treeInactiveIndentGuidesStroke:te(v_e),treeStickyScrollBackground:void 0,treeStickyScrollBorder:void 0,treeStickyScrollShadow:te(eA),tableColumnsBorder:te(C_e),tableOddRowsBackgroundColor:te(y_e)};function f_(s){return Gke(s,lC)}const Xke={selectBackground:te(sA),selectListBackground:te(Hme),selectForeground:te(FW),decoratorRightForeground:te(qJ),selectBorder:te(BW),focusBorder:te(Nc),listFocusBackground:te(hS),listInactiveSelectionIconForeground:te(VW),listFocusForeground:te(dS),listFocusOutline:EJ(ri,q.transparent.toString()),listHoverBackground:te(zJ),listHoverForeground:te($J),listHoverOutline:te(ri),selectListBorder:te(TW),listBackground:void 0,listActiveSelectionBackground:void 0,listActiveSelectionForeground:void 0,listActiveSelectionIconForeground:void 0,listFocusAndSelectionBackground:void 0,listDropOverBackground:void 0,listDropBetweenBackground:void 0,listInactiveSelectionBackground:void 0,listInactiveSelectionForeground:void 0,listInactiveFocusBackground:void 0,listInactiveFocusOutline:void 0,listSelectionOutline:void 0,listFocusAndSelectionForeground:void 0,listFocusAndSelectionOutline:void 0,listInactiveFocusForeground:void 0,tableColumnsBorder:void 0,tableOddRowsBackgroundColor:void 0,treeIndentGuidesStroke:void 0,treeInactiveIndentGuidesStroke:void 0,treeStickyScrollBackground:void 0,treeStickyScrollBorder:void 0,treeStickyScrollShadow:void 0},Qke={shadowColor:te(tC),borderColor:te(x_e),foregroundColor:te(L_e),backgroundColor:te(k_e),selectionForegroundColor:te(E_e),selectionBackgroundColor:te(I_e),selectionBorderColor:te(D_e),separatorColor:te(T_e),scrollbarShadow:te(eA),scrollbarSliderBackground:te(RJ),scrollbarSliderHoverBackground:te(AJ),scrollbarSliderActiveBackground:te(MJ)};function Jke(s,e){if(Vc)return!1;const t=eEe(s,e),i=s.getValue("window");return i?.menuStyle==="native"?!(!Ye&&!t):i?.menuStyle==="custom"?!1:t}function eEe(s,e){return e||(e=xie(s)),e==="native"}function xie(s){if(Vc)return"custom";const e=s.getValue("window");if(e){if(Ye&&e.nativeTabs===!0||Ye&&e.nativeFullScreen===!1)return"native";const n=e.titleBarStyle;if(n==="native"||n==="custom")return n}return"custom"}function tEe(s){if(Vc||Ye||xie(s)==="native")return"native";const t=s.getValue("window")?.controlsStyle;return t==="custom"||t==="hidden"?t:"native"}var SA=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},Zs=function(s,e){return function(t,i){e(t,i,s)}};function iEe(s,e){const t=[];return nEe(s,t),t}function nEe(s,e,t){const i=Vd.getInstance(),n=i.keyStatus.altKey||(Bn||vo)&&i.keyStatus.shiftKey;Eie(s,e,n,o=>o==="navigation")}function Lie(s,e,t,i){const n={primary:[],secondary:[]};return kie(s,n,e,t,i),n}function sEe(s,e,t,i){const n=[];return kie(s,n,e,t,i),n}function kie(s,e,t,i,n){Eie(s,e,!1,typeof t=="string"?r=>r===t:t,i,n)}function Eie(s,e,t,i=r=>r==="navigation",n=()=>!1,o=!1){let r,a;Array.isArray(e)?(r=e,a=e):(r=e.primary,a=e.secondary);const l=new Set;for(const[c,d]of s){let h;i(c)?(h=r,h.length>0&&o&&h.push(new tn)):(h=a,h.length>0&&h.push(new tn));for(let u of d){t&&(u=u instanceof yr&&u.alt?u.alt:u);const g=h.push(u);u instanceof ov&&l.add({group:c,action:u,index:g-1})}}for(const{group:c,action:d,index:h}of l){const u=i(c)?r:a,g=d.actions;n(d,c,u.length)&&u.splice(h,1,...g)}}let bg=class extends vv{constructor(e,t,i,n,o,r,a,l){super(void 0,e,{icon:!!(e.class||e.item.icon),label:!e.class&&!e.item.icon,draggable:t?.draggable,keybinding:t?.keybinding,hoverDelegate:t?.hoverDelegate,keybindingNotRenderedWithLabel:t?.keybindingNotRenderedWithLabel}),this._options=t,this._keybindingService=i,this._notificationService=n,this._contextKeyService=o,this._themeService=r,this._contextMenuService=a,this._accessibilityService=l,this._wantsAltCommand=!1,this._itemClassDispose=this._register(new vt),this._altKey=Vd.getInstance()}get _menuItemAction(){return this._action}get _commandAction(){return this._wantsAltCommand&&this._menuItemAction.alt||this._menuItemAction}async onClick(e){e.preventDefault(),e.stopPropagation();try{await this.actionRunner.run(this._commandAction,this._context)}catch(t){this._notificationService.error(t)}}render(e){if(super.render(e),e.classList.add("menu-entry"),this.options.icon&&this._updateItemClass(this._menuItemAction.item),this._menuItemAction.alt){let t=!1;const i=()=>{const n=!!this._menuItemAction.alt?.enabled&&(!this._accessibilityService.isMotionReduced()||t)&&(this._altKey.keyStatus.altKey||this._altKey.keyStatus.shiftKey&&t);n!==this._wantsAltCommand&&(this._wantsAltCommand=n,this.updateLabel(),this.updateTooltip(),this.updateClass())};this._register(this._altKey.event(i)),this._register(W(e,"mouseleave",n=>{t=!1,i()})),this._register(W(e,"mouseenter",n=>{t=!0,i()})),i()}}updateLabel(){this.options.label&&this.label&&(this.label.textContent=this._commandAction.label)}getTooltip(){const e=this._keybindingService.lookupKeybinding(this._commandAction.id,this._contextKeyService),t=e&&e.getLabel(),i=this._commandAction.tooltip||this._commandAction.label;let n=t?m(1644,"{0} ({1})",i,t):i;if(!this._wantsAltCommand&&this._menuItemAction.alt?.enabled){const o=this._menuItemAction.alt.tooltip||this._menuItemAction.alt.label,r=this._keybindingService.lookupKeybinding(this._menuItemAction.alt.id,this._contextKeyService),a=r&&r.getLabel(),l=a?m(1645,"{0} ({1})",o,a):o;n=m(1646,`{0}
[{1}] {2}`,n,z7.modifierLabels[Fo].altKey,l)}return n}updateClass(){this.options.icon&&(this._commandAction!==this._menuItemAction?this._menuItemAction.alt&&this._updateItemClass(this._menuItemAction.alt.item):this._updateItemClass(this._menuItemAction.item))}_updateItemClass(e){this._itemClassDispose.value=void 0;const{element:t,label:i}=this;if(!t||!i)return;const n=this._commandAction.checked&&Uke(e.toggled)&&e.toggled.icon?e.toggled.icon:e.icon;if(n)if(xe.isThemeIcon(n)){const o=xe.asClassNameArray(n);i.classList.add(...o),this._itemClassDispose.value=pe(()=>{i.classList.remove(...o)})}else i.style.backgroundImage=wh(this._themeService.getColorTheme().type)?Mc(n.dark):Mc(n.light),i.classList.add("icon"),this._itemClassDispose.value=Ra(pe(()=>{i.style.backgroundImage="",i.classList.remove("icon")}),this._themeService.onDidColorThemeChange(()=>{this.updateClass()}))}};bg=SA([Zs(2,ft),Zs(3,Ii),Zs(4,Ae),Zs(5,wi),Zs(6,Nr),Zs(7,Wn)],bg);class U7 extends bg{render(e){this.options.label=!0,this.options.icon=!1,super.render(e),e.classList.add("text-only"),e.classList.toggle("use-comma",this._options?.useComma??!1)}updateLabel(){const e=this._keybindingService.lookupKeybinding(this._action.id,this._contextKeyService);if(!e)return super.updateLabel();if(this.label){const t=U7._symbolPrintEnter(e);this._options?.conversational?this.label.textContent=m(1647,"{1} to {0}",this._action.label,t):this.label.textContent=m(1648,"{0} ({1})",this._action.label,t)}}static _symbolPrintEnter(e){return e.getLabel()?.replace(/\benter\b/gi,"⏎").replace(/\bEscape\b/gi,"Esc")}}let y4=class extends DI{constructor(e,t,i,n,o){const r={...t,menuAsChild:t?.menuAsChild??!1,classNames:t?.classNames??(xe.isThemeIcon(e.item.icon)?xe.asClassName(e.item.icon):void 0),keybindingProvider:t?.keybindingProvider??(a=>i.lookupKeybinding(a.id))};super(e,{getActions:()=>e.actions},n,r),this._keybindingService=i,this._contextMenuService=n,this._themeService=o}render(e){super.render(e),dt(this.element),e.classList.add("menu-entry");const t=this._action,{icon:i}=t.item;if(i&&!xe.isThemeIcon(i)){this.element.classList.add("icon");const n=()=>{this.element&&(this.element.style.backgroundImage=wh(this._themeService.getColorTheme().type)?Mc(i.dark):Mc(i.light))};n(),this._register(this._themeService.onDidColorThemeChange(()=>{n()}))}}};y4=SA([Zs(2,ft),Zs(3,Nr),Zs(4,wi)],y4);let S4=class extends ll{constructor(e,t,i,n,o,r,a,l){super(null,e),this._keybindingService=i,this._notificationService=n,this._contextMenuService=o,this._menuService=r,this._instaService=a,this._storageService=l,this._defaultActionDisposables=this._register(new U),this._container=null,this._options=t,this._storageKey=`${e.item.submenu.id}_lastActionId`;let c;const d=t?.togglePrimaryAction?l.get(this._storageKey,1):void 0;d&&(c=e.actions.find(u=>d===u.id)),c||(c=e.actions[0]),this._defaultAction=this._defaultActionDisposables.add(this._instaService.createInstance(bg,c,{keybinding:this._getDefaultActionKeybindingLabel(c)}));const h={keybindingProvider:u=>this._keybindingService.lookupKeybinding(u.id),...t,menuAsChild:t?.menuAsChild??!0,classNames:t?.classNames??["codicon","codicon-chevron-down"],actionRunner:t?.actionRunner??this._register(new Lm)};this._dropdown=this._register(new DI(e,e.actions,this._contextMenuService,h)),t?.togglePrimaryAction&&this._register(this._dropdown.actionRunner.onDidRun(u=>{u.action instanceof yr&&this.update(u.action)}))}update(e){this._options?.togglePrimaryAction&&this._storageService.store(this._storageKey,e.id,1,1),this._defaultActionDisposables.clear(),this._defaultAction=this._defaultActionDisposables.add(this._instaService.createInstance(bg,e,{keybinding:this._getDefaultActionKeybindingLabel(e)})),this._defaultAction.actionRunner=this._defaultActionDisposables.add(new class extends Lm{async runAction(t,i){await t.run(void 0)}}),this._container&&this._defaultAction.render(RR(this._container,ie(".action-container")))}_getDefaultActionKeybindingLabel(e){let t;if(this._options?.renderKeybindingWithDefaultActionLabel){const i=this._keybindingService.lookupKeybinding(e.id);i&&(t=`(${i.getLabel()})`)}return t}setActionContext(e){super.setActionContext(e),this._defaultAction.setActionContext(e),this._dropdown.setActionContext(e)}set actionRunner(e){super.actionRunner=e,this._defaultAction.actionRunner=e,this._dropdown.actionRunner=e}get actionRunner(){return super.actionRunner}render(e){this._container=e,super.render(this._container),this._container.classList.add("monaco-dropdown-with-default");const t=ie(".action-container");this._defaultAction.render(X(this._container,t)),this._register(W(t,se.KEY_DOWN,n=>{const o=new Pt(n);o.equals(17)&&(this._defaultAction.element.tabIndex=-1,this._dropdown.focus(),o.stopPropagation())}));const i=ie(".dropdown-action-container");this._dropdown.render(X(this._container,i)),this._register(W(i,se.KEY_DOWN,n=>{const o=new Pt(n);o.equals(15)&&(this._defaultAction.element.tabIndex=0,this._dropdown.setFocusable(!1),this._defaultAction.element?.focus(),o.stopPropagation())}))}focus(e){e?this._dropdown.focus():(this._defaultAction.element.tabIndex=0,this._defaultAction.element.focus())}blur(){this._defaultAction.element.tabIndex=-1,this._dropdown.blur(),this._container.blur()}setFocusable(e){e?this._defaultAction.element.tabIndex=0:(this._defaultAction.element.tabIndex=-1,this._dropdown.setFocusable(!1))}};S4=SA([Zs(2,ft),Zs(3,Ii),Zs(4,Nr),Zs(5,da),Zs(6,_e),Zs(7,Bs)],S4);let x4=class extends OLe{constructor(e,t,i){super(null,e,e.actions.map(n=>({text:n.id===tn.ID?"─────────":n.label,isDisabled:!n.enabled})),0,t,Xke,{ariaLabel:e.tooltip,optionsAsChildren:!0,useCustomDrawn:!Jke(i)}),this.select(Math.max(0,e.actions.findIndex(n=>n.checked)))}render(e){super.render(e),e.style.borderColor=te(BW)}runAction(e,t){const i=this.action.actions[t];i&&this.actionRunner.run(i)}};x4=SA([Zs(1,Rh),Zs(2,He)],x4);function q7(s,e,t){return e instanceof yr?s.createInstance(bg,e,t):e instanceof ap?e.item.isSelection?s.createInstance(x4,e):e.item.isSplitButton?s.createInstance(S4,e,{...t,togglePrimaryAction:typeof e.item.isSplitButton!="boolean"?e.item.isSplitButton.togglePrimaryAction:!1}):s.createInstance(y4,e,t):void 0}var xA=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},Ec=function(s,e){return function(t,i){e(t,i,s)}},ub,u1;let L4=class{constructor(e,t,i){this._commandService=e,this._keybindingService=t,this._hiddenStates=new k4(i)}createMenu(e,t,i){return new RI(e,this._hiddenStates,{emitEventsForSubmenuChanges:!1,eventDebounceDelay:50,...i},this._commandService,this._keybindingService,t)}getMenuActions(e,t,i){const n=new RI(e,this._hiddenStates,{emitEventsForSubmenuChanges:!1,eventDebounceDelay:50,...i},this._commandService,this._keybindingService,t),o=n.getActions(i);return n.dispose(),o}resetHiddenStates(e){this._hiddenStates.reset(e)}};L4=xA([Ec(0,Gt),Ec(1,ft),Ec(2,Bs)],L4);var kp;let k4=(kp=class{constructor(e){this._storageService=e,this._disposables=new U,this._onDidChange=new M,this.onDidChange=this._onDidChange.event,this._ignoreChangeEvent=!1,this._hiddenByDefaultCache=new Map;try{const t=e.get(ub._key,0,"{}");this._data=JSON.parse(t)}catch{this._data=Object.create(null)}this._disposables.add(e.onDidChangeValue(0,ub._key,this._disposables)(()=>{if(!this._ignoreChangeEvent)try{const t=e.get(ub._key,0,"{}");this._data=JSON.parse(t)}catch(t){console.log("FAILED to read storage after UPDATE",t)}this._onDidChange.fire()}))}dispose(){this._onDidChange.dispose(),this._disposables.dispose()}_isHiddenByDefault(e,t){return this._hiddenByDefaultCache.get(`${e.id}/${t}`)??!1}setDefaultState(e,t,i){this._hiddenByDefaultCache.set(`${e.id}/${t}`,i)}isHidden(e,t){const i=this._isHiddenByDefault(e,t),n=this._data[e.id]?.includes(t)??!1;return i?!n:n}updateHidden(e,t,i){this._isHiddenByDefault(e,t)&&(i=!i);const o=this._data[e.id];if(i)o?o.indexOf(t)<0&&o.push(t):this._data[e.id]=[t];else if(o){const r=o.indexOf(t);r>=0&&xhe(o,r),o.length===0&&delete this._data[e.id]}this._persist()}reset(e){if(e===void 0)this._data=Object.create(null),this._persist();else{for(const{id:t}of e)this._data[t]&&delete this._data[t];this._persist()}}_persist(){try{this._ignoreChangeEvent=!0;const e=JSON.stringify(this._data);this._storageService.store(ub._key,e,0,0)}finally{this._ignoreChangeEvent=!1}}},ub=kp,kp._key="menu.hiddenCommands",kp);k4=ub=xA([Ec(0,Bs)],k4);class Y1{constructor(e,t){this._id=e,this._collectContextKeysForSubmenus=t,this._menuGroups=[],this._allMenuIds=new Set,this._structureContextKeys=new Set,this._preconditionContextKeys=new Set,this._toggledContextKeys=new Set,this.refresh()}get allMenuIds(){return this._allMenuIds}get structureContextKeys(){return this._structureContextKeys}get preconditionContextKeys(){return this._preconditionContextKeys}get toggledContextKeys(){return this._toggledContextKeys}refresh(){this._menuGroups.length=0,this._allMenuIds.clear(),this._structureContextKeys.clear(),this._preconditionContextKeys.clear(),this._toggledContextKeys.clear();const e=this._sort(En.getMenuItems(this._id));let t;for(const i of e){const n=i.group||"";(!t||t[0]!==n)&&(t=[n,[]],this._menuGroups.push(t)),t[1].push(i),this._collectContextKeysAndSubmenuIds(i)}this._allMenuIds.add(this._id)}_sort(e){return e}_collectContextKeysAndSubmenuIds(e){if(Y1._fillInKbExprKeys(e.when,this._structureContextKeys),aw(e)){if(e.command.precondition&&Y1._fillInKbExprKeys(e.command.precondition,this._preconditionContextKeys),e.command.toggled){const t=e.command.toggled.condition||e.command.toggled;Y1._fillInKbExprKeys(t,this._toggledContextKeys)}}else this._collectContextKeysForSubmenus&&(En.getMenuItems(e.submenu).forEach(this._collectContextKeysAndSubmenuIds,this),this._allMenuIds.add(e.submenu))}static _fillInKbExprKeys(e,t){if(e)for(const i of e.keys())t.add(i)}}let E4=u1=class extends Y1{constructor(e,t,i,n,o,r){super(e,i),this._hiddenStates=t,this._commandService=n,this._keybindingService=o,this._contextKeyService=r,this.refresh()}createActionGroups(e){const t=[];for(const i of this._menuGroups){const[n,o]=i;let r;for(const a of o)if(this._contextKeyService.contextMatchesRules(a.when)){const l=aw(a);l&&this._hiddenStates.setDefaultState(this._id,a.command.id,!!a.isHiddenByDefault);const c=oEe(this._id,l?a.command:a,this._hiddenStates);if(l){const d=Iie(this._commandService,this._keybindingService,a.command.id,a.when);(r??=[]).push(new yr(a.command,a.alt,e,c,d,this._contextKeyService,this._commandService))}else{const d=new u1(a.submenu,this._hiddenStates,this._collectContextKeysForSubmenus,this._commandService,this._keybindingService,this._contextKeyService).createActionGroups(e),h=tn.join(...d.map(u=>u[1]));h.length>0&&(r??=[]).push(new ap(a,c,h))}}r&&r.length>0&&t.push([n,r])}return t}_sort(e){return e.sort(u1._compareMenuItems)}static _compareMenuItems(e,t){const i=e.group,n=t.group;if(i!==n){if(i){if(!n)return-1}else return 1;if(i==="navigation")return-1;if(n==="navigation")return 1;const a=i.localeCompare(n);if(a!==0)return a}const o=e.order||0,r=t.order||0;return o<r?-1:o>r?1:u1._compareTitles(aw(e)?e.command.title:e.title,aw(t)?t.command.title:t.title)}static _compareTitles(e,t){const i=typeof e=="string"?e:e.original,n=typeof t=="string"?t:t.original;return i.localeCompare(n)}};E4=u1=xA([Ec(3,Gt),Ec(4,ft),Ec(5,Ae)],E4);let RI=class{constructor(e,t,i,n,o,r){this._disposables=new U,this._menuInfo=new E4(e,t,i.emitEventsForSubmenuChanges,n,o,r);const a=new Rt(()=>{this._menuInfo.refresh(),this._onDidChange.fire({menu:this,isStructuralChange:!0,isEnablementChange:!0,isToggleChange:!0})},i.eventDebounceDelay);this._disposables.add(a),this._disposables.add(En.onDidChangeMenu(h=>{for(const u of this._menuInfo.allMenuIds)if(h.has(u)){a.schedule();break}}));const l=this._disposables.add(new U),c=h=>{let u=!1,g=!1,f=!1;for(const p of h)if(u=u||p.isStructuralChange,g=g||p.isEnablementChange,f=f||p.isToggleChange,u&&g&&f)break;return{menu:this,isStructuralChange:u,isEnablementChange:g,isToggleChange:f}},d=()=>{l.add(r.onDidChangeContext(h=>{const u=h.affectsSome(this._menuInfo.structureContextKeys),g=h.affectsSome(this._menuInfo.preconditionContextKeys),f=h.affectsSome(this._menuInfo.toggledContextKeys);(u||g||f)&&this._onDidChange.fire({menu:this,isStructuralChange:u,isEnablementChange:g,isToggleChange:f})})),l.add(t.onDidChange(h=>{this._onDidChange.fire({menu:this,isStructuralChange:!0,isEnablementChange:!1,isToggleChange:!1})}))};this._onDidChange=new qX({onWillAddFirstListener:d,onDidRemoveLastListener:l.clear.bind(l),delay:i.eventDebounceDelay,merge:c}),this.onDidChange=this._onDidChange.event}getActions(e){return this._menuInfo.createActionGroups(e)}dispose(){this._disposables.dispose(),this._onDidChange.dispose()}};RI=xA([Ec(3,Gt),Ec(4,ft),Ec(5,Ae)],RI);function oEe(s,e,t){const i=Vue(e)?e.submenu.id:e.id,n=typeof e.title=="string"?e.title:e.title.value,o=rp({id:`hide/${s.id}/${i}`,label:m(1651,"Hide '{0}'",n),run(){t.updateHidden(s,i,!0)}}),r=rp({id:`toggle/${s.id}/${i}`,label:n,get checked(){return!t.isHidden(s,i)},run(){t.updateHidden(s,i,!!this.checked)}});return{hide:o,toggle:r,get isHidden(){return!r.checked}}}function Iie(s,e,t,i=void 0,n=!0){return rp({id:`configureKeybinding/${t}`,label:m(1652,"Configure Keybinding"),enabled:n,run(){const r=!!!e.lookupKeybinding(t)&&i?i.serialize():void 0;s.executeCommand("workbench.action.openGlobalKeybindings",`@command:${t}`+(r?` +when:${r}`:""))}})}const Die=qe("IActionViewItemService");class rEe{constructor(){this._providers=new Map,this._onDidChange=new M,this.onDidChange=this._onDidChange.event}dispose(){this._onDidChange.dispose()}lookUp(e,t){return this._providers.get(this._makeKey(e,t))}_makeKey(e,t){return`${e.id}/${t instanceof fe?t.id:t}`}}tt(Die,rEe,1);var Tie=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},hr=function(s,e){return function(t,i){e(t,i,s)}};let AI=class extends Bke{constructor(e,t,i,n,o,r,a,l){super(e,o,{getKeyBinding:d=>r.lookupKeybinding(d.id)??void 0,...t,allowContextMenu:!0,skipTelemetry:typeof t?.telemetrySource=="string"}),this._options=t,this._menuService=i,this._contextKeyService=n,this._contextMenuService=o,this._keybindingService=r,this._commandService=a,this._sessionDisposables=this._store.add(new U);const c=t?.telemetrySource;c&&this._store.add(this.actionBar.onDidRun(d=>l.publicLog2("workbenchActionExecuted",{id:d.action.id,from:c})))}setActions(e,t=[],i){this._sessionDisposables.clear();const n=e.slice(),o=t.slice(),r=[];let a=0;const l=[];let c=!1;if(this._options?.hiddenItemStrategy!==-1)for(let d=0;d<n.length;d++){const h=n[d];!(h instanceof yr)&&!(h instanceof ap)||h.hideActions&&(r.push(h.hideActions.toggle),h.hideActions.toggle.checked&&a++,h.hideActions.isHidden&&(c=!0,n[d]=void 0,this._options?.hiddenItemStrategy!==0&&(l[d]=h)))}if(this._options?.overflowBehavior!==void 0){const d=gte(new Set(this._options.overflowBehavior.exempted),ot.map(n,g=>g?.id)),h=this._options.overflowBehavior.maxItems-d.size;let u=0;for(let g=0;g<n.length;g++){const f=n[g];f&&(u++,!d.has(f.id)&&u>=h&&(n[g]=void 0,l[g]=f))}}Pz(n),Pz(l),super.setActions(n,tn.join(l,o)),(r.length>0||n.length>0)&&this._sessionDisposables.add(W(this.getElement(),"contextmenu",d=>{const h=new is(be(this.getElement()),d),u=this.getItemAction(h.target);if(!u)return;h.preventDefault(),h.stopPropagation();const g=[];if(u instanceof yr&&u.menuKeybinding)g.push(u.menuKeybinding);else if(!(u instanceof ap||u instanceof FS)){const p=!!this._keybindingService.lookupKeybinding(u.id);g.push(Iie(this._commandService,this._keybindingService,u.id,void 0,p))}if(r.length>0){let p=!1;if(a===1&&this._options?.hiddenItemStrategy===0){p=!0;for(let _=0;_<r.length;_++)if(r[_].checked){r[_]=rp({id:u.id,label:u.label,checked:!0,enabled:!1,run(){}});break}}if(!p&&(u instanceof yr||u instanceof ap)){if(!u.hideActions)return;g.push(u.hideActions.hide)}else g.push(rp({id:"label",label:m(1649,"Hide"),enabled:!1,run(){}}))}const f=tn.join(g,r);this._options?.resetMenu&&!i&&(i=[this._options.resetMenu]),c&&i&&(f.push(new tn),f.push(rp({id:"resetThisMenu",label:m(1650,"Reset Menu"),run:()=>this._menuService.resetHiddenStates(i)}))),f.length!==0&&this._contextMenuService.showContextMenu({getAnchor:()=>h,getActions:()=>f,menuId:this._options?.contextMenu,menuActionOptions:{renderShortTitle:!0,...this._options?.menuOptions},skipTelemetry:typeof this._options?.telemetrySource=="string",contextKeyService:this._contextKeyService})}))}};AI=Tie([hr(2,da),hr(3,Ae),hr(4,Nr),hr(5,ft),hr(6,Gt),hr(7,vs)],AI);let BS=class extends AI{get onDidChangeMenuItems(){return this._onDidChangeMenuItems.event}constructor(e,t,i,n,o,r,a,l,c,d,h){super(e,{resetMenu:t,...i,actionViewItemProvider:(f,p)=>{let _=d.lookUp(t,f instanceof ap?f.item.submenu.id:f.id);_||(_=i?.actionViewItemProvider);const b=_?.(f,p,h,be(e).vscodeWindowId);return b||q7(h,f,p)}},n,o,r,a,l,c),this._onDidChangeMenuItems=this._store.add(new M);const u=this._store.add(n.createMenu(t,o,{emitEventsForSubmenuChanges:!0,eventDebounceDelay:i?.eventDebounceDelay})),g=()=>{const{primary:f,secondary:p}=Lie(u.getActions(i?.menuOptions),i?.toolbarOptions?.primaryGroup,i?.toolbarOptions?.shouldInlineSubmenu,i?.toolbarOptions?.useSeparatorsInPrimaryActions);e.classList.toggle("has-no-actions",f.length===0&&p.length===0),super.setActions(f,p)};this._store.add(u.onDidChange(()=>{g(),this._onDidChangeMenuItems.fire(this)})),this._store.add(d.onDidChange(f=>{f===t&&g()})),g()}setActions(){throw new Ce("This toolbar is populated from a menu.")}};BS=Tie([hr(3,da),hr(4,Ae),hr(5,Nr),hr(6,ft),hr(7,Gt),hr(8,vs),hr(9,Die),hr(10,_e)],BS);var aEe=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},ej=function(s,e){return function(t,i){e(t,i,s)}};const so=qe("hoverService");let Cv=class extends F{get delay(){return this.isInstantlyHovering()?0:this.hoverOptions?.dynamicDelay?e=>this.hoverOptions?.dynamicDelay?.(e)??this._delay:this._delay}constructor(e,t,i={},n,o){super(),this.placement=e,this.hoverOptions=t,this.overrideOptions=i,this.configurationService=n,this.hoverService=o,this.lastHoverHideTime=0,this.timeLimit=200,this.hoverDisposables=this._register(new U),this._delay=this.configurationService.getValue("workbench.hover.delay"),this._register(this.configurationService.onDidChangeConfiguration(r=>{r.affectsConfiguration("workbench.hover.delay")&&(this._delay=this.configurationService.getValue("workbench.hover.delay"))}))}showHover(e,t){const i=typeof this.overrideOptions=="function"?this.overrideOptions(e,t):this.overrideOptions;this.hoverDisposables.clear();const n=ki(e.target)?[e.target]:e.target.targetElements;for(const r of n)this.hoverDisposables.add(Fi(r,"keydown",a=>{a.equals(9)&&this.hoverService.hideHover()}));const o=ki(e.content)?void 0:typeof e.content=="string"?e.content.toString():e.content.value;return this.hoverService.showInstantHover({...e,...i,persistence:{hideOnKeyDown:!0,...i.persistence},id:o,appearance:{...e.appearance,compact:!0,skipFadeInAnimation:this.isInstantlyHovering(),...i.appearance}},t)}isInstantlyHovering(){return!!this.hoverOptions?.instantHover&&Date.now()-this.lastHoverHideTime<this.timeLimit}onDidHideHover(){this.hoverDisposables.clear(),this.hoverOptions?.instantHover&&(this.lastHoverHideTime=Date.now())}};Cv=aEe([ej(3,He),ej(4,so)],Cv);const lEe={showHover:function(){throw new Error("Native hover function not implemented.")},delay:0,showNativeHover:!0};class Fm extends JW{constructor(e){super(),this._textModel=e}getValueOfRange(e){return this._textModel.getValueInRange(e)}getLineLength(e){return this._textModel.getLineLength(e)}get length(){const e=this._textModel.getLineCount(),t=this._textModel.getLineLength(e);return new cn(e-1,t)}}class Nie extends Lm{constructor(e){super(),this._getContext=e}runAction(e,t){const i=this._getContext();return super.runAction(e,i)}}class cEe extends F{constructor(e,t,i){super(),this._editor=e,this._domNode=t,this.itemProvider=i,this.scrollTop=bt(this,this._editor.onDidScrollChange,r=>this._editor.getScrollTop()),this.isScrollTopZero=this.scrollTop.map(r=>r===0),this.modelAttached=bt(this,this._editor.onDidChangeModel,r=>this._editor.hasModel()),this.editorOnDidChangeViewZones=Po("onDidChangeViewZones",this._editor.onDidChangeViewZones),this.editorOnDidContentSizeChange=Po("onDidContentSizeChange",this._editor.onDidContentSizeChange),this.domNodeSizeChanged=Yr("domNodeSizeChanged"),this.views=new Map,this._domNode.className="gutter monaco-editor";const n=this._domNode.appendChild(ct("div.scroll-decoration",{role:"presentation",ariaHidden:"true",style:{width:"100%"}}).root),o=new ResizeObserver(()=>{Vt(r=>{this.domNodeSizeChanged.trigger(r)})});o.observe(this._domNode),this._register(pe(()=>o.disconnect())),this._register(De(r=>{n.className=this.isScrollTopZero.read(r)?"":"scroll-decoration"})),this._register(De(r=>this.render(r)))}dispose(){super.dispose(),vn(this._domNode)}render(e){if(!this.modelAttached.read(e))return;this.domNodeSizeChanged.read(e),this.editorOnDidChangeViewZones.read(e),this.editorOnDidContentSizeChange.read(e);const t=this.scrollTop.read(e),i=this._editor.getVisibleRanges(),n=new Set(this.views.keys()),o=me.ofStartAndLength(0,this._domNode.clientHeight);if(!o.isEmpty)for(const r of i){const a=new Ne(r.startLineNumber,r.endLineNumber+1),l=this.itemProvider.getIntersectingGutterItems(a,e);Vt(c=>{for(const d of l){if(!d.range.intersect(a))continue;n.delete(d.id);let h=this.views.get(d.id);if(h)h.item.set(d,c);else{const p=document.createElement("div");this._domNode.appendChild(p);const _=Re("item",d),b=this.itemProvider.createView(_,p);h=new dEe(_,b,p),this.views.set(d.id,h)}const u=d.range.startLineNumber<=this._editor.getModel().getLineCount()?this._editor.getTopForLineNumber(d.range.startLineNumber,!0)-t:d.range.startLineNumber>1?this._editor.getBottomForLineNumber(d.range.startLineNumber-1,!1)-t:0,f=(d.range.endLineNumberExclusive===1?Math.max(u,this._editor.getTopForLineNumber(d.range.startLineNumber,!1)-t):Math.max(u,this._editor.getBottomForLineNumber(d.range.endLineNumberExclusive-1,!0)-t))-u;h.domNode.style.top=`${u}px`,h.domNode.style.height=`${f}px`,h.gutterItemView.layout(me.ofStartAndLength(u,f),o)}})}for(const r of n){const a=this.views.get(r);a.gutterItemView.dispose(),a.domNode.remove(),this.views.delete(r)}}}class dEe{constructor(e,t,i){this.item=e,this.gutterItemView=t,this.domNode=i}}var Rie=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},Wk=function(s,e){return function(t,i){e(t,i,s)}};const HM=[],CL=35;let I4=class extends F{constructor(e,t,i,n,o,r,a,l,c){super(),this._diffModel=t,this._editors=i,this._options=n,this._sashLayout=o,this._boundarySashes=r,this._instantiationService=a,this._contextKeyService=l,this._menuService=c,this._menu=this._register(this._menuService.createMenu(fe.DiffEditorHunkToolbar,this._contextKeyService)),this._actions=bt(this,this._menu.onDidChange,()=>this._menu.getActions()),this._hasActions=this._actions.map(h=>h.length>0),this._showSash=j(this,h=>this._options.renderSideBySide.read(h)&&this._hasActions.read(h)),this.width=j(this,h=>this._hasActions.read(h)?CL:0),this.elements=ct("div.gutter@gutter",{style:{position:"absolute",height:"100%",width:CL+"px"}},[]),this._currentDiff=j(this,h=>{const u=this._diffModel.read(h);if(!u)return;const g=u.diff.read(h)?.mappings,f=this._editors.modifiedCursor.read(h);if(f)return g?.find(p=>p.lineRangeMapping.modified.contains(f.lineNumber))}),this._selectedDiffs=j(this,h=>{const g=this._diffModel.read(h)?.diff.read(h);if(!g)return HM;const f=this._editors.modifiedSelections.read(h);if(f.every(w=>w.isEmpty()))return HM;const p=new Kr(f.map(w=>Ne.fromRangeInclusive(w))),b=g.mappings.filter(w=>w.lineRangeMapping.innerChanges&&p.intersects(w.lineRangeMapping.modified)).map(w=>({mapping:w,rangeMappings:w.lineRangeMapping.innerChanges.filter(C=>f.some(v=>k.areIntersecting(C.modifiedRange,v)))}));return b.length===0||b.every(w=>w.rangeMappings.length===0)?HM:b}),this._register(FLe(e,this.elements.root)),this._register(W(this.elements.root,"click",()=>{this._editors.modified.focus()})),this._register(_g(this.elements.root,{display:this._hasActions.map(h=>h?"block":"none")})),Wr(this,h=>this._showSash.read(h)?new Cie(e,this._sashLayout.dimensions,this._options.enableSplitViewResizing,this._boundarySashes,oW(this,g=>this._sashLayout.sashLeft.read(g)-CL,(g,f)=>this._sashLayout.sashLeft.set(g+CL,f)),()=>this._sashLayout.resetSash()):void 0).recomputeInitiallyAndOnChange(this._store);const d=j(this,h=>{const u=this._diffModel.read(h);if(!u)return[];const g=u.diff.read(h);if(!g)return[];const f=this._selectedDiffs.read(h);if(f.length>0){const _=Lr.fromRangeMappings(f.flatMap(b=>b.rangeMappings));return[new tj(_,!0,fe.DiffEditorSelectionToolbar,void 0,u.model.original.uri,u.model.modified.uri)]}const p=this._currentDiff.read(h);return g.mappings.map(_=>new tj(_.lineRangeMapping.withInnerChangesFromLineRanges(),_.lineRangeMapping===p?.lineRangeMapping,fe.DiffEditorHunkToolbar,void 0,u.model.original.uri,u.model.modified.uri))});this._register(new cEe(this._editors.modified,this.elements.root,{getIntersectingGutterItems:(h,u)=>d.read(u),createView:(h,u)=>this._instantiationService.createInstance(D4,h,u,this)})),this._register(W(this.elements.gutter,se.MOUSE_WHEEL,h=>{this._editors.modified.getOption(117).handleMouseWheel&&this._editors.modified.delegateScrollFromMouseWheelEvent(h)},{passive:!1}))}computeStagedValue(e){const t=e.innerChanges??[],i=new Fm(this._editors.modifiedModel.get()),n=new Fm(this._editors.original.getModel());return new Na(t.map(a=>a.toTextEdit(i))).apply(n)}layout(e){this.elements.gutter.style.left=e+"px"}};I4=Rie([Wk(6,_e),Wk(7,Ae),Wk(8,da)],I4);class tj{constructor(e,t,i,n,o,r){this.mapping=e,this.showAlways=t,this.menuId=i,this.rangeOverride=n,this.originalUri=o,this.modifiedUri=r}get id(){return this.mapping.modified.toString()}get range(){return this.rangeOverride??this.mapping.modified}}let D4=class extends F{constructor(e,t,i,n){super(),this._item=e,this._elements=ct("div.gutterItem",{style:{height:"20px",width:"34px"}},[ct("div.background@background",{},[]),ct("div.buttons@buttons",{},[])]),this._showAlways=this._item.map(this,r=>r.showAlways),this._menuId=this._item.map(this,r=>r.menuId),this._isSmall=Re(this,!1),this._lastItemRange=void 0,this._lastViewRange=void 0;const o=this._register(n.createInstance(Cv,"element",{instantHover:!0},{position:{hoverPosition:1}}));this._register(Nb(t,this._elements.root)),this._register(De(r=>{const a=this._showAlways.read(r);this._elements.root.classList.toggle("noTransition",!0),this._elements.root.classList.toggle("showAlways",a),setTimeout(()=>{this._elements.root.classList.toggle("noTransition",!1)},0)})),this._register(ps((r,a)=>{this._elements.buttons.replaceChildren();const l=a.add(n.createInstance(BS,this._elements.buttons,this._menuId.read(r),{orientation:1,hoverDelegate:o,toolbarOptions:{primaryGroup:c=>c.startsWith("primary")},overflowBehavior:{maxItems:this._isSmall.read(r)?1:3},hiddenItemStrategy:0,actionRunner:a.add(new Nie(()=>{const c=this._item.read(void 0),d=c.mapping;return{mapping:d,originalWithModifiedChanges:i.computeStagedValue(d),originalUri:c.originalUri,modifiedUri:c.modifiedUri}})),menuOptions:{shouldForwardArgs:!0}}));a.add(l.onDidChangeMenuItems(()=>{this._lastItemRange&&this.layout(this._lastItemRange,this._lastViewRange)}))}))}layout(e,t){this._lastItemRange=e,this._lastViewRange=t;let i=this._elements.buttons.clientHeight;this._isSmall.set(this._item.get().mapping.original.startLineNumber===1&&e.length<30,void 0),i=this._elements.buttons.clientHeight;const n=e.length/2-i/2,o=i;let r=e.start+n;const a=me.tryCreate(o,t.endExclusive-o-i),l=me.tryCreate(e.start+o,e.endExclusive-i-o);l&&a&&l.start<l.endExclusive&&(r=a.clip(r),r=l.clip(r)),this._elements.buttons.style.top=`${r-e.start}px`}};D4=Rie([Wk(3,_e)],D4);var hEe=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},uEe=function(s,e){return function(t,i){e(t,i,s)}},T4,$u;let MI=($u=class extends F{static setBreadcrumbsSourceFactory(e){this._breadcrumbsSourceFactory.set(e,void 0)}get isUpdatingHiddenAreas(){return this._isUpdatingHiddenAreas}constructor(e,t,i,n){super(),this._editors=e,this._diffModel=t,this._options=i,this._instantiationService=n,this._modifiedOutlineSource=Wr(this,l=>{const c=this._editors.modifiedModel.read(l),d=T4._breadcrumbsSourceFactory.read(l);return!c||!d?void 0:d(c,this._instantiationService)}),this._isUpdatingHiddenAreas=!1,this._register(this._editors.original.onDidChangeCursorPosition(l=>{if(l.reason===1)return;const c=this._diffModel.get();Vt(d=>{for(const h of this._editors.original.getSelections()||[])c?.ensureOriginalLineIsVisible(h.getStartPosition().lineNumber,0,d),c?.ensureOriginalLineIsVisible(h.getEndPosition().lineNumber,0,d)})})),this._register(this._editors.modified.onDidChangeCursorPosition(l=>{if(l.reason===1)return;const c=this._diffModel.get();Vt(d=>{for(const h of this._editors.modified.getSelections()||[])c?.ensureModifiedLineIsVisible(h.getStartPosition().lineNumber,0,d),c?.ensureModifiedLineIsVisible(h.getEndPosition().lineNumber,0,d)})}));const o=this._diffModel.map((l,c)=>{const d=l?.unchangedRegions.read(c)??[];return d.length===1&&d[0].modifiedLineNumber===1&&d[0].lineCount===this._editors.modifiedModel.read(c)?.getLineCount()?[]:d});this.viewZones=j(this,l=>{const c=this._modifiedOutlineSource.read(l);if(!c)return{origViewZones:[],modViewZones:[]};const d=[],h=[],u=this._options.renderSideBySide.read(l),g=this._options.compactMode.read(l),f=o.read(l);for(let p=0;p<f.length;p++){const _=f[p];if(!_.shouldHideControls(l)&&!(g&&(p===0||p===f.length-1)))if(g){{const b=j(this,C=>_.getHiddenOriginalRange(C).startLineNumber-1),w=new Rb(b,12);d.push(w),l.store.add(new ij(this._editors.original,w,_,!u))}{const b=j(this,C=>_.getHiddenModifiedRange(C).startLineNumber-1),w=new Rb(b,12);h.push(w),l.store.add(new ij(this._editors.modified,w,_))}}else{{const b=j(this,C=>_.getHiddenOriginalRange(C).startLineNumber-1),w=new Rb(b,24);d.push(w),l.store.add(new nj(this._editors.original,w,_,_.originalUnchangedRange,!u,c,C=>this._diffModel.get().ensureModifiedLineIsVisible(C,2,void 0),this._options))}{const b=j(this,C=>_.getHiddenModifiedRange(C).startLineNumber-1),w=new Rb(b,24);h.push(w),l.store.add(new nj(this._editors.modified,w,_,_.modifiedUnchangedRange,!1,c,C=>this._diffModel.get().ensureModifiedLineIsVisible(C,2,void 0),this._options))}}}return{origViewZones:d,modViewZones:h}});const r={description:"unchanged lines",className:"diff-unchanged-lines",isWholeLine:!0},a={description:"Fold Unchanged",glyphMarginHoverMessage:new hs(void 0,{isTrusted:!0,supportThemeIcons:!0}).appendMarkdown(m(124,"Fold Unchanged Region")),glyphMarginClassName:"fold-unchanged "+xe.asClassName(Z.fold),zIndex:10001};this._register(CI(this._editors.original,j(this,l=>{const c=o.read(l),d=c.map(h=>({range:h.originalUnchangedRange.toInclusiveRange(),options:r}));for(const h of c)h.shouldHideControls(l)&&d.push({range:k.fromPositions(new A(h.originalLineNumber,1)),options:a});return d}))),this._register(CI(this._editors.modified,j(this,l=>{const c=o.read(l),d=c.map(h=>({range:h.modifiedUnchangedRange.toInclusiveRange(),options:r}));for(const h of c)h.shouldHideControls(l)&&d.push({range:Ne.ofLength(h.modifiedLineNumber,1).toInclusiveRange(),options:a});return d}))),this._register(De(l=>{const c=o.read(l);this._isUpdatingHiddenAreas=!0;try{this._editors.original.setHiddenAreas(c.map(d=>d.getHiddenOriginalRange(l).toInclusiveRange()).filter(Ln)),this._editors.modified.setHiddenAreas(c.map(d=>d.getHiddenModifiedRange(l).toInclusiveRange()).filter(Ln))}finally{this._isUpdatingHiddenAreas=!1}})),this._register(this._editors.modified.onMouseUp(l=>{if(!l.event.rightButton&&l.target.position&&l.target.element?.className.includes("fold-unchanged")){const c=l.target.position.lineNumber,d=this._diffModel.get();if(!d)return;const h=d.unchangedRegions.get().find(u=>u.modifiedUnchangedRange.contains(c));if(!h)return;h.collapseAll(void 0),l.event.stopPropagation(),l.event.preventDefault()}})),this._register(this._editors.original.onMouseUp(l=>{if(!l.event.rightButton&&l.target.position&&l.target.element?.className.includes("fold-unchanged")){const c=l.target.position.lineNumber,d=this._diffModel.get();if(!d)return;const h=d.unchangedRegions.get().find(u=>u.originalUnchangedRange.contains(c));if(!h)return;h.collapseAll(void 0),l.event.stopPropagation(),l.event.preventDefault()}}))}},T4=$u,$u._breadcrumbsSourceFactory=Re($u,()=>({dispose(){},getBreadcrumbItems(e,t){return[]}})),$u);MI=T4=hEe([uEe(3,_e)],MI);class ij extends M7{constructor(e,t,i,n=!1){const o=ct("div.diff-hidden-lines-widget");super(e,t,o.root),this._unchangedRegion=i,this._hide=n,this._nodes=ct("div.diff-hidden-lines-compact",[ct("div.line-left",[]),ct("div.text@text",[]),ct("div.line-right",[])]),o.root.appendChild(this._nodes.root),this._hide&&this._nodes.root.replaceChildren(),this._register(De(r=>{if(!this._hide){const a=this._unchangedRegion.getHiddenModifiedRange(r).length,l=m(125,"{0} hidden lines",a);this._nodes.text.innerText=l}}))}}class nj extends M7{constructor(e,t,i,n,o,r,a,l){const c=ct("div.diff-hidden-lines-widget");super(e,t,c.root),this._editor=e,this._unchangedRegion=i,this._unchangedRegionRange=n,this._hide=o,this._modifiedOutlineSource=r,this._revealModifiedHiddenLine=a,this._options=l,this._nodes=ct("div.diff-hidden-lines",[ct("div.top@top",{title:m(126,"Click or drag to show more above")}),ct("div.center@content",{style:{display:"flex"}},[ct("div@first",{style:{display:"flex",justifyContent:"center",alignItems:"center",flexShrink:"0"}},[ie("a",{title:m(127,"Show Unchanged Region"),role:"button",onclick:()=>{this._unchangedRegion.showAll(void 0)}},...Pu("$(unfold)"))]),ct("div@others",{style:{display:"flex",justifyContent:"center",alignItems:"center"}})]),ct("div.bottom@bottom",{title:m(128,"Click or drag to show more below"),role:"button"})]),c.root.appendChild(this._nodes.root),this._hide?vn(this._nodes.first):this._register(_g(this._nodes.first,{width:di(this._editor).layoutInfoContentLeft})),this._register(De(h=>{const u=this._unchangedRegion.visibleLineCountTop.read(h)+this._unchangedRegion.visibleLineCountBottom.read(h)===this._unchangedRegion.lineCount;this._nodes.bottom.classList.toggle("canMoveTop",!u),this._nodes.bottom.classList.toggle("canMoveBottom",this._unchangedRegion.visibleLineCountBottom.read(h)>0),this._nodes.top.classList.toggle("canMoveTop",this._unchangedRegion.visibleLineCountTop.read(h)>0),this._nodes.top.classList.toggle("canMoveBottom",!u);const g=this._unchangedRegion.isDragged.read(h),f=this._editor.getDomNode();f&&(f.classList.toggle("draggingUnchangedRegion",!!g),g==="top"?(f.classList.toggle("canMoveTop",this._unchangedRegion.visibleLineCountTop.read(h)>0),f.classList.toggle("canMoveBottom",!u)):g==="bottom"?(f.classList.toggle("canMoveTop",!u),f.classList.toggle("canMoveBottom",this._unchangedRegion.visibleLineCountBottom.read(h)>0)):(f.classList.toggle("canMoveTop",!1),f.classList.toggle("canMoveBottom",!1)))}));const d=this._editor;this._register(W(this._nodes.top,"mousedown",h=>{if(h.button!==0)return;this._nodes.top.classList.toggle("dragging",!0),this._nodes.root.classList.toggle("dragging",!0),h.preventDefault();const u=h.clientY;let g=!1;const f=this._unchangedRegion.visibleLineCountTop.get();this._unchangedRegion.isDragged.set("top",void 0);const p=be(this._nodes.top),_=W(p,"mousemove",w=>{const v=w.clientY-u;g=g||Math.abs(v)>2;const y=Math.round(v/d.getOption(75)),x=Math.max(0,Math.min(f+y,this._unchangedRegion.getMaxVisibleLineCountTop()));this._unchangedRegion.visibleLineCountTop.set(x,void 0)}),b=W(p,"mouseup",w=>{g||this._unchangedRegion.showMoreAbove(this._options.hideUnchangedRegionsRevealLineCount.get(),void 0),this._nodes.top.classList.toggle("dragging",!1),this._nodes.root.classList.toggle("dragging",!1),this._unchangedRegion.isDragged.set(void 0,void 0),_.dispose(),b.dispose()})})),this._register(W(this._nodes.bottom,"mousedown",h=>{if(h.button!==0)return;this._nodes.bottom.classList.toggle("dragging",!0),this._nodes.root.classList.toggle("dragging",!0),h.preventDefault();const u=h.clientY;let g=!1;const f=this._unchangedRegion.visibleLineCountBottom.get();this._unchangedRegion.isDragged.set("bottom",void 0);const p=be(this._nodes.bottom),_=W(p,"mousemove",w=>{const v=w.clientY-u;g=g||Math.abs(v)>2;const y=Math.round(v/d.getOption(75)),x=Math.max(0,Math.min(f-y,this._unchangedRegion.getMaxVisibleLineCountBottom())),L=this._unchangedRegionRange.endLineNumberExclusive>d.getModel().getLineCount()?d.getContentHeight():d.getTopForLineNumber(this._unchangedRegionRange.endLineNumberExclusive);this._unchangedRegion.visibleLineCountBottom.set(x,void 0);const E=this._unchangedRegionRange.endLineNumberExclusive>d.getModel().getLineCount()?d.getContentHeight():d.getTopForLineNumber(this._unchangedRegionRange.endLineNumberExclusive);d.setScrollTop(d.getScrollTop()+(E-L))}),b=W(p,"mouseup",w=>{if(this._unchangedRegion.isDragged.set(void 0,void 0),!g){const C=d.getTopForLineNumber(this._unchangedRegionRange.endLineNumberExclusive);this._unchangedRegion.showMoreBelow(this._options.hideUnchangedRegionsRevealLineCount.get(),void 0);const v=d.getTopForLineNumber(this._unchangedRegionRange.endLineNumberExclusive);d.setScrollTop(d.getScrollTop()+(v-C))}this._nodes.bottom.classList.toggle("dragging",!1),this._nodes.root.classList.toggle("dragging",!1),_.dispose(),b.dispose()})})),this._register(De(h=>{const u=[];if(!this._hide){const g=i.getHiddenModifiedRange(h).length,f=m(129,"{0} hidden lines",g),p=ie("span",{title:m(130,"Double click to unfold")},f);p.addEventListener("dblclick",w=>{w.button===0&&(w.preventDefault(),this._unchangedRegion.showAll(void 0))}),u.push(p);const _=this._unchangedRegion.getHiddenModifiedRange(h),b=this._modifiedOutlineSource.getBreadcrumbItems(_,h);if(b.length>0){u.push(ie("span",void 0,"  |  "));for(let w=0;w<b.length;w++){const C=b[w],v=iI.toIcon(C.kind),y=ct("div.breadcrumb-item",{style:{display:"flex",alignItems:"center"}},[Dl(v)," ",C.name,...w===b.length-1?[]:[Dl(Z.chevronRight)]]).root;u.push(y),y.onclick=()=>{this._revealModifiedHiddenLine(C.startLineNumber)}}}}vn(this._nodes.others,...u)}))}}const VM=[];class gEe extends F{constructor(e,t,i,n){super(),this._editors=e,this._diffModel=t,this._options=i,this._widget=n,this._selectedDiffs=j(this,o=>{const a=this._diffModel.read(o)?.diff.read(o);if(!a)return VM;const l=this._editors.modifiedSelections.read(o);if(l.every(u=>u.isEmpty()))return VM;const c=new Kr(l.map(u=>Ne.fromRangeInclusive(u))),h=a.mappings.filter(u=>u.lineRangeMapping.innerChanges&&c.intersects(u.lineRangeMapping.modified)).map(u=>({mapping:u,rangeMappings:u.lineRangeMapping.innerChanges.filter(g=>l.some(f=>k.areIntersecting(g.modifiedRange,f)))}));return h.length===0||h.every(u=>u.rangeMappings.length===0)?VM:h}),this._register(ps((o,r)=>{if(!this._options.shouldRenderOldRevertArrows.read(o))return;const a=this._diffModel.read(o),l=a?.diff.read(o);if(!a||!l||a.movedTextToCompare.read(o))return;const c=[],d=this._selectedDiffs.read(o),h=new Set(d.map(u=>u.mapping));if(d.length>0){const u=this._editors.modifiedSelections.read(o),g=r.add(new OI(u[u.length-1].positionLineNumber,this._widget,d.flatMap(f=>f.rangeMappings),!0));this._editors.modified.addGlyphMarginWidget(g),c.push(g)}for(const u of l.mappings)if(!h.has(u)&&!u.lineRangeMapping.modified.isEmpty&&u.lineRangeMapping.innerChanges){const g=r.add(new OI(u.lineRangeMapping.modified.startLineNumber,this._widget,u.lineRangeMapping,!1));this._editors.modified.addGlyphMarginWidget(g),c.push(g)}r.add(pe(()=>{for(const u of c)this._editors.modified.removeGlyphMarginWidget(u)}))}))}}const gN=class gN extends F{getId(){return this._id}constructor(e,t,i,n){super(),this._lineNumber=e,this._widget=t,this._diffs=i,this._revertSelection=n,this._id=`revertButton${gN.counter++}`,this._domNode=ct("div.revertButton",{title:this._revertSelection?m(135,"Revert Selected Changes"):m(136,"Revert Change")},[Dl(Z.arrowRight)]).root,this._register(W(this._domNode,se.MOUSE_DOWN,o=>{o.button!==2&&(o.stopPropagation(),o.preventDefault())})),this._register(W(this._domNode,se.MOUSE_UP,o=>{o.stopPropagation(),o.preventDefault()})),this._register(W(this._domNode,se.CLICK,o=>{this._diffs instanceof Ts?this._widget.revert(this._diffs):this._widget.revertRangeMappings(this._diffs),o.stopPropagation(),o.preventDefault()}))}getDomNode(){return this._domNode}getPosition(){return{lane:Ll.Right,range:{startColumn:1,startLineNumber:this._lineNumber,endColumn:1,endLineNumber:this._lineNumber},zIndex:10001}}};gN.counter=0;let OI=gN;var fEe=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},VC=function(s,e){return function(t,i){e(t,i,s)}};let Oc=class extends w4{get onDidContentSizeChange(){return this._editors.onDidContentSizeChange}constructor(e,t,i,n,o,r,a,l){super(),this._domElement=e,this._parentContextKeyService=n,this._parentInstantiationService=o,this._codeEditorService=r,this._accessibilitySignalService=a,this._editorProgressService=l,this.elements=ct("div.monaco-diff-editor.side-by-side",{style:{position:"relative",height:"100%"}},[ct("div.editor.original@original",{style:{position:"absolute",height:"100%"}}),ct("div.editor.modified@modified",{style:{position:"absolute",height:"100%"}}),ct("div.accessibleDiffViewer@accessibleDiffViewer",{style:{position:"absolute",height:"100%"}})]),this._diffModelSrc=this._register(lW(this,void 0)),this._diffModel=j(this,v=>this._diffModelSrc.read(v)?.object),this.onDidChangeModel=oe.fromObservableLight(this._diffModel),this._contextKeyService=this._register(this._parentContextKeyService.createScoped(this._domElement)),this._instantiationService=this._register(this._parentInstantiationService.createChild(new rC([Ae,this._contextKeyService]))),this._boundarySashes=Re(this,void 0),this._accessibleDiffViewerShouldBeVisible=Re(this,!1),this._accessibleDiffViewerVisible=j(this,v=>this._options.onlyShowAccessibleDiffViewer.read(v)?!0:this._accessibleDiffViewerShouldBeVisible.read(v)),this._movedBlocksLinesPart=Re(this,void 0),this._layoutInfo=j(this,v=>{const y=this._rootSizeObserver.width.read(v),x=this._rootSizeObserver.height.read(v);this._rootSizeObserver.automaticLayout?this.elements.root.style.height="100%":this.elements.root.style.height=x+"px";const L=this._sash.read(v),E=this._gutter.read(v),I=E?.width.read(v)??0,R=this._overviewRulerPart.read(v)?.width??0;let O,P,B,$,z;if(!!L){const J=L.sashLeft.read(v),ne=this._movedBlocksLinesPart.read(v)?.width.read(v)??0;O=0,P=J-I-ne,z=J-I,B=J,$=y-B-R}else{z=0;const J=this._options.inlineViewHideOriginalLineNumbers.read(v);O=I,J?P=0:P=Math.max(5,this._editors.originalObs.layoutInfoDecorationsLeft.read(v)),B=I+P,$=y-B-R}return this.elements.original.style.left=O+"px",this.elements.original.style.width=P+"px",this._editors.original.layout({width:P,height:x},!0),E?.layout(z),this.elements.modified.style.left=B+"px",this.elements.modified.style.width=$+"px",this._editors.modified.layout({width:$,height:x},!0),{modifiedEditor:this._editors.modified.getLayoutInfo(),originalEditor:this._editors.original.getLayoutInfo()}}),this._diffValue=this._diffModel.map((v,y)=>v?.diff.read(y)),this.onDidUpdateDiff=oe.fromObservableLight(this._diffValue),this._codeEditorService.willCreateDiffEditor(),this._contextKeyService.createKey("isInDiffEditor",!0),this._domElement.appendChild(this.elements.root),this._register(pe(()=>this.elements.root.remove())),this._rootSizeObserver=this._register(new hie(this.elements.root,t.dimension)),this._rootSizeObserver.setAutomaticLayout(t.automaticLayout??!1),this._options=this._instantiationService.createInstance(v4,t),this._register(De(v=>{this._options.setWidth(this._rootSizeObserver.width.read(v))})),this._contextKeyService.createKey(N.isEmbeddedDiffEditor.key,!1),this._register(jl(N.isEmbeddedDiffEditor,this._contextKeyService,v=>this._options.isInEmbeddedEditor.read(v))),this._register(jl(N.comparingMovedCode,this._contextKeyService,v=>!!this._diffModel.read(v)?.movedTextToCompare.read(v))),this._register(jl(N.diffEditorRenderSideBySideInlineBreakpointReached,this._contextKeyService,v=>this._options.couldShowInlineViewBecauseOfSize.read(v))),this._register(jl(N.diffEditorInlineMode,this._contextKeyService,v=>!this._options.renderSideBySide.read(v))),this._register(jl(N.hasChanges,this._contextKeyService,v=>(this._diffModel.read(v)?.diff.read(v)?.mappings.length??0)>0)),this._editors=this._register(this._instantiationService.createInstance(b4,this.elements.original,this.elements.modified,this._options,i,(v,y,x,L)=>this._createInnerEditor(v,y,x,L))),this._register(jl(N.diffEditorOriginalWritable,this._contextKeyService,v=>this._options.originalEditable.read(v))),this._register(jl(N.diffEditorModifiedWritable,this._contextKeyService,v=>!this._options.readOnly.read(v))),this._register(jl(N.diffEditorOriginalUri,this._contextKeyService,v=>this._diffModel.read(v)?.model.original.uri.toString()??"")),this._register(jl(N.diffEditorModifiedUri,this._contextKeyService,v=>this._diffModel.read(v)?.model.modified.uri.toString()??"")),this._overviewRulerPart=Wr(this,v=>this._options.renderOverviewRuler.read(v)?this._instantiationService.createInstance(fd(PS),this._editors,this.elements.root,this._diffModel,this._rootSizeObserver.width,this._rootSizeObserver.height,this._layoutInfo.map(y=>y.modifiedEditor)):void 0).recomputeInitiallyAndOnChange(this._store);const c={height:this._rootSizeObserver.height,width:this._rootSizeObserver.width.map((v,y)=>v-(this._overviewRulerPart.read(y)?.width??0))};this._sashLayout=new Dke(this._options,c),this._sash=Wr(this,v=>{const y=this._options.renderSideBySide.read(v);return this.elements.root.classList.toggle("side-by-side",y),y?new Cie(this.elements.root,c,this._options.enableSplitViewResizing,this._boundarySashes,this._sashLayout.sashLeft,()=>this._sashLayout.resetSash()):void 0}).recomputeInitiallyAndOnChange(this._store);const d=Wr(this,v=>this._instantiationService.createInstance(fd(MI),this._editors,this._diffModel,this._options)).recomputeInitiallyAndOnChange(this._store);Wr(this,v=>this._instantiationService.createInstance(fd(vke),this._editors,this._diffModel,this._options,this)).recomputeInitiallyAndOnChange(this._store);const h=new Set,u=new Set;let g=!1;const f=Wr(this,v=>this._instantiationService.createInstance(fd(m4),be(this._domElement),this._editors,this._diffModel,this._options,this,()=>g||d.read(void 0).isUpdatingHiddenAreas,h,u)).recomputeInitiallyAndOnChange(this._store),p=j(this,v=>{const y=f.read(v).viewZones.read(v).orig,x=d.read(v).viewZones.read(v).origViewZones;return y.concat(x)}),_=j(this,v=>{const y=f.read(v).viewZones.read(v).mod,x=d.read(v).viewZones.read(v).modViewZones;return y.concat(x)});this._register(yI(this._editors.original,p,v=>{g=v},h));let b;this._register(yI(this._editors.modified,_,v=>{g=v,g?b=Il.capture(this._editors.modified):(b?.restore(this._editors.modified),b=void 0)},u)),this._accessibleDiffViewer=Wr(this,v=>this._instantiationService.createInstance(fd(Zf),this.elements.accessibleDiffViewer,this._accessibleDiffViewerVisible,(y,x)=>this._accessibleDiffViewerShouldBeVisible.set(y,x),this._options.onlyShowAccessibleDiffViewer.map(y=>!y),this._rootSizeObserver.width,this._rootSizeObserver.height,this._diffModel.map((y,x)=>y?.diff.read(x)?.mappings.map(L=>L.lineRangeMapping)),new ske(this._editors))).recomputeInitiallyAndOnChange(this._store);const w=this._accessibleDiffViewerVisible.map(v=>v?"hidden":"visible");this._register(_g(this.elements.modified,{visibility:w})),this._register(_g(this.elements.original,{visibility:w})),this._createDiffEditorContributions(),this._codeEditorService.addDiffEditor(this),this._register(pe(()=>{this._codeEditorService.removeDiffEditor(this)})),this._gutter=Wr(this,v=>this._options.shouldRenderGutterMenu.read(v)?this._instantiationService.createInstance(fd(I4),this.elements.root,this._diffModel,this._editors,this._options,this._sashLayout,this._boundarySashes):void 0),this._register(sv(this._layoutInfo)),Wr(this,v=>new(fd(Cw))(this.elements.root,this._diffModel,this._layoutInfo.map(y=>y.originalEditor),this._layoutInfo.map(y=>y.modifiedEditor),this._editors)).recomputeInitiallyAndOnChange(this._store,v=>{this._movedBlocksLinesPart.set(v,void 0)}),this._register(oe.runAndSubscribe(this._editors.modified.onDidChangeCursorPosition,v=>this._handleCursorPositionChange(v,!0))),this._register(oe.runAndSubscribe(this._editors.original.onDidChangeCursorPosition,v=>this._handleCursorPositionChange(v,!1)));const C=this._diffModel.map(this,(v,y)=>{if(v)return v.diff.read(y)===void 0&&!v.isDiffUpToDate.read(y)});this._register(ps((v,y)=>{if(C.read(v)===!0){const x=this._editorProgressService.show(!0,1e3);y.add(pe(()=>x.done()))}})),this._register(ps((v,y)=>{y.add(new(fd(gEe))(this._editors,this._diffModel,this._options,this))})),this._register(ps((v,y)=>{const x=this._diffModel.read(v);if(x)for(const L of[x.model.original,x.model.modified])y.add(L.onWillDispose(E=>{Oe(new Ce("TextModel got disposed before DiffEditorWidget model got reset")),this.setModel(null)}))})),this._register(De(v=>{this._options.setModel(this._diffModel.read(v))}))}_createInnerEditor(e,t,i,n){return e.createInstance(Am,t,i,n)}_createDiffEditorContributions(){const e=lw.getDiffEditorContributions();for(const t of e)try{this._register(this._instantiationService.createInstance(t.ctor,this))}catch(i){Oe(i)}}get _targetEditor(){return this._editors.modified}getEditorType(){return sx.IDiffEditor}layout(e){this._rootSizeObserver.observe(e)}hasTextFocus(){return this._editors.original.hasTextFocus()||this._editors.modified.hasTextFocus()}saveViewState(){const e=this._editors.original.saveViewState(),t=this._editors.modified.saveViewState();return{original:e,modified:t,modelState:this._diffModel.get()?.serializeState()}}restoreViewState(e){if(e&&e.original&&e.modified){const t=e;this._editors.original.restoreViewState(t.original),this._editors.modified.restoreViewState(t.modified),t.modelState&&this._diffModel.get()?.restoreSerializedState(t.modelState)}}handleInitialized(){this._editors.original.handleInitialized(),this._editors.modified.handleInitialized()}createViewModel(e){return this._instantiationService.createInstance(p4,e,this._options)}getModel(){return this._diffModel.get()?.model??null}setModel(e){const t=e?"model"in e?SI.create(e).createNewRef(this):SI.create(this.createViewModel(e),this):null;this.setDiffModel(t)}setDiffModel(e,t){const i=this._diffModel.get();!e&&i&&this._accessibleDiffViewer.get().close(),this._diffModel.get()!==e?.object&&nv(t,n=>{const o=e?.object;bt.batchEventsGlobally(n,()=>{this._editors.original.setModel(o?o.model.original:null),this._editors.modified.setModel(o?o.model.modified:null)});const r=this._diffModelSrc.get()?.createNewRef(this);this._diffModelSrc.set(e?.createNewRef(this),n),setTimeout(()=>{r?.dispose()},0)})}updateOptions(e){this._options.updateOptions(e)}getContainerDomNode(){return this._domElement}getOriginalEditor(){return this._editors.original}getModifiedEditor(){return this._editors.modified}getLineChanges(){const e=this._diffModel.get()?.diff.get();return e?pEe(e):null}getDiffComputationResult(){const e=this._diffModel.get()?.diff.get();return e?{changes:this.getLineChanges(),changes2:e.mappings.map(t=>t.lineRangeMapping),identical:e.identical,quitEarly:e.quitEarly}:null}revert(e){const t=this._diffModel.get();!t||!t.isDiffUpToDate.get()||(this._editors.modified.pushUndoStop(),this._editors.modified.executeEdits("diffEditor",[{range:e.modified.toExclusiveRange(),text:t.model.original.getValueInRange(e.original.toExclusiveRange())}]),this._editors.modified.pushUndoStop())}revertRangeMappings(e){const t=this._diffModel.get();if(!t||!t.isDiffUpToDate.get())return;const i=e.map(n=>({range:n.modifiedRange,text:t.model.original.getValueInRange(n.originalRange)}));this._editors.modified.pushUndoStop(),this._editors.modified.executeEdits("diffEditor",i),this._editors.modified.pushUndoStop()}revertFocusedRangeMappings(){const e=this._diffModel.get();if(!e||!e.isDiffUpToDate.get())return;const t=this._diffModel.get()?.diff.get()?.mappings;if(!t||t.length===0)return;const i=this._editors.modified;if(!i.hasTextFocus())return;const n=i.getPosition().lineNumber,o=i.getSelection(),r=Ne.fromRange(o||new k(n,0,n,0)),a=t.filter(l=>l.lineRangeMapping.modified.intersect(r));i.pushUndoStop(),i.executeEdits("diffEditor",a.map(l=>({range:l.lineRangeMapping.modified.toExclusiveRange(),text:e.model.original.getValueInRange(l.lineRangeMapping.original.toExclusiveRange())}))),i.pushUndoStop()}_goTo(e){this._editors.modified.setPosition(new A(e.lineRangeMapping.modified.startLineNumber,1)),this._editors.modified.revealRangeInCenter(e.lineRangeMapping.modified.toExclusiveRange())}goToDiff(e){const t=this._diffModel.get()?.diff.get()?.mappings;if(!t||t.length===0)return;const i=this._editors.modified.getPosition().lineNumber;let n;e==="next"?this._editors.modified.getModel().getLineCount()===i?n=t[0]:n=t.find(r=>r.lineRangeMapping.modified.startLineNumber>i)??t[0]:n=gS(t,o=>o.lineRangeMapping.modified.startLineNumber<i)??t[t.length-1],this._goTo(n),n.lineRangeMapping.modified.isEmpty?this._accessibilitySignalService.playSignal(Gs.diffLineDeleted,{source:"diffEditor.goToDiff"}):n.lineRangeMapping.original.isEmpty?this._accessibilitySignalService.playSignal(Gs.diffLineInserted,{source:"diffEditor.goToDiff"}):n&&this._accessibilitySignalService.playSignal(Gs.diffLineModified,{source:"diffEditor.goToDiff"})}revealFirstDiff(){const e=this._diffModel.get();e&&this.waitForDiff().then(()=>{const t=e.diff.get()?.mappings;!t||t.length===0||this._goTo(t[0])})}accessibleDiffViewerNext(){this._accessibleDiffViewer.get().next()}accessibleDiffViewerPrev(){this._accessibleDiffViewer.get().prev()}async waitForDiff(){const e=this._diffModel.get();e&&await e.waitForDiff()}mapToOtherSide(){const e=this._editors.modified.hasWidgetFocus(),t=e?this._editors.modified:this._editors.original,i=e?this._editors.original:this._editors.modified;let n;const o=t.getSelection();if(o){const r=this._diffModel.get()?.diff.get()?.mappings.map(a=>e?a.lineRangeMapping.flip():a.lineRangeMapping);if(r){const a=Aq(o.getStartPosition(),r),l=Aq(o.getEndPosition(),r);n=k.plusRange(a,l)}}return{destination:i,destinationSelection:n}}switchSide(){const{destination:e,destinationSelection:t}=this.mapToOtherSide();e.focus(),t&&e.setSelection(t)}exitCompareMove(){const e=this._diffModel.get();e&&e.movedTextToCompare.set(void 0,void 0)}collapseAllUnchangedRegions(){const e=this._diffModel.get()?.unchangedRegions.get();e&&Vt(t=>{for(const i of e)i.collapseAll(t)})}showAllUnchangedRegions(){const e=this._diffModel.get()?.unchangedRegions.get();e&&Vt(t=>{for(const i of e)i.showAll(t)})}_handleCursorPositionChange(e,t){if(e?.reason===3){const i=this._diffModel.get()?.diff.get()?.mappings.find(n=>t?n.lineRangeMapping.modified.contains(e.position.lineNumber):n.lineRangeMapping.original.contains(e.position.lineNumber));i?.lineRangeMapping.modified.isEmpty?this._accessibilitySignalService.playSignal(Gs.diffLineDeleted,{source:"diffEditor.cursorPositionChanged"}):i?.lineRangeMapping.original.isEmpty?this._accessibilitySignalService.playSignal(Gs.diffLineInserted,{source:"diffEditor.cursorPositionChanged"}):i&&this._accessibilitySignalService.playSignal(Gs.diffLineModified,{source:"diffEditor.cursorPositionChanged"})}}};Oc=fEe([VC(3,Ae),VC(4,_e),VC(5,ht),VC(6,Th),VC(7,Ch)],Oc);function pEe(s){return s.mappings.map(e=>{const t=e.lineRangeMapping;let i,n,o,r,a=t.innerChanges;return t.original.isEmpty?(i=t.original.startLineNumber-1,n=0,a=void 0):(i=t.original.startLineNumber,n=t.original.endLineNumberExclusive-1),t.modified.isEmpty?(o=t.modified.startLineNumber-1,r=0,a=void 0):(o=t.modified.startLineNumber,r=t.modified.endLineNumberExclusive-1),{originalStartLineNumber:i,originalEndLineNumber:n,modifiedStartLineNumber:o,modifiedEndLineNumber:r,charChanges:a?.map(l=>({originalStartLineNumber:l.originalRange.startLineNumber,originalStartColumn:l.originalRange.startColumn,originalEndLineNumber:l.originalRange.endLineNumber,originalEndColumn:l.originalRange.endColumn,modifiedStartLineNumber:l.modifiedRange.startLineNumber,modifiedStartColumn:l.modifiedRange.startColumn,modifiedEndLineNumber:l.modifiedRange.endLineNumber,modifiedEndColumn:l.modifiedRange.endColumn}))}})}class mEe extends Tn{constructor(){super({id:"diffEditor.toggleCollapseUnchangedRegions",title:V(85,"Toggle Collapse Unchanged Regions"),icon:Z.map,toggled:K.has("config.diffEditor.hideUnchangedRegions.enabled"),precondition:K.has("isInDiffEditor"),menu:{when:K.has("isInDiffEditor"),id:fe.EditorTitle,order:22,group:"navigation"}})}run(e,...t){const i=e.get(He),n=!i.getValue("diffEditor.hideUnchangedRegions.enabled");i.updateValue("diffEditor.hideUnchangedRegions.enabled",n)}}class Aie extends Tn{constructor(){super({id:"diffEditor.toggleShowMovedCodeBlocks",title:V(86,"Toggle Show Moved Code Blocks"),precondition:K.has("isInDiffEditor")})}run(e,...t){const i=e.get(He),n=!i.getValue("diffEditor.experimental.showMoves");i.updateValue("diffEditor.experimental.showMoves",n)}}class Mie extends Tn{constructor(){super({id:"diffEditor.toggleUseInlineViewWhenSpaceIsLimited",title:V(87,"Toggle Use Inline View When Space Is Limited"),precondition:K.has("isInDiffEditor")})}run(e,...t){const i=e.get(He),n=!i.getValue("diffEditor.useInlineViewWhenSpaceIsLimited");i.updateValue("diffEditor.useInlineViewWhenSpaceIsLimited",n)}}const gx=V(88,"Diff Editor");class _Ee extends za{constructor(){super({id:"diffEditor.switchSide",title:V(89,"Switch Side"),icon:Z.arrowSwap,precondition:K.has("isInDiffEditor"),f1:!0,category:gx})}runEditorCommand(e,t,i){const n=p_(e);if(n instanceof Oc){if(i&&i.dryRun)return{destinationSelection:n.mapToOtherSide().destinationSelection};n.switchSide()}}}class bEe extends za{constructor(){super({id:"diffEditor.exitCompareMove",title:V(90,"Exit Compare Move"),icon:Z.close,precondition:N.comparingMovedCode,f1:!1,category:gx,keybinding:{weight:1e4,primary:9}})}runEditorCommand(e,t,...i){const n=p_(e);n instanceof Oc&&n.exitCompareMove()}}class wEe extends za{constructor(){super({id:"diffEditor.collapseAllUnchangedRegions",title:V(91,"Collapse All Unchanged Regions"),icon:Z.fold,precondition:K.has("isInDiffEditor"),f1:!0,category:gx})}runEditorCommand(e,t,...i){const n=p_(e);n instanceof Oc&&n.collapseAllUnchangedRegions()}}class vEe extends za{constructor(){super({id:"diffEditor.showAllUnchangedRegions",title:V(92,"Show All Unchanged Regions"),icon:Z.unfold,precondition:K.has("isInDiffEditor"),f1:!0,category:gx})}runEditorCommand(e,t,...i){const n=p_(e);n instanceof Oc&&n.showAllUnchangedRegions()}}class N4 extends Tn{constructor(){super({id:"diffEditor.revert",title:V(93,"Revert"),f1:!0,category:gx,precondition:K.has("isInDiffEditor")})}run(e,t){return t?this.runViaToolbarContext(e,t):this.runViaCursorOrSelection(e)}runViaCursorOrSelection(e){const t=p_(e);t instanceof Oc&&t.revertFocusedRangeMappings()}runViaToolbarContext(e,t){const i=CEe(e,t.originalUri,t.modifiedUri);i instanceof Oc&&i.revertRangeMappings(t.mapping.innerChanges??[])}}const Oie=V(94,"Accessible Diff Viewer"),fN=class fN extends Tn{constructor(){super({id:fN.id,title:V(95,"Go to Next Difference"),category:Oie,precondition:K.has("isInDiffEditor"),keybinding:{primary:65,weight:100},f1:!0})}run(e){p_(e)?.accessibleDiffViewerNext()}};fN.id="editor.action.accessibleDiffViewer.next";let WS=fN;const pN=class pN extends Tn{constructor(){super({id:pN.id,title:V(96,"Go to Previous Difference"),category:Oie,precondition:K.has("isInDiffEditor"),keybinding:{primary:1089,weight:100},f1:!0})}run(e){p_(e)?.accessibleDiffViewerPrev()}};pN.id="editor.action.accessibleDiffViewer.prev";let PI=pN;function CEe(s,e,t){return s.get(ht).listDiffEditors().find(o=>{const r=o.getModifiedEditor(),a=o.getOriginalEditor();return r&&r.getModel()?.uri.toString()===t.toString()&&a&&a.getModel()?.uri.toString()===e.toString()})||null}function p_(s){const t=s.get(ht).listDiffEditors(),i=ln();if(i){for(const n of t)if(n.getContainerDomNode().contains(i))return n}return null}It(mEe);It(Aie);It(Mie);En.appendMenuItem(fe.EditorTitle,{command:{id:new Mie().desc.id,title:m(119,"Use Inline View When Space Is Limited"),toggled:K.has("config.diffEditor.useInlineViewWhenSpaceIsLimited"),precondition:K.has("isInDiffEditor")},order:11,group:"1_diff",when:K.and(N.diffEditorRenderSideBySideInlineBreakpointReached,K.has("isInDiffEditor"))});En.appendMenuItem(fe.EditorTitle,{command:{id:new Aie().desc.id,title:m(120,"Show Moved Code Blocks"),icon:Z.move,toggled:Kv.create("config.diffEditor.experimental.showMoves",!0),precondition:K.has("isInDiffEditor")},order:10,group:"1_diff",when:K.has("isInDiffEditor")});It(N4);for(const s of[{icon:Z.arrowRight,key:N.diffEditorInlineMode.toNegated()},{icon:Z.discard,key:N.diffEditorInlineMode}])En.appendMenuItem(fe.DiffEditorHunkToolbar,{command:{id:new N4().desc.id,title:m(121,"Revert Block"),icon:s.icon},when:K.and(N.diffEditorModifiedWritable,s.key),order:5,group:"primary"}),En.appendMenuItem(fe.DiffEditorSelectionToolbar,{command:{id:new N4().desc.id,title:m(122,"Revert Selection"),icon:s.icon},when:K.and(N.diffEditorModifiedWritable,s.key),order:5,group:"primary"});It(_Ee);It(bEe);It(wEe);It(vEe);En.appendMenuItem(fe.EditorTitle,{command:{id:WS.id,title:m(123,"Open Accessible Diff Viewer"),precondition:K.has("isInDiffEditor")},order:10,group:"2_diff",when:K.and(N.accessibleDiffViewerVisible.negate(),K.has("isInDiffEditor"))});at.registerCommandAlias("editor.action.diffReview.next",WS.id);It(WS);at.registerCommandAlias("editor.action.diffReview.prev",PI.id);It(PI);var yEe=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},SEe=function(s,e){return function(t,i){e(t,i,s)}},R4;const LA=new ce("selectionAnchorSet",!1);var Ep;let wg=(Ep=class{static get(e){return e.getContribution(R4.ID)}constructor(e,t){this.editor=e,this.selectionAnchorSetContextKey=LA.bindTo(t),this.modelChangeListener=e.onDidChangeModel(()=>this.selectionAnchorSetContextKey.reset())}setSelectionAnchor(){if(this.editor.hasModel()){const e=this.editor.getPosition();this.editor.changeDecorations(t=>{this.decorationId&&t.removeDecoration(this.decorationId),this.decorationId=t.addDecoration(he.fromPositions(e,e),{description:"selection-anchor",stickiness:1,hoverMessage:new hs().appendText(m(798,"Selection Anchor")),className:"selection-anchor"})}),this.selectionAnchorSetContextKey.set(!!this.decorationId),Js(m(799,"Anchor set at {0}:{1}",e.lineNumber,e.column))}}goToSelectionAnchor(){if(this.editor.hasModel()&&this.decorationId){const e=this.editor.getModel().getDecorationRange(this.decorationId);e&&this.editor.setPosition(e.getStartPosition())}}selectFromAnchorToCursor(){if(this.editor.hasModel()&&this.decorationId){const e=this.editor.getModel().getDecorationRange(this.decorationId);if(e){const t=this.editor.getPosition();this.editor.setSelection(he.fromPositions(e.getStartPosition(),t)),this.cancelSelectionAnchor()}}}cancelSelectionAnchor(){if(this.decorationId){const e=this.decorationId;this.editor.changeDecorations(t=>{t.removeDecoration(e),this.decorationId=void 0}),this.selectionAnchorSetContextKey.set(!1)}}dispose(){this.cancelSelectionAnchor(),this.modelChangeListener.dispose()}},R4=Ep,Ep.ID="editor.contrib.selectionAnchorController",Ep);wg=R4=yEe([SEe(1,Ae)],wg);class xEe extends ue{constructor(){super({id:"editor.action.setSelectionAnchor",label:V(800,"Set Selection Anchor"),precondition:void 0,kbOpts:{kbExpr:N.editorTextFocus,primary:Gi(2089,2080),weight:100}})}async run(e,t){wg.get(t)?.setSelectionAnchor()}}class LEe extends ue{constructor(){super({id:"editor.action.goToSelectionAnchor",label:V(801,"Go to Selection Anchor"),precondition:LA})}async run(e,t){wg.get(t)?.goToSelectionAnchor()}}class kEe extends ue{constructor(){super({id:"editor.action.selectFromAnchorToCursor",label:V(802,"Select from Anchor to Cursor"),precondition:LA,kbOpts:{kbExpr:N.editorTextFocus,primary:Gi(2089,2089),weight:100}})}async run(e,t){wg.get(t)?.selectFromAnchorToCursor()}}class EEe extends ue{constructor(){super({id:"editor.action.cancelSelectionAnchor",label:V(803,"Cancel Selection Anchor"),precondition:LA,kbOpts:{kbExpr:N.editorTextFocus,primary:9,weight:100}})}async run(e,t){wg.get(t)?.cancelSelectionAnchor()}}lt(wg.ID,wg,4);ae(xEe);ae(LEe);ae(kEe);ae(EEe);const IEe=T("editorOverviewRuler.bracketMatchForeground","#A0A0A0",m(804,"Overview ruler marker color for matching brackets."));class DEe extends ue{constructor(){super({id:"editor.action.jumpToBracket",label:V(806,"Go to Bracket"),precondition:void 0,kbOpts:{kbExpr:N.editorTextFocus,primary:3165,weight:100}})}run(e,t){Bm.get(t)?.jumpToBracket()}}class TEe extends ue{constructor(){super({id:"editor.action.selectToBracket",label:V(807,"Select to Bracket"),precondition:void 0,metadata:{description:V(808,"Select the text inside and including the brackets or curly braces"),args:[{name:"args",schema:{type:"object",properties:{selectBrackets:{type:"boolean",default:!0}}}}]}})}run(e,t,i){let n=!0;i&&i.selectBrackets===!1&&(n=!1),Bm.get(t)?.selectToBracket(n)}}class NEe extends ue{constructor(){super({id:"editor.action.removeBrackets",label:V(809,"Remove Brackets"),precondition:void 0,kbOpts:{kbExpr:N.editorTextFocus,primary:2561,weight:100},canTriggerInlineEdits:!0})}run(e,t){Bm.get(t)?.removeBrackets(this.id)}}class REe{constructor(e,t,i){this.position=e,this.brackets=t,this.options=i}}const Jh=class Jh extends F{static get(e){return e.getContribution(Jh.ID)}constructor(e){super(),this._editor=e,this._lastBracketsData=[],this._lastVersionId=0,this._decorations=this._editor.createDecorationsCollection(),this._updateBracketsSoon=this._register(new Rt(()=>this._updateBrackets(),50)),this._matchBrackets=this._editor.getOption(80),this._updateBracketsSoon.schedule(),this._register(e.onDidChangeCursorPosition(t=>{this._matchBrackets!=="never"&&this._updateBracketsSoon.schedule()})),this._register(e.onDidChangeModelContent(t=>{this._updateBracketsSoon.schedule()})),this._register(e.onDidChangeModel(t=>{this._lastBracketsData=[],this._updateBracketsSoon.schedule()})),this._register(e.onDidChangeModelLanguageConfiguration(t=>{this._lastBracketsData=[],this._updateBracketsSoon.schedule()})),this._register(e.onDidChangeConfiguration(t=>{t.hasChanged(80)&&(this._matchBrackets=this._editor.getOption(80),this._decorations.clear(),this._lastBracketsData=[],this._lastVersionId=0,this._updateBracketsSoon.schedule())})),this._register(e.onDidBlurEditorWidget(()=>{this._updateBracketsSoon.schedule()})),this._register(e.onDidFocusEditorWidget(()=>{this._updateBracketsSoon.schedule()}))}jumpToBracket(){if(!this._editor.hasModel())return;const e=this._editor.getModel(),t=this._editor.getSelections().map(i=>{const n=i.getStartPosition(),o=e.bracketPairs.matchBracket(n);let r=null;if(o)o[0].containsPosition(n)&&!o[1].containsPosition(n)?r=o[1].getStartPosition():o[1].containsPosition(n)&&(r=o[0].getStartPosition());else{const a=e.bracketPairs.findEnclosingBrackets(n);if(a)r=a[1].getStartPosition();else{const l=e.bracketPairs.findNextBracket(n);l&&l.range&&(r=l.range.getStartPosition())}}return r?new he(r.lineNumber,r.column,r.lineNumber,r.column):new he(n.lineNumber,n.column,n.lineNumber,n.column)});this._editor.setSelections(t),this._editor.revealRange(t[0])}selectToBracket(e){if(!this._editor.hasModel())return;const t=this._editor.getModel(),i=[];this._editor.getSelections().forEach(n=>{const o=n.getStartPosition();let r=t.bracketPairs.matchBracket(o);if(!r&&(r=t.bracketPairs.findEnclosingBrackets(o),!r)){const c=t.bracketPairs.findNextBracket(o);c&&c.range&&(r=t.bracketPairs.matchBracket(c.range.getStartPosition()))}let a=null,l=null;if(r){r.sort(k.compareRangesUsingStarts);const[c,d]=r;if(a=e?c.getStartPosition():c.getEndPosition(),l=e?d.getEndPosition():d.getStartPosition(),d.containsPosition(o)){const h=a;a=l,l=h}}a&&l&&i.push(new he(a.lineNumber,a.column,l.lineNumber,l.column))}),i.length>0&&(this._editor.setSelections(i),this._editor.revealRange(i[0]))}removeBrackets(e){if(!this._editor.hasModel())return;const t=this._editor.getModel();this._editor.getSelections().forEach(i=>{const n=i.getPosition();let o=t.bracketPairs.matchBracket(n);o||(o=t.bracketPairs.findEnclosingBrackets(n)),o&&(this._editor.pushUndoStop(),this._editor.executeEdits(e,[{range:o[0],text:""},{range:o[1],text:""}]),this._editor.pushUndoStop())})}_updateBrackets(){if(this._matchBrackets==="never")return;this._recomputeBrackets();const e=[];let t=0;for(const i of this._lastBracketsData){const n=i.brackets;n&&(e[t++]={range:n[0],options:i.options},e[t++]={range:n[1],options:i.options})}this._decorations.set(e)}_recomputeBrackets(){if(!this._editor.hasModel()||!this._editor.hasWidgetFocus()){this._lastBracketsData=[],this._lastVersionId=0;return}const e=this._editor.getSelections();if(e.length>100){this._lastBracketsData=[],this._lastVersionId=0;return}const t=this._editor.getModel(),i=t.getVersionId();let n=[];this._lastVersionId===i&&(n=this._lastBracketsData);const o=[];let r=0;for(let h=0,u=e.length;h<u;h++){const g=e[h];g.isEmpty()&&(o[r++]=g.getStartPosition())}o.length>1&&o.sort(A.compare);const a=[];let l=0,c=0;const d=n.length;for(let h=0,u=o.length;h<u;h++){const g=o[h];for(;c<d&&n[c].position.isBefore(g);)c++;if(c<d&&n[c].position.equals(g))a[l++]=n[c];else{let f=t.bracketPairs.matchBracket(g,20),p=Jh._DECORATION_OPTIONS_WITH_OVERVIEW_RULER;!f&&this._matchBrackets==="always"&&(f=t.bracketPairs.findEnclosingBrackets(g,20),p=Jh._DECORATION_OPTIONS_WITHOUT_OVERVIEW_RULER),a[l++]=new REe(g,f,p)}}this._lastBracketsData=a,this._lastVersionId=i}};Jh.ID="editor.contrib.bracketMatchingController",Jh._DECORATION_OPTIONS_WITH_OVERVIEW_RULER=Fe.register({description:"bracket-match-overview",stickiness:1,className:"bracket-match",overviewRuler:{color:Ci(IEe),position:Sr.Center}}),Jh._DECORATION_OPTIONS_WITHOUT_OVERVIEW_RULER=Fe.register({description:"bracket-match-no-overview",stickiness:1,className:"bracket-match"});let Bm=Jh;lt(Bm.ID,Bm,1);ae(TEe);ae(DEe);ae(NEe);En.appendMenuItem(fe.MenubarGoMenu,{group:"5_infile_nav",command:{id:"editor.action.jumpToBracket",title:m(805,"Go to &&Bracket")},order:2});class AEe{constructor(e,t){this._selection=e,this._isMovingLeft=t}getEditOperations(e,t){if(this._selection.startLineNumber!==this._selection.endLineNumber||this._selection.isEmpty())return;const i=this._selection.startLineNumber,n=this._selection.startColumn,o=this._selection.endColumn;if(!(this._isMovingLeft&&n===1)&&!(!this._isMovingLeft&&o===e.getLineMaxColumn(i)))if(this._isMovingLeft){const r=new k(i,n-1,i,n),a=e.getValueInRange(r);t.addEditOperation(r,null),t.addEditOperation(new k(i,o,i,o),a)}else{const r=new k(i,o,i,o+1),a=e.getValueInRange(r);t.addEditOperation(r,null),t.addEditOperation(new k(i,n,i,n),a)}}computeCursorState(e,t){return this._isMovingLeft?new he(this._selection.startLineNumber,this._selection.startColumn-1,this._selection.endLineNumber,this._selection.endColumn-1):new he(this._selection.startLineNumber,this._selection.startColumn+1,this._selection.endLineNumber,this._selection.endColumn+1)}}class Pie extends ue{constructor(e,t){super(t),this.left=e}run(e,t){if(!t.hasModel())return;const i=[],n=t.getSelections();for(const o of n)i.push(new AEe(o,this.left));t.pushUndoStop(),t.executeCommands(this.id,i),t.pushUndoStop()}}class MEe extends Pie{constructor(){super(!0,{id:"editor.action.moveCarretLeftAction",label:V(810,"Move Selected Text Left"),precondition:N.writable})}}class OEe extends Pie{constructor(){super(!1,{id:"editor.action.moveCarretRightAction",label:V(811,"Move Selected Text Right"),precondition:N.writable})}}ae(MEe);ae(OEe);class PEe extends ue{constructor(){super({id:"editor.action.transposeLetters",label:V(812,"Transpose Letters"),precondition:N.writable,kbOpts:{kbExpr:N.textInputFocus,primary:0,mac:{primary:306},weight:100}})}run(e,t){if(!t.hasModel())return;const i=t.getModel(),n=[],o=t.getSelections();for(const r of o){if(!r.isEmpty())continue;const a=r.startLineNumber,l=r.startColumn,c=i.getLineMaxColumn(a);if(a===1&&(l===1||l===2&&c===2))continue;const d=l===c?r.getPosition():Ct.rightPosition(i,r.getPosition().lineNumber,r.getPosition().column),h=Ct.leftPosition(i,d),u=Ct.leftPosition(i,h),g=i.getValueInRange(k.fromPositions(u,h)),f=i.getValueInRange(k.fromPositions(h,d)),p=k.fromPositions(u,d);n.push(new ts(p,f+g))}n.length>0&&(t.pushUndoStop(),t.executeCommands(this.id,n),t.pushUndoStop())}}ae(PEe);const FEe=qe("productService");function j7(s,e){return{id:e,asString:async()=>s,asFile:()=>{},value:typeof s=="string"?s:void 0}}function BEe(s,e,t,i){const n={id:d_(),name:s,uri:e,data:t};return{id:i,asString:async()=>"",asFile:()=>n,value:void 0}}class Fie{constructor(){this._entries=new Map}get size(){let e=0;for(const t of this._entries)e++;return e}has(e){return this._entries.has(this.toKey(e))}matches(e){const t=[...this._entries.keys()];return ot.some(this,([i,n])=>n.asFile())&&t.push("files"),Bie(FI(e),t)}get(e){return this._entries.get(this.toKey(e))?.[0]}append(e,t){const i=this._entries.get(e);i?i.push(t):this._entries.set(this.toKey(e),[t])}replace(e,t){this._entries.set(this.toKey(e),[t])}delete(e){this._entries.delete(this.toKey(e))}*[Symbol.iterator](){for(const[e,t]of this._entries)for(const i of t)yield[e,i]}toKey(e){return FI(e)}}function FI(s){return s.toLowerCase()}function sj(s,e){return Bie(FI(s),e.map(FI))}function Bie(s,e){if(s==="*/*")return e.length>0;if(e.includes(s))return!0;const t=s.match(/^([a-z]+)\/([a-z]+|\*)$/i);if(!t)return!1;const[i,n,o]=t;return o==="*"?e.some(r=>r.startsWith(n+"/")):!1}const kA=Object.freeze({create:s=>fh(s.map(e=>e.toString())).join(`\r
`),split:s=>s.split(`\r
`),parse:s=>kA.split(s).filter(e=>!e.startsWith("#"))}),ec=class ec{constructor(e){this.value=e}equals(e){return this.value===e.value}contains(e){return this.equals(e)||this.value===""||e.value.startsWith(this.value+ec.sep)}intersects(e){return this.contains(e)||e.contains(this)}append(...e){return new ec((this.value?[this.value,...e]:e).join(ec.sep))}};ec.sep=".",ec.None=new ec("@@none@@"),ec.Empty=new ec("");let _i=ec;const WEe={ctrlCmd:!1,alt:!1};var HS;(function(s){s[s.Blur=1]="Blur",s[s.Gesture=2]="Gesture",s[s.Other=3]="Other"})(HS||(HS={}));var nl;(function(s){s[s.NONE=0]="NONE",s[s.FIRST=1]="FIRST",s[s.SECOND=2]="SECOND",s[s.LAST=3]="LAST"})(nl||(nl={}));var Kt;(function(s){s[s.First=1]="First",s[s.Second=2]="Second",s[s.Last=3]="Last",s[s.Next=4]="Next",s[s.Previous=5]="Previous",s[s.NextPage=6]="NextPage",s[s.PreviousPage=7]="PreviousPage",s[s.NextSeparator=8]="NextSeparator",s[s.PreviousSeparator=9]="PreviousSeparator"})(Kt||(Kt={}));var BI;(function(s){s[s.Title=1]="Title",s[s.Inline=2]="Inline",s[s.Input=3]="Input"})(BI||(BI={}));const bs=qe("quickInputService"),oj={EDITORS:"CodeEditors",FILES:"CodeFiles"};class HEe{}const VEe={DragAndDropContribution:"workbench.contributions.dragAndDrop"};bi.add(VEe.DragAndDropContribution,new HEe);const Iy=class Iy{constructor(){}static getInstance(){return Iy.INSTANCE}hasData(e){return e&&e===this.proto}getData(e){if(this.hasData(e))return this.data}};Iy.INSTANCE=new Iy;let A4=Iy;function Wie(s){if(Sl&&typeof globalThis.vscode?.webUtils?.getPathForFile=="function")return globalThis.vscode?.webUtils?.getPathForFile(s)}function Hie(s){const e=new Fie;for(const t of s.items){const i=t.type;if(t.kind==="string"){const n=new Promise(o=>t.getAsString(o));e.append(i,j7(n))}else if(t.kind==="file"){const n=t.getAsFile();n&&e.append(i,zEe(n))}}return e}function zEe(s){const e=Wie(s),t=e?Le.parse(e):void 0;return BEe(s.name,t,async()=>new Uint8Array(await s.arrayBuffer()))}const $Ee=Object.freeze([oj.EDITORS,oj.FILES,Mm.RESOURCES,Mm.INTERNAL_URI_LIST]);function Vie(s,e=!1){const t=Hie(s),i=t.get(Mm.INTERNAL_URI_LIST);if(i)t.replace(Ti.uriList,i);else if(e||!t.has(Ti.uriList)){const n=[];for(const o of s.items){const r=o.getAsFile();if(r){const a=Wie(r);try{a?n.push(Le.file(a).toString()):n.push(Le.parse(r.name,!0).toString())}catch{}}}n.length&&t.replace(Ti.uriList,j7(kA.create(n)))}for(const n of $Ee)t.delete(n);return t}const fx=qe("IWorkspaceEditService");class G7{constructor(e){this.metadata=e}static convert(e){return e.edits.map(t=>{if(Bu.is(t))return Bu.lift(t);if(Sw.is(t))return Sw.lift(t);throw new Error("Unsupported edit")})}}class Bu extends G7{static is(e){return e instanceof Bu?!0:rn(e)&&Le.isUri(e.resource)&&rn(e.textEdit)}static lift(e){return e instanceof Bu?e:new Bu(e.resource,e.textEdit,e.versionId,e.metadata)}constructor(e,t,i=void 0,n){super(n),this.resource=e,this.textEdit=t,this.versionId=i}}class Sw extends G7{static is(e){return e instanceof Sw?!0:rn(e)&&(!!e.newResource||!!e.oldResource)}static lift(e){return e instanceof Sw?e:new Sw(e.oldResource,e.newResource,e.options,e.metadata)}constructor(e,t,i={},n){super(n),this.oldResource=e,this.newResource=t,this.options=i}}const K7=qe("IEditorCancelService"),zie=new ce("cancellableOperation",!1,m(939,"Whether the editor runs a cancellable operation, e.g. like 'Peek References'"));tt(K7,class{constructor(){this._tokens=new WeakMap}add(s,e){let t=this._tokens.get(s);t||(t=s.invokeWithinContext(n=>{const o=zie.bindTo(n.get(Ae)),r=new Ns;return{key:o,tokens:r}}),this._tokens.set(s,t));let i;return t.key.set(!0),i=t.tokens.push(e),()=>{i&&(i(),t.key.set(!t.tokens.isEmpty()),i=void 0)}}cancel(s){const e=this._tokens.get(s);if(!e)return;const t=e.tokens.pop();t&&(t.cancel(),e.key.set(!e.tokens.isEmpty()))}},1);class UEe extends si{constructor(e,t){super(t),this.editor=e,this._unregister=e.invokeWithinContext(i=>i.get(K7).add(e,this))}dispose(){this._unregister(),super.dispose()}}le(new class extends hn{constructor(){super({id:"editor.cancelOperation",kbOpts:{weight:100,primary:9},precondition:zie})}runEditorCommand(s,e){s.get(K7).cancel(e)}});let $ie=class M4{constructor(e,t){if(this.flags=t,(this.flags&1)!==0){const i=e.getModel();this.modelVersionId=i?xm("{0}#{1}",i.uri.toString(),i.getVersionId()):null}else this.modelVersionId=null;(this.flags&4)!==0?this.position=e.getPosition():this.position=null,(this.flags&2)!==0?this.selection=e.getSelection():this.selection=null,(this.flags&8)!==0?(this.scrollLeft=e.getScrollLeft(),this.scrollTop=e.getScrollTop()):(this.scrollLeft=-1,this.scrollTop=-1)}_equals(e){if(!(e instanceof M4))return!1;const t=e;return!(this.modelVersionId!==t.modelVersionId||this.scrollLeft!==t.scrollLeft||this.scrollTop!==t.scrollTop||!this.position&&t.position||this.position&&!t.position||this.position&&t.position&&!this.position.equals(t.position)||!this.selection&&t.selection||this.selection&&!t.selection||this.selection&&t.selection&&!this.selection.equalsRange(t.selection))}validate(e){return this._equals(new M4(e,this.flags))}};class yh extends UEe{constructor(e,t,i,n){super(e,n),this._listener=new U,t&4&&this._listener.add(e.onDidChangeCursorPosition(o=>{(!i||!k.containsPosition(i,o.position))&&this.cancel()})),t&2&&this._listener.add(e.onDidChangeCursorSelection(o=>{(!i||!k.containsRange(i,o.selection))&&this.cancel()})),t&8&&this._listener.add(e.onDidScrollChange(o=>this.cancel())),t&1&&(this._listener.add(e.onDidChangeModel(o=>this.cancel())),this._listener.add(e.onDidChangeModelContent(o=>this.cancel())))}dispose(){this._listener.dispose(),super.dispose()}}class Y7 extends si{constructor(e,t){super(t),this._listener=e.onDidChangeContent(()=>this.cancel())}dispose(){this._listener.dispose(),super.dispose()}}var qEe=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},jEe=function(s,e){return function(t,i){e(t,i,s)}};const GEe=Fe.register({description:"inline-progress-widget",stickiness:1,showIfCollapsed:!0,after:{content:oQ,inlineClassName:"inline-editor-progress-decoration",inlineClassNameAffectsLetterSpacing:!0}}),mN=class mN extends F{constructor(e,t,i,n,o){super(),this.typeId=e,this.editor=t,this.range=i,this.delegate=o,this.allowEditorOverflow=!1,this.suppressMouseDown=!0,this.create(n),this.editor.addContentWidget(this),this.editor.layoutContentWidget(this)}create(e){this.domNode=ie(".inline-progress-widget"),this.domNode.role="button",this.domNode.title=e;const t=ie("span.icon");this.domNode.append(t),t.classList.add(...xe.asClassNameArray(Z.loading),"codicon-modifier-spin");const i=()=>{const n=this.editor.getOption(75);this.domNode.style.height=`${n}px`,this.domNode.style.width=`${Math.ceil(.8*n)}px`};i(),this._register(this.editor.onDidChangeConfiguration(n=>{(n.hasChanged(61)||n.hasChanged(75))&&i()})),this._register(W(this.domNode,se.CLICK,n=>{this.delegate.cancel()}))}getId(){return mN.baseId+"."+this.typeId}getDomNode(){return this.domNode}getPosition(){return{position:{lineNumber:this.range.startLineNumber,column:this.range.startColumn},preference:[0]}}dispose(){super.dispose(),this.editor.removeContentWidget(this)}};mN.baseId="editor.widget.inlineProgressWidget";let O4=mN,WI=class extends F{constructor(e,t,i){super(),this.id=e,this._editor=t,this._instantiationService=i,this._showDelay=500,this._showPromise=this._register(new vt),this._currentWidget=this._register(new vt),this._operationIdPool=0,this._currentDecorations=t.createDecorationsCollection()}dispose(){super.dispose(),this._currentDecorations.clear()}async showWhile(e,t,i,n,o){const r=this._operationIdPool++;this._currentOperation=r,this.clear(),this._showPromise.value=hh(()=>{const a=k.fromPositions(e);this._currentDecorations.set([{range:a,options:GEe}]).length>0&&(this._currentWidget.value=this._instantiationService.createInstance(O4,this.id,this._editor,a,t,n))},o??this._showDelay);try{return await i}finally{this._currentOperation===r&&(this.clear(),this._currentOperation=void 0)}}clear(){this._showPromise.clear(),this._currentDecorations.clear(),this._currentWidget.clear()}};WI=qEe([jEe(2,_e)],WI);const Ah=qe("openerService");function KEe(s,e){return s.with({fragment:`${e.startLineNumber},${e.startColumn}${e.endLineNumber?`-${e.endLineNumber}${e.endColumn?`,${e.endColumn}`:""}`:""}`})}function YEe(s){let e;const t=/^L?(\d+)(?:,(\d+))?(-L?(\d+)(?:,(\d+))?)?/.exec(s.fragment);return t&&(e={startLineNumber:parseInt(t[1]),startColumn:t[2]?parseInt(t[2]):1,endLineNumber:t[4]?parseInt(t[4]):void 0,endColumn:t[4]?t[5]?parseInt(t[5]):1:void 0},s=s.with({fragment:""})),{selection:e,uri:s}}var ZEe=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},XEe=function(s,e){return function(t,i){e(t,i,s)}};const Ua=qe("markdownRendererService");let P4=class{constructor(e){this._openerService=e}render(e,t,i){const n={...t};n.actionHandler||(n.actionHandler=(r,a)=>Uie(this._openerService,r,a.isTrusted)),n.codeBlockRenderer||(n.codeBlockRenderer=(r,a)=>this._defaultCodeBlockRenderer?.renderCodeBlock(r,a,n??{})??Promise.resolve(document.createElement("span")));const o=lx(e,n,i);return o.element.classList.add("rendered-markdown"),o}setDefaultCodeBlockRenderer(e){this._defaultCodeBlockRenderer=e}};P4=ZEe([XEe(0,Ah)],P4);async function Uie(s,e,t,i){try{return await s.open(e,{fromUserGesture:!0,allowContributedOpeners:!0,allowCommands:QEe(t),skipValidation:i})}catch(n){return Oe(n),!1}}function QEe(s){return s===!0?!0:s&&Array.isArray(s.enabledCommands)?s.enabledCommands:!1}tt(Ua,P4,1);var JEe=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},rj=function(s,e){return function(t,i){e(t,i,s)}},Hk,Uu;let Vo=(Uu=class{static get(e){return e.getContribution(Hk.ID)}constructor(e,t,i){this._openerService=i,this._messageWidget=new vt,this._messageListeners=new U,this._mouseOverMessage=!1,this._editor=e,this._visible=Hk.MESSAGE_VISIBLE.bindTo(t)}dispose(){this._messageListeners.dispose(),this._messageWidget.dispose(),this._visible.reset()}showMessage(e,t){if(Js(oh(e)?e.value:e),this._visible.set(!0),this._messageWidget.clear(),this._messageListeners.clear(),oh(e)){const n=this._messageListeners.add(lx(e,{actionHandler:(o,r)=>{this.closeMessage(),Uie(this._openerService,o,r.isTrusted)}}));this._messageWidget.value=new zM(this._editor,t,n.element)}else this._messageWidget.value=new zM(this._editor,t,e);this._messageListeners.add(oe.debounce(this._editor.onDidBlurEditorText,(n,o)=>o,0)(()=>{this._mouseOverMessage||this._messageWidget.value&&wn(ln(),this._messageWidget.value.getDomNode())||this.closeMessage()})),this._messageListeners.add(this._editor.onDidChangeCursorPosition(()=>this.closeMessage())),this._messageListeners.add(this._editor.onDidDispose(()=>this.closeMessage())),this._messageListeners.add(this._editor.onDidChangeModel(()=>this.closeMessage())),this._messageListeners.add(W(this._messageWidget.value.getDomNode(),se.MOUSE_ENTER,()=>this._mouseOverMessage=!0,!0)),this._messageListeners.add(W(this._messageWidget.value.getDomNode(),se.MOUSE_LEAVE,()=>this._mouseOverMessage=!1,!0));let i;this._messageListeners.add(this._editor.onMouseMove(n=>{n.target.position&&(i?i.containsPosition(n.target.position)||this.closeMessage():i=new k(t.lineNumber-3,1,n.target.position.lineNumber+3,1))}))}closeMessage(){this._visible.reset(),this._messageListeners.clear(),this._messageWidget.value&&this._messageListeners.add(zM.fadeOut(this._messageWidget.value))}},Hk=Uu,Uu.ID="editor.contrib.messageController",Uu.MESSAGE_VISIBLE=new ce("messageVisible",!1,m(1287,"Whether the editor is currently showing an inline message")),Uu);Vo=Hk=JEe([rj(1,Ae),rj(2,Ah)],Vo);const eIe=hn.bindToContribution(Vo.get);le(new eIe({id:"leaveEditorMessage",precondition:Vo.MESSAGE_VISIBLE,handler:s=>s.closeMessage(),kbOpts:{weight:130,primary:9}}));let zM=class{static fadeOut(e){const t=()=>{e.dispose(),clearTimeout(i),e.getDomNode().removeEventListener("animationend",t)},i=setTimeout(t,110);return e.getDomNode().addEventListener("animationend",t),e.getDomNode().classList.add("fadeOut"),{dispose:t}}constructor(e,{lineNumber:t,column:i},n){this.allowEditorOverflow=!0,this.suppressMouseDown=!1,this._editor=e,this._editor.revealLinesInCenterIfOutsideViewport(t,t,0),this._position={lineNumber:t,column:i},this._domNode=document.createElement("div"),this._domNode.classList.add("monaco-editor-overlaymessage"),this._domNode.style.marginLeft="-6px";const o=document.createElement("div");o.classList.add("anchor","top"),this._domNode.appendChild(o);const r=document.createElement("div");typeof n=="string"?(r.classList.add("message"),r.textContent=n):(n.classList.add("message"),r.appendChild(n)),this._domNode.appendChild(r);const a=document.createElement("div");a.classList.add("anchor","below"),this._domNode.appendChild(a),this._editor.addContentWidget(this),this._domNode.classList.add("fadeIn")}dispose(){this._editor.removeContentWidget(this)}getId(){return"messageoverlay"}getDomNode(){return this._domNode}getPosition(){return{position:this._position,preference:[1,2],positionAffinity:1}}afterRender(e){this._domNode.classList.toggle("below",e===2)}};lt(Vo.ID,Vo,4);class tIe{constructor(){this._value="",this._pos=0}reset(e){return this._value=e,this._pos=0,this}next(){return this._pos+=1,this}hasNext(){return this._pos<this._value.length-1}cmp(e){const t=e.charCodeAt(0),i=this._value.charCodeAt(this._pos);return t-i}value(){return this._value[this._pos]}}class iIe{constructor(e=!0){this._caseSensitive=e}reset(e){return this._value=e,this._from=0,this._to=0,this.next()}hasNext(){return this._to<this._value.length}next(){this._from=this._to;let e=!0;for(;this._to<this._value.length;this._to++)if(this._value.charCodeAt(this._to)===46)if(e)this._from++;else break;else e=!1;return this}cmp(e){return this._caseSensitive?q6(e,this._value,0,e.length,this._from,this._to):Uv(e,this._value,0,e.length,this._from,this._to)}value(){return this._value.substring(this._from,this._to)}}class nIe{constructor(e=!0,t=!0){this._splitOnBackslash=e,this._caseSensitive=t}reset(e){this._from=0,this._to=0,this._value=e,this._valueLen=e.length;for(let t=e.length-1;t>=0;t--,this._valueLen--){const i=this._value.charCodeAt(t);if(!(i===47||this._splitOnBackslash&&i===92))break}return this.next()}hasNext(){return this._to<this._valueLen}next(){this._from=this._to;let e=!0;for(;this._to<this._valueLen;this._to++){const t=this._value.charCodeAt(this._to);if(t===47||this._splitOnBackslash&&t===92)if(e)this._from++;else break;else e=!1}return this}cmp(e){return this._caseSensitive?q6(e,this._value,0,e.length,this._from,this._to):Uv(e,this._value,0,e.length,this._from,this._to)}value(){return this._value.substring(this._from,this._to)}}class sIe{constructor(e,t){this._ignorePathCasing=e,this._ignoreQueryAndFragment=t,this._states=[],this._stateIdx=0}reset(e){return this._value=e,this._states=[],this._value.scheme&&this._states.push(1),this._value.authority&&this._states.push(2),this._value.path&&(this._pathIterator=new nIe(!1,!this._ignorePathCasing(e)),this._pathIterator.reset(e.path),this._pathIterator.value()&&this._states.push(3)),this._ignoreQueryAndFragment(e)||(this._value.query&&this._states.push(4),this._value.fragment&&this._states.push(5)),this._stateIdx=0,this}next(){return this._states[this._stateIdx]===3&&this._pathIterator.hasNext()?this._pathIterator.next():this._stateIdx+=1,this}hasNext(){return this._states[this._stateIdx]===3&&this._pathIterator.hasNext()||this._stateIdx<this._states.length-1}cmp(e){if(this._states[this._stateIdx]===1)return hP(e,this._value.scheme);if(this._states[this._stateIdx]===2)return hP(e,this._value.authority);if(this._states[this._stateIdx]===3)return this._pathIterator.cmp(e);if(this._states[this._stateIdx]===4)return qy(e,this._value.query);if(this._states[this._stateIdx]===5)return qy(e,this._value.fragment);throw new Error}value(){if(this._states[this._stateIdx]===1)return this._value.scheme;if(this._states[this._stateIdx]===2)return this._value.authority;if(this._states[this._stateIdx]===3)return this._pathIterator.value();if(this._states[this._stateIdx]===4)return this._value.query;if(this._states[this._stateIdx]===5)return this._value.fragment;throw new Error}}const Dy=class Dy{static wrap(e){return e===void 0?Dy.Val:e}static unwrap(e){return e===Dy.Val?void 0:e}};Dy.Val=Symbol("undefined_placeholder");let wd=Dy;class yL{constructor(){this.height=1,this.value=void 0,this.key=void 0,this.left=void 0,this.mid=void 0,this.right=void 0}rotateLeft(){const e=this.right;return this.right=e.left,e.left=this,this.updateHeight(),e.updateHeight(),e}rotateRight(){const e=this.left;return this.left=e.right,e.right=this,this.updateHeight(),e.updateHeight(),e}updateHeight(){this.height=1+Math.max(this.heightLeft,this.heightRight)}balanceFactor(){return this.heightRight-this.heightLeft}get heightLeft(){return this.left?.height??0}get heightRight(){return this.right?.height??0}}class xw{static forUris(e=()=>!1,t=()=>!1){return new xw(new sIe(e,t))}static forStrings(){return new xw(new tIe)}static forConfigKeys(){return new xw(new iIe)}constructor(e){this._iter=e}clear(){this._root=void 0}set(e,t){const i=this._iter.reset(e);let n;this._root||(this._root=new yL,this._root.segment=i.value());const o=[];for(n=this._root;;){const a=i.cmp(n.segment);if(a>0)n.left||(n.left=new yL,n.left.segment=i.value()),o.push([-1,n]),n=n.left;else if(a<0)n.right||(n.right=new yL,n.right.segment=i.value()),o.push([1,n]),n=n.right;else if(i.hasNext())i.next(),n.mid||(n.mid=new yL,n.mid.segment=i.value()),o.push([0,n]),n=n.mid;else break}const r=wd.unwrap(n.value);n.value=wd.wrap(t),n.key=e;for(let a=o.length-1;a>=0;a--){const l=o[a][1];l.updateHeight();const c=l.balanceFactor();if(c<-1||c>1){const d=o[a][0],h=o[a+1][0];if(d===1&&h===1)o[a][1]=l.rotateLeft();else if(d===-1&&h===-1)o[a][1]=l.rotateRight();else if(d===1&&h===-1)l.right=o[a+1][1]=o[a+1][1].rotateRight(),o[a][1]=l.rotateLeft();else if(d===-1&&h===1)l.left=o[a+1][1]=o[a+1][1].rotateLeft(),o[a][1]=l.rotateRight();else throw new Error;if(a>0)switch(o[a-1][0]){case-1:o[a-1][1].left=o[a][1];break;case 1:o[a-1][1].right=o[a][1];break;case 0:o[a-1][1].mid=o[a][1];break}else this._root=o[0][1]}}return r}get(e){return wd.unwrap(this._getNode(e)?.value)}_getNode(e){const t=this._iter.reset(e);let i=this._root;for(;i;){const n=t.cmp(i.segment);if(n>0)i=i.left;else if(n<0)i=i.right;else if(t.hasNext())t.next(),i=i.mid;else break}return i}has(e){const t=this._getNode(e);return!(t?.value===void 0&&t?.mid===void 0)}delete(e){return this._delete(e,!1)}deleteSuperstr(e){return this._delete(e,!0)}_delete(e,t){const i=this._iter.reset(e),n=[];let o=this._root;for(;o;){const r=i.cmp(o.segment);if(r>0)n.push([-1,o]),o=o.left;else if(r<0)n.push([1,o]),o=o.right;else if(i.hasNext())i.next(),n.push([0,o]),o=o.mid;else break}if(o){if(t?(o.left=void 0,o.mid=void 0,o.right=void 0,o.height=1):(o.key=void 0,o.value=void 0),!o.mid&&!o.value)if(o.left&&o.right){const r=[[1,o]],a=this._min(o.right,r);if(a.key){o.key=a.key,o.value=a.value,o.segment=a.segment;const l=a.right;if(r.length>1){const[d,h]=r[r.length-1];switch(d){case-1:h.left=l;break;case 0:Vy(!1);case 1:Vy(!1)}}else o.right=l;const c=this._balanceByStack(r);if(n.length>0){const[d,h]=n[n.length-1];switch(d){case-1:h.left=c;break;case 0:h.mid=c;break;case 1:h.right=c;break}}else this._root=c}}else{const r=o.left??o.right;if(n.length>0){const[a,l]=n[n.length-1];switch(a){case-1:l.left=r;break;case 0:l.mid=r;break;case 1:l.right=r;break}}else this._root=r}this._root=this._balanceByStack(n)??this._root}}_min(e,t){for(;e.left;)t.push([-1,e]),e=e.left;return e}_balanceByStack(e){for(let t=e.length-1;t>=0;t--){const i=e[t][1];i.updateHeight();const n=i.balanceFactor();if(n>1?(i.right.balanceFactor()>=0||(i.right=i.right.rotateRight()),e[t][1]=i.rotateLeft()):n<-1&&(i.left.balanceFactor()<=0||(i.left=i.left.rotateLeft()),e[t][1]=i.rotateRight()),t>0)switch(e[t-1][0]){case-1:e[t-1][1].left=e[t][1];break;case 1:e[t-1][1].right=e[t][1];break;case 0:e[t-1][1].mid=e[t][1];break}else return e[0][1]}}findSubstr(e){const t=this._iter.reset(e);let i=this._root,n;for(;i;){const o=t.cmp(i.segment);if(o>0)i=i.left;else if(o<0)i=i.right;else if(t.hasNext())t.next(),n=wd.unwrap(i.value)||n,i=i.mid;else break}return i&&wd.unwrap(i.value)||n}findSuperstr(e){return this._findSuperstrOrElement(e,!1)}_findSuperstrOrElement(e,t){const i=this._iter.reset(e);let n=this._root;for(;n;){const o=i.cmp(n.segment);if(o>0)n=n.left;else if(o<0)n=n.right;else if(i.hasNext())i.next(),n=n.mid;else return n.mid?this._entries(n.mid):t?wd.unwrap(n.value):void 0}}forEach(e){for(const[t,i]of this)e(i,t)}*[Symbol.iterator](){yield*this._entries(this._root)}_entries(e){const t=[];return this._dfsEntries(e,t),t[Symbol.iterator]()}_dfsEntries(e,t){e&&(e.left&&this._dfsEntries(e.left,t),e.value!==void 0&&t.push([e.key,wd.unwrap(e.value)]),e.mid&&this._dfsEntries(e.mid,t),e.right&&this._dfsEntries(e.right,t))}}const Sh=qe("contextService");function F4(s){const e=s;return typeof e?.id=="string"&&Le.isUri(e.uri)}function oIe(s){return typeof s?.id=="string"&&!F4(s)&&!lIe(s)}const rIe={id:"empty-window"};function aIe(s,e){if(typeof s=="string"||typeof s>"u")return typeof s=="string"?{id:th(s)}:rIe;const t=s;return t.configuration?{id:t.id,configPath:t.configuration}:t.folders.length===1?{id:t.id,uri:t.folders[0].uri}:{id:t.id}}function lIe(s){const e=s;return typeof e?.id=="string"&&Le.isUri(e.configPath)}let cIe=class{constructor(e,t){this.raw=t,this.uri=e.uri,this.index=e.index,this.name=e.name}toJSON(){return{uri:this.uri,name:this.name,index:this.index}}};const B4="code-workspace";m(2050,"Code Workspace");const qie="4064f6ec-cb38-4ad0-af64-ee6467e63c82";function dIe(s){return s.id===qie}var Z7=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},VS=function(s,e){return function(t,i){e(t,i,s)}};class X7{constructor(e){this.copyMimeTypes=[],this.kind=e,this.providedDropEditKinds=[this.kind],this.providedPasteEditKinds=[this.kind]}async provideDocumentPasteEdits(e,t,i,n,o){const r=await this.getEdit(i,o);if(r)return{edits:[{insertText:r.insertText,title:r.title,kind:r.kind,handledMimeType:r.handledMimeType,yieldTo:r.yieldTo}],dispose(){}}}async provideDocumentDropEdits(e,t,i,n){const o=await this.getEdit(i,n);if(o)return{edits:[{insertText:o.insertText,title:o.title,kind:o.kind,handledMimeType:o.handledMimeType,yieldTo:o.yieldTo}],dispose(){}}}}const _N=class _N extends X7{constructor(){super(_i.Empty.append("text","plain")),this.id=_N.id,this.dropMimeTypes=[Ti.text],this.pasteMimeTypes=[Ti.text]}async getEdit(e,t){const i=e.get(Ti.text);if(!i||e.has(Ti.uriList))return;const n=await i.asString();return{handledMimeType:Ti.text,title:m(926,"Insert Plain Text"),insertText:n,kind:this.kind}}};_N.id="text";let Wm=_N;class jie extends X7{constructor(){super(_i.Empty.append("uri","path","absolute")),this.dropMimeTypes=[Ti.uriList],this.pasteMimeTypes=[Ti.uriList]}async getEdit(e,t){const i=await Gie(e);if(!i.length||t.isCancellationRequested)return;let n=0;const o=i.map(({uri:a,originalText:l})=>a.scheme===Te.file?a.fsPath:(n++,l)).join(" ");let r;return n>0?r=i.length>1?m(927,"Insert Uris"):m(928,"Insert Uri"):r=i.length>1?m(929,"Insert Paths"):m(930,"Insert Path"),{handledMimeType:Ti.uriList,insertText:o,title:r,kind:this.kind}}}let HI=class extends X7{constructor(e){super(_i.Empty.append("uri","path","relative")),this._workspaceContextService=e,this.dropMimeTypes=[Ti.uriList],this.pasteMimeTypes=[Ti.uriList]}async getEdit(e,t){const i=await Gie(e);if(!i.length||t.isCancellationRequested)return;const n=Al(i.map(({uri:o})=>{const r=this._workspaceContextService.getWorkspaceFolder(o);return r?ege(r.uri,o):void 0}));if(n.length)return{handledMimeType:Ti.uriList,insertText:n.join(" "),title:i.length>1?m(931,"Insert Relative Paths"):m(932,"Insert Relative Path"),kind:this.kind}}};HI=Z7([VS(0,Sh)],HI);class hIe{constructor(){this.kind=new _i("html"),this.providedPasteEditKinds=[this.kind],this.copyMimeTypes=[],this.pasteMimeTypes=["text/html"],this._yieldTo=[{mimeType:Ti.text}]}async provideDocumentPasteEdits(e,t,i,n,o){if(n.triggerKind!==mS.PasteAs&&!n.only?.contains(this.kind))return;const a=await i.get("text/html")?.asString();if(!(!a||o.isCancellationRequested))return{dispose(){},edits:[{insertText:a,yieldTo:this._yieldTo,title:m(933,"Insert HTML"),kind:this.kind}]}}}async function Gie(s){const e=s.get(Ti.uriList);if(!e)return[];const t=await e.asString(),i=[];for(const n of kA.parse(t))try{i.push({uri:Le.parse(n),originalText:n})}catch{}return i}const Xf={scheme:"*",hasAccessToAllModels:!0};let W4=class extends F{constructor(e,t){super(),this._register(e.documentDropEditProvider.register(Xf,new Wm)),this._register(e.documentDropEditProvider.register(Xf,new jie)),this._register(e.documentDropEditProvider.register(Xf,new HI(t)))}};W4=Z7([VS(0,ge),VS(1,Sh)],W4);let H4=class extends F{constructor(e,t){super(),this._register(e.documentPasteEditProvider.register(Xf,new Wm)),this._register(e.documentPasteEditProvider.register(Xf,new jie)),this._register(e.documentPasteEditProvider.register(Xf,new HI(t))),this._register(e.documentPasteEditProvider.register(Xf,new hIe))}};H4=Z7([VS(0,ge),VS(1,Sh)],H4);const Ca=class Ca{constructor(){this.value="",this.pos=0}static isDigitCharacter(e){return e>=48&&e<=57}static isVariableCharacter(e){return e===95||e>=97&&e<=122||e>=65&&e<=90}text(e){this.value=e,this.pos=0}tokenText(e){return this.value.substr(e.pos,e.len)}next(){if(this.pos>=this.value.length)return{type:14,pos:this.pos,len:0};const e=this.pos;let t=0,i=this.value.charCodeAt(e),n;if(n=Ca._table[i],typeof n=="number")return this.pos+=1,{type:n,pos:e,len:1};if(Ca.isDigitCharacter(i)){n=8;do t+=1,i=this.value.charCodeAt(e+t);while(Ca.isDigitCharacter(i));return this.pos+=t,{type:n,pos:e,len:t}}if(Ca.isVariableCharacter(i)){n=9;do i=this.value.charCodeAt(e+ ++t);while(Ca.isVariableCharacter(i)||Ca.isDigitCharacter(i));return this.pos+=t,{type:n,pos:e,len:t}}n=10;do t+=1,i=this.value.charCodeAt(e+t);while(!isNaN(i)&&typeof Ca._table[i]>"u"&&!Ca.isDigitCharacter(i)&&!Ca.isVariableCharacter(i));return this.pos+=t,{type:n,pos:e,len:t}}};Ca._table={36:0,58:1,44:2,123:3,125:4,92:5,47:6,124:7,43:11,45:12,63:13};let V4=Ca;class cC{constructor(){this._children=[]}appendChild(e){return e instanceof Xs&&this._children[this._children.length-1]instanceof Xs?this._children[this._children.length-1].value+=e.value:(e.parent=this,this._children.push(e)),this}replace(e,t){const{parent:i}=e,n=i.children.indexOf(e),o=i.children.slice(0);o.splice(n,1,...t),i._children=o,(function r(a,l){for(const c of a)c.parent=l,r(c.children,c)})(t,i)}get children(){return this._children}get rightMostDescendant(){return this._children.length>0?this._children[this._children.length-1].rightMostDescendant:this}get snippet(){let e=this;for(;;){if(!e)return;if(e instanceof px)return e;e=e.parent}}toString(){return this.children.reduce((e,t)=>e+t.toString(),"")}len(){return 0}}class Xs extends cC{constructor(e){super(),this.value=e}toString(){return this.value}len(){return this.value.length}clone(){return new Xs(this.value)}}class Kie extends cC{}class $r extends Kie{static compareByIndex(e,t){return e.index===t.index?0:e.isFinalTabstop?1:t.isFinalTabstop||e.index<t.index?-1:e.index>t.index?1:0}constructor(e){super(),this.index=e}get isFinalTabstop(){return this.index===0}get choice(){return this._children.length===1&&this._children[0]instanceof dC?this._children[0]:void 0}clone(){const e=new $r(this.index);return this.transform&&(e.transform=this.transform.clone()),e._children=this.children.map(t=>t.clone()),e}}class dC extends cC{constructor(){super(...arguments),this.options=[]}appendChild(e){return e instanceof Xs&&(e.parent=this,this.options.push(e)),this}toString(){return this.options[0].value}len(){return this.options[0].len()}clone(){const e=new dC;return this.options.forEach(e.appendChild,e),e}}class Q7 extends cC{constructor(){super(...arguments),this.regexp=new RegExp("")}resolve(e){const t=this;let i=!1,n=e.replace(this.regexp,function(){return i=!0,t._replace(Array.prototype.slice.call(arguments,0,-2))});return!i&&this._children.some(o=>o instanceof sl&&!!o.elseValue)&&(n=this._replace([])),n}_replace(e){let t="";for(const i of this._children)if(i instanceof sl){let n=e[i.index]||"";n=i.resolve(n),t+=n}else t+=i.toString();return t}toString(){return""}clone(){const e=new Q7;return e.regexp=new RegExp(this.regexp.source,(this.regexp.ignoreCase?"i":"")+(this.regexp.global?"g":"")),e._children=this.children.map(t=>t.clone()),e}}class sl extends cC{constructor(e,t,i,n){super(),this.index=e,this.shorthandName=t,this.ifValue=i,this.elseValue=n}resolve(e){return this.shorthandName==="upcase"?e?e.toLocaleUpperCase():"":this.shorthandName==="downcase"?e?e.toLocaleLowerCase():"":this.shorthandName==="capitalize"?e?e[0].toLocaleUpperCase()+e.substr(1):"":this.shorthandName==="pascalcase"?e?this._toPascalCase(e):"":this.shorthandName==="camelcase"?e?this._toCamelCase(e):"":e&&typeof this.ifValue=="string"?this.ifValue:!e&&typeof this.elseValue=="string"?this.elseValue:e||""}_toPascalCase(e){const t=e.match(/[a-z0-9]+/gi);return t?t.map(i=>i.charAt(0).toUpperCase()+i.substr(1)).join(""):e}_toCamelCase(e){const t=e.match(/[a-z0-9]+/gi);return t?t.map((i,n)=>n===0?i.charAt(0).toLowerCase()+i.substr(1):i.charAt(0).toUpperCase()+i.substr(1)).join(""):e}clone(){return new sl(this.index,this.shorthandName,this.ifValue,this.elseValue)}}class zS extends Kie{constructor(e){super(),this.name=e}resolve(e){let t=e.resolve(this);return this.transform&&(t=this.transform.resolve(t||"")),t!==void 0?(this._children=[new Xs(t)],!0):!1}clone(){const e=new zS(this.name);return this.transform&&(e.transform=this.transform.clone()),e._children=this.children.map(t=>t.clone()),e}}function aj(s,e){const t=[...s];for(;t.length>0;){const i=t.shift();if(!e(i))break;t.unshift(...i.children)}}class px extends cC{get placeholderInfo(){if(!this._placeholders){const e=[];let t;this.walk(function(i){return i instanceof $r&&(e.push(i),t=!t||t.index<i.index?i:t),!0}),this._placeholders={all:e,last:t}}return this._placeholders}get placeholders(){const{all:e}=this.placeholderInfo;return e}offset(e){let t=0,i=!1;return this.walk(n=>n===e?(i=!0,!1):(t+=n.len(),!0)),i?t:-1}fullLen(e){let t=0;return aj([e],i=>(t+=i.len(),!0)),t}enclosingPlaceholders(e){const t=[];let{parent:i}=e;for(;i;)i instanceof $r&&t.push(i),i=i.parent;return t}resolveVariables(e){return this.walk(t=>(t instanceof zS&&t.resolve(e)&&(this._placeholders=void 0),!0)),this}appendChild(e){return this._placeholders=void 0,super.appendChild(e)}replace(e,t){return this._placeholders=void 0,super.replace(e,t)}clone(){const e=new px;return this._children=this.children.map(t=>t.clone()),e}walk(e){aj(this.children,e)}}class Hm{constructor(){this._scanner=new V4,this._token={type:14,pos:0,len:0}}static escape(e){return e.replace(/\$|}|\\/g,"\\$&")}static guessNeedsClipboard(e){return/\${?CLIPBOARD/.test(e)}parse(e,t,i){const n=new px;return this.parseFragment(e,n),this.ensureFinalTabstop(n,i??!1,t??!1),n}parseFragment(e,t){const i=t.children.length;for(this._scanner.text(e),this._token=this._scanner.next();this._parse(t););const n=new Map,o=[];t.walk(l=>(l instanceof $r&&(l.isFinalTabstop?n.set(0,void 0):!n.has(l.index)&&l.children.length>0?n.set(l.index,l.children):o.push(l)),!0));const r=(l,c)=>{const d=n.get(l.index);if(!d)return;const h=new $r(l.index);h.transform=l.transform;for(const u of d){const g=u.clone();h.appendChild(g),g instanceof $r&&n.has(g.index)&&!c.has(g.index)&&(c.add(g.index),r(g,c),c.delete(g.index))}t.replace(l,[h])},a=new Set;for(const l of o)r(l,a);return t.children.slice(i)}ensureFinalTabstop(e,t,i){(t||i&&e.placeholders.length>0)&&(e.placeholders.find(o=>o.index===0)||e.appendChild(new $r(0)))}_accept(e,t){if(e===void 0||this._token.type===e){const i=t?this._scanner.tokenText(this._token):!0;return this._token=this._scanner.next(),i}return!1}_backTo(e){return this._scanner.pos=e.pos+e.len,this._token=e,!1}_until(e){const t=this._token;for(;this._token.type!==e;){if(this._token.type===14)return!1;if(this._token.type===5){const n=this._scanner.next();if(n.type!==0&&n.type!==4&&n.type!==5)return!1}this._token=this._scanner.next()}const i=this._scanner.value.substring(t.pos,this._token.pos).replace(/\\(\$|}|\\)/g,"$1");return this._token=this._scanner.next(),i}_parse(e){return this._parseEscaped(e)||this._parseTabstopOrVariableName(e)||this._parseComplexPlaceholder(e)||this._parseComplexVariable(e)||this._parseAnything(e)}_parseEscaped(e){let t;return(t=this._accept(5,!0))?(t=this._accept(0,!0)||this._accept(4,!0)||this._accept(5,!0)||t,e.appendChild(new Xs(t)),!0):!1}_parseTabstopOrVariableName(e){let t;const i=this._token;return this._accept(0)&&(t=this._accept(9,!0)||this._accept(8,!0))?(e.appendChild(/^\d+$/.test(t)?new $r(Number(t)):new zS(t)),!0):this._backTo(i)}_parseComplexPlaceholder(e){let t;const i=this._token;if(!(this._accept(0)&&this._accept(3)&&(t=this._accept(8,!0))))return this._backTo(i);const o=new $r(Number(t));if(this._accept(1))for(;;){if(this._accept(4))return e.appendChild(o),!0;if(!this._parse(o))return e.appendChild(new Xs("${"+t+":")),o.children.forEach(e.appendChild,e),!0}else if(o.index>0&&this._accept(7)){const r=new dC;for(;;){if(this._parseChoiceElement(r)){if(this._accept(2))continue;if(this._accept(7)&&(o.appendChild(r),this._accept(4)))return e.appendChild(o),!0}return this._backTo(i),!1}}else return this._accept(6)?this._parseTransform(o)?(e.appendChild(o),!0):(this._backTo(i),!1):this._accept(4)?(e.appendChild(o),!0):this._backTo(i)}_parseChoiceElement(e){const t=this._token,i=[];for(;!(this._token.type===2||this._token.type===7);){let n;if((n=this._accept(5,!0))?n=this._accept(2,!0)||this._accept(7,!0)||this._accept(5,!0)||n:n=this._accept(void 0,!0),!n)return this._backTo(t),!1;i.push(n)}return i.length===0?(this._backTo(t),!1):(e.appendChild(new Xs(i.join(""))),!0)}_parseComplexVariable(e){let t;const i=this._token;if(!(this._accept(0)&&this._accept(3)&&(t=this._accept(9,!0))))return this._backTo(i);const o=new zS(t);if(this._accept(1))for(;;){if(this._accept(4))return e.appendChild(o),!0;if(!this._parse(o))return e.appendChild(new Xs("${"+t+":")),o.children.forEach(e.appendChild,e),!0}else return this._accept(6)?this._parseTransform(o)?(e.appendChild(o),!0):(this._backTo(i),!1):this._accept(4)?(e.appendChild(o),!0):this._backTo(i)}_parseTransform(e){const t=new Q7;let i="",n="";for(;!this._accept(6);){let o;if(o=this._accept(5,!0)){o=this._accept(6,!0)||o,i+=o;continue}if(this._token.type!==14){i+=this._accept(void 0,!0);continue}return!1}for(;!this._accept(6);){let o;if(o=this._accept(5,!0)){o=this._accept(5,!0)||this._accept(6,!0)||o,t.appendChild(new Xs(o));continue}if(!(this._parseFormatString(t)||this._parseAnything(t)))return!1}for(;!this._accept(4);){if(this._token.type!==14){n+=this._accept(void 0,!0);continue}return!1}try{t.regexp=new RegExp(i,n)}catch{return!1}return e.transform=t,!0}_parseFormatString(e){const t=this._token;if(!this._accept(0))return!1;let i=!1;this._accept(3)&&(i=!0);const n=this._accept(8,!0);if(n)if(i){if(this._accept(4))return e.appendChild(new sl(Number(n))),!0;if(!this._accept(1))return this._backTo(t),!1}else return e.appendChild(new sl(Number(n))),!0;else return this._backTo(t),!1;if(this._accept(6)){const o=this._accept(9,!0);return!o||!this._accept(4)?(this._backTo(t),!1):(e.appendChild(new sl(Number(n),o)),!0)}else if(this._accept(11)){const o=this._until(4);if(o)return e.appendChild(new sl(Number(n),void 0,o,void 0)),!0}else if(this._accept(12)){const o=this._until(4);if(o)return e.appendChild(new sl(Number(n),void 0,void 0,o)),!0}else if(this._accept(13)){const o=this._until(1);if(o){const r=this._until(4);if(r)return e.appendChild(new sl(Number(n),void 0,o,r)),!0}}else{const o=this._until(4);if(o)return e.appendChild(new sl(Number(n),void 0,void 0,o)),!0}return this._backTo(t),!1}_parseAnything(e){return this._token.type!==14?(e.appendChild(new Xs(this._scanner.tokenText(this._token))),this._accept(void 0),!0):!1}}function Yie(s,e,t){return(typeof t.insertText=="string"?t.insertText==="":t.insertText.snippet==="")?{edits:t.additionalEdit?.edits??[]}:{edits:[...e.map(i=>new Bu(s,{range:i,text:typeof t.insertText=="string"?Hm.escape(t.insertText)+"$0":t.insertText.snippet,insertAsSnippet:!0})),...t.additionalEdit?.edits??[]]}}function Zie(s){function e(r,a){return"mimeType"in r?r.mimeType===a.handledMimeType:!!a.kind&&r.kind.contains(a.kind)}const t=new Map;for(const r of s)for(const a of r.yieldTo??[])for(const l of s)if(l!==r&&e(a,l)){let c=t.get(r);c||(c=[],t.set(r,c)),c.push(l)}if(!t.size)return Array.from(s);const i=new Set,n=[];function o(r){if(!r.length)return[];const a=r[0];if(n.includes(a))return console.warn("Yield to cycle detected",a),r;if(i.has(a))return o(r.slice(1));let l=[];const c=t.get(a);return c&&(n.push(a),l=o(c),n.pop()),i.add(a),[...l,a,...o(r.slice(1))]}return o(Array.from(s))}q.white.toString(),q.white.toString();const uIe=Object.freeze({allowedTags:{override:["b","i","u","code","span"]},allowedAttributes:{override:["class"]}});class VI extends F{get onDidClick(){return this._onDidClick.event}constructor(e,t){super(),this._label="",this._onDidClick=this._register(new M),this._onDidEscape=this._register(new M),this.options=t,this._element=document.createElement("a"),this._element.classList.add("monaco-button"),this._element.tabIndex=0,this._element.setAttribute("role","button"),this._element.classList.toggle("secondary",!!t.secondary);const i=t.secondary?t.buttonSecondaryBackground:t.buttonBackground,n=t.secondary?t.buttonSecondaryForeground:t.buttonForeground;this._element.style.color=n||"",this._element.style.backgroundColor=i||"",t.supportShortLabel&&(this._labelShortElement=document.createElement("div"),this._labelShortElement.classList.add("monaco-button-label-short"),this._element.appendChild(this._labelShortElement),this._labelElement=document.createElement("div"),this._labelElement.classList.add("monaco-button-label"),this._element.appendChild(this._labelElement),this._element.classList.add("monaco-text-button-with-short-label")),typeof t.title=="string"&&this.setTitle(t.title),typeof t.ariaLabel=="string"&&this._element.setAttribute("aria-label",t.ariaLabel),e.appendChild(this._element),this.enabled=!t.disabled,this._register(_s.addTarget(this._element)),[se.CLICK,qt.Tap].forEach(o=>{this._register(W(this._element,o,r=>{if(!this.enabled){gt.stop(r);return}this._onDidClick.fire(r)}))}),this._register(W(this._element,se.KEY_DOWN,o=>{const r=new Pt(o);let a=!1;this.enabled&&(r.equals(3)||r.equals(10))?(this._onDidClick.fire(o),a=!0):r.equals(9)&&(this._onDidEscape.fire(o),this._element.blur(),a=!0),a&&gt.stop(r,!0)})),this._register(W(this._element,se.MOUSE_OVER,o=>{this._element.classList.contains("disabled")||this.updateBackground(!0)})),this._register(W(this._element,se.MOUSE_OUT,o=>{this.updateBackground(!1)})),this.focusTracker=this._register(oa(this._element)),this._register(this.focusTracker.onDidFocus(()=>{this.enabled&&this.updateBackground(!0)})),this._register(this.focusTracker.onDidBlur(()=>{this.enabled&&this.updateBackground(!1)}))}dispose(){super.dispose(),this._element.remove()}getContentElements(e){const t=[];for(let i of Pu(e))if(typeof i=="string"){if(i=i.trim(),i==="")continue;const n=document.createElement("span");n.textContent=i,t.push(n)}else t.push(i);return t}updateBackground(e){let t;this.options.secondary?t=e?this.options.buttonSecondaryHoverBackground:this.options.buttonSecondaryBackground:t=e?this.options.buttonHoverBackground:this.options.buttonBackground,t&&(this._element.style.backgroundColor=t)}get element(){return this._element}set label(e){if(this._label===e||oh(this._label)&&oh(e)&&w0e(this._label,e))return;this._element.classList.add("monaco-text-button");const t=this.options.supportShortLabel?this._labelElement:this._element;if(oh(e)){const n=lx(e,void 0,document.createElement("span"));n.dispose();const o=n.element.querySelector("p")?.innerHTML;o?Jte(t,o,uIe):vn(t)}else this.options.supportIcons?vn(t,...this.getContentElements(e)):t.textContent=e;let i="";typeof this.options.title=="string"?i=this.options.title:this.options.title&&(i=Wxe(e)),this.setTitle(i),this._setAriaLabel(),this._label=e}get label(){return this._label}_setAriaLabel(){typeof this.options.ariaLabel=="string"?this._element.setAttribute("aria-label",this.options.ariaLabel):typeof this.options.title=="string"&&this._element.setAttribute("aria-label",this.options.title)}set icon(e){this._setAriaLabel();const t=Array.from(this._element.classList).filter(i=>i.startsWith("codicon-"));this._element.classList.remove(...t),this._element.classList.add(...xe.asClassNameArray(e))}set enabled(e){e?(this._element.classList.remove("disabled"),this._element.setAttribute("aria-disabled",String(!1)),this._element.tabIndex=0):(this._element.classList.add("disabled"),this._element.setAttribute("aria-disabled",String(!0)))}get enabled(){return!this._element.classList.contains("disabled")}setTitle(e){!this._hover&&e!==""?this._hover=this._register($a().setupManagedHover(this.options.hoverDelegate??Uc("element"),this._element,e)):this._hover&&this._hover.update(e)}}function $M(s,e){return e&&(s.stack||s.stacktrace)?m(29,"{0}: {1}",cj(s),lj(s.stack)||lj(s.stacktrace)):cj(s)}function lj(s){return Array.isArray(s)?s.join(`
`):s}function cj(s){return s.code==="ERR_UNC_HOST_NOT_ALLOWED"?`${s.message}. Please update the 'security.allowedUNCHosts' setting if you want to allow this host.`:typeof s.code=="string"&&typeof s.errno=="number"&&typeof s.syscall=="string"?m(30,"A system error occurred ({0})",s.message):s.message||m(31,"An unknown error occurred. Please consult the log for more details.")}function zI(s=null,e=!1){if(!s)return m(32,"An unknown error occurred. Please consult the log for more details.");if(Array.isArray(s)){const t=Al(s),i=zI(t[0],e);return t.length>1?m(33,"{0} ({1} errors in total)",i,t.length):i}if(bn(s))return s;if(s.detail){const t=s.detail;if(t.error)return $M(t.error,e);if(t.exception)return $M(t.exception,e)}return s.stack?$M(s,e):s.message?s.message:m(34,"An unknown error occurred. Please consult the log for more details.")}const SL=ie,gIe={keybindingLabelBackground:void 0,keybindingLabelForeground:void 0,keybindingLabelBorder:void 0,keybindingLabelBottomBorder:void 0,keybindingLabelShadow:void 0};class hC extends F{constructor(e,t,i){super(),this.os=t,this.keyElements=new Set,this.options=i||Object.create(null);const n=this.options.keybindingLabelForeground;this.domNode=X(e,SL(".monaco-keybinding")),n&&(this.domNode.style.color=n),this.hover=this._register($a().setupManagedHover(Uc("mouse"),this.domNode,"")),this.didEverRender=!1,e.appendChild(this.domNode)}set(e,t){this.didEverRender&&this.keybinding===e&&hC.areSame(this.matches,t)||(this.keybinding=e,this.matches=t,this.render())}render(){if(this.clear(),this.keybinding){const e=this.keybinding.getChords();e[0]&&this.renderChord(this.domNode,e[0],this.matches?this.matches.firstPart:null);for(let i=1;i<e.length;i++)X(this.domNode,SL("span.monaco-keybinding-key-chord-separator",void 0," ")),this.renderChord(this.domNode,e[i],this.matches?this.matches.chordPart:null);const t=this.options.disableTitle??!1?void 0:this.keybinding.getAriaLabel()||void 0;this.hover.update(t),this.domNode.setAttribute("aria-label",t||"")}else this.options&&this.options.renderUnboundKeybindings&&this.renderUnbound(this.domNode);this.didEverRender=!0}clear(){Pn(this.domNode),this.keyElements.clear()}renderChord(e,t,i){const n=z7.modifierLabels[this.os];t.ctrlKey&&this.renderKey(e,n.ctrlKey,!!i?.ctrlKey,n.separator),t.shiftKey&&this.renderKey(e,n.shiftKey,!!i?.shiftKey,n.separator),t.altKey&&this.renderKey(e,n.altKey,!!i?.altKey,n.separator),t.metaKey&&this.renderKey(e,n.metaKey,!!i?.metaKey,n.separator);const o=t.keyLabel;o&&this.renderKey(e,o,!!i?.keyCode,"")}renderKey(e,t,i,n){X(e,this.createKeyElement(t,i?".highlight":"")),n&&X(e,SL("span.monaco-keybinding-key-separator",void 0,n))}renderUnbound(e){X(e,this.createKeyElement(m(15,"Unbound")))}createKeyElement(e,t=""){const i=SL("span.monaco-keybinding-key"+t,void 0,e);return this.keyElements.add(i),this.options.keybindingLabelBackground&&(i.style.backgroundColor=this.options.keybindingLabelBackground),this.options.keybindingLabelBorder&&(i.style.borderColor=this.options.keybindingLabelBorder),this.options.keybindingLabelBottomBorder&&(i.style.borderBottomColor=this.options.keybindingLabelBottomBorder),this.options.keybindingLabelShadow&&(i.style.boxShadow=`inset 0 -1px 0 ${this.options.keybindingLabelShadow}`),i}static areSame(e,t){return e===t||!e&&!t?!0:!!e&&!!t&&Ho(e.firstPart,t.firstPart)&&Ho(e.chordPart,t.chordPart)}}const qc=qe("layoutService");var Xie=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},Vk=function(s,e){return function(t,i){e(t,i,s)}};const Qie="acceptSelectedCodeAction",Jie="previewSelectedCodeAction";class fIe{get templateId(){return"header"}renderTemplate(e){e.classList.add("group-header");const t=document.createElement("span");return e.append(t),{container:e,text:t}}renderElement(e,t,i){i.text.textContent=e.group?.title??e.label??""}disposeTemplate(e){}}class pIe{get templateId(){return"separator"}renderTemplate(e){e.classList.add("separator");const t=document.createElement("span");return e.append(t),{container:e,text:t}}renderElement(e,t,i){i.text.textContent=e.label??""}disposeTemplate(e){}}let z4=class{get templateId(){return"action"}constructor(e,t){this._supportsPreview=e,this._keybindingService=t}renderTemplate(e){e.classList.add(this.templateId);const t=document.createElement("div");t.className="icon",e.append(t);const i=document.createElement("span");i.className="title",e.append(i);const n=document.createElement("span");n.className="description",e.append(n);const o=new hC(e,Fo);return{container:e,icon:t,text:i,description:n,keybinding:o}}renderElement(e,t,i){if(e.group?.icon?(i.icon.className=xe.asClassName(e.group.icon),e.group.icon.color&&(i.icon.style.color=te(e.group.icon.color.id))):(i.icon.className=xe.asClassName(Z.lightBulb),i.icon.style.color="var(--vscode-editorLightBulb-foreground)"),!e.item||!e.label)return;fue(!e.hideIcon,i.icon),i.text.textContent=$I(e.label),e.keybinding?(i.description.textContent=e.keybinding.getLabel(),i.description.style.display="inline",i.description.style.letterSpacing="0.5px"):e.description?(i.description.textContent=$I(e.description),i.description.style.display="inline"):(i.description.textContent="",i.description.style.display="none");const n=this._keybindingService.lookupKeybinding(Qie)?.getLabel(),o=this._keybindingService.lookupKeybinding(Jie)?.getLabel();i.container.classList.toggle("option-disabled",e.disabled),e.tooltip?i.container.title=e.tooltip:e.disabled?i.container.title=e.label:n&&o?this._supportsPreview&&e.canPreview?i.container.title=m(1653,"{0} to Apply, {1} to Preview",n,o):i.container.title=m(1654,"{0} to Apply",n):i.container.title=""}disposeTemplate(e){e.keybinding.dispose()}};z4=Xie([Vk(1,ft)],z4);class mIe extends UIEvent{constructor(){super("acceptSelectedAction")}}class dj extends UIEvent{constructor(){super("previewSelectedAction")}}function _Ie(s){if(s.kind==="action")return s.label}let $4=class extends F{constructor(e,t,i,n,o,r,a,l){super(),this._delegate=n,this._contextViewService=r,this._keybindingService=a,this._layoutService=l,this._actionLineHeight=28,this._headerLineHeight=28,this._separatorLineHeight=8,this.cts=this._register(new si),this.domNode=document.createElement("div"),this.domNode.classList.add("actionList");const c={getHeight:d=>{switch(d.kind){case"header":return this._headerLineHeight;case"separator":return this._separatorLineHeight;default:return this._actionLineHeight}},getTemplateId:d=>d.kind};this._list=this._register(new Tr(e,this.domNode,c,[new z4(t,this._keybindingService),new fIe,new pIe],{keyboardSupport:!1,typeNavigationEnabled:!0,keyboardNavigationLabelProvider:{getKeyboardNavigationLabel:_Ie},accessibilityProvider:{getAriaLabel:d=>{if(d.kind==="action"){let h=d.label?$I(d?.label):"";return d.description&&(h=h+", "+$I(d.description)),d.disabled&&(h=m(1655,"{0}, Disabled Reason: {1}",h,d.disabled)),h}return null},getWidgetAriaLabel:()=>m(1656,"Action Widget"),getRole:d=>{switch(d.kind){case"action":return"option";case"separator":return"separator";default:return"separator"}},getWidgetRole:()=>"listbox",...o}})),this._list.style(lC),this._register(this._list.onMouseClick(d=>this.onListClick(d))),this._register(this._list.onMouseOver(d=>this.onListHover(d))),this._register(this._list.onDidChangeFocus(()=>this.onFocus())),this._register(this._list.onDidChangeSelection(d=>this.onListSelection(d))),this._allMenuItems=i,this._list.splice(0,this._list.length,this._allMenuItems),this._list.length&&this.focusNext()}focusCondition(e){return!e.disabled&&e.kind==="action"}hide(e){this._delegate.onHide(e),this.cts.cancel(),this._contextViewService.hideContextView()}layout(e){const t=this._allMenuItems.filter(d=>d.kind==="header").length,i=this._allMenuItems.filter(d=>d.kind==="separator").length,r=this._allMenuItems.length*this._actionLineHeight+t*this._headerLineHeight-t*this._actionLineHeight+i*this._separatorLineHeight-i*this._actionLineHeight;this._list.layout(r);let a=e;if(this._allMenuItems.length>=50)a=380;else{const d=this._allMenuItems.map((h,u)=>{const g=this.domNode.ownerDocument.getElementById(this._list.getElementID(u));if(g){g.style.width="auto";const f=g.getBoundingClientRect().width;return g.style.width="",f}return 0});a=Math.max(...d,e)}const c=Math.min(r,this._layoutService.getContainer(be(this.domNode)).clientHeight*.7);return this._list.layout(c,a),this.domNode.style.height=`${c}px`,this._list.domFocus(),a}focusPrevious(){this._list.focusPrevious(1,!0,void 0,this.focusCondition)}focusNext(){this._list.focusNext(1,!0,void 0,this.focusCondition)}acceptSelected(e){const t=this._list.getFocus();if(t.length===0)return;const i=t[0],n=this._list.element(i);if(!this.focusCondition(n))return;const o=e?new dj:new mIe;this._list.setSelection([i],o)}onListSelection(e){if(!e.elements.length)return;const t=e.elements[0];t.item&&this.focusCondition(t)?this._delegate.onSelect(t.item,e.browserEvent instanceof dj):this._list.setSelection([])}onFocus(){const e=this._list.getFocus();if(e.length===0)return;const t=e[0],i=this._list.element(t);this._delegate.onFocus?.(i.item)}async onListHover(e){const t=e.element;if(t&&t.item&&this.focusCondition(t)){if(this._delegate.onHover&&!t.disabled&&t.kind==="action"){const i=await this._delegate.onHover(t.item,this.cts.token);t.canPreview=i?i.canPreview:void 0}e.index&&this._list.splice(e.index,1,[t])}this._list.setFocus(typeof e.index=="number"?[e.index]:[])}onListClick(e){e.element&&this.focusCondition(e.element)&&this._list.setFocus([])}};$4=Xie([Vk(5,Rh),Vk(6,ft),Vk(7,qc)],$4);function $I(s){return s.replace(/\r\n|\r|\n/g," ")}var bIe=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},UM=function(s,e){return function(t,i){e(t,i,s)}};T("actionBar.toggledBackground",iC,m(1657,"Background color for toggled action items in action bar."));const Vm={Visible:new ce("codeActionMenuVisible",!1,m(1658,"Whether the action widget list is visible"))},Rg=qe("actionWidgetService");let zm=class extends F{get isVisible(){return Vm.Visible.getValue(this._contextKeyService)||!1}constructor(e,t,i){super(),this._contextViewService=e,this._contextKeyService=t,this._instantiationService=i,this._list=this._register(new vt)}show(e,t,i,n,o,r,a,l){const c=Vm.Visible.bindTo(this._contextKeyService),d=this._instantiationService.createInstance($4,e,t,i,n,l);this._contextViewService.showContextView({getAnchor:()=>o,render:h=>(c.set(!0),this._renderWidget(h,d,a??[])),onHide:h=>{c.reset(),this._onWidgetClosed(h)}},r,!1)}acceptSelected(e){this._list.value?.acceptSelected(e)}focusPrevious(){this._list?.value?.focusPrevious()}focusNext(){this._list?.value?.focusNext()}hide(e){this._list.value?.hide(e),this._list.clear()}_renderWidget(e,t,i){const n=document.createElement("div");if(n.classList.add("action-widget"),e.appendChild(n),this._list.value=t,this._list.value)n.appendChild(this._list.value.domNode);else throw new Error("List has no value");const o=new U,r=document.createElement("div"),a=e.appendChild(r);a.classList.add("context-view-block"),o.add(W(a,se.MOUSE_DOWN,g=>g.stopPropagation()));const l=document.createElement("div"),c=e.appendChild(l);c.classList.add("context-view-pointerBlock"),o.add(W(c,se.POINTER_MOVE,()=>c.remove())),o.add(W(c,se.MOUSE_DOWN,()=>c.remove()));let d=0;if(i.length){const g=this._createActionBar(".action-widget-action-bar",i);g&&(n.appendChild(g.getContainer().parentElement),o.add(g),d=g.getContainer().offsetWidth)}const h=this._list.value?.layout(d);n.style.width=`${h}px`;const u=o.add(oa(e));return o.add(u.onDidBlur(()=>this.hide(!0))),o}_createActionBar(e,t){if(!t.length)return;const i=ie(e),n=new bo(i);return n.push(t,{icon:!1,label:!0}),n}_onWidgetClosed(e){this._list.value?.hide(e)}};zm=bIe([UM(0,Rh),UM(1,Ae),UM(2,_e)],zm);tt(Rg,zm,1);const mx=1100;It(class extends Tn{constructor(){super({id:"hideCodeActionWidget",title:V(1659,"Hide action widget"),precondition:Vm.Visible,keybinding:{weight:mx,primary:9,secondary:[1033]}})}run(s){s.get(Rg).hide(!0)}});It(class extends Tn{constructor(){super({id:"selectPrevCodeAction",title:V(1660,"Select previous action"),precondition:Vm.Visible,keybinding:{weight:mx,primary:16,secondary:[2064],mac:{primary:16,secondary:[2064,302]}}})}run(s){const e=s.get(Rg);e instanceof zm&&e.focusPrevious()}});It(class extends Tn{constructor(){super({id:"selectNextCodeAction",title:V(1661,"Select next action"),precondition:Vm.Visible,keybinding:{weight:mx,primary:18,secondary:[2066],mac:{primary:18,secondary:[2066,300]}}})}run(s){const e=s.get(Rg);e instanceof zm&&e.focusNext()}});It(class extends Tn{constructor(){super({id:Qie,title:V(1662,"Accept selected action"),precondition:Vm.Visible,keybinding:{weight:mx,primary:3,secondary:[2137]}})}run(s){const e=s.get(Rg);e instanceof zm&&e.acceptSelected()}});It(class extends Tn{constructor(){super({id:Jie,title:V(1663,"Preview selected action"),precondition:Vm.Visible,keybinding:{weight:mx,primary:2051}})}run(s){const e=s.get(Rg);e instanceof zm&&e.acceptSelected(!0)}});var ene=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},Lw=function(s,e){return function(t,i){e(t,i,s)}},U4,Ip;let q4=(Ip=class extends F{constructor(e,t,i,n,o,r,a,l,c,d,h){super(),this.typeId=e,this.editor=t,this.showCommand=n,this.range=o,this.edits=r,this.onSelectNewEdit=a,this.additionalActions=l,this._keybindingService=d,this._actionWidgetService=h,this.allowEditorOverflow=!0,this.suppressMouseDown=!0,this.create(),this.visibleContext=i.bindTo(c),this.visibleContext.set(!0),this._register(pe(()=>this.visibleContext.reset())),this.editor.addContentWidget(this),this.editor.layoutContentWidget(this),this._register(pe((()=>this.editor.removeContentWidget(this)))),this._register(this.editor.onDidChangeCursorPosition(u=>{this.dispose()})),this._register(oe.runAndSubscribe(d.onDidUpdateKeybindings,()=>{this._updateButtonTitle()}))}_updateButtonTitle(){const e=this._keybindingService.lookupKeybinding(this.showCommand.id)?.getLabel();this.button.element.title=this.showCommand.label+(e?` (${e})`:"")}create(){this.domNode=ie(".post-edit-widget"),this.button=this._register(new VI(this.domNode,{supportIcons:!0})),this.button.label="$(insert)",this._register(W(this.domNode,se.CLICK,()=>this.showSelector()))}getId(){return U4.baseId+"."+this.typeId}getDomNode(){return this.domNode}getPosition(){return{position:this.range.getEndPosition(),preference:[2]}}showSelector(){const e=Li(this.button.element),t={x:e.left+e.width,y:e.top+e.height};this._actionWidgetService.show("postEditWidget",!1,this.edits.allEdits.map((i,n)=>({kind:"action",item:i,label:i.title,disabled:!1,canPreview:!1,group:{title:"",icon:xe.fromId(n===this.edits.activeEditIndex?Z.check.id:Z.blank.id)}})),{onHide:()=>{this.editor.focus()},onSelect:i=>{this._actionWidgetService.hide(!1);const n=this.edits.allEdits.findIndex(o=>o===i);if(n!==this.edits.activeEditIndex)return this.onSelectNewEdit(n)}},t,this.editor.getDomNode()??void 0,this.additionalActions)}},U4=Ip,Ip.baseId="editor.widget.postEditWidget",Ip);q4=U4=ene([Lw(8,Ae),Lw(9,ft),Lw(10,Rg)],q4);let UI=class extends F{constructor(e,t,i,n,o,r,a,l){super(),this._id=e,this._editor=t,this._visibleContext=i,this._showCommand=n,this._getAdditionalActions=o,this._instantiationService=r,this._bulkEditService=a,this._notificationService=l,this._currentWidget=this._register(new vt),this._register(oe.any(t.onDidChangeModel,t.onDidChangeModelContent)(()=>this.clear()))}async applyEditAndShowIfNeeded(e,t,i,n,o){if(!e.length||!this._editor.hasModel())return;const r=this._editor.getModel(),a=t.allEdits.at(t.activeEditIndex);if(!a)return;const l=async b=>{const w=this._editor.getModel();w&&(await w.undo(),this.applyEditAndShowIfNeeded(e,{activeEditIndex:b,allEdits:t.allEdits},i,n,o))},c=(b,w)=>{Dr(b)||(this._notificationService.error(w),i&&this.show(e[0],t,l))},d=new yh(this._editor,3,void 0,o);let h;try{h=await Sde(n(a,d.token),d.token)}catch(b){return c(b,m(937,`Error resolving edit '{0}':
{1}`,a.title,zI(b)))}finally{d.dispose()}if(o.isCancellationRequested)return;const u=Yie(r.uri,e,h),g=e[0],f=r.deltaDecorations([],[{range:g,options:{description:"paste-line-suffix",stickiness:0}}]);this._editor.focus();let p,_;try{p=await this._bulkEditService.apply(u,{editor:this._editor,token:o}),_=r.getDecorationRange(f[0])}catch(b){return c(b,m(938,`Error applying edit '{0}':
{1}`,a.title,zI(b)))}finally{r.deltaDecorations(f,[])}o.isCancellationRequested||i&&p.isApplied&&t.allEdits.length>1&&this.show(_??g,t,l)}show(e,t,i){this.clear(),this._editor.hasModel()&&(this._currentWidget.value=this._instantiationService.createInstance(q4,this._id,this._editor,this._visibleContext,this._showCommand,e,t,i,this._getAdditionalActions()))}clear(){this._currentWidget.clear()}tryShowSelector(){this._currentWidget.value?.showSelector()}};UI=ene([Lw(5,_e),Lw(6,fx),Lw(7,Ii)],UI);var wIe=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},od=function(s,e){return function(t,i){e(t,i,s)}},ba;const tne="editor.changePasteType",vIe="editor.pasteAs.preferences",J7=new ce("pasteWidgetVisible",!1,m(917,"Whether the paste widget is showing")),qM="application/vnd.code.copymetadata";var Dp;let xh=(Dp=class extends F{static get(e){return e.getContribution(ba.ID)}constructor(e,t,i,n,o,r,a,l,c,d){super(),this._logService=i,this._bulkEditService=n,this._clipboardService=o,this._commandService=r,this._configService=a,this._languageFeaturesService=l,this._quickInputService=c,this._progressService=d,this._editor=e;const h=e.getContainerDomNode();this._register(W(h,"copy",u=>this.handleCopy(u))),this._register(W(h,"cut",u=>this.handleCopy(u))),this._register(W(h,"paste",u=>this.handlePaste(u),!0)),this._pasteProgressManager=this._register(new WI("pasteIntoEditor",e,t)),this._postPasteWidgetManager=this._register(t.createInstance(UI,"pasteIntoEditor",e,J7,{id:tne,label:m(918,"Show paste options...")},()=>ba._configureDefaultAction?[ba._configureDefaultAction]:[]))}changePasteType(){this._postPasteWidgetManager.tryShowSelector()}async pasteAs(e){this._logService.trace("CopyPasteController.pasteAs"),this._editor.focus();try{this._logService.trace("Before calling editor.action.clipboardPasteAction"),this._pasteAsActionContext={preferred:e},await this._commandService.executeCommand("editor.action.clipboardPasteAction")}finally{this._pasteAsActionContext=void 0}}clearWidgets(){this._postPasteWidgetManager.clear()}isPasteAsEnabled(){return this._editor.getOption(97).enabled}async finishedPaste(){await this._currentPasteOperation}handleCopy(e){let t=null;if(e.clipboardData){const[_,b]=lp.getTextData(e.clipboardData);t=(b||xc.INSTANCE.get(_))?.id||null,this._logService.trace("CopyPasteController#handleCopy for id : ",t," with text.length : ",_.length)}else this._logService.trace("CopyPasteController#handleCopy");if(!this._editor.hasTextFocus()||(this._clipboardService.clearInternalState?.(),!e.clipboardData||!this.isPasteAsEnabled()))return;const i=this._editor.getModel(),n=this._editor.getSelections();if(!i||!n?.length)return;const o=this._editor.getOption(45);let r=n;const a=n.length===1&&n[0].isEmpty();if(a){if(!o)return;r=[new k(r[0].startLineNumber,1,r[0].startLineNumber,1+i.getLineLength(r[0].startLineNumber))]}const l=this._editor._getViewModel()?.getPlainTextToCopy(n,o,Bn),d={multicursorText:Array.isArray(l)?l:null,pasteOnNewLine:a,mode:null},h=this._languageFeaturesService.documentPasteEditProvider.ordered(i).filter(_=>!!_.prepareDocumentPaste);if(!h.length){this.setCopyMetadata(e.clipboardData,{defaultPastePayload:d});return}const u=Hie(e.clipboardData),g=h.flatMap(_=>_.copyMimeTypes??[]),f=t??d_();this.setCopyMetadata(e.clipboardData,{id:f,providerCopyMimeTypes:g,defaultPastePayload:d});const p=h.map(_=>({providerMimeTypes:_.copyMimeTypes,operation:an(b=>_.prepareDocumentPaste(i,r,u,b).catch(w=>{console.error(w)}))}));ba._currentCopyOperation?.operations.forEach(_=>_.operation.cancel()),ba._currentCopyOperation={handle:f,operations:p}}async handlePaste(e){if(e.clipboardData){const[c,d]=lp.getTextData(e.clipboardData),h=d||xc.INSTANCE.get(c);this._logService.trace("CopyPasteController#handlePaste for id : ",h?.id)}else this._logService.trace("CopyPasteController#handlePaste");if(!e.clipboardData||!this._editor.hasTextFocus())return;Vo.get(this._editor)?.closeMessage(),this._currentPasteOperation?.cancel(),this._currentPasteOperation=void 0;const t=this._editor.getModel(),i=this._editor.getSelections();if(!i?.length||!t||this._editor.getOption(104)||!this.isPasteAsEnabled()&&!this._pasteAsActionContext)return;const n=this.fetchCopyMetadata(e);this._logService.trace("CopyPasteController#handlePaste with metadata : ",n?.id," and text.length : ",e.clipboardData.getData("text/plain").length);const o=Vie(e.clipboardData);o.delete(qM);const r=Array.from(e.clipboardData.files).map(c=>c.type),a=[...e.clipboardData.types,...r,...n?.providerCopyMimeTypes??[],Ti.uriList],l=this._languageFeaturesService.documentPasteEditProvider.ordered(t).filter(c=>{const d=this._pasteAsActionContext?.preferred;return d&&!this.providerMatchesPreference(c,d)?!1:c.pasteMimeTypes?.some(h=>sj(h,a))});if(!l.length){this._pasteAsActionContext?.preferred&&(this.showPasteAsNoEditMessage(i,this._pasteAsActionContext.preferred),e.preventDefault(),e.stopImmediatePropagation());return}e.preventDefault(),e.stopImmediatePropagation(),this._pasteAsActionContext?this.showPasteAsPick(this._pasteAsActionContext.preferred,l,i,o,n):this.doPasteInline(l,i,o,n,e)}showPasteAsNoEditMessage(e,t){const i="only"in t?t.only.value:"preferences"in t?t.preferences.length?t.preferences.map(n=>n.value).join(", "):m(919,"empty"):t.providerId;Vo.get(this._editor)?.showMessage(m(920,"No paste edits for '{0}' found",i),e[0].getStartPosition())}doPasteInline(e,t,i,n,o){this._logService.trace("CopyPasteController#doPasteInline");const r=this._editor;if(!r.hasModel())return;const a=new yh(r,3,void 0),l=an(async c=>{const d=this._editor;if(!d.hasModel())return;const h=d.getModel(),u=new U,g=u.add(new si(c));u.add(a.token.onCancellationRequested(()=>g.cancel()));const f=g.token;try{if(await this.mergeInDataFromCopy(e,i,n,f),f.isCancellationRequested)return;const p=e.filter(w=>this.isSupportedPasteProvider(w,i));if(!p.length||p.length===1&&p[0]instanceof Wm)return this.applyDefaultPasteHandler(i,n,f,o);const _={triggerKind:mS.Automatic},b=await this.getPasteEdits(p,i,h,t,_,f);if(u.add(b),f.isCancellationRequested)return;if(b.edits.length===1&&b.edits[0].provider instanceof Wm)return this.applyDefaultPasteHandler(i,n,f,o);if(b.edits.length){const w=d.getOption(97).showPasteSelector==="afterPaste";return this._postPasteWidgetManager.applyEditAndShowIfNeeded(t,{activeEditIndex:this.getInitialActiveEditIndex(h,b.edits),allEdits:b.edits},w,async(C,v)=>{if(!C.provider.resolveDocumentPasteEdit)return C;const y=C.provider.resolveDocumentPasteEdit(C,v),x=new n_,L=await this._pasteProgressManager.showWhile(t[0].getEndPosition(),m(921,"Resolving paste edit for '{0}'. Click to cancel",C.title),tv(Promise.race([x.p,y]),v),{cancel:()=>x.cancel()},0);return L&&(C.insertText=L.insertText,C.additionalEdit=L.additionalEdit),C},f)}await this.applyDefaultPasteHandler(i,n,f,o)}finally{u.dispose(),this._currentPasteOperation===l&&(this._currentPasteOperation=void 0)}});this._pasteProgressManager.showWhile(t[0].getEndPosition(),m(922,"Running paste handlers. Click to cancel and do basic paste"),l,{cancel:async()=>{l.cancel(),!a.token.isCancellationRequested&&await this.applyDefaultPasteHandler(i,n,a.token,o)}}).finally(()=>{a.dispose()}),this._currentPasteOperation=l}showPasteAsPick(e,t,i,n,o){this._logService.trace("CopyPasteController#showPasteAsPick");const r=an(async a=>{const l=this._editor;if(!l.hasModel())return;const c=l.getModel(),d=new U,h=d.add(new yh(l,3,void 0,a));try{if(await this.mergeInDataFromCopy(t,n,o,h.token),h.token.isCancellationRequested)return;let u=t.filter(b=>this.isSupportedPasteProvider(b,n,e));e&&(u=u.filter(b=>this.providerMatchesPreference(b,e)));const g={triggerKind:mS.PasteAs,only:e&&"only"in e?e.only:void 0};let f=d.add(await this.getPasteEdits(u,n,c,i,g,h.token));if(h.token.isCancellationRequested)return;if(e&&(f={edits:f.edits.filter(b=>"only"in e?e.only.contains(b.kind):"preferences"in e?e.preferences.some(w=>w.contains(b.kind)):e.providerId===b.provider.id),dispose:f.dispose}),!f.edits.length){e&&this.showPasteAsNoEditMessage(i,e);return}let p;if(e)p=f.edits.at(0);else{const b={id:"editor.pasteAs.default",label:m(923,"Configure default paste action"),edit:void 0},w=await this._quickInputService.pick([...f.edits.map(C=>({label:C.title,description:C.kind?.value,edit:C})),...ba._configureDefaultAction?[{type:"separator"},{label:ba._configureDefaultAction.label,edit:void 0}]:[]],{placeHolder:m(924,"Select Paste Action")});if(w===b){ba._configureDefaultAction?.run();return}p=w?.edit}if(!p)return;const _=Yie(c.uri,i,p);await this._bulkEditService.apply(_,{editor:this._editor})}finally{d.dispose(),this._currentPasteOperation===r&&(this._currentPasteOperation=void 0)}});this._progressService.withProgress({location:10,title:m(925,"Running paste handlers")},()=>r)}setCopyMetadata(e,t){this._logService.trace("CopyPasteController#setCopyMetadata new id : ",t.id),e.setData(qM,JSON.stringify(t))}fetchCopyMetadata(e){if(this._logService.trace("CopyPasteController#fetchCopyMetadata"),!e.clipboardData)return;const t=e.clipboardData.getData(qM);if(t)try{return JSON.parse(t)}catch{return}const[i,n]=lp.getTextData(e.clipboardData);if(n)return{defaultPastePayload:{mode:n.mode,multicursorText:n.multicursorText??null,pasteOnNewLine:!!n.isFromEmptySelection}}}async mergeInDataFromCopy(e,t,i,n){if(this._logService.trace("CopyPasteController#mergeInDataFromCopy with metadata : ",i?.id),i?.id&&ba._currentCopyOperation?.handle===i.id){const o=ba._currentCopyOperation.operations.filter(a=>e.some(l=>l.pasteMimeTypes.some(c=>sj(c,a.providerMimeTypes)))).map(a=>a.operation),r=await Promise.all(o);if(n.isCancellationRequested)return;for(const a of r.reverse())if(a)for(const[l,c]of a)t.replace(l,c)}if(!t.has(Ti.uriList)){const o=await this._clipboardService.readResources();if(n.isCancellationRequested)return;o.length&&t.append(Ti.uriList,j7(kA.create(o)))}}async getPasteEdits(e,t,i,n,o,r){const a=new U,l=await tv(Promise.all(e.map(async d=>{try{const h=await d.provideDocumentPasteEdits?.(i,n,t,o,r);return h&&a.add(h),h?.edits?.map(u=>({...u,provider:d}))}catch(h){Dr(h)||console.error(h);return}})),r),c=Al(l??[]).flat().filter(d=>!o.only||o.only.contains(d.kind));return{edits:Zie(c),dispose:()=>a.dispose()}}async applyDefaultPasteHandler(e,t,i,n){const r=await(e.get(Ti.text)??e.get("text"))?.asString()??"";if(i.isCancellationRequested)return;const a={clipboardEvent:n,text:r,pasteOnNewLine:t?.defaultPastePayload.pasteOnNewLine??!1,multicursorText:t?.defaultPastePayload.multicursorText??null,mode:null};this._logService.trace("CopyPasteController#applyDefaultPasteHandler for id : ",t?.id),this._editor.trigger("keyboard","paste",a)}isSupportedPasteProvider(e,t,i){return e.pasteMimeTypes?.some(n=>t.matches(n))?!i||this.providerMatchesPreference(e,i):!1}providerMatchesPreference(e,t){return"only"in t?e.providedPasteEditKinds.some(i=>t.only.contains(i)):"preferences"in t?t.preferences.some(i=>t.preferences.some(n=>n.contains(i))):e.id===t.providerId}getInitialActiveEditIndex(e,t){const i=this._configService.getValue(vIe,{resource:e.uri});for(const n of Array.isArray(i)?i:[]){const o=new _i(n),r=t.findIndex(a=>o.contains(a.kind));if(r>=0)return r}return 0}},ba=Dp,Dp.ID="editor.contrib.copyPasteActionController",Dp);xh=ba=wIe([od(1,_e),od(2,jt),od(3,fx),od(4,Ko),od(5,Gt),od(6,He),od(7,ge),od(8,bs),od(9,bte)],xh);const $m="9_cutcopypaste",CIe=Sl||document.queryCommandSupported("cut"),ine=Sl||document.queryCommandSupported("copy"),yIe=typeof navigator.clipboard>"u"||wo?document.queryCommandSupported("paste"):!0;function eH(s){return s.register(),s}const SIe=CIe?eH(new Yv({id:"editor.action.clipboardCutAction",precondition:void 0,kbOpts:Sl?{primary:2102,win:{primary:2102,secondary:[1044]},weight:100}:void 0,menuOpts:[{menuId:fe.MenubarEditMenu,group:"2_ccp",title:m(813,"Cu&&t"),order:1},{menuId:fe.EditorContext,group:$m,title:m(814,"Cut"),when:N.writable,order:1},{menuId:fe.CommandPalette,group:"",title:m(815,"Cut"),order:1},{menuId:fe.SimpleEditorContext,group:$m,title:m(816,"Cut"),when:N.writable,order:1}]})):void 0,xIe=ine?eH(new Yv({id:"editor.action.clipboardCopyAction",precondition:void 0,kbOpts:Sl?{primary:2081,win:{primary:2081,secondary:[2067]},weight:100}:void 0,menuOpts:[{menuId:fe.MenubarEditMenu,group:"2_ccp",title:m(817,"&&Copy"),order:2},{menuId:fe.EditorContext,group:$m,title:m(818,"Copy"),order:2},{menuId:fe.CommandPalette,group:"",title:m(819,"Copy"),order:1},{menuId:fe.SimpleEditorContext,group:$m,title:m(820,"Copy"),order:2}]})):void 0;En.appendMenuItem(fe.MenubarEditMenu,{submenu:fe.MenubarCopy,title:V(825,"Copy As"),group:"2_ccp",order:3});En.appendMenuItem(fe.EditorContext,{submenu:fe.EditorContextCopy,title:V(826,"Copy As"),group:$m,order:3});En.appendMenuItem(fe.EditorContext,{submenu:fe.EditorContextShare,title:V(827,"Share"),group:"11_share",order:-1,when:K.and(K.notEquals("resourceScheme","output"),N.editorTextFocus)});En.appendMenuItem(fe.ExplorerContext,{submenu:fe.ExplorerContextShare,title:V(828,"Share"),group:"11_share",order:-1});const jM=yIe?eH(new Yv({id:"editor.action.clipboardPasteAction",precondition:void 0,kbOpts:Sl?{primary:2100,win:{primary:2100,secondary:[1043]},linux:{primary:2100,secondary:[1043]},weight:100}:void 0,menuOpts:[{menuId:fe.MenubarEditMenu,group:"2_ccp",title:m(821,"&&Paste"),order:4},{menuId:fe.EditorContext,group:$m,title:m(822,"Paste"),when:N.writable,order:4},{menuId:fe.CommandPalette,group:"",title:m(823,"Paste"),order:1},{menuId:fe.SimpleEditorContext,group:$m,title:m(824,"Paste"),when:N.writable,order:4}]})):void 0;class LIe extends ue{constructor(){super({id:"editor.action.clipboardCopyWithSyntaxHighlightingAction",label:V(829,"Copy with Syntax Highlighting"),precondition:void 0,kbOpts:{kbExpr:N.textInputFocus,primary:0,weight:100}})}run(e,t){const i=e.get(jt);i.trace("ExecCommandCopyWithSyntaxHighlightingAction#run"),!(!t.hasModel()||!t.getOption(45)&&t.getSelection().isEmpty())&&(nF.forceCopyWithSyntaxHighlighting=!0,t.focus(),i.trace("ExecCommandCopyWithSyntaxHighlightingAction (before execCommand copy)"),t.getContainerDomNode().ownerDocument.execCommand("copy"),i.trace("ExecCommandCopyWithSyntaxHighlightingAction (after execCommand copy)"),nF.forceCopyWithSyntaxHighlighting=!1)}}function nne(s,e){s&&(s.addImplementation(1e4,"code-editor",(t,i)=>{const n=t.get(jt);n.trace("registerExecCommandImpl (addImplementation code-editor for : ",e,")");const o=t.get(ht).getFocusedCodeEditor();if(o&&o.hasTextFocus()){const r=o.getOption(45),a=o.getSelection();return a&&a.isEmpty()&&!r||(o.getOption(170)&&e==="cut"?(hj(o),n.trace("registerExecCommandImpl (before execCommand copy)"),o.getContainerDomNode().ownerDocument.execCommand("copy"),o.trigger(void 0,"cut",void 0),n.trace("registerExecCommandImpl (after execCommand copy)")):(hj(o),n.trace("registerExecCommandImpl (before execCommand "+e+")"),o.getContainerDomNode().ownerDocument.execCommand(e),n.trace("registerExecCommandImpl (after execCommand "+e+")"))),!0}return!1}),s.addImplementation(0,"generic-dom",(t,i)=>{const n=t.get(jt);return n.trace("registerExecCommandImpl (addImplementation generic-dom for : ",e,")"),n.trace("registerExecCommandImpl (before execCommand "+e+")"),K0().execCommand(e),n.trace("registerExecCommandImpl (after execCommand "+e+")"),!0}))}function hj(s){if(s.getOption(170)){const t=e7.get(s.getId());t&&t.onWillCopy()}}nne(SIe,"cut");nne(xIe,"copy");jM&&(jM.addImplementation(1e4,"code-editor",(s,e)=>{const t=s.get(jt);t.trace("registerExecCommandImpl (addImplementation code-editor for : paste)");const i=s.get(ht),n=s.get(Ko),o=s.get(vs),r=s.get(FEe),a=i.getFocusedCodeEditor();if(a&&a.hasModel()&&a.hasTextFocus()){if(a.getOption(170)){const h=e7.get(a.getId());h&&h.onWillPaste()}const c=Cn.create(!0);t.trace("registerExecCommandImpl (before triggerPaste)");const d=n.triggerPaste(kt().vscodeWindowId);return d?(t.trace("registerExecCommandImpl (triggerPaste defined)"),d.then(async()=>{if(t.trace("registerExecCommandImpl (after triggerPaste)"),r.quality!=="stable"){const h=c.elapsed();o.publicLog2("editorAsyncPaste",{duration:h})}return xh.get(a)?.finishedPaste()??Promise.resolve()})):(t.trace("registerExecCommandImpl (triggerPaste undefined)"),Vc?(t.trace("registerExecCommandImpl (Paste handling on web)"),(async()=>{const h=await n.readText();if(h!==""){const u=xc.INSTANCE.get(h);let g=!1,f=null,p=null;u&&(g=a.getOption(45)&&!!u.isFromEmptySelection,f=typeof u.multicursorText<"u"?u.multicursorText:null,p=u.mode),t.trace("registerExecCommandImpl (clipboardText.length : ",h.length," id : ",u?.id,")"),a.trigger("keyboard","paste",{text:h,pasteOnNewLine:g,multicursorText:f,mode:p})}})()):!0)}return!1}),jM.addImplementation(0,"generic-dom",(s,e)=>(s.get(jt).trace("registerExecCommandImpl (addImplementation generic-dom for : paste)"),s.get(Ko).triggerPaste(kt().vscodeWindowId)??!1)));ine&&ae(LIe);const EA=Object.freeze({id:"editor",order:5,type:"object",title:m(147,"Editor"),scope:6}),qI={...EA,properties:{"editor.tabSize":{type:"number",default:Qn.tabSize,minimum:1,maximum:100,markdownDescription:m(148,"The number of spaces a tab is equal to. This setting is overridden based on the file contents when {0} is on.","`#editor.detectIndentation#`")},"editor.indentSize":{anyOf:[{type:"string",enum:["tabSize"]},{type:"number",minimum:1}],default:"tabSize",markdownDescription:m(149,'The number of spaces used for indentation or `"tabSize"` to use the value from `#editor.tabSize#`. This setting is overridden based on the file contents when `#editor.detectIndentation#` is on.')},"editor.insertSpaces":{type:"boolean",default:Qn.insertSpaces,markdownDescription:m(150,"Insert spaces when pressing `Tab`. This setting is overridden based on the file contents when {0} is on.","`#editor.detectIndentation#`")},"editor.detectIndentation":{type:"boolean",default:Qn.detectIndentation,markdownDescription:m(151,"Controls whether {0} and {1} will be automatically detected when a file is opened based on the file contents.","`#editor.tabSize#`","`#editor.insertSpaces#`")},"editor.trimAutoWhitespace":{type:"boolean",default:Qn.trimAutoWhitespace,description:m(152,"Remove trailing auto inserted whitespace.")},"editor.largeFileOptimizations":{type:"boolean",default:Qn.largeFileOptimizations,description:m(153,"Special handling for large files to disable certain memory intensive features.")},"editor.wordBasedSuggestions":{enum:["off","currentDocument","matchingDocuments","allDocuments"],default:"matchingDocuments",enumDescriptions:[m(154,"Turn off Word Based Suggestions."),m(155,"Only suggest words from the active document."),m(156,"Suggest words from all open documents of the same language."),m(157,"Suggest words from all open documents.")],description:m(158,"Controls whether completions should be computed based on words in the document and from which documents they are computed.")},"editor.semanticHighlighting.enabled":{enum:[!0,!1,"configuredByTheme"],enumDescriptions:[m(159,"Semantic highlighting enabled for all color themes."),m(160,"Semantic highlighting disabled for all color themes."),m(161,"Semantic highlighting is configured by the current color theme's `semanticHighlighting` setting.")],default:"configuredByTheme",description:m(162,"Controls whether the semanticHighlighting is shown for the languages that support it.")},"editor.stablePeek":{type:"boolean",default:!1,markdownDescription:m(163,"Keep peek editors open even when double-clicking their content or when hitting `Escape`.")},"editor.maxTokenizationLineLength":{type:"integer",default:2e4,description:m(164,"Lines above this length will not be tokenized for performance reasons")},"editor.experimental.asyncTokenization":{type:"boolean",default:!0,description:m(165,"Controls whether the tokenization should happen asynchronously on a web worker."),tags:["experimental"]},"editor.experimental.asyncTokenizationLogging":{type:"boolean",default:!1,description:m(166,"Controls whether async tokenization should be logged. For debugging only.")},"editor.experimental.asyncTokenizationVerification":{type:"boolean",default:!1,description:m(167,"Controls whether async tokenization should be verified against legacy background tokenization. Might slow down tokenization. For debugging only."),tags:["experimental"]},"editor.experimental.treeSitterTelemetry":{type:"boolean",default:!1,markdownDescription:m(168,"Controls whether tree sitter parsing should be turned on and telemetry collected. Setting `#editor.experimental.preferTreeSitter#` for specific languages will take precedence."),tags:["experimental"],experiment:{mode:"auto"}},"editor.experimental.preferTreeSitter.css":{type:"boolean",default:!1,markdownDescription:m(169,"Controls whether tree sitter parsing should be turned on for css. This will take precedence over `#editor.experimental.treeSitterTelemetry#` for css."),tags:["experimental"],experiment:{mode:"auto"}},"editor.experimental.preferTreeSitter.typescript":{type:"boolean",default:!1,markdownDescription:m(170,"Controls whether tree sitter parsing should be turned on for typescript. This will take precedence over `#editor.experimental.treeSitterTelemetry#` for typescript."),tags:["experimental"],experiment:{mode:"auto"}},"editor.experimental.preferTreeSitter.ini":{type:"boolean",default:!1,markdownDescription:m(171,"Controls whether tree sitter parsing should be turned on for ini. This will take precedence over `#editor.experimental.treeSitterTelemetry#` for ini."),tags:["experimental"],experiment:{mode:"auto"}},"editor.experimental.preferTreeSitter.regex":{type:"boolean",default:!1,markdownDescription:m(172,"Controls whether tree sitter parsing should be turned on for regex. This will take precedence over `#editor.experimental.treeSitterTelemetry#` for regex."),tags:["experimental"],experiment:{mode:"auto"}},"editor.language.brackets":{type:["array","null"],default:null,description:m(173,"Defines the bracket symbols that increase or decrease the indentation."),items:{type:"array",items:[{type:"string",description:m(174,"The opening bracket character or string sequence.")},{type:"string",description:m(175,"The closing bracket character or string sequence.")}]}},"editor.language.colorizedBracketPairs":{type:["array","null"],default:null,description:m(176,"Defines the bracket pairs that are colorized by their nesting level if bracket pair colorization is enabled."),items:{type:"array",items:[{type:"string",description:m(177,"The opening bracket character or string sequence.")},{type:"string",description:m(178,"The closing bracket character or string sequence.")}]}},"diffEditor.maxComputationTime":{type:"number",default:$n.maxComputationTime,description:m(179,"Timeout in milliseconds after which diff computation is cancelled. Use 0 for no timeout.")},"diffEditor.maxFileSize":{type:"number",default:$n.maxFileSize,description:m(180,"Maximum file size in MB for which to compute diffs. Use 0 for no limit.")},"diffEditor.renderSideBySide":{type:"boolean",default:$n.renderSideBySide,description:m(181,"Controls whether the diff editor shows the diff side by side or inline.")},"diffEditor.renderSideBySideInlineBreakpoint":{type:"number",default:$n.renderSideBySideInlineBreakpoint,description:m(182,"If the diff editor width is smaller than this value, the inline view is used.")},"diffEditor.useInlineViewWhenSpaceIsLimited":{type:"boolean",default:$n.useInlineViewWhenSpaceIsLimited,description:m(183,"If enabled and the editor width is too small, the inline view is used.")},"diffEditor.renderMarginRevertIcon":{type:"boolean",default:$n.renderMarginRevertIcon,description:m(184,"When enabled, the diff editor shows arrows in its glyph margin to revert changes.")},"diffEditor.renderGutterMenu":{type:"boolean",default:$n.renderGutterMenu,description:m(185,"When enabled, the diff editor shows a special gutter for revert and stage actions.")},"diffEditor.ignoreTrimWhitespace":{type:"boolean",default:$n.ignoreTrimWhitespace,description:m(186,"When enabled, the diff editor ignores changes in leading or trailing whitespace.")},"diffEditor.renderIndicators":{type:"boolean",default:$n.renderIndicators,description:m(187,"Controls whether the diff editor shows +/- indicators for added/removed changes.")},"diffEditor.codeLens":{type:"boolean",default:$n.diffCodeLens,description:m(188,"Controls whether the editor shows CodeLens.")},"diffEditor.wordWrap":{type:"string",enum:["off","on","inherit"],default:$n.diffWordWrap,markdownEnumDescriptions:[m(189,"Lines will never wrap."),m(190,"Lines will wrap at the viewport width."),m(191,"Lines will wrap according to the {0} setting.","`#editor.wordWrap#`")]},"diffEditor.diffAlgorithm":{type:"string",enum:["legacy","advanced"],default:$n.diffAlgorithm,markdownEnumDescriptions:[m(192,"Uses the legacy diffing algorithm."),m(193,"Uses the advanced diffing algorithm.")]},"diffEditor.hideUnchangedRegions.enabled":{type:"boolean",default:$n.hideUnchangedRegions.enabled,markdownDescription:m(194,"Controls whether the diff editor shows unchanged regions.")},"diffEditor.hideUnchangedRegions.revealLineCount":{type:"integer",default:$n.hideUnchangedRegions.revealLineCount,markdownDescription:m(195,"Controls how many lines are used for unchanged regions."),minimum:1},"diffEditor.hideUnchangedRegions.minimumLineCount":{type:"integer",default:$n.hideUnchangedRegions.minimumLineCount,markdownDescription:m(196,"Controls how many lines are used as a minimum for unchanged regions."),minimum:1},"diffEditor.hideUnchangedRegions.contextLineCount":{type:"integer",default:$n.hideUnchangedRegions.contextLineCount,markdownDescription:m(197,"Controls how many lines are used as context when comparing unchanged regions."),minimum:1},"diffEditor.experimental.showMoves":{type:"boolean",default:$n.experimental.showMoves,markdownDescription:m(198,"Controls whether the diff editor should show detected code moves.")},"diffEditor.experimental.showEmptyDecorations":{type:"boolean",default:$n.experimental.showEmptyDecorations,description:m(199,"Controls whether the diff editor shows empty decorations to see where characters got inserted or deleted.")},"diffEditor.experimental.useTrueInlineView":{type:"boolean",default:$n.experimental.useTrueInlineView,description:m(200,"If enabled and the editor uses the inline view, word changes are rendered inline.")}}};function kIe(s){return typeof s.type<"u"||typeof s.anyOf<"u"}for(const s of xb){const e=s.schema;if(typeof e<"u")if(kIe(e))qI.properties[`editor.${s.name}`]=e;else for(const t in e)Object.hasOwnProperty.call(e,t)&&(qI.properties[t]=e[t])}let xL=null;function sne(){return xL===null&&(xL=Object.create(null),Object.keys(qI.properties).forEach(s=>{xL[s]=!0})),xL}function EIe(s){return sne()[`editor.${s}`]||!1}function IIe(s){return sne()[`diffEditor.${s}`]||!1}const DIe=bi.as(Ml.Configuration);DIe.registerConfiguration(qI);const ii=new class{constructor(){this.QuickFix=new _i("quickfix"),this.Refactor=new _i("refactor"),this.RefactorExtract=this.Refactor.append("extract"),this.RefactorInline=this.Refactor.append("inline"),this.RefactorMove=this.Refactor.append("move"),this.RefactorRewrite=this.Refactor.append("rewrite"),this.Notebook=new _i("notebook"),this.Source=new _i("source"),this.SourceOrganizeImports=this.Source.append("organizeImports"),this.SourceFixAll=this.Source.append("fixAll"),this.SurroundWith=this.Refactor.append("surround")}};var Yo;(function(s){s.Refactor="refactor",s.RefactorPreview="refactor preview",s.Lightbulb="lightbulb",s.Default="other (default)",s.SourceAction="source action",s.QuickFix="quick fix action",s.FixAll="fix all",s.OrganizeImports="organize imports",s.AutoFix="auto fix",s.QuickFixHover="quick fix hover window",s.OnSave="save participants",s.ProblemsView="problems view"})(Yo||(Yo={}));function TIe(s,e){return!(s.include&&!s.include.intersects(e)||s.excludes&&s.excludes.some(t=>one(e,t,s.include))||!s.includeSourceActions&&ii.Source.contains(e))}function NIe(s,e){const t=e.kind?new _i(e.kind):void 0;return!(s.include&&(!t||!s.include.contains(t))||s.excludes&&t&&s.excludes.some(i=>one(t,i,s.include))||!s.includeSourceActions&&t&&ii.Source.contains(t)||s.onlyIncludePreferredActions&&!e.isPreferred)}function one(s,e,t){return!(!e.contains(s)||t&&e.contains(t))}class gc{static fromUser(e,t){return!e||typeof e!="object"?new gc(t.kind,t.apply,!1):new gc(gc.getKindFromUser(e,t.kind),gc.getApplyFromUser(e,t.apply),gc.getPreferredUser(e))}static getApplyFromUser(e,t){switch(typeof e.apply=="string"?e.apply.toLowerCase():""){case"first":return"first";case"never":return"never";case"ifsingle":return"ifSingle";default:return t}}static getKindFromUser(e,t){return typeof e.kind=="string"?new _i(e.kind):t}static getPreferredUser(e){return typeof e.preferred=="boolean"?e.preferred:!1}constructor(e,t,i){this.kind=e,this.apply=t,this.preferred=i}}class RIe{constructor(e,t,i){this.action=e,this.provider=t,this.highlightRange=i}async resolve(e){if(this.provider?.resolveCodeAction&&!this.action.edit){let t;try{t=await this.provider.resolveCodeAction(this.action,e)}catch(i){ji(i)}t&&(this.action.edit=t.edit)}return this}}const rne="editor.action.codeAction",tH="editor.action.quickFix",ane="editor.action.autoFix",lne="editor.action.refactor",cne="editor.action.sourceAction",j4="editor.action.organizeImports",G4="editor.action.fixAll",AIe=1e3;class Z1 extends F{static codeActionsPreferredComparator(e,t){return e.isPreferred&&!t.isPreferred?-1:!e.isPreferred&&t.isPreferred?1:0}static codeActionsComparator({action:e},{action:t}){return e.isAI&&!t.isAI?1:!e.isAI&&t.isAI?-1:Ms(e.diagnostics)?Ms(t.diagnostics)?Z1.codeActionsPreferredComparator(e,t):-1:Ms(t.diagnostics)?1:Z1.codeActionsPreferredComparator(e,t)}constructor(e,t,i){super(),this.documentation=t,this._register(i),this.allActions=[...e].sort(Z1.codeActionsComparator),this.validActions=this.allActions.filter(({action:n})=>!n.disabled)}get hasAutoFix(){return this.validActions.some(({action:e})=>!!e.kind&&ii.QuickFix.contains(new _i(e.kind))&&!!e.isPreferred)}get hasAIFix(){return this.validActions.some(({action:e})=>!!e.isAI)}get allAIFixes(){return this.validActions.every(({action:e})=>!!e.isAI)}}const uj={actions:[],documentation:void 0};async function Ab(s,e,t,i,n,o){const r=i.filter||{},a={...r,excludes:[...r.excludes||[],ii.Notebook]},l={only:r.include?.value,trigger:i.type},c=new Y7(e,o),d=i.type===2,h=MIe(s,e,d?a:r),u=new U,g=h.map(async p=>{const _=setTimeout(()=>n.report(p),1250);try{const b=await p.provideCodeActions(e,t,l,c.token);if(c.token.isCancellationRequested)return b?.dispose(),uj;b&&u.add(b);const w=(b?.actions||[]).filter(v=>v&&NIe(r,v)),C=PIe(p,w,r.include);return{actions:w.map(v=>new RIe(v,p)),documentation:C}}catch(b){if(Dr(b))throw b;return ji(b),uj}finally{clearTimeout(_)}}),f=s.onDidChange(()=>{const p=s.all(e);ni(p,h)||c.cancel()});try{const p=await Promise.all(g),_=p.map(C=>C.actions).flat(),b=[...Al(p.map(C=>C.documentation)),...OIe(s,e,i,_)],w=new Z1(_,b,u);return u.add(w),w}catch(p){throw u.dispose(),p}finally{f.dispose(),c.dispose()}}function MIe(s,e,t){return s.all(e).filter(i=>i.providedCodeActionKinds?i.providedCodeActionKinds.some(n=>TIe(t,new _i(n))):!0)}function*OIe(s,e,t,i){if(e&&i.length)for(const n of s.all(e))n._getAdditionalMenuItems&&(yield*n._getAdditionalMenuItems?.({trigger:t.type,only:t.filter?.include?.value},i.map(o=>o.action)))}function PIe(s,e,t){if(!s.documentation)return;const i=s.documentation.map(n=>({kind:new _i(n.kind),command:n.command}));if(t){let n;for(const o of i)o.kind.contains(t)&&(n?n.kind.contains(o.kind)&&(n=o):n=o);if(n)return n?.command}for(const n of e)if(n.kind){for(const o of i)if(o.kind.contains(new _i(n.kind)))return o.command}}var Lu;(function(s){s.OnSave="onSave",s.FromProblemsView="fromProblemsView",s.FromCodeActions="fromCodeActions",s.FromAILightbulb="fromAILightbulb",s.FromProblemsHover="fromProblemsHover"})(Lu||(Lu={}));async function FIe(s,e,t,i,n=Ke.None){const o=s.get(fx),r=s.get(Gt),a=s.get(vs),l=s.get(Ii),c=s.get(Th);if(a.publicLog2("codeAction.applyCodeAction",{codeActionTitle:e.action.title,codeActionKind:e.action.kind,codeActionIsPreferred:!!e.action.isPreferred,reason:t}),c.playSignal(Gs.codeActionTriggered),await e.resolve(n),!n.isCancellationRequested&&!(e.action.edit?.edits.length&&!(await o.apply(e.action.edit,{editor:i?.editor,label:e.action.title,quotableLabel:e.action.title,code:"undoredo.codeAction",respectAutoSaveConfig:t!==Lu.OnSave,showPreview:i?.preview,reason:cs.codeAction({kind:e.action.kind,providerId:fA.fromExtensionId(e.provider?.extensionId)})})).isApplied)){if(e.action.command)try{await r.executeCommand(e.action.command.id,...e.action.command.arguments||[])}catch(d){const h=BIe(d);l.error(typeof h=="string"?h:m(830,"An unknown error occurred while applying the code action"))}setTimeout(()=>c.playSignal(Gs.codeActionApplied),AIe)}}function BIe(s){return typeof s=="string"?s:s instanceof Error&&typeof s.message=="string"?s.message:void 0}at.registerCommand("_executeCodeActionProvider",async function(s,e,t,i,n){if(!(e instanceof Le))throw Qr();const{codeActionProvider:o}=s.get(ge),r=s.get(hi).getModel(e);if(!r)throw Qr();const a=he.isISelection(t)?he.liftSelection(t):k.isIRange(t)?r.validateRange(t):void 0;if(!a)throw Qr();const l=typeof i=="string"?new _i(i):void 0,c=await Ab(o,r,a,{type:1,triggerAction:Yo.Default,filter:{includeSourceActions:!0,include:l}},ml.None,Ke.None),d=[],h=Math.min(c.validActions.length,typeof n=="number"?n:0);for(let u=0;u<h;u++)d.push(c.validActions[u].resolve(Ke.None));try{return await Promise.all(d),c.validActions.map(u=>u.action)}finally{setTimeout(()=>c.dispose(),100)}});var mi;(function(s){s[s.Hint=1]="Hint",s[s.Info=2]="Info",s[s.Warning=4]="Warning",s[s.Error=8]="Error"})(mi||(mi={}));(function(s){function e(l,c){return c-l}s.compare=e;const t=Object.create(null);t[s.Error]=m(1732,"Error"),t[s.Warning]=m(1733,"Warning"),t[s.Info]=m(1734,"Info");function i(l){return t[l]||""}s.toString=i;const n=Object.create(null);n[s.Error]=m(1735,"Errors"),n[s.Warning]=m(1736,"Warnings"),n[s.Info]=m(1737,"Infos");function o(l){return n[l]||""}s.toStringPlural=o;function r(l){switch(l){case fi.Error:return s.Error;case fi.Warning:return s.Warning;case fi.Info:return s.Info;case fi.Ignore:return s.Hint}}s.fromSeverity=r;function a(l){switch(l){case s.Error:return fi.Error;case s.Warning:return fi.Warning;case s.Info:return fi.Info;case s.Hint:return fi.Ignore}}s.toSeverity=a})(mi||(mi={}));var jI;(function(s){function t(n){return i(n,!0)}s.makeKey=t;function i(n,o){const r=[""];return n.source?r.push(n.source.replace("¦","\\¦")):r.push(""),n.code?typeof n.code=="string"?r.push(n.code.replace("¦","\\¦")):r.push(n.code.value.replace("¦","\\¦")):r.push(""),n.severity!==void 0&&n.severity!==null?r.push(mi.toString(n.severity)):r.push(""),n.message&&o?r.push(n.message.replace("¦","\\¦")):r.push(""),n.startLineNumber!==void 0&&n.startLineNumber!==null?r.push(n.startLineNumber.toString()):r.push(""),n.startColumn!==void 0&&n.startColumn!==null?r.push(n.startColumn.toString()):r.push(""),n.endLineNumber!==void 0&&n.endLineNumber!==null?r.push(n.endLineNumber.toString()):r.push(""),n.endColumn!==void 0&&n.endColumn!==null?r.push(n.endColumn.toString()):r.push(""),r.push(""),r.join("¦")}s.makeKeyOptionalMessage=i})(jI||(jI={}));const jc=qe("markerService");var WIe=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},HIe=function(s,e){return function(t,i){e(t,i,s)}},K4,Tp;let Y4=(Tp=class{constructor(e){this.keybindingService=e}getResolver(){const e=new es(()=>this.keybindingService.getKeybindings().filter(t=>K4.codeActionCommands.indexOf(t.command)>=0).filter(t=>t.resolvedKeybinding).map(t=>{let i=t.commandArgs;return t.command===j4?i={kind:ii.SourceOrganizeImports.value}:t.command===G4&&(i={kind:ii.SourceFixAll.value}),{resolvedKeybinding:t.resolvedKeybinding,...gc.fromUser(i,{kind:_i.None,apply:"never"})}}));return t=>{if(t.kind)return this.bestKeybindingForCodeAction(t,e.value)?.resolvedKeybinding}}bestKeybindingForCodeAction(e,t){if(!e.kind)return;const i=new _i(e.kind);return t.filter(n=>n.kind.contains(i)).filter(n=>n.preferred?e.isPreferred:!0).reduceRight((n,o)=>n?n.kind.contains(o.kind)?o:n:o,void 0)}},K4=Tp,Tp.codeActionCommands=[lne,rne,cne,j4,G4],Tp);Y4=K4=WIe([HIe(0,ft)],Y4);T("symbolIcon.arrayForeground",et,m(1495,"The foreground color for array symbols. These symbols appear in the outline, breadcrumb, and suggest widget."));T("symbolIcon.booleanForeground",et,m(1496,"The foreground color for boolean symbols. These symbols appear in the outline, breadcrumb, and suggest widget."));T("symbolIcon.classForeground",{dark:"#EE9D28",light:"#D67E00",hcDark:"#EE9D28",hcLight:"#D67E00"},m(1497,"The foreground color for class symbols. These symbols appear in the outline, breadcrumb, and suggest widget."));T("symbolIcon.colorForeground",et,m(1498,"The foreground color for color symbols. These symbols appear in the outline, breadcrumb, and suggest widget."));T("symbolIcon.constantForeground",et,m(1499,"The foreground color for constant symbols. These symbols appear in the outline, breadcrumb, and suggest widget."));T("symbolIcon.constructorForeground",{dark:"#B180D7",light:"#652D90",hcDark:"#B180D7",hcLight:"#652D90"},m(1500,"The foreground color for constructor symbols. These symbols appear in the outline, breadcrumb, and suggest widget."));T("symbolIcon.enumeratorForeground",{dark:"#EE9D28",light:"#D67E00",hcDark:"#EE9D28",hcLight:"#D67E00"},m(1501,"The foreground color for enumerator symbols. These symbols appear in the outline, breadcrumb, and suggest widget."));T("symbolIcon.enumeratorMemberForeground",{dark:"#75BEFF",light:"#007ACC",hcDark:"#75BEFF",hcLight:"#007ACC"},m(1502,"The foreground color for enumerator member symbols. These symbols appear in the outline, breadcrumb, and suggest widget."));T("symbolIcon.eventForeground",{dark:"#EE9D28",light:"#D67E00",hcDark:"#EE9D28",hcLight:"#D67E00"},m(1503,"The foreground color for event symbols. These symbols appear in the outline, breadcrumb, and suggest widget."));T("symbolIcon.fieldForeground",{dark:"#75BEFF",light:"#007ACC",hcDark:"#75BEFF",hcLight:"#007ACC"},m(1504,"The foreground color for field symbols. These symbols appear in the outline, breadcrumb, and suggest widget."));T("symbolIcon.fileForeground",et,m(1505,"The foreground color for file symbols. These symbols appear in the outline, breadcrumb, and suggest widget."));T("symbolIcon.folderForeground",et,m(1506,"The foreground color for folder symbols. These symbols appear in the outline, breadcrumb, and suggest widget."));T("symbolIcon.functionForeground",{dark:"#B180D7",light:"#652D90",hcDark:"#B180D7",hcLight:"#652D90"},m(1507,"The foreground color for function symbols. These symbols appear in the outline, breadcrumb, and suggest widget."));T("symbolIcon.interfaceForeground",{dark:"#75BEFF",light:"#007ACC",hcDark:"#75BEFF",hcLight:"#007ACC"},m(1508,"The foreground color for interface symbols. These symbols appear in the outline, breadcrumb, and suggest widget."));T("symbolIcon.keyForeground",et,m(1509,"The foreground color for key symbols. These symbols appear in the outline, breadcrumb, and suggest widget."));T("symbolIcon.keywordForeground",et,m(1510,"The foreground color for keyword symbols. These symbols appear in the outline, breadcrumb, and suggest widget."));T("symbolIcon.methodForeground",{dark:"#B180D7",light:"#652D90",hcDark:"#B180D7",hcLight:"#652D90"},m(1511,"The foreground color for method symbols. These symbols appear in the outline, breadcrumb, and suggest widget."));T("symbolIcon.moduleForeground",et,m(1512,"The foreground color for module symbols. These symbols appear in the outline, breadcrumb, and suggest widget."));T("symbolIcon.namespaceForeground",et,m(1513,"The foreground color for namespace symbols. These symbols appear in the outline, breadcrumb, and suggest widget."));T("symbolIcon.nullForeground",et,m(1514,"The foreground color for null symbols. These symbols appear in the outline, breadcrumb, and suggest widget."));T("symbolIcon.numberForeground",et,m(1515,"The foreground color for number symbols. These symbols appear in the outline, breadcrumb, and suggest widget."));T("symbolIcon.objectForeground",et,m(1516,"The foreground color for object symbols. These symbols appear in the outline, breadcrumb, and suggest widget."));T("symbolIcon.operatorForeground",et,m(1517,"The foreground color for operator symbols. These symbols appear in the outline, breadcrumb, and suggest widget."));T("symbolIcon.packageForeground",et,m(1518,"The foreground color for package symbols. These symbols appear in the outline, breadcrumb, and suggest widget."));T("symbolIcon.propertyForeground",et,m(1519,"The foreground color for property symbols. These symbols appear in the outline, breadcrumb, and suggest widget."));T("symbolIcon.referenceForeground",et,m(1520,"The foreground color for reference symbols. These symbols appear in the outline, breadcrumb, and suggest widget."));T("symbolIcon.snippetForeground",et,m(1521,"The foreground color for snippet symbols. These symbols appear in the outline, breadcrumb, and suggest widget."));T("symbolIcon.stringForeground",et,m(1522,"The foreground color for string symbols. These symbols appear in the outline, breadcrumb, and suggest widget."));T("symbolIcon.structForeground",et,m(1523,"The foreground color for struct symbols. These symbols appear in the outline, breadcrumb, and suggest widget."));T("symbolIcon.textForeground",et,m(1524,"The foreground color for text symbols. These symbols appear in the outline, breadcrumb, and suggest widget."));T("symbolIcon.typeParameterForeground",et,m(1525,"The foreground color for type parameter symbols. These symbols appear in the outline, breadcrumb, and suggest widget."));T("symbolIcon.unitForeground",et,m(1526,"The foreground color for unit symbols. These symbols appear in the outline, breadcrumb, and suggest widget."));T("symbolIcon.variableForeground",{dark:"#75BEFF",light:"#007ACC",hcDark:"#75BEFF",hcLight:"#007ACC"},m(1527,"The foreground color for variable symbols. These symbols appear in the outline, breadcrumb, and suggest widget."));const dne=Object.freeze({kind:_i.Empty,title:m(866,"More Actions...")}),VIe=Object.freeze([{kind:ii.QuickFix,title:m(867,"Quick Fix")},{kind:ii.RefactorExtract,title:m(868,"Extract"),icon:Z.wrench},{kind:ii.RefactorInline,title:m(869,"Inline"),icon:Z.wrench},{kind:ii.RefactorRewrite,title:m(870,"Rewrite"),icon:Z.wrench},{kind:ii.RefactorMove,title:m(871,"Move"),icon:Z.wrench},{kind:ii.SurroundWith,title:m(872,"Surround With"),icon:Z.surroundWith},{kind:ii.Source,title:m(873,"Source Action"),icon:Z.symbolFile},dne]);function zIe(s,e,t){if(!e)return s.map(o=>({kind:"action",item:o,group:dne,disabled:!!o.action.disabled,label:o.action.disabled||o.action.title,canPreview:!!o.action.edit?.edits.length}));const i=VIe.map(o=>({group:o,actions:[]}));for(const o of s){const r=o.action.kind?new _i(o.action.kind):_i.None;for(const a of i)if(a.group.kind.contains(r)){a.actions.push(o);break}}const n=[];for(const o of i)if(o.actions.length){n.push({kind:"header",group:o.group});for(const r of o.actions){const a=o.group;n.push({kind:"action",item:r,group:r.action.isAI?{title:a.title,kind:a.kind,icon:Z.sparkle}:a,label:r.action.title,disabled:!!r.action.disabled,keybinding:t(r.action)})}}return n}const hne=new ce("supportedCodeAction",""),gj="_typescript.applyFixAllCodeAction";class $Ie extends F{constructor(e,t,i,n=250){super(),this._editor=e,this._markerService=t,this._signalChange=i,this._delay=n,this._autoTriggerTimer=this._register(new $o),this._register(this._markerService.onMarkerChanged(o=>this._onMarkerChanges(o))),this._register(this._editor.onDidChangeCursorPosition(()=>this._tryAutoTrigger()))}trigger(e){const t=this._getRangeOfSelectionUnlessWhitespaceEnclosed(e);this._signalChange(t?{trigger:e,selection:t}:void 0)}_onMarkerChanges(e){const t=this._editor.getModel();t&&e.some(i=>dg(i,t.uri))&&this._tryAutoTrigger()}_tryAutoTrigger(){this._autoTriggerTimer.cancelAndSet(()=>{this.trigger({type:2,triggerAction:Yo.Default})},this._delay)}_getRangeOfSelectionUnlessWhitespaceEnclosed(e){if(!this._editor.hasModel())return;const t=this._editor.getSelection();if(e.type===1)return t;const i=this._editor.getOption(73).enabled;if(i!==xa.Off){{if(i===xa.On)return t;if(i===xa.OnCode){if(!t.isEmpty())return t;const o=this._editor.getModel(),{lineNumber:r,column:a}=t.getPosition(),l=o.getLineContent(r);if(l.length===0)return;if(a===1){if(/\s/.test(l[0]))return}else if(a===o.getLineMaxColumn(r)){if(/\s/.test(l[l.length-1]))return}else if(/\s/.test(l[a-2])&&/\s/.test(l[a-1]))return}}return t}}}var yf;(function(s){s.Empty={type:0};class e{constructor(i,n,o){this.trigger=i,this.position=n,this._cancellablePromise=o,this.type=1,this.actions=o.catch(r=>{if(Dr(r))return Z4;throw r})}cancel(){this._cancellablePromise.cancel()}}s.Triggered=e})(yf||(yf={}));const Z4=Object.freeze({allActions:[],validActions:[],dispose:()=>{},documentation:[],hasAutoFix:!1,hasAIFix:!1,allAIFixes:!1});class UIe extends F{constructor(e,t,i,n,o,r){super(),this._editor=e,this._registry=t,this._markerService=i,this._progressService=o,this._configurationService=r,this._codeActionOracle=this._register(new vt),this._state=yf.Empty,this._onDidChangeState=this._register(new M),this.onDidChangeState=this._onDidChangeState.event,this.codeActionsDisposable=this._register(new vt),this._disposed=!1,this._supportedCodeActions=hne.bindTo(n),this._register(this._editor.onDidChangeModel(()=>this._update())),this._register(this._editor.onDidChangeModelLanguage(()=>this._update())),this._register(this._registry.onDidChange(()=>this._update())),this._register(this._editor.onDidChangeConfiguration(a=>{a.hasChanged(73)&&this._update()})),this._update()}dispose(){this._disposed||(this._disposed=!0,super.dispose(),this.setState(yf.Empty,!0))}_settingEnabledNearbyQuickfixes(){const e=this._editor?.getModel();return this._configurationService?this._configurationService.getValue("editor.codeActionWidget.includeNearbyQuickFixes",{resource:e?.uri}):!1}_update(){if(this._disposed)return;this._codeActionOracle.value=void 0,this.setState(yf.Empty);const e=this._editor.getModel();if(e&&this._registry.has(e)&&!this._editor.getOption(104)){const t=this._registry.all(e).flatMap(i=>i.providedCodeActionKinds??[]);this._supportedCodeActions.set(t.join(" ")),this._codeActionOracle.value=new $Ie(this._editor,this._markerService,i=>{if(!i){this.setState(yf.Empty);return}const n=i.selection.getStartPosition(),o=an(async l=>{if(this._settingEnabledNearbyQuickfixes()&&i.trigger.type===1&&(i.trigger.triggerAction===Yo.QuickFix||i.trigger.filter?.include?.contains(ii.QuickFix))){const d=await Ab(this._registry,e,i.selection,i.trigger,ml.None,l);this.codeActionsDisposable.value=d;const h=[...d.allActions];if(l.isCancellationRequested)return d.dispose(),Z4;const u=d.validActions?.some(f=>f.action.kind&&ii.QuickFix.contains(new _i(f.action.kind))&&!f.action.isAI),g=this._markerService.read({resource:e.uri});if(u){for(const f of d.validActions)f.action.command?.arguments?.some(p=>typeof p=="string"&&p.includes(gj))&&(f.action.diagnostics=[...g.filter(p=>p.relatedInformation)]);return{validActions:d.validActions,allActions:h,documentation:d.documentation,hasAutoFix:d.hasAutoFix,hasAIFix:d.hasAIFix,allAIFixes:d.allAIFixes,dispose:()=>{this.codeActionsDisposable.value=d}}}else if(!u&&g.length>0){const f=i.selection.getPosition();let p=f,_=Number.MAX_VALUE;const b=[...d.validActions];for(const C of g){const v=C.endColumn,y=C.endLineNumber,x=C.startLineNumber;if(y===f.lineNumber||x===f.lineNumber){p=new A(y,v);const L={type:i.trigger.type,triggerAction:i.trigger.triggerAction,filter:{include:i.trigger.filter?.include?i.trigger.filter?.include:ii.QuickFix},autoApply:i.trigger.autoApply,context:{notAvailableMessage:i.trigger.context?.notAvailableMessage||"",position:p}},E=new he(p.lineNumber,p.column,p.lineNumber,p.column),I=await Ab(this._registry,e,E,L,ml.None,l);if(l.isCancellationRequested)return I.dispose(),Z4;if(I.validActions.length!==0){for(const R of I.validActions)R.action.command?.arguments?.some(O=>typeof O=="string"&&O.includes(gj))&&(R.action.diagnostics=[...g.filter(O=>O.relatedInformation)]);d.allActions.length===0&&h.push(...I.allActions),Math.abs(f.column-v)<_?b.unshift(...I.validActions):b.push(...I.validActions)}_=Math.abs(f.column-v)}}const w=b.filter((C,v,y)=>y.findIndex(x=>x.action.title===C.action.title)===v);return w.sort((C,v)=>C.action.isPreferred&&!v.action.isPreferred?-1:!C.action.isPreferred&&v.action.isPreferred||C.action.isAI&&!v.action.isAI?1:!C.action.isAI&&v.action.isAI?-1:0),{validActions:w,allActions:h,documentation:d.documentation,hasAutoFix:d.hasAutoFix,hasAIFix:d.hasAIFix,allAIFixes:d.allAIFixes,dispose:()=>{this.codeActionsDisposable.value=d}}}}if(i.trigger.type===1){const d=await Ab(this._registry,e,i.selection,i.trigger,ml.None,l);return this.codeActionsDisposable.value=d,d}const c=await Ab(this._registry,e,i.selection,i.trigger,ml.None,l);return this.codeActionsDisposable.value=c,c});i.trigger.type===1&&this._progressService?.showWhile(o,250);const r=new yf.Triggered(i.trigger,n,o);let a=!1;this._state.type===1&&(a=this._state.trigger.type===1&&r.type===1&&r.trigger.type===2&&this._state.position!==r.position),a?setTimeout(()=>{this.setState(r)},500):this.setState(r)},void 0),this._codeActionOracle.value.trigger({type:2,triggerAction:Yo.Default})}else this._supportedCodeActions.reset()}trigger(e){this._codeActionOracle.value?.trigger(e),this.codeActionsDisposable.dispose()}setState(e,t){e!==this._state&&(this._state.type===1&&this._state.cancel(),this._state=e,!t&&!this._disposed&&this._onDidChangeState.fire(e))}}var qIe=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},jIe=function(s,e){return function(t,i){e(t,i,s)}},gb;const fj=Qt("gutter-lightbulb",Z.lightBulb,m(874,"Icon which spawns code actions menu from the gutter when there is no space in the editor.")),pj=Qt("gutter-lightbulb-auto-fix",Z.lightbulbAutofix,m(875,"Icon which spawns code actions menu from the gutter when there is no space in the editor and a quick fix is available.")),mj=Qt("gutter-lightbulb-sparkle",Z.lightbulbSparkle,m(876,"Icon which spawns code actions menu from the gutter when there is no space in the editor and an AI fix is available.")),_j=Qt("gutter-lightbulb-aifix-auto-fix",Z.lightbulbSparkleAutofix,m(877,"Icon which spawns code actions menu from the gutter when there is no space in the editor and an AI fix and a quick fix is available.")),bj=Qt("gutter-lightbulb-sparkle-filled",Z.sparkleFilled,m(878,"Icon which spawns code actions menu from the gutter when there is no space in the editor and an AI fix and a quick fix is available."));var Qa;(function(s){s.Hidden={type:0};class e{constructor(i,n,o,r){this.actions=i,this.trigger=n,this.editorPosition=o,this.widgetPosition=r,this.type=1}}s.Showing=e})(Qa||(Qa={}));var Xd;let $S=(Xd=class extends F{constructor(e,t){super(),this._editor=e,this._keybindingService=t,this._onClick=this._register(new M),this.onClick=this._onClick.event,this._state=Qa.Hidden,this._gutterState=Qa.Hidden,this._iconClasses=[],this.lightbulbClasses=["codicon-"+fj.id,"codicon-"+_j.id,"codicon-"+pj.id,"codicon-"+mj.id,"codicon-"+bj.id],this.gutterDecoration=gb.GUTTER_DECORATION,this._domNode=ie("div.lightBulbWidget"),this._domNode.role="listbox",this._register(_s.ignoreTarget(this._domNode)),this._editor.addContentWidget(this),this._register(this._editor.onDidChangeModelContent(i=>{const n=this._editor.getModel();(this.state.type!==1||!n||this.state.editorPosition.lineNumber>=n.getLineCount())&&this.hide(),(this.gutterState.type!==1||!n||this.gutterState.editorPosition.lineNumber>=n.getLineCount())&&this.gutterHide()})),this._register(iue(this._domNode,i=>{if(this.state.type!==1)return;this._editor.focus(),i.preventDefault();const{top:n,height:o}=Li(this._domNode),r=this._editor.getOption(75);let a=Math.floor(r/3);this.state.widgetPosition.position!==null&&this.state.widgetPosition.position.lineNumber<this.state.editorPosition.lineNumber&&(a+=r),this._onClick.fire({x:i.posx,y:n+o+a,actions:this.state.actions,trigger:this.state.trigger})})),this._register(W(this._domNode,"mouseenter",i=>{(i.buttons&1)===1&&this.hide()})),this._register(oe.runAndSubscribe(this._keybindingService.onDidUpdateKeybindings,()=>{this._preferredKbLabel=this._keybindingService.lookupKeybinding(ane)?.getLabel()??void 0,this._quickFixKbLabel=this._keybindingService.lookupKeybinding(tH)?.getLabel()??void 0,this._updateLightBulbTitleAndIcon()})),this._register(this._editor.onMouseDown(async i=>{if(!i.target.element||!this.lightbulbClasses.some(l=>i.target.element&&i.target.element.classList.contains(l))||this.gutterState.type!==1)return;this._editor.focus();const{top:n,height:o}=Li(i.target.element),r=this._editor.getOption(75);let a=Math.floor(r/3);this.gutterState.widgetPosition.position!==null&&this.gutterState.widgetPosition.position.lineNumber<this.gutterState.editorPosition.lineNumber&&(a+=r),this._onClick.fire({x:i.event.posx,y:n+o+a,actions:this.gutterState.actions,trigger:this.gutterState.trigger})}))}dispose(){super.dispose(),this._editor.removeContentWidget(this),this._gutterDecorationID&&this._removeGutterDecoration(this._gutterDecorationID)}getId(){return"LightBulbWidget"}getDomNode(){return this._domNode}getPosition(){return this._state.type===1?this._state.widgetPosition:null}update(e,t,i){if(e.validActions.length<=0)return this.gutterHide(),this.hide();if(!this._editor.hasTextFocus())return this.gutterHide(),this.hide();if(!this._editor.getOptions().get(73).enabled)return this.gutterHide(),this.hide();const r=this._editor.getModel();if(!r)return this.gutterHide(),this.hide();const{lineNumber:a,column:l}=r.validatePosition(i),c=r.getOptions().tabSize,d=this._editor.getOptions().get(59),h=r.getLineContent(a),u=uA(h,c),g=d.spaceWidth*u>22,f=y=>y>2&&this._editor.getTopForLineNumber(y)===this._editor.getTopForLineNumber(y-1),p=this._editor.getLineDecorations(a);let _=!1;if(p)for(const y of p){const x=y.options.glyphMarginClassName;if(x&&!this.lightbulbClasses.some(L=>x.includes(L))){_=!0;break}}let b=a,w=1;if(!g){const y=x=>{const L=r.getLineContent(x);return/^\s*$|^\s+/.test(L)||L.length<=w};if(a>1&&!f(a-1)){const x=r.getLineCount(),L=a===x,E=a>1&&y(a-1),I=!L&&y(a+1),R=y(a),O=!I&&!E;if(!I&&!E&&!_)return this.gutterState=new Qa.Showing(e,t,i,{position:{lineNumber:b,column:w},preference:gb._posPref}),this.renderGutterLightbub(),this.hide();E||L||E&&!R?b-=1:(I||O&&R)&&(b+=1)}else if(a===1&&(a===r.getLineCount()||!y(a+1)&&!y(a)))if(this.gutterState=new Qa.Showing(e,t,i,{position:{lineNumber:b,column:w},preference:gb._posPref}),_)this.gutterHide();else return this.renderGutterLightbub(),this.hide();else if(a<r.getLineCount()&&!f(a+1))b+=1;else if(l*d.spaceWidth<22)return this.hide();w=/^\S\s*$/.test(r.getLineContent(b))?2:1}this.state=new Qa.Showing(e,t,i,{position:{lineNumber:b,column:w},preference:gb._posPref}),this._gutterDecorationID&&(this._removeGutterDecoration(this._gutterDecorationID),this.gutterHide());const C=e.validActions,v=e.validActions[0].action.kind;if(C.length!==1||!v){this._editor.layoutContentWidget(this);return}this._editor.layoutContentWidget(this)}hide(){this.state!==Qa.Hidden&&(this.state=Qa.Hidden,this._editor.layoutContentWidget(this))}gutterHide(){this.gutterState!==Qa.Hidden&&(this._gutterDecorationID&&this._removeGutterDecoration(this._gutterDecorationID),this.gutterState=Qa.Hidden)}get state(){return this._state}set state(e){this._state=e,this._updateLightBulbTitleAndIcon()}get gutterState(){return this._gutterState}set gutterState(e){this._gutterState=e,this._updateGutterLightBulbTitleAndIcon()}_updateLightBulbTitleAndIcon(){if(this._domNode.classList.remove(...this._iconClasses),this._iconClasses=[],this.state.type!==1)return;let e,t=!1;this.state.actions.allAIFixes?(e=Z.sparkleFilled,this.state.actions.validActions.length===1&&(t=!0)):this.state.actions.hasAutoFix?this.state.actions.hasAIFix?e=Z.lightbulbSparkleAutofix:e=Z.lightbulbAutofix:this.state.actions.hasAIFix?e=Z.lightbulbSparkle:e=Z.lightBulb,this._updateLightbulbTitle(this.state.actions.hasAutoFix,t),this._iconClasses=xe.asClassNameArray(e),this._domNode.classList.add(...this._iconClasses)}_updateGutterLightBulbTitleAndIcon(){if(this.gutterState.type!==1)return;let e,t=!1;this.gutterState.actions.allAIFixes?(e=bj,this.gutterState.actions.validActions.length===1&&(t=!0)):this.gutterState.actions.hasAutoFix?this.gutterState.actions.hasAIFix?e=_j:e=pj:this.gutterState.actions.hasAIFix?e=mj:e=fj,this._updateLightbulbTitle(this.gutterState.actions.hasAutoFix,t);const i=Fe.register({description:"codicon-gutter-lightbulb-decoration",glyphMarginClassName:xe.asClassName(e),glyphMargin:{position:Ll.Left},stickiness:1});this.gutterDecoration=i}renderGutterLightbub(){const e=this._editor.getSelection();e&&(this._gutterDecorationID===void 0?this._addGutterDecoration(e.startLineNumber):this._updateGutterDecoration(this._gutterDecorationID,e.startLineNumber))}_addGutterDecoration(e){this._editor.changeDecorations(t=>{this._gutterDecorationID=t.addDecoration(new k(e,0,e,0),this.gutterDecoration)})}_removeGutterDecoration(e){this._editor.changeDecorations(t=>{t.removeDecoration(e),this._gutterDecorationID=void 0})}_updateGutterDecoration(e,t){this._editor.changeDecorations(i=>{i.changeDecoration(e,new k(t,0,t,0)),i.changeDecorationOptions(e,this.gutterDecoration)})}_updateLightbulbTitle(e,t){this.state.type===1&&(t?this.title=m(879,"Run: {0}",this.state.actions.validActions[0].action.title):e&&this._preferredKbLabel?this.title=m(880,"Show Code Actions. Preferred Quick Fix Available ({0})",this._preferredKbLabel):!e&&this._quickFixKbLabel?this.title=m(881,"Show Code Actions ({0})",this._quickFixKbLabel):e||(this.title=m(882,"Show Code Actions")))}set title(e){this._domNode.title=e}},gb=Xd,Xd.GUTTER_DECORATION=Fe.register({description:"codicon-gutter-lightbulb-decoration",glyphMarginClassName:xe.asClassName(Z.lightBulb),glyphMargin:{position:Ll.Left},stickiness:1}),Xd.ID="editor.contrib.lightbulbWidget",Xd._posPref=[0],Xd);$S=gb=qIe([jIe(1,ft)],$S);var GIe=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},Hl=function(s,e){return function(t,i){e(t,i,s)}},fb;const KIe="quickfix-edit-highlight";var qu;let Um=(qu=class extends F{static get(e){return e.getContribution(fb.ID)}constructor(e,t,i,n,o,r,a,l,c,d,h){super(),this._commandService=a,this._configurationService=l,this._actionWidgetService=c,this._instantiationService=d,this._progressService=h,this._activeCodeActions=this._register(new vt),this._showDisabled=!1,this._disposed=!1,this._editor=e,this._model=this._register(new UIe(this._editor,o.codeActionProvider,t,i,r,l)),this._register(this._model.onDidChangeState(u=>this.update(u))),this._lightBulbWidget=new es(()=>{const u=this._editor.getContribution($S.ID);return u&&this._register(u.onClick(g=>this.showCodeActionsFromLightbulb(g.actions,g))),u}),this._resolver=n.createInstance(Y4),this._register(this._editor.onDidLayoutChange(()=>this._actionWidgetService.hide()))}dispose(){this._disposed=!0,super.dispose()}async showCodeActionsFromLightbulb(e,t){if(e.allAIFixes&&e.validActions.length===1){const i=e.validActions[0],n=i.action.command;n&&n.id==="inlineChat.start"&&n.arguments&&n.arguments.length>=1&&n.arguments[0]&&(n.arguments[0]={...n.arguments[0],autoSend:!1}),await this.applyCodeAction(i,!1,!1,Lu.FromAILightbulb);return}await this.showCodeActionList(e,t,{includeDisabledActions:!1,fromLightbulb:!0})}showCodeActions(e,t,i){return this.showCodeActionList(t,i,{includeDisabledActions:!1,fromLightbulb:!1})}manualTriggerAtCurrentPosition(e,t,i,n){if(!this._editor.hasModel())return;Vo.get(this._editor)?.closeMessage();const o=this._editor.getPosition();this._trigger({type:1,triggerAction:t,filter:i,autoApply:n,context:{notAvailableMessage:e,position:o}})}_trigger(e){return this._model.trigger(e)}async applyCodeAction(e,t,i,n){const o=this._progressService.show(!0,500);try{await this._instantiationService.invokeFunction(FIe,e,n,{preview:i,editor:this._editor})}finally{t&&this._trigger({type:2,triggerAction:Yo.QuickFix,filter:{}}),o.done()}}hideLightBulbWidget(){this._lightBulbWidget.rawValue?.hide(),this._lightBulbWidget.rawValue?.gutterHide()}async update(e){if(e.type!==1){this.hideLightBulbWidget();return}let t;try{t=await e.actions}catch(n){Oe(n);return}if(!(this._disposed||this._editor.getSelection()?.startLineNumber!==e.position.lineNumber))if(this._lightBulbWidget.value?.update(t,e.trigger,e.position),e.trigger.type===1){if(e.trigger.filter?.include){const o=this.tryGetValidActionToApply(e.trigger,t);if(o){try{this.hideLightBulbWidget(),await this.applyCodeAction(o,!1,!1,Lu.FromCodeActions)}finally{t.dispose()}return}if(e.trigger.context){const r=this.getInvalidActionThatWouldHaveBeenApplied(e.trigger,t);if(r&&r.action.disabled){Vo.get(this._editor)?.showMessage(r.action.disabled,e.trigger.context.position),t.dispose();return}}}const n=!!e.trigger.filter?.include;if(e.trigger.context&&(!t.allActions.length||!n&&!t.validActions.length)){Vo.get(this._editor)?.showMessage(e.trigger.context.notAvailableMessage,e.trigger.context.position),this._activeCodeActions.value=t,t.dispose();return}this._activeCodeActions.value=t,this.showCodeActionList(t,this.toCoords(e.position),{includeDisabledActions:n,fromLightbulb:!1})}else this._actionWidgetService.isVisible?t.dispose():this._activeCodeActions.value=t}getInvalidActionThatWouldHaveBeenApplied(e,t){if(t.allActions.length&&(e.autoApply==="first"&&t.validActions.length===0||e.autoApply==="ifSingle"&&t.allActions.length===1))return t.allActions.find(({action:i})=>i.disabled)}tryGetValidActionToApply(e,t){if(t.validActions.length&&(e.autoApply==="first"&&t.validActions.length>0||e.autoApply==="ifSingle"&&t.validActions.length===1))return t.validActions[0]}async showCodeActionList(e,t,i){const n=this._editor.createDecorationsCollection(),o=this._editor.getDomNode();if(!o)return;const r=i.includeDisabledActions&&(this._showDisabled||e.validActions.length===0)?e.allActions:e.validActions;if(!r.length)return;const a=A.isIPosition(t)?this.toCoords(t):t,l={onSelect:async(c,d)=>{this.applyCodeAction(c,!0,!!d,i.fromLightbulb?Lu.FromAILightbulb:Lu.FromCodeActions),this._actionWidgetService.hide(!1),n.clear()},onHide:c=>{this._editor?.focus(),n.clear()},onHover:async(c,d)=>{if(d.isCancellationRequested)return;let h=!1;const u=c.action.kind;if(u){const g=new _i(u);h=[ii.RefactorExtract,ii.RefactorInline,ii.RefactorRewrite,ii.RefactorMove,ii.Source].some(p=>p.contains(g))}return{canPreview:h||!!c.action.edit?.edits.length}},onFocus:c=>{if(c&&c.action){const d=c.action.ranges,h=c.action.diagnostics;if(n.clear(),d&&d.length>0){const u=h&&h?.length>1?h.map(g=>({range:g,options:fb.DECORATION})):d.map(g=>({range:g,options:fb.DECORATION}));n.set(u)}else if(h&&h.length>0){const u=h.map(f=>({range:f,options:fb.DECORATION}));n.set(u);const g=h[0];if(g.startLineNumber&&g.startColumn){const f=this._editor.getModel()?.getWordAtPosition({lineNumber:g.startLineNumber,column:g.startColumn})?.word;xl(m(863,"Context: {0} at line {1} and column {2}.",f,g.startLineNumber,g.startColumn))}}}else n.clear()}};this._actionWidgetService.show("codeActionWidget",!0,zIe(r,this._shouldShowHeaders(),this._resolver.getResolver()),l,a,o,this._getActionBarActions(e,t,i))}toCoords(e){if(!this._editor.hasModel())return{x:0,y:0};this._editor.revealPosition(e,1),this._editor.render();const t=this._editor.getScrolledVisiblePosition(e),i=Li(this._editor.getDomNode()),n=i.left+t.left,o=i.top+t.top+t.height;return{x:n,y:o}}_shouldShowHeaders(){const e=this._editor?.getModel();return this._configurationService.getValue("editor.codeActionWidget.showHeaders",{resource:e?.uri})}_getActionBarActions(e,t,i){if(i.fromLightbulb)return[];const n=e.documentation.map(o=>({id:o.id,label:o.title,tooltip:o.tooltip??"",class:void 0,enabled:!0,run:()=>this._commandService.executeCommand(o.id,...o.arguments??[])}));return i.includeDisabledActions&&e.validActions.length>0&&e.allActions.length!==e.validActions.length&&n.push(this._showDisabled?{id:"hideMoreActions",label:m(864,"Hide Disabled"),enabled:!0,tooltip:"",class:void 0,run:()=>(this._showDisabled=!1,this.showCodeActionList(e,t,i))}:{id:"showMoreActions",label:m(865,"Show Disabled"),enabled:!0,tooltip:"",class:void 0,run:()=>(this._showDisabled=!0,this.showCodeActionList(e,t,i))}),n}},fb=qu,qu.ID="editor.contrib.codeActionController",qu.DECORATION=Fe.register({description:"quickfix-highlight",className:KIe}),qu);Um=fb=GIe([Hl(1,jc),Hl(2,Ae),Hl(3,_e),Hl(4,ge),Hl(5,Ch),Hl(6,Gt),Hl(7,He),Hl(8,Rg),Hl(9,_e),Hl(10,Ch)],Um);ua((s,e)=>{((n,o)=>{o&&e.addRule(`.monaco-editor ${n} { background-color: ${o}; }`)})(".quickfix-edit-highlight",s.getColor(zd));const i=s.getColor(_u);i&&e.addRule(`.monaco-editor .quickfix-edit-highlight { border: 1px ${Cl(s.type)?"dotted":"solid"} ${i}; box-sizing: border-box; }`)});function _x(s){return K.regex(hne.keys()[0],new RegExp("(\\s|^)"+Uo(s.value)+"\\b"))}const iH={type:"object",defaultSnippets:[{body:{kind:""}}],properties:{kind:{type:"string",description:m(831,"Kind of the code action to run.")},apply:{type:"string",description:m(832,"Controls when the returned actions are applied."),default:"ifSingle",enum:["first","ifSingle","never"],enumDescriptions:[m(833,"Always apply the first returned code action."),m(834,"Apply the first returned code action if it is the only one."),m(835,"Do not apply the returned code actions.")]},preferred:{type:"boolean",default:!1,description:m(836,"Controls if only preferred code actions should be returned.")}}};function m_(s,e,t,i,n=Yo.Default){s.hasModel()&&Um.get(s)?.manualTriggerAtCurrentPosition(e,n,t,i)}class YIe extends ue{constructor(){super({id:tH,label:V(853,"Quick Fix..."),precondition:K.and(N.writable,N.hasCodeActionsProvider),kbOpts:{kbExpr:N.textInputFocus,primary:2137,weight:100}})}run(e,t){return m_(t,m(837,"No code actions available"),void 0,void 0,Yo.QuickFix)}}class ZIe extends hn{constructor(){super({id:rne,precondition:K.and(N.writable,N.hasCodeActionsProvider),metadata:{description:"Trigger a code action",args:[{name:"args",schema:iH}]}})}runEditorCommand(e,t,i){const n=gc.fromUser(i,{kind:_i.Empty,apply:"ifSingle"});return m_(t,typeof i?.kind=="string"?n.preferred?m(838,"No preferred code actions for '{0}' available",i.kind):m(839,"No code actions for '{0}' available",i.kind):n.preferred?m(840,"No preferred code actions available"):m(841,"No code actions available"),{include:n.kind,includeSourceActions:!0,onlyIncludePreferredActions:n.preferred},n.apply)}}class XIe extends ue{constructor(){super({id:lne,label:V(854,"Refactor..."),precondition:K.and(N.writable,N.hasCodeActionsProvider),kbOpts:{kbExpr:N.textInputFocus,primary:3120,mac:{primary:1328},weight:100},contextMenuOpts:{group:"1_modification",order:2,when:K.and(N.writable,_x(ii.Refactor))},metadata:{description:"Refactor...",args:[{name:"args",schema:iH}]}})}run(e,t,i){const n=gc.fromUser(i,{kind:ii.Refactor,apply:"never"});return m_(t,typeof i?.kind=="string"?n.preferred?m(842,"No preferred refactorings for '{0}' available",i.kind):m(843,"No refactorings for '{0}' available",i.kind):n.preferred?m(844,"No preferred refactorings available"):m(845,"No refactorings available"),{include:ii.Refactor.contains(n.kind)?n.kind:_i.None,onlyIncludePreferredActions:n.preferred},n.apply,Yo.Refactor)}}class QIe extends ue{constructor(){super({id:cne,label:V(855,"Source Action..."),precondition:K.and(N.writable,N.hasCodeActionsProvider),contextMenuOpts:{group:"1_modification",order:2.1,when:K.and(N.writable,_x(ii.Source))},metadata:{description:"Source Action...",args:[{name:"args",schema:iH}]}})}run(e,t,i){const n=gc.fromUser(i,{kind:ii.Source,apply:"never"});return m_(t,typeof i?.kind=="string"?n.preferred?m(846,"No preferred source actions for '{0}' available",i.kind):m(847,"No source actions for '{0}' available",i.kind):n.preferred?m(848,"No preferred source actions available"):m(849,"No source actions available"),{include:ii.Source.contains(n.kind)?n.kind:_i.None,includeSourceActions:!0,onlyIncludePreferredActions:n.preferred},n.apply,Yo.SourceAction)}}class JIe extends ue{constructor(){super({id:j4,label:V(856,"Organize Imports"),precondition:K.and(N.writable,_x(ii.SourceOrganizeImports)),kbOpts:{kbExpr:N.textInputFocus,primary:1581,weight:100},metadata:{description:V(857,"Organize imports in the current file. Also called 'Optimize Imports' by some tools")}})}run(e,t){return m_(t,m(850,"No organize imports action available"),{include:ii.SourceOrganizeImports,includeSourceActions:!0},"ifSingle",Yo.OrganizeImports)}}class eDe extends ue{constructor(){super({id:G4,label:V(858,"Fix All"),precondition:K.and(N.writable,_x(ii.SourceFixAll))})}run(e,t){return m_(t,m(851,"No fix all action available"),{include:ii.SourceFixAll,includeSourceActions:!0},"ifSingle",Yo.FixAll)}}class tDe extends ue{constructor(){super({id:ane,label:V(859,"Auto Fix..."),precondition:K.and(N.writable,_x(ii.QuickFix)),kbOpts:{kbExpr:N.textInputFocus,primary:1625,mac:{primary:2649},weight:100}})}run(e,t){return m_(t,m(852,"No auto fixes available"),{include:ii.QuickFix,onlyIncludePreferredActions:!0},"ifSingle",Yo.AutoFix)}}lt(Um.ID,Um,3);lt($S.ID,$S,4);ae(YIe);ae(XIe);ae(QIe);ae(JIe);ae(tDe);ae(eDe);le(new ZIe);bi.as(Ml.Configuration).registerConfiguration({...EA,properties:{"editor.codeActionWidget.showHeaders":{type:"boolean",scope:6,description:m(860,"Enable/disable showing group headers in the Code Action menu."),default:!0}}});bi.as(Ml.Configuration).registerConfiguration({...EA,properties:{"editor.codeActionWidget.includeNearbyQuickFixes":{type:"boolean",scope:6,description:m(861,"Enable/disable showing nearest Quick Fix within a line when not currently on a diagnostic."),default:!0}}});bi.as(Ml.Configuration).registerConfiguration({...EA,properties:{"editor.codeActions.triggerOnFocusChange":{type:"boolean",scope:6,markdownDescription:m(862,"Enable triggering {0} when {1} is set to {2}. Code Actions must be set to {3} to be triggered for window and focus changes.","`#editor.codeActionsOnSave#`","`#files.autoSave#`","`afterDelay`","`always`"),default:!1}}});const bN=class bN{constructor(){this.lenses=[]}dispose(){this._store?.dispose()}get isDisposed(){return this._store?.isDisposed??!1}add(e,t){i_(e)&&(this._store??=new U,this._store.add(e));for(const i of e.lenses)this.lenses.push({symbol:i,provider:t})}};bN.Empty=new bN;let yv=bN;async function une(s,e,t){const i=s.ordered(e),n=new Map,o=new yv,r=i.map(async(a,l)=>{n.set(a,l);try{const c=await Promise.resolve(a.provideCodeLenses(e,t));c&&o.add(c,a)}catch(c){ji(c)}});return await Promise.all(r),t.isCancellationRequested?(o.dispose(),yv.Empty):(o.lenses=o.lenses.sort((a,l)=>a.symbol.range.startLineNumber<l.symbol.range.startLineNumber?-1:a.symbol.range.startLineNumber>l.symbol.range.startLineNumber?1:n.get(a.provider)<n.get(l.provider)?-1:n.get(a.provider)>n.get(l.provider)?1:a.symbol.range.startColumn<l.symbol.range.startColumn?-1:a.symbol.range.startColumn>l.symbol.range.startColumn?1:0),o)}at.registerCommand("_executeCodeLensProvider",function(s,...e){let[t,i]=e;dt(Le.isUri(t)),dt(typeof i=="number"||!i);const{codeLensProvider:n}=s.get(ge),o=s.get(hi).getModel(t);if(!o)throw Qr();const r=[],a=new U;return une(n,o,Ke.None).then(l=>{a.add(l);const c=[];for(const d of l.lenses)i==null||d.symbol.command?r.push(d.symbol):i-- >0&&d.provider.resolveCodeLens&&c.push(Promise.resolve(d.provider.resolveCodeLens(o,d.symbol,Ke.None)).then(h=>r.push(h||d.symbol)));return Promise.all(c)}).then(()=>r).finally(()=>{setTimeout(()=>a.dispose(),100)})});var iDe=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},nDe=function(s,e){return function(t,i){e(t,i,s)}};const gne=qe("ICodeLensCache");class wj{constructor(e,t){this.lineCount=e,this.data=t}}let X4=class{constructor(e){this._fakeProvider=new class{provideCodeLenses(){throw new Error("not supported")}},this._cache=new Va(20,.75);const t="codelens/cache";r1(Nt,()=>e.remove(t,1));const i="codelens/cache2",n=e.get(i,1,"{}");this._deserialize(n);const o=oe.filter(e.onWillSaveState,r=>r.reason===Fu.SHUTDOWN);oe.once(o)(r=>{e.store(i,this._serialize(),1,1)})}put(e,t){const i=t.lenses.map(r=>({range:r.symbol.range,command:r.symbol.command&&{id:"",title:r.symbol.command?.title}})),n=new yv;n.add({lenses:i},this._fakeProvider);const o=new wj(e.getLineCount(),n);this._cache.set(e.uri.toString(),o)}get(e){const t=this._cache.get(e.uri.toString());return t&&t.lineCount===e.getLineCount()?t.data:void 0}delete(e){this._cache.delete(e.uri.toString())}_serialize(){const e=Object.create(null);for(const[t,i]of this._cache){const n=new Set;for(const o of i.data.lenses)n.add(o.symbol.range.startLineNumber);e[t]={lineCount:i.lineCount,lines:[...n.values()]}}return JSON.stringify(e)}_deserialize(e){try{const t=JSON.parse(e);for(const i in t){const n=t[i],o=[];for(const a of n.lines)o.push({range:new k(a,1,a,11)});const r=new yv;r.add({lenses:o},this._fakeProvider),this._cache.set(i,new wj(n.lineCount,r))}}catch{}}};X4=iDe([nDe(0,Bs)],X4);tt(gne,X4,1);class sDe{constructor(e,t,i){this.afterColumn=1073741824,this.afterLineNumber=e,this.heightInPx=t,this._onHeight=i,this.suppressMouseDown=!0,this.domNode=document.createElement("div")}onComputedHeight(e){this._lastHeight===void 0?this._lastHeight=e:this._lastHeight!==e&&(this._lastHeight=e,this._onHeight())}isVisible(){return this._lastHeight!==0&&this.domNode.hasAttribute("monaco-visible-view-zone")}}const Ty=class Ty{constructor(e,t){this.allowEditorOverflow=!1,this.suppressMouseDown=!0,this._commands=new Map,this._isEmpty=!0,this._editor=e,this._id=`codelens.widget-${Ty._idPool++}`,this.updatePosition(t),this._domNode=document.createElement("span"),this._domNode.className="codelens-decoration"}withCommands(e,t){this._commands.clear();const i=[];let n=!1;for(let o=0;o<e.length;o++){const r=e[o];if(r&&(n=!0,r.command)){const a=Pu(r.command.title.trim());if(r.command.id){const l=`c${Ty._idPool++}`;i.push(ie("a",{id:l,title:r.command.tooltip,role:"button"},...a)),this._commands.set(l,r.command)}else i.push(ie("span",{title:r.command.tooltip},...a));o+1<e.length&&i.push(ie("span",void 0," | "))}}n?(vn(this._domNode,...i),this._isEmpty&&t&&this._domNode.classList.add("fadein"),this._isEmpty=!1):vn(this._domNode,ie("span",void 0,"no commands"))}getCommand(e){return e.parentElement===this._domNode?this._commands.get(e.id):void 0}getId(){return this._id}getDomNode(){return this._domNode}updatePosition(e){const t=this._editor.getModel().getLineFirstNonWhitespaceColumn(e);this._widgetPosition={position:{lineNumber:e,column:t},preference:[1]}}getPosition(){return this._widgetPosition||null}};Ty._idPool=0;let Q4=Ty;class GM{constructor(){this._removeDecorations=[],this._addDecorations=[],this._addDecorationsCallbacks=[]}addDecoration(e,t){this._addDecorations.push(e),this._addDecorationsCallbacks.push(t)}removeDecoration(e){this._removeDecorations.push(e)}commit(e){const t=e.deltaDecorations(this._removeDecorations,this._addDecorations);for(let i=0,n=t.length;i<n;i++)this._addDecorationsCallbacks[i](t[i])}}const vj=Fe.register({collapseOnReplaceEdit:!0,description:"codelens"});class Cj{constructor(e,t,i,n,o,r){this._isDisposed=!1,this._editor=t,this._data=e,this._decorationIds=[];let a;const l=[];this._data.forEach((c,d)=>{c.symbol.command&&l.push(c.symbol),i.addDecoration({range:c.symbol.range,options:vj},h=>this._decorationIds[d]=h),a?a=k.plusRange(a,c.symbol.range):a=k.lift(c.symbol.range)}),this._viewZone=new sDe(a.startLineNumber-1,o,r),this._viewZoneId=n.addZone(this._viewZone),l.length>0&&(this._createContentWidgetIfNecessary(),this._contentWidget.withCommands(l,!1))}_createContentWidgetIfNecessary(){this._contentWidget?this._editor.layoutContentWidget(this._contentWidget):(this._contentWidget=new Q4(this._editor,this._viewZone.afterLineNumber+1),this._editor.addContentWidget(this._contentWidget))}dispose(e,t){this._decorationIds.forEach(e.removeDecoration,e),this._decorationIds=[],t?.removeZone(this._viewZoneId),this._contentWidget&&(this._editor.removeContentWidget(this._contentWidget),this._contentWidget=void 0),this._isDisposed=!0}isDisposed(){return this._isDisposed}isValid(){return this._decorationIds.some((e,t)=>{const i=this._editor.getModel().getDecorationRange(e),n=this._data[t].symbol;return!!(i&&k.isEmpty(n.range)===i.isEmpty())})}updateCodeLensSymbols(e,t){this._decorationIds.forEach(t.removeDecoration,t),this._decorationIds=[],this._data=e,this._data.forEach((i,n)=>{t.addDecoration({range:i.symbol.range,options:vj},o=>this._decorationIds[n]=o)})}updateHeight(e,t){this._viewZone.heightInPx=e,t.layoutZone(this._viewZoneId),this._contentWidget&&this._editor.layoutContentWidget(this._contentWidget)}computeIfNecessary(e){if(!this._viewZone.isVisible())return null;for(let t=0;t<this._decorationIds.length;t++){const i=e.getDecorationRange(this._decorationIds[t]);i&&(this._data[t].symbol.range=i)}return this._data}updateCommands(e){this._createContentWidgetIfNecessary(),this._contentWidget.withCommands(e,!0);for(let t=0;t<this._data.length;t++){const i=e[t];if(i){const{symbol:n}=this._data[t];n.command=i.command||n.command}}}getCommand(e){return this._contentWidget?.getCommand(e)}getLineNumber(){const e=this._editor.getModel().getDecorationRange(this._decorationIds[0]);return e?e.startLineNumber:-1}update(e){if(this.isValid()){const t=this._editor.getModel().getDecorationRange(this._decorationIds[0]);t&&(this._viewZone.afterLineNumber=t.startLineNumber-1,e.layoutZone(this._viewZoneId),this._contentWidget&&(this._contentWidget.updatePosition(t.startLineNumber),this._editor.layoutContentWidget(this._contentWidget)))}}}const nH=qe("environmentService");var oDe=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},yj=function(s,e){return function(t,i){e(t,i,s)}};const ga=qe("ILanguageFeatureDebounceService");var GI;(function(s){const e=new WeakMap;let t=0;function i(n){let o=e.get(n);return o===void 0&&(o=++t,e.set(n,o)),o}s.of=i})(GI||(GI={}));class rDe{constructor(e){this._default=e}get(e){return this._default}update(e,t){return this._default}default(){return this._default}}class aDe{constructor(e,t,i,n,o,r){this._logService=e,this._name=t,this._registry=i,this._default=n,this._min=o,this._max=r,this._cache=new Va(50,.7)}_key(e){return e.id+this._registry.all(e).reduce((t,i)=>ER(GI.of(i),t),0)}get(e){const t=this._key(e),i=this._cache.get(t);return i?qs(i.value,this._min,this._max):this.default()}update(e,t){const i=this._key(e);let n=this._cache.get(i);n||(n=new rLe(6),this._cache.set(i,n));const o=qs(n.update(t),this._min,this._max);return kR(e.uri,"output")||this._logService.trace(`[DEBOUNCE: ${this._name}] for ${e.uri.toString()} is ${o}ms`),o}_overall(){const e=new sie;for(const[,t]of this._cache)e.update(t.value);return e.value}default(){const e=this._overall()|0||this._default;return qs(e,this._min,this._max)}}let J4=class{constructor(e,t){this._logService=e,this._data=new Map,this._isDev=t.isExtensionDevelopment||!t.isBuilt}for(e,t,i){const n=i?.min??50,o=i?.max??n**2,r=i?.key??void 0,a=`${GI.of(e)},${n}${r?","+r:""}`;let l=this._data.get(a);return l||(this._isDev?(this._logService.debug(`[DEBOUNCE: ${t}] is disabled in developed mode`),l=new rDe(n*1.5)):l=new aDe(this._logService,t,e,this._overallAverage()|0||n*1.5,n,o),this._data.set(a,l)),l}_overallAverage(){const e=new sie;for(const t of this._data.values())e.update(t.default());return e.value}};J4=oDe([yj(0,jt),yj(1,nH)],J4);tt(ga,J4,1);var lDe=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},zC=function(s,e){return function(t,i){e(t,i,s)}},zw;let US=(zw=class{constructor(e,t,i,n,o,r){this._editor=e,this._languageFeaturesService=t,this._commandService=n,this._notificationService=o,this._codeLensCache=r,this._disposables=new U,this._localToDispose=new U,this._lenses=[],this._oldCodeLensModels=new U,this._provideCodeLensDebounce=i.for(t.codeLensProvider,"CodeLensProvide",{min:250}),this._resolveCodeLensesDebounce=i.for(t.codeLensProvider,"CodeLensResolve",{min:250,salt:"resolve"}),this._resolveCodeLensesScheduler=new Rt(()=>this._resolveCodeLensesInViewport(),this._resolveCodeLensesDebounce.default()),this._disposables.add(this._editor.onDidChangeModel(()=>this._onModelChange())),this._disposables.add(this._editor.onDidChangeModelLanguage(()=>this._onModelChange())),this._disposables.add(this._editor.onDidChangeConfiguration(a=>{(a.hasChanged(59)||a.hasChanged(25)||a.hasChanged(24))&&this._updateLensStyle(),a.hasChanged(23)&&this._onModelChange()})),this._disposables.add(t.codeLensProvider.onDidChange(this._onModelChange,this)),this._onModelChange(),this._updateLensStyle()}dispose(){this._localDispose(),this._localToDispose.dispose(),this._disposables.dispose(),this._oldCodeLensModels.dispose(),this._currentCodeLensModel?.dispose()}_getLayoutInfo(){const e=Math.max(1.3,this._editor.getOption(75)/this._editor.getOption(61));let t=this._editor.getOption(25);return(!t||t<5)&&(t=this._editor.getOption(61)*.9|0),{fontSize:t,codeLensHeight:t*e|0}}_updateLensStyle(){const{codeLensHeight:e,fontSize:t}=this._getLayoutInfo(),i=this._editor.getOption(24),n=this._editor.getOption(59),{style:o}=this._editor.getContainerDomNode();o.setProperty("--vscode-editorCodeLens-lineHeight",`${e}px`),o.setProperty("--vscode-editorCodeLens-fontSize",`${t}px`),o.setProperty("--vscode-editorCodeLens-fontFeatureSettings",n.fontFeatureSettings),i&&(o.setProperty("--vscode-editorCodeLens-fontFamily",i),o.setProperty("--vscode-editorCodeLens-fontFamilyDefault",_o.fontFamily)),this._editor.changeViewZones(r=>{for(const a of this._lenses)a.updateHeight(e,r)})}_localDispose(){this._getCodeLensModelPromise?.cancel(),this._getCodeLensModelPromise=void 0,this._resolveCodeLensesPromise?.cancel(),this._resolveCodeLensesPromise=void 0,this._localToDispose.clear(),this._oldCodeLensModels.clear(),this._currentCodeLensModel?.dispose()}_onModelChange(){this._localDispose();const e=this._editor.getModel();if(!e||!this._editor.getOption(23)||e.isTooLargeForTokenization())return;const t=this._codeLensCache.get(e);if(t&&this._renderCodeLensSymbols(t),!this._languageFeaturesService.codeLensProvider.has(e)){t&&hh(()=>{const n=this._codeLensCache.get(e);t===n&&(this._codeLensCache.delete(e),this._onModelChange())},30*1e3,this._localToDispose);return}for(const n of this._languageFeaturesService.codeLensProvider.all(e))if(typeof n.onDidChange=="function"){const o=n.onDidChange(()=>i.schedule());this._localToDispose.add(o)}const i=new Rt(()=>{const n=Date.now();this._getCodeLensModelPromise?.cancel(),this._getCodeLensModelPromise=an(o=>une(this._languageFeaturesService.codeLensProvider,e,o)),this._getCodeLensModelPromise.then(o=>{this._currentCodeLensModel&&this._oldCodeLensModels.add(this._currentCodeLensModel),this._currentCodeLensModel=o,this._codeLensCache.put(e,o);const r=this._provideCodeLensDebounce.update(e,Date.now()-n);i.delay=r,this._renderCodeLensSymbols(o),this._resolveCodeLensesInViewportSoon()},Oe)},this._provideCodeLensDebounce.get(e));this._localToDispose.add(i),this._localToDispose.add(pe(()=>this._resolveCodeLensesScheduler.cancel())),this._localToDispose.add(this._editor.onDidChangeModelContent(()=>{this._editor.changeDecorations(n=>{this._editor.changeViewZones(o=>{const r=[];let a=-1;this._lenses.forEach(c=>{!c.isValid()||a===c.getLineNumber()?r.push(c):(c.update(o),a=c.getLineNumber())});const l=new GM;r.forEach(c=>{c.dispose(l,o),this._lenses.splice(this._lenses.indexOf(c),1)}),l.commit(n)})}),i.schedule(),this._resolveCodeLensesScheduler.cancel(),this._resolveCodeLensesPromise?.cancel(),this._resolveCodeLensesPromise=void 0})),this._localToDispose.add(this._editor.onDidFocusEditorText(()=>{i.schedule()})),this._localToDispose.add(this._editor.onDidBlurEditorText(()=>{i.cancel()})),this._localToDispose.add(this._editor.onDidScrollChange(n=>{n.scrollTopChanged&&this._lenses.length>0&&this._resolveCodeLensesInViewportSoon()})),this._localToDispose.add(this._editor.onDidLayoutChange(()=>{this._resolveCodeLensesInViewportSoon()})),this._localToDispose.add(pe(()=>{if(this._editor.getModel()){const n=Il.capture(this._editor);this._editor.changeDecorations(o=>{this._editor.changeViewZones(r=>{this._disposeAllLenses(o,r)})}),n.restore(this._editor)}else this._disposeAllLenses(void 0,void 0)})),this._localToDispose.add(this._editor.onMouseDown(n=>{if(n.target.type!==9)return;let o=n.target.element;if(o?.tagName==="SPAN"&&(o=o.parentElement),o?.tagName==="A")for(const r of this._lenses){const a=r.getCommand(o);if(a){this._commandService.executeCommand(a.id,...a.arguments||[]).catch(l=>this._notificationService.error(l));break}}})),i.schedule()}_disposeAllLenses(e,t){const i=new GM;for(const n of this._lenses)n.dispose(i,t);e&&i.commit(e),this._lenses.length=0}_renderCodeLensSymbols(e){if(!this._editor.hasModel())return;const t=this._editor.getModel().getLineCount(),i=[];let n;for(const a of e.lenses){const l=a.symbol.range.startLineNumber;l<1||l>t||(n&&n[n.length-1].symbol.range.startLineNumber===l?n.push(a):(n=[a],i.push(n)))}if(!i.length&&!this._lenses.length)return;const o=Il.capture(this._editor),r=this._getLayoutInfo();this._editor.changeDecorations(a=>{this._editor.changeViewZones(l=>{const c=new GM;let d=0,h=0;for(;h<i.length&&d<this._lenses.length;){const u=i[h][0].symbol.range.startLineNumber,g=this._lenses[d].getLineNumber();g<u?(this._lenses[d].dispose(c,l),this._lenses.splice(d,1)):g===u?(this._lenses[d].updateCodeLensSymbols(i[h],c),h++,d++):(this._lenses.splice(d,0,new Cj(i[h],this._editor,c,l,r.codeLensHeight,()=>this._resolveCodeLensesInViewportSoon())),d++,h++)}for(;d<this._lenses.length;)this._lenses[d].dispose(c,l),this._lenses.splice(d,1);for(;h<i.length;)this._lenses.push(new Cj(i[h],this._editor,c,l,r.codeLensHeight,()=>this._resolveCodeLensesInViewportSoon())),h++;c.commit(a)})}),o.restore(this._editor)}_resolveCodeLensesInViewportSoon(){this._editor.getModel()&&this._resolveCodeLensesScheduler.schedule()}_resolveCodeLensesInViewport(){this._resolveCodeLensesPromise?.cancel(),this._resolveCodeLensesPromise=void 0;const e=this._editor.getModel();if(!e)return;const t=[],i=[];if(this._lenses.forEach(r=>{const a=r.computeIfNecessary(e);a&&(t.push(a),i.push(r))}),t.length===0){this._oldCodeLensModels.clear();return}const n=Date.now(),o=an(r=>{const a=t.map((l,c)=>{const d=new Array(l.length),h=l.map((u,g)=>!u.symbol.command&&typeof u.provider.resolveCodeLens=="function"?Promise.resolve(u.provider.resolveCodeLens(e,u.symbol,r)).then(f=>{d[g]=f},ji):(d[g]=u.symbol,Promise.resolve(void 0)));return Promise.all(h).then(()=>{!r.isCancellationRequested&&!i[c].isDisposed()&&i[c].updateCommands(d)})});return Promise.all(a)});this._resolveCodeLensesPromise=o,this._resolveCodeLensesPromise.then(()=>{const r=this._resolveCodeLensesDebounce.update(e,Date.now()-n);this._resolveCodeLensesScheduler.delay=r,this._currentCodeLensModel&&this._codeLensCache.put(e,this._currentCodeLensModel),this._oldCodeLensModels.clear(),o===this._resolveCodeLensesPromise&&(this._resolveCodeLensesPromise=void 0)},r=>{Oe(r),o===this._resolveCodeLensesPromise&&(this._resolveCodeLensesPromise=void 0)})}async getModel(){return await this._getCodeLensModelPromise,await this._resolveCodeLensesPromise,this._currentCodeLensModel?.isDisposed?void 0:this._currentCodeLensModel}},zw.ID="css.editor.codeLens",zw);US=lDe([zC(1,ge),zC(2,ga),zC(3,Gt),zC(4,Ii),zC(5,gne)],US);lt(US.ID,US,1);ae(class extends ue{constructor(){super({id:"codelens.showLensesInCurrentLine",precondition:N.hasCodeLensProvider,label:V(884,"Show CodeLens Commands for Current Line")})}async run(e,t){if(!t.hasModel())return;const i=e.get(bs),n=e.get(Gt),o=e.get(Ii),r=t.getSelection().positionLineNumber,a=t.getContribution(US.ID);if(!a)return;const l=await a.getModel();if(!l)return;const c=[];for(const u of l.lenses)u.symbol.command&&u.symbol.range.startLineNumber===r&&c.push({label:u.symbol.command.title,command:u.symbol.command});if(c.length===0)return;const d=await i.pick(c,{canPickMany:!1,placeHolder:m(883,"Select a command")});if(!d)return;let h=d.command;if(l.isDisposed){const g=(await a.getModel())?.lenses.find(f=>f.symbol.range.startLineNumber===r&&f.symbol.command?.title===h.title);if(!g||!g.symbol.command)return;h=g.symbol.command}try{await n.executeCommand(h.id,...h.arguments||[])}catch(u){o.error(u)}}});const fne=[];function uC(s){fne.push(s)}function cDe(){return fne.slice(0)}class KM{constructor(e,t,i,n){this.priority=e,this.range=t,this.initialMousePosX=i,this.initialMousePosY=n,this.type=1}equals(e){return e.type===1&&this.range.equalsRange(e.range)}canAdoptVisibleHover(e,t){return e.type===1&&t.lineNumber===this.range.startLineNumber}}class g1{constructor(e,t,i,n,o,r){this.priority=e,this.owner=t,this.range=i,this.initialMousePosX=n,this.initialMousePosY=o,this.supportsMarkerHover=r,this.type=2}equals(e){return e.type===2&&this.owner===e.owner}canAdoptVisibleHover(e,t){return e.type===2&&this.owner===e.owner}}class qm{constructor(e,t){this.renderedHoverParts=e,this.disposables=t}dispose(){for(const e of this.renderedHoverParts)e.dispose();this.disposables?.dispose()}}const __=new class{constructor(){this._participants=[]}register(e){this._participants.push(e)}getAll(){return this._participants}};var pne=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},e5=function(s,e){return function(t,i){e(t,i,s)}};let qS=class{constructor(e){this._editorWorkerService=e}async provideDocumentColors(e,t){return this._editorWorkerService.computeDefaultDocumentColors(e.uri)}provideColorPresentations(e,t,i){const n=t.range,o=t.color,r=o.alpha,a=new q(new G(Math.round(255*o.red),Math.round(255*o.green),Math.round(255*o.blue),r)),l=r?q.Format.CSS.formatRGBA(a):q.Format.CSS.formatRGB(a),c=r?q.Format.CSS.formatHSLA(a):q.Format.CSS.formatHSL(a),d=r?q.Format.CSS.formatHexA(a):q.Format.CSS.formatHex(a),h=[];return h.push({label:l,textEdit:{range:n,text:l}}),h.push({label:c,textEdit:{range:n,text:c}}),h.push({label:d,textEdit:{range:n,text:d}}),h}};qS=pne([e5(0,ko)],qS);let t5=class extends F{constructor(e,t){super(),this._register(e.colorProvider.register("*",new qS(t)))}};t5=pne([e5(0,ge),e5(1,ko)],t5);async function mne(s,e,t,i="auto"){return sH(new dDe,s,e,t,i)}function _ne(s,e,t,i){return Promise.resolve(t.provideColorPresentations(s,e,i))}class dDe{constructor(){}async compute(e,t,i,n){const o=await e.provideDocumentColors(t,i);if(Array.isArray(o))for(const r of o)n.push({colorInfo:r,provider:e});return Array.isArray(o)}}class hDe{constructor(){}async compute(e,t,i,n){const o=await e.provideDocumentColors(t,i);if(Array.isArray(o))for(const r of o)n.push({range:r.range,color:[r.color.red,r.color.green,r.color.blue,r.color.alpha]});return Array.isArray(o)}}class uDe{constructor(e){this.colorInfo=e}async compute(e,t,i,n){const o=await e.provideColorPresentations(t,this.colorInfo,Ke.None);return Array.isArray(o)&&n.push(...o),Array.isArray(o)}}async function sH(s,e,t,i,n){let o=!1,r;const a=[],l=e.ordered(t);for(let c=l.length-1;c>=0;c--){const d=l[c];if(n!=="always"&&d instanceof qS)r=d;else try{await s.compute(d,t,i,a)&&(o=!0)}catch(h){ji(h)}}return o?a:r&&n!=="never"?(await s.compute(r,t,i,a),a):[]}function bne(s,e){const{colorProvider:t}=s.get(ge),i=s.get(hi).getModel(e);if(!i)throw Qr();const n=s.get(He).getValue("editor.defaultColorDecorators",{resource:e});return{model:i,colorProviderRegistry:t,defaultColorDecoratorsEnablement:n}}var gDe=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},YM=function(s,e){return function(t,i){e(t,i,s)}},i5;const wne=Object.create({});var ju;let Sv=(ju=class extends F{constructor(e,t,i,n){super(),this._editor=e,this._configurationService=t,this._languageFeaturesService=i,this._localToDispose=this._register(new U),this._decorationsIds=[],this._colorDatas=new Map,this._decoratorLimitReporter=this._register(new fDe),this._colorDecorationClassRefs=this._register(new U),this._colorDecoratorIds=this._editor.createDecorationsCollection(),this._ruleFactory=this._register(new jE(this._editor)),this._debounceInformation=n.for(i.colorProvider,"Document Colors",{min:i5.RECOMPUTE_TIME}),this._register(e.onDidChangeModel(()=>{this._isColorDecoratorsEnabled=this.isEnabled(),this.updateColors()})),this._register(e.onDidChangeModelLanguage(()=>this.updateColors())),this._register(i.colorProvider.onDidChange(()=>this.updateColors())),this._register(e.onDidChangeConfiguration(o=>{const r=this._isColorDecoratorsEnabled;this._isColorDecoratorsEnabled=this.isEnabled(),this._defaultColorDecoratorsEnablement=this._editor.getOption(167);const a=r!==this._isColorDecoratorsEnabled||o.hasChanged(27),l=o.hasChanged(167);(a||l)&&(this._isColorDecoratorsEnabled?this.updateColors():this.removeAllDecorations())})),this._timeoutTimer=null,this._computePromise=null,this._isColorDecoratorsEnabled=this.isEnabled(),this._defaultColorDecoratorsEnablement=this._editor.getOption(167),this.updateColors()}isEnabled(){const e=this._editor.getModel();if(!e)return!1;const t=e.getLanguageId(),i=this._configurationService.getValue(t);if(i&&typeof i=="object"){const n=i.colorDecorators;if(n&&n.enable!==void 0&&!n.enable)return n.enable}return this._editor.getOption(26)}static get(e){return e.getContribution(this.ID)}dispose(){this.stop(),this.removeAllDecorations(),super.dispose()}updateColors(){if(this.stop(),!this._isColorDecoratorsEnabled)return;const e=this._editor.getModel();!e||!this._languageFeaturesService.colorProvider.has(e)||(this._localToDispose.add(this._editor.onDidChangeModelContent(()=>{this._timeoutTimer||(this._timeoutTimer=new $o,this._timeoutTimer.cancelAndSet(()=>{this._timeoutTimer=null,this.beginCompute()},this._debounceInformation.get(e)))})),this.beginCompute())}async beginCompute(){this._computePromise=an(async e=>{const t=this._editor.getModel();if(!t)return[];const i=new Cn(!1),n=await mne(this._languageFeaturesService.colorProvider,t,e,this._defaultColorDecoratorsEnablement);return this._debounceInformation.update(t,i.elapsed()),n});try{const e=await this._computePromise;this.updateDecorations(e),this.updateColorDecorators(e),this._computePromise=null}catch(e){Oe(e)}}stop(){this._timeoutTimer&&(this._timeoutTimer.cancel(),this._timeoutTimer=null),this._computePromise&&(this._computePromise.cancel(),this._computePromise=null),this._localToDispose.clear()}updateDecorations(e){const t=e.map(i=>({range:{startLineNumber:i.colorInfo.range.startLineNumber,startColumn:i.colorInfo.range.startColumn,endLineNumber:i.colorInfo.range.endLineNumber,endColumn:i.colorInfo.range.endColumn},options:Fe.EMPTY}));this._editor.changeDecorations(i=>{this._decorationsIds=i.deltaDecorations(this._decorationsIds,t),this._colorDatas=new Map,this._decorationsIds.forEach((n,o)=>this._colorDatas.set(n,e[o]))})}updateColorDecorators(e){this._colorDecorationClassRefs.clear();const t=[],i=this._editor.getOption(27);for(let o=0;o<e.length&&t.length<i;o++){const{red:r,green:a,blue:l,alpha:c}=e[o].colorInfo.color,d=new G(Math.round(r*255),Math.round(a*255),Math.round(l*255),c),h=`rgba(${d.r}, ${d.g}, ${d.b}, ${d.a})`,u=this._colorDecorationClassRefs.add(this._ruleFactory.createClassNameRef({backgroundColor:h}));t.push({range:{startLineNumber:e[o].colorInfo.range.startLineNumber,startColumn:e[o].colorInfo.range.startColumn,endLineNumber:e[o].colorInfo.range.endLineNumber,endColumn:e[o].colorInfo.range.endColumn},options:{description:"colorDetector",before:{content:oQ,inlineClassName:`${u.className} colorpicker-color-decoration`,inlineClassNameAffectsLetterSpacing:!0,attachedData:wne}}})}const n=i<e.length?i:!1;this._decoratorLimitReporter.update(e.length,n),this._colorDecoratorIds.set(t)}removeAllDecorations(){this._editor.removeDecorations(this._decorationsIds),this._decorationsIds=[],this._colorDecoratorIds.clear(),this._colorDecorationClassRefs.clear()}getColorData(e){const t=this._editor.getModel();if(!t)return null;const i=t.getDecorationsInRange(k.fromPositions(e,e)).filter(n=>this._colorDatas.has(n.id));return i.length===0?null:this._colorDatas.get(i[0].id)}isColorDecoration(e){return this._colorDecoratorIds.has(e)}},i5=ju,ju.ID="editor.contrib.colorDetector",ju.RECOMPUTE_TIME=1e3,ju);Sv=i5=gDe([YM(1,He),YM(2,ge),YM(3,ga)],Sv);class fDe extends F{constructor(){super(...arguments),this._onDidChange=this._register(new M),this._computed=0,this._limited=!1}update(e,t){(e!==this._computed||t!==this._limited)&&(this._computed=e,this._limited=t,this._onDidChange.fire())}}const vne="editor.action.showHover",pDe="editor.action.showDefinitionPreviewHover",mDe="editor.action.hideHover",_De="editor.action.scrollUpHover",bDe="editor.action.scrollDownHover",wDe="editor.action.scrollLeftHover",vDe="editor.action.scrollRightHover",CDe="editor.action.pageUpHover",yDe="editor.action.pageDownHover",SDe="editor.action.goToTopHover",xDe="editor.action.goToBottomHover",IA="editor.action.increaseHoverVerbosityLevel",LDe=m(1102,"Increase Hover Verbosity Level"),DA="editor.action.decreaseHoverVerbosityLevel",kDe=m(1103,"Decrease Hover Verbosity Level"),jS="editor.action.inlineSuggest.commit",Cne="editor.action.inlineSuggest.showPrevious",yne="editor.action.inlineSuggest.showNext",EDe="editor.action.inlineSuggest.jump",Sne="editor.action.inlineSuggest.hide",n5="editor.action.inlineSuggest.toggleShowCollapsed";var oH=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},Ia=function(s,e){return function(t,i){e(t,i,s)}},zk;let s5=class extends F{constructor(e,t,i){super(),this.editor=e,this.model=t,this.instantiationService=i,this.alwaysShowToolbar=bt(this,this.editor.onDidChangeConfiguration,()=>this.editor.getOption(71).showToolbar==="always"),this.sessionPosition=void 0,this.position=j(this,n=>{const o=this.model.read(n)?.primaryGhostText.read(n);if(!this.alwaysShowToolbar.read(n)||!o||o.parts.length===0)return this.sessionPosition=void 0,null;const r=o.parts[0].column;this.sessionPosition&&this.sessionPosition.lineNumber!==o.lineNumber&&(this.sessionPosition=void 0);const a=new A(o.lineNumber,Math.min(r,this.sessionPosition?.column??Number.MAX_SAFE_INTEGER));return this.sessionPosition=a,a}),this._register(ps((n,o)=>{const r=this.model.read(n);if(!r||!this.alwaysShowToolbar.read(n))return;const a=j(c=>{const d=c.store.add(this.instantiationService.createInstance(xv.hot.read(c),this.editor,!0,this.position,r.selectedInlineCompletionIndex,r.inlineCompletionsCount,r.activeCommands,r.warning,()=>{}));return e.addContentWidget(d),c.store.add(pe(()=>e.removeContentWidget(d))),c.store.add(De(h=>{this.position.read(h)&&r.lastTriggerKind.read(h)!==ho.Explicit&&r.triggerExplicitly()})),d}),l=Ih(this,(c,d)=>!!this.position.read(c)||!!d);o.add(De(c=>{l.read(c)&&a.read(c)}))}))}};s5=oH([Ia(2,_e)],s5);const IDe=Qt("inline-suggestion-hints-next",Z.chevronRight,m(1207,"Icon for show next parameter hint.")),DDe=Qt("inline-suggestion-hints-previous",Z.chevronLeft,m(1208,"Icon for show previous parameter hint."));var mc;let xv=(mc=class extends F{static get dropDownVisible(){return this._dropDownVisible}createCommandAction(e,t,i){const n=new Cr(e,t,i,!0,()=>this._commandService.executeCommand(e)),o=this.keybindingService.lookupKeybinding(e,this._contextKeyService);let r=t;return o&&(r=m(1209,"{0} ({1})",t,o.getLabel())),n.tooltip=r,n}constructor(e,t,i,n,o,r,a,l,c,d,h,u,g){super(),this.editor=e,this.withBorder=t,this._position=i,this._currentSuggestionIdx=n,this._suggestionCount=o,this._extraCommands=r,this._warning=a,this._relayout=l,this._commandService=c,this.keybindingService=h,this._contextKeyService=u,this._menuService=g,this.id=`InlineSuggestionHintsContentWidget${zk.id++}`,this.allowEditorOverflow=!0,this.suppressMouseDown=!1,this._warningMessageContentNode=j(f=>{const p=this._warning.read(f);return p?typeof p.message=="string"?p.message:f.store.add(lx(p.message)).element:void 0}),this._warningMessageNode=Ve.div({class:"warningMessage",style:{maxWidth:400,margin:4,marginBottom:4,display:j(f=>this._warning.read(f)?"block":"none")}},[this._warningMessageContentNode]).keepUpdated(this._store),this.nodes=ct("div.inlineSuggestionsHints",{className:this.withBorder?"monaco-hover monaco-hover-content":""},[this._warningMessageNode.element,ct("div@toolBar")]),this.previousAction=this._register(this.createCommandAction(Cne,m(1210,"Previous"),xe.asClassName(DDe))),this.availableSuggestionCountAction=this._register(new Cr("inlineSuggestionHints.availableSuggestionCount","",void 0,!1)),this.nextAction=this._register(this.createCommandAction(yne,m(1211,"Next"),xe.asClassName(IDe))),this.inlineCompletionsActionsMenus=this._register(this._menuService.createMenu(fe.InlineCompletionsActions,this._contextKeyService)),this.clearAvailableSuggestionCountLabelDebounced=this._register(new Rt(()=>{this.availableSuggestionCountAction.label=""},100)),this.disableButtonsDebounced=this._register(new Rt(()=>{this.previousAction.enabled=this.nextAction.enabled=!1},100)),this._register(De(f=>{this._warningMessageContentNode.read(f),this._warningMessageNode.readEffect(f),this._relayout()})),this.toolBar=this._register(d.createInstance(o5,this.nodes.toolBar,fe.InlineSuggestionToolbar,{menuOptions:{renderShortTitle:!0},toolbarOptions:{primaryGroup:f=>f.startsWith("primary")},actionViewItemProvider:(f,p)=>{if(f instanceof yr)return d.createInstance(NDe,f,void 0);if(f===this.availableSuggestionCountAction){const _=new TDe(void 0,f,{label:!0,icon:!1});return _.setClass("availableSuggestionCount"),_}},telemetrySource:"InlineSuggestionToolbar"})),this.toolBar.setPrependedPrimaryActions([this.previousAction,this.availableSuggestionCountAction,this.nextAction]),this._register(this.toolBar.onDidChangeDropdownVisibility(f=>{zk._dropDownVisible=f})),this._register(De(f=>{this._position.read(f),this.editor.layoutContentWidget(this)})),this._register(De(f=>{const p=this._suggestionCount.read(f),_=this._currentSuggestionIdx.read(f);p!==void 0?(this.clearAvailableSuggestionCountLabelDebounced.cancel(),this.availableSuggestionCountAction.label=`${_+1}/${p}`):this.clearAvailableSuggestionCountLabelDebounced.schedule(),p!==void 0&&p>1?(this.disableButtonsDebounced.cancel(),this.previousAction.enabled=this.nextAction.enabled=!0):this.disableButtonsDebounced.schedule()})),this._register(De(f=>{const _=this._extraCommands.read(f).map(b=>({class:void 0,id:b.command.id,enabled:!0,tooltip:b.command.tooltip||"",label:b.command.title,run:w=>this._commandService.executeCommand(b.command.id)}));for(const[b,w]of this.inlineCompletionsActionsMenus.getActions())for(const C of w)C instanceof yr&&_.push(C);_.length>0&&_.unshift(new tn),this.toolBar.setAdditionalSecondaryActions(_)}))}getId(){return this.id}getDomNode(){return this.nodes.root}getPosition(){return{position:this._position.get(),preference:[1,2],positionAffinity:3}}},zk=mc,mc.hot=bA(mc),mc._dropDownVisible=!1,mc.id=0,mc);xv=zk=oH([Ia(8,Gt),Ia(9,_e),Ia(10,ft),Ia(11,Ae),Ia(12,da)],xv);class TDe extends vv{constructor(){super(...arguments),this._className=void 0}setClass(e){this._className=e}render(e){super.render(e),this._className&&e.classList.add(this._className)}updateTooltip(){}}class NDe extends bg{updateLabel(){const e=this._keybindingService.lookupKeybinding(this._action.id,this._contextKeyService,!0);if(!e)return super.updateLabel();if(this.label){const t=ct("div.keybinding").root;this._register(new hC(t,Fo,{disableTitle:!0,...gIe})).set(e),this.label.textContent=this._action.label,this.label.appendChild(t),this.label.classList.add("inlineSuggestionStatusBarItemLabel")}}updateTooltip(){}}let o5=class extends AI{constructor(e,t,i,n,o,r,a,l,c){super(e,{resetMenu:t,...i},n,o,r,a,l,c),this.menuId=t,this.options2=i,this.menuService=n,this.contextKeyService=o,this.menu=this._store.add(this.menuService.createMenu(this.menuId,this.contextKeyService,{emitEventsForSubmenuChanges:!0})),this.additionalActions=[],this.prependedPrimaryActions=[],this.additionalPrimaryActions=[],this._store.add(this.menu.onDidChange(()=>this.updateToolbar())),this.updateToolbar()}updateToolbar(){const{primary:e,secondary:t}=Lie(this.menu.getActions(this.options2?.menuOptions),this.options2?.toolbarOptions?.primaryGroup,this.options2?.toolbarOptions?.shouldInlineSubmenu,this.options2?.toolbarOptions?.useSeparatorsInPrimaryActions);t.push(...this.additionalActions),e.unshift(...this.prependedPrimaryActions),e.push(...this.additionalPrimaryActions),this.setActions(e,t)}setPrependedPrimaryActions(e){ni(this.prependedPrimaryActions,e,(t,i)=>t===i)||(this.prependedPrimaryActions=e,this.updateToolbar())}setAdditionalSecondaryActions(e){ni(this.additionalActions,e,(t,i)=>t===i)||(this.additionalActions=e,this.updateToolbar())}};o5=oH([Ia(3,da),Ia(4,Ae),Ia(5,Nr),Ia(6,ft),Ia(7,Gt),Ia(8,vs)],o5);function TA(s,e,t){const i=Li(s);return!(e<i.left||e>i.left+i.width||t<i.top||t>i.top+i.height)}class RDe{constructor(e,t,i,n){this.value=e,this.isComplete=t,this.hasLoadingMessage=i,this.options=n}}class xne extends F{constructor(e,t){super(),this._editor=e,this._computer=t,this._onResult=this._register(new M),this.onResult=this._onResult.event,this._asyncComputationScheduler=this._register(new ZM(i=>this._triggerAsyncComputation(i),0)),this._syncComputationScheduler=this._register(new ZM(i=>this._triggerSyncComputation(i),0)),this._loadingMessageScheduler=this._register(new ZM(i=>this._triggerLoadingMessage(i),0)),this._state=0,this._asyncIterable=null,this._asyncIterableDone=!1,this._result=[]}dispose(){this._asyncIterable&&(this._asyncIterable.cancel(),this._asyncIterable=null),this._options=void 0,super.dispose()}get _hoverTime(){return this._editor.getOption(69).delay}get _firstWaitTime(){return this._hoverTime/2}get _secondWaitTime(){return this._hoverTime-this._firstWaitTime}get _loadingMessageTime(){return 3*this._hoverTime}_setState(e,t){this._options=t,this._state=e,this._fireResult(t)}_triggerAsyncComputation(e){this._setState(2,e),this._syncComputationScheduler.schedule(e,this._secondWaitTime),this._computer.computeAsync?(this._asyncIterableDone=!1,this._asyncIterable=Dde(t=>this._computer.computeAsync(e,t)),(async()=>{try{for await(const t of this._asyncIterable)t&&(this._result.push(t),this._fireResult(e));this._asyncIterableDone=!0,(this._state===3||this._state===4)&&this._setState(0,e)}catch(t){Oe(t)}})()):this._asyncIterableDone=!0}_triggerSyncComputation(e){this._computer.computeSync&&(this._result=this._result.concat(this._computer.computeSync(e))),this._setState(this._asyncIterableDone?0:3,e)}_triggerLoadingMessage(e){this._state===3&&this._setState(4,e)}_fireResult(e){if(this._state===1||this._state===2)return;const t=this._state===0,i=this._state===4;this._onResult.fire(new RDe(this._result.slice(0),t,i,e))}start(e,t){if(e===0)this._state===0&&(this._setState(1,t),this._asyncComputationScheduler.schedule(t,this._firstWaitTime),this._loadingMessageScheduler.schedule(t,this._loadingMessageTime));else switch(this._state){case 0:this._triggerAsyncComputation(t),this._syncComputationScheduler.cancel(),this._triggerSyncComputation(t);break;case 2:this._syncComputationScheduler.cancel(),this._triggerSyncComputation(t);break}}cancel(){this._asyncComputationScheduler.cancel(),this._syncComputationScheduler.cancel(),this._loadingMessageScheduler.cancel(),this._asyncIterable&&(this._asyncIterable.cancel(),this._asyncIterable=null),this._result=[],this._options=void 0,this._state=0}get options(){return this._options}}class ZM extends F{constructor(e,t){super(),this._scheduler=this._register(new Rt(()=>e(this._options),t))}schedule(e,t){this._options=e,this._scheduler.schedule(t)}cancel(){this._scheduler.cancel()}}class rH{get onDidWillResize(){return this._onDidWillResize.event}get onDidResize(){return this._onDidResize.event}constructor(){this._onDidWillResize=new M,this._onDidResize=new M,this._sashListener=new U,this._size=new St(0,0),this._minSize=new St(0,0),this._maxSize=new St(Number.MAX_SAFE_INTEGER,Number.MAX_SAFE_INTEGER),this.domNode=document.createElement("div"),this._eastSash=new ls(this.domNode,{getVerticalSashLeft:()=>this._size.width},{orientation:0}),this._westSash=new ls(this.domNode,{getVerticalSashLeft:()=>0},{orientation:0}),this._northSash=new ls(this.domNode,{getHorizontalSashTop:()=>0},{orientation:1,orthogonalEdge:EI.North}),this._southSash=new ls(this.domNode,{getHorizontalSashTop:()=>this._size.height},{orientation:1,orthogonalEdge:EI.South}),this._northSash.orthogonalStartSash=this._westSash,this._northSash.orthogonalEndSash=this._eastSash,this._southSash.orthogonalStartSash=this._westSash,this._southSash.orthogonalEndSash=this._eastSash;let e,t=0,i=0;this._sashListener.add(oe.any(this._northSash.onDidStart,this._eastSash.onDidStart,this._southSash.onDidStart,this._westSash.onDidStart)(()=>{e===void 0&&(this._onDidWillResize.fire(),e=this._size,t=0,i=0)})),this._sashListener.add(oe.any(this._northSash.onDidEnd,this._eastSash.onDidEnd,this._southSash.onDidEnd,this._westSash.onDidEnd)(()=>{e!==void 0&&(e=void 0,t=0,i=0,this._onDidResize.fire({dimension:this._size,done:!0}))})),this._sashListener.add(this._eastSash.onDidChange(n=>{e&&(i=n.currentX-n.startX,this.layout(e.height+t,e.width+i),this._onDidResize.fire({dimension:this._size,done:!1,east:!0}))})),this._sashListener.add(this._westSash.onDidChange(n=>{e&&(i=-(n.currentX-n.startX),this.layout(e.height+t,e.width+i),this._onDidResize.fire({dimension:this._size,done:!1,west:!0}))})),this._sashListener.add(this._northSash.onDidChange(n=>{e&&(t=-(n.currentY-n.startY),this.layout(e.height+t,e.width+i),this._onDidResize.fire({dimension:this._size,done:!1,north:!0}))})),this._sashListener.add(this._southSash.onDidChange(n=>{e&&(t=n.currentY-n.startY,this.layout(e.height+t,e.width+i),this._onDidResize.fire({dimension:this._size,done:!1,south:!0}))})),this._sashListener.add(oe.any(this._eastSash.onDidReset,this._westSash.onDidReset)(n=>{this._preferredSize&&(this.layout(this._size.height,this._preferredSize.width),this._onDidResize.fire({dimension:this._size,done:!0}))})),this._sashListener.add(oe.any(this._northSash.onDidReset,this._southSash.onDidReset)(n=>{this._preferredSize&&(this.layout(this._preferredSize.height,this._size.width),this._onDidResize.fire({dimension:this._size,done:!0}))}))}dispose(){this._northSash.dispose(),this._southSash.dispose(),this._eastSash.dispose(),this._westSash.dispose(),this._sashListener.dispose(),this._onDidResize.dispose(),this._onDidWillResize.dispose(),this.domNode.remove()}enableSashes(e,t,i,n){this._northSash.state=e?3:0,this._eastSash.state=t?3:0,this._southSash.state=i?3:0,this._westSash.state=n?3:0}layout(e=this.size.height,t=this.size.width){const{height:i,width:n}=this._minSize,{height:o,width:r}=this._maxSize;e=Math.max(i,Math.min(o,e)),t=Math.max(n,Math.min(r,t));const a=new St(t,e);St.equals(a,this._size)||(this.domNode.style.height=e+"px",this.domNode.style.width=t+"px",this._size=a,this._northSash.layout(),this._eastSash.layout(),this._southSash.layout(),this._westSash.layout())}clearSashHoverState(){this._eastSash.clearSashHoverState(),this._westSash.clearSashHoverState(),this._northSash.clearSashHoverState(),this._southSash.clearSashHoverState()}get size(){return this._size}set maxSize(e){this._maxSize=e}get maxSize(){return this._maxSize}set minSize(e){this._minSize=e}get minSize(){return this._minSize}set preferredSize(e){this._preferredSize=e}get preferredSize(){return this._preferredSize}}const ADe=30,MDe=24;class ODe extends F{constructor(e,t=new St(10,10)){super(),this._editor=e,this.allowEditorOverflow=!0,this.suppressMouseDown=!1,this._resizableNode=this._register(new rH),this._contentPosition=null,this._isResizing=!1,this._resizableNode.domNode.style.position="absolute",this._resizableNode.minSize=St.lift(t),this._resizableNode.layout(t.height,t.width),this._resizableNode.enableSashes(!0,!0,!0,!0),this._register(this._resizableNode.onDidResize(i=>{this._resize(new St(i.dimension.width,i.dimension.height)),i.done&&(this._isResizing=!1)})),this._register(this._resizableNode.onDidWillResize(()=>{this._isResizing=!0}))}get isResizing(){return this._isResizing}getDomNode(){return this._resizableNode.domNode}getPosition(){return this._contentPosition}get position(){return this._contentPosition?.position?A.lift(this._contentPosition.position):void 0}_availableVerticalSpaceAbove(e){const t=this._editor.getDomNode(),i=this._editor.getScrolledVisiblePosition(e);return!t||!i?void 0:Li(t).top+i.top-ADe}_availableVerticalSpaceBelow(e){const t=this._editor.getDomNode(),i=this._editor.getScrolledVisiblePosition(e);if(!t||!i)return;const n=Li(t),o=lg(t.ownerDocument.body),r=n.top+i.top+i.height;return o.height-r-MDe}_findPositionPreference(e,t){const i=Math.min(this._availableVerticalSpaceBelow(t)??1/0,e),n=Math.min(this._availableVerticalSpaceAbove(t)??1/0,e),o=Math.min(Math.max(n,i),e),r=Math.min(e,o);let a;return this._editor.getOption(69).above?a=r<=n?1:2:a=r<=i?2:1,a===1?this._resizableNode.enableSashes(!0,!0,!1,!1):this._resizableNode.enableSashes(!1,!0,!0,!1),a}_resize(e){this._resizableNode.layout(e.height,e.width)}}const LL=ie;let aH=class extends F{constructor(e){super(),this.containerDomNode=document.createElement("div"),this.containerDomNode.className="monaco-hover",this.containerDomNode.classList.toggle("fade-in",!!e),this.containerDomNode.tabIndex=0,this.containerDomNode.setAttribute("role","tooltip"),this.contentsDomNode=document.createElement("div"),this.contentsDomNode.className="monaco-hover-content",this.scrollbar=this._register(new tx(this.contentsDomNode,{consumeMouseWheelIfScrollbarIsNeeded:!0})),this.containerDomNode.appendChild(this.scrollbar.getDomNode())}onContentsChanged(){this.scrollbar.scanDomNode()}};class NA extends F{static render(e,t,i){return new NA(e,t,i)}constructor(e,t,i){super(),this.actionLabel=t.label,this.actionKeybindingLabel=i,this.actionContainer=X(e,LL("div.action-container")),this.actionContainer.setAttribute("tabindex","0"),this.action=X(this.actionContainer,LL("a.action")),this.action.setAttribute("role","button"),t.iconClass&&X(this.action,LL("span.icon")).classList.add(...t.iconClass.split(" ")),this.actionRenderedLabel=i?`${t.label} (${i})`:t.label;const n=X(this.action,LL("span"));n.textContent=this.actionRenderedLabel,this._store.add(new kne(this.actionContainer,t.run)),this._store.add(new Ene(this.actionContainer,t.run,[3,10])),this.setEnabled(!0)}setEnabled(e){e?(this.actionContainer.classList.remove("disabled"),this.actionContainer.removeAttribute("aria-disabled")):(this.actionContainer.classList.add("disabled"),this.actionContainer.setAttribute("aria-disabled","true"))}}function Lne(s,e){return s&&e?m(7,"Inspect this in the accessible view with {0}.",e):s?m(8,"Inspect this in the accessible view via the command Open Accessible View which is currently not triggerable via keybinding."):""}class kne extends F{constructor(e,t){super(),this._register(W(e,se.CLICK,i=>{i.stopPropagation(),i.preventDefault(),t(e)}))}}class Ene extends F{constructor(e,t,i){super(),this._register(W(e,se.KEY_DOWN,n=>{const o=new Pt(n);i.some(r=>o.equals(r))&&(n.stopPropagation(),n.preventDefault(),t(e))}))}}var PDe=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},kL=function(s,e){return function(t,i){e(t,i,s)}},wa;const Sj=30;var Gu;let r5=(Gu=class extends ODe{get isVisibleFromKeyboard(){return this._renderedHover?.source===2}get isVisible(){return this._hoverVisibleKey.get()??!1}get isFocused(){return this._hoverFocusedKey.get()??!1}constructor(e,t,i,n,o){const r=e.getOption(75)+8,a=150,l=new St(a,r);super(e,l),this._configurationService=i,this._accessibilityService=n,this._keybindingService=o,this._hover=this._register(new aH(!0)),this._onDidResize=this._register(new M),this.onDidResize=this._onDidResize.event,this._onDidScroll=this._register(new M),this.onDidScroll=this._onDidScroll.event,this._onContentsChanged=this._register(new M),this.onContentsChanged=this._onContentsChanged.event,this._minimumSize=l,this._hoverVisibleKey=N.hoverVisible.bindTo(t),this._hoverFocusedKey=N.hoverFocused.bindTo(t),X(this._resizableNode.domNode,this._hover.containerDomNode),this._resizableNode.domNode.style.zIndex="50",this._resizableNode.domNode.className="monaco-resizable-hover",this._register(this._editor.onDidLayoutChange(()=>{this.isVisible&&this._updateMaxDimensions()})),this._register(this._editor.onDidChangeConfiguration(d=>{d.hasChanged(59)&&this._updateFont()}));const c=this._register(oa(this._resizableNode.domNode));this._register(c.onDidFocus(()=>{this._hoverFocusedKey.set(!0)})),this._register(c.onDidBlur(()=>{this._hoverFocusedKey.set(!1)})),this._register(this._hover.scrollbar.onScroll(d=>{this._onDidScroll.fire(d)})),this._setRenderedHover(void 0),this._editor.addContentWidget(this)}dispose(){super.dispose(),this._renderedHover?.dispose(),this._editor.removeContentWidget(this)}getId(){return wa.ID}static _applyDimensions(e,t,i){const n=typeof t=="number"?`${t}px`:t,o=typeof i=="number"?`${i}px`:i;e.style.width=n,e.style.height=o}_setContentsDomNodeDimensions(e,t){const i=this._hover.contentsDomNode;return wa._applyDimensions(i,e,t)}_setContainerDomNodeDimensions(e,t){const i=this._hover.containerDomNode;return wa._applyDimensions(i,e,t)}_setScrollableElementDimensions(e,t){const i=this._hover.scrollbar.getDomNode();return wa._applyDimensions(i,e,t)}_setHoverWidgetDimensions(e,t){this._setContainerDomNodeDimensions(e,t),this._setScrollableElementDimensions(e,t),this._setContentsDomNodeDimensions(e,t),this._layoutContentWidget()}static _applyMaxDimensions(e,t,i){const n=typeof t=="number"?`${t}px`:t,o=typeof i=="number"?`${i}px`:i;e.style.maxWidth=n,e.style.maxHeight=o}_setHoverWidgetMaxDimensions(e,t){wa._applyMaxDimensions(this._hover.contentsDomNode,e,t),wa._applyMaxDimensions(this._hover.scrollbar.getDomNode(),e,t),wa._applyMaxDimensions(this._hover.containerDomNode,e,t),this._hover.containerDomNode.style.setProperty("--vscode-hover-maxWidth",typeof e=="number"?`${e}px`:e),this._layoutContentWidget()}_setAdjustedHoverWidgetDimensions(e){this._setHoverWidgetMaxDimensions("none","none"),this._setHoverWidgetDimensions(e.width,e.height)}_updateResizableNodeMaxDimensions(){const e=this._findMaximumRenderingWidth()??1/0,t=this._findMaximumRenderingHeight()??1/0;this._resizableNode.maxSize=new St(e,t),this._setHoverWidgetMaxDimensions(e,t)}_resize(e){wa._lastDimensions=new St(e.width,e.height),this._setAdjustedHoverWidgetDimensions(e),this._resizableNode.layout(e.height,e.width),this._updateResizableNodeMaxDimensions(),this._hover.scrollbar.scanDomNode(),this._editor.layoutContentWidget(this),this._onDidResize.fire()}_findAvailableSpaceVertically(){const e=this._renderedHover?.showAtPosition;if(e)return this._positionPreference===1?this._availableVerticalSpaceAbove(e):this._availableVerticalSpaceBelow(e)}_findMaximumRenderingHeight(){const e=this._findAvailableSpaceVertically();if(!e)return;let i=this._hover.contentsDomNode.children.length-1;return Array.from(this._hover.contentsDomNode.children).forEach(n=>{i+=n.clientHeight}),Math.min(e,i)}_isHoverTextOverflowing(){this._hover.containerDomNode.style.setProperty("--vscode-hover-whiteSpace","nowrap"),this._hover.containerDomNode.style.setProperty("--vscode-hover-sourceWhiteSpace","nowrap");const e=Array.from(this._hover.contentsDomNode.children).some(t=>t.scrollWidth>t.clientWidth);return this._hover.containerDomNode.style.removeProperty("--vscode-hover-whiteSpace"),this._hover.containerDomNode.style.removeProperty("--vscode-hover-sourceWhiteSpace"),e}_findMaximumRenderingWidth(){if(!this._editor||!this._editor.hasModel())return;const e=this._isHoverTextOverflowing(),t=typeof this._contentWidth>"u"?0:this._contentWidth;return e||this._hover.containerDomNode.clientWidth<t?lg(this._hover.containerDomNode.ownerDocument.body).width-14:this._hover.containerDomNode.clientWidth}isMouseGettingCloser(e,t){if(!this._renderedHover)return!1;if(this._renderedHover.initialMousePosX===void 0||this._renderedHover.initialMousePosY===void 0)return this._renderedHover.initialMousePosX=e,this._renderedHover.initialMousePosY=t,!1;const i=Li(this.getDomNode());this._renderedHover.closestMouseDistance===void 0&&(this._renderedHover.closestMouseDistance=xj(this._renderedHover.initialMousePosX,this._renderedHover.initialMousePosY,i.left,i.top,i.width,i.height));const n=xj(e,t,i.left,i.top,i.width,i.height);return n>this._renderedHover.closestMouseDistance+4?!1:(this._renderedHover.closestMouseDistance=Math.min(this._renderedHover.closestMouseDistance,n),!0)}_setRenderedHover(e){this._renderedHover?.dispose(),this._renderedHover=e,this._hoverVisibleKey.set(!!e),this._hover.containerDomNode.classList.toggle("hidden",!e)}_updateFont(){const{fontSize:e,lineHeight:t}=this._editor.getOption(59),i=this._hover.contentsDomNode;i.style.fontSize=`${e}px`,i.style.lineHeight=`${t/e}`,Array.prototype.slice.call(this._hover.contentsDomNode.getElementsByClassName("code")).forEach(o=>this._editor.applyFontInfo(o))}_updateContent(e){const t=this._hover.contentsDomNode;t.style.paddingBottom="",t.textContent="",t.appendChild(e)}_layoutContentWidget(){this._editor.layoutContentWidget(this),this._hover.onContentsChanged()}_updateMaxDimensions(){const e=Math.max(this._editor.getLayoutInfo().height/4,250,wa._lastDimensions.height),t=Math.max(this._editor.getLayoutInfo().width*.66,750,wa._lastDimensions.width);this._resizableNode.maxSize=new St(t,e),this._setHoverWidgetMaxDimensions(t,e)}_render(e){this._setRenderedHover(e),this._updateFont(),this._updateContent(e.domNode),this.handleContentsChanged(),this._editor.render()}getPosition(){return this._renderedHover?{position:this._renderedHover.showAtPosition,secondaryPosition:this._renderedHover.showAtSecondaryPosition,positionAffinity:this._renderedHover.shouldAppearBeforeContent?3:void 0,preference:[this._positionPreference??1]}:null}show(e){if(!this._editor||!this._editor.hasModel())return;this._render(e);const t=Hd(this._hover.containerDomNode),i=e.showAtPosition;this._positionPreference=this._findPositionPreference(t,i)??1,this.handleContentsChanged(),e.shouldFocus&&this._hover.containerDomNode.focus(),this._onDidResize.fire();const o=this._hover.containerDomNode.ownerDocument.activeElement===this._hover.containerDomNode&&Lne(this._configurationService.getValue("accessibility.verbosity.hover")===!0&&this._accessibilityService.isScreenReaderOptimized(),this._keybindingService.lookupKeybinding("editor.action.accessibleView")?.getAriaLabel()??"");o&&(this._hover.contentsDomNode.ariaLabel=this._hover.contentsDomNode.textContent+", "+o)}hide(){if(!this._renderedHover)return;const e=this._renderedHover.shouldFocus||this._hoverFocusedKey.get();this._setRenderedHover(void 0),this._resizableNode.maxSize=new St(1/0,1/0),this._resizableNode.clearSashHoverState(),this._hoverFocusedKey.set(!1),this._editor.layoutContentWidget(this),e&&this._editor.focus()}_removeConstraintsRenderNormally(){const e=this._editor.getLayoutInfo();this._resizableNode.layout(e.height,e.width),this._setHoverWidgetDimensions("auto","auto"),this._updateMaxDimensions()}setMinimumDimensions(e){this._minimumSize=new St(Math.max(this._minimumSize.width,e.width),Math.max(this._minimumSize.height,e.height)),this._updateMinimumWidth()}_updateMinimumWidth(){const e=typeof this._contentWidth>"u"?this._minimumSize.width:Math.min(this._contentWidth,this._minimumSize.width);this._resizableNode.minSize=new St(e,this._minimumSize.height)}handleContentsChanged(){this._removeConstraintsRenderNormally();const e=this._hover.contentsDomNode;let t=Hd(e),i=Ro(e)+2;if(this._resizableNode.layout(t,i),this._setHoverWidgetDimensions(i,t),t=Hd(e),i=Ro(e),this._contentWidth=i,this._updateMinimumWidth(),this._resizableNode.layout(t,i),this._renderedHover?.showAtPosition){const n=Hd(this._hover.containerDomNode);this._positionPreference=this._findPositionPreference(n,this._renderedHover.showAtPosition)}this._layoutContentWidget(),this._onContentsChanged.fire()}focus(){this._hover.containerDomNode.focus()}scrollUp(){const e=this._hover.scrollbar.getScrollPosition().scrollTop,t=this._editor.getOption(59);this._hover.scrollbar.setScrollPosition({scrollTop:e-t.lineHeight})}scrollDown(){const e=this._hover.scrollbar.getScrollPosition().scrollTop,t=this._editor.getOption(59);this._hover.scrollbar.setScrollPosition({scrollTop:e+t.lineHeight})}scrollLeft(){const e=this._hover.scrollbar.getScrollPosition().scrollLeft;this._hover.scrollbar.setScrollPosition({scrollLeft:e-Sj})}scrollRight(){const e=this._hover.scrollbar.getScrollPosition().scrollLeft;this._hover.scrollbar.setScrollPosition({scrollLeft:e+Sj})}pageUp(){const e=this._hover.scrollbar.getScrollPosition().scrollTop,t=this._hover.scrollbar.getScrollDimensions().height;this._hover.scrollbar.setScrollPosition({scrollTop:e-t})}pageDown(){const e=this._hover.scrollbar.getScrollPosition().scrollTop,t=this._hover.scrollbar.getScrollDimensions().height;this._hover.scrollbar.setScrollPosition({scrollTop:e+t})}goToTop(){this._hover.scrollbar.setScrollPosition({scrollTop:0})}goToBottom(){this._hover.scrollbar.setScrollPosition({scrollTop:this._hover.scrollbar.getScrollDimensions().scrollHeight})}},wa=Gu,Gu.ID="editor.contrib.resizableContentHoverWidget",Gu._lastDimensions=new St(0,0),Gu);r5=wa=PDe([kL(1,Ae),kL(2,He),kL(3,Wn),kL(4,ft)],r5);function xj(s,e,t,i,n,o){const r=t+n/2,a=i+o/2,l=Math.max(Math.abs(s-r)-n/2,0),c=Math.max(Math.abs(e-a)-o/2,0);return Math.sqrt(l*l+c*c)}class KI{constructor(e,t){this._editor=e,this._participants=t}static _getLineDecorations(e,t){if(t.type!==1&&!t.supportsMarkerHover)return[];const i=e.getModel(),n=t.range.startLineNumber;if(n>i.getLineCount())return[];const o=i.getLineMaxColumn(n);return e.getLineDecorations(n).filter(r=>{if(r.options.isWholeLine)return!0;const a=r.range.startLineNumber===n?r.range.startColumn:1,l=r.range.endLineNumber===n?r.range.endColumn:o;if(r.options.showIfCollapsed){if(a>t.range.startColumn+1||t.range.endColumn-1>l)return!1}else if(a>t.range.startColumn||t.range.endColumn>l)return!1;return!0})}computeAsync(e,t){const i=e.anchor;if(!this._editor.hasModel()||!i)return Jr.EMPTY;const n=KI._getLineDecorations(this._editor,i);return Jr.merge(this._participants.map(o=>o.computeAsync?o.computeAsync(i,n,e.source,t):Jr.EMPTY))}computeSync(e){if(!this._editor.hasModel())return[];const t=e.anchor,i=KI._getLineDecorations(this._editor,t);let n=[];for(const o of this._participants)n=n.concat(o.computeSync(t,i,e.source));return Al(n)}}class Ine{constructor(e,t,i){this.hoverParts=e,this.isComplete=t,this.options=i}filter(e){const t=this.hoverParts.filter(i=>i.isValidForHoverAnchor(e));return t.length===this.hoverParts.length?this:new FDe(this,t,this.isComplete,this.options)}}class FDe extends Ine{constructor(e,t,i,n){super(t,i,n),this.original=e}filter(e){return this.original.filter(e)}}var BDe=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},Lj=function(s,e){return function(t,i){e(t,i,s)}};const kj=ie;let YI=class extends F{get hasContent(){return this._hasContent}constructor(e,t){super(),this._keybindingService=e,this._hoverService=t,this.actions=[],this._hasContent=!1,this.hoverElement=kj("div.hover-row.status-bar"),this.hoverElement.tabIndex=0,this.actionsElement=X(this.hoverElement,kj("div.actions"))}addAction(e){const t=this._keybindingService.lookupKeybinding(e.commandId),i=t?t.getLabel():null;this._hasContent=!0;const n=this._register(NA.render(this.actionsElement,e,i));return this._register(this._hoverService.setupManagedHover(Uc("element"),n.actionContainer,n.actionRenderedLabel)),this.actions.push(n),n}append(e){const t=X(this.actionsElement,e);return this._hasContent=!0,t}};YI=BDe([Lj(0,ft),Lj(1,so)],YI);const Dne={inputActiveOptionBorder:"#007ACC00",inputActiveOptionForeground:"#FFFFFF",inputActiveOptionBackground:"#0E639C50"};class Mh extends Xo{get onChange(){return this._onChange.event}get onKeyDown(){return this._onKeyDown.event}constructor(e){super(),this._onChange=this._register(new M),this._onKeyDown=this._register(new M),this._opts=e,this._title=this._opts.title,this._checked=this._opts.isChecked;const t=["monaco-custom-toggle"];this._opts.icon&&(this._icon=this._opts.icon,t.push(...xe.asClassNameArray(this._icon))),this._opts.actionClassName&&t.push(...this._opts.actionClassName.split(" ")),this._checked&&t.push("checked"),this.domNode=document.createElement("div"),this._register($a().setupDelayedHover(this.domNode,()=>({content:this._title,style:1}),this._opts.hoverLifecycleOptions)),this.domNode.classList.add(...t),this._opts.notFocusable||(this.domNode.tabIndex=0),this.domNode.setAttribute("role","checkbox"),this.domNode.setAttribute("aria-checked",String(this._checked)),this.setTitle(this._opts.title),this.applyStyles(),this.onclick(this.domNode,i=>{this.enabled&&(this.checked=!this._checked,this._onChange.fire(!1),i.preventDefault())}),this._register(this.ignoreGesture(this.domNode)),this.onkeydown(this.domNode,i=>{if(this.enabled){if(i.keyCode===10||i.keyCode===3){this.checked=!this._checked,this._onChange.fire(!0),i.preventDefault(),i.stopPropagation();return}this._onKeyDown.fire(i)}})}get enabled(){return this.domNode.getAttribute("aria-disabled")!=="true"}focus(){this.domNode.focus()}get checked(){return this._checked}set checked(e){this._checked=e,this.domNode.setAttribute("aria-checked",String(this._checked)),this.domNode.classList.toggle("checked",this._checked),this.applyStyles()}setIcon(e){this._icon&&this.domNode.classList.remove(...xe.asClassNameArray(this._icon)),this._icon=e,this._icon&&this.domNode.classList.add(...xe.asClassNameArray(this._icon))}width(){return 22}applyStyles(){this.domNode&&(this.domNode.style.borderColor=this._checked&&this._opts.inputActiveOptionBorder||"",this.domNode.style.color=this._checked&&this._opts.inputActiveOptionForeground||"inherit",this.domNode.style.backgroundColor=this._checked&&this._opts.inputActiveOptionBackground||"")}enable(){this.domNode.setAttribute("aria-disabled",String(!1)),this.domNode.classList.remove("disabled")}disable(){this.domNode.setAttribute("aria-disabled",String(!0)),this.domNode.classList.add("disabled")}setTitle(e){this._title=e,this.domNode.setAttribute("aria-label",e)}set visible(e){this.domNode.style.display=e?"":"none"}get visible(){return this.domNode.style.display!=="none"}}const VV=class VV extends Xo{constructor(e,t,i){super(),this.checkbox=e,this.domNode=t,this.styles=i,this._onChange=this._register(new M),this.onChange=this._onChange.event,this.applyStyles()}get enabled(){return this.checkbox.enabled}enable(){this.checkbox.enable(),this.applyStyles(!0)}disable(){this.checkbox.disable(),this.applyStyles(!1)}setTitle(e){this.checkbox.setTitle(e)}applyStyles(e=this.enabled){this.domNode.style.color=(e?this.styles.checkboxForeground:this.styles.checkboxDisabledForeground)||"",this.domNode.style.backgroundColor=(e?this.styles.checkboxBackground:this.styles.checkboxDisabledBackground)||"",this.domNode.style.borderColor=(e?this.styles.checkboxBorder:this.styles.checkboxDisabledBackground)||"";const t=this.styles.size||18;this.domNode.style.width=this.domNode.style.height=this.domNode.style.fontSize=`${t}px`,this.domNode.style.fontSize=`${t-2}px`}};VV.CLASS_NAME="monaco-checkbox";let GS=VV;class Tne extends GS{constructor(e,t,i){const n=new Mh({title:e,isChecked:t,icon:Z.check,actionClassName:GS.CLASS_NAME,hoverLifecycleOptions:i.hoverLifecycleOptions,...Dne});super(n,n.domNode,i),this._register(n),this._register(this.checkbox.onChange(o=>{this.applyStyles(),this._onChange.fire(o)}))}get checked(){return this.checkbox.checked}set checked(e){this.checkbox.checked=e,this.applyStyles()}applyStyles(e){this.checkbox.checked?this.checkbox.setIcon(Z.check):this.checkbox.setIcon(void 0),super.applyStyles(e)}}class Nne extends GS{constructor(e,t,i){let n;switch(t){case!0:n=Z.check;break;case"mixed":n=Z.dash;break;case!1:n=void 0;break}const o=new Mh({title:e,isChecked:t===!0,icon:n,actionClassName:Tne.CLASS_NAME,hoverLifecycleOptions:i.hoverLifecycleOptions,...Dne});super(o,o.domNode,i),this._state=t,this._register(o),this._register(this.checkbox.onChange(r=>{this._state=this.checkbox.checked,this.applyStyles(),this._onChange.fire(r)}))}get checked(){return this._state}set checked(e){this._state!==e&&(this._state=e,this.checkbox.checked=e===!0,this.applyStyles())}applyStyles(e){switch(this._state){case!0:this.checkbox.setIcon(Z.check);break;case"mixed":this.checkbox.setIcon(Z.dash);break;case!1:this.checkbox.setIcon(void 0);break}super.applyStyles(e)}}const Ss=class Ss{constructor(e){this._editor=e,this._decorations=[],this._overviewRulerApproximateDecorations=[],this._findScopeDecorationIds=[],this._rangeHighlightDecorationId=null,this._highlightedDecorationId=null,this._startPosition=this._editor.getPosition()}dispose(){this._editor.removeDecorations(this._allDecorations()),this._decorations=[],this._overviewRulerApproximateDecorations=[],this._findScopeDecorationIds=[],this._rangeHighlightDecorationId=null,this._highlightedDecorationId=null}reset(){this._decorations=[],this._overviewRulerApproximateDecorations=[],this._findScopeDecorationIds=[],this._rangeHighlightDecorationId=null,this._highlightedDecorationId=null}getCount(){return this._decorations.length}getFindScope(){return this._findScopeDecorationIds[0]?this._editor.getModel().getDecorationRange(this._findScopeDecorationIds[0]):null}getFindScopes(){if(this._findScopeDecorationIds.length){const e=this._findScopeDecorationIds.map(t=>this._editor.getModel().getDecorationRange(t)).filter(t=>!!t);if(e.length)return e}return null}getStartPosition(){return this._startPosition}setStartPosition(e){this._startPosition=e,this.setCurrentFindMatch(null)}_getDecorationIndex(e){const t=this._decorations.indexOf(e);return t>=0?t+1:1}getDecorationRangeAt(e){const t=e<this._decorations.length?this._decorations[e]:null;return t?this._editor.getModel().getDecorationRange(t):null}getCurrentMatchesPosition(e){const t=this._editor.getModel().getDecorationsInRange(e);for(const i of t){const n=i.options;if(n===Ss._FIND_MATCH_DECORATION||n===Ss._CURRENT_FIND_MATCH_DECORATION)return this._getDecorationIndex(i.id)}return 0}setCurrentFindMatch(e){let t=null,i=0;if(e)for(let n=0,o=this._decorations.length;n<o;n++){const r=this._editor.getModel().getDecorationRange(this._decorations[n]);if(e.equalsRange(r)){t=this._decorations[n],i=n+1;break}}return(this._highlightedDecorationId!==null||t!==null)&&this._editor.changeDecorations(n=>{if(this._highlightedDecorationId!==null&&(n.changeDecorationOptions(this._highlightedDecorationId,Ss._FIND_MATCH_DECORATION),this._highlightedDecorationId=null),t!==null&&(this._highlightedDecorationId=t,n.changeDecorationOptions(this._highlightedDecorationId,Ss._CURRENT_FIND_MATCH_DECORATION)),this._rangeHighlightDecorationId!==null&&(n.removeDecoration(this._rangeHighlightDecorationId),this._rangeHighlightDecorationId=null),t!==null){let o=this._editor.getModel().getDecorationRange(t);if(o.startLineNumber!==o.endLineNumber&&o.endColumn===1){const r=o.endLineNumber-1,a=this._editor.getModel().getLineMaxColumn(r);o=new k(o.startLineNumber,o.startColumn,r,a)}this._rangeHighlightDecorationId=n.addDecoration(o,Ss._RANGE_HIGHLIGHT_DECORATION)}}),i}set(e,t){this._editor.changeDecorations(i=>{let n=Ss._FIND_MATCH_DECORATION;const o=[];if(e.length>1e3){n=Ss._FIND_MATCH_NO_OVERVIEW_DECORATION;const a=this._editor.getModel().getLineCount(),c=this._editor.getLayoutInfo().height/a,d=Math.max(2,Math.ceil(3/c));let h=e[0].range.startLineNumber,u=e[0].range.endLineNumber;for(let g=1,f=e.length;g<f;g++){const p=e[g].range;u+d>=p.startLineNumber?p.endLineNumber>u&&(u=p.endLineNumber):(o.push({range:new k(h,1,u,1),options:Ss._FIND_MATCH_ONLY_OVERVIEW_DECORATION}),h=p.startLineNumber,u=p.endLineNumber)}o.push({range:new k(h,1,u,1),options:Ss._FIND_MATCH_ONLY_OVERVIEW_DECORATION})}const r=new Array(e.length);for(let a=0,l=e.length;a<l;a++)r[a]={range:e[a].range,options:n};this._decorations=i.deltaDecorations(this._decorations,r),this._overviewRulerApproximateDecorations=i.deltaDecorations(this._overviewRulerApproximateDecorations,o),this._rangeHighlightDecorationId&&(i.removeDecoration(this._rangeHighlightDecorationId),this._rangeHighlightDecorationId=null),this._findScopeDecorationIds.length&&(this._findScopeDecorationIds.forEach(a=>i.removeDecoration(a)),this._findScopeDecorationIds=[]),t?.length&&(this._findScopeDecorationIds=t.map(a=>i.addDecoration(a,Ss._FIND_SCOPE_DECORATION)))})}matchBeforePosition(e){if(this._decorations.length===0)return null;for(let t=this._decorations.length-1;t>=0;t--){const i=this._decorations[t],n=this._editor.getModel().getDecorationRange(i);if(!(!n||n.endLineNumber>e.lineNumber)){if(n.endLineNumber<e.lineNumber)return n;if(!(n.endColumn>e.column))return n}}return this._editor.getModel().getDecorationRange(this._decorations[this._decorations.length-1])}matchAfterPosition(e){if(this._decorations.length===0)return null;for(let t=0,i=this._decorations.length;t<i;t++){const n=this._decorations[t],o=this._editor.getModel().getDecorationRange(n);if(!(!o||o.startLineNumber<e.lineNumber)){if(o.startLineNumber>e.lineNumber)return o;if(!(o.startColumn<e.column))return o}}return this._editor.getModel().getDecorationRange(this._decorations[0])}_allDecorations(){let e=[];return e=e.concat(this._decorations),e=e.concat(this._overviewRulerApproximateDecorations),this._findScopeDecorationIds.length&&e.push(...this._findScopeDecorationIds),this._rangeHighlightDecorationId&&e.push(this._rangeHighlightDecorationId),e}};Ss._CURRENT_FIND_MATCH_DECORATION=Fe.register({description:"current-find-match",stickiness:1,zIndex:13,className:"currentFindMatch",inlineClassName:"currentFindMatchInline",showIfCollapsed:!0,overviewRuler:{color:Ci(eM),position:Sr.Center},minimap:{color:Ci(QP),position:1}}),Ss._FIND_MATCH_DECORATION=Fe.register({description:"find-match",stickiness:1,zIndex:10,className:"findMatch",inlineClassName:"findMatchInline",showIfCollapsed:!0,overviewRuler:{color:Ci(eM),position:Sr.Center},minimap:{color:Ci(QP),position:1}}),Ss._FIND_MATCH_NO_OVERVIEW_DECORATION=Fe.register({description:"find-match-no-overview",stickiness:1,className:"findMatch",showIfCollapsed:!0}),Ss._FIND_MATCH_ONLY_OVERVIEW_DECORATION=Fe.register({description:"find-match-only-overview",stickiness:1,overviewRuler:{color:Ci(eM),position:Sr.Center}}),Ss._RANGE_HIGHLIGHT_DECORATION=Fe.register({description:"find-range-highlight",stickiness:1,className:"rangeHighlight",isWholeLine:!0}),Ss._FIND_SCOPE_DECORATION=Fe.register({description:"find-scope",className:"findScope",isWholeLine:!0});let a5=Ss;class WDe{constructor(e,t,i){this._editorSelection=e,this._ranges=t,this._replaceStrings=i,this._trackedEditorSelectionId=null}getEditOperations(e,t){if(this._ranges.length>0){const i=[];for(let r=0;r<this._ranges.length;r++)i.push({range:this._ranges[r],text:this._replaceStrings[r]});i.sort((r,a)=>k.compareRangesUsingStarts(r.range,a.range));const n=[];let o=i[0];for(let r=1;r<i.length;r++)o.range.endLineNumber===i[r].range.startLineNumber&&o.range.endColumn===i[r].range.startColumn?(o.range=o.range.plusRange(i[r].range),o.text=o.text+i[r].text):(n.push(o),o=i[r]);n.push(o);for(const r of n)t.addEditOperation(r.range,r.text)}this._trackedEditorSelectionId=t.trackSelection(this._editorSelection)}computeCursorState(e,t){return t.getTrackedSelection(this._trackedEditorSelectionId)}}function Rne(s,e){if(s&&s[0]!==""){const t=Ej(s,e,"-"),i=Ej(s,e,"_");return t&&!i?Ij(s,e,"-"):!t&&i?Ij(s,e,"_"):s[0].toUpperCase()===s[0]?e.toUpperCase():s[0].toLowerCase()===s[0]?e.toLowerCase():$de(s[0][0])&&e.length>0?e[0].toUpperCase()+e.substr(1):s[0][0].toUpperCase()!==s[0][0]&&e.length>0?e[0].toLowerCase()+e.substr(1):e}else return e}function Ej(s,e,t){return s[0].indexOf(t)!==-1&&e.indexOf(t)!==-1&&s[0].split(t).length===e.split(t).length}function Ij(s,e,t){const i=e.split(t),n=s[0].split(t);let o="";return i.forEach((r,a)=>{o+=Rne([n[a]],r)+t}),o.slice(0,-1)}class Dj{constructor(e){this.staticValue=e,this.kind=0}}class HDe{constructor(e){this.pieces=e,this.kind=1}}class Lv{static fromStaticValue(e){return new Lv([pp.staticValue(e)])}get hasReplacementPatterns(){return this._state.kind===1}constructor(e){!e||e.length===0?this._state=new Dj(""):e.length===1&&e[0].staticValue!==null?this._state=new Dj(e[0].staticValue):this._state=new HDe(e)}buildReplaceString(e,t){if(this._state.kind===0)return t?Rne(e,this._state.staticValue):this._state.staticValue;let i="";for(let n=0,o=this._state.pieces.length;n<o;n++){const r=this._state.pieces[n];if(r.staticValue!==null){i+=r.staticValue;continue}let a=Lv._substitute(r.matchIndex,e);if(r.caseOps!==null&&r.caseOps.length>0){const l=[],c=r.caseOps.length;let d=0;for(let h=0,u=a.length;h<u;h++){if(d>=c){l.push(a.slice(h));break}switch(r.caseOps[d]){case"U":l.push(a[h].toUpperCase());break;case"u":l.push(a[h].toUpperCase()),d++;break;case"L":l.push(a[h].toLowerCase());break;case"l":l.push(a[h].toLowerCase()),d++;break;default:l.push(a[h])}}a=l.join("")}i+=a}return i}static _substitute(e,t){if(t===null)return"";if(e===0)return t[0];let i="";for(;e>0;){if(e<t.length)return(t[e]||"")+i;i=String(e%10)+i,e=Math.floor(e/10)}return"$"+i}}class pp{static staticValue(e){return new pp(e,-1,null)}static caseOps(e,t){return new pp(null,e,t)}constructor(e,t,i){this.staticValue=e,this.matchIndex=t,!i||i.length===0?this.caseOps=null:this.caseOps=i.slice(0)}}class VDe{constructor(e){this._source=e,this._lastCharIndex=0,this._result=[],this._resultLen=0,this._currentStaticPiece=""}emitUnchanged(e){this._emitStatic(this._source.substring(this._lastCharIndex,e)),this._lastCharIndex=e}emitStatic(e,t){this._emitStatic(e),this._lastCharIndex=t}_emitStatic(e){e.length!==0&&(this._currentStaticPiece+=e)}emitMatchIndex(e,t,i){this._currentStaticPiece.length!==0&&(this._result[this._resultLen++]=pp.staticValue(this._currentStaticPiece),this._currentStaticPiece=""),this._result[this._resultLen++]=pp.caseOps(e,i),this._lastCharIndex=t}finalize(){return this.emitUnchanged(this._source.length),this._currentStaticPiece.length!==0&&(this._result[this._resultLen++]=pp.staticValue(this._currentStaticPiece),this._currentStaticPiece=""),new Lv(this._result)}}function zDe(s){if(!s||s.length===0)return new Lv(null);const e=[],t=new VDe(s);for(let i=0,n=s.length;i<n;i++){const o=s.charCodeAt(i);if(o===92){if(i++,i>=n)break;const r=s.charCodeAt(i);switch(r){case 92:t.emitUnchanged(i-1),t.emitStatic("\\",i+1);break;case 110:t.emitUnchanged(i-1),t.emitStatic(`
`,i+1);break;case 116:t.emitUnchanged(i-1),t.emitStatic("	",i+1);break;case 117:case 85:case 108:case 76:t.emitUnchanged(i-1),t.emitStatic("",i+1),e.push(String.fromCharCode(r));break}continue}if(o===36){if(i++,i>=n)break;const r=s.charCodeAt(i);if(r===36){t.emitUnchanged(i-1),t.emitStatic("$",i+1);continue}if(r===48||r===38){t.emitUnchanged(i-1),t.emitMatchIndex(0,i+1,e),e.length=0;continue}if(49<=r&&r<=57){let a=r-48;if(i+1<n){const l=s.charCodeAt(i+1);if(48<=l&&l<=57){i++,a=a*10+(l-48),t.emitUnchanged(i-2),t.emitMatchIndex(a,i+1,e),e.length=0;continue}}t.emitUnchanged(i-1),t.emitMatchIndex(a,i+1,e),e.length=0;continue}}}return t.finalize()}const Oh=new ce("findWidgetVisible",!1);Oh.toNegated();const RA=new ce("findInputFocussed",!1),lH=new ce("replaceInputFocussed",!1),EL={primary:545,mac:{primary:2593}},IL={primary:565,mac:{primary:2613}},DL={primary:560,mac:{primary:2608}},TL={primary:554,mac:{primary:2602}},NL={primary:558,mac:{primary:2606}},Ut={StartFindAction:"actions.find",StartFindWithSelection:"actions.findWithSelection",StartFindWithArgs:"editor.actions.findWithArgs",NextMatchFindAction:"editor.action.nextMatchFindAction",PreviousMatchFindAction:"editor.action.previousMatchFindAction",GoToMatchFindAction:"editor.action.goToMatchFindAction",NextSelectionMatchFindAction:"editor.action.nextSelectionMatchFindAction",PreviousSelectionMatchFindAction:"editor.action.previousSelectionMatchFindAction",StartFindReplaceAction:"editor.action.startFindReplaceAction",CloseFindWidgetCommand:"closeFindWidget",ToggleCaseSensitiveCommand:"toggleFindCaseSensitive",ToggleWholeWordCommand:"toggleFindWholeWord",ToggleRegexCommand:"toggleFindRegex",ToggleSearchScopeCommand:"toggleFindInSelection",TogglePreserveCaseCommand:"togglePreserveCase",ReplaceOneAction:"editor.action.replaceOne",ReplaceAllAction:"editor.action.replaceAll",SelectAllMatchesAction:"editor.action.selectAllMatches"},ku=19999,$De=240;class X1{constructor(e,t){this._toDispose=new U,this._editor=e,this._state=t,this._isDisposed=!1,this._startSearchingTimer=new $o,this._decorations=new a5(e),this._toDispose.add(this._decorations),this._updateDecorationsScheduler=new Rt(()=>{if(this._editor.hasModel())return this.research(!1)},100),this._toDispose.add(this._updateDecorationsScheduler),this._toDispose.add(this._editor.onDidChangeCursorPosition(i=>{(i.reason===3||i.reason===5||i.reason===6)&&this._decorations.setStartPosition(this._editor.getPosition())})),this._ignoreModelContentChanged=!1,this._toDispose.add(this._editor.onDidChangeModelContent(i=>{this._ignoreModelContentChanged||(i.isFlush&&this._decorations.reset(),this._decorations.setStartPosition(this._editor.getPosition()),this._updateDecorationsScheduler.schedule())})),this._toDispose.add(this._state.onFindReplaceStateChange(i=>this._onStateChanged(i))),this.research(!1,this._state.searchScope)}dispose(){this._isDisposed=!0,xt(this._startSearchingTimer),this._toDispose.dispose()}_onStateChanged(e){this._isDisposed||this._editor.hasModel()&&(e.searchString||e.isReplaceRevealed||e.isRegex||e.wholeWord||e.matchCase||e.searchScope)&&(this._editor.getModel().isTooLargeForSyncing()?(this._startSearchingTimer.cancel(),this._startSearchingTimer.setIfNotSet(()=>{e.searchScope?this.research(e.moveCursor,this._state.searchScope):this.research(e.moveCursor)},$De)):e.searchScope?this.research(e.moveCursor,this._state.searchScope):this.research(e.moveCursor))}static _getSearchRange(e,t){return t||e.getFullModelRange()}research(e,t){let i=null;typeof t<"u"?t!==null&&(Array.isArray(t)?i=t:i=[t]):i=this._decorations.getFindScopes(),i!==null&&(i=i.map(a=>{if(a.startLineNumber!==a.endLineNumber){let l=a.endLineNumber;return a.endColumn===1&&(l=l-1),new k(a.startLineNumber,1,l,this._editor.getModel().getLineMaxColumn(l))}return a}));const n=this._findMatches(i,!1,ku);this._decorations.set(n,i);const o=this._editor.getSelection();let r=this._decorations.getCurrentMatchesPosition(o);if(r===0&&n.length>0){const a=fS(n.map(l=>l.range),l=>k.compareRangesUsingStarts(l,o)>=0);r=a>0?a-1+1:r}this._state.changeMatchInfo(r,this._decorations.getCount(),void 0),e&&this._editor.getOption(50).cursorMoveOnType&&this._moveToNextMatch(this._decorations.getStartPosition())}_hasMatches(){return this._state.matchesCount>0}_cannotFind(){if(!this._hasMatches()){const e=this._decorations.getFindScope();return e&&this._editor.revealRangeInCenterIfOutsideViewport(e,0),!0}return!1}_setCurrentFindMatch(e){const t=this._decorations.setCurrentFindMatch(e);this._state.changeMatchInfo(t,this._decorations.getCount(),e),this._editor.setSelection(e),this._editor.revealRangeInCenterIfOutsideViewport(e,0)}_prevSearchPosition(e){const t=this._state.isRegex&&(this._state.searchString.indexOf("^")>=0||this._state.searchString.indexOf("$")>=0);let{lineNumber:i,column:n}=e;const o=this._editor.getModel();return t||n===1?(i===1?i=o.getLineCount():i--,n=o.getLineMaxColumn(i)):n--,new A(i,n)}_moveToPrevMatch(e,t=!1){if(!this._state.canNavigateBack()){const d=this._decorations.matchAfterPosition(e);d&&this._setCurrentFindMatch(d);return}if(this._decorations.getCount()<ku){let d=this._decorations.matchBeforePosition(e);d&&d.isEmpty()&&d.getStartPosition().equals(e)&&(e=this._prevSearchPosition(e),d=this._decorations.matchBeforePosition(e)),d&&this._setCurrentFindMatch(d);return}if(this._cannotFind())return;const i=this._decorations.getFindScope(),n=X1._getSearchRange(this._editor.getModel(),i);n.getEndPosition().isBefore(e)&&(e=n.getEndPosition()),e.isBefore(n.getStartPosition())&&(e=n.getEndPosition());const{lineNumber:o,column:r}=e,a=this._editor.getModel();let l=new A(o,r),c=a.findPreviousMatch(this._state.searchString,l,this._state.isRegex,this._state.matchCase,this._state.wholeWord?this._editor.getOption(148):null,!1);if(c&&c.range.isEmpty()&&c.range.getStartPosition().equals(l)&&(l=this._prevSearchPosition(l),c=a.findPreviousMatch(this._state.searchString,l,this._state.isRegex,this._state.matchCase,this._state.wholeWord?this._editor.getOption(148):null,!1)),!!c){if(!t&&!n.containsRange(c.range))return this._moveToPrevMatch(c.range.getStartPosition(),!0);this._setCurrentFindMatch(c.range)}}moveToPrevMatch(){this._moveToPrevMatch(this._editor.getSelection().getStartPosition())}_nextSearchPosition(e){const t=this._state.isRegex&&(this._state.searchString.indexOf("^")>=0||this._state.searchString.indexOf("$")>=0);let{lineNumber:i,column:n}=e;const o=this._editor.getModel();return t||n===o.getLineMaxColumn(i)?(i===o.getLineCount()?i=1:i++,n=1):n++,new A(i,n)}_moveToNextMatch(e){if(!this._state.canNavigateForward()){const i=this._decorations.matchBeforePosition(e);i&&this._setCurrentFindMatch(i);return}if(this._decorations.getCount()<ku){let i=this._decorations.matchAfterPosition(e);i&&i.isEmpty()&&i.getStartPosition().equals(e)&&(e=this._nextSearchPosition(e),i=this._decorations.matchAfterPosition(e)),i&&this._setCurrentFindMatch(i);return}const t=this._getNextMatch(e,!1,!0);t&&this._setCurrentFindMatch(t.range)}_getNextMatch(e,t,i,n=!1){if(this._cannotFind())return null;const o=this._decorations.getFindScope(),r=X1._getSearchRange(this._editor.getModel(),o);r.getEndPosition().isBefore(e)&&(e=r.getStartPosition()),e.isBefore(r.getStartPosition())&&(e=r.getStartPosition());const{lineNumber:a,column:l}=e,c=this._editor.getModel();let d=new A(a,l),h=c.findNextMatch(this._state.searchString,d,this._state.isRegex,this._state.matchCase,this._state.wholeWord?this._editor.getOption(148):null,t);return i&&h&&h.range.isEmpty()&&h.range.getStartPosition().equals(d)&&(d=this._nextSearchPosition(d),h=c.findNextMatch(this._state.searchString,d,this._state.isRegex,this._state.matchCase,this._state.wholeWord?this._editor.getOption(148):null,t)),h?!n&&!r.containsRange(h.range)?this._getNextMatch(h.range.getEndPosition(),t,i,!0):h:null}moveToNextMatch(){this._moveToNextMatch(this._editor.getSelection().getEndPosition())}_moveToMatch(e){const t=this._decorations.getDecorationRangeAt(e);t&&this._setCurrentFindMatch(t)}moveToMatch(e){this._moveToMatch(e)}_getReplacePattern(){return this._state.isRegex?zDe(this._state.replaceString):Lv.fromStaticValue(this._state.replaceString)}replace(){if(!this._hasMatches())return;const e=this._getReplacePattern(),t=this._editor.getSelection(),i=this._getNextMatch(t.getStartPosition(),!0,!1);if(i)if(t.equalsRange(i.range)){const n=e.buildReplaceString(i.matches,this._state.preserveCase),o=new ts(t,n);this._executeEditorCommand("replace",o),this._decorations.setStartPosition(new A(t.startLineNumber,t.startColumn+n.length)),this.research(!0)}else this._decorations.setStartPosition(this._editor.getPosition()),this._setCurrentFindMatch(i.range)}_findMatches(e,t,i){const n=(e||[null]).map(o=>X1._getSearchRange(this._editor.getModel(),o));return this._editor.getModel().findMatches(this._state.searchString,n,this._state.isRegex,this._state.matchCase,this._state.wholeWord?this._editor.getOption(148):null,t,i)}replaceAll(){if(!this._hasMatches())return;const e=this._decorations.getFindScopes();e===null&&this._state.matchesCount>=ku?this._largeReplaceAll():this._regularReplaceAll(e),this.research(!1)}_largeReplaceAll(){const t=new nf(this._state.searchString,this._state.isRegex,this._state.matchCase,this._state.wholeWord?this._editor.getOption(148):null).parseSearchRequest();if(!t)return;let i=t.regex;if(!i.multiline){let h="mu";i.ignoreCase&&(h+="i"),i.global&&(h+="g"),i=new RegExp(i.source,h)}const n=this._editor.getModel(),o=n.getValue(1),r=n.getFullModelRange(),a=this._getReplacePattern();let l;const c=this._state.preserveCase;a.hasReplacementPatterns||c?l=o.replace(i,function(){return a.buildReplaceString(arguments,c)}):l=o.replace(i,a.buildReplaceString(null,c));const d=new pW(r,l,this._editor.getSelection());this._executeEditorCommand("replaceAll",d)}_regularReplaceAll(e){const t=this._getReplacePattern(),i=this._findMatches(e,t.hasReplacementPatterns||this._state.preserveCase,1073741824),n=[];for(let r=0,a=i.length;r<a;r++)n[r]=t.buildReplaceString(i[r].matches,this._state.preserveCase);const o=new WDe(this._editor.getSelection(),i.map(r=>r.range),n);this._executeEditorCommand("replaceAll",o)}selectAllMatches(){if(!this._hasMatches())return;const e=this._decorations.getFindScopes();let i=this._findMatches(e,!1,1073741824).map(o=>new he(o.range.startLineNumber,o.range.startColumn,o.range.endLineNumber,o.range.endColumn));const n=this._editor.getSelection();for(let o=0,r=i.length;o<r;o++)if(i[o].equalsRange(n)){i=[n].concat(i.slice(0,o)).concat(i.slice(o+1));break}this._editor.setSelections(i)}_executeEditorCommand(e,t){try{this._ignoreModelContentChanged=!0,this._editor.pushUndoStop(),this._editor.executeCommand(e,t),this._editor.pushUndoStop()}finally{this._ignoreModelContentChanged=!1}}}const UDe=m(2,"Match Case"),qDe=m(3,"Match Whole Word"),jDe=m(4,"Use Regular Expression");class Ane extends Mh{constructor(e){super({icon:Z.caseSensitive,title:UDe+e.appendTitle,isChecked:e.isChecked,hoverLifecycleOptions:e.hoverLifecycleOptions,inputActiveOptionBorder:e.inputActiveOptionBorder,inputActiveOptionForeground:e.inputActiveOptionForeground,inputActiveOptionBackground:e.inputActiveOptionBackground})}}class Mne extends Mh{constructor(e){super({icon:Z.wholeWord,title:qDe+e.appendTitle,isChecked:e.isChecked,hoverLifecycleOptions:e.hoverLifecycleOptions,inputActiveOptionBorder:e.inputActiveOptionBorder,inputActiveOptionForeground:e.inputActiveOptionForeground,inputActiveOptionBackground:e.inputActiveOptionBackground})}}class One extends Mh{constructor(e){super({icon:Z.regex,title:jDe+e.appendTitle,isChecked:e.isChecked,hoverLifecycleOptions:e.hoverLifecycleOptions,inputActiveOptionBorder:e.inputActiveOptionBorder,inputActiveOptionForeground:e.inputActiveOptionForeground,inputActiveOptionBackground:e.inputActiveOptionBackground})}}function GDe(s,e,t){const i=t??document.createElement("div");return i.textContent=s,i}function KDe(s,e,t){const i=t??document.createElement("div");return i.textContent="",Pne(i,ZDe(s),e?.actionHandler,e?.renderCodeSegments),i}class YDe{constructor(e){this.source=e,this.index=0}eos(){return this.index>=this.source.length}next(){const e=this.peek();return this.advance(),e}peek(){return this.source[this.index]}advance(){this.index++}}function Pne(s,e,t,i){let n;if(e.type===2)n=document.createTextNode(e.content||"");else if(e.type===3)n=document.createElement("b");else if(e.type===4)n=document.createElement("i");else if(e.type===7&&i)n=document.createElement("code");else if(e.type===5&&t){const o=document.createElement("a");t.disposables.add(Fi(o,"click",r=>{t.callback(String(e.index),r)})),n=o}else e.type===8?n=document.createElement("br"):e.type===1&&(n=s);n&&s!==n&&s.appendChild(n),n&&Array.isArray(e.children)&&e.children.forEach(o=>{Pne(n,o,t,i)})}function ZDe(s,e){const t={type:1,children:[]};let i=0,n=t;const o=[],r=new YDe(s);for(;!r.eos();){let a=r.next();const l=a==="\\"&&l5(r.peek())!==0;if(l&&(a=r.next()),!l&&XDe(a)&&a===r.peek()){r.advance(),n.type===2&&(n=o.pop());const c=l5(a);if(n.type===c||n.type===5&&c===6)n=o.pop();else{const d={type:c,children:[]};c===5&&(d.index=i,i++),n.children.push(d),o.push(n),n=d}}else if(a===`
`)n.type===2&&(n=o.pop()),n.children.push({type:8});else if(n.type!==2){const c={type:2,content:a};n.children.push(c),o.push(n),n=c}else n.content+=a}return n.type===2&&(n=o.pop()),t}function XDe(s,e){return l5(s)!==0}function l5(s,e){switch(s){case"*":return 3;case"_":return 4;case"[":return 5;case"]":return 6;case"`":return 0;default:return 0}}class QDe{constructor(e,t=0,i=e.length,n=t-1){this.items=e,this.start=t,this.end=i,this.index=n}current(){return this.index===this.start-1||this.index===this.end?null:this.items[this.index]}next(){return this.index=Math.min(this.index+1,this.end),this.current()}previous(){return this.index=Math.max(this.index-1,this.start-1),this.current()}first(){return this.index=this.start,this.current()}last(){return this.index=this.end-1,this.current()}}class JDe{constructor(e=new Set,t=10){this._history=e,this._limit=t,this._onChange(),this._history.onDidChange&&(this._disposable=this._history.onDidChange(()=>this._onChange()))}getHistory(){return this._elements}add(e){this._history.delete(e),this._history.add(e),this._onChange()}next(){return this._navigator.next()}previous(){return this._currentPosition()!==0?this._navigator.previous():null}current(){return this._navigator.current()}first(){return this._navigator.first()}last(){return this._navigator.last()}isLast(){return this._currentPosition()>=this._elements.length-1}isNowhere(){return this._navigator.current()===null}has(e){return this._history.has(e)}_onChange(){this._reduceToLimit();const e=this._elements;this._navigator=new QDe(e,0,e.length,e.length)}_reduceToLimit(){const e=this._elements;if(e.length>this._limit){const t=e.slice(e.length-this._limit);this._history.replace?this._history.replace(t):this._history=new Set(t)}}_currentPosition(){const e=this._navigator.current();return e?this._elements.indexOf(e):-1}get _elements(){const e=[];return this._history.forEach(t=>e.push(t)),e}dispose(){this._disposable&&(this._disposable.dispose(),this._disposable=void 0)}}const q_=ie;let eTe=class extends Xo{get onDidChange(){return this._onDidChange.event}get onDidHeightChange(){return this._onDidHeightChange.event}constructor(e,t,i){super(),this.state="idle",this.maxHeight=Number.POSITIVE_INFINITY,this.hover=this._register(new vt),this._onDidChange=this._register(new M),this._onDidHeightChange=this._register(new M),this.contextViewProvider=t,this.options=i,this.message=null,this.placeholder=this.options.placeholder||"",this.tooltip=this.options.tooltip??(this.placeholder||""),this.ariaLabel=this.options.ariaLabel||"",this.options.validationOptions&&(this.validation=this.options.validationOptions.validation),this.element=X(e,q_(".monaco-inputbox.idle"));const n=this.options.flexibleHeight?"textarea":"input",o=X(this.element,q_(".ibwrapper"));if(this.input=X(o,q_(n+".input.empty")),this.input.setAttribute("autocorrect","off"),this.input.setAttribute("autocapitalize","off"),this.input.setAttribute("spellcheck","false"),this.onfocus(this.input,()=>this.element.classList.add("synthetic-focus")),this.onblur(this.input,()=>this.element.classList.remove("synthetic-focus")),this.options.flexibleHeight){this.maxHeight=typeof this.options.flexibleMaxHeight=="number"?this.options.flexibleMaxHeight:Number.POSITIVE_INFINITY,this.mirror=X(o,q_("div.mirror")),this.mirror.innerText=" ",this.scrollableElement=new tee(this.element,{vertical:1}),this.options.flexibleWidth&&(this.input.setAttribute("wrap","off"),this.mirror.style.whiteSpace="pre",this.mirror.style.wordWrap="initial"),X(e,this.scrollableElement.getDomNode()),this._register(this.scrollableElement),this._register(this.scrollableElement.onScroll(l=>this.input.scrollTop=l.scrollTop));const r=this._register(new Lt(e.ownerDocument,"selectionchange")),a=oe.filter(r.event,()=>e.ownerDocument.getSelection()?.anchorNode===o);this._register(a(this.updateScrollDimensions,this)),this._register(this.onDidHeightChange(this.updateScrollDimensions,this))}else this.input.type=this.options.type||"text",this.input.setAttribute("wrap","off");this.ariaLabel&&this.input.setAttribute("aria-label",this.ariaLabel),this.placeholder&&!this.options.showPlaceholderOnFocus&&this.setPlaceHolder(this.placeholder),this.tooltip&&this.setTooltip(this.tooltip),this.oninput(this.input,()=>this.onValueChange()),this.onblur(this.input,()=>this.onBlur()),this.onfocus(this.input,()=>this.onFocus()),this._register(this.ignoreGesture(this.input)),setTimeout(()=>this.updateMirror(),0),this.options.actions&&(this.actionbar=this._register(new bo(this.element)),this.actionbar.push(this.options.actions,{icon:!0,label:!1})),this.applyStyles()}onBlur(){this._hideMessage(),this.options.showPlaceholderOnFocus&&this.input.setAttribute("placeholder","")}onFocus(){this._showMessage(),this.options.showPlaceholderOnFocus&&this.input.setAttribute("placeholder",this.placeholder||"")}setPlaceHolder(e){this.placeholder=e,this.input.setAttribute("placeholder",e)}setTooltip(e){this.tooltip=e,this.hover.value||(this.hover.value=this._register($a().setupDelayedHoverAtMouse(this.input,()=>({content:this.tooltip,appearance:{compact:!0}}))))}get inputElement(){return this.input}get value(){return this.input.value}set value(e){this.input.value!==e&&(this.input.value=e,this.onValueChange())}get height(){return typeof this.cachedHeight=="number"?this.cachedHeight:Hd(this.element)}focus(){this.input.focus()}blur(){this.input.blur()}hasFocus(){return NR(this.input)}select(e=null){this.input.select(),e&&(this.input.setSelectionRange(e.start,e.end),e.end===this.input.value.length&&(this.input.scrollLeft=this.input.scrollWidth))}isSelectionAtEnd(){return this.input.selectionEnd===this.input.value.length&&this.input.selectionStart===this.input.selectionEnd}getSelection(){const e=this.input.selectionStart;if(e===null)return null;const t=this.input.selectionEnd??e;return{start:e,end:t}}enable(){this.input.removeAttribute("disabled")}disable(){this.blur(),this.input.disabled=!0,this._hideMessage()}set paddingRight(e){this.input.style.width=`calc(100% - ${e}px)`,this.mirror&&(this.mirror.style.paddingRight=e+"px")}updateScrollDimensions(){if(typeof this.cachedContentHeight!="number"||typeof this.cachedHeight!="number"||!this.scrollableElement)return;const e=this.cachedContentHeight,t=this.cachedHeight,i=this.input.scrollTop;this.scrollableElement.setScrollDimensions({scrollHeight:e,height:t}),this.scrollableElement.setScrollPosition({scrollTop:i})}showMessage(e,t){if(this.state==="open"&&Ho(this.message,e))return;this.message=e,this.element.classList.remove("idle"),this.element.classList.remove("info"),this.element.classList.remove("warning"),this.element.classList.remove("error"),this.element.classList.add(this.classForType(e.type));const i=this.stylesForType(this.message.type);this.element.style.border=`1px solid ${sh(i.border,"transparent")}`,this.message.content&&(this.hasFocus()||t)&&this._showMessage()}hideMessage(){this.message=null,this.element.classList.remove("info"),this.element.classList.remove("warning"),this.element.classList.remove("error"),this.element.classList.add("idle"),this._hideMessage(),this.applyStyles()}validate(){let e=null;return this.validation&&(e=this.validation(this.value),e?(this.inputElement.setAttribute("aria-invalid","true"),this.showMessage(e)):this.inputElement.hasAttribute("aria-invalid")&&(this.inputElement.removeAttribute("aria-invalid"),this.hideMessage())),e?.type}stylesForType(e){const t=this.options.inputBoxStyles;switch(e){case 1:return{border:t.inputValidationInfoBorder,background:t.inputValidationInfoBackground,foreground:t.inputValidationInfoForeground};case 2:return{border:t.inputValidationWarningBorder,background:t.inputValidationWarningBackground,foreground:t.inputValidationWarningForeground};default:return{border:t.inputValidationErrorBorder,background:t.inputValidationErrorBackground,foreground:t.inputValidationErrorForeground}}}classForType(e){switch(e){case 1:return"info";case 2:return"warning";default:return"error"}}_showMessage(){if(!this.contextViewProvider||!this.message)return;let e;const t=()=>e.style.width=Ro(this.element)+"px";this.contextViewProvider.showContextView({getAnchor:()=>this.element,anchorAlignment:1,render:n=>{if(!this.message)return null;e=X(n,q_(".monaco-inputbox-container")),t();const o=q_("span.monaco-inputbox-message");this.message.formatContent?KDe(this.message.content,void 0,o):GDe(this.message.content,void 0,o),o.classList.add(this.classForType(this.message.type));const r=this.stylesForType(this.message.type);return o.style.backgroundColor=r.background??"",o.style.color=r.foreground??"",o.style.border=r.border?`1px solid ${r.border}`:"",X(e,o),null},onHide:()=>{this.state="closed"},layout:t});let i;this.message.type===3?i=m(9,"Error: {0}",this.message.content):this.message.type===2?i=m(10,"Warning: {0}",this.message.content):i=m(11,"Info: {0}",this.message.content),Js(i),this.state="open"}_hideMessage(){this.contextViewProvider&&(this.state==="open"&&this.contextViewProvider.hideContextView(),this.state="idle")}onValueChange(){this._onDidChange.fire(this.value),this.validate(),this.updateMirror(),this.input.classList.toggle("empty",!this.value),this.state==="open"&&this.contextViewProvider&&this.contextViewProvider.layout()}updateMirror(){if(!this.mirror)return;const e=this.value,i=e.charCodeAt(e.length-1)===10?" ":"";(e+i).replace(/\u000c/g,"")?this.mirror.textContent=e+i:this.mirror.innerText=" ",this.layout()}applyStyles(){const e=this.options.inputBoxStyles,t=e.inputBackground??"",i=e.inputForeground??"",n=e.inputBorder??"";this.element.style.backgroundColor=t,this.element.style.color=i,this.input.style.backgroundColor="inherit",this.input.style.color=i,this.element.style.border=`1px solid ${sh(n,"transparent")}`}layout(){if(!this.mirror)return;const e=this.cachedContentHeight;this.cachedContentHeight=Hd(this.mirror),e!==this.cachedContentHeight&&(this.cachedHeight=Math.min(this.cachedContentHeight,this.maxHeight),this.input.style.height=this.cachedHeight+"px",this._onDidHeightChange.fire(this.cachedContentHeight))}insertAtCursor(e){const t=this.inputElement,i=t.selectionStart,n=t.selectionEnd,o=t.value;i!==null&&n!==null&&(this.value=o.substr(0,i)+e+o.substr(n),t.setSelectionRange(i+1,i+1),this.layout())}dispose(){this._hideMessage(),this.message=null,this.actionbar?.dispose(),super.dispose()}};class Fne extends eTe{constructor(e,t,i){const n=m(12," or {0} for history","⇅"),o=m(13," ({0} for history)","⇅");super(e,t,i),this._onDidFocus=this._register(new M),this.onDidFocus=this._onDidFocus.event,this._onDidBlur=this._register(new M),this.onDidBlur=this._onDidBlur.event,this.history=this._register(new JDe(i.history,100));const r=()=>{if(i.showHistoryHint&&i.showHistoryHint()&&!this.placeholder.endsWith(n)&&!this.placeholder.endsWith(o)&&this.history.getHistory().length){const a=this.placeholder.endsWith(")")?n:o,l=this.placeholder+a;i.showPlaceholderOnFocus&&!NR(this.input)?this.placeholder=l:this.setPlaceHolder(l)}};this.observer=new MutationObserver((a,l)=>{a.forEach(c=>{c.target.textContent||r()})}),this.observer.observe(this.input,{attributeFilter:["class"]}),this.onfocus(this.input,()=>r()),this.onblur(this.input,()=>{const a=l=>{if(this.placeholder.endsWith(l)){const c=this.placeholder.slice(0,this.placeholder.length-l.length);return i.showPlaceholderOnFocus?this.placeholder=c:this.setPlaceHolder(c),!0}else return!1};a(o)||a(n)})}dispose(){super.dispose(),this.observer&&(this.observer.disconnect(),this.observer=void 0)}addToHistory(e){this.value&&(e||this.value!==this.getCurrentValue())&&this.history.add(this.value)}isAtLastInHistory(){return this.history.isLast()}isNowhereInHistory(){return this.history.isNowhere()}showNextValue(){this.history.has(this.value)||this.addToHistory();let e=this.getNextValue();e&&(e=e===this.value?this.getNextValue():e),this.value=e??"",xl(this.value?this.value:m(14,"Cleared Input"))}showPreviousValue(){this.history.has(this.value)||this.addToHistory();let e=this.getPreviousValue();e&&(e=e===this.value?this.getPreviousValue():e),e&&(this.value=e,xl(this.value))}setPlaceHolder(e){super.setPlaceHolder(e),this.setTooltip(e)}onBlur(){super.onBlur(),this._onDidBlur.fire()}onFocus(){super.onFocus(),this._onDidFocus.fire()}getCurrentValue(){let e=this.history.current();return e||(e=this.history.last(),this.history.next()),e}getPreviousValue(){return this.history.previous()||this.history.first()}getNextValue(){return this.history.next()}}const tTe=m(1,"input");class Bne extends Xo{get onDidOptionChange(){return this._onDidOptionChange.event}get onKeyDown(){return this._onKeyDown.event}get onMouseDown(){return this._onMouseDown.event}get onCaseSensitiveKeyDown(){return this._onCaseSensitiveKeyDown.event}get onRegexKeyDown(){return this._onRegexKeyDown.event}constructor(e,t,i){super(),this.fixFocusOnOptionClickEnabled=!0,this.imeSessionInProgress=!1,this.additionalTogglesDisposables=this._register(new vt),this.additionalToggles=[],this._onDidOptionChange=this._register(new M),this._onKeyDown=this._register(new M),this._onMouseDown=this._register(new M),this._onInput=this._register(new M),this._onKeyUp=this._register(new M),this._onCaseSensitiveKeyDown=this._register(new M),this._onRegexKeyDown=this._register(new M),this._lastHighlightFindOptions=0,this.placeholder=i.placeholder||"",this.validation=i.validation,this.label=i.label||tTe,this.showCommonFindToggles=!!i.showCommonFindToggles;const n=i.appendCaseSensitiveLabel||"",o=i.appendWholeWordsLabel||"",r=i.appendRegexLabel||"",a=!!i.flexibleHeight,l=!!i.flexibleWidth,c=i.flexibleMaxHeight;if(this.domNode=document.createElement("div"),this.domNode.classList.add("monaco-findInput"),this.inputBox=this._register(new Fne(this.domNode,t,{placeholder:this.placeholder||"",ariaLabel:this.label||"",validationOptions:{validation:this.validation},showHistoryHint:i.showHistoryHint,flexibleHeight:a,flexibleWidth:l,flexibleMaxHeight:c,inputBoxStyles:i.inputBoxStyles,history:i.history})),this.showCommonFindToggles){const d=i?.hoverLifecycleOptions||{groupId:"find-input"};this.regex=this._register(new One({appendTitle:r,isChecked:!1,hoverLifecycleOptions:d,...i.toggleStyles})),this._register(this.regex.onChange(u=>{this._onDidOptionChange.fire(u),!u&&this.fixFocusOnOptionClickEnabled&&this.inputBox.focus(),this.validate()})),this._register(this.regex.onKeyDown(u=>{this._onRegexKeyDown.fire(u)})),this.wholeWords=this._register(new Mne({appendTitle:o,isChecked:!1,hoverLifecycleOptions:d,...i.toggleStyles})),this._register(this.wholeWords.onChange(u=>{this._onDidOptionChange.fire(u),!u&&this.fixFocusOnOptionClickEnabled&&this.inputBox.focus(),this.validate()})),this.caseSensitive=this._register(new Ane({appendTitle:n,isChecked:!1,hoverLifecycleOptions:d,...i.toggleStyles})),this._register(this.caseSensitive.onChange(u=>{this._onDidOptionChange.fire(u),!u&&this.fixFocusOnOptionClickEnabled&&this.inputBox.focus(),this.validate()})),this._register(this.caseSensitive.onKeyDown(u=>{this._onCaseSensitiveKeyDown.fire(u)}));const h=[this.caseSensitive.domNode,this.wholeWords.domNode,this.regex.domNode];this.onkeydown(this.domNode,u=>{if(u.equals(15)||u.equals(17)||u.equals(9)){const g=h.indexOf(this.domNode.ownerDocument.activeElement);if(g>=0){let f=-1;u.equals(17)?f=(g+1)%h.length:u.equals(15)&&(g===0?f=h.length-1:f=g-1),u.equals(9)?(h[g].blur(),this.inputBox.focus()):f>=0&&h[f].focus(),gt.stop(u,!0)}}})}this.controls=document.createElement("div"),this.controls.className="controls",this.controls.style.display=this.showCommonFindToggles?"":"none",this.caseSensitive&&this.controls.append(this.caseSensitive.domNode),this.wholeWords&&this.controls.appendChild(this.wholeWords.domNode),this.regex&&this.controls.appendChild(this.regex.domNode),this.setAdditionalToggles(i?.additionalToggles),this.controls&&this.domNode.appendChild(this.controls),e?.appendChild(this.domNode),this._register(W(this.inputBox.inputElement,"compositionstart",d=>{this.imeSessionInProgress=!0})),this._register(W(this.inputBox.inputElement,"compositionend",d=>{this.imeSessionInProgress=!1,this._onInput.fire()})),this.onkeydown(this.inputBox.inputElement,d=>this._onKeyDown.fire(d)),this.onkeyup(this.inputBox.inputElement,d=>this._onKeyUp.fire(d)),this.oninput(this.inputBox.inputElement,d=>this._onInput.fire()),this.onmousedown(this.inputBox.inputElement,d=>this._onMouseDown.fire(d))}get onDidChange(){return this.inputBox.onDidChange}layout(e){this.inputBox.layout(),this.updateInputBoxPadding(e.collapsedFindWidget)}enable(){this.domNode.classList.remove("disabled"),this.inputBox.enable(),this.regex?.enable(),this.wholeWords?.enable(),this.caseSensitive?.enable();for(const e of this.additionalToggles)e.enable()}disable(){this.domNode.classList.add("disabled"),this.inputBox.disable(),this.regex?.disable(),this.wholeWords?.disable(),this.caseSensitive?.disable();for(const e of this.additionalToggles)e.disable()}setFocusInputOnOptionClick(e){this.fixFocusOnOptionClickEnabled=e}setEnabled(e){e?this.enable():this.disable()}setAdditionalToggles(e){for(const t of this.additionalToggles)t.domNode.remove();this.additionalToggles=[],this.additionalTogglesDisposables.value=new U;for(const t of e??[])this.additionalTogglesDisposables.value.add(t),this.controls.appendChild(t.domNode),this.additionalTogglesDisposables.value.add(t.onChange(i=>{this._onDidOptionChange.fire(i),!i&&this.fixFocusOnOptionClickEnabled&&this.inputBox.focus()})),this.additionalToggles.push(t);this.additionalToggles.length>0&&(this.controls.style.display=""),this.updateInputBoxPadding()}updateInputBoxPadding(e=!1){e?this.inputBox.paddingRight=0:this.inputBox.paddingRight=(this.caseSensitive?.width()??0)+(this.wholeWords?.width()??0)+(this.regex?.width()??0)+this.additionalToggles.reduce((t,i)=>t+i.width(),0)}getValue(){return this.inputBox.value}setValue(e){this.inputBox.value!==e&&(this.inputBox.value=e)}select(){this.inputBox.select()}focus(){this.inputBox.focus()}getCaseSensitive(){return this.caseSensitive?.checked??!1}setCaseSensitive(e){this.caseSensitive&&(this.caseSensitive.checked=e)}getWholeWords(){return this.wholeWords?.checked??!1}setWholeWords(e){this.wholeWords&&(this.wholeWords.checked=e)}getRegex(){return this.regex?.checked??!1}setRegex(e){this.regex&&(this.regex.checked=e,this.validate())}focusOnCaseSensitive(){this.caseSensitive?.focus()}highlightFindOptions(){this.domNode.classList.remove("highlight-"+this._lastHighlightFindOptions),this._lastHighlightFindOptions=1-this._lastHighlightFindOptions,this.domNode.classList.add("highlight-"+this._lastHighlightFindOptions)}validate(){this.inputBox.validate()}showMessage(e){this.inputBox.showMessage(e)}clearMessage(){this.inputBox.hideMessage()}}const iTe=m(5,"input"),nTe=m(6,"Preserve Case");class sTe extends Mh{constructor(e){super({icon:Z.preserveCase,title:nTe+e.appendTitle,isChecked:e.isChecked,hoverLifecycleOptions:e.hoverLifecycleOptions,inputActiveOptionBorder:e.inputActiveOptionBorder,inputActiveOptionForeground:e.inputActiveOptionForeground,inputActiveOptionBackground:e.inputActiveOptionBackground})}}class oTe extends Xo{get onDidOptionChange(){return this._onDidOptionChange.event}get onKeyDown(){return this._onKeyDown.event}get onPreserveCaseKeyDown(){return this._onPreserveCaseKeyDown.event}constructor(e,t,i,n){super(),this._showOptionButtons=i,this.fixFocusOnOptionClickEnabled=!0,this.cachedOptionsWidth=0,this._onDidOptionChange=this._register(new M),this._onKeyDown=this._register(new M),this._onMouseDown=this._register(new M),this._onInput=this._register(new M),this._onKeyUp=this._register(new M),this._onPreserveCaseKeyDown=this._register(new M),this.contextViewProvider=t,this.placeholder=n.placeholder||"",this.validation=n.validation,this.label=n.label||iTe;const o=n.appendPreserveCaseLabel||"",r=n.history||new Set([]),a=!!n.flexibleHeight,l=!!n.flexibleWidth,c=n.flexibleMaxHeight;this.domNode=document.createElement("div"),this.domNode.classList.add("monaco-findInput"),this.inputBox=this._register(new Fne(this.domNode,this.contextViewProvider,{ariaLabel:this.label||"",placeholder:this.placeholder||"",validationOptions:{validation:this.validation},history:r,showHistoryHint:n.showHistoryHint,flexibleHeight:a,flexibleWidth:l,flexibleMaxHeight:c,inputBoxStyles:n.inputBoxStyles})),this.preserveCase=this._register(new sTe({appendTitle:o,isChecked:!1,hoverLifecycleOptions:n.hoverLifecycleOptions,...n.toggleStyles})),this._register(this.preserveCase.onChange(u=>{this._onDidOptionChange.fire(u),!u&&this.fixFocusOnOptionClickEnabled&&this.inputBox.focus(),this.validate()})),this._register(this.preserveCase.onKeyDown(u=>{this._onPreserveCaseKeyDown.fire(u)})),this._showOptionButtons?this.cachedOptionsWidth=this.preserveCase.width():this.cachedOptionsWidth=0;const d=[this.preserveCase.domNode];this.onkeydown(this.domNode,u=>{if(u.equals(15)||u.equals(17)||u.equals(9)){const g=d.indexOf(this.domNode.ownerDocument.activeElement);if(g>=0){let f=-1;u.equals(17)?f=(g+1)%d.length:u.equals(15)&&(g===0?f=d.length-1:f=g-1),u.equals(9)?(d[g].blur(),this.inputBox.focus()):f>=0&&d[f].focus(),gt.stop(u,!0)}}});const h=document.createElement("div");h.className="controls",h.style.display=this._showOptionButtons?"block":"none",h.appendChild(this.preserveCase.domNode),this.domNode.appendChild(h),e?.appendChild(this.domNode),this.onkeydown(this.inputBox.inputElement,u=>this._onKeyDown.fire(u)),this.onkeyup(this.inputBox.inputElement,u=>this._onKeyUp.fire(u)),this.oninput(this.inputBox.inputElement,u=>this._onInput.fire()),this.onmousedown(this.inputBox.inputElement,u=>this._onMouseDown.fire(u))}enable(){this.domNode.classList.remove("disabled"),this.inputBox.enable(),this.preserveCase.enable()}disable(){this.domNode.classList.add("disabled"),this.inputBox.disable(),this.preserveCase.disable()}setEnabled(e){e?this.enable():this.disable()}select(){this.inputBox.select()}focus(){this.inputBox.focus()}getPreserveCase(){return this.preserveCase.checked}setPreserveCase(e){this.preserveCase.checked=e}focusOnPreserve(){this.preserveCase.focus()}validate(){this.inputBox?.validate()}set width(e){this.inputBox.paddingRight=this.cachedOptionsWidth,this.domNode.style.width=e+"px"}dispose(){super.dispose()}}var Wne=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},Hne=function(s,e){return function(t,i){e(t,i,s)}};const cH=new ce("suggestWidgetVisible",!1,m(1698,"Whether suggestion are visible")),dH="historyNavigationWidgetFocus",Vne="historyNavigationForwardsEnabled",zne="historyNavigationBackwardsEnabled";let Q1;const RL=[];function $ne(s,e){if(RL.includes(e))throw new Error("Cannot register the same widget multiple times");RL.push(e);const t=new U,i=new ce(dH,!1).bindTo(s),n=new ce(Vne,!0).bindTo(s),o=new ce(zne,!0).bindTo(s),r=()=>{i.set(!0),Q1=e},a=()=>{i.set(!1),Q1===e&&(Q1=void 0)};return NR(e.element)&&r(),t.add(e.onDidFocus(()=>r())),t.add(e.onDidBlur(()=>a())),t.add(pe(()=>{RL.splice(RL.indexOf(e),1),a()})),{historyNavigationForwardsEnablement:n,historyNavigationBackwardsEnablement:o,dispose(){t.dispose()}}}let c5=class extends Bne{constructor(e,t,i,n){super(e,t,i);const o=this._register(n.createScoped(this.inputBox.element));this._register($ne(o,this.inputBox))}};c5=Wne([Hne(3,Ae)],c5);let d5=class extends oTe{constructor(e,t,i,n,o=!1){super(e,t,o,i);const r=this._register(n.createScoped(this.inputBox.element));this._register($ne(r,this.inputBox))}};d5=Wne([Hne(3,Ae)],d5);Dn.registerCommandAndKeybindingRule({id:"history.showPrevious",weight:200,when:K.and(K.has(dH),K.equals(zne,!0),K.not("isComposing"),cH.isEqualTo(!1)),primary:16,secondary:[528],handler:s=>{Q1?.showPreviousValue()}});Dn.registerCommandAndKeybindingRule({id:"history.showNext",weight:200,when:K.and(K.has(dH),K.equals(Vne,!0),K.not("isComposing"),cH.isEqualTo(!1)),primary:18,secondary:[530],handler:s=>{Q1?.showNextValue()}});function Tj(s){return s.lookupKeybinding("history.showPrevious")?.getElectronAccelerator()==="Up"&&s.lookupKeybinding("history.showNext")?.getElectronAccelerator()==="Down"}const Nj=Qt("find-collapsed",Z.chevronRight,m(956,"Icon to indicate that the editor find widget is collapsed.")),Rj=Qt("find-expanded",Z.chevronDown,m(957,"Icon to indicate that the editor find widget is expanded.")),rTe=Qt("find-selection",Z.selection,m(958,"Icon for 'Find in Selection' in the editor find widget.")),aTe=Qt("find-replace",Z.replace,m(959,"Icon for 'Replace' in the editor find widget.")),lTe=Qt("find-replace-all",Z.replaceAll,m(960,"Icon for 'Replace All' in the editor find widget.")),cTe=Qt("find-previous-match",Z.arrowUp,m(961,"Icon for 'Find Previous' in the editor find widget.")),dTe=Qt("find-next-match",Z.arrowDown,m(962,"Icon for 'Find Next' in the editor find widget.")),hTe=m(963,"Find / Replace"),uTe=m(964,"Find"),gTe=m(965,"Find"),fTe=m(966,"Previous Match"),pTe=m(967,"Next Match"),mTe=m(968,"Find in Selection"),_Te=m(969,"Close"),bTe=m(970,"Replace"),wTe=m(971,"Replace"),vTe=m(972,"Replace"),CTe=m(973,"Replace All"),yTe=m(974,"Toggle Replace"),STe=m(975,"Only the first {0} results are highlighted, but all find operations work on the entire text.",ku),xTe=m(976,"{0} of {1}"),Aj=m(977,"No results"),Vl=419,LTe=275,kTe=LTe-54;let $C=69;const ETe=33,Mj=Ye?256:2048;class XM{constructor(e){this.afterLineNumber=e,this.heightInPx=ETe,this.suppressMouseDown=!1,this.domNode=document.createElement("div"),this.domNode.className="dock-find-viewzone"}}function Oj(s,e,t){const i=!!e.match(/\n/);if(t&&i&&t.selectionStart>0){s.stopPropagation();return}}function Pj(s,e,t){const i=!!e.match(/\n/);if(t&&i&&t.selectionEnd<t.value.length){s.stopPropagation();return}}const wN=class wN extends Xo{constructor(e,t,i,n,o,r,a,l,c){super(),this._hoverService=a,this._findWidgetSearchHistory=l,this._replaceWidgetHistory=c,this._cachedHeight=null,this._revealTimeouts=[],this._codeEditor=e,this._controller=t,this._state=i,this._contextViewProvider=n,this._keybindingService=o,this._contextKeyService=r,this._isVisible=!1,this._isReplaceVisible=!1,this._ignoreChangeEvent=!1,this._updateHistoryDelayer=new sa(500),this._register(pe(()=>this._updateHistoryDelayer.cancel())),this._register(this._state.onFindReplaceStateChange(d=>this._onStateChanged(d))),this._buildDomNode(),this._updateButtons(),this._tryUpdateWidgetWidth(),this._findInput.inputBox.layout(),this._register(this._codeEditor.onDidChangeConfiguration(d=>{if(d.hasChanged(104)&&(this._codeEditor.getOption(104)&&this._state.change({isReplaceRevealed:!1},!1),this._updateButtons()),d.hasChanged(165)&&this._tryUpdateWidgetWidth(),d.hasChanged(2)&&this.updateAccessibilitySupport(),d.hasChanged(50)){const h=this._codeEditor.getOption(50).loop;this._state.change({loop:h},!1);const u=this._codeEditor.getOption(50).addExtraSpaceOnTop;u&&!this._viewZone&&(this._viewZone=new XM(0),this._showViewZone()),!u&&this._viewZone&&this._removeViewZone()}})),this.updateAccessibilitySupport(),this._register(this._codeEditor.onDidChangeCursorSelection(()=>{this._isVisible&&this._updateToggleSelectionFindButton()})),this._register(this._codeEditor.onDidFocusEditorWidget(async()=>{if(this._isVisible){const d=await this._controller.getGlobalBufferTerm();d&&d!==this._state.searchString&&(this._state.change({searchString:d},!1),this._findInput.select())}})),this._findInputFocused=RA.bindTo(r),this._findFocusTracker=this._register(oa(this._findInput.inputBox.inputElement)),this._register(this._findFocusTracker.onDidFocus(()=>{this._findInputFocused.set(!0),this._updateSearchScope()})),this._register(this._findFocusTracker.onDidBlur(()=>{this._findInputFocused.set(!1)})),this._replaceInputFocused=lH.bindTo(r),this._replaceFocusTracker=this._register(oa(this._replaceInput.inputBox.inputElement)),this._register(this._replaceFocusTracker.onDidFocus(()=>{this._replaceInputFocused.set(!0),this._updateSearchScope()})),this._register(this._replaceFocusTracker.onDidBlur(()=>{this._replaceInputFocused.set(!1)})),this._codeEditor.addOverlayWidget(this),this._codeEditor.getOption(50).addExtraSpaceOnTop&&(this._viewZone=new XM(0)),this._register(this._codeEditor.onDidChangeModel(()=>{this._isVisible&&(this._viewZoneId=void 0)})),this._register(this._codeEditor.onDidScrollChange(d=>{if(d.scrollTopChanged){this._layoutViewZone();return}setTimeout(()=>{this._layoutViewZone()},0)}))}getId(){return wN.ID}getDomNode(){return this._domNode}getPosition(){return this._isVisible?{preference:0}:null}_onStateChanged(e){if(e.searchString){try{this._ignoreChangeEvent=!0,this._findInput.setValue(this._state.searchString)}finally{this._ignoreChangeEvent=!1}this._updateButtons()}if(e.replaceString&&(this._replaceInput.inputBox.value=this._state.replaceString),e.isRevealed&&(this._state.isRevealed?this._reveal():this._hide(!0)),e.isReplaceRevealed&&(this._state.isReplaceRevealed?!this._codeEditor.getOption(104)&&!this._isReplaceVisible&&(this._isReplaceVisible=!0,this._replaceInput.width=Ro(this._findInput.domNode),this._updateButtons(),this._replaceInput.inputBox.layout()):this._isReplaceVisible&&(this._isReplaceVisible=!1,this._updateButtons())),(e.isRevealed||e.isReplaceRevealed)&&(this._state.isRevealed||this._state.isReplaceRevealed)&&this._tryUpdateHeight()&&this._showViewZone(),e.isRegex&&this._findInput.setRegex(this._state.isRegex),e.wholeWord&&this._findInput.setWholeWords(this._state.wholeWord),e.matchCase&&this._findInput.setCaseSensitive(this._state.matchCase),e.preserveCase&&this._replaceInput.setPreserveCase(this._state.preserveCase),e.searchScope&&(this._state.searchScope?this._toggleSelectionFind.checked=!0:this._toggleSelectionFind.checked=!1,this._updateToggleSelectionFindButton()),e.searchString||e.matchesCount||e.matchesPosition){const t=this._state.searchString.length>0&&this._state.matchesCount===0;this._domNode.classList.toggle("no-results",t),this._updateMatchesCount(),this._updateButtons()}(e.searchString||e.currentMatch)&&this._layoutViewZone(),e.updateHistory&&this._delayedUpdateHistory(),e.loop&&this._updateButtons()}_delayedUpdateHistory(){this._updateHistoryDelayer.trigger(this._updateHistory.bind(this)).then(void 0,Oe)}_updateHistory(){this._state.searchString&&this._findInput.inputBox.addToHistory(),this._state.replaceString&&this._replaceInput.inputBox.addToHistory()}_updateMatchesCount(){this._matchesCount.style.minWidth=$C+"px",this._state.matchesCount>=ku?this._matchesCount.title=STe:this._matchesCount.title="",this._matchesCount.firstChild?.remove();let e;if(this._state.matchesCount>0){let t=String(this._state.matchesCount);this._state.matchesCount>=ku&&(t+="+");let i=String(this._state.matchesPosition);i==="0"&&(i="?"),e=xm(xTe,i,t)}else e=Aj;this._matchesCount.appendChild(document.createTextNode(e)),Js(this._getAriaLabel(e,this._state.currentMatch,this._state.searchString)),$C=Math.max($C,this._matchesCount.clientWidth)}_getAriaLabel(e,t,i){if(e===Aj)return i===""?m(978,"{0} found",e):m(979,"{0} found for '{1}'",e,i);if(t){const n=m(980,"{0} found for '{1}', at {2}",e,i,t.startLineNumber+":"+t.startColumn),o=this._codeEditor.getModel();return o&&t.startLineNumber<=o.getLineCount()&&t.startLineNumber>=1?`${o.getLineContent(t.startLineNumber)}, ${n}`:n}return m(981,"{0} found for '{1}'",e,i)}_updateToggleSelectionFindButton(){const e=this._codeEditor.getSelection(),t=e?e.startLineNumber!==e.endLineNumber||e.startColumn!==e.endColumn:!1,i=this._toggleSelectionFind.checked;this._isVisible&&(i||t)?this._toggleSelectionFind.enable():this._toggleSelectionFind.disable()}_updateButtons(){this._findInput.setEnabled(this._isVisible),this._replaceInput.setEnabled(this._isVisible&&this._isReplaceVisible),this._updateToggleSelectionFindButton(),this._closeBtn.setEnabled(this._isVisible);const e=this._state.searchString.length>0,t=!!this._state.matchesCount;this._prevBtn.setEnabled(this._isVisible&&e&&t&&this._state.canNavigateBack()),this._nextBtn.setEnabled(this._isVisible&&e&&t&&this._state.canNavigateForward()),this._replaceBtn.setEnabled(this._isVisible&&this._isReplaceVisible&&e),this._replaceAllBtn.setEnabled(this._isVisible&&this._isReplaceVisible&&e),this._domNode.classList.toggle("replaceToggled",this._isReplaceVisible),this._toggleReplaceBtn.setExpanded(this._isReplaceVisible);const i=!this._codeEditor.getOption(104);this._toggleReplaceBtn.setEnabled(this._isVisible&&i)}_reveal(){if(this._revealTimeouts.forEach(e=>{clearTimeout(e)}),this._revealTimeouts=[],!this._isVisible){this._isVisible=!0;const e=this._codeEditor.getSelection();switch(this._codeEditor.getOption(50).autoFindInSelection){case"always":this._toggleSelectionFind.checked=!0;break;case"never":this._toggleSelectionFind.checked=!1;break;case"multiline":{const i=!!e&&e.startLineNumber!==e.endLineNumber;this._toggleSelectionFind.checked=i;break}}this._tryUpdateWidgetWidth(),this._updateButtons(),this._revealTimeouts.push(setTimeout(()=>{this._domNode.classList.add("visible"),this._domNode.setAttribute("aria-hidden","false")},0)),this._revealTimeouts.push(setTimeout(()=>{this._findInput.validate()},200)),this._codeEditor.layoutOverlayWidget(this);let t=!0;if(this._codeEditor.getOption(50).seedSearchStringFromSelection&&e){const i=this._codeEditor.getDomNode();if(i){const n=Li(i),o=this._codeEditor.getScrolledVisiblePosition(e.getStartPosition()),r=n.left+(o?o.left:0),a=o?o.top:0;if(this._viewZone&&a<this._viewZone.heightInPx){e.endLineNumber>e.startLineNumber&&(t=!1);const l=DQ(this._domNode).left;r>l&&(t=!1);const c=this._codeEditor.getScrolledVisiblePosition(e.getEndPosition());n.left+(c?c.left:0)>l&&(t=!1)}}}this._showViewZone(t)}}_hide(e){this._revealTimeouts.forEach(t=>{clearTimeout(t)}),this._revealTimeouts=[],this._isVisible&&(this._isVisible=!1,this._updateButtons(),this._domNode.classList.remove("visible"),this._domNode.setAttribute("aria-hidden","true"),this._findInput.clearMessage(),e&&this._codeEditor.focus(),this._codeEditor.layoutOverlayWidget(this),this._removeViewZone())}_layoutViewZone(e){if(!this._codeEditor.getOption(50).addExtraSpaceOnTop){this._removeViewZone();return}if(!this._isVisible)return;const i=this._viewZone;this._viewZoneId!==void 0||!i||this._codeEditor.changeViewZones(n=>{i.heightInPx=this._getHeight(),this._viewZoneId=n.addZone(i),this._codeEditor.setScrollTop(e||this._codeEditor.getScrollTop()+i.heightInPx)})}_showViewZone(e=!0){if(!this._isVisible||!this._codeEditor.getOption(50).addExtraSpaceOnTop)return;this._viewZone===void 0&&(this._viewZone=new XM(0));const i=this._viewZone;this._codeEditor.changeViewZones(n=>{if(this._viewZoneId!==void 0){const o=this._getHeight();if(o===i.heightInPx)return;const r=o-i.heightInPx;i.heightInPx=o,n.layoutZone(this._viewZoneId),e&&this._codeEditor.setScrollTop(this._codeEditor.getScrollTop()+r);return}else{let o=this._getHeight();if(o-=this._codeEditor.getOption(96).top,o<=0)return;i.heightInPx=o,this._viewZoneId=n.addZone(i),e&&this._codeEditor.setScrollTop(this._codeEditor.getScrollTop()+o)}})}_removeViewZone(){this._codeEditor.changeViewZones(e=>{this._viewZoneId!==void 0&&(e.removeZone(this._viewZoneId),this._viewZoneId=void 0,this._viewZone&&(this._codeEditor.setScrollTop(this._codeEditor.getScrollTop()-this._viewZone.heightInPx),this._viewZone=void 0))})}_tryUpdateWidgetWidth(){if(!this._isVisible||!this._domNode.isConnected)return;const e=this._codeEditor.getLayoutInfo();if(e.contentWidth<=0){this._domNode.classList.add("hiddenEditor");return}else this._domNode.classList.contains("hiddenEditor")&&this._domNode.classList.remove("hiddenEditor");const i=e.width,n=e.minimap.minimapWidth;let o=!1,r=!1,a=!1;if(this._resized&&Ro(this._domNode)>Vl){this._domNode.style.maxWidth=`${i-28-n-15}px`,this._replaceInput.width=Ro(this._findInput.domNode);return}if(Vl+28+n>=i&&(r=!0),Vl+28+n-$C>=i&&(a=!0),Vl+28+n-$C>=i+50&&(o=!0),this._domNode.classList.toggle("collapsed-find-widget",o),this._domNode.classList.toggle("narrow-find-widget",a),this._domNode.classList.toggle("reduced-find-widget",r),!a&&!o&&(this._domNode.style.maxWidth=`${i-28-n-15}px`),this._findInput.layout({collapsedFindWidget:o,narrowFindWidget:a,reducedFindWidget:r}),this._resized){const l=this._findInput.inputBox.element.clientWidth;l>0&&(this._replaceInput.width=l)}else this._isReplaceVisible&&(this._replaceInput.width=Ro(this._findInput.domNode))}_getHeight(){let e=0;return e+=4,e+=this._findInput.inputBox.height+2,this._isReplaceVisible&&(e+=4,e+=this._replaceInput.inputBox.height+2),e+=4,e}_tryUpdateHeight(){const e=this._getHeight();return this._cachedHeight!==null&&this._cachedHeight===e?!1:(this._cachedHeight=e,this._domNode.style.height=`${e}px`,!0)}focusFindInput(){this._findInput.select(),this._findInput.focus()}focusReplaceInput(){this._replaceInput.select(),this._replaceInput.focus()}highlightFindOptions(){this._findInput.highlightFindOptions()}_updateSearchScope(){if(this._codeEditor.hasModel()&&this._toggleSelectionFind.checked){const e=this._codeEditor.getSelections();e.map(t=>{t.endColumn===1&&t.endLineNumber>t.startLineNumber&&(t=t.setEndPosition(t.endLineNumber-1,this._codeEditor.getModel().getLineMaxColumn(t.endLineNumber-1)));const i=this._state.currentMatch;return t.startLineNumber!==t.endLineNumber&&!k.equalsRange(t,i)?t:null}).filter(t=>!!t),e.length&&this._state.change({searchScope:e},!0)}}_onFindInputMouseDown(e){e.middleButton&&e.stopPropagation()}_onFindInputKeyDown(e){if(e.equals(Mj|3))if(this._keybindingService.dispatchEvent(e,e.target)){e.preventDefault();return}else{this._findInput.inputBox.insertAtCursor(`
`),e.preventDefault();return}if(e.equals(2)){this._isReplaceVisible?this._replaceInput.focus():this._findInput.focusOnCaseSensitive(),e.preventDefault();return}if(e.equals(2066)){this._codeEditor.focus(),e.preventDefault();return}if(e.equals(16))return Oj(e,this._findInput.getValue(),this._findInput.domNode.querySelector("textarea"));if(e.equals(18))return Pj(e,this._findInput.getValue(),this._findInput.domNode.querySelector("textarea"))}_onReplaceInputKeyDown(e){if(e.equals(Mj|3))if(this._keybindingService.dispatchEvent(e,e.target)){e.preventDefault();return}else{this._replaceInput.inputBox.insertAtCursor(`
`),e.preventDefault();return}if(e.equals(2)){this._findInput.focusOnCaseSensitive(),e.preventDefault();return}if(e.equals(1026)){this._findInput.focus(),e.preventDefault();return}if(e.equals(2066)){this._codeEditor.focus(),e.preventDefault();return}if(e.equals(16))return Oj(e,this._replaceInput.inputBox.value,this._replaceInput.inputBox.element.querySelector("textarea"));if(e.equals(18))return Pj(e,this._replaceInput.inputBox.value,this._replaceInput.inputBox.element.querySelector("textarea"))}getVerticalSashLeft(e){return 0}_keybindingLabelFor(e){const t=this._keybindingService.lookupKeybinding(e);return t?` (${t.getLabel()})`:""}_buildDomNode(){const i=this._codeEditor.getOption(50).history,n=this._codeEditor.getOption(50).replaceHistory;this._findInput=this._register(new c5(null,this._contextViewProvider,{width:kTe,label:uTe,placeholder:gTe,appendCaseSensitiveLabel:this._keybindingLabelFor(Ut.ToggleCaseSensitiveCommand),appendWholeWordsLabel:this._keybindingLabelFor(Ut.ToggleWholeWordCommand),appendRegexLabel:this._keybindingLabelFor(Ut.ToggleRegexCommand),validation:h=>{if(h.length===0||!this._findInput.getRegex())return null;try{return new RegExp(h,"gu"),null}catch(u){return{content:u.message}}},flexibleHeight:!0,flexibleWidth:!0,flexibleMaxHeight:118,showCommonFindToggles:!0,showHistoryHint:()=>Tj(this._keybindingService),inputBoxStyles:NI,toggleStyles:TI,history:i==="workspace"?this._findWidgetSearchHistory:new Set([])},this._contextKeyService)),this._findInput.setRegex(!!this._state.isRegex),this._findInput.setCaseSensitive(!!this._state.matchCase),this._findInput.setWholeWords(!!this._state.wholeWord),this._register(this._findInput.onKeyDown(h=>{h.equals(3)&&!this._codeEditor.getOption(50).findOnType&&this._state.change({searchString:this._findInput.getValue()},!0),this._onFindInputKeyDown(h)})),this._register(this._findInput.inputBox.onDidChange(()=>{this._ignoreChangeEvent||!this._codeEditor.getOption(50).findOnType||this._state.change({searchString:this._findInput.getValue()},!0)})),this._register(this._findInput.onDidOptionChange(()=>{this._state.change({isRegex:this._findInput.getRegex(),wholeWord:this._findInput.getWholeWords(),matchCase:this._findInput.getCaseSensitive()},!0)})),this._register(this._findInput.onCaseSensitiveKeyDown(h=>{h.equals(1026)&&this._isReplaceVisible&&(this._replaceInput.focus(),h.preventDefault())})),this._register(this._findInput.onRegexKeyDown(h=>{h.equals(2)&&this._isReplaceVisible&&(this._replaceInput.focusOnPreserve(),h.preventDefault())})),this._register(this._findInput.inputBox.onDidHeightChange(h=>{this._tryUpdateHeight()&&this._showViewZone()})),vo&&this._register(this._findInput.onMouseDown(h=>this._onFindInputMouseDown(h))),this._matchesCount=document.createElement("div"),this._matchesCount.className="matchesCount",this._updateMatchesCount();const o={groupId:"find-widget"};this._prevBtn=this._register(new af({label:fTe+this._keybindingLabelFor(Ut.PreviousMatchFindAction),icon:cTe,hoverLifecycleOptions:o,onTrigger:()=>{gu(this._codeEditor.getAction(Ut.PreviousMatchFindAction)).run().then(void 0,Oe)}},this._hoverService)),this._nextBtn=this._register(new af({label:pTe+this._keybindingLabelFor(Ut.NextMatchFindAction),icon:dTe,hoverLifecycleOptions:o,onTrigger:()=>{gu(this._codeEditor.getAction(Ut.NextMatchFindAction)).run().then(void 0,Oe)}},this._hoverService));const r=document.createElement("div");r.className="find-part",r.appendChild(this._findInput.domNode);const a=document.createElement("div");a.className="find-actions",r.appendChild(a),a.appendChild(this._matchesCount),a.appendChild(this._prevBtn.domNode),a.appendChild(this._nextBtn.domNode),this._toggleSelectionFind=this._register(new Mh({icon:rTe,title:mTe+this._keybindingLabelFor(Ut.ToggleSearchScopeCommand),isChecked:!1,hoverLifecycleOptions:o,inputActiveOptionBackground:te(iC),inputActiveOptionBorder:te(X0),inputActiveOptionForeground:te(Q0)})),this._register(this._toggleSelectionFind.onChange(()=>{if(this._toggleSelectionFind.checked){if(this._codeEditor.hasModel()){let h=this._codeEditor.getSelections();h=h.map(u=>(u.endColumn===1&&u.endLineNumber>u.startLineNumber&&(u=u.setEndPosition(u.endLineNumber-1,this._codeEditor.getModel().getLineMaxColumn(u.endLineNumber-1))),u.isEmpty()?null:u)).filter(u=>!!u),h.length&&this._state.change({searchScope:h},!0)}}else this._state.change({searchScope:null},!0)})),a.appendChild(this._toggleSelectionFind.domNode),this._closeBtn=this._register(new af({label:_Te+this._keybindingLabelFor(Ut.CloseFindWidgetCommand),icon:pie,hoverLifecycleOptions:o,onTrigger:()=>{this._state.change({isRevealed:!1,searchScope:null},!1)},onKeyDown:h=>{h.equals(2)&&this._isReplaceVisible&&(this._replaceBtn.isEnabled()?this._replaceBtn.focus():this._codeEditor.focus(),h.preventDefault())}},this._hoverService)),this._replaceInput=this._register(new d5(null,void 0,{label:bTe,placeholder:wTe,appendPreserveCaseLabel:this._keybindingLabelFor(Ut.TogglePreserveCaseCommand),history:n==="workspace"?this._replaceWidgetHistory:new Set([]),flexibleHeight:!0,flexibleWidth:!0,flexibleMaxHeight:118,showHistoryHint:()=>Tj(this._keybindingService),inputBoxStyles:NI,toggleStyles:TI,hoverLifecycleOptions:o},this._contextKeyService,!0)),this._replaceInput.setPreserveCase(!!this._state.preserveCase),this._register(this._replaceInput.onKeyDown(h=>this._onReplaceInputKeyDown(h))),this._register(this._replaceInput.inputBox.onDidChange(()=>{this._state.change({replaceString:this._replaceInput.inputBox.value},!1)})),this._register(this._replaceInput.inputBox.onDidHeightChange(h=>{this._isReplaceVisible&&this._tryUpdateHeight()&&this._showViewZone()})),this._register(this._replaceInput.onDidOptionChange(()=>{this._state.change({preserveCase:this._replaceInput.getPreserveCase()},!0)})),this._register(this._replaceInput.onPreserveCaseKeyDown(h=>{h.equals(2)&&(this._prevBtn.isEnabled()?this._prevBtn.focus():this._nextBtn.isEnabled()?this._nextBtn.focus():this._toggleSelectionFind.enabled?this._toggleSelectionFind.focus():this._closeBtn.isEnabled()&&this._closeBtn.focus(),h.preventDefault())})),this._replaceBtn=this._register(new af({label:vTe+this._keybindingLabelFor(Ut.ReplaceOneAction),icon:aTe,hoverLifecycleOptions:o,onTrigger:()=>{this._controller.replace()},onKeyDown:h=>{h.equals(1026)&&(this._closeBtn.focus(),h.preventDefault())}},this._hoverService)),this._replaceAllBtn=this._register(new af({label:CTe+this._keybindingLabelFor(Ut.ReplaceAllAction),icon:lTe,hoverLifecycleOptions:o,onTrigger:()=>{this._controller.replaceAll()}},this._hoverService));const l=document.createElement("div");l.className="replace-part",l.appendChild(this._replaceInput.domNode);const c=document.createElement("div");c.className="replace-actions",l.appendChild(c),c.appendChild(this._replaceBtn.domNode),c.appendChild(this._replaceAllBtn.domNode),this._toggleReplaceBtn=this._register(new af({label:yTe,className:"codicon toggle left",onTrigger:()=>{this._state.change({isReplaceRevealed:!this._isReplaceVisible},!1),this._isReplaceVisible&&(this._replaceInput.width=Ro(this._findInput.domNode),this._replaceInput.inputBox.layout()),this._showViewZone()}},this._hoverService)),this._toggleReplaceBtn.setExpanded(this._isReplaceVisible),this._domNode=document.createElement("div"),this._domNode.className="editor-widget find-widget",this._domNode.setAttribute("aria-hidden","true"),this._domNode.ariaLabel=hTe,this._domNode.role="dialog",this._domNode.style.width=`${Vl}px`,this._domNode.appendChild(this._toggleReplaceBtn.domNode),this._domNode.appendChild(r),this._domNode.appendChild(this._closeBtn.domNode),this._domNode.appendChild(l),this._resizeSash=this._register(new ls(this._domNode,this,{orientation:0,size:2})),this._resized=!1;let d=Vl;this._register(this._resizeSash.onDidStart(()=>{d=Ro(this._domNode)})),this._register(this._resizeSash.onDidChange(h=>{this._resized=!0;const u=d+h.startX-h.currentX;if(u<Vl)return;const g=parseFloat(TR(this._domNode).maxWidth)||0;u>g||(this._domNode.style.width=`${u}px`,this._isReplaceVisible&&(this._replaceInput.width=Ro(this._findInput.domNode)),this._findInput.inputBox.layout(),this._tryUpdateHeight())})),this._register(this._resizeSash.onDidReset(()=>{const h=Ro(this._domNode);if(h<Vl)return;let u=Vl;if(!this._resized||h===Vl){const g=this._codeEditor.getLayoutInfo();u=g.width-28-g.minimap.minimapWidth-15,this._resized=!0}this._domNode.style.width=`${u}px`,this._isReplaceVisible&&(this._replaceInput.width=Ro(this._findInput.domNode)),this._findInput.inputBox.layout()}))}updateAccessibilitySupport(){const e=this._codeEditor.getOption(2);this._findInput.setFocusInputOnOptionClick(e!==2)}getViewState(){let e=!1;return this._viewZone&&this._viewZoneId&&(e=this._viewZone.heightInPx>this._codeEditor.getScrollTop()),{widgetViewZoneVisible:e,scrollTop:this._codeEditor.getScrollTop()}}setViewState(e){e&&e.widgetViewZoneVisible&&this._layoutViewZone(e.scrollTop)}};wN.ID="editor.contrib.findWidget";let h5=wN;class af extends Xo{constructor(e,t){super(),this._opts=e;let i="button";this._opts.className&&(i=i+" "+this._opts.className),this._opts.icon&&(i=i+" "+xe.asClassName(this._opts.icon)),this._domNode=document.createElement("div"),this._domNode.tabIndex=0,this._domNode.className=i,this._domNode.setAttribute("role","button"),this._domNode.setAttribute("aria-label",this._opts.label),this._register(t.setupDelayedHover(this._domNode,{content:this._opts.label,style:1},e.hoverLifecycleOptions)),this.onclick(this._domNode,n=>{this._opts.onTrigger(),n.preventDefault()}),this.onkeydown(this._domNode,n=>{if(n.equals(10)||n.equals(3)){this._opts.onTrigger(),n.preventDefault();return}this._opts.onKeyDown?.(n)})}get domNode(){return this._domNode}isEnabled(){return this._domNode.tabIndex>=0}focus(){this._domNode.focus()}setEnabled(e){this._domNode.classList.toggle("disabled",!e),this._domNode.setAttribute("aria-disabled",String(!e)),this._domNode.tabIndex=e?0:-1}setExpanded(e){this._domNode.setAttribute("aria-expanded",String(!!e)),e?(this._domNode.classList.remove(...xe.asClassNameArray(Nj)),this._domNode.classList.add(...xe.asClassNameArray(Rj))):(this._domNode.classList.remove(...xe.asClassNameArray(Rj)),this._domNode.classList.add(...xe.asClassNameArray(Nj)))}}ua((s,e)=>{const t=s.getColor(_u);t&&e.addRule(`.monaco-editor .findMatch { border: 1px ${Cl(s.type)?"dotted":"solid"} ${t}; box-sizing: border-box; }`);const i=s.getColor(hme);i&&e.addRule(`.monaco-editor .findScope { border: 1px ${Cl(s.type)?"dashed":"solid"} ${i}; }`);const n=s.getColor(rt);n&&e.addRule(`.monaco-editor .find-widget { border: 1px solid ${n}; }`);const o=s.getColor(cme);o&&e.addRule(`.monaco-editor .findMatchInline { color: ${o}; }`);const r=s.getColor(dme);r&&e.addRule(`.monaco-editor .currentFindMatchInline { color: ${r}; }`)});var ITe=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},Fj=function(s,e){return function(t,i){e(t,i,s)}};let u5=class extends F{constructor(e,t,i,n){super(),this._container=e,this._getContent=t,this._clipboardService=i,this._hoverService=n,this._container.classList.add("hover-row-with-copy"),this._button=this._register(new af({label:m(1128,"Copy"),icon:Z.copy,onTrigger:()=>this._copyContent(),className:"hover-copy-button"},this._hoverService)),this._container.appendChild(this._button.domNode)}async _copyContent(){const e=this._getContent();e&&(await this._clipboardService.writeText(e),xl(m(1129,"Copied to clipboard")))}};u5=ITe([Fj(2,Ko),Fj(3,so)],u5);class DTe{constructor(e,t,i){this.provider=e,this.hover=t,this.ordinal=i}}async function TTe(s,e,t,i,n){const o=await Promise.resolve(s.provideHover(t,i,n)).catch(ji);if(!(!o||!NTe(o)))return new DTe(s,o,e)}function hH(s,e,t,i,n=!1){const r=s.ordered(e,n).map((a,l)=>TTe(a,l,e,t,i));return Jr.fromPromisesResolveOrder(r).coalesce()}async function Une(s,e,t,i,n=!1){const o=[];for await(const r of hH(s,e,t,i,n))o.push(r.hover);return o}Lo("_executeHoverProvider",(s,e,t)=>{const i=s.get(ge);return Une(i.hoverProvider,e,t,Ke.None)});Lo("_executeHoverProvider_recursive",(s,e,t)=>{const i=s.get(ge);return Une(i.hoverProvider,e,t,Ke.None,!0)});function NTe(s){const e=typeof s.range<"u",t=typeof s.contents<"u"&&s.contents&&s.contents.length>0;return e&&t}var RTe=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},j_=function(s,e){return function(t,i){e(t,i,s)}};const Qf=ie,ATe=Qt("hover-increase-verbosity",Z.add,m(1130,"Icon for increaseing hover verbosity.")),MTe=Qt("hover-decrease-verbosity",Z.remove,m(1131,"Icon for decreasing hover verbosity."));class Ta{constructor(e,t,i,n,o,r=void 0){this.owner=e,this.range=t,this.contents=i,this.isBeforeContent=n,this.ordinal=o,this.source=r}isValidForHoverAnchor(e){return e.type===1&&this.range.startColumn<=e.range.startColumn&&this.range.endColumn>=e.range.endColumn}}class qne{constructor(e,t,i){this.hover=e,this.hoverProvider=t,this.hoverPosition=i}supportsVerbosityAction(e){switch(e){case Ao.Increase:return this.hover.canIncreaseVerbosity??!1;case Ao.Decrease:return this.hover.canDecreaseVerbosity??!1}}}let KS=class{constructor(e,t,i,n,o,r,a){this._editor=e,this._markdownRendererService=t,this._configurationService=i,this._languageFeaturesService=n,this._keybindingService=o,this._hoverService=r,this._commandService=a,this.hoverOrdinal=3}createLoadingMessage(e){return new Ta(this,e.range,[new hs().appendText(m(1132,"Loading..."))],!1,2e3)}computeSync(e,t){if(!this._editor.hasModel()||e.type!==1)return[];const i=this._editor.getModel(),n=e.range.startLineNumber,o=i.getLineMaxColumn(n),r=[];let a=1e3;const l=i.getLineLength(n),c=i.getLanguageIdAtPosition(e.range.startLineNumber,e.range.startColumn),d=this._editor.getOption(133),h=this._configurationService.getValue("editor.maxTokenizationLineLength",{overrideIdentifier:c});let u=!1;d>=0&&l>d&&e.range.startColumn>=d&&(u=!0,r.push(new Ta(this,e.range,[{value:m(1133,"Rendering paused for long line for performance reasons. This can be configured via `editor.stopRenderingLineAfter`.")}],!1,a++))),!u&&typeof h=="number"&&l>=h&&r.push(new Ta(this,e.range,[{value:m(1134,"Tokenization is skipped for long lines for performance reasons. This can be configured via `editor.maxTokenizationLineLength`.")}],!1,a++));let g=!1;for(const f of t){const p=f.range.startLineNumber===n?f.range.startColumn:1,_=f.range.endLineNumber===n?f.range.endColumn:o,b=f.options.hoverMessage;if(!b||wv(b))continue;f.options.beforeContentClassName&&(g=!0);const w=new k(e.range.startLineNumber,p,e.range.startLineNumber,_);r.push(new Ta(this,w,tW(b),g,a++))}return r}computeAsync(e,t,i,n){if(!this._editor.hasModel()||e.type!==1)return Jr.EMPTY;const o=this._editor.getModel(),r=this._languageFeaturesService.hoverProvider;return r.has(o)?this._getMarkdownHovers(r,o,e,n):Jr.EMPTY}async*_getMarkdownHovers(e,t,i,n){const o=i.range.getStartPosition(),r=hH(e,t,o,n);for await(const a of r)if(!wv(a.hover.contents)){const l=a.hover.range?k.lift(a.hover.range):i.range,c=new qne(a.hover,a.provider,o);yield new Ta(this,l,a.hover.contents,!1,a.ordinal,c)}}renderHoverParts(e,t){return this._renderedHoverParts=new OTe(t,e.fragment,this,this._editor,this._commandService,this._keybindingService,this._hoverService,this._configurationService,this._markdownRendererService,e.onContentsChanged),this._renderedHoverParts}handleScroll(e){this._renderedHoverParts?.handleScroll(e)}getAccessibleContent(e){return this._renderedHoverParts?.getAccessibleContent(e)??""}updateMarkdownHoverVerbosityLevel(e,t){return Promise.resolve(this._renderedHoverParts?.updateMarkdownHoverPartVerbosityLevel(e,t))}};KS=RTe([j_(1,Ua),j_(2,He),j_(3,ge),j_(4,ft),j_(5,so),j_(6,Gt)],KS);class AL{constructor(e,t,i,n){this.hoverPart=e,this.hoverElement=t,this.disposables=i,this.actionsContainer=n}dispose(){this.disposables.dispose()}}class OTe{constructor(e,t,i,n,o,r,a,l,c,d){this._hoverParticipant=i,this._editor=n,this._commandService=o,this._keybindingService=r,this._hoverService=a,this._configurationService=l,this._markdownRendererService=c,this._onFinishedRendering=d,this._ongoingHoverOperations=new Map,this._disposables=new U,this.renderedHoverParts=this._renderHoverParts(e,t,this._onFinishedRendering),this._disposables.add(pe(()=>{this.renderedHoverParts.forEach(h=>{h.dispose()}),this._ongoingHoverOperations.forEach(h=>{h.tokenSource.dispose(!0)})}))}_renderHoverParts(e,t,i){return e.sort(ns(n=>n.ordinal,Wo)),e.map(n=>{const o=this._renderHoverPart(n,i);return t.appendChild(o.hoverElement),o})}_renderHoverPart(e,t){const i=this._renderMarkdownHover(e,t),n=i.hoverElement,o=e.source,r=new U;if(r.add(i),!o)return new AL(e,n,r);const a=o.supportsVerbosityAction(Ao.Increase),l=o.supportsVerbosityAction(Ao.Decrease);if(!a&&!l)return new AL(e,n,r);const c=Qf("div.verbosity-actions");n.prepend(c);const d=Qf("div.verbosity-actions-inner");return c.append(d),r.add(this._renderHoverExpansionAction(d,Ao.Increase,a)),r.add(this._renderHoverExpansionAction(d,Ao.Decrease,l)),new AL(e,n,r,d)}_renderMarkdownHover(e,t){return jne(this._editor,e,this._markdownRendererService,t)}_renderHoverExpansionAction(e,t,i){const n=new U,o=t===Ao.Increase,r=X(e,Qf(xe.asCSSSelector(o?ATe:MTe)));r.tabIndex=0;const a=new Cv("mouse",void 0,{target:e,position:{hoverPosition:0}},this._configurationService,this._hoverService);if(n.add(this._hoverService.setupManagedHover(a,r,FTe(this._keybindingService,t))),!i)return r.classList.add("disabled"),n;r.classList.add("enabled");const l=()=>this._commandService.executeCommand(t===Ao.Increase?IA:DA,{focus:!0});return n.add(new kne(r,l)),n.add(new Ene(r,l,[3,10])),n}handleScroll(e){this.renderedHoverParts.forEach(t=>{const i=t.actionsContainer;if(!i)return;const n=t.hoverElement,r=e.scrollTop+e.height,a=n.offsetTop,l=n.clientHeight,c=a+l,d=22;let h;c<=r||a>=r?h=l-d:h=r-a-d,i.style.top=`${h}px`})}async updateMarkdownHoverPartVerbosityLevel(e,t){const i=this._editor.getModel();if(!i)return;const n=this._getRenderedHoverPartAtIndex(t),o=n?.hoverPart.source;if(!n||!o?.supportsVerbosityAction(e))return;const r=await this._fetchHover(o,i,e);if(!r)return;const a=new qne(r,o.hoverProvider,o.hoverPosition),l=n.hoverPart,c=new Ta(this._hoverParticipant,l.range,r.contents,l.isBeforeContent,l.ordinal,a),d=this._updateRenderedHoverPart(t,c);if(d)return{hoverPart:c,hoverElement:d.hoverElement}}getAccessibleContent(e){const t=this.renderedHoverParts.findIndex(r=>r.hoverPart===e);if(t===-1)return;const i=this._getRenderedHoverPartAtIndex(t);return i?i.hoverElement.innerText.replace(/[^\S\n\r]+/gu," "):void 0}async _fetchHover(e,t,i){let n=i===Ao.Increase?1:-1;const o=e.hoverProvider,r=this._ongoingHoverOperations.get(o);r&&(r.tokenSource.cancel(),n+=r.verbosityDelta);const a=new si;this._ongoingHoverOperations.set(o,{verbosityDelta:n,tokenSource:a});const l={verbosityRequest:{verbosityDelta:n,previousHover:e.hover}};let c;try{c=await Promise.resolve(o.provideHover(t,e.hoverPosition,a.token,l))}catch(d){ji(d)}return a.dispose(),this._ongoingHoverOperations.delete(o),c}_updateRenderedHoverPart(e,t){if(e>=this.renderedHoverParts.length||e<0)return;const i=this._renderHoverPart(t,this._onFinishedRendering),n=this.renderedHoverParts[e],o=n.hoverElement,r=i.hoverElement,a=Array.from(r.children);o.replaceChildren(...a);const l=new AL(t,o,i.disposables,i.actionsContainer);return n.dispose(),this.renderedHoverParts[e]=l,l}_getRenderedHoverPartAtIndex(e){return this.renderedHoverParts[e]}dispose(){this._disposables.dispose()}}function PTe(s,e,t,i){e.sort(ns(o=>o.ordinal,Wo));const n=[];for(const o of e){const r=jne(t,o,i,s.onContentsChanged);s.fragment.appendChild(r.hoverElement),n.push(r)}return new qm(n)}function jne(s,e,t,i){const n=new U,o=Qf("div.hover-row"),r=Qf("div.hover-row-contents");o.appendChild(r);const a=e.contents;for(const c of a){if(wv(c))continue;const d=Qf("div.markdown-hover"),h=X(d,Qf("div.hover-contents")),u=n.add(t.render(c,{context:s,asyncRenderCallback:()=>{h.className="hover-contents code-hover-contents",i()}}));h.appendChild(u.element),r.appendChild(d)}return{hoverPart:e,hoverElement:o,dispose(){n.dispose()}}}function FTe(s,e){switch(e){case Ao.Increase:{const t=s.lookupKeybinding(IA);return t?m(1135,"Increase Hover Verbosity ({0})",t.getLabel()):m(1136,"Increase Hover Verbosity")}case Ao.Decrease:{const t=s.lookupKeybinding(DA);return t?m(1137,"Decrease Hover Verbosity ({0})",t.getLabel()):m(1138,"Decrease Hover Verbosity")}}}const Bj=ie;class BTe extends F{constructor(e,t,i){super(),this.model=t,this.pixelRatio=i,this._onDidChange=new M,this.onDidChange=this._onDidChange.event,this._onColorFlushed=new M,this.onColorFlushed=this._onColorFlushed.event,this._domNode=Bj(".saturation-wrap"),X(e,this._domNode),this._canvas=document.createElement("canvas"),this._canvas.className="saturation-box",X(this._domNode,this._canvas),this.selection=Bj(".saturation-selection"),X(this._domNode,this.selection),this.layout(),this._register(W(this._domNode,se.POINTER_DOWN,n=>this.onPointerDown(n))),this._register(this.model.onDidChangeColor(this.onDidChangeColor,this)),this.monitor=null}get domNode(){return this._domNode}onPointerDown(e){if(!e.target||!(e.target instanceof Element))return;this.monitor=this._register(new eC);const t=Li(this._domNode);e.target!==this.selection&&this.onDidChangePosition(e.offsetX,e.offsetY),this.monitor.startMonitoring(e.target,e.pointerId,e.buttons,n=>this.onDidChangePosition(n.pageX-t.left,n.pageY-t.top),()=>null);const i=W(e.target.ownerDocument,se.POINTER_UP,()=>{this._onColorFlushed.fire(),i.dispose(),this.monitor&&(this.monitor.stopMonitoring(!0),this.monitor=null)},!0)}onDidChangePosition(e,t){const i=Math.max(0,Math.min(1,e/this.width)),n=Math.max(0,Math.min(1,1-t/this.height));this.paintSelection(i,n),this._onDidChange.fire({s:i,v:n})}layout(){this.width=this._domNode.offsetWidth,this.height=this._domNode.offsetHeight,this._canvas.width=this.width*this.pixelRatio,this._canvas.height=this.height*this.pixelRatio,this.paint();const e=this.model.color.hsva;this.paintSelection(e.s,e.v)}paint(){const e=this.model.color.hsva,t=new q(new vc(e.h,1,1,1)),i=this._canvas.getContext("2d"),n=i.createLinearGradient(0,0,this._canvas.width,0);n.addColorStop(0,"rgba(255, 255, 255, 1)"),n.addColorStop(.5,"rgba(255, 255, 255, 0.5)"),n.addColorStop(1,"rgba(255, 255, 255, 0)");const o=i.createLinearGradient(0,0,0,this._canvas.height);o.addColorStop(0,"rgba(0, 0, 0, 0)"),o.addColorStop(1,"rgba(0, 0, 0, 1)"),i.rect(0,0,this._canvas.width,this._canvas.height),i.fillStyle=q.Format.CSS.format(t),i.fill(),i.fillStyle=n,i.fill(),i.fillStyle=o,i.fill()}paintSelection(e,t){this.selection.style.left=`${e*this.width}px`,this.selection.style.top=`${this.height-t*this.height}px`}onDidChangeColor(e){if(this.monitor&&this.monitor.isMonitoring())return;this.paint();const t=e.hsva;this.paintSelection(t.s,t.v)}}class WTe extends F{constructor(e){super(),this._onClicked=this._register(new M),this.onClicked=this._onClicked.event,this._button=X(e,document.createElement("button")),this._button.classList.add("insert-button"),this._button.textContent="Insert",this._register(W(this._button,se.CLICK,()=>{this._onClicked.fire()}))}get button(){return this._button}}const UC=ie;class Gne extends F{constructor(e,t,i){super(),this.model=t,this._onDidChange=new M,this.onDidChange=this._onDidChange.event,this._onColorFlushed=new M,this.onColorFlushed=this._onColorFlushed.event,i==="standalone"?(this.domNode=X(e,UC(".standalone-strip")),this.overlay=X(this.domNode,UC(".standalone-overlay"))):(this.domNode=X(e,UC(".strip")),this.overlay=X(this.domNode,UC(".overlay"))),this.slider=X(this.domNode,UC(".slider")),this.slider.style.top="0px",this._register(W(this.domNode,se.POINTER_DOWN,n=>this.onPointerDown(n))),this._register(t.onDidChangeColor(this.onDidChangeColor,this)),this.layout()}layout(){this.height=this.domNode.offsetHeight-this.slider.offsetHeight;const e=this.getValue(this.model.color);this.updateSliderPosition(e)}onDidChangeColor(e){const t=this.getValue(e);this.updateSliderPosition(t)}onPointerDown(e){if(!e.target||!(e.target instanceof Element))return;const t=this._register(new eC),i=Li(this.domNode);this.domNode.classList.add("grabbing"),e.target!==this.slider&&this.onDidChangeTop(e.offsetY),t.startMonitoring(e.target,e.pointerId,e.buttons,o=>this.onDidChangeTop(o.pageY-i.top),()=>null);const n=W(e.target.ownerDocument,se.POINTER_UP,()=>{this._onColorFlushed.fire(),n.dispose(),t.stopMonitoring(!0),this.domNode.classList.remove("grabbing")},!0)}onDidChangeTop(e){const t=Math.max(0,Math.min(1,1-e/this.height));this.updateSliderPosition(t),this._onDidChange.fire(t)}updateSliderPosition(e){this.slider.style.top=`${(1-e)*this.height}px`}}class HTe extends Gne{constructor(e,t,i){super(e,t,i),this.domNode.classList.add("opacity-strip"),this.onDidChangeColor(this.model.color)}onDidChangeColor(e){super.onDidChangeColor(e);const{r:t,g:i,b:n}=e.rgba,o=new q(new G(t,i,n,1)),r=new q(new G(t,i,n,0));this.overlay.style.background=`linear-gradient(to bottom, ${o} 0%, ${r} 100%)`}getValue(e){return e.hsva.a}}class VTe extends Gne{constructor(e,t,i){super(e,t,i),this.domNode.classList.add("hue-strip")}getValue(e){return 1-e.hsva.h/360}}const zTe=ie;class $Te extends F{constructor(e,t,i,n){super(),this.model=t,this.pixelRatio=i,this._insertButton=null,this._domNode=zTe(".colorpicker-body"),X(e,this._domNode),this._saturationBox=new BTe(this._domNode,this.model,this.pixelRatio),this._register(this._saturationBox),this._register(this._saturationBox.onDidChange(this.onDidSaturationValueChange,this)),this._register(this._saturationBox.onColorFlushed(this.flushColor,this)),this._opacityStrip=new HTe(this._domNode,this.model,n),this._register(this._opacityStrip),this._register(this._opacityStrip.onDidChange(this.onDidOpacityChange,this)),this._register(this._opacityStrip.onColorFlushed(this.flushColor,this)),this._hueStrip=new VTe(this._domNode,this.model,n),this._register(this._hueStrip),this._register(this._hueStrip.onDidChange(this.onDidHueChange,this)),this._register(this._hueStrip.onColorFlushed(this.flushColor,this)),n==="standalone"&&(this._insertButton=this._register(new WTe(this._domNode)),this._domNode.classList.add("standalone-colorpicker"))}flushColor(){this.model.flushColor()}onDidSaturationValueChange({s:e,v:t}){const i=this.model.color.hsva;this.model.color=new q(new vc(i.h,e,t,i.a))}onDidOpacityChange(e){const t=this.model.color.hsva;this.model.color=new q(new vc(t.h,t.s,t.v,e))}onDidHueChange(e){const t=this.model.color.hsva,i=(1-e)*360;this.model.color=new q(new vc(i===360?0:i,t.s,t.v,t.a))}get domNode(){return this._domNode}get saturationBox(){return this._saturationBox}get enterButton(){return this._insertButton}layout(){this._saturationBox.layout(),this._opacityStrip.layout(),this._hueStrip.layout()}}const UTe=ie;class qTe extends F{constructor(e){super(),this._onClicked=this._register(new M),this.onClicked=this._onClicked.event,this._button=document.createElement("div"),this._button.classList.add("close-button"),X(e,this._button);const t=document.createElement("div");t.classList.add("close-button-inner-div"),X(this._button,t),X(t,UTe(".button"+xe.asCSSSelector(Qt("color-picker-close",Z.close,m(885,"Icon to close the color picker"))))).classList.add("close-icon"),this._register(W(this._button,se.CLICK,()=>{this._onClicked.fire()}))}}const ML=ie;class jTe extends F{constructor(e,t,i,n){super(),this.model=t,this.type=n,this._closeButton=null,this._domNode=ML(".colorpicker-header"),X(e,this._domNode),this._pickedColorNode=X(this._domNode,ML(".picked-color")),X(this._pickedColorNode,ML("span.codicon.codicon-color-mode")),this._pickedColorPresentation=X(this._pickedColorNode,document.createElement("span")),this._pickedColorPresentation.classList.add("picked-color-presentation");const o=m(886,"Click to toggle color options (rgb/hsl/hex)");this._pickedColorNode.setAttribute("title",o),this._originalColorNode=X(this._domNode,ML(".original-color")),this._originalColorNode.style.backgroundColor=q.Format.CSS.format(this.model.originalColor)||"",this.backgroundColor=i.getColorTheme().getColor(UE)||q.white,this._register(i.onDidColorThemeChange(r=>{this.backgroundColor=r.getColor(UE)||q.white})),this._register(W(this._pickedColorNode,se.CLICK,()=>this.model.selectNextColorPresentation())),this._register(W(this._originalColorNode,se.CLICK,()=>{this.model.color=this.model.originalColor,this.model.flushColor()})),this._register(t.onDidChangeColor(this.onDidChangeColor,this)),this._register(t.onDidChangePresentation(this.onDidChangePresentation,this)),this._pickedColorNode.style.backgroundColor=q.Format.CSS.format(t.color)||"",this._pickedColorNode.classList.toggle("light",t.color.rgba.a<.5?this.backgroundColor.isLighter():t.color.isLighter()),this.onDidChangeColor(this.model.color),this.type==="standalone"&&(this._domNode.classList.add("standalone-colorpicker"),this._closeButton=this._register(new qTe(this._domNode)))}get closeButton(){return this._closeButton}get pickedColorNode(){return this._pickedColorNode}get originalColorNode(){return this._originalColorNode}onDidChangeColor(e){this._pickedColorNode.style.backgroundColor=q.Format.CSS.format(e)||"",this._pickedColorNode.classList.toggle("light",e.rgba.a<.5?this.backgroundColor.isLighter():e.isLighter()),this.onDidChangePresentation()}onDidChangePresentation(){this._pickedColorPresentation.textContent=this.model.presentation?this.model.presentation.label:""}}const GTe=ie;class Kne extends Xo{constructor(e,t,i,n,o){super(),this.model=t,this.pixelRatio=i,this._register(sS.getInstance(be(e)).onDidChange(()=>this.layout())),this._domNode=GTe(".colorpicker-widget"),e.appendChild(this._domNode),this.header=this._register(new jTe(this._domNode,this.model,n,o)),this.body=this._register(new $Te(this._domNode,this.model,this.pixelRatio,o))}layout(){this.body.layout()}get domNode(){return this._domNode}}class KTe{get color(){return this._color}set color(e){this._color.equals(e)||(this._color=e,this._onDidChangeColor.fire(e))}get presentation(){return this.colorPresentations[this.presentationIndex]}get colorPresentations(){return this._colorPresentations}set colorPresentations(e){this._colorPresentations=e,this.presentationIndex>e.length-1&&(this.presentationIndex=0),this._onDidChangePresentation.fire(this.presentation)}constructor(e,t,i){this.presentationIndex=i,this._onColorFlushed=new M,this.onColorFlushed=this._onColorFlushed.event,this._onDidChangeColor=new M,this.onDidChangeColor=this._onDidChangeColor.event,this._onDidChangePresentation=new M,this.onDidChangePresentation=this._onDidChangePresentation.event,this.originalColor=e,this._color=e,this._colorPresentations=t}selectNextColorPresentation(){this.presentationIndex=(this.presentationIndex+1)%this.colorPresentations.length,this.flushColor(),this._onDidChangePresentation.fire(this.presentation)}guessColorPresentation(e,t){let i=-1;for(let n=0;n<this.colorPresentations.length;n++)if(t.toLowerCase()===this.colorPresentations[n].label){i=n;break}if(i===-1){const n=t.split("(")[0].toLowerCase();for(let o=0;o<this.colorPresentations.length;o++)if(this.colorPresentations[o].label.toLowerCase().startsWith(n)){i=o;break}}i!==-1&&i!==this.presentationIndex&&(this.presentationIndex=i,this._onDidChangePresentation.fire(this.presentation))}flushColor(){this._onColorFlushed.fire(this._color)}}async function Yne(s,e,t){const i=s.getValueInRange(e.range),{red:n,green:o,blue:r,alpha:a}=e.color,l=new G(Math.round(n*255),Math.round(o*255),Math.round(r*255),a),c=new q(l),d=await _ne(s,e,t,Ke.None),h=new KTe(c,[],0);return h.colorPresentations=d||[],h.guessColorPresentation(c,i),{range:k.lift(e.range),model:h,provider:t}}function Zne(s,e,t){const i=[],n=t.presentation.textEdit??{range:e,text:t.presentation.label,forceMoveMarkers:!1};i.push(n),t.presentation.additionalTextEdits&&i.push(...t.presentation.additionalTextEdits);const o=k.lift(n.range),r=s.getModel()._setTrackedRange(null,o,3);return s.executeEdits("colorpicker",i),s.pushUndoStop(),s.getModel()._getTrackedRange(r)??o}async function YS(s,e,t,i,n){const o=await _ne(s,{range:i,color:{red:t.rgba.r/255,green:t.rgba.g/255,blue:t.rgba.b/255,alpha:t.rgba.a}},n.provider,Ke.None);e.colorPresentations=o||[]}var YTe=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},ZTe=function(s,e){return function(t,i){e(t,i,s)}};class ZI{constructor(e,t,i,n){this.owner=e,this.range=t,this.model=i,this.provider=n,this.forceShowAtRange=!0}isValidForHoverAnchor(e){return e.type===1&&this.range.startColumn<=e.range.startColumn&&this.range.endColumn>=e.range.endColumn}static fromBaseColor(e,t){return new ZI(e,t.range,t.model,t.provider)}}let XI=class{constructor(e,t){this._editor=e,this._themeService=t,this.hoverOrdinal=2}computeSync(e,t,i){return[]}computeAsync(e,t,i,n){return Jr.fromPromise(this._computeAsync(e,t,i))}async _computeAsync(e,t,i){if(!this._editor.hasModel())return[];if(!this._isValidRequest(i))return[];const n=Sv.get(this._editor);if(!n)return[];for(const o of t){if(!n.isColorDecoration(o))continue;const r=n.getColorData(o.range.getStartPosition());if(r)return[ZI.fromBaseColor(this,await Yne(this._editor.getModel(),r.colorInfo,r.provider))]}return[]}_isValidRequest(e){const t=this._editor.getOption(168);switch(e){case 0:return t==="hover"||t==="clickAndHover";case 1:return t==="click"||t==="clickAndHover";case 2:return!0}}renderHoverParts(e,t){const i=this._editor;if(t.length===0||!i.hasModel())return new qm([]);const n=i.getOption(75)+8;e.setMinimumDimensions(new St(302,n));const o=new U,r=t[0],a=i.getModel(),l=r.model;this._colorPicker=o.add(new Kne(e.fragment,l,i.getOption(163),this._themeService,"hover"));let c=!1,d=new k(r.range.startLineNumber,r.range.startColumn,r.range.endLineNumber,r.range.endColumn);o.add(l.onColorFlushed(async u=>{await YS(a,l,u,d,r),c=!0,d=Zne(i,d,l)})),o.add(l.onDidChangeColor(u=>{YS(a,l,u,d,r)})),o.add(i.onDidChangeModelContent(u=>{c?c=!1:(e.hide(),i.focus())}));const h={hoverPart:ZI.fromBaseColor(this,r),hoverElement:this._colorPicker.domNode,dispose(){o.dispose()}};return new qm([h])}getAccessibleContent(e){return m(887,"There is a color picker here.")}handleResize(){this._colorPicker?.layout()}handleContentsChanged(){this._colorPicker?.layout()}handleHide(){this._colorPicker?.dispose(),this._colorPicker=void 0}isColorPickerVisible(){return!!this._colorPicker}};XI=YTe([ZTe(1,wi)],XI);class Si{static insert(e,t){return{range:new k(e.lineNumber,e.column,e.lineNumber,e.column),text:t,forceMoveMarkers:!0}}static delete(e){return{range:e,text:null}}static replace(e,t){return{range:e,text:t}}static replaceMove(e,t){return{range:e,text:t,forceMoveMarkers:!0}}}function g5(s,e){return!!s[e]}class QM{constructor(e,t){this.target=e.target,this.isLeftClick=e.event.leftButton,this.isMiddleClick=e.event.middleButton,this.isRightClick=e.event.rightButton,this.mouseMiddleClickAction=t.mouseMiddleClickAction,this.hasTriggerModifier=g5(e.event,t.triggerModifier),this.isMiddleClick&&t.mouseMiddleClickAction==="ctrlLeftClick"&&(this.isMiddleClick=!1,this.isLeftClick=!0,this.hasTriggerModifier=!0),this.hasSideBySideModifier=g5(e.event,t.triggerSideBySideModifier),this.isNoneOrSingleMouseDown=e.event.detail<=1}}class Wj{constructor(e,t){this.keyCodeIsTriggerKey=e.keyCode===t.triggerKey,this.keyCodeIsSideBySideKey=e.keyCode===t.triggerSideBySideKey,this.hasTriggerModifier=g5(e,t.triggerModifier)}}class OL{constructor(e,t,i,n,o){this.mouseMiddleClickAction=o,this.triggerKey=e,this.triggerModifier=t,this.triggerSideBySideKey=i,this.triggerSideBySideModifier=n}equals(e){return this.triggerKey===e.triggerKey&&this.triggerModifier===e.triggerModifier&&this.triggerSideBySideKey===e.triggerSideBySideKey&&this.triggerSideBySideModifier===e.triggerSideBySideModifier&&this.mouseMiddleClickAction===e.mouseMiddleClickAction}}function Hj(s,e){return s==="altKey"?Ye?new OL(57,"metaKey",6,"altKey",e):new OL(5,"ctrlKey",6,"altKey",e):Ye?new OL(6,"altKey",57,"metaKey",e):new OL(6,"altKey",5,"ctrlKey",e)}class AA extends F{constructor(e,t){super(),this._onMouseMoveOrRelevantKeyDown=this._register(new M),this.onMouseMoveOrRelevantKeyDown=this._onMouseMoveOrRelevantKeyDown.event,this._onExecute=this._register(new M),this.onExecute=this._onExecute.event,this._onCancel=this._register(new M),this.onCancel=this._onCancel.event,this._editor=e,this._extractLineNumberFromMouseEvent=t?.extractLineNumberFromMouseEvent??(i=>i.target.position?i.target.position.lineNumber:0),this._opts=Hj(this._editor.getOption(86),this._editor.getOption(87)),this._lastMouseMoveEvent=null,this._hasTriggerKeyOnMouseDown=!1,this._lineNumberOnMouseDown=0,this._register(this._editor.onDidChangeConfiguration(i=>{if(i.hasChanged(86)||i.hasChanged(87)){const n=Hj(this._editor.getOption(86),this._editor.getOption(87));if(this._opts.equals(n))return;this._opts=n,this._lastMouseMoveEvent=null,this._hasTriggerKeyOnMouseDown=!1,this._lineNumberOnMouseDown=0,this._onCancel.fire()}})),this._register(this._editor.onMouseMove(i=>this._onEditorMouseMove(new QM(i,this._opts)))),this._register(this._editor.onMouseDown(i=>this._onEditorMouseDown(new QM(i,this._opts)))),this._register(this._editor.onMouseUp(i=>this._onEditorMouseUp(new QM(i,this._opts)))),this._register(this._editor.onKeyDown(i=>this._onEditorKeyDown(new Wj(i,this._opts)))),this._register(this._editor.onKeyUp(i=>this._onEditorKeyUp(new Wj(i,this._opts)))),this._register(this._editor.onMouseDrag(()=>this._resetHandler())),this._register(this._editor.onDidChangeCursorSelection(i=>this._onDidChangeCursorSelection(i))),this._register(this._editor.onDidChangeModel(i=>this._resetHandler())),this._register(this._editor.onDidChangeModelContent(()=>this._resetHandler())),this._register(this._editor.onDidScrollChange(i=>{(i.scrollTopChanged||i.scrollLeftChanged)&&this._resetHandler()}))}_onDidChangeCursorSelection(e){e.selection&&e.selection.startColumn!==e.selection.endColumn&&this._resetHandler()}_onEditorMouseMove(e){this._lastMouseMoveEvent=e,this._onMouseMoveOrRelevantKeyDown.fire([e,null])}_onEditorMouseDown(e){this._hasTriggerKeyOnMouseDown=e.hasTriggerModifier,this._lineNumberOnMouseDown=this._extractLineNumberFromMouseEvent(e)}_onEditorMouseUp(e){const t=this._extractLineNumberFromMouseEvent(e);this._lineNumberOnMouseDown&&this._lineNumberOnMouseDown===t&&(this._hasTriggerKeyOnMouseDown||e.isMiddleClick&&e.mouseMiddleClickAction==="openLink")&&this._onExecute.fire(e)}_onEditorKeyDown(e){this._lastMouseMoveEvent&&(e.keyCodeIsTriggerKey||e.keyCodeIsSideBySideKey&&e.hasTriggerModifier)?this._onMouseMoveOrRelevantKeyDown.fire([this._lastMouseMoveEvent,e]):e.hasTriggerModifier&&this._onCancel.fire()}_onEditorKeyUp(e){e.keyCodeIsTriggerKey&&this._onCancel.fire()}_resetHandler(){this._lastMouseMoveEvent=null,this._hasTriggerKeyOnMouseDown=!1,this._onCancel.fire()}}class Xne{constructor(e,t){this.range=e,this.direction=t}}class uH{constructor(e,t,i){this.hint=e,this.anchor=t,this.provider=i,this._isResolved=!1}with(e){const t=new uH(this.hint,e.anchor,this.provider);return t._isResolved=this._isResolved,t._currentResolve=this._currentResolve,t}async resolve(e){if(typeof this.provider.resolveInlayHint=="function"){if(this._currentResolve)return await this._currentResolve,e.isCancellationRequested?void 0:this.resolve(e);this._isResolved||(this._currentResolve=this._doResolve(e).finally(()=>this._currentResolve=void 0)),await this._currentResolve}}async _doResolve(e){try{const t=await Promise.resolve(this.provider.resolveInlayHint(this.hint,e));this.hint.tooltip=t?.tooltip??this.hint.tooltip,this.hint.label=t?.label??this.hint.label,this.hint.textEdits=t?.textEdits??this.hint.textEdits,this._isResolved=!0}catch(t){ji(t),this._isResolved=!1}}}const ew=class ew{static async create(e,t,i,n){const o=[],r=e.ordered(t).reverse().map(a=>i.map(async l=>{try{const c=await a.provideInlayHints(t,l,n);(c?.hints.length||a.onDidChangeInlayHints)&&o.push([c??ew._emptyInlayHintList,a])}catch(c){ji(c)}}));if(await Promise.all(r.flat()),n.isCancellationRequested||t.isDisposed())throw new ia;return new ew(i,o,t)}constructor(e,t,i){this._disposables=new U,this.ranges=e,this.provider=new Set;const n=[];for(const[o,r]of t){this._disposables.add(o),this.provider.add(r);for(const a of o.hints){const l=i.validatePosition(a.position);let c="before";const d=ew._getRangeAtPosition(i,l);let h;d.getStartPosition().isBefore(l)?(h=k.fromPositions(d.getStartPosition(),l),c="after"):(h=k.fromPositions(l,d.getEndPosition()),c="before"),n.push(new uH(a,new Xne(h,c),r))}}this.items=n.sort((o,r)=>A.compare(o.hint.position,r.hint.position))}dispose(){this._disposables.dispose()}static _getRangeAtPosition(e,t){const i=t.lineNumber,n=e.getWordAtPosition(t);if(n)return new k(i,n.startColumn,i,n.endColumn);e.tokenization.tokenizeIfCheap(i);const o=e.tokenization.getLineTokens(i),r=t.column-1,a=o.findTokenIndexAtOffset(r);let l=o.getStartOffset(a),c=o.getEndOffset(a);return c-l===1&&(l===r&&a>1?(l=o.getStartOffset(a-1),c=o.getEndOffset(a-1)):c===r&&a<o.getCount()-1&&(l=o.getStartOffset(a+1),c=o.getEndOffset(a+1))),new k(i,l+1,i,c+1)}};ew._emptyInlayHintList=Object.freeze({dispose(){},hints:[]});let QI=ew;function XTe(s){return Rte(s.id,...s.arguments??[]).toString()}function Tl(s){return s&&typeof s.getEditorType=="function"?s.getEditorType()===sx.ICodeEditor:!1}function gH(s){return s&&typeof s.getEditorType=="function"?s.getEditorType()===sx.IDiffEditor:!1}function QTe(s){return!!s&&typeof s=="object"&&typeof s.onDidChangeActiveEditor=="function"}function Qne(s){return Tl(s)?s:gH(s)?s.getModifiedEditor():QTe(s)&&Tl(s.activeCodeEditor)?s.activeCodeEditor:null}var JTe=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},rd=function(s,e){return function(t,i){e(t,i,s)}};let Lh=class extends Am{constructor(e,t,i,n,o,r,a,l,c,d,h,u,g){super(e,{...n.getRawOptions(),overflowWidgetsDomNode:n.getOverflowWidgetsDomNode()},i,o,r,a,l,c,d,h,u,g),this._parentEditor=n,this._overwriteOptions=t,super.updateOptions(this._overwriteOptions),this._register(n.onDidChangeConfiguration(f=>this._onParentConfigurationChanged(f)))}getParentEditor(){return this._parentEditor}_onParentConfigurationChanged(e){super.updateOptions(this._parentEditor.getRawOptions()),super.updateOptions(this._overwriteOptions)}updateOptions(e){GR(this._overwriteOptions,e,!0),super.updateOptions(this._overwriteOptions)}};Lh=JTe([rd(4,_e),rd(5,ht),rd(6,Gt),rd(7,Ae),rd(8,wi),rd(9,Ii),rd(10,Wn),rd(11,ui),rd(12,ge)],Lh);function Jne(s){const e=s.get(ht).getFocusedCodeEditor();return e instanceof Lh?e.getParentEditor():e}const Vj=new q(new G(0,122,204)),eNe={showArrow:!0,showFrame:!0,className:"",frameColor:Vj,arrowColor:Vj,keepEditorSelection:!1},tNe="vs.editor.contrib.zoneWidget";class iNe{constructor(e,t,i,n,o,r,a,l){this.id="",this.domNode=e,this.afterLineNumber=t,this.afterColumn=i,this.heightInLines=n,this.showInHiddenAreas=a,this.ordinal=l,this._onDomNodeTop=o,this._onComputedHeight=r}onDomNodeTop(e){this._onDomNodeTop(e)}onComputedHeight(e){this._onComputedHeight(e)}}class nNe{constructor(e,t){this._id=e,this._domNode=t}getId(){return this._id}getDomNode(){return this._domNode}getPosition(){return null}}const vN=class vN{constructor(e){this._editor=e,this._ruleName=vN._IdGenerator.nextId(),this._color=null,this._height=-1,this._decorations=this._editor.createDecorationsCollection()}dispose(){this.hide(),jP(this._ruleName)}set color(e){this._color!==e&&(this._color=e,this._updateStyle())}set height(e){this._height!==e&&(this._height=e,this._updateStyle())}_updateStyle(){jP(this._ruleName),zE(`.monaco-editor ${this._ruleName}`,`border-style: solid; border-color: transparent; border-bottom-color: ${this._color}; border-width: ${this._height}px; bottom: -${this._height}px !important; margin-left: -${this._height}px; `)}show(e){e.column===1&&(e={lineNumber:e.lineNumber,column:2}),this._decorations.set([{range:k.fromPositions(e),options:{description:"zone-widget-arrow",className:this._ruleName,stickiness:1}}])}hide(){this._decorations.clear()}};vN._IdGenerator=new k7(".arrow-decoration-");let f5=vN;class sNe{constructor(e,t={}){this._arrow=null,this._overlayWidget=null,this._resizeSash=null,this._isSashResizeHeight=!1,this._viewZone=null,this._disposables=new U,this.container=null,this._isShowing=!1,this.editor=e,this._positionMarkerId=this.editor.createDecorationsCollection(),this.options=rc(t),GR(this.options,eNe,!1),this.domNode=document.createElement("div"),this.options.isAccessible||(this.domNode.setAttribute("aria-hidden","true"),this.domNode.setAttribute("role","presentation")),this._disposables.add(this.editor.onDidLayoutChange(i=>{const n=this._getWidth(i);this.domNode.style.width=n+"px",this.domNode.style.left=this._getLeft(i)+"px",this._onWidth(n)}))}dispose(){this._overlayWidget&&(this.editor.removeOverlayWidget(this._overlayWidget),this._overlayWidget=null),this._viewZone&&this.editor.changeViewZones(e=>{this._viewZone&&e.removeZone(this._viewZone.id),this._viewZone=null}),this._positionMarkerId.clear(),this._disposables.dispose()}create(){this.domNode.classList.add("zone-widget"),this.options.className&&this.domNode.classList.add(this.options.className),this.container=document.createElement("div"),this.container.classList.add("zone-widget-container"),this.domNode.appendChild(this.container),this.options.showArrow&&(this._arrow=new f5(this.editor),this._disposables.add(this._arrow)),this._fillContainer(this.container),this._initSash(),this._applyStyles()}style(e){e.frameColor&&(this.options.frameColor=e.frameColor),e.arrowColor&&(this.options.arrowColor=e.arrowColor),this._applyStyles()}_applyStyles(){if(this.container&&this.options.frameColor){const e=this.options.frameColor.toString();this.container.style.borderTopColor=e,this.container.style.borderBottomColor=e}if(this._arrow&&this.options.arrowColor){const e=this.options.arrowColor.toString();this._arrow.color=e}}_getWidth(e){return e.width-e.minimap.minimapWidth-e.verticalScrollbarWidth}_getLeft(e){return e.minimap.minimapWidth>0&&e.minimap.minimapLeft===0?e.minimap.minimapWidth:0}_onViewZoneTop(e){this.domNode.style.top=e+"px"}_onViewZoneHeight(e){if(this.domNode.style.height=`${e}px`,this.container){const t=e-this._decoratingElementsHeight();this.container.style.height=`${t}px`;const i=this.editor.getLayoutInfo();this._doLayout(t,this._getWidth(i))}this._resizeSash?.layout()}get position(){const e=this._positionMarkerId.getRange(0);if(e)return e.getStartPosition()}show(e,t){const i=k.isIRange(e)?k.lift(e):k.fromPositions(e);this._isShowing=!0,this._showImpl(i,t),this._isShowing=!1,this._positionMarkerId.set([{range:i,options:Fe.EMPTY}])}hide(){this._viewZone&&(this.editor.changeViewZones(e=>{this._viewZone&&e.removeZone(this._viewZone.id)}),this._viewZone=null),this._overlayWidget&&(this.editor.removeOverlayWidget(this._overlayWidget),this._overlayWidget=null),this._arrow?.hide(),this._positionMarkerId.clear(),this._isSashResizeHeight=!1}_decoratingElementsHeight(){const e=this.editor.getOption(75);let t=0;if(this.options.showArrow){const i=Math.round(e/3);t+=2*i}if(this.options.showFrame){const i=this.options.frameWidth??Math.round(e/9);t+=2*i}return t}_getMaximumHeightInLines(){return Math.max(12,this.editor.getLayoutInfo().height/this.editor.getOption(75)*.8)}_showImpl(e,t){const i=e.getStartPosition(),n=this.editor.getLayoutInfo(),o=this._getWidth(n);this.domNode.style.width=`${o}px`,this.domNode.style.left=this._getLeft(n)+"px";const r=document.createElement("div");r.style.overflow="hidden";const a=this.editor.getOption(75),l=this._getMaximumHeightInLines();l!==void 0&&(t=Math.min(t,l));let c=0,d=0;if(this._arrow&&this.options.showArrow&&(c=Math.round(a/3),this._arrow.height=c,this._arrow.show(i)),this.options.showFrame&&(d=Math.round(a/9)),this.editor.changeViewZones(g=>{this._viewZone&&g.removeZone(this._viewZone.id),this._overlayWidget&&(this.editor.removeOverlayWidget(this._overlayWidget),this._overlayWidget=null),this.domNode.style.top="-1000px",this._viewZone=new iNe(r,i.lineNumber,i.column,t,f=>this._onViewZoneTop(f),f=>this._onViewZoneHeight(f),this.options.showInHiddenAreas,this.options.ordinal),this._viewZone.id=g.addZone(this._viewZone),this._overlayWidget=new nNe(tNe+this._viewZone.id,this.domNode),this.editor.addOverlayWidget(this._overlayWidget)}),this._updateSashEnablement(),this.container&&this.options.showFrame){const g=this.options.frameWidth?this.options.frameWidth:d;this.container.style.borderTopWidth=g+"px",this.container.style.borderBottomWidth=g+"px"}const h=t*a-this._decoratingElementsHeight();this.container&&(this.container.style.top=c+"px",this.container.style.height=h+"px",this.container.style.overflow="hidden"),this._doLayout(h,o),this.options.keepEditorSelection||this.editor.setSelection(e);const u=this.editor.getModel();if(u){const g=u.validateRange(new k(e.startLineNumber,1,e.endLineNumber+1,1));this.revealRange(g,g.startLineNumber===u.getLineCount())}}revealRange(e,t){t?this.editor.revealLineNearTop(e.endLineNumber,0):this.editor.revealRange(e,0)}setCssClass(e,t){this.container&&(t&&this.container.classList.remove(t),this.container.classList.add(e))}_onWidth(e){}_doLayout(e,t){}_relayout(e,t){const i=this._getMaximumHeightInLines(),n=t&&i!==void 0?Math.min(i,e):e;this._viewZone&&this._viewZone.heightInLines!==n&&(this.editor.changeViewZones(o=>{this._viewZone&&(this._viewZone.heightInLines=n,o.layoutZone(this._viewZone.id))}),this._updateSashEnablement())}_initSash(){if(this._resizeSash)return;this._resizeSash=this._disposables.add(new ls(this.domNode,this,{orientation:1})),this.options.isResizeable||(this._resizeSash.state=0);let e;this._disposables.add(this._resizeSash.onDidStart(t=>{this._viewZone&&(e={startY:t.startY,heightInLines:this._viewZone.heightInLines,...this._getResizeBounds()})})),this._disposables.add(this._resizeSash.onDidEnd(()=>{e=void 0})),this._disposables.add(this._resizeSash.onDidChange(t=>{if(e){const i=(t.currentY-e.startY)/this.editor.getOption(75),n=i<0?Math.ceil(i):Math.floor(i),o=e.heightInLines+n;o>e.minLines&&o<e.maxLines&&(this._isSashResizeHeight=!0,this._relayout(o))}}))}_updateSashEnablement(){if(this._resizeSash){const{minLines:e,maxLines:t}=this._getResizeBounds();this._resizeSash.state=e===t?0:3}}_getResizeBounds(){return{minLines:5,maxLines:35}}getHorizontalSashLeft(){return 0}getHorizontalSashTop(){return(this.domNode.style.height===null?0:parseInt(this.domNode.style.height))-this._decoratingElementsHeight()/2}getHorizontalSashWidth(){const e=this.editor.getLayoutInfo();return e.width-e.minimap.minimapWidth}}var ese=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},tse=function(s,e){return function(t,i){e(t,i,s)}};const ise=qe("IPeekViewService");tt(ise,class{constructor(){this._widgets=new Map}addExclusiveWidget(s,e){const t=this._widgets.get(s);t&&(t.listener.dispose(),t.widget.dispose());const i=()=>{const n=this._widgets.get(s);n&&n.widget===e&&(n.listener.dispose(),this._widgets.delete(s))};this._widgets.set(s,{widget:e,listener:e.onDidClose(i)})}},1);var So;(function(s){s.inPeekEditor=new ce("inReferenceSearchEditor",!0,m(1316,"Whether the current code editor is embedded inside peek")),s.notInPeekEditor=s.inPeekEditor.toNegated()})(So||(So={}));var $w;let JI=($w=class{constructor(e,t){e instanceof Lh&&So.inPeekEditor.bindTo(t)}dispose(){}},$w.ID="editor.contrib.referenceController",$w);JI=ese([tse(1,Ae)],JI);lt(JI.ID,JI,0);const oNe={headerBackgroundColor:q.white,primaryHeadingColor:q.fromHex("#333333"),secondaryHeadingColor:q.fromHex("#6c6c6cb3")};let eD=class extends sNe{constructor(e,t,i){super(e,t),this.instantiationService=i,this._onDidClose=new M,this.onDidClose=this._onDidClose.event,GR(this.options,oNe,!1);const n=di(this.editor);n.openedPeekWidgets.set(n.openedPeekWidgets.get()+1,void 0)}dispose(){if(!this.disposed){this.disposed=!0,super.dispose(),this._onDidClose.fire(this);const e=di(this.editor);e.openedPeekWidgets.set(e.openedPeekWidgets.get()-1,void 0)}}style(e){const t=this.options;e.headerBackgroundColor&&(t.headerBackgroundColor=e.headerBackgroundColor),e.primaryHeadingColor&&(t.primaryHeadingColor=e.primaryHeadingColor),e.secondaryHeadingColor&&(t.secondaryHeadingColor=e.secondaryHeadingColor),super.style(e)}_applyStyles(){super._applyStyles();const e=this.options;this._headElement&&e.headerBackgroundColor&&(this._headElement.style.backgroundColor=e.headerBackgroundColor.toString()),this._primaryHeading&&e.primaryHeadingColor&&(this._primaryHeading.style.color=e.primaryHeadingColor.toString()),this._secondaryHeading&&e.secondaryHeadingColor&&(this._secondaryHeading.style.color=e.secondaryHeadingColor.toString()),this._bodyElement&&e.frameColor&&(this._bodyElement.style.borderColor=e.frameColor.toString())}_fillContainer(e){this.setCssClass("peekview-widget"),this._headElement=ie(".head"),this._bodyElement=ie(".body"),this._fillHead(this._headElement),this._fillBody(this._bodyElement),e.appendChild(this._headElement),e.appendChild(this._bodyElement)}_fillHead(e,t){this._titleElement=ie(".peekview-title"),this.options.supportOnTitleClick&&(this._titleElement.classList.add("clickable"),Fi(this._titleElement,"click",o=>this._onTitleClick(o))),X(this._headElement,this._titleElement),this._fillTitleIcon(this._titleElement),this._primaryHeading=ie("span.filename"),this._secondaryHeading=ie("span.dirname"),this._metaHeading=ie("span.meta"),X(this._titleElement,this._primaryHeading,this._secondaryHeading,this._metaHeading);const i=ie(".peekview-actions");X(this._headElement,i);const n=this._getActionBarOptions();this._actionbarWidget=new bo(i,n),this._disposables.add(this._actionbarWidget),t||this._actionbarWidget.push(this._disposables.add(new Cr("peekview.close",m(1317,"Close"),xe.asClassName(Z.close),!0,()=>(this.dispose(),Promise.resolve()))),{label:!1,icon:!0})}_fillTitleIcon(e){}_getActionBarOptions(){return{actionViewItemProvider:q7.bind(void 0,this.instantiationService),orientation:0}}_onTitleClick(e){}setTitle(e,t){this._primaryHeading&&this._secondaryHeading&&(this._primaryHeading.innerText=e,this._primaryHeading.setAttribute("title",e),t?this._secondaryHeading.innerText=t:Pn(this._secondaryHeading))}setMetaTitle(e){this._metaHeading&&(e?(this._metaHeading.innerText=e,Mo(this._metaHeading)):js(this._metaHeading))}_doLayout(e,t){if(!this._isShowing&&e<0){this.dispose();return}const i=Math.ceil(this.editor.getOption(75)*1.2),n=Math.round(e-(i+1));this._doLayoutHead(i,t),this._doLayoutBody(n,t)}_doLayoutHead(e,t){this._headElement&&(this._headElement.style.height=`${e}px`,this._headElement.style.lineHeight=this._headElement.style.height)}_doLayoutBody(e,t){this._bodyElement&&(this._bodyElement.style.height=`${e}px`)}};eD=ese([tse(2,_e)],eD);const rNe=T("peekViewTitle.background",{dark:"#252526",light:"#F3F3F3",hcDark:q.black,hcLight:q.white},m(1318,"Background color of the peek view title area.")),nse=T("peekViewTitleLabel.foreground",{dark:q.white,light:q.black,hcDark:q.white,hcLight:zc},m(1319,"Color of the peek view title.")),sse=T("peekViewTitleDescription.foreground",{dark:"#ccccccb3",light:"#616161",hcDark:"#FFFFFF99",hcLight:"#292929"},m(1320,"Color of the peek view title info.")),aNe=T("peekView.border",{dark:Rc,light:Rc,hcDark:rt,hcLight:rt},m(1321,"Color of the peek view borders and arrow.")),lNe=T("peekViewResult.background",{dark:"#252526",light:"#F3F3F3",hcDark:q.black,hcLight:q.white},m(1322,"Background color of the peek view result list."));T("peekViewResult.lineForeground",{dark:"#bbbbbb",light:"#646465",hcDark:q.white,hcLight:zc},m(1323,"Foreground color for line nodes in the peek view result list."));T("peekViewResult.fileForeground",{dark:q.white,light:"#1E1E1E",hcDark:q.white,hcLight:zc},m(1324,"Foreground color for file nodes in the peek view result list."));T("peekViewResult.selectionBackground",{dark:"#3399ff33",light:"#3399ff33",hcDark:null,hcLight:null},m(1325,"Background color of the selected entry in the peek view result list."));T("peekViewResult.selectionForeground",{dark:q.white,light:"#6C6C6C",hcDark:q.white,hcLight:zc},m(1326,"Foreground color of the selected entry in the peek view result list."));const fH=T("peekViewEditor.background",{dark:"#001F33",light:"#F2F8FC",hcDark:q.black,hcLight:q.white},m(1327,"Background color of the peek view editor."));T("peekViewEditorGutter.background",fH,m(1328,"Background color of the gutter in the peek view editor."));T("peekViewEditorStickyScroll.background",fH,m(1329,"Background color of sticky scroll in the peek view editor."));T("peekViewEditorStickyScrollGutter.background",fH,m(1330,"Background color of the gutter part of sticky scroll in the peek view editor."));T("peekViewResult.matchHighlightBackground",{dark:"#ea5c004d",light:"#ea5c004d",hcDark:null,hcLight:null},m(1331,"Match highlight color in the peek view result list."));T("peekViewEditor.matchHighlightBackground",{dark:"#ff8f0099",light:"#f5d802de",hcDark:null,hcLight:null},m(1332,"Match highlight color in the peek view editor."));T("peekViewEditor.matchHighlightBorder",{dark:null,light:null,hcDark:ri,hcLight:ri},m(1333,"Match highlight border in the peek view editor."));class cNe{get templateId(){return this.renderer.templateId}constructor(e,t){this.renderer=e,this.modelProvider=t}renderTemplate(e){return{data:this.renderer.renderTemplate(e),disposable:F.None}}renderElement(e,t,i,n){if(i.disposable?.dispose(),!i.data)return;const o=this.modelProvider();if(o.isResolved(e))return this.renderer.renderElement(o.get(e),e,i.data,n);const r=new si,a=o.resolve(e,r.token);i.disposable={dispose:()=>r.cancel()},this.renderer.renderPlaceholder(e,i.data),a.then(l=>this.renderer.renderElement(l,e,i.data,n))}disposeTemplate(e){e.disposable&&(e.disposable.dispose(),e.disposable=void 0),e.data&&(this.renderer.disposeTemplate(e.data),e.data=void 0)}}class dNe{constructor(e,t){this.modelProvider=e,this.accessibilityProvider=t}getWidgetAriaLabel(){return this.accessibilityProvider.getWidgetAriaLabel()}getAriaLabel(e){const t=this.modelProvider();return t.isResolved(e)?this.accessibilityProvider.getAriaLabel(t.get(e)):null}}function hNe(s,e){return{...e,accessibilityProvider:e.accessibilityProvider&&new dNe(s,e.accessibilityProvider)}}class uNe{constructor(e,t,i,n,o={}){this.modelDisposables=new U;const r=()=>this.model,a=n.map(l=>new cNe(l,r));this.list=new Tr(e,t,i,a,hNe(r,o))}updateOptions(e){this.list.updateOptions(e)}getHTMLElement(){return this.list.getHTMLElement()}get onDidFocus(){return this.list.onDidFocus}get widget(){return this.list}get onDidDispose(){return this.list.onDidDispose}get onMouseDblClick(){return oe.map(this.list.onMouseDblClick,({element:e,index:t,browserEvent:i})=>({element:e===void 0?void 0:this._model.get(e),index:t,browserEvent:i}))}get onPointer(){return oe.map(this.list.onPointer,({element:e,index:t,browserEvent:i})=>({element:e===void 0?void 0:this._model.get(e),index:t,browserEvent:i}))}get onDidChangeSelection(){return oe.map(this.list.onDidChangeSelection,({elements:e,indexes:t,browserEvent:i})=>({elements:e.map(n=>this._model.get(n)),indexes:t,browserEvent:i}))}get model(){return this._model}set model(e){this.modelDisposables.clear(),this._model=e,this.list.splice(0,this.list.length,Us(e.length)),this.modelDisposables.add(e.onDidIncrementLength(t=>this.list.splice(this.list.length,0,Us(this.list.length,t))))}getFocus(){return this.list.getFocus()}getSelection(){return this.list.getSelection()}getSelectedElements(){return this.getSelection().map(e=>this.model.get(e))}style(e){this.list.style(e)}dispose(){this.list.dispose(),this.modelDisposables.dispose()}}const gNe={separatorBorder:q.transparent};class ose{set size(e){this._size=e}get size(){return this._size}get visible(){return typeof this._cachedVisibleSize>"u"}setVisible(e,t){if(e!==this.visible){e?(this.size=qs(this._cachedVisibleSize,this.viewMinimumSize,this.viewMaximumSize),this._cachedVisibleSize=void 0):(this._cachedVisibleSize=typeof t=="number"?t:this.size,this.size=0),this.container.classList.toggle("visible",e);try{this.view.setVisible?.(e)}catch(i){console.error("Splitview: Failed to set visible view"),console.error(i)}}}get minimumSize(){return this.visible?this.view.minimumSize:0}get viewMinimumSize(){return this.view.minimumSize}get maximumSize(){return this.visible?this.view.maximumSize:0}get viewMaximumSize(){return this.view.maximumSize}get priority(){return this.view.priority}get proportionalLayout(){return this.view.proportionalLayout??!0}get snap(){return!!this.view.snap}set enabled(e){this.container.style.pointerEvents=e?"":"none"}constructor(e,t,i,n){this.container=e,this.view=t,this.disposable=n,this._cachedVisibleSize=void 0,typeof i=="number"?(this._size=i,this._cachedVisibleSize=void 0,e.classList.add("visible")):(this._size=0,this._cachedVisibleSize=i.cachedVisibleSize)}layout(e,t){this.layoutContainer(e);try{this.view.layout(this.size,e,t)}catch(i){console.error("Splitview: Failed to layout view"),console.error(i)}}dispose(){this.disposable.dispose()}}class fNe extends ose{layoutContainer(e){this.container.style.top=`${e}px`,this.container.style.height=`${this.size}px`}}class pNe extends ose{layoutContainer(e){this.container.style.left=`${e}px`,this.container.style.width=`${this.size}px`}}var vd;(function(s){s[s.Idle=0]="Idle",s[s.Busy=1]="Busy"})(vd||(vd={}));var tD;(function(s){s.Distribute={type:"distribute"};function e(n){return{type:"split",index:n}}s.Split=e;function t(n){return{type:"auto",index:n}}s.Auto=t;function i(n){return{type:"invisible",cachedVisibleSize:n}}s.Invisible=i})(tD||(tD={}));class rse extends F{get orthogonalStartSash(){return this._orthogonalStartSash}get orthogonalEndSash(){return this._orthogonalEndSash}get startSnappingEnabled(){return this._startSnappingEnabled}get endSnappingEnabled(){return this._endSnappingEnabled}set orthogonalStartSash(e){for(const t of this.sashItems)t.sash.orthogonalStartSash=e;this._orthogonalStartSash=e}set orthogonalEndSash(e){for(const t of this.sashItems)t.sash.orthogonalEndSash=e;this._orthogonalEndSash=e}set startSnappingEnabled(e){this._startSnappingEnabled!==e&&(this._startSnappingEnabled=e,this.updateSashEnablement())}set endSnappingEnabled(e){this._endSnappingEnabled!==e&&(this._endSnappingEnabled=e,this.updateSashEnablement())}constructor(e,t={}){super(),this.size=0,this._contentSize=0,this.proportions=void 0,this.viewItems=[],this.sashItems=[],this.state=vd.Idle,this._onDidSashChange=this._register(new M),this._onDidSashReset=this._register(new M),this._startSnappingEnabled=!0,this._endSnappingEnabled=!0,this.onDidSashChange=this._onDidSashChange.event,this.onDidSashReset=this._onDidSashReset.event,this.orientation=t.orientation??0,this.inverseAltBehavior=t.inverseAltBehavior??!1,this.proportionalLayout=t.proportionalLayout??!0,this.getSashOrthogonalSize=t.getSashOrthogonalSize,this.el=document.createElement("div"),this.el.classList.add("monaco-split-view2"),this.el.classList.add(this.orientation===0?"vertical":"horizontal"),e.appendChild(this.el),this.sashContainer=X(this.el,ie(".sash-container")),this.viewContainer=ie(".split-view-container"),this.scrollable=this._register(new sC({forceIntegerValues:!0,smoothScrollDuration:125,scheduleAtNextAnimationFrame:n=>Co(be(this.el),n)})),this.scrollableElement=this._register(new cA(this.viewContainer,{vertical:this.orientation===0?t.scrollbarVisibility??1:2,horizontal:this.orientation===1?t.scrollbarVisibility??1:2},this.scrollable));const i=this._register(new Lt(this.viewContainer,"scroll")).event;this._register(i(n=>{const o=this.scrollableElement.getScrollPosition(),r=Math.abs(this.viewContainer.scrollLeft-o.scrollLeft)<=1?void 0:this.viewContainer.scrollLeft,a=Math.abs(this.viewContainer.scrollTop-o.scrollTop)<=1?void 0:this.viewContainer.scrollTop;(r!==void 0||a!==void 0)&&this.scrollableElement.setScrollPosition({scrollLeft:r,scrollTop:a})})),this.onDidScroll=this.scrollableElement.onScroll,this._register(this.onDidScroll(n=>{n.scrollTopChanged&&(this.viewContainer.scrollTop=n.scrollTop),n.scrollLeftChanged&&(this.viewContainer.scrollLeft=n.scrollLeft)})),X(this.el,this.scrollableElement.getDomNode()),this.style(t.styles||gNe),t.descriptor&&(this.size=t.descriptor.size,t.descriptor.views.forEach((n,o)=>{const r=fs(n.visible)||n.visible?n.size:{type:"invisible",cachedVisibleSize:n.size},a=n.view;this.doAddView(a,r,o,!0)}),this._contentSize=this.viewItems.reduce((n,o)=>n+o.size,0),this.saveProportions())}style(e){e.separatorBorder.isTransparent()?(this.el.classList.remove("separator-border"),this.el.style.removeProperty("--separator-border")):(this.el.classList.add("separator-border"),this.el.style.setProperty("--separator-border",e.separatorBorder.toString()))}addView(e,t,i=this.viewItems.length,n){this.doAddView(e,t,i,n)}layout(e,t){const i=Math.max(this.size,this._contentSize);if(this.size=e,this.layoutContext=t,this.proportions){let n=0;for(let o=0;o<this.viewItems.length;o++){const r=this.viewItems[o],a=this.proportions[o];typeof a=="number"?n+=a:e-=r.size}for(let o=0;o<this.viewItems.length;o++){const r=this.viewItems[o],a=this.proportions[o];typeof a=="number"&&n>0&&(r.size=qs(Math.round(a*e/n),r.minimumSize,r.maximumSize))}}else{const n=Us(this.viewItems.length),o=n.filter(a=>this.viewItems[a].priority===1),r=n.filter(a=>this.viewItems[a].priority===2);this.resize(this.viewItems.length-1,e-i,void 0,o,r)}this.distributeEmptySpace(),this.layoutViews()}saveProportions(){this.proportionalLayout&&this._contentSize>0&&(this.proportions=this.viewItems.map(e=>e.proportionalLayout&&e.visible?e.size/this._contentSize:void 0))}onSashStart({sash:e,start:t,alt:i}){for(const a of this.viewItems)a.enabled=!1;const n=this.sashItems.findIndex(a=>a.sash===e),o=Ra(W(this.el.ownerDocument.body,"keydown",a=>r(this.sashDragState.current,a.altKey)),W(this.el.ownerDocument.body,"keyup",()=>r(this.sashDragState.current,!1))),r=(a,l)=>{const c=this.viewItems.map(f=>f.size);let d=Number.NEGATIVE_INFINITY,h=Number.POSITIVE_INFINITY;if(this.inverseAltBehavior&&(l=!l),l)if(n===this.sashItems.length-1){const p=this.viewItems[n];d=(p.minimumSize-p.size)/2,h=(p.maximumSize-p.size)/2}else{const p=this.viewItems[n+1];d=(p.size-p.maximumSize)/2,h=(p.size-p.minimumSize)/2}let u,g;if(!l){const f=Us(n,-1),p=Us(n+1,this.viewItems.length),_=f.reduce((E,I)=>E+(this.viewItems[I].minimumSize-c[I]),0),b=f.reduce((E,I)=>E+(this.viewItems[I].viewMaximumSize-c[I]),0),w=p.length===0?Number.POSITIVE_INFINITY:p.reduce((E,I)=>E+(c[I]-this.viewItems[I].minimumSize),0),C=p.length===0?Number.NEGATIVE_INFINITY:p.reduce((E,I)=>E+(c[I]-this.viewItems[I].viewMaximumSize),0),v=Math.max(_,C),y=Math.min(w,b),x=this.findFirstSnapIndex(f),L=this.findFirstSnapIndex(p);if(typeof x=="number"){const E=this.viewItems[x],I=Math.floor(E.viewMinimumSize/2);u={index:x,limitDelta:E.visible?v-I:v+I,size:E.size}}if(typeof L=="number"){const E=this.viewItems[L],I=Math.floor(E.viewMinimumSize/2);g={index:L,limitDelta:E.visible?y+I:y-I,size:E.size}}}this.sashDragState={start:a,current:a,index:n,sizes:c,minDelta:d,maxDelta:h,alt:l,snapBefore:u,snapAfter:g,disposable:o}};r(t,i)}onSashChange({current:e}){const{index:t,start:i,sizes:n,alt:o,minDelta:r,maxDelta:a,snapBefore:l,snapAfter:c}=this.sashDragState;this.sashDragState.current=e;const d=e-i,h=this.resize(t,d,n,void 0,void 0,r,a,l,c);if(o){const u=t===this.sashItems.length-1,g=this.viewItems.map(C=>C.size),f=u?t:t+1,p=this.viewItems[f],_=p.size-p.maximumSize,b=p.size-p.minimumSize,w=u?t-1:t+1;this.resize(w,-h,g,void 0,void 0,_,b)}this.distributeEmptySpace(),this.layoutViews()}onSashEnd(e){this._onDidSashChange.fire(e),this.sashDragState.disposable.dispose(),this.saveProportions();for(const t of this.viewItems)t.enabled=!0}onViewChange(e,t){const i=this.viewItems.indexOf(e);i<0||i>=this.viewItems.length||(t=typeof t=="number"?t:e.size,t=qs(t,e.minimumSize,e.maximumSize),this.inverseAltBehavior&&i>0?(this.resize(i-1,Math.floor((e.size-t)/2)),this.distributeEmptySpace(),this.layoutViews()):(e.size=t,this.relayout([i],void 0)))}resizeView(e,t){if(!(e<0||e>=this.viewItems.length)){if(this.state!==vd.Idle)throw new Error("Cant modify splitview");this.state=vd.Busy;try{const i=Us(this.viewItems.length).filter(a=>a!==e),n=[...i.filter(a=>this.viewItems[a].priority===1),e],o=i.filter(a=>this.viewItems[a].priority===2),r=this.viewItems[e];t=Math.round(t),t=qs(t,r.minimumSize,Math.min(r.maximumSize,this.size)),r.size=t,this.relayout(n,o)}finally{this.state=vd.Idle}}}distributeViewSizes(){const e=[];let t=0;for(const a of this.viewItems)a.maximumSize-a.minimumSize>0&&(e.push(a),t+=a.size);const i=Math.floor(t/e.length);for(const a of e)a.size=qs(i,a.minimumSize,a.maximumSize);const n=Us(this.viewItems.length),o=n.filter(a=>this.viewItems[a].priority===1),r=n.filter(a=>this.viewItems[a].priority===2);this.relayout(o,r)}getViewSize(e){return e<0||e>=this.viewItems.length?-1:this.viewItems[e].size}doAddView(e,t,i=this.viewItems.length,n){if(this.state!==vd.Idle)throw new Error("Cant modify splitview");this.state=vd.Busy;try{const o=ie(".split-view-view");i===this.viewItems.length?this.viewContainer.appendChild(o):this.viewContainer.insertBefore(o,this.viewContainer.children.item(i));const r=e.onDidChange(u=>this.onViewChange(d,u)),a=pe(()=>o.remove()),l=Ra(r,a);let c;typeof t=="number"?c=t:(t.type==="auto"&&(this.areViewsDistributed()?t={type:"distribute"}:t={type:"split",index:t.index}),t.type==="split"?c=this.getViewSize(t.index)/2:t.type==="invisible"?c={cachedVisibleSize:t.cachedVisibleSize}:c=e.minimumSize);const d=this.orientation===0?new fNe(o,e,c,l):new pNe(o,e,c,l);if(this.viewItems.splice(i,0,d),this.viewItems.length>1){const u={orthogonalStartSash:this.orthogonalStartSash,orthogonalEndSash:this.orthogonalEndSash},g=this.orientation===0?new ls(this.sashContainer,{getHorizontalSashTop:E=>this.getSashPosition(E),getHorizontalSashWidth:this.getSashOrthogonalSize},{...u,orientation:1}):new ls(this.sashContainer,{getVerticalSashLeft:E=>this.getSashPosition(E),getVerticalSashHeight:this.getSashOrthogonalSize},{...u,orientation:0}),f=this.orientation===0?E=>({sash:g,start:E.startY,current:E.currentY,alt:E.altKey}):E=>({sash:g,start:E.startX,current:E.currentX,alt:E.altKey}),_=oe.map(g.onDidStart,f)(this.onSashStart,this),w=oe.map(g.onDidChange,f)(this.onSashChange,this),v=oe.map(g.onDidEnd,()=>this.sashItems.findIndex(E=>E.sash===g))(this.onSashEnd,this),y=g.onDidReset(()=>{const E=this.sashItems.findIndex(B=>B.sash===g),I=Us(E,-1),R=Us(E+1,this.viewItems.length),O=this.findFirstSnapIndex(I),P=this.findFirstSnapIndex(R);typeof O=="number"&&!this.viewItems[O].visible||typeof P=="number"&&!this.viewItems[P].visible||this._onDidSashReset.fire(E)}),x=Ra(_,w,v,y,g),L={sash:g,disposable:x};this.sashItems.splice(i-1,0,L)}o.appendChild(e.element);let h;typeof t!="number"&&t.type==="split"&&(h=[t.index]),n||this.relayout([i],h),!n&&typeof t!="number"&&t.type==="distribute"&&this.distributeViewSizes()}finally{this.state=vd.Idle}}relayout(e,t){const i=this.viewItems.reduce((n,o)=>n+o.size,0);this.resize(this.viewItems.length-1,this.size-i,void 0,e,t),this.distributeEmptySpace(),this.layoutViews(),this.saveProportions()}resize(e,t,i=this.viewItems.map(d=>d.size),n,o,r=Number.NEGATIVE_INFINITY,a=Number.POSITIVE_INFINITY,l,c){if(e<0||e>=this.viewItems.length)return 0;const d=Us(e,-1),h=Us(e+1,this.viewItems.length);if(o)for(const L of o)F2(d,L),F2(h,L);if(n)for(const L of n)jx(d,L),jx(h,L);const u=d.map(L=>this.viewItems[L]),g=d.map(L=>i[L]),f=h.map(L=>this.viewItems[L]),p=h.map(L=>i[L]),_=d.reduce((L,E)=>L+(this.viewItems[E].minimumSize-i[E]),0),b=d.reduce((L,E)=>L+(this.viewItems[E].maximumSize-i[E]),0),w=h.length===0?Number.POSITIVE_INFINITY:h.reduce((L,E)=>L+(i[E]-this.viewItems[E].minimumSize),0),C=h.length===0?Number.NEGATIVE_INFINITY:h.reduce((L,E)=>L+(i[E]-this.viewItems[E].maximumSize),0),v=Math.max(_,C,r),y=Math.min(w,b,a);let x=!1;if(l){const L=this.viewItems[l.index],E=t>=l.limitDelta;x=E!==L.visible,L.setVisible(E,l.size)}if(!x&&c){const L=this.viewItems[c.index],E=t<c.limitDelta;x=E!==L.visible,L.setVisible(E,c.size)}if(x)return this.resize(e,t,i,n,o,r,a);t=qs(t,v,y);for(let L=0,E=t;L<u.length;L++){const I=u[L],R=qs(g[L]+E,I.minimumSize,I.maximumSize),O=R-g[L];E-=O,I.size=R}for(let L=0,E=t;L<f.length;L++){const I=f[L],R=qs(p[L]-E,I.minimumSize,I.maximumSize),O=R-p[L];E+=O,I.size=R}return t}distributeEmptySpace(e){const t=this.viewItems.reduce((a,l)=>a+l.size,0);let i=this.size-t;const n=Us(this.viewItems.length-1,-1),o=n.filter(a=>this.viewItems[a].priority===1),r=n.filter(a=>this.viewItems[a].priority===2);for(const a of r)F2(n,a);for(const a of o)jx(n,a);typeof e=="number"&&jx(n,e);for(let a=0;i!==0&&a<n.length;a++){const l=this.viewItems[n[a]],c=qs(l.size+i,l.minimumSize,l.maximumSize),d=c-l.size;i-=d,l.size=c}}layoutViews(){this._contentSize=this.viewItems.reduce((t,i)=>t+i.size,0);let e=0;for(const t of this.viewItems)t.layout(e,this.layoutContext),e+=t.size;this.sashItems.forEach(t=>t.sash.layout()),this.updateSashEnablement(),this.updateScrollableElement()}updateScrollableElement(){this.orientation===0?this.scrollableElement.setScrollDimensions({height:this.size,scrollHeight:this._contentSize}):this.scrollableElement.setScrollDimensions({width:this.size,scrollWidth:this._contentSize})}updateSashEnablement(){let e=!1;const t=this.viewItems.map(l=>e=l.size-l.minimumSize>0||e);e=!1;const i=this.viewItems.map(l=>e=l.maximumSize-l.size>0||e),n=[...this.viewItems].reverse();e=!1;const o=n.map(l=>e=l.size-l.minimumSize>0||e).reverse();e=!1;const r=n.map(l=>e=l.maximumSize-l.size>0||e).reverse();let a=0;for(let l=0;l<this.sashItems.length;l++){const{sash:c}=this.sashItems[l],d=this.viewItems[l];a+=d.size;const h=!(t[l]&&r[l+1]),u=!(i[l]&&o[l+1]);if(h&&u){const g=Us(l,-1),f=Us(l+1,this.viewItems.length),p=this.findFirstSnapIndex(g),_=this.findFirstSnapIndex(f),b=typeof p=="number"&&!this.viewItems[p].visible,w=typeof _=="number"&&!this.viewItems[_].visible;b&&o[l]&&(a>0||this.startSnappingEnabled)?c.state=1:w&&t[l]&&(a<this._contentSize||this.endSnappingEnabled)?c.state=2:c.state=0}else h&&!u?c.state=1:!h&&u?c.state=2:c.state=3}}getSashPosition(e){let t=0;for(let i=0;i<this.sashItems.length;i++)if(t+=this.viewItems[i].size,this.sashItems[i].sash===e)return t;return 0}findFirstSnapIndex(e){for(const t of e){const i=this.viewItems[t];if(i.visible&&i.snap)return t}for(const t of e){const i=this.viewItems[t];if(i.visible&&i.maximumSize-i.minimumSize>0)return;if(!i.visible&&i.snap)return t}}areViewsDistributed(){let e,t;for(const i of this.viewItems)if(e=e===void 0?i.size:Math.min(e,i.size),t=t===void 0?i.size:Math.max(t,i.size),t-e>2)return!1;return!0}dispose(){this.sashDragState?.disposable.dispose(),xt(this.viewItems),this.viewItems=[],this.sashItems.forEach(e=>e.disposable.dispose()),this.sashItems=[],super.dispose()}}const CN=class CN{constructor(e,t,i){this.columns=e,this.getColumnSize=i,this.templateId=CN.TemplateId,this.renderedTemplates=new Set;const n=new Map(t.map(o=>[o.templateId,o]));this.renderers=[];for(const o of e){const r=n.get(o.templateId);if(!r)throw new Error(`Table cell renderer for template id ${o.templateId} not found.`);this.renderers.push(r)}}renderTemplate(e){const t=X(e,ie(".monaco-table-tr")),i=[],n=[];for(let r=0;r<this.columns.length;r++){const a=this.renderers[r],l=X(t,ie(".monaco-table-td",{"data-col-index":r}));l.style.width=`${this.getColumnSize(r)}px`,i.push(l),n.push(a.renderTemplate(l))}const o={container:e,cellContainers:i,cellTemplateData:n};return this.renderedTemplates.add(o),o}renderElement(e,t,i,n){for(let o=0;o<this.columns.length;o++){const a=this.columns[o].project(e);this.renderers[o].renderElement(a,t,i.cellTemplateData[o],n)}}disposeElement(e,t,i,n){for(let o=0;o<this.columns.length;o++){const r=this.renderers[o];if(r.disposeElement){const l=this.columns[o].project(e);r.disposeElement(l,t,i.cellTemplateData[o],n)}}}disposeTemplate(e){for(let t=0;t<this.columns.length;t++)this.renderers[t].disposeTemplate(e.cellTemplateData[t]);Pn(e.container),this.renderedTemplates.delete(e)}layoutColumn(e,t){for(const{cellContainers:i}of this.renderedTemplates)i[e].style.width=`${t}px`}};CN.TemplateId="row";let iD=CN;function mNe(s){return{getHeight(e){return s.getHeight(e)},getTemplateId(){return iD.TemplateId}}}class _Ne extends F{get minimumSize(){return this.column.minimumWidth??120}get maximumSize(){return this.column.maximumWidth??Number.POSITIVE_INFINITY}get onDidChange(){return this.column.onDidChangeWidthConstraints??oe.None}constructor(e,t){super(),this.column=e,this.index=t,this._onDidLayout=new M,this.onDidLayout=this._onDidLayout.event,this.element=ie(".monaco-table-th",{"data-col-index":t},e.label),e.tooltip&&this._register($a().setupManagedHover(Uc("mouse"),this.element,e.tooltip))}layout(e){this._onDidLayout.fire([this.index,e])}}const yN=class yN{get onDidChangeFocus(){return this.list.onDidChangeFocus}get onDidChangeSelection(){return this.list.onDidChangeSelection}get onDidScroll(){return this.list.onDidScroll}get onMouseDblClick(){return this.list.onMouseDblClick}get onPointer(){return this.list.onPointer}get onDidFocus(){return this.list.onDidFocus}get scrollTop(){return this.list.scrollTop}set scrollTop(e){this.list.scrollTop=e}get scrollHeight(){return this.list.scrollHeight}get renderHeight(){return this.list.renderHeight}get onDidDispose(){return this.list.onDidDispose}constructor(e,t,i,n,o,r){this.virtualDelegate=i,this.columns=n,this.domId=`table_id_${++yN.InstanceCount}`,this.disposables=new U,this.cachedWidth=0,this.cachedHeight=0,this.domNode=X(t,ie(`.monaco-table.${this.domId}`));const a=n.map((d,h)=>this.disposables.add(new _Ne(d,h))),l={size:a.reduce((d,h)=>d+h.column.weight,0),views:a.map(d=>({size:d.column.weight,view:d}))};this.splitview=this.disposables.add(new rse(this.domNode,{orientation:1,scrollbarVisibility:2,getSashOrthogonalSize:()=>this.cachedHeight,descriptor:l})),this.splitview.el.style.height=`${i.headerRowHeight}px`,this.splitview.el.style.lineHeight=`${i.headerRowHeight}px`;const c=new iD(n,o,d=>this.splitview.getViewSize(d));this.list=this.disposables.add(new Tr(e,this.domNode,mNe(i),[c],r)),oe.any(...a.map(d=>d.onDidLayout))(([d,h])=>c.layoutColumn(d,h),null,this.disposables),this.splitview.onDidSashReset(d=>{const h=n.reduce((g,f)=>g+f.weight,0),u=n[d].weight/h*this.cachedWidth;this.splitview.resizeView(d,u)},null,this.disposables),this.styleElement=ca(this.domNode),this.style(LLe)}updateOptions(e){this.list.updateOptions(e)}splice(e,t,i=[]){this.list.splice(e,t,i)}getHTMLElement(){return this.domNode}style(e){const t=[];t.push(`.monaco-table.${this.domId} > .monaco-split-view2 .monaco-sash.vertical::before {
			top: ${this.virtualDelegate.headerRowHeight+1}px;
			height: calc(100% - ${this.virtualDelegate.headerRowHeight}px);
		}`),this.styleElement.textContent=t.join(`
`),this.list.style(e)}getSelectedElements(){return this.list.getSelectedElements()}getSelection(){return this.list.getSelection()}getFocus(){return this.list.getFocus()}dispose(){this.disposables.dispose()}};yN.InstanceCount=0;let p5=yN;var ur;(function(s){s[s.Expanded=0]="Expanded",s[s.Collapsed=1]="Collapsed",s[s.PreserveOrExpanded=2]="PreserveOrExpanded",s[s.PreserveOrCollapsed=3]="PreserveOrCollapsed"})(ur||(ur={}));var Jf;(function(s){s[s.Unknown=0]="Unknown",s[s.Twistie=1]="Twistie",s[s.Element=2]="Element",s[s.Filter=3]="Filter"})(Jf||(Jf={}));class mr extends Error{constructor(e,t){super(`TreeError [${e}] ${t}`)}}class pH{constructor(e){this.fn=e,this._map=new WeakMap}map(e){let t=this._map.get(e);return t||(t=this.fn(e),this._map.set(e,t)),t}}class qh{constructor(e,t,i,n){this.originalStart=e,this.originalLength=t,this.modifiedStart=i,this.modifiedLength=n}getOriginalEnd(){return this.originalStart+this.originalLength}getModifiedEnd(){return this.modifiedStart+this.modifiedLength}}class zj{constructor(e){this.source=e}getElements(){const e=this.source,t=new Int32Array(e.length);for(let i=0,n=e.length;i<n;i++)t[i]=e.charCodeAt(i);return t}}function bNe(s,e,t){return new fc(new zj(s),new zj(e)).ComputeDiff(t).changes}class G_{static Assert(e,t){if(!e)throw new Error(t)}}class K_{static Copy(e,t,i,n,o){for(let r=0;r<o;r++)i[n+r]=e[t+r]}static Copy2(e,t,i,n,o){for(let r=0;r<o;r++)i[n+r]=e[t+r]}}class $j{constructor(){this.m_changes=[],this.m_originalStart=1073741824,this.m_modifiedStart=1073741824,this.m_originalCount=0,this.m_modifiedCount=0}MarkNextChange(){(this.m_originalCount>0||this.m_modifiedCount>0)&&this.m_changes.push(new qh(this.m_originalStart,this.m_originalCount,this.m_modifiedStart,this.m_modifiedCount)),this.m_originalCount=0,this.m_modifiedCount=0,this.m_originalStart=1073741824,this.m_modifiedStart=1073741824}AddOriginalElement(e,t){this.m_originalStart=Math.min(this.m_originalStart,e),this.m_modifiedStart=Math.min(this.m_modifiedStart,t),this.m_originalCount++}AddModifiedElement(e,t){this.m_originalStart=Math.min(this.m_originalStart,e),this.m_modifiedStart=Math.min(this.m_modifiedStart,t),this.m_modifiedCount++}getChanges(){return(this.m_originalCount>0||this.m_modifiedCount>0)&&this.MarkNextChange(),this.m_changes}getReverseChanges(){return(this.m_originalCount>0||this.m_modifiedCount>0)&&this.MarkNextChange(),this.m_changes.reverse(),this.m_changes}}class fc{constructor(e,t,i=null){this.ContinueProcessingPredicate=i,this._originalSequence=e,this._modifiedSequence=t;const[n,o,r]=fc._getElements(e),[a,l,c]=fc._getElements(t);this._hasStrings=r&&c,this._originalStringElements=n,this._originalElementsOrHash=o,this._modifiedStringElements=a,this._modifiedElementsOrHash=l,this.m_forwardHistory=[],this.m_reverseHistory=[]}static _isStringArray(e){return e.length>0&&typeof e[0]=="string"}static _getElements(e){const t=e.getElements();if(fc._isStringArray(t)){const i=new Int32Array(t.length);for(let n=0,o=t.length;n<o;n++)i[n]=Q6(t[n],0);return[t,i,!0]}return t instanceof Int32Array?[[],t,!1]:[[],new Int32Array(t),!1]}ElementsAreEqual(e,t){return this._originalElementsOrHash[e]!==this._modifiedElementsOrHash[t]?!1:this._hasStrings?this._originalStringElements[e]===this._modifiedStringElements[t]:!0}ElementsAreStrictEqual(e,t){if(!this.ElementsAreEqual(e,t))return!1;const i=fc._getStrictElement(this._originalSequence,e),n=fc._getStrictElement(this._modifiedSequence,t);return i===n}static _getStrictElement(e,t){return typeof e.getStrictElement=="function"?e.getStrictElement(t):null}OriginalElementsAreEqual(e,t){return this._originalElementsOrHash[e]!==this._originalElementsOrHash[t]?!1:this._hasStrings?this._originalStringElements[e]===this._originalStringElements[t]:!0}ModifiedElementsAreEqual(e,t){return this._modifiedElementsOrHash[e]!==this._modifiedElementsOrHash[t]?!1:this._hasStrings?this._modifiedStringElements[e]===this._modifiedStringElements[t]:!0}ComputeDiff(e){return this._ComputeDiff(0,this._originalElementsOrHash.length-1,0,this._modifiedElementsOrHash.length-1,e)}_ComputeDiff(e,t,i,n,o){const r=[!1];let a=this.ComputeDiffRecursive(e,t,i,n,r);return o&&(a=this.PrettifyChanges(a)),{quitEarly:r[0],changes:a}}ComputeDiffRecursive(e,t,i,n,o){for(o[0]=!1;e<=t&&i<=n&&this.ElementsAreEqual(e,i);)e++,i++;for(;t>=e&&n>=i&&this.ElementsAreEqual(t,n);)t--,n--;if(e>t||i>n){let h;return i<=n?(G_.Assert(e===t+1,"originalStart should only be one more than originalEnd"),h=[new qh(e,0,i,n-i+1)]):e<=t?(G_.Assert(i===n+1,"modifiedStart should only be one more than modifiedEnd"),h=[new qh(e,t-e+1,i,0)]):(G_.Assert(e===t+1,"originalStart should only be one more than originalEnd"),G_.Assert(i===n+1,"modifiedStart should only be one more than modifiedEnd"),h=[]),h}const r=[0],a=[0],l=this.ComputeRecursionPoint(e,t,i,n,r,a,o),c=r[0],d=a[0];if(l!==null)return l;if(!o[0]){const h=this.ComputeDiffRecursive(e,c,i,d,o);let u=[];return o[0]?u=[new qh(c+1,t-(c+1)+1,d+1,n-(d+1)+1)]:u=this.ComputeDiffRecursive(c+1,t,d+1,n,o),this.ConcatenateChanges(h,u)}return[new qh(e,t-e+1,i,n-i+1)]}WALKTRACE(e,t,i,n,o,r,a,l,c,d,h,u,g,f,p,_,b,w){let C=null,v=null,y=new $j,x=t,L=i,E=g[0]-_[0]-n,I=-1073741824,R=this.m_forwardHistory.length-1;do{const O=E+e;O===x||O<L&&c[O-1]<c[O+1]?(h=c[O+1],f=h-E-n,h<I&&y.MarkNextChange(),I=h,y.AddModifiedElement(h+1,f),E=O+1-e):(h=c[O-1]+1,f=h-E-n,h<I&&y.MarkNextChange(),I=h-1,y.AddOriginalElement(h,f+1),E=O-1-e),R>=0&&(c=this.m_forwardHistory[R],e=c[0],x=1,L=c.length-1)}while(--R>=-1);if(C=y.getReverseChanges(),w[0]){let O=g[0]+1,P=_[0]+1;if(C!==null&&C.length>0){const B=C[C.length-1];O=Math.max(O,B.getOriginalEnd()),P=Math.max(P,B.getModifiedEnd())}v=[new qh(O,u-O+1,P,p-P+1)]}else{y=new $j,x=r,L=a,E=g[0]-_[0]-l,I=1073741824,R=b?this.m_reverseHistory.length-1:this.m_reverseHistory.length-2;do{const O=E+o;O===x||O<L&&d[O-1]>=d[O+1]?(h=d[O+1]-1,f=h-E-l,h>I&&y.MarkNextChange(),I=h+1,y.AddOriginalElement(h+1,f+1),E=O+1-o):(h=d[O-1],f=h-E-l,h>I&&y.MarkNextChange(),I=h,y.AddModifiedElement(h+1,f+1),E=O-1-o),R>=0&&(d=this.m_reverseHistory[R],o=d[0],x=1,L=d.length-1)}while(--R>=-1);v=y.getChanges()}return this.ConcatenateChanges(C,v)}ComputeRecursionPoint(e,t,i,n,o,r,a){let l=0,c=0,d=0,h=0,u=0,g=0;e--,i--,o[0]=0,r[0]=0,this.m_forwardHistory=[],this.m_reverseHistory=[];const f=t-e+(n-i),p=f+1,_=new Int32Array(p),b=new Int32Array(p),w=n-i,C=t-e,v=e-i,y=t-n,L=(C-w)%2===0;_[w]=e,b[C]=t,a[0]=!1;for(let E=1;E<=f/2+1;E++){let I=0,R=0;d=this.ClipDiagonalBound(w-E,E,w,p),h=this.ClipDiagonalBound(w+E,E,w,p);for(let P=d;P<=h;P+=2){P===d||P<h&&_[P-1]<_[P+1]?l=_[P+1]:l=_[P-1]+1,c=l-(P-w)-v;const B=l;for(;l<t&&c<n&&this.ElementsAreEqual(l+1,c+1);)l++,c++;if(_[P]=l,l+c>I+R&&(I=l,R=c),!L&&Math.abs(P-C)<=E-1&&l>=b[P])return o[0]=l,r[0]=c,B<=b[P]&&E<=1448?this.WALKTRACE(w,d,h,v,C,u,g,y,_,b,l,t,o,c,n,r,L,a):null}const O=(I-e+(R-i)-E)/2;if(this.ContinueProcessingPredicate!==null&&!this.ContinueProcessingPredicate(I,O))return a[0]=!0,o[0]=I,r[0]=R,O>0&&E<=1448?this.WALKTRACE(w,d,h,v,C,u,g,y,_,b,l,t,o,c,n,r,L,a):(e++,i++,[new qh(e,t-e+1,i,n-i+1)]);u=this.ClipDiagonalBound(C-E,E,C,p),g=this.ClipDiagonalBound(C+E,E,C,p);for(let P=u;P<=g;P+=2){P===u||P<g&&b[P-1]>=b[P+1]?l=b[P+1]-1:l=b[P-1],c=l-(P-C)-y;const B=l;for(;l>e&&c>i&&this.ElementsAreEqual(l,c);)l--,c--;if(b[P]=l,L&&Math.abs(P-w)<=E&&l<=_[P])return o[0]=l,r[0]=c,B>=_[P]&&E<=1448?this.WALKTRACE(w,d,h,v,C,u,g,y,_,b,l,t,o,c,n,r,L,a):null}if(E<=1447){let P=new Int32Array(h-d+2);P[0]=w-d+1,K_.Copy2(_,d,P,1,h-d+1),this.m_forwardHistory.push(P),P=new Int32Array(g-u+2),P[0]=C-u+1,K_.Copy2(b,u,P,1,g-u+1),this.m_reverseHistory.push(P)}}return this.WALKTRACE(w,d,h,v,C,u,g,y,_,b,l,t,o,c,n,r,L,a)}PrettifyChanges(e){for(let t=0;t<e.length;t++){const i=e[t],n=t<e.length-1?e[t+1].originalStart:this._originalElementsOrHash.length,o=t<e.length-1?e[t+1].modifiedStart:this._modifiedElementsOrHash.length,r=i.originalLength>0,a=i.modifiedLength>0;for(;i.originalStart+i.originalLength<n&&i.modifiedStart+i.modifiedLength<o&&(!r||this.OriginalElementsAreEqual(i.originalStart,i.originalStart+i.originalLength))&&(!a||this.ModifiedElementsAreEqual(i.modifiedStart,i.modifiedStart+i.modifiedLength));){const c=this.ElementsAreStrictEqual(i.originalStart,i.modifiedStart);if(this.ElementsAreStrictEqual(i.originalStart+i.originalLength,i.modifiedStart+i.modifiedLength)&&!c)break;i.originalStart++,i.modifiedStart++}const l=[null];if(t<e.length-1&&this.ChangesOverlap(e[t],e[t+1],l)){e[t]=l[0],e.splice(t+1,1),t--;continue}}for(let t=e.length-1;t>=0;t--){const i=e[t];let n=0,o=0;if(t>0){const h=e[t-1];n=h.originalStart+h.originalLength,o=h.modifiedStart+h.modifiedLength}const r=i.originalLength>0,a=i.modifiedLength>0;let l=0,c=this._boundaryScore(i.originalStart,i.originalLength,i.modifiedStart,i.modifiedLength);for(let h=1;;h++){const u=i.originalStart-h,g=i.modifiedStart-h;if(u<n||g<o||r&&!this.OriginalElementsAreEqual(u,u+i.originalLength)||a&&!this.ModifiedElementsAreEqual(g,g+i.modifiedLength))break;const p=(u===n&&g===o?5:0)+this._boundaryScore(u,i.originalLength,g,i.modifiedLength);p>c&&(c=p,l=h)}i.originalStart-=l,i.modifiedStart-=l;const d=[null];if(t>0&&this.ChangesOverlap(e[t-1],e[t],d)){e[t-1]=d[0],e.splice(t,1),t++;continue}}if(this._hasStrings)for(let t=1,i=e.length;t<i;t++){const n=e[t-1],o=e[t],r=o.originalStart-n.originalStart-n.originalLength,a=n.originalStart,l=o.originalStart+o.originalLength,c=l-a,d=n.modifiedStart,h=o.modifiedStart+o.modifiedLength,u=h-d;if(r<5&&c<20&&u<20){const g=this._findBetterContiguousSequence(a,c,d,u,r);if(g){const[f,p]=g;(f!==n.originalStart+n.originalLength||p!==n.modifiedStart+n.modifiedLength)&&(n.originalLength=f-n.originalStart,n.modifiedLength=p-n.modifiedStart,o.originalStart=f+r,o.modifiedStart=p+r,o.originalLength=l-o.originalStart,o.modifiedLength=h-o.modifiedStart)}}}return e}_findBetterContiguousSequence(e,t,i,n,o){if(t<o||n<o)return null;const r=e+t-o+1,a=i+n-o+1;let l=0,c=0,d=0;for(let h=e;h<r;h++)for(let u=i;u<a;u++){const g=this._contiguousSequenceScore(h,u,o);g>0&&g>l&&(l=g,c=h,d=u)}return l>0?[c,d]:null}_contiguousSequenceScore(e,t,i){let n=0;for(let o=0;o<i;o++){if(!this.ElementsAreEqual(e+o,t+o))return 0;n+=this._originalStringElements[e+o].length}return n}_OriginalIsBoundary(e){return e<=0||e>=this._originalElementsOrHash.length-1?!0:this._hasStrings&&/^\s*$/.test(this._originalStringElements[e])}_OriginalRegionIsBoundary(e,t){if(this._OriginalIsBoundary(e)||this._OriginalIsBoundary(e-1))return!0;if(t>0){const i=e+t;if(this._OriginalIsBoundary(i-1)||this._OriginalIsBoundary(i))return!0}return!1}_ModifiedIsBoundary(e){return e<=0||e>=this._modifiedElementsOrHash.length-1?!0:this._hasStrings&&/^\s*$/.test(this._modifiedStringElements[e])}_ModifiedRegionIsBoundary(e,t){if(this._ModifiedIsBoundary(e)||this._ModifiedIsBoundary(e-1))return!0;if(t>0){const i=e+t;if(this._ModifiedIsBoundary(i-1)||this._ModifiedIsBoundary(i))return!0}return!1}_boundaryScore(e,t,i,n){const o=this._OriginalRegionIsBoundary(e,t)?1:0,r=this._ModifiedRegionIsBoundary(i,n)?1:0;return o+r}ConcatenateChanges(e,t){const i=[];if(e.length===0||t.length===0)return t.length>0?t:e;if(this.ChangesOverlap(e[e.length-1],t[0],i)){const n=new Array(e.length+t.length-1);return K_.Copy(e,0,n,0,e.length-1),n[e.length-1]=i[0],K_.Copy(t,1,n,e.length,t.length-1),n}else{const n=new Array(e.length+t.length);return K_.Copy(e,0,n,0,e.length),K_.Copy(t,0,n,e.length,t.length),n}}ChangesOverlap(e,t,i){if(G_.Assert(e.originalStart<=t.originalStart,"Left change is not less than or equal to right change"),G_.Assert(e.modifiedStart<=t.modifiedStart,"Left change is not less than or equal to right change"),e.originalStart+e.originalLength>=t.originalStart||e.modifiedStart+e.modifiedLength>=t.modifiedStart){const n=e.originalStart;let o=e.originalLength;const r=e.modifiedStart;let a=e.modifiedLength;return e.originalStart+e.originalLength>=t.originalStart&&(o=t.originalStart+t.originalLength-e.originalStart),e.modifiedStart+e.modifiedLength>=t.modifiedStart&&(a=t.modifiedStart+t.modifiedLength-e.modifiedStart),i[0]=new qh(n,o,r,a),!0}else return i[0]=null,!1}ClipDiagonalBound(e,t,i,n){if(e>=0&&e<n)return e;const o=i,r=n-i-1,a=t%2===0;if(e<0){const l=o%2===0;return a===l?0:1}else{const l=r%2===0;return a===l?n-1:n-2}}}function bx(s){return!!s&&s.visibility!==void 0}function jm(s){switch(s){case!0:return 1;case!1:return 0;default:return s}}function JM(s){return"collapsible"in s}class wNe{constructor(e,t,i={}){this.user=e,this.rootRef=[],this.eventBufferer=new $0,this._onDidSpliceModel=new M,this.onDidSpliceModel=this._onDidSpliceModel.event,this._onDidSpliceRenderedNodes=new M,this.onDidSpliceRenderedNodes=this._onDidSpliceRenderedNodes.event,this._onDidChangeCollapseState=new M,this.onDidChangeCollapseState=this.eventBufferer.wrapEvent(this._onDidChangeCollapseState.event),this._onDidChangeRenderNodeCount=new M,this.onDidChangeRenderNodeCount=this.eventBufferer.wrapEvent(this._onDidChangeRenderNodeCount.event),this.refilterDelayer=new sa(GX),this.collapseByDefault=typeof i.collapseByDefault>"u"?!1:i.collapseByDefault,this.allowNonCollapsibleParents=i.allowNonCollapsibleParents??!1,this.filter=i.filter,this.autoExpandSingleChildren=typeof i.autoExpandSingleChildren>"u"?!1:i.autoExpandSingleChildren,this.root={parent:void 0,element:t,children:[],depth:0,visibleChildrenCount:0,visibleChildIndex:-1,collapsible:!1,collapsed:!1,renderNodeCount:0,visibility:1,visible:!0,filterData:void 0}}splice(e,t,i=ot.empty(),n={}){if(e.length===0)throw new mr(this.user,"Invalid tree location");n.diffIdentityProvider?this.spliceSmart(n.diffIdentityProvider,e,t,i,n):this.spliceSimple(e,t,i,n)}spliceSmart(e,t,i,n=ot.empty(),o,r=o.diffDepth??0){const{parentNode:a}=this.getParentNodeWithListIndex(t);if(!a.lastDiffIds)return this.spliceSimple(t,i,n,o);const l=[...n],c=t[t.length-1],d=new fc({getElements:()=>a.lastDiffIds},{getElements:()=>[...a.children.slice(0,c),...l,...a.children.slice(c+i)].map(p=>e.getId(p.element).toString())}).ComputeDiff(!1);if(d.quitEarly)return a.lastDiffIds=void 0,this.spliceSimple(t,i,l,o);const h=t.slice(0,-1),u=(p,_,b)=>{if(r>0)for(let w=0;w<b;w++)p--,_--,this.spliceSmart(e,[...h,p,0],Number.MAX_SAFE_INTEGER,l[_].children,o,r-1)};let g=Math.min(a.children.length,c+i),f=l.length;for(const p of d.changes.sort((_,b)=>b.originalStart-_.originalStart))u(g,f,g-(p.originalStart+p.originalLength)),g=p.originalStart,f=p.modifiedStart-c,this.spliceSimple([...h,g],p.originalLength,ot.slice(l,f,f+p.modifiedLength),o);u(g,f,g)}spliceSimple(e,t,i=ot.empty(),{onDidCreateNode:n,onDidDeleteNode:o,diffIdentityProvider:r}){const{parentNode:a,listIndex:l,revealed:c,visible:d}=this.getParentNodeWithListIndex(e),h=[],u=ot.map(i,y=>this.createTreeNode(y,a,a.visible?1:0,c,h,n)),g=e[e.length-1];let f=0;for(let y=g;y>=0&&y<a.children.length;y--){const x=a.children[y];if(x.visible){f=x.visibleChildIndex;break}}const p=[];let _=0,b=0;for(const y of u)p.push(y),b+=y.renderNodeCount,y.visible&&(y.visibleChildIndex=f+_++);const w=CE(a.children,g,t,p);r?a.lastDiffIds?CE(a.lastDiffIds,g,t,p.map(y=>r.getId(y.element).toString())):a.lastDiffIds=a.children.map(y=>r.getId(y.element).toString()):a.lastDiffIds=void 0;let C=0;for(const y of w)y.visible&&C++;if(C!==0)for(let y=g+p.length;y<a.children.length;y++){const x=a.children[y];x.visible&&(x.visibleChildIndex-=C)}if(a.visibleChildrenCount+=_-C,w.length>0&&o){const y=x=>{o(x),x.children.forEach(y)};w.forEach(y)}if(c&&d){const y=w.reduce((x,L)=>x+(L.visible?L.renderNodeCount:0),0);this._updateAncestorsRenderNodeCount(a,b-y),this._onDidSpliceRenderedNodes.fire({start:l,deleteCount:y,elements:h})}this._onDidSpliceModel.fire({insertedNodes:p,deletedNodes:w});let v=a;for(;v;){if(v.visibility===2){this.refilterDelayer.trigger(()=>this.refilter());break}v=v.parent}}rerender(e){if(e.length===0)throw new mr(this.user,"Invalid tree location");const{node:t,listIndex:i,revealed:n}=this.getTreeNodeWithListIndex(e);t.visible&&n&&this._onDidSpliceRenderedNodes.fire({start:i,deleteCount:1,elements:[t]})}has(e){return this.hasTreeNode(e)}getListIndex(e){const{listIndex:t,visible:i,revealed:n}=this.getTreeNodeWithListIndex(e);return i&&n?t:-1}getListRenderCount(e){return this.getTreeNode(e).renderNodeCount}isCollapsible(e){return this.getTreeNode(e).collapsible}setCollapsible(e,t){const i=this.getTreeNode(e);typeof t>"u"&&(t=!i.collapsible);const n={collapsible:t};return this.eventBufferer.bufferEvents(()=>this._setCollapseState(e,n))}isCollapsed(e){return this.getTreeNode(e).collapsed}setCollapsed(e,t,i){const n=this.getTreeNode(e);typeof t>"u"&&(t=!n.collapsed);const o={collapsed:t,recursive:i||!1};return this.eventBufferer.bufferEvents(()=>this._setCollapseState(e,o))}_setCollapseState(e,t){const{node:i,listIndex:n,revealed:o}=this.getTreeNodeWithListIndex(e),r=this._setListNodeCollapseState(i,n,o,t);if(i!==this.root&&this.autoExpandSingleChildren&&r&&!JM(t)&&i.collapsible&&!i.collapsed&&!t.recursive){let a=-1;for(let l=0;l<i.children.length;l++)if(i.children[l].visible)if(a>-1){a=-1;break}else a=l;a>-1&&this._setCollapseState([...e,a],t)}return r}_setListNodeCollapseState(e,t,i,n){const o=this._setNodeCollapseState(e,n,!1);if(!i||!e.visible||!o)return o;const r=e.renderNodeCount,a=this.updateNodeAfterCollapseChange(e),l=r-(t===-1?0:1);return this._onDidSpliceRenderedNodes.fire({start:t+1,deleteCount:l,elements:a.slice(1)}),o}_setNodeCollapseState(e,t,i){let n;if(e===this.root?n=!1:(JM(t)?(n=e.collapsible!==t.collapsible,e.collapsible=t.collapsible):e.collapsible?(n=e.collapsed!==t.collapsed,e.collapsed=t.collapsed):n=!1,n&&this._onDidChangeCollapseState.fire({node:e,deep:i})),!JM(t)&&t.recursive)for(const o of e.children)n=this._setNodeCollapseState(o,t,!0)||n;return n}expandTo(e){this.eventBufferer.bufferEvents(()=>{let t=this.getTreeNode(e);for(;t.parent;)t=t.parent,e=e.slice(0,e.length-1),t.collapsed&&this._setCollapseState(e,{collapsed:!1,recursive:!1})})}refilter(){const e=this.root.renderNodeCount,t=this.updateNodeAfterFilterChange(this.root);this._onDidSpliceRenderedNodes.fire({start:0,deleteCount:e,elements:t}),this.refilterDelayer.cancel()}createTreeNode(e,t,i,n,o,r){const a={parent:t,element:e.element,children:[],depth:t.depth+1,visibleChildrenCount:0,visibleChildIndex:-1,collapsible:typeof e.collapsible=="boolean"?e.collapsible:typeof e.collapsed<"u",collapsed:typeof e.collapsed>"u"?this.collapseByDefault:e.collapsed,renderNodeCount:1,visibility:1,visible:!0,filterData:void 0},l=this._filterNode(a,i);a.visibility=l,n&&o.push(a);const c=e.children||ot.empty(),d=n&&l!==0&&!a.collapsed;let h=0,u=1;for(const g of c){const f=this.createTreeNode(g,a,l,d,o,r);a.children.push(f),u+=f.renderNodeCount,f.visible&&(f.visibleChildIndex=h++)}return this.allowNonCollapsibleParents||(a.collapsible=a.collapsible||a.children.length>0),a.visibleChildrenCount=h,a.visible=l===2?h>0:l===1,a.visible?a.collapsed||(a.renderNodeCount=u):(a.renderNodeCount=0,n&&o.pop()),r?.(a),a}updateNodeAfterCollapseChange(e){const t=e.renderNodeCount,i=[];return this._updateNodeAfterCollapseChange(e,i),this._updateAncestorsRenderNodeCount(e.parent,i.length-t),i}_updateNodeAfterCollapseChange(e,t){if(e.visible===!1)return 0;if(t.push(e),e.renderNodeCount=1,!e.collapsed)for(const i of e.children)e.renderNodeCount+=this._updateNodeAfterCollapseChange(i,t);return this._onDidChangeRenderNodeCount.fire(e),e.renderNodeCount}updateNodeAfterFilterChange(e){const t=e.renderNodeCount,i=[];return this._updateNodeAfterFilterChange(e,e.visible?1:0,i),this._updateAncestorsRenderNodeCount(e.parent,i.length-t),i}_updateNodeAfterFilterChange(e,t,i,n=!0){let o;if(e!==this.root){if(o=this._filterNode(e,t),o===0)return e.visible=!1,e.renderNodeCount=0,!1;n&&i.push(e)}const r=i.length;e.renderNodeCount=e===this.root?0:1;let a=!1;if(!e.collapsed||o!==0){let l=0;for(const c of e.children)a=this._updateNodeAfterFilterChange(c,o,i,n&&!e.collapsed)||a,c.visible&&(c.visibleChildIndex=l++);e.visibleChildrenCount=l}else e.visibleChildrenCount=0;return e!==this.root&&(e.visible=o===2?a:o===1,e.visibility=o),e.visible?e.collapsed||(e.renderNodeCount+=i.length-r):(e.renderNodeCount=0,n&&i.pop()),this._onDidChangeRenderNodeCount.fire(e),e.visible}_updateAncestorsRenderNodeCount(e,t){if(t!==0)for(;e;)e.renderNodeCount+=t,this._onDidChangeRenderNodeCount.fire(e),e=e.parent}_filterNode(e,t){const i=this.filter?this.filter.filter(e.element,t):1;return typeof i=="boolean"?(e.filterData=void 0,i?1:0):bx(i)?(e.filterData=i.data,jm(i.visibility)):(e.filterData=void 0,jm(i))}hasTreeNode(e,t=this.root){if(!e||e.length===0)return!0;const[i,...n]=e;return i<0||i>t.children.length?!1:this.hasTreeNode(n,t.children[i])}getTreeNode(e,t=this.root){if(!e||e.length===0)return t;const[i,...n]=e;if(i<0||i>t.children.length)throw new mr(this.user,"Invalid tree location");return this.getTreeNode(n,t.children[i])}getTreeNodeWithListIndex(e){if(e.length===0)return{node:this.root,listIndex:-1,revealed:!0,visible:!1};const{parentNode:t,listIndex:i,revealed:n,visible:o}=this.getParentNodeWithListIndex(e),r=e[e.length-1];if(r<0||r>t.children.length)throw new mr(this.user,"Invalid tree location");const a=t.children[r];return{node:a,listIndex:i,revealed:n,visible:o&&a.visible}}getParentNodeWithListIndex(e,t=this.root,i=0,n=!0,o=!0){const[r,...a]=e;if(r<0||r>t.children.length)throw new mr(this.user,"Invalid tree location");for(let l=0;l<r;l++)i+=t.children[l].renderNodeCount;return n=n&&!t.collapsed,o=o&&t.visible,a.length===0?{parentNode:t,listIndex:i,revealed:n,visible:o}:this.getParentNodeWithListIndex(a,t.children[r],i+1,n,o)}getNode(e=[]){return this.getTreeNode(e)}getNodeLocation(e){const t=[];let i=e;for(;i.parent;)t.push(i.parent.children.indexOf(i)),i=i.parent;return t.reverse()}getParentNodeLocation(e){if(e.length!==0)return e.length===1?[]:She(e)[0]}getFirstElementChild(e){const t=this.getTreeNode(e);if(t.children.length!==0)return t.children[0].element}}class vNe extends cx{constructor(e){super(e.elements.map(t=>t.element)),this.data=e}}function eO(s){return s instanceof cx?new vNe(s):s}class CNe{constructor(e,t){this.modelProvider=e,this.dnd=t,this.autoExpandDisposable=F.None,this.disposables=new U}getDragURI(e){return this.dnd.getDragURI(e.element)}getDragLabel(e,t){if(this.dnd.getDragLabel)return this.dnd.getDragLabel(e.map(i=>i.element),t)}onDragStart(e,t){this.dnd.onDragStart?.(eO(e),t)}onDragOver(e,t,i,n,o,r=!0){const a=this.dnd.onDragOver(eO(e),t&&t.element,i,n,o),l=this.autoExpandNode!==t;if(l&&(this.autoExpandDisposable.dispose(),this.autoExpandNode=t),typeof t>"u")return a;if(l&&typeof a!="boolean"&&a.autoExpand&&(this.autoExpandDisposable=hh(()=>{const g=this.modelProvider(),f=g.getNodeLocation(t);g.isCollapsed(f)&&g.setCollapsed(f,!1),this.autoExpandNode=void 0},500,this.disposables)),typeof a=="boolean"||!a.accept||typeof a.bubble>"u"||a.feedback){if(!r){const g=typeof a=="boolean"?a:a.accept,f=typeof a=="boolean"?void 0:a.effect;return{accept:g,effect:f,feedback:[i]}}return a}if(a.bubble===1){const g=this.modelProvider(),f=g.getNodeLocation(t),p=g.getParentNodeLocation(f),_=g.getNode(p),b=p&&g.getListIndex(p);return this.onDragOver(e,_,b,n,o,!1)}const c=this.modelProvider(),d=c.getNodeLocation(t),h=c.getListIndex(d),u=c.getListRenderCount(d);return{...a,feedback:Us(h,h+u)}}drop(e,t,i,n,o){this.autoExpandDisposable.dispose(),this.autoExpandNode=void 0,this.dnd.drop(eO(e),t&&t.element,i,n,o)}onDragEnd(e){this.dnd.onDragEnd?.(e)}dispose(){this.disposables.dispose(),this.dnd.dispose()}}function yNe(s,e,t){return t&&{...t,identityProvider:t.identityProvider&&{getId(i){return t.identityProvider.getId(i.element)}},dnd:t.dnd&&e.add(new CNe(s,t.dnd)),multipleSelectionController:t.multipleSelectionController&&{isSelectionSingleChangeEvent(i){return t.multipleSelectionController.isSelectionSingleChangeEvent({...i,element:i.element})},isSelectionRangeChangeEvent(i){return t.multipleSelectionController.isSelectionRangeChangeEvent({...i,element:i.element})}},accessibilityProvider:t.accessibilityProvider&&{...t.accessibilityProvider,getSetSize(i){const n=s(),o=n.getNodeLocation(i),r=n.getParentNodeLocation(o);return n.getNode(r).visibleChildrenCount},getPosInSet(i){return i.visibleChildIndex+1},isChecked:t.accessibilityProvider&&t.accessibilityProvider.isChecked?i=>t.accessibilityProvider.isChecked(i.element):void 0,getRole:t.accessibilityProvider&&t.accessibilityProvider.getRole?i=>t.accessibilityProvider.getRole(i.element):()=>"treeitem",getAriaLabel(i){return t.accessibilityProvider.getAriaLabel(i.element)},getWidgetAriaLabel(){return t.accessibilityProvider.getWidgetAriaLabel()},getWidgetRole:t.accessibilityProvider&&t.accessibilityProvider.getWidgetRole?()=>t.accessibilityProvider.getWidgetRole():()=>"tree",getAriaLevel:t.accessibilityProvider&&t.accessibilityProvider.getAriaLevel?i=>t.accessibilityProvider.getAriaLevel(i.element):i=>i.depth,getActiveDescendantId:t.accessibilityProvider.getActiveDescendantId&&(i=>t.accessibilityProvider.getActiveDescendantId(i.element))},keyboardNavigationLabelProvider:t.keyboardNavigationLabelProvider&&{...t.keyboardNavigationLabelProvider,getKeyboardNavigationLabel(i){return t.keyboardNavigationLabelProvider.getKeyboardNavigationLabel(i.element)}}}}class mH{constructor(e){this.delegate=e}getHeight(e){return this.delegate.getHeight(e.element)}getTemplateId(e){return this.delegate.getTemplateId(e.element)}hasDynamicHeight(e){return!!this.delegate.hasDynamicHeight&&this.delegate.hasDynamicHeight(e.element)}setDynamicHeight(e,t){this.delegate.setDynamicHeight?.(e.element,t)}}var Gm;(function(s){s.None="none",s.OnHover="onHover",s.Always="always"})(Gm||(Gm={}));class SNe{get elements(){return this._elements}constructor(e,t=[]){this._elements=t,this.disposables=new U,this.onDidChange=oe.forEach(e,i=>this._elements=i,this.disposables)}dispose(){this.disposables.dispose()}}const tw=class tw{constructor(e,t,i,n,o,r={}){this.renderer=e,this.model=t,this.activeNodes=n,this.renderedIndentGuides=o,this.renderedElements=new Map,this.renderedNodes=new Map,this.indent=tw.DefaultIndent,this.hideTwistiesOfChildlessElements=!1,this.shouldRenderIndentGuides=!1,this.activeIndentNodes=new Set,this.indentGuidesDisposable=F.None,this.disposables=new U,this.templateId=e.templateId,this.updateOptions(r),oe.map(i,a=>a.node)(this.onDidChangeNodeTwistieState,this,this.disposables),e.onDidChangeTwistieState?.(this.onDidChangeTwistieState,this,this.disposables)}updateOptions(e={}){if(typeof e.indent<"u"){const t=qs(e.indent,0,40);if(t!==this.indent){this.indent=t;for(const[i,n]of this.renderedNodes)n.indentSize=tw.DefaultIndent+(i.depth-1)*this.indent,this.renderTreeElement(i,n)}}if(typeof e.renderIndentGuides<"u"){const t=e.renderIndentGuides!==Gm.None;if(t!==this.shouldRenderIndentGuides){this.shouldRenderIndentGuides=t;for(const[i,n]of this.renderedNodes)this._renderIndentGuides(i,n);if(this.indentGuidesDisposable.dispose(),t){const i=new U;this.activeNodes.onDidChange(this._onDidChangeActiveNodes,this,i),this.indentGuidesDisposable=i,this._onDidChangeActiveNodes(this.activeNodes.elements)}}}typeof e.hideTwistiesOfChildlessElements<"u"&&(this.hideTwistiesOfChildlessElements=e.hideTwistiesOfChildlessElements)}renderTemplate(e){const t=X(e,ie(".monaco-tl-row")),i=X(t,ie(".monaco-tl-indent")),n=X(t,ie(".monaco-tl-twistie")),o=X(t,ie(".monaco-tl-contents")),r=this.renderer.renderTemplate(o);return{container:e,indent:i,twistie:n,indentGuidesDisposable:F.None,indentSize:0,templateData:r}}renderElement(e,t,i,n){i.indentSize=tw.DefaultIndent+(e.depth-1)*this.indent,this.renderedNodes.set(e,i),this.renderedElements.set(e.element,e),this.renderTreeElement(e,i),this.renderer.renderElement(e,t,i.templateData,{...n,indent:i.indentSize})}disposeElement(e,t,i,n){i.indentGuidesDisposable.dispose(),this.renderer.disposeElement?.(e,t,i.templateData,{...n,indent:i.indentSize}),typeof n?.height=="number"&&(this.renderedNodes.delete(e),this.renderedElements.delete(e.element))}disposeTemplate(e){this.renderer.disposeTemplate(e.templateData)}onDidChangeTwistieState(e){const t=this.renderedElements.get(e);t&&this.onDidChangeNodeTwistieState(t)}onDidChangeNodeTwistieState(e){const t=this.renderedNodes.get(e);t&&(this._onDidChangeActiveNodes(this.activeNodes.elements),this.renderTreeElement(e,t))}renderTreeElement(e,t){t.twistie.style.paddingLeft=`${t.indentSize}px`,t.indent.style.width=`${t.indentSize+this.indent-16}px`,e.collapsible?t.container.setAttribute("aria-expanded",String(!e.collapsed)):t.container.removeAttribute("aria-expanded"),t.twistie.classList.remove(...xe.asClassNameArray(Z.treeItemExpanded));let i=!1;this.renderer.renderTwistie&&(i=this.renderer.renderTwistie(e.element,t.twistie)),e.collapsible&&(!this.hideTwistiesOfChildlessElements||e.visibleChildrenCount>0)?(i||t.twistie.classList.add(...xe.asClassNameArray(Z.treeItemExpanded)),t.twistie.classList.add("collapsible"),t.twistie.classList.toggle("collapsed",e.collapsed)):t.twistie.classList.remove("collapsible","collapsed"),this._renderIndentGuides(e,t)}_renderIndentGuides(e,t){if(Pn(t.indent),t.indentGuidesDisposable.dispose(),!this.shouldRenderIndentGuides)return;const i=new U;for(;;){const n=this.model.getNodeLocation(e),o=this.model.getParentNodeLocation(n);if(!o)break;const r=this.model.getNode(o),a=ie(".indent-guide",{style:`width: ${this.indent}px`});this.activeIndentNodes.has(r)&&a.classList.add("active"),t.indent.childElementCount===0?t.indent.appendChild(a):t.indent.insertBefore(a,t.indent.firstElementChild),this.renderedIndentGuides.add(r,a),i.add(pe(()=>this.renderedIndentGuides.delete(r,a))),e=r}t.indentGuidesDisposable=i}_onDidChangeActiveNodes(e){if(!this.shouldRenderIndentGuides)return;const t=new Set;e.forEach(i=>{const n=this.model.getNodeLocation(i);try{const o=this.model.getParentNodeLocation(n);i.collapsible&&i.children.length>0&&!i.collapsed?t.add(i):o&&t.add(this.model.getNode(o))}catch{}}),this.activeIndentNodes.forEach(i=>{t.has(i)||this.renderedIndentGuides.forEach(i,n=>n.classList.remove("active"))}),t.forEach(i=>{this.activeIndentNodes.has(i)||this.renderedIndentGuides.forEach(i,n=>n.classList.add("active"))}),this.activeIndentNodes=t}dispose(){this.renderedNodes.clear(),this.renderedElements.clear(),this.indentGuidesDisposable.dispose(),xt(this.disposables)}};tw.DefaultIndent=8;let m5=tw;function xNe(s,e){const t=e.toLowerCase().indexOf(s);let i;if(t>-1){i=[Number.MAX_SAFE_INTEGER,0];for(let n=s.length;n>0;n--)i.push(t+n-1)}return i}class ase{get totalCount(){return this._totalCount}get matchCount(){return this._matchCount}set findMatchType(e){this._findMatchType=e}get findMatchType(){return this._findMatchType}set findMode(e){this._findMode=e}get findMode(){return this._findMode}constructor(e,t,i){this._keyboardNavigationLabelProvider=e,this._filter=t,this._defaultFindVisibility=i,this._totalCount=0,this._matchCount=0,this._findMatchType=fl.Fuzzy,this._findMode=_r.Highlight,this._pattern="",this._lowercasePattern="",this.disposables=new U}filter(e,t){let i=1;if(this._filter){const r=this._filter.filter(e,t);if(typeof r=="boolean"?i=r?1:0:bx(r)?i=jm(r.visibility):i=r,i===0)return!1}if(this._totalCount++,!this._pattern)return this._matchCount++,{data:Oa.Default,visibility:i};const n=this._keyboardNavigationLabelProvider.getKeyboardNavigationLabel(e),o=Array.isArray(n)?n:[n];for(const r of o){const a=r&&r.toString();if(typeof a>"u")return{data:Oa.Default,visibility:i};let l;if(this._findMatchType===fl.Contiguous?l=xNe(this._lowercasePattern,a.toLowerCase()):l=Om(this._pattern,this._lowercasePattern,0,a,a.toLowerCase(),0,{firstMatchCanBeWeak:!0,boostFullMatch:!0}),l)return this._matchCount++,o.length===1?{data:l,visibility:i}:{data:{label:a,score:l},visibility:i}}return this._findMode===_r.Filter?typeof this._defaultFindVisibility=="number"?this._defaultFindVisibility:this._defaultFindVisibility?this._defaultFindVisibility(e):2:{data:Oa.Default,visibility:i}}reset(){this._totalCount=0,this._matchCount=0}dispose(){xt(this.disposables)}}class LNe{constructor(e){this.stateMap=new Map(e.map(t=>[t.id,{...t}]))}get(e){const t=this.stateMap.get(e);if(t===void 0)throw new Error(`No state found for toggle id ${e}`);return t.isChecked}set(e,t){const i=this.stateMap.get(e);if(i===void 0)throw new Error(`No state found for toggle id ${e}`);return i.isChecked===t?!1:(i.isChecked=t,!0)}}var _r;(function(s){s[s.Highlight=0]="Highlight",s[s.Filter=1]="Filter"})(_r||(_r={}));var fl;(function(s){s[s.Fuzzy=0]="Fuzzy",s[s.Contiguous=1]="Contiguous"})(fl||(fl={}));var Yh;(function(s){s.Mode="mode",s.MatchType="matchType"})(Yh||(Yh={}));class kNe{get pattern(){return this._pattern}get placeholder(){return this._placeholder}set placeholder(e){this._placeholder=e,this.widget?.setPlaceHolder(e)}constructor(e,t,i,n={}){this.tree=e,this.filter=t,this.contextViewProvider=i,this.options=n,this._pattern="",this._onDidChangePattern=new M,this._onDidChangeOpenState=new M,this.onDidChangeOpenState=this._onDidChangeOpenState.event,this.enabledDisposables=new U,this.disposables=new U,this.toggles=new LNe(n.toggles??[]),this._placeholder=n.placeholder??m(20,"Type to search")}isOpened(){return!!this.widget}updateToggleState(e,t){this.toggles.set(e,t),this.widget?.setToggleState(e,t)}renderMessage(e,t){e?this.tree.options.showNotFoundMessage??!0?this.widget?.showMessage({type:2,content:t??m(21,"No results found.")}):this.widget?.showMessage({type:2}):this.widget?.clearMessage()}alertResults(e){Js(e?m(23,"{0} results",e):m(22,"No results"))}dispose(){this._history=void 0,this._onDidChangePattern.dispose(),this.enabledDisposables.dispose(),this.disposables.dispose()}}let lse=class extends kNe{get mode(){return this.toggles.get(Yh.Mode)?_r.Filter:_r.Highlight}set mode(e){if(e===this.mode)return;const t=e===_r.Filter;this.updateToggleState(Yh.Mode,t),this.placeholder=t?m(24,"Type to filter"):m(25,"Type to search"),this.filter.findMode=e,this.tree.refilter(),this.render(),this._onDidChangeMode.fire(e)}get matchType(){return this.toggles.get(Yh.MatchType)?fl.Fuzzy:fl.Contiguous}set matchType(e){e!==this.matchType&&(this.updateToggleState(Yh.MatchType,e===fl.Fuzzy),this.filter.findMatchType=e,this.tree.refilter(),this.render(),this._onDidChangeMatchType.fire(e))}constructor(e,t,i,n={}){const o=n.defaultFindMode??_r.Highlight,r=n.defaultFindMatchType??fl.Fuzzy,a=[{id:Yh.Mode,icon:Z.listFilter,title:m(26,"Filter"),isChecked:o===_r.Filter},{id:Yh.MatchType,icon:Z.searchFuzzy,title:m(27,"Fuzzy Match"),isChecked:r===fl.Fuzzy}];t.findMatchType=r,t.findMode=o,super(e,t,i,{...n,toggles:a}),this.filter=t,this._onDidChangeMode=new M,this.onDidChangeMode=this._onDidChangeMode.event,this._onDidChangeMatchType=new M,this.onDidChangeMatchType=this._onDidChangeMatchType.event,this.disposables.add(this.tree.onDidChangeModel(()=>{this.isOpened()&&(this.pattern.length!==0&&this.tree.refilter(),this.render())})),this.disposables.add(this.tree.onWillRefilter(()=>this.filter.reset()))}updateOptions(e={}){e.defaultFindMode!==void 0&&(this.mode=e.defaultFindMode),e.defaultFindMatchType!==void 0&&(this.matchType=e.defaultFindMatchType)}shouldAllowFocus(e){return!this.isOpened()||!this.pattern||this.filter.totalCount>0&&this.filter.matchCount<=1?!0:!Oa.isDefault(e.filterData)}render(){const t=this.filter.matchCount===0&&this.filter.totalCount>0&&this.pattern.length>0;this.renderMessage(t),this.pattern.length&&this.alertResults(this.filter.matchCount)}};function ENe(s,e){return s.position===e.position&&cse(s,e)}function cse(s,e){return s.node.element===e.node.element&&s.startIndex===e.startIndex&&s.height===e.height&&s.endIndex===e.endIndex}class INe{constructor(e=[]){this.stickyNodes=e}get count(){return this.stickyNodes.length}equal(e){return ni(this.stickyNodes,e.stickyNodes,ENe)}contains(e){return this.stickyNodes.some(t=>t.node.element===e.element)}lastNodePartiallyVisible(){if(this.count===0)return!1;const e=this.stickyNodes[this.count-1];if(this.count===1)return e.position!==0;const t=this.stickyNodes[this.count-2];return t.position+t.height!==e.position}animationStateChanged(e){if(!ni(this.stickyNodes,e.stickyNodes,cse)||this.count===0)return!1;const t=this.stickyNodes[this.count-1],i=e.stickyNodes[e.count-1];return t.position!==i.position}}class DNe{constrainStickyScrollNodes(e,t,i){for(let n=0;n<e.length;n++){const o=e[n];if(o.position+o.height>i||n>=t)return e.slice(0,n)}return e}}let Uj=class extends F{constructor(e,t,i,n,o,r={}){super(),this.tree=e,this.model=t,this.view=i,this.treeDelegate=o,this.maxWidgetViewRatio=.4;const a=this.validateStickySettings(r);this.stickyScrollMaxItemCount=a.stickyScrollMaxItemCount,this.stickyScrollDelegate=r.stickyScrollDelegate??new DNe,this.paddingTop=r.paddingTop??0,this._widget=this._register(new TNe(i.getScrollableElement(),i,e,n,o,r.accessibilityProvider)),this.onDidChangeHasFocus=this._widget.onDidChangeHasFocus,this.onContextMenu=this._widget.onContextMenu,this._register(i.onDidScroll(()=>this.update())),this._register(i.onDidChangeContentHeight(()=>this.update())),this._register(e.onDidChangeCollapseState(()=>this.update())),this._register(t.onDidSpliceRenderedNodes(l=>{const c=this._widget.state;if(!c)return;if(l.deleteCount>0&&c.stickyNodes.some(u=>!this.model.has(this.model.getNodeLocation(u.node)))){this.update();return}c.stickyNodes.some(u=>{const g=this.model.getListIndex(this.model.getNodeLocation(u.node));return g>=l.start&&g<l.start+l.deleteCount&&c.contains(u.node)})&&this._widget.rerender()})),this.update()}get height(){return this._widget.height}getNodeAtHeight(e){let t;if(e===0?t=this.view.firstVisibleIndex:t=this.view.indexAt(e+this.view.scrollTop),!(t<0||t>=this.view.length))return this.view.element(t)}update(){const e=this.getNodeAtHeight(this.paddingTop);if(!e||this.tree.scrollTop<=this.paddingTop){this._widget.setState(void 0);return}const t=this.findStickyState(e);this._widget.setState(t)}findStickyState(e){const t=[];let i=e,n=0,o=this.getNextStickyNode(i,void 0,n);for(;o&&(t.push(o),n+=o.height,!(t.length<=this.stickyScrollMaxItemCount&&(i=this.getNextVisibleNode(o),!i)));)o=this.getNextStickyNode(i,o.node,n);const r=this.constrainStickyNodes(t);return r.length?new INe(r):void 0}getNextVisibleNode(e){return this.getNodeAtHeight(e.position+e.height)}getNextStickyNode(e,t,i){const n=this.getAncestorUnderPrevious(e,t);if(n&&!(n===e&&(!this.nodeIsUncollapsedParent(e)||this.nodeTopAlignsWithStickyNodesBottom(e,i))))return this.createStickyScrollNode(n,i)}nodeTopAlignsWithStickyNodesBottom(e,t){const i=this.getNodeIndex(e),n=this.view.getElementTop(i),o=t;return this.view.scrollTop===n-o}createStickyScrollNode(e,t){const i=this.treeDelegate.getHeight(e),{startIndex:n,endIndex:o}=this.getNodeRange(e),r=this.calculateStickyNodePosition(o,t,i);return{node:e,position:r,height:i,startIndex:n,endIndex:o}}getAncestorUnderPrevious(e,t=void 0){let i=e,n=this.getParentNode(i);for(;n;){if(n===t)return i;i=n,n=this.getParentNode(i)}if(t===void 0)return i}calculateStickyNodePosition(e,t,i){let n=this.view.getRelativeTop(e);if(n===null&&this.view.firstVisibleIndex===e&&e+1<this.view.length){const c=this.treeDelegate.getHeight(this.view.element(e)),d=this.view.getRelativeTop(e+1);n=d?d-c/this.view.renderHeight:null}if(n===null)return t;const o=this.view.element(e),r=this.treeDelegate.getHeight(o),l=n*this.view.renderHeight+r;return t+i>l&&t<=l?l-i:t}constrainStickyNodes(e){if(e.length===0)return[];const t=this.view.renderHeight*this.maxWidgetViewRatio,i=e[e.length-1];if(e.length<=this.stickyScrollMaxItemCount&&i.position+i.height<=t)return e;const n=this.stickyScrollDelegate.constrainStickyScrollNodes(e,this.stickyScrollMaxItemCount,t);if(!n.length)return[];const o=n[n.length-1];if(n.length>this.stickyScrollMaxItemCount||o.position+o.height>t)throw new Error("stickyScrollDelegate violates constraints");return n}getParentNode(e){const t=this.model.getNodeLocation(e),i=this.model.getParentNodeLocation(t);return i?this.model.getNode(i):void 0}nodeIsUncollapsedParent(e){const t=this.model.getNodeLocation(e);return this.model.getListRenderCount(t)>1}getNodeIndex(e){const t=this.model.getNodeLocation(e);return this.model.getListIndex(t)}getNodeRange(e){const t=this.model.getNodeLocation(e),i=this.model.getListIndex(t);if(i<0)throw new Error("Node not found in tree");const n=this.model.getListRenderCount(t),o=i+n-1;return{startIndex:i,endIndex:o}}nodePositionTopBelowWidget(e){const t=[];let i=this.getParentNode(e);for(;i;)t.push(i),i=this.getParentNode(i);let n=0;for(let o=0;o<t.length&&o<this.stickyScrollMaxItemCount;o++)n+=this.treeDelegate.getHeight(t[o]);return n}domFocus(){this._widget.domFocus()}focusedLast(){return this._widget.focusedLast()}updateOptions(e={}){if(e.paddingTop!==void 0&&(this.paddingTop=e.paddingTop),e.stickyScrollMaxItemCount!==void 0){const t=this.validateStickySettings(e);this.stickyScrollMaxItemCount!==t.stickyScrollMaxItemCount&&(this.stickyScrollMaxItemCount=t.stickyScrollMaxItemCount,this.update())}}validateStickySettings(e){let t=7;return typeof e.stickyScrollMaxItemCount=="number"&&(t=Math.max(e.stickyScrollMaxItemCount,1)),{stickyScrollMaxItemCount:t}}},TNe=class{get state(){return this._previousState}constructor(e,t,i,n,o,r){this.view=t,this.tree=i,this.treeRenderers=n,this.treeDelegate=o,this.accessibilityProvider=r,this._previousElements=[],this._previousStateDisposables=new U,this._rootDomNode=ie(".monaco-tree-sticky-container.empty"),e.appendChild(this._rootDomNode);const a=ie(".monaco-tree-sticky-container-shadow");this._rootDomNode.appendChild(a),this.stickyScrollFocus=new NNe(this._rootDomNode,t),this.onDidChangeHasFocus=this.stickyScrollFocus.onDidChangeHasFocus,this.onContextMenu=this.stickyScrollFocus.onContextMenu}get height(){if(!this._previousState)return 0;const e=this._previousState.stickyNodes[this._previousState.count-1];return e.position+e.height}setState(e){const t=!!this._previousState&&this._previousState.count>0,i=!!e&&e.count>0;if(!t&&!i||t&&i&&this._previousState.equal(e))return;if(t!==i&&this.setVisible(i),!i){this._previousState=void 0,this._previousElements=[],this._previousStateDisposables.clear();return}const n=e.stickyNodes[e.count-1];this._previousState&&e.animationStateChanged(this._previousState)?this._previousElements[this._previousState.count-1].style.top=`${n.position}px`:this.renderState(e),this._previousState=e,this._rootDomNode.style.height=`${n.position+n.height}px`}renderState(e){this._previousStateDisposables.clear();const t=Array(e.count);for(let i=e.count-1;i>=0;i--){const n=e.stickyNodes[i],{element:o,disposable:r}=this.createElement(n,i,e.count);t[i]=o,this._rootDomNode.appendChild(o),this._previousStateDisposables.add(r)}this.stickyScrollFocus.updateElements(t,e),this._previousElements=t}rerender(){this._previousState&&this.renderState(this._previousState)}createElement(e,t,i){const n=e.startIndex,o=document.createElement("div");o.style.top=`${e.position}px`,this.tree.options.setRowHeight!==!1&&(o.style.height=`${e.height}px`),this.tree.options.setRowLineHeight!==!1&&(o.style.lineHeight=`${e.height}px`),o.classList.add("monaco-tree-sticky-row"),o.classList.add("monaco-list-row"),o.setAttribute("data-index",`${n}`),o.setAttribute("data-parity",n%2===0?"even":"odd"),o.setAttribute("id",this.view.getElementID(n));const r=this.setAccessibilityAttributes(o,e.node.element,t,i),a=this.treeDelegate.getTemplateId(e.node),l=this.treeRenderers.find(u=>u.templateId===a);if(!l)throw new Error(`No renderer found for template id ${a}`);let c=e.node;c===this.tree.getNode(this.tree.getNodeLocation(e.node))&&(c=new Proxy(e.node,{}));const d=l.renderTemplate(o);l.renderElement(c,e.startIndex,d,{height:e.height});const h=pe(()=>{r.dispose(),l.disposeElement(c,e.startIndex,d,{height:e.height}),l.disposeTemplate(d),o.remove()});return{element:o,disposable:h}}setAccessibilityAttributes(e,t,i,n){if(!this.accessibilityProvider)return F.None;this.accessibilityProvider.getSetSize&&e.setAttribute("aria-setsize",String(this.accessibilityProvider.getSetSize(t,i,n))),this.accessibilityProvider.getPosInSet&&e.setAttribute("aria-posinset",String(this.accessibilityProvider.getPosInSet(t,i))),this.accessibilityProvider.getRole&&e.setAttribute("role",this.accessibilityProvider.getRole(t)??"treeitem");const o=this.accessibilityProvider.getAriaLabel(t),r=o&&typeof o!="string"?o:zt(o),a=De(c=>{const d=c.readObservable(r);d?e.setAttribute("aria-label",d):e.removeAttribute("aria-label")});typeof o=="string"||o&&e.setAttribute("aria-label",o.get());const l=this.accessibilityProvider.getAriaLevel&&this.accessibilityProvider.getAriaLevel(t);return typeof l=="number"&&e.setAttribute("aria-level",`${l}`),e.setAttribute("aria-selected",String(!1)),a}setVisible(e){this._rootDomNode.classList.toggle("empty",!e),e||this.stickyScrollFocus.updateElements([],void 0)}domFocus(){this.stickyScrollFocus.domFocus()}focusedLast(){return this.stickyScrollFocus.focusedLast()}dispose(){this.stickyScrollFocus.dispose(),this._previousStateDisposables.dispose(),this._rootDomNode.remove()}};class NNe extends F{get domHasFocus(){return this._domHasFocus}set domHasFocus(e){e!==this._domHasFocus&&(this._onDidChangeHasFocus.fire(e),this._domHasFocus=e)}constructor(e,t){super(),this.container=e,this.view=t,this.focusedIndex=-1,this.elements=[],this._onDidChangeHasFocus=new M,this.onDidChangeHasFocus=this._onDidChangeHasFocus.event,this._onContextMenu=new M,this.onContextMenu=this._onContextMenu.event,this._domHasFocus=!1,this._register(W(this.container,"focus",()=>this.onFocus())),this._register(W(this.container,"blur",()=>this.onBlur())),this._register(this.view.onDidFocus(()=>this.toggleStickyScrollFocused(!1))),this._register(this.view.onKeyDown(i=>this.onKeyDown(i))),this._register(this.view.onMouseDown(i=>this.onMouseDown(i))),this._register(this.view.onContextMenu(i=>this.handleContextMenu(i)))}handleContextMenu(e){const t=e.browserEvent.target;if(!MS(t)&&!G1(t)){this.focusedLast()&&this.view.domFocus();return}if(!Id(e.browserEvent)){if(!this.state)throw new Error("Context menu should not be triggered when state is undefined");const r=this.state.stickyNodes.findIndex(a=>a.node.element===e.element?.element);if(r===-1)throw new Error("Context menu should not be triggered when element is not in sticky scroll widget");this.container.focus(),this.setFocus(r);return}if(!this.state||this.focusedIndex<0)throw new Error("Context menu key should not be triggered when focus is not in sticky scroll widget");const n=this.state.stickyNodes[this.focusedIndex].node.element,o=this.elements[this.focusedIndex];this._onContextMenu.fire({element:n,anchor:o,browserEvent:e.browserEvent,isStickyScroll:!0})}onKeyDown(e){if(this.domHasFocus&&this.state){if(e.key==="ArrowUp")this.setFocusedElement(Math.max(0,this.focusedIndex-1)),e.preventDefault(),e.stopPropagation();else if(e.key==="ArrowDown"||e.key==="ArrowRight"){if(this.focusedIndex>=this.state.count-1){const t=this.state.stickyNodes[this.state.count-1].startIndex+1;this.view.domFocus(),this.view.setFocus([t]),this.scrollNodeUnderWidget(t,this.state)}else this.setFocusedElement(this.focusedIndex+1);e.preventDefault(),e.stopPropagation()}}}onMouseDown(e){const t=e.browserEvent.target;!MS(t)&&!G1(t)||(e.browserEvent.preventDefault(),e.browserEvent.stopPropagation())}updateElements(e,t){if(t&&t.count===0)throw new Error("Sticky scroll state must be undefined when there are no sticky nodes");if(t&&t.count!==e.length)throw new Error("Sticky scroll focus received illigel state");const i=this.focusedIndex;if(this.removeFocus(),this.elements=e,this.state=t,t){const n=qs(i,0,t.count-1);this.setFocus(n)}else this.domHasFocus&&this.view.domFocus();this.container.tabIndex=t?0:-1}setFocusedElement(e){const t=this.state;if(!t)throw new Error("Cannot set focus when state is undefined");if(this.setFocus(e),!(e<t.count-1)&&t.lastNodePartiallyVisible()){const i=t.stickyNodes[e];this.scrollNodeUnderWidget(i.endIndex+1,t)}}scrollNodeUnderWidget(e,t){const i=t.stickyNodes[t.count-1],n=t.count>1?t.stickyNodes[t.count-2]:void 0,o=this.view.getElementTop(e),r=n?n.position+n.height+i.height:i.height;this.view.scrollTop=o-r}domFocus(){if(!this.state)throw new Error("Cannot focus when state is undefined");this.container.focus()}focusedLast(){return this.state?this.view.getHTMLElement().classList.contains("sticky-scroll-focused"):!1}removeFocus(){this.focusedIndex!==-1&&(this.toggleElementFocus(this.elements[this.focusedIndex],!1),this.focusedIndex=-1)}setFocus(e){if(0>e)throw new Error("addFocus() can not remove focus");if(!this.state&&e>=0)throw new Error("Cannot set focus index when state is undefined");if(this.state&&e>=this.state.count)throw new Error("Cannot set focus index to an index that does not exist");const t=this.focusedIndex;t>=0&&this.toggleElementFocus(this.elements[t],!1),e>=0&&this.toggleElementFocus(this.elements[e],!0),this.focusedIndex=e}toggleElementFocus(e,t){this.toggleElementActiveFocus(e,t&&this.domHasFocus),this.toggleElementPassiveFocus(e,t)}toggleCurrentElementActiveFocus(e){this.focusedIndex!==-1&&this.toggleElementActiveFocus(this.elements[this.focusedIndex],e)}toggleElementActiveFocus(e,t){e.classList.toggle("focused",t)}toggleElementPassiveFocus(e,t){e.classList.toggle("passive-focused",t)}toggleStickyScrollFocused(e){this.view.getHTMLElement().classList.toggle("sticky-scroll-focused",e)}onFocus(){if(!this.state||this.elements.length===0)throw new Error("Cannot focus when state is undefined or elements are empty");this.domHasFocus=!0,this.toggleStickyScrollFocused(!0),this.toggleCurrentElementActiveFocus(!0),this.focusedIndex===-1&&this.setFocus(0)}onBlur(){this.domHasFocus=!1,this.toggleCurrentElementActiveFocus(!1)}dispose(){this.toggleStickyScrollFocused(!1),this._onDidChangeHasFocus.fire(!1),super.dispose()}}function PL(s){let e=Jf.Unknown;return V2(s.browserEvent.target,"monaco-tl-twistie","monaco-tl-row")?e=Jf.Twistie:V2(s.browserEvent.target,"monaco-tl-contents","monaco-tl-row")?e=Jf.Element:V2(s.browserEvent.target,"monaco-tree-type-filter","monaco-list")&&(e=Jf.Filter),{browserEvent:s.browserEvent,element:s.element?s.element.element:null,target:e}}function RNe(s){const e=MS(s.browserEvent.target);return{element:s.element?s.element.element:null,browserEvent:s.browserEvent,anchor:s.anchor,isStickyScroll:e}}function $k(s,e){e(s),s.children.forEach(t=>$k(t,e))}class tO{get nodeSet(){return this._nodeSet||(this._nodeSet=this.createNodeSet()),this._nodeSet}constructor(e,t){this.getFirstViewElementWithTrait=e,this.identityProvider=t,this.nodes=[],this._onDidChange=new M,this.onDidChange=this._onDidChange.event}set(e,t){!t?.__forceEvent&&ni(this.nodes,e)||this._set(e,!1,t)}_set(e,t,i){if(this.nodes=[...e],this.elements=void 0,this._nodeSet=void 0,!t){const n=this;this._onDidChange.fire({get elements(){return n.get()},browserEvent:i})}}get(){return this.elements||(this.elements=this.nodes.map(e=>e.element)),[...this.elements]}getNodes(){return this.nodes}has(e){return this.nodeSet.has(e)}onDidModelSplice({insertedNodes:e,deletedNodes:t}){if(!this.identityProvider){const l=this.createNodeSet(),c=d=>l.delete(d);t.forEach(d=>$k(d,c)),this.set([...l.values()]);return}const i=new Set,n=l=>i.add(this.identityProvider.getId(l.element).toString());t.forEach(l=>$k(l,n));const o=new Map,r=l=>o.set(this.identityProvider.getId(l.element).toString(),l);e.forEach(l=>$k(l,r));const a=[];for(const l of this.nodes){const c=this.identityProvider.getId(l.element).toString();if(!i.has(c))a.push(l);else{const h=o.get(c);h&&h.visible&&a.push(h)}}if(this.nodes.length>0&&a.length===0){const l=this.getFirstViewElementWithTrait();l&&a.push(l)}this._set(a,!0)}createNodeSet(){const e=new Set;for(const t of this.nodes)e.add(t);return e}}class ANe extends cie{constructor(e,t,i){super(e),this.tree=t,this.stickyScrollProvider=i}onViewPointer(e){if(oie(e.browserEvent.target)||wl(e.browserEvent.target)||h1(e.browserEvent.target)||e.browserEvent.isHandledByList)return;const t=e.element;if(!t)return super.onViewPointer(e);if(this.isSelectionRangeChangeEvent(e)||this.isSelectionSingleChangeEvent(e))return super.onViewPointer(e);const i=e.browserEvent.target,n=i.classList.contains("monaco-tl-twistie")||i.classList.contains("monaco-icon-label")&&i.classList.contains("folder-icon")&&e.browserEvent.offsetX<16,o=G1(e.browserEvent.target);let r=!1;if(o?r=!0:typeof this.tree.expandOnlyOnTwistieClick=="function"?r=this.tree.expandOnlyOnTwistieClick(t.element):r=!!this.tree.expandOnlyOnTwistieClick,o)this.handleStickyScrollMouseEvent(e,t);else{if(r&&!n&&e.browserEvent.detail!==2)return super.onViewPointer(e);if(!this.tree.expandOnDoubleClick&&e.browserEvent.detail===2)return super.onViewPointer(e)}if(t.collapsible&&(!o||n)){const a=this.tree.getNodeLocation(t),l=e.browserEvent.altKey;if(this.tree.setFocus([a]),this.tree.toggleCollapsed(a,l),n){e.browserEvent.isHandledByList=!0;return}}o||super.onViewPointer(e)}handleStickyScrollMouseEvent(e,t){if(bLe(e.browserEvent.target)||wLe(e.browserEvent.target))return;const i=this.stickyScrollProvider();if(!i)throw new Error("Sticky scroll controller not found");const n=this.list.indexOf(t),o=this.list.getElementTop(n),r=i.nodePositionTopBelowWidget(t);this.tree.scrollTop=o-r,this.list.domFocus(),this.list.setFocus([n]),this.list.setSelection([n])}onDoubleClick(e){e.browserEvent.target.classList.contains("monaco-tl-twistie")||!this.tree.expandOnDoubleClick||e.browserEvent.isHandledByList||super.onDoubleClick(e)}onMouseDown(e){const t=e.browserEvent.target;if(!MS(t)&&!G1(t)){super.onMouseDown(e);return}}onContextMenu(e){const t=e.browserEvent.target;if(!MS(t)&&!G1(t)){super.onContextMenu(e);return}}}class MNe extends Tr{constructor(e,t,i,n,o,r,a,l){super(e,t,i,n,l),this.focusTrait=o,this.selectionTrait=r,this.anchorTrait=a}createMouseController(e){return new ANe(this,e.tree,e.stickyScrollProvider)}splice(e,t,i=[]){if(super.splice(e,t,i),i.length===0)return;const n=[],o=[];let r;i.forEach((a,l)=>{this.focusTrait.has(a)&&n.push(e+l),this.selectionTrait.has(a)&&o.push(e+l),this.anchorTrait.has(a)&&(r=e+l)}),n.length>0&&super.setFocus(fh([...super.getFocus(),...n])),o.length>0&&super.setSelection(fh([...super.getSelection(),...o])),typeof r=="number"&&super.setAnchor(r)}setFocus(e,t,i=!1){super.setFocus(e,t),i||this.focusTrait.set(e.map(n=>this.element(n)),t)}setSelection(e,t,i=!1){super.setSelection(e,t),i||this.selectionTrait.set(e.map(n=>this.element(n)),t)}setAnchor(e,t=!1){super.setAnchor(e),t||(typeof e>"u"?this.anchorTrait.set([]):this.anchorTrait.set([this.element(e)]))}}class dse{get onDidScroll(){return this.view.onDidScroll}get onDidChangeFocus(){return this.eventBufferer.wrapEvent(this.focus.onDidChange)}get onDidChangeSelection(){return this.eventBufferer.wrapEvent(this.selection.onDidChange)}get onMouseDblClick(){return oe.filter(oe.map(this.view.onMouseDblClick,PL),e=>e.target!==Jf.Filter)}get onMouseOver(){return oe.map(this.view.onMouseOver,PL)}get onMouseOut(){return oe.map(this.view.onMouseOut,PL)}get onContextMenu(){return oe.any(oe.filter(oe.map(this.view.onContextMenu,RNe),e=>!e.isStickyScroll),this.stickyScrollController?.onContextMenu??oe.None)}get onPointer(){return oe.map(this.view.onPointer,PL)}get onKeyDown(){return this.view.onKeyDown}get onDidFocus(){return this.view.onDidFocus}get onDidChangeModel(){return oe.any(this.onDidChangeModelRelay.event,this.onDidSwapModel.event)}get onDidChangeCollapseState(){return this.onDidChangeCollapseStateRelay.event}get expandOnDoubleClick(){return typeof this._options.expandOnDoubleClick>"u"?!0:this._options.expandOnDoubleClick}get expandOnlyOnTwistieClick(){return typeof this._options.expandOnlyOnTwistieClick>"u"?!0:this._options.expandOnlyOnTwistieClick}get onDidDispose(){return this.view.onDidDispose}constructor(e,t,i,n,o={}){this._user=e,this._options=o,this.eventBufferer=new $0,this.onDidChangeFindOpenState=oe.None,this.onDidChangeStickyScrollFocused=oe.None,this.disposables=new U,this.onDidSwapModel=this.disposables.add(new M),this.onDidChangeModelRelay=this.disposables.add(new yC),this.onDidSpliceModelRelay=this.disposables.add(new yC),this.onDidChangeCollapseStateRelay=this.disposables.add(new yC),this.onDidChangeRenderNodeCountRelay=this.disposables.add(new yC),this.onDidChangeActiveNodesRelay=this.disposables.add(new yC),this._onWillRefilter=new M,this.onWillRefilter=this._onWillRefilter.event,this._onDidUpdateOptions=new M,this.modelDisposables=new U,o.keyboardNavigationLabelProvider&&(o.findWidgetEnabled??!0)&&(this.findFilter=new ase(o.keyboardNavigationLabelProvider,o.filter,o.defaultFindVisibility),o={...o,filter:this.findFilter},this.disposables.add(this.findFilter)),this.model=this.createModel(e,o),this.treeDelegate=new mH(i);const r=this.disposables.add(new SNe(this.onDidChangeActiveNodesRelay.event)),a=new zQ;this.renderers=n.map(l=>new m5(l,this.model,this.onDidChangeCollapseStateRelay.event,r,a,o));for(const l of this.renderers)this.disposables.add(l);if(this.focus=new tO(()=>this.view.getFocusedElements()[0],o.identityProvider),this.selection=new tO(()=>this.view.getSelectedElements()[0],o.identityProvider),this.anchor=new tO(()=>this.view.getAnchorElement(),o.identityProvider),this.view=new MNe(e,t,this.treeDelegate,this.renderers,this.focus,this.selection,this.anchor,{...yNe(()=>this.model,this.disposables,o),tree:this,stickyScrollProvider:()=>this.stickyScrollController}),this.setupModel(this.model),o.keyboardSupport!==!1){const l=oe.chain(this.view.onKeyDown,c=>c.filter(d=>!wl(d.target)).map(d=>new Pt(d)));oe.chain(l,c=>c.filter(d=>d.keyCode===15))(this.onLeftArrow,this,this.disposables),oe.chain(l,c=>c.filter(d=>d.keyCode===17))(this.onRightArrow,this,this.disposables),oe.chain(l,c=>c.filter(d=>d.keyCode===10))(this.onSpace,this,this.disposables)}if((o.findWidgetEnabled??!0)&&o.keyboardNavigationLabelProvider&&o.contextViewProvider){const l={styles:o.findWidgetStyles,defaultFindMode:o.defaultFindMode,defaultFindMatchType:o.defaultFindMatchType,showNotFoundMessage:o.showNotFoundMessage};this.findController=this.disposables.add(new lse(this,this.findFilter,o.contextViewProvider,l)),this.focusNavigationFilter=c=>this.findController.shouldAllowFocus(c),this.onDidChangeFindOpenState=this.findController.onDidChangeOpenState,this.onDidChangeFindMode=this.findController.onDidChangeMode,this.onDidChangeFindMatchType=this.findController.onDidChangeMatchType}else this.onDidChangeFindMode=oe.None,this.onDidChangeFindMatchType=oe.None;o.enableStickyScroll&&(this.stickyScrollController=new Uj(this,this.model,this.view,this.renderers,this.treeDelegate,o),this.onDidChangeStickyScrollFocused=this.stickyScrollController.onDidChangeHasFocus),this.styleElement=ca(this.view.getHTMLElement()),this.getHTMLElement().classList.toggle("always",this._options.renderIndentGuides===Gm.Always)}updateOptions(e={}){this._options={...this._options,...e};for(const t of this.renderers)t.updateOptions(e);this.view.updateOptions(this._options),this.findController?.updateOptions(e),this.updateStickyScroll(e),this._onDidUpdateOptions.fire(this._options),this.getHTMLElement().classList.toggle("always",this._options.renderIndentGuides===Gm.Always)}get options(){return this._options}updateStickyScroll(e){!this.stickyScrollController&&this._options.enableStickyScroll?(this.stickyScrollController=new Uj(this,this.model,this.view,this.renderers,this.treeDelegate,this._options),this.onDidChangeStickyScrollFocused=this.stickyScrollController.onDidChangeHasFocus):this.stickyScrollController&&!this._options.enableStickyScroll&&(this.onDidChangeStickyScrollFocused=oe.None,this.stickyScrollController.dispose(),this.stickyScrollController=void 0),this.stickyScrollController?.updateOptions(e)}getHTMLElement(){return this.view.getHTMLElement()}get onDidChangeContentHeight(){return this.view.onDidChangeContentHeight}get scrollTop(){return this.view.scrollTop}set scrollTop(e){this.view.scrollTop=e}get scrollHeight(){return this.view.scrollHeight}get renderHeight(){return this.view.renderHeight}get ariaLabel(){return this.view.ariaLabel}set ariaLabel(e){this.view.ariaLabel=e}domFocus(){this.stickyScrollController?.focusedLast()?this.stickyScrollController.domFocus():this.view.domFocus()}layout(e,t){this.view.layout(e,t)}style(e){const t=`.${this.view.domId}`,i=[];e.treeIndentGuidesStroke&&(i.push(`.monaco-list${t}:hover .monaco-tl-indent > .indent-guide, .monaco-list${t}.always .monaco-tl-indent > .indent-guide  { opacity: 1; border-color: ${e.treeInactiveIndentGuidesStroke}; }`),i.push(`.monaco-list${t} .monaco-tl-indent > .indent-guide.active { opacity: 1; border-color: ${e.treeIndentGuidesStroke}; }`));const n=e.treeStickyScrollBackground??e.listBackground;n&&(i.push(`.monaco-list${t} .monaco-scrollable-element .monaco-tree-sticky-container { background-color: ${n}; }`),i.push(`.monaco-list${t} .monaco-scrollable-element .monaco-tree-sticky-container .monaco-tree-sticky-row { background-color: ${n}; }`)),e.treeStickyScrollBorder&&i.push(`.monaco-list${t} .monaco-scrollable-element .monaco-tree-sticky-container { border-bottom: 1px solid ${e.treeStickyScrollBorder}; }`),e.treeStickyScrollShadow&&i.push(`.monaco-list${t} .monaco-scrollable-element .monaco-tree-sticky-container .monaco-tree-sticky-container-shadow { box-shadow: ${e.treeStickyScrollShadow} 0 6px 6px -6px inset; height: 3px; }`),e.listFocusForeground&&(i.push(`.monaco-list${t}.sticky-scroll-focused .monaco-scrollable-element .monaco-tree-sticky-container:focus .monaco-list-row.focused { color: ${e.listFocusForeground}; }`),i.push(`.monaco-list${t}:not(.sticky-scroll-focused) .monaco-scrollable-element .monaco-tree-sticky-container .monaco-list-row.focused { color: inherit; }`));const o=sh(e.listFocusAndSelectionOutline,sh(e.listSelectionOutline,e.listFocusOutline??""));o&&(i.push(`.monaco-list${t}.sticky-scroll-focused .monaco-scrollable-element .monaco-tree-sticky-container:focus .monaco-list-row.focused.selected { outline: 1px solid ${o}; outline-offset: -1px;}`),i.push(`.monaco-list${t}:not(.sticky-scroll-focused) .monaco-scrollable-element .monaco-tree-sticky-container .monaco-list-row.focused.selected { outline: inherit;}`)),e.listFocusOutline&&(i.push(`.monaco-list${t}.sticky-scroll-focused .monaco-scrollable-element .monaco-tree-sticky-container:focus .monaco-list-row.focused { outline: 1px solid ${e.listFocusOutline}; outline-offset: -1px; }`),i.push(`.monaco-list${t}:not(.sticky-scroll-focused) .monaco-scrollable-element .monaco-tree-sticky-container .monaco-list-row.focused { outline: inherit; }`),i.push(`.context-menu-visible .monaco-list${t}.last-focused.sticky-scroll-focused .monaco-scrollable-element .monaco-tree-sticky-container .monaco-list-row.passive-focused { outline: 1px solid ${e.listFocusOutline}; outline-offset: -1px; }`),i.push(`.context-menu-visible .monaco-list${t}.last-focused.sticky-scroll-focused .monaco-list-rows .monaco-list-row.focused { outline: inherit; }`),i.push(`.context-menu-visible .monaco-list${t}.last-focused:not(.sticky-scroll-focused) .monaco-tree-sticky-container .monaco-list-rows .monaco-list-row.focused { outline: inherit; }`)),this.styleElement.textContent=i.join(`
`),this.view.style(e)}getParentElement(e){const t=this.model.getParentNodeLocation(e);return this.model.getNode(t).element}getFirstElementChild(e){return this.model.getFirstElementChild(e)}getNode(e){return this.model.getNode(e)}getNodeLocation(e){return this.model.getNodeLocation(e)}collapse(e,t=!1){return this.model.setCollapsed(e,!0,t)}expand(e,t=!1){return this.model.setCollapsed(e,!1,t)}toggleCollapsed(e,t=!1){return this.model.setCollapsed(e,void 0,t)}isCollapsible(e){return this.model.isCollapsible(e)}setCollapsible(e,t){return this.model.setCollapsible(e,t)}isCollapsed(e){return this.model.isCollapsed(e)}refilter(){this._onWillRefilter.fire(void 0),this.model.refilter()}setSelection(e,t){this.eventBufferer.bufferEvents(()=>{const i=e.map(o=>this.model.getNode(o));this.selection.set(i,t);const n=e.map(o=>this.model.getListIndex(o)).filter(o=>o>-1);this.view.setSelection(n,t,!0)})}getSelection(){return this.selection.get()}setFocus(e,t){this.eventBufferer.bufferEvents(()=>{const i=e.map(o=>this.model.getNode(o));this.focus.set(i,t);const n=e.map(o=>this.model.getListIndex(o)).filter(o=>o>-1);this.view.setFocus(n,t,!0)})}focusNext(e=1,t=!1,i,n=Id(i)&&i.altKey?void 0:this.focusNavigationFilter){this.view.focusNext(e,t,i,n)}focusPrevious(e=1,t=!1,i,n=Id(i)&&i.altKey?void 0:this.focusNavigationFilter){this.view.focusPrevious(e,t,i,n)}focusNextPage(e,t=Id(e)&&e.altKey?void 0:this.focusNavigationFilter){return this.view.focusNextPage(e,t)}focusPreviousPage(e,t=Id(e)&&e.altKey?void 0:this.focusNavigationFilter){return this.view.focusPreviousPage(e,t,()=>this.stickyScrollController?.height??0)}focusLast(e,t=Id(e)&&e.altKey?void 0:this.focusNavigationFilter){this.view.focusLast(e,t)}focusFirst(e,t=Id(e)&&e.altKey?void 0:this.focusNavigationFilter){this.view.focusFirst(e,t)}getFocus(){return this.focus.get()}reveal(e,t){this.model.expandTo(e);const i=this.model.getListIndex(e);if(i!==-1)if(!this.stickyScrollController)this.view.reveal(i,t);else{const n=this.stickyScrollController.nodePositionTopBelowWidget(this.getNode(e));this.view.reveal(i,t,n)}}onLeftArrow(e){e.preventDefault(),e.stopPropagation();const t=this.view.getFocusedElements();if(t.length===0)return;const i=t[0],n=this.model.getNodeLocation(i);if(!this.model.setCollapsed(n,!0)){const r=this.model.getParentNodeLocation(n);if(!r)return;const a=this.model.getListIndex(r);this.view.reveal(a),this.view.setFocus([a])}}onRightArrow(e){e.preventDefault(),e.stopPropagation();const t=this.view.getFocusedElements();if(t.length===0)return;const i=t[0],n=this.model.getNodeLocation(i);if(!this.model.setCollapsed(n,!1)){if(!i.children.some(l=>l.visible))return;const[r]=this.view.getFocus(),a=r+1;this.view.reveal(a),this.view.setFocus([a])}}onSpace(e){e.preventDefault(),e.stopPropagation();const t=this.view.getFocusedElements();if(t.length===0)return;const i=t[0],n=this.model.getNodeLocation(i),o=e.browserEvent.altKey;this.model.setCollapsed(n,void 0,o)}setupModel(e){this.modelDisposables.clear(),this.modelDisposables.add(e.onDidSpliceRenderedNodes(({start:o,deleteCount:r,elements:a})=>this.view.splice(o,r,a)));const t=oe.forEach(e.onDidSpliceModel,o=>{this.eventBufferer.bufferEvents(()=>{this.focus.onDidModelSplice(o),this.selection.onDidModelSplice(o)})},this.modelDisposables);t(()=>null,null,this.modelDisposables);const i=this.modelDisposables.add(new M),n=this.modelDisposables.add(new sa(0));this.modelDisposables.add(oe.any(t,this.focus.onDidChange,this.selection.onDidChange)(()=>{n.trigger(()=>{const o=new Set;for(const r of this.focus.getNodes())o.add(r);for(const r of this.selection.getNodes())o.add(r);i.fire([...o.values()])})})),this.onDidChangeActiveNodesRelay.input=i.event,this.onDidChangeModelRelay.input=oe.signal(e.onDidSpliceModel),this.onDidChangeCollapseStateRelay.input=e.onDidChangeCollapseState,this.onDidChangeRenderNodeCountRelay.input=e.onDidChangeRenderNodeCount,this.onDidSpliceModelRelay.input=e.onDidSpliceModel}dispose(){xt(this.disposables),this.stickyScrollController?.dispose(),this.view.dispose(),this.modelDisposables.dispose()}}class _H{constructor(e,t={}){this.user=e,this.rootRef=null,this.nodes=new Map,this.nodesByIdentity=new Map,this.model=new wNe(e,null,t),this.onDidSpliceModel=this.model.onDidSpliceModel,this.onDidSpliceRenderedNodes=this.model.onDidSpliceRenderedNodes,this.onDidChangeCollapseState=this.model.onDidChangeCollapseState,this.onDidChangeRenderNodeCount=this.model.onDidChangeRenderNodeCount,t.sorter&&(this.sorter={compare(i,n){return t.sorter.compare(i.element,n.element)}}),this.identityProvider=t.identityProvider}setChildren(e,t=ot.empty(),i={}){const n=this.getElementLocation(e);this._setChildren(n,this.preserveCollapseState(t),i)}_setChildren(e,t=ot.empty(),i){const n=new Set,o=new Set,r=l=>{if(l.element===null)return;const c=l;if(n.add(c.element),this.nodes.set(c.element,c),this.identityProvider){const d=this.identityProvider.getId(c.element).toString();o.add(d),this.nodesByIdentity.set(d,c)}i.onDidCreateNode?.(c)},a=l=>{if(l.element===null)return;const c=l;if(n.has(c.element)||this.nodes.delete(c.element),this.identityProvider){const d=this.identityProvider.getId(c.element).toString();o.has(d)||this.nodesByIdentity.delete(d)}i.onDidDeleteNode?.(c)};this.model.splice([...e,0],Number.MAX_VALUE,t,{...i,onDidCreateNode:r,onDidDeleteNode:a})}preserveCollapseState(e=ot.empty()){return this.sorter&&(e=[...e].sort(this.sorter.compare.bind(this.sorter))),ot.map(e,t=>{let i=this.nodes.get(t.element);if(!i&&this.identityProvider){const r=this.identityProvider.getId(t.element).toString();i=this.nodesByIdentity.get(r)}if(!i){let r;return typeof t.collapsed>"u"?r=void 0:t.collapsed===ur.Collapsed||t.collapsed===ur.PreserveOrCollapsed?r=!0:t.collapsed===ur.Expanded||t.collapsed===ur.PreserveOrExpanded?r=!1:r=!!t.collapsed,{...t,children:this.preserveCollapseState(t.children),collapsed:r}}const n=typeof t.collapsible=="boolean"?t.collapsible:i.collapsible;let o;return typeof t.collapsed>"u"||t.collapsed===ur.PreserveOrCollapsed||t.collapsed===ur.PreserveOrExpanded?o=i.collapsed:t.collapsed===ur.Collapsed?o=!0:t.collapsed===ur.Expanded?o=!1:o=!!t.collapsed,{...t,collapsible:n,collapsed:o,children:this.preserveCollapseState(t.children)}})}rerender(e){const t=this.getElementLocation(e);this.model.rerender(t)}resort(e=null,t=!0){if(!this.sorter)return;const i=this.getElementLocation(e),n=this.model.getNode(i);this._setChildren(i,this.resortChildren(n,t),{})}resortChildren(e,t,i=!0){let n=[...e.children];return(t||i)&&(n=n.sort(this.sorter.compare.bind(this.sorter))),ot.map(n,o=>({element:o.element,collapsible:o.collapsible,collapsed:o.collapsed,children:this.resortChildren(o,t,!1)}))}getFirstElementChild(e=null){const t=this.getElementLocation(e);return this.model.getFirstElementChild(t)}has(e){return this.nodes.has(e)}getListIndex(e){const t=this.getElementLocation(e);return this.model.getListIndex(t)}getListRenderCount(e){const t=this.getElementLocation(e);return this.model.getListRenderCount(t)}isCollapsible(e){const t=this.getElementLocation(e);return this.model.isCollapsible(t)}setCollapsible(e,t){const i=this.getElementLocation(e);return this.model.setCollapsible(i,t)}isCollapsed(e){const t=this.getElementLocation(e);return this.model.isCollapsed(t)}setCollapsed(e,t,i){const n=this.getElementLocation(e);return this.model.setCollapsed(n,t,i)}expandTo(e){const t=this.getElementLocation(e);this.model.expandTo(t)}refilter(){this.model.refilter()}getNode(e=null){if(e===null)return this.model.getNode(this.model.rootRef);const t=this.nodes.get(e);if(!t)throw new mr(this.user,`Tree element not found: ${e}`);return t}getNodeLocation(e){return e.element}getParentNodeLocation(e){if(e===null)throw new mr(this.user,"Invalid getParentNodeLocation call");const t=this.nodes.get(e);if(!t)throw new mr(this.user,`Tree element not found: ${e}`);const i=this.model.getNodeLocation(t),n=this.model.getParentNodeLocation(i);return this.model.getNode(n).element}getElementLocation(e){if(e===null)return[];const t=this.nodes.get(e);if(!t)throw new mr(this.user,`Tree element not found: ${e}`);return this.model.getNodeLocation(t)}}function Uk(s){const e=[s.element],t=s.incompressible||!1;return{element:{elements:e,incompressible:t},children:ot.map(ot.from(s.children),Uk),collapsible:s.collapsible,collapsed:s.collapsed}}function qk(s){const e=[s.element],t=s.incompressible||!1;let i,n;for(;[n,i]=ot.consume(ot.from(s.children),2),!(n.length!==1||n[0].incompressible);)s=n[0],e.push(s.element);return{element:{elements:e,incompressible:t},children:ot.map(ot.concat(n,i),qk),collapsible:s.collapsible,collapsed:s.collapsed}}function _5(s,e=0){let t;return e<s.element.elements.length-1?t=[_5(s,e+1)]:t=ot.map(ot.from(s.children),i=>_5(i,0)),e===0&&s.element.incompressible?{element:s.element.elements[e],children:t,incompressible:!0,collapsible:s.collapsible,collapsed:s.collapsed}:{element:s.element.elements[e],children:t,collapsible:s.collapsible,collapsed:s.collapsed}}function qj(s){return _5(s,0)}function hse(s,e,t){return s.element===e?{...s,children:t}:{...s,children:ot.map(ot.from(s.children),i=>hse(i,e,t))}}const ONe=s=>({getId(e){return e.elements.map(t=>s.getId(t).toString()).join("\0")}});class PNe{get onDidSpliceRenderedNodes(){return this.model.onDidSpliceRenderedNodes}get onDidSpliceModel(){return this.model.onDidSpliceModel}get onDidChangeCollapseState(){return this.model.onDidChangeCollapseState}get onDidChangeRenderNodeCount(){return this.model.onDidChangeRenderNodeCount}constructor(e,t={}){this.user=e,this.rootRef=null,this.nodes=new Map,this.model=new _H(e,t),this.enabled=typeof t.compressionEnabled>"u"?!0:t.compressionEnabled,this.identityProvider=t.identityProvider}setChildren(e,t=ot.empty(),i){const n=i.diffIdentityProvider&&ONe(i.diffIdentityProvider);if(e===null){const f=ot.map(t,this.enabled?qk:Uk);this._setChildren(null,f,{diffIdentityProvider:n,diffDepth:1/0});return}const o=this.nodes.get(e);if(!o)throw new mr(this.user,"Unknown compressed tree node");const r=this.model.getNode(o),a=this.model.getParentNodeLocation(o),l=this.model.getNode(a),c=qj(r),d=hse(c,e,t),h=(this.enabled?qk:Uk)(d),u=i.diffIdentityProvider?((f,p)=>i.diffIdentityProvider.getId(f)===i.diffIdentityProvider.getId(p)):void 0;if(ni(h.element.elements,r.element.elements,u)){this._setChildren(o,h.children||ot.empty(),{diffIdentityProvider:n,diffDepth:1});return}const g=l.children.map(f=>f===r?h:f);this._setChildren(l.element,g,{diffIdentityProvider:n,diffDepth:r.depth-l.depth})}isCompressionEnabled(){return this.enabled}setCompressionEnabled(e){if(e===this.enabled)return;this.enabled=e;const i=this.model.getNode().children,n=ot.map(i,qj),o=ot.map(n,e?qk:Uk);this._setChildren(null,o,{diffIdentityProvider:this.identityProvider,diffDepth:1/0})}_setChildren(e,t,i){const n=new Set,o=a=>{for(const l of a.element.elements)n.add(l),this.nodes.set(l,a.element)},r=a=>{for(const l of a.element.elements)n.has(l)||this.nodes.delete(l)};this.model.setChildren(e,t,{...i,onDidCreateNode:o,onDidDeleteNode:r})}has(e){return this.nodes.has(e)}getListIndex(e){const t=this.getCompressedNode(e);return this.model.getListIndex(t)}getListRenderCount(e){const t=this.getCompressedNode(e);return this.model.getListRenderCount(t)}getNode(e){if(typeof e>"u")return this.model.getNode();const t=this.getCompressedNode(e);return this.model.getNode(t)}getNodeLocation(e){const t=this.model.getNodeLocation(e);return t===null?null:t.elements[t.elements.length-1]}getParentNodeLocation(e){const t=this.getCompressedNode(e),i=this.model.getParentNodeLocation(t);return i===null?null:i.elements[i.elements.length-1]}getFirstElementChild(e){const t=this.getCompressedNode(e);return this.model.getFirstElementChild(t)}isCollapsible(e){const t=this.getCompressedNode(e);return this.model.isCollapsible(t)}setCollapsible(e,t){const i=this.getCompressedNode(e);return this.model.setCollapsible(i,t)}isCollapsed(e){const t=this.getCompressedNode(e);return this.model.isCollapsed(t)}setCollapsed(e,t,i){const n=this.getCompressedNode(e);return this.model.setCollapsed(n,t,i)}expandTo(e){const t=this.getCompressedNode(e);this.model.expandTo(t)}rerender(e){const t=this.getCompressedNode(e);this.model.rerender(t)}refilter(){this.model.refilter()}resort(e=null,t=!0){const i=this.getCompressedNode(e);this.model.resort(i,t)}getCompressedNode(e){if(e===null)return null;const t=this.nodes.get(e);if(!t)throw new mr(this.user,`Tree element not found: ${e}`);return t}}const FNe=s=>s[s.length-1];class bH{get element(){return this.node.element===null?null:this.unwrapper(this.node.element)}get children(){return this.node.children.map(e=>new bH(this.unwrapper,e))}get depth(){return this.node.depth}get visibleChildrenCount(){return this.node.visibleChildrenCount}get visibleChildIndex(){return this.node.visibleChildIndex}get collapsible(){return this.node.collapsible}get collapsed(){return this.node.collapsed}get visible(){return this.node.visible}get filterData(){return this.node.filterData}constructor(e,t){this.unwrapper=e,this.node=t}}function BNe(s,e){return{...e,identityProvider:e.identityProvider&&{getId(t){return e.identityProvider.getId(s(t))}},sorter:e.sorter&&{compare(t,i){return e.sorter.compare(t.elements[0],i.elements[0])}},filter:e.filter&&{filter(t,i){const n=t.elements;for(let o=0;o<n.length-1;o++){const r=e.filter.filter(n[o],i);i=jm(bx(r)?r.visibility:r)}return e.filter.filter(n[n.length-1],i)}}}}class WNe{get onDidSpliceModel(){return oe.map(this.model.onDidSpliceModel,({insertedNodes:e,deletedNodes:t})=>({insertedNodes:e.map(i=>this.nodeMapper.map(i)),deletedNodes:t.map(i=>this.nodeMapper.map(i))}))}get onDidSpliceRenderedNodes(){return oe.map(this.model.onDidSpliceRenderedNodes,({start:e,deleteCount:t,elements:i})=>({start:e,deleteCount:t,elements:i.map(n=>this.nodeMapper.map(n))}))}get onDidChangeCollapseState(){return oe.map(this.model.onDidChangeCollapseState,({node:e,deep:t})=>({node:this.nodeMapper.map(e),deep:t}))}get onDidChangeRenderNodeCount(){return oe.map(this.model.onDidChangeRenderNodeCount,e=>this.nodeMapper.map(e))}constructor(e,t={}){this.rootRef=null,this.elementMapper=t.elementMapper||FNe;const i=n=>this.elementMapper(n.elements);this.nodeMapper=new pH(n=>new bH(i,n)),this.model=new PNe(e,BNe(i,t))}setChildren(e,t=ot.empty(),i={}){this.model.setChildren(e,t,i)}isCompressionEnabled(){return this.model.isCompressionEnabled()}setCompressionEnabled(e){this.model.setCompressionEnabled(e)}has(e){return this.model.has(e)}getListIndex(e){return this.model.getListIndex(e)}getListRenderCount(e){return this.model.getListRenderCount(e)}getNode(e){return this.nodeMapper.map(this.model.getNode(e))}getNodeLocation(e){return e.element}getParentNodeLocation(e){return this.model.getParentNodeLocation(e)}getFirstElementChild(e){const t=this.model.getFirstElementChild(e);return t===null||typeof t>"u"?t:this.elementMapper(t.elements)}isCollapsible(e){return this.model.isCollapsible(e)}setCollapsible(e,t){return this.model.setCollapsible(e,t)}isCollapsed(e){return this.model.isCollapsed(e)}setCollapsed(e,t,i){return this.model.setCollapsed(e,t,i)}expandTo(e){return this.model.expandTo(e)}rerender(e){return this.model.rerender(e)}refilter(){return this.model.refilter()}resort(e=null,t=!0){return this.model.resort(e,t)}getCompressedTreeNode(e=null){return this.model.getNode(e)}}var HNe=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o};class wH extends dse{get onDidChangeCollapseState(){return this.model.onDidChangeCollapseState}constructor(e,t,i,n,o={}){super(e,t,i,n,o),this.user=e}setChildren(e,t=ot.empty(),i){this.model.setChildren(e,t,i)}rerender(e){if(e===void 0){this.view.rerender();return}this.model.rerender(e)}resort(e,t=!0){this.model.resort(e,t)}hasElement(e){return this.model.has(e)}createModel(e,t){return new _H(e,t)}}class use{get compressedTreeNodeProvider(){return this._compressedTreeNodeProvider()}constructor(e,t,i){this._compressedTreeNodeProvider=e,this.stickyScrollDelegate=t,this.renderer=i,this.templateId=i.templateId,i.onDidChangeTwistieState&&(this.onDidChangeTwistieState=i.onDidChangeTwistieState)}renderTemplate(e){return{compressedTreeNode:void 0,data:this.renderer.renderTemplate(e)}}renderElement(e,t,i,n){let o=this.stickyScrollDelegate.getCompressedNode(e);o||(o=this.compressedTreeNodeProvider.getCompressedTreeNode(e.element)),o.element.elements.length===1?(i.compressedTreeNode=void 0,this.renderer.renderElement(e,t,i.data,n)):(i.compressedTreeNode=o,this.renderer.renderCompressedElements(o,t,i.data,n))}disposeElement(e,t,i,n){i.compressedTreeNode?this.renderer.disposeCompressedElements?.(i.compressedTreeNode,t,i.data,n):this.renderer.disposeElement?.(e,t,i.data,n)}disposeTemplate(e){this.renderer.disposeTemplate(e.data)}renderTwistie(e,t){return this.renderer.renderTwistie?.(e,t)??!1}}HNe([Ai],use.prototype,"compressedTreeNodeProvider",null);class VNe{constructor(e){this.modelProvider=e,this.compressedStickyNodes=new Map}getCompressedNode(e){return this.compressedStickyNodes.get(e)}constrainStickyScrollNodes(e,t,i){if(this.compressedStickyNodes.clear(),e.length===0)return[];for(let n=0;n<e.length;n++){const o=e[n],r=o.position+o.height;if(n+1<e.length&&r+e[n+1].height>i||n>=t-1&&t<e.length){const l=e.slice(0,n),c=e.slice(n),d=this.compressStickyNodes(c);return[...l,d]}}return e}compressStickyNodes(e){if(e.length===0)throw new Error("Can't compress empty sticky nodes");const t=this.modelProvider();if(!t.isCompressionEnabled())return e[0];const i=[];for(let c=0;c<e.length;c++){const d=e[c],h=t.getCompressedTreeNode(d.node.element);if(h.element){if(c!==0&&h.element.incompressible)break;i.push(...h.element.elements)}}if(i.length<2)return e[0];const n=e[e.length-1],o={elements:i,incompressible:!1},r={...n.node,children:[],element:o},a=new Proxy(e[0].node,{}),l={node:a,startIndex:e[0].startIndex,endIndex:n.endIndex,position:e[0].position,height:e[0].height};return this.compressedStickyNodes.set(a,r),l}}function zNe(s,e){return e&&{...e,keyboardNavigationLabelProvider:e.keyboardNavigationLabelProvider&&{getKeyboardNavigationLabel(t){let i;try{i=s().getCompressedTreeNode(t)}catch{return e.keyboardNavigationLabelProvider.getKeyboardNavigationLabel(t)}return i.element.elements.length===1?e.keyboardNavigationLabelProvider.getKeyboardNavigationLabel(t):e.keyboardNavigationLabelProvider.getCompressedNodeKeyboardNavigationLabel(i.element.elements)}}}}class gse extends wH{constructor(e,t,i,n,o={}){const r=()=>this,a=new VNe(()=>this.model),l=n.map(c=>new use(r,a,c));super(e,t,i,l,{...zNe(r,o),stickyScrollDelegate:a})}setChildren(e,t=ot.empty(),i){this.model.setChildren(e,t,i)}createModel(e,t){return new WNe(e,t)}updateOptions(e={}){super.updateOptions(e),typeof e.compressionEnabled<"u"&&this.model.setCompressionEnabled(e.compressionEnabled)}getCompressedTreeNode(e=null){return this.model.getCompressedTreeNode(e)}}function iO(s){return{...s,children:[],refreshPromise:void 0,stale:!0,slow:!1,forceExpanded:!1}}function b5(s,e){return e.parent?e.parent===s?!0:b5(s,e.parent):!1}function $Ne(s,e){return s===e||b5(s,e)||b5(e,s)}class vH{get element(){return this.node.element.element}get children(){return this.node.children.map(e=>new vH(e))}get depth(){return this.node.depth}get visibleChildrenCount(){return this.node.visibleChildrenCount}get visibleChildIndex(){return this.node.visibleChildIndex}get collapsible(){return this.node.collapsible}get collapsed(){return this.node.collapsed}get visible(){return this.node.visible}get filterData(){return this.node.filterData}constructor(e){this.node=e}}class UNe{constructor(e,t,i){this.renderer=e,this.nodeMapper=t,this.onDidChangeTwistieState=i,this.renderedNodes=new Map,this.templateId=e.templateId}renderTemplate(e){return{templateData:this.renderer.renderTemplate(e)}}renderElement(e,t,i,n){this.renderer.renderElement(this.nodeMapper.map(e),t,i.templateData,n)}renderTwistie(e,t){return e.slow?(t.classList.add(...xe.asClassNameArray(Z.treeItemLoading)),!0):(t.classList.remove(...xe.asClassNameArray(Z.treeItemLoading)),!1)}disposeElement(e,t,i,n){this.renderer.disposeElement?.(this.nodeMapper.map(e),t,i.templateData,n)}disposeTemplate(e){this.renderer.disposeTemplate(e.templateData)}dispose(){this.renderedNodes.clear()}}function jj(s){return{browserEvent:s.browserEvent,elements:s.elements.map(e=>e.element)}}function Gj(s){return{browserEvent:s.browserEvent,element:s.element&&s.element.element,target:s.target}}class qNe extends cx{constructor(e){super(e.elements.map(t=>t.element)),this.data=e}}function nO(s){return s instanceof cx?new qNe(s):s}class jNe{constructor(e){this.dnd=e}getDragURI(e){return this.dnd.getDragURI(e.element)}getDragLabel(e,t){if(this.dnd.getDragLabel)return this.dnd.getDragLabel(e.map(i=>i.element),t)}onDragStart(e,t){this.dnd.onDragStart?.(nO(e),t)}onDragOver(e,t,i,n,o,r=!0){return this.dnd.onDragOver(nO(e),t&&t.element,i,n,o)}drop(e,t,i,n,o){this.dnd.drop(nO(e),t&&t.element,i,n,o)}onDragEnd(e){this.dnd.onDragEnd?.(e)}dispose(){this.dnd.dispose()}}class GNe extends ase{constructor(e,t,i){super(t,i),this.findProvider=e,this.isFindSessionActive=!1}filter(e,t){const i=super.filter(e,t);if(!this.isFindSessionActive||this.findMode===_r.Highlight||!this.findProvider.isVisible)return i;const n=bx(i)?i.visibility:i;return jm(n)===0?0:this.findProvider.isVisible(e)?i:0}}class KNe extends lse{constructor(e,t,i,n,o){super(e,i,n,o),this.findProvider=t,this.filter=i,this.activeSession=!1,this.asyncWorkInProgress=!1,this.disposables.add(pe(async()=>{this.activeSession&&await this.findProvider.endSession?.()}))}render(){if(this.asyncWorkInProgress||!this.activeFindMetadata)return;const e=this.activeFindMetadata.matchCount===0&&this.pattern.length>0;this.renderMessage(e),this.pattern.length&&this.alertResults(this.activeFindMetadata.matchCount)}shouldAllowFocus(e){return this.shouldFocusWhenNavigating(e)}shouldFocusWhenNavigating(e){if(!this.activeSession||!this.activeFindMetadata)return!0;const t=e.element?.element;return t&&this.activeFindMetadata.isMatch(t)?!0:!Oa.isDefault(e.filterData)}}function fse(s){return s&&{...s,collapseByDefault:!0,identityProvider:s.identityProvider&&{getId(e){return s.identityProvider.getId(e.element)}},dnd:s.dnd&&new jNe(s.dnd),multipleSelectionController:s.multipleSelectionController&&{isSelectionSingleChangeEvent(e){return s.multipleSelectionController.isSelectionSingleChangeEvent({...e,element:e.element})},isSelectionRangeChangeEvent(e){return s.multipleSelectionController.isSelectionRangeChangeEvent({...e,element:e.element})}},accessibilityProvider:s.accessibilityProvider&&{...s.accessibilityProvider,getPosInSet:void 0,getSetSize:void 0,getRole:s.accessibilityProvider.getRole?e=>s.accessibilityProvider.getRole(e.element):()=>"treeitem",isChecked:s.accessibilityProvider.isChecked?e=>!!s.accessibilityProvider?.isChecked(e.element):void 0,getAriaLabel(e){return s.accessibilityProvider.getAriaLabel(e.element)},getWidgetAriaLabel(){return s.accessibilityProvider.getWidgetAriaLabel()},getWidgetRole:s.accessibilityProvider.getWidgetRole?()=>s.accessibilityProvider.getWidgetRole():()=>"tree",getAriaLevel:s.accessibilityProvider.getAriaLevel&&(e=>s.accessibilityProvider.getAriaLevel(e.element)),getActiveDescendantId:s.accessibilityProvider.getActiveDescendantId&&(e=>s.accessibilityProvider.getActiveDescendantId(e.element))},filter:s.filter&&{filter(e,t){return s.filter.filter(e.element,t)}},keyboardNavigationLabelProvider:s.keyboardNavigationLabelProvider&&{...s.keyboardNavigationLabelProvider,getKeyboardNavigationLabel(e){return s.keyboardNavigationLabelProvider.getKeyboardNavigationLabel(e.element)}},sorter:void 0,expandOnlyOnTwistieClick:typeof s.expandOnlyOnTwistieClick>"u"?void 0:typeof s.expandOnlyOnTwistieClick!="function"?s.expandOnlyOnTwistieClick:(e=>s.expandOnlyOnTwistieClick(e.element)),defaultFindVisibility:e=>e.hasChildren&&e.stale?1:typeof s.defaultFindVisibility=="number"?s.defaultFindVisibility:typeof s.defaultFindVisibility>"u"?2:s.defaultFindVisibility(e.element),stickyScrollDelegate:s.stickyScrollDelegate}}function w5(s,e){e(s),s.children.forEach(t=>w5(t,e))}class pse{get onDidScroll(){return this.tree.onDidScroll}get onDidChangeFocus(){return oe.map(this.tree.onDidChangeFocus,jj)}get onDidChangeSelection(){return oe.map(this.tree.onDidChangeSelection,jj)}get onMouseDblClick(){return oe.map(this.tree.onMouseDblClick,Gj)}get onPointer(){return oe.map(this.tree.onPointer,Gj)}get onDidFocus(){return this.tree.onDidFocus}get onDidChangeModel(){return this.tree.onDidChangeModel}get onDidChangeCollapseState(){return this.tree.onDidChangeCollapseState}get onDidChangeStickyScrollFocused(){return this.tree.onDidChangeStickyScrollFocused}get onDidDispose(){return this.tree.onDidDispose}constructor(e,t,i,n,o,r={}){this.user=e,this.dataSource=o,this.nodes=new Map,this.subTreeRefreshPromises=new Map,this.refreshPromises=new Map,this._onDidRender=new M,this._onDidChangeNodeSlowState=new M,this.nodeMapper=new pH(c=>new vH(c)),this.disposables=new U,this.identityProvider=r.identityProvider,this.autoExpandSingleChildren=typeof r.autoExpandSingleChildren>"u"?!1:r.autoExpandSingleChildren,this.sorter=r.sorter,this.getDefaultCollapseState=c=>r.collapseByDefault?r.collapseByDefault(c)?ur.PreserveOrCollapsed:ur.PreserveOrExpanded:void 0;let a=!1,l;if(r.findProvider&&(r.findWidgetEnabled??!0)&&r.keyboardNavigationLabelProvider&&r.contextViewProvider&&(a=!0,l=new GNe(r.findProvider,r.keyboardNavigationLabelProvider,r.filter)),this.tree=this.createTree(e,t,i,n,{...r,findWidgetEnabled:!a,filter:l??r.filter}),this.root=iO({element:void 0,parent:null,hasChildren:!0,defaultCollapseState:void 0}),this.identityProvider&&(this.root={...this.root,id:null}),this.nodes.set(null,this.root),this.tree.onDidChangeCollapseState(this._onDidChangeCollapseState,this,this.disposables),a){const c={styles:r.findWidgetStyles,showNotFoundMessage:r.showNotFoundMessage,defaultFindMatchType:r.defaultFindMatchType,defaultFindMode:r.defaultFindMode};this.findController=this.disposables.add(new KNe(this.tree,r.findProvider,l,this.tree.options.contextViewProvider,c)),this.focusNavigationFilter=d=>this.findController.shouldFocusWhenNavigating(d),this.onDidChangeFindOpenState=this.findController.onDidChangeOpenState,this.onDidChangeFindMode=this.findController.onDidChangeMode,this.onDidChangeFindMatchType=this.findController.onDidChangeMatchType}else this.onDidChangeFindOpenState=this.tree.onDidChangeFindOpenState,this.onDidChangeFindMode=this.tree.onDidChangeFindMode,this.onDidChangeFindMatchType=this.tree.onDidChangeFindMatchType}createTree(e,t,i,n,o){const r=new mH(i),a=n.map(c=>new UNe(c,this.nodeMapper,this._onDidChangeNodeSlowState.event)),l=fse(o)||{};return new wH(e,t,r,a,l)}updateOptions(e={}){this.findController&&(e.defaultFindMode!==void 0&&(this.findController.mode=e.defaultFindMode),e.defaultFindMatchType!==void 0&&(this.findController.matchType=e.defaultFindMatchType)),this.tree.updateOptions(e)}getHTMLElement(){return this.tree.getHTMLElement()}get scrollTop(){return this.tree.scrollTop}set scrollTop(e){this.tree.scrollTop=e}get scrollHeight(){return this.tree.scrollHeight}get renderHeight(){return this.tree.renderHeight}domFocus(){this.tree.domFocus()}layout(e,t){this.tree.layout(e,t)}style(e){this.tree.style(e)}getInput(){return this.root.element}async setInput(e,t){this.cancelAllRefreshPromises(),this.root.element=e;const i=t&&{viewState:t,focus:[],selection:[]};await this._updateChildren(e,!0,!1,i),i&&(this.tree.setFocus(i.focus),this.tree.setSelection(i.selection)),t&&typeof t.scrollTop=="number"&&(this.scrollTop=t.scrollTop)}cancelAllRefreshPromises(e=!1){this.refreshPromises.forEach(t=>t.cancel()),this.refreshPromises.clear(),e&&(this.subTreeRefreshPromises.forEach(t=>t.cancel()),this.subTreeRefreshPromises.clear())}async _updateChildren(e=this.root.element,t=!0,i=!1,n,o){if(typeof this.root.element>"u")throw new mr(this.user,"Tree input not set");this.root.refreshPromise&&(await this.root.refreshPromise,await oe.toPromise(this._onDidRender.event));const r=this.getDataNode(e);if(await this.refreshAndRenderNode(r,t,n,o),i)try{this.tree.rerender(r)}catch{}}rerender(e){if(e===void 0||e===this.root.element){this.tree.rerender();return}const t=this.getDataNode(e);this.tree.rerender(t)}getNode(e=this.root.element){const t=this.getDataNode(e),i=this.tree.getNode(t===this.root?null:t);return this.nodeMapper.map(i)}collapse(e,t=!1){const i=this.getDataNode(e);return this.tree.collapse(i===this.root?null:i,t)}async expand(e,t=!1){if(typeof this.root.element>"u")throw new mr(this.user,"Tree input not set");this.root.refreshPromise&&(await this.root.refreshPromise,await oe.toPromise(this._onDidRender.event));const i=this.getDataNode(e);if(this.tree.hasElement(i)&&!this.tree.isCollapsible(i)||(i.refreshPromise&&(await i.refreshPromise,await oe.toPromise(this._onDidRender.event)),i!==this.root&&!i.refreshPromise&&!this.tree.isCollapsed(i)))return!1;const n=this.tree.expand(i===this.root?null:i,t);return i.refreshPromise&&(await i.refreshPromise,await oe.toPromise(this._onDidRender.event)),n}setSelection(e,t){const i=e.map(n=>this.getDataNode(n));this.tree.setSelection(i,t)}getSelection(){return this.tree.getSelection().map(t=>t.element)}setFocus(e,t){const i=e.map(n=>this.getDataNode(n));this.tree.setFocus(i,t)}getFocus(){return this.tree.getFocus().map(t=>t.element)}reveal(e,t){this.tree.reveal(this.getDataNode(e),t)}getParentElement(e){const t=this.tree.getParentElement(this.getDataNode(e));return t&&t.element}getFirstElementChild(e=this.root.element){const t=this.getDataNode(e),i=this.tree.getFirstElementChild(t===this.root?null:t);return i&&i.element}getDataNode(e){const t=this.nodes.get(e===this.root.element?null:e);if(!t){const i=this.identityProvider?.getId(e).toString();throw new mr(this.user,`Data tree node not found${i?`: ${i}`:""}`)}return t}async refreshAndRenderNode(e,t,i,n){this.disposables.isDisposed||(await this.refreshNode(e,t,i),!this.disposables.isDisposed&&this.render(e,i,n))}async refreshNode(e,t,i){let n;if(this.subTreeRefreshPromises.forEach((o,r)=>{!n&&$Ne(r,e)&&(n=o.then(()=>this.refreshNode(e,t,i)))}),n)return n;if(e!==this.root&&this.tree.getNode(e).collapsed){e.hasChildren=!!this.dataSource.hasChildren(e.element),e.stale=!0,this.setChildren(e,[],t,i);return}return this.doRefreshSubTree(e,t,i)}async doRefreshSubTree(e,t,i){const n=an(async()=>{const o=await this.doRefreshNode(e,t,i);e.stale=!1,await Uy.settled(o.map(r=>this.doRefreshSubTree(r,t,i)))});return e.refreshPromise=n,this.subTreeRefreshPromises.set(e,n),n.finally(()=>{e.refreshPromise=void 0,this.subTreeRefreshPromises.delete(e)}),n}async doRefreshNode(e,t,i){e.hasChildren=!!this.dataSource.hasChildren(e.element);let n;if(!e.hasChildren)n=Promise.resolve(ot.empty());else{const o=this.doGetChildren(e);if(tP(o))n=Promise.resolve(o);else{const r=Tc(800);r.then(()=>{e.slow=!0,this._onDidChangeNodeSlowState.fire(e)},a=>null),n=o.finally(()=>r.cancel())}}try{const o=await n;return this.setChildren(e,o,t,i)}catch(o){if(e!==this.root&&this.tree.hasElement(e)&&this.tree.collapse(e),Dr(o))return[];throw o}finally{e.slow&&(e.slow=!1,this._onDidChangeNodeSlowState.fire(e))}}doGetChildren(e){let t=this.refreshPromises.get(e);if(t)return t;const i=this.dataSource.getChildren(e.element);return tP(i)?this.processChildren(i):(t=an(async()=>this.processChildren(await i)),this.refreshPromises.set(e,t),t.finally(()=>{this.refreshPromises.delete(e)}))}_onDidChangeCollapseState({node:e,deep:t}){e.element!==null&&!e.collapsed&&e.element.stale&&(t?this.collapse(e.element.element):this.refreshAndRenderNode(e.element,!1).catch(Oe))}setChildren(e,t,i,n){const o=[...t];if(e.children.length===0&&o.length===0)return[];const r=new Map,a=new Map;for(const d of e.children)r.set(d.element,d),this.identityProvider&&a.set(d.id,{node:d,collapsed:this.tree.hasElement(d)&&this.tree.isCollapsed(d)});const l=[],c=o.map(d=>{const h=!!this.dataSource.hasChildren(d);if(!this.identityProvider){const p=iO({element:d,parent:e,hasChildren:h,defaultCollapseState:this.getDefaultCollapseState(d)});return h&&p.defaultCollapseState===ur.PreserveOrExpanded&&l.push(p),p}const u=this.identityProvider.getId(d).toString(),g=a.get(u);if(g){const p=g.node;return r.delete(p.element),this.nodes.delete(p.element),this.nodes.set(d,p),p.element=d,p.hasChildren=h,i?g.collapsed?(p.children.forEach(_=>w5(_,b=>this.nodes.delete(b.element))),p.children.splice(0,p.children.length),p.stale=!0):l.push(p):h&&!g.collapsed&&l.push(p),p}const f=iO({element:d,parent:e,id:u,hasChildren:h,defaultCollapseState:this.getDefaultCollapseState(d)});return n&&n.viewState.focus&&n.viewState.focus.indexOf(u)>-1&&n.focus.push(f),n&&n.viewState.selection&&n.viewState.selection.indexOf(u)>-1&&n.selection.push(f),(n&&n.viewState.expanded&&n.viewState.expanded.indexOf(u)>-1||h&&f.defaultCollapseState===ur.PreserveOrExpanded)&&l.push(f),f});for(const d of r.values())w5(d,h=>this.nodes.delete(h.element));for(const d of c)this.nodes.set(d.element,d);return CE(e.children,0,e.children.length,c),e!==this.root&&this.autoExpandSingleChildren&&c.length===1&&l.length===0&&(c[0].forceExpanded=!0,l.push(c[0])),l}render(e,t,i){const n=e.children.map(r=>this.asTreeElement(r,t)),o=i&&{...i,diffIdentityProvider:i.diffIdentityProvider&&{getId(r){return i.diffIdentityProvider.getId(r.element)}}};this.tree.setChildren(e===this.root?null:e,n,o),e!==this.root&&this.tree.setCollapsible(e,e.hasChildren),this._onDidRender.fire()}asTreeElement(e,t){if(e.stale)return{element:e,collapsible:e.hasChildren,collapsed:!0};let i;return t&&t.viewState.expanded&&e.id&&t.viewState.expanded.indexOf(e.id)>-1?i=!1:e.forceExpanded?(i=!1,e.forceExpanded=!1):i=e.defaultCollapseState,{element:e,children:e.hasChildren?ot.map(e.children,n=>this.asTreeElement(n,t)):[],collapsible:e.hasChildren,collapsed:i}}processChildren(e){return this.sorter&&(e=[...e].sort(this.sorter.compare.bind(this.sorter))),e}dispose(){this.disposables.dispose(),this.tree.dispose()}}class CH{get element(){return{elements:this.node.element.elements.map(e=>e.element),incompressible:this.node.element.incompressible}}get children(){return this.node.children.map(e=>new CH(e))}get depth(){return this.node.depth}get visibleChildrenCount(){return this.node.visibleChildrenCount}get visibleChildIndex(){return this.node.visibleChildIndex}get collapsible(){return this.node.collapsible}get collapsed(){return this.node.collapsed}get visible(){return this.node.visible}get filterData(){return this.node.filterData}constructor(e){this.node=e}}class YNe{constructor(e,t,i,n){this.renderer=e,this.nodeMapper=t,this.compressibleNodeMapperProvider=i,this.onDidChangeTwistieState=n,this.renderedNodes=new Map,this.disposables=[],this.templateId=e.templateId}renderTemplate(e){return{templateData:this.renderer.renderTemplate(e)}}renderElement(e,t,i,n){this.renderer.renderElement(this.nodeMapper.map(e),t,i.templateData,n)}renderCompressedElements(e,t,i,n){this.renderer.renderCompressedElements(this.compressibleNodeMapperProvider().map(e),t,i.templateData,n)}renderTwistie(e,t){return e.slow?(t.classList.add(...xe.asClassNameArray(Z.treeItemLoading)),!0):(t.classList.remove(...xe.asClassNameArray(Z.treeItemLoading)),!1)}disposeElement(e,t,i,n){this.renderer.disposeElement?.(this.nodeMapper.map(e),t,i.templateData,n)}disposeCompressedElements(e,t,i,n){this.renderer.disposeCompressedElements?.(this.compressibleNodeMapperProvider().map(e),t,i.templateData,n)}disposeTemplate(e){this.renderer.disposeTemplate(e.templateData)}dispose(){this.renderedNodes.clear(),this.disposables=xt(this.disposables)}}function ZNe(s){const e=s&&fse(s);return e&&{...e,keyboardNavigationLabelProvider:e.keyboardNavigationLabelProvider&&{...e.keyboardNavigationLabelProvider,getCompressedNodeKeyboardNavigationLabel(t){return s.keyboardNavigationLabelProvider.getCompressedNodeKeyboardNavigationLabel(t.map(i=>i.element))}},stickyScrollDelegate:e.stickyScrollDelegate}}class XNe extends pse{constructor(e,t,i,n,o,r,a={}){super(e,t,i,o,r,a),this.compressionDelegate=n,this.compressibleNodeMapper=new pH(l=>new CH(l)),this.filter=a.filter}createTree(e,t,i,n,o){const r=new mH(i),a=n.map(c=>new YNe(c,this.nodeMapper,()=>this.compressibleNodeMapper,this._onDidChangeNodeSlowState.event)),l=ZNe(o)||{};return new gse(e,t,r,a,l)}asTreeElement(e,t){return{incompressible:this.compressionDelegate.isIncompressible(e.element),...super.asTreeElement(e,t)}}render(e,t,i){if(!this.identityProvider)return super.render(e,t);const n=g=>this.identityProvider.getId(g).toString(),o=g=>{const f=new Set;for(const p of g){const _=this.tree.getCompressedTreeNode(p===this.root?null:p);if(_.element)for(const b of _.element.elements)f.add(n(b.element))}return f},r=o(this.tree.getSelection()),a=o(this.tree.getFocus());super.render(e,t,i);const l=this.getSelection();let c=!1;const d=this.getFocus();let h=!1;const u=g=>{const f=g.element;if(f)for(let p=0;p<f.elements.length;p++){const _=n(f.elements[p].element),b=f.elements[f.elements.length-1].element;r.has(_)&&l.indexOf(b)===-1&&(l.push(b),c=!0),a.has(_)&&d.indexOf(b)===-1&&(d.push(b),h=!0)}g.children.forEach(u)};u(this.tree.getCompressedTreeNode(e===this.root?null:e)),c&&this.setSelection(l),h&&this.setFocus(d)}processChildren(e){return this.filter&&(e=ot.filter(e,t=>{const i=this.filter.filter(t,1),n=QNe(i);if(n===2)throw new Error("Recursive tree visibility not supported in async data compressed trees");return n===1})),super.processChildren(e)}}function QNe(s){return typeof s=="boolean"?s?1:0:bx(s)?jm(s.visibility):jm(s)}class JNe extends dse{constructor(e,t,i,n,o,r={}){super(e,t,i,n,r),this.user=e,this.dataSource=o,this.identityProvider=r.identityProvider}createModel(e,t){return new _H(e,t)}}new ce("isMac",Ye,m(1684,"Whether the operating system is macOS"));new ce("isLinux",vo,m(1685,"Whether the operating system is Linux"));const MA=new ce("isWindows",Bn,m(1686,"Whether the operating system is Windows")),mse=new ce("isWeb",Vc,m(1687,"Whether the platform is a web browser"));new ce("isMacNative",Ye&&!Vc,m(1688,"Whether the operating system is macOS on a non-browser platform"));new ce("isIOS",na,m(1689,"Whether the operating system is iOS"));new ce("isMobile",HX,m(1690,"Whether the platform is a mobile web browser"));new ce("isDevelopment",!1,!0);new ce("productQualityType","",m(1691,"Quality type of VS Code"));const _se="inputFocus",yH=new ce(_se,!1,m(1692,"Whether keyboard focus is inside an input box"));var Ph=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},Xt=function(s,e){return function(t,i){e(t,i,s)}};const fa=qe("listService");class eRe{get lastFocusedList(){return this._lastFocusedWidget}constructor(){this.disposables=new U,this.lists=[],this._lastFocusedWidget=void 0}setLastFocusedList(e){e!==this._lastFocusedWidget&&(this._lastFocusedWidget?.getHTMLElement().classList.remove("last-focused"),this._lastFocusedWidget=e,this._lastFocusedWidget?.getHTMLElement().classList.add("last-focused"))}register(e,t){if(this.lists.some(n=>n.widget===e))throw new Error("Cannot register the same widget multiple times");const i={widget:e,extraContextKeys:t};return this.lists.push(i),NR(e.getHTMLElement())&&this.setLastFocusedList(e),Ra(e.onDidFocus(()=>this.setLastFocusedList(e)),pe(()=>this.lists.splice(this.lists.indexOf(i),1)),e.onDidDispose(()=>{this.lists=this.lists.filter(n=>n!==i),this._lastFocusedWidget===e&&this.setLastFocusedList(void 0)}))}dispose(){this.disposables.dispose()}}const ZS=new ce("listScrollAtBoundary","none");K.or(ZS.isEqualTo("top"),ZS.isEqualTo("both"));K.or(ZS.isEqualTo("bottom"),ZS.isEqualTo("both"));const bse=new ce("listFocus",!0),wse=new ce("treestickyScrollFocused",!1),OA=new ce("listSupportsMultiselect",!0),vse=K.and(bse,K.not(_se),wse.negate()),SH=new ce("listHasSelectionOrFocus",!1),xH=new ce("listDoubleSelection",!1),LH=new ce("listMultiSelection",!1),PA=new ce("listSelectionNavigation",!1),tRe=new ce("listSupportsFind",!0),kH=new ce("treeElementCanCollapse",!1),iRe=new ce("treeElementHasParent",!1),EH=new ce("treeElementCanExpand",!1),nRe=new ce("treeElementHasChild",!1),sRe=new ce("treeFindOpen",!1),Cse="listTypeNavigationMode",yse="listAutomaticKeyboardNavigation";function FA(s,e){const t=s.createScoped(e.getHTMLElement());return bse.bindTo(t),t}function BA(s,e){const t=ZS.bindTo(s),i=()=>{const n=e.scrollTop===0,o=e.scrollHeight-e.renderHeight-e.scrollTop<1;n&&o?t.set("both"):n?t.set("top"):o?t.set("bottom"):t.set("none")};return i(),e.onDidScroll(i)}const b_="workbench.list.multiSelectModifier",jk="workbench.list.openMode",ta="workbench.list.horizontalScrolling",IH="workbench.list.defaultFindMode",DH="workbench.list.typeNavigationMode",nD="workbench.list.keyboardNavigation",Nl="workbench.list.scrollByPage",TH="workbench.list.defaultFindMatchType",XS="workbench.tree.indent",sD="workbench.tree.renderIndentGuides",Rl="workbench.list.smoothScrolling",Pc="workbench.list.mouseWheelScrollSensitivity",Fc="workbench.list.fastScrollSensitivity",oD="workbench.tree.expandMode",rD="workbench.tree.enableStickyScroll",aD="workbench.tree.stickyScrollMaxItemCount";function Bc(s){return s.getValue(b_)==="alt"}class oRe extends F{constructor(e){super(),this.configurationService=e,this.useAltAsMultipleSelectionModifier=Bc(e),this.registerListeners()}registerListeners(){this._register(this.configurationService.onDidChangeConfiguration(e=>{e.affectsConfiguration(b_)&&(this.useAltAsMultipleSelectionModifier=Bc(this.configurationService))}))}isSelectionSingleChangeEvent(e){return this.useAltAsMultipleSelectionModifier?e.browserEvent.altKey:aie(e)}isSelectionRangeChangeEvent(e){return lie(e)}}function WA(s,e){const t=s.get(He),i=s.get(ft),n=new U;return[{...e,keyboardNavigationDelegate:{mightProducePrintableCharacter(r){return i.mightProducePrintableCharacter(r)}},smoothScrolling:!!t.getValue(Rl),mouseWheelScrollSensitivity:t.getValue(Pc),fastScrollSensitivity:t.getValue(Fc),multipleSelectionController:e.multipleSelectionController??n.add(new oRe(t)),keyboardNavigationEventFilter:lRe(i),scrollByPage:!!t.getValue(Nl)},n]}let Kj=class extends Tr{constructor(e,t,i,n,o,r,a,l,c){const d=typeof o.horizontalScrolling<"u"?o.horizontalScrolling:!!l.getValue(ta),[h,u]=c.invokeFunction(WA,o);super(e,t,i,n,{keyboardSupport:!1,...h,horizontalScrolling:d}),this.disposables.add(u),this.contextKeyService=FA(r,this),this.disposables.add(BA(this.contextKeyService,this)),this.listSupportsMultiSelect=OA.bindTo(this.contextKeyService),this.listSupportsMultiSelect.set(o.multipleSelectionSupport!==!1),PA.bindTo(this.contextKeyService).set(!!o.selectionNavigation),this.listHasSelectionOrFocus=SH.bindTo(this.contextKeyService),this.listDoubleSelection=xH.bindTo(this.contextKeyService),this.listMultiSelection=LH.bindTo(this.contextKeyService),this.horizontalScrolling=o.horizontalScrolling,this._useAltAsMultipleSelectionModifier=Bc(l),this.disposables.add(this.contextKeyService),this.disposables.add(a.register(this)),this.updateStyles(o.overrideStyles),this.disposables.add(this.onDidChangeSelection(()=>{const f=this.getSelection(),p=this.getFocus();this.contextKeyService.bufferChangeEvents(()=>{this.listHasSelectionOrFocus.set(f.length>0||p.length>0),this.listMultiSelection.set(f.length>1),this.listDoubleSelection.set(f.length===2)})})),this.disposables.add(this.onDidChangeFocus(()=>{const f=this.getSelection(),p=this.getFocus();this.listHasSelectionOrFocus.set(f.length>0||p.length>0)})),this.disposables.add(l.onDidChangeConfiguration(f=>{f.affectsConfiguration(b_)&&(this._useAltAsMultipleSelectionModifier=Bc(l));let p={};if(f.affectsConfiguration(ta)&&this.horizontalScrolling===void 0){const _=!!l.getValue(ta);p={...p,horizontalScrolling:_}}if(f.affectsConfiguration(Nl)){const _=!!l.getValue(Nl);p={...p,scrollByPage:_}}if(f.affectsConfiguration(Rl)){const _=!!l.getValue(Rl);p={...p,smoothScrolling:_}}if(f.affectsConfiguration(Pc)){const _=l.getValue(Pc);p={...p,mouseWheelScrollSensitivity:_}}if(f.affectsConfiguration(Fc)){const _=l.getValue(Fc);p={...p,fastScrollSensitivity:_}}Object.keys(p).length>0&&this.updateOptions(p)})),this.navigator=new Sse(this,{configurationService:l,...o}),this.disposables.add(this.navigator)}updateOptions(e){super.updateOptions(e),e.overrideStyles!==void 0&&this.updateStyles(e.overrideStyles),e.multipleSelectionSupport!==void 0&&this.listSupportsMultiSelect.set(!!e.multipleSelectionSupport)}updateStyles(e){this.style(e?f_(e):lC)}};Kj=Ph([Xt(5,Ae),Xt(6,fa),Xt(7,He),Xt(8,_e)],Kj);let Yj=class extends uNe{constructor(e,t,i,n,o,r,a,l,c){const d=typeof o.horizontalScrolling<"u"?o.horizontalScrolling:!!l.getValue(ta),[h,u]=c.invokeFunction(WA,o);super(e,t,i,n,{keyboardSupport:!1,...h,horizontalScrolling:d}),this.disposables=new U,this.disposables.add(u),this.contextKeyService=FA(r,this),this.disposables.add(BA(this.contextKeyService,this.widget)),this.horizontalScrolling=o.horizontalScrolling,this.listSupportsMultiSelect=OA.bindTo(this.contextKeyService),this.listSupportsMultiSelect.set(o.multipleSelectionSupport!==!1),PA.bindTo(this.contextKeyService).set(!!o.selectionNavigation),this._useAltAsMultipleSelectionModifier=Bc(l),this.disposables.add(this.contextKeyService),this.disposables.add(a.register(this)),this.updateStyles(o.overrideStyles),this.disposables.add(l.onDidChangeConfiguration(f=>{f.affectsConfiguration(b_)&&(this._useAltAsMultipleSelectionModifier=Bc(l));let p={};if(f.affectsConfiguration(ta)&&this.horizontalScrolling===void 0){const _=!!l.getValue(ta);p={...p,horizontalScrolling:_}}if(f.affectsConfiguration(Nl)){const _=!!l.getValue(Nl);p={...p,scrollByPage:_}}if(f.affectsConfiguration(Rl)){const _=!!l.getValue(Rl);p={...p,smoothScrolling:_}}if(f.affectsConfiguration(Pc)){const _=l.getValue(Pc);p={...p,mouseWheelScrollSensitivity:_}}if(f.affectsConfiguration(Fc)){const _=l.getValue(Fc);p={...p,fastScrollSensitivity:_}}Object.keys(p).length>0&&this.updateOptions(p)})),this.navigator=new Sse(this,{configurationService:l,...o}),this.disposables.add(this.navigator)}updateOptions(e){super.updateOptions(e),e.overrideStyles!==void 0&&this.updateStyles(e.overrideStyles),e.multipleSelectionSupport!==void 0&&this.listSupportsMultiSelect.set(!!e.multipleSelectionSupport)}updateStyles(e){this.style(e?f_(e):lC)}dispose(){this.disposables.dispose(),super.dispose()}};Yj=Ph([Xt(5,Ae),Xt(6,fa),Xt(7,He),Xt(8,_e)],Yj);let Zj=class extends p5{constructor(e,t,i,n,o,r,a,l,c,d){const h=typeof r.horizontalScrolling<"u"?r.horizontalScrolling:!!c.getValue(ta),[u,g]=d.invokeFunction(WA,r);super(e,t,i,n,o,{keyboardSupport:!1,...u,horizontalScrolling:h}),this.disposables.add(g),this.contextKeyService=FA(a,this),this.disposables.add(BA(this.contextKeyService,this)),this.listSupportsMultiSelect=OA.bindTo(this.contextKeyService),this.listSupportsMultiSelect.set(r.multipleSelectionSupport!==!1),PA.bindTo(this.contextKeyService).set(!!r.selectionNavigation),this.listHasSelectionOrFocus=SH.bindTo(this.contextKeyService),this.listDoubleSelection=xH.bindTo(this.contextKeyService),this.listMultiSelection=LH.bindTo(this.contextKeyService),this.horizontalScrolling=r.horizontalScrolling,this._useAltAsMultipleSelectionModifier=Bc(c),this.disposables.add(this.contextKeyService),this.disposables.add(l.register(this)),this.updateStyles(r.overrideStyles),this.disposables.add(this.onDidChangeSelection(()=>{const p=this.getSelection(),_=this.getFocus();this.contextKeyService.bufferChangeEvents(()=>{this.listHasSelectionOrFocus.set(p.length>0||_.length>0),this.listMultiSelection.set(p.length>1),this.listDoubleSelection.set(p.length===2)})})),this.disposables.add(this.onDidChangeFocus(()=>{const p=this.getSelection(),_=this.getFocus();this.listHasSelectionOrFocus.set(p.length>0||_.length>0)})),this.disposables.add(c.onDidChangeConfiguration(p=>{p.affectsConfiguration(b_)&&(this._useAltAsMultipleSelectionModifier=Bc(c));let _={};if(p.affectsConfiguration(ta)&&this.horizontalScrolling===void 0){const b=!!c.getValue(ta);_={..._,horizontalScrolling:b}}if(p.affectsConfiguration(Nl)){const b=!!c.getValue(Nl);_={..._,scrollByPage:b}}if(p.affectsConfiguration(Rl)){const b=!!c.getValue(Rl);_={..._,smoothScrolling:b}}if(p.affectsConfiguration(Pc)){const b=c.getValue(Pc);_={..._,mouseWheelScrollSensitivity:b}}if(p.affectsConfiguration(Fc)){const b=c.getValue(Fc);_={..._,fastScrollSensitivity:b}}Object.keys(_).length>0&&this.updateOptions(_)})),this.navigator=new rRe(this,{configurationService:c,...r}),this.disposables.add(this.navigator)}updateOptions(e){super.updateOptions(e),e.overrideStyles!==void 0&&this.updateStyles(e.overrideStyles),e.multipleSelectionSupport!==void 0&&this.listSupportsMultiSelect.set(!!e.multipleSelectionSupport)}updateStyles(e){this.style(e?f_(e):lC)}dispose(){this.disposables.dispose(),super.dispose()}};Zj=Ph([Xt(6,Ae),Xt(7,fa),Xt(8,He),Xt(9,_e)],Zj);class NH extends F{constructor(e,t){super(),this.widget=e,this._onDidOpen=this._register(new M),this.onDidOpen=this._onDidOpen.event,this._register(oe.filter(this.widget.onDidChangeSelection,i=>Id(i.browserEvent))(i=>this.onSelectionFromKeyboard(i))),this._register(this.widget.onPointer(i=>this.onPointer(i.element,i.browserEvent))),this._register(this.widget.onMouseDblClick(i=>this.onMouseDblClick(i.element,i.browserEvent))),typeof t?.openOnSingleClick!="boolean"&&t?.configurationService?(this.openOnSingleClick=t?.configurationService.getValue(jk)!=="doubleClick",this._register(t?.configurationService.onDidChangeConfiguration(i=>{i.affectsConfiguration(jk)&&(this.openOnSingleClick=t?.configurationService.getValue(jk)!=="doubleClick")}))):this.openOnSingleClick=t?.openOnSingleClick??!0}onSelectionFromKeyboard(e){if(e.elements.length!==1)return;const t=e.browserEvent,i=typeof t.preserveFocus=="boolean"?t.preserveFocus:!0,n=typeof t.pinned=="boolean"?t.pinned:!i;this._open(this.getSelectedElement(),i,n,!1,e.browserEvent)}onPointer(e,t){if(!this.openOnSingleClick||t.detail===2)return;const n=t.button===1,o=!0,r=n,a=t.ctrlKey||t.metaKey||t.altKey;this._open(e,o,r,a,t)}onMouseDblClick(e,t){if(!t)return;const i=t.target;if(i.classList.contains("monaco-tl-twistie")||i.classList.contains("monaco-icon-label")&&i.classList.contains("folder-icon")&&t.offsetX<16)return;const o=!1,r=!0,a=t.ctrlKey||t.metaKey||t.altKey;this._open(e,o,r,a,t)}_open(e,t,i,n,o){e&&this._onDidOpen.fire({editorOptions:{preserveFocus:t,pinned:i,revealIfVisible:!0},sideBySide:n,element:e,browserEvent:o})}}class Sse extends NH{constructor(e,t){super(e,t),this.widget=e}getSelectedElement(){return this.widget.getSelectedElements()[0]}}class rRe extends NH{constructor(e,t){super(e,t)}getSelectedElement(){return this.widget.getSelectedElements()[0]}}class aRe extends NH{constructor(e,t){super(e,t)}getSelectedElement(){return this.widget.getSelection()[0]??void 0}}function lRe(s){let e=!1;return t=>{if(t.toKeyCodeChord().isModifierKey())return!1;if(e)return e=!1,!1;const i=s.softDispatch(t,t.target);return i.kind===1?(e=!0,!1):(e=!1,i.kind===0)}}let lD=class extends wH{get onDidOpen(){return this.internals.onDidOpen}constructor(e,t,i,n,o,r,a,l,c){const{options:d,getTypeNavigationMode:h,disposable:u}=r.invokeFunction(wx,o);super(e,t,i,n,d),this.disposables.add(u),this.internals=new Km(this,o,h,o.overrideStyles,a,l,c),this.disposables.add(this.internals)}updateOptions(e){super.updateOptions(e),this.internals.updateOptions(e)}};lD=Ph([Xt(5,_e),Xt(6,Ae),Xt(7,fa),Xt(8,He)],lD);let Xj=class extends gse{constructor(e,t,i,n,o,r,a,l,c){const{options:d,getTypeNavigationMode:h,disposable:u}=r.invokeFunction(wx,o);super(e,t,i,n,d),this.disposables.add(u),this.internals=new Km(this,o,h,o.overrideStyles,a,l,c),this.disposables.add(this.internals)}updateOptions(e={}){super.updateOptions(e),e.overrideStyles&&this.internals.updateStyleOverrides(e.overrideStyles),this.internals.updateOptions(e)}};Xj=Ph([Xt(5,_e),Xt(6,Ae),Xt(7,fa),Xt(8,He)],Xj);let Qj=class extends JNe{constructor(e,t,i,n,o,r,a,l,c,d){const{options:h,getTypeNavigationMode:u,disposable:g}=a.invokeFunction(wx,r);super(e,t,i,n,o,h),this.disposables.add(g),this.internals=new Km(this,r,u,r.overrideStyles,l,c,d),this.disposables.add(this.internals)}updateOptions(e={}){super.updateOptions(e),e.overrideStyles!==void 0&&this.internals.updateStyleOverrides(e.overrideStyles),this.internals.updateOptions(e)}};Qj=Ph([Xt(6,_e),Xt(7,Ae),Xt(8,fa),Xt(9,He)],Qj);let v5=class extends pse{get onDidOpen(){return this.internals.onDidOpen}constructor(e,t,i,n,o,r,a,l,c,d){const{options:h,getTypeNavigationMode:u,disposable:g}=a.invokeFunction(wx,r);super(e,t,i,n,o,h),this.disposables.add(g),this.internals=new Km(this,r,u,r.overrideStyles,l,c,d),this.disposables.add(this.internals)}updateOptions(e={}){super.updateOptions(e),e.overrideStyles&&this.internals.updateStyleOverrides(e.overrideStyles),this.internals.updateOptions(e)}};v5=Ph([Xt(6,_e),Xt(7,Ae),Xt(8,fa),Xt(9,He)],v5);let Jj=class extends XNe{constructor(e,t,i,n,o,r,a,l,c,d,h){const{options:u,getTypeNavigationMode:g,disposable:f}=l.invokeFunction(wx,a);super(e,t,i,n,o,r,u),this.disposables.add(f),this.internals=new Km(this,a,g,a.overrideStyles,c,d,h),this.disposables.add(this.internals)}updateOptions(e){super.updateOptions(e),this.internals.updateOptions(e)}};Jj=Ph([Xt(7,_e),Xt(8,Ae),Xt(9,fa),Xt(10,He)],Jj);function xse(s){const e=s.getValue(IH);if(e==="highlight")return _r.Highlight;if(e==="filter")return _r.Filter;const t=s.getValue(nD);if(t==="simple"||t==="highlight")return _r.Highlight;if(t==="filter")return _r.Filter}function Lse(s){const e=s.getValue(TH);if(e==="fuzzy")return fl.Fuzzy;if(e==="contiguous")return fl.Contiguous}function wx(s,e){const t=s.get(He),i=s.get(Rh),n=s.get(Ae),o=s.get(_e),r=()=>{const u=n.getContextKeyValue(Cse);if(u==="automatic")return uc.Automatic;if(u==="trigger"||n.getContextKeyValue(yse)===!1)return uc.Trigger;const f=t.getValue(DH);if(f==="automatic")return uc.Automatic;if(f==="trigger")return uc.Trigger},a=e.horizontalScrolling!==void 0?e.horizontalScrolling:!!t.getValue(ta),[l,c]=o.invokeFunction(WA,e),d=e.paddingBottom,h=e.renderIndentGuides!==void 0?e.renderIndentGuides:t.getValue(sD);return{getTypeNavigationMode:r,disposable:c,options:{keyboardSupport:!1,...l,indent:typeof t.getValue(XS)=="number"?t.getValue(XS):void 0,renderIndentGuides:h,smoothScrolling:!!t.getValue(Rl),defaultFindMode:e.defaultFindMode??xse(t),defaultFindMatchType:e.defaultFindMatchType??Lse(t),horizontalScrolling:a,scrollByPage:!!t.getValue(Nl),paddingBottom:d,hideTwistiesOfChildlessElements:e.hideTwistiesOfChildlessElements,expandOnlyOnTwistieClick:e.expandOnlyOnTwistieClick??t.getValue(oD)==="doubleClick",contextViewProvider:i,findWidgetStyles:Zke,enableStickyScroll:!!t.getValue(rD),stickyScrollMaxItemCount:Number(t.getValue(aD))}}}let Km=class{get onDidOpen(){return this.navigator.onDidOpen}constructor(e,t,i,n,o,r,a){this.tree=e,this.disposables=[],this.contextKeyService=FA(o,e),this.disposables.push(BA(this.contextKeyService,e)),this.listSupportsMultiSelect=OA.bindTo(this.contextKeyService),this.listSupportsMultiSelect.set(t.multipleSelectionSupport!==!1),PA.bindTo(this.contextKeyService).set(!!t.selectionNavigation),this.listSupportFindWidget=tRe.bindTo(this.contextKeyService),this.listSupportFindWidget.set(t.findWidgetEnabled??!0),this.hasSelectionOrFocus=SH.bindTo(this.contextKeyService),this.hasDoubleSelection=xH.bindTo(this.contextKeyService),this.hasMultiSelection=LH.bindTo(this.contextKeyService),this.treeElementCanCollapse=kH.bindTo(this.contextKeyService),this.treeElementHasParent=iRe.bindTo(this.contextKeyService),this.treeElementCanExpand=EH.bindTo(this.contextKeyService),this.treeElementHasChild=nRe.bindTo(this.contextKeyService),this.treeFindOpen=sRe.bindTo(this.contextKeyService),this.treeStickyScrollFocused=wse.bindTo(this.contextKeyService),this._useAltAsMultipleSelectionModifier=Bc(a),this.updateStyleOverrides(n);const c=()=>{const h=e.getFocus()[0];if(!h)return;const u=e.getNode(h);this.treeElementCanCollapse.set(u.collapsible&&!u.collapsed),this.treeElementHasParent.set(!!e.getParentElement(h)),this.treeElementCanExpand.set(u.collapsible&&u.collapsed),this.treeElementHasChild.set(!!e.getFirstElementChild(h))},d=new Set;d.add(Cse),d.add(yse),this.disposables.push(this.contextKeyService,r.register(e),e.onDidChangeSelection(()=>{const h=e.getSelection(),u=e.getFocus();this.contextKeyService.bufferChangeEvents(()=>{this.hasSelectionOrFocus.set(h.length>0||u.length>0),this.hasMultiSelection.set(h.length>1),this.hasDoubleSelection.set(h.length===2)})}),e.onDidChangeFocus(()=>{const h=e.getSelection(),u=e.getFocus();this.hasSelectionOrFocus.set(h.length>0||u.length>0),c()}),e.onDidChangeCollapseState(c),e.onDidChangeModel(c),e.onDidChangeFindOpenState(h=>this.treeFindOpen.set(h)),e.onDidChangeStickyScrollFocused(h=>this.treeStickyScrollFocused.set(h)),a.onDidChangeConfiguration(h=>{let u={};if(h.affectsConfiguration(b_)&&(this._useAltAsMultipleSelectionModifier=Bc(a)),h.affectsConfiguration(XS)){const g=a.getValue(XS);u={...u,indent:g}}if(h.affectsConfiguration(sD)&&t.renderIndentGuides===void 0){const g=a.getValue(sD);u={...u,renderIndentGuides:g}}if(h.affectsConfiguration(Rl)){const g=!!a.getValue(Rl);u={...u,smoothScrolling:g}}if(h.affectsConfiguration(IH)||h.affectsConfiguration(nD)){const g=xse(a);u={...u,defaultFindMode:g}}if(h.affectsConfiguration(DH)||h.affectsConfiguration(nD)){const g=i();u={...u,typeNavigationMode:g}}if(h.affectsConfiguration(TH)){const g=Lse(a);u={...u,defaultFindMatchType:g}}if(h.affectsConfiguration(ta)&&t.horizontalScrolling===void 0){const g=!!a.getValue(ta);u={...u,horizontalScrolling:g}}if(h.affectsConfiguration(Nl)){const g=!!a.getValue(Nl);u={...u,scrollByPage:g}}if(h.affectsConfiguration(oD)&&t.expandOnlyOnTwistieClick===void 0&&(u={...u,expandOnlyOnTwistieClick:a.getValue(oD)==="doubleClick"}),h.affectsConfiguration(rD)){const g=a.getValue(rD);u={...u,enableStickyScroll:g}}if(h.affectsConfiguration(aD)){const g=Math.max(1,a.getValue(aD));u={...u,stickyScrollMaxItemCount:g}}if(h.affectsConfiguration(Pc)){const g=a.getValue(Pc);u={...u,mouseWheelScrollSensitivity:g}}if(h.affectsConfiguration(Fc)){const g=a.getValue(Fc);u={...u,fastScrollSensitivity:g}}Object.keys(u).length>0&&e.updateOptions(u)}),this.contextKeyService.onDidChangeContext(h=>{h.affectsSome(d)&&e.updateOptions({typeNavigationMode:i()})})),this.navigator=new aRe(e,{configurationService:a,...t}),this.disposables.push(this.navigator)}updateOptions(e){e.multipleSelectionSupport!==void 0&&this.listSupportsMultiSelect.set(!!e.multipleSelectionSupport)}updateStyleOverrides(e){this.tree.style(e?f_(e):lC)}dispose(){this.disposables=xt(this.disposables)}};Km=Ph([Xt(4,Ae),Xt(5,fa),Xt(6,He)],Km);const cRe=bi.as(Ml.Configuration);cRe.registerConfiguration({id:"workbench",order:7,title:m(1705,"Workbench"),type:"object",properties:{[b_]:{type:"string",enum:["ctrlCmd","alt"],markdownEnumDescriptions:[m(1706,"Maps to `Control` on Windows and Linux and to `Command` on macOS."),m(1707,"Maps to `Alt` on Windows and Linux and to `Option` on macOS.")],default:"ctrlCmd",description:m(1708,"The modifier to be used to add an item in trees and lists to a multi-selection with the mouse (for example in the explorer, open editors and scm view). The 'Open to Side' mouse gestures - if supported - will adapt such that they do not conflict with the multiselect modifier.")},[jk]:{type:"string",enum:["singleClick","doubleClick"],default:"singleClick",description:m(1709,"Controls how to open items in trees and lists using the mouse (if supported). Note that some trees and lists might choose to ignore this setting if it is not applicable.")},[ta]:{type:"boolean",default:!1,description:m(1710,"Controls whether lists and trees support horizontal scrolling in the workbench. Warning: turning on this setting has a performance implication.")},[Nl]:{type:"boolean",default:!1,description:m(1711,"Controls whether clicks in the scrollbar scroll page by page.")},[XS]:{type:"number",default:8,minimum:4,maximum:40,description:m(1712,"Controls tree indentation in pixels.")},[sD]:{type:"string",enum:["none","onHover","always"],default:"onHover",description:m(1713,"Controls whether the tree should render indent guides.")},[Rl]:{type:"boolean",default:!1,description:m(1714,"Controls whether lists and trees have smooth scrolling.")},[Pc]:{type:"number",default:1,markdownDescription:m(1715,"A multiplier to be used on the `deltaX` and `deltaY` of mouse wheel scroll events.")},[Fc]:{type:"number",default:5,markdownDescription:m(1716,"Scrolling speed multiplier when pressing `Alt`.")},[IH]:{type:"string",enum:["highlight","filter"],enumDescriptions:[m(1717,"Highlight elements when searching. Further up and down navigation will traverse only the highlighted elements."),m(1718,"Filter elements when searching.")],default:"highlight",description:m(1719,"Controls the default find mode for lists and trees in the workbench.")},[nD]:{type:"string",enum:["simple","highlight","filter"],enumDescriptions:[m(1720,"Simple keyboard navigation focuses elements which match the keyboard input. Matching is done only on prefixes."),m(1721,"Highlight keyboard navigation highlights elements which match the keyboard input. Further up and down navigation will traverse only the highlighted elements."),m(1722,"Filter keyboard navigation will filter out and hide all the elements which do not match the keyboard input.")],default:"highlight",description:m(1723,"Controls the keyboard navigation style for lists and trees in the workbench. Can be simple, highlight and filter."),deprecated:!0,deprecationMessage:m(1724,"Please use 'workbench.list.defaultFindMode' and	'workbench.list.typeNavigationMode' instead.")},[TH]:{type:"string",enum:["fuzzy","contiguous"],enumDescriptions:[m(1725,"Use fuzzy matching when searching."),m(1726,"Use contiguous matching when searching.")],default:"fuzzy",description:m(1727,"Controls the type of matching used when searching lists and trees in the workbench.")},[oD]:{type:"string",enum:["singleClick","doubleClick"],default:"singleClick",description:m(1728,"Controls how tree folders are expanded when clicking the folder names. Note that some trees and lists might choose to ignore this setting if it is not applicable.")},[rD]:{type:"boolean",default:!0,description:m(1729,"Controls whether sticky scrolling is enabled in trees.")},[aD]:{type:"number",minimum:1,default:7,markdownDescription:m(1730,"Controls the number of sticky elements displayed in the tree when {0} is enabled.","`#workbench.tree.enableStickyScroll#`")},[DH]:{type:"string",enum:["automatic","trigger"],default:"automatic",markdownDescription:m(1731,"Controls how type navigation works in lists and trees in the workbench. When set to `trigger`, type navigation begins once the `list.triggerTypeNavigation` command is run.")}}});class kh{constructor(e,t,i,n){this.isProviderFirst=e,this.parent=t,this.link=i,this._rangeCallback=n,this.id=XF.nextId()}get uri(){return this.link.uri}get range(){return this._range??this.link.targetSelectionRange??this.link.range}set range(e){this._range=e,this._rangeCallback(this)}get ariaMessage(){const e=this.parent.getPreview(this)?.preview(this.range);return e?m(1088,"{0} in {1} on line {2} at column {3}",e.value,ra(this.uri),this.range.startLineNumber,this.range.startColumn):m(1087,"in {0} on line {1} at column {2}",ra(this.uri),this.range.startLineNumber,this.range.startColumn)}}class dRe{constructor(e){this._modelReference=e}dispose(){this._modelReference.dispose()}preview(e,t=8){const i=this._modelReference.object.textEditorModel;if(!i)return;const{startLineNumber:n,startColumn:o,endLineNumber:r,endColumn:a}=e,l=i.getWordUntilPosition({lineNumber:n,column:o-t}),c=new k(n,l.startColumn,n,o),d=new k(r,a,r,1073741824),h=i.getValueInRange(c).replace(/^\s+/,""),u=i.getValueInRange(e),g=i.getValueInRange(d).replace(/\s+$/,"");return{value:h+u+g,highlight:{start:h.length,end:h.length+u.length}}}}class kv{constructor(e,t){this.parent=e,this.uri=t,this.children=[],this._previews=new Wi}dispose(){xt(this._previews.values()),this._previews.clear()}getPreview(e){return this._previews.get(e.uri)}get ariaMessage(){const e=this.children.length;return e===1?m(1089,"1 symbol in {0}, full path {1}",ra(this.uri),this.uri.fsPath):m(1090,"{0} symbols in {1}, full path {2}",e,ra(this.uri),this.uri.fsPath)}async resolve(e){if(this._previews.size!==0)return this;for(const t of this.children)if(!this._previews.has(t.uri))try{const i=await e.createModelReference(t.uri);this._previews.set(t.uri,new dRe(i))}catch(i){Oe(i)}return this}}class zo{constructor(e,t){this.groups=[],this.references=[],this._onDidChangeReferenceRange=new M,this.onDidChangeReferenceRange=this._onDidChangeReferenceRange.event,this._links=e,this._title=t;const[i]=e;e.sort(zo._compareReferences);let n;for(const o of e)if((!n||!oi.isEqual(n.uri,o.uri,!0))&&(n=new kv(this,o.uri),this.groups.push(n)),n.children.length===0||zo._compareReferences(o,n.children[n.children.length-1])!==0){const r=new kh(i===o,n,o,a=>this._onDidChangeReferenceRange.fire(a));this.references.push(r),n.children.push(r)}}dispose(){xt(this.groups),this._onDidChangeReferenceRange.dispose(),this.groups.length=0}clone(){return new zo(this._links,this._title)}get title(){return this._title}get isEmpty(){return this.groups.length===0}get ariaMessage(){return this.isEmpty?m(1091,"No results found"):this.references.length===1?m(1092,"Found 1 symbol in {0}",this.references[0].uri.fsPath):this.groups.length===1?m(1093,"Found {0} symbols in {1}",this.references.length,this.groups[0].uri.fsPath):m(1094,"Found {0} symbols in {1} files",this.references.length,this.groups.length)}nextOrPreviousReference(e,t){const{parent:i}=e;let n=i.children.indexOf(e);const o=i.children.length,r=i.parent.groups.length;return r===1||t&&n+1<o||!t&&n>0?(t?n=(n+1)%o:n=(n+o-1)%o,i.children[n]):(n=i.parent.groups.indexOf(i),t?(n=(n+1)%r,i.parent.groups[n].children[0]):(n=(n+r-1)%r,i.parent.groups[n].children[i.parent.groups[n].children.length-1]))}nearestReference(e,t){const i=this.references.map((n,o)=>({idx:o,prefixLen:Fa(n.uri.toString(),e.toString()),offsetDist:Math.abs(n.range.startLineNumber-t.lineNumber)*100+Math.abs(n.range.startColumn-t.column)})).sort((n,o)=>n.prefixLen>o.prefixLen?-1:n.prefixLen<o.prefixLen?1:n.offsetDist<o.offsetDist?-1:n.offsetDist>o.offsetDist?1:0)[0];if(i)return this.references[i.idx]}referenceAt(e,t){for(const i of this.references)if(i.uri.toString()===e.toString()&&k.containsPosition(i.range,t))return i}firstReference(){for(const e of this.references)if(e.isProviderFirst)return e;return this.references[0]}static _compareReferences(e,t){return oi.compare(e.uri,t.uri)||k.compareRangesUsingStarts(e.range,t.range)}}class C5 extends F{constructor(e,t,i){super(),this.options=t,this.styles=i,this.count=0,this.hover=this._register(new vt),this.element=X(e,ie(".monaco-count-badge")),this._register(pe(()=>e.removeChild(this.element))),this.countFormat=this.options.countFormat||"{0}",this.titleFormat=this.options.titleFormat||"",this.setCount(this.options.count||0),this.updateHover()}setCount(e){this.count=e,this.render()}setTitleFormat(e){this.titleFormat=e,this.updateHover(),this.render()}updateHover(){this.titleFormat!==""&&!this.hover.value?this.hover.value=$a().setupDelayedHoverAtMouse(this.element,()=>({content:xm(this.titleFormat,this.count),appearance:{compact:!0}})):this.titleFormat===""&&this.hover.value&&(this.hover.value=void 0)}render(){this.element.textContent=xm(this.countFormat,this.count),this.element.style.backgroundColor=this.styles.badgeBackground??"",this.element.style.color=this.styles.badgeForeground??"",this.styles.badgeBorder&&(this.element.style.border=`1px solid ${this.styles.badgeBorder}`)}}class Wu extends F{constructor(e,t){super(),this.options=t,this.text="",this.title="",this.highlights=[],this.didEverRender=!1,this.domNode=X(e,ie("span.monaco-highlighted-label"))}get element(){return this.domNode}set(e,t=[],i="",n,o){e||(e=""),n&&(e=Wu.escapeNewLines(e,t)),!(this.didEverRender&&this.text===e&&this.title===i&&Ho(this.highlights,t))&&(this.text=e,this.title=i,this.highlights=t,this.render(o))}render(e){const t=[];let i=0;for(const n of this.highlights){if(n.end===n.start)continue;if(i<n.start){const a=this.text.substring(i,n.start);e?t.push(...Pu(a)):t.push(a),i=n.start}const o=this.text.substring(i,n.end),r=ie("span.highlight",void 0,...e?Pu(o):[o]);n.extraClasses&&r.classList.add(...n.extraClasses),t.push(r),i=n.end}if(i<this.text.length){const n=this.text.substring(i);e?t.push(...Pu(n)):t.push(n)}if(vn(this.domNode,...t),!this.customHover&&this.title!==""){const n=this.options?.hoverDelegate??Uc("mouse");this.customHover=this._register($a().setupManagedHover(n,this.domNode,this.title))}else this.customHover&&this.customHover.update(this.title);this.didEverRender=!0}static escapeNewLines(e,t){let i=0,n=0;return e.replace(/\r\n|\r|\n/g,(o,r)=>{n=o===`\r
`?-1:0,r+=i;for(const a of t)a.end<=r||(a.start>=r&&(a.start+=n),a.end>=r&&(a.end+=n));return i+=n,"⏎"})}}class qC{constructor(e){this._element=e}get element(){return this._element}set textContent(e){this.disposed||e===this._textContent||(this._textContent=e,this._element.textContent=e)}set classNames(e){this.disposed||Ho(e,this._classNames)||(this._classNames=e,this._element.classList.value="",this._element.classList.add(...e))}set empty(e){this.disposed||e===this._empty||(this._empty=e,this._element.style.marginLeft=e?"0":"")}dispose(){this.disposed=!0}}class QS extends F{constructor(e,t){super(),this.customHovers=new Map,this.creationOptions=t,this.domNode=this._register(new qC(X(e,ie(".monaco-icon-label")))),this.labelContainer=X(this.domNode.element,ie(".monaco-icon-label-container")),this.nameContainer=X(this.labelContainer,ie("span.monaco-icon-name-container")),t?.supportHighlights||t?.supportIcons?this.nameNode=this._register(new gRe(this.nameContainer,!!t.supportIcons)):this.nameNode=new hRe(this.nameContainer),this.hoverDelegate=t?.hoverDelegate??Uc("mouse")}get element(){return this.domNode.element}setLabel(e,t,i){const n=["monaco-icon-label"],o=["monaco-icon-label-container"];let r="";i&&(i.extraClasses&&n.push(...i.extraClasses),i.bold&&n.push("bold"),i.italic&&n.push("italic"),i.strikethrough&&n.push("strikethrough"),i.disabledCommand&&o.push("disabled"),i.title&&(typeof i.title=="string"?r+=i.title:r+=e));const a=this.domNode.element.querySelector(".monaco-icon-label-iconpath");if(i?.iconPath){let l;if(!a||!ki(a)?(l=ie(".monaco-icon-label-iconpath"),this.domNode.element.prepend(l)):l=a,xe.isThemeIcon(i.iconPath)){const c=xe.asClassName(i.iconPath);l.className=`monaco-icon-label-iconpath ${c}`,l.style.backgroundImage=""}else l.style.backgroundImage=Mc(i?.iconPath);l.style.backgroundRepeat="no-repeat",l.style.backgroundPosition="center",l.style.backgroundSize="contain"}else a&&a.remove();if(this.domNode.classNames=n,this.domNode.element.setAttribute("aria-label",r),this.labelContainer.classList.value="",this.labelContainer.classList.add(...o),this.setupHover(i?.descriptionTitle?this.labelContainer:this.element,i?.title),this.nameNode.setLabel(e,i),t||this.descriptionNode){const l=this.getOrCreateDescriptionNode();if(l instanceof Wu){const c=i?.supportIcons??this.creationOptions?.supportIcons;l.set(t||"",i?i.descriptionMatches:void 0,void 0,i?.labelEscapeNewLines,c),this.setupHover(l.element,i?.descriptionTitle)}else l.textContent=t&&i?.labelEscapeNewLines?Wu.escapeNewLines(t,[]):t||"",this.setupHover(l.element,i?.descriptionTitle||""),l.empty=!t}if(i?.suffix||this.suffixNode){const l=this.getOrCreateSuffixNode();l.textContent=i?.suffix??""}}setupHover(e,t){const i=this.customHovers.get(e);if(i&&(i.dispose(),this.customHovers.delete(e)),!t){e.removeAttribute("title");return}let n=e;if(this.creationOptions?.hoverTargetOverride){if(!wn(e,this.creationOptions.hoverTargetOverride))throw new Error("hoverTargetOverrride must be an ancestor of the htmlElement");n=this.creationOptions.hoverTargetOverride}const o=$a().setupManagedHover(this.hoverDelegate,n,t);o&&this.customHovers.set(e,o)}dispose(){super.dispose();for(const e of this.customHovers.values())e.dispose();this.customHovers.clear()}getOrCreateSuffixNode(){if(!this.suffixNode){const e=this._register(new qC(uue(this.nameContainer,ie("span.monaco-icon-suffix-container"))));this.suffixNode=this._register(new qC(X(e.element,ie("span.label-suffix"))))}return this.suffixNode}getOrCreateDescriptionNode(){if(!this.descriptionNode){const e=this._register(new qC(X(this.labelContainer,ie("span.monaco-icon-description-container"))));this.creationOptions?.supportDescriptionHighlights?this.descriptionNode=this._register(new Wu(X(e.element,ie("span.label-description")))):this.descriptionNode=this._register(new qC(X(e.element,ie("span.label-description"))))}return this.descriptionNode}}class hRe{constructor(e){this.container=e,this.label=void 0,this.singleLabel=void 0}setLabel(e,t){if(!(this.label===e&&Ho(this.options,t)))if(this.label=e,this.options=t,typeof e=="string")this.singleLabel||(this.container.textContent="",this.container.classList.remove("multiple"),this.singleLabel=X(this.container,ie("a.label-name",{id:t?.domId}))),this.singleLabel.textContent=e;else{this.container.textContent="",this.container.classList.add("multiple"),this.singleLabel=void 0;for(let i=0;i<e.length;i++){const n=e[i],o=t?.domId&&`${t?.domId}_${i}`;X(this.container,ie("a.label-name",{id:o,"data-icon-label-count":e.length,"data-icon-label-index":i,role:"treeitem"},n)),i<e.length-1&&X(this.container,ie("span.label-separator",void 0,t?.separator||"/"))}}}}function uRe(s,e,t){if(!t)return;let i=0;return s.map(n=>{const o={start:i,end:i+n.length},r=t.map(a=>as.intersect(o,a)).filter(a=>!as.isEmpty(a)).map(({start:a,end:l})=>({start:a-i,end:l-i}));return i=o.end+e.length,r})}class gRe extends F{constructor(e,t){super(),this.container=e,this.supportIcons=t,this.label=void 0,this.singleLabel=void 0}setLabel(e,t){if(this.label===e&&Ho(this.options,t))return;this.label=e,this.options=t;const i=t?.supportIcons??this.supportIcons;if(typeof e=="string")this.singleLabel||(this.container.textContent="",this.container.classList.remove("multiple"),this.singleLabel=this._register(new Wu(X(this.container,ie("a.label-name",{id:t?.domId}))))),this.singleLabel.set(e,t?.matches,void 0,t?.labelEscapeNewLines,i);else{this.container.textContent="",this.container.classList.add("multiple"),this.singleLabel=void 0;const n=t?.separator||"/",o=uRe(e,n,t?.matches);for(let r=0;r<e.length;r++){const a=e[r],l=o?o[r]:void 0,c=t?.domId&&`${t?.domId}_${r}`,d=ie("a.label-name",{id:c,"data-icon-label-count":e.length,"data-icon-label-index":r,role:"treeitem"});this._register(new Wu(X(this.container,d))).set(a,l,void 0,t?.labelEscapeNewLines,i),r<e.length-1&&X(d,ie("span.label-separator",void 0,n))}}}}const Ym=qe("labelService");var HA=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},VA=function(s,e){return function(t,i){e(t,i,s)}},y5;let S5=class{constructor(e){this._resolverService=e}hasChildren(e){return e instanceof zo||e instanceof kv}getChildren(e){if(e instanceof zo)return e.groups;if(e instanceof kv)return e.resolve(this._resolverService).then(t=>t.children);throw new Error("bad tree")}};S5=HA([VA(0,Fs)],S5);class fRe{getHeight(){return 23}getTemplateId(e){return e instanceof kv?cD.id:dD.id}}let x5=class{constructor(e){this._keybindingService=e}getKeyboardNavigationLabel(e){if(e instanceof kh){const t=e.parent.getPreview(e)?.preview(e.range);if(t)return t.value}return ra(e.uri)}};x5=HA([VA(0,ft)],x5);class pRe{getId(e){return e instanceof kh?e.id:e.uri}}let L5=class extends F{constructor(e,t){super(),this._labelService=t;const i=document.createElement("div");i.classList.add("reference-file"),this.file=this._register(new QS(i,{supportHighlights:!0})),this.badge=this._register(new C5(X(i,ie(".count")),{},Sie)),e.appendChild(i)}set(e,t){const i=zR(e.uri);this.file.setLabel(this._labelService.getUriBasenameLabel(e.uri),this._labelService.getUriLabel(i,{relative:!0}),{title:this._labelService.getUriLabel(e.uri),matches:t});const n=e.children.length;this.badge.setCount(n),n>1?this.badge.setTitleFormat(m(1081,"{0} references",n)):this.badge.setTitleFormat(m(1082,"{0} reference",n))}};L5=HA([VA(1,Ym)],L5);var Np;let cD=(Np=class{constructor(e){this._instantiationService=e,this.templateId=y5.id}renderTemplate(e){return this._instantiationService.createInstance(L5,e)}renderElement(e,t,i){i.set(e.element,ox(e.filterData))}disposeTemplate(e){e.dispose()}},y5=Np,Np.id="FileReferencesRenderer",Np);cD=y5=HA([VA(0,_e)],cD);class mRe extends F{constructor(e){super(),this.label=this._register(new Wu(e))}set(e,t){const i=e.parent.getPreview(e)?.preview(e.range);if(!i||!i.value)this.label.set(`${ra(e.uri)}:${e.range.startLineNumber+1}:${e.range.startColumn+1}`);else{const{value:n,highlight:o}=i;t&&!Oa.isDefault(t)?(this.label.element.classList.toggle("referenceMatch",!1),this.label.set(n,ox(t))):(this.label.element.classList.toggle("referenceMatch",!0),this.label.set(n,[o]))}}}const SN=class SN{constructor(){this.templateId=SN.id}renderTemplate(e){return new mRe(e)}renderElement(e,t,i){i.set(e.element,e.filterData)}disposeTemplate(e){e.dispose()}};SN.id="OneReferenceRenderer";let dD=SN;class _Re{getWidgetAriaLabel(){return m(1083,"References")}getAriaLabel(e){return e.ariaMessage}}var kse=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},lf=function(s,e){return function(t,i){e(t,i,s)}};const xN=class xN{constructor(e,t){this._editor=e,this._model=t,this._decorations=new Map,this._decorationIgnoreSet=new Set,this._callOnDispose=new U,this._callOnModelChange=new U,this._callOnDispose.add(this._editor.onDidChangeModel(()=>this._onModelChanged())),this._onModelChanged()}dispose(){this._callOnModelChange.dispose(),this._callOnDispose.dispose(),this.removeDecorations()}_onModelChanged(){this._callOnModelChange.clear();const e=this._editor.getModel();if(e){for(const t of this._model.references)if(t.uri.toString()===e.uri.toString()){this._addDecorations(t.parent);return}}}_addDecorations(e){if(!this._editor.hasModel())return;this._callOnModelChange.add(this._editor.getModel().onDidChangeDecorations(()=>this._onDecorationChanged()));const t=[],i=[];for(let n=0,o=e.children.length;n<o;n++){const r=e.children[n];this._decorationIgnoreSet.has(r.id)||r.uri.toString()===this._editor.getModel().uri.toString()&&(t.push({range:r.range,options:xN.DecorationOptions}),i.push(n))}this._editor.changeDecorations(n=>{const o=n.deltaDecorations([],t);for(let r=0;r<o.length;r++)this._decorations.set(o[r],e.children[i[r]])})}_onDecorationChanged(){const e=[],t=this._editor.getModel();if(t){for(const[i,n]of this._decorations){const o=t.getDecorationRange(i);if(!o)continue;let r=!1;if(!k.equalsRange(o,n.range)){if(k.spansMultipleLines(o))r=!0;else{const a=n.range.endColumn-n.range.startColumn,l=o.endColumn-o.startColumn;a!==l&&(r=!0)}r?(this._decorationIgnoreSet.add(n.id),e.push(i)):n.range=o}}for(let i=0,n=e.length;i<n;i++)this._decorations.delete(e[i]);this._editor.removeDecorations(e)}}removeDecorations(){this._editor.removeDecorations([...this._decorations.keys()]),this._decorations.clear()}};xN.DecorationOptions=Fe.register({description:"reference-decoration",stickiness:1,className:"reference-decoration"});let k5=xN;class bRe{constructor(){this.ratio=.7,this.heightInLines=18}static fromJSON(e){let t,i;try{const n=JSON.parse(e);t=n.ratio,i=n.heightInLines}catch{}return{ratio:t||.7,heightInLines:i||18}}}class wRe extends v5{}let E5=class{constructor(e){this.labelService=e,this.disposables=new U}getDragURI(e){return e instanceof kv?e.uri.toString():e instanceof kh?KEe(e.uri,e.range).toString():null}getDragLabel(e){return e.length===0?void 0:e.map(i=>this.labelService.getUriBasenameLabel(i.uri)).join(", ")}onDragStart(e,t){if(!t.dataTransfer)return;const n=e.elements.map(o=>this.getDragURI(o)).filter(Boolean);n.length&&(t.dataTransfer.setData(Mm.RESOURCES,JSON.stringify(n)),t.dataTransfer.setData(Mm.TEXT,n.join(`
`)))}onDragOver(){return!1}drop(){}dispose(){this.disposables.dispose()}};E5=kse([lf(0,Ym)],E5);let I5=class extends eD{constructor(e,t,i,n,o,r,a,l,c){super(e,{showFrame:!1,showArrow:!0,isResizeable:!0,isAccessible:!0,supportOnTitleClick:!0},r),this._defaultTreeKeyboardSupport=t,this.layoutData=i,this._textModelResolverService=o,this._instantiationService=r,this._peekViewService=a,this._uriLabel=l,this._keybindingService=c,this._disposeOnNewModel=new U,this._callOnDispose=new U,this._onDidSelectReference=new M,this.onDidSelectReference=this._onDidSelectReference.event,this._dim=new St(0,0),this._isClosing=!1,this._applyTheme(n.getColorTheme()),this._callOnDispose.add(n.onDidColorThemeChange(this._applyTheme.bind(this))),this._peekViewService.addExclusiveWidget(e,this),this.create()}get isClosing(){return this._isClosing}dispose(){this._isClosing=!0,this.setModel(void 0),this._callOnDispose.dispose(),this._disposeOnNewModel.dispose(),xt(this._preview),xt(this._previewNotAvailableMessage),xt(this._tree),xt(this._previewModelReference),this._splitView.dispose(),super.dispose()}_applyTheme(e){const t=e.getColor(aNe)||q.transparent;this.style({arrowColor:t,frameColor:t,headerBackgroundColor:e.getColor(rNe)||q.transparent,primaryHeadingColor:e.getColor(nse),secondaryHeadingColor:e.getColor(sse)})}show(e){super.show(e,this.layoutData.heightInLines||18)}focusOnReferenceTree(){this._tree.domFocus()}focusOnPreviewEditor(){this._preview.focus()}isPreviewEditorFocused(){return this._preview.hasTextFocus()}_onTitleClick(e){this._preview&&this._preview.getModel()&&this._onDidSelectReference.fire({element:this._getFocusedReference(),kind:e.ctrlKey||e.metaKey||e.altKey?"side":"open",source:"title"})}_fillBody(e){this.setCssClass("reference-zone-widget"),this._messageContainer=X(e,ie("div.messages")),js(this._messageContainer),this._splitView=new rse(e,{orientation:1}),this._previewContainer=X(e,ie("div.preview.inline"));const t={scrollBeyondLastLine:!1,scrollbar:{verticalScrollbarSize:14,horizontal:"auto",useShadows:!0,verticalHasArrows:!1,horizontalHasArrows:!1,alwaysConsumeMouseWheel:!0},overviewRulerLanes:2,fixedOverflowWidgets:!0,minimap:{enabled:!1}};this._preview=this._instantiationService.createInstance(Lh,this._previewContainer,t,{},this.editor),js(this._previewContainer),this._previewNotAvailableMessage=this._instantiationService.createInstance(Rm,m(1084,"no preview available"),xr,Rm.DEFAULT_CREATION_OPTIONS,null),this._treeContainer=X(e,ie("div.ref-tree.inline"));const i={keyboardSupport:this._defaultTreeKeyboardSupport,accessibilityProvider:new _Re,keyboardNavigationLabelProvider:this._instantiationService.createInstance(x5),identityProvider:new pRe,openOnSingleClick:!0,selectionNavigation:!0,overrideStyles:{listBackground:lNe},dnd:this._instantiationService.createInstance(E5)};this._defaultTreeKeyboardSupport&&this._callOnDispose.add(Fi(this._treeContainer,"keydown",o=>{o.equals(9)&&(this._keybindingService.dispatchEvent(o,o.target),o.stopPropagation())},!0)),this._tree=this._instantiationService.createInstance(wRe,"ReferencesWidget",this._treeContainer,new fRe,[this._instantiationService.createInstance(cD),this._instantiationService.createInstance(dD)],this._instantiationService.createInstance(S5),i),this._splitView.addView({onDidChange:oe.None,element:this._previewContainer,minimumSize:200,maximumSize:Number.MAX_VALUE,layout:o=>{this._preview.layout({height:this._dim.height,width:o})}},tD.Distribute),this._splitView.addView({onDidChange:oe.None,element:this._treeContainer,minimumSize:100,maximumSize:Number.MAX_VALUE,layout:o=>{this._treeContainer.style.height=`${this._dim.height}px`,this._treeContainer.style.width=`${o}px`,this._tree.layout(this._dim.height,o)}},tD.Distribute),this._disposables.add(this._splitView.onDidSashChange(()=>{this._dim.width&&(this.layoutData.ratio=this._splitView.getViewSize(0)/this._dim.width)},void 0));const n=(o,r)=>{o instanceof kh&&(r==="show"&&this._revealReference(o,!1),this._onDidSelectReference.fire({element:o,kind:r,source:"tree"}))};this._disposables.add(this._tree.onDidOpen(o=>{o.sideBySide?n(o.element,"side"):o.editorOptions.pinned?n(o.element,"goto"):n(o.element,"show")})),js(this._treeContainer)}_onWidth(e){this._dim&&this._doLayoutBody(this._dim.height,e)}_doLayoutBody(e,t){super._doLayoutBody(e,t),this._dim=new St(t,e),this.layoutData.heightInLines=this._viewZone?this._viewZone.heightInLines:this.layoutData.heightInLines,this._splitView.layout(t),this._splitView.resizeView(0,t*this.layoutData.ratio)}setSelection(e){return this._revealReference(e,!0).then(()=>{this._model&&(this._tree.setSelection([e]),this._tree.setFocus([e]))})}setModel(e){return this._disposeOnNewModel.clear(),this._model=e,this._model?this._onNewModel():Promise.resolve()}_onNewModel(){return this._model?this._model.isEmpty?(this.setTitle(""),this._messageContainer.innerText=m(1085,"No results"),Mo(this._messageContainer),Promise.resolve(void 0)):(js(this._messageContainer),this._decorationsManager=new k5(this._preview,this._model),this._disposeOnNewModel.add(this._decorationsManager),this._disposeOnNewModel.add(this._model.onDidChangeReferenceRange(e=>this._tree.rerender(e))),this._disposeOnNewModel.add(this._preview.onMouseDown(e=>{const{event:t,target:i}=e;if(t.detail!==2)return;const n=this._getFocusedReference();n&&this._onDidSelectReference.fire({element:{uri:n.uri,range:i.range},kind:t.ctrlKey||t.metaKey||t.altKey?"side":"open",source:"editor"})})),this.container.classList.add("results-loaded"),Mo(this._treeContainer),Mo(this._previewContainer),this._splitView.layout(this._dim.width),this.focusOnReferenceTree(),this._tree.setInput(this._model.groups.length===1?this._model.groups[0]:this._model)):Promise.resolve(void 0)}_getFocusedReference(){const[e]=this._tree.getFocus();if(e instanceof kh)return e;if(e instanceof kv&&e.children.length>0)return e.children[0]}async revealReference(e){await this._revealReference(e,!1),this._onDidSelectReference.fire({element:e,kind:"goto",source:"tree"})}async _revealReference(e,t){if(this._revealedReference===e)return;this._revealedReference=e,e.uri.scheme!==Te.inMemory?this.setTitle(Xue(e.uri),this._uriLabel.getUriLabel(zR(e.uri))):this.setTitle(m(1086,"References"));const i=this._textModelResolverService.createModelReference(e.uri);this._tree.getInput()===e.parent?this._tree.reveal(e):(t&&this._tree.reveal(e.parent),await this._tree.expand(e.parent),this._tree.reveal(e));const n=await i;if(!this._model){n.dispose();return}xt(this._previewModelReference);const o=n.object;if(o){const r=this._preview.getModel()===o.textEditorModel?0:1,a=k.lift(e.range).collapseToStart();this._previewModelReference=n,this._preview.setModel(o.textEditorModel),this._preview.setSelection(a),this._preview.revealRangeInCenter(a,r)}else this._preview.setModel(this._previewNotAvailableMessage),n.dispose()}};I5=kse([lf(3,wi),lf(4,Fs),lf(5,_e),lf(6,ise),lf(7,Ym),lf(8,ft)],I5);var vRe=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},Y_=function(s,e){return function(t,i){e(t,i,s)}},Gk;const w_=new ce("referenceSearchVisible",!1,m(1078,"Whether reference peek is visible, like 'Peek References' or 'Peek Definition'"));var Rp;let Zm=(Rp=class{static get(e){return e.getContribution(Gk.ID)}constructor(e,t,i,n,o,r,a,l){this._defaultTreeKeyboardSupport=e,this._editor=t,this._editorService=n,this._notificationService=o,this._instantiationService=r,this._storageService=a,this._configurationService=l,this._disposables=new U,this._requestIdPool=0,this._ignoreModelChangeEvent=!1,this._referenceSearchVisible=w_.bindTo(i)}dispose(){this._referenceSearchVisible.reset(),this._disposables.dispose(),this._widget?.dispose(),this._model?.dispose(),this._widget=void 0,this._model=void 0}toggleWidget(e,t,i){let n;if(this._widget&&(n=this._widget.position),this.closeWidget(),n&&e.containsPosition(n))return;this._peekMode=i,this._referenceSearchVisible.set(!0),this._disposables.add(this._editor.onDidChangeModelLanguage(()=>{this.closeWidget()})),this._disposables.add(this._editor.onDidChangeModel(()=>{this._ignoreModelChangeEvent||this.closeWidget()}));const o="peekViewLayout",r=bRe.fromJSON(this._storageService.get(o,0,"{}"));this._widget=this._instantiationService.createInstance(I5,this._editor,this._defaultTreeKeyboardSupport,r),this._widget.setTitle(m(1079,"Loading...")),this._widget.show(e),this._disposables.add(this._widget.onDidClose(()=>{t.cancel(),this._widget?(this._storageService.store(o,JSON.stringify(this._widget.layoutData),0,1),this._widget.isClosing||this.closeWidget(),this._widget=void 0):this.closeWidget()})),this._disposables.add(this._widget.onDidSelectReference(l=>{const{element:c,kind:d}=l;if(c)switch(d){case"open":(l.source!=="editor"||!this._configurationService.getValue("editor.stablePeek"))&&this.openReference(c,!1,!1);break;case"side":this.openReference(c,!0,!1);break;case"goto":i?this._gotoReference(c,!0):this.openReference(c,!1,!0);break}}));const a=++this._requestIdPool;t.then(l=>{if(a!==this._requestIdPool||!this._widget){l.dispose();return}return this._model?.dispose(),this._model=l,this._widget.setModel(this._model).then(()=>{if(this._widget&&this._model&&this._editor.hasModel()){this._model.isEmpty?this._widget.setMetaTitle(""):this._widget.setMetaTitle(m(1080,"{0} ({1})",this._model.title,this._model.references.length));const c=this._editor.getModel().uri,d=new A(e.startLineNumber,e.startColumn),h=this._model.nearestReference(c,d);if(h)return this._widget.setSelection(h).then(()=>{this._widget&&this._editor.getOption(99)==="editor"&&this._widget.focusOnPreviewEditor()})}})},l=>{this._notificationService.error(l)})}changeFocusBetweenPreviewAndReferences(){this._widget&&(this._widget.isPreviewEditorFocused()?this._widget.focusOnReferenceTree():this._widget.focusOnPreviewEditor())}async goToNextOrPreviousReference(e){if(!this._editor.hasModel()||!this._model||!this._widget)return;const t=this._widget.position;if(!t)return;const i=this._model.nearestReference(this._editor.getModel().uri,t);if(!i)return;const n=this._model.nextOrPreviousReference(i,e),o=this._editor.hasTextFocus(),r=this._widget.isPreviewEditorFocused();await this._widget.setSelection(n),await this._gotoReference(n,!1),o?this._editor.focus():this._widget&&r&&this._widget.focusOnPreviewEditor()}async revealReference(e){!this._editor.hasModel()||!this._model||!this._widget||await this._widget.revealReference(e)}closeWidget(e=!0){this._widget?.dispose(),this._model?.dispose(),this._referenceSearchVisible.reset(),this._disposables.clear(),this._widget=void 0,this._model=void 0,e&&this._editor.focus(),this._requestIdPool+=1}_gotoReference(e,t){this._widget?.hide(),this._ignoreModelChangeEvent=!0;const i=k.lift(e.range).collapseToStart();return this._editorService.openCodeEditor({resource:e.uri,options:{selection:i,selectionSource:"code.jump",pinned:t}},this._editor).then(n=>{if(this._ignoreModelChangeEvent=!1,!n||!this._widget){this.closeWidget();return}if(this._editor===n)this._widget.show(i),this._widget.focusOnReferenceTree();else{const o=Gk.get(n),r=this._model.clone();this.closeWidget(),n.focus(),o?.toggleWidget(i,an(a=>Promise.resolve(r)),this._peekMode??!1)}},n=>{this._ignoreModelChangeEvent=!1,Oe(n)})}openReference(e,t,i){t||this.closeWidget();const{uri:n,range:o}=e;this._editorService.openCodeEditor({resource:n,options:{selection:o,selectionSource:"code.jump",pinned:i}},this._editor,t)}},Gk=Rp,Rp.ID="editor.contrib.referencesController",Rp);Zm=Gk=vRe([Y_(2,Ae),Y_(3,ht),Y_(4,Ii),Y_(5,_e),Y_(6,Bs),Y_(7,He)],Zm);function v_(s,e){const t=Jne(s);if(!t)return;const i=Zm.get(t);i&&e(i)}Dn.registerCommandAndKeybindingRule({id:"togglePeekWidgetFocus",weight:100,primary:Gi(2089,60),when:K.or(w_,So.inPeekEditor),handler(s){v_(s,e=>{e.changeFocusBetweenPreviewAndReferences()})}});Dn.registerCommandAndKeybindingRule({id:"goToNextReference",weight:90,primary:62,secondary:[70],when:K.or(w_,So.inPeekEditor),handler(s){v_(s,e=>{e.goToNextOrPreviousReference(!0)})}});Dn.registerCommandAndKeybindingRule({id:"goToPreviousReference",weight:90,primary:1086,secondary:[1094],when:K.or(w_,So.inPeekEditor),handler(s){v_(s,e=>{e.goToNextOrPreviousReference(!1)})}});at.registerCommandAlias("goToNextReferenceFromEmbeddedEditor","goToNextReference");at.registerCommandAlias("goToPreviousReferenceFromEmbeddedEditor","goToPreviousReference");at.registerCommandAlias("closeReferenceSearchEditor","closeReferenceSearch");at.registerCommand("closeReferenceSearch",s=>v_(s,e=>e.closeWidget()));Dn.registerKeybindingRule({id:"closeReferenceSearch",weight:-1,primary:9,secondary:[1033],when:K.and(So.inPeekEditor,K.not("config.editor.stablePeek"))});Dn.registerKeybindingRule({id:"closeReferenceSearch",weight:250,primary:9,secondary:[1033],when:K.and(w_,K.not("config.editor.stablePeek"),K.or(N.editorTextFocus,yH.negate()))});Dn.registerCommandAndKeybindingRule({id:"revealReference",weight:200,primary:3,mac:{primary:3,secondary:[2066]},when:K.and(w_,vse,kH.negate(),EH.negate()),handler(s){const t=s.get(fa).lastFocusedList?.getFocus();Array.isArray(t)&&t[0]instanceof kh&&v_(s,i=>i.revealReference(t[0]))}});Dn.registerCommandAndKeybindingRule({id:"openReferenceToSide",weight:100,primary:2051,mac:{primary:259},when:K.and(w_,vse,kH.negate(),EH.negate()),handler(s){const t=s.get(fa).lastFocusedList?.getFocus();Array.isArray(t)&&t[0]instanceof kh&&v_(s,i=>i.openReference(t[0],!0,!0))}});at.registerCommand("openReference",s=>{const t=s.get(fa).lastFocusedList?.getFocus();Array.isArray(t)&&t[0]instanceof kh&&v_(s,i=>i.openReference(t[0],!1,!0))});var Ese=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},f1=function(s,e){return function(t,i){e(t,i,s)}};const RH=new ce("hasSymbols",!1,m(1095,"Whether there are symbol locations that can be navigated via keyboard-only.")),zA=qe("ISymbolNavigationService");let D5=class{constructor(e,t,i,n){this._editorService=t,this._notificationService=i,this._keybindingService=n,this._currentModel=void 0,this._currentIdx=-1,this._ignoreEditorChange=!1,this._ctxHasSymbols=RH.bindTo(e)}reset(){this._ctxHasSymbols.reset(),this._currentState?.dispose(),this._currentMessage?.close(),this._currentModel=void 0,this._currentIdx=-1}put(e){const t=e.parent.parent;if(t.references.length<=1){this.reset();return}this._currentModel=t,this._currentIdx=t.references.indexOf(e),this._ctxHasSymbols.set(!0),this._showMessage();const i=new T5(this._editorService),n=i.onDidChange(o=>{if(this._ignoreEditorChange)return;const r=this._editorService.getActiveCodeEditor();if(!r)return;const a=r.getModel(),l=r.getPosition();if(!a||!l)return;let c=!1,d=!1;for(const h of t.references)if(dg(h.uri,a.uri))c=!0,d=d||k.containsPosition(h.range,l);else if(c)break;(!c||!d)&&this.reset()});this._currentState=Ra(i,n)}revealNext(e){if(!this._currentModel)return Promise.resolve();this._currentIdx+=1,this._currentIdx%=this._currentModel.references.length;const t=this._currentModel.references[this._currentIdx];return this._showMessage(),this._ignoreEditorChange=!0,this._editorService.openCodeEditor({resource:t.uri,options:{selection:k.collapseToStart(t.range),selectionRevealType:3}},e).finally(()=>{this._ignoreEditorChange=!1})}_showMessage(){this._currentMessage?.close();const e=this._keybindingService.lookupKeybinding("editor.gotoNextSymbolFromResult"),t=e?m(1096,"Symbol {0} of {1}, {2} for next",this._currentIdx+1,this._currentModel.references.length,e.getLabel()):m(1097,"Symbol {0} of {1}",this._currentIdx+1,this._currentModel.references.length);this._currentMessage=this._notificationService.status(t)}};D5=Ese([f1(0,Ae),f1(1,ht),f1(2,Ii),f1(3,ft)],D5);tt(zA,D5,1);le(new class extends hn{constructor(){super({id:"editor.gotoNextSymbolFromResult",precondition:RH,kbOpts:{weight:100,primary:70}})}runEditorCommand(s,e){return s.get(zA).revealNext(e)}});Dn.registerCommandAndKeybindingRule({id:"editor.gotoNextSymbolFromResult.cancel",weight:100,when:RH,primary:9,handler(s){s.get(zA).reset()}});let T5=class{constructor(e){this._listener=new Map,this._disposables=new U,this._onDidChange=new M,this.onDidChange=this._onDidChange.event,this._disposables.add(e.onCodeEditorRemove(this._onDidRemoveEditor,this)),this._disposables.add(e.onCodeEditorAdd(this._onDidAddEditor,this)),e.listCodeEditors().forEach(this._onDidAddEditor,this)}dispose(){this._disposables.dispose(),this._onDidChange.dispose(),xt(this._listener.values())}_onDidAddEditor(e){this._listener.set(e,Ra(e.onDidChangeCursorPosition(t=>this._onDidChange.fire({editor:e})),e.onDidChangeModelContent(t=>this._onDidChange.fire({editor:e}))))}_onDidRemoveEditor(e){this._listener.get(e)?.dispose(),this._listener.delete(e)}};T5=Ese([f1(0,ht)],T5);function N5(s,e){return e.uri.scheme===s.uri.scheme?!0:!fP(e.uri,Te.walkThroughSnippet,Te.vscodeChatCodeBlock,Te.vscodeChatCodeCompareBlock)}async function vx(s,e,t,i,n){const r=t.ordered(s,i).map(l=>Promise.resolve(n(l,s,e)).then(void 0,c=>{ji(c)})),a=await Promise.all(r);return Al(a.flat()).filter(l=>N5(s,l))}function Cx(s,e,t,i,n){return vx(e,t,s,i,(o,r,a)=>o.provideDefinition(r,a,n))}function AH(s,e,t,i,n){return vx(e,t,s,i,(o,r,a)=>o.provideDeclaration(r,a,n))}function MH(s,e,t,i,n){return vx(e,t,s,i,(o,r,a)=>o.provideImplementation(r,a,n))}function OH(s,e,t,i,n){return vx(e,t,s,i,(o,r,a)=>o.provideTypeDefinition(r,a,n))}function yx(s,e,t,i,n,o){return vx(e,t,s,n,async(r,a,l)=>{const c=(await r.provideReferences(a,l,{includeDeclaration:!0},o))?.filter(h=>N5(a,h));if(!i||!c||c.length!==2)return c;const d=(await r.provideReferences(a,l,{includeDeclaration:!1},o))?.filter(h=>N5(a,h));return d&&d.length===1?d:c})}async function Gc(s){const e=await s(),t=new zo(e,""),i=t.references.map(n=>n.link);return t.dispose(),i}Lo("_executeDefinitionProvider",(s,e,t)=>{const i=s.get(ge),n=Cx(i.definitionProvider,e,t,!1,Ke.None);return Gc(()=>n)});Lo("_executeDefinitionProvider_recursive",(s,e,t)=>{const i=s.get(ge),n=Cx(i.definitionProvider,e,t,!0,Ke.None);return Gc(()=>n)});Lo("_executeTypeDefinitionProvider",(s,e,t)=>{const i=s.get(ge),n=OH(i.typeDefinitionProvider,e,t,!1,Ke.None);return Gc(()=>n)});Lo("_executeTypeDefinitionProvider_recursive",(s,e,t)=>{const i=s.get(ge),n=OH(i.typeDefinitionProvider,e,t,!0,Ke.None);return Gc(()=>n)});Lo("_executeDeclarationProvider",(s,e,t)=>{const i=s.get(ge),n=AH(i.declarationProvider,e,t,!1,Ke.None);return Gc(()=>n)});Lo("_executeDeclarationProvider_recursive",(s,e,t)=>{const i=s.get(ge),n=AH(i.declarationProvider,e,t,!0,Ke.None);return Gc(()=>n)});Lo("_executeReferenceProvider",(s,e,t)=>{const i=s.get(ge),n=yx(i.referenceProvider,e,t,!1,!1,Ke.None);return Gc(()=>n)});Lo("_executeReferenceProvider_recursive",(s,e,t)=>{const i=s.get(ge),n=yx(i.referenceProvider,e,t,!1,!0,Ke.None);return Gc(()=>n)});Lo("_executeImplementationProvider",(s,e,t)=>{const i=s.get(ge),n=MH(i.implementationProvider,e,t,!1,Ke.None);return Gc(()=>n)});Lo("_executeImplementationProvider_recursive",(s,e,t)=>{const i=s.get(ge),n=MH(i.implementationProvider,e,t,!0,Ke.None);return Gc(()=>n)});En.appendMenuItem(fe.EditorContext,{submenu:fe.EditorContextPeek,title:m(1038,"Peek"),group:"navigation",order:100});class Ev{static is(e){return!e||typeof e!="object"?!1:!!(e instanceof Ev||A.isIPosition(e.position)&&e.model)}constructor(e,t){this.model=e,this.position=t}}const ya=class ya extends za{static all(){return ya._allSymbolNavigationCommands.values()}static _patchConfig(e){const t={...e,f1:!0};if(t.menu)for(const i of ot.wrap(t.menu))(i.id===fe.EditorContext||i.id===fe.EditorContextPeek)&&(i.when=K.and(e.precondition,i.when));return t}constructor(e,t){super(ya._patchConfig(t)),this.configuration=e,ya._allSymbolNavigationCommands.set(t.id,this)}runEditorCommand(e,t,i,n){if(!t.hasModel())return Promise.resolve(void 0);const o=e.get(Ii),r=e.get(ht),a=e.get(Ch),l=e.get(zA),c=e.get(ge),d=e.get(_e),h=t.getModel(),u=t.getPosition(),g=Ev.is(i)?i:new Ev(h,u),f=new yh(t,5),p=tv(this._getLocationModel(c,g.model,g.position,f.token),f.token).then(async _=>{if(!_||f.token.isCancellationRequested)return;Js(_.ariaMessage);let b;if(_.referenceAt(h.uri,u)){const C=this._getAlternativeCommand(t);C!==void 0&&!ya._activeAlternativeCommands.has(C)&&ya._allSymbolNavigationCommands.has(C)&&(b=ya._allSymbolNavigationCommands.get(C))}const w=_.references.length;if(w===0){if(!this.configuration.muteMessage){const C=h.getWordAtPosition(u);Vo.get(t)?.showMessage(this._getNoResultFoundMessage(C),u)}}else if(w===1&&b)ya._activeAlternativeCommands.add(this.desc.id),d.invokeFunction(C=>b.runEditorCommand(C,t,i,n).finally(()=>{ya._activeAlternativeCommands.delete(this.desc.id)}));else return this._onResult(r,l,t,_,n)},_=>{o.error(_)}).finally(()=>{f.dispose()});return a.showWhile(p,250),p}async _onResult(e,t,i,n,o){const r=this._getGoToPreference(i);if(!(i instanceof Lh)&&(this.configuration.openInPeek||r==="peek"&&n.references.length>1))this._openInPeek(i,n,o);else{const a=n.firstReference(),l=n.references.length>1&&r==="gotoAndPeek",c=await this._openReference(i,e,a,this.configuration.openToSide,!l);l&&c?this._openInPeek(c,n,o):n.dispose(),r==="goto"&&t.put(a)}}async _openReference(e,t,i,n,o){let r;if(Xwe(i)&&(r=i.targetSelectionRange),r||(r=i.range),!r)return;const a=await t.openCodeEditor({resource:i.uri,options:{selection:k.collapseToStart(r),selectionRevealType:3,selectionSource:"code.jump"}},e,n);if(a){if(o){const l=a.getModel(),c=a.createDecorationsCollection([{range:r,options:{description:"symbol-navigate-action-highlight",className:"symbolHighlight"}}]);setTimeout(()=>{a.getModel()===l&&c.clear()},350)}return a}}_openInPeek(e,t,i){const n=Zm.get(e);n&&e.hasModel()?n.toggleWidget(i??e.getSelection(),an(o=>Promise.resolve(t)),this.configuration.openInPeek):t.dispose()}};ya._allSymbolNavigationCommands=new Map,ya._activeAlternativeCommands=new Set;let Eh=ya;class Sx extends Eh{async _getLocationModel(e,t,i,n){return new zo(await Cx(e.definitionProvider,t,i,!1,n),m(1039,"Definitions"))}_getNoResultFoundMessage(e){return e&&e.word?m(1040,"No definition found for '{0}'",e.word):m(1041,"No definition found")}_getAlternativeCommand(e){return e.getOption(67).alternativeDefinitionCommand}_getGoToPreference(e){return e.getOption(67).multipleDefinitions}}var Ku;It((Ku=class extends Sx{constructor(){super({openToSide:!1,openInPeek:!1,muteMessage:!1},{id:Ku.id,title:{...V(1065,"Go to Definition"),mnemonicTitle:m(1042,"Go to &&Definition")},precondition:N.hasDefinitionProvider,keybinding:[{when:N.editorTextFocus,primary:70,weight:100},{when:K.and(N.editorTextFocus,mse),primary:2118,weight:100}],menu:[{id:fe.EditorContext,group:"navigation",order:1.1},{id:fe.MenubarGoMenu,precondition:null,group:"4_symbol_nav",order:2}]}),at.registerCommandAlias("editor.action.goToDeclaration",Ku.id)}},Ku.id="editor.action.revealDefinition",Ku));var Yu;It((Yu=class extends Sx{constructor(){super({openToSide:!0,openInPeek:!1,muteMessage:!1},{id:Yu.id,title:V(1066,"Open Definition to the Side"),precondition:K.and(N.hasDefinitionProvider,N.isInEmbeddedEditor.toNegated()),keybinding:[{when:N.editorTextFocus,primary:Gi(2089,70),weight:100},{when:K.and(N.editorTextFocus,mse),primary:Gi(2089,2118),weight:100}]}),at.registerCommandAlias("editor.action.openDeclarationToTheSide",Yu.id)}},Yu.id="editor.action.revealDefinitionAside",Yu));var Zu;It((Zu=class extends Sx{constructor(){super({openToSide:!1,openInPeek:!0,muteMessage:!1},{id:Zu.id,title:V(1067,"Peek Definition"),precondition:K.and(N.hasDefinitionProvider,So.notInPeekEditor,N.isInEmbeddedEditor.toNegated()),keybinding:{when:N.editorTextFocus,primary:582,linux:{primary:3140},weight:100},menu:{id:fe.EditorContextPeek,group:"peek",order:2}}),at.registerCommandAlias("editor.action.previewDeclaration",Zu.id)}},Zu.id="editor.action.peekDefinition",Zu));class Ise extends Eh{async _getLocationModel(e,t,i,n){return new zo(await AH(e.declarationProvider,t,i,!1,n),m(1043,"Declarations"))}_getNoResultFoundMessage(e){return e&&e.word?m(1044,"No declaration found for '{0}'",e.word):m(1045,"No declaration found")}_getAlternativeCommand(e){return e.getOption(67).alternativeDeclarationCommand}_getGoToPreference(e){return e.getOption(67).multipleDeclarations}}var Ap;It((Ap=class extends Ise{constructor(){super({openToSide:!1,openInPeek:!1,muteMessage:!1},{id:Ap.id,title:{...V(1068,"Go to Declaration"),mnemonicTitle:m(1046,"Go to &&Declaration")},precondition:K.and(N.hasDeclarationProvider,N.isInEmbeddedEditor.toNegated()),menu:[{id:fe.EditorContext,group:"navigation",order:1.3},{id:fe.MenubarGoMenu,precondition:null,group:"4_symbol_nav",order:3}]})}_getNoResultFoundMessage(e){return e&&e.word?m(1047,"No declaration found for '{0}'",e.word):m(1048,"No declaration found")}},Ap.id="editor.action.revealDeclaration",Ap));It(class extends Ise{constructor(){super({openToSide:!1,openInPeek:!0,muteMessage:!1},{id:"editor.action.peekDeclaration",title:V(1069,"Peek Declaration"),precondition:K.and(N.hasDeclarationProvider,So.notInPeekEditor,N.isInEmbeddedEditor.toNegated()),menu:{id:fe.EditorContextPeek,group:"peek",order:3}})}});class Dse extends Eh{async _getLocationModel(e,t,i,n){return new zo(await OH(e.typeDefinitionProvider,t,i,!1,n),m(1049,"Type Definitions"))}_getNoResultFoundMessage(e){return e&&e.word?m(1050,"No type definition found for '{0}'",e.word):m(1051,"No type definition found")}_getAlternativeCommand(e){return e.getOption(67).alternativeTypeDefinitionCommand}_getGoToPreference(e){return e.getOption(67).multipleTypeDefinitions}}var Mp;It((Mp=class extends Dse{constructor(){super({openToSide:!1,openInPeek:!1,muteMessage:!1},{id:Mp.ID,title:{...V(1070,"Go to Type Definition"),mnemonicTitle:m(1052,"Go to &&Type Definition")},precondition:N.hasTypeDefinitionProvider,keybinding:{when:N.editorTextFocus,primary:0,weight:100},menu:[{id:fe.EditorContext,group:"navigation",order:1.4},{id:fe.MenubarGoMenu,precondition:null,group:"4_symbol_nav",order:3}]})}},Mp.ID="editor.action.goToTypeDefinition",Mp));var Op;It((Op=class extends Dse{constructor(){super({openToSide:!1,openInPeek:!0,muteMessage:!1},{id:Op.ID,title:V(1071,"Peek Type Definition"),precondition:K.and(N.hasTypeDefinitionProvider,So.notInPeekEditor,N.isInEmbeddedEditor.toNegated()),menu:{id:fe.EditorContextPeek,group:"peek",order:4}})}},Op.ID="editor.action.peekTypeDefinition",Op));class Tse extends Eh{async _getLocationModel(e,t,i,n){return new zo(await MH(e.implementationProvider,t,i,!1,n),m(1053,"Implementations"))}_getNoResultFoundMessage(e){return e&&e.word?m(1054,"No implementation found for '{0}'",e.word):m(1055,"No implementation found")}_getAlternativeCommand(e){return e.getOption(67).alternativeImplementationCommand}_getGoToPreference(e){return e.getOption(67).multipleImplementations}}var Pp;It((Pp=class extends Tse{constructor(){super({openToSide:!1,openInPeek:!1,muteMessage:!1},{id:Pp.ID,title:{...V(1072,"Go to Implementations"),mnemonicTitle:m(1056,"Go to &&Implementations")},precondition:N.hasImplementationProvider,keybinding:{when:N.editorTextFocus,primary:2118,weight:100},menu:[{id:fe.EditorContext,group:"navigation",order:1.45},{id:fe.MenubarGoMenu,precondition:null,group:"4_symbol_nav",order:4}]})}},Pp.ID="editor.action.goToImplementation",Pp));var Fp;It((Fp=class extends Tse{constructor(){super({openToSide:!1,openInPeek:!0,muteMessage:!1},{id:Fp.ID,title:V(1073,"Peek Implementations"),precondition:K.and(N.hasImplementationProvider,So.notInPeekEditor,N.isInEmbeddedEditor.toNegated()),keybinding:{when:N.editorTextFocus,primary:3142,weight:100},menu:{id:fe.EditorContextPeek,group:"peek",order:5}})}},Fp.ID="editor.action.peekImplementation",Fp));class Nse extends Eh{_getNoResultFoundMessage(e){return e?m(1057,"No references found for '{0}'",e.word):m(1058,"No references found")}_getAlternativeCommand(e){return e.getOption(67).alternativeReferenceCommand}_getGoToPreference(e){return e.getOption(67).multipleReferences}}It(class extends Nse{constructor(){super({openToSide:!1,openInPeek:!1,muteMessage:!1},{id:"editor.action.goToReferences",title:{...V(1074,"Go to References"),mnemonicTitle:m(1059,"Go to &&References")},precondition:K.and(N.hasReferenceProvider,So.notInPeekEditor,N.isInEmbeddedEditor.toNegated()),keybinding:{when:N.editorTextFocus,primary:1094,weight:100},menu:[{id:fe.EditorContext,group:"navigation",order:1.45},{id:fe.MenubarGoMenu,precondition:null,group:"4_symbol_nav",order:5}]})}async _getLocationModel(e,t,i,n){return new zo(await yx(e.referenceProvider,t,i,!0,!1,n),m(1060,"References"))}});It(class extends Nse{constructor(){super({openToSide:!1,openInPeek:!0,muteMessage:!1},{id:"editor.action.referenceSearch.trigger",title:V(1075,"Peek References"),precondition:K.and(N.hasReferenceProvider,So.notInPeekEditor,N.isInEmbeddedEditor.toNegated()),menu:{id:fe.EditorContextPeek,group:"peek",order:6}})}async _getLocationModel(e,t,i,n){return new zo(await yx(e.referenceProvider,t,i,!1,!1,n),m(1061,"References"))}});class CRe extends Eh{constructor(e,t,i){super(e,{id:"editor.action.goToLocation",title:V(1076,"Go to Any Symbol"),precondition:K.and(So.notInPeekEditor,N.isInEmbeddedEditor.toNegated())}),this._references=t,this._gotoMultipleBehaviour=i}async _getLocationModel(e,t,i,n){return new zo(this._references,m(1062,"Locations"))}_getNoResultFoundMessage(e){return e&&m(1063,"No results for '{0}'",e.word)||""}_getGoToPreference(e){return this._gotoMultipleBehaviour??e.getOption(67).multipleReferences}_getAlternativeCommand(){}}at.registerCommand({id:"editor.action.goToLocations",metadata:{description:"Go to locations from a position in a file",args:[{name:"uri",description:"The text document in which to start",constraint:Le},{name:"position",description:"The position at which to start",constraint:A.isIPosition},{name:"locations",description:"An array of locations.",constraint:Array},{name:"multiple",description:"Define what to do when having multiple results, either `peek`, `gotoAndPeek`, or `goto`"},{name:"noResultsMessage",description:"Human readable message that shows when locations is empty."}]},handler:async(s,e,t,i,n,o,r)=>{dt(Le.isUri(e)),dt(A.isIPosition(t)),dt(Array.isArray(i)),dt(typeof n>"u"||typeof n=="string"),dt(typeof r>"u"||typeof r=="boolean");const a=s.get(ht),l=await a.openCodeEditor({resource:e},a.getFocusedCodeEditor());if(Tl(l))return l.setPosition(t),l.revealPositionInCenterIfOutsideViewport(t,0),l.invokeWithinContext(c=>{const d=new class extends CRe{_getNoResultFoundMessage(h){return o||super._getNoResultFoundMessage(h)}}({muteMessage:!o,openInPeek:!!r,openToSide:!1},i,n);c.get(_e).invokeFunction(d.run.bind(d),l)})}});at.registerCommand({id:"editor.action.peekLocations",metadata:{description:"Peek locations from a position in a file",args:[{name:"uri",description:"The text document in which to start",constraint:Le},{name:"position",description:"The position at which to start",constraint:A.isIPosition},{name:"locations",description:"An array of locations.",constraint:Array},{name:"multiple",description:"Define what to do when having multiple results, either `peek`, `gotoAndPeek`, or `goto`"}]},handler:async(s,e,t,i,n)=>{s.get(Gt).executeCommand("editor.action.goToLocations",e,t,i,n,void 0,!0)}});at.registerCommand({id:"editor.action.findReferences",handler:(s,e,t)=>{dt(Le.isUri(e)),dt(A.isIPosition(t));const i=s.get(ge),n=s.get(ht);return n.openCodeEditor({resource:e},n.getFocusedCodeEditor()).then(o=>{if(!Tl(o)||!o.hasModel())return;const r=Zm.get(o);if(!r)return;const a=an(c=>yx(i.referenceProvider,o.getModel(),A.lift(t),!1,!1,c).then(d=>new zo(d,m(1064,"References")))),l=new k(t.lineNumber,t.column,t.lineNumber,t.column);return Promise.resolve(r.toggleWidget(l,a,!1))})}});at.registerCommandAlias("editor.action.showReferences","editor.action.peekLocations");async function yRe(s,e,t,i){const n=s.get(Fs),o=s.get(Nr),r=s.get(Gt),a=s.get(_e),l=s.get(Ii);if(await i.item.resolve(Ke.None),!i.part.location)return;const c=i.part.location,d=[],h=new Set(En.getMenuItems(fe.EditorContext).map(g=>aw(g)?g.command.id:d_()));for(const g of Eh.all())h.has(g.desc.id)&&d.push(new Cr(g.desc.id,yr.label(g.desc,{renderShortTitle:!0}),void 0,!0,async()=>{const f=await n.createModelReference(c.uri);try{const p=new Ev(f.object.textEditorModel,k.getStartPosition(c.range)),_=i.item.anchor.range;await a.invokeFunction(g.runEditorCommand.bind(g),e,p,_)}finally{f.dispose()}}));if(i.part.command){const{command:g}=i.part;d.push(new tn),d.push(new Cr(g.id,g.title,void 0,!0,async()=>{try{await r.executeCommand(g.id,...g.arguments??[])}catch(f){l.notify({severity:oC.Error,source:i.item.provider.displayName,message:f})}}))}const u=e.getOption(144);o.showContextMenu({domForShadowRoot:u?e.getDomNode()??void 0:void 0,getAnchor:()=>{const g=Li(t);return{x:g.left,y:g.top+g.height+8}},getActions:()=>d,onHide:()=>{e.focus()},autoSelectFirstItem:!0})}async function Rse(s,e,t,i){const o=await s.get(Fs).createModelReference(i.uri);await t.invokeWithinContext(async r=>{const a=e.hasSideBySideModifier,l=r.get(Ae),c=So.inPeekEditor.getValue(l),d=!a&&t.getOption(101)&&!c;return new Sx({openToSide:a,openInPeek:d,muteMessage:!0},{title:{value:"",original:""},id:"",precondition:void 0}).run(r,new Ev(o.object.textEditorModel,k.getStartPosition(i.range)),k.lift(i.range))}),o.dispose()}var SRe=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},Z_=function(s,e){return function(t,i){e(t,i,s)}},pb;class hD{constructor(){this._entries=new Va(50)}get(e){const t=hD._key(e);return this._entries.get(t)}set(e,t){const i=hD._key(e);this._entries.set(i,t)}static _key(e){return`${e.uri.toString()}/${e.getVersionId()}`}}const Ase=qe("IInlayHintsCache");tt(Ase,hD,1);class R5{constructor(e,t){this.item=e,this.index=t}get part(){const e=this.item.hint.label;return typeof e=="string"?{label:e}:e[this.index]}}class xRe{constructor(e,t){this.part=e,this.hasTriggerModifier=t}}class LRe{constructor(){this._store=new vt,this._tokenSource=new si}dispose(){this._store.dispose(),this._tokenSource.dispose(!0)}reset(){return this._tokenSource.dispose(!0),this._tokenSource=new si,this._store.value=new U,{store:this._store.value,token:this._tokenSource.token}}}var Qd;let JS=(Qd=class{static get(e){return e.getContribution(pb.ID)??void 0}constructor(e,t,i,n,o,r,a){this._editor=e,this._languageFeaturesService=t,this._inlayHintsCache=n,this._commandService=o,this._notificationService=r,this._instaService=a,this._disposables=new U,this._sessionDisposables=new U,this._decorationsMetadata=new Map,this._activeRenderMode=0,this._ruleFactory=this._disposables.add(new jE(this._editor)),this._debounceInfo=i.for(t.inlayHintsProvider,"InlayHint",{min:25}),this._disposables.add(t.inlayHintsProvider.onDidChange(()=>this._update())),this._disposables.add(e.onDidChangeModel(()=>this._update())),this._disposables.add(e.onDidChangeModelLanguage(()=>this._update())),this._disposables.add(e.onDidChangeConfiguration(l=>{l.hasChanged(159)&&this._update()})),this._update()}dispose(){this._sessionDisposables.dispose(),this._removeAllDecorations(),this._disposables.dispose()}_update(){this._sessionDisposables.clear(),this._removeAllDecorations();const e=this._editor.getOption(159);if(e.enabled==="off")return;const t=this._editor.getModel();if(!t||!this._languageFeaturesService.inlayHintsProvider.has(t))return;if(e.enabled==="on")this._activeRenderMode=0;else{let c,d;e.enabled==="onUnlessPressed"?(c=0,d=1):(c=1,d=0),this._activeRenderMode=c,this._sessionDisposables.add(Vd.getInstance().event(h=>{if(!this._editor.hasModel())return;const u=h.altKey&&h.ctrlKey&&!(h.shiftKey||h.metaKey)?d:c;if(u!==this._activeRenderMode){this._activeRenderMode=u;const g=this._editor.getModel(),f=this._copyInlayHintsWithCurrentAnchor(g);this._updateHintsDecorators([g.getFullModelRange()],f),a.schedule(0)}}))}const i=this._inlayHintsCache.get(t);i&&this._updateHintsDecorators([t.getFullModelRange()],i),this._sessionDisposables.add(pe(()=>{t.isDisposed()||this._cacheHintsForFastRestore(t)}));let n;const o=new Set;this._sessionDisposables.add(t.onWillDispose(()=>n?.cancel()));const r=this._sessionDisposables.add(new LRe),a=new Rt(async()=>{const c=Date.now(),{store:d,token:h}=r.reset();try{const u=await QI.create(this._languageFeaturesService.inlayHintsProvider,t,this._getHintsRanges(),h);if(a.delay=this._debounceInfo.update(t,Date.now()-c),h.isCancellationRequested){u.dispose();return}for(const g of u.provider)typeof g.onDidChangeInlayHints=="function"&&!o.has(g)&&(o.add(g),d.add(g.onDidChangeInlayHints(()=>{a.isScheduled()||a.schedule()})));d.add(u),this._updateHintsDecorators(u.ranges,u.items),this._cacheHintsForFastRestore(t)}catch(u){Oe(u)}},this._debounceInfo.get(t));this._sessionDisposables.add(a),a.schedule(0),this._sessionDisposables.add(this._editor.onDidScrollChange(c=>{(c.scrollTopChanged||!a.isScheduled())&&a.schedule()}));const l=this._sessionDisposables.add(new vt);this._sessionDisposables.add(this._editor.onDidChangeModelContent(c=>{const d=Math.max(a.delay,800);this._cursorInfo={position:this._editor.getPosition(),notEarlierThan:Date.now()+d},l.value=hh(()=>a.schedule(0),d),a.schedule()})),this._sessionDisposables.add(this._editor.onDidChangeConfiguration(c=>{c.hasChanged(159)&&a.schedule()})),this._sessionDisposables.add(this._installDblClickGesture(()=>a.schedule(0))),this._sessionDisposables.add(this._installLinkGesture()),this._sessionDisposables.add(this._installContextMenu())}_installLinkGesture(){const e=new U,t=e.add(new AA(this._editor)),i=new U;return e.add(i),e.add(t.onMouseMoveOrRelevantKeyDown(n=>{const[o]=n,r=this._getInlayHintLabelPart(o),a=this._editor.getModel();if(!r||!a){i.clear();return}const l=new si;i.add(pe(()=>l.dispose(!0))),r.item.resolve(l.token),this._activeInlayHintPart=r.part.command||r.part.location?new xRe(r,o.hasTriggerModifier):void 0;const c=a.validatePosition(r.item.hint.position).lineNumber,d=new k(c,1,c,a.getLineMaxColumn(c)),h=this._getInlineHintsForRange(d);this._updateHintsDecorators([d],h),i.add(pe(()=>{this._activeInlayHintPart=void 0,this._updateHintsDecorators([d],h)}))})),e.add(t.onCancel(()=>i.clear())),e.add(t.onExecute(async n=>{const o=this._getInlayHintLabelPart(n);if(o){const r=o.part;r.location?this._instaService.invokeFunction(Rse,n,this._editor,r.location):hF.is(r.command)&&await this._invokeCommand(r.command,o.item)}})),e}_getInlineHintsForRange(e){const t=new Set;for(const i of this._decorationsMetadata.values())e.containsRange(i.item.anchor.range)&&t.add(i.item);return Array.from(t)}_installDblClickGesture(e){return this._editor.onMouseUp(async t=>{if(t.event.detail!==2)return;const i=this._getInlayHintLabelPart(t);if(i&&(t.event.preventDefault(),await i.item.resolve(Ke.None),Ms(i.item.hint.textEdits))){const n=i.item.hint.textEdits.map(o=>Si.replace(k.lift(o.range),o.text));this._editor.executeEdits("inlayHint.default",n),e()}})}_installContextMenu(){return this._editor.onContextMenu(async e=>{if(!ki(e.event.target))return;const t=this._getInlayHintLabelPart(e);t&&await this._instaService.invokeFunction(yRe,this._editor,e.event.target,t)})}_getInlayHintLabelPart(e){if(e.target.type!==6)return;const t=e.target.detail.injectedText?.options;if(t instanceof mg&&t?.attachedData instanceof R5)return t.attachedData}async _invokeCommand(e,t){try{await this._commandService.executeCommand(e.id,...e.arguments??[])}catch(i){this._notificationService.notify({severity:oC.Error,source:t.provider.displayName,message:i})}}_cacheHintsForFastRestore(e){const t=this._copyInlayHintsWithCurrentAnchor(e);this._inlayHintsCache.set(e,t)}_copyInlayHintsWithCurrentAnchor(e){const t=new Map;for(const[i,n]of this._decorationsMetadata){if(t.has(n.item))continue;const o=e.getDecorationRange(i);if(o){const r=new Xne(o,n.item.anchor.direction),a=n.item.with({anchor:r});t.set(n.item,a)}}return Array.from(t.values())}_getHintsRanges(){const t=this._editor.getModel(),i=this._editor.getVisibleRangesPlusViewportAboveBelow(),n=[];for(const o of i.sort(k.compareRangesUsingStarts)){const r=t.validateRange(new k(o.startLineNumber-30,o.startColumn,o.endLineNumber+30,o.endColumn));n.length===0||!k.areIntersectingOrTouching(n[n.length-1],r)?n.push(r):n[n.length-1]=k.plusRange(n[n.length-1],r)}return n}_updateHintsDecorators(e,t){const i=new Map;if(this._cursorInfo&&this._cursorInfo.notEarlierThan>Date.now()&&e.some(_=>_.containsPosition(this._cursorInfo.position))){const{position:_}=this._cursorInfo;this._cursorInfo=void 0;const b=new Map;for(const y of this._editor.getLineDecorations(_.lineNumber)??[]){const x=this._decorationsMetadata.get(y.id);if(y.range.startColumn>_.column)continue;const L=x?.decoration.options[x.item.anchor.direction];if(L&&L.attachedData!==pb._whitespaceData){const E=b.get(x.item)??0;b.set(x.item,E+L.content.length)}}const w=t.filter(y=>y.anchor.range.startLineNumber===_.lineNumber&&y.anchor.range.endColumn<=_.column),C=Array.from(b.values());let v;for(;;){const y=w.shift(),x=C.shift();if(!x&&!y)break;if(y)i.set(y,x??0),v=y;else if(v&&x){let L=i.get(v);L+=x,L+=C.reduce((E,I)=>E+I,0),C.length=0;break}}}const n=[],o=(_,b,w,C,v)=>{const y={content:w,inlineClassNameAffectsLetterSpacing:!0,inlineClassName:b.className,cursorStops:C,attachedData:v};n.push({item:_,classNameRef:b,decoration:{range:_.anchor.range,options:{description:"InlayHint",showIfCollapsed:_.anchor.range.isEmpty(),collapseOnReplaceEdit:!_.anchor.range.isEmpty(),stickiness:0,[_.anchor.direction]:this._activeRenderMode===0?y:void 0}}})},r=(_,b)=>{const w=this._ruleFactory.createClassNameRef({width:`${a/3|0}px`,display:"inline-block"});o(_,w," ",b?jr.Right:jr.None,pb._whitespaceData)},{fontSize:a,fontFamily:l,padding:c,isUniform:d}=this._getLayoutInfo(),h=this._editor.getOption(159).maximumLength,u="--code-editorInlayHintsFontFamily";this._editor.getContainerDomNode().style.setProperty(u,l);let g={line:0,totalLen:0};for(let _=0;_<t.length;_++){const b=t[_];if(g.line!==b.anchor.range.startLineNumber&&(g={line:b.anchor.range.startLineNumber,totalLen:0}),h&&g.totalLen>h)continue;b.hint.paddingLeft&&r(b,!1);const w=typeof b.hint.label=="string"?[{label:b.hint.label}]:b.hint.label,C=i.get(b);let v=0;for(let y=0;y<w.length;y++){const x=w[y],L=y===0,E=y===w.length-1,I={fontSize:`${a}px`,fontFamily:`var(${u}), ${_o.fontFamily}`,verticalAlign:d?"baseline":"middle",unicodeBidi:"isolate"};Ms(b.hint.textEdits)&&(I.cursor="default"),this._fillInColors(I,b.hint),(x.command||x.location)&&this._activeInlayHintPart?.part.item===b&&this._activeInlayHintPart.part.index===y&&(I.textDecoration="underline",this._activeInlayHintPart.hasTriggerModifier&&(I.color=Ci(ame),I.cursor="pointer"));let R=x.label;g.totalLen+=R.length;let O=!1;const P=h!==0?g.totalLen-h:0;if(P>0&&(R=R.slice(0,-P)+"…",O=!0),v+=R.length,C!==void 0){const B=v-C;B>=0&&(v-=B,R=R.slice(0,-(1+B))+"…",O=!0)}if(c&&(L&&(E||O)?(I.padding=`1px ${Math.max(1,a/4)|0}px`,I.borderRadius=`${a/4|0}px`):L?(I.padding=`1px 0 1px ${Math.max(1,a/4)|0}px`,I.borderRadius=`${a/4|0}px 0 0 ${a/4|0}px`):E||O?(I.padding=`1px ${Math.max(1,a/4)|0}px 1px 0`,I.borderRadius=`0 ${a/4|0}px ${a/4|0}px 0`):I.padding="1px 0 1px 0"),o(b,this._ruleFactory.createClassNameRef(I),kRe(R),E&&!b.hint.paddingRight?jr.Right:jr.None,new R5(b,y)),O)break}if(C!==void 0&&v<C){const y=C-v;o(b,this._ruleFactory.createClassNameRef({})," ".repeat(y),jr.None)}if(b.hint.paddingRight&&r(b,!0),n.length>pb._MAX_DECORATORS)break}const f=[];for(const[_,b]of this._decorationsMetadata){const w=this._editor.getModel()?.getDecorationRange(_);w&&e.some(C=>C.containsRange(w))&&(f.push(_),b.classNameRef.dispose(),this._decorationsMetadata.delete(_))}const p=Il.capture(this._editor);this._editor.changeDecorations(_=>{const b=_.deltaDecorations(f,n.map(w=>w.decoration));for(let w=0;w<b.length;w++){const C=n[w];this._decorationsMetadata.set(b[w],C)}}),p.restore(this._editor)}_fillInColors(e,t){t.kind===nI.Parameter?(e.backgroundColor=Ci(pme),e.color=Ci(fme)):t.kind===nI.Type?(e.backgroundColor=Ci(gme),e.color=Ci(ume)):(e.backgroundColor=Ci(MW),e.color=Ci(AW))}_getLayoutInfo(){const e=this._editor.getOption(159),t=e.padding,i=this._editor.getOption(61),n=this._editor.getOption(58);let o=e.fontSize;(!o||o<5||o>i)&&(o=i);const r=e.fontFamily||n;return{fontSize:o,fontFamily:r,padding:t,isUniform:!t&&r===n&&o===i}}_removeAllDecorations(){this._editor.removeDecorations(Array.from(this._decorationsMetadata.keys()));for(const e of this._decorationsMetadata.values())e.classNameRef.dispose();this._decorationsMetadata.clear()}},pb=Qd,Qd.ID="editor.contrib.InlayHints",Qd._MAX_DECORATORS=1500,Qd._whitespaceData={},Qd);JS=pb=SRe([Z_(1,ge),Z_(2,ga),Z_(3,Ase),Z_(4,Gt),Z_(5,Ii),Z_(6,_e)],JS);function kRe(s){return s.replace(/[ \t]/g," ")}at.registerCommand("_executeInlayHintProvider",async(s,...e)=>{const[t,i]=e;dt(Le.isUri(t)),dt(k.isIRange(i));const{inlayHintsProvider:n}=s.get(ge),o=await s.get(Fs).createModelReference(t);try{const r=await QI.create(n,o.object.textEditorModel,[k.lift(i)],Ke.None),a=r.items.map(l=>l.hint);return setTimeout(()=>r.dispose(),0),a}finally{o.dispose()}});var ERe=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},jg=function(s,e){return function(t,i){e(t,i,s)}};class eG extends g1{constructor(e,t,i,n){super(10,t,e.item.anchor.range,i,n,!0),this.part=e}}let uD=class extends KS{constructor(e,t,i,n,o,r,a,l){super(e,t,o,a,i,n,l),this._resolverService=r,this.hoverOrdinal=6}suggestHoverAnchor(e){if(!JS.get(this._editor)||e.target.type!==6)return null;const i=e.target.detail.injectedText?.options;return i instanceof mg&&i.attachedData instanceof R5?new eG(i.attachedData,this,e.event.posx,e.event.posy):null}computeSync(){return[]}computeAsync(e,t,i,n){return e instanceof eG?new Jr(async o=>{const{part:r}=e;if(await r.item.resolve(n),n.isCancellationRequested)return;let a;typeof r.item.hint.tooltip=="string"?a=new hs().appendText(r.item.hint.tooltip):r.item.hint.tooltip&&(a=r.item.hint.tooltip),a&&o.emitOne(new Ta(this,e.range,[a],!1,0)),Ms(r.item.hint.textEdits)&&o.emitOne(new Ta(this,e.range,[new hs().appendText(m(1164,"Double-click to insert"))],!1,10001));let l;if(typeof r.part.tooltip=="string"?l=new hs().appendText(r.part.tooltip):r.part.tooltip&&(l=r.part.tooltip),l&&o.emitOne(new Ta(this,e.range,[l],!1,1)),r.part.location||r.part.command){let d;const u=this._editor.getOption(86)==="altKey"?Ye?m(1165,"cmd + click"):m(1166,"ctrl + click"):Ye?m(1167,"option + click"):m(1168,"alt + click");r.part.location&&r.part.command?d=new hs().appendText(m(1169,"Go to Definition ({0}), right click for more",u)):r.part.location?d=new hs().appendText(m(1170,"Go to Definition ({0})",u)):r.part.command&&(d=new hs(`[${m(1171,"Execute Command")}](${XTe(r.part.command)} "${r.part.command.title}") (${u})`,{isTrusted:!0})),d&&o.emitOne(new Ta(this,e.range,[d],!1,1e4))}const c=this._resolveInlayHintLabelPartHover(r,n);for await(const d of c)o.emitOne(d)}):Jr.EMPTY}async*_resolveInlayHintLabelPartHover(e,t){if(!e.part.location)return;const{uri:i,range:n}=e.part.location,o=await this._resolverService.createModelReference(i);try{const r=o.object.textEditorModel;if(!this._languageFeaturesService.hoverProvider.has(r))return;for await(const a of hH(this._languageFeaturesService.hoverProvider,r,new A(n.startLineNumber,n.startColumn),t))wv(a.hover.contents)||(yield new Ta(this,e.item.anchor.range,a.hover.contents,!1,2+a.ordinal))}finally{o.dispose()}}};uD=ERe([jg(1,Ua),jg(2,ft),jg(3,so),jg(4,He),jg(5,Fs),jg(6,ge),jg(7,Gt)],uD);var Mse=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},gD=function(s,e){return function(t,i){e(t,i,s)}};class tG{constructor(e,t,i){this.marker=e,this.index=t,this.total=i}}let A5=class{constructor(e,t,i){this._markerService=t,this._configService=i,this._onDidChange=new M,this.onDidChange=this._onDidChange.event,this._dispoables=new U,this._markers=[],this._nextIdx=-1,Le.isUri(e)?this._resourceFilter=a=>a.toString()===e.toString():e&&(this._resourceFilter=e);const n=this._configService.getValue("problems.sortOrder"),o=(a,l)=>{let c=qy(a.resource.toString(),l.resource.toString());return c===0&&(n==="position"?c=k.compareRangesUsingStarts(a,l)||mi.compare(a.severity,l.severity):c=mi.compare(a.severity,l.severity)||k.compareRangesUsingStarts(a,l)),c},r=()=>{let a=this._markerService.read({resource:Le.isUri(e)?e:void 0,severities:mi.Error|mi.Warning|mi.Info});return typeof e=="function"&&(a=a.filter(l=>this._resourceFilter(l.resource))),a.sort(o),ni(a,this._markers,(l,c)=>l.resource.toString()===c.resource.toString()&&l.startLineNumber===c.startLineNumber&&l.startColumn===c.startColumn&&l.endLineNumber===c.endLineNumber&&l.endColumn===c.endColumn&&l.severity===c.severity&&l.message===c.message)?!1:(this._markers=a,!0)};r(),this._dispoables.add(t.onMarkerChanged(a=>{(!this._resourceFilter||a.some(l=>this._resourceFilter(l)))&&r()&&(this._nextIdx=-1,this._onDidChange.fire())}))}dispose(){this._dispoables.dispose(),this._onDidChange.dispose()}matches(e){return!this._resourceFilter&&!e?!0:!this._resourceFilter||!e?!1:this._resourceFilter(e)}get selected(){const e=this._markers[this._nextIdx];return e&&new tG(e,this._nextIdx+1,this._markers.length)}_initIdx(e,t,i){let n=this._markers.findIndex(o=>dg(o.resource,e.uri));if(n<0)n=J6(this._markers.length,o=>qy(this._markers[o].resource.toString(),e.uri.toString())),n<0&&(n=~n),i?this._nextIdx=n:this._nextIdx=(this._markers.length+n-1)%this._markers.length;else{let o=!1,r=!1;for(let a=n;a<this._markers.length;a++){let l=k.lift(this._markers[a]);if(l.isEmpty()){const c=e.getWordAtPosition(l.getStartPosition());c&&(l=new k(l.startLineNumber,c.startColumn,l.startLineNumber,c.endColumn))}if(t&&(l.containsPosition(t)||t.isBeforeOrEqual(l.getStartPosition()))){this._nextIdx=a,o=!0,r=!l.containsPosition(t);break}if(this._markers[a].resource.toString()!==e.uri.toString())break}o?r&&!i&&(this._nextIdx-=1):this._nextIdx=i?0:this._markers.length-1}this._nextIdx<0&&(this._nextIdx=this._markers.length-1)}resetIndex(){this._nextIdx=-1}move(e,t,i){if(this._markers.length===0)return!1;const n=this._nextIdx;return this._nextIdx===-1?this._initIdx(t,i,e):e?this._nextIdx=(this._nextIdx+1)%this._markers.length:e||(this._nextIdx=(this._nextIdx-1+this._markers.length)%this._markers.length),n!==this._nextIdx}find(e,t){let i=this._markers.findIndex(n=>n.resource.toString()===e.toString());if(!(i<0)){for(;i<this._markers.length;i++)if(k.containsPosition(this._markers[i],t))return new tG(this._markers[i],i+1,this._markers.length)}}};A5=Mse([gD(1,jc),gD(2,He)],A5);const Ose=qe("IMarkerNavigationService");let M5=class{constructor(e,t){this._markerService=e,this._configService=t,this._provider=new Ns}getMarkerList(e){for(const t of this._provider){const i=t.getMarkerList(e);if(i)return i}return new A5(e,this._markerService,this._configService)}};M5=Mse([gD(0,jc),gD(1,He)],M5);tt(Ose,M5,1);var O5;(function(s){function e(t){switch(t){case fi.Ignore:return"severity-ignore "+xe.asClassName(Z.info);case fi.Info:return xe.asClassName(Z.info);case fi.Warning:return xe.asClassName(Z.warning);case fi.Error:return xe.asClassName(Z.error);default:return""}}s.className=e})(O5||(O5={}));var IRe=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},X_=function(s,e){return function(t,i){e(t,i,s)}},P5;class DRe{constructor(e,t,i,n,o){this._openerService=n,this._labelService=o,this._lines=0,this._longestLineLength=0,this._relatedDiagnostics=new WeakMap,this._disposables=new U,this._editor=t;const r=document.createElement("div");r.className="descriptioncontainer",this._messageBlock=document.createElement("div"),this._messageBlock.classList.add("message"),this._messageBlock.setAttribute("aria-live","assertive"),this._messageBlock.setAttribute("role","alert"),r.appendChild(this._messageBlock),this._relatedBlock=document.createElement("div"),r.appendChild(this._relatedBlock),this._disposables.add(Fi(this._relatedBlock,"click",a=>{a.preventDefault();const l=this._relatedDiagnostics.get(a.target);l&&i(l)})),this._scrollable=new tee(r,{horizontal:1,vertical:1,useShadows:!1,horizontalScrollbarSize:6,verticalScrollbarSize:6}),e.appendChild(this._scrollable.getDomNode()),this._disposables.add(this._scrollable.onScroll(a=>{r.style.left=`-${a.scrollLeft}px`,r.style.top=`-${a.scrollTop}px`})),this._disposables.add(this._scrollable)}dispose(){xt(this._disposables)}update(e){const{source:t,message:i,relatedInformation:n,code:o}=e;let r=(t?.length||0)+2;o&&(typeof o=="string"?r+=o.length:r+=o.value.length);const a=qo(i);this._lines=a.length,this._longestLineLength=0;for(const u of a)this._longestLineLength=Math.max(u.length+r,this._longestLineLength);Pn(this._messageBlock),this._messageBlock.setAttribute("aria-label",this.getAriaLabel(e)),this._editor.applyFontInfo(this._messageBlock);let l=this._messageBlock;for(const u of a)l=document.createElement("div"),l.innerText=u,u===""&&(l.style.height=this._messageBlock.style.lineHeight),this._messageBlock.appendChild(l);if(t||o){const u=document.createElement("span");if(u.classList.add("details"),l.appendChild(u),t){const g=document.createElement("span");g.innerText=t,g.classList.add("source"),u.appendChild(g)}if(o)if(typeof o=="string"){const g=document.createElement("span");g.innerText=`(${o})`,g.classList.add("code"),u.appendChild(g)}else{this._codeLink=ie("a.code-link"),this._codeLink.setAttribute("href",`${o.target.toString()}`),this._codeLink.onclick=f=>{this._openerService.open(o.target,{allowCommands:!0}),f.preventDefault(),f.stopPropagation()};const g=X(this._codeLink,ie("span"));g.innerText=o.value,u.appendChild(this._codeLink)}}if(Pn(this._relatedBlock),this._editor.applyFontInfo(this._relatedBlock),Ms(n)){const u=this._relatedBlock.appendChild(document.createElement("div"));u.style.paddingTop=`${Math.floor(this._editor.getOption(75)*.66)}px`,this._lines+=1;for(const g of n){const f=document.createElement("div"),p=document.createElement("a");p.classList.add("filename"),p.innerText=`${this._labelService.getUriBasenameLabel(g.resource)}(${g.startLineNumber}, ${g.startColumn}): `,p.title=this._labelService.getUriLabel(g.resource),this._relatedDiagnostics.set(p,g);const _=document.createElement("span");_.innerText=g.message,f.appendChild(p),f.appendChild(_),this._lines+=1,u.appendChild(f)}}const c=this._editor.getOption(59),d=Math.ceil(c.typicalFullwidthCharacterWidth*this._longestLineLength*.75),h=c.lineHeight*this._lines;this._scrollable.setScrollDimensions({scrollWidth:d,scrollHeight:h})}layout(e,t){this._scrollable.getDomNode().style.height=`${e}px`,this._scrollable.getDomNode().style.width=`${t}px`,this._scrollable.setScrollDimensions({width:t,height:e})}getHeightInLines(){return Math.min(17,this._lines)}getAriaLabel(e){let t="";switch(e.severity){case mi.Error:t=m(1024,"Error");break;case mi.Warning:t=m(1025,"Warning");break;case mi.Info:t=m(1026,"Info");break;case mi.Hint:t=m(1027,"Hint");break}let i=m(1028,"{0} at {1}. ",t,e.startLineNumber+":"+e.startColumn);const n=this._editor.getModel();return n&&e.startLineNumber<=n.getLineCount()&&e.startLineNumber>=1&&(i=`${n.getLineContent(e.startLineNumber)}, ${i}`),i}}var Bp;let e0=(Bp=class extends eD{constructor(e,t,i,n,o,r,a){super(e,{showArrow:!0,showFrame:!0,isAccessible:!0,frameWidth:1},o),this._themeService=t,this._openerService=i,this._menuService=n,this._contextKeyService=r,this._labelService=a,this._callOnDispose=new U,this._onDidSelectRelatedInformation=new M,this.onDidSelectRelatedInformation=this._onDidSelectRelatedInformation.event,this._severity=mi.Warning,this._backgroundColor=q.white,this._applyTheme(t.getColorTheme()),this._callOnDispose.add(t.onDidColorThemeChange(this._applyTheme.bind(this))),this.create()}_applyTheme(e){this._backgroundColor=e.getColor(ARe);let t=F5,i=TRe;this._severity===mi.Warning?(t=Kk,i=NRe):this._severity===mi.Info&&(t=B5,i=RRe);const n=e.getColor(t),o=e.getColor(i);this.style({arrowColor:n,frameColor:n,headerBackgroundColor:o,primaryHeadingColor:e.getColor(nse),secondaryHeadingColor:e.getColor(sse)})}_applyStyles(){this._parentContainer&&(this._parentContainer.style.backgroundColor=this._backgroundColor?this._backgroundColor.toString():""),super._applyStyles()}dispose(){this._callOnDispose.dispose(),super.dispose()}_fillHead(e){super._fillHead(e),this._disposables.add(this._actionbarWidget.actionRunner.onWillRun(n=>this.editor.focus()));const t=this._menuService.getMenuActions(P5.TitleMenu,this._contextKeyService),i=sEe(t);this._actionbarWidget.push(i,{label:!1,icon:!0,index:0})}_fillTitleIcon(e){this._icon=X(e,ie(""))}_fillBody(e){this._parentContainer=e,e.classList.add("marker-widget"),this._parentContainer.tabIndex=0,this._parentContainer.setAttribute("role","tooltip"),this._container=document.createElement("div"),e.appendChild(this._container),this._message=new DRe(this._container,this.editor,t=>this._onDidSelectRelatedInformation.fire(t),this._openerService,this._labelService),this._disposables.add(this._message)}show(){throw new Error("call showAtMarker")}showAtMarker(e,t,i){this._container.classList.remove("stale"),this._message.update(e),this._severity=e.severity,this._applyTheme(this._themeService.getColorTheme());const n=k.lift(e),o=this.editor.getPosition(),r=o&&n.containsPosition(o)?o:n.getStartPosition();super.show(r,this.computeRequiredHeight());const a=this.editor.getModel();if(a){const l=i>1?m(1029,"{0} of {1} problems",t,i):m(1030,"{0} of {1} problem",t,i);this.setTitle(ra(a.uri),l)}this._icon.className=`codicon ${O5.className(mi.toSeverity(this._severity))}`,this.editor.revealPositionNearTop(r,0),this.editor.focus()}updateMarker(e){this._container.classList.remove("stale"),this._message.update(e)}showStale(){this._container.classList.add("stale"),this._relayout()}_doLayoutBody(e,t){super._doLayoutBody(e,t),this._heightInPixel=e,this._message.layout(e,t),this._container.style.height=`${e}px`}_onWidth(e){this._message.layout(this._heightInPixel,e)}_relayout(){super._relayout(this.computeRequiredHeight())}computeRequiredHeight(){return 3+this._message.getHeightInLines()}},P5=Bp,Bp.TitleMenu=new fe("gotoErrorTitleMenu"),Bp);e0=P5=IRe([X_(1,wi),X_(2,Ah),X_(3,da),X_(4,_e),X_(5,Ae),X_(6,Ym)],e0);const iG=rS(iA,sme),nG=rS(_h,aS),sG=rS(Rc,lS),F5=T("editorMarkerNavigationError.background",{dark:iG,light:iG,hcDark:rt,hcLight:rt},m(1031,"Editor marker navigation widget error color.")),TRe=T("editorMarkerNavigationError.headerBackground",{dark:Be(F5,.1),light:Be(F5,.1),hcDark:null,hcLight:null},m(1032,"Editor marker navigation widget error heading background.")),Kk=T("editorMarkerNavigationWarning.background",{dark:nG,light:nG,hcDark:rt,hcLight:rt},m(1033,"Editor marker navigation widget warning color.")),NRe=T("editorMarkerNavigationWarning.headerBackground",{dark:Be(Kk,.1),light:Be(Kk,.1),hcDark:"#0C141F",hcLight:Be(Kk,.2)},m(1034,"Editor marker navigation widget warning heading background.")),B5=T("editorMarkerNavigationInfo.background",{dark:sG,light:sG,hcDark:rt,hcLight:rt},m(1035,"Editor marker navigation widget info color.")),RRe=T("editorMarkerNavigationInfo.headerBackground",{dark:Be(B5,.1),light:Be(B5,.1),hcDark:null,hcLight:null},m(1036,"Editor marker navigation widget info heading background.")),ARe=T("editorMarkerNavigation.background",Bi,m(1037,"Editor marker navigation widget background."));var MRe=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},FL=function(s,e){return function(t,i){e(t,i,s)}},p1,Wp;let Xm=(Wp=class{static get(e){return e.getContribution(p1.ID)}constructor(e,t,i,n,o){this._markerNavigationService=t,this._contextKeyService=i,this._editorService=n,this._instantiationService=o,this._sessionDispoables=new U,this._editor=e,this._widgetVisible=Pse.bindTo(this._contextKeyService)}dispose(){this._cleanUp(),this._sessionDispoables.dispose()}_cleanUp(){this._widgetVisible.reset(),this._sessionDispoables.clear(),this._widget=void 0,this._model=void 0}_getOrCreateModel(e){if(this._model&&this._model.matches(e))return this._model;let t=!1;return this._model&&(t=!0,this._cleanUp()),this._model=this._markerNavigationService.getMarkerList(e),t&&this._model.move(!0,this._editor.getModel(),this._editor.getPosition()),this._widget=this._instantiationService.createInstance(e0,this._editor),this._widget.onDidClose(()=>this.close(),this,this._sessionDispoables),this._widgetVisible.set(!0),this._sessionDispoables.add(this._model),this._sessionDispoables.add(this._widget),this._sessionDispoables.add(this._editor.onDidChangeCursorPosition(i=>{(!this._model?.selected||!k.containsPosition(this._model?.selected.marker,i.position))&&this._model?.resetIndex()})),this._sessionDispoables.add(this._model.onDidChange(()=>{if(!this._widget||!this._widget.position||!this._model)return;const i=this._model.find(this._editor.getModel().uri,this._widget.position);i?this._widget.updateMarker(i.marker):this._widget.showStale()})),this._sessionDispoables.add(this._widget.onDidSelectRelatedInformation(i=>{this._editorService.openCodeEditor({resource:i.resource,options:{pinned:!0,revealIfOpened:!0,selection:k.lift(i).collapseToStart()}},this._editor),this.close(!1)})),this._sessionDispoables.add(this._editor.onDidChangeModel(()=>this._cleanUp())),this._model}close(e=!0){this._cleanUp(),e&&this._editor.focus()}showAtMarker(e){if(!this._editor.hasModel())return;const t=this._editor.getModel(),i=this._getOrCreateModel(t.uri);i.resetIndex(),i.move(!0,t,new A(e.startLineNumber,e.startColumn)),i.selected&&this._widget.showAtMarker(i.selected.marker,i.selected.index,i.selected.total)}async navigate(e,t){if(!this._editor.hasModel())return;const i=this._editor.getModel(),n=this._getOrCreateModel(t?void 0:i.uri);if(n.move(e,i,this._editor.getPosition()),!!n.selected)if(n.selected.marker.resource.toString()!==i.uri.toString()){this._cleanUp();const o=await this._editorService.openCodeEditor({resource:n.selected.marker.resource,options:{pinned:!1,revealIfOpened:!0,selectionRevealType:2,selection:n.selected.marker}},this._editor);o&&(p1.get(o)?.close(),p1.get(o)?.navigate(e,t))}else this._widget.showAtMarker(n.selected.marker,n.selected.index,n.selected.total)}},p1=Wp,Wp.ID="editor.contrib.markerController",Wp);Xm=p1=MRe([FL(1,Ose),FL(2,Ae),FL(3,ht),FL(4,_e)],Xm);class $A extends ue{constructor(e,t,i){super(i),this._next=e,this._multiFile=t}async run(e,t){t.hasModel()&&await Xm.get(t)?.navigate(this._next,this._multiFile)}}const Of=class Of extends $A{constructor(){super(!0,!1,{id:Of.ID,label:Of.LABEL,precondition:void 0,kbOpts:{kbExpr:N.focus,primary:578,weight:100},menuOpts:{menuId:e0.TitleMenu,title:Of.LABEL.value,icon:Qt("marker-navigation-next",Z.arrowDown,m(1016,"Icon for goto next marker.")),group:"navigation",order:1}})}};Of.ID="editor.action.marker.next",Of.LABEL=V(1020,"Go to Next Problem (Error, Warning, Info)");let fD=Of;const Pf=class Pf extends $A{constructor(){super(!1,!1,{id:Pf.ID,label:Pf.LABEL,precondition:void 0,kbOpts:{kbExpr:N.focus,primary:1602,weight:100},menuOpts:{menuId:e0.TitleMenu,title:Pf.LABEL.value,icon:Qt("marker-navigation-previous",Z.arrowUp,m(1017,"Icon for goto previous marker.")),group:"navigation",order:2}})}};Pf.ID="editor.action.marker.prev",Pf.LABEL=V(1021,"Go to Previous Problem (Error, Warning, Info)");let W5=Pf;class ORe extends $A{constructor(){super(!0,!0,{id:"editor.action.marker.nextInFiles",label:V(1022,"Go to Next Problem in Files (Error, Warning, Info)"),precondition:void 0,kbOpts:{kbExpr:N.focus,primary:66,weight:100},menuOpts:{menuId:fe.MenubarGoMenu,title:m(1018,"Next &&Problem"),group:"6_problem_nav",order:1}})}}class PRe extends $A{constructor(){super(!1,!0,{id:"editor.action.marker.prevInFiles",label:V(1023,"Go to Previous Problem in Files (Error, Warning, Info)"),precondition:void 0,kbOpts:{kbExpr:N.focus,primary:1090,weight:100},menuOpts:{menuId:fe.MenubarGoMenu,title:m(1019,"Previous &&Problem"),group:"6_problem_nav",order:2}})}}lt(Xm.ID,Xm,4);ae(fD);ae(W5);ae(ORe);ae(PRe);const Pse=new ce("markersNavigationVisible",!1),FRe=hn.bindToContribution(Xm.get);le(new FRe({id:"closeMarkersNavigation",precondition:Pse,handler:s=>s.close(),kbOpts:{weight:150,kbExpr:N.focus,primary:9,secondary:[1033]}}));var BRe=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},sO=function(s,e){return function(t,i){e(t,i,s)}};const pa=ie;class Fse{constructor(e,t,i){this.owner=e,this.range=t,this.marker=i}isValidForHoverAnchor(e){return e.type===1&&this.range.startColumn<=e.range.startColumn&&this.range.endColumn>=e.range.endColumn}}const oG={type:1,filter:{include:ii.QuickFix},triggerAction:Yo.QuickFixHover};let H5=class{constructor(e,t,i,n){this._editor=e,this._markerDecorationsService=t,this._openerService=i,this._languageFeaturesService=n,this.hoverOrdinal=1,this.recentMarkerCodeActionsInfo=void 0}computeSync(e,t){if(!this._editor.hasModel()||e.type!==1&&!e.supportsMarkerHover)return[];const i=this._editor.getModel(),n=e.range;if(!i.isValidRange(e.range))return[];const o=n.startLineNumber,r=i.getLineMaxColumn(o),a=[];for(const l of t){const c=l.range.startLineNumber===o?l.range.startColumn:1,d=l.range.endLineNumber===o?l.range.endColumn:r,h=this._markerDecorationsService.getMarker(i.uri,l);if(!h)continue;const u=new k(e.range.startLineNumber,c,e.range.startLineNumber,d);a.push(new Fse(this,u,h))}return a}renderHoverParts(e,t){if(!t.length)return new qm([]);const i=[];t.forEach(r=>{const a=this._renderMarkerHover(r);e.fragment.appendChild(a.hoverElement),i.push(a)});const n=t.length===1?t[0]:t.sort((r,a)=>mi.compare(r.marker.severity,a.marker.severity))[0],o=this._renderMarkerStatusbar(e,n);return new qm(i,o)}getAccessibleContent(e){return e.marker.message}_renderMarkerHover(e){const t=new U,i=pa("div.hover-row"),n=X(i,pa("div.marker.hover-contents")),{source:o,message:r,code:a,relatedInformation:l}=e.marker;this._editor.applyFontInfo(n);const c=X(n,pa("span"));if(c.style.whiteSpace="pre-wrap",c.innerText=r,o||a)if(a&&typeof a!="string"){const h=pa("span");if(o){const p=X(h,pa("span"));p.innerText=o}const u=X(h,pa("a.code-link"));u.setAttribute("href",a.target.toString(!0)),t.add(W(u,"click",p=>{this._openerService.open(a.target,{allowCommands:!0}),p.preventDefault(),p.stopPropagation()}));const g=X(u,pa("span"));g.innerText=a.value;const f=X(n,h);f.style.opacity="0.6",f.style.paddingLeft="6px"}else{const h=X(n,pa("span"));h.style.opacity="0.6",h.style.paddingLeft="6px",h.innerText=o&&a?`${o}(${a})`:o||`(${a})`}if(Ms(l))for(const{message:h,resource:u,startLineNumber:g,startColumn:f}of l){const p=X(n,pa("div"));p.style.marginTop="8px";const _=X(p,pa("a"));_.innerText=`${ra(u)}(${g}, ${f}): `,_.style.cursor="pointer",t.add(W(_,"click",w=>{if(w.stopPropagation(),w.preventDefault(),this._openerService){const C={selection:{startLineNumber:g,startColumn:f}};this._openerService.open(u,{fromUserGesture:!0,editorOptions:C}).catch(Oe)}}));const b=X(p,pa("span"));b.innerText=h,this._editor.applyFontInfo(b)}return{hoverPart:e,hoverElement:i,dispose:()=>t.dispose()}}_renderMarkerStatusbar(e,t){const i=new U;if(t.marker.severity===mi.Error||t.marker.severity===mi.Warning||t.marker.severity===mi.Info){const n=Xm.get(this._editor);n&&e.statusBar.addAction({label:m(1139,"View Problem"),commandId:fD.ID,run:()=>{e.hide(),n.showAtMarker(t.marker),this._editor.focus()}})}if(!this._editor.getOption(104)){const n=e.statusBar.append(pa("div"));this.recentMarkerCodeActionsInfo&&(jI.makeKey(this.recentMarkerCodeActionsInfo.marker)===jI.makeKey(t.marker)?this.recentMarkerCodeActionsInfo.hasCodeActions||(n.textContent=m(1140,"No quick fixes available")):this.recentMarkerCodeActionsInfo=void 0);const o=this.recentMarkerCodeActionsInfo&&!this.recentMarkerCodeActionsInfo.hasCodeActions?F.None:hh(()=>n.textContent=m(1141,"Checking for quick fixes..."),200,i);n.textContent||(n.textContent=" ");const r=this.getCodeActions(t.marker);i.add(pe(()=>r.cancel())),r.then(a=>{if(o.dispose(),this.recentMarkerCodeActionsInfo={marker:t.marker,hasCodeActions:a.validActions.length>0},!this.recentMarkerCodeActionsInfo.hasCodeActions){a.dispose(),n.textContent=m(1142,"No quick fixes available");return}n.style.display="none";let l=!1;i.add(pe(()=>{l||a.dispose()})),e.statusBar.addAction({label:m(1143,"Quick Fix..."),commandId:tH,run:d=>{l=!0;const h=Um.get(this._editor),u=Li(d);e.hide(),h?.showCodeActions(oG,a,{x:u.left,y:u.top,width:u.width,height:u.height})}});const c=a.validActions.find(d=>d.action.isAI);c&&e.statusBar.addAction({label:c.action.title,commandId:c.action.command?.id??"",iconClass:xe.asClassName(Z.sparkle),run:()=>{Um.get(this._editor)?.applyCodeAction(c,!1,!1,Lu.FromProblemsHover)}}),e.onContentsChanged()},Oe)}return i}getCodeActions(e){return an(t=>Ab(this._languageFeaturesService.codeActionProvider,this._editor.getModel(),new k(e.startLineNumber,e.startColumn,e.endLineNumber,e.endColumn),oG,ml.None,t))}};H5=BRe([sO(1,EW),sO(2,Ah),sO(3,ge)],H5);var Bse=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},kw=function(s,e){return function(t,i){e(t,i,s)}},V5,z5;let $5=V5=class extends F{constructor(e,t,i,n,o,r,a){super();const l=t.hoverParts;this._renderedHoverParts=this._register(new U5(e,i,l,n,o,r,a));const c=t.options,d=c.anchor,{showAtPosition:h,showAtSecondaryPosition:u}=V5.computeHoverPositions(e,d.range,l);this.shouldAppearBeforeContent=l.some(g=>g.isBeforeContent),this.showAtPosition=h,this.showAtSecondaryPosition=u,this.initialMousePosX=d.initialMousePosX,this.initialMousePosY=d.initialMousePosY,this.shouldFocus=c.shouldFocus,this.source=c.source}get domNode(){return this._renderedHoverParts.domNode}get domNodeHasChildren(){return this._renderedHoverParts.domNodeHasChildren}get focusedHoverPartIndex(){return this._renderedHoverParts.focusedHoverPartIndex}get hoverPartsCount(){return this._renderedHoverParts.hoverPartsCount}focusHoverPartWithIndex(e){this._renderedHoverParts.focusHoverPartWithIndex(e)}async updateHoverVerbosityLevel(e,t,i){this._renderedHoverParts.updateHoverVerbosityLevel(e,t,i)}isColorPickerVisible(){return this._renderedHoverParts.isColorPickerVisible()}static computeHoverPositions(e,t,i){let n=1;if(e.hasModel()){const d=e._getViewModel(),h=d.coordinatesConverter,u=h.convertModelRangeToViewRange(t),g=d.getLineMinColumn(u.startLineNumber),f=new A(u.startLineNumber,g);n=h.convertViewPositionToModelPosition(f).column}const o=t.startLineNumber;let r=t.startColumn,a;for(const d of i){const h=d.range,u=h.startLineNumber===o,g=h.endLineNumber===o;if(u&&g){const p=h.startColumn,_=Math.min(r,p);r=Math.max(_,n)}d.forceShowAtRange&&(a=h)}let l,c;if(a){const d=a.getStartPosition();l=d,c=d}else l=t.getStartPosition(),c=new A(o,r);return{showAtPosition:l,showAtSecondaryPosition:c}}};$5=V5=Bse([kw(4,ft),kw(5,so),kw(6,Ko)],$5);class WRe{constructor(e,t){this._statusBar=t,e.appendChild(this._statusBar.hoverElement)}get hoverElement(){return this._statusBar.hoverElement}get actions(){return this._statusBar.actions}dispose(){this._statusBar.dispose()}}var Hp;let U5=(Hp=class extends F{constructor(e,t,i,n,o,r,a){super(),this._hoverService=r,this._clipboardService=a,this._renderedParts=[],this._focusedHoverPartIndex=-1,this._context=n,this._fragment=document.createDocumentFragment(),this._register(this._renderParts(t,i,n,o,this._hoverService)),this._register(this._registerListenersOnRenderedParts()),this._register(this._createEditorDecorations(e,i)),this._updateMarkdownAndColorParticipantInfo(t)}_createEditorDecorations(e,t){if(t.length===0)return F.None;let i=t[0].range;for(const o of t){const r=o.range;i=k.plusRange(i,r)}const n=e.createDecorationsCollection();return n.set([{range:i,options:z5._DECORATION_OPTIONS}]),pe(()=>{n.clear()})}_renderParts(e,t,i,n,o){const r=new YI(n,o),a={fragment:this._fragment,statusBar:r,...i},l=new U;l.add(r);for(const d of e){const h=this._renderHoverPartsForParticipant(t,d,a);l.add(h);for(const u of h.renderedHoverParts)this._renderedParts.push({type:"hoverPart",participant:d,hoverPart:u.hoverPart,hoverElement:u.hoverElement})}const c=this._renderStatusBar(this._fragment,r);return c&&(l.add(c),this._renderedParts.push({type:"statusBar",hoverElement:c.hoverElement,actions:c.actions})),l}_renderHoverPartsForParticipant(e,t,i){const n=e.filter(r=>r.owner===t);return n.length>0?t.renderHoverParts(i,n):new qm([])}_renderStatusBar(e,t){if(t.hasContent)return new WRe(e,t)}_registerListenersOnRenderedParts(){const e=new U;return this._renderedParts.forEach((t,i)=>{const n=t.hoverElement;n.tabIndex=0,e.add(W(n,se.FOCUS_IN,o=>{o.stopPropagation(),this._focusedHoverPartIndex=i})),e.add(W(n,se.FOCUS_OUT,o=>{o.stopPropagation(),this._focusedHoverPartIndex=-1})),t.type==="hoverPart"&&t.hoverPart instanceof Fse&&e.add(new u5(n,()=>t.participant.getAccessibleContent(t.hoverPart),this._clipboardService,this._hoverService))}),e}_updateMarkdownAndColorParticipantInfo(e){const t=e.find(i=>i instanceof KS&&!(i instanceof uD));t&&(this._markdownHoverParticipant=t),this._colorHoverParticipant=e.find(i=>i instanceof XI)}focusHoverPartWithIndex(e){e<0||e>=this._renderedParts.length||this._renderedParts[e].hoverElement.focus()}async updateHoverVerbosityLevel(e,t,i){if(!this._markdownHoverParticipant)return;let n;t>=0?n={start:t,endExclusive:t+1}:n=this._findRangeOfMarkdownHoverParts(this._markdownHoverParticipant);for(let o=n.start;o<n.endExclusive;o++){const r=this._normalizedIndexToMarkdownHoverIndexRange(this._markdownHoverParticipant,o);if(r===void 0)continue;const a=await this._markdownHoverParticipant.updateMarkdownHoverVerbosityLevel(e,r);a&&(this._renderedParts[o]={type:"hoverPart",participant:this._markdownHoverParticipant,hoverPart:a.hoverPart,hoverElement:a.hoverElement})}i&&(t>=0?this.focusHoverPartWithIndex(t):this._context.focus()),this._context.onContentsChanged()}isColorPickerVisible(){return this._colorHoverParticipant?.isColorPickerVisible()??!1}_normalizedIndexToMarkdownHoverIndexRange(e,t){const i=this._renderedParts[t];if(!i||i.type!=="hoverPart"||!(i.participant===e))return;const o=this._renderedParts.findIndex(r=>r.type==="hoverPart"&&r.participant===e);if(o===-1)throw new Ce;return t-o}_findRangeOfMarkdownHoverParts(e){const t=this._renderedParts.slice(),i=t.findIndex(r=>r.type==="hoverPart"&&r.participant===e),n=t.reverse().findIndex(r=>r.type==="hoverPart"&&r.participant===e),o=n>=0?t.length-n:n;return{start:i,endExclusive:o+1}}get domNode(){return this._fragment}get domNodeHasChildren(){return this._fragment.hasChildNodes()}get focusedHoverPartIndex(){return this._focusedHoverPartIndex}get hoverPartsCount(){return this._renderedParts.length}},z5=Hp,Hp._DECORATION_OPTIONS=Fe.register({description:"content-hover-highlight",className:"hoverHighlight"}),Hp);U5=z5=Bse([kw(4,ft),kw(5,so),kw(6,Ko)],U5);var HRe=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},BL=function(s,e){return function(t,i){e(t,i,s)}};let q5=class extends F{constructor(e,t,i,n,o){super(),this._editor=e,this._instantiationService=t,this._keybindingService=i,this._hoverService=n,this._clipboardService=o,this._currentResult=null,this._renderedContentHover=this._register(new vt),this._onContentsChanged=this._register(new M),this.onContentsChanged=this._onContentsChanged.event,this._contentHoverWidget=this._register(this._instantiationService.createInstance(r5,this._editor)),this._participants=this._initializeHoverParticipants(),this._hoverOperation=this._register(new xne(this._editor,new KI(this._editor,this._participants))),this._registerListeners()}_initializeHoverParticipants(){const e=[];for(const t of __.getAll()){const i=this._instantiationService.createInstance(t,this._editor);e.push(i)}return e.sort((t,i)=>t.hoverOrdinal-i.hoverOrdinal),this._register(this._contentHoverWidget.onDidResize(()=>{this._participants.forEach(t=>t.handleResize?.())})),this._register(this._contentHoverWidget.onDidScroll(t=>{this._participants.forEach(i=>i.handleScroll?.(t))})),this._register(this._contentHoverWidget.onContentsChanged(()=>{this._participants.forEach(t=>t.handleContentsChanged?.())})),e}_registerListeners(){this._register(this._hoverOperation.onResult(t=>{const i=t.hasLoadingMessage?this._addLoadingMessage(t):t.value;this._withResult(new Ine(i,t.isComplete,t.options))}));const e=this._contentHoverWidget.getDomNode();this._register(Fi(e,"keydown",t=>{t.equals(9)&&this.hide()})),this._register(Fi(e,"mouseleave",t=>{this._onMouseLeave(t)})),this._register(yi.onDidChange(()=>{this._contentHoverWidget.position&&this._currentResult&&this._setCurrentResult(this._currentResult)})),this._register(this._contentHoverWidget.onContentsChanged(()=>{this._onContentsChanged.fire()}))}_startShowingOrUpdateHover(e,t,i,n,o){if(!(this._contentHoverWidget.position&&this._currentResult))return e?(this._startHoverOperationIfNecessary(e,t,i,n,!1),!0):!1;const a=this._editor.getOption(69).sticky,l=o&&this._contentHoverWidget.isMouseGettingCloser(o.event.posx,o.event.posy);return a&&l?(e&&this._startHoverOperationIfNecessary(e,t,i,n,!0),!0):e?this._currentResult&&this._currentResult.options.anchor.equals(e)?!0:this._currentResult&&e.canAdoptVisibleHover(this._currentResult.options.anchor,this._contentHoverWidget.position)?(this._currentResult&&this._setCurrentResult(this._currentResult.filter(e)),this._startHoverOperationIfNecessary(e,t,i,n,!1),!0):(this._setCurrentResult(null),this._startHoverOperationIfNecessary(e,t,i,n,!1),!0):(this._setCurrentResult(null),!1)}_startHoverOperationIfNecessary(e,t,i,n,o){if(this._hoverOperation.options&&this._hoverOperation.options.anchor.equals(e))return;this._hoverOperation.cancel();const a={anchor:e,source:i,shouldFocus:n,insistOnKeepingHoverVisible:o};this._hoverOperation.start(t,a)}_setCurrentResult(e){let t=e;if(this._currentResult===t)return;t&&t.hoverParts.length===0&&(t=null),this._currentResult=t,this._currentResult?this._showHover(this._currentResult):this._hideHover()}_addLoadingMessage(e){for(const t of this._participants){if(!t.createLoadingMessage)continue;const i=t.createLoadingMessage(e.options.anchor);if(i)return e.value.slice(0).concat([i])}return e.value}_withResult(e){if(this._contentHoverWidget.position&&this._currentResult&&this._currentResult.isComplete||this._setCurrentResult(e),!e.isComplete)return;const n=e.hoverParts.length===0,o=e.options.insistOnKeepingHoverVisible;n&&o||this._setCurrentResult(e)}_showHover(e){const t=this._getHoverContext();this._renderedContentHover.value=new $5(this._editor,e,this._participants,t,this._keybindingService,this._hoverService,this._clipboardService),this._renderedContentHover.value.domNodeHasChildren?this._contentHoverWidget.show(this._renderedContentHover.value):this._renderedContentHover.clear()}_hideHover(){this._contentHoverWidget.hide(),this._participants.forEach(e=>e.handleHide?.())}_getHoverContext(){return{hide:()=>{this.hide()},onContentsChanged:()=>{this._contentHoverWidget.handleContentsChanged()},setMinimumDimensions:o=>{this._contentHoverWidget.setMinimumDimensions(o)},focus:()=>this.focus()}}showsOrWillShow(e){if(this._contentHoverWidget.isResizing)return!0;const i=this._findHoverAnchorCandidates(e);if(!(i.length>0))return this._startShowingOrUpdateHover(null,0,0,!1,e);const o=i[0];return this._startShowingOrUpdateHover(o,0,0,!1,e)}_findHoverAnchorCandidates(e){const t=[];for(const n of this._participants){if(!n.suggestHoverAnchor)continue;const o=n.suggestHoverAnchor(e);o&&t.push(o)}const i=e.target;switch(i.type){case 6:{t.push(new KM(0,i.range,e.event.posx,e.event.posy));break}case 7:{const n=this._editor.getOption(59).typicalHalfwidthCharacterWidth/2;if(!(!i.detail.isAfterLines&&typeof i.detail.horizontalDistanceToText=="number"&&i.detail.horizontalDistanceToText<n))break;t.push(new KM(0,i.range,e.event.posx,e.event.posy));break}}return t.sort((n,o)=>o.priority-n.priority),t}_onMouseLeave(e){const t=this._editor.getDomNode();(!t||!TA(t,e.x,e.y))&&this.hide()}startShowingAtRange(e,t,i,n){this._startShowingOrUpdateHover(new KM(0,e,void 0,void 0),t,i,n,null)}async updateHoverVerbosityLevel(e,t,i){this._renderedContentHover.value?.updateHoverVerbosityLevel(e,t,i)}focusedHoverPartIndex(){return this._renderedContentHover.value?.focusedHoverPartIndex??-1}containsNode(e){return e?this._contentHoverWidget.getDomNode().contains(e):!1}focus(){if(this._renderedContentHover.value?.hoverPartsCount===1){this.focusHoverPartWithIndex(0);return}this._contentHoverWidget.focus()}focusHoverPartWithIndex(e){this._renderedContentHover.value?.focusHoverPartWithIndex(e)}scrollUp(){this._contentHoverWidget.scrollUp()}scrollDown(){this._contentHoverWidget.scrollDown()}scrollLeft(){this._contentHoverWidget.scrollLeft()}scrollRight(){this._contentHoverWidget.scrollRight()}pageUp(){this._contentHoverWidget.pageUp()}pageDown(){this._contentHoverWidget.pageDown()}goToTop(){this._contentHoverWidget.goToTop()}goToBottom(){this._contentHoverWidget.goToBottom()}hide(){this._hoverOperation.cancel(),this._setCurrentResult(null)}getDomNode(){return this._contentHoverWidget.getDomNode()}get isColorPickerVisible(){return this._renderedContentHover.value?.isColorPickerVisible()??!1}get isVisibleFromKeyboard(){return this._contentHoverWidget.isVisibleFromKeyboard}get isVisible(){return this._contentHoverWidget.isVisible}get isFocused(){return this._contentHoverWidget.isFocused}get isResizing(){return this._contentHoverWidget.isResizing}get widget(){return this._contentHoverWidget}};q5=HRe([BL(1,_e),BL(2,ft),BL(3,so),BL(4,Ko)],q5);function Wse(s){const e=s.target;return!!e&&e.type===6&&e.detail.injectedText?.options.attachedData===wne}var VRe=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},oO=function(s,e){return function(t,i){e(t,i,s)}},j5,Vp;let ws=(Vp=class extends F{constructor(e,t,i,n){super(),this._editor=e,this._instantiationService=i,this._keybindingService=n,this._onHoverContentsChanged=this._register(new M),this.shouldKeepOpenOnEditorMouseMoveOrLeave=!1,this._listenersStore=new U,this._isMouseDown=!1,this._ignoreMouseEvents=!1,this._reactToEditorMouseMoveRunner=this._register(new Rt(()=>{this._mouseMoveEvent&&this._reactToEditorMouseMove(this._mouseMoveEvent)},0)),this._register(t.onDidShowContextMenu(()=>{this.hideContentHover(),this._ignoreMouseEvents=!0})),this._register(t.onDidHideContextMenu(()=>{this._ignoreMouseEvents=!1})),this._hookListeners(),this._register(this._editor.onDidChangeConfiguration(o=>{o.hasChanged(69)&&(this._unhookListeners(),this._hookListeners())}))}static get(e){return e.getContribution(j5.ID)}_hookListeners(){const e=this._editor.getOption(69);this._hoverSettings={enabled:e.enabled,sticky:e.sticky,hidingDelay:e.hidingDelay},e.enabled||this._cancelSchedulerAndHide(),this._listenersStore.add(this._editor.onMouseDown(t=>this._onEditorMouseDown(t))),this._listenersStore.add(this._editor.onMouseUp(()=>this._onEditorMouseUp())),this._listenersStore.add(this._editor.onMouseMove(t=>this._onEditorMouseMove(t))),this._listenersStore.add(this._editor.onKeyDown(t=>this._onKeyDown(t))),this._listenersStore.add(this._editor.onMouseLeave(t=>this._onEditorMouseLeave(t))),this._listenersStore.add(this._editor.onDidChangeModel(()=>this._cancelSchedulerAndHide())),this._listenersStore.add(this._editor.onDidChangeModelContent(()=>this._cancelScheduler())),this._listenersStore.add(this._editor.onDidScrollChange(t=>this._onEditorScrollChanged(t)))}_unhookListeners(){this._listenersStore.clear()}_cancelSchedulerAndHide(){this._cancelScheduler(),this.hideContentHover()}_cancelScheduler(){this._mouseMoveEvent=void 0,this._reactToEditorMouseMoveRunner.cancel()}_onEditorScrollChanged(e){this._ignoreMouseEvents||(e.scrollTopChanged||e.scrollLeftChanged)&&this.hideContentHover()}_onEditorMouseDown(e){this._ignoreMouseEvents||(this._isMouseDown=!0,this._shouldKeepHoverWidgetVisible(e))||this.hideContentHover()}_shouldKeepHoverWidgetVisible(e){return this._isMouseOnContentHoverWidget(e)||this._isContentWidgetResizing()||Wse(e)}_isMouseOnContentHoverWidget(e){return this._contentWidget?TA(this._contentWidget.getDomNode(),e.event.posx,e.event.posy):!1}_onEditorMouseUp(){this._ignoreMouseEvents||(this._isMouseDown=!1)}_onEditorMouseLeave(e){this._ignoreMouseEvents||this.shouldKeepOpenOnEditorMouseMoveOrLeave||(this._cancelScheduler(),this._shouldKeepHoverWidgetVisible(e))||this.hideContentHover()}_shouldKeepCurrentHover(e){const t=this._contentWidget;if(!t)return!1;const i=this._hoverSettings.sticky,n=(d,h)=>{const u=this._isMouseOnContentHoverWidget(d);return h&&u},o=d=>{const h=t.isColorPickerVisible,u=this._isMouseOnContentHoverWidget(d),g=h&&u,f=h&&this._isMouseDown;return g||f},r=(d,h)=>{const u=d.event.browserEvent.view;return u?h&&t.containsNode(u.document.activeElement)&&!u.getSelection()?.isCollapsed:!1},a=t.isFocused,l=t.isResizing,c=this._hoverSettings.sticky&&t.isVisibleFromKeyboard;return this.shouldKeepOpenOnEditorMouseMoveOrLeave||a||l||c||n(e,i)||o(e)||r(e,i)}_onEditorMouseMove(e){if(this._ignoreMouseEvents)return;if(this._mouseMoveEvent=e,this._shouldKeepCurrentHover(e)){this._reactToEditorMouseMoveRunner.cancel();return}if(this._shouldRescheduleHoverComputation()){this._reactToEditorMouseMoveRunner.isScheduled()||this._reactToEditorMouseMoveRunner.schedule(this._hoverSettings.hidingDelay);return}this._reactToEditorMouseMove(e)}_shouldRescheduleHoverComputation(){const e=this._hoverSettings.hidingDelay;return(this._contentWidget?.isVisible??!1)&&this._hoverSettings.sticky&&e>0}_reactToEditorMouseMove(e){this._hoverSettings.enabled&&this._getOrCreateContentWidget().showsOrWillShow(e)||this.hideContentHover()}_onKeyDown(e){if(this._ignoreMouseEvents||!this._contentWidget)return;const t=this._isPotentialKeyboardShortcut(e),i=this._isModifierKeyPressed(e);t||i||this._contentWidget.isFocused&&e.keyCode===2||this.hideContentHover()}_isPotentialKeyboardShortcut(e){if(!this._editor.hasModel()||!this._contentWidget)return!1;const t=this._keybindingService.softDispatch(e,this._editor.getDomNode()),i=t.kind===1,n=t.kind===2&&(t.commandId===vne||t.commandId===IA||t.commandId===DA)&&this._contentWidget.isVisible;return i||n}_isModifierKeyPressed(e){return e.keyCode===5||e.keyCode===6||e.keyCode===57||e.keyCode===4}hideContentHover(){xv.dropDownVisible||this._contentWidget?.hide()}_getOrCreateContentWidget(){return this._contentWidget||(this._contentWidget=this._instantiationService.createInstance(q5,this._editor),this._listenersStore.add(this._contentWidget.onContentsChanged(()=>this._onHoverContentsChanged.fire()))),this._contentWidget}showContentHover(e,t,i,n){this._getOrCreateContentWidget().startShowingAtRange(e,t,i,n)}_isContentWidgetResizing(){return this._contentWidget?.widget.isResizing||!1}focusedHoverPartIndex(){return this._getOrCreateContentWidget().focusedHoverPartIndex()}updateHoverVerbosityLevel(e,t,i){this._getOrCreateContentWidget().updateHoverVerbosityLevel(e,t,i)}focus(){this._contentWidget?.focus()}scrollUp(){this._contentWidget?.scrollUp()}scrollDown(){this._contentWidget?.scrollDown()}scrollLeft(){this._contentWidget?.scrollLeft()}scrollRight(){this._contentWidget?.scrollRight()}pageUp(){this._contentWidget?.pageUp()}pageDown(){this._contentWidget?.pageDown()}goToTop(){this._contentWidget?.goToTop()}goToBottom(){this._contentWidget?.goToBottom()}get isColorPickerVisible(){return this._contentWidget?.isColorPickerVisible}get isHoverVisible(){return this._contentWidget?.isVisible}dispose(){super.dispose(),this._unhookListeners(),this._listenersStore.dispose(),this._contentWidget?.dispose()}},j5=Vp,Vp.ID="editor.contrib.contentHover",Vp);ws=j5=VRe([oO(1,Nr),oO(2,_e),oO(3,ft)],ws);const zV=class zV extends F{constructor(e){super(),this._editor=e,this._register(e.onMouseDown(t=>this.onMouseDown(t)))}dispose(){super.dispose()}onMouseDown(e){const t=this._editor.getOption(168);if(t!=="click"&&t!=="clickAndHover"||!Wse(e))return;const i=this._editor.getContribution(ws.ID);if(!i||i.isColorPickerVisible)return;const n=e.target.range;if(!n)return;const o=new k(n.startLineNumber,n.startColumn+1,n.endLineNumber,n.endColumn+1);i.showContentHover(o,1,1,!1)}};zV.ID="editor.contrib.colorContribution";let pD=zV;var zRe=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},$Re=function(s,e){return function(t,i){e(t,i,s)}};class PH{constructor(e,t,i,n){this.owner=e,this.range=t,this.model=i,this.provider=n}static fromBaseColor(e,t){return new PH(e,t.range,t.model,t.provider)}}class URe extends F{constructor(e,t,i,n){super();const o=e.getModel(),r=i.model;this.color=i.model.color,this.colorPicker=this._register(new Kne(t.fragment,r,e.getOption(163),n,"standalone")),this._register(r.onColorFlushed(a=>{this.color=a})),this._register(r.onDidChangeColor(a=>{YS(o,r,a,i.range,i)})),this._register(e.onDidChangeModelContent(a=>{t.hide(),e.focus()})),YS(o,r,this.color,i.range,i)}}let G5=class{constructor(e,t){this._editor=e,this._themeService=t}async createColorHover(e,t,i){if(!this._editor.hasModel()||!Sv.get(this._editor))return null;const o=await mne(i,this._editor.getModel(),Ke.None);let r=null,a=null;for(const u of o){const g=u.colorInfo;k.containsRange(g.range,e.range)&&(r=g,a=u.provider)}const l=r??e,c=a??t,d=!!r;return{colorHover:PH.fromBaseColor(this,await Yne(this._editor.getModel(),l,c)),foundInEditor:d}}async updateEditorModel(e){if(!this._editor.hasModel())return;const t=e.model;let i=new k(e.range.startLineNumber,e.range.startColumn,e.range.endLineNumber,e.range.endColumn);this._color&&(await YS(this._editor.getModel(),t,this._color,i,e),i=Zne(this._editor,i,t))}renderHoverParts(e,t){if(!(t.length===0||!this._editor.hasModel()))return this._setMinimumDimensions(e),this._renderedParts=new URe(this._editor,e,t[0],this._themeService),this._renderedParts}_setMinimumDimensions(e){const t=this._editor.getOption(75)+8;e.setMinimumDimensions(new St(302,t))}get _color(){return this._renderedParts?.color}};G5=zRe([$Re(1,wi)],G5);var qRe=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},jC=function(s,e){return function(t,i){e(t,i,s)}},K5;class jRe{constructor(e,t){this.value=e,this.foundInEditor=t}}const rG=8,GRe=22;var zp;let Y5=(zp=class extends F{constructor(e,t,i,n,o,r,a,l){super(),this._editor=e,this._standaloneColorPickerVisible=t,this._standaloneColorPickerFocused=i,this._keybindingService=o,this._languageFeaturesService=r,this._editorWorkerService=a,this._hoverService=l,this.allowEditorOverflow=!0,this._position=void 0,this._body=document.createElement("div"),this._colorHover=null,this._selectionSetInEditor=!1,this._onResult=this._register(new M),this.onResult=this._onResult.event,this._renderedHoverParts=this._register(new vt),this._renderedStatusBar=this._register(new vt),this._standaloneColorPickerVisible.set(!0),this._standaloneColorPickerParticipant=n.createInstance(G5,this._editor),this._position=this._editor._getViewModel()?.getPrimaryCursorState().modelState.position;const c=this._editor.getSelection(),d=c?{startLineNumber:c.startLineNumber,startColumn:c.startColumn,endLineNumber:c.endLineNumber,endColumn:c.endColumn}:{startLineNumber:0,endLineNumber:0,endColumn:0,startColumn:0},h=this._register(oa(this._body));this._register(h.onDidBlur(u=>{this.hide()})),this._register(h.onDidFocus(u=>{this.focus()})),this._register(this._editor.onDidChangeCursorPosition(()=>{this._selectionSetInEditor?this._selectionSetInEditor=!1:this.hide()})),this._register(this._editor.onMouseMove(u=>{const g=u.target.element?.classList;g&&g.contains("colorpicker-color-decoration")&&this.hide()})),this._register(this.onResult(u=>{this._render(u.value,u.foundInEditor)})),this._start(d),this._body.style.zIndex="50",this._editor.addContentWidget(this)}updateEditor(){this._colorHover&&this._standaloneColorPickerParticipant.updateEditorModel(this._colorHover)}getId(){return K5.ID}getDomNode(){return this._body}getPosition(){if(!this._position)return null;const e=this._editor.getOption(69).above;return{position:this._position,secondaryPosition:this._position,preference:e?[1,2]:[2,1],positionAffinity:2}}hide(){this.dispose(),this._standaloneColorPickerVisible.set(!1),this._standaloneColorPickerFocused.set(!1),this._editor.removeContentWidget(this),this._editor.focus()}focus(){this._standaloneColorPickerFocused.set(!0),this._body.focus()}async _start(e){const t=await this._computeAsync(e);t&&this._onResult.fire(new jRe(t.result,t.foundInEditor))}async _computeAsync(e){if(!this._editor.hasModel())return null;const t={range:e,color:{red:0,green:0,blue:0,alpha:1}},i=await this._standaloneColorPickerParticipant.createColorHover(t,new qS(this._editorWorkerService),this._languageFeaturesService.colorProvider);return i?{result:i.colorHover,foundInEditor:i.foundInEditor}:null}_render(e,t){const i=document.createDocumentFragment();this._renderedStatusBar.value=this._register(new YI(this._keybindingService,this._hoverService));const n={fragment:i,statusBar:this._renderedStatusBar.value,onContentsChanged:()=>{},setMinimumDimensions:()=>{},hide:()=>this.hide(),focus:()=>this.focus()};if(this._colorHover=e,this._renderedHoverParts.value=this._standaloneColorPickerParticipant.renderHoverParts(n,[e]),!this._renderedHoverParts.value){this._renderedStatusBar.clear(),this._renderedHoverParts.clear();return}const o=this._renderedHoverParts.value.colorPicker;this._body.classList.add("standalone-colorpicker-body"),this._body.style.maxHeight=Math.max(this._editor.getLayoutInfo().height/4,250)+"px",this._body.style.maxWidth=Math.max(this._editor.getLayoutInfo().width*.66,500)+"px",this._body.tabIndex=0,this._body.appendChild(i),o.layout();const r=o.body,a=r.saturationBox.domNode.clientWidth,l=r.domNode.clientWidth-a-GRe-rG,c=o.body.enterButton;c?.onClicked(()=>{this.updateEditor(),this.hide()});const d=o.header,h=d.pickedColorNode;h.style.width=a+rG+"px";const u=d.originalColorNode;u.style.width=l+"px",o.header.closeButton?.onClicked(()=>{this.hide()}),t&&(c&&(c.button.textContent="Replace"),this._selectionSetInEditor=!0,this._editor.setSelection(e.range)),this._editor.layoutContentWidget(this)}},K5=zp,zp.ID="editor.contrib.standaloneColorPickerWidget",zp);Y5=K5=qRe([jC(3,_e),jC(4,ft),jC(5,ge),jC(6,ko),jC(7,so)],Y5);var KRe=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},aG=function(s,e){return function(t,i){e(t,i,s)}},Z5,$p;let Qm=($p=class extends F{constructor(e,t,i){super(),this._editor=e,this._instantiationService=i,this._standaloneColorPickerWidget=null,this._standaloneColorPickerVisible=N.standaloneColorPickerVisible.bindTo(t),this._standaloneColorPickerFocused=N.standaloneColorPickerFocused.bindTo(t)}showOrFocus(){this._editor.hasModel()&&(this._standaloneColorPickerVisible.get()?this._standaloneColorPickerFocused.get()||this._standaloneColorPickerWidget?.focus():this._standaloneColorPickerWidget=this._instantiationService.createInstance(Y5,this._editor,this._standaloneColorPickerVisible,this._standaloneColorPickerFocused))}hide(){this._standaloneColorPickerFocused.set(!1),this._standaloneColorPickerVisible.set(!1),this._standaloneColorPickerWidget?.hide(),this._editor.focus()}insertColor(){this._standaloneColorPickerWidget?.updateEditor(),this.hide()}static get(e){return e.getContribution(Z5.ID)}},Z5=$p,$p.ID="editor.contrib.standaloneColorPickerController",$p);Qm=Z5=KRe([aG(1,Ae),aG(2,_e)],Qm);class YRe extends za{constructor(){super({id:"editor.action.showOrFocusStandaloneColorPicker",title:{...V(889,"Show or Focus Standalone Color Picker"),mnemonicTitle:m(888,"&&Show or Focus Standalone Color Picker")},precondition:void 0,menu:[{id:fe.CommandPalette}],metadata:{description:V(890,"Show or focus a standalone color picker which uses the default color provider. It displays hex/rgb/hsl colors.")}})}runEditorCommand(e,t){Qm.get(t)?.showOrFocus()}}class ZRe extends ue{constructor(){super({id:"editor.action.hideColorPicker",label:V(891,"Hide the Color Picker"),precondition:N.standaloneColorPickerVisible.isEqualTo(!0),kbOpts:{primary:9,weight:100},metadata:{description:V(892,"Hide the standalone color picker.")}})}run(e,t){Qm.get(t)?.hide()}}class XRe extends ue{constructor(){super({id:"editor.action.insertColorWithStandaloneColorPicker",label:V(893,"Insert Color with Standalone Color Picker"),precondition:N.standaloneColorPickerFocused.isEqualTo(!0),kbOpts:{primary:3,weight:100},metadata:{description:V(894,"Insert hex/rgb/hsl colors with the focused standalone color picker.")}})}run(e,t){Qm.get(t)?.insertColor()}}ae(ZRe);ae(XRe);It(YRe);lt(pD.ID,pD,2);lt(Qm.ID,Qm,1);lt(Sv.ID,Sv,1);uC(t5);__.register(XI);at.registerCommand("_executeDocumentColorProvider",function(s,...e){const[t]=e;if(!(t instanceof Le))throw Qr();const{model:i,colorProviderRegistry:n,defaultColorDecoratorsEnablement:o}=bne(s,t);return sH(new hDe,n,i,Ke.None,o)});at.registerCommand("_executeColorPresentationProvider",function(s,...e){const[t,i]=e;if(!i)return;const{uri:n,range:o}=i;if(!(n instanceof Le)||!Array.isArray(t)||t.length!==4||!k.isIRange(o))throw Qr();const{model:r,colorProviderRegistry:a,defaultColorDecoratorsEnablement:l}=bne(s,n),[c,d,h,u]=t;return sH(new uDe({range:o,color:{red:c,green:d,blue:h,alpha:u}}),a,r,Ke.None,l)});class Eu{constructor(e,t,i){this.languageConfigurationService=i,this._selection=e,this._insertSpace=t,this._usedEndToken=null}static _haystackHasNeedleAtOffset(e,t,i){if(i<0)return!1;const n=t.length,o=e.length;if(i+n>o)return!1;for(let r=0;r<n;r++){const a=e.charCodeAt(i+r),l=t.charCodeAt(r);if(a!==l&&!(a>=65&&a<=90&&a+32===l)&&!(l>=65&&l<=90&&l+32===a))return!1}return!0}_createOperationsForBlockComment(e,t,i,n,o,r){const a=e.startLineNumber,l=e.startColumn,c=e.endLineNumber,d=e.endColumn,h=o.getLineContent(a),u=o.getLineContent(c);let g=h.lastIndexOf(t,l-1+t.length),f=u.indexOf(i,d-1-i.length);if(g!==-1&&f!==-1)if(a===c)h.substring(g+t.length,f).indexOf(i)>=0&&(g=-1,f=-1);else{const _=h.substring(g+t.length),b=u.substring(0,f);(_.indexOf(i)>=0||b.indexOf(i)>=0)&&(g=-1,f=-1)}let p;g!==-1&&f!==-1?(n&&g+t.length<h.length&&h.charCodeAt(g+t.length)===32&&(t=t+" "),n&&f>0&&u.charCodeAt(f-1)===32&&(i=" "+i,f-=1),p=Eu._createRemoveBlockCommentOperations(new k(a,g+t.length+1,c,f+1),t,i)):(p=Eu._createAddBlockCommentOperations(e,t,i,this._insertSpace),this._usedEndToken=p.length===1?i:null);for(const _ of p)r.addTrackedEditOperation(_.range,_.text)}static _createRemoveBlockCommentOperations(e,t,i){const n=[];return k.isEmpty(e)?n.push(Si.delete(new k(e.startLineNumber,e.startColumn-t.length,e.endLineNumber,e.endColumn+i.length))):(n.push(Si.delete(new k(e.startLineNumber,e.startColumn-t.length,e.startLineNumber,e.startColumn))),n.push(Si.delete(new k(e.endLineNumber,e.endColumn,e.endLineNumber,e.endColumn+i.length)))),n}static _createAddBlockCommentOperations(e,t,i,n){const o=[];return k.isEmpty(e)?o.push(Si.replace(new k(e.startLineNumber,e.startColumn,e.endLineNumber,e.endColumn),t+"  "+i)):(o.push(Si.insert(new A(e.startLineNumber,e.startColumn),t+(n?" ":""))),o.push(Si.insert(new A(e.endLineNumber,e.endColumn),(n?" ":"")+i))),o}getEditOperations(e,t){const i=this._selection.startLineNumber,n=this._selection.startColumn;e.tokenization.tokenizeIfCheap(i);const o=e.getLanguageIdAtPosition(i,n),r=this.languageConfigurationService.getLanguageConfiguration(o).comments;!r||!r.blockCommentStartToken||!r.blockCommentEndToken||this._createOperationsForBlockComment(this._selection,r.blockCommentStartToken,r.blockCommentEndToken,this._insertSpace,e,t)}computeCursorState(e,t){const i=t.getInverseEditOperations();if(i.length===2){const n=i[0],o=i[1];return new he(n.range.endLineNumber,n.range.endColumn,o.range.startLineNumber,o.range.startColumn)}else{const n=i[0].range,o=this._usedEndToken?-this._usedEndToken.length-1:0;return new he(n.endLineNumber,n.endColumn+o,n.endLineNumber,n.endColumn+o)}}}class Cd{constructor(e,t,i,n,o,r,a){this.languageConfigurationService=e,this._selection=t,this._indentSize=i,this._type=n,this._insertSpace=o,this._selectionId=null,this._deltaColumn=0,this._moveEndPositionDown=!1,this._ignoreEmptyLines=r,this._ignoreFirstLine=a||!1}static _gatherPreflightCommentStrings(e,t,i,n){e.tokenization.tokenizeIfCheap(t);const o=e.getLanguageIdAtPosition(t,1),r=n.getLanguageConfiguration(o).comments,a=r?r.lineCommentToken:null;if(!a)return null;const l=[];for(let c=0,d=i-t+1;c<d;c++)l[c]={ignore:!1,commentStr:a,commentStrOffset:0,commentStrLength:a.length};return l}static _analyzeLines(e,t,i,n,o,r,a,l,c){let d=!0;const u=l.getLanguageConfiguration(c).comments?.lineCommentNoIndent??!1;let g;e===0?g=!0:e===1?g=!1:g=!0;for(let f=0,p=n.length;f<p;f++){const _=n[f],b=o+f;if(b===o&&a){_.ignore=!0;continue}const w=i.getLineContent(b),C=As(w);if(C===-1){_.ignore=r,_.commentStrOffset=u?0:w.length;continue}d=!1;const v=u?0:C;if(_.ignore=!1,_.commentStrOffset=v,g&&!Eu._haystackHasNeedleAtOffset(w,_.commentStr,v)&&(e===0?g=!1:e===1||(_.ignore=!0)),g&&t){const y=C+_.commentStrLength;y<w.length&&w.charCodeAt(y)===32&&(_.commentStrLength+=1)}}if(e===0&&d){g=!1;for(let f=0,p=n.length;f<p;f++)n[f].ignore=!1}return{supported:!0,shouldRemoveComments:g,lines:n}}static _gatherPreflightData(e,t,i,n,o,r,a,l){const c=Cd._gatherPreflightCommentStrings(i,n,o,l),d=i.getLanguageIdAtPosition(n,1);return c===null?{supported:!1}:Cd._analyzeLines(e,t,i,c,n,r,a,l,d)}_executeLineComments(e,t,i,n){let o;i.shouldRemoveComments?o=Cd._createRemoveLineCommentsOperations(i.lines,n.startLineNumber):(Cd._normalizeInsertionPoint(e,i.lines,n.startLineNumber,this._indentSize),o=this._createAddLineCommentsOperations(i.lines,n.startLineNumber));const r=new A(n.positionLineNumber,n.positionColumn);for(let a=0,l=o.length;a<l;a++)t.addEditOperation(o[a].range,o[a].text),k.isEmpty(o[a].range)&&k.getStartPosition(o[a].range).equals(r)&&e.getLineContent(r.lineNumber).length+1===r.column&&(this._deltaColumn=(o[a].text||"").length);this._selectionId=t.trackSelection(n)}_attemptRemoveBlockComment(e,t,i,n){let o=t.startLineNumber,r=t.endLineNumber;const a=n.length+Math.max(e.getLineFirstNonWhitespaceColumn(t.startLineNumber),t.startColumn);let l=e.getLineContent(o).lastIndexOf(i,a-1),c=e.getLineContent(r).indexOf(n,t.endColumn-1-i.length);return l!==-1&&c===-1&&(c=e.getLineContent(o).indexOf(n,l+i.length),r=o),l===-1&&c!==-1&&(l=e.getLineContent(r).lastIndexOf(i,c),o=r),t.isEmpty()&&(l===-1||c===-1)&&(l=e.getLineContent(o).indexOf(i),l!==-1&&(c=e.getLineContent(o).indexOf(n,l+i.length))),l!==-1&&e.getLineContent(o).charCodeAt(l+i.length)===32&&(i+=" "),c!==-1&&e.getLineContent(r).charCodeAt(c-1)===32&&(n=" "+n,c-=1),l!==-1&&c!==-1?Eu._createRemoveBlockCommentOperations(new k(o,l+i.length+1,r,c+1),i,n):null}_executeBlockComment(e,t,i){e.tokenization.tokenizeIfCheap(i.startLineNumber);const n=e.getLanguageIdAtPosition(i.startLineNumber,1),o=this.languageConfigurationService.getLanguageConfiguration(n).comments;if(!o||!o.blockCommentStartToken||!o.blockCommentEndToken)return;const r=o.blockCommentStartToken,a=o.blockCommentEndToken;let l=this._attemptRemoveBlockComment(e,i,r,a);if(!l){if(i.isEmpty()){const c=e.getLineContent(i.startLineNumber);let d=As(c);d===-1&&(d=c.length),l=Eu._createAddBlockCommentOperations(new k(i.startLineNumber,d+1,i.startLineNumber,c.length+1),r,a,this._insertSpace)}else l=Eu._createAddBlockCommentOperations(new k(i.startLineNumber,e.getLineFirstNonWhitespaceColumn(i.startLineNumber),i.endLineNumber,e.getLineMaxColumn(i.endLineNumber)),r,a,this._insertSpace);l.length===1&&(this._deltaColumn=r.length+1)}this._selectionId=t.trackSelection(i);for(const c of l)t.addEditOperation(c.range,c.text)}getEditOperations(e,t){let i=this._selection;if(this._moveEndPositionDown=!1,i.startLineNumber===i.endLineNumber&&this._ignoreFirstLine){t.addEditOperation(new k(i.startLineNumber,e.getLineMaxColumn(i.startLineNumber),i.startLineNumber+1,1),i.startLineNumber===e.getLineCount()?"":`
`),this._selectionId=t.trackSelection(i);return}i.startLineNumber<i.endLineNumber&&i.endColumn===1&&(this._moveEndPositionDown=!0,i=i.setEndPosition(i.endLineNumber-1,e.getLineMaxColumn(i.endLineNumber-1)));const n=Cd._gatherPreflightData(this._type,this._insertSpace,e,i.startLineNumber,i.endLineNumber,this._ignoreEmptyLines,this._ignoreFirstLine,this.languageConfigurationService);return n.supported?this._executeLineComments(e,t,n,i):this._executeBlockComment(e,t,i)}computeCursorState(e,t){let i=t.getTrackedSelection(this._selectionId);return this._moveEndPositionDown&&(i=i.setEndPosition(i.endLineNumber+1,1)),new he(i.selectionStartLineNumber,i.selectionStartColumn+this._deltaColumn,i.positionLineNumber,i.positionColumn+this._deltaColumn)}static _createRemoveLineCommentsOperations(e,t){const i=[];for(let n=0,o=e.length;n<o;n++){const r=e[n];r.ignore||i.push(Si.delete(new k(t+n,r.commentStrOffset+1,t+n,r.commentStrOffset+r.commentStrLength+1)))}return i}_createAddLineCommentsOperations(e,t){const i=[],n=this._insertSpace?" ":"";for(let o=0,r=e.length;o<r;o++){const a=e[o];a.ignore||i.push(Si.insert(new A(t+o,a.commentStrOffset+1),a.commentStr+n))}return i}static nextVisibleColumn(e,t,i,n){return i?e+(t-e%t):e+n}static _normalizeInsertionPoint(e,t,i,n){let o=1073741824,r,a;for(let l=0,c=t.length;l<c;l++){if(t[l].ignore)continue;const d=e.getLineContent(i+l);let h=0;for(let u=0,g=t[l].commentStrOffset;h<o&&u<g;u++)h=Cd.nextVisibleColumn(h,n,d.charCodeAt(u)===9,1);h<o&&(o=h)}o=Math.floor(o/n)*n;for(let l=0,c=t.length;l<c;l++){if(t[l].ignore)continue;const d=e.getLineContent(i+l);let h=0;for(r=0,a=t[l].commentStrOffset;h<o&&r<a;r++)h=Cd.nextVisibleColumn(h,n,d.charCodeAt(r)===9,1);h>o?t[l].commentStrOffset=r-1:t[l].commentStrOffset=r}}}class FH extends ue{constructor(e,t){super(t),this._type=e}run(e,t){const i=e.get(ui);if(!t.hasModel())return;const n=t.getModel(),o=[],r=n.getOptions(),a=t.getOption(29),l=t.getSelections().map((d,h)=>({selection:d,index:h,ignoreFirstLine:!1}));l.sort((d,h)=>k.compareRangesUsingStarts(d.selection,h.selection));let c=l[0];for(let d=1;d<l.length;d++){const h=l[d];c.selection.endLineNumber===h.selection.startLineNumber&&(c.index<h.index?h.ignoreFirstLine=!0:(c.ignoreFirstLine=!0,c=h))}for(const d of l)o.push(new Cd(i,d.selection,r.indentSize,this._type,a.insertSpace,a.ignoreEmptyLines,d.ignoreFirstLine));t.pushUndoStop(),t.executeCommands(this.id,o),t.pushUndoStop()}}class QRe extends FH{constructor(){super(0,{id:"editor.action.commentLine",label:V(897,"Toggle Line Comment"),precondition:N.writable,kbOpts:{kbExpr:N.editorTextFocus,primary:2138,weight:100},menuOpts:{menuId:fe.MenubarEditMenu,group:"5_insert",title:m(895,"&&Toggle Line Comment"),order:1},canTriggerInlineEdits:!0})}}class JRe extends FH{constructor(){super(1,{id:"editor.action.addCommentLine",label:V(898,"Add Line Comment"),precondition:N.writable,kbOpts:{kbExpr:N.editorTextFocus,primary:Gi(2089,2081),weight:100},canTriggerInlineEdits:!0})}}class eAe extends FH{constructor(){super(2,{id:"editor.action.removeCommentLine",label:V(899,"Remove Line Comment"),precondition:N.writable,kbOpts:{kbExpr:N.editorTextFocus,primary:Gi(2089,2099),weight:100},canTriggerInlineEdits:!0})}}class tAe extends ue{constructor(){super({id:"editor.action.blockComment",label:V(900,"Toggle Block Comment"),precondition:N.writable,kbOpts:{kbExpr:N.editorTextFocus,primary:1567,linux:{primary:3103},weight:100},menuOpts:{menuId:fe.MenubarEditMenu,group:"5_insert",title:m(896,"Toggle &&Block Comment"),order:2},canTriggerInlineEdits:!0})}run(e,t){const i=e.get(ui);if(!t.hasModel())return;const n=t.getOption(29),o=[],r=t.getSelections();for(const a of r)o.push(new Eu(a,n.insertSpace,i));t.pushUndoStop(),t.executeCommands(this.id,o),t.pushUndoStop()}}ae(QRe);ae(JRe);ae(eAe);ae(tAe);var iAe=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},Gg=function(s,e){return function(t,i){e(t,i,s)}},X5,Up;let t0=(Up=class{static get(e){return e.getContribution(X5.ID)}constructor(e,t,i,n,o,r,a,l){this._contextMenuService=t,this._contextViewService=i,this._contextKeyService=n,this._keybindingService=o,this._menuService=r,this._configurationService=a,this._workspaceContextService=l,this._toDispose=new U,this._contextMenuIsBeingShownCount=0,this._editor=e,this._toDispose.add(this._editor.onContextMenu(c=>this._onContextMenu(c))),this._toDispose.add(this._editor.onMouseWheel(c=>{if(this._contextMenuIsBeingShownCount>0){const d=this._contextViewService.getContextViewElement(),h=c.srcElement;h.shadowRoot&&cg(d)===h.shadowRoot||this._contextViewService.hideContextView()}})),this._toDispose.add(this._editor.onKeyDown(c=>{this._editor.getOption(30)&&c.keyCode===58&&(c.preventDefault(),c.stopPropagation(),this.showContextMenu())}))}_onContextMenu(e){if(!this._editor.hasModel())return;if(!this._editor.getOption(30)){this._editor.focus(),e.target.position&&!this._editor.getSelection().containsPosition(e.target.position)&&this._editor.setPosition(e.target.position);return}if(e.target.type===12||e.target.type===6&&e.target.detail.injectedText)return;if(e.event.preventDefault(),e.event.stopPropagation(),e.target.type===11)return this._showScrollbarContextMenu(e.event);if(e.target.type!==6&&e.target.type!==7&&e.target.type!==1)return;if(this._editor.focus(),e.target.position){let i=!1;for(const n of this._editor.getSelections())if(n.containsPosition(e.target.position)){i=!0;break}i||this._editor.setPosition(e.target.position)}let t=null;e.target.type!==1&&(t=e.event),this.showContextMenu(t)}showContextMenu(e){if(!this._editor.getOption(30)||!this._editor.hasModel())return;const t=this._getMenuActions(this._editor.getModel(),this._editor.contextMenuId);t.length>0&&this._doShowContextMenu(t,e)}_getMenuActions(e,t){const i=[],n=this._menuService.getMenuActions(t,this._contextKeyService,{arg:e.uri});for(const o of n){const[,r]=o;let a=0;for(const l of r)if(l instanceof ap){const c=this._getMenuActions(e,l.item.submenu);c.length>0&&(i.push(new ov(l.id,l.label,c)),a++)}else i.push(l),a++;a&&i.push(new tn)}return i.length&&i.pop(),i}_doShowContextMenu(e,t=null){if(!this._editor.hasModel())return;let i=t;if(!i){this._editor.revealPosition(this._editor.getPosition(),1),this._editor.render();const o=this._editor.getScrolledVisiblePosition(this._editor.getPosition()),r=Li(this._editor.getDomNode()),a=r.left+o.left,l=r.top+o.top+o.height;i={x:a,y:l}}const n=this._editor.getOption(144)&&!na;this._contextMenuIsBeingShownCount++,this._contextMenuService.showContextMenu({domForShadowRoot:n?this._editor.getOverflowWidgetsDomNode()??this._editor.getDomNode():void 0,getAnchor:()=>i,getActions:()=>e,getActionViewItem:o=>{const r=this._keybindingFor(o);if(r)return new vv(o,o,{label:!0,keybinding:r.getLabel(),isMenu:!0});const a=o;return typeof a.getActionViewItem=="function"?a.getActionViewItem():new vv(o,o,{icon:!0,label:!0,isMenu:!0})},getKeyBinding:o=>this._keybindingFor(o),onHide:o=>{this._contextMenuIsBeingShownCount--}})}_showScrollbarContextMenu(e){if(!this._editor.hasModel()||dIe(this._workspaceContextService.getWorkspace()))return;const t=this._editor.getOption(81);let i=0;const n=c=>({id:`menu-action-${++i}`,label:c.label,tooltip:"",class:void 0,enabled:typeof c.enabled>"u"?!0:c.enabled,checked:c.checked,run:c.run}),o=(c,d)=>new ov(`menu-action-${++i}`,c,d,void 0),r=(c,d,h,u,g)=>{if(!d)return n({label:c,enabled:d,run:()=>{}});const f=_=>()=>{this._configurationService.updateValue(h,_)},p=[];for(const _ of g)p.push(n({label:_.label,checked:u===_.value,run:f(_.value)}));return o(c,p)},a=[];a.push(n({label:m(901,"Minimap"),checked:t.enabled,run:()=>{this._configurationService.updateValue("editor.minimap.enabled",!t.enabled)}})),a.push(new tn),a.push(n({label:m(902,"Render Characters"),enabled:t.enabled,checked:t.renderCharacters,run:()=>{this._configurationService.updateValue("editor.minimap.renderCharacters",!t.renderCharacters)}})),a.push(r(m(903,"Vertical size"),t.enabled,"editor.minimap.size",t.size,[{label:m(904,"Proportional"),value:"proportional"},{label:m(905,"Fill"),value:"fill"},{label:m(906,"Fit"),value:"fit"}])),a.push(r(m(907,"Slider"),t.enabled,"editor.minimap.showSlider",t.showSlider,[{label:m(908,"Mouse Over"),value:"mouseover"},{label:m(909,"Always"),value:"always"}]));const l=this._editor.getOption(144)&&!na;this._contextMenuIsBeingShownCount++,this._contextMenuService.showContextMenu({domForShadowRoot:l?this._editor.getDomNode():void 0,getAnchor:()=>e,getActions:()=>a,onHide:c=>{this._contextMenuIsBeingShownCount--,this._editor.focus()}})}_keybindingFor(e){return this._keybindingService.lookupKeybinding(e.id)}dispose(){this._contextMenuIsBeingShownCount>0&&this._contextViewService.hideContextView(),this._toDispose.dispose()}},X5=Up,Up.ID="editor.contrib.contextmenu",Up);t0=X5=iAe([Gg(1,Nr),Gg(2,Rh),Gg(3,Ae),Gg(4,ft),Gg(5,da),Gg(6,He),Gg(7,Sh)],t0);class nAe extends ue{constructor(){super({id:"editor.action.showContextMenu",label:V(910,"Show Editor Context Menu"),precondition:void 0,kbOpts:{kbExpr:N.textInputFocus,primary:1092,weight:100}})}run(e,t){t0.get(t)?.showContextMenu()}}lt(t0.ID,t0,2);ae(nAe);class rO{constructor(e){this.selections=e}equals(e){const t=this.selections.length,i=e.selections.length;if(t!==i)return!1;for(let n=0;n<t;n++)if(!this.selections[n].equalsSelection(e.selections[n]))return!1;return!0}}class aO{constructor(e,t,i){this.cursorState=e,this.scrollTop=t,this.scrollLeft=i}}const LN=class LN extends F{static get(e){return e.getContribution(LN.ID)}constructor(e){super(),this._editor=e,this._isCursorUndoRedo=!1,this._undoStack=[],this._redoStack=[],this._register(e.onDidChangeModel(t=>{this._undoStack=[],this._redoStack=[]})),this._register(e.onDidChangeModelContent(t=>{this._undoStack=[],this._redoStack=[]})),this._register(e.onDidChangeCursorSelection(t=>{if(this._isCursorUndoRedo||!t.oldSelections||t.oldModelVersionId!==t.modelVersionId)return;const i=new rO(t.oldSelections);this._undoStack.length>0&&this._undoStack[this._undoStack.length-1].cursorState.equals(i)||(this._undoStack.push(new aO(i,e.getScrollTop(),e.getScrollLeft())),this._redoStack=[],this._undoStack.length>50&&this._undoStack.shift())}))}cursorUndo(){!this._editor.hasModel()||this._undoStack.length===0||(this._redoStack.push(new aO(new rO(this._editor.getSelections()),this._editor.getScrollTop(),this._editor.getScrollLeft())),this._applyState(this._undoStack.pop()))}cursorRedo(){!this._editor.hasModel()||this._redoStack.length===0||(this._undoStack.push(new aO(new rO(this._editor.getSelections()),this._editor.getScrollTop(),this._editor.getScrollLeft())),this._applyState(this._redoStack.pop()))}_applyState(e){this._isCursorUndoRedo=!0,this._editor.setSelections(e.cursorState.selections),this._editor.setScrollPosition({scrollTop:e.scrollTop,scrollLeft:e.scrollLeft}),this._isCursorUndoRedo=!1}};LN.ID="editor.contrib.cursorUndoRedoController";let Iv=LN;class sAe extends ue{constructor(){super({id:"cursorUndo",label:V(911,"Cursor Undo"),precondition:void 0,kbOpts:{kbExpr:N.textInputFocus,primary:2099,weight:100}})}run(e,t,i){Iv.get(t)?.cursorUndo()}}class oAe extends ue{constructor(){super({id:"cursorRedo",label:V(912,"Cursor Redo"),precondition:void 0})}run(e,t,i){Iv.get(t)?.cursorRedo()}}lt(Iv.ID,Iv,0);ae(sAe);ae(oAe);class rAe{constructor(e,t,i){this.selection=e,this.targetPosition=t,this.copy=i,this.targetSelection=null}getEditOperations(e,t){const i=e.getValueInRange(this.selection);if(this.copy||t.addEditOperation(this.selection,null),t.addEditOperation(new k(this.targetPosition.lineNumber,this.targetPosition.column,this.targetPosition.lineNumber,this.targetPosition.column),i),this.selection.containsPosition(this.targetPosition)&&!(this.copy&&(this.selection.getEndPosition().equals(this.targetPosition)||this.selection.getStartPosition().equals(this.targetPosition)))){this.targetSelection=this.selection;return}if(this.copy){this.targetSelection=new he(this.targetPosition.lineNumber,this.targetPosition.column,this.selection.endLineNumber-this.selection.startLineNumber+this.targetPosition.lineNumber,this.selection.startLineNumber===this.selection.endLineNumber?this.targetPosition.column+this.selection.endColumn-this.selection.startColumn:this.selection.endColumn);return}if(this.targetPosition.lineNumber>this.selection.endLineNumber){this.targetSelection=new he(this.targetPosition.lineNumber-this.selection.endLineNumber+this.selection.startLineNumber,this.targetPosition.column,this.targetPosition.lineNumber,this.selection.startLineNumber===this.selection.endLineNumber?this.targetPosition.column+this.selection.endColumn-this.selection.startColumn:this.selection.endColumn);return}if(this.targetPosition.lineNumber<this.selection.endLineNumber){this.targetSelection=new he(this.targetPosition.lineNumber,this.targetPosition.column,this.targetPosition.lineNumber+this.selection.endLineNumber-this.selection.startLineNumber,this.selection.startLineNumber===this.selection.endLineNumber?this.targetPosition.column+this.selection.endColumn-this.selection.startColumn:this.selection.endColumn);return}this.selection.endColumn<=this.targetPosition.column?this.targetSelection=new he(this.targetPosition.lineNumber-this.selection.endLineNumber+this.selection.startLineNumber,this.selection.startLineNumber===this.selection.endLineNumber?this.targetPosition.column-this.selection.endColumn+this.selection.startColumn:this.targetPosition.column-this.selection.endColumn+this.selection.startColumn,this.targetPosition.lineNumber,this.selection.startLineNumber===this.selection.endLineNumber?this.targetPosition.column:this.selection.endColumn):this.targetSelection=new he(this.targetPosition.lineNumber-this.selection.endLineNumber+this.selection.startLineNumber,this.targetPosition.column,this.targetPosition.lineNumber,this.targetPosition.column+this.selection.endColumn-this.selection.startColumn)}computeCursorState(e,t){return this.targetSelection}}function Q_(s){return Ye?s.altKey:s.ctrlKey}const eu=class eu extends F{constructor(e){super(),this._editor=e,this._dndDecorationIds=this._editor.createDecorationsCollection(),this._register(this._editor.onMouseDown(t=>this._onEditorMouseDown(t))),this._register(this._editor.onMouseUp(t=>this._onEditorMouseUp(t))),this._register(this._editor.onMouseDrag(t=>this._onEditorMouseDrag(t))),this._register(this._editor.onMouseDrop(t=>this._onEditorMouseDrop(t))),this._register(this._editor.onMouseDropCanceled(()=>this._onEditorMouseDropCanceled())),this._register(this._editor.onKeyDown(t=>this.onEditorKeyDown(t))),this._register(this._editor.onKeyUp(t=>this.onEditorKeyUp(t))),this._register(this._editor.onDidBlurEditorWidget(()=>this.onEditorBlur())),this._register(this._editor.onDidBlurEditorText(()=>this.onEditorBlur())),this._mouseDown=!1,this._modifierPressed=!1,this._dragSelection=null}onEditorBlur(){this._removeDecoration(),this._dragSelection=null,this._mouseDown=!1,this._modifierPressed=!1}onEditorKeyDown(e){!this._editor.getOption(42)||this._editor.getOption(28)||(Q_(e)&&(this._modifierPressed=!0),this._mouseDown&&Q_(e)&&this._editor.updateOptions({mouseStyle:"copy"}))}onEditorKeyUp(e){!this._editor.getOption(42)||this._editor.getOption(28)||(Q_(e)&&(this._modifierPressed=!1),this._mouseDown&&e.keyCode===eu.TRIGGER_KEY_VALUE&&this._editor.updateOptions({mouseStyle:"default"}))}_onEditorMouseDown(e){this._mouseDown=!0}_onEditorMouseUp(e){this._mouseDown=!1,this._editor.updateOptions({mouseStyle:"text"})}_onEditorMouseDrag(e){const t=e.target;if(this._dragSelection===null){const n=(this._editor.getSelections()||[]).filter(o=>t.position&&o.containsPosition(t.position));if(n.length===1)this._dragSelection=n[0];else return}Q_(e.event)?this._editor.updateOptions({mouseStyle:"copy"}):this._editor.updateOptions({mouseStyle:"default"}),t.position&&(this._dragSelection.containsPosition(t.position)?this._removeDecoration():this.showAt(t.position))}_onEditorMouseDropCanceled(){this._editor.updateOptions({mouseStyle:"text"}),this._removeDecoration(),this._dragSelection=null,this._mouseDown=!1}_onEditorMouseDrop(e){if(e.target&&(this._hitContent(e.target)||this._hitMargin(e.target))&&e.target.position){const t=new A(e.target.position.lineNumber,e.target.position.column);if(this._dragSelection===null){let i=null;if(e.event.shiftKey){const n=this._editor.getSelection();if(n){const{selectionStartLineNumber:o,selectionStartColumn:r}=n;i=[new he(o,r,t.lineNumber,t.column)]}}else i=(this._editor.getSelections()||[]).map(n=>n.containsPosition(t)?new he(t.lineNumber,t.column,t.lineNumber,t.column):n);this._editor.setSelections(i||[],"mouse",3)}else(!this._dragSelection.containsPosition(t)||(Q_(e.event)||this._modifierPressed)&&(this._dragSelection.getEndPosition().equals(t)||this._dragSelection.getStartPosition().equals(t)))&&(this._editor.pushUndoStop(),this._editor.executeCommand(eu.ID,new rAe(this._dragSelection,t,Q_(e.event)||this._modifierPressed)),this._editor.pushUndoStop())}this._editor.updateOptions({mouseStyle:"text"}),this._removeDecoration(),this._dragSelection=null,this._mouseDown=!1}showAt(e){this._dndDecorationIds.set([{range:new k(e.lineNumber,e.column,e.lineNumber,e.column),options:eu._DECORATION_OPTIONS}]),this._editor.revealPosition(e,1)}_removeDecoration(){this._dndDecorationIds.clear()}_hitContent(e){return e.type===6||e.type===7}_hitMargin(e){return e.type===2||e.type===3||e.type===4}dispose(){this._removeDecoration(),this._dragSelection=null,this._mouseDown=!1,this._modifierPressed=!1,super.dispose()}};eu.ID="editor.contrib.dragAndDrop",eu.TRIGGER_KEY_VALUE=Ye?6:5,eu._DECORATION_OPTIONS=Fe.register({description:"dnd-target",className:"dnd-target"});let mD=eu;lt(mD.ID,mD,2);const aAe="editor.action.pasteAs";lt(xh.ID,xh,0);uC(H4);le(new class extends hn{constructor(){super({id:tne,precondition:J7,kbOpts:{weight:100,primary:2137}})}runEditorCommand(s,e){return xh.get(e)?.changePasteType()}});le(new class extends hn{constructor(){super({id:"editor.hidePasteWidget",precondition:J7,kbOpts:{weight:100,primary:9}})}runEditorCommand(s,e){xh.get(e)?.clearWidgets()}});var qp;ae((qp=class extends ue{constructor(){super({id:aAe,label:V(915,"Paste As..."),precondition:N.writable,metadata:{description:"Paste as",args:[{name:"args",schema:qp.argsSchema}]},canTriggerInlineEdits:!0})}run(e,t,i){let n;return i&&("kind"in i?n={only:new _i(i.kind)}:"preferences"in i&&(n={preferences:i.preferences.map(o=>new _i(o))})),xh.get(t)?.pasteAs(n)}},qp.argsSchema={oneOf:[{type:"object",required:["kind"],properties:{kind:{type:"string",description:m(913,`The kind of the paste edit to try pasting with.
If there are multiple edits for this kind, the editor will show a picker. If there are no edits of this kind, the editor will show an error message.`)}}},{type:"object",required:["preferences"],properties:{preferences:{type:"array",description:m(914,`List of preferred paste edit kind to try applying.
The first edit matching the preferences will be applied.`),items:{type:"string"}}}}]},qp));ae(class extends ue{constructor(){super({id:"editor.action.pasteAsText",label:V(916,"Paste as Text"),precondition:N.writable,canTriggerInlineEdits:!0})}run(s,e){return xh.get(e)?.pasteAs({providerId:Wm.id})}});class lAe{constructor(){this._dragOperations=new Map}removeDragOperationTransfer(e){if(e&&this._dragOperations.has(e)){const t=this._dragOperations.get(e);return this._dragOperations.delete(e),t}}}class lG{constructor(e){this.identifier=e}}const Hse=qe("treeViewsDndService");tt(Hse,lAe,1);var cAe=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},WL=function(s,e){return function(t,i){e(t,i,s)}},md;const dAe="editor.dropIntoEditor.preferences",Vse="editor.changeDropType",BH=new ce("dropWidgetVisible",!1,m(934,"Whether the drop widget is showing"));var jp;let Dv=(jp=class extends F{static get(e){return e.getContribution(md.ID)}constructor(e,t,i,n,o){super(),this._configService=i,this._languageFeaturesService=n,this._treeViewsDragAndDropService=o,this.treeItemsTransfer=A4.getInstance(),this._dropProgressManager=this._register(t.createInstance(WI,"dropIntoEditor",e)),this._postDropWidgetManager=this._register(t.createInstance(UI,"dropIntoEditor",e,BH,{id:Vse,label:m(935,"Show drop options...")},()=>md._configureDefaultAction?[md._configureDefaultAction]:[])),this._register(e.onDropIntoEditor(r=>this.onDropIntoEditor(e,r.position,r.event)))}clearWidgets(){this._postDropWidgetManager.clear()}changeDropType(){this._postDropWidgetManager.tryShowSelector()}async onDropIntoEditor(e,t,i){if(!i.dataTransfer||!e.hasModel())return;md._currentDropOperation?.cancel(),e.focus(),e.setPosition(t);const n=an(async o=>{const r=new U,a=r.add(new yh(e,1,void 0,o));try{const l=await this.extractDataTransferData(i);if(l.size===0||a.token.isCancellationRequested)return;const c=e.getModel();if(!c)return;const d=this._languageFeaturesService.documentDropEditProvider.ordered(c).filter(u=>u.dropMimeTypes?u.dropMimeTypes.some(g=>l.matches(g)):!0),h=r.add(await this.getDropEdits(d,c,t,l,a.token));if(a.token.isCancellationRequested)return;if(h.edits.length){const u=this.getInitialActiveEditIndex(c,h.edits),g=e.getOption(43).showDropSelector==="afterDrop";await this._postDropWidgetManager.applyEditAndShowIfNeeded([k.fromPositions(t)],{activeEditIndex:u,allEdits:h.edits},g,async f=>f,o)}}finally{r.dispose(),md._currentDropOperation===n&&(md._currentDropOperation=void 0)}});this._dropProgressManager.showWhile(t,m(936,"Running drop handlers. Click to cancel"),n,{cancel:()=>n.cancel()}),md._currentDropOperation=n}async getDropEdits(e,t,i,n,o){const r=new U,a=await tv(Promise.all(e.map(async c=>{try{const d=await c.provideDocumentDropEdits(t,i,n,o);return d&&r.add(d),d?.edits.map(h=>({...h,providerId:c.id}))}catch(d){Dr(d)||console.error(d),console.error(d)}})),o),l=Al(a??[]).flat();return{edits:Zie(l),dispose:()=>r.dispose()}}getInitialActiveEditIndex(e,t){const i=this._configService.getValue(dAe,{resource:e.uri});for(const n of Array.isArray(i)?i:[]){const o=new _i(n),r=t.findIndex(a=>a.kind&&o.contains(a.kind));if(r>=0)return r}return 0}async extractDataTransferData(e){if(!e.dataTransfer)return new Fie;const t=Vie(e.dataTransfer);if(this.treeItemsTransfer.hasData(lG.prototype)){const i=this.treeItemsTransfer.getData(lG.prototype);if(Array.isArray(i))for(const n of i){const o=await this._treeViewsDragAndDropService.removeDragOperationTransfer(n.identifier);if(o)for(const[r,a]of o)t.replace(r,a)}}return t}},md=jp,jp.ID="editor.contrib.dropIntoEditorController",jp);Dv=md=cAe([WL(1,_e),WL(2,He),WL(3,ge),WL(4,Hse)],Dv);lt(Dv.ID,Dv,2);uC(W4);le(new class extends hn{constructor(){super({id:Vse,precondition:BH,kbOpts:{weight:100,primary:2137}})}runEditorCommand(s,e,t){Dv.get(e)?.changeDropType()}});le(new class extends hn{constructor(){super({id:"editor.hideDropWidget",precondition:BH,kbOpts:{weight:100,primary:9}})}runEditorCommand(s,e,t){Dv.get(e)?.clearWidgets()}});const kN=class kN extends Xo{constructor(e,t,i){super(),this._hideSoon=this._register(new Rt(()=>this._hide(),2e3)),this._isVisible=!1,this._editor=e,this._state=t,this._keybindingService=i,this._domNode=document.createElement("div"),this._domNode.className="findOptionsWidget",this._domNode.style.display="none",this._domNode.style.top="10px",this._domNode.style.zIndex="12",this._domNode.setAttribute("role","presentation"),this._domNode.setAttribute("aria-hidden","true");const n={inputActiveOptionBorder:te(X0),inputActiveOptionForeground:te(Q0),inputActiveOptionBackground:te(iC)},o={groupId:"find-options-widget"};this.caseSensitive=this._register(new Ane({appendTitle:this._keybindingLabelFor(Ut.ToggleCaseSensitiveCommand),isChecked:this._state.matchCase,hoverLifecycleOptions:o,...n})),this._domNode.appendChild(this.caseSensitive.domNode),this._register(this.caseSensitive.onChange(()=>{this._state.change({matchCase:this.caseSensitive.checked},!1)})),this.wholeWords=this._register(new Mne({appendTitle:this._keybindingLabelFor(Ut.ToggleWholeWordCommand),isChecked:this._state.wholeWord,hoverLifecycleOptions:o,...n})),this._domNode.appendChild(this.wholeWords.domNode),this._register(this.wholeWords.onChange(()=>{this._state.change({wholeWord:this.wholeWords.checked},!1)})),this.regex=this._register(new One({appendTitle:this._keybindingLabelFor(Ut.ToggleRegexCommand),isChecked:this._state.isRegex,hoverLifecycleOptions:o,...n})),this._domNode.appendChild(this.regex.domNode),this._register(this.regex.onChange(()=>{this._state.change({isRegex:this.regex.checked},!1)})),this._editor.addOverlayWidget(this),this._register(this._state.onFindReplaceStateChange(r=>{let a=!1;r.isRegex&&(this.regex.checked=this._state.isRegex,a=!0),r.wholeWord&&(this.wholeWords.checked=this._state.wholeWord,a=!0),r.matchCase&&(this.caseSensitive.checked=this._state.matchCase,a=!0),!this._state.isRevealed&&a&&this._revealTemporarily()})),this._register(W(this._domNode,se.MOUSE_LEAVE,r=>this._onMouseLeave())),this._register(W(this._domNode,"mouseover",r=>this._onMouseOver()))}_keybindingLabelFor(e){const t=this._keybindingService.lookupKeybinding(e);return t?` (${t.getLabel()})`:""}dispose(){this._editor.removeOverlayWidget(this),super.dispose()}getId(){return kN.ID}getDomNode(){return this._domNode}getPosition(){return{preference:0}}highlightFindOptions(){this._revealTemporarily()}_revealTemporarily(){this._show(),this._hideSoon.schedule()}_onMouseLeave(){this._hideSoon.schedule()}_onMouseOver(){this._hideSoon.cancel()}_show(){this._isVisible||(this._isVisible=!0,this._domNode.style.display="block")}_hide(){this._isVisible&&(this._isVisible=!1,this._domNode.style.display="none")}};kN.ID="editor.contrib.findOptionsWidget";let Q5=kN;function HL(s,e){return s===1?!0:s===2?!1:e}class hAe extends F{get searchString(){return this._searchString}get replaceString(){return this._replaceString}get isRevealed(){return this._isRevealed}get isReplaceRevealed(){return this._isReplaceRevealed}get isRegex(){return HL(this._isRegexOverride,this._isRegex)}get wholeWord(){return HL(this._wholeWordOverride,this._wholeWord)}get matchCase(){return HL(this._matchCaseOverride,this._matchCase)}get preserveCase(){return HL(this._preserveCaseOverride,this._preserveCase)}get actualIsRegex(){return this._isRegex}get actualWholeWord(){return this._wholeWord}get actualMatchCase(){return this._matchCase}get actualPreserveCase(){return this._preserveCase}get searchScope(){return this._searchScope}get matchesPosition(){return this._matchesPosition}get matchesCount(){return this._matchesCount}get currentMatch(){return this._currentMatch}constructor(){super(),this._onFindReplaceStateChange=this._register(new M),this.onFindReplaceStateChange=this._onFindReplaceStateChange.event,this._searchString="",this._replaceString="",this._isRevealed=!1,this._isReplaceRevealed=!1,this._isRegex=!1,this._isRegexOverride=0,this._wholeWord=!1,this._wholeWordOverride=0,this._matchCase=!1,this._matchCaseOverride=0,this._preserveCase=!1,this._preserveCaseOverride=0,this._searchScope=null,this._matchesPosition=0,this._matchesCount=0,this._currentMatch=null,this._loop=!0,this._isSearching=!1,this._filters=null}changeMatchInfo(e,t,i){const n={moveCursor:!1,updateHistory:!1,searchString:!1,replaceString:!1,isRevealed:!1,isReplaceRevealed:!1,isRegex:!1,wholeWord:!1,matchCase:!1,preserveCase:!1,searchScope:!1,matchesPosition:!1,matchesCount:!1,currentMatch:!1,loop:!1,isSearching:!1,filters:!1};let o=!1;t===0&&(e=0),e>t&&(e=t),this._matchesPosition!==e&&(this._matchesPosition=e,n.matchesPosition=!0,o=!0),this._matchesCount!==t&&(this._matchesCount=t,n.matchesCount=!0,o=!0),typeof i<"u"&&(k.equalsRange(this._currentMatch,i)||(this._currentMatch=i,n.currentMatch=!0,o=!0)),o&&this._onFindReplaceStateChange.fire(n)}change(e,t,i=!0){const n={moveCursor:t,updateHistory:i,searchString:!1,replaceString:!1,isRevealed:!1,isReplaceRevealed:!1,isRegex:!1,wholeWord:!1,matchCase:!1,preserveCase:!1,searchScope:!1,matchesPosition:!1,matchesCount:!1,currentMatch:!1,loop:!1,isSearching:!1,filters:!1};let o=!1;const r=this.isRegex,a=this.wholeWord,l=this.matchCase,c=this.preserveCase;typeof e.searchString<"u"&&this._searchString!==e.searchString&&(this._searchString=e.searchString,n.searchString=!0,o=!0),typeof e.replaceString<"u"&&this._replaceString!==e.replaceString&&(this._replaceString=e.replaceString,n.replaceString=!0,o=!0),typeof e.isRevealed<"u"&&this._isRevealed!==e.isRevealed&&(this._isRevealed=e.isRevealed,n.isRevealed=!0,o=!0),typeof e.isReplaceRevealed<"u"&&this._isReplaceRevealed!==e.isReplaceRevealed&&(this._isReplaceRevealed=e.isReplaceRevealed,n.isReplaceRevealed=!0,o=!0),typeof e.isRegex<"u"&&(this._isRegex=e.isRegex),typeof e.wholeWord<"u"&&(this._wholeWord=e.wholeWord),typeof e.matchCase<"u"&&(this._matchCase=e.matchCase),typeof e.preserveCase<"u"&&(this._preserveCase=e.preserveCase),typeof e.searchScope<"u"&&(e.searchScope?.every(d=>this._searchScope?.some(h=>!k.equalsRange(h,d)))||(this._searchScope=e.searchScope,n.searchScope=!0,o=!0)),typeof e.loop<"u"&&this._loop!==e.loop&&(this._loop=e.loop,n.loop=!0,o=!0),typeof e.isSearching<"u"&&this._isSearching!==e.isSearching&&(this._isSearching=e.isSearching,n.isSearching=!0,o=!0),typeof e.filters<"u"&&(this._filters?this._filters.update(e.filters):this._filters=e.filters,n.filters=!0,o=!0),this._isRegexOverride=typeof e.isRegexOverride<"u"?e.isRegexOverride:0,this._wholeWordOverride=typeof e.wholeWordOverride<"u"?e.wholeWordOverride:0,this._matchCaseOverride=typeof e.matchCaseOverride<"u"?e.matchCaseOverride:0,this._preserveCaseOverride=typeof e.preserveCaseOverride<"u"?e.preserveCaseOverride:0,r!==this.isRegex&&(o=!0,n.isRegex=!0),a!==this.wholeWord&&(o=!0,n.wholeWord=!0),l!==this.matchCase&&(o=!0,n.matchCase=!0),c!==this.preserveCase&&(o=!0,n.preserveCase=!0),o&&this._onFindReplaceStateChange.fire(n)}canNavigateBack(){return this.canNavigateInLoop()||this.matchesPosition!==1}canNavigateForward(){return this.canNavigateInLoop()||this.matchesPosition<this.matchesCount}canNavigateInLoop(){return this._loop||this.matchesCount>=ku}}var uAe=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},gAe=function(s,e){return function(t,i){e(t,i,s)}},jh,Xu;let J5=(Xu=class{static getOrCreate(e){return jh._instance||(jh._instance=new jh(e)),jh._instance}constructor(e){this.storageService=e,this.inMemoryValues=new Set,this._onDidChangeEmitter=new M,this.onDidChange=this._onDidChangeEmitter.event,this.load()}delete(e){const t=this.inMemoryValues.delete(e);return this.save(),t}add(e){return this.inMemoryValues.add(e),this.save(),this}has(e){return this.inMemoryValues.has(e)}forEach(e,t){return this.load(),this.inMemoryValues.forEach(e)}replace(e){this.inMemoryValues=new Set(e),this.save()}load(){let e;const t=this.storageService.get(jh.FIND_HISTORY_KEY,1);if(t)try{e=JSON.parse(t)}catch{}this.inMemoryValues=new Set(e||[])}save(){const e=[];return this.inMemoryValues.forEach(t=>e.push(t)),new Promise(t=>{this.storageService.store(jh.FIND_HISTORY_KEY,JSON.stringify(e),1,0),this._onDidChangeEmitter.fire(e),t()})}},jh=Xu,Xu.FIND_HISTORY_KEY="workbench.find.history",Xu._instance=null,Xu);J5=jh=uAe([gAe(0,Bs)],J5);var fAe=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},pAe=function(s,e){return function(t,i){e(t,i,s)}},Gh,Qu;let e3=(Qu=class{static getOrCreate(e){return Gh._instance||(Gh._instance=new Gh(e)),Gh._instance}constructor(e){this.storageService=e,this.inMemoryValues=new Set,this._onDidChangeEmitter=new M,this.onDidChange=this._onDidChangeEmitter.event,this.load()}delete(e){const t=this.inMemoryValues.delete(e);return this.save(),t}add(e){return this.inMemoryValues.add(e),this.save(),this}has(e){return this.inMemoryValues.has(e)}forEach(e,t){return this.load(),this.inMemoryValues.forEach(e)}replace(e){this.inMemoryValues=new Set(e),this.save()}load(){let e;const t=this.storageService.get(Gh.FIND_HISTORY_KEY,1);if(t)try{e=JSON.parse(t)}catch{}this.inMemoryValues=new Set(e||[])}save(){const e=[];return this.inMemoryValues.forEach(t=>e.push(t)),new Promise(t=>{this.storageService.store(Gh.FIND_HISTORY_KEY,JSON.stringify(e),1,0),this._onDidChangeEmitter.fire(e),t()})}},Gh=Qu,Qu.FIND_HISTORY_KEY="workbench.replace.history",Qu._instance=null,Qu);e3=Gh=fAe([pAe(0,Bs)],e3);var zse=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},La=function(s,e){return function(t,i){e(t,i,s)}},t3;const mAe=524288;function i3(s,e="single",t=!1){if(!s.hasModel())return null;const i=s.getSelection();if(e==="single"&&i.startLineNumber===i.endLineNumber||e==="multiple"){if(i.isEmpty()){const n=s.getConfiguredWordAtPosition(i.getStartPosition());if(n&&t===!1)return n.word}else if(s.getModel().getValueLengthInRange(i)<mAe)return s.getModel().getValueInRange(i)}return null}var Gp;let xo=(Gp=class extends F{get editor(){return this._editor}static get(e){return e.getContribution(t3.ID)}constructor(e,t,i,n,o,r){super(),this._editor=e,this._findWidgetVisible=Oh.bindTo(t),this._contextKeyService=t,this._storageService=i,this._clipboardService=n,this._notificationService=o,this._hoverService=r,this._updateHistoryDelayer=new sa(500),this._state=this._register(new hAe),this.loadQueryState(),this._register(this._state.onFindReplaceStateChange(a=>this._onStateChanged(a))),this._model=null,this._register(this._editor.onDidChangeModel(()=>{const a=this._editor.getModel()&&this._state.isRevealed;this.disposeModel(),this._state.change({searchScope:null,matchCase:this._storageService.getBoolean("editor.matchCase",1,!1),wholeWord:this._storageService.getBoolean("editor.wholeWord",1,!1),isRegex:this._storageService.getBoolean("editor.isRegex",1,!1),preserveCase:this._storageService.getBoolean("editor.preserveCase",1,!1)},!1),a&&this._start({forceRevealReplace:!1,seedSearchStringFromSelection:"none",seedSearchStringFromNonEmptySelection:!1,seedSearchStringFromGlobalClipboard:!1,shouldFocus:0,shouldAnimate:!1,updateSearchScope:!1,loop:this._editor.getOption(50).loop})}))}dispose(){this.disposeModel(),super.dispose()}disposeModel(){this._model&&(this._model.dispose(),this._model=null)}_onStateChanged(e){this.saveQueryState(e),e.isRevealed&&(this._state.isRevealed?this._findWidgetVisible.set(!0):(this._findWidgetVisible.reset(),this.disposeModel())),e.searchString&&this.setGlobalBufferTerm(this._state.searchString)}saveQueryState(e){e.isRegex&&this._storageService.store("editor.isRegex",this._state.actualIsRegex,1,1),e.wholeWord&&this._storageService.store("editor.wholeWord",this._state.actualWholeWord,1,1),e.matchCase&&this._storageService.store("editor.matchCase",this._state.actualMatchCase,1,1),e.preserveCase&&this._storageService.store("editor.preserveCase",this._state.actualPreserveCase,1,1)}loadQueryState(){this._state.change({matchCase:this._storageService.getBoolean("editor.matchCase",1,this._state.matchCase),wholeWord:this._storageService.getBoolean("editor.wholeWord",1,this._state.wholeWord),isRegex:this._storageService.getBoolean("editor.isRegex",1,this._state.isRegex),preserveCase:this._storageService.getBoolean("editor.preserveCase",1,this._state.preserveCase)},!1)}isFindInputFocused(){return!!RA.getValue(this._contextKeyService)}getState(){return this._state}closeFindWidget(){this._state.change({isRevealed:!1,searchScope:null},!1),this._editor.focus()}toggleCaseSensitive(){this._state.change({matchCase:!this._state.matchCase},!1),this._state.isRevealed||this.highlightFindOptions()}toggleWholeWords(){this._state.change({wholeWord:!this._state.wholeWord},!1),this._state.isRevealed||this.highlightFindOptions()}toggleRegex(){this._state.change({isRegex:!this._state.isRegex},!1),this._state.isRevealed||this.highlightFindOptions()}togglePreserveCase(){this._state.change({preserveCase:!this._state.preserveCase},!1),this._state.isRevealed||this.highlightFindOptions()}toggleSearchScope(){if(this._state.searchScope)this._state.change({searchScope:null},!0);else if(this._editor.hasModel()){let e=this._editor.getSelections();e=e.map(t=>(t.endColumn===1&&t.endLineNumber>t.startLineNumber&&(t=t.setEndPosition(t.endLineNumber-1,this._editor.getModel().getLineMaxColumn(t.endLineNumber-1))),t.isEmpty()?null:t)).filter(t=>!!t),e.length&&this._state.change({searchScope:e},!0)}}setSearchString(e){this._state.isRegex&&(e=Uo(e)),this._state.change({searchString:e},!1)}highlightFindOptions(e=!1){}async _start(e,t){if(this.disposeModel(),!this._editor.hasModel())return;const i={...t,isRevealed:!0};if(e.seedSearchStringFromSelection==="single"){const n=i3(this._editor,e.seedSearchStringFromSelection,e.seedSearchStringFromNonEmptySelection);n&&(this._state.isRegex?i.searchString=Uo(n):i.searchString=n)}else if(e.seedSearchStringFromSelection==="multiple"&&!e.updateSearchScope){const n=i3(this._editor,e.seedSearchStringFromSelection);n&&(i.searchString=n)}if(!i.searchString&&e.seedSearchStringFromGlobalClipboard){const n=await this.getGlobalBufferTerm();if(!this._editor.hasModel())return;n&&(i.searchString=n)}if(e.forceRevealReplace||i.isReplaceRevealed?i.isReplaceRevealed=!0:this._findWidgetVisible.get()||(i.isReplaceRevealed=!1),e.updateSearchScope){const n=this._editor.getSelections();n.some(o=>!o.isEmpty())&&(i.searchScope=n)}i.loop=e.loop,this._state.change(i,!1),this._model||(this._model=new X1(this._editor,this._state))}start(e,t){return this._start(e,t)}moveToNextMatch(){return this._model?(this._model.moveToNextMatch(),!0):!1}moveToPrevMatch(){return this._model?(this._model.moveToPrevMatch(),!0):!1}goToMatch(e){return this._model?(this._model.moveToMatch(e),!0):!1}replace(){return this._model?(this._model.replace(),!0):!1}replaceAll(){return this._model?this._editor.getModel()?.isTooLargeForHeapOperation()?(this._notificationService.warn(m(940,"The file is too large to perform a replace all operation.")),!1):(this._model.replaceAll(),!0):!1}selectAllMatches(){return this._model?(this._model.selectAllMatches(),this._editor.focus(),!0):!1}async getGlobalBufferTerm(){return this._editor.getOption(50).globalFindClipboard&&this._editor.hasModel()&&!this._editor.getModel().isTooLargeForSyncing()?this._clipboardService.readFindText():""}setGlobalBufferTerm(e){this._editor.getOption(50).globalFindClipboard&&this._editor.hasModel()&&!this._editor.getModel().isTooLargeForSyncing()&&this._clipboardService.writeFindText(e)}},t3=Gp,Gp.ID="editor.contrib.findController",Gp);xo=t3=zse([La(1,Ae),La(2,Bs),La(3,Ko),La(4,Ii),La(5,so)],xo);let n3=class extends xo{constructor(e,t,i,n,o,r,a,l){super(e,i,r,a,o,l),this._contextViewService=t,this._keybindingService=n,this._widget=null,this._findOptionsWidget=null,this._findWidgetSearchHistory=J5.getOrCreate(r),this._replaceWidgetHistory=e3.getOrCreate(r)}async _start(e,t){this._widget||this._createFindWidget();const i=this._editor.getSelection();let n=!1;switch(this._editor.getOption(50).autoFindInSelection){case"always":n=!0;break;case"never":n=!1;break;case"multiline":{n=!!i&&i.startLineNumber!==i.endLineNumber;break}}e.updateSearchScope=e.updateSearchScope||n,await super._start(e,t),this._widget&&(e.shouldFocus===2?this._widget.focusReplaceInput():e.shouldFocus===1&&this._widget.focusFindInput())}highlightFindOptions(e=!1){this._widget||this._createFindWidget(),this._state.isRevealed&&!e?this._widget.highlightFindOptions():this._findOptionsWidget.highlightFindOptions()}_createFindWidget(){this._widget=this._register(new h5(this._editor,this,this._state,this._contextViewService,this._keybindingService,this._contextKeyService,this._hoverService,this._findWidgetSearchHistory,this._replaceWidgetHistory)),this._findOptionsWidget=this._register(new Q5(this._editor,this._state,this._keybindingService))}saveViewState(){return this._widget?.getViewState()}restoreViewState(e){this._widget?.setViewState(e)}};n3=zse([La(1,Rh),La(2,Ae),La(3,ft),La(4,Ii),La(5,Bs),La(6,Ko),La(7,so)],n3);const _Ae=qR(new UR({id:Ut.StartFindAction,label:V(947,"Find"),precondition:K.or(N.focus,K.has("editorIsOpen")),kbOpts:{kbExpr:null,primary:2084,weight:100},menuOpts:{menuId:fe.MenubarEditMenu,group:"3_find",title:m(941,"&&Find"),order:1}}));_Ae.addImplementation(0,(s,e,t)=>{const i=xo.get(e);return i?i.start({forceRevealReplace:!1,seedSearchStringFromSelection:e.getOption(50).seedSearchStringFromSelection!=="never"?"single":"none",seedSearchStringFromNonEmptySelection:e.getOption(50).seedSearchStringFromSelection==="selection",seedSearchStringFromGlobalClipboard:e.getOption(50).globalFindClipboard,shouldFocus:1,shouldAnimate:!0,updateSearchScope:!1,loop:e.getOption(50).loop}):!1});const bAe={description:"Open a new In-Editor Find Widget.",args:[{name:"Open a new In-Editor Find Widget args",schema:{properties:{searchString:{type:"string"},replaceString:{type:"string"},isRegex:{type:"boolean"},matchWholeWord:{type:"boolean"},isCaseSensitive:{type:"boolean"},preserveCase:{type:"boolean"},findInSelection:{type:"boolean"}}}}]};class wAe extends ue{constructor(){super({id:Ut.StartFindWithArgs,label:V(948,"Find with Arguments"),precondition:void 0,kbOpts:{kbExpr:null,primary:0,weight:100},metadata:bAe})}async run(e,t,i){const n=xo.get(t);if(n){const o=i?{searchString:i.searchString,replaceString:i.replaceString,isReplaceRevealed:i.replaceString!==void 0,isRegex:i.isRegex,wholeWord:i.matchWholeWord,matchCase:i.isCaseSensitive,preserveCase:i.preserveCase}:{};await n.start({forceRevealReplace:!1,seedSearchStringFromSelection:n.getState().searchString.length===0&&t.getOption(50).seedSearchStringFromSelection!=="never"?"single":"none",seedSearchStringFromNonEmptySelection:t.getOption(50).seedSearchStringFromSelection==="selection",seedSearchStringFromGlobalClipboard:!0,shouldFocus:1,shouldAnimate:!0,updateSearchScope:i?.findInSelection||!1,loop:t.getOption(50).loop},o),n.setGlobalBufferTerm(n.getState().searchString)}}}class vAe extends ue{constructor(){super({id:Ut.StartFindWithSelection,label:V(949,"Find with Selection"),precondition:void 0,kbOpts:{kbExpr:null,primary:0,mac:{primary:2083},weight:100}})}async run(e,t){const i=xo.get(t);i&&(await i.start({forceRevealReplace:!1,seedSearchStringFromSelection:"multiple",seedSearchStringFromNonEmptySelection:!1,seedSearchStringFromGlobalClipboard:!1,shouldFocus:0,shouldAnimate:!0,updateSearchScope:!1,loop:t.getOption(50).loop}),i.setGlobalBufferTerm(i.getState().searchString))}}async function $se(s,e){const t=xo.get(s);if(!t)return;const i=()=>(e?t.moveToNextMatch():t.moveToPrevMatch())?(t.editor.pushUndoStop(),!0):!1;i()||(await t.start({forceRevealReplace:!1,seedSearchStringFromSelection:t.getState().searchString.length===0&&s.getOption(50).seedSearchStringFromSelection!=="never"?"single":"none",seedSearchStringFromNonEmptySelection:s.getOption(50).seedSearchStringFromSelection==="selection",seedSearchStringFromGlobalClipboard:!0,shouldFocus:0,shouldAnimate:!0,updateSearchScope:!1,loop:s.getOption(50).loop}),i())}const CAe=qR(new UR({id:Ut.NextMatchFindAction,label:V(950,"Find Next"),precondition:void 0,kbOpts:[{kbExpr:N.focus,primary:61,mac:{primary:2085,secondary:[61]},weight:100},{kbExpr:K.and(N.focus,RA),primary:3,weight:100}]}));CAe.addImplementation(0,async(s,e,t)=>$se(e,!0));const yAe=qR(new UR({id:Ut.PreviousMatchFindAction,label:V(951,"Find Previous"),precondition:void 0,kbOpts:[{kbExpr:N.focus,primary:1085,mac:{primary:3109,secondary:[1085]},weight:100},{kbExpr:K.and(N.focus,RA),primary:1027,weight:100}]}));yAe.addImplementation(0,async(s,e,t)=>$se(e,!1));class SAe extends ue{constructor(){super({id:Ut.GoToMatchFindAction,label:V(952,"Go to Match..."),precondition:Oh}),this._highlightDecorations=[]}run(e,t,i){const n=xo.get(t);if(!n)return;const o=n.getState().matchesCount;if(o<1){e.get(Ii).notify({severity:oC.Warning,message:m(942,"No matches. Try searching for something else.")});return}const r=e.get(bs),a=new U,l=a.add(r.createInputBox());l.placeholder=m(943,"Type a number to go to a specific match (between 1 and {0})",o);const c=h=>{const u=parseInt(h);if(isNaN(u))return;const g=n.getState().matchesCount;if(u>0&&u<=g)return u-1;if(u<0&&u>=-g)return g+u},d=h=>{const u=c(h);if(typeof u=="number"){l.validationMessage=void 0,n.goToMatch(u);const g=n.getState().currentMatch;g&&this.addDecorations(t,g)}else l.validationMessage=m(944,"Please type a number between 1 and {0}",n.getState().matchesCount),this.clearDecorations(t)};a.add(l.onDidChangeValue(h=>{d(h)})),a.add(l.onDidAccept(()=>{const h=c(l.value);typeof h=="number"?(n.goToMatch(h),l.hide()):l.validationMessage=m(945,"Please type a number between 1 and {0}",n.getState().matchesCount)})),a.add(l.onDidHide(()=>{this.clearDecorations(t),a.dispose()})),l.show()}clearDecorations(e){e.changeDecorations(t=>{this._highlightDecorations=t.deltaDecorations(this._highlightDecorations,[])})}addDecorations(e,t){e.changeDecorations(i=>{this._highlightDecorations=i.deltaDecorations(this._highlightDecorations,[{range:t,options:{description:"find-match-quick-access-range-highlight",className:"rangeHighlight",isWholeLine:!0}},{range:t,options:{description:"find-match-quick-access-range-highlight-overview",overviewRuler:{color:Ci(fee),position:Sr.Full}}}])})}}class Use extends ue{async run(e,t){const i=xo.get(t);if(!i)return;const n=i3(t,"single",!1);n&&i.setSearchString(n),this._run(i)||(await i.start({forceRevealReplace:!1,seedSearchStringFromSelection:"none",seedSearchStringFromNonEmptySelection:!1,seedSearchStringFromGlobalClipboard:!1,shouldFocus:0,shouldAnimate:!0,updateSearchScope:!1,loop:t.getOption(50).loop}),this._run(i))}}class xAe extends Use{constructor(){super({id:Ut.NextSelectionMatchFindAction,label:V(953,"Find Next Selection"),precondition:void 0,kbOpts:{kbExpr:N.focus,primary:2109,weight:100}})}_run(e){return e.moveToNextMatch()}}class LAe extends Use{constructor(){super({id:Ut.PreviousSelectionMatchFindAction,label:V(954,"Find Previous Selection"),precondition:void 0,kbOpts:{kbExpr:N.focus,primary:3133,weight:100}})}_run(e){return e.moveToPrevMatch()}}const kAe=qR(new UR({id:Ut.StartFindReplaceAction,label:V(955,"Replace"),precondition:K.or(N.focus,K.has("editorIsOpen")),kbOpts:{kbExpr:null,primary:2086,mac:{primary:2596},weight:100},menuOpts:{menuId:fe.MenubarEditMenu,group:"3_find",title:m(946,"&&Replace"),order:2}}));kAe.addImplementation(0,(s,e,t)=>{if(!e.hasModel()||e.getOption(104))return!1;const i=xo.get(e);if(!i)return!1;const n=e.getSelection(),o=i.isFindInputFocused(),r=!n.isEmpty()&&n.startLineNumber===n.endLineNumber&&e.getOption(50).seedSearchStringFromSelection!=="never"&&!o,a=o||r?2:1;return i.start({forceRevealReplace:!0,seedSearchStringFromSelection:r?"single":"none",seedSearchStringFromNonEmptySelection:e.getOption(50).seedSearchStringFromSelection==="selection",seedSearchStringFromGlobalClipboard:e.getOption(50).seedSearchStringFromSelection!=="never",shouldFocus:a,shouldAnimate:!0,updateSearchScope:!1,loop:e.getOption(50).loop})});lt(xo.ID,n3,0);ae(wAe);ae(vAe);ae(SAe);ae(xAe);ae(LAe);const Ol=hn.bindToContribution(xo.get);le(new Ol({id:Ut.CloseFindWidgetCommand,precondition:Oh,handler:s=>s.closeFindWidget(),kbOpts:{weight:105,kbExpr:K.and(N.focus,K.not("isComposing")),primary:9,secondary:[1033]}}));le(new Ol({id:Ut.ToggleCaseSensitiveCommand,precondition:void 0,handler:s=>s.toggleCaseSensitive(),kbOpts:{weight:105,kbExpr:N.focus,primary:EL.primary,mac:EL.mac,win:EL.win,linux:EL.linux}}));le(new Ol({id:Ut.ToggleWholeWordCommand,precondition:void 0,handler:s=>s.toggleWholeWords(),kbOpts:{weight:105,kbExpr:N.focus,primary:IL.primary,mac:IL.mac,win:IL.win,linux:IL.linux}}));le(new Ol({id:Ut.ToggleRegexCommand,precondition:void 0,handler:s=>s.toggleRegex(),kbOpts:{weight:105,kbExpr:N.focus,primary:DL.primary,mac:DL.mac,win:DL.win,linux:DL.linux}}));le(new Ol({id:Ut.ToggleSearchScopeCommand,precondition:void 0,handler:s=>s.toggleSearchScope(),kbOpts:{weight:105,kbExpr:N.focus,primary:TL.primary,mac:TL.mac,win:TL.win,linux:TL.linux}}));le(new Ol({id:Ut.TogglePreserveCaseCommand,precondition:void 0,handler:s=>s.togglePreserveCase(),kbOpts:{weight:105,kbExpr:N.focus,primary:NL.primary,mac:NL.mac,win:NL.win,linux:NL.linux}}));le(new Ol({id:Ut.ReplaceOneAction,precondition:Oh,handler:s=>s.replace(),kbOpts:{weight:105,kbExpr:N.focus,primary:3094}}));le(new Ol({id:Ut.ReplaceOneAction,precondition:Oh,handler:s=>s.replace(),kbOpts:{weight:105,kbExpr:K.and(N.focus,lH),primary:3}}));le(new Ol({id:Ut.ReplaceAllAction,precondition:Oh,handler:s=>s.replaceAll(),kbOpts:{weight:105,kbExpr:N.focus,primary:2563}}));le(new Ol({id:Ut.ReplaceAllAction,precondition:Oh,handler:s=>s.replaceAll(),kbOpts:{weight:105,kbExpr:K.and(N.focus,lH),primary:void 0,mac:{primary:2051}}}));le(new Ol({id:Ut.SelectAllMatchesAction,precondition:Oh,handler:s=>s.selectAllMatches(),kbOpts:{weight:105,kbExpr:N.focus,primary:515}}));const EAe={0:" ",1:"u",2:"r"},cG=65535,ol=16777215,dG=4278190080;class lO{constructor(e){const t=Math.ceil(e/32);this._states=new Uint32Array(t)}get(e){const t=e/32|0,i=e%32;return(this._states[t]&1<<i)!==0}set(e,t){const i=e/32|0,n=e%32,o=this._states[i];t?this._states[i]=o|1<<n:this._states[i]=o&~(1<<n)}}class pr{constructor(e,t,i){if(e.length!==t.length||e.length>cG)throw new Error("invalid startIndexes or endIndexes size");this._startIndexes=e,this._endIndexes=t,this._collapseStates=new lO(e.length),this._userDefinedStates=new lO(e.length),this._recoveredStates=new lO(e.length),this._types=i,this._parentsComputed=!1}ensureParentIndices(){if(!this._parentsComputed){this._parentsComputed=!0;const e=[],t=(i,n)=>{const o=e[e.length-1];return this.getStartLineNumber(o)<=i&&this.getEndLineNumber(o)>=n};for(let i=0,n=this._startIndexes.length;i<n;i++){const o=this._startIndexes[i],r=this._endIndexes[i];if(o>ol||r>ol)throw new Error("startLineNumber or endLineNumber must not exceed "+ol);for(;e.length>0&&!t(o,r);)e.pop();const a=e.length>0?e[e.length-1]:-1;e.push(i),this._startIndexes[i]=o+((a&255)<<24),this._endIndexes[i]=r+((a&65280)<<16)}}}get length(){return this._startIndexes.length}getStartLineNumber(e){return this._startIndexes[e]&ol}getEndLineNumber(e){return this._endIndexes[e]&ol}getType(e){return this._types?this._types[e]:void 0}hasTypes(){return!!this._types}isCollapsed(e){return this._collapseStates.get(e)}setCollapsed(e,t){this._collapseStates.set(e,t)}isUserDefined(e){return this._userDefinedStates.get(e)}setUserDefined(e,t){return this._userDefinedStates.set(e,t)}isRecovered(e){return this._recoveredStates.get(e)}setRecovered(e,t){return this._recoveredStates.set(e,t)}getSource(e){return this.isUserDefined(e)?1:this.isRecovered(e)?2:0}setSource(e,t){t===1?(this.setUserDefined(e,!0),this.setRecovered(e,!1)):t===2?(this.setUserDefined(e,!1),this.setRecovered(e,!0)):(this.setUserDefined(e,!1),this.setRecovered(e,!1))}setCollapsedAllOfType(e,t){let i=!1;if(this._types)for(let n=0;n<this._types.length;n++)this._types[n]===e&&(this.setCollapsed(n,t),i=!0);return i}toRegion(e){return new IAe(this,e)}getParentIndex(e){this.ensureParentIndices();const t=((this._startIndexes[e]&dG)>>>24)+((this._endIndexes[e]&dG)>>>16);return t===cG?-1:t}contains(e,t){return this.getStartLineNumber(e)<=t&&this.getEndLineNumber(e)>=t}findIndex(e){let t=0,i=this._startIndexes.length;if(i===0)return-1;for(;t<i;){const n=Math.floor((t+i)/2);e<this.getStartLineNumber(n)?i=n:t=n+1}return t-1}findRange(e){let t=this.findIndex(e);if(t>=0){if(this.getEndLineNumber(t)>=e)return t;for(t=this.getParentIndex(t);t!==-1;){if(this.contains(t,e))return t;t=this.getParentIndex(t)}}return-1}toString(){const e=[];for(let t=0;t<this.length;t++)e[t]=`[${EAe[this.getSource(t)]}${this.isCollapsed(t)?"+":"-"}] ${this.getStartLineNumber(t)}/${this.getEndLineNumber(t)}`;return e.join(", ")}toFoldRange(e){return{startLineNumber:this._startIndexes[e]&ol,endLineNumber:this._endIndexes[e]&ol,type:this._types?this._types[e]:void 0,isCollapsed:this.isCollapsed(e),source:this.getSource(e)}}static fromFoldRanges(e){const t=e.length,i=new Uint32Array(t),n=new Uint32Array(t);let o=[],r=!1;for(let l=0;l<t;l++){const c=e[l];i[l]=c.startLineNumber,n[l]=c.endLineNumber,o.push(c.type),c.type&&(r=!0)}r||(o=void 0);const a=new pr(i,n,o);for(let l=0;l<t;l++)e[l].isCollapsed&&a.setCollapsed(l,!0),a.setSource(l,e[l].source);return a}static sanitizeAndMerge(e,t,i,n){i=i??Number.MAX_VALUE;const o=(_,b)=>Array.isArray(_)?(w=>w<b?_[w]:void 0):(w=>w<b?_.toFoldRange(w):void 0),r=o(e,e.length),a=o(t,t.length);let l=0,c=0,d=r(0),h=a(0);const u=[];let g,f=0;const p=[];for(;d||h;){let _;if(h&&(!d||d.startLineNumber>=h.startLineNumber))d&&d.startLineNumber===h.startLineNumber?(h.source===1?_=h:(_=d,_.isCollapsed=h.isCollapsed&&(d.endLineNumber===h.endLineNumber||!n?.startsInside(d.startLineNumber+1,d.endLineNumber+1)),_.source=0),d=r(++l)):(_=h,h.isCollapsed&&h.source===0&&(_.source=2)),h=a(++c);else{let b=c,w=h;for(;;){if(!w||w.startLineNumber>d.endLineNumber){_=d;break}if(w.source===1&&w.endLineNumber>d.endLineNumber)break;w=a(++b)}d=r(++l)}if(_){for(;g&&g.endLineNumber<_.startLineNumber;)g=u.pop();_.endLineNumber>_.startLineNumber&&_.startLineNumber>f&&_.endLineNumber<=i&&(!g||g.endLineNumber>=_.endLineNumber)&&(p.push(_),f=_.startLineNumber,g&&u.push(g),g=_)}}return p}}class IAe{constructor(e,t){this.ranges=e,this.index=t}get startLineNumber(){return this.ranges.getStartLineNumber(this.index)}get endLineNumber(){return this.ranges.getEndLineNumber(this.index)}get regionIndex(){return this.index}get parentIndex(){return this.ranges.getParentIndex(this.index)}get isCollapsed(){return this.ranges.isCollapsed(this.index)}containedBy(e){return e.startLineNumber<=this.startLineNumber&&e.endLineNumber>=this.endLineNumber}containsLine(e){return this.startLineNumber<=e&&e<=this.endLineNumber}}class DAe{get regions(){return this._regions}get textModel(){return this._textModel}constructor(e,t){this._updateEventEmitter=new M,this.onDidChange=this._updateEventEmitter.event,this._textModel=e,this._decorationProvider=t,this._regions=new pr(new Uint32Array(0),new Uint32Array(0)),this._editorDecorationIds=[]}toggleCollapseState(e){if(!e.length)return;e=e.sort((i,n)=>i.regionIndex-n.regionIndex);const t={};this._decorationProvider.changeDecorations(i=>{let n=0,o=-1,r=-1;const a=l=>{for(;n<l;){const c=this._regions.getEndLineNumber(n),d=this._regions.isCollapsed(n);if(c<=o){const h=this.regions.getSource(n)!==0;i.changeDecorationOptions(this._editorDecorationIds[n],this._decorationProvider.getDecorationOption(d,c<=r,h))}d&&c>r&&(r=c),n++}};for(const l of e){const c=l.regionIndex,d=this._editorDecorationIds[c];if(d&&!t[d]){t[d]=!0,a(c);const h=!this._regions.isCollapsed(c);this._regions.setCollapsed(c,h),o=Math.max(o,this._regions.getEndLineNumber(c))}}a(this._regions.length)}),this._updateEventEmitter.fire({model:this,collapseStateChanged:e})}removeManualRanges(e){const t=new Array,i=n=>{for(const o of e)if(!(o.startLineNumber>n.endLineNumber||n.startLineNumber>o.endLineNumber))return!0;return!1};for(let n=0;n<this._regions.length;n++){const o=this._regions.toFoldRange(n);(o.source===0||!i(o))&&t.push(o)}this.updatePost(pr.fromFoldRanges(t))}update(e,t){const i=this._currentFoldedOrManualRanges(t),n=pr.sanitizeAndMerge(e,i,this._textModel.getLineCount(),t);this.updatePost(pr.fromFoldRanges(n))}updatePost(e){const t=[];let i=-1;for(let n=0,o=e.length;n<o;n++){const r=e.getStartLineNumber(n),a=e.getEndLineNumber(n),l=e.isCollapsed(n),c=e.getSource(n)!==0,d={startLineNumber:r,startColumn:this._textModel.getLineMaxColumn(r),endLineNumber:a,endColumn:this._textModel.getLineMaxColumn(a)+1};t.push({range:d,options:this._decorationProvider.getDecorationOption(l,a<=i,c)}),l&&a>i&&(i=a)}this._decorationProvider.changeDecorations(n=>this._editorDecorationIds=n.deltaDecorations(this._editorDecorationIds,t)),this._regions=e,this._updateEventEmitter.fire({model:this})}_currentFoldedOrManualRanges(e){const t=[];for(let i=0,n=this._regions.length;i<n;i++){let o=this.regions.isCollapsed(i);const r=this.regions.getSource(i);if(o||r!==0){const a=this._regions.toFoldRange(i),l=this._textModel.getDecorationRange(this._editorDecorationIds[i]);l&&(o&&e?.startsInside(l.startLineNumber+1,l.endLineNumber)&&(o=!1),t.push({startLineNumber:l.startLineNumber,endLineNumber:l.endLineNumber,type:a.type,isCollapsed:o,source:r}))}}return t}getMemento(){const e=this._currentFoldedOrManualRanges(),t=[],i=this._textModel.getLineCount();for(let n=0,o=e.length;n<o;n++){const r=e[n];if(r.startLineNumber>=r.endLineNumber||r.startLineNumber<1||r.endLineNumber>i)continue;const a=this._getLinesChecksum(r.startLineNumber+1,r.endLineNumber);t.push({startLineNumber:r.startLineNumber,endLineNumber:r.endLineNumber,isCollapsed:r.isCollapsed,source:r.source,checksum:a})}return t.length>0?t:void 0}applyMemento(e){if(!Array.isArray(e))return;const t=[],i=this._textModel.getLineCount();for(const o of e){if(o.startLineNumber>=o.endLineNumber||o.startLineNumber<1||o.endLineNumber>i)continue;const r=this._getLinesChecksum(o.startLineNumber+1,o.endLineNumber);(!o.checksum||r===o.checksum)&&t.push({startLineNumber:o.startLineNumber,endLineNumber:o.endLineNumber,type:void 0,isCollapsed:o.isCollapsed??!0,source:o.source??0})}const n=pr.sanitizeAndMerge(this._regions,t,i);this.updatePost(pr.fromFoldRanges(n))}_getLinesChecksum(e,t){return j0(this._textModel.getLineContent(e)+this._textModel.getLineContent(t))%1e6}dispose(){this._decorationProvider.removeDecorations(this._editorDecorationIds)}getAllRegionsAtLine(e,t){const i=[];if(this._regions){let n=this._regions.findRange(e),o=1;for(;n>=0;){const r=this._regions.toRegion(n);(!t||t(r,o))&&i.push(r),o++,n=r.parentIndex}}return i}getRegionAtLine(e){if(this._regions){const t=this._regions.findRange(e);if(t>=0)return this._regions.toRegion(t)}return null}getRegionsInside(e,t){const i=[],n=e?e.regionIndex+1:0,o=e?e.endLineNumber:Number.MAX_VALUE;if(t&&t.length===2){const r=[];for(let a=n,l=this._regions.length;a<l;a++){const c=this._regions.toRegion(a);if(this._regions.getStartLineNumber(a)<o){for(;r.length>0&&!c.containedBy(r[r.length-1]);)r.pop();r.push(c),t(c,r.length)&&i.push(c)}else break}}else for(let r=n,a=this._regions.length;r<a;r++){const l=this._regions.toRegion(r);if(this._regions.getStartLineNumber(r)<o)(!t||t(l))&&i.push(l);else break}return i}}function WH(s,e,t){const i=[];for(const n of t){const o=s.getRegionAtLine(n);if(o){const r=!o.isCollapsed;if(i.push(o),e>1){const a=s.getRegionsInside(o,(l,c)=>l.isCollapsed!==r&&c<e);i.push(...a)}}}s.toggleCollapseState(i)}function gC(s,e,t=Number.MAX_VALUE,i){const n=[];if(i&&i.length>0)for(const o of i){const r=s.getRegionAtLine(o);if(r&&(r.isCollapsed!==e&&n.push(r),t>1)){const a=s.getRegionsInside(r,(l,c)=>l.isCollapsed!==e&&c<t);n.push(...a)}}else{const o=s.getRegionsInside(null,(r,a)=>r.isCollapsed!==e&&a<t);n.push(...o)}s.toggleCollapseState(n)}function qse(s,e,t,i){const n=[];for(const o of i){const r=s.getAllRegionsAtLine(o,(a,l)=>a.isCollapsed!==e&&l<=t);n.push(...r)}s.toggleCollapseState(n)}function TAe(s,e,t){const i=[];for(const n of t){const o=s.getAllRegionsAtLine(n,r=>r.isCollapsed!==e);o.length>0&&i.push(o[0])}s.toggleCollapseState(i)}function NAe(s,e,t,i){const n=(r,a)=>a===e&&r.isCollapsed!==t&&!i.some(l=>r.containsLine(l)),o=s.getRegionsInside(null,n);s.toggleCollapseState(o)}function jse(s,e,t){const i=[];for(const r of t){const a=s.getAllRegionsAtLine(r,void 0);a.length>0&&i.push(a[0])}const n=r=>i.every(a=>!a.containedBy(r)&&!r.containedBy(a))&&r.isCollapsed!==e,o=s.getRegionsInside(null,n);s.toggleCollapseState(o)}function HH(s,e,t){const i=s.textModel,n=s.regions,o=[];for(let r=n.length-1;r>=0;r--)if(t!==n.isCollapsed(r)){const a=n.getStartLineNumber(r);e.test(i.getLineContent(a))&&o.push(n.toRegion(r))}s.toggleCollapseState(o)}function VH(s,e,t){const i=s.regions,n=[];for(let o=i.length-1;o>=0;o--)t!==i.isCollapsed(o)&&e===i.getType(o)&&n.push(i.toRegion(o));s.toggleCollapseState(n)}function RAe(s,e){let t=null;const i=e.getRegionAtLine(s);if(i!==null&&(t=i.startLineNumber,s===t)){const n=i.parentIndex;n!==-1?t=e.regions.getStartLineNumber(n):t=null}return t}function AAe(s,e){let t=e.getRegionAtLine(s);if(t!==null&&t.startLineNumber===s){if(s!==t.startLineNumber)return t.startLineNumber;{const i=t.parentIndex;let n=0;for(i!==-1&&(n=e.regions.getStartLineNumber(t.parentIndex));t!==null;)if(t.regionIndex>0){if(t=e.regions.toRegion(t.regionIndex-1),t.startLineNumber<=n)return null;if(t.parentIndex===i)return t.startLineNumber}else return null}}else if(e.regions.length>0)for(t=e.regions.toRegion(e.regions.length-1);t!==null;){if(t.startLineNumber<s)return t.startLineNumber;t.regionIndex>0?t=e.regions.toRegion(t.regionIndex-1):t=null}return null}function MAe(s,e){let t=e.getRegionAtLine(s);if(t!==null&&t.startLineNumber===s){const i=t.parentIndex;let n=0;if(i!==-1)n=e.regions.getEndLineNumber(t.parentIndex);else{if(e.regions.length===0)return null;n=e.regions.getEndLineNumber(e.regions.length-1)}for(;t!==null;)if(t.regionIndex<e.regions.length){if(t=e.regions.toRegion(t.regionIndex+1),t.startLineNumber>=n)return null;if(t.parentIndex===i)return t.startLineNumber}else return null}else if(e.regions.length>0)for(t=e.regions.toRegion(0);t!==null;){if(t.startLineNumber>s)return t.startLineNumber;t.regionIndex<e.regions.length?t=e.regions.toRegion(t.regionIndex+1):t=null}return null}class OAe{get onDidChange(){return this._updateEventEmitter.event}get hiddenRanges(){return this._hiddenRanges}constructor(e){this._updateEventEmitter=new M,this._hasLineChanges=!1,this._foldingModel=e,this._foldingModelListener=e.onDidChange(t=>this.updateHiddenRanges()),this._hiddenRanges=[],e.regions.length&&this.updateHiddenRanges()}notifyChangeModelContent(e){this._hiddenRanges.length&&!this._hasLineChanges&&(this._hasLineChanges=e.changes.some(t=>t.range.endLineNumber!==t.range.startLineNumber||fg(t.text)[0]!==0))}updateHiddenRanges(){let e=!1;const t=[];let i=0,n=0,o=Number.MAX_VALUE,r=-1;const a=this._foldingModel.regions;for(;i<a.length;i++){if(!a.isCollapsed(i))continue;const l=a.getStartLineNumber(i)+1,c=a.getEndLineNumber(i);o<=l&&c<=r||(!e&&n<this._hiddenRanges.length&&this._hiddenRanges[n].startLineNumber===l&&this._hiddenRanges[n].endLineNumber===c?(t.push(this._hiddenRanges[n]),n++):(e=!0,t.push(new k(l,1,c,1))),o=l,r=c)}(this._hasLineChanges||e||n<this._hiddenRanges.length)&&this.applyHiddenRanges(t)}applyHiddenRanges(e){this._hiddenRanges=e,this._hasLineChanges=!1,this._updateEventEmitter.fire(e)}hasRanges(){return this._hiddenRanges.length>0}isHidden(e){return hG(this._hiddenRanges,e)!==null}adjustSelections(e){let t=!1;const i=this._foldingModel.textModel;let n=null;const o=r=>((!n||!PAe(r,n))&&(n=hG(this._hiddenRanges,r)),n?n.startLineNumber-1:null);for(let r=0,a=e.length;r<a;r++){let l=e[r];const c=o(l.startLineNumber);c&&(l=l.setStartPosition(c,i.getLineMaxColumn(c)),t=!0);const d=o(l.endLineNumber);d&&(l=l.setEndPosition(d,i.getLineMaxColumn(d)),t=!0),e[r]=l}return t}dispose(){this.hiddenRanges.length>0&&(this._hiddenRanges=[],this._updateEventEmitter.fire(this._hiddenRanges)),this._foldingModelListener&&(this._foldingModelListener.dispose(),this._foldingModelListener=null)}}function PAe(s,e){return s>=e.startLineNumber&&s<=e.endLineNumber}function hG(s,e){const t=fS(s,i=>e<i.startLineNumber)-1;return t>=0&&s[t].endLineNumber>=e?s[t]:null}const FAe=5e3,BAe="indent";class zH{constructor(e,t,i){this.editorModel=e,this.languageConfigurationService=t,this.foldingRangesLimit=i,this.id=BAe}dispose(){}compute(e){const t=this.languageConfigurationService.getLanguageConfiguration(this.editorModel.getLanguageId()).foldingRules,i=t&&!!t.offSide,n=t&&t.markers;return Promise.resolve(VAe(this.editorModel,i,n,this.foldingRangesLimit))}}let WAe=class{constructor(e){this._startIndexes=[],this._endIndexes=[],this._indentOccurrences=[],this._length=0,this._foldingRangesLimit=e}insertFirst(e,t,i){if(e>ol||t>ol)return;const n=this._length;this._startIndexes[n]=e,this._endIndexes[n]=t,this._length++,i<1e3&&(this._indentOccurrences[i]=(this._indentOccurrences[i]||0)+1)}toIndentRanges(e){const t=this._foldingRangesLimit.limit;if(this._length<=t){this._foldingRangesLimit.update(this._length,!1);const i=new Uint32Array(this._length),n=new Uint32Array(this._length);for(let o=this._length-1,r=0;o>=0;o--,r++)i[r]=this._startIndexes[o],n[r]=this._endIndexes[o];return new pr(i,n)}else{this._foldingRangesLimit.update(this._length,t);let i=0,n=this._indentOccurrences.length;for(let l=0;l<this._indentOccurrences.length;l++){const c=this._indentOccurrences[l];if(c){if(c+i>t){n=l;break}i+=c}}const o=e.getOptions().tabSize,r=new Uint32Array(t),a=new Uint32Array(t);for(let l=this._length-1,c=0;l>=0;l--){const d=this._startIndexes[l],h=e.getLineContent(d),u=uA(h,o);(u<n||u===n&&i++<t)&&(r[c]=d,a[c]=this._endIndexes[l],c++)}return new pr(r,a)}}};const HAe={limit:FAe,update:()=>{}};function VAe(s,e,t,i=HAe){const n=s.getOptions().tabSize,o=new WAe(i);let r;t&&(r=new RegExp(`(${t.start.source})|(?:${t.end.source})`));const a=[],l=s.getLineCount()+1;a.push({indent:-1,endAbove:l,line:l});for(let c=s.getLineCount();c>0;c--){const d=s.getLineContent(c),h=uA(d,n);let u=a[a.length-1];if(h===-1){e&&(u.endAbove=c);continue}let g;if(r&&(g=d.match(r)))if(g[1]){let f=a.length-1;for(;f>0&&a[f].indent!==-2;)f--;if(f>0){a.length=f+1,u=a[f],o.insertFirst(c,u.line,h),u.line=c,u.indent=h,u.endAbove=c;continue}}else{a.push({indent:-2,endAbove:c,line:c});continue}if(u.indent>h){do a.pop(),u=a[a.length-1];while(u.indent>h);const f=u.endAbove-1;f-c>=1&&o.insertFirst(c,f,h)}u.indent===h?u.endAbove=c:a.push({indent:h,endAbove:c,line:c})}return o.toIndentRanges(s)}const zAe=T("editor.foldBackground",{light:Be(mu,.3),dark:Be(mu,.3),hcDark:null,hcLight:null},m(1002,"Background color behind folded ranges. The color must not be opaque so as not to hide underlying decorations."),!0);T("editor.foldPlaceholderForeground",{light:"#808080",dark:"#808080",hcDark:null,hcLight:null},m(1003,"Color of the collapsed text after the first line of a folded range."));T("editorGutter.foldingControlForeground",$E,m(1004,"Color of the folding control in the editor gutter."));const _D=Qt("folding-expanded",Z.chevronDown,m(1005,"Icon for expanded ranges in the editor glyph margin.")),bD=Qt("folding-collapsed",Z.chevronRight,m(1006,"Icon for collapsed ranges in the editor glyph margin.")),uG=Qt("folding-manual-collapsed",bD,m(1007,"Icon for manually collapsed ranges in the editor glyph margin.")),gG=Qt("folding-manual-expanded",_D,m(1008,"Icon for manually expanded ranges in the editor glyph margin.")),cO={color:Ci(zAe),position:1},J_=m(1009,"Click to expand the range."),VL=m(1010,"Click to collapse the range."),Yi=class Yi{constructor(e){this.editor=e,this.showFoldingControls="mouseover",this.showFoldingHighlights=!0}getDecorationOption(e,t,i){return t?Yi.HIDDEN_RANGE_DECORATION:this.showFoldingControls==="never"?e?this.showFoldingHighlights?Yi.NO_CONTROLS_COLLAPSED_HIGHLIGHTED_RANGE_DECORATION:Yi.NO_CONTROLS_COLLAPSED_RANGE_DECORATION:Yi.NO_CONTROLS_EXPANDED_RANGE_DECORATION:e?i?this.showFoldingHighlights?Yi.MANUALLY_COLLAPSED_HIGHLIGHTED_VISUAL_DECORATION:Yi.MANUALLY_COLLAPSED_VISUAL_DECORATION:this.showFoldingHighlights?Yi.COLLAPSED_HIGHLIGHTED_VISUAL_DECORATION:Yi.COLLAPSED_VISUAL_DECORATION:this.showFoldingControls==="mouseover"?i?Yi.MANUALLY_EXPANDED_AUTO_HIDE_VISUAL_DECORATION:Yi.EXPANDED_AUTO_HIDE_VISUAL_DECORATION:i?Yi.MANUALLY_EXPANDED_VISUAL_DECORATION:Yi.EXPANDED_VISUAL_DECORATION}changeDecorations(e){return this.editor.changeDecorations(e)}removeDecorations(e){this.editor.removeDecorations(e)}};Yi.COLLAPSED_VISUAL_DECORATION=Fe.register({description:"folding-collapsed-visual-decoration",stickiness:0,afterContentClassName:"inline-folded",isWholeLine:!0,linesDecorationsTooltip:J_,firstLineDecorationClassName:xe.asClassName(bD)}),Yi.COLLAPSED_HIGHLIGHTED_VISUAL_DECORATION=Fe.register({description:"folding-collapsed-highlighted-visual-decoration",stickiness:0,afterContentClassName:"inline-folded",className:"folded-background",minimap:cO,isWholeLine:!0,linesDecorationsTooltip:J_,firstLineDecorationClassName:xe.asClassName(bD)}),Yi.MANUALLY_COLLAPSED_VISUAL_DECORATION=Fe.register({description:"folding-manually-collapsed-visual-decoration",stickiness:0,afterContentClassName:"inline-folded",isWholeLine:!0,linesDecorationsTooltip:J_,firstLineDecorationClassName:xe.asClassName(uG)}),Yi.MANUALLY_COLLAPSED_HIGHLIGHTED_VISUAL_DECORATION=Fe.register({description:"folding-manually-collapsed-highlighted-visual-decoration",stickiness:0,afterContentClassName:"inline-folded",className:"folded-background",minimap:cO,isWholeLine:!0,linesDecorationsTooltip:J_,firstLineDecorationClassName:xe.asClassName(uG)}),Yi.NO_CONTROLS_COLLAPSED_RANGE_DECORATION=Fe.register({description:"folding-no-controls-range-decoration",stickiness:0,afterContentClassName:"inline-folded",isWholeLine:!0,linesDecorationsTooltip:J_}),Yi.NO_CONTROLS_COLLAPSED_HIGHLIGHTED_RANGE_DECORATION=Fe.register({description:"folding-no-controls-range-decoration",stickiness:0,afterContentClassName:"inline-folded",className:"folded-background",minimap:cO,isWholeLine:!0,linesDecorationsTooltip:J_}),Yi.EXPANDED_VISUAL_DECORATION=Fe.register({description:"folding-expanded-visual-decoration",stickiness:1,isWholeLine:!0,firstLineDecorationClassName:"alwaysShowFoldIcons "+xe.asClassName(_D),linesDecorationsTooltip:VL}),Yi.EXPANDED_AUTO_HIDE_VISUAL_DECORATION=Fe.register({description:"folding-expanded-auto-hide-visual-decoration",stickiness:1,isWholeLine:!0,firstLineDecorationClassName:xe.asClassName(_D),linesDecorationsTooltip:VL}),Yi.MANUALLY_EXPANDED_VISUAL_DECORATION=Fe.register({description:"folding-manually-expanded-visual-decoration",stickiness:0,isWholeLine:!0,firstLineDecorationClassName:"alwaysShowFoldIcons "+xe.asClassName(gG),linesDecorationsTooltip:VL}),Yi.MANUALLY_EXPANDED_AUTO_HIDE_VISUAL_DECORATION=Fe.register({description:"folding-manually-expanded-auto-hide-visual-decoration",stickiness:0,isWholeLine:!0,firstLineDecorationClassName:xe.asClassName(gG),linesDecorationsTooltip:VL}),Yi.NO_CONTROLS_EXPANDED_RANGE_DECORATION=Fe.register({description:"folding-no-controls-range-decoration",stickiness:0,isWholeLine:!0}),Yi.HIDDEN_RANGE_DECORATION=Fe.register({description:"folding-hidden-range-decoration",stickiness:1});let s3=Yi;const $Ae={},UAe="syntax";class $H{constructor(e,t,i,n,o){this.editorModel=e,this.providers=t,this.handleFoldingRangesChange=i,this.foldingRangesLimit=n,this.fallbackRangeProvider=o,this.id=UAe,this.disposables=new U,o&&this.disposables.add(o);for(const r of t)typeof r.onDidChange=="function"&&this.disposables.add(r.onDidChange(i))}compute(e){return qAe(this.providers,this.editorModel,e).then(t=>this.editorModel.isDisposed()?null:t?GAe(t,this.foldingRangesLimit):this.fallbackRangeProvider?.compute(e)??null)}dispose(){this.disposables.dispose()}}function qAe(s,e,t){let i=null;const n=s.map((o,r)=>Promise.resolve(o.provideFoldingRanges(e,$Ae,t)).then(a=>{if(!t.isCancellationRequested&&Array.isArray(a)){Array.isArray(i)||(i=[]);const l=e.getLineCount();for(const c of a)c.start>0&&c.end>c.start&&c.end<=l&&i.push({start:c.start,end:c.end,rank:r,kind:c.kind})}},ji));return Promise.all(n).then(o=>i)}class jAe{constructor(e){this._startIndexes=[],this._endIndexes=[],this._nestingLevels=[],this._nestingLevelCounts=[],this._types=[],this._length=0,this._foldingRangesLimit=e}add(e,t,i,n){if(e>ol||t>ol)return;const o=this._length;this._startIndexes[o]=e,this._endIndexes[o]=t,this._nestingLevels[o]=n,this._types[o]=i,this._length++,n<30&&(this._nestingLevelCounts[n]=(this._nestingLevelCounts[n]||0)+1)}toIndentRanges(){const e=this._foldingRangesLimit.limit;if(this._length<=e){this._foldingRangesLimit.update(this._length,!1);const t=new Uint32Array(this._length),i=new Uint32Array(this._length);for(let n=0;n<this._length;n++)t[n]=this._startIndexes[n],i[n]=this._endIndexes[n];return new pr(t,i,this._types)}else{this._foldingRangesLimit.update(this._length,e);let t=0,i=this._nestingLevelCounts.length;for(let a=0;a<this._nestingLevelCounts.length;a++){const l=this._nestingLevelCounts[a];if(l){if(l+t>e){i=a;break}t+=l}}const n=new Uint32Array(e),o=new Uint32Array(e),r=[];for(let a=0,l=0;a<this._length;a++){const c=this._nestingLevels[a];(c<i||c===i&&t++<e)&&(n[l]=this._startIndexes[a],o[l]=this._endIndexes[a],r[l]=this._types[a],l++)}return new pr(n,o,r)}}}function GAe(s,e){const t=s.sort((r,a)=>{let l=r.start-a.start;return l===0&&(l=r.rank-a.rank),l}),i=new jAe(e);let n;const o=[];for(const r of t)if(!n)n=r,i.add(r.start,r.end,r.kind&&r.kind.value,o.length);else if(r.start>n.start)if(r.end<=n.end)o.push(n),n=r,i.add(r.start,r.end,r.kind&&r.kind.value,o.length);else{if(r.start>n.end){do n=o.pop();while(n&&r.start>n.end);n&&o.push(n),n=r}i.add(r.start,r.end,r.kind&&r.kind.value,o.length)}return i.toIndentRanges()}var KAe=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},GC=function(s,e){return function(t,i){e(t,i,s)}},m1;const Hn=new ce("foldingEnabled",!1);var Kp;let vg=(Kp=class extends F{static get(e){return e.getContribution(m1.ID)}static getFoldingRangeProviders(e,t){const i=e.foldingRangeProvider.ordered(t);return m1._foldingRangeSelector?.(i,t)??i}constructor(e,t,i,n,o,r){super(),this.contextKeyService=t,this.languageConfigurationService=i,this.languageFeaturesService=r,this.localToDispose=this._register(new U),this.editor=e,this._foldingLimitReporter=this._register(new Gse(e));const a=this.editor.getOptions();this._isEnabled=a.get(52),this._useFoldingProviders=a.get(53)!=="indentation",this._unfoldOnClickAfterEndOfLine=a.get(57),this._restoringViewState=!1,this._currentModelHasFoldedImports=!1,this._foldingImportsByDefault=a.get(55),this.updateDebounceInfo=o.for(r.foldingRangeProvider,"Folding",{min:200}),this.foldingModel=null,this.hiddenRangeModel=null,this.rangeProvider=null,this.foldingRegionPromise=null,this.foldingModelPromise=null,this.updateScheduler=null,this.cursorChangedScheduler=null,this.mouseDownInfo=null,this.foldingDecorationProvider=new s3(e),this.foldingDecorationProvider.showFoldingControls=a.get(126),this.foldingDecorationProvider.showFoldingHighlights=a.get(54),this.foldingEnabled=Hn.bindTo(this.contextKeyService),this.foldingEnabled.set(this._isEnabled),this._register(this.editor.onDidChangeModel(()=>this.onModelChanged())),this._register(this.editor.onDidChangeConfiguration(l=>{if(l.hasChanged(52)&&(this._isEnabled=this.editor.getOptions().get(52),this.foldingEnabled.set(this._isEnabled),this.onModelChanged()),l.hasChanged(56)&&this.onModelChanged(),l.hasChanged(126)||l.hasChanged(54)){const c=this.editor.getOptions();this.foldingDecorationProvider.showFoldingControls=c.get(126),this.foldingDecorationProvider.showFoldingHighlights=c.get(54),this.triggerFoldingModelChanged()}l.hasChanged(53)&&(this._useFoldingProviders=this.editor.getOptions().get(53)!=="indentation",this.onFoldingStrategyChanged()),l.hasChanged(57)&&(this._unfoldOnClickAfterEndOfLine=this.editor.getOptions().get(57)),l.hasChanged(55)&&(this._foldingImportsByDefault=this.editor.getOptions().get(55))})),this.onModelChanged()}saveViewState(){const e=this.editor.getModel();if(!e||!this._isEnabled||e.isTooLargeForTokenization())return{};if(this.foldingModel){const t=this.foldingModel.getMemento(),i=this.rangeProvider?this.rangeProvider.id:void 0;return{collapsedRegions:t,lineCount:e.getLineCount(),provider:i,foldedImports:this._currentModelHasFoldedImports}}}restoreViewState(e){const t=this.editor.getModel();if(!(!t||!this._isEnabled||t.isTooLargeForTokenization()||!this.hiddenRangeModel)&&e&&(this._currentModelHasFoldedImports=!!e.foldedImports,e.collapsedRegions&&e.collapsedRegions.length>0&&this.foldingModel)){this._restoringViewState=!0;try{this.foldingModel.applyMemento(e.collapsedRegions)}finally{this._restoringViewState=!1}}}onModelChanged(){this.localToDispose.clear();const e=this.editor.getModel();!this._isEnabled||!e||e.isTooLargeForTokenization()||(this._currentModelHasFoldedImports=!1,this.foldingModel=new DAe(e,this.foldingDecorationProvider),this.localToDispose.add(this.foldingModel),this.hiddenRangeModel=new OAe(this.foldingModel),this.localToDispose.add(this.hiddenRangeModel),this.localToDispose.add(this.hiddenRangeModel.onDidChange(t=>this.onHiddenRangesChanges(t))),this.updateScheduler=new sa(this.updateDebounceInfo.get(e)),this.localToDispose.add(this.updateScheduler),this.cursorChangedScheduler=new Rt(()=>this.revealCursor(),200),this.localToDispose.add(this.cursorChangedScheduler),this.localToDispose.add(this.languageFeaturesService.foldingRangeProvider.onDidChange(()=>this.onFoldingStrategyChanged())),this.localToDispose.add(this.editor.onDidChangeModelLanguageConfiguration(()=>this.onFoldingStrategyChanged())),this.localToDispose.add(this.editor.onDidChangeModelContent(t=>this.onDidChangeModelContent(t))),this.localToDispose.add(this.editor.onDidChangeCursorPosition(()=>this.onCursorPositionChanged())),this.localToDispose.add(this.editor.onMouseDown(t=>this.onEditorMouseDown(t))),this.localToDispose.add(this.editor.onMouseUp(t=>this.onEditorMouseUp(t))),this.localToDispose.add({dispose:()=>{this.foldingRegionPromise&&(this.foldingRegionPromise.cancel(),this.foldingRegionPromise=null),this.updateScheduler?.cancel(),this.updateScheduler=null,this.foldingModel=null,this.foldingModelPromise=null,this.hiddenRangeModel=null,this.cursorChangedScheduler=null,this.rangeProvider?.dispose(),this.rangeProvider=null}}),this.triggerFoldingModelChanged())}onFoldingStrategyChanged(){this.rangeProvider?.dispose(),this.rangeProvider=null,this.triggerFoldingModelChanged()}getRangeProvider(e){if(this.rangeProvider)return this.rangeProvider;const t=new zH(e,this.languageConfigurationService,this._foldingLimitReporter);if(this.rangeProvider=t,this._useFoldingProviders&&this.foldingModel){const i=m1.getFoldingRangeProviders(this.languageFeaturesService,e);i.length>0&&(this.rangeProvider=new $H(e,i,()=>this.triggerFoldingModelChanged(),this._foldingLimitReporter,t))}return this.rangeProvider}getFoldingModel(){return this.foldingModelPromise}onDidChangeModelContent(e){this.hiddenRangeModel?.notifyChangeModelContent(e),this.triggerFoldingModelChanged()}triggerFoldingModelChanged(){this.updateScheduler&&(this.foldingRegionPromise&&(this.foldingRegionPromise.cancel(),this.foldingRegionPromise=null),this.foldingModelPromise=this.updateScheduler.trigger(()=>{const e=this.foldingModel;if(!e)return null;const t=new Cn,i=this.getRangeProvider(e.textModel),n=this.foldingRegionPromise=an(o=>i.compute(o));return n.then(o=>{if(o&&n===this.foldingRegionPromise){let r;if(this._foldingImportsByDefault&&!this._currentModelHasFoldedImports){const c=o.setCollapsedAllOfType(h_.Imports.value,!0);c&&(r=Il.capture(this.editor),this._currentModelHasFoldedImports=c)}const a=this.editor.getSelections();e.update(o,YAe(a)),r?.restore(this.editor);const l=this.updateDebounceInfo.update(e.textModel,t.elapsed());this.updateScheduler&&(this.updateScheduler.defaultDelay=l)}return e})}).then(void 0,e=>(Oe(e),null)))}onHiddenRangesChanges(e){if(this.hiddenRangeModel&&e.length&&!this._restoringViewState){const t=this.editor.getSelections();t&&this.hiddenRangeModel.adjustSelections(t)&&this.editor.setSelections(t)}this.editor.setHiddenAreas(e,this)}onCursorPositionChanged(){this.hiddenRangeModel&&this.hiddenRangeModel.hasRanges()&&this.cursorChangedScheduler.schedule()}revealCursor(){const e=this.getFoldingModel();e&&e.then(t=>{if(t){const i=this.editor.getSelections();if(i&&i.length>0){const n=[];for(const o of i){const r=o.selectionStartLineNumber;this.hiddenRangeModel&&this.hiddenRangeModel.isHidden(r)&&n.push(...t.getAllRegionsAtLine(r,a=>a.isCollapsed&&r>a.startLineNumber))}n.length&&(t.toggleCollapseState(n),this.reveal(i[0].getPosition()))}}}).then(void 0,Oe)}onEditorMouseDown(e){if(this.mouseDownInfo=null,!this.hiddenRangeModel||!e.target||!e.target.range||!e.event.leftButton&&!e.event.middleButton)return;const t=e.target.range;let i=!1;switch(e.target.type){case 4:{const n=e.target.detail,o=e.target.element.offsetLeft;if(n.offsetX-o<4)return;i=!0;break}case 7:{if(this._unfoldOnClickAfterEndOfLine&&this.hiddenRangeModel.hasRanges()&&!e.target.detail.isAfterLines)break;return}case 6:{if(this.hiddenRangeModel.hasRanges()){const n=this.editor.getModel();if(n&&t.startColumn===n.getLineMaxColumn(t.startLineNumber))break}return}default:return}this.mouseDownInfo={lineNumber:t.startLineNumber,iconClicked:i}}onEditorMouseUp(e){const t=this.foldingModel;if(!t||!this.mouseDownInfo||!e.target)return;const i=this.mouseDownInfo.lineNumber,n=this.mouseDownInfo.iconClicked,o=e.target.range;if(!o||o.startLineNumber!==i)return;if(n){if(e.target.type!==4)return}else{const a=this.editor.getModel();if(!a||o.startColumn!==a.getLineMaxColumn(i))return}const r=t.getRegionAtLine(i);if(r&&r.startLineNumber===i){const a=r.isCollapsed;if(n||a){const l=e.event.altKey;let c=[];if(l){const d=u=>!u.containedBy(r)&&!r.containedBy(u),h=t.getRegionsInside(null,d);for(const u of h)u.isCollapsed&&c.push(u);c.length===0&&(c=h)}else{const d=e.event.middleButton||e.event.shiftKey;if(d)for(const h of t.getRegionsInside(r))h.isCollapsed===a&&c.push(h);(a||!d||c.length===0)&&c.push(r)}t.toggleCollapseState(c),this.reveal({lineNumber:i,column:1})}}}reveal(e){this.editor.revealPositionInCenterIfOutsideViewport(e,0)}},m1=Kp,Kp.ID="editor.contrib.folding",Kp);vg=m1=KAe([GC(1,Ae),GC(2,ui),GC(3,Ii),GC(4,ga),GC(5,ge)],vg);class Gse extends F{constructor(e){super(),this.editor=e,this._onDidChange=this._register(new M),this._computed=0,this._limited=!1}get limit(){return this.editor.getOptions().get(56)}update(e,t){(e!==this._computed||t!==this._limited)&&(this._computed=e,this._limited=t,this._onDidChange.fire())}}class rs extends ue{runEditorCommand(e,t,i){const n=e.get(ui),o=vg.get(t);if(!o)return;const r=o.getFoldingModel();if(r)return this.reportTelemetry(e,t),r.then(a=>{if(a){this.invoke(o,a,t,i,n);const l=t.getSelection();l&&o.reveal(l.getStartPosition())}})}getSelectedLines(e){const t=e.getSelections();return t?t.map(i=>i.startLineNumber):[]}getLineNumbers(e,t){return e&&e.selectionLines?e.selectionLines.map(i=>i+1):this.getSelectedLines(t)}run(e,t){}}function YAe(s){return!s||s.length===0?{startsInside:()=>!1}:{startsInside(e,t){for(const i of s){const n=i.startLineNumber;if(n>=e&&n<=t)return!0}return!1}}}function Kse(s){if(!fs(s)){if(!rn(s))return!1;const e=s;if(!fs(e.levels)&&!ch(e.levels)||!fs(e.direction)&&!bn(e.direction)||!fs(e.selectionLines)&&(!Array.isArray(e.selectionLines)||!e.selectionLines.every(ch)))return!1}return!0}class ZAe extends rs{constructor(){super({id:"editor.unfold",label:V(982,"Unfold"),precondition:Hn,kbOpts:{kbExpr:N.editorTextFocus,primary:3166,mac:{primary:2654},weight:100},metadata:{description:"Unfold the content in the editor",args:[{name:"Unfold editor argument",description:`Property-value pairs that can be passed through this argument:
						* 'levels': Number of levels to unfold. If not set, defaults to 1.
						* 'direction': If 'up', unfold given number of levels up otherwise unfolds down.
						* 'selectionLines': Array of the start lines (0-based) of the editor selections to apply the unfold action to. If not set, the active selection(s) will be used.
						`,constraint:Kse,schema:{type:"object",properties:{levels:{type:"number",default:1},direction:{type:"string",enum:["up","down"],default:"down"},selectionLines:{type:"array",items:{type:"number"}}}}}]}})}invoke(e,t,i,n){const o=n&&n.levels||1,r=this.getLineNumbers(n,i);n&&n.direction==="up"?qse(t,!1,o,r):gC(t,!1,o,r)}}class XAe extends rs{constructor(){super({id:"editor.unfoldRecursively",label:V(983,"Unfold Recursively"),precondition:Hn,kbOpts:{kbExpr:N.editorTextFocus,primary:Gi(2089,2142),weight:100}})}invoke(e,t,i,n){gC(t,!1,Number.MAX_VALUE,this.getSelectedLines(i))}}class QAe extends rs{constructor(){super({id:"editor.fold",label:V(984,"Fold"),precondition:Hn,kbOpts:{kbExpr:N.editorTextFocus,primary:3164,mac:{primary:2652},weight:100},metadata:{description:"Fold the content in the editor",args:[{name:"Fold editor argument",description:`Property-value pairs that can be passed through this argument:
							* 'levels': Number of levels to fold.
							* 'direction': If 'up', folds given number of levels up otherwise folds down.
							* 'selectionLines': Array of the start lines (0-based) of the editor selections to apply the fold action to. If not set, the active selection(s) will be used.
							If no levels or direction is set, folds the region at the locations or if already collapsed, the first uncollapsed parent instead.
						`,constraint:Kse,schema:{type:"object",properties:{levels:{type:"number"},direction:{type:"string",enum:["up","down"]},selectionLines:{type:"array",items:{type:"number"}}}}}]}})}invoke(e,t,i,n){const o=this.getLineNumbers(n,i),r=n&&n.levels,a=n&&n.direction;typeof r!="number"&&typeof a!="string"?TAe(t,!0,o):a==="up"?qse(t,!0,r||1,o):gC(t,!0,r||1,o)}}class JAe extends rs{constructor(){super({id:"editor.toggleFold",label:V(985,"Toggle Fold"),precondition:Hn,kbOpts:{kbExpr:N.editorTextFocus,primary:Gi(2089,2090),weight:100}})}invoke(e,t,i){const n=this.getSelectedLines(i);WH(t,1,n)}}class e2e extends rs{constructor(){super({id:"editor.foldRecursively",label:V(986,"Fold Recursively"),precondition:Hn,kbOpts:{kbExpr:N.editorTextFocus,primary:Gi(2089,2140),weight:100}})}invoke(e,t,i){const n=this.getSelectedLines(i);gC(t,!0,Number.MAX_VALUE,n)}}class t2e extends rs{constructor(){super({id:"editor.toggleFoldRecursively",label:V(987,"Toggle Fold Recursively"),precondition:Hn,kbOpts:{kbExpr:N.editorTextFocus,primary:Gi(2089,3114),weight:100}})}invoke(e,t,i){const n=this.getSelectedLines(i);WH(t,Number.MAX_VALUE,n)}}class i2e extends rs{constructor(){super({id:"editor.foldAllBlockComments",label:V(988,"Fold All Block Comments"),precondition:Hn,kbOpts:{kbExpr:N.editorTextFocus,primary:Gi(2089,2138),weight:100}})}invoke(e,t,i,n,o){if(t.regions.hasTypes())VH(t,h_.Comment.value,!0);else{const r=i.getModel();if(!r)return;const a=o.getLanguageConfiguration(r.getLanguageId()).comments;if(a&&a.blockCommentStartToken){const l=new RegExp("^\\s*"+Uo(a.blockCommentStartToken));HH(t,l,!0)}}}}class n2e extends rs{constructor(){super({id:"editor.foldAllMarkerRegions",label:V(989,"Fold All Regions"),precondition:Hn,kbOpts:{kbExpr:N.editorTextFocus,primary:Gi(2089,2077),weight:100}})}invoke(e,t,i,n,o){if(t.regions.hasTypes())VH(t,h_.Region.value,!0);else{const r=i.getModel();if(!r)return;const a=o.getLanguageConfiguration(r.getLanguageId()).foldingRules;if(a&&a.markers&&a.markers.start){const l=new RegExp(a.markers.start);HH(t,l,!0)}}}}class s2e extends rs{constructor(){super({id:"editor.unfoldAllMarkerRegions",label:V(990,"Unfold All Regions"),precondition:Hn,kbOpts:{kbExpr:N.editorTextFocus,primary:Gi(2089,2078),weight:100}})}invoke(e,t,i,n,o){if(t.regions.hasTypes())VH(t,h_.Region.value,!1);else{const r=i.getModel();if(!r)return;const a=o.getLanguageConfiguration(r.getLanguageId()).foldingRules;if(a&&a.markers&&a.markers.start){const l=new RegExp(a.markers.start);HH(t,l,!1)}}}}class o2e extends rs{constructor(){super({id:"editor.foldAllExcept",label:V(991,"Fold All Except Selected"),precondition:Hn,kbOpts:{kbExpr:N.editorTextFocus,primary:Gi(2089,2136),weight:100}})}invoke(e,t,i){const n=this.getSelectedLines(i);jse(t,!0,n)}}class r2e extends rs{constructor(){super({id:"editor.unfoldAllExcept",label:V(992,"Unfold All Except Selected"),precondition:Hn,kbOpts:{kbExpr:N.editorTextFocus,primary:Gi(2089,2134),weight:100}})}invoke(e,t,i){const n=this.getSelectedLines(i);jse(t,!1,n)}}class a2e extends rs{constructor(){super({id:"editor.foldAll",label:V(993,"Fold All"),precondition:Hn,kbOpts:{kbExpr:N.editorTextFocus,primary:Gi(2089,2069),weight:100}})}invoke(e,t,i){gC(t,!0)}}class l2e extends rs{constructor(){super({id:"editor.unfoldAll",label:V(994,"Unfold All"),precondition:Hn,kbOpts:{kbExpr:N.editorTextFocus,primary:Gi(2089,2088),weight:100}})}invoke(e,t,i){gC(t,!1)}}const iw=class iw extends rs{getFoldingLevel(){return parseInt(this.id.substr(iw.ID_PREFIX.length))}invoke(e,t,i){NAe(t,this.getFoldingLevel(),!0,this.getSelectedLines(i))}};iw.ID_PREFIX="editor.foldLevel",iw.ID=e=>iw.ID_PREFIX+e;let wD=iw;class c2e extends rs{constructor(){super({id:"editor.gotoParentFold",label:V(995,"Go to Parent Fold"),precondition:Hn,kbOpts:{kbExpr:N.editorTextFocus,weight:100}})}invoke(e,t,i){const n=this.getSelectedLines(i);if(n.length>0){const o=RAe(n[0],t);o!==null&&i.setSelection({startLineNumber:o,startColumn:1,endLineNumber:o,endColumn:1})}}}class d2e extends rs{constructor(){super({id:"editor.gotoPreviousFold",label:V(996,"Go to Previous Folding Range"),precondition:Hn,kbOpts:{kbExpr:N.editorTextFocus,weight:100}})}invoke(e,t,i){const n=this.getSelectedLines(i);if(n.length>0){const o=AAe(n[0],t);o!==null&&i.setSelection({startLineNumber:o,startColumn:1,endLineNumber:o,endColumn:1})}}}class h2e extends rs{constructor(){super({id:"editor.gotoNextFold",label:V(997,"Go to Next Folding Range"),precondition:Hn,kbOpts:{kbExpr:N.editorTextFocus,weight:100}})}invoke(e,t,i){const n=this.getSelectedLines(i);if(n.length>0){const o=MAe(n[0],t);o!==null&&i.setSelection({startLineNumber:o,startColumn:1,endLineNumber:o,endColumn:1})}}}class u2e extends rs{constructor(){super({id:"editor.createFoldingRangeFromSelection",label:V(998,"Create Folding Range from Selection"),precondition:Hn,kbOpts:{kbExpr:N.editorTextFocus,primary:Gi(2089,2135),weight:100}})}invoke(e,t,i){const n=[],o=i.getSelections();if(o){for(const r of o){let a=r.endLineNumber;r.endColumn===1&&--a,a>r.startLineNumber&&(n.push({startLineNumber:r.startLineNumber,endLineNumber:a,type:void 0,isCollapsed:!0,source:1}),i.setSelection({startLineNumber:r.startLineNumber,startColumn:1,endLineNumber:r.startLineNumber,endColumn:1}))}if(n.length>0){n.sort((a,l)=>a.startLineNumber-l.startLineNumber);const r=pr.sanitizeAndMerge(t.regions,n,i.getModel()?.getLineCount());t.updatePost(pr.fromFoldRanges(r))}}}}class g2e extends rs{constructor(){super({id:"editor.removeManualFoldingRanges",label:V(999,"Remove Manual Folding Ranges"),precondition:Hn,kbOpts:{kbExpr:N.editorTextFocus,primary:Gi(2089,2137),weight:100}})}invoke(e,t,i){const n=i.getSelections();if(n){const o=[];for(const r of n){const{startLineNumber:a,endLineNumber:l}=r;o.push(l>=a?{startLineNumber:a,endLineNumber:l}:{endLineNumber:l,startLineNumber:a})}t.removeManualRanges(o),e.triggerFoldingModelChanged()}}}class f2e extends rs{constructor(){super({id:"editor.toggleImportFold",label:V(1e3,"Toggle Import Fold"),precondition:Hn,kbOpts:{kbExpr:N.editorTextFocus,weight:100}})}async invoke(e,t){const i=[],n=t.regions;for(let o=n.length-1;o>=0;o--)n.getType(o)===h_.Imports.value&&i.push(n.toRegion(o));t.toggleCollapseState(i),e.triggerFoldingModelChanged()}}lt(vg.ID,vg,0);ae(ZAe);ae(XAe);ae(QAe);ae(e2e);ae(t2e);ae(a2e);ae(l2e);ae(i2e);ae(n2e);ae(s2e);ae(o2e);ae(r2e);ae(JAe);ae(c2e);ae(d2e);ae(h2e);ae(u2e);ae(g2e);ae(f2e);for(let s=1;s<=7;s++)age(new wD({id:wD.ID(s),label:V(1001,"Fold Level {0}",s),precondition:Hn,kbOpts:{kbExpr:N.editorTextFocus,primary:Gi(2089,2048|21+s),weight:100}}));at.registerCommand("_executeFoldingRangeProvider",async function(s,...e){const[t]=e;if(!(t instanceof Le))throw Qr();const i=s.get(ge),n=s.get(hi).getModel(t);if(!n)throw Qr();const o=s.get(He);if(!o.getValue("editor.folding",{resource:t}))return[];const r=s.get(ui),a=o.getValue("editor.foldingStrategy",{resource:t}),l={get limit(){return o.getValue("editor.foldingMaximumRegions",{resource:t})},update:(g,f)=>{}},c=new zH(n,r,l);let d=c;if(a!=="indentation"){const g=vg.getFoldingRangeProviders(i,n);g.length&&(d=new $H(n,g,()=>{},l,c))}const h=await d.compute(Ke.None),u=[];try{if(h)for(let g=0;g<h.length;g++){const f=h.getType(g);u.push({start:h.getStartLineNumber(g),end:h.getEndLineNumber(g),kind:f?h_.fromValue(f):void 0})}return u}finally{d.dispose()}});class p2e extends ue{constructor(){super({id:"editor.action.fontZoomIn",label:V(1011,"Increase Editor Font Size"),precondition:void 0})}run(e,t){Gr.setZoomLevel(Gr.getZoomLevel()+1)}}class m2e extends ue{constructor(){super({id:"editor.action.fontZoomOut",label:V(1012,"Decrease Editor Font Size"),precondition:void 0})}run(e,t){Gr.setZoomLevel(Gr.getZoomLevel()-1)}}class _2e extends ue{constructor(){super({id:"editor.action.fontZoomReset",label:V(1013,"Reset Editor Font Size"),precondition:void 0})}run(e,t){Gr.setZoomLevel(0)}}ae(p2e);ae(m2e);ae(_2e);class Tv{static _handleEolEdits(e,t){let i;const n=[];for(const o of t)typeof o.eol=="number"&&(i=o.eol),o.range&&typeof o.text=="string"&&n.push(o);return typeof i=="number"&&e.hasModel()&&e.getModel().pushEOL(i),n}static _isFullModelReplaceEdit(e,t){if(!e.hasModel())return!1;const i=e.getModel(),n=i.validateRange(t.range);return i.getFullModelRange().equalsRange(n)}static execute(e,t,i){i&&e.pushUndoStop();const n=Il.capture(e),o=Tv._handleEolEdits(e,t);o.length===1&&Tv._isFullModelReplaceEdit(e,o[0])?e.executeEdits("formatEditsCommand",o.map(r=>Si.replace(k.lift(r.range),r.text))):e.executeEdits("formatEditsCommand",o.map(r=>Si.replaceMove(k.lift(r.range),r.text))),i&&e.pushUndoStop(),n.restoreRelativeVerticalPositionOfCursor(e)}}class fG{constructor(e){this.value=e,this._lower=e.toLowerCase()}static toKey(e){return typeof e=="string"?e.toLowerCase():e._lower}}class b2e{constructor(e){if(this._set=new Set,e)for(const t of e)this.add(t)}add(e){this._set.add(fG.toKey(e))}has(e){return this._set.has(fG.toKey(e))}}function Yse(s,e,t){const i=[],n=new b2e,o=s.ordered(t);for(const a of o)i.push(a),a.extensionId&&n.add(a.extensionId);const r=e.ordered(t);for(const a of r){if(a.extensionId){if(n.has(a.extensionId))continue;n.add(a.extensionId)}i.push({displayName:a.displayName,extensionId:a.extensionId,provideDocumentFormattingEdits(l,c,d){return a.provideDocumentRangeFormattingEdits(l,l.getFullModelRange(),c,d)}})}return i}const Ny=class Ny{static setFormatterSelector(e){return{dispose:Ny._selectors.unshift(e)}}static async select(e,t,i,n){if(e.length===0)return;const o=ot.first(Ny._selectors);if(o)return await o(e,t,i,n)}};Ny._selectors=new Ns;let i0=Ny;async function Zse(s,e,t,i,n,o,r){const a=s.get(_e),{documentRangeFormattingEditProvider:l}=s.get(ge),c=Tl(e)?e.getModel():e,d=l.ordered(c),h=await i0.select(d,c,i,2);h&&(n.report(h),await a.invokeFunction(w2e,h,e,t,o,r))}async function w2e(s,e,t,i,n,o){const r=s.get(ko),a=s.get(jt),l=s.get(Th);let c,d;Tl(t)?(c=t.getModel(),d=new yh(t,5,void 0,n)):(c=t,d=new Y7(t,n));const h=[];let u=0;for(const b of tW(i).sort(k.compareRangesUsingStarts))u>0&&k.areIntersectingOrTouching(h[u-1],b)?h[u-1]=k.fromPositions(h[u-1].getStartPosition(),b.getEndPosition()):u=h.push(b);const g=async b=>{a.trace("[format][provideDocumentRangeFormattingEdits] (request)",e.extensionId?.value,b);const w=await e.provideDocumentRangeFormattingEdits(c,b,c.getFormattingOptions(),d.token)||[];return a.trace("[format][provideDocumentRangeFormattingEdits] (response)",e.extensionId?.value,w),w},f=(b,w)=>{if(!b.length||!w.length)return!1;const C=b.reduce((v,y)=>k.plusRange(v,y.range),b[0].range);if(!w.some(v=>k.intersectRanges(C,v.range)))return!1;for(const v of b)for(const y of w)if(k.intersectRanges(v.range,y.range))return!0;return!1},p=[],_=[];try{if(typeof e.provideDocumentRangesFormattingEdits=="function"){a.trace("[format][provideDocumentRangeFormattingEdits] (request)",e.extensionId?.value,h);const b=await e.provideDocumentRangesFormattingEdits(c,h,c.getFormattingOptions(),d.token)||[];a.trace("[format][provideDocumentRangeFormattingEdits] (response)",e.extensionId?.value,b),_.push(b)}else{for(const b of h){if(d.token.isCancellationRequested)return!0;_.push(await g(b))}for(let b=0;b<h.length;++b)for(let w=b+1;w<h.length;++w){if(d.token.isCancellationRequested)return!0;if(f(_[b],_[w])){const C=k.plusRange(h[b],h[w]),v=await g(C);h.splice(w,1),h.splice(b,1),h.push(C),_.splice(w,1),_.splice(b,1),_.push(v),b=0,w=0}}}for(const b of _){if(d.token.isCancellationRequested)return!0;const w=await r.computeMoreMinimalEdits(c.uri,b);w&&p.push(...w)}if(d.token.isCancellationRequested)return!0}finally{d.dispose()}if(p.length===0)return!1;if(Tl(t))Tv.execute(t,p,!0),t.revealPositionInCenterIfOutsideViewport(t.getPosition(),1);else{const[{range:b}]=p,w=new he(b.startLineNumber,b.startColumn,b.endLineNumber,b.endColumn);c.pushEditOperations([w],p.map(C=>({text:C.text,range:k.lift(C.range),forceMoveMarkers:!0})),C=>{for(const{range:v}of C)if(k.areIntersectingOrTouching(v,w))return[new he(v.startLineNumber,v.startColumn,v.endLineNumber,v.endColumn)];return null})}return l.playSignal(Gs.format,{userGesture:o}),!0}async function v2e(s,e,t,i,n,o){const r=s.get(_e),a=s.get(ge),l=Tl(e)?e.getModel():e,c=Yse(a.documentFormattingEditProvider,a.documentRangeFormattingEditProvider,l),d=await i0.select(c,l,t,1);d&&(i.report(d),await r.invokeFunction(C2e,d,e,t,n,o))}async function C2e(s,e,t,i,n,o){const r=s.get(ko),a=s.get(Th);let l,c;Tl(t)?(l=t.getModel(),c=new yh(t,5,void 0,n)):(l=t,c=new Y7(t,n));let d;try{const h=await e.provideDocumentFormattingEdits(l,l.getFormattingOptions(),c.token);if(d=await r.computeMoreMinimalEdits(l.uri,h),c.token.isCancellationRequested)return!0}finally{c.dispose()}if(!d||d.length===0)return!1;if(Tl(t))Tv.execute(t,d,i!==2),i!==2&&t.revealPositionInCenterIfOutsideViewport(t.getPosition(),1);else{const[{range:h}]=d,u=new he(h.startLineNumber,h.startColumn,h.endLineNumber,h.endColumn);l.pushEditOperations([u],d.map(g=>({text:g.text,range:k.lift(g.range),forceMoveMarkers:!0})),g=>{for(const{range:f}of g)if(k.areIntersectingOrTouching(f,u))return[new he(f.startLineNumber,f.startColumn,f.endLineNumber,f.endColumn)];return null})}return a.playSignal(Gs.format,{userGesture:o}),!0}async function y2e(s,e,t,i,n,o){const r=e.documentRangeFormattingEditProvider.ordered(t);for(const a of r){const l=await Promise.resolve(a.provideDocumentRangeFormattingEdits(t,i,n,o)).catch(ji);if(Ms(l))return await s.computeMoreMinimalEdits(t.uri,l)}}async function S2e(s,e,t,i,n){const o=Yse(e.documentFormattingEditProvider,e.documentRangeFormattingEditProvider,t);for(const r of o){const a=await Promise.resolve(r.provideDocumentFormattingEdits(t,i,n)).catch(ji);if(Ms(a))return await s.computeMoreMinimalEdits(t.uri,a)}}function Xse(s,e,t,i,n,o,r){const a=e.onTypeFormattingEditProvider.ordered(t);return a.length===0||a[0].autoFormatTriggerCharacters.indexOf(n)<0?Promise.resolve(void 0):Promise.resolve(a[0].provideOnTypeFormattingEdits(t,i,n,o,r)).catch(ji).then(l=>s.computeMoreMinimalEdits(t.uri,l))}at.registerCommand("_executeFormatRangeProvider",async function(s,...e){const[t,i,n]=e;dt(Le.isUri(t)),dt(k.isIRange(i));const o=s.get(Fs),r=s.get(ko),a=s.get(ge),l=await o.createModelReference(t);try{return y2e(r,a,l.object.textEditorModel,k.lift(i),n,Ke.None)}finally{l.dispose()}});at.registerCommand("_executeFormatDocumentProvider",async function(s,...e){const[t,i]=e;dt(Le.isUri(t));const n=s.get(Fs),o=s.get(ko),r=s.get(ge),a=await n.createModelReference(t);try{return S2e(o,r,a.object.textEditorModel,i,Ke.None)}finally{a.dispose()}});at.registerCommand("_executeFormatOnTypeProvider",async function(s,...e){const[t,i,n,o]=e;dt(Le.isUri(t)),dt(A.isIPosition(i)),dt(typeof n=="string");const r=s.get(Fs),a=s.get(ko),l=s.get(ge),c=await r.createModelReference(t);try{return Xse(a,l,c.object.textEditorModel,A.lift(i),n,o,Ke.None)}finally{c.dispose()}});var Qse=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},J1=function(s,e){return function(t,i){e(t,i,s)}},Uw;let vD=(Uw=class{constructor(e,t,i,n){this._editor=e,this._languageFeaturesService=t,this._workerService=i,this._accessibilitySignalService=n,this._disposables=new U,this._sessionDisposables=new U,this._disposables.add(t.onTypeFormattingEditProvider.onDidChange(this._update,this)),this._disposables.add(e.onDidChangeModel(()=>this._update())),this._disposables.add(e.onDidChangeModelLanguage(()=>this._update())),this._disposables.add(e.onDidChangeConfiguration(o=>{o.hasChanged(65)&&this._update()})),this._update()}dispose(){this._disposables.dispose(),this._sessionDisposables.dispose()}_update(){if(this._sessionDisposables.clear(),!this._editor.getOption(65)||!this._editor.hasModel())return;const e=this._editor.getModel(),[t]=this._languageFeaturesService.onTypeFormattingEditProvider.ordered(e);if(!t||!t.autoFormatTriggerCharacters)return;const i=new TE;for(const n of t.autoFormatTriggerCharacters)i.add(n.charCodeAt(0));this._sessionDisposables.add(this._editor.onDidType(n=>{const o=n.charCodeAt(n.length-1);i.has(o)&&this._trigger(String.fromCharCode(o))}))}_trigger(e){if(!this._editor.hasModel()||this._editor.getSelections().length>1||!this._editor.getSelection().isEmpty())return;const t=this._editor.getModel(),i=this._editor.getPosition(),n=new si,o=this._editor.onDidChangeModelContent(r=>{if(r.isFlush){n.cancel(),o.dispose();return}for(let a=0,l=r.changes.length;a<l;a++)if(r.changes[a].range.endLineNumber<=i.lineNumber){n.cancel(),o.dispose();return}});Xse(this._workerService,this._languageFeaturesService,t,i,e,t.getFormattingOptions(),n.token).then(r=>{n.token.isCancellationRequested||Ms(r)&&(this._accessibilitySignalService.playSignal(Gs.format,{userGesture:!1}),Tv.execute(this._editor,r,!0))}).finally(()=>{o.dispose()})}},Uw.ID="editor.contrib.autoFormat",Uw);vD=Qse([J1(1,ge),J1(2,ko),J1(3,Th)],vD);var qw;let CD=(qw=class{constructor(e,t,i){this.editor=e,this._languageFeaturesService=t,this._instantiationService=i,this._callOnDispose=new U,this._callOnModel=new U,this._callOnDispose.add(e.onDidChangeConfiguration(()=>this._update())),this._callOnDispose.add(e.onDidChangeModel(()=>this._update())),this._callOnDispose.add(e.onDidChangeModelLanguage(()=>this._update())),this._callOnDispose.add(t.documentRangeFormattingEditProvider.onDidChange(this._update,this))}dispose(){this._callOnDispose.dispose(),this._callOnModel.dispose()}_update(){this._callOnModel.clear(),this.editor.getOption(64)&&this.editor.hasModel()&&this._languageFeaturesService.documentRangeFormattingEditProvider.has(this.editor.getModel())&&this._callOnModel.add(this.editor.onDidPaste(({range:e})=>this._trigger(e)))}_trigger(e){this.editor.hasModel()&&(this.editor.getSelections().length>1||this._instantiationService.invokeFunction(Zse,this.editor,e,2,ml.None,Ke.None,!1).catch(Oe))}},qw.ID="editor.contrib.formatOnPaste",qw);CD=Qse([J1(1,ge),J1(2,_e)],CD);class x2e extends ue{constructor(){super({id:"editor.action.formatDocument",label:V(1014,"Format Document"),precondition:K.and(N.notInCompositeEditor,N.writable,N.hasDocumentFormattingProvider),kbOpts:{kbExpr:N.editorTextFocus,primary:1572,linux:{primary:3111},weight:100},contextMenuOpts:{group:"1_modification",order:1.3}})}async run(e,t){if(t.hasModel()){const i=e.get(_e);await e.get(Ch).showWhile(i.invokeFunction(v2e,t,1,ml.None,Ke.None,!0),250)}}}class L2e extends ue{constructor(){super({id:"editor.action.formatSelection",label:V(1015,"Format Selection"),precondition:K.and(N.writable,N.hasDocumentSelectionFormattingProvider),kbOpts:{kbExpr:N.editorTextFocus,primary:Gi(2089,2084),weight:100},contextMenuOpts:{when:N.hasNonEmptySelection,group:"1_modification",order:1.31}})}async run(e,t){if(!t.hasModel())return;const i=e.get(_e),n=t.getModel(),o=t.getSelections().map(a=>a.isEmpty()?new k(a.startLineNumber,1,a.startLineNumber,n.getLineMaxColumn(a.startLineNumber)):a);await e.get(Ch).showWhile(i.invokeFunction(Zse,t,o,1,ml.None,Ke.None,!0),250)}}lt(vD.ID,vD,2);lt(CD.ID,CD,2);ae(x2e);ae(L2e);at.registerCommand("editor.action.format",async s=>{const e=s.get(ht).getFocusedCodeEditor();if(!e||!e.hasModel())return;const t=s.get(Gt);e.getSelection().isEmpty()?await t.executeCommand("editor.action.formatDocument"):await t.executeCommand("editor.action.formatSelection")});var k2e=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},dO=function(s,e){return function(t,i){e(t,i,s)}};class Mb{remove(){this.parent?.children.delete(this.id)}static findId(e,t){let i;typeof e=="string"?i=`${t.id}/${e}`:(i=`${t.id}/${e.name}`,t.children.get(i)!==void 0&&(i=`${t.id}/${e.name}_${e.range.startLineNumber}_${e.range.startColumn}`));let n=i;for(let o=0;t.children.get(n)!==void 0;o++)n=`${i}_${o}`;return n}static empty(e){return e.children.size===0}}class o3 extends Mb{constructor(e,t,i){super(),this.id=e,this.parent=t,this.symbol=i,this.children=new Map}}class Jse extends Mb{constructor(e,t,i,n){super(),this.id=e,this.parent=t,this.label=i,this.order=n,this.children=new Map}}class Rd extends Mb{static create(e,t,i){const n=new si(i),o=new Rd(t.uri),r=e.ordered(t),a=r.map((c,d)=>{const h=Mb.findId(`provider_${d}`,o),u=new Jse(h,o,c.displayName??"Unknown Outline Provider",d);return Promise.resolve(c.provideDocumentSymbols(t,n.token)).then(g=>{for(const f of g||[])Rd._makeOutlineElement(f,u);return u},g=>(ji(g),u)).then(g=>{Mb.empty(g)?g.remove():o._groups.set(h,g)})}),l=e.onDidChange(()=>{const c=e.ordered(t);ni(c,r)||n.cancel()});return Promise.all(a).then(()=>n.token.isCancellationRequested&&!i.isCancellationRequested?Rd.create(e,t,i):o._compact()).finally(()=>{n.dispose(),l.dispose(),n.dispose()})}static _makeOutlineElement(e,t){const i=Mb.findId(e,t),n=new o3(i,t,e);if(e.children)for(const o of e.children)Rd._makeOutlineElement(o,n);t.children.set(n.id,n)}constructor(e){super(),this.uri=e,this.id="root",this.parent=void 0,this._groups=new Map,this.children=new Map,this.id="root",this.parent=void 0}_compact(){let e=0;for(const[t,i]of this._groups)i.children.size===0?this._groups.delete(t):e+=1;if(e!==1)this.children=this._groups;else{const t=ot.first(this._groups.values());for(const[,i]of t.children)i.parent=this,this.children.set(i.id,i)}return this}getTopLevelSymbols(){const e=[];for(const t of this.children.values())t instanceof o3?e.push(t.symbol):e.push(...ot.map(t.children.values(),i=>i.symbol));return e.sort((t,i)=>k.compareRangesUsingStarts(t.range,i.range))}asListOfDocumentSymbols(){const e=this.getTopLevelSymbols(),t=[];return Rd._flattenDocumentSymbols(t,e,""),t.sort((i,n)=>A.compare(k.getStartPosition(i.range),k.getStartPosition(n.range))||A.compare(k.getEndPosition(n.range),k.getEndPosition(i.range)))}static _flattenDocumentSymbols(e,t,i){for(const n of t)e.push({kind:n.kind,tags:n.tags,name:n.name,detail:n.detail,containerName:n.containerName||i,range:n.range,selectionRange:n.selectionRange,children:void 0}),n.children&&Rd._flattenDocumentSymbols(e,n.children,n.name)}}const xx=qe("IOutlineModelService");let r3=class{constructor(e,t,i){this._languageFeaturesService=e,this._disposables=new U,this._cache=new Va(15,.7),this._debounceInformation=t.for(e.documentSymbolProvider,"DocumentSymbols",{min:350}),this._disposables.add(i.onModelRemoved(n=>{this._cache.delete(n.id)}))}dispose(){this._disposables.dispose()}async getOrCreate(e,t){const i=this._languageFeaturesService.documentSymbolProvider,n=i.ordered(e);let o=this._cache.get(e.id);if(!o||o.versionId!==e.getVersionId()||!ni(o.provider,n)){const a=new si;o={versionId:e.getVersionId(),provider:n,promiseCnt:0,source:a,promise:Rd.create(i,e,a.token),model:void 0},this._cache.set(e.id,o);const l=Date.now();o.promise.then(c=>{o.model=c,this._debounceInformation.update(e,Date.now()-l)}).catch(c=>{this._cache.delete(e.id)})}if(o.model)return o.model;o.promiseCnt+=1;const r=t.onCancellationRequested(()=>{--o.promiseCnt===0&&(o.source.cancel(),this._cache.delete(e.id))});try{return await o.promise}finally{r.dispose()}}};r3=k2e([dO(0,ge),dO(1,ga),dO(2,hi)],r3);tt(xx,r3,1);at.registerCommand("_executeDocumentSymbolProvider",async function(s,...e){const[t]=e;dt(Le.isUri(t));const i=s.get(xx),o=await s.get(Fs).createModelReference(t);try{return(await i.getOrCreate(o.object.textEditorModel,Ke.None)).getTopLevelSymbols()}finally{o.dispose()}});const UA=new class{constructor(){this._implementations=[]}register(e){return this._implementations.push(e),{dispose:()=>{const t=this._implementations.indexOf(e);t!==-1&&this._implementations.splice(t,1)}}}getImplementations(){return this._implementations}};var E2e=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},I2e=function(s,e){return function(t,i){e(t,i,s)}};function D2e(s){return s instanceof eoe?s._instance:s}let eoe=class{constructor(e){this.instantiationService=e}init(...e){}};function T2e(s){return s.get()}let pG=class extends eoe{constructor(e,t){super(t),this.init(e)}};pG=E2e([I2e(1,_e)],pG);class Lx{validateOrThrow(e){const t=this.validate(e);if(t.error)throw new Error(t.error.message);return t.content}}class qA extends Lx{constructor(e){super(),this.type=e}validate(e){return typeof e!==this.type?{content:void 0,error:{message:`Expected ${this.type}, but got ${typeof e}`}}:{content:e,error:void 0}}getJSONSchema(){return{type:this.type}}}const N2e=new qA("string");function R2e(){return N2e}new qA("number");const A2e=new qA("boolean");function mG(){return A2e}new qA("object");class toe extends Lx{validate(e){return e!==void 0?{content:void 0,error:{message:`Expected undefined, but got ${typeof e}`}}:{content:void 0,error:void 0}}getJSONSchema(){return{}}}function M2e(){return new toe}class a3{constructor(e){this.validator=e}}function hO(s){return new a3(s)}class O2e extends Lx{constructor(e){super(),this.properties=e}validate(e){if(typeof e!="object"||e===null)return{content:void 0,error:{message:"Expected object"}};const t={};for(const i in this.properties){const n=this.properties[i],o=e[i],r=n instanceof a3,a=r?n.validator:n;if(r&&o===void 0)continue;const{content:l,error:c}=a.validate(o);if(c)return{content:void 0,error:{message:`Error in property '${i}': ${c.message}`}};t[i]=l}return{content:t,error:void 0}}getJSONSchema(){const e=[],t={};for(const[n,o]of Object.entries(this.properties)){const r=o instanceof a3,a=r?o.validator:o;t[n]=a.getJSONSchema(),r||e.push(n)}return{type:"object",properties:t,...e.length>0?{required:e}:{}}}}function P2e(s){return new O2e(s)}class F2e extends Lx{constructor(e){super(),this.validators=e}validate(e){let t;for(const i of this.validators){const{content:n,error:o}=i.validate(e);if(!o)return{content:n,error:void 0};t=o}return{content:void 0,error:t}}getJSONSchema(){return{oneOf:khe(this.validators,e=>{if(!(e instanceof toe))return e.getJSONSchema()})}}}function B2e(...s){return new F2e(s)}class W2e extends Lx{constructor(e,t){super(),this._ref=e,this._validator=t}validate(e){return this._validator.validate(e)}getJSONSchema(){return{$ref:this._ref}}}function H2e(s,e){return new W2e(s,e)}const je={Visible:cH,HasFocusedSuggestion:new ce("suggestWidgetHasFocusedSuggestion",!1,m(1455,"Whether any suggestion is focused")),DetailsVisible:new ce("suggestWidgetDetailsVisible",!1,m(1456,"Whether suggestion details are visible")),MultipleSuggestions:new ce("suggestWidgetMultipleSuggestions",!1,m(1457,"Whether there are multiple suggestions to pick from")),MakesTextEdit:new ce("suggestionMakesTextEdit",!0,m(1458,"Whether inserting the current suggestion yields in a change or has everything already been typed")),AcceptSuggestionsOnEnter:new ce("acceptSuggestionOnEnter",!0,m(1459,"Whether suggestions are inserted when pressing Enter")),HasInsertAndReplaceRange:new ce("suggestionHasInsertAndReplaceRange",!1,m(1460,"Whether the current suggestion has insert and replace behaviour")),InsertMode:new ce("suggestionInsertMode",void 0,{type:"string",description:m(1461,"Whether the default behaviour is to insert or replace")}),CanResolve:new ce("suggestionCanResolve",!1,m(1462,"Whether the current suggestion supports to resolve further details"))},Hu=new fe("suggestWidgetStatusBar");let V2e=class{constructor(e,t,i,n){this.position=e,this.completion=t,this.container=i,this.provider=n,this.isInvalid=!1,this.score=Oa.Default,this.distance=0,this.textLabel=typeof t.label=="string"?t.label:t.label?.label,this.labelLow=this.textLabel.toLowerCase(),this.isInvalid=!this.textLabel,this.sortTextLow=t.sortText&&t.sortText.toLowerCase(),this.filterTextLow=t.filterText&&t.filterText.toLowerCase(),this.extensionId=t.extensionId,k.isIRange(t.range)?(this.editStart=new A(t.range.startLineNumber,t.range.startColumn),this.editInsertEnd=new A(t.range.endLineNumber,t.range.endColumn),this.editReplaceEnd=new A(t.range.endLineNumber,t.range.endColumn),this.isInvalid=this.isInvalid||k.spansMultipleLines(t.range)||t.range.startLineNumber!==e.lineNumber):(this.editStart=new A(t.range.insert.startLineNumber,t.range.insert.startColumn),this.editInsertEnd=new A(t.range.insert.endLineNumber,t.range.insert.endColumn),this.editReplaceEnd=new A(t.range.replace.endLineNumber,t.range.replace.endColumn),this.isInvalid=this.isInvalid||k.spansMultipleLines(t.range.insert)||k.spansMultipleLines(t.range.replace)||t.range.insert.startLineNumber!==e.lineNumber||t.range.replace.startLineNumber!==e.lineNumber||t.range.insert.startColumn!==t.range.replace.startColumn),typeof n.resolveCompletionItem!="function"&&(this._resolveCache=Promise.resolve(),this._resolveDuration=0)}get isResolved(){return this._resolveDuration!==void 0}get resolveDuration(){return this._resolveDuration!==void 0?this._resolveDuration:-1}async resolve(e){if(!this._resolveCache){const t=e.onCancellationRequested(()=>{this._resolveCache=void 0,this._resolveDuration=void 0}),i=new Cn(!0);this._resolveCache=Promise.resolve(this.provider.resolveCompletionItem(this.completion,e)).then(n=>{Object.assign(this.completion,n),this._resolveDuration=i.elapsed()},n=>{Dr(n)&&(this._resolveCache=void 0,this._resolveDuration=void 0)}).finally(()=>{t.dispose()})}return this._resolveCache}};const EN=class EN{constructor(e=2,t=new Set,i=new Set,n=new Map,o=!0){this.snippetSortOrder=e,this.kindFilter=t,this.providerFilter=i,this.providerItemsToReuse=n,this.showDeprecated=o}};EN.default=new EN;let n0=EN;class z2e{constructor(e,t,i,n){this.items=e,this.needsClipboard=t,this.durations=i,this.disposable=n}}async function UH(s,e,t,i=n0.default,n={triggerKind:0},o=Ke.None){const r=new Cn;t=t.clone();const a=e.getWordAtPosition(t),l=a?new k(t.lineNumber,a.startColumn,t.lineNumber,a.endColumn):k.fromPositions(t),c={replace:l,insert:l.setEndPosition(t.lineNumber,t.column)},d=[],h=new U,u=[];let g=!1;const f=(_,b,w)=>{let C=!1;if(!b)return C;for(const v of b.suggestions)if(!i.kindFilter.has(v.kind)){if(!i.showDeprecated&&v?.tags?.includes(1))continue;v.range||(v.range=c),v.sortText||(v.sortText=typeof v.label=="string"?v.label:v.label.label),!g&&v.insertTextRules&&v.insertTextRules&4&&(g=Hm.guessNeedsClipboard(v.insertText)),d.push(new V2e(t,v,b,_)),C=!0}return i_(b)&&h.add(b),u.push({providerName:_._debugDisplayName??"unknown_provider",elapsedProvider:b.duration??-1,elapsedOverall:w.elapsed()}),C},p=(async()=>{})();for(const _ of s.orderedGroups(e)){let b=!1;if(await Promise.all(_.map(async w=>{if(i.providerItemsToReuse.has(w)){const C=i.providerItemsToReuse.get(w);C.forEach(v=>d.push(v)),b=b||C.length>0;return}if(!(i.providerFilter.size>0&&!i.providerFilter.has(w)))try{const C=new Cn,v=await w.provideCompletionItems(e,t,n,o);b=f(w,v,C)||b}catch(C){ji(C)}})),b||o.isCancellationRequested)break}return await p,o.isCancellationRequested?(h.dispose(),Promise.reject(new ia)):new z2e(d.sort(q2e(i.snippetSortOrder)),g,{entries:u,elapsed:r.elapsed()},h)}function qH(s,e){if(s.sortTextLow&&e.sortTextLow){if(s.sortTextLow<e.sortTextLow)return-1;if(s.sortTextLow>e.sortTextLow)return 1}return s.textLabel<e.textLabel?-1:s.textLabel>e.textLabel?1:s.completion.kind-e.completion.kind}function $2e(s,e){if(s.completion.kind!==e.completion.kind){if(s.completion.kind===28)return-1;if(e.completion.kind===28)return 1}return qH(s,e)}function U2e(s,e){if(s.completion.kind!==e.completion.kind){if(s.completion.kind===28)return 1;if(e.completion.kind===28)return-1}return qH(s,e)}const jA=new Map;jA.set(0,$2e);jA.set(2,U2e);jA.set(1,qH);function q2e(s){return jA.get(s)}at.registerCommand("_executeCompletionItemProvider",async(s,...e)=>{const[t,i,n,o]=e;dt(Le.isUri(t)),dt(A.isIPosition(i)),dt(typeof n=="string"||!n),dt(typeof o=="number"||!o);const{completionProvider:r}=s.get(ge),a=await s.get(Fs).createModelReference(t);try{const l={incomplete:!1,suggestions:[]},c=[],d=a.object.textEditorModel.validatePosition(i),h=await UH(r,a.object.textEditorModel,d,void 0,{triggerCharacter:n??void 0,triggerKind:n?1:0});for(const u of h.items)c.length<(o??0)&&c.push(u.resolve(Ke.None)),l.incomplete=l.incomplete||u.container.incomplete,l.suggestions.push(u.completion);try{return await Promise.all(c),l}finally{setTimeout(()=>h.disposable.dispose(),100)}}finally{a.dispose()}});function j2e(s,e){s.getContribution("editor.contrib.suggestController")?.triggerSuggest(new Set().add(e),void 0,!0)}class Ob{static isAllOff(e){return e.other==="off"&&e.comments==="off"&&e.strings==="off"}static isAllOn(e){return e.other==="on"&&e.comments==="on"&&e.strings==="on"}static valueFor(e,t){switch(t){case 1:return e.comments;case 2:return e.strings;default:return e.other}}}const rr=class rr{};rr.inlineSuggestionVisible=new ce("inlineSuggestionVisible",!1,m(1192,"Whether an inline suggestion is visible")),rr.inlineSuggestionHasIndentation=new ce("inlineSuggestionHasIndentation",!1,m(1193,"Whether the inline suggestion starts with whitespace")),rr.inlineSuggestionHasIndentationLessThanTabSize=new ce("inlineSuggestionHasIndentationLessThanTabSize",!0,m(1194,"Whether the inline suggestion starts with whitespace that is less than what would be inserted by tab")),rr.suppressSuggestions=new ce("inlineSuggestionSuppressSuggestions",void 0,m(1195,"Whether suggestions should be suppressed for the current suggestion")),rr.cursorBeforeGhostText=new ce("cursorBeforeGhostText",!1,m(1196,"Whether the cursor is at ghost text")),rr.cursorInIndentation=new ce("cursorInIndentation",!1,m(1197,"Whether the cursor is in indentation")),rr.hasSelection=new ce("editor.hasSelection",!1,m(1198,"Whether the editor has a selection")),rr.cursorAtInlineEdit=new ce("cursorAtInlineEdit",!1,m(1199,"Whether the cursor is at an inline edit")),rr.inlineEditVisible=new ce("inlineEditIsVisible",!1,m(1200,"Whether an inline edit is visible")),rr.tabShouldJumpToInlineEdit=new ce("tabShouldJumpToInlineEdit",!1,m(1201,"Whether tab should jump to an inline edit.")),rr.tabShouldAcceptInlineEdit=new ce("tabShouldAcceptInlineEdit",!1,m(1202,"Whether tab should accept the inline edit.")),rr.inInlineEditsPreviewEditor=new ce("inInlineEditsPreviewEditor",!0,m(1203,"Whether the current code editor is showing an inline edits preview"));let Ot=rr;class G2e{constructor(e,t,i){this._selection=e,this._cursors=t,this._selectionId=null,this._trimInRegexesAndStrings=i}getEditOperations(e,t){const i=K2e(e,this._cursors,this._trimInRegexesAndStrings);for(let n=0,o=i.length;n<o;n++){const r=i[n];t.addEditOperation(r.range,r.text)}this._selectionId=t.trackSelection(this._selection)}computeCursorState(e,t){return t.getTrackedSelection(this._selectionId)}}function K2e(s,e,t){e.sort((a,l)=>a.lineNumber===l.lineNumber?a.column-l.column:a.lineNumber-l.lineNumber);for(let a=e.length-2;a>=0;a--)e[a].lineNumber===e[a+1].lineNumber&&e.splice(a,1);const i=[];let n=0,o=0;const r=e.length;for(let a=1,l=s.getLineCount();a<=l;a++){const c=s.getLineContent(a),d=c.length+1;let h=0;if(o<r&&e[o].lineNumber===a&&(h=e[o].column,o++,h===d)||c.length===0)continue;const u=Ma(c);let g=0;if(u===-1)g=1;else if(u!==c.length-1)g=u+2;else continue;if(!t){if(!s.tokenization.hasAccurateTokensForLine(a))continue;const f=s.tokenization.getLineTokens(a),p=f.getStandardTokenType(f.findTokenIndexAtOffset(g));if(p===2||p===3)continue}g=Math.max(h,g),i[n++]=Si.delete(new k(a,g,a,d))}return i}class ioe{constructor(e,t,i){this._selection=e,this._isCopyingDown=t,this._noop=i||!1,this._selectionDirection=0,this._selectionId=null,this._startLineNumberDelta=0,this._endLineNumberDelta=0}getEditOperations(e,t){let i=this._selection;this._startLineNumberDelta=0,this._endLineNumberDelta=0,i.startLineNumber<i.endLineNumber&&i.endColumn===1&&(this._endLineNumberDelta=1,i=i.setEndPosition(i.endLineNumber-1,e.getLineMaxColumn(i.endLineNumber-1)));const n=[];for(let r=i.startLineNumber;r<=i.endLineNumber;r++)n.push(e.getLineContent(r));const o=n.join(`
`);o===""&&this._isCopyingDown&&(this._startLineNumberDelta++,this._endLineNumberDelta++),this._noop?t.addEditOperation(new k(i.endLineNumber,e.getLineMaxColumn(i.endLineNumber),i.endLineNumber+1,1),i.endLineNumber===e.getLineCount()?"":`
`):this._isCopyingDown?t.addEditOperation(new k(i.startLineNumber,1,i.startLineNumber,1),o+`
`):t.addEditOperation(new k(i.endLineNumber,e.getLineMaxColumn(i.endLineNumber),i.endLineNumber,e.getLineMaxColumn(i.endLineNumber)),`
`+o),this._selectionId=t.trackSelection(i),this._selectionDirection=this._selection.getDirection()}computeCursorState(e,t){let i=t.getTrackedSelection(this._selectionId);if(this._startLineNumberDelta!==0||this._endLineNumberDelta!==0){let n=i.startLineNumber,o=i.startColumn,r=i.endLineNumber,a=i.endColumn;this._startLineNumberDelta!==0&&(n=n+this._startLineNumberDelta,o=1),this._endLineNumberDelta!==0&&(r=r+this._endLineNumberDelta,a=1),i=he.createWithDirection(n,o,r,a,this._selectionDirection)}return i}}function To(s,e){let t=0;for(let i=0;i<s.length;i++)s.charAt(i)==="	"?t+=e:t++;return t}function ey(s,e,t){s=s<0?0:s;let i="";if(!t){const n=Math.floor(s/e);s=s%e;for(let o=0;o<n;o++)i+="	"}for(let n=0;n<s;n++)i+=" ";return i}var Y2e=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},Z2e=function(s,e){return function(t,i){e(t,i,s)}};let l3=class{constructor(e,t,i,n){this._languageConfigurationService=n,this._selection=e,this._isMovingDown=t,this._autoIndent=i,this._selectionId=null,this._moveEndLineSelectionShrink=!1}getEditOperations(e,t){const i=()=>e.getLanguageId(),n=(h,u)=>e.getLanguageIdAtPosition(h,u),o=e.getLineCount();if(this._isMovingDown&&this._selection.endLineNumber===o){this._selectionId=t.trackSelection(this._selection);return}if(!this._isMovingDown&&this._selection.startLineNumber===1){this._selectionId=t.trackSelection(this._selection);return}this._moveEndPositionDown=!1;let r=this._selection;r.startLineNumber<r.endLineNumber&&r.endColumn===1&&(this._moveEndPositionDown=!0,r=r.setEndPosition(r.endLineNumber-1,e.getLineMaxColumn(r.endLineNumber-1)));const{tabSize:a,indentSize:l,insertSpaces:c}=e.getOptions(),d=this.buildIndentConverter(a,l,c);if(r.startLineNumber===r.endLineNumber&&e.getLineMaxColumn(r.startLineNumber)===1){const h=r.startLineNumber,u=this._isMovingDown?h+1:h-1;e.getLineMaxColumn(u)===1?t.addEditOperation(new k(1,1,1,1),null):(t.addEditOperation(new k(h,1,h,1),e.getLineContent(u)),t.addEditOperation(new k(u,1,u,e.getLineMaxColumn(u)),null)),r=new he(u,1,u,1)}else{let h,u;if(this._isMovingDown){h=r.endLineNumber+1,u=e.getLineContent(h),t.addEditOperation(new k(h-1,e.getLineMaxColumn(h-1),h,e.getLineMaxColumn(h)),null);let g=u;if(this.shouldAutoIndent(e,r)){const f=this.matchEnterRule(e,d,a,h,r.startLineNumber-1);if(f!==null){const _=Bt(e.getLineContent(h)),b=f+To(_,a);g=ey(b,a,c)+this.trimStart(u)}else{const _={tokenization:{getLineTokens:w=>w===r.startLineNumber?e.tokenization.getLineTokens(h):e.tokenization.getLineTokens(w),getLanguageId:i,getLanguageIdAtPosition:n},getLineContent:w=>w===r.startLineNumber?e.getLineContent(h):e.getLineContent(w)},b=D1(this._autoIndent,_,e.getLanguageIdAtPosition(h,1),r.startLineNumber,d,this._languageConfigurationService);if(b!==null){const w=Bt(e.getLineContent(h)),C=To(b,a),v=To(w,a);C!==v&&(g=ey(C,a,c)+this.trimStart(u))}}t.addEditOperation(new k(r.startLineNumber,1,r.startLineNumber,1),g+`
`);const p=this.matchEnterRuleMovingDown(e,d,a,r.startLineNumber,h,g);if(p!==null)p!==0&&this.getIndentEditsOfMovingBlock(e,t,r,a,c,p);else{const _={tokenization:{getLineTokens:w=>w===r.startLineNumber?e.tokenization.getLineTokens(h):w>=r.startLineNumber+1&&w<=r.endLineNumber+1?e.tokenization.getLineTokens(w-1):e.tokenization.getLineTokens(w),getLanguageId:i,getLanguageIdAtPosition:n},getLineContent:w=>w===r.startLineNumber?g:w>=r.startLineNumber+1&&w<=r.endLineNumber+1?e.getLineContent(w-1):e.getLineContent(w)},b=D1(this._autoIndent,_,e.getLanguageIdAtPosition(h,1),r.startLineNumber+1,d,this._languageConfigurationService);if(b!==null){const w=Bt(e.getLineContent(r.startLineNumber)),C=To(b,a),v=To(w,a);if(C!==v){const y=C-v;this.getIndentEditsOfMovingBlock(e,t,r,a,c,y)}}}}else t.addEditOperation(new k(r.startLineNumber,1,r.startLineNumber,1),g+`
`)}else if(h=r.startLineNumber-1,u=e.getLineContent(h),t.addEditOperation(new k(h,1,h+1,1),null),t.addEditOperation(new k(r.endLineNumber,e.getLineMaxColumn(r.endLineNumber),r.endLineNumber,e.getLineMaxColumn(r.endLineNumber)),`
`+u),this.shouldAutoIndent(e,r)){const g={tokenization:{getLineTokens:p=>p===h?e.tokenization.getLineTokens(r.startLineNumber):e.tokenization.getLineTokens(p),getLanguageId:i,getLanguageIdAtPosition:n},getLineContent:p=>p===h?e.getLineContent(r.startLineNumber):e.getLineContent(p)},f=this.matchEnterRule(e,d,a,r.startLineNumber,r.startLineNumber-2);if(f!==null)f!==0&&this.getIndentEditsOfMovingBlock(e,t,r,a,c,f);else{const p=D1(this._autoIndent,g,e.getLanguageIdAtPosition(r.startLineNumber,1),h,d,this._languageConfigurationService);if(p!==null){const _=Bt(e.getLineContent(r.startLineNumber)),b=To(p,a),w=To(_,a);if(b!==w){const C=b-w;this.getIndentEditsOfMovingBlock(e,t,r,a,c,C)}}}}}this._selectionId=t.trackSelection(r)}buildIndentConverter(e,t,i){return{shiftIndent:n=>la.shiftIndent(n,n.length+1,e,t,i),unshiftIndent:n=>la.unshiftIndent(n,n.length+1,e,t,i)}}parseEnterResult(e,t,i,n,o){if(o){let r=o.indentation;o.indentAction===Xi.None||o.indentAction===Xi.Indent?r=o.indentation+o.appendText:o.indentAction===Xi.IndentOutdent?r=o.indentation:o.indentAction===Xi.Outdent&&(r=t.unshiftIndent(o.indentation)+o.appendText);const a=e.getLineContent(n);if(this.trimStart(a).indexOf(this.trimStart(r))>=0){const l=Bt(e.getLineContent(n));let c=Bt(r);const d=_J(e,n,this._languageConfigurationService);d!==null&&d&2&&(c=t.unshiftIndent(c));const h=To(c,i),u=To(l,i);return h-u}}return null}matchEnterRuleMovingDown(e,t,i,n,o,r){if(Ma(r)>=0){const a=e.getLineMaxColumn(o),l=dw(this._autoIndent,e,new k(o,a,o,a),this._languageConfigurationService);return this.parseEnterResult(e,t,i,n,l)}else{let a=n-1;for(;a>=1;){const d=e.getLineContent(a);if(Ma(d)>=0)break;a--}if(a<1||n>e.getLineCount())return null;const l=e.getLineMaxColumn(a),c=dw(this._autoIndent,e,new k(a,l,a,l),this._languageConfigurationService);return this.parseEnterResult(e,t,i,n,c)}}matchEnterRule(e,t,i,n,o,r){let a=o;for(;a>=1;){let d;if(a===o&&r!==void 0?d=r:d=e.getLineContent(a),Ma(d)>=0)break;a--}if(a<1||n>e.getLineCount())return null;const l=e.getLineMaxColumn(a),c=dw(this._autoIndent,e,new k(a,l,a,l),this._languageConfigurationService);return this.parseEnterResult(e,t,i,n,c)}trimStart(e){return e.replace(/^\s+/,"")}shouldAutoIndent(e,t){if(this._autoIndent<4||!e.tokenization.isCheapToTokenize(t.startLineNumber))return!1;const i=e.getLanguageIdAtPosition(t.startLineNumber,1),n=e.getLanguageIdAtPosition(t.endLineNumber,1);return!(i!==n||this._languageConfigurationService.getLanguageConfiguration(i).indentRulesSupport===null)}getIndentEditsOfMovingBlock(e,t,i,n,o,r){for(let a=i.startLineNumber;a<=i.endLineNumber;a++){const l=e.getLineContent(a),c=Bt(l),h=To(c,n)+r,u=ey(h,n,o);u!==c&&(t.addEditOperation(new k(a,1,a,c.length+1),u),a===i.endLineNumber&&i.endColumn<=c.length+1&&u===""&&(this._moveEndLineSelectionShrink=!0))}}computeCursorState(e,t){let i=t.getTrackedSelection(this._selectionId);return this._moveEndPositionDown&&(i=i.setEndPosition(i.endLineNumber+1,1)),this._moveEndLineSelectionShrink&&i.startLineNumber<i.endLineNumber&&(i=i.setEndPosition(i.endLineNumber,2)),i}};l3=Y2e([Z2e(3,ui)],l3);const $V=class $V{constructor(e,t){this.selection=e,this.descending=t,this.selectionId=null}getEditOperations(e,t){const i=X2e(e,this.selection,this.descending);i&&t.addEditOperation(i.range,i.text),this.selectionId=t.trackSelection(this.selection)}computeCursorState(e,t){return t.getTrackedSelection(this.selectionId)}static canRun(e,t,i){if(e===null)return!1;const n=noe(e,t,i);if(!n)return!1;for(let o=0,r=n.before.length;o<r;o++)if(n.before[o]!==n.after[o])return!0;return!1}};$V._COLLATOR=a_.Collator();let s0=$V;function noe(s,e,t){const i=e.startLineNumber;let n=e.endLineNumber;if(e.endColumn===1&&n--,i>=n)return null;const o=[];for(let a=i;a<=n;a++)o.push(s.getLineContent(a));let r=o.slice(0);return r.sort(s0._COLLATOR.value.compare),t===!0&&(r=r.reverse()),{startLineNumber:i,endLineNumber:n,before:o,after:r}}function X2e(s,e,t){const i=noe(s,e,t);return i?Si.replace(new k(i.startLineNumber,1,i.endLineNumber,s.getLineMaxColumn(i.endLineNumber)),i.after.join(`
`)):null}class soe extends ue{constructor(e,t){super(t),this.down=e}run(e,t){if(!t.hasModel())return;const i=t.getSelections().map((r,a)=>({selection:r,index:a,ignore:!1}));i.sort((r,a)=>k.compareRangesUsingStarts(r.selection,a.selection));let n=i[0];for(let r=1;r<i.length;r++){const a=i[r];n.selection.endLineNumber===a.selection.startLineNumber&&(n.index<a.index?a.ignore=!0:(n.ignore=!0,n=a))}const o=[];for(const r of i)o.push(new ioe(r.selection,this.down,r.ignore));t.pushUndoStop(),t.executeCommands(this.id,o),t.pushUndoStop()}}class Q2e extends soe{constructor(){super(!1,{id:"editor.action.copyLinesUpAction",label:V(1249,"Copy Line Up"),precondition:N.writable,kbOpts:{kbExpr:N.editorTextFocus,primary:1552,linux:{primary:3600},weight:100},menuOpts:{menuId:fe.MenubarSelectionMenu,group:"2_line",title:m(1244,"&&Copy Line Up"),order:1},canTriggerInlineEdits:!0})}}class J2e extends soe{constructor(){super(!0,{id:"editor.action.copyLinesDownAction",label:V(1250,"Copy Line Down"),precondition:N.writable,kbOpts:{kbExpr:N.editorTextFocus,primary:1554,linux:{primary:3602},weight:100},menuOpts:{menuId:fe.MenubarSelectionMenu,group:"2_line",title:m(1245,"Co&&py Line Down"),order:2},canTriggerInlineEdits:!0})}}class eMe extends ue{constructor(){super({id:"editor.action.duplicateSelection",label:V(1251,"Duplicate Selection"),precondition:N.writable,menuOpts:{menuId:fe.MenubarSelectionMenu,group:"2_line",title:m(1246,"&&Duplicate Selection"),order:5},canTriggerInlineEdits:!0})}run(e,t,i){if(!t.hasModel())return;const n=[],o=t.getSelections(),r=t.getModel();for(const a of o)if(a.isEmpty())n.push(new ioe(a,!0));else{const l=new he(a.endLineNumber,a.endColumn,a.endLineNumber,a.endColumn);n.push(new bge(l,r.getValueInRange(a)))}t.pushUndoStop(),t.executeCommands(this.id,n),t.pushUndoStop()}}class ooe extends ue{constructor(e,t){super(t),this.down=e}run(e,t){const i=e.get(ui),n=[],o=t.getSelections()||[],r=t.getOption(16);for(const a of o)n.push(new l3(a,this.down,r,i));t.pushUndoStop(),t.executeCommands(this.id,n),t.pushUndoStop()}}class tMe extends ooe{constructor(){super(!1,{id:"editor.action.moveLinesUpAction",label:V(1252,"Move Line Up"),precondition:N.writable,kbOpts:{kbExpr:N.editorTextFocus,primary:528,linux:{primary:528},weight:100},menuOpts:{menuId:fe.MenubarSelectionMenu,group:"2_line",title:m(1247,"Mo&&ve Line Up"),order:3},canTriggerInlineEdits:!0})}}class iMe extends ooe{constructor(){super(!0,{id:"editor.action.moveLinesDownAction",label:V(1253,"Move Line Down"),precondition:N.writable,kbOpts:{kbExpr:N.editorTextFocus,primary:530,linux:{primary:530},weight:100},menuOpts:{menuId:fe.MenubarSelectionMenu,group:"2_line",title:m(1248,"Move &&Line Down"),order:4},canTriggerInlineEdits:!0})}}class roe extends ue{constructor(e,t){super(t),this.descending=e}run(e,t){if(!t.hasModel())return;const i=t.getModel();let n=t.getSelections();n.length===1&&n[0].isEmpty()&&(n=[new he(1,1,i.getLineCount(),i.getLineMaxColumn(i.getLineCount()))]);for(const r of n)if(!s0.canRun(t.getModel(),r,this.descending))return;const o=[];for(let r=0,a=n.length;r<a;r++)o[r]=new s0(n[r],this.descending);t.pushUndoStop(),t.executeCommands(this.id,o),t.pushUndoStop()}}class nMe extends roe{constructor(){super(!1,{id:"editor.action.sortLinesAscending",label:V(1254,"Sort Lines Ascending"),precondition:N.writable,canTriggerInlineEdits:!0})}}class sMe extends roe{constructor(){super(!0,{id:"editor.action.sortLinesDescending",label:V(1255,"Sort Lines Descending"),precondition:N.writable,canTriggerInlineEdits:!0})}}class oMe extends ue{constructor(){super({id:"editor.action.removeDuplicateLines",label:V(1256,"Delete Duplicate Lines"),precondition:N.writable,canTriggerInlineEdits:!0})}run(e,t){if(!t.hasModel())return;const i=t.getModel();if(i.getLineCount()===1&&i.getLineMaxColumn(1)===1)return;const n=[],o=[];let r=0,a=!0,l=t.getSelections();l.length===1&&l[0].isEmpty()&&(l=[new he(1,1,i.getLineCount(),i.getLineMaxColumn(i.getLineCount()))],a=!1);for(const c of l){const d=new Set,h=[];for(let p=c.startLineNumber;p<=c.endLineNumber;p++){const _=i.getLineContent(p);d.has(_)||(h.push(_),d.add(_))}const u=new he(c.startLineNumber,1,c.endLineNumber,i.getLineMaxColumn(c.endLineNumber)),g=c.startLineNumber-r,f=new he(g,1,g+h.length-1,h[h.length-1].length);n.push(Si.replace(u,h.join(`
`))),o.push(f),r+=c.endLineNumber-c.startLineNumber+1-h.length}t.pushUndoStop(),t.executeEdits(this.id,n,a?o:void 0),t.pushUndoStop()}}class rMe extends ue{constructor(){super({id:"editor.action.reverseLines",label:V(1257,"Reverse lines"),precondition:N.writable,canTriggerInlineEdits:!0})}run(e,t){if(!t.hasModel())return;const i=t.getModel(),n=t.getSelections();let o=n;o.length===1&&o[0].isEmpty()&&(o=[new he(1,1,i.getLineCount(),i.getLineMaxColumn(i.getLineCount()))]);const r=[],a=[];for(let l=0;l<o.length;l++){const c=o[l],d=n[l];let h=c.endLineNumber;c.startLineNumber<c.endLineNumber&&c.endColumn===1&&h--;let u=new k(c.startLineNumber,1,h,i.getLineMaxColumn(h));h===i.getLineCount()&&i.getLineContent(u.endLineNumber)===""&&(u=u.setEndPosition(u.endLineNumber-1,i.getLineMaxColumn(u.endLineNumber-1)));const g=[];for(let b=u.endLineNumber;b>=u.startLineNumber;b--)g.push(i.getLineContent(b));const f=Si.replace(u,g.join(`
`));r.push(f);const p=function(b){return b<=u.endLineNumber?u.endLineNumber-b+u.startLineNumber:b},_=function(b){if(b.isEmpty())return new he(p(b.positionLineNumber),b.positionColumn,p(b.positionLineNumber),b.positionColumn);{const w=p(b.selectionStartLineNumber),C=p(b.positionLineNumber),v=b.selectionStartColumn,y=b.positionColumn;return new he(w,v,C,y)}};a.push(_(d))}t.pushUndoStop(),t.executeEdits(this.id,r,a),t.pushUndoStop()}}const IN=class IN extends ue{constructor(){super({id:IN.ID,label:V(1258,"Trim Trailing Whitespace"),precondition:N.writable,kbOpts:{kbExpr:N.editorTextFocus,primary:Gi(2089,2102),weight:100}})}run(e,t,i){let n=[];i.reason==="auto-save"&&(n=(t.getSelections()||[]).map(d=>new A(d.positionLineNumber,d.positionColumn)));const o=t.getSelection();if(o===null)return;const r=e.get(He),a=t.getModel(),l=r.getValue("files.trimTrailingWhitespaceInRegexAndStrings",{overrideIdentifier:a?.getLanguageId(),resource:a?.uri}),c=new G2e(o,n,l);t.pushUndoStop(),t.executeCommands(this.id,[c]),t.pushUndoStop()}};IN.ID="editor.action.trimTrailingWhitespace";let c3=IN;class aMe extends ue{constructor(){super({id:"editor.action.deleteLines",label:V(1259,"Delete Line"),precondition:N.writable,kbOpts:{kbExpr:N.textInputFocus,primary:3113,weight:100},canTriggerInlineEdits:!0})}run(e,t){if(!t.hasModel())return;const i=this._getLinesToRemove(t),n=t.getModel();if(n.getLineCount()===1&&n.getLineMaxColumn(1)===1)return;let o=0;const r=[],a=[];for(let l=0,c=i.length;l<c;l++){const d=i[l];let h=d.startLineNumber,u=d.endLineNumber,g=1,f=n.getLineMaxColumn(u);u<n.getLineCount()?(u+=1,f=1):h>1&&(h-=1,g=n.getLineMaxColumn(h)),r.push(Si.replace(new he(h,g,u,f),"")),a.push(new he(h-o,d.positionColumn,h-o,d.positionColumn)),o+=d.endLineNumber-d.startLineNumber+1}t.pushUndoStop(),t.executeEdits(this.id,r,a),t.pushUndoStop()}_getLinesToRemove(e){const t=e.getSelections().map(o=>{let r=o.endLineNumber;return o.startLineNumber<o.endLineNumber&&o.endColumn===1&&(r-=1),{startLineNumber:o.startLineNumber,selectionStartColumn:o.selectionStartColumn,endLineNumber:r,positionColumn:o.positionColumn}});t.sort((o,r)=>o.startLineNumber===r.startLineNumber?o.endLineNumber-r.endLineNumber:o.startLineNumber-r.startLineNumber);const i=[];let n=t[0];for(let o=1;o<t.length;o++)n.endLineNumber+1>=t[o].startLineNumber?n.endLineNumber=t[o].endLineNumber:(i.push(n),n=t[o]);return i.push(n),i}}class lMe extends ue{constructor(){super({id:"editor.action.indentLines",label:V(1260,"Indent Line"),precondition:N.writable,kbOpts:{kbExpr:N.editorTextFocus,primary:2142,weight:100},canTriggerInlineEdits:!0})}run(e,t){const i=t._getViewModel();i&&(t.pushUndoStop(),t.executeCommands(this.id,ou.indent(i.cursorConfig,t.getModel(),t.getSelections())),t.pushUndoStop())}}class cMe extends ue{constructor(){super({id:"editor.action.outdentLines",label:V(1261,"Outdent Line"),precondition:N.writable,kbOpts:{kbExpr:N.editorTextFocus,primary:2140,weight:100},canTriggerInlineEdits:!0})}run(e,t){hw.Outdent.runEditorCommand(e,t,null)}}const DN=class DN extends ue{constructor(){super({id:DN.ID,label:V(1262,"Insert Line Above"),precondition:N.writable,kbOpts:{kbExpr:N.editorTextFocus,primary:3075,weight:100},canTriggerInlineEdits:!0})}run(e,t){const i=t._getViewModel();i&&(t.pushUndoStop(),t.executeCommands(this.id,ZR.lineInsertBefore(i.cursorConfig,t.getModel(),t.getSelections())))}};DN.ID="editor.action.insertLineBefore";let yD=DN;const TN=class TN extends ue{constructor(){super({id:TN.ID,label:V(1263,"Insert Line Below"),precondition:N.writable,kbOpts:{kbExpr:N.editorTextFocus,primary:2051,weight:100},canTriggerInlineEdits:!0})}run(e,t){const i=t._getViewModel();i&&(t.pushUndoStop(),t.executeCommands(this.id,ZR.lineInsertAfter(i.cursorConfig,t.getModel(),t.getSelections())))}};TN.ID="editor.action.insertLineAfter";let SD=TN;class aoe extends ue{run(e,t){if(!t.hasModel())return;const i=t.getSelection(),n=this._getRangesToDelete(t),o=[];for(let l=0,c=n.length-1;l<c;l++){const d=n[l],h=n[l+1];k.intersectRanges(d,h)===null?o.push(d):n[l+1]=k.plusRange(d,h)}o.push(n[n.length-1]);const r=this._getEndCursorState(i,o),a=o.map(l=>Si.replace(l,""));t.pushUndoStop(),t.executeEdits(this.id,a,r),t.pushUndoStop()}}class dMe extends aoe{constructor(){super({id:"deleteAllLeft",label:V(1264,"Delete All Left"),precondition:N.writable,kbOpts:{kbExpr:N.textInputFocus,primary:0,mac:{primary:2049},weight:100},canTriggerInlineEdits:!0})}_getEndCursorState(e,t){let i=null;const n=[];let o=0;return t.forEach(r=>{let a;if(r.endColumn===1&&o>0){const l=r.startLineNumber-o;a=new he(l,r.startColumn,l,r.startColumn)}else a=new he(r.startLineNumber,r.startColumn,r.startLineNumber,r.startColumn);o+=r.endLineNumber-r.startLineNumber,r.intersectRanges(e)?i=a:n.push(a)}),i&&n.unshift(i),n}_getRangesToDelete(e){const t=e.getSelections();if(t===null)return[];let i=t;const n=e.getModel();return n===null?[]:(i.sort(k.compareRangesUsingStarts),i=i.map(o=>{if(o.isEmpty())if(o.startColumn===1){const r=Math.max(1,o.startLineNumber-1),a=o.startLineNumber===1?1:n.getLineLength(r)+1;return new k(r,a,o.startLineNumber,1)}else return new k(o.startLineNumber,1,o.startLineNumber,o.startColumn);else return new k(o.startLineNumber,1,o.endLineNumber,o.endColumn)}),i)}}class hMe extends aoe{constructor(){super({id:"deleteAllRight",label:V(1265,"Delete All Right"),precondition:N.writable,kbOpts:{kbExpr:N.textInputFocus,primary:0,mac:{primary:297,secondary:[2068]},weight:100},canTriggerInlineEdits:!0})}_getEndCursorState(e,t){let i=null;const n=[];for(let o=0,r=t.length,a=0;o<r;o++){const l=t[o],c=new he(l.startLineNumber-a,l.startColumn,l.startLineNumber-a,l.startColumn);l.intersectRanges(e)?i=c:n.push(c)}return i&&n.unshift(i),n}_getRangesToDelete(e){const t=e.getModel();if(t===null)return[];const i=e.getSelections();if(i===null)return[];const n=i.map(o=>{if(o.isEmpty()){const r=t.getLineMaxColumn(o.startLineNumber);return o.startColumn===r?new k(o.startLineNumber,o.startColumn,o.startLineNumber+1,1):new k(o.startLineNumber,o.startColumn,o.startLineNumber,r)}return o});return n.sort(k.compareRangesUsingStarts),n}}class uMe extends ue{constructor(){super({id:"editor.action.joinLines",label:V(1266,"Join Lines"),precondition:N.writable,kbOpts:{kbExpr:N.editorTextFocus,primary:0,mac:{primary:296},weight:100},canTriggerInlineEdits:!0})}run(e,t){const i=t.getSelections();if(i===null)return;let n=t.getSelection();if(n===null)return;i.sort(k.compareRangesUsingStarts);const o=[],r=i.reduce((u,g)=>u.isEmpty()?u.endLineNumber===g.startLineNumber?(n.equalsSelection(u)&&(n=g),g):g.startLineNumber>u.endLineNumber+1?(o.push(u),g):new he(u.startLineNumber,u.startColumn,g.endLineNumber,g.endColumn):g.startLineNumber>u.endLineNumber?(o.push(u),g):new he(u.startLineNumber,u.startColumn,g.endLineNumber,g.endColumn));o.push(r);const a=t.getModel();if(a===null)return;const l=[],c=[];let d=n,h=0;for(let u=0,g=o.length;u<g;u++){const f=o[u],p=f.startLineNumber,_=1;let b=0,w,C;const v=a.getLineLength(f.endLineNumber)-f.endColumn;if(f.isEmpty()||f.startLineNumber===f.endLineNumber){const L=f.getStartPosition();L.lineNumber<a.getLineCount()?(w=p+1,C=a.getLineMaxColumn(w)):(w=L.lineNumber,C=a.getLineMaxColumn(L.lineNumber))}else w=f.endLineNumber,C=a.getLineMaxColumn(w);let y=a.getLineContent(p);for(let L=p+1;L<=w;L++){const E=a.getLineContent(L),I=a.getLineFirstNonWhitespaceColumn(L);if(I>=1){let R=!0;y===""&&(R=!1),R&&(y.charAt(y.length-1)===" "||y.charAt(y.length-1)==="	")&&(R=!1,y=y.replace(/[\s\uFEFF\xA0]+$/g," "));const O=E.substr(I-1);y+=(R?" ":"")+O,R?b=O.length+1:b=O.length}else b=0}const x=new k(p,_,w,C);if(!x.isEmpty()){let L;f.isEmpty()?(l.push(Si.replace(x,y)),L=new he(x.startLineNumber-h,y.length-b+1,p-h,y.length-b+1)):f.startLineNumber===f.endLineNumber?(l.push(Si.replace(x,y)),L=new he(f.startLineNumber-h,f.startColumn,f.endLineNumber-h,f.endColumn)):(l.push(Si.replace(x,y)),L=new he(f.startLineNumber-h,f.startColumn,f.startLineNumber-h,y.length-v)),k.intersectRanges(x,n)!==null?d=L:c.push(L)}h+=x.endLineNumber-x.startLineNumber}c.unshift(d),t.pushUndoStop(),t.executeEdits(this.id,l,c),t.pushUndoStop()}}class gMe extends ue{constructor(){super({id:"editor.action.transpose",label:V(1267,"Transpose Characters around the Cursor"),precondition:N.writable,canTriggerInlineEdits:!0})}run(e,t){const i=t.getSelections();if(i===null)return;const n=t.getModel();if(n===null)return;const o=[];for(let r=0,a=i.length;r<a;r++){const l=i[r];if(!l.isEmpty())continue;const c=l.getStartPosition(),d=n.getLineMaxColumn(c.lineNumber);if(c.column>=d){if(c.lineNumber===n.getLineCount())continue;const h=new k(c.lineNumber,Math.max(1,c.column-1),c.lineNumber+1,1),u=n.getValueInRange(h).split("").reverse().join("");o.push(new ts(new he(c.lineNumber,Math.max(1,c.column-1),c.lineNumber+1,1),u))}else{const h=new k(c.lineNumber,Math.max(1,c.column-1),c.lineNumber,c.column+1),u=n.getValueInRange(h).split("").reverse().join("");o.push(new pW(h,u,new he(c.lineNumber,c.column+1,c.lineNumber,c.column+1)))}}t.pushUndoStop(),t.executeCommands(this.id,o),t.pushUndoStop()}}class C_ extends ue{run(e,t){const i=t.getSelections();if(i===null)return;const n=t.getModel();if(n===null)return;const o=t.getOption(148),r=[];for(const a of i)if(a.isEmpty()){const l=a.getStartPosition(),c=t.getConfiguredWordAtPosition(l);if(!c)continue;const d=new k(l.lineNumber,c.startColumn,l.lineNumber,c.endColumn),h=n.getValueInRange(d);r.push(Si.replace(d,this._modifyText(h,o)))}else{const l=n.getValueInRange(a);r.push(Si.replace(a,this._modifyText(l,o)))}t.pushUndoStop(),t.executeEdits(this.id,r),t.pushUndoStop()}}class fMe extends C_{constructor(){super({id:"editor.action.transformToUppercase",label:V(1268,"Transform to Uppercase"),precondition:N.writable,canTriggerInlineEdits:!0})}_modifyText(e,t){return e.toLocaleUpperCase()}}class pMe extends C_{constructor(){super({id:"editor.action.transformToLowercase",label:V(1269,"Transform to Lowercase"),precondition:N.writable,canTriggerInlineEdits:!0})}_modifyText(e,t){return e.toLocaleLowerCase()}}class Pa{constructor(e,t){this._pattern=e,this._flags=t,this._actual=null,this._evaluated=!1}get(){if(!this._evaluated){this._evaluated=!0;try{this._actual=new RegExp(this._pattern,this._flags)}catch{}}return this._actual}isSupported(){return this.get()!==null}}const NN=class NN extends C_{constructor(){super({id:"editor.action.transformToTitlecase",label:V(1270,"Transform to Title Case"),precondition:N.writable,canTriggerInlineEdits:!0})}_modifyText(e,t){const i=NN.titleBoundary.get();return i?e.toLocaleLowerCase().replace(i,n=>n.toLocaleUpperCase()):e}};NN.titleBoundary=new Pa("(^|[^\\p{L}\\p{N}']|((^|\\P{L})'))\\p{L}","gmu");let xD=NN;const nw=class nw extends C_{constructor(){super({id:"editor.action.transformToSnakecase",label:V(1271,"Transform to Snake Case"),precondition:N.writable,canTriggerInlineEdits:!0})}_modifyText(e,t){const i=nw.caseBoundary.get(),n=nw.singleLetters.get();return!i||!n?e:e.replace(i,"$1_$2").replace(n,"$1_$2$3").toLocaleLowerCase()}};nw.caseBoundary=new Pa("(\\p{Ll})(\\p{Lu})","gmu"),nw.singleLetters=new Pa("(\\p{Lu}|\\p{N})(\\p{Lu})(\\p{Ll})","gmu");let ty=nw;const tu=class tu extends C_{constructor(){super({id:"editor.action.transformToCamelcase",label:V(1272,"Transform to Camel Case"),precondition:N.writable,canTriggerInlineEdits:!0})}_modifyText(e,t){const i=/\r\n|\r|\n/.test(e)?tu.multiLineWordBoundary.get():tu.singleLineWordBoundary.get(),n=tu.validWordStart.get();if(!i||!n)return e;const o=e.split(i);return o.shift()?.replace(n,a=>a.toLocaleLowerCase())+o.map(a=>a.substring(0,1).toLocaleUpperCase()+a.substring(1)).join("")}};tu.singleLineWordBoundary=new Pa("[_\\s-]+","gm"),tu.multiLineWordBoundary=new Pa("[_-]+","gm"),tu.validWordStart=new Pa("^(\\p{Lu}[^\\p{Lu}])","gmu");let iy=tu;const iu=class iu extends C_{constructor(){super({id:"editor.action.transformToPascalcase",label:V(1273,"Transform to Pascal Case"),precondition:N.writable,canTriggerInlineEdits:!0})}_modifyText(e,t){const i=iu.wordBoundary.get(),n=iu.wordBoundaryToMaintain.get(),o=iu.upperCaseWordMatcher.get();return!i||!n||!o?e:e.split(n).map(l=>l.split(i)).flat().map(l=>{const c=l.charAt(0).toLocaleUpperCase()+l.slice(1);return c.length>1&&o.test(c)?c.charAt(0)+c.slice(1).toLocaleLowerCase():c}).join("")}};iu.wordBoundary=new Pa("[_ \\t-]","gm"),iu.wordBoundaryToMaintain=new Pa("(?<=\\.)","gm"),iu.upperCaseWordMatcher=new Pa("^\\p{Lu}+$","mu");let LD=iu;const nu=class nu extends C_{static isSupported(){return[this.caseBoundary,this.singleLetters,this.underscoreBoundary].every(t=>t.isSupported())}constructor(){super({id:"editor.action.transformToKebabcase",label:V(1274,"Transform to Kebab Case"),precondition:N.writable,canTriggerInlineEdits:!0})}_modifyText(e,t){const i=nu.caseBoundary.get(),n=nu.singleLetters.get(),o=nu.underscoreBoundary.get();return!i||!n||!o?e:e.replace(o,"$1-$3").replace(i,"$1-$2").replace(n,"$1-$2").toLocaleLowerCase()}};nu.caseBoundary=new Pa("(\\p{Ll})(\\p{Lu})","gmu"),nu.singleLetters=new Pa("(\\p{Lu}|\\p{N})(\\p{Lu}\\p{Ll})","gmu"),nu.underscoreBoundary=new Pa("(\\S)(_)(\\S)","gm");let kD=nu;ae(Q2e);ae(J2e);ae(eMe);ae(tMe);ae(iMe);ae(nMe);ae(sMe);ae(oMe);ae(c3);ae(aMe);ae(lMe);ae(cMe);ae(yD);ae(SD);ae(dMe);ae(hMe);ae(uMe);ae(gMe);ae(fMe);ae(pMe);ae(rMe);ty.caseBoundary.isSupported()&&ty.singleLetters.isSupported()&&ae(ty);iy.singleLineWordBoundary.isSupported()&&iy.multiLineWordBoundary.isSupported()&&ae(iy);LD.wordBoundary.isSupported()&&ae(LD);xD.titleBoundary.isSupported()&&ae(xD);kD.isSupported()&&ae(kD);const jH=qe("dataChannelService");class mMe{getDataChannel(e){return{sendData:()=>{}}}}var _Me=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},_G=function(s,e){return function(t,i){e(t,i,s)}};function bMe(s){return s.sourceId+" @@ "+JSON.stringify({...s,modelUri:s.modelUri?.toString(),sourceId:void 0})}let ED=class extends F{static cast(){return this}constructor(e,t,i){super(),this._key=e,this._contextKeyService=t,this._dataChannelService=i,this._isEnabledContextKeyValue=wMe("structuredLogger.enabled:"+this._key,this._contextKeyService).recomputeInitiallyAndOnChange(this._store),this.isEnabled=this._isEnabledContextKeyValue.map(n=>n!==void 0)}log(e){return this._isEnabledContextKeyValue.get()?(this._dataChannelService.getDataChannel("structuredLogger:"+this._key).sendData(e),!0):!1}};ED=_Me([_G(1,Ae),_G(2,jH)],ED);function wMe(s,e){return bt(e.onDidChangeContext,()=>e.getContextKeyValue(s))}var vMe=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},bG=function(s,e){return function(t,i){e(t,i,s)}};let d3=class extends F{constructor(e,t,i){super(),this._editor=e,this._instantiationService=t,this._loggerService=i,this._structuredLogger=this._register(this._instantiationService.createInstance(ED.cast(),"editor.inlineSuggest.logChangeReason.commandId"));const n=this._loggerService?.createLogger("textModelChanges",{hidden:!1,name:"Text Model Changes Reason"}),o=bt(this,n.onDidChangeLogLevel,()=>n.getLevel());this._register(De(r=>{KQ(o.read(r),pn.Trace)&&r.store.add(this._editor.onDidChangeModelContent(a=>{this._editor.getModel()?.uri.scheme!=="output"&&n.trace("onDidChangeModelContent: "+a.detailedReasons.map(l=>l.toKey(Number.MAX_VALUE)).join(", "))}))})),this._register(De(r=>{this._editor instanceof Am&&this._structuredLogger.isEnabled.read(r)&&r.store.add(this._editor.onDidChangeModelContent(a=>{const l=this._editor.getModel();if(!l)return;const c=a.detailedReasons[0],d={...c.metadata,sourceId:"TextModel.setChangeReason",source:c.metadata.source,time:Date.now(),modelUri:l.uri,modelVersion:l.getVersionId()};setTimeout(()=>{this._structuredLogger.log(d)},0)}))}))}};d3=vMe([bG(1,_e),bG(2,jQ)],d3);function wG(s,e=Bn){return Yue(s,e)?s.charAt(0).toUpperCase()+s.slice(1):s}var CMe=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},yMe=function(s,e){return function(t,i){e(t,i,s)}};class vG{constructor(e){this._delegates=e}resolve(e){for(const t of this._delegates){const i=t.resolve(e);if(i!==void 0)return i}}}class CG{constructor(e,t,i,n){this._model=e,this._selection=t,this._selectionIdx=i,this._overtypingCapturer=n}resolve(e){const{name:t}=e;if(t==="SELECTION"||t==="TM_SELECTED_TEXT"){let i=this._model.getValueInRange(this._selection)||void 0,n=this._selection.startLineNumber!==this._selection.endLineNumber;if(!i&&this._overtypingCapturer){const o=this._overtypingCapturer.getLastOvertypedInfo(this._selectionIdx);o&&(i=o.value,n=o.multiline)}if(i&&n&&e.snippet){const o=this._model.getLineContent(this._selection.startLineNumber),r=Bt(o,0,this._selection.startColumn-1);let a=r;e.snippet.walk(c=>c===e?!1:(c instanceof Xs&&(a=Bt(qo(c.value).pop())),!0));const l=Fa(a,r);i=i.replace(/(\r\n|\r|\n)(.*)/g,(c,d,h)=>`${d}${a.substr(l)}${h}`)}return i}else{if(t==="TM_CURRENT_LINE")return this._model.getLineContent(this._selection.positionLineNumber);if(t==="TM_CURRENT_WORD"){const i=this._model.getWordAtPosition({lineNumber:this._selection.positionLineNumber,column:this._selection.positionColumn});return i&&i.word||void 0}else{if(t==="TM_LINE_INDEX")return String(this._selection.positionLineNumber-1);if(t==="TM_LINE_NUMBER")return String(this._selection.positionLineNumber);if(t==="CURSOR_INDEX")return String(this._selectionIdx);if(t==="CURSOR_NUMBER")return String(this._selectionIdx+1)}}}}class yG{constructor(e,t){this._labelService=e,this._model=t}resolve(e){const{name:t}=e;if(t==="TM_FILENAME")return th(this._model.uri.fsPath);if(t==="TM_FILENAME_BASE"){const i=th(this._model.uri.fsPath),n=i.lastIndexOf(".");return n<=0?i:i.slice(0,n)}else{if(t==="TM_DIRECTORY")return vk(this._model.uri.fsPath)==="."?"":this._labelService.getUriLabel(zR(this._model.uri));if(t==="TM_DIRECTORY_BASE")return vk(this._model.uri.fsPath)==="."?"":th(vk(this._model.uri.fsPath));if(t==="TM_FILEPATH")return this._labelService.getUriLabel(this._model.uri);if(t==="RELATIVE_FILEPATH")return this._labelService.getUriLabel(this._model.uri,{relative:!0,noPrefix:!0})}}}class SG{constructor(e,t,i,n){this._readClipboardText=e,this._selectionIdx=t,this._selectionCount=i,this._spread=n}resolve(e){if(e.name!=="CLIPBOARD")return;const t=this._readClipboardText();if(t){if(this._spread){const i=t.split(/\r\n|\n|\r/).filter(n=>!XX(n));if(i.length===this._selectionCount)return i[this._selectionIdx]}return t}}}let ID=class{constructor(e,t,i){this._model=e,this._selection=t,this._languageConfigurationService=i}resolve(e){const{name:t}=e,i=this._model.getLanguageIdAtPosition(this._selection.selectionStartLineNumber,this._selection.selectionStartColumn),n=this._languageConfigurationService.getLanguageConfiguration(i).comments;if(n){if(t==="LINE_COMMENT")return n.lineCommentToken||void 0;if(t==="BLOCK_COMMENT_START")return n.blockCommentStartToken||void 0;if(t==="BLOCK_COMMENT_END")return n.blockCommentEndToken||void 0}}};ID=CMe([yMe(2,ui)],ID);const tc=class tc{constructor(){this._date=new Date}resolve(e){const{name:t}=e;if(t==="CURRENT_YEAR")return String(this._date.getFullYear());if(t==="CURRENT_YEAR_SHORT")return String(this._date.getFullYear()).slice(-2);if(t==="CURRENT_MONTH")return String(this._date.getMonth().valueOf()+1).padStart(2,"0");if(t==="CURRENT_DATE")return String(this._date.getDate().valueOf()).padStart(2,"0");if(t==="CURRENT_HOUR")return String(this._date.getHours().valueOf()).padStart(2,"0");if(t==="CURRENT_MINUTE")return String(this._date.getMinutes().valueOf()).padStart(2,"0");if(t==="CURRENT_SECOND")return String(this._date.getSeconds().valueOf()).padStart(2,"0");if(t==="CURRENT_DAY_NAME")return tc.dayNames[this._date.getDay()];if(t==="CURRENT_DAY_NAME_SHORT")return tc.dayNamesShort[this._date.getDay()];if(t==="CURRENT_MONTH_NAME")return tc.monthNames[this._date.getMonth()];if(t==="CURRENT_MONTH_NAME_SHORT")return tc.monthNamesShort[this._date.getMonth()];if(t==="CURRENT_SECONDS_UNIX")return String(Math.floor(this._date.getTime()/1e3));if(t==="CURRENT_TIMEZONE_OFFSET"){const i=this._date.getTimezoneOffset(),n=i>0?"-":"+",o=Math.trunc(Math.abs(i/60)),r=o<10?"0"+o:o,a=Math.abs(i)-o*60,l=a<10?"0"+a:a;return n+r+":"+l}}};tc.dayNames=[m(1406,"Sunday"),m(1407,"Monday"),m(1408,"Tuesday"),m(1409,"Wednesday"),m(1410,"Thursday"),m(1411,"Friday"),m(1412,"Saturday")],tc.dayNamesShort=[m(1413,"Sun"),m(1414,"Mon"),m(1415,"Tue"),m(1416,"Wed"),m(1417,"Thu"),m(1418,"Fri"),m(1419,"Sat")],tc.monthNames=[m(1420,"January"),m(1421,"February"),m(1422,"March"),m(1423,"April"),m(1424,"May"),m(1425,"June"),m(1426,"July"),m(1427,"August"),m(1428,"September"),m(1429,"October"),m(1430,"November"),m(1431,"December")],tc.monthNamesShort=[m(1432,"Jan"),m(1433,"Feb"),m(1434,"Mar"),m(1435,"Apr"),m(1436,"May"),m(1437,"Jun"),m(1438,"Jul"),m(1439,"Aug"),m(1440,"Sep"),m(1441,"Oct"),m(1442,"Nov"),m(1443,"Dec")];let DD=tc;class xG{constructor(e){this._workspaceService=e}resolve(e){if(!this._workspaceService)return;const t=aIe(this._workspaceService.getWorkspace());if(!oIe(t)){if(e.name==="WORKSPACE_NAME")return this._resolveWorkspaceName(t);if(e.name==="WORKSPACE_FOLDER")return this._resoveWorkspacePath(t)}}_resolveWorkspaceName(e){if(F4(e))return th(e.uri.path);let t=th(e.configPath.path);return t.endsWith(B4)&&(t=t.substr(0,t.length-B4.length-1)),t}_resoveWorkspacePath(e){if(F4(e))return wG(e.uri.fsPath);const t=th(e.configPath.path);let i=e.configPath.fsPath;return i.endsWith(t)&&(i=i.substr(0,i.length-t.length-1)),i?wG(i):"/"}}class LG{resolve(e){const{name:t}=e;if(t==="RANDOM")return Math.random().toString().slice(-6);if(t==="RANDOM_HEX")return Math.random().toString(16).slice(-6);if(t==="UUID")return d_()}}var SMe=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},xMe=function(s,e){return function(t,i){e(t,i,s)}},Ya;const Sa=class Sa{constructor(e,t,i){this._editor=e,this._snippet=t,this._snippetLineLeadingWhitespace=i,this._offset=-1,this._nestingLevel=1,this._placeholderGroups=Oz(t.placeholders,$r.compareByIndex),this._placeholderGroupsIdx=-1}initialize(e){this._offset=e.newPosition}dispose(){this._placeholderDecorations&&this._editor.removeDecorations([...this._placeholderDecorations.values()]),this._placeholderGroups.length=0}_initDecorations(){if(this._offset===-1)throw new Error("Snippet not initialized!");if(this._placeholderDecorations)return;this._placeholderDecorations=new Map;const e=this._editor.getModel();this._editor.changeDecorations(t=>{for(const i of this._snippet.placeholders){const n=this._snippet.offset(i),o=this._snippet.fullLen(i),r=k.fromPositions(e.getPositionAt(this._offset+n),e.getPositionAt(this._offset+n+o)),a=i.isFinalTabstop?Sa._decor.inactiveFinal:Sa._decor.inactive,l=t.addDecoration(r,a);this._placeholderDecorations.set(i,l)}})}move(e){if(!this._editor.hasModel())return[];if(this._initDecorations(),this._placeholderGroupsIdx>=0){const n=[];for(const o of this._placeholderGroups[this._placeholderGroupsIdx])if(o.transform){const r=this._placeholderDecorations.get(o),a=this._editor.getModel().getDecorationRange(r),l=this._editor.getModel().getValueInRange(a),c=o.transform.resolve(l).split(/\r\n|\r|\n/);for(let d=1;d<c.length;d++)c[d]=this._editor.getModel().normalizeIndentation(this._snippetLineLeadingWhitespace+c[d]);n.push(Si.replace(a,c.join(this._editor.getModel().getEOL())))}n.length>0&&this._editor.executeEdits("snippet.placeholderTransform",n)}let t=!1;e===!0&&this._placeholderGroupsIdx<this._placeholderGroups.length-1?(this._placeholderGroupsIdx+=1,t=!0):e===!1&&this._placeholderGroupsIdx>0&&(this._placeholderGroupsIdx-=1,t=!0);const i=this._editor.getModel().changeDecorations(n=>{const o=new Set,r=[];for(const a of this._placeholderGroups[this._placeholderGroupsIdx]){const l=this._placeholderDecorations.get(a),c=this._editor.getModel().getDecorationRange(l);r.push(new he(c.startLineNumber,c.startColumn,c.endLineNumber,c.endColumn)),t=t&&this._hasPlaceholderBeenCollapsed(a),n.changeDecorationOptions(l,a.isFinalTabstop?Sa._decor.activeFinal:Sa._decor.active),o.add(a);for(const d of this._snippet.enclosingPlaceholders(a)){const h=this._placeholderDecorations.get(d);n.changeDecorationOptions(h,d.isFinalTabstop?Sa._decor.activeFinal:Sa._decor.active),o.add(d)}}for(const[a,l]of this._placeholderDecorations)o.has(a)||n.changeDecorationOptions(l,a.isFinalTabstop?Sa._decor.inactiveFinal:Sa._decor.inactive);return r});return t?this.move(e):i??[]}_hasPlaceholderBeenCollapsed(e){let t=e;for(;t;){if(t instanceof $r){const i=this._placeholderDecorations.get(t);if(this._editor.getModel().getDecorationRange(i).isEmpty()&&t.toString().length>0)return!0}t=t.parent}return!1}get isAtFirstPlaceholder(){return this._placeholderGroupsIdx<=0||this._placeholderGroups.length===0}get isAtLastPlaceholder(){return this._placeholderGroupsIdx===this._placeholderGroups.length-1}get hasPlaceholder(){return this._snippet.placeholders.length>0}get isTrivialSnippet(){if(this._snippet.placeholders.length===0)return!0;if(this._snippet.placeholders.length===1){const[e]=this._snippet.placeholders;if(e.isFinalTabstop&&this._snippet.rightMostDescendant===e)return!0}return!1}computePossibleSelections(){const e=new Map;for(const t of this._placeholderGroups){let i;for(const n of t){if(n.isFinalTabstop)break;i||(i=[],e.set(n.index,i));const o=this._placeholderDecorations.get(n),r=this._editor.getModel().getDecorationRange(o);if(!r){e.delete(n.index);break}i.push(r)}}return e}get activeChoice(){if(!this._placeholderDecorations)return;const e=this._placeholderGroups[this._placeholderGroupsIdx][0];if(!e?.choice)return;const t=this._placeholderDecorations.get(e);if(!t)return;const i=this._editor.getModel().getDecorationRange(t);if(i)return{range:i,choice:e.choice}}get hasChoice(){let e=!1;return this._snippet.walk(t=>(e=t instanceof dC,!e)),e}merge(e){const t=this._editor.getModel();this._nestingLevel*=10,this._editor.changeDecorations(i=>{for(const n of this._placeholderGroups[this._placeholderGroupsIdx]){const o=e.shift();console.assert(o._offset!==-1),console.assert(!o._placeholderDecorations);const r=o._snippet.placeholderInfo.last.index;for(const l of o._snippet.placeholderInfo.all)l.isFinalTabstop?l.index=n.index+(r+1)/this._nestingLevel:l.index=n.index+l.index/this._nestingLevel;this._snippet.replace(n,o._snippet.children);const a=this._placeholderDecorations.get(n);i.removeDecoration(a),this._placeholderDecorations.delete(n);for(const l of o._snippet.placeholders){const c=o._snippet.offset(l),d=o._snippet.fullLen(l),h=k.fromPositions(t.getPositionAt(o._offset+c),t.getPositionAt(o._offset+c+d)),u=i.addDecoration(h,Sa._decor.inactive);this._placeholderDecorations.set(l,u)}}this._placeholderGroups=Oz(this._snippet.placeholders,$r.compareByIndex)})}};Sa._decor={active:Fe.register({description:"snippet-placeholder-1",stickiness:0,className:"snippet-placeholder"}),inactive:Fe.register({description:"snippet-placeholder-2",stickiness:1,className:"snippet-placeholder"}),activeFinal:Fe.register({description:"snippet-placeholder-3",stickiness:1,className:"finish-snippet-placeholder"}),inactiveFinal:Fe.register({description:"snippet-placeholder-4",stickiness:1,className:"finish-snippet-placeholder"})};let TD=Sa;const kG={overwriteBefore:0,overwriteAfter:0,adjustWhitespace:!0,clipboardText:void 0,overtypingCapturer:void 0};let ND=Ya=class{static adjustWhitespace(e,t,i,n,o){const r=e.getLineContent(t.lineNumber),a=Bt(r,0,t.column-1);let l;return n.walk(c=>{if(!(c instanceof Xs)||c.parent instanceof dC||o&&!o.has(c))return!0;const d=c.value.split(/\r\n|\r|\n/);if(i){const u=n.offset(c);if(u===0)d[0]=e.normalizeIndentation(d[0]);else{l=l??n.toString();const g=l.charCodeAt(u-1);(g===10||g===13)&&(d[0]=e.normalizeIndentation(a+d[0]))}for(let g=1;g<d.length;g++)d[g]=e.normalizeIndentation(a+d[g])}const h=d.join(e.getEOL());return h!==c.value&&(c.parent.replace(c,[new Xs(h)]),l=void 0),!0}),a}static adjustSelection(e,t,i,n){if(i!==0||n!==0){const{positionLineNumber:o,positionColumn:r}=t,a=r-i,l=r+n,c=e.validateRange({startLineNumber:o,startColumn:a,endLineNumber:o,endColumn:l});t=he.createWithDirection(c.startLineNumber,c.startColumn,c.endLineNumber,c.endColumn,t.getDirection())}return t}static createEditsAndSnippetsFromSelections(e,t,i,n,o,r,a,l,c){const d=[],h=[];if(!e.hasModel())return{edits:d,snippets:h};const u=e.getModel(),g=e.invokeWithinContext(v=>v.get(Sh)),f=e.invokeWithinContext(v=>new yG(v.get(Ym),u)),p=()=>a,_=u.getValueInRange(Ya.adjustSelection(u,e.getSelection(),i,0)),b=u.getValueInRange(Ya.adjustSelection(u,e.getSelection(),0,n)),w=u.getLineFirstNonWhitespaceColumn(e.getSelection().positionLineNumber),C=e.getSelections().map((v,y)=>({selection:v,idx:y})).sort((v,y)=>k.compareRangesUsingStarts(v.selection,y.selection));for(const{selection:v,idx:y}of C){let x=Ya.adjustSelection(u,v,i,0),L=Ya.adjustSelection(u,v,0,n);_!==u.getValueInRange(x)&&(x=v),b!==u.getValueInRange(L)&&(L=v);const E=v.setStartPosition(x.startLineNumber,x.startColumn).setEndPosition(L.endLineNumber,L.endColumn),I=new Hm().parse(t,!0,o),R=E.getStartPosition(),O=Ya.adjustWhitespace(u,R,r||y>0&&w!==u.getLineFirstNonWhitespaceColumn(v.positionLineNumber),I);I.resolveVariables(new vG([f,new SG(p,y,C.length,e.getOption(88)==="spread"),new CG(u,v,y,l),new ID(u,v,c),new DD,new xG(g),new LG])),d[y]=Si.replace(E,I.toString()),d[y].identifier={major:y,minor:0},d[y]._isTracked=!0,h[y]=new TD(e,I,O)}return{edits:d,snippets:h}}static createEditsAndSnippetsFromEdits(e,t,i,n,o,r,a){if(!e.hasModel()||t.length===0)return{edits:[],snippets:[]};const l=[],c=e.getModel(),d=new Hm,h=new px,u=new vG([e.invokeWithinContext(f=>new yG(f.get(Ym),c)),new SG(()=>o,0,e.getSelections().length,e.getOption(88)==="spread"),new CG(c,e.getSelection(),0,r),new ID(c,e.getSelection(),a),new DD,new xG(e.invokeWithinContext(f=>f.get(Sh))),new LG]);t=t.sort((f,p)=>k.compareRangesUsingStarts(f.range,p.range));let g=0;for(let f=0;f<t.length;f++){const{range:p,template:_,keepWhitespace:b}=t[f];if(f>0){const x=t[f-1].range,L=k.fromPositions(x.getEndPosition(),p.getStartPosition()),E=new Xs(c.getValueInRange(L));h.appendChild(E),g+=E.value.length}const w=d.parseFragment(_,h);Ya.adjustWhitespace(c,p.getStartPosition(),b!==void 0?!b:n,h,new Set(w)),h.resolveVariables(u);const C=h.toString(),v=C.slice(g);g=C.length;const y=Si.replace(p,v);y.identifier={major:f,minor:0},y._isTracked=!0,l.push(y)}return d.ensureFinalTabstop(h,i,!0),{edits:l,snippets:[new TD(e,h,"")]}}constructor(e,t,i=kG,n){this._editor=e,this._template=t,this._options=i,this._languageConfigurationService=n,this._templateMerges=[],this._snippets=[]}dispose(){xt(this._snippets)}_logInfo(){return`template="${this._template}", merged_templates="${this._templateMerges.join(" -> ")}"`}insert(e){if(!this._editor.hasModel())return;const{edits:t,snippets:i}=typeof this._template=="string"?Ya.createEditsAndSnippetsFromSelections(this._editor,this._template,this._options.overwriteBefore,this._options.overwriteAfter,!1,this._options.adjustWhitespace,this._options.clipboardText,this._options.overtypingCapturer,this._languageConfigurationService):Ya.createEditsAndSnippetsFromEdits(this._editor,this._template,!1,this._options.adjustWhitespace,this._options.clipboardText,this._options.overtypingCapturer,this._languageConfigurationService);this._snippets=i,this._editor.executeEdits(e??cs.snippet(),t,n=>{const o=n.filter(r=>!!r.identifier);for(let r=0;r<i.length;r++)i[r].initialize(o[r].textChange);return this._snippets[0].hasPlaceholder?this._move(!0):o.map(r=>he.fromPositions(r.range.getEndPosition()))}),this._editor.revealRange(this._editor.getSelections()[0])}merge(e,t=kG){if(!this._editor.hasModel())return;this._templateMerges.push([this._snippets[0]._nestingLevel,this._snippets[0]._placeholderGroupsIdx,e]);const{edits:i,snippets:n}=Ya.createEditsAndSnippetsFromSelections(this._editor,e,t.overwriteBefore,t.overwriteAfter,!0,t.adjustWhitespace,t.clipboardText,t.overtypingCapturer,this._languageConfigurationService);this._editor.executeEdits("snippet",i,o=>{const r=o.filter(l=>!!l.identifier);for(let l=0;l<n.length;l++)n[l].initialize(r[l].textChange);const a=n[0].isTrivialSnippet;if(!a){for(const l of this._snippets)l.merge(n);console.assert(n.length===0)}return this._snippets[0].hasPlaceholder&&!a?this._move(void 0):r.map(l=>he.fromPositions(l.range.getEndPosition()))})}next(){const e=this._move(!0);this._editor.setSelections(e),this._editor.revealPositionInCenterIfOutsideViewport(e[0].getPosition())}prev(){const e=this._move(!1);this._editor.setSelections(e),this._editor.revealPositionInCenterIfOutsideViewport(e[0].getPosition())}_move(e){const t=[];for(const i of this._snippets){const n=i.move(e);t.push(...n)}return t}get isAtFirstPlaceholder(){return this._snippets[0].isAtFirstPlaceholder}get isAtLastPlaceholder(){return this._snippets[0].isAtLastPlaceholder}get hasPlaceholder(){return this._snippets[0].hasPlaceholder}get hasChoice(){return this._snippets[0].hasChoice}get activeChoice(){return this._snippets[0].activeChoice}isSelectionWithinPlaceholders(){if(!this.hasPlaceholder)return!1;const e=this._editor.getSelections();if(e.length<this._snippets.length)return!1;const t=new Map;for(const i of this._snippets){const n=i.computePossibleSelections();if(t.size===0)for(const[o,r]of n){r.sort(k.compareRangesUsingStarts);for(const a of e)if(r[0].containsRange(a)){t.set(o,[]);break}}if(t.size===0)return!1;t.forEach((o,r)=>{o.push(...n.get(r))})}e.sort(k.compareRangesUsingStarts);for(const[i,n]of t){if(n.length!==e.length){t.delete(i);continue}n.sort(k.compareRangesUsingStarts);for(let o=0;o<n.length;o++)if(!n[o].containsRange(e[o])){t.delete(i);continue}}return t.size>0}};ND=Ya=SMe([xMe(3,ui)],ND);var LMe=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},zL=function(s,e){return function(t,i){e(t,i,s)}},mb;const EG={overwriteBefore:0,overwriteAfter:0,undoStopBefore:!0,undoStopAfter:!0,adjustWhitespace:!0,clipboardText:void 0,overtypingCapturer:void 0};var _c;let Ps=(_c=class{static get(e){return e.getContribution(mb.ID)}constructor(e,t,i,n,o){this._editor=e,this._logService=t,this._languageFeaturesService=i,this._languageConfigurationService=o,this._inSnippetObservable=Re(this,!1),this._snippetListener=new U,this._modelVersionId=-1,this._inSnippet=mb.InSnippetMode.bindTo(n),this._hasNextTabstop=mb.HasNextTabstop.bindTo(n),this._hasPrevTabstop=mb.HasPrevTabstop.bindTo(n)}dispose(){this._inSnippet.reset(),this._inSnippetObservable.set(!1,void 0),this._hasPrevTabstop.reset(),this._hasNextTabstop.reset(),this._session?.dispose(),this._snippetListener.dispose()}insert(e,t){try{this._doInsert(e,typeof t>"u"?EG:{...EG,...t})}catch(i){this.cancel(),this._logService.error(i),this._logService.error("snippet_error"),this._logService.error("insert_template=",e),this._logService.error("existing_template=",this._session?this._session._logInfo():"<no_session>")}}_doInsert(e,t){if(this._editor.hasModel()){if(this._snippetListener.clear(),t.undoStopBefore&&this._editor.getModel().pushStackElement(),this._session&&typeof e!="string"&&this.cancel(),this._session?(dt(typeof e=="string"),this._session.merge(e,t)):(this._modelVersionId=this._editor.getModel().getAlternativeVersionId(),this._session=new ND(this._editor,e,t,this._languageConfigurationService),this._session.insert(t.reason)),t.undoStopAfter&&this._editor.getModel().pushStackElement(),this._session?.hasChoice){const i={_debugDisplayName:"snippetChoiceCompletions",provideCompletionItems:(c,d)=>{if(!this._session||c!==this._editor.getModel()||!A.equals(this._editor.getPosition(),d))return;const{activeChoice:h}=this._session;if(!h||h.choice.options.length===0)return;const u=c.getValueInRange(h.range),g=!!h.choice.options.find(p=>p.value===u),f=[];for(let p=0;p<h.choice.options.length;p++){const _=h.choice.options[p];f.push({kind:13,label:_.value,insertText:_.value,sortText:"a".repeat(p+1),range:h.range,filterText:g?`${u}_${_.value}`:void 0,command:{id:"jumpToNextSnippetPlaceholder",title:m(1405,"Go to next placeholder...")}})}return{suggestions:f}}},n=this._editor.getModel();let o,r=!1;const a=()=>{o?.dispose(),r=!1},l=()=>{r||(o=this._languageFeaturesService.completionProvider.register({language:n.getLanguageId(),pattern:n.uri.fsPath,scheme:n.uri.scheme,exclusive:!0},i),this._snippetListener.add(o),r=!0)};this._choiceCompletions={provider:i,enable:l,disable:a}}this._updateState(),this._snippetListener.add(this._editor.onDidChangeModelContent(i=>i.isFlush&&this.cancel())),this._snippetListener.add(this._editor.onDidChangeModel(()=>this.cancel())),this._snippetListener.add(this._editor.onDidChangeCursorSelection(()=>this._updateState()))}}_updateState(){if(!(!this._session||!this._editor.hasModel())){if(this._modelVersionId===this._editor.getModel().getAlternativeVersionId())return this.cancel();if(!this._session.hasPlaceholder)return this.cancel();if(this._session.isAtLastPlaceholder||!this._session.isSelectionWithinPlaceholders())return this._editor.getModel().pushStackElement(),this.cancel();this._inSnippet.set(!0),this._inSnippetObservable.set(!0,void 0),this._hasPrevTabstop.set(!this._session.isAtFirstPlaceholder),this._hasNextTabstop.set(!this._session.isAtLastPlaceholder),this._handleChoice()}}_handleChoice(){if(!this._session||!this._editor.hasModel()){this._currentChoice=void 0;return}const{activeChoice:e}=this._session;if(!e||!this._choiceCompletions){this._choiceCompletions?.disable(),this._currentChoice=void 0;return}this._currentChoice!==e.choice&&(this._currentChoice=e.choice,this._choiceCompletions.enable(),queueMicrotask(()=>{j2e(this._editor,this._choiceCompletions.provider)}))}finish(){for(;this._inSnippet.get();)this.next()}cancel(e=!1){this._inSnippet.reset(),this._inSnippetObservable.set(!1,void 0),this._hasPrevTabstop.reset(),this._hasNextTabstop.reset(),this._snippetListener.clear(),this._currentChoice=void 0,this._session?.dispose(),this._session=void 0,this._modelVersionId=-1,e&&this._editor.setSelections([this._editor.getSelection()])}prev(){this._session?.prev(),this._updateState()}next(){this._session?.next(),this._updateState()}isInSnippet(){return!!this._inSnippet.get()}get isInSnippetObservable(){return this._inSnippetObservable}},mb=_c,_c.ID="snippetController2",_c.InSnippetMode=new ce("inSnippetMode",!1,m(1402,"Whether the editor in current in snippet mode")),_c.HasNextTabstop=new ce("hasNextTabstop",!1,m(1403,"Whether there is a next tab stop when in snippet mode")),_c.HasPrevTabstop=new ce("hasPrevTabstop",!1,m(1404,"Whether there is a previous tab stop when in snippet mode")),_c);Ps=mb=LMe([zL(1,jt),zL(2,ge),zL(3,Ae),zL(4,ui)],Ps);lt(Ps.ID,Ps,4);const GA=hn.bindToContribution(Ps.get);le(new GA({id:"jumpToNextSnippetPlaceholder",precondition:K.and(Ps.InSnippetMode,Ps.HasNextTabstop),handler:s=>s.next(),kbOpts:{weight:130,kbExpr:N.textInputFocus,primary:2}}));le(new GA({id:"jumpToPrevSnippetPlaceholder",precondition:K.and(Ps.InSnippetMode,Ps.HasPrevTabstop),handler:s=>s.prev(),kbOpts:{weight:130,kbExpr:N.textInputFocus,primary:1026}}));le(new GA({id:"leaveSnippet",precondition:Ps.InSnippetMode,handler:s=>s.cancel(!0),kbOpts:{weight:130,kbExpr:N.textInputFocus,primary:9,secondary:[1033]}}));le(new GA({id:"acceptSnippet",precondition:Ps.InSnippetMode,handler:s=>s.finish()}));function o0(s){return new kMe(s)}class kMe extends Bee{constructor(e){super(),this._textModel=e}getOffset(e){return this._textModel.getOffsetAt(e)}getPosition(e){return this._textModel.getPositionAt(e)}}const EMe=[];function IMe(){return EMe}function IG(s){return DMe(s).map(t=>t.getEndPosition())}function DMe(s){const e=yE.createSortPermutation(s,ns(n=>n.range,k.compareRangesUsingStarts)),i=new Na(e.apply(s)).getNewRanges();return e.inverse().apply(i)}function TMe(s,e){const t=o0(e),i=e.getValue();return s.map(r=>t.getStringReplacement(r)).map(r=>r.removeCommonSuffixPrefix(i)).map(r=>t.getTextReplacement(r))}function NMe(s,e){const t=Re("result",[]),i=[];return e.add(De(n=>{const o=s.read(n);Vt(r=>{if(o.length!==i.length){i.length=o.length;for(let a=0;a<i.length;a++)i[a]||(i[a]=Re("item",o[a]));t.set([...i],r)}i.forEach((a,l)=>a.set(o[l],r))})})),t}class RMe{constructor(e){this._contextKeyService=e}bind(e,t){return jl(e,this._contextKeyService,t instanceof Function?t:i=>t.read(i))}}function DG(s,e){return new Promise(t=>{let i;const n=setTimeout(()=>{i&&i.dispose(),t()},s);e&&(i=e.onCancellationRequested(()=>{clearTimeout(n),i&&i.dispose(),t()}))})}class AMe{constructor(e,t,i,n=MMe){this.startValue=e,this.endValue=t,this.durationMs=i,this._interpolationFunction=n,this.startTimeMs=Date.now(),e===t&&(this.durationMs=0)}isFinished(){return Date.now()>=this.startTimeMs+this.durationMs}getValue(){const e=Date.now()-this.startTimeMs;return e>=this.durationMs?this.endValue:this._interpolationFunction(e,this.startValue,this.endValue-this.startValue,this.durationMs)}}function MMe(s,e,t,i){return s===i?e+t:t*(-Math.pow(2,-10*s/i)+1)+e}function OMe(s,e,t,i){return t*((s=s/i-1)*s*s+1)+e}class PMe{constructor(e){this._value=Re(this,e)}getValue(e){const t=this._value.read(e);return t.isFinished()||RD.instance.invalidateOnNextAnimationFrame(e),t.getValue()}}const RN=class RN{constructor(){this._counter=Yr(this),this._isScheduled=!1}invalidateOnNextAnimationFrame(e){this._counter.read(e),this._isScheduled||(this._isScheduled=!0,kt().requestAnimationFrame(()=>{this._isScheduled=!1,this._update()}))}_update(){this._counter.trigger(void 0)}};RN.instance=new RN;let RD=RN;class r0{constructor(e,t){this.lineNumber=e,this.parts=t,rg(()=>H0(t,(i,n)=>i.column<=n.column))}equals(e){return this.lineNumber===e.lineNumber&&this.parts.length===e.parts.length&&this.parts.every((t,i)=>t.equals(e.parts[i]))}renderForScreenReader(e){if(this.parts.length===0)return"";const t=this.parts[this.parts.length-1],i=e.substr(0,t.column-1);return new Na([...this.parts.map(o=>new Ui(k.fromPositions(new A(1,o.column)),o.lines.map(r=>r.line).join(`
`)))]).applyToString(i).substring(this.parts[0].column-1)}isEmpty(){return this.parts.every(e=>e.lines.length===0)}get lineCount(){return 1+this.parts.reduce((e,t)=>e+t.lines.length-1,0)}}class AD{constructor(e,t,i,n=[]){this.column=e,this.text=t,this.preview=i,this._inlineDecorations=n,this.lines=qo(this.text).map((o,r)=>({line:o,lineDecorations:Rs.filter(this._inlineDecorations,r+1,1,o.length+1)}))}equals(e){return this.column===e.column&&this.lines.length===e.lines.length&&this.lines.every((t,i)=>t.line===e.lines[i].line&&Rs.equalsArr(t.lineDecorations,e.lines[i].lineDecorations))}}class h3{constructor(e,t,i,n=0){this.lineNumber=e,this.columnRange=t,this.text=i,this.additionalReservedLineCount=n,this.parts=[new AD(this.columnRange.endColumnExclusive,this.text,!1)],this.newLines=qo(this.text)}renderForScreenReader(e){return this.newLines.join(`
`)}get lineCount(){return this.newLines.length}isEmpty(){return this.parts.every(e=>e.lines.length===0)}equals(e){return this.lineNumber===e.lineNumber&&this.columnRange.equals(e.columnRange)&&this.newLines.length===e.newLines.length&&this.newLines.every((t,i)=>t===e.newLines[i])&&this.additionalReservedLineCount===e.additionalReservedLineCount}}function TG(s,e){return ni(s,e,loe)}function loe(s,e){return s===e?!0:!s||!e?!1:s instanceof r0&&e instanceof r0||s instanceof h3&&e instanceof h3?s.equals(e):!1}function jd(s,e,t){const i=t?s.range.intersectRanges(t):s.range;if(!i)return s;const n=s.text.replaceAll(`\r
`,`
`),o=e.getValueInRange(i,1),r=Fa(o,n),a=cn.ofText(o.substring(0,r)).addToPosition(s.range.getStartPosition()),l=n.substring(r),c=k.fromPositions(a,s.range.getEndPosition());return new Ui(c,l)}function coe(s,e){return s.text.startsWith(e.text)&&FMe(s.range,e.range)}function FMe(s,e){return e.getStartPosition().equals(s.getStartPosition())&&e.getEndPosition().isBeforeOrEqual(s.getEndPosition())}function NG(s,e,t,i,n=0){let o=jd(s,e);if(o.range.endLineNumber!==o.range.startLineNumber)return;const r=e.getLineContent(o.range.startLineNumber),a=Bt(r).length;if(o.range.startColumn-1<=a){const f=Bt(o.text).length,p=r.substring(o.range.startColumn-1,a),[_,b]=[o.range.getStartPosition(),o.range.getEndPosition()],w=_.column+p.length<=b.column?_.delta(0,p.length):b,C=k.fromPositions(w,b),v=o.text.startsWith(p)?o.text.substring(p.length):o.text.substring(f);o=new Ui(C,v)}const c=e.getValueInRange(o.range),d=BMe(c,o.text);if(!d)return;const h=o.range.startLineNumber,u=new Array;if(t==="prefix"){const f=d.filter(p=>p.originalLength===0);if(f.length>1||f.length===1&&f[0].originalStart!==c.length)return}const g=o.text.length-n;for(const f of d){const p=o.range.startColumn+f.originalStart+f.originalLength;if(t==="subwordSmart"&&i&&i.lineNumber===o.range.startLineNumber&&p<i.column||f.originalLength>0)return;if(f.modifiedLength===0)continue;const _=f.modifiedStart+f.modifiedLength,b=Math.max(f.modifiedStart,Math.min(_,g)),w=o.text.substring(f.modifiedStart,b),C=o.text.substring(b,Math.max(f.modifiedStart,_));w.length>0&&u.push(new AD(p,w,!1)),C.length>0&&u.push(new AD(p,C,!0))}return new r0(h,u)}let $L;function BMe(s,e){if($L?.originalValue===s&&$L?.newValue===e)return $L?.changes;{let t=AG(s,e,!0);if(t){const i=RG(t);if(i>0){const n=AG(s,e,!1);n&&RG(n)<i&&(t=n)}}return $L={originalValue:s,newValue:e,changes:t},t}}function RG(s){let e=0;for(const t of s)e+=t.originalLength;return e}function AG(s,e,t){if(s.length>5e3||e.length>5e3)return;function i(c){let d=0;for(let h=0,u=c.length;h<u;h++){const g=c.charCodeAt(h);g>d&&(d=g)}return d}const n=Math.max(i(s),i(e));function o(c){if(c<0)throw new Error("unexpected");return n+c+1}function r(c){let d=0,h=0;const u=new Int32Array(c.length);for(let g=0,f=c.length;g<f;g++)if(t&&c[g]==="("){const p=h*100+d;u[g]=o(2*p),d++}else if(t&&c[g]===")"){d=Math.max(d-1,0);const p=h*100+d;u[g]=o(2*p+1),d===0&&h++}else u[g]=c.charCodeAt(g);return u}const a=r(s),l=r(e);return new fc({getElements:()=>a},{getElements:()=>l}).ComputeDiff(!1).changes}function WMe(s,e){let t,i=!1;const n=new yQ(new os(s,void 0,e.update),(o,r)=>{i||(t=e.initial instanceof Function?e.initial():e.initial,i=!0);const a=e.update(o,t,r);return t=a,a},e.changeTracker,()=>{i&&(e.disposeFinal?.(t),i=!1)},e.equalityComparer??kr,(o,r,a)=>{if(!i)throw new Ce("Can only set when there is a listener! This is to prevent leaks.");nv(r,l=>{t=o,n.setValue(o,l,a)})},dn.ofCaller());return n}var HMe=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},MG=function(s,e){return function(t,i){e(t,i,s)}};class VMe{constructor(e,t){this._baseService=e,this._intercept=t}publicLog2(e,t){this._intercept(e,t),this._baseService.publicLog2(e,t)}}let u3=class extends VMe{constructor(e,t){super(e,(i,n)=>{let o=!0;n&&g3 in n&&(o=!!n[g3]),o&&t.getDataChannel("editTelemetry").sendData({eventName:i,data:n??{}})})}};u3=HMe([MG(0,vs),MG(1,jH)],u3);const g3=Symbol("shouldForwardToChannel");function zMe(s){return{[g3]:s}}function OG(s){if(!s)return!1;const e=s.toLowerCase();return e==="github.copilot"||e==="github.copilot-chat"}function $Me(s){const e=s.map(i=>new On(me.ofStartAndLength(i.rangeOffset,i.rangeLength),i.text));return e.reverse(),new vh(e)}function UMe(s,e){s.publicLog2("inlineCompletion.endOfLife",e)}class Yk{constructor(e,t,i){this.changes=e,this.moves=t,this.hitTimeout=i}}class doe{constructor(e,t){this.lineRangeMapping=e,this.changes=t}}const qMe=3;class jMe{computeDiff(e,t,i){const o=new YMe(e,t,{maxComputationTime:i.maxComputationTimeMs,shouldIgnoreTrimWhitespace:i.ignoreTrimWhitespace,shouldComputeCharChanges:!0,shouldMakePrettyDiff:!0,shouldPostProcessCharChanges:!0}).computeDiff(),r=[];let a=null;for(const l of o.changes){let c;l.originalEndLineNumber===0?c=new Ne(l.originalStartLineNumber+1,l.originalStartLineNumber+1):c=new Ne(l.originalStartLineNumber,l.originalEndLineNumber+1);let d;l.modifiedEndLineNumber===0?d=new Ne(l.modifiedStartLineNumber+1,l.modifiedStartLineNumber+1):d=new Ne(l.modifiedStartLineNumber,l.modifiedEndLineNumber+1);let h=new Lr(c,d,l.charChanges?.map(u=>new Ks(new k(u.originalStartLineNumber,u.originalStartColumn,u.originalEndLineNumber,u.originalEndColumn),new k(u.modifiedStartLineNumber,u.modifiedStartColumn,u.modifiedEndLineNumber,u.modifiedEndColumn))));a&&(a.modified.endLineNumberExclusive===h.modified.startLineNumber||a.original.endLineNumberExclusive===h.original.startLineNumber)&&(h=new Lr(a.original.join(h.original),a.modified.join(h.modified),a.innerChanges&&h.innerChanges?a.innerChanges.concat(h.innerChanges):void 0),r.pop()),r.push(h),a=h}return rg(()=>H0(r,(l,c)=>c.original.startLineNumber-l.original.endLineNumberExclusive===c.modified.startLineNumber-l.modified.endLineNumberExclusive&&l.original.endLineNumberExclusive<c.original.startLineNumber&&l.modified.endLineNumberExclusive<c.modified.startLineNumber)),new Yk(r,[],o.quitEarly)}}function hoe(s,e,t,i){return new fc(s,e,t).ComputeDiff(i)}let PG=class{constructor(e){const t=[],i=[];for(let n=0,o=e.length;n<o;n++)t[n]=f3(e[n],1),i[n]=p3(e[n],1);this.lines=e,this._startColumns=t,this._endColumns=i}getElements(){const e=[];for(let t=0,i=this.lines.length;t<i;t++)e[t]=this.lines[t].substring(this._startColumns[t]-1,this._endColumns[t]-1);return e}getStrictElement(e){return this.lines[e]}getStartLineNumber(e){return e+1}getEndLineNumber(e){return e+1}createCharSequence(e,t,i){const n=[],o=[],r=[];let a=0;for(let l=t;l<=i;l++){const c=this.lines[l],d=e?this._startColumns[l]:1,h=e?this._endColumns[l]:c.length+1;for(let u=d;u<h;u++)n[a]=c.charCodeAt(u-1),o[a]=l+1,r[a]=u,a++;!e&&l<i&&(n[a]=10,o[a]=l+1,r[a]=c.length+1,a++)}return new GMe(n,o,r)}};class GMe{constructor(e,t,i){this._charCodes=e,this._lineNumbers=t,this._columns=i}toString(){return"["+this._charCodes.map((e,t)=>(e===10?"\\n":String.fromCharCode(e))+`-(${this._lineNumbers[t]},${this._columns[t]})`).join(", ")+"]"}_assertIndex(e,t){if(e<0||e>=t.length)throw new Error("Illegal index")}getElements(){return this._charCodes}getStartLineNumber(e){return e>0&&e===this._lineNumbers.length?this.getEndLineNumber(e-1):(this._assertIndex(e,this._lineNumbers),this._lineNumbers[e])}getEndLineNumber(e){return e===-1?this.getStartLineNumber(e+1):(this._assertIndex(e,this._lineNumbers),this._charCodes[e]===10?this._lineNumbers[e]+1:this._lineNumbers[e])}getStartColumn(e){return e>0&&e===this._columns.length?this.getEndColumn(e-1):(this._assertIndex(e,this._columns),this._columns[e])}getEndColumn(e){return e===-1?this.getStartColumn(e+1):(this._assertIndex(e,this._columns),this._charCodes[e]===10?1:this._columns[e]+1)}}class Ew{constructor(e,t,i,n,o,r,a,l){this.originalStartLineNumber=e,this.originalStartColumn=t,this.originalEndLineNumber=i,this.originalEndColumn=n,this.modifiedStartLineNumber=o,this.modifiedStartColumn=r,this.modifiedEndLineNumber=a,this.modifiedEndColumn=l}static createFromDiffChange(e,t,i){const n=t.getStartLineNumber(e.originalStart),o=t.getStartColumn(e.originalStart),r=t.getEndLineNumber(e.originalStart+e.originalLength-1),a=t.getEndColumn(e.originalStart+e.originalLength-1),l=i.getStartLineNumber(e.modifiedStart),c=i.getStartColumn(e.modifiedStart),d=i.getEndLineNumber(e.modifiedStart+e.modifiedLength-1),h=i.getEndColumn(e.modifiedStart+e.modifiedLength-1);return new Ew(n,o,r,a,l,c,d,h)}}function KMe(s){if(s.length<=1)return s;const e=[s[0]];let t=e[0];for(let i=1,n=s.length;i<n;i++){const o=s[i],r=o.originalStart-(t.originalStart+t.originalLength),a=o.modifiedStart-(t.modifiedStart+t.modifiedLength);Math.min(r,a)<qMe?(t.originalLength=o.originalStart+o.originalLength-t.originalStart,t.modifiedLength=o.modifiedStart+o.modifiedLength-t.modifiedStart):(e.push(o),t=o)}return e}class ny{constructor(e,t,i,n,o){this.originalStartLineNumber=e,this.originalEndLineNumber=t,this.modifiedStartLineNumber=i,this.modifiedEndLineNumber=n,this.charChanges=o}static createFromDiffResult(e,t,i,n,o,r,a){let l,c,d,h,u;if(t.originalLength===0?(l=i.getStartLineNumber(t.originalStart)-1,c=0):(l=i.getStartLineNumber(t.originalStart),c=i.getEndLineNumber(t.originalStart+t.originalLength-1)),t.modifiedLength===0?(d=n.getStartLineNumber(t.modifiedStart)-1,h=0):(d=n.getStartLineNumber(t.modifiedStart),h=n.getEndLineNumber(t.modifiedStart+t.modifiedLength-1)),r&&t.originalLength>0&&t.originalLength<20&&t.modifiedLength>0&&t.modifiedLength<20&&o()){const g=i.createCharSequence(e,t.originalStart,t.originalStart+t.originalLength-1),f=n.createCharSequence(e,t.modifiedStart,t.modifiedStart+t.modifiedLength-1);if(g.getElements().length>0&&f.getElements().length>0){let p=hoe(g,f,o,!0).changes;a&&(p=KMe(p)),u=[];for(let _=0,b=p.length;_<b;_++)u.push(Ew.createFromDiffChange(p[_],g,f))}}return new ny(l,c,d,h,u)}}class YMe{constructor(e,t,i){this.shouldComputeCharChanges=i.shouldComputeCharChanges,this.shouldPostProcessCharChanges=i.shouldPostProcessCharChanges,this.shouldIgnoreTrimWhitespace=i.shouldIgnoreTrimWhitespace,this.shouldMakePrettyDiff=i.shouldMakePrettyDiff,this.originalLines=e,this.modifiedLines=t,this.original=new PG(e),this.modified=new PG(t),this.continueLineDiff=FG(i.maxComputationTime),this.continueCharDiff=FG(i.maxComputationTime===0?0:Math.min(i.maxComputationTime,5e3))}computeDiff(){if(this.original.lines.length===1&&this.original.lines[0].length===0)return this.modified.lines.length===1&&this.modified.lines[0].length===0?{quitEarly:!1,changes:[]}:{quitEarly:!1,changes:[{originalStartLineNumber:1,originalEndLineNumber:1,modifiedStartLineNumber:1,modifiedEndLineNumber:this.modified.lines.length,charChanges:void 0}]};if(this.modified.lines.length===1&&this.modified.lines[0].length===0)return{quitEarly:!1,changes:[{originalStartLineNumber:1,originalEndLineNumber:this.original.lines.length,modifiedStartLineNumber:1,modifiedEndLineNumber:1,charChanges:void 0}]};const e=hoe(this.original,this.modified,this.continueLineDiff,this.shouldMakePrettyDiff),t=e.changes,i=e.quitEarly;if(this.shouldIgnoreTrimWhitespace){const a=[];for(let l=0,c=t.length;l<c;l++)a.push(ny.createFromDiffResult(this.shouldIgnoreTrimWhitespace,t[l],this.original,this.modified,this.continueCharDiff,this.shouldComputeCharChanges,this.shouldPostProcessCharChanges));return{quitEarly:i,changes:a}}const n=[];let o=0,r=0;for(let a=-1,l=t.length;a<l;a++){const c=a+1<l?t[a+1]:null,d=c?c.originalStart:this.originalLines.length,h=c?c.modifiedStart:this.modifiedLines.length;for(;o<d&&r<h;){const u=this.originalLines[o],g=this.modifiedLines[r];if(u!==g){{let f=f3(u,1),p=f3(g,1);for(;f>1&&p>1;){const _=u.charCodeAt(f-2),b=g.charCodeAt(p-2);if(_!==b)break;f--,p--}(f>1||p>1)&&this._pushTrimWhitespaceCharChange(n,o+1,1,f,r+1,1,p)}{let f=p3(u,1),p=p3(g,1);const _=u.length+1,b=g.length+1;for(;f<_&&p<b;){const w=u.charCodeAt(f-1),C=u.charCodeAt(p-1);if(w!==C)break;f++,p++}(f<_||p<b)&&this._pushTrimWhitespaceCharChange(n,o+1,f,_,r+1,p,b)}}o++,r++}c&&(n.push(ny.createFromDiffResult(this.shouldIgnoreTrimWhitespace,c,this.original,this.modified,this.continueCharDiff,this.shouldComputeCharChanges,this.shouldPostProcessCharChanges)),o+=c.originalLength,r+=c.modifiedLength)}return{quitEarly:i,changes:n}}_pushTrimWhitespaceCharChange(e,t,i,n,o,r,a){if(this._mergeTrimWhitespaceCharChange(e,t,i,n,o,r,a))return;let l;this.shouldComputeCharChanges&&(l=[new Ew(t,i,t,n,o,r,o,a)]),e.push(new ny(t,t,o,o,l))}_mergeTrimWhitespaceCharChange(e,t,i,n,o,r,a){const l=e.length;if(l===0)return!1;const c=e[l-1];return c.originalEndLineNumber===0||c.modifiedEndLineNumber===0?!1:c.originalEndLineNumber===t&&c.modifiedEndLineNumber===o?(this.shouldComputeCharChanges&&c.charChanges&&c.charChanges.push(new Ew(t,i,t,n,o,r,o,a)),!0):c.originalEndLineNumber+1===t&&c.modifiedEndLineNumber+1===o?(c.originalEndLineNumber=t,c.modifiedEndLineNumber=o,this.shouldComputeCharChanges&&c.charChanges&&c.charChanges.push(new Ew(t,i,t,n,o,r,o,a)),!0):!1}}function f3(s,e){const t=As(s);return t===-1?e:t+1}function p3(s,e){const t=Ma(s);return t===-1?e:t+2}function FG(s){if(s===0)return()=>!0;const e=Date.now();return()=>Date.now()-e<s}class ZMe{compute(e,t,i=OS.instance,n){if(e.length===0||t.length===0)return rh.trivial(e,t);const o=new FM(e.length,t.length),r=new FM(e.length,t.length),a=new FM(e.length,t.length);for(let f=0;f<e.length;f++)for(let p=0;p<t.length;p++){if(!i.isValid())return rh.trivialTimedOut(e,t);const _=f===0?0:o.get(f-1,p),b=p===0?0:o.get(f,p-1);let w;e.getElement(f)===t.getElement(p)?(f===0||p===0?w=0:w=o.get(f-1,p-1),f>0&&p>0&&r.get(f-1,p-1)===3&&(w+=a.get(f-1,p-1)),w+=n?n(f,p):1):w=-1;const C=Math.max(_,b,w);if(C===w){const v=f>0&&p>0?a.get(f-1,p-1):0;a.set(f,p,v+1),r.set(f,p,3)}else C===_?(a.set(f,p,0),r.set(f,p,1)):C===b&&(a.set(f,p,0),r.set(f,p,2));o.set(f,p,C)}const l=[];let c=e.length,d=t.length;function h(f,p){(f+1!==c||p+1!==d)&&l.push(new fn(new me(f+1,c),new me(p+1,d))),c=f,d=p}let u=e.length-1,g=t.length-1;for(;u>=0&&g>=0;)r.get(u,g)===3?(h(u,g),u--,g--):r.get(u,g)===1?u--:g--;return h(-1,-1),l.reverse(),new rh(l,!1)}}class uoe{compute(e,t,i=OS.instance){if(e.length===0||t.length===0)return rh.trivial(e,t);const n=e,o=t;function r(p,_){for(;p<n.length&&_<o.length&&n.getElement(p)===o.getElement(_);)p++,_++;return p}let a=0;const l=new XMe;l.set(0,r(0,0));const c=new QMe;c.set(0,l.get(0)===0?null:new BG(null,0,0,l.get(0)));let d=0;e:for(;;){if(a++,!i.isValid())return rh.trivialTimedOut(n,o);const p=-Math.min(a,o.length+a%2),_=Math.min(a,n.length+a%2);for(d=p;d<=_;d+=2){const b=d===_?-1:l.get(d+1),w=d===p?-1:l.get(d-1)+1,C=Math.min(Math.max(b,w),n.length),v=C-d;if(C>n.length||v>o.length)continue;const y=r(C,v);l.set(d,y);const x=C===b?c.get(d+1):c.get(d-1);if(c.set(d,y!==C?new BG(x,C,v,y-C):x),l.get(d)===n.length&&l.get(d)-d===o.length)break e}}let h=c.get(d);const u=[];let g=n.length,f=o.length;for(;;){const p=h?h.x+h.length:0,_=h?h.y+h.length:0;if((p!==g||_!==f)&&u.push(new fn(new me(p,g),new me(_,f))),!h)break;g=h.x,f=h.y,h=h.prev}return u.reverse(),new rh(u,!1)}}class BG{constructor(e,t,i,n){this.prev=e,this.x=t,this.y=i,this.length=n}}class XMe{constructor(){this.positiveArr=new Int32Array(10),this.negativeArr=new Int32Array(10)}get(e){return e<0?(e=-e-1,this.negativeArr[e]):this.positiveArr[e]}set(e,t){if(e<0){if(e=-e-1,e>=this.negativeArr.length){const i=this.negativeArr;this.negativeArr=new Int32Array(i.length*2),this.negativeArr.set(i)}this.negativeArr[e]=t}else{if(e>=this.positiveArr.length){const i=this.positiveArr;this.positiveArr=new Int32Array(i.length*2),this.positiveArr.set(i)}this.positiveArr[e]=t}}}class QMe{constructor(){this.positiveArr=[],this.negativeArr=[]}get(e){return e<0?(e=-e-1,this.negativeArr[e]):this.positiveArr[e]}set(e,t){e<0?(e=-e-1,this.negativeArr[e]=t):this.positiveArr[e]=t}}class MD{constructor(e,t,i){this.lines=e,this.range=t,this.considerWhitespaceChanges=i,this.elements=[],this.firstElementOffsetByLineIdx=[],this.lineStartOffsets=[],this.trimmedWsLengthsByLineIdx=[],this.firstElementOffsetByLineIdx.push(0);for(let n=this.range.startLineNumber;n<=this.range.endLineNumber;n++){let o=e[n-1],r=0;n===this.range.startLineNumber&&this.range.startColumn>1&&(r=this.range.startColumn-1,o=o.substring(r)),this.lineStartOffsets.push(r);let a=0;if(!i){const c=o.trimStart();a=o.length-c.length,o=c.trimEnd()}this.trimmedWsLengthsByLineIdx.push(a);const l=n===this.range.endLineNumber?Math.min(this.range.endColumn-1-r-a,o.length):o.length;for(let c=0;c<l;c++)this.elements.push(o.charCodeAt(c));n<this.range.endLineNumber&&(this.elements.push(10),this.firstElementOffsetByLineIdx.push(this.elements.length))}}toString(){return`Slice: "${this.text}"`}get text(){return this.getText(new me(0,this.length))}getText(e){return this.elements.slice(e.start,e.endExclusive).map(t=>String.fromCharCode(t)).join("")}getElement(e){return this.elements[e]}get length(){return this.elements.length}getBoundaryScore(e){const t=VG(e>0?this.elements[e-1]:-1),i=VG(e<this.elements.length?this.elements[e]:-1);if(t===7&&i===8)return 0;if(t===8)return 150;let n=0;return t!==i&&(n+=10,t===0&&i===1&&(n+=1)),n+=HG(t),n+=HG(i),n}translateOffset(e,t="right"){const i=pv(this.firstElementOffsetByLineIdx,o=>o<=e),n=e-this.firstElementOffsetByLineIdx[i];return new A(this.range.startLineNumber+i,1+this.lineStartOffsets[i]+n+(n===0&&t==="left"?0:this.trimmedWsLengthsByLineIdx[i]))}translateRange(e){const t=this.translateOffset(e.start,"right"),i=this.translateOffset(e.endExclusive,"left");return i.isBefore(t)?k.fromPositions(i,i):k.fromPositions(t,i)}findWordContaining(e){if(e<0||e>=this.elements.length||!eb(this.elements[e]))return;let t=e;for(;t>0&&eb(this.elements[t-1]);)t--;let i=e;for(;i<this.elements.length&&eb(this.elements[i]);)i++;return new me(t,i)}findSubWordContaining(e){if(e<0||e>=this.elements.length||!eb(this.elements[e]))return;let t=e;for(;t>0&&eb(this.elements[t-1])&&!WG(this.elements[t]);)t--;let i=e;for(;i<this.elements.length&&eb(this.elements[i])&&!WG(this.elements[i]);)i++;return new me(t,i)}countLinesIn(e){return this.translateOffset(e.endExclusive).lineNumber-this.translateOffset(e.start).lineNumber}isStronglyEqual(e,t){return this.elements[e]===this.elements[t]}extendToFullLines(e){const t=fv(this.firstElementOffsetByLineIdx,n=>n<=e.start)??0,i=Wwe(this.firstElementOffsetByLineIdx,n=>e.endExclusive<=n)??this.elements.length;return new me(t,i)}}function eb(s){return s>=97&&s<=122||s>=65&&s<=90||s>=48&&s<=57}function WG(s){return s>=65&&s<=90}const JMe={0:0,1:0,2:0,3:10,4:2,5:30,6:3,7:10,8:10};function HG(s){return JMe[s]}function VG(s){return s===10?8:s===13?7:f4(s)?6:s>=97&&s<=122?0:s>=65&&s<=90?1:s>=48&&s<=57?2:s===-1?3:s===44||s===59?5:4}function eOe(s,e,t,i,n,o){let{moves:r,excludedChanges:a}=iOe(s,e,t,o);if(!o.isValid())return[];const l=s.filter(d=>!a.has(d)),c=nOe(l,i,n,e,t,o);return wP(r,c),r=sOe(r),r=r.filter(d=>{const h=d.original.toOffsetRange().slice(e).map(g=>g.trim());return h.join(`
`).length>=15&&tOe(h,g=>g.length>=2)>=2}),r=oOe(s,r),r}function tOe(s,e){let t=0;for(const i of s)e(i)&&t++;return t}function iOe(s,e,t,i){const n=[],o=s.filter(l=>l.modified.isEmpty&&l.original.length>=3).map(l=>new LI(l.original,e,l)),r=new Set(s.filter(l=>l.original.isEmpty&&l.modified.length>=3).map(l=>new LI(l.modified,t,l))),a=new Set;for(const l of o){let c=-1,d;for(const h of r){const u=l.computeSimilarity(h);u>c&&(c=u,d=h)}if(c>.9&&d&&(r.delete(d),n.push(new Ts(l.range,d.range)),a.add(l.source),a.add(d.source)),!i.isValid())return{moves:n,excludedChanges:a}}return{moves:n,excludedChanges:a}}function nOe(s,e,t,i,n,o){const r=[],a=new zQ;for(const u of s)for(let g=u.original.startLineNumber;g<u.original.endLineNumberExclusive-2;g++){const f=`${e[g-1]}:${e[g+1-1]}:${e[g+2-1]}`;a.add(f,{range:new Ne(g,g+3)})}const l=[];s.sort(ns(u=>u.modified.startLineNumber,Wo));for(const u of s){let g=[];for(let f=u.modified.startLineNumber;f<u.modified.endLineNumberExclusive-2;f++){const p=`${t[f-1]}:${t[f+1-1]}:${t[f+2-1]}`,_=new Ne(f,f+3),b=[];a.forEach(p,({range:w})=>{for(const v of g)if(v.originalLineRange.endLineNumberExclusive+1===w.endLineNumberExclusive&&v.modifiedLineRange.endLineNumberExclusive+1===_.endLineNumberExclusive){v.originalLineRange=new Ne(v.originalLineRange.startLineNumber,w.endLineNumberExclusive),v.modifiedLineRange=new Ne(v.modifiedLineRange.startLineNumber,_.endLineNumberExclusive),b.push(v);return}const C={modifiedLineRange:_,originalLineRange:w};l.push(C),b.push(C)}),g=b}if(!o.isValid())return[]}l.sort(bQ(ns(u=>u.modifiedLineRange.length,Wo)));const c=new Kr,d=new Kr;for(const u of l){const g=u.modifiedLineRange.startLineNumber-u.originalLineRange.startLineNumber,f=c.subtractFrom(u.modifiedLineRange),p=d.subtractFrom(u.originalLineRange).getWithDelta(g),_=f.getIntersection(p);for(const b of _.ranges){if(b.length<3)continue;const w=b,C=b.delta(-g);r.push(new Ts(C,w)),c.addRange(w),d.addRange(C)}}r.sort(ns(u=>u.original.startLineNumber,Wo));const h=new QE(s);for(let u=0;u<r.length;u++){const g=r[u],f=h.findLastMonotonous(x=>x.original.startLineNumber<=g.original.startLineNumber),p=fv(s,x=>x.modified.startLineNumber<=g.modified.startLineNumber),_=Math.max(g.original.startLineNumber-f.original.startLineNumber,g.modified.startLineNumber-p.modified.startLineNumber),b=h.findLastMonotonous(x=>x.original.startLineNumber<g.original.endLineNumberExclusive),w=fv(s,x=>x.modified.startLineNumber<g.modified.endLineNumberExclusive),C=Math.max(b.original.endLineNumberExclusive-g.original.endLineNumberExclusive,w.modified.endLineNumberExclusive-g.modified.endLineNumberExclusive);let v;for(v=0;v<_;v++){const x=g.original.startLineNumber-v-1,L=g.modified.startLineNumber-v-1;if(x>i.length||L>n.length||c.contains(L)||d.contains(x)||!zG(i[x-1],n[L-1],o))break}v>0&&(d.addRange(new Ne(g.original.startLineNumber-v,g.original.startLineNumber)),c.addRange(new Ne(g.modified.startLineNumber-v,g.modified.startLineNumber)));let y;for(y=0;y<C;y++){const x=g.original.endLineNumberExclusive+y,L=g.modified.endLineNumberExclusive+y;if(x>i.length||L>n.length||c.contains(L)||d.contains(x)||!zG(i[x-1],n[L-1],o))break}y>0&&(d.addRange(new Ne(g.original.endLineNumberExclusive,g.original.endLineNumberExclusive+y)),c.addRange(new Ne(g.modified.endLineNumberExclusive,g.modified.endLineNumberExclusive+y))),(v>0||y>0)&&(r[u]=new Ts(new Ne(g.original.startLineNumber-v,g.original.endLineNumberExclusive+y),new Ne(g.modified.startLineNumber-v,g.modified.endLineNumberExclusive+y)))}return r}function zG(s,e,t){if(s.trim()===e.trim())return!0;if(s.length>300&&e.length>300)return!1;const n=new uoe().compute(new MD([s],new k(1,1,1,s.length),!1),new MD([e],new k(1,1,1,e.length),!1),t);let o=0;const r=fn.invert(n.diffs,s.length);for(const d of r)d.seq1Range.forEach(h=>{f4(s.charCodeAt(h))||o++});function a(d){let h=0;for(let u=0;u<s.length;u++)f4(d.charCodeAt(u))||h++;return h}const l=a(s.length>e.length?s:e);return o/l>.6&&l>10}function sOe(s){if(s.length===0)return s;s.sort(ns(t=>t.original.startLineNumber,Wo));const e=[s[0]];for(let t=1;t<s.length;t++){const i=e[e.length-1],n=s[t],o=n.original.startLineNumber-i.original.endLineNumberExclusive,r=n.modified.startLineNumber-i.modified.endLineNumberExclusive;if(o>=0&&r>=0&&o+r<=2){e[e.length-1]=i.join(n);continue}e.push(n)}return e}function oOe(s,e){const t=new QE(s);return e=e.filter(i=>{const n=t.findLastMonotonous(a=>a.original.startLineNumber<i.original.endLineNumberExclusive)||new Ts(new Ne(1,1),new Ne(1,1)),o=fv(s,a=>a.modified.startLineNumber<i.modified.endLineNumberExclusive);return n!==o}),e}function $G(s,e,t){let i=t;return i=UG(s,e,i),i=UG(s,e,i),i=rOe(s,e,i),i}function UG(s,e,t){if(t.length===0)return t;const i=[];i.push(t[0]);for(let o=1;o<t.length;o++){const r=i[i.length-1];let a=t[o];if(a.seq1Range.isEmpty||a.seq2Range.isEmpty){const l=a.seq1Range.start-r.seq1Range.endExclusive;let c;for(c=1;c<=l&&!(s.getElement(a.seq1Range.start-c)!==s.getElement(a.seq1Range.endExclusive-c)||e.getElement(a.seq2Range.start-c)!==e.getElement(a.seq2Range.endExclusive-c));c++);if(c--,c===l){i[i.length-1]=new fn(new me(r.seq1Range.start,a.seq1Range.endExclusive-l),new me(r.seq2Range.start,a.seq2Range.endExclusive-l));continue}a=a.delta(-c)}i.push(a)}const n=[];for(let o=0;o<i.length-1;o++){const r=i[o+1];let a=i[o];if(a.seq1Range.isEmpty||a.seq2Range.isEmpty){const l=r.seq1Range.start-a.seq1Range.endExclusive;let c;for(c=0;c<l&&!(!s.isStronglyEqual(a.seq1Range.start+c,a.seq1Range.endExclusive+c)||!e.isStronglyEqual(a.seq2Range.start+c,a.seq2Range.endExclusive+c));c++);if(c===l){i[o+1]=new fn(new me(a.seq1Range.start+l,r.seq1Range.endExclusive),new me(a.seq2Range.start+l,r.seq2Range.endExclusive));continue}c>0&&(a=a.delta(c))}n.push(a)}return i.length>0&&n.push(i[i.length-1]),n}function rOe(s,e,t){if(!s.getBoundaryScore||!e.getBoundaryScore)return t;for(let i=0;i<t.length;i++){const n=i>0?t[i-1]:void 0,o=t[i],r=i+1<t.length?t[i+1]:void 0,a=new me(n?n.seq1Range.endExclusive+1:0,r?r.seq1Range.start-1:s.length),l=new me(n?n.seq2Range.endExclusive+1:0,r?r.seq2Range.start-1:e.length);o.seq1Range.isEmpty?t[i]=qG(o,s,e,a,l):o.seq2Range.isEmpty&&(t[i]=qG(o.swap(),e,s,l,a).swap())}return t}function qG(s,e,t,i,n){let r=1;for(;s.seq1Range.start-r>=i.start&&s.seq2Range.start-r>=n.start&&t.isStronglyEqual(s.seq2Range.start-r,s.seq2Range.endExclusive-r)&&r<100;)r++;r--;let a=0;for(;s.seq1Range.start+a<i.endExclusive&&s.seq2Range.endExclusive+a<n.endExclusive&&t.isStronglyEqual(s.seq2Range.start+a,s.seq2Range.endExclusive+a)&&a<100;)a++;if(r===0&&a===0)return s;let l=0,c=-1;for(let d=-r;d<=a;d++){const h=s.seq2Range.start+d,u=s.seq2Range.endExclusive+d,g=s.seq1Range.start+d,f=e.getBoundaryScore(g)+t.getBoundaryScore(h)+t.getBoundaryScore(u);f>c&&(c=f,l=d)}return s.delta(l)}function aOe(s,e,t){const i=[];for(const n of t){const o=i[i.length-1];if(!o){i.push(n);continue}n.seq1Range.start-o.seq1Range.endExclusive<=2||n.seq2Range.start-o.seq2Range.endExclusive<=2?i[i.length-1]=new fn(o.seq1Range.join(n.seq1Range),o.seq2Range.join(n.seq2Range)):i.push(n)}return i}function jG(s,e,t,i,n=!1){const o=fn.invert(t,s.length),r=[];let a=new qd(0,0);function l(d,h){if(d.offset1<a.offset1||d.offset2<a.offset2)return;const u=i(s,d.offset1),g=i(e,d.offset2);if(!u||!g)return;let f=new fn(u,g);const p=f.intersect(h);let _=p.seq1Range.length,b=p.seq2Range.length;for(;o.length>0;){const w=o[0];if(!(w.seq1Range.intersects(f.seq1Range)||w.seq2Range.intersects(f.seq2Range)))break;const v=i(s,w.seq1Range.start),y=i(e,w.seq2Range.start),x=new fn(v,y),L=x.intersect(w);if(_+=L.seq1Range.length,b+=L.seq2Range.length,f=f.join(x),f.seq1Range.endExclusive>=w.seq1Range.endExclusive)o.shift();else break}(n&&_+b<f.seq1Range.length+f.seq2Range.length||_+b<(f.seq1Range.length+f.seq2Range.length)*2/3)&&r.push(f),a=f.getEndExclusives()}for(;o.length>0;){const d=o.shift();d.seq1Range.isEmpty||(l(d.getStarts(),d),l(d.getEndExclusives().delta(-1),d))}return lOe(t,r)}function lOe(s,e){const t=[];for(;s.length>0||e.length>0;){const i=s[0],n=e[0];let o;i&&(!n||i.seq1Range.start<n.seq1Range.start)?o=s.shift():o=e.shift(),t.length>0&&t[t.length-1].seq1Range.endExclusive>=o.seq1Range.start?t[t.length-1]=t[t.length-1].join(o):t.push(o)}return t}function cOe(s,e,t){let i=t;if(i.length===0)return i;let n=0,o;do{o=!1;const r=[i[0]];for(let a=1;a<i.length;a++){let d=function(u,g){const f=new me(c.seq1Range.endExclusive,l.seq1Range.start);return s.getText(f).replace(/\s/g,"").length<=4&&(u.seq1Range.length+u.seq2Range.length>5||g.seq1Range.length+g.seq2Range.length>5)};const l=i[a],c=r[r.length-1];d(c,l)?(o=!0,r[r.length-1]=r[r.length-1].join(l)):r.push(l)}i=r}while(n++<10&&o);return i}function dOe(s,e,t){let i=t;if(i.length===0)return i;let n=0,o;do{o=!1;const a=[i[0]];for(let l=1;l<i.length;l++){let h=function(g,f){const p=new me(d.seq1Range.endExclusive,c.seq1Range.start);if(s.countLinesIn(p)>5||p.length>500)return!1;const b=s.getText(p).trim();if(b.length>20||b.split(/\r\n|\r|\n/).length>1)return!1;const w=s.countLinesIn(g.seq1Range),C=g.seq1Range.length,v=e.countLinesIn(g.seq2Range),y=g.seq2Range.length,x=s.countLinesIn(f.seq1Range),L=f.seq1Range.length,E=e.countLinesIn(f.seq2Range),I=f.seq2Range.length,R=130;function O(P){return Math.min(P,R)}return Math.pow(Math.pow(O(w*40+C),1.5)+Math.pow(O(v*40+y),1.5),1.5)+Math.pow(Math.pow(O(x*40+L),1.5)+Math.pow(O(E*40+I),1.5),1.5)>(R**1.5)**1.5*1.3};const c=i[l],d=a[a.length-1];h(d,c)?(o=!0,a[a.length-1]=a[a.length-1].join(c)):a.push(c)}i=a}while(n++<10&&o);const r=[];return Lhe(i,(a,l,c)=>{let d=l;function h(b){return b.length>0&&b.trim().length<=3&&l.seq1Range.length+l.seq2Range.length>100}const u=s.extendToFullLines(l.seq1Range),g=s.getText(new me(u.start,l.seq1Range.start));h(g)&&(d=d.deltaStart(-g.length));const f=s.getText(new me(l.seq1Range.endExclusive,u.endExclusive));h(f)&&(d=d.deltaEnd(f.length));const p=fn.fromOffsetPairs(a?a.getEndExclusives():qd.zero,c?c.getStarts():qd.max),_=d.intersect(p);r.length>0&&_.getStarts().equals(r[r.length-1].getEndExclusives())?r[r.length-1]=r[r.length-1].join(_):r.push(_)}),r}class GG{constructor(e,t){this.trimmedHash=e,this.lines=t}getElement(e){return this.trimmedHash[e]}get length(){return this.trimmedHash.length}getBoundaryScore(e){const t=e===0?0:KG(this.lines[e-1]),i=e===this.lines.length?0:KG(this.lines[e]);return 1e3-(t+i)}getText(e){return this.lines.slice(e.start,e.endExclusive).join(`
`)}isStronglyEqual(e,t){return this.lines[e]===this.lines[t]}}function KG(s){let e=0;for(;e<s.length&&(s.charCodeAt(e)===32||s.charCodeAt(e)===9);)e++;return e}class hOe{constructor(){this.dynamicProgrammingDiffing=new ZMe,this.myersDiffingAlgorithm=new uoe}computeDiff(e,t,i){if(e.length<=1&&ni(e,t,(L,E)=>L===E))return new Yk([],[],!1);if(e.length===1&&e[0].length===0||t.length===1&&t[0].length===0)return new Yk([new Lr(new Ne(1,e.length+1),new Ne(1,t.length+1),[new Ks(new k(1,1,e.length,e[e.length-1].length+1),new k(1,1,t.length,t[t.length-1].length+1))])],[],!1);const n=i.maxComputationTimeMs===0?OS.instance:new dke(i.maxComputationTimeMs),o=!i.ignoreTrimWhitespace,r=new Map;function a(L){let E=r.get(L);return E===void 0&&(E=r.size,r.set(L,E)),E}const l=e.map(L=>a(L.trim())),c=t.map(L=>a(L.trim())),d=new GG(l,e),h=new GG(c,t),u=d.length+h.length<1700?this.dynamicProgrammingDiffing.compute(d,h,n,(L,E)=>e[L]===t[E]?t[E].length===0?.1:1+Math.log(1+t[E].length):.99):this.myersDiffingAlgorithm.compute(d,h,n);let g=u.diffs,f=u.hitTimeout;g=$G(d,h,g),g=cOe(d,h,g);const p=[],_=L=>{if(o)for(let E=0;E<L;E++){const I=b+E,R=w+E;if(e[I]!==t[R]){const O=this.refineDiff(e,t,new fn(new me(I,I+1),new me(R,R+1)),n,o,i);for(const P of O.mappings)p.push(P);O.hitTimeout&&(f=!0)}}};let b=0,w=0;for(const L of g){rg(()=>L.seq1Range.start-b===L.seq2Range.start-w);const E=L.seq1Range.start-b;_(E),b=L.seq1Range.endExclusive,w=L.seq2Range.endExclusive;const I=this.refineDiff(e,t,L,n,o,i);I.hitTimeout&&(f=!0);for(const R of I.mappings)p.push(R)}_(e.length-b);const C=new eL(e),v=new eL(t),y=xI(p,C,v);let x=[];return i.computeMoves&&(x=this.computeMoves(y,e,t,l,c,n,o,i)),rg(()=>{function L(I,R){if(I.lineNumber<1||I.lineNumber>R.length)return!1;const O=R[I.lineNumber-1];return!(I.column<1||I.column>O.length+1)}function E(I,R){return!(I.startLineNumber<1||I.startLineNumber>R.length+1||I.endLineNumberExclusive<1||I.endLineNumberExclusive>R.length+1)}for(const I of y){if(!I.innerChanges)return!1;for(const R of I.innerChanges)if(!(L(R.modifiedRange.getStartPosition(),t)&&L(R.modifiedRange.getEndPosition(),t)&&L(R.originalRange.getStartPosition(),e)&&L(R.originalRange.getEndPosition(),e)))return!1;if(!E(I.modified,t)||!E(I.original,e))return!1}return!0}),new Yk(y,x,f)}computeMoves(e,t,i,n,o,r,a,l){return eOe(e,t,i,n,o,r).map(h=>{const u=this.refineDiff(t,i,new fn(h.original.toOffsetRange(),h.modified.toOffsetRange()),r,a,l),g=xI(u.mappings,new eL(t),new eL(i),!0);return new doe(h,g)})}refineDiff(e,t,i,n,o,r){const l=uOe(i).toRangeMapping2(e,t),c=new MD(e,l.originalRange,o),d=new MD(t,l.modifiedRange,o),h=c.length+d.length<500?this.dynamicProgrammingDiffing.compute(c,d,n):this.myersDiffingAlgorithm.compute(c,d,n);let u=h.diffs;return u=$G(c,d,u),u=jG(c,d,u,(f,p)=>f.findWordContaining(p)),r.extendToSubwords&&(u=jG(c,d,u,(f,p)=>f.findSubWordContaining(p),!0)),u=aOe(c,d,u),u=dOe(c,d,u),{mappings:u.map(f=>new Ks(c.translateRange(f.seq1Range),d.translateRange(f.seq2Range))),hitTimeout:h.hitTimeout}}}function uOe(s){return new Ts(new Ne(s.seq1Range.start+1,s.seq1Range.endExclusive+1),new Ne(s.seq2Range.start+1,s.seq2Range.endExclusive+1))}const m3={getLegacy:()=>new jMe,getDefault:()=>new hOe};var _3;(function(s){function e(t,i){return!t.isInlineEdit&&!t.uri?gOe.create(t,i):Iw.create(t,i)}s.create=e})(_3||(_3={}));class goe{constructor(e,t,i){this._data=e,this.identity=t,this.hint=i}get source(){return this._data.source}get isFromExplicitRequest(){return this._data.context.triggerKind===ho.Explicit}get forwardStable(){return this.source.inlineSuggestions.enableForwardStability??!1}get editRange(){return this.getSingleTextEdit().range}get targetRange(){return this.hint?.range&&!this.hint.jumpToEdit?this.hint?.range:this.editRange}get insertText(){return this.getSingleTextEdit().text}get semanticId(){return this.hash}get action(){return this._sourceInlineCompletion.gutterMenuLinkAction}get command(){return this._sourceInlineCompletion.command}get warning(){return this._sourceInlineCompletion.warning}get showInlineEditMenu(){return!!this._sourceInlineCompletion.showInlineEditMenu}get hash(){return JSON.stringify([this.getSingleTextEdit().text,this.getSingleTextEdit().range.getStartPosition().toString()])}get requestUuid(){return this._data.context.requestUuid}get partialAccepts(){return this._data.partialAccepts}get _sourceInlineCompletion(){return this._data.sourceInlineCompletion}addRef(){this.identity.addRef(),this.source.addRef()}removeRef(){this.identity.removeRef(),this.source.removeRef()}reportInlineEditShown(e,t,i){this._data.reportInlineEditShown(e,this.insertText,t,i)}reportPartialAccept(e,t,i){this._data.reportPartialAccept(e,t,i)}reportEndOfLife(e){this._data.reportEndOfLife(e)}setEndOfLifeReason(e){this._data.setEndOfLifeReason(e)}setIsPreceeded(e){this._data.setIsPreceeded(e.partialAccepts)}setNotShownReasonIfNotSet(e){this._data.setNotShownReason(e)}getSourceCompletion(){return this._sourceInlineCompletion}}const AN=class AN{constructor(){this._onDispose=Yr(this),this._jumpedTo=Re(this,!1),this._refCount=1,this.id="InlineCompletionIdentity"+AN.idCounter++}get jumpedTo(){return this._jumpedTo}addRef(){this._refCount++}removeRef(){this._refCount--,this._refCount===0&&this._onDispose.trigger(void 0)}setJumpTo(e){this._jumpedTo.set(!0,e)}};AN.idCounter=0;let OD=AN;class a0{static create(e){return new a0(k.lift(e.range),e.content,e.style,e.jumpToEdit)}constructor(e,t,i,n){this.range=e,this.content=t,this.style=i,this.jumpToEdit=n}withEdit(e,t){const i=new me(t.getOffset(this.range.getStartPosition()),t.getOffset(this.range.getEndPosition())),n=Fee([i],e)[0];if(!n)return;const o=t.getRange(n);return new a0(o,this.content,this.style,this.jumpToEdit)}}let gOe=class Zk extends goe{static create(e,t){const i=new OD,n=o0(t),o=e.insertText.replace(/\r\n|\r|\n/g,t.getEOL()),r=pOe(new On(n.getOffsetRange(e.range),o),t),a=r.removeCommonSuffixAndPrefix(t.getValue()),l=n.getTextReplacement(r),c=e.hint?a0.create(e.hint):void 0;return new Zk(r,a,l,l.range,e.snippetInfo,e.additionalTextEdits,e,i,c)}constructor(e,t,i,n,o,r,a,l,c){super(a,l,c),this._edit=e,this._trimmedEdit=t,this._textEdit=i,this._originalRange=n,this.snippetInfo=o,this.additionalTextEdits=r,this.isInlineEdit=!1}get hash(){return JSON.stringify(this._trimmedEdit.toJson())}getSingleTextEdit(){return this._textEdit}withIdentity(e){return new Zk(this._edit,this._trimmedEdit,this._textEdit,this._originalRange,this.snippetInfo,this.additionalTextEdits,this._data,e,this.hint)}withEdit(e,t){const i=Fee([this._edit.replaceRange],e);if(i.length===0)return;const n=new On(i[0],this._textEdit.text),o=o0(t),r=o.getTextReplacement(n);let a=this.hint;if(a&&(a=a.withEdit(e,o),!a))return;const l=n.removeCommonSuffixAndPrefix(t.getValue());return new Zk(n,l,r,this._originalRange,this.snippetInfo,this.additionalTextEdits,this._data,this.identity,a)}canBeReused(e,t){const i=this._textEdit.range;return!!i&&i.containsPosition(t)&&this.isVisible(e,t)&&cn.ofRange(i).isGreaterThanOrEqualTo(cn.ofRange(this._originalRange))}isVisible(e,t){const i=this.getSingleTextEdit();return foe(i,this._originalRange,e,t)}};function foe(s,e,t,i){const n=jd(s,t),o=s.range;if(!o||e&&!e.getStartPosition().equals(o.getStartPosition())||i.lineNumber!==n.range.startLineNumber||n.isEmpty)return!1;const r=t.getValueInRange(n.range,1),a=n.text,l=Math.max(0,i.column-n.range.startColumn);let c=a.substring(0,l),d=a.substring(l),h=r.substring(0,l),u=r.substring(l);const g=t.getLineIndentColumn(n.range.startLineNumber);return n.range.startColumn<=g&&(h=h.trimStart(),h.length===0&&(u=u.trimStart()),c=c.trimStart(),c.length===0&&(d=d.trimStart())),c.startsWith(h)&&!!yte(u,d)}class Iw extends goe{static create(e,t){const i=fOe(t,e.range,e.insertText),n=new Fm(t),o=Na.fromStringEdit(i,n),r=i.isEmpty()?new Ui(new k(1,1,1,1),""):o.toReplacement(n),a=new OD,l=i.replacements.map(d=>{const h=k.fromPositions(t.getPositionAt(d.replaceRange.start),t.getPositionAt(d.replaceRange.endExclusive)),u=t.getValueInRange(h);return PD.create(d,u)}),c=e.hint?a0.create(e.hint):void 0;return new Iw(i,r,e.uri,e,a,l,c,!1,t.getVersionId())}constructor(e,t,i,n,o,r,a,l=!1,c){super(n,o,a),this._edit=e,this._textEdit=t,this.uri=i,this._edits=r,this._lastChangePartOfInlineEdit=l,this._inlineEditModelVersion=c,this.snippetInfo=void 0,this.additionalTextEdits=[],this.isInlineEdit=!0}get updatedEditModelVersion(){return this._inlineEditModelVersion}get updatedEdit(){return this._edit}getSingleTextEdit(){return this._textEdit}withIdentity(e){return new Iw(this._edit,this._textEdit,this.uri,this._data,e,this._edits,this.hint,this._lastChangePartOfInlineEdit,this._inlineEditModelVersion)}canBeReused(e,t){return this._lastChangePartOfInlineEdit&&this.updatedEditModelVersion===e.getVersionId()}withEdit(e,t){return this._applyTextModelChanges(e,this._edits,t)}_applyTextModelChanges(e,t,i){if(t=t.map(h=>h.applyTextModelChanges(e)),t.some(h=>h.edit===void 0))return;const n=i.getVersionId();let o=this._inlineEditModelVersion;const r=t.some(h=>h.lastChangeUpdatedEdit);if(r&&(o=n??-1),n===null||o+20<n||(t=t.filter(h=>!h.edit.isEmpty),t.length===0))return;const a=new vh(t.map(h=>h.edit)),l=o0(i),c=l.getTextEdit(a).toReplacement(new Fm(i));let d=this.hint;if(!(d&&(d=d.withEdit(e,l),!d)))return new Iw(a,c,this.uri,this._data,this.identity,t,d,r,o)}}function fOe(s,e,t){const i=s.getEOL(),n=s.getValueInRange(e),o=t.replace(/\r\n|\r|\n/g,i),l=m3.getDefault().computeDiff(qo(n),qo(o),{ignoreTrimWhitespace:!1,computeMoves:!1,extendToSubwords:!0,maxComputationTimeMs:500}).changes.flatMap(u=>u.innerChanges??[]);function c(u,g){const f=cn.fromPosition(g.getStartPosition());return cn.ofRange(g).createRange(f.addToPosition(u))}const d=new vu(o);return new vh(l.map(u=>{const g=c(e.getStartPosition(),u.originalRange),f=o0(s).getOffsetRange(g),p=d.getValueOfRange(u.modifiedRange),_=new On(f,p),b=s.getValueInRange(g);return mOe(_,b,l.length,s)}))}class PD{static create(e,t){const i=Fa(e.newText,t),n=uh(e.newText,t),o=e.newText.substring(i,e.newText.length-n);return new PD(e,o,i,n)}get edit(){return this._edit}get lastChangeUpdatedEdit(){return this._lastChangeUpdatedEdit}constructor(e,t,i,n,o=!1){this._edit=e,this._trimmedNewText=t,this._prefixLength=i,this._suffixLength=n,this._lastChangeUpdatedEdit=o}applyTextModelChanges(e){const t=this._clone();return t._applyTextModelChanges(e),t}_clone(){return new PD(this._edit,this._trimmedNewText,this._prefixLength,this._suffixLength,this._lastChangeUpdatedEdit)}_applyTextModelChanges(e){if(this._lastChangeUpdatedEdit=!1,!this._edit)throw new Ce("UpdatedInnerEdits: No edit to apply changes to");const t=this._applyChanges(this._edit,e);if(!t){this._edit=void 0;return}this._edit=t.edit,this._lastChangeUpdatedEdit=t.editHasChanged}_applyChanges(e,t){let i=e.replaceRange.start,n=e.replaceRange.endExclusive,o=e.newText,r=!1;const a=this._prefixLength>0||this._suffixLength>0;for(let l=t.replacements.length-1;l>=0;l--){const c=t.replacements[l],d=c.newText.length>0&&c.replaceRange.isEmpty;if(d&&!a&&c.replaceRange.start===i&&o.startsWith(c.newText)){i+=c.newText.length,o=o.substring(c.newText.length),n=Math.max(i,n),r=!0;continue}if(d&&a&&c.replaceRange.start===i+this._prefixLength&&this._trimmedNewText.startsWith(c.newText)){n+=c.newText.length,r=!0,this._prefixLength+=c.newText.length,this._trimmedNewText=this._trimmedNewText.substring(c.newText.length);continue}if(c.newText.length===0&&c.replaceRange.length>0&&c.replaceRange.start>=i+this._prefixLength&&c.replaceRange.endExclusive<=n-this._suffixLength){n-=c.replaceRange.length,r=!0;continue}if(c.equals(e)){r=!0,i=c.replaceRange.endExclusive,o="";continue}if(!(c.replaceRange.start>n)){if(c.replaceRange.endExclusive<i){i+=c.newText.length-c.replaceRange.length,n+=c.newText.length-c.replaceRange.length;continue}return}}return this._trimmedNewText.length===0&&i+this._prefixLength===n-this._suffixLength?{edit:new On(new me(i+this._prefixLength,i+this._prefixLength),""),editHasChanged:!0}:{edit:new On(new me(i,n),o),editHasChanged:r}}}function pOe(s,e){const t=e.getEOL();return s.replaceRange.isEmpty&&s.newText.includes(t)&&(s=poe(s,e)),s}function mOe(s,e,t,i){const n=i.getEOL();if(s.newText.endsWith(n)&&e.endsWith(n)&&(s=new On(s.replaceRange.deltaEnd(-n.length),s.newText.slice(0,-n.length))),t===1&&s.replaceRange.isEmpty&&s.newText.includes(n)){const o=i.getPositionAt(s.replaceRange.start);i.getLineLength(o.lineNumber)!==0&&(s=poe(s,i))}if(t===1){const o=Fa(e,s.newText),r=uh(e.slice(o),s.newText.slice(o));if(o+r===e.length)return new On(s.replaceRange.deltaStart(o).deltaEnd(-r),s.newText.substring(o,s.newText.length-r));if(o+r===s.newText.length)return new On(s.replaceRange.deltaStart(o).deltaEnd(-r),"")}return s}function poe(s,e){if(!s.replaceRange.isEmpty)throw new Ce("Unexpected original range");if(s.replaceRange.start===0)return s;const t=e.getEOL(),i=e.getPositionAt(s.replaceRange.start),n=i.column,o=i.lineNumber;return n===1&&o>1&&s.newText.endsWith(t)&&!s.newText.startsWith(t)?new On(s.replaceRange.delta(-1),t+s.newText.slice(0,-t.length)):s}function _Oe(s,e){const t=new Uee,i=new jee(t,c=>e.getLanguageConfiguration(c)),n=new qee(new bOe([s]),i),o=DF(n,[],void 0,!0);let r="";const a=s.getLineContent();function l(c,d){if(c.kind===2)if(l(c.openingBracket,d),d=xi(d,c.openingBracket.length),c.child&&(l(c.child,d),d=xi(d,c.child.length)),c.closingBracket)l(c.closingBracket,d),d=xi(d,c.closingBracket.length);else{const u=i.getSingleLanguageBracketTokens(c.openingBracket.languageId).findClosingTokenText(c.openingBracket.bracketIds);r+=u}else if(c.kind!==3){if(c.kind===0||c.kind===1)r+=a.substring(d,xi(d,c.length));else if(c.kind===4)for(const h of c.children)l(h,d),d=xi(d,h.length)}}return l(o,Qs),r}class bOe{constructor(e){this.lines=e,this.tokenization={getLineTokens:t=>this.lines[t-1]}}getLineCount(){return this.lines.length}getLineLength(e){return this.lines[e-1].getLineContent().length}}class GH{constructor(){this._nodes=new Set,this._outgoingEdges=new Map}static from(e,t){const i=new GH;for(const n of e)i._nodes.add(n);for(const n of e){const o=t(n);if(o.length>0){const r=new Set;for(const a of o)r.add(a);i._outgoingEdges.set(n,r)}}return i}removeCycles(){const e=[],t=new Set,i=new Set,n=[],o=r=>{t.add(r),i.add(r);const a=this._outgoingEdges.get(r);if(a)for(const l of a)t.has(l)?i.has(l)&&(e.push(l),n.push({from:r,to:l})):o(l);i.delete(r)};for(const r of this._nodes)t.has(r)||o(r);for(const{from:r,to:a}of n){const l=this._outgoingEdges.get(r);l&&l.delete(a)}return{foundCycles:e}}getOutgoing(e){const t=this._outgoingEdges.get(e);return t?Array.from(t):[]}}var us;(function(s){s.Jump="jump",s.Accept="accept",s.Inactive="inactive"})(us||(us={}));var mt;(function(s){s.GhostText="ghostText",s.Custom="custom",s.SideBySide="sideBySide",s.Deletion="deletion",s.InsertionInline="insertionInline",s.InsertionMultiLine="insertionMultiLine",s.WordReplacements="wordReplacements",s.LineReplacement="lineReplacement",s.Collapsed="collapsed"})(mt||(mt={}));function wOe(s,e,t,i,n,o){const r=oee("icr"),a=new si;let l;const c={...i,requestUuid:r},d=SOe(e,t),h=pye(s,b=>b.groupId),u=GH.from(s,b=>b.yieldsToGroupIds?.flatMap(w=>h.get(w)??[])??[]),{foundCycles:g}=u.removeCycles();g.length>0&&ji(new Error(`Inline completions: cyclic yield-to dependency detected. Path: ${g.map(b=>b.toString?b.toString():""+b).join(" -> ")}`));let f=0;const p=new dP(async b=>{try{if(f++,a.token.isCancellationRequested)return;const w=u.getOutgoing(b);for(const E of w){const I=await p.get(E);if(I)for(const R of I.inlineSuggestions.items){if(R.isInlineEdit||typeof R.insertText!="string"&&R.insertText!==void 0)return;if(R.insertText!==void 0){const O=new Ui(k.lift(R.range)??d,R.insertText);if(foe(O,void 0,t,e))return}}}let C;const v=Date.now();try{C=await b.provideInlineCompletions(t,e,c,a.token)}catch(E){ji(E);return}const y=Date.now();if(!C)return;const x=[],L=new yOe(C,x,b);if(L.addRef(),moe(a.token,()=>L.removeRef(l)),a.token.isCancellationRequested)return;for(const E of C.items)x.push(vOe(E,L,d,t,o,c,n,{startTime:v,endTime:y}));return L}finally{f--}}),_=Jr.fromPromisesResolveOrder(s.map(b=>p.get(b))).filter(Ln);return{contextWithUuid:c,get didAllProvidersReturn(){return f===0},lists:_,cancelAndDispose:b=>{l===void 0&&(l=b,a.dispose(!0))}}}function moe(s,e){if(s.isCancellationRequested)return e(),F.None;{const t=s.onCancellationRequested(()=>{t.dispose(),e()});return{dispose:()=>t.dispose()}}}function vOe(s,e,t,i,n,o,r,a){let l,c,d=s.range?k.lift(s.range):t;if(typeof s.insertText=="string"){if(l=s.insertText,n&&s.completeBracketPairs){l=YG(l,d.getStartPosition(),i,n);const h=l.length-s.insertText.length;h!==0&&(d=new k(d.startLineNumber,d.startColumn,d.endLineNumber,d.endColumn+h))}c=void 0}else if(s.insertText===void 0)l="",c=void 0,d=new k(1,1,1,1);else if("snippet"in s.insertText){const h=s.insertText.snippet.length;if(n&&s.completeBracketPairs){s.insertText.snippet=YG(s.insertText.snippet,d.getStartPosition(),i,n);const g=s.insertText.snippet.length-h;g!==0&&(d=new k(d.startLineNumber,d.startColumn,d.endLineNumber,d.endColumn+g))}const u=new Hm().parse(s.insertText.snippet);u.children.length===1&&u.children[0]instanceof Xs?(l=u.children[0].value,c=void 0):(l=u.toString(),c={snippet:s.insertText.snippet,range:d})}else W0(s.insertText);return new COe(d,l,c,Le.revive(s.uri),s.hint,s.additionalTextEdits||IMe(),s,e,o,s.isInlineEdit??!1,r,a,s.correlationId)}class COe{constructor(e,t,i,n,o,r,a,l,c,d,h,u,g){this.range=e,this.insertText=t,this.snippetInfo=i,this.uri=n,this.hint=o,this.additionalTextEdits=r,this.sourceInlineCompletion=a,this.source=l,this.context=c,this.isInlineEdit=d,this._requestInfo=h,this._providerRequestInfo=u,this._correlationId=g,this._didShow=!1,this._timeUntilShown=void 0,this._showStartTime=void 0,this._shownDuration=0,this._showUncollapsedStartTime=void 0,this._showUncollapsedDuration=0,this._notShownReason=void 0,this._didReportEndOfLife=!1,this._lastSetEndOfLifeReason=void 0,this._isPreceeded=!1,this._partiallyAcceptedCount=0,this._partiallyAcceptedSinceOriginal={characters:0,ratio:0,count:0},this._viewData={editorType:h.editorType}}get showInlineEditMenu(){return this.sourceInlineCompletion.showInlineEditMenu??!1}get partialAccepts(){return this._partiallyAcceptedSinceOriginal}async reportInlineEditShown(e,t,i,n){if(this.updateShownDuration(i),this._didShow)return;this._didShow=!0,this._viewData.viewKind=i,this._viewData.renderData=n,this._timeUntilShown=Date.now()-this._requestInfo.startTime;const o=new ES(n.lineCountModified,n.lineCountOriginal,n.characterCountModified,n.characterCountOriginal);this.source.provider.handleItemDidShow?.(this.source.inlineSuggestions,this.sourceInlineCompletion,t,o),this.sourceInlineCompletion.shownCommand&&await e.executeCommand(this.sourceInlineCompletion.shownCommand.id,...this.sourceInlineCompletion.shownCommand.arguments||[])}reportPartialAccept(e,t,i){this._partiallyAcceptedCount++,this._partiallyAcceptedSinceOriginal.characters+=i.characters,this._partiallyAcceptedSinceOriginal.ratio=Math.min(this._partiallyAcceptedSinceOriginal.ratio+(1-this._partiallyAcceptedSinceOriginal.ratio)*i.ratio,1),this._partiallyAcceptedSinceOriginal.count+=i.count,this.source.provider.handlePartialAccept?.(this.source.inlineSuggestions,this.sourceInlineCompletion,e,t)}reportEndOfLife(e){if(!this._didReportEndOfLife&&(this._didReportEndOfLife=!0,this.reportInlineEditHidden(),e||(e=this._lastSetEndOfLifeReason??{kind:Dm.Ignored,userTypingDisagreed:!1,supersededBy:void 0}),e.kind===Dm.Rejected&&this.source.provider.handleRejection&&this.source.provider.handleRejection(this.source.inlineSuggestions,this.sourceInlineCompletion),this.source.provider.handleEndOfLifetime)){const t={requestUuid:this.context.requestUuid,correlationId:this._correlationId,selectedSuggestionInfo:!!this.context.selectedSuggestionInfo,partiallyAccepted:this._partiallyAcceptedCount,partiallyAcceptedCountSinceOriginal:this._partiallyAcceptedSinceOriginal.count,partiallyAcceptedRatioSinceOriginal:this._partiallyAcceptedSinceOriginal.ratio,partiallyAcceptedCharactersSinceOriginal:this._partiallyAcceptedSinceOriginal.characters,shown:this._didShow,shownDuration:this._shownDuration,shownDurationUncollapsed:this._showUncollapsedDuration,preceeded:this._isPreceeded,timeUntilShown:this._timeUntilShown,timeUntilProviderRequest:this._providerRequestInfo.startTime-this._requestInfo.startTime,timeUntilProviderResponse:this._providerRequestInfo.endTime-this._requestInfo.startTime,editorType:this._viewData.editorType,languageId:this._requestInfo.languageId,requestReason:this._requestInfo.reason,viewKind:this._viewData.viewKind,notShownReason:this._notShownReason,typingInterval:this._requestInfo.typingInterval,typingIntervalCharacterCount:this._requestInfo.typingIntervalCharacterCount,availableProviders:this._requestInfo.availableProviders.map(i=>i.toString()).join(","),...this._viewData.renderData};this.source.provider.handleEndOfLifetime(this.source.inlineSuggestions,this.sourceInlineCompletion,e,t)}}setIsPreceeded(e){this._isPreceeded=!0,(this._partiallyAcceptedSinceOriginal.characters!==0||this._partiallyAcceptedSinceOriginal.ratio!==0||this._partiallyAcceptedSinceOriginal.count!==0)&&console.warn("Expected partiallyAcceptedCountSinceOriginal to be { characters: 0, rate: 0, partialAcceptances: 0 } before setIsPreceeded."),this._partiallyAcceptedSinceOriginal=e}setNotShownReason(e){this._notShownReason??=e}setEndOfLifeReason(e){this.reportInlineEditHidden(),this._lastSetEndOfLifeReason=e}updateShownDuration(e){const t=Date.now();this._showStartTime||(this._showStartTime=t);const i=e===mt.Collapsed;!i&&this._showUncollapsedStartTime===void 0&&(this._showUncollapsedStartTime=t),i&&this._showUncollapsedStartTime!==void 0&&(this._showUncollapsedDuration+=t-this._showUncollapsedStartTime)}reportInlineEditHidden(){if(this._showStartTime===void 0)return;const e=Date.now();this._shownDuration+=e-this._showStartTime,this._showStartTime=void 0,this._showUncollapsedStartTime!==void 0&&(this._showUncollapsedDuration+=e-this._showUncollapsedStartTime,this._showUncollapsedStartTime=void 0)}}var sy;(function(s){s.TextEditor="textEditor",s.DiffEditor="diffEditor",s.Notebook="notebook"})(sy||(sy={}));class yOe{constructor(e,t,i){this.inlineSuggestions=e,this.inlineSuggestionsData=t,this.provider=i,this.refCount=0}addRef(){this.refCount++}removeRef(e={kind:"other"}){if(this.refCount--,this.refCount===0){for(const t of this.inlineSuggestionsData)t.reportEndOfLife();this.provider.disposeInlineCompletions(this.inlineSuggestions,e)}}}function SOe(s,e){const t=e.getWordAtPosition(s),i=e.getLineMaxColumn(s.lineNumber);return t?new k(s.lineNumber,t.startColumn,s.lineNumber,i):k.fromPositions(s,s.with(void 0,i))}function YG(s,e,t,i){const n=t.getLineContent(e.lineNumber),o=On.replace(new me(e.column-1,n.length),s),a=t.tokenization.tokenizeLinesAt(e.lineNumber,[o.replace(n)])?.[0].sliceZeroCopy(o.getRangeAfterReplace());return a?_Oe(a,i):s}var xOe=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},KC=function(s,e){return function(t,i){e(t,i,s)}},b3,Yp;let w3=(Yp=class extends F{constructor(e,t,i,n,o,r,a,l,c){super(),this._textModel=e,this._versionId=t,this._debounceValue=i,this._cursorPosition=n,this._languageConfigurationService=o,this._logService=r,this._configurationService=a,this._instantiationService=l,this._contextKeyService=c,this._updateOperation=this._register(new vt),this._state=WMe(this,{initial:()=>({inlineCompletions:ka.createEmpty(),suggestWidgetInlineCompletions:ka.createEmpty()}),disposeFinal:h=>{h.inlineCompletions.dispose(),h.suggestWidgetInlineCompletions.dispose()},changeTracker:rye(()=>({versionId:this._versionId})),update:(h,u,g)=>{const f=vh.compose(g.changes.map(p=>p.change?$Me(p.change.changes):vh.empty).filter(Ln));if(f.isEmpty())return u;try{return{inlineCompletions:u.inlineCompletions.createStateWithAppliedEdit(f,this._textModel),suggestWidgetInlineCompletions:u.suggestWidgetInlineCompletions.createStateWithAppliedEdit(f,this._textModel)}}finally{u.inlineCompletions.dispose(),u.suggestWidgetInlineCompletions.dispose()}}}),this.inlineCompletions=this._state.map(this,h=>h.inlineCompletions),this.suggestWidgetInlineCompletions=this._state.map(this,h=>h.suggestWidgetInlineCompletions),this._completionsEnabled=void 0,this.clearOperationOnTextModelChange=j(this,h=>{this._versionId.read(h),this._updateOperation.clear()}),this._loadingCount=Re(this,0),this._loggingEnabled=oq("editor.inlineSuggest.logFetch",!1,this._configurationService).recomputeInitiallyAndOnChange(this._store),this._sendRequestData=oq("editor.inlineSuggest.emptyResponseInformation",!0,this._configurationService).recomputeInitiallyAndOnChange(this._store),this._structuredFetchLogger=this._register(this._instantiationService.createInstance(ED.cast(),"editor.inlineSuggest.logFetch.commandId")),this.clearOperationOnTextModelChange.recomputeInitiallyAndOnChange(this._store);const d=Ek.defaultChatAgent?.completionsEnablementSetting??void 0;d&&(this._updateCompletionsEnablement(d),this._register(this._configurationService.onDidChangeConfiguration(h=>{h.affectsConfiguration(d)&&this._updateCompletionsEnablement(d)}))),this._state.recomputeInitiallyAndOnChange(this._store)}_updateCompletionsEnablement(e){const t=this._configurationService.getValue(e);rn(t)?this._completionsEnabled=t:this._completionsEnabled=void 0}_log(e){this._loggingEnabled.get()&&this._logService.info(bMe(e)),this._structuredFetchLogger.log(e)}fetch(e,t,i,n,o,r,a){const l=this._cursorPosition.get(),c=new LOe(l,i,this._textModel.getVersionId(),new Set(e)),d=i.selectedSuggestionInfo?this.suggestWidgetInlineCompletions.get():this.inlineCompletions.get();if(this._updateOperation.value?.request.satisfies(c))return this._updateOperation.value.promise;if(d?.request?.satisfies(c))return Promise.resolve(!0);const h=!!this._updateOperation.value;this._updateOperation.clear();const u=new si,g=(async()=>{const p=new U;this._loadingCount.set(this._loadingCount.get()+1,void 0);let _=!1;const b=()=>{_||(_=!0,this._loadingCount.set(this._loadingCount.get()-1,void 0))};p.add(new Rt(()=>b(),10*1e3)).schedule();const C=e.filter(y=>y.providerId),v=new kOe(i,a,C);try{const y=this._debounceValue.get(this._textModel),x=See(e.map(J=>J.debounceDelayMs),Dhe(Wo))??y;if((h||o&&i.triggerKind===ho.Automatic)&&await DG(x,u.token),u.token.isCancellationRequested||this._store.isDisposed||this._textModel.getVersionId()!==c.versionId)return v.setNoSuggestionReasonIfNotSet("canceled:beforeFetch"),!1;const E=b3._requestId++;(this._loggingEnabled.get()||this._structuredFetchLogger.isEnabled.get())&&this._log({sourceId:"InlineCompletions.fetch",kind:"start",requestId:E,modelUri:this._textModel.uri,modelVersion:this._textModel.getVersionId(),context:{triggerKind:i.triggerKind,suggestInfo:i.selectedSuggestionInfo?!0:void 0},time:Date.now(),provider:t});const I=new Date,R=wOe(e,this._cursorPosition.get(),this._textModel,i,a,this._languageConfigurationService);moe(u.token,()=>R.cancelAndDispose({kind:"tokenCancellation"}));let O=!1,P=!1;const B=[];for await(const J of R.lists)if(J){J.addRef(),p.add(pe(()=>J.removeRef(J.inlineSuggestionsData.length===0?{kind:"empty"}:{kind:"notTaken"})));for(const ne of J.inlineSuggestionsData){if(P=!0,!i.includeInlineEdits&&(ne.isInlineEdit||ne.showInlineEditMenu)){ne.setNotShownReason("notInlineEditRequested");continue}if(!i.includeInlineCompletions&&!(ne.isInlineEdit||ne.showInlineEditMenu)){ne.setNotShownReason("notInlineCompletionRequested");continue}const de=_3.create(ne,this._textModel);B.push(de),!de.isInlineEdit&&!de.showInlineEditMenu&&i.triggerKind===ho.Automatic&&de.isVisible(this._textModel,this._cursorPosition.get())&&(O=!0)}if(O)break}if(R.cancelAndDispose({kind:"lostRace"}),this._loggingEnabled.get()||this._structuredFetchLogger.isEnabled.get()){const J=R.didAllProvidersReturn;let ne;(u.token.isCancellationRequested||this._store.isDisposed||this._textModel.getVersionId()!==c.versionId)&&(ne="canceled");const de=B.map(ke=>({range:ke.editRange.toString(),text:ke.insertText,hint:ke.hint,isInlineEdit:ke.isInlineEdit,showInlineEditMenu:ke.showInlineEditMenu,providerId:ke.source.provider.providerId?.toString()}));this._log({sourceId:"InlineCompletions.fetch",kind:"end",requestId:E,durationMs:Date.now()-I.getTime(),error:ne,result:de,time:Date.now(),didAllProvidersReturn:J})}if(v.setRequestUuid(R.contextWithUuid.requestUuid),P)v.setHasProducedSuggestion(),B.length>0&&u.token.isCancellationRequested&&B.forEach(J=>J.setNotShownReasonIfNotSet("canceled:whileAwaitingOtherProviders"));else if(u.token.isCancellationRequested)v.setNoSuggestionReasonIfNotSet("canceled:whileFetching");else{const J=this._contextKeyService.getContextKeyValue("completionsQuotaExceeded");v.setNoSuggestionReasonIfNotSet(J?"completionsQuotaExceeded":"noSuggestion")}const $=i.earliestShownDateTime-Date.now();if($>0&&await DG($,u.token),u.token.isCancellationRequested||this._store.isDisposed||this._textModel.getVersionId()!==c.versionId||r.get()){const J=u.token.isCancellationRequested?"canceled:afterMinShowDelay":this._store.isDisposed?"canceled:disposed":this._textModel.getVersionId()!==c.versionId?"canceled:documentChanged":r.get()?"canceled:userJumped":"unknown";return B.forEach(ne=>ne.setNotShownReasonIfNotSet(J)),!1}const z=new Date;this._debounceValue.update(this._textModel,z.getTime()-I.getTime());const Q=this._cursorPosition.get();this._updateOperation.clear(),Vt(J=>{const ne=this._state.get();i.selectedSuggestionInfo?this._state.set({inlineCompletions:ka.createEmpty(),suggestWidgetInlineCompletions:ne.suggestWidgetInlineCompletions.createStateWithAppliedResults(B,c,this._textModel,Q,n)},J):this._state.set({inlineCompletions:ne.inlineCompletions.createStateWithAppliedResults(B,c,this._textModel,Q,n),suggestWidgetInlineCompletions:ka.createEmpty()},J),ne.inlineCompletions.dispose(),ne.suggestWidgetInlineCompletions.dispose()})}finally{p.dispose(),b(),this.sendInlineCompletionsRequestTelemetry(v)}return!0})(),f=new DOe(c,u,g);return this._updateOperation.value=f,g}clear(e){this._updateOperation.clear();const t=this._state.get();this._state.set({inlineCompletions:ka.createEmpty(),suggestWidgetInlineCompletions:ka.createEmpty()},e),t.inlineCompletions.dispose(),t.suggestWidgetInlineCompletions.dispose()}seedInlineCompletionsWithSuggestWidget(){const e=this.inlineCompletions.get(),t=this.suggestWidgetInlineCompletions.get();t&&Vt(i=>{if(!e||(t.request?.versionId??-1)>(e.request?.versionId??-1)){e?.dispose();const n=this._state.get();this._state.set({inlineCompletions:t.clone(),suggestWidgetInlineCompletions:ka.createEmpty()},i),n.inlineCompletions.dispose(),n.suggestWidgetInlineCompletions.dispose()}this.clearSuggestWidgetInlineCompletions(i)})}sendInlineCompletionsRequestTelemetry(e){if(!this._sendRequestData.get()&&!this._contextKeyService.getContextKeyValue("isRunningUnificationExperiment")||e.requestUuid===void 0||e.hasProducedSuggestion||!IOe(this._completionsEnabled,this._textModel.getLanguageId())||!e.providers.some(n=>OG(n.providerId?.extensionId)))return;const t={opportunityId:e.requestUuid,noSuggestionReason:e.noSuggestionReason??"unknown",extensionId:"vscode-core",extensionVersion:"0.0.0",groupId:"empty",shown:!1,editorType:e.requestInfo.editorType,requestReason:e.requestInfo.reason,typingInterval:e.requestInfo.typingInterval,typingIntervalCharacterCount:e.requestInfo.typingIntervalCharacterCount,languageId:e.requestInfo.languageId,selectedSuggestionInfo:!!e.context.selectedSuggestionInfo,availableProviders:e.providers.map(n=>n.providerId?.toString()).filter(Ln).join(","),...zMe(e.providers.some(n=>OG(n.providerId?.extensionId))),timeUntilProviderRequest:void 0,timeUntilProviderResponse:void 0,viewKind:void 0,preceeded:void 0,superseded:void 0,reason:void 0,correlationId:void 0,shownDuration:void 0,shownDurationUncollapsed:void 0,timeUntilShown:void 0,partiallyAccepted:void 0,partiallyAcceptedCountSinceOriginal:void 0,partiallyAcceptedRatioSinceOriginal:void 0,partiallyAcceptedCharactersSinceOriginal:void 0,cursorColumnDistance:void 0,cursorLineDistance:void 0,lineCountOriginal:void 0,lineCountModified:void 0,characterCountOriginal:void 0,characterCountModified:void 0,disjointReplacements:void 0,sameShapeReplacements:void 0,notShownReason:void 0},i=this._instantiationService.createInstance(u3);UMe(i,t)}clearSuggestWidgetInlineCompletions(e){this._updateOperation.value?.request.context.selectedSuggestionInfo&&this._updateOperation.clear()}cancelUpdate(){this._updateOperation.clear()}},b3=Yp,Yp._requestId=0,Yp);w3=b3=xOe([KC(4,ui),KC(5,jt),KC(6,He),KC(7,_e),KC(8,Ae)],w3);class LOe{constructor(e,t,i,n){this.position=e,this.context=t,this.versionId=i,this.providers=n}satisfies(e){return this.position.equals(e.position)&&Vf(this.context.selectedSuggestionInfo,e.context.selectedSuggestionInfo,nW())&&(e.context.triggerKind===ho.Automatic||this.context.triggerKind===ho.Explicit)&&this.versionId===e.versionId&&EOe(e.providers,this.providers)}}class kOe{constructor(e,t,i){this.context=e,this.requestInfo=t,this.providers=i,this.hasProducedSuggestion=!1}setRequestUuid(e){this.requestUuid=e}setNoSuggestionReasonIfNotSet(e){this.noSuggestionReason??=e}setHasProducedSuggestion(){this.hasProducedSuggestion=!0}}function EOe(s,e){return[...s].every(t=>e.has(t))}function IOe(s,e="*"){return s===void 0?!1:typeof s[e]<"u"?!!s[e]:!!s["*"]}class DOe{constructor(e,t,i){this.request=e,this.cancellationTokenSource=t,this.promise=i}dispose(){this.cancellationTokenSource.cancel()}}class ka extends F{static createEmpty(){return new ka([],void 0)}constructor(e,t){for(const i of e)i.addRef();super(),this.inlineCompletions=e,this.request=t,this._register({dispose:()=>{for(const i of this.inlineCompletions)i.removeRef()}})}_findById(e){return this.inlineCompletions.find(t=>t.identity===e)}_findByHash(e){return this.inlineCompletions.find(t=>t.hash===e)}createStateWithAppliedEdit(e,t){const i=this.inlineCompletions.map(n=>n.withEdit(e,t)).filter(Ln);return new ka(i,this.request)}createStateWithAppliedResults(e,t,i,n,o){let r;if(o){const c=this._findById(o);if(c&&c.canBeReused(i,t.position)){r=c;const d=e.find(h=>h.hash===c.hash);d?e=NOe(d,e):e=[c,...e]}}const a=r?!r.isInlineEdit:e.some(c=>!c.isInlineEdit&&c.isVisible(i,n));let l=[];for(const c of e){const d=this._findByHash(c.hash);let h;d&&d!==c?(h=c.withIdentity(d.identity),c.setIsPreceeded(d),d.setEndOfLifeReason({kind:Dm.Ignored,userTypingDisagreed:!1,supersededBy:c.getSourceCompletion()})):h=c,a!==h.isInlineEdit&&l.push(h)}return l.sort(ns(c=>c.showInlineEditMenu,_Q)),l=TOe(l,c=>c.semanticId),new ka(l,t)}clone(){return new ka(this.inlineCompletions,this.request)}}function TOe(s,e){const t=new Set;return s.filter(i=>{const n=e(i);return t.has(n)?!1:(t.add(n),!0)})}function NOe(s,e){const t=e.indexOf(s);return t>-1?[s,...e.slice(0,t),...e.slice(t+1)]:e}class ROe{constructor(e,t,i){this.edit=e,this.commands=t,this.inlineCompletion=i}equals(e){return this.edit.equals(e.edit)&&this.inlineCompletion===e.inlineCompletion}}var AOe=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},ZG=function(s,e){return function(t,i){e(t,i,s)}};const KA=qe("IInlineCompletionsService"),_oe=new ce("inlineCompletions.snoozed",!1,m(79,"Whether inline completions are currently snoozed"));let v3=class extends F{get snoozeTimeLeft(){return this._snoozeTimeEnd===void 0?0:Math.max(0,this._snoozeTimeEnd-Date.now())}constructor(e,t){super(),this._contextKeyService=e,this._telemetryService=t,this._onDidChangeIsSnoozing=this._register(new M),this.onDidChangeIsSnoozing=this._onDidChangeIsSnoozing.event,this._snoozeTimeEnd=void 0,this._recentCompletionIds=[],this._timer=this._register(new $o);const i=_oe.bindTo(this._contextKeyService);this._register(this.onDidChangeIsSnoozing(()=>i.set(this.isSnoozing())))}setSnoozeDuration(e){if(e<0)throw new Ce(`Invalid snooze duration: ${e}. Duration must be non-negative.`);if(e===0){this.cancelSnooze();return}const t=this.isSnoozing(),i=this.snoozeTimeLeft;this._snoozeTimeEnd=Date.now()+e,t||this._onDidChangeIsSnoozing.fire(!0),this._timer.cancelAndSet(()=>{if(!this.isSnoozing())this._onDidChangeIsSnoozing.fire(!1);else throw new Ce("Snooze timer did not fire as expected")},this.snoozeTimeLeft+1),this._reportSnooze(e-i,e)}isSnoozing(){return this.snoozeTimeLeft>0}cancelSnooze(){this.isSnoozing()&&(this._reportSnooze(-this.snoozeTimeLeft,0),this._snoozeTimeEnd=void 0,this._timer.cancel(),this._onDidChangeIsSnoozing.fire(!1))}reportNewCompletion(e){this._lastCompletionId=e,this._recentCompletionIds.unshift(e),this._recentCompletionIds.length>5&&this._recentCompletionIds.pop()}_reportSnooze(e,t){const i=Math.round(e/1e3),n=Math.round(t/1e3);this._telemetryService.publicLog2("inlineCompletions.snooze",{deltaSeconds:i,totalSeconds:n,lastCompletionId:this._lastCompletionId,recentCompletionIds:this._recentCompletionIds})}};v3=AOe([ZG(0,Ae),ZG(1,vs)],v3);tt(KA,v3,1);const MOe="editor.action.inlineSuggest.snooze",OOe="editor.action.inlineSuggest.cancelSnooze",XG="inlineCompletions.lastSnoozeDuration",MN=class MN extends Tn{constructor(){super({id:MN.ID,title:V(81,"Snooze Inline Suggestions"),precondition:K.true(),f1:!0})}async run(e,...t){const i=e.get(bs),n=e.get(KA),o=e.get(Bs);let r;t.length>0&&typeof t[0]=="number"&&(r=t[0]*6e4),r||(r=await this.getDurationFromUser(i,o)),r&&n.setSnoozeDuration(r)}async getDurationFromUser(e,t){const i=t.getNumber(XG,0,3e5),n=[{label:"1 minute",id:"1",value:6e4},{label:"5 minutes",id:"5",value:3e5},{label:"10 minutes",id:"10",value:6e5},{label:"15 minutes",id:"15",value:9e5},{label:"30 minutes",id:"30",value:18e5},{label:"60 minutes",id:"60",value:36e5}],o=await e.pick(n,{placeHolder:m(80,"Select snooze duration for Inline Suggestions"),activeItem:n.find(r=>r.value===i)});if(o)return t.store(XG,o.value,0,0),o.value}};MN.ID=MOe;let C3=MN;const ON=class ON extends Tn{constructor(){super({id:ON.ID,title:V(82,"Cancel Snooze Inline Suggestions"),precondition:_oe,f1:!0})}async run(e){e.get(KA).cancelSnooze()}};ON.ID=OOe;let y3=ON;const Pr=class Pr extends F{getTypingInterval(){return(this._cacheInvalidated||this._cachedTypingIntervalResult===null)&&(this._cachedTypingIntervalResult=this._calculateTypingInterval(),this._cacheInvalidated=!1),this._cachedTypingIntervalResult}constructor(e){super(),this._textModel=e,this._typingSessions=[],this._currentSession=null,this._lastChangeTime=0,this._cachedTypingIntervalResult=null,this._cacheInvalidated=!0,this._register(this._textModel.onDidChangeContent(t=>this._updateTypingSpeed(t)))}_updateTypingSpeed(e){const t=Date.now();if(!this._isUserTyping(e)){this._finalizeCurrentSession();return}this._currentSession&&t-this._lastChangeTime>Pr.MAX_SESSION_GAP_MS&&this._finalizeCurrentSession(),this._currentSession||(this._currentSession={startTime:t,endTime:t,characterCount:0}),this._currentSession.endTime=t,this._currentSession.characterCount+=this._getActualCharacterCount(e),this._lastChangeTime=t,this._cacheInvalidated=!0}_getActualCharacterCount(e){let t=0;for(const i of e.changes)t+=Math.max(i.text.length,i.rangeLength);return t}_isUserTyping(e){if(!e.detailedReasons||e.detailedReasons.length===0)return!1;for(const t of e.detailedReasons)if(this._isUserTypingReason(t))return!0;return!1}_isUserTypingReason(e){if(e.metadata.isUndoing||e.metadata.isRedoing)return!1;if(e.metadata.source==="cursor"){const t=e.metadata.kind;return t==="type"||t==="compositionType"||t==="compositionEnd"}else return!1}_finalizeCurrentSession(){if(!this._currentSession)return;this._currentSession.endTime-this._currentSession.startTime>=Pr.MIN_SESSION_DURATION_MS&&this._currentSession.characterCount>0&&(this._typingSessions.push(this._currentSession),this._typingSessions.length>Pr.SESSION_HISTORY_LIMIT&&this._typingSessions.shift()),this._currentSession=null}_calculateTypingInterval(){if(this._currentSession){const e={...this._currentSession};if(e.endTime-e.startTime>=Pr.MIN_SESSION_DURATION_MS&&e.characterCount>0){const i=[...this._typingSessions,e];return this._calculateSpeedFromSessions(i)}}return this._calculateSpeedFromSessions(this._typingSessions)}_calculateSpeedFromSessions(e){if(e.length===0)return{averageInterval:0,characterCount:0};const t=[...e].sort((d,h)=>h.endTime-d.endTime),i=Date.now()-Pr.TYPING_SPEED_WINDOW_MS,n=t.filter(d=>d.endTime>i),o=t.splice(n.length);let r=SE(n.map(d=>d.characterCount));for(let d=0;d<o.length&&r<Pr.MIN_CHARS_FOR_RELIABLE_SPEED;d++)n.push(o[d]),r+=o[d].characterCount;const a=SE(n.map(d=>d.endTime-d.startTime));if(a===0||r<=1)return{averageInterval:0,characterCount:r};const l=Math.max(1,r-1),c=a/l;return{averageInterval:Math.round(c),characterCount:r}}dispose(){this._finalizeCurrentSession(),super.dispose()}};Pr.MAX_SESSION_GAP_MS=3e3,Pr.MIN_SESSION_DURATION_MS=1e3,Pr.SESSION_HISTORY_LIMIT=50,Pr.TYPING_SPEED_WINDOW_MS=3e5,Pr.MIN_CHARS_FOR_RELIABLE_SPEED=20;let S3=Pr;var POe=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},Kg=function(s,e){return function(t,i){e(t,i,s)}};let x3=class extends F{get isAcceptingPartially(){return this._isAcceptingPartially}constructor(e,t,i,n,o,r,a,l,c,d,h,u,g,f){super(),this.textModel=e,this._selectedSuggestItem=t,this._textModelVersionId=i,this._positions=n,this._debounceValue=o,this._enabled=r,this._editor=a,this._instantiationService=l,this._commandService=c,this._languageConfigurationService=d,this._accessibilityService=h,this._languageFeaturesService=u,this._codeEditorService=g,this._inlineCompletionsService=f,this._isActive=Re(this,!1),this._onlyRequestInlineEditsSignal=Yr(this),this._forceUpdateExplicitlySignal=Yr(this),this._noDelaySignal=Yr(this),this._fetchSpecificProviderSignal=Yr(this),this._selectedInlineCompletionId=Re(this,void 0),this.primaryPosition=j(this,v=>this._positions.read(v)[0]??new A(1,1)),this.allPositions=j(this,v=>this._positions.read(v)),this._isAcceptingPartially=!1,this._appearedInsideViewport=j(this,v=>{const y=this.state.read(v);return!y||!y.inlineCompletion?!1:BOe(this._editor,y.inlineCompletion)}),this._onDidAccept=new M,this.onDidAccept=this._onDidAccept.event,this._lastShownInlineCompletionInfo=void 0,this._lastAcceptedInlineCompletionInfo=void 0,this._didUndoInlineEdits=Vz({owner:this,changeTracker:{createChangeSummary:()=>({didUndo:!1}),handleChange:(v,y)=>(y.didUndo=v.didChange(this._textModelVersionId)&&!!v.change?.isUndoing,!0)}},(v,y)=>{const x=this._textModelVersionId.read(v);return x!==null&&this._lastAcceptedInlineCompletionInfo&&this._lastAcceptedInlineCompletionInfo.textModelVersionIdAfter===x-1&&this._lastAcceptedInlineCompletionInfo.inlineCompletion.isInlineEdit&&y.didUndo?(this._lastAcceptedInlineCompletionInfo=void 0,!0):!1}),this._preserveCurrentCompletionReasons=new Set([yd.Redo,yd.Undo,yd.AcceptWord]),this.dontRefetchSignal=Yr(this),this._fetchInlineCompletionsPromise=Vz({owner:this,changeTracker:{createChangeSummary:()=>({dontRefetch:!1,preserveCurrentCompletion:!1,inlineCompletionTriggerKind:ho.Automatic,onlyRequestInlineEdits:!1,shouldDebounce:!0,provider:void 0,textChange:!1,changeReason:""}),handleChange:(v,y)=>{if(v.didChange(this._textModelVersionId)){this._preserveCurrentCompletionReasons.has(this._getReason(v.change))&&(y.preserveCurrentCompletion=!0);const x=v.change?.detailedReasons??[];y.changeReason=x.length>0?x[0].getType():"",y.textChange=!0}else v.didChange(this._forceUpdateExplicitlySignal)?(y.preserveCurrentCompletion=!0,y.inlineCompletionTriggerKind=ho.Explicit):v.didChange(this.dontRefetchSignal)?y.dontRefetch=!0:v.didChange(this._onlyRequestInlineEditsSignal)?y.onlyRequestInlineEdits=!0:v.didChange(this._fetchSpecificProviderSignal)&&(y.provider=v.change);return!0}}},(v,y)=>{if(this._source.clearOperationOnTextModelChange.read(v),this._noDelaySignal.read(v),this.dontRefetchSignal.read(v),this._onlyRequestInlineEditsSignal.read(v),this._forceUpdateExplicitlySignal.read(v),this._fetchSpecificProviderSignal.read(v),!((this._enabled.read(v)&&this._selectedSuggestItem.read(v)||this._isActive.read(v))&&(!this._inlineCompletionsService.isSnoozing()||y.inlineCompletionTriggerKind===ho.Explicit))){this._source.cancelUpdate();return}this._textModelVersionId.read(v);const L=this._source.suggestWidgetInlineCompletions.read(void 0);let E=this._selectedSuggestItem.read(v);if(this._shouldShowOnSuggestConflict.read(void 0)&&(E=void 0),L&&!E&&this._source.seedInlineCompletionsWithSuggestWidget(),y.dontRefetch)return Promise.resolve(!0);if(this._didUndoInlineEdits.read(v)&&y.inlineCompletionTriggerKind!==ho.Explicit){Vt(ne=>{this._source.clear(ne)});return}let I="";y.provider?I+="providerOnDidChange":y.inlineCompletionTriggerKind===ho.Explicit&&(I+="explicit"),y.changeReason&&(I+=I.length>0?`:${y.changeReason}`:y.changeReason);const R=this._typing.getTypingInterval(),O={editorType:this.editorType,startTime:Date.now(),languageId:this.textModel.getLanguageId(),reason:I,typingInterval:R.averageInterval,typingIntervalCharacterCount:R.characterCount,availableProviders:[]};let P={triggerKind:y.inlineCompletionTriggerKind,selectedSuggestionInfo:E?.toSelectedSuggestionInfo(),includeInlineCompletions:!y.onlyRequestInlineEdits,includeInlineEdits:this._inlineEditsEnabled.read(v),requestIssuedDateTime:O.startTime,earliestShownDateTime:O.startTime+(y.inlineCompletionTriggerKind===ho.Explicit||this.inAcceptFlow.read(void 0)?0:this._minShowDelay.read(void 0))};P.triggerKind===ho.Automatic&&y.textChange&&this.textModel.getAlternativeVersionId()===this._lastShownInlineCompletionInfo?.alternateTextModelVersionId&&(P={...P,includeInlineCompletions:!this._lastShownInlineCompletionInfo.inlineCompletion.isInlineEdit,includeInlineEdits:this._lastShownInlineCompletionInfo.inlineCompletion.isInlineEdit});const B=this.selectedInlineCompletion.read(void 0)??this._inlineCompletionItems.read(void 0)?.inlineEdit,$=y.preserveCurrentCompletion||B?.forwardStable?B:void 0,z=this._jumpedToId.map(ne=>!!ne&&ne===this._inlineCompletionItems.read(void 0)?.inlineEdit?.semanticId),Q=y.provider?{providers:[y.provider],label:"single:"+y.provider.providerId?.toString()}:{providers:this._languageFeaturesService.inlineCompletionsProvider.all(this.textModel),label:void 0},J=this.getAvailableProviders(Q.providers);return O.availableProviders=J.map(ne=>ne.providerId).filter(Ln),this._source.fetch(J,Q.label,P,$?.identity,y.shouldDebounce,z,O)}),this._inlineCompletionItems=Jn({owner:this},v=>{const y=this._source.inlineCompletions.read(v);if(!y)return;const x=this.primaryPosition.read(v);let L;const E=[];for(const I of y.inlineCompletions)I.isInlineEdit?L=I:I.isVisible(this.textModel,x)&&E.push(I);return E.length!==0&&(L=void 0),{inlineCompletions:E,inlineEdit:L}}),this._filteredInlineCompletionItems=Jn({owner:this,equalsFn:Gy()},v=>this._inlineCompletionItems.read(v)?.inlineCompletions??[]),this.selectedInlineCompletionIndex=j(this,v=>{const y=this._selectedInlineCompletionId.read(v),x=this._filteredInlineCompletionItems.read(v),L=this._selectedInlineCompletionId===void 0?-1:x.findIndex(E=>E.semanticId===y);return L===-1?(this._selectedInlineCompletionId.set(void 0,void 0),0):L}),this.selectedInlineCompletion=j(this,v=>{const y=this._filteredInlineCompletionItems.read(v),x=this.selectedInlineCompletionIndex.read(v);return y[x]}),this.activeCommands=Jn({owner:this,equalsFn:Gy()},v=>this.selectedInlineCompletion.read(v)?.source.inlineSuggestions.commands??[]),this.inlineCompletionsCount=j(this,v=>{if(this.lastTriggerKind.read(v)===ho.Explicit)return this._filteredInlineCompletionItems.read(v).length}),this._hasVisiblePeekWidgets=j(this,v=>this._editorObs.openedPeekWidgets.read(v)>0),this._shouldShowOnSuggestConflict=j(this,v=>{const y=this._showOnSuggestConflict.read(v);if(y!=="never"&&!!this.selectedInlineCompletion.read(v)){const L=this._selectedSuggestItem.read(v);return L?y==="whenSuggestListIsIncomplete"?L.listIncomplete:!0:!1}return!1}),this.state=Jn({owner:this,equalsFn:(v,y)=>!v||!y?v===y:v.kind==="ghostText"&&y.kind==="ghostText"?TG(v.ghostTexts,y.ghostTexts)&&v.inlineCompletion===y.inlineCompletion&&v.suggestItem===y.suggestItem:v.kind==="inlineEdit"&&y.kind==="inlineEdit"?v.inlineEdit.equals(y.inlineEdit):!1},v=>{const y=this.textModel;if(this._suppressInSnippetMode.read(v)&&this._isInSnippetMode.read(v))return;const x=this._inlineCompletionItems.read(v),L=x?.inlineEdit;if(L){if(this._hasVisiblePeekWidgets.read(v))return;let I=L.getSingleTextEdit();I=jd(I,y);const R=this.primaryPosition.map(Q=>Ne.fromRangeInclusive(L.targetRange).addMargin(1,1).contains(Q.lineNumber)),O=L.source.inlineSuggestions.commands,P=new ROe(I,O??[],L),B=L.updatedEdit,$=B?Na.fromStringEdit(B,new Fm(this.textModel)).replacements:[I],z=(x.inlineEdit?.command?.id==="vscode.open"||x.inlineEdit?.command?.id==="_workbench.open")&&x.inlineEdit?.command.arguments?.length?Le.from(x.inlineEdit?.command.arguments[0]):void 0;return{kind:"inlineEdit",inlineEdit:P,inlineCompletion:L,edits:$,cursorAtInlineEdit:R,nextEditUri:z}}const E=this._selectedSuggestItem.read(v);if(!this._shouldShowOnSuggestConflict.read(v)&&E){const I=jd(E.getSingleTextEdit(),y),R=this._computeAugmentation(I,v);if(!this._suggestPreviewEnabled.read(v)&&!R)return;const P=R?.edit??I,B=R?R.edit.text.length-I.text.length:0,$=this._suggestPreviewMode.read(v),z=this._positions.read(v),J=[P,...uO(this.textModel,z,P)].map((Me,Ze)=>({edit:Me,ghostText:Me?NG(Me,y,$,z[Ze],B):void 0})).filter(({edit:Me,ghostText:Ze})=>Me!==void 0&&Ze!==void 0),ne=J.map(({edit:Me})=>Me),de=J.map(({ghostText:Me})=>Me),ke=de[0]??new r0(P.range.endLineNumber,[]);return{kind:"ghostText",edits:ne,primaryGhostText:ke,ghostTexts:de,inlineCompletion:R?.completion,suggestItem:E}}else{if(!this._isActive.read(v))return;const I=this.selectedInlineCompletion.read(v);if(!I)return;const R=I.getSingleTextEdit(),O=this._inlineSuggestMode.read(v),P=this._positions.read(v),$=[R,...uO(this.textModel,P,R)].map((J,ne)=>({edit:J,ghostText:J?NG(J,y,O,P[ne],0):void 0})).filter(({edit:J,ghostText:ne})=>J!==void 0&&ne!==void 0),z=$.map(({edit:J})=>J),Q=$.map(({ghostText:J})=>J);return Q[0]?{kind:"ghostText",edits:z,primaryGhostText:Q[0],ghostTexts:Q,inlineCompletion:I,suggestItem:void 0}:void 0}}),this.inlineCompletionState=j(this,v=>{const y=this.state.read(v);if(!(!y||y.kind!=="ghostText")&&!this._editorObs.inComposition.read(v))return y}),this.inlineEditState=j(this,v=>{const y=this.state.read(v);if(!(!y||y.kind!=="inlineEdit"))return y}),this.inlineEditAvailable=j(this,v=>!!this.inlineEditState.read(v)),this.warning=j(this,v=>this.inlineCompletionState.read(v)?.inlineCompletion?.warning),this.ghostTexts=Jn({owner:this,equalsFn:TG},v=>{const y=this.inlineCompletionState.read(v);if(y)return y.ghostTexts}),this.primaryGhostText=Jn({owner:this,equalsFn:loe},v=>{const y=this.inlineCompletionState.read(v);if(y)return y?.primaryGhostText}),this.showCollapsed=j(this,v=>{const y=this.state.read(v);if(!y||y.kind!=="inlineEdit"||y.inlineCompletion.hint)return!1;const x=y.inlineCompletion.updatedEditModelVersion===this._textModelVersionId.read(v);return(this._inlineEditsShowCollapsedEnabled.read(v)||!x)&&this._jumpedToId.read(v)!==y.inlineCompletion.semanticId&&!this._inAcceptFlow.read(v)}),this._tabShouldIndent=j(this,v=>{if(this._inAcceptFlow.read(v))return!1;function y(E){return E.startLineNumber!==E.endLineNumber}function x(E,I){const R=E.getLineIndentColumn(I),O=E.getLineLastNonWhitespaceColumn(I),P=Math.max(O,R);return new k(I,R,I,P)}return this._editorObs.selections.read(v)?.some(E=>E.isEmpty()?this.textModel.getLineLength(E.startLineNumber)===0:y(E)||E.containsRange(x(this.textModel,E.startLineNumber)))}),this.tabShouldJumpToInlineEdit=j(this,v=>{if(this._tabShouldIndent.read(v))return!1;const y=this.inlineEditState.read(v);return y?this.showCollapsed.read(v)?!0:this._inAcceptFlow.read(v)&&this._appearedInsideViewport.read(v)&&!y.inlineCompletion.hint?.jumpToEdit?!1:!y.cursorAtInlineEdit.read(v):!1}),this.tabShouldAcceptInlineEdit=j(this,v=>{const y=this.inlineEditState.read(v);return!y||this.showCollapsed.read(v)||this._tabShouldIndent.read(v)?!1:this._inAcceptFlow.read(v)&&this._appearedInsideViewport.read(v)&&!y.inlineCompletion.hint?.jumpToEdit||y.inlineCompletion.targetRange.startLineNumber===this._editorObs.cursorLineNumber.read(v)||this._jumpedToId.read(v)===y.inlineCompletion.semanticId?!0:y.cursorAtInlineEdit.read(v)}),this._jumpedToId=Re(this,void 0),this._inAcceptFlow=Re(this,!1),this.inAcceptFlow=this._inAcceptFlow,this._source=this._register(this._instantiationService.createInstance(w3,this.textModel,this._textModelVersionId,this._debounceValue,this.primaryPosition)),this.lastTriggerKind=this._source.inlineCompletions.map(this,v=>v?.request?.context.triggerKind),this._editorObs=di(this._editor);const p=this._editorObs.getOption(134);this._suggestPreviewEnabled=p.map(v=>v.preview),this._suggestPreviewMode=p.map(v=>v.previewMode);const _=this._editorObs.getOption(71);this._inlineSuggestMode=_.map(v=>v.mode),this._suppressedInlineCompletionGroupIds=_.map(v=>new Set(v.experimental.suppressInlineSuggestions.split(","))),this._inlineEditsEnabled=_.map(v=>!!v.edits.enabled),this._inlineEditsShowCollapsedEnabled=_.map(v=>v.edits.showCollapsed),this._triggerCommandOnProviderChange=_.map(v=>v.triggerCommandOnProviderChange),this._minShowDelay=_.map(v=>v.minShowDelay),this._showOnSuggestConflict=_.map(v=>v.experimental.showOnSuggestConflict),this._suppressInSnippetMode=_.map(v=>v.suppressInSnippetMode);const b=Ps.get(this._editor);this._isInSnippetMode=b?.isInSnippetObservable??zt(!1),this._typing=this._register(new S3(this.textModel)),this._register(this._inlineCompletionsService.onDidChangeIsSnoozing(v=>{v&&this.stop()}));{const v=this.textModel.uri.scheme==="vscode-notebook-cell",[y]=this._codeEditorService.listDiffEditors().filter(x=>x.getOriginalEditor().getId()===this._editor.getId()||x.getModifiedEditor().getId()===this._editor.getId());this.isInDiffEditor=!!y,this.editorType=v?sy.Notebook:this.isInDiffEditor?sy.DiffEditor:sy.TextEditor}this._register(sv(this.state,v=>{v&&v.inlineCompletion&&this._inlineCompletionsService.reportNewCompletion(v.inlineCompletion.requestUuid)})),this._register(sv(this._fetchInlineCompletionsPromise)),this._register(De(v=>{this._editorObs.versionId.read(v),this._inAcceptFlow.set(!1,void 0)})),this._register(De(v=>{this.state.map((x,L)=>!x||x.kind==="inlineEdit"&&!x.cursorAtInlineEdit.read(L)).read(v)&&this._jumpedToId.set(void 0,void 0)}));const w=this.inlineEditState.map(v=>v?.inlineCompletion.semanticId);this._register(De(v=>{w.read(v)&&(this._editor.pushUndoStop(),this._lastShownInlineCompletionInfo={alternateTextModelVersionId:this.textModel.getAlternativeVersionId(),inlineCompletion:this.state.get().inlineCompletion})}));const C=bt(this._languageFeaturesService.inlineCompletionsProvider.onDidChange,()=>this._languageFeaturesService.inlineCompletionsProvider.all(e));rW(this,C,(v,y)=>{v.onDidChangeInlineCompletions&&y.add(v.onDidChangeInlineCompletions(()=>{if(!this._enabled.get()||(this._codeEditorService.getFocusedCodeEditor()||this._codeEditorService.getActiveCodeEditor())!==this._editor)return;if(this._triggerCommandOnProviderChange.get()){this.trigger(void 0,{onlyFetchInlineEdits:!0});return}const L=this.state.get();L&&(L.inlineCompletion||L.edits)&&L.inlineCompletion?.source.provider!==v||Vt(E=>{this._fetchSpecificProviderSignal.trigger(E,v),this.trigger(E)})}))}).recomputeInitiallyAndOnChange(this._store),this._didUndoInlineEdits.recomputeInitiallyAndOnChange(this._store)}getIndentationInfo(e){let t=!1,i=!0;const n=this?.primaryGhostText.read(e);if(this?._selectedSuggestItem&&n&&n.parts.length>0){const{column:o,lines:r}=n.parts[0],a=r[0].line,l=this.textModel.getLineIndentColumn(n.lineNumber);if(o<=l){let d=As(a);d===-1&&(d=a.length-1),t=d>0;const h=this.textModel.getOptions().tabSize;i=pi.visibleColumnFromColumn(a,d+1,h)<h}}return{startsWithIndentation:t,startsWithIndentationLessThanTabSize:i}}_getReason(e){return e?.isUndoing?yd.Undo:e?.isRedoing?yd.Redo:this.isAcceptingPartially?yd.AcceptWord:yd.Other}getAvailableProviders(e){const t=this._suppressedInlineCompletionGroupIds.get(),i=e.filter(r=>!(r.groupId&&t.has(r.groupId))),n=new Set;for(const r of i)r.excludesGroupIds?.forEach(a=>n.add(a));const o=[];for(const r of i)r.groupId&&n.has(r.groupId)||o.push(r);return o}async trigger(e,t={}){nv(e,i=>{t.onlyFetchInlineEdits&&this._onlyRequestInlineEditsSignal.trigger(i),t.noDelay&&this._noDelaySignal.trigger(i),this._isActive.set(!0,i),t.explicit&&(this._inAcceptFlow.set(!0,i),this._forceUpdateExplicitlySignal.trigger(i)),t.provider&&this._fetchSpecificProviderSignal.trigger(i,t.provider)}),await this._fetchInlineCompletionsPromise.get()}async triggerExplicitly(e,t=!1){return this.trigger(e,{onlyFetchInlineEdits:t,explicit:!0})}stop(e="automatic",t){nv(t,i=>{if(e==="explicitCancel"){const n=this.state.get()?.inlineCompletion;n&&n.reportEndOfLife({kind:Dm.Rejected})}this._isActive.set(!1,i),this._source.clear(i)})}_computeAugmentation(e,t){const i=this.textModel,n=this._source.suggestWidgetInlineCompletions.read(t),o=n?n.inlineCompletions.filter(a=>!a.isInlineEdit):[this.selectedInlineCompletion.read(t)].filter(Ln);return Vwe(o,a=>{let l=a.getSingleTextEdit();return l=jd(l,i,k.fromPositions(l.range.getStartPosition(),e.range.getEndPosition())),coe(l,e)?{completion:a,edit:l}:void 0})}async _deltaSelectedInlineCompletionIndex(e){await this.triggerExplicitly();const t=this._filteredInlineCompletionItems.get()||[];if(t.length>0){const i=(this.selectedInlineCompletionIndex.get()+e+t.length)%t.length;this._selectedInlineCompletionId.set(t[i].semanticId,void 0)}else this._selectedInlineCompletionId.set(void 0,void 0)}async next(){await this._deltaSelectedInlineCompletionIndex(1)}async previous(){await this._deltaSelectedInlineCompletionIndex(-1)}_getMetadata(e,t,i=void 0){return i?cs.inlineCompletionPartialAccept({nes:e.isInlineEdit,requestUuid:e.requestUuid,providerId:e.source.provider.providerId,languageId:t,type:i}):cs.inlineCompletionAccept({nes:e.isInlineEdit,requestUuid:e.requestUuid,providerId:e.source.provider.providerId,languageId:t})}async accept(e=this._editor){if(e.getModel()!==this.textModel)throw new Ce;let t,i=!1;const n=this.state.get();if(n?.kind==="ghostText"){if(!n||n.primaryGhostText.isEmpty()||!n.inlineCompletion)return;t=n.inlineCompletion}else if(n?.kind==="inlineEdit")t=n.inlineCompletion,i=!!n.nextEditUri;else return;t.addRef();try{if(e.pushUndoStop(),!i)if(t.snippetInfo){const o=Ui.delete(t.editRange),r=t.additionalTextEdits.map(l=>new Ui(k.lift(l.range),l.text??"")),a=Na.fromParallelReplacementsUnsorted([o,...r]);e.edit(a,this._getMetadata(t,this.textModel.getLanguageId())),e.setPosition(t.snippetInfo.range.getStartPosition(),"inlineCompletionAccept"),Ps.get(e)?.insert(t.snippetInfo.snippet,{undoStopBefore:!1})}else{const o=n.edits;let r=o;n.kind==="ghostText"&&(r=TMe(o,this.textModel));const a=IG(r).map(d=>he.fromPositions(d)),l=t.additionalTextEdits.map(d=>new Ui(k.lift(d.range),d.text??"")),c=Na.fromParallelReplacementsUnsorted([...o,...l]);if(e.edit(c,this._getMetadata(t,this.textModel.getLanguageId())),t.hint===void 0&&e.setSelections(n.kind==="inlineEdit"?a.slice(-1):a,"inlineCompletionAccept"),n.kind==="inlineEdit"&&!this._accessibilityService.isMotionReduced()){const d=c.getNewRanges(),h=this._store.add(new FOe(e,d,()=>{this._store.delete(h)}))}}this._onDidAccept.fire(),this.stop(),t.command&&await this._commandService.executeCommand(t.command.id,...t.command.arguments||[]).then(void 0,ji),t.reportEndOfLife({kind:Dm.Accepted})}finally{t.removeRef(),this._inAcceptFlow.set(!0,void 0),this._lastAcceptedInlineCompletionInfo={textModelVersionIdAfter:this.textModel.getVersionId(),inlineCompletion:t}}}async acceptNextWord(){await this._acceptNext(this._editor,"word",(e,t)=>{const i=this.textModel.getLanguageIdAtPosition(e.lineNumber,e.column),n=this._languageConfigurationService.getLanguageConfiguration(i),o=new RegExp(n.wordDefinition.source,n.wordDefinition.flags.replace("g","")),r=t.match(o);let a=0;r&&r.index!==void 0?r.index===0?a=r[0].length:a=r.index:a=t.length;const c=/\s+/g.exec(t);return c&&c.index!==void 0&&c.index+c[0].length<a&&(a=c.index+c[0].length),a},0)}async acceptNextLine(){await this._acceptNext(this._editor,"line",(e,t)=>{const i=t.match(/\n/);return i&&i.index!==void 0?i.index+1:t.length},1)}async _acceptNext(e,t,i,n){if(e.getModel()!==this.textModel)throw new Ce;const o=this.inlineCompletionState.get();if(!o||o.primaryGhostText.isEmpty()||!o.inlineCompletion)return;const r=o.primaryGhostText,a=o.inlineCompletion;if(a.snippetInfo){await this.accept(e);return}const l=r.parts[0],c=new A(r.lineNumber,l.column),d=l.text,h=i(c,d);if(h===d.length&&r.parts.length===1){this.accept(e);return}const u=d.substring(0,h),g=this._positions.get(),f=g[0];a.addRef();try{this._isAcceptingPartially=!0;try{e.pushUndoStop();const w=k.fromPositions(f,c),C=e.getModel().getValueInRange(w)+u,v=new Ui(w,C),y=[v,...uO(this.textModel,g,v)].filter(Ln),x=IG(y).map(L=>he.fromPositions(L));e.edit(Na.fromParallelReplacementsUnsorted(y),this._getMetadata(a,t)),e.setSelections(x,"inlineCompletionPartialAccept"),e.revealPositionInCenterIfOutsideViewport(e.getPosition(),1)}finally{this._isAcceptingPartially=!1}const p=k.fromPositions(a.editRange.getStartPosition(),cn.ofText(u).addToPosition(c)),b=e.getModel().getValueInRange(p,1).length;a.reportPartialAccept(b,{kind:n,acceptedLength:b},{characters:h,ratio:h/d.length,count:1})}finally{a.removeRef()}}handleSuggestAccepted(e){const t=jd(e.getSingleTextEdit(),this.textModel),i=this._computeAugmentation(t,void 0);if(!i)return;const o=this.textModel.getValueInRange(i.completion.editRange,1).length+t.text.length;i.completion.reportPartialAccept(t.text.length,{kind:2,acceptedLength:o},{characters:t.text.length,count:1,ratio:1})}extractReproSample(){const e=this.textModel.getValue(),t=this.state.get()?.inlineCompletion;return{documentValue:e,inlineCompletion:t?.getSourceCompletion()}}jump(){const e=this.inlineEditState.get();e&&Vt(t=>{this._jumpedToId.set(e.inlineCompletion.semanticId,t),this.dontRefetchSignal.trigger(t);const i=e.inlineCompletion.targetRange,n=i.getStartPosition();if(this._editor.setPosition(n,"inlineCompletions.jump"),i.isSingleLine()&&(e.inlineCompletion.hint||!e.inlineCompletion.insertText.includes(`
`)))this._editor.revealPosition(n);else{const r=new k(i.startLineNumber-1,1,i.endLineNumber+1,1);this._editor.revealRange(r,1)}e.inlineCompletion.identity.setJumpTo(t),this._editor.focus()})}async handleInlineSuggestionShown(e,t,i){await e.reportInlineEditShown(this._commandService,t,i)}};x3=POe([Kg(7,_e),Kg(8,Gt),Kg(9,ui),Kg(10,Wn),Kg(11,ge),Kg(12,ht),Kg(13,KA)],x3);var yd;(function(s){s[s.Undo=0]="Undo",s[s.Redo=1]="Redo",s[s.AcceptWord=2]="AcceptWord",s[s.Other=3]="Other"})(yd||(yd={}));function uO(s,e,t){if(e.length===1)return[];const i=new Fm(s),n=i.getTransformer(),o=n.getOffset(e[0]),r=e.slice(1).map(u=>n.getOffset(u));t=t.removeCommonPrefixAndSuffix(i);const a=n.getStringReplacement(t),l=a.replaceRange.start-o,c=a.replaceRange.join(me.emptyAt(o)),d=i.getValueOfOffsetRange(c);return r.map(u=>{const g=u+l,f=g+a.replaceRange.length,p=new me(g,f),_=p.join(me.emptyAt(u));if(i.getValueOfOffsetRange(_)!==d)return;const w=new On(p,a.newText);return n.getTextReplacement(w)}).filter(Ln)}class FOe extends F{constructor(e,t,i){super(),i&&this._register({dispose:()=>i()}),this._register(di(e).setDecorations(zt(t.map(r=>({range:r,options:{description:"animation",className:"edits-fadeout-decoration",zIndex:1}})))));const n=new AMe(1,0,1e3,OMe),o=new PMe(n);this._register(De(r=>{const a=o.getValue(r);e.getContainerDomNode().style.setProperty("--animation-opacity",a.toString()),n.isFinished()&&this.dispose()}))}}function BOe(s,e){const t=e.targetRange,i=s.getVisibleRanges();return i.length<1?!1:new k(i[0].startLineNumber,i[0].startColumn,i[i.length-1].endLineNumber,i[i.length-1].endColumn).containsRange(t)}var WOe=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},QG=function(s,e){return function(t,i){e(t,i,s)}},_1;class KH{constructor(e){this.name=e}select(e,t,i){if(i.length===0)return 0;const n=i[0].score[0];for(let o=0;o<i.length;o++){const{score:r,completion:a}=i[o];if(r[0]!==n)break;if(a.preselect)return o}return 0}}class JG extends KH{constructor(){super("first")}memorize(e,t,i){}toJSON(){}fromJSON(){}}class HOe extends KH{constructor(){super("recentlyUsed"),this._cache=new Va(300,.66),this._seq=0}memorize(e,t,i){const n=`${e.getLanguageId()}/${i.textLabel}`;this._cache.set(n,{touch:this._seq++,type:i.completion.kind,insertText:i.completion.insertText})}select(e,t,i){if(i.length===0)return 0;const n=e.getLineContent(t.lineNumber).substr(t.column-10,t.column-1);if(/\s$/.test(n))return super.select(e,t,i);const o=i[0].score[0];let r=-1,a=-1,l=-1;for(let c=0;c<i.length&&i[c].score[0]===o;c++){const d=`${e.getLanguageId()}/${i[c].textLabel}`,h=this._cache.peek(d);if(h&&h.touch>l&&h.type===i[c].completion.kind&&h.insertText===i[c].completion.insertText&&(l=h.touch,a=c),i[c].completion.preselect&&r===-1)return r=c}return a!==-1?a:r!==-1?r:0}toJSON(){return this._cache.toJSON()}fromJSON(e){this._cache.clear();const t=0;for(const[i,n]of e)n.touch=t,n.type=typeof n.type=="number"?n.type:mv.fromString(n.type),this._cache.set(i,n);this._seq=this._cache.size}}class VOe extends KH{constructor(){super("recentlyUsedByPrefix"),this._trie=xw.forStrings(),this._seq=0}memorize(e,t,i){const{word:n}=e.getWordUntilPosition(t),o=`${e.getLanguageId()}/${n}`;this._trie.set(o,{type:i.completion.kind,insertText:i.completion.insertText,touch:this._seq++})}select(e,t,i){const{word:n}=e.getWordUntilPosition(t);if(!n)return super.select(e,t,i);const o=`${e.getLanguageId()}/${n}`;let r=this._trie.get(o);if(r||(r=this._trie.findSubstr(o)),r)for(let a=0;a<i.length;a++){const{kind:l,insertText:c}=i[a].completion;if(l===r.type&&c===r.insertText)return a}return super.select(e,t,i)}toJSON(){const e=[];return this._trie.forEach((t,i)=>e.push([i,t])),e.sort((t,i)=>-(t[1].touch-i[1].touch)).forEach((t,i)=>t[1].touch=i),e.slice(0,200)}fromJSON(e){if(this._trie.clear(),e.length>0){this._seq=e[0][1].touch+1;for(const[t,i]of e)i.type=typeof i.type=="number"?i.type:mv.fromString(i.type),this._trie.set(t,i)}}}var Ju;let L3=(Ju=class{constructor(e,t){this._storageService=e,this._configService=t,this._disposables=new U,this._persistSoon=new Rt(()=>this._saveState(),500),this._disposables.add(e.onWillSaveState(i=>{i.reason===Fu.SHUTDOWN&&this._saveState()}))}dispose(){this._disposables.dispose(),this._persistSoon.dispose()}memorize(e,t,i){this._withStrategy(e,t).memorize(e,t,i),this._persistSoon.schedule()}select(e,t,i){return this._withStrategy(e,t).select(e,t,i)}_withStrategy(e,t){const i=this._configService.getValue("editor.suggestSelection",{overrideIdentifier:e.getLanguageIdAtPosition(t.lineNumber,t.column),resource:e.uri});if(this._strategy?.name!==i){this._saveState();const n=_1._strategyCtors.get(i)||JG;this._strategy=new n;try{const r=this._configService.getValue("editor.suggest.shareSuggestSelections")?0:1,a=this._storageService.get(`${_1._storagePrefix}/${i}`,r);a&&this._strategy.fromJSON(JSON.parse(a))}catch{}}return this._strategy}_saveState(){if(this._strategy){const t=this._configService.getValue("editor.suggest.shareSuggestSelections")?0:1,i=JSON.stringify(this._strategy);this._storageService.store(`${_1._storagePrefix}/${this._strategy.name}`,i,t,1)}}},_1=Ju,Ju._strategyCtors=new Map([["recentlyUsedByPrefix",VOe],["recentlyUsed",HOe],["first",JG]]),Ju._storagePrefix="suggest/memories",Ju);L3=_1=WOe([QG(0,Bs),QG(1,He)],L3);const YA=qe("ISuggestMemories");tt(YA,L3,1);var zOe=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},$Oe=function(s,e){return function(t,i){e(t,i,s)}},k3,Zp;let FD=(Zp=class{constructor(e,t){this._editor=e,this._enabled=!1,this._ckAtEnd=k3.AtEnd.bindTo(t),this._configListener=this._editor.onDidChangeConfiguration(i=>i.hasChanged(139)&&this._update()),this._update()}dispose(){this._configListener.dispose(),this._selectionListener?.dispose(),this._ckAtEnd.reset()}_update(){const e=this._editor.getOption(139)==="on";if(this._enabled!==e)if(this._enabled=e,this._enabled){const t=()=>{if(!this._editor.hasModel()){this._ckAtEnd.set(!1);return}const i=this._editor.getModel(),n=this._editor.getSelection(),o=i.getWordAtPosition(n.getStartPosition());if(!o){this._ckAtEnd.set(!1);return}this._ckAtEnd.set(o.endColumn===n.getStartPosition().column&&n.getStartPosition().lineNumber===n.getEndPosition().lineNumber)};this._selectionListener=this._editor.onDidChangeCursorSelection(t),t()}else this._selectionListener&&(this._ckAtEnd.reset(),this._selectionListener.dispose(),this._selectionListener=void 0)}},k3=Zp,Zp.AtEnd=new ce("atEndOfWord",!1,{type:"boolean",description:m(1494,"A context key that is true when at the end of a word. Note that this is only defined when tab-completions are enabled")}),Zp);FD=k3=zOe([$Oe(1,Ae)],FD);var UOe=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},qOe=function(s,e){return function(t,i){e(t,i,s)}},b1,Xp;let Nv=(Xp=class{constructor(e,t){this._editor=e,this._index=0,this._ckOtherSuggestions=b1.OtherSuggestions.bindTo(t)}dispose(){this.reset()}reset(){this._ckOtherSuggestions.reset(),this._listener?.dispose(),this._model=void 0,this._acceptNext=void 0,this._ignore=!1}set({model:e,index:t},i){if(e.items.length===0){this.reset();return}if(b1._moveIndex(!0,e,t)===t){this.reset();return}this._acceptNext=i,this._model=e,this._index=t,this._listener=this._editor.onDidChangeCursorPosition(()=>{this._ignore||this.reset()}),this._ckOtherSuggestions.set(!0)}static _moveIndex(e,t,i){let n=i;for(let o=t.items.length;o>0&&(n=(n+t.items.length+(e?1:-1))%t.items.length,!(n===i||!t.items[n].completion.additionalTextEdits));o--);return n}next(){this._move(!0)}prev(){this._move(!1)}_move(e){if(this._model)try{this._ignore=!0,this._index=b1._moveIndex(e,this._model,this._index),this._acceptNext({index:this._index,item:this._model.items[this._index],model:this._model})}finally{this._ignore=!1}}},b1=Xp,Xp.OtherSuggestions=new ce("hasOtherSuggestions",!1),Xp);Nv=b1=UOe([qOe(1,Ae)],Nv);class jOe{constructor(e,t,i,n){this._disposables=new U,this._disposables.add(i.onDidSuggest(o=>{o.completionModel.items.length===0&&this.reset()})),this._disposables.add(i.onDidCancel(o=>{this.reset()})),this._disposables.add(t.onDidShow(()=>this._onItem(t.getFocusedItem()))),this._disposables.add(t.onDidFocus(this._onItem,this)),this._disposables.add(t.onDidHide(this.reset,this)),this._disposables.add(e.onWillType(o=>{if(this._active&&!t.isFrozen()&&i.state!==0){const r=o.charCodeAt(o.length-1);this._active.acceptCharacters.has(r)&&e.getOption(0)&&n(this._active.item)}}))}_onItem(e){if(!e||!Ms(e.item.completion.commitCharacters)){this.reset();return}if(this._active&&this._active.item.item===e.item)return;const t=new TE;for(const i of e.item.completion.commitCharacters)i.length>0&&t.add(i.charCodeAt(0));this._active={acceptCharacters:t,item:e}}reset(){this._active=void 0}dispose(){this._disposables.dispose()}}const Fr=class Fr{async provideSelectionRanges(e,t){const i=[];for(const n of t){const o=[];i.push(o);const r=new Map;await new Promise(a=>Fr._bracketsRightYield(a,0,e,n,r)),await new Promise(a=>Fr._bracketsLeftYield(a,0,e,n,r,o))}return i}static _bracketsRightYield(e,t,i,n,o){const r=new Map,a=Date.now();for(;;){if(t>=Fr._maxRounds){e();break}if(!n){e();break}const l=i.bracketPairs.findNextBracket(n);if(!l){e();break}if(Date.now()-a>Fr._maxDuration){setTimeout(()=>Fr._bracketsRightYield(e,t+1,i,n,o));break}if(l.bracketInfo.isOpeningBracket){const d=l.bracketInfo.bracketText,h=r.has(d)?r.get(d):0;r.set(d,h+1)}else{const d=l.bracketInfo.getOpeningBrackets()[0].bracketText;let h=r.has(d)?r.get(d):0;if(h-=1,r.set(d,Math.max(0,h)),h<0){let u=o.get(d);u||(u=new Ns,o.set(d,u)),u.push(l.range)}}n=l.range.getEndPosition()}}static _bracketsLeftYield(e,t,i,n,o,r){const a=new Map,l=Date.now();for(;;){if(t>=Fr._maxRounds&&o.size===0){e();break}if(!n){e();break}const c=i.bracketPairs.findPrevBracket(n);if(!c){e();break}if(Date.now()-l>Fr._maxDuration){setTimeout(()=>Fr._bracketsLeftYield(e,t+1,i,n,o,r));break}if(c.bracketInfo.isOpeningBracket){const h=c.bracketInfo.bracketText;let u=a.has(h)?a.get(h):0;if(u-=1,a.set(h,Math.max(0,u)),u<0){const g=o.get(h);if(g){const f=g.shift();g.size===0&&o.delete(h);const p=k.fromPositions(c.range.getEndPosition(),f.getStartPosition()),_=k.fromPositions(c.range.getStartPosition(),f.getEndPosition());r.push({range:p}),r.push({range:_}),Fr._addBracketLeading(i,_,r)}}}else{const h=c.bracketInfo.getOpeningBrackets()[0].bracketText,u=a.has(h)?a.get(h):0;a.set(h,u+1)}n=c.range.getStartPosition()}}static _addBracketLeading(e,t,i){if(t.startLineNumber===t.endLineNumber)return;const n=t.startLineNumber,o=e.getLineFirstNonWhitespaceColumn(n);o!==0&&o!==t.startColumn&&(i.push({range:k.fromPositions(new A(n,o),t.getEndPosition())}),i.push({range:k.fromPositions(new A(n,1),t.getEndPosition())}));const r=n-1;if(r>0){const a=e.getLineFirstNonWhitespaceColumn(r);a===t.startColumn&&a!==e.getLineLastNonWhitespaceColumn(r)&&(i.push({range:k.fromPositions(new A(r,a),t.getEndPosition())}),i.push({range:k.fromPositions(new A(r,1),t.getEndPosition())}))}}};Fr._maxDuration=30,Fr._maxRounds=2;let BD=Fr;const ic=class ic{static async create(e,t){if(!t.getOption(134).localityBonus||!t.hasModel())return ic.None;const i=t.getModel(),n=t.getPosition();if(!e.canComputeWordRanges(i.uri))return ic.None;const[o]=await new BD().provideSelectionRanges(i,[n]);if(o.length===0)return ic.None;const r=await e.computeWordRanges(i.uri,o[0].range);if(!r)return ic.None;const a=i.getWordUntilPosition(n);return delete r[a.word],new class extends ic{distance(l,c){if(!n.equals(t.getPosition()))return 0;if(c.kind===17)return 2<<20;const d=typeof c.label=="string"?c.label:c.label.label,h=r[d];if(pQ(h))return 2<<20;const u=vE(h,k.fromPositions(l),k.compareRangesUsingStarts),g=u>=0?h[u]:h[Math.max(0,~u-1)];let f=o.length;for(const p of o){if(!k.containsRange(p.range,g))break;f-=1}return f}}}};ic.None=new class extends ic{distance(){return 0}};let WD=ic,eK=class{constructor(e,t){this.leadingLineContent=e,this.characterCountDelta=t}};class cu{constructor(e,t,i,n,o,r,a=TS.default,l=void 0){this.clipboardText=l,this._snippetCompareFn=cu._compareCompletionItems,this._items=e,this._column=t,this._wordDistance=n,this._options=o,this._refilterKind=1,this._lineContext=i,this._fuzzyScoreOptions=a,r==="top"?this._snippetCompareFn=cu._compareCompletionItemsSnippetsUp:r==="bottom"&&(this._snippetCompareFn=cu._compareCompletionItemsSnippetsDown)}get lineContext(){return this._lineContext}set lineContext(e){(this._lineContext.leadingLineContent!==e.leadingLineContent||this._lineContext.characterCountDelta!==e.characterCountDelta)&&(this._refilterKind=this._lineContext.characterCountDelta<e.characterCountDelta&&this._filteredItems?2:1,this._lineContext=e)}get items(){return this._ensureCachedState(),this._filteredItems}getItemsByProvider(){return this._ensureCachedState(),this._itemsByProvider}getIncompleteProvider(){this._ensureCachedState();const e=new Set;for(const[t,i]of this.getItemsByProvider())i.length>0&&i[0].container.incomplete&&e.add(t);return e}get stats(){return this._ensureCachedState(),this._stats}_ensureCachedState(){this._refilterKind!==0&&this._createCachedState()}_createCachedState(){this._itemsByProvider=new Map;const e=[],{leadingLineContent:t,characterCountDelta:i}=this._lineContext;let n="",o="";const r=this._refilterKind===1?this._items:this._filteredItems,a=[],l=!this._options.filterGraceful||r.length>2e3?Om:d0e;for(let c=0;c<r.length;c++){const d=r[c];if(d.isInvalid)continue;const h=this._itemsByProvider.get(d.provider);h?h.push(d):this._itemsByProvider.set(d.provider,[d]);const u=d.position.column-d.editStart.column,g=u+i-(d.position.column-this._column);if(n.length!==g&&(n=g===0?"":t.slice(-g),o=n.toLowerCase()),d.word=n,g===0)d.score=Oa.Default;else{let f=0;for(;f<u;){const p=n.charCodeAt(f);if(p===32||p===9)f+=1;else break}if(f>=g)d.score=Oa.Default;else if(typeof d.completion.filterText=="string"){const p=l(n,o,f,d.completion.filterText,d.filterTextLow,0,this._fuzzyScoreOptions);if(!p)continue;hP(d.completion.filterText,d.textLabel)===0?d.score=p:(d.score=r0e(n,o,f,d.textLabel,d.labelLow,0),d.score[0]=p[0])}else{const p=l(n,o,f,d.textLabel,d.labelLow,0,this._fuzzyScoreOptions);if(!p)continue;d.score=p}}d.idx=c,d.distance=this._wordDistance.distance(d.position,d.completion),a.push(d),e.push(d.textLabel.length)}this._filteredItems=a.sort(this._snippetCompareFn),this._refilterKind=0,this._stats={pLabelLen:e.length?bP(e.length-.85,e,(c,d)=>c-d):0}}static _compareCompletionItems(e,t){return e.score[0]>t.score[0]?-1:e.score[0]<t.score[0]?1:e.distance<t.distance?-1:e.distance>t.distance?1:e.idx<t.idx?-1:e.idx>t.idx?1:0}static _compareCompletionItemsSnippetsDown(e,t){if(e.completion.kind!==t.completion.kind){if(e.completion.kind===28)return 1;if(t.completion.kind===28)return-1}return cu._compareCompletionItems(e,t)}static _compareCompletionItemsSnippetsUp(e,t){if(e.completion.kind!==t.completion.kind){if(e.completion.kind===28)return-1;if(t.completion.kind===28)return 1}return cu._compareCompletionItems(e,t)}}var GOe=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},Vh=function(s,e){return function(t,i){e(t,i,s)}},E3;class Yg{static shouldAutoTrigger(e){if(!e.hasModel())return!1;const t=e.getModel(),i=e.getPosition();t.tokenization.tokenizeIfCheap(i.lineNumber);const n=t.getWordAtPosition(i);return!(!n||n.endColumn!==i.column&&n.startColumn+1!==i.column||!isNaN(Number(n.word)))}constructor(e,t,i){this.leadingLineContent=e.getLineContent(t.lineNumber).substr(0,t.column-1),this.leadingWord=e.getWordUntilPosition(t),this.lineNumber=t.lineNumber,this.column=t.column,this.triggerOptions=i}}function KOe(s,e,t){if(!e.getContextKeyValue(Ot.inlineSuggestionVisible.key))return!0;const i=e.getContextKeyValue(Ot.suppressSuggestions.key);return i!==void 0?!i:!s.getOption(71).suppressSuggestions}function YOe(s,e,t){if(!e.getContextKeyValue("inlineSuggestionVisible"))return!0;const i=e.getContextKeyValue(Ot.suppressSuggestions.key);return i!==void 0?!i:!s.getOption(71).suppressSuggestions}let HD=E3=class{constructor(e,t,i,n,o,r,a,l,c){this._editor=e,this._editorWorkerService=t,this._clipboardService=i,this._telemetryService=n,this._logService=o,this._contextKeyService=r,this._configurationService=a,this._languageFeaturesService=l,this._envService=c,this._toDispose=new U,this._triggerCharacterListener=new U,this._triggerQuickSuggest=new $o,this._triggerState=void 0,this._completionDisposables=new U,this._onDidCancel=new M,this._onDidTrigger=new M,this._onDidSuggest=new M,this.onDidCancel=this._onDidCancel.event,this.onDidTrigger=this._onDidTrigger.event,this.onDidSuggest=this._onDidSuggest.event,this._currentSelection=this._editor.getSelection()||new he(1,1,1,1),this._toDispose.add(this._editor.onDidChangeModel(()=>{this._updateTriggerCharacters(),this.cancel()})),this._toDispose.add(this._editor.onDidChangeModelLanguage(()=>{this._updateTriggerCharacters(),this.cancel()})),this._toDispose.add(this._editor.onDidChangeConfiguration(()=>{this._updateTriggerCharacters()})),this._toDispose.add(this._languageFeaturesService.completionProvider.onDidChange(()=>{this._updateTriggerCharacters(),this._updateActiveSuggestSession()}));let d=!1;this._toDispose.add(this._editor.onDidCompositionStart(()=>{d=!0})),this._toDispose.add(this._editor.onDidCompositionEnd(()=>{d=!1,this._onCompositionEnd()})),this._toDispose.add(this._editor.onDidChangeCursorSelection(h=>{d||this._onCursorChange(h)})),this._toDispose.add(this._editor.onDidChangeModelContent(()=>{!d&&this._triggerState!==void 0&&this._refilterCompletionItems()})),this._updateTriggerCharacters()}dispose(){xt(this._triggerCharacterListener),xt([this._onDidCancel,this._onDidSuggest,this._onDidTrigger,this._triggerQuickSuggest]),this._toDispose.dispose(),this._completionDisposables.dispose(),this.cancel()}_updateTriggerCharacters(){if(this._triggerCharacterListener.clear(),this._editor.getOption(104)||!this._editor.hasModel()||!this._editor.getOption(137))return;const e=new Map;for(const i of this._languageFeaturesService.completionProvider.all(this._editor.getModel()))for(const n of i.triggerCharacters||[]){let o=e.get(n);o||(o=new Set,e.set(n,o)),o.add(i)}const t=i=>{if(!YOe(this._editor,this._contextKeyService,this._configurationService)||Yg.shouldAutoTrigger(this._editor))return;if(!i){const r=this._editor.getPosition();i=this._editor.getModel().getLineContent(r.lineNumber).substr(0,r.column-1)}let n="";ag(i.charCodeAt(i.length-1))?sn(i.charCodeAt(i.length-2))&&(n=i.substr(i.length-2)):n=i.charAt(i.length-1);const o=e.get(n);if(o){const r=new Map;if(this._completionModel)for(const[a,l]of this._completionModel.getItemsByProvider())o.has(a)||r.set(a,l);this.trigger({auto:!0,triggerKind:1,triggerCharacter:n,retrigger:!!this._completionModel,clipboardText:this._completionModel?.clipboardText,completionOptions:{providerFilter:o,providerItemsToReuse:r}})}};this._triggerCharacterListener.add(this._editor.onDidType(t)),this._triggerCharacterListener.add(this._editor.onDidCompositionEnd(()=>t()))}get state(){return this._triggerState?this._triggerState.auto?2:1:0}cancel(e=!1){this._triggerState!==void 0&&(this._triggerQuickSuggest.cancel(),this._requestToken?.cancel(),this._requestToken=void 0,this._triggerState=void 0,this._completionModel=void 0,this._context=void 0,this._onDidCancel.fire({retrigger:e}))}clear(){this._completionDisposables.clear()}_updateActiveSuggestSession(){this._triggerState!==void 0&&(!this._editor.hasModel()||!this._languageFeaturesService.completionProvider.has(this._editor.getModel())?this.cancel():this.trigger({auto:this._triggerState.auto,retrigger:!0}))}_onCursorChange(e){if(!this._editor.hasModel())return;const t=this._currentSelection;if(this._currentSelection=this._editor.getSelection(),!e.selection.isEmpty()||e.reason!==0&&e.reason!==3||e.source!=="keyboard"&&e.source!=="deleteLeft"){this.cancel();return}this._triggerState===void 0&&e.reason===0?(t.containsRange(this._currentSelection)||t.getEndPosition().isBeforeOrEqual(this._currentSelection.getPosition()))&&this._doTriggerQuickSuggest():this._triggerState!==void 0&&e.reason===3&&this._refilterCompletionItems()}_onCompositionEnd(){this._triggerState===void 0?this._doTriggerQuickSuggest():this._refilterCompletionItems()}_doTriggerQuickSuggest(){Ob.isAllOff(this._editor.getOption(102))||this._editor.getOption(134).snippetsPreventQuickSuggestions&&Ps.get(this._editor)?.isInSnippet()||(this.cancel(),this._triggerQuickSuggest.cancelAndSet(()=>{if(this._triggerState!==void 0||!Yg.shouldAutoTrigger(this._editor)||!this._editor.hasModel()||!this._editor.hasWidgetFocus())return;const e=this._editor.getModel(),t=this._editor.getPosition(),i=this._editor.getOption(102);if(!Ob.isAllOff(i)){if(!Ob.isAllOn(i)){e.tokenization.tokenizeIfCheap(t.lineNumber);const n=e.tokenization.getLineTokens(t.lineNumber),o=n.getStandardTokenType(n.findTokenIndexAtOffset(Math.max(t.column-1-1,0)));if(Ob.valueFor(i,o)!=="on")return}KOe(this._editor,this._contextKeyService,this._configurationService)&&this._languageFeaturesService.completionProvider.has(e)&&this.trigger({auto:!0})}},this._editor.getOption(103)))}_refilterCompletionItems(){dt(this._editor.hasModel()),dt(this._triggerState!==void 0);const e=this._editor.getModel(),t=this._editor.getPosition(),i=new Yg(e,t,{...this._triggerState,refilter:!0});this._onNewContext(i)}trigger(e){if(!this._editor.hasModel())return;const t=this._editor.getModel(),i=new Yg(t,this._editor.getPosition(),e);this.cancel(e.retrigger),this._triggerState=e,this._onDidTrigger.fire({auto:e.auto,shy:e.shy??!1,position:this._editor.getPosition()}),this._context=i;let n={triggerKind:e.triggerKind??0};e.triggerCharacter&&(n={triggerKind:1,triggerCharacter:e.triggerCharacter}),this._requestToken=new si;const o=this._editor.getOption(128);let r=1;switch(o){case"top":r=0;break;case"bottom":r=2;break}const{itemKind:a,showDeprecated:l}=E3.createSuggestFilter(this._editor),c=new n0(r,e.completionOptions?.kindFilter??a,e.completionOptions?.providerFilter,e.completionOptions?.providerItemsToReuse,l),d=WD.create(this._editorWorkerService,this._editor),h=UH(this._languageFeaturesService.completionProvider,t,this._editor.getPosition(),c,n,this._requestToken.token);Promise.all([h,d]).then(async([u,g])=>{if(this._requestToken?.dispose(),!this._editor.hasModel()){u.disposable.dispose();return}let f=e?.clipboardText;if(!f&&u.needsClipboard&&(f=await this._clipboardService.readText()),this._triggerState===void 0){u.disposable.dispose();return}const p=this._editor.getModel(),_=new Yg(p,this._editor.getPosition(),e),b={...TS.default,firstMatchCanBeWeak:!this._editor.getOption(134).matchOnWordStartOnly};if(this._completionModel=new cu(u.items,this._context.column,{leadingLineContent:_.leadingLineContent,characterCountDelta:_.column-this._context.column},g,this._editor.getOption(134),this._editor.getOption(128),b,f),this._completionDisposables.add(u.disposable),this._onNewContext(_),this._reportDurationsTelemetry(u.durations),!this._envService.isBuilt||this._envService.isExtensionDevelopment)for(const w of u.items)w.isInvalid&&this._logService.warn(`[suggest] did IGNORE invalid completion item from ${w.provider._debugDisplayName}`,w.completion)}).catch(Oe)}_reportDurationsTelemetry(e){Math.random()>1e-4||setTimeout(()=>{this._telemetryService.publicLog2("suggest.durations.json",{data:JSON.stringify(e)}),this._logService.debug("suggest.durations.json",e)})}static createSuggestFilter(e){const t=new Set;e.getOption(128)==="none"&&t.add(28);const n=e.getOption(134);return n.showMethods||t.add(0),n.showFunctions||t.add(1),n.showConstructors||t.add(2),n.showFields||t.add(3),n.showVariables||t.add(4),n.showClasses||t.add(5),n.showStructs||t.add(6),n.showInterfaces||t.add(7),n.showModules||t.add(8),n.showProperties||t.add(9),n.showEvents||t.add(10),n.showOperators||t.add(11),n.showUnits||t.add(12),n.showValues||t.add(13),n.showConstants||t.add(14),n.showEnums||t.add(15),n.showEnumMembers||t.add(16),n.showKeywords||t.add(17),n.showWords||t.add(18),n.showColors||t.add(19),n.showFiles||t.add(20),n.showReferences||t.add(21),n.showColors||t.add(22),n.showFolders||t.add(23),n.showTypeParameters||t.add(24),n.showSnippets||t.add(28),n.showUsers||t.add(25),n.showIssues||t.add(26),{itemKind:t,showDeprecated:n.showDeprecated}}_onNewContext(e){if(this._context){if(e.lineNumber!==this._context.lineNumber){this.cancel();return}if(Bt(e.leadingLineContent)!==Bt(this._context.leadingLineContent)){this.cancel();return}if(e.column<this._context.column){e.leadingWord.word?this.trigger({auto:this._context.triggerOptions.auto,retrigger:!0}):this.cancel();return}if(this._completionModel){if(e.leadingWord.word.length!==0&&e.leadingWord.startColumn>this._context.leadingWord.startColumn){if(Yg.shouldAutoTrigger(this._editor)&&this._context){const i=this._completionModel.getItemsByProvider();this.trigger({auto:this._context.triggerOptions.auto,retrigger:!0,clipboardText:this._completionModel.clipboardText,completionOptions:{providerItemsToReuse:i}})}return}if(e.column>this._context.column&&this._completionModel.getIncompleteProvider().size>0&&e.leadingWord.word.length!==0){const t=new Map,i=new Set;for(const[n,o]of this._completionModel.getItemsByProvider())o.length>0&&o[0].container.incomplete?i.add(n):t.set(n,o);this.trigger({auto:this._context.triggerOptions.auto,triggerKind:2,retrigger:!0,clipboardText:this._completionModel.clipboardText,completionOptions:{providerFilter:i,providerItemsToReuse:t}})}else{const t=this._completionModel.lineContext;let i=!1;if(this._completionModel.lineContext={leadingLineContent:e.leadingLineContent,characterCountDelta:e.column-this._context.column},this._completionModel.items.length===0){const n=Yg.shouldAutoTrigger(this._editor);if(!this._context){this.cancel();return}if(n&&this._context.leadingWord.endColumn<e.leadingWord.startColumn){this.trigger({auto:this._context.triggerOptions.auto,retrigger:!0});return}if(this._context.triggerOptions.auto){this.cancel();return}else if(this._completionModel.lineContext=t,i=this._completionModel.items.length>0,i&&e.leadingWord.word.length===0){this.cancel();return}}this._onDidSuggest.fire({completionModel:this._completionModel,triggerOptions:e.triggerOptions,isFrozen:i})}}}}};HD=E3=GOe([Vh(1,ko),Vh(2,Ko),Vh(3,vs),Vh(4,jt),Vh(5,Ae),Vh(6,He),Vh(7,ge),Vh(8,nH)],HD);const PN=class PN{constructor(e,t){this._disposables=new U,this._lastOvertyped=[],this._locked=!1,this._disposables.add(e.onWillType(()=>{if(this._locked||!e.hasModel())return;const i=e.getSelections(),n=i.length;let o=!1;for(let a=0;a<n;a++)if(!i[a].isEmpty()){o=!0;break}if(!o){this._lastOvertyped.length!==0&&(this._lastOvertyped.length=0);return}this._lastOvertyped=[];const r=e.getModel();for(let a=0;a<n;a++){const l=i[a];if(r.getValueLengthInRange(l)>PN._maxSelectionLength)return;this._lastOvertyped[a]={value:r.getValueInRange(l),multiline:l.startLineNumber!==l.endLineNumber}}})),this._disposables.add(t.onDidTrigger(i=>{this._locked=!0})),this._disposables.add(t.onDidCancel(i=>{this._locked=!1}))}getLastOvertypedInfo(e){if(e>=0&&e<this._lastOvertyped.length)return this._lastOvertyped[e]}dispose(){this._disposables.dispose()}};PN._maxSelectionLength=51200;let I3=PN;var ZOe=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},gO=function(s,e){return function(t,i){e(t,i,s)}};let D3=class{constructor(e,t,i,n,o){this._menuId=t,this._menuService=n,this._contextKeyService=o,this._menuDisposables=new U,this.element=X(e,ie(".suggest-status-bar"));const r=(a=>a instanceof yr?i.createInstance(U7,a,{useComma:!1}):void 0);this._leftActions=new bo(this.element,{actionViewItemProvider:r}),this._rightActions=new bo(this.element,{actionViewItemProvider:r}),this._leftActions.domNode.classList.add("left"),this._rightActions.domNode.classList.add("right")}dispose(){this._menuDisposables.dispose(),this._leftActions.dispose(),this._rightActions.dispose(),this.element.remove()}show(){const e=this._menuService.createMenu(this._menuId,this._contextKeyService),t=()=>{const i=[],n=[];for(const[o,r]of e.getActions())o==="left"?i.push(...r):n.push(...r);this._leftActions.clear(),this._leftActions.push(i),this._rightActions.clear(),this._rightActions.push(n)};this._menuDisposables.add(e.onDidChange(()=>t())),this._menuDisposables.add(e)}hide(){this._menuDisposables.clear()}};D3=ZOe([gO(2,_e),gO(3,da),gO(4,Ae)],D3);var XOe=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},tK=function(s,e){return function(t,i){e(t,i,s)}};function YH(s){return!!s&&!!(s.completion.documentation||s.completion.detail&&s.completion.detail!==s.completion.label)}let T3=class{constructor(e,t,i){this._editor=e,this._themeService=t,this._markdownRendererService=i,this._onDidClose=new M,this.onDidClose=this._onDidClose.event,this._onDidChangeContents=new M,this.onDidChangeContents=this._onDidChangeContents.event,this._disposables=new U,this._renderDisposeable=new U,this._size=new St(330,0),this.domNode=ie(".suggest-details"),this.domNode.classList.add("no-docs"),this._body=ie(".body"),this._scrollbar=new tx(this._body,{alwaysConsumeMouseWheel:!0}),X(this.domNode,this._scrollbar.getDomNode()),this._disposables.add(this._scrollbar),this._header=X(this._body,ie(".header")),this._close=X(this._header,ie("span"+xe.asCSSSelector(Z.close))),this._close.title=m(1490,"Close"),this._close.role="button",this._close.tabIndex=-1,this._type=X(this._header,ie("p.type")),this._docs=X(this._body,ie("p.docs")),this._configureFont(),this._disposables.add(this._editor.onDidChangeConfiguration(n=>{n.hasChanged(59)&&this._configureFont()}))}dispose(){this._disposables.dispose(),this._renderDisposeable.dispose()}_configureFont(){const e=this._editor.getOptions(),t=e.get(59),i=t.getMassagedFontFamily(),n=e.get(135)||t.fontSize,o=e.get(136)||t.lineHeight,r=t.fontWeight,a=`${n}px`,l=`${o}px`;this.domNode.style.fontSize=a,this.domNode.style.lineHeight=`${o/n}`,this.domNode.style.fontWeight=r,this.domNode.style.fontFeatureSettings=t.fontFeatureSettings,this._type.style.fontFamily=i,this._close.style.height=l,this._close.style.width=l}getLayoutInfo(){const e=this._editor.getOption(136)||this._editor.getOption(59).lineHeight,t=Cl(this._themeService.getColorTheme().type)?2:1,i=t*2;return{lineHeight:e,borderWidth:t,borderHeight:i,verticalPadding:22,horizontalPadding:14}}renderLoading(){this._type.textContent=m(1491,"Loading..."),this._docs.textContent="",this.domNode.classList.remove("no-docs","no-type"),this.layout(this.size.width,this.getLayoutInfo().lineHeight*2),this._onDidChangeContents.fire(this)}renderItem(e,t){this._renderDisposeable.clear();let{detail:i,documentation:n}=e.completion;if(t){let o="";o+=`score: ${e.score[0]}
`,o+=`prefix: ${e.word??"(no prefix)"}
`,o+=`word: ${e.completion.filterText?e.completion.filterText+" (filterText)":e.textLabel}
`,o+=`distance: ${e.distance} (localityBonus-setting)
`,o+=`index: ${e.idx}, based on ${e.completion.sortText&&`sortText: "${e.completion.sortText}"`||"label"}
`,o+=`commit_chars: ${e.completion.commitCharacters?.join("")}
`,n=new hs().appendCodeblock("empty",o),i=`Provider: ${e.provider._debugDisplayName}`}if(!t&&!YH(e)){this.clearContents();return}if(this.domNode.classList.remove("no-docs","no-type"),i){const o=i.length>1e5?`${i.substr(0,1e5)}…`:i;this._type.textContent=o,this._type.title=o,Mo(this._type),this._type.classList.toggle("auto-wrap",!/\r?\n^\s+/gmi.test(o))}else Pn(this._type),this._type.title="",js(this._type),this.domNode.classList.add("no-type");if(Pn(this._docs),typeof n=="string")this._docs.classList.remove("markdown-docs"),this._docs.textContent=n;else if(n){this._docs.classList.add("markdown-docs"),Pn(this._docs);const o=this._markdownRendererService.render(n,{context:this._editor,asyncRenderCallback:()=>{this.layout(this._size.width,this._type.clientHeight+this._docs.clientHeight),this._onDidChangeContents.fire(this)}});this._docs.appendChild(o.element),this._renderDisposeable.add(o)}this.domNode.classList.toggle("detail-and-doc",!!i&&!!n),this.domNode.style.userSelect="text",this.domNode.tabIndex=-1,this._close.onmousedown=o=>{o.preventDefault(),o.stopPropagation()},this._close.onclick=o=>{o.preventDefault(),o.stopPropagation(),this._onDidClose.fire()},this._body.scrollTop=0,this.layout(this._size.width,this._type.clientHeight+this._docs.clientHeight),this._onDidChangeContents.fire(this)}clearContents(){this.domNode.classList.add("no-docs"),this._type.textContent="",this._docs.textContent=""}get isEmpty(){return this.domNode.classList.contains("no-docs")}get size(){return this._size}layout(e,t){const i=new St(e,t);St.equals(i,this._size)||(this._size=i,sue(this.domNode,e,t)),this._scrollbar.scanDomNode()}scrollDown(e=8){this._body.scrollTop+=e}scrollUp(e=8){this._body.scrollTop-=e}scrollTop(){this._body.scrollTop=0}scrollBottom(){this._body.scrollTop=this._body.scrollHeight}pageDown(){this.scrollDown(80)}pageUp(){this.scrollUp(80)}focus(){this.domNode.focus()}};T3=XOe([tK(1,wi),tK(2,Ua)],T3);class QOe{constructor(e,t){this.widget=e,this._editor=t,this.allowEditorOverflow=!0,this._disposables=new U,this._added=!1,this._preferAlignAtTop=!0,this._resizable=new rH,this._resizable.domNode.classList.add("suggest-details-container"),this._resizable.domNode.appendChild(e.domNode),this._resizable.enableSashes(!1,!0,!0,!1);let i,n,o=0,r=0;this._disposables.add(this._resizable.onDidWillResize(()=>{i=this._topLeft,n=this._resizable.size})),this._disposables.add(this._resizable.onDidResize(a=>{if(i&&n){this.widget.layout(a.dimension.width,a.dimension.height);let l=!1;a.west&&(r=n.width-a.dimension.width,l=!0),a.north&&(o=n.height-a.dimension.height,l=!0),l&&this._applyTopLeft({top:i.top+o,left:i.left+r})}a.done&&(i=void 0,n=void 0,o=0,r=0,this._userSize=a.dimension)})),this._disposables.add(this.widget.onDidChangeContents(()=>{this._anchorBox&&this._placeAtAnchor(this._anchorBox,this._userSize??this.widget.size,this._preferAlignAtTop)}))}dispose(){this._resizable.dispose(),this._disposables.dispose(),this.hide()}getId(){return"suggest.details"}getDomNode(){return this._resizable.domNode}getPosition(){return this._topLeft?{preference:this._topLeft}:null}show(){this._added||(this._editor.addOverlayWidget(this),this._added=!0)}hide(e=!1){this._resizable.clearSashHoverState(),this._added&&(this._editor.removeOverlayWidget(this),this._added=!1,this._anchorBox=void 0,this._topLeft=void 0),e&&(this._userSize=void 0,this.widget.clearContents())}placeAtAnchor(e,t){const i=e.getBoundingClientRect();this._anchorBox=i,this._preferAlignAtTop=t,this._placeAtAnchor(this._anchorBox,this._userSize??this.widget.size,t)}_placeAtAnchor(e,t,i){const n=lg(this.getDomNode().ownerDocument.body),o=this.widget.getLayoutInfo(),r=new St(220,2*o.lineHeight),a=e.top,l=(function(){const y=n.width-(e.left+e.width+o.borderWidth+o.horizontalPadding),x=-o.borderWidth+e.left+e.width,L=new St(y,n.height-e.top-o.borderHeight-o.verticalPadding),E=L.with(void 0,e.top+e.height-o.borderHeight-o.verticalPadding);return{top:a,left:x,fit:y-t.width,maxSizeTop:L,maxSizeBottom:E,minSize:r.with(Math.min(y,r.width))}})(),c=(function(){const y=e.left-o.borderWidth-o.horizontalPadding,x=Math.max(o.horizontalPadding,e.left-t.width-o.borderWidth),L=new St(y,n.height-e.top-o.borderHeight-o.verticalPadding),E=L.with(void 0,e.top+e.height-o.borderHeight-o.verticalPadding);return{top:a,left:x,fit:y-t.width,maxSizeTop:L,maxSizeBottom:E,minSize:r.with(Math.min(y,r.width))}})(),d=(function(){const y=e.left,x=-o.borderWidth+e.top+e.height,L=new St(e.width-o.borderHeight,n.height-e.top-e.height-o.verticalPadding);return{top:x,left:y,fit:L.height-t.height,maxSizeBottom:L,maxSizeTop:L,minSize:r.with(L.width)}})(),h=[l,c,d],u=h.find(y=>y.fit>=0)??h.sort((y,x)=>x.fit-y.fit)[0],g=e.top+e.height-o.borderHeight;let f,p=t.height;const _=Math.max(u.maxSizeTop.height,u.maxSizeBottom.height);p>_&&(p=_);let b;i?p<=u.maxSizeTop.height?(f=!0,b=u.maxSizeTop):(f=!1,b=u.maxSizeBottom):p<=u.maxSizeBottom.height?(f=!1,b=u.maxSizeBottom):(f=!0,b=u.maxSizeTop);let{top:w,left:C}=u;!f&&p>e.height&&(w=g-p);const v=this._editor.getDomNode();if(v){const y=v.getBoundingClientRect();w-=y.top,C-=y.left}this._applyTopLeft({left:C,top:w}),this._resizable.enableSashes(!f,u===l,f,u!==l),this._resizable.minSize=u.minSize,this._resizable.maxSize=b,this._resizable.layout(p,Math.min(b.width,t.width)),this.widget.layout(this._resizable.size.width,this._resizable.size.height)}_applyTopLeft(e){this._topLeft=e,this._editor.layoutOverlayWidget(this)}}const iK=qe("fileService");var Sc;(function(s){s[s.FILE=0]="FILE",s[s.FOLDER=1]="FOLDER",s[s.ROOT_FOLDER=2]="ROOT_FOLDER"})(Sc||(Sc={}));const JOe=/(?:\/|^)(?:([^\/]+)\/)?([^\/]+)$/;function UL(s,e,t,i,n){if(xe.isThemeIcon(n))return[`codicon-${n.id}`,"predefined-file-icon"];if(Le.isUri(n))return[];const o=i===Sc.ROOT_FOLDER?["rootfolder-icon"]:i===Sc.FOLDER?["folder-icon"]:["file-icon"];if(t){let r;if(t.scheme===Te.data)r=hg.parseMetaData(t).get(hg.META_DATA_LABEL);else{const a=t.path.match(JOe);a?(r=qL(a[2].toLowerCase()),a[1]&&o.push(`${qL(a[1].toLowerCase())}-name-dir-icon`)):r=qL(t.authority.toLowerCase())}if(i===Sc.ROOT_FOLDER)o.push(`${r}-root-name-folder-icon`);else if(i===Sc.FOLDER)o.push(`${r}-name-folder-icon`);else{if(r){if(o.push(`${r}-name-file-icon`),o.push("name-file-icon"),r.length<=255){const l=r.split(".");for(let c=1;c<l.length;c++)o.push(`${l.slice(c).join(".")}-ext-file-icon`)}o.push("ext-file-icon")}const a=ePe(s,e,t);a&&o.push(`${qL(a)}-lang-file-icon`)}}return o}function ePe(s,e,t){if(!t)return null;let i=null;if(t.scheme===Te.data){const o=hg.parseMetaData(t).get(hg.META_DATA_MIME);o&&(i=e.getLanguageIdByMimeType(o))}else{const n=s.getModel(t);n&&(i=n.getLanguageId())}return i&&i!==xr?i:e.guessLanguageIdByFilepathOrFirstLine(t)}function qL(s){return s.replace(/[\s]/g,"/")}var tPe=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},fO=function(s,e){return function(t,i){e(t,i,s)}};const iPe=Qt("suggest-more-info",Z.chevronRight,m(1492,"Icon for more information in the suggest widget."));var hl;const nPe=new(hl=class{extract(e,t){if(e.textLabel.match(hl._regexStrict))return t[0]=e.textLabel,!0;if(e.completion.detail&&e.completion.detail.match(hl._regexStrict))return t[0]=e.completion.detail,!0;if(e.completion.documentation){const i=typeof e.completion.documentation=="string"?e.completion.documentation:e.completion.documentation.value,n=hl._regexRelaxed.exec(i);if(n&&(n.index===0||n.index+n[0].length===i.length))return t[0]=n[0],!0}return!1}},hl._regexRelaxed=/(#([\da-fA-F]{3}){1,2}|(rgb|hsl)a\(\s*(\d{1,3}%?\s*,\s*){3}(1|0?\.\d+)\)|(rgb|hsl)\(\s*\d{1,3}%?(\s*,\s*\d{1,3}%?){2}\s*\))/,hl._regexStrict=new RegExp(`^${hl._regexRelaxed.source}$`,"i"),hl);let N3=class{constructor(e,t,i,n){this._editor=e,this._modelService=t,this._languageService=i,this._themeService=n,this._onDidToggleDetails=new M,this.onDidToggleDetails=this._onDidToggleDetails.event,this.templateId="suggestion"}dispose(){this._onDidToggleDetails.dispose()}renderTemplate(e){const t=new U,i=e;i.classList.add("show-file-icons");const n=X(e,ie(".icon")),o=X(n,ie("span.colorspan")),r=X(e,ie(".contents")),a=X(r,ie(".main")),l=X(a,ie(".icon-label.codicon")),c=X(a,ie("span.left")),d=X(a,ie("span.right")),h=new QS(c,{supportHighlights:!0,supportIcons:!0});t.add(h);const u=X(c,ie("span.signature-label")),g=X(c,ie("span.qualifier-label")),f=X(d,ie("span.details-label")),p=X(d,ie("span.readMore"+xe.asCSSSelector(iPe)));return p.title=m(1493,"Read More"),{root:i,left:c,right:d,icon:n,colorspan:o,iconLabel:h,iconContainer:l,parametersLabel:u,qualifierLabel:g,detailsLabel:f,readMore:p,disposables:t,configureFont:()=>{const b=this._editor.getOptions(),w=b.get(59),C=w.getMassagedFontFamily(),v=w.fontFeatureSettings,y=w.fontVariationSettings,x=b.get(135)||w.fontSize,L=b.get(136)||w.lineHeight,E=w.fontWeight,I=w.letterSpacing,R=`${x}px`,O=`${L}px`,P=`${I}px`;i.style.fontSize=R,i.style.fontWeight=E,i.style.letterSpacing=P,a.style.fontFamily=C,a.style.fontFeatureSettings=v,a.style.fontVariationSettings=y,a.style.lineHeight=O,n.style.height=O,n.style.width=O,p.style.height=O,p.style.width=O}}}renderElement(e,t,i){i.configureFont();const{completion:n}=e;i.colorspan.style.backgroundColor="";const o={labelEscapeNewLines:!0,matches:ox(e.score)},r=[];if(n.kind===19&&nPe.extract(e,r))i.icon.className="icon customcolor",i.iconContainer.className="icon hide",i.colorspan.style.backgroundColor=r[0];else if(n.kind===20&&this._themeService.getFileIconTheme().hasFileIcons){i.icon.className="icon hide",i.iconContainer.className="icon hide";const a=UL(this._modelService,this._languageService,Le.from({scheme:"fake",path:e.textLabel}),Sc.FILE),l=UL(this._modelService,this._languageService,Le.from({scheme:"fake",path:n.detail}),Sc.FILE);o.extraClasses=a.length>l.length?a:l}else n.kind===23&&this._themeService.getFileIconTheme().hasFolderIcons?(i.icon.className="icon hide",i.iconContainer.className="icon hide",o.extraClasses=[UL(this._modelService,this._languageService,Le.from({scheme:"fake",path:e.textLabel}),Sc.FOLDER),UL(this._modelService,this._languageService,Le.from({scheme:"fake",path:n.detail}),Sc.FOLDER)].flat()):(i.icon.className="icon hide",i.iconContainer.className="",i.iconContainer.classList.add("suggest-icon",...xe.asClassNameArray(mv.toIcon(n.kind))));n.tags&&n.tags.indexOf(1)>=0&&(o.extraClasses=(o.extraClasses||[]).concat(["deprecated"]),o.matches=[]),i.iconLabel.setLabel(e.textLabel,void 0,o),typeof n.label=="string"?(i.parametersLabel.textContent="",i.detailsLabel.textContent=pO(n.detail||""),i.root.classList.add("string-label")):(i.parametersLabel.textContent=pO(n.label.detail||""),i.detailsLabel.textContent=pO(n.label.description||""),i.root.classList.remove("string-label")),this._editor.getOption(134).showInlineDetails?Mo(i.detailsLabel):js(i.detailsLabel),YH(e)?(i.right.classList.add("can-expand-details"),Mo(i.readMore),i.readMore.onmousedown=a=>{a.stopPropagation(),a.preventDefault()},i.readMore.onclick=a=>{a.stopPropagation(),a.preventDefault(),this._onDidToggleDetails.fire()}):(i.right.classList.remove("can-expand-details"),js(i.readMore),i.readMore.onmousedown=null,i.readMore.onclick=null)}disposeTemplate(e){e.disposables.dispose()}};N3=tPe([fO(1,hi),fO(2,Ei),fO(3,wi)],N3);function pO(s){return s.replace(/\r\n|\r|\n/g,"")}var sPe=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},jL=function(s,e){return function(t,i){e(t,i,s)}},_b;T("editorSuggestWidget.background",wr,m(1473,"Background color of the suggest widget."));T("editorSuggestWidget.border",TW,m(1474,"Border color of the suggest widget."));const oPe=T("editorSuggestWidget.foreground",zc,m(1475,"Foreground color of the suggest widget."));T("editorSuggestWidget.selectedForeground",dS,m(1476,"Foreground color of the selected entry in the suggest widget."));T("editorSuggestWidget.selectedIconForeground",VW,m(1477,"Icon foreground color of the selected entry in the suggest widget."));const rPe=T("editorSuggestWidget.selectedBackground",hS,m(1478,"Background color of the selected entry in the suggest widget."));T("editorSuggestWidget.highlightForeground",Lb,m(1479,"Color of the match highlights in the suggest widget."));T("editorSuggestWidget.focusHighlightForeground",p_e,m(1480,"Color of the match highlights in the suggest widget when an item is focused."));T("editorSuggestWidgetStatus.foreground",Be(oPe,.5),m(1481,"Foreground color of the suggest widget status."));class aPe{constructor(e,t){this._service=e,this._key=`suggestWidget.size/${t.getEditorType()}/${t instanceof Lh}`}restore(){const e=this._service.get(this._key,0)??"";try{const t=JSON.parse(e);if(St.is(t))return St.lift(t)}catch{}}store(e){this._service.store(this._key,JSON.stringify(e),0,1)}reset(){this._service.remove(this._key,0)}}var eg;let R3=(eg=class{constructor(e,t,i,n,o){this.editor=e,this._storageService=t,this._state=0,this._isAuto=!1,this._pendingLayout=new vt,this._pendingShowDetails=new vt,this._ignoreFocusEvents=!1,this._forceRenderingAbove=!1,this._explainMode=!1,this._showTimeout=new $o,this._disposables=new U,this._onDidSelect=new Sm,this._onDidFocus=new Sm,this._onDidHide=new M,this._onDidShow=new M,this.onDidSelect=this._onDidSelect.event,this.onDidFocus=this._onDidFocus.event,this.onDidHide=this._onDidHide.event,this.onDidShow=this._onDidShow.event,this._onDetailsKeydown=new M,this.onDetailsKeyDown=this._onDetailsKeydown.event,this.element=new rH,this.element.domNode.classList.add("editor-widget","suggest-widget"),this._contentWidget=new lPe(this,e),this._persistedSize=new aPe(t,e);class r{constructor(g,f,p=!1,_=!1){this.persistedSize=g,this.currentSize=f,this.persistHeight=p,this.persistWidth=_}}let a;this._disposables.add(this.element.onDidWillResize(()=>{this._contentWidget.lockPreference(),a=new r(this._persistedSize.restore(),this.element.size)})),this._disposables.add(this.element.onDidResize(u=>{if(this._resize(u.dimension.width,u.dimension.height),a&&(a.persistHeight=a.persistHeight||!!u.north||!!u.south,a.persistWidth=a.persistWidth||!!u.east||!!u.west),!!u.done){if(a){const{itemHeight:g,defaultSize:f}=this.getLayoutInfo(),p=Math.round(g/2);let{width:_,height:b}=this.element.size;(!a.persistHeight||Math.abs(a.currentSize.height-b)<=p)&&(b=a.persistedSize?.height??f.height),(!a.persistWidth||Math.abs(a.currentSize.width-_)<=p)&&(_=a.persistedSize?.width??f.width),this._persistedSize.store(new St(_,b))}this._contentWidget.unlockPreference(),a=void 0}})),this._messageElement=X(this.element.domNode,ie(".message")),this._listElement=X(this.element.domNode,ie(".tree"));const l=this._disposables.add(o.createInstance(T3,this.editor));l.onDidClose(()=>this.toggleDetails(),this,this._disposables),this._details=new QOe(l,this.editor);const c=()=>this.element.domNode.classList.toggle("no-icons",!this.editor.getOption(134).showIcons);c();const d=o.createInstance(N3,this.editor);this._disposables.add(d),this._disposables.add(d.onDidToggleDetails(()=>this.toggleDetails())),this._list=new Tr("SuggestWidget",this._listElement,{getHeight:u=>this.getLayoutInfo().itemHeight,getTemplateId:u=>"suggestion"},[d],{alwaysConsumeMouseWheel:!0,useShadows:!1,mouseSupport:!1,multipleSelectionSupport:!1,accessibilityProvider:{getRole:()=>Bn?"listitem":"option",getWidgetAriaLabel:()=>m(1484,"Suggest"),getWidgetRole:()=>"listbox",getAriaLabel:u=>{let g=u.textLabel;const f=mv.toLabel(u.completion.kind);if(typeof u.completion.label!="string"){const{detail:w,description:C}=u.completion.label;w&&C?g=m(1485,"{0} {1}, {2}, {3}",g,w,C,f):w?g=m(1486,"{0} {1}, {2}",g,w,f):C&&(g=m(1487,"{0}, {1}, {2}",g,C,f))}else g=m(1488,"{0}, {1}",g,f);if(!u.isResolved||!this._isDetailsVisible())return g;const{documentation:p,detail:_}=u.completion,b=xm("{0}{1}",_||"",p?typeof p=="string"?p:p.value:"");return m(1489,"{0}, docs: {1}",g,b)}}}),this._list.style(f_({listInactiveFocusBackground:rPe,listInactiveFocusOutline:ri})),this._status=o.createInstance(D3,this.element.domNode,Hu);const h=()=>this.element.domNode.classList.toggle("with-status-bar",this.editor.getOption(134).showStatusBar);h(),this._disposables.add(this._list.onMouseDown(u=>this._onListMouseDownOrTap(u))),this._disposables.add(this._list.onTap(u=>this._onListMouseDownOrTap(u))),this._disposables.add(this._list.onDidChangeSelection(u=>this._onListSelection(u))),this._disposables.add(this._list.onDidChangeFocus(u=>this._onListFocus(u))),this._disposables.add(this.editor.onDidChangeCursorSelection(()=>this._onCursorSelectionChanged())),this._disposables.add(this.editor.onDidChangeConfiguration(u=>{u.hasChanged(134)&&(h(),c()),this._completionModel&&(u.hasChanged(59)||u.hasChanged(135)||u.hasChanged(136))&&this._list.splice(0,this._list.length,this._completionModel.items)})),this._ctxSuggestWidgetVisible=je.Visible.bindTo(i),this._ctxSuggestWidgetDetailsVisible=je.DetailsVisible.bindTo(i),this._ctxSuggestWidgetMultipleSuggestions=je.MultipleSuggestions.bindTo(i),this._ctxSuggestWidgetHasFocusedSuggestion=je.HasFocusedSuggestion.bindTo(i),this._disposables.add(Fi(this._details.widget.domNode,"keydown",u=>{this._onDetailsKeydown.fire(u)})),this._disposables.add(this.editor.onMouseDown(u=>this._onEditorMouseDown(u)))}dispose(){this._details.widget.dispose(),this._details.dispose(),this._list.dispose(),this._status.dispose(),this._disposables.dispose(),this._loadingTimeout?.dispose(),this._pendingLayout.dispose(),this._pendingShowDetails.dispose(),this._showTimeout.dispose(),this._contentWidget.dispose(),this.element.dispose()}_onEditorMouseDown(e){this._details.widget.domNode.contains(e.target.element)?this._details.widget.domNode.focus():this.element.domNode.contains(e.target.element)&&this.editor.focus()}_onCursorSelectionChanged(){this._state!==0&&this._contentWidget.layout()}_onListMouseDownOrTap(e){typeof e.element>"u"||typeof e.index>"u"||(e.browserEvent.preventDefault(),e.browserEvent.stopPropagation(),this._select(e.element,e.index))}_onListSelection(e){e.elements.length&&this._select(e.elements[0],e.indexes[0])}_select(e,t){const i=this._completionModel;i&&(this._onDidSelect.fire({item:e,index:t,model:i}),this.editor.focus())}_onListFocus(e){if(this._ignoreFocusEvents)return;if(this._state===5&&this._setState(3),!e.elements.length){this._currentSuggestionDetails&&(this._currentSuggestionDetails.cancel(),this._currentSuggestionDetails=void 0,this._focusedItem=void 0),this.editor.setAriaOptions({activeDescendant:void 0}),this._ctxSuggestWidgetHasFocusedSuggestion.set(!1);return}if(!this._completionModel)return;this._ctxSuggestWidgetHasFocusedSuggestion.set(!0);const t=e.elements[0],i=e.indexes[0];t!==this._focusedItem&&(this._currentSuggestionDetails?.cancel(),this._currentSuggestionDetails=void 0,this._focusedItem=t,this._list.reveal(i),this._currentSuggestionDetails=an(async n=>{const o=hh(()=>{this._isDetailsVisible()&&this._showDetails(!0,!1)},250),r=n.onCancellationRequested(()=>o.dispose());try{return await t.resolve(n)}finally{o.dispose(),r.dispose()}}),this._currentSuggestionDetails.then(()=>{i>=this._list.length||t!==this._list.element(i)||(this._ignoreFocusEvents=!0,this._list.splice(i,1,[t]),this._list.setFocus([i]),this._ignoreFocusEvents=!1,this._isDetailsVisible()?this._showDetails(!1,!1):this.element.domNode.classList.remove("docs-side"),this.editor.setAriaOptions({activeDescendant:this._list.getElementID(i)}))}).catch(Oe)),this._onDidFocus.fire({item:t,index:i,model:this._completionModel})}_setState(e){if(this._state!==e)switch(this._state=e,this.element.domNode.classList.toggle("frozen",e===4),this.element.domNode.classList.remove("message"),e){case 0:js(this._messageElement,this._listElement,this._status.element),this._details.hide(!0),this._status.hide(),this._contentWidget.hide(),this._ctxSuggestWidgetVisible.reset(),this._ctxSuggestWidgetMultipleSuggestions.reset(),this._ctxSuggestWidgetHasFocusedSuggestion.reset(),this._showTimeout.cancel(),this.element.domNode.classList.remove("visible"),this._list.splice(0,this._list.length),this._focusedItem=void 0,this._cappedHeight=void 0,this._explainMode=!1;break;case 1:this.element.domNode.classList.add("message"),this._messageElement.textContent=_b.LOADING_MESSAGE,js(this._listElement,this._status.element),Mo(this._messageElement),this._details.hide(),this._show(),this._focusedItem=void 0,xl(_b.LOADING_MESSAGE);break;case 2:this.element.domNode.classList.add("message"),this._messageElement.textContent=_b.NO_SUGGESTIONS_MESSAGE,js(this._listElement,this._status.element),Mo(this._messageElement),this._details.hide(),this._show(),this._focusedItem=void 0,xl(_b.NO_SUGGESTIONS_MESSAGE);break;case 3:js(this._messageElement),Mo(this._listElement,this._status.element),this._show();break;case 4:js(this._messageElement),Mo(this._listElement,this._status.element),this._show();break;case 5:js(this._messageElement),Mo(this._listElement,this._status.element),this._details.show(),this._show(),this._details.widget.focus();break}}_show(){this._status.show(),this._contentWidget.show(),this._layout(this._persistedSize.restore()),this._ctxSuggestWidgetVisible.set(!0),this._showTimeout.cancelAndSet(()=>{this.element.domNode.classList.add("visible"),this._onDidShow.fire(this)},100)}showTriggered(e,t){this._state===0&&(this._contentWidget.setPosition(this.editor.getPosition()),this._isAuto=!!e,this._isAuto||(this._loadingTimeout=hh(()=>this._setState(1),t)))}showSuggestions(e,t,i,n,o){if(this._contentWidget.setPosition(this.editor.getPosition()),this._loadingTimeout?.dispose(),this._currentSuggestionDetails?.cancel(),this._currentSuggestionDetails=void 0,this._completionModel!==e&&(this._completionModel=e),i&&this._state!==2&&this._state!==0){this._setState(4);return}const r=this._completionModel.items.length,a=r===0;if(this._ctxSuggestWidgetMultipleSuggestions.set(r>1),a){this._setState(n?0:2),this._completionModel=void 0;return}this._focusedItem=void 0,this._onDidFocus.pause(),this._onDidSelect.pause();try{this._list.splice(0,this._list.length,this._completionModel.items),this._setState(i?4:3),this._list.reveal(t,0,t===0?0:this.getLayoutInfo().itemHeight*.33),this._list.setFocus(o?[]:[t])}finally{this._onDidFocus.resume(),this._onDidSelect.resume()}this._pendingLayout.value=xE(be(this.element.domNode),()=>{this._pendingLayout.clear(),this._layout(this.element.size),this._details.widget.domNode.classList.remove("focused")})}focusSelected(){this._list.length>0&&this._list.setFocus([0])}selectNextPage(){switch(this._state){case 0:return!1;case 5:return this._details.widget.pageDown(),!0;case 1:return!this._isAuto;default:return this._list.focusNextPage(),!0}}selectNext(){switch(this._state){case 0:return!1;case 1:return!this._isAuto;default:return this._list.focusNext(1,!0),!0}}selectLast(){switch(this._state){case 0:return!1;case 5:return this._details.widget.scrollBottom(),!0;case 1:return!this._isAuto;default:return this._list.focusLast(),!0}}selectPreviousPage(){switch(this._state){case 0:return!1;case 5:return this._details.widget.pageUp(),!0;case 1:return!this._isAuto;default:return this._list.focusPreviousPage(),!0}}selectPrevious(){switch(this._state){case 0:return!1;case 1:return!this._isAuto;default:return this._list.focusPrevious(1,!0),!1}}selectFirst(){switch(this._state){case 0:return!1;case 5:return this._details.widget.scrollTop(),!0;case 1:return!this._isAuto;default:return this._list.focusFirst(),!0}}getFocusedItem(){if(this._state!==0&&this._state!==2&&this._state!==1&&this._completionModel&&this._list.getFocus().length>0)return{item:this._list.getFocusedElements()[0],index:this._list.getFocus()[0],model:this._completionModel}}toggleDetailsFocus(){this._state===5?(this._list.setFocus(this._list.getFocus()),this._setState(3)):this._state===3&&(this._setState(5),this._isDetailsVisible()?this._details.widget.focus():this.toggleDetails(!0))}toggleDetails(e=!1){this._isDetailsVisible()?(this._pendingShowDetails.clear(),this._ctxSuggestWidgetDetailsVisible.set(!1),this._setDetailsVisible(!1),this._details.hide(),this.element.domNode.classList.remove("shows-details")):(YH(this._list.getFocusedElements()[0])||this._explainMode)&&(this._state===3||this._state===5||this._state===4)&&(this._ctxSuggestWidgetDetailsVisible.set(!0),this._setDetailsVisible(!0),this._showDetails(!1,e))}_showDetails(e,t){this._pendingShowDetails.value=xE(be(this.element.domNode),()=>{this._pendingShowDetails.clear(),this._details.show();let i=!1;e?this._details.widget.renderLoading():this._details.widget.renderItem(this._list.getFocusedElements()[0],this._explainMode),this._details.widget.isEmpty?this._details.hide():(this._positionDetails(),this.element.domNode.classList.add("shows-details"),t&&(this._details.widget.focus(),i=!0)),i||this.editor.focus()})}toggleExplainMode(){this._list.getFocusedElements()[0]&&(this._explainMode=!this._explainMode,this._isDetailsVisible()?this._showDetails(!1,!1):this.toggleDetails())}resetPersistedSize(){this._persistedSize.reset()}hideWidget(){this._pendingLayout.clear(),this._pendingShowDetails.clear(),this._loadingTimeout?.dispose(),this._setState(0),this._onDidHide.fire(this),this.element.clearSashHoverState();const e=this._persistedSize.restore(),t=Math.ceil(this.getLayoutInfo().itemHeight*4.3);e&&e.height<t&&this._persistedSize.store(e.with(void 0,t))}isFrozen(){return this._state===4}_afterRender(e){if(e===null){this._isDetailsVisible()&&this._details.hide();return}this._state===2||this._state===1||(this._isDetailsVisible()&&!this._details.widget.isEmpty&&this._details.show(),this._positionDetails())}_layout(e){if(!this.editor.hasModel()||!this.editor.getDomNode())return;const t=lg(this.element.domNode.ownerDocument.body),i=this.getLayoutInfo();e||(e=i.defaultSize);let n=e.height,o=e.width;if(this._status.element.style.height=`${i.itemHeight}px`,this._state===2||this._state===1)n=i.itemHeight+i.borderHeight,o=i.defaultSize.width/2,this.element.enableSashes(!1,!1,!1,!1),this.element.minSize=this.element.maxSize=new St(o,n),this._contentWidget.setPreference(2);else{const r=t.width-i.borderHeight-2*i.horizontalPadding;o>r&&(o=r);const a=this._completionModel?this._completionModel.stats.pLabelLen*i.typicalHalfwidthCharacterWidth:o,l=i.statusBarHeight+this._list.contentHeight+i.borderHeight,c=i.itemHeight+i.statusBarHeight,d=Li(this.editor.getDomNode()),h=this.editor.getScrolledVisiblePosition(this.editor.getPosition()),u=d.top+h.top+h.height,g=Math.min(t.height-u-i.verticalPadding,l),f=d.top+h.top-i.verticalPadding,p=Math.min(f,l);let _=Math.min(Math.max(p,g)+i.borderHeight,l);n===this._cappedHeight?.capped&&(n=this._cappedHeight.wanted),n<c&&(n=c),n>_&&(n=_),n>g&&p>g||this._forceRenderingAbove&&f>150?(this._contentWidget.setPreference(1),this.element.enableSashes(!0,!0,!1,!1),_=p):(this._contentWidget.setPreference(2),this.element.enableSashes(!1,!0,!0,!1),_=g),this.element.preferredSize=new St(a,i.defaultSize.height),this.element.maxSize=new St(r,_),this.element.minSize=new St(220,c),this._cappedHeight=n===l?{wanted:this._cappedHeight?.wanted??e.height,capped:n}:void 0}this._resize(o,n)}_resize(e,t){const{width:i,height:n}=this.element.maxSize;e=Math.min(i,e),t=Math.min(n,t);const{statusBarHeight:o}=this.getLayoutInfo();this._list.layout(t-o,e),this._listElement.style.height=`${t-o}px`,this.element.layout(t,e),this._contentWidget.layout(),this._positionDetails()}_positionDetails(){this._isDetailsVisible()&&this._details.placeAtAnchor(this.element.domNode,this._contentWidget.getPosition()?.preference[0]===2)}getLayoutInfo(){const e=this.editor.getOption(59),t=qs(this.editor.getOption(136)||e.lineHeight,8,1e3),i=!this.editor.getOption(134).showStatusBar||this._state===2||this._state===1?0:t,n=this._details.widget.getLayoutInfo().borderWidth,o=2*n;return{itemHeight:t,statusBarHeight:i,borderWidth:n,borderHeight:o,typicalHalfwidthCharacterWidth:e.typicalHalfwidthCharacterWidth,verticalPadding:22,horizontalPadding:14,defaultSize:new St(430,i+12*t)}}_isDetailsVisible(){return this._storageService.getBoolean("expandSuggestionDocs",0,!1)}_setDetailsVisible(e){this._storageService.store("expandSuggestionDocs",e,0,0)}forceRenderingAbove(){this._forceRenderingAbove||(this._forceRenderingAbove=!0,this._layout(this._persistedSize.restore()))}stopForceRenderingAbove(){this._forceRenderingAbove=!1}},_b=eg,eg.LOADING_MESSAGE=m(1482,"Loading..."),eg.NO_SUGGESTIONS_MESSAGE=m(1483,"No suggestions."),eg);R3=_b=sPe([jL(1,Bs),jL(2,Ae),jL(3,wi),jL(4,_e)],R3);class lPe{constructor(e,t){this._widget=e,this._editor=t,this.allowEditorOverflow=!0,this.suppressMouseDown=!1,this._preferenceLocked=!1,this._added=!1,this._hidden=!1}dispose(){this._added&&(this._added=!1,this._editor.removeContentWidget(this))}getId(){return"editor.widget.suggestWidget"}getDomNode(){return this._widget.element.domNode}show(){this._hidden=!1,this._added||(this._added=!0,this._editor.addContentWidget(this))}hide(){this._hidden||(this._hidden=!0,this.layout())}layout(){this._editor.layoutContentWidget(this)}getPosition(){return this._hidden||!this._position||!this._preference?null:{position:this._position,preference:[this._preference]}}beforeRender(){const{height:e,width:t}=this._widget.element.size,{borderWidth:i,horizontalPadding:n}=this._widget.getLayoutInfo();return new St(t+2*i+n,e+2*i)}afterRender(e){this._widget._afterRender(e)}setPreference(e){this._preferenceLocked||(this._preference=e)}lockPreference(){this._preferenceLocked=!0}unlockPreference(){this._preferenceLocked=!1}setPosition(e){this._position=e}}var cPe=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},tb=function(s,e){return function(t,i){e(t,i,s)}},A3;class dPe{constructor(e,t){if(this._model=e,this._position=t,this._decorationOptions=Fe.register({description:"suggest-line-suffix",stickiness:1}),e.getLineMaxColumn(t.lineNumber)!==t.column){const n=e.getOffsetAt(t),o=e.getPositionAt(n+1);e.changeDecorations(r=>{this._marker&&r.removeDecoration(this._marker),this._marker=r.addDecoration(k.fromPositions(t,o),this._decorationOptions)})}}dispose(){this._marker&&!this._model.isDisposed()&&this._model.changeDecorations(e=>{e.removeDecoration(this._marker),this._marker=void 0})}delta(e){if(this._model.isDisposed()||this._position.lineNumber!==e.lineNumber)return 0;if(this._marker){const t=this._model.getDecorationRange(this._marker);return this._model.getOffsetAt(t.getStartPosition())-this._model.getOffsetAt(e)}else return this._model.getLineMaxColumn(e.lineNumber)-e.column}}var Qp;let _l=(Qp=class{static get(e){return e.getContribution(A3.ID)}get onWillInsertSuggestItem(){return this._onWillInsertSuggestItem.event}constructor(e,t,i,n,o,r,a){this._memoryService=t,this._commandService=i,this._contextKeyService=n,this._instantiationService=o,this._logService=r,this._telemetryService=a,this._lineSuffix=new vt,this._toDispose=new U,this._selectors=new hPe(h=>h.priority),this._onWillInsertSuggestItem=new M,this._wantsForceRenderingAbove=!1,this.editor=e,this.model=o.createInstance(HD,this.editor),this._selectors.register({priority:0,select:(h,u,g)=>this._memoryService.select(h,u,g)});const l=je.InsertMode.bindTo(n);l.set(e.getOption(134).insertMode),this._toDispose.add(this.model.onDidTrigger(()=>l.set(e.getOption(134).insertMode))),this.widget=this._toDispose.add(new B2(be(e.getDomNode()),()=>{const h=this._instantiationService.createInstance(R3,this.editor);this._toDispose.add(h),this._toDispose.add(h.onDidSelect(_=>this._insertSuggestion(_,0),this));const u=new jOe(this.editor,h,this.model,_=>this._insertSuggestion(_,2));this._toDispose.add(u);const g=je.MakesTextEdit.bindTo(this._contextKeyService),f=je.HasInsertAndReplaceRange.bindTo(this._contextKeyService),p=je.CanResolve.bindTo(this._contextKeyService);return this._toDispose.add(pe(()=>{g.reset(),f.reset(),p.reset()})),this._toDispose.add(h.onDidFocus(({item:_})=>{const b=this.editor.getPosition(),w=_.editStart.column,C=b.column;let v=!0;this.editor.getOption(1)==="smart"&&this.model.state===2&&!_.completion.additionalTextEdits&&!(_.completion.insertTextRules&4)&&C-w===_.completion.insertText.length&&(v=this.editor.getModel().getValueInRange({startLineNumber:b.lineNumber,startColumn:w,endLineNumber:b.lineNumber,endColumn:C})!==_.completion.insertText),g.set(v),f.set(!A.equals(_.editInsertEnd,_.editReplaceEnd)),p.set(!!_.provider.resolveCompletionItem||!!_.completion.documentation||_.completion.detail!==_.completion.label)})),this._toDispose.add(h.onDetailsKeyDown(_=>{if(_.toKeyCodeChord().equals(new dh(!0,!1,!1,!1,33))||Ye&&_.toKeyCodeChord().equals(new dh(!1,!1,!1,!0,33))){_.stopPropagation();return}_.toKeyCodeChord().isModifierKey()||this.editor.focus()})),this._wantsForceRenderingAbove&&h.forceRenderingAbove(),h})),this._overtypingCapturer=this._toDispose.add(new B2(be(e.getDomNode()),()=>this._toDispose.add(new I3(this.editor,this.model)))),this._alternatives=this._toDispose.add(new B2(be(e.getDomNode()),()=>this._toDispose.add(new Nv(this.editor,this._contextKeyService)))),this._toDispose.add(o.createInstance(FD,e)),this._toDispose.add(this.model.onDidTrigger(h=>{this.widget.value.showTriggered(h.auto,h.shy?250:50),this._lineSuffix.value=new dPe(this.editor.getModel(),h.position)})),this._toDispose.add(this.model.onDidSuggest(h=>{if(h.triggerOptions.shy)return;let u=-1;for(const f of this._selectors.itemsOrderedByPriorityDesc)if(u=f.select(this.editor.getModel(),this.editor.getPosition(),h.completionModel.items),u!==-1)break;if(u===-1&&(u=0),this.model.state===0)return;let g=!1;if(h.triggerOptions.auto){const f=this.editor.getOption(134);f.selectionMode==="never"||f.selectionMode==="always"?g=f.selectionMode==="never":f.selectionMode==="whenTriggerCharacter"?g=h.triggerOptions.triggerKind!==1:f.selectionMode==="whenQuickSuggestion"&&(g=h.triggerOptions.triggerKind===1&&!h.triggerOptions.refilter)}this.widget.value.showSuggestions(h.completionModel,u,h.isFrozen,h.triggerOptions.auto,g)})),this._toDispose.add(this.model.onDidCancel(h=>{h.retrigger||this.widget.value.hideWidget()})),this._toDispose.add(this.editor.onDidBlurEditorWidget(()=>{this.model.cancel(),this.model.clear()}));const c=je.AcceptSuggestionsOnEnter.bindTo(n),d=()=>{const h=this.editor.getOption(1);c.set(h==="on"||h==="smart")};this._toDispose.add(this.editor.onDidChangeConfiguration(()=>d())),d()}dispose(){this._alternatives.dispose(),this._toDispose.dispose(),this.widget.dispose(),this.model.dispose(),this._lineSuffix.dispose(),this._onWillInsertSuggestItem.dispose()}_insertSuggestion(e,t){if(!e||!e.item){this._alternatives.value.reset(),this.model.cancel(),this.model.clear();return}if(!this.editor.hasModel())return;const i=Ps.get(this.editor);if(!i)return;this._onWillInsertSuggestItem.fire({item:e.item});const n=this.editor.getModel(),o=n.getAlternativeVersionId(),{item:r}=e,a=[],l=new si;t&1||this.editor.pushUndoStop();const c=this.getOverwriteInfo(r,!!(t&8));this._memoryService.memorize(n,this.editor.getPosition(),r);const d=r.isResolved;let h=-1,u=-1;if(Array.isArray(r.completion.additionalTextEdits)){this.model.cancel();const f=Il.capture(this.editor);this.editor.executeEdits("suggestController.additionalTextEdits.sync",r.completion.additionalTextEdits.map(p=>{let _=k.lift(p.range);if(_.startLineNumber===r.position.lineNumber&&_.startColumn>r.position.column){const b=this.editor.getPosition().column-r.position.column,w=b,C=k.spansMultipleLines(_)?0:b;_=new k(_.startLineNumber,_.startColumn+w,_.endLineNumber,_.endColumn+C)}return Si.replaceMove(_,p.text)})),f.restoreRelativeVerticalPositionOfCursor(this.editor)}else if(!d){const f=new Cn;let p;const _=n.onDidChangeContent(v=>{if(v.isFlush){l.cancel(),_.dispose();return}for(const y of v.changes){const x=k.getEndPosition(y.range);(!p||A.isBefore(x,p))&&(p=x)}}),b=t;t|=2;let w=!1;const C=this.editor.onWillType(()=>{C.dispose(),w=!0,b&2||this.editor.pushUndoStop()});a.push(r.resolve(l.token).then(()=>{if(!r.completion.additionalTextEdits||l.token.isCancellationRequested)return;if(p&&r.completion.additionalTextEdits.some(y=>A.isBefore(p,k.getStartPosition(y.range))))return!1;w&&this.editor.pushUndoStop();const v=Il.capture(this.editor);return this.editor.executeEdits("suggestController.additionalTextEdits.async",r.completion.additionalTextEdits.map(y=>Si.replaceMove(k.lift(y.range),y.text))),v.restoreRelativeVerticalPositionOfCursor(this.editor),(w||!(b&2))&&this.editor.pushUndoStop(),!0}).then(v=>{this._logService.trace("[suggest] async resolving of edits DONE (ms, applied?)",f.elapsed(),v),u=v===!0?1:v===!1?0:-2}).finally(()=>{_.dispose(),C.dispose()}))}let{insertText:g}=r.completion;if(r.completion.insertTextRules&4||(g=Hm.escape(g)),this.model.cancel(),i.insert(g,{overwriteBefore:c.overwriteBefore,overwriteAfter:c.overwriteAfter,undoStopBefore:!1,undoStopAfter:!1,adjustWhitespace:!(r.completion.insertTextRules&1),clipboardText:e.model.clipboardText,overtypingCapturer:this._overtypingCapturer.value,reason:cs.suggest({providerId:fA.fromExtensionId(r.extensionId?.value)})}),t&2||this.editor.pushUndoStop(),r.completion.command)if(r.completion.command.id===VD.id)this.model.trigger({auto:!0,retrigger:!0});else{const f=new Cn;a.push(this._commandService.executeCommand(r.completion.command.id,...r.completion.command.arguments?[...r.completion.command.arguments]:[]).catch(p=>{r.completion.extensionId?ji(p):Oe(p)}).finally(()=>{h=f.elapsed()}))}t&4&&this._alternatives.value.set(e,f=>{for(l.cancel();n.canUndo();){o!==n.getAlternativeVersionId()&&n.undo(),this._insertSuggestion(f,3|(t&8?8:0));break}}),this._alertCompletionItem(r),Promise.all(a).finally(()=>{this._reportSuggestionAcceptedTelemetry(r,n,d,h,u,e.index,e.model.items),this.model.clear(),l.dispose()})}_reportSuggestionAcceptedTelemetry(e,t,i,n,o,r,a){if(Math.random()>1e-4)return;const l=new Map;for(let u=0;u<Math.min(30,a.length);u++){const g=a[u].textLabel;l.has(g)?l.get(g).push(u):l.set(g,[u])}const c=l.get(e.textLabel),h=c&&c.length>1?c[0]:-1;this._telemetryService.publicLog2("suggest.acceptedSuggestion",{extensionId:e.extensionId?.value??"unknown",providerId:e.provider._debugDisplayName??"unknown",kind:e.completion.kind,basenameHash:j0(ra(t.uri)).toString(16),languageId:t.getLanguageId(),fileExtension:Que(t.uri),resolveInfo:e.provider.resolveCompletionItem?i?1:0:-1,resolveDuration:e.resolveDuration,commandDuration:n,additionalEditsAsync:o,index:r,firstIndex:h})}getOverwriteInfo(e,t){dt(this.editor.hasModel());let i=this.editor.getOption(134).insertMode==="replace";t&&(i=!i);const n=e.position.column-e.editStart.column,o=(i?e.editReplaceEnd.column:e.editInsertEnd.column)-e.position.column,r=this.editor.getPosition().column-e.position.column,a=this._lineSuffix.value?this._lineSuffix.value.delta(this.editor.getPosition()):0;return{overwriteBefore:n+r,overwriteAfter:o+a}}_alertCompletionItem(e){if(Ms(e.completion.additionalTextEdits)){const t=m(1463,"Accepting '{0}' made {1} additional edits",e.textLabel,e.completion.additionalTextEdits.length);Js(t)}}triggerSuggest(e,t,i){this.editor.hasModel()&&(this.model.trigger({auto:t??!1,completionOptions:{providerFilter:e,kindFilter:i?new Set:void 0}}),this.editor.revealPosition(this.editor.getPosition(),0),this.editor.focus())}triggerSuggestAndAcceptBest(e){if(!this.editor.hasModel())return;const t=this.editor.getPosition(),i=()=>{t.equals(this.editor.getPosition())&&this._commandService.executeCommand(e.fallback)},n=o=>{if(o.completion.insertTextRules&4||o.completion.additionalTextEdits)return!0;const r=this.editor.getPosition(),a=o.editStart.column,l=r.column;return l-a!==o.completion.insertText.length?!0:this.editor.getModel().getValueInRange({startLineNumber:r.lineNumber,startColumn:a,endLineNumber:r.lineNumber,endColumn:l})!==o.completion.insertText};oe.once(this.model.onDidTrigger)(o=>{const r=[];oe.any(this.model.onDidTrigger,this.model.onDidCancel)(()=>{xt(r),i()},void 0,r),this.model.onDidSuggest(({completionModel:a})=>{if(xt(r),a.items.length===0){i();return}const l=this._memoryService.select(this.editor.getModel(),this.editor.getPosition(),a.items),c=a.items[l];if(!n(c)){i();return}this.editor.pushUndoStop(),this._insertSuggestion({index:l,item:c,model:a},7)},void 0,r)}),this.model.trigger({auto:!1,shy:!0}),this.editor.revealPosition(t,0),this.editor.focus()}acceptSelectedSuggestion(e,t){const i=this.widget.value.getFocusedItem();let n=0;e&&(n|=4),t&&(n|=8),this._insertSuggestion(i,n)}acceptNextSuggestion(){this._alternatives.value.next()}acceptPrevSuggestion(){this._alternatives.value.prev()}cancelSuggestWidget(){this.model.cancel(),this.model.clear(),this.widget.value.hideWidget()}focusSuggestion(){this.widget.value.focusSelected()}selectNextSuggestion(){this.widget.value.selectNext()}selectNextPageSuggestion(){this.widget.value.selectNextPage()}selectLastSuggestion(){this.widget.value.selectLast()}selectPrevSuggestion(){this.widget.value.selectPrevious()}selectPrevPageSuggestion(){this.widget.value.selectPreviousPage()}selectFirstSuggestion(){this.widget.value.selectFirst()}toggleSuggestionDetails(){this.widget.value.toggleDetails()}toggleExplainMode(){this.widget.value.toggleExplainMode()}toggleSuggestionFocus(){this.widget.value.toggleDetailsFocus()}resetWidgetSize(){this.widget.value.resetPersistedSize()}forceRenderingAbove(){this.widget.isInitialized?this.widget.value.forceRenderingAbove():this._wantsForceRenderingAbove=!0}stopForceRenderingAbove(){this.widget.isInitialized?this.widget.value.stopForceRenderingAbove():this._wantsForceRenderingAbove=!1}registerSelector(e){return this._selectors.register(e)}},A3=Qp,Qp.ID="editor.contrib.suggestController",Qp);_l=A3=cPe([tb(1,YA),tb(2,Gt),tb(3,Ae),tb(4,_e),tb(5,jt),tb(6,vs)],_l);class hPe{constructor(e){this.prioritySelector=e,this._items=new Array}register(e){if(this._items.indexOf(e)!==-1)throw new Error("Value is already registered");return this._items.push(e),this._items.sort((t,i)=>this.prioritySelector(i)-this.prioritySelector(t)),{dispose:()=>{const t=this._items.indexOf(e);t>=0&&this._items.splice(t,1)}}}get itemsOrderedByPriorityDesc(){return this._items}}const FN=class FN extends ue{constructor(){super({id:FN.id,label:V(1471,"Trigger Suggest"),precondition:K.and(N.writable,N.hasCompletionItemProvider,je.Visible.toNegated()),kbOpts:{kbExpr:N.textInputFocus,primary:2058,secondary:[2087],mac:{primary:266,secondary:[521,2087]},weight:100}})}run(e,t,i){const n=_l.get(t);if(!n)return;let o;i&&typeof i=="object"&&i.auto===!0&&(o=!0),n.triggerSuggest(void 0,o,void 0)}};FN.id="editor.action.triggerSuggest";let VD=FN;lt(_l.ID,_l,2);ae(VD);const Ir=190,oo=hn.bindToContribution(_l.get);le(new oo({id:"acceptSelectedSuggestion",precondition:K.and(je.Visible,je.HasFocusedSuggestion),handler(s){s.acceptSelectedSuggestion(!0,!1)},kbOpts:[{primary:2,kbExpr:K.and(je.Visible,N.textInputFocus),weight:Ir},{primary:3,kbExpr:K.and(je.Visible,N.textInputFocus,je.AcceptSuggestionsOnEnter,je.MakesTextEdit),weight:Ir}],menuOpts:[{menuId:Hu,title:m(1464,"Insert"),group:"left",order:1,when:je.HasInsertAndReplaceRange.toNegated()},{menuId:Hu,title:m(1465,"Insert"),group:"left",order:1,when:K.and(je.HasInsertAndReplaceRange,je.InsertMode.isEqualTo("insert"))},{menuId:Hu,title:m(1466,"Replace"),group:"left",order:1,when:K.and(je.HasInsertAndReplaceRange,je.InsertMode.isEqualTo("replace"))}]}));le(new oo({id:"acceptAlternativeSelectedSuggestion",precondition:K.and(je.Visible,N.textInputFocus,je.HasFocusedSuggestion),kbOpts:{weight:Ir,kbExpr:N.textInputFocus,primary:1027,secondary:[1026]},handler(s){s.acceptSelectedSuggestion(!1,!0)},menuOpts:[{menuId:Hu,group:"left",order:2,when:K.and(je.HasInsertAndReplaceRange,je.InsertMode.isEqualTo("insert")),title:m(1467,"Replace")},{menuId:Hu,group:"left",order:2,when:K.and(je.HasInsertAndReplaceRange,je.InsertMode.isEqualTo("replace")),title:m(1468,"Insert")}]}));at.registerCommandAlias("acceptSelectedSuggestionOnEnter","acceptSelectedSuggestion");le(new oo({id:"hideSuggestWidget",precondition:je.Visible,handler:s=>s.cancelSuggestWidget(),kbOpts:{weight:Ir,kbExpr:N.textInputFocus,primary:9,secondary:[1033]}}));le(new oo({id:"selectNextSuggestion",precondition:K.and(je.Visible,K.or(je.MultipleSuggestions,je.HasFocusedSuggestion.negate())),handler:s=>s.selectNextSuggestion(),kbOpts:{weight:Ir,kbExpr:N.textInputFocus,primary:18,secondary:[2066],mac:{primary:18,secondary:[2066,300]}}}));le(new oo({id:"selectNextPageSuggestion",precondition:K.and(je.Visible,K.or(je.MultipleSuggestions,je.HasFocusedSuggestion.negate())),handler:s=>s.selectNextPageSuggestion(),kbOpts:{weight:Ir,kbExpr:N.textInputFocus,primary:12,secondary:[2060]}}));le(new oo({id:"selectLastSuggestion",precondition:K.and(je.Visible,K.or(je.MultipleSuggestions,je.HasFocusedSuggestion.negate())),handler:s=>s.selectLastSuggestion()}));le(new oo({id:"selectPrevSuggestion",precondition:K.and(je.Visible,K.or(je.MultipleSuggestions,je.HasFocusedSuggestion.negate())),handler:s=>s.selectPrevSuggestion(),kbOpts:{weight:Ir,kbExpr:N.textInputFocus,primary:16,secondary:[2064],mac:{primary:16,secondary:[2064,302]}}}));le(new oo({id:"selectPrevPageSuggestion",precondition:K.and(je.Visible,K.or(je.MultipleSuggestions,je.HasFocusedSuggestion.negate())),handler:s=>s.selectPrevPageSuggestion(),kbOpts:{weight:Ir,kbExpr:N.textInputFocus,primary:11,secondary:[2059]}}));le(new oo({id:"selectFirstSuggestion",precondition:K.and(je.Visible,K.or(je.MultipleSuggestions,je.HasFocusedSuggestion.negate())),handler:s=>s.selectFirstSuggestion()}));le(new oo({id:"focusSuggestion",precondition:K.and(je.Visible,je.HasFocusedSuggestion.negate()),handler:s=>s.focusSuggestion(),kbOpts:{weight:Ir,kbExpr:N.textInputFocus,primary:2058,secondary:[2087],mac:{primary:266,secondary:[2087]}}}));le(new oo({id:"focusAndAcceptSuggestion",precondition:K.and(je.Visible,je.HasFocusedSuggestion.negate()),handler:s=>{s.focusSuggestion(),s.acceptSelectedSuggestion(!0,!1)}}));le(new oo({id:"toggleSuggestionDetails",precondition:K.and(je.Visible,je.HasFocusedSuggestion),handler:s=>s.toggleSuggestionDetails(),kbOpts:{weight:Ir,kbExpr:N.textInputFocus,primary:2058,secondary:[2087],mac:{primary:266,secondary:[2087]}},menuOpts:[{menuId:Hu,group:"right",order:1,when:K.and(je.DetailsVisible,je.CanResolve),title:m(1469,"Show Less")},{menuId:Hu,group:"right",order:1,when:K.and(je.DetailsVisible.toNegated(),je.CanResolve),title:m(1470,"Show More")}]}));le(new oo({id:"toggleExplainMode",precondition:je.Visible,handler:s=>s.toggleExplainMode(),kbOpts:{weight:100,primary:2138}}));le(new oo({id:"toggleSuggestionFocus",precondition:je.Visible,handler:s=>s.toggleSuggestionFocus(),kbOpts:{weight:Ir,kbExpr:N.textInputFocus,primary:2570,mac:{primary:778}}}));le(new oo({id:"insertBestCompletion",precondition:K.and(N.textInputFocus,K.equals("config.editor.tabCompletion","on"),FD.AtEnd,je.Visible.toNegated(),Nv.OtherSuggestions.toNegated(),Ps.InSnippetMode.toNegated()),handler:(s,e)=>{s.triggerSuggestAndAcceptBest(rn(e)?{fallback:"tab",...e}:{fallback:"tab"})},kbOpts:{weight:Ir,primary:2}}));le(new oo({id:"insertNextSuggestion",precondition:K.and(N.textInputFocus,K.equals("config.editor.tabCompletion","on"),Nv.OtherSuggestions,je.Visible.toNegated(),Ps.InSnippetMode.toNegated()),handler:s=>s.acceptNextSuggestion(),kbOpts:{weight:Ir,kbExpr:N.textInputFocus,primary:2}}));le(new oo({id:"insertPrevSuggestion",precondition:K.and(N.textInputFocus,K.equals("config.editor.tabCompletion","on"),Nv.OtherSuggestions,je.Visible.toNegated(),Ps.InSnippetMode.toNegated()),handler:s=>s.acceptPrevSuggestion(),kbOpts:{weight:Ir,kbExpr:N.textInputFocus,primary:1026}}));ae(class extends ue{constructor(){super({id:"editor.action.resetSuggestSize",label:V(1472,"Reset Suggest Widget Size"),precondition:void 0})}run(s,e){_l.get(e)?.resetWidgetSize()}});class uPe extends F{get selectedItem(){return this._currentSuggestItemInfo}constructor(e,t,i){super(),this.editor=e,this.suggestControllerPreselector=t,this.onWillAccept=i,this.isSuggestWidgetVisible=!1,this.isShiftKeyPressed=!1,this._isActive=!1,this._currentSuggestItemInfo=void 0,this._onDidSelectedItemChange=this._register(new M),this.onDidSelectedItemChange=this._onDidSelectedItemChange.event,this._register(e.onKeyDown(o=>{o.shiftKey&&!this.isShiftKeyPressed&&(this.isShiftKeyPressed=!0,this.update(this._isActive))})),this._register(e.onKeyUp(o=>{o.shiftKey&&this.isShiftKeyPressed&&(this.isShiftKeyPressed=!1,this.update(this._isActive))}));const n=_l.get(this.editor);if(n){this._register(n.registerSelector({priority:100,select:(a,l,c)=>{const d=this.editor.getModel();if(!d)return-1;const h=this.suggestControllerPreselector(),u=h?jd(h,d):void 0;if(!u)return-1;const g=A.lift(l),f=c.map((_,b)=>{const w=oy.fromSuggestion(n,d,g,_,this.isShiftKeyPressed),C=jd(w.getSingleTextEdit(),d),v=coe(u,C);return{index:b,valid:v,prefixLength:C.text.length,suggestItem:_}}).filter(_=>_&&_.valid&&_.prefixLength>0),p=jW(f,ns(_=>_.prefixLength,Wo));return p?p.index:-1}}));let o=!1;const r=()=>{o||(o=!0,this._register(n.widget.value.onDidShow(()=>{this.isSuggestWidgetVisible=!0,this.update(!0)})),this._register(n.widget.value.onDidHide(()=>{this.isSuggestWidgetVisible=!1,this.update(!1)})),this._register(n.widget.value.onDidFocus(()=>{this.isSuggestWidgetVisible=!0,this.update(!0)})))};this._register(oe.once(n.model.onDidTrigger)(a=>{r()})),this._register(n.onWillInsertSuggestItem(a=>{const l=this.editor.getPosition(),c=this.editor.getModel();if(!l||!c)return;const d=oy.fromSuggestion(n,c,l,a.item,this.isShiftKeyPressed);this.onWillAccept(d)}))}this.update(this._isActive)}update(e){const t=this.getSuggestItemInfo();(this._isActive!==e||!gPe(this._currentSuggestItemInfo,t))&&(this._isActive=e,this._currentSuggestItemInfo=t,this._onDidSelectedItemChange.fire())}getSuggestItemInfo(){const e=_l.get(this.editor);if(!e||!this.isSuggestWidgetVisible)return;const t=e.widget.value.getFocusedItem(),i=this.editor.getPosition(),n=this.editor.getModel();if(!(!t||!i||!n))return oy.fromSuggestion(e,n,i,t.item,this.isShiftKeyPressed)}stopForceRenderingAbove(){_l.get(this.editor)?.stopForceRenderingAbove()}forceRenderingAbove(){_l.get(this.editor)?.forceRenderingAbove()}}class oy{static fromSuggestion(e,t,i,n,o){let{insertText:r}=n.completion,a=!1;if(n.completion.insertTextRules&4){const c=new Hm().parse(r);c.children.length<100&&ND.adjustWhitespace(t,i,!0,c),r=c.toString(),a=!0}const l=e.getOverwriteInfo(n,o);return new oy(k.fromPositions(i.delta(0,-l.overwriteBefore),i.delta(0,Math.max(l.overwriteAfter,0))),r,n.completion.kind,a,n.container.incomplete??!1)}constructor(e,t,i,n,o){this.range=e,this.insertText=t,this.completionItemKind=i,this.isSnippetText=n,this.listIncomplete=o}equals(e){return this.range.equalsRange(e.range)&&this.insertText===e.insertText&&this.completionItemKind===e.completionItemKind&&this.isSnippetText===e.isSnippetText}toSelectedSuggestionInfo(){return new Eee(this.range,this.insertText,this.completionItemKind,this.isSnippetText)}getSingleTextEdit(){return new Ui(this.range,this.insertText)}}function gPe(s,e){return s===e?!0:!s||!e?!1:s.equals(e)}class fPe extends F{constructor(e,t,i){super(),this._editorObs=e,this._handleSuggestAccepted=t,this._suggestControllerPreselector=i,this._suggestWidgetAdaptor=this._register(new uPe(this._editorObs.editor,()=>(this._editorObs.forceUpdate(),this._suggestControllerPreselector()),n=>this._editorObs.forceUpdate(o=>{this._handleSuggestAccepted(n)}))),this.selectedItem=bt(this,n=>this._suggestWidgetAdaptor.onDidSelectedItemChange(()=>{this._editorObs.forceUpdate(o=>n(void 0))}),()=>this._suggestWidgetAdaptor.selectedItem)}stopForceRenderingAbove(){this._suggestWidgetAdaptor.stopForceRenderingAbove()}forceRenderingAbove(){this._suggestWidgetAdaptor.forceRenderingAbove()}}class pPe{constructor(e,t){this.lineNumber=e,this.columnRange=t}}class nK{constructor(e,t){if(this.startColumn=e,this.endColumnExclusive=t,e>t)throw new Ce(`startColumn ${e} cannot be after endColumnExclusive ${t}`)}toRange(e){return new k(e,this.startColumn,e,this.endColumnExclusive)}equals(e){return this.startColumn===e.startColumn&&this.endColumnExclusive===e.endColumnExclusive}toZeroBasedOffsetRange(){return new me(this.startColumn-1,this.endColumnExclusive-1)}}class ep{static fromLineTokens(e){const t=[];for(let i=0;i<e.getCount();i++)t.push(new sK(e.getTokenText(i),e.getMetadata(i)));return ep.create(t)}static create(e){return new ep(e)}constructor(e){this._tokenInfo=e}toLineTokens(e){return Ri.createFromTextAndMetadata(this.map((t,i)=>({text:i.text,metadata:i.metadata})),e)}map(e){const t=[];let i=0;for(const n of this._tokenInfo){const o=new me(i,i+n.text.length);t.push(e(o,n)),i+=n.text.length}return t}slice(e){const t=[];let i=0;for(const n of this._tokenInfo){const o=i,r=o+n.text.length;if(r>e.start){if(o>=e.endExclusive)break;const a=Math.max(0,e.start-o),l=Math.max(0,r-e.endExclusive);t.push(new sK(n.text.slice(a,n.text.length-l),n.metadata))}i+=n.text.length}return ep.create(t)}append(e){const t=this._tokenInfo.concat(e._tokenInfo);return ep.create(t)}}class sK{constructor(e,t){this.text=e,this.metadata=t}}var mPe=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},_Pe=function(s,e){return function(t,i){e(t,i,s)}};const boe="ghost-text";var Jp;let l0=(Jp=class extends F{constructor(e,t,i,n,o,r){super(),this._editor=e,this._model=t,this._options=i,this._shouldKeepCursorStable=n,this._isClickable=o,this._languageService=r,this._isDisposed=Re(this,!1),this._editorObs=di(this._editor),this._warningState=j(a=>{const l=this._model.ghostText.read(a);if(!l)return;const c=this._model.warning.read(a);if(c)return{lineNumber:l.lineNumber,position:new A(l.lineNumber,l.parts[0].column),icon:c.icon}}),this._onDidClick=this._register(new M),this.onDidClick=this._onDidClick.event,this._useSyntaxHighlighting=this._options.map(a=>a.syntaxHighlightingEnabled),this._extraClassNames=j(this,a=>{const l=[...this._options.read(a).extraClasses??[]];return this._useSyntaxHighlighting.read(a)&&l.push("syntax-highlighted"),this._warningState.read(a)&&l.push("warning"),l.map(d=>` ${d}`).join("")}),this.uiState=j(this,a=>{if(this._isDisposed.read(a))return;const l=this._editorObs.model.read(a);if(l!==this._model.targetTextModel.read(a))return;const c=this._model.ghostText.read(a);if(!c)return;const d=c instanceof h3?c.columnRange:void 0,h=this._useSyntaxHighlighting.read(a),u=this._extraClassNames.read(a),{inlineTexts:g,additionalLines:f,hiddenRange:p,additionalLinesOriginalSuffix:_}=bPe(c,l,boe+u),b=l.getLineContent(c.lineNumber),w=new vh(g.map(O=>On.insert(O.column-1,O.text))),C=h?l.tokenization.tokenizeLinesAt(c.lineNumber,[w.apply(b),...f.map(O=>O.content)]):void 0,v=w.getNewRanges(),y=g.map((O,P)=>({...O,tokens:C?.[0]?.getTokensInRange(v[P])})),x=f.map((O,P)=>{let B=C?.[P+1]??Ri.createEmpty(O.content,this._languageService.languageIdCodec);if(P===f.length-1&&_){const z=ep.fromLineTokens(l.tokenization.getLineTokens(_.lineNumber)).slice(_.columnRange.toZeroBasedOffsetRange());B=ep.fromLineTokens(B).append(z).toLineTokens(B.languageIdCodec)}return{content:B,decorations:O.decorations}}),L=this._editor.getSelection()?.getStartPosition().column,E=y.filter(O=>O.text!==""),I=E.length!==0,R={cursorColumnDistance:(I?E[0].column:1)-L,cursorLineDistance:I?0:f.findIndex(O=>O.content!=="")+1,lineCountOriginal:I?1:0,lineCountModified:f.length+(I?1:0),characterCountOriginal:0,characterCountModified:SE(E.map(O=>O.text.length))+SE(x.map(O=>O.content.getTextLength())),disjointReplacements:E.length+(f.length>0?1:0),sameShapeReplacements:E.length>1&&x.length===0?E.every(O=>O.text===E[0].text):void 0};return this._model.handleInlineCompletionShown.read(a)?.(R),{replacedRange:d,inlineTexts:y,additionalLines:x,hiddenRange:p,lineNumber:c.lineNumber,additionalReservedLineCount:this._model.minReservedLineCount.read(a),targetTextModel:l,syntaxHighlightingEnabled:h}}),this.decorations=j(this,a=>{const l=this.uiState.read(a);if(!l)return[];const c=[],d=this._extraClassNames.read(a);l.replacedRange&&c.push({range:l.replacedRange.toRange(l.lineNumber),options:{inlineClassName:"inline-completion-text-to-replace"+d,description:"GhostTextReplacement"}}),l.hiddenRange&&c.push({range:l.hiddenRange.toRange(l.lineNumber),options:{inlineClassName:"ghost-text-hidden",description:"ghost-text-hidden"}});for(const h of l.inlineTexts)c.push({range:k.fromPositions(new A(l.lineNumber,h.column)),options:{description:"ghost-text-decoration",after:{content:h.text,tokens:h.tokens,inlineClassName:(h.preview?"ghost-text-decoration-preview":"ghost-text-decoration")+(this._isClickable?" clickable":"")+d+h.lineDecorations.map(u=>" "+u.className).join(" "),cursorStops:jr.Left,attachedData:new mO(this)},showIfCollapsed:!0}});return c}),this._additionalLinesWidget=this._register(new wPe(this._editor,j(a=>{const l=this.uiState.read(a);return l?{lineNumber:l.lineNumber,additionalLines:l.additionalLines,minReservedLineCount:l.additionalReservedLineCount,targetTextModel:l.targetTextModel}:void 0}),this._shouldKeepCursorStable,this._isClickable)),this._isInlineTextHovered=this._editorObs.isTargetHovered(a=>a.target.type===6&&a.target.detail.injectedText?.options.attachedData instanceof mO&&a.target.detail.injectedText.options.attachedData.owner===this,this._store),this.isHovered=j(this,a=>this._isDisposed.read(a)?!1:this._isInlineTextHovered.read(a)||this._additionalLinesWidget.isHovered.read(a)),this.height=j(this,a=>this._editorObs.getOption(75).read(a)+(this._additionalLinesWidget.viewZoneHeight.read(a)??0)),this._register(pe(()=>{this._isDisposed.set(!0,void 0)})),this._register(this._editorObs.setDecorations(this.decorations)),this._isClickable&&(this._register(this._additionalLinesWidget.onDidClick(a=>this._onDidClick.fire(a))),this._register(this._editor.onMouseUp(a=>{if(a.target.type!==6)return;const l=a.target.detail.injectedText?.options.attachedData;l instanceof mO&&l.owner===this&&this._onDidClick.fire(a.event)}))),this._register(ps((a,l)=>{}))}static getWarningWidgetContext(e){const t=e.ghostTextViewWarningWidgetData;if(t)return t;if(e.parentElement)return this.getWarningWidgetContext(e.parentElement)}ownsViewZone(e){return this._additionalLinesWidget.viewZoneId===e}},Jp.hot=bA(Jp),Jp);l0=mPe([_Pe(5,Ei)],l0);class mO{constructor(e){this.owner=e}}function bPe(s,e,t){const i=[],n=[];function o(h,u){if(n.length>0){const g=n[n.length-1];u&&g.decorations.push(new Rs(g.content.length+1,g.content.length+1+h[0].line.length,u,0)),g.content+=h[0].line,h=h.slice(1)}for(const g of h)n.push({content:g.line,decorations:u?[new Rs(1,g.line.length+1,u,0),...g.lineDecorations]:[...g.lineDecorations]})}const r=e.getLineContent(s.lineNumber);let a,l=0;for(const h of s.parts){let u=h.lines;a===void 0?(i.push({column:h.column,text:u[0].line,preview:h.preview,lineDecorations:u[0].lineDecorations}),u=u.slice(1)):o([{line:r.substring(l,h.column-1),lineDecorations:[]}],void 0),u.length>0&&(o(u,t),a===void 0&&h.column<=r.length&&(a=h.column)),l=h.column-1}let c;a!==void 0&&(c=new pPe(s.lineNumber,new nK(l+1,r.length+1)));const d=a!==void 0?new nK(a,r.length+1):void 0;return{inlineTexts:i,additionalLines:n,hiddenRange:d,additionalLinesOriginalSuffix:c}}class wPe extends F{get viewZoneId(){return this._viewZoneInfo?.viewZoneId}get viewZoneHeight(){return this._viewZoneHeight}constructor(e,t,i,n){super(),this._editor=e,this._lines=t,this._shouldKeepCursorStable=i,this._isClickable=n,this._viewZoneHeight=Re("viewZoneHeight",void 0),this.editorOptionsChanged=Po("editorOptionChanged",oe.filter(this._editor.onDidChangeConfiguration,o=>o.hasChanged(40)||o.hasChanged(133)||o.hasChanged(113)||o.hasChanged(108)||o.hasChanged(60)||o.hasChanged(59)||o.hasChanged(75))),this._onDidClick=this._register(new M),this.onDidClick=this._onDidClick.event,this._viewZoneListener=this._register(new vt),this.isHovered=di(this._editor).isTargetHovered(o=>oK(o.target.element),this._store),this.hasBeenAccepted=!1,this._editor instanceof Am&&this._shouldKeepCursorStable&&this._register(this._editor.onBeforeExecuteEdit(o=>this.hasBeenAccepted=o.source==="inlineSuggestion.accept")),this._register(De(o=>{const r=this._lines.read(o);this.editorOptionsChanged.read(o),r?(this.hasBeenAccepted=!1,this.updateLines(r.lineNumber,r.additionalLines,r.minReservedLineCount)):this.clear()}))}dispose(){super.dispose(),this.clear()}clear(){this._viewZoneListener.clear(),this._editor.changeViewZones(e=>{this.removeActiveViewZone(e)})}updateLines(e,t,i){const n=this._editor.getModel();if(!n)return;const{tabSize:o}=n.getOptions();this._editor.changeViewZones(r=>{const a=new U;this.removeActiveViewZone(r);const l=Math.max(t.length,i);if(l>0){const c=document.createElement("div");vPe(c,o,t,this._editor.getOptions(),this._isClickable),this._isClickable&&(a.add(W(c,"mousedown",d=>{d.preventDefault()})),a.add(W(c,"click",d=>{oK(d.target)&&this._onDidClick.fire(new is(be(d),d))}))),this.addViewZone(r,e,l,c)}this._viewZoneListener.value=a})}addViewZone(e,t,i,n){const o=e.addZone({afterLineNumber:t,heightInLines:i,domNode:n,afterColumnAffinity:1,onComputedHeight:r=>{this._viewZoneHeight.set(r,void 0)}});this.keepCursorStable(t,i),this._viewZoneInfo={viewZoneId:o,heightInLines:i,lineNumber:t}}removeActiveViewZone(e){this._viewZoneInfo&&(e.removeZone(this._viewZoneInfo.viewZoneId),this.hasBeenAccepted||this.keepCursorStable(this._viewZoneInfo.lineNumber,-this._viewZoneInfo.heightInLines),this._viewZoneInfo=void 0,this._viewZoneHeight.set(void 0,void 0))}keepCursorStable(e,t){if(!this._shouldKeepCursorStable)return;const i=this._editor.getSelection()?.getStartPosition()?.lineNumber;i!==void 0&&e<i&&this._editor.setScrollTop(this._editor.getScrollTop()+t*this._editor.getOption(75))}}function oK(s){return ki(s)&&s.classList.contains(boe)}function vPe(s,e,t,i,n){const o=i.get(40),r=i.get(133),a="none",l=i.get(108),c=i.get(60),d=i.get(59),h=i.get(75);let u="suggest-preview-text";n&&(u+=" clickable");const g=new Eg(1e4);g.appendString(`<div class="${u}">`);for(let _=0,b=t.length;_<b;_++){const w=t[_],C=w.content;g.appendString('<div class="view-line'),g.appendString('" style="top:'),g.appendString(String(_*h)),g.appendString('px;width:1000000px;">');const v=C.getLineContent(),y=q0(v),x=iv(v);nC(new Dh(d.isMonospace&&!o,d.canUseHalfwidthRightwardsArrow,v,!1,y,x,0,C,w.decorations,e,0,d.spaceWidth,d.middotWidth,d.wsmiddotWidth,r,a,l,c!==mh.OFF,null,null,0),g),g.appendString("</div>")}g.appendString("</div>"),In(s,d);const f=g.build(),p=rK?rK.createHTML(f):f;s.innerHTML=p}const rK=$c("editorGhostText",{createHTML:s=>s}),BN=class BN{constructor(e){this.replacements=e,Vy(H0(e,(t,i)=>t.lineRange.endLineNumberExclusive<=i.lineRange.startLineNumber))}toString(){return this.replacements.map(e=>e.toString()).join(",")}getNewLineRanges(){const e=[];let t=0;for(const i of this.replacements)e.push(Ne.ofLength(i.lineRange.startLineNumber+t,i.newLines.length)),t+=i.newLines.length-i.lineRange.length;return e}};BN.empty=new BN([]);let M3=BN;class zD{static fromSingleTextEdit(e,t){const i=qo(e.text);let n=e.range.startLineNumber;const o=t.getValueOfRange(k.fromPositions(new A(e.range.startLineNumber,1),e.range.getStartPosition()));i[0]=o+i[0];let r=e.range.endLineNumber+1;const a=t.getTransformer().getLineLength(e.range.endLineNumber)+1,l=t.getValueOfRange(k.fromPositions(e.range.getEndPosition(),new A(e.range.endLineNumber,a)));i[i.length-1]=i[i.length-1]+l;const c=e.range.startColumn===t.getTransformer().getLineLength(e.range.startLineNumber)+1,d=e.range.endColumn===1;return c&&i[0].length===o.length&&(n++,i.shift()),i.length>0&&n<r&&d&&i[i.length-1].length===l.length&&(r--,i.pop()),new zD(new Ne(n,r),i)}constructor(e,t){this.lineRange=e,this.newLines=t}toString(){return`${this.lineRange}->${JSON.stringify(this.newLines)}`}toLineEdit(){return new M3([this])}}class woe{get lineEdit(){return this.edit.replacements.length===0?new zD(new Ne(1,1),[]):zD.fromSingleTextEdit(this.edit.toReplacement(this.originalText),this.originalText)}get originalLineRange(){return this.lineEdit.lineRange}get modifiedLineRange(){return this.lineEdit.toLineEdit().getNewLineRanges()[0]}get displayRange(){return this.originalText.lineRange.intersect(this.originalLineRange.join(Ne.ofLength(this.originalLineRange.startLineNumber,this.lineEdit.newLines.length)))}constructor(e,t,i,n,o,r){this.originalText=e,this.edit=t,this.cursorPosition=i,this.multiCursorPositions=n,this.commands=o,this.inlineCompletion=r}}class voe{constructor(e,t,i){this._model=e,this.inlineEdit=t,this.tabAction=i,this.action=this.inlineEdit.inlineCompletion.action,this.displayName=this.inlineEdit.inlineCompletion.source.provider.displayName??m(1219,"Inline Edit"),this.extensionCommands=this.inlineEdit.inlineCompletion.source.inlineSuggestions.commands??[],this.isInDiffEditor=this._model.isInDiffEditor,this.displayLocation=this.inlineEdit.inlineCompletion.hint,this.showCollapsed=this._model.showCollapsed}accept(){this._model.accept()}jump(){this._model.jump()}handleInlineEditShown(e,t){this._model.handleInlineSuggestionShown(this.inlineEdit.inlineCompletion,e,t)}}class CPe{constructor(e){this._model=e,this.onDidAccept=this._model.onDidAccept,this.inAcceptFlow=this._model.inAcceptFlow}}class yPe{constructor(e,t,i,n){this.lineRange=i;const o=di(e),r=j(this,a=>o.isFocused.read(a)&&n.showInlineEditMenu?us.Accept:us.Inactive);this.model=new voe(t,new woe(new vu(""),new Na([n.getSingleTextEdit()]),t.primaryPosition.get(),t.allPositions.get(),n.source.inlineSuggestions.commands??[],n),r)}}class Ft{static fromPoints(e,t){return new Ft(e.x,e.y,t.x,t.y)}static fromPointSize(e,t){return new Ft(e.x,e.y,e.x+t.x,e.y+t.y)}static fromLeftTopRightBottom(e,t,i,n){return new Ft(e,t,i,n)}static fromLeftTopWidthHeight(e,t,i,n){return new Ft(e,t,e+i,t+n)}static fromRanges(e,t){return new Ft(e.start,t.start,e.endExclusive,t.endExclusive)}static hull(e){let t=Number.MAX_SAFE_INTEGER,i=Number.MAX_SAFE_INTEGER,n=Number.MIN_SAFE_INTEGER,o=Number.MIN_SAFE_INTEGER;for(const r of e)t=Math.min(t,r.left),i=Math.min(i,r.top),n=Math.max(n,r.right),o=Math.max(o,r.bottom);return new Ft(t,i,n,o)}get width(){return this.right-this.left}get height(){return this.bottom-this.top}constructor(e,t,i,n){if(this.left=e,this.top=t,this.right=i,this.bottom=n,e>i)throw new Ce("Invalid arguments: Horizontally offset by "+(e-i));if(t>n)throw new Ce("Invalid arguments: Vertically offset by "+(t-n))}withMargin(e,t,i,n){let o,r,a,l;return t===void 0&&i===void 0&&n===void 0?o=r=a=l=e:i===void 0&&n===void 0?(o=r=t,a=l=e):(o=n,r=t,a=e,l=i),new Ft(this.left-o,this.top-a,this.right+r,this.bottom+l)}intersectVertical(e){const t=Math.max(this.top,e.start),i=Math.min(this.bottom,e.endExclusive);return new Ft(this.left,t,this.right,Math.max(t,i))}intersectHorizontal(e){const t=Math.max(this.left,e.start),i=Math.min(this.right,e.endExclusive);return new Ft(t,this.top,Math.max(t,i),this.bottom)}toString(){return`Rect{(${this.left},${this.top}), (${this.right},${this.bottom})}`}intersect(e){const t=Math.max(this.left,e.left),i=Math.min(this.right,e.right),n=Math.max(this.top,e.top),o=Math.min(this.bottom,e.bottom);if(!(t>i||n>o))return new Ft(t,n,i,o)}containsRect(e){return this.left<=e.left&&this.top<=e.top&&this.right>=e.right&&this.bottom>=e.bottom}containsPoint(e){return this.left<=e.x&&this.top<=e.y&&this.right>=e.x&&this.bottom>=e.y}moveToBeContainedIn(e){const t=this.width,i=this.height;let n=this.left,o=this.top;return n<e.left?n=e.left:n+t>e.right&&(n=e.right-t),o<e.top?o=e.top:o+i>e.bottom&&(o=e.bottom-i),new Ft(n,o,n+t,o+i)}withWidth(e){return new Ft(this.left,this.top,this.left+e,this.bottom)}withHeight(e){return new Ft(this.left,this.top,this.right,this.top+e)}withTop(e){return new Ft(this.left,e,this.right,this.bottom)}withLeft(e){return new Ft(e,this.top,this.right,this.bottom)}translateX(e){return new Ft(this.left+e,this.top,this.right+e,this.bottom)}translateY(e){return new Ft(this.left,this.top+e,this.right,this.bottom+e)}getLeftBottom(){return new _n(this.left,this.bottom)}getRightBottom(){return new _n(this.right,this.bottom)}getRightTop(){return new _n(this.right,this.top)}toStyles(){return{position:"absolute",left:`${this.left}px`,top:`${this.top}px`,width:`${this.width}px`,height:`${this.height}px`}}}class ry{constructor(e,t,i,n=null){this.startLineNumbers=e,this.endLineNumbers=t,this.lastLineRelativePosition=i,this.showEndForLine=n}equals(e){return!!e&&this.lastLineRelativePosition===e.lastLineRelativePosition&&this.showEndForLine===e.showEndForLine&&ni(this.startLineNumbers,e.startLineNumbers)&&ni(this.endLineNumbers,e.endLineNumbers)}static get Empty(){return new ry([],[],0)}}const aK=$c("stickyScrollViewLayer",{createHTML:s=>s}),_O="data-sticky-line-index",lK="data-sticky-is-line",SPe="data-sticky-is-line-number",cK="data-sticky-is-folding-icon";class xPe extends F{get height(){return this._height}constructor(e){super(),this._foldingIconStore=this._register(new U),this._rootDomNode=document.createElement("div"),this._lineNumbersDomNode=document.createElement("div"),this._linesDomNodeScrollable=document.createElement("div"),this._linesDomNode=document.createElement("div"),this._renderedStickyLines=[],this._lineNumbers=[],this._lastLineRelativePosition=0,this._minContentWidthInPx=0,this._isOnGlyphMargin=!1,this._height=-1,this._onDidChangeStickyScrollHeight=this._register(new M),this.onDidChangeStickyScrollHeight=this._onDidChangeStickyScrollHeight.event,this._editor=e,this._lineHeight=e.getOption(75),this._lineNumbersDomNode.className="sticky-widget-line-numbers",this._lineNumbersDomNode.setAttribute("role","none"),this._linesDomNode.className="sticky-widget-lines",this._linesDomNode.setAttribute("role","list"),this._linesDomNodeScrollable.className="sticky-widget-lines-scrollable",this._linesDomNodeScrollable.appendChild(this._linesDomNode),this._rootDomNode.className="sticky-widget",this._rootDomNode.classList.toggle("peek",e instanceof Lh),this._rootDomNode.appendChild(this._lineNumbersDomNode),this._rootDomNode.appendChild(this._linesDomNodeScrollable),this._setHeight(0);const t=()=>{this._linesDomNode.style.left=this._editor.getOption(131).scrollWithEditor?`-${this._editor.getScrollLeft()}px`:"0px"};this._register(this._editor.onDidChangeConfiguration(i=>{i.hasChanged(131)&&t(),i.hasChanged(75)&&(this._lineHeight=this._editor.getOption(75))})),this._register(this._editor.onDidScrollChange(i=>{i.scrollLeftChanged&&t(),i.scrollWidthChanged&&this._updateWidgetWidth()})),this._register(this._editor.onDidChangeModel(()=>{t(),this._updateWidgetWidth()})),t(),this._register(this._editor.onDidLayoutChange(i=>{this._updateWidgetWidth()})),this._updateWidgetWidth()}get lineNumbers(){return this._lineNumbers}get lineNumberCount(){return this._lineNumbers.length}getRenderedStickyLine(e){return this._renderedStickyLines.find(t=>t.lineNumber===e)}getCurrentLines(){return this._lineNumbers}setState(e,t,i){const n=!this._state&&!e,o=this._state&&this._state.equals(e);if(i===void 0&&(n||o))return;const r=this._findRenderingData(e),a=this._lineNumbers;this._lineNumbers=r.lineNumbers,this._lastLineRelativePosition=r.lastLineRelativePosition;const l=this._findIndexToRebuildFrom(a,this._lineNumbers,i);this._renderRootNode(this._lineNumbers,this._lastLineRelativePosition,t,l),this._state=e}_findRenderingData(e){if(!e)return{lineNumbers:[],lastLineRelativePosition:0};const t=[...e.startLineNumbers];e.showEndForLine!==null&&(t[e.showEndForLine]=e.endLineNumbers[e.showEndForLine]);let i=0;for(let n=0;n<t.length;n++)i+=this._editor.getLineHeightForPosition(new A(t[n],1));return i===0?{lineNumbers:[],lastLineRelativePosition:0}:{lineNumbers:t,lastLineRelativePosition:e.lastLineRelativePosition}}_findIndexToRebuildFrom(e,t,i){if(t.length===0)return 0;if(i!==void 0)return i;const n=t.findIndex(o=>!e.includes(o));return n===-1?0:n}_updateWidgetWidth(){const e=this._editor.getLayoutInfo(),t=e.contentLeft;this._lineNumbersDomNode.style.width=`${t}px`,this._linesDomNodeScrollable.style.setProperty("--vscode-editorStickyScroll-scrollableWidth",`${this._editor.getScrollWidth()-e.verticalScrollbarWidth}px`),this._rootDomNode.style.width=`${e.width-e.verticalScrollbarWidth}px`}_useFoldingOpacityTransition(e){this._lineNumbersDomNode.style.setProperty("--vscode-editorStickyScroll-foldingOpacityTransition",`opacity ${e?.5:0}s`)}_setFoldingIconsVisibility(e){for(const t of this._renderedStickyLines){const i=t.foldingIcon;i&&i.setVisible(e?!0:i.isCollapsed)}}async _renderRootNode(e,t,i,n){const o=this._editor._getViewModel();if(!o){this._clearWidget();return}if(e.length===0){this._clearWidget();return}const r=[],a=e[e.length-1];let l=0;for(let d=0;d<this._renderedStickyLines.length;d++)if(d<n){const h=this._renderedStickyLines[d];r.push(this._updatePosition(h,l,h.lineNumber===a)),l+=h.height}else{const h=this._renderedStickyLines[d];h.lineNumberDomNode.remove(),h.lineDomNode.remove()}const c=this._editor.getLayoutInfo();for(let d=n;d<e.length;d++){const h=this._renderChildNode(o,d,e[d],l,a===e[d],i,c);l+=h.height,this._linesDomNode.appendChild(h.lineDomNode),this._lineNumbersDomNode.appendChild(h.lineNumberDomNode),r.push(h)}i&&(this._setFoldingHoverListeners(),this._useFoldingOpacityTransition(!this._isOnGlyphMargin)),this._minContentWidthInPx=Math.max(...this._renderedStickyLines.map(d=>d.scrollWidth))+c.verticalScrollbarWidth,this._renderedStickyLines=r,this._setHeight(l+t),this._editor.layoutOverlayWidget(this)}_clearWidget(){for(let e=0;e<this._renderedStickyLines.length;e++){const t=this._renderedStickyLines[e];t.lineNumberDomNode.remove(),t.lineDomNode.remove()}this._setHeight(0)}_setHeight(e){this._height!==e&&(this._height=e,this._height===0?this._rootDomNode.style.display="none":(this._rootDomNode.style.display="block",this._lineNumbersDomNode.style.height=`${this._height}px`,this._linesDomNodeScrollable.style.height=`${this._height}px`,this._rootDomNode.style.height=`${this._height}px`),this._onDidChangeStickyScrollHeight.fire({height:this._height}))}_setFoldingHoverListeners(){this._editor.getOption(126)==="mouseover"&&(this._foldingIconStore.clear(),this._foldingIconStore.add(W(this._lineNumbersDomNode,se.MOUSE_ENTER,()=>{this._isOnGlyphMargin=!0,this._setFoldingIconsVisibility(!0)})),this._foldingIconStore.add(W(this._lineNumbersDomNode,se.MOUSE_LEAVE,()=>{this._isOnGlyphMargin=!1,this._useFoldingOpacityTransition(!0),this._setFoldingIconsVisibility(!1)})))}_renderChildNode(e,t,i,n,o,r,a){const l=e.coordinatesConverter.convertModelPositionToViewPosition(new A(i,1)).lineNumber,c=e.getViewLineRenderingData(l),d=this._editor.getOption(76),h=this._editor.getOption(117).verticalScrollbarSize;let u;try{u=Rs.filter(c.inlineDecorations,l,c.minColumn,c.maxColumn)}catch{u=[]}const g=this._editor.getLineHeightForPosition(new A(i,1)),f=e.getTextDirection(i),p=new Dh(!0,!0,c.content,c.continuesWithWrappedLine,c.isBasicASCII,c.containsRTL,0,c.tokens,u,c.tabSize,c.startVisibleColumn,1,1,1,500,"none",!0,!0,null,f,h),_=new Eg(2e3),b=nC(p,_);let w;aK?w=aK.createHTML(_.build()):w=_.build();const C=document.createElement("span");C.setAttribute(_O,String(t)),C.setAttribute(lK,""),C.setAttribute("role","listitem"),C.tabIndex=0,C.className="sticky-line-content",C.classList.add(`stickyLine${i}`),C.style.lineHeight=`${g}px`,C.innerHTML=w;const v=document.createElement("span");v.setAttribute(_O,String(t)),v.setAttribute(SPe,""),v.className="sticky-line-number",v.style.lineHeight=`${g}px`;const y=a.contentLeft;v.style.width=`${y}px`;const x=document.createElement("span");d.renderType===1||d.renderType===3&&i%10===0?x.innerText=i.toString():d.renderType===2&&(x.innerText=Math.abs(i-this._editor.getPosition().lineNumber).toString()),x.className="sticky-line-number-inner",x.style.width=`${a.lineNumbersWidth}px`,x.style.paddingLeft=`${a.lineNumbersLeft}px`,v.appendChild(x);const L=this._renderFoldingIconForLine(r,i);L&&(v.appendChild(L.domNode),L.domNode.style.left=`${a.lineNumbersWidth+a.lineNumbersLeft}px`,L.domNode.style.lineHeight=`${g}px`),this._editor.applyFontInfo(C),this._editor.applyFontInfo(v),v.style.lineHeight=`${g}px`,C.style.lineHeight=`${g}px`,v.style.height=`${g}px`,C.style.height=`${g}px`;const E=new LPe(t,i,C,v,L,b.characterMapping,C.scrollWidth,g);return this._updatePosition(E,n,o)}_updatePosition(e,t,i){const n=e.lineDomNode,o=e.lineNumberDomNode;if(i){n.style.zIndex="0",o.style.zIndex="0";const a=`${t+this._lastLineRelativePosition+(e.foldingIcon?.isCollapsed?1:0)}px`;n.style.top=a,o.style.top=a}else n.style.zIndex="1",o.style.zIndex="1",n.style.top=`${t}px`,o.style.top=`${t}px`;return e}_renderFoldingIconForLine(e,t){const i=this._editor.getOption(126);if(!e||i==="never")return;const n=e.regions,o=n.findRange(t),r=n.getStartLineNumber(o);if(!(t===r))return;const l=n.isCollapsed(o),c=new kPe(l,r,n.getEndLineNumber(o),this._lineHeight);return c.setVisible(this._isOnGlyphMargin?!0:l||i==="always"),c.domNode.setAttribute(cK,""),c}getId(){return"editor.contrib.stickyScrollWidget"}getDomNode(){return this._rootDomNode}getPosition(){return{preference:2,stackOridinal:10}}getMinContentWidthInPx(){return this._minContentWidthInPx}focusLineWithIndex(e){0<=e&&e<this._renderedStickyLines.length&&this._renderedStickyLines[e].lineDomNode.focus()}getEditorPositionFromNode(e){if(!e||e.children.length>0)return null;const t=this._getRenderedStickyLineFromChildDomNode(e);if(!t)return null;const i=dv(t.characterMapping,e,0);return new A(t.lineNumber,i)}getLineNumberFromChildDomNode(e){return this._getRenderedStickyLineFromChildDomNode(e)?.lineNumber??null}_getRenderedStickyLineFromChildDomNode(e){const t=this.getLineIndexFromChildDomNode(e);return t===null||t<0||t>=this._renderedStickyLines.length?null:this._renderedStickyLines[t]}getLineIndexFromChildDomNode(e){const t=this._getAttributeValue(e,_O);return t?parseInt(t,10):null}isInStickyLine(e){return this._getAttributeValue(e,lK)!==void 0}isInFoldingIconDomNode(e){return this._getAttributeValue(e,cK)!==void 0}_getAttributeValue(e,t){for(;e&&e!==this._rootDomNode;){const i=e.getAttribute(t);if(i!==null)return i;e=e.parentElement}}}class LPe{constructor(e,t,i,n,o,r,a,l){this.index=e,this.lineNumber=t,this.lineDomNode=i,this.lineNumberDomNode=n,this.foldingIcon=o,this.characterMapping=r,this.scrollWidth=a,this.height=l}}class kPe{constructor(e,t,i,n){this.isCollapsed=e,this.foldingStartLine=t,this.foldingEndLine=i,this.dimension=n,this.domNode=document.createElement("div"),this.domNode.style.width="26px",this.domNode.style.height=`${n}px`,this.domNode.style.lineHeight=`${n}px`,this.domNode.className=xe.asClassName(e?bD:_D)}setVisible(e){this.domNode.style.cursor=e?"pointer":"default",this.domNode.style.opacity=e?"1":"0"}}class ay{constructor(e,t){this.startLineNumber=e,this.endLineNumber=t}}class $D{constructor(e,t,i){this.range=e,this.children=t,this.parent=i}}class Coe{constructor(e,t,i,n){this.uri=e,this.version=t,this.element=i,this.outlineProviderId=n}}var ZA=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},c0=function(s,e){return function(t,i){e(t,i,s)}},ly;(function(s){s.OUTLINE_MODEL="outlineModel",s.FOLDING_PROVIDER_MODEL="foldingProviderModel",s.INDENTATION_MODEL="indentationModel"})(ly||(ly={}));var Iu;(function(s){s[s.VALID=0]="VALID",s[s.INVALID=1]="INVALID",s[s.CANCELED=2]="CANCELED"})(Iu||(Iu={}));let O3=class extends F{constructor(e,t,i,n){switch(super(),this._editor=e,this._modelProviders=[],this._modelPromise=null,this._updateScheduler=this._register(new sa(300)),this._updateOperation=this._register(new U),this._editor.getOption(131).defaultModel){case ly.OUTLINE_MODEL:this._modelProviders.push(new P3(this._editor,n));case ly.FOLDING_PROVIDER_MODEL:this._modelProviders.push(new B3(this._editor,t,n));case ly.INDENTATION_MODEL:this._modelProviders.push(new F3(this._editor,i));break}}dispose(){this._modelProviders.forEach(e=>e.dispose()),this._updateOperation.clear(),this._cancelModelPromise(),super.dispose()}_cancelModelPromise(){this._modelPromise&&(this._modelPromise.cancel(),this._modelPromise=null)}async update(e){return this._updateOperation.clear(),this._updateOperation.add({dispose:()=>{this._cancelModelPromise(),this._updateScheduler.cancel()}}),this._cancelModelPromise(),await this._updateScheduler.trigger(async()=>{for(const t of this._modelProviders){const{statusPromise:i,modelPromise:n}=t.computeStickyModel(e);this._modelPromise=n;const o=await i;if(this._modelPromise!==n)return null;switch(o){case Iu.CANCELED:return this._updateOperation.clear(),null;case Iu.VALID:return t.stickyModel}}return null}).catch(t=>(Oe(t),null))}};O3=ZA([c0(2,_e),c0(3,ge)],O3);class yoe extends F{constructor(e){super(),this._editor=e,this._stickyModel=null}get stickyModel(){return this._stickyModel}_invalid(){return this._stickyModel=null,Iu.INVALID}computeStickyModel(e){if(e.isCancellationRequested||!this.isProviderValid())return{statusPromise:this._invalid(),modelPromise:null};const t=an(i=>this.createModelFromProvider(i));return{statusPromise:t.then(i=>this.isModelValid(i)?e.isCancellationRequested?Iu.CANCELED:(this._stickyModel=this.createStickyModel(e,i),Iu.VALID):this._invalid()).then(void 0,i=>(Oe(i),Iu.CANCELED)),modelPromise:t}}isModelValid(e){return!0}isProviderValid(){return!0}}let P3=class extends yoe{constructor(e,t){super(e),this._languageFeaturesService=t}createModelFromProvider(e){return Rd.create(this._languageFeaturesService.documentSymbolProvider,this._editor.getModel(),e)}createStickyModel(e,t){const{stickyOutlineElement:i,providerID:n}=this._stickyModelFromOutlineModel(t,this._stickyModel?.outlineProviderId),o=this._editor.getModel();return new Coe(o.uri,o.getVersionId(),i,n)}isModelValid(e){return e&&e.children.size>0}_stickyModelFromOutlineModel(e,t){let i;if(ot.first(e.children.values())instanceof Jse){const a=ot.find(e.children.values(),l=>l.id===t);if(a)i=a.children;else{let l="",c=-1,d;for(const[h,u]of e.children.entries()){const g=this._findSumOfRangesOfGroup(u);g>c&&(d=u,c=g,l=u.id)}t=l,i=d.children}}else i=e.children;const n=[],o=Array.from(i.values()).sort((a,l)=>{const c=new ay(a.symbol.range.startLineNumber,a.symbol.range.endLineNumber),d=new ay(l.symbol.range.startLineNumber,l.symbol.range.endLineNumber);return this._comparator(c,d)});for(const a of o)n.push(this._stickyModelFromOutlineElement(a,a.symbol.selectionRange.startLineNumber));return{stickyOutlineElement:new $D(void 0,n,void 0),providerID:t}}_stickyModelFromOutlineElement(e,t){const i=[];for(const o of e.children.values())if(o.symbol.selectionRange.startLineNumber!==o.symbol.range.endLineNumber)if(o.symbol.selectionRange.startLineNumber!==t)i.push(this._stickyModelFromOutlineElement(o,o.symbol.selectionRange.startLineNumber));else for(const r of o.children.values())i.push(this._stickyModelFromOutlineElement(r,o.symbol.selectionRange.startLineNumber));i.sort((o,r)=>this._comparator(o.range,r.range));const n=new ay(e.symbol.selectionRange.startLineNumber,e.symbol.range.endLineNumber);return new $D(n,i,void 0)}_comparator(e,t){return e.startLineNumber!==t.startLineNumber?e.startLineNumber-t.startLineNumber:t.endLineNumber-e.endLineNumber}_findSumOfRangesOfGroup(e){let t=0;for(const i of e.children.values())t+=this._findSumOfRangesOfGroup(i);return e instanceof o3?t+e.symbol.range.endLineNumber-e.symbol.selectionRange.startLineNumber:t}};P3=ZA([c0(1,ge)],P3);class Soe extends yoe{constructor(e){super(e),this._foldingLimitReporter=this._register(new Gse(e))}createStickyModel(e,t){const i=this._fromFoldingRegions(t),n=this._editor.getModel();return new Coe(n.uri,n.getVersionId(),i,void 0)}isModelValid(e){return e!==null}_fromFoldingRegions(e){const t=e.length,i=[],n=new $D(void 0,[],void 0);for(let o=0;o<t;o++){const r=e.getParentIndex(o);let a;r!==-1?a=i[r]:a=n;const l=new $D(new ay(e.getStartLineNumber(o),e.getEndLineNumber(o)+1),[],a);a.children.push(l),i.push(l)}return n}}let F3=class extends Soe{constructor(e,t){super(e),this._languageConfigurationService=t,this.provider=this._register(new zH(e.getModel(),this._languageConfigurationService,this._foldingLimitReporter))}async createModelFromProvider(e){return this.provider.compute(e)}};F3=ZA([c0(1,ui)],F3);let B3=class extends Soe{constructor(e,t,i){super(e),this._languageFeaturesService=i,this.provider=this._register(new vt),this._register(this._languageFeaturesService.foldingRangeProvider.onDidChange(()=>{this._updateProvider(e,t)})),this._updateProvider(e,t)}_updateProvider(e,t){const i=vg.getFoldingRangeProviders(this._languageFeaturesService,e.getModel());i.length!==0&&(this.provider.value=new $H(e.getModel(),i,t,this._foldingLimitReporter,void 0))}isProviderValid(){return this.provider!==void 0}async createModelFromProvider(e){return this.provider.value?.compute(e)??null}};B3=ZA([c0(2,ge)],B3);var EPe=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},dK=function(s,e){return function(t,i){e(t,i,s)}};class IPe{constructor(e,t,i,n){this.startLineNumber=e,this.endLineNumber=t,this.top=i,this.height=n}}let W3=class extends F{constructor(e,t,i){super(),this._languageFeaturesService=t,this._languageConfigurationService=i,this._onDidChangeStickyScroll=this._register(new M),this.onDidChangeStickyScroll=this._onDidChangeStickyScroll.event,this._model=null,this._cts=null,this._stickyModelProvider=null,this._editor=e,this._sessionStore=this._register(new U),this._updateSoon=this._register(new Rt(()=>this.update(),50)),this._register(this._editor.onDidChangeConfiguration(n=>{n.hasChanged(131)&&this.readConfiguration()})),this.readConfiguration()}readConfiguration(){this._sessionStore.clear(),this._editor.getOption(131).enabled&&(this._sessionStore.add(this._editor.onDidChangeModel(()=>{this._model=null,this.updateStickyModelProvider(),this._onDidChangeStickyScroll.fire(),this.update()})),this._sessionStore.add(this._editor.onDidChangeHiddenAreas(()=>this.update())),this._sessionStore.add(this._editor.onDidChangeModelContent(()=>this._updateSoon.schedule())),this._sessionStore.add(this._languageFeaturesService.documentSymbolProvider.onDidChange(()=>this.update())),this._sessionStore.add(pe(()=>{this._stickyModelProvider?.dispose(),this._stickyModelProvider=null})),this.updateStickyModelProvider(),this.update())}getVersionId(){return this._model?.version}updateStickyModelProvider(){this._stickyModelProvider?.dispose(),this._stickyModelProvider=null,this._editor.hasModel()&&(this._stickyModelProvider=new O3(this._editor,()=>this._updateSoon.schedule(),this._languageConfigurationService,this._languageFeaturesService))}async update(){this._cts?.dispose(!0),this._cts=new si,await this.updateStickyModel(this._cts.token),this._onDidChangeStickyScroll.fire()}async updateStickyModel(e){if(!this._editor.hasModel()||!this._stickyModelProvider||this._editor.getModel().isTooLargeForTokenization()){this._model=null;return}const t=await this._stickyModelProvider.update(e);e.isCancellationRequested||(this._model=t)}getCandidateStickyLinesIntersecting(e){if(!this._model?.element)return[];const t=[];return this.getCandidateStickyLinesIntersectingFromStickyModel(e,this._model.element,t,0,0,-1),this.filterHiddenRanges(t)}getCandidateStickyLinesIntersectingFromStickyModel(e,t,i,n,o,r){const a=this._editor.getModel();if(!a||t.children.length===0)return;let l=r;const c=[];for(let u=0;u<t.children.length;u++){const g=t.children[u];g.range&&c.push(g.range.startLineNumber)}const d=this.updateIndex(vE(c,e.startLineNumber,(u,g)=>u-g)),h=this.updateIndex(vE(c,e.endLineNumber,(u,g)=>u-g));for(let u=d;u<=h;u++){const g=t.children[u];if(!g||!g.range)continue;const{startLineNumber:f,endLineNumber:p}=g.range;if(p>f+1&&e.startLineNumber<=p+1&&f-1<=e.endLineNumber&&f!==l&&a.isValidRange(new k(f,1,p,1))){l=f;const _=this._editor.getLineHeightForPosition(new A(f,1));i.push(new IPe(f,p-1,o,_)),this.getCandidateStickyLinesIntersectingFromStickyModel(e,g,i,n+1,o+_,f)}}}filterHiddenRanges(e){const t=this._editor._getViewModel()?.getHiddenAreas();return t?e.filter(i=>!t.some(n=>i.startLineNumber>=n.startLineNumber&&i.endLineNumber<=n.endLineNumber+1)):e}updateIndex(e){return e===-1?0:e<0?-e-2:e}};W3=EPe([dK(1,ge),dK(2,ui)],W3);var DPe=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},ib=function(s,e){return function(t,i){e(t,i,s)}},H3,em;let Ha=(em=class extends F{constructor(e,t,i,n,o,r,a){super(),this._editor=e,this._contextMenuService=t,this._languageFeaturesService=i,this._instaService=n,this._contextKeyService=a,this._sessionStore=new U,this._maxStickyLines=Number.MAX_SAFE_INTEGER,this._candidateDefinitionsLength=-1,this._focusedStickyElementIndex=-1,this._enabled=!1,this._focused=!1,this._positionRevealed=!1,this._onMouseDown=!1,this._endLineNumbers=[],this._mouseTarget=null,this._onDidChangeStickyScrollHeight=this._register(new M),this.onDidChangeStickyScrollHeight=this._onDidChangeStickyScrollHeight.event,this._stickyScrollWidget=new xPe(this._editor),this._stickyLineCandidateProvider=new W3(this._editor,i,o),this._register(this._stickyScrollWidget),this._register(this._stickyLineCandidateProvider),this._widgetState=ry.Empty;const l=this._stickyScrollWidget.getDomNode();this._register(this._editor.onDidChangeLineHeight(d=>{d.changes.forEach(h=>{const u=h.lineNumber;this._widgetState.startLineNumbers.includes(u)&&this._renderStickyScroll(u)})})),this._register(this._editor.onDidChangeFont(d=>{d.changes.forEach(h=>{const u=h.lineNumber;this._widgetState.startLineNumbers.includes(u)&&this._renderStickyScroll(u)})})),this._register(this._editor.onDidChangeConfiguration(d=>{this._readConfigurationChange(d)})),this._register(W(l,se.CONTEXT_MENU,async d=>{this._onContextMenu(be(l),d)})),this._stickyScrollFocusedContextKey=N.stickyScrollFocused.bindTo(this._contextKeyService),this._stickyScrollVisibleContextKey=N.stickyScrollVisible.bindTo(this._contextKeyService);const c=this._register(oa(l));this._register(c.onDidBlur(d=>{this._positionRevealed===!1&&l.clientHeight===0?(this._focusedStickyElementIndex=-1,this.focus()):this._disposeFocusStickyScrollStore()})),this._register(c.onDidFocus(d=>{this.focus()})),this._registerMouseListeners(),this._register(W(l,se.MOUSE_DOWN,d=>{this._onMouseDown=!0})),this._register(this._stickyScrollWidget.onDidChangeStickyScrollHeight(d=>{this._onDidChangeStickyScrollHeight.fire(d)})),this._onDidResize(),this._readConfiguration()}get stickyScrollWidgetHeight(){return this._stickyScrollWidget.height}static get(e){return e.getContribution(H3.ID)}_disposeFocusStickyScrollStore(){this._stickyScrollFocusedContextKey.set(!1),this._focusDisposableStore?.dispose(),this._focused=!1,this._positionRevealed=!1,this._onMouseDown=!1}isFocused(){return this._focused}focus(){if(this._onMouseDown){this._onMouseDown=!1,this._editor.focus();return}this._stickyScrollFocusedContextKey.get()!==!0&&(this._focused=!0,this._focusDisposableStore=new U,this._stickyScrollFocusedContextKey.set(!0),this._focusedStickyElementIndex=this._stickyScrollWidget.lineNumbers.length-1,this._stickyScrollWidget.focusLineWithIndex(this._focusedStickyElementIndex))}focusNext(){this._focusedStickyElementIndex<this._stickyScrollWidget.lineNumberCount-1&&this._focusNav(!0)}focusPrevious(){this._focusedStickyElementIndex>0&&this._focusNav(!1)}selectEditor(){this._editor.focus()}_focusNav(e){this._focusedStickyElementIndex=e?this._focusedStickyElementIndex+1:this._focusedStickyElementIndex-1,this._stickyScrollWidget.focusLineWithIndex(this._focusedStickyElementIndex)}goToFocused(){const e=this._stickyScrollWidget.lineNumbers;this._disposeFocusStickyScrollStore(),this._revealPosition({lineNumber:e[this._focusedStickyElementIndex],column:1})}_revealPosition(e){this._reveaInEditor(e,()=>this._editor.revealPosition(e))}_revealLineInCenterIfOutsideViewport(e){this._reveaInEditor(e,()=>this._editor.revealLineInCenterIfOutsideViewport(e.lineNumber,0))}_reveaInEditor(e,t){this._focused&&this._disposeFocusStickyScrollStore(),this._positionRevealed=!0,t(),this._editor.setSelection(k.fromPositions(e)),this._editor.focus()}_registerMouseListeners(){const e=this._register(new U),t=this._register(new AA(this._editor,{extractLineNumberFromMouseEvent:o=>{const r=this._stickyScrollWidget.getEditorPositionFromNode(o.target.element);return r?r.lineNumber:0}})),i=o=>{if(!this._editor.hasModel()||o.target.type!==12||o.target.detail!==this._stickyScrollWidget.getId())return null;const r=o.target.element;if(!r||r.innerText!==r.innerHTML)return null;const a=this._stickyScrollWidget.getEditorPositionFromNode(r);return a?{range:new k(a.lineNumber,a.column,a.lineNumber,a.column+r.innerText.length),textElement:r}:null},n=this._stickyScrollWidget.getDomNode();this._register(Fi(n,se.CLICK,o=>{if(o.ctrlKey||o.altKey||o.metaKey||!o.leftButton)return;if(o.shiftKey){const c=this._stickyScrollWidget.getLineIndexFromChildDomNode(o.target);if(c===null)return;const d=new A(this._endLineNumbers[c],1);this._revealLineInCenterIfOutsideViewport(d);return}if(this._stickyScrollWidget.isInFoldingIconDomNode(o.target)){const c=this._stickyScrollWidget.getLineNumberFromChildDomNode(o.target);this._toggleFoldingRegionForLine(c);return}if(!this._stickyScrollWidget.isInStickyLine(o.target))return;let l=this._stickyScrollWidget.getEditorPositionFromNode(o.target);if(!l){const c=this._stickyScrollWidget.getLineNumberFromChildDomNode(o.target);if(c===null)return;l=new A(c,1)}this._revealPosition(l)})),this._register(W(Nt,se.MOUSE_MOVE,o=>{this._mouseTarget=o.target,this._onMouseMoveOrKeyDown(o)})),this._register(W(Nt,se.KEY_DOWN,o=>{this._onMouseMoveOrKeyDown(o)})),this._register(W(Nt,se.KEY_UP,()=>{this._showEndForLine!==void 0&&(this._showEndForLine=void 0,this._renderStickyScroll())})),this._register(t.onMouseMoveOrRelevantKeyDown(([o,r])=>{const a=i(o);if(!a||!o.hasTriggerModifier||!this._editor.hasModel()){e.clear();return}const{range:l,textElement:c}=a;if(!l.equalsRange(this._stickyRangeProjectedOnEditor))this._stickyRangeProjectedOnEditor=l,e.clear();else if(c.style.textDecoration==="underline")return;const d=new si;e.add(pe(()=>d.dispose(!0)));let h;Cx(this._languageFeaturesService.definitionProvider,this._editor.getModel(),new A(l.startLineNumber,l.startColumn+1),!1,d.token).then((u=>{if(!d.token.isCancellationRequested)if(u.length!==0){this._candidateDefinitionsLength=u.length;const g=c;h!==g?(e.clear(),h=g,h.style.textDecoration="underline",e.add(pe(()=>{h.style.textDecoration="none"}))):h||(h=g,h.style.textDecoration="underline",e.add(pe(()=>{h.style.textDecoration="none"})))}else e.clear()}))})),this._register(t.onCancel(()=>{e.clear()})),this._register(t.onExecute(async o=>{if(o.target.type!==12||o.target.detail!==this._stickyScrollWidget.getId())return;const r=this._stickyScrollWidget.getEditorPositionFromNode(o.target.element);r&&(!this._editor.hasModel()||!this._stickyRangeProjectedOnEditor||(this._candidateDefinitionsLength>1&&(this._focused&&this._disposeFocusStickyScrollStore(),this._revealPosition({lineNumber:r.lineNumber,column:1})),this._instaService.invokeFunction(Rse,o,this._editor,{uri:this._editor.getModel().uri,range:this._stickyRangeProjectedOnEditor})))}))}_onContextMenu(e,t){const i=new is(e,t);this._contextMenuService.showContextMenu({menuId:fe.StickyScrollContext,getAnchor:()=>i})}_onMouseMoveOrKeyDown(e){if(!e.shiftKey||!this._mouseTarget||!ki(this._mouseTarget))return;const t=this._stickyScrollWidget.getLineIndexFromChildDomNode(this._mouseTarget);t===null||this._showEndForLine===t||(this._showEndForLine=t,this._renderStickyScroll())}_toggleFoldingRegionForLine(e){if(!this._foldingModel||e===null)return;const t=this._stickyScrollWidget.getRenderedStickyLine(e),i=t?.foldingIcon;if(!i)return;WH(this._foldingModel,1,[e]),i.isCollapsed=!i.isCollapsed;const n=(i.isCollapsed?this._editor.getTopForLineNumber(i.foldingEndLine):this._editor.getTopForLineNumber(i.foldingStartLine))-this._editor.getOption(75)*t.index+1;this._editor.setScrollTop(n),this._renderStickyScroll(e)}_readConfiguration(){const e=this._editor.getOption(131);if(e.enabled===!1){this._editor.removeOverlayWidget(this._stickyScrollWidget),this._resetState(),this._sessionStore.clear(),this._enabled=!1;return}else e.enabled&&!this._enabled&&(this._editor.addOverlayWidget(this._stickyScrollWidget),this._sessionStore.add(this._editor.onDidScrollChange(i=>{i.scrollTopChanged&&(this._showEndForLine=void 0,this._renderStickyScroll())})),this._sessionStore.add(this._editor.onDidLayoutChange(()=>this._onDidResize())),this._sessionStore.add(this._editor.onDidChangeModelTokens(i=>this._onTokensChange(i))),this._sessionStore.add(this._stickyLineCandidateProvider.onDidChangeStickyScroll(()=>{this._showEndForLine=void 0,this._renderStickyScroll()})),this._enabled=!0);this._editor.getOption(76).renderType===2&&this._sessionStore.add(this._editor.onDidChangeCursorPosition(()=>{this._showEndForLine=void 0,this._renderStickyScroll(0)}))}_readConfigurationChange(e){(e.hasChanged(131)||e.hasChanged(81)||e.hasChanged(75)||e.hasChanged(126)||e.hasChanged(76))&&this._readConfiguration(),(e.hasChanged(76)||e.hasChanged(52)||e.hasChanged(126))&&this._renderStickyScroll(0)}_needsUpdate(e){const t=this._stickyScrollWidget.getCurrentLines();for(const i of t)for(const n of e.ranges)if(i>=n.fromLineNumber&&i<=n.toLineNumber)return!0;return!1}_onTokensChange(e){this._needsUpdate(e)&&this._renderStickyScroll(0)}_onDidResize(){const t=this._editor.getLayoutInfo().height/this._editor.getOption(75);this._maxStickyLines=Math.round(t*.25),this._renderStickyScroll(0)}async _renderStickyScroll(e){const t=this._editor.getModel();if(!t||t.isTooLargeForTokenization()){this._resetState();return}const i=this._updateAndGetMinRebuildFromLine(e),n=this._stickyLineCandidateProvider.getVersionId();if(n===void 0||n===t.getVersionId())if(!this._focused)await this._updateState(i);else if(this._focusedStickyElementIndex===-1)await this._updateState(i),this._focusedStickyElementIndex=this._stickyScrollWidget.lineNumberCount-1,this._focusedStickyElementIndex!==-1&&this._stickyScrollWidget.focusLineWithIndex(this._focusedStickyElementIndex);else{const r=this._stickyScrollWidget.lineNumbers[this._focusedStickyElementIndex];await this._updateState(i),this._stickyScrollWidget.lineNumberCount===0?this._focusedStickyElementIndex=-1:(this._stickyScrollWidget.lineNumbers.includes(r)||(this._focusedStickyElementIndex=this._stickyScrollWidget.lineNumberCount-1),this._stickyScrollWidget.focusLineWithIndex(this._focusedStickyElementIndex))}}_updateAndGetMinRebuildFromLine(e){if(e!==void 0){const t=this._minRebuildFromLine!==void 0?this._minRebuildFromLine:1/0;this._minRebuildFromLine=Math.min(e,t)}return this._minRebuildFromLine}async _updateState(e){this._minRebuildFromLine=void 0,this._foldingModel=await vg.get(this._editor)?.getFoldingModel()??void 0,this._widgetState=this.findScrollWidgetState();const t=this._widgetState.startLineNumbers.length>0;this._stickyScrollVisibleContextKey.set(t),this._stickyScrollWidget.setState(this._widgetState,this._foldingModel,e)}async _resetState(){this._minRebuildFromLine=void 0,this._foldingModel=void 0,this._widgetState=ry.Empty,this._stickyScrollVisibleContextKey.set(!1),this._stickyScrollWidget.setState(void 0,void 0)}findScrollWidgetState(){const e=Math.min(this._maxStickyLines,this._editor.getOption(131).maxLineCount),t=this._editor.getScrollTop();let i=0;const n=[],o=[],r=this._editor.getVisibleRanges();if(r.length!==0){const a=new ay(r[0].startLineNumber,r[r.length-1].endLineNumber),l=this._stickyLineCandidateProvider.getCandidateStickyLinesIntersecting(a);for(const c of l){const d=c.startLineNumber,h=c.endLineNumber,u=c.top,g=u+c.height,f=this._editor.getTopForLineNumber(d)-t,p=this._editor.getBottomForLineNumber(h)-t;if(u>f&&u<=p&&(n.push(d),o.push(h+1),g>p&&(i=p-g)),n.length===e)break}}return this._endLineNumbers=o,new ry(n,o,i,this._showEndForLine)}dispose(){super.dispose(),this._sessionStore.dispose()}},H3=em,em.ID="store.contrib.stickyScrollController",em);Ha=H3=DPe([ib(1,Nr),ib(2,ge),ib(3,_e),ib(4,ui),ib(5,ga),ib(6,Ae)],Ha);const cy=T("inlineEdit.originalBackground",Be(bu,.2),m(1220,"Background color for the original text in inline edits."),!0),xoe=T("inlineEdit.modifiedBackground",Be(qf,.3),m(1221,"Background color for the modified text in inline edits."),!0);T("inlineEdit.originalChangedLineBackground",Be(bu,.8),m(1222,"Background color for the changed lines in the original text of inline edits."),!0);const TPe=T("inlineEdit.originalChangedTextBackground",Be(bu,.8),m(1223,"Overlay color for the changed text in the original text of inline edits."),!0),NPe=T("inlineEdit.modifiedChangedLineBackground",{light:Be(Xx,.7),dark:Be(Xx,.7),hcDark:Xx,hcLight:Xx},m(1224,"Background color for the changed lines in the modified text of inline edits."),!0),RPe=T("inlineEdit.modifiedChangedTextBackground",Be(qf,.7),m(1225,"Overlay color for the changed text in the modified text of inline edits."),!0),APe=T("inlineEdit.gutterIndicator.primaryForeground",oA,m(1226,"Foreground color for the primary inline edit gutter indicator.")),Pb=T("inlineEdit.gutterIndicator.primaryBorder",jf,m(1227,"Border color for the primary inline edit gutter indicator.")),Loe=T("inlineEdit.gutterIndicator.primaryBackground",{light:Be(Pb,.5),dark:Be(Pb,.4),hcDark:Be(Pb,.4),hcLight:Be(Pb,.5)},m(1228,"Background color for the primary inline edit gutter indicator.")),MPe=T("inlineEdit.gutterIndicator.secondaryForeground",HJ,m(1229,"Foreground color for the secondary inline edit gutter indicator.")),koe=T("inlineEdit.gutterIndicator.secondaryBorder",qE,m(1230,"Border color for the secondary inline edit gutter indicator.")),Eoe=T("inlineEdit.gutterIndicator.secondaryBackground",koe,m(1231,"Background color for the secondary inline edit gutter indicator.")),OPe=T("inlineEdit.gutterIndicator.successfulForeground",oA,m(1232,"Foreground color for the successful inline edit gutter indicator.")),Ioe=T("inlineEdit.gutterIndicator.successfulBorder",jf,m(1233,"Border color for the successful inline edit gutter indicator.")),Doe=T("inlineEdit.gutterIndicator.successfulBackground",Ioe,m(1234,"Background color for the successful inline edit gutter indicator.")),PPe=T("inlineEdit.gutterIndicator.background",{hcDark:Be("tab.inactiveBackground",.5),hcLight:Be("tab.inactiveBackground",.5),dark:Be("tab.inactiveBackground",.5),light:"#5f5f5f18"},m(1235,"Background color for the inline edit gutter indicator.")),w1=T("inlineEdit.originalBorder",{light:bu,dark:bu,hcDark:bu,hcLight:bu},m(1236,"Border color for the original text in inline edits.")),v1=T("inlineEdit.modifiedBorder",{light:mo(qf,.6),dark:qf,hcDark:qf,hcLight:qf},m(1237,"Border color for the modified text in inline edits.")),FPe=T("inlineEdit.tabWillAcceptModifiedBorder",{light:mo(v1,0),dark:mo(v1,0),hcDark:mo(v1,0),hcLight:mo(v1,0)},m(1238,"Modified border color for the inline edits widget when tab will accept it.")),BPe=T("inlineEdit.tabWillAcceptOriginalBorder",{light:mo(w1,0),dark:mo(w1,0),hcDark:mo(w1,0),hcLight:mo(w1,0)},m(1239,"Original border color for the inline edits widget over the original text when tab will accept it."));function d0(s){return s.map(e=>e===us.Accept?FPe:v1)}function XA(s){return s.map(e=>e===us.Accept?BPe:w1)}function Br(s,e){let t;typeof s=="string"?t=bO(s,e):t=s.map((n,o)=>bO(n,e).read(o));const i=bO(Bi,e);return t.map((n,o)=>n.makeOpaque(i.read(o)))}function bO(s,e){return SQ({owner:{observeColor:s},equalsFn:(t,i)=>t.equals(i),debugName:()=>`observeColor(${s})`},e.onDidColorThemeChange,()=>{const t=e.getColorTheme().getColor(s);if(!t)throw new Ce(`Missing color: ${s}`);return t})}function Du(s,e,t){s.layoutInfo.read(t),s.value.read(t);const i=s.model.read(t);if(!i)return 0;let n=0;s.scrollTop.read(t);for(let r=e.startLineNumber;r<e.endLineNumberExclusive;r++){const a=i.getLineMaxColumn(r);let l=s.editor.getOffsetForColumn(r,a);if(l===-1){const c=s.editor.getOption(59).typicalHalfwidthCharacterWidth;l=a*c}n=Math.max(n,l)}const o=e.mapToLineArray(r=>i.getLineContent(r));return n<5&&o.some(r=>r.length>0)&&i.uri.scheme!=="file"&&console.error("unexpected width"),n}function WPe(s,e,t){return s.layoutInfo.read(t),s.value.read(t),s.model.read(t)?(s.scrollTop.read(t),s.editor.getOffsetForColumn(e.lineNumber,e.column)):0}function ZH(s,e,t,i,n=void 0){const o=i.getModel();if(!o)return{prefixTrim:0,prefixLeftOffset:0};const r=s.map(u=>u.isSingleLine()?u.startColumn-1:0),a=e.mapToLineArray(u=>VF(o.getLineContent(u))),l=t.filter(u=>u!=="").map(u=>VF(u)),c=Math.min(...r,...a,...l);let d;if(o.getLineIndentColumn(e.startLineNumber)>=c+1)di(i).scrollTop.read(n),d=i.getOffsetForColumn(e.startLineNumber,c+1);else if(t.length>0)d=XH(t[0].slice(0,c),i,o);else return{prefixTrim:0,prefixLeftOffset:0};return{prefixTrim:c,prefixLeftOffset:d}}function XH(s,e,t){const i=e.getOption(59).typicalHalfwidthCharacterWidth,n=t.getOptions().tabSize*i,o=s.split("	").length-1;return(s.length-o)*i+o*n}function QH(s){const e=s.layoutInfoContentLeft,t=j({name:"editor.validOverlay.width"},n=>{const o=s.layoutInfoMinimap.read(n).minimapLeft!==0,r=s.layoutInfoWidth.read(n)-e.read(n);if(o){const a=s.layoutInfoMinimap.read(n).minimapWidth+s.layoutInfoVerticalScrollbarWidth.read(n);return r-a}return r}),i=j({name:"editor.validOverlay.height"},n=>s.layoutInfoHeight.read(n)+s.contentHeight.read(n));return j({name:"editor.validOverlay"},n=>Ft.fromLeftTopWidthHeight(e.read(n),0,t.read(n),i.read(n)))}function HPe(s,e){const t=[];for(const i of s){const n=e.mapRange(i.modifiedRange);t.push(new Ks(i.originalRange,n))}return t}function wO(...s){return s.filter(e=>typeof e=="string").join(" ")}function VPe(s,e){return new k(e.lineNumber,e.column+s.start,e.lineNumber,e.column+s.endExclusive)}function zPe(s,e){let t=0;e:for(let i=0,n=s.length;i<n;i++)switch(s.charCodeAt(i)){case 9:t+=e;break;case 32:t++;break;default:break e}return t-t%e}function $Pe(s,e,t){let i=e-e%t,n=0;for(;n<s.length&&i!==0;n++)switch(s.charCodeAt(n)){case 9:i-=t;break;case 32:i--;break;default:throw new Ce("Unexpected character found while calculating indent length")}return n}function UPe(s,e,t){const i=qo(s),n=[],o=xee(e.mapToLineArray(r=>zPe(i[r-1],t)),Wo);return e.forEach(r=>{const a=$Pe(i[r-1],o,t);n.push(new Ui(VPe(new me(0,a),new A(r,1)),""))}),new Na(n)}class V3{constructor(){this._data=""}moveTo(e){return this._data+=`M ${e.x} ${e.y} `,this}lineTo(e){return this._data+=`L ${e.x} ${e.y} `,this}build(){return this._data}}function Wc(s){const e=Ih(void 0,(t,i)=>s.read(t)||i);return Jn({debugName:()=>`${s.debugName}.mapOutFalsy`},t=>{if(e.read(t),!!s.read(t))return e})}function Zr(s,e=dn.ofCaller()){return{left:j({name:"editor.validOverlay.left"},t=>s(t).left,e),top:j({name:"editor.validOverlay.top"},t=>s(t).top,e),width:j({name:"editor.validOverlay.width"},t=>s(t).right-s(t).left,e),height:j({name:"editor.validOverlay.height"},t=>s(t).bottom-s(t).top,e)}}var qPe=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},vO=function(s,e){return function(t,i){e(t,i,s)}};let z3=class{constructor(e,t,i,n,o,r){this._model=e,this._close=t,this._editorObs=i,this._contextKeyService=n,this._keybindingService=o,this._commandService=r,this._inlineEditsShowCollapsed=this._editorObs.getOption(71).map(a=>a.edits.showCollapsed)}toDisposableLiveElement(){return this._createHoverContent().toDisposableLiveElement()}_createHoverContent(){const e=Re("active",void 0),t=u=>({title:u.title,icon:u.icon,keybinding:typeof u.commandId=="string"?this._getKeybinding(u.commandArgs?void 0:u.commandId):j(this,g=>typeof u.commandId=="string"?void 0:this._getKeybinding(u.commandArgs?void 0:u.commandId.read(g)).read(g)),isActive:e.map(g=>g===u.id),onHoverChange:g=>e.set(g?u.id:void 0,void 0),onAction:()=>(this._close(!0),this._commandService.executeCommand(typeof u.commandId=="string"?u.commandId:u.commandId.get(),...u.commandArgs??[]))}),i=GPe(this._model.displayName),n=Zg(t({id:"gotoAndAccept",title:`${m(1212,"Go To")} / ${m(1213,"Accept")}`,icon:Z.check,commandId:jS})),o=Zg(t({id:"reject",title:m(1214,"Reject"),icon:Z.close,commandId:Sne})),r=this._model.extensionCommands.map((u,g)=>Zg(t({id:u.command.id+"_"+g,title:u.command.title,icon:u.icon??Z.symbolEvent,commandId:u.command.id,commandArgs:u.command.arguments}))),a=this._inlineEditsShowCollapsed.map(u=>Zg(t(u?{id:"showExpanded",title:m(1215,"Show Expanded"),icon:Z.expandAll,commandId:n5}:{id:"showCollapsed",title:m(1216,"Show Collapsed"),icon:Z.collapseAll,commandId:n5}))),l=Zg(t({id:"snooze",title:m(1217,"Snooze"),icon:Z.bellSlash,commandId:"editor.action.inlineSuggest.snooze"})),c=Zg(t({id:"settings",title:m(1218,"Settings"),icon:Z.gear,commandId:"workbench.action.openSettings",commandArgs:["@tag:nextEditSuggestions"]})),d=this._model.action?[this._model.action]:[],h=d.length>0?KPe(d.map(u=>({id:u.id,label:u.title+"...",enabled:!0,run:()=>this._commandService.executeCommand(u.id,...u.arguments??[]),class:void 0,tooltip:u.tooltip??u.title})),{hoverDelegate:lEe}):void 0;return jPe([i,n,o,a,r.length?hK():void 0,l,c,...r,h?hK():void 0,h])}_getKeybinding(e){return e?bt(this._contextKeyService.onDidChangeContext,()=>this._keybindingService.lookupKeybinding(e)):zt(void 0)}};z3=qPe([vO(3,Ae),vO(4,ft),vO(5,Gt)],z3);function jPe(s){return Ve.div({class:"content",style:{margin:4,minWidth:180}},s)}function GPe(s){return Ve.div({class:"header",style:{color:te(NJ),fontSize:"13px",fontWeight:"600",padding:"0 4px",lineHeight:28}},[s])}function Zg(s){return j({name:"inlineEdits.option"},e=>Ve.div({class:["monaco-menu-option",s.isActive?.map(t=>t&&"active")],onmouseenter:()=>s.onHoverChange?.(!0),onmouseleave:()=>s.onHoverChange?.(!1),onclick:s.onAction,onkeydown:t=>{t.key==="Enter"&&s.onAction?.()},tabIndex:0,style:{borderRadius:3}},[Ve.elem("span",{style:{fontSize:16,display:"flex"}},[xe.isThemeIcon(s.icon)?Dl(s.icon):s.icon.map(t=>Dl(t))]),Ve.elem("span",{},[s.title]),Ve.div({style:{marginLeft:"auto"},ref:t=>{const i=e.store.add(new hC(t,Fo,{disableTitle:!0,...yie,keybindingLabelShadow:void 0,keybindingLabelForeground:te(NJ),keybindingLabelBackground:"transparent",keybindingLabelBorder:"transparent",keybindingLabelBottomBorder:void 0}));e.store.add(De(n=>{i.set(s.keybinding.read(n))}))}})]))}function KPe(s,e){return j({name:"inlineEdits.actionBar"},t=>Ve.div({class:["action-widget-action-bar"],style:{padding:"3px 24px"}},[Ve.div({ref:i=>{t.store.add(new bo(i,e)).push(s,{icon:!1,label:!0})}})]))}function hK(){return Ve.div({id:"inline-edit-gutter-indicator-menu-separator",class:"menu-separator",style:{color:te(S_e),padding:"2px 0"}},Ve.div({style:{borderBottom:`1px solid ${te(RW)}`}}))}var YPe=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},GL=function(s,e){return function(t,i){e(t,i,s)}};let $3=class extends F{get model(){const e=this._model.get();if(!e)throw new Ce("Inline Edit Model not available");return e}constructor(e,t,i,n,o,r,a,l,c,d){super(),this._editorObs=e,this._originalRange=t,this._verticalOffset=i,this._model=n,this._isHoveringOverInlineEdit=o,this._focusIsInMenu=r,this._hoverService=a,this._instantiationService=l,this._accessibilityService=c,this._tabAction=j(this,h=>{const u=this._model.read(h);return u?u.tabAction.read(h):us.Inactive}),this._hoverVisible=Re(this,!1),this.isHoverVisible=this._hoverVisible,this._isHoveredOverIcon=Re(this,!1),this._isHoveredOverIconDebounced=$z(this._isHoveredOverIcon,100),this.isHoveredOverIcon=this._isHoveredOverIconDebounced,this._isHoveredOverInlineEditDebounced=$z(this._isHoveringOverInlineEdit,100),this._gutterIndicatorStyles=this._tabAction.map(this,(h,u)=>{switch(h){case us.Inactive:return{background:Br(Eoe,d).read(u).toString(),foreground:Br(MPe,d).read(u).toString(),border:Br(koe,d).read(u).toString()};case us.Jump:return{background:Br(Loe,d).read(u).toString(),foreground:Br(APe,d).read(u).toString(),border:Br(Pb,d).read(u).toString()};case us.Accept:return{background:Br(Doe,d).read(u).toString(),foreground:Br(OPe,d).read(u).toString(),border:Br(Ioe,d).read(u).toString()};default:W0()}}),this._originalRangeObs=Wc(this._originalRange),this._state=j(this,h=>{const u=this._originalRangeObs.read(h);if(u)return{range:u,lineOffsetRange:this._editorObs.observeLineOffsetRange(u,h.store)}}),this._stickyScrollController=Ha.get(this._editorObs.editor),this._stickyScrollHeight=this._stickyScrollController?bt(this._stickyScrollController.onDidChangeStickyScrollHeight,()=>this._stickyScrollController.stickyScrollWidgetHeight):zt(0),this._lineNumberToRender=j(this,h=>{if(this._verticalOffset.read(h)!==0)return"";const u=this._originalRange.read(h)?.startLineNumber,g=this._editorObs.getOption(76).read(h);if(u===void 0||g.renderType===0)return"";if(g.renderType===3){const f=this._editorObs.cursorPosition.read(h);return u%10===0||f&&f.lineNumber===u?u.toString():""}if(g.renderType===2){const f=this._editorObs.cursorPosition.read(h);if(!f)return"";const p=Math.abs(u-f.lineNumber);return p===0?u.toString():p.toString()}return g.renderType===4?g.renderFn?g.renderFn(u):"":u.toString()}),this._availableWidthForIcon=j(this,h=>{const u=this._editorObs.editor.getModel(),g=this._editorObs.editor,f=this._editorObs.layoutInfo.read(h),p=f.decorationsLeft+f.decorationsWidth-f.glyphMarginLeft;if(!u||p<=0)return()=>0;if(f.lineNumbersLeft===0)return()=>p;const _=this._editorObs.getOption(76).read(h);if(_.renderType===2||_.renderType===0)return()=>p;const b=g.getOption(59).typicalHalfwidthCharacterWidth,w=f.lineNumbersLeft+f.lineNumbersWidth,v=(u.getLineCount()+1).toString().length,y=[];for(let x=1;x<=v;x++){const L=10**(x-1),E=g.getTopForLineNumber(L),I=x*b,R=Math.min(p,Math.max(0,w-I-f.glyphMarginLeft));y.push({firstLineNumberWithDigitCount:L,topOfLineNumber:E,usableWidthLeftOfLineNumber:R})}return x=>{for(let L=y.length-1;L>=0;L--)if(x>=y[L].topOfLineNumber)return y[L].usableWidthLeftOfLineNumber;throw new Ce("Could not find avilable width for icon")}}),this._layout=j(this,h=>{const u=this._state.read(h);if(!u)return;const g=this._editorObs.layoutInfo.read(h),f=this._editorObs.observeLineHeightForLine(u.range.map(Ze=>Ze.startLineNumber)).read(h),p=2,_=g.decorationsLeft+g.decorationsWidth-g.glyphMarginLeft-2*p,b=g.height-2*p,w=Ft.fromLeftTopWidthHeight(p,p,_,b),C=w.withTop(this._stickyScrollHeight.read(h)),v=w.withTop(C.top+p),y=u.lineOffsetRange.read(h),x=Ft.fromRanges(me.fromTo(v.left,v.right),y),L=f,E=this._verticalOffset.read(h),I=x.withHeight(L).translateY(E),R=C.containsRect(I),O=this._tabAction.map(Ze=>Ze===us.Accept?Z.keyboardTab:Z.arrowRight),P=j(this,Ze=>{if(this._isHoveredOverIconDebounced.read(Ze)||this._isHoveredOverInlineEditDebounced.read(Ze))return Z.check;if(this._tabAction.read(Ze)===us.Accept)return Z.keyboardTab;const Xe=this._editorObs.cursorLineNumber.read(Ze)??0,At=u.range.read(Ze).startLineNumber;return Xe<=At?Z.keyboardTabAbove:Z.keyboardTabBelow}),B=22,$=16,z=Ze=>{const Xe=this._availableWidthForIcon.read(void 0)(Ze.bottom+this._editorObs.editor.getScrollTop())-p;return Math.max(Math.min(Xe,B),$)};if(R){const Ze=I;let Xe;g.lineNumbersWidth===0?Xe=Math.min(Math.max(g.lineNumbersLeft-w.left,0),Ze.width-B):Xe=Math.max(g.lineNumbersLeft+g.lineNumbersWidth-w.left,0);const At=Ze.withWidth(Xe),Et=Math.max(Math.min(g.decorationsWidth,B),$),un=Ze.withWidth(Et).translateX(Xe);return{gutterEditArea:x,icon:P,iconDirection:"right",iconRect:un,pillRect:Ze,lineNumberRect:At}}const Q=w.intersect(x);if(Q&&Q.height>=L){const Ze=I.moveToBeContainedIn(v).moveToBeContainedIn(Q),Xe=Ze.withWidth(z(Ze));return{gutterEditArea:x,icon:P,iconDirection:"right",iconRect:Xe,pillRect:Xe}}const ne=I.moveToBeContainedIn(w),de=ne.withWidth(z(ne)),ke=de,Me=de.top<I.top?"top":"bottom";return{gutterEditArea:x,icon:O,iconDirection:Me,iconRect:ke,pillRect:de}}),this._iconRef=Ve.ref(),this.isVisible=this._layout.map(h=>!!h),this._indicator=Ve.div({class:"inline-edits-view-gutter-indicator",onclick:()=>{const u=this._layout.get()?.icon.get()===Z.check;this._editorObs.editor.focus(),u?this.model.accept():this.model.jump()},tabIndex:0,style:{position:"absolute",overflow:"visible"}},Wc(this._layout).map(h=>h?[Ve.div({style:{position:"absolute",background:te(PPe),borderRadius:"4px",...Zr(u=>h.read(u).gutterEditArea)}}),Ve.div({class:"icon",ref:this._iconRef,onmouseenter:()=>{this._showHover()},style:{cursor:"pointer",zIndex:"20",position:"absolute",backgroundColor:this._gutterIndicatorStyles.map(u=>u.background),"--vscodeIconForeground":this._gutterIndicatorStyles.map(u=>u.foreground),border:this._gutterIndicatorStyles.map(u=>`1px solid ${u.border}`),boxSizing:"border-box",borderRadius:"4px",display:"flex",justifyContent:"flex-end",transition:"background-color 0.2s ease-in-out, width 0.2s ease-in-out",...Zr(u=>h.read(u).pillRect)}},[Ve.div({className:"line-number",style:{lineHeight:h.map(u=>u.lineNumberRect?u.lineNumberRect.height:0),display:h.map(u=>u.lineNumberRect?"flex":"none"),alignItems:"center",justifyContent:"flex-end",width:h.map(u=>u.lineNumberRect?u.lineNumberRect.width:0),height:"100%",color:this._gutterIndicatorStyles.map(u=>u.foreground)}},this._lineNumberToRender),Ve.div({style:{rotate:h.map(u=>`${ZPe(u.iconDirection)}deg`),transition:"rotate 0.2s ease-in-out",display:"flex",alignItems:"center",justifyContent:"center",height:"100%",marginRight:h.map(u=>u.pillRect.width-u.iconRect.width-(u.lineNumberRect?.width??0)),width:h.map(u=>u.iconRect.width)}},[h.map((u,g)=>Dl(u.icon.read(g)))])])]:[])).keepUpdated(this._store),this._register(this._editorObs.createOverlayWidget({domNode:this._indicator.element,position:zt(null),allowEditorOverflow:!1,minContentWidthInPx:zt(0)})),this._register(this._editorObs.editor.onMouseMove(h=>{if(this._state.get()===void 0)return;const f=this._iconRef.element.getBoundingClientRect(),p=Ft.fromLeftTopWidthHeight(f.left,f.top,f.width,f.height),_=new _n(h.event.posx,h.event.posy);this._isHoveredOverIcon.set(p.containsPoint(_),void 0)})),this._register(this._editorObs.editor.onDidScrollChange(()=>{this._isHoveredOverIcon.set(!1,void 0)})),this._register(yl(this._isHoveredOverInlineEditDebounced,h=>{h&&this.triggerAnimation()})),this._register(De(h=>{this._indicator.readEffect(h),this._indicator.element&&this._editorObs.editor.applyFontInfo(this._indicator.element)}))}triggerAnimation(){return this._accessibilityService.isMotionReduced()?new Animation(null,null).finished:this._iconRef.element.animate([{outline:`2px solid ${this._gutterIndicatorStyles.map(t=>t.border).get()}`,outlineOffset:"-1px",offset:0},{outline:"2px solid transparent",outlineOffset:"10px",offset:1}],{duration:500}).finished}_showHover(){if(this._hoverVisible.get())return;const e=new U,t=e.add(this._instantiationService.createInstance(z3,this.model,o=>{o&&this._editorObs.editor.focus(),n?.dispose()},this._editorObs).toDisposableLiveElement()),i=e.add(oa(t.element));e.add(i.onDidBlur(()=>this._focusIsInMenu.set(!1,void 0))),e.add(i.onDidFocus(()=>this._focusIsInMenu.set(!0,void 0))),e.add(pe(()=>this._focusIsInMenu.set(!1,void 0)));const n=this._hoverService.showInstantHover({target:this._iconRef.element,content:t.element});n?(this._hoverVisible.set(!0,void 0),e.add(this._editorObs.editor.onDidScrollChange(()=>n.dispose())),e.add(n.onDispose(()=>{this._hoverVisible.set(!1,void 0),e.dispose()}))):e.dispose()}};$3=YPe([GL(6,so),GL(7,_e),GL(8,Wn),GL(9,wi)],$3);function ZPe(s){switch(s){case"top":return 90;case"bottom":return-90;case"right":return 0}}var XPe=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},uK=function(s,e){return function(t,i){e(t,i,s)}},sr;(function(s){s.FirstTime="firstTime",s.SecondTime="secondTime",s.Active="active"})(sr||(sr={}));let U3=class extends F{constructor(e,t,i,n,o,r){super(),this._host=e,this._model=t,this._indicator=i,this._collapsedView=n,this._storageService=o,this._configurationService=r,this._disposables=this._register(new vt),this._setupDone=Re({name:"setupDone"},!1),this._activeCompletionId=j(a=>{const l=this._model.read(a);if(!l||!this._setupDone.read(a))return;const c=this._indicator.read(a);if(!(!c||!c.isVisible.read(a)))return l.inlineEdit.inlineCompletion.identity.id}),this._register(this._initializeDebugSetting()),this._disposables.value=this.setupNewUserExperience(),this._setupDone.set(!0,void 0)}setupNewUserExperience(){if(this.getNewUserType()===sr.Active)return;const e=new U;let t=!1,i=!1,n=0,o=0;return e.add(qve(this._activeCompletionId,async(r,a,l,c)=>{if(r===void 0)return;let d=this.getNewUserType();switch(d){case sr.FirstTime:{(n++>=5||t)&&(d=sr.SecondTime,this.setNewUserType(d));break}case sr.SecondTime:{o++>=3&&i&&(d=sr.Active,this.setNewUserType(d));break}}switch(d){case sr.FirstTime:{for(let h=0;h<3&&!c.isCancellationRequested;h++)await this._indicator.get()?.triggerAnimation(),await Tc(500);break}case sr.SecondTime:{this._indicator.get()?.triggerAnimation();break}}})),e.add(De(r=>{this._collapsedView.isVisible.read(r)&&this.getNewUserType()!==sr.Active&&this._collapsedView.triggerAnimation()})),e.add(ps((r,a)=>{const l=this._indicator.read(r);l&&a.add(yl(l.isHoveredOverIcon,async c=>{c&&(t=!0)}))})),e.add(ps((r,a)=>{const l=this._host.read(r);l&&a.add(l.onDidAccept(()=>{i=!0}))})),e}getNewUserType(){return this._storageService.get("inlineEditsGutterIndicatorUserKind",-1,sr.FirstTime)}setNewUserType(e){switch(e){case sr.FirstTime:throw new Ce("UserKind should not be set to first time");case sr.SecondTime:break;case sr.Active:this._disposables.clear();break}this._storageService.store("inlineEditsGutterIndicatorUserKind",e,-1,0)}_initializeDebugSetting(){const e="editor.inlineSuggest.edits.resetNewUserExperience";return this._configurationService.getValue(e)&&this._storageService.remove("inlineEditsGutterIndicatorUserKind",-1),this._configurationService.onDidChangeConfiguration(i=>{i.affectsConfiguration(e)&&this._configurationService.getValue(e)&&(this._storageService.remove("inlineEditsGutterIndicatorUserKind",-1),this._disposables.value=this.setupNewUserExperience())})}};U3=XPe([uK(4,Bs),uK(5,He)],U3);var QPe=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},JPe=function(s,e){return function(t,i){e(t,i,s)}};let q3=class extends F{constructor(e,t,i){super(),this._editor=e,this._edit=t,this._accessibilityService=i,this._onDidClick=this._register(new M),this.onDidClick=this._onDidClick.event,this._iconRef=Ve.ref(),this.isHovered=zt(!1),this._editorObs=di(this._editor);const o=this._edit.map(c=>c?.edit.replacements[0]??null).map(c=>c?jd(c,this._editor.getModel()).range.getStartPosition():null),r=this._editorObs.observePosition(o,this._store),a=j(c=>{const d=r.read(c);if(!d)return null;const h=this._editorObs.layoutInfoContentLeft.read(c),u=this._editorObs.scrollLeft.read(c);return new _n(h+d.x-u,d.y)}),l=Ve.div({class:"inline-edits-collapsed-view",style:{position:"absolute",overflow:"visible",top:"0px",left:"0px",display:"block"}},[[this.getCollapsedIndicator(a)]]).keepUpdated(this._store).element;this._register(this._editorObs.createOverlayWidget({domNode:l,position:zt(null),allowEditorOverflow:!1,minContentWidthInPx:zt(0)})),this.isVisible=this._edit.map((c,d)=>!!c&&a.read(d)!==null)}triggerAnimation(){return this._accessibilityService.isMotionReduced()?new Animation(null,null).finished:this._iconRef.element.animate([{offset:0,transform:"translateY(-3px)"},{offset:.2,transform:"translateY(1px)"},{offset:.36,transform:"translateY(-1px)"},{offset:.52,transform:"translateY(1px)"},{offset:.68,transform:"translateY(-1px)"},{offset:.84,transform:"translateY(1px)"},{offset:1,transform:"translateY(0px)"}],{duration:2e3}).finished}getCollapsedIndicator(e){const t=this._editorObs.layoutInfoContentLeft,i=e.map((o,r)=>o?o.deltaX(-t.read(r)):null),n=this.createIconPath(i);return Ve.svg({class:"collapsedView",ref:this._iconRef,style:{position:"absolute",...Zr(o=>QH(this._editorObs).read(o)),overflow:"hidden",pointerEvents:"none"}},[Ve.svgElem("path",{class:"collapsedViewPath",d:n,fill:te(Pb)})])}createIconPath(e){return e.map(o=>{if(!o)return new V3().build();const r=o.deltaX(-6/2).deltaY(-1),a=r.deltaX(6),l=r.deltaY(1),c=a.deltaY(1),d=l.deltaX(6/2).deltaY(3);return new V3().moveTo(r).lineTo(a).lineTo(c).lineTo(d).lineTo(l).lineTo(r).build()})}};q3=QPe([JPe(2,Wn)],q3);var eFe=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},gK=function(s,e){return function(t,i){e(t,i,s)}};const KL=14,CO=0,ad=4,YL=4,fK=2;let j3=class extends F{constructor(e,t,i,n,o){super(),this._editor=e,this._languageService=o,this._onDidClick=this._register(new M),this.onDidClick=this._onDidClick.event,this._isHovered=Re(this,!1),this.isHovered=this._isHovered,this._viewRef=Ve.ref(),this._editorObs=di(this._editor);const r=i.map((d,h)=>{let u;switch(d){case us.Inactive:u=Eoe;break;case us.Jump:u=Loe;break;case us.Accept:u=Doe;break}return{border:Br(u,n).read(h).toString(),background:te(Bi)}}),a=t.map(d=>d?this.getState(d):void 0),l=a.map(d=>d?this.getRendering(d,r):void 0);this.minEditorScrollHeight=j(this,d=>{const h=a.read(d);return h?h.rect.read(d).bottom+this._editor.getScrollTop():0});const c=Ve.div({class:"inline-edits-custom-view",style:{position:"absolute",overflow:"visible",top:"0px",left:"0px",display:"block"}},[l]).keepUpdated(this._store);this._register(this._editorObs.createOverlayWidget({domNode:c.element,position:zt(null),allowEditorOverflow:!1,minContentWidthInPx:Ih(this,(d,h)=>{const u=a.read(d);if(!u)return h??0;const g=u.rect.map(f=>f.right).read(d)+this._editorObs.layoutInfoVerticalScrollbarWidth.read(d)+ad-this._editorObs.layoutInfoContentLeft.read(d);return Math.max(h??0,g)}).recomputeInitiallyAndOnChange(this._store)})),this._register(De(d=>{if(!l.read(d)){this._isHovered.set(!1,void 0);return}this._isHovered.set(c.isHovered.read(d),void 0)}))}fitsInsideViewport(e,t,i){const n=this._editorObs.layoutInfoWidth.read(i),o=this._editorObs.layoutInfoContentLeft.read(i),r=this._editor.getLayoutInfo().verticalScrollbarWidth,a=this._editorObs.layoutInfoMinimap.read(i).minimapLeft!==0?this._editorObs.layoutInfoMinimap.read(i).minimapWidth:0,l=Du(this._editorObs,e,void 0),c=XH(t,this._editor,this._editor.getModel()),d=ad+KL;return l+c+d<n-o-r-a}getState(e){const t=j(this,a=>{const l=e.range.startLineNumber,c=e.range.endLineNumber,d=e.range.startColumn,h=e.range.endColumn,u=this._editor.getModel()?.getLineCount()??0,g=Du(this._editorObs,new Ne(l,l+1),a),f=l+1<=u?Du(this._editorObs,new Ne(l+1,l+2),a):void 0,p=l-1>=1?Du(this._editorObs,new Ne(l-1,l),a):void 0,_=this._editor.getOffsetForColumn(l,d),b=this._editor.getOffsetForColumn(c,h);return{lineWidth:g,lineWidthBelow:f,lineWidthAbove:p,startContentLeftOffset:_,endContentLeftOffset:b}}),i=e.range.startLineNumber,n=e.range.endLineNumber,o=this.fitsInsideViewport(new Ne(i,n+1),e.content,void 0);return{rect:j(this,a=>{const l=this._editorObs.getOption(59).read(a).typicalHalfwidthCharacterWidth,{lineWidth:c,lineWidthBelow:d,lineWidthAbove:h,startContentLeftOffset:u,endContentLeftOffset:g}=t.read(a),f=this._editorObs.layoutInfoContentLeft.read(a),p=this._editorObs.observeLineHeightForLine(i).recomputeInitiallyAndOnChange(a.store).read(a),_=this._editorObs.scrollTop.read(a),b=this._editorObs.scrollLeft.read(a);let w;i===n&&g+5*l>=c&&o?w="end":d!==void 0&&d+KL-YL-ad<u?w="below":h!==void 0&&h+KL-YL-ad<u?w="above":w="end";let C,v,y=0,x=0;switch(w){case"end":{C=this._editorObs.editor.getTopForLineNumber(i),v=c,y=ad+KL;break}case"below":{C=this._editorObs.editor.getTopForLineNumber(i+1),v=u,y=ad+YL,x=CO+fK;break}case"above":{C=this._editorObs.editor.getTopForLineNumber(i-1),v=u,y=ad+YL,x=-CO+fK;break}}return Ft.fromLeftTopWidthHeight(f+v-b,C-_,l*e.content.length,p).withMargin(CO,ad).translateX(y).translateY(x)}),label:e.content,kind:e.style}}getRendering(e,t){const i=document.createElement("div"),n=this._editor.getModel().tokenization.tokenizeLinesAt(1,[e.label])?.[0];let o;n&&e.kind===cF.Code?o=ih.fromLineTokens(n).toLineTokens(e.label,this._languageService.languageIdCodec):o=Ri.createEmpty(e.label,this._languageService.languageIdCodec);const r=hx(new ux([o]),ah.fromEditor(this._editor).withSetWidth(!1).withScrollBeyondLastColumn(0),[],i,!0);i.style.width=`${r.minWidthInPx}px`;const a=e.rect.map(l=>l.withMargin(0,ad));return Ve.div({class:"collapsedView",ref:this._viewRef,style:{position:"absolute",...Zr(l=>a.read(l)),overflow:"hidden",boxSizing:"border-box",cursor:"pointer",border:t.map(l=>`1px solid ${l.border}`),borderRadius:"4px",backgroundColor:t.map(l=>l.background),display:"flex",alignItems:"center",justifyContent:"center",whiteSpace:"nowrap"},onmousedown:l=>{l.preventDefault()},onclick:l=>{this._onDidClick.fire(new is(be(l),l))}},[i])}};j3=eFe([gK(3,wi),gK(4,Ei)],j3);const tFe=0,iFe=0,yO=1,nFe=1,sFe=3,SO=4;class oFe extends F{constructor(e,t,i,n){super(),this._editor=e,this._edit=t,this._uiState=i,this._tabAction=n,this._onDidClick=this._register(new M),this.onDidClick=this._onDidClick.event,this._display=j(this,a=>this._uiState.read(a)?"block":"none"),this._editorMaxContentWidthInRange=j(this,a=>{const l=this._originalDisplayRange.read(a);return l?(this._editorObs.versionId.read(a),Ih(this,(c,d)=>{const h=Du(this._editorObs,l,c);return Math.max(h,d??0)})):zt(0)}).map((a,l)=>a.read(l)),this._maxPrefixTrim=j(this,a=>{const l=this._uiState.read(a);return l?ZH(l.deletions,l.originalRange,[],this._editor):{prefixTrim:0,prefixLeftOffset:0}}),this._editorLayoutInfo=j(this,a=>{const l=this._edit.read(a);if(!l||!this._uiState.read(a))return null;const d=this._editorObs.layoutInfo.read(a),h=this._editorObs.scrollLeft.read(a),u=this._editorObs.getOption(59).map(C=>C.typicalHalfwidthCharacterWidth).read(a),g=d.contentLeft+Math.max(this._editorMaxContentWidthInRange.read(a),u)-h,f=l.originalLineRange,p=this._originalVerticalStartPosition.read(a)??this._editor.getTopForLineNumber(f.startLineNumber)-this._editorObs.scrollTop.read(a),_=this._originalVerticalEndPosition.read(a)??this._editor.getTopForLineNumber(f.endLineNumberExclusive)-this._editorObs.scrollTop.read(a),b=d.contentLeft+this._maxPrefixTrim.read(a).prefixLeftOffset-h;return g<=b?null:{codeRect:Ft.fromLeftTopRightBottom(b,p,g,_).withMargin(iFe,tFe),contentLeft:d.contentLeft}}).recomputeInitiallyAndOnChange(this._store),this._originalOverlay=Ve.div({style:{pointerEvents:"none"}},j(this,a=>{const l=Wc(this._editorLayoutInfo).read(a);if(!l)return;const c=l.map(g=>Ft.fromLeftTopRightBottom(g.contentLeft-SO-yO,g.codeRect.top,g.contentLeft,g.codeRect.bottom)),d=j(this,g=>{const f=l.read(g).codeRect,p=c.read(g);return f.intersectHorizontal(new me(p.left,Number.MAX_SAFE_INTEGER))}),h=this._uiState.map(g=>g?.inDiffEditor?sFe:nFe).read(a),u=d.map(g=>g.withMargin(h,h));return[Ve.div({class:"originalSeparatorDeletion",style:{...u.read(a).toStyles(),borderRadius:`${SO}px`,border:`${yO+h}px solid ${te(Bi)}`,boxSizing:"border-box"}}),Ve.div({class:"originalOverlayDeletion",style:{...d.read(a).toStyles(),borderRadius:`${SO}px`,border:XA(this._tabAction).map(g=>`${yO}px solid ${te(g)}`),boxSizing:"border-box",backgroundColor:te(cy)}}),Ve.div({class:"originalOverlayHiderDeletion",style:{...c.read(a).toStyles(),backgroundColor:te(Bi)}})]})).keepUpdated(this._store),this._nonOverflowView=Ve.div({class:"inline-edits-view",style:{position:"absolute",overflow:"visible",top:"0px",left:"0px",display:this._display}},[[this._originalOverlay]]).keepUpdated(this._store),this.isHovered=zt(!1),this._editorObs=di(this._editor);const o=j(this,a=>{const l=this._edit.read(a);return l?new A(l.originalLineRange.startLineNumber,1):null}),r=j(this,a=>{const l=this._edit.read(a);return l?new A(l.originalLineRange.endLineNumberExclusive,1):null});this._originalDisplayRange=this._uiState.map(a=>a?.originalRange),this._originalVerticalStartPosition=this._editorObs.observePosition(o,this._store).map(a=>a?.y),this._originalVerticalEndPosition=this._editorObs.observePosition(r,this._store).map(a=>a?.y),this._register(this._editorObs.createOverlayWidget({domNode:this._nonOverflowView.element,position:zt(null),allowEditorOverflow:!1,minContentWidthInPx:j(this,a=>{const l=this._editorLayoutInfo.read(a);return l===null?0:l.codeRect.width})}))}}var rFe=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},pK=function(s,e){return function(t,i){e(t,i,s)}};const YC=1,aFe=1,lFe=3,xO=4;let G3=class extends F{constructor(e,t,i,n,o){super(),this._editor=e,this._input=t,this._tabAction=i,this._languageService=o,this._onDidClick=this._register(new M),this.onDidClick=this._onDidClick.event,this._state=j(this,r=>{const a=this._input.read(r);if(!a)return;const l=this._editor.getModel(),c=l.getEOL();if(a.startColumn===1&&a.lineNumber>1&&l.getLineLength(a.lineNumber)!==0&&a.text.endsWith(c)&&!a.text.startsWith(c)){const d=l.getLineLength(a.lineNumber-1)+1;return{lineNumber:a.lineNumber-1,column:d,text:c+a.text.slice(0,-c.length)}}return{lineNumber:a.lineNumber,column:a.startColumn,text:a.text}}),this._trimVertically=j(this,r=>{const a=this._state.read(r),l=a?.text;if(!l||l.trim()==="")return{topOffset:0,bottomOffset:0,linesTop:0,linesBottom:0};const c=this._editor.getLineHeightForPosition(new A(a.lineNumber,1)),d=this._editor.getModel().getEOL();let h=0,u=0,g=0;for(;g<l.length&&l.startsWith(d,g);g+=d.length)h+=1;for(let f=l.length;f>g&&l.endsWith(d,f);f-=d.length)u+=1;return{topOffset:h*c,bottomOffset:u*c,linesTop:h,linesBottom:u}}),this._maxPrefixTrim=j(this,r=>{const a=this._state.read(r);if(!a)return{prefixLeftOffset:0,prefixTrim:0};const l=this._editor.getModel(),c=l.getEOL(),d=this._trimVertically.read(r),h=a.text.split(c),u=h.slice(d.linesTop,h.length-d.linesBottom);d.linesTop===0&&(u[0]=l.getLineContent(a.lineNumber)+u[0]);const g=new Ne(a.lineNumber,a.lineNumber+(d.linesTop>0?0:1));return ZH([],g,u,this._editor)}),this._ghostText=j(r=>{const a=this._state.read(r),l=this._maxPrefixTrim.read(r);if(!a)return;const d=this._editor.getModel().getEOL(),u=a.text.split(d).map((g,f)=>new fp(new k(f+1,f===0?1:l.prefixTrim+1,f+1,g.length+1),"modified-background",0));return new r0(a.lineNumber,[new AD(a.column,a.text,!1,u)])}),this._display=j(this,r=>this._state.read(r)?"block":"none"),this._editorMaxContentWidthInRange=j(this,r=>{const a=this._state.read(r);if(!a)return 0;this._editorObs.versionId.read(r);const l=this._editor.getModel(),c=l.getEOL(),d=a.text.startsWith(c)?"":l.getValueInRange(new k(a.lineNumber,1,a.lineNumber,a.column)),h=l.getValueInRange(new k(a.lineNumber,a.column,a.lineNumber,l.getLineLength(a.lineNumber)+1)),g=(d+a.text+h).split(c),f=ah.fromEditor(this._editor).withSetWidth(!1).withScrollBeyondLastColumn(0),p=g.map(_=>{const b=l.tokenization.tokenizeLinesAt(a.lineNumber,[_])?.[0];let w;return b?w=ih.fromLineTokens(b).toLineTokens(_,this._languageService.languageIdCodec):w=Ri.createEmpty(_,this._languageService.languageIdCodec),hx(new ux([w]),f,[],ie("div"),!0).minWidthInPx});return Math.max(...p)}),this.startLineOffset=this._trimVertically.map(r=>r.topOffset),this.originalLines=this._state.map(r=>r?new Ne(r.lineNumber,Math.min(r.lineNumber+2,this._editor.getModel().getLineCount()+1)):void 0),this._overlayLayout=j(this,r=>{this._ghostText.read(r);const a=this._state.read(r);if(!a)return null;this._editorObs.observePosition(Re(this,new A(a.lineNumber,a.column)),r.store).read(r);const l=this._editorObs.layoutInfo.read(r),c=this._editorObs.scrollLeft.read(r),d=this._editorObs.layoutInfoVerticalScrollbarWidth.read(r),h=l.contentLeft+this._editorMaxContentWidthInRange.read(r)-c,u=this._maxPrefixTrim.read(r).prefixLeftOffset??0,g=l.contentLeft+u-c;if(h<=g)return null;const{topOffset:f,bottomOffset:p}=this._trimVertically.read(r),_=this._editorObs.scrollTop.read(r),b=this._ghostTextView.height.read(r)-f-p,w=this._editor.getTopForLineNumber(a.lineNumber)-_+f,C=w+b,v=new Ft(g,w,h,C);return{overlay:v,startsAtContentLeft:u===0,contentLeft:l.contentLeft,minContentWidthRequired:u+v.width+d}}).recomputeInitiallyAndOnChange(this._store),this._modifiedOverlay=Ve.div({style:{pointerEvents:"none"}},j(this,r=>{const a=Wc(this._overlayLayout).read(r);if(!a)return;const l=a.map(u=>Ft.fromLeftTopRightBottom(u.contentLeft-xO-YC,u.overlay.top,u.contentLeft,u.overlay.bottom)).read(r),c=this._input.map(u=>u?.inDiffEditor?lFe:aFe).read(r),d=a.map(u=>u.overlay.withMargin(0,YC,0,u.startsAtContentLeft?0:YC).intersectHorizontal(new me(l.left,Number.MAX_SAFE_INTEGER))),h=d.map(u=>u.withMargin(c,c));return[Ve.div({class:"originalUnderlayInsertion",style:{...h.read(r).toStyles(),borderRadius:xO,border:`${YC+c}px solid ${te(Bi)}`,boxSizing:"border-box"}}),Ve.div({class:"originalOverlayInsertion",style:{...d.read(r).toStyles(),borderRadius:xO,border:d0(this._tabAction).map(u=>`${YC}px solid ${te(u)}`),boxSizing:"border-box",backgroundColor:te(xoe)}}),Ve.div({class:"originalOverlayHiderInsertion",style:{...l.toStyles(),backgroundColor:te(Bi)}})]})).keepUpdated(this._store),this._view=Ve.div({class:"inline-edits-view",style:{position:"absolute",overflow:"visible",top:"0px",left:"0px",display:this._display}},[[this._modifiedOverlay]]).keepUpdated(this._store),this._editorObs=di(this._editor),this._ghostTextView=this._register(n.createInstance(l0,this._editor,{ghostText:this._ghostText,minReservedLineCount:zt(0),targetTextModel:this._editorObs.model.map(r=>r??void 0),warning:zt(void 0),handleInlineCompletionShown:zt(()=>{})},Re(this,{syntaxHighlightingEnabled:!0,extraClasses:["inline-edit"]}),!0,!0)),this.isHovered=this._ghostTextView.isHovered,this._register(this._ghostTextView.onDidClick(r=>{this._onDidClick.fire(r)})),this._register(this._editorObs.createOverlayWidget({domNode:this._view.element,position:zt(null),allowEditorOverflow:!1,minContentWidthInPx:j(this,r=>{const a=this._overlayLayout.read(r);return a===null?0:a.minContentWidthRequired})}))}};G3=rFe([pK(3,_e),pK(4,Ei)],G3);var cFe=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},mK=function(s,e){return function(t,i){e(t,i,s)}};let K3=class extends F{constructor(e,t,i,n,o,r){super(),this._editor=e,this._edit=t,this._isInDiffEditor=i,this._tabAction=n,this._languageService=o,this._themeService=r,this._onDidClick=this._register(new M),this.onDidClick=this._onDidClick.event,this._maxPrefixTrim=this._edit.map((a,l)=>a?ZH(a.replacements.flatMap(c=>[c.originalRange,c.modifiedRange]),a.originalRange,a.modifiedLines,this._editor.editor,l):void 0),this._modifiedLineElements=j(this,a=>{const l=[];let c=0;const d=this._maxPrefixTrim.read(a),h=this._edit.read(a);if(!h||!d)return;const u=d.prefixTrim,g=dFe(h.replacements.map(_=>_.modifiedRange)).map(_=>new k(_.startLineNumber,_.startColumn-u,_.endLineNumber,_.endColumn-u)),f=this._editor.model.get(),p=h.modifiedRange.startLineNumber;for(let _=0;_<h.modifiedRange.length;_++){const b=document.createElement("div"),w=p+_,C=h.modifiedLines[_].slice(u),v=f.tokenization.tokenizeLinesAt(w,[C])?.[0];let y;v?y=ih.fromLineTokens(v).toLineTokens(C,this._languageService.languageIdCodec):y=Ri.createEmpty(C,this._languageService.languageIdCodec);const x=[];for(const E of g.filter(I=>I.startLineNumber===w)){const I=Math.min(E.endColumn,C.length+1);x.push(new fp(new k(1,E.startColumn,1,I),"inlineCompletions-modified-bubble",0))}const L=hx(new ux([y]),ah.fromEditor(this._editor.editor).withSetWidth(!1).withScrollBeyondLastColumn(0),x,b,!0);this._editor.getOption(59).read(a),c=Math.max(c,L.minWidthInPx),l.push(b)}return{lines:l,requiredWidth:c}}),this._layout=j(this,a=>{const l=this._modifiedLineElements.read(a),c=this._maxPrefixTrim.read(a),d=this._edit.read(a);if(!l||!c||!d)return;const{prefixLeftOffset:h}=c,{requiredWidth:u}=l,g=this._editor.observeLineHeightsForLineRange(d.originalRange).read(a),f=(()=>{const Q=g.slice(0,d.modifiedRange.length);for(;Q.length<d.modifiedRange.length;)Q.push(g[g.length-1]);return Q})(),p=this._editor.layoutInfoContentLeft.read(a),_=this._editor.layoutInfoVerticalScrollbarWidth.read(a),b=this._editor.scrollLeft.read(a),w=this._editor.scrollTop.read(a),C=p-b,v=this._editor.editor.getModel(),y=d.originalRange.mapToLineArray(Q=>this._editor.editor.getOffsetForColumn(Q,v.getLineMaxColumn(Q))-h),x=Math.max(...y,u),L=d.originalRange.startLineNumber,E=d.originalRange.endLineNumberExclusive-1,I=this._editor.editor.getTopForLineNumber(L)-w,R=this._editor.editor.getBottomForLineNumber(E)-w,O=Ft.fromLeftTopWidthHeight(C+h,I,x,R-I),P=Ft.fromLeftTopWidthHeight(O.left,O.bottom,O.width,f.reduce((Q,J)=>Q+J,0)),B=Ft.hull([O,P]),$=B.intersectVertical(new me(O.bottom,Number.MAX_SAFE_INTEGER)),z=new Ft($.left,$.top,$.right,$.bottom);return{originalLinesOverlay:O,modifiedLinesOverlay:P,background:B,lowerBackground:$,lowerText:z,modifiedLineHeights:f,minContentWidthRequired:h+x+_}}),this._viewZoneInfo=j(a=>{if(!this._editor.getOption(71).map(g=>g.edits.allowCodeShifting==="always").read(a))return;const c=this._layout.read(a),d=this._edit.read(a);if(!c||!d)return;const h=c.lowerBackground.height,u=d.originalRange.endLineNumberExclusive;return{height:h,lineNumber:u}}),this.minEditorScrollHeight=j(this,a=>{const l=Wc(this._layout).read(a);return!l||this._viewZoneInfo.read(a)!==void 0?0:l.read(a).lowerText.bottom+this._editor.editor.getScrollTop()}),this._div=Ve.div({class:"line-replacement"},[j(this,a=>{const l=Wc(this._layout).read(a),c=this._modifiedLineElements.read(a);if(!l||!c)return[];const d=l.read(a),h=this._editor.layoutInfoContentLeft.read(a),u=this._isInDiffEditor.read(a)?3:1;c.lines.forEach((p,_)=>{p.style.width=`${d.lowerText.width}px`,p.style.height=`${d.modifiedLineHeights[_]}px`,p.style.position="relative"});const g=d0(this._tabAction).read(a),f=XA(this._tabAction).read(a);return[Ve.div({style:{position:"absolute",...Zr(p=>QH(this._editor).read(p)),overflow:"hidden",pointerEvents:"none"}},[Ve.div({class:"borderAroundLineReplacement",style:{position:"absolute",...Zr(p=>l.read(p).background.translateX(-h).withMargin(u)),borderRadius:"4px",border:`${u+1}px solid ${te(Bi)}`,boxSizing:"border-box",pointerEvents:"none"}}),Ve.div({class:"originalOverlayLineReplacement",style:{position:"absolute",...Zr(p=>l.read(p).background.translateX(-h)),borderRadius:"4px",border:Br(f,this._themeService).map(p=>`1px solid ${p.toString()}`),pointerEvents:"none",boxSizing:"border-box",background:te(cy)}}),Ve.div({class:"modifiedOverlayLineReplacement",style:{position:"absolute",...Zr(p=>l.read(p).lowerBackground.translateX(-h)),borderRadius:"0 0 4px 4px",background:te(Bi),boxShadow:`${te(eA)} 0 6px 6px -6px`,border:`1px solid ${te(g)}`,boxSizing:"border-box",overflow:"hidden",cursor:"pointer",pointerEvents:"auto"},onmousedown:p=>{p.preventDefault()},onclick:p=>this._onDidClick.fire(new is(be(p),p))},[Ve.div({style:{position:"absolute",top:0,left:0,width:"100%",height:"100%",background:te(NPe)}})]),Ve.div({class:"modifiedLinesLineReplacement",style:{position:"absolute",boxSizing:"border-box",...Zr(p=>l.read(p).lowerText.translateX(-h)),fontFamily:this._editor.getOption(58),fontSize:this._editor.getOption(61),fontWeight:this._editor.getOption(62),pointerEvents:"none",whiteSpace:"nowrap",borderRadius:"0 0 4px 4px",overflow:"hidden"}},[...c.lines])])]})]).keepUpdated(this._store),this.isHovered=this._editor.isTargetHovered(a=>this._isMouseOverWidget(a),this._store),this._previousViewZoneInfo=void 0,this._register(pe(()=>this._editor.editor.changeViewZones(a=>this.removePreviousViewZone(a)))),this._register(Vhe(this._viewZoneInfo,({lastValue:a,newValue:l})=>{a===l||a?.height===l?.height&&a?.lineNumber===l?.lineNumber||this._editor.editor.changeViewZones(c=>{this.removePreviousViewZone(c),l&&this.addViewZone(l,c)})})),this._register(this._editor.createOverlayWidget({domNode:this._div.element,minContentWidthInPx:j(this,a=>this._layout.read(a)?.minContentWidthRequired??0),position:zt({preference:{top:0,left:0}}),allowEditorOverflow:!1}))}_isMouseOverWidget(e){const t=this._layout.get();return!t||!(e.event instanceof bh)?!1:t.lowerBackground.containsPoint(new _n(e.event.relativePos.x,e.event.relativePos.y))}removePreviousViewZone(e){if(!this._previousViewZoneInfo)return;e.removeZone(this._previousViewZoneInfo.id);const t=this._editor.cursorLineNumber.get();t!==null&&t>=this._previousViewZoneInfo.lineNumber&&this._editor.editor.setScrollTop(this._editor.scrollTop.get()-this._previousViewZoneInfo.height),this._previousViewZoneInfo=void 0}addViewZone(e,t){const i=t.addZone({afterLineNumber:e.lineNumber-1,heightInPx:e.height,domNode:ie("div")});this._previousViewZoneInfo={height:e.height,lineNumber:e.lineNumber,id:i};const n=this._editor.cursorLineNumber.get();n!==null&&n>=e.lineNumber&&this._editor.editor.setScrollTop(this._editor.scrollTop.get()+e.height)}};K3=cFe([mK(4,Ei),mK(5,wi)],K3);function dFe(s){const e=[];for(;s.length;){let t=s.shift();t.startLineNumber!==t.endLineNumber&&(s.push(new k(t.startLineNumber+1,1,t.endLineNumber,t.endColumn)),t=new k(t.startLineNumber,t.startColumn,t.startLineNumber,Number.MAX_SAFE_INTEGER)),e.push(t)}return e}var hFe=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},_K=function(s,e){return function(t,i){e(t,i,s)}};const nb=0,ZC=0,Ga=1,uFe=1,gFe=3,ld=4,LO=20,bK=12;let UD=class extends F{static fitsInsideViewport(e,t,i,n){const o=di(e),r=o.layoutInfoWidth.read(n),a=o.layoutInfoContentLeft.read(n),l=e.getLayoutInfo().verticalScrollbarWidth,c=o.layoutInfoMinimap.read(n).minimapLeft!==0?o.layoutInfoMinimap.read(n).minimapWidth:0,d=Du(o,i.displayRange,void 0),h=i.lineEdit.newLines.reduce((f,p)=>Math.max(f,XH(p,e,t)),0),u=LO,g=bK+2*Ga;return d+h+u+g<r-a-l-c}constructor(e,t,i,n,o,r,a){super(),this._editor=e,this._edit=t,this._previewTextModel=i,this._uiState=n,this._tabAction=o,this._instantiationService=r,this._themeService=a,this._editorObs=di(this._editor),this._onDidClick=this._register(new M),this.onDidClick=this._onDidClick.event,this._display=j(this,c=>this._uiState.read(c)?"block":"none"),this.previewRef=Ve.ref();const l=this._uiState.map(c=>c?.isInDiffEditor?gFe:uFe);this._editorContainer=Ve.div({class:["editorContainer"],style:{position:"absolute",overflow:"hidden",cursor:"pointer"},onmousedown:c=>{c.preventDefault()},onclick:c=>{this._onDidClick.fire(new is(be(c),c))}},[Ve.div({class:"preview",style:{pointerEvents:"none"},ref:this.previewRef})]).keepUpdated(this._store),this.isHovered=this._editorContainer.didMouseMoveDuringHover,this.previewEditor=this._register(this._instantiationService.createInstance(Lh,this.previewRef.element,{glyphMargin:!1,lineNumbers:"off",minimap:{enabled:!1},guides:{indentation:!1,bracketPairs:!1,bracketPairsHorizontal:!1,highlightActiveIndentation:!1},rulers:[],padding:{top:0,bottom:0},folding:!1,selectOnLineNumbers:!1,selectionHighlight:!1,columnSelection:!1,overviewRulerBorder:!1,overviewRulerLanes:0,lineDecorationsWidth:0,lineNumbersMinChars:0,revealHorizontalRightPadding:0,bracketPairColorization:{enabled:!0,independentColorPoolPerBracketType:!1},scrollBeyondLastLine:!1,scrollbar:{vertical:"hidden",horizontal:"hidden",handleMouseWheel:!1},readOnly:!0,wordWrap:"off",wordWrapOverride1:"off",wordWrapOverride2:"off"},{contextKeyValues:{[Ot.inInlineEditsPreviewEditor.key]:!0},contributions:[]},this._editor)),this._previewEditorObs=di(this.previewEditor),this._activeViewZones=[],this._updatePreviewEditor=j(this,c=>{this._editorContainer.readEffect(c),this._previewEditorObs.model.read(c),this._display.read(c),this._nonOverflowView&&(this._nonOverflowView.element.style.display=this._display.read(c));const d=this._uiState.read(c),h=this._edit.read(c);if(!d||!h)return;const u=h.originalLineRange,g=[];u.startLineNumber>1&&g.push(new k(1,1,u.startLineNumber-1,1)),u.startLineNumber+d.newTextLineCount<this._previewTextModel.getLineCount()+1&&g.push(new k(u.startLineNumber+d.newTextLineCount,1,this._previewTextModel.getLineCount()+1,1)),this.previewEditor.setHiddenAreas(g,void 0,!0);const f=[...this._activeViewZones];this._activeViewZones=[];const p=u.endLineNumberExclusive-u.startLineNumber-d.newTextLineCount;this.previewEditor.changeViewZones(_=>{f.forEach(b=>_.removeZone(b)),p>0&&this._activeViewZones.push(_.addZone({afterLineNumber:u.startLineNumber+d.newTextLineCount-1,heightInLines:p,showInHiddenAreas:!0,domNode:ie("div.diagonal-fill.inline-edits-view-zone")}))})}),this._previewEditorWidth=j(this,c=>{const d=this._edit.read(c);return d?(this._updatePreviewEditor.read(c),Du(this._previewEditorObs,d.modifiedLineRange,c)):0}),this._cursorPosIfTouchesEdit=j(this,c=>{const d=this._editorObs.cursorPosition.read(c),h=this._edit.read(c);if(!(!h||!d))return h.modifiedLineRange.contains(d.lineNumber)?d:void 0}),this._originalStartPosition=j(this,c=>{const d=this._edit.read(c);return d?new A(d.originalLineRange.startLineNumber,1):null}),this._originalEndPosition=j(this,c=>{const d=this._edit.read(c);return d?new A(d.originalLineRange.endLineNumberExclusive,1):null}),this._originalVerticalStartPosition=this._editorObs.observePosition(this._originalStartPosition,this._store).map(c=>c?.y),this._originalVerticalEndPosition=this._editorObs.observePosition(this._originalEndPosition,this._store).map(c=>c?.y),this._originalDisplayRange=this._edit.map(c=>c?.displayRange),this._editorMaxContentWidthInRange=j(this,c=>{const d=this._originalDisplayRange.read(c);return d?(this._editorObs.versionId.read(c),Ih(this,(h,u)=>{const g=Du(this._editorObs,d,h);return Math.max(g,u??0)})):zt(0)}).map((c,d)=>c.read(d)),this._previewEditorLayoutInfo=j(this,c=>{const d=this._edit.read(c);if(!d||!this._uiState.read(c))return null;const u=d.originalLineRange,g=this._editorObs.scrollLeft.read(c),f=this._editorMaxContentWidthInRange.read(c),p=this._editorObs.layoutInfo.read(c),_=this._previewEditorWidth.read(c),b=p.contentWidth-p.verticalScrollbarWidth,w=this._editor.getContainerDomNode().getBoundingClientRect(),C=p.contentLeft+p.contentWidth+w.left,v=be(this._editor.getContainerDomNode()).innerWidth-C,y=be(this._editor.getContainerDomNode()).innerWidth-w.right,x=Math.min(p.contentWidth*.3,_,100),L=0,E=L+v,I=this._cursorPosIfTouchesEdit.read(c),R=Math.max(b+g-L-Math.max(0,x-E),Math.min(I?WPe(this._editorObs,I,c)+50:0,b+g)),O=Math.min(f+LO,R),P=f+LO+_+70,B=R-O;let $,z;O>g?($=0,z=p.contentLeft+O-g):($=g-O,z=p.contentLeft);const Q=this._originalVerticalStartPosition.read(c)??this._editor.getTopForLineNumber(u.startLineNumber)-this._editorObs.scrollTop.read(c),J=this._originalVerticalEndPosition.read(c)??this._editor.getBottomForLineNumber(u.endLineNumberExclusive-1)-this._editorObs.scrollTop.read(c),ne=p.contentLeft-g;let de=Ft.fromLeftTopRightBottom(ne,Q,z,J);const ke=de.height===0;ke||(de=de.withMargin(ZC,nb));const Ze=this._previewEditorObs.observeLineHeightsForLineRange(d.modifiedLineRange).read(c).reduce((wC,Rx)=>wC+Rx,0),Xe=J-Q,At=Math.max(Xe,Ze),Et=B===0,un=0,Vi=Math.min(_+bK,y+p.width-p.contentLeft-un);let Ws=Ft.fromLeftTopWidthHeight(de.right+un,Q,Vi,At);return ke?Ws=Ws.withMargin(ZC,nb).translateY(ZC):Ws=Ws.withMargin(ZC,nb).translateX(nb+Ga),{codeRect:de,editRect:Ws,codeScrollLeft:g,contentLeft:p.contentLeft,isInsertion:ke,maxContentWidth:P,shouldShowShadow:Et,desiredPreviewEditorScrollLeft:$,previewEditorWidth:Vi}}),this._stickyScrollController=Ha.get(this._editorObs.editor),this._stickyScrollHeight=this._stickyScrollController?bt(this._stickyScrollController.onDidChangeStickyScrollHeight,()=>this._stickyScrollController.stickyScrollWidgetHeight):zt(0),this._shouldOverflow=j(this,c=>!1),this._originalBackgroundColor=bt(this,this._themeService.onDidColorThemeChange,()=>this._themeService.getColorTheme().getColor(cy)??q.transparent),this._backgroundSvg=Ve.svg({transform:"translate(-0.5 -0.5)",style:{overflow:"visible",pointerEvents:"none",position:"absolute"}},[Ve.svgElem("path",{class:"rightOfModifiedBackgroundCoverUp",d:j(this,c=>{const d=this._previewEditorLayoutInfo.read(c);if(!(!d||this._originalBackgroundColor.read(c).isTransparent()))return new V3().moveTo(d.codeRect.getRightTop()).lineTo(d.codeRect.getRightTop().deltaX(1e3)).lineTo(d.codeRect.getRightBottom().deltaX(1e3)).lineTo(d.codeRect.getRightBottom()).build()}),style:{fill:EJ(Bi,"transparent")}})]).keepUpdated(this._store),this._originalOverlay=Ve.div({style:{pointerEvents:"none",display:this._previewEditorLayoutInfo.map(c=>c?.isInsertion?"none":"block")}},j(this,c=>{const d=Wc(this._previewEditorLayoutInfo).read(c);if(!d)return;const h=l.read(c),u=XA(this._tabAction).map(x=>`${Ga}px solid ${te(x)}`),g=`${Ga+h}px solid ${te(Bi)}`,f=d.read(c).codeScrollLeft!==0,p=d.map(x=>x.codeRect.bottom<x.editRect.bottom),_=ld*2+Ga*2,b=d.map(x=>Ft.fromLeftTopRightBottom(x.contentLeft-ld-Ga,x.codeRect.top,x.contentLeft,x.codeRect.bottom+_)).read(c),w=new me(b.left,Number.MAX_SAFE_INTEGER),C=d.map(x=>x.codeRect.intersectHorizontal(w)),v=C.map(x=>x.withMargin(h,0,h,h).intersectHorizontal(w)),y=C.map(x=>Ft.fromLeftTopWidthHeight(x.right-_+Ga,x.bottom-Ga,_,_).intersectHorizontal(w));return[Ve.div({class:"originalSeparatorSideBySide",style:{...v.read(c).toStyles(),boxSizing:"border-box",borderRadius:`${ld}px 0 0 ${ld}px`,borderTop:g,borderBottom:g,borderLeft:f?"none":g}}),Ve.div({class:"originalOverlaySideBySide",style:{...C.read(c).toStyles(),boxSizing:"border-box",borderRadius:`${ld}px 0 0 ${ld}px`,borderTop:u,borderBottom:u,borderLeft:f?"none":u,backgroundColor:te(cy)}}),Ve.div({class:"originalCornerCutoutSideBySide",style:{pointerEvents:"none",display:p.map(x=>x?"block":"none"),...y.read(c).toStyles()}},[Ve.div({class:"originalCornerCutoutBackground",style:{position:"absolute",top:"0px",left:"0px",width:"100%",height:"100%",backgroundColor:Br(cy,this._themeService).map(x=>x.toString())}}),Ve.div({class:"originalCornerCutoutBorder",style:{position:"absolute",top:"0px",left:"0px",width:"100%",height:"100%",boxSizing:"border-box",borderTop:u,borderRight:u,borderRadius:"0 100% 0 0",backgroundColor:te(Bi)}})]),Ve.div({class:"originalOverlaySideBySideHider",style:{...b.toStyles(),backgroundColor:te(Bi)}})]})).keepUpdated(this._store),this._modifiedOverlay=Ve.div({style:{pointerEvents:"none"}},j(this,c=>{const d=Wc(this._previewEditorLayoutInfo).read(c);if(!d)return;const h=d.map(C=>C.codeRect.bottom<C.editRect.bottom),u=l.read(c),g=h.map(C=>`0 ${ld}px ${ld}px ${C?ld:0}px`),f=Br(d0(this._tabAction),this._themeService).map(C=>`1px solid ${C.toString()}`),p=`${Ga+u}px solid ${te(Bi)}`,_=d.map(C=>C.editRect.withMargin(0,Ga)),b=_.map(C=>C.withMargin(u,u,u,0)),w=j(this,C=>{const v=_.read(C),y=d.read(C);return!y.isInsertion||y.contentLeft>=v.left?Ft.fromLeftTopWidthHeight(v.left,v.top,0,0):new Ft(y.contentLeft,v.top,v.left,v.top+Ga*2)});return[Ve.div({class:"modifiedInsertionSideBySide",style:{...w.read(c).toStyles(),backgroundColor:d0(this._tabAction).map(C=>te(C))}}),Ve.div({class:"modifiedSeparatorSideBySide",style:{...b.read(c).toStyles(),borderRadius:g,borderTop:p,borderBottom:p,borderRight:p,boxSizing:"border-box"}}),Ve.div({class:"modifiedOverlaySideBySide",style:{..._.read(c).toStyles(),borderRadius:g,border:f,boxSizing:"border-box",backgroundColor:te(xoe)}})]})).keepUpdated(this._store),this._nonOverflowView=Ve.div({class:"inline-edits-view",style:{position:"absolute",overflow:"visible",top:"0px",left:"0px",display:this._display}},[this._backgroundSvg,j(this,c=>this._shouldOverflow.read(c)?[]:[this._editorContainer,this._originalOverlay,this._modifiedOverlay])]).keepUpdated(this._store),this._register(this._editorObs.createOverlayWidget({domNode:this._nonOverflowView.element,position:zt(null),allowEditorOverflow:!1,minContentWidthInPx:j(this,c=>{const d=this._previewEditorLayoutInfo.read(c)?.maxContentWidth;return d===void 0?0:d})})),this.previewEditor.setModel(this._previewTextModel),this._register(De(c=>{const d=this._previewEditorLayoutInfo.read(c);if(!d)return;const h=d.editRect.withMargin(-ZC,-nb);this.previewEditor.layout({height:h.height,width:d.previewEditorWidth+15}),this._editorContainer.element.style.top=`${h.top}px`,this._editorContainer.element.style.left=`${h.left}px`,this._editorContainer.element.style.width=`${d.previewEditorWidth+nb}px`})),this._register(De(c=>{const d=this._previewEditorLayoutInfo.read(c);d&&this._previewEditorObs.editor.setScrollLeft(d.desiredPreviewEditorScrollLeft)})),this._updatePreviewEditor.recomputeInitiallyAndOnChange(this._store)}};UD=hFe([_K(5,_e),_K(6,wi)],UD);var fFe=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},pFe=function(s,e){return function(t,i){e(t,i,s)}};const cd=1;var jw;let Fb=(jw=class extends F{constructor(e,t,i,n){super(),this._editor=e,this._edit=t,this._tabAction=i,this._languageService=n,this._onDidClick=this._register(new M),this.onDidClick=this._onDidClick.event,this._start=this._editor.observePosition(zt(this._edit.range.getStartPosition()),this._store),this._end=this._editor.observePosition(zt(this._edit.range.getEndPosition()),this._store),this._line=document.createElement("div"),this._hoverableElement=Re(this,null),this.isHovered=this._hoverableElement.map((r,a)=>r?.didMouseMoveDuringHover.read(a)??!1),this._renderTextEffect=j(this,r=>{const a=this._editor.model.get(),l=a.getLineContent(this._edit.range.startLineNumber),c=On.replace(new me(this._edit.range.startColumn-1,this._edit.range.endColumn-1),this._edit.text),d=c.replace(l),h=a.tokenization.tokenizeLinesAt(this._edit.range.startLineNumber,[d])?.[0];let u;h?u=ih.fromLineTokens(h).slice(c.getRangeAfterReplace()).toLineTokens(this._edit.text,this._languageService.languageIdCodec):u=Ri.createEmpty(this._edit.text,this._languageService.languageIdCodec);const g=hx(new ux([u]),ah.fromEditor(this._editor.editor).withSetWidth(!1).withScrollBeyondLastColumn(0),[],this._line,!0);this._line.style.width=`${g.minWidthInPx}px`});const o=this._editor.observeLineHeightForPosition(this._edit.range.getStartPosition());this._layout=j(this,r=>{this._renderTextEffect.read(r);const a=this._start.read(r),l=this._end.read(r);if(!a||!l||a.x>l.x||a.y>l.y)return;const c=o.read(r),d=this._editor.scrollLeft.read(r),h=this._editor.getOption(59).read(r).typicalHalfwidthCharacterWidth,u=3*h,g=4,f=new _n(u,g),p=Ft.fromPoints(a,l).withHeight(c).translateX(-d),_=Ft.fromPointSize(p.getLeftBottom().add(f),new _n(this._edit.text.length*h,p.height)),b=_.withLeft(p.left);return{originalLine:p,modifiedLine:_,lowerBackground:b,lineHeight:c}}),this.minEditorScrollHeight=j(this,r=>{const a=Wc(this._layout).read(r);return a?a.read(r).modifiedLine.bottom+cd+this._editor.editor.getScrollTop():0}),this._root=Ve.div({class:"word-replacement"},[j(this,r=>{const a=Wc(this._layout).read(r);if(!a)return[];const l=XA(this._tabAction).map(d=>te(d)).read(r),c=d0(this._tabAction).map(d=>te(d)).read(r);return[Ve.div({style:{position:"absolute",...Zr(d=>QH(this._editor).read(d)),overflow:"hidden",pointerEvents:"none"}},[Ve.div({style:{position:"absolute",...Zr(d=>a.read(d).lowerBackground.withMargin(cd,2*cd,cd,0)),background:te(Bi),cursor:"pointer",pointerEvents:"auto"},onmousedown:d=>{d.preventDefault()},onmouseup:d=>this._onDidClick.fire(new is(be(d),d)),obsRef:d=>{this._hoverableElement.set(d,void 0)}}),Ve.div({style:{position:"absolute",...Zr(d=>a.read(d).modifiedLine.withMargin(cd,2*cd)),fontFamily:this._editor.getOption(58),fontSize:this._editor.getOption(61),fontWeight:this._editor.getOption(62),pointerEvents:"none",boxSizing:"border-box",borderRadius:"4px",border:`${cd}px solid ${c}`,background:te(RPe),display:"flex",justifyContent:"center",alignItems:"center",outline:`2px solid ${te(Bi)}`}},[this._line]),Ve.div({style:{position:"absolute",...Zr(d=>a.read(d).originalLine.withMargin(cd)),boxSizing:"border-box",borderRadius:"4px",border:`${cd}px solid ${l}`,background:te(TPe),pointerEvents:"none"}},[]),Ve.svg({width:11,height:14,viewBox:"0 0 11 14",fill:"none",style:{position:"absolute",left:a.map(d=>d.modifiedLine.left-16),top:a.map(d=>d.modifiedLine.top+Math.round((d.lineHeight-14-5)/2))}},[Ve.svgElem("path",{d:"M1 0C1 2.98966 1 5.92087 1 8.49952C1 9.60409 1.89543 10.5 3 10.5H10.5",stroke:te(N$)}),Ve.svgElem("path",{d:"M6 7.5L9.99999 10.49998L6 13.5",stroke:te(N$)})])])]})]).keepUpdated(this._store),this._register(this._editor.createOverlayWidget({domNode:this._root.element,minContentWidthInPx:zt(0),position:zt({preference:{top:0,left:0}}),allowEditorOverflow:!1}))}},jw.MAX_LENGTH=100,jw);Fb=fFe([pFe(3,Ei)],Fb);class mFe extends F{constructor(e,t,i){super(),this._originalEditor=e,this._state=t,this._modifiedTextModel=i,this._onDidClick=this._register(new M),this.onDidClick=this._onDidClick.event,this.isHovered=di(this._originalEditor).isTargetHovered(o=>o.target.type===6&&o.target.detail.injectedText?.options.attachedData instanceof kO&&o.target.detail.injectedText.options.attachedData.owner===this,this._store),this._tokenizationFinished=wFe(this._modifiedTextModel),this._decorations=j(this,o=>{const r=this._state.read(o);if(!r)return;const a=r.modifiedText,l=r.mode==="insertionInline",c=r.diff.length===1&&r.diff[0].innerChanges?.length===1,d=!0,h=[],u=[],g=Fe.register({className:"inlineCompletions-line-insert",description:"line-insert",isWholeLine:!0,marginClassName:"gutter-insert"}),f=Fe.register({className:"inlineCompletions-line-delete",description:"line-delete",isWholeLine:!0,marginClassName:"gutter-delete"}),p=Fe.register({className:"inlineCompletions-char-delete",description:"char-delete",isWholeLine:!1,zIndex:1}),_=Fe.register({className:"inlineCompletions-char-insert",description:"char-insert",isWholeLine:!0}),b=Fe.register({className:"inlineCompletions-char-insert",description:"char-insert",shouldFillLineOnLineBreak:!0}),w=Fe.register({className:"inlineCompletions-char-insert diff-range-empty",description:"char-insert diff-range-empty"}),C=Fe.register({className:"inlineCompletions-original-lines",description:"inlineCompletions-original-lines",isWholeLine:!1,shouldFillLineOnLineBreak:!0}),v=r.mode!=="sideBySide"&&r.mode!=="deletion"&&r.mode!=="insertionInline"&&r.mode!=="lineReplacement",y=r.mode==="lineReplacement";for(const x of r.diff)if(v&&(x.original.isEmpty||h.push({range:x.original.toInclusiveRange(),options:f}),x.modified.isEmpty||u.push({range:x.modified.toInclusiveRange(),options:g})),x.modified.isEmpty||x.original.isEmpty)x.original.isEmpty||h.push({range:x.original.toInclusiveRange(),options:p}),x.modified.isEmpty||u.push({range:x.modified.toInclusiveRange(),options:_});else{const L=l&&_Fe(x);for(const E of x.innerChanges||[]){if(x.original.contains(E.originalRange.startLineNumber)&&!(y&&E.originalRange.isEmpty())){const I=this._originalEditor.getModel()?.getValueInRange(E.originalRange,1);h.push({range:E.originalRange,options:{description:"char-delete",shouldFillLineOnLineBreak:!1,className:wO("inlineCompletions-char-delete",E.originalRange.isSingleLine()&&r.mode==="insertionInline"&&"single-line-inline",E.originalRange.isEmpty()&&"empty",(E.originalRange.isEmpty()&&c||r.mode==="deletion"&&I===`
`)&&d&&!L&&"diff-range-empty"),inlineClassName:L?wO("strike-through","inlineCompletions"):null,zIndex:1}})}if(x.modified.contains(E.modifiedRange.startLineNumber)&&u.push({range:E.modifiedRange,options:E.modifiedRange.isEmpty()&&d&&!L&&c?w:b}),L){const I=a.getValueOfRange(E.modifiedRange),R=I.length>3?[{text:I.slice(0,1),extraClasses:["start"],offsetRange:new me(E.modifiedRange.startColumn-1,E.modifiedRange.startColumn)},{text:I.slice(1,-1),extraClasses:[],offsetRange:new me(E.modifiedRange.startColumn,E.modifiedRange.endColumn-2)},{text:I.slice(-1),extraClasses:["end"],offsetRange:new me(E.modifiedRange.endColumn-2,E.modifiedRange.endColumn-1)}]:[{text:I,extraClasses:["start","end"],offsetRange:new me(E.modifiedRange.startColumn-1,E.modifiedRange.endColumn)}];this._tokenizationFinished.read(o);const O=this._modifiedTextModel.tokenization.getLineTokens(E.modifiedRange.startLineNumber);for(const{text:P,extraClasses:B,offsetRange:$}of R)h.push({range:k.fromPositions(E.originalRange.getEndPosition()),options:{description:"inserted-text",before:{tokens:O.getTokensInRange($),content:P,inlineClassName:wO("inlineCompletions-char-insert",E.modifiedRange.isSingleLine()&&r.mode==="insertionInline"&&"single-line-inline",...B),cursorStops:jr.None,attachedData:new kO(this)},zIndex:2,showIfCollapsed:!0}})}}}if(r.isInDiffEditor)for(const x of r.diff)x.original.isEmpty||h.push({range:x.original.toExclusiveRange(),options:C});return{originalDecorations:h,modifiedDecorations:u}}),this._register(di(this._originalEditor).setDecorations(this._decorations.map(o=>o?.originalDecorations??[])));const n=this._state.map(o=>o?.modifiedCodeEditor);this._register(ps((o,r)=>{const a=n.read(o);a&&r.add(di(a).setDecorations(this._decorations.map(l=>l?.modifiedDecorations??[])))})),this._register(this._originalEditor.onMouseUp(o=>{if(o.target.type!==6)return;const r=o.target.detail.injectedText?.options.attachedData;r instanceof kO&&r.owner===this&&this._onDidClick.fire(o.event)}))}}class kO{constructor(e){this.owner=e}}function _Fe(s){return s.innerChanges?s.innerChanges.every(e=>kI(e.modifiedRange)&&kI(e.originalRange)):!1}let bFe=0;function wFe(s){return bt(s.onDidChangeTokens,()=>bFe++)}var vFe=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},CFe=function(s,e){return function(t,i){e(t,i,s)}};let Y3=class extends F{constructor(e,t,i,n,o,r){super(),this._editor=e,this._host=t,this._model=i,this._ghostTextIndicator=n,this._focusIsInMenu=o,this._instantiationService=r,this._editorObs=di(this._editor),this._tabAction=j(d=>this._model.read(d)?.tabAction.read(d)??us.Inactive),this._constructorDone=Re(this,!1),this._uiState=j(this,d=>{const h=this._model.read(d);if(!h||!this._constructorDone.read(d))return;const u=h.inlineEdit;let g=Ks.fromEdit(u.edit),f=u.edit.apply(u.originalText),p=xI(g,u.originalText,new vu(f)),_=this.determineRenderState(h,d,p,new vu(f));if(!_){Oe(new Error(`unable to determine view: tried to render ${this._previousView?.view}`));return}if(_.kind===mt.SideBySide){const w=UPe(f,u.modifiedLineRange,l.getOptions().tabSize);f=w.applyToString(f),g=HPe(g,w),p=xI(g,u.originalText,new vu(f))}return this._previewTextModel.setLanguage(this._editor.getModel().getLanguageId()),this._previewTextModel.getValue()!==f&&this._previewTextModel.setValue(f),h.showCollapsed.read(d)&&!this._indicator.read(d)?.isHoverVisible.read(d)&&(_={kind:mt.Collapsed,viewData:_.viewData}),h.handleInlineEditShown(_.kind,_.viewData),{state:_,diff:p,edit:u,newText:f,newTextLineCount:u.modifiedLineRange.length,isInDiffEditor:h.isInDiffEditor}}),this._previewTextModel=this._register(this._instantiationService.createInstance(Rm,"",this._editor.getModel().getLanguageId(),{...Rm.DEFAULT_CREATION_OPTIONS,bracketPairColorizationOptions:{enabled:!0,independentColorPoolPerBracketType:!1}},null)),this._indicatorCyclicDependencyCircuitBreaker=Re(this,!1),this._indicator=j(this,d=>{if(!this._indicatorCyclicDependencyCircuitBreaker.read(d))return;const h=Jn({owner:this,equalsFn:Vf(nW())},g=>{const f=this._ghostTextIndicator.read(g);if(f)return f.lineRange;const p=this._uiState.read(g);if(p){if(p.state?.kind==="custom"){const _=p.state.displayLocation?.range;if(!_)throw new Ce("custom view should have a range");return new Ne(_.startLineNumber,_.endLineNumber)}return p.state?.kind==="insertionMultiLine"?this._insertion.originalLines.read(g):p.edit.displayRange}}),u=j(this,g=>{const f=this._model.read(g);if(f)return f;const p=this._ghostTextIndicator.read(g);return p?p.model:f});return d.store.add(this._instantiationService.createInstance($3,this._editorObs,h,this._gutterIndicatorOffset,u,this._inlineEditsIsHovered,this._focusIsInMenu))}),this._inlineEditsIsHovered=j(this,d=>this._sideBySide.isHovered.read(d)||this._wordReplacementViews.read(d).some(h=>h.isHovered.read(d))||this._deletion.isHovered.read(d)||this._inlineDiffView.isHovered.read(d)||this._lineReplacementView.isHovered.read(d)||this._insertion.isHovered.read(d)||this._customView.isHovered.read(d)),this._gutterIndicatorOffset=j(this,d=>{if(this._uiState.read(d)?.state?.kind==="insertionMultiLine")return this._insertion.startLineOffset.read(d);const h=this._ghostTextIndicator.read(d);return h?xFe(h,this._editor):0}),this._sideBySide=this._register(this._instantiationService.createInstance(UD,this._editor,this._model.map(d=>d?.inlineEdit),this._previewTextModel,this._uiState.map(d=>d&&d.state?.kind===mt.SideBySide?{newTextLineCount:d.newTextLineCount,isInDiffEditor:d.isInDiffEditor}:void 0),this._tabAction)),this._deletion=this._register(this._instantiationService.createInstance(oFe,this._editor,this._model.map(d=>d?.inlineEdit),this._uiState.map(d=>d&&d.state?.kind===mt.Deletion?{originalRange:d.state.originalRange,deletions:d.state.deletions,inDiffEditor:d.isInDiffEditor}:void 0),this._tabAction)),this._insertion=this._register(this._instantiationService.createInstance(G3,this._editor,this._uiState.map(d=>d&&d.state?.kind===mt.InsertionMultiLine?{lineNumber:d.state.lineNumber,startColumn:d.state.column,text:d.state.text,inDiffEditor:d.isInDiffEditor}:void 0),this._tabAction)),this._inlineDiffViewState=j(this,d=>{const h=this._uiState.read(d);if(!(!h||!h.state)&&!(h.state.kind==="wordReplacements"||h.state.kind==="insertionMultiLine"||h.state.kind==="collapsed"||h.state.kind==="custom"))return{modifiedText:new vu(h.newText),diff:h.diff,mode:h.state.kind,modifiedCodeEditor:this._sideBySide.previewEditor,isInDiffEditor:h.isInDiffEditor}}),this._inlineCollapsedView=this._register(this._instantiationService.createInstance(q3,this._editor,this._model.map((d,h)=>this._uiState.read(h)?.state?.kind==="collapsed"?d?.inlineEdit:void 0))),this._customView=this._register(this._instantiationService.createInstance(j3,this._editor,this._model.map((d,h)=>this._uiState.read(h)?.state?.kind==="custom"?d?.displayLocation:void 0),this._tabAction)),this._inlineDiffView=this._register(new mFe(this._editor,this._inlineDiffViewState,this._previewTextModel)),this._wordReplacementViews=rW(this,this._uiState.map(d=>d?.state?.kind==="wordReplacements"?d.state.replacements:[]),(d,h)=>h.add(this._instantiationService.createInstance(Fb,this._editorObs,d,this._tabAction))),this._lineReplacementView=this._register(this._instantiationService.createInstance(K3,this._editorObs,this._uiState.map(d=>d?.state?.kind===mt.LineReplacement?{originalRange:d.state.originalRange,modifiedRange:d.state.modifiedRange,modifiedLines:d.state.modifiedLines,replacements:d.state.replacements}:void 0),this._uiState.map(d=>d?.isInDiffEditor??!1),this._tabAction)),this._useCodeShifting=this._editorObs.getOption(71).map(d=>d.edits.allowCodeShifting),this._renderSideBySide=this._editorObs.getOption(71).map(d=>d.edits.renderSideBySide),this._register(ps((d,h)=>{const u=this._model.read(d);u&&h.add(oe.any(this._sideBySide.onDidClick,this._deletion.onDidClick,this._lineReplacementView.onDidClick,this._insertion.onDidClick,...this._wordReplacementViews.read(d).map(g=>g.onDidClick),this._inlineDiffView.onDidClick,this._customView.onDidClick)(g=>{this._viewHasBeenShownLongerThan(350)&&(g.preventDefault(),u.accept())}))})),this._indicator.recomputeInitiallyAndOnChange(this._store),this._wordReplacementViews.recomputeInitiallyAndOnChange(this._store),this._indicatorCyclicDependencyCircuitBreaker.set(!0,void 0),this._register(this._instantiationService.createInstance(U3,this._host,this._model,this._indicator,this._inlineCollapsedView));const a=j(this,d=>Math.max(...this._wordReplacementViews.read(d).map(h=>h.minEditorScrollHeight.read(d)),this._lineReplacementView.minEditorScrollHeight.read(d),this._customView.minEditorScrollHeight.read(d))).recomputeInitiallyAndOnChange(this._store),l=this._editor.getModel();let c;this._register(De(d=>{const h=a.read(d);this._editor.changeViewZones(u=>{const g=this._editor.getScrollHeight(),f=h-g+1;f!==0&&c&&(u.removeZone(c),c=void 0),!(f<=0)&&(c=u.addZone({afterLineNumber:l.getLineCount(),heightInPx:f,domNode:ie("div.minScrollHeightViewZone")}))})})),this._constructorDone.set(!0,void 0)}getCacheId(e){return e.inlineEdit.inlineCompletion.identity.id}determineView(e,t,i,n){const o=e.inlineEdit,r=this._previousView?.id===this.getCacheId(e)&&!e.displayLocation?.jumpToEdit,a=this._previousView?.editorWidth!==this._editorObs.layoutInfoWidth.read(t)&&(this._previousView?.view===mt.SideBySide||this._previousView?.view===mt.LineReplacement);if(r&&!a)return this._previousView.view;if(e.inlineEdit.inlineCompletion instanceof Iw&&e.inlineEdit.inlineCompletion.uri||e.displayLocation&&!e.inlineEdit.inlineCompletion.identity.jumpedTo.read(t))return mt.Custom;const l=o.originalLineRange.length,c=o.modifiedLineRange.length,d=i.flatMap(u=>u.innerChanges??[]),h=d.length===1;if(!e.isInDiffEditor){if(h&&this._useCodeShifting.read(t)!=="never"&&Toe(i))return yFe(i,o.cursorPosition)?mt.InsertionInline:mt.LineReplacement;if(vK(d,o,n))return mt.Deletion;if(wK(i)&&this._useCodeShifting.read(t)==="always")return mt.InsertionMultiLine;if(d.every(g=>cn.ofRange(g.originalRange).columnCount<Fb.MAX_LENGTH&&cn.ofRange(g.modifiedRange).columnCount<Fb.MAX_LENGTH)&&h&&l===1&&c===1){const g=d.map(p=>n.getValueOfRange(p.modifiedRange)),f=d.map(p=>e.inlineEdit.originalText.getValueOfRange(p.originalRange));if(!g.some(p=>p.includes("	"))&&!f.some(p=>p.includes("	"))&&(!d.some(p=>p.originalRange.isEmpty())||!CK(d.map(p=>new Ui(p.originalRange,"")),o.originalText).some(p=>p.range.isEmpty()&&cn.ofRange(p.range).columnCount<Fb.MAX_LENGTH)))return mt.WordReplacements}}if(l>0&&c>0)return l===1&&c===1&&!e.isInDiffEditor?mt.LineReplacement:this._renderSideBySide.read(t)!=="never"&&UD.fitsInsideViewport(this._editor,this._previewTextModel,o,t)?mt.SideBySide:mt.LineReplacement;if(e.isInDiffEditor){if(vK(d,o,n))return mt.Deletion;if(wK(i)&&this._useCodeShifting.read(t)==="always")return mt.InsertionMultiLine}return mt.SideBySide}determineRenderState(e,t,i,n){const o=e.inlineEdit;let r=this.determineView(e,t,i,n);if(this._willRenderAboveCursor(t,o,r))switch(r){case mt.LineReplacement:case mt.WordReplacements:r=mt.SideBySide;break}this._previousView={id:this.getCacheId(e),view:r,editorWidth:this._editor.getLayoutInfo().width,timestamp:Date.now()};const a=i.flatMap(f=>f.innerChanges??[]),l=this._editor.getModel(),c=a.map(f=>({originalRange:f.originalRange,modifiedRange:f.modifiedRange,original:l.getValueInRange(f.originalRange),modified:n.getValueOfRange(f.modifiedRange)})),d=o.cursorPosition,h=c.length===0?!1:c[0].modified.startsWith(l.getEOL()),u={cursorColumnDistance:o.edit.replacements.length===0?0:o.edit.replacements[0].range.getStartPosition().column-d.column,cursorLineDistance:o.lineEdit.lineRange.startLineNumber-d.lineNumber+(h&&o.lineEdit.lineRange.startLineNumber>=d.lineNumber?1:0),lineCountOriginal:o.lineEdit.lineRange.length,lineCountModified:o.lineEdit.newLines.length,characterCountOriginal:c.reduce((f,p)=>f+p.original.length,0),characterCountModified:c.reduce((f,p)=>f+p.modified.length,0),disjointReplacements:c.length,sameShapeReplacements:c.every(f=>f.original===c[0].original&&f.modified===c[0].modified)};switch(r){case mt.InsertionInline:return{kind:mt.InsertionInline,viewData:u};case mt.SideBySide:return{kind:mt.SideBySide,viewData:u};case mt.Collapsed:return{kind:mt.Collapsed,viewData:u};case mt.Custom:return{kind:mt.Custom,displayLocation:e.displayLocation,viewData:u}}if(r===mt.Deletion)return{kind:mt.Deletion,originalRange:o.originalLineRange,deletions:a.map(f=>f.originalRange),viewData:u};if(r===mt.InsertionMultiLine){const f=a[0];return{kind:mt.InsertionMultiLine,lineNumber:f.originalRange.startLineNumber,column:f.originalRange.startColumn,text:n.getValueOfRange(f.modifiedRange),viewData:u}}const g=c.map(f=>new Ui(f.originalRange,f.modified));if(g.length!==0){if(r===mt.WordReplacements){let f=SFe(g,o.originalText);return f.some(p=>p.range.isEmpty())&&(f=CK(g,o.originalText)),{kind:mt.WordReplacements,replacements:f,viewData:u}}if(r===mt.LineReplacement)return{kind:mt.LineReplacement,originalRange:o.originalLineRange,modifiedRange:o.modifiedLineRange,modifiedLines:o.modifiedLineRange.mapToLineArray(f=>n.getLineAt(f)),replacements:a.map(f=>({originalRange:f.originalRange,modifiedRange:f.modifiedRange})),viewData:u}}}_willRenderAboveCursor(e,t,i){if(this._useCodeShifting.read(e)==="always")return!1;for(const o of t.multiCursorPositions)if(i===mt.WordReplacements&&o.lineNumber===t.originalLineRange.startLineNumber+1||i===mt.LineReplacement&&o.lineNumber>=t.originalLineRange.endLineNumberExclusive&&o.lineNumber<t.modifiedLineRange.endLineNumberExclusive+t.modifiedLineRange.length)return!0;return!1}_viewHasBeenShownLongerThan(e){const t=this._previousView?.timestamp;if(!t)throw new Ce("viewHasBeenShownLongThan called before a view has been shown");return Date.now()-t>=e}};Y3=vFe([CFe(5,_e)],Y3);function Toe(s){return s.every(t=>t.innerChanges.every(i=>e(i)));function e(t){return!(!t.originalRange.isEmpty()||!(t.modifiedRange.startLineNumber===t.modifiedRange.endLineNumber))}}function yFe(s,e){if(!e||!Toe(s))return!1;const t=e;return s.every(n=>n.innerChanges.every(o=>i(o)));function i(n){const o=n.originalRange.getStartPosition();return!!(t.isBeforeOrEqual(o)||o.lineNumber<t.lineNumber)}}function wK(s){const e=s.flatMap(i=>i.innerChanges??[]);if(e.length!==1)return!1;const t=e[0];return!(!t.originalRange.isEmpty()||t.modifiedRange.startLineNumber===t.modifiedRange.endLineNumber)}function vK(s,e,t){return s.map(n=>({original:e.originalText.getValueOfRange(n.originalRange),modified:t.getValueOfRange(n.modifiedRange)})).every(({original:n,modified:o})=>o.trim()===""&&n.length>0&&(n.length>o.length||n.trim()!==""))}function SFe(s,e){return Noe(s,e,t=>/^[a-zA-Z]$/.test(t))}function CK(s,e){return Noe(s,e,t=>!/^\s$/.test(t))}function Noe(s,e,t){const i=[];s.sort((o,r)=>k.compareRangesUsingStarts(o.range,r.range));for(const o of s){let r=o.range.startColumn-1,a=o.range.endColumn-2,l="",c="";const d=e.getLineAt(o.range.startLineNumber),h=e.getLineAt(o.range.endLineNumber);if(n(d[r]))for(;n(d[r-1]);)l=d[r-1]+l,r--;if(n(h[a])||a<r)for(;n(h[a+1]);)c+=h[a+1],a++;let u=new Ui(new k(o.range.startLineNumber,r+1,o.range.endLineNumber,a+2),l+o.text+c);i.length>0&&k.areIntersectingOrTouching(i[i.length-1].range,u.range)&&(u=Ui.joinReplacements([i.pop(),u],e)),i.push(u)}function n(o){return o===void 0?!1:t(o)}return i}function xFe(s,e){const t=s.model.inlineEdit.edit.replacements;if(t.length!==1)return 0;const i=e.getModel();if(!i)return 0;const n=i.getEOL(),o=t[0];if(o.range.isEmpty()&&o.text.startsWith(n)){const r=e.getLineHeightForPosition(o.range.getStartPosition());return LFe(o.text,n)*r}return 0}function LFe(s,e){if(!e.length)return 0;let t=0,i=0;for(;s.startsWith(e,i);)t++,i+=e.length;return t}var kFe=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},EFe=function(s,e){return function(t,i){e(t,i,s)}},tm;let Z3=(tm=class extends F{constructor(e,t,i,n,o){super(),this._editor=e,this._edit=t,this._model=i,this._focusIsInMenu=n,this._inlineEdit=j(this,r=>{const a=this._model.read(r);if(!a)return;const l=this._edit.read(r);if(!l)return;const c=this._editor.getModel();if(!c)return;const d=a.inlineEditState.read(void 0)?.inlineCompletion.updatedEdit;if(!d)return;const h=d.replacements.map(f=>{const p=k.fromPositions(c.getPositionAt(f.replaceRange.start),c.getPositionAt(f.replaceRange.endExclusive));return new Ui(p,f.newText)}),u=new Na(h),g=new Fm(c);return new woe(g,u,a.primaryPosition.read(void 0),a.allPositions.read(void 0),l.commands,l.inlineCompletion)}),this._inlineEditModel=j(this,r=>{const a=this._model.read(r);if(!a)return;const l=this._inlineEdit.read(r);if(!l)return;const c=j(this,d=>{if(this._editorObs.isFocused.read(d)){if(a.tabShouldJumpToInlineEdit.read(d))return us.Jump;if(a.tabShouldAcceptInlineEdit.read(d))return us.Accept}return us.Inactive});return new voe(a,l,c)}),this._inlineEditHost=j(this,r=>{const a=this._model.read(r);if(a)return new CPe(a)}),this._ghostTextIndicator=j(this,r=>{const a=this._model.read(r);if(!a)return;const l=a.inlineCompletionState.read(r);if(!l)return;const c=l.inlineCompletion;if(!c||!c.showInlineEditMenu)return;const d=Ne.ofLength(l.primaryGhostText.lineNumber,1);return new yPe(this._editor,a,d,c)}),this._editorObs=di(this._editor),this._register(o.createInstance(Y3,this._editor,this._inlineEditHost,this._inlineEditModel,this._ghostTextIndicator,this._focusIsInMenu))}},tm.hot=bA(tm),tm);Z3=kFe([EFe(4,_e)],Z3);var IFe=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},DFe=function(s,e){return function(t,i){e(t,i,s)}};let X3=class extends F{constructor(e,t,i,n){super(),this._editor=e,this._model=t,this._focusIsInMenu=i,this._instantiationService=n,this._ghostTexts=j(this,o=>this._model.read(o)?.ghostTexts.read(o)??[]),this._stablizedGhostTexts=NMe(this._ghostTexts,this._store),this._editorObs=di(this._editor),this._ghostTextWidgets=rW(this,this._stablizedGhostTexts,(o,r)=>Wr(a=>this._instantiationService.createInstance(l0.hot.read(a),this._editor,{ghostText:o,warning:this._model.map((l,c)=>{const d=l?.warning?.read(c);return d?{icon:d.icon}:void 0}),minReservedLineCount:zt(0),targetTextModel:this._model.map(l=>l?.textModel),handleInlineCompletionShown:this._model.map((l,c)=>{const d=l?.inlineCompletionState.read(c)?.inlineCompletion;return d?h=>l.handleInlineSuggestionShown(d,mt.GhostText,h):()=>{}})},this._editorObs.getOption(71).map(l=>({syntaxHighlightingEnabled:l.syntaxHighlightingEnabled})),!1,!1)).recomputeInitiallyAndOnChange(r)).recomputeInitiallyAndOnChange(this._store),this._inlineEdit=j(this,o=>this._model.read(o)?.inlineEditState.read(o)?.inlineEdit),this._everHadInlineEdit=Ih(this,(o,r)=>r||!!this._inlineEdit.read(o)||!!this._model.read(o)?.inlineCompletionState.read(o)?.inlineCompletion?.showInlineEditMenu),this._inlineEditWidget=Wr(o=>{if(this._everHadInlineEdit.read(o))return this._instantiationService.createInstance(Z3.hot.read(o),this._editor,this._inlineEdit,this._model,this._focusIsInMenu)}).recomputeInitiallyAndOnChange(this._store),this._fontFamily=this._editorObs.getOption(71).map(o=>o.fontFamily),this._register(Ype(j(o=>`
.monaco-editor .ghost-text-decoration,
.monaco-editor .ghost-text-decoration-preview,
.monaco-editor .ghost-text {
	font-family: ${this._fontFamily.read(o)};
}`))),this._register(new s5(this._editor,this._model,this._instantiationService))}shouldShowHoverAtViewZone(e){return this._ghostTextWidgets.get()[0]?.get().ownsViewZone(e)??!1}};X3=IFe([DFe(3,_e)],X3);var TFe=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},dd=function(s,e){return function(t,i){e(t,i,s)}},Gl,bc;let Zo=(bc=class extends F{static getInFocusedEditorOrParent(e){const t=Jne(e);return t?Gl.get(t):null}static get(e){return D2e(e.getContribution(Gl.ID))}constructor(e,t,i,n,o,r,a,l,c,d){super(),this.editor=e,this._instantiationService=t,this._contextKeyService=i,this._configurationService=n,this._commandService=o,this._debounceService=r,this._languageFeaturesService=a,this._accessibilitySignalService=l,this._keybindingService=c,this._accessibilityService=d,this._editorObs=di(this.editor),this._positions=j(this,p=>this._editorObs.selections.read(p)?.map(_=>_.getEndPosition())??[new A(1,1)]),this._suggestWidgetAdapter=this._register(new fPe(this._editorObs,p=>this.model.get()?.handleSuggestAccepted(p),()=>this.model.get()?.selectedInlineCompletion.get()?.getSingleTextEdit())),this._enabledInConfig=bt(this,this.editor.onDidChangeConfiguration,()=>this.editor.getOption(71).enabled),this._isScreenReaderEnabled=bt(this,this._accessibilityService.onDidChangeScreenReaderOptimized,()=>this._accessibilityService.isScreenReaderOptimized()),this._editorDictationInProgress=bt(this,this._contextKeyService.onDidChangeContext,()=>this._contextKeyService.getContext(this.editor.getDomNode()).getValue("editorDictation.inProgress")===!0),this._enabled=j(this,p=>this._enabledInConfig.read(p)&&(!this._isScreenReaderEnabled.read(p)||!this._editorDictationInProgress.read(p))),this._debounceValue=this._debounceService.for(this._languageFeaturesService.inlineCompletionsProvider,"InlineCompletionsDebounce",{min:50,max:50}),this._focusIsInMenu=Re(this,!1),this._focusIsInEditorOrMenu=j(this,p=>{const _=this._editorObs.isFocused.read(p),b=this._focusIsInMenu.read(p);return _||b}),this._cursorIsInIndentation=j(this,p=>{const _=this._editorObs.cursorPosition.read(p);if(_===null)return!1;const b=this._editorObs.model.read(p);if(!b)return!1;this._editorObs.versionId.read(p);const w=b.getLineIndentColumn(_.lineNumber);return _.column<=w}),this.model=Wr(this,p=>{if(this._editorObs.isReadonly.read(p))return;const _=this._editorObs.model.read(p);return _?this._instantiationService.createInstance(x3,_,this._suggestWidgetAdapter.selectedItem,this._editorObs.versionId,this._positions,this._debounceValue,this._enabled,this.editor):void 0}).recomputeInitiallyAndOnChange(this._store),this._playAccessibilitySignal=Yr(this),this._hideInlineEditOnSelectionChange=this._editorObs.getOption(71).map(p=>!0),this._view=this._register(this._instantiationService.createInstance(X3,this.editor,this.model,this._focusIsInMenu)),Gl._instances.add(this),this._register(pe(()=>Gl._instances.delete(this))),this._register(De(p=>{const _=this.model.read(p);if(!_)return;const b=_.state.read(p);if(!b||!this._focusIsInEditorOrMenu.read(void 0))return;const w=b.kind==="inlineEdit"?b.nextEditUri:void 0;for(const C of Gl._instances)C!==this&&(w&&dg(w,C.editor.getModel()?.uri)?C.model.read(void 0)?.trigger():C.reject())})),this._register(De(p=>{const _=this.model.read(p),b=this.editor.getModel()?.uri;!_||!b||p.store.add(_.onDidAccept(()=>{for(const w of Gl._instances){if(w===this)continue;const C=w.model.read(void 0)?.state.read(void 0);C?.kind==="inlineEdit"&&dg(C.nextEditUri,b)&&w.model.read(void 0)?.stop("automatic")}}))})),this._register(yl(this._editorObs.onDidType,(p,_)=>{this._enabled.get()&&this.model.get()?.trigger()})),this._register(yl(this._editorObs.onDidPaste,(p,_)=>{this._enabled.get()&&this.model.get()?.trigger()}));const h=new Set([hw.Tab.id,hw.DeleteLeft.id,hw.DeleteRight.id,jS,"acceptSelectedSuggestion",SD.ID,yD.ID,Ut.NextMatchFindAction,...EE.getRegisteredCommands()]);this._register(this._commandService.onDidExecuteCommand(p=>{if(h.has(p.commandId)&&e.hasTextFocus()&&this._enabled.get()){let _=!1;p.commandId===jS&&(_=!0),this._editorObs.forceUpdate(b=>{this.model.get()?.trigger(b,{noDelay:_})})}})),this._register(yl(this._editorObs.selections,(p,_,b)=>{if(b.some(w=>w.reason===3||w.source==="api")){if(!this._hideInlineEditOnSelectionChange.get()&&this.model.get()?.state.get()?.kind==="inlineEdit")return;const w=this.model.get();if(!w)return;w.state.get()?.kind==="ghostText"&&this.model.get()?.stop()}})),this._register(De(p=>{const _=this._focusIsInEditorOrMenu.read(p),b=this.model.read(void 0);if(_){const w=b?.state.read(void 0);(!w||w.kind!=="inlineEdit"||!w.nextEditUri)&&Vt(C=>{for(const v of Gl._instances)v!==this&&v.model.read(void 0)?.stop("automatic",C)});return}this._contextKeyService.getContextKeyValue("accessibleViewIsShown")||this._configurationService.getValue("editor.inlineSuggest.keepOnBlur")||e.getOption(71).keepOnBlur||xv.dropDownVisible||b&&(b.state.read(void 0)?.inlineCompletion?.isFromExplicitRequest&&b.inlineEditAvailable.read(void 0)||Vt(w=>{b.stop("automatic",w)}))})),this._register(De(p=>{const _=this.model.read(p)?.inlineCompletionState.read(p);_?.suggestItem?_.primaryGhostText.lineCount>=2&&this._suggestWidgetAdapter.forceRenderingAbove():this._suggestWidgetAdapter.stopForceRenderingAbove()})),this._register(pe(()=>{this._suggestWidgetAdapter.stopForceRenderingAbove()}));const u=Ih(this,(p,_)=>{const w=this.model.read(p)?.state.read(p);return this._suggestWidgetAdapter.selectedItem.get()?_:w?.inlineCompletion?.semanticId});this._register(Ree(j(p=>(this._playAccessibilitySignal.read(p),u.read(p),{})),async(p,_,b,w)=>{let C=this.model.get(),v=C?.state.get();if(!v||!C||(await Tc(50,lP(w)),await wie(this._suggestWidgetAdapter.selectedItem,fs,()=>!1,lP(w)),C=this.model.get(),v=C?.state.get(),!v||!C))return;const y=v.kind==="ghostText"?C.textModel.getLineContent(v.primaryGhostText.lineNumber):"";this._accessibilitySignalService.playSignal(v.kind==="ghostText"?Gs.inlineSuggestion:Gs.nextEditSuggestion),this.editor.getOption(12)&&(v.kind==="ghostText"?this._provideScreenReaderUpdate(v.primaryGhostText.renderForScreenReader(y)):this._provideScreenReaderUpdate(""))})),this._register(this._configurationService.onDidChangeConfiguration(p=>{p.affectsConfiguration("accessibility.verbosity.inlineCompletions")&&this.editor.updateOptions({inlineCompletionsAccessibilityVerbose:this._configurationService.getValue("accessibility.verbosity.inlineCompletions")})})),this.editor.updateOptions({inlineCompletionsAccessibilityVerbose:this._configurationService.getValue("accessibility.verbosity.inlineCompletions")});const g=new RMe(this._contextKeyService);this._register(g.bind(Ot.cursorInIndentation,this._cursorIsInIndentation)),this._register(g.bind(Ot.hasSelection,p=>!this._editorObs.cursorSelection.read(p)?.isEmpty())),this._register(g.bind(Ot.cursorAtInlineEdit,this.model.map((p,_)=>p?.inlineEditState?.read(_)?.cursorAtInlineEdit.read(_)))),this._register(g.bind(Ot.tabShouldAcceptInlineEdit,this.model.map((p,_)=>!!p?.tabShouldAcceptInlineEdit.read(_)))),this._register(g.bind(Ot.tabShouldJumpToInlineEdit,this.model.map((p,_)=>!!p?.tabShouldJumpToInlineEdit.read(_)))),this._register(g.bind(Ot.inlineEditVisible,p=>this.model.read(p)?.inlineEditState.read(p)!==void 0)),this._register(g.bind(Ot.inlineSuggestionHasIndentation,p=>this.model.read(p)?.getIndentationInfo(p)?.startsWithIndentation)),this._register(g.bind(Ot.inlineSuggestionHasIndentationLessThanTabSize,p=>this.model.read(p)?.getIndentationInfo(p)?.startsWithIndentationLessThanTabSize)),this._register(g.bind(Ot.suppressSuggestions,p=>{const b=this.model.read(p)?.inlineCompletionState.read(p);return b?.primaryGhostText&&b?.inlineCompletion?b.inlineCompletion.source.inlineSuggestions.suppressSuggestions:void 0})),this._register(g.bind(Ot.inlineSuggestionVisible,p=>{const b=this.model.read(p)?.inlineCompletionState.read(p);return!!b?.inlineCompletion&&b?.primaryGhostText!==void 0&&!b?.primaryGhostText.isEmpty()}));const f=j(this,p=>{const w=this.model.read(p)?.inlineCompletionState.read(p)?.primaryGhostText;return!w||w.isEmpty()?void 0:new A(w.lineNumber,w.parts[0].column)});this._register(g.bind(Ot.cursorBeforeGhostText,p=>{const _=f.read(p);if(!_)return!1;const b=this._editorObs.cursorPosition.read(p);return b?_.equals(b):!1})),this._register(this._instantiationService.createInstance(d3,this.editor))}playAccessibilitySignal(e){this._playAccessibilitySignal.trigger(e)}_provideScreenReaderUpdate(e){const t=this._contextKeyService.getContextKeyValue("accessibleViewIsShown"),i=this._keybindingService.lookupKeybinding("editor.action.accessibleView");let n;!t&&i&&this.editor.getOption(169)&&(n=m(1204,"Inspect this in the accessible view ({0})",i.getAriaLabel())),Js(n?e+", "+n:e)}shouldShowHoverAt(e){const t=this.model.get()?.primaryGhostText.get();return t?t.parts.some(i=>e.containsPosition(new A(t.lineNumber,i.column))):!1}shouldShowHoverAtViewZone(e){return this._view.shouldShowHoverAtViewZone(e)}reject(){Vt(e=>{const t=this.model.get();if(t&&(t.stop("explicitCancel",e),this._focusIsInEditorOrMenu.get()))for(const i of Gl._instances)i!==this&&i.model.get()?.stop("automatic",e)})}jump(){const e=this.model.get();e&&e.jump()}},Gl=bc,bc._instances=new Set,bc.hot=bA(bc),bc.ID="editor.contrib.inlineCompletionsController",bc);Zo=Gl=TFe([dd(1,_e),dd(2,Ae),dd(3,He),dd(4,Gt),dd(5,ga),dd(6,ge),dd(7,Th),dd(8,ft),dd(9,Wn)],Zo);const WN=class WN extends ue{constructor(){super({id:WN.ID,label:V(1183,"Show Next Inline Suggestion"),precondition:K.and(N.writable,Ot.inlineSuggestionVisible),kbOpts:{weight:100,primary:606}})}async run(e,t){Zo.get(t)?.model.get()?.next()}};WN.ID=yne;let Q3=WN;const HN=class HN extends ue{constructor(){super({id:HN.ID,label:V(1184,"Show Previous Inline Suggestion"),precondition:K.and(N.writable,Ot.inlineSuggestionVisible),kbOpts:{weight:100,primary:604}})}async run(e,t){Zo.get(t)?.model.get()?.previous()}};HN.ID=Cne;let J3=HN;const NFe="vscode://schemas/inlineCompletionProviderIdArgs";function RFe(s){const e=[];return s.providerId&&(e.push(s.providerId.toStringWithoutVersion()),e.push(s.providerId.extensionId+":*")),e}const yK=B2e(P2e({showNoResultNotification:hO(mG()),providerId:hO(H2e(NFe,R2e())),explicit:hO(mG())}),M2e());class AFe extends ue{constructor(){super({id:"editor.action.inlineSuggest.trigger",label:V(1185,"Trigger Inline Suggestion"),precondition:N.writable,metadata:{description:m(1172,"Triggers an inline suggestion in the editor."),args:[{name:"args",description:m(1173,"Options for triggering inline suggestions."),isOptional:!0,schema:yK.getJSONSchema()}]}})}async run(e,t,i){const n=e.get(Ii),o=e.get(ge),r=Zo.get(t),a=yK.validateOrThrow(i),l=a?.providerId?o.inlineCompletionsProvider.all(t.getModel()).find(c=>RFe(c).some(d=>d===a.providerId)):void 0;await zhe(async c=>{await r?.model.get()?.trigger(c,{provider:l,explicit:a?.explicit??!0}),r?.playAccessibilitySignal(c)}),a?.showNoResultNotification&&(r?.model.get()?.state.get()||n.notify({severity:oC.Info,message:m(1174,"No inline suggestion is available.")}))}}class MFe extends ue{constructor(){super({id:"editor.action.inlineSuggest.acceptNextWord",label:V(1186,"Accept Next Word Of Inline Suggestion"),precondition:K.and(N.writable,Ot.inlineSuggestionVisible),kbOpts:{weight:101,primary:2065,kbExpr:K.and(N.writable,Ot.inlineSuggestionVisible,Ot.cursorBeforeGhostText,Jv.negate())},menuOpts:[{menuId:fe.InlineSuggestionToolbar,title:m(1175,"Accept Word"),group:"primary",order:2}]})}async run(e,t){await Zo.get(t)?.model.get()?.acceptNextWord()}}class OFe extends ue{constructor(){super({id:"editor.action.inlineSuggest.acceptNextLine",label:V(1187,"Accept Next Line Of Inline Suggestion"),precondition:K.and(N.writable,Ot.inlineSuggestionVisible),kbOpts:{weight:101},menuOpts:[{menuId:fe.InlineSuggestionToolbar,title:m(1176,"Accept Line"),group:"secondary",order:2}]})}async run(e,t){await Zo.get(t)?.model.get()?.acceptNextLine()}}class PFe extends ue{constructor(){super({id:jS,label:V(1188,"Accept Inline Suggestion"),precondition:K.or(Ot.inlineSuggestionVisible,Ot.inlineEditVisible),menuOpts:[{menuId:fe.InlineSuggestionToolbar,title:m(1177,"Accept"),group:"primary",order:2},{menuId:fe.InlineEditsActions,title:m(1178,"Accept"),group:"primary",order:2}],kbOpts:[{primary:2,weight:200,kbExpr:K.or(K.and(Ot.inlineSuggestionVisible,N.tabMovesFocus.toNegated(),je.Visible.toNegated(),N.hoverFocused.toNegated(),Ot.hasSelection.toNegated(),Ot.inlineSuggestionHasIndentationLessThanTabSize),K.and(Ot.inlineEditVisible,N.tabMovesFocus.toNegated(),je.Visible.toNegated(),N.hoverFocused.toNegated(),Ot.tabShouldAcceptInlineEdit))}]})}async run(e,t){const i=Zo.getInFocusedEditorOrParent(e);i&&(i.model.get()?.accept(i.editor),i.editor.focus())}}Dn.registerKeybindingRule({id:jS,weight:202,primary:2,when:K.and(Ot.inInlineEditsPreviewEditor)});class FFe extends ue{constructor(){super({id:EDe,label:V(1189,"Jump to next inline edit"),precondition:Ot.inlineEditVisible,menuOpts:[{menuId:fe.InlineEditsActions,title:m(1179,"Jump"),group:"primary",order:1,when:Ot.cursorAtInlineEdit.toNegated()}],kbOpts:{primary:2,weight:201,kbExpr:K.and(Ot.inlineEditVisible,N.tabMovesFocus.toNegated(),je.Visible.toNegated(),N.hoverFocused.toNegated(),Ot.tabShouldJumpToInlineEdit)}})}async run(e,t){const i=Zo.get(t);i&&i.jump()}}const VN=class VN extends ue{constructor(){super({id:VN.ID,label:V(1190,"Hide Inline Suggestion"),precondition:K.or(Ot.inlineSuggestionVisible,Ot.inlineEditVisible),kbOpts:{weight:190,primary:9},menuOpts:[{menuId:fe.InlineEditsActions,title:m(1180,"Reject"),group:"primary",order:3}]})}async run(e,t){const i=Zo.getInFocusedEditorOrParent(e);Vt(n=>{i?.model.get()?.stop("explicitCancel",n)}),i?.editor.focus()}};VN.ID=Sne;let qD=VN;const zN=class zN extends ue{constructor(){super({id:zN.ID,label:V(1191,"Toggle Inline Suggestions Show Collapsed"),precondition:K.true()})}async run(e,t){const i=e.get(He),n=i.getValue("editor.inlineSuggest.edits.showCollapsed");i.updateValue("editor.inlineSuggest.edits.showCollapsed",!n)}};zN.ID=n5;let e8=zN;Dn.registerKeybindingRule({id:qD.ID,weight:-1,primary:9,secondary:[1033],when:K.and(Ot.inInlineEditsPreviewEditor)});const $N=class $N extends Tn{constructor(){super({id:$N.ID,title:m(1181,"Always Show Toolbar"),f1:!1,precondition:void 0,menu:[{id:fe.InlineSuggestionToolbar,group:"secondary",order:10}],toggled:K.equals("config.editor.inlineSuggest.showToolbar","always")})}async run(e){const t=e.get(He),n=t.getValue("editor.inlineSuggest.showToolbar")==="always"?"onHover":"always";t.updateValue("editor.inlineSuggest.showToolbar",n)}};$N.ID="editor.action.inlineSuggest.toggleAlwaysShowToolbar";let t8=$N;class BFe extends ue{constructor(){super({id:"editor.action.inlineSuggest.dev.extractRepro",label:m(1182,"Developer: Extract Inline Suggest State"),alias:"Developer: Inline Suggest Extract Repro",precondition:K.or(Ot.inlineEditVisible,Ot.inlineSuggestionVisible)})}async run(e,t){const i=e.get(Ko),o=Zo.get(t)?.model.get();if(!o)return;const r=o.extractReproSample(),l=qo(JSON.stringify({inlineCompletion:r.inlineCompletion},null,4)).map(d=>"// "+d).join(`
`),c=`${r.documentValue}

// <json>
${l}
// </json>
`;return await i.writeText(c),{reproCase:c}}}var WFe=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},ZL=function(s,e){return function(t,i){e(t,i,s)}};class HFe{constructor(e,t,i){this.owner=e,this.range=t,this.controller=i}isValidForHoverAnchor(e){return e.type===1&&this.range.startColumn<=e.range.startColumn&&this.range.endColumn>=e.range.endColumn}}let i8=class{constructor(e,t,i,n,o){this._editor=e,this.accessibilityService=t,this._instantiationService=i,this._telemetryService=n,this._markdownRendererService=o,this.hoverOrdinal=4}suggestHoverAnchor(e){const t=Zo.get(this._editor);if(!t)return null;const i=e.target;if(i.type===8){const n=i.detail;if(t.shouldShowHoverAtViewZone(n.viewZoneId))return new g1(1e3,this,k.fromPositions(this._editor.getModel().validatePosition(n.positionBefore||n.position)),e.event.posx,e.event.posy,!1)}if(i.type===7&&t.shouldShowHoverAt(i.range))return new g1(1e3,this,i.range,e.event.posx,e.event.posy,!1);if(i.type===6&&i.detail.mightBeForeignElement&&t.shouldShowHoverAt(i.range))return new g1(1e3,this,i.range,e.event.posx,e.event.posy,!1);if(i.type===9&&i.element){const n=l0.getWarningWidgetContext(i.element);if(n&&t.shouldShowHoverAt(n.range))return new g1(1e3,this,n.range,e.event.posx,e.event.posy,!1)}return null}computeSync(e,t){if(this._editor.getOption(71).showToolbar!=="onHover")return[];const i=Zo.get(this._editor);return i&&i.shouldShowHoverAt(e.range)?[new HFe(this,e.range,i)]:[]}renderHoverParts(e,t){const i=new U,n=t[0];this._telemetryService.publicLog2("inlineCompletionHover.shown"),this.accessibilityService.isScreenReaderOptimized()&&!this._editor.getOption(12)&&i.add(this.renderScreenReaderText(e,n));const o=n.controller.model.get(),r=document.createElement("div");e.fragment.appendChild(r),i.add(ps((l,c)=>{const d=c.add(this._instantiationService.createInstance(xv.hot.read(l),this._editor,!1,zt(null),o.selectedInlineCompletionIndex,o.inlineCompletionsCount,o.activeCommands,o.warning,()=>{e.onContentsChanged()}));r.replaceChildren(d.getDomNode())})),o.triggerExplicitly();const a={hoverPart:n,hoverElement:r,dispose(){i.dispose()}};return new qm([a])}getAccessibleContent(e){return m(1205,"There are inline completions here")}renderScreenReaderText(e,t){const i=new U,n=ie,o=n("div.hover-row.markdown-hover"),r=X(o,n("div.hover-contents",{"aria-live":"assertive"})),a=l=>{const c=m(1206,"Suggestion:"),d=i.add(this._markdownRendererService.render(new hs().appendText(c).appendCodeblock("text",l),{context:this._editor,asyncRenderCallback:()=>{r.className="hover-contents code-hover-contents",e.onContentsChanged()}}));r.replaceChildren(d.element)};return i.add(De(l=>{const c=t.controller.model.read(l)?.primaryGhostText.read(l);if(c){const d=this._editor.getModel().getLineContent(c.lineNumber);a(c.renderForScreenReader(d))}else vn(r)})),e.fragment.appendChild(o),i}};i8=WFe([ZL(1,Wn),ZL(2,_e),ZL(3,vs),ZL(4,Ua)],i8);class VFe{}lt(Zo.ID,T2e(Zo.hot),3);ae(AFe);ae(Q3);ae(J3);ae(MFe);ae(OFe);ae(PFe);ae(e8);ae(qD);ae(FFe);It(t8);ae(BFe);It(C3);It(y3);__.register(i8);UA.register(new VFe);var zFe=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},EO=function(s,e){return function(t,i){e(t,i,s)}},C1,tg;let h0=(tg=class{constructor(e,t,i,n){this.textModelResolverService=t,this.languageService=i,this.languageFeaturesService=n,this.toUnhook=new U,this.toUnhookForKeyboard=new U,this.currentWordAtPosition=null,this.previousPromise=null,this.editor=e,this.linkDecorations=this.editor.createDecorationsCollection();const o=new AA(e);this.toUnhook.add(o),this.toUnhook.add(o.onMouseMoveOrRelevantKeyDown(([r,a])=>{this.startFindDefinitionFromMouse(r,a??void 0)})),this.toUnhook.add(o.onExecute(r=>{this.isEnabled(r)&&this.gotoDefinition(r.target.position,r.hasSideBySideModifier).catch(a=>{Oe(a)}).finally(()=>{this.removeLinkDecorations()})})),this.toUnhook.add(o.onCancel(()=>{this.removeLinkDecorations(),this.currentWordAtPosition=null}))}static get(e){return e.getContribution(C1.ID)}async startFindDefinitionFromCursor(e){await this.startFindDefinition(e),this.toUnhookForKeyboard.add(this.editor.onDidChangeCursorPosition(()=>{this.currentWordAtPosition=null,this.removeLinkDecorations(),this.toUnhookForKeyboard.clear()})),this.toUnhookForKeyboard.add(this.editor.onKeyDown(t=>{t&&(this.currentWordAtPosition=null,this.removeLinkDecorations(),this.toUnhookForKeyboard.clear())}))}startFindDefinitionFromMouse(e,t){if(e.target.type===9&&this.linkDecorations.length>0)return;if(!this.editor.hasModel()||!this.isEnabled(e,t)){this.currentWordAtPosition=null,this.removeLinkDecorations();return}const i=e.target.position;this.startFindDefinition(i)}async startFindDefinition(e){this.toUnhookForKeyboard.clear();const t=e?this.editor.getModel()?.getWordAtPosition(e):null;if(!t){this.currentWordAtPosition=null,this.removeLinkDecorations();return}if(this.currentWordAtPosition&&this.currentWordAtPosition.startColumn===t.startColumn&&this.currentWordAtPosition.endColumn===t.endColumn&&this.currentWordAtPosition.word===t.word)return;this.currentWordAtPosition=t;const i=new $ie(this.editor,15);this.previousPromise&&(this.previousPromise.cancel(),this.previousPromise=null),this.previousPromise=an(r=>this.findDefinition(e,r));let n;try{n=await this.previousPromise}catch(r){Oe(r);return}if(!n||!n.length||!i.validate(this.editor)){this.removeLinkDecorations();return}const o=n[0].originSelectionRange?k.lift(n[0].originSelectionRange):new k(e.lineNumber,t.startColumn,e.lineNumber,t.endColumn);if(n.length>1){let r=o;for(const{originSelectionRange:a}of n)a&&(r=k.plusRange(r,a));this.addDecoration(r,new hs().appendText(m(1077,"Click to show {0} definitions.",n.length)))}else{const r=n[0];return r.uri?this.textModelResolverService.createModelReference(r.uri).then(a=>{if(!a.object||!a.object.textEditorModel){a.dispose();return}const{object:{textEditorModel:l}}=a,{startLineNumber:c}=r.range;if(c<1||c>l.getLineCount()){a.dispose();return}const d=this.getPreviewValue(l,c,r),h=this.languageService.guessLanguageIdByFilepathOrFirstLine(l.uri);this.addDecoration(o,d?new hs().appendCodeblock(h||"",d):void 0),a.dispose()}):void 0}}getPreviewValue(e,t,i){let n=i.range;return n.endLineNumber-n.startLineNumber>=C1.MAX_SOURCE_PREVIEW_LINES&&(n=this.getPreviewRangeBasedOnIndentation(e,t)),n=e.validateRange(n),this.stripIndentationFromPreviewRange(e,t,n)}stripIndentationFromPreviewRange(e,t,i){let o=e.getLineFirstNonWhitespaceColumn(t);for(let a=t+1;a<i.endLineNumber;a++){const l=e.getLineFirstNonWhitespaceColumn(a);o=Math.min(o,l)}return e.getValueInRange(i).replace(new RegExp(`^\\s{${o-1}}`,"gm"),"").trim()}getPreviewRangeBasedOnIndentation(e,t){const i=e.getLineFirstNonWhitespaceColumn(t),n=Math.min(e.getLineCount(),t+C1.MAX_SOURCE_PREVIEW_LINES);let o=t+1;for(;o<n;o++){const r=e.getLineFirstNonWhitespaceColumn(o);if(i===r)break}return new k(t,1,o+1,1)}addDecoration(e,t){const i={range:e,options:{description:"goto-definition-link",inlineClassName:"goto-definition-link",hoverMessage:t}};this.linkDecorations.set([i])}removeLinkDecorations(){this.linkDecorations.clear()}isEnabled(e,t){return this.editor.hasModel()&&e.isLeftClick&&e.isNoneOrSingleMouseDown&&e.target.type===6&&!(e.target.detail.injectedText?.options instanceof mg)&&(e.hasTriggerModifier||(t?t.keyCodeIsTriggerKey:!1))&&this.languageFeaturesService.definitionProvider.has(this.editor.getModel())}findDefinition(e,t){const i=this.editor.getModel();return i?Cx(this.languageFeaturesService.definitionProvider,i,e,!1,t):Promise.resolve(null)}async gotoDefinition(e,t){return this.editor.setPosition(e),this.editor.invokeWithinContext(i=>{const n=!t&&this.editor.getOption(101)&&!this.isInPeekEditor(i);return new Sx({openToSide:t,openInPeek:n,muteMessage:!0},{title:{value:"",original:""},id:"",precondition:void 0}).run(i)})}isInPeekEditor(e){const t=e.get(Ae);return So.inPeekEditor.getValue(t)}dispose(){this.toUnhook.dispose(),this.toUnhookForKeyboard.dispose()}},C1=tg,tg.ID="editor.contrib.gotodefinitionatposition",tg.MAX_SOURCE_PREVIEW_LINES=8,tg);h0=C1=zFe([EO(1,Fs),EO(2,Ei),EO(3,ge)],h0);lt(h0.ID,h0,2);class $Fe extends ue{constructor(){super({id:"editor.action.debugEditorGpuRenderer",label:V(1101,"Developer: Debug Editor GPU Renderer"),precondition:K.true()})}async run(e,t){const i=e.get(_e),o=await e.get(bs).pick([{label:m(1098,"Log Texture Atlas Stats"),id:"logTextureAtlasStats"},{label:m(1099,"Save Texture Atlas"),id:"saveTextureAtlas"},{label:m(1100,"Draw Glyph"),id:"drawGlyph"}],{canPickMany:!1});if(o)switch(o.id){case"logTextureAtlasStats":i.invokeFunction(r=>{const a=r.get(jt),l=Ds.atlas;if(!Ds.atlas){a.error("No texture atlas found");return}const c=l.getStats();a.info(["Texture atlas stats",...c].join(`

`))});break;case"saveTextureAtlas":i.invokeFunction(async r=>{const a=r.get(Sh),l=r.get(iK),c=a.getWorkspace().folders;if(c.length>0){const d=Ds.atlas,h=[];for(const[u,g]of d.pages.entries())h.push(l.writeFile(Le.joinPath(c[0].uri,`textureAtlasPage${u}_actual.png`),Ru.wrap(new Uint8Array(await(await g.source.convertToBlob()).arrayBuffer()))),l.writeFile(Le.joinPath(c[0].uri,`textureAtlasPage${u}_usage.png`),Ru.wrap(new Uint8Array(await(await g.getUsagePreview()).arrayBuffer()))));await Promise.all(h)}});break;case"drawGlyph":i.invokeFunction(async r=>{const a=r.get(He),l=r.get(iK),c=r.get(bs),h=r.get(Sh).getWorkspace().folders;if(h.length===0)return;const u=Ds.atlas,g=a.getValue("editor.fontFamily"),f=a.getValue("editor.fontSize"),p=new CS(f,g,kt().devicePixelRatio,Ds.decorationStyleCache);let _=await c.input({prompt:"Enter a character to draw (prefix with 0x for code point))"});if(!_)return;const b=_.match(/0x(?<codePoint>[0-9a-f]+)/i)?.groups?.codePoint;b!==void 0&&(_=String.fromCodePoint(parseInt(b,16)));const w=0,v=u.getGlyph(p,_,w,0,0);if(!v)return;const y=u.pages[v.pageIndex].source.getContext("2d")?.getImageData(v.x,v.y,v.w,v.h);if(!y)return;const x=new OffscreenCanvas(y.width,y.height);Tm(x.getContext("2d")).putImageData(y,0,0);const E=await x.convertToBlob({type:"image/png"}),I=Le.joinPath(h[0].uri,`glyph_${_}_${w}_${f}px_${g.replaceAll(/[,\\\/\.'\s]/g,"_")}.png`);await l.writeFile(I,Ru.wrap(new Uint8Array(await E.arrayBuffer())))});break}}}ae($Fe);var Ja;(function(s){s.NoAutoFocus="noAutoFocus",s.FocusIfVisible="focusIfVisible",s.AutoFocusImmediately="autoFocusImmediately"})(Ja||(Ja={}));class UFe extends ue{constructor(){super({id:vne,label:V(1107,"Show or Focus Hover"),metadata:{description:V(1108,"Show or focus the editor hover which shows documentation, references, and other content for a symbol at the current cursor position."),args:[{name:"args",schema:{type:"object",properties:{focus:{description:"Controls if and when the hover should take focus upon being triggered by this action.",enum:[Ja.NoAutoFocus,Ja.FocusIfVisible,Ja.AutoFocusImmediately],enumDescriptions:[m(1104,"The hover will not automatically take focus."),m(1105,"The hover will take focus only if it is already visible."),m(1106,"The hover will automatically take focus when it appears.")],default:Ja.FocusIfVisible}}}}]},precondition:void 0,kbOpts:{kbExpr:N.editorTextFocus,primary:Gi(2089,2087),weight:100}})}run(e,t,i){if(!t.hasModel())return;const n=ws.get(t);if(!n)return;const o=i?.focus;let r=Ja.FocusIfVisible;Object.values(Ja).includes(o)?r=o:typeof o=="boolean"&&o&&(r=Ja.AutoFocusImmediately);const a=c=>{const d=t.getPosition(),h=new k(d.lineNumber,d.column,d.lineNumber,d.column);n.showContentHover(h,1,2,c)},l=t.getOption(2)===2;n.isHoverVisible?r!==Ja.NoAutoFocus?n.focus():a(l):a(l||r===Ja.AutoFocusImmediately)}}class qFe extends ue{constructor(){super({id:pDe,label:V(1109,"Show Definition Preview Hover"),precondition:void 0,metadata:{description:V(1110,"Show the definition preview hover in the editor.")}})}run(e,t){const i=ws.get(t);if(!i)return;const n=t.getPosition();if(!n)return;const o=new k(n.lineNumber,n.column,n.lineNumber,n.column),r=h0.get(t);if(!r)return;r.startFindDefinitionFromCursor(n).then(()=>{i.showContentHover(o,1,2,!0)})}}class jFe extends ue{constructor(){super({id:mDe,label:V(1111,"Hide Hover"),alias:"Hide Content Hover",precondition:void 0})}run(e,t){ws.get(t)?.hideContentHover()}}class GFe extends ue{constructor(){super({id:_De,label:V(1112,"Scroll Up Hover"),precondition:N.hoverFocused,kbOpts:{kbExpr:N.hoverFocused,primary:16,weight:100},metadata:{description:V(1113,"Scroll up the editor hover.")}})}run(e,t){const i=ws.get(t);i&&i.scrollUp()}}class KFe extends ue{constructor(){super({id:bDe,label:V(1114,"Scroll Down Hover"),precondition:N.hoverFocused,kbOpts:{kbExpr:N.hoverFocused,primary:18,weight:100},metadata:{description:V(1115,"Scroll down the editor hover.")}})}run(e,t){const i=ws.get(t);i&&i.scrollDown()}}class YFe extends ue{constructor(){super({id:wDe,label:V(1116,"Scroll Left Hover"),precondition:N.hoverFocused,kbOpts:{kbExpr:N.hoverFocused,primary:15,weight:100},metadata:{description:V(1117,"Scroll left the editor hover.")}})}run(e,t){const i=ws.get(t);i&&i.scrollLeft()}}class ZFe extends ue{constructor(){super({id:vDe,label:V(1118,"Scroll Right Hover"),precondition:N.hoverFocused,kbOpts:{kbExpr:N.hoverFocused,primary:17,weight:100},metadata:{description:V(1119,"Scroll right the editor hover.")}})}run(e,t){const i=ws.get(t);i&&i.scrollRight()}}class XFe extends ue{constructor(){super({id:CDe,label:V(1120,"Page Up Hover"),precondition:N.hoverFocused,kbOpts:{kbExpr:N.hoverFocused,primary:11,secondary:[528],weight:100},metadata:{description:V(1121,"Page up the editor hover.")}})}run(e,t){const i=ws.get(t);i&&i.pageUp()}}class QFe extends ue{constructor(){super({id:yDe,label:V(1122,"Page Down Hover"),precondition:N.hoverFocused,kbOpts:{kbExpr:N.hoverFocused,primary:12,secondary:[530],weight:100},metadata:{description:V(1123,"Page down the editor hover.")}})}run(e,t){const i=ws.get(t);i&&i.pageDown()}}class JFe extends ue{constructor(){super({id:SDe,label:V(1124,"Go To Top Hover"),precondition:N.hoverFocused,kbOpts:{kbExpr:N.hoverFocused,primary:14,secondary:[2064],weight:100},metadata:{description:V(1125,"Go to the top of the editor hover.")}})}run(e,t){const i=ws.get(t);i&&i.goToTop()}}class e4e extends ue{constructor(){super({id:xDe,label:V(1126,"Go To Bottom Hover"),precondition:N.hoverFocused,kbOpts:{kbExpr:N.hoverFocused,primary:13,secondary:[2066],weight:100},metadata:{description:V(1127,"Go to the bottom of the editor hover.")}})}run(e,t){const i=ws.get(t);i&&i.goToBottom()}}class t4e extends ue{constructor(){super({id:IA,label:LDe,alias:"Increase Hover Verbosity Level",precondition:N.hoverVisible})}run(e,t,i){const n=ws.get(t);if(!n)return;const o=i?.index!==void 0?i.index:n.focusedHoverPartIndex();n.updateHoverVerbosityLevel(Ao.Increase,o,i?.focus)}}class i4e extends ue{constructor(){super({id:DA,label:kDe,alias:"Decrease Hover Verbosity Level",precondition:N.hoverVisible})}run(e,t,i){const n=ws.get(t);if(!n)return;const o=i?.index!==void 0?i.index:n.focusedHoverPartIndex();ws.get(t)?.updateHoverVerbosityLevel(Ao.Decrease,o,i?.focus)}}class n4e{constructor(e){this._editor=e}computeSync(e){const t=r=>({value:r}),i=this._editor.getLineDecorations(e.lineNumber),n=[],o=e.laneOrLine==="lineNo";if(!i)return n;for(const r of i){const a=r.options.glyphMargin?.position??Ll.Center;if(!o&&a!==e.laneOrLine)continue;const l=o?r.options.lineNumberHoverMessage:r.options.glyphMarginHoverMessage;!l||wv(l)||n.push(...tW(l).map(t))}return n}}var s4e=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},o4e=function(s,e){return function(t,i){e(t,i,s)}},n8;const SK=ie;var im;let s8=(im=class extends F{constructor(e,t){super(),this._markdownRendererService=t,this.allowEditorOverflow=!0,this._renderDisposeables=this._register(new U),this._editor=e,this._isVisible=!1,this._messages=[],this._hover=this._register(new aH(!0)),this._hover.containerDomNode.classList.toggle("hidden",!this._isVisible),this._hoverOperation=this._register(new xne(this._editor,new n4e(this._editor))),this._register(this._hoverOperation.onResult(i=>this._withResult(i))),this._register(this._editor.onDidChangeModelDecorations(()=>this._onModelDecorationsChanged())),this._register(this._editor.onDidChangeConfiguration(i=>{i.hasChanged(59)&&this._updateFont()})),this._register(Fi(this._hover.containerDomNode,"mouseleave",i=>{this._onMouseLeave(i)})),this._editor.addOverlayWidget(this)}dispose(){this._hoverComputerOptions=void 0,this._editor.removeOverlayWidget(this),super.dispose()}getId(){return n8.ID}getDomNode(){return this._hover.containerDomNode}getPosition(){return null}_updateFont(){Array.prototype.slice.call(this._hover.contentsDomNode.getElementsByClassName("code")).forEach(t=>this._editor.applyFontInfo(t))}_onModelDecorationsChanged(){this._isVisible&&this._hoverComputerOptions&&(this._hoverOperation.cancel(),this._hoverOperation.start(0,this._hoverComputerOptions))}showsOrWillShow(e){const t=e.target;return t.type===2&&t.detail.glyphMarginLane?(this._startShowingAt(t.position.lineNumber,t.detail.glyphMarginLane),!0):t.type===3?(this._startShowingAt(t.position.lineNumber,"lineNo"),!0):!1}_startShowingAt(e,t){this._hoverComputerOptions&&this._hoverComputerOptions.lineNumber===e&&this._hoverComputerOptions.laneOrLine===t||(this._hoverOperation.cancel(),this.hide(),this._hoverComputerOptions={lineNumber:e,laneOrLine:t},this._hoverOperation.start(0,this._hoverComputerOptions))}hide(){this._hoverComputerOptions=void 0,this._hoverOperation.cancel(),this._isVisible&&(this._isVisible=!1,this._hover.containerDomNode.classList.toggle("hidden",!this._isVisible))}_withResult(e){this._messages=e.value,this._messages.length>0?this._renderMessages(e.options.lineNumber,e.options.laneOrLine,this._messages):this.hide()}_renderMessages(e,t,i){this._renderDisposeables.clear();const n=document.createDocumentFragment();for(const o of i){const r=SK("div.hover-row.markdown-hover"),a=X(r,SK("div.hover-contents")),l=this._renderDisposeables.add(this._markdownRendererService.render(o.value,{context:this._editor}));a.appendChild(l.element),n.appendChild(r)}this._updateContents(n),this._showAt(e,t)}_updateContents(e){this._hover.contentsDomNode.textContent="",this._hover.contentsDomNode.appendChild(e),this._updateFont()}_showAt(e,t){this._isVisible||(this._isVisible=!0,this._hover.containerDomNode.classList.toggle("hidden",!this._isVisible));const i=this._editor.getLayoutInfo(),n=this._editor.getTopForLineNumber(e),o=this._editor.getScrollTop(),r=this._editor.getOption(75),a=this._hover.containerDomNode.clientHeight,l=n-o-(a-r)/2,c=i.glyphMarginLeft+i.glyphMarginWidth+(t==="lineNo"?i.lineNumbersWidth:0),h=i.height-a,u=Math.max(0,Math.min(Math.round(l),h));if(this._editor.getOption(51)){const f=this._editor.getDomNode();if(f){const p=Li(f);this._hover.containerDomNode.style.position="fixed",this._hover.containerDomNode.style.left=`${p.left+c}px`,this._hover.containerDomNode.style.top=`${p.top+u}px`}}else this._hover.containerDomNode.style.position="absolute",this._hover.containerDomNode.style.left=`${c}px`,this._hover.containerDomNode.style.top=`${u}px`;this._hover.containerDomNode.style.zIndex="11"}_onMouseLeave(e){const t=this._editor.getDomNode();(!t||!TA(t,e.x,e.y))&&this.hide()}},n8=im,im.ID="editor.contrib.modesGlyphHoverWidget",im);s8=n8=s4e([o4e(1,Ua)],s8);var r4e=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},a4e=function(s,e){return function(t,i){e(t,i,s)}},Gw;let jD=(Gw=class extends F{constructor(e,t){super(),this._editor=e,this._instantiationService=t,this.shouldKeepOpenOnEditorMouseMoveOrLeave=!1,this._listenersStore=new U,this._hoverState={mouseDown:!1},this._reactToEditorMouseMoveRunner=this._register(new Rt(()=>this._reactToEditorMouseMove(this._mouseMoveEvent),0)),this._hookListeners(),this._register(this._editor.onDidChangeConfiguration(i=>{i.hasChanged(69)&&(this._unhookListeners(),this._hookListeners())}))}_hookListeners(){const e=this._editor.getOption(69);this._hoverSettings={enabled:e.enabled,sticky:e.sticky,hidingDelay:e.hidingDelay},e.enabled?(this._listenersStore.add(this._editor.onMouseDown(t=>this._onEditorMouseDown(t))),this._listenersStore.add(this._editor.onMouseUp(()=>this._onEditorMouseUp())),this._listenersStore.add(this._editor.onMouseMove(t=>this._onEditorMouseMove(t))),this._listenersStore.add(this._editor.onKeyDown(t=>this._onKeyDown(t)))):(this._listenersStore.add(this._editor.onMouseMove(t=>this._onEditorMouseMove(t))),this._listenersStore.add(this._editor.onKeyDown(t=>this._onKeyDown(t)))),this._listenersStore.add(this._editor.onMouseLeave(t=>this._onEditorMouseLeave(t))),this._listenersStore.add(this._editor.onDidChangeModel(()=>{this._cancelScheduler(),this.hideGlyphHover()})),this._listenersStore.add(this._editor.onDidChangeModelContent(()=>this._cancelScheduler())),this._listenersStore.add(this._editor.onDidScrollChange(t=>this._onEditorScrollChanged(t)))}_unhookListeners(){this._listenersStore.clear()}_cancelScheduler(){this._mouseMoveEvent=void 0,this._reactToEditorMouseMoveRunner.cancel()}_onEditorScrollChanged(e){(e.scrollTopChanged||e.scrollLeftChanged)&&this.hideGlyphHover()}_onEditorMouseDown(e){this._hoverState.mouseDown=!0,!this._isMouseOnGlyphHoverWidget(e)&&this.hideGlyphHover()}_isMouseOnGlyphHoverWidget(e){const t=this._glyphWidget?.getDomNode();return t?TA(t,e.event.posx,e.event.posy):!1}_onEditorMouseUp(){this._hoverState.mouseDown=!1}_onEditorMouseLeave(e){this.shouldKeepOpenOnEditorMouseMoveOrLeave||(this._cancelScheduler(),this._isMouseOnGlyphHoverWidget(e))||this.hideGlyphHover()}_shouldNotRecomputeCurrentHoverWidget(e){const t=this._hoverSettings.sticky,i=this._isMouseOnGlyphHoverWidget(e);return t&&i}_onEditorMouseMove(e){if(this.shouldKeepOpenOnEditorMouseMoveOrLeave)return;if(this._mouseMoveEvent=e,this._shouldNotRecomputeCurrentHoverWidget(e)){this._reactToEditorMouseMoveRunner.cancel();return}this._reactToEditorMouseMove(e)}_reactToEditorMouseMove(e){!e||this._tryShowHoverWidget(e)||this.hideGlyphHover()}_tryShowHoverWidget(e){return this._getOrCreateGlyphWidget().showsOrWillShow(e)}_onKeyDown(e){this._editor.hasModel()&&(e.keyCode===5||e.keyCode===6||e.keyCode===57||e.keyCode===4||this.hideGlyphHover())}hideGlyphHover(){this._glyphWidget?.hide()}_getOrCreateGlyphWidget(){return this._glyphWidget||(this._glyphWidget=this._instantiationService.createInstance(s8,this._editor)),this._glyphWidget}dispose(){super.dispose(),this._unhookListeners(),this._listenersStore.dispose(),this._glyphWidget?.dispose()}},Gw.ID="editor.contrib.marginHover",Gw);jD=r4e([a4e(1,_e)],jD);class l4e{}class c4e{}class d4e{}lt(ws.ID,ws,2);lt(jD.ID,jD,2);ae(UFe);ae(qFe);ae(jFe);ae(GFe);ae(KFe);ae(YFe);ae(ZFe);ae(XFe);ae(QFe);ae(JFe);ae(e4e);ae(t4e);ae(i4e);__.register(KS);__.register(H5);ua((s,e)=>{const t=s.getColor(RW);t&&(e.addRule(`.monaco-editor .monaco-hover .hover-row:not(:first-child):not(:empty) { border-top: 1px solid ${t.transparent(.5)}; }`),e.addRule(`.monaco-editor .monaco-hover hr { border-top: 1px solid ${t.transparent(.5)}; }`),e.addRule(`.monaco-editor .monaco-hover hr { border-bottom: 0px solid ${t.transparent(.5)}; }`))});UA.register(new l4e);UA.register(new c4e);UA.register(new d4e);function Roe(s,e,t,i){if(s.getLineCount()===1&&s.getLineMaxColumn(1)===1)return[];const n=e.getLanguageConfiguration(s.getLanguageId()).indentRulesSupport;if(!n)return[];const o=new CW(s,n,e);for(i=Math.min(i,s.getLineCount());t<=i&&o.shouldIgnore(t);)t++;if(t>i-1)return[];const{tabSize:r,indentSize:a,insertSpaces:l}=s.getOptions(),c=(p,_)=>(_=_||1,la.shiftIndent(p,p.length+_,r,a,l)),d=(p,_)=>(_=_||1,la.unshiftIndent(p,p.length+_,r,a,l)),h=[],u=s.getLineContent(t);let g=Bt(u),f=g;o.shouldIncrease(t)?(f=c(f),g=c(g)):o.shouldIndentNextLine(t)&&(f=c(f)),t++;for(let p=t;p<=i;p++){if(h4e(s,p))continue;const _=s.getLineContent(p),b=Bt(_),w=f;o.shouldDecrease(p,w)&&(f=d(f),g=d(g)),b!==f&&h.push(Si.replaceMove(new he(p,1,p,b.length+1),fW(f,a,l))),!o.shouldIgnore(p)&&(o.shouldIncrease(p,w)?(g=c(g),f=g):o.shouldIndentNextLine(p,w)?f=c(f):f=g)}return h}function h4e(s,e){return s.tokenization.isCheapToTokenize(e)?s.tokenization.getLineTokens(e).getStandardTokenType(0)===2:!1}var u4e=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},g4e=function(s,e){return function(t,i){e(t,i,s)}};const UN=class UN extends ue{constructor(){super({id:UN.ID,label:V(1148,"Convert Indentation to Spaces"),precondition:N.writable,metadata:{description:V(1149,"Convert the tab indentation to spaces.")}})}run(e,t){const i=t.getModel();if(!i)return;const n=i.getOptions(),o=t.getSelection();if(!o)return;const r=new b4e(o,n.tabSize);t.pushUndoStop(),t.executeCommands(this.id,[r]),t.pushUndoStop(),i.updateOptions({insertSpaces:!0})}};UN.ID="editor.action.indentationToSpaces";let o8=UN;const qN=class qN extends ue{constructor(){super({id:qN.ID,label:V(1150,"Convert Indentation to Tabs"),precondition:N.writable,metadata:{description:V(1151,"Convert the spaces indentation to tabs.")}})}run(e,t){const i=t.getModel();if(!i)return;const n=i.getOptions(),o=t.getSelection();if(!o)return;const r=new w4e(o,n.tabSize);t.pushUndoStop(),t.executeCommands(this.id,[r]),t.pushUndoStop(),i.updateOptions({insertSpaces:!1})}};qN.ID="editor.action.indentationToTabs";let r8=qN;class JH extends ue{constructor(e,t,i){super(i),this.insertSpaces=e,this.displaySizeOnly=t}run(e,t){const i=e.get(bs),n=e.get(hi),o=t.getModel();if(!o)return;const r=n.getCreationOptions(o.getLanguageId(),o.uri,o.isForSimpleWidget),a=o.getOptions(),l=[1,2,3,4,5,6,7,8].map(d=>({id:d.toString(),label:d.toString(),description:d===r.tabSize&&d===a.tabSize?m(1144,"Configured Tab Size"):d===r.tabSize?m(1145,"Default Tab Size"):d===a.tabSize?m(1146,"Current Tab Size"):void 0})),c=Math.min(o.getOptions().tabSize-1,7);setTimeout(()=>{i.pick(l,{placeHolder:m(1147,"Select Tab Size for Current File"),activeItem:l[c]}).then(d=>{if(d&&o&&!o.isDisposed()){const h=parseInt(d.label,10);this.displaySizeOnly?o.updateOptions({tabSize:h}):o.updateOptions({tabSize:h,indentSize:h,insertSpaces:this.insertSpaces})}})},50)}}const jN=class jN extends JH{constructor(){super(!1,!1,{id:jN.ID,label:V(1152,"Indent Using Tabs"),precondition:void 0,metadata:{description:V(1153,"Use indentation with tabs.")}})}};jN.ID="editor.action.indentUsingTabs";let a8=jN;const GN=class GN extends JH{constructor(){super(!0,!1,{id:GN.ID,label:V(1154,"Indent Using Spaces"),precondition:void 0,metadata:{description:V(1155,"Use indentation with spaces.")}})}};GN.ID="editor.action.indentUsingSpaces";let l8=GN;const KN=class KN extends JH{constructor(){super(!0,!0,{id:KN.ID,label:V(1156,"Change Tab Display Size"),precondition:void 0,metadata:{description:V(1157,"Change the space size equivalent of the tab.")}})}};KN.ID="editor.action.changeTabDisplaySize";let c8=KN;const YN=class YN extends ue{constructor(){super({id:YN.ID,label:V(1158,"Detect Indentation from Content"),precondition:void 0,metadata:{description:V(1159,"Detect the indentation from content.")}})}run(e,t){const i=e.get(hi),n=t.getModel();if(!n)return;const o=i.getCreationOptions(n.getLanguageId(),n.uri,n.isForSimpleWidget);n.detectIndentation(o.insertSpaces,o.tabSize)}};YN.ID="editor.action.detectIndentation";let d8=YN;class f4e extends ue{constructor(){super({id:"editor.action.reindentlines",label:V(1160,"Reindent Lines"),precondition:N.writable,metadata:{description:V(1161,"Reindent the lines of the editor.")},canTriggerInlineEdits:!0})}run(e,t){const i=e.get(ui),n=t.getModel();if(!n)return;const o=Roe(n,i,1,n.getLineCount());o.length>0&&(t.pushUndoStop(),t.executeEdits(this.id,o),t.pushUndoStop())}}class p4e extends ue{constructor(){super({id:"editor.action.reindentselectedlines",label:V(1162,"Reindent Selected Lines"),precondition:N.writable,metadata:{description:V(1163,"Reindent the selected lines of the editor.")},canTriggerInlineEdits:!0})}run(e,t){const i=e.get(ui),n=t.getModel();if(!n)return;const o=t.getSelections();if(o===null)return;const r=[];for(const a of o){let l=a.startLineNumber,c=a.endLineNumber;if(l!==c&&a.endColumn===1&&c--,l===1){if(l===c)continue}else l--;const d=Roe(n,i,l,c);r.push(...d)}r.length>0&&(t.pushUndoStop(),t.executeEdits(this.id,r),t.pushUndoStop())}}class m4e{constructor(e,t){this._initialSelection=t,this._edits=[],this._selectionId=null;for(const i of e)i.range&&typeof i.text=="string"&&this._edits.push(i)}getEditOperations(e,t){for(const n of this._edits)t.addEditOperation(k.lift(n.range),n.text);let i=!1;Array.isArray(this._edits)&&this._edits.length===1&&this._initialSelection.isEmpty()&&(this._edits[0].range.startColumn===this._initialSelection.endColumn&&this._edits[0].range.startLineNumber===this._initialSelection.endLineNumber?(i=!0,this._selectionId=t.trackSelection(this._initialSelection,!0)):this._edits[0].range.endColumn===this._initialSelection.startColumn&&this._edits[0].range.endLineNumber===this._initialSelection.startLineNumber&&(i=!0,this._selectionId=t.trackSelection(this._initialSelection,!1))),i||(this._selectionId=t.trackSelection(this._initialSelection))}computeCursorState(e,t){return t.getTrackedSelection(this._selectionId)}}var Kw;let GD=(Kw=class{constructor(e,t){this.editor=e,this._languageConfigurationService=t,this.callOnDispose=new U,this.callOnModel=new U,this.callOnDispose.add(e.onDidChangeConfiguration(()=>this.update())),this.callOnDispose.add(e.onDidChangeModel(()=>this.update())),this.callOnDispose.add(e.onDidChangeModelLanguage(()=>this.update()))}update(){this.callOnModel.clear(),!(!this.editor.getOption(17)||this.editor.getOption(16)<4)&&this.editor.hasModel()&&this.callOnModel.add(this.editor.onDidPaste(({range:e})=>{this.trigger(e)}))}trigger(e){const t=this.editor.getSelections();if(t===null||t.length>1)return;const i=this.editor.getModel();if(!i||this.rangeContainsOnlyWhitespaceCharacters(i,e)||!this.editor.getOption(18)&&_4e(i,e)||!i.tokenization.isCheapToTokenize(e.getStartPosition().lineNumber))return;const o=this.editor.getOption(16),{tabSize:r,indentSize:a,insertSpaces:l}=i.getOptions(),c=[],d={shiftIndent:f=>la.shiftIndent(f,f.length+1,r,a,l),unshiftIndent:f=>la.unshiftIndent(f,f.length+1,r,a,l)};let h=e.startLineNumber,u=i.getLineContent(h);if(!/\S/.test(u.substring(0,e.startColumn-1))){const f=D1(o,i,i.getLanguageId(),h,d,this._languageConfigurationService);if(f!==null){const p=Bt(u),_=To(f,r),b=To(p,r);if(_!==b){const w=ey(_,r,l);c.push({range:new k(h,1,h,p.length+1),text:w}),u=w+u.substring(p.length)}else{const w=_J(i,h,this._languageConfigurationService);if(w===0||w===8)return}}}const g=h;for(;h<e.endLineNumber;){if(!/\S/.test(i.getLineContent(h+1))){h++;continue}break}if(h!==e.endLineNumber){const p=D1(o,{tokenization:{getLineTokens:_=>i.tokenization.getLineTokens(_),getLanguageId:()=>i.getLanguageId(),getLanguageIdAtPosition:(_,b)=>i.getLanguageIdAtPosition(_,b)},getLineContent:_=>_===g?u:i.getLineContent(_)},i.getLanguageId(),h+1,d,this._languageConfigurationService);if(p!==null){const _=To(p,r),b=To(Bt(i.getLineContent(h+1)),r);if(_!==b){const w=_-b;for(let C=h+1;C<=e.endLineNumber;C++){const v=i.getLineContent(C),y=Bt(v),L=To(y,r)+w,E=ey(L,r,l);E!==y&&c.push({range:new k(C,1,C,y.length+1),text:E})}}}}if(c.length>0){this.editor.pushUndoStop();const f=new m4e(c,this.editor.getSelection());this.editor.executeCommand("autoIndentOnPaste",f),this.editor.pushUndoStop()}}rangeContainsOnlyWhitespaceCharacters(e,t){const i=o=>o.trim().length===0;let n=!0;if(t.startLineNumber===t.endLineNumber){const r=e.getLineContent(t.startLineNumber).substring(t.startColumn-1,t.endColumn-1);n=i(r)}else for(let o=t.startLineNumber;o<=t.endLineNumber;o++){const r=e.getLineContent(o);if(o===t.startLineNumber){const a=r.substring(t.startColumn-1);n=i(a)}else if(o===t.endLineNumber){const a=r.substring(0,t.endColumn-1);n=i(a)}else n=e.getLineFirstNonWhitespaceColumn(o)===0;if(!n)break}return n}dispose(){this.callOnDispose.dispose(),this.callOnModel.dispose()}},Kw.ID="editor.contrib.autoIndentOnPaste",Kw);GD=u4e([g4e(1,ui)],GD);function _4e(s,e){const t=i=>afe(s,i)===2;return t(e.getStartPosition())||t(e.getEndPosition())}function Aoe(s,e,t,i){if(s.getLineCount()===1&&s.getLineMaxColumn(1)===1)return;let n="";for(let r=0;r<t;r++)n+=" ";const o=new RegExp(n,"gi");for(let r=1,a=s.getLineCount();r<=a;r++){let l=s.getLineFirstNonWhitespaceColumn(r);if(l===0&&(l=s.getLineMaxColumn(r)),l===1)continue;const c=new k(r,1,r,l),d=s.getValueInRange(c),h=i?d.replace(/\t/ig,n):d.replace(o,"	");e.addEditOperation(c,h)}}class b4e{constructor(e,t){this.selection=e,this.tabSize=t,this.selectionId=null}getEditOperations(e,t){this.selectionId=t.trackSelection(this.selection),Aoe(e,t,this.tabSize,!0)}computeCursorState(e,t){return t.getTrackedSelection(this.selectionId)}}class w4e{constructor(e,t){this.selection=e,this.tabSize=t,this.selectionId=null}getEditOperations(e,t){this.selectionId=t.trackSelection(this.selection),Aoe(e,t,this.tabSize,!1)}computeCursorState(e,t){return t.getTrackedSelection(this.selectionId)}}lt(GD.ID,GD,2);ae(o8);ae(r8);ae(a8);ae(l8);ae(c8);ae(d8);ae(f4e);ae(p4e);lt(JS.ID,JS,1);__.register(uD);class v4e{constructor(e,t,i){this._editRange=e,this._originalSelection=t,this._text=i}getEditOperations(e,t){t.addTrackedEditOperation(this._editRange,this._text)}computeCursorState(e,t){const n=t.getInverseEditOperations()[0].range;return this._originalSelection.isEmpty()?new he(n.endLineNumber,Math.min(this._originalSelection.positionColumn,n.endColumn),n.endLineNumber,Math.min(this._originalSelection.positionColumn,n.endColumn)):new he(n.endLineNumber,n.endColumn-this._text.length,n.endLineNumber,n.endColumn)}}var C4e=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},y4e=function(s,e){return function(t,i){e(t,i,s)}},Xk,ig;let Rv=(ig=class{static get(e){return e.getContribution(Xk.ID)}constructor(e,t){this.editor=e,this.editorWorkerService=t,this.decorations=this.editor.createDecorationsCollection()}dispose(){}run(e,t){this.currentRequest?.cancel();const i=this.editor.getSelection(),n=this.editor.getModel();if(!n||!i)return;let o=i;if(o.startLineNumber!==o.endLineNumber)return;const r=new $ie(this.editor,5),a=n.uri;return this.editorWorkerService.canNavigateValueSet(a)?(this.currentRequest=an(l=>this.editorWorkerService.navigateValueSet(a,o,t)),this.currentRequest.then(l=>{if(!l||!l.range||!l.value||!r.validate(this.editor))return;const c=k.lift(l.range);let d=l.range;const h=l.value.length-(o.endColumn-o.startColumn);d={startLineNumber:d.startLineNumber,startColumn:d.startColumn,endLineNumber:d.endLineNumber,endColumn:d.startColumn+l.value.length},h>1&&(o=new he(o.startLineNumber,o.startColumn,o.endLineNumber,o.endColumn+h-1));const u=new v4e(c,o,l.value);this.editor.pushUndoStop(),this.editor.executeCommand(e,u),this.editor.pushUndoStop(),this.decorations.set([{range:d,options:Xk.DECORATION}]),this.decorationRemover?.cancel(),this.decorationRemover=Tc(350),this.decorationRemover.then(()=>this.decorations.clear()).catch(Oe)}).catch(Oe)):Promise.resolve(void 0)}},Xk=ig,ig.ID="editor.contrib.inPlaceReplaceController",ig.DECORATION=Fe.register({description:"in-place-replace",className:"valueSetReplacement"}),ig);Rv=Xk=C4e([y4e(1,ko)],Rv);class S4e extends ue{constructor(){super({id:"editor.action.inPlaceReplace.up",label:V(1240,"Replace with Previous Value"),precondition:N.writable,kbOpts:{kbExpr:N.editorTextFocus,primary:3159,weight:100}})}run(e,t){const i=Rv.get(t);return i?i.run(this.id,!1):Promise.resolve(void 0)}}class x4e extends ue{constructor(){super({id:"editor.action.inPlaceReplace.down",label:V(1241,"Replace with Next Value"),precondition:N.writable,kbOpts:{kbExpr:N.editorTextFocus,primary:3161,weight:100}})}run(e,t){const i=Rv.get(t);return i?i.run(this.id,!0):Promise.resolve(void 0)}}lt(Rv.ID,Rv,4);ae(S4e);ae(x4e);class L4e{constructor(e){this._selection=e,this._selectionId=null}getEditOperations(e,t){const i=k4e(e);i&&t.addEditOperation(i.range,i.text),this._selectionId=t.trackSelection(this._selection)}computeCursorState(e,t){return t.getTrackedSelection(this._selectionId)}}function k4e(s){const e=s.getLineCount(),t=s.getLineContent(e),i=Ma(t)===-1;if(!(!e||i))return Si.insert(new A(e,s.getLineMaxColumn(e)),s.getEOL())}const ZN=class ZN extends ue{constructor(){super({id:ZN.ID,label:V(1242,"Insert Final New Line"),precondition:N.writable})}run(e,t,i){const n=t.getSelection();if(n===null)return;const o=new L4e(n);t.pushUndoStop(),t.executeCommands(this.id,[o]),t.pushUndoStop()}};ZN.ID="editor.action.insertFinalNewLine";let h8=ZN;ae(h8);class E4e extends ue{constructor(){super({id:"expandLineSelection",label:V(1243,"Expand Line Selection"),precondition:void 0,kbOpts:{weight:0,kbExpr:N.textInputFocus,primary:2090}})}run(e,t,i){if(i=i||{},!t.hasModel())return;const n=t._getViewModel();n.model.pushStackElement(),n.setCursorStates(i.source,3,Rn.expandLineSelection(n,n.getCursorStates())),n.revealAllCursors(i.source,!0)}}ae(E4e);var I4e=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},XL=function(s,e){return function(t,i){e(t,i,s)}},Qk;const Moe=new ce("LinkedEditingInputVisible",!1),D4e="linked-editing-decoration";var ng;let Av=(ng=class extends F{static get(e){return e.getContribution(Qk.ID)}constructor(e,t,i,n,o){super(),this.languageConfigurationService=n,this._syncRangesToken=0,this._localToDispose=this._register(new U),this._editor=e,this._providers=i.linkedEditingRangeProvider,this._enabled=!1,this._visibleContextKey=Moe.bindTo(t),this._debounceInformation=o.for(this._providers,"Linked Editing",{max:200}),this._currentDecorations=this._editor.createDecorationsCollection(),this._languageWordPattern=null,this._currentWordPattern=null,this._ignoreChangeEvent=!1,this._localToDispose=this._register(new U),this._rangeUpdateTriggerPromise=null,this._rangeSyncTriggerPromise=null,this._currentRequestCts=null,this._currentRequestPosition=null,this._currentRequestModelVersion=null,this._register(this._editor.onDidChangeModel(()=>this.reinitialize(!0))),this._register(this._editor.onDidChangeConfiguration(r=>{(r.hasChanged(78)||r.hasChanged(106))&&this.reinitialize(!1)})),this._register(this._providers.onDidChange(()=>this.reinitialize(!1))),this._register(this._editor.onDidChangeModelLanguage(()=>this.reinitialize(!0))),this.reinitialize(!0)}reinitialize(e){const t=this._editor.getModel(),i=t!==null&&(this._editor.getOption(78)||this._editor.getOption(106))&&this._providers.has(t);if(i===this._enabled&&!e||(this._enabled=i,this.clearRanges(),this._localToDispose.clear(),!i||t===null))return;this._localToDispose.add(oe.runAndSubscribe(t.onDidChangeLanguageConfiguration,()=>{this._languageWordPattern=this.languageConfigurationService.getLanguageConfiguration(t.getLanguageId()).getWordDefinition()}));const n=new sa(this._debounceInformation.get(t)),o=()=>{this._rangeUpdateTriggerPromise=n.trigger(()=>this.updateRanges(),this._debounceDuration??this._debounceInformation.get(t))},r=new sa(0),a=l=>{this._rangeSyncTriggerPromise=r.trigger(()=>this._syncRanges(l))};this._localToDispose.add(this._editor.onDidChangeCursorPosition(()=>{o()})),this._localToDispose.add(this._editor.onDidChangeModelContent(l=>{if(!this._ignoreChangeEvent&&this._currentDecorations.length>0){const c=this._currentDecorations.getRange(0);if(c&&l.changes.every(d=>c.intersectRanges(d.range))){a(this._syncRangesToken);return}}o()})),this._localToDispose.add({dispose:()=>{n.dispose(),r.dispose()}}),this.updateRanges()}_syncRanges(e){if(!this._editor.hasModel()||e!==this._syncRangesToken||this._currentDecorations.length===0)return;const t=this._editor.getModel(),i=this._currentDecorations.getRange(0);if(!i||i.startLineNumber!==i.endLineNumber)return this.clearRanges();const n=t.getValueInRange(i);if(this._currentWordPattern){const r=n.match(this._currentWordPattern);if((r?r[0].length:0)!==n.length)return this.clearRanges()}const o=[];for(let r=1,a=this._currentDecorations.length;r<a;r++){const l=this._currentDecorations.getRange(r);if(l)if(l.startLineNumber!==l.endLineNumber)o.push({range:l,text:n});else{let c=t.getValueInRange(l),d=n,h=l.startColumn,u=l.endColumn;const g=Fa(c,d);h+=g,c=c.substr(g),d=d.substr(g);const f=uh(c,d);u-=f,c=c.substr(0,c.length-f),d=d.substr(0,d.length-f),(h!==u||d.length!==0)&&o.push({range:new k(l.startLineNumber,h,l.endLineNumber,u),text:d})}}if(o.length!==0)try{this._editor.popUndoStop(),this._ignoreChangeEvent=!0;const r=this._editor._getViewModel().getPrevEditOperationType();this._editor.executeEdits("linkedEditing",o),this._editor._getViewModel().setPrevEditOperationType(r)}finally{this._ignoreChangeEvent=!1}}dispose(){this.clearRanges(),super.dispose()}clearRanges(){this._visibleContextKey.set(!1),this._currentDecorations.clear(),this._currentRequestCts&&(this._currentRequestCts.cancel(),this._currentRequestCts=null,this._currentRequestPosition=null)}async updateRanges(e=!1){if(!this._editor.hasModel()){this.clearRanges();return}const t=this._editor.getPosition();if(!this._enabled&&!e||this._editor.getSelections().length>1){this.clearRanges();return}const i=this._editor.getModel(),n=i.getVersionId();if(this._currentRequestPosition&&this._currentRequestModelVersion===n){if(t.equals(this._currentRequestPosition))return;if(this._currentDecorations.length>0){const r=this._currentDecorations.getRange(0);if(r&&r.containsPosition(t))return}}this._currentRequestPosition?.equals(t)||this._currentDecorations.getRange(0)?.containsPosition(t)||this.clearRanges(),this._currentRequestPosition=t,this._currentRequestModelVersion=n;const o=this._currentRequestCts=new si;try{const r=new Cn(!1),a=await Ooe(this._providers,i,t,o.token);if(this._debounceInformation.update(i,r.elapsed()),o!==this._currentRequestCts||(this._currentRequestCts=null,n!==i.getVersionId()))return;let l=[];a?.ranges&&(l=a.ranges),this._currentWordPattern=a?.wordPattern||this._languageWordPattern;let c=!1;for(let h=0,u=l.length;h<u;h++)if(k.containsPosition(l[h],t)){if(c=!0,h!==0){const g=l[h];l.splice(h,1),l.unshift(g)}break}if(!c){this.clearRanges();return}const d=l.map(h=>({range:h,options:Qk.DECORATION}));this._visibleContextKey.set(!0),this._currentDecorations.set(d),this._syncRangesToken++}catch(r){Dr(r)||Oe(r),(this._currentRequestCts===o||!this._currentRequestCts)&&this.clearRanges()}}},Qk=ng,ng.ID="editor.contrib.linkedEditing",ng.DECORATION=Fe.register({description:"linked-editing",stickiness:0,className:D4e}),ng);Av=Qk=I4e([XL(1,Ae),XL(2,ge),XL(3,ui),XL(4,ga)],Av);class T4e extends ue{constructor(){super({id:"editor.action.linkedEditing",label:V(1276,"Start Linked Editing"),precondition:K.and(N.writable,N.hasRenameProvider),kbOpts:{kbExpr:N.editorTextFocus,primary:3132,weight:100}})}runCommand(e,t){const i=e.get(ht),[n,o]=Array.isArray(t)&&t||[void 0,void 0];return Le.isUri(n)&&A.isIPosition(o)?i.openCodeEditor({resource:n},i.getActiveCodeEditor()).then(r=>{r&&(r.setPosition(o),r.invokeWithinContext(a=>(this.reportTelemetry(a,r),this.run(a,r))))},Oe):super.runCommand(e,t)}run(e,t){const i=Av.get(t);return i?Promise.resolve(i.updateRanges(!0)):Promise.resolve()}}const N4e=hn.bindToContribution(Av.get);le(new N4e({id:"cancelLinkedEditingInput",precondition:Moe,handler:s=>s.clearRanges(),kbOpts:{kbExpr:N.editorTextFocus,weight:199,primary:9,secondary:[1033]}}));function Ooe(s,e,t,i){const n=s.ordered(e);return z6(n.map(o=>async()=>{try{return await o.provideLinkedEditingRanges(e,t,i)}catch(r){ji(r);return}}),o=>!!o&&Ms(o?.ranges))}T("editor.linkedEditingBackground",{dark:q.fromHex("#f00").transparent(.3),light:q.fromHex("#f00").transparent(.3),hcDark:q.fromHex("#f00").transparent(.3),hcLight:q.white},m(1275,"Background color when the editor auto renames on type."));Lo("_executeLinkedEditingProvider",(s,e,t)=>{const{linkedEditingRangeProvider:i}=s.get(ge);return Ooe(i,e,t,Ke.None)});lt(Av.ID,Av,1);ae(T4e);let R4e=class{constructor(e,t){this._link=e,this._provider=t}toJSON(){return{range:this.range,url:this.url,tooltip:this.tooltip}}get range(){return this._link.range}get url(){return this._link.url}get tooltip(){return this._link.tooltip}async resolve(e){return this._link.url?this._link.url:typeof this._provider.resolveLink=="function"?Promise.resolve(this._provider.resolveLink(this._link,e)).then(t=>(this._link=t||this._link,this._link.url?this.resolve(e):Promise.reject(new Error("missing")))):Promise.reject(new Error("missing"))}};const Ry=class Ry{constructor(e){this._disposables=new U;let t=[];for(const[i,n]of e){const o=i.links.map(r=>new R4e(r,n));t=Ry._union(t,o),i_(i)&&(this._disposables??=new U,this._disposables.add(i))}this.links=t}dispose(){this._disposables?.dispose(),this.links.length=0}static _union(e,t){const i=[];let n,o,r,a;for(n=0,r=0,o=e.length,a=t.length;n<o&&r<a;){const l=e[n],c=t[r];if(k.areIntersectingOrTouching(l.range,c.range)){n++;continue}k.compareRangesUsingStarts(l.range,c.range)<0?(i.push(l),n++):(i.push(c),r++)}for(;n<o;n++)i.push(e[n]);for(;r<a;r++)i.push(t[r]);return i}};Ry.Empty=new Ry([]);let KD=Ry;async function Poe(s,e,t){const i=[],n=s.ordered(e).reverse().map(async(r,a)=>{try{const l=await r.provideLinks(e,t);l&&(i[a]=[l,r])}catch(l){ji(l)}});await Promise.all(n);let o=new KD(Al(i));return t.isCancellationRequested&&(o.dispose(),o=KD.Empty),o}at.registerCommand("_executeLinkProvider",async(s,...e)=>{let[t,i]=e;dt(t instanceof Le),typeof i!="number"&&(i=0);const{linkProvider:n}=s.get(ge),o=s.get(hi).getModel(t);if(!o)return[];const r=await Poe(n,o,Ke.None);if(!r)return[];for(let l=0;l<Math.min(i,r.links.length);l++)await r.links[l].resolve(Ke.None);const a=r.links.slice(0);return r.dispose(),a});var A4e=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},QL=function(s,e){return function(t,i){e(t,i,s)}},u8,nm;let u0=(nm=class extends F{static get(e){return e.getContribution(u8.ID)}constructor(e,t,i,n,o){super(),this.editor=e,this.openerService=t,this.notificationService=i,this.languageFeaturesService=n,this.providers=this.languageFeaturesService.linkProvider,this.debounceInformation=o.for(this.providers,"Links",{min:1e3,max:4e3}),this.computeLinks=this._register(new Rt(()=>this.computeLinksNow(),1e3)),this.computePromise=null,this.activeLinksList=null,this.currentOccurrences={},this.activeLinkDecorationId=null;const r=this._register(new AA(e));this._register(r.onMouseMoveOrRelevantKeyDown(([a,l])=>{this._onEditorMouseMove(a,l)})),this._register(r.onExecute(a=>{this.onEditorMouseUp(a)})),this._register(r.onCancel(a=>{this.cleanUpActiveLinkDecoration()})),this._register(e.onDidChangeConfiguration(a=>{a.hasChanged(79)&&(this.updateDecorations([]),this.stop(),this.computeLinks.schedule(0))})),this._register(e.onDidChangeModelContent(a=>{this.editor.hasModel()&&this.computeLinks.schedule(this.debounceInformation.get(this.editor.getModel()))})),this._register(e.onDidChangeModel(a=>{this.currentOccurrences={},this.activeLinkDecorationId=null,this.stop(),this.computeLinks.schedule(0)})),this._register(e.onDidChangeModelLanguage(a=>{this.stop(),this.computeLinks.schedule(0)})),this._register(this.providers.onDidChange(a=>{this.stop(),this.computeLinks.schedule(0)})),this.computeLinks.schedule(0)}async computeLinksNow(){if(!this.editor.hasModel()||!this.editor.getOption(79))return;const e=this.editor.getModel();if(!e.isTooLargeForSyncing()&&this.providers.has(e)){this.activeLinksList&&(this.activeLinksList.dispose(),this.activeLinksList=null),this.computePromise=an(t=>Poe(this.providers,e,t));try{const t=new Cn(!1);if(this.activeLinksList=await this.computePromise,this.debounceInformation.update(e,t.elapsed()),e.isDisposed())return;this.updateDecorations(this.activeLinksList.links)}catch(t){Oe(t)}finally{this.computePromise=null}}}updateDecorations(e){const t=this.editor.getOption(86)==="altKey",i=[],n=Object.keys(this.currentOccurrences);for(const r of n){const a=this.currentOccurrences[r];i.push(a.decorationId)}const o=[];if(e)for(const r of e)o.push(Dw.decoration(r,t));this.editor.changeDecorations(r=>{const a=r.deltaDecorations(i,o);this.currentOccurrences={},this.activeLinkDecorationId=null;for(let l=0,c=a.length;l<c;l++){const d=new Dw(e[l],a[l]);this.currentOccurrences[d.decorationId]=d}})}_onEditorMouseMove(e,t){const i=this.editor.getOption(86)==="altKey";if(this.isEnabled(e,t)){this.cleanUpActiveLinkDecoration();const n=this.getLinkOccurrence(e.target.position);n&&this.editor.changeDecorations(o=>{n.activate(o,i),this.activeLinkDecorationId=n.decorationId})}else this.cleanUpActiveLinkDecoration()}cleanUpActiveLinkDecoration(){const e=this.editor.getOption(86)==="altKey";if(this.activeLinkDecorationId){const t=this.currentOccurrences[this.activeLinkDecorationId];t&&this.editor.changeDecorations(i=>{t.deactivate(i,e)}),this.activeLinkDecorationId=null}}onEditorMouseUp(e){if(!this.isEnabled(e))return;const t=this.getLinkOccurrence(e.target.position);t&&this.openLinkOccurrence(t,e.hasSideBySideModifier,!0)}openLinkOccurrence(e,t,i=!1){if(!this.openerService)return;const{link:n}=e;n.resolve(Ke.None).then(o=>{if(typeof o=="string"&&this.editor.hasModel()){const r=this.editor.getModel().uri;if(r.scheme===Te.file&&o.startsWith(`${Te.file}:`)){const a=Le.parse(o);if(a.scheme===Te.file){const l=$l(a);let c=null;l.startsWith("/./")||l.startsWith("\\.\\")?c=`.${l.substr(1)}`:(l.startsWith("//./")||l.startsWith("\\\\.\\"))&&(c=`.${l.substr(2)}`),c&&(o=qQ(r,c))}}}return this.openerService.open(o,{openToSide:t,fromUserGesture:i,allowContributedOpeners:!0,allowCommands:!0,fromWorkspace:!0})},o=>{const r=o instanceof Error?o.message:o;r==="invalid"?this.notificationService.warn(m(1277,"Failed to open this link because it is not well-formed: {0}",n.url.toString())):r==="missing"?this.notificationService.warn(m(1278,"Failed to open this link because its target is missing.")):Oe(o)})}getLinkOccurrence(e){if(!this.editor.hasModel()||!e)return null;const t=this.editor.getModel().getDecorationsInRange({startLineNumber:e.lineNumber,startColumn:e.column,endLineNumber:e.lineNumber,endColumn:e.column},0,!0);for(const i of t){const n=this.currentOccurrences[i.id];if(n)return n}return null}isEnabled(e,t){return!!(e.target.type===6&&(e.hasTriggerModifier||t&&t.keyCodeIsTriggerKey||e.isMiddleClick&&e.mouseMiddleClickAction==="openLink"))}stop(){this.computeLinks.cancel(),this.activeLinksList&&(this.activeLinksList?.dispose(),this.activeLinksList=null),this.computePromise&&(this.computePromise.cancel(),this.computePromise=null)}dispose(){super.dispose(),this.stop()}},u8=nm,nm.ID="editor.linkDetector",nm);u0=u8=A4e([QL(1,Ah),QL(2,Ii),QL(3,ge),QL(4,ga)],u0);const xK={general:Fe.register({description:"detected-link",stickiness:1,collapseOnReplaceEdit:!0,inlineClassName:"detected-link"}),active:Fe.register({description:"detected-link-active",stickiness:1,collapseOnReplaceEdit:!0,inlineClassName:"detected-link-active"})};class Dw{static decoration(e,t){return{range:e.range,options:Dw._getOptions(e,t,!1)}}static _getOptions(e,t,i){const n={...i?xK.active:xK.general};return n.hoverMessage=M4e(e,t),n}constructor(e,t){this.link=e,this.decorationId=t}activate(e,t){e.changeDecorationOptions(this.decorationId,Dw._getOptions(this.link,t,!0))}deactivate(e,t){e.changeDecorationOptions(this.decorationId,Dw._getOptions(this.link,t,!1))}}function M4e(s,e){const t=s.url&&/^command:/i.test(s.url.toString()),i=s.tooltip?s.tooltip:t?m(1279,"Execute command"):m(1280,"Follow link"),n=e?Ye?m(1281,"cmd + click"):m(1282,"ctrl + click"):Ye?m(1283,"option + click"):m(1284,"alt + click");if(s.url){let o="";if(/^command:/i.test(s.url.toString())){const a=s.url.toString().match(/^command:([^?#]+)/);if(a){const l=a[1];o=m(1285,"Execute command {0}",l)}}return new hs("",!0).appendLink(s.url.toString(!0).replace(/ /g,"%20"),i,o).appendMarkdown(` (${n})`)}else return new hs().appendText(`${i} (${n})`)}class O4e extends ue{constructor(){super({id:"editor.action.openLink",label:V(1286,"Open Link"),precondition:void 0})}run(e,t){const i=u0.get(t);if(!i||!t.hasModel())return;const n=t.getSelections();for(const o of n){const r=i.getLinkOccurrence(o.getEndPosition());r&&i.openLinkOccurrence(r,!1)}}}lt(u0.ID,u0,1);ae(O4e);const UV=class UV extends F{constructor(e){super(),this._editor=e,this._register(this._editor.onMouseDown(t=>{const i=this._editor.getOption(133);i>=0&&t.target.type===6&&t.target.position.column>=i&&this._editor.updateOptions({stopRenderingLineAfter:-1})}))}};UV.ID="editor.contrib.longLinesHelper";let YD=UV;lt(YD.ID,YD,2);const qV=class qV extends F{constructor(e){super(),this._editor=e;const t=di(this._editor),i=t.getOption(171);this._register(De(n=>{if(!i.read(n))return;const o=t.domNode.read(n);if(!o)return;const r=n.store.add(lW("scrollingSession",void 0));n.store.add(this._editor.onMouseDown(l=>{if(r.read(void 0)){r.set(void 0,void 0);return}if(!l.event.middleButton)return;l.event.stopPropagation(),l.event.preventDefault();const d=new U,h=new _n(l.event.posx,l.event.posy),g=P4e(be(o),h,d).map(_=>_.subtract(h).withThreshold(5)),f=o.getBoundingClientRect(),p=new _n(h.x-f.left,h.y-f.top);r.set({mouseDeltaAfterThreshold:g,initialMousePosInEditor:p,didScroll:!1,dispose:()=>d.dispose()},void 0),d.add(this._editor.onMouseUp(_=>{const b=r.read(void 0);b&&b.didScroll&&r.set(void 0,void 0)})),d.add(this._editor.onKeyDown(_=>{r.set(void 0,void 0)}))})),n.store.add(De(l=>{const c=r.read(l);if(!c)return;let d=Date.now();l.store.add(De(u=>{RD.instance.invalidateOnNextAnimationFrame(u);const g=Date.now(),f=g-d;d=g;const p=c.mouseDeltaAfterThreshold.read(void 0),_=f/32,b=p.scale(_),w=new _n(this._editor.getScrollLeft(),this._editor.getScrollTop());this._editor.setScrollPosition(F4e(w.add(b))),b.isZero()||(c.didScroll=!0)}));const h=j(u=>{const g=c.mouseDeltaAfterThreshold.read(u);let f="";return f+=g.y<0?"n":g.y>0?"s":"",f+=g.x<0?"w":g.x>0?"e":"",f});l.store.add(De(u=>{o.setAttribute("data-scroll-direction",h.read(u))}))}));const a=n.store.add(Ve.div({class:["scroll-editor-on-middle-click-dot",r.map(l=>l?"":"hidden")],style:{left:r.map(l=>l?l.initialMousePosInEditor.x:0),top:r.map(l=>l?l.initialMousePosInEditor.y:0)}}).toDisposableLiveElement());n.store.add(Nb(o,a.element)),n.store.add(De(l=>{const c=r.read(l);o.classList.toggle("scroll-editor-on-middle-click-editor",!!c)}))}))}};qV.ID="editor.contrib.middleScroll";let ZD=qV;function P4e(s,e,t){const i=Re("pos",e);return t.add(W(s,"mousemove",n=>{i.set(new _n(n.pageX,n.pageY),void 0)})),i}function F4e(s){return{scrollLeft:s.x,scrollTop:s.y}}lt(ZD.ID,ZD,2);const B4e=T("editor.wordHighlightBackground",{dark:"#575757B8",light:"#57575740",hcDark:null,hcLight:null},m(1563,"Background color of a symbol during read-access, like reading a variable. The color must not be opaque so as not to hide underlying decorations."),!0);T("editor.wordHighlightStrongBackground",{dark:"#004972B8",light:"#0e639c40",hcDark:null,hcLight:null},m(1564,"Background color of a symbol during write-access, like writing to a variable. The color must not be opaque so as not to hide underlying decorations."),!0);T("editor.wordHighlightTextBackground",B4e,m(1565,"Background color of a textual occurrence for a symbol. The color must not be opaque so as not to hide underlying decorations."),!0);const W4e=T("editor.wordHighlightBorder",{light:null,dark:null,hcDark:ri,hcLight:ri},m(1566,"Border color of a symbol during read-access, like reading a variable."));T("editor.wordHighlightStrongBorder",{light:null,dark:null,hcDark:ri,hcLight:ri},m(1567,"Border color of a symbol during write-access, like writing to a variable."));T("editor.wordHighlightTextBorder",W4e,m(1568,"Border color of a textual occurrence for a symbol."));const H4e=T("editorOverviewRuler.wordHighlightForeground","#A0A0A0CC",m(1569,"Overview ruler marker color for symbol highlights. The color must not be opaque so as not to hide underlying decorations."),!0),V4e=T("editorOverviewRuler.wordHighlightStrongForeground","#C0A0C0CC",m(1570,"Overview ruler marker color for write-access symbol highlights. The color must not be opaque so as not to hide underlying decorations."),!0),z4e=T("editorOverviewRuler.wordHighlightTextForeground",FJ,m(1571,"Overview ruler marker color of a textual occurrence for a symbol. The color must not be opaque so as not to hide underlying decorations."),!0),$4e=Fe.register({description:"word-highlight-strong",stickiness:1,className:"wordHighlightStrong",overviewRuler:{color:Ci(V4e),position:Sr.Center},minimap:{color:Ci(nA),position:1}}),U4e=Fe.register({description:"word-highlight-text",stickiness:1,className:"wordHighlightText",overviewRuler:{color:Ci(z4e),position:Sr.Center},minimap:{color:Ci(nA),position:1}}),q4e=Fe.register({description:"selection-highlight-overview",stickiness:1,className:"selectionHighlight",overviewRuler:{color:Ci(FJ),position:Sr.Center},minimap:{color:Ci(nA),position:1}}),j4e=Fe.register({description:"selection-highlight",stickiness:1,className:"selectionHighlight"}),G4e=Fe.register({description:"word-highlight",stickiness:1,className:"wordHighlight",overviewRuler:{color:Ci(H4e),position:Sr.Center},minimap:{color:Ci(nA),position:1}});function K4e(s){return s===_v.Write?$4e:s===_v.Text?U4e:G4e}function Y4e(s){return s?j4e:q4e}ua((s,e)=>{const t=s.getColor(NW);t&&e.addRule(`.monaco-editor .selectionHighlight { background-color: ${t.transparent(.5)}; }`)});var Z4e=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},X4e=function(s,e){return function(t,i){e(t,i,s)}},g8;function Ag(s,e){const t=e.filter(i=>!s.find(n=>n.equals(i)));if(t.length>=1){const i=t.map(o=>`line ${o.viewState.position.lineNumber} column ${o.viewState.position.column}`).join(", "),n=t.length===1?m(1288,"Cursor added: {0}",i):m(1289,"Cursors added: {0}",i);xl(n)}}class Q4e extends ue{constructor(){super({id:"editor.action.insertCursorAbove",label:V(1298,"Add Cursor Above"),precondition:void 0,kbOpts:{kbExpr:N.editorTextFocus,primary:2576,linux:{primary:1552,secondary:[3088]},weight:100},menuOpts:{menuId:fe.MenubarSelectionMenu,group:"3_multi",title:m(1290,"&&Add Cursor Above"),order:2}})}run(e,t,i){if(!t.hasModel())return;let n=!0;i&&i.logicalLine===!1&&(n=!1);const o=t._getViewModel();if(o.cursorConfig.readOnly)return;o.model.pushStackElement();const r=o.getCursorStates();o.setCursorStates(i.source,3,Rn.addCursorUp(o,r,n)),o.revealTopMostCursor(i.source),Ag(r,o.getCursorStates())}}class J4e extends ue{constructor(){super({id:"editor.action.insertCursorBelow",label:V(1299,"Add Cursor Below"),precondition:void 0,kbOpts:{kbExpr:N.editorTextFocus,primary:2578,linux:{primary:1554,secondary:[3090]},weight:100},menuOpts:{menuId:fe.MenubarSelectionMenu,group:"3_multi",title:m(1291,"A&&dd Cursor Below"),order:3}})}run(e,t,i){if(!t.hasModel())return;let n=!0;i&&i.logicalLine===!1&&(n=!1);const o=t._getViewModel();if(o.cursorConfig.readOnly)return;o.model.pushStackElement();const r=o.getCursorStates();o.setCursorStates(i.source,3,Rn.addCursorDown(o,r,n)),o.revealBottomMostCursor(i.source),Ag(r,o.getCursorStates())}}class e5e extends ue{constructor(){super({id:"editor.action.insertCursorAtEndOfEachLineSelected",label:V(1300,"Add Cursors to Line Ends"),precondition:void 0,kbOpts:{kbExpr:N.editorTextFocus,primary:1575,weight:100},menuOpts:{menuId:fe.MenubarSelectionMenu,group:"3_multi",title:m(1292,"Add C&&ursors to Line Ends"),order:4}})}getCursorsForSelection(e,t,i){if(!e.isEmpty()){for(let n=e.startLineNumber;n<e.endLineNumber;n++){const o=t.getLineMaxColumn(n);i.push(new he(n,o,n,o))}e.endColumn>1&&i.push(new he(e.endLineNumber,e.endColumn,e.endLineNumber,e.endColumn))}}run(e,t){if(!t.hasModel())return;const i=t.getModel(),n=t.getSelections(),o=t._getViewModel(),r=o.getCursorStates(),a=[];n.forEach(l=>this.getCursorsForSelection(l,i,a)),a.length>0&&t.setSelections(a),Ag(r,o.getCursorStates())}}class t5e extends ue{constructor(){super({id:"editor.action.addCursorsToBottom",label:V(1301,"Add Cursors to Bottom"),precondition:void 0})}run(e,t){if(!t.hasModel())return;const i=t.getSelections(),n=t.getModel().getLineCount(),o=[];for(let l=i[0].startLineNumber;l<=n;l++)o.push(new he(l,i[0].startColumn,l,i[0].endColumn));const r=t._getViewModel(),a=r.getCursorStates();o.length>0&&t.setSelections(o),Ag(a,r.getCursorStates())}}class i5e extends ue{constructor(){super({id:"editor.action.addCursorsToTop",label:V(1302,"Add Cursors to Top"),precondition:void 0})}run(e,t){if(!t.hasModel())return;const i=t.getSelections(),n=[];for(let a=i[0].startLineNumber;a>=1;a--)n.push(new he(a,i[0].startColumn,a,i[0].endColumn));const o=t._getViewModel(),r=o.getCursorStates();n.length>0&&t.setSelections(n),Ag(r,o.getCursorStates())}}class JL{constructor(e,t,i){this.selections=e,this.revealRange=t,this.revealScrollType=i}}class g0{static create(e,t){if(!e.hasModel())return null;const i=t.getState();if(!e.hasTextFocus()&&i.isRevealed&&i.searchString.length>0)return new g0(e,t,!1,i.searchString,i.wholeWord,i.matchCase,null);let n=!1,o,r;const a=e.getSelections();a.length===1&&a[0].isEmpty()?(n=!0,o=!0,r=!0):(o=i.wholeWord,r=i.matchCase);const l=e.getSelection();let c,d=null;if(l.isEmpty()){const h=e.getConfiguredWordAtPosition(l.getStartPosition());if(!h)return null;c=h.word,d=new he(l.startLineNumber,h.startColumn,l.startLineNumber,h.endColumn)}else c=e.getModel().getValueInRange(l).replace(/\r\n/g,`
`);return new g0(e,t,n,c,o,r,d)}constructor(e,t,i,n,o,r,a){this._editor=e,this.findController=t,this.isDisconnectedFromFindController=i,this.searchText=n,this.wholeWord=o,this.matchCase=r,this.currentMatch=a}addSelectionToNextFindMatch(){if(!this._editor.hasModel())return null;const e=this._getNextMatch();if(!e)return null;const t=this._editor.getSelections();return new JL(t.concat(e),e,0)}moveSelectionToNextFindMatch(){if(!this._editor.hasModel())return null;const e=this._getNextMatch();if(!e)return null;const t=this._editor.getSelections();return new JL(t.slice(0,t.length-1).concat(e),e,0)}_getNextMatch(){if(!this._editor.hasModel())return null;if(this.currentMatch){const n=this.currentMatch;return this.currentMatch=null,n}this.findController.highlightFindOptions();const e=this._editor.getSelections(),t=e[e.length-1],i=this._editor.getModel().findNextMatch(this.searchText,t.getEndPosition(),!1,this.matchCase,this.wholeWord?this._editor.getOption(148):null,!1);return i?new he(i.range.startLineNumber,i.range.startColumn,i.range.endLineNumber,i.range.endColumn):null}addSelectionToPreviousFindMatch(){if(!this._editor.hasModel())return null;const e=this._getPreviousMatch();if(!e)return null;const t=this._editor.getSelections();return new JL(t.concat(e),e,0)}moveSelectionToPreviousFindMatch(){if(!this._editor.hasModel())return null;const e=this._getPreviousMatch();if(!e)return null;const t=this._editor.getSelections();return new JL(t.slice(0,t.length-1).concat(e),e,0)}_getPreviousMatch(){if(!this._editor.hasModel())return null;if(this.currentMatch){const n=this.currentMatch;return this.currentMatch=null,n}this.findController.highlightFindOptions();const e=this._editor.getSelections(),t=e[e.length-1],i=this._editor.getModel().findPreviousMatch(this.searchText,t.getStartPosition(),!1,this.matchCase,this.wholeWord?this._editor.getOption(148):null,!1);return i?new he(i.range.startLineNumber,i.range.startColumn,i.range.endLineNumber,i.range.endColumn):null}selectAll(e){if(!this._editor.hasModel())return[];this.findController.highlightFindOptions();const t=this._editor.getModel();return e?t.findMatches(this.searchText,e,!1,this.matchCase,this.wholeWord?this._editor.getOption(148):null,!1,1073741824):t.findMatches(this.searchText,!0,!1,this.matchCase,this.wholeWord?this._editor.getOption(148):null,!1,1073741824)}}const XN=class XN extends F{static get(e){return e.getContribution(XN.ID)}constructor(e){super(),this._sessionDispose=this._register(new U),this._editor=e,this._ignoreSelectionChange=!1,this._session=null}dispose(){this._endSession(),super.dispose()}_beginSessionIfNeeded(e){if(!this._session){const t=g0.create(this._editor,e);if(!t)return;this._session=t;const i={searchString:this._session.searchText};this._session.isDisconnectedFromFindController&&(i.wholeWordOverride=1,i.matchCaseOverride=1,i.isRegexOverride=2),e.getState().change(i,!1),this._sessionDispose.add(this._editor.onDidChangeCursorSelection(n=>{this._ignoreSelectionChange||this._endSession()})),this._sessionDispose.add(this._editor.onDidBlurEditorText(()=>{this._endSession()})),this._sessionDispose.add(e.getState().onFindReplaceStateChange(n=>{(n.matchCase||n.wholeWord)&&this._endSession()}))}}_endSession(){if(this._sessionDispose.clear(),this._session&&this._session.isDisconnectedFromFindController){const e={wholeWordOverride:0,matchCaseOverride:0,isRegexOverride:0};this._session.findController.getState().change(e,!1)}this._session=null}_setSelections(e){this._ignoreSelectionChange=!0,this._editor.setSelections(e),this._ignoreSelectionChange=!1}_expandEmptyToWord(e,t){if(!t.isEmpty())return t;const i=this._editor.getConfiguredWordAtPosition(t.getStartPosition());return i?new he(t.startLineNumber,i.startColumn,t.startLineNumber,i.endColumn):t}_applySessionResult(e){e&&(this._setSelections(e.selections),e.revealRange&&this._editor.revealRangeInCenterIfOutsideViewport(e.revealRange,e.revealScrollType))}getSession(e){return this._session}addSelectionToNextFindMatch(e){if(this._editor.hasModel()){if(!this._session){const t=this._editor.getSelections();if(t.length>1){const n=e.getState().matchCase;if(!Foe(this._editor.getModel(),t,n)){const r=this._editor.getModel(),a=[];for(let l=0,c=t.length;l<c;l++)a[l]=this._expandEmptyToWord(r,t[l]);this._editor.setSelections(a);return}}}this._beginSessionIfNeeded(e),this._session&&this._applySessionResult(this._session.addSelectionToNextFindMatch())}}addSelectionToPreviousFindMatch(e){this._beginSessionIfNeeded(e),this._session&&this._applySessionResult(this._session.addSelectionToPreviousFindMatch())}moveSelectionToNextFindMatch(e){this._beginSessionIfNeeded(e),this._session&&this._applySessionResult(this._session.moveSelectionToNextFindMatch())}moveSelectionToPreviousFindMatch(e){this._beginSessionIfNeeded(e),this._session&&this._applySessionResult(this._session.moveSelectionToPreviousFindMatch())}selectAll(e){if(!this._editor.hasModel())return;let t=null;const i=e.getState();if(i.isRevealed&&i.searchString.length>0&&i.isRegex){const n=this._editor.getModel();i.searchScope?t=n.findMatches(i.searchString,i.searchScope,i.isRegex,i.matchCase,i.wholeWord?this._editor.getOption(148):null,!1,1073741824):t=n.findMatches(i.searchString,!0,i.isRegex,i.matchCase,i.wholeWord?this._editor.getOption(148):null,!1,1073741824)}else{if(this._beginSessionIfNeeded(e),!this._session)return;t=this._session.selectAll(i.searchScope)}if(t.length>0){const n=this._editor.getSelection();for(let o=0,r=t.length;o<r;o++){const a=t[o];if(a.range.intersectRanges(n)){t[o]=t[0],t[0]=a;break}}this._setSelections(t.map(o=>new he(o.range.startLineNumber,o.range.startColumn,o.range.endLineNumber,o.range.endColumn)))}}};XN.ID="editor.contrib.multiCursorController";let Mv=XN;class fC extends ue{run(e,t){const i=Mv.get(t);if(!i)return;const n=t._getViewModel();if(n){const o=n.getCursorStates(),r=xo.get(t);if(r)this._run(i,r);else{const a=e.get(_e).createInstance(xo,t);this._run(i,a),a.dispose()}Ag(o,n.getCursorStates())}}}class n5e extends fC{constructor(){super({id:"editor.action.addSelectionToNextFindMatch",label:V(1303,"Add Selection to Next Find Match"),precondition:void 0,kbOpts:{kbExpr:N.focus,primary:2082,weight:100},menuOpts:{menuId:fe.MenubarSelectionMenu,group:"3_multi",title:m(1293,"Add &&Next Occurrence"),order:5}})}_run(e,t){e.addSelectionToNextFindMatch(t)}}class s5e extends fC{constructor(){super({id:"editor.action.addSelectionToPreviousFindMatch",label:V(1304,"Add Selection to Previous Find Match"),precondition:void 0,menuOpts:{menuId:fe.MenubarSelectionMenu,group:"3_multi",title:m(1294,"Add P&&revious Occurrence"),order:6}})}_run(e,t){e.addSelectionToPreviousFindMatch(t)}}class o5e extends fC{constructor(){super({id:"editor.action.moveSelectionToNextFindMatch",label:V(1305,"Move Last Selection to Next Find Match"),precondition:void 0,kbOpts:{kbExpr:N.focus,primary:Gi(2089,2082),weight:100}})}_run(e,t){e.moveSelectionToNextFindMatch(t)}}class r5e extends fC{constructor(){super({id:"editor.action.moveSelectionToPreviousFindMatch",label:V(1306,"Move Last Selection to Previous Find Match"),precondition:void 0})}_run(e,t){e.moveSelectionToPreviousFindMatch(t)}}class a5e extends fC{constructor(){super({id:"editor.action.selectHighlights",label:V(1307,"Select All Occurrences of Find Match"),precondition:void 0,kbOpts:{kbExpr:N.focus,primary:3114,weight:100},menuOpts:{menuId:fe.MenubarSelectionMenu,group:"3_multi",title:m(1295,"Select All &&Occurrences"),order:7}})}_run(e,t){e.selectAll(t)}}class l5e extends fC{constructor(){super({id:"editor.action.changeAll",label:V(1308,"Change All Occurrences"),precondition:K.and(N.writable,N.editorTextFocus),kbOpts:{kbExpr:N.editorTextFocus,primary:2108,weight:100},contextMenuOpts:{group:"1_modification",order:1.2}})}_run(e,t){e.selectAll(t)}}class c5e{constructor(e,t,i,n,o){this._model=e,this._searchText=t,this._matchCase=i,this._wordSeparators=n,this._cachedFindMatches=null,this._modelVersionId=this._model.getVersionId(),o&&this._model===o._model&&this._searchText===o._searchText&&this._matchCase===o._matchCase&&this._wordSeparators===o._wordSeparators&&this._modelVersionId===o._modelVersionId&&(this._cachedFindMatches=o._cachedFindMatches)}findMatches(){return this._cachedFindMatches===null&&(this._cachedFindMatches=this._model.findMatches(this._searchText,!0,!1,this._matchCase,this._wordSeparators,!1).map(e=>e.range),this._cachedFindMatches.sort(k.compareRangesUsingStarts)),this._cachedFindMatches}}var sm;let XD=(sm=class extends F{constructor(e,t){super(),this._languageFeaturesService=t,this.editor=e,this._isEnabled=e.getOption(122),this._isEnabledMultiline=e.getOption(124),this._maxLength=e.getOption(123),this._decorations=e.createDecorationsCollection(),this.updateSoon=this._register(new Rt(()=>this._update(),300)),this.state=null,this._register(e.onDidChangeConfiguration(n=>{this._isEnabled=e.getOption(122),this._isEnabledMultiline=e.getOption(124),this._maxLength=e.getOption(123)})),this._register(e.onDidChangeCursorSelection(n=>{this._isEnabled&&(n.selection.isEmpty()?n.reason===3?(this.state&&this._setState(null),this.updateSoon.schedule()):this._setState(null):this._update())})),this._register(e.onDidChangeModel(n=>{this._setState(null)})),this._register(e.onDidChangeModelContent(n=>{this._isEnabled&&this.updateSoon.schedule()}));const i=xo.get(e);i&&this._register(i.getState().onFindReplaceStateChange(n=>{this._update()})),this.updateSoon.schedule()}_update(){this._setState(g8._createState(this.state,this._isEnabled,this._isEnabledMultiline,this._maxLength,this.editor))}static _createState(e,t,i,n,o){if(!t||!o.hasModel())return null;if(!i){const h=o.getSelection();if(h.startLineNumber!==h.endLineNumber)return null}const r=Mv.get(o);if(!r)return null;const a=xo.get(o);if(!a)return null;let l=r.getSession(a);if(!l){const h=o.getSelections();if(h.length>1){const g=a.getState().matchCase;if(!Foe(o.getModel(),h,g))return null}l=g0.create(o,a)}if(!l||l.currentMatch||/^[ \t]+$/.test(l.searchText)||n>0&&l.searchText.length>n)return null;const c=a.getState(),d=c.matchCase;if(c.isRevealed){let h=c.searchString;d||(h=h.toLowerCase());let u=l.searchText;if(d||(u=u.toLowerCase()),h===u&&l.matchCase===c.matchCase&&l.wholeWord===c.wholeWord&&!c.isRegex)return null}return new c5e(o.getModel(),l.searchText,l.matchCase,l.wholeWord?o.getOption(148):null,e)}_setState(e){if(this.state=e,!this.state){this._decorations.clear();return}if(!this.editor.hasModel())return;const t=this.editor.getModel();if(t.isTooLargeForTokenization())return;const i=this.state.findMatches(),n=this.editor.getSelections();n.sort(k.compareRangesUsingStarts);const o=[];for(let c=0,d=0,h=i.length,u=n.length;c<h;){const g=i[c];if(d>=u)o.push(g),c++;else{const f=k.compareRangesUsingStarts(g,n[d]);f<0?((n[d].isEmpty()||!k.areIntersecting(g,n[d]))&&o.push(g),c++):(f>0||c++,d++)}}const r=this.editor.getOption(90)!=="off",a=this._languageFeaturesService.documentHighlightProvider.has(t)&&r,l=o.map(c=>({range:c,options:Y4e(a)}));this._decorations.set(l)}dispose(){this._setState(null),super.dispose()}},g8=sm,sm.ID="editor.contrib.selectionHighlighter",sm);XD=g8=Z4e([X4e(1,ge)],XD);function Foe(s,e,t){const i=LK(s,e[0],!t);for(let n=1,o=e.length;n<o;n++){const r=e[n];if(r.isEmpty())return!1;const a=LK(s,r,!t);if(i!==a)return!1}return!0}function LK(s,e,t){const i=s.getValueInRange(e);return t?i.toLowerCase():i}class d5e extends ue{constructor(){super({id:"editor.action.focusNextCursor",label:V(1309,"Focus Next Cursor"),metadata:{description:m(1296,"Focuses the next cursor"),args:[]},precondition:void 0})}run(e,t,i){if(!t.hasModel())return;const n=t._getViewModel();if(n.cursorConfig.readOnly)return;n.model.pushStackElement();const o=Array.from(n.getCursorStates()),r=o.shift();r&&(o.push(r),n.setCursorStates(i.source,3,o),n.revealPrimaryCursor(i.source,!0),Ag(o,n.getCursorStates()))}}class h5e extends ue{constructor(){super({id:"editor.action.focusPreviousCursor",label:V(1310,"Focus Previous Cursor"),metadata:{description:m(1297,"Focuses the previous cursor"),args:[]},precondition:void 0})}run(e,t,i){if(!t.hasModel())return;const n=t._getViewModel();if(n.cursorConfig.readOnly)return;n.model.pushStackElement();const o=Array.from(n.getCursorStates()),r=o.pop();r&&(o.unshift(r),n.setCursorStates(i.source,3,o),n.revealPrimaryCursor(i.source,!0),Ag(o,n.getCursorStates()))}}lt(Mv.ID,Mv,4);lt(XD.ID,XD,1);ae(Q4e);ae(J4e);ae(e5e);ae(n5e);ae(s5e);ae(o5e);ae(r5e);ae(a5e);ae(l5e);ae(t5e);ae(i5e);ae(d5e);ae(h5e);const Jm={Visible:new ce("parameterHintsVisible",!1),MultipleSignatures:new ce("parameterHintsMultipleSignatures",!1)};async function Boe(s,e,t,i,n){const o=s.ordered(e);for(const r of o)try{const a=await r.provideSignatureHelp(e,t,n,i);if(a)return a}catch(a){ji(a)}}at.registerCommand("_executeSignatureHelpProvider",async(s,...e)=>{const[t,i,n]=e;dt(Le.isUri(t)),dt(A.isIPosition(i)),dt(typeof n=="string"||!n);const o=s.get(ge),r=await s.get(Fs).createModelReference(t);try{const a=await Boe(o.signatureHelpProvider,r.object.textEditorModel,A.lift(i),{triggerKind:Lc.Invoke,isRetrigger:!1,triggerCharacter:n},Ke.None);return a?(setTimeout(()=>a.dispose(),0),a.value):void 0}finally{r.dispose()}});var Zh;(function(s){s.Default={type:0};class e{constructor(n,o){this.request=n,this.previouslyActiveHints=o,this.type=2}}s.Pending=e;class t{constructor(n){this.hints=n,this.type=1}}s.Active=t})(Zh||(Zh={}));const QN=class QN extends F{constructor(e,t,i=QN.DEFAULT_DELAY){super(),this._onChangedHints=this._register(new M),this.onChangedHints=this._onChangedHints.event,this.triggerOnType=!1,this._state=Zh.Default,this._pendingTriggers=[],this._lastSignatureHelpResult=this._register(new vt),this.triggerChars=new TE,this.retriggerChars=new TE,this.triggerId=0,this.editor=e,this.providers=t,this.throttledDelayer=new sa(i),this._register(this.editor.onDidBlurEditorWidget(()=>this.cancel())),this._register(this.editor.onDidChangeConfiguration(()=>this.onEditorConfigurationChange())),this._register(this.editor.onDidChangeModel(n=>this.onModelChanged())),this._register(this.editor.onDidChangeModelLanguage(n=>this.onModelChanged())),this._register(this.editor.onDidChangeCursorSelection(n=>this.onCursorChange(n))),this._register(this.editor.onDidChangeModelContent(n=>this.onModelContentChange())),this._register(this.providers.onDidChange(this.onModelChanged,this)),this._register(this.editor.onDidType(n=>this.onDidType(n))),this.onEditorConfigurationChange(),this.onModelChanged()}get state(){return this._state}set state(e){this._state.type===2&&this._state.request.cancel(),this._state=e}cancel(e=!1){this.state=Zh.Default,this.throttledDelayer.cancel(),e||this._onChangedHints.fire(void 0)}trigger(e,t){const i=this.editor.getModel();if(!i||!this.providers.has(i))return;const n=++this.triggerId;this._pendingTriggers.push(e),this.throttledDelayer.trigger(()=>this.doTrigger(n),t).catch(Oe)}next(){if(this.state.type!==1)return;const e=this.state.hints.signatures.length,t=this.state.hints.activeSignature,i=t%e===e-1,n=this.editor.getOption(98).cycle;if((e<2||i)&&!n){this.cancel();return}this.updateActiveSignature(i&&n?0:t+1)}previous(){if(this.state.type!==1)return;const e=this.state.hints.signatures.length,t=this.state.hints.activeSignature,i=t===0,n=this.editor.getOption(98).cycle;if((e<2||i)&&!n){this.cancel();return}this.updateActiveSignature(i&&n?e-1:t-1)}updateActiveSignature(e){this.state.type===1&&(this.state=new Zh.Active({...this.state.hints,activeSignature:e}),this._onChangedHints.fire(this.state.hints))}async doTrigger(e){const t=this.state.type===1||this.state.type===2,i=this.getLastActiveHints();if(this.cancel(!0),this._pendingTriggers.length===0)return!1;const n=this._pendingTriggers.reduce(u5e);this._pendingTriggers=[];const o={triggerKind:n.triggerKind,triggerCharacter:n.triggerCharacter,isRetrigger:t,activeSignatureHelp:i};if(!this.editor.hasModel())return!1;const r=this.editor.getModel(),a=this.editor.getPosition();this.state=new Zh.Pending(an(l=>Boe(this.providers,r,a,o,l)),i);try{const l=await this.state.request;return e!==this.triggerId?(l?.dispose(),!1):!l||!l.value.signatures||l.value.signatures.length===0?(l?.dispose(),this._lastSignatureHelpResult.clear(),this.cancel(),!1):(this.state=new Zh.Active(l.value),this._lastSignatureHelpResult.value=l,this._onChangedHints.fire(this.state.hints),!0)}catch(l){return e===this.triggerId&&(this.state=Zh.Default),Oe(l),!1}}getLastActiveHints(){switch(this.state.type){case 1:return this.state.hints;case 2:return this.state.previouslyActiveHints;default:return}}get isTriggered(){return this.state.type===1||this.state.type===2||this.throttledDelayer.isTriggered()}onModelChanged(){this.cancel(),this.triggerChars.clear(),this.retriggerChars.clear();const e=this.editor.getModel();if(e)for(const t of this.providers.ordered(e)){for(const i of t.signatureHelpTriggerCharacters||[])if(i.length){const n=i.charCodeAt(0);this.triggerChars.add(n),this.retriggerChars.add(n)}for(const i of t.signatureHelpRetriggerCharacters||[])i.length&&this.retriggerChars.add(i.charCodeAt(0))}}onDidType(e){if(!this.triggerOnType)return;const t=e.length-1,i=e.charCodeAt(t);(this.triggerChars.has(i)||this.isTriggered&&this.retriggerChars.has(i))&&this.trigger({triggerKind:Lc.TriggerCharacter,triggerCharacter:e.charAt(t)})}onCursorChange(e){e.source==="mouse"?this.cancel():this.isTriggered&&this.trigger({triggerKind:Lc.ContentChange})}onModelContentChange(){this.isTriggered&&this.trigger({triggerKind:Lc.ContentChange})}onEditorConfigurationChange(){this.triggerOnType=this.editor.getOption(98).enabled,this.triggerOnType||this.cancel()}dispose(){this.cancel(!0),super.dispose()}};QN.DEFAULT_DELAY=120;let f8=QN;function u5e(s,e){switch(e.triggerKind){case Lc.Invoke:return e;case Lc.ContentChange:return s;case Lc.TriggerCharacter:default:return e}}var g5e=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},kK=function(s,e){return function(t,i){e(t,i,s)}},p8;const er=ie,f5e=Qt("parameter-hints-next",Z.chevronDown,m(1312,"Icon for show next parameter hint.")),p5e=Qt("parameter-hints-previous",Z.chevronUp,m(1313,"Icon for show previous parameter hint."));var om;let m8=(om=class extends F{constructor(e,t,i,n){super(),this.editor=e,this.model=t,this.markdownRendererService=n,this.renderDisposeables=this._register(new U),this.visible=!1,this.announcedLabel=null,this.allowEditorOverflow=!0,this.keyVisible=Jm.Visible.bindTo(i),this.keyMultipleSignatures=Jm.MultipleSignatures.bindTo(i)}createParameterHintDOMNodes(){const e=er(".editor-widget.parameter-hints-widget"),t=X(e,er(".phwrapper"));t.tabIndex=-1;const i=X(t,er(".controls")),n=X(i,er(".button"+xe.asCSSSelector(p5e))),o=X(i,er(".overloads")),r=X(i,er(".button"+xe.asCSSSelector(f5e)));this._register(W(n,"click",u=>{gt.stop(u),this.previous()})),this._register(W(r,"click",u=>{gt.stop(u),this.next()}));const a=er(".body"),l=new tx(a,{alwaysConsumeMouseWheel:!0});this._register(l),t.appendChild(l.getDomNode());const c=X(a,er(".signature")),d=X(a,er(".docs"));e.style.userSelect="text",this.domNodes={element:e,signature:c,overloads:o,docs:d,scrollbar:l},this.editor.addContentWidget(this),this.hide(),this._register(this.editor.onDidChangeCursorSelection(u=>{this.visible&&this.editor.layoutContentWidget(this)}));const h=()=>{if(!this.domNodes)return;const u=this.editor.getOption(59),g=this.domNodes.element;g.style.fontSize=`${u.fontSize}px`,g.style.lineHeight=`${u.lineHeight/u.fontSize}`,g.style.setProperty("--vscode-parameterHintsWidget-editorFontFamily",u.fontFamily),g.style.setProperty("--vscode-parameterHintsWidget-editorFontFamilyDefault",_o.fontFamily)};h(),this._register(oe.chain(this.editor.onDidChangeConfiguration.bind(this.editor),u=>u.filter(g=>g.hasChanged(59)))(h)),this._register(this.editor.onDidLayoutChange(u=>this.updateMaxHeight())),this.updateMaxHeight()}show(){this.visible||(this.domNodes||this.createParameterHintDOMNodes(),this.keyVisible.set(!0),this.visible=!0,setTimeout(()=>{this.domNodes?.element.classList.add("visible")},100),this.editor.layoutContentWidget(this))}hide(){this.renderDisposeables.clear(),this.visible&&(this.keyVisible.reset(),this.visible=!1,this.announcedLabel=null,this.domNodes?.element.classList.remove("visible"),this.editor.layoutContentWidget(this))}getPosition(){return this.visible?{position:this.editor.getPosition(),preference:[1,2]}:null}render(e){if(this.renderDisposeables.clear(),!this.domNodes)return;const t=e.signatures.length>1;this.domNodes.element.classList.toggle("multiple",t),this.keyMultipleSignatures.set(t),this.domNodes.signature.innerText="",this.domNodes.docs.innerText="";const i=e.signatures[e.activeSignature];if(!i)return;const n=X(this.domNodes.signature,er(".code")),o=i.parameters.length>0,r=i.activeParameter??e.activeParameter;if(o)this.renderParameters(n,i,r);else{const c=X(n,er("span"));c.textContent=i.label}const a=i.parameters[r];if(a?.documentation){const c=er("span.documentation");if(typeof a.documentation=="string")c.textContent=a.documentation;else{const d=this.renderMarkdownDocs(a.documentation);c.appendChild(d.element)}X(this.domNodes.docs,er("p",{},c))}if(i.documentation!==void 0)if(typeof i.documentation=="string")X(this.domNodes.docs,er("p",{},i.documentation));else{const c=this.renderMarkdownDocs(i.documentation);X(this.domNodes.docs,c.element)}const l=this.hasDocs(i,a);if(this.domNodes.signature.classList.toggle("has-docs",l),this.domNodes.docs.classList.toggle("empty",!l),this.domNodes.overloads.textContent=String(e.activeSignature+1).padStart(e.signatures.length.toString().length,"0")+"/"+e.signatures.length,a){let c="";const d=i.parameters[r];Array.isArray(d.label)?c=i.label.substring(d.label[0],d.label[1]):c=d.label,d.documentation&&(c+=typeof d.documentation=="string"?`, ${d.documentation}`:`, ${d.documentation.value}`),i.documentation&&(c+=typeof i.documentation=="string"?`, ${i.documentation}`:`, ${i.documentation.value}`),this.announcedLabel!==c&&(Js(m(1314,"{0}, hint",c)),this.announcedLabel=c)}this.editor.layoutContentWidget(this),this.domNodes.scrollbar.scanDomNode()}renderMarkdownDocs(e){const t=this.renderDisposeables.add(this.markdownRendererService.render(e,{context:this.editor,asyncRenderCallback:()=>{this.domNodes?.scrollbar.scanDomNode()}}));return t.element.classList.add("markdown-docs"),t}hasDocs(e,t){return!!(t&&typeof t.documentation=="string"&&gu(t.documentation).length>0||t&&typeof t.documentation=="object"&&gu(t.documentation).value.length>0||e.documentation&&typeof e.documentation=="string"&&gu(e.documentation).length>0||e.documentation&&typeof e.documentation=="object"&&gu(e.documentation.value).length>0)}renderParameters(e,t,i){const[n,o]=this.getParameterLabelOffsets(t,i),r=document.createElement("span");r.textContent=t.label.substring(0,n);const a=document.createElement("span");a.textContent=t.label.substring(n,o),a.className="parameter active";const l=document.createElement("span");l.textContent=t.label.substring(o),X(e,r,a,l)}getParameterLabelOffsets(e,t){const i=e.parameters[t];if(i){if(Array.isArray(i.label))return i.label;if(i.label.length){const n=new RegExp(`(\\W|^)${Uo(i.label)}(?=\\W|$)`,"g");n.test(e.label);const o=n.lastIndex-i.label.length;return o>=0?[o,n.lastIndex]:[0,0]}else return[0,0]}else return[0,0]}next(){this.editor.focus(),this.model.next()}previous(){this.editor.focus(),this.model.previous()}getDomNode(){return this.domNodes||this.createParameterHintDOMNodes(),this.domNodes.element}getId(){return p8.ID}updateMaxHeight(){if(!this.domNodes)return;const t=`${Math.max(this.editor.getLayoutInfo().height/4,250)}px`;this.domNodes.element.style.maxHeight=t;const i=this.domNodes.element.getElementsByClassName("phwrapper");i.length&&(i[0].style.maxHeight=t)}},p8=om,om.ID="editor.widget.parameterHintsWidget",om);m8=p8=g5e([kK(2,Ae),kK(3,Ua)],m8);T("editorHoverWidget.highlightForeground",Lb,m(1315,"Foreground color of the active item in the parameter hint."));var m5e=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},EK=function(s,e){return function(t,i){e(t,i,s)}},_8,rm;let Ov=(rm=class extends F{static get(e){return e.getContribution(_8.ID)}constructor(e,t,i){super(),this.editor=e,this.model=this._register(new f8(e,i.signatureHelpProvider)),this._register(this.model.onChangedHints(n=>{n?(this.widget.value.show(),this.widget.value.render(n)):this.widget.rawValue?.hide()})),this.widget=new es(()=>this._register(t.createInstance(m8,this.editor,this.model)))}cancel(){this.model.cancel()}previous(){this.widget.rawValue?.previous()}next(){this.widget.rawValue?.next()}trigger(e){this.model.trigger(e,0)}},_8=rm,rm.ID="editor.controller.parameterHints",rm);Ov=_8=m5e([EK(1,_e),EK(2,ge)],Ov);class _5e extends ue{constructor(){super({id:"editor.action.triggerParameterHints",label:V(1311,"Trigger Parameter Hints"),precondition:N.hasSignatureHelpProvider,kbOpts:{kbExpr:N.editorTextFocus,primary:3082,weight:100}})}run(e,t){Ov.get(t)?.trigger({triggerKind:Lc.Invoke})}}lt(Ov.ID,Ov,2);ae(_5e);const eV=175,tV=hn.bindToContribution(Ov.get);le(new tV({id:"closeParameterHints",precondition:Jm.Visible,handler:s=>s.cancel(),kbOpts:{weight:eV,kbExpr:N.focus,primary:9,secondary:[1033]}}));le(new tV({id:"showPrevParameterHint",precondition:K.and(Jm.Visible,Jm.MultipleSignatures),handler:s=>s.previous(),kbOpts:{weight:eV,kbExpr:N.focus,primary:16,secondary:[528],mac:{primary:16,secondary:[528,302]}}}));le(new tV({id:"showNextParameterHint",precondition:K.and(Jm.Visible,Jm.MultipleSignatures),handler:s=>s.next(),kbOpts:{weight:eV,kbExpr:N.focus,primary:18,secondary:[530],mac:{primary:18,secondary:[530,300]}}}));const jV=class jV extends F{constructor(e){super(),this._editor=e,this._editorObs=di(this._editor),this._placeholderText=this._editorObs.getOption(100),this._state=Jn({owner:this,equalsFn:vP},t=>{const i=this._placeholderText.read(t);if(i&&this._editorObs.valueIsEmpty.read(t))return{placeholder:i}}),this._shouldViewBeAlive=b5e(this,t=>this._state.read(t)?.placeholder!==void 0),this._view=j(t=>{if(!this._shouldViewBeAlive.read(t))return;const i=ct("div.editorPlaceholder");t.store.add(De(n=>{const o=this._state.read(n),r=o?.placeholder!==void 0;i.root.style.display=r?"block":"none",i.root.innerText=o?.placeholder??""})),t.store.add(De(n=>{const o=this._editorObs.layoutInfo.read(n);i.root.style.left=`${o.contentLeft}px`,i.root.style.width=o.contentWidth-o.verticalScrollbarWidth+"px",i.root.style.top=`${this._editor.getTopForLineNumber(0)}px`})),t.store.add(De(n=>{i.root.style.fontFamily=this._editorObs.getOption(58).read(n),i.root.style.fontSize=this._editorObs.getOption(61).read(n)+"px",i.root.style.lineHeight=this._editorObs.getOption(75).read(n)+"px"})),t.store.add(this._editorObs.createOverlayWidget({allowEditorOverflow:!1,minContentWidthInPx:zt(0),position:zt(null),domNode:i.root}))}),this._view.recomputeInitiallyAndOnChange(this._store)}};jV.ID="editor.contrib.placeholderText";let QD=jV;function b5e(s,e){return Ih(s,(t,i)=>i===!0?!0:e(t))}var w5e=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},v5e=function(s,e){return function(t,i){e(t,i,s)}};class C5e{constructor(e){this.instantiationService=e}init(...e){}}function y5e(s){return s()}let IK=class extends C5e{constructor(e,t){super(t),this.init(e)}};IK=w5e([v5e(1,_e)],IK);lt(QD.ID,y5e(()=>QD),0);T("editor.placeholder.foreground",dwe,m(1334,"Foreground color of the placeholder text in the editor."));const QA=qe("textResourceConfigurationService"),Woe=qe("textResourcePropertiesService");var S5e=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},ek=function(s,e){return function(t,i){e(t,i,s)}};const pC=new ce("renameInputVisible",!1,m(1391,"Whether the rename input widget is visible"));new ce("renameInputFocused",!1,m(1392,"Whether the rename input widget is focused"));let b8=class{constructor(e,t,i,n,o,r){this._editor=e,this._acceptKeybindings=t,this._themeService=i,this._keybindingService=n,this._logService=r,this.allowEditorOverflow=!0,this._disposables=new U,this._visibleContextKey=pC.bindTo(o),this._isEditingRenameCandidate=!1,this._nRenameSuggestionsInvocations=0,this._hadAutomaticRenameSuggestionsInvocation=!1,this._candidates=new Set,this._beforeFirstInputFieldEditSW=new Cn,this._inputWithButton=new x5e,this._disposables.add(this._inputWithButton),this._editor.addContentWidget(this),this._disposables.add(this._editor.onDidChangeConfiguration(a=>{a.hasChanged(59)&&this._updateFont()})),this._disposables.add(i.onDidColorThemeChange(this._updateStyles,this))}dispose(){this._disposables.dispose(),this._editor.removeContentWidget(this)}getId(){return"__renameInputWidget"}getDomNode(){return this._domNode||(this._domNode=document.createElement("div"),this._domNode.className="monaco-editor rename-box",this._domNode.appendChild(this._inputWithButton.domNode),this._renameCandidateListView=this._disposables.add(new iV(this._domNode,{fontInfo:this._editor.getOption(59),onFocusChange:e=>{this._inputWithButton.input.value=e,this._isEditingRenameCandidate=!1},onSelectionChange:()=>{this._isEditingRenameCandidate=!1,this.acceptInput(!1)}})),this._disposables.add(this._inputWithButton.onDidInputChange(()=>{this._renameCandidateListView?.focusedCandidate!==void 0&&(this._isEditingRenameCandidate=!0),this._timeBeforeFirstInputFieldEdit??=this._beforeFirstInputFieldEditSW.elapsed(),this._renameCandidateProvidersCts?.token.isCancellationRequested===!1&&this._renameCandidateProvidersCts.cancel(),this._renameCandidateListView?.clearFocus()})),this._label=document.createElement("div"),this._label.className="rename-label",this._domNode.appendChild(this._label),this._updateFont(),this._updateStyles(this._themeService.getColorTheme())),this._domNode}_updateStyles(e){if(!this._domNode)return;const t=e.getColor(tC),i=e.getColor(OW);this._domNode.style.backgroundColor=String(e.getColor(wr)??""),this._domNode.style.boxShadow=t?` 0 0 8px 2px ${t}`:"",this._domNode.style.border=i?`1px solid ${i}`:"",this._domNode.style.color=String(e.getColor(BJ)??"");const n=e.getColor(WJ);this._inputWithButton.domNode.style.backgroundColor=String(e.getColor(JP)??""),this._inputWithButton.input.style.backgroundColor=String(e.getColor(JP)??""),this._inputWithButton.domNode.style.borderWidth=n?"1px":"0px",this._inputWithButton.domNode.style.borderStyle=n?"solid":"none",this._inputWithButton.domNode.style.borderColor=n?.toString()??"none"}_updateFont(){if(this._domNode===void 0)return;dt(this._label!==void 0,"RenameWidget#_updateFont: _label must not be undefined given _domNode is defined"),this._editor.applyFontInfo(this._inputWithButton.input);const e=this._editor.getOption(59);this._label.style.fontSize=`${this._computeLabelFontSize(e.fontSize)}px`}_computeLabelFontSize(e){return e*.8}getPosition(){if(!this._visible||!this._editor.hasModel()||!this._editor.getDomNode())return null;const e=lg(this.getDomNode().ownerDocument.body),t=Li(this._editor.getDomNode()),i=this._getTopForPosition();this._nPxAvailableAbove=i+t.top,this._nPxAvailableBelow=e.height-this._nPxAvailableAbove;const n=this._editor.getOption(75),{totalHeight:o}=f0.getLayoutInfo({lineHeight:n}),r=this._nPxAvailableBelow>o*6?[2,1]:[1,2];return{position:this._position,preference:r}}beforeRender(){const[e,t]=this._acceptKeybindings;return this._label.innerText=m(1393,"{0} to Rename, {1} to Preview",this._keybindingService.lookupKeybinding(e)?.getLabel(),this._keybindingService.lookupKeybinding(t)?.getLabel()),this._domNode.style.minWidth="200px",null}afterRender(e){if(e===null){this.cancelInput(!0,"afterRender (because position is null)");return}if(!this._editor.hasModel()||!this._editor.getDomNode())return;dt(this._renameCandidateListView),dt(this._nPxAvailableAbove!==void 0),dt(this._nPxAvailableBelow!==void 0);const t=Hd(this._inputWithButton.domNode),i=Hd(this._label);let n;e===2?n=this._nPxAvailableBelow:n=this._nPxAvailableAbove,this._renameCandidateListView.layout({height:n-i-t,width:Ro(this._inputWithButton.domNode)})}acceptInput(e){this._trace("invoking acceptInput"),this._currentAcceptInput?.(e)}cancelInput(e,t){this._currentCancelInput?.(e)}focusNextRenameSuggestion(){this._renameCandidateListView?.focusNext()||(this._inputWithButton.input.value=this._currentName)}focusPreviousRenameSuggestion(){this._renameCandidateListView?.focusPrevious()||(this._inputWithButton.input.value=this._currentName)}getInput(e,t,i,n,o){const{start:r,end:a}=this._getSelection(e,t);this._renameCts=o;const l=new U;this._nRenameSuggestionsInvocations=0,this._hadAutomaticRenameSuggestionsInvocation=!1,n===void 0?this._inputWithButton.button.style.display="none":(this._inputWithButton.button.style.display="flex",this._requestRenameCandidatesOnce=n,this._requestRenameCandidates(t,!1),l.add(W(this._inputWithButton.button,"click",()=>this._requestRenameCandidates(t,!0))),l.add(W(this._inputWithButton.button,se.KEY_DOWN,d=>{const h=new Pt(d);(h.equals(3)||h.equals(10))&&(h.stopPropagation(),h.preventDefault(),this._requestRenameCandidates(t,!0))}))),this._isEditingRenameCandidate=!1,this._domNode.classList.toggle("preview",i),this._position=new A(e.startLineNumber,e.startColumn),this._currentName=t,this._inputWithButton.input.value=t,this._inputWithButton.input.setAttribute("selectionStart",r.toString()),this._inputWithButton.input.setAttribute("selectionEnd",a.toString()),this._inputWithButton.input.size=Math.max((e.endColumn-e.startColumn)*1.1,20),this._beforeFirstInputFieldEditSW.reset(),l.add(pe(()=>{this._renameCts=void 0,o.dispose(!0)})),l.add(pe(()=>{this._renameCandidateProvidersCts!==void 0&&(this._renameCandidateProvidersCts.dispose(!0),this._renameCandidateProvidersCts=void 0)})),l.add(pe(()=>this._candidates.clear()));const c=new n_;return c.p.finally(()=>{l.dispose(),this._hide()}),this._currentCancelInput=d=>(this._trace("invoking _currentCancelInput"),this._currentAcceptInput=void 0,this._currentCancelInput=void 0,this._renameCandidateListView?.clearCandidates(),c.complete(d),!0),this._currentAcceptInput=d=>{this._trace("invoking _currentAcceptInput"),dt(this._renameCandidateListView!==void 0);const h=this._renameCandidateListView.nCandidates;let u,g;const f=this._renameCandidateListView.focusedCandidate;if(f!==void 0?(this._trace("using new name from renameSuggestion"),u=f,g={k:"renameSuggestion"}):(this._trace("using new name from inputField"),u=this._inputWithButton.input.value,g=this._isEditingRenameCandidate?{k:"userEditedRenameSuggestion"}:{k:"inputField"}),u===t||u.trim().length===0){this.cancelInput(!0,"_currentAcceptInput (because newName === value || newName.trim().length === 0)");return}this._currentAcceptInput=void 0,this._currentCancelInput=void 0,this._renameCandidateListView.clearCandidates(),c.complete({newName:u,wantsPreview:i&&d,stats:{source:g,nRenameSuggestions:h,timeBeforeFirstInputFieldEdit:this._timeBeforeFirstInputFieldEdit,nRenameSuggestionsInvocations:this._nRenameSuggestionsInvocations,hadAutomaticRenameSuggestionsInvocation:this._hadAutomaticRenameSuggestionsInvocation}})},l.add(o.token.onCancellationRequested(()=>this.cancelInput(!0,"cts.token.onCancellationRequested"))),l.add(this._editor.onDidBlurEditorWidget(()=>this.cancelInput(!this._domNode?.ownerDocument.hasFocus(),"editor.onDidBlurEditorWidget"))),this._show(),c.p}_requestRenameCandidates(e,t){if(this._requestRenameCandidatesOnce!==void 0&&(this._renameCandidateProvidersCts!==void 0&&this._renameCandidateProvidersCts.dispose(!0),dt(this._renameCts),this._inputWithButton.buttonState!=="stop")){this._renameCandidateProvidersCts=new si;const i=t?_S.Invoke:_S.Automatic,n=this._requestRenameCandidatesOnce(i,this._renameCandidateProvidersCts.token);if(n.length===0){this._inputWithButton.setSparkleButton();return}t||(this._hadAutomaticRenameSuggestionsInvocation=!0),this._nRenameSuggestionsInvocations+=1,this._inputWithButton.setStopButton(),this._updateRenameCandidates(n,e,this._renameCts.token)}}_getSelection(e,t){dt(this._editor.hasModel());const i=this._editor.getSelection();let n=0,o=t.length;return!k.isEmpty(i)&&!k.spansMultipleLines(i)&&k.containsRange(e,i)&&(n=Math.max(0,i.startColumn-e.startColumn),o=Math.min(e.endColumn,i.endColumn)-e.startColumn),{start:n,end:o}}_show(){this._trace("invoking _show"),this._editor.revealLineInCenterIfOutsideViewport(this._position.lineNumber,0),this._visible=!0,this._visibleContextKey.set(!0),this._editor.layoutContentWidget(this),setTimeout(()=>{this._inputWithButton.input.focus(),this._inputWithButton.input.setSelectionRange(parseInt(this._inputWithButton.input.getAttribute("selectionStart")),parseInt(this._inputWithButton.input.getAttribute("selectionEnd")))},100)}async _updateRenameCandidates(e,t,i){const n=(...c)=>this._trace("_updateRenameCandidates",...c);n("start");const o=await tv(Promise.allSettled(e),i);if(this._inputWithButton.setSparkleButton(),o===void 0){n("returning early - received updateRenameCandidates results - undefined");return}const r=o.flatMap(c=>c.status==="fulfilled"&&Ln(c.value)?c.value:[]);n(`received updateRenameCandidates results - total (unfiltered) ${r.length} candidates.`);const a=fh(r,c=>c.newSymbolName);n(`distinct candidates - ${a.length} candidates.`);const l=a.filter(({newSymbolName:c})=>c.trim().length>0&&c!==this._inputWithButton.input.value&&c!==t&&!this._candidates.has(c));if(n(`valid distinct candidates - ${r.length} candidates.`),l.forEach(c=>this._candidates.add(c.newSymbolName)),l.length<1){n("returning early - no valid distinct candidates");return}n("setting candidates"),this._renameCandidateListView.setCandidates(l),n("asking editor to re-layout"),this._editor.layoutContentWidget(this)}_hide(){this._trace("invoked _hide"),this._visible=!1,this._visibleContextKey.reset(),this._editor.layoutContentWidget(this)}_getTopForPosition(){const e=this._editor.getVisibleRanges();let t;return e.length>0?t=e[0].startLineNumber:(this._logService.warn("RenameWidget#_getTopForPosition: this should not happen - visibleRanges is empty"),t=Math.max(1,this._position.lineNumber-5)),this._editor.getTopForLineNumber(this._position.lineNumber)-this._editor.getTopForLineNumber(t)}_trace(...e){this._logService.trace("RenameWidget",...e)}};b8=S5e([ek(2,wi),ek(3,ft),ek(4,Ae),ek(5,jt)],b8);class iV{constructor(e,t){this._disposables=new U,this._availableHeight=0,this._minimumWidth=0,this._lineHeight=t.fontInfo.lineHeight,this._typicalHalfwidthCharacterWidth=t.fontInfo.typicalHalfwidthCharacterWidth,this._listContainer=document.createElement("div"),this._listContainer.className="rename-box rename-candidate-list-container",e.appendChild(this._listContainer),this._listWidget=iV._createListWidget(this._listContainer,this._candidateViewHeight,t.fontInfo),this._disposables.add(this._listWidget.onDidChangeFocus(i=>{i.elements.length===1&&t.onFocusChange(i.elements[0].newSymbolName)},this._disposables)),this._disposables.add(this._listWidget.onDidChangeSelection(i=>{i.elements.length===1&&t.onSelectionChange()},this._disposables)),this._disposables.add(this._listWidget.onDidBlur(i=>{this._listWidget.setFocus([])})),this._listWidget.style(f_({listInactiveFocusForeground:dS,listInactiveFocusBackground:hS}))}dispose(){this._listWidget.dispose(),this._disposables.dispose()}layout({height:e,width:t}){this._availableHeight=e,this._minimumWidth=t}setCandidates(e){this._listWidget.splice(0,0,e);const t=this._pickListHeight(this._listWidget.length),i=this._pickListWidth(e);this._listWidget.layout(t,i),this._listContainer.style.height=`${t}px`,this._listContainer.style.width=`${i}px`,xl(m(1394,"Received {0} rename suggestions",e.length))}clearCandidates(){this._listContainer.style.height="0px",this._listContainer.style.width="0px",this._listWidget.splice(0,this._listWidget.length,[])}get nCandidates(){return this._listWidget.length}get focusedCandidate(){if(this._listWidget.length===0)return;const e=this._listWidget.getSelectedElements()[0];if(e!==void 0)return e.newSymbolName;const t=this._listWidget.getFocusedElements()[0];if(t!==void 0)return t.newSymbolName}focusNext(){if(this._listWidget.length===0)return!1;const e=this._listWidget.getFocus();if(e.length===0)return this._listWidget.focusFirst(),this._listWidget.reveal(0),!0;if(e[0]===this._listWidget.length-1)return this._listWidget.setFocus([]),this._listWidget.reveal(0),!1;{this._listWidget.focusNext();const t=this._listWidget.getFocus()[0];return this._listWidget.reveal(t),!0}}focusPrevious(){if(this._listWidget.length===0)return!1;const e=this._listWidget.getFocus();if(e.length===0){this._listWidget.focusLast();const t=this._listWidget.getFocus()[0];return this._listWidget.reveal(t),!0}else{if(e[0]===0)return this._listWidget.setFocus([]),!1;{this._listWidget.focusPrevious();const t=this._listWidget.getFocus()[0];return this._listWidget.reveal(t),!0}}}clearFocus(){this._listWidget.setFocus([])}get _candidateViewHeight(){const{totalHeight:e}=f0.getLayoutInfo({lineHeight:this._lineHeight});return e}_pickListHeight(e){const t=this._candidateViewHeight*e;return Math.min(t,this._availableHeight,this._candidateViewHeight*7)}_pickListWidth(e){const t=Math.ceil(Math.max(...e.map(n=>n.newSymbolName.length))*this._typicalHalfwidthCharacterWidth);return Math.max(this._minimumWidth,25+t+10)}static _createListWidget(e,t,i){const n=new class{getTemplateId(r){return"candidate"}getHeight(r){return t}},o=new class{constructor(){this.templateId="candidate"}renderTemplate(r){return new f0(r,i)}renderElement(r,a,l){l.populate(r)}disposeTemplate(r){r.dispose()}};return new Tr("NewSymbolNameCandidates",e,n,[o],{keyboardSupport:!1,mouseSupport:!0,multipleSelectionSupport:!1})}}class x5e{constructor(){this._buttonHoverContent="",this._onDidInputChange=new M,this.onDidInputChange=this._onDidInputChange.event,this._disposables=new U}get domNode(){return this._domNode||(this._domNode=document.createElement("div"),this._domNode.className="rename-input-with-button",this._domNode.style.display="flex",this._domNode.style.flexDirection="row",this._domNode.style.alignItems="center",this._inputNode=document.createElement("input"),this._inputNode.className="rename-input",this._inputNode.type="text",this._inputNode.style.border="none",this._inputNode.setAttribute("aria-label",m(1395,"Rename input. Type new name and press Enter to commit.")),this._domNode.appendChild(this._inputNode),this._buttonNode=document.createElement("div"),this._buttonNode.className="rename-suggestions-button",this._buttonNode.setAttribute("tabindex","0"),this._buttonGenHoverText=m(1396,"Generate New Name Suggestions"),this._buttonCancelHoverText=m(1397,"Cancel"),this._buttonHoverContent=this._buttonGenHoverText,this._disposables.add($a().setupDelayedHover(this._buttonNode,()=>({content:this._buttonHoverContent,style:1}))),this._domNode.appendChild(this._buttonNode),this._disposables.add(W(this.input,se.INPUT,()=>this._onDidInputChange.fire())),this._disposables.add(W(this.input,se.KEY_DOWN,e=>{const t=new Pt(e);(t.keyCode===15||t.keyCode===17)&&this._onDidInputChange.fire()})),this._disposables.add(W(this.input,se.CLICK,()=>this._onDidInputChange.fire())),this._disposables.add(W(this.input,se.FOCUS,()=>{this.domNode.style.outlineWidth="1px",this.domNode.style.outlineStyle="solid",this.domNode.style.outlineOffset="-1px",this.domNode.style.outlineColor="var(--vscode-focusBorder)"})),this._disposables.add(W(this.input,se.BLUR,()=>{this.domNode.style.outline="none"}))),this._domNode}get input(){return dt(this._inputNode),this._inputNode}get button(){return dt(this._buttonNode),this._buttonNode}get buttonState(){return this._buttonState}setSparkleButton(){this._buttonState="sparkle",this._sparkleIcon??=Dl(Z.sparkle),Pn(this.button),this.button.appendChild(this._sparkleIcon),this.button.setAttribute("aria-label","Generating new name suggestions"),this._buttonHoverContent=this._buttonGenHoverText,this.input.focus()}setStopButton(){this._buttonState="stop",this._stopIcon??=Dl(Z.stopCircle),Pn(this.button),this.button.appendChild(this._stopIcon),this.button.setAttribute("aria-label","Cancel generating new name suggestions"),this._buttonHoverContent=this._buttonCancelHoverText,this.input.focus()}dispose(){this._disposables.dispose()}}const Ay=class Ay{constructor(e,t){this._domNode=document.createElement("div"),this._domNode.className="rename-box rename-candidate",this._domNode.style.display="flex",this._domNode.style.columnGap="5px",this._domNode.style.alignItems="center",this._domNode.style.height=`${t.lineHeight}px`,this._domNode.style.padding=`${Ay._PADDING}px`;const i=document.createElement("div");i.style.display="flex",i.style.alignItems="center",i.style.width=i.style.height=`${t.lineHeight*.8}px`,this._domNode.appendChild(i),this._icon=Dl(Z.sparkle),this._icon.style.display="none",i.appendChild(this._icon),this._label=document.createElement("div"),In(this._label,t),this._domNode.appendChild(this._label),e.appendChild(this._domNode)}populate(e){this._updateIcon(e),this._updateLabel(e)}_updateIcon(e){const t=!!e.tags?.includes(dF.AIGenerated);this._icon.style.display=t?"inherit":"none"}_updateLabel(e){this._label.innerText=e.newSymbolName}static getLayoutInfo({lineHeight:e}){return{totalHeight:e+Ay._PADDING*2}}dispose(){}};Ay._PADDING=2;let f0=Ay;var L5e=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},Xg=function(s,e){return function(t,i){e(t,i,s)}},w8;class nV{constructor(e,t,i){this.model=e,this.position=t,this._providerRenameIdx=0,this._providers=i.ordered(e)}hasProvider(){return this._providers.length>0}async resolveRenameLocation(e){const t=[];for(this._providerRenameIdx=0;this._providerRenameIdx<this._providers.length;this._providerRenameIdx++){const n=this._providers[this._providerRenameIdx];if(!n.resolveRenameLocation)break;const o=await n.resolveRenameLocation(this.model,this.position,e);if(o){if(o.rejectReason){t.push(o.rejectReason);continue}return o}}this._providerRenameIdx=0;const i=this.model.getWordAtPosition(this.position);return i?{range:new k(this.position.lineNumber,i.startColumn,this.position.lineNumber,i.endColumn),text:i.word,rejectReason:t.length>0?t.join(`
`):void 0}:{range:k.fromPositions(this.position),text:"",rejectReason:t.length>0?t.join(`
`):void 0}}async provideRenameEdits(e,t){return this._provideRenameEdits(e,this._providerRenameIdx,[],t)}async _provideRenameEdits(e,t,i,n){const o=this._providers[t];if(!o)return{edits:[],rejectReason:i.join(`
`)};const r=await o.provideRenameEdits(this.model,this.position,e,n);if(r){if(r.rejectReason)return this._provideRenameEdits(e,t+1,i.concat(r.rejectReason),n)}else return this._provideRenameEdits(e,t+1,i.concat(m(1380,"No result.")),n);return r}}async function k5e(s,e,t,i){const n=new nV(e,t,s),o=await n.resolveRenameLocation(Ke.None);return o?.rejectReason?{edits:[],rejectReason:o.rejectReason}:n.provideRenameEdits(i,Ke.None)}var am;let Cg=(am=class{static get(e){return e.getContribution(w8.ID)}constructor(e,t,i,n,o,r,a,l){this.editor=e,this._instaService=t,this._notificationService=i,this._bulkEditService=n,this._progressService=o,this._logService=r,this._configService=a,this._languageFeaturesService=l,this._disposableStore=new U,this._cts=new si,this._renameWidget=this._disposableStore.add(this._instaService.createInstance(b8,this.editor,["acceptRenameInput","acceptRenameInputWithPreview"]))}dispose(){this._disposableStore.dispose(),this._cts.dispose(!0)}async run(){const e=this._logService.trace.bind(this._logService,"[rename]");if(this._cts.dispose(!0),this._cts=new si,!this.editor.hasModel()){e("editor has no model");return}const t=this.editor.getPosition(),i=new nV(this.editor.getModel(),t,this._languageFeaturesService.renameProvider);if(!i.hasProvider()){e("skeleton has no provider");return}const n=new yh(this.editor,5,void 0,this._cts.token);let o;try{e("resolving rename location");const f=i.resolveRenameLocation(n.token);this._progressService.showWhile(f,250),o=await f,e("resolved rename location")}catch(f){f instanceof ia?e("resolve rename location cancelled",JSON.stringify(f,null,"	")):(e("resolve rename location failed",f instanceof Error?f:JSON.stringify(f,null,"	")),(typeof f=="string"||oh(f))&&Vo.get(this.editor)?.showMessage(f||m(1381,"An unknown error occurred while resolving rename location"),t));return}finally{n.dispose()}if(!o){e("returning early - no loc");return}if(o.rejectReason){e(`returning early - rejected with reason: ${o.rejectReason}`,o.rejectReason),Vo.get(this.editor)?.showMessage(o.rejectReason,t);return}if(n.token.isCancellationRequested){e("returning early - cts1 cancelled");return}const r=new yh(this.editor,5,o.range,this._cts.token),a=this.editor.getModel(),l=this._languageFeaturesService.newSymbolNamesProvider.all(a),c=await Promise.all(l.map(async f=>[f,await f.supportsAutomaticNewSymbolNamesTriggerKind??!1])),d=(f,p)=>{let _=c.slice();return f===_S.Automatic&&(_=_.filter(([b,w])=>w)),_.map(([b])=>b.provideNewSymbolNames(a,o.range,f,p))};e("creating rename input field and awaiting its result");const h=this._bulkEditService.hasPreviewHandler()&&this._configService.getValue(this.editor.getModel().uri,"editor.rename.enablePreview"),u=await this._renameWidget.getInput(o.range,o.text,h,l.length>0?d:void 0,r);if(e("received response from rename input field"),typeof u=="boolean"){e(`returning early - rename input field response - ${u}`),u&&this.editor.focus(),r.dispose();return}this.editor.focus(),e("requesting rename edits");const g=tv(i.provideRenameEdits(u.newName,r.token),r.token).then(async f=>{if(!f){e("returning early - no rename edits result");return}if(!this.editor.hasModel()){e("returning early - no model after rename edits are provided");return}if(f.rejectReason){e(`returning early - rejected with reason: ${f.rejectReason}`),this._notificationService.info(f.rejectReason);return}this.editor.setSelection(k.fromPositions(this.editor.getSelection().getPosition())),e("applying edits"),this._bulkEditService.apply(f,{editor:this.editor,showPreview:u.wantsPreview,label:m(1382,"Renaming '{0}' to '{1}'",o?.text,u.newName),code:"undoredo.rename",quotableLabel:m(1383,"Renaming {0} to {1}",o?.text,u.newName),respectAutoSaveConfig:!0,reason:cs.rename()}).then(p=>{e("edits applied"),p.ariaSummary&&Js(m(1384,"Successfully renamed '{0}' to '{1}'. Summary: {2}",o.text,u.newName,p.ariaSummary))}).catch(p=>{e(`error when applying edits ${JSON.stringify(p,null,"	")}`),this._notificationService.error(m(1385,"Rename failed to apply edits")),this._logService.error(p)})},f=>{e("error when providing rename edits",JSON.stringify(f,null,"	")),this._notificationService.error(m(1386,"Rename failed to compute edits")),this._logService.error(f)}).finally(()=>{r.dispose()});return e("returning rename operation"),this._progressService.showWhile(g,250),g}acceptRenameInput(e){this._renameWidget.acceptInput(e)}cancelRenameInput(){this._renameWidget.cancelInput(!0,"cancelRenameInput command")}focusNextRenameSuggestion(){this._renameWidget.focusNextRenameSuggestion()}focusPreviousRenameSuggestion(){this._renameWidget.focusPreviousRenameSuggestion()}},w8=am,am.ID="editor.contrib.renameController",am);Cg=w8=L5e([Xg(1,_e),Xg(2,Ii),Xg(3,fx),Xg(4,Ch),Xg(5,jt),Xg(6,QA),Xg(7,ge)],Cg);class E5e extends ue{constructor(){super({id:"editor.action.rename",label:V(1388,"Rename Symbol"),precondition:K.and(N.writable,N.hasRenameProvider),kbOpts:{kbExpr:N.editorTextFocus,primary:60,weight:100},contextMenuOpts:{group:"1_modification",order:1.1},canTriggerInlineEdits:!0})}runCommand(e,t){const i=e.get(ht),[n,o]=Array.isArray(t)&&t||[void 0,void 0];return Le.isUri(n)&&A.isIPosition(o)?i.openCodeEditor({resource:n},i.getActiveCodeEditor()).then(r=>{r&&(r.setPosition(o),r.invokeWithinContext(a=>(this.reportTelemetry(a,r),this.run(a,r))))},Oe):super.runCommand(e,t)}run(e,t){const i=e.get(jt),n=Cg.get(t);return n?(i.trace("[RenameAction] got controller, running..."),n.run()):(i.trace("[RenameAction] returning early - controller missing"),Promise.resolve())}}lt(Cg.ID,Cg,4);ae(E5e);const sV=hn.bindToContribution(Cg.get);le(new sV({id:"acceptRenameInput",precondition:pC,handler:s=>s.acceptRenameInput(!1),kbOpts:{weight:199,kbExpr:K.and(N.focus,K.not("isComposing")),primary:3}}));le(new sV({id:"acceptRenameInputWithPreview",precondition:K.and(pC,K.has("config.editor.rename.enablePreview")),handler:s=>s.acceptRenameInput(!0),kbOpts:{weight:199,kbExpr:K.and(N.focus,K.not("isComposing")),primary:2051}}));le(new sV({id:"cancelRenameInput",precondition:pC,handler:s=>s.cancelRenameInput(),kbOpts:{weight:199,kbExpr:N.focus,primary:9,secondary:[1033]}}));It(class extends Tn{constructor(){super({id:"focusNextRenameSuggestion",title:{...V(1389,"Focus Next Rename Suggestion")},precondition:pC,keybinding:[{primary:18,weight:199}]})}run(e){const t=e.get(ht).getFocusedCodeEditor();if(!t)return;const i=Cg.get(t);i&&i.focusNextRenameSuggestion()}});It(class extends Tn{constructor(){super({id:"focusPreviousRenameSuggestion",title:{...V(1390,"Focus Previous Rename Suggestion")},precondition:pC,keybinding:[{primary:16,weight:199}]})}run(e){const t=e.get(ht).getFocusedCodeEditor();if(!t)return;const i=Cg.get(t);i&&i.focusPreviousRenameSuggestion()}});Lo("_executeDocumentRenameProvider",function(s,e,t,...i){const[n]=i;dt(typeof n=="string");const{renameProvider:o}=s.get(ge);return k5e(o,e,t,n)});Lo("_executePrepareRename",async function(s,e,t){const{renameProvider:i}=s.get(ge),o=await new nV(e,t,i).resolveRenameLocation(Ke.None);if(o?.rejectReason)throw new Error(o.rejectReason);return o});bi.as(Ml.Configuration).registerConfiguration({id:"editor",properties:{"editor.rename.enablePreview":{scope:6,description:m(1387,"Enable/disable the ability to preview changes before renaming"),default:!0,type:"boolean"}}});var I5e=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},DK=function(s,e){return function(t,i){e(t,i,s)}},Yw;let JD=(Yw=class extends F{constructor(e,t,i){super(),this.editor=e,this.languageConfigurationService=t,this.editorWorkerService=i,this.decorations=this.editor.createDecorationsCollection(),this.options=this.createOptions(e.getOption(81)),this.computePromise=null,this.currentOccurrences={},this._register(e.onDidChangeModel(n=>{this.currentOccurrences={},this.options=this.createOptions(e.getOption(81)),this.stop(),this.computeSectionHeaders.schedule(0)})),this._register(e.onDidChangeModelLanguage(n=>{this.currentOccurrences={},this.options=this.createOptions(e.getOption(81)),this.stop(),this.computeSectionHeaders.schedule(0)})),this._register(t.onDidChange(n=>{const o=this.editor.getModel()?.getLanguageId();o&&n.affects(o)&&(this.currentOccurrences={},this.options=this.createOptions(e.getOption(81)),this.stop(),this.computeSectionHeaders.schedule(0))})),this._register(e.onDidChangeConfiguration(n=>{this.options&&!n.hasChanged(81)||(this.options=this.createOptions(e.getOption(81)),this.updateDecorations([]),this.stop(),this.computeSectionHeaders.schedule(0))})),this._register(this.editor.onDidChangeModelContent(n=>{this.computeSectionHeaders.schedule()})),this._register(e.onDidChangeModelTokens(n=>{this.computeSectionHeaders.isScheduled()||this.computeSectionHeaders.schedule(1e3)})),this.computeSectionHeaders=this._register(new Rt(()=>{this.findSectionHeaders()},250)),this.computeSectionHeaders.schedule(0)}createOptions(e){if(!e||!this.editor.hasModel())return;const t=this.editor.getModel().getLanguageId();if(!t)return;const i=this.languageConfigurationService.getLanguageConfiguration(t).comments,n=this.languageConfigurationService.getLanguageConfiguration(t).foldingRules;if(!(!i&&!n?.markers))return{foldingRules:n,markSectionHeaderRegex:e.markSectionHeaderRegex,findMarkSectionHeaders:e.showMarkSectionHeaders,findRegionSectionHeaders:e.showRegionSectionHeaders}}findSectionHeaders(){if(!this.editor.hasModel()||!this.options?.findMarkSectionHeaders&&!this.options?.findRegionSectionHeaders)return;const e=this.editor.getModel();if(e.isDisposed()||e.isTooLargeForSyncing())return;const t=e.getVersionId();this.editorWorkerService.findSectionHeaders(e.uri,this.options).then(i=>{e.isDisposed()||e.getVersionId()!==t||this.updateDecorations(i)})}updateDecorations(e){const t=this.editor.getModel();t&&(e=e.filter(o=>{if(!o.shouldBeInComments)return!0;const r=t.validateRange(o.range),a=t.tokenization.getLineTokens(r.startLineNumber),l=a.findTokenIndexAtOffset(r.startColumn-1),c=a.getStandardTokenType(l);return a.getLanguageId(l)===t.getLanguageId()&&c===1}));const i=Object.values(this.currentOccurrences).map(o=>o.decorationId),n=e.map(o=>D5e(o));this.editor.changeDecorations(o=>{const r=o.deltaDecorations(i,n);this.currentOccurrences={};for(let a=0,l=r.length;a<l;a++){const c={sectionHeader:e[a],decorationId:r[a]};this.currentOccurrences[c.decorationId]=c}})}stop(){this.computeSectionHeaders.cancel(),this.computePromise&&(this.computePromise.cancel(),this.computePromise=null)}dispose(){super.dispose(),this.stop(),this.decorations.clear()}},Yw.ID="editor.sectionHeaderDetector",Yw);JD=I5e([DK(1,ui),DK(2,ko)],JD);function D5e(s){return{range:s.range,options:Fe.createDynamic({description:"section-header",stickiness:3,collapseOnReplaceEdit:!0,minimap:{color:void 0,position:1,sectionHeaderStyle:s.hasSeparatorLine?2:1,sectionHeaderText:s.text}})}}lt(JD.ID,JD,1);class T5e{constructor(e=5){this.timesPerSecond=e,this._lastRun=0,this._minimumTimeBetweenRuns=1e3/e}runIfNotLimited(e){const t=Date.now();t-this._lastRun>=this._minimumTimeBetweenRuns&&(this._lastRun=t,e())}}class dy{static create(e,t){return new dy(e,new v8(t))}get startLineNumber(){return this._startLineNumber}get endLineNumber(){return this._endLineNumber}constructor(e,t){this._startLineNumber=e,this._tokens=t,this._endLineNumber=this._startLineNumber+this._tokens.getMaxDeltaLine()}toString(){return this._tokens.toString(this._startLineNumber)}_updateEndLineNumber(){this._endLineNumber=this._startLineNumber+this._tokens.getMaxDeltaLine()}isEmpty(){return this._tokens.isEmpty()}getLineTokens(e){return this._startLineNumber<=e&&e<=this._endLineNumber?this._tokens.getLineTokens(e-this._startLineNumber):null}getRange(){const e=this._tokens.getRange();return e&&new k(this._startLineNumber+e.startLineNumber,e.startColumn,this._startLineNumber+e.endLineNumber,e.endColumn)}removeTokens(e){const t=e.startLineNumber-this._startLineNumber,i=e.endLineNumber-this._startLineNumber;this._startLineNumber+=this._tokens.removeTokens(t,e.startColumn-1,i,e.endColumn-1),this._updateEndLineNumber()}split(e){const t=e.startLineNumber-this._startLineNumber,i=e.endLineNumber-this._startLineNumber,[n,o,r]=this._tokens.split(t,e.startColumn-1,i,e.endColumn-1);return[new dy(this._startLineNumber,n),new dy(this._startLineNumber+r,o)]}applyEdit(e,t){const[i,n,o]=fg(t);this.acceptEdit(e,i,n,o,t.length>0?t.charCodeAt(0):0)}acceptEdit(e,t,i,n,o){this._acceptDeleteRange(e),this._acceptInsertText(new A(e.startLineNumber,e.startColumn),t,i,n,o),this._updateEndLineNumber()}_acceptDeleteRange(e){if(e.startLineNumber===e.endLineNumber&&e.startColumn===e.endColumn)return;const t=e.startLineNumber-this._startLineNumber,i=e.endLineNumber-this._startLineNumber;if(i<0){const o=i-t;this._startLineNumber-=o;return}const n=this._tokens.getMaxDeltaLine();if(!(t>=n+1)){if(t<0&&i>=n+1){this._startLineNumber=0,this._tokens.clear();return}if(t<0){const o=-t;this._startLineNumber-=o,this._tokens.acceptDeleteRange(e.startColumn-1,0,0,i,e.endColumn-1)}else this._tokens.acceptDeleteRange(0,t,e.startColumn-1,i,e.endColumn-1)}}_acceptInsertText(e,t,i,n,o){if(t===0&&i===0)return;const r=e.lineNumber-this._startLineNumber;if(r<0){this._startLineNumber+=t;return}const a=this._tokens.getMaxDeltaLine();r>=a+1||this._tokens.acceptInsertText(r,e.column-1,t,i,n,o)}reportIfInvalid(e){this._tokens.reportIfInvalid(e,this._startLineNumber)}}const su=class su{constructor(e){this._tokens=e,this._tokenCount=e.length/4}toString(e){const t=[];for(let i=0;i<this._tokenCount;i++)t.push(`(${this._getDeltaLine(i)+e},${this._getStartCharacter(i)}-${this._getEndCharacter(i)})`);return`[${t.join(",")}]`}getMaxDeltaLine(){const e=this._getTokenCount();return e===0?-1:this._getDeltaLine(e-1)}getRange(){const e=this._getTokenCount();if(e===0)return null;const t=this._getStartCharacter(0),i=this._getDeltaLine(e-1),n=this._getEndCharacter(e-1);return new k(0,t+1,i,n+1)}_getTokenCount(){return this._tokenCount}_getDeltaLine(e){return this._tokens[4*e]}_getStartCharacter(e){return this._tokens[4*e+1]}_getEndCharacter(e){return this._tokens[4*e+2]}isEmpty(){return this._getTokenCount()===0}getLineTokens(e){let t=0,i=this._getTokenCount()-1;for(;t<i;){const n=t+Math.floor((i-t)/2),o=this._getDeltaLine(n);if(o<e)t=n+1;else if(o>e)i=n-1;else{let r=n;for(;r>t&&this._getDeltaLine(r-1)===e;)r--;let a=n;for(;a<i&&this._getDeltaLine(a+1)===e;)a++;return new TK(this._tokens.subarray(4*r,4*a+4))}}return this._getDeltaLine(t)===e?new TK(this._tokens.subarray(4*t,4*t+4)):null}clear(){this._tokenCount=0}removeTokens(e,t,i,n){const o=this._tokens,r=this._tokenCount;let a=0,l=!1,c=0;for(let d=0;d<r;d++){const h=4*d,u=o[h],g=o[h+1],f=o[h+2],p=o[h+3];if((u>e||u===e&&f>=t)&&(u<i||u===i&&g<=n))l=!0;else{if(a===0&&(c=u),l){const _=4*a;o[_]=u-c,o[_+1]=g,o[_+2]=f,o[_+3]=p}else c!==0&&(o[h]=u-c);a++}}return this._tokenCount=a,c}split(e,t,i,n){const o=this._tokens,r=this._tokenCount,a=[],l=[];let c=a,d=0,h=0;for(let u=0;u<r;u++){const g=4*u,f=o[g],p=o[g+1],_=o[g+2],b=o[g+3];if(f>e||f===e&&_>=t){if(f<i||f===i&&p<=n)continue;c!==l&&(c=l,d=0,h=f)}c[d++]=f-h,c[d++]=p,c[d++]=_,c[d++]=b}return[new su(new Uint32Array(a)),new su(new Uint32Array(l)),h]}acceptDeleteRange(e,t,i,n,o){const r=this._tokens,a=this._tokenCount,l=n-t;let c=0,d=!1;for(let h=0;h<a;h++){const u=4*h;let g=r[u],f=r[u+1],p=r[u+2];const _=r[u+3];if(g<t||g===t&&p<=i){c++;continue}else if(g===t&&f<i)g===n&&p>o?p-=o-i:p=i;else if(g===t&&f===i)if(g===n&&p>o)p-=o-i;else{d=!0;continue}else if(g<n||g===n&&f<o)if(g===n&&p>o)g=t,f=i,p=f+(p-o);else{d=!0;continue}else if(g>n){if(l===0&&!d){c=a;break}g-=l}else if(g===n&&f>=o)e&&g===0&&(f+=e,p+=e),g-=l,f-=o-i,p-=o-i;else throw new Error("Not possible!");const b=4*c;r[b]=g,r[b+1]=f,r[b+2]=p,r[b+3]=_,c++}this._tokenCount=c}acceptInsertText(e,t,i,n,o,r){const a=i===0&&n===1&&(r>=48&&r<=57||r>=65&&r<=90||r>=97&&r<=122),l=this._tokens,c=this._tokenCount;for(let d=0;d<c;d++){const h=4*d;let u=l[h],g=l[h+1],f=l[h+2];if(!(u<e||u===e&&f<t)){if(u===e&&f===t)if(a)f+=1;else continue;else if(u===e&&g<t&&t<f)i===0?f+=n:f=t;else{if(u===e&&g===t&&a)continue;if(u===e)if(u+=i,i===0)g+=n,f+=n;else{const p=f-g;g=o+(g-t),f=g+p}else u+=i}l[h]=u,l[h+1]=g,l[h+2]=f}}}reportIfInvalid(e,t){for(let i=0;i<this._tokenCount;i++){const n=this._getDeltaLine(i)+t;n<1?su._rateLimiter.runIfNotLimited(()=>{console.error("Invalid Semantic Tokens Data From Extension: lineNumber < 1")}):n>e.getLineCount()?su._rateLimiter.runIfNotLimited(()=>{console.error("Invalid Semantic Tokens Data From Extension: lineNumber > model.getLineCount()")}):this._getEndCharacter(i)>e.getLineLength(n)&&su._rateLimiter.runIfNotLimited(()=>{console.error("Invalid Semantic Tokens Data From Extension: end character > model.getLineLength(lineNumber)")})}}};su._rateLimiter=new T5e(10/60);let v8=su;class TK{constructor(e){this._tokens=e}getCount(){return this._tokens.length/4}getStartCharacter(e){return this._tokens[4*e+1]}getEndCharacter(e){return this._tokens[4*e+2]}getMetadata(e){return this._tokens[4*e+3]}}var N5e=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},IO=function(s,e){return function(t,i){e(t,i,s)}};let C8=class{constructor(e,t,i,n){this._legend=e,this._themeService=t,this._languageService=i,this._logService=n,this._hasWarnedOverlappingTokens=!1,this._hasWarnedInvalidLengthTokens=!1,this._hasWarnedInvalidEditStart=!1,this._hashTable=new y8}getMetadata(e,t,i){const n=this._languageService.languageIdCodec.encodeLanguageId(i),o=this._hashTable.get(e,t,n);let r;if(o)r=o.metadata;else{let a=this._legend.tokenTypes[e];const l=[];if(a){let c=t;for(let h=0;c>0&&h<this._legend.tokenModifiers.length;h++)c&1&&l.push(this._legend.tokenModifiers[h]),c=c>>1;const d=this._themeService.getColorTheme().getTokenStyleMetadata(a,l,i);if(typeof d>"u")r=2147483647;else{if(r=0,typeof d.italic<"u"){const h=(d.italic?1:0)<<11;r|=h|1}if(typeof d.bold<"u"){const h=(d.bold?2:0)<<11;r|=h|2}if(typeof d.underline<"u"){const h=(d.underline?4:0)<<11;r|=h|4}if(typeof d.strikethrough<"u"){const h=(d.strikethrough?8:0)<<11;r|=h|8}if(d.foreground){const h=d.foreground<<15;r|=h|16}r===0&&(r=2147483647)}}else r=2147483647,a="not-in-legend";this._hashTable.add(e,t,n,r)}return r}warnOverlappingSemanticTokens(e,t){this._hasWarnedOverlappingTokens||(this._hasWarnedOverlappingTokens=!0,this._logService.warn(`Overlapping semantic tokens detected at lineNumber ${e}, column ${t}`))}warnInvalidLengthSemanticTokens(e,t){this._hasWarnedInvalidLengthTokens||(this._hasWarnedInvalidLengthTokens=!0,this._logService.warn(`Semantic token with invalid length detected at lineNumber ${e}, column ${t}`))}warnInvalidEditStart(e,t,i,n,o){this._hasWarnedInvalidEditStart||(this._hasWarnedInvalidEditStart=!0,this._logService.warn(`Invalid semantic tokens edit detected (previousResultId: ${e}, resultId: ${t}) at edit #${i}: The provided start offset ${n} is outside the previous data (length ${o}).`))}};C8=N5e([IO(1,wi),IO(2,Ei),IO(3,jt)],C8);function Hoe(s,e,t){const i=s.data,n=s.data.length/5|0,o=Math.max(Math.ceil(n/1024),400),r=[];let a=0,l=1,c=0;for(;a<n;){const d=a;let h=Math.min(d+o,n);if(h<n){let w=h;for(;w-1>d&&i[5*w]===0;)w--;if(w-1===d){let C=h;for(;C+1<n&&i[5*C]===0;)C++;h=C}else h=w}let u=new Uint32Array((h-d)*4),g=0,f=0,p=0,_=0;for(;a<h;){const w=5*a,C=i[w],v=i[w+1],y=l+C|0,x=C===0?c+v|0:v,L=i[w+2],E=x+L|0,I=i[w+3],R=i[w+4];if(E<=x)e.warnInvalidLengthSemanticTokens(y,x+1);else if(p===y&&_>x)e.warnOverlappingSemanticTokens(y,x+1);else{const O=e.getMetadata(I,R,t);O!==2147483647&&(f===0&&(f=y),u[g]=y-f,u[g+1]=x,u[g+2]=E,u[g+3]=O,g+=4,p=y,_=E)}l=y,c=x,a++}g!==u.length&&(u=u.subarray(0,g));const b=dy.create(f,u);r.push(b)}return r}class R5e{constructor(e,t,i,n){this.tokenTypeIndex=e,this.tokenModifierSet=t,this.languageId=i,this.metadata=n,this.next=null}}const Ed=class Ed{constructor(){this._elementsCount=0,this._currentLengthIndex=0,this._currentLength=Ed._SIZES[this._currentLengthIndex],this._growCount=Math.round(this._currentLengthIndex+1<Ed._SIZES.length?2/3*this._currentLength:0),this._elements=[],Ed._nullOutEntries(this._elements,this._currentLength)}static _nullOutEntries(e,t){for(let i=0;i<t;i++)e[i]=null}_hash2(e,t){return(e<<5)-e+t|0}_hashFunc(e,t,i){return this._hash2(this._hash2(e,t),i)%this._currentLength}get(e,t,i){const n=this._hashFunc(e,t,i);let o=this._elements[n];for(;o;){if(o.tokenTypeIndex===e&&o.tokenModifierSet===t&&o.languageId===i)return o;o=o.next}return null}add(e,t,i,n){if(this._elementsCount++,this._growCount!==0&&this._elementsCount>=this._growCount){const o=this._elements;this._currentLengthIndex++,this._currentLength=Ed._SIZES[this._currentLengthIndex],this._growCount=Math.round(this._currentLengthIndex+1<Ed._SIZES.length?2/3*this._currentLength:0),this._elements=[],Ed._nullOutEntries(this._elements,this._currentLength);for(const r of o){let a=r;for(;a;){const l=a.next;a.next=null,this._add(a),a=l}}}this._add(new R5e(e,t,i,n))}_add(e){const t=this._hashFunc(e.tokenTypeIndex,e.tokenModifierSet,e.languageId);e.next=this._elements[t],this._elements[t]=e}};Ed._SIZES=[3,7,13,31,61,127,251,509,1021,2039,4093,8191,16381,32749,65521,131071,262139,524287,1048573,2097143];let y8=Ed;const JA=qe("semanticTokensStylingService");function A5e(s){for(let e=0,t=s.length;e<t;e+=4){const i=s[e+0],n=s[e+1],o=s[e+2],r=s[e+3];s[e+0]=r,s[e+1]=o,s[e+2]=n,s[e+3]=i}}function M5e(s){const e=new Uint8Array(s.buffer,s.byteOffset,s.length*4);return VX()||A5e(e),Ru.wrap(e)}function Voe(s){const e=new Uint32Array(O5e(s));let t=0;if(e[t++]=s.id,s.type==="full")e[t++]=1,e[t++]=s.data.length,e.set(s.data,t),t+=s.data.length;else{e[t++]=2,e[t++]=s.deltas.length;for(const i of s.deltas)e[t++]=i.start,e[t++]=i.deleteCount,i.data?(e[t++]=i.data.length,e.set(i.data,t),t+=i.data.length):e[t++]=0}return M5e(e)}function O5e(s){let e=0;if(e+=2,s.type==="full")e+=1+s.data.length;else{e+=1,e+=3*s.deltas.length;for(const t of s.deltas)t.data&&(e+=t.data.length)}return e}function e2(s){return s&&!!s.data}function zoe(s){return s&&Array.isArray(s.edits)}class P5e{constructor(e,t,i){this.provider=e,this.tokens=t,this.error=i}}function $oe(s,e){return s.has(e)}function F5e(s,e){const t=s.orderedGroups(e);return t.length>0?t[0]:[]}async function Uoe(s,e,t,i,n){const o=F5e(s,e),r=await Promise.all(o.map(async a=>{let l,c=null;try{l=await a.provideDocumentSemanticTokens(e,a===t?i:null,n)}catch(d){c=d,l=null}return(!l||!e2(l)&&!zoe(l))&&(l=null),new P5e(a,l,c)}));for(const a of r){if(a.error)throw a.error;if(a.tokens)return a}return r.length>0?r[0]:null}function B5e(s,e){const t=s.orderedGroups(e);return t.length>0?t[0]:null}class W5e{constructor(e,t){this.provider=e,this.tokens=t}}function H5e(s,e){return s.has(e)}function qoe(s,e){const t=s.orderedGroups(e);return t.length>0?t[0]:[]}async function oV(s,e,t,i){const n=qoe(s,e),o=await Promise.all(n.map(async r=>{let a;try{a=await r.provideDocumentRangeSemanticTokens(e,t,i)}catch(l){ji(l),a=null}return(!a||!e2(a))&&(a=null),new W5e(r,a)}));for(const r of o)if(r.tokens)return r;return o.length>0?o[0]:null}at.registerCommand("_provideDocumentSemanticTokensLegend",async(s,...e)=>{const[t]=e;dt(t instanceof Le);const i=s.get(hi).getModel(t);if(!i)return;const{documentSemanticTokensProvider:n}=s.get(ge),o=B5e(n,i);return o?o[0].getLegend():s.get(Gt).executeCommand("_provideDocumentRangeSemanticTokensLegend",t)});at.registerCommand("_provideDocumentSemanticTokens",async(s,...e)=>{const[t]=e;dt(t instanceof Le);const i=s.get(hi).getModel(t);if(!i)return;const{documentSemanticTokensProvider:n}=s.get(ge);if(!$oe(n,i))return s.get(Gt).executeCommand("_provideDocumentRangeSemanticTokens",t,i.getFullModelRange());const o=await Uoe(n,i,null,null,Ke.None);if(!o)return;const{provider:r,tokens:a}=o;if(!a||!e2(a))return;const l=Voe({id:0,type:"full",data:a.data});return a.resultId&&r.releaseDocumentSemanticTokens(a.resultId),l});at.registerCommand("_provideDocumentRangeSemanticTokensLegend",async(s,...e)=>{const[t,i]=e;dt(t instanceof Le);const n=s.get(hi).getModel(t);if(!n)return;const{documentRangeSemanticTokensProvider:o}=s.get(ge),r=qoe(o,n);if(r.length===0)return;if(r.length===1)return r[0].getLegend();if(!i||!k.isIRange(i))return console.warn("provideDocumentRangeSemanticTokensLegend might be out-of-sync with provideDocumentRangeSemanticTokens unless a range argument is passed in"),r[0].getLegend();const a=await oV(o,n,k.lift(i),Ke.None);if(a)return a.provider.getLegend()});at.registerCommand("_provideDocumentRangeSemanticTokens",async(s,...e)=>{const[t,i]=e;dt(t instanceof Le),dt(k.isIRange(i));const n=s.get(hi).getModel(t);if(!n)return;const{documentRangeSemanticTokensProvider:o}=s.get(ge),r=await oV(o,n,k.lift(i),Ke.None);if(!(!r||!r.tokens))return Voe({id:0,type:"full",data:r.tokens.data})});const rV="editor.semanticHighlighting";function Jk(s,e,t){const i=t.getValue(rV,{overrideIdentifier:s.getLanguageId(),resource:s.uri})?.enabled;return typeof i=="boolean"?i:e.getColorTheme().semanticHighlighting}var joe=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},cc=function(s,e){return function(t,i){e(t,i,s)}},Kh;let S8=class extends F{constructor(e,t,i,n,o,r){super(),this._watchers=new Wi;const a=d=>{this._watchers.get(d.uri)?.dispose(),this._watchers.set(d.uri,new x8(d,e,i,o,r))},l=(d,h)=>{h.dispose(),this._watchers.delete(d.uri)},c=()=>{for(const d of t.getModels()){const h=this._watchers.get(d.uri);Jk(d,i,n)?h||a(d):h&&l(d,h)}};t.getModels().forEach(d=>{Jk(d,i,n)&&a(d)}),this._register(t.onModelAdded(d=>{Jk(d,i,n)&&a(d)})),this._register(t.onModelRemoved(d=>{const h=this._watchers.get(d.uri);h&&l(d,h)})),this._register(n.onDidChangeConfiguration(d=>{d.affectsConfiguration(rV)&&c()})),this._register(i.onDidColorThemeChange(c))}dispose(){xt(this._watchers.values()),this._watchers.clear(),super.dispose()}};S8=joe([cc(0,JA),cc(1,hi),cc(2,wi),cc(3,He),cc(4,ga),cc(5,ge)],S8);var sg;let x8=(sg=class extends F{constructor(e,t,i,n,o){super(),this._semanticTokensStylingService=t,this._isDisposed=!1,this._model=e,this._provider=o.documentSemanticTokensProvider,this._debounceInformation=n.for(this._provider,"DocumentSemanticTokens",{min:Kh.REQUEST_MIN_DELAY,max:Kh.REQUEST_MAX_DELAY}),this._fetchDocumentSemanticTokens=this._register(new Rt(()=>this._fetchDocumentSemanticTokensNow(),Kh.REQUEST_MIN_DELAY)),this._currentDocumentResponse=null,this._currentDocumentRequestCancellationTokenSource=null,this._documentProvidersChangeListeners=[],this._providersChangedDuringRequest=!1,this._register(this._model.onDidChangeContent(()=>{this._fetchDocumentSemanticTokens.isScheduled()||this._fetchDocumentSemanticTokens.schedule(this._debounceInformation.get(this._model))})),this._register(this._model.onDidChangeAttached(()=>{this._fetchDocumentSemanticTokens.isScheduled()||this._fetchDocumentSemanticTokens.schedule(this._debounceInformation.get(this._model))})),this._register(this._model.onDidChangeLanguage(()=>{this._currentDocumentResponse&&(this._currentDocumentResponse.dispose(),this._currentDocumentResponse=null),this._currentDocumentRequestCancellationTokenSource&&(this._currentDocumentRequestCancellationTokenSource.cancel(),this._currentDocumentRequestCancellationTokenSource=null),this._setDocumentSemanticTokens(null,null,null,[]),this._fetchDocumentSemanticTokens.schedule(0)}));const r=()=>{xt(this._documentProvidersChangeListeners),this._documentProvidersChangeListeners=[];for(const a of this._provider.all(e))typeof a.onDidChange=="function"&&this._documentProvidersChangeListeners.push(a.onDidChange(()=>{if(this._currentDocumentRequestCancellationTokenSource){this._providersChangedDuringRequest=!0;return}this._fetchDocumentSemanticTokens.schedule(0)}))};r(),this._register(this._provider.onDidChange(()=>{r(),this._fetchDocumentSemanticTokens.schedule(this._debounceInformation.get(this._model))})),this._register(i.onDidColorThemeChange(a=>{this._setDocumentSemanticTokens(null,null,null,[]),this._fetchDocumentSemanticTokens.schedule(this._debounceInformation.get(this._model))})),this._fetchDocumentSemanticTokens.schedule(0)}dispose(){this._currentDocumentResponse&&(this._currentDocumentResponse.dispose(),this._currentDocumentResponse=null),this._currentDocumentRequestCancellationTokenSource&&(this._currentDocumentRequestCancellationTokenSource.cancel(),this._currentDocumentRequestCancellationTokenSource=null),xt(this._documentProvidersChangeListeners),this._documentProvidersChangeListeners=[],this._setDocumentSemanticTokens(null,null,null,[]),this._isDisposed=!0,super.dispose()}_fetchDocumentSemanticTokensNow(){if(this._currentDocumentRequestCancellationTokenSource)return;if(!$oe(this._provider,this._model)){this._currentDocumentResponse&&this._model.tokenization.setSemanticTokens(null,!1);return}if(!this._model.isAttachedToEditor())return;const e=new si,t=this._currentDocumentResponse?this._currentDocumentResponse.provider:null,i=this._currentDocumentResponse&&this._currentDocumentResponse.resultId||null,n=Uoe(this._provider,this._model,t,i,e.token);this._currentDocumentRequestCancellationTokenSource=e,this._providersChangedDuringRequest=!1;const o=[],r=this._model.onDidChangeContent(l=>{o.push(l)}),a=new Cn(!1);n.then(l=>{if(this._debounceInformation.update(this._model,a.elapsed()),this._currentDocumentRequestCancellationTokenSource=null,r.dispose(),!l)this._setDocumentSemanticTokens(null,null,null,o);else{const{provider:c,tokens:d}=l,h=this._semanticTokensStylingService.getStyling(c);this._setDocumentSemanticTokens(c,d||null,h,o)}},l=>{l&&(Dr(l)||typeof l.message=="string"&&l.message.indexOf("busy")!==-1)||Oe(l),this._currentDocumentRequestCancellationTokenSource=null,r.dispose(),(o.length>0||this._providersChangedDuringRequest)&&(this._fetchDocumentSemanticTokens.isScheduled()||this._fetchDocumentSemanticTokens.schedule(this._debounceInformation.get(this._model)))})}static _copy(e,t,i,n,o){o=Math.min(o,i.length-n,e.length-t);for(let r=0;r<o;r++)i[n+r]=e[t+r]}_setDocumentSemanticTokens(e,t,i,n){const o=this._currentDocumentResponse,r=()=>{(n.length>0||this._providersChangedDuringRequest)&&!this._fetchDocumentSemanticTokens.isScheduled()&&this._fetchDocumentSemanticTokens.schedule(this._debounceInformation.get(this._model))};if(this._currentDocumentResponse&&(this._currentDocumentResponse.dispose(),this._currentDocumentResponse=null),this._isDisposed){e&&t&&e.releaseDocumentSemanticTokens(t.resultId);return}if(!e||!i){this._model.tokenization.setSemanticTokens(null,!1);return}if(!t){this._model.tokenization.setSemanticTokens(null,!0),r();return}if(zoe(t)){if(!o){this._model.tokenization.setSemanticTokens(null,!0);return}if(t.edits.length===0)t={resultId:t.resultId,data:o.data};else{let a=0;for(const u of t.edits)a+=(u.data?u.data.length:0)-u.deleteCount;const l=o.data,c=new Uint32Array(l.length+a);let d=l.length,h=c.length;for(let u=t.edits.length-1;u>=0;u--){const g=t.edits[u];if(g.start>l.length){i.warnInvalidEditStart(o.resultId,t.resultId,u,g.start,l.length),this._model.tokenization.setSemanticTokens(null,!0);return}const f=d-(g.start+g.deleteCount);f>0&&(Kh._copy(l,d-f,c,h-f,f),h-=f),g.data&&(Kh._copy(g.data,0,c,h-g.data.length,g.data.length),h-=g.data.length),d=g.start}d>0&&Kh._copy(l,0,c,0,d),t={resultId:t.resultId,data:c}}}if(e2(t)){this._currentDocumentResponse=new V5e(e,t.resultId,t.data);const a=Hoe(t,i,this._model.getLanguageId());if(n.length>0)for(const l of n)for(const c of a)for(const d of l.changes)c.applyEdit(d.range,d.text);this._model.tokenization.setSemanticTokens(a,!0)}else this._model.tokenization.setSemanticTokens(null,!0);r()}},Kh=sg,sg.REQUEST_MIN_DELAY=300,sg.REQUEST_MAX_DELAY=2e3,sg);x8=Kh=joe([cc(1,JA),cc(2,wi),cc(3,ga),cc(4,ge)],x8);class V5e{constructor(e,t,i){this.provider=e,this.resultId=t,this.data=i}dispose(){this.provider.releaseDocumentSemanticTokens(this.resultId)}}uC(S8);var z5e=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},XC=function(s,e){return function(t,i){e(t,i,s)}},Zw;let eT=(Zw=class extends F{constructor(e,t,i,n,o,r){super(),this._semanticTokensStylingService=t,this._themeService=i,this._configurationService=n,this._editor=e,this._provider=r.documentRangeSemanticTokensProvider,this._debounceInformation=o.for(this._provider,"DocumentRangeSemanticTokens",{min:100,max:500}),this._tokenizeViewport=this._register(new Rt(()=>this._tokenizeViewportNow(),100)),this._outstandingRequests=[],this._rangeProvidersChangeListeners=[];const a=()=>{this._editor.hasModel()&&this._tokenizeViewport.schedule(this._debounceInformation.get(this._editor.getModel()))},l=()=>{if(this._cleanupProviderListeners(),this._editor.hasModel()){const c=this._editor.getModel();for(const d of this._provider.all(c)){const h=d.onDidChange?.(()=>{this._cancelAll(),a()});h&&this._rangeProvidersChangeListeners.push(h)}}};this._register(this._editor.onDidScrollChange(()=>{a()})),this._register(this._editor.onDidChangeModel(()=>{l(),this._cancelAll(),a()})),this._register(this._editor.onDidChangeModelLanguage(()=>{l(),this._cancelAll(),a()})),this._register(this._editor.onDidChangeModelContent(c=>{this._cancelAll(),a()})),l(),this._register(this._provider.onDidChange(()=>{l(),this._cancelAll(),a()})),this._register(this._configurationService.onDidChangeConfiguration(c=>{c.affectsConfiguration(rV)&&(this._cancelAll(),a())})),this._register(this._themeService.onDidColorThemeChange(()=>{this._cancelAll(),a()})),a()}dispose(){this._cleanupProviderListeners(),super.dispose()}_cleanupProviderListeners(){xt(this._rangeProvidersChangeListeners),this._rangeProvidersChangeListeners=[]}_cancelAll(){for(const e of this._outstandingRequests)e.cancel();this._outstandingRequests=[]}_removeOutstandingRequest(e){for(let t=0,i=this._outstandingRequests.length;t<i;t++)if(this._outstandingRequests[t]===e){this._outstandingRequests.splice(t,1);return}}_tokenizeViewportNow(){if(!this._editor.hasModel())return;const e=this._editor.getModel();if(e.tokenization.hasCompleteSemanticTokens())return;if(!Jk(e,this._themeService,this._configurationService)){e.tokenization.hasSomeSemanticTokens()&&e.tokenization.setSemanticTokens(null,!1);return}if(!H5e(this._provider,e)){e.tokenization.hasSomeSemanticTokens()&&e.tokenization.setSemanticTokens(null,!1);return}const t=this._editor.getVisibleRangesPlusViewportAboveBelow();this._outstandingRequests=this._outstandingRequests.concat(t.map(i=>this._requestRange(e,i)))}_requestRange(e,t){const i=e.getVersionId(),n=an(r=>Promise.resolve(oV(this._provider,e,t,r))),o=new Cn(!1);return n.then(r=>{if(this._debounceInformation.update(e,o.elapsed()),!r||!r.tokens||e.isDisposed()||e.getVersionId()!==i)return;const{provider:a,tokens:l}=r,c=this._semanticTokensStylingService.getStyling(a);e.tokenization.setPartialSemanticTokens(t,Hoe(l,c,e.getLanguageId()))}).then(()=>this._removeOutstandingRequest(n),()=>this._removeOutstandingRequest(n)),n}},Zw.ID="editor.contrib.viewportSemanticTokens",Zw);eT=z5e([XC(1,JA),XC(2,wi),XC(3,He),XC(4,ga),XC(5,ge)],eT);lt(eT.ID,eT,1);class $5e{constructor(e=!0){this.selectSubwords=e}provideSelectionRanges(e,t){const i=[];for(const n of t){const o=[];i.push(o),this.selectSubwords&&this._addInWordRanges(o,e,n),this._addWordRanges(o,e,n),this._addWhitespaceLine(o,e,n),o.push({range:e.getFullModelRange()})}return i}_addInWordRanges(e,t,i){const n=t.getWordAtPosition(i);if(!n)return;const{word:o,startColumn:r}=n,a=i.column-r;let l=a,c=a,d=0;for(;l>=0;l--){const h=o.charCodeAt(l);if(l!==a&&(h===95||h===45))break;if(fu(h)&&oc(d))break;d=h}for(l+=1;c<o.length;c++){const h=o.charCodeAt(c);if(oc(h)&&fu(d))break;if(h===95||h===45)break;d=h}l<c&&e.push({range:new k(i.lineNumber,r+l,i.lineNumber,r+c)})}_addWordRanges(e,t,i){const n=t.getWordAtPosition(i);n&&e.push({range:new k(i.lineNumber,n.startColumn,i.lineNumber,n.endColumn)})}_addWhitespaceLine(e,t,i){t.getLineLength(i.lineNumber)>0&&t.getLineFirstNonWhitespaceColumn(i.lineNumber)===0&&t.getLineLastNonWhitespaceColumn(i.lineNumber)===0&&e.push({range:new k(i.lineNumber,1,i.lineNumber,t.getLineMaxColumn(i.lineNumber))})}}var U5e=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},q5e=function(s,e){return function(t,i){e(t,i,s)}},L8;class aV{constructor(e,t){this.index=e,this.ranges=t}mov(e){const t=this.index+(e?1:-1);if(t<0||t>=this.ranges.length)return this;const i=new aV(t,this.ranges);return i.ranges[t].equalsRange(this.ranges[this.index])?i.mov(e):i}}var lm;let p0=(lm=class{static get(e){return e.getContribution(L8.ID)}constructor(e,t){this._editor=e,this._languageFeaturesService=t,this._ignoreSelection=!1}dispose(){this._selectionListener?.dispose()}async run(e){if(!this._editor.hasModel())return;const t=this._editor.getSelections(),i=this._editor.getModel();if(this._state||await Koe(this._languageFeaturesService.selectionRangeProvider,i,t.map(o=>o.getPosition()),this._editor.getOption(129),Ke.None).then(o=>{if(!(!Ms(o)||o.length!==t.length)&&!(!this._editor.hasModel()||!ni(this._editor.getSelections(),t,(r,a)=>r.equalsSelection(a)))){for(let r=0;r<o.length;r++)o[r]=o[r].filter(a=>a.containsPosition(t[r].getStartPosition())&&a.containsPosition(t[r].getEndPosition())),o[r].unshift(t[r]);this._state=o.map(r=>new aV(0,r)),this._selectionListener?.dispose(),this._selectionListener=this._editor.onDidChangeCursorPosition(()=>{this._ignoreSelection||(this._selectionListener?.dispose(),this._state=void 0)})}}),!this._state)return;this._state=this._state.map(o=>o.mov(e));const n=this._state.map(o=>he.fromPositions(o.ranges[o.index].getStartPosition(),o.ranges[o.index].getEndPosition()));this._ignoreSelection=!0;try{this._editor.setSelections(n)}finally{this._ignoreSelection=!1}}},L8=lm,lm.ID="editor.contrib.smartSelectController",lm);p0=L8=U5e([q5e(1,ge)],p0);class Goe extends ue{constructor(e,t){super(t),this._forward=e}async run(e,t){const i=p0.get(t);i&&await i.run(this._forward)}}class j5e extends Goe{constructor(){super(!0,{id:"editor.action.smartSelect.expand",label:V(1400,"Expand Selection"),precondition:void 0,kbOpts:{kbExpr:N.editorTextFocus,primary:1553,mac:{primary:3345,secondary:[1297]},weight:100},menuOpts:{menuId:fe.MenubarSelectionMenu,group:"1_basic",title:m(1398,"&&Expand Selection"),order:2}})}}at.registerCommandAlias("editor.action.smartSelect.grow","editor.action.smartSelect.expand");class G5e extends Goe{constructor(){super(!1,{id:"editor.action.smartSelect.shrink",label:V(1401,"Shrink Selection"),precondition:void 0,kbOpts:{kbExpr:N.editorTextFocus,primary:1551,mac:{primary:3343,secondary:[1295]},weight:100},menuOpts:{menuId:fe.MenubarSelectionMenu,group:"1_basic",title:m(1399,"&&Shrink Selection"),order:3}})}}lt(p0.ID,p0,4);ae(j5e);ae(G5e);async function Koe(s,e,t,i,n){const o=s.all(e).concat(new $5e(i.selectSubwords));o.length===1&&o.unshift(new BD);const r=[],a=[];for(const l of o)r.push(Promise.resolve(l.provideSelectionRanges(e,t,n)).then(c=>{if(Ms(c)&&c.length===t.length)for(let d=0;d<t.length;d++){a[d]||(a[d]=[]);for(const h of c[d])k.isIRange(h.range)&&k.containsPosition(h.range,t[d])&&a[d].push(k.lift(h.range))}},ji));return await Promise.all(r),a.map(l=>{if(l.length===0)return[];l.sort((u,g)=>A.isBefore(u.getStartPosition(),g.getStartPosition())?1:A.isBefore(g.getStartPosition(),u.getStartPosition())||A.isBefore(u.getEndPosition(),g.getEndPosition())?-1:A.isBefore(g.getEndPosition(),u.getEndPosition())?1:0);const c=[];let d;for(const u of l)(!d||k.containsRange(u,d)&&!k.equalsRange(u,d))&&(c.push(u),d=u);if(!i.selectLeadingAndTrailingWhitespace)return c;const h=[c[0]];for(let u=1;u<c.length;u++){const g=c[u-1],f=c[u];if(f.startLineNumber!==g.startLineNumber||f.endLineNumber!==g.endLineNumber){const p=new k(g.startLineNumber,e.getLineFirstNonWhitespaceColumn(g.startLineNumber),g.endLineNumber,e.getLineLastNonWhitespaceColumn(g.endLineNumber));p.containsRange(g)&&!p.equalsRange(g)&&f.containsRange(p)&&!f.equalsRange(p)&&h.push(p);const _=new k(g.startLineNumber,1,g.endLineNumber,e.getLineMaxColumn(g.endLineNumber));_.containsRange(g)&&!_.equalsRange(p)&&f.containsRange(_)&&!f.equalsRange(_)&&h.push(_)}h.push(f)}return h})}at.registerCommand("_executeSelectionRangeProvider",async function(s,...e){const[t,i]=e;dt(Le.isUri(t)),dt(Vce(i,r=>A.isIPosition(r)));const n=s.get(ge).selectionRangeProvider,o=await s.get(Fs).createModelReference(t);try{return Koe(n,o.object.textEditorModel,i.map(A.lift),{selectLeadingAndTrailingWhitespace:!0,selectSubwords:!0},Ke.None)}finally{o.dispose()}});const k8=Object.freeze({View:V(1638,"View"),Help:V(1639,"Help"),Test:V(1640,"Test"),File:V(1641,"File"),Preferences:V(1642,"Preferences"),Developer:V(1643,"Developer")});class K5e extends za{constructor(){super({id:"editor.action.toggleStickyScroll",title:{...V(1448,"Toggle Editor Sticky Scroll"),mnemonicTitle:m(1444,"&&Toggle Editor Sticky Scroll")},metadata:{description:V(1449,"Toggle/enable the editor sticky scroll which shows the nested scopes at the top of the viewport")},category:k8.View,toggled:{condition:K.equals("config.editor.stickyScroll.enabled",!0),title:m(1445,"Sticky Scroll"),mnemonicTitle:m(1446,"&&Sticky Scroll")},menu:[{id:fe.CommandPalette},{id:fe.MenubarAppearanceMenu,group:"4_editor",order:3},{id:fe.StickyScrollContext}]})}async runEditorCommand(e,t){const i=e.get(He),n=!i.getValue("editor.stickyScroll.enabled"),o=Ha.get(t)?.isFocused();i.updateValue("editor.stickyScroll.enabled",n),o&&t.focus()}}const t2=100;class Y5e extends za{constructor(){super({id:"editor.action.focusStickyScroll",title:{...V(1450,"Focus Editor Sticky Scroll"),mnemonicTitle:m(1447,"&&Focus Editor Sticky Scroll")},precondition:K.and(K.has("config.editor.stickyScroll.enabled"),N.stickyScrollVisible),menu:[{id:fe.CommandPalette}]})}runEditorCommand(e,t){Ha.get(t)?.focus()}}class Z5e extends za{constructor(){super({id:"editor.action.selectNextStickyScrollLine",title:V(1451,"Select the next editor sticky scroll line"),precondition:N.stickyScrollFocused.isEqualTo(!0),keybinding:{weight:t2,primary:18}})}runEditorCommand(e,t){Ha.get(t)?.focusNext()}}class X5e extends za{constructor(){super({id:"editor.action.selectPreviousStickyScrollLine",title:V(1452,"Select the previous sticky scroll line"),precondition:N.stickyScrollFocused.isEqualTo(!0),keybinding:{weight:t2,primary:16}})}runEditorCommand(e,t){Ha.get(t)?.focusPrevious()}}class Q5e extends za{constructor(){super({id:"editor.action.goToFocusedStickyScrollLine",title:V(1453,"Go to the focused sticky scroll line"),precondition:N.stickyScrollFocused.isEqualTo(!0),keybinding:{weight:t2,primary:3}})}runEditorCommand(e,t){Ha.get(t)?.goToFocused()}}class J5e extends za{constructor(){super({id:"editor.action.selectEditor",title:V(1454,"Select Editor"),precondition:N.stickyScrollFocused.isEqualTo(!0),keybinding:{weight:t2,primary:9}})}runEditorCommand(e,t){Ha.get(t)?.selectEditor()}}lt(Ha.ID,Ha,1);It(K5e);It(Y5e);It(X5e);It(Z5e);It(Q5e);It(J5e);var Yoe=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},y1=function(s,e){return function(t,i){e(t,i,s)}};class e3e{constructor(e,t,i,n,o,r,a){this.range=e,this.insertText=t,this.filterText=i,this.additionalTextEdits=n,this.command=o,this.gutterMenuLinkAction=r,this.completion=a}}let E8=class extends gde{constructor(e,t,i,n,o,r){super(o.disposable),this.model=e,this.line=t,this.word=i,this.completionModel=n,this._suggestMemoryService=r}canBeReused(e,t,i){return this.model===e&&this.line===t&&this.word.word.length>0&&this.word.startColumn===i.startColumn&&this.word.endColumn<i.endColumn&&this.completionModel.getIncompleteProvider().size===0}get items(){const e=[],{items:t}=this.completionModel,i=this._suggestMemoryService.select(this.model,{lineNumber:this.line,column:this.word.endColumn+this.completionModel.lineContext.characterCountDelta},t),n=ot.slice(t,i),o=ot.slice(t,0,i);let r=5;for(const a of ot.concat(n,o)){if(a.score===Oa.Default)continue;const l=new k(a.editStart.lineNumber,a.editStart.column,a.editInsertEnd.lineNumber,a.editInsertEnd.column+this.completionModel.lineContext.characterCountDelta),c=a.completion.insertTextRules&&a.completion.insertTextRules&4?{snippet:a.completion.insertText}:a.completion.insertText;e.push(new e3e(l,c,a.filterTextLow??a.labelLow,a.completion.additionalTextEdits,a.completion.command,a.completion.action,a)),r-->=0&&a.resolve(Ke.None)}return e}};E8=Yoe([y1(5,YA)],E8);let I8=class extends F{constructor(e,t,i,n){super(),this._languageFeatureService=e,this._clipboardService=t,this._suggestMemoryService=i,this._editorService=n,this._store.add(e.inlineCompletionsProvider.register("*",this))}async provideInlineCompletions(e,t,i,n){if(i.selectedSuggestionInfo)return;let o;for(const g of this._editorService.listCodeEditors())if(g.getModel()===e){o=g;break}if(!o)return;const r=o.getOption(102);if(Ob.isAllOff(r))return;e.tokenization.tokenizeIfCheap(t.lineNumber);const a=e.tokenization.getLineTokens(t.lineNumber),l=a.getStandardTokenType(a.findTokenIndexAtOffset(Math.max(t.column-1-1,0)));if(Ob.valueFor(r,l)!=="inline")return;let c=e.getWordAtPosition(t),d;if(c?.word||(d=this._getTriggerCharacterInfo(e,t)),!c?.word&&!d||(c||(c=e.getWordUntilPosition(t)),c.endColumn!==t.column))return;let h;const u=e.getValueInRange(new k(t.lineNumber,1,t.lineNumber,t.column));if(!d&&this._lastResult?.canBeReused(e,t.lineNumber,c)){const g=new eK(u,t.column-this._lastResult.word.endColumn);this._lastResult.completionModel.lineContext=g,this._lastResult.acquire(),h=this._lastResult}else{const g=await UH(this._languageFeatureService.completionProvider,e,t,new n0(void 0,HD.createSuggestFilter(o).itemKind,d?.providers),d&&{triggerKind:1,triggerCharacter:d.ch},n);let f;g.needsClipboard&&(f=await this._clipboardService.readText());const p=new cu(g.items,t.column,new eK(u,0),WD.None,o.getOption(134),o.getOption(128),{boostFullMatch:!1,firstMatchCanBeWeak:!1},f);h=new E8(e,t.lineNumber,c,p,g,this._suggestMemoryService)}return this._lastResult=h,h}handleItemDidShow(e,t){t.completion.resolve(Ke.None)}disposeInlineCompletions(e){e.release()}_getTriggerCharacterInfo(e,t){const i=e.getValueInRange(k.fromPositions({lineNumber:t.lineNumber,column:t.column-1},t)),n=new Set;for(const o of this._languageFeatureService.completionProvider.all(e))o.triggerCharacters?.includes(i)&&n.add(o);if(n.size!==0)return{providers:n,ch:i}}};I8=Yoe([y1(0,ge),y1(1,Ko),y1(2,YA),y1(3,ht)],I8);uC(I8);class t3e extends ue{constructor(){super({id:"editor.action.forceRetokenize",label:V(1532,"Developer: Force Retokenize"),precondition:void 0})}run(e,t){if(!t.hasModel())return;const i=t.getModel();i.tokenization.resetTokenization();const n=new Cn;i.tokenization.forceTokenization(i.getLineCount()),n.stop(),console.log(`tokenization took ${n.elapsed()}`)}}ae(t3e);const JN=class JN extends Tn{constructor(){super({id:JN.ID,title:V(1530,"Toggle Tab Key Moves Focus"),precondition:void 0,keybinding:{primary:2091,mac:{primary:1323},weight:100},metadata:{description:V(1531,"Determines whether the tab key moves focus around the workbench or inserts the tab character in the current editor. This is also called tab trapping, tab navigation, or tab focus mode.")},f1:!0})}run(){const t=!av.getTabFocusMode();av.setTabFocusMode(t),Js(t?m(1528,"Pressing Tab will now move focus to the next focusable element"):m(1529,"Pressing Tab will now insert the tab character"))}};JN.ID="editor.action.toggleTabFocusMode";let D8=JN;It(D8);class lV{static computeUnicodeHighlights(e,t,i){const n=i?i.startLineNumber:1,o=i?i.endLineNumber:e.getLineCount(),r=new NK(t),a=r.getCandidateCodePoints();let l;a==="allNonBasicAscii"?l=new RegExp("[^\\t\\n\\r\\x20-\\x7E]","g"):l=new RegExp(`${i3e(Array.from(a))}`,"g");const c=new Db(null,l),d=[];let h=!1,u,g=0,f=0,p=0;e:for(let _=n,b=o;_<=b;_++){const w=e.getLineContent(_),C=w.length;c.reset(0);do if(u=c.next(w),u){let v=u.index,y=u.index+u[0].length;if(v>0){const I=w.charCodeAt(v-1);sn(I)&&v--}if(y+1<C){const I=w.charCodeAt(y-1);sn(I)&&y++}const x=w.substring(v,y);let L=iS(v+1,mW,w,0);L&&L.endColumn<=v+1&&(L=null);const E=r.shouldHighlightNonBasicASCII(x,L?L.word:null);if(E!==0){if(E===3?g++:E===2?f++:E===1?p++:W0(),d.length>=1e3){h=!0;break e}d.push(new k(_,v+1,_,y+1))}}while(u)}return{ranges:d,hasMore:h,ambiguousCharacterCount:g,invisibleCharacterCount:f,nonBasicAsciiCharacterCount:p}}static computeUnicodeHighlightReason(e,t){const i=new NK(t);switch(i.shouldHighlightNonBasicASCII(e,null)){case 0:return null;case 2:return{kind:1};case 3:{const o=e.codePointAt(0),r=i.ambiguousCharacters.getPrimaryConfusable(o),a=jy.getLocales().filter(l=>!jy.getInstance(new Set([...t.allowedLocales,l])).isAmbiguous(o));return{kind:0,confusableWith:String.fromCodePoint(r),notAmbiguousInLocales:a}}case 1:return{kind:2}}}}function i3e(s,e){return`[${Uo(s.map(i=>String.fromCodePoint(i)).join(""))}]`}class NK{constructor(e){this.options=e,this.allowedCodePoints=new Set(e.allowedCodePoints),this.ambiguousCharacters=jy.getInstance(new Set(e.allowedLocales))}getCandidateCodePoints(){if(this.options.nonBasicASCII)return"allNonBasicAscii";const e=new Set;if(this.options.invisibleCharacters)for(const t of np.codePoints)RK(String.fromCodePoint(t))||e.add(t);if(this.options.ambiguousCharacters)for(const t of this.ambiguousCharacters.getConfusableCodePoints())e.add(t);for(const t of this.allowedCodePoints)e.delete(t);return e}shouldHighlightNonBasicASCII(e,t){const i=e.codePointAt(0);if(this.allowedCodePoints.has(i))return 0;if(this.options.nonBasicASCII)return 1;let n=!1,o=!1;if(t)for(const r of t){const a=r.codePointAt(0),l=q0(r);n=n||l,!l&&!this.ambiguousCharacters.isAmbiguous(a)&&!np.isInvisibleCharacter(a)&&(o=!0)}return!n&&o?0:this.options.invisibleCharacters&&!RK(e)&&np.isInvisibleCharacter(i)?2:this.options.ambiguousCharacters&&this.ambiguousCharacters.isAmbiguous(i)?3:0}}function RK(s){return s===" "||s===`
`||s==="	"}var n3e=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},AK=function(s,e){return function(t,i){e(t,i,s)}};let T8=class extends F{get enabled(){return this._enabled}set enabled(e){e?(this.el.setAttribute("aria-disabled","false"),this.el.tabIndex=0,this.el.style.pointerEvents="auto",this.el.style.opacity="1",this.el.style.cursor="pointer",this._enabled=!1):(this.el.setAttribute("aria-disabled","true"),this.el.tabIndex=-1,this.el.style.pointerEvents="none",this.el.style.opacity="0.4",this.el.style.cursor="default",this._enabled=!0),this._enabled=e}constructor(e,t,i={},n,o){super(),this._link=t,this._hoverService=n,this._enabled=!0,this.el=X(e,ie("a.monaco-link",{tabIndex:t.tabIndex??0,href:t.href},t.label)),this.hoverDelegate=i.hoverDelegate??Uc("mouse"),this.setTooltip(t.title),this.el.setAttribute("role","button");const r=this._register(new Lt(this.el,"click")),a=this._register(new Lt(this.el,"keypress")),l=oe.chain(a.event,h=>h.map(u=>new Pt(u)).filter(u=>u.keyCode===3)),c=this._register(new Lt(this.el,qt.Tap)).event;this._register(_s.addTarget(this.el));const d=oe.any(r.event,l,c);this._register(d(h=>{this.enabled&&(gt.stop(h,!0),i?.opener?i.opener(this._link.href):o.open(this._link.href,{allowCommands:!0}))})),this.enabled=!0}setTooltip(e){!this.hover&&e?this.hover=this._register(this._hoverService.setupManagedHover(this.hoverDelegate,this.el,e)):this.hover&&this.hover.update(e)}};T8=n3e([AK(3,so),AK(4,Ah)],T8);var Zoe=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},N8=function(s,e){return function(t,i){e(t,i,s)}};const s3e=26;let R8=class extends F{constructor(e,t){super(),this._editor=e,this.instantiationService=t,this.banner=this._register(this.instantiationService.createInstance(A8))}hide(){this._editor.setBanner(null,0),this.banner.clear()}show(e){this.banner.show({...e,onClose:()=>{this.hide(),e.onClose?.()}}),this._editor.setBanner(this.banner.element,s3e)}};R8=Zoe([N8(1,_e)],R8);let A8=class extends F{constructor(e,t){super(),this.instantiationService=e,this.markdownRendererService=t,this.element=ie("div.editor-banner"),this.element.tabIndex=0}getAriaLabel(e){if(e.ariaLabel)return e.ariaLabel;if(typeof e.message=="string")return e.message}getBannerMessage(e){if(typeof e=="string"){const t=ie("span");return t.innerText=e,t}return this.markdownRendererService.render(e).element}clear(){Pn(this.element)}show(e){Pn(this.element);const t=this.getAriaLabel(e);t&&this.element.setAttribute("aria-label",t);const i=X(this.element,ie("div.icon-container"));i.setAttribute("aria-hidden","true"),e.icon&&i.appendChild(ie(`div${xe.asCSSSelector(e.icon)}`));const n=X(this.element,ie("div.message-container"));if(n.setAttribute("aria-hidden","true"),n.appendChild(this.getBannerMessage(e.message)),this.messageActionsContainer=X(this.element,ie("div.message-actions-container")),e.actions)for(const r of e.actions)this._register(this.instantiationService.createInstance(T8,this.messageActionsContainer,{...r,tabIndex:-1},{}));const o=X(this.element,ie("div.action-container"));this.actionBar=this._register(new bo(o)),this.actionBar.push(this._register(new Cr("banner.close",m(1533,"Close Banner"),xe.asClassName(pie),!0,()=>{typeof e.onClose=="function"&&e.onClose()})),{icon:!0,label:!1}),this.actionBar.setFocusable(!1)}};A8=Zoe([N8(0,_e),N8(1,Ua)],A8);const Xoe=qe("workspaceTrustManagementService");var cV=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},hy=function(s,e){return function(t,i){e(t,i,s)}};const o3e=Qt("extensions-warning-message",Z.warning,m(1534,"Icon shown with a warning message in the extensions editor."));var Xw;let m0=(Xw=class extends F{constructor(e,t,i,n){super(),this._editor=e,this._editorWorkerService=t,this._workspaceTrustService=i,this._highlighter=null,this._bannerClosed=!1,this._updateState=o=>{if(o&&o.hasMore){if(this._bannerClosed)return;const r=Math.max(o.ambiguousCharacterCount,o.nonBasicAsciiCharacterCount,o.invisibleCharacterCount);let a;if(o.nonBasicAsciiCharacterCount>=r)a={message:m(1535,"This document contains many non-basic ASCII unicode characters"),command:new b0};else if(o.ambiguousCharacterCount>=r)a={message:m(1536,"This document contains many ambiguous unicode characters"),command:new e_};else if(o.invisibleCharacterCount>=r)a={message:m(1537,"This document contains many invisible unicode characters"),command:new _0};else throw new Error("Unreachable");this._bannerController.show({id:"unicodeHighlightBanner",message:a.message,icon:o3e,actions:[{label:a.command.shortLabel,href:`command:${a.command.desc.id}`}],onClose:()=>{this._bannerClosed=!0}})}else this._bannerController.hide()},this._bannerController=this._register(n.createInstance(R8,e)),this._register(this._editor.onDidChangeModel(()=>{this._bannerClosed=!1,this._updateHighlighter()})),this._options=e.getOption(142),this._register(i.onDidChangeTrust(o=>{this._updateHighlighter()})),this._register(e.onDidChangeConfiguration(o=>{o.hasChanged(142)&&(this._options=e.getOption(142),this._updateHighlighter())})),this._updateHighlighter()}dispose(){this._highlighter&&(this._highlighter.dispose(),this._highlighter=null),super.dispose()}_updateHighlighter(){if(this._updateState(null),this._highlighter&&(this._highlighter.dispose(),this._highlighter=null),!this._editor.hasModel())return;const e=r3e(this._workspaceTrustService.isWorkspaceTrusted(),this._options);if([e.nonBasicASCII,e.ambiguousCharacters,e.invisibleCharacters].every(i=>i===!1))return;const t={nonBasicASCII:e.nonBasicASCII,ambiguousCharacters:e.ambiguousCharacters,invisibleCharacters:e.invisibleCharacters,includeComments:e.includeComments,includeStrings:e.includeStrings,allowedCodePoints:Object.keys(e.allowedCharacters).map(i=>i.codePointAt(0)),allowedLocales:Object.keys(e.allowedLocales).map(i=>i==="_os"?a_.NumberFormat().value.resolvedOptions().locale:i==="_vscode"?Zce:i)};this._editorWorkerService.canComputeUnicodeHighlights(this._editor.getModel().uri)?this._highlighter=new M8(this._editor,t,this._updateState,this._editorWorkerService):this._highlighter=new a3e(this._editor,t,this._updateState)}getDecorationInfo(e){return this._highlighter?this._highlighter.getDecorationInfo(e):null}},Xw.ID="editor.contrib.unicodeHighlighter",Xw);m0=cV([hy(1,ko),hy(2,Xoe),hy(3,_e)],m0);function r3e(s,e){return{nonBasicASCII:e.nonBasicASCII===lr?!s:e.nonBasicASCII,ambiguousCharacters:e.ambiguousCharacters,invisibleCharacters:e.invisibleCharacters,includeComments:e.includeComments===lr?!s:e.includeComments,includeStrings:e.includeStrings===lr?!s:e.includeStrings,allowedCharacters:e.allowedCharacters,allowedLocales:e.allowedLocales}}let M8=class extends F{constructor(e,t,i,n){super(),this._editor=e,this._options=t,this._updateState=i,this._editorWorkerService=n,this._model=this._editor.getModel(),this._decorations=this._editor.createDecorationsCollection(),this._updateSoon=this._register(new Rt(()=>this._update(),250)),this._register(this._editor.onDidChangeModelContent(()=>{this._updateSoon.schedule()})),this._updateSoon.schedule()}dispose(){this._decorations.clear(),super.dispose()}_update(){if(this._model.isDisposed())return;if(!this._model.mightContainNonBasicASCII()){this._decorations.clear();return}const e=this._model.getVersionId();this._editorWorkerService.computedUnicodeHighlights(this._model.uri,this._options).then(t=>{if(this._model.isDisposed()||this._model.getVersionId()!==e)return;this._updateState(t);const i=[];if(!t.hasMore)for(const n of t.ranges)i.push({range:n,options:tT.instance.getDecorationFromOptions(this._options)});this._decorations.set(i)})}getDecorationInfo(e){if(!this._decorations.has(e))return null;const t=this._editor.getModel();if(!KW(t,e))return null;const i=t.getValueInRange(e.range);return{reason:Joe(i,this._options),inComment:YW(t,e),inString:ZW(t,e)}}};M8=cV([hy(3,ko)],M8);class a3e extends F{constructor(e,t,i){super(),this._editor=e,this._options=t,this._updateState=i,this._model=this._editor.getModel(),this._decorations=this._editor.createDecorationsCollection(),this._updateSoon=this._register(new Rt(()=>this._update(),250)),this._register(this._editor.onDidLayoutChange(()=>{this._updateSoon.schedule()})),this._register(this._editor.onDidScrollChange(()=>{this._updateSoon.schedule()})),this._register(this._editor.onDidChangeHiddenAreas(()=>{this._updateSoon.schedule()})),this._register(this._editor.onDidChangeModelContent(()=>{this._updateSoon.schedule()})),this._updateSoon.schedule()}dispose(){this._decorations.clear(),super.dispose()}_update(){if(this._model.isDisposed())return;if(!this._model.mightContainNonBasicASCII()){this._decorations.clear();return}const e=this._editor.getVisibleRanges(),t=[],i={ranges:[],ambiguousCharacterCount:0,invisibleCharacterCount:0,nonBasicAsciiCharacterCount:0,hasMore:!1};for(const n of e){const o=lV.computeUnicodeHighlights(this._model,this._options,n);for(const r of o.ranges)i.ranges.push(r);i.ambiguousCharacterCount+=i.ambiguousCharacterCount,i.invisibleCharacterCount+=i.invisibleCharacterCount,i.nonBasicAsciiCharacterCount+=i.nonBasicAsciiCharacterCount,i.hasMore=i.hasMore||o.hasMore}if(!i.hasMore)for(const n of i.ranges)t.push({range:n,options:tT.instance.getDecorationFromOptions(this._options)});this._updateState(i),this._decorations.set(t)}getDecorationInfo(e){if(!this._decorations.has(e))return null;const t=this._editor.getModel(),i=t.getValueInRange(e.range);return KW(t,e)?{reason:Joe(i,this._options),inComment:YW(t,e),inString:ZW(t,e)}:null}}const Qoe=m(1538,"Configure Unicode Highlight Options");let O8=class{constructor(e,t){this._editor=e,this._markdownRendererService=t,this.hoverOrdinal=5}computeSync(e,t){if(!this._editor.hasModel()||e.type!==1)return[];const i=this._editor.getModel(),n=this._editor.getContribution(m0.ID);if(!n)return[];const o=[],r=new Set;let a=300;for(const l of t){const c=n.getDecorationInfo(l);if(!c)continue;const h=i.getValueInRange(l.range).codePointAt(0),u=DO(h);let g;switch(c.reason.kind){case 0:{q0(c.reason.confusableWith)?g=m(1539,"The character {0} could be confused with the ASCII character {1}, which is more common in source code.",u,DO(c.reason.confusableWith.codePointAt(0))):g=m(1540,"The character {0} could be confused with the character {1}, which is more common in source code.",u,DO(c.reason.confusableWith.codePointAt(0)));break}case 1:g=m(1541,"The character {0} is invisible.",u);break;case 2:g=m(1542,"The character {0} is not a basic ASCII character.",u);break}if(r.has(g))continue;r.add(g);const f={codePoint:h,reason:c.reason,inComment:c.inComment,inString:c.inString},p=m(1543,"Adjust settings"),_=Rte(iT.ID,f),b=new hs("",!0).appendMarkdown(g).appendText(" ").appendLink(_,p,Qoe);o.push(new Ta(this,l.range,[b],!1,a++))}return o}renderHoverParts(e,t){return PTe(e,t,this._editor,this._markdownRendererService)}getAccessibleContent(e){return e.contents.map(t=>t.value).join(`
`)}};O8=cV([hy(1,Ua)],O8);function P8(s){return`U+${s.toString(16).padStart(4,"0")}`}function DO(s){let e=`\`${P8(s)}\``;return np.isInvisibleCharacter(s)||(e+=` "${`${l3e(s)}`}"`),e}function l3e(s){return s===96?"`` ` ``":"`"+String.fromCodePoint(s)+"`"}function Joe(s,e){return lV.computeUnicodeHighlightReason(s,e)}const eR=class eR{constructor(){this.map=new Map}getDecorationFromOptions(e){return this.getDecoration(!e.includeComments,!e.includeStrings)}getDecoration(e,t){const i=`${e}${t}`;let n=this.map.get(i);return n||(n=Fe.createDynamic({description:"unicode-highlight",stickiness:1,className:"unicode-highlight",showIfCollapsed:!0,overviewRuler:null,minimap:null,hideInCommentTokens:e,hideInStringTokens:t}),this.map.set(i,n)),n}};eR.instance=new eR;let tT=eR;class c3e extends ue{constructor(){super({id:e_.ID,label:V(1552,"Disable highlighting of characters in comments"),precondition:void 0}),this.shortLabel=m(1544,"Disable Highlight In Comments")}async run(e,t,i){const n=e.get(He);n&&this.runAction(n)}async runAction(e){await e.updateValue(go.includeComments,!1,2)}}class d3e extends ue{constructor(){super({id:e_.ID,label:V(1553,"Disable highlighting of characters in strings"),precondition:void 0}),this.shortLabel=m(1545,"Disable Highlight In Strings")}async run(e,t,i){const n=e.get(He);n&&this.runAction(n)}async runAction(e){await e.updateValue(go.includeStrings,!1,2)}}const tR=class tR extends Tn{constructor(){super({id:tR.ID,title:V(1554,"Disable highlighting of ambiguous characters"),precondition:void 0,f1:!1}),this.shortLabel=m(1546,"Disable Ambiguous Highlight")}async run(e,t,i){const n=e.get(He);n&&this.runAction(n)}async runAction(e){await e.updateValue(go.ambiguousCharacters,!1,2)}};tR.ID="editor.action.unicodeHighlight.disableHighlightingOfAmbiguousCharacters";let e_=tR;const iR=class iR extends Tn{constructor(){super({id:iR.ID,title:V(1555,"Disable highlighting of invisible characters"),precondition:void 0,f1:!1}),this.shortLabel=m(1547,"Disable Invisible Highlight")}async run(e,t,i){const n=e.get(He);n&&this.runAction(n)}async runAction(e){await e.updateValue(go.invisibleCharacters,!1,2)}};iR.ID="editor.action.unicodeHighlight.disableHighlightingOfInvisibleCharacters";let _0=iR;const nR=class nR extends Tn{constructor(){super({id:nR.ID,title:V(1556,"Disable highlighting of non basic ASCII characters"),precondition:void 0,f1:!1}),this.shortLabel=m(1548,"Disable Non ASCII Highlight")}async run(e,t,i){const n=e.get(He);n&&this.runAction(n)}async runAction(e){await e.updateValue(go.nonBasicASCII,!1,2)}};nR.ID="editor.action.unicodeHighlight.disableHighlightingOfNonBasicAsciiCharacters";let b0=nR;const sR=class sR extends Tn{constructor(){super({id:sR.ID,title:V(1557,"Show Exclude Options"),precondition:void 0,f1:!1})}async run(e,t){const{codePoint:i,reason:n,inString:o,inComment:r}=t,a=String.fromCodePoint(i),l=e.get(bs),c=e.get(He);function d(f){return np.isInvisibleCharacter(f)?m(1549,"Exclude {0} (invisible character) from being highlighted",P8(f)):m(1550,"Exclude {0} from being highlighted",`${P8(f)} "${a}"`)}const h=[];if(n.kind===0)for(const f of n.notAmbiguousInLocales)h.push({label:m(1551,'Allow unicode characters that are more common in the language "{0}".',f),run:async()=>{u3e(c,[f])}});if(h.push({label:d(i),run:()=>h3e(c,[i])}),r){const f=new c3e;h.push({label:f.label,run:async()=>f.runAction(c)})}else if(o){const f=new d3e;h.push({label:f.label,run:async()=>f.runAction(c)})}function u(f){return typeof f.desc.title=="string"?f.desc.title:f.desc.title.value}if(n.kind===0){const f=new e_;h.push({label:u(f),run:async()=>f.runAction(c)})}else if(n.kind===1){const f=new _0;h.push({label:u(f),run:async()=>f.runAction(c)})}else if(n.kind===2){const f=new b0;h.push({label:u(f),run:async()=>f.runAction(c)})}else g3e(n);const g=await l.pick(h,{title:Qoe});g&&await g.run()}};sR.ID="editor.action.unicodeHighlight.showExcludeOptions";let iT=sR;async function h3e(s,e){const t=s.getValue(go.allowedCharacters);let i;typeof t=="object"&&t?i=t:i={};for(const n of e)i[String.fromCodePoint(n)]=!0;await s.updateValue(go.allowedCharacters,i,2)}async function u3e(s,e){const t=s.inspect(go.allowedLocales).user?.value;let i;typeof t=="object"&&t?i=Object.assign({},t):i={};for(const n of e)i[n]=!0;await s.updateValue(go.allowedLocales,i,2)}function g3e(s){throw new Error(`Unexpected value: ${s}`)}It(e_);It(_0);It(b0);It(iT);lt(m0.ID,m0,1);__.register(O8);const kx=qe("dialogService");var f3e=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},MK=function(s,e){return function(t,i){e(t,i,s)}};const ere="ignoreUnusualLineTerminators";function p3e(s,e,t){s.setModelProperty(e.uri,ere,t)}function m3e(s,e){return s.getModelProperty(e.uri,ere)}var Qw;let nT=(Qw=class extends F{constructor(e,t,i){super(),this._editor=e,this._dialogService=t,this._codeEditorService=i,this._isPresentingDialog=!1,this._config=this._editor.getOption(143),this._register(this._editor.onDidChangeConfiguration(n=>{n.hasChanged(143)&&(this._config=this._editor.getOption(143),this._checkForUnusualLineTerminators())})),this._register(this._editor.onDidChangeModel(()=>{this._checkForUnusualLineTerminators()})),this._register(this._editor.onDidChangeModelContent(n=>{n.isUndoing||this._checkForUnusualLineTerminators()})),this._checkForUnusualLineTerminators()}async _checkForUnusualLineTerminators(){if(this._config==="off"||!this._editor.hasModel())return;const e=this._editor.getModel();if(!e.mightContainUnusualLineTerminators()||m3e(this._codeEditorService,e)===!0||this._editor.getOption(104))return;if(this._config==="auto"){e.removeUnusualLineTerminators(this._editor.getSelections());return}if(this._isPresentingDialog)return;let i;try{this._isPresentingDialog=!0,i=await this._dialogService.confirm({title:m(1558,"Unusual Line Terminators"),message:m(1559,"Detected unusual line terminators"),detail:m(1560,"The file '{0}' contains one or more unusual line terminator characters, like Line Separator (LS) or Paragraph Separator (PS).\n\nIt is recommended to remove them from the file. This can be configured via `editor.unusualLineTerminators`.",ra(e.uri)),primaryButton:m(1561,"&&Remove Unusual Line Terminators"),cancelButton:m(1562,"Ignore")})}finally{this._isPresentingDialog=!1}if(!i.confirmed){p3e(this._codeEditorService,e,!0);return}e.removeUnusualLineTerminators(this._editor.getSelections())}},Qw.ID="editor.contrib.unusualLineTerminatorsDetector",Qw);nT=f3e([MK(1,kx),MK(2,ht)],nT);lt(nT.ID,nT,1);const tk="**",OK="/",eE="[/\\\\]",tE="[^/\\\\]",_3e=/\//g;function PK(s,e){switch(s){case 0:return"";case 1:return`${tE}*?`;default:return`(?:${eE}|${tE}+${eE}${e?`|${eE}${tE}+`:""})*?`}}function FK(s,e){if(!s)return[];const t=[];let i=!1,n=!1,o="";for(const r of s){switch(r){case e:if(!i&&!n){t.push(o),o="";continue}break;case"{":i=!0;break;case"}":i=!1;break;case"[":n=!0;break;case"]":n=!1;break}o+=r}return o&&t.push(o),t}function tre(s){if(!s)return"";let e="";const t=FK(s,OK);if(t.every(i=>i===tk))e=".*";else{let i=!1;t.forEach((n,o)=>{if(n===tk){if(i)return;e+=PK(2,o===t.length-1)}else{let r=!1,a="",l=!1,c="";for(const d of n){if(d!=="}"&&r){a+=d;continue}if(l&&(d!=="]"||!c)){let h;d==="-"?h=d:(d==="^"||d==="!")&&!c?h="^":d===OK?h="":h=Uo(d),c+=h;continue}switch(d){case"{":r=!0;continue;case"[":l=!0;continue;case"}":{const u=`(?:${FK(a,",").map(g=>tre(g)).join("|")})`;e+=u,r=!1,a="";break}case"]":{e+="["+c+"]",l=!1,c="";break}case"?":e+=tE;continue;case"*":e+=PK(1);continue;default:e+=Uo(d)}}o<t.length-1&&(t[o+1]!==tk||o+2<t.length)&&(e+=eE)}i=n===tk})}return e}const b3e=/^\*\*\/\*\.[\w\.-]+$/,w3e=/^\*\*\/([\w\.-]+)\/?$/,v3e=/^{\*\*\/\*?[\w\.-]+\/?(,\*\*\/\*?[\w\.-]+\/?)*}$/,C3e=/^{\*\*\/\*?[\w\.-]+(\/(\*\*)?)?(,\*\*\/\*?[\w\.-]+(\/(\*\*)?)?)*}$/,y3e=/^\*\*((\/[\w\.-]+)+)\/?$/,S3e=/^([\w\.-]+(\/[\w\.-]+)*)\/?$/,BK=new Va(1e4),WK=function(){return!1},Ic=function(){return null};function dV(s,e){if(!s)return Ic;let t;typeof s!="string"?t=s.pattern:t=s,t=t.trim();const i=e.ignoreCase??!1,n={...e,equals:i?Hf:(l,c)=>l===c,endsWith:i?Fde:(l,c)=>l.endsWith(c),isEqualOrParent:(l,c)=>NP(l,c,!vo||i)},o=`${i?t.toLowerCase():t}_${!!e.trimForExclusions}_${i}`;let r=BK.get(o);if(r)return HK(r,s,n);let a;return b3e.test(t)?r=x3e(t.substring(4),t,n):(a=w3e.exec(TO(t,n)))?r=L3e(a[1],t,n):(e.trimForExclusions?C3e:v3e).test(t)?r=k3e(t,n):(a=y3e.exec(TO(t,n)))?r=VK(a[1].substring(1),t,!0,n):(a=S3e.exec(TO(t,n)))?r=VK(a[1],t,!1,n):r=E3e(t,n),BK.set(o,r),HK(r,s,n)}function HK(s,e,t){if(typeof e=="string")return s;const i=function(n,o){return t.isEqualOrParent(n,e.base)?s(U0(n.substring(e.base.length),bl),o):null};return i.allBasenames=s.allBasenames,i.allPaths=s.allPaths,i.basenames=s.basenames,i.patterns=s.patterns,i}function TO(s,e){return e.trimForExclusions&&s.endsWith("/**")?s.substring(0,s.length-2):s}function x3e(s,e,t){return function(i,n){return typeof i=="string"&&t.endsWith(i,s)?e:null}}function L3e(s,e,t){const i=`/${s}`,n=`\\${s}`,o=function(a,l){return typeof a!="string"?null:l?t.equals(l,s)?e:null:t.equals(a,s)||t.endsWith(a,i)||t.endsWith(a,n)?e:null},r=[s];return o.basenames=r,o.patterns=[e],o.allBasenames=r,o}function k3e(s,e){const t=nre(s.slice(1,-1).split(",").map(a=>dV(a,e)).filter(a=>a!==Ic),s),i=t.length;if(!i)return Ic;if(i===1)return t[0];const n=function(a,l){for(let c=0,d=t.length;c<d;c++)if(t[c](a,l))return s;return null},o=t.find(a=>!!a.allBasenames);o&&(n.allBasenames=o.allBasenames);const r=t.reduce((a,l)=>l.allPaths?a.concat(l.allPaths):a,[]);return r.length&&(n.allPaths=r),n}function VK(s,e,t,i){const n=bl===Zi.sep,o=n?s:s.replace(_3e,bl),r=bl+o,a=Zi.sep+s;let l;return t?l=function(c,d){return typeof c=="string"&&(i.equals(c,o)||i.endsWith(c,r)||!n&&(i.equals(c,s)||i.endsWith(c,a)))?e:null}:l=function(c,d){return typeof c=="string"&&(i.equals(c,o)||!n&&i.equals(c,s))?e:null},l.allPaths=[(t?"*/":"./")+s],l}function E3e(s,e){try{const t=new RegExp(`^${tre(s)}$`,e.ignoreCase?"i":void 0);return function(i){return t.lastIndex=0,typeof i=="string"&&t.test(i)?s:null}}catch{return Ic}}function I3e(s,e,t){return!s||typeof e!="string"?!1:ire(s,t)(e)}function ire(s,e={}){if(!s)return WK;if(typeof s=="string"||D3e(s)){const t=dV(s,e);if(t===Ic)return WK;const i=function(n,o){return!!t(n,o)};return t.allBasenames&&(i.allBasenames=t.allBasenames),t.allPaths&&(i.allPaths=t.allPaths),i}return T3e(s,e)}function D3e(s){const e=s;return e?typeof e.base=="string"&&typeof e.pattern=="string":!1}function T3e(s,e){const t=nre(Object.getOwnPropertyNames(s).map(a=>N3e(a,s[a],e)).filter(a=>a!==Ic)),i=t.length;if(!i)return Ic;if(!t.some(a=>!!a.requiresSiblings)){if(i===1)return t[0];const a=function(d,h){let u;for(let g=0,f=t.length;g<f;g++){const p=t[g](d,h);if(typeof p=="string")return p;cP(p)&&(u||(u=[]),u.push(p))}return u?(async()=>{for(const g of u){const f=await g;if(typeof f=="string")return f}return null})():null},l=t.find(d=>!!d.allBasenames);l&&(a.allBasenames=l.allBasenames);const c=t.reduce((d,h)=>h.allPaths?d.concat(h.allPaths):d,[]);return c.length&&(a.allPaths=c),a}const n=function(a,l,c){let d,h;for(let u=0,g=t.length;u<g;u++){const f=t[u];f.requiresSiblings&&c&&(l||(l=th(a)),d||(d=l.substring(0,l.length-ohe(a).length)));const p=f(a,l,d,c);if(typeof p=="string")return p;cP(p)&&(h||(h=[]),h.push(p))}return h?(async()=>{for(const u of h){const g=await u;if(typeof g=="string")return g}return null})():null},o=t.find(a=>!!a.allBasenames);o&&(n.allBasenames=o.allBasenames);const r=t.reduce((a,l)=>l.allPaths?a.concat(l.allPaths):a,[]);return r.length&&(n.allPaths=r),n}function N3e(s,e,t){if(e===!1)return Ic;const i=dV(s,t);if(i===Ic)return Ic;if(typeof e=="boolean")return i;if(e){const n=e.when;if(typeof n=="string"){const o=(r,a,l,c)=>{if(!c||!i(r,a))return null;const d=n.replace("$(basename)",()=>l),h=c(d);return cP(h)?h.then(u=>u?s:null):h?s:null};return o.requiresSiblings=!0,o}}return i}function nre(s,e){const t=s.filter(a=>!!a.basenames);if(t.length<2)return s;const i=t.reduce((a,l)=>{const c=l.basenames;return c?a.concat(c):a},[]);let n;if(e){n=[];for(let a=0,l=i.length;a<l;a++)n.push(e)}else n=t.reduce((a,l)=>{const c=l.patterns;return c?a.concat(c):a},[]);const o=function(a,l){if(typeof a!="string")return null;if(!l){let d;for(d=a.length;d>0;d--){const h=a.charCodeAt(d-1);if(h===47||h===92)break}l=a.substring(d)}const c=i.indexOf(l);return c!==-1?n[c]:null};o.basenames=i,o.patterns=n,o.allBasenames=i;const r=s.filter(a=>!a.basenames);return r.push(o),r}function hV(s,e,t,i,n,o){if(Array.isArray(s)){let r=0;for(const a of s){const l=hV(a,e,t,i,n,o);if(l===10)return l;l>r&&(r=l)}return r}else{if(typeof s=="string")return i?s==="*"?5:s===t?10:0:0;if(s){const{language:r,pattern:a,scheme:l,hasAccessToAllModels:c,notebookType:d}=s;if(!i&&!c)return 0;d&&n&&(e=n);let h=0;if(l)if(l===e.scheme)h=10;else if(l==="*")h=5;else return 0;if(r)if(r===t)h=10;else if(r==="*")h=Math.max(h,5);else return 0;if(d)if(d===o)h=10;else if(d==="*"&&o!==void 0)h=Math.max(h,5);else return 0;if(a){let u;if(typeof a=="string"?u=a:u={...a,base:lQ(a.base)},u===e.fsPath||I3e(u,e.fsPath))h=10;else return 0}return h}else return 0}}var R3e=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},A3e=function(s,e){return function(t,i){e(t,i,s)}};class zK{constructor(){this.selector={language:"*"}}provideDocumentHighlights(e,t,i){const n=[],o=e.getWordAtPosition({lineNumber:t.lineNumber,column:t.column});return o?e.isDisposed()?void 0:e.findMatches(o.word,!0,!1,!0,RE,!1).map(a=>({range:a.range,kind:_v.Text})):Promise.resolve(n)}provideMultiDocumentHighlights(e,t,i,n){const o=new Wi,r=e.getWordAtPosition({lineNumber:t.lineNumber,column:t.column});if(!r)return Promise.resolve(o);for(const a of[e,...i]){if(a.isDisposed())continue;const c=a.findMatches(r.word,!0,!1,!0,RE,!1).map(d=>({range:d.range,kind:_v.Text}));c&&o.set(a.uri,c)}return o}}let F8=class extends F{constructor(e){super(),this._register(e.documentHighlightProvider.register("*",new zK)),this._register(e.multiDocumentHighlightProvider.register("*",new zK))}};F8=R3e([A3e(0,ge)],F8);var sre=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},dc=function(s,e){return function(t,i){e(t,i,s)}},vi,B8;const uV=new ce("hasWordHighlights",!1);function ore(s,e,t,i){const n=s.ordered(e);return z6(n.map(o=>()=>Promise.resolve(o.provideDocumentHighlights(e,t,i)).then(void 0,ji)),o=>o!=null).then(o=>{if(o){const r=new Wi;return r.set(e.uri,o),r}return new Wi})}function M3e(s,e,t,i,n){const o=s.ordered(e);return z6(o.map(r=>()=>{const a=n.filter(l=>nJ(l)).filter(l=>hV(r.selector,l.uri,l.getLanguageId(),!0,void 0,void 0)>0);return Promise.resolve(r.provideMultiDocumentHighlights(e,t,a,i)).then(void 0,ji)}),r=>r!=null)}class rre{constructor(e,t,i){this._model=e,this._selection=t,this._wordSeparators=i,this._wordRange=this._getCurrentWordRange(e,t),this._result=null}get result(){return this._result||(this._result=an(e=>this._compute(this._model,this._selection,this._wordSeparators,e))),this._result}_getCurrentWordRange(e,t){const i=e.getWordAtPosition(t.getPosition());return i?new k(t.startLineNumber,i.startColumn,t.startLineNumber,i.endColumn):null}cancel(){this.result.cancel()}}class O3e extends rre{constructor(e,t,i,n){super(e,t,i),this._providers=n}_compute(e,t,i,n){return ore(this._providers,e,t.getPosition(),n).then(o=>o||new Wi)}}class P3e extends rre{constructor(e,t,i,n,o){super(e,t,i),this._providers=n,this._otherModels=o}_compute(e,t,i,n){return M3e(this._providers,e,t.getPosition(),n,this._otherModels).then(o=>o||new Wi)}}function F3e(s,e,t,i){return new O3e(e,t,i,s)}function B3e(s,e,t,i,n){return new P3e(e,t,i,s,n)}Lo("_executeDocumentHighlights",async(s,e,t)=>{const i=s.get(ge);return(await ore(i.documentHighlightProvider,e,t,Ke.None))?.get(e.uri)});var og;let W8=(og=class{constructor(e,t,i,n,o,r,a,l){this.toUnhook=new U,this.workerRequestTokenId=0,this.workerRequestCompleted=!1,this.workerRequestValue=new Wi,this.lastCursorPositionChangeTime=0,this.renderDecorationsTimer=void 0,this.runDelayer=this.toUnhook.add(new sa(50)),this.editor=e,this.providers=t,this.multiDocumentProviders=i,this.codeEditorService=r,this.textModelService=o,this.configurationService=a,this.logService=l,this._hasWordHighlights=uV.bindTo(n),this._ignorePositionChangeEvent=!1,this.occurrencesHighlightEnablement=this.editor.getOption(90),this.occurrencesHighlightDelay=this.configurationService.getValue("editor.occurrencesHighlightDelay"),this.model=this.editor.getModel(),this.toUnhook.add(e.onDidChangeCursorPosition(c=>{this._ignorePositionChangeEvent||this.occurrencesHighlightEnablement!=="off"&&this.runDelayer.trigger(()=>{this._onPositionChanged(c)})})),this.toUnhook.add(e.onDidFocusEditorText(c=>{this.occurrencesHighlightEnablement!=="off"&&(this.workerRequest||this.runDelayer.trigger(()=>{this._run()}))})),this.toUnhook.add(e.onDidChangeModelContent(c=>{kR(this.model.uri,"output")||this._stopAll()})),this.toUnhook.add(e.onDidChangeModel(c=>{!c.newModelUrl&&c.oldModelUrl?this._stopSingular():vi.query&&this._run()})),this.toUnhook.add(e.onDidChangeConfiguration(c=>{const d=this.editor.getOption(90);if(this.occurrencesHighlightEnablement!==d)switch(this.occurrencesHighlightEnablement=d,d){case"off":this._stopAll();break;case"singleFile":this._stopAll(vi.query?.modelInfo?.modelURI);break;case"multiFile":vi.query&&this._run(!0);break;default:console.warn("Unknown occurrencesHighlight setting value:",d);break}})),this.toUnhook.add(this.configurationService.onDidChangeConfiguration(c=>{if(c.affectsConfiguration("editor.occurrencesHighlightDelay")){const d=a.getValue("editor.occurrencesHighlightDelay");this.occurrencesHighlightDelay!==d&&(this.occurrencesHighlightDelay=d)}})),this.toUnhook.add(e.onDidBlurEditorWidget(()=>{const c=this.codeEditorService.getFocusedCodeEditor();c?c.getModel()?.uri.scheme===Te.vscodeNotebookCell&&this.editor.getModel()?.uri.scheme!==Te.vscodeNotebookCell&&this._stopAll():this._stopAll()})),this.decorations=this.editor.createDecorationsCollection(),this.workerRequestTokenId=0,this.workerRequest=null,this.workerRequestCompleted=!1,this.lastCursorPositionChangeTime=0,this.renderDecorationsTimer=void 0,vi.query&&this._run()}hasDecorations(){return this.decorations.length>0}restore(e){this.occurrencesHighlightEnablement!=="off"&&(this.runDelayer.cancel(),this.runDelayer.trigger(()=>{this._run(!1,e)}))}stop(){this.occurrencesHighlightEnablement!=="off"&&this._stopAll()}_getSortedHighlights(){return this.decorations.getRanges().sort(k.compareRangesUsingStarts)}moveNext(){const e=this._getSortedHighlights(),i=(e.findIndex(o=>o.containsPosition(this.editor.getPosition()))+1)%e.length,n=e[i];try{this._ignorePositionChangeEvent=!0,this.editor.setPosition(n.getStartPosition()),this.editor.revealRangeInCenterIfOutsideViewport(n);const o=this._getWord();if(o){const r=this.editor.getModel().getLineContent(n.startLineNumber);Js(`${r}, ${i+1} of ${e.length} for '${o.word}'`)}}finally{this._ignorePositionChangeEvent=!1}}moveBack(){const e=this._getSortedHighlights(),i=(e.findIndex(o=>o.containsPosition(this.editor.getPosition()))-1+e.length)%e.length,n=e[i];try{this._ignorePositionChangeEvent=!0,this.editor.setPosition(n.getStartPosition()),this.editor.revealRangeInCenterIfOutsideViewport(n);const o=this._getWord();if(o){const r=this.editor.getModel().getLineContent(n.startLineNumber);Js(`${r}, ${i+1} of ${e.length} for '${o.word}'`)}}finally{this._ignorePositionChangeEvent=!1}}_removeSingleDecorations(){if(!this.editor.hasModel())return;const e=vi.storedDecorationIDs.get(this.editor.getModel().uri);e&&(this.editor.removeDecorations(e),vi.storedDecorationIDs.delete(this.editor.getModel().uri),this.decorations.length>0&&(this.decorations.clear(),this._hasWordHighlights.set(!1)))}_removeAllDecorations(e){const t=this.codeEditorService.listCodeEditors(),i=[];for(const n of t){if(!n.hasModel()||dg(n.getModel().uri,e))continue;const o=vi.storedDecorationIDs.get(n.getModel().uri);if(!o)continue;n.removeDecorations(o),i.push(n.getModel().uri);const r=yg.get(n);r?.wordHighlighter&&r.wordHighlighter.decorations.length>0&&(r.wordHighlighter.decorations.clear(),r.wordHighlighter.workerRequest=null,r.wordHighlighter._hasWordHighlights.set(!1))}for(const n of i)vi.storedDecorationIDs.delete(n)}_stopSingular(){this._removeSingleDecorations(),this.editor.hasTextFocus()&&(this.editor.getModel()?.uri.scheme!==Te.vscodeNotebookCell&&vi.query?.modelInfo?.modelURI.scheme!==Te.vscodeNotebookCell?(vi.query=null,this._run()):vi.query?.modelInfo&&(vi.query.modelInfo=null)),this.renderDecorationsTimer!==void 0&&(clearTimeout(this.renderDecorationsTimer),this.renderDecorationsTimer=void 0),this.workerRequest!==null&&(this.workerRequest.cancel(),this.workerRequest=null),this.workerRequestCompleted||(this.workerRequestTokenId++,this.workerRequestCompleted=!0)}_stopAll(e){this._removeAllDecorations(e),this.renderDecorationsTimer!==void 0&&(clearTimeout(this.renderDecorationsTimer),this.renderDecorationsTimer=void 0),this.workerRequest!==null&&(this.workerRequest.cancel(),this.workerRequest=null),this.workerRequestCompleted||(this.workerRequestTokenId++,this.workerRequestCompleted=!0)}_onPositionChanged(e){if(this.occurrencesHighlightEnablement==="off"){this._stopAll();return}if(e.source!=="api"&&e.reason!==3){this._stopAll();return}this._run()}_getWord(){const e=this.editor.getSelection(),t=e.startLineNumber,i=e.startColumn;return this.model.isDisposed()?null:this.model.getWordAtPosition({lineNumber:t,column:i})}getOtherModelsToHighlight(e){if(!e)return[];if(e.uri.scheme===Te.vscodeNotebookCell){const o=[],r=this.codeEditorService.listCodeEditors();for(const a of r){const l=a.getModel();l&&l!==e&&l.uri.scheme===Te.vscodeNotebookCell&&o.push(l)}return o}const i=[],n=this.codeEditorService.listCodeEditors();for(const o of n){if(!gH(o))continue;const r=o.getModel();r&&e===r.modified&&i.push(r.modified)}if(i.length)return i;if(this.occurrencesHighlightEnablement==="singleFile")return[];for(const o of n){const r=o.getModel();r&&r!==e&&i.push(r)}return i}async _run(e,t){if(this.editor.hasTextFocus()){const n=this.editor.getSelection();if(!n||n.startLineNumber!==n.endLineNumber){vi.query=null,this._stopAll();return}const o=n.startColumn,r=n.endColumn,a=this._getWord();if(!a||a.startColumn>o||a.endColumn<r){vi.query=null,this._stopAll();return}vi.query={modelInfo:{modelURI:this.model.uri,selection:n}}}else if(!vi.query){this._stopAll();return}if(this.lastCursorPositionChangeTime=new Date().getTime(),dg(this.editor.getModel().uri,vi.query.modelInfo?.modelURI)){if(!e){const a=this.decorations.getRanges();for(const l of a)if(l.containsPosition(this.editor.getPosition()))return}this._stopAll(e?this.model.uri:void 0);const n=++this.workerRequestTokenId;this.workerRequestCompleted=!1;const o=this.getOtherModelsToHighlight(this.editor.getModel());if(!vi.query||!vi.query.modelInfo)return;const r=await this.textModelService.createModelReference(vi.query.modelInfo.modelURI);try{this.workerRequest=this.computeWithModel(r.object.textEditorModel,vi.query.modelInfo.selection,o),this.workerRequest?.result.then(a=>{n===this.workerRequestTokenId&&(this.workerRequestCompleted=!0,this.workerRequestValue=a||[],this._beginRenderDecorations(t??this.occurrencesHighlightDelay))},Oe)}catch(a){this.logService.error("Unexpected error during occurrence request. Log: ",a)}finally{r.dispose()}}else if(this.model.uri.scheme===Te.vscodeNotebookCell){const n=++this.workerRequestTokenId;if(this.workerRequestCompleted=!1,!vi.query||!vi.query.modelInfo)return;const o=await this.textModelService.createModelReference(vi.query.modelInfo.modelURI);try{this.workerRequest=this.computeWithModel(o.object.textEditorModel,vi.query.modelInfo.selection,[this.model]),this.workerRequest?.result.then(r=>{n===this.workerRequestTokenId&&(this.workerRequestCompleted=!0,this.workerRequestValue=r||[],this._beginRenderDecorations(t??this.occurrencesHighlightDelay))},Oe)}catch(r){this.logService.error("Unexpected error during occurrence request. Log: ",r)}finally{o.dispose()}}}computeWithModel(e,t,i){return i.length?B3e(this.multiDocumentProviders,e,t,this.editor.getOption(148),i):F3e(this.providers,e,t,this.editor.getOption(148))}_beginRenderDecorations(e){const t=new Date().getTime(),i=this.lastCursorPositionChangeTime+e;t>=i?(this.renderDecorationsTimer=void 0,this.renderDecorations()):this.renderDecorationsTimer=setTimeout(()=>{this.renderDecorations()},i-t)}renderDecorations(){this.renderDecorationsTimer=void 0;const e=this.codeEditorService.listCodeEditors();for(const t of e){const i=yg.get(t);if(!i)continue;const n=[],o=t.getModel()?.uri;if(o&&this.workerRequestValue.has(o)){const r=vi.storedDecorationIDs.get(o),a=this.workerRequestValue.get(o);if(a)for(const c of a)c.range&&n.push({range:c.range,options:K4e(c.kind)});let l=[];t.changeDecorations(c=>{l=c.deltaDecorations(r??[],n)}),vi.storedDecorationIDs=vi.storedDecorationIDs.set(o,l),n.length>0&&(i.wordHighlighter?.decorations.set(n),i.wordHighlighter?._hasWordHighlights.set(!0))}}this.workerRequest=null}dispose(){this._stopSingular(),this.toUnhook.dispose()}},vi=og,og.storedDecorationIDs=new Wi,og.query=null,og);W8=vi=sre([dc(4,Fs),dc(5,ht),dc(6,He),dc(7,jt)],W8);var cm;let yg=(cm=class extends F{static get(e){return e.getContribution(B8.ID)}constructor(e,t,i,n,o,r,a){super(),this._wordHighlighter=null;const l=()=>{e.hasModel()&&!e.getModel().isTooLargeForTokenization()&&e.getModel().uri.scheme!==Te.accessibleView&&(this._wordHighlighter=new W8(e,i.documentHighlightProvider,i.multiDocumentHighlightProvider,t,o,n,r,a))};this._register(e.onDidChangeModel(c=>{this._wordHighlighter&&(!c.newModelUrl&&c.oldModelUrl?.scheme!==Te.vscodeNotebookCell&&this.wordHighlighter?.stop(),this._wordHighlighter.dispose(),this._wordHighlighter=null),l()})),l()}get wordHighlighter(){return this._wordHighlighter}saveViewState(){return!!(this._wordHighlighter&&this._wordHighlighter.hasDecorations())}moveNext(){this._wordHighlighter?.moveNext()}moveBack(){this._wordHighlighter?.moveBack()}restoreViewState(e){this._wordHighlighter&&e&&this._wordHighlighter.restore(250)}dispose(){this._wordHighlighter&&(this._wordHighlighter.dispose(),this._wordHighlighter=null),super.dispose()}},B8=cm,cm.ID="editor.contrib.wordHighlighter",cm);yg=B8=sre([dc(1,Ae),dc(2,ge),dc(3,ht),dc(4,Fs),dc(5,He),dc(6,jt)],yg);class are extends ue{constructor(e,t){super(t),this._isNext=e}run(e,t){const i=yg.get(t);i&&(this._isNext?i.moveNext():i.moveBack())}}class W3e extends are{constructor(){super(!0,{id:"editor.action.wordHighlight.next",label:V(1572,"Go to Next Symbol Highlight"),precondition:uV,kbOpts:{kbExpr:N.editorTextFocus,primary:65,weight:100}})}}class H3e extends are{constructor(){super(!1,{id:"editor.action.wordHighlight.prev",label:V(1573,"Go to Previous Symbol Highlight"),precondition:uV,kbOpts:{kbExpr:N.editorTextFocus,primary:1089,weight:100}})}}class V3e extends ue{constructor(){super({id:"editor.action.wordHighlight.trigger",label:V(1574,"Trigger Symbol Highlight"),precondition:void 0,kbOpts:{kbExpr:N.editorTextFocus,primary:0,weight:100}})}run(e,t,i){const n=yg.get(t);n&&n.restoreViewState(!0)}}lt(yg.ID,yg,0);ae(W3e);ae(H3e);ae(V3e);uC(F8);class i2 extends hn{constructor(e){super(e),this._inSelectionMode=e.inSelectionMode,this._wordNavigationType=e.wordNavigationType}runEditorCommand(e,t,i){if(!t.hasModel())return;const n=aa(t.getOption(148),t.getOption(147)),o=t.getModel(),r=t.getSelections(),a=r.length>1,l=r.map(c=>{const d=new A(c.positionLineNumber,c.positionColumn),h=this._move(n,o,d,this._wordNavigationType,a);return this._moveTo(c,h,this._inSelectionMode)});if(o.pushStackElement(),t._getViewModel().setCursorStates("moveWordCommand",3,l.map(c=>ut.fromModelSelection(c))),l.length===1){const c=new A(l[0].positionLineNumber,l[0].positionColumn);t.revealPosition(c,0)}}_moveTo(e,t,i){return i?new he(e.selectionStartLineNumber,e.selectionStartColumn,t.lineNumber,t.column):new he(t.lineNumber,t.column,t.lineNumber,t.column)}}class Mg extends i2{_move(e,t,i,n,o){return yt.moveWordLeft(e,t,i,n,o)}}class Og extends i2{_move(e,t,i,n,o){return yt.moveWordRight(e,t,i,n)}}class z3e extends Mg{constructor(){super({inSelectionMode:!1,wordNavigationType:0,id:"cursorWordStartLeft",precondition:void 0})}}class $3e extends Mg{constructor(){super({inSelectionMode:!1,wordNavigationType:2,id:"cursorWordEndLeft",precondition:void 0})}}class U3e extends Mg{constructor(){super({inSelectionMode:!1,wordNavigationType:1,id:"cursorWordLeft",precondition:void 0,kbOpts:{kbExpr:K.and(N.textInputFocus,K.and(Jv,MA)?.negate()),primary:2063,mac:{primary:527},weight:100}})}}class q3e extends Mg{constructor(){super({inSelectionMode:!0,wordNavigationType:0,id:"cursorWordStartLeftSelect",precondition:void 0})}}class j3e extends Mg{constructor(){super({inSelectionMode:!0,wordNavigationType:2,id:"cursorWordEndLeftSelect",precondition:void 0})}}class G3e extends Mg{constructor(){super({inSelectionMode:!0,wordNavigationType:1,id:"cursorWordLeftSelect",precondition:void 0,kbOpts:{kbExpr:K.and(N.textInputFocus,K.and(Jv,MA)?.negate()),primary:3087,mac:{primary:1551},weight:100}})}}class K3e extends Mg{constructor(){super({inSelectionMode:!1,wordNavigationType:3,id:"cursorWordAccessibilityLeft",precondition:void 0})}_move(e,t,i,n,o){return super._move(aa(Is.wordSeparators.defaultValue,e.intlSegmenterLocales),t,i,n,o)}}class Y3e extends Mg{constructor(){super({inSelectionMode:!0,wordNavigationType:3,id:"cursorWordAccessibilityLeftSelect",precondition:void 0})}_move(e,t,i,n,o){return super._move(aa(Is.wordSeparators.defaultValue,e.intlSegmenterLocales),t,i,n,o)}}class Z3e extends Og{constructor(){super({inSelectionMode:!1,wordNavigationType:0,id:"cursorWordStartRight",precondition:void 0})}}class X3e extends Og{constructor(){super({inSelectionMode:!1,wordNavigationType:2,id:"cursorWordEndRight",precondition:void 0,kbOpts:{kbExpr:K.and(N.textInputFocus,K.and(Jv,MA)?.negate()),primary:2065,mac:{primary:529},weight:100}})}}class Q3e extends Og{constructor(){super({inSelectionMode:!1,wordNavigationType:2,id:"cursorWordRight",precondition:void 0})}}class J3e extends Og{constructor(){super({inSelectionMode:!0,wordNavigationType:0,id:"cursorWordStartRightSelect",precondition:void 0})}}class e8e extends Og{constructor(){super({inSelectionMode:!0,wordNavigationType:2,id:"cursorWordEndRightSelect",precondition:void 0,kbOpts:{kbExpr:K.and(N.textInputFocus,K.and(Jv,MA)?.negate()),primary:3089,mac:{primary:1553},weight:100}})}}class t8e extends Og{constructor(){super({inSelectionMode:!0,wordNavigationType:2,id:"cursorWordRightSelect",precondition:void 0})}}class i8e extends Og{constructor(){super({inSelectionMode:!1,wordNavigationType:3,id:"cursorWordAccessibilityRight",precondition:void 0})}_move(e,t,i,n,o){return super._move(aa(Is.wordSeparators.defaultValue,e.intlSegmenterLocales),t,i,n,o)}}class n8e extends Og{constructor(){super({inSelectionMode:!0,wordNavigationType:3,id:"cursorWordAccessibilityRightSelect",precondition:void 0})}_move(e,t,i,n,o){return super._move(aa(Is.wordSeparators.defaultValue,e.intlSegmenterLocales),t,i,n,o)}}class n2 extends hn{constructor(e){super({canTriggerInlineEdits:!0,...e}),this._whitespaceHeuristics=e.whitespaceHeuristics,this._wordNavigationType=e.wordNavigationType}runEditorCommand(e,t,i){const n=e?.get(ui);if(!t.hasModel()||!n)return;const o=aa(t.getOption(148),t.getOption(147)),r=t.getModel(),a=t.getSelections(),l=t.getOption(10),c=t.getOption(15),d=n.getLanguageConfiguration(r.getLanguageId()).getAutoClosingPairs(),h=t._getViewModel(),u=a.map(g=>{const f=this._delete({wordSeparators:o,model:r,selection:g,whitespaceHeuristics:this._whitespaceHeuristics,autoClosingDelete:t.getOption(13),autoClosingBrackets:l,autoClosingQuotes:c,autoClosingPairs:d,autoClosedCharacters:h.getCursorAutoClosedCharacters()},this._wordNavigationType);return new ts(f,"")});t.pushUndoStop(),t.executeCommands(this.id,u),t.pushUndoStop()}}class gV extends n2{_delete(e,t){const i=yt.deleteWordLeft(e,t);return i||new k(1,1,1,1)}}class fV extends n2{_delete(e,t){const i=yt.deleteWordRight(e,t);if(i)return i;const n=e.model.getLineCount(),o=e.model.getLineMaxColumn(n);return new k(n,o,n,o)}}class s8e extends gV{constructor(){super({whitespaceHeuristics:!1,wordNavigationType:0,id:"deleteWordStartLeft",precondition:N.writable})}}class o8e extends gV{constructor(){super({whitespaceHeuristics:!1,wordNavigationType:2,id:"deleteWordEndLeft",precondition:N.writable})}}class r8e extends gV{constructor(){super({whitespaceHeuristics:!0,wordNavigationType:0,id:"deleteWordLeft",precondition:N.writable,kbOpts:{kbExpr:N.textInputFocus,primary:2049,mac:{primary:513},weight:100}})}}class a8e extends fV{constructor(){super({whitespaceHeuristics:!1,wordNavigationType:0,id:"deleteWordStartRight",precondition:N.writable})}}class l8e extends fV{constructor(){super({whitespaceHeuristics:!1,wordNavigationType:2,id:"deleteWordEndRight",precondition:N.writable})}}class c8e extends fV{constructor(){super({whitespaceHeuristics:!0,wordNavigationType:2,id:"deleteWordRight",precondition:N.writable,kbOpts:{kbExpr:N.textInputFocus,primary:2068,mac:{primary:532},weight:100}})}}class d8e extends ue{constructor(){super({id:"deleteInsideWord",precondition:N.writable,label:V(1575,"Delete Word")})}run(e,t,i){if(!t.hasModel())return;const n=aa(t.getOption(148),t.getOption(147)),o=t.getModel(),a=t.getSelections().map(l=>{const c=yt.deleteInsideWord(n,o,l);return new ts(c,"")});t.pushUndoStop(),t.executeCommands(this.id,a),t.pushUndoStop()}}le(new z3e);le(new $3e);le(new U3e);le(new q3e);le(new j3e);le(new G3e);le(new Z3e);le(new X3e);le(new Q3e);le(new J3e);le(new e8e);le(new t8e);le(new K3e);le(new Y3e);le(new i8e);le(new n8e);le(new s8e);le(new o8e);le(new r8e);le(new a8e);le(new l8e);le(new c8e);ae(d8e);class h8e extends n2{constructor(){super({whitespaceHeuristics:!0,wordNavigationType:0,id:"deleteWordPartLeft",precondition:N.writable,kbOpts:{kbExpr:N.textInputFocus,primary:0,mac:{primary:769},weight:100}})}_delete(e,t){const i=jR.deleteWordPartLeft(e);return i||new k(1,1,1,1)}}class u8e extends n2{constructor(){super({whitespaceHeuristics:!0,wordNavigationType:2,id:"deleteWordPartRight",precondition:N.writable,kbOpts:{kbExpr:N.textInputFocus,primary:0,mac:{primary:788},weight:100}})}_delete(e,t){const i=jR.deleteWordPartRight(e);if(i)return i;const n=e.model.getLineCount(),o=e.model.getLineMaxColumn(n);return new k(n,o,n,o)}}class lre extends i2{_move(e,t,i,n,o){return jR.moveWordPartLeft(e,t,i,o)}}class g8e extends lre{constructor(){super({inSelectionMode:!1,wordNavigationType:0,id:"cursorWordPartLeft",precondition:void 0,kbOpts:{kbExpr:N.textInputFocus,primary:0,mac:{primary:783},weight:100}})}}at.registerCommandAlias("cursorWordPartStartLeft","cursorWordPartLeft");class f8e extends lre{constructor(){super({inSelectionMode:!0,wordNavigationType:0,id:"cursorWordPartLeftSelect",precondition:void 0,kbOpts:{kbExpr:N.textInputFocus,primary:0,mac:{primary:1807},weight:100}})}}at.registerCommandAlias("cursorWordPartStartLeftSelect","cursorWordPartLeftSelect");class cre extends i2{_move(e,t,i,n,o){return jR.moveWordPartRight(e,t,i)}}class p8e extends cre{constructor(){super({inSelectionMode:!1,wordNavigationType:2,id:"cursorWordPartRight",precondition:void 0,kbOpts:{kbExpr:N.textInputFocus,primary:0,mac:{primary:785},weight:100}})}}class m8e extends cre{constructor(){super({inSelectionMode:!0,wordNavigationType:2,id:"cursorWordPartRightSelect",precondition:void 0,kbOpts:{kbExpr:N.textInputFocus,primary:0,mac:{primary:1809},weight:100}})}}le(new h8e);le(new u8e);le(new g8e);le(new f8e);le(new p8e);le(new m8e);const GV=class GV extends F{constructor(e){super(),this.editor=e,this._register(this.editor.onDidAttemptReadOnlyEdit(()=>this._onDidAttemptReadOnlyEdit()))}_onDidAttemptReadOnlyEdit(){const e=Vo.get(this.editor);if(e&&this.editor.hasModel()){let t=this.editor.getOptions().get(105);t||(this.editor.isSimpleWidget?t=new hs(m(1378,"Cannot edit in read-only input")):t=new hs(m(1379,"Cannot edit in read-only editor"))),e.showMessage(t,this.editor.getPosition())}}};GV.ID="editor.contrib.readOnlyMessageController";let sT=GV;lt(sT.ID,sT,2);var _8e=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},$K=function(s,e){return function(t,i){e(t,i,s)}};let H8=class extends F{constructor(e,t,i){super(),this._textModel=e,this._languageFeaturesService=t,this._outlineModelService=i,this._currentModel=Re(this,void 0);const n=Po("documentSymbolProvider.onDidChange",this._languageFeaturesService.documentSymbolProvider.onDidChange),o=Po("_textModel.onDidChangeContent",oe.debounce(r=>this._textModel.onDidChangeContent(r),()=>{},100));this._register(ps(async(r,a)=>{n.read(r),o.read(r);const l=a.add(new WLe),c=await this._outlineModelService.getOrCreate(this._textModel,l.token);a.isDisposed||this._currentModel.set(c,void 0)}))}getBreadcrumbItems(e,t){const i=this._currentModel.read(t);if(!i)return[];const n=i.asListOfDocumentSymbols().filter(o=>e.contains(o.range.startLineNumber)&&!e.contains(o.range.endLineNumber));return n.sort(bQ(ns(o=>o.range.endLineNumber-o.range.startLineNumber,Wo))),n.map(o=>({name:o.name,kind:o.kind,startLineNumber:o.range.startLineNumber}))}};H8=_8e([$K(1,ge),$K(2,xx)],H8);MI.setBreadcrumbsSourceFactory((s,e)=>e.createInstance(H8,s));var b8e=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},NO=function(s,e){return function(t,i){e(t,i,s)}},Jw;let oT=(Jw=class extends F{constructor(e,t,i,n){super();const o=this._register(di(e)),r=this._register(n.createMenu(fe.EditorContent,e.contextKeyService)),a=bt(this,r.onDidChange,()=>r.getActions().length===0);this._register(De(l=>{if(a.read(l))return;const d=ct("div.floating-menu-overlay-widget");d.root.style.height="28px";const h=t.createInstance(BS,d.root,fe.EditorContent,{actionViewItemProvider:(u,g)=>{if(!(u instanceof yr))return;const f=i.lookupKeybinding(u.id);if(f)return t.createInstance(class extends bg{updateLabel(){this.options.label&&this.label&&(this.label.textContent=`${this._commandAction.label} (${f.getLabel()})`)}},u,{...g,keybindingNotRenderedWithLabel:!0})},hiddenItemStrategy:0,menuOptions:{shouldForwardArgs:!0},telemetrySource:"editor.overlayToolbar",toolbarOptions:{primaryGroup:()=>!0,useSeparatorsInPrimaryActions:!0}});l.store.add(h),l.store.add(De(u=>{const g=o.model.read(u);h.context=g?.uri})),l.store.add(o.createOverlayWidget({allowEditorOverflow:!1,domNode:d.root,minContentWidthInPx:zt(0),position:zt({preference:1})}))}))}},Jw.ID="editor.contrib.floatingToolbar",Jw);oT=b8e([NO(1,_e),NO(2,ft),NO(3,da)],oT);lt(oT.ID,oT,1);var V8;(function(s){s.inspectTokensAction=m(786,"Developer: Inspect Tokens")})(V8||(V8={}));var rT;(function(s){s.gotoLineActionLabel=m(787,"Go to Line/Column...")})(rT||(rT={}));var z8;(function(s){s.helpQuickAccessActionLabel=m(788,"Show all Quick Access Providers")})(z8||(z8={}));var aT;(function(s){s.quickCommandActionLabel=m(789,"Command Palette"),s.quickCommandHelp=m(790,"Show And Run Commands")})(aT||(aT={}));var w0;(function(s){s.quickOutlineActionLabel=m(791,"Go to Symbol..."),s.quickOutlineByCategoryActionLabel=m(792,"Go to Symbol by Category...")})(w0||(w0={}));var $8;(function(s){s.editorViewAccessibleLabel=m(793,"Editor content")})($8||($8={}));var U8;(function(s){s.toggleHighContrast=m(794,"Toggle High Contrast Theme")})(U8||(U8={}));var q8;(function(s){s.bulkEditServiceSummary=m(795,"Made {0} edits in {1} files")})(q8||(q8={}));const KV=class KV extends F{constructor(e){super(),this.editor=e,this.widget=null,na&&(this._register(e.onDidChangeConfiguration(()=>this.update())),this.update())}update(){const e=!this.editor.getOption(104);!this.widget&&e?this.widget=new j8(this.editor):this.widget&&!e&&(this.widget.dispose(),this.widget=null)}dispose(){super.dispose(),this.widget&&(this.widget.dispose(),this.widget=null)}};KV.ID="editor.contrib.iPadShowKeyboard";let lT=KV;const oR=class oR extends F{constructor(e){super(),this.editor=e,this._domNode=document.createElement("textarea"),this._domNode.className="iPadShowKeyboard",this._register(W(this._domNode,"touchstart",t=>{this.editor.focus()})),this._register(W(this._domNode,"focus",t=>{this.editor.focus()})),this.editor.addOverlayWidget(this)}dispose(){this.editor.removeOverlayWidget(this),super.dispose()}getId(){return oR.ID}getDomNode(){return this._domNode}getPosition(){return{preference:1}}};oR.ID="editor.contrib.ShowKeyboardWidget";let j8=oR;lt(lT.ID,lT,3);const Rr=qe("themeService");var w8e=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},UK=function(s,e){return function(t,i){e(t,i,s)}},G8,dm;let v0=(dm=class extends F{static get(e){return e.getContribution(G8.ID)}constructor(e,t,i){super(),this._editor=e,this._languageService=i,this._widget=null,this._register(this._editor.onDidChangeModel(n=>this.stop())),this._register(this._editor.onDidChangeModelLanguage(n=>this.stop())),this._register(yi.onDidChange(n=>this.stop())),this._register(this._editor.onKeyUp(n=>n.keyCode===9&&this.stop()))}dispose(){this.stop(),super.dispose()}launch(){this._widget||this._editor.hasModel()&&(this._widget=new K8(this._editor,this._languageService))}stop(){this._widget&&(this._widget.dispose(),this._widget=null)}},G8=dm,dm.ID="editor.contrib.inspectTokens",dm);v0=G8=w8e([UK(1,Rr),UK(2,Ei)],v0);class v8e extends ue{constructor(){super({id:"editor.action.inspectTokens",label:V8.inspectTokensAction,alias:"Developer: Inspect Tokens",precondition:void 0})}run(e,t){v0.get(t)?.launch()}}function C8e(s){let e="";for(let t=0,i=s.length;t<i;t++){const n=s.charCodeAt(t);switch(n){case 9:e+="→";break;case 32:e+="·";break;default:e+=String.fromCharCode(n)}}return e}function y8e(s,e){const t=yi.get(e);if(t)return t;const i=s.encodeLanguageId(e);return{getInitialState:()=>bv,tokenize:(n,o,r)=>h7(e,r),tokenizeEncoded:(n,o,r)=>mA(i,r)}}const rR=class rR extends F{constructor(e,t){super(),this.allowEditorOverflow=!0,this._editor=e,this._languageService=t,this._model=this._editor.getModel(),this._domNode=document.createElement("div"),this._domNode.className="tokens-inspect-widget",this._tokenizationSupport=y8e(this._languageService.languageIdCodec,this._model.getLanguageId()),this._compute(this._editor.getPosition()),this._register(this._editor.onDidChangeCursorPosition(i=>this._compute(this._editor.getPosition()))),this._editor.addContentWidget(this)}dispose(){this._editor.removeContentWidget(this),super.dispose()}getId(){return rR._ID}_compute(e){const t=this._getTokensAtLine(e.lineNumber);let i=0;for(let l=t.tokens1.length-1;l>=0;l--){const c=t.tokens1[l];if(e.column-1>=c.offset){i=l;break}}let n=0;for(let l=t.tokens2.length>>>1;l>=0;l--)if(e.column-1>=t.tokens2[l<<1]){n=l;break}const o=this._model.getLineContent(e.lineNumber);let r="";if(i<t.tokens1.length){const l=t.tokens1[i].offset,c=i+1<t.tokens1.length?t.tokens1[i+1].offset:o.length;r=o.substring(l,c)}vn(this._domNode,ie("h2.tm-token",void 0,C8e(r),ie("span.tm-token-length",void 0,`${r.length} ${r.length===1?"char":"chars"}`))),X(this._domNode,ie("hr.tokens-inspect-separator",{style:"clear:both"}));const a=(n<<1)+1<t.tokens2.length?this._decodeMetadata(t.tokens2[(n<<1)+1]):null;X(this._domNode,ie("table.tm-metadata-table",void 0,ie("tbody",void 0,ie("tr",void 0,ie("td.tm-metadata-key",void 0,"language"),ie("td.tm-metadata-value",void 0,`${a?a.languageId:"-?-"}`)),ie("tr",void 0,ie("td.tm-metadata-key",void 0,"token type"),ie("td.tm-metadata-value",void 0,`${a?this._tokenTypeToString(a.tokenType):"-?-"}`)),ie("tr",void 0,ie("td.tm-metadata-key",void 0,"font style"),ie("td.tm-metadata-value",void 0,`${a?this._fontStyleToString(a.fontStyle):"-?-"}`)),ie("tr",void 0,ie("td.tm-metadata-key",void 0,"foreground"),ie("td.tm-metadata-value",void 0,`${a?q.Format.CSS.formatHex(a.foreground):"-?-"}`)),ie("tr",void 0,ie("td.tm-metadata-key",void 0,"background"),ie("td.tm-metadata-value",void 0,`${a?q.Format.CSS.formatHex(a.background):"-?-"}`))))),X(this._domNode,ie("hr.tokens-inspect-separator")),i<t.tokens1.length&&X(this._domNode,ie("span.tm-token-type",void 0,t.tokens1[i].type)),this._editor.layoutContentWidget(this)}_decodeMetadata(e){const t=yi.getColorMap(),i=ds.getLanguageId(e),n=ds.getTokenType(e),o=ds.getFontStyle(e),r=ds.getForeground(e),a=ds.getBackground(e);return{languageId:this._languageService.languageIdCodec.decodeLanguageId(i),tokenType:n,fontStyle:o,foreground:t[r],background:t[a]}}_tokenTypeToString(e){switch(e){case 0:return"Other";case 1:return"Comment";case 2:return"String";case 3:return"RegEx";default:return"??"}}_fontStyleToString(e){let t="";return e&1&&(t+="italic "),e&2&&(t+="bold "),e&4&&(t+="underline "),e&8&&(t+="strikethrough "),t.length===0&&(t="---"),t}_getTokensAtLine(e){const t=this._getStateBeforeLine(e),i=this._tokenizationSupport.tokenize(this._model.getLineContent(e),!0,t),n=this._tokenizationSupport.tokenizeEncoded(this._model.getLineContent(e),!0,t);return{startState:t,tokens1:i.tokens,tokens2:n.tokens,endState:i.endState}}_getStateBeforeLine(e){let t=this._tokenizationSupport.getInitialState();for(let i=1;i<e;i++)t=this._tokenizationSupport.tokenize(this._model.getLineContent(i),!0,t).endState;return t}getDomNode(){return this._domNode}getPosition(){return{position:this._editor.getPosition(),preference:[2,1]}}};rR._ID="editor.contrib.inspectTokensWidget";let K8=rR;lt(v0.ID,v0,4);ae(v8e);var Y8;(function(s){s[s.PRESERVE=0]="PRESERVE",s[s.LAST=1]="LAST"})(Y8||(Y8={}));const y_={Quickaccess:"workbench.contributions.quickaccess"};class S8e{constructor(){this.providers=[],this.defaultProvider=void 0}registerQuickAccessProvider(e){return e.prefix.length===0?this.defaultProvider=e:this.providers.push(e),this.providers.sort((t,i)=>i.prefix.length-t.prefix.length),pe(()=>{this.providers.splice(this.providers.indexOf(e),1),this.defaultProvider===e&&(this.defaultProvider=void 0)})}getQuickAccessProviders(){return Al([this.defaultProvider,...this.providers])}getQuickAccessProvider(e){return e&&this.providers.find(i=>e.startsWith(i.prefix))||void 0||this.defaultProvider}}bi.add(y_.Quickaccess,new S8e);var x8e=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},qK=function(s,e){return function(t,i){e(t,i,s)}},S1,hm;let Z8=(hm=class{constructor(e,t){this.quickInputService=e,this.keybindingService=t,this.registry=bi.as(y_.Quickaccess)}provide(e){const t=new U;return t.add(e.onDidAccept(()=>{const[i]=e.selectedItems;i&&this.quickInputService.quickAccess.show(i.prefix,{preserveValue:!0})})),t.add(e.onDidChangeValue(i=>{const n=this.registry.getQuickAccessProvider(i.substr(S1.PREFIX.length));n&&n.prefix&&n.prefix!==S1.PREFIX&&this.quickInputService.quickAccess.show(n.prefix,{preserveValue:!0})})),e.items=this.getQuickAccessProviders().filter(i=>i.prefix!==S1.PREFIX),t}getQuickAccessProviders(){return this.registry.getQuickAccessProviders().sort((t,i)=>t.prefix.localeCompare(i.prefix)).flatMap(t=>this.createPicks(t))}createPicks(e){return e.helpEntries.map(t=>{const i=t.prefix||e.prefix,n=i||"…";return{prefix:i,label:n,keybinding:t.commandId?this.keybindingService.lookupKeybinding(t.commandId):void 0,ariaLabel:m(1747,"{0}, {1}",n,t.description),description:t.description}})}},S1=hm,hm.PREFIX="?",hm);Z8=S1=x8e([qK(0,bs),qK(1,ft)],Z8);bi.as(y_.Quickaccess).registerQuickAccessProvider({ctor:Z8,prefix:"",helpEntries:[{description:z8.helpQuickAccessActionLabel}]});class dre{constructor(e){this.options=e,this.rangeHighlightDecorationId=void 0}provide(e,t,i){const n=new U;e.canAcceptInBackground=!!this.options?.canAcceptInBackground,e.matchOnLabel=e.matchOnDescription=e.matchOnDetail=e.sortByLabel=!1;const o=n.add(new vt);return o.value=this.doProvide(e,t,i),n.add(this.onDidActiveTextEditorControlChange(()=>{o.value=void 0,o.value=this.doProvide(e,t)})),n}doProvide(e,t,i){const n=new U,o=this.activeTextEditorControl;if(o&&this.canProvideWithTextEditor(o)){const r={editor:o},a=Qne(o);if(a){let l=o.saveViewState()??void 0;n.add(a.onDidChangeCursorPosition(()=>{l=o.saveViewState()??void 0})),r.restoreViewState=()=>{l&&o===this.activeTextEditorControl&&o.restoreViewState(l)},n.add(ym(t.onCancellationRequested)(()=>r.restoreViewState?.()))}n.add(pe(()=>this.clearDecorations(o))),n.add(this.provideWithTextEditor(r,e,t,i))}else n.add(this.provideWithoutTextEditor(e,t));return n}canProvideWithTextEditor(e){return!0}gotoLocation({editor:e},t){e.setSelection(t.range,"code.jump"),e.revealRangeInCenter(t.range,0),t.preserveFocus||e.focus();const i=e.getModel();i&&"getLineContent"in i&&xl(`${i.getLineContent(t.range.startLineNumber)}`)}getModel(e){return gH(e)?e.getModel()?.modified:e.getModel()}addDecorations(e,t){e.changeDecorations(i=>{const n=[];this.rangeHighlightDecorationId&&(n.push(this.rangeHighlightDecorationId.overviewRulerDecorationId),n.push(this.rangeHighlightDecorationId.rangeHighlightId),this.rangeHighlightDecorationId=void 0);const o=[{range:t,options:{description:"quick-access-range-highlight",className:"rangeHighlight",isWholeLine:!0}},{range:t,options:{description:"quick-access-range-highlight-overview",overviewRuler:{color:Ci(fee),position:Sr.Full}}}],[r,a]=i.deltaDecorations(n,o);this.rangeHighlightDecorationId={rangeHighlightId:r,overviewRulerDecorationId:a}})}clearDecorations(e){const t=this.rangeHighlightDecorationId;t&&(e.changeDecorations(i=>{i.deltaDecorations([t.overviewRulerDecorationId,t.rangeHighlightId],[])}),this.rangeHighlightDecorationId=void 0)}}const Ff=class Ff extends dre{constructor(){super({canAcceptInBackground:!0})}get useZeroBasedOffset(){return this.storageService.getBoolean(Ff.ZERO_BASED_OFFSET_STORAGE_KEY,-1,!1)}set useZeroBasedOffset(e){this.storageService.store(Ff.ZERO_BASED_OFFSET_STORAGE_KEY,e,-1,0)}provideWithoutTextEditor(e){const t=m(1335,"Open a text editor first to go to a line.");return e.items=[{label:t}],e.ariaLabel=t,F.None}provideWithTextEditor(e,t,i){const n=e.editor,o=new U;o.add(t.onDidAccept(c=>{const[d]=t.selectedItems;if(d){if(!d.lineNumber)return;this.gotoLocation(e,{range:this.toRange(d.lineNumber,d.column),keyMods:t.keyMods,preserveFocus:c.inBackground}),c.inBackground||t.hide()}}));const r=()=>{const c=t.value.trim().substring(Ff.PREFIX.length),{inOffsetMode:d,lineNumber:h,column:u,label:g}=this.parsePosition(n,c);if(a.visible=!!d,t.items=[{lineNumber:h,column:u,label:g}],t.ariaLabel=g,!h){this.clearDecorations(n);return}const f=this.toRange(h,u);n.revealRangeInCenter(f,0),this.addDecorations(n,f)},a=new Mh({title:m(1336,"Use Zero-Based Offset"),icon:Z.indexZero,isChecked:this.useZeroBasedOffset,inputActiveOptionBorder:te(X0),inputActiveOptionForeground:te(Q0),inputActiveOptionBackground:te(iC)});o.add(a.onChange(()=>{this.useZeroBasedOffset=!this.useZeroBasedOffset,r()})),t.toggles=[a],r(),o.add(t.onDidChangeValue(()=>r()));const l=Qne(n);return l&&l.getOptions().get(76).renderType===2&&(l.updateOptions({lineNumbers:"on"}),o.add(pe(()=>l.updateOptions({lineNumbers:"relative"})))),o}toRange(e=1,t=1){return{startLineNumber:e,startColumn:t,endLineNumber:e,endColumn:t}}parsePosition(e,t){const i=this.getModel(e);if(!i)return{label:m(1337,"Open a text editor first to go to a line.")};if(t.startsWith(":")){let n=parseInt(t.substring(1),10);const o=i.getValueLength();if(isNaN(n))return{inOffsetMode:!0,label:this.useZeroBasedOffset?m(1338,"Type a character position to go to (from 0 to {0}).",o-1):m(1339,"Type a character position to go to (from 1 to {0}).",o)};{const r=n<0;this.useZeroBasedOffset||(n-=Math.sign(n)),r&&(n+=o);const a=i.getPositionAt(n);return{...a,inOffsetMode:!0,label:m(1340,"Press 'Enter' to go to line {0} at column {1}.",a.lineNumber,a.column)}}}else{const n=t.split(/,|:|#/),o=i.getLineCount();let r=parseInt(n[0]?.trim(),10);if(n.length<1||isNaN(r))return{label:m(1341,"Type a line number to go to (from 1 to {0}).",o)};r=r>=0?r:o+1+r,r=Math.min(Math.max(1,r),o);const a=i.getLineMaxColumn(r);let l=parseInt(n[1]?.trim(),10);return n.length<2||isNaN(l)?{lineNumber:r,column:1,label:n.length<2?m(1342,"Press 'Enter' to go to line {0} or enter : to add a column number.",r):m(1343,"Press 'Enter' to go to line {0} or enter a column number (from 1 to {1}).",r,a)}:(l=l>=0?l:a+l,l=Math.min(Math.max(1,l),a),{lineNumber:r,column:l,label:m(1344,"Press 'Enter' to go to line {0} at column {1}.",r,l)})}}};Ff.PREFIX=":",Ff.ZERO_BASED_OFFSET_STORAGE_KEY="gotoLine.useZeroBasedOffset";let X8=Ff;var L8e=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},jK=function(s,e){return function(t,i){e(t,i,s)}};let C0=class extends X8{constructor(e,t){super(),this.editorService=e,this.storageService=t,this.onDidActiveTextEditorControlChange=oe.None}get activeTextEditorControl(){return this.editorService.getFocusedCodeEditor()??void 0}};C0=L8e([jK(0,ht),jK(1,Bs)],C0);var um;let hre=(um=class extends ue{constructor(){super({id:um.ID,label:rT.gotoLineActionLabel,alias:"Go to Line/Column...",precondition:void 0,kbOpts:{kbExpr:N.focus,primary:2085,mac:{primary:293},weight:100}})}run(e){e.get(bs).quickAccess.show(C0.PREFIX)}},um.ID="editor.action.gotoLine",um);ae(hre);bi.as(y_.Quickaccess).registerQuickAccessProvider({ctor:C0,prefix:C0.PREFIX,helpEntries:[{description:rT.gotoLineActionLabel,commandId:hre.ID}]});const ure=[void 0,[]];function RO(s,e,t=0,i=0){const n=e;return n.values&&n.values.length>1?k8e(s,n.values,t,i):gre(s,e,t,i)}function k8e(s,e,t,i){let n=0;const o=[];for(const r of e){const[a,l]=gre(s,r,t,i);if(typeof a!="number")return ure;n+=a,o.push(...l)}return[n,E8e(o)]}function gre(s,e,t,i){const n=Om(e.original,e.originalLowercase,t,s,s.toLowerCase(),i,{firstMatchCanBeWeak:!0,boostFullMatch:!0});return n?[n[0],ox(n)]:ure}function E8e(s){const e=s.sort((n,o)=>n.start-o.start),t=[];let i;for(const n of e)!i||!I8e(i,n)?(i=n,t.push(n)):(i.start=Math.min(i.start,n.start),i.end=Math.max(i.end,n.end));return t}function I8e(s,e){return!(s.end<e.start||e.end<s.start)}function GK(s){return s.startsWith('"')&&s.endsWith('"')}const fre=" ";function Q8(s){typeof s!="string"&&(s="");const e=s.toLowerCase(),{pathNormalized:t,normalized:i,normalizedLowercase:n}=KK(s),o=t.indexOf(bl)>=0,r=GK(s);let a;const l=s.split(fre);if(l.length>1)for(const c of l){const d=GK(c),{pathNormalized:h,normalized:u,normalizedLowercase:g}=KK(c);u&&(a||(a=[]),a.push({original:c,originalLowercase:c.toLowerCase(),pathNormalized:h,normalized:u,normalizedLowercase:g,expectContiguousMatch:d}))}return{original:s,originalLowercase:e,pathNormalized:t,normalized:i,normalizedLowercase:n,values:a,containsPathSeparator:o,expectContiguousMatch:r}}function KK(s){let e;Bn?e=s.replace(/\//g,bl):e=s.replace(/\\/g,bl);const t=e.replace(/[\*\u2026\s"]/g,"");return{pathNormalized:e,normalized:t,normalizedLowercase:t.toLowerCase()}}function YK(s){return Array.isArray(s)?Q8(s.map(e=>e.original).join(fre)):Q8(s.original)}var D8e=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},ZK=function(s,e){return function(t,i){e(t,i,s)}},iE,ul;let mp=(ul=class extends dre{constructor(e,t,i=Object.create(null)){super(i),this._languageFeaturesService=e,this._outlineModelService=t,this.options=i,this.options.canAcceptInBackground=!0}provideWithoutTextEditor(e){return this.provideLabelPick(e,m(1345,"To go to a symbol, first open a text editor with symbol information.")),F.None}provideWithTextEditor(e,t,i,n){const o=e.editor,r=this.getModel(o);return r?this._languageFeaturesService.documentSymbolProvider.has(r)?this.doProvideWithEditorSymbols(e,r,t,i,n):this.doProvideWithoutEditorSymbols(e,r,t,i):F.None}doProvideWithoutEditorSymbols(e,t,i,n){const o=new U;return this.provideLabelPick(i,m(1346,"The active text editor does not provide symbol information.")),(async()=>!await this.waitForLanguageSymbolRegistry(t,o)||n.isCancellationRequested||o.add(this.doProvideWithEditorSymbols(e,t,i,n)))(),o}provideLabelPick(e,t){e.items=[{label:t,index:0,kind:14}],e.ariaLabel=t}async waitForLanguageSymbolRegistry(e,t){if(this._languageFeaturesService.documentSymbolProvider.has(e))return!0;const i=new n_,n=t.add(this._languageFeaturesService.documentSymbolProvider.onDidChange(()=>{this._languageFeaturesService.documentSymbolProvider.has(e)&&(n.dispose(),i.complete(!0))}));return t.add(pe(()=>i.complete(!1))),i.p}doProvideWithEditorSymbols(e,t,i,n,o){const r=e.editor,a=new U;a.add(i.onDidAccept(h=>{const[u]=i.selectedItems;u&&u.range&&(this.gotoLocation(e,{range:u.range.selection,keyMods:i.keyMods,preserveFocus:h.inBackground}),o?.handleAccept?.(u,h.inBackground),h.inBackground||i.hide())})),a.add(i.onDidTriggerItemButton(({item:h})=>{h&&h.range&&(this.gotoLocation(e,{range:h.range.selection,keyMods:i.keyMods,forceSideBySide:!0}),i.hide())}));const l=this.getDocumentSymbols(t,n),c=a.add(new vt),d=async h=>{c?.value?.cancel(),i.busy=!1,c.value=new si,i.busy=!0;try{const u=Q8(i.value.substr(iE.PREFIX.length).trim()),g=await this.doGetSymbolPicks(l,u,void 0,c.value.token,t);if(n.isCancellationRequested)return;if(g.length>0){if(i.items=g,h&&u.original.length===0){const f=gS(g,p=>!!(p.type!=="separator"&&p.range&&k.containsPosition(p.range.decoration,h)));f&&(i.activeItems=[f])}}else u.original.length>0?this.provideLabelPick(i,m(1347,"No matching editor symbols")):this.provideLabelPick(i,m(1348,"No editor symbols"))}finally{n.isCancellationRequested||(i.busy=!1)}};return a.add(i.onDidChangeValue(()=>d(void 0))),d(r.getSelection()?.getPosition()),a.add(i.onDidChangeActive(()=>{const[h]=i.activeItems;h&&h.range&&(r.revealRangeInCenter(h.range.selection,0),this.addDecorations(r,h.range.decoration))})),a}async doGetSymbolPicks(e,t,i,n,o){const r=await e;if(n.isCancellationRequested)return[];const a=t.original.indexOf(iE.SCOPE_PREFIX)===0,l=a?1:0;let c,d;t.values&&t.values.length>1?(c=YK(t.values[0]),d=YK(t.values.slice(1))):c=t;let h;const u=this.options?.openSideBySideDirection?.();u&&(h=[{iconClass:u==="right"?xe.asClassName(Z.splitHorizontal):xe.asClassName(Z.splitVertical),tooltip:u==="right"?m(1349,"Open to the Side"):m(1350,"Open to the Bottom")}]);const g=[];for(let _=0;_<r.length;_++){const b=r[_],w=Ode(b.name),C=`$(${iI.toIcon(b.kind).id}) ${w}`,v=C.length-w.length;let y=b.containerName;i?.extraContainerLabel&&(y?y=`${i.extraContainerLabel} • ${y}`:y=i.extraContainerLabel);let x,L,E,I;if(t.original.length>l){let O=!1;if(c!==t&&([x,L]=RO(C,{...t,values:void 0},l,v),typeof x=="number"&&(O=!0)),typeof x!="number"&&([x,L]=RO(C,c,l,v),typeof x!="number"))continue;if(!O&&d){if(y&&d.original.length>0&&([E,I]=RO(y,d)),typeof E!="number")continue;typeof x=="number"&&(x+=E)}}const R=b.tags&&b.tags.indexOf(1)>=0;g.push({index:_,kind:b.kind,score:x,label:C,ariaLabel:Jwe(b.name,b.kind),description:y,highlights:R?void 0:{label:L,description:I},range:{selection:k.collapseToStart(b.selectionRange),decoration:b.range},uri:o.uri,symbolName:w,strikethrough:R,buttons:h})}const f=g.sort((_,b)=>a?this.compareByKindAndScore(_,b):this.compareByScore(_,b));let p=[];if(a){let C=function(){b&&typeof _=="number"&&w>0&&(b.label=xm(MO[_]||AO,w))},_,b,w=0;for(const v of f)_!==v.kind?(C(),_=v.kind,w=1,b={type:"separator"},p.push(b)):w++,p.push(v);C()}else f.length>0&&(p=[{label:m(1351,"symbols ({0})",g.length),type:"separator"},...f]);return p}compareByScore(e,t){if(typeof e.score!="number"&&typeof t.score=="number")return 1;if(typeof e.score=="number"&&typeof t.score!="number")return-1;if(typeof e.score=="number"&&typeof t.score=="number"){if(e.score>t.score)return-1;if(e.score<t.score)return 1}return e.index<t.index?-1:e.index>t.index?1:0}compareByKindAndScore(e,t){const i=MO[e.kind]||AO,n=MO[t.kind]||AO,o=i.localeCompare(n);return o===0?this.compareByScore(e,t):o}async getDocumentSymbols(e,t){const i=await this._outlineModelService.getOrCreate(e,t);return t.isCancellationRequested?[]:i.asListOfDocumentSymbols()}},iE=ul,ul.PREFIX="@",ul.SCOPE_PREFIX=":",ul.PREFIX_BY_CATEGORY=`${ul.PREFIX}${ul.SCOPE_PREFIX}`,ul);mp=iE=D8e([ZK(0,ge),ZK(1,xx)],mp);const AO=m(1352,"properties ({0})"),MO={5:m(1353,"methods ({0})"),11:m(1354,"functions ({0})"),8:m(1355,"constructors ({0})"),12:m(1356,"variables ({0})"),4:m(1357,"classes ({0})"),22:m(1358,"structs ({0})"),23:m(1359,"events ({0})"),24:m(1360,"operators ({0})"),10:m(1361,"interfaces ({0})"),2:m(1362,"namespaces ({0})"),3:m(1363,"packages ({0})"),25:m(1364,"type parameters ({0})"),1:m(1365,"modules ({0})"),6:m(1366,"properties ({0})"),9:m(1367,"enumerations ({0})"),21:m(1368,"enumeration members ({0})"),14:m(1369,"strings ({0})"),0:m(1370,"files ({0})"),17:m(1371,"arrays ({0})"),15:m(1372,"numbers ({0})"),16:m(1373,"booleans ({0})"),18:m(1374,"objects ({0})"),19:m(1375,"keys ({0})"),7:m(1376,"fields ({0})"),13:m(1377,"constants ({0})")};var T8e=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},OO=function(s,e){return function(t,i){e(t,i,s)}};let J8=class extends mp{constructor(e,t,i){super(t,i),this.editorService=e,this.onDidActiveTextEditorControlChange=oe.None}get activeTextEditorControl(){return this.editorService.getFocusedCodeEditor()??void 0}};J8=T8e([OO(0,ht),OO(1,ge),OO(2,xx)],J8);const aR=class aR extends ue{constructor(){super({id:aR.ID,label:w0.quickOutlineActionLabel,alias:"Go to Symbol...",precondition:N.hasDocumentSymbolProvider,kbOpts:{kbExpr:N.focus,primary:3117,weight:100},contextMenuOpts:{group:"navigation",order:3}})}run(e){e.get(bs).quickAccess.show(mp.PREFIX,{itemActivation:nl.NONE})}};aR.ID="editor.action.quickOutline";let cT=aR;ae(cT);bi.as(y_.Quickaccess).registerQuickAccessProvider({ctor:J8,prefix:mp.PREFIX,helpEntries:[{description:w0.quickOutlineActionLabel,prefix:mp.PREFIX,commandId:cT.ID},{description:w0.quickOutlineByCategoryActionLabel,prefix:mp.PREFIX_BY_CATEGORY}]});function N8e(s){const e=new Map;for(const t of s)e.set(t,(e.get(t)??0)+1);return e}class uy{constructor(){this.chunkCount=0,this.chunkOccurrences=new Map,this.documents=new Map}calculateScores(e,t){const i=this.computeEmbedding(e),n=new Map,o=[];for(const[r,a]of this.documents){if(t.isCancellationRequested)return[];for(const l of a.chunks){const c=this.computeSimilarityScore(l,i,n);c>0&&o.push({key:r,score:c})}}return o}static termFrequencies(e){return N8e(uy.splitTerms(e))}static*splitTerms(e){const t=i=>i.toLowerCase();for(const[i]of e.matchAll(new RegExp("\\b\\p{Letter}[\\p{Letter}\\d]{2,}\\b","gu"))){yield t(i);const n=i.replace(/([a-z])([A-Z])/g,"$1 $2").split(/\s+/g);if(n.length>1)for(const o of n)o.length>2&&new RegExp("\\p{Letter}{3,}","gu").test(o)&&(yield t(o))}}updateDocuments(e){for(const{key:t}of e)this.deleteDocument(t);for(const t of e){const i=[];for(const n of t.textChunks){const o=uy.termFrequencies(n);for(const r of o.keys())this.chunkOccurrences.set(r,(this.chunkOccurrences.get(r)??0)+1);i.push({text:n,tf:o})}this.chunkCount+=i.length,this.documents.set(t.key,{chunks:i})}return this}deleteDocument(e){const t=this.documents.get(e);if(t){this.documents.delete(e),this.chunkCount-=t.chunks.length;for(const i of t.chunks)for(const n of i.tf.keys()){const o=this.chunkOccurrences.get(n);if(typeof o=="number"){const r=o-1;r<=0?this.chunkOccurrences.delete(n):this.chunkOccurrences.set(n,r)}}}}computeSimilarityScore(e,t,i){let n=0;for(const[o,r]of Object.entries(t)){const a=e.tf.get(o);if(!a)continue;let l=i.get(o);typeof l!="number"&&(l=this.computeIdf(o),i.set(o,l));const c=a*l;n+=c*r}return n}computeEmbedding(e){const t=uy.termFrequencies(e);return this.computeTfidf(t)}computeIdf(e){const t=this.chunkOccurrences.get(e)??0;return t>0?Math.log((this.chunkCount+1)/t):0}computeTfidf(e){const t=Object.create(null);for(const[i,n]of e){const o=this.computeIdf(i);o>0&&(t[i]=n*o)}return t}}function R8e(s){const e=s.slice(0);e.sort((i,n)=>n.score-i.score);const t=e[0]?.score??0;if(t>0)for(const i of e)i.score/=t;return e}var Bb;(function(s){s[s.NO_ACTION=0]="NO_ACTION",s[s.CLOSE_PICKER=1]="CLOSE_PICKER",s[s.REFRESH_PICKER=2]="REFRESH_PICKER",s[s.REMOVE_ITEM=3]="REMOVE_ITEM"})(Bb||(Bb={}));function PO(s){return Array.isArray(s.items)}function XK(s){const e=s;return!!e.picks&&e.additionalPicks instanceof Promise}class A8e extends F{constructor(e,t){super(),this.prefix=e,this.options=t}provide(e,t,i){const n=new U;e.canAcceptInBackground=!!this.options?.canAcceptInBackground,e.matchOnLabel=e.matchOnDescription=e.matchOnDetail=e.sortByLabel=!1;let o;const r=n.add(new vt),a=async()=>{o?.dispose(!0),e.busy=!1;const c=r.value=new U;o=c.add(new si(t));const d=o.token;let h=e.value.substring(this.prefix.length);this.options?.shouldSkipTrimPickFilter||(h=h.trim());const u=this._getPicks(h,c,d,i),g=(p,_)=>{let b,w;if(PO(p)?(b=p.items,w=p.active):b=p,b.length===0){if(_)return!1;(h.length>0||e.hideInput)&&this.options?.noResultsPick&&(vm(this.options.noResultsPick)?b=[this.options.noResultsPick(h)]:b=[this.options.noResultsPick])}return e.items=b,w&&(e.activeItems=[w]),!0},f=async p=>{let _=!1,b=!1;await Promise.all([(async()=>{typeof p.mergeDelay=="number"&&(await Tc(p.mergeDelay),d.isCancellationRequested)||b||(_=g(p.picks,!0))})(),(async()=>{e.busy=!0;try{const w=await p.additionalPicks;if(d.isCancellationRequested)return;let C,v;PO(p.picks)?(C=p.picks.items,v=p.picks.active):C=p.picks;let y,x;if(PO(w)?(y=w.items,x=w.active):y=w,y.length>0||!_){let L;if(!v&&!x){const E=e.activeItems[0];E&&C.indexOf(E)!==-1&&(L=E)}g({items:[...C,...y],active:v||x||L})}}finally{d.isCancellationRequested||(e.busy=!1),b=!0}})()])};if(u!==null)if(XK(u))await f(u);else if(!(u instanceof Promise))g(u);else{e.busy=!0;try{const p=await u;if(d.isCancellationRequested)return;XK(p)?await f(p):g(p)}finally{d.isCancellationRequested||(e.busy=!1)}}};n.add(e.onDidChangeValue(()=>a())),a(),n.add(e.onDidAccept(c=>{if(i?.handleAccept){c.inBackground||e.hide(),i.handleAccept?.(e.activeItems[0],c.inBackground);return}const[d]=e.selectedItems;typeof d?.accept=="function"&&(c.inBackground||e.hide(),d.accept(e.keyMods,c))}));const l=async(c,d)=>{if(typeof d.trigger!="function")return;const h=d.buttons?.indexOf(c)??-1;if(h>=0){const u=d.trigger(h,e.keyMods),g=typeof u=="number"?u:await u;if(t.isCancellationRequested)return;switch(g){case Bb.NO_ACTION:break;case Bb.CLOSE_PICKER:e.hide();break;case Bb.REFRESH_PICKER:a();break;case Bb.REMOVE_ITEM:{const f=e.items.indexOf(d);if(f!==-1){const p=e.items.slice(),_=p.splice(f,1),b=e.activeItems.filter(C=>C!==_[0]),w=e.keepScrollPosition;e.keepScrollPosition=!0,e.items=p,b&&(e.activeItems=b),e.keepScrollPosition=w}break}}}};return n.add(e.onDidTriggerItemButton(({button:c,item:d})=>l(c,d))),n.add(e.onDidTriggerSeparatorButton(({button:c,separator:d})=>l(c,d))),n}}new Va(1e4);const M8e=new Va(1e4);function O8e(s){return F8e(s,"NFD",M8e)}const P8e=/[^\u0000-\u0080]/;function F8e(s,e,t){if(!s)return s;const i=t.get(s);if(i)return i;let n;return P8e.test(s)?n=s.normalize(e):n=s,t.set(s,n),n}const B8e=(function(){const s=/[\u0300-\u036f]/g;return function(e){return O8e(e).replace(s,"")}})();var pre=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},du=function(s,e){return function(t,i){e(t,i,s)}},cf,en,wc;let eB=(wc=class extends A8e{constructor(e,t,i,n,o,r){super(cf.PREFIX,e),this.keybindingService=i,this.commandService=n,this.telemetryService=o,this.dialogService=r,this.commandsHistory=this._register(t.createInstance(tB)),this.options=e}async _getPicks(e,t,i,n){const o=await this.getCommandPicks(i);if(i.isCancellationRequested)return[];const r=ym(()=>{const f=new uy;f.updateDocuments(o.map(_=>({key:_.commandId,textChunks:[this.getTfIdfChunk(_)]})));const p=f.calculateScores(e,i);return R8e(p).filter(_=>_.score>cf.TFIDF_THRESHOLD).slice(0,cf.TFIDF_MAX_RESULTS)}),a=this.normalizeForFiltering(e),l=[];for(const f of o){f.labelNoAccents??=this.normalizeForFiltering(f.label);const p=cf.WORD_FILTER(a,f.labelNoAccents)??void 0;let _;if(f.commandAlias&&(f.aliasNoAccents??=this.normalizeForFiltering(f.commandAlias),_=cf.WORD_FILTER(a,f.aliasNoAccents)??void 0),p||_)f.highlights={label:p,detail:this.options.showAlias?_:void 0},l.push(f);else if(e===f.commandId)l.push(f);else if(e.length>=3){const b=r();if(i.isCancellationRequested)return[];const w=b.find(C=>C.key===f.commandId);w&&(f.tfIdfScore=w.score,l.push(f))}}const c=new Map;for(const f of l){const p=c.get(f.label);p?(f.description=f.commandId,p.description=p.commandId):c.set(f.label,f)}l.sort((f,p)=>{if(f.tfIdfScore&&p.tfIdfScore)return f.tfIdfScore===p.tfIdfScore?f.label.localeCompare(p.label):p.tfIdfScore-f.tfIdfScore;if(f.tfIdfScore)return 1;if(p.tfIdfScore)return-1;const _=this.commandsHistory.peek(f.commandId),b=this.commandsHistory.peek(p.commandId);if(_&&b)return _>b?-1:1;if(_)return-1;if(b)return 1;if(this.options.suggestedCommandIds){const v=this.options.suggestedCommandIds.has(f.commandId),y=this.options.suggestedCommandIds.has(p.commandId);if(v&&y)return 0;if(v)return-1;if(y)return 1}const w=f.commandCategory===k8.Developer.value,C=p.commandCategory===k8.Developer.value;return w&&!C?1:!w&&C?-1:f.label.localeCompare(p.label)});const d=[];let h=!1,u=!0,g=!!this.options.suggestedCommandIds;for(let f=0;f<l.length;f++){const p=l[f];f===0&&this.commandsHistory.peek(p.commandId)&&(d.push({type:"separator",label:m(1740,"recently used")}),h=!0),u&&p.tfIdfScore!==void 0&&(d.push({type:"separator",label:m(1741,"similar commands")}),u=!1),g&&p.tfIdfScore===void 0&&!this.commandsHistory.peek(p.commandId)&&this.options.suggestedCommandIds?.has(p.commandId)&&(d.push({type:"separator",label:m(1742,"commonly used")}),h=!0,g=!1),h&&p.tfIdfScore===void 0&&!this.commandsHistory.peek(p.commandId)&&!this.options.suggestedCommandIds?.has(p.commandId)&&(d.push({type:"separator",label:m(1743,"other commands")}),h=!1),d.push(this.toCommandPick(p,n))}return this.hasAdditionalCommandPicks(e,i)?{picks:d,additionalPicks:(async()=>{const f=await this.getAdditionalCommandPicks(o,l,e,i);if(i.isCancellationRequested)return[];const p=f.map(_=>this.toCommandPick(_,n));return u&&p[0]?.type!=="separator"&&p.unshift({type:"separator",label:m(1744,"similar commands")}),p})()}:d}toCommandPick(e,t){if(e.type==="separator")return e;const i=this.keybindingService.lookupKeybinding(e.commandId),n=i?m(1745,"{0}, {1}",e.label,i.getAriaLabel()):e.label;return{...e,ariaLabel:n,detail:this.options.showAlias&&e.commandAlias!==e.label?e.commandAlias:void 0,keybinding:i,accept:async()=>{this.commandsHistory.push(e.commandId),this.telemetryService.publicLog2("workbenchActionExecuted",{id:e.commandId,from:t?.from??"quick open"});try{e.args?.length?await this.commandService.executeCommand(e.commandId,...e.args):await this.commandService.executeCommand(e.commandId)}catch(o){Dr(o)||this.dialogService.error(m(1746,"Command '{0}' resulted in an error",e.label),zI(o))}}}}getTfIdfChunk({label:e,commandAlias:t,commandDescription:i}){let n=e;return t&&t!==e&&(n+=` - ${t}`),i&&i.value!==e&&(n+=` - ${i.value===i.original?i.value:`${i.value} (${i.original})`}`),n}normalizeForFiltering(e){const t=B8e(e);return t.length!==e.length?(this.telemetryService.publicLog2("QuickAccess:FilterLengthMismatch",{originalLength:e.length,normalizedLength:t.length}),e):t}},cf=wc,wc.PREFIX=">",wc.TFIDF_THRESHOLD=.5,wc.TFIDF_MAX_RESULTS=5,wc.WORD_FILTER=w7(DS,i0e,Cte),wc);eB=cf=pre([du(1,_e),du(2,ft),du(3,Gt),du(4,vs),du(5,kx)],eB);var gl;let tB=(gl=class extends F{constructor(e,t,i){super(),this.storageService=e,this.configurationService=t,this.logService=i,this.configuredCommandsHistoryLength=0,this.updateConfiguration(),this.load(),this.registerListeners()}registerListeners(){this._register(this.configurationService.onDidChangeConfiguration(e=>this.updateConfiguration(e))),this._register(this.storageService.onWillSaveState(e=>{e.reason===Fu.SHUTDOWN&&this.saveState()}))}updateConfiguration(e){e&&!e.affectsConfiguration("workbench.commandPalette.history")||(this.configuredCommandsHistoryLength=en.getConfiguredCommandHistoryLength(this.configurationService),en.cache&&en.cache.limit!==this.configuredCommandsHistoryLength&&(en.cache.limit=this.configuredCommandsHistoryLength,en.hasChanges=!0))}load(){const e=this.storageService.get(en.PREF_KEY_CACHE,0);let t;if(e)try{t=JSON.parse(e)}catch(n){this.logService.error(`[CommandsHistory] invalid data: ${n}`)}const i=en.cache=new Va(this.configuredCommandsHistoryLength,1);if(t){let n;t.usesLRU?n=t.entries:n=t.entries.sort((o,r)=>o.value-r.value),n.forEach(o=>i.set(o.key,o.value))}en.counter=this.storageService.getNumber(en.PREF_KEY_COUNTER,0,en.counter)}push(e){en.cache&&(en.cache.set(e,en.counter++),en.hasChanges=!0)}peek(e){return en.cache?.peek(e)}saveState(){if(!en.cache||!en.hasChanges)return;const e={usesLRU:!0,entries:[]};en.cache.forEach((t,i)=>e.entries.push({key:i,value:t})),this.storageService.store(en.PREF_KEY_CACHE,JSON.stringify(e),0,0),this.storageService.store(en.PREF_KEY_COUNTER,en.counter,0,0),en.hasChanges=!1}static getConfiguredCommandHistoryLength(e){const i=e.getValue().workbench?.commandPalette?.history;return typeof i=="number"?i:en.DEFAULT_COMMANDS_HISTORY_LENGTH}},en=gl,gl.DEFAULT_COMMANDS_HISTORY_LENGTH=50,gl.PREF_KEY_CACHE="commandPalette.mru.cache",gl.PREF_KEY_COUNTER="commandPalette.mru.counter",gl.counter=1,gl.hasChanges=!1,gl);tB=en=pre([du(0,Bs),du(1,He),du(2,jt)],tB);class W8e extends eB{constructor(e,t,i,n,o,r){super(e,t,i,n,o,r)}getCodeEditorCommandPicks(){const e=this.activeTextEditorControl;if(!e)return[];const t=[];for(const i of e.getSupportedActions()){let n;i.metadata?.description&&($ke(i.metadata.description)?n=i.metadata.description:n={original:i.metadata.description,value:i.metadata.description}),t.push({commandId:i.id,commandAlias:i.alias,commandDescription:n,label:L7(i.label)||i.id})}return t}}var H8e=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},sb=function(s,e){return function(t,i){e(t,i,s)}};let y0=class extends W8e{get activeTextEditorControl(){return this.codeEditorService.getFocusedCodeEditor()??void 0}constructor(e,t,i,n,o,r){super({showAlias:!1},e,i,n,o,r),this.codeEditorService=t}async getCommandPicks(){return this.getCodeEditorCommandPicks()}hasAdditionalCommandPicks(){return!1}async getAdditionalCommandPicks(){return[]}};y0=H8e([sb(0,_e),sb(1,ht),sb(2,ft),sb(3,Gt),sb(4,vs),sb(5,kx)],y0);const lR=class lR extends ue{constructor(){super({id:lR.ID,label:aT.quickCommandActionLabel,alias:"Command Palette",precondition:void 0,kbOpts:{kbExpr:N.focus,primary:59,weight:100},contextMenuOpts:{group:"z_commands",order:1}})}run(e){e.get(bs).quickAccess.show(y0.PREFIX)}};lR.ID="editor.action.quickCommand";let dT=lR;ae(dT);bi.as(y_.Quickaccess).registerQuickAccessProvider({ctor:y0,prefix:y0.PREFIX,helpEntries:[{description:aT.quickCommandHelp,commandId:dT.ID}]});var V8e=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},ob=function(s,e){return function(t,i){e(t,i,s)}};let iB=class extends Zm{constructor(e,t,i,n,o,r,a){super(!0,e,t,i,n,o,r,a)}};iB=V8e([ob(1,Ae),ob(2,ht),ob(3,Ii),ob(4,_e),ob(5,Bs),ob(6,He)],iB);lt(Zm.ID,iB,4);class z8e{constructor(e,t,i,n,o){this._parsedThemeRuleBrand=void 0,this.token=e,this.index=t,this.fontStyle=i,this.foreground=n,this.background=o}}function $8e(s){if(!s||!Array.isArray(s))return[];const e=[];let t=0;for(let i=0,n=s.length;i<n;i++){const o=s[i];let r=-1;if(typeof o.fontStyle=="string"){r=0;const c=o.fontStyle.split(" ");for(let d=0,h=c.length;d<h;d++)switch(c[d]){case"italic":r=r|1;break;case"bold":r=r|2;break;case"underline":r=r|4;break;case"strikethrough":r=r|8;break}}let a=null;typeof o.foreground=="string"&&(a=o.foreground);let l=null;typeof o.background=="string"&&(l=o.background),e[t++]=new z8e(o.token||"",i,r,a,l)}return e}function U8e(s,e){s.sort((d,h)=>{const u=Y8e(d.token,h.token);return u!==0?u:d.index-h.index});let t=0,i="000000",n="ffffff";for(;s.length>=1&&s[0].token==="";){const d=s.shift();d.fontStyle!==-1&&(t=d.fontStyle),d.foreground!==null&&(i=d.foreground),d.background!==null&&(n=d.background)}const o=new j8e;for(const d of e)o.getId(d);const r=o.getId(i),a=o.getId(n),l=new pV(t,r,a),c=new mV(l);for(let d=0,h=s.length;d<h;d++){const u=s[d];c.insert(u.token,u.fontStyle,o.getId(u.foreground),o.getId(u.background))}return new mre(o,c)}const q8e=/^#?([0-9A-Fa-f]{6})([0-9A-Fa-f]{2})?$/;class j8e{constructor(){this._lastColorId=0,this._id2color=[],this._color2id=new Map}getId(e){if(e===null)return 0;const t=e.match(q8e);if(!t)throw new Error("Illegal value for token color: "+e);e=t[1].toUpperCase();let i=this._color2id.get(e);return i||(i=++this._lastColorId,this._color2id.set(e,i),this._id2color[i]=q.fromHex("#"+e),i)}getColorMap(){return this._id2color.slice(0)}}class mre{static createFromRawTokenTheme(e,t){return this.createFromParsedTokenTheme($8e(e),t)}static createFromParsedTokenTheme(e,t){return U8e(e,t)}constructor(e,t){this._colorMap=e,this._root=t,this._cache=new Map}getColorMap(){return this._colorMap.getColorMap()}_match(e){return this._root.match(e)}match(e,t){let i=this._cache.get(t);if(typeof i>"u"){const n=this._match(t),o=K8e(t);i=(n.metadata|o<<8)>>>0,this._cache.set(t,i)}return(i|e<<0)>>>0}}const G8e=/\b(comment|string|regex|regexp)\b/;function K8e(s){const e=s.match(G8e);if(!e)return 0;switch(e[1]){case"comment":return 1;case"string":return 2;case"regex":return 3;case"regexp":return 3}throw new Error("Unexpected match for standard token type!")}function Y8e(s,e){return s<e?-1:s>e?1:0}class pV{constructor(e,t,i){this._themeTrieElementRuleBrand=void 0,this._fontStyle=e,this._foreground=t,this._background=i,this.metadata=(this._fontStyle<<11|this._foreground<<15|this._background<<24)>>>0}clone(){return new pV(this._fontStyle,this._foreground,this._background)}acceptOverwrite(e,t,i){e!==-1&&(this._fontStyle=e),t!==0&&(this._foreground=t),i!==0&&(this._background=i),this.metadata=(this._fontStyle<<11|this._foreground<<15|this._background<<24)>>>0}}class mV{constructor(e){this._themeTrieElementBrand=void 0,this._mainRule=e,this._children=new Map}match(e){if(e==="")return this._mainRule;const t=e.indexOf(".");let i,n;t===-1?(i=e,n=""):(i=e.substring(0,t),n=e.substring(t+1));const o=this._children.get(i);return typeof o<"u"?o.match(n):this._mainRule}insert(e,t,i,n){if(e===""){this._mainRule.acceptOverwrite(t,i,n);return}const o=e.indexOf(".");let r,a;o===-1?(r=e,a=""):(r=e.substring(0,o),a=e.substring(o+1));let l=this._children.get(r);typeof l>"u"&&(l=new mV(this._mainRule.clone()),this._children.set(r,l)),l.insert(a,t,i,n)}}function Z8e(s){const e=[];for(let t=1,i=s.length;t<i;t++){const n=s[t];e[t]=`.mtk${t} { color: ${n}; }`}return e.push(".mtki { font-style: italic; }"),e.push(".mtkb { font-weight: bold; }"),e.push(".mtku { text-decoration: underline; text-underline-position: under; }"),e.push(".mtks { text-decoration: line-through; }"),e.push(".mtks.mtku { text-decoration: underline line-through; text-underline-position: under; }"),e.join(`
`)}const X8e={base:"vs",inherit:!1,rules:[{token:"",foreground:"000000",background:"fffffe"},{token:"invalid",foreground:"cd3131"},{token:"emphasis",fontStyle:"italic"},{token:"strong",fontStyle:"bold"},{token:"variable",foreground:"001188"},{token:"variable.predefined",foreground:"4864AA"},{token:"constant",foreground:"dd0000"},{token:"comment",foreground:"008000"},{token:"number",foreground:"098658"},{token:"number.hex",foreground:"3030c0"},{token:"regexp",foreground:"800000"},{token:"annotation",foreground:"808080"},{token:"type",foreground:"008080"},{token:"delimiter",foreground:"000000"},{token:"delimiter.html",foreground:"383838"},{token:"delimiter.xml",foreground:"0000FF"},{token:"tag",foreground:"800000"},{token:"tag.id.pug",foreground:"4F76AC"},{token:"tag.class.pug",foreground:"4F76AC"},{token:"meta.scss",foreground:"800000"},{token:"metatag",foreground:"e00000"},{token:"metatag.content.html",foreground:"FF0000"},{token:"metatag.html",foreground:"808080"},{token:"metatag.xml",foreground:"808080"},{token:"metatag.php",fontStyle:"bold"},{token:"key",foreground:"863B00"},{token:"string.key.json",foreground:"A31515"},{token:"string.value.json",foreground:"0451A5"},{token:"attribute.name",foreground:"FF0000"},{token:"attribute.value",foreground:"0451A5"},{token:"attribute.value.number",foreground:"098658"},{token:"attribute.value.unit",foreground:"098658"},{token:"attribute.value.html",foreground:"0000FF"},{token:"attribute.value.xml",foreground:"0000FF"},{token:"string",foreground:"A31515"},{token:"string.html",foreground:"0000FF"},{token:"string.sql",foreground:"FF0000"},{token:"string.yaml",foreground:"0451A5"},{token:"keyword",foreground:"0000FF"},{token:"keyword.json",foreground:"0451A5"},{token:"keyword.flow",foreground:"AF00DB"},{token:"keyword.flow.scss",foreground:"0000FF"},{token:"operator.scss",foreground:"666666"},{token:"operator.sql",foreground:"778899"},{token:"operator.swift",foreground:"666666"},{token:"predefined.sql",foreground:"C700C7"}],colors:{[Bi]:"#FFFFFE",[zc]:"#000000",[OJ]:"#E5EBF1",[ix]:"#D3D3D3",[nx]:"#939393",[NW]:"#ADD6FF4D"}},Q8e={base:"vs-dark",inherit:!1,rules:[{token:"",foreground:"D4D4D4",background:"1E1E1E"},{token:"invalid",foreground:"f44747"},{token:"emphasis",fontStyle:"italic"},{token:"strong",fontStyle:"bold"},{token:"variable",foreground:"74B0DF"},{token:"variable.predefined",foreground:"4864AA"},{token:"variable.parameter",foreground:"9CDCFE"},{token:"constant",foreground:"569CD6"},{token:"comment",foreground:"608B4E"},{token:"number",foreground:"B5CEA8"},{token:"number.hex",foreground:"5BB498"},{token:"regexp",foreground:"B46695"},{token:"annotation",foreground:"cc6666"},{token:"type",foreground:"3DC9B0"},{token:"delimiter",foreground:"DCDCDC"},{token:"delimiter.html",foreground:"808080"},{token:"delimiter.xml",foreground:"808080"},{token:"tag",foreground:"569CD6"},{token:"tag.id.pug",foreground:"4F76AC"},{token:"tag.class.pug",foreground:"4F76AC"},{token:"meta.scss",foreground:"A79873"},{token:"meta.tag",foreground:"CE9178"},{token:"metatag",foreground:"DD6A6F"},{token:"metatag.content.html",foreground:"9CDCFE"},{token:"metatag.html",foreground:"569CD6"},{token:"metatag.xml",foreground:"569CD6"},{token:"metatag.php",fontStyle:"bold"},{token:"key",foreground:"9CDCFE"},{token:"string.key.json",foreground:"9CDCFE"},{token:"string.value.json",foreground:"CE9178"},{token:"attribute.name",foreground:"9CDCFE"},{token:"attribute.value",foreground:"CE9178"},{token:"attribute.value.number.css",foreground:"B5CEA8"},{token:"attribute.value.unit.css",foreground:"B5CEA8"},{token:"attribute.value.hex.css",foreground:"D4D4D4"},{token:"string",foreground:"CE9178"},{token:"string.sql",foreground:"FF0000"},{token:"keyword",foreground:"569CD6"},{token:"keyword.flow",foreground:"C586C0"},{token:"keyword.json",foreground:"CE9178"},{token:"keyword.flow.scss",foreground:"569CD6"},{token:"operator.scss",foreground:"909090"},{token:"operator.sql",foreground:"778899"},{token:"operator.swift",foreground:"909090"},{token:"predefined.sql",foreground:"FF00FF"}],colors:{[Bi]:"#1E1E1E",[zc]:"#D4D4D4",[OJ]:"#3A3D41",[ix]:"#404040",[nx]:"#707070",[NW]:"#ADD6FF26"}},J8e={base:"hc-black",inherit:!1,rules:[{token:"",foreground:"FFFFFF",background:"000000"},{token:"invalid",foreground:"f44747"},{token:"emphasis",fontStyle:"italic"},{token:"strong",fontStyle:"bold"},{token:"variable",foreground:"1AEBFF"},{token:"variable.parameter",foreground:"9CDCFE"},{token:"constant",foreground:"569CD6"},{token:"comment",foreground:"608B4E"},{token:"number",foreground:"FFFFFF"},{token:"regexp",foreground:"C0C0C0"},{token:"annotation",foreground:"569CD6"},{token:"type",foreground:"3DC9B0"},{token:"delimiter",foreground:"FFFF00"},{token:"delimiter.html",foreground:"FFFF00"},{token:"tag",foreground:"569CD6"},{token:"tag.id.pug",foreground:"4F76AC"},{token:"tag.class.pug",foreground:"4F76AC"},{token:"meta",foreground:"D4D4D4"},{token:"meta.tag",foreground:"CE9178"},{token:"metatag",foreground:"569CD6"},{token:"metatag.content.html",foreground:"1AEBFF"},{token:"metatag.html",foreground:"569CD6"},{token:"metatag.xml",foreground:"569CD6"},{token:"metatag.php",fontStyle:"bold"},{token:"key",foreground:"9CDCFE"},{token:"string.key",foreground:"9CDCFE"},{token:"string.value",foreground:"CE9178"},{token:"attribute.name",foreground:"569CD6"},{token:"attribute.value",foreground:"3FF23F"},{token:"string",foreground:"CE9178"},{token:"string.sql",foreground:"FF0000"},{token:"keyword",foreground:"569CD6"},{token:"keyword.flow",foreground:"C586C0"},{token:"operator.sql",foreground:"778899"},{token:"operator.swift",foreground:"909090"},{token:"predefined.sql",foreground:"FF00FF"}],colors:{[Bi]:"#000000",[zc]:"#FFFFFF",[ix]:"#FFFFFF",[nx]:"#FFFFFF"}},eBe={base:"hc-light",inherit:!1,rules:[{token:"",foreground:"292929",background:"FFFFFF"},{token:"invalid",foreground:"B5200D"},{token:"emphasis",fontStyle:"italic"},{token:"strong",fontStyle:"bold"},{token:"variable",foreground:"264F70"},{token:"variable.predefined",foreground:"4864AA"},{token:"constant",foreground:"dd0000"},{token:"comment",foreground:"008000"},{token:"number",foreground:"098658"},{token:"number.hex",foreground:"3030c0"},{token:"regexp",foreground:"800000"},{token:"annotation",foreground:"808080"},{token:"type",foreground:"008080"},{token:"delimiter",foreground:"000000"},{token:"delimiter.html",foreground:"383838"},{token:"tag",foreground:"800000"},{token:"tag.id.pug",foreground:"4F76AC"},{token:"tag.class.pug",foreground:"4F76AC"},{token:"meta.scss",foreground:"800000"},{token:"metatag",foreground:"e00000"},{token:"metatag.content.html",foreground:"B5200D"},{token:"metatag.html",foreground:"808080"},{token:"metatag.xml",foreground:"808080"},{token:"metatag.php",fontStyle:"bold"},{token:"key",foreground:"863B00"},{token:"string.key.json",foreground:"A31515"},{token:"string.value.json",foreground:"0451A5"},{token:"attribute.name",foreground:"264F78"},{token:"attribute.value",foreground:"0451A5"},{token:"string",foreground:"A31515"},{token:"string.sql",foreground:"B5200D"},{token:"keyword",foreground:"0000FF"},{token:"keyword.flow",foreground:"AF00DB"},{token:"operator.sql",foreground:"778899"},{token:"operator.swift",foreground:"666666"},{token:"predefined.sql",foreground:"C700C7"}],colors:{[Bi]:"#FFFFFF",[zc]:"#292929",[ix]:"#292929",[nx]:"#292929"}};function tBe(s){const e=new U,t=e.add(new M),i=uie();return e.add(i.onDidChange(()=>t.fire())),s&&e.add(s.onDidProductIconThemeChange(()=>t.fire())),{dispose:()=>e.dispose(),onDidChange:t.event,getCSS(){const n=s?s.getProductIconTheme():new _re,o={},r=new CM,a=new CM;for(const l of i.getIcons()){const c=n.getIcon(l);if(!c)continue;const d=c.font,h=Do`--vscode-icon-${fL(l.id)}-font-family`,u=Do`--vscode-icon-${fL(l.id)}-content`;d?(o[d.id]=d.definition,a.push(Do`${h}: ${Hh(d.id)};`,Do`${u}: ${Hh(c.fontCharacter)};`),r.push(Do`.codicon-${fL(l.id)}:before { content: ${Hh(c.fontCharacter)}; font-family: ${Hh(d.id)}; }`)):(a.push(Do`${u}: ${Hh(c.fontCharacter)}; ${h}: 'codicon';`),r.push(Do`.codicon-${fL(l.id)}:before { content: ${Hh(c.fontCharacter)}; }`))}for(const l in o){const c=o[l],d=c.weight?Do`font-weight: ${rq(c.weight)};`:Do``,h=c.style?Do`font-style: ${rq(c.style)};`:Do``,u=new CM;for(const g of c.src)u.push(Do`${Mc(g.location)} format(${Hh(g.format)})`);r.push(Do`@font-face { src: ${u.join(", ")}; font-family: ${Hh(l)};${d}${h} font-display: block; }`)}return r.push(Do`:root { ${a.join(" ")} }`),r.join(`
`)}}}class _re{getIcon(e){const t=uie();let i=e.defaults;for(;xe.isThemeIcon(i);){const n=t.getIcon(i.id);if(!n)return;i=n.defaults}return i}}const Fd="vs",Tw="vs-dark",_p="hc-black",bp="hc-light",bre=bi.as(IJ.ColorContribution),iBe=bi.as(cee.ThemingContribution);class wre{constructor(e,t){this.semanticHighlighting=!1,this.themeData=t;const i=t.base;e.length>0?(nE(e)?this.id=e:this.id=i+" "+e,this.themeName=e):(this.id=i,this.themeName=i),this.colors=null,this.defaultColors=Object.create(null),this._tokenTheme=null}get base(){return this.themeData.base}notifyBaseUpdated(){this.themeData.inherit&&(this.colors=null,this._tokenTheme=null)}getColors(){if(!this.colors){const e=new Map;for(const t in this.themeData.colors)e.set(t,q.fromHex(this.themeData.colors[t]));if(this.themeData.inherit){const t=nB(this.themeData.base);for(const i in t.colors)e.has(i)||e.set(i,q.fromHex(t.colors[i]))}this.colors=e}return this.colors}getColor(e,t){const i=this.getColors().get(e);if(i)return i;if(t!==!1)return this.getDefault(e)}getDefault(e){let t=this.defaultColors[e];return t||(t=bre.resolveDefaultColor(e,this),this.defaultColors[e]=t,t)}defines(e){return this.getColors().has(e)}get type(){switch(this.base){case Fd:return Xr.LIGHT;case _p:return Xr.HIGH_CONTRAST_DARK;case bp:return Xr.HIGH_CONTRAST_LIGHT;default:return Xr.DARK}}get tokenTheme(){if(!this._tokenTheme){let e=[],t=[];if(this.themeData.inherit){const o=nB(this.themeData.base);e=o.rules,o.encodedTokensColors&&(t=o.encodedTokensColors)}const i=this.themeData.colors["editor.foreground"],n=this.themeData.colors["editor.background"];if(i||n){const o={token:""};i&&(o.foreground=i),n&&(o.background=n),e.push(o)}e=e.concat(this.themeData.rules),this.themeData.encodedTokensColors&&(t=this.themeData.encodedTokensColors),this._tokenTheme=mre.createFromRawTokenTheme(e,t)}return this._tokenTheme}getTokenStyleMetadata(e,t,i){const o=this.tokenTheme._match([e].concat(t).join(".")).metadata,r=ds.getForeground(o),a=ds.getFontStyle(o);return{foreground:r,italic:!!(a&1),bold:!!(a&2),underline:!!(a&4),strikethrough:!!(a&8)}}get tokenColorMap(){return[]}}function nE(s){return s===Fd||s===Tw||s===_p||s===bp}function nB(s){switch(s){case Fd:return X8e;case Tw:return Q8e;case _p:return J8e;case bp:return eBe}}function ik(s){const e=nB(s);return new wre(s,e)}class nBe extends F{constructor(){super(),this._onColorThemeChange=this._register(new M),this.onDidColorThemeChange=this._onColorThemeChange.event,this._onProductIconThemeChange=this._register(new M),this.onDidProductIconThemeChange=this._onProductIconThemeChange.event,this._environment=Object.create(null),this._builtInProductIconTheme=new _re,this._autoDetectHighContrast=!0,this._knownThemes=new Map,this._knownThemes.set(Fd,ik(Fd)),this._knownThemes.set(Tw,ik(Tw)),this._knownThemes.set(_p,ik(_p)),this._knownThemes.set(bp,ik(bp));const e=this._register(tBe(this));this._codiconCSS=e.getCSS(),this._themeCSS="",this._allCSS=`${this._codiconCSS}
${this._themeCSS}`,this._globalStyleElement=null,this._styleElements=[],this._colorMapOverride=null,this.setTheme(Fd),this._onOSSchemeChanged(),this._register(e.onDidChange(()=>{this._codiconCSS=e.getCSS(),this._updateCSS()})),PX(Nt,"(forced-colors: active)",()=>{this._onOSSchemeChanged()})}registerEditorContainer(e){return LE(e)?this._registerShadowDomContainer(e):this._registerRegularEditorContainer()}_registerRegularEditorContainer(){return this._globalStyleElement||(this._globalStyleElement=ca(void 0,e=>{e.className="monaco-colors",e.textContent=this._allCSS}),this._styleElements.push(this._globalStyleElement)),F.None}_registerShadowDomContainer(e){const t=ca(e,i=>{i.className="monaco-colors",i.textContent=this._allCSS});return this._styleElements.push(t),{dispose:()=>{for(let i=0;i<this._styleElements.length;i++)if(this._styleElements[i]===t){this._styleElements.splice(i,1);return}}}}defineTheme(e,t){if(!/^[a-z0-9\-]+$/i.test(e))throw new Error("Illegal theme name!");if(!nE(t.base)&&!nE(e))throw new Error("Illegal theme base!");this._knownThemes.set(e,new wre(e,t)),nE(e)&&this._knownThemes.forEach(i=>{i.base===e&&i.notifyBaseUpdated()}),this._theme.themeName===e&&this.setTheme(e)}getColorTheme(){return this._theme}setColorMapOverride(e){this._colorMapOverride=e,this._updateThemeOrColorMap()}setTheme(e){let t;this._knownThemes.has(e)?t=this._knownThemes.get(e):t=this._knownThemes.get(Fd),this._updateActualTheme(t)}_updateActualTheme(e){!e||this._theme===e||(this._theme=e,this._updateThemeOrColorMap())}_onOSSchemeChanged(){if(this._autoDetectHighContrast){const e=Nt.matchMedia("(forced-colors: active)").matches;if(e!==Cl(this._theme.type)){let t;wh(this._theme.type)?t=e?_p:Tw:t=e?bp:Fd,this._updateActualTheme(this._knownThemes.get(t))}}}setAutoDetectHighContrast(e){this._autoDetectHighContrast=e,this._onOSSchemeChanged()}_updateThemeOrColorMap(){const e=[],t={},i={addRule:r=>{t[r]||(e.push(r),t[r]=!0)}};iBe.getThemingParticipants().forEach(r=>r(this._theme,i,this._environment));const n=[];for(const r of bre.getColors()){const a=this._theme.getColor(r.id,!0);a&&n.push(`${DW(r.id)}: ${a.toString()};`)}i.addRule(`.monaco-editor, .monaco-diff-editor, .monaco-component { ${n.join(`
`)} }`);const o=this._colorMapOverride||this._theme.tokenTheme.getColorMap();i.addRule(Z8e(o)),i.addRule(".monaco-editor, .monaco-diff-editor, .monaco-component { forced-color-adjust: none; }"),this._themeCSS=e.join(`
`),this._updateCSS(),yi.setColorMap(o),this._onColorThemeChange.fire(this._theme)}_updateCSS(){this._allCSS=`${this._codiconCSS}
${this._themeCSS}`,this._styleElements.forEach(e=>e.textContent=this._allCSS)}getFileIconTheme(){return{hasFileIcons:!1,hasFolderIcons:!1,hidesExplorerArrows:!1}}getProductIconTheme(){return this._builtInProductIconTheme}}class sBe extends ue{constructor(){super({id:"editor.action.toggleHighContrast",label:U8.toggleHighContrast,alias:"Toggle High Contrast Theme",precondition:void 0}),this._originalThemeName=null}run(e,t){const i=e.get(Rr),n=i.getColorTheme();Cl(n.type)?(i.setTheme(this._originalThemeName||(wh(n.type)?Tw:Fd)),this._originalThemeName=null):(i.setTheme(wh(n.type)?_p:bp),this._originalThemeName=n.themeName)}}ae(sBe);var sB;(function(s){s[s.Unknown=0]="Unknown",s[s.Disabled=1]="Disabled",s[s.Enabled=2]="Enabled"})(sB||(sB={}));var oB;(function(s){s[s.Invoke=1]="Invoke",s[s.Auto=2]="Auto"})(oB||(oB={}));var rB;(function(s){s[s.None=0]="None",s[s.KeepWhitespace=1]="KeepWhitespace",s[s.InsertAsSnippet=4]="InsertAsSnippet"})(rB||(rB={}));var aB;(function(s){s[s.Method=0]="Method",s[s.Function=1]="Function",s[s.Constructor=2]="Constructor",s[s.Field=3]="Field",s[s.Variable=4]="Variable",s[s.Class=5]="Class",s[s.Struct=6]="Struct",s[s.Interface=7]="Interface",s[s.Module=8]="Module",s[s.Property=9]="Property",s[s.Event=10]="Event",s[s.Operator=11]="Operator",s[s.Unit=12]="Unit",s[s.Value=13]="Value",s[s.Constant=14]="Constant",s[s.Enum=15]="Enum",s[s.EnumMember=16]="EnumMember",s[s.Keyword=17]="Keyword",s[s.Text=18]="Text",s[s.Color=19]="Color",s[s.File=20]="File",s[s.Reference=21]="Reference",s[s.Customcolor=22]="Customcolor",s[s.Folder=23]="Folder",s[s.TypeParameter=24]="TypeParameter",s[s.User=25]="User",s[s.Issue=26]="Issue",s[s.Tool=27]="Tool",s[s.Snippet=28]="Snippet"})(aB||(aB={}));var lB;(function(s){s[s.Deprecated=1]="Deprecated"})(lB||(lB={}));var cB;(function(s){s[s.Invoke=0]="Invoke",s[s.TriggerCharacter=1]="TriggerCharacter",s[s.TriggerForIncompleteCompletions=2]="TriggerForIncompleteCompletions"})(cB||(cB={}));var dB;(function(s){s[s.EXACT=0]="EXACT",s[s.ABOVE=1]="ABOVE",s[s.BELOW=2]="BELOW"})(dB||(dB={}));var hB;(function(s){s[s.NotSet=0]="NotSet",s[s.ContentFlush=1]="ContentFlush",s[s.RecoverFromMarkers=2]="RecoverFromMarkers",s[s.Explicit=3]="Explicit",s[s.Paste=4]="Paste",s[s.Undo=5]="Undo",s[s.Redo=6]="Redo"})(hB||(hB={}));var uB;(function(s){s[s.LF=1]="LF",s[s.CRLF=2]="CRLF"})(uB||(uB={}));var gB;(function(s){s[s.Text=0]="Text",s[s.Read=1]="Read",s[s.Write=2]="Write"})(gB||(gB={}));var fB;(function(s){s[s.None=0]="None",s[s.Keep=1]="Keep",s[s.Brackets=2]="Brackets",s[s.Advanced=3]="Advanced",s[s.Full=4]="Full"})(fB||(fB={}));var pB;(function(s){s[s.acceptSuggestionOnCommitCharacter=0]="acceptSuggestionOnCommitCharacter",s[s.acceptSuggestionOnEnter=1]="acceptSuggestionOnEnter",s[s.accessibilitySupport=2]="accessibilitySupport",s[s.accessibilityPageSize=3]="accessibilityPageSize",s[s.allowOverflow=4]="allowOverflow",s[s.allowVariableLineHeights=5]="allowVariableLineHeights",s[s.allowVariableFonts=6]="allowVariableFonts",s[s.allowVariableFontsInAccessibilityMode=7]="allowVariableFontsInAccessibilityMode",s[s.ariaLabel=8]="ariaLabel",s[s.ariaRequired=9]="ariaRequired",s[s.autoClosingBrackets=10]="autoClosingBrackets",s[s.autoClosingComments=11]="autoClosingComments",s[s.screenReaderAnnounceInlineSuggestion=12]="screenReaderAnnounceInlineSuggestion",s[s.autoClosingDelete=13]="autoClosingDelete",s[s.autoClosingOvertype=14]="autoClosingOvertype",s[s.autoClosingQuotes=15]="autoClosingQuotes",s[s.autoIndent=16]="autoIndent",s[s.autoIndentOnPaste=17]="autoIndentOnPaste",s[s.autoIndentOnPasteWithinString=18]="autoIndentOnPasteWithinString",s[s.automaticLayout=19]="automaticLayout",s[s.autoSurround=20]="autoSurround",s[s.bracketPairColorization=21]="bracketPairColorization",s[s.guides=22]="guides",s[s.codeLens=23]="codeLens",s[s.codeLensFontFamily=24]="codeLensFontFamily",s[s.codeLensFontSize=25]="codeLensFontSize",s[s.colorDecorators=26]="colorDecorators",s[s.colorDecoratorsLimit=27]="colorDecoratorsLimit",s[s.columnSelection=28]="columnSelection",s[s.comments=29]="comments",s[s.contextmenu=30]="contextmenu",s[s.copyWithSyntaxHighlighting=31]="copyWithSyntaxHighlighting",s[s.cursorBlinking=32]="cursorBlinking",s[s.cursorSmoothCaretAnimation=33]="cursorSmoothCaretAnimation",s[s.cursorStyle=34]="cursorStyle",s[s.cursorSurroundingLines=35]="cursorSurroundingLines",s[s.cursorSurroundingLinesStyle=36]="cursorSurroundingLinesStyle",s[s.cursorWidth=37]="cursorWidth",s[s.cursorHeight=38]="cursorHeight",s[s.disableLayerHinting=39]="disableLayerHinting",s[s.disableMonospaceOptimizations=40]="disableMonospaceOptimizations",s[s.domReadOnly=41]="domReadOnly",s[s.dragAndDrop=42]="dragAndDrop",s[s.dropIntoEditor=43]="dropIntoEditor",s[s.editContext=44]="editContext",s[s.emptySelectionClipboard=45]="emptySelectionClipboard",s[s.experimentalGpuAcceleration=46]="experimentalGpuAcceleration",s[s.experimentalWhitespaceRendering=47]="experimentalWhitespaceRendering",s[s.extraEditorClassName=48]="extraEditorClassName",s[s.fastScrollSensitivity=49]="fastScrollSensitivity",s[s.find=50]="find",s[s.fixedOverflowWidgets=51]="fixedOverflowWidgets",s[s.folding=52]="folding",s[s.foldingStrategy=53]="foldingStrategy",s[s.foldingHighlight=54]="foldingHighlight",s[s.foldingImportsByDefault=55]="foldingImportsByDefault",s[s.foldingMaximumRegions=56]="foldingMaximumRegions",s[s.unfoldOnClickAfterEndOfLine=57]="unfoldOnClickAfterEndOfLine",s[s.fontFamily=58]="fontFamily",s[s.fontInfo=59]="fontInfo",s[s.fontLigatures=60]="fontLigatures",s[s.fontSize=61]="fontSize",s[s.fontWeight=62]="fontWeight",s[s.fontVariations=63]="fontVariations",s[s.formatOnPaste=64]="formatOnPaste",s[s.formatOnType=65]="formatOnType",s[s.glyphMargin=66]="glyphMargin",s[s.gotoLocation=67]="gotoLocation",s[s.hideCursorInOverviewRuler=68]="hideCursorInOverviewRuler",s[s.hover=69]="hover",s[s.inDiffEditor=70]="inDiffEditor",s[s.inlineSuggest=71]="inlineSuggest",s[s.letterSpacing=72]="letterSpacing",s[s.lightbulb=73]="lightbulb",s[s.lineDecorationsWidth=74]="lineDecorationsWidth",s[s.lineHeight=75]="lineHeight",s[s.lineNumbers=76]="lineNumbers",s[s.lineNumbersMinChars=77]="lineNumbersMinChars",s[s.linkedEditing=78]="linkedEditing",s[s.links=79]="links",s[s.matchBrackets=80]="matchBrackets",s[s.minimap=81]="minimap",s[s.mouseStyle=82]="mouseStyle",s[s.mouseWheelScrollSensitivity=83]="mouseWheelScrollSensitivity",s[s.mouseWheelZoom=84]="mouseWheelZoom",s[s.multiCursorMergeOverlapping=85]="multiCursorMergeOverlapping",s[s.multiCursorModifier=86]="multiCursorModifier",s[s.mouseMiddleClickAction=87]="mouseMiddleClickAction",s[s.multiCursorPaste=88]="multiCursorPaste",s[s.multiCursorLimit=89]="multiCursorLimit",s[s.occurrencesHighlight=90]="occurrencesHighlight",s[s.occurrencesHighlightDelay=91]="occurrencesHighlightDelay",s[s.overtypeCursorStyle=92]="overtypeCursorStyle",s[s.overtypeOnPaste=93]="overtypeOnPaste",s[s.overviewRulerBorder=94]="overviewRulerBorder",s[s.overviewRulerLanes=95]="overviewRulerLanes",s[s.padding=96]="padding",s[s.pasteAs=97]="pasteAs",s[s.parameterHints=98]="parameterHints",s[s.peekWidgetDefaultFocus=99]="peekWidgetDefaultFocus",s[s.placeholder=100]="placeholder",s[s.definitionLinkOpensInPeek=101]="definitionLinkOpensInPeek",s[s.quickSuggestions=102]="quickSuggestions",s[s.quickSuggestionsDelay=103]="quickSuggestionsDelay",s[s.readOnly=104]="readOnly",s[s.readOnlyMessage=105]="readOnlyMessage",s[s.renameOnType=106]="renameOnType",s[s.renderRichScreenReaderContent=107]="renderRichScreenReaderContent",s[s.renderControlCharacters=108]="renderControlCharacters",s[s.renderFinalNewline=109]="renderFinalNewline",s[s.renderLineHighlight=110]="renderLineHighlight",s[s.renderLineHighlightOnlyWhenFocus=111]="renderLineHighlightOnlyWhenFocus",s[s.renderValidationDecorations=112]="renderValidationDecorations",s[s.renderWhitespace=113]="renderWhitespace",s[s.revealHorizontalRightPadding=114]="revealHorizontalRightPadding",s[s.roundedSelection=115]="roundedSelection",s[s.rulers=116]="rulers",s[s.scrollbar=117]="scrollbar",s[s.scrollBeyondLastColumn=118]="scrollBeyondLastColumn",s[s.scrollBeyondLastLine=119]="scrollBeyondLastLine",s[s.scrollPredominantAxis=120]="scrollPredominantAxis",s[s.selectionClipboard=121]="selectionClipboard",s[s.selectionHighlight=122]="selectionHighlight",s[s.selectionHighlightMaxLength=123]="selectionHighlightMaxLength",s[s.selectionHighlightMultiline=124]="selectionHighlightMultiline",s[s.selectOnLineNumbers=125]="selectOnLineNumbers",s[s.showFoldingControls=126]="showFoldingControls",s[s.showUnused=127]="showUnused",s[s.snippetSuggestions=128]="snippetSuggestions",s[s.smartSelect=129]="smartSelect",s[s.smoothScrolling=130]="smoothScrolling",s[s.stickyScroll=131]="stickyScroll",s[s.stickyTabStops=132]="stickyTabStops",s[s.stopRenderingLineAfter=133]="stopRenderingLineAfter",s[s.suggest=134]="suggest",s[s.suggestFontSize=135]="suggestFontSize",s[s.suggestLineHeight=136]="suggestLineHeight",s[s.suggestOnTriggerCharacters=137]="suggestOnTriggerCharacters",s[s.suggestSelection=138]="suggestSelection",s[s.tabCompletion=139]="tabCompletion",s[s.tabIndex=140]="tabIndex",s[s.trimWhitespaceOnDelete=141]="trimWhitespaceOnDelete",s[s.unicodeHighlighting=142]="unicodeHighlighting",s[s.unusualLineTerminators=143]="unusualLineTerminators",s[s.useShadowDOM=144]="useShadowDOM",s[s.useTabStops=145]="useTabStops",s[s.wordBreak=146]="wordBreak",s[s.wordSegmenterLocales=147]="wordSegmenterLocales",s[s.wordSeparators=148]="wordSeparators",s[s.wordWrap=149]="wordWrap",s[s.wordWrapBreakAfterCharacters=150]="wordWrapBreakAfterCharacters",s[s.wordWrapBreakBeforeCharacters=151]="wordWrapBreakBeforeCharacters",s[s.wordWrapColumn=152]="wordWrapColumn",s[s.wordWrapOverride1=153]="wordWrapOverride1",s[s.wordWrapOverride2=154]="wordWrapOverride2",s[s.wrappingIndent=155]="wrappingIndent",s[s.wrappingStrategy=156]="wrappingStrategy",s[s.showDeprecated=157]="showDeprecated",s[s.inertialScroll=158]="inertialScroll",s[s.inlayHints=159]="inlayHints",s[s.wrapOnEscapedLineFeeds=160]="wrapOnEscapedLineFeeds",s[s.effectiveCursorStyle=161]="effectiveCursorStyle",s[s.editorClassName=162]="editorClassName",s[s.pixelRatio=163]="pixelRatio",s[s.tabFocusMode=164]="tabFocusMode",s[s.layoutInfo=165]="layoutInfo",s[s.wrappingInfo=166]="wrappingInfo",s[s.defaultColorDecorators=167]="defaultColorDecorators",s[s.colorDecoratorsActivatedOn=168]="colorDecoratorsActivatedOn",s[s.inlineCompletionsAccessibilityVerbose=169]="inlineCompletionsAccessibilityVerbose",s[s.effectiveEditContext=170]="effectiveEditContext",s[s.scrollOnMiddleClick=171]="scrollOnMiddleClick",s[s.effectiveAllowVariableFonts=172]="effectiveAllowVariableFonts"})(pB||(pB={}));var mB;(function(s){s[s.TextDefined=0]="TextDefined",s[s.LF=1]="LF",s[s.CRLF=2]="CRLF"})(mB||(mB={}));var _B;(function(s){s[s.LF=0]="LF",s[s.CRLF=1]="CRLF"})(_B||(_B={}));var bB;(function(s){s[s.Left=1]="Left",s[s.Center=2]="Center",s[s.Right=3]="Right"})(bB||(bB={}));var wB;(function(s){s[s.Increase=0]="Increase",s[s.Decrease=1]="Decrease"})(wB||(wB={}));var vB;(function(s){s[s.None=0]="None",s[s.Indent=1]="Indent",s[s.IndentOutdent=2]="IndentOutdent",s[s.Outdent=3]="Outdent"})(vB||(vB={}));var CB;(function(s){s[s.Both=0]="Both",s[s.Right=1]="Right",s[s.Left=2]="Left",s[s.None=3]="None"})(CB||(CB={}));var yB;(function(s){s[s.Type=1]="Type",s[s.Parameter=2]="Parameter"})(yB||(yB={}));var SB;(function(s){s[s.Accepted=0]="Accepted",s[s.Rejected=1]="Rejected",s[s.Ignored=2]="Ignored"})(SB||(SB={}));var xB;(function(s){s[s.Code=1]="Code",s[s.Label=2]="Label"})(xB||(xB={}));var LB;(function(s){s[s.Automatic=0]="Automatic",s[s.Explicit=1]="Explicit"})(LB||(LB={}));var kB;(function(s){s[s.DependsOnKbLayout=-1]="DependsOnKbLayout",s[s.Unknown=0]="Unknown",s[s.Backspace=1]="Backspace",s[s.Tab=2]="Tab",s[s.Enter=3]="Enter",s[s.Shift=4]="Shift",s[s.Ctrl=5]="Ctrl",s[s.Alt=6]="Alt",s[s.PauseBreak=7]="PauseBreak",s[s.CapsLock=8]="CapsLock",s[s.Escape=9]="Escape",s[s.Space=10]="Space",s[s.PageUp=11]="PageUp",s[s.PageDown=12]="PageDown",s[s.End=13]="End",s[s.Home=14]="Home",s[s.LeftArrow=15]="LeftArrow",s[s.UpArrow=16]="UpArrow",s[s.RightArrow=17]="RightArrow",s[s.DownArrow=18]="DownArrow",s[s.Insert=19]="Insert",s[s.Delete=20]="Delete",s[s.Digit0=21]="Digit0",s[s.Digit1=22]="Digit1",s[s.Digit2=23]="Digit2",s[s.Digit3=24]="Digit3",s[s.Digit4=25]="Digit4",s[s.Digit5=26]="Digit5",s[s.Digit6=27]="Digit6",s[s.Digit7=28]="Digit7",s[s.Digit8=29]="Digit8",s[s.Digit9=30]="Digit9",s[s.KeyA=31]="KeyA",s[s.KeyB=32]="KeyB",s[s.KeyC=33]="KeyC",s[s.KeyD=34]="KeyD",s[s.KeyE=35]="KeyE",s[s.KeyF=36]="KeyF",s[s.KeyG=37]="KeyG",s[s.KeyH=38]="KeyH",s[s.KeyI=39]="KeyI",s[s.KeyJ=40]="KeyJ",s[s.KeyK=41]="KeyK",s[s.KeyL=42]="KeyL",s[s.KeyM=43]="KeyM",s[s.KeyN=44]="KeyN",s[s.KeyO=45]="KeyO",s[s.KeyP=46]="KeyP",s[s.KeyQ=47]="KeyQ",s[s.KeyR=48]="KeyR",s[s.KeyS=49]="KeyS",s[s.KeyT=50]="KeyT",s[s.KeyU=51]="KeyU",s[s.KeyV=52]="KeyV",s[s.KeyW=53]="KeyW",s[s.KeyX=54]="KeyX",s[s.KeyY=55]="KeyY",s[s.KeyZ=56]="KeyZ",s[s.Meta=57]="Meta",s[s.ContextMenu=58]="ContextMenu",s[s.F1=59]="F1",s[s.F2=60]="F2",s[s.F3=61]="F3",s[s.F4=62]="F4",s[s.F5=63]="F5",s[s.F6=64]="F6",s[s.F7=65]="F7",s[s.F8=66]="F8",s[s.F9=67]="F9",s[s.F10=68]="F10",s[s.F11=69]="F11",s[s.F12=70]="F12",s[s.F13=71]="F13",s[s.F14=72]="F14",s[s.F15=73]="F15",s[s.F16=74]="F16",s[s.F17=75]="F17",s[s.F18=76]="F18",s[s.F19=77]="F19",s[s.F20=78]="F20",s[s.F21=79]="F21",s[s.F22=80]="F22",s[s.F23=81]="F23",s[s.F24=82]="F24",s[s.NumLock=83]="NumLock",s[s.ScrollLock=84]="ScrollLock",s[s.Semicolon=85]="Semicolon",s[s.Equal=86]="Equal",s[s.Comma=87]="Comma",s[s.Minus=88]="Minus",s[s.Period=89]="Period",s[s.Slash=90]="Slash",s[s.Backquote=91]="Backquote",s[s.BracketLeft=92]="BracketLeft",s[s.Backslash=93]="Backslash",s[s.BracketRight=94]="BracketRight",s[s.Quote=95]="Quote",s[s.OEM_8=96]="OEM_8",s[s.IntlBackslash=97]="IntlBackslash",s[s.Numpad0=98]="Numpad0",s[s.Numpad1=99]="Numpad1",s[s.Numpad2=100]="Numpad2",s[s.Numpad3=101]="Numpad3",s[s.Numpad4=102]="Numpad4",s[s.Numpad5=103]="Numpad5",s[s.Numpad6=104]="Numpad6",s[s.Numpad7=105]="Numpad7",s[s.Numpad8=106]="Numpad8",s[s.Numpad9=107]="Numpad9",s[s.NumpadMultiply=108]="NumpadMultiply",s[s.NumpadAdd=109]="NumpadAdd",s[s.NUMPAD_SEPARATOR=110]="NUMPAD_SEPARATOR",s[s.NumpadSubtract=111]="NumpadSubtract",s[s.NumpadDecimal=112]="NumpadDecimal",s[s.NumpadDivide=113]="NumpadDivide",s[s.KEY_IN_COMPOSITION=114]="KEY_IN_COMPOSITION",s[s.ABNT_C1=115]="ABNT_C1",s[s.ABNT_C2=116]="ABNT_C2",s[s.AudioVolumeMute=117]="AudioVolumeMute",s[s.AudioVolumeUp=118]="AudioVolumeUp",s[s.AudioVolumeDown=119]="AudioVolumeDown",s[s.BrowserSearch=120]="BrowserSearch",s[s.BrowserHome=121]="BrowserHome",s[s.BrowserBack=122]="BrowserBack",s[s.BrowserForward=123]="BrowserForward",s[s.MediaTrackNext=124]="MediaTrackNext",s[s.MediaTrackPrevious=125]="MediaTrackPrevious",s[s.MediaStop=126]="MediaStop",s[s.MediaPlayPause=127]="MediaPlayPause",s[s.LaunchMediaPlayer=128]="LaunchMediaPlayer",s[s.LaunchMail=129]="LaunchMail",s[s.LaunchApp2=130]="LaunchApp2",s[s.Clear=131]="Clear",s[s.MAX_VALUE=132]="MAX_VALUE"})(kB||(kB={}));var EB;(function(s){s[s.Hint=1]="Hint",s[s.Info=2]="Info",s[s.Warning=4]="Warning",s[s.Error=8]="Error"})(EB||(EB={}));var IB;(function(s){s[s.Unnecessary=1]="Unnecessary",s[s.Deprecated=2]="Deprecated"})(IB||(IB={}));var DB;(function(s){s[s.Inline=1]="Inline",s[s.Gutter=2]="Gutter"})(DB||(DB={}));var TB;(function(s){s[s.Normal=1]="Normal",s[s.Underlined=2]="Underlined"})(TB||(TB={}));var NB;(function(s){s[s.UNKNOWN=0]="UNKNOWN",s[s.TEXTAREA=1]="TEXTAREA",s[s.GUTTER_GLYPH_MARGIN=2]="GUTTER_GLYPH_MARGIN",s[s.GUTTER_LINE_NUMBERS=3]="GUTTER_LINE_NUMBERS",s[s.GUTTER_LINE_DECORATIONS=4]="GUTTER_LINE_DECORATIONS",s[s.GUTTER_VIEW_ZONE=5]="GUTTER_VIEW_ZONE",s[s.CONTENT_TEXT=6]="CONTENT_TEXT",s[s.CONTENT_EMPTY=7]="CONTENT_EMPTY",s[s.CONTENT_VIEW_ZONE=8]="CONTENT_VIEW_ZONE",s[s.CONTENT_WIDGET=9]="CONTENT_WIDGET",s[s.OVERVIEW_RULER=10]="OVERVIEW_RULER",s[s.SCROLLBAR=11]="SCROLLBAR",s[s.OVERLAY_WIDGET=12]="OVERLAY_WIDGET",s[s.OUTSIDE_EDITOR=13]="OUTSIDE_EDITOR"})(NB||(NB={}));var RB;(function(s){s[s.AIGenerated=1]="AIGenerated"})(RB||(RB={}));var AB;(function(s){s[s.Invoke=0]="Invoke",s[s.Automatic=1]="Automatic"})(AB||(AB={}));var MB;(function(s){s[s.TOP_RIGHT_CORNER=0]="TOP_RIGHT_CORNER",s[s.BOTTOM_RIGHT_CORNER=1]="BOTTOM_RIGHT_CORNER",s[s.TOP_CENTER=2]="TOP_CENTER"})(MB||(MB={}));var OB;(function(s){s[s.Left=1]="Left",s[s.Center=2]="Center",s[s.Right=4]="Right",s[s.Full=7]="Full"})(OB||(OB={}));var PB;(function(s){s[s.Word=0]="Word",s[s.Line=1]="Line",s[s.Suggest=2]="Suggest"})(PB||(PB={}));var FB;(function(s){s[s.Left=0]="Left",s[s.Right=1]="Right",s[s.None=2]="None",s[s.LeftOfInjectedText=3]="LeftOfInjectedText",s[s.RightOfInjectedText=4]="RightOfInjectedText"})(FB||(FB={}));var BB;(function(s){s[s.Off=0]="Off",s[s.On=1]="On",s[s.Relative=2]="Relative",s[s.Interval=3]="Interval",s[s.Custom=4]="Custom"})(BB||(BB={}));var WB;(function(s){s[s.None=0]="None",s[s.Text=1]="Text",s[s.Blocks=2]="Blocks"})(WB||(WB={}));var HB;(function(s){s[s.Smooth=0]="Smooth",s[s.Immediate=1]="Immediate"})(HB||(HB={}));var VB;(function(s){s[s.Auto=1]="Auto",s[s.Hidden=2]="Hidden",s[s.Visible=3]="Visible"})(VB||(VB={}));var zB;(function(s){s[s.LTR=0]="LTR",s[s.RTL=1]="RTL"})(zB||(zB={}));var $B;(function(s){s.Off="off",s.OnCode="onCode",s.On="on"})($B||($B={}));var UB;(function(s){s[s.Invoke=1]="Invoke",s[s.TriggerCharacter=2]="TriggerCharacter",s[s.ContentChange=3]="ContentChange"})(UB||(UB={}));var qB;(function(s){s[s.File=0]="File",s[s.Module=1]="Module",s[s.Namespace=2]="Namespace",s[s.Package=3]="Package",s[s.Class=4]="Class",s[s.Method=5]="Method",s[s.Property=6]="Property",s[s.Field=7]="Field",s[s.Constructor=8]="Constructor",s[s.Enum=9]="Enum",s[s.Interface=10]="Interface",s[s.Function=11]="Function",s[s.Variable=12]="Variable",s[s.Constant=13]="Constant",s[s.String=14]="String",s[s.Number=15]="Number",s[s.Boolean=16]="Boolean",s[s.Array=17]="Array",s[s.Object=18]="Object",s[s.Key=19]="Key",s[s.Null=20]="Null",s[s.EnumMember=21]="EnumMember",s[s.Struct=22]="Struct",s[s.Event=23]="Event",s[s.Operator=24]="Operator",s[s.TypeParameter=25]="TypeParameter"})(qB||(qB={}));var jB;(function(s){s[s.Deprecated=1]="Deprecated"})(jB||(jB={}));var GB;(function(s){s[s.LTR=0]="LTR",s[s.RTL=1]="RTL"})(GB||(GB={}));var KB;(function(s){s[s.Hidden=0]="Hidden",s[s.Blink=1]="Blink",s[s.Smooth=2]="Smooth",s[s.Phase=3]="Phase",s[s.Expand=4]="Expand",s[s.Solid=5]="Solid"})(KB||(KB={}));var YB;(function(s){s[s.Line=1]="Line",s[s.Block=2]="Block",s[s.Underline=3]="Underline",s[s.LineThin=4]="LineThin",s[s.BlockOutline=5]="BlockOutline",s[s.UnderlineThin=6]="UnderlineThin"})(YB||(YB={}));var ZB;(function(s){s[s.AlwaysGrowsWhenTypingAtEdges=0]="AlwaysGrowsWhenTypingAtEdges",s[s.NeverGrowsWhenTypingAtEdges=1]="NeverGrowsWhenTypingAtEdges",s[s.GrowsOnlyWhenTypingBefore=2]="GrowsOnlyWhenTypingBefore",s[s.GrowsOnlyWhenTypingAfter=3]="GrowsOnlyWhenTypingAfter"})(ZB||(ZB={}));var XB;(function(s){s[s.None=0]="None",s[s.Same=1]="Same",s[s.Indent=2]="Indent",s[s.DeepIndent=3]="DeepIndent"})(XB||(XB={}));var Jd;let oBe=(Jd=class{static chord(e,t){return Gi(e,t)}},Jd.CtrlCmd=2048,Jd.Shift=1024,Jd.Alt=512,Jd.WinCtrl=256,Jd);function vre(){return{editor:void 0,languages:void 0,CancellationTokenSource:si,Emitter:M,KeyCode:kB,KeyMod:oBe,Position:A,Range:k,Selection:he,SelectionDirection:zB,MarkerSeverity:EB,MarkerTag:IB,Uri:Le,Token:pS}}const QK="default",rBe="$initialize";let JK=!1;function QB(s){Vc&&(JK||(JK=!0,console.warn("Could not create web worker(s). Falling back to loading web worker code in main thread, which might cause UI freezes. Please see https://github.com/microsoft/monaco-editor#faq")),console.warn(s.message))}class aBe{constructor(e,t,i,n,o){this.vsWorker=e,this.req=t,this.channel=i,this.method=n,this.args=o,this.type=0}}class eY{constructor(e,t,i,n){this.vsWorker=e,this.seq=t,this.res=i,this.err=n,this.type=1}}class lBe{constructor(e,t,i,n,o){this.vsWorker=e,this.req=t,this.channel=i,this.eventName=n,this.arg=o,this.type=2}}class cBe{constructor(e,t,i){this.vsWorker=e,this.req=t,this.event=i,this.type=3}}class dBe{constructor(e,t){this.vsWorker=e,this.req=t,this.type=4}}class hBe{constructor(e){this._workerId=-1,this._handler=e,this._lastSentReq=0,this._pendingReplies=Object.create(null),this._pendingEmitters=new Map,this._pendingEvents=new Map}setWorkerId(e){this._workerId=e}async sendMessage(e,t,i){const n=String(++this._lastSentReq);return new Promise((o,r)=>{this._pendingReplies[n]={resolve:o,reject:r},this._send(new aBe(this._workerId,n,e,t,i))})}listen(e,t,i){let n=null;const o=new M({onWillAddFirstListener:()=>{n=String(++this._lastSentReq),this._pendingEmitters.set(n,o),this._send(new lBe(this._workerId,n,e,t,i))},onDidRemoveLastListener:()=>{this._pendingEmitters.delete(n),this._send(new dBe(this._workerId,n)),n=null}});return o.event}handleMessage(e){!e||!e.vsWorker||this._workerId!==-1&&e.vsWorker!==this._workerId||this._handleMessage(e)}createProxyToRemoteChannel(e,t){const i={get:(n,o)=>(typeof o=="string"&&!n[o]&&(yre(o)?n[o]=r=>this.listen(e,o,r):Cre(o)?n[o]=this.listen(e,o,void 0):o.charCodeAt(0)===36&&(n[o]=async(...r)=>(await t?.(),this.sendMessage(e,o,r)))),n[o])};return new Proxy(Object.create(null),i)}_handleMessage(e){switch(e.type){case 1:return this._handleReplyMessage(e);case 0:return this._handleRequestMessage(e);case 2:return this._handleSubscribeEventMessage(e);case 3:return this._handleEventMessage(e);case 4:return this._handleUnsubscribeEventMessage(e)}}_handleReplyMessage(e){if(!this._pendingReplies[e.seq]){console.warn("Got reply to unknown seq");return}const t=this._pendingReplies[e.seq];if(delete this._pendingReplies[e.seq],e.err){let i=e.err;if(e.err.$isError){const n=new Error;n.name=e.err.name,n.message=e.err.message,n.stack=e.err.stack,i=n}t.reject(i);return}t.resolve(e.res)}_handleRequestMessage(e){const t=e.req;this._handler.handleMessage(e.channel,e.method,e.args).then(n=>{this._send(new eY(this._workerId,t,n,void 0))},n=>{n.detail instanceof Error&&(n.detail=eP(n.detail)),this._send(new eY(this._workerId,t,void 0,eP(n)))})}_handleSubscribeEventMessage(e){const t=e.req,i=this._handler.handleEvent(e.channel,e.eventName,e.arg)(n=>{this._send(new cBe(this._workerId,t,n))});this._pendingEvents.set(t,i)}_handleEventMessage(e){if(!this._pendingEmitters.has(e.req)){console.warn("Got event for unknown req");return}this._pendingEmitters.get(e.req).fire(e.event)}_handleUnsubscribeEventMessage(e){if(!this._pendingEvents.has(e.req)){console.warn("Got unsubscribe for unknown req");return}this._pendingEvents.get(e.req).dispose(),this._pendingEvents.delete(e.req)}_send(e){const t=[];if(e.type===0)for(let i=0;i<e.args.length;i++){const n=e.args[i];n instanceof ArrayBuffer&&t.push(n)}else e.type===1&&e.res instanceof ArrayBuffer&&t.push(e.res);this._handler.sendMessage(e,t)}}class uBe extends F{constructor(e){super(),this._localChannels=new Map,this._worker=this._register(e),this._register(this._worker.onMessage(t=>{this._protocol.handleMessage(t)})),this._register(this._worker.onError(t=>{QB(t),Oe(t)})),this._protocol=new hBe({sendMessage:(t,i)=>{this._worker.postMessage(t,i)},handleMessage:(t,i,n)=>this._handleMessage(t,i,n),handleEvent:(t,i,n)=>this._handleEvent(t,i,n)}),this._protocol.setWorkerId(this._worker.getId()),this._onModuleLoaded=this._protocol.sendMessage(QK,rBe,[this._worker.getId()]).then(()=>{}),this.proxy=this._protocol.createProxyToRemoteChannel(QK,async()=>{await this._onModuleLoaded}),this._onModuleLoaded.catch(t=>{this._onError("Worker failed to load ",t)})}_handleMessage(e,t,i){const n=this._localChannels.get(e);if(!n)return Promise.reject(new Error(`Missing channel ${e} on main thread`));const o=n[t];if(typeof o!="function")return Promise.reject(new Error(`Missing method ${t} on main thread channel ${e}`));try{return Promise.resolve(o.apply(n,i))}catch(r){return Promise.reject(r)}}_handleEvent(e,t,i){const n=this._localChannels.get(e);if(!n)throw new Error(`Missing channel ${e} on main thread`);if(yre(t)){const o=n[t];if(typeof o!="function")throw new Error(`Missing dynamic event ${t} on main thread channel ${e}.`);const r=o.call(n,i);if(typeof r!="function")throw new Error(`Missing dynamic event ${t} on main thread channel ${e}.`);return r}if(Cre(t)){const o=n[t];if(typeof o!="function")throw new Error(`Missing event ${t} on main thread channel ${e}.`);return o}throw new Error(`Malformed event name ${t}`)}setChannel(e,t){this._localChannels.set(e,t)}_onError(e,t){console.error(e),console.info(t)}}function Cre(s){return s[0]==="o"&&s[1]==="n"&&oc(s.charCodeAt(2))}function yre(s){return/^onDynamic/.test(s)&&oc(s.charCodeAt(9))}let Wb;typeof self=="object"&&self.constructor&&self.constructor.name==="DedicatedWorkerGlobalScope"&&globalThis.workerttPolicy!==void 0?Wb=globalThis.workerttPolicy:Wb=$c("defaultWorkerFactory",{createScriptURL:s=>s});function gBe(s,e){const t=s.label||"anonymous"+e,i=M6();if(i){if(typeof i.getWorker=="function")return i.getWorker("workerMain.js",t);if(typeof i.getWorkerUrl=="function"){const o=i.getWorkerUrl("workerMain.js",t);return new Worker(Wb?Wb.createScriptURL(o):o,{name:t,type:"module"})}}const n=s.esmModuleLocation;if(n){const o=fBe(t,n.toString(!0)),r=new Worker(Wb?Wb.createScriptURL(o):o,{name:t,type:"module"});return pBe(r)}throw new Error("You must define a function MonacoEnvironment.getWorkerUrl or MonacoEnvironment.getWorker")}function fBe(s,e){if(!(/^((http:)|(https:)|(file:))/.test(e)&&e.substring(0,globalThis.origin.length)!==globalThis.origin)){const i=e.lastIndexOf("?"),n=e.lastIndexOf("#",i),o=i>0?new URLSearchParams(e.substring(i+1,~n?n:void 0)):new URLSearchParams;mP.addSearchParam(o,!0,!0),o.toString()?e=`${e}?${o.toString()}#${s}`:e=`${e}#${s}`}const t=new Blob([Al([`/*${s}*/`,`globalThis._VSCODE_NLS_MESSAGES = ${JSON.stringify(MX())};`,`globalThis._VSCODE_NLS_LANGUAGE = ${JSON.stringify(A6())};`,`globalThis._VSCODE_FILE_ROOT = ${JSON.stringify(globalThis._VSCODE_FILE_ROOT)};`,"const ttPolicy = globalThis.trustedTypes?.createPolicy('defaultWorkerFactory', { createScriptURL: value => value });","globalThis.workerttPolicy = ttPolicy;",`await import(ttPolicy?.createScriptURL(${JSON.stringify(e)}) ?? ${JSON.stringify(e)});`,"globalThis.postMessage({ type: 'vscode-worker-ready' });",`/*${s}*/`]).join("")],{type:"application/javascript"});return URL.createObjectURL(t)}function pBe(s){return new Promise((e,t)=>{s.onmessage=function(i){i.data.type==="vscode-worker-ready"&&(s.onmessage=null,e(s))},s.onerror=t})}function mBe(s){return!!s&&typeof s.then=="function"}const cR=class cR extends F{constructor(e){super(),this._onMessage=this._register(new M),this.onMessage=this._onMessage.event,this._onError=this._register(new M),this.onError=this._onError.event,this.id=++cR.LAST_WORKER_ID;const t=e instanceof Worker||"then"in e?e:gBe(e,this.id);mBe(t)?this.worker=t:this.worker=Promise.resolve(t),this.postMessage("-please-ignore-",[]);const i=n=>{this._onError.fire(n)};this.worker.then(n=>{n.onmessage=o=>{this._onMessage.fire(o.data)},n.onmessageerror=o=>{this._onError.fire(o)},typeof n.addEventListener=="function"&&n.addEventListener("error",i)}),this._register(pe(()=>{this.worker?.then(n=>{n.onmessage=null,n.onmessageerror=null,n.removeEventListener("error",i),n.terminate()}),this.worker=null}))}getId(){return this.id}postMessage(e,t){this.worker?.then(i=>{try{i.postMessage(e,t)}catch(n){Oe(n),Oe(new Error("FAILED to post message to worker",{cause:n}))}})}};cR.LAST_WORKER_ID=0;let JB=cR;class _Be{constructor(e,t){this.esmModuleLocation=e,this.label=t}}function bBe(s,e){const t=Le.isUri(s)?new _Be(s,e):s;return new uBe(new JB(t))}class wBe{constructor(e,t,i){const n=new Uint8Array(e*t);for(let o=0,r=e*t;o<r;o++)n[o]=i;this._data=n,this.rows=e,this.cols=t}get(e,t){return this._data[e*this.cols+t]}set(e,t,i){this._data[e*this.cols+t]=i}}class vBe{constructor(e){let t=0,i=0;for(let o=0,r=e.length;o<r;o++){const[a,l,c]=e[o];l>t&&(t=l),a>i&&(i=a),c>i&&(i=c)}t++,i++;const n=new wBe(i,t,0);for(let o=0,r=e.length;o<r;o++){const[a,l,c]=e[o];n.set(a,l,c)}this._states=n,this._maxCharCode=t}nextState(e,t){return t<0||t>=this._maxCharCode?0:this._states.get(e,t)}}let FO=null;function CBe(){return FO===null&&(FO=new vBe([[1,104,2],[1,72,2],[1,102,6],[1,70,6],[2,116,3],[2,84,3],[3,116,4],[3,84,4],[4,112,5],[4,80,5],[5,115,9],[5,83,9],[5,58,10],[6,105,7],[6,73,7],[7,108,8],[7,76,8],[8,101,9],[8,69,9],[9,58,10],[10,47,11],[11,47,12]])),FO}let QC=null;function yBe(){if(QC===null){QC=new Zv(0);const s=` 	<>'"、。｡､，．：；‘〈「『〔（［｛｢｣｝］）〕』」〉’｀～…|`;for(let t=0;t<s.length;t++)QC.set(s.charCodeAt(t),1);const e=".,;:";for(let t=0;t<e.length;t++)QC.set(e.charCodeAt(t),2)}return QC}class hT{static _createLink(e,t,i,n,o){let r=o-1;do{const a=t.charCodeAt(r);if(e.get(a)!==2)break;r--}while(r>n);if(n>0){const a=t.charCodeAt(n-1),l=t.charCodeAt(r);(a===40&&l===41||a===91&&l===93||a===123&&l===125)&&r--}return{range:{startLineNumber:i,startColumn:n+1,endLineNumber:i,endColumn:r+2},url:t.substring(n,r+1)}}static computeLinks(e,t=CBe()){const i=yBe(),n=[];for(let o=1,r=e.getLineCount();o<=r;o++){const a=e.getLineContent(o),l=a.length;let c=0,d=0,h=0,u=1,g=!1,f=!1,p=!1,_=!1;for(;c<l;){let b=!1;const w=a.charCodeAt(c);if(u===13){let C;switch(w){case 40:g=!0,C=0;break;case 41:C=g?0:1;break;case 91:p=!0,f=!0,C=0;break;case 93:p=!1,C=f?0:1;break;case 123:_=!0,C=0;break;case 125:C=_?0:1;break;case 39:case 34:case 96:h===w?C=1:h===39||h===34||h===96?C=0:C=1;break;case 42:C=h===42?1:0;break;case 32:C=p?0:1;break;default:C=i.get(w)}C===1&&(n.push(hT._createLink(i,a,o,d,c)),b=!0)}else if(u===12){let C;w===91?(f=!0,C=0):C=i.get(w),C===1?b=!0:u=13}else u=t.nextState(u,w),u===0&&(b=!0);b&&(u=1,g=!1,f=!1,_=!1,d=c+1,h=w),c++}u===13&&n.push(hT._createLink(i,a,o,d,l))}return n}}function SBe(s){return!s||typeof s.getLineCount!="function"||typeof s.getLineContent!="function"?[]:hT.computeLinks(s)}const dR=class dR{constructor(){this._defaultValueSet=[["true","false"],["True","False"],["Private","Public","Friend","ReadOnly","Partial","Protected","WriteOnly"],["public","protected","private"]]}navigateValueSet(e,t,i,n,o){if(e&&t){const r=this.doNavigateValueSet(t,o);if(r)return{range:e,value:r}}if(i&&n){const r=this.doNavigateValueSet(n,o);if(r)return{range:i,value:r}}return null}doNavigateValueSet(e,t){const i=this.numberReplace(e,t);return i!==null?i:this.textReplace(e,t)}numberReplace(e,t){const i=Math.pow(10,e.length-(e.lastIndexOf(".")+1));let n=Number(e);const o=parseFloat(e);return!isNaN(n)&&!isNaN(o)&&n===o?n===0&&!t?null:(n=Math.floor(n*i),n+=t?i:-i,String(n/i)):null}textReplace(e,t){return this.valueSetsReplace(this._defaultValueSet,e,t)}valueSetsReplace(e,t,i){let n=null;for(let o=0,r=e.length;n===null&&o<r;o++)n=this.valueSetReplace(e[o],t,i);return n}valueSetReplace(e,t,i){let n=e.indexOf(t);return n>=0?(n+=i?1:-1,n<0?n=e.length-1:n%=e.length,e[n]):null}};dR.INSTANCE=new dR;let e9=dR;function Sre(s){const e=[];for(const t of s){const i=Number(t);(i||i===0&&t.replace(/\s/g,"")!=="")&&e.push(i)}return e}function _V(s,e,t,i){return{red:s/255,blue:t/255,green:e/255,alpha:i}}function JC(s,e){const t=e.index,i=e[0].length;if(t===void 0)return;const n=s.positionAt(t);return{startLineNumber:n.lineNumber,startColumn:n.column,endLineNumber:n.lineNumber,endColumn:n.column+i}}function xBe(s,e){if(!s)return;const t=q.Format.CSS.parseHex(e);if(t)return{range:s,color:_V(t.rgba.r,t.rgba.g,t.rgba.b,t.rgba.a)}}function tY(s,e,t){if(!s||e.length!==1)return;const n=e[0].values(),o=Sre(n);return{range:s,color:_V(o[0],o[1],o[2],t?o[3]:1)}}function iY(s,e,t){if(!s||e.length!==1)return;const n=e[0].values(),o=Sre(n),r=new q(new Ea(o[0],o[1]/100,o[2]/100,t?o[3]:1));return{range:s,color:_V(r.rgba.r,r.rgba.g,r.rgba.b,r.rgba.a)}}function e1(s,e){return typeof s=="string"?[...s.matchAll(e)]:s.findMatches(e)}function LBe(s){const e=[],t=new RegExp(`\\b(rgb|rgba|hsl|hsla)(\\([0-9\\s,.\\%]*\\))|^(#)([A-Fa-f0-9]{3})\\b|^(#)([A-Fa-f0-9]{4})\\b|^(#)([A-Fa-f0-9]{6})\\b|^(#)([A-Fa-f0-9]{8})\\b|(?<=['"\\s])(#)([A-Fa-f0-9]{3})\\b|(?<=['"\\s])(#)([A-Fa-f0-9]{4})\\b|(?<=['"\\s])(#)([A-Fa-f0-9]{6})\\b|(?<=['"\\s])(#)([A-Fa-f0-9]{8})\\b`,"gm"),i=e1(s,t);if(i.length>0)for(const n of i){const o=n.filter(c=>c!==void 0),r=o[1],a=o[2];if(!a)continue;let l;if(r==="rgb"){const c=/^\(\s*(25[0-5]|2[0-4][0-9]|1[0-9]{2}|[1-9][0-9]|[0-9])\s*,\s*(25[0-5]|2[0-4][0-9]|1[0-9]{2}|[1-9][0-9]|[0-9])\s*,\s*(25[0-5]|2[0-4][0-9]|1[0-9]{2}|[1-9][0-9]|[0-9])\s*\)$/gm;l=tY(JC(s,n),e1(a,c),!1)}else if(r==="rgba"){const c=/^\(\s*(25[0-5]|2[0-4][0-9]|1[0-9]{2}|[1-9][0-9]|[0-9])\s*,\s*(25[0-5]|2[0-4][0-9]|1[0-9]{2}|[1-9][0-9]|[0-9])\s*,\s*(25[0-5]|2[0-4][0-9]|1[0-9]{2}|[1-9][0-9]|[0-9])\s*,\s*(0[.][0-9]+|[.][0-9]+|[01][.]|[01])\s*\)$/gm;l=tY(JC(s,n),e1(a,c),!0)}else if(r==="hsl"){const c=/^\(\s*((?:360(?:\.0+)?|(?:36[0]|3[0-5][0-9]|[12][0-9][0-9]|[1-9]?[0-9])(?:\.\d+)?))\s*[\s,]\s*(100|\d{1,2}[.]\d*|\d{1,2})%\s*[\s,]\s*(100|\d{1,2}[.]\d*|\d{1,2})%\s*\)$/gm;l=iY(JC(s,n),e1(a,c),!1)}else if(r==="hsla"){const c=/^\(\s*((?:360(?:\.0+)?|(?:36[0]|3[0-5][0-9]|[12][0-9][0-9]|[1-9]?[0-9])(?:\.\d+)?))\s*[\s,]\s*(100|\d{1,2}[.]\d*|\d{1,2})%\s*[\s,]\s*(100|\d{1,2}[.]\d*|\d{1,2})%\s*[\s,]\s*(0[.][0-9]+|[.][0-9]+|[01][.]0*|[01])\s*\)$/gm;l=iY(JC(s,n),e1(a,c),!0)}else r==="#"&&(l=xBe(JC(s,n),r+a));l&&e.push(l)}return e}function kBe(s){return!s||typeof s.getValue!="function"||typeof s.positionAt!="function"?[]:LBe(s)}const EBe=/^-+|-+$/g,nY=100,IBe=5;function DBe(s,e){let t=[];if(e.findRegionSectionHeaders&&e.foldingRules?.markers){const i=TBe(s,e);t=t.concat(i)}if(e.findMarkSectionHeaders){const i=NBe(s,e);t=t.concat(i)}return t}function TBe(s,e){const t=[],i=s.getLineCount();for(let n=1;n<=i;n++){const o=s.getLineContent(n),r=o.match(e.foldingRules.markers.start);if(r){const a={startLineNumber:n,startColumn:r[0].length+1,endLineNumber:n,endColumn:o.length+1};if(a.endColumn>a.startColumn){const l={range:a,...RBe(o.substring(r[0].length)),shouldBeInComments:!1};(l.text||l.hasSeparatorLine)&&t.push(l)}}}return t}function NBe(s,e){const t=[],i=s.getLineCount();if(!e.markSectionHeaderRegex||e.markSectionHeaderRegex.trim()==="")return t;const n=ote(e.markSectionHeaderRegex),o=new RegExp(e.markSectionHeaderRegex,`gdm${n?"s":""}`);if(eQ(o))return t;for(let r=1;r<=i;r+=nY-IBe){const a=Math.min(r+nY-1,i),l=[];for(let h=r;h<=a;h++)l.push(s.getLineContent(h));const c=l.join(`
`);o.lastIndex=0;let d;for(;(d=o.exec(c))!==null;){const h=c.substring(0,d.index),u=(h.match(/\n/g)||[]).length,g=r+u,f=d[0].split(`
`),p=f.length,_=g+p-1,b=h.lastIndexOf(`
`)+1,w=d.index-b+1,C=f[f.length-1],v=p===1?w+d[0].length:C.length+1,y={startLineNumber:g,startColumn:w,endLineNumber:_,endColumn:v},x=(d.groups??{}).label??"",L=((d.groups??{}).separator??"")!=="",E={range:y,text:x,hasSeparatorLine:L,shouldBeInComments:!0};(E.text||E.hasSeparatorLine)&&(t.length===0||t[t.length-1].range.endLineNumber<E.range.startLineNumber)&&t.push(E),o.lastIndex=d.index+d[0].length}}return t}function RBe(s){s=s.trim();const e=s.startsWith("-");return s=s.replace(EBe,""),{text:s,hasSeparatorLine:e}}class ABe{constructor(e,t,i,n){this._uri=e,this._lines=t,this._eol=i,this._versionId=n,this._lineStarts=null,this._cachedTextValue=null}dispose(){this._lines.length=0}get version(){return this._versionId}getText(){return this._cachedTextValue===null&&(this._cachedTextValue=this._lines.join(this._eol)),this._cachedTextValue}onEvents(e){e.eol&&e.eol!==this._eol&&(this._eol=e.eol,this._lineStarts=null);const t=e.changes;for(const i of t)this._acceptDeleteRange(i.range),this._acceptInsertText(new A(i.range.startLineNumber,i.range.startColumn),i.text);this._versionId=e.versionId,this._cachedTextValue=null}_ensureLineStarts(){if(!this._lineStarts){const e=this._eol.length,t=this._lines.length,i=new Uint32Array(t);for(let n=0;n<t;n++)i[n]=this._lines[n].length+e;this._lineStarts=new CSe(i)}}_setLineText(e,t){this._lines[e]=t,this._lineStarts&&this._lineStarts.setValue(e,this._lines[e].length+this._eol.length)}_acceptDeleteRange(e){if(e.startLineNumber===e.endLineNumber){if(e.startColumn===e.endColumn)return;this._setLineText(e.startLineNumber-1,this._lines[e.startLineNumber-1].substring(0,e.startColumn-1)+this._lines[e.startLineNumber-1].substring(e.endColumn-1));return}this._setLineText(e.startLineNumber-1,this._lines[e.startLineNumber-1].substring(0,e.startColumn-1)+this._lines[e.endLineNumber-1].substring(e.endColumn-1)),this._lines.splice(e.startLineNumber,e.endLineNumber-e.startLineNumber),this._lineStarts&&this._lineStarts.removeValues(e.startLineNumber,e.endLineNumber-e.startLineNumber)}_acceptInsertText(e,t){if(t.length===0)return;const i=qo(t);if(i.length===1){this._setLineText(e.lineNumber-1,this._lines[e.lineNumber-1].substring(0,e.column-1)+i[0]+this._lines[e.lineNumber-1].substring(e.column-1));return}i[i.length-1]+=this._lines[e.lineNumber-1].substring(e.column-1),this._setLineText(e.lineNumber-1,this._lines[e.lineNumber-1].substring(0,e.column-1)+i[0]);const n=new Uint32Array(i.length-1);for(let o=1;o<i.length;o++)this._lines.splice(e.lineNumber+o-1,0,i[o]),n[o-1]=i[o].length+this._eol.length;this._lineStarts&&this._lineStarts.insertValues(e.lineNumber,n)}}const sY=60*1e3;class MBe extends F{constructor(e,t,i=!1){if(super(),this._syncedModels=Object.create(null),this._syncedModelsLastUsedTime=Object.create(null),this._proxy=e,this._modelService=t,!i){const n=new $6;n.cancelAndSet(()=>this._checkStopModelSync(),Math.round(sY/2)),this._register(n)}}dispose(){for(const e in this._syncedModels)xt(this._syncedModels[e]);this._syncedModels=Object.create(null),this._syncedModelsLastUsedTime=Object.create(null),super.dispose()}ensureSyncedResources(e,t=!1){for(const i of e){const n=i.toString();this._syncedModels[n]||this._beginModelSync(i,t),this._syncedModels[n]&&(this._syncedModelsLastUsedTime[n]=new Date().getTime())}}_checkStopModelSync(){const e=new Date().getTime(),t=[];for(const i in this._syncedModelsLastUsedTime)e-this._syncedModelsLastUsedTime[i]>sY&&t.push(i);for(const i of t)this._stopModelSync(i)}_beginModelSync(e,t){const i=this._modelService.getModel(e);if(!i||!t&&i.isTooLargeForSyncing())return;const n=e.toString();this._proxy.$acceptNewModel({url:i.uri.toString(),lines:i.getLinesContent(),EOL:i.getEOL(),versionId:i.getVersionId()});const o=new U;o.add(i.onDidChangeContent(r=>{this._proxy.$acceptModelChanged(n.toString(),r)})),o.add(i.onWillDispose(()=>{this._stopModelSync(n)})),o.add(pe(()=>{this._proxy.$acceptRemovedModel(n)})),this._syncedModels[n]=o}_stopModelSync(e){const t=this._syncedModels[e];delete this._syncedModels[e],delete this._syncedModelsLastUsedTime[e],xt(t)}}class OBe{constructor(){this._models=Object.create(null)}getModel(e){return this._models[e]}getModels(){const e=[];return Object.keys(this._models).forEach(t=>e.push(this._models[t])),e}$acceptNewModel(e){this._models[e.url]=new PBe(Le.parse(e.url),e.lines,e.EOL,e.versionId)}$acceptModelChanged(e,t){if(!this._models[e])return;this._models[e].onEvents(t)}$acceptRemovedModel(e){this._models[e]&&delete this._models[e]}}class PBe extends ABe{get uri(){return this._uri}get eol(){return this._eol}getValue(){return this.getText()}findMatches(e){const t=[];for(let i=0;i<this._lines.length;i++){const n=this._lines[i],o=this.offsetAt(new A(i+1,1)),r=n.matchAll(e);for(const a of r)(a.index||a.index===0)&&(a.index=a.index+o),t.push(a)}return t}getLinesContent(){return this._lines.slice(0)}getLineCount(){return this._lines.length}getLineContent(e){return this._lines[e-1]}getWordAtPosition(e,t){const i=iS(e.column,_W(t),this._lines[e.lineNumber-1],0);return i?new k(e.lineNumber,i.startColumn,e.lineNumber,i.endColumn):null}words(e){const t=this._lines,i=this._wordenize.bind(this);let n=0,o="",r=0,a=[];return{*[Symbol.iterator](){for(;;)if(r<a.length){const l=o.substring(a[r].start,a[r].end);r+=1,yield l}else if(n<t.length)o=t[n],a=i(o,e),r=0,n+=1;else break}}}getLineWords(e,t){const i=this._lines[e-1],n=this._wordenize(i,t),o=[];for(const r of n)o.push({word:i.substring(r.start,r.end),startColumn:r.start+1,endColumn:r.end+1});return o}_wordenize(e,t){const i=[];let n;for(t.lastIndex=0;(n=t.exec(e))&&n[0].length!==0;)i.push({start:n.index,end:n.index+n[0].length});return i}getValueInRange(e){if(e=this._validateRange(e),e.startLineNumber===e.endLineNumber)return this._lines[e.startLineNumber-1].substring(e.startColumn-1,e.endColumn-1);const t=this._eol,i=e.startLineNumber-1,n=e.endLineNumber-1,o=[];o.push(this._lines[i].substring(e.startColumn-1));for(let r=i+1;r<n;r++)o.push(this._lines[r]);return o.push(this._lines[n].substring(0,e.endColumn-1)),o.join(t)}offsetAt(e){return e=this._validatePosition(e),this._ensureLineStarts(),this._lineStarts.getPrefixSum(e.lineNumber-2)+(e.column-1)}positionAt(e){e=Math.floor(e),e=Math.max(0,e),this._ensureLineStarts();const t=this._lineStarts.getIndexOf(e),i=this._lines[t.index].length;return{lineNumber:1+t.index,column:1+Math.min(t.remainder,i)}}_validateRange(e){const t=this._validatePosition({lineNumber:e.startLineNumber,column:e.startColumn}),i=this._validatePosition({lineNumber:e.endLineNumber,column:e.endColumn});return t.lineNumber!==e.startLineNumber||t.column!==e.startColumn||i.lineNumber!==e.endLineNumber||i.column!==e.endColumn?{startLineNumber:t.lineNumber,startColumn:t.column,endLineNumber:i.lineNumber,endColumn:i.column}:e}_validatePosition(e){if(!A.isIPosition(e))throw new Error("bad position");let{lineNumber:t,column:i}=e,n=!1;if(t<1)t=1,i=1,n=!0;else if(t>this._lines.length)t=this._lines.length,i=this._lines[t-1].length+1,n=!0;else{const o=this._lines[t-1].length+1;i<1?(i=1,n=!0):i>o&&(i=o,n=!0)}return n?{lineNumber:t,column:i}:e}}const Bf=class Bf{constructor(e=null){this._foreignModule=e,this._requestHandlerBrand=void 0,this._workerTextModelSyncServer=new OBe}dispose(){}async $ping(){return"pong"}_getModel(e){return this._workerTextModelSyncServer.getModel(e)}getModels(){return this._workerTextModelSyncServer.getModels()}$acceptNewModel(e){this._workerTextModelSyncServer.$acceptNewModel(e)}$acceptModelChanged(e,t){this._workerTextModelSyncServer.$acceptModelChanged(e,t)}$acceptRemovedModel(e){this._workerTextModelSyncServer.$acceptRemovedModel(e)}async $computeUnicodeHighlights(e,t,i){const n=this._getModel(e);return n?lV.computeUnicodeHighlights(n,t,i):{ranges:[],hasMore:!1,ambiguousCharacterCount:0,invisibleCharacterCount:0,nonBasicAsciiCharacterCount:0}}async $findSectionHeaders(e,t){const i=this._getModel(e);return i?DBe(i,t):[]}async $computeDiff(e,t,i,n){const o=this._getModel(e),r=this._getModel(t);return!o||!r?null:Bf.computeDiff(o,r,i,n)}static computeDiff(e,t,i,n){const o=n==="advanced"?m3.getDefault():m3.getLegacy(),r=e.getLinesContent(),a=t.getLinesContent(),l=o.computeDiff(r,a,i),c=l.changes.length>0?!1:this._modelsAreIdentical(e,t);function d(h){return h.map(u=>[u.original.startLineNumber,u.original.endLineNumberExclusive,u.modified.startLineNumber,u.modified.endLineNumberExclusive,u.innerChanges?.map(g=>[g.originalRange.startLineNumber,g.originalRange.startColumn,g.originalRange.endLineNumber,g.originalRange.endColumn,g.modifiedRange.startLineNumber,g.modifiedRange.startColumn,g.modifiedRange.endLineNumber,g.modifiedRange.endColumn])])}return{identical:c,quitEarly:l.hitTimeout,changes:d(l.changes),moves:l.moves.map(h=>[h.lineRangeMapping.original.startLineNumber,h.lineRangeMapping.original.endLineNumberExclusive,h.lineRangeMapping.modified.startLineNumber,h.lineRangeMapping.modified.endLineNumberExclusive,d(h.changes)])}}static _modelsAreIdentical(e,t){const i=e.getLineCount(),n=t.getLineCount();if(i!==n)return!1;for(let o=1;o<=i;o++){const r=e.getLineContent(o),a=t.getLineContent(o);if(r!==a)return!1}return!0}async $computeMoreMinimalEdits(e,t,i){const n=this._getModel(e);if(!n)return t;const o=[];let r;t=t.slice(0).sort((l,c)=>{if(l.range&&c.range)return k.compareRangesUsingStarts(l.range,c.range);const d=l.range?0:1,h=c.range?0:1;return d-h});let a=0;for(let l=1;l<t.length;l++)k.getEndPosition(t[a].range).equals(k.getStartPosition(t[l].range))?(t[a].range=k.fromPositions(k.getStartPosition(t[a].range),k.getEndPosition(t[l].range)),t[a].text+=t[l].text):(a++,t[a]=t[l]);t.length=a+1;for(let{range:l,text:c,eol:d}of t){if(typeof d=="number"&&(r=d),k.isEmpty(l)&&!c)continue;const h=n.getValueInRange(l);if(c=c.replace(/\r\n|\n|\r/g,n.eol),h===c)continue;if(Math.max(c.length,h.length)>Bf._diffLimit){o.push({range:l,text:c});continue}const u=bNe(h,c,i),g=n.offsetAt(k.lift(l).getStartPosition());for(const f of u){const p=n.positionAt(g+f.originalStart),_=n.positionAt(g+f.originalStart+f.originalLength),b={text:c.substr(f.modifiedStart,f.modifiedLength),range:{startLineNumber:p.lineNumber,startColumn:p.column,endLineNumber:_.lineNumber,endColumn:_.column}};n.getValueInRange(b.range)!==b.text&&o.push(b)}}return typeof r=="number"&&o.push({eol:r,text:"",range:{startLineNumber:0,startColumn:0,endLineNumber:0,endColumn:0}}),o}async $computeLinks(e){const t=this._getModel(e);return t?SBe(t):null}async $computeDefaultDocumentColors(e){const t=this._getModel(e);return t?kBe(t):null}async $textualSuggest(e,t,i,n){const o=new Cn,r=new RegExp(i,n),a=new Set;e:for(const l of e){const c=this._getModel(l);if(c){for(const d of c.words(r))if(!(d===t||!isNaN(Number(d)))&&(a.add(d),a.size>Bf._suggestionsLimit))break e}}return{words:Array.from(a),duration:o.elapsed()}}async $computeWordRanges(e,t,i,n){const o=this._getModel(e);if(!o)return Object.create(null);const r=new RegExp(i,n),a=Object.create(null);for(let l=t.startLineNumber;l<t.endLineNumber;l++){const c=o.getLineWords(l,r);for(const d of c){if(!isNaN(Number(d.word)))continue;let h=a[d.word];h||(h=[],a[d.word]=h),h.push({startLineNumber:l,startColumn:d.startColumn,endLineNumber:l,endColumn:d.endColumn})}}return a}async $navigateValueSet(e,t,i,n,o){const r=this._getModel(e);if(!r)return null;const a=new RegExp(n,o);t.startColumn===t.endColumn&&(t={startLineNumber:t.startLineNumber,startColumn:t.startColumn,endLineNumber:t.endLineNumber,endColumn:t.endColumn+1});const l=r.getValueInRange(t),c=r.getWordAtPosition({lineNumber:t.startLineNumber,column:t.startColumn},a);if(!c)return null;const d=r.getValueInRange(c);return e9.INSTANCE.navigateValueSet(t,l,c,d,i)}$fmr(e,t){if(!this._foreignModule||typeof this._foreignModule[e]!="function")return Promise.reject(new Error("Missing requestHandler or method: "+e));try{return Promise.resolve(this._foreignModule[e].apply(this._foreignModule,t))}catch(i){return Promise.reject(i)}}};Bf._diffLimit=1e5,Bf._suggestionsLimit=1e4;let t9=Bf;typeof importScripts=="function"&&(globalThis.monaco=vre());const My=class My{static getChannel(e){return e.getChannel(My.CHANNEL_NAME)}static setChannel(e,t){e.setChannel(My.CHANNEL_NAME,t)}};My.CHANNEL_NAME="editorWorkerHost";let i9=My;var bV=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},Sf=function(s,e){return function(t,i){e(t,i,s)}};const oY=300*1e3;function xf(s,e){const t=s.getModel(e);return!(!t||t.isTooLargeForSyncing())}let n9=class extends F{constructor(e,t,i,n,o,r){super(),this._languageConfigurationService=o,this._modelService=t,this._workerManager=this._register(new s9(e,this._modelService)),this._logService=n,this._register(r.linkProvider.register({language:"*",hasAccessToAllModels:!0},{provideLinks:async(a,l)=>{if(!xf(this._modelService,a.uri))return Promise.resolve({links:[]});const d=await(await this._workerWithResources([a.uri])).$computeLinks(a.uri.toString());return d&&{links:d}}})),this._register(r.completionProvider.register("*",new FBe(this._workerManager,i,this._modelService,this._languageConfigurationService,this._logService)))}dispose(){super.dispose()}canComputeUnicodeHighlights(e){return xf(this._modelService,e)}async computedUnicodeHighlights(e,t,i){return(await this._workerWithResources([e])).$computeUnicodeHighlights(e.toString(),t,i)}async computeDiff(e,t,i,n){const r=await(await this._workerWithResources([e,t],!0)).$computeDiff(e.toString(),t.toString(),i,n);if(!r)return null;return{identical:r.identical,quitEarly:r.quitEarly,changes:l(r.changes),moves:r.moves.map(c=>new doe(new Ts(new Ne(c[0],c[1]),new Ne(c[2],c[3])),l(c[4])))};function l(c){return c.map(d=>new Lr(new Ne(d[0],d[1]),new Ne(d[2],d[3]),d[4]?.map(h=>new Ks(new k(h[0],h[1],h[2],h[3]),new k(h[4],h[5],h[6],h[7])))))}}async computeMoreMinimalEdits(e,t,i=!1){if(Ms(t)){if(!xf(this._modelService,e))return Promise.resolve(t);const n=Cn.create(),o=this._workerWithResources([e]).then(r=>r.$computeMoreMinimalEdits(e.toString(),t,i));return o.finally(()=>this._logService.trace("FORMAT#computeMoreMinimalEdits",e.toString(!0),n.elapsed())),Promise.race([o,Tc(1e3).then(()=>t)])}else return Promise.resolve(void 0)}canNavigateValueSet(e){return xf(this._modelService,e)}async navigateValueSet(e,t,i){const n=this._modelService.getModel(e);if(!n)return null;const o=this._languageConfigurationService.getLanguageConfiguration(n.getLanguageId()).getWordDefinition(),r=o.source,a=o.flags;return(await this._workerWithResources([e])).$navigateValueSet(e.toString(),t,i,r,a)}canComputeWordRanges(e){return xf(this._modelService,e)}async computeWordRanges(e,t){const i=this._modelService.getModel(e);if(!i)return Promise.resolve(null);const n=this._languageConfigurationService.getLanguageConfiguration(i.getLanguageId()).getWordDefinition(),o=n.source,r=n.flags;return(await this._workerWithResources([e])).$computeWordRanges(e.toString(),t,o,r)}async findSectionHeaders(e,t){return(await this._workerWithResources([e])).$findSectionHeaders(e.toString(),t)}async computeDefaultDocumentColors(e){return(await this._workerWithResources([e])).$computeDefaultDocumentColors(e.toString())}async _workerWithResources(e,t=!1){return await(await this._workerManager.withWorker()).workerWithSyncedResources(e,t)}};n9=bV([Sf(1,hi),Sf(2,QA),Sf(3,jt),Sf(4,ui),Sf(5,ge)],n9);class FBe{constructor(e,t,i,n,o){this.languageConfigurationService=n,this.logService=o,this._debugDisplayName="wordbasedCompletions",this._workerManager=e,this._configurationService=t,this._modelService=i}async provideCompletionItems(e,t){const i=this._configurationService.getValue(e.uri,t,"editor");if(i.wordBasedSuggestions==="off")return;const n=[];if(i.wordBasedSuggestions==="currentDocument")xf(this._modelService,e.uri)&&n.push(e.uri);else for(const h of this._modelService.getModels())xf(this._modelService,h.uri)&&(h===e?n.unshift(h.uri):(i.wordBasedSuggestions==="allDocuments"||h.getLanguageId()===e.getLanguageId())&&n.push(h.uri));if(n.length===0)return;const o=this.languageConfigurationService.getLanguageConfiguration(e.getLanguageId()).getWordDefinition(),r=e.getWordAtPosition(t),a=r?new k(t.lineNumber,r.startColumn,t.lineNumber,r.endColumn):k.fromPositions(t),l=a.setEndPosition(t.lineNumber,t.column);this.logService.trace("[WordBasedCompletionItemProvider]",`word: "${r?.word||""}", wordDef: "${o}", replace: [${a.toString()}], insert: [${l.toString()}]`);const d=await(await this._workerManager.withWorker()).textualSuggest(n,r?.word,o);if(d)return{duration:d.duration,suggestions:d.words.map(h=>({kind:18,label:h,insertText:h,range:{insert:l,replace:a}}))}}}let s9=class extends F{constructor(e,t){super(),this._workerDescriptor=e,this._modelService=t,this._editorWorkerClient=null,this._lastWorkerUsedTime=new Date().getTime(),this._register(new cW).cancelAndSet(()=>this._checkStopIdleWorker(),Math.round(oY/2),Nt),this._register(this._modelService.onModelRemoved(n=>this._checkStopEmptyWorker()))}dispose(){this._editorWorkerClient&&(this._editorWorkerClient.dispose(),this._editorWorkerClient=null),super.dispose()}_checkStopEmptyWorker(){if(!this._editorWorkerClient)return;this._modelService.getModels().length===0&&(this._editorWorkerClient.dispose(),this._editorWorkerClient=null)}_checkStopIdleWorker(){if(!this._editorWorkerClient)return;new Date().getTime()-this._lastWorkerUsedTime>oY&&(this._editorWorkerClient.dispose(),this._editorWorkerClient=null)}withWorker(){return this._lastWorkerUsedTime=new Date().getTime(),this._editorWorkerClient||(this._editorWorkerClient=new uT(this._workerDescriptor,!1,this._modelService)),Promise.resolve(this._editorWorkerClient)}};s9=bV([Sf(1,hi)],s9);class BBe{constructor(e){this._instance=e,this.proxy=this._instance}dispose(){this._instance.dispose()}setChannel(e,t){throw new Error("Not supported")}}let uT=class extends F{constructor(e,t,i){super(),this._workerDescriptorOrWorker=e,this._disposed=!1,this._modelService=i,this._keepIdleModels=t,this._worker=null,this._modelManager=null}fhr(e,t){throw new Error("Not implemented!")}_getOrCreateWorker(){if(!this._worker)try{this._worker=this._register(bBe(this._workerDescriptorOrWorker)),i9.setChannel(this._worker,this._createEditorWorkerHost())}catch(e){QB(e),this._worker=this._createFallbackLocalWorker()}return this._worker}async _getProxy(){try{const e=this._getOrCreateWorker().proxy;return await e.$ping(),e}catch(e){return QB(e),this._worker=this._createFallbackLocalWorker(),this._worker.proxy}}_createFallbackLocalWorker(){return new BBe(new t9(null))}_createEditorWorkerHost(){return{$fhr:(e,t)=>this.fhr(e,t)}}_getOrCreateModelManager(e){return this._modelManager||(this._modelManager=this._register(new MBe(e,this._modelService,this._keepIdleModels))),this._modelManager}async workerWithSyncedResources(e,t=!1){if(this._disposed)return Promise.reject(Wce());const i=await this._getProxy();return this._getOrCreateModelManager(i).ensureSyncedResources(e,t),i}async textualSuggest(e,t,i){const n=await this.workerWithSyncedResources(e),o=i.source,r=i.flags;return n.$textualSuggest(e.map(a=>a.toString()),t,o,r)}dispose(){super.dispose(),this._disposed=!0}};uT=bV([Sf(2,hi)],uT);function WBe(s,e){return new HBe(s,e)}class HBe extends uT{constructor(e,t){super(t.worker,t.keepIdleModels||!1,e),this._foreignModuleHost=t.host||null,this._foreignProxy=this._getProxy().then(i=>new Proxy({},{get(n,o,r){if(o!=="then"){if(typeof o!="string")throw new Error("Not supported");return(...a)=>i.$fmr(o,a)}}}))}fhr(e,t){if(!this._foreignModuleHost||typeof this._foreignModuleHost[e]!="function")return Promise.reject(new Error("Missing method "+e+" or missing main thread foreign host."));try{return Promise.resolve(this._foreignModuleHost[e].apply(this._foreignModuleHost,t))}catch(i){return Promise.reject(i)}}getProxy(){return this._foreignProxy}withSyncedResources(e){return this.workerWithSyncedResources(e).then(t=>this.getProxy())}}function VBe(s){return Array.isArray(s)}function zBe(s){return!VBe(s)}function xre(s){return typeof s=="string"}function rY(s){return!xre(s)}function Lf(s){return!s}function lh(s,e){return s.ignoreCase&&e?e.toLowerCase():e}function aY(s){return s.replace(/[&<>'"_]/g,"-")}function $Be(s,e){console.log(`${s.languageId}: ${e}`)}function ei(s,e){return new Error(`${s.languageId}: ${e}`)}function hu(s,e,t,i,n){const o=/\$((\$)|(#)|(\d\d?)|[sS](\d\d?)|@(\w+))/g;let r=null;return e.replace(o,function(a,l,c,d,h,u,g,f,p){return Lf(c)?Lf(d)?!Lf(h)&&h<i.length?lh(s,i[h]):!Lf(g)&&s&&typeof s[g]=="string"?s[g]:(r===null&&(r=n.split("."),r.unshift(n)),!Lf(u)&&u<r.length?lh(s,r[u]):""):lh(s,t):"$"})}function UBe(s,e,t){const i=/\$[sS](\d\d?)/g;let n=null;return e.replace(i,function(o,r){return n===null&&(n=t.split("."),n.unshift(t)),!Lf(r)&&r<n.length?Uo(lh(s,n[r])):""})}function nk(s,e){let t=e;for(;t&&t.length>0;){const i=s.tokenizer[t];if(i)return i;const n=t.lastIndexOf(".");n<0?t=null:t=t.substr(0,n)}return null}function qBe(s,e){let t=e;for(;t&&t.length>0;){if(s.stateNames[t])return!0;const n=t.lastIndexOf(".");n<0?t=null:t=t.substr(0,n)}return!1}var jBe=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},GBe=function(s,e){return function(t,i){e(t,i,s)}},o9;const Lre=5,hR=class hR{static create(e,t){return this._INSTANCE.create(e,t)}constructor(e){this._maxCacheDepth=e,this._entries=Object.create(null)}create(e,t){if(e!==null&&e.depth>=this._maxCacheDepth)return new Nw(e,t);let i=Nw.getStackElementId(e);i.length>0&&(i+="|"),i+=t;let n=this._entries[i];return n||(n=new Nw(e,t),this._entries[i]=n,n)}};hR._INSTANCE=new hR(Lre);let S0=hR;class Nw{constructor(e,t){this.parent=e,this.state=t,this.depth=(this.parent?this.parent.depth:0)+1}static getStackElementId(e){let t="";for(;e!==null;)t.length>0&&(t+="|"),t+=e.state,e=e.parent;return t}static _equals(e,t){for(;e!==null&&t!==null;){if(e===t)return!0;if(e.state!==t.state)return!1;e=e.parent,t=t.parent}return e===null&&t===null}equals(e){return Nw._equals(this,e)}push(e){return S0.create(this,e)}pop(){return this.parent}popall(){let e=this;for(;e.parent;)e=e.parent;return e}switchTo(e){return S0.create(this.parent,e)}}class Hb{constructor(e,t){this.languageId=e,this.state=t}equals(e){return this.languageId===e.languageId&&this.state.equals(e.state)}clone(){return this.state.clone()===this.state?this:new Hb(this.languageId,this.state)}}const uR=class uR{static create(e,t){return this._INSTANCE.create(e,t)}constructor(e){this._maxCacheDepth=e,this._entries=Object.create(null)}create(e,t){if(t!==null)return new gy(e,t);if(e!==null&&e.depth>=this._maxCacheDepth)return new gy(e,t);const i=Nw.getStackElementId(e);let n=this._entries[i];return n||(n=new gy(e,null),this._entries[i]=n,n)}};uR._INSTANCE=new uR(Lre);let uu=uR;class gy{constructor(e,t){this.stack=e,this.embeddedLanguageData=t}clone(){return(this.embeddedLanguageData?this.embeddedLanguageData.clone():null)===this.embeddedLanguageData?this:uu.create(this.stack,this.embeddedLanguageData)}equals(e){return!(e instanceof gy)||!this.stack.equals(e.stack)?!1:this.embeddedLanguageData===null&&e.embeddedLanguageData===null?!0:this.embeddedLanguageData===null||e.embeddedLanguageData===null?!1:this.embeddedLanguageData.equals(e.embeddedLanguageData)}}class KBe{constructor(){this._tokens=[],this._languageId=null,this._lastTokenType=null,this._lastTokenLanguage=null}enterLanguage(e){this._languageId=e}emit(e,t){this._lastTokenType===t&&this._lastTokenLanguage===this._languageId||(this._lastTokenType=t,this._lastTokenLanguage=this._languageId,this._tokens.push(new pS(e,t,this._languageId)))}nestedLanguageTokenize(e,t,i,n){const o=i.languageId,r=i.state,a=yi.get(o);if(!a)return this.enterLanguage(o),this.emit(n,""),r;const l=a.tokenize(e,t,r);if(n!==0)for(const c of l.tokens)this._tokens.push(new pS(c.offset+n,c.type,c.language));else this._tokens=this._tokens.concat(l.tokens);return this._lastTokenType=null,this._lastTokenLanguage=null,this._languageId=null,l.endState}finalize(e){return new GW(this._tokens,e)}}class gT{constructor(e,t){this._languageService=e,this._theme=t,this._prependTokens=null,this._tokens=[],this._currentLanguageId=0,this._lastTokenMetadata=0}enterLanguage(e){this._currentLanguageId=this._languageService.languageIdCodec.encodeLanguageId(e)}emit(e,t){const i=this._theme.match(this._currentLanguageId,t)|1024;this._lastTokenMetadata!==i&&(this._lastTokenMetadata=i,this._tokens.push(e),this._tokens.push(i))}static _merge(e,t,i){const n=e!==null?e.length:0,o=t.length,r=i!==null?i.length:0;if(n===0&&o===0&&r===0)return new Uint32Array(0);if(n===0&&o===0)return i;if(o===0&&r===0)return e;const a=new Uint32Array(n+o+r);e!==null&&a.set(e);for(let l=0;l<o;l++)a[n+l]=t[l];return i!==null&&a.set(i,n+o),a}nestedLanguageTokenize(e,t,i,n){const o=i.languageId,r=i.state,a=yi.get(o);if(!a)return this.enterLanguage(o),this.emit(n,""),r;const l=a.tokenizeEncoded(e,t,r);if(n!==0)for(let c=0,d=l.tokens.length;c<d;c+=2)l.tokens[c]+=n;return this._prependTokens=gT._merge(this._prependTokens,this._tokens,l.tokens),this._tokens=[],this._currentLanguageId=0,this._lastTokenMetadata=0,l.endState}finalize(e){return new gA(gT._merge(this._prependTokens,this._tokens,null),e)}}let x0=o9=class extends F{constructor(e,t,i,n,o){super(),this._configurationService=o,this._languageService=e,this._standaloneThemeService=t,this._languageId=i,this._lexer=n,this._embeddedLanguages=Object.create(null),this.embeddedLoaded=Promise.resolve(void 0);let r=!1;this._register(yi.onDidChange(a=>{if(r)return;let l=!1;for(let c=0,d=a.changedLanguages.length;c<d;c++){const h=a.changedLanguages[c];if(this._embeddedLanguages[h]){l=!0;break}}l&&(r=!0,yi.handleChange([this._languageId]),r=!1)})),this._maxTokenizationLineLength=this._configurationService.getValue("editor.maxTokenizationLineLength",{overrideIdentifier:this._languageId}),this._register(this._configurationService.onDidChangeConfiguration(a=>{a.affectsConfiguration("editor.maxTokenizationLineLength")&&(this._maxTokenizationLineLength=this._configurationService.getValue("editor.maxTokenizationLineLength",{overrideIdentifier:this._languageId}))}))}getLoadStatus(){const e=[];for(const t in this._embeddedLanguages){const i=yi.get(t);if(i){if(i instanceof o9){const n=i.getLoadStatus();n.loaded===!1&&e.push(n.promise)}continue}yi.isResolved(t)||e.push(yi.getOrCreate(t))}return e.length===0?{loaded:!0}:{loaded:!1,promise:Promise.all(e).then(t=>{})}}getInitialState(){const e=S0.create(null,this._lexer.start);return uu.create(e,null)}tokenize(e,t,i){if(e.length>=this._maxTokenizationLineLength)return h7(this._languageId,i);const n=new KBe,o=this._tokenize(e,t,i,n);return n.finalize(o)}tokenizeEncoded(e,t,i){if(e.length>=this._maxTokenizationLineLength)return mA(this._languageService.languageIdCodec.encodeLanguageId(this._languageId),i);const n=new gT(this._languageService,this._standaloneThemeService.getColorTheme().tokenTheme),o=this._tokenize(e,t,i,n);return n.finalize(o)}_tokenize(e,t,i,n){return i.embeddedLanguageData?this._nestedTokenize(e,t,i,0,n):this._myTokenize(e,t,i,0,n)}_findLeavingNestedLanguageOffset(e,t){let i=this._lexer.tokenizer[t.stack.state];if(!i&&(i=nk(this._lexer,t.stack.state),!i))throw ei(this._lexer,"tokenizer state is not defined: "+t.stack.state);let n=-1,o=!1;for(const r of i){if(!rY(r.action)||!(r.action.nextEmbedded==="@pop"||r.action.hasEmbeddedEndInCases))continue;o=!0;let a=r.resolveRegex(t.stack.state);const l=a.source;if(l.substr(0,4)==="^(?:"&&l.substr(l.length-1,1)===")"){const d=(a.ignoreCase?"i":"")+(a.unicode?"u":"");a=new RegExp(l.substr(4,l.length-5),d)}const c=e.search(a);c===-1||c!==0&&r.matchOnlyAtLineStart||(n===-1||c<n)&&(n=c)}if(!o)throw ei(this._lexer,'no rule containing nextEmbedded: "@pop" in tokenizer embedded state: '+t.stack.state);return n}_nestedTokenize(e,t,i,n,o){const r=this._findLeavingNestedLanguageOffset(e,i);if(r===-1){const c=o.nestedLanguageTokenize(e,t,i.embeddedLanguageData,n);return uu.create(i.stack,new Hb(i.embeddedLanguageData.languageId,c))}const a=e.substring(0,r);a.length>0&&o.nestedLanguageTokenize(a,!1,i.embeddedLanguageData,n);const l=e.substring(r);return this._myTokenize(l,t,i,n+r,o)}_safeRuleName(e){return e?e.name:"(unknown)"}_myTokenize(e,t,i,n,o){o.enterLanguage(this._languageId);const r=e.length,a=t&&this._lexer.includeLF?e+`
`:e,l=a.length;let c=i.embeddedLanguageData,d=i.stack,h=0,u=null,g=!0;for(;g||h<l;){const f=h,p=d.depth,_=u?u.groups.length:0,b=d.state;let w=null,C=null,v=null,y=null,x=null;if(u){w=u.matches;const I=u.groups.shift();C=I.matched,v=I.action,y=u.rule,u.groups.length===0&&(u=null)}else{if(!g&&h>=l)break;g=!1;let I=this._lexer.tokenizer[b];if(!I&&(I=nk(this._lexer,b),!I))throw ei(this._lexer,"tokenizer state is not defined: "+b);const R=a.substr(h);for(const O of I)if((h===0||!O.matchOnlyAtLineStart)&&(w=R.match(O.resolveRegex(b)),w)){C=w[0],v=O.action;break}}if(w||(w=[""],C=""),v||(h<l&&(w=[a.charAt(h)],C=w[0]),v=this._lexer.defaultToken),C===null)break;for(h+=C.length;zBe(v)&&rY(v)&&v.test;)v=v.test(C,w,b,h===l);let L=null;if(typeof v=="string"||Array.isArray(v))L=v;else if(v.group)L=v.group;else if(v.token!==null&&v.token!==void 0){if(v.tokenSubst?L=hu(this._lexer,v.token,C,w,b):L=v.token,v.nextEmbedded)if(v.nextEmbedded==="@pop"){if(!c)throw ei(this._lexer,"cannot pop embedded language if not inside one");c=null}else{if(c)throw ei(this._lexer,"cannot enter embedded language from within an embedded language");x=hu(this._lexer,v.nextEmbedded,C,w,b)}if(v.goBack&&(h=Math.max(0,h-v.goBack)),v.switchTo&&typeof v.switchTo=="string"){let I=hu(this._lexer,v.switchTo,C,w,b);if(I[0]==="@"&&(I=I.substr(1)),nk(this._lexer,I))d=d.switchTo(I);else throw ei(this._lexer,"trying to switch to a state '"+I+"' that is undefined in rule: "+this._safeRuleName(y))}else{if(v.transform&&typeof v.transform=="function")throw ei(this._lexer,"action.transform not supported");if(v.next)if(v.next==="@push"){if(d.depth>=this._lexer.maxStack)throw ei(this._lexer,"maximum tokenizer stack size reached: ["+d.state+","+d.parent.state+",...]");d=d.push(b)}else if(v.next==="@pop"){if(d.depth<=1)throw ei(this._lexer,"trying to pop an empty stack in rule: "+this._safeRuleName(y));d=d.pop()}else if(v.next==="@popall")d=d.popall();else{let I=hu(this._lexer,v.next,C,w,b);if(I[0]==="@"&&(I=I.substr(1)),nk(this._lexer,I))d=d.push(I);else throw ei(this._lexer,"trying to set a next state '"+I+"' that is undefined in rule: "+this._safeRuleName(y))}}v.log&&typeof v.log=="string"&&$Be(this._lexer,this._lexer.languageId+": "+hu(this._lexer,v.log,C,w,b))}if(L===null)throw ei(this._lexer,"lexer rule has no well-defined action in rule: "+this._safeRuleName(y));const E=I=>{const R=this._languageService.getLanguageIdByLanguageName(I)||this._languageService.getLanguageIdByMimeType(I)||I,O=this._getNestedEmbeddedLanguageData(R);if(h<l){const P=e.substr(h);return this._nestedTokenize(P,t,uu.create(d,O),n+h,o)}else return uu.create(d,O)};if(Array.isArray(L)){if(u&&u.groups.length>0)throw ei(this._lexer,"groups cannot be nested: "+this._safeRuleName(y));if(w.length!==L.length+1)throw ei(this._lexer,"matched number of groups does not match the number of actions in rule: "+this._safeRuleName(y));let I=0;for(let R=1;R<w.length;R++)I+=w[R].length;if(I!==C.length)throw ei(this._lexer,"with groups, all characters should be matched in consecutive groups in rule: "+this._safeRuleName(y));u={rule:y,matches:w,groups:[]};for(let R=0;R<L.length;R++)u.groups[R]={action:L[R],matched:w[R+1]};h-=C.length;continue}else{if(L==="@rematch"&&(h-=C.length,C="",w=null,L="",x!==null))return E(x);if(C.length===0){if(l===0||p!==d.depth||b!==d.state||(u?u.groups.length:0)!==_)continue;throw ei(this._lexer,"no progress in tokenizer in rule: "+this._safeRuleName(y))}let I=null;if(xre(L)&&L.indexOf("@brackets")===0){const R=L.substr(9),O=YBe(this._lexer,C);if(!O)throw ei(this._lexer,"@brackets token returned but no bracket defined as: "+C);I=aY(O.token+R)}else{const R=L===""?"":L+this._lexer.tokenPostfix;I=aY(R)}f<r&&o.emit(f+n,I)}if(x!==null)return E(x)}return uu.create(d,c)}_getNestedEmbeddedLanguageData(e){if(!this._languageService.isRegisteredLanguageId(e))return new Hb(e,bv);e!==this._languageId&&(this._languageService.requestBasicLanguageFeatures(e),yi.getOrCreate(e),this._embeddedLanguages[e]=!0);const t=yi.get(e);return t?new Hb(e,t.getInitialState()):new Hb(e,bv)}};x0=o9=jBe([GBe(4,He)],x0);function YBe(s,e){if(!e)return null;e=lh(s,e);const t=s.brackets;for(const i of t){if(i.open===e)return{token:i.token,bracketType:1};if(i.close===e)return{token:i.token,bracketType:-1}}return null}const ZBe=$c("standaloneColorizer",{createHTML:s=>s});class wV{static colorizeElement(e,t,i,n){n=n||{};const o=n.theme||"vs",r=n.mimeType||i.getAttribute("lang")||i.getAttribute("data-lang");if(!r)return console.error("Mode not detected"),Promise.resolve();const a=t.getLanguageIdByMimeType(r)||r;e.setTheme(o);const l=i.firstChild?i.firstChild.nodeValue:"";i.className+=" "+o;const c=d=>{const h=ZBe?.createHTML(d)??d;i.innerHTML=h};return this.colorize(t,l||"",a,n).then(c,d=>console.error(d))}static async colorize(e,t,i,n){const o=e.languageIdCodec;let r=4;n&&typeof n.tabSize=="number"&&(r=n.tabSize),X6(t)&&(t=t.substr(1));const a=qo(t);if(!e.isRegisteredLanguageId(i))return lY(a,r,o);const l=await yi.getOrCreate(i);return l?XBe(a,r,l,o):lY(a,r,o)}static colorizeLine(e,t,i,n,o=4){const r=Er.isBasicASCII(e,t),a=Er.containsRTL(e,r,i);return lA(new Dh(!1,!0,e,!1,r,a,0,n,[],o,0,0,0,0,-1,"none",!1,!1,null,null,0)).html}static colorizeModelLine(e,t,i=4){const n=e.getLineContent(t);e.tokenization.forceTokenization(t);const r=e.tokenization.getLineTokens(t).inflate();return this.colorizeLine(n,e.mightContainNonBasicASCII(),e.mightContainRTL(),r,i)}}function XBe(s,e,t,i){return new Promise((n,o)=>{const r=()=>{const a=QBe(s,e,t,i);if(t instanceof x0){const l=t.getLoadStatus();if(l.loaded===!1){l.promise.then(r,o);return}}n(a)};r()})}function lY(s,e,t){let i=[];const o=new Uint32Array(2);o[0]=0,o[1]=33587200;for(let r=0,a=s.length;r<a;r++){const l=s[r];o[0]=l.length;const c=new Ri(o,l,t),d=Er.isBasicASCII(l,!0),h=Er.containsRTL(l,d,!0),u=lA(new Dh(!1,!0,l,!1,d,h,0,c,[],e,0,0,0,0,-1,"none",!1,!1,null,null,0));i=i.concat(u.html),i.push("<br/>")}return i.join("")}function QBe(s,e,t,i){let n=[],o=t.getInitialState();for(let r=0,a=s.length;r<a;r++){const l=s[r],c=t.tokenizeEncoded(l,!0,o);Ri.convertToEndOffset(c.tokens,l.length);const d=new Ri(c.tokens,l,i),h=Er.isBasicASCII(l,!0),u=Er.containsRTL(l,h,!0),g=lA(new Dh(!1,!0,l,!1,h,u,0,d.inflate(),[],e,0,0,0,0,-1,"none",!1,!1,null,null,0));n=n.concat(g.html),n.push("<br/>"),o=c.endState}return n.join("")}var JBe=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},e9e=function(s,e){return function(t,i){e(t,i,s)}};let r9=class extends F{constructor(e){super(),this._themeService=e,this._onWillCreateCodeEditor=this._register(new M),this._onCodeEditorAdd=this._register(new M),this.onCodeEditorAdd=this._onCodeEditorAdd.event,this._onCodeEditorRemove=this._register(new M),this.onCodeEditorRemove=this._onCodeEditorRemove.event,this._onWillCreateDiffEditor=this._register(new M),this._onDiffEditorAdd=this._register(new M),this.onDiffEditorAdd=this._onDiffEditorAdd.event,this._onDiffEditorRemove=this._register(new M),this.onDiffEditorRemove=this._onDiffEditorRemove.event,this._decorationOptionProviders=new Map,this._codeEditorOpenHandlers=new Ns,this._modelProperties=new Map,this._codeEditors=Object.create(null),this._diffEditors=Object.create(null),this._globalStyleSheet=null}willCreateCodeEditor(){this._onWillCreateCodeEditor.fire()}addCodeEditor(e){this._codeEditors[e.getId()]=e,this._onCodeEditorAdd.fire(e)}removeCodeEditor(e){delete this._codeEditors[e.getId()]&&this._onCodeEditorRemove.fire(e)}listCodeEditors(){return Object.keys(this._codeEditors).map(e=>this._codeEditors[e])}willCreateDiffEditor(){this._onWillCreateDiffEditor.fire()}addDiffEditor(e){this._diffEditors[e.getId()]=e,this._onDiffEditorAdd.fire(e)}removeDiffEditor(e){delete this._diffEditors[e.getId()]&&this._onDiffEditorRemove.fire(e)}listDiffEditors(){return Object.keys(this._diffEditors).map(e=>this._diffEditors[e])}getFocusedCodeEditor(){let e=null;const t=this.listCodeEditors();for(const i of t){if(i.hasTextFocus())return i;i.hasWidgetFocus()&&(e=i)}return e}removeDecorationType(e){const t=this._decorationOptionProviders.get(e);t&&(t.refCount--,t.refCount<=0&&(this._decorationOptionProviders.delete(e),t.dispose(),this.listCodeEditors().forEach(i=>i.removeDecorationsByType(e))))}setModelProperty(e,t,i){const n=e.toString();let o;this._modelProperties.has(n)?o=this._modelProperties.get(n):(o=new Map,this._modelProperties.set(n,o)),o.set(t,i)}getModelProperty(e,t){const i=e.toString();if(this._modelProperties.has(i))return this._modelProperties.get(i).get(t)}async openCodeEditor(e,t,i){for(const n of this._codeEditorOpenHandlers){const o=await n(e,t,i);if(o!==null)return o}return null}registerCodeEditorOpenHandler(e){const t=this._codeEditorOpenHandlers.unshift(e);return pe(t)}};r9=JBe([e9e(0,wi)],r9);var t9e=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},cY=function(s,e){return function(t,i){e(t,i,s)}};let fT=class extends r9{constructor(e,t){super(t),this._register(this.onCodeEditorAdd(()=>this._checkContextKey())),this._register(this.onCodeEditorRemove(()=>this._checkContextKey())),this._editorIsOpen=e.createKey("editorIsOpen",!1),this._activeCodeEditor=null,this._register(this.registerCodeEditorOpenHandler(async(i,n,o)=>n?this.doOpenEditor(n,i):null))}_checkContextKey(){let e=!1;for(const t of this.listCodeEditors())if(!t.isSimpleWidget){e=!0;break}this._editorIsOpen.set(e)}setActiveCodeEditor(e){this._activeCodeEditor=e}getActiveCodeEditor(){return this._activeCodeEditor}doOpenEditor(e,t){if(!this.findModel(e,t.resource)){if(t.resource){const o=t.resource.scheme;if(o===Te.http||o===Te.https)return OQ(t.resource.toString()),e}return null}const n=t.options?t.options.selection:null;if(n)if(typeof n.endLineNumber=="number"&&typeof n.endColumn=="number")e.setSelection(n),e.revealRangeInCenter(n,1);else{const o={lineNumber:n.startLineNumber,column:n.startColumn};e.setPosition(o),e.revealPositionInCenter(o,1)}return e}findModel(e,t){const i=e.getModel();return i&&i.uri.toString()!==t.toString()?null:i}};fT=t9e([cY(0,Ae),cY(1,wi)],fT);tt(ht,fT,0);var kre=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},Ere=function(s,e){return function(t,i){e(t,i,s)}};let pT=class{get mainContainer(){return this._codeEditorService.listCodeEditors().at(0)?.getContainerDomNode()??Nt.document.body}get activeContainer(){return(this._codeEditorService.getFocusedCodeEditor()??this._codeEditorService.getActiveCodeEditor())?.getContainerDomNode()??this.mainContainer}get mainContainerDimension(){return lg(this.mainContainer)}get activeContainerDimension(){return lg(this.activeContainer)}get containers(){return Al(this._codeEditorService.listCodeEditors().map(e=>e.getContainerDomNode()))}getContainer(){return this.activeContainer}whenContainerStylesLoaded(){}focus(){this._codeEditorService.getFocusedCodeEditor()?.focus()}constructor(e){this._codeEditorService=e,this.onDidLayoutMainContainer=oe.None,this.onDidLayoutActiveContainer=oe.None,this.onDidLayoutContainer=oe.None,this.onDidChangeActiveContainer=oe.None,this.onDidAddContainer=oe.None,this.mainContainerOffset={top:0,quickPickTop:0},this.activeContainerOffset={top:0,quickPickTop:0}}};pT=kre([Ere(0,ht)],pT);let a9=class extends pT{get mainContainer(){return this._container}constructor(e,t){super(t),this._container=e}};a9=kre([Ere(1,ht)],a9);tt(qc,pT,1);var i9e=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},dY=function(s,e){return function(t,i){e(t,i,s)}};function sk(s){return s.scheme===Te.file?s.fsPath:s.path}let Ire=0;class ok{constructor(e,t,i,n,o,r,a){this.id=++Ire,this.type=0,this.actual=e,this.label=e.label,this.confirmBeforeUndo=e.confirmBeforeUndo||!1,this.resourceLabel=t,this.strResource=i,this.resourceLabels=[this.resourceLabel],this.strResources=[this.strResource],this.groupId=n,this.groupOrder=o,this.sourceId=r,this.sourceOrder=a,this.isValid=!0}setValid(e){this.isValid=e}toString(){return`[id:${this.id}] [group:${this.groupId}] [${this.isValid?"  VALID":"INVALID"}] ${this.actual.constructor.name} - ${this.actual}`}}class hY{constructor(e,t){this.resourceLabel=e,this.reason=t}}class uY{constructor(){this.elements=new Map}createMessage(){const e=[],t=[];for(const[,n]of this.elements)(n.reason===0?e:t).push(n.resourceLabel);const i=[];return e.length>0&&i.push(m(2030,"The following files have been closed and modified on disk: {0}.",e.join(", "))),t.length>0&&i.push(m(2031,"The following files have been modified in an incompatible way: {0}.",t.join(", "))),i.join(`
`)}get size(){return this.elements.size}has(e){return this.elements.has(e)}set(e,t){this.elements.set(e,t)}delete(e){return this.elements.delete(e)}}class n9e{constructor(e,t,i,n,o,r,a){this.id=++Ire,this.type=1,this.actual=e,this.label=e.label,this.confirmBeforeUndo=e.confirmBeforeUndo||!1,this.resourceLabels=t,this.strResources=i,this.groupId=n,this.groupOrder=o,this.sourceId=r,this.sourceOrder=a,this.removedResources=null,this.invalidatedResources=null}canSplit(){return typeof this.actual.split=="function"}removeResource(e,t,i){this.removedResources||(this.removedResources=new uY),this.removedResources.has(t)||this.removedResources.set(t,new hY(e,i))}setValid(e,t,i){i?this.invalidatedResources&&(this.invalidatedResources.delete(t),this.invalidatedResources.size===0&&(this.invalidatedResources=null)):(this.invalidatedResources||(this.invalidatedResources=new uY),this.invalidatedResources.has(t)||this.invalidatedResources.set(t,new hY(e,0)))}toString(){return`[id:${this.id}] [group:${this.groupId}] [${this.invalidatedResources?"INVALID":"  VALID"}] ${this.actual.constructor.name} - ${this.actual}`}}class Dre{constructor(e,t){this.resourceLabel=e,this.strResource=t,this._past=[],this._future=[],this.locked=!1,this.versionId=1}dispose(){for(const e of this._past)e.type===1&&e.removeResource(this.resourceLabel,this.strResource,0);for(const e of this._future)e.type===1&&e.removeResource(this.resourceLabel,this.strResource,0);this.versionId++}toString(){const e=[];e.push(`* ${this.strResource}:`);for(let t=0;t<this._past.length;t++)e.push(`   * [UNDO] ${this._past[t]}`);for(let t=this._future.length-1;t>=0;t--)e.push(`   * [REDO] ${this._future[t]}`);return e.join(`
`)}flushAllElements(){this._past=[],this._future=[],this.versionId++}_setElementValidFlag(e,t){e.type===1?e.setValid(this.resourceLabel,this.strResource,t):e.setValid(t)}setElementsValidFlag(e,t){for(const i of this._past)t(i.actual)&&this._setElementValidFlag(i,e);for(const i of this._future)t(i.actual)&&this._setElementValidFlag(i,e)}pushElement(e){for(const t of this._future)t.type===1&&t.removeResource(this.resourceLabel,this.strResource,1);this._future=[],this._past.push(e),this.versionId++}createSnapshot(e){const t=[];for(let i=0,n=this._past.length;i<n;i++)t.push(this._past[i].id);for(let i=this._future.length-1;i>=0;i--)t.push(this._future[i].id);return new ute(e,t)}restoreSnapshot(e){const t=e.elements.length;let i=!0,n=0,o=-1;for(let a=0,l=this._past.length;a<l;a++,n++){const c=this._past[a];i&&(n>=t||c.id!==e.elements[n])&&(i=!1,o=0),!i&&c.type===1&&c.removeResource(this.resourceLabel,this.strResource,0)}let r=-1;for(let a=this._future.length-1;a>=0;a--,n++){const l=this._future[a];i&&(n>=t||l.id!==e.elements[n])&&(i=!1,r=a),!i&&l.type===1&&l.removeResource(this.resourceLabel,this.strResource,0)}o!==-1&&(this._past=this._past.slice(0,o)),r!==-1&&(this._future=this._future.slice(r+1)),this.versionId++}getElements(){const e=[],t=[];for(const i of this._past)e.push(i.actual);for(const i of this._future)t.push(i.actual);return{past:e,future:t}}getClosestPastElement(){return this._past.length===0?null:this._past[this._past.length-1]}getSecondClosestPastElement(){return this._past.length<2?null:this._past[this._past.length-2]}getClosestFutureElement(){return this._future.length===0?null:this._future[this._future.length-1]}hasPastElements(){return this._past.length>0}hasFutureElements(){return this._future.length>0}splitPastWorkspaceElement(e,t){for(let i=this._past.length-1;i>=0;i--)if(this._past[i]===e){t.has(this.strResource)?this._past[i]=t.get(this.strResource):this._past.splice(i,1);break}this.versionId++}splitFutureWorkspaceElement(e,t){for(let i=this._future.length-1;i>=0;i--)if(this._future[i]===e){t.has(this.strResource)?this._future[i]=t.get(this.strResource):this._future.splice(i,1);break}this.versionId++}moveBackward(e){this._past.pop(),this._future.push(e),this.versionId++}moveForward(e){this._future.pop(),this._past.push(e),this.versionId++}}class BO{constructor(e){this.editStacks=e,this._versionIds=[];for(let t=0,i=this.editStacks.length;t<i;t++)this._versionIds[t]=this.editStacks[t].versionId}isValid(){for(let e=0,t=this.editStacks.length;e<t;e++)if(this._versionIds[e]!==this.editStacks[e].versionId)return!1;return!0}}const Tre=new Dre("","");Tre.locked=!0;let l9=class{constructor(e,t){this._dialogService=e,this._notificationService=t,this._editStacks=new Map,this._uriComparisonKeyComputers=[]}getUriComparisonKey(e){for(const t of this._uriComparisonKeyComputers)if(t[0]===e.scheme)return t[1].getComparisonKey(e);return e.toString()}_print(e){console.log("------------------------------------"),console.log(`AFTER ${e}: `);const t=[];for(const i of this._editStacks)t.push(i[1].toString());console.log(t.join(`
`))}pushElement(e,t=HF.None,i=vf.None){if(e.type===0){const n=sk(e.resource),o=this.getUriComparisonKey(e.resource);this._pushElement(new ok(e,n,o,t.id,t.nextOrder(),i.id,i.nextOrder()))}else{const n=new Set,o=[],r=[];for(const a of e.resources){const l=sk(a),c=this.getUriComparisonKey(a);n.has(c)||(n.add(c),o.push(l),r.push(c))}o.length===1?this._pushElement(new ok(e,o[0],r[0],t.id,t.nextOrder(),i.id,i.nextOrder())):this._pushElement(new n9e(e,o,r,t.id,t.nextOrder(),i.id,i.nextOrder()))}}_pushElement(e){for(let t=0,i=e.strResources.length;t<i;t++){const n=e.resourceLabels[t],o=e.strResources[t];let r;this._editStacks.has(o)?r=this._editStacks.get(o):(r=new Dre(n,o),this._editStacks.set(o,r)),r.pushElement(e)}}getLastElement(e){const t=this.getUriComparisonKey(e);if(this._editStacks.has(t)){const i=this._editStacks.get(t);if(i.hasFutureElements())return null;const n=i.getClosestPastElement();return n?n.actual:null}return null}_splitPastWorkspaceElement(e,t){const i=e.actual.split(),n=new Map;for(const o of i){const r=sk(o.resource),a=this.getUriComparisonKey(o.resource),l=new ok(o,r,a,0,0,0,0);n.set(l.strResource,l)}for(const o of e.strResources){if(t&&t.has(o))continue;this._editStacks.get(o).splitPastWorkspaceElement(e,n)}}_splitFutureWorkspaceElement(e,t){const i=e.actual.split(),n=new Map;for(const o of i){const r=sk(o.resource),a=this.getUriComparisonKey(o.resource),l=new ok(o,r,a,0,0,0,0);n.set(l.strResource,l)}for(const o of e.strResources){if(t&&t.has(o))continue;this._editStacks.get(o).splitFutureWorkspaceElement(e,n)}}removeElements(e){const t=typeof e=="string"?e:this.getUriComparisonKey(e);this._editStacks.has(t)&&(this._editStacks.get(t).dispose(),this._editStacks.delete(t))}setElementsValidFlag(e,t,i){const n=this.getUriComparisonKey(e);this._editStacks.has(n)&&this._editStacks.get(n).setElementsValidFlag(t,i)}createSnapshot(e){const t=this.getUriComparisonKey(e);return this._editStacks.has(t)?this._editStacks.get(t).createSnapshot(e):new ute(e,[])}restoreSnapshot(e){const t=this.getUriComparisonKey(e.resource);if(this._editStacks.has(t)){const i=this._editStacks.get(t);i.restoreSnapshot(e),!i.hasPastElements()&&!i.hasFutureElements()&&(i.dispose(),this._editStacks.delete(t))}}getElements(e){const t=this.getUriComparisonKey(e);return this._editStacks.has(t)?this._editStacks.get(t).getElements():{past:[],future:[]}}_findClosestUndoElementWithSource(e){if(!e)return[null,null];let t=null,i=null;for(const[n,o]of this._editStacks){const r=o.getClosestPastElement();r&&r.sourceId===e&&(!t||r.sourceOrder>t.sourceOrder)&&(t=r,i=n)}return[t,i]}canUndo(e){if(e instanceof vf){const[,i]=this._findClosestUndoElementWithSource(e.id);return!!i}const t=this.getUriComparisonKey(e);return this._editStacks.has(t)?this._editStacks.get(t).hasPastElements():!1}_onError(e,t){Oe(e);for(const i of t.strResources)this.removeElements(i);this._notificationService.error(e)}_acquireLocks(e){for(const t of e.editStacks)if(t.locked)throw new Error("Cannot acquire edit stack lock");for(const t of e.editStacks)t.locked=!0;return()=>{for(const t of e.editStacks)t.locked=!1}}_safeInvokeWithLocks(e,t,i,n,o){const r=this._acquireLocks(i);let a;try{a=t()}catch(l){return r(),n.dispose(),this._onError(l,e)}return a?a.then(()=>(r(),n.dispose(),o()),l=>(r(),n.dispose(),this._onError(l,e))):(r(),n.dispose(),o())}async _invokeWorkspacePrepare(e){if(typeof e.actual.prepareUndoRedo>"u")return F.None;const t=e.actual.prepareUndoRedo();return typeof t>"u"?F.None:t}_invokeResourcePrepare(e,t){if(e.actual.type!==1||typeof e.actual.prepareUndoRedo>"u")return t(F.None);const i=e.actual.prepareUndoRedo();return i?i_(i)?t(i):i.then(n=>t(n)):t(F.None)}_getAffectedEditStacks(e){const t=[];for(const i of e.strResources)t.push(this._editStacks.get(i)||Tre);return new BO(t)}_tryToSplitAndUndo(e,t,i,n){if(t.canSplit())return this._splitPastWorkspaceElement(t,i),this._notificationService.warn(n),new rk(this._undo(e,0,!0));for(const o of t.strResources)this.removeElements(o);return this._notificationService.warn(n),new rk}_checkWorkspaceUndo(e,t,i,n){if(t.removedResources)return this._tryToSplitAndUndo(e,t,t.removedResources,m(2032,"Could not undo '{0}' across all files. {1}",t.label,t.removedResources.createMessage()));if(n&&t.invalidatedResources)return this._tryToSplitAndUndo(e,t,t.invalidatedResources,m(2033,"Could not undo '{0}' across all files. {1}",t.label,t.invalidatedResources.createMessage()));const o=[];for(const a of i.editStacks)a.getClosestPastElement()!==t&&o.push(a.resourceLabel);if(o.length>0)return this._tryToSplitAndUndo(e,t,null,m(2034,"Could not undo '{0}' across all files because changes were made to {1}",t.label,o.join(", ")));const r=[];for(const a of i.editStacks)a.locked&&r.push(a.resourceLabel);return r.length>0?this._tryToSplitAndUndo(e,t,null,m(2035,"Could not undo '{0}' across all files because there is already an undo or redo operation running on {1}",t.label,r.join(", "))):i.isValid()?null:this._tryToSplitAndUndo(e,t,null,m(2036,"Could not undo '{0}' across all files because an undo or redo operation occurred in the meantime",t.label))}_workspaceUndo(e,t,i){const n=this._getAffectedEditStacks(t),o=this._checkWorkspaceUndo(e,t,n,!1);return o?o.returnValue:this._confirmAndExecuteWorkspaceUndo(e,t,n,i)}_isPartOfUndoGroup(e){if(!e.groupId)return!1;for(const[,t]of this._editStacks){const i=t.getClosestPastElement();if(i){if(i===e){const n=t.getSecondClosestPastElement();if(n&&n.groupId===e.groupId)return!0}if(i.groupId===e.groupId)return!0}}return!1}async _confirmAndExecuteWorkspaceUndo(e,t,i,n){if(t.canSplit()&&!this._isPartOfUndoGroup(t)){let a;(function(d){d[d.All=0]="All",d[d.This=1]="This",d[d.Cancel=2]="Cancel"})(a||(a={}));const{result:l}=await this._dialogService.prompt({type:fi.Info,message:m(2037,"Would you like to undo '{0}' across all files?",t.label),buttons:[{label:m(2038,"&&Undo in {0} Files",i.editStacks.length),run:()=>a.All},{label:m(2039,"Undo this &&File"),run:()=>a.This}],cancelButton:{run:()=>a.Cancel}});if(l===a.Cancel)return;if(l===a.This)return this._splitPastWorkspaceElement(t,null),this._undo(e,0,!0);const c=this._checkWorkspaceUndo(e,t,i,!1);if(c)return c.returnValue;n=!0}let o;try{o=await this._invokeWorkspacePrepare(t)}catch(a){return this._onError(a,t)}const r=this._checkWorkspaceUndo(e,t,i,!0);if(r)return o.dispose(),r.returnValue;for(const a of i.editStacks)a.moveBackward(t);return this._safeInvokeWithLocks(t,()=>t.actual.undo(),i,o,()=>this._continueUndoInGroup(t.groupId,n))}_resourceUndo(e,t,i){if(!t.isValid){e.flushAllElements();return}if(e.locked){const n=m(2040,"Could not undo '{0}' because there is already an undo or redo operation running.",t.label);this._notificationService.warn(n);return}return this._invokeResourcePrepare(t,n=>(e.moveBackward(t),this._safeInvokeWithLocks(t,()=>t.actual.undo(),new BO([e]),n,()=>this._continueUndoInGroup(t.groupId,i))))}_findClosestUndoElementInGroup(e){if(!e)return[null,null];let t=null,i=null;for(const[n,o]of this._editStacks){const r=o.getClosestPastElement();r&&r.groupId===e&&(!t||r.groupOrder>t.groupOrder)&&(t=r,i=n)}return[t,i]}_continueUndoInGroup(e,t){if(!e)return;const[,i]=this._findClosestUndoElementInGroup(e);if(i)return this._undo(i,0,t)}undo(e){if(e instanceof vf){const[,t]=this._findClosestUndoElementWithSource(e.id);return t?this._undo(t,e.id,!1):void 0}return typeof e=="string"?this._undo(e,0,!1):this._undo(this.getUriComparisonKey(e),0,!1)}_undo(e,t=0,i){if(!this._editStacks.has(e))return;const n=this._editStacks.get(e),o=n.getClosestPastElement();if(!o)return;if(o.groupId){const[a,l]=this._findClosestUndoElementInGroup(o.groupId);if(o!==a&&l)return this._undo(l,t,i)}return(o.sourceId!==t||o.confirmBeforeUndo)&&!i?this._confirmAndContinueUndo(e,t,o):o.type===1?this._workspaceUndo(e,o,i):this._resourceUndo(n,o,i)}async _confirmAndContinueUndo(e,t,i){if((await this._dialogService.confirm({message:m(2041,"Would you like to undo '{0}'?",i.label),primaryButton:m(2042,"&&Yes"),cancelButton:m(2043,"No")})).confirmed)return this._undo(e,t,!0)}_findClosestRedoElementWithSource(e){if(!e)return[null,null];let t=null,i=null;for(const[n,o]of this._editStacks){const r=o.getClosestFutureElement();r&&r.sourceId===e&&(!t||r.sourceOrder<t.sourceOrder)&&(t=r,i=n)}return[t,i]}canRedo(e){if(e instanceof vf){const[,i]=this._findClosestRedoElementWithSource(e.id);return!!i}const t=this.getUriComparisonKey(e);return this._editStacks.has(t)?this._editStacks.get(t).hasFutureElements():!1}_tryToSplitAndRedo(e,t,i,n){if(t.canSplit())return this._splitFutureWorkspaceElement(t,i),this._notificationService.warn(n),new rk(this._redo(e));for(const o of t.strResources)this.removeElements(o);return this._notificationService.warn(n),new rk}_checkWorkspaceRedo(e,t,i,n){if(t.removedResources)return this._tryToSplitAndRedo(e,t,t.removedResources,m(2044,"Could not redo '{0}' across all files. {1}",t.label,t.removedResources.createMessage()));if(n&&t.invalidatedResources)return this._tryToSplitAndRedo(e,t,t.invalidatedResources,m(2045,"Could not redo '{0}' across all files. {1}",t.label,t.invalidatedResources.createMessage()));const o=[];for(const a of i.editStacks)a.getClosestFutureElement()!==t&&o.push(a.resourceLabel);if(o.length>0)return this._tryToSplitAndRedo(e,t,null,m(2046,"Could not redo '{0}' across all files because changes were made to {1}",t.label,o.join(", ")));const r=[];for(const a of i.editStacks)a.locked&&r.push(a.resourceLabel);return r.length>0?this._tryToSplitAndRedo(e,t,null,m(2047,"Could not redo '{0}' across all files because there is already an undo or redo operation running on {1}",t.label,r.join(", "))):i.isValid()?null:this._tryToSplitAndRedo(e,t,null,m(2048,"Could not redo '{0}' across all files because an undo or redo operation occurred in the meantime",t.label))}_workspaceRedo(e,t){const i=this._getAffectedEditStacks(t),n=this._checkWorkspaceRedo(e,t,i,!1);return n?n.returnValue:this._executeWorkspaceRedo(e,t,i)}async _executeWorkspaceRedo(e,t,i){let n;try{n=await this._invokeWorkspacePrepare(t)}catch(r){return this._onError(r,t)}const o=this._checkWorkspaceRedo(e,t,i,!0);if(o)return n.dispose(),o.returnValue;for(const r of i.editStacks)r.moveForward(t);return this._safeInvokeWithLocks(t,()=>t.actual.redo(),i,n,()=>this._continueRedoInGroup(t.groupId))}_resourceRedo(e,t){if(!t.isValid){e.flushAllElements();return}if(e.locked){const i=m(2049,"Could not redo '{0}' because there is already an undo or redo operation running.",t.label);this._notificationService.warn(i);return}return this._invokeResourcePrepare(t,i=>(e.moveForward(t),this._safeInvokeWithLocks(t,()=>t.actual.redo(),new BO([e]),i,()=>this._continueRedoInGroup(t.groupId))))}_findClosestRedoElementInGroup(e){if(!e)return[null,null];let t=null,i=null;for(const[n,o]of this._editStacks){const r=o.getClosestFutureElement();r&&r.groupId===e&&(!t||r.groupOrder<t.groupOrder)&&(t=r,i=n)}return[t,i]}_continueRedoInGroup(e){if(!e)return;const[,t]=this._findClosestRedoElementInGroup(e);if(t)return this._redo(t)}redo(e){if(e instanceof vf){const[,t]=this._findClosestRedoElementWithSource(e.id);return t?this._redo(t):void 0}return typeof e=="string"?this._redo(e):this._redo(this.getUriComparisonKey(e))}_redo(e){if(!this._editStacks.has(e))return;const t=this._editStacks.get(e),i=t.getClosestFutureElement();if(i){if(i.groupId){const[n,o]=this._findClosestRedoElementInGroup(i.groupId);if(i!==n&&o)return this._redo(o)}return i.type===1?this._workspaceRedo(e,i):this._resourceRedo(t,i)}}};l9=i9e([dY(0,kx),dY(1,Ii)],l9);class rk{constructor(e){this.returnValue=e}}tt(f7,l9,1);var s9e=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},WO=function(s,e){return function(t,i){e(t,i,s)}};let c9=class extends F{constructor(e,t,i){super(),this._themeService=e,this._logService=t,this._languageService=i,this._caches=new WeakMap,this._register(this._themeService.onDidColorThemeChange(()=>{this._caches=new WeakMap}))}getStyling(e){return this._caches.has(e)||this._caches.set(e,new C8(e.getLegend(),this._themeService,this._languageService,this._logService)),this._caches.get(e)}};c9=s9e([WO(0,wi),WO(1,jt),WO(2,Ei)],c9);tt(JA,c9,1);function Nre(s){return typeof s=="string"?!1:Array.isArray(s)?s.every(Nre):!!s.exclusive}class gY{constructor(e,t,i,n,o){this.uri=e,this.languageId=t,this.notebookUri=i,this.notebookType=n,this.recursive=o}equals(e){return this.notebookType===e.notebookType&&this.languageId===e.languageId&&this.uri.toString()===e.uri.toString()&&this.notebookUri?.toString()===e.notebookUri?.toString()&&this.recursive===e.recursive}}class gi{get onDidChange(){return this._onDidChange.event}constructor(e){this._notebookInfoResolver=e,this._clock=0,this._entries=[],this._onDidChange=new M}register(e,t){let i={selector:e,provider:t,_score:-1,_time:this._clock++};return this._entries.push(i),this._lastCandidate=void 0,this._onDidChange.fire(this._entries.length),pe(()=>{if(i){const n=this._entries.indexOf(i);n>=0&&(this._entries.splice(n,1),this._lastCandidate=void 0,this._onDidChange.fire(this._entries.length),i=void 0)}})}has(e){return this.all(e).length>0}all(e){if(!e)return[];this._updateScores(e,!1);const t=[];for(const i of this._entries)i._score>0&&t.push(i.provider);return t}ordered(e,t=!1){const i=[];return this._orderedForEach(e,t,n=>i.push(n.provider)),i}orderedGroups(e){const t=[];let i,n;return this._orderedForEach(e,!1,o=>{i&&n===o._score?i.push(o.provider):(n=o._score,i=[o.provider],t.push(i))}),t}_orderedForEach(e,t,i){this._updateScores(e,t);for(const n of this._entries)n._score>0&&i(n)}_updateScores(e,t){const i=this._notebookInfoResolver?.(e.uri),n=i?new gY(e.uri,e.getLanguageId(),i.uri,i.type,t):new gY(e.uri,e.getLanguageId(),void 0,void 0,t);if(!this._lastCandidate?.equals(n)){this._lastCandidate=n;for(const o of this._entries)if(o._score=hV(o.selector,n.uri,n.languageId,nJ(e),n.notebookUri,n.notebookType),Nre(o.selector)&&o._score>0)if(t)o._score=0;else{for(const r of this._entries)r._score=0;o._score=1e3;break}this._entries.sort(gi._compareByScoreAndTime)}}static _compareByScoreAndTime(e,t){return e._score<t._score?1:e._score>t._score?-1:x1(e.selector)&&!x1(t.selector)?1:!x1(e.selector)&&x1(t.selector)?-1:e._time<t._time?1:e._time>t._time?-1:0}}function x1(s){return typeof s=="string"?!1:Array.isArray(s)?s.some(x1):!!s.isBuiltin}class o9e{constructor(){this.referenceProvider=new gi(this._score.bind(this)),this.renameProvider=new gi(this._score.bind(this)),this.newSymbolNamesProvider=new gi(this._score.bind(this)),this.codeActionProvider=new gi(this._score.bind(this)),this.definitionProvider=new gi(this._score.bind(this)),this.typeDefinitionProvider=new gi(this._score.bind(this)),this.declarationProvider=new gi(this._score.bind(this)),this.implementationProvider=new gi(this._score.bind(this)),this.documentSymbolProvider=new gi(this._score.bind(this)),this.inlayHintsProvider=new gi(this._score.bind(this)),this.colorProvider=new gi(this._score.bind(this)),this.codeLensProvider=new gi(this._score.bind(this)),this.documentFormattingEditProvider=new gi(this._score.bind(this)),this.documentRangeFormattingEditProvider=new gi(this._score.bind(this)),this.onTypeFormattingEditProvider=new gi(this._score.bind(this)),this.signatureHelpProvider=new gi(this._score.bind(this)),this.hoverProvider=new gi(this._score.bind(this)),this.documentHighlightProvider=new gi(this._score.bind(this)),this.multiDocumentHighlightProvider=new gi(this._score.bind(this)),this.selectionRangeProvider=new gi(this._score.bind(this)),this.foldingRangeProvider=new gi(this._score.bind(this)),this.linkProvider=new gi(this._score.bind(this)),this.inlineCompletionsProvider=new gi(this._score.bind(this)),this.completionProvider=new gi(this._score.bind(this)),this.linkedEditingRangeProvider=new gi(this._score.bind(this)),this.documentRangeSemanticTokensProvider=new gi(this._score.bind(this)),this.documentSemanticTokensProvider=new gi(this._score.bind(this)),this.documentDropEditProvider=new gi(this._score.bind(this)),this.documentPasteEditProvider=new gi(this._score.bind(this))}_score(e){return this._notebookTypeResolver?.(e)}}tt(ge,o9e,1);var r9e=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},ak=function(s,e){return function(t,i){e(t,i,s)}};const zl=ie;let d9=class extends Xo{get _targetWindow(){return be(this._target.targetElements[0])}get _targetDocumentElement(){return be(this._target.targetElements[0]).document.documentElement}get isDisposed(){return this._isDisposed}get isMouseIn(){return this._lockMouseTracker.isMouseIn}get domNode(){return this._hover.containerDomNode}get onDispose(){return this._onDispose.event}get onRequestLayout(){return this._onRequestLayout.event}get anchor(){return this._hoverPosition===2?0:1}get x(){return this._x}get y(){return this._y}get isLocked(){return this._isLocked}set isLocked(e){this._isLocked!==e&&(this._isLocked=e,this._hoverContainer.classList.toggle("locked",this._isLocked))}constructor(e,t,i,n,o){if(super(),this._keybindingService=t,this._configurationService=i,this._markdownRenderer=n,this._accessibilityService=o,this._messageListeners=new U,this._isDisposed=!1,this._forcePosition=!1,this._x=0,this._y=0,this._isLocked=!1,this._enableFocusTraps=!1,this._addedFocusTrap=!1,this._maxHeightRatioRelativeToWindow=.5,this._onDispose=this._register(new M),this._onRequestLayout=this._register(new M),this._linkHandler=e.linkHandler,this._target="targetElements"in e.target?e.target:new a9e(e.target),e.style)switch(e.style){case 1:{e.appearance??={},e.appearance.compact??=!0,e.appearance.showPointer??=!0;break}case 2:{e.appearance??={},e.appearance.compact??=!0;break}}this._hoverPointer=e.appearance?.showPointer?zl("div.workbench-hover-pointer"):void 0,this._hover=this._register(new aH(!e.appearance?.skipFadeInAnimation)),this._hover.containerDomNode.classList.add("workbench-hover"),e.appearance?.compact&&this._hover.containerDomNode.classList.add("workbench-hover","compact"),e.additionalClasses&&this._hover.containerDomNode.classList.add(...e.additionalClasses),e.position?.forcePosition&&(this._forcePosition=!0),e.trapFocus&&(this._enableFocusTraps=!0);const r=e.appearance?.maxHeightRatio;r!==void 0&&r>0&&r<=1&&(this._maxHeightRatioRelativeToWindow=r),this._hoverPosition=e.position?.hoverPosition===void 0?3:ch(e.position.hoverPosition)?e.position.hoverPosition:2,this.onmousedown(this._hover.containerDomNode,u=>u.stopPropagation()),this.onkeydown(this._hover.containerDomNode,u=>{u.equals(9)&&this.dispose()}),this._register(W(this._targetWindow,"blur",()=>this.dispose()));const a=zl("div.hover-row.markdown-hover"),l=zl("div.hover-contents");if(typeof e.content=="string")l.textContent=e.content,l.style.whiteSpace="pre-wrap";else if(ki(e.content))l.appendChild(e.content),l.classList.add("html-hover-contents");else{const u=e.content,{element:g}=this._register(this._markdownRenderer.render(u,{actionHandler:this._linkHandler,asyncRenderCallback:()=>{l.classList.add("code-hover-contents"),this.layout(),this._onRequestLayout.fire()}}));l.appendChild(g)}if(a.appendChild(l),this._hover.contentsDomNode.appendChild(a),e.actions&&e.actions.length>0){const u=zl("div.hover-row.status-bar"),g=zl("div.actions");e.actions.forEach(f=>{const p=this._keybindingService.lookupKeybinding(f.commandId),_=p?p.getLabel():null;this._register(NA.render(g,{label:f.label,commandId:f.commandId,run:b=>{f.run(b),this.dispose()},iconClass:f.iconClass},_))}),u.appendChild(g),this._hover.containerDomNode.appendChild(u)}this._hoverContainer=zl("div.workbench-hover-container"),this._hoverPointer&&this._hoverContainer.appendChild(this._hoverPointer),this._hoverContainer.appendChild(this._hover.containerDomNode);let c;if(e.actions&&e.actions.length>0?c=!1:e.persistence?.hideOnHover===void 0?c=typeof e.content=="string"||oh(e.content)&&!e.content.value.includes("](")&&!e.content.value.includes("</a>"):c=e.persistence.hideOnHover,e.appearance?.showHoverHint){const u=zl("div.hover-row.status-bar"),g=zl("div.info");g.textContent=m(1699,"Hold {0} key to mouse over",Ye?"Option":"Alt"),u.appendChild(g),this._hover.containerDomNode.appendChild(u)}const d=[...this._target.targetElements];c||d.push(this._hoverContainer);const h=this._register(new fY(d));if(this._register(h.onMouseOut(()=>{this._isLocked||this.dispose()})),c){const u=[...this._target.targetElements,this._hoverContainer];this._lockMouseTracker=this._register(new fY(u)),this._register(this._lockMouseTracker.onMouseOut(()=>{this._isLocked||this.dispose()}))}else this._lockMouseTracker=h}addFocusTrap(){if(!this._enableFocusTraps||this._addedFocusTrap)return;this._addedFocusTrap=!0;const e=this._hover.containerDomNode,t=this.findLastFocusableChild(this._hover.containerDomNode);if(t){const i=RR(this._hoverContainer,zl("div")),n=X(this._hoverContainer,zl("div"));i.tabIndex=0,n.tabIndex=0,this._register(W(n,"focus",o=>{e.focus(),o.preventDefault()})),this._register(W(i,"focus",o=>{t.focus(),o.preventDefault()}))}}findLastFocusableChild(e){if(e.hasChildNodes())for(let t=0;t<e.childNodes.length;t++){const i=e.childNodes.item(e.childNodes.length-t-1);if(i.nodeType===i.ELEMENT_NODE){const o=i;if(typeof o.tabIndex=="number"&&o.tabIndex>=0)return o}const n=this.findLastFocusableChild(i);if(n)return n}}render(e){e.appendChild(this._hoverContainer);const i=this._hoverContainer.contains(this._hoverContainer.ownerDocument.activeElement)&&Lne(this._configurationService.getValue("accessibility.verbosity.hover")===!0&&this._accessibilityService.isScreenReaderOptimized(),this._keybindingService.lookupKeybinding("editor.action.accessibleView")?.getAriaLabel());i&&xl(i),this.layout(),this.addFocusTrap()}layout(){this._hover.containerDomNode.classList.remove("right-aligned"),this._hover.contentsDomNode.style.maxHeight="";const e=d=>{const h=TQ(d),u=d.getBoundingClientRect();return{top:u.top*h,bottom:u.bottom*h,right:u.right*h,left:u.left*h}},t=this._target.targetElements.map(d=>e(d)),{top:i,right:n,bottom:o,left:r}=t[0],a=n-r,l=o-i,c={top:i,right:n,bottom:o,left:r,width:a,height:l,center:{x:r+a/2,y:i+l/2}};if(this.adjustHorizontalHoverPosition(c),this.adjustVerticalHoverPosition(c),this.adjustHoverMaxHeight(c),this._hoverContainer.style.padding="",this._hoverContainer.style.margin="",this._hoverPointer){switch(this._hoverPosition){case 1:c.left+=3,c.right+=3,this._hoverContainer.style.paddingLeft="3px",this._hoverContainer.style.marginLeft="-3px";break;case 0:c.left-=3,c.right-=3,this._hoverContainer.style.paddingRight="3px",this._hoverContainer.style.marginRight="-3px";break;case 2:c.top+=3,c.bottom+=3,this._hoverContainer.style.paddingTop="3px",this._hoverContainer.style.marginTop="-3px";break;case 3:c.top-=3,c.bottom-=3,this._hoverContainer.style.paddingBottom="3px",this._hoverContainer.style.marginBottom="-3px";break}c.center.x=c.left+a/2,c.center.y=c.top+l/2}this.computeXCordinate(c),this.computeYCordinate(c),this._hoverPointer&&(this._hoverPointer.classList.remove("top"),this._hoverPointer.classList.remove("left"),this._hoverPointer.classList.remove("right"),this._hoverPointer.classList.remove("bottom"),this.setHoverPointerPosition(c)),this._hover.onContentsChanged()}computeXCordinate(e){const t=this._hover.containerDomNode.clientWidth+2;this._target.x!==void 0?this._x=this._target.x:this._hoverPosition===1?this._x=e.right:this._hoverPosition===0?this._x=e.left-t:(this._hoverPointer?this._x=e.center.x-this._hover.containerDomNode.clientWidth/2:this._x=e.left,this._x+t>=this._targetDocumentElement.clientWidth&&(this._hover.containerDomNode.classList.add("right-aligned"),this._x=Math.max(this._targetDocumentElement.clientWidth-t-2,this._targetDocumentElement.clientLeft))),this._x<this._targetDocumentElement.clientLeft&&(this._x=e.left+2)}computeYCordinate(e){this._target.y!==void 0?this._y=this._target.y:this._hoverPosition===3?this._y=e.top:this._hoverPosition===2?this._y=e.bottom-2:this._hoverPointer?this._y=e.center.y+this._hover.containerDomNode.clientHeight/2:this._y=e.bottom,this._y>this._targetWindow.innerHeight&&(this._y=e.bottom)}adjustHorizontalHoverPosition(e){if(this._target.x!==void 0)return;const t=this._hoverPointer?3:0;if(this._forcePosition){const i=t+2;this._hoverPosition===1?this._hover.containerDomNode.style.maxWidth=`${this._targetDocumentElement.clientWidth-e.right-i}px`:this._hoverPosition===0&&(this._hover.containerDomNode.style.maxWidth=`${e.left-i}px`);return}this._hoverPosition===1?this._targetDocumentElement.clientWidth-e.right<this._hover.containerDomNode.clientWidth+t&&(e.left>=this._hover.containerDomNode.clientWidth+t?this._hoverPosition=0:this._hoverPosition=2):this._hoverPosition===0&&(e.left<this._hover.containerDomNode.clientWidth+t&&(this._targetDocumentElement.clientWidth-e.right>=this._hover.containerDomNode.clientWidth+t?this._hoverPosition=1:this._hoverPosition=2),e.left-this._hover.containerDomNode.clientWidth-t<=this._targetDocumentElement.clientLeft&&(this._hoverPosition=1))}adjustVerticalHoverPosition(e){if(this._target.y!==void 0||this._forcePosition)return;const t=this._hoverPointer?3:0;this._hoverPosition===3?e.top-this._hover.containerDomNode.clientHeight-t<0&&(this._hoverPosition=2):this._hoverPosition===2&&e.bottom+this._hover.containerDomNode.offsetHeight+t>this._targetWindow.innerHeight&&(this._hoverPosition=3)}adjustHoverMaxHeight(e){let t=this._targetWindow.innerHeight*this._maxHeightRatioRelativeToWindow;if(this._forcePosition){const i=(this._hoverPointer?3:0)+2;this._hoverPosition===3?t=Math.min(t,e.top-i):this._hoverPosition===2&&(t=Math.min(t,this._targetWindow.innerHeight-e.bottom-i))}if(this._hover.containerDomNode.style.maxHeight=`${t}px`,this._hover.contentsDomNode.clientHeight<this._hover.contentsDomNode.scrollHeight){const i=`${this._hover.scrollbar.options.verticalScrollbarSize}px`;this._hover.contentsDomNode.style.paddingRight!==i&&(this._hover.contentsDomNode.style.paddingRight=i)}}setHoverPointerPosition(e){if(this._hoverPointer)switch(this._hoverPosition){case 0:case 1:{this._hoverPointer.classList.add(this._hoverPosition===0?"right":"left");const t=this._hover.containerDomNode.clientHeight;t>e.height?this._hoverPointer.style.top=`${e.center.y-(this._y-t)-3}px`:this._hoverPointer.style.top=`${Math.round(t/2)-3}px`;break}case 3:case 2:{this._hoverPointer.classList.add(this._hoverPosition===3?"bottom":"top");const t=this._hover.containerDomNode.clientWidth;let i=Math.round(t/2)-3;const n=this._x+i;(n<e.left||n>e.right)&&(i=e.center.x-this._x-3),this._hoverPointer.style.left=`${i}px`;break}}}focus(){this._hover.containerDomNode.focus()}dispose(){this._isDisposed||(this._onDispose.fire(),this._target.dispose?.(),this._hoverContainer.remove(),this._messageListeners.dispose(),super.dispose()),this._isDisposed=!0}};d9=r9e([ak(1,ft),ak(2,He),ak(3,Ua),ak(4,Wn)],d9);class fY extends Xo{get onMouseOut(){return this._onMouseOut.event}get isMouseIn(){return this._isMouseIn}constructor(e,t=200){super(),this._elements=e,this._eventDebounceDelay=t,this._isMouseIn=!0,this._mouseTimer=this._register(new vt),this._onMouseOut=this._register(new M);for(const i of this._elements)this.onmouseover(i,()=>this._onTargetMouseOver()),this.onmouseleave(i,()=>this._onTargetMouseLeave())}_onTargetMouseOver(){this._isMouseIn=!0,this._mouseTimer.clear()}_onTargetMouseLeave(){this._isMouseIn=!1,this._mouseTimer.value=new $o(()=>this._fireIfMouseOutside(),this._eventDebounceDelay)}_fireIfMouseOutside(){this._isMouseIn||this._onMouseOut.fire()}}class a9e{constructor(e){this._element=e,this.targetElements=[this._element]}dispose(){}}function l9e(s){const e=s;return!!e&&typeof e.x=="number"&&typeof e.y=="number"}var Tu;(function(s){s[s.AVOID=0]="AVOID",s[s.ALIGN=1]="ALIGN"})(Tu||(Tu={}));function Vb(s,e,t){const i=t.mode===Tu.ALIGN?t.offset:t.offset+t.size,n=t.mode===Tu.ALIGN?t.offset+t.size:t.offset;return t.position===0?e<=s-i?i:e<=n?n-e:Math.max(s-e,0):e<=n?n-e:e<=s-i?i:0}const sw=class sw extends F{constructor(e,t){super(),this.container=null,this.useFixedPosition=!1,this.useShadowDOM=!1,this.delegate=null,this.toDisposeOnClean=F.None,this.toDisposeOnSetContainer=F.None,this.shadowRoot=null,this.shadowRootHostElement=null,this.view=ie(".context-view"),js(this.view),this.setContainer(e,t),this._register(pe(()=>this.setContainer(null,1)))}setContainer(e,t){this.useFixedPosition=t!==1;const i=this.useShadowDOM;if(this.useShadowDOM=t===3,!(e===this.container&&i===this.useShadowDOM)&&(this.container&&(this.toDisposeOnSetContainer.dispose(),this.view.remove(),this.shadowRoot&&(this.shadowRoot=null,this.shadowRootHostElement?.remove(),this.shadowRootHostElement=null),this.container=null),e)){if(this.container=e,this.useShadowDOM){this.shadowRootHostElement=ie(".shadow-root-host"),this.container.appendChild(this.shadowRootHostElement),this.shadowRoot=this.shadowRootHostElement.attachShadow({mode:"open"});const o=document.createElement("style");o.textContent=c9e,this.shadowRoot.appendChild(o),this.shadowRoot.appendChild(this.view),this.shadowRoot.appendChild(ie("slot"))}else this.container.appendChild(this.view);const n=new U;sw.BUBBLE_UP_EVENTS.forEach(o=>{n.add(Fi(this.container,o,r=>{this.onDOMEvent(r,!1)}))}),sw.BUBBLE_DOWN_EVENTS.forEach(o=>{n.add(Fi(this.container,o,r=>{this.onDOMEvent(r,!0)},!0))}),this.toDisposeOnSetContainer=n}}show(e){this.isVisible()&&this.hide(),Pn(this.view),this.view.className="context-view monaco-component",this.view.style.top="0px",this.view.style.left="0px",this.view.style.zIndex=`${2575+(e.layer??0)}`,this.view.style.position=this.useFixedPosition?"fixed":"absolute",Mo(this.view),this.toDisposeOnClean=e.render(this.view)||F.None,this.delegate=e,this.doLayout(),this.delegate.focus?.()}getViewElement(){return this.view}layout(){if(this.isVisible()){if(this.delegate.canRelayout===!1&&!(na&&z0.pointerEvents)){this.hide();return}this.delegate?.layout?.(),this.doLayout()}}doLayout(){if(!this.isVisible())return;const e=this.delegate.getAnchor();let t;if(ki(e)){const f=Li(e),p=TQ(e);t={top:f.top*p,left:f.left*p,width:f.width*p,height:f.height*p}}else l9e(e)?t={top:e.y,left:e.x,width:e.width||1,height:e.height||2}:t={top:e.posy,left:e.posx,width:2,height:2};const i=Ro(this.view),n=Hd(this.view),o=this.delegate.anchorPosition??0,r=this.delegate.anchorAlignment??0,a=this.delegate.anchorAxisAlignment??0;let l,c;const d=kt();if(a===0){const f={offset:t.top-d.pageYOffset,size:t.height,position:o===0?0:1},p={offset:t.left,size:t.width,position:r===0?0:1,mode:Tu.ALIGN};l=Vb(d.innerHeight,n,f)+d.pageYOffset,as.intersects({start:l,end:l+n},{start:f.offset,end:f.offset+f.size})&&(p.mode=Tu.AVOID),c=Vb(d.innerWidth,i,p)}else{const f={offset:t.left,size:t.width,position:r===0?0:1},p={offset:t.top,size:t.height,position:o===0?0:1,mode:Tu.ALIGN};c=Vb(d.innerWidth,i,f),as.intersects({start:c,end:c+i},{start:f.offset,end:f.offset+f.size})&&(p.mode=Tu.AVOID),l=Vb(d.innerHeight,n,p)+d.pageYOffset}this.view.classList.remove("top","bottom","left","right"),this.view.classList.add(o===0?"bottom":"top"),this.view.classList.add(r===0?"left":"right"),this.view.classList.toggle("fixed",this.useFixedPosition);const h=Li(this.container),u=this.container.scrollTop||0,g=this.container.scrollLeft||0;this.view.style.top=`${l-(this.useFixedPosition?Li(this.view).top:h.top)+u}px`,this.view.style.left=`${c-(this.useFixedPosition?Li(this.view).left:h.left)+g}px`,this.view.style.width="initial"}hide(e){const t=this.delegate;this.delegate=null,t?.onHide&&t.onHide(e),this.toDisposeOnClean.dispose(),js(this.view)}isVisible(){return!!this.delegate}onDOMEvent(e,t){this.delegate&&(this.delegate.onDOMEvent?this.delegate.onDOMEvent(e,be(e).document.activeElement):t&&!wn(e.target,this.container)&&this.hide())}dispose(){this.hide(),super.dispose()}};sw.BUBBLE_UP_EVENTS=["click","keydown","focus","blur"],sw.BUBBLE_DOWN_EVENTS=["click"];let h9=sw;const c9e=`
	:host {
		all: initial; /* 1st rule so subsequent properties are reset. */
	}

	.codicon[class*='codicon-'] {
		font: normal normal normal 16px/1 codicon;
		display: inline-block;
		text-decoration: none;
		text-rendering: auto;
		text-align: center;
		-webkit-font-smoothing: antialiased;
		-moz-osx-font-smoothing: grayscale;
		user-select: none;
		-webkit-user-select: none;
		-ms-user-select: none;
	}

	:host {
		font-family: -apple-system, BlinkMacSystemFont, "Segoe WPC", "Segoe UI", "HelveticaNeue-Light", system-ui, "Ubuntu", "Droid Sans", sans-serif;
	}

	:host-context(.mac) { font-family: -apple-system, BlinkMacSystemFont, sans-serif; }
	:host-context(.mac:lang(zh-Hans)) { font-family: -apple-system, BlinkMacSystemFont, "PingFang SC", "Hiragino Sans GB", sans-serif; }
	:host-context(.mac:lang(zh-Hant)) { font-family: -apple-system, BlinkMacSystemFont, "PingFang TC", sans-serif; }
	:host-context(.mac:lang(ja)) { font-family: -apple-system, BlinkMacSystemFont, "Hiragino Kaku Gothic Pro", sans-serif; }
	:host-context(.mac:lang(ko)) { font-family: -apple-system, BlinkMacSystemFont, "Apple SD Gothic Neo", "Nanum Gothic", "AppleGothic", sans-serif; }

	:host-context(.windows) { font-family: "Segoe WPC", "Segoe UI", sans-serif; }
	:host-context(.windows:lang(zh-Hans)) { font-family: "Segoe WPC", "Segoe UI", "Microsoft YaHei", sans-serif; }
	:host-context(.windows:lang(zh-Hant)) { font-family: "Segoe WPC", "Segoe UI", "Microsoft Jhenghei", sans-serif; }
	:host-context(.windows:lang(ja)) { font-family: "Segoe WPC", "Segoe UI", "Yu Gothic UI", "Meiryo UI", sans-serif; }
	:host-context(.windows:lang(ko)) { font-family: "Segoe WPC", "Segoe UI", "Malgun Gothic", "Dotom", sans-serif; }

	:host-context(.linux) { font-family: system-ui, "Ubuntu", "Droid Sans", sans-serif; }
	:host-context(.linux:lang(zh-Hans)) { font-family: system-ui, "Ubuntu", "Droid Sans", "Source Han Sans SC", "Source Han Sans CN", "Source Han Sans", sans-serif; }
	:host-context(.linux:lang(zh-Hant)) { font-family: system-ui, "Ubuntu", "Droid Sans", "Source Han Sans TC", "Source Han Sans TW", "Source Han Sans", sans-serif; }
	:host-context(.linux:lang(ja)) { font-family: system-ui, "Ubuntu", "Droid Sans", "Source Han Sans J", "Source Han Sans JP", "Source Han Sans", sans-serif; }
	:host-context(.linux:lang(ko)) { font-family: system-ui, "Ubuntu", "Droid Sans", "Source Han Sans K", "Source Han Sans JR", "Source Han Sans", "UnDotum", "FBaekmuk Gulim", sans-serif; }
`;var d9e=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},h9e=function(s,e){return function(t,i){e(t,i,s)}};let mT=class extends F{constructor(e){super(),this.layoutService=e,this.contextView=this._register(new h9(this.layoutService.mainContainer,1)),this.layout(),this._register(e.onDidLayoutContainer(()=>this.layout()))}showContextView(e,t,i){let n;t?t===this.layoutService.getContainer(be(t))?n=1:i?n=3:n=2:n=1,this.contextView.setContainer(t??this.layoutService.activeContainer,n),this.contextView.show(e);const o={close:()=>{this.openContextView===o&&this.hideContextView()}};return this.openContextView=o,o}layout(){this.contextView.layout()}hideContextView(e){this.contextView.hide(e),this.openContextView=void 0}};mT=d9e([h9e(0,qc)],mT);class u9e extends mT{getContextViewElement(){return this.contextView.getViewElement()}}function Rre(s){const e=s;return typeof e=="object"&&"markdown"in e&&"markdownNotSupportedFallback"in e}class g9e{constructor(e,t,i){this.hoverDelegate=e,this.target=t,this.fadeInAnimation=i}async update(e,t,i){if(this._cancellationTokenSource&&(this._cancellationTokenSource.dispose(!0),this._cancellationTokenSource=void 0),this.isDisposed)return;let n;if(bn(e)||ki(e)||e===void 0)n=e;else{this._cancellationTokenSource=new si;const o=this._cancellationTokenSource.token;let r;if(Rre(e)?vm(e.markdown)?r=e.markdown(o).then(a=>a??e.markdownNotSupportedFallback):r=e.markdown??e.markdownNotSupportedFallback:r=e.element(o),r instanceof Promise?(this._hoverWidget||this.show(m(1700,"Loading..."),t,i),n=await r):n=r,this.isDisposed||o.isCancellationRequested)return}this.show(n,t,i)}show(e,t,i){const n=this._hoverWidget;if(this.hasContent(e)){const o={content:e,target:this.target,actions:i?.actions,linkHandler:i?.linkHandler,trapFocus:i?.trapFocus,appearance:{showPointer:this.hoverDelegate.placement==="element",skipFadeInAnimation:!this.fadeInAnimation||!!n,showHoverHint:i?.appearance?.showHoverHint},position:{hoverPosition:2}};this._hoverWidget=this.hoverDelegate.showHover(o,t)}n?.dispose()}hasContent(e){return e?oh(e)?!!e.value:!0:!1}get isDisposed(){return this._hoverWidget?.isDisposed}dispose(){this._hoverWidget?.dispose(),this._cancellationTokenSource?.dispose(!0),this._cancellationTokenSource=void 0}}var f9e=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},rb=function(s,e){return function(t,i){e(t,i,s)}};let u9=class extends F{constructor(e,t,i,n,o,r){super(),this._instantiationService=e,this._configurationService=t,this._keybindingService=n,this._layoutService=o,this._accessibilityService=r,this._currentDelayedHoverWasShown=!1,this._delayedHovers=new Map,this._managedHovers=new Map,this._register(i.onDidShowContextMenu(()=>this.hideHover())),this._contextViewHandler=this._register(new mT(this._layoutService)),this._register(Dn.registerCommandAndKeybindingRule({id:"workbench.action.showHover",weight:0,primary:Gi(2089,2087),handler:()=>{this._showAndFocusHoverForActiveElement()}}))}showInstantHover(e,t,i,n){const o=this._createHover(e,i);if(o)return this._showHover(o,e,t),o}showDelayedHover(e,t){if(e.id===void 0&&(e.id=pY(e.content)),!this._currentDelayedHover||this._currentDelayedHoverWasShown){if(this._currentHover?.isLocked)return;if(zh(this._currentHoverOptions)===zh(e))return this._currentHover;if(this._currentHover&&!this._currentHover.isDisposed&&this._currentDelayedHoverGroupId!==void 0&&this._currentDelayedHoverGroupId===t?.groupId)return this.showInstantHover({...e,appearance:{...e.appearance,skipFadeInAnimation:!0}})}else if(this._currentDelayedHover&&zh(this._currentHoverOptions)===zh(e))return this._currentDelayedHover;const i=this._createHover(e,void 0);if(!i){this._currentDelayedHover=void 0,this._currentDelayedHoverWasShown=!1,this._currentDelayedHoverGroupId=void 0;return}return this._currentDelayedHover=i,this._currentDelayedHoverWasShown=!1,this._currentDelayedHoverGroupId=t?.groupId,Tc(this._configurationService.getValue("workbench.hover.delay")).then(()=>{i&&!i.isDisposed&&(this._currentDelayedHoverWasShown=!0,this._showHover(i,e))}),i}setupDelayedHover(e,t,i){const n=()=>({...typeof t=="function"?t():t,target:e});return this._setupDelayedHover(e,n,i)}setupDelayedHoverAtMouse(e,t,i){const n=o=>({...typeof t=="function"?t():t,target:{targetElements:[e],x:o!==void 0?o.x+10:void 0}});return this._setupDelayedHover(e,n,i)}_setupDelayedHover(e,t,i){const n=new U;return n.add(W(e,se.MOUSE_OVER,o=>{this.showDelayedHover(t(o),{groupId:i?.groupId})})),i?.setupKeyboardEvents&&n.add(W(e,se.KEY_DOWN,o=>{const r=new Pt(o);(r.equals(10)||r.equals(3))&&this.showInstantHover(t(),!0)})),this._delayedHovers.set(e,{show:o=>{this.showInstantHover(t(),o)}}),n.add(pe(()=>this._delayedHovers.delete(e))),n}_createHover(e,t){if(this._currentDelayedHover=void 0,e.content===""||this._currentHover?.isLocked||(e.id===void 0&&(e.id=pY(e.content)),zh(this._currentHoverOptions)===zh(e)))return;this._currentHoverOptions=e,this._lastHoverOptions=e;const i=e.trapFocus||this._accessibilityService.isScreenReaderOptimized(),n=ln();t||(i&&n?n.classList.contains("monaco-hover")||(this._lastFocusedElementBeforeOpen=n):this._lastFocusedElementBeforeOpen=void 0);const o=new U,r=this._instantiationService.createInstance(d9,e);if(e.persistence?.sticky&&(r.isLocked=!0),e.position?.hoverPosition&&!ch(e.position.hoverPosition)&&(e.target={targetElements:ki(e.target)?[e.target]:e.target.targetElements,x:e.position.hoverPosition.x+10}),r.onDispose(()=>{this._currentHover?.domNode&&RQ(this._currentHover.domNode)&&this._lastFocusedElementBeforeOpen?.focus(),zh(this._currentHoverOptions)===zh(e)&&this.doHideHover(),o.dispose()},void 0,o),!e.container){const a=ki(e.target)?e.target:e.target.targetElements[0];e.container=this._layoutService.getContainer(be(a))}if(r.onRequestLayout(()=>this._contextViewHandler.layout(),void 0,o),e.persistence?.sticky)o.add(W(be(e.container).document,se.MOUSE_DOWN,a=>{wn(a.target,r.domNode)||this.doHideHover()}));else{if("targetElements"in e.target)for(const l of e.target.targetElements)o.add(W(l,se.CLICK,()=>this.hideHover()));else o.add(W(e.target,se.CLICK,()=>this.hideHover()));const a=ln();if(a){const l=be(a).document;o.add(W(a,se.KEY_DOWN,c=>this._keyDown(c,r,!!e.persistence?.hideOnKeyDown))),o.add(W(l,se.KEY_DOWN,c=>this._keyDown(c,r,!!e.persistence?.hideOnKeyDown))),o.add(W(a,se.KEY_UP,c=>this._keyUp(c,r))),o.add(W(l,se.KEY_UP,c=>this._keyUp(c,r)))}}if("IntersectionObserver"in Nt){const a=new IntersectionObserver(c=>this._intersectionChange(c,r),{threshold:0}),l="targetElements"in e.target?e.target.targetElements[0]:e.target;a.observe(l),o.add(pe(()=>a.disconnect()))}return this._currentHover=r,r}_showHover(e,t,i){this._contextViewHandler.showContextView(new m9e(e,i),t.container)}hideHover(e){!e&&this._currentHover?.isLocked||!this._currentHoverOptions||this.doHideHover()}doHideHover(){this._currentHover=void 0,this._currentHoverOptions=void 0,this._contextViewHandler.hideContextView()}_intersectionChange(e,t){e[e.length-1].isIntersecting||t.dispose()}showAndFocusLastHover(){this._lastHoverOptions&&this.showInstantHover(this._lastHoverOptions,!0,!0)}_showAndFocusHoverForActiveElement(){let e=ln();for(;e;){const t=this._delayedHovers.get(e)??this._managedHovers.get(e);if(t){t.show(!0);return}e=e.parentElement}}_keyDown(e,t,i){if(e.key==="Alt"){t.isLocked=!0;return}const n=new Pt(e);this._keybindingService.resolveKeyboardEvent(n).getSingleModifierDispatchChords().some(r=>!!r)||this._keybindingService.softDispatch(n,n.target).kind!==0||i&&(!this._currentHoverOptions?.trapFocus||e.key!=="Tab")&&(this.hideHover(),this._lastFocusedElementBeforeOpen?.focus())}_keyUp(e,t){e.key==="Alt"&&(t.isLocked=!1,t.isMouseIn||(this.hideHover(),this._lastFocusedElementBeforeOpen?.focus()))}setupManagedHover(e,t,i,n){if(e.showNativeHover)return p9e(t,i);t.setAttribute("custom-hover","true"),t.title!==""&&(console.warn("HTML element already has a title attribute, which will conflict with the custom hover. Please remove the title attribute."),console.trace("Stack trace:",t.title),t.title="");let o,r;const a=(g,f)=>{const p=r!==void 0;g&&(r?.dispose(),r=void 0),f&&(o?.dispose(),o=void 0),p&&(e.onDidHideHover?.(),r=void 0)},l=(g,f,p,_)=>new $o(async()=>{(!r||r.isDisposed)&&(r=new g9e(e,p||t,g>0),await r.update(typeof i=="function"?i():i,f,{...n,trapFocus:_}))},g),c=new U;let d=!1;c.add(W(t,se.MOUSE_DOWN,()=>{d=!0,a(!0,!0)},!0)),c.add(W(t,se.MOUSE_UP,()=>{d=!1},!0)),c.add(W(t,se.MOUSE_LEAVE,g=>{d=!1,a(!1,g.fromElement===t)},!0)),c.add(W(t,se.MOUSE_OVER,g=>{if(o)return;const f=new U,p={targetElements:[t],dispose:()=>{}};if(e.placement===void 0||e.placement==="mouse"){const _=b=>{p.x=b.x+10,HO(b,t)||a(!0,!0)};f.add(W(t,se.MOUSE_MOVE,_,!0))}o=f,HO(g,t)&&f.add(l(typeof e.delay=="function"?e.delay(i):e.delay,!1,p))},!0));const h=g=>{if(d||o||!HO(g,t))return;const f={targetElements:[t],dispose:()=>{}},p=new U,_=()=>a(!0,!0);p.add(W(t,se.BLUR,_,!0)),p.add(l(typeof e.delay=="function"?e.delay(i):e.delay,!1,f)),o=p};wl(t)||c.add(W(t,se.FOCUS,h,!0));const u={show:g=>{a(!1,!0),l(0,g,void 0,g)},hide:()=>{a(!0,!0)},update:async(g,f)=>{i=g,await r?.update(i,void 0,f)},dispose:()=>{this._managedHovers.delete(t),c.dispose(),a(!0,!0)}};return this._managedHovers.set(t,u),u}showManagedHover(e){const t=this._managedHovers.get(e);t&&t.show(!0)}dispose(){this._managedHovers.forEach(e=>e.dispose()),super.dispose()}};u9=f9e([rb(0,_e),rb(1,He),rb(2,Nr),rb(3,ft),rb(4,qc),rb(5,Wn)],u9);function zh(s){if(s!==void 0)return s?.id??s}function pY(s){if(!ki(s))return typeof s=="string"?s.toString():s.value}function mY(s){const e=typeof s=="function"?s():s;if(bn(e))return L7(e);if(Rre(e))return e.markdownNotSupportedFallback}function p9e(s,e){function t(i){i?s.setAttribute("title",i):s.removeAttribute("title")}return t(mY(e)),{update:i=>t(mY(i)),show:()=>{},hide:()=>{},dispose:()=>t(void 0)}}class m9e{get anchorPosition(){return this._hover.anchor}constructor(e,t=!1){this._hover=e,this._focus=t,this.layer=1}render(e){return this._hover.render(e),this._focus&&this._hover.focus(),this._hover}getAnchor(){return{x:this._hover.x,y:this._hover.y}}layout(){this._hover.layout()}}function HO(s,e){return ki(s.target)&&_9e(s.target,e)===e}function _9e(s,e){for(e=e??be(s).document.body;!s.hasAttribute("custom-hover")&&s!==e;)s=s.parentElement;return s}tt(so,u9,1);ua((s,e)=>{const t=s.getColor(RW);t&&(e.addRule(`.monaco-hover.workbench-hover .hover-row:not(:first-child):not(:empty) { border-top: 1px solid ${t.transparent(.5)}; }`),e.addRule(`.monaco-hover.workbench-hover hr { border-top: 1px solid ${t.transparent(.5)}; }`))});function lk(s){return Object.isFrozen(s)?s:Cge(s)}class Mn{static createEmptyModel(e){return new Mn({},[],[],void 0,e)}constructor(e,t,i,n,o){this._contents=e,this._keys=t,this._overrides=i,this.raw=n,this.logService=o,this.overrideConfigurations=new Map}get rawConfiguration(){if(!this._rawConfiguration)if(this.raw){const e=(Array.isArray(this.raw)?this.raw:[this.raw]).map(t=>{if(t instanceof Mn)return t;const i=new b9e("",this.logService);return i.parseRaw(t),i.configurationModel});this._rawConfiguration=e.reduce((t,i)=>i===t?i:t.merge(i),e[0])}else this._rawConfiguration=this;return this._rawConfiguration}get contents(){return this._contents}get overrides(){return this._overrides}get keys(){return this._keys}isEmpty(){return this._keys.length===0&&Object.keys(this._contents).length===0&&this._overrides.length===0}getValue(e){return e?u$(this.contents,e):this.contents}inspect(e,t){const i=this;return{get value(){return lk(i.rawConfiguration.getValue(e))},get override(){return t?lk(i.rawConfiguration.getOverrideValue(e,t)):void 0},get merged(){return lk(t?i.rawConfiguration.override(t).getValue(e):i.rawConfiguration.getValue(e))},get overrides(){const n=[];for(const{contents:o,identifiers:r,keys:a}of i.rawConfiguration.overrides){const l=new Mn(o,a,[],void 0,i.logService).getValue(e);l!==void 0&&n.push({identifiers:r,value:l})}return n.length?lk(n):void 0}}}getOverrideValue(e,t){const i=this.getContentsForOverrideIdentifer(t);return i?e?u$(i,e):i:void 0}override(e){let t=this.overrideConfigurations.get(e);return t||(t=this.createOverrideConfigurationModel(e),this.overrideConfigurations.set(e,t)),t}merge(...e){const t=rc(this.contents),i=rc(this.overrides),n=[...this.keys],o=this.raw?Array.isArray(this.raw)?[...this.raw]:[this.raw]:[this];for(const r of e)if(o.push(...r.raw?Array.isArray(r.raw)?r.raw:[r.raw]:[r]),!r.isEmpty()){this.mergeContents(t,r.contents);for(const a of r.overrides){const[l]=i.filter(c=>ni(c.identifiers,a.identifiers));l?(this.mergeContents(l.contents,a.contents),l.keys.push(...a.keys),l.keys=fh(l.keys)):i.push(rc(a))}for(const a of r.keys)n.indexOf(a)===-1&&n.push(a)}return new Mn(t,n,i,!o.length||o.every(r=>r instanceof Mn)?void 0:o,this.logService)}createOverrideConfigurationModel(e){const t=this.getContentsForOverrideIdentifer(e);if(!t||typeof t!="object"||!Object.keys(t).length)return this;const i={};for(const n of fh([...Object.keys(this.contents),...Object.keys(t)])){let o=this.contents[n];const r=t[n];r&&(typeof o=="object"&&typeof r=="object"?(o=rc(o),this.mergeContents(o,r)):o=r),i[n]=o}return new Mn(i,this.keys,this.overrides,void 0,this.logService)}mergeContents(e,t){for(const i of Object.keys(t)){if(i in e&&rn(e[i])&&rn(t[i])){this.mergeContents(e[i],t[i]);continue}e[i]=rc(t[i])}}getContentsForOverrideIdentifer(e){let t=null,i=null;const n=o=>{o&&(i?this.mergeContents(i,o):i=rc(o))};for(const o of this.overrides)o.identifiers.length===1&&o.identifiers[0]===e?t=o.contents:o.identifiers.includes(e)&&n(o.contents);return n(t),i}toJSON(){return{contents:this.contents,overrides:this.overrides,keys:this.keys}}setValue(e,t){this.updateValue(e,t,!1)}removeValue(e){const t=this.keys.indexOf(e);t!==-1&&(this.keys.splice(t,1),Hge(this.contents,e),gg.test(e)&&this.overrides.splice(this.overrides.findIndex(i=>ni(i.identifiers,ME(e))),1))}updateValue(e,t,i){if(cJ(this.contents,e,t,n=>this.logService.error(n)),i=i||this.keys.indexOf(e)===-1,i&&this.keys.push(e),gg.test(e)){const n=ME(e),o={identifiers:n,keys:Object.keys(this.contents[e]),contents:PP(this.contents[e],a=>this.logService.error(a))},r=this.overrides.findIndex(a=>ni(a.identifiers,n));r!==-1?this.overrides[r]=o:this.overrides.push(o)}}}class b9e{constructor(e,t){this._name=e,this.logService=t,this._raw=null,this._configurationModel=null,this._restrictedConfigurations=[]}get configurationModel(){return this._configurationModel||Mn.createEmptyModel(this.logService)}parseRaw(e,t){this._raw=e;const{contents:i,keys:n,overrides:o,restricted:r,hasExcludedProperties:a}=this.doParseRaw(e,t);this._configurationModel=new Mn(i,n,o,a?[e]:void 0,this.logService),this._restrictedConfigurations=r||[]}doParseRaw(e,t){const i=bi.as(Ml.Configuration),n=i.getConfigurationProperties(),o=i.getExcludedConfigurationProperties(),r=this.filter(e,n,o,!0,t);e=r.raw;const a=PP(e,d=>this.logService.error(`Conflict in settings file ${this._name}: ${d}`)),l=Object.keys(e),c=this.toOverrides(e,d=>this.logService.error(`Conflict in settings file ${this._name}: ${d}`));return{contents:a,keys:l,overrides:c,restricted:r.restricted,hasExcludedProperties:r.hasExcludedProperties}}filter(e,t,i,n,o){let r=!1;if(!o?.scopes&&!o?.skipRestricted&&!o?.skipUnregistered&&!o?.exclude?.length)return{raw:e,restricted:[],hasExcludedProperties:r};const a={},l=[];for(const c in e)if(gg.test(c)&&n){const d=this.filter(e[c],t,i,!1,o);a[c]=d.raw,r=r||d.hasExcludedProperties,l.push(...d.restricted)}else{const d=t[c];d?.restricted&&l.push(c),this.shouldInclude(c,d,i,o)?a[c]=e[c]:r=!0}return{raw:a,restricted:l,hasExcludedProperties:r}}shouldInclude(e,t,i,n){if(n.exclude?.includes(e))return!1;if(n.include?.includes(e))return!0;if(n.skipRestricted&&t?.restricted||n.skipUnregistered&&!t)return!1;const o=t??i[e],r=o?typeof o.scope<"u"?o.scope:4:void 0;return r===void 0||n.scopes===void 0?!0:n.scopes.includes(r)}toOverrides(e,t){const i=[];for(const n of Object.keys(e))if(gg.test(n)){const o={};for(const r in e[n])o[r]=e[n][r];i.push({identifiers:ME(n),keys:Object.keys(o),contents:PP(o,t)})}return i}}class w9e{constructor(e,t,i,n,o,r,a,l,c,d,h,u,g){this.key=e,this.overrides=t,this._value=i,this.overrideIdentifiers=n,this.defaultConfiguration=o,this.policyConfiguration=r,this.applicationConfiguration=a,this.userConfiguration=l,this.localUserConfiguration=c,this.remoteUserConfiguration=d,this.workspaceConfiguration=h,this.folderConfigurationModel=u,this.memoryConfigurationModel=g}toInspectValue(e){return e?.value!==void 0||e?.override!==void 0||e?.overrides!==void 0?e:void 0}get userInspectValue(){return this._userInspectValue||(this._userInspectValue=this.userConfiguration.inspect(this.key,this.overrides.overrideIdentifier)),this._userInspectValue}get user(){return this.toInspectValue(this.userInspectValue)}}class s2{constructor(e,t,i,n,o,r,a,l,c,d){this._defaultConfiguration=e,this._policyConfiguration=t,this._applicationConfiguration=i,this._localUserConfiguration=n,this._remoteUserConfiguration=o,this._workspaceConfiguration=r,this._folderConfigurations=a,this._memoryConfiguration=l,this._memoryConfigurationByResource=c,this.logService=d,this._workspaceConsolidatedConfiguration=null,this._foldersConsolidatedConfigurations=new Wi,this._userConfiguration=null}getValue(e,t,i){return this.getConsolidatedConfigurationModel(e,t,i).getValue(e)}updateValue(e,t,i={}){let n;i.resource?(n=this._memoryConfigurationByResource.get(i.resource),n||(n=Mn.createEmptyModel(this.logService),this._memoryConfigurationByResource.set(i.resource,n))):n=this._memoryConfiguration,t===void 0?n.removeValue(e):n.setValue(e,t),i.resource||(this._workspaceConsolidatedConfiguration=null)}inspect(e,t,i){const n=this.getConsolidatedConfigurationModel(e,t,i),o=this.getFolderConfigurationModelForResource(t.resource,i),r=t.resource?this._memoryConfigurationByResource.get(t.resource)||this._memoryConfiguration:this._memoryConfiguration,a=new Set;for(const l of n.overrides)for(const c of l.identifiers)n.getOverrideValue(e,c)!==void 0&&a.add(c);return new w9e(e,t,n.getValue(e),a.size?[...a]:void 0,this._defaultConfiguration,this._policyConfiguration.isEmpty()?void 0:this._policyConfiguration,this.applicationConfiguration.isEmpty()?void 0:this.applicationConfiguration,this.userConfiguration,this.localUserConfiguration,this.remoteUserConfiguration,i?this._workspaceConfiguration:void 0,o||void 0,r)}get applicationConfiguration(){return this._applicationConfiguration}get userConfiguration(){if(!this._userConfiguration)if(this._remoteUserConfiguration.isEmpty())this._userConfiguration=this._localUserConfiguration;else{const e=this._localUserConfiguration.merge(this._remoteUserConfiguration);this._userConfiguration=new Mn(e.contents,e.keys,e.overrides,void 0,this.logService)}return this._userConfiguration}get localUserConfiguration(){return this._localUserConfiguration}get remoteUserConfiguration(){return this._remoteUserConfiguration}getConsolidatedConfigurationModel(e,t,i){let n=this.getConsolidatedConfigurationModelForResource(t,i);if(t.overrideIdentifier&&(n=n.override(t.overrideIdentifier)),!this._policyConfiguration.isEmpty()&&this._policyConfiguration.getValue(e)!==void 0){n=n.merge();for(const o of this._policyConfiguration.keys)n.setValue(o,this._policyConfiguration.getValue(o))}return n}getConsolidatedConfigurationModelForResource({resource:e},t){let i=this.getWorkspaceConsolidatedConfiguration();if(t&&e){const n=t.getFolder(e);n&&(i=this.getFolderConsolidatedConfiguration(n.uri)||i);const o=this._memoryConfigurationByResource.get(e);o&&(i=i.merge(o))}return i}getWorkspaceConsolidatedConfiguration(){return this._workspaceConsolidatedConfiguration||(this._workspaceConsolidatedConfiguration=this._defaultConfiguration.merge(this.applicationConfiguration,this.userConfiguration,this._workspaceConfiguration,this._memoryConfiguration)),this._workspaceConsolidatedConfiguration}getFolderConsolidatedConfiguration(e){let t=this._foldersConsolidatedConfigurations.get(e);if(!t){const i=this.getWorkspaceConsolidatedConfiguration(),n=this._folderConfigurations.get(e);n?(t=i.merge(n),this._foldersConsolidatedConfigurations.set(e,t)):t=i}return t}getFolderConfigurationModelForResource(e,t){if(t&&e){const i=t.getFolder(e);if(i)return this._folderConfigurations.get(i.uri)}}toData(){return{defaults:{contents:this._defaultConfiguration.contents,overrides:this._defaultConfiguration.overrides,keys:this._defaultConfiguration.keys},policy:{contents:this._policyConfiguration.contents,overrides:this._policyConfiguration.overrides,keys:this._policyConfiguration.keys},application:{contents:this.applicationConfiguration.contents,overrides:this.applicationConfiguration.overrides,keys:this.applicationConfiguration.keys,raw:Array.isArray(this.applicationConfiguration.raw)?void 0:this.applicationConfiguration.raw},userLocal:{contents:this.localUserConfiguration.contents,overrides:this.localUserConfiguration.overrides,keys:this.localUserConfiguration.keys,raw:Array.isArray(this.localUserConfiguration.raw)?void 0:this.localUserConfiguration.raw},userRemote:{contents:this.remoteUserConfiguration.contents,overrides:this.remoteUserConfiguration.overrides,keys:this.remoteUserConfiguration.keys,raw:Array.isArray(this.remoteUserConfiguration.raw)?void 0:this.remoteUserConfiguration.raw},workspace:{contents:this._workspaceConfiguration.contents,overrides:this._workspaceConfiguration.overrides,keys:this._workspaceConfiguration.keys},folders:[...this._folderConfigurations.keys()].reduce((e,t)=>{const{contents:i,overrides:n,keys:o}=this._folderConfigurations.get(t);return e.push([t,{contents:i,overrides:n,keys:o}]),e},[])}}static parse(e,t){const i=this.parseConfigurationModel(e.defaults,t),n=this.parseConfigurationModel(e.policy,t),o=this.parseConfigurationModel(e.application,t),r=this.parseConfigurationModel(e.userLocal,t),a=this.parseConfigurationModel(e.userRemote,t),l=this.parseConfigurationModel(e.workspace,t),c=e.folders.reduce((d,h)=>(d.set(Le.revive(h[0]),this.parseConfigurationModel(h[1],t)),d),new Wi);return new s2(i,n,o,r,a,l,c,Mn.createEmptyModel(t),new Wi,t)}static parseConfigurationModel(e,t){return new Mn(e.contents,e.keys,e.overrides,e.raw,t)}}class v9e{constructor(e,t,i,n,o){this.change=e,this.previous=t,this.currentConfiguraiton=i,this.currentWorkspace=n,this.logService=o,this._marker=`
`,this._markerCode1=this._marker.charCodeAt(0),this._markerCode2=46,this.affectedKeys=new Set,this._previousConfiguration=void 0;for(const r of e.keys)this.affectedKeys.add(r);for(const[,r]of e.overrides)for(const a of r)this.affectedKeys.add(a);this._affectsConfigStr=this._marker;for(const r of this.affectedKeys)this._affectsConfigStr+=r+this._marker}get previousConfiguration(){return!this._previousConfiguration&&this.previous&&(this._previousConfiguration=s2.parse(this.previous.data,this.logService)),this._previousConfiguration}affectsConfiguration(e,t){const i=this._marker+e,n=this._affectsConfigStr.indexOf(i);if(n<0)return!1;const o=n+i.length;if(o>=this._affectsConfigStr.length)return!1;const r=this._affectsConfigStr.charCodeAt(o);if(r!==this._markerCode1&&r!==this._markerCode2)return!1;if(t){const a=this.previousConfiguration?this.previousConfiguration.getValue(e,t,this.previous?.workspace):void 0,l=this.currentConfiguraiton.getValue(e,t,this.currentWorkspace);return!Ho(a,l)}return!0}}const _T={kind:0},C9e={kind:1};function y9e(s,e,t){return{kind:2,commandId:s,commandArgs:e,isBubble:t}}class fy{constructor(e,t,i){this._log=i,this._defaultKeybindings=e,this._defaultBoundCommands=new Map;for(const n of e){const o=n.command;o&&o.charAt(0)!=="-"&&this._defaultBoundCommands.set(o,!0)}this._map=new Map,this._lookupMap=new Map,this._keybindings=fy.handleRemovals([].concat(e).concat(t));for(let n=0,o=this._keybindings.length;n<o;n++){const r=this._keybindings[n];if(r.chords.length===0)continue;const a=r.when?.substituteConstants();a&&a.type===0||this._addKeyPress(r.chords[0],r)}}static _isTargetedForRemoval(e,t,i){if(t){for(let n=0;n<t.length;n++)if(t[n]!==e.chords[n])return!1}return!(i&&i.type!==1&&(!e.when||!Pue(i,e.when)))}static handleRemovals(e){const t=new Map;for(let n=0,o=e.length;n<o;n++){const r=e[n];if(r.command&&r.command.charAt(0)==="-"){const a=r.command.substring(1);t.has(a)?t.get(a).push(r):t.set(a,[r])}}if(t.size===0)return e;const i=[];for(let n=0,o=e.length;n<o;n++){const r=e[n];if(!r.command||r.command.length===0){i.push(r);continue}if(r.command.charAt(0)==="-")continue;const a=t.get(r.command);if(!a||!r.isDefault){i.push(r);continue}let l=!1;for(const c of a){const d=c.when;if(this._isTargetedForRemoval(r,c.chords,d)){l=!0;break}}if(!l){i.push(r);continue}}return i}_addKeyPress(e,t){const i=this._map.get(e);if(typeof i>"u"){this._map.set(e,[t]),this._addToLookupMap(t);return}for(let n=i.length-1;n>=0;n--){const o=i[n];if(o.command===t.command)continue;let r=!0;for(let a=1;a<o.chords.length&&a<t.chords.length;a++)if(o.chords[a]!==t.chords[a]){r=!1;break}r&&fy.whenIsEntirelyIncluded(o.when,t.when)&&this._removeFromLookupMap(o)}i.push(t),this._addToLookupMap(t)}_addToLookupMap(e){if(!e.command)return;let t=this._lookupMap.get(e.command);typeof t>"u"?(t=[e],this._lookupMap.set(e.command,t)):t.push(e)}_removeFromLookupMap(e){if(!e.command)return;const t=this._lookupMap.get(e.command);if(!(typeof t>"u")){for(let i=0,n=t.length;i<n;i++)if(t[i]===e){t.splice(i,1);return}}}static whenIsEntirelyIncluded(e,t){return!t||t.type===1?!0:!e||e.type===1?!1:TP(e,t)}getKeybindings(){return this._keybindings}lookupPrimaryKeybinding(e,t,i=!1){const n=this._lookupMap.get(e);if(typeof n>"u"||n.length===0)return null;if(n.length===1&&!i)return n[0];for(let o=n.length-1;o>=0;o--){const r=n[o];if(t.contextMatchesRules(r.when))return r}return i?null:n[n.length-1]}resolve(e,t,i){const n=[...t,i];this._log(`| Resolving ${n}`);const o=this._map.get(n[0]);if(o===void 0)return this._log("\\ No keybinding entries."),_T;let r=null;if(n.length<2)r=o;else{r=[];for(let l=0,c=o.length;l<c;l++){const d=o[l];if(n.length>d.chords.length)continue;let h=!0;for(let u=1;u<n.length;u++)if(d.chords[u]!==n[u]){h=!1;break}h&&r.push(d)}}const a=this._findCommand(e,r);return a?n.length<a.chords.length?(this._log(`\\ From ${r.length} keybinding entries, awaiting ${a.chords.length-n.length} more chord(s), when: ${_Y(a.when)}, source: ${bY(a)}.`),C9e):(this._log(`\\ From ${r.length} keybinding entries, matched ${a.command}, when: ${_Y(a.when)}, source: ${bY(a)}.`),y9e(a.command,a.commandArgs,a.bubble)):(this._log(`\\ From ${r.length} keybinding entries, no when clauses matched the context.`),_T)}_findCommand(e,t){for(let i=t.length-1;i>=0;i--){const n=t[i];if(fy._contextMatchesRules(e,n.when))return n}return null}static _contextMatchesRules(e,t){return t?t.evaluate(e):!0}}function _Y(s){return s?`${s.serialize()}`:"no when condition"}function bY(s){return s.extensionId?s.isBuiltinExtension?`built-in extension ${s.extensionId}`:`user extension ${s.extensionId}`:s.isDefault?"built-in":"user"}const S9e=/^(cursor|delete|undo|redo|tab|editor\.action\.clipboard)/;class x9e extends F{get onDidUpdateKeybindings(){return this._onDidUpdateKeybindings?this._onDidUpdateKeybindings.event:oe.None}get inChordMode(){return this._currentChords.length>0}constructor(e,t,i,n,o){super(),this._contextKeyService=e,this._commandService=t,this._telemetryService=i,this._notificationService=n,this._logService=o,this._onDidUpdateKeybindings=this._register(new M),this._currentChords=[],this._currentChordChecker=new $6,this._currentChordStatusMessage=null,this._ignoreSingleModifiers=zb.EMPTY,this._currentSingleModifier=null,this._currentSingleModifierClearTimeout=new $o,this._currentlyDispatchingCommandId=null,this._logging=!1}dispose(){super.dispose()}_log(e){this._logging&&this._logService.info(`[KeybindingService]: ${e}`)}getKeybindings(){return this._getResolver().getKeybindings()}lookupKeybinding(e,t,i=!1){const n=this._getResolver().lookupPrimaryKeybinding(e,t||this._contextKeyService,i);if(n)return n.resolvedKeybinding}dispatchEvent(e,t){return this._dispatch(e,t)}softDispatch(e,t){this._log("/ Soft dispatching keyboard event");const i=this.resolveKeyboardEvent(e);if(i.hasMultipleChords())return console.warn("keyboard event should not be mapped to multiple chords"),_T;const[n]=i.getDispatchChords();if(n===null)return this._log("\\ Keyboard event cannot be dispatched"),_T;const o=this._contextKeyService.getContext(t),r=this._currentChords.map((({keypress:a})=>a));return this._getResolver().resolve(o,r,n)}_scheduleLeaveChordMode(){const e=Date.now();this._currentChordChecker.cancelAndSet(()=>{if(!this._documentHasFocus()){this._leaveChordMode();return}Date.now()-e>5e3&&this._leaveChordMode()},500)}_expectAnotherChord(e,t){switch(this._currentChords.push({keypress:e,label:t}),this._currentChords.length){case 0:throw uE("impossible");case 1:this._currentChordStatusMessage=this._notificationService.status(m(1701,"({0}) was pressed. Waiting for second key of chord...",t));break;default:{const i=this._currentChords.map(({label:n})=>n).join(", ");this._currentChordStatusMessage=this._notificationService.status(m(1702,"({0}) was pressed. Waiting for next key of chord...",i))}}this._scheduleLeaveChordMode(),kc.enabled&&kc.disable()}_leaveChordMode(){this._currentChordStatusMessage&&(this._currentChordStatusMessage.close(),this._currentChordStatusMessage=null),this._currentChordChecker.cancel(),this._currentChords=[],kc.enable()}_dispatch(e,t){return this._doDispatch(this.resolveKeyboardEvent(e),t,!1)}_singleModifierDispatch(e,t){const i=this.resolveKeyboardEvent(e),[n]=i.getSingleModifierDispatchChords();if(n)return this._ignoreSingleModifiers.has(n)?(this._log(`+ Ignoring single modifier ${n} due to it being pressed together with other keys.`),this._ignoreSingleModifiers=zb.EMPTY,this._currentSingleModifierClearTimeout.cancel(),this._currentSingleModifier=null,!1):(this._ignoreSingleModifiers=zb.EMPTY,this._currentSingleModifier===null?(this._log(`+ Storing single modifier for possible chord ${n}.`),this._currentSingleModifier=n,this._currentSingleModifierClearTimeout.cancelAndSet(()=>{this._log("+ Clearing single modifier due to 300ms elapsed."),this._currentSingleModifier=null},300),!1):n===this._currentSingleModifier?(this._log(`/ Dispatching single modifier chord ${n} ${n}`),this._currentSingleModifierClearTimeout.cancel(),this._currentSingleModifier=null,this._doDispatch(i,t,!0)):(this._log(`+ Clearing single modifier due to modifier mismatch: ${this._currentSingleModifier} ${n}`),this._currentSingleModifierClearTimeout.cancel(),this._currentSingleModifier=null,!1));const[o]=i.getChords();return this._ignoreSingleModifiers=new zb(o),this._currentSingleModifier!==null&&this._log("+ Clearing single modifier due to other key up."),this._currentSingleModifierClearTimeout.cancel(),this._currentSingleModifier=null,!1}_doDispatch(e,t,i=!1){let n=!1;if(e.hasMultipleChords())return console.warn("Unexpected keyboard event mapped to multiple chords"),!1;let o=null,r=null;if(i){const[d]=e.getSingleModifierDispatchChords();o=d,r=d?[d]:[]}else[o]=e.getDispatchChords(),r=this._currentChords.map(({keypress:d})=>d);if(o===null)return this._log("\\ Keyboard event cannot be dispatched in keydown phase."),n;const a=this._contextKeyService.getContext(t),l=e.getLabel(),c=this._getResolver().resolve(a,r,o);switch(c.kind){case 0:{if(this._logService.trace("KeybindingService#dispatch",l,"[ No matching keybinding ]"),this.inChordMode){const d=this._currentChords.map(({label:h})=>h).join(", ");this._log(`+ Leaving multi-chord mode: Nothing bound to "${d}, ${l}".`),this._notificationService.status(m(1703,"The key combination ({0}, {1}) is not a command.",d,l),{hideAfter:10*1e3}),this._leaveChordMode(),n=!0}return n}case 1:return this._logService.trace("KeybindingService#dispatch",l,"[ Several keybindings match - more chords needed ]"),n=!0,this._expectAnotherChord(o,l),this._log(this._currentChords.length===1?"+ Entering multi-chord mode...":"+ Continuing multi-chord mode..."),n;case 2:{if(this._logService.trace("KeybindingService#dispatch",l,`[ Will dispatch command ${c.commandId} ]`),c.commandId===null||c.commandId===""){if(this.inChordMode){const d=this._currentChords.map(({label:h})=>h).join(", ");this._log(`+ Leaving chord mode: Nothing bound to "${d}, ${l}".`),this._notificationService.status(m(1704,"The key combination ({0}, {1}) is not a command.",d,l),{hideAfter:10*1e3}),this._leaveChordMode(),n=!0}}else{this.inChordMode&&this._leaveChordMode(),c.isBubble||(n=!0),this._log(`+ Invoking command ${c.commandId}.`),this._currentlyDispatchingCommandId=c.commandId;try{typeof c.commandArgs>"u"?this._commandService.executeCommand(c.commandId).then(void 0,d=>this._notificationService.warn(d)):this._commandService.executeCommand(c.commandId,c.commandArgs).then(void 0,d=>this._notificationService.warn(d))}finally{this._currentlyDispatchingCommandId=null}S9e.test(c.commandId)||this._telemetryService.publicLog2("workbenchActionExecuted",{id:c.commandId,from:"keybinding",detail:e.getUserSettingsLabel()??void 0})}return n}}}mightProducePrintableCharacter(e){return e.ctrlKey||e.metaKey?!1:e.keyCode>=31&&e.keyCode<=56||e.keyCode>=21&&e.keyCode<=30}}const gR=class gR{constructor(e){this._ctrlKey=e?e.ctrlKey:!1,this._shiftKey=e?e.shiftKey:!1,this._altKey=e?e.altKey:!1,this._metaKey=e?e.metaKey:!1}has(e){switch(e){case"ctrl":return this._ctrlKey;case"shift":return this._shiftKey;case"alt":return this._altKey;case"meta":return this._metaKey}}};gR.EMPTY=new gR(null);let zb=gR;class wY{constructor(e,t,i,n,o,r,a){this._resolvedKeybindingItemBrand=void 0,this.resolvedKeybinding=e,this.chords=e?g9(e.getDispatchChords()):[],e&&this.chords.length===0&&(this.chords=g9(e.getSingleModifierDispatchChords())),this.bubble=t?t.charCodeAt(0)===94:!1,this.command=this.bubble?t.substr(1):t,this.commandArgs=i,this.when=n,this.isDefault=o,this.extensionId=r,this.isBuiltinExtension=a}}function g9(s){const e=[];for(let t=0,i=s.length;t<i;t++){const n=s[t];if(!n)return[];e.push(n)}return e}class L9e extends sde{constructor(e,t){if(super(),t.length===0)throw Qr("chords");this._os=e,this._chords=t}getLabel(){return z7.toLabel(this._os,this._chords,e=>this._getLabel(e))}getAriaLabel(){return Wke.toLabel(this._os,this._chords,e=>this._getAriaLabel(e))}getElectronAccelerator(){return this._chords.length>1||this._chords[0].isDuplicateModifierCase()?null:Hke.toLabel(this._os,this._chords,e=>this._getElectronAccelerator(e))}getUserSettingsLabel(){return Vke.toLabel(this._os,this._chords,e=>this._getUserSettingsLabel(e))}hasMultipleChords(){return this._chords.length>1}getChords(){return this._chords.map(e=>this._getChord(e))}_getChord(e){return new nde(e.ctrlKey,e.shiftKey,e.altKey,e.metaKey,this._getLabel(e),this._getAriaLabel(e))}getDispatchChords(){return this._chords.map(e=>this._getChordDispatch(e))}getSingleModifierDispatchChords(){return this._chords.map(e=>this._getSingleModifierChordDispatch(e))}}class L0 extends L9e{constructor(e,t){super(t,e)}_keyCodeToUILabel(e){if(this._os===2)switch(e){case 15:return"←";case 16:return"↑";case 17:return"→";case 18:return"↓"}return Ad.toString(e)}_getLabel(e){return e.isDuplicateModifierCase()?"":this._keyCodeToUILabel(e.keyCode)}_getAriaLabel(e){return e.isDuplicateModifierCase()?"":Ad.toString(e.keyCode)}_getElectronAccelerator(e){return Ad.toElectronAccelerator(e.keyCode)}_getUserSettingsLabel(e){if(e.isDuplicateModifierCase())return"";const t=Ad.toUserSettingsUS(e.keyCode);return t&&t.toLowerCase()}_getChordDispatch(e){return L0.getDispatchStr(e)}static getDispatchStr(e){if(e.isModifierKey())return null;let t="";return e.ctrlKey&&(t+="ctrl+"),e.shiftKey&&(t+="shift+"),e.altKey&&(t+="alt+"),e.metaKey&&(t+="meta+"),t+=Ad.toString(e.keyCode),t}_getSingleModifierChordDispatch(e){return e.keyCode===5&&!e.shiftKey&&!e.altKey&&!e.metaKey?"ctrl":e.keyCode===4&&!e.ctrlKey&&!e.altKey&&!e.metaKey?"shift":e.keyCode===6&&!e.ctrlKey&&!e.shiftKey&&!e.metaKey?"alt":e.keyCode===57&&!e.ctrlKey&&!e.shiftKey&&!e.altKey?"meta":null}static _scanCodeToKeyCode(e){const t=H6[e];if(t!==-1)return t;switch(e){case 10:return 31;case 11:return 32;case 12:return 33;case 13:return 34;case 14:return 35;case 15:return 36;case 16:return 37;case 17:return 38;case 18:return 39;case 19:return 40;case 20:return 41;case 21:return 42;case 22:return 43;case 23:return 44;case 24:return 45;case 25:return 46;case 26:return 47;case 27:return 48;case 28:return 49;case 29:return 50;case 30:return 51;case 31:return 52;case 32:return 53;case 33:return 54;case 34:return 55;case 35:return 56;case 36:return 22;case 37:return 23;case 38:return 24;case 39:return 25;case 40:return 26;case 41:return 27;case 42:return 28;case 43:return 29;case 44:return 30;case 45:return 21;case 51:return 88;case 52:return 86;case 53:return 92;case 54:return 94;case 55:return 93;case 56:return 0;case 57:return 85;case 58:return 95;case 59:return 91;case 60:return 87;case 61:return 89;case 62:return 90;case 106:return 97}return 0}static _toKeyCodeChord(e){if(!e)return null;if(e instanceof dh)return e;const t=this._scanCodeToKeyCode(e.scanCode);return t===0?null:new dh(e.ctrlKey,e.shiftKey,e.altKey,e.metaKey,t)}static resolveKeybinding(e,t){const i=g9(e.chords.map(n=>this._toKeyCodeChord(n)));return i.length>0?[new L0(i,t)]:[]}}let Pv=[],vV=[],Are=[];function ck(s,e=!1){k9e(s,!1,e)}function k9e(s,e,t){const i=E9e(s,e);Pv.push(i),i.userConfigured?Are.push(i):vV.push(i),t&&!i.userConfigured&&Pv.forEach(n=>{n.mime===i.mime||n.userConfigured||(i.extension&&n.extension===i.extension&&console.warn(`Overwriting extension <<${i.extension}>> to now point to mime <<${i.mime}>>`),i.filename&&n.filename===i.filename&&console.warn(`Overwriting filename <<${i.filename}>> to now point to mime <<${i.mime}>>`),i.filepattern&&n.filepattern===i.filepattern&&console.warn(`Overwriting filepattern <<${i.filepattern}>> to now point to mime <<${i.mime}>>`),i.firstline&&n.firstline===i.firstline&&console.warn(`Overwriting firstline <<${i.firstline}>> to now point to mime <<${i.mime}>>`))})}function E9e(s,e){return{id:s.id,mime:s.mime,filename:s.filename,extension:s.extension,filepattern:s.filepattern,firstline:s.firstline,userConfigured:e,filenameLowercase:s.filename?s.filename.toLowerCase():void 0,extensionLowercase:s.extension?s.extension.toLowerCase():void 0,filepatternLowercase:s.filepattern?ire(s.filepattern.toLowerCase()):void 0,filepatternOnPath:s.filepattern?s.filepattern.indexOf(Zi.sep)>=0:!1}}function I9e(){Pv=Pv.filter(s=>s.userConfigured),vV=[]}function D9e(s,e){return T9e(s,e).map(t=>t.id)}function T9e(s,e){let t;if(s)switch(s.scheme){case Te.file:t=s.fsPath;break;case Te.data:{t=hg.parseMetaData(s).get(hg.META_DATA_LABEL);break}case Te.vscodeNotebookCell:t=void 0;break;default:t=s.path}if(!t)return[{id:"unknown",mime:Ti.unknown}];t=t.toLowerCase();const i=th(t),n=vY(t,i,Are);if(n)return[n,{id:xr,mime:Ti.text}];const o=vY(t,i,vV);if(o)return[o,{id:xr,mime:Ti.text}];if(e){const r=N9e(e);if(r)return[r,{id:xr,mime:Ti.text}]}return[{id:"unknown",mime:Ti.unknown}]}function vY(s,e,t){let i,n,o;for(let r=t.length-1;r>=0;r--){const a=t[r];if(e===a.filenameLowercase){i=a;break}if(a.filepattern&&(!n||a.filepattern.length>n.filepattern.length)){const l=a.filepatternOnPath?s:e;a.filepatternLowercase?.(l)&&(n=a)}a.extension&&(!o||a.extension.length>o.extension.length)&&e.endsWith(a.extensionLowercase)&&(o=a)}if(i)return i;if(n)return n;if(o)return o}function N9e(s){if(X6(s)&&(s=s.substr(1)),s.length>0)for(let e=Pv.length-1;e>=0;e--){const t=Pv[e];if(!t.firstline)continue;const i=s.match(t.firstline);if(i&&i.length>0)return t}}const dk=Object.prototype.hasOwnProperty,CY="vs.editor.nullLanguage";class R9e{constructor(){this._languageIdToLanguage=[],this._languageToLanguageId=new Map,this._register(CY,0),this._register(xr,1),this._nextLanguageId=2}_register(e,t){this._languageIdToLanguage[t]=e,this._languageToLanguageId.set(e,t)}register(e){if(this._languageToLanguageId.has(e))return;const t=this._nextLanguageId++;this._register(e,t)}encodeLanguageId(e){return this._languageToLanguageId.get(e)||0}decodeLanguageId(e){return this._languageIdToLanguage[e]||CY}}const Oy=class Oy extends F{constructor(e=!0,t=!1){super(),this._onDidChange=this._register(new M),this.onDidChange=this._onDidChange.event,Oy.instanceCount++,this._warnOnOverwrite=t,this.languageIdCodec=new R9e,this._dynamicLanguages=[],this._languages={},this._mimeTypesMap={},this._nameMap={},this._lowercaseNameMap={},e&&(this._initializeFromRegistry(),this._register(rv.onDidChangeLanguages(i=>{this._initializeFromRegistry()})))}dispose(){Oy.instanceCount--,super.dispose()}_initializeFromRegistry(){this._languages={},this._mimeTypesMap={},this._nameMap={},this._lowercaseNameMap={},I9e();const e=[].concat(rv.getLanguages()).concat(this._dynamicLanguages);this._registerLanguages(e)}_registerLanguages(e){for(const t of e)this._registerLanguage(t);this._mimeTypesMap={},this._nameMap={},this._lowercaseNameMap={},Object.keys(this._languages).forEach(t=>{const i=this._languages[t];i.name&&(this._nameMap[i.name]=i.identifier),i.aliases.forEach(n=>{this._lowercaseNameMap[n.toLowerCase()]=i.identifier}),i.mimetypes.forEach(n=>{this._mimeTypesMap[n]=i.identifier})}),bi.as(Ml.Configuration).registerOverrideIdentifiers(this.getRegisteredLanguageIds()),this._onDidChange.fire()}_registerLanguage(e){const t=e.id;let i;dk.call(this._languages,t)?i=this._languages[t]:(this.languageIdCodec.register(t),i={identifier:t,name:null,mimetypes:[],aliases:[],extensions:[],filenames:[],configurationFiles:[],icons:[]},this._languages[t]=i),this._mergeLanguage(i,e)}_mergeLanguage(e,t){const i=t.id;let n=null;if(Array.isArray(t.mimetypes)&&t.mimetypes.length>0&&(e.mimetypes.push(...t.mimetypes),n=t.mimetypes[0]),n||(n=`text/x-${i}`,e.mimetypes.push(n)),Array.isArray(t.extensions)){t.configuration?e.extensions=t.extensions.concat(e.extensions):e.extensions=e.extensions.concat(t.extensions);for(const a of t.extensions)ck({id:i,mime:n,extension:a},this._warnOnOverwrite)}if(Array.isArray(t.filenames))for(const a of t.filenames)ck({id:i,mime:n,filename:a},this._warnOnOverwrite),e.filenames.push(a);if(Array.isArray(t.filenamePatterns))for(const a of t.filenamePatterns)ck({id:i,mime:n,filepattern:a},this._warnOnOverwrite);if(typeof t.firstLine=="string"&&t.firstLine.length>0){let a=t.firstLine;a.charAt(0)!=="^"&&(a="^"+a);try{const l=new RegExp(a);eQ(l)||ck({id:i,mime:n,firstline:l},this._warnOnOverwrite)}catch(l){console.warn(`[${t.id}]: Invalid regular expression \`${a}\`: `,l)}}e.aliases.push(i);let o=null;if(typeof t.aliases<"u"&&Array.isArray(t.aliases)&&(t.aliases.length===0?o=[null]:o=t.aliases),o!==null)for(const a of o)!a||a.length===0||e.aliases.push(a);const r=o!==null&&o.length>0;if(!(r&&o[0]===null)){const a=(r?o[0]:null)||i;(r||!e.name)&&(e.name=a)}t.configuration&&e.configurationFiles.push(t.configuration),t.icon&&e.icons.push(t.icon)}isRegisteredLanguageId(e){return e?dk.call(this._languages,e):!1}getRegisteredLanguageIds(){return Object.keys(this._languages)}getLanguageIdByLanguageName(e){const t=e.toLowerCase();return dk.call(this._lowercaseNameMap,t)?this._lowercaseNameMap[t]:null}getLanguageIdByMimeType(e){return e&&dk.call(this._mimeTypesMap,e)?this._mimeTypesMap[e]:null}guessLanguageIdByFilepathOrFirstLine(e,t){return!e&&!t?[]:D9e(e,t)}};Oy.instanceCount=0;let f9=Oy;const Py=class Py extends F{constructor(e=!1){super(),this._onDidRequestBasicLanguageFeatures=this._register(new M),this.onDidRequestBasicLanguageFeatures=this._onDidRequestBasicLanguageFeatures.event,this._onDidRequestRichLanguageFeatures=this._register(new M),this.onDidRequestRichLanguageFeatures=this._onDidRequestRichLanguageFeatures.event,this._onDidChange=this._register(new M({leakWarningThreshold:200})),this.onDidChange=this._onDidChange.event,this._requestedBasicLanguages=new Set,this._requestedRichLanguages=new Set,Py.instanceCount++,this._registry=this._register(new f9(!0,e)),this.languageIdCodec=this._registry.languageIdCodec,this._register(this._registry.onDidChange(()=>this._onDidChange.fire()))}dispose(){Py.instanceCount--,super.dispose()}isRegisteredLanguageId(e){return this._registry.isRegisteredLanguageId(e)}getLanguageIdByLanguageName(e){return this._registry.getLanguageIdByLanguageName(e)}getLanguageIdByMimeType(e){return this._registry.getLanguageIdByMimeType(e)}guessLanguageIdByFilepathOrFirstLine(e,t){return this._registry.guessLanguageIdByFilepathOrFirstLine(e,t).at(0)??null}createById(e){return new yY(this.onDidChange,()=>this._createAndGetLanguageIdentifier(e))}createByFilepathOrFirstLine(e,t){return new yY(this.onDidChange,()=>{const i=this.guessLanguageIdByFilepathOrFirstLine(e,t);return this._createAndGetLanguageIdentifier(i)})}_createAndGetLanguageIdentifier(e){return(!e||!this.isRegisteredLanguageId(e))&&(e=xr),e}requestBasicLanguageFeatures(e){this._requestedBasicLanguages.has(e)||(this._requestedBasicLanguages.add(e),this._onDidRequestBasicLanguageFeatures.fire(e))}requestRichLanguageFeatures(e){this._requestedRichLanguages.has(e)||(this._requestedRichLanguages.add(e),this.requestBasicLanguageFeatures(e),yi.getOrCreate(e),this._onDidRequestRichLanguageFeatures.fire(e))}};Py.instanceCount=0;let p9=Py;class yY{constructor(e,t){this._value=bt(this,e,()=>t()),this.onDidChange=oe.fromObservable(this._value)}get languageId(){return this._value.get()}}const m9=/\(&([^\s&])\)|(^|[^&])&([^\s&])/,VO=/(&amp;)?(&amp;)([^\s&])/g;var bT;(function(s){s[s.Right=0]="Right",s[s.Left=1]="Left"})(bT||(bT={}));var _9;(function(s){s[s.Above=0]="Above",s[s.Below=1]="Below"})(_9||(_9={}));class Rw extends bo{constructor(e,t,i,n){e.classList.add("monaco-menu-container"),e.setAttribute("role","presentation");const o=document.createElement("div");o.classList.add("monaco-menu"),o.setAttribute("role","presentation"),super(o,{orientation:1,actionViewItemProvider:c=>this.doGetActionViewItem(c,i,r),context:i.context,actionRunner:i.actionRunner,ariaLabel:i.ariaLabel,ariaRole:"menu",focusOnlyEnabledItems:!0,triggerKeys:{keys:[3,...Ye||vo?[10]:[]],keyDown:!0}}),this.menuStyles=n,this.menuElement=o,this.actionsList.tabIndex=0,this.initializeOrUpdateStyleSheet(e,n),this._register(_s.addTarget(o)),this._register(W(o,se.KEY_DOWN,c=>{new Pt(c).equals(2)&&c.preventDefault()})),i.enableMnemonics&&this._register(W(o,se.KEY_DOWN,c=>{const d=c.key.toLocaleLowerCase();if(this.mnemonics.has(d)){gt.stop(c,!0);const h=this.mnemonics.get(d);if(h.length===1&&(h[0]instanceof SY&&h[0].container&&this.focusItemByElement(h[0].container),h[0].onClick(c)),h.length>1){const u=h.shift();u&&u.container&&(this.focusItemByElement(u.container),h.push(u)),this.mnemonics.set(d,h)}}})),vo&&this._register(W(o,se.KEY_DOWN,c=>{const d=new Pt(c);d.equals(14)||d.equals(11)?(this.focusedItem=this.viewItems.length-1,this.focusNext(),gt.stop(c,!0)):(d.equals(13)||d.equals(12))&&(this.focusedItem=0,this.focusPrevious(),gt.stop(c,!0))})),this._register(W(this.domNode,se.MOUSE_OUT,c=>{const d=c.relatedTarget;wn(d,this.domNode)||(this.focusedItem=void 0,this.updateFocus(),c.stopPropagation())})),this._register(W(this.actionsList,se.MOUSE_OVER,c=>{let d=c.target;if(!(!d||!wn(d,this.actionsList)||d===this.actionsList)){for(;d.parentElement!==this.actionsList&&d.parentElement!==null;)d=d.parentElement;if(d.classList.contains("action-item")){const h=this.focusedItem;this.setFocusedItem(d),h!==this.focusedItem&&this.updateFocus()}}})),this._register(_s.addTarget(this.actionsList)),this._register(W(this.actionsList,qt.Tap,c=>{let d=c.initialTarget;if(!(!d||!wn(d,this.actionsList)||d===this.actionsList)){for(;d.parentElement!==this.actionsList&&d.parentElement!==null;)d=d.parentElement;if(d.classList.contains("action-item")){const h=this.focusedItem;this.setFocusedItem(d),h!==this.focusedItem&&this.updateFocus()}}}));const r={parent:this};this.mnemonics=new Map,this.scrollableElement=this._register(new tx(o,{alwaysConsumeMouseWheel:!0,horizontal:2,vertical:3,verticalScrollbarSize:7,handleMouseWheel:!0,useShadows:!0}));const a=this.scrollableElement.getDomNode();a.style.position="",this.styleScrollElement(a,n),this._register(W(o,qt.Change,c=>{gt.stop(c,!0);const d=this.scrollableElement.getScrollPosition().scrollTop;this.scrollableElement.setScrollPosition({scrollTop:d-c.translationY})})),this._register(W(a,se.MOUSE_UP,c=>{c.preventDefault()}));const l=be(e);o.style.maxHeight=`${Math.max(10,l.innerHeight-e.getBoundingClientRect().top-35)}px`,t=t.filter((c,d)=>i.submenuIds?.has(c.id)?(console.warn(`Found submenu cycle: ${c.id}`),!1):!(c instanceof tn&&(d===t.length-1||d===0||t[d-1]instanceof tn))),this.push(t,{icon:!0,label:!0,isMenu:!0}),e.appendChild(this.scrollableElement.getDomNode()),this.scrollableElement.scanDomNode(),this.viewItems.filter(c=>!(c instanceof xY)).forEach((c,d,h)=>{c.updatePositionInSet(d+1,h.length)})}initializeOrUpdateStyleSheet(e,t){this.styleSheet||(LE(e)?this.styleSheet=ca(e):(Rw.globalStyleSheet||(Rw.globalStyleSheet=ca()),this.styleSheet=Rw.globalStyleSheet)),this.styleSheet.textContent=M9e(t,LE(e))}styleScrollElement(e,t){const i=t.foregroundColor??"",n=t.backgroundColor??"",o=t.borderColor?`1px solid ${t.borderColor}`:"",r="5px",a=t.shadowColor?`0 2px 8px ${t.shadowColor}`:"";e.style.outline=o,e.style.borderRadius=r,e.style.color=i,e.style.backgroundColor=n,e.style.boxShadow=a}getContainer(){return this.scrollableElement.getDomNode()}get onScroll(){return this.scrollableElement.onScroll}focusItemByElement(e){const t=this.focusedItem;this.setFocusedItem(e),t!==this.focusedItem&&this.updateFocus()}setFocusedItem(e){for(let t=0;t<this.actionsList.children.length;t++){const i=this.actionsList.children[t];if(e===i){this.focusedItem=t;break}}}updateFocus(e){super.updateFocus(e,!0,!0),typeof this.focusedItem<"u"&&this.scrollableElement.setScrollPosition({scrollTop:Math.round(this.menuElement.scrollTop)})}doGetActionViewItem(e,t,i){if(e instanceof tn)return new xY(t.context,e,{icon:!0},this.menuStyles);if(e instanceof ov){const n=new SY(e,e.actions,i,{...t,submenuIds:new Set([...t.submenuIds||[],e.id])},this.menuStyles);if(t.enableMnemonics){const o=n.getMnemonic();if(o&&n.isEnabled()){let r=[];this.mnemonics.has(o)&&(r=this.mnemonics.get(o)),r.push(n),this.mnemonics.set(o,r)}}return n}else{const n=t.getKeyBinding?.(e)?.getLabel(),o={enableMnemonics:t.enableMnemonics,useEventAsContext:t.useEventAsContext,keybinding:n},r=new Mre(t.context,e,o,this.menuStyles);if(t.enableMnemonics){const a=r.getMnemonic();if(a&&r.isEnabled()){let l=[];this.mnemonics.has(a)&&(l=this.mnemonics.get(a)),l.push(r),this.mnemonics.set(a,l)}}return r}}}class Mre extends ll{constructor(e,t,i,n){if(i={...i,isMenu:!0,icon:i.icon!==void 0?i.icon:!1,label:i.label!==void 0?i.label:!0},super(t,t,i),this.menuStyle=n,this.options=i,this.cssClass="",this.options.label&&i.enableMnemonics){const o=this.action.label;if(o){const r=m9.exec(o);r&&(this.mnemonic=(r[1]?r[1]:r[3]).toLocaleLowerCase())}}this.runOnceToEnableMouseUp=new Rt(()=>{this.element&&(this._register(W(this.element,se.MOUSE_UP,o=>{if(gt.stop(o,!0),wo){if(new is(be(this.element),o).rightButton)return;this.onClick(o)}else setTimeout(()=>{this.onClick(o)},0)})),this._register(W(this.element,se.CONTEXT_MENU,o=>{gt.stop(o,!0)})))},100),this._register(this.runOnceToEnableMouseUp)}render(e){super.render(e),this.element&&(this.container=e,this.item=X(this.element,ie("a.action-menu-item")),this._action.id===tn.ID?this.item.setAttribute("role","presentation"):(this.item.setAttribute("role","menuitem"),this.mnemonic&&this.item.setAttribute("aria-keyshortcuts",`${this.mnemonic}`)),this.check=X(this.item,ie("span.menu-item-check"+xe.asCSSSelector(Z.menuSelection))),this.check.setAttribute("role","none"),this.label=X(this.item,ie("span.action-label")),this.options.label&&this.options.keybinding&&(X(this.item,ie("span.keybinding")).textContent=this.options.keybinding),this.runOnceToEnableMouseUp.schedule(),this.updateClass(),this.updateLabel(),this.updateTooltip(),this.updateEnabled(),this.updateChecked(),this.applyStyle())}blur(){super.blur(),this.applyStyle()}focus(){super.focus(),this.item?.focus(),this.applyStyle()}updatePositionInSet(e,t){this.item&&(this.item.setAttribute("aria-posinset",`${e}`),this.item.setAttribute("aria-setsize",`${t}`))}updateLabel(){if(this.label&&this.options.label){Pn(this.label);let e=L7(this.action.label);if(e){const t=A9e(e);this.options.enableMnemonics||(e=t),this.label.setAttribute("aria-label",t.replace(/&&/g,"&"));const i=m9.exec(e);if(i){e=Aa(e),VO.lastIndex=0;let n=VO.exec(e);for(;n&&n[1];)n=VO.exec(e);const o=r=>r.replace(/&amp;&amp;/g,"&amp;");n?this.label.append(U0(o(e.substr(0,n.index))," "),ie("u",{"aria-hidden":"true"},n[3]),QX(o(e.substr(n.index+n[0].length))," ")):this.label.textContent=o(e).trim(),this.item?.setAttribute("aria-keyshortcuts",(i[1]?i[1]:i[3]).toLocaleLowerCase())}else this.label.textContent=e.replace(/&&/g,"&").trim()}}}updateTooltip(){}updateClass(){this.cssClass&&this.item&&this.item.classList.remove(...this.cssClass.split(" ")),this.options.icon&&this.label?(this.cssClass=this.action.class||"",this.label.classList.add("icon"),this.cssClass&&this.label.classList.add(...this.cssClass.split(" ")),this.updateEnabled()):this.label&&this.label.classList.remove("icon")}updateEnabled(){this.action.enabled?(this.element&&(this.element.classList.remove("disabled"),this.element.removeAttribute("aria-disabled")),this.item&&(this.item.classList.remove("disabled"),this.item.removeAttribute("aria-disabled"),this.item.tabIndex=0)):(this.element&&(this.element.classList.add("disabled"),this.element.setAttribute("aria-disabled","true")),this.item&&(this.item.classList.add("disabled"),this.item.setAttribute("aria-disabled","true")))}updateChecked(){if(!this.item)return;const e=this.action.checked;this.item.classList.toggle("checked",!!e),e!==void 0?(this.item.setAttribute("role","menuitemcheckbox"),this.item.setAttribute("aria-checked",e?"true":"false")):(this.item.setAttribute("role","menuitem"),this.item.setAttribute("aria-checked",""))}getMnemonic(){return this.mnemonic}applyStyle(){const e=this.element&&this.element.classList.contains("focused"),t=e&&this.menuStyle.selectionForegroundColor?this.menuStyle.selectionForegroundColor:this.menuStyle.foregroundColor,i=e&&this.menuStyle.selectionBackgroundColor?this.menuStyle.selectionBackgroundColor:void 0,n=e&&this.menuStyle.selectionBorderColor?`1px solid ${this.menuStyle.selectionBorderColor}`:"",o=e&&this.menuStyle.selectionBorderColor?"-1px":"";this.item&&(this.item.style.color=t??"",this.item.style.backgroundColor=i??"",this.item.style.outline=n,this.item.style.outlineOffset=o),this.check&&(this.check.style.color=t??"")}}class SY extends Mre{constructor(e,t,i,n,o){super(e,e,n,o),this.submenuActions=t,this.parentData=i,this.submenuOptions=n,this.mysubmenu=null,this.submenuDisposables=this._register(new U),this.mouseOver=!1,this.expandDirection=n&&n.expandDirection!==void 0?n.expandDirection:{horizontal:bT.Right,vertical:_9.Below},this.showScheduler=new Rt(()=>{this.mouseOver&&(this.cleanupExistingSubmenu(!1),this.createSubmenu(!1))},250),this.hideScheduler=new Rt(()=>{this.element&&!wn(ln(),this.element)&&this.parentData.submenu===this.mysubmenu&&(this.parentData.parent.focus(!1),this.cleanupExistingSubmenu(!0))},750)}render(e){super.render(e),this.element&&(this.item&&(this.item.classList.add("monaco-submenu-item"),this.item.tabIndex=0,this.item.setAttribute("aria-haspopup","true"),this.updateAriaExpanded("false"),this.submenuIndicator=X(this.item,ie("span.submenu-indicator"+xe.asCSSSelector(Z.menuSubmenu))),this.submenuIndicator.setAttribute("aria-hidden","true")),this._register(W(this.element,se.KEY_UP,t=>{const i=new Pt(t);(i.equals(17)||i.equals(3))&&(gt.stop(t,!0),this.createSubmenu(!0))})),this._register(W(this.element,se.KEY_DOWN,t=>{const i=new Pt(t);ln()===this.item&&(i.equals(17)||i.equals(3))&&gt.stop(t,!0)})),this._register(W(this.element,se.MOUSE_OVER,t=>{this.mouseOver||(this.mouseOver=!0,this.showScheduler.schedule())})),this._register(W(this.element,se.MOUSE_LEAVE,t=>{this.mouseOver=!1})),this._register(W(this.element,se.FOCUS_OUT,t=>{this.element&&!wn(ln(),this.element)&&this.hideScheduler.schedule()})),this._register(this.parentData.parent.onScroll(()=>{this.parentData.submenu===this.mysubmenu&&(this.parentData.parent.focus(!1),this.cleanupExistingSubmenu(!0))})))}updateEnabled(){}onClick(e){gt.stop(e,!0),this.cleanupExistingSubmenu(!1),this.createSubmenu(!0)}cleanupExistingSubmenu(e){if(this.parentData.submenu&&(e||this.parentData.submenu!==this.mysubmenu)){try{this.parentData.submenu.dispose()}catch{}this.parentData.submenu=void 0,this.updateAriaExpanded("false"),this.submenuContainer&&(this.submenuDisposables.clear(),this.submenuContainer=void 0)}}calculateSubmenuMenuLayout(e,t,i,n){const o={top:0,left:0};return o.left=Vb(e.width,t.width,{position:n.horizontal===bT.Right?0:1,offset:i.left,size:i.width}),o.left>=i.left&&o.left<i.left+i.width&&(i.left+10+t.width<=e.width&&(o.left=i.left+10),i.top+=10,i.height=0),o.top=Vb(e.height,t.height,{position:0,offset:i.top,size:0}),o.top+t.height===i.top&&o.top+i.height+t.height<=e.height&&(o.top+=i.height),o}createSubmenu(e=!0){if(this.element)if(this.parentData.submenu)this.parentData.submenu.focus(!1);else{this.updateAriaExpanded("true"),this.submenuContainer=X(this.element,ie("div.monaco-submenu")),this.submenuContainer.classList.add("menubar-menu-items-holder","context-view");const t=be(this.parentData.parent.domNode).getComputedStyle(this.parentData.parent.domNode),i=parseFloat(t.paddingTop||"0")||0;this.submenuContainer.style.position="fixed",this.submenuContainer.style.top="0",this.submenuContainer.style.left="0",this.submenuContainer.style.zIndex="1",this.parentData.submenu=new Rw(this.submenuContainer,this.submenuActions.length?this.submenuActions:[new EP],this.submenuOptions,this.menuStyle);const n=this.element.getBoundingClientRect(),o={top:n.top-i,left:n.left,height:n.height+2*i,width:n.width},r=this.submenuContainer.getBoundingClientRect(),a=be(this.element),{top:l,left:c}=this.calculateSubmenuMenuLayout(new St(a.innerWidth,a.innerHeight),St.lift(r),o,this.expandDirection);this.submenuContainer.style.left=`${c-r.left}px`,this.submenuContainer.style.top=`${l-r.top}px`,this.submenuDisposables.add(W(this.submenuContainer,se.KEY_UP,d=>{new Pt(d).equals(15)&&(gt.stop(d,!0),this.parentData.parent.focus(),this.cleanupExistingSubmenu(!0))})),this.submenuDisposables.add(W(this.submenuContainer,se.KEY_DOWN,d=>{new Pt(d).equals(15)&&gt.stop(d,!0)})),this.submenuDisposables.add(this.parentData.submenu.onDidCancel(()=>{this.parentData.parent.focus(),this.cleanupExistingSubmenu(!0)})),this.parentData.submenu.focus(e),this.mysubmenu=this.parentData.submenu}}updateAriaExpanded(e){this.item&&this.item?.setAttribute("aria-expanded",e)}applyStyle(){super.applyStyle();const t=this.element&&this.element.classList.contains("focused")&&this.menuStyle.selectionForegroundColor?this.menuStyle.selectionForegroundColor:this.menuStyle.foregroundColor;this.submenuIndicator&&(this.submenuIndicator.style.color=t??"")}dispose(){super.dispose(),this.hideScheduler.dispose(),this.mysubmenu&&(this.mysubmenu.dispose(),this.mysubmenu=null),this.submenuContainer&&(this.submenuContainer=void 0)}}class xY extends vv{constructor(e,t,i,n){super(e,t,i),this.menuStyles=n}render(e){super.render(e),this.label&&(this.label.style.borderBottomColor=this.menuStyles.separatorColor?`${this.menuStyles.separatorColor}`:"")}}function A9e(s){const e=m9,t=e.exec(s);if(!t)return s;const i=!t[1];return s.replace(e,i?"$2$3":"").trim()}function LY(s){const e=WQ()[s.id];return`.codicon-${s.id}:before { content: '\\${e.toString(16)}'; }`}function M9e(s,e){let t=`
.monaco-menu {
	font-size: 13px;
	border-radius: 5px;
	min-width: 160px;
}

${LY(Z.menuSelection)}
${LY(Z.menuSubmenu)}

.monaco-menu .monaco-action-bar {
	text-align: right;
	overflow: hidden;
	white-space: nowrap;
}

.monaco-menu .monaco-action-bar .actions-container {
	display: flex;
	margin: 0 auto;
	padding: 0;
	width: 100%;
	justify-content: flex-end;
}

.monaco-menu .monaco-action-bar.vertical .actions-container {
	display: inline-block;
}

.monaco-menu .monaco-action-bar.reverse .actions-container {
	flex-direction: row-reverse;
}

.monaco-menu .monaco-action-bar .action-item {
	cursor: pointer;
	display: inline-block;
	transition: transform 50ms ease;
	position: relative;  /* DO NOT REMOVE - this is the key to preventing the ghosting icon bug in Chrome 42 */
}

.monaco-menu .monaco-action-bar .action-item.disabled {
	cursor: default;
}

.monaco-menu .monaco-action-bar .action-item .icon,
.monaco-menu .monaco-action-bar .action-item .codicon {
	display: inline-block;
}

.monaco-menu .monaco-action-bar .action-item .codicon {
	display: flex;
	align-items: center;
}

.monaco-menu .monaco-action-bar .action-label {
	font-size: 11px;
	margin-right: 4px;
}

.monaco-menu .monaco-action-bar .action-item.disabled .action-label,
.monaco-menu .monaco-action-bar .action-item.disabled .action-label:hover {
	color: var(--vscode-disabledForeground);
}

/* Vertical actions */

.monaco-menu .monaco-action-bar.vertical {
	text-align: left;
}

.monaco-menu .monaco-action-bar.vertical .action-item {
	display: block;
}

.monaco-menu .monaco-action-bar.vertical .action-label.separator {
	display: block;
	border-bottom: 1px solid var(--vscode-menu-separatorBackground);
	padding-top: 1px;
	padding: 30px;
}

.monaco-menu .secondary-actions .monaco-action-bar .action-label {
	margin-left: 6px;
}

/* Action Items */
.monaco-menu .monaco-action-bar .action-item.select-container {
	overflow: hidden; /* somehow the dropdown overflows its container, we prevent it here to not push */
	flex: 1;
	max-width: 170px;
	min-width: 60px;
	display: flex;
	align-items: center;
	justify-content: center;
	margin-right: 10px;
}

.monaco-menu .monaco-action-bar.vertical {
	margin-left: 0;
	overflow: visible;
}

.monaco-menu .monaco-action-bar.vertical .actions-container {
	display: block;
}

.monaco-menu .monaco-action-bar.vertical .action-item {
	padding: 0;
	transform: none;
	display: flex;
}

.monaco-menu .monaco-action-bar.vertical .action-item.active {
	transform: none;
}

.monaco-menu .monaco-action-bar.vertical .action-menu-item {
	flex: 1 1 auto;
	display: flex;
	height: 2em;
	align-items: center;
	position: relative;
	margin: 0 4px;
	border-radius: 4px;
}

.monaco-menu .monaco-action-bar.vertical .action-menu-item:hover .keybinding,
.monaco-menu .monaco-action-bar.vertical .action-menu-item:focus .keybinding {
	opacity: unset;
}

.monaco-menu .monaco-action-bar.vertical .action-label {
	flex: 1 1 auto;
	text-decoration: none;
	padding: 0 1em;
	background: none;
	font-size: 12px;
	line-height: 1;
}

.monaco-menu .monaco-action-bar.vertical .keybinding,
.monaco-menu .monaco-action-bar.vertical .submenu-indicator {
	display: inline-block;
	flex: 2 1 auto;
	padding: 0 1em;
	text-align: right;
	font-size: 12px;
	line-height: 1;
	opacity: 0.7;
}

.monaco-menu .monaco-action-bar.vertical .submenu-indicator {
	height: 100%;
}

.monaco-menu .monaco-action-bar.vertical .submenu-indicator.codicon {
	font-size: 16px !important;
	display: flex;
	align-items: center;
}

.monaco-menu .monaco-action-bar.vertical .submenu-indicator.codicon::before {
	margin-left: auto;
	margin-right: -20px;
}

.monaco-menu .monaco-action-bar.vertical .action-item.disabled .keybinding,
.monaco-menu .monaco-action-bar.vertical .action-item.disabled .submenu-indicator {
	opacity: 0.4;
}

.monaco-menu .monaco-action-bar.vertical .action-label:not(.separator) {
	display: inline-block;
	box-sizing: border-box;
	margin: 0;
}

.monaco-menu .monaco-action-bar.vertical .action-item {
	position: static;
	overflow: visible;
}

.monaco-menu .monaco-action-bar.vertical .action-item .monaco-submenu {
	position: absolute;
}

.monaco-menu .monaco-action-bar.vertical .action-label.separator {
	width: 100%;
	height: 0px !important;
	opacity: 1;
}

.monaco-menu .monaco-action-bar.vertical .action-label.separator.text {
	padding: 0.7em 1em 0.1em 1em;
	font-weight: bold;
	opacity: 1;
}

.monaco-menu .monaco-action-bar.vertical .action-label:hover {
	color: inherit;
}

.monaco-menu .monaco-action-bar.vertical .menu-item-check {
	position: absolute;
	visibility: hidden;
	width: 1em;
	height: 100%;
}

.monaco-menu .monaco-action-bar.vertical .action-menu-item.checked .menu-item-check {
	visibility: visible;
	display: flex;
	align-items: center;
	justify-content: center;
}

/* Context Menu */

.context-view.monaco-menu-container {
	outline: 0;
	border: none;
	animation: fadeIn 0.083s linear;
	-webkit-app-region: no-drag;
}

.context-view.monaco-menu-container :focus,
.context-view.monaco-menu-container .monaco-action-bar.vertical:focus,
.context-view.monaco-menu-container .monaco-action-bar.vertical :focus {
	outline: 0;
}

.hc-black .context-view.monaco-menu-container,
.hc-light .context-view.monaco-menu-container,
:host-context(.hc-black) .context-view.monaco-menu-container,
:host-context(.hc-light) .context-view.monaco-menu-container {
	box-shadow: none;
}

.hc-black .monaco-menu .monaco-action-bar.vertical .action-item.focused,
.hc-light .monaco-menu .monaco-action-bar.vertical .action-item.focused,
:host-context(.hc-black) .monaco-menu .monaco-action-bar.vertical .action-item.focused,
:host-context(.hc-light) .monaco-menu .monaco-action-bar.vertical .action-item.focused {
	background: none;
}

/* Vertical Action Bar Styles */

.monaco-menu .monaco-action-bar.vertical {
	padding: 4px 0;
}

.monaco-menu .monaco-action-bar.vertical .action-menu-item {
	height: 2em;
}

.monaco-menu .monaco-action-bar.vertical .action-label:not(.separator),
.monaco-menu .monaco-action-bar.vertical .keybinding {
	font-size: inherit;
	padding: 0 2em;
	max-height: 100%;
}

.monaco-menu .monaco-action-bar.vertical .menu-item-check {
	font-size: inherit;
	width: 2em;
}

.monaco-menu .monaco-action-bar.vertical .action-label.separator {
	font-size: inherit;
	margin: 5px 0 !important;
	padding: 0;
	border-radius: 0;
}

.linux .monaco-menu .monaco-action-bar.vertical .action-label.separator,
:host-context(.linux) .monaco-menu .monaco-action-bar.vertical .action-label.separator {
	margin-left: 0;
	margin-right: 0;
}

.monaco-menu .monaco-action-bar.vertical .submenu-indicator {
	font-size: 60%;
	padding: 0 1.8em;
}

.linux .monaco-menu .monaco-action-bar.vertical .submenu-indicator,
:host-context(.linux) .monaco-menu .monaco-action-bar.vertical .submenu-indicator {
	height: 100%;
	mask-size: 10px 10px;
	-webkit-mask-size: 10px 10px;
}

.monaco-menu .action-item {
	cursor: default;
}`;if(e){t+=`
			/* Arrows */
			.monaco-scrollable-element > .scrollbar > .scra {
				cursor: pointer;
				font-size: 11px !important;
			}

			.monaco-scrollable-element > .visible {
				opacity: 1;

				/* Background rule added for IE9 - to allow clicks on dom node */
				background:rgba(0,0,0,0);

				transition: opacity 100ms linear;
			}
			.monaco-scrollable-element > .invisible {
				opacity: 0;
				pointer-events: none;
			}
			.monaco-scrollable-element > .invisible.fade {
				transition: opacity 800ms linear;
			}

			/* Scrollable Content Inset Shadow */
			.monaco-scrollable-element > .shadow {
				position: absolute;
				display: none;
			}
			.monaco-scrollable-element > .shadow.top {
				display: block;
				top: 0;
				left: 3px;
				height: 3px;
				width: 100%;
			}
			.monaco-scrollable-element > .shadow.left {
				display: block;
				top: 3px;
				left: 0;
				height: 100%;
				width: 3px;
			}
			.monaco-scrollable-element > .shadow.top-left-corner {
				display: block;
				top: 0;
				left: 0;
				height: 3px;
				width: 3px;
			}
			/* Fix for https://github.com/microsoft/vscode/issues/103170 */
			.monaco-menu .action-item .monaco-submenu {
				z-index: 1;
			}
		`;const i=s.scrollbarShadow;i&&(t+=`
				.monaco-scrollable-element > .shadow.top {
					box-shadow: ${i} 0 6px 6px -6px inset;
				}

				.monaco-scrollable-element > .shadow.left {
					box-shadow: ${i} 6px 0 6px -6px inset;
				}

				.monaco-scrollable-element > .shadow.top.left {
					box-shadow: ${i} 6px 6px 6px -6px inset;
				}
			`);const n=s.scrollbarSliderBackground;n&&(t+=`
				.monaco-scrollable-element > .scrollbar > .slider {
					background: ${n};
				}
			`);const o=s.scrollbarSliderHoverBackground;o&&(t+=`
				.monaco-scrollable-element > .scrollbar > .slider:hover {
					background: ${o};
				}
			`);const r=s.scrollbarSliderActiveBackground;r&&(t+=`
				.monaco-scrollable-element > .scrollbar > .slider.active {
					background: ${r};
				}
			`)}return t}class O9e{constructor(e,t,i,n){this.contextViewService=e,this.telemetryService=t,this.notificationService=i,this.keybindingService=n,this.focusToReturn=null,this.lastContainer=null,this.block=null,this.blockDisposable=null,this.options={blockMouse:!0}}configure(e){this.options=e}showContextMenu(e){const t=e.getActions();if(!t.length)return;this.focusToReturn=ln();let i;const n=ki(e.domForShadowRoot)?e.domForShadowRoot:void 0;this.contextViewService.showContextView({getAnchor:()=>e.getAnchor(),canRelayout:!1,anchorAlignment:e.anchorAlignment,anchorAxisAlignment:e.anchorAxisAlignment,layer:e.layer,render:o=>{this.lastContainer=o;const r=e.getMenuClassName?e.getMenuClassName():"";r&&(o.className+=" "+r),this.options.blockMouse&&(this.block=o.appendChild(ie(".context-view-block")),this.block.style.position="fixed",this.block.style.cursor="initial",this.block.style.left="0",this.block.style.top="0",this.block.style.width="100%",this.block.style.height="100%",this.block.style.zIndex="-1",this.blockDisposable?.dispose(),this.blockDisposable=W(this.block,se.MOUSE_DOWN,d=>d.stopPropagation()));const a=new U,l=e.actionRunner||a.add(new Lm);l.onWillRun(d=>this.onActionRun(d,!e.skipTelemetry),this,a),l.onDidRun(this.onDidActionRun,this,a),i=new Rw(o,t,{actionViewItemProvider:e.getActionViewItem,context:e.getActionsContext?e.getActionsContext():null,actionRunner:l,getKeyBinding:e.getKeyBinding?e.getKeyBinding:d=>this.keybindingService.lookupKeybinding(d.id)},Qke),i.onDidCancel(()=>this.contextViewService.hideContextView(!0),null,a),i.onDidBlur(()=>this.contextViewService.hideContextView(!0),null,a);const c=be(o);return a.add(W(c,se.BLUR,()=>this.contextViewService.hideContextView(!0))),a.add(W(c,se.MOUSE_DOWN,d=>{if(d.defaultPrevented)return;const h=new is(c,d);let u=h.target;if(!h.rightButton){for(;u;){if(u===o)return;u=u.parentElement}this.contextViewService.hideContextView(!0)}})),Ra(a,i)},focus:()=>{i?.focus(!!e.autoSelectFirstItem)},onHide:o=>{e.onHide?.(!!o),this.block&&(this.block.remove(),this.block=null),this.blockDisposable?.dispose(),this.blockDisposable=null,this.lastContainer&&(ln()===this.lastContainer||wn(ln(),this.lastContainer))&&this.focusToReturn?.focus(),this.lastContainer=null}},n,!!n)}onActionRun(e,t){t&&this.telemetryService.publicLog2("workbenchActionExecuted",{id:e.action.id,from:"contextMenu"}),this.contextViewService.hideContextView(!1)}onDidActionRun(e){e.error&&!Dr(e.error)&&this.notificationService.error(e.error)}}var P9e=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},ab=function(s,e){return function(t,i){e(t,i,s)}};let b9=class extends F{get contextMenuHandler(){return this._contextMenuHandler||(this._contextMenuHandler=new O9e(this.contextViewService,this.telemetryService,this.notificationService,this.keybindingService)),this._contextMenuHandler}constructor(e,t,i,n,o,r){super(),this.telemetryService=e,this.notificationService=t,this.contextViewService=i,this.keybindingService=n,this.menuService=o,this.contextKeyService=r,this._contextMenuHandler=void 0,this._onDidShowContextMenu=this._store.add(new M),this.onDidShowContextMenu=this._onDidShowContextMenu.event,this._onDidHideContextMenu=this._store.add(new M),this.onDidHideContextMenu=this._onDidHideContextMenu.event}configure(e){this.contextMenuHandler.configure(e)}showContextMenu(e){e=w9.transform(e,this.menuService,this.contextKeyService),this.contextMenuHandler.showContextMenu({...e,onHide:t=>{e.onHide?.(t),this._onDidHideContextMenu.fire()}}),Vd.getInstance().resetKeyStatus(),this._onDidShowContextMenu.fire()}};b9=P9e([ab(0,vs),ab(1,Ii),ab(2,Rh),ab(3,ft),ab(4,da),ab(5,Ae)],b9);var w9;(function(s){function e(i){return i&&i.menuId instanceof fe}function t(i,n,o){if(!e(i))return i;const{menuId:r,menuActionOptions:a,contextKeyService:l}=i;return{...i,getActions:()=>{let c=[];if(r){const d=n.getMenuActions(r,l??o,a);c=iEe(d)}return i.getActions?tn.join(i.getActions(),c):c}}}s.transform=t})(w9||(w9={}));var wT;(function(s){s[s.API=0]="API",s[s.USER=1]="USER"})(wT||(wT={}));var CV=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},vT=function(s,e){return function(t,i){e(t,i,s)}};let v9=class{constructor(e){this._commandService=e}async open(e,t){if(!kR(e,Te.command))return!1;if(!t?.allowCommands||(typeof e=="string"&&(e=Le.parse(e)),Array.isArray(t.allowCommands)&&!t.allowCommands.includes(e.path)))return!0;let i=[];try{i=JF(decodeURIComponent(e.query))}catch{try{i=JF(e.query)}catch{}}return Array.isArray(i)||(i=[i]),await this._commandService.executeCommand(e.path,...i),!0}};v9=CV([vT(0,Gt)],v9);let C9=class{constructor(e){this._editorService=e}async open(e,t){typeof e=="string"&&(e=Le.parse(e));const{selection:i,uri:n}=YEe(e);return e=n,e.scheme===Te.file&&(e=Jue(e)),await this._editorService.openCodeEditor({resource:e,options:{selection:i,source:t?.fromUserGesture?wT.USER:wT.API,...t?.editorOptions}},this._editorService.getFocusedCodeEditor(),t?.openToSide),!0}};C9=CV([vT(0,ht)],C9);let y9=class{constructor(e,t){this._openers=new Ns,this._validators=new Ns,this._resolvers=new Ns,this._resolvedUriTargets=new Wi(i=>i.with({path:null,fragment:null,query:null}).toString()),this._externalOpeners=new Ns,this._defaultExternalOpener={openExternal:async i=>(fP(i,Te.http,Te.https)?OQ(i):Nt.location.href=i,!0)},this._openers.push({open:async(i,n)=>n?.openExternal||fP(i,Te.mailto,Te.http,Te.https,Te.vsls)?(await this._doOpenExternal(i,n),!0):!1}),this._openers.push(new v9(t)),this._openers.push(new C9(e))}registerOpener(e){return{dispose:this._openers.unshift(e)}}async open(e,t){if(!t?.skipValidation){const i=typeof e=="string"?Le.parse(e):e,n=this._resolvedUriTargets.get(i)??e;for(const o of this._validators)if(!await o.shouldOpen(n,t))return!1}for(const i of this._openers)if(await i.open(e,t))return!0;return!1}async resolveExternalUri(e,t){for(const i of this._resolvers)try{const n=await i.resolveExternalUri(e,t);if(n)return this._resolvedUriTargets.has(n.resolved)||this._resolvedUriTargets.set(n.resolved,e),n}catch{}throw new Error("Could not resolve external URI: "+e.toString())}async _doOpenExternal(e,t){const i=typeof e=="string"?Le.parse(e):e;let n;try{n=(await this.resolveExternalUri(i,t)).resolved}catch{n=i}let o;if(typeof e=="string"&&i.toString()===n.toString()?o=e:o=encodeURI(n.toString(!0)),t?.allowContributedOpeners){const r=typeof t?.allowContributedOpeners=="string"?t?.allowContributedOpeners:void 0;for(const a of this._externalOpeners)if(await a.openExternal(o,{sourceUri:i,preferredOpenerId:r},Ke.None))return!0}return this._defaultExternalOpener.openExternal(o,{sourceUri:i},Ke.None)}dispose(){this._validators.clear()}};y9=CV([vT(0,ht),vT(1,Gt)],y9);var F9e=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},kY=function(s,e){return function(t,i){e(t,i,s)}};let S9=class extends F{constructor(e,t){super(),this._markerService=t,this._onDidChangeMarker=this._register(new M),this._suppressedRanges=new Wi,this._markerDecorations=new Wi,e.getModels().forEach(i=>this._onModelAdded(i)),this._register(e.onModelAdded(this._onModelAdded,this)),this._register(e.onModelRemoved(this._onModelRemoved,this)),this._register(this._markerService.onMarkerChanged(this._handleMarkerChange,this))}dispose(){super.dispose(),this._markerDecorations.forEach(e=>e.dispose()),this._markerDecorations.clear()}getMarker(e,t){const i=this._markerDecorations.get(e);return i&&i.getMarker(t)||null}_handleMarkerChange(e){e.forEach(t=>{const i=this._markerDecorations.get(t);i&&this._updateDecorations(i)})}_onModelAdded(e){const t=new B9e(e);this._markerDecorations.set(e.uri,t),this._updateDecorations(t)}_onModelRemoved(e){const t=this._markerDecorations.get(e.uri);t&&(t.dispose(),this._markerDecorations.delete(e.uri)),(e.uri.scheme===Te.inMemory||e.uri.scheme===Te.internal||e.uri.scheme===Te.vscode)&&this._markerService?.read({resource:e.uri}).map(i=>i.owner).forEach(i=>this._markerService.remove(i,[e.uri]))}_updateDecorations(e){let t=this._markerService.read({resource:e.model.uri,take:500});const i=this._suppressedRanges.get(e.model.uri);i&&(t=t.filter(n=>!ot.some(i,o=>k.areIntersectingOrTouching(o,n)))),e.update(t)&&this._onDidChangeMarker.fire(e.model)}};S9=F9e([kY(0,hi),kY(1,jc)],S9);class B9e extends F{constructor(e){super(),this.model=e,this._map=new Gue,this._register(pe(()=>{this.model.deltaDecorations([...this._map.values()],[]),this._map.clear()}))}update(e){const{added:t,removed:i}=mye(new Set(this._map.keys()),new Set(e));if(t.length===0&&i.length===0)return!1;const n=i.map(a=>this._map.get(a)),o=t.map(a=>({range:this._createDecorationRange(this.model,a),options:this._createDecorationOption(a)})),r=this.model.deltaDecorations(n,o);for(const a of i)this._map.delete(a);for(let a=0;a<r.length;a++)this._map.set(t[a],r[a]);return!0}getMarker(e){return this._map.getKey(e.id)}_createDecorationRange(e,t){let i=k.lift(t);if(t.severity===mi.Hint&&!this._hasMarkerTag(t,1)&&!this._hasMarkerTag(t,2)&&(i=i.setEndPosition(i.startLineNumber,i.startColumn+2)),i=e.validateRange(i),i.isEmpty()){const n=e.getLineLastNonWhitespaceColumn(i.startLineNumber)||e.getLineMaxColumn(i.startLineNumber);if(n===1||i.endColumn>=n)return i;const o=e.getWordAtPosition(i.getStartPosition());o&&(i=new k(i.startLineNumber,o.startColumn,i.endLineNumber,o.endColumn))}else if(t.endColumn===Number.MAX_VALUE&&t.startColumn===1&&i.startLineNumber===i.endLineNumber){const n=e.getLineFirstNonWhitespaceColumn(t.startLineNumber);n<i.endColumn&&(i=new k(i.startLineNumber,n,i.endLineNumber,i.endColumn),t.startColumn=n)}return i}_createDecorationOption(e){let t,i,n,o,r;switch(e.severity){case mi.Hint:this._hasMarkerTag(e,2)?t=void 0:this._hasMarkerTag(e,1)?t="squiggly-unnecessary":t="squiggly-hint",n=0;break;case mi.Info:t="squiggly-info",i=Ci(fwe),n=10,r={color:Ci(Lme),position:1};break;case mi.Warning:t="squiggly-warning",i=Ci(gwe),n=20,r={color:Ci(kme),position:1};break;case mi.Error:default:t="squiggly-error",i=Ci(uwe),n=30,r={color:Ci(Eme),position:1};break}return e.tags&&(e.tags.indexOf(1)!==-1&&(o="squiggly-inline-unnecessary"),e.tags.indexOf(2)!==-1&&(o="squiggly-inline-deprecated")),{description:"marker-decoration",stickiness:1,className:t,showIfCollapsed:!0,overviewRuler:{color:i,position:Sr.Right},minimap:r,zIndex:n,inlineClassName:o}}_hasMarkerTag(e,t){return e.tags?e.tags.indexOf(t)>=0:!1}}var W9e=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},hk=function(s,e){return function(t,i){e(t,i,s)}},bb;function Qg(s){return s.toString()}class H9e{constructor(e,t,i){this.model=e,this._modelEventListeners=new U,this.model=e,this._modelEventListeners.add(e.onWillDispose(()=>t(e))),this._modelEventListeners.add(e.onDidChangeLanguage(n=>i(e,n)))}dispose(){this._modelEventListeners.dispose()}}const V9e=vo||Ye?1:2;class z9e{constructor(e,t,i,n,o,r,a,l){this.uri=e,this.initialUndoRedoSnapshot=t,this.time=i,this.sharesUndoRedoStack=n,this.heapSize=o,this.sha1=r,this.versionId=a,this.alternativeVersionId=l}}var gm;let x9=(gm=class extends F{constructor(e,t,i,n){super(),this._configurationService=e,this._resourcePropertiesService=t,this._undoRedoService=i,this._instantiationService=n,this._onModelAdded=this._register(new M),this.onModelAdded=this._onModelAdded.event,this._onModelRemoved=this._register(new M),this.onModelRemoved=this._onModelRemoved.event,this._onModelModeChanged=this._register(new M),this.onModelLanguageChanged=this._onModelModeChanged.event,this._modelCreationOptionsByLanguageAndResource=Object.create(null),this._models={},this._disposedModels=new Map,this._disposedModelsHeapSize=0,this._register(this._configurationService.onDidChangeConfiguration(o=>this._updateModelOptions(o))),this._updateModelOptions(void 0)}static _readModelOptions(e,t){let i=Qn.tabSize;e.editor&&typeof e.editor.tabSize<"u"&&(i=Dd(e.editor.tabSize,Qn.tabSize,1,100));let n="tabSize";e.editor&&typeof e.editor.indentSize<"u"&&e.editor.indentSize!=="tabSize"&&(n=Dd(e.editor.indentSize,"tabSize",1,100));let o=Qn.insertSpaces;e.editor&&typeof e.editor.insertSpaces<"u"&&(o=e.editor.insertSpaces==="false"?!1:!!e.editor.insertSpaces);let r=V9e;const a=e.eol;a===`\r
`?r=2:a===`
`&&(r=1);let l=Qn.trimAutoWhitespace;e.editor&&typeof e.editor.trimAutoWhitespace<"u"&&(l=e.editor.trimAutoWhitespace==="false"?!1:!!e.editor.trimAutoWhitespace);let c=Qn.detectIndentation;e.editor&&typeof e.editor.detectIndentation<"u"&&(c=e.editor.detectIndentation==="false"?!1:!!e.editor.detectIndentation);let d=Qn.largeFileOptimizations;e.editor&&typeof e.editor.largeFileOptimizations<"u"&&(d=e.editor.largeFileOptimizations==="false"?!1:!!e.editor.largeFileOptimizations);let h=Qn.bracketPairColorizationOptions;if(e.editor?.bracketPairColorization&&typeof e.editor.bracketPairColorization=="object"){const u=e.editor.bracketPairColorization;h={enabled:!!u.enabled,independentColorPoolPerBracketType:!!u.independentColorPoolPerBracketType}}return{isForSimpleWidget:t,tabSize:i,indentSize:n,insertSpaces:o,detectIndentation:c,defaultEOL:r,trimAutoWhitespace:l,largeFileOptimizations:d,bracketPairColorizationOptions:h}}_getEOL(e,t){if(e)return this._resourcePropertiesService.getEOL(e,t);const i=this._configurationService.getValue("files.eol",{overrideIdentifier:t});return i&&typeof i=="string"&&i!=="auto"?i:Fo===3||Fo===2?`
`:`\r
`}_shouldRestoreUndoStack(){const e=this._configurationService.getValue("files.restoreUndoStack");return typeof e=="boolean"?e:!0}getCreationOptions(e,t,i){const n=typeof e=="string"?e:e.languageId;let o=this._modelCreationOptionsByLanguageAndResource[n+t];if(!o){const r=this._configurationService.getValue("editor",{overrideIdentifier:n,resource:t}),a=this._getEOL(t,n);o=bb._readModelOptions({editor:r,eol:a},i),this._modelCreationOptionsByLanguageAndResource[n+t]=o}return o}_updateModelOptions(e){const t=this._modelCreationOptionsByLanguageAndResource;this._modelCreationOptionsByLanguageAndResource=Object.create(null);const i=Object.keys(this._models);for(let n=0,o=i.length;n<o;n++){const r=i[n],a=this._models[r],l=a.model.getLanguageId(),c=a.model.uri;if(e&&!e.affectsConfiguration("editor",{overrideIdentifier:l,resource:c})&&!e.affectsConfiguration("files.eol",{overrideIdentifier:l,resource:c}))continue;const d=t[l+c],h=this.getCreationOptions(l,c,a.model.isForSimpleWidget);bb._setModelOptionsForModel(a.model,h,d)}}static _setModelOptionsForModel(e,t,i){i&&i.defaultEOL!==t.defaultEOL&&e.getLineCount()===1&&e.setEOL(t.defaultEOL===1?0:1),!(i&&i.detectIndentation===t.detectIndentation&&i.insertSpaces===t.insertSpaces&&i.tabSize===t.tabSize&&i.indentSize===t.indentSize&&i.trimAutoWhitespace===t.trimAutoWhitespace&&Ho(i.bracketPairColorizationOptions,t.bracketPairColorizationOptions))&&(t.detectIndentation?(e.detectIndentation(t.insertSpaces,t.tabSize),e.updateOptions({trimAutoWhitespace:t.trimAutoWhitespace,bracketColorizationOptions:t.bracketPairColorizationOptions})):e.updateOptions({insertSpaces:t.insertSpaces,tabSize:t.tabSize,indentSize:t.indentSize,trimAutoWhitespace:t.trimAutoWhitespace,bracketColorizationOptions:t.bracketPairColorizationOptions}))}_insertDisposedModel(e){this._disposedModels.set(Qg(e.uri),e),this._disposedModelsHeapSize+=e.heapSize}_removeDisposedModel(e){const t=this._disposedModels.get(Qg(e));return t&&(this._disposedModelsHeapSize-=t.heapSize),this._disposedModels.delete(Qg(e)),t}_ensureDisposedModelsHeapSize(e){if(this._disposedModelsHeapSize>e){const t=[];for(this._disposedModels.forEach(i=>{i.sharesUndoRedoStack||t.push(i)}),t.sort((i,n)=>i.time-n.time);t.length>0&&this._disposedModelsHeapSize>e;){const i=t.shift();this._removeDisposedModel(i.uri),i.initialUndoRedoSnapshot!==null&&this._undoRedoService.restoreSnapshot(i.initialUndoRedoSnapshot)}}}_createModelData(e,t,i,n){const o=this.getCreationOptions(t,i,n),r=this._instantiationService.createInstance(Rm,e,t,o,i);if(i&&this._disposedModels.has(Qg(i))){const c=this._removeDisposedModel(i),d=this._undoRedoService.getElements(i),h=this._getSHA1Computer(),u=h.canComputeSHA1(r)?h.computeSHA1(r)===c.sha1:!1;if(u||c.sharesUndoRedoStack){for(const g of d.past)Nd(g)&&g.matchesResource(i)&&g.setModel(r);for(const g of d.future)Nd(g)&&g.matchesResource(i)&&g.setModel(r);this._undoRedoService.setElementsValidFlag(i,!0,g=>Nd(g)&&g.matchesResource(i)),u&&(r._overwriteVersionId(c.versionId),r._overwriteAlternativeVersionId(c.alternativeVersionId),r._overwriteInitialUndoRedoSnapshot(c.initialUndoRedoSnapshot))}else c.initialUndoRedoSnapshot!==null&&this._undoRedoService.restoreSnapshot(c.initialUndoRedoSnapshot)}const a=Qg(r.uri);if(this._models[a])throw new Error("ModelService: Cannot add model because it already exists!");const l=new H9e(r,c=>this._onWillDispose(c),(c,d)=>this._onDidChangeLanguage(c,d));return this._models[a]=l,l}createModel(e,t,i,n=!1){let o;return t?o=this._createModelData(e,t,i,n):o=this._createModelData(e,xr,i,n),this._onModelAdded.fire(o.model),o.model}getModels(){const e=[],t=Object.keys(this._models);for(let i=0,n=t.length;i<n;i++){const o=t[i];e.push(this._models[o].model)}return e}getModel(e){const t=Qg(e),i=this._models[t];return i?i.model:null}_schemaShouldMaintainUndoRedoElements(e){return e.scheme===Te.file||e.scheme===Te.vscodeRemote||e.scheme===Te.vscodeUserData||e.scheme===Te.vscodeNotebookCell||e.scheme==="fake-fs"}_onWillDispose(e){const t=Qg(e.uri),i=this._models[t],n=this._undoRedoService.getUriComparisonKey(e.uri)!==e.uri.toString();let o=!1,r=0;if(n||this._shouldRestoreUndoStack()&&this._schemaShouldMaintainUndoRedoElements(e.uri)){const c=this._undoRedoService.getElements(e.uri);if(c.past.length>0||c.future.length>0){for(const d of c.past)Nd(d)&&d.matchesResource(e.uri)&&(o=!0,r+=d.heapSize(e.uri),d.setModel(e.uri));for(const d of c.future)Nd(d)&&d.matchesResource(e.uri)&&(o=!0,r+=d.heapSize(e.uri),d.setModel(e.uri))}}const a=bb.MAX_MEMORY_FOR_CLOSED_FILES_UNDO_STACK,l=this._getSHA1Computer();if(o)if(!n&&(r>a||!l.canComputeSHA1(e))){const c=i.model.getInitialUndoRedoSnapshot();c!==null&&this._undoRedoService.restoreSnapshot(c)}else this._ensureDisposedModelsHeapSize(a-r),this._undoRedoService.setElementsValidFlag(e.uri,!1,c=>Nd(c)&&c.matchesResource(e.uri)),this._insertDisposedModel(new z9e(e.uri,i.model.getInitialUndoRedoSnapshot(),Date.now(),n,r,l.computeSHA1(e),e.getVersionId(),e.getAlternativeVersionId()));else if(!n){const c=i.model.getInitialUndoRedoSnapshot();c!==null&&this._undoRedoService.restoreSnapshot(c)}delete this._models[t],i.dispose(),delete this._modelCreationOptionsByLanguageAndResource[e.getLanguageId()+e.uri],this._onModelRemoved.fire(e)}_onDidChangeLanguage(e,t){const i=t.oldLanguage,n=e.getLanguageId(),o=this.getCreationOptions(i,e.uri,e.isForSimpleWidget),r=this.getCreationOptions(n,e.uri,e.isForSimpleWidget);bb._setModelOptionsForModel(e,r,o),this._onModelModeChanged.fire({model:e,oldLanguageId:i})}_getSHA1Computer(){return new L9}},bb=gm,gm.MAX_MEMORY_FOR_CLOSED_FILES_UNDO_STACK=20*1024*1024,gm);x9=bb=W9e([hk(0,He),hk(1,Woe),hk(2,f7),hk(3,_e)],x9);const fR=class fR{canComputeSHA1(e){return e.getValueLength()<=fR.MAX_MODEL_SIZE}computeSHA1(e){const t=new _P,i=e.createSnapshot();let n;for(;n=i.read();)t.update(n);return t.digest()}};fR.MAX_MODEL_SIZE=10*1024*1024;let L9=fR;var $9e=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},EY=function(s,e){return function(t,i){e(t,i,s)}};let k9=class extends F{constructor(e,t){super(),this.quickInputService=e,this.instantiationService=t,this.registry=bi.as(y_.Quickaccess),this.mapProviderToDescriptor=new Map,this.lastAcceptedPickerValues=new Map,this.visibleQuickAccess=void 0,this._register(pe(()=>{for(const i of this.mapProviderToDescriptor.values())i_(i)&&i.dispose();this.visibleQuickAccess?.picker.dispose()}))}show(e="",t){this.doShowOrPick(e,!1,t)}doShowOrPick(e,t,i){const[n,o]=this.getOrInstantiateProvider(e,i?.enabledProviderPrefixes),r=this.visibleQuickAccess,a=r?.descriptor;if(r&&o&&a===o){e!==o.prefix&&!i?.preserveValue&&(r.picker.value=e),this.adjustValueSelection(r.picker,o,i);return}if(o&&!i?.preserveValue){let f;if(r&&a&&a!==o){const p=r.value.substr(a.prefix.length);p&&(f=`${o.prefix}${p}`)}if(!f){const p=n?.defaultFilterValue;p===Y8.LAST?f=this.lastAcceptedPickerValues.get(o):typeof p=="string"&&(f=`${o.prefix}${p}`)}typeof f=="string"&&(e=f)}const l=r?.picker?.valueSelection,c=r?.picker?.value,d=new U,h=d.add(this.quickInputService.createQuickPick({useSeparators:!0}));h.value=e,this.adjustValueSelection(h,o,i),h.placeholder=i?.placeholder??o?.placeholder,h.quickNavigate=i?.quickNavigateConfiguration,h.hideInput=!!h.quickNavigate&&!r,(typeof i?.itemActivation=="number"||i?.quickNavigateConfiguration)&&(h.itemActivation=i?.itemActivation??nl.SECOND),h.contextKey=o?.contextKey,h.filterValue=f=>f.substring(o?o.prefix.length:0);let u;t&&(u=new n_,d.add(oe.once(h.onWillAccept)(f=>{f.veto(),h.hide()}))),d.add(this.registerPickerListeners(h,n,o,e,i));const g=d.add(new si);if(n&&d.add(n.provide(h,g.token,i?.providerOptions)),oe.once(h.onDidHide)(()=>{h.selectedItems.length===0&&g.cancel(),d.dispose(),u?.complete(h.selectedItems.slice(0))}),h.show(),l&&c===e&&(h.valueSelection=l),t)return u?.p}adjustValueSelection(e,t,i){let n;i?.preserveValue?n=[e.value.length,e.value.length]:n=[t?.prefix.length??0,e.value.length],e.valueSelection=n}registerPickerListeners(e,t,i,n,o){const r=new U,a=this.visibleQuickAccess={picker:e,descriptor:i,value:n};return r.add(pe(()=>{a===this.visibleQuickAccess&&(this.visibleQuickAccess=void 0)})),r.add(e.onDidChangeValue(l=>{const[c]=this.getOrInstantiateProvider(l,o?.enabledProviderPrefixes);c!==t?this.show(l,{enabledProviderPrefixes:o?.enabledProviderPrefixes,preserveValue:!0,providerOptions:o?.providerOptions}):a.value=l})),i&&r.add(e.onDidAccept(()=>{this.lastAcceptedPickerValues.set(i,e.value)})),r}getOrInstantiateProvider(e,t){const i=this.registry.getQuickAccessProvider(e);if(!i||t&&!t?.includes(i.prefix))return[void 0,void 0];let n=this.mapProviderToDescriptor.get(i);return n||(n=this.instantiationService.createInstance(i.ctor),this.mapProviderToDescriptor.set(i,n)),[n,i]}};k9=$9e([EY(0,bs),EY(1,_e)],k9);var U9e=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o};class Ore{constructor(e){this.nodes=e}toString(){return this.nodes.map(e=>typeof e=="string"?e:e.label).join("")}}U9e([Ai],Ore.prototype,"toString",null);const q9e=/\[([^\]]+)\]\(((?:https?:\/\/|command:|file:)[^\)\s]+)(?: (["'])(.+?)(\3))?\)/gi;function j9e(s){const e=[];let t=0,i;for(;i=q9e.exec(s);){i.index-t>0&&e.push(s.substring(t,i.index));const[,n,o,,r]=i;r?e.push({label:n,href:o,title:r}):e.push({label:n,href:o}),t=i.index+i[0].length}return t<s.length&&e.push(s.substring(t)),new Ore(e)}const zO={},G9e=new k7("quick-input-button-icon-");function K9e(s){if(!s)return;let e;const t=s.dark.toString();return zO[t]?e=zO[t]:(e=G9e.nextId(),zE(`.${e}, .hc-light .${e}`,`background-image: ${Mc(s.light||s.dark)}`),zE(`.vs-dark .${e}, .hc-black .${e}`,`background-image: ${Mc(s.dark)}`),zO[t]=e),e}function Aw(s,e,t){let i=s.iconClass||K9e(s.iconPath);return s.alwaysVisible&&(i=i?`${i} always-visible`:"always-visible"),{id:e,label:"",tooltip:s.tooltip||"",class:i,enabled:!0,run:t}}function Y9e(s,e,t){vn(e);const i=j9e(s);let n=0;for(const o of i.nodes)if(typeof o=="string")e.append(...Pu(o));else{let r=o.title;!r&&o.href.startsWith("command:")?r=m(1771,"Click to execute command '{0}'",o.href.substring(8)):r||(r=o.href);const a=ie("a",{href:o.href,title:r,tabIndex:n++},o.label);a.style.textDecoration="underline";const l=g=>{lue(g)&&gt.stop(g,!0),t.callback(o.href)},c=t.disposables.add(new Lt(a,se.CLICK)).event,d=t.disposables.add(new Lt(a,se.KEY_DOWN)).event,h=oe.chain(d,g=>g.filter(f=>{const p=new Pt(f);return p.equals(10)||p.equals(3)}));t.disposables.add(_s.addTarget(a));const u=t.disposables.add(new Lt(a,qt.Tap)).event;oe.any(c,u,h)(l,null,t.disposables),e.appendChild(a)}}var Z9e=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},IY=function(s,e){return function(t,i){e(t,i,s)}};const Pre="inQuickInput",X9e=new ce(Pre,!1,m(1748,"Whether keyboard focus is inside the quick input control")),o2=K.has(Pre),Q9e="quickInputAlignment",J9e=new ce(Q9e,"top",m(1749,"The alignment of the quick input")),k0="quickInputType",e6e=new ce(k0,void 0,m(1750,"The type of the currently visible quick input")),Fre="cursorAtEndOfQuickInputBox",t6e=new ce(Fre,!1,m(1751,"Whether the cursor in the quick input is at the end of the input box")),i6e=K.has(Fre),E9={iconClass:xe.asClassName(Z.quickInputBack),tooltip:m(1752,"Back")},pR=class pR extends F{constructor(e){super(),this.ui=e,this._visible=Re("visible",!1),this._widgetUpdated=!1,this._enabled=!0,this._busy=!1,this._ignoreFocusOut=!1,this._leftButtons=[],this._rightButtons=[],this._inlineButtons=[],this.buttonsUpdated=!1,this._toggles=[],this.togglesUpdated=!1,this.noValidationMessage=pR.noPromptMessage,this._severity=fi.Ignore,this.onDidTriggerButtonEmitter=this._register(new M),this.onDidHideEmitter=this._register(new M),this.onWillHideEmitter=this._register(new M),this.onDisposeEmitter=this._register(new M),this.visibleDisposables=this._register(new U),this.onDidHide=this.onDidHideEmitter.event}get visible(){return this._visible.get()}get title(){return this._title}set title(e){this._title=e,this.update()}get description(){return this._description}set description(e){this._description=e,this.update()}get step(){return this._steps}set step(e){this._steps=e,this.update()}get totalSteps(){return this._totalSteps}set totalSteps(e){this._totalSteps=e,this.update()}get enabled(){return this._enabled}set enabled(e){this._enabled=e,this.update()}get contextKey(){return this._contextKey}set contextKey(e){this._contextKey=e,this.update()}get busy(){return this._busy}set busy(e){this._busy=e,this.update()}get ignoreFocusOut(){return this._ignoreFocusOut}set ignoreFocusOut(e){const t=this._ignoreFocusOut!==e&&!na;this._ignoreFocusOut=e&&!na,t&&this.update()}get titleButtons(){return this._leftButtons.length?[...this._leftButtons,this._rightButtons]:this._rightButtons}get buttons(){return[...this._leftButtons,...this._rightButtons,...this._inlineButtons]}set buttons(e){this._leftButtons=e.filter(t=>t===E9),this._rightButtons=e.filter(t=>t!==E9&&t.location!==BI.Inline),this._inlineButtons=e.filter(t=>t.location===BI.Inline),this.buttonsUpdated=!0,this.update()}get toggles(){return this._toggles}set toggles(e){this._toggles=e??[],this.togglesUpdated=!0,this.update()}get validationMessage(){return this._validationMessage}set validationMessage(e){this._validationMessage=e,this.update()}get severity(){return this._severity}set severity(e){this._severity=e,this.update()}show(){this.visible||(this.visibleDisposables.add(this.ui.onDidTriggerButton(e=>{this.buttons.indexOf(e)!==-1&&this.onDidTriggerButtonEmitter.fire(e)})),this.ui.show(this),this._visible.set(!0,void 0),this._lastValidationMessage=void 0,this._lastSeverity=void 0,this.buttons.length&&(this.buttonsUpdated=!0),this.toggles.length&&(this.togglesUpdated=!0),this.update())}hide(){this.visible&&this.ui.hide()}didHide(e=HS.Other){this._visible.set(!1,void 0),this.visibleDisposables.clear(),this.onDidHideEmitter.fire({reason:e})}willHide(e=HS.Other){this.onWillHideEmitter.fire({reason:e})}update(){if(!this.visible)return;const e=this.getTitle();e&&this.ui.title.textContent!==e?this.ui.title.textContent=e:!e&&this.ui.title.innerHTML!=="&nbsp;"&&(this.ui.title.innerText=" ");const t=this.getDescription();if(this.ui.description1.textContent!==t&&(this.ui.description1.textContent=t),this.ui.description2.textContent!==t&&(this.ui.description2.textContent=t),this._widgetUpdated&&(this._widgetUpdated=!1,this._widget?vn(this.ui.widget,this._widget):vn(this.ui.widget)),this.busy&&!this.busyDelay&&(this.busyDelay=new $o,this.busyDelay.setIfNotSet(()=>{this.visible&&(this.ui.progressBar.infinite(),this.ui.progressBar.getContainer().removeAttribute("aria-hidden"))},800)),!this.busy&&this.busyDelay&&(this.ui.progressBar.stop(),this.ui.progressBar.getContainer().setAttribute("aria-hidden","true"),this.busyDelay.cancel(),this.busyDelay=void 0),this.buttonsUpdated){this.buttonsUpdated=!1,this.ui.leftActionBar.clear();const n=this._leftButtons.map((a,l)=>Aw(a,`id-${l}`,async()=>this.onDidTriggerButtonEmitter.fire(a)));this.ui.leftActionBar.push(n,{icon:!0,label:!1}),this.ui.rightActionBar.clear();const o=this._rightButtons.map((a,l)=>Aw(a,`id-${l}`,async()=>this.onDidTriggerButtonEmitter.fire(a)));this.ui.rightActionBar.push(o,{icon:!0,label:!1}),this.ui.inlineActionBar.clear();const r=this._inlineButtons.map((a,l)=>Aw(a,`id-${l}`,async()=>this.onDidTriggerButtonEmitter.fire(a)));this.ui.inlineActionBar.push(r,{icon:!0,label:!1})}if(this.togglesUpdated){this.togglesUpdated=!1;const n=this.toggles?.filter(r=>r instanceof Mh)??[];this.ui.inputBox.toggles=n;const o=n.length*22;this.ui.countContainer.style.right=o>0?`${4+o}px`:"4px",this.ui.visibleCountContainer.style.right=o>0?`${4+o}px`:"4px"}this.ui.ignoreFocusOut=this.ignoreFocusOut,this.ui.setEnabled(this.enabled),this.ui.setContextKey(this.contextKey);const i=this.validationMessage||this.noValidationMessage;this._lastValidationMessage!==i&&(this._lastValidationMessage=i,vn(this.ui.message),i&&Y9e(i,this.ui.message,{callback:n=>{this.ui.linkOpenerDelegate(n)},disposables:this.visibleDisposables})),this._lastSeverity!==this.severity&&(this._lastSeverity=this.severity,this.showMessageDecoration(this.severity))}getTitle(){return this.title&&this.step?`${this.title} (${this.getSteps()})`:this.title?this.title:this.step?this.getSteps():""}getDescription(){return this.description||""}getSteps(){return this.step&&this.totalSteps?m(1754,"{0}/{1}",this.step,this.totalSteps):this.step?String(this.step):""}showMessageDecoration(e){if(this.ui.inputBox.showDecoration(e),e!==fi.Ignore){const t=this.ui.inputBox.stylesForType(e);this.ui.message.style.color=t.foreground?`${t.foreground}`:"",this.ui.message.style.backgroundColor=t.background?`${t.background}`:"",this.ui.message.style.border=t.border?`1px solid ${t.border}`:"",this.ui.message.style.marginBottom="-2px"}else this.ui.message.style.color="",this.ui.message.style.backgroundColor="",this.ui.message.style.border="",this.ui.message.style.marginBottom=""}dispose(){this.hide(),this.onDisposeEmitter.fire(),super.dispose()}};pR.noPromptMessage=m(1753,"Press 'Enter' to confirm your input or 'Escape' to cancel");let E0=pR;const mR=class mR extends E0{constructor(e){super(e),this._value="",this.onDidChangeValueEmitter=this._register(new M),this.onWillAcceptEmitter=this._register(new M),this.onDidAcceptEmitter=this._register(new M),this.onDidCustomEmitter=this._register(new M),this._items=[],this.itemsUpdated=!1,this._canSelectMany=!1,this._canAcceptInBackground=!1,this._matchOnDescription=!1,this._matchOnDetail=!1,this._matchOnLabel=!0,this._matchOnLabelMode="fuzzy",this._sortByLabel=!0,this._keepScrollPosition=!1,this._itemActivation=nl.FIRST,this._activeItems=[],this.activeItemsUpdated=!1,this.activeItemsToConfirm=[],this.onDidChangeActiveEmitter=this._register(new M),this._selectedItems=[],this.selectedItemsUpdated=!1,this.selectedItemsToConfirm=[],this.onDidChangeSelectionEmitter=this._register(new M),this.onDidTriggerItemButtonEmitter=this._register(new M),this.onDidTriggerSeparatorButtonEmitter=this._register(new M),this.valueSelectionUpdated=!0,this._ok="default",this._customButton=!1,this._focusEventBufferer=new $0,this.type="quickPick",this.filterValue=t=>t,this.onDidChangeValue=this.onDidChangeValueEmitter.event,this.onWillAccept=this.onWillAcceptEmitter.event,this.onDidAccept=this.onDidAcceptEmitter.event,this.onDidChangeActive=this.onDidChangeActiveEmitter.event,this.onDidChangeSelection=this.onDidChangeSelectionEmitter.event,this.onDidTriggerItemButton=this.onDidTriggerItemButtonEmitter.event,this.onDidTriggerSeparatorButton=this.onDidTriggerSeparatorButtonEmitter.event,this.noValidationMessage=void 0}get quickNavigate(){return this._quickNavigate}set quickNavigate(e){this._quickNavigate=e,this.update()}get value(){return this._value}set value(e){this.doSetValue(e)}doSetValue(e,t){this._value!==e&&(this._value=e,t||this.update(),this.visible&&this.ui.list.filter(this.filterValue(this._value))&&this.trySelectFirst(),this.onDidChangeValueEmitter.fire(this._value))}set ariaLabel(e){this._ariaLabel=e,this.update()}get ariaLabel(){return this._ariaLabel}get placeholder(){return this._placeholder}set placeholder(e){this._placeholder=e,this.update()}get prompt(){return this.noValidationMessage}set prompt(e){this.noValidationMessage=e,this.update()}get items(){return this._items}get scrollTop(){return this.ui.list.scrollTop}set scrollTop(e){this.ui.list.scrollTop=e}set items(e){this._items=e,this.itemsUpdated=!0,this.update()}get canSelectMany(){return this._canSelectMany}set canSelectMany(e){this._canSelectMany=e,this.update()}get canAcceptInBackground(){return this._canAcceptInBackground}set canAcceptInBackground(e){this._canAcceptInBackground=e}get matchOnDescription(){return this._matchOnDescription}set matchOnDescription(e){this._matchOnDescription=e,this.update()}get matchOnDetail(){return this._matchOnDetail}set matchOnDetail(e){this._matchOnDetail=e,this.update()}get matchOnLabel(){return this._matchOnLabel}set matchOnLabel(e){this._matchOnLabel=e,this.update()}get matchOnLabelMode(){return this._matchOnLabelMode}set matchOnLabelMode(e){this._matchOnLabelMode=e,this.update()}get sortByLabel(){return this._sortByLabel}set sortByLabel(e){this._sortByLabel=e,this.update()}get keepScrollPosition(){return this._keepScrollPosition}set keepScrollPosition(e){this._keepScrollPosition=e}get itemActivation(){return this._itemActivation}set itemActivation(e){this._itemActivation=e}get activeItems(){return this._activeItems}set activeItems(e){this._activeItems=e,this.activeItemsUpdated=!0,this.update()}get selectedItems(){return this._selectedItems}set selectedItems(e){this._selectedItems=e,this.selectedItemsUpdated=!0,this.update()}get keyMods(){return this._quickNavigate?WEe:this.ui.keyMods}get valueSelection(){const e=this.ui.inputBox.getSelection();if(e)return[e.start,e.end]}set valueSelection(e){this._valueSelection=e,this.valueSelectionUpdated=!0,this.update()}get customButton(){return this._customButton}set customButton(e){this._customButton=e,this.update()}get customLabel(){return this._customButtonLabel}set customLabel(e){this._customButtonLabel=e,this.update()}get customHover(){return this._customButtonHover}set customHover(e){this._customButtonHover=e,this.update()}get ok(){return this._ok}set ok(e){this._ok=e,this.update()}get okLabel(){return this._okLabel??m(1756,"OK")}set okLabel(e){this._okLabel=e,this.update()}get hideInput(){return!!this._hideInput}set hideInput(e){this._hideInput=e,this.update()}trySelectFirst(){this.canSelectMany||this.ui.list.focus(Kt.First)}show(){this.visible||(this.visibleDisposables.add(this.ui.inputBox.onDidChange(e=>{this.doSetValue(e,!0)})),this.visibleDisposables.add(this.ui.onDidAccept(()=>{this.canSelectMany?this.ui.list.getCheckedElements().length||(this._selectedItems=[],this.onDidChangeSelectionEmitter.fire(this.selectedItems)):this.activeItems[0]&&(this._selectedItems=[this.activeItems[0]],this.onDidChangeSelectionEmitter.fire(this.selectedItems)),this.handleAccept(!1)})),this.visibleDisposables.add(this.ui.onDidCustom(()=>{this.onDidCustomEmitter.fire()})),this.visibleDisposables.add(this._focusEventBufferer.wrapEvent(this.ui.list.onDidChangeFocus,(e,t)=>t)(e=>{this.activeItemsUpdated||this.activeItemsToConfirm!==this._activeItems&&ni(e,this._activeItems,(t,i)=>t===i)||(this._activeItems=e,this.onDidChangeActiveEmitter.fire(e))})),this.visibleDisposables.add(this.ui.list.onDidChangeSelection(({items:e,event:t})=>{if(this.canSelectMany&&!e.some(i=>i.pickable===!1)){e.length&&this.ui.list.setSelectedElements([]);return}this.selectedItemsToConfirm!==this._selectedItems&&ni(e,this._selectedItems,(i,n)=>i===n)||(this._selectedItems=e,this.onDidChangeSelectionEmitter.fire(e),e.length&&this.handleAccept(dW(t)&&t.button===1))})),this.visibleDisposables.add(this.ui.list.onChangedCheckedElements(e=>{!this.canSelectMany||!this.visible||this.selectedItemsToConfirm!==this._selectedItems&&ni(e,this._selectedItems,(t,i)=>t===i)||(this._selectedItems=e,this.onDidChangeSelectionEmitter.fire(e))})),this.visibleDisposables.add(this.ui.list.onButtonTriggered(e=>this.onDidTriggerItemButtonEmitter.fire(e))),this.visibleDisposables.add(this.ui.list.onSeparatorButtonTriggered(e=>this.onDidTriggerSeparatorButtonEmitter.fire(e))),this.visibleDisposables.add(this.registerQuickNavigation()),this.valueSelectionUpdated=!0),super.show()}handleAccept(e){let t=!1;this.onWillAcceptEmitter.fire({veto:()=>t=!0}),t||this.onDidAcceptEmitter.fire({inBackground:e})}registerQuickNavigation(){return W(this.ui.container,se.KEY_UP,e=>{if(this.canSelectMany||!this._quickNavigate)return;const t=new Pt(e),i=t.keyCode;this._quickNavigate.keybindings.some(r=>{const a=r.getChords();return a.length>1?!1:a[0].shiftKey&&i===4?!(t.ctrlKey||t.altKey||t.metaKey):!!(a[0].altKey&&i===6||a[0].ctrlKey&&i===5||a[0].metaKey&&i===57)})&&(this.activeItems[0]&&(this._selectedItems=[this.activeItems[0]],this.onDidChangeSelectionEmitter.fire(this.selectedItems),this.handleAccept(!1)),this._quickNavigate=void 0)})}update(){if(!this.visible)return;const e=this.keepScrollPosition?this.scrollTop:0,t=!!this.description,i={title:!!this.title||!!this.step||!!this.titleButtons.length,description:t,checkAll:this.canSelectMany&&!this._hideCheckAll,checkBox:this.canSelectMany,inputBox:!this._hideInput,progressBar:!this._hideInput||t,visibleCount:!0,count:this.canSelectMany&&!this._hideCountBadge,ok:this.ok==="default"?this.canSelectMany:this.ok,list:!0,message:!!this.validationMessage||!!this.prompt,customButton:this.customButton};this.ui.setVisibilities(i),super.update(),this.ui.inputBox.value!==this.value&&(this.ui.inputBox.value=this.value),this.valueSelectionUpdated&&(this.valueSelectionUpdated=!1,this.ui.inputBox.select(this._valueSelection&&{start:this._valueSelection[0],end:this._valueSelection[1]})),this.ui.inputBox.placeholder!==(this.placeholder||"")&&(this.ui.inputBox.placeholder=this.placeholder||"");let n=this.ariaLabel;!n&&i.inputBox&&(n=this.placeholder,this.title&&(n=n?`${n} - ${this.title}`:this.title),n||(n=mR.DEFAULT_ARIA_LABEL)),this.ui.list.ariaLabel!==n&&(this.ui.list.ariaLabel=n??null),this.ui.inputBox.ariaLabel!==n&&(this.ui.inputBox.ariaLabel=n??"input"),this.ui.list.matchOnDescription=this.matchOnDescription,this.ui.list.matchOnDetail=this.matchOnDetail,this.ui.list.matchOnLabel=this.matchOnLabel,this.ui.list.matchOnLabelMode=this.matchOnLabelMode,this.ui.list.sortByLabel=this.sortByLabel,this.itemsUpdated&&(this.itemsUpdated=!1,this._focusEventBufferer.bufferEvents(()=>{switch(this.ui.list.setElements(this.items),this.ui.list.shouldLoop=!this.canSelectMany,this.ui.list.filter(this.filterValue(this.ui.inputBox.value)),this._itemActivation){case nl.NONE:this._itemActivation=nl.FIRST;break;case nl.SECOND:this.ui.list.focus(Kt.Second),this._itemActivation=nl.FIRST;break;case nl.LAST:this.ui.list.focus(Kt.Last),this._itemActivation=nl.FIRST;break;default:this.trySelectFirst();break}})),this.ui.container.classList.contains("show-checkboxes")!==!!this.canSelectMany&&(this.canSelectMany?this.ui.list.clearFocus():this.trySelectFirst()),this.activeItemsUpdated&&(this.activeItemsUpdated=!1,this.activeItemsToConfirm=this._activeItems,this.ui.list.setFocusedElements(this.activeItems),this.activeItemsToConfirm===this._activeItems&&(this.activeItemsToConfirm=null)),this.selectedItemsUpdated&&(this.selectedItemsUpdated=!1,this.selectedItemsToConfirm=this._selectedItems,this.canSelectMany?this.ui.list.setCheckedElements(this.selectedItems):this.ui.list.setSelectedElements(this.selectedItems),this.selectedItemsToConfirm===this._selectedItems&&(this.selectedItemsToConfirm=null)),this.ui.ok.label=this.okLabel||"",this.ui.customButton.label=this.customLabel||"",this.ui.customButton.element.title=this.customHover||"",i.inputBox||(this.ui.list.domFocus(),this.canSelectMany&&this.ui.list.focus(Kt.First)),this.keepScrollPosition&&(this.scrollTop=e)}focus(e){this.ui.list.focus(e),this.canSelectMany&&this.ui.list.domFocus()}accept(e){e&&!this._canAcceptInBackground||(this.activeItems[0]&&!this._canSelectMany&&(this._selectedItems=[this.activeItems[0]],this.onDidChangeSelectionEmitter.fire(this.selectedItems)),this.handleAccept(e??!1))}};mR.DEFAULT_ARIA_LABEL=m(1755,"Type to narrow down results.");let py=mR;class n6e extends E0{constructor(){super(...arguments),this._value="",this.valueSelectionUpdated=!0,this._password=!1,this.onDidValueChangeEmitter=this._register(new M),this.onDidAcceptEmitter=this._register(new M),this.type="inputBox",this.onDidChangeValue=this.onDidValueChangeEmitter.event,this.onDidAccept=this.onDidAcceptEmitter.event}get value(){return this._value}set value(e){this._value=e||"",this.update()}get valueSelection(){const e=this.ui.inputBox.getSelection();if(e)return[e.start,e.end]}set valueSelection(e){this._valueSelection=e,this.valueSelectionUpdated=!0,this.update()}get placeholder(){return this._placeholder}set placeholder(e){this._placeholder=e,this.update()}get ariaLabel(){return this._ariaLabel}set ariaLabel(e){this._ariaLabel=e,this.update()}get password(){return this._password}set password(e){this._password=e,this.update()}get prompt(){return this._prompt}set prompt(e){this._prompt=e,this.noValidationMessage=e?m(1757,"{0} (Press 'Enter' to confirm or 'Escape' to cancel)",e):E0.noPromptMessage,this.update()}show(){this.visible||(this.visibleDisposables.add(this.ui.inputBox.onDidChange(e=>{e!==this.value&&(this._value=e,this.onDidValueChangeEmitter.fire(e))})),this.visibleDisposables.add(this.ui.onDidAccept(()=>this.onDidAcceptEmitter.fire())),this.valueSelectionUpdated=!0),super.show()}accept(){this.onDidAcceptEmitter.fire()}update(){if(!this.visible)return;this.ui.container.classList.remove("hidden-input");const e={title:!!this.title||!!this.step||!!this.titleButtons.length,description:!!this.description||!!this.step,inputBox:!0,message:!0,progressBar:!0};this.ui.setVisibilities(e),super.update(),this.ui.inputBox.value!==this.value&&(this.ui.inputBox.value=this.value),this.valueSelectionUpdated&&(this.valueSelectionUpdated=!1,this.ui.inputBox.select(this._valueSelection&&{start:this._valueSelection[0],end:this._valueSelection[1]})),this.ui.inputBox.placeholder!==(this.placeholder||"")&&(this.ui.inputBox.placeholder=this.placeholder||""),this.ui.inputBox.password!==this.password&&(this.ui.inputBox.password=this.password);let t=this.ariaLabel;!t&&e.inputBox&&(t=this.placeholder?this.title?`${this.placeholder} - ${this.title}`:this.placeholder:this.title?this.title:"input"),this.ui.inputBox.ariaLabel!==t&&(this.ui.inputBox.ariaLabel=t||"input")}}let I9=class extends Cv{constructor(e,t){super("mouse",void 0,i=>this.getOverrideOptions(i),e,t)}getOverrideOptions(e){const t=(ki(e.content)?e.content.textContent??"":typeof e.content=="string"?e.content:e.content.value).includes(`
`);return{persistence:{hideOnKeyDown:!1},appearance:{showHoverHint:t,skipFadeInAnimation:!0}}}};I9=Z9e([IY(0,He),IY(1,so)],I9);const DY="done",TY="active",$O="infinite",UO="infinite-long-running",NY="discrete",_R=class _R extends F{constructor(e,t){super(),this.progressSignal=this._register(new vt),this.workedVal=0,this.showDelayedScheduler=this._register(new Rt(()=>Mo(this.element),0)),this.longRunningScheduler=this._register(new Rt(()=>this.infiniteLongRunning(),_R.LONG_RUNNING_INFINITE_THRESHOLD)),this.create(e,t)}create(e,t){this.element=document.createElement("div"),this.element.classList.add("monaco-progress-container"),this.element.setAttribute("role","progressbar"),this.element.setAttribute("aria-valuemin","0"),e.appendChild(this.element),this.bit=document.createElement("div"),this.bit.classList.add("progress-bit"),this.bit.style.backgroundColor=t?.progressBarBackground||"#0E70C0",this.element.appendChild(this.bit)}off(){this.bit.style.width="inherit",this.bit.style.opacity="1",this.element.classList.remove(TY,$O,UO,NY),this.workedVal=0,this.totalWork=void 0,this.longRunningScheduler.cancel(),this.progressSignal.clear()}stop(){return this.doDone(!1)}doDone(e){return this.element.classList.add(DY),this.element.classList.contains($O)?(this.bit.style.opacity="0",e?setTimeout(()=>this.off(),200):this.off()):(this.bit.style.width="inherit",e?setTimeout(()=>this.off(),200):this.off()),this}infinite(){return this.bit.style.width="2%",this.bit.style.opacity="1",this.element.classList.remove(NY,DY,UO),this.element.classList.add(TY,$O),this.longRunningScheduler.schedule(),this}infiniteLongRunning(){this.element.classList.add(UO)}getContainer(){return this.element}};_R.LONG_RUNNING_INFINITE_THRESHOLD=1e4;let D9=_R;const s6e=ie;class o6e extends F{constructor(e,t,i){super(),this.parent=e,this.onDidChange=o=>this.findInput.onDidChange(o),this.container=X(this.parent,s6e(".quick-input-box")),this.findInput=this._register(new Bne(this.container,void 0,{label:"",inputBoxStyles:t,toggleStyles:i}));const n=this.findInput.inputBox.inputElement;n.role="textbox",n.ariaHasPopup="menu",n.ariaAutoComplete="list"}get onKeyDown(){return this.findInput.onKeyDown}get value(){return this.findInput.getValue()}set value(e){this.findInput.setValue(e)}select(e=null){this.findInput.inputBox.select(e)}getSelection(){return this.findInput.inputBox.getSelection()}isSelectionAtEnd(){return this.findInput.inputBox.isSelectionAtEnd()}get placeholder(){return this.findInput.inputBox.inputElement.getAttribute("placeholder")||""}set placeholder(e){this.findInput.inputBox.setPlaceHolder(e)}get password(){return this.findInput.inputBox.inputElement.type==="password"}set password(e){this.findInput.inputBox.inputElement.type=e?"password":"text"}set enabled(e){this.findInput.inputBox.inputElement.toggleAttribute("readonly",!e)}set toggles(e){this.findInput.setAdditionalToggles(e)}get ariaLabel(){return this.findInput.inputBox.inputElement.getAttribute("aria-label")||""}set ariaLabel(e){this.findInput.inputBox.inputElement.setAttribute("aria-label",e)}hasFocus(){return this.findInput.inputBox.hasFocus()}setAttribute(e,t){this.findInput.inputBox.inputElement.setAttribute(e,t)}removeAttribute(e){this.findInput.inputBox.inputElement.removeAttribute(e)}showDecoration(e){e===fi.Ignore?this.findInput.clearMessage():this.findInput.showMessage({type:e===fi.Info?1:e===fi.Warning?2:3,content:""})}stylesForType(e){return this.findInput.inputBox.stylesForType(e===fi.Info?1:e===fi.Warning?2:3)}setFocus(){this.findInput.focus()}layout(){this.findInput.inputBox.layout()}}const RY=new es(()=>{const s=a_.Collator(void 0,{numeric:!0,sensitivity:"base"}).value;return{collator:s,collatorIsNumeric:s.resolvedOptions().numeric}});new es(()=>({collator:a_.Collator(void 0,{numeric:!0}).value}));new es(()=>({collator:a_.Collator(void 0,{numeric:!0,sensitivity:"accent"}).value}));function r6e(s,e,t=!1){const i=s||"",n=e||"",o=RY.value.collator.compare(i,n);return RY.value.collatorIsNumeric&&o===0&&i!==n?i<n?-1:1:o}function a6e(s,e,t){const i=s.toLowerCase(),n=e.toLowerCase(),o=l6e(s,e,t);if(o)return o;const r=i.endsWith(t),a=n.endsWith(t);if(r!==a)return r?-1:1;const l=r6e(i,n);return l!==0?l:i.localeCompare(n)}function l6e(s,e,t){const i=s.toLowerCase(),n=e.toLowerCase(),o=i.startsWith(t),r=n.startsWith(t);if(o!==r)return o?-1:1;if(o&&r){if(i.length<n.length)return-1;if(i.length>n.length)return 1}return 0}var r2=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},T9=function(s,e){return function(t,i){e(t,i,s)}},N9;const Kl=ie;class Bre{constructor(e,t,i){this.index=e,this.hasCheckbox=t,this._hidden=!1,this._init=new es(()=>{const n=i.label??"",o=Kf(n).text.trim(),r=i.ariaLabel||[n,this.saneDescription,this.saneDetail].map(a=>Nte(a)).filter(a=>!!a).join(", ");return{saneLabel:n,saneSortLabel:o,saneAriaLabel:r}}),this._saneDescription=i.description,this._saneTooltip=i.tooltip}get saneLabel(){return this._init.value.saneLabel}get saneSortLabel(){return this._init.value.saneSortLabel}get saneAriaLabel(){return this._init.value.saneAriaLabel}get element(){return this._element}set element(e){this._element=e}get hidden(){return this._hidden}set hidden(e){this._hidden=e}get saneDescription(){return this._saneDescription}set saneDescription(e){this._saneDescription=e}get saneDetail(){return this._saneDetail}set saneDetail(e){this._saneDetail=e}get saneTooltip(){return this._saneTooltip}set saneTooltip(e){this._saneTooltip=e}get labelHighlights(){return this._labelHighlights}set labelHighlights(e){this._labelHighlights=e}get descriptionHighlights(){return this._descriptionHighlights}set descriptionHighlights(e){this._descriptionHighlights=e}get detailHighlights(){return this._detailHighlights}set detailHighlights(e){this._detailHighlights=e}}class qn extends Bre{constructor(e,t,i,n,o,r,a){super(e,i,r),this.childIndex=t,this.fireButtonTriggered=n,this._onChecked=o,this.item=r,this._separator=a,this._checked=!1,this.onChecked=i?oe.map(oe.filter(this._onChecked.event,l=>l.element===this),l=>l.checked):oe.None,this._saneDetail=r.detail,this._labelHighlights=r.highlights?.label,this._descriptionHighlights=r.highlights?.description,this._detailHighlights=r.highlights?.detail}get separator(){return this._separator}set separator(e){this._separator=e}get checked(){return this._checked}set checked(e){e!==this._checked&&(this._checked=e,this._onChecked.fire({element:this,checked:e}))}get checkboxDisabled(){return!!this.item.disabled}}var sc;(function(s){s[s.NONE=0]="NONE",s[s.MOUSE_HOVER=1]="MOUSE_HOVER",s[s.ACTIVE_ITEM=2]="ACTIVE_ITEM"})(sc||(sc={}));class df extends Bre{constructor(e,t,i){super(e,!1,i),this.fireSeparatorButtonTriggered=t,this.separator=i,this.children=new Array,this.focusInsideSeparator=sc.NONE}}class c6e{getHeight(e){return e instanceof df?30:e.saneDetail?44:22}getTemplateId(e){return e instanceof qn?CT.ID:yT.ID}}class d6e{getWidgetAriaLabel(){return m(1770,"Quick Input")}getAriaLabel(e){return e.separator?.label?`${e.saneAriaLabel}, ${e.separator.label}`:e.saneAriaLabel}getWidgetRole(){return"listbox"}getRole(e){return e.hasCheckbox?"checkbox":"option"}isChecked(e){if(!(!e.hasCheckbox||!(e instanceof qn)))return{get value(){return e.checked},onDidChange:t=>e.onChecked(()=>t())}}}class Wre{constructor(e){this.hoverDelegate=e}renderTemplate(e){const t=Object.create(null);t.toDisposeElement=new U,t.toDisposeTemplate=new U,t.entry=X(e,Kl(".quick-input-list-entry"));const i=X(t.entry,Kl("label.quick-input-list-label"));t.outerLabel=i,t.checkbox=t.toDisposeTemplate.add(new vt),t.toDisposeTemplate.add(Fi(i,se.CLICK,c=>{if(t.checkbox.value&&!c.defaultPrevented&&t.checkbox.value.enabled){const d=!t.checkbox.value.checked;t.checkbox.value.checked=d,t.element.checked=d}}));const n=X(i,Kl(".quick-input-list-rows")),o=X(n,Kl(".quick-input-list-row")),r=X(n,Kl(".quick-input-list-row"));t.label=new QS(o,{supportHighlights:!0,supportDescriptionHighlights:!0,supportIcons:!0,hoverDelegate:this.hoverDelegate}),t.toDisposeTemplate.add(t.label),t.icon=RR(t.label.element,Kl(".quick-input-list-icon"));const a=X(o,Kl(".quick-input-list-entry-keybinding"));t.keybinding=new hC(a,Fo),t.toDisposeTemplate.add(t.keybinding);const l=X(r,Kl(".quick-input-list-label-meta"));return t.detail=new QS(l,{supportHighlights:!0,supportIcons:!0,hoverDelegate:this.hoverDelegate}),t.toDisposeTemplate.add(t.detail),t.separator=X(t.entry,Kl(".quick-input-list-separator")),t.actionBar=new bo(t.entry,this.hoverDelegate?{hoverDelegate:this.hoverDelegate}:void 0),t.actionBar.domNode.classList.add("quick-input-list-entry-action-bar"),t.toDisposeTemplate.add(t.actionBar),t}disposeTemplate(e){e.toDisposeElement.dispose(),e.toDisposeTemplate.dispose()}disposeElement(e,t,i){i.toDisposeElement.clear(),i.actionBar.clear()}}var fm;let CT=(fm=class extends Wre{constructor(e,t){super(e),this.themeService=t,this._itemsWithSeparatorsFrequency=new Map}get templateId(){return N9.ID}ensureCheckbox(e,t){if(!e.hasCheckbox){t.checkbox.value?.domNode.remove(),t.checkbox.clear();return}let i=t.checkbox.value;i?i.setTitle(e.saneLabel):(i=new Tne(e.saneLabel,e.checked,{...$7,size:15}),t.checkbox.value=i,t.outerLabel.prepend(i.domNode)),e.checkboxDisabled?i.disable():i.enable(),i.checked=e.checked,t.toDisposeElement.add(e.onChecked(n=>i.checked=n)),t.toDisposeElement.add(i.onChange(()=>e.checked=i.checked))}renderElement(e,t,i){const n=e.element;i.element=n,n.element=i.entry??void 0;const o=n.item;n.element.classList.toggle("not-pickable",n.item.pickable===!1),this.ensureCheckbox(n,i);const{labelHighlights:r,descriptionHighlights:a,detailHighlights:l}=n;if(o.iconPath){const u=wh(this.themeService.getColorTheme().type)?o.iconPath.dark:o.iconPath.light??o.iconPath.dark,g=Le.revive(u);i.icon.className="quick-input-list-icon",i.icon.style.backgroundImage=Mc(g)}else i.icon.style.backgroundImage="",i.icon.className=o.iconClass?`quick-input-list-icon ${o.iconClass}`:"";let c;!n.saneTooltip&&n.saneDescription&&(c={markdown:{value:Aa(n.saneDescription),supportThemeIcons:!0},markdownNotSupportedFallback:n.saneDescription});const d={matches:r||[],descriptionTitle:c,descriptionMatches:a||[],labelEscapeNewLines:!0};if(d.extraClasses=o.iconClasses,d.italic=o.italic,d.strikethrough=o.strikethrough,i.entry.classList.remove("quick-input-list-separator-as-item"),i.label.setLabel(n.saneLabel,n.saneDescription,d),i.keybinding.set(o.keybinding),n.saneDetail){let u;n.saneTooltip||(u={markdown:{value:Aa(n.saneDetail),supportThemeIcons:!0},markdownNotSupportedFallback:n.saneDetail}),i.detail.element.style.display="",i.detail.setLabel(n.saneDetail,void 0,{matches:l,title:u,labelEscapeNewLines:!0})}else i.detail.element.style.display="none";n.separator?.label?(i.separator.textContent=n.separator.label,i.separator.style.display="",this.addItemWithSeparator(n)):i.separator.style.display="none",i.entry.classList.toggle("quick-input-list-separator-border",!!n.separator&&n.childIndex!==0);const h=o.buttons;h&&h.length?(i.actionBar.push(h.map((u,g)=>Aw(u,`id-${g}`,()=>n.fireButtonTriggered({button:u,item:n.item}))),{icon:!0,label:!1}),i.entry.classList.add("has-actions")):i.entry.classList.remove("has-actions")}disposeElement(e,t,i){this.removeItemWithSeparator(e.element),super.disposeElement(e,t,i)}isItemWithSeparatorVisible(e){return this._itemsWithSeparatorsFrequency.has(e)}addItemWithSeparator(e){this._itemsWithSeparatorsFrequency.set(e,(this._itemsWithSeparatorsFrequency.get(e)||0)+1)}removeItemWithSeparator(e){const t=this._itemsWithSeparatorsFrequency.get(e)||0;t>1?this._itemsWithSeparatorsFrequency.set(e,t-1):this._itemsWithSeparatorsFrequency.delete(e)}},N9=fm,fm.ID="quickpickitem",fm);CT=N9=r2([T9(1,wi)],CT);const bR=class bR extends Wre{constructor(){super(...arguments),this._visibleSeparatorsFrequency=new Map}get templateId(){return bR.ID}get visibleSeparators(){return[...this._visibleSeparatorsFrequency.keys()]}isSeparatorVisible(e){return this._visibleSeparatorsFrequency.has(e)}renderElement(e,t,i){const n=e.element;i.element=n,n.element=i.entry??void 0,n.element.classList.toggle("focus-inside",!!n.focusInsideSeparator);const o=n.separator,{labelHighlights:r,descriptionHighlights:a}=n;i.icon.style.backgroundImage="",i.icon.className="";let l;!n.saneTooltip&&n.saneDescription&&(l={markdown:{value:Aa(n.saneDescription),supportThemeIcons:!0},markdownNotSupportedFallback:n.saneDescription});const c={matches:r||[],descriptionTitle:l,descriptionMatches:a||[],labelEscapeNewLines:!0};i.entry.classList.add("quick-input-list-separator-as-item"),i.label.setLabel(n.saneLabel,n.saneDescription,c),i.separator.style.display="none",i.entry.classList.add("quick-input-list-separator-border");const d=o.buttons;d&&d.length?(i.actionBar.push(d.map((h,u)=>Aw(h,`id-${u}`,()=>n.fireSeparatorButtonTriggered({button:h,separator:n.separator}))),{icon:!0,label:!1}),i.entry.classList.add("has-actions")):i.entry.classList.remove("has-actions"),this.addSeparator(n)}disposeElement(e,t,i){this.removeSeparator(e.element),this.isSeparatorVisible(e.element)||e.element.element?.classList.remove("focus-inside"),super.disposeElement(e,t,i)}addSeparator(e){this._visibleSeparatorsFrequency.set(e,(this._visibleSeparatorsFrequency.get(e)||0)+1)}removeSeparator(e){const t=this._visibleSeparatorsFrequency.get(e)||0;t>1?this._visibleSeparatorsFrequency.set(e,t-1):this._visibleSeparatorsFrequency.delete(e)}};bR.ID="quickpickseparator";let yT=bR,I0=class extends F{constructor(e,t,i,n,o,r){super(),this.parent=e,this.hoverDelegate=t,this.linkOpenerDelegate=i,this.accessibilityService=r,this._onKeyDown=new M,this._onLeave=new M,this.onLeave=this._onLeave.event,this._visibleCountObservable=Re("VisibleCount",0),this.onChangedVisibleCount=oe.fromObservable(this._visibleCountObservable,this._store),this._allVisibleCheckedObservable=Re("AllVisibleChecked",!1),this.onChangedAllVisibleChecked=oe.fromObservable(this._allVisibleCheckedObservable,this._store),this._checkedCountObservable=Re("CheckedCount",0),this.onChangedCheckedCount=oe.fromObservable(this._checkedCountObservable,this._store),this._checkedElementsObservable=H1({equalsFn:ni},new Array),this.onChangedCheckedElements=oe.fromObservable(this._checkedElementsObservable,this._store),this._onButtonTriggered=new M,this.onButtonTriggered=this._onButtonTriggered.event,this._onSeparatorButtonTriggered=new M,this.onSeparatorButtonTriggered=this._onSeparatorButtonTriggered.event,this._elementChecked=new M,this._elementCheckedEventBufferer=new $0,this._hasCheckboxes=!1,this._inputElements=new Array,this._elementTree=new Array,this._itemElements=new Array,this._elementDisposable=this._register(new U),this._matchOnDescription=!1,this._matchOnDetail=!1,this._matchOnLabel=!0,this._matchOnLabelMode="fuzzy",this._sortByLabel=!0,this._shouldLoop=!0,this._container=X(this.parent,Kl(".quick-input-list")),this._separatorRenderer=new yT(t),this._itemRenderer=o.createInstance(CT,t),this._tree=this._register(o.createInstance(lD,"QuickInput",this._container,new c6e,[this._itemRenderer,this._separatorRenderer],{filter:{filter(a){return a.hidden?0:a instanceof df?2:1}},sorter:{compare:(a,l)=>{if(!this.sortByLabel||!this._lastQueryString)return 0;const c=this._lastQueryString.toLowerCase();return u6e(a,l,c)}},accessibilityProvider:new d6e,setRowLineHeight:!1,multipleSelectionSupport:!1,hideTwistiesOfChildlessElements:!0,renderIndentGuides:Gm.None,findWidgetEnabled:!1,indent:0,horizontalScrolling:!1,allowNonCollapsibleParents:!0,alwaysConsumeMouseWheel:!0})),this._tree.getHTMLElement().id=n,this._registerListeners()}get onDidChangeFocus(){return oe.map(this._tree.onDidChangeFocus,e=>e.elements.filter(t=>t instanceof qn).map(t=>t.item),this._store)}get onDidChangeSelection(){return oe.map(this._tree.onDidChangeSelection,e=>({items:e.elements.filter(t=>t instanceof qn).map(t=>t.item),event:e.browserEvent}),this._store)}get displayed(){return this._container.style.display!=="none"}set displayed(e){this._container.style.display=e?"":"none"}get scrollTop(){return this._tree.scrollTop}set scrollTop(e){this._tree.scrollTop=e}get ariaLabel(){return this._tree.ariaLabel}set ariaLabel(e){this._tree.ariaLabel=e??""}set enabled(e){this._tree.getHTMLElement().style.pointerEvents=e?"":"none"}get matchOnDescription(){return this._matchOnDescription}set matchOnDescription(e){this._matchOnDescription=e}get matchOnDetail(){return this._matchOnDetail}set matchOnDetail(e){this._matchOnDetail=e}get matchOnLabel(){return this._matchOnLabel}set matchOnLabel(e){this._matchOnLabel=e}get matchOnLabelMode(){return this._matchOnLabelMode}set matchOnLabelMode(e){this._matchOnLabelMode=e}get sortByLabel(){return this._sortByLabel}set sortByLabel(e){this._sortByLabel=e}get shouldLoop(){return this._shouldLoop}set shouldLoop(e){this._shouldLoop=e}_registerListeners(){this._registerOnKeyDown(),this._registerOnContainerClick(),this._registerOnMouseMiddleClick(),this._registerOnTreeModelChanged(),this._registerOnElementChecked(),this._registerOnContextMenu(),this._registerHoverListeners(),this._registerSelectionChangeListener(),this._registerSeparatorActionShowingListeners()}_registerOnKeyDown(){this._register(this._tree.onKeyDown(e=>{const t=new Pt(e);t.keyCode===10&&this.toggleCheckbox(),this._onKeyDown.fire(t)}))}_registerOnContainerClick(){this._register(W(this._container,se.CLICK,e=>{(e.x||e.y)&&this._onLeave.fire()}))}_registerOnMouseMiddleClick(){this._register(W(this._container,se.AUXCLICK,e=>{e.button===1&&this._onLeave.fire()}))}_registerOnTreeModelChanged(){this._register(this._tree.onDidChangeModel(()=>{const e=this._itemElements.filter(t=>!t.hidden).length;this._visibleCountObservable.set(e,void 0),this._hasCheckboxes&&this._updateCheckedObservables()}))}_registerOnElementChecked(){this._register(this._elementCheckedEventBufferer.wrapEvent(this._elementChecked.event,(e,t)=>t)(e=>this._updateCheckedObservables()))}_registerOnContextMenu(){this._register(this._tree.onContextMenu(e=>{e.element&&(e.browserEvent.preventDefault(),this._tree.setSelection([e.element]))}))}_registerHoverListeners(){const e=this._register(new KX(typeof this.hoverDelegate.delay=="function"?this.hoverDelegate.delay():this.hoverDelegate.delay));this._register(this._tree.onMouseOver(async t=>{if(Gz(t.browserEvent.target)){e.cancel();return}if(!(!Gz(t.browserEvent.relatedTarget)&&wn(t.browserEvent.relatedTarget,t.element?.element)))try{await e.trigger(async()=>{t.element instanceof qn&&this.showHover(t.element)})}catch(i){if(!Dr(i))throw i}})),this._register(this._tree.onMouseOut(t=>{wn(t.browserEvent.relatedTarget,t.element?.element)||e.cancel()}))}_registerSeparatorActionShowingListeners(){this._register(this._tree.onDidChangeFocus(e=>{const t=e.elements[0]?this._tree.getParentElement(e.elements[0]):null;for(const i of this._separatorRenderer.visibleSeparators){const n=i===t;!!(i.focusInsideSeparator&sc.ACTIVE_ITEM)!==n&&(n?i.focusInsideSeparator|=sc.ACTIVE_ITEM:i.focusInsideSeparator&=~sc.ACTIVE_ITEM,this._tree.rerender(i))}})),this._register(this._tree.onMouseOver(e=>{const t=e.element?this._tree.getParentElement(e.element):null;for(const i of this._separatorRenderer.visibleSeparators){if(i!==t)continue;i.focusInsideSeparator&sc.MOUSE_HOVER||(i.focusInsideSeparator|=sc.MOUSE_HOVER,this._tree.rerender(i))}})),this._register(this._tree.onMouseOut(e=>{const t=e.element?this._tree.getParentElement(e.element):null;for(const i of this._separatorRenderer.visibleSeparators){if(i!==t)continue;i.focusInsideSeparator&sc.MOUSE_HOVER&&(i.focusInsideSeparator&=~sc.MOUSE_HOVER,this._tree.rerender(i))}}))}_registerSelectionChangeListener(){this._register(this._tree.onDidChangeSelection(e=>{const t=e.elements.filter(i=>i instanceof qn);t.length!==e.elements.length&&(e.elements.length===1&&e.elements[0]instanceof df&&(this._tree.setFocus([e.elements[0].children[0]]),this._tree.reveal(e.elements[0],0)),this._tree.setSelection(t))}))}setAllVisibleChecked(e){this._elementCheckedEventBufferer.bufferEvents(()=>{this._itemElements.forEach(t=>{!t.hidden&&!t.checkboxDisabled&&t.item.pickable!==!1&&(t.checked=e)})})}setElements(e){this._elementDisposable.clear(),this._lastQueryString=void 0,this._inputElements=e,this._hasCheckboxes=this.parent.classList.contains("show-checkboxes");let t;this._itemElements=new Array,this._elementTree=e.reduce((i,n,o)=>{let r;if(n.type==="separator"){if(!n.buttons)return i;t=new df(o,a=>this._onSeparatorButtonTriggered.fire(a),n),r=t}else{const a=o>0?e[o-1]:void 0;let l;a&&a.type==="separator"&&!a.buttons&&(l=a);const c=new qn(o,t?.children?t.children.length:o,this._hasCheckboxes&&n.pickable!==!1,d=>this._onButtonTriggered.fire(d),this._elementChecked,n,l);if(this._itemElements.push(c),t)return t.children.push(c),i;r=c}return i.push(r),i},new Array),this._setElementsToTree(this._elementTree),this.accessibilityService.isScreenReaderOptimized()&&setTimeout(()=>{const i=this._tree.getHTMLElement().querySelector(".monaco-list-row.focused"),n=i?.parentNode;if(i&&n){const o=i.nextSibling;i.remove(),n.insertBefore(i,o)}},0)}setFocusedElements(e){const t=e.map(i=>this._itemElements.find(n=>n.item===i)).filter(i=>!!i).filter(i=>!i.hidden);if(this._tree.setFocus(t),e.length>0){const i=this._tree.getFocus()[0];i&&this._tree.reveal(i)}}getActiveDescendant(){return this._tree.getHTMLElement().getAttribute("aria-activedescendant")}setSelectedElements(e){const t=e.map(i=>this._itemElements.find(n=>n.item===i)).filter(i=>!!i);this._tree.setSelection(t)}getCheckedElements(){return this._itemElements.filter(e=>e.checked).map(e=>e.item)}setCheckedElements(e){this._elementCheckedEventBufferer.bufferEvents(()=>{const t=new Set;for(const i of e)t.add(i);for(const i of this._itemElements)i.checked=t.has(i.item)})}focus(e){if(this._itemElements.length)switch(e===Kt.Second&&this._itemElements.length<2&&(e=Kt.First),e){case Kt.First:this._tree.scrollTop=0,this._tree.focusFirst(void 0,t=>t.element instanceof qn);break;case Kt.Second:{this._tree.scrollTop=0;let t=!1;this._tree.focusFirst(void 0,i=>i.element instanceof qn?t?!0:(t=!t,!1):!1);break}case Kt.Last:this._tree.scrollTop=this._tree.scrollHeight,this._tree.focusLast(void 0,t=>t.element instanceof qn);break;case Kt.Next:{const t=this._tree.getFocus();this._tree.focusNext(void 0,this._shouldLoop,void 0,n=>n.element instanceof qn?(this._tree.reveal(n.element),!0):!1);const i=this._tree.getFocus();t.length&&t[0]===i[0]&&this._onLeave.fire();break}case Kt.Previous:{const t=this._tree.getFocus();this._tree.focusPrevious(void 0,this._shouldLoop,void 0,n=>{if(!(n.element instanceof qn))return!1;const o=this._tree.getParentElement(n.element);return o===null||o.children[0]!==n.element?this._tree.reveal(n.element):this._tree.reveal(o),!0});const i=this._tree.getFocus();t.length&&t[0]===i[0]&&this._onLeave.fire();break}case Kt.NextPage:this._tree.focusNextPage(void 0,t=>t.element instanceof qn?(this._tree.reveal(t.element),!0):!1);break;case Kt.PreviousPage:this._tree.focusPreviousPage(void 0,t=>{if(!(t.element instanceof qn))return!1;const i=this._tree.getParentElement(t.element);return i===null||i.children[0]!==t.element?this._tree.reveal(t.element):this._tree.reveal(i),!0});break;case Kt.NextSeparator:{let t=!1;const i=this._tree.getFocus()[0];this._tree.focusNext(void 0,!0,void 0,o=>{if(t)return!0;if(o.element instanceof df)t=!0,this._separatorRenderer.isSeparatorVisible(o.element)?this._tree.reveal(o.element.children[0]):this._tree.reveal(o.element,0);else if(o.element instanceof qn){if(o.element.separator)return this._itemRenderer.isItemWithSeparatorVisible(o.element)?this._tree.reveal(o.element):this._tree.reveal(o.element,0),!0;if(o.element===this._elementTree[0])return this._tree.reveal(o.element,0),!0}return!1});const n=this._tree.getFocus()[0];i===n&&(this._tree.scrollTop=this._tree.scrollHeight,this._tree.focusLast(void 0,o=>o.element instanceof qn));break}case Kt.PreviousSeparator:{let t,i=!!this._tree.getFocus()[0]?.separator;this._tree.focusPrevious(void 0,!0,void 0,n=>{if(n.element instanceof df)i?t||(this._separatorRenderer.isSeparatorVisible(n.element)?this._tree.reveal(n.element):this._tree.reveal(n.element,0),t=n.element.children[0]):i=!0;else if(n.element instanceof qn&&!t){if(n.element.separator)this._itemRenderer.isItemWithSeparatorVisible(n.element)?this._tree.reveal(n.element):this._tree.reveal(n.element,0),t=n.element;else if(n.element===this._elementTree[0])return this._tree.reveal(n.element,0),!0}return!1}),t&&this._tree.setFocus([t]);break}}}clearFocus(){this._tree.setFocus([])}domFocus(){this._tree.domFocus()}layout(e){this._tree.getHTMLElement().style.maxHeight=e?`${Math.floor(e/44)*44+6}px`:"",this._tree.layout()}filter(e){if(this._lastQueryString=e,!(this._sortByLabel||this._matchOnLabel||this._matchOnDescription||this._matchOnDetail))return this._tree.layout(),!1;const t=e;if(e=e.trim(),!e||!(this.matchOnLabel||this.matchOnDescription||this.matchOnDetail))this._itemElements.forEach(i=>{i.labelHighlights=void 0,i.descriptionHighlights=void 0,i.detailHighlights=void 0,i.hidden=!1;const n=i.index&&this._inputElements[i.index-1];i.item&&(i.separator=n&&n.type==="separator"&&!n.buttons?n:void 0)});else{let i;this._itemElements.forEach(n=>{let o;this.matchOnLabelMode==="fuzzy"?o=this.matchOnLabel?q1(e,Kf(n.saneLabel))??void 0:void 0:o=this.matchOnLabel?h6e(t,Kf(n.saneLabel))??void 0:void 0;const r=this.matchOnDescription?q1(e,Kf(n.saneDescription||""))??void 0:void 0,a=this.matchOnDetail?q1(e,Kf(n.saneDetail||""))??void 0:void 0;if(o||r||a?(n.labelHighlights=o,n.descriptionHighlights=r,n.detailHighlights=a,n.hidden=!1):(n.labelHighlights=void 0,n.descriptionHighlights=void 0,n.detailHighlights=void 0,n.hidden=n.item?!n.item.alwaysShow:!0),n.item?n.separator=void 0:n.separator&&(n.hidden=!0),!this.sortByLabel){const l=n.index&&this._inputElements[n.index-1]||void 0;l?.type==="separator"&&!l.buttons&&(i=l),i&&!n.hidden&&(n.separator=i,i=void 0)}})}return this._setElementsToTree(this._sortByLabel&&e?this._itemElements:this._elementTree),this._tree.layout(),!0}toggleCheckbox(){this._elementCheckedEventBufferer.bufferEvents(()=>{const e=this._tree.getFocus().filter(i=>i instanceof qn),t=this._allVisibleChecked(e);for(const i of e)i.checkboxDisabled||(i.checked=!t)})}style(e){this._tree.style(e)}toggleHover(){const e=this._tree.getFocus()[0];if(!e?.saneTooltip||!(e instanceof qn))return;if(this._lastHover&&!this._lastHover.isDisposed){this._lastHover.dispose();return}this.showHover(e);const t=new U;t.add(this._tree.onDidChangeFocus(i=>{i.elements[0]instanceof qn&&this.showHover(i.elements[0])})),this._lastHover&&t.add(this._lastHover),this._elementDisposable.add(t)}_setElementsToTree(e){const t=new Array;for(const i of e)i instanceof df?t.push({element:i,collapsible:!1,collapsed:!1,children:i.children.map(n=>({element:n,collapsible:!1,collapsed:!1}))}):t.push({element:i,collapsible:!1,collapsed:!1});this._tree.setChildren(null,t)}_allVisibleChecked(e,t=!0){for(let i=0,n=e.length;i<n;i++){const o=e[i];if(!o.hidden&&o.item.pickable!==!1)if(o.checked)t=!0;else return!1}return t}_updateCheckedObservables(){Vt(e=>{this._allVisibleCheckedObservable.set(this._allVisibleChecked(this._itemElements,!1),e);const t=this._itemElements.filter(i=>i.checked).length;this._checkedCountObservable.set(t,e),this._checkedElementsObservable.set(this.getCheckedElements(),e)})}showHover(e){this._lastHover&&!this._lastHover.isDisposed&&(this.hoverDelegate.onDidHideHover?.(),this._lastHover?.dispose()),!(!e.element||!e.saneTooltip)&&(this._lastHover=this.hoverDelegate.showHover({content:e.saneTooltip,target:e.element,linkHandler:t=>{this.linkOpenerDelegate(t)},appearance:{showPointer:!0},container:this._container,position:{hoverPosition:1}},!1))}};r2([Ai],I0.prototype,"onDidChangeFocus",null);r2([Ai],I0.prototype,"onDidChangeSelection",null);I0=r2([T9(4,_e),T9(5,Wn)],I0);function h6e(s,e){const{text:t,iconOffsets:i}=e;if(!i||i.length===0)return AY(s,t);const n=U0(t," "),o=t.length-n.length,r=AY(s,n);if(r)for(const a of r){const l=i[a.start+o]+o;a.start+=l,a.end+=l}return r}function AY(s,e){const t=e.toLowerCase().indexOf(s.toLowerCase());return t!==-1?[{start:t,end:t+s.length}]:null}function u6e(s,e,t){const i=s.labelHighlights||[],n=e.labelHighlights||[];return i.length&&!n.length?-1:!i.length&&n.length?1:i.length===0&&n.length===0?0:a6e(s.saneSortLabel,e.saneSortLabel,t)}function g6e(s,e={}){Dn.registerCommandAndKeybindingRule({weight:200,when:o2,metadata:{description:m(1758,"Used while in the context of any kind of quick input. If you change one keybinding for this command, you should change all of the other keybindings (modifier variants) of this command as well.")},...s,secondary:yV(s.primary,s.secondary??[],e)})}function uo(s,e={}){Dn.registerCommandAndKeybindingRule({weight:200,when:K.and(K.or(K.equals(k0,"quickPick"),K.equals(k0,"quickTree")),o2),metadata:{description:m(1759,"Used while in the context of the quick pick. If you change one keybinding for this command, you should change all of the other keybindings (modifier variants) of this command as well.")},...s,secondary:yV(s.primary,s.secondary??[],e)})}const D0=Ye?256:2048;function yV(s,e,t={}){return t.withAltMod&&e.push(512+s),t.withCtrlMod&&(e.push(D0+s),t.withAltMod&&e.push(512+D0+s)),t.withCmdMod&&Ye&&(e.push(2048+s),t.withCtrlMod&&e.push(2304+s),t.withAltMod&&(e.push(2560+s),t.withCtrlMod&&e.push(2816+s))),e}function gr(s,e){return t=>{const i=t.get(bs).currentQuickInput;if(i)return e&&i.quickNavigate?i.focus(e):i.focus(s)}}uo({id:"quickInput.pageNext",primary:12,handler:gr(Kt.NextPage)},{withAltMod:!0,withCtrlMod:!0,withCmdMod:!0});uo({id:"quickInput.pagePrevious",primary:11,handler:gr(Kt.PreviousPage)},{withAltMod:!0,withCtrlMod:!0,withCmdMod:!0});uo({id:"quickInput.first",primary:D0+14,handler:gr(Kt.First)},{withAltMod:!0,withCmdMod:!0});uo({id:"quickInput.last",primary:D0+13,handler:gr(Kt.Last)},{withAltMod:!0,withCmdMod:!0});uo({id:"quickInput.next",primary:18,handler:gr(Kt.Next)},{withCtrlMod:!0});uo({id:"quickInput.previous",primary:16,handler:gr(Kt.Previous)},{withCtrlMod:!0});const MY=m(1760,"If we're in quick access mode, this will navigate to the next item. If we are not in quick access mode, this will navigate to the next separator."),OY=m(1761,"If we're in quick access mode, this will navigate to the previous item. If we are not in quick access mode, this will navigate to the previous separator.");Ye?(uo({id:"quickInput.nextSeparatorWithQuickAccessFallback",primary:2066,handler:gr(Kt.NextSeparator,Kt.Next),metadata:{description:MY}}),uo({id:"quickInput.nextSeparator",primary:2578,secondary:[2322],handler:gr(Kt.NextSeparator)},{withCtrlMod:!0}),uo({id:"quickInput.previousSeparatorWithQuickAccessFallback",primary:2064,handler:gr(Kt.PreviousSeparator,Kt.Previous),metadata:{description:OY}}),uo({id:"quickInput.previousSeparator",primary:2576,secondary:[2320],handler:gr(Kt.PreviousSeparator)},{withCtrlMod:!0})):(uo({id:"quickInput.nextSeparatorWithQuickAccessFallback",primary:530,handler:gr(Kt.NextSeparator,Kt.Next),metadata:{description:MY}}),uo({id:"quickInput.nextSeparator",primary:2578,handler:gr(Kt.NextSeparator)}),uo({id:"quickInput.previousSeparatorWithQuickAccessFallback",primary:528,handler:gr(Kt.PreviousSeparator,Kt.Previous),metadata:{description:OY}}),uo({id:"quickInput.previousSeparator",primary:2576,handler:gr(Kt.PreviousSeparator)}));Dn.registerCommandAndKeybindingRule({id:"quickInput.accept",primary:3,weight:200,when:K.and(K.notEquals(k0,"quickWidget"),o2,K.not("isComposing")),metadata:{description:m(1762,"Used while in the context of some quick input. If you change one keybinding for this command, you should change all of the other keybindings (modifier variants) of this command as well.")},handler:s=>{s.get(bs).currentQuickInput?.accept()},secondary:yV(3,[],{withAltMod:!0,withCtrlMod:!0,withCmdMod:!0})});uo({id:"quickInput.acceptInBackground",when:K.and(o2,K.equals(k0,"quickPick"),K.or(yH.negate(),i6e)),primary:17,weight:250,handler:s=>{s.get(bs).currentQuickInput?.accept(!0)}},{withAltMod:!0,withCtrlMod:!0,withCmdMod:!0});g6e({id:"quickInput.hide",primary:9,handler:s=>{s.get(bs).currentQuickInput?.hide()}},{withAltMod:!0,withCtrlMod:!0,withCmdMod:!0});uo({id:"quickInput.toggleHover",primary:D0|10,handler:s=>{s.get(bs).toggleHover()}});var f6e=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},p6e=function(s,e){return function(t,i){e(t,i,s)}},R9;const t1=ie;var pm;let ST=(pm=class extends F{constructor(e,t,i,n){super(),this._hoverDelegate=e,this._buttonTriggeredEmitter=t,this.onCheckedEvent=i,this._themeService=n,this.templateId=R9.ID}renderTemplate(e){const t=new U,i=X(e,t1(".quick-input-tree-entry")),n=t.add(new Nne("",!1,{...$7,size:15}));i.appendChild(n.domNode);const o=X(i,t1("label.quick-input-tree-label")),r=X(o,t1(".quick-input-tree-rows")),a=X(r,t1(".quick-input-tree-row")),l=RR(a,t1(".quick-input-tree-icon")),c=t.add(new QS(a,{supportHighlights:!0,supportDescriptionHighlights:!0,supportIcons:!0,hoverDelegate:this._hoverDelegate})),d=t.add(new bo(i,this._hoverDelegate?{hoverDelegate:this._hoverDelegate}:void 0));return d.domNode.classList.add("quick-input-tree-entry-action-bar"),{toDisposeTemplate:t,entry:i,checkbox:n,icon:l,label:c,actionBar:d,toDisposeElement:new U}}renderElement(e,t,i,n){const o=i.toDisposeElement,r=e.element;if(r.pickable===!1?i.checkbox.domNode.style.display="none":(i.checkbox.domNode.style.display="",i.checkbox.checked=r.checked??!1,o.add(oe.filter(this.onCheckedEvent,h=>h.item===r)(h=>i.checkbox.checked=h.checked)),r.disabled&&i.checkbox.disable()),r.iconPath){const h=wh(this._themeService.getColorTheme().type)?r.iconPath.dark:r.iconPath.light??r.iconPath.dark,u=Le.revive(h);i.icon.className="quick-input-tree-icon",i.icon.style.backgroundImage=Mc(u)}else i.icon.style.backgroundImage="",i.icon.className=r.iconClass?`quick-input-tree-icon ${r.iconClass}`:"";const{labelHighlights:a,descriptionHighlights:l}=e.filterData||{};let c;r.description&&(c={markdown:{value:Aa(r.description),supportThemeIcons:!0},markdownNotSupportedFallback:r.description}),i.label.setLabel(r.label,r.description,{matches:a,descriptionMatches:l,extraClasses:r.iconClasses,italic:r.italic,strikethrough:r.strikethrough,labelEscapeNewLines:!0,descriptionTitle:c});const d=r.buttons;d&&d.length?(i.actionBar.push(d.map((h,u)=>Aw(h,`tree-${u}`,()=>this._buttonTriggeredEmitter.fire({item:r,button:h}))),{icon:!0,label:!1}),i.entry.classList.add("has-actions")):i.entry.classList.remove("has-actions")}disposeElement(e,t,i,n){i.toDisposeElement.clear(),i.actionBar.clear()}disposeTemplate(e){e.toDisposeElement.dispose(),e.toDisposeTemplate.dispose()}},R9=pm,pm.ID="quickInputTreeElement",pm);ST=R9=f6e([p6e(3,wi)],ST);class m6e{getHeight(e){return 22}getTemplateId(e){return ST.ID}}function _6e(s){let e=!1,t=!1,i=!1;for(const o of s){switch(o.element?.checked){case"mixed":i=!0;break;case!0:e=!0;break;default:t=!0;break}if(e&&t&&i)break}return t?i||e?"mixed":!1:i?"mixed":e}class b6e{constructor(e){this.onCheckedEvent=e}getWidgetAriaLabel(){return m(1772,"Quick Tree")}getAriaLabel(e){return e.ariaLabel||[e.label,e.description].map(t=>Nte(t)).filter(t=>!!t).join(", ")}getWidgetRole(){return"tree"}getRole(e){return"checkbox"}isChecked(e){return{get value(){return e.checked==="mixed"?"mixed":!!e.checked},onDidChange:t=>oe.filter(this.onCheckedEvent,i=>i.item===e)(i=>t())}}}class w6e{constructor(){this.filterValue="",this.matchOnLabel=!0,this.matchOnDescription=!1}filter(e,t){if(!this.filterValue||!(this.matchOnLabel||this.matchOnDescription))return e.children?{visibility:2,data:{}}:{visibility:1,data:{}};const i=this.matchOnLabel?q1(this.filterValue,Kf(e.label))??void 0:void 0,n=this.matchOnDescription?q1(this.filterValue,Kf(e.description||""))??void 0:void 0;return{visibility:t===1||i||n?1:e.children?2:0,data:{labelHighlights:i,descriptionHighlights:n}}}}class v6e extends F{constructor(){super(...arguments),this._sortByLabel=!0}get sortByLabel(){return this._sortByLabel}set sortByLabel(e){this._sortByLabel=e}compare(e,t){if(!this._sortByLabel)return 0;if(e.label<t.label)return-1;if(e.label>t.label)return 1;if(e.description&&t.description){if(e.description<t.description)return-1;if(e.description>t.description)return 1}else{if(e.description)return-1;if(t.description)return 1}return 0}}var C6e=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},y6e=function(s,e){return function(t,i){e(t,i,s)}};const S6e=ie;let A9=class extends F{constructor(e,t,i){super(),this.instantiationService=i,this._onDidTriggerButton=this._register(new M),this._onDidChangeCheckboxState=this._register(new M),this.onDidChangeCheckboxState=this._onDidChangeCheckboxState.event,this._onDidCheckedLeafItemsChange=this._register(new M),this._onLeave=new M,this.onLeave=this._onLeave.event,this._onDidAccept=this._register(new M),this.onDidAccept=this._onDidAccept.event,this._container=X(e,S6e(".quick-input-tree")),this._renderer=this._register(this.instantiationService.createInstance(ST,t,this._onDidTriggerButton,this.onDidChangeCheckboxState)),this._filter=this.instantiationService.createInstance(w6e),this._sorter=this._register(new v6e),this._tree=this._register(this.instantiationService.createInstance(lD,"QuickInputTree",this._container,new m6e,[this._renderer],{accessibilityProvider:new b6e(this.onDidChangeCheckboxState),horizontalScrolling:!1,multipleSelectionSupport:!1,findWidgetEnabled:!1,alwaysConsumeMouseWheel:!0,hideTwistiesOfChildlessElements:!0,renderIndentGuides:Gm.None,expandOnDoubleClick:!0,expandOnlyOnTwistieClick:!0,disableExpandOnSpacebar:!0,sorter:this._sorter,filter:this._filter})),this.registerOnOpenListener()}get tree(){return this._tree}get displayed(){return this._container.style.display!=="none"}set displayed(e){this._container.style.display=e?"":"none"}get sortByLabel(){return this._sorter.sortByLabel}set sortByLabel(e){this._sorter.sortByLabel=e,this._tree.resort(null,!0)}getActiveDescendant(){return this._tree.getHTMLElement().getAttribute("aria-activedescendant")}updateFilterOptions(e){e.matchOnLabel!==void 0&&(this._filter.matchOnLabel=e.matchOnLabel),e.matchOnDescription!==void 0&&(this._filter.matchOnDescription=e.matchOnDescription),this._tree.refilter()}layout(e){this._tree.getHTMLElement().style.maxHeight=e?`${Math.floor(e/44)*44+6}px`:"",this._tree.layout()}registerOnOpenListener(){this._register(this._tree.onDidOpen(e=>{const t=e.element;if(!t||t.disabled)return;if(t.pickable===!1){this._tree.setFocus([t]),this._onDidAccept.fire();return}const i=t.checked!==!0;if((t.checked??!1)===i)return;t.checked=i,this._tree.rerender(t);const n=new Set,o=[...this._tree.getNode(t).children];for(;o.length;){const a=o.shift();a?.element&&!n.has(a.element)&&(n.add(a.element),(a.element.checked??!1)!==t.checked&&(a.element.checked=t.checked,this._tree.rerender(a.element)),o.push(...a.children))}let r=this._tree.getParentElement(t);for(;r;){const a=[...this._tree.getNode(r).children],l=_6e(a);(r.checked??!1)!==l&&(r.checked=l,this._tree.rerender(r)),r=this._tree.getParentElement(r)}this._onDidChangeCheckboxState.fire({item:t,checked:t.checked??!1}),this._onDidCheckedLeafItemsChange.fire(this.getCheckedLeafItems())}))}getCheckedLeafItems(){const e=new Set,t=[...this._tree.getNode().children],i=new Array;for(;t.length;){const n=t.shift();!n?.element||e.has(n.element)||n.element.checked&&(e.add(n.element),t.push(...n.children),n.element.children||i.push(n.element))}return i}};A9=C6e([y6e(2,_e)],A9);var Hre=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},tp=function(s,e){return function(t,i){e(t,i,s)}},M9;const tr=ie,qO="workbench.quickInput.viewState";var mm;let O9=(mm=class extends F{get currentQuickInput(){return this.controller??void 0}get container(){return this._container}constructor(e,t,i,n,o){super(),this.options=e,this.layoutService=t,this.instantiationService=i,this.storageService=o,this.enabled=!0,this.onDidAcceptEmitter=this._register(new M),this.onDidCustomEmitter=this._register(new M),this.onDidTriggerButtonEmitter=this._register(new M),this.keyMods={ctrlCmd:!1,alt:!1},this.controller=null,this.onShowEmitter=this._register(new M),this.onShow=this.onShowEmitter.event,this.onHideEmitter=this._register(new M),this.onHide=this.onHideEmitter.event,this.inQuickInputContext=X9e.bindTo(n),this.quickInputTypeContext=e6e.bindTo(n),this.endOfQuickInputBoxContext=t6e.bindTo(n),this.idPrefix=e.idPrefix,this._container=e.container,this.styles=e.styles,this._register(oe.runAndSubscribe(G0,({window:r,disposables:a})=>this.registerKeyModsListeners(r,a),{window:Nt,disposables:this._store})),this._register(Qhe(r=>{this.ui&&be(this.ui.container)===r&&(this.reparentUI(this.layoutService.mainContainer),this.layout(this.layoutService.mainContainerDimension,this.layoutService.mainContainerOffset.quickPickTop))})),this.viewState=this.loadViewState()}registerKeyModsListeners(e,t){const i=n=>{this.keyMods.ctrlCmd=n.ctrlKey||n.metaKey,this.keyMods.alt=n.altKey};for(const n of[se.KEY_DOWN,se.KEY_UP,se.MOUSE_DOWN])t.add(W(e,n,i,!0))}getUI(e){if(this.ui)return e&&be(this._container)!==be(this.layoutService.activeContainer)&&(this.reparentUI(this.layoutService.activeContainer),this.layout(this.layoutService.activeContainerDimension,this.layoutService.activeContainerOffset.quickPickTop)),this.ui;const t=X(this._container,tr(".quick-input-widget.show-file-icons"));t.tabIndex=-1,t.style.display="none";const i=ca(t),n=X(t,tr(".quick-input-titlebar")),o=this._register(new bo(n,{hoverDelegate:this.options.hoverDelegate}));o.domNode.classList.add("quick-input-left-action-bar");const r=X(n,tr(".quick-input-title")),a=this._register(new bo(n,{hoverDelegate:this.options.hoverDelegate}));a.domNode.classList.add("quick-input-right-action-bar");const l=X(t,tr(".quick-input-header")),c=this._register(new Nne(m(1763,"Toggle all checkboxes"),!1,{...$7,size:15}));X(l,c.domNode),this._register(c.onChange(()=>{const z=c.checked;P.setAllVisibleChecked(z===!0)})),this._register(W(c.domNode,se.CLICK,z=>{(z.x||z.y)&&g.setFocus()}));const d=X(l,tr(".quick-input-description")),h=X(l,tr(".quick-input-and-message")),u=X(h,tr(".quick-input-filter")),g=this._register(new o6e(u,this.styles.inputBox,this.styles.toggle));g.setAttribute("aria-describedby",`${this.idPrefix}message`);const f=X(u,tr(".quick-input-visible-count"));f.setAttribute("aria-live","polite"),f.setAttribute("aria-atomic","true");const p=this._register(new C5(f,{countFormat:m(1764,"{0} Results")},this.styles.countBadge)),_=X(u,tr(".quick-input-count"));_.setAttribute("aria-live","polite");const b=this._register(new C5(_,{countFormat:m(1765,"{0} Selected")},this.styles.countBadge)),w=this._register(new bo(l,{hoverDelegate:this.options.hoverDelegate}));w.domNode.classList.add("quick-input-inline-action-bar");const C=X(l,tr(".quick-input-action")),v=this._register(new VI(C,this.styles.button));v.label=m(1766,"OK"),this._register(v.onDidClick(z=>{this.onDidAcceptEmitter.fire()}));const y=X(l,tr(".quick-input-action")),x=this._register(new VI(y,{...this.styles.button,supportIcons:!0}));x.label=m(1767,"Custom"),this._register(x.onDidClick(z=>{this.onDidCustomEmitter.fire()}));const L=X(h,tr(`#${this.idPrefix}message.quick-input-message`)),E=this._register(new D9(t,this.styles.progressBar));E.getContainer().classList.add("quick-input-progress");const I=X(t,tr(".quick-input-html-widget"));I.tabIndex=-1;const R=X(t,tr(".quick-input-description")),O=this.idPrefix+"list",P=this._register(this.instantiationService.createInstance(I0,t,this.options.hoverDelegate,this.options.linkOpenerDelegate,O));g.setAttribute("aria-controls",O),this._register(P.onDidChangeFocus(()=>{g.hasFocus()&&g.setAttribute("aria-activedescendant",P.getActiveDescendant()??"")})),this._register(P.onChangedAllVisibleChecked(z=>{c.checked=z})),this._register(P.onChangedVisibleCount(z=>{p.setCount(z)})),this._register(P.onChangedCheckedCount(z=>{V0(()=>b.setCount(z))})),this._register(P.onLeave(()=>{setTimeout(()=>{this.controller&&(g.setFocus(),this.controller instanceof py&&this.controller.canSelectMany&&P.clearFocus())},0)}));const B=this._register(this.instantiationService.createInstance(A9,t,this.options.hoverDelegate));this._register(B.tree.onDidChangeFocus(()=>{g.hasFocus()&&g.setAttribute("aria-activedescendant",B.getActiveDescendant()??"")})),this._register(B.onLeave(()=>{setTimeout(()=>{this.controller&&(g.setFocus(),B.tree.setFocus([]))},0)})),this._register(B.onDidAccept(()=>{this.onDidAcceptEmitter.fire()})),this._register(B.tree.onDidChangeContentHeight(()=>this.updateLayout()));const $=oa(t);return this._register($),this._register(W(t,se.FOCUS,z=>{const Q=this.getUI();if(wn(z.relatedTarget,Q.inputContainer)){const J=Q.inputBox.isSelectionAtEnd();this.endOfQuickInputBoxContext.get()!==J&&this.endOfQuickInputBoxContext.set(J)}wn(z.relatedTarget,Q.container)||(this.inQuickInputContext.set(!0),this.previousFocusElement=ki(z.relatedTarget)?z.relatedTarget:void 0)},!0)),this._register($.onDidBlur(()=>{!this.getUI().ignoreFocusOut&&!this.options.ignoreFocusOut()&&this.hide(HS.Blur),this.inQuickInputContext.set(!1),this.endOfQuickInputBoxContext.set(!1),this.previousFocusElement=void 0})),this._register(g.onKeyDown(z=>{const Q=this.getUI().inputBox.isSelectionAtEnd();this.endOfQuickInputBoxContext.get()!==Q&&this.endOfQuickInputBoxContext.set(Q),g.removeAttribute("aria-activedescendant")})),this._register(W(t,se.FOCUS,z=>{g.setFocus()})),this.dndController=this._register(this.instantiationService.createInstance(P9,this._container,t,[{node:n,includeChildren:!0},{node:l,includeChildren:!1}],this.viewState)),this._register(De(z=>{const Q=this.dndController?.dndViewState.read(z);Q&&(Q.top!==void 0&&Q.left!==void 0?this.viewState={...this.viewState,top:Q.top,left:Q.left}:this.viewState=void 0,this.updateLayout(),Q.done&&this.saveViewState(this.viewState))})),this.ui={container:t,styleSheet:i,leftActionBar:o,titleBar:n,title:r,description1:R,description2:d,widget:I,rightActionBar:a,inlineActionBar:w,checkAll:c,inputContainer:h,filterContainer:u,inputBox:g,visibleCountContainer:f,visibleCount:p,countContainer:_,count:b,okContainer:C,ok:v,message:L,customButtonContainer:y,customButton:x,list:P,tree:B,progressBar:E,onDidAccept:this.onDidAcceptEmitter.event,onDidCustom:this.onDidCustomEmitter.event,onDidTriggerButton:this.onDidTriggerButtonEmitter.event,ignoreFocusOut:!1,keyMods:this.keyMods,show:z=>this.show(z),hide:()=>this.hide(),setVisibilities:z=>this.setVisibilities(z),setEnabled:z=>this.setEnabled(z),setContextKey:z=>this.options.setContextKey(z),linkOpenerDelegate:z=>this.options.linkOpenerDelegate(z)},this.updateStyles(),this.ui}reparentUI(e){this.ui&&(this._container=e,X(this._container,this.ui.container),this.dndController?.reparentUI(this._container))}pick(e,t={},i=Ke.None){return new Promise((n,o)=>{let r=d=>{r=n,t.onKeyMods?.(a.keyMods),n(d)};if(i.isCancellationRequested){r(void 0);return}const a=this.createQuickPick({useSeparators:!0});let l;const c=[a,a.onDidAccept(()=>{if(a.canSelectMany)r(a.selectedItems.slice()),a.hide();else{const d=a.activeItems[0];d&&(r(d),a.hide())}}),a.onDidChangeActive(d=>{const h=d[0];h&&t.onDidFocus&&t.onDidFocus(h)}),a.onDidChangeSelection(d=>{if(!a.canSelectMany){const h=d[0];h&&(r(h),a.hide())}}),a.onDidTriggerItemButton(d=>t.onDidTriggerItemButton&&t.onDidTriggerItemButton({...d,removeItem:()=>{const h=a.items.indexOf(d.item);if(h!==-1){const u=a.items.slice(),g=u.splice(h,1),f=a.activeItems.filter(_=>_!==g[0]),p=a.keepScrollPosition;a.keepScrollPosition=!0,a.items=u,f&&(a.activeItems=f),a.keepScrollPosition=p}}})),a.onDidTriggerSeparatorButton(d=>t.onDidTriggerSeparatorButton?.(d)),a.onDidChangeValue(d=>{l&&!d&&(a.activeItems.length!==1||a.activeItems[0]!==l)&&(a.activeItems=[l])}),i.onCancellationRequested(()=>{a.hide()}),a.onDidHide(()=>{xt(c),r(void 0)})];a.title=t.title,t.value&&(a.value=t.value),a.canSelectMany=!!t.canPickMany,a.placeholder=t.placeHolder,a.prompt=t.prompt,a.ignoreFocusOut=!!t.ignoreFocusLost,a.matchOnDescription=!!t.matchOnDescription,a.matchOnDetail=!!t.matchOnDetail,t.sortByLabel!==void 0&&(a.sortByLabel=t.sortByLabel),a.matchOnLabel=t.matchOnLabel===void 0||t.matchOnLabel,a.quickNavigate=t.quickNavigate,a.hideInput=!!t.hideInput,a.contextKey=t.contextKey,a.busy=!0,Promise.all([e,t.activeItem]).then(([d,h])=>{l=h,a.busy=!1,a.items=d,a.canSelectMany&&(a.selectedItems=d.filter(u=>u.type!=="separator"&&u.picked)),l&&(a.activeItems=[l])}),a.show(),Promise.resolve(e).then(void 0,d=>{o(d),a.hide()})})}setValidationOnInput(e,t){t&&bn(t)?(e.severity=fi.Error,e.validationMessage=t):t&&!bn(t)?(e.severity=t.severity,e.validationMessage=t.content):(e.severity=fi.Ignore,e.validationMessage=void 0)}input(e={},t=Ke.None){return new Promise(i=>{if(t.isCancellationRequested){i(void 0);return}const n=this.createInputBox(),o=e.validateInput||(()=>Promise.resolve(void 0)),r=oe.debounce(n.onDidChangeValue,(d,h)=>h,100);let a=e.value||"",l=Promise.resolve(o(a));const c=[n,r(d=>{d!==a&&(l=Promise.resolve(o(d)),a=d),l.then(h=>{d===a&&this.setValidationOnInput(n,h)})}),n.onDidAccept(()=>{const d=n.value;d!==a&&(l=Promise.resolve(o(d)),a=d),l.then(h=>{!h||!bn(h)&&h.severity!==fi.Error?(i(d),n.hide()):d===a&&this.setValidationOnInput(n,h)})}),t.onCancellationRequested(()=>{n.hide()}),n.onDidHide(()=>{xt(c),i(void 0)})];n.title=e.title,n.value=e.value||"",n.valueSelection=e.valueSelection,n.prompt=e.prompt,n.placeholder=e.placeHolder,n.password=!!e.password,n.ignoreFocusOut=!!e.ignoreFocusLost,n.show()})}createQuickPick(e={useSeparators:!1}){const t=this.getUI(!0);return new py(t)}createInputBox(){const e=this.getUI(!0);return new n6e(e)}show(e){const t=this.getUI(!0);this.onShowEmitter.fire();const i=this.controller;this.controller=e,i?.didHide(),this.setEnabled(!0),t.leftActionBar.clear(),t.title.textContent="",t.description1.textContent="",t.description2.textContent="",vn(t.widget),t.rightActionBar.clear(),t.inlineActionBar.clear(),t.checkAll.checked=!1,t.inputBox.placeholder="",t.inputBox.password=!1,t.inputBox.showDecoration(fi.Ignore),t.visibleCount.setCount(0),t.count.setCount(0),vn(t.message),t.progressBar.stop(),t.progressBar.getContainer().setAttribute("aria-hidden","true"),t.list.setElements([]),t.list.matchOnDescription=!1,t.list.matchOnDetail=!1,t.list.matchOnLabel=!0,t.list.sortByLabel=!0,t.tree.updateFilterOptions({matchOnDescription:!1,matchOnLabel:!0}),t.tree.sortByLabel=!0,t.ignoreFocusOut=!1,t.inputBox.toggles=void 0;const n=this.options.backKeybindingLabel();E9.tooltip=n?m(1768,"Back ({0})",n):m(1769,"Back"),t.container.style.display="",this.updateLayout(),this.dndController?.layoutContainer(),t.inputBox.setFocus(),this.quickInputTypeContext.set(e.type)}isVisible(){return!!this.ui&&this.ui.container.style.display!=="none"}setVisibilities(e){const t=this.getUI();t.title.style.display=e.title?"":"none",t.description1.style.display=e.description&&(e.inputBox||e.checkAll)?"":"none",t.description2.style.display=e.description&&!(e.inputBox||e.checkAll)?"":"none",t.checkAll.domNode.style.display=e.checkAll?"":"none",t.inputContainer.style.display=e.inputBox?"":"none",t.filterContainer.style.display=e.inputBox?"":"none",t.visibleCountContainer.style.display=e.visibleCount?"":"none",t.countContainer.style.display=e.count?"":"none",t.okContainer.style.display=e.ok?"":"none",t.customButtonContainer.style.display=e.customButton?"":"none",t.message.style.display=e.message?"":"none",t.progressBar.getContainer().style.display=e.progressBar?"":"none",t.list.displayed=!!e.list,t.tree.displayed=!!e.tree,t.container.classList.toggle("show-checkboxes",!!e.checkBox),t.container.classList.toggle("hidden-input",!e.inputBox&&!e.description),this.updateLayout()}setEnabled(e){if(e!==this.enabled){this.enabled=e;const t=this.getUI();for(const i of t.leftActionBar.viewItems)i.action.enabled=e;for(const i of t.rightActionBar.viewItems)i.action.enabled=e;e?t.checkAll.enable():t.checkAll.disable(),t.inputBox.enabled=e,t.ok.enabled=e,t.list.enabled=e}}hide(e){const t=this.controller;if(!t)return;t.willHide(e);const i=this.ui?.container,n=i&&!RQ(i);if(this.controller=null,this.onHideEmitter.fire(),i&&(i.style.display="none"),!n){let o=this.previousFocusElement;for(;o&&!o.offsetParent;)o=o.parentElement??void 0;o?.offsetParent?(o.focus(),this.previousFocusElement=void 0):this.options.returnFocus()}t.didHide(e)}toggleHover(){this.isVisible()&&this.controller instanceof py&&this.getUI().list.toggleHover()}layout(e,t){this.dimension=e,this.titleBarOffset=t,this.updateLayout()}updateLayout(){if(this.ui&&this.isVisible()){const e=this.ui.container.style,t=Math.min(this.dimension.width*.62,M9.MAX_WIDTH);e.width=t+"px",e.top=`${this.viewState?.top?Math.round(this.dimension.height*this.viewState.top):this.titleBarOffset}px`,e.left=`${Math.round(this.dimension.width*(this.viewState?.left??.5)-t/2)}px`,this.ui.inputBox.layout(),this.ui.list.layout(this.dimension&&this.dimension.height*.4),this.ui.tree.layout(this.dimension&&this.dimension.height*.4)}}applyStyles(e){this.styles=e,this.updateStyles()}updateStyles(){if(this.ui){const{quickInputTitleBackground:e,quickInputBackground:t,quickInputForeground:i,widgetBorder:n,widgetShadow:o}=this.styles.widget;this.ui.titleBar.style.backgroundColor=e??"",this.ui.container.style.backgroundColor=t??"",this.ui.container.style.color=i??"",this.ui.container.style.border=n?`1px solid ${n}`:"",this.ui.container.style.boxShadow=o?`0 0 8px 2px ${o}`:"",this.ui.list.style(this.styles.list),this.ui.tree.tree.style(this.styles.list);const r=[];this.styles.pickerGroup.pickerGroupBorder&&r.push(`.quick-input-list .quick-input-list-entry { border-top-color:  ${this.styles.pickerGroup.pickerGroupBorder}; }`),this.styles.pickerGroup.pickerGroupForeground&&r.push(`.quick-input-list .quick-input-list-separator { color:  ${this.styles.pickerGroup.pickerGroupForeground}; }`),this.styles.pickerGroup.pickerGroupForeground&&r.push(".quick-input-list .quick-input-list-separator-as-item { color: var(--vscode-descriptionForeground); }"),(this.styles.keybindingLabel.keybindingLabelBackground||this.styles.keybindingLabel.keybindingLabelBorder||this.styles.keybindingLabel.keybindingLabelBottomBorder||this.styles.keybindingLabel.keybindingLabelShadow||this.styles.keybindingLabel.keybindingLabelForeground)&&(r.push(".quick-input-list .monaco-keybinding > .monaco-keybinding-key {"),this.styles.keybindingLabel.keybindingLabelBackground&&r.push(`background-color: ${this.styles.keybindingLabel.keybindingLabelBackground};`),this.styles.keybindingLabel.keybindingLabelBorder&&r.push(`border-color: ${this.styles.keybindingLabel.keybindingLabelBorder};`),this.styles.keybindingLabel.keybindingLabelBottomBorder&&r.push(`border-bottom-color: ${this.styles.keybindingLabel.keybindingLabelBottomBorder};`),this.styles.keybindingLabel.keybindingLabelShadow&&r.push(`box-shadow: inset 0 -1px 0 ${this.styles.keybindingLabel.keybindingLabelShadow};`),this.styles.keybindingLabel.keybindingLabelForeground&&r.push(`color: ${this.styles.keybindingLabel.keybindingLabelForeground};`),r.push("}"));const a=r.join(`
`);a!==this.ui.styleSheet.textContent&&(this.ui.styleSheet.textContent=a)}}loadViewState(){try{const e=JSON.parse(this.storageService.get(qO,-1,"{}"));if(e.top!==void 0||e.left!==void 0)return e}catch{}}saveViewState(e){this.layoutService.activeContainer===this.layoutService.mainContainer&&(e!==void 0?this.storageService.store(qO,JSON.stringify(e),-1,1):this.storageService.remove(qO,-1))}},M9=mm,mm.MAX_WIDTH=600,mm);O9=M9=Hre([tp(1,qc),tp(2,_e),tp(3,Ae),tp(4,Bs)],O9);let P9=class extends F{constructor(e,t,i,n,o,r,a){super(),this._container=e,this._quickInputContainer=t,this._quickInputDragAreas=i,this._layoutService=o,this.configurationService=a,this.dndViewState=Re(this,void 0),this._snapThreshold=20,this._snapLineHorizontalRatio=.25,this._quickInputAlignmentContext=J9e.bindTo(r);const l=tEe(this.configurationService)==="custom";this._controlsOnLeft=l&&T2===1,this._controlsOnRight=l&&(T2===3||T2===2),this._registerLayoutListener(),this.registerMouseListeners(),this.dndViewState.set({...n,done:!0},void 0)}reparentUI(e){this._container=e}layoutContainer(e=this._layoutService.activeContainerDimension){const t=this.dndViewState.get(),i=this._quickInputContainer.getBoundingClientRect();if(t?.top&&t?.left){const n=Math.round(t.left*100)/100,o=e.width,r=i.width,a=n*o-r/2;this._layout(t.top*e.height,a)}}_registerLayoutListener(){this._register(oe.filter(this._layoutService.onDidLayoutContainer,e=>e.container===this._container)(e=>this.layoutContainer(e.dimension)))}registerMouseListeners(){const e=this._quickInputContainer;this._register(jz(e,t=>{const i=new is(be(e),t);i.detail===2&&this._quickInputDragAreas.some(({node:n,includeChildren:o})=>o?wn(i.target,n):i.target===n)&&this.dndViewState.set({top:void 0,left:void 0,done:!0},void 0)})),this._register(IQ(e,t=>{const i=be(this._layoutService.activeContainer),n=new is(i,t);if(!this._quickInputDragAreas.some(({node:h,includeChildren:u})=>u?wn(n.target,h):n.target===h))return;const o=this._quickInputContainer.getBoundingClientRect(),r=n.browserEvent.clientX-o.left,a=n.browserEvent.clientY-o.top;let l=!1;const c=nue(i,h=>{new is(i,h).preventDefault(),l||(l=!0),this._layout(h.clientY-a,h.clientX-r)}),d=jz(i,h=>{if(l){const u=this.dndViewState.get();this.dndViewState.set({top:u?.top,left:u?.left,done:!0},void 0)}c.dispose(),d.dispose()})}))}_layout(e,t){const i=this._getTopSnapValue(),n=this._getCenterYSnapValue(),o=this._getCenterXSnapValue();e=Math.max(0,Math.min(e,this._container.clientHeight-this._quickInputContainer.clientHeight)),e<this._layoutService.activeContainerOffset.top&&(this._controlsOnLeft?t=Math.max(t,80/JO(kt())):this._controlsOnRight&&(t=Math.min(t,this._container.clientWidth-this._quickInputContainer.clientWidth-140/JO(kt()))));const r=Math.abs(e-i)<this._snapThreshold;e=r?i:e;const a=Math.abs(e-n)<this._snapThreshold;e=a?n:e;const l=e/this._container.clientHeight;t=Math.max(0,Math.min(t,this._container.clientWidth-this._quickInputContainer.clientWidth));const c=Math.abs(t-o)<this._snapThreshold;t=c?o:t;const d=this._container.clientWidth,h=this._quickInputContainer.clientWidth,g=(t+h/2)/d;if(this.dndViewState.set({top:l,left:g,done:!1},void 0),c){if(r){this._quickInputAlignmentContext.set("top");return}else if(a){this._quickInputAlignmentContext.set("center");return}}this._quickInputAlignmentContext.set(void 0)}_getTopSnapValue(){return this._layoutService.activeContainerOffset.quickPickTop}_getCenterYSnapValue(){return Math.round(this._container.clientHeight*this._snapLineHorizontalRatio)}_getCenterXSnapValue(){return Math.round(this._container.clientWidth/2)-Math.round(this._quickInputContainer.clientWidth/2)}};P9=Hre([tp(4,qc),tp(5,Ae),tp(6,He)],P9);var x6e=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},i1=function(s,e){return function(t,i){e(t,i,s)}};let F9=class extends zbe{get controller(){return this._controller||(this._controller=this._register(this.createController())),this._controller}get hasController(){return!!this._controller}get currentQuickInput(){return this.controller.currentQuickInput}get quickAccess(){return this._quickAccess||(this._quickAccess=this._register(this.instantiationService.createInstance(k9))),this._quickAccess}constructor(e,t,i,n,o){super(i),this.instantiationService=e,this.contextKeyService=t,this.layoutService=n,this.configurationService=o,this._onShow=this._register(new M),this._onHide=this._register(new M),this.contexts=new Map}createController(e=this.layoutService,t){const i={idPrefix:"quickInput_",container:e.activeContainer,ignoreFocusOut:()=>!1,backKeybindingLabel:()=>{},setContextKey:o=>this.setContextKey(o),linkOpenerDelegate:o=>{this.instantiationService.invokeFunction(r=>{r.get(Ah).open(o,{allowCommands:!0,fromUserGesture:!0})})},returnFocus:()=>e.focus(),styles:this.computeStyles(),hoverDelegate:this._register(this.instantiationService.createInstance(I9))},n=this._register(this.instantiationService.createInstance(O9,{...i,...t}));return n.layout(e.activeContainerDimension,e.activeContainerOffset.quickPickTop),this._register(e.onDidLayoutActiveContainer(o=>{be(e.activeContainer)===be(n.container)&&n.layout(o,e.activeContainerOffset.quickPickTop)})),this._register(e.onDidChangeActiveContainer(()=>{n.isVisible()||n.layout(e.activeContainerDimension,e.activeContainerOffset.quickPickTop)})),this._register(n.onShow(()=>{this.resetContextKeys(),this._onShow.fire()})),this._register(n.onHide(()=>{this.resetContextKeys(),this._onHide.fire()})),n}setContextKey(e){let t;e&&(t=this.contexts.get(e),t||(t=new ce(e,!1).bindTo(this.contextKeyService),this.contexts.set(e,t))),!(t&&t.get())&&(this.resetContextKeys(),t?.set(!0))}resetContextKeys(){this.contexts.forEach(e=>{e.get()&&e.reset()})}pick(e,t,i=Ke.None){return this.controller.pick(e,t,i)}input(e={},t=Ke.None){return this.controller.input(e,t)}createQuickPick(e={useSeparators:!1}){return this.controller.createQuickPick(e)}createInputBox(){return this.controller.createInputBox()}toggleHover(){this.hasController&&this.controller.toggleHover()}updateStyles(){this.hasController&&this.controller.applyStyles(this.computeStyles())}computeStyles(){return{widget:{quickInputBackground:te(tM),quickInputForeground:te(N_e),quickInputTitleBackground:te(R_e),widgetBorder:te(OW),widgetShadow:te(tC)},inputBox:NI,toggle:TI,countBadge:Sie,button:Kke,progressBar:Yke,keybindingLabel:yie,list:f_({listBackground:tM,listFocusBackground:hS,listFocusForeground:dS,listInactiveFocusForeground:dS,listInactiveSelectionIconForeground:VW,listInactiveFocusBackground:hS,listFocusOutline:ri,listInactiveFocusOutline:ri,treeStickyScrollBackground:tM}),pickerGroup:{pickerGroupBorder:te(A_e),pickerGroupForeground:te(qJ)}}}};F9=x6e([i1(0,_e),i1(1,Ae),i1(2,wi),i1(3,qc),i1(4,He)],F9);var Vre=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},kf=function(s,e){return function(t,i){e(t,i,s)}};let B9=class extends F9{constructor(e,t,i,n,o,r){super(t,i,n,new a9(e.getContainerDomNode(),o),r),this.host=void 0;const a=T0.get(e);if(a){const l=a.widget;this.host={_serviceBrand:void 0,get mainContainer(){return l.getDomNode()},getContainer(){return l.getDomNode()},whenContainerStylesLoaded(){},get containers(){return[l.getDomNode()]},get activeContainer(){return l.getDomNode()},get mainContainerDimension(){return e.getLayoutInfo()},get activeContainerDimension(){return e.getLayoutInfo()},get onDidLayoutMainContainer(){return e.onDidLayoutChange},get onDidLayoutActiveContainer(){return e.onDidLayoutChange},get onDidLayoutContainer(){return oe.map(e.onDidLayoutChange,c=>({container:l.getDomNode(),dimension:c}))},get onDidChangeActiveContainer(){return oe.None},get onDidAddContainer(){return oe.None},get mainContainerOffset(){return{top:0,quickPickTop:0}},get activeContainerOffset(){return{top:0,quickPickTop:0}},focus:()=>e.focus()}}else this.host=void 0}createController(){return super.createController(this.host)}};B9=Vre([kf(1,_e),kf(2,Ae),kf(3,wi),kf(4,ht),kf(5,He)],B9);let W9=class{get activeService(){const e=this.codeEditorService.getFocusedCodeEditor();if(!e)throw new Error("Quick input service needs a focused editor to work.");let t=this.mapEditorToService.get(e);if(!t){const i=t=this.instantiationService.createInstance(B9,e);this.mapEditorToService.set(e,t),ym(e.onDidDispose)(()=>{i.dispose(),this.mapEditorToService.delete(e)})}return t}get currentQuickInput(){return this.activeService.currentQuickInput}get quickAccess(){return this.activeService.quickAccess}constructor(e,t){this.instantiationService=e,this.codeEditorService=t,this.mapEditorToService=new Map}pick(e,t,i=Ke.None){return this.activeService.pick(e,t,i)}input(e,t){return this.activeService.input(e,t)}createQuickPick(e={useSeparators:!1}){return this.activeService.createQuickPick(e)}createInputBox(){return this.activeService.createInputBox()}toggleHover(){return this.activeService.toggleHover()}};W9=Vre([kf(0,_e),kf(1,ht)],W9);const wR=class wR{static get(e){return e.getContribution(wR.ID)}constructor(e){this.editor=e,this.widget=new H9(this.editor)}dispose(){this.widget.dispose()}};wR.ID="editor.controller.quickInput";let T0=wR;const vR=class vR{constructor(e){this.codeEditor=e,this.domNode=document.createElement("div"),this.codeEditor.addOverlayWidget(this)}getId(){return vR.ID}getDomNode(){return this.domNode}getPosition(){return{preference:{top:0,left:0}}}dispose(){this.codeEditor.removeOverlayWidget(this)}};vR.ID="editor.contrib.quickInputWidget";let H9=vR;lt(T0.ID,T0,4);var L6e=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},jO=function(s,e){return function(t,i){e(t,i,s)}};let V9=class extends F{constructor(e,t,i){super(),this._contextKeyService=e,this._layoutService=t,this._configurationService=i,this._accessibilitySupport=0,this._onDidChangeScreenReaderOptimized=new M,this._onDidChangeReducedMotion=new M,this._onDidChangeLinkUnderline=new M,this._accessibilityModeEnabledContext=Jv.bindTo(this._contextKeyService);const n=()=>this._accessibilityModeEnabledContext.set(this.isScreenReaderOptimized());this._register(this._configurationService.onDidChangeConfiguration(r=>{r.affectsConfiguration("editor.accessibilitySupport")&&(n(),this._onDidChangeScreenReaderOptimized.fire()),r.affectsConfiguration("workbench.reduceMotion")&&(this._configMotionReduced=this._configurationService.getValue("workbench.reduceMotion"),this._onDidChangeReducedMotion.fire())})),n(),this._register(this.onDidChangeScreenReaderOptimized(()=>n()));const o=Nt.matchMedia("(prefers-reduced-motion: reduce)");this._systemMotionReduced=o.matches,this._configMotionReduced=this._configurationService.getValue("workbench.reduceMotion"),this._linkUnderlinesEnabled=this._configurationService.getValue("accessibility.underlineLinks"),this.initReducedMotionListeners(o),this.initLinkUnderlineListeners()}initReducedMotionListeners(e){this._register(W(e,"change",()=>{this._systemMotionReduced=e.matches,this._configMotionReduced==="auto"&&this._onDidChangeReducedMotion.fire()}));const t=()=>{const i=this.isMotionReduced();this._layoutService.mainContainer.classList.toggle("monaco-reduce-motion",i),this._layoutService.mainContainer.classList.toggle("monaco-enable-motion",!i)};t(),this._register(this.onDidChangeReducedMotion(()=>t()))}initLinkUnderlineListeners(){this._register(this._configurationService.onDidChangeConfiguration(t=>{if(t.affectsConfiguration("accessibility.underlineLinks")){const i=this._configurationService.getValue("accessibility.underlineLinks");this._linkUnderlinesEnabled=i,this._onDidChangeLinkUnderline.fire()}}));const e=()=>{const t=this._linkUnderlinesEnabled;this._layoutService.mainContainer.classList.toggle("underline-links",t)};e(),this._register(this.onDidChangeLinkUnderlines(()=>e()))}onDidChangeLinkUnderlines(e){return this._onDidChangeLinkUnderline.event(e)}get onDidChangeScreenReaderOptimized(){return this._onDidChangeScreenReaderOptimized.event}isScreenReaderOptimized(){const e=this._configurationService.getValue("editor.accessibilitySupport");return e==="on"||e==="auto"&&this._accessibilitySupport===2}get onDidChangeReducedMotion(){return this._onDidChangeReducedMotion.event}isMotionReduced(){const e=this._configMotionReduced;return e==="on"||e==="auto"&&this._systemMotionReduced}getAccessibilitySupport(){return this._accessibilitySupport}};V9=L6e([jO(0,Ae),jO(1,qc),jO(2,He)],V9);var k6e=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},PY=function(s,e){return function(t,i){e(t,i,s)}},z9;const FY="application/vnd.code.resources";var _m;let $9=(_m=class extends F{constructor(e,t){super(),this.layoutService=e,this.logService=t,this.mapTextToType=new Map,this.findText="",this.resources=[],this.resourcesStateHash=void 0,(wm||FX)&&this.installWebKitWriteTextWorkaround(),this._register(oe.runAndSubscribe(G0,({window:i,disposables:n})=>{n.add(W(i.document,"copy",()=>this.clearResourcesState()))},{window:Nt,disposables:this._store}))}triggerPaste(){this.logService.trace("BrowserClipboardService#triggerPaste")}installWebKitWriteTextWorkaround(){const e=()=>{const t=new n_;this.webKitPendingClipboardWritePromise&&!this.webKitPendingClipboardWritePromise.isSettled&&this.webKitPendingClipboardWritePromise.cancel(),this.webKitPendingClipboardWritePromise=t,kt().navigator.clipboard.write([new ClipboardItem({"text/plain":t.p})]).catch(async i=>{(!(i instanceof Error)||i.name!=="NotAllowedError"||!t.isRejected)&&this.logService.error(i)})};this._register(oe.runAndSubscribe(this.layoutService.onDidAddContainer,({container:t,disposables:i})=>{i.add(W(t,"click",e)),i.add(W(t,"keydown",e))},{container:this.layoutService.mainContainer,disposables:this._store}))}async writeText(e,t){if(this.logService.trace("BrowserClipboardService#writeText called with type:",t," text.length:",e.length),this.clearResourcesState(),t){this.mapTextToType.set(t,e),this.logService.trace("BrowserClipboardService#writeText");return}if(this.webKitPendingClipboardWritePromise)return this.webKitPendingClipboardWritePromise.complete(e);try{return this.logService.trace("before navigator.clipboard.writeText"),await kt().navigator.clipboard.writeText(e)}catch(i){console.error(i)}this.fallbackWriteText(e)}fallbackWriteText(e){this.logService.trace("BrowserClipboardService#fallbackWriteText");const t=K0(),i=t.activeElement,n=t.body.appendChild(ie("textarea",{"aria-hidden":!0}));n.style.height="1px",n.style.width="1px",n.style.position="absolute",n.value=e,n.focus(),n.select(),t.execCommand("copy"),ki(i)&&i.focus(),n.remove()}async readText(e){if(this.logService.trace("BrowserClipboardService#readText called with type:",e),e){const t=this.mapTextToType.get(e)||"";return this.logService.trace("BrowserClipboardService#readText text.length:",t.length),t}try{const t=await kt().navigator.clipboard.readText();return this.logService.trace("BrowserClipboardService#readText text.length:",t.length),t}catch(t){console.error(t)}return""}async readFindText(){return this.findText}async writeFindText(e){this.findText=e}async readResources(){try{const t=await kt().navigator.clipboard.read();for(const i of t)if(i.types.includes(`web ${FY}`)){const n=await i.getType(`web ${FY}`);return JSON.parse(await n.text()).map(r=>Le.from(r))}}catch{}const e=await this.computeResourcesStateHash();return this.resourcesStateHash!==e&&this.clearResourcesState(),this.resources}async computeResourcesStateHash(){if(this.resources.length===0)return;const e=await this.readText();return j0(e.substring(0,z9.MAX_RESOURCE_STATE_SOURCE_LENGTH))}clearInternalState(){this.clearResourcesState()}clearResourcesState(){this.resources=[],this.resourcesStateHash=void 0}},z9=_m,_m.MAX_RESOURCE_STATE_SOURCE_LENGTH=1e3,_m);$9=z9=k6e([PY(0,qc),PY(1,jt)],$9);var E6e=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},I6e=function(s,e){return function(t,i){e(t,i,s)}};const my="data-keybinding-context";class SV{constructor(e,t){this._id=e,this._parent=t,this._value=Object.create(null),this._value._contextId=e}get value(){return{...this._value}}setValue(e,t){return this._value[e]!==t?(this._value[e]=t,!0):!1}removeValue(e){return e in this._value?(delete this._value[e],!0):!1}getValue(e){const t=this._value[e];return typeof t>"u"&&this._parent?this._parent.getValue(e):t}}const CR=class CR extends SV{constructor(){super(-1,null)}setValue(e,t){return!1}removeValue(e){return!1}getValue(e){}};CR.INSTANCE=new CR;let Fv=CR;const Fy=class Fy extends SV{constructor(e,t,i){super(e,null),this._configurationService=t,this._values=xw.forConfigKeys(),this._listener=this._configurationService.onDidChangeConfiguration(n=>{if(n.source===7){const o=Array.from(this._values,([r])=>r);this._values.clear(),i.fire(new WY(o))}else{const o=[];for(const r of n.affectedKeys){const a=`config.${r}`,l=this._values.findSuperstr(a);l!==void 0&&(o.push(...ot.map(l,([c])=>c)),this._values.deleteSuperstr(a)),this._values.has(a)&&(o.push(a),this._values.delete(a))}i.fire(new WY(o))}})}dispose(){this._listener.dispose()}getValue(e){if(e.indexOf(Fy._keyPrefix)!==0)return super.getValue(e);if(this._values.has(e))return this._values.get(e);const t=e.substr(Fy._keyPrefix.length),i=this._configurationService.getValue(t);let n;switch(typeof i){case"number":case"boolean":case"string":n=i;break;default:Array.isArray(i)?n=JSON.stringify(i):n=i}return this._values.set(e,n),n}setValue(e,t){return super.setValue(e,t)}removeValue(e){return super.removeValue(e)}};Fy._keyPrefix="config.";let U9=Fy;class D6e{constructor(e,t,i){this._service=e,this._key=t,this._defaultValue=i,this.reset()}set(e){this._service.setContext(this._key,e)}reset(){typeof this._defaultValue>"u"?this._service.removeContext(this._key):this._service.setContext(this._key,this._defaultValue)}get(){return this._service.getContextKeyValue(this._key)}}class BY{constructor(e){this.key=e}affectsSome(e){return e.has(this.key)}allKeysContainedIn(e){return this.affectsSome(e)}}class WY{constructor(e){this.keys=e}affectsSome(e){for(const t of this.keys)if(e.has(t))return!0;return!1}allKeysContainedIn(e){return this.keys.every(t=>e.has(t))}}class T6e{constructor(e){this.events=e}affectsSome(e){for(const t of this.events)if(t.affectsSome(e))return!0;return!1}allKeysContainedIn(e){return this.events.every(t=>t.allKeysContainedIn(e))}}function N6e(s,e){return s.allKeysContainedIn(new Set(Object.keys(e)))}class zre extends F{get onDidChangeContext(){return this._onDidChangeContext.event}constructor(e){super(),this._onDidChangeContext=this._register(new Sm({merge:t=>new T6e(t)})),this._isDisposed=!1,this._myContextId=e}createKey(e,t){if(this._isDisposed)throw new Error("AbstractContextKeyService has been disposed");return new D6e(this,e,t)}bufferChangeEvents(e){this._onDidChangeContext.pause();try{e()}finally{this._onDidChangeContext.resume()}}createScoped(e){if(this._isDisposed)throw new Error("AbstractContextKeyService has been disposed");return new R6e(this,e)}contextMatchesRules(e){if(this._isDisposed)throw new Error("AbstractContextKeyService has been disposed");const t=this.getContextValuesContainer(this._myContextId);return e?e.evaluate(t):!0}getContextKeyValue(e){if(!this._isDisposed)return this.getContextValuesContainer(this._myContextId).getValue(e)}setContext(e,t){if(this._isDisposed)return;const i=this.getContextValuesContainer(this._myContextId);i&&i.setValue(e,t)&&this._onDidChangeContext.fire(new BY(e))}removeContext(e){this._isDisposed||this.getContextValuesContainer(this._myContextId).removeValue(e)&&this._onDidChangeContext.fire(new BY(e))}getContext(e){return this._isDisposed?Fv.INSTANCE:this.getContextValuesContainer(A6e(e))}dispose(){super.dispose(),this._isDisposed=!0}}let q9=class extends zre{constructor(e){super(0),this._contexts=new Map,this._lastContextId=0,this.inputFocusedContext=yH.bindTo(this);const t=this._register(new U9(this._myContextId,e,this._onDidChangeContext));this._contexts.set(this._myContextId,t),this._register(oe.runAndSubscribe(G0,({window:i,disposables:n})=>{const o=n.add(new vt);n.add(W(i,se.FOCUS_IN,()=>{o.value=new U,this.updateInputContextKeys(i.document,o.value)},!0))},{window:Nt,disposables:this._store}))}updateInputContextKeys(e,t){function i(){return!!e.activeElement&&wl(e.activeElement)}const n=i();if(this.inputFocusedContext.set(n),n){const o=t.add(oa(e.activeElement));oe.once(o.onDidBlur)(()=>{kt().document===e&&this.inputFocusedContext.set(i()),o.dispose()},void 0,t)}}getContextValuesContainer(e){return this._isDisposed?Fv.INSTANCE:this._contexts.get(e)||Fv.INSTANCE}createChildContext(e=this._myContextId){if(this._isDisposed)throw new Error("ContextKeyService has been disposed");const t=++this._lastContextId;return this._contexts.set(t,new SV(t,this.getContextValuesContainer(e))),t}disposeContext(e){this._isDisposed||this._contexts.delete(e)}};q9=E6e([I6e(0,He)],q9);class R6e extends zre{constructor(e,t){if(super(e.createChildContext()),this._parentChangeListener=this._register(new vt),this._parent=e,this._updateParentChangeListener(),this._domNode=t,this._domNode.hasAttribute(my)){let i="";this._domNode.classList&&(i=Array.from(this._domNode.classList.values()).join(", ")),console.error(`Element already has context attribute${i?": "+i:""}`)}this._domNode.setAttribute(my,String(this._myContextId))}_updateParentChangeListener(){this._parentChangeListener.value=this._parent.onDidChangeContext(e=>{const i=this._parent.getContextValuesContainer(this._myContextId).value;N6e(e,i)||this._onDidChangeContext.fire(e)})}dispose(){this._isDisposed||(this._parent.disposeContext(this._myContextId),this._domNode.removeAttribute(my),super.dispose())}getContextValuesContainer(e){return this._isDisposed?Fv.INSTANCE:this._parent.getContextValuesContainer(e)}createChildContext(e=this._myContextId){if(this._isDisposed)throw new Error("ScopedContextKeyService has been disposed");return this._parent.createChildContext(e)}disposeContext(e){this._isDisposed||this._parent.disposeContext(e)}}function A6e(s){for(;s;){if(s.hasAttribute(my)){const e=s.getAttribute(my);return e?parseInt(e,10):NaN}s=s.parentElement}return 0}function M6e(s,e,t){s.get(Ae).createKey(String(e),O6e(t))}function O6e(s){return iJ(s,e=>{if(typeof e=="object"&&e.$mid===1)return Le.revive(e).toString();if(e instanceof Le)return e.toString()})}at.registerCommand("_setContext",M6e);at.registerCommand({id:"getContextKeyInfo",handler(){return[...ce.all()].sort((s,e)=>s.key.localeCompare(e.key))},metadata:{description:m(1674,"A command that returns information about context keys"),args:[]}});at.registerCommand("_generateContextKeyInfo",function(){const s=[],e=new Set;for(const t of ce.all())e.has(t.key)||(e.add(t.key),s.push(t));s.sort((t,i)=>t.key.localeCompare(i.key)),console.log(JSON.stringify(s,void 0,2))});let P6e=class{constructor(e,t){this.key=e,this.data=t,this.incoming=new Map,this.outgoing=new Map}};class HY{constructor(e){this._hashFn=e,this._nodes=new Map}roots(){const e=[];for(const t of this._nodes.values())t.outgoing.size===0&&e.push(t);return e}insertEdge(e,t){const i=this.lookupOrInsertNode(e),n=this.lookupOrInsertNode(t);i.outgoing.set(n.key,n),n.incoming.set(i.key,i)}removeNode(e){const t=this._hashFn(e);this._nodes.delete(t);for(const i of this._nodes.values())i.outgoing.delete(t),i.incoming.delete(t)}lookupOrInsertNode(e){const t=this._hashFn(e);let i=this._nodes.get(t);return i||(i=new P6e(t,e),this._nodes.set(t,i)),i}isEmpty(){return this._nodes.size===0}toString(){const e=[];for(const[t,i]of this._nodes)e.push(`${t}
	(-> incoming)[${[...i.incoming.keys()].join(", ")}]
	(outgoing ->)[${[...i.outgoing.keys()].join(",")}]
`);return e.join(`
`)}findCycleSlow(){for(const[e,t]of this._nodes){const i=new Set([e]),n=this._findCycle(t,i);if(n)return n}}_findCycle(e,t){for(const[i,n]of e.outgoing){if(t.has(i))return[...t,i].join(" -> ");t.add(i);const o=this._findCycle(n,t);if(o)return o;t.delete(i)}}}const F6e=!1;class VY extends Error{constructor(e){super("cyclic dependency between services"),this.message=e.findCycleSlow()??`UNABLE to detect cycle, dumping graph: 
${e.toString()}`}}class xT{constructor(e=new rC,t=!1,i,n=F6e){this._services=e,this._strict=t,this._parent=i,this._enableTracing=n,this._isDisposed=!1,this._servicesToMaybeDispose=new Set,this._children=new Set,this._activeInstantiations=new Set,this._services.set(_e,this),this._globalGraph=n?i?._globalGraph??new HY(o=>o):void 0}dispose(){if(!this._isDisposed){this._isDisposed=!0,xt(this._children),this._children.clear();for(const e of this._servicesToMaybeDispose)i_(e)&&e.dispose();this._servicesToMaybeDispose.clear()}}_throwIfDisposed(){if(this._isDisposed)throw new Error("InstantiationService has been disposed")}createChild(e,t){this._throwIfDisposed();const i=this,n=new class extends xT{dispose(){i._children.delete(n),super.dispose()}}(e,this._strict,this,this._enableTracing);return this._children.add(n),t?.add(n),n}invokeFunction(e,...t){this._throwIfDisposed();const i=_y.traceInvocation(this._enableTracing,e);let n=!1;try{return e({get:r=>{if(n)throw uE("service accessor is only valid during the invocation of its target method");const a=this._getOrCreateServiceInstance(r,i);if(!a)throw new Error(`[invokeFunction] unknown service '${r}'`);return a},getIfExists:r=>{if(n)throw uE("service accessor is only valid during the invocation of its target method");return this._getOrCreateServiceInstance(r,i)}},...t)}finally{n=!0,i.stop()}}createInstance(e,...t){this._throwIfDisposed();let i,n;return e instanceof ac?(i=_y.traceCreation(this._enableTracing,e.ctor),n=this._createInstance(e.ctor,e.staticArguments.concat(t),i)):(i=_y.traceCreation(this._enableTracing,e),n=this._createInstance(e,t,i)),i.stop(),n}_createInstance(e,t=[],i){const n=pl.getServiceDependencies(e).sort((a,l)=>a.index-l.index),o=[];for(const a of n){const l=this._getOrCreateServiceInstance(a.id,i);l||this._throwIfStrict(`[createInstance] ${e.name} depends on UNKNOWN service ${a.id}.`,!1),o.push(l)}const r=n.length>0?n[0].index:t.length;if(t.length!==r){console.trace(`[createInstance] First service dependency of ${e.name} at position ${r+1} conflicts with ${t.length} static arguments`);const a=r-t.length;a>0?t=t.concat(new Array(a)):t=t.slice(0,r)}return Reflect.construct(e,t.concat(o))}_setCreatedServiceInstance(e,t){if(this._services.get(e)instanceof ac)this._services.set(e,t);else if(this._parent)this._parent._setCreatedServiceInstance(e,t);else throw new Error("illegalState - setting UNKNOWN service instance")}_getServiceInstanceOrDescriptor(e){const t=this._services.get(e);return!t&&this._parent?this._parent._getServiceInstanceOrDescriptor(e):t}_getOrCreateServiceInstance(e,t){this._globalGraph&&this._globalGraphImplicitDependency&&this._globalGraph.insertEdge(this._globalGraphImplicitDependency,String(e));const i=this._getServiceInstanceOrDescriptor(e);return i instanceof ac?this._safeCreateAndCacheServiceInstance(e,i,t.branch(e,!0)):(t.branch(e,!1),i)}_safeCreateAndCacheServiceInstance(e,t,i){if(this._activeInstantiations.has(e))throw new Error(`illegal state - RECURSIVELY instantiating service '${e}'`);this._activeInstantiations.add(e);try{return this._createAndCacheServiceInstance(e,t,i)}finally{this._activeInstantiations.delete(e)}}_createAndCacheServiceInstance(e,t,i){const n=new HY(l=>l.id.toString());let o=0;const r=[{id:e,desc:t,_trace:i}],a=new Set;for(;r.length;){const l=r.pop();if(!a.has(String(l.id))){if(a.add(String(l.id)),n.lookupOrInsertNode(l),o++>1e3)throw new VY(n);for(const c of pl.getServiceDependencies(l.desc.ctor)){const d=this._getServiceInstanceOrDescriptor(c.id);if(d||this._throwIfStrict(`[createInstance] ${e} depends on ${c.id} which is NOT registered.`,!0),this._globalGraph?.insertEdge(String(l.id),String(c.id)),d instanceof ac){const h={id:c.id,desc:d,_trace:l._trace.branch(c.id,!0)};n.insertEdge(l,h),r.push(h)}}}}for(;;){const l=n.roots();if(l.length===0){if(!n.isEmpty())throw new VY(n);break}for(const{data:c}of l){if(this._getServiceInstanceOrDescriptor(c.id)instanceof ac){const h=this._createServiceInstanceWithOwner(c.id,c.desc.ctor,c.desc.staticArguments,c.desc.supportsDelayedInstantiation,c._trace);this._setCreatedServiceInstance(c.id,h)}n.removeNode(c)}}return this._getServiceInstanceOrDescriptor(e)}_createServiceInstanceWithOwner(e,t,i=[],n,o){if(this._services.get(e)instanceof ac)return this._createServiceInstance(e,t,i,n,o,this._servicesToMaybeDispose);if(this._parent)return this._parent._createServiceInstanceWithOwner(e,t,i,n,o);throw new Error(`illegalState - creating UNKNOWN service instance ${t.name}`)}_createServiceInstance(e,t,i=[],n,o,r){if(n){const a=new xT(void 0,this._strict,this,this._enableTracing);a._globalGraphImplicitDependency=String(e);const l=new Map,c=new Ide(()=>{const d=a._createInstance(t,i,o);for(const[h,u]of l){const g=d[h];if(typeof g=="function")for(const f of u)f.disposable=g.apply(d,f.listener)}return l.clear(),r.add(d),d});return new Proxy(Object.create(null),{get(d,h){if(!c.isInitialized&&typeof h=="string"&&(h.startsWith("onDid")||h.startsWith("onWill"))){let f=l.get(h);return f||(f=new Ns,l.set(h,f)),(_,b,w)=>{if(c.isInitialized)return c.value[h](_,b,w);{const C={listener:[_,b,w],disposable:void 0},v=f.push(C);return pe(()=>{v(),C.disposable?.dispose()})}}}if(h in d)return d[h];const u=c.value;let g=u[h];return typeof g!="function"||(g=g.bind(u),d[h]=g),g},set(d,h,u){return c.value[h]=u,!0},getPrototypeOf(d){return t.prototype}})}else{const a=this._createInstance(t,i,o);return r.add(a),a}}_throwIfStrict(e,t){if(t&&console.warn(e),this._strict)throw new Error(e)}}const ar=class ar{static traceInvocation(e,t){return e?new ar(2,t.name||new Error().stack.split(`
`).slice(3,4).join(`
`)):ar._None}static traceCreation(e,t){return e?new ar(1,t.name):ar._None}constructor(e,t){this.type=e,this.name=t,this._start=Date.now(),this._dep=[]}branch(e,t){const i=new ar(3,e.toString());return this._dep.push([e,t,i]),i}stop(){const e=Date.now()-this._start;ar._totals+=e;let t=!1;function i(o,r){const a=[],l=new Array(o+1).join("	");for(const[c,d,h]of r._dep)if(d&&h){t=!0,a.push(`${l}CREATES -> ${c}`);const u=i(o+1,h);u&&a.push(u)}else a.push(`${l}uses -> ${c}`);return a.join(`
`)}const n=[`${this.type===1?"CREATE":"CALL"} ${this.name}`,`${i(1,this)}`,`DONE, took ${e.toFixed(2)}ms (grand total ${ar._totals.toFixed(2)}ms)`];(e>2||t)&&ar.all.add(n.join(`
`))}};ar.all=new Set,ar._None=new class extends ar{constructor(){super(0,null)}stop(){}branch(){return this}},ar._totals=0;let _y=ar;const B6e=new Set([Te.inMemory,Te.vscodeSourceControl,Te.walkThrough,Te.walkThroughSnippet,Te.vscodeChatCodeBlock,Te.vscodeTerminal]);class W6e{constructor(){this._byResource=new Wi,this._byOwner=new Map}set(e,t,i){let n=this._byResource.get(e);n||(n=new Map,this._byResource.set(e,n)),n.set(t,i);let o=this._byOwner.get(t);o||(o=new Wi,this._byOwner.set(t,o)),o.set(e,i)}get(e,t){return this._byResource.get(e)?.get(t)}delete(e,t){let i=!1,n=!1;const o=this._byResource.get(e);o&&(i=o.delete(t));const r=this._byOwner.get(t);if(r&&(n=r.delete(e)),i!==n)throw new Error("illegal state");return i&&n}values(e){return typeof e=="string"?this._byOwner.get(e)?.values()??ot.empty():Le.isUri(e)?this._byResource.get(e)?.values()??ot.empty():ot.map(ot.concat(...this._byOwner.values()),t=>t[1])}}class H6e{constructor(e){this.errors=0,this.infos=0,this.warnings=0,this.unknowns=0,this._data=new Wi,this._service=e,this._subscription=e.onMarkerChanged(this._update,this)}dispose(){this._subscription.dispose()}_update(e){for(const t of e){const i=this._data.get(t);i&&this._substract(i);const n=this._resourceStats(t);this._add(n),this._data.set(t,n)}}_resourceStats(e){const t={errors:0,warnings:0,infos:0,unknowns:0};if(B6e.has(e.scheme))return t;for(const{severity:i}of this._service.read({resource:e}))i===mi.Error?t.errors+=1:i===mi.Warning?t.warnings+=1:i===mi.Info?t.infos+=1:t.unknowns+=1;return t}_substract(e){this.errors-=e.errors,this.warnings-=e.warnings,this.infos-=e.infos,this.unknowns-=e.unknowns}_add(e){this.errors+=e.errors,this.warnings+=e.warnings,this.infos+=e.infos,this.unknowns+=e.unknowns}}class Ef{constructor(){this._onMarkerChanged=new qX({delay:0,merge:Ef._merge}),this.onMarkerChanged=this._onMarkerChanged.event,this._data=new W6e,this._stats=new H6e(this),this._filteredResources=new Wi}dispose(){this._stats.dispose(),this._onMarkerChanged.dispose()}remove(e,t){for(const i of t||[])this.changeOne(e,i,[])}changeOne(e,t,i){if(pQ(i))this._data.delete(t,e)&&this._onMarkerChanged.fire([t]);else{const n=[];for(const o of i){const r=Ef._toMarker(e,t,o);r&&n.push(r)}this._data.set(t,e,n),this._onMarkerChanged.fire([t])}}static _toMarker(e,t,i){let{code:n,severity:o,message:r,source:a,startLineNumber:l,startColumn:c,endLineNumber:d,endColumn:h,relatedInformation:u,tags:g,origin:f}=i;if(r)return l=l>0?l:1,c=c>0?c:1,d=d>=l?d:l,h=h>0?h:c,{resource:t,owner:e,code:n,severity:o,message:r,source:a,startLineNumber:l,startColumn:c,endLineNumber:d,endColumn:h,relatedInformation:u,tags:g,origin:f}}changeAll(e,t){const i=[],n=this._data.values(e);if(n)for(const o of n){const r=ot.first(o);r&&(i.push(r.resource),this._data.delete(r.resource,e))}if(Ms(t)){const o=new Wi;for(const{resource:r,marker:a}of t){const l=Ef._toMarker(e,r,a);if(!l)continue;const c=o.get(r);c?c.push(l):(o.set(r,[l]),i.push(r))}for(const[r,a]of o)this._data.set(r,e,a)}i.length>0&&this._onMarkerChanged.fire(i)}_createFilteredMarker(e,t){const i=t.length===1?m(1738,'Problems are paused because: "{0}"',t[0]):m(1739,'Problems are paused because: "{0}" and {1} more',t[0],t.length-1);return{owner:"markersFilter",resource:e,severity:mi.Info,message:i,startLineNumber:1,startColumn:1,endLineNumber:1,endColumn:1}}read(e=Object.create(null)){let{owner:t,resource:i,severities:n,take:o}=e;if((!o||o<0)&&(o=-1),t&&i){const r=e.ignoreResourceFilters?void 0:this._filteredResources.get(i);if(r?.length)return[this._createFilteredMarker(i,r)];const a=this._data.get(i,t);if(!a)return[];const l=[];for(const c of a){if(o>0&&l.length===o)break;const d=e.ignoreResourceFilters?void 0:this._filteredResources.get(i);d?.length?l.push(this._createFilteredMarker(i,d)):Ef._accept(c,n)&&l.push(c)}return l}else{const r=!t&&!i?this._data.values():this._data.values(i??t),a=[],l=new Uue;for(const c of r)for(const d of c){if(l.has(d.resource))continue;if(o>0&&a.length===o)break;const h=e.ignoreResourceFilters?void 0:this._filteredResources.get(d.resource);h?.length?(a.push(this._createFilteredMarker(d.resource,h)),l.add(d.resource)):Ef._accept(d,n)&&a.push(d)}return a}}static _accept(e,t){return t===void 0||(t&e.severity)===e.severity}static _merge(e){const t=new Wi;for(const i of e)for(const n of i)t.set(n,!0);return Array.from(t.keys())}}class V6e extends F{get configurationModel(){return this._configurationModel}constructor(e){super(),this.logService=e,this._configurationModel=Mn.createEmptyModel(e)}reload(){return this.resetConfigurationModel(),this.configurationModel}getConfigurationDefaultOverrides(){return{}}resetConfigurationModel(){this._configurationModel=Mn.createEmptyModel(this.logService);const e=bi.as(Ml.Configuration).getConfigurationProperties();this.updateConfigurationModel(Object.keys(e),e)}updateConfigurationModel(e,t){const i=this.getConfigurationDefaultOverrides();for(const n of e){const o=i[n],r=t[n];o!==void 0?this._configurationModel.setValue(n,o):r?this._configurationModel.setValue(n,rc(r.default)):this._configurationModel.removeValue(n)}}}class z6e extends F{constructor(e,t=[]){super(),this.logger=new ige([e,...t]),this._register(e.onDidChangeLogLevel(i=>this.setLevel(i)))}get onDidChangeLogLevel(){return this.logger.onDidChangeLogLevel}setLevel(e){this.logger.setLevel(e)}getLevel(){return this.logger.getLevel()}trace(e,...t){this.logger.trace(e,...t)}debug(e,...t){this.logger.debug(e,...t)}info(e,...t){this.logger.info(e,...t)}warn(e,...t){this.logger.warn(e,...t)}error(e,...t){this.logger.error(e,...t)}}class $6e{getParserClass(){throw new Error("not implemented in StandaloneTreeSitterLibraryService")}supportsLanguage(e,t){return!1}getLanguage(e,t,i){}getInjectionQueries(e,t){return null}getHighlightingQueries(e,t){return null}}var Kc=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},qi=function(s,e){return function(t,i){e(t,i,s)}};class U6e{constructor(e){this.disposed=!1,this.model=e,this._onWillDispose=new M}get textEditorModel(){return this.model}dispose(){this.disposed=!0,this._onWillDispose.fire()}}let j9=class{constructor(e){this.modelService=e}createModelReference(e){const t=this.modelService.getModel(e);return t?Promise.resolve(new fde(new U6e(t))):Promise.reject(new Error("Model not found"))}};j9=Kc([qi(0,hi)],j9);const yR=class yR{show(){return yR.NULL_PROGRESS_RUNNER}async showWhile(e,t){await e}};yR.NULL_PROGRESS_RUNNER={done:()=>{},total:()=>{},worked:()=>{}};let G9=yR;class q6e{withProgress(e,t,i){return t({report:()=>{}})}}class j6e{constructor(){this.isExtensionDevelopment=!1,this.isBuilt=!1}}class G6e{async confirm(e){return{confirmed:this.doConfirm(e.message,e.detail),checkboxChecked:!1}}doConfirm(e,t){let i=e;return t&&(i=i+`

`+t),Nt.confirm(i)}async prompt(e){let t;if(this.doConfirm(e.message,e.detail)){const n=[...e.buttons??[]];e.cancelButton&&typeof e.cancelButton!="string"&&typeof e.cancelButton!="boolean"&&n.push(e.cancelButton),t=await n[0]?.run({checkboxChecked:!1})}return{result:t}}async error(e,t){await this.prompt({type:fi.Error,message:e,detail:t})}}const By=class By{info(e){return this.notify({severity:fi.Info,message:e})}warn(e){return this.notify({severity:fi.Warning,message:e})}error(e){return this.notify({severity:fi.Error,message:e})}notify(e){switch(e.severity){case fi.Error:console.error(e.message);break;case fi.Warning:console.warn(e.message);break;default:console.log(e.message);break}return By.NO_OP}prompt(e,t,i,n){return By.NO_OP}status(e,t){return{close:()=>{}}}};By.NO_OP=new Pve;let K9=By,Y9=class{constructor(e){this._onWillExecuteCommand=new M,this._onDidExecuteCommand=new M,this.onDidExecuteCommand=this._onDidExecuteCommand.event,this._instantiationService=e}executeCommand(e,...t){const i=at.getCommand(e);if(!i)return Promise.reject(new Error(`command '${e}' not found`));try{this._onWillExecuteCommand.fire({commandId:e,args:t});const n=this._instantiationService.invokeFunction.apply(this._instantiationService,[i.handler,...t]);return this._onDidExecuteCommand.fire({commandId:e,args:t}),Promise.resolve(n)}catch(n){return Promise.reject(n)}}};Y9=Kc([qi(0,_e)],Y9);let Bv=class extends x9e{constructor(e,t,i,n,o,r){super(e,t,i,n,o),this._cachedResolver=null,this._dynamicKeybindings=[],this._domNodeListeners=[];const a=g=>{const f=new U;f.add(W(g,se.KEY_DOWN,p=>{const _=new Pt(p);this._dispatch(_,_.target)&&(_.preventDefault(),_.stopPropagation())})),f.add(W(g,se.KEY_UP,p=>{const _=new Pt(p);this._singleModifierDispatch(_,_.target)&&_.preventDefault()})),this._domNodeListeners.push(new K6e(g,f))},l=g=>{for(let f=0;f<this._domNodeListeners.length;f++){const p=this._domNodeListeners[f];p.domNode===g&&(this._domNodeListeners.splice(f,1),p.dispose())}},c=g=>{g.getOption(70)||a(g.getContainerDomNode())},d=g=>{g.getOption(70)||l(g.getContainerDomNode())};this._register(r.onCodeEditorAdd(c)),this._register(r.onCodeEditorRemove(d)),r.listCodeEditors().forEach(c);const h=g=>{a(g.getContainerDomNode())},u=g=>{l(g.getContainerDomNode())};this._register(r.onDiffEditorAdd(h)),this._register(r.onDiffEditorRemove(u)),r.listDiffEditors().forEach(h)}addDynamicKeybinding(e,t,i,n){return Ra(at.registerCommand(e,i),this.addDynamicKeybindings([{keybinding:t,command:e,when:n}]))}addDynamicKeybindings(e){const t=e.map(i=>({keybinding:oP(i.keybinding,Fo),command:i.command??null,commandArgs:i.commandArgs,when:i.when,weight1:1e3,weight2:0,extensionId:null,isBuiltinExtension:!1}));return this._dynamicKeybindings=this._dynamicKeybindings.concat(t),this.updateResolver(),pe(()=>{for(let i=0;i<this._dynamicKeybindings.length;i++)if(this._dynamicKeybindings[i]===t[0]){this._dynamicKeybindings.splice(i,t.length),this.updateResolver();return}})}updateResolver(){this._cachedResolver=null,this._onDidUpdateKeybindings.fire()}_getResolver(){if(!this._cachedResolver){const e=this._toNormalizedKeybindingItems(Dn.getDefaultKeybindings(),!0),t=this._toNormalizedKeybindingItems(this._dynamicKeybindings,!1);this._cachedResolver=new fy(e,t,i=>this._log(i))}return this._cachedResolver}_documentHasFocus(){return Nt.document.hasFocus()}_toNormalizedKeybindingItems(e,t){const i=[];let n=0;for(const o of e){const r=o.when||void 0,a=o.keybinding;if(!a)i[n++]=new wY(void 0,o.command,o.commandArgs,r,t,null,!1);else{const l=L0.resolveKeybinding(a,Fo);for(const c of l)i[n++]=new wY(c,o.command,o.commandArgs,r,t,null,!1)}}return i}resolveKeyboardEvent(e){const t=new dh(e.ctrlKey,e.shiftKey,e.altKey,e.metaKey,e.keyCode);return new L0([t],Fo)}};Bv=Kc([qi(0,Ae),qi(1,Gt),qi(2,vs),qi(3,Ii),qi(4,jt),qi(5,ht)],Bv);class K6e extends F{constructor(e,t){super(),this.domNode=e,this._register(t)}}function zY(s){return!!s&&typeof s=="object"&&(!s.overrideIdentifier||typeof s.overrideIdentifier=="string")&&(!s.resource||s.resource instanceof Le)}let LT=class{constructor(e){this.logService=e,this._onDidChangeConfiguration=new M,this.onDidChangeConfiguration=this._onDidChangeConfiguration.event;const t=new V6e(e);this._configuration=new s2(t.reload(),Mn.createEmptyModel(e),Mn.createEmptyModel(e),Mn.createEmptyModel(e),Mn.createEmptyModel(e),Mn.createEmptyModel(e),new Wi,Mn.createEmptyModel(e),new Wi,e),t.dispose()}getValue(e,t){const i=typeof e=="string"?e:void 0,n=zY(e)?e:zY(t)?t:{};return this._configuration.getValue(i,n,void 0)}updateValues(e){const t={data:this._configuration.toData()},i=[];for(const n of e){const[o,r]=n;this.getValue(o)!==r&&(this._configuration.updateValue(o,r),i.push(o))}if(i.length>0){const n=new v9e({keys:i,overrides:[]},t,this._configuration,void 0,this.logService);n.source=8,this._onDidChangeConfiguration.fire(n)}return Promise.resolve()}updateValue(e,t,i,n){return this.updateValues([[e,t]])}inspect(e,t={}){return this._configuration.inspect(e,t,void 0)}};LT=Kc([qi(0,jt)],LT);let Z9=class{constructor(e,t,i){this.configurationService=e,this.modelService=t,this.languageService=i,this._onDidChangeConfiguration=new M,this.configurationService.onDidChangeConfiguration(n=>{this._onDidChangeConfiguration.fire({affectedKeys:n.affectedKeys,affectsConfiguration:(o,r)=>n.affectsConfiguration(r)})})}getValue(e,t,i){const n=A.isIPosition(t)?t:null,o=n?typeof i=="string"?i:void 0:typeof t=="string"?t:void 0,r=e?this.getLanguage(e,n):void 0;return typeof o>"u"?this.configurationService.getValue({resource:e,overrideIdentifier:r}):this.configurationService.getValue(o,{resource:e,overrideIdentifier:r})}getLanguage(e,t){const i=this.modelService.getModel(e);return i?t?i.getLanguageIdAtPosition(t.lineNumber,t.column):i.getLanguageId():this.languageService.guessLanguageIdByFilepathOrFirstLine(e)}};Z9=Kc([qi(0,He),qi(1,hi),qi(2,Ei)],Z9);let X9=class{constructor(e){this.configurationService=e}getEOL(e,t){const i=this.configurationService.getValue("files.eol",{overrideIdentifier:t,resource:e});return i&&typeof i=="string"&&i!=="auto"?i:vo||Ye?`
`:`\r
`}};X9=Kc([qi(0,He)],X9);class Y6e{publicLog2(){}}const Wy=class Wy{constructor(){const e=Le.from({scheme:Wy.SCHEME,authority:"model",path:"/"});this.workspace={id:qie,folders:[new cIe({uri:e,name:"",index:0})]}}getWorkspace(){return this.workspace}getWorkspaceFolder(e){return e&&e.scheme===Wy.SCHEME?this.workspace.folders[0]:null}};Wy.SCHEME="inmemory";let Q9=Wy;function kT(s,e,t){if(!e||!(s instanceof LT))return;const i=[];Object.keys(e).forEach(n=>{EIe(n)&&i.push([`editor.${n}`,e[n]]),t&&IIe(n)&&i.push([`diffEditor.${n}`,e[n]])}),i.length>0&&s.updateValues(i)}let J9=class{constructor(e){this._modelService=e}hasPreviewHandler(){return!1}async apply(e,t){const i=Array.isArray(e)?e:G7.convert(e),n=new Map;for(const a of i){if(!(a instanceof Bu))throw new Error("bad edit - only text edits are supported");const l=this._modelService.getModel(a.resource);if(!l)throw new Error("bad edit - model not found");if(typeof a.versionId=="number"&&l.getVersionId()!==a.versionId)throw new Error("bad state - model changed in the meantime");let c=n.get(l);c||(c=[],n.set(l,c)),c.push(Si.replaceMove(k.lift(a.textEdit.range),a.textEdit.text))}let o=0,r=0;for(const[a,l]of n)a.pushStackElement(),a.pushEditOperations([],l,()=>[]),a.pushStackElement(),r+=1,o+=l.length;return{ariaSummary:xm(q8.bulkEditServiceSummary,o,r),isApplied:o>0}}};J9=Kc([qi(0,hi)],J9);class Z6e{getUriLabel(e,t){return e.scheme==="file"?e.fsPath:e.path}getUriBasenameLabel(e){return ra(e)}}let e6=class extends u9e{constructor(e,t){super(e),this._codeEditorService=t}showContextView(e,t,i){if(!t){const n=this._codeEditorService.getFocusedCodeEditor()||this._codeEditorService.getActiveCodeEditor();n&&(t=n.getContainerDomNode())}return super.showContextView(e,t,i)}};e6=Kc([qi(0,qc),qi(1,ht)],e6);class X6e{constructor(){this._neverEmitter=new M,this.onDidChangeTrust=this._neverEmitter.event}isWorkspaceTrusted(){return!0}}class Q6e extends p9{constructor(){super()}}class J6e extends z6e{constructor(){super(new tge)}}let t6=class extends b9{constructor(e,t,i,n,o,r){super(e,t,i,n,o,r),this.configure({blockMouse:!1})}};t6=Kc([qi(0,vs),qi(1,Ii),qi(2,Rh),qi(3,ft),qi(4,da),qi(5,Ae)],t6);const eWe={esmModuleLocation:void 0,label:"editorWorkerService"};let i6=class extends n9{constructor(e,t,i,n,o){super(eWe,e,t,i,n,o)}};i6=Kc([qi(0,hi),qi(1,QA),qi(2,jt),qi(3,ui),qi(4,ge)],i6);class tWe{async playSignal(e,t){}}tt(jt,J6e,0);tt(He,LT,0);tt(QA,Z9,0);tt(Woe,X9,0);tt(Sh,Q9,0);tt(Ym,Z6e,0);tt(vs,Y6e,0);tt(kx,G6e,0);tt(nH,j6e,0);tt(Ii,K9,0);tt(jc,Ef,0);tt(Ei,Q6e,0);tt(Rr,nBe,0);tt(hi,x9,0);tt(EW,S9,0);tt(Ae,q9,0);tt(bte,q6e,0);tt(Ch,G9,0);tt(Bs,jke,0);tt(ko,i6,0);tt(fx,J9,0);tt(Xoe,X6e,0);tt(Fs,j9,0);tt(Wn,V9,0);tt(fa,eRe,0);tt(Gt,Y9,0);tt(ft,Bv,0);tt(bs,W9,0);tt(Rh,e6,0);tt(Ah,y9,0);tt(Ko,$9,0);tt(Nr,t6,0);tt(da,L4,0);tt(Th,tWe,0);tt(u7,$6e,0);tt(jQ,oge,0);tt(jH,mMe,0);var Pe;(function(s){const e=new rC;for(const[l,c]of g$())e.set(l,c);const t=new xT(e,!0);e.set(_e,t);function i(l){n||r({});const c=e.get(l);if(!c)throw new Error("Missing service "+l);return c instanceof ac?t.invokeFunction(d=>d.get(l)):c}s.get=i;let n=!1;const o=new M;function r(l){if(n)return t;n=!0;for(const[d,h]of g$())e.get(d)||e.set(d,h);for(const d in l)if(l.hasOwnProperty(d)){const h=qe(d);e.get(h)instanceof ac&&e.set(h,l[d])}const c=cDe();for(const d of c)try{t.createInstance(d)}catch(h){Oe(h)}return o.fire(),t}s.initialize=r;function a(l){if(n)return l();const c=new U,d=c.add(o.event(()=>{d.dispose(),c.add(l())}));return c}s.withServices=a})(Pe||(Pe={}));var iWe=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},$Y=function(s,e){return function(t,i){e(t,i,s)}},sE,bm;let n6=(bm=class{constructor(e,t){this._configurationService=e,this._languageService=t}async renderCodeBlock(e,t,i){const n=Tl(i.context)?i.context:void 0;let o;e?o=this._languageService.getLanguageIdByLanguageName(e):n&&(o=n.getModel()?.getLanguageId()),o||(o=xr);const r=await lSe(this._languageService,t,o),a=sE._ttpTokenizer?sE._ttpTokenizer.createHTML(r)??r:r,l=document.createElement("span");l.innerHTML=a;const c=l.querySelector(".monaco-tokenized-source");return ki(c)?(In(c,this.getFontInfo(n)),l):document.createElement("span")}getFontInfo(e){return e?e.getOption(59):Bpe({fontFamily:this._configurationService.getValue("editor").fontFamily},1)}},sE=bm,bm._ttpTokenizer=$c("tokenizeToString",{createHTML(e){return e}}),bm);n6=sE=iWe([$Y(0,He),$Y(1,Ei)],n6);var xV=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},Ht=function(s,e){return function(t,i){e(t,i,s)}};let nWe=0,UY=!1;function sWe(s){if(!s){if(UY)return;UY=!0}vue(s||Nt.document.body)}let ET=class extends Am{constructor(e,t,i,n,o,r,a,l,c,d,h,u,g,f){const p={...t};p.ariaLabel=p.ariaLabel||$8.editorViewAccessibleLabel,super(e,p,{},i,n,o,r,c,d,h,u,g),l instanceof Bv?this._standaloneKeybindingService=l:this._standaloneKeybindingService=null,sWe(p.ariaContainerElement),KSe((_,b)=>i.createInstance(Cv,_,{instantHover:b},{})),sLe(a),f.setDefaultCodeBlockRenderer(i.createInstance(n6))}addCommand(e,t,i){if(!this._standaloneKeybindingService)return console.warn("Cannot add command because the editor is configured with an unrecognized KeybindingService"),null;const n="DYNAMIC_"+ ++nWe,o=K.deserialize(i);return this._standaloneKeybindingService.addDynamicKeybinding(n,e,t,o),n}createContextKey(e,t){return this._contextKeyService.createKey(e,t)}addAction(e){if(typeof e.id!="string"||typeof e.label!="string"||typeof e.run!="function")throw new Error("Invalid action descriptor, `id`, `label` and `run` are required properties!");if(!this._standaloneKeybindingService)return console.warn("Cannot add keybinding because the editor is configured with an unrecognized KeybindingService"),F.None;const t=e.id,i=e.label,n=K.and(K.equals("editorId",this.getId()),K.deserialize(e.precondition)),o=e.keybindings,r=K.and(n,K.deserialize(e.keybindingContext)),a=e.contextMenuGroupId||null,l=e.contextMenuOrder||0,c=(g,...f)=>Promise.resolve(e.run(this,...f)),d=new U,h=this.getId()+":"+t;if(d.add(at.registerCommand(h,c)),a){const g={command:{id:h,title:i},when:n,group:a,order:l};d.add(En.appendMenuItem(fe.EditorContext,g))}if(Array.isArray(o))for(const g of o)d.add(this._standaloneKeybindingService.addDynamicKeybinding(h,g,c,r));const u=new zee(h,i,i,void 0,n,(...g)=>Promise.resolve(e.run(this,...g)),this._contextKeyService);return this._actions.set(t,u),d.add(pe(()=>{this._actions.delete(t)})),d}_triggerCommand(e,t){if(this._codeEditorService instanceof fT)try{this._codeEditorService.setActiveCodeEditor(this),super._triggerCommand(e,t)}finally{this._codeEditorService.setActiveCodeEditor(null)}else super._triggerCommand(e,t)}};ET=xV([Ht(2,_e),Ht(3,ht),Ht(4,Gt),Ht(5,Ae),Ht(6,so),Ht(7,ft),Ht(8,wi),Ht(9,Ii),Ht(10,Wn),Ht(11,ui),Ht(12,ge),Ht(13,Ua)],ET);let s6=class extends ET{constructor(e,t,i,n,o,r,a,l,c,d,h,u,g,f,p,_,b){const w={...t};kT(h,w,!1);const C=c.registerEditorContainer(e);typeof w.theme=="string"&&c.setTheme(w.theme),typeof w.autoDetectHighContrast<"u"&&c.setAutoDetectHighContrast(!!w.autoDetectHighContrast);const v=w.model;delete w.model,super(e,w,i,n,o,r,a,l,c,d,u,p,_,b),this._configurationService=h,this._standaloneThemeService=c,this._register(C);let y;if(typeof v>"u"){const x=f.getLanguageIdByMimeType(w.language)||w.language||xr;y=$re(g,f,w.value||"",x,void 0),this._ownsModel=!0}else y=v,this._ownsModel=!1;if(this._attachModel(y),y){const x={oldModelUrl:null,newModelUrl:y.uri};this._onDidChangeModel.fire(x)}}dispose(){super.dispose()}updateOptions(e){kT(this._configurationService,e,!1),typeof e.theme=="string"&&this._standaloneThemeService.setTheme(e.theme),typeof e.autoDetectHighContrast<"u"&&this._standaloneThemeService.setAutoDetectHighContrast(!!e.autoDetectHighContrast),super.updateOptions(e)}_postDetachModelCleanup(e){super._postDetachModelCleanup(e),e&&this._ownsModel&&(e.dispose(),this._ownsModel=!1)}};s6=xV([Ht(2,_e),Ht(3,ht),Ht(4,Gt),Ht(5,Ae),Ht(6,so),Ht(7,ft),Ht(8,Rr),Ht(9,Ii),Ht(10,He),Ht(11,Wn),Ht(12,hi),Ht(13,Ei),Ht(14,ui),Ht(15,ge),Ht(16,Ua)],s6);let o6=class extends Oc{constructor(e,t,i,n,o,r,a,l,c,d,h,u){const g={...t};kT(l,g,!0);const f=r.registerEditorContainer(e);typeof g.theme=="string"&&r.setTheme(g.theme),typeof g.autoDetectHighContrast<"u"&&r.setAutoDetectHighContrast(!!g.autoDetectHighContrast),super(e,g,{},n,i,o,u,d),this._configurationService=l,this._standaloneThemeService=r,this._register(f)}dispose(){super.dispose()}updateOptions(e){kT(this._configurationService,e,!0),typeof e.theme=="string"&&this._standaloneThemeService.setTheme(e.theme),typeof e.autoDetectHighContrast<"u"&&this._standaloneThemeService.setAutoDetectHighContrast(!!e.autoDetectHighContrast),super.updateOptions(e)}_createInnerEditor(e,t,i){return e.createInstance(ET,t,i)}getOriginalEditor(){return super.getOriginalEditor()}getModifiedEditor(){return super.getModifiedEditor()}addCommand(e,t,i){return this.getModifiedEditor().addCommand(e,t,i)}createContextKey(e,t){return this.getModifiedEditor().createContextKey(e,t)}addAction(e){return this.getModifiedEditor().addAction(e)}};o6=xV([Ht(2,_e),Ht(3,Ae),Ht(4,ht),Ht(5,Rr),Ht(6,Ii),Ht(7,He),Ht(8,Nr),Ht(9,Ch),Ht(10,Ko),Ht(11,Th)],o6);function $re(s,e,t,i,n){if(t=t||"",!i){const o=t.indexOf(`
`);let r=t;return o!==-1&&(r=t.substring(0,o)),qY(s,t,e.createByFilepathOrFirstLine(n||null,r),n)}return qY(s,t,e.createById(i),n)}function qY(s,e,t,i){return s.createModel(e,t,i)}T("multiDiffEditor.headerBackground",{dark:"#262626",light:"tab.inactiveBackground",hcDark:"tab.inactiveBackground",hcLight:"tab.inactiveBackground"},m(142,"The background color of the diff editor's header"));T("multiDiffEditor.background",Bi,m(143,"The background color of the multi file diff editor"));T("multiDiffEditor.border",{dark:"sideBarSectionHeader.border",light:"#cccccc",hcDark:"sideBarSectionHeader.border",hcLight:"#cccccc"},m(144,"The border color of the multi file diff editor"));var oWe=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},jY=function(s,e){return function(t,i){e(t,i,s)}};class rWe{constructor(e,t){this.viewModel=e,this.deltaScrollVertical=t}getId(){return this.viewModel}}let r6=class extends F{constructor(e,t,i,n,o){super(),this._container=e,this._overflowWidgetsDomNode=t,this._workbenchUIElementFactory=i,this._instantiationService=n,this._viewModel=Re(this,void 0),this._collapsed=j(this,l=>this._viewModel.read(l)?.collapsed.read(l)),this._editorContentHeight=Re(this,500),this.contentHeight=j(this,l=>(this._collapsed.read(l)?0:this._editorContentHeight.read(l))+this._outerEditorHeight),this._modifiedContentWidth=Re(this,0),this._modifiedWidth=Re(this,0),this._originalContentWidth=Re(this,0),this._originalWidth=Re(this,0),this.maxScroll=j(this,l=>{const c=this._modifiedContentWidth.read(l)-this._modifiedWidth.read(l),d=this._originalContentWidth.read(l)-this._originalWidth.read(l);return c>d?{maxScroll:c,width:this._modifiedWidth.read(l)}:{maxScroll:d,width:this._originalWidth.read(l)}}),this._elements=ct("div.multiDiffEntry",[ct("div.header@header",[ct("div.header-content",[ct("div.collapse-button@collapseButton"),ct("div.file-path",[ct("div.title.modified.show-file-icons@primaryPath",[]),ct("div.status.deleted@status",["R"]),ct("div.title.original.show-file-icons@secondaryPath",[])]),ct("div.actions@actions")])]),ct("div.editorParent",[ct("div.editorContainer@editor")])]),this.editor=this._register(this._instantiationService.createInstance(Oc,this._elements.editor,{overflowWidgetsDomNode:this._overflowWidgetsDomNode,fixedOverflowWidgets:!0},{})),this.isModifedFocused=di(this.editor.getModifiedEditor()).isFocused,this.isOriginalFocused=di(this.editor.getOriginalEditor()).isFocused,this.isFocused=j(this,l=>this.isModifedFocused.read(l)||this.isOriginalFocused.read(l)),this._resourceLabel=this._workbenchUIElementFactory.createResourceLabel?this._register(this._workbenchUIElementFactory.createResourceLabel(this._elements.primaryPath)):void 0,this._resourceLabel2=this._workbenchUIElementFactory.createResourceLabel?this._register(this._workbenchUIElementFactory.createResourceLabel(this._elements.secondaryPath)):void 0,this._dataStore=this._register(new U),this._headerHeight=40,this._lastScrollTop=-1,this._isSettingScrollTop=!1;const r=new VI(this._elements.collapseButton,{});this._register(De(l=>{r.element.className="",r.icon=this._collapsed.read(l)?Z.chevronRight:Z.chevronDown})),this._register(r.onDidClick(()=>{this._viewModel.get()?.collapsed.set(!this._collapsed.get(),void 0)})),this._register(De(l=>{this._elements.editor.style.display=this._collapsed.read(l)?"none":"block"})),this._register(this.editor.getModifiedEditor().onDidLayoutChange(l=>{const c=this.editor.getModifiedEditor().getLayoutInfo().contentWidth;this._modifiedWidth.set(c,void 0)})),this._register(this.editor.getOriginalEditor().onDidLayoutChange(l=>{const c=this.editor.getOriginalEditor().getLayoutInfo().contentWidth;this._originalWidth.set(c,void 0)})),this._register(this.editor.onDidContentSizeChange(l=>{o1(c=>{this._editorContentHeight.set(l.contentHeight,c),this._modifiedContentWidth.set(this.editor.getModifiedEditor().getContentWidth(),c),this._originalContentWidth.set(this.editor.getOriginalEditor().getContentWidth(),c)})})),this._register(this.editor.getOriginalEditor().onDidScrollChange(l=>{if(this._isSettingScrollTop||!l.scrollTopChanged||!this._data)return;const c=l.scrollTop-this._lastScrollTop;this._data.deltaScrollVertical(c)})),this._register(De(l=>{const c=this._viewModel.read(l)?.isActive.read(l);this._elements.root.classList.toggle("active",c)})),this._container.appendChild(this._elements.root),this._outerEditorHeight=this._headerHeight,this._contextKeyService=this._register(o.createScoped(this._elements.actions));const a=this._register(this._instantiationService.createChild(new rC([Ae,this._contextKeyService])));this._register(a.createInstance(BS,this._elements.actions,fe.MultiDiffEditorFileToolbar,{actionRunner:this._register(new Nie(()=>this._viewModel.get()?.modifiedUri??this._viewModel.get()?.originalUri)),menuOptions:{shouldForwardArgs:!0},toolbarOptions:{primaryGroup:l=>l.startsWith("navigation")},actionViewItemProvider:(l,c)=>q7(a,l,c)}))}setScrollLeft(e){this._modifiedContentWidth.get()-this._modifiedWidth.get()>this._originalContentWidth.get()-this._originalWidth.get()?this.editor.getModifiedEditor().setScrollLeft(e):this.editor.getOriginalEditor().setScrollLeft(e)}setData(e){this._data=e;function t(n){return{...n,scrollBeyondLastLine:!1,hideUnchangedRegions:{enabled:!0},scrollbar:{vertical:"hidden",horizontal:"hidden",handleMouseWheel:!1,useShadows:!1},renderOverviewRuler:!1,fixedOverflowWidgets:!0,overviewRulerBorder:!1}}if(!e){o1(n=>{this._viewModel.set(void 0,n),this.editor.setDiffModel(null,n),this._dataStore.clear()});return}const i=e.viewModel.documentDiffItem;if(o1(n=>{this._resourceLabel?.setUri(e.viewModel.modifiedUri??e.viewModel.originalUri,{strikethrough:e.viewModel.modifiedUri===void 0});let o=!1,r=!1,a=!1,l="";e.viewModel.modifiedUri&&e.viewModel.originalUri&&e.viewModel.modifiedUri.path!==e.viewModel.originalUri.path?(l="R",o=!0):e.viewModel.modifiedUri?e.viewModel.originalUri||(l="A",a=!0):(l="D",r=!0),this._elements.status.classList.toggle("renamed",o),this._elements.status.classList.toggle("deleted",r),this._elements.status.classList.toggle("added",a),this._elements.status.innerText=l,this._resourceLabel2?.setUri(o?e.viewModel.originalUri:void 0,{strikethrough:!0}),this._dataStore.clear(),this._viewModel.set(e.viewModel,n),this.editor.setDiffModel(e.viewModel.diffEditorViewModelRef,n),this.editor.updateOptions(t(i.options??{}))}),i.onOptionsDidChange&&this._dataStore.add(i.onOptionsDidChange(()=>{this.editor.updateOptions(t(i.options??{}))})),e.viewModel.isAlive.recomputeInitiallyAndOnChange(this._dataStore,n=>{n||this.setData(void 0)}),e.viewModel.documentDiffItem.contextKeys)for(const[n,o]of Object.entries(e.viewModel.documentDiffItem.contextKeys))this._contextKeyService.createKey(n,o)}render(e,t,i,n){this._elements.root.style.visibility="visible",this._elements.root.style.top=`${e.start}px`,this._elements.root.style.height=`${e.length}px`,this._elements.root.style.width=`${t}px`,this._elements.root.style.position="absolute";const o=e.length-this._headerHeight,r=Math.max(0,Math.min(n.start-e.start,o));this._elements.header.style.transform=`translateY(${r}px)`,o1(a=>{this.editor.layout({width:t-16-2,height:e.length-this._outerEditorHeight})});try{this._isSettingScrollTop=!0,this._lastScrollTop=i,this.editor.getOriginalEditor().setScrollTop(i)}finally{this._isSettingScrollTop=!1}this._elements.header.classList.toggle("shadow",r>0||i>0),this._elements.header.classList.toggle("collapsed",r===o)}hide(){this._elements.root.style.top="-100000px",this._elements.root.style.visibility="hidden"}};r6=oWe([jY(3,_e),jY(4,Ae)],r6);class aWe{constructor(e){this._create=e,this._unused=new Set,this._used=new Set,this._itemData=new Map}getUnusedObj(e){let t;if(this._unused.size===0)t=this._create(e),this._itemData.set(t,e);else{const i=[...this._unused.values()];t=i.find(n=>this._itemData.get(n).getId()===e.getId())??i[0],this._unused.delete(t),this._itemData.set(t,e),t.setData(e)}return this._used.add(t),{object:t,dispose:()=>{this._used.delete(t),this._unused.size>5?t.dispose():this._unused.add(t)}}}dispose(){for(const e of this._used)e.dispose();for(const e of this._unused)e.dispose();this._used.clear(),this._unused.clear()}}var lWe=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},GY=function(s,e){return function(t,i){e(t,i,s)}};let a6=class extends F{constructor(e,t,i,n,o,r){super(),this._element=e,this._dimension=t,this._viewModel=i,this._workbenchUIElementFactory=n,this._parentContextKeyService=o,this._parentInstantiationService=r,this._scrollableElements=ct("div.scrollContent",[ct("div@content",{style:{overflow:"hidden"}}),ct("div.monaco-editor@overflowWidgetsDomNode",{})]),this._scrollable=this._register(new sC({forceIntegerValues:!1,scheduleAtNextAnimationFrame:c=>Co(be(this._element),c),smoothScrollDuration:100})),this._scrollableElement=this._register(new cA(this._scrollableElements.root,{vertical:1,horizontal:1,useShadows:!1},this._scrollable)),this._elements=ct("div.monaco-component.multiDiffEditor",{},[ct("div",{},[this._scrollableElement.getDomNode()]),ct("div.placeholder@placeholder",{},[ct("div")])]),this._sizeObserver=this._register(new hie(this._element,void 0)),this._objectPool=this._register(new aWe(c=>{const d=this._instantiationService.createInstance(r6,this._scrollableElements.content,this._scrollableElements.overflowWidgetsDomNode,this._workbenchUIElementFactory);return d.setData(c),d})),this.scrollTop=bt(this,this._scrollableElement.onScroll,()=>this._scrollableElement.getScrollPosition().scrollTop),this.scrollLeft=bt(this,this._scrollableElement.onScroll,()=>this._scrollableElement.getScrollPosition().scrollLeft),this._viewItemsInfo=j(this,c=>{const d=this._viewModel.read(c);if(!d)return{items:[],getItem:f=>{throw new Ce}};const h=d.items.read(c),u=new Map;return{items:h.map(f=>{const p=c.store.add(new dWe(f,this._objectPool,this.scrollLeft,b=>{this._scrollableElement.setScrollPosition({scrollTop:this._scrollableElement.getScrollPosition().scrollTop+b})})),_=this._lastDocStates?.[p.getKey()];return _&&Vt(b=>{p.setViewState(_,b)}),u.set(f,p),p}),getItem:f=>u.get(f)}}),this._viewItems=this._viewItemsInfo.map(this,c=>c.items),this._spaceBetweenPx=0,this._totalHeight=this._viewItems.map(this,(c,d)=>c.reduce((h,u)=>h+u.contentHeight.read(d)+this._spaceBetweenPx,0)),this.activeControl=j(this,c=>{const d=this._viewModel.read(c)?.activeDiffItem.read(c);return d?this._viewItemsInfo.read(c).getItem(d).template.read(c)?.editor:void 0}),this._contextKeyService=this._register(this._parentContextKeyService.createScoped(this._element)),this._instantiationService=this._register(this._parentInstantiationService.createChild(new rC([Ae,this._contextKeyService]))),this._contextKeyService.createKey(N.inMultiDiffEditor.key,!0),this._lastDocStates={},this._register(ps((c,d)=>{const h=this._viewModel.read(c);if(h&&h.contextKeys)for(const[u,g]of Object.entries(h.contextKeys)){const f=this._contextKeyService.createKey(u,void 0);f.set(g),d.add(pe(()=>f.reset()))}}));const a=this._parentContextKeyService.createKey(N.multiDiffEditorAllCollapsed.key,!1);this._register(De(c=>{const d=this._viewModel.read(c);if(d){const h=d.items.read(c).every(u=>u.collapsed.read(c));a.set(h)}})),this._register(De(c=>{const d=this._dimension.read(c);this._sizeObserver.observe(d)}));const l=j(c=>{if(this._viewItems.read(c).length>0)return;const h=this._viewModel.read(c);return!h||h.isLoading.read(c)?m(145,"Loading..."):m(146,"No Changed Files")});this._register(De(c=>{const d=l.read(c);this._elements.placeholder.innerText=d??"",this._elements.placeholder.classList.toggle("visible",!!d)})),this._scrollableElements.content.style.position="relative",this._register(De(c=>{const d=this._sizeObserver.height.read(c);this._scrollableElements.root.style.height=`${d}px`;const h=this._totalHeight.read(c);this._scrollableElements.content.style.height=`${h}px`;const u=this._sizeObserver.width.read(c);let g=u;const f=this._viewItems.read(c),p=jW(f,ns(_=>_.maxScroll.read(c).maxScroll,Wo));if(p){const _=p.maxScroll.read(c);g=u+_.maxScroll}this._scrollableElement.setScrollDimensions({width:u,height:d,scrollHeight:h,scrollWidth:g})})),e.replaceChildren(this._elements.root),this._register(pe(()=>{e.replaceChildren()})),this._register(De(c=>{const d=this._viewModel.read(c);if(d&&!d.isLoading.read(c)){if(d.items.read(c).length===0||d.activeDiffItem.read(c))return;this.goToNextChange()}})),this._register(this._register(De(c=>{o1(d=>{this.render(c)})})))}reveal(e,t){const i=this._viewItems.get(),n=i.findIndex(c=>c.viewModel.originalUri?.toString()===e.original?.toString()&&c.viewModel.modifiedUri?.toString()===e.modified?.toString());if(n===-1)throw new Ce("Resource not found in diff editor");const o=i[n];this._viewModel.get().activeDiffItem.setCache(o.viewModel,void 0);let r=0;for(let c=0;c<n;c++)r+=i[c].contentHeight.get()+this._spaceBetweenPx;this._scrollableElement.setScrollPosition({scrollTop:r});const a=o.template.get()?.editor,l="original"in e?a?.getOriginalEditor():a?.getModifiedEditor();l&&t?.range&&(l.revealRangeInCenter(t.range),cWe(l,t.range))}goToNextChange(){this._navigateToChange("next")}_navigateToChange(e){const t=this._viewItems.get();if(t.length===0)return;const i=this._viewModel.get()?.activeDiffItem.get(),n=i?t.findIndex(l=>l.viewModel===i):-1;if(n===-1){this._goToFile(0,"first");return}const o=t[n];o.viewModel.collapsed.get()&&o.viewModel.collapsed.set(!1,void 0);const r=o.template.get()?.editor;if(r?.getDiffComputationResult()?.changes2?.length){const l=r.getModifiedEditor().getPosition()?.lineNumber||1,c=r.getDiffComputationResult().changes2;if(e==="next"?c.some(h=>h.modified.startLineNumber>l):c.some(h=>h.modified.endLineNumberExclusive<=l)){r.goToDiff(e);return}}const a=(n+(e==="next"?1:-1)+t.length)%t.length;this._goToFile(a,e==="next"?"first":"last")}_goToFile(e,t){const i=this._viewItems.get()[e];i.viewModel.collapsed.get()&&i.viewModel.collapsed.set(!1,void 0),this.reveal({original:i.viewModel.originalUri,modified:i.viewModel.modifiedUri});const n=i.template.get()?.editor;if(n?.getDiffComputationResult()?.changes2?.length)if(t==="first")n.revealFirstDiff();else{const o=n.getDiffComputationResult().changes2.at(-1),r=n.getModifiedEditor();r.setPosition({lineNumber:o.modified.startLineNumber,column:1}),r.revealLineInCenter(o.modified.startLineNumber)}n?.focus()}render(e){const t=this.scrollTop.read(e);let i=0,n=0,o=0;const r=this._sizeObserver.height.read(e),a=me.ofStartAndLength(t,r),l=this._sizeObserver.width.read(e);for(const c of this._viewItems.read(e)){const d=c.contentHeight.read(e),h=Math.min(d,r),u=me.ofStartAndLength(n,h),g=me.ofStartAndLength(o,d);if(g.isBefore(a))i-=d-h,c.hide();else if(g.isAfter(a))c.hide();else{const f=Math.max(0,Math.min(a.start-g.start,d-h));i-=f;const p=me.ofStartAndLength(t+i,r);c.render(u,f,l,p)}n+=h+this._spaceBetweenPx,o+=d+this._spaceBetweenPx}this._scrollableElements.content.style.transform=`translateY(${-(t+i)}px)`}};a6=lWe([GY(4,Ae),GY(5,_e)],a6);function cWe(s,e){const t=s.getModel(),i=s.createDecorationsCollection([{range:e,options:{description:"symbol-navigate-action-highlight",className:"symbolHighlight"}}]);setTimeout(()=>{s.getModel()===t&&i.clear()},350)}class dWe extends F{constructor(e,t,i,n){super(),this.viewModel=e,this._objectPool=t,this._scrollLeft=i,this._deltaScrollVertical=n,this._templateRef=this._register(lW(this,void 0)),this.contentHeight=j(this,o=>this._templateRef.read(o)?.object.contentHeight?.read(o)??this.viewModel.lastTemplateData.read(o).contentHeight),this.maxScroll=j(this,o=>this._templateRef.read(o)?.object.maxScroll.read(o)??{maxScroll:0,scrollWidth:0}),this.template=j(this,o=>this._templateRef.read(o)?.object),this._isHidden=Re(this,!1),this._isFocused=j(this,o=>this.template.read(o)?.isFocused.read(o)??!1),this.viewModel.setIsFocused(this._isFocused,void 0),this._register(De(o=>{const r=this._scrollLeft.read(o);this._templateRef.read(o)?.object.setScrollLeft(r)})),this._register(De(o=>{const r=this._templateRef.read(o);!r||!this._isHidden.read(o)||r.object.isFocused.read(o)||this._clear()}))}dispose(){this._clear(),super.dispose()}toString(){return`VirtualViewItem(${this.viewModel.documentDiffItem.modified?.uri.toString()})`}getKey(){return this.viewModel.getKey()}setViewState(e,t){this.viewModel.collapsed.set(e.collapsed,t),this._updateTemplateData(t);const i=this.viewModel.lastTemplateData.get(),n=e.selections?.map(he.liftSelection);this.viewModel.lastTemplateData.set({...i,selections:n},t);const o=this._templateRef.get();o&&n&&o.object.editor.setSelections(n)}_updateTemplateData(e){const t=this._templateRef.get();t&&this.viewModel.lastTemplateData.set({contentHeight:t.object.contentHeight.get(),selections:t.object.editor.getSelections()??void 0},e)}_clear(){const e=this._templateRef.get();e&&Vt(t=>{this._updateTemplateData(t),e.object.hide(),this._templateRef.set(void 0,t)})}hide(){this._isHidden.set(!0,void 0)}render(e,t,i,n){this._isHidden.set(!1,void 0);let o=this._templateRef.get();if(!o){o=this._objectPool.getUnusedObj(new rWe(this.viewModel,this._deltaScrollVertical)),this._templateRef.set(o,void 0);const r=this.viewModel.lastTemplateData.get().selections;r&&o.object.editor.setSelections(r)}o.object.render(e,i,t,n)}}var hWe=function(s,e,t,i){var n=arguments.length,o=n<3?e:i===null?i=Object.getOwnPropertyDescriptor(e,t):i,r;if(typeof Reflect=="object"&&typeof Reflect.decorate=="function")o=Reflect.decorate(s,e,t,i);else for(var a=s.length-1;a>=0;a--)(r=s[a])&&(o=(n<3?r(o):n>3?r(e,t,o):r(e,t))||o);return n>3&&o&&Object.defineProperty(e,t,o),o},uWe=function(s,e){return function(t,i){e(t,i,s)}};let l6=class extends F{constructor(e,t,i){super(),this._element=e,this._workbenchUIElementFactory=t,this._instantiationService=i,this._dimension=Re(this,void 0),this._viewModel=Re(this,void 0),this._widgetImpl=j(this,n=>n.store.add(this._instantiationService.createInstance(fd(a6),this._element,this._dimension,this._viewModel,this._workbenchUIElementFactory))),this._register(sv(this._widgetImpl))}};l6=hWe([uWe(2,_e)],l6);function gWe(s,e,t){return Pe.initialize(t||{}).createInstance(s6,s,e)}function fWe(s){return Pe.get(ht).onCodeEditorAdd(t=>{s(t)})}function pWe(s){return Pe.get(ht).onDiffEditorAdd(t=>{s(t)})}function mWe(){return Pe.get(ht).listCodeEditors()}function _We(){return Pe.get(ht).listDiffEditors()}function bWe(s,e,t){return Pe.initialize(t||{}).createInstance(o6,s,e)}function wWe(s,e){const t=Pe.initialize(e||{});return new l6(s,{},t)}function vWe(s){if(typeof s.id!="string"||typeof s.run!="function")throw new Error("Invalid command descriptor, `id` and `run` are required properties!");return at.registerCommand(s.id,s.run)}function CWe(s){if(typeof s.id!="string"||typeof s.label!="string"||typeof s.run!="function")throw new Error("Invalid action descriptor, `id`, `label` and `run` are required properties!");const e=K.deserialize(s.precondition),t=(n,...o)=>hn.runEditorCommand(n,o,e,(r,a,l)=>Promise.resolve(s.run(a,...l))),i=new U;if(i.add(at.registerCommand(s.id,t)),s.contextMenuGroupId){const n={command:{id:s.id,title:s.label},when:e,group:s.contextMenuGroupId,order:s.contextMenuOrder||0};i.add(En.appendMenuItem(fe.EditorContext,n))}if(Array.isArray(s.keybindings)){const n=Pe.get(ft);if(!(n instanceof Bv))console.warn("Cannot add keybinding because the editor is configured with an unrecognized KeybindingService");else{const o=K.and(e,K.deserialize(s.keybindingContext));i.add(n.addDynamicKeybindings(s.keybindings.map(r=>({keybinding:r,command:s.id,when:o}))))}}return i}function yWe(s){return Ure([s])}function Ure(s){const e=Pe.get(ft);return e instanceof Bv?e.addDynamicKeybindings(s.map(t=>({keybinding:t.keybinding,command:t.command,commandArgs:t.commandArgs,when:K.deserialize(t.when)}))):(console.warn("Cannot add keybinding because the editor is configured with an unrecognized KeybindingService"),F.None)}function SWe(s,e,t){const i=Pe.get(Ei),n=i.getLanguageIdByMimeType(e)||e;return $re(Pe.get(hi),i,s,n,t)}function xWe(s,e){const t=Pe.get(Ei),i=t.getLanguageIdByMimeType(e)||e||xr;s.setLanguage(t.createById(i))}function LWe(s,e,t){s&&Pe.get(jc).changeOne(e,s.uri,t)}function kWe(s){Pe.get(jc).changeAll(s,[])}function EWe(s){return Pe.get(jc).read(s)}function IWe(s){return Pe.get(jc).onMarkerChanged(s)}function DWe(s){return Pe.get(hi).getModel(s)}function TWe(){return Pe.get(hi).getModels()}function NWe(s){return Pe.get(hi).onModelAdded(s)}function RWe(s){return Pe.get(hi).onModelRemoved(s)}function AWe(s){return Pe.get(hi).onModelLanguageChanged(t=>{s({model:t.model,oldLanguage:t.oldLanguageId})})}function MWe(s){return WBe(Pe.get(hi),s)}function OWe(s,e){const t=Pe.get(Ei),i=Pe.get(Rr);return wV.colorizeElement(i,t,s,e).then(()=>{i.registerEditorContainer(s)})}function PWe(s,e,t){const i=Pe.get(Ei);return Pe.get(Rr).registerEditorContainer(Nt.document.body),wV.colorize(i,s,e,t)}function FWe(s,e,t=4){return Pe.get(Rr).registerEditorContainer(Nt.document.body),wV.colorizeModelLine(s,e,t)}function BWe(s){const e=yi.get(s);return e||{getInitialState:()=>bv,tokenize:(t,i,n)=>h7(s,n)}}function WWe(s,e){yi.getOrCreate(e);const t=BWe(e),i=qo(s),n=[];let o=t.getInitialState();for(let r=0,a=i.length;r<a;r++){const l=i[r],c=t.tokenize(l,!0,o);n[r]=c.tokens,o=c.endState}return n}function HWe(s,e){Pe.get(Rr).defineTheme(s,e)}function VWe(s){Pe.get(Rr).setTheme(s)}function zWe(){UP.clearAllFontInfos()}function $We(s,e){return at.registerCommand({id:s,handler:e})}function UWe(s){return Pe.get(Ah).registerOpener({async open(t){return typeof t=="string"&&(t=Le.parse(t)),s.open(t)}})}function qWe(s){return Pe.get(ht).registerCodeEditorOpenHandler(async(t,i,n)=>{if(!i)return null;const o=t.options?.selection;let r;return o&&typeof o.endLineNumber=="number"&&typeof o.endColumn=="number"?r=o:o&&(r={lineNumber:o.startLineNumber,column:o.startColumn}),await s.openCodeEditor(i,t.resource,r)?i:null})}function jWe(){return{create:gWe,getEditors:mWe,getDiffEditors:_We,onDidCreateEditor:fWe,onDidCreateDiffEditor:pWe,createDiffEditor:bWe,addCommand:vWe,addEditorAction:CWe,addKeybindingRule:yWe,addKeybindingRules:Ure,createModel:SWe,setModelLanguage:xWe,setModelMarkers:LWe,getModelMarkers:EWe,removeAllMarkers:kWe,onDidChangeMarkers:IWe,getModels:TWe,getModel:DWe,onDidCreateModel:NWe,onWillDisposeModel:RWe,onDidChangeModelLanguage:AWe,createWebWorker:MWe,colorizeElement:OWe,colorize:PWe,colorizeModelLine:FWe,tokenize:WWe,defineTheme:HWe,setTheme:VWe,remeasureFonts:zWe,registerCommand:$We,registerLinkOpener:UWe,registerEditorOpener:qWe,AccessibilitySupport:sB,ContentWidgetPositionPreference:dB,CursorChangeReason:hB,DefaultEndOfLine:uB,EditorAutoIndentStrategy:fB,EditorOption:pB,EndOfLinePreference:mB,EndOfLineSequence:_B,MinimapPosition:DB,MinimapSectionHeaderStyle:TB,MouseTargetType:NB,OverlayWidgetPositionPreference:MB,OverviewRulerLane:OB,GlyphMarginLane:bB,RenderLineNumbersType:BB,RenderMinimap:WB,ScrollbarVisibility:VB,ScrollType:HB,TextEditorCursorBlinkingStyle:KB,TextEditorCursorStyle:YB,TrackedRangeStickiness:ZB,WrappingIndent:XB,InjectedTextCursorStops:CB,PositionAffinity:FB,ShowLightbulbIconMode:$B,TextDirection:GB,ConfigurationChangedEvent:SJ,BareFontInfo:Em,FontInfo:BE,TextModelResolvedOptions:kk,FindMatch:tS,ApplyUpdateResult:T1,EditorZoom:Gr,createMultiFileDiffEditor:wWe,EditorType:sx,EditorOptions:Is}}function GWe(s,e){if(!e||!Array.isArray(e))return!1;for(const t of e)if(!s(t))return!1;return!0}function uk(s,e){return typeof s=="boolean"?s:e}function KY(s,e){return typeof s=="string"?s:e}function KWe(s){const e={};for(const t of s)e[t]=!0;return e}function YY(s,e=!1){e&&(s=s.map(function(i){return i.toLowerCase()}));const t=KWe(s);return e?function(i){return t[i.toLowerCase()]!==void 0&&t.hasOwnProperty(i.toLowerCase())}:function(i){return t[i]!==void 0&&t.hasOwnProperty(i)}}function c6(s,e,t){e=e.replace(/@@/g,"");let i=0,n;do n=!1,e=e.replace(/@(\w+)/g,function(r,a){n=!0;let l="";if(typeof s[a]=="string")l=s[a];else if(s[a]&&s[a]instanceof RegExp)l=s[a].source;else throw s[a]===void 0?ei(s,"language definition does not contain attribute '"+a+"', used at: "+e):ei(s,"attribute reference '"+a+"' must be a string, used at: "+e);return Lf(l)?"":"(?:"+l+")"}),i++;while(n&&i<5);e=e.replace(/\x01/g,"@");const o=(s.ignoreCase?"i":"")+(s.unicode?"u":"");if(t&&e.match(/\$[sS](\d\d?)/g)){let a=null,l=null;return c=>(l&&a===c||(a=c,l=new RegExp(UBe(s,e,c),o)),l)}return new RegExp(e,o)}function YWe(s,e,t,i){if(i<0)return s;if(i<e.length)return e[i];if(i>=100){i=i-100;const n=t.split(".");if(n.unshift(t),i<n.length)return n[i]}return null}function ZWe(s,e,t,i){let n=-1,o=t,r=t.match(/^\$(([sS]?)(\d\d?)|#)(.*)$/);r&&(r[3]&&(n=parseInt(r[3]),r[2]&&(n=n+100)),o=r[4]);let a="~",l=o;!o||o.length===0?(a="!=",l=""):/^\w*$/.test(l)?a="==":(r=o.match(/^(@|!@|~|!~|==|!=)(.*)$/),r&&(a=r[1],l=r[2]));let c;if((a==="~"||a==="!~")&&/^(\w|\|)*$/.test(l)){const d=YY(l.split("|"),s.ignoreCase);c=function(h){return a==="~"?d(h):!d(h)}}else if(a==="@"||a==="!@"){const d=s[l];if(!d)throw ei(s,"the @ match target '"+l+"' is not defined, in rule: "+e);if(!GWe(function(u){return typeof u=="string"},d))throw ei(s,"the @ match target '"+l+"' must be an array of strings, in rule: "+e);const h=YY(d,s.ignoreCase);c=function(u){return a==="@"?h(u):!h(u)}}else if(a==="~"||a==="!~")if(l.indexOf("$")<0){const d=c6(s,"^"+l+"$",!1);c=function(h){return a==="~"?d.test(h):!d.test(h)}}else c=function(d,h,u,g){return c6(s,"^"+hu(s,l,h,u,g)+"$",!1).test(d)};else if(l.indexOf("$")<0){const d=lh(s,l);c=function(h){return a==="=="?h===d:h!==d}}else{const d=lh(s,l);c=function(h,u,g,f,p){const _=hu(s,d,u,g,f);return a==="=="?h===_:h!==_}}return n===-1?{name:t,value:i,test:function(d,h,u,g){return c(d,d,h,u,g)}}:{name:t,value:i,test:function(d,h,u,g){const f=YWe(d,h,u,n);return c(f||"",d,h,u,g)}}}function d6(s,e,t){if(t){if(typeof t=="string")return t;if(t.token||t.token===""){if(typeof t.token!="string")throw ei(s,"a 'token' attribute must be of type string, in rule: "+e);{const i={token:t.token};if(t.token.indexOf("$")>=0&&(i.tokenSubst=!0),typeof t.bracket=="string")if(t.bracket==="@open")i.bracket=1;else if(t.bracket==="@close")i.bracket=-1;else throw ei(s,"a 'bracket' attribute must be either '@open' or '@close', in rule: "+e);if(t.next){if(typeof t.next!="string")throw ei(s,"the next state must be a string value in rule: "+e);{let n=t.next;if(!/^(@pop|@push|@popall)$/.test(n)&&(n[0]==="@"&&(n=n.substr(1)),n.indexOf("$")<0&&!qBe(s,hu(s,n,"",[],""))))throw ei(s,"the next state '"+t.next+"' is not defined in rule: "+e);i.next=n}}return typeof t.goBack=="number"&&(i.goBack=t.goBack),typeof t.switchTo=="string"&&(i.switchTo=t.switchTo),typeof t.log=="string"&&(i.log=t.log),typeof t.nextEmbedded=="string"&&(i.nextEmbedded=t.nextEmbedded,s.usesEmbedded=!0),i}}else if(Array.isArray(t)){const i=[];for(let n=0,o=t.length;n<o;n++)i[n]=d6(s,e,t[n]);return{group:i}}else if(t.cases){const i=[];let n=!1;for(const r in t.cases)if(t.cases.hasOwnProperty(r)){const a=d6(s,e,t.cases[r]);r==="@default"||r==="@"||r===""?i.push({test:void 0,value:a,name:r}):r==="@eos"?i.push({test:function(l,c,d,h){return h},value:a,name:r}):i.push(ZWe(s,e,r,a)),n||(n=!bn(a)&&(a.hasEmbeddedEndInCases||["@pop","@popall"].includes(a.nextEmbedded||"")))}const o=s.defaultToken;return{hasEmbeddedEndInCases:n,test:function(r,a,l,c){for(const d of i)if(!d.test||d.test(r,a,l,c))return d.value;return o}}}else throw ei(s,"an action must be a string, an object with a 'token' or 'cases' attribute, or an array of actions; in rule: "+e)}else return{token:""}}class XWe{constructor(e){this.regex=new RegExp(""),this.action={token:""},this.matchOnlyAtLineStart=!1,this.name="",this.name=e}setRegex(e,t){let i;if(typeof t=="string")i=t;else if(t instanceof RegExp)i=t.source;else throw ei(e,"rules must start with a match string or regular expression: "+this.name);this.matchOnlyAtLineStart=i.length>0&&i[0]==="^",this.name=this.name+": "+i,this.regex=c6(e,"^(?:"+(this.matchOnlyAtLineStart?i.substr(1):i)+")",!0)}setAction(e,t){this.action=d6(e,this.name,t)}resolveRegex(e){return this.regex instanceof RegExp?this.regex:this.regex(e)}}function qre(s,e){if(!e||typeof e!="object")throw new Error("Monarch: expecting a language definition object");const t={languageId:s,includeLF:uk(e.includeLF,!1),noThrow:!1,maxStack:100,start:typeof e.start=="string"?e.start:null,ignoreCase:uk(e.ignoreCase,!1),unicode:uk(e.unicode,!1),tokenPostfix:KY(e.tokenPostfix,"."+s),defaultToken:KY(e.defaultToken,"source"),usesEmbedded:!1,stateNames:{},tokenizer:{},brackets:[]},i=e;i.languageId=s,i.includeLF=t.includeLF,i.ignoreCase=t.ignoreCase,i.unicode=t.unicode,i.noThrow=t.noThrow,i.usesEmbedded=t.usesEmbedded,i.stateNames=e.tokenizer,i.defaultToken=t.defaultToken;function n(r,a,l){for(const c of l){let d=c.include;if(d){if(typeof d!="string")throw ei(t,"an 'include' attribute must be a string at: "+r);if(d[0]==="@"&&(d=d.substr(1)),!e.tokenizer[d])throw ei(t,"include target '"+d+"' is not defined at: "+r);n(r+"."+d,a,e.tokenizer[d])}else{const h=new XWe(r);if(Array.isArray(c)&&c.length>=1&&c.length<=3)if(h.setRegex(i,c[0]),c.length>=3)if(typeof c[1]=="string")h.setAction(i,{token:c[1],next:c[2]});else if(typeof c[1]=="object"){const u=c[1];u.next=c[2],h.setAction(i,u)}else throw ei(t,"a next state as the last element of a rule can only be given if the action is either an object or a string, at: "+r);else h.setAction(i,c[1]);else{if(!c.regex)throw ei(t,"a rule must either be an array, or an object with a 'regex' or 'include' field at: "+r);c.name&&typeof c.name=="string"&&(h.name=c.name),c.matchOnlyAtStart&&(h.matchOnlyAtLineStart=uk(c.matchOnlyAtLineStart,!1)),h.setRegex(i,c.regex),h.setAction(i,c.action)}a.push(h)}}}if(!e.tokenizer||typeof e.tokenizer!="object")throw ei(t,"a language definition must define the 'tokenizer' attribute as an object");t.tokenizer=[];for(const r in e.tokenizer)if(e.tokenizer.hasOwnProperty(r)){t.start||(t.start=r);const a=e.tokenizer[r];t.tokenizer[r]=new Array,n("tokenizer."+r,t.tokenizer[r],a)}if(t.usesEmbedded=i.usesEmbedded,e.brackets){if(!Array.isArray(e.brackets))throw ei(t,"the 'brackets' attribute must be defined as an array")}else e.brackets=[{open:"{",close:"}",token:"delimiter.curly"},{open:"[",close:"]",token:"delimiter.square"},{open:"(",close:")",token:"delimiter.parenthesis"},{open:"<",close:">",token:"delimiter.angle"}];const o=[];for(const r of e.brackets){let a=r;if(a&&Array.isArray(a)&&a.length===3&&(a={token:a[2],open:a[0],close:a[1]}),a.open===a.close)throw ei(t,"open and close brackets in a 'brackets' attribute must be different: "+a.open+`
 hint: use the 'bracket' attribute if matching on equal brackets is required.`);if(typeof a.open=="string"&&typeof a.token=="string"&&typeof a.close=="string")o.push({token:a.token+t.tokenPostfix,open:lh(t,a.open),close:lh(t,a.close)});else throw ei(t,"every element in the 'brackets' array must be a '{open,close,token}' object or array")}return t.brackets=o,t.noThrow=!0,t}function QWe(s){rv.registerLanguage(s)}function JWe(){let s=[];return s=s.concat(rv.getLanguages()),s}function e7e(s){return Pe.get(Ei).languageIdCodec.encodeLanguageId(s)}function t7e(s,e){return Pe.withServices(()=>{const i=Pe.get(Ei).onDidRequestRichLanguageFeatures(n=>{n===s&&(i.dispose(),e())});return i})}function i7e(s,e){return Pe.withServices(()=>{const i=Pe.get(Ei).onDidRequestBasicLanguageFeatures(n=>{n===s&&(i.dispose(),e())});return i})}function n7e(s,e){if(!Pe.get(Ei).isRegisteredLanguageId(s))throw new Error(`Cannot set configuration for unknown language ${s}`);return Pe.get(ui).register(s,e,100)}class s7e{constructor(e,t){this._languageId=e,this._actual=t}dispose(){}getInitialState(){return this._actual.getInitialState()}tokenize(e,t,i){if(typeof this._actual.tokenize=="function")return N0.adaptTokenize(this._languageId,this._actual,e,i);throw new Error("Not supported!")}tokenizeEncoded(e,t,i){const n=this._actual.tokenizeEncoded(e,i);return new gA(n.tokens,n.endState)}}class N0{constructor(e,t,i,n){this._languageId=e,this._actual=t,this._languageService=i,this._standaloneThemeService=n}dispose(){}getInitialState(){return this._actual.getInitialState()}static _toClassicTokens(e,t){const i=[];let n=0;for(let o=0,r=e.length;o<r;o++){const a=e[o];let l=a.startIndex;o===0?l=0:l<n&&(l=n),i[o]=new pS(l,a.scopes,t),n=l}return i}static adaptTokenize(e,t,i,n){const o=t.tokenize(i,n),r=N0._toClassicTokens(o.tokens,e);let a;return o.endState.equals(n)?a=n:a=o.endState,new GW(r,a)}tokenize(e,t,i){return N0.adaptTokenize(this._languageId,this._actual,e,i)}_toBinaryTokens(e,t){const i=e.encodeLanguageId(this._languageId),n=this._standaloneThemeService.getColorTheme().tokenTheme,o=[];let r=0,a=0;for(let c=0,d=t.length;c<d;c++){const h=t[c],u=n.match(i,h.scopes)|1024;if(r>0&&o[r-1]===u)continue;let g=h.startIndex;c===0?g=0:g<a&&(g=a),o[r++]=g,o[r++]=u,a=g}const l=new Uint32Array(r);for(let c=0;c<r;c++)l[c]=o[c];return l}tokenizeEncoded(e,t,i){const n=this._actual.tokenize(e,i),o=this._toBinaryTokens(this._languageService.languageIdCodec,n.tokens);let r;return n.endState.equals(i)?r=i:r=n.endState,new gA(o,r)}}function o7e(s){return typeof s.getInitialState=="function"}function r7e(s){return"tokenizeEncoded"in s}function jre(s){return s&&typeof s.then=="function"}function a7e(s){const e=Pe.get(Rr);if(s){const t=[null];for(let i=1,n=s.length;i<n;i++)t[i]=q.fromHex(s[i]);e.setColorMapOverride(t)}else e.setColorMapOverride(null)}function Gre(s,e){return r7e(e)?new s7e(s,e):new N0(s,e,Pe.get(Ei),Pe.get(Rr))}function LV(s,e){const t=new eve(async()=>{const i=await Promise.resolve(e.create());return i?o7e(i)?Gre(s,i):new x0(Pe.get(Ei),Pe.get(Rr),s,qre(s,i),Pe.get(He)):null});return yi.registerFactory(s,t)}function l7e(s,e){if(!Pe.get(Ei).isRegisteredLanguageId(s))throw new Error(`Cannot set tokens provider for unknown language ${s}`);return jre(e)?LV(s,{create:()=>e}):yi.register(s,Gre(s,e))}function c7e(s,e){const t=i=>new x0(Pe.get(Ei),Pe.get(Rr),s,qre(s,i),Pe.get(He));return jre(e)?LV(s,{create:()=>e}):yi.register(s,t(e))}function d7e(s,e){return Pe.get(ge).referenceProvider.register(s,e)}function h7e(s,e){return Pe.get(ge).renameProvider.register(s,e)}function u7e(s,e){return Pe.get(ge).newSymbolNamesProvider.register(s,e)}function g7e(s,e){return Pe.get(ge).signatureHelpProvider.register(s,e)}function f7e(s,e){return Pe.get(ge).hoverProvider.register(s,{provideHover:async(i,n,o,r)=>{const a=i.getWordAtPosition(n);return Promise.resolve(e.provideHover(i,n,o,r)).then(l=>{if(l)return!l.range&&a&&(l.range=new k(n.lineNumber,a.startColumn,n.lineNumber,a.endColumn)),l.range||(l.range=new k(n.lineNumber,n.column,n.lineNumber,n.column)),l})}})}function p7e(s,e){return Pe.get(ge).documentSymbolProvider.register(s,e)}function m7e(s,e){return Pe.get(ge).documentHighlightProvider.register(s,e)}function _7e(s,e){return Pe.get(ge).linkedEditingRangeProvider.register(s,e)}function b7e(s,e){return Pe.get(ge).definitionProvider.register(s,e)}function w7e(s,e){return Pe.get(ge).implementationProvider.register(s,e)}function v7e(s,e){return Pe.get(ge).typeDefinitionProvider.register(s,e)}function C7e(s,e){return Pe.get(ge).codeLensProvider.register(s,e)}function y7e(s,e,t){return Pe.get(ge).codeActionProvider.register(s,{providedCodeActionKinds:t?.providedCodeActionKinds,documentation:t?.documentation,provideCodeActions:(n,o,r,a)=>{const c=Pe.get(jc).read({resource:n.uri}).filter(d=>k.areIntersectingOrTouching(d,o));return e.provideCodeActions(n,o,{markers:c,only:r.only,trigger:r.trigger},a)},resolveCodeAction:e.resolveCodeAction})}function S7e(s,e){return Pe.get(ge).documentFormattingEditProvider.register(s,e)}function x7e(s,e){return Pe.get(ge).documentRangeFormattingEditProvider.register(s,e)}function L7e(s,e){return Pe.get(ge).onTypeFormattingEditProvider.register(s,e)}function k7e(s,e){return Pe.get(ge).linkProvider.register(s,e)}function E7e(s,e){return Pe.get(ge).completionProvider.register(s,e)}function I7e(s,e){return Pe.get(ge).colorProvider.register(s,e)}function D7e(s,e){return Pe.get(ge).foldingRangeProvider.register(s,e)}function T7e(s,e){return Pe.get(ge).declarationProvider.register(s,e)}function N7e(s,e){return Pe.get(ge).selectionRangeProvider.register(s,e)}function R7e(s,e){return Pe.get(ge).documentSemanticTokensProvider.register(s,e)}function A7e(s,e){return Pe.get(ge).documentRangeSemanticTokensProvider.register(s,e)}function M7e(s,e){return Pe.get(ge).inlineCompletionsProvider.register(s,e)}function O7e(s,e){return Pe.get(ge).inlayHintsProvider.register(s,e)}function P7e(){return{register:QWe,getLanguages:JWe,onLanguage:t7e,onLanguageEncountered:i7e,getEncodedLanguageId:e7e,setLanguageConfiguration:n7e,setColorMap:a7e,registerTokensProviderFactory:LV,setTokensProvider:l7e,setMonarchTokensProvider:c7e,registerReferenceProvider:d7e,registerRenameProvider:h7e,registerNewSymbolNameProvider:u7e,registerCompletionItemProvider:E7e,registerSignatureHelpProvider:g7e,registerHoverProvider:f7e,registerDocumentSymbolProvider:p7e,registerDocumentHighlightProvider:m7e,registerLinkedEditingRangeProvider:_7e,registerDefinitionProvider:b7e,registerImplementationProvider:w7e,registerTypeDefinitionProvider:v7e,registerCodeLensProvider:C7e,registerCodeActionProvider:y7e,registerDocumentFormattingEditProvider:S7e,registerDocumentRangeFormattingEditProvider:x7e,registerOnTypeFormattingEditProvider:L7e,registerLinkProvider:k7e,registerColorProvider:I7e,registerFoldingRangeProvider:D7e,registerDeclarationProvider:T7e,registerSelectionRangeProvider:N7e,registerDocumentSemanticTokensProvider:R7e,registerDocumentRangeSemanticTokensProvider:A7e,registerInlineCompletionsProvider:M7e,registerInlayHintsProvider:O7e,DocumentHighlightKind:gB,CompletionItemKind:aB,CompletionItemTag:lB,CompletionItemInsertTextRule:rB,SymbolKind:qB,SymbolTag:jB,IndentAction:vB,CompletionTriggerKind:cB,SignatureHelpTriggerKind:UB,InlayHintKind:yB,InlineCompletionTriggerKind:LB,CodeActionTriggerType:oB,NewSymbolNameTag:RB,NewSymbolNameTriggerKind:AB,PartialAcceptTriggerKind:PB,HoverVerbosityAction:wB,InlineCompletionEndOfLifeReasonKind:SB,InlineCompletionHintStyle:xB,FoldingRangeKind:h_,SelectedSuggestionInfo:Eee,EditDeltaInfo:ES}}Is.wrappingIndent.defaultValue=0;Is.glyphMargin.defaultValue=!1;Is.autoIndent.defaultValue=3;Is.overviewRulerLanes.defaultValue=2;i0.setFormatterSelector((s,e,t)=>Promise.resolve(s[0]));const ro=vre();ro.editor=jWe();ro.languages=P7e();const Kre=ro.CancellationTokenSource,Sg=ro.Emitter,Yre=ro.KeyCode,Zre=ro.KeyMod,Fh=ro.Position,Hc=ro.Range,Xre=ro.Selection,Qre=ro.SelectionDirection,kn=ro.MarkerSeverity,Wv=ro.MarkerTag,Yc=ro.Uri,Jre=ro.Token,ti=ro.editor,H=ro.languages,F7e=M6(),Mw=globalThis;(F7e?.globalAPI||typeof Mw.define=="function"&&Mw.define.amd)&&(Mw.monaco=ro);typeof Mw.require<"u"&&typeof Mw.require.config=="function"&&Mw.require.config({ignoreDuplicateModules:["vscode-languageserver-types","vscode-languageserver-types/main","vscode-languageserver-textdocument","vscode-languageserver-textdocument/main","vscode-nls","vscode-nls/vscode-nls","jsonc-parser","jsonc-parser/main","vscode-uri","vscode-uri/index","vs/basic-languages/typescript/typescript"]});const B7e=Object.freeze(Object.defineProperty({__proto__:null,CancellationTokenSource:Kre,Emitter:Sg,KeyCode:Yre,KeyMod:Zre,MarkerSeverity:kn,MarkerTag:Wv,Position:Fh,Range:Hc,Selection:Xre,SelectionDirection:Qre,Token:Jre,Uri:Yc,editor:ti,languages:H},Symbol.toStringTag,{value:"Module"}));let kV=class{constructor(e,t,i){this._onDidChange=new Sg,this._languageId=e,this.setOptions(t),this.setModeConfiguration(i)}get onDidChange(){return this._onDidChange.event}get languageId(){return this._languageId}get modeConfiguration(){return this._modeConfiguration}get diagnosticsOptions(){return this.options}get options(){return this._options}setOptions(e){this._options=e||Object.create(null),this._onDidChange.fire(this)}setDiagnosticsOptions(e){this.setOptions(e)}setModeConfiguration(e){this._modeConfiguration=e||Object.create(null),this._onDidChange.fire(this)}};const EV={validate:!0,lint:{compatibleVendorPrefixes:"ignore",vendorPrefix:"warning",duplicateProperties:"warning",emptyRules:"warning",importStatement:"ignore",boxModel:"ignore",universalSelector:"ignore",zeroUnits:"ignore",fontFaceProperties:"warning",hexColorLength:"error",argumentsInColorFunction:"error",unknownProperties:"warning",ieHack:"ignore",unknownVendorSpecificProperties:"ignore",propertyIgnoredDueToDisplay:"warning",important:"ignore",float:"ignore",idSelector:"ignore"},data:{useDefaultDataProvider:!0},format:{newlineBetweenSelectors:!0,newlineBetweenRules:!0,spaceAroundSelectorSeparator:!1,braceStyle:"collapse",maxPreserveNewLines:void 0,preserveNewLines:!0}},IV={completionItems:!0,hovers:!0,documentSymbols:!0,definitions:!0,references:!0,documentHighlights:!0,rename:!0,colors:!0,foldingRanges:!0,diagnostics:!0,selectionRanges:!0,documentFormattingEdits:!0,documentRangeFormattingEdits:!0},eae=new kV("css",EV,IV),tae=new kV("scss",EV,IV),iae=new kV("less",EV,IV);function DV(){return Se(()=>Promise.resolve().then(()=>wze),void 0)}H.onLanguage("less",()=>{DV().then(s=>s.setupMode(iae))});H.onLanguage("scss",()=>{DV().then(s=>s.setupMode(tae))});H.onLanguage("css",()=>{DV().then(s=>s.setupMode(eae))});const nae=Object.freeze(Object.defineProperty({__proto__:null,cssDefaults:eae,lessDefaults:iae,scssDefaults:tae},Symbol.toStringTag,{value:"Module"}));let W7e=class{constructor(e,t,i){this._onDidChange=new Sg,this._languageId=e,this.setOptions(t),this.setModeConfiguration(i)}get onDidChange(){return this._onDidChange.event}get languageId(){return this._languageId}get options(){return this._options}get modeConfiguration(){return this._modeConfiguration}setOptions(e){this._options=e||Object.create(null),this._onDidChange.fire(this)}setModeConfiguration(e){this._modeConfiguration=e||Object.create(null),this._onDidChange.fire(this)}};const H7e={tabSize:4,insertSpaces:!1,wrapLineLength:120,unformatted:'default": "a, abbr, acronym, b, bdo, big, br, button, cite, code, dfn, em, i, img, input, kbd, label, map, object, q, samp, select, small, span, strong, sub, sup, textarea, tt, var',contentUnformatted:"pre",indentInnerHtml:!1,preserveNewLines:!0,maxPreserveNewLines:void 0,indentHandlebars:!1,endWithNewline:!1,extraLiners:"head, body, /html",wrapAttributes:"auto"},a2={format:H7e,suggest:{},data:{useDefaultDataProvider:!0}};function l2(s){return{completionItems:!0,hovers:!0,documentSymbols:!0,links:!0,documentHighlights:!0,rename:!0,colors:!0,foldingRanges:!0,selectionRanges:!0,diagnostics:s===by,documentFormattingEdits:s===by,documentRangeFormattingEdits:s===by}}const by="html",ZY="handlebars",XY="razor",sae=c2(by,a2,l2(by)),V7e=sae.defaults,oae=c2(ZY,a2,l2(ZY)),z7e=oae.defaults,rae=c2(XY,a2,l2(XY)),$7e=rae.defaults;function U7e(){return Se(()=>Promise.resolve().then(()=>Sze),void 0)}function c2(s,e=a2,t=l2(s)){const i=new W7e(s,e,t);let n;const o=H.onLanguage(s,async()=>{n=(await U7e()).setupMode(i)});return{defaults:i,dispose(){o.dispose(),n?.dispose(),n=void 0}}}const aae=Object.freeze(Object.defineProperty({__proto__:null,handlebarDefaults:z7e,handlebarLanguageService:oae,htmlDefaults:V7e,htmlLanguageService:sae,razorDefaults:$7e,razorLanguageService:rae,registerHTMLLanguageService:c2},Symbol.toStringTag,{value:"Module"}));let q7e=class{constructor(e,t,i){this._onDidChange=new Sg,this._languageId=e,this.setDiagnosticsOptions(t),this.setModeConfiguration(i)}get onDidChange(){return this._onDidChange.event}get languageId(){return this._languageId}get modeConfiguration(){return this._modeConfiguration}get diagnosticsOptions(){return this._diagnosticsOptions}setDiagnosticsOptions(e){this._diagnosticsOptions=e||Object.create(null),this._onDidChange.fire(this)}setModeConfiguration(e){this._modeConfiguration=e||Object.create(null),this._onDidChange.fire(this)}};const j7e={validate:!0,allowComments:!0,schemas:[],enableSchemaRequest:!1,schemaRequest:"warning",schemaValidation:"warning",comments:"error",trailingCommas:"error"},G7e={documentFormattingEdits:!0,documentRangeFormattingEdits:!0,completionItems:!0,hovers:!0,documentSymbols:!0,tokens:!0,colors:!0,foldingRanges:!0,diagnostics:!0,selectionRanges:!0},lae=new q7e("json",j7e,G7e),K7e=()=>cae().then(s=>s.getWorker());function cae(){return Se(()=>Promise.resolve().then(()=>zze),void 0)}H.register({id:"json",extensions:[".json",".bowerrc",".jshintrc",".jscsrc",".eslintrc",".babelrc",".har"],aliases:["JSON","json"],mimetypes:["application/json"]});H.onLanguage("json",()=>{cae().then(s=>s.setupMode(lae))});const dae=Object.freeze(Object.defineProperty({__proto__:null,getWorker:K7e,jsonDefaults:lae},Symbol.toStringTag,{value:"Module"})),Y7e="5.9.3";var hae=(s=>(s[s.None=0]="None",s[s.CommonJS=1]="CommonJS",s[s.AMD=2]="AMD",s[s.UMD=3]="UMD",s[s.System=4]="System",s[s.ES2015=5]="ES2015",s[s.ESNext=99]="ESNext",s))(hae||{}),uae=(s=>(s[s.None=0]="None",s[s.Preserve=1]="Preserve",s[s.React=2]="React",s[s.ReactNative=3]="ReactNative",s[s.ReactJSX=4]="ReactJSX",s[s.ReactJSXDev=5]="ReactJSXDev",s))(uae||{}),gae=(s=>(s[s.CarriageReturnLineFeed=0]="CarriageReturnLineFeed",s[s.LineFeed=1]="LineFeed",s))(gae||{}),fae=(s=>(s[s.ES3=0]="ES3",s[s.ES5=1]="ES5",s[s.ES2015=2]="ES2015",s[s.ES2016=3]="ES2016",s[s.ES2017=4]="ES2017",s[s.ES2018=5]="ES2018",s[s.ES2019=6]="ES2019",s[s.ES2020=7]="ES2020",s[s.ESNext=99]="ESNext",s[s.JSON=100]="JSON",s[s.Latest=99]="Latest",s))(fae||{}),pae=(s=>(s[s.Classic=1]="Classic",s[s.NodeJs=2]="NodeJs",s))(pae||{});class mae{constructor(e,t,i,n,o){this._onDidChange=new Sg,this._onDidExtraLibsChange=new Sg,this._extraLibs=Object.create(null),this._removedExtraLibs=Object.create(null),this._eagerModelSync=!1,this.setCompilerOptions(e),this.setDiagnosticsOptions(t),this.setWorkerOptions(i),this.setInlayHintsOptions(n),this.setModeConfiguration(o),this._onDidExtraLibsChangeTimeout=-1}get onDidChange(){return this._onDidChange.event}get onDidExtraLibsChange(){return this._onDidExtraLibsChange.event}get modeConfiguration(){return this._modeConfiguration}get workerOptions(){return this._workerOptions}get inlayHintsOptions(){return this._inlayHintsOptions}getExtraLibs(){return this._extraLibs}addExtraLib(e,t){let i;if(typeof t>"u"?i=`ts:extralib-${Math.random().toString(36).substring(2,15)}`:i=t,this._extraLibs[i]&&this._extraLibs[i].content===e)return{dispose:()=>{}};let n=1;return this._removedExtraLibs[i]&&(n=this._removedExtraLibs[i]+1),this._extraLibs[i]&&(n=this._extraLibs[i].version+1),this._extraLibs[i]={content:e,version:n},this._fireOnDidExtraLibsChangeSoon(),{dispose:()=>{let o=this._extraLibs[i];o&&o.version===n&&(delete this._extraLibs[i],this._removedExtraLibs[i]=n,this._fireOnDidExtraLibsChangeSoon())}}}setExtraLibs(e){for(const t in this._extraLibs)this._removedExtraLibs[t]=this._extraLibs[t].version;if(this._extraLibs=Object.create(null),e&&e.length>0)for(const t of e){const i=t.filePath||`ts:extralib-${Math.random().toString(36).substring(2,15)}`,n=t.content;let o=1;this._removedExtraLibs[i]&&(o=this._removedExtraLibs[i]+1),this._extraLibs[i]={content:n,version:o}}this._fireOnDidExtraLibsChangeSoon()}_fireOnDidExtraLibsChangeSoon(){this._onDidExtraLibsChangeTimeout===-1&&(this._onDidExtraLibsChangeTimeout=window.setTimeout(()=>{this._onDidExtraLibsChangeTimeout=-1,this._onDidExtraLibsChange.fire(void 0)},0))}getCompilerOptions(){return this._compilerOptions}setCompilerOptions(e){this._compilerOptions=e||Object.create(null),this._onDidChange.fire(void 0)}getDiagnosticsOptions(){return this._diagnosticsOptions}setDiagnosticsOptions(e){this._diagnosticsOptions=e||Object.create(null),this._onDidChange.fire(void 0)}setWorkerOptions(e){this._workerOptions=e||Object.create(null),this._onDidChange.fire(void 0)}setInlayHintsOptions(e){this._inlayHintsOptions=e||Object.create(null),this._onDidChange.fire(void 0)}setMaximumWorkerIdleTime(e){}setEagerModelSync(e){this._eagerModelSync=e}getEagerModelSync(){return this._eagerModelSync}setModeConfiguration(e){this._modeConfiguration=e||Object.create(null),this._onDidChange.fire(void 0)}}const Z7e=Y7e,_ae={completionItems:!0,hovers:!0,documentSymbols:!0,definitions:!0,references:!0,documentHighlights:!0,rename:!0,diagnostics:!0,documentRangeFormattingEdits:!0,signatureHelp:!0,onTypeFormattingEdits:!0,codeActions:!0,inlayHints:!0},TV=new mae({allowNonTsExtensions:!0,target:99},{noSemanticValidation:!1,noSyntaxValidation:!1,onlyVisible:!1},{},{},_ae),bae=new mae({allowNonTsExtensions:!0,allowJs:!0,target:99},{noSemanticValidation:!0,noSyntaxValidation:!1,onlyVisible:!1},{},{},_ae),X7e=()=>d2().then(s=>s.getTypeScriptWorker()),Q7e=()=>d2().then(s=>s.getJavaScriptWorker());function d2(){return Se(()=>Promise.resolve().then(()=>Kze),void 0)}H.onLanguage("typescript",()=>d2().then(s=>s.setupTypeScript(TV)));H.onLanguage("javascript",()=>d2().then(s=>s.setupJavaScript(bae)));const wae=Object.freeze(Object.defineProperty({__proto__:null,JsxEmit:uae,ModuleKind:hae,ModuleResolutionKind:pae,NewLineKind:gae,ScriptTarget:fae,getJavaScriptWorker:Q7e,getTypeScriptWorker:X7e,javascriptDefaults:bae,typescriptDefaults:TV,typescriptVersion:Z7e},Symbol.toStringTag,{value:"Module"})),vae={},GO={};class NV{static getOrCreate(e){return GO[e]||(GO[e]=new NV(e)),GO[e]}constructor(e){this._languageId=e,this._loadingTriggered=!1,this._lazyLoadPromise=new Promise((t,i)=>{this._lazyLoadPromiseResolve=t,this._lazyLoadPromiseReject=i})}load(){return this._loadingTriggered||(this._loadingTriggered=!0,vae[this._languageId].loader().then(e=>this._lazyLoadPromiseResolve(e),e=>this._lazyLoadPromiseReject(e))),this._lazyLoadPromise}}function Ie(s){const e=s.id;vae[e]=s,H.register(s);const t=NV.getOrCreate(e);H.registerTokensProviderFactory(e,{create:async()=>(await t.load()).language}),H.onLanguageEncountered(e,async()=>{const i=await t.load();H.setLanguageConfiguration(e,i.conf)})}Ie({id:"abap",extensions:[".abap"],aliases:["abap","ABAP"],loader:()=>Se(()=>Promise.resolve().then(()=>Xze),void 0)});Ie({id:"apex",extensions:[".cls"],aliases:["Apex","apex"],mimetypes:["text/x-apex-source","text/x-apex"],loader:()=>Se(()=>Promise.resolve().then(()=>i$e),void 0)});Ie({id:"azcli",extensions:[".azcli"],aliases:["Azure CLI","azcli"],loader:()=>Se(()=>Promise.resolve().then(()=>o$e),void 0)});Ie({id:"bat",extensions:[".bat",".cmd"],aliases:["Batch","bat"],loader:()=>Se(()=>Promise.resolve().then(()=>l$e),void 0)});Ie({id:"bicep",extensions:[".bicep"],aliases:["Bicep"],loader:()=>Se(()=>Promise.resolve().then(()=>w$e),void 0)});Ie({id:"cameligo",extensions:[".mligo"],aliases:["Cameligo"],loader:()=>Se(()=>Promise.resolve().then(()=>y$e),void 0)});Ie({id:"clojure",extensions:[".clj",".cljs",".cljc",".edn"],aliases:["clojure","Clojure"],loader:()=>Se(()=>Promise.resolve().then(()=>L$e),void 0)});Ie({id:"coffeescript",extensions:[".coffee"],aliases:["CoffeeScript","coffeescript","coffee"],mimetypes:["text/x-coffeescript","text/coffeescript"],loader:()=>Se(()=>Promise.resolve().then(()=>I$e),void 0)});Ie({id:"c",extensions:[".c",".h"],aliases:["C","c"],loader:()=>Se(()=>Promise.resolve().then(()=>ple),void 0)});Ie({id:"cpp",extensions:[".cpp",".cc",".cxx",".hpp",".hh",".hxx"],aliases:["C++","Cpp","cpp"],loader:()=>Se(()=>Promise.resolve().then(()=>ple),void 0)});Ie({id:"csharp",extensions:[".cs",".csx",".cake"],aliases:["C#","csharp"],loader:()=>Se(()=>Promise.resolve().then(()=>A$e),void 0)});Ie({id:"csp",extensions:[".csp"],aliases:["CSP","csp"],loader:()=>Se(()=>Promise.resolve().then(()=>P$e),void 0)});Ie({id:"css",extensions:[".css"],aliases:["CSS","css"],mimetypes:["text/css"],loader:()=>Se(()=>Promise.resolve().then(()=>W$e),void 0)});Ie({id:"cypher",extensions:[".cypher",".cyp"],aliases:["Cypher","OpenCypher"],loader:()=>Se(()=>Promise.resolve().then(()=>z$e),void 0)});Ie({id:"dart",extensions:[".dart"],aliases:["Dart","dart"],mimetypes:["text/x-dart-source","text/x-dart"],loader:()=>Se(()=>Promise.resolve().then(()=>q$e),void 0)});Ie({id:"dockerfile",extensions:[".dockerfile"],filenames:["Dockerfile"],aliases:["Dockerfile"],loader:()=>Se(()=>Promise.resolve().then(()=>K$e),void 0)});Ie({id:"ecl",extensions:[".ecl"],aliases:["ECL","Ecl","ecl"],loader:()=>Se(()=>Promise.resolve().then(()=>X$e),void 0)});Ie({id:"elixir",extensions:[".ex",".exs"],aliases:["Elixir","elixir","ex"],loader:()=>Se(()=>Promise.resolve().then(()=>eUe),void 0)});Ie({id:"flow9",extensions:[".flow"],aliases:["Flow9","Flow","flow9","flow"],loader:()=>Se(()=>Promise.resolve().then(()=>nUe),void 0)});Ie({id:"fsharp",extensions:[".fs",".fsi",".ml",".mli",".fsx",".fsscript"],aliases:["F#","FSharp","fsharp"],loader:()=>Se(()=>Promise.resolve().then(()=>rUe),void 0)});Ie({id:"freemarker2",extensions:[".ftl",".ftlh",".ftlx"],aliases:["FreeMarker2","Apache FreeMarker2"],loader:()=>Se(()=>Promise.resolve().then(()=>D_),void 0).then(s=>s.TagAutoInterpolationDollar)});Ie({id:"freemarker2.tag-angle.interpolation-dollar",aliases:["FreeMarker2 (Angle/Dollar)","Apache FreeMarker2 (Angle/Dollar)"],loader:()=>Se(()=>Promise.resolve().then(()=>D_),void 0).then(s=>s.TagAngleInterpolationDollar)});Ie({id:"freemarker2.tag-bracket.interpolation-dollar",aliases:["FreeMarker2 (Bracket/Dollar)","Apache FreeMarker2 (Bracket/Dollar)"],loader:()=>Se(()=>Promise.resolve().then(()=>D_),void 0).then(s=>s.TagBracketInterpolationDollar)});Ie({id:"freemarker2.tag-angle.interpolation-bracket",aliases:["FreeMarker2 (Angle/Bracket)","Apache FreeMarker2 (Angle/Bracket)"],loader:()=>Se(()=>Promise.resolve().then(()=>D_),void 0).then(s=>s.TagAngleInterpolationBracket)});Ie({id:"freemarker2.tag-bracket.interpolation-bracket",aliases:["FreeMarker2 (Bracket/Bracket)","Apache FreeMarker2 (Bracket/Bracket)"],loader:()=>Se(()=>Promise.resolve().then(()=>D_),void 0).then(s=>s.TagBracketInterpolationBracket)});Ie({id:"freemarker2.tag-auto.interpolation-dollar",aliases:["FreeMarker2 (Auto/Dollar)","Apache FreeMarker2 (Auto/Dollar)"],loader:()=>Se(()=>Promise.resolve().then(()=>D_),void 0).then(s=>s.TagAutoInterpolationDollar)});Ie({id:"freemarker2.tag-auto.interpolation-bracket",aliases:["FreeMarker2 (Auto/Bracket)","Apache FreeMarker2 (Auto/Bracket)"],loader:()=>Se(()=>Promise.resolve().then(()=>D_),void 0).then(s=>s.TagAutoInterpolationBracket)});Ie({id:"go",extensions:[".go"],aliases:["Go"],loader:()=>Se(()=>Promise.resolve().then(()=>mUe),void 0)});Ie({id:"graphql",extensions:[".graphql",".gql"],aliases:["GraphQL","graphql","gql"],mimetypes:["application/graphql"],loader:()=>Se(()=>Promise.resolve().then(()=>wUe),void 0)});Ie({id:"handlebars",extensions:[".handlebars",".hbs"],aliases:["Handlebars","handlebars","hbs"],mimetypes:["text/x-handlebars-template"],loader:()=>Se(()=>Promise.resolve().then(()=>yUe),void 0)});Ie({id:"hcl",extensions:[".tf",".tfvars",".hcl"],aliases:["Terraform","tf","HCL","hcl"],loader:()=>Se(()=>Promise.resolve().then(()=>LUe),void 0)});Ie({id:"html",extensions:[".html",".htm",".shtml",".xhtml",".mdoc",".jsp",".asp",".aspx",".jshtm"],aliases:["HTML","htm","html","xhtml"],mimetypes:["text/html","text/x-jshtm","text/template","text/ng-template"],loader:()=>Se(()=>Promise.resolve().then(()=>IUe),void 0)});Ie({id:"ini",extensions:[".ini",".properties",".gitconfig"],filenames:["config",".gitattributes",".gitconfig",".editorconfig"],aliases:["Ini","ini"],loader:()=>Se(()=>Promise.resolve().then(()=>NUe),void 0)});Ie({id:"java",extensions:[".java",".jav"],aliases:["Java","java"],mimetypes:["text/x-java-source","text/x-java"],loader:()=>Se(()=>Promise.resolve().then(()=>MUe),void 0)});Ie({id:"javascript",extensions:[".js",".es6",".jsx",".mjs",".cjs"],firstLine:"^#!.*\\bnode",filenames:["jakefile"],aliases:["JavaScript","javascript","js"],mimetypes:["text/javascript"],loader:()=>Se(()=>Promise.resolve().then(()=>BUe),void 0)});Ie({id:"julia",extensions:[".jl"],aliases:["julia","Julia"],loader:()=>Se(()=>Promise.resolve().then(()=>VUe),void 0)});Ie({id:"kotlin",extensions:[".kt",".kts"],aliases:["Kotlin","kotlin"],mimetypes:["text/x-kotlin-source","text/x-kotlin"],loader:()=>Se(()=>Promise.resolve().then(()=>UUe),void 0)});Ie({id:"less",extensions:[".less"],aliases:["Less","less"],mimetypes:["text/x-less","text/less"],loader:()=>Se(()=>Promise.resolve().then(()=>GUe),void 0)});Ie({id:"lexon",extensions:[".lex"],aliases:["Lexon"],loader:()=>Se(()=>Promise.resolve().then(()=>ZUe),void 0)});Ie({id:"lua",extensions:[".lua"],aliases:["Lua","lua"],loader:()=>Se(()=>Promise.resolve().then(()=>JUe),void 0)});Ie({id:"liquid",extensions:[".liquid",".html.liquid"],aliases:["Liquid","liquid"],mimetypes:["application/liquid"],loader:()=>Se(()=>Promise.resolve().then(()=>iqe),void 0)});Ie({id:"m3",extensions:[".m3",".i3",".mg",".ig"],aliases:["Modula-3","Modula3","modula3","m3"],loader:()=>Se(()=>Promise.resolve().then(()=>oqe),void 0)});Ie({id:"markdown",extensions:[".md",".markdown",".mdown",".mkdn",".mkd",".mdwn",".mdtxt",".mdtext"],aliases:["Markdown","markdown"],loader:()=>Se(()=>Promise.resolve().then(()=>lqe),void 0)});Ie({id:"mdx",extensions:[".mdx"],aliases:["MDX","mdx"],loader:()=>Se(()=>Promise.resolve().then(()=>hqe),void 0)});Ie({id:"mips",extensions:[".s"],aliases:["MIPS","MIPS-V"],mimetypes:["text/x-mips","text/mips","text/plaintext"],loader:()=>Se(()=>Promise.resolve().then(()=>fqe),void 0)});Ie({id:"msdax",extensions:[".dax",".msdax"],aliases:["DAX","MSDAX"],loader:()=>Se(()=>Promise.resolve().then(()=>_qe),void 0)});Ie({id:"mysql",extensions:[],aliases:["MySQL","mysql"],loader:()=>Se(()=>Promise.resolve().then(()=>vqe),void 0)});Ie({id:"objective-c",extensions:[".m"],aliases:["Objective-C"],loader:()=>Se(()=>Promise.resolve().then(()=>Sqe),void 0)});Ie({id:"pascal",extensions:[".pas",".p",".pp"],aliases:["Pascal","pas"],mimetypes:["text/x-pascal-source","text/x-pascal"],loader:()=>Se(()=>Promise.resolve().then(()=>kqe),void 0)});Ie({id:"pascaligo",extensions:[".ligo"],aliases:["Pascaligo","ligo"],loader:()=>Se(()=>Promise.resolve().then(()=>Dqe),void 0)});Ie({id:"perl",extensions:[".pl",".pm"],aliases:["Perl","pl"],loader:()=>Se(()=>Promise.resolve().then(()=>Rqe),void 0)});Ie({id:"pgsql",extensions:[],aliases:["PostgreSQL","postgres","pg","postgre"],loader:()=>Se(()=>Promise.resolve().then(()=>Oqe),void 0)});Ie({id:"php",extensions:[".php",".php4",".php5",".phtml",".ctp"],aliases:["PHP","php"],mimetypes:["application/x-php"],loader:()=>Se(()=>Promise.resolve().then(()=>Bqe),void 0)});Ie({id:"pla",extensions:[".pla"],loader:()=>Se(()=>Promise.resolve().then(()=>Vqe),void 0)});Ie({id:"postiats",extensions:[".dats",".sats",".hats"],aliases:["ATS","ATS/Postiats"],loader:()=>Se(()=>Promise.resolve().then(()=>Uqe),void 0)});Ie({id:"powerquery",extensions:[".pq",".pqm"],aliases:["PQ","M","Power Query","Power Query M"],loader:()=>Se(()=>Promise.resolve().then(()=>Gqe),void 0)});Ie({id:"powershell",extensions:[".ps1",".psm1",".psd1"],aliases:["PowerShell","powershell","ps","ps1"],loader:()=>Se(()=>Promise.resolve().then(()=>Zqe),void 0)});Ie({id:"proto",extensions:[".proto"],aliases:["protobuf","Protocol Buffers"],loader:()=>Se(()=>Promise.resolve().then(()=>eje),void 0)});Ie({id:"pug",extensions:[".jade",".pug"],aliases:["Pug","Jade","jade"],loader:()=>Se(()=>Promise.resolve().then(()=>nje),void 0)});Ie({id:"python",extensions:[".py",".rpy",".pyw",".cpy",".gyp",".gypi"],aliases:["Python","py"],firstLine:"^#!/.*\\bpython[0-9.-]*\\b",loader:()=>Se(()=>Promise.resolve().then(()=>rje),void 0)});Ie({id:"qsharp",extensions:[".qs"],aliases:["Q#","qsharp"],loader:()=>Se(()=>Promise.resolve().then(()=>cje),void 0)});Ie({id:"r",extensions:[".r",".rhistory",".rmd",".rprofile",".rt"],aliases:["R","r"],loader:()=>Se(()=>Promise.resolve().then(()=>uje),void 0)});Ie({id:"razor",extensions:[".cshtml"],aliases:["Razor","razor"],mimetypes:["text/x-cshtml"],loader:()=>Se(()=>Promise.resolve().then(()=>pje),void 0)});Ie({id:"redis",extensions:[".redis"],aliases:["redis"],loader:()=>Se(()=>Promise.resolve().then(()=>bje),void 0)});Ie({id:"redshift",extensions:[],aliases:["Redshift","redshift"],loader:()=>Se(()=>Promise.resolve().then(()=>Cje),void 0)});Ie({id:"restructuredtext",extensions:[".rst"],aliases:["reStructuredText","restructuredtext"],loader:()=>Se(()=>Promise.resolve().then(()=>xje),void 0)});Ie({id:"ruby",extensions:[".rb",".rbx",".rjs",".gemspec",".pp"],filenames:["rakefile","Gemfile"],aliases:["Ruby","rb"],loader:()=>Se(()=>Promise.resolve().then(()=>Eje),void 0)});Ie({id:"rust",extensions:[".rs",".rlib"],aliases:["Rust","rust"],loader:()=>Se(()=>Promise.resolve().then(()=>Tje),void 0)});Ie({id:"sb",extensions:[".sb"],aliases:["Small Basic","sb"],loader:()=>Se(()=>Promise.resolve().then(()=>Aje),void 0)});Ie({id:"scala",extensions:[".scala",".sc",".sbt"],aliases:["Scala","scala","SBT","Sbt","sbt","Dotty","dotty"],mimetypes:["text/x-scala-source","text/x-scala","text/x-sbt","text/x-dotty"],loader:()=>Se(()=>Promise.resolve().then(()=>Pje),void 0)});Ie({id:"scheme",extensions:[".scm",".ss",".sch",".rkt"],aliases:["scheme","Scheme"],loader:()=>Se(()=>Promise.resolve().then(()=>Wje),void 0)});Ie({id:"scss",extensions:[".scss"],aliases:["Sass","sass","scss"],mimetypes:["text/x-scss","text/scss"],loader:()=>Se(()=>Promise.resolve().then(()=>zje),void 0)});Ie({id:"shell",extensions:[".sh",".bash"],aliases:["Shell","sh"],loader:()=>Se(()=>Promise.resolve().then(()=>qje),void 0)});Ie({id:"sol",extensions:[".sol"],aliases:["sol","solidity","Solidity"],loader:()=>Se(()=>Promise.resolve().then(()=>Kje),void 0)});Ie({id:"aes",extensions:[".aes"],aliases:["aes","sophia","Sophia"],loader:()=>Se(()=>Promise.resolve().then(()=>Xje),void 0)});Ie({id:"sparql",extensions:[".rq"],aliases:["sparql","SPARQL"],loader:()=>Se(()=>Promise.resolve().then(()=>eGe),void 0)});Ie({id:"sql",extensions:[".sql"],aliases:["SQL"],loader:()=>Se(()=>Promise.resolve().then(()=>nGe),void 0)});Ie({id:"st",extensions:[".st",".iecst",".iecplc",".lc3lib",".TcPOU",".TcDUT",".TcGVL",".TcIO"],aliases:["StructuredText","scl","stl"],loader:()=>Se(()=>Promise.resolve().then(()=>rGe),void 0)});Ie({id:"swift",aliases:["Swift","swift"],extensions:[".swift"],mimetypes:["text/swift"],loader:()=>Se(()=>Promise.resolve().then(()=>cGe),void 0)});Ie({id:"systemverilog",extensions:[".sv",".svh"],aliases:["SV","sv","SystemVerilog","systemverilog"],loader:()=>Se(()=>Promise.resolve().then(()=>wle),void 0)});Ie({id:"verilog",extensions:[".v",".vh"],aliases:["V","v","Verilog","verilog"],loader:()=>Se(()=>Promise.resolve().then(()=>wle),void 0)});Ie({id:"tcl",extensions:[".tcl"],aliases:["tcl","Tcl","tcltk","TclTk","tcl/tk","Tcl/Tk"],loader:()=>Se(()=>Promise.resolve().then(()=>fGe),void 0)});Ie({id:"twig",extensions:[".twig"],aliases:["Twig","twig"],mimetypes:["text/x-twig"],loader:()=>Se(()=>Promise.resolve().then(()=>_Ge),void 0)});Ie({id:"typescript",extensions:[".ts",".tsx",".cts",".mts"],aliases:["TypeScript","ts","typescript"],mimetypes:["text/typescript"],loader:()=>Se(()=>Promise.resolve().then(()=>OUe),void 0)});Ie({id:"typespec",extensions:[".tsp"],aliases:["TypeSpec"],loader:()=>Se(()=>Promise.resolve().then(()=>EGe),void 0)});Ie({id:"vb",extensions:[".vb"],aliases:["Visual Basic","vb"],loader:()=>Se(()=>Promise.resolve().then(()=>TGe),void 0)});Ie({id:"wgsl",extensions:[".wgsl"],aliases:["WebGPU Shading Language","WGSL","wgsl"],loader:()=>Se(()=>Promise.resolve().then(()=>$Ge),void 0)});Ie({id:"xml",extensions:[".xml",".xsd",".dtd",".ascx",".csproj",".config",".props",".targets",".wxi",".wxl",".wxs",".xaml",".svg",".svgz",".opf",".xslt",".xsl"],firstLine:"(\\<\\?xml.*)|(\\<svg)|(\\<\\!doctype\\s+svg)",aliases:["XML","xml"],mimetypes:["text/xml","application/xml","application/xaml+xml","application/xml-dtd"],loader:()=>Se(()=>Promise.resolve().then(()=>jGe),void 0)});Ie({id:"yaml",extensions:[".yaml",".yml"],aliases:["YAML","yaml","YML","yml"],mimetypes:["application/x-yaml","text/x-yaml"],loader:()=>Se(()=>Promise.resolve().then(()=>YGe),void 0)});var J7e=Object.defineProperty,eHe=(s,e,t)=>e in s?J7e(s,e,{enumerable:!0,configurable:!0,writable:!0,value:t}):s[e]=t,ye=(s,e,t)=>eHe(s,typeof e!="symbol"?e+"":e,t),gk,QY,fk,KO,pk;function tHe(s){return s.method!==void 0}var JY;(function(s){function e(t){return t}s.create=e})(JY||(JY={}));var Da;(function(s){s.parseError=-32700,s.invalidRequest=-32600,s.methodNotFound=-32601,s.invalidParams=-32602,s.internalError=-32603;function e(o){return-32099<=o&&o<=-32e3}s.isServerError=e;function t(o){if(!e(o))throw new Error("Invalid range for a server error.");return o}s.serverError=t,s.unexpectedServerError=-32e3;function i(o){return!0}s.isApplicationError=i;function n(o){return o}s.applicationError=n,s.genericApplicationError=-320100})(Da||(Da={}));var RV=class{constructor(){ye(this,"listeners",new Set),ye(this,"event",s=>(this.listeners.add(s),{dispose:()=>{this.listeners.delete(s)}}))}fire(s){this.listeners.forEach(e=>e(s))}},iHe=class{constructor(s){ye(this,"_value"),ye(this,"eventEmitter"),this._value=s,this.eventEmitter=new RV}get value(){return this._value}set value(s){this._value!==s&&(this._value=s,this.eventEmitter.fire(s))}get onChange(){return this.eventEmitter.event}};function nHe(s,e){const t=setTimeout(e,s);return{dispose:()=>clearTimeout(t)}}function YO(s,e,t){return s instanceof Set?(s.add(e),{dispose:()=>s.delete(e)}):(s.set(e,t),{dispose:()=>s.delete(e)})}var sHe=class{constructor(){ye(this,"_state","none"),ye(this,"promise"),ye(this,"resolve",()=>{}),ye(this,"reject",()=>{}),this.promise=new Promise((s,e)=>{this.resolve=s,this.reject=e})}get state(){return this._state}},Cae=(gk=class{constructor(){ye(this,"_unprocessedMessages",[]),ye(this,"_messageListener"),ye(this,"id",gk.id++),ye(this,"_state",new iHe({state:"open"})),ye(this,"state",this._state)}setListener(s){if(this._messageListener=s,!!s)for(;this._unprocessedMessages.length>0&&this._messageListener!==void 0;){const e=this._unprocessedMessages.shift();this._messageListener(e)}}send(s){return this._sendImpl(s)}_dispatchReceivedMessage(s){this._unprocessedMessages.length===0&&this._messageListener?this._messageListener(s):this._unprocessedMessages.push(s)}_onConnectionClosed(){this._state.value={state:"closed",error:void 0}}log(s){return new oHe(this,s??new rHe)}},ye(gk,"id",0),gk),oHe=class{constructor(s,e){ye(this,"baseStream"),ye(this,"logger"),this.baseStream=s,this.logger=e}get state(){return this.baseStream.state}setListener(s){if(s===void 0){this.baseStream.setListener(void 0);return}this.baseStream.setListener(e=>{this.logger.log(this.baseStream,"incoming",e),s(e)})}send(s){return this.logger.log(this.baseStream,"outgoing",s),this.baseStream.send(s)}toString(){return`StreamLogger/${this.baseStream.toString()}`}},rHe=class{log(s,e,t){console.log(`${e==="incoming"?"<-":"->"} [${s.toString()}] ${JSON.stringify(t)}`)}},aHe=class yae{constructor(e){ye(this,"connect"),this.connect=e}mapContext(e){return new yae(t=>this.connect(t?lHe(t,e):void 0))}};function lHe(s,e){return{handleNotification:(t,i)=>s.handleNotification(t,e(i)),handleRequest:(t,i,n)=>s.handleRequest(t,i,e(n))}}var cHe=class Sae{constructor(e,t,i){ye(this,"_stream"),ye(this,"_listener"),ye(this,"_logger"),ye(this,"_unprocessedResponses",new Map),ye(this,"_lastUsedRequestId",0),this._stream=e,this._listener=t,this._logger=i,this._stream.setListener(n=>{tHe(n)?n.id===void 0?this._processNotification(n):this._processRequest(n):this._processResponse(n)})}static createChannel(e,t){let i=!1;return new aHe(n=>{if(i)throw new Error(`A channel to the stream ${e} was already constructed!`);return i=!0,new Sae(e,n,t)})}get state(){return this._stream.state}async _processNotification(e){if(e.id!==void 0)throw new Error;if(!this._listener){this._logger&&this._logger.debug({text:"Notification ignored",message:e});return}try{await this._listener.handleNotification({method:e.method,params:e.params||null})}catch(t){this._logger&&this._logger.warn({text:`Exception was thrown while handling notification: ${t}`,exception:t,message:e})}}async _processRequest(e){if(e.id===void 0)throw new Error;let t;if(this._listener)try{t=await this._listener.handleRequest({method:e.method,params:e.params||null},e.id)}catch(n){this._logger&&this._logger.warn({text:`Exception was thrown while handling request: ${n}`,message:e,exception:n}),t={error:{code:Da.internalError,message:"An unexpected exception was thrown.",data:void 0}}}else this._logger&&this._logger.debug({text:"Received request even though not listening for requests",message:e}),t={error:{code:Da.methodNotFound,message:"This endpoint does not listen for requests or notifications.",data:void 0}};let i;"result"in t?i={jsonrpc:"2.0",id:e.id,result:t.result}:i={jsonrpc:"2.0",id:e.id,error:t.error},await this._stream.send(i)}_processResponse(e){const t=""+e.id,i=this._unprocessedResponses.get(t);if(!i){this._logger&&this._logger.debug({text:"Got an unexpected response message",message:e});return}this._unprocessedResponses.delete(t),i(e)}_newRequestId(){return this._lastUsedRequestId++}sendRequest(e,t,i){const n={jsonrpc:"2.0",id:this._newRequestId(),method:e.method,params:e.params||void 0};return i&&i(n.id),new Promise((o,r)=>{const a=""+n.id;this._unprocessedResponses.set(a,l=>{"result"in l?o({result:l.result}):(l.error||r(new Error("Response had neither 'result' nor 'error' field set.")),o({error:l.error}))}),this._stream.send(n).then(void 0,l=>{this._unprocessedResponses.delete(a),r(l)})})}sendNotification(e,t){const i={jsonrpc:"2.0",id:void 0,method:e.method,params:e.params||void 0};return this._stream.send(i)}toString(){return"StreamChannel/"+this._stream.toString()}},Ow;(function(s){function e(){return{deserializeFromJson:n=>({hasErrors:!1,value:n}),serializeToJson:n=>n}}s.sAny=e;function t(){return{deserializeFromJson:n=>({hasErrors:!1,value:{}}),serializeToJson:n=>({})}}s.sEmptyObject=t;function i(){return{deserializeFromJson:n=>({hasErrors:!1,value:void 0}),serializeToJson:n=>null}}s.sVoidFromNull=i})(Ow||(Ow={}));const xae=Symbol("OptionalMethodNotFound");var Lae=class{contextualize(s){return new dHe(this,s)}},dHe=class extends Lae{constructor(s,e){super(),ye(this,"underylingTypedChannel"),ye(this,"converters"),this.underylingTypedChannel=s,this.converters=e}async request(s,e,t){const i=await this.converters.getSendContext(t);return this.underylingTypedChannel.request(s,e,i)}async notify(s,e,t){const i=await this.converters.getSendContext(t);return this.underylingTypedChannel.notify(s,e,i)}registerNotificationHandler(s,e){return this.underylingTypedChannel.registerNotificationHandler(s,async(t,i)=>await e(t,await this.converters.getNewContext(i)))}registerRequestHandler(s,e){return this.underylingTypedChannel.registerRequestHandler(s,async(t,i,n)=>await e(t,i,await this.converters.getNewContext(n)))}},h6=class kae extends Lae{constructor(e,t={}){super(),ye(this,"channelCtor"),ye(this,"_requestSender"),ye(this,"_handler",new Map),ye(this,"_unknownNotificationHandler",new Set),ye(this,"_timeout"),ye(this,"sendExceptionDetails",!1),ye(this,"_logger"),ye(this,"listeningDeferred",new sHe),ye(this,"onListening",this.listeningDeferred.promise),ye(this,"_requestDidErrorEventEmitter",new RV),ye(this,"onRequestDidError",this._requestDidErrorEventEmitter.event),this.channelCtor=e,this._logger=t.logger,this.sendExceptionDetails=!!t.sendExceptionDetails,this._timeout=nHe(1e3,()=>{this._requestSender||console.warn(`"${this.startListen.name}" has not been called within 1 second after construction of this channel. Did you forget to call it?`,this)})}static fromTransport(e,t={}){return new kae(cHe.createChannel(e,t.logger),t)}startListen(){if(this._requestSender)throw new Error(`"${this.startListen.name}" can be called only once, but it already has been called.`);this._timeout&&(this._timeout.dispose(),this._timeout=void 0),this._requestSender=this.channelCtor.connect({handleRequest:(e,t,i)=>this.handleRequest(e,t,i),handleNotification:(e,t)=>this.handleNotification(e,t)}),this.listeningDeferred.resolve()}checkChannel(e){if(!e)throw new Error(`"${this.startListen.name}" must be called before any messages can be sent or received.`);return!0}async handleRequest(e,t,i){const n=this._handler.get(e.method);if(!n)return this._logger&&this._logger.debug({text:`No request handler for "${e.method}".`,data:{requestObject:e}}),{error:{code:Da.methodNotFound,message:`No request handler for "${e.method}".`,data:{method:e.method}}};if(n.kind!="request"){const r=`"${e.method}" is registered as notification, but was sent as request.`;return this._logger&&this._logger.debug({text:r,data:{requestObject:e}}),{error:{code:Da.invalidRequest,message:r,data:{method:e.method}}}}const o=n.requestType.paramsSerializer.deserializeFromJson(e.params);if(o.hasErrors){const r=`Got invalid params: ${o.errorMessage}`;return this._logger&&this._logger.debug({text:r,data:{requestObject:e,errorMessage:o.errorMessage}}),{error:{code:Da.invalidParams,message:r,data:{errors:o.errorMessage}}}}else{const r=o.value;let a;try{const l=await n.handler(r,t,i);if("error"in l||"errorMessage"in l){const c=l.error?n.requestType.errorSerializer.serializeToJson(l.error):void 0;a={error:{code:l.errorCode||Da.genericApplicationError,message:l.errorMessage||"An error was returned",data:c}}}else a={result:n.requestType.resultSerializer.serializeToJson(l.ok)}}catch(l){l instanceof tZ?a={error:{code:l.code,message:l.message}}:(this._logger&&this._logger.warn({text:`An exception was thrown while handling a request: ${l}.`,exception:l,data:{requestObject:e}}),a={error:{code:Da.unexpectedServerError,message:this.sendExceptionDetails?`An exception was thrown while handling a request: ${l}.`:"Server has thrown an unexpected exception"}})}return a}}async handleNotification(e,t){const i=this._handler.get(e.method);if(!i){for(const r of this._unknownNotificationHandler)r(e);this._unknownNotificationHandler.size===0&&this._logger&&this._logger.debug({text:`Unhandled notification "${e.method}"`,data:{requestObject:e}});return}if(i.kind!="notification"){this._logger&&this._logger.debug({text:`"${e.method}" is registered as request, but was sent as notification.`,data:{requestObject:e}});return}const n=i.notificationType.paramsSerializer.deserializeFromJson(e.params);if(n.hasErrors){this._logger&&this._logger.debug({text:`Got invalid params: ${n}`,data:{requestObject:e,errorMessage:n.errorMessage}});return}const o=n.value;for(const r of i.handlers)try{r(o,t)}catch(a){this._logger&&this._logger.warn({text:`An exception was thrown while handling a notification: ${a}.`,exception:a,data:{requestObject:e}})}}registerUnknownNotificationHandler(e){return YO(this._unknownNotificationHandler,e)}registerRequestHandler(e,t){if(this._handler.get(e.method))throw new Error(`Handler with method "${e.method}" already registered.`);return YO(this._handler,e.method,{kind:"request",requestType:e,handler:t})}registerNotificationHandler(e,t){let i=this._handler.get(e.method);if(!i)i={kind:"notification",notificationType:e,handlers:new Set},this._handler.set(e.method,i);else{if(i.kind!=="notification")throw new Error(`Method "${e.method}" was already registered as request handler.`);if(i.notificationType!==e)throw new Error(`Method "${e.method}" was registered for a different type.`)}return YO(i.handlers,t)}getRegisteredTypes(){const e=[];for(const t of this._handler.values())t.kind==="notification"?e.push(t.notificationType):t.kind==="request"&&e.push(t.requestType);return e}async request(e,t,i){if(!this.checkChannel(this._requestSender))throw new Error("Impossible");const n=e.paramsSerializer.serializeToJson(t);eZ(n);const o=await this._requestSender.sendRequest({method:e.method,params:n},i);if("error"in o){if(e.isOptional&&o.error.code===Da.methodNotFound)return xae;let r;if(o.error.data!==void 0){const l=e.errorSerializer.deserializeFromJson(o.error.data);if(l.hasErrors)throw new Error(l.errorMessage);r=l.value}else r=void 0;const a=new tZ(o.error.message,r,o.error.code);throw this._requestDidErrorEventEmitter.fire({error:a}),a}else{const r=e.resultSerializer.deserializeFromJson(o.result);if(r.hasErrors)throw new Error("Could not deserialize response: "+r.errorMessage+`

${JSON.stringify(o,null,2)}`);return r.value}}async notify(e,t,i){if(!this.checkChannel(this._requestSender))throw new Error;const n=e.paramsSerializer.serializeToJson(t);eZ(n),this._requestSender.sendNotification({method:e.method,params:n},i)}};function eZ(s){if(s!==null&&Array.isArray(s)&&typeof s!="object")throw new Error("Invalid value! Only null, array and object is allowed.")}var tZ=class Eae extends Error{constructor(e,t,i=Da.genericApplicationError){super(e),ye(this,"data"),ye(this,"code"),this.data=t,this.code=i,Object.setPrototypeOf(this,Eae.prototype)}},hHe=class u6{constructor(e,t,i,n,o=!1){ye(this,"method"),ye(this,"paramsSerializer"),ye(this,"resultSerializer"),ye(this,"errorSerializer"),ye(this,"isOptional"),ye(this,"kind","request"),this.method=e,this.paramsSerializer=t,this.resultSerializer=i,this.errorSerializer=n,this.isOptional=o}withMethod(e){return new u6(e,this.paramsSerializer,this.resultSerializer,this.errorSerializer)}optional(){return new u6(this.method,this.paramsSerializer,this.resultSerializer,this.errorSerializer,!0)}},uHe=class Iae{constructor(e,t){ye(this,"method"),ye(this,"paramsSerializer"),ye(this,"kind","notification"),this.method=e,this.paramsSerializer=t}withMethod(e){return new Iae(e,this.paramsSerializer)}};function ze(s){return new hHe((s||{}).method,Ow.sAny(),Ow.sAny(),Ow.sAny())}function Mi(s){return new uHe((s||{}).method,Ow.sAny())}const gHe=Symbol();var iZ=(QY=gHe,fk=class{constructor(s){ye(this,"error"),ye(this,QY),this.error=s}},ye(fk,"factory",s=>new fk(s)),fk);function fHe(s){const e=nZ(s.server),t=nZ(s.client);return new pHe(s.tags||[],e,t)}function nZ(s){const e={};for(const[t,i]of Object.entries(s)){const n=i.method?i.method:t;e[t]=i.withMethod(n)}return e}var pHe=class Dae{constructor(e=[],t,i){ye(this,"tags"),ye(this,"server"),ye(this,"client"),this.tags=e,this.server=t,this.client=i}_onlyDesignTime(){return new Error("This property is not meant to be accessed at runtime")}get TContractObject(){throw this._onlyDesignTime()}get TClientInterface(){throw this._onlyDesignTime()}get TServerInterface(){throw this._onlyDesignTime()}get TClientHandler(){throw this._onlyDesignTime()}get TServerHandler(){throw this._onlyDesignTime()}get TTags(){throw this._onlyDesignTime()}getInterface(e,t,i,n){const o=this.buildCounterpart(e,i),r=this.registerHandlers(e,t,n,o);return{counterpart:o,dispose:()=>r.dispose()}}buildCounterpart(e,t){const i={};for(const[n,o]of Object.entries(t)){let r;o.kind==="request"?o.isOptional?r=async(a,l)=>{a===void 0&&(a={});try{return await e.request(o,a,l)}catch(c){if(c&&c.code===Da.methodNotFound)return xae;throw c}}:r=(a,l)=>(a===void 0&&(a={}),e.request(o,a,l)):r=(a,l)=>(a===void 0&&(a={}),e.notify(o,a,l)),i[n]=r}return i}registerHandlers(e,t,i,n){const o=[];for(const[r,a]of Object.entries(t))if(a.kind==="request"){let l=i[r];if(!l)continue;const c=this.createRequestHandler(n,l);o.push(e.registerRequestHandler(a,c))}else{const l=i[r];l&&o.push(e.registerNotificationHandler(a,(c,d)=>{l(c,{context:d,counterpart:n})}))}return{dispose:()=>o.forEach(r=>r.dispose())}}createRequestHandler(e,t){return async(i,n,o)=>{const r=await t(i,{context:o,counterpart:e,newErr:iZ.factory,requestId:n});return r instanceof iZ?r.error:{ok:r}}}static getServerFromStream(e,t,i,n){const o=h6.fromTransport(t,i),{server:r}=e.getServer(o,n);return o.startListen(),{channel:o,server:r}}static registerServerToStream(e,t,i,n){const o=h6.fromTransport(t,i),{client:r}=e.registerServer(o,n);return o.startListen(),{channel:o,client:r}}getServer(e,t){const{counterpart:i,dispose:n}=this.getInterface(e,this.client,this.server,t);return{server:i,dispose:n}}registerServer(e,t){const{counterpart:i,dispose:n}=this.getInterface(e,this.server,this.client,t);return{client:i,dispose:n}}withContext(){return new Dae(this.tags,this.server,this.client)}};let Pw=(function(s){return s.Comment="comment",s.Imports="imports",s.Region="region",s})({}),Ki=(function(s){return s[s.File=1]="File",s[s.Module=2]="Module",s[s.Namespace=3]="Namespace",s[s.Package=4]="Package",s[s.Class=5]="Class",s[s.Method=6]="Method",s[s.Property=7]="Property",s[s.Field=8]="Field",s[s.Constructor=9]="Constructor",s[s.Enum=10]="Enum",s[s.Interface=11]="Interface",s[s.Function=12]="Function",s[s.Variable=13]="Variable",s[s.Constant=14]="Constant",s[s.String=15]="String",s[s.Number=16]="Number",s[s.Boolean=17]="Boolean",s[s.Array=18]="Array",s[s.Object=19]="Object",s[s.Key=20]="Key",s[s.Null=21]="Null",s[s.EnumMember=22]="EnumMember",s[s.Struct=23]="Struct",s[s.Event=24]="Event",s[s.Operator=25]="Operator",s[s.TypeParameter=26]="TypeParameter",s})({}),Tae=(function(s){return s[s.Deprecated=1]="Deprecated",s})({}),sZ=(function(s){return s[s.Type=1]="Type",s[s.Parameter=2]="Parameter",s})({}),mHe=(function(s){return s[s.None=0]="None",s[s.Full=1]="Full",s[s.Incremental=2]="Incremental",s})({}),Ji=(function(s){return s[s.Text=1]="Text",s[s.Method=2]="Method",s[s.Function=3]="Function",s[s.Constructor=4]="Constructor",s[s.Field=5]="Field",s[s.Variable=6]="Variable",s[s.Class=7]="Class",s[s.Interface=8]="Interface",s[s.Module=9]="Module",s[s.Property=10]="Property",s[s.Unit=11]="Unit",s[s.Value=12]="Value",s[s.Enum=13]="Enum",s[s.Keyword=14]="Keyword",s[s.Snippet=15]="Snippet",s[s.Color=16]="Color",s[s.File=17]="File",s[s.Reference=18]="Reference",s[s.Folder=19]="Folder",s[s.EnumMember=20]="EnumMember",s[s.Constant=21]="Constant",s[s.Struct=22]="Struct",s[s.Event=23]="Event",s[s.Operator=24]="Operator",s[s.TypeParameter=25]="TypeParameter",s})({}),_He=(function(s){return s[s.Deprecated=1]="Deprecated",s})({}),bHe=(function(s){return s[s.PlainText=1]="PlainText",s[s.Snippet=2]="Snippet",s})({}),ZO=(function(s){return s[s.Text=1]="Text",s[s.Read=2]="Read",s[s.Write=3]="Write",s})({}),hd=(function(s){return s.Empty="",s.QuickFix="quickfix",s.Refactor="refactor",s.RefactorExtract="refactor.extract",s.RefactorInline="refactor.inline",s.RefactorRewrite="refactor.rewrite",s.Source="source",s.SourceOrganizeImports="source.organizeImports",s.SourceFixAll="source.fixAll",s})({}),IT=(function(s){return s.PlainText="plaintext",s.Markdown="markdown",s})({}),Gd=(function(s){return s[s.Error=1]="Error",s[s.Warning=2]="Warning",s[s.Information=3]="Information",s[s.Hint=4]="Hint",s})({}),oZ=(function(s){return s[s.Unnecessary=1]="Unnecessary",s[s.Deprecated=2]="Deprecated",s})({}),oE=(function(s){return s[s.Invoked=1]="Invoked",s[s.TriggerCharacter=2]="TriggerCharacter",s[s.TriggerForIncompleteCompletions=3]="TriggerForIncompleteCompletions",s})({}),rE=(function(s){return s[s.Invoked=1]="Invoked",s[s.TriggerCharacter=2]="TriggerCharacter",s[s.ContentChange=3]="ContentChange",s})({}),g6=(function(s){return s[s.Invoked=1]="Invoked",s[s.Automatic=2]="Automatic",s})({}),wHe=(function(s){return s.Relative="relative",s})({});var pt=class{constructor(s){this.method=s}};const nt={textDocumentImplementation:new pt("textDocument/implementation"),textDocumentTypeDefinition:new pt("textDocument/typeDefinition"),textDocumentDocumentColor:new pt("textDocument/documentColor"),textDocumentColorPresentation:new pt("textDocument/colorPresentation"),textDocumentFoldingRange:new pt("textDocument/foldingRange"),textDocumentDeclaration:new pt("textDocument/declaration"),textDocumentSelectionRange:new pt("textDocument/selectionRange"),textDocumentPrepareCallHierarchy:new pt("textDocument/prepareCallHierarchy"),textDocumentSemanticTokensFull:new pt("textDocument/semanticTokens/full"),textDocumentSemanticTokensFullDelta:new pt("textDocument/semanticTokens/full/delta"),textDocumentLinkedEditingRange:new pt("textDocument/linkedEditingRange"),workspaceWillCreateFiles:new pt("workspace/willCreateFiles"),workspaceWillRenameFiles:new pt("workspace/willRenameFiles"),workspaceWillDeleteFiles:new pt("workspace/willDeleteFiles"),textDocumentMoniker:new pt("textDocument/moniker"),textDocumentPrepareTypeHierarchy:new pt("textDocument/prepareTypeHierarchy"),textDocumentInlineValue:new pt("textDocument/inlineValue"),textDocumentInlayHint:new pt("textDocument/inlayHint"),textDocumentDiagnostic:new pt("textDocument/diagnostic"),textDocumentInlineCompletion:new pt("textDocument/inlineCompletion"),textDocumentWillSaveWaitUntil:new pt("textDocument/willSaveWaitUntil"),textDocumentCompletion:new pt("textDocument/completion"),textDocumentHover:new pt("textDocument/hover"),textDocumentSignatureHelp:new pt("textDocument/signatureHelp"),textDocumentDefinition:new pt("textDocument/definition"),textDocumentReferences:new pt("textDocument/references"),textDocumentDocumentHighlight:new pt("textDocument/documentHighlight"),textDocumentDocumentSymbol:new pt("textDocument/documentSymbol"),textDocumentCodeAction:new pt("textDocument/codeAction"),workspaceSymbol:new pt("workspace/symbol"),textDocumentCodeLens:new pt("textDocument/codeLens"),textDocumentDocumentLink:new pt("textDocument/documentLink"),textDocumentFormatting:new pt("textDocument/formatting"),textDocumentRangeFormatting:new pt("textDocument/rangeFormatting"),textDocumentRangesFormatting:new pt("textDocument/rangesFormatting"),textDocumentOnTypeFormatting:new pt("textDocument/onTypeFormatting"),textDocumentRename:new pt("textDocument/rename"),workspaceExecuteCommand:new pt("workspace/executeCommand"),workspaceDidCreateFiles:new pt("workspace/didCreateFiles"),workspaceDidRenameFiles:new pt("workspace/didRenameFiles"),workspaceDidDeleteFiles:new pt("workspace/didDeleteFiles"),workspaceDidChangeConfiguration:new pt("workspace/didChangeConfiguration"),textDocumentDidOpen:new pt("textDocument/didOpen"),textDocumentDidChange:new pt("textDocument/didChange"),textDocumentDidClose:new pt("textDocument/didClose"),textDocumentDidSave:new pt("textDocument/didSave"),textDocumentWillSave:new pt("textDocument/willSave"),workspaceDidChangeWatchedFiles:new pt("workspace/didChangeWatchedFiles")},R0=fHe({server:{textDocumentImplementation:ze({method:"textDocument/implementation"}),textDocumentTypeDefinition:ze({method:"textDocument/typeDefinition"}),textDocumentDocumentColor:ze({method:"textDocument/documentColor"}),textDocumentColorPresentation:ze({method:"textDocument/colorPresentation"}),textDocumentFoldingRange:ze({method:"textDocument/foldingRange"}),textDocumentDeclaration:ze({method:"textDocument/declaration"}),textDocumentSelectionRange:ze({method:"textDocument/selectionRange"}),textDocumentPrepareCallHierarchy:ze({method:"textDocument/prepareCallHierarchy"}),callHierarchyIncomingCalls:ze({method:"callHierarchy/incomingCalls"}),callHierarchyOutgoingCalls:ze({method:"callHierarchy/outgoingCalls"}),textDocumentSemanticTokensFull:ze({method:"textDocument/semanticTokens/full"}),textDocumentSemanticTokensFullDelta:ze({method:"textDocument/semanticTokens/full/delta"}),textDocumentSemanticTokensRange:ze({method:"textDocument/semanticTokens/range"}),textDocumentLinkedEditingRange:ze({method:"textDocument/linkedEditingRange"}),workspaceWillCreateFiles:ze({method:"workspace/willCreateFiles"}),workspaceWillRenameFiles:ze({method:"workspace/willRenameFiles"}),workspaceWillDeleteFiles:ze({method:"workspace/willDeleteFiles"}),textDocumentMoniker:ze({method:"textDocument/moniker"}),textDocumentPrepareTypeHierarchy:ze({method:"textDocument/prepareTypeHierarchy"}),typeHierarchySupertypes:ze({method:"typeHierarchy/supertypes"}),typeHierarchySubtypes:ze({method:"typeHierarchy/subtypes"}),textDocumentInlineValue:ze({method:"textDocument/inlineValue"}),textDocumentInlayHint:ze({method:"textDocument/inlayHint"}),inlayHintResolve:ze({method:"inlayHint/resolve"}),textDocumentDiagnostic:ze({method:"textDocument/diagnostic"}),workspaceDiagnostic:ze({method:"workspace/diagnostic"}),textDocumentInlineCompletion:ze({method:"textDocument/inlineCompletion"}),initialize:ze({method:"initialize"}),shutdown:ze({method:"shutdown"}),textDocumentWillSaveWaitUntil:ze({method:"textDocument/willSaveWaitUntil"}),textDocumentCompletion:ze({method:"textDocument/completion"}),completionItemResolve:ze({method:"completionItem/resolve"}),textDocumentHover:ze({method:"textDocument/hover"}),textDocumentSignatureHelp:ze({method:"textDocument/signatureHelp"}),textDocumentDefinition:ze({method:"textDocument/definition"}),textDocumentReferences:ze({method:"textDocument/references"}),textDocumentDocumentHighlight:ze({method:"textDocument/documentHighlight"}),textDocumentDocumentSymbol:ze({method:"textDocument/documentSymbol"}),textDocumentCodeAction:ze({method:"textDocument/codeAction"}),codeActionResolve:ze({method:"codeAction/resolve"}),workspaceSymbol:ze({method:"workspace/symbol"}),workspaceSymbolResolve:ze({method:"workspaceSymbol/resolve"}),textDocumentCodeLens:ze({method:"textDocument/codeLens"}),codeLensResolve:ze({method:"codeLens/resolve"}),textDocumentDocumentLink:ze({method:"textDocument/documentLink"}),documentLinkResolve:ze({method:"documentLink/resolve"}),textDocumentFormatting:ze({method:"textDocument/formatting"}),textDocumentRangeFormatting:ze({method:"textDocument/rangeFormatting"}),textDocumentRangesFormatting:ze({method:"textDocument/rangesFormatting"}),textDocumentOnTypeFormatting:ze({method:"textDocument/onTypeFormatting"}),textDocumentRename:ze({method:"textDocument/rename"}),textDocumentPrepareRename:ze({method:"textDocument/prepareRename"}),workspaceExecuteCommand:ze({method:"workspace/executeCommand"}),workspaceDidChangeWorkspaceFolders:Mi({method:"workspace/didChangeWorkspaceFolders"}),windowWorkDoneProgressCancel:Mi({method:"window/workDoneProgress/cancel"}),workspaceDidCreateFiles:Mi({method:"workspace/didCreateFiles"}),workspaceDidRenameFiles:Mi({method:"workspace/didRenameFiles"}),workspaceDidDeleteFiles:Mi({method:"workspace/didDeleteFiles"}),notebookDocumentDidOpen:Mi({method:"notebookDocument/didOpen"}),notebookDocumentDidChange:Mi({method:"notebookDocument/didChange"}),notebookDocumentDidSave:Mi({method:"notebookDocument/didSave"}),notebookDocumentDidClose:Mi({method:"notebookDocument/didClose"}),initialized:Mi({method:"initialized"}),exit:Mi({method:"exit"}),workspaceDidChangeConfiguration:Mi({method:"workspace/didChangeConfiguration"}),textDocumentDidOpen:Mi({method:"textDocument/didOpen"}),textDocumentDidChange:Mi({method:"textDocument/didChange"}),textDocumentDidClose:Mi({method:"textDocument/didClose"}),textDocumentDidSave:Mi({method:"textDocument/didSave"}),textDocumentWillSave:Mi({method:"textDocument/willSave"}),workspaceDidChangeWatchedFiles:Mi({method:"workspace/didChangeWatchedFiles"}),setTrace:Mi({method:"$/setTrace"}),cancelRequest:Mi({method:"$/cancelRequest"}),progress:Mi({method:"$/progress"})},client:{workspaceWorkspaceFolders:ze({method:"workspace/workspaceFolders"}).optional(),workspaceConfiguration:ze({method:"workspace/configuration"}).optional(),workspaceFoldingRangeRefresh:ze({method:"workspace/foldingRange/refresh"}).optional(),windowWorkDoneProgressCreate:ze({method:"window/workDoneProgress/create"}).optional(),workspaceSemanticTokensRefresh:ze({method:"workspace/semanticTokens/refresh"}).optional(),windowShowDocument:ze({method:"window/showDocument"}).optional(),workspaceInlineValueRefresh:ze({method:"workspace/inlineValue/refresh"}).optional(),workspaceInlayHintRefresh:ze({method:"workspace/inlayHint/refresh"}).optional(),workspaceDiagnosticRefresh:ze({method:"workspace/diagnostic/refresh"}).optional(),clientRegisterCapability:ze({method:"client/registerCapability"}).optional(),clientUnregisterCapability:ze({method:"client/unregisterCapability"}).optional(),windowShowMessageRequest:ze({method:"window/showMessageRequest"}).optional(),workspaceCodeLensRefresh:ze({method:"workspace/codeLens/refresh"}).optional(),workspaceApplyEdit:ze({method:"workspace/applyEdit"}).optional(),windowShowMessage:Mi({method:"window/showMessage"}),windowLogMessage:Mi({method:"window/logMessage"}),telemetryEvent:Mi({method:"telemetry/event"}),textDocumentPublishDiagnostics:Mi({method:"textDocument/publishDiagnostics"}),logTrace:Mi({method:"$/logTrace"}),cancelRequest:Mi({method:"$/cancelRequest"}),progress:Mi({method:"$/progress"})}});function wp(s,e){if(s.textModel!==e)throw new Error(`Expected text model to be ${e}, but got ${s.textModel}`);return s}var Qi=(KO=class{constructor(){ye(this,"_store",new AV)}dispose(){this._store.dispose()}_register(s){if(s===this)throw new Error("Cannot register a disposable on itself!");return this._store.add(s)}},ye(KO,"None",Object.freeze({dispose(){}})),KO),AV=(pk=class{constructor(){ye(this,"_toDispose",new Set),ye(this,"_isDisposed",!1)}dispose(){this._isDisposed||(this._isDisposed=!0,this.clear())}clear(){if(this._toDispose.size!==0)try{for(const s of this._toDispose)s.dispose()}finally{this._toDispose.clear()}}add(s){if(!s)return s;if(s===this)throw new Error("Cannot register a disposable on itself!");return this._isDisposed?pk.DISABLE_DISPOSED_WARNING||console.warn(new Error("Trying to add a disposable to a DisposableStore that has already been disposed of. The added object will be leaked!").stack):this._toDispose.add(s),s}},ye(pk,"DISABLE_DISPOSED_WARNING",!1),pk);const Nae=new Map([[hd.Empty,""],[hd.QuickFix,"quickfix"],[hd.Refactor,"refactor"],[hd.RefactorExtract,"refactor.extract"],[hd.RefactorInline,"refactor.inline"],[hd.RefactorRewrite,"refactor.rewrite"],[hd.Source,"source"],[hd.SourceOrganizeImports,"source.organizeImports"],[hd.SourceFixAll,"source.fixAll"]]);function vHe(s){if(s)return Nae.get(s)??s}const CHe=new Map([[H.CodeActionTriggerType.Invoke,g6.Invoked],[H.CodeActionTriggerType.Auto,g6.Automatic]]);function yHe(s){return CHe.get(s)??g6.Invoked}const Rae=new Map([[Ji.Text,H.CompletionItemKind.Text],[Ji.Method,H.CompletionItemKind.Method],[Ji.Function,H.CompletionItemKind.Function],[Ji.Constructor,H.CompletionItemKind.Constructor],[Ji.Field,H.CompletionItemKind.Field],[Ji.Variable,H.CompletionItemKind.Variable],[Ji.Class,H.CompletionItemKind.Class],[Ji.Interface,H.CompletionItemKind.Interface],[Ji.Module,H.CompletionItemKind.Module],[Ji.Property,H.CompletionItemKind.Property],[Ji.Unit,H.CompletionItemKind.Unit],[Ji.Value,H.CompletionItemKind.Value],[Ji.Enum,H.CompletionItemKind.Enum],[Ji.Keyword,H.CompletionItemKind.Keyword],[Ji.Snippet,H.CompletionItemKind.Snippet],[Ji.Color,H.CompletionItemKind.Color],[Ji.File,H.CompletionItemKind.File],[Ji.Reference,H.CompletionItemKind.Reference],[Ji.Folder,H.CompletionItemKind.Folder],[Ji.EnumMember,H.CompletionItemKind.EnumMember],[Ji.Constant,H.CompletionItemKind.Constant],[Ji.Struct,H.CompletionItemKind.Struct],[Ji.Event,H.CompletionItemKind.Event],[Ji.Operator,H.CompletionItemKind.Operator],[Ji.TypeParameter,H.CompletionItemKind.TypeParameter]]);function SHe(s){return s?Rae.get(s)??H.CompletionItemKind.Text:H.CompletionItemKind.Text}const Aae=new Map([[_He.Deprecated,H.CompletionItemTag.Deprecated]]);function xHe(s){return Aae.get(s)}const LHe=new Map([[H.CompletionTriggerKind.Invoke,oE.Invoked],[H.CompletionTriggerKind.TriggerCharacter,oE.TriggerCharacter],[H.CompletionTriggerKind.TriggerForIncompleteCompletions,oE.TriggerForIncompleteCompletions]]);function kHe(s){return LHe.get(s)??oE.Invoked}const EHe=new Map([[bHe.Snippet,H.CompletionItemInsertTextRule.InsertAsSnippet]]);function IHe(s){if(s)return EHe.get(s)}const Mae=new Map([[Ki.File,H.SymbolKind.File],[Ki.Module,H.SymbolKind.Module],[Ki.Namespace,H.SymbolKind.Namespace],[Ki.Package,H.SymbolKind.Package],[Ki.Class,H.SymbolKind.Class],[Ki.Method,H.SymbolKind.Method],[Ki.Property,H.SymbolKind.Property],[Ki.Field,H.SymbolKind.Field],[Ki.Constructor,H.SymbolKind.Constructor],[Ki.Enum,H.SymbolKind.Enum],[Ki.Interface,H.SymbolKind.Interface],[Ki.Function,H.SymbolKind.Function],[Ki.Variable,H.SymbolKind.Variable],[Ki.Constant,H.SymbolKind.Constant],[Ki.String,H.SymbolKind.String],[Ki.Number,H.SymbolKind.Number],[Ki.Boolean,H.SymbolKind.Boolean],[Ki.Array,H.SymbolKind.Array],[Ki.Object,H.SymbolKind.Object],[Ki.Key,H.SymbolKind.Key],[Ki.Null,H.SymbolKind.Null],[Ki.EnumMember,H.SymbolKind.EnumMember],[Ki.Struct,H.SymbolKind.Struct],[Ki.Event,H.SymbolKind.Event],[Ki.Operator,H.SymbolKind.Operator],[Ki.TypeParameter,H.SymbolKind.TypeParameter]]);function Oae(s){return Mae.get(s)??H.SymbolKind.File}const DHe=new Map([[Tae.Deprecated,H.SymbolTag.Deprecated]]);function Pae(s){return DHe.get(s)}const THe=new Map([[ZO.Text,H.DocumentHighlightKind.Text],[ZO.Read,H.DocumentHighlightKind.Read],[ZO.Write,H.DocumentHighlightKind.Write]]);function NHe(s){return s?THe.get(s)??H.DocumentHighlightKind.Text:H.DocumentHighlightKind.Text}const RHe=new Map([[Pw.Comment,H.FoldingRangeKind.Comment],[Pw.Imports,H.FoldingRangeKind.Imports],[Pw.Region,H.FoldingRangeKind.Region]]);function AHe(s){if(s)return RHe.get(s)}const MHe=new Map([[kn.Error,Gd.Error],[kn.Warning,Gd.Warning],[kn.Info,Gd.Information],[kn.Hint,Gd.Hint]]);function OHe(s){return MHe.get(s)??Gd.Error}const PHe=new Map([[Gd.Error,kn.Error],[Gd.Warning,kn.Warning],[Gd.Information,kn.Info],[Gd.Hint,kn.Hint]]);function FHe(s){return s?PHe.get(s)??kn.Error:kn.Error}const Fae=new Map([[oZ.Unnecessary,Wv.Unnecessary],[oZ.Deprecated,Wv.Deprecated]]);function BHe(s){return Fae.get(s)}const WHe=new Map([[H.SignatureHelpTriggerKind.Invoke,rE.Invoked],[H.SignatureHelpTriggerKind.TriggerCharacter,rE.TriggerCharacter],[H.SignatureHelpTriggerKind.ContentChange,rE.ContentChange]]);function HHe(s){return WHe.get(s)??rE.Invoked}function Fw(s){if(s)return{id:s.command,title:s.title,arguments:s.arguments}}const VHe=new Map([[sZ.Type,H.InlayHintKind.Type],[sZ.Parameter,H.InlayHintKind.Parameter]]);function zHe(s){return s?VHe.get(s)??H.InlayHintKind.Type:H.InlayHintKind.Type}function xg(s,e){if("targetUri"in s){const t=e.bridge.translateBackRange({uri:s.targetUri},s.targetRange);return{uri:t.textModel.uri,range:t.range,originSelectionRange:s.originSelectionRange?e.bridge.translateBackRange({uri:s.targetUri},s.originSelectionRange).range:void 0,targetSelectionRange:s.targetSelectionRange?e.bridge.translateBackRange({uri:s.targetUri},s.targetSelectionRange).range:void 0}}else{const t=e.bridge.translateBackRange({uri:s.uri},s.range);return{uri:t.textModel.uri,range:t.range}}}function Vn(s){return!s||s.length===0?{language:"*"}:s.map(e=>"notebook"in e?typeof e.notebook=="string"?{notebookType:e.notebook,language:e.language}:{notebookType:e.notebook.notebookType,language:e.language,pattern:e.notebook.pattern,scheme:e.notebook.scheme}:{language:e.language,pattern:e.pattern,scheme:e.scheme})}function $He(s,e){if(!e)return!0;const t=s.getLanguageId();if(s.uri.toString(!0),!e||e.length===0)return!0;for(const i of e)if(!(i.language&&i.language!=="*"&&i.language!==t))return!0;return!1}function f6(s){const e={severity:FHe(s.severity),startLineNumber:s.range.start.line+1,startColumn:s.range.start.character+1,endLineNumber:s.range.end.line+1,endColumn:s.range.end.character+1,message:s.message,source:s.source,code:typeof s.code=="string"?s.code:s.code?.toString()};return s.tags&&(e.tags=s.tags.map(t=>BHe(t)).filter(t=>t!==void 0)),s.relatedInformation&&(e.relatedInformation=s.relatedInformation.map(t=>({resource:Yc.parse(t.location.uri),startLineNumber:t.location.range.start.line+1,startColumn:t.location.range.start.character+1,endLineNumber:t.location.range.end.line+1,endColumn:t.location.range.end.character+1,message:t.message}))),e}var UHe=class extends Qi{constructor(s){super(),this._connection=s,this._register(this._connection.capabilities.addStaticClientCapabilities({textDocument:{completion:{dynamicRegistration:!0,contextSupport:!0,completionItemKind:{valueSet:Array.from(Rae.keys())},completionItem:{tagSupport:{valueSet:Array.from(Aae.keys())},commitCharactersSupport:!0,deprecatedSupport:!0,preselectSupport:!0}}}})),this._register(this._connection.capabilities.registerCapabilityHandler(nt.textDocumentCompletion,!0,e=>H.registerCompletionItemProvider(Vn(e.documentSelector),new qHe(this._connection,e))))}},qHe=class{constructor(s,e){ye(this,"resolveCompletionItem"),this._client=s,this._capabilities=e,e.resolveProvider&&(this.resolveCompletionItem=async(t,i)=>(Bae(t,await this._client.server.completionItemResolve(t._lspItem),this._client.bridge,t._translated,t._model),t))}get triggerCharacters(){return this._capabilities.triggerCharacters}async provideCompletionItems(s,e,t,i){const n=this._client.bridge.translate(s,e),o=await this._client.server.textDocumentCompletion({textDocument:n.textDocument,position:n.position,context:t.triggerCharacter?{triggerKind:kHe(t.triggerKind),triggerCharacter:t.triggerCharacter}:void 0});return o?{suggestions:(Array.isArray(o)?o:o.items).map(r=>({...jHe(r,this._client.bridge,n,s,e),_lspItem:r,_translated:n,_model:s}))}:{suggestions:[]}}};function jHe(s,e,t,i,n){let o=s.insertText||s.label,r;s.textEdit&&("range"in s.textEdit?(o=s.textEdit.newText,r=wp(e.translateBackRange(t.textDocument,s.textEdit.range),i).range):(o=s.textEdit.newText,r={insert:wp(e.translateBackRange(t.textDocument,s.textEdit.insert),i).range,replace:wp(e.translateBackRange(t.textDocument,s.textEdit.replace),i).range})),r||(r=Hc.fromPositions(n,n));const a={label:s.label,kind:SHe(s.kind),insertText:o,sortText:s.sortText,filterText:s.filterText,preselect:s.preselect,commitCharacters:s.commitCharacters,range:r};return Bae(a,s,e,t,i),a}function Bae(s,e,t,i,n){e.detail!==void 0&&(s.detail=e.detail),e.documentation!==void 0&&(s.documentation=GHe(e.documentation)),e.insertTextFormat!==void 0&&(s.insertTextRules=IHe(e.insertTextFormat)),e.tags&&e.tags.length>0&&(s.tags=e.tags.map(xHe).filter(o=>o!==void 0)),e.additionalTextEdits&&e.additionalTextEdits.length>0&&(s.additionalTextEdits=e.additionalTextEdits.map(o=>({range:wp(t.translateBackRange(i.textDocument,o.range),n).range,text:o.newText}))),e.command&&(s.command=Fw(e.command))}function GHe(s){if(s)return typeof s=="string"?s:{value:s.value,isTrusted:!0}}var KHe=class extends Qi{constructor(s){super(),this._connection=s,this._register(this._connection.capabilities.addStaticClientCapabilities({textDocument:{hover:{dynamicRegistration:!0,contentFormat:[IT.Markdown,IT.PlainText]}}})),this._register(this._connection.capabilities.registerCapabilityHandler(nt.textDocumentHover,!0,e=>H.registerHoverProvider(Vn(e.documentSelector),new YHe(this._connection,e))))}},YHe=class{constructor(s,e){this._client=s,this._capabilities=e}async provideHover(s,e,t){const i=this._client.bridge.translate(s,e),n=await this._client.server.textDocumentHover({textDocument:i.textDocument,position:i.position});return!n||!n.contents?null:{contents:ZHe(n.contents),range:n.range?this._client.bridge.translateBackRange(i.textDocument,n.range).range:void 0}}};function ZHe(s){return Array.isArray(s)?s.map(e=>rZ(e)):[rZ(s)]}function rZ(s){return typeof s=="string"?{value:s,isTrusted:!0}:"kind"in s?{value:s.value,isTrusted:!0}:{value:`\`\`\`${s.language}
${s.value}
\`\`\``,isTrusted:!0}}var XHe=class extends Qi{constructor(s){super(),this._connection=s,this._register(this._connection.capabilities.addStaticClientCapabilities({textDocument:{signatureHelp:{dynamicRegistration:!0,contextSupport:!0,signatureInformation:{documentationFormat:[IT.Markdown,IT.PlainText],parameterInformation:{labelOffsetSupport:!0},activeParameterSupport:!0}}}})),this._register(this._connection.capabilities.registerCapabilityHandler(nt.textDocumentSignatureHelp,!0,e=>H.registerSignatureHelpProvider(Vn(e.documentSelector),new QHe(this._connection,e))))}},QHe=class{constructor(s,e){ye(this,"signatureHelpTriggerCharacters"),ye(this,"signatureHelpRetriggerCharacters"),this._client=s,this._capabilities=e,this.signatureHelpTriggerCharacters=e.triggerCharacters,this.signatureHelpRetriggerCharacters=e.retriggerCharacters}async provideSignatureHelp(s,e,t,i){const n=this._client.bridge.translate(s,e),o=await this._client.server.textDocumentSignatureHelp({textDocument:n.textDocument,position:n.position,context:{triggerKind:HHe(i.triggerKind),triggerCharacter:i.triggerCharacter,isRetrigger:i.isRetrigger}});return o?{value:{signatures:o.signatures.map(r=>({label:r.label,documentation:aZ(r.documentation),parameters:r.parameters?.map(a=>({label:a.label,documentation:aZ(a.documentation)}))||[],activeParameter:r.activeParameter})),activeSignature:o.activeSignature||0,activeParameter:o.activeParameter||0},dispose:()=>{}}:null}};function aZ(s){if(s)return typeof s=="string"?s:{value:s.value,isTrusted:!0}}var JHe=class extends Qi{constructor(s){super(),this._connection=s,this._register(this._connection.capabilities.addStaticClientCapabilities({textDocument:{definition:{dynamicRegistration:!0,linkSupport:!0}}})),this._register(this._connection.capabilities.registerCapabilityHandler(nt.textDocumentDefinition,!0,e=>H.registerDefinitionProvider(Vn(e.documentSelector),new eVe(this._connection,e))))}},eVe=class{constructor(s,e){this._client=s,this._capabilities=e}async provideDefinition(s,e,t){const i=this._client.bridge.translate(s,e),n=await this._client.server.textDocumentDefinition({textDocument:i.textDocument,position:i.position});return n?Array.isArray(n)?n.map(o=>xg(o,this._client)):xg(n,this._client):null}},tVe=class extends Qi{constructor(s){super(),this._connection=s,this._register(this._connection.capabilities.addStaticClientCapabilities({textDocument:{declaration:{dynamicRegistration:!0,linkSupport:!0}}})),this._register(this._connection.capabilities.registerCapabilityHandler(nt.textDocumentDeclaration,!0,e=>H.registerDeclarationProvider(Vn(e.documentSelector),new iVe(this._connection,e))))}},iVe=class{constructor(s,e){this._client=s,this._capabilities=e}async provideDeclaration(s,e,t){const i=this._client.bridge.translate(s,e),n=await this._client.server.textDocumentDeclaration({textDocument:i.textDocument,position:i.position});return n?Array.isArray(n)?n.map(o=>xg(o,this._client)):xg(n,this._client):null}},nVe=class extends Qi{constructor(s){super(),this._connection=s,this._register(this._connection.capabilities.addStaticClientCapabilities({textDocument:{typeDefinition:{dynamicRegistration:!0,linkSupport:!0}}})),this._register(this._connection.capabilities.registerCapabilityHandler(nt.textDocumentTypeDefinition,!0,e=>H.registerTypeDefinitionProvider(Vn(e.documentSelector),new sVe(this._connection,e))))}},sVe=class{constructor(s,e){this._client=s,this._capabilities=e}async provideTypeDefinition(s,e,t){const i=this._client.bridge.translate(s,e),n=await this._client.server.textDocumentTypeDefinition({textDocument:i.textDocument,position:i.position});return n?Array.isArray(n)?n.map(o=>xg(o,this._client)):xg(n,this._client):null}},oVe=class extends Qi{constructor(s){super(),this._connection=s,this._register(this._connection.capabilities.addStaticClientCapabilities({textDocument:{implementation:{dynamicRegistration:!0,linkSupport:!0}}})),this._register(this._connection.capabilities.registerCapabilityHandler(nt.textDocumentImplementation,!0,e=>H.registerImplementationProvider(Vn(e.documentSelector),new rVe(this._connection,e))))}},rVe=class{constructor(s,e){this._client=s,this._capabilities=e}async provideImplementation(s,e,t){const i=this._client.bridge.translate(s,e),n=await this._client.server.textDocumentImplementation({textDocument:i.textDocument,position:i.position});return n?Array.isArray(n)?n.map(o=>xg(o,this._client)):xg(n,this._client):null}},aVe=class extends Qi{constructor(s){super(),this._connection=s,this._register(this._connection.capabilities.addStaticClientCapabilities({textDocument:{references:{dynamicRegistration:!0}}})),this._register(this._connection.capabilities.registerCapabilityHandler(nt.textDocumentReferences,!0,e=>H.registerReferenceProvider(Vn(e.documentSelector),new lVe(this._connection,e))))}},lVe=class{constructor(s,e){this._client=s,this._capabilities=e}async provideReferences(s,e,t,i){const n=this._client.bridge.translate(s,e),o=await this._client.server.textDocumentReferences({textDocument:n.textDocument,position:n.position,context:{includeDeclaration:t.includeDeclaration}});return o?o.map(r=>{const a=this._client.bridge.translateBackRange({uri:r.uri},r.range);return{uri:a.textModel.uri,range:a.range}}):null}},cVe=class extends Qi{constructor(s){super(),this._connection=s,this._register(this._connection.capabilities.addStaticClientCapabilities({textDocument:{documentHighlight:{dynamicRegistration:!0}}})),this._register(this._connection.capabilities.registerCapabilityHandler(nt.textDocumentDocumentHighlight,!0,e=>H.registerDocumentHighlightProvider(Vn(e.documentSelector),new dVe(this._connection,e))))}},dVe=class{constructor(s,e){this._client=s,this._capabilities=e}async provideDocumentHighlights(s,e,t){const i=this._client.bridge.translate(s,e),n=await this._client.server.textDocumentDocumentHighlight({textDocument:i.textDocument,position:i.position});return n?n.map(o=>({range:this._client.bridge.translateBackRange(i.textDocument,o.range).range,kind:NHe(o.kind)})):null}},hVe=class extends Qi{constructor(s){super(),this._connection=s,this._register(this._connection.capabilities.addStaticClientCapabilities({textDocument:{documentSymbol:{dynamicRegistration:!0,hierarchicalDocumentSymbolSupport:!0,symbolKind:{valueSet:Array.from(Mae.keys())},tagSupport:{valueSet:[Tae.Deprecated]}}}})),this._register(this._connection.capabilities.registerCapabilityHandler(nt.textDocumentDocumentSymbol,!0,e=>H.registerDocumentSymbolProvider(Vn(e.documentSelector),new uVe(this._connection,e))))}},uVe=class{constructor(s,e){this._client=s,this._capabilities=e}async provideDocumentSymbols(s,e){const t=this._client.bridge.translate(s,new Fh(1,1)),i=await this._client.server.textDocumentDocumentSymbol({textDocument:t.textDocument});return i?Array.isArray(i)&&i.length>0?"location"in i[0]?i.map(n=>gVe(n,this._client)):i.map(n=>Wae(n,this._client,t.textDocument)):[]:null}};function Wae(s,e,t){return{name:s.name,detail:s.detail||"",kind:Oae(s.kind),tags:s.tags?.map(i=>Pae(i)).filter(i=>i!==void 0)||[],range:e.bridge.translateBackRange(t,s.range).range,selectionRange:e.bridge.translateBackRange(t,s.selectionRange).range,children:s.children?.map(i=>Wae(i,e,t))||[]}}function gVe(s,e){return{name:s.name,detail:"",kind:Oae(s.kind),tags:s.tags?.map(t=>Pae(t)).filter(t=>t!==void 0)||[],range:e.bridge.translateBackRange({uri:s.location.uri},s.location.range).range,selectionRange:e.bridge.translateBackRange({uri:s.location.uri},s.location.range).range,children:[]}}var fVe=class extends Qi{constructor(s){super(),this._connection=s,this._register(this._connection.capabilities.addStaticClientCapabilities({textDocument:{rename:{dynamicRegistration:!0,prepareSupport:!0}}})),this._register(this._connection.capabilities.registerCapabilityHandler(nt.textDocumentRename,!0,e=>H.registerRenameProvider(Vn(e.documentSelector),new pVe(this._connection,e))))}},pVe=class{constructor(s,e){this._client=s,this._capabilities=e}async provideRenameEdits(s,e,t,i){const n=this._client.bridge.translate(s,e),o=await this._client.server.textDocumentRename({textDocument:n.textDocument,position:n.position,newName:t});return o?mVe(o,this._client):null}async resolveRenameLocation(s,e,t){if(!this._capabilities.prepareProvider)return null;const i=this._client.bridge.translate(s,e),n=await this._client.server.textDocumentPrepareRename({textDocument:i.textDocument,position:i.position});if(!n)return null;if("range"in n&&"placeholder"in n)return{range:this._client.bridge.translateBackRange(i.textDocument,n.range).range,text:n.placeholder};if("defaultBehavior"in n)return null;if("start"in n&&"end"in n){const o=this._client.bridge.translateBackRange(i.textDocument,n).range;return{range:o,text:s.getValueInRange(o)}}return null}};function mVe(s,e){const t=[];if(s.changes)for(const i in s.changes){const n=s.changes[i];for(const o of n){const r=e.bridge.translateBackRange({uri:i},o.range);t.push({resource:r.textModel.uri,versionId:void 0,textEdit:{range:r.range,text:o.newText}})}}if(s.documentChanges){for(const i of s.documentChanges)if("textDocument"in i){const n=i.textDocument.uri;for(const o of i.edits){const r=e.bridge.translateBackRange({uri:n},o.range);t.push({resource:r.textModel.uri,versionId:i.textDocument.version,textEdit:{range:r.range,text:o.newText}})}}}return{edits:t}}var _Ve=class extends Qi{constructor(s){super(),this._connection=s,this._register(this._connection.capabilities.addStaticClientCapabilities({textDocument:{codeAction:{dynamicRegistration:!0,codeActionLiteralSupport:{codeActionKind:{valueSet:Array.from(Nae.keys())}},isPreferredSupport:!0,disabledSupport:!0,dataSupport:!0,resolveSupport:{properties:["edit"]}}}})),this._register(this._connection.capabilities.registerCapabilityHandler(nt.textDocumentCodeAction,!0,e=>H.registerCodeActionProvider(Vn(e.documentSelector),new bVe(this._connection,e))))}},bVe=class{constructor(s,e){ye(this,"resolveCodeAction"),this._client=s,this._capabilities=e,e.resolveProvider&&(this.resolveCodeAction=async(t,i)=>{if(t._lspAction){const n=await this._client.server.codeActionResolve(t._lspAction);n.edit&&(t.edit=lZ(n.edit,this._client)),n.command&&(t.command=Fw(n.command))}return t})}async provideCodeActions(s,e,t,i){const n=this._client.bridge.translate(s,e.getStartPosition()),o=await this._client.server.textDocumentCodeAction({textDocument:n.textDocument,range:this._client.bridge.translateRange(s,e),context:{diagnostics:t.markers.map(r=>({range:this._client.bridge.translateRange(s,Hc.lift(r)),message:r.message,severity:OHe(r.severity)})),triggerKind:yHe(t.trigger)}});return o?{actions:(Array.isArray(o)?o:[o]).map(r=>{if("title"in r&&!("kind"in r)){const a=r;return{title:a.title,command:Fw(a)}}else{const a=r;return{title:a.title,kind:vHe(a.kind),isPreferred:a.isPreferred,disabled:a.disabled?.reason,edit:a.edit?lZ(a.edit,this._client):void 0,command:Fw(a.command),_lspAction:a}}}),dispose:()=>{}}:null}};function lZ(s,e){const t=[];if(s.changes)for(const i in s.changes){const n=s.changes[i];for(const o of n){const r=e.bridge.translateBackRange({uri:i},o.range);t.push({resource:r.textModel.uri,versionId:void 0,textEdit:{range:r.range,text:o.newText}})}}if(s.documentChanges){for(const i of s.documentChanges)if("textDocument"in i){const n=i.textDocument.uri;for(const o of i.edits){const r=e.bridge.translateBackRange({uri:n},o.range);t.push({resource:r.textModel.uri,versionId:i.textDocument.version??void 0,textEdit:{range:r.range,text:o.newText}})}}}return{edits:t}}var wVe=class extends Qi{constructor(s){super(),this._connection=s,this._register(this._connection.capabilities.addStaticClientCapabilities({textDocument:{codeLens:{dynamicRegistration:!0}}})),this._register(this._connection.capabilities.registerCapabilityHandler(nt.textDocumentCodeLens,!0,e=>H.registerCodeLensProvider(Vn(e.documentSelector),new vVe(this._connection,e))))}},vVe=class{constructor(s,e){this._client=s,this._capabilities=e}async provideCodeLenses(s,e){const t=this._client.bridge.translate(s,new Fh(1,1)),i=await this._client.server.textDocumentCodeLens({textDocument:t.textDocument});return i?{lenses:i.map(n=>({range:wp(this._client.bridge.translateBackRange(t.textDocument,n.range),s).range,command:Fw(n.command),_lspCodeLens:n})),dispose:()=>{}}:null}async resolveCodeLens(s,e,t){if(!this._capabilities.resolveProvider||!e._lspCodeLens)return e;const i=await this._client.server.codeLensResolve(e._lspCodeLens);return i.command&&(e.command={id:i.command.command,title:i.command.title,arguments:i.command.arguments}),e}},CVe=class extends Qi{constructor(s){super(),this._connection=s,this._register(this._connection.capabilities.addStaticClientCapabilities({textDocument:{documentLink:{dynamicRegistration:!0,tooltipSupport:!0}}})),this._register(this._connection.capabilities.registerCapabilityHandler(nt.textDocumentDocumentLink,!0,e=>H.registerLinkProvider(Vn(e.documentSelector),new yVe(this._connection,e))))}},yVe=class{constructor(s,e){this._client=s,this._capabilities=e}async provideLinks(s,e){const t=this._client.bridge.translate(s,new Fh(1,1)),i=await this._client.server.textDocumentDocumentLink({textDocument:t.textDocument});return i?{links:i.map(n=>({range:this._client.bridge.translateBackRange(t.textDocument,n.range).range,url:n.target,tooltip:n.tooltip}))}:null}async resolveLink(s,e){return this._capabilities.resolveProvider,s}},SVe=class extends Qi{constructor(s){super(),this._connection=s,this._register(this._connection.capabilities.addStaticClientCapabilities({textDocument:{formatting:{dynamicRegistration:!0}}})),this._register(this._connection.capabilities.registerCapabilityHandler(nt.textDocumentFormatting,!0,e=>H.registerDocumentFormattingEditProvider(Vn(e.documentSelector),new xVe(this._connection,e))))}},xVe=class{constructor(s,e){this._client=s,this._capabilities=e}async provideDocumentFormattingEdits(s,e,t){const i=this._client.bridge.translate(s,new Fh(1,1)),n=await this._client.server.textDocumentFormatting({textDocument:i.textDocument,options:{tabSize:e.tabSize,insertSpaces:e.insertSpaces}});return n?n.map(o=>({range:this._client.bridge.translateBackRange(i.textDocument,o.range).range,text:o.newText})):null}},LVe=class extends Qi{constructor(s){super(),this._connection=s,this._register(this._connection.capabilities.addStaticClientCapabilities({textDocument:{rangeFormatting:{dynamicRegistration:!0}}})),this._register(this._connection.capabilities.registerCapabilityHandler(nt.textDocumentRangeFormatting,!0,e=>H.registerDocumentRangeFormattingEditProvider(Vn(e.documentSelector),new kVe(this._connection,e))))}},kVe=class{constructor(s,e){this._client=s,this._capabilities=e}async provideDocumentRangeFormattingEdits(s,e,t,i){const n=this._client.bridge.translate(s,e.getStartPosition()),o=await this._client.server.textDocumentRangeFormatting({textDocument:n.textDocument,range:this._client.bridge.translateRange(s,e),options:{tabSize:t.tabSize,insertSpaces:t.insertSpaces}});return o?o.map(r=>({range:this._client.bridge.translateBackRange(n.textDocument,r.range).range,text:r.newText})):null}},EVe=class extends Qi{constructor(s){super(),this._connection=s,this._register(this._connection.capabilities.addStaticClientCapabilities({textDocument:{onTypeFormatting:{dynamicRegistration:!0}}})),this._register(this._connection.capabilities.registerCapabilityHandler(nt.textDocumentOnTypeFormatting,!0,e=>H.registerOnTypeFormattingEditProvider(Vn(e.documentSelector),new IVe(this._connection,e))))}},IVe=class{constructor(s,e){ye(this,"autoFormatTriggerCharacters"),this._client=s,this._capabilities=e,this.autoFormatTriggerCharacters=[e.firstTriggerCharacter,...e.moreTriggerCharacter||[]]}async provideOnTypeFormattingEdits(s,e,t,i,n){const o=this._client.bridge.translate(s,e),r=await this._client.server.textDocumentOnTypeFormatting({textDocument:o.textDocument,position:o.position,ch:t,options:{tabSize:i.tabSize,insertSpaces:i.insertSpaces}});return r?r.map(a=>({range:this._client.bridge.translateBackRange(o.textDocument,a.range).range,text:a.newText})):null}},DVe=class extends Qi{constructor(s){super(),this._connection=s,this._register(this._connection.capabilities.addStaticClientCapabilities({textDocument:{foldingRange:{dynamicRegistration:!0,rangeLimit:5e3,lineFoldingOnly:!1,foldingRangeKind:{valueSet:[Pw.Comment,Pw.Imports,Pw.Region]}}}})),this._register(this._connection.capabilities.registerCapabilityHandler(nt.textDocumentFoldingRange,!0,e=>H.registerFoldingRangeProvider(Vn(e.documentSelector),new TVe(this._connection,e))))}},TVe=class{constructor(s,e){this._client=s,this._capabilities=e}async provideFoldingRanges(s,e,t){const i=this._client.bridge.translate(s,new Fh(1,1)),n=await this._client.server.textDocumentFoldingRange({textDocument:i.textDocument});return n?n.map(o=>({start:o.startLine+1,end:o.endLine+1,kind:AHe(o.kind)})):null}},NVe=class extends Qi{constructor(s){super(),this._connection=s,this._register(this._connection.capabilities.addStaticClientCapabilities({textDocument:{selectionRange:{dynamicRegistration:!0}}})),this._register(this._connection.capabilities.registerCapabilityHandler(nt.textDocumentSelectionRange,!0,e=>H.registerSelectionRangeProvider(Vn(e.documentSelector),new RVe(this._connection,e))))}},RVe=class{constructor(s,e){this._client=s,this._capabilities=e}async provideSelectionRanges(s,e,t){const i=this._client.bridge.translate(s,e[0]),n=await this._client.server.textDocumentSelectionRange({textDocument:i.textDocument,positions:e.map(o=>this._client.bridge.translate(s,o).position)});return n?n.map(o=>this.convertSelectionRange(o,i.textDocument)):null}convertSelectionRange(s,e){const t=[];let i=s;for(;i;)t.push({range:this._client.bridge.translateBackRange(e,i.range).range}),i=i.parent;return t}},AVe=class extends Qi{constructor(s){super(),ye(this,"_providers",new Set),this._connection=s,this._register(this._connection.capabilities.addStaticClientCapabilities({textDocument:{inlayHint:{dynamicRegistration:!0,resolveSupport:{properties:["tooltip","textEdits","label.tooltip","label.location","label.command"]}}},workspace:{inlayHint:{refreshSupport:!0}}})),this._register(this._connection.connection.registerRequestHandler(R0.client.workspaceInlayHintRefresh,async()=>{for(const e of this._providers)e.refresh();return{ok:null}})),this._register(this._connection.capabilities.registerCapabilityHandler(nt.textDocumentInlayHint,!0,e=>{const t=new MVe(this._connection,e);this._providers.add(t);const i=H.registerInlayHintsProvider(Vn(e.documentSelector),t);return{dispose:()=>{this._providers.delete(t),i.dispose()}}}))}},MVe=class{constructor(s,e){ye(this,"_onDidChangeInlayHints",new Sg),ye(this,"onDidChangeInlayHints",this._onDidChangeInlayHints.event),ye(this,"resolveInlayHint"),this._client=s,this._capabilities=e,e.resolveProvider&&(this.resolveInlayHint=async(t,i)=>{const n=await this._client.server.inlayHintResolve(t._lspInlayHint);return n.tooltip&&(t.tooltip=p6(n.tooltip)),n.label!==t._lspInlayHint.label&&(t.label=cZ(n.label)),n.textEdits&&(t.textEdits=n.textEdits.map(o=>({range:this._client.bridge.translateBackRange({uri:t._targetUri},o.range).range,text:o.newText}))),t})}refresh(){this._onDidChangeInlayHints.fire()}async provideInlayHints(s,e,t){const i=this._client.bridge.translate(s,e.getStartPosition()),n=await OVe(async()=>await this._client.server.textDocumentInlayHint({textDocument:i.textDocument,range:this._client.bridge.translateRange(s,e)}));return n?{hints:n.map(o=>({label:cZ(o.label),position:wp(this._client.bridge.translateBack(i.textDocument,o.position),s).position,kind:zHe(o.kind),tooltip:p6(o.tooltip),paddingLeft:o.paddingLeft,paddingRight:o.paddingRight,textEdits:o.textEdits?.map(r=>({range:wp(this._client.bridge.translateBackRange(i.textDocument,r.range),s).range,text:r.newText})),_lspInlayHint:o,_targetUri:i.textDocument.uri})),dispose:()=>{}}:null}};async function OVe(s){for(let e=3;;e--)try{return await s()}catch(t){if(t.message==="content modified"&&e>0)continue;throw t}}function cZ(s){return typeof s=="string"?s:s.map(e=>{const t={label:e.value,tooltip:p6(e.tooltip),command:Fw(e.command)};return e.location&&(t.location={uri:Yc.parse(e.location.uri),range:new Hc(e.location.range.start.line+1,e.location.range.start.character+1,e.location.range.end.line+1,e.location.range.end.character+1)}),t})}function p6(s){if(s)return typeof s=="string"?s:{value:s.value,isTrusted:!0}}var PVe=class extends Qi{constructor(s){super(),this._connection=s,this._register(this._connection.capabilities.addStaticClientCapabilities({textDocument:{semanticTokens:{dynamicRegistration:!0,requests:{range:!0,full:{delta:!0}},tokenTypes:["namespace","type","class","enum","interface","struct","typeParameter","parameter","variable","property","enumMember","event","function","method","macro","keyword","modifier","comment","string","number","regexp","operator","decorator"],tokenModifiers:["declaration","definition","readonly","static","deprecated","abstract","async","modification","documentation","defaultLibrary"],formats:[wHe.Relative],overlappingTokenSupport:!1,multilineTokenSupport:!0}}})),this._register(this._connection.capabilities.registerCapabilityHandler(nt.textDocumentSemanticTokensFull,!0,e=>H.registerDocumentSemanticTokensProvider(Vn(e.documentSelector),new FVe(this._connection,e))))}},FVe=class{constructor(s,e){this._client=s,this._capabilities=e}getLegend(){return{tokenTypes:this._capabilities.legend.tokenTypes,tokenModifiers:this._capabilities.legend.tokenModifiers}}releaseDocumentSemanticTokens(s){}async provideDocumentSemanticTokens(s,e,t){const i=this._client.bridge.translate(s,s.getPositionAt(0)),n=this._capabilities.full;if(e&&n&&typeof n=="object"&&n.delta){const r=await this._client.server.textDocumentSemanticTokensFullDelta({textDocument:i.textDocument,previousResultId:e});return r?"edits"in r?{resultId:r.resultId,edits:r.edits.map(a=>({start:a.start,deleteCount:a.deleteCount,data:a.data?new Uint32Array(a.data):void 0}))}:{resultId:r.resultId,data:new Uint32Array(r.data)}:null}const o=await this._client.server.textDocumentSemanticTokensFull({textDocument:i.textDocument});return o?{resultId:o.resultId,data:new Uint32Array(o.data)}:null}async provideDocumentSemanticTokensEdits(s,e,t){return this.provideDocumentSemanticTokens(s,e,t)}},BVe=class extends Qi{constructor(s){super(),ye(this,"_diagnosticsMarkerOwner","lsp"),ye(this,"_pullDiagnosticProviders",new Map),this._connection=s,this._register(this._connection.capabilities.addStaticClientCapabilities({textDocument:{publishDiagnostics:{relatedInformation:!0,tagSupport:{valueSet:[...Fae.keys()]},versionSupport:!0,codeDescriptionSupport:!0,dataSupport:!0},diagnostic:{dynamicRegistration:!0,relatedDocumentSupport:!0}}}));debugger;this._register(this._connection.connection.registerNotificationHandler(R0.client.textDocumentPublishDiagnostics,e=>this._handlePublishDiagnostics(e))),this._register(this._connection.capabilities.registerCapabilityHandler(nt.textDocumentDiagnostic,!0,e=>{const t=new AV;for(const i of ti.getModels())this._addPullDiagnosticProvider(i,e,t);return t.add(ti.onDidCreateModel(i=>{this._addPullDiagnosticProvider(i,e,t)})),t}))}_addPullDiagnosticProvider(s,e,t){if(s.getLanguageId(),!$He(s,e.documentSelector))return;const i=new WVe(s,this._connection,this._diagnosticsMarkerOwner,e);this._pullDiagnosticProviders.set(s,i),t.add(i),t.add(s.onWillDispose(()=>{this._pullDiagnosticProviders.delete(s)}))}_handlePublishDiagnostics(s){const e=s.uri;try{const t=this._connection.bridge.translateBack({uri:e},{line:0,character:0}).textModel;if(!t||t.isDisposed())return;const i=s.diagnostics.map(n=>f6(n));ti.setModelMarkers(t,this._diagnosticsMarkerOwner,i)}catch(t){console.debug(`Could not set diagnostics for ${e}:`,t)}}},WVe=class extends Qi{constructor(s,e,t,i){super(),ye(this,"_updateHandle"),ye(this,"_previousResultId"),this._model=s,this._connection=e,this._markerOwner=t,this._capability=i,this._register(this._model.onDidChangeContent(()=>{this._scheduleDiagnosticUpdate()})),this._scheduleDiagnosticUpdate()}_scheduleDiagnosticUpdate(){this._updateHandle!==void 0&&clearTimeout(this._updateHandle),this._updateHandle=window.setTimeout(()=>{this._updateHandle=void 0,this._requestDiagnostics()},500)}async _requestDiagnostics(){if(!this._model.isDisposed())try{const s=this._connection.bridge.translate(this._model,new Fh(1,1)),e=await this._connection.server.textDocumentDiagnostic({textDocument:s.textDocument,identifier:this._capability.identifier,previousResultId:this._previousResultId});if(this._model.isDisposed())return;this._handleDiagnosticReport(e)}catch(s){console.error("Error requesting diagnostics:",s)}}_handleDiagnosticReport(s){if(s.kind==="full"){this._previousResultId=s.resultId;const e=s.items.map(t=>f6(t));ti.setModelMarkers(this._model,this._markerOwner,e),"relatedDocuments"in s&&s.relatedDocuments&&this._handleRelatedDocuments(s.relatedDocuments)}else s.kind==="unchanged"&&(this._previousResultId=s.resultId)}_handleRelatedDocuments(s){for(const[e,t]of Object.entries(s))try{const i=this._connection.bridge.translateBack({uri:e},{line:0,character:0}).textModel;if(!i||i.isDisposed())continue;if(t.kind==="full"){const n=t.items.map(o=>f6(o));ti.setModelMarkers(i,this._markerOwner,n)}}catch(i){console.debug(`Could not set related diagnostics for ${e}:`,i)}}dispose(){this._updateHandle!==void 0&&(clearTimeout(this._updateHandle),this._updateHandle=void 0),super.dispose()}},HVe=class{constructor(s,e,t,i){this.server=s,this.bridge=e,this.capabilities=t,this.connection=i}},VVe=class extends Qi{constructor(s){super(),ye(this,"_staticCapabilities",new Set),ye(this,"_dynamicFromStatic",UVe.create()),ye(this,"_registrations",new Map),ye(this,"_serverCapabilities"),this._connection=s,this._register(this._connection.registerRequestHandler(R0.client.clientRegisterCapability,async e=>{for(const t of e.registrations){const i=hZ(t.method),n=new dZ(t.id,i,t.registerOptions,!1);this._registerCapabilityOptions(n)}return{ok:null}})),this._register(this._connection.registerRequestHandler(R0.client.clientUnregisterCapability,async e=>{for(const t of e.unregisterations){const i=hZ(t.method),n=this._registrations.get(i),o=n?.registrations.get(t.id);if(!o)throw new Error(`No registration for method ${t.method} with id ${t.id}`);o?.handlerDisposables.forEach(r=>r.dispose()),n?.registrations.delete(t.id)}return{ok:null}}))}_registerCapabilityOptions(s){let e=this._registrations.get(s.capability);if(e||(e=new uZ,this._registrations.set(s.capability,e)),e.registrations.has(s.id))throw new Error(`Handler for method ${s.capability.method} with id ${s.id} already registered`);e.registrations.set(s.id,s);for(const t of e.handlers)!t.handleStaticCapability&&s.isFromStatic||s.handlerDisposables.set(t,t.handler(s.options))}setServerCapabilities(s){if(this._serverCapabilities)throw new Error("Server capabilities already set");this._serverCapabilities=s;for(const e of Object.values(nt)){const t=this._dynamicFromStatic.getOptions(e,s);t&&this._registerCapabilityOptions(new dZ(e.method,e,t,!0))}}getClientCapabilities(){const s={};for(const e of this._staticCapabilities)Vae(s,e.cap);return s}addStaticClientCapabilities(s){const e={cap:s};return this._staticCapabilities.add(e),{dispose:()=>{this._staticCapabilities.delete(e)}}}registerCapabilityHandler(s,e,t){let i=this._registrations.get(s);i||(i=new uZ,this._registrations.set(s,i));const n=new zVe(s,e,t);i.handlers.add(n);for(const o of i.registrations.values())!n.handleStaticCapability&&o.isFromStatic||o.handlerDisposables.set(n,t(o.options));return{dispose:()=>{i.handlers.delete(n);for(const o of i.registrations.values()){const r=o.handlerDisposables.get(n);r&&(r.dispose(),o.handlerDisposables.delete(n))}}}}},zVe=class{constructor(s,e,t){this.capability=s,this.handleStaticCapability=e,this.handler=t}},dZ=class{constructor(s,e,t,i){ye(this,"handlerDisposables",new Map),this.id=s,this.capability=e,this.options=t,this.isFromStatic=i}};const $Ve=new Map([...Object.values(nt)].map(s=>[s.method,s]));function hZ(s){const e=$Ve.get(s);if(!e)throw new Error(`No capability found for method ${s}`);return e}var uZ=class{constructor(){ye(this,"handlers",new Set),ye(this,"registrations",new Map)}},UVe=class Hae{constructor(){ye(this,"_mappings",new Map)}static create(){const e=new Hae;return e.set(nt.textDocumentDidChange,t=>{if(t.textDocumentSync!==void 0)return typeof t.textDocumentSync=="object"?{syncKind:t.textDocumentSync.change??mHe.None,documentSelector:null}:{syncKind:t.textDocumentSync,documentSelector:null}}),e.set(nt.textDocumentCompletion,t=>t.completionProvider),e.set(nt.textDocumentHover,t=>t.hoverProvider),e.set(nt.textDocumentSignatureHelp,t=>t.signatureHelpProvider),e.set(nt.textDocumentDefinition,t=>t.definitionProvider),e.set(nt.textDocumentReferences,t=>t.referencesProvider),e.set(nt.textDocumentDocumentHighlight,t=>t.documentHighlightProvider),e.set(nt.textDocumentDocumentSymbol,t=>t.documentSymbolProvider),e.set(nt.textDocumentCodeAction,t=>t.codeActionProvider),e.set(nt.textDocumentCodeLens,t=>t.codeLensProvider),e.set(nt.textDocumentDocumentLink,t=>t.documentLinkProvider),e.set(nt.textDocumentFormatting,t=>t.documentFormattingProvider),e.set(nt.textDocumentRangeFormatting,t=>t.documentRangeFormattingProvider),e.set(nt.textDocumentOnTypeFormatting,t=>t.documentOnTypeFormattingProvider),e.set(nt.textDocumentRename,t=>t.renameProvider),e.set(nt.textDocumentFoldingRange,t=>t.foldingRangeProvider),e.set(nt.textDocumentDeclaration,t=>t.declarationProvider),e.set(nt.textDocumentTypeDefinition,t=>t.typeDefinitionProvider),e.set(nt.textDocumentImplementation,t=>t.implementationProvider),e.set(nt.textDocumentDocumentColor,t=>t.colorProvider),e.set(nt.textDocumentSelectionRange,t=>t.selectionRangeProvider),e.set(nt.textDocumentLinkedEditingRange,t=>t.linkedEditingRangeProvider),e.set(nt.textDocumentPrepareCallHierarchy,t=>t.callHierarchyProvider),e.set(nt.textDocumentSemanticTokensFull,t=>t.semanticTokensProvider),e.set(nt.textDocumentInlayHint,t=>t.inlayHintProvider),e.set(nt.textDocumentInlineValue,t=>t.inlineValueProvider),e.set(nt.textDocumentDiagnostic,t=>t.diagnosticProvider),e.set(nt.textDocumentMoniker,t=>t.monikerProvider),e.set(nt.textDocumentPrepareTypeHierarchy,t=>t.typeHierarchyProvider),e.set(nt.workspaceSymbol,t=>t.workspaceSymbolProvider),e.set(nt.workspaceExecuteCommand,t=>t.executeCommandProvider),e}set(e,t){if(this._mappings.has(e.method))throw new Error(`Capability for method ${e.method} already registered`);this._mappings.set(e.method,t)}getOptions(e,t){const i=this._mappings.get(e.method);if(i)return i(t)}};function Vae(s,e){for(const t of Object.keys(e)){const i=e[t];if(i===void 0)continue;const n=s[t];if(n===void 0){s[t]=i;continue}if(typeof i!="object"||i===null){s[t]=i;continue}if(typeof n!="object"||n===null){s[t]=i;continue}Vae(n,i)}}var qVe=class extends Qi{constructor(s,e){super(),ye(this,"_managedModels",new Map),ye(this,"_managedModelsReverse",new Map),ye(this,"_started",!1),this._server=s,this._capabilities=e,this._register(this._capabilities.addStaticClientCapabilities({textDocument:{synchronization:{dynamicRegistration:!0,willSave:!1,willSaveWaitUntil:!1,didSave:!1}}})),this._register(e.registerCapabilityHandler(nt.textDocumentDidChange,!0,t=>{if(this._started)return{dispose:()=>{}};this._started=!0,this._register(ti.onDidCreateModel(i=>{this._getOrCreateManagedModel(i)}));for(const i of ti.getModels())this._getOrCreateManagedModel(i);return{dispose:()=>{}}}))}_getOrCreateManagedModel(s){if(!this._started)throw new Error("Not started");const e=s.uri.toString(!0).toLowerCase();let t=this._managedModels.get(s);return t||(t=new jVe(s,this._server),this._managedModels.set(s,t),this._managedModelsReverse.set(e,s)),s.onWillDispose(()=>{t.dispose(),this._managedModels.delete(s),this._managedModelsReverse.delete(e)}),t}translateBack(s,e){const t=s.uri.toLowerCase(),i=this._managedModelsReverse.get(t);if(!i)throw new Error(`No text model for uri ${t}`);return{textModel:i,position:new Fh(e.line+1,e.character+1)}}translateBackRange(s,e){const t=s.uri.toLowerCase(),i=this._managedModelsReverse.get(t);if(!i)throw new Error(`No text model for uri ${t}`);return{textModel:i,range:new Hc(e.start.line+1,e.start.character+1,e.end.line+1,e.end.character+1)}}translate(s,e){return{textDocument:{uri:s.uri.toString(!0)},position:{line:e.lineNumber-1,character:e.column-1}}}translateRange(s,e){return{start:{line:e.startLineNumber-1,character:e.startColumn-1},end:{line:e.endLineNumber-1,character:e.endColumn-1}}}},jVe=class extends Qi{constructor(s,e){super(),this._textModel=s,this._api=e;const t=s.uri.toString(!0).toLowerCase();this._api.textDocumentDidOpen({textDocument:{languageId:s.getLanguageId(),uri:t,version:s.getVersionId(),text:s.getValue()}}),this._register(s.onDidChangeContent(i=>{const n=i.changes.map(o=>GVe(o));this._api.textDocumentDidChange({textDocument:{uri:t,version:s.getVersionId()},contentChanges:n})})),this._register({dispose:()=>{this._api.textDocumentDidClose({textDocument:{uri:t}})}})}};function GVe(s){return{range:KVe(s.range),rangeLength:s.rangeLength,text:s.text}}function KVe(s){return{start:{line:s.startLineNumber-1,character:s.startColumn-1},end:{line:s.endLineNumber-1,character:s.endColumn-1}}}var YVe=class{constructor(s){ye(this,"_connection"),ye(this,"_capabilitiesRegistry"),ye(this,"_bridge"),ye(this,"_initPromise");const e=h6.fromTransport(s),t=R0.getServer(e,{});e.startListen(),this._capabilitiesRegistry=new VVe(e),this._bridge=new qVe(t.server,this._capabilitiesRegistry),this._connection=new HVe(t.server,this._bridge,this._capabilitiesRegistry,e),this.createFeatures(),this._initPromise=this._init()}async _init(){const s=await this._connection.server.initialize({processId:null,capabilities:this._capabilitiesRegistry.getClientCapabilities(),rootUri:null});this._connection.server.initialized({}),this._capabilitiesRegistry.setServerCapabilities(s.capabilities)}createFeatures(){const s=new AV;return s.add(new UHe(this._connection)),s.add(new KHe(this._connection)),s.add(new XHe(this._connection)),s.add(new JHe(this._connection)),s.add(new tVe(this._connection)),s.add(new nVe(this._connection)),s.add(new oVe(this._connection)),s.add(new aVe(this._connection)),s.add(new cVe(this._connection)),s.add(new hVe(this._connection)),s.add(new fVe(this._connection)),s.add(new _Ve(this._connection)),s.add(new wVe(this._connection)),s.add(new CVe(this._connection)),s.add(new SVe(this._connection)),s.add(new LVe(this._connection)),s.add(new EVe(this._connection)),s.add(new DVe(this._connection)),s.add(new NVe(this._connection)),s.add(new AVe(this._connection)),s.add(new PVe(this._connection)),s.add(new BVe(this._connection)),s}},wb=null;typeof WebSocket<"u"?wb=WebSocket:typeof MozWebSocket<"u"?wb=MozWebSocket:typeof global<"u"?wb=global.WebSocket||global.MozWebSocket:typeof window<"u"?wb=window.WebSocket||window.MozWebSocket:typeof self<"u"&&(wb=self.WebSocket||self.MozWebSocket);var ZVe=wb;function XVe(s){return"host"in s?{address:`${s.forceTls?"wss":"ws"}://${s.host}:${s.port}`}:s}var QVe=class m6 extends Cae{constructor(e){super(),ye(this,"socket"),ye(this,"errorEmitter",new RV),ye(this,"onError",this.errorEmitter),this.socket=e,e.onmessage=t=>{try{const i=t.data;if(typeof i=="string"){const n=JSON.parse(i);this._dispatchReceivedMessage(n)}else throw new Error("Not supported")}catch(i){this.errorEmitter.fire({error:i})}},e.onclose=t=>{this._onConnectionClosed()}}static connectTo(e){const t=new ZVe(XVe(e).address);return new Promise((i,n)=>{t.onerror=o=>{n(o)},t.onopen=()=>{i(new m6(t))}})}static fromWebSocket(e){return new m6(e)}close(){this.socket.close()}dispose(){this.close()}_sendImpl(e){const t=JSON.stringify(e);return new Promise((i,n)=>{this.socket.send(t,o=>{o?n(o):i()})})}toString(){return`${this.id}@${this.socket.url}`}},zae=class extends Cae{constructor(s,e=void 0,t=void 0){super(),ye(this,"_windowLike"),ye(this,"_source"),ye(this,"_loadingState"),ye(this,"_disposed",!1),ye(this,"_messageHandler",({data:i,source:n})=>{this._source&&n!==this._source||typeof i=="object"&&i&&this._dispatchReceivedMessage(i)}),this._windowLike=s,this._source=e,this._loadingState=t,this._windowLike.addEventListener("message",this._messageHandler)}async _sendImpl(s){if(this._disposed)throw new Error("Transport is disposed");this._loadingState&&!this._loadingState.loaded&&await this._loadingState.onLoaded,this._windowLike.postMessage(s)}toString(){return`${this.id}@${this._windowLike}`}dispose(){this._disposed||(this._disposed=!0,this._windowLike.removeEventListener("message",this._messageHandler))}};function JVe(s){if(typeof window>"u")throw new Error("call this function from the main browser thread");return new zae(s)}function eze(s){if(typeof window>"u")throw new Error("call this function from the main browser thread");return new zae(s.contentWindow,s.contentWindow,{loaded:window.document.readyState==="complete",onLoaded:new Promise(e=>{window.addEventListener("load",()=>e())})})}const tze=Object.freeze(Object.defineProperty({__proto__:null,MonacoLspClient:YVe,WebSocketTransport:QVe,createTransportToIFrame:eze,createTransportToWorker:JVe},Symbol.toStringTag,{value:"Module"}));function ize(s,e){const t=globalThis.MonacoEnvironment;if(t?.createTrustedTypesPolicy)try{return t.createTrustedTypesPolicy(s,e)}catch(i){console.error(i);return}try{return globalThis.trustedTypes?.createPolicy(s,e)}catch(i){console.error(i);return}}let DT;typeof self=="object"&&self.constructor&&self.constructor.name==="DedicatedWorkerGlobalScope"&&globalThis.workerttPolicy!==void 0?DT=globalThis.workerttPolicy:DT=ize("defaultWorkerFactory",{createScriptURL:s=>s});function nze(s){const e=s.label,t=globalThis.MonacoEnvironment;if(t){if(typeof t.getWorker=="function")return t.getWorker("workerMain.js",e);if(typeof t.getWorkerUrl=="function"){const i=t.getWorkerUrl("workerMain.js",e);return new Worker(DT?DT.createScriptURL(i):i,{name:e,type:"module"})}}if(s.createWorker)return s.createWorker();throw new Error("You must define a function MonacoEnvironment.getWorkerUrl or MonacoEnvironment.getWorker")}function Ex(s){const e=Promise.resolve(nze({label:s.label??"monaco-editor-worker",moduleId:s.moduleId,createWorker:s.createWorker})).then(t=>(t.postMessage("ignore"),t.postMessage(s.createData),t));return ti.createWebWorker({worker:e,host:s.host,keepIdleModels:s.keepIdleModels})}function $ae(){return B7e}const sze=globalThis.MonacoEnvironment;sze?.globalAPI&&(globalThis.monaco=$ae());const h2=$ae();h2.languages.css=nae;h2.languages.html=aae;h2.languages.typescript=wae;h2.languages.json=dae;const iQe=Object.freeze(Object.defineProperty({__proto__:null,CancellationTokenSource:Kre,Emitter:Sg,KeyCode:Yre,KeyMod:Zre,MarkerSeverity:kn,MarkerTag:Wv,Position:Fh,Range:Hc,Selection:Xre,SelectionDirection:Qre,Token:Jre,Uri:Yc,createWebWorker:Ex,css:nae,editor:ti,html:aae,json:dae,languages:H,lsp:tze,typescript:wae},Symbol.toStringTag,{value:"Module"}));function nQe(s){return new Worker("/assets/editor.worker-CdQrwHl8.js",{name:s?.name})}function sQe(s){return new Worker("/assets/json.worker-BoL8UZqY.js",{name:s?.name})}const oQe={base:"vs",inherit:!0,rules:[{background:"ffffff",token:""},{foreground:"6a737d",token:"comment"},{foreground:"6a737d",token:"punctuation.definition.comment"},{foreground:"6a737d",token:"string.comment"},{foreground:"005cc5",token:"constant"},{foreground:"005cc5",token:"entity.name.constant"},{foreground:"005cc5",token:"variable.other.constant"},{foreground:"005cc5",token:"variable.language"},{foreground:"6f42c1",token:"entity"},{foreground:"6f42c1",token:"entity.name"},{foreground:"24292e",token:"variable.parameter.function"},{foreground:"22863a",token:"entity.name.tag"},{foreground:"d73a49",token:"keyword"},{foreground:"d73a49",token:"storage"},{foreground:"d73a49",token:"storage.type"},{foreground:"24292e",token:"storage.modifier.package"},{foreground:"24292e",token:"storage.modifier.import"},{foreground:"24292e",token:"storage.type.java"},{foreground:"032f62",token:"string"},{foreground:"032f62",token:"punctuation.definition.string"},{foreground:"032f62",token:"string punctuation.section.embedded source"},{foreground:"005cc5",token:"support"},{foreground:"005cc5",token:"meta.property-name"},{foreground:"e36209",token:"variable"},{foreground:"24292e",token:"variable.other"},{foreground:"b31d28",fontStyle:"bold italic underline",token:"invalid.broken"},{foreground:"b31d28",fontStyle:"bold italic underline",token:"invalid.deprecated"},{foreground:"fafbfc",background:"b31d28",fontStyle:"italic underline",token:"invalid.illegal"},{foreground:"fafbfc",background:"d73a49",fontStyle:"italic underline",token:"carriage-return"},{foreground:"b31d28",fontStyle:"bold italic underline",token:"invalid.unimplemented"},{foreground:"b31d28",token:"message.error"},{foreground:"24292e",token:"string source"},{foreground:"005cc5",token:"string variable"},{foreground:"032f62",token:"source.regexp"},{foreground:"032f62",token:"string.regexp"},{foreground:"032f62",token:"string.regexp.character-class"},{foreground:"032f62",token:"string.regexp constant.character.escape"},{foreground:"032f62",token:"string.regexp source.ruby.embedded"},{foreground:"032f62",token:"string.regexp string.regexp.arbitrary-repitition"},{foreground:"22863a",fontStyle:"bold",token:"string.regexp constant.character.escape"},{foreground:"005cc5",token:"support.constant"},{foreground:"005cc5",token:"support.variable"},{foreground:"005cc5",token:"meta.module-reference"},{foreground:"735c0f",token:"markup.list"},{foreground:"005cc5",fontStyle:"bold",token:"markup.heading"},{foreground:"005cc5",fontStyle:"bold",token:"markup.heading entity.name"},{foreground:"22863a",token:"markup.quote"},{foreground:"24292e",fontStyle:"italic",token:"markup.italic"},{foreground:"24292e",fontStyle:"bold",token:"markup.bold"},{foreground:"005cc5",token:"markup.raw"},{foreground:"b31d28",background:"ffeef0",token:"markup.deleted"},{foreground:"b31d28",background:"ffeef0",token:"meta.diff.header.from-file"},{foreground:"b31d28",background:"ffeef0",token:"punctuation.definition.deleted"},{foreground:"22863a",background:"f0fff4",token:"markup.inserted"},{foreground:"22863a",background:"f0fff4",token:"meta.diff.header.to-file"},{foreground:"22863a",background:"f0fff4",token:"punctuation.definition.inserted"},{foreground:"e36209",background:"ffebda",token:"markup.changed"},{foreground:"e36209",background:"ffebda",token:"punctuation.definition.changed"},{foreground:"f6f8fa",background:"005cc5",token:"markup.ignored"},{foreground:"f6f8fa",background:"005cc5",token:"markup.untracked"},{foreground:"6f42c1",fontStyle:"bold",token:"meta.diff.range"},{foreground:"005cc5",token:"meta.diff.header"},{foreground:"005cc5",fontStyle:"bold",token:"meta.separator"},{foreground:"005cc5",token:"meta.output"},{foreground:"586069",token:"brackethighlighter.tag"},{foreground:"586069",token:"brackethighlighter.curly"},{foreground:"586069",token:"brackethighlighter.round"},{foreground:"586069",token:"brackethighlighter.square"},{foreground:"586069",token:"brackethighlighter.angle"},{foreground:"586069",token:"brackethighlighter.quote"},{foreground:"b31d28",token:"brackethighlighter.unmatched"},{foreground:"b31d28",token:"sublimelinter.mark.error"},{foreground:"e36209",token:"sublimelinter.mark.warning"},{foreground:"959da5",token:"sublimelinter.gutter-mark"},{foreground:"032f62",fontStyle:"underline",token:"constant.other.reference.link"},{foreground:"032f62",fontStyle:"underline",token:"string.other.link"}],colors:{"editor.foreground":"#24292e","editor.background":"#ffffff","editor.selectionBackground":"#c8c8fa","editor.inactiveSelectionBackground":"#fafbfc","editor.lineHighlightBackground":"#fafbfc","editorCursor.foreground":"#24292e","editorWhitespace.foreground":"#959da5","editorIndentGuide.background":"#959da5","editorIndentGuide.activeBackground":"#24292e","editor.selectionHighlightBorder":"#fafbfc"}},oze=120*1e3;let Uae=class{constructor(e){this._defaults=e,this._worker=null,this._client=null,this._idleCheckInterval=window.setInterval(()=>this._checkIfIdle(),30*1e3),this._lastUsedTime=0,this._configChangeListener=this._defaults.onDidChange(()=>this._stopWorker())}_stopWorker(){this._worker&&(this._worker.dispose(),this._worker=null),this._client=null}dispose(){clearInterval(this._idleCheckInterval),this._configChangeListener.dispose(),this._stopWorker()}_checkIfIdle(){if(!this._worker)return;Date.now()-this._lastUsedTime>oze&&this._stopWorker()}_getClient(){return this._lastUsedTime=Date.now(),this._client||(this._worker=Ex({moduleId:"vs/language/css/cssWorker",createWorker:()=>new Worker(new URL("/assets/css.worker-DBVD8oXr.js",import.meta.url),{type:"module"}),label:this._defaults.languageId,createData:{options:this._defaults.options,languageId:this._defaults.languageId}}),this._client=this._worker.getProxy()),this._client}getLanguageServiceWorker(...e){let t;return this._getClient().then(i=>{t=i}).then(i=>{if(this._worker)return this._worker.withSyncedResources(e)}).then(i=>t)}};var gZ;(function(s){function e(t){return typeof t=="string"}s.is=e})(gZ||(gZ={}));var _6;(function(s){function e(t){return typeof t=="string"}s.is=e})(_6||(_6={}));var fZ;(function(s){s.MIN_VALUE=-2147483648,s.MAX_VALUE=2147483647;function e(t){return typeof t=="number"&&s.MIN_VALUE<=t&&t<=s.MAX_VALUE}s.is=e})(fZ||(fZ={}));var TT;(function(s){s.MIN_VALUE=0,s.MAX_VALUE=2147483647;function e(t){return typeof t=="number"&&s.MIN_VALUE<=t&&t<=s.MAX_VALUE}s.is=e})(TT||(TT={}));var cl;(function(s){function e(i,n){return i===Number.MAX_VALUE&&(i=TT.MAX_VALUE),n===Number.MAX_VALUE&&(n=TT.MAX_VALUE),{line:i,character:n}}s.create=e;function t(i){let n=i;return re.objectLiteral(n)&&re.uinteger(n.line)&&re.uinteger(n.character)}s.is=t})(cl||(cl={}));var Fn;(function(s){function e(i,n,o,r){if(re.uinteger(i)&&re.uinteger(n)&&re.uinteger(o)&&re.uinteger(r))return{start:cl.create(i,n),end:cl.create(o,r)};if(cl.is(i)&&cl.is(n))return{start:i,end:n};throw new Error(`Range#create called with invalid arguments[${i}, ${n}, ${o}, ${r}]`)}s.create=e;function t(i){let n=i;return re.objectLiteral(n)&&cl.is(n.start)&&cl.is(n.end)}s.is=t})(Fn||(Fn={}));var NT;(function(s){function e(i,n){return{uri:i,range:n}}s.create=e;function t(i){let n=i;return re.objectLiteral(n)&&Fn.is(n.range)&&(re.string(n.uri)||re.undefined(n.uri))}s.is=t})(NT||(NT={}));var pZ;(function(s){function e(i,n,o,r){return{targetUri:i,targetRange:n,targetSelectionRange:o,originSelectionRange:r}}s.create=e;function t(i){let n=i;return re.objectLiteral(n)&&Fn.is(n.targetRange)&&re.string(n.targetUri)&&Fn.is(n.targetSelectionRange)&&(Fn.is(n.originSelectionRange)||re.undefined(n.originSelectionRange))}s.is=t})(pZ||(pZ={}));var b6;(function(s){function e(i,n,o,r){return{red:i,green:n,blue:o,alpha:r}}s.create=e;function t(i){const n=i;return re.objectLiteral(n)&&re.numberRange(n.red,0,1)&&re.numberRange(n.green,0,1)&&re.numberRange(n.blue,0,1)&&re.numberRange(n.alpha,0,1)}s.is=t})(b6||(b6={}));var mZ;(function(s){function e(i,n){return{range:i,color:n}}s.create=e;function t(i){const n=i;return re.objectLiteral(n)&&Fn.is(n.range)&&b6.is(n.color)}s.is=t})(mZ||(mZ={}));var _Z;(function(s){function e(i,n,o){return{label:i,textEdit:n,additionalTextEdits:o}}s.create=e;function t(i){const n=i;return re.objectLiteral(n)&&re.string(n.label)&&(re.undefined(n.textEdit)||Vv.is(n))&&(re.undefined(n.additionalTextEdits)||re.typedArray(n.additionalTextEdits,Vv.is))}s.is=t})(_Z||(_Z={}));var wy;(function(s){s.Comment="comment",s.Imports="imports",s.Region="region"})(wy||(wy={}));var bZ;(function(s){function e(i,n,o,r,a,l){const c={startLine:i,endLine:n};return re.defined(o)&&(c.startCharacter=o),re.defined(r)&&(c.endCharacter=r),re.defined(a)&&(c.kind=a),re.defined(l)&&(c.collapsedText=l),c}s.create=e;function t(i){const n=i;return re.objectLiteral(n)&&re.uinteger(n.startLine)&&re.uinteger(n.startLine)&&(re.undefined(n.startCharacter)||re.uinteger(n.startCharacter))&&(re.undefined(n.endCharacter)||re.uinteger(n.endCharacter))&&(re.undefined(n.kind)||re.string(n.kind))}s.is=t})(bZ||(bZ={}));var w6;(function(s){function e(i,n){return{location:i,message:n}}s.create=e;function t(i){let n=i;return re.defined(n)&&NT.is(n.location)&&re.string(n.message)}s.is=t})(w6||(w6={}));var $b;(function(s){s.Error=1,s.Warning=2,s.Information=3,s.Hint=4})($b||($b={}));var wZ;(function(s){s.Unnecessary=1,s.Deprecated=2})(wZ||(wZ={}));var vZ;(function(s){function e(t){const i=t;return re.objectLiteral(i)&&re.string(i.href)}s.is=e})(vZ||(vZ={}));var RT;(function(s){function e(i,n,o,r,a,l){let c={range:i,message:n};return re.defined(o)&&(c.severity=o),re.defined(r)&&(c.code=r),re.defined(a)&&(c.source=a),re.defined(l)&&(c.relatedInformation=l),c}s.create=e;function t(i){var n;let o=i;return re.defined(o)&&Fn.is(o.range)&&re.string(o.message)&&(re.number(o.severity)||re.undefined(o.severity))&&(re.integer(o.code)||re.string(o.code)||re.undefined(o.code))&&(re.undefined(o.codeDescription)||re.string((n=o.codeDescription)===null||n===void 0?void 0:n.href))&&(re.string(o.source)||re.undefined(o.source))&&(re.undefined(o.relatedInformation)||re.typedArray(o.relatedInformation,w6.is))}s.is=t})(RT||(RT={}));var Hv;(function(s){function e(i,n,...o){let r={title:i,command:n};return re.defined(o)&&o.length>0&&(r.arguments=o),r}s.create=e;function t(i){let n=i;return re.defined(n)&&re.string(n.title)&&re.string(n.command)}s.is=t})(Hv||(Hv={}));var Vv;(function(s){function e(o,r){return{range:o,newText:r}}s.replace=e;function t(o,r){return{range:{start:o,end:o},newText:r}}s.insert=t;function i(o){return{range:o,newText:""}}s.del=i;function n(o){const r=o;return re.objectLiteral(r)&&re.string(r.newText)&&Fn.is(r.range)}s.is=n})(Vv||(Vv={}));var v6;(function(s){function e(i,n,o){const r={label:i};return n!==void 0&&(r.needsConfirmation=n),o!==void 0&&(r.description=o),r}s.create=e;function t(i){const n=i;return re.objectLiteral(n)&&re.string(n.label)&&(re.boolean(n.needsConfirmation)||n.needsConfirmation===void 0)&&(re.string(n.description)||n.description===void 0)}s.is=t})(v6||(v6={}));var zv;(function(s){function e(t){const i=t;return re.string(i)}s.is=e})(zv||(zv={}));var CZ;(function(s){function e(o,r,a){return{range:o,newText:r,annotationId:a}}s.replace=e;function t(o,r,a){return{range:{start:o,end:o},newText:r,annotationId:a}}s.insert=t;function i(o,r){return{range:o,newText:"",annotationId:r}}s.del=i;function n(o){const r=o;return Vv.is(r)&&(v6.is(r.annotationId)||zv.is(r.annotationId))}s.is=n})(CZ||(CZ={}));var C6;(function(s){function e(i,n){return{textDocument:i,edits:n}}s.create=e;function t(i){let n=i;return re.defined(n)&&k6.is(n.textDocument)&&Array.isArray(n.edits)}s.is=t})(C6||(C6={}));var y6;(function(s){function e(i,n,o){let r={kind:"create",uri:i};return n!==void 0&&(n.overwrite!==void 0||n.ignoreIfExists!==void 0)&&(r.options=n),o!==void 0&&(r.annotationId=o),r}s.create=e;function t(i){let n=i;return n&&n.kind==="create"&&re.string(n.uri)&&(n.options===void 0||(n.options.overwrite===void 0||re.boolean(n.options.overwrite))&&(n.options.ignoreIfExists===void 0||re.boolean(n.options.ignoreIfExists)))&&(n.annotationId===void 0||zv.is(n.annotationId))}s.is=t})(y6||(y6={}));var S6;(function(s){function e(i,n,o,r){let a={kind:"rename",oldUri:i,newUri:n};return o!==void 0&&(o.overwrite!==void 0||o.ignoreIfExists!==void 0)&&(a.options=o),r!==void 0&&(a.annotationId=r),a}s.create=e;function t(i){let n=i;return n&&n.kind==="rename"&&re.string(n.oldUri)&&re.string(n.newUri)&&(n.options===void 0||(n.options.overwrite===void 0||re.boolean(n.options.overwrite))&&(n.options.ignoreIfExists===void 0||re.boolean(n.options.ignoreIfExists)))&&(n.annotationId===void 0||zv.is(n.annotationId))}s.is=t})(S6||(S6={}));var x6;(function(s){function e(i,n,o){let r={kind:"delete",uri:i};return n!==void 0&&(n.recursive!==void 0||n.ignoreIfNotExists!==void 0)&&(r.options=n),o!==void 0&&(r.annotationId=o),r}s.create=e;function t(i){let n=i;return n&&n.kind==="delete"&&re.string(n.uri)&&(n.options===void 0||(n.options.recursive===void 0||re.boolean(n.options.recursive))&&(n.options.ignoreIfNotExists===void 0||re.boolean(n.options.ignoreIfNotExists)))&&(n.annotationId===void 0||zv.is(n.annotationId))}s.is=t})(x6||(x6={}));var L6;(function(s){function e(t){let i=t;return i&&(i.changes!==void 0||i.documentChanges!==void 0)&&(i.documentChanges===void 0||i.documentChanges.every(n=>re.string(n.kind)?y6.is(n)||S6.is(n)||x6.is(n):C6.is(n)))}s.is=e})(L6||(L6={}));var yZ;(function(s){function e(i){return{uri:i}}s.create=e;function t(i){let n=i;return re.defined(n)&&re.string(n.uri)}s.is=t})(yZ||(yZ={}));var SZ;(function(s){function e(i,n){return{uri:i,version:n}}s.create=e;function t(i){let n=i;return re.defined(n)&&re.string(n.uri)&&re.integer(n.version)}s.is=t})(SZ||(SZ={}));var k6;(function(s){function e(i,n){return{uri:i,version:n}}s.create=e;function t(i){let n=i;return re.defined(n)&&re.string(n.uri)&&(n.version===null||re.integer(n.version))}s.is=t})(k6||(k6={}));var xZ;(function(s){function e(i,n,o,r){return{uri:i,languageId:n,version:o,text:r}}s.create=e;function t(i){let n=i;return re.defined(n)&&re.string(n.uri)&&re.string(n.languageId)&&re.integer(n.version)&&re.string(n.text)}s.is=t})(xZ||(xZ={}));var E6;(function(s){s.PlainText="plaintext",s.Markdown="markdown";function e(t){const i=t;return i===s.PlainText||i===s.Markdown}s.is=e})(E6||(E6={}));var A0;(function(s){function e(t){const i=t;return re.objectLiteral(t)&&E6.is(i.kind)&&re.string(i.value)}s.is=e})(A0||(A0={}));var jn;(function(s){s.Text=1,s.Method=2,s.Function=3,s.Constructor=4,s.Field=5,s.Variable=6,s.Class=7,s.Interface=8,s.Module=9,s.Property=10,s.Unit=11,s.Value=12,s.Enum=13,s.Keyword=14,s.Snippet=15,s.Color=16,s.File=17,s.Reference=18,s.Folder=19,s.EnumMember=20,s.Constant=21,s.Struct=22,s.Event=23,s.Operator=24,s.TypeParameter=25})(jn||(jn={}));var I6;(function(s){s.PlainText=1,s.Snippet=2})(I6||(I6={}));var LZ;(function(s){s.Deprecated=1})(LZ||(LZ={}));var kZ;(function(s){function e(i,n,o){return{newText:i,insert:n,replace:o}}s.create=e;function t(i){const n=i;return n&&re.string(n.newText)&&Fn.is(n.insert)&&Fn.is(n.replace)}s.is=t})(kZ||(kZ={}));var EZ;(function(s){s.asIs=1,s.adjustIndentation=2})(EZ||(EZ={}));var IZ;(function(s){function e(t){const i=t;return i&&(re.string(i.detail)||i.detail===void 0)&&(re.string(i.description)||i.description===void 0)}s.is=e})(IZ||(IZ={}));var DZ;(function(s){function e(t){return{label:t}}s.create=e})(DZ||(DZ={}));var TZ;(function(s){function e(t,i){return{items:t||[],isIncomplete:!!i}}s.create=e})(TZ||(TZ={}));var AT;(function(s){function e(i){return i.replace(/[\\`*_{}[\]()#+\-.!]/g,"\\$&")}s.fromPlainText=e;function t(i){const n=i;return re.string(n)||re.objectLiteral(n)&&re.string(n.language)&&re.string(n.value)}s.is=t})(AT||(AT={}));var NZ;(function(s){function e(t){let i=t;return!!i&&re.objectLiteral(i)&&(A0.is(i.contents)||AT.is(i.contents)||re.typedArray(i.contents,AT.is))&&(t.range===void 0||Fn.is(t.range))}s.is=e})(NZ||(NZ={}));var RZ;(function(s){function e(t,i){return i?{label:t,documentation:i}:{label:t}}s.create=e})(RZ||(RZ={}));var AZ;(function(s){function e(t,i,...n){let o={label:t};return re.defined(i)&&(o.documentation=i),re.defined(n)?o.parameters=n:o.parameters=[],o}s.create=e})(AZ||(AZ={}));var vy;(function(s){s.Text=1,s.Read=2,s.Write=3})(vy||(vy={}));var MZ;(function(s){function e(t,i){let n={range:t};return re.number(i)&&(n.kind=i),n}s.create=e})(MZ||(MZ={}));var Gn;(function(s){s.File=1,s.Module=2,s.Namespace=3,s.Package=4,s.Class=5,s.Method=6,s.Property=7,s.Field=8,s.Constructor=9,s.Enum=10,s.Interface=11,s.Function=12,s.Variable=13,s.Constant=14,s.String=15,s.Number=16,s.Boolean=17,s.Array=18,s.Object=19,s.Key=20,s.Null=21,s.EnumMember=22,s.Struct=23,s.Event=24,s.Operator=25,s.TypeParameter=26})(Gn||(Gn={}));var OZ;(function(s){s.Deprecated=1})(OZ||(OZ={}));var PZ;(function(s){function e(t,i,n,o,r){let a={name:t,kind:i,location:{uri:o,range:n}};return r&&(a.containerName=r),a}s.create=e})(PZ||(PZ={}));var FZ;(function(s){function e(t,i,n,o){return o!==void 0?{name:t,kind:i,location:{uri:n,range:o}}:{name:t,kind:i,location:{uri:n}}}s.create=e})(FZ||(FZ={}));var BZ;(function(s){function e(i,n,o,r,a,l){let c={name:i,detail:n,kind:o,range:r,selectionRange:a};return l!==void 0&&(c.children=l),c}s.create=e;function t(i){let n=i;return n&&re.string(n.name)&&re.number(n.kind)&&Fn.is(n.range)&&Fn.is(n.selectionRange)&&(n.detail===void 0||re.string(n.detail))&&(n.deprecated===void 0||re.boolean(n.deprecated))&&(n.children===void 0||Array.isArray(n.children))&&(n.tags===void 0||Array.isArray(n.tags))}s.is=t})(BZ||(BZ={}));var WZ;(function(s){s.Empty="",s.QuickFix="quickfix",s.Refactor="refactor",s.RefactorExtract="refactor.extract",s.RefactorInline="refactor.inline",s.RefactorRewrite="refactor.rewrite",s.Source="source",s.SourceOrganizeImports="source.organizeImports",s.SourceFixAll="source.fixAll"})(WZ||(WZ={}));var MT;(function(s){s.Invoked=1,s.Automatic=2})(MT||(MT={}));var HZ;(function(s){function e(i,n,o){let r={diagnostics:i};return n!=null&&(r.only=n),o!=null&&(r.triggerKind=o),r}s.create=e;function t(i){let n=i;return re.defined(n)&&re.typedArray(n.diagnostics,RT.is)&&(n.only===void 0||re.typedArray(n.only,re.string))&&(n.triggerKind===void 0||n.triggerKind===MT.Invoked||n.triggerKind===MT.Automatic)}s.is=t})(HZ||(HZ={}));var VZ;(function(s){function e(i,n,o){let r={title:i},a=!0;return typeof n=="string"?(a=!1,r.kind=n):Hv.is(n)?r.command=n:r.edit=n,a&&o!==void 0&&(r.kind=o),r}s.create=e;function t(i){let n=i;return n&&re.string(n.title)&&(n.diagnostics===void 0||re.typedArray(n.diagnostics,RT.is))&&(n.kind===void 0||re.string(n.kind))&&(n.edit!==void 0||n.command!==void 0)&&(n.command===void 0||Hv.is(n.command))&&(n.isPreferred===void 0||re.boolean(n.isPreferred))&&(n.edit===void 0||L6.is(n.edit))}s.is=t})(VZ||(VZ={}));var zZ;(function(s){function e(i,n){let o={range:i};return re.defined(n)&&(o.data=n),o}s.create=e;function t(i){let n=i;return re.defined(n)&&Fn.is(n.range)&&(re.undefined(n.command)||Hv.is(n.command))}s.is=t})(zZ||(zZ={}));var $Z;(function(s){function e(i,n){return{tabSize:i,insertSpaces:n}}s.create=e;function t(i){let n=i;return re.defined(n)&&re.uinteger(n.tabSize)&&re.boolean(n.insertSpaces)}s.is=t})($Z||($Z={}));var UZ;(function(s){function e(i,n,o){return{range:i,target:n,data:o}}s.create=e;function t(i){let n=i;return re.defined(n)&&Fn.is(n.range)&&(re.undefined(n.target)||re.string(n.target))}s.is=t})(UZ||(UZ={}));var qZ;(function(s){function e(i,n){return{range:i,parent:n}}s.create=e;function t(i){let n=i;return re.objectLiteral(n)&&Fn.is(n.range)&&(n.parent===void 0||s.is(n.parent))}s.is=t})(qZ||(qZ={}));var jZ;(function(s){s.namespace="namespace",s.type="type",s.class="class",s.enum="enum",s.interface="interface",s.struct="struct",s.typeParameter="typeParameter",s.parameter="parameter",s.variable="variable",s.property="property",s.enumMember="enumMember",s.event="event",s.function="function",s.method="method",s.macro="macro",s.keyword="keyword",s.modifier="modifier",s.comment="comment",s.string="string",s.number="number",s.regexp="regexp",s.operator="operator",s.decorator="decorator"})(jZ||(jZ={}));var GZ;(function(s){s.declaration="declaration",s.definition="definition",s.readonly="readonly",s.static="static",s.deprecated="deprecated",s.abstract="abstract",s.async="async",s.modification="modification",s.documentation="documentation",s.defaultLibrary="defaultLibrary"})(GZ||(GZ={}));var KZ;(function(s){function e(t){const i=t;return re.objectLiteral(i)&&(i.resultId===void 0||typeof i.resultId=="string")&&Array.isArray(i.data)&&(i.data.length===0||typeof i.data[0]=="number")}s.is=e})(KZ||(KZ={}));var YZ;(function(s){function e(i,n){return{range:i,text:n}}s.create=e;function t(i){const n=i;return n!=null&&Fn.is(n.range)&&re.string(n.text)}s.is=t})(YZ||(YZ={}));var ZZ;(function(s){function e(i,n,o){return{range:i,variableName:n,caseSensitiveLookup:o}}s.create=e;function t(i){const n=i;return n!=null&&Fn.is(n.range)&&re.boolean(n.caseSensitiveLookup)&&(re.string(n.variableName)||n.variableName===void 0)}s.is=t})(ZZ||(ZZ={}));var XZ;(function(s){function e(i,n){return{range:i,expression:n}}s.create=e;function t(i){const n=i;return n!=null&&Fn.is(n.range)&&(re.string(n.expression)||n.expression===void 0)}s.is=t})(XZ||(XZ={}));var QZ;(function(s){function e(i,n){return{frameId:i,stoppedLocation:n}}s.create=e;function t(i){const n=i;return re.defined(n)&&Fn.is(i.stoppedLocation)}s.is=t})(QZ||(QZ={}));var D6;(function(s){s.Type=1,s.Parameter=2;function e(t){return t===1||t===2}s.is=e})(D6||(D6={}));var T6;(function(s){function e(i){return{value:i}}s.create=e;function t(i){const n=i;return re.objectLiteral(n)&&(n.tooltip===void 0||re.string(n.tooltip)||A0.is(n.tooltip))&&(n.location===void 0||NT.is(n.location))&&(n.command===void 0||Hv.is(n.command))}s.is=t})(T6||(T6={}));var JZ;(function(s){function e(i,n,o){const r={position:i,label:n};return o!==void 0&&(r.kind=o),r}s.create=e;function t(i){const n=i;return re.objectLiteral(n)&&cl.is(n.position)&&(re.string(n.label)||re.typedArray(n.label,T6.is))&&(n.kind===void 0||D6.is(n.kind))&&n.textEdits===void 0||re.typedArray(n.textEdits,Vv.is)&&(n.tooltip===void 0||re.string(n.tooltip)||A0.is(n.tooltip))&&(n.paddingLeft===void 0||re.boolean(n.paddingLeft))&&(n.paddingRight===void 0||re.boolean(n.paddingRight))}s.is=t})(JZ||(JZ={}));var eX;(function(s){function e(t){return{kind:"snippet",value:t}}s.createSnippet=e})(eX||(eX={}));var tX;(function(s){function e(t,i,n,o){return{insertText:t,filterText:i,range:n,command:o}}s.create=e})(tX||(tX={}));var iX;(function(s){function e(t){return{items:t}}s.create=e})(iX||(iX={}));var nX;(function(s){s.Invoked=0,s.Automatic=1})(nX||(nX={}));var sX;(function(s){function e(t,i){return{range:t,text:i}}s.create=e})(sX||(sX={}));var oX;(function(s){function e(t,i){return{triggerKind:t,selectedCompletionInfo:i}}s.create=e})(oX||(oX={}));var rX;(function(s){function e(t){const i=t;return re.objectLiteral(i)&&_6.is(i.uri)&&re.string(i.name)}s.is=e})(rX||(rX={}));var aX;(function(s){function e(o,r,a,l){return new rze(o,r,a,l)}s.create=e;function t(o){let r=o;return!!(re.defined(r)&&re.string(r.uri)&&(re.undefined(r.languageId)||re.string(r.languageId))&&re.uinteger(r.lineCount)&&re.func(r.getText)&&re.func(r.positionAt)&&re.func(r.offsetAt))}s.is=t;function i(o,r){let a=o.getText(),l=n(r,(d,h)=>{let u=d.range.start.line-h.range.start.line;return u===0?d.range.start.character-h.range.start.character:u}),c=a.length;for(let d=l.length-1;d>=0;d--){let h=l[d],u=o.offsetAt(h.range.start),g=o.offsetAt(h.range.end);if(g<=c)a=a.substring(0,u)+h.newText+a.substring(g,a.length);else throw new Error("Overlapping edit");c=u}return a}s.applyEdits=i;function n(o,r){if(o.length<=1)return o;const a=o.length/2|0,l=o.slice(0,a),c=o.slice(a);n(l,r),n(c,r);let d=0,h=0,u=0;for(;d<l.length&&h<c.length;)r(l[d],c[h])<=0?o[u++]=l[d++]:o[u++]=c[h++];for(;d<l.length;)o[u++]=l[d++];for(;h<c.length;)o[u++]=c[h++];return o}})(aX||(aX={}));class rze{constructor(e,t,i,n){this._uri=e,this._languageId=t,this._version=i,this._content=n,this._lineOffsets=void 0}get uri(){return this._uri}get languageId(){return this._languageId}get version(){return this._version}getText(e){if(e){let t=this.offsetAt(e.start),i=this.offsetAt(e.end);return this._content.substring(t,i)}return this._content}update(e,t){this._content=e.text,this._version=t,this._lineOffsets=void 0}getLineOffsets(){if(this._lineOffsets===void 0){let e=[],t=this._content,i=!0;for(let n=0;n<t.length;n++){i&&(e.push(n),i=!1);let o=t.charAt(n);i=o==="\r"||o===`
`,o==="\r"&&n+1<t.length&&t.charAt(n+1)===`
`&&n++}i&&t.length>0&&e.push(t.length),this._lineOffsets=e}return this._lineOffsets}positionAt(e){e=Math.max(Math.min(e,this._content.length),0);let t=this.getLineOffsets(),i=0,n=t.length;if(n===0)return cl.create(0,e);for(;i<n;){let r=Math.floor((i+n)/2);t[r]>e?n=r:i=r+1}let o=i-1;return cl.create(o,e-t[o])}offsetAt(e){let t=this.getLineOffsets();if(e.line>=t.length)return this._content.length;if(e.line<0)return 0;let i=t[e.line],n=e.line+1<t.length?t[e.line+1]:this._content.length;return Math.max(Math.min(i+e.character,n),i)}get lineCount(){return this.getLineOffsets().length}}var re;(function(s){const e=Object.prototype.toString;function t(g){return typeof g<"u"}s.defined=t;function i(g){return typeof g>"u"}s.undefined=i;function n(g){return g===!0||g===!1}s.boolean=n;function o(g){return e.call(g)==="[object String]"}s.string=o;function r(g){return e.call(g)==="[object Number]"}s.number=r;function a(g,f,p){return e.call(g)==="[object Number]"&&f<=g&&g<=p}s.numberRange=a;function l(g){return e.call(g)==="[object Number]"&&-2147483648<=g&&g<=2147483647}s.integer=l;function c(g){return e.call(g)==="[object Number]"&&0<=g&&g<=2147483647}s.uinteger=c;function d(g){return e.call(g)==="[object Function]"}s.func=d;function h(g){return g!==null&&typeof g=="object"}s.objectLiteral=h;function u(g,f){return Array.isArray(g)&&g.every(f)}s.typedArray=u})(re||(re={}));let Ix=class{constructor(e,t,i){this._languageId=e,this._worker=t,this._disposables=[],this._listener=Object.create(null);const n=r=>{let a=r.getLanguageId();if(a!==this._languageId)return;let l;this._listener[r.uri.toString()]=r.onDidChangeContent(()=>{window.clearTimeout(l),l=window.setTimeout(()=>this._doValidate(r.uri,a),500)}),this._doValidate(r.uri,a)},o=r=>{ti.setModelMarkers(r,this._languageId,[]);let a=r.uri.toString(),l=this._listener[a];l&&(l.dispose(),delete this._listener[a])};this._disposables.push(ti.onDidCreateModel(n)),this._disposables.push(ti.onWillDisposeModel(o)),this._disposables.push(ti.onDidChangeModelLanguage(r=>{o(r.model),n(r.model)})),this._disposables.push(i(r=>{ti.getModels().forEach(a=>{a.getLanguageId()===this._languageId&&(o(a),n(a))})})),this._disposables.push({dispose:()=>{ti.getModels().forEach(o);for(let r in this._listener)this._listener[r].dispose()}}),ti.getModels().forEach(n)}dispose(){this._disposables.forEach(e=>e&&e.dispose()),this._disposables.length=0}_doValidate(e,t){this._worker(e).then(i=>i.doValidation(e.toString())).then(i=>{const n=i.map(r=>lze(e,r));let o=ti.getModel(e);o&&o.getLanguageId()===t&&ti.setModelMarkers(o,t,n)}).then(void 0,i=>{console.error(i)})}};function aze(s){switch(s){case $b.Error:return kn.Error;case $b.Warning:return kn.Warning;case $b.Information:return kn.Info;case $b.Hint:return kn.Hint;default:return kn.Info}}function lze(s,e){let t=typeof e.code=="number"?String(e.code):e.code;return{severity:aze(e.severity),startLineNumber:e.range.start.line+1,startColumn:e.range.start.character+1,endLineNumber:e.range.end.line+1,endColumn:e.range.end.character+1,message:e.message,code:t,source:e.source}}class mC{constructor(e,t){this._worker=e,this._triggerCharacters=t}get triggerCharacters(){return this._triggerCharacters}provideCompletionItems(e,t,i,n){const o=e.uri;return this._worker(o).then(r=>r.doComplete(o.toString(),Zc(t))).then(r=>{if(!r)return;const a=e.getWordUntilPosition(t),l=new Hc(t.lineNumber,a.startColumn,t.lineNumber,a.endColumn),c=r.items.map(d=>{const h={label:d.label,insertText:d.insertText||d.label,sortText:d.sortText,filterText:d.filterText,documentation:d.documentation,detail:d.detail,command:hze(d.command),range:l,kind:dze(d.kind)};return d.textEdit&&(cze(d.textEdit)?h.range={insert:Os(d.textEdit.insert),replace:Os(d.textEdit.replace)}:h.range=Os(d.textEdit.range),h.insertText=d.textEdit.newText),d.additionalTextEdits&&(h.additionalTextEdits=d.additionalTextEdits.map(Lg)),d.insertTextFormat===I6.Snippet&&(h.insertTextRules=H.CompletionItemInsertTextRule.InsertAsSnippet),h});return{isIncomplete:r.isIncomplete,suggestions:c}})}}function Zc(s){if(s)return{character:s.column-1,line:s.lineNumber-1}}function Dx(s){if(s)return{start:{line:s.startLineNumber-1,character:s.startColumn-1},end:{line:s.endLineNumber-1,character:s.endColumn-1}}}function Os(s){if(s)return new Hc(s.start.line+1,s.start.character+1,s.end.line+1,s.end.character+1)}function cze(s){return typeof s.insert<"u"&&typeof s.replace<"u"}function dze(s){const e=H.CompletionItemKind;switch(s){case jn.Text:return e.Text;case jn.Method:return e.Method;case jn.Function:return e.Function;case jn.Constructor:return e.Constructor;case jn.Field:return e.Field;case jn.Variable:return e.Variable;case jn.Class:return e.Class;case jn.Interface:return e.Interface;case jn.Module:return e.Module;case jn.Property:return e.Property;case jn.Unit:return e.Unit;case jn.Value:return e.Value;case jn.Enum:return e.Enum;case jn.Keyword:return e.Keyword;case jn.Snippet:return e.Snippet;case jn.Color:return e.Color;case jn.File:return e.File;case jn.Reference:return e.Reference}return e.Property}function Lg(s){if(s)return{range:Os(s.range),text:s.newText}}function hze(s){return s&&s.command==="editor.action.triggerSuggest"?{id:s.command,title:s.title,arguments:s.arguments}:void 0}class S_{constructor(e){this._worker=e}provideHover(e,t,i){let n=e.uri;return this._worker(n).then(o=>o.doHover(n.toString(),Zc(t))).then(o=>{if(o)return{range:Os(o.range),contents:gze(o.contents)}})}}function uze(s){return s&&typeof s=="object"&&typeof s.kind=="string"}function lX(s){return typeof s=="string"?{value:s}:uze(s)?s.kind==="plaintext"?{value:s.value.replace(/[\\`*_{}[\]()#+\-.!]/g,"\\$&")}:{value:s.value}:{value:"```"+s.language+`
`+s.value+"\n```\n"}}function gze(s){if(s)return Array.isArray(s)?s.map(lX):[lX(s)]}let _C=class{constructor(e){this._worker=e}provideDocumentHighlights(e,t,i){const n=e.uri;return this._worker(n).then(o=>o.findDocumentHighlights(n.toString(),Zc(t))).then(o=>{if(o)return o.map(r=>({range:Os(r.range),kind:fze(r.kind)}))})}};function fze(s){switch(s){case vy.Read:return H.DocumentHighlightKind.Read;case vy.Write:return H.DocumentHighlightKind.Write;case vy.Text:return H.DocumentHighlightKind.Text}return H.DocumentHighlightKind.Text}let u2=class{constructor(e){this._worker=e}provideDefinition(e,t,i){const n=e.uri;return this._worker(n).then(o=>o.findDefinition(n.toString(),Zc(t))).then(o=>{if(o)return[qae(o)]})}};function qae(s){return{uri:Yc.parse(s.uri),range:Os(s.range)}}let g2=class{constructor(e){this._worker=e}provideReferences(e,t,i,n){const o=e.uri;return this._worker(o).then(r=>r.findReferences(o.toString(),Zc(t))).then(r=>{if(r)return r.map(qae)})}},bC=class{constructor(e){this._worker=e}provideRenameEdits(e,t,i,n){const o=e.uri;return this._worker(o).then(r=>r.doRename(o.toString(),Zc(t),i)).then(r=>pze(r))}};function pze(s){if(!s||!s.changes)return;let e=[];for(let t in s.changes){const i=Yc.parse(t);for(let n of s.changes[t])e.push({resource:i,versionId:void 0,textEdit:{range:Os(n.range),text:n.newText}})}return{edits:e}}class x_{constructor(e){this._worker=e}provideDocumentSymbols(e,t){const i=e.uri;return this._worker(i).then(n=>n.findDocumentSymbols(i.toString())).then(n=>{if(n)return n.map(o=>mze(o)?jae(o):{name:o.name,detail:"",containerName:o.containerName,kind:Gae(o.kind),range:Os(o.location.range),selectionRange:Os(o.location.range),tags:[]})})}}function mze(s){return"children"in s}function jae(s){return{name:s.name,detail:s.detail??"",kind:Gae(s.kind),range:Os(s.range),selectionRange:Os(s.selectionRange),tags:s.tags??[],children:(s.children??[]).map(e=>jae(e))}}function Gae(s){let e=H.SymbolKind;switch(s){case Gn.File:return e.File;case Gn.Module:return e.Module;case Gn.Namespace:return e.Namespace;case Gn.Package:return e.Package;case Gn.Class:return e.Class;case Gn.Method:return e.Method;case Gn.Property:return e.Property;case Gn.Field:return e.Field;case Gn.Constructor:return e.Constructor;case Gn.Enum:return e.Enum;case Gn.Interface:return e.Interface;case Gn.Function:return e.Function;case Gn.Variable:return e.Variable;case Gn.Constant:return e.Constant;case Gn.String:return e.String;case Gn.Number:return e.Number;case Gn.Boolean:return e.Boolean;case Gn.Array:return e.Array}return e.Function}class Tx{constructor(e){this._worker=e}provideLinks(e,t){const i=e.uri;return this._worker(i).then(n=>n.findDocumentLinks(i.toString())).then(n=>{if(n)return{links:n.map(o=>({range:Os(o.range),url:o.target}))}})}}class L_{constructor(e){this._worker=e}provideDocumentFormattingEdits(e,t,i){const n=e.uri;return this._worker(n).then(o=>o.format(n.toString(),null,Kae(t)).then(r=>{if(!(!r||r.length===0))return r.map(Lg)}))}}class k_{constructor(e){this._worker=e,this.canFormatMultipleRanges=!1}provideDocumentRangeFormattingEdits(e,t,i,n){const o=e.uri;return this._worker(o).then(r=>r.format(o.toString(),Dx(t),Kae(i)).then(a=>{if(!(!a||a.length===0))return a.map(Lg)}))}}function Kae(s){return{tabSize:s.tabSize,insertSpaces:s.insertSpaces}}class Nx{constructor(e){this._worker=e}provideDocumentColors(e,t){const i=e.uri;return this._worker(i).then(n=>n.findDocumentColors(i.toString())).then(n=>{if(n)return n.map(o=>({color:o.color,range:Os(o.range)}))})}provideColorPresentations(e,t,i){const n=e.uri;return this._worker(n).then(o=>o.getColorPresentations(n.toString(),t.color,Dx(t.range))).then(o=>{if(o)return o.map(r=>{let a={label:r.label};return r.textEdit&&(a.textEdit=Lg(r.textEdit)),r.additionalTextEdits&&(a.additionalTextEdits=r.additionalTextEdits.map(Lg)),a})})}}class E_{constructor(e){this._worker=e}provideFoldingRanges(e,t,i){const n=e.uri;return this._worker(n).then(o=>o.getFoldingRanges(n.toString(),t)).then(o=>{if(o)return o.map(r=>{const a={start:r.startLine+1,end:r.endLine+1};return typeof r.kind<"u"&&(a.kind=_ze(r.kind)),a})})}}function _ze(s){switch(s){case wy.Comment:return H.FoldingRangeKind.Comment;case wy.Imports:return H.FoldingRangeKind.Imports;case wy.Region:return H.FoldingRangeKind.Region}}class I_{constructor(e){this._worker=e}provideSelectionRanges(e,t,i){const n=e.uri;return this._worker(n).then(o=>o.getSelectionRanges(n.toString(),t.map(Zc))).then(o=>{if(o)return o.map(r=>{const a=[];for(;r;)a.push({range:Os(r.range)}),r=r.parent;return a})})}}function bze(s){const e=[],t=[],i=new Uae(s);e.push(i);const n=(...r)=>i.getLanguageServiceWorker(...r);function o(){const{languageId:r,modeConfiguration:a}=s;Yae(t),a.completionItems&&t.push(H.registerCompletionItemProvider(r,new mC(n,["/","-",":"]))),a.hovers&&t.push(H.registerHoverProvider(r,new S_(n))),a.documentHighlights&&t.push(H.registerDocumentHighlightProvider(r,new _C(n))),a.definitions&&t.push(H.registerDefinitionProvider(r,new u2(n))),a.references&&t.push(H.registerReferenceProvider(r,new g2(n))),a.documentSymbols&&t.push(H.registerDocumentSymbolProvider(r,new x_(n))),a.rename&&t.push(H.registerRenameProvider(r,new bC(n))),a.colors&&t.push(H.registerColorProvider(r,new Nx(n))),a.foldingRanges&&t.push(H.registerFoldingRangeProvider(r,new E_(n))),a.diagnostics&&t.push(new Ix(r,n,s.onDidChange)),a.selectionRanges&&t.push(H.registerSelectionRangeProvider(r,new I_(n))),a.documentFormattingEdits&&t.push(H.registerDocumentFormattingEditProvider(r,new L_(n))),a.documentRangeFormattingEdits&&t.push(H.registerDocumentRangeFormattingEditProvider(r,new k_(n)))}return o(),e.push(cX(t)),cX(e)}function cX(s){return{dispose:()=>Yae(s)}}function Yae(s){for(;s.length;)s.pop().dispose()}const wze=Object.freeze(Object.defineProperty({__proto__:null,CompletionAdapter:mC,DefinitionAdapter:u2,DiagnosticsAdapter:Ix,DocumentColorAdapter:Nx,DocumentFormattingEditProvider:L_,DocumentHighlightAdapter:_C,DocumentLinkAdapter:Tx,DocumentRangeFormattingEditProvider:k_,DocumentSymbolAdapter:x_,FoldingRangeAdapter:E_,HoverAdapter:S_,ReferenceAdapter:g2,RenameAdapter:bC,SelectionRangeAdapter:I_,WorkerManager:Uae,fromPosition:Zc,fromRange:Dx,setupMode:bze,toRange:Os,toTextEdit:Lg},Symbol.toStringTag,{value:"Module"})),vze=120*1e3;let MV=class{constructor(e){this._defaults=e,this._worker=null,this._client=null,this._idleCheckInterval=window.setInterval(()=>this._checkIfIdle(),30*1e3),this._lastUsedTime=0,this._configChangeListener=this._defaults.onDidChange(()=>this._stopWorker())}_stopWorker(){this._worker&&(this._worker.dispose(),this._worker=null),this._client=null}dispose(){clearInterval(this._idleCheckInterval),this._configChangeListener.dispose(),this._stopWorker()}_checkIfIdle(){if(!this._worker)return;Date.now()-this._lastUsedTime>vze&&this._stopWorker()}_getClient(){return this._lastUsedTime=Date.now(),this._client||(this._worker=Ex({moduleId:"vs/language/html/htmlWorker",createWorker:()=>new Worker(new URL("/assets/html.worker-CwpTb9lJ.js",import.meta.url),{type:"module"}),createData:{languageSettings:this._defaults.options,languageId:this._defaults.languageId},label:this._defaults.languageId}),this._client=this._worker.getProxy()),this._client}getLanguageServiceWorker(...e){let t;return this._getClient().then(i=>{t=i}).then(i=>{if(this._worker)return this._worker.withSyncedResources(e)}).then(i=>t)}};class Zae extends mC{constructor(e){super(e,[".",":","<",'"',"=","/"])}}function Cze(s){const e=new MV(s),t=(...n)=>e.getLanguageServiceWorker(...n);let i=s.languageId;H.registerCompletionItemProvider(i,new Zae(t)),H.registerHoverProvider(i,new S_(t)),H.registerDocumentHighlightProvider(i,new _C(t)),H.registerLinkProvider(i,new Tx(t)),H.registerFoldingRangeProvider(i,new E_(t)),H.registerDocumentSymbolProvider(i,new x_(t)),H.registerSelectionRangeProvider(i,new I_(t)),H.registerRenameProvider(i,new bC(t)),i==="html"&&(H.registerDocumentFormattingEditProvider(i,new L_(t)),H.registerDocumentRangeFormattingEditProvider(i,new k_(t)))}function yze(s){const e=[],t=[],i=new MV(s);e.push(i);const n=(...r)=>i.getLanguageServiceWorker(...r);function o(){const{languageId:r,modeConfiguration:a}=s;Xae(t),a.completionItems&&t.push(H.registerCompletionItemProvider(r,new Zae(n))),a.hovers&&t.push(H.registerHoverProvider(r,new S_(n))),a.documentHighlights&&t.push(H.registerDocumentHighlightProvider(r,new _C(n))),a.links&&t.push(H.registerLinkProvider(r,new Tx(n))),a.documentSymbols&&t.push(H.registerDocumentSymbolProvider(r,new x_(n))),a.rename&&t.push(H.registerRenameProvider(r,new bC(n))),a.foldingRanges&&t.push(H.registerFoldingRangeProvider(r,new E_(n))),a.selectionRanges&&t.push(H.registerSelectionRangeProvider(r,new I_(n))),a.documentFormattingEdits&&t.push(H.registerDocumentFormattingEditProvider(r,new L_(n))),a.documentRangeFormattingEdits&&t.push(H.registerDocumentRangeFormattingEditProvider(r,new k_(n)))}return o(),e.push(dX(t)),dX(e)}function dX(s){return{dispose:()=>Xae(s)}}function Xae(s){for(;s.length;)s.pop().dispose()}const Sze=Object.freeze(Object.defineProperty({__proto__:null,CompletionAdapter:mC,DefinitionAdapter:u2,DiagnosticsAdapter:Ix,DocumentColorAdapter:Nx,DocumentFormattingEditProvider:L_,DocumentHighlightAdapter:_C,DocumentLinkAdapter:Tx,DocumentRangeFormattingEditProvider:k_,DocumentSymbolAdapter:x_,FoldingRangeAdapter:E_,HoverAdapter:S_,ReferenceAdapter:g2,RenameAdapter:bC,SelectionRangeAdapter:I_,WorkerManager:MV,fromPosition:Zc,fromRange:Dx,setupMode:yze,setupMode1:Cze,toRange:Os,toTextEdit:Lg},Symbol.toStringTag,{value:"Module"})),xze=120*1e3;let Qae=class{constructor(e){this._defaults=e,this._worker=null,this._client=null,this._idleCheckInterval=window.setInterval(()=>this._checkIfIdle(),30*1e3),this._lastUsedTime=0,this._configChangeListener=this._defaults.onDidChange(()=>this._stopWorker())}_stopWorker(){this._worker&&(this._worker.dispose(),this._worker=null),this._client=null}dispose(){clearInterval(this._idleCheckInterval),this._configChangeListener.dispose(),this._stopWorker()}_checkIfIdle(){if(!this._worker)return;Date.now()-this._lastUsedTime>xze&&this._stopWorker()}_getClient(){return this._lastUsedTime=Date.now(),this._client||(this._worker=Ex({moduleId:"vs/language/json/jsonWorker",createWorker:()=>new Worker(new URL("/assets/json.worker-BoL8UZqY.js",import.meta.url),{type:"module"}),label:this._defaults.languageId,createData:{languageSettings:this._defaults.diagnosticsOptions,languageId:this._defaults.languageId,enableSchemaRequest:this._defaults.diagnosticsOptions.enableSchemaRequest}}),this._client=this._worker.getProxy()),this._client}getLanguageServiceWorker(...e){let t;return this._getClient().then(i=>{t=i}).then(i=>{if(this._worker)return this._worker.withSyncedResources(e)}).then(i=>t)}};function Lze(s,e=!1){const t=s.length;let i=0,n="",o=0,r=16,a=0,l=0,c=0,d=0,h=0;function u(C,v){let y=0,x=0;for(;y<C;){let L=s.charCodeAt(i);if(L>=48&&L<=57)x=x*16+L-48;else if(L>=65&&L<=70)x=x*16+L-65+10;else if(L>=97&&L<=102)x=x*16+L-97+10;else break;i++,y++}return y<C&&(x=-1),x}function g(C){i=C,n="",o=0,r=16,h=0}function f(){let C=i;if(s.charCodeAt(i)===48)i++;else for(i++;i<s.length&&lb(s.charCodeAt(i));)i++;if(i<s.length&&s.charCodeAt(i)===46)if(i++,i<s.length&&lb(s.charCodeAt(i)))for(i++;i<s.length&&lb(s.charCodeAt(i));)i++;else return h=3,s.substring(C,i);let v=i;if(i<s.length&&(s.charCodeAt(i)===69||s.charCodeAt(i)===101))if(i++,(i<s.length&&s.charCodeAt(i)===43||s.charCodeAt(i)===45)&&i++,i<s.length&&lb(s.charCodeAt(i))){for(i++;i<s.length&&lb(s.charCodeAt(i));)i++;v=i}else h=3;return s.substring(C,v)}function p(){let C="",v=i;for(;;){if(i>=t){C+=s.substring(v,i),h=2;break}const y=s.charCodeAt(i);if(y===34){C+=s.substring(v,i),i++;break}if(y===92){if(C+=s.substring(v,i),i++,i>=t){h=2;break}switch(s.charCodeAt(i++)){case 34:C+='"';break;case 92:C+="\\";break;case 47:C+="/";break;case 98:C+="\b";break;case 102:C+="\f";break;case 110:C+=`
`;break;case 114:C+="\r";break;case 116:C+="	";break;case 117:const L=u(4);L>=0?C+=String.fromCharCode(L):h=4;break;default:h=5}v=i;continue}if(y>=0&&y<=31)if(n1(y)){C+=s.substring(v,i),h=2;break}else h=6;i++}return C}function _(){if(n="",h=0,o=i,l=a,d=c,i>=t)return o=t,r=17;let C=s.charCodeAt(i);if(XO(C)){do i++,n+=String.fromCharCode(C),C=s.charCodeAt(i);while(XO(C));return r=15}if(n1(C))return i++,n+=String.fromCharCode(C),C===13&&s.charCodeAt(i)===10&&(i++,n+=`
`),a++,c=i,r=14;switch(C){case 123:return i++,r=1;case 125:return i++,r=2;case 91:return i++,r=3;case 93:return i++,r=4;case 58:return i++,r=6;case 44:return i++,r=5;case 34:return i++,n=p(),r=10;case 47:const v=i-1;if(s.charCodeAt(i+1)===47){for(i+=2;i<t&&!n1(s.charCodeAt(i));)i++;return n=s.substring(v,i),r=12}if(s.charCodeAt(i+1)===42){i+=2;const y=t-1;let x=!1;for(;i<y;){const L=s.charCodeAt(i);if(L===42&&s.charCodeAt(i+1)===47){i+=2,x=!0;break}i++,n1(L)&&(L===13&&s.charCodeAt(i)===10&&i++,a++,c=i)}return x||(i++,h=1),n=s.substring(v,i),r=13}return n+=String.fromCharCode(C),i++,r=16;case 45:if(n+=String.fromCharCode(C),i++,i===t||!lb(s.charCodeAt(i)))return r=16;case 48:case 49:case 50:case 51:case 52:case 53:case 54:case 55:case 56:case 57:return n+=f(),r=11;default:for(;i<t&&b(C);)i++,C=s.charCodeAt(i);if(o!==i){switch(n=s.substring(o,i),n){case"true":return r=8;case"false":return r=9;case"null":return r=7}return r=16}return n+=String.fromCharCode(C),i++,r=16}}function b(C){if(XO(C)||n1(C))return!1;switch(C){case 125:case 93:case 123:case 91:case 34:case 58:case 44:case 47:return!1}return!0}function w(){let C;do C=_();while(C>=12&&C<=15);return C}return{setPosition:g,getPosition:()=>i,scan:e?w:_,getToken:()=>r,getTokenValue:()=>n,getTokenOffset:()=>o,getTokenLength:()=>i-o,getTokenStartLine:()=>l,getTokenStartCharacter:()=>o-d,getTokenError:()=>h}}function XO(s){return s===32||s===9}function n1(s){return s===10||s===13}function lb(s){return s>=48&&s<=57}var hX;(function(s){s[s.lineFeed=10]="lineFeed",s[s.carriageReturn=13]="carriageReturn",s[s.space=32]="space",s[s._0=48]="_0",s[s._1=49]="_1",s[s._2=50]="_2",s[s._3=51]="_3",s[s._4=52]="_4",s[s._5=53]="_5",s[s._6=54]="_6",s[s._7=55]="_7",s[s._8=56]="_8",s[s._9=57]="_9",s[s.a=97]="a",s[s.b=98]="b",s[s.c=99]="c",s[s.d=100]="d",s[s.e=101]="e",s[s.f=102]="f",s[s.g=103]="g",s[s.h=104]="h",s[s.i=105]="i",s[s.j=106]="j",s[s.k=107]="k",s[s.l=108]="l",s[s.m=109]="m",s[s.n=110]="n",s[s.o=111]="o",s[s.p=112]="p",s[s.q=113]="q",s[s.r=114]="r",s[s.s=115]="s",s[s.t=116]="t",s[s.u=117]="u",s[s.v=118]="v",s[s.w=119]="w",s[s.x=120]="x",s[s.y=121]="y",s[s.z=122]="z",s[s.A=65]="A",s[s.B=66]="B",s[s.C=67]="C",s[s.D=68]="D",s[s.E=69]="E",s[s.F=70]="F",s[s.G=71]="G",s[s.H=72]="H",s[s.I=73]="I",s[s.J=74]="J",s[s.K=75]="K",s[s.L=76]="L",s[s.M=77]="M",s[s.N=78]="N",s[s.O=79]="O",s[s.P=80]="P",s[s.Q=81]="Q",s[s.R=82]="R",s[s.S=83]="S",s[s.T=84]="T",s[s.U=85]="U",s[s.V=86]="V",s[s.W=87]="W",s[s.X=88]="X",s[s.Y=89]="Y",s[s.Z=90]="Z",s[s.asterisk=42]="asterisk",s[s.backslash=92]="backslash",s[s.closeBrace=125]="closeBrace",s[s.closeBracket=93]="closeBracket",s[s.colon=58]="colon",s[s.comma=44]="comma",s[s.dot=46]="dot",s[s.doubleQuote=34]="doubleQuote",s[s.minus=45]="minus",s[s.openBrace=123]="openBrace",s[s.openBracket=91]="openBracket",s[s.plus=43]="plus",s[s.slash=47]="slash",s[s.formFeed=12]="formFeed",s[s.tab=9]="tab"})(hX||(hX={}));new Array(20).fill(0).map((s,e)=>" ".repeat(e));const cb=200;new Array(cb).fill(0).map((s,e)=>`
`+" ".repeat(e)),new Array(cb).fill(0).map((s,e)=>"\r"+" ".repeat(e)),new Array(cb).fill(0).map((s,e)=>`\r
`+" ".repeat(e)),new Array(cb).fill(0).map((s,e)=>`
`+"	".repeat(e)),new Array(cb).fill(0).map((s,e)=>"\r"+"	".repeat(e)),new Array(cb).fill(0).map((s,e)=>`\r
`+"	".repeat(e));var uX;(function(s){s.DEFAULT={allowTrailingComma:!1}})(uX||(uX={}));const kze=Lze;var gX;(function(s){s[s.None=0]="None",s[s.UnexpectedEndOfComment=1]="UnexpectedEndOfComment",s[s.UnexpectedEndOfString=2]="UnexpectedEndOfString",s[s.UnexpectedEndOfNumber=3]="UnexpectedEndOfNumber",s[s.InvalidUnicode=4]="InvalidUnicode",s[s.InvalidEscapeCharacter=5]="InvalidEscapeCharacter",s[s.InvalidCharacter=6]="InvalidCharacter"})(gX||(gX={}));var fX;(function(s){s[s.OpenBraceToken=1]="OpenBraceToken",s[s.CloseBraceToken=2]="CloseBraceToken",s[s.OpenBracketToken=3]="OpenBracketToken",s[s.CloseBracketToken=4]="CloseBracketToken",s[s.CommaToken=5]="CommaToken",s[s.ColonToken=6]="ColonToken",s[s.NullKeyword=7]="NullKeyword",s[s.TrueKeyword=8]="TrueKeyword",s[s.FalseKeyword=9]="FalseKeyword",s[s.StringLiteral=10]="StringLiteral",s[s.NumericLiteral=11]="NumericLiteral",s[s.LineCommentTrivia=12]="LineCommentTrivia",s[s.BlockCommentTrivia=13]="BlockCommentTrivia",s[s.LineBreakTrivia=14]="LineBreakTrivia",s[s.Trivia=15]="Trivia",s[s.Unknown=16]="Unknown",s[s.EOF=17]="EOF"})(fX||(fX={}));var pX;(function(s){s[s.InvalidSymbol=1]="InvalidSymbol",s[s.InvalidNumberFormat=2]="InvalidNumberFormat",s[s.PropertyNameExpected=3]="PropertyNameExpected",s[s.ValueExpected=4]="ValueExpected",s[s.ColonExpected=5]="ColonExpected",s[s.CommaExpected=6]="CommaExpected",s[s.CloseBraceExpected=7]="CloseBraceExpected",s[s.CloseBracketExpected=8]="CloseBracketExpected",s[s.EndOfFileExpected=9]="EndOfFileExpected",s[s.InvalidCommentToken=10]="InvalidCommentToken",s[s.UnexpectedEndOfComment=11]="UnexpectedEndOfComment",s[s.UnexpectedEndOfString=12]="UnexpectedEndOfString",s[s.UnexpectedEndOfNumber=13]="UnexpectedEndOfNumber",s[s.InvalidUnicode=14]="InvalidUnicode",s[s.InvalidEscapeCharacter=15]="InvalidEscapeCharacter",s[s.InvalidCharacter=16]="InvalidCharacter"})(pX||(pX={}));function Eze(s){return{getInitialState:()=>new M0(null,null,!1,null),tokenize:(e,t)=>Fze(s,e,t)}}const mX="delimiter.bracket.json",_X="delimiter.array.json",Ize="delimiter.colon.json",Dze="delimiter.comma.json",Tze="keyword.json",Nze="keyword.json",Rze="string.value.json",Aze="number.json",Mze="string.key.json",Oze="comment.block.json",Pze="comment.line.json";class ip{constructor(e,t){this.parent=e,this.type=t}static pop(e){return e?e.parent:null}static push(e,t){return new ip(e,t)}static equals(e,t){if(!e&&!t)return!0;if(!e||!t)return!1;for(;e&&t;){if(e===t)return!0;if(e.type!==t.type)return!1;e=e.parent,t=t.parent}return!0}}class M0{constructor(e,t,i,n){this._state=e,this.scanError=t,this.lastWasColon=i,this.parents=n}clone(){return new M0(this._state,this.scanError,this.lastWasColon,this.parents)}equals(e){return e===this?!0:!e||!(e instanceof M0)?!1:this.scanError===e.scanError&&this.lastWasColon===e.lastWasColon&&ip.equals(this.parents,e.parents)}getStateData(){return this._state}setStateData(e){this._state=e}}function Fze(s,e,t,i=0){let n=0,o=!1;switch(t.scanError){case 2:e='"'+e,n=1;break;case 1:e="/*"+e,n=2;break}const r=kze(e);let a=t.lastWasColon,l=t.parents;const c={tokens:[],endState:t.clone()};for(;;){let d=i+r.getPosition(),h="";const u=r.scan();if(u===17)break;if(d===i+r.getPosition())throw new Error("Scanner did not advance, next 3 characters are: "+e.substr(r.getPosition(),3));switch(o&&(d-=n),o=n>0,u){case 1:l=ip.push(l,0),h=mX,a=!1;break;case 2:l=ip.pop(l),h=mX,a=!1;break;case 3:l=ip.push(l,1),h=_X,a=!1;break;case 4:l=ip.pop(l),h=_X,a=!1;break;case 6:h=Ize,a=!0;break;case 5:h=Dze,a=!1;break;case 8:case 9:h=Tze,a=!1;break;case 7:h=Nze,a=!1;break;case 10:const f=(l?l.type:0)===1;h=a||f?Rze:Mze,a=!1;break;case 11:h=Aze,a=!1;break}switch(u){case 12:h=Pze;break;case 13:h=Oze;break}c.endState=new M0(t.getStateData(),r.getTokenError(),a,l),c.tokens.push({startIndex:d,scopes:h})}return c}let va;function Bze(){return new Promise((s,e)=>{if(!va)return e("JSON not registered!");s(va)})}class Wze extends Ix{constructor(e,t,i){super(e,t,i.onDidChange),this._disposables.push(ti.onWillDisposeModel(n=>{this._resetSchema(n.uri)})),this._disposables.push(ti.onDidChangeModelLanguage(n=>{this._resetSchema(n.model.uri)}))}_resetSchema(e){this._worker().then(t=>{t.resetSchema(e.toString())})}}function Hze(s){const e=[],t=[],i=new Qae(s);e.push(i),va=(...r)=>i.getLanguageServiceWorker(...r);function n(){const{languageId:r,modeConfiguration:a}=s;Jae(t),a.documentFormattingEdits&&t.push(H.registerDocumentFormattingEditProvider(r,new L_(va))),a.documentRangeFormattingEdits&&t.push(H.registerDocumentRangeFormattingEditProvider(r,new k_(va))),a.completionItems&&t.push(H.registerCompletionItemProvider(r,new mC(va,[" ",":",'"']))),a.hovers&&t.push(H.registerHoverProvider(r,new S_(va))),a.documentSymbols&&t.push(H.registerDocumentSymbolProvider(r,new x_(va))),a.tokens&&t.push(H.setTokensProvider(r,Eze(!0))),a.colors&&t.push(H.registerColorProvider(r,new Nx(va))),a.foldingRanges&&t.push(H.registerFoldingRangeProvider(r,new E_(va))),a.diagnostics&&t.push(new Wze(r,va,s)),a.selectionRanges&&t.push(H.registerSelectionRangeProvider(r,new I_(va)))}n(),e.push(H.setLanguageConfiguration(s.languageId,Vze));let o=s.modeConfiguration;return s.onDidChange(r=>{r.modeConfiguration!==o&&(o=r.modeConfiguration,n())}),e.push(bX(t)),bX(e)}function bX(s){return{dispose:()=>Jae(s)}}function Jae(s){for(;s.length;)s.pop().dispose()}const Vze={wordPattern:/(-?\d*\.\d\w*)|([^\[\{\]\}\:\"\,\s]+)/g,comments:{lineComment:"//",blockComment:["/*","*/"]},brackets:[["{","}"],["[","]"]],autoClosingPairs:[{open:"{",close:"}",notIn:["string"]},{open:"[",close:"]",notIn:["string"]},{open:'"',close:'"',notIn:["string"]}]},zze=Object.freeze(Object.defineProperty({__proto__:null,CompletionAdapter:mC,DefinitionAdapter:u2,DiagnosticsAdapter:Ix,DocumentColorAdapter:Nx,DocumentFormattingEditProvider:L_,DocumentHighlightAdapter:_C,DocumentLinkAdapter:Tx,DocumentRangeFormattingEditProvider:k_,DocumentSymbolAdapter:x_,FoldingRangeAdapter:E_,HoverAdapter:S_,ReferenceAdapter:g2,RenameAdapter:bC,SelectionRangeAdapter:I_,WorkerManager:Qae,fromPosition:Zc,fromRange:Dx,getWorker:Bze,setupMode:Hze,toRange:Os,toTextEdit:Lg},Symbol.toStringTag,{value:"Module"}));class ele{constructor(e,t){this._modeId=e,this._defaults=t,this._worker=null,this._client=null,this._configChangeListener=this._defaults.onDidChange(()=>this._stopWorker()),this._updateExtraLibsToken=0,this._extraLibsChangeListener=this._defaults.onDidExtraLibsChange(()=>this._updateExtraLibs())}dispose(){this._configChangeListener.dispose(),this._extraLibsChangeListener.dispose(),this._stopWorker()}_stopWorker(){this._worker&&(this._worker.dispose(),this._worker=null),this._client=null}async _updateExtraLibs(){if(!this._worker)return;const e=++this._updateExtraLibsToken,t=await this._worker.getProxy();this._updateExtraLibsToken===e&&t.updateExtraLibs(this._defaults.getExtraLibs())}_getClient(){return this._client||(this._client=(async()=>(this._worker=Ex({moduleId:"vs/language/typescript/tsWorker",createWorker:()=>new Worker(new URL("/assets/ts.worker-BH9nVgjN.js",import.meta.url),{type:"module"}),label:this._modeId,keepIdleModels:!0,createData:{compilerOptions:this._defaults.getCompilerOptions(),extraLibs:this._defaults.getExtraLibs(),customWorkerPath:this._defaults.workerOptions.customWorkerPath,inlayHintsOptions:this._defaults.inlayHintsOptions}}),this._defaults.getEagerModelSync()?await this._worker.withSyncedResources(ti.getModels().filter(e=>e.getLanguageId()===this._modeId).map(e=>e.uri)):await this._worker.getProxy()))()),this._client}async getLanguageServiceWorker(...e){const t=await this._getClient();return this._worker&&await this._worker.withSyncedResources(e),t}}const ve={};ve["lib.d.ts"]=!0;ve["lib.decorators.d.ts"]=!0;ve["lib.decorators.legacy.d.ts"]=!0;ve["lib.dom.asynciterable.d.ts"]=!0;ve["lib.dom.d.ts"]=!0;ve["lib.dom.iterable.d.ts"]=!0;ve["lib.es2015.collection.d.ts"]=!0;ve["lib.es2015.core.d.ts"]=!0;ve["lib.es2015.d.ts"]=!0;ve["lib.es2015.generator.d.ts"]=!0;ve["lib.es2015.iterable.d.ts"]=!0;ve["lib.es2015.promise.d.ts"]=!0;ve["lib.es2015.proxy.d.ts"]=!0;ve["lib.es2015.reflect.d.ts"]=!0;ve["lib.es2015.symbol.d.ts"]=!0;ve["lib.es2015.symbol.wellknown.d.ts"]=!0;ve["lib.es2016.array.include.d.ts"]=!0;ve["lib.es2016.d.ts"]=!0;ve["lib.es2016.full.d.ts"]=!0;ve["lib.es2016.intl.d.ts"]=!0;ve["lib.es2017.arraybuffer.d.ts"]=!0;ve["lib.es2017.d.ts"]=!0;ve["lib.es2017.date.d.ts"]=!0;ve["lib.es2017.full.d.ts"]=!0;ve["lib.es2017.intl.d.ts"]=!0;ve["lib.es2017.object.d.ts"]=!0;ve["lib.es2017.sharedmemory.d.ts"]=!0;ve["lib.es2017.string.d.ts"]=!0;ve["lib.es2017.typedarrays.d.ts"]=!0;ve["lib.es2018.asyncgenerator.d.ts"]=!0;ve["lib.es2018.asynciterable.d.ts"]=!0;ve["lib.es2018.d.ts"]=!0;ve["lib.es2018.full.d.ts"]=!0;ve["lib.es2018.intl.d.ts"]=!0;ve["lib.es2018.promise.d.ts"]=!0;ve["lib.es2018.regexp.d.ts"]=!0;ve["lib.es2019.array.d.ts"]=!0;ve["lib.es2019.d.ts"]=!0;ve["lib.es2019.full.d.ts"]=!0;ve["lib.es2019.intl.d.ts"]=!0;ve["lib.es2019.object.d.ts"]=!0;ve["lib.es2019.string.d.ts"]=!0;ve["lib.es2019.symbol.d.ts"]=!0;ve["lib.es2020.bigint.d.ts"]=!0;ve["lib.es2020.d.ts"]=!0;ve["lib.es2020.date.d.ts"]=!0;ve["lib.es2020.full.d.ts"]=!0;ve["lib.es2020.intl.d.ts"]=!0;ve["lib.es2020.number.d.ts"]=!0;ve["lib.es2020.promise.d.ts"]=!0;ve["lib.es2020.sharedmemory.d.ts"]=!0;ve["lib.es2020.string.d.ts"]=!0;ve["lib.es2020.symbol.wellknown.d.ts"]=!0;ve["lib.es2021.d.ts"]=!0;ve["lib.es2021.full.d.ts"]=!0;ve["lib.es2021.intl.d.ts"]=!0;ve["lib.es2021.promise.d.ts"]=!0;ve["lib.es2021.string.d.ts"]=!0;ve["lib.es2021.weakref.d.ts"]=!0;ve["lib.es2022.array.d.ts"]=!0;ve["lib.es2022.d.ts"]=!0;ve["lib.es2022.error.d.ts"]=!0;ve["lib.es2022.full.d.ts"]=!0;ve["lib.es2022.intl.d.ts"]=!0;ve["lib.es2022.object.d.ts"]=!0;ve["lib.es2022.regexp.d.ts"]=!0;ve["lib.es2022.string.d.ts"]=!0;ve["lib.es2023.array.d.ts"]=!0;ve["lib.es2023.collection.d.ts"]=!0;ve["lib.es2023.d.ts"]=!0;ve["lib.es2023.full.d.ts"]=!0;ve["lib.es2023.intl.d.ts"]=!0;ve["lib.es2024.arraybuffer.d.ts"]=!0;ve["lib.es2024.collection.d.ts"]=!0;ve["lib.es2024.d.ts"]=!0;ve["lib.es2024.full.d.ts"]=!0;ve["lib.es2024.object.d.ts"]=!0;ve["lib.es2024.promise.d.ts"]=!0;ve["lib.es2024.regexp.d.ts"]=!0;ve["lib.es2024.sharedmemory.d.ts"]=!0;ve["lib.es2024.string.d.ts"]=!0;ve["lib.es5.d.ts"]=!0;ve["lib.es6.d.ts"]=!0;ve["lib.esnext.array.d.ts"]=!0;ve["lib.esnext.collection.d.ts"]=!0;ve["lib.esnext.d.ts"]=!0;ve["lib.esnext.decorators.d.ts"]=!0;ve["lib.esnext.disposable.d.ts"]=!0;ve["lib.esnext.error.d.ts"]=!0;ve["lib.esnext.float16.d.ts"]=!0;ve["lib.esnext.full.d.ts"]=!0;ve["lib.esnext.intl.d.ts"]=!0;ve["lib.esnext.iterator.d.ts"]=!0;ve["lib.esnext.promise.d.ts"]=!0;ve["lib.esnext.sharedmemory.d.ts"]=!0;ve["lib.scripthost.d.ts"]=!0;ve["lib.webworker.asynciterable.d.ts"]=!0;ve["lib.webworker.d.ts"]=!0;ve["lib.webworker.importscripts.d.ts"]=!0;ve["lib.webworker.iterable.d.ts"]=!0;function OT(s,e,t=0){if(typeof s=="string")return s;if(s===void 0)return"";let i="";if(t){i+=e;for(let n=0;n<t;n++)i+="  "}if(i+=s.messageText,t++,s.next)for(const n of s.next)i+=OT(n,e,t);return i}function hc(s){return s?s.map(e=>e.text).join(""):""}class qa{constructor(e){this._worker=e}_textSpanToRange(e,t){let i=e.getPositionAt(t.start),n=e.getPositionAt(t.start+t.length),{lineNumber:o,column:r}=i,{lineNumber:a,column:l}=n;return{startLineNumber:o,startColumn:r,endLineNumber:a,endColumn:l}}}class tle{constructor(e){this._worker=e,this._libFiles={},this._hasFetchedLibFiles=!1,this._fetchLibFilesPromise=null}isLibFile(e){return e&&e.path.indexOf("/lib.")===0?!!ve[e.path.slice(1)]:!1}getOrCreateModel(e){const t=Yc.parse(e),i=ti.getModel(t);if(i)return i;if(this.isLibFile(t)&&this._hasFetchedLibFiles)return ti.createModel(this._libFiles[t.path.slice(1)],"typescript",t);const n=TV.getExtraLibs()[e];return n?ti.createModel(n.content,"typescript",t):null}_containsLibFile(e){for(let t of e)if(this.isLibFile(t))return!0;return!1}async fetchLibFilesIfNecessary(e){this._containsLibFile(e)&&await this._fetchLibFiles()}_fetchLibFiles(){return this._fetchLibFilesPromise||(this._fetchLibFilesPromise=this._worker().then(e=>e.getLibFiles()).then(e=>{this._hasFetchedLibFiles=!0,this._libFiles=e})),this._fetchLibFilesPromise}}class ile extends qa{constructor(e,t,i,n){super(n),this._libFiles=e,this._defaults=t,this._selector=i,this._disposables=[],this._listener=Object.create(null);const o=l=>{if(l.getLanguageId()!==i)return;const c=()=>{const{onlyVisible:g}=this._defaults.getDiagnosticsOptions();g?l.isAttachedToEditor()&&this._doValidate(l):this._doValidate(l)};let d;const h=l.onDidChangeContent(()=>{clearTimeout(d),d=window.setTimeout(c,500)}),u=l.onDidChangeAttached(()=>{const{onlyVisible:g}=this._defaults.getDiagnosticsOptions();g&&(l.isAttachedToEditor()?c():ti.setModelMarkers(l,this._selector,[]))});this._listener[l.uri.toString()]={dispose(){h.dispose(),u.dispose(),clearTimeout(d)}},c()},r=l=>{ti.setModelMarkers(l,this._selector,[]);const c=l.uri.toString();this._listener[c]&&(this._listener[c].dispose(),delete this._listener[c])};this._disposables.push(ti.onDidCreateModel(l=>o(l))),this._disposables.push(ti.onWillDisposeModel(r)),this._disposables.push(ti.onDidChangeModelLanguage(l=>{r(l.model),o(l.model)})),this._disposables.push({dispose(){for(const l of ti.getModels())r(l)}});const a=()=>{for(const l of ti.getModels())r(l),o(l)};this._disposables.push(this._defaults.onDidChange(a)),this._disposables.push(this._defaults.onDidExtraLibsChange(a)),ti.getModels().forEach(l=>o(l))}dispose(){this._disposables.forEach(e=>e&&e.dispose()),this._disposables=[]}async _doValidate(e){const t=await this._worker(e.uri);if(e.isDisposed())return;const i=[],{noSyntaxValidation:n,noSemanticValidation:o,noSuggestionDiagnostics:r}=this._defaults.getDiagnosticsOptions();n||i.push(t.getSyntacticDiagnostics(e.uri.toString())),o||i.push(t.getSemanticDiagnostics(e.uri.toString())),r||i.push(t.getSuggestionDiagnostics(e.uri.toString()));const a=await Promise.all(i);if(!a||e.isDisposed())return;const l=a.reduce((d,h)=>h.concat(d),[]).filter(d=>(this._defaults.getDiagnosticsOptions().diagnosticCodesToIgnore||[]).indexOf(d.code)===-1),c=l.map(d=>d.relatedInformation||[]).reduce((d,h)=>h.concat(d),[]).map(d=>d.file?Yc.parse(d.file.fileName):null);await this._libFiles.fetchLibFilesIfNecessary(c),!e.isDisposed()&&ti.setModelMarkers(e,this._selector,l.map(d=>this._convertDiagnostics(e,d)))}_convertDiagnostics(e,t){const i=t.start||0,n=t.length||1,{lineNumber:o,column:r}=e.getPositionAt(i),{lineNumber:a,column:l}=e.getPositionAt(i+n),c=[];return t.reportsUnnecessary&&c.push(Wv.Unnecessary),t.reportsDeprecated&&c.push(Wv.Deprecated),{severity:this._tsDiagnosticCategoryToMarkerSeverity(t.category),startLineNumber:o,startColumn:r,endLineNumber:a,endColumn:l,message:OT(t.messageText,`
`),code:t.code.toString(),tags:c,relatedInformation:this._convertRelatedInformation(e,t.relatedInformation)}}_convertRelatedInformation(e,t){if(!t)return[];const i=[];return t.forEach(n=>{let o=e;if(n.file&&(o=this._libFiles.getOrCreateModel(n.file.fileName)),!o)return;const r=n.start||0,a=n.length||1,{lineNumber:l,column:c}=o.getPositionAt(r),{lineNumber:d,column:h}=o.getPositionAt(r+a);i.push({resource:o.uri,startLineNumber:l,startColumn:c,endLineNumber:d,endColumn:h,message:OT(n.messageText,`
`)})}),i}_tsDiagnosticCategoryToMarkerSeverity(e){switch(e){case 1:return kn.Error;case 3:return kn.Info;case 0:return kn.Warning;case 2:return kn.Hint}return kn.Info}}class Bw extends qa{get triggerCharacters(){return["."]}async provideCompletionItems(e,t,i,n){const o=e.getWordUntilPosition(t),r=new Hc(t.lineNumber,o.startColumn,t.lineNumber,o.endColumn),a=e.uri,l=e.getOffsetAt(t),c=await this._worker(a);if(e.isDisposed())return;const d=await c.getCompletionsAtPosition(a.toString(),l);return!d||e.isDisposed()?void 0:{suggestions:d.entries.map(u=>{let g=r;if(u.replacementSpan){const p=e.getPositionAt(u.replacementSpan.start),_=e.getPositionAt(u.replacementSpan.start+u.replacementSpan.length);g=new Hc(p.lineNumber,p.column,_.lineNumber,_.column)}const f=[];return u.kindModifiers!==void 0&&u.kindModifiers.indexOf("deprecated")!==-1&&f.push(H.CompletionItemTag.Deprecated),{uri:a,position:t,offset:l,range:g,label:u.name,insertText:u.name,sortText:u.sortText,kind:Bw.convertKind(u.kind),tags:f}})}}async resolveCompletionItem(e,t){const i=e,n=i.uri,o=i.position,r=i.offset,l=await(await this._worker(n)).getCompletionEntryDetails(n.toString(),r,i.label);return l?{uri:n,position:o,label:l.name,kind:Bw.convertKind(l.kind),detail:hc(l.displayParts),documentation:{value:Bw.createDocumentationString(l)}}:i}static convertKind(e){switch(e){case Je.primitiveType:case Je.keyword:return H.CompletionItemKind.Keyword;case Je.variable:case Je.localVariable:return H.CompletionItemKind.Variable;case Je.memberVariable:case Je.memberGetAccessor:case Je.memberSetAccessor:return H.CompletionItemKind.Field;case Je.function:case Je.memberFunction:case Je.constructSignature:case Je.callSignature:case Je.indexSignature:return H.CompletionItemKind.Function;case Je.enum:return H.CompletionItemKind.Enum;case Je.module:return H.CompletionItemKind.Module;case Je.class:return H.CompletionItemKind.Class;case Je.interface:return H.CompletionItemKind.Interface;case Je.warning:return H.CompletionItemKind.File}return H.CompletionItemKind.Property}static createDocumentationString(e){let t=hc(e.documentation);if(e.tags)for(const i of e.tags)t+=`

${nle(i)}`;return t}}function nle(s){let e=`*@${s.name}*`;if(s.name==="param"&&s.text){const[t,...i]=s.text;e+=`\`${t.text}\``,i.length>0&&(e+=` — ${i.map(n=>n.text).join(" ")}`)}else Array.isArray(s.text)?e+=` — ${s.text.map(t=>t.text).join(" ")}`:s.text&&(e+=` — ${s.text}`);return e}class f2 extends qa{constructor(){super(...arguments),this.signatureHelpTriggerCharacters=["(",","]}static _toSignatureHelpTriggerReason(e){switch(e.triggerKind){case H.SignatureHelpTriggerKind.TriggerCharacter:return e.triggerCharacter?e.isRetrigger?{kind:"retrigger",triggerCharacter:e.triggerCharacter}:{kind:"characterTyped",triggerCharacter:e.triggerCharacter}:{kind:"invoked"};case H.SignatureHelpTriggerKind.ContentChange:return e.isRetrigger?{kind:"retrigger"}:{kind:"invoked"};case H.SignatureHelpTriggerKind.Invoke:default:return{kind:"invoked"}}}async provideSignatureHelp(e,t,i,n){const o=e.uri,r=e.getOffsetAt(t),a=await this._worker(o);if(e.isDisposed())return;const l=await a.getSignatureHelpItems(o.toString(),r,{triggerReason:f2._toSignatureHelpTriggerReason(n)});if(!l||e.isDisposed())return;const c={activeSignature:l.selectedItemIndex,activeParameter:l.argumentIndex,signatures:[]};return l.items.forEach(d=>{const h={label:"",parameters:[]};h.documentation={value:hc(d.documentation)},h.label+=hc(d.prefixDisplayParts),d.parameters.forEach((u,g,f)=>{const p=hc(u.displayParts),_={label:p,documentation:{value:hc(u.documentation)}};h.label+=p,h.parameters.push(_),g<f.length-1&&(h.label+=hc(d.separatorDisplayParts))}),h.label+=hc(d.suffixDisplayParts),c.signatures.push(h)}),{value:c,dispose(){}}}}class sle extends qa{async provideHover(e,t,i){const n=e.uri,o=e.getOffsetAt(t),r=await this._worker(n);if(e.isDisposed())return;const a=await r.getQuickInfoAtPosition(n.toString(),o);if(!a||e.isDisposed())return;const l=hc(a.documentation),c=a.tags?a.tags.map(h=>nle(h)).join(`  

`):"",d=hc(a.displayParts);return{range:this._textSpanToRange(e,a.textSpan),contents:[{value:"```typescript\n"+d+"\n```\n"},{value:l+(c?`

`+c:"")}]}}}class ole extends qa{async provideDocumentHighlights(e,t,i){const n=e.uri,o=e.getOffsetAt(t),r=await this._worker(n);if(e.isDisposed())return;const a=await r.getDocumentHighlights(n.toString(),o,[n.toString()]);if(!(!a||e.isDisposed()))return a.flatMap(l=>l.highlightSpans.map(c=>({range:this._textSpanToRange(e,c.textSpan),kind:c.kind==="writtenReference"?H.DocumentHighlightKind.Write:H.DocumentHighlightKind.Text})))}}class rle extends qa{constructor(e,t){super(t),this._libFiles=e}async provideDefinition(e,t,i){const n=e.uri,o=e.getOffsetAt(t),r=await this._worker(n);if(e.isDisposed())return;const a=await r.getDefinitionAtPosition(n.toString(),o);if(!a||e.isDisposed()||(await this._libFiles.fetchLibFilesIfNecessary(a.map(c=>Yc.parse(c.fileName))),e.isDisposed()))return;const l=[];for(let c of a){const d=this._libFiles.getOrCreateModel(c.fileName);d&&l.push({uri:d.uri,range:this._textSpanToRange(d,c.textSpan)})}return l}}class ale extends qa{constructor(e,t){super(t),this._libFiles=e}async provideReferences(e,t,i,n){const o=e.uri,r=e.getOffsetAt(t),a=await this._worker(o);if(e.isDisposed())return;const l=await a.getReferencesAtPosition(o.toString(),r);if(!l||e.isDisposed()||(await this._libFiles.fetchLibFilesIfNecessary(l.map(d=>Yc.parse(d.fileName))),e.isDisposed()))return;const c=[];for(let d of l){const h=this._libFiles.getOrCreateModel(d.fileName);h&&c.push({uri:h.uri,range:this._textSpanToRange(h,d.textSpan)})}return c}}class lle extends qa{async provideDocumentSymbols(e,t){const i=e.uri,n=await this._worker(i);if(e.isDisposed())return;const o=await n.getNavigationTree(i.toString());if(!o||e.isDisposed())return;const r=(l,c)=>({name:l.text,detail:"",kind:Qo[l.kind]||H.SymbolKind.Variable,range:this._textSpanToRange(e,l.spans[0]),selectionRange:this._textSpanToRange(e,l.spans[0]),tags:[],children:l.childItems?.map(h=>r(h,l.text)),containerName:c});return o.childItems?o.childItems.map(l=>r(l)):[]}}class Je{}Je.unknown="";Je.keyword="keyword";Je.script="script";Je.module="module";Je.class="class";Je.interface="interface";Je.type="type";Je.enum="enum";Je.variable="var";Je.localVariable="local var";Je.function="function";Je.localFunction="local function";Je.memberFunction="method";Je.memberGetAccessor="getter";Je.memberSetAccessor="setter";Je.memberVariable="property";Je.constructorImplementation="constructor";Je.callSignature="call";Je.indexSignature="index";Je.constructSignature="construct";Je.parameter="parameter";Je.typeParameter="type parameter";Je.primitiveType="primitive type";Je.label="label";Je.alias="alias";Je.const="const";Je.let="let";Je.warning="warning";let Qo=Object.create(null);Qo[Je.module]=H.SymbolKind.Module;Qo[Je.class]=H.SymbolKind.Class;Qo[Je.enum]=H.SymbolKind.Enum;Qo[Je.interface]=H.SymbolKind.Interface;Qo[Je.memberFunction]=H.SymbolKind.Method;Qo[Je.memberVariable]=H.SymbolKind.Property;Qo[Je.memberGetAccessor]=H.SymbolKind.Property;Qo[Je.memberSetAccessor]=H.SymbolKind.Property;Qo[Je.variable]=H.SymbolKind.Variable;Qo[Je.const]=H.SymbolKind.Variable;Qo[Je.localVariable]=H.SymbolKind.Variable;Qo[Je.variable]=H.SymbolKind.Variable;Qo[Je.function]=H.SymbolKind.Function;Qo[Je.localFunction]=H.SymbolKind.Function;class t_ extends qa{static _convertOptions(e){return{ConvertTabsToSpaces:e.insertSpaces,TabSize:e.tabSize,IndentSize:e.tabSize,IndentStyle:2,NewLineCharacter:`
`,InsertSpaceAfterCommaDelimiter:!0,InsertSpaceAfterSemicolonInForStatements:!0,InsertSpaceBeforeAndAfterBinaryOperators:!0,InsertSpaceAfterKeywordsInControlFlowStatements:!0,InsertSpaceAfterFunctionKeywordForAnonymousFunctions:!0,InsertSpaceAfterOpeningAndBeforeClosingNonemptyParenthesis:!1,InsertSpaceAfterOpeningAndBeforeClosingNonemptyBrackets:!1,InsertSpaceAfterOpeningAndBeforeClosingTemplateStringBraces:!1,PlaceOpenBraceOnNewLineForControlBlocks:!1,PlaceOpenBraceOnNewLineForFunctions:!1}}_convertTextChanges(e,t){return{text:t.newText,range:this._textSpanToRange(e,t.span)}}}class cle extends t_{constructor(){super(...arguments),this.canFormatMultipleRanges=!1}async provideDocumentRangeFormattingEdits(e,t,i,n){const o=e.uri,r=e.getOffsetAt({lineNumber:t.startLineNumber,column:t.startColumn}),a=e.getOffsetAt({lineNumber:t.endLineNumber,column:t.endColumn}),l=await this._worker(o);if(e.isDisposed())return;const c=await l.getFormattingEditsForRange(o.toString(),r,a,t_._convertOptions(i));if(!(!c||e.isDisposed()))return c.map(d=>this._convertTextChanges(e,d))}}class dle extends t_{get autoFormatTriggerCharacters(){return[";","}",`
`]}async provideOnTypeFormattingEdits(e,t,i,n,o){const r=e.uri,a=e.getOffsetAt(t),l=await this._worker(r);if(e.isDisposed())return;const c=await l.getFormattingEditsAfterKeystroke(r.toString(),a,i,t_._convertOptions(n));if(!(!c||e.isDisposed()))return c.map(d=>this._convertTextChanges(e,d))}}class hle extends t_{async provideCodeActions(e,t,i,n){const o=e.uri,r=e.getOffsetAt({lineNumber:t.startLineNumber,column:t.startColumn}),a=e.getOffsetAt({lineNumber:t.endLineNumber,column:t.endColumn}),l=t_._convertOptions(e.getOptions()),c=i.markers.filter(g=>g.code).map(g=>g.code).map(Number),d=await this._worker(o);if(e.isDisposed())return;const h=await d.getCodeFixesAtPosition(o.toString(),r,a,c,l);return!h||e.isDisposed()?{actions:[],dispose:()=>{}}:{actions:h.filter(g=>g.changes.filter(f=>f.isNewFile).length===0).map(g=>this._tsCodeFixActionToMonacoCodeAction(e,i,g)),dispose:()=>{}}}_tsCodeFixActionToMonacoCodeAction(e,t,i){const n=[];for(const r of i.changes)for(const a of r.textChanges)n.push({resource:e.uri,versionId:void 0,textEdit:{range:this._textSpanToRange(e,a.span),text:a.newText}});return{title:i.description,edit:{edits:n},diagnostics:t.markers,kind:"quickfix"}}}class ule extends qa{constructor(e,t){super(t),this._libFiles=e}async provideRenameEdits(e,t,i,n){const o=e.uri,r=o.toString(),a=e.getOffsetAt(t),l=await this._worker(o);if(e.isDisposed())return;const c=await l.getRenameInfo(r,a,{allowRenameOfImportPath:!1});if(c.canRename===!1)return{edits:[],rejectReason:c.localizedErrorMessage};if(c.fileToRename!==void 0)throw new Error("Renaming files is not supported.");const d=await l.findRenameLocations(r,a,!1,!1,!1);if(!d||e.isDisposed())return;const h=[];for(const u of d){const g=this._libFiles.getOrCreateModel(u.fileName);if(g)h.push({resource:g.uri,versionId:void 0,textEdit:{range:this._textSpanToRange(g,u.textSpan),text:i}});else throw new Error(`Unknown file ${u.fileName}.`)}return{edits:h}}}class gle extends qa{async provideInlayHints(e,t,i){const n=e.uri,o=n.toString(),r=e.getOffsetAt({lineNumber:t.startLineNumber,column:t.startColumn}),a=e.getOffsetAt({lineNumber:t.endLineNumber,column:t.endColumn}),l=await this._worker(n);return e.isDisposed()?null:{hints:(await l.provideInlayHints(o,r,a)).map(h=>({...h,label:h.text,position:e.getPositionAt(h.position),kind:this._convertHintKind(h.kind)})),dispose:()=>{}}}_convertHintKind(e){switch(e){case"Parameter":return H.InlayHintKind.Parameter;case"Type":return H.InlayHintKind.Type;default:return H.InlayHintKind.Type}}}let N6,R6;function $ze(s){R6=fle(s,"typescript")}function Uze(s){N6=fle(s,"javascript")}function qze(){return new Promise((s,e)=>{if(!N6)return e("JavaScript not registered!");s(N6)})}function jze(){return new Promise((s,e)=>{if(!R6)return e("TypeScript not registered!");s(R6)})}function fle(s,e){const t=[],i=new ele(e,s),n=(...a)=>i.getLanguageServiceWorker(...a),o=new tle(n);function r(){const{modeConfiguration:a}=s;Gze(t),a.completionItems&&t.push(H.registerCompletionItemProvider(e,new Bw(n))),a.signatureHelp&&t.push(H.registerSignatureHelpProvider(e,new f2(n))),a.hovers&&t.push(H.registerHoverProvider(e,new sle(n))),a.documentHighlights&&t.push(H.registerDocumentHighlightProvider(e,new ole(n))),a.definitions&&t.push(H.registerDefinitionProvider(e,new rle(o,n))),a.references&&t.push(H.registerReferenceProvider(e,new ale(o,n))),a.documentSymbols&&t.push(H.registerDocumentSymbolProvider(e,new lle(n))),a.rename&&t.push(H.registerRenameProvider(e,new ule(o,n))),a.documentRangeFormattingEdits&&t.push(H.registerDocumentRangeFormattingEditProvider(e,new cle(n))),a.onTypeFormattingEdits&&t.push(H.registerOnTypeFormattingEditProvider(e,new dle(n))),a.codeActions&&t.push(H.registerCodeActionProvider(e,new hle(n))),a.inlayHints&&t.push(H.registerInlayHintsProvider(e,new gle(n))),a.diagnostics&&t.push(new ile(o,s,e,n))}return r(),n}function Gze(s){for(;s.length;)s.pop().dispose()}const Kze=Object.freeze(Object.defineProperty({__proto__:null,Adapter:qa,CodeActionAdaptor:hle,DefinitionAdapter:rle,DiagnosticsAdapter:ile,DocumentHighlightAdapter:ole,FormatAdapter:cle,FormatHelper:t_,FormatOnTypeAdapter:dle,InlayHintsAdapter:gle,Kind:Je,LibFiles:tle,OutlineAdapter:lle,QuickInfoAdapter:sle,ReferenceAdapter:ale,RenameAdapter:ule,SignatureHelpAdapter:f2,SuggestAdapter:Bw,WorkerManager:ele,flattenDiagnosticMessageText:OT,getJavaScriptWorker:qze,getTypeScriptWorker:jze,setupJavaScript:Uze,setupTypeScript:$ze},Symbol.toStringTag,{value:"Module"})),Yze={comments:{lineComment:"*"},brackets:[["[","]"],["(",")"]]},Zze={defaultToken:"invalid",ignoreCase:!0,tokenPostfix:".abap",keywords:["abap-source","abbreviated","abstract","accept","accepting","according","activation","actual","add","add-corresponding","adjacent","after","alias","aliases","align","all","allocate","alpha","analysis","analyzer","and","append","appendage","appending","application","archive","area","arithmetic","as","ascending","aspect","assert","assign","assigned","assigning","association","asynchronous","at","attributes","authority","authority-check","avg","back","background","backup","backward","badi","base","before","begin","between","big","binary","bintohex","bit","black","blank","blanks","blob","block","blocks","blue","bound","boundaries","bounds","boxed","break-point","buffer","by","bypassing","byte","byte-order","call","calling","case","cast","casting","catch","center","centered","chain","chain-input","chain-request","change","changing","channels","character","char-to-hex","check","checkbox","ci_","circular","class","class-coding","class-data","class-events","class-methods","class-pool","cleanup","clear","client","clob","clock","close","coalesce","code","coding","col_background","col_group","col_heading","col_key","col_negative","col_normal","col_positive","col_total","collect","color","column","columns","comment","comments","commit","common","communication","comparing","component","components","compression","compute","concat","concat_with_space","concatenate","cond","condense","condition","connect","connection","constants","context","contexts","continue","control","controls","conv","conversion","convert","copies","copy","corresponding","country","cover","cpi","create","creating","critical","currency","currency_conversion","current","cursor","cursor-selection","customer","customer-function","dangerous","data","database","datainfo","dataset","date","dats_add_days","dats_add_months","dats_days_between","dats_is_valid","daylight","dd/mm/yy","dd/mm/yyyy","ddmmyy","deallocate","decimal_shift","decimals","declarations","deep","default","deferred","define","defining","definition","delete","deleting","demand","department","descending","describe","destination","detail","dialog","directory","disconnect","display","display-mode","distinct","divide","divide-corresponding","division","do","dummy","duplicate","duplicates","duration","during","dynamic","dynpro","edit","editor-call","else","elseif","empty","enabled","enabling","encoding","end","endat","endcase","endcatch","endchain","endclass","enddo","endenhancement","end-enhancement-section","endexec","endform","endfunction","endian","endif","ending","endinterface","end-lines","endloop","endmethod","endmodule","end-of-definition","end-of-editing","end-of-file","end-of-page","end-of-selection","endon","endprovide","endselect","end-test-injection","end-test-seam","endtry","endwhile","endwith","engineering","enhancement","enhancement-point","enhancements","enhancement-section","entries","entry","enum","environment","equiv","errormessage","errors","escaping","event","events","exact","except","exception","exceptions","exception-table","exclude","excluding","exec","execute","exists","exit","exit-command","expand","expanding","expiration","explicit","exponent","export","exporting","extend","extended","extension","extract","fail","fetch","field","field-groups","fields","field-symbol","field-symbols","file","filter","filters","filter-table","final","find","first","first-line","fixed-point","fkeq","fkge","flush","font","for","form","format","forward","found","frame","frames","free","friends","from","function","functionality","function-pool","further","gaps","generate","get","giving","gkeq","gkge","global","grant","green","group","groups","handle","handler","harmless","hashed","having","hdb","header","headers","heading","head-lines","help-id","help-request","hextobin","hide","high","hint","hold","hotspot","icon","id","identification","identifier","ids","if","ignore","ignoring","immediately","implementation","implementations","implemented","implicit","import","importing","in","inactive","incl","include","includes","including","increment","index","index-line","infotypes","inheriting","init","initial","initialization","inner","inout","input","insert","instance","instances","instr","intensified","interface","interface-pool","interfaces","internal","intervals","into","inverse","inverted-date","is","iso","job","join","keep","keeping","kernel","key","keys","keywords","kind","language","last","late","layout","leading","leave","left","left-justified","leftplus","leftspace","legacy","length","let","level","levels","like","line","lines","line-count","linefeed","line-selection","line-size","list","listbox","list-processing","little","llang","load","load-of-program","lob","local","locale","locator","logfile","logical","log-point","long","loop","low","lower","lpad","lpi","ltrim","mail","main","major-id","mapping","margin","mark","mask","match","matchcode","max","maximum","medium","members","memory","mesh","message","message-id","messages","messaging","method","methods","min","minimum","minor-id","mm/dd/yy","mm/dd/yyyy","mmddyy","mode","modif","modifier","modify","module","move","move-corresponding","multiply","multiply-corresponding","name","nametab","native","nested","nesting","new","new-line","new-page","new-section","next","no","no-display","no-extension","no-gap","no-gaps","no-grouping","no-heading","no-scrolling","no-sign","no-title","no-topofpage","no-zero","node","nodes","non-unicode","non-unique","not","null","number","object","objects","obligatory","occurrence","occurrences","occurs","of","off","offset","ole","on","only","open","option","optional","options","or","order","other","others","out","outer","output","output-length","overflow","overlay","pack","package","pad","padding","page","pages","parameter","parameters","parameter-table","part","partially","pattern","percentage","perform","performing","person","pf1","pf10","pf11","pf12","pf13","pf14","pf15","pf2","pf3","pf4","pf5","pf6","pf7","pf8","pf9","pf-status","pink","places","pool","pos_high","pos_low","position","pragmas","precompiled","preferred","preserving","primary","print","print-control","priority","private","procedure","process","program","property","protected","provide","public","push","pushbutton","put","queue-only","quickinfo","radiobutton","raise","raising","range","ranges","read","reader","read-only","receive","received","receiver","receiving","red","redefinition","reduce","reduced","ref","reference","refresh","regex","reject","remote","renaming","replace","replacement","replacing","report","request","requested","reserve","reset","resolution","respecting","responsible","result","results","resumable","resume","retry","return","returncode","returning","returns","right","right-justified","rightplus","rightspace","risk","rmc_communication_failure","rmc_invalid_status","rmc_system_failure","role","rollback","rows","rpad","rtrim","run","sap","sap-spool","saving","scale_preserving","scale_preserving_scientific","scan","scientific","scientific_with_leading_zero","scroll","scroll-boundary","scrolling","search","secondary","seconds","section","select","selection","selections","selection-screen","selection-set","selection-sets","selection-table","select-options","send","separate","separated","set","shared","shift","short","shortdump-id","sign_as_postfix","single","size","skip","skipping","smart","some","sort","sortable","sorted","source","specified","split","spool","spots","sql","sqlscript","stable","stamp","standard","starting","start-of-editing","start-of-selection","state","statement","statements","static","statics","statusinfo","step-loop","stop","structure","structures","style","subkey","submatches","submit","subroutine","subscreen","subtract","subtract-corresponding","suffix","sum","summary","summing","supplied","supply","suppress","switch","switchstates","symbol","syncpoints","syntax","syntax-check","syntax-trace","system-call","system-exceptions","system-exit","tab","tabbed","table","tables","tableview","tabstrip","target","task","tasks","test","testing","test-injection","test-seam","text","textpool","then","throw","time","times","timestamp","timezone","tims_is_valid","title","titlebar","title-lines","to","tokenization","tokens","top-lines","top-of-page","trace-file","trace-table","trailing","transaction","transfer","transformation","translate","transporting","trmac","truncate","truncation","try","tstmp_add_seconds","tstmp_current_utctimestamp","tstmp_is_valid","tstmp_seconds_between","type","type-pool","type-pools","types","uline","unassign","under","unicode","union","unique","unit_conversion","unix","unpack","until","unwind","up","update","upper","user","user-command","using","utf-8","valid","value","value-request","values","vary","varying","verification-message","version","via","view","visible","wait","warning","when","whenever","where","while","width","window","windows","with","with-heading","without","with-title","word","work","write","writer","xml","xsd","yellow","yes","yymmdd","zero","zone","abap_system_timezone","abap_user_timezone","access","action","adabas","adjust_numbers","allow_precision_loss","allowed","amdp","applicationuser","as_geo_json","as400","associations","balance","behavior","breakup","bulk","cds","cds_client","check_before_save","child","clients","corr","corr_spearman","cross","cycles","datn_add_days","datn_add_months","datn_days_between","dats_from_datn","dats_tims_to_tstmp","dats_to_datn","db2","db6","ddl","dense_rank","depth","deterministic","discarding","entities","entity","error","failed","finalize","first_value","fltp_to_dec","following","fractional","full","graph","grouping","hierarchy","hierarchy_ancestors","hierarchy_ancestors_aggregate","hierarchy_descendants","hierarchy_descendants_aggregate","hierarchy_siblings","incremental","indicators","lag","last_value","lead","leaves","like_regexpr","link","locale_sap","lock","locks","many","mapped","matched","measures","median","mssqlnt","multiple","nodetype","ntile","nulls","occurrences_regexpr","one","operations","oracle","orphans","over","parent","parents","partition","pcre","period","pfcg_mapping","preceding","privileged","product","projection","rank","redirected","replace_regexpr","reported","response","responses","root","row","row_number","sap_system_date","save","schema","session","sets","shortdump","siblings","spantree","start","stddev","string_agg","subtotal","sybase","tims_from_timn","tims_to_timn","to_blob","to_clob","total","trace-entry","tstmp_to_dats","tstmp_to_dst","tstmp_to_tims","tstmpl_from_utcl","tstmpl_to_utcl","unbounded","utcl_add_seconds","utcl_current","utcl_seconds_between","uuid","var","verbatim"],builtinFunctions:["abs","acos","asin","atan","bit-set","boolc","boolx","ceil","char_off","charlen","cmax","cmin","concat_lines_of","contains","contains_any_not_of","contains_any_of","cos","cosh","count","count_any_not_of","count_any_of","dbmaxlen","distance","escape","exp","find_any_not_of","find_any_of","find_end","floor","frac","from_mixed","ipow","line_exists","line_index","log","log10","matches","nmax","nmin","numofchar","repeat","rescale","reverse","round","segment","shift_left","shift_right","sign","sin","sinh","sqrt","strlen","substring","substring_after","substring_before","substring_from","substring_to","tan","tanh","to_lower","to_mixed","to_upper","trunc","utclong_add","utclong_current","utclong_diff","xsdbool","xstrlen"],typeKeywords:["b","c","d","decfloat16","decfloat34","f","i","int8","n","p","s","string","t","utclong","x","xstring","any","clike","csequence","decfloat","numeric","simple","xsequence","accp","char","clnt","cuky","curr","datn","dats","d16d","d16n","d16r","d34d","d34n","d34r","dec","df16_dec","df16_raw","df34_dec","df34_raw","fltp","geom_ewkb","int1","int2","int4","lang","lchr","lraw","numc","quan","raw","rawstring","sstring","timn","tims","unit","utcl","df16_scl","df34_scl","prec","varc","abap_bool","abap_false","abap_true","abap_undefined","me","screen","space","super","sy","syst","table_line","*sys*"],builtinMethods:["class_constructor","constructor"],derivedTypes:["%CID","%CID_REF","%CONTROL","%DATA","%ELEMENT","%FAIL","%KEY","%MSG","%PARAM","%PID","%PID_ASSOC","%PID_PARENT","%_HINTS"],cdsLanguage:["@AbapAnnotation","@AbapCatalog","@AccessControl","@API","@ClientDependent","@ClientHandling","@CompatibilityContract","@DataAging","@EndUserText","@Environment","@LanguageDependency","@MappingRole","@Metadata","@MetadataExtension","@ObjectModel","@Scope","@Semantics","$EXTENSION","$SELF"],selectors:["->","->*","=>","~","~*"],operators:[" +"," -","/","*","**","div","mod","=","#","@","+=","-=","*=","/=","**=","&&=","?=","&","&&","bit-and","bit-not","bit-or","bit-xor","m","o","z","<"," >","<=",">=","<>","><","=<","=>","bt","byte-ca","byte-cn","byte-co","byte-cs","byte-na","byte-ns","ca","cn","co","cp","cs","eq","ge","gt","le","lt","na","nb","ne","np","ns","*/","*:","--","/*","//"],symbols:/[=><!~?&+\-*\/\^%#@]+/,tokenizer:{root:[[/[a-z_\/$%@]([\w\/$%]|-(?!>))*/,{cases:{"@typeKeywords":"type","@keywords":"keyword","@cdsLanguage":"annotation","@derivedTypes":"type","@builtinFunctions":"type","@builtinMethods":"type","@operators":"key","@default":"identifier"}}],[/<[\w]+>/,"identifier"],[/##[\w|_]+/,"comment"],{include:"@whitespace"},[/[:,.]/,"delimiter"],[/[{}()\[\]]/,"@brackets"],[/@symbols/,{cases:{"@selectors":"tag","@operators":"key","@default":""}}],[/'/,{token:"string",bracket:"@open",next:"@stringquote"}],[/`/,{token:"string",bracket:"@open",next:"@stringping"}],[/\|/,{token:"string",bracket:"@open",next:"@stringtemplate"}],[/\d+/,"number"]],stringtemplate:[[/[^\\\|]+/,"string"],[/\\\|/,"string"],[/\|/,{token:"string",bracket:"@close",next:"@pop"}]],stringping:[[/[^\\`]+/,"string"],[/`/,{token:"string",bracket:"@close",next:"@pop"}]],stringquote:[[/[^\\']+/,"string"],[/'/,{token:"string",bracket:"@close",next:"@pop"}]],whitespace:[[/[ \t\r\n]+/,""],[/^\*.*$/,"comment"],[/\".*$/,"comment"]]}},Xze=Object.freeze(Object.defineProperty({__proto__:null,conf:Yze,language:Zze},Symbol.toStringTag,{value:"Module"})),Qze={wordPattern:/(-?\d*\.\d\w*)|([^\`\~\!\#\%\^\&\*\(\)\-\=\+\[\{\]\}\\\|\;\:\'\"\,\.\<\>\/\?\s]+)/g,comments:{lineComment:"//",blockComment:["/*","*/"]},brackets:[["{","}"],["[","]"],["(",")"]],autoClosingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:'"',close:'"'},{open:"'",close:"'"}],surroundingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:'"',close:'"'},{open:"'",close:"'"},{open:"<",close:">"}],folding:{markers:{start:new RegExp("^\\s*//\\s*(?:(?:#?region\\b)|(?:<editor-fold\\b))"),end:new RegExp("^\\s*//\\s*(?:(?:#?endregion\\b)|(?:</editor-fold>))")}}},Jze=["abstract","activate","and","any","array","as","asc","assert","autonomous","begin","bigdecimal","blob","boolean","break","bulk","by","case","cast","catch","char","class","collect","commit","const","continue","convertcurrency","decimal","default","delete","desc","do","double","else","end","enum","exception","exit","export","extends","false","final","finally","float","for","from","future","get","global","goto","group","having","hint","if","implements","import","in","inner","insert","instanceof","int","interface","into","join","last_90_days","last_month","last_n_days","last_week","like","limit","list","long","loop","map","merge","native","new","next_90_days","next_month","next_n_days","next_week","not","null","nulls","number","object","of","on","or","outer","override","package","parallel","pragma","private","protected","public","retrieve","return","returning","rollback","savepoint","search","select","set","short","sort","stat","static","strictfp","super","switch","synchronized","system","testmethod","then","this","this_month","this_week","throw","throws","today","tolabel","tomorrow","transaction","transient","trigger","true","try","type","undelete","update","upsert","using","virtual","void","volatile","webservice","when","where","while","yesterday"],e$e=s=>s.charAt(0).toUpperCase()+s.substr(1);let aE=[];Jze.forEach(s=>{aE.push(s),aE.push(s.toUpperCase()),aE.push(e$e(s))});const t$e={defaultToken:"",tokenPostfix:".apex",keywords:aE,operators:["=",">","<","!","~","?",":","==","<=",">=","!=","&&","||","++","--","+","-","*","/","&","|","^","%","<<",">>",">>>","+=","-=","*=","/=","&=","|=","^=","%=","<<=",">>=",">>>="],symbols:/[=><!~?:&|+\-*\/\^%]+/,escapes:/\\(?:[abfnrtv\\"']|x[0-9A-Fa-f]{1,4}|u[0-9A-Fa-f]{4}|U[0-9A-Fa-f]{8})/,digits:/\d+(_+\d+)*/,octaldigits:/[0-7]+(_+[0-7]+)*/,binarydigits:/[0-1]+(_+[0-1]+)*/,hexdigits:/[[0-9a-fA-F]+(_+[0-9a-fA-F]+)*/,tokenizer:{root:[[/[a-z_$][\w$]*/,{cases:{"@keywords":{token:"keyword.$0"},"@default":"identifier"}}],[/[A-Z][\w\$]*/,{cases:{"@keywords":{token:"keyword.$0"},"@default":"type.identifier"}}],{include:"@whitespace"},[/[{}()\[\]]/,"@brackets"],[/[<>](?!@symbols)/,"@brackets"],[/@symbols/,{cases:{"@operators":"delimiter","@default":""}}],[/@\s*[a-zA-Z_\$][\w\$]*/,"annotation"],[/(@digits)[eE]([\-+]?(@digits))?[fFdD]?/,"number.float"],[/(@digits)\.(@digits)([eE][\-+]?(@digits))?[fFdD]?/,"number.float"],[/(@digits)[fFdD]/,"number.float"],[/(@digits)[lL]?/,"number"],[/[;,.]/,"delimiter"],[/"([^"\\]|\\.)*$/,"string.invalid"],[/'([^'\\]|\\.)*$/,"string.invalid"],[/"/,"string",'@string."'],[/'/,"string","@string.'"],[/'[^\\']'/,"string"],[/(')(@escapes)(')/,["string","string.escape","string"]],[/'/,"string.invalid"]],whitespace:[[/[ \t\r\n]+/,""],[/\/\*\*(?!\/)/,"comment.doc","@apexdoc"],[/\/\*/,"comment","@comment"],[/\/\/.*$/,"comment"]],comment:[[/[^\/*]+/,"comment"],[/\*\//,"comment","@pop"],[/[\/*]/,"comment"]],apexdoc:[[/[^\/*]+/,"comment.doc"],[/\*\//,"comment.doc","@pop"],[/[\/*]/,"comment.doc"]],string:[[/[^\\"']+/,"string"],[/@escapes/,"string.escape"],[/\\./,"string.escape.invalid"],[/["']/,{cases:{"$#==$S2":{token:"string",next:"@pop"},"@default":"string"}}]]}},i$e=Object.freeze(Object.defineProperty({__proto__:null,conf:Qze,language:t$e},Symbol.toStringTag,{value:"Module"})),n$e={comments:{lineComment:"#"}},s$e={defaultToken:"keyword",ignoreCase:!0,tokenPostfix:".azcli",str:/[^#\s]/,tokenizer:{root:[{include:"@comment"},[/\s-+@str*\s*/,{cases:{"@eos":{token:"key.identifier",next:"@popall"},"@default":{token:"key.identifier",next:"@type"}}}],[/^-+@str*\s*/,{cases:{"@eos":{token:"key.identifier",next:"@popall"},"@default":{token:"key.identifier",next:"@type"}}}]],type:[{include:"@comment"},[/-+@str*\s*/,{cases:{"@eos":{token:"key.identifier",next:"@popall"},"@default":"key.identifier"}}],[/@str+\s*/,{cases:{"@eos":{token:"string",next:"@popall"},"@default":"string"}}]],comment:[[/#.*$/,{cases:{"@eos":{token:"comment",next:"@popall"}}}]]}},o$e=Object.freeze(Object.defineProperty({__proto__:null,conf:n$e,language:s$e},Symbol.toStringTag,{value:"Module"})),r$e={comments:{lineComment:"REM"},brackets:[["{","}"],["[","]"],["(",")"]],autoClosingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:'"',close:'"'}],surroundingPairs:[{open:"[",close:"]"},{open:"(",close:")"},{open:'"',close:'"'}],folding:{markers:{start:new RegExp("^\\s*(::\\s*|REM\\s+)#region"),end:new RegExp("^\\s*(::\\s*|REM\\s+)#endregion")}}},a$e={defaultToken:"",ignoreCase:!0,tokenPostfix:".bat",brackets:[{token:"delimiter.bracket",open:"{",close:"}"},{token:"delimiter.parenthesis",open:"(",close:")"},{token:"delimiter.square",open:"[",close:"]"}],keywords:/call|defined|echo|errorlevel|exist|for|goto|if|pause|set|shift|start|title|not|pushd|popd/,symbols:/[=><!~?&|+\-*\/\^;\.,]+/,escapes:/\\(?:[abfnrtv\\"']|x[0-9A-Fa-f]{1,4}|u[0-9A-Fa-f]{4}|U[0-9A-Fa-f]{8})/,tokenizer:{root:[[/^(\s*)(rem(?:\s.*|))$/,["","comment"]],[/(\@?)(@keywords)(?!\w)/,[{token:"keyword"},{token:"keyword.$2"}]],[/[ \t\r\n]+/,""],[/setlocal(?!\w)/,"keyword.tag-setlocal"],[/endlocal(?!\w)/,"keyword.tag-setlocal"],[/[a-zA-Z_]\w*/,""],[/:\w*/,"metatag"],[/%[^%]+%/,"variable"],[/%%[\w]+(?!\w)/,"variable"],[/[{}()\[\]]/,"@brackets"],[/@symbols/,"delimiter"],[/\d*\.\d+([eE][\-+]?\d+)?/,"number.float"],[/0[xX][0-9a-fA-F_]*[0-9a-fA-F]/,"number.hex"],[/\d+/,"number"],[/[;,.]/,"delimiter"],[/"/,"string",'@string."'],[/'/,"string","@string.'"]],string:[[/[^\\"'%]+/,{cases:{"@eos":{token:"string",next:"@popall"},"@default":"string"}}],[/@escapes/,"string.escape"],[/\\./,"string.escape.invalid"],[/%[\w ]+%/,"variable"],[/%%[\w]+(?!\w)/,"variable"],[/["']/,{cases:{"$#==$S2":{token:"string",next:"@pop"},"@default":"string"}}],[/$/,"string","@popall"]]}},l$e=Object.freeze(Object.defineProperty({__proto__:null,conf:r$e,language:a$e},Symbol.toStringTag,{value:"Module"})),c$e=s=>`\\b${s}\\b`,d$e="[_a-zA-Z]",h$e="[_a-zA-Z0-9]",u$e=c$e(`${d$e}${h$e}*`),g$e=["targetScope","resource","module","param","var","output","for","in","if","existing"],f$e=["true","false","null"],p$e="[ \\t\\r\\n]",m$e="[0-9]+",_$e={comments:{lineComment:"//",blockComment:["/*","*/"]},brackets:[["{","}"],["[","]"],["(",")"]],surroundingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:"'",close:"'"},{open:"'''",close:"'''"}],autoClosingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:"'",close:"'",notIn:["string","comment"]},{open:"'''",close:"'''",notIn:["string","comment"]}],autoCloseBefore:`:.,=}])' 
	`,indentationRules:{increaseIndentPattern:new RegExp("^((?!\\/\\/).)*(\\{[^}\"'`]*|\\([^)\"'`]*|\\[[^\\]\"'`]*)$"),decreaseIndentPattern:new RegExp("^((?!.*?\\/\\*).*\\*/)?\\s*[\\}\\]].*$")}},b$e={defaultToken:"",tokenPostfix:".bicep",brackets:[{open:"{",close:"}",token:"delimiter.curly"},{open:"[",close:"]",token:"delimiter.square"},{open:"(",close:")",token:"delimiter.parenthesis"}],symbols:/[=><!~?:&|+\-*/^%]+/,keywords:g$e,namedLiterals:f$e,escapes:"\\\\(u{[0-9A-Fa-f]+}|n|r|t|\\\\|'|\\${)",tokenizer:{root:[{include:"@expression"},{include:"@whitespace"}],stringVerbatim:[{regex:"(|'|'')[^']",action:{token:"string"}},{regex:"'''",action:{token:"string.quote",next:"@pop"}}],stringLiteral:[{regex:"\\${",action:{token:"delimiter.bracket",next:"@bracketCounting"}},{regex:"[^\\\\'$]+",action:{token:"string"}},{regex:"@escapes",action:{token:"string.escape"}},{regex:"\\\\.",action:{token:"string.escape.invalid"}},{regex:"'",action:{token:"string",next:"@pop"}}],bracketCounting:[{regex:"{",action:{token:"delimiter.bracket",next:"@bracketCounting"}},{regex:"}",action:{token:"delimiter.bracket",next:"@pop"}},{include:"expression"}],comment:[{regex:"[^\\*]+",action:{token:"comment"}},{regex:"\\*\\/",action:{token:"comment",next:"@pop"}},{regex:"[\\/*]",action:{token:"comment"}}],whitespace:[{regex:p$e},{regex:"\\/\\*",action:{token:"comment",next:"@comment"}},{regex:"\\/\\/.*$",action:{token:"comment"}}],expression:[{regex:"'''",action:{token:"string.quote",next:"@stringVerbatim"}},{regex:"'",action:{token:"string.quote",next:"@stringLiteral"}},{regex:m$e,action:{token:"number"}},{regex:u$e,action:{cases:{"@keywords":{token:"keyword"},"@namedLiterals":{token:"keyword"},"@default":{token:"identifier"}}}}]}},w$e=Object.freeze(Object.defineProperty({__proto__:null,conf:_$e,language:b$e},Symbol.toStringTag,{value:"Module"})),v$e={comments:{lineComment:"//",blockComment:["(*","*)"]},brackets:[["{","}"],["[","]"],["(",")"],["<",">"]],autoClosingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:"<",close:">"},{open:"'",close:"'"},{open:'"',close:'"'},{open:"(*",close:"*)"}],surroundingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:"<",close:">"},{open:"'",close:"'"},{open:'"',close:'"'},{open:"(*",close:"*)"}]},C$e={defaultToken:"",tokenPostfix:".cameligo",ignoreCase:!0,brackets:[{open:"{",close:"}",token:"delimiter.curly"},{open:"[",close:"]",token:"delimiter.square"},{open:"(",close:")",token:"delimiter.parenthesis"},{open:"<",close:">",token:"delimiter.angle"}],keywords:["abs","assert","block","Bytes","case","Crypto","Current","else","failwith","false","for","fun","if","in","let","let%entry","let%init","List","list","Map","map","match","match%nat","mod","not","operation","Operation","of","record","Set","set","sender","skip","source","String","then","to","true","type","with"],typeKeywords:["int","unit","string","tz","nat","bool"],operators:["=",">","<","<=",">=","<>",":",":=","and","mod","or","+","-","*","/","@","&","^","%","->","<-","&&","||"],symbols:/[=><:@\^&|+\-*\/\^%]+/,tokenizer:{root:[[/[a-zA-Z_][\w]*/,{cases:{"@keywords":{token:"keyword.$0"},"@default":"identifier"}}],{include:"@whitespace"},[/[{}()\[\]]/,"@brackets"],[/[<>](?!@symbols)/,"@brackets"],[/@symbols/,{cases:{"@operators":"delimiter","@default":""}}],[/\d*\.\d+([eE][\-+]?\d+)?/,"number.float"],[/\$[0-9a-fA-F]{1,16}/,"number.hex"],[/\d+/,"number"],[/[;,.]/,"delimiter"],[/'([^'\\]|\\.)*$/,"string.invalid"],[/'/,"string","@string"],[/'[^\\']'/,"string"],[/'/,"string.invalid"],[/\#\d+/,"string"]],comment:[[/[^\(\*]+/,"comment"],[/\*\)/,"comment","@pop"],[/\(\*/,"comment"]],string:[[/[^\\']+/,"string"],[/\\./,"string.escape.invalid"],[/'/,{token:"string.quote",bracket:"@close",next:"@pop"}]],whitespace:[[/[ \t\r\n]+/,"white"],[/\(\*/,"comment","@comment"],[/\/\/.*$/,"comment"]]}},y$e=Object.freeze(Object.defineProperty({__proto__:null,conf:v$e,language:C$e},Symbol.toStringTag,{value:"Module"})),S$e={comments:{lineComment:";;"},brackets:[["[","]"],["(",")"],["{","}"]],autoClosingPairs:[{open:"[",close:"]"},{open:'"',close:'"'},{open:"(",close:")"},{open:"{",close:"}"}],surroundingPairs:[{open:"[",close:"]"},{open:'"',close:'"'},{open:"(",close:")"},{open:"{",close:"}"}]},x$e={defaultToken:"",ignoreCase:!0,tokenPostfix:".clj",brackets:[{open:"[",close:"]",token:"delimiter.square"},{open:"(",close:")",token:"delimiter.parenthesis"},{open:"{",close:"}",token:"delimiter.curly"}],constants:["true","false","nil"],numbers:/^(?:[+\-]?\d+(?:(?:N|(?:[eE][+\-]?\d+))|(?:\.?\d*(?:M|(?:[eE][+\-]?\d+))?)|\/\d+|[xX][0-9a-fA-F]+|r[0-9a-zA-Z]+)?(?=[\\\[\]\s"#'(),;@^`{}~]|$))/,characters:/^(?:\\(?:backspace|formfeed|newline|return|space|tab|o[0-7]{3}|u[0-9A-Fa-f]{4}|x[0-9A-Fa-f]{4}|.)?(?=[\\\[\]\s"(),;@^`{}~]|$))/,escapes:/^\\(?:["'\\bfnrt]|x[0-9A-Fa-f]{1,4}|u[0-9A-Fa-f]{4}|U[0-9A-Fa-f]{8})/,qualifiedSymbols:/^(?:(?:[^\\\/\[\]\d\s"#'(),;@^`{}~][^\\\[\]\s"(),;@^`{}~]*(?:\.[^\\\/\[\]\d\s"#'(),;@^`{}~][^\\\[\]\s"(),;@^`{}~]*)*\/)?(?:\/|[^\\\/\[\]\d\s"#'(),;@^`{}~][^\\\[\]\s"(),;@^`{}~]*)*(?=[\\\[\]\s"(),;@^`{}~]|$))/,specialForms:[".","catch","def","do","if","monitor-enter","monitor-exit","new","quote","recur","set!","throw","try","var"],coreSymbols:["*","*'","*1","*2","*3","*agent*","*allow-unresolved-vars*","*assert*","*clojure-version*","*command-line-args*","*compile-files*","*compile-path*","*compiler-options*","*data-readers*","*default-data-reader-fn*","*e","*err*","*file*","*flush-on-newline*","*fn-loader*","*in*","*math-context*","*ns*","*out*","*print-dup*","*print-length*","*print-level*","*print-meta*","*print-namespace-maps*","*print-readably*","*read-eval*","*reader-resolver*","*source-path*","*suppress-read*","*unchecked-math*","*use-context-classloader*","*verbose-defrecords*","*warn-on-reflection*","+","+'","-","-'","->","->>","->ArrayChunk","->Eduction","->Vec","->VecNode","->VecSeq","-cache-protocol-fn","-reset-methods","..","/","<","<=","=","==",">",">=","EMPTY-NODE","Inst","StackTraceElement->vec","Throwable->map","accessor","aclone","add-classpath","add-watch","agent","agent-error","agent-errors","aget","alength","alias","all-ns","alter","alter-meta!","alter-var-root","amap","ancestors","and","any?","apply","areduce","array-map","as->","aset","aset-boolean","aset-byte","aset-char","aset-double","aset-float","aset-int","aset-long","aset-short","assert","assoc","assoc!","assoc-in","associative?","atom","await","await-for","await1","bases","bean","bigdec","bigint","biginteger","binding","bit-and","bit-and-not","bit-clear","bit-flip","bit-not","bit-or","bit-set","bit-shift-left","bit-shift-right","bit-test","bit-xor","boolean","boolean-array","boolean?","booleans","bound-fn","bound-fn*","bound?","bounded-count","butlast","byte","byte-array","bytes","bytes?","case","cast","cat","char","char-array","char-escape-string","char-name-string","char?","chars","chunk","chunk-append","chunk-buffer","chunk-cons","chunk-first","chunk-next","chunk-rest","chunked-seq?","class","class?","clear-agent-errors","clojure-version","coll?","comment","commute","comp","comparator","compare","compare-and-set!","compile","complement","completing","concat","cond","cond->","cond->>","condp","conj","conj!","cons","constantly","construct-proxy","contains?","count","counted?","create-ns","create-struct","cycle","dec","dec'","decimal?","declare","dedupe","default-data-readers","definline","definterface","defmacro","defmethod","defmulti","defn","defn-","defonce","defprotocol","defrecord","defstruct","deftype","delay","delay?","deliver","denominator","deref","derive","descendants","destructure","disj","disj!","dissoc","dissoc!","distinct","distinct?","doall","dorun","doseq","dosync","dotimes","doto","double","double-array","double?","doubles","drop","drop-last","drop-while","eduction","empty","empty?","ensure","ensure-reduced","enumeration-seq","error-handler","error-mode","eval","even?","every-pred","every?","ex-data","ex-info","extend","extend-protocol","extend-type","extenders","extends?","false?","ffirst","file-seq","filter","filterv","find","find-keyword","find-ns","find-protocol-impl","find-protocol-method","find-var","first","flatten","float","float-array","float?","floats","flush","fn","fn?","fnext","fnil","for","force","format","frequencies","future","future-call","future-cancel","future-cancelled?","future-done?","future?","gen-class","gen-interface","gensym","get","get-in","get-method","get-proxy-class","get-thread-bindings","get-validator","group-by","halt-when","hash","hash-combine","hash-map","hash-ordered-coll","hash-set","hash-unordered-coll","ident?","identical?","identity","if-let","if-not","if-some","ifn?","import","in-ns","inc","inc'","indexed?","init-proxy","inst-ms","inst-ms*","inst?","instance?","int","int-array","int?","integer?","interleave","intern","interpose","into","into-array","ints","io!","isa?","iterate","iterator-seq","juxt","keep","keep-indexed","key","keys","keyword","keyword?","last","lazy-cat","lazy-seq","let","letfn","line-seq","list","list*","list?","load","load-file","load-reader","load-string","loaded-libs","locking","long","long-array","longs","loop","macroexpand","macroexpand-1","make-array","make-hierarchy","map","map-entry?","map-indexed","map?","mapcat","mapv","max","max-key","memfn","memoize","merge","merge-with","meta","method-sig","methods","min","min-key","mix-collection-hash","mod","munge","name","namespace","namespace-munge","nat-int?","neg-int?","neg?","newline","next","nfirst","nil?","nnext","not","not-any?","not-empty","not-every?","not=","ns","ns-aliases","ns-imports","ns-interns","ns-map","ns-name","ns-publics","ns-refers","ns-resolve","ns-unalias","ns-unmap","nth","nthnext","nthrest","num","number?","numerator","object-array","odd?","or","parents","partial","partition","partition-all","partition-by","pcalls","peek","persistent!","pmap","pop","pop!","pop-thread-bindings","pos-int?","pos?","pr","pr-str","prefer-method","prefers","primitives-classnames","print","print-ctor","print-dup","print-method","print-simple","print-str","printf","println","println-str","prn","prn-str","promise","proxy","proxy-call-with-super","proxy-mappings","proxy-name","proxy-super","push-thread-bindings","pvalues","qualified-ident?","qualified-keyword?","qualified-symbol?","quot","rand","rand-int","rand-nth","random-sample","range","ratio?","rational?","rationalize","re-find","re-groups","re-matcher","re-matches","re-pattern","re-seq","read","read-line","read-string","reader-conditional","reader-conditional?","realized?","record?","reduce","reduce-kv","reduced","reduced?","reductions","ref","ref-history-count","ref-max-history","ref-min-history","ref-set","refer","refer-clojure","reify","release-pending-sends","rem","remove","remove-all-methods","remove-method","remove-ns","remove-watch","repeat","repeatedly","replace","replicate","require","reset!","reset-meta!","reset-vals!","resolve","rest","restart-agent","resultset-seq","reverse","reversible?","rseq","rsubseq","run!","satisfies?","second","select-keys","send","send-off","send-via","seq","seq?","seqable?","seque","sequence","sequential?","set","set-agent-send-executor!","set-agent-send-off-executor!","set-error-handler!","set-error-mode!","set-validator!","set?","short","short-array","shorts","shuffle","shutdown-agents","simple-ident?","simple-keyword?","simple-symbol?","slurp","some","some->","some->>","some-fn","some?","sort","sort-by","sorted-map","sorted-map-by","sorted-set","sorted-set-by","sorted?","special-symbol?","spit","split-at","split-with","str","string?","struct","struct-map","subs","subseq","subvec","supers","swap!","swap-vals!","symbol","symbol?","sync","tagged-literal","tagged-literal?","take","take-last","take-nth","take-while","test","the-ns","thread-bound?","time","to-array","to-array-2d","trampoline","transduce","transient","tree-seq","true?","type","unchecked-add","unchecked-add-int","unchecked-byte","unchecked-char","unchecked-dec","unchecked-dec-int","unchecked-divide-int","unchecked-double","unchecked-float","unchecked-inc","unchecked-inc-int","unchecked-int","unchecked-long","unchecked-multiply","unchecked-multiply-int","unchecked-negate","unchecked-negate-int","unchecked-remainder-int","unchecked-short","unchecked-subtract","unchecked-subtract-int","underive","unquote","unquote-splicing","unreduced","unsigned-bit-shift-right","update","update-in","update-proxy","uri?","use","uuid?","val","vals","var-get","var-set","var?","vary-meta","vec","vector","vector-of","vector?","volatile!","volatile?","vreset!","vswap!","when","when-first","when-let","when-not","when-some","while","with-bindings","with-bindings*","with-in-str","with-loading-context","with-local-vars","with-meta","with-open","with-out-str","with-precision","with-redefs","with-redefs-fn","xml-seq","zero?","zipmap"],tokenizer:{root:[{include:"@whitespace"},[/@numbers/,"number"],[/@characters/,"string"],{include:"@string"},[/[()\[\]{}]/,"@brackets"],[/\/#"(?:\.|(?:")|[^"\n])*"\/g/,"regexp"],[/[#'@^`~]/,"meta"],[/@qualifiedSymbols/,{cases:{"^:.+$":"constant","@specialForms":"keyword","@coreSymbols":"keyword","@constants":"constant","@default":"identifier"}}]],whitespace:[[/[\s,]+/,"white"],[/;.*$/,"comment"],[/\(comment\b/,"comment","@comment"]],comment:[[/\(/,"comment","@push"],[/\)/,"comment","@pop"],[/[^()]/,"comment"]],string:[[/"/,"string","@multiLineString"]],multiLineString:[[/"/,"string","@popall"],[/@escapes/,"string.escape"],[/./,"string"]]}},L$e=Object.freeze(Object.defineProperty({__proto__:null,conf:S$e,language:x$e},Symbol.toStringTag,{value:"Module"})),k$e={wordPattern:/(-?\d*\.\d\w*)|([^\`\~\!\@\#%\^\&\*\(\)\=\$\-\+\[\{\]\}\\\|\;\:\'\"\,\.\<\>\/\?\s]+)/g,comments:{blockComment:["###","###"],lineComment:"#"},brackets:[["{","}"],["[","]"],["(",")"]],autoClosingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:'"',close:'"'},{open:"'",close:"'"}],surroundingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:'"',close:'"'},{open:"'",close:"'"}],folding:{markers:{start:new RegExp("^\\s*#region\\b"),end:new RegExp("^\\s*#endregion\\b")}}},E$e={defaultToken:"",ignoreCase:!0,tokenPostfix:".coffee",brackets:[{open:"{",close:"}",token:"delimiter.curly"},{open:"[",close:"]",token:"delimiter.square"},{open:"(",close:")",token:"delimiter.parenthesis"}],regEx:/\/(?!\/\/)(?:[^\/\\]|\\.)*\/[igm]*/,keywords:["and","or","is","isnt","not","on","yes","@","no","off","true","false","null","this","new","delete","typeof","in","instanceof","return","throw","break","continue","debugger","if","else","switch","for","while","do","try","catch","finally","class","extends","super","undefined","then","unless","until","loop","of","by","when"],symbols:/[=><!~?&%|+\-*\/\^\.,\:]+/,escapes:/\\(?:[abfnrtv\\"'$]|x[0-9A-Fa-f]{1,4}|u[0-9A-Fa-f]{4}|U[0-9A-Fa-f]{8})/,tokenizer:{root:[[/\@[a-zA-Z_]\w*/,"variable.predefined"],[/[a-zA-Z_]\w*/,{cases:{this:"variable.predefined","@keywords":{token:"keyword.$0"},"@default":""}}],[/[ \t\r\n]+/,""],[/###/,"comment","@comment"],[/#.*$/,"comment"],["///",{token:"regexp",next:"@hereregexp"}],[/^(\s*)(@regEx)/,["","regexp"]],[/(\()(\s*)(@regEx)/,["@brackets","","regexp"]],[/(\,)(\s*)(@regEx)/,["delimiter","","regexp"]],[/(\=)(\s*)(@regEx)/,["delimiter","","regexp"]],[/(\:)(\s*)(@regEx)/,["delimiter","","regexp"]],[/(\[)(\s*)(@regEx)/,["@brackets","","regexp"]],[/(\!)(\s*)(@regEx)/,["delimiter","","regexp"]],[/(\&)(\s*)(@regEx)/,["delimiter","","regexp"]],[/(\|)(\s*)(@regEx)/,["delimiter","","regexp"]],[/(\?)(\s*)(@regEx)/,["delimiter","","regexp"]],[/(\{)(\s*)(@regEx)/,["@brackets","","regexp"]],[/(\;)(\s*)(@regEx)/,["","","regexp"]],[/}/,{cases:{"$S2==interpolatedstring":{token:"string",next:"@pop"},"@default":"@brackets"}}],[/[{}()\[\]]/,"@brackets"],[/@symbols/,"delimiter"],[/\d+[eE]([\-+]?\d+)?/,"number.float"],[/\d+\.\d+([eE][\-+]?\d+)?/,"number.float"],[/0[xX][0-9a-fA-F]+/,"number.hex"],[/0[0-7]+(?!\d)/,"number.octal"],[/\d+/,"number"],[/[,.]/,"delimiter"],[/"""/,"string",'@herestring."""'],[/'''/,"string","@herestring.'''"],[/"/,{cases:{"@eos":"string","@default":{token:"string",next:'@string."'}}}],[/'/,{cases:{"@eos":"string","@default":{token:"string",next:"@string.'"}}}]],string:[[/[^"'\#\\]+/,"string"],[/@escapes/,"string.escape"],[/\./,"string.escape.invalid"],[/\./,"string.escape.invalid"],[/#{/,{cases:{'$S2=="':{token:"string",next:"root.interpolatedstring"},"@default":"string"}}],[/["']/,{cases:{"$#==$S2":{token:"string",next:"@pop"},"@default":"string"}}],[/#/,"string"]],herestring:[[/("""|''')/,{cases:{"$1==$S2":{token:"string",next:"@pop"},"@default":"string"}}],[/[^#\\'"]+/,"string"],[/['"]+/,"string"],[/@escapes/,"string.escape"],[/\./,"string.escape.invalid"],[/#{/,{token:"string.quote",next:"root.interpolatedstring"}],[/#/,"string"]],comment:[[/[^#]+/,"comment"],[/###/,"comment","@pop"],[/#/,"comment"]],hereregexp:[[/[^\\\/#]+/,"regexp"],[/\\./,"regexp"],[/#.*$/,"comment"],["///[igm]*",{token:"regexp",next:"@pop"}],[/\//,"regexp"]]}},I$e=Object.freeze(Object.defineProperty({__proto__:null,conf:k$e,language:E$e},Symbol.toStringTag,{value:"Module"})),D$e={comments:{lineComment:"//",blockComment:["/*","*/"]},brackets:[["{","}"],["[","]"],["(",")"]],autoClosingPairs:[{open:"[",close:"]"},{open:"{",close:"}"},{open:"(",close:")"},{open:"'",close:"'",notIn:["string","comment"]},{open:'"',close:'"',notIn:["string"]}],surroundingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:'"',close:'"'},{open:"'",close:"'"}],folding:{markers:{start:new RegExp("^\\s*#pragma\\s+region\\b"),end:new RegExp("^\\s*#pragma\\s+endregion\\b")}}},T$e={defaultToken:"",tokenPostfix:".cpp",brackets:[{token:"delimiter.curly",open:"{",close:"}"},{token:"delimiter.parenthesis",open:"(",close:")"},{token:"delimiter.square",open:"[",close:"]"},{token:"delimiter.angle",open:"<",close:">"}],keywords:["abstract","amp","array","auto","bool","break","case","catch","char","class","const","constexpr","const_cast","continue","cpu","decltype","default","delegate","delete","do","double","dynamic_cast","each","else","enum","event","explicit","export","extern","false","final","finally","float","for","friend","gcnew","generic","goto","if","in","initonly","inline","int","interface","interior_ptr","internal","literal","long","mutable","namespace","new","noexcept","nullptr","__nullptr","operator","override","partial","pascal","pin_ptr","private","property","protected","public","ref","register","reinterpret_cast","restrict","return","safe_cast","sealed","short","signed","sizeof","static","static_assert","static_cast","struct","switch","template","this","thread_local","throw","tile_static","true","try","typedef","typeid","typename","union","unsigned","using","virtual","void","volatile","wchar_t","where","while","_asm","_based","_cdecl","_declspec","_fastcall","_if_exists","_if_not_exists","_inline","_multiple_inheritance","_pascal","_single_inheritance","_stdcall","_virtual_inheritance","_w64","__abstract","__alignof","__asm","__assume","__based","__box","__builtin_alignof","__cdecl","__clrcall","__declspec","__delegate","__event","__except","__fastcall","__finally","__forceinline","__gc","__hook","__identifier","__if_exists","__if_not_exists","__inline","__int128","__int16","__int32","__int64","__int8","__interface","__leave","__m128","__m128d","__m128i","__m256","__m256d","__m256i","__m512","__m512d","__m512i","__m64","__multiple_inheritance","__newslot","__nogc","__noop","__nounwind","__novtordisp","__pascal","__pin","__pragma","__property","__ptr32","__ptr64","__raise","__restrict","__resume","__sealed","__single_inheritance","__stdcall","__super","__thiscall","__try","__try_cast","__typeof","__unaligned","__unhook","__uuidof","__value","__virtual_inheritance","__w64","__wchar_t"],operators:["=",">","<","!","~","?",":","==","<=",">=","!=","&&","||","++","--","+","-","*","/","&","|","^","%","<<",">>","+=","-=","*=","/=","&=","|=","^=","%=","<<=",">>="],symbols:/[=><!~?:&|+\-*\/\^%]+/,escapes:/\\(?:[0abfnrtv\\"']|x[0-9A-Fa-f]{1,4}|u[0-9A-Fa-f]{4}|U[0-9A-Fa-f]{8})/,integersuffix:/([uU](ll|LL|l|L)|(ll|LL|l|L)?[uU]?)/,floatsuffix:/[fFlL]?/,encoding:/u|u8|U|L/,tokenizer:{root:[[/@encoding?R\"(?:([^ ()\\\t]*))\(/,{token:"string.raw.begin",next:"@raw.$1"}],[/[a-zA-Z_]\w*/,{cases:{"@keywords":{token:"keyword.$0"},"@default":"identifier"}}],[/^\s*#\s*include/,{token:"keyword.directive.include",next:"@include"}],[/^\s*#\s*\w+/,"keyword.directive"],{include:"@whitespace"},[/\[\s*\[/,{token:"annotation",next:"@annotation"}],[/[{}()<>\[\]]/,"@brackets"],[/@symbols/,{cases:{"@operators":"delimiter","@default":""}}],[/\d*\d+[eE]([\-+]?\d+)?(@floatsuffix)/,"number.float"],[/\d*\.\d+([eE][\-+]?\d+)?(@floatsuffix)/,"number.float"],[/0[xX][0-9a-fA-F']*[0-9a-fA-F](@integersuffix)/,"number.hex"],[/0[0-7']*[0-7](@integersuffix)/,"number.octal"],[/0[bB][0-1']*[0-1](@integersuffix)/,"number.binary"],[/\d[\d']*\d(@integersuffix)/,"number"],[/\d(@integersuffix)/,"number"],[/[;,.]/,"delimiter"],[/"([^"\\]|\\.)*$/,"string.invalid"],[/"/,"string","@string"],[/'[^\\']'/,"string"],[/(')(@escapes)(')/,["string","string.escape","string"]],[/'/,"string.invalid"]],whitespace:[[/[ \t\r\n]+/,""],[/\/\*\*(?!\/)/,"comment.doc","@doccomment"],[/\/\*/,"comment","@comment"],[/\/\/.*\\$/,"comment","@linecomment"],[/\/\/.*$/,"comment"]],comment:[[/[^\/*]+/,"comment"],[/\*\//,"comment","@pop"],[/[\/*]/,"comment"]],linecomment:[[/.*[^\\]$/,"comment","@pop"],[/[^]+/,"comment"]],doccomment:[[/[^\/*]+/,"comment.doc"],[/\*\//,"comment.doc","@pop"],[/[\/*]/,"comment.doc"]],string:[[/[^\\"]+/,"string"],[/@escapes/,"string.escape"],[/\\./,"string.escape.invalid"],[/"/,"string","@pop"]],raw:[[/[^)]+/,"string.raw"],[/\)$S2\"/,{token:"string.raw.end",next:"@pop"}],[/\)/,"string.raw"]],annotation:[{include:"@whitespace"},[/using|alignas/,"keyword"],[/[a-zA-Z0-9_]+/,"annotation"],[/[,:]/,"delimiter"],[/[()]/,"@brackets"],[/\]\s*\]/,{token:"annotation",next:"@pop"}]],include:[[/(\s*)(<)([^<>]*)(>)/,["","keyword.directive.include.begin","string.include.identifier",{token:"keyword.directive.include.end",next:"@pop"}]],[/(\s*)(")([^"]*)(")/,["","keyword.directive.include.begin","string.include.identifier",{token:"keyword.directive.include.end",next:"@pop"}]]]}},ple=Object.freeze(Object.defineProperty({__proto__:null,conf:D$e,language:T$e},Symbol.toStringTag,{value:"Module"})),N$e={wordPattern:/(-?\d*\.\d\w*)|([^\`\~\!\#\$\%\^\&\*\(\)\-\=\+\[\{\]\}\\\|\;\:\'\"\,\.\<\>\/\?\s]+)/g,comments:{lineComment:"//",blockComment:["/*","*/"]},brackets:[["{","}"],["[","]"],["(",")"]],autoClosingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:"'",close:"'",notIn:["string","comment"]},{open:'"',close:'"',notIn:["string","comment"]}],surroundingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:"<",close:">"},{open:"'",close:"'"},{open:'"',close:'"'}],folding:{markers:{start:new RegExp("^\\s*#region\\b"),end:new RegExp("^\\s*#endregion\\b")}}},R$e={defaultToken:"",tokenPostfix:".cs",brackets:[{open:"{",close:"}",token:"delimiter.curly"},{open:"[",close:"]",token:"delimiter.square"},{open:"(",close:")",token:"delimiter.parenthesis"},{open:"<",close:">",token:"delimiter.angle"}],keywords:["extern","alias","using","bool","decimal","sbyte","byte","short","ushort","int","uint","long","ulong","char","float","double","object","dynamic","string","assembly","is","as","ref","out","this","base","new","typeof","void","checked","unchecked","default","delegate","var","const","if","else","switch","case","while","do","for","foreach","in","break","continue","goto","return","throw","try","catch","finally","lock","yield","from","let","where","join","on","equals","into","orderby","ascending","descending","select","group","by","namespace","partial","class","field","event","method","param","public","protected","internal","private","abstract","sealed","static","struct","readonly","volatile","virtual","override","params","get","set","add","remove","operator","true","false","implicit","explicit","interface","enum","null","async","await","fixed","sizeof","stackalloc","unsafe","nameof","when"],namespaceFollows:["namespace","using"],parenFollows:["if","for","while","switch","foreach","using","catch","when"],operators:["=","??","||","&&","|","^","&","==","!=","<=",">=","<<","+","-","*","/","%","!","~","++","--","+=","-=","*=","/=","%=","&=","|=","^=","<<=",">>=",">>","=>"],symbols:/[=><!~?:&|+\-*\/\^%]+/,escapes:/\\(?:[abfnrtv\\"']|x[0-9A-Fa-f]{1,4}|u[0-9A-Fa-f]{4}|U[0-9A-Fa-f]{8})/,tokenizer:{root:[[/\@?[a-zA-Z_]\w*/,{cases:{"@namespaceFollows":{token:"keyword.$0",next:"@namespace"},"@keywords":{token:"keyword.$0",next:"@qualified"},"@default":{token:"identifier",next:"@qualified"}}}],{include:"@whitespace"},[/}/,{cases:{"$S2==interpolatedstring":{token:"string.quote",next:"@pop"},"$S2==litinterpstring":{token:"string.quote",next:"@pop"},"@default":"@brackets"}}],[/[{}()\[\]]/,"@brackets"],[/[<>](?!@symbols)/,"@brackets"],[/@symbols/,{cases:{"@operators":"delimiter","@default":""}}],[/[0-9_]*\.[0-9_]+([eE][\-+]?\d+)?[fFdD]?/,"number.float"],[/0[xX][0-9a-fA-F_]+/,"number.hex"],[/0[bB][01_]+/,"number.hex"],[/[0-9_]+/,"number"],[/[;,.]/,"delimiter"],[/"([^"\\]|\\.)*$/,"string.invalid"],[/"/,{token:"string.quote",next:"@string"}],[/\$\@"/,{token:"string.quote",next:"@litinterpstring"}],[/\@"/,{token:"string.quote",next:"@litstring"}],[/\$"/,{token:"string.quote",next:"@interpolatedstring"}],[/'[^\\']'/,"string"],[/(')(@escapes)(')/,["string","string.escape","string"]],[/'/,"string.invalid"]],qualified:[[/[a-zA-Z_][\w]*/,{cases:{"@keywords":{token:"keyword.$0"},"@default":"identifier"}}],[/\./,"delimiter"],["","","@pop"]],namespace:[{include:"@whitespace"},[/[A-Z]\w*/,"namespace"],[/[\.=]/,"delimiter"],["","","@pop"]],comment:[[/[^\/*]+/,"comment"],["\\*/","comment","@pop"],[/[\/*]/,"comment"]],string:[[/[^\\"]+/,"string"],[/@escapes/,"string.escape"],[/\\./,"string.escape.invalid"],[/"/,{token:"string.quote",next:"@pop"}]],litstring:[[/[^"]+/,"string"],[/""/,"string.escape"],[/"/,{token:"string.quote",next:"@pop"}]],litinterpstring:[[/[^"{]+/,"string"],[/""/,"string.escape"],[/{{/,"string.escape"],[/}}/,"string.escape"],[/{/,{token:"string.quote",next:"root.litinterpstring"}],[/"/,{token:"string.quote",next:"@pop"}]],interpolatedstring:[[/[^\\"{]+/,"string"],[/@escapes/,"string.escape"],[/\\./,"string.escape.invalid"],[/{{/,"string.escape"],[/}}/,"string.escape"],[/{/,{token:"string.quote",next:"root.interpolatedstring"}],[/"/,{token:"string.quote",next:"@pop"}]],whitespace:[[/^[ \t\v\f]*#((r)|(load))(?=\s)/,"directive.csx"],[/^[ \t\v\f]*#\w.*$/,"namespace.cpp"],[/[ \t\v\f\r\n]+/,""],[/\/\*/,"comment","@comment"],[/\/\/.*$/,"comment"]]}},A$e=Object.freeze(Object.defineProperty({__proto__:null,conf:N$e,language:R$e},Symbol.toStringTag,{value:"Module"})),M$e={brackets:[],autoClosingPairs:[],surroundingPairs:[]},O$e={keywords:[],typeKeywords:[],tokenPostfix:".csp",operators:[],symbols:/[=><!~?:&|+\-*\/\^%]+/,escapes:/\\(?:[abfnrtv\\"']|x[0-9A-Fa-f]{1,4}|u[0-9A-Fa-f]{4}|U[0-9A-Fa-f]{8})/,tokenizer:{root:[[/child-src/,"string.quote"],[/connect-src/,"string.quote"],[/default-src/,"string.quote"],[/font-src/,"string.quote"],[/frame-src/,"string.quote"],[/img-src/,"string.quote"],[/manifest-src/,"string.quote"],[/media-src/,"string.quote"],[/object-src/,"string.quote"],[/script-src/,"string.quote"],[/style-src/,"string.quote"],[/worker-src/,"string.quote"],[/base-uri/,"string.quote"],[/plugin-types/,"string.quote"],[/sandbox/,"string.quote"],[/disown-opener/,"string.quote"],[/form-action/,"string.quote"],[/frame-ancestors/,"string.quote"],[/report-uri/,"string.quote"],[/report-to/,"string.quote"],[/upgrade-insecure-requests/,"string.quote"],[/block-all-mixed-content/,"string.quote"],[/require-sri-for/,"string.quote"],[/reflected-xss/,"string.quote"],[/referrer/,"string.quote"],[/policy-uri/,"string.quote"],[/'self'/,"string.quote"],[/'unsafe-inline'/,"string.quote"],[/'unsafe-eval'/,"string.quote"],[/'strict-dynamic'/,"string.quote"],[/'unsafe-hashed-attributes'/,"string.quote"]]}},P$e=Object.freeze(Object.defineProperty({__proto__:null,conf:M$e,language:O$e},Symbol.toStringTag,{value:"Module"})),F$e={wordPattern:/(#?-?\d*\.\d\w*%?)|((::|[@#.!:])?[\w-?]+%?)|::|[@#.!:]/g,comments:{blockComment:["/*","*/"]},brackets:[["{","}"],["[","]"],["(",")"]],autoClosingPairs:[{open:"{",close:"}",notIn:["string","comment"]},{open:"[",close:"]",notIn:["string","comment"]},{open:"(",close:")",notIn:["string","comment"]},{open:'"',close:'"',notIn:["string","comment"]},{open:"'",close:"'",notIn:["string","comment"]}],surroundingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:'"',close:'"'},{open:"'",close:"'"}],folding:{markers:{start:new RegExp("^\\s*\\/\\*\\s*#region\\b\\s*(.*?)\\s*\\*\\/"),end:new RegExp("^\\s*\\/\\*\\s*#endregion\\b.*\\*\\/")}}},B$e={defaultToken:"",tokenPostfix:".css",ws:`[ 	
\r\f]*`,identifier:"-?-?([a-zA-Z]|(\\\\(([0-9a-fA-F]{1,6}\\s?)|[^[0-9a-fA-F])))([\\w\\-]|(\\\\(([0-9a-fA-F]{1,6}\\s?)|[^[0-9a-fA-F])))*",brackets:[{open:"{",close:"}",token:"delimiter.bracket"},{open:"[",close:"]",token:"delimiter.bracket"},{open:"(",close:")",token:"delimiter.parenthesis"},{open:"<",close:">",token:"delimiter.angle"}],tokenizer:{root:[{include:"@selector"}],selector:[{include:"@comments"},{include:"@import"},{include:"@strings"},["[@](keyframes|-webkit-keyframes|-moz-keyframes|-o-keyframes)",{token:"keyword",next:"@keyframedeclaration"}],["[@](page|content|font-face|-moz-document)",{token:"keyword"}],["[@](charset|namespace)",{token:"keyword",next:"@declarationbody"}],["(url-prefix)(\\()",["attribute.value",{token:"delimiter.parenthesis",next:"@urldeclaration"}]],["(url)(\\()",["attribute.value",{token:"delimiter.parenthesis",next:"@urldeclaration"}]],{include:"@selectorname"},["[\\*]","tag"],["[>\\+,]","delimiter"],["\\[",{token:"delimiter.bracket",next:"@selectorattribute"}],["{",{token:"delimiter.bracket",next:"@selectorbody"}]],selectorbody:[{include:"@comments"},["[*_]?@identifier@ws:(?=(\\s|\\d|[^{;}]*[;}]))","attribute.name","@rulevalue"],["}",{token:"delimiter.bracket",next:"@pop"}]],selectorname:[["(\\.|#(?=[^{])|%|(@identifier)|:)+","tag"]],selectorattribute:[{include:"@term"},["]",{token:"delimiter.bracket",next:"@pop"}]],term:[{include:"@comments"},["(url-prefix)(\\()",["attribute.value",{token:"delimiter.parenthesis",next:"@urldeclaration"}]],["(url)(\\()",["attribute.value",{token:"delimiter.parenthesis",next:"@urldeclaration"}]],{include:"@functioninvocation"},{include:"@numbers"},{include:"@name"},{include:"@strings"},["([<>=\\+\\-\\*\\/\\^\\|\\~,])","delimiter"],[",","delimiter"]],rulevalue:[{include:"@comments"},{include:"@strings"},{include:"@term"},["!important","keyword"],[";","delimiter","@pop"],["(?=})",{token:"",next:"@pop"}]],warndebug:[["[@](warn|debug)",{token:"keyword",next:"@declarationbody"}]],import:[["[@](import)",{token:"keyword",next:"@declarationbody"}]],urldeclaration:[{include:"@strings"},[`[^)\r
]+`,"string"],["\\)",{token:"delimiter.parenthesis",next:"@pop"}]],parenthizedterm:[{include:"@term"},["\\)",{token:"delimiter.parenthesis",next:"@pop"}]],declarationbody:[{include:"@term"},[";","delimiter","@pop"],["(?=})",{token:"",next:"@pop"}]],comments:[["\\/\\*","comment","@comment"],["\\/\\/+.*","comment"]],comment:[["\\*\\/","comment","@pop"],[/[^*/]+/,"comment"],[/./,"comment"]],name:[["@identifier","attribute.value"]],numbers:[["-?(\\d*\\.)?\\d+([eE][\\-+]?\\d+)?",{token:"attribute.value.number",next:"@units"}],["#[0-9a-fA-F_]+(?!\\w)","attribute.value.hex"]],units:[["(em|ex|ch|rem|fr|vmin|vmax|vw|vh|vm|cm|mm|in|px|pt|pc|deg|grad|rad|turn|s|ms|Hz|kHz|%)?","attribute.value.unit","@pop"]],keyframedeclaration:[["@identifier","attribute.value"],["{",{token:"delimiter.bracket",switchTo:"@keyframebody"}]],keyframebody:[{include:"@term"},["{",{token:"delimiter.bracket",next:"@selectorbody"}],["}",{token:"delimiter.bracket",next:"@pop"}]],functioninvocation:[["@identifier\\(",{token:"attribute.value",next:"@functionarguments"}]],functionarguments:[["\\$@identifier@ws:","attribute.name"],["[,]","delimiter"],{include:"@term"},["\\)",{token:"attribute.value",next:"@pop"}]],strings:[['~?"',{token:"string",next:"@stringenddoublequote"}],["~?'",{token:"string",next:"@stringendquote"}]],stringenddoublequote:[["\\\\.","string"],['"',{token:"string",next:"@pop"}],[/[^\\"]+/,"string"],[".","string"]],stringendquote:[["\\\\.","string"],["'",{token:"string",next:"@pop"}],[/[^\\']+/,"string"],[".","string"]]}},W$e=Object.freeze(Object.defineProperty({__proto__:null,conf:F$e,language:B$e},Symbol.toStringTag,{value:"Module"})),H$e={comments:{lineComment:"//",blockComment:["/*","*/"]},brackets:[["{","}"],["[","]"],["(",")"]],autoClosingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:'"',close:'"'},{open:"'",close:"'"},{open:"`",close:"`"}],surroundingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:'"',close:'"'},{open:"'",close:"'"},{open:"`",close:"`"}]},V$e={defaultToken:"",tokenPostfix:".cypher",ignoreCase:!0,brackets:[{open:"{",close:"}",token:"delimiter.curly"},{open:"[",close:"]",token:"delimiter.bracket"},{open:"(",close:")",token:"delimiter.parenthesis"}],keywords:["ALL","AND","AS","ASC","ASCENDING","BY","CALL","CASE","CONTAINS","CREATE","DELETE","DESC","DESCENDING","DETACH","DISTINCT","ELSE","END","ENDS","EXISTS","IN","IS","LIMIT","MANDATORY","MATCH","MERGE","NOT","ON","ON","OPTIONAL","OR","ORDER","REMOVE","RETURN","SET","SKIP","STARTS","THEN","UNION","UNWIND","WHEN","WHERE","WITH","XOR","YIELD"],builtinLiterals:["true","TRUE","false","FALSE","null","NULL"],builtinFunctions:["abs","acos","asin","atan","atan2","avg","ceil","coalesce","collect","cos","cot","count","degrees","e","endNode","exists","exp","floor","head","id","keys","labels","last","left","length","log","log10","lTrim","max","min","nodes","percentileCont","percentileDisc","pi","properties","radians","rand","range","relationships","replace","reverse","right","round","rTrim","sign","sin","size","split","sqrt","startNode","stDev","stDevP","substring","sum","tail","tan","timestamp","toBoolean","toFloat","toInteger","toLower","toString","toUpper","trim","type"],operators:["+","-","*","/","%","^","=","<>","<",">","<=",">=","->","<-","-->","<--"],escapes:/\\(?:[tbnrf\\"'`]|u[0-9A-Fa-f]{4}|U[0-9A-Fa-f]{8})/,digits:/\d+/,octaldigits:/[0-7]+/,hexdigits:/[0-9a-fA-F]+/,tokenizer:{root:[[/[{}[\]()]/,"@brackets"],{include:"common"}],common:[{include:"@whitespace"},{include:"@numbers"},{include:"@strings"},[/:[a-zA-Z_][\w]*/,"type.identifier"],[/[a-zA-Z_][\w]*(?=\()/,{cases:{"@builtinFunctions":"predefined.function"}}],[/[a-zA-Z_$][\w$]*/,{cases:{"@keywords":"keyword","@builtinLiterals":"predefined.literal","@default":"identifier"}}],[/`/,"identifier.escape","@identifierBacktick"],[/[;,.:|]/,"delimiter"],[/[<>=%+\-*/^]+/,{cases:{"@operators":"delimiter","@default":""}}]],numbers:[[/-?(@digits)[eE](-?(@digits))?/,"number.float"],[/-?(@digits)?\.(@digits)([eE]-?(@digits))?/,"number.float"],[/-?0x(@hexdigits)/,"number.hex"],[/-?0(@octaldigits)/,"number.octal"],[/-?(@digits)/,"number"]],strings:[[/"([^"\\]|\\.)*$/,"string.invalid"],[/'([^'\\]|\\.)*$/,"string.invalid"],[/"/,"string","@stringDouble"],[/'/,"string","@stringSingle"]],whitespace:[[/[ \t\r\n]+/,"white"],[/\/\*/,"comment","@comment"],[/\/\/.*$/,"comment"]],comment:[[/\/\/.*/,"comment"],[/[^/*]+/,"comment"],[/\*\//,"comment","@pop"],[/[/*]/,"comment"]],stringDouble:[[/[^\\"]+/,"string"],[/@escapes/,"string"],[/\\./,"string.invalid"],[/"/,"string","@pop"]],stringSingle:[[/[^\\']+/,"string"],[/@escapes/,"string"],[/\\./,"string.invalid"],[/'/,"string","@pop"]],identifierBacktick:[[/[^\\`]+/,"identifier.escape"],[/@escapes/,"identifier.escape"],[/\\./,"identifier.escape.invalid"],[/`/,"identifier.escape","@pop"]]}},z$e=Object.freeze(Object.defineProperty({__proto__:null,conf:H$e,language:V$e},Symbol.toStringTag,{value:"Module"})),$$e={comments:{lineComment:"//",blockComment:["/*","*/"]},brackets:[["{","}"],["[","]"],["(",")"]],autoClosingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:"'",close:"'",notIn:["string","comment"]},{open:'"',close:'"',notIn:["string"]},{open:"`",close:"`",notIn:["string","comment"]},{open:"/**",close:" */",notIn:["string"]}],surroundingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:"<",close:">"},{open:"'",close:"'"},{open:"(",close:")"},{open:'"',close:'"'},{open:"`",close:"`"}],folding:{markers:{start:/^\s*\s*#?region\b/,end:/^\s*\s*#?endregion\b/}}},U$e={defaultToken:"invalid",tokenPostfix:".dart",keywords:["abstract","dynamic","implements","show","as","else","import","static","assert","enum","in","super","async","export","interface","switch","await","extends","is","sync","break","external","library","this","case","factory","mixin","throw","catch","false","new","true","class","final","null","try","const","finally","on","typedef","continue","for","operator","var","covariant","Function","part","void","default","get","rethrow","while","deferred","hide","return","with","do","if","set","yield"],typeKeywords:["int","double","String","bool"],operators:["+","-","*","/","~/","%","++","--","==","!=",">","<",">=","<=","=","-=","/=","%=",">>=","^=","+=","*=","~/=","<<=","&=","!=","||","&&","&","|","^","~","<<",">>","!",">>>","??","?",":","|="],symbols:/[=><!~?:&|+\-*\/\^%]+/,escapes:/\\(?:[abfnrtv\\"']|x[0-9A-Fa-f]{1,4}|u[0-9A-Fa-f]{4}|U[0-9A-Fa-f]{8})/,digits:/\d+(_+\d+)*/,octaldigits:/[0-7]+(_+[0-7]+)*/,binarydigits:/[0-1]+(_+[0-1]+)*/,hexdigits:/[[0-9a-fA-F]+(_+[0-9a-fA-F]+)*/,regexpctl:/[(){}\[\]\$\^|\-*+?\.]/,regexpesc:/\\(?:[bBdDfnrstvwWn0\\\/]|@regexpctl|c[A-Z]|x[0-9a-fA-F]{2}|u[0-9a-fA-F]{4})/,tokenizer:{root:[[/[{}]/,"delimiter.bracket"],{include:"common"}],common:[[/[a-z_$][\w$]*/,{cases:{"@typeKeywords":"type.identifier","@keywords":"keyword","@default":"identifier"}}],[/[A-Z_$][\w\$]*/,"type.identifier"],{include:"@whitespace"},[/\/(?=([^\\\/]|\\.)+\/([gimsuy]*)(\s*)(\.|;|,|\)|\]|\}|$))/,{token:"regexp",bracket:"@open",next:"@regexp"}],[/@[a-zA-Z]+/,"annotation"],[/[()\[\]]/,"@brackets"],[/[<>](?!@symbols)/,"@brackets"],[/!(?=([^=]|$))/,"delimiter"],[/@symbols/,{cases:{"@operators":"delimiter","@default":""}}],[/(@digits)[eE]([\-+]?(@digits))?/,"number.float"],[/(@digits)\.(@digits)([eE][\-+]?(@digits))?/,"number.float"],[/0[xX](@hexdigits)n?/,"number.hex"],[/0[oO]?(@octaldigits)n?/,"number.octal"],[/0[bB](@binarydigits)n?/,"number.binary"],[/(@digits)n?/,"number"],[/[;,.]/,"delimiter"],[/"([^"\\]|\\.)*$/,"string.invalid"],[/'([^'\\]|\\.)*$/,"string.invalid"],[/"/,"string","@string_double"],[/'/,"string","@string_single"]],whitespace:[[/[ \t\r\n]+/,""],[/\/\*\*(?!\/)/,"comment.doc","@jsdoc"],[/\/\*/,"comment","@comment"],[/\/\/\/.*$/,"comment.doc"],[/\/\/.*$/,"comment"]],comment:[[/[^\/*]+/,"comment"],[/\*\//,"comment","@pop"],[/[\/*]/,"comment"]],jsdoc:[[/[^\/*]+/,"comment.doc"],[/\*\//,"comment.doc","@pop"],[/[\/*]/,"comment.doc"]],regexp:[[/(\{)(\d+(?:,\d*)?)(\})/,["regexp.escape.control","regexp.escape.control","regexp.escape.control"]],[/(\[)(\^?)(?=(?:[^\]\\\/]|\\.)+)/,["regexp.escape.control",{token:"regexp.escape.control",next:"@regexrange"}]],[/(\()(\?:|\?=|\?!)/,["regexp.escape.control","regexp.escape.control"]],[/[()]/,"regexp.escape.control"],[/@regexpctl/,"regexp.escape.control"],[/[^\\\/]/,"regexp"],[/@regexpesc/,"regexp.escape"],[/\\\./,"regexp.invalid"],[/(\/)([gimsuy]*)/,[{token:"regexp",bracket:"@close",next:"@pop"},"keyword.other"]]],regexrange:[[/-/,"regexp.escape.control"],[/\^/,"regexp.invalid"],[/@regexpesc/,"regexp.escape"],[/[^\]]/,"regexp"],[/\]/,{token:"regexp.escape.control",next:"@pop",bracket:"@close"}]],string_double:[[/[^\\"\$]+/,"string"],[/[^\\"]+/,"string"],[/@escapes/,"string.escape"],[/\\./,"string.escape.invalid"],[/"/,"string","@pop"],[/\$\w+/,"identifier"]],string_single:[[/[^\\'\$]+/,"string"],[/@escapes/,"string.escape"],[/\\./,"string.escape.invalid"],[/'/,"string","@pop"],[/\$\w+/,"identifier"]]}},q$e=Object.freeze(Object.defineProperty({__proto__:null,conf:$$e,language:U$e},Symbol.toStringTag,{value:"Module"})),j$e={brackets:[["{","}"],["[","]"],["(",")"]],autoClosingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:'"',close:'"'},{open:"'",close:"'"}],surroundingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:'"',close:'"'},{open:"'",close:"'"}]},G$e={defaultToken:"",tokenPostfix:".dockerfile",variable:/\${?[\w]+}?/,tokenizer:{root:[{include:"@whitespace"},{include:"@comment"},[/(ONBUILD)(\s+)/,["keyword",""]],[/(ENV)(\s+)([\w]+)/,["keyword","",{token:"variable",next:"@arguments"}]],[/(FROM|MAINTAINER|RUN|EXPOSE|ENV|ADD|ARG|VOLUME|LABEL|USER|WORKDIR|COPY|CMD|STOPSIGNAL|SHELL|HEALTHCHECK|ENTRYPOINT)/,{token:"keyword",next:"@arguments"}]],arguments:[{include:"@whitespace"},{include:"@strings"},[/(@variable)/,{cases:{"@eos":{token:"variable",next:"@popall"},"@default":"variable"}}],[/\\/,{cases:{"@eos":"","@default":""}}],[/./,{cases:{"@eos":{token:"",next:"@popall"},"@default":""}}]],whitespace:[[/\s+/,{cases:{"@eos":{token:"",next:"@popall"},"@default":""}}]],comment:[[/(^#.*$)/,"comment","@popall"]],strings:[[/\\'$/,"","@popall"],[/\\'/,""],[/'$/,"string","@popall"],[/'/,"string","@stringBody"],[/"$/,"string","@popall"],[/"/,"string","@dblStringBody"]],stringBody:[[/[^\\\$']/,{cases:{"@eos":{token:"string",next:"@popall"},"@default":"string"}}],[/\\./,"string.escape"],[/'$/,"string","@popall"],[/'/,"string","@pop"],[/(@variable)/,"variable"],[/\\$/,"string"],[/$/,"string","@popall"]],dblStringBody:[[/[^\\\$"]/,{cases:{"@eos":{token:"string",next:"@popall"},"@default":"string"}}],[/\\./,"string.escape"],[/"$/,"string","@popall"],[/"/,"string","@pop"],[/(@variable)/,"variable"],[/\\$/,"string"],[/$/,"string","@popall"]]}},K$e=Object.freeze(Object.defineProperty({__proto__:null,conf:j$e,language:G$e},Symbol.toStringTag,{value:"Module"})),Y$e={comments:{lineComment:"//",blockComment:["/*","*/"]},brackets:[["{","}"],["[","]"],["(",")"]],autoClosingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:"'",close:"'",notIn:["string","comment"]},{open:'"',close:'"',notIn:["string","comment"]}],surroundingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:"<",close:">"},{open:"'",close:"'"},{open:'"',close:'"'}]},Z$e={defaultToken:"",tokenPostfix:".ecl",ignoreCase:!0,brackets:[{open:"{",close:"}",token:"delimiter.curly"},{open:"[",close:"]",token:"delimiter.square"},{open:"(",close:")",token:"delimiter.parenthesis"},{open:"<",close:">",token:"delimiter.angle"}],pounds:["append","break","declare","demangle","end","for","getdatatype","if","inmodule","loop","mangle","onwarning","option","set","stored","uniquename"].join("|"),keywords:["__compressed__","after","all","and","any","as","atmost","before","beginc","best","between","case","cluster","compressed","compression","const","counter","csv","default","descend","embed","encoding","encrypt","end","endc","endembed","endmacro","enum","escape","except","exclusive","expire","export","extend","fail","few","fileposition","first","flat","forward","from","full","function","functionmacro","group","grouped","heading","hole","ifblock","import","in","inner","interface","internal","joined","keep","keyed","last","left","limit","linkcounted","literal","little_endian","load","local","locale","lookup","lzw","macro","many","maxcount","maxlength","min skew","module","mofn","multiple","named","namespace","nocase","noroot","noscan","nosort","not","noxpath","of","onfail","only","opt","or","outer","overwrite","packed","partition","penalty","physicallength","pipe","prefetch","quote","record","repeat","retry","return","right","right1","right2","rows","rowset","scan","scope","self","separator","service","shared","skew","skip","smart","soapaction","sql","stable","store","terminator","thor","threshold","timelimit","timeout","token","transform","trim","type","unicodeorder","unordered","unsorted","unstable","update","use","validate","virtual","whole","width","wild","within","wnotrim","xml","xpath"],functions:["abs","acos","aggregate","allnodes","apply","ascii","asin","assert","asstring","atan","atan2","ave","build","buildindex","case","catch","choose","choosen","choosesets","clustersize","combine","correlation","cos","cosh","count","covariance","cron","dataset","dedup","define","denormalize","dictionary","distribute","distributed","distribution","ebcdic","enth","error","evaluate","event","eventextra","eventname","exists","exp","fail","failcode","failmessage","fetch","fromunicode","fromxml","getenv","getisvalid","global","graph","group","hash","hash32","hash64","hashcrc","hashmd5","having","httpcall","httpheader","if","iff","index","intformat","isvalid","iterate","join","keydiff","keypatch","keyunicode","length","library","limit","ln","loadxml","local","log","loop","map","matched","matchlength","matchposition","matchtext","matchunicode","max","merge","mergejoin","min","nofold","nolocal","nonempty","normalize","nothor","notify","output","parallel","parse","pipe","power","preload","process","project","pull","random","range","rank","ranked","realformat","recordof","regexfind","regexreplace","regroup","rejected","rollup","round","roundup","row","rowdiff","sample","sequential","set","sin","sinh","sizeof","soapcall","sort","sorted","sqrt","stepped","stored","sum","table","tan","tanh","thisnode","topn","tounicode","toxml","transfer","transform","trim","truncate","typeof","ungroup","unicodeorder","variance","wait","which","workunit","xmldecode","xmlencode","xmltext","xmlunicode"],typesint:["integer","unsigned"].join("|"),typesnum:["data","qstring","string","unicode","utf8","varstring","varunicode"],typesone:["ascii","big_endian","boolean","data","decimal","ebcdic","grouped","integer","linkcounted","pattern","qstring","real","record","rule","set of","streamed","string","token","udecimal","unicode","unsigned","utf8","varstring","varunicode"].join("|"),operators:["+","-","/",":=","<","<>","=",">","\\","and","in","not","or"],symbols:/[=><!~?:&|+\-*\/\^%]+/,escapes:/\\(?:[abfnrtv\\"']|x[0-9A-Fa-f]{1,4}|u[0-9A-Fa-f]{4}|U[0-9A-Fa-f]{8})/,tokenizer:{root:[[/@typesint[4|8]/,"type"],[/#(@pounds)/,"type"],[/@typesone/,"type"],[/[a-zA-Z_$][\w-$]*/,{cases:{"@functions":"keyword.function","@keywords":"keyword","@operators":"operator"}}],{include:"@whitespace"},[/[{}()\[\]]/,"@brackets"],[/[<>](?!@symbols)/,"@brackets"],[/@symbols/,{cases:{"@operators":"delimiter","@default":""}}],[/[0-9_]*\.[0-9_]+([eE][\-+]?\d+)?/,"number.float"],[/0[xX][0-9a-fA-F_]+/,"number.hex"],[/0[bB][01]+/,"number.hex"],[/[0-9_]+/,"number"],[/[;,.]/,"delimiter"],[/"([^"\\]|\\.)*$/,"string.invalid"],[/"/,"string","@string"],[/'[^\\']'/,"string"],[/(')(@escapes)(')/,["string","string.escape","string"]],[/'/,"string.invalid"]],whitespace:[[/[ \t\v\f\r\n]+/,""],[/\/\*/,"comment","@comment"],[/\/\/.*$/,"comment"]],comment:[[/[^\/*]+/,"comment"],[/\*\//,"comment","@pop"],[/[\/*]/,"comment"]],string:[[/[^\\']+/,"string"],[/@escapes/,"string.escape"],[/\\./,"string.escape.invalid"],[/'/,"string","@pop"]]}},X$e=Object.freeze(Object.defineProperty({__proto__:null,conf:Y$e,language:Z$e},Symbol.toStringTag,{value:"Module"})),Q$e={comments:{lineComment:"#"},brackets:[["{","}"],["[","]"],["(",")"]],surroundingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:"'",close:"'"},{open:'"',close:'"'}],autoClosingPairs:[{open:"'",close:"'",notIn:["string","comment"]},{open:'"',close:'"',notIn:["comment"]},{open:'"""',close:'"""'},{open:"`",close:"`",notIn:["string","comment"]},{open:"(",close:")"},{open:"{",close:"}"},{open:"[",close:"]"},{open:"<<",close:">>"}],indentationRules:{increaseIndentPattern:/^\s*(after|else|catch|rescue|fn|[^#]*(do|<\-|\->|\{|\[|\=))\s*$/,decreaseIndentPattern:/^\s*((\}|\])\s*$|(after|else|catch|rescue|end)\b)/}},J$e={defaultToken:"source",tokenPostfix:".elixir",brackets:[{open:"[",close:"]",token:"delimiter.square"},{open:"(",close:")",token:"delimiter.parenthesis"},{open:"{",close:"}",token:"delimiter.curly"},{open:"<<",close:">>",token:"delimiter.angle.special"}],declarationKeywords:["def","defp","defn","defnp","defguard","defguardp","defmacro","defmacrop","defdelegate","defcallback","defmacrocallback","defmodule","defprotocol","defexception","defimpl","defstruct"],operatorKeywords:["and","in","not","or","when"],namespaceKeywords:["alias","import","require","use"],otherKeywords:["after","case","catch","cond","do","else","end","fn","for","if","quote","raise","receive","rescue","super","throw","try","unless","unquote_splicing","unquote","with"],constants:["true","false","nil"],nameBuiltin:["__MODULE__","__DIR__","__ENV__","__CALLER__","__STACKTRACE__"],operator:/-[->]?|!={0,2}|\*{1,2}|\/|\\\\|&{1,3}|\.\.?|\^(?:\^\^)?|\+\+?|<(?:-|<<|=|>|\|>|~>?)?|=~|={1,3}|>(?:=|>>)?|\|~>|\|>|\|{1,3}|~>>?|~~~|::/,variableName:/[a-z_][a-zA-Z0-9_]*[?!]?/,atomName:/[a-zA-Z_][a-zA-Z0-9_@]*[?!]?|@specialAtomName|@operator/,specialAtomName:/\.\.\.|<<>>|%\{\}|%|\{\}/,aliasPart:/[A-Z][a-zA-Z0-9_]*/,moduleName:/@aliasPart(?:\.@aliasPart)*/,sigilSymmetricDelimiter:/"""|'''|"|'|\/|\|/,sigilStartDelimiter:/@sigilSymmetricDelimiter|<|\{|\[|\(/,sigilEndDelimiter:/@sigilSymmetricDelimiter|>|\}|\]|\)/,sigilModifiers:/[a-zA-Z0-9]*/,decimal:/\d(?:_?\d)*/,hex:/[0-9a-fA-F](_?[0-9a-fA-F])*/,octal:/[0-7](_?[0-7])*/,binary:/[01](_?[01])*/,escape:/\\u[0-9a-fA-F]{4}|\\x[0-9a-fA-F]{2}|\\./,tokenizer:{root:[{include:"@whitespace"},{include:"@comments"},{include:"@keywordsShorthand"},{include:"@numbers"},{include:"@identifiers"},{include:"@strings"},{include:"@atoms"},{include:"@sigils"},{include:"@attributes"},{include:"@symbols"}],whitespace:[[/\s+/,"white"]],comments:[[/(#)(.*)/,["comment.punctuation","comment"]]],keywordsShorthand:[[/(@atomName)(:)(\s+)/,["constant","constant.punctuation","white"]],[/"(?=([^"]|#\{.*?\}|\\")*":)/,{token:"constant.delimiter",next:"@doubleQuotedStringKeyword"}],[/'(?=([^']|#\{.*?\}|\\')*':)/,{token:"constant.delimiter",next:"@singleQuotedStringKeyword"}]],doubleQuotedStringKeyword:[[/":/,{token:"constant.delimiter",next:"@pop"}],{include:"@stringConstantContentInterpol"}],singleQuotedStringKeyword:[[/':/,{token:"constant.delimiter",next:"@pop"}],{include:"@stringConstantContentInterpol"}],numbers:[[/0b@binary/,"number.binary"],[/0o@octal/,"number.octal"],[/0x@hex/,"number.hex"],[/@decimal\.@decimal([eE]-?@decimal)?/,"number.float"],[/@decimal/,"number"]],identifiers:[[/\b(defp?|defnp?|defmacrop?|defguardp?|defdelegate)(\s+)(@variableName)(?!\s+@operator)/,["keyword.declaration","white",{cases:{unquote:"keyword","@default":"function"}}]],[/(@variableName)(?=\s*\.?\s*\()/,{cases:{"@declarationKeywords":"keyword.declaration","@namespaceKeywords":"keyword","@otherKeywords":"keyword","@default":"function.call"}}],[/(@moduleName)(\s*)(\.)(\s*)(@variableName)/,["type.identifier","white","operator","white","function.call"]],[/(:)(@atomName)(\s*)(\.)(\s*)(@variableName)/,["constant.punctuation","constant","white","operator","white","function.call"]],[/(\|>)(\s*)(@variableName)/,["operator","white",{cases:{"@otherKeywords":"keyword","@default":"function.call"}}]],[/(&)(\s*)(@variableName)/,["operator","white","function.call"]],[/@variableName/,{cases:{"@declarationKeywords":"keyword.declaration","@operatorKeywords":"keyword.operator","@namespaceKeywords":"keyword","@otherKeywords":"keyword","@constants":"constant.language","@nameBuiltin":"variable.language","_.*":"comment.unused","@default":"identifier"}}],[/@moduleName/,"type.identifier"]],strings:[[/"""/,{token:"string.delimiter",next:"@doubleQuotedHeredoc"}],[/'''/,{token:"string.delimiter",next:"@singleQuotedHeredoc"}],[/"/,{token:"string.delimiter",next:"@doubleQuotedString"}],[/'/,{token:"string.delimiter",next:"@singleQuotedString"}]],doubleQuotedHeredoc:[[/"""/,{token:"string.delimiter",next:"@pop"}],{include:"@stringContentInterpol"}],singleQuotedHeredoc:[[/'''/,{token:"string.delimiter",next:"@pop"}],{include:"@stringContentInterpol"}],doubleQuotedString:[[/"/,{token:"string.delimiter",next:"@pop"}],{include:"@stringContentInterpol"}],singleQuotedString:[[/'/,{token:"string.delimiter",next:"@pop"}],{include:"@stringContentInterpol"}],atoms:[[/(:)(@atomName)/,["constant.punctuation","constant"]],[/:"/,{token:"constant.delimiter",next:"@doubleQuotedStringAtom"}],[/:'/,{token:"constant.delimiter",next:"@singleQuotedStringAtom"}]],doubleQuotedStringAtom:[[/"/,{token:"constant.delimiter",next:"@pop"}],{include:"@stringConstantContentInterpol"}],singleQuotedStringAtom:[[/'/,{token:"constant.delimiter",next:"@pop"}],{include:"@stringConstantContentInterpol"}],sigils:[[/~[a-z]@sigilStartDelimiter/,{token:"@rematch",next:"@sigil.interpol"}],[/~([A-Z]+)@sigilStartDelimiter/,{token:"@rematch",next:"@sigil.noInterpol"}]],sigil:[[/~([a-z]|[A-Z]+)\{/,{token:"@rematch",switchTo:"@sigilStart.$S2.$1.{.}"}],[/~([a-z]|[A-Z]+)\[/,{token:"@rematch",switchTo:"@sigilStart.$S2.$1.[.]"}],[/~([a-z]|[A-Z]+)\(/,{token:"@rematch",switchTo:"@sigilStart.$S2.$1.(.)"}],[/~([a-z]|[A-Z]+)\</,{token:"@rematch",switchTo:"@sigilStart.$S2.$1.<.>"}],[/~([a-z]|[A-Z]+)(@sigilSymmetricDelimiter)/,{token:"@rematch",switchTo:"@sigilStart.$S2.$1.$2.$2"}]],"sigilStart.interpol.s":[[/~s@sigilStartDelimiter/,{token:"string.delimiter",switchTo:"@sigilContinue.$S2.$S3.$S4.$S5"}]],"sigilContinue.interpol.s":[[/(@sigilEndDelimiter)@sigilModifiers/,{cases:{"$1==$S5":{token:"string.delimiter",next:"@pop"},"@default":"string"}}],{include:"@stringContentInterpol"}],"sigilStart.noInterpol.S":[[/~S@sigilStartDelimiter/,{token:"string.delimiter",switchTo:"@sigilContinue.$S2.$S3.$S4.$S5"}]],"sigilContinue.noInterpol.S":[[/(^|[^\\])\\@sigilEndDelimiter/,"string"],[/(@sigilEndDelimiter)@sigilModifiers/,{cases:{"$1==$S5":{token:"string.delimiter",next:"@pop"},"@default":"string"}}],{include:"@stringContent"}],"sigilStart.interpol.r":[[/~r@sigilStartDelimiter/,{token:"regexp.delimiter",switchTo:"@sigilContinue.$S2.$S3.$S4.$S5"}]],"sigilContinue.interpol.r":[[/(@sigilEndDelimiter)@sigilModifiers/,{cases:{"$1==$S5":{token:"regexp.delimiter",next:"@pop"},"@default":"regexp"}}],{include:"@regexpContentInterpol"}],"sigilStart.noInterpol.R":[[/~R@sigilStartDelimiter/,{token:"regexp.delimiter",switchTo:"@sigilContinue.$S2.$S3.$S4.$S5"}]],"sigilContinue.noInterpol.R":[[/(^|[^\\])\\@sigilEndDelimiter/,"regexp"],[/(@sigilEndDelimiter)@sigilModifiers/,{cases:{"$1==$S5":{token:"regexp.delimiter",next:"@pop"},"@default":"regexp"}}],{include:"@regexpContent"}],"sigilStart.interpol":[[/~([a-z]|[A-Z]+)@sigilStartDelimiter/,{token:"sigil.delimiter",switchTo:"@sigilContinue.$S2.$S3.$S4.$S5"}]],"sigilContinue.interpol":[[/(@sigilEndDelimiter)@sigilModifiers/,{cases:{"$1==$S5":{token:"sigil.delimiter",next:"@pop"},"@default":"sigil"}}],{include:"@sigilContentInterpol"}],"sigilStart.noInterpol":[[/~([a-z]|[A-Z]+)@sigilStartDelimiter/,{token:"sigil.delimiter",switchTo:"@sigilContinue.$S2.$S3.$S4.$S5"}]],"sigilContinue.noInterpol":[[/(^|[^\\])\\@sigilEndDelimiter/,"sigil"],[/(@sigilEndDelimiter)@sigilModifiers/,{cases:{"$1==$S5":{token:"sigil.delimiter",next:"@pop"},"@default":"sigil"}}],{include:"@sigilContent"}],attributes:[[/\@(module|type)?doc (~[sS])?"""/,{token:"comment.block.documentation",next:"@doubleQuotedHeredocDocstring"}],[/\@(module|type)?doc (~[sS])?'''/,{token:"comment.block.documentation",next:"@singleQuotedHeredocDocstring"}],[/\@(module|type)?doc (~[sS])?"/,{token:"comment.block.documentation",next:"@doubleQuotedStringDocstring"}],[/\@(module|type)?doc (~[sS])?'/,{token:"comment.block.documentation",next:"@singleQuotedStringDocstring"}],[/\@(module|type)?doc false/,"comment.block.documentation"],[/\@(@variableName)/,"variable"]],doubleQuotedHeredocDocstring:[[/"""/,{token:"comment.block.documentation",next:"@pop"}],{include:"@docstringContent"}],singleQuotedHeredocDocstring:[[/'''/,{token:"comment.block.documentation",next:"@pop"}],{include:"@docstringContent"}],doubleQuotedStringDocstring:[[/"/,{token:"comment.block.documentation",next:"@pop"}],{include:"@docstringContent"}],singleQuotedStringDocstring:[[/'/,{token:"comment.block.documentation",next:"@pop"}],{include:"@docstringContent"}],symbols:[[/\?(\\.|[^\\\s])/,"number.constant"],[/&\d+/,"operator"],[/<<<|>>>/,"operator"],[/[()\[\]\{\}]|<<|>>/,"@brackets"],[/\.\.\./,"identifier"],[/=>/,"punctuation"],[/@operator/,"operator"],[/[:;,.%]/,"punctuation"]],stringContentInterpol:[{include:"@interpolation"},{include:"@escapeChar"},{include:"@stringContent"}],stringContent:[[/./,"string"]],stringConstantContentInterpol:[{include:"@interpolation"},{include:"@escapeChar"},{include:"@stringConstantContent"}],stringConstantContent:[[/./,"constant"]],regexpContentInterpol:[{include:"@interpolation"},{include:"@escapeChar"},{include:"@regexpContent"}],regexpContent:[[/(\s)(#)(\s.*)$/,["white","comment.punctuation","comment"]],[/./,"regexp"]],sigilContentInterpol:[{include:"@interpolation"},{include:"@escapeChar"},{include:"@sigilContent"}],sigilContent:[[/./,"sigil"]],docstringContent:[[/./,"comment.block.documentation"]],escapeChar:[[/@escape/,"constant.character.escape"]],interpolation:[[/#{/,{token:"delimiter.bracket.embed",next:"@interpolationContinue"}]],interpolationContinue:[[/}/,{token:"delimiter.bracket.embed",next:"@pop"}],{include:"@root"}]}},eUe=Object.freeze(Object.defineProperty({__proto__:null,conf:Q$e,language:J$e},Symbol.toStringTag,{value:"Module"})),tUe={comments:{blockComment:["/*","*/"],lineComment:"//"},brackets:[["{","}"],["[","]"],["(",")"]],autoClosingPairs:[{open:"{",close:"}",notIn:["string"]},{open:"[",close:"]",notIn:["string"]},{open:"(",close:")",notIn:["string"]},{open:'"',close:'"',notIn:["string"]},{open:"'",close:"'",notIn:["string"]}],surroundingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:'"',close:'"'},{open:"'",close:"'"},{open:"<",close:">"}]},iUe={defaultToken:"",tokenPostfix:".flow",keywords:["import","require","export","forbid","native","if","else","cast","unsafe","switch","default"],types:["io","mutable","bool","int","double","string","flow","void","ref","true","false","with"],operators:["=",">","<","<=",">=","==","!","!=",":=","::=","&&","||","+","-","*","/","@","&","%",":","->","\\","$","??","^"],symbols:/[@$=><!~?:&|+\-*\\\/\^%]+/,escapes:/\\(?:[abfnrtv\\"']|x[0-9A-Fa-f]{1,4}|u[0-9A-Fa-f]{4}|U[0-9A-Fa-f]{8})/,tokenizer:{root:[[/[a-zA-Z_]\w*/,{cases:{"@keywords":"keyword","@types":"type","@default":"identifier"}}],{include:"@whitespace"},[/[{}()\[\]]/,"delimiter"],[/[<>](?!@symbols)/,"delimiter"],[/@symbols/,{cases:{"@operators":"delimiter","@default":""}}],[/((0(x|X)[0-9a-fA-F]*)|(([0-9]+\.?[0-9]*)|(\.[0-9]+))((e|E)(\+|-)?[0-9]+)?)/,"number"],[/[;,.]/,"delimiter"],[/"([^"\\]|\\.)*$/,"string.invalid"],[/"/,"string","@string"]],whitespace:[[/[ \t\r\n]+/,""],[/\/\*/,"comment","@comment"],[/\/\/.*$/,"comment"]],comment:[[/[^\/*]+/,"comment"],[/\*\//,"comment","@pop"],[/[\/*]/,"comment"]],string:[[/[^\\"]+/,"string"],[/@escapes/,"string.escape"],[/\\./,"string.escape.invalid"],[/"/,"string","@pop"]]}},nUe=Object.freeze(Object.defineProperty({__proto__:null,conf:tUe,language:iUe},Symbol.toStringTag,{value:"Module"})),sUe={comments:{lineComment:"//",blockComment:["(*","*)"]},brackets:[["{","}"],["[","]"],["(",")"]],autoClosingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:'"',close:'"'}],surroundingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:'"',close:'"'},{open:"'",close:"'"}],folding:{markers:{start:new RegExp("^\\s*//\\s*#region\\b|^\\s*\\(\\*\\s*#region(.*)\\*\\)"),end:new RegExp("^\\s*//\\s*#endregion\\b|^\\s*\\(\\*\\s*#endregion\\s*\\*\\)")}}},oUe={defaultToken:"",tokenPostfix:".fs",keywords:["abstract","and","atomic","as","assert","asr","base","begin","break","checked","component","const","constraint","constructor","continue","class","default","delegate","do","done","downcast","downto","elif","else","end","exception","eager","event","external","extern","false","finally","for","fun","function","fixed","functor","global","if","in","include","inherit","inline","interface","internal","land","lor","lsl","lsr","lxor","lazy","let","match","member","mod","module","mutable","namespace","method","mixin","new","not","null","of","open","or","object","override","private","parallel","process","protected","pure","public","rec","return","static","sealed","struct","sig","then","to","true","tailcall","trait","try","type","upcast","use","val","void","virtual","volatile","when","while","with","yield"],symbols:/[=><!~?:&|+\-*\^%;\.,\/]+/,escapes:/\\(?:[abfnrtv\\"']|x[0-9A-Fa-f]{1,4}|u[0-9A-Fa-f]{4}|U[0-9A-Fa-f]{8})/,integersuffix:/[uU]?[yslnLI]?/,floatsuffix:/[fFmM]?/,tokenizer:{root:[[/[a-zA-Z_]\w*/,{cases:{"@keywords":{token:"keyword.$0"},"@default":"identifier"}}],{include:"@whitespace"},[/\[<.*>\]/,"annotation"],[/^#(if|else|endif)/,"keyword"],[/[{}()\[\]]/,"@brackets"],[/[<>](?!@symbols)/,"@brackets"],[/@symbols/,"delimiter"],[/\d*\d+[eE]([\-+]?\d+)?(@floatsuffix)/,"number.float"],[/\d*\.\d+([eE][\-+]?\d+)?(@floatsuffix)/,"number.float"],[/0x[0-9a-fA-F]+LF/,"number.float"],[/0x[0-9a-fA-F]+(@integersuffix)/,"number.hex"],[/0b[0-1]+(@integersuffix)/,"number.bin"],[/\d+(@integersuffix)/,"number"],[/[;,.]/,"delimiter"],[/"([^"\\]|\\.)*$/,"string.invalid"],[/"""/,"string",'@string."""'],[/"/,"string",'@string."'],[/\@"/,{token:"string.quote",next:"@litstring"}],[/'[^\\']'B?/,"string"],[/(')(@escapes)(')/,["string","string.escape","string"]],[/'/,"string.invalid"]],whitespace:[[/[ \t\r\n]+/,""],[/\(\*(?!\))/,"comment","@comment"],[/\/\/.*$/,"comment"]],comment:[[/[^*(]+/,"comment"],[/\*\)/,"comment","@pop"],[/\*/,"comment"],[/\(\*\)/,"comment"],[/\(/,"comment"]],string:[[/[^\\"]+/,"string"],[/@escapes/,"string.escape"],[/\\./,"string.escape.invalid"],[/("""|"B?)/,{cases:{"$#==$S2":{token:"string",next:"@pop"},"@default":"string"}}]],litstring:[[/[^"]+/,"string"],[/""/,"string.escape"],[/"/,{token:"string.quote",next:"@pop"}]]}},rUe=Object.freeze(Object.defineProperty({__proto__:null,conf:sUe,language:oUe},Symbol.toStringTag,{value:"Module"})),PT=["assign","flush","ftl","return","global","import","include","break","continue","local","nested","nt","setting","stop","t","lt","rt","fallback"],FT=["attempt","autoesc","autoEsc","compress","comment","escape","noescape","function","if","list","items","sep","macro","noparse","noParse","noautoesc","noAutoEsc","outputformat","switch","visit","recurse"],O0={close:">",id:"angle",open:"<"},P0={close:"\\]",id:"bracket",open:"\\["},aUe={close:"[>\\]]",id:"auto",open:"[<\\[]"},OV={close:"\\}",id:"dollar",open1:"\\$",open2:"\\{"},PV={close:"\\]",id:"bracket",open1:"\\[",open2:"="};function p2(s){return{brackets:[["<",">"],["[","]"],["(",")"],["{","}"]],comments:{blockComment:[`${s.open}--`,`--${s.close}`]},autoCloseBefore:`
\r	 }]),.:;=`,autoClosingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:'"',close:'"',notIn:["string"]},{open:"'",close:"'",notIn:["string"]}],surroundingPairs:[{open:'"',close:'"'},{open:"'",close:"'"},{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:"<",close:">"}],folding:{markers:{start:new RegExp(`${s.open}#(?:${FT.join("|")})([^/${s.close}]*(?!/)${s.close})[^${s.open}]*$`),end:new RegExp(`${s.open}/#(?:${FT.join("|")})[\\r\\n\\t ]*>`)}},onEnterRules:[{beforeText:new RegExp(`${s.open}#(?!(?:${PT.join("|")}))([a-zA-Z_]+)([^/${s.close}]*(?!/)${s.close})[^${s.open}]*$`),afterText:new RegExp(`^${s.open}/#([a-zA-Z_]+)[\\r\\n\\t ]*${s.close}$`),action:{indentAction:H.IndentAction.IndentOutdent}},{beforeText:new RegExp(`${s.open}#(?!(?:${PT.join("|")}))([a-zA-Z_]+)([^/${s.close}]*(?!/)${s.close})[^${s.open}]*$`),action:{indentAction:H.IndentAction.Indent}}]}}function mle(){return{brackets:[["<",">"],["[","]"],["(",")"],["{","}"]],autoCloseBefore:`
\r	 }]),.:;=`,autoClosingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:'"',close:'"',notIn:["string"]},{open:"'",close:"'",notIn:["string"]}],surroundingPairs:[{open:'"',close:'"'},{open:"'",close:"'"},{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:"<",close:">"}],folding:{markers:{start:new RegExp(`[<\\[]#(?:${FT.join("|")})([^/>\\]]*(?!/)[>\\]])[^<\\[]*$`),end:new RegExp(`[<\\[]/#(?:${FT.join("|")})[\\r\\n\\t ]*>`)}},onEnterRules:[{beforeText:new RegExp(`[<\\[]#(?!(?:${PT.join("|")}))([a-zA-Z_]+)([^/>\\]]*(?!/)[>\\]])[^[<\\[]]*$`),afterText:new RegExp("^[<\\[]/#([a-zA-Z_]+)[\\r\\n\\t ]*[>\\]]$"),action:{indentAction:H.IndentAction.IndentOutdent}},{beforeText:new RegExp(`[<\\[]#(?!(?:${PT.join("|")}))([a-zA-Z_]+)([^/>\\]]*(?!/)[>\\]])[^[<\\[]]*$`),action:{indentAction:H.IndentAction.Indent}}]}}function vp(s,e){const t=`_${s.id}_${e.id}`,i=o=>o.replace(/__id__/g,t),n=o=>{const r=o.source.replace(/__id__/g,t);return new RegExp(r,o.flags)};return{unicode:!0,includeLF:!1,start:i("default__id__"),ignoreCase:!1,defaultToken:"invalid",tokenPostfix:".freemarker2",brackets:[{open:"{",close:"}",token:"delimiter.curly"},{open:"[",close:"]",token:"delimiter.square"},{open:"(",close:")",token:"delimiter.parenthesis"},{open:"<",close:">",token:"delimiter.angle"}],[i("open__id__")]:new RegExp(s.open),[i("close__id__")]:new RegExp(s.close),[i("iOpen1__id__")]:new RegExp(e.open1),[i("iOpen2__id__")]:new RegExp(e.open2),[i("iClose__id__")]:new RegExp(e.close),[i("startTag__id__")]:n(/(@open__id__)(#)/),[i("endTag__id__")]:n(/(@open__id__)(\/#)/),[i("startOrEndTag__id__")]:n(/(@open__id__)(\/?#)/),[i("closeTag1__id__")]:n(/((?:@blank)*)(@close__id__)/),[i("closeTag2__id__")]:n(/((?:@blank)*\/?)(@close__id__)/),blank:/[ \t\n\r]/,keywords:["false","true","in","as","using"],directiveStartCloseTag1:/attempt|recover|sep|auto[eE]sc|no(?:autoe|AutoE)sc|compress|default|no[eE]scape|comment|no[pP]arse/,directiveStartCloseTag2:/else|break|continue|return|stop|flush|t|lt|rt|nt|nested|recurse|fallback|ftl/,directiveStartBlank:/if|else[iI]f|list|for[eE]ach|switch|case|assign|global|local|include|import|function|macro|transform|visit|stop|return|call|setting|output[fF]ormat|nested|recurse|escape|ftl|items/,directiveEndCloseTag1:/if|list|items|sep|recover|attempt|for[eE]ach|local|global|assign|function|macro|output[fF]ormat|auto[eE]sc|no(?:autoe|AutoE)sc|compress|transform|switch|escape|no[eE]scape/,escapedChar:/\\(?:[ntrfbgla\\'"\{=]|(?:x[0-9A-Fa-f]{1,4}))/,asciiDigit:/[0-9]/,integer:/[0-9]+/,nonEscapedIdStartChar:/[\$@-Z_a-z\u00AA\u00B5\u00BA\u00C0-\u00D6\u00D8-\u00F6\u00F8-\u1FFF\u2071\u207F\u2090-\u209C\u2102\u2107\u210A-\u2113\u2115\u2119-\u211D\u2124\u2126\u2128\u212A-\u212D\u212F-\u2139\u213C-\u213F\u2145-\u2149\u214E\u2183-\u2184\u2C00-\u2C2E\u2C30-\u2C5E\u2C60-\u2CE4\u2CEB-\u2CEE\u2CF2-\u2CF3\u2D00-\u2D25\u2D27\u2D2D\u2D30-\u2D67\u2D6F\u2D80-\u2D96\u2DA0-\u2DA6\u2DA8-\u2DAE\u2DB0-\u2DB6\u2DB8-\u2DBE\u2DC0-\u2DC6\u2DC8-\u2DCE\u2DD0-\u2DD6\u2DD8-\u2DDE\u2E2F\u3005-\u3006\u3031-\u3035\u303B-\u303C\u3040-\u318F\u31A0-\u31BA\u31F0-\u31FF\u3300-\u337F\u3400-\u4DB5\u4E00-\uA48C\uA4D0-\uA4FD\uA500-\uA60C\uA610-\uA62B\uA640-\uA66E\uA67F-\uA697\uA6A0-\uA6E5\uA717-\uA71F\uA722-\uA788\uA78B-\uA78E\uA790-\uA793\uA7A0-\uA7AA\uA7F8-\uA801\uA803-\uA805\uA807-\uA80A\uA80C-\uA822\uA840-\uA873\uA882-\uA8B3\uA8D0-\uA8D9\uA8F2-\uA8F7\uA8FB\uA900-\uA925\uA930-\uA946\uA960-\uA97C\uA984-\uA9B2\uA9CF-\uA9D9\uAA00-\uAA28\uAA40-\uAA42\uAA44-\uAA4B\uAA50-\uAA59\uAA60-\uAA76\uAA7A\uAA80-\uAAAF\uAAB1\uAAB5-\uAAB6\uAAB9-\uAABD\uAAC0\uAAC2\uAADB-\uAADD\uAAE0-\uAAEA\uAAF2-\uAAF4\uAB01-\uAB06\uAB09-\uAB0E\uAB11-\uAB16\uAB20-\uAB26\uAB28-\uAB2E\uABC0-\uABE2\uABF0-\uABF9\uAC00-\uD7A3\uD7B0-\uD7C6\uD7CB-\uD7FB\uF900-\uFB06\uFB13-\uFB17\uFB1D\uFB1F-\uFB28\uFB2A-\uFB36\uFB38-\uFB3C\uFB3E\uFB40-\uFB41\uFB43-\uFB44\uFB46-\uFBB1\uFBD3-\uFD3D\uFD50-\uFD8F\uFD92-\uFDC7\uFDF0-\uFDFB\uFE70-\uFE74\uFE76-\uFEFC\uFF10-\uFF19\uFF21-\uFF3A\uFF41-\uFF5A\uFF66-\uFFBE\uFFC2-\uFFC7\uFFCA-\uFFCF\uFFD2-\uFFD7\uFFDA-\uFFDC]/,escapedIdChar:/\\[\-\.:#]/,idStartChar:/(?:@nonEscapedIdStartChar)|(?:@escapedIdChar)/,id:/(?:@idStartChar)(?:(?:@idStartChar)|(?:@asciiDigit))*/,specialHashKeys:/\*\*|\*|false|true|in|as|using/,namedSymbols:/&lt;=|&gt;=|\\lte|\\lt|&lt;|\\gte|\\gt|&gt;|&amp;&amp;|\\and|-&gt;|->|==|!=|\+=|-=|\*=|\/=|%=|\+\+|--|<=|&&|\|\||:|\.\.\.|\.\.\*|\.\.<|\.\.!|\?\?|=|<|\+|-|\*|\/|%|\||\.\.|\?|!|&|\.|,|;/,arrows:["->","-&gt;"],delimiters:[";",":",",","."],stringOperators:["lte","lt","gte","gt"],noParseTags:["noparse","noParse","comment"],tokenizer:{[i("default__id__")]:[{include:i("@directive_token__id__")},{include:i("@interpolation_and_text_token__id__")}],[i("fmExpression__id__.directive")]:[{include:i("@blank_and_expression_comment_token__id__")},{include:i("@directive_end_token__id__")},{include:i("@expression_token__id__")}],[i("fmExpression__id__.interpolation")]:[{include:i("@blank_and_expression_comment_token__id__")},{include:i("@expression_token__id__")},{include:i("@greater_operators_token__id__")}],[i("inParen__id__.plain")]:[{include:i("@blank_and_expression_comment_token__id__")},{include:i("@directive_end_token__id__")},{include:i("@expression_token__id__")}],[i("inParen__id__.gt")]:[{include:i("@blank_and_expression_comment_token__id__")},{include:i("@expression_token__id__")},{include:i("@greater_operators_token__id__")}],[i("noSpaceExpression__id__")]:[{include:i("@no_space_expression_end_token__id__")},{include:i("@directive_end_token__id__")},{include:i("@expression_token__id__")}],[i("unifiedCall__id__")]:[{include:i("@unified_call_token__id__")}],[i("singleString__id__")]:[{include:i("@string_single_token__id__")}],[i("doubleString__id__")]:[{include:i("@string_double_token__id__")}],[i("rawSingleString__id__")]:[{include:i("@string_single_raw_token__id__")}],[i("rawDoubleString__id__")]:[{include:i("@string_double_raw_token__id__")}],[i("expressionComment__id__")]:[{include:i("@expression_comment_token__id__")}],[i("noParse__id__")]:[{include:i("@no_parse_token__id__")}],[i("terseComment__id__")]:[{include:i("@terse_comment_token__id__")}],[i("directive_token__id__")]:[[n(/(?:@startTag__id__)(@directiveStartCloseTag1)(?:@closeTag1__id__)/),s.id==="auto"?{cases:{"$1==<":{token:"@rematch",switchTo:`@default_angle_${e.id}`},"$1==[":{token:"@rematch",switchTo:`@default_bracket_${e.id}`}}}:[{token:"@brackets.directive"},{token:"delimiter.directive"},{cases:{"@noParseTags":{token:"tag",next:i("@noParse__id__.$3")},"@default":{token:"tag"}}},{token:"delimiter.directive"},{token:"@brackets.directive"}]],[n(/(?:@startTag__id__)(@directiveStartCloseTag2)(?:@closeTag2__id__)/),s.id==="auto"?{cases:{"$1==<":{token:"@rematch",switchTo:`@default_angle_${e.id}`},"$1==[":{token:"@rematch",switchTo:`@default_bracket_${e.id}`}}}:[{token:"@brackets.directive"},{token:"delimiter.directive"},{token:"tag"},{token:"delimiter.directive"},{token:"@brackets.directive"}]],[n(/(?:@startTag__id__)(@directiveStartBlank)(@blank)/),s.id==="auto"?{cases:{"$1==<":{token:"@rematch",switchTo:`@default_angle_${e.id}`},"$1==[":{token:"@rematch",switchTo:`@default_bracket_${e.id}`}}}:[{token:"@brackets.directive"},{token:"delimiter.directive"},{token:"tag"},{token:"",next:i("@fmExpression__id__.directive")}]],[n(/(?:@endTag__id__)(@directiveEndCloseTag1)(?:@closeTag1__id__)/),s.id==="auto"?{cases:{"$1==<":{token:"@rematch",switchTo:`@default_angle_${e.id}`},"$1==[":{token:"@rematch",switchTo:`@default_bracket_${e.id}`}}}:[{token:"@brackets.directive"},{token:"delimiter.directive"},{token:"tag"},{token:"delimiter.directive"},{token:"@brackets.directive"}]],[n(/(@open__id__)(@)/),s.id==="auto"?{cases:{"$1==<":{token:"@rematch",switchTo:`@default_angle_${e.id}`},"$1==[":{token:"@rematch",switchTo:`@default_bracket_${e.id}`}}}:[{token:"@brackets.directive"},{token:"delimiter.directive",next:i("@unifiedCall__id__")}]],[n(/(@open__id__)(\/@)((?:(?:@id)(?:\.(?:@id))*)?)(?:@closeTag1__id__)/),[{token:"@brackets.directive"},{token:"delimiter.directive"},{token:"tag"},{token:"delimiter.directive"},{token:"@brackets.directive"}]],[n(/(@open__id__)#--/),s.id==="auto"?{cases:{"$1==<":{token:"@rematch",switchTo:`@default_angle_${e.id}`},"$1==[":{token:"@rematch",switchTo:`@default_bracket_${e.id}`}}}:{token:"comment",next:i("@terseComment__id__")}],[n(/(?:@startOrEndTag__id__)([a-zA-Z_]+)/),s.id==="auto"?{cases:{"$1==<":{token:"@rematch",switchTo:`@default_angle_${e.id}`},"$1==[":{token:"@rematch",switchTo:`@default_bracket_${e.id}`}}}:[{token:"@brackets.directive"},{token:"delimiter.directive"},{token:"tag.invalid",next:i("@fmExpression__id__.directive")}]]],[i("interpolation_and_text_token__id__")]:[[n(/(@iOpen1__id__)(@iOpen2__id__)/),[{token:e.id==="bracket"?"@brackets.interpolation":"delimiter.interpolation"},{token:e.id==="bracket"?"delimiter.interpolation":"@brackets.interpolation",next:i("@fmExpression__id__.interpolation")}]],[/[\$#<\[\{]|(?:@blank)+|[^\$<#\[\{\n\r\t ]+/,{token:"source"}]],[i("string_single_token__id__")]:[[/[^'\\]/,{token:"string"}],[/@escapedChar/,{token:"string.escape"}],[/'/,{token:"string",next:"@pop"}]],[i("string_double_token__id__")]:[[/[^"\\]/,{token:"string"}],[/@escapedChar/,{token:"string.escape"}],[/"/,{token:"string",next:"@pop"}]],[i("string_single_raw_token__id__")]:[[/[^']+/,{token:"string.raw"}],[/'/,{token:"string.raw",next:"@pop"}]],[i("string_double_raw_token__id__")]:[[/[^"]+/,{token:"string.raw"}],[/"/,{token:"string.raw",next:"@pop"}]],[i("expression_token__id__")]:[[/(r?)(['"])/,{cases:{"r'":[{token:"keyword"},{token:"string.raw",next:i("@rawSingleString__id__")}],'r"':[{token:"keyword"},{token:"string.raw",next:i("@rawDoubleString__id__")}],"'":[{token:"source"},{token:"string",next:i("@singleString__id__")}],'"':[{token:"source"},{token:"string",next:i("@doubleString__id__")}]}}],[/(?:@integer)(?:\.(?:@integer))?/,{cases:{"(?:@integer)":{token:"number"},"@default":{token:"number.float"}}}],[/(\.)(@blank*)(@specialHashKeys)/,[{token:"delimiter"},{token:""},{token:"identifier"}]],[/(?:@namedSymbols)/,{cases:{"@arrows":{token:"meta.arrow"},"@delimiters":{token:"delimiter"},"@default":{token:"operators"}}}],[/@id/,{cases:{"@keywords":{token:"keyword.$0"},"@stringOperators":{token:"operators"},"@default":{token:"identifier"}}}],[/[\[\]\(\)\{\}]/,{cases:{"\\[":{cases:{"$S2==gt":{token:"@brackets",next:i("@inParen__id__.gt")},"@default":{token:"@brackets",next:i("@inParen__id__.plain")}}},"\\]":{cases:{...e.id==="bracket"?{"$S2==interpolation":{token:"@brackets.interpolation",next:"@popall"}}:{},...s.id==="bracket"?{"$S2==directive":{token:"@brackets.directive",next:"@popall"}}:{},[i("$S1==inParen__id__")]:{token:"@brackets",next:"@pop"},"@default":{token:"@brackets"}}},"\\(":{token:"@brackets",next:i("@inParen__id__.gt")},"\\)":{cases:{[i("$S1==inParen__id__")]:{token:"@brackets",next:"@pop"},"@default":{token:"@brackets"}}},"\\{":{cases:{"$S2==gt":{token:"@brackets",next:i("@inParen__id__.gt")},"@default":{token:"@brackets",next:i("@inParen__id__.plain")}}},"\\}":{cases:{...e.id==="bracket"?{}:{"$S2==interpolation":{token:"@brackets.interpolation",next:"@popall"}},[i("$S1==inParen__id__")]:{token:"@brackets",next:"@pop"},"@default":{token:"@brackets"}}}}}],[/\$\{/,{token:"delimiter.invalid"}]],[i("blank_and_expression_comment_token__id__")]:[[/(?:@blank)+/,{token:""}],[/[<\[][#!]--/,{token:"comment",next:i("@expressionComment__id__")}]],[i("directive_end_token__id__")]:[[/>/,s.id==="bracket"?{token:"operators"}:{token:"@brackets.directive",next:"@popall"}],[n(/(\/)(@close__id__)/),[{token:"delimiter.directive"},{token:"@brackets.directive",next:"@popall"}]]],[i("greater_operators_token__id__")]:[[/>/,{token:"operators"}],[/>=/,{token:"operators"}]],[i("no_space_expression_end_token__id__")]:[[/(?:@blank)+/,{token:"",switchTo:i("@fmExpression__id__.directive")}]],[i("unified_call_token__id__")]:[[/(@id)((?:@blank)+)/,[{token:"tag"},{token:"",next:i("@fmExpression__id__.directive")}]],[n(/(@id)(\/?)(@close__id__)/),[{token:"tag"},{token:"delimiter.directive"},{token:"@brackets.directive",next:"@popall"}]],[/./,{token:"@rematch",next:i("@noSpaceExpression__id__")}]],[i("no_parse_token__id__")]:[[n(/(@open__id__)(\/#?)([a-zA-Z]+)((?:@blank)*)(@close__id__)/),{cases:{"$S2==$3":[{token:"@brackets.directive"},{token:"delimiter.directive"},{token:"tag"},{token:""},{token:"@brackets.directive",next:"@popall"}],"$S2==comment":[{token:"comment"},{token:"comment"},{token:"comment"},{token:"comment"},{token:"comment"}],"@default":[{token:"source"},{token:"source"},{token:"source"},{token:"source"},{token:"source"}]}}],[/[^<\[\-]+|[<\[\-]/,{cases:{"$S2==comment":{token:"comment"},"@default":{token:"source"}}}]],[i("expression_comment_token__id__")]:[[/--[>\]]/,{token:"comment",next:"@pop"}],[/[^\->\]]+|[>\]\-]/,{token:"comment"}]],[i("terse_comment_token__id__")]:[[n(/--(?:@close__id__)/),{token:"comment",next:"@popall"}],[/[^<\[\-]+|[<\[\-]/,{token:"comment"}]]}}}function _le(s){const e=vp(O0,s),t=vp(P0,s),i=vp(aUe,s);return{...e,...t,...i,unicode:!0,includeLF:!1,start:`default_auto_${s.id}`,ignoreCase:!1,defaultToken:"invalid",tokenPostfix:".freemarker2",brackets:[{open:"{",close:"}",token:"delimiter.curly"},{open:"[",close:"]",token:"delimiter.square"},{open:"(",close:")",token:"delimiter.parenthesis"},{open:"<",close:">",token:"delimiter.angle"}],tokenizer:{...e.tokenizer,...t.tokenizer,...i.tokenizer}}}const lUe={conf:p2(O0),language:vp(O0,OV)},cUe={conf:p2(P0),language:vp(P0,OV)},dUe={conf:p2(O0),language:vp(O0,PV)},hUe={conf:p2(P0),language:vp(P0,PV)},uUe={conf:mle(),language:_le(OV)},gUe={conf:mle(),language:_le(PV)},D_=Object.freeze(Object.defineProperty({__proto__:null,TagAngleInterpolationBracket:dUe,TagAngleInterpolationDollar:lUe,TagAutoInterpolationBracket:gUe,TagAutoInterpolationDollar:uUe,TagBracketInterpolationBracket:hUe,TagBracketInterpolationDollar:cUe},Symbol.toStringTag,{value:"Module"})),fUe={comments:{lineComment:"//",blockComment:["/*","*/"]},brackets:[["{","}"],["[","]"],["(",")"]],autoClosingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:"`",close:"`",notIn:["string"]},{open:'"',close:'"',notIn:["string"]},{open:"'",close:"'",notIn:["string","comment"]}],surroundingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:"`",close:"`"},{open:'"',close:'"'},{open:"'",close:"'"}]},pUe={defaultToken:"",tokenPostfix:".go",keywords:["break","case","chan","const","continue","default","defer","else","fallthrough","for","func","go","goto","if","import","interface","map","package","range","return","select","struct","switch","type","var","bool","true","false","uint8","uint16","uint32","uint64","int8","int16","int32","int64","float32","float64","complex64","complex128","byte","rune","uint","int","uintptr","string","nil"],operators:["+","-","*","/","%","&","|","^","<<",">>","&^","+=","-=","*=","/=","%=","&=","|=","^=","<<=",">>=","&^=","&&","||","<-","++","--","==","<",">","=","!","!=","<=",">=",":=","...","(",")","","]","{","}",",",";",".",":"],symbols:/[=><!~?:&|+\-*\/\^%]+/,escapes:/\\(?:[abfnrtv\\"']|x[0-9A-Fa-f]{1,4}|u[0-9A-Fa-f]{4}|U[0-9A-Fa-f]{8})/,tokenizer:{root:[[/[a-zA-Z_]\w*/,{cases:{"@keywords":{token:"keyword.$0"},"@default":"identifier"}}],{include:"@whitespace"},[/\[\[.*\]\]/,"annotation"],[/^\s*#\w+/,"keyword"],[/[{}()\[\]]/,"@brackets"],[/[<>](?!@symbols)/,"@brackets"],[/@symbols/,{cases:{"@operators":"delimiter","@default":""}}],[/\d*\d+[eE]([\-+]?\d+)?/,"number.float"],[/\d*\.\d+([eE][\-+]?\d+)?/,"number.float"],[/0[xX][0-9a-fA-F']*[0-9a-fA-F]/,"number.hex"],[/0[0-7']*[0-7]/,"number.octal"],[/0[bB][0-1']*[0-1]/,"number.binary"],[/\d[\d']*/,"number"],[/\d/,"number"],[/[;,.]/,"delimiter"],[/"([^"\\]|\\.)*$/,"string.invalid"],[/"/,"string","@string"],[/`/,"string","@rawstring"],[/'[^\\']'/,"string"],[/(')(@escapes)(')/,["string","string.escape","string"]],[/'/,"string.invalid"]],whitespace:[[/[ \t\r\n]+/,""],[/\/\*\*(?!\/)/,"comment.doc","@doccomment"],[/\/\*/,"comment","@comment"],[/\/\/.*$/,"comment"]],comment:[[/[^\/*]+/,"comment"],[/\*\//,"comment","@pop"],[/[\/*]/,"comment"]],doccomment:[[/[^\/*]+/,"comment.doc"],[/\/\*/,"comment.doc.invalid"],[/\*\//,"comment.doc","@pop"],[/[\/*]/,"comment.doc"]],string:[[/[^\\"]+/,"string"],[/@escapes/,"string.escape"],[/\\./,"string.escape.invalid"],[/"/,"string","@pop"]],rawstring:[[/[^\`]/,"string"],[/`/,"string","@pop"]]}},mUe=Object.freeze(Object.defineProperty({__proto__:null,conf:fUe,language:pUe},Symbol.toStringTag,{value:"Module"})),_Ue={comments:{lineComment:"#"},brackets:[["{","}"],["[","]"],["(",")"]],autoClosingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:'"""',close:'"""',notIn:["string","comment"]},{open:'"',close:'"',notIn:["string","comment"]}],surroundingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:'"""',close:'"""'},{open:'"',close:'"'}],folding:{offSide:!0}},bUe={defaultToken:"invalid",tokenPostfix:".gql",keywords:["null","true","false","query","mutation","subscription","extend","schema","directive","scalar","type","interface","union","enum","input","implements","fragment","on"],typeKeywords:["Int","Float","String","Boolean","ID"],directiveLocations:["SCHEMA","SCALAR","OBJECT","FIELD_DEFINITION","ARGUMENT_DEFINITION","INTERFACE","UNION","ENUM","ENUM_VALUE","INPUT_OBJECT","INPUT_FIELD_DEFINITION","QUERY","MUTATION","SUBSCRIPTION","FIELD","FRAGMENT_DEFINITION","FRAGMENT_SPREAD","INLINE_FRAGMENT","VARIABLE_DEFINITION"],operators:["=","!","?",":","&","|"],symbols:/[=!?:&|]+/,escapes:/\\(?:["\\\/bfnrt]|u[0-9A-Fa-f]{4})/,tokenizer:{root:[[/[a-z_][\w$]*/,{cases:{"@keywords":"keyword","@default":"key.identifier"}}],[/[$][\w$]*/,{cases:{"@keywords":"keyword","@default":"argument.identifier"}}],[/[A-Z][\w\$]*/,{cases:{"@typeKeywords":"keyword","@default":"type.identifier"}}],{include:"@whitespace"},[/[{}()\[\]]/,"@brackets"],[/@symbols/,{cases:{"@operators":"operator","@default":""}}],[/@\s*[a-zA-Z_\$][\w\$]*/,{token:"annotation",log:"annotation token: $0"}],[/\d*\.\d+([eE][\-+]?\d+)?/,"number.float"],[/0[xX][0-9a-fA-F]+/,"number.hex"],[/\d+/,"number"],[/[;,.]/,"delimiter"],[/"""/,{token:"string",next:"@mlstring",nextEmbedded:"markdown"}],[/"([^"\\]|\\.)*$/,"string.invalid"],[/"/,{token:"string.quote",bracket:"@open",next:"@string"}]],mlstring:[[/[^"]+/,"string"],['"""',{token:"string",next:"@pop",nextEmbedded:"@pop"}]],string:[[/[^\\"]+/,"string"],[/@escapes/,"string.escape"],[/\\./,"string.escape.invalid"],[/"/,{token:"string.quote",bracket:"@close",next:"@pop"}]],whitespace:[[/[ \t\r\n]+/,""],[/#.*$/,"comment"]]}},wUe=Object.freeze(Object.defineProperty({__proto__:null,conf:_Ue,language:bUe},Symbol.toStringTag,{value:"Module"})),wX=["area","base","br","col","embed","hr","img","input","keygen","link","menuitem","meta","param","source","track","wbr"],vUe={wordPattern:/(-?\d*\.\d\w*)|([^\`\~\!\@\$\^\&\*\(\)\=\+\[\{\]\}\\\|\;\:\'\"\,\.\<\>\/\s]+)/g,comments:{blockComment:["{{!--","--}}"]},brackets:[["<!--","-->"],["<",">"],["{{","}}"],["{","}"],["(",")"]],autoClosingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:'"',close:'"'},{open:"'",close:"'"}],surroundingPairs:[{open:"<",close:">"},{open:'"',close:'"'},{open:"'",close:"'"}],onEnterRules:[{beforeText:new RegExp(`<(?!(?:${wX.join("|")}))(\\w[\\w\\d]*)([^/>]*(?!/)>)[^<]*$`,"i"),afterText:/^<\/(\w[\w\d]*)\s*>$/i,action:{indentAction:H.IndentAction.IndentOutdent}},{beforeText:new RegExp(`<(?!(?:${wX.join("|")}))(\\w[\\w\\d]*)([^/>]*(?!/)>)[^<]*$`,"i"),action:{indentAction:H.IndentAction.Indent}}]},CUe={defaultToken:"",tokenPostfix:"",tokenizer:{root:[[/\{\{!--/,"comment.block.start.handlebars","@commentBlock"],[/\{\{!/,"comment.start.handlebars","@comment"],[/\{\{/,{token:"@rematch",switchTo:"@handlebarsInSimpleState.root"}],[/<!DOCTYPE/,"metatag.html","@doctype"],[/<!--/,"comment.html","@commentHtml"],[/(<)(\w+)(\/>)/,["delimiter.html","tag.html","delimiter.html"]],[/(<)(script)/,["delimiter.html",{token:"tag.html",next:"@script"}]],[/(<)(style)/,["delimiter.html",{token:"tag.html",next:"@style"}]],[/(<)([:\w]+)/,["delimiter.html",{token:"tag.html",next:"@otherTag"}]],[/(<\/)(\w+)/,["delimiter.html",{token:"tag.html",next:"@otherTag"}]],[/</,"delimiter.html"],[/\{/,"delimiter.html"],[/[^<{]+/]],doctype:[[/\{\{/,{token:"@rematch",switchTo:"@handlebarsInSimpleState.comment"}],[/[^>]+/,"metatag.content.html"],[/>/,"metatag.html","@pop"]],comment:[[/\}\}/,"comment.end.handlebars","@pop"],[/./,"comment.content.handlebars"]],commentBlock:[[/--\}\}/,"comment.block.end.handlebars","@pop"],[/./,"comment.content.handlebars"]],commentHtml:[[/\{\{/,{token:"@rematch",switchTo:"@handlebarsInSimpleState.comment"}],[/-->/,"comment.html","@pop"],[/[^-]+/,"comment.content.html"],[/./,"comment.content.html"]],otherTag:[[/\{\{/,{token:"@rematch",switchTo:"@handlebarsInSimpleState.otherTag"}],[/\/?>/,"delimiter.html","@pop"],[/"([^"]*)"/,"attribute.value"],[/'([^']*)'/,"attribute.value"],[/[\w\-]+/,"attribute.name"],[/=/,"delimiter"],[/[ \t\r\n]+/]],script:[[/\{\{/,{token:"@rematch",switchTo:"@handlebarsInSimpleState.script"}],[/type/,"attribute.name","@scriptAfterType"],[/"([^"]*)"/,"attribute.value"],[/'([^']*)'/,"attribute.value"],[/[\w\-]+/,"attribute.name"],[/=/,"delimiter"],[/>/,{token:"delimiter.html",next:"@scriptEmbedded.text/javascript",nextEmbedded:"text/javascript"}],[/[ \t\r\n]+/],[/(<\/)(script\s*)(>)/,["delimiter.html","tag.html",{token:"delimiter.html",next:"@pop"}]]],scriptAfterType:[[/\{\{/,{token:"@rematch",switchTo:"@handlebarsInSimpleState.scriptAfterType"}],[/=/,"delimiter","@scriptAfterTypeEquals"],[/>/,{token:"delimiter.html",next:"@scriptEmbedded.text/javascript",nextEmbedded:"text/javascript"}],[/[ \t\r\n]+/],[/<\/script\s*>/,{token:"@rematch",next:"@pop"}]],scriptAfterTypeEquals:[[/\{\{/,{token:"@rematch",switchTo:"@handlebarsInSimpleState.scriptAfterTypeEquals"}],[/"([^"]*)"/,{token:"attribute.value",switchTo:"@scriptWithCustomType.$1"}],[/'([^']*)'/,{token:"attribute.value",switchTo:"@scriptWithCustomType.$1"}],[/>/,{token:"delimiter.html",next:"@scriptEmbedded.text/javascript",nextEmbedded:"text/javascript"}],[/[ \t\r\n]+/],[/<\/script\s*>/,{token:"@rematch",next:"@pop"}]],scriptWithCustomType:[[/\{\{/,{token:"@rematch",switchTo:"@handlebarsInSimpleState.scriptWithCustomType.$S2"}],[/>/,{token:"delimiter.html",next:"@scriptEmbedded.$S2",nextEmbedded:"$S2"}],[/"([^"]*)"/,"attribute.value"],[/'([^']*)'/,"attribute.value"],[/[\w\-]+/,"attribute.name"],[/=/,"delimiter"],[/[ \t\r\n]+/],[/<\/script\s*>/,{token:"@rematch",next:"@pop"}]],scriptEmbedded:[[/\{\{/,{token:"@rematch",switchTo:"@handlebarsInEmbeddedState.scriptEmbedded.$S2",nextEmbedded:"@pop"}],[/<\/script/,{token:"@rematch",next:"@pop",nextEmbedded:"@pop"}]],style:[[/\{\{/,{token:"@rematch",switchTo:"@handlebarsInSimpleState.style"}],[/type/,"attribute.name","@styleAfterType"],[/"([^"]*)"/,"attribute.value"],[/'([^']*)'/,"attribute.value"],[/[\w\-]+/,"attribute.name"],[/=/,"delimiter"],[/>/,{token:"delimiter.html",next:"@styleEmbedded.text/css",nextEmbedded:"text/css"}],[/[ \t\r\n]+/],[/(<\/)(style\s*)(>)/,["delimiter.html","tag.html",{token:"delimiter.html",next:"@pop"}]]],styleAfterType:[[/\{\{/,{token:"@rematch",switchTo:"@handlebarsInSimpleState.styleAfterType"}],[/=/,"delimiter","@styleAfterTypeEquals"],[/>/,{token:"delimiter.html",next:"@styleEmbedded.text/css",nextEmbedded:"text/css"}],[/[ \t\r\n]+/],[/<\/style\s*>/,{token:"@rematch",next:"@pop"}]],styleAfterTypeEquals:[[/\{\{/,{token:"@rematch",switchTo:"@handlebarsInSimpleState.styleAfterTypeEquals"}],[/"([^"]*)"/,{token:"attribute.value",switchTo:"@styleWithCustomType.$1"}],[/'([^']*)'/,{token:"attribute.value",switchTo:"@styleWithCustomType.$1"}],[/>/,{token:"delimiter.html",next:"@styleEmbedded.text/css",nextEmbedded:"text/css"}],[/[ \t\r\n]+/],[/<\/style\s*>/,{token:"@rematch",next:"@pop"}]],styleWithCustomType:[[/\{\{/,{token:"@rematch",switchTo:"@handlebarsInSimpleState.styleWithCustomType.$S2"}],[/>/,{token:"delimiter.html",next:"@styleEmbedded.$S2",nextEmbedded:"$S2"}],[/"([^"]*)"/,"attribute.value"],[/'([^']*)'/,"attribute.value"],[/[\w\-]+/,"attribute.name"],[/=/,"delimiter"],[/[ \t\r\n]+/],[/<\/style\s*>/,{token:"@rematch",next:"@pop"}]],styleEmbedded:[[/\{\{/,{token:"@rematch",switchTo:"@handlebarsInEmbeddedState.styleEmbedded.$S2",nextEmbedded:"@pop"}],[/<\/style/,{token:"@rematch",next:"@pop",nextEmbedded:"@pop"}]],handlebarsInSimpleState:[[/\{\{\{?/,"delimiter.handlebars"],[/\}\}\}?/,{token:"delimiter.handlebars",switchTo:"@$S2.$S3"}],{include:"handlebarsRoot"}],handlebarsInEmbeddedState:[[/\{\{\{?/,"delimiter.handlebars"],[/\}\}\}?/,{token:"delimiter.handlebars",switchTo:"@$S2.$S3",nextEmbedded:"$S3"}],{include:"handlebarsRoot"}],handlebarsRoot:[[/"[^"]*"/,"string.handlebars"],[/[#/][^\s}]+/,"keyword.helper.handlebars"],[/else\b/,"keyword.helper.handlebars"],[/[\s]+/],[/[^}]/,"variable.parameter.handlebars"]]}},yUe=Object.freeze(Object.defineProperty({__proto__:null,conf:vUe,language:CUe},Symbol.toStringTag,{value:"Module"})),SUe={comments:{lineComment:"#",blockComment:["/*","*/"]},brackets:[["{","}"],["[","]"],["(",")"]],autoClosingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:'"',close:'"',notIn:["string"]}],surroundingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:'"',close:'"'}]},xUe={defaultToken:"",tokenPostfix:".hcl",keywords:["var","local","path","for_each","any","string","number","bool","true","false","null","if ","else ","endif ","for ","in","endfor"],operators:["=",">=","<=","==","!=","+","-","*","/","%","&&","||","!","<",">","?","...",":"],symbols:/[=><!~?:&|+\-*\/\^%]+/,escapes:/\\(?:[abfnrtv\\"']|x[0-9A-Fa-f]{1,4}|u[0-9A-Fa-f]{4}|U[0-9A-Fa-f]{8})/,terraformFunctions:/(abs|ceil|floor|log|max|min|pow|signum|chomp|format|formatlist|indent|join|lower|regex|regexall|replace|split|strrev|substr|title|trimspace|upper|chunklist|coalesce|coalescelist|compact|concat|contains|distinct|element|flatten|index|keys|length|list|lookup|map|matchkeys|merge|range|reverse|setintersection|setproduct|setunion|slice|sort|transpose|values|zipmap|base64decode|base64encode|base64gzip|csvdecode|jsondecode|jsonencode|urlencode|yamldecode|yamlencode|abspath|dirname|pathexpand|basename|file|fileexists|fileset|filebase64|templatefile|formatdate|timeadd|timestamp|base64sha256|base64sha512|bcrypt|filebase64sha256|filebase64sha512|filemd5|filemd1|filesha256|filesha512|md5|rsadecrypt|sha1|sha256|sha512|uuid|uuidv5|cidrhost|cidrnetmask|cidrsubnet|tobool|tolist|tomap|tonumber|toset|tostring)/,terraformMainBlocks:/(module|data|terraform|resource|provider|variable|output|locals)/,tokenizer:{root:[[/^@terraformMainBlocks([ \t]*)([\w-]+|"[\w-]+"|)([ \t]*)([\w-]+|"[\w-]+"|)([ \t]*)(\{)/,["type","","string","","string","","@brackets"]],[/(\w+[ \t]+)([ \t]*)([\w-]+|"[\w-]+"|)([ \t]*)([\w-]+|"[\w-]+"|)([ \t]*)(\{)/,["identifier","","string","","string","","@brackets"]],[/(\w+[ \t]+)([ \t]*)([\w-]+|"[\w-]+"|)([ \t]*)([\w-]+|"[\w-]+"|)(=)(\{)/,["identifier","","string","","operator","","@brackets"]],{include:"@terraform"}],terraform:[[/@terraformFunctions(\()/,["type","@brackets"]],[/[a-zA-Z_]\w*-*/,{cases:{"@keywords":{token:"keyword.$0"},"@default":"variable"}}],{include:"@whitespace"},{include:"@heredoc"},[/[{}()\[\]]/,"@brackets"],[/[<>](?!@symbols)/,"@brackets"],[/@symbols/,{cases:{"@operators":"operator","@default":""}}],[/\d*\d+[eE]([\-+]?\d+)?/,"number.float"],[/\d*\.\d+([eE][\-+]?\d+)?/,"number.float"],[/\d[\d']*/,"number"],[/\d/,"number"],[/[;,.]/,"delimiter"],[/"/,"string","@string"],[/'/,"invalid"]],heredoc:[[/<<[-]*\s*["]?([\w\-]+)["]?/,{token:"string.heredoc.delimiter",next:"@heredocBody.$1"}]],heredocBody:[[/([\w\-]+)$/,{cases:{"$1==$S2":[{token:"string.heredoc.delimiter",next:"@popall"}],"@default":"string.heredoc"}}],[/./,"string.heredoc"]],whitespace:[[/[ \t\r\n]+/,""],[/\/\*/,"comment","@comment"],[/\/\/.*$/,"comment"],[/#.*$/,"comment"]],comment:[[/[^\/*]+/,"comment"],[/\*\//,"comment","@pop"],[/[\/*]/,"comment"]],string:[[/\$\{/,{token:"delimiter",next:"@stringExpression"}],[/[^\\"\$]+/,"string"],[/@escapes/,"string.escape"],[/\\./,"string.escape.invalid"],[/"/,"string","@popall"]],stringInsideExpression:[[/[^\\"]+/,"string"],[/@escapes/,"string.escape"],[/\\./,"string.escape.invalid"],[/"/,"string","@pop"]],stringExpression:[[/\}/,{token:"delimiter",next:"@pop"}],[/"/,"string","@stringInsideExpression"],{include:"@terraform"}]}},LUe=Object.freeze(Object.defineProperty({__proto__:null,conf:SUe,language:xUe},Symbol.toStringTag,{value:"Module"})),vX=["area","base","br","col","embed","hr","img","input","keygen","link","menuitem","meta","param","source","track","wbr"],kUe={wordPattern:/(-?\d*\.\d\w*)|([^\`\~\!\@\$\^\&\*\(\)\=\+\[\{\]\}\\\|\;\:\'\"\,\.\<\>\/\s]+)/g,comments:{blockComment:["<!--","-->"]},brackets:[["<!--","-->"],["<",">"],["{","}"],["(",")"]],autoClosingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:'"',close:'"'},{open:"'",close:"'"}],surroundingPairs:[{open:'"',close:'"'},{open:"'",close:"'"},{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:"<",close:">"}],onEnterRules:[{beforeText:new RegExp(`<(?!(?:${vX.join("|")}))([_:\\w][_:\\w-.\\d]*)([^/>]*(?!/)>)[^<]*$`,"i"),afterText:/^<\/([_:\w][_:\w-.\d]*)\s*>$/i,action:{indentAction:H.IndentAction.IndentOutdent}},{beforeText:new RegExp(`<(?!(?:${vX.join("|")}))(\\w[\\w\\d]*)([^/>]*(?!/)>)[^<]*$`,"i"),action:{indentAction:H.IndentAction.Indent}}],folding:{markers:{start:new RegExp("^\\s*<!--\\s*#region\\b.*-->"),end:new RegExp("^\\s*<!--\\s*#endregion\\b.*-->")}}},EUe={defaultToken:"",tokenPostfix:".html",ignoreCase:!0,tokenizer:{root:[[/<!DOCTYPE/,"metatag","@doctype"],[/<!--/,"comment","@comment"],[/(<)((?:[\w\-]+:)?[\w\-]+)(\s*)(\/>)/,["delimiter","tag","","delimiter"]],[/(<)(script)/,["delimiter",{token:"tag",next:"@script"}]],[/(<)(style)/,["delimiter",{token:"tag",next:"@style"}]],[/(<)((?:[\w\-]+:)?[\w\-]+)/,["delimiter",{token:"tag",next:"@otherTag"}]],[/(<\/)((?:[\w\-]+:)?[\w\-]+)/,["delimiter",{token:"tag",next:"@otherTag"}]],[/</,"delimiter"],[/[^<]+/]],doctype:[[/[^>]+/,"metatag.content"],[/>/,"metatag","@pop"]],comment:[[/-->/,"comment","@pop"],[/[^-]+/,"comment.content"],[/./,"comment.content"]],otherTag:[[/\/?>/,"delimiter","@pop"],[/"([^"]*)"/,"attribute.value"],[/'([^']*)'/,"attribute.value"],[/[\w\-]+/,"attribute.name"],[/=/,"delimiter"],[/[ \t\r\n]+/]],script:[[/type/,"attribute.name","@scriptAfterType"],[/"([^"]*)"/,"attribute.value"],[/'([^']*)'/,"attribute.value"],[/[\w\-]+/,"attribute.name"],[/=/,"delimiter"],[/>/,{token:"delimiter",next:"@scriptEmbedded",nextEmbedded:"text/javascript"}],[/[ \t\r\n]+/],[/(<\/)(script\s*)(>)/,["delimiter","tag",{token:"delimiter",next:"@pop"}]]],scriptAfterType:[[/=/,"delimiter","@scriptAfterTypeEquals"],[/>/,{token:"delimiter",next:"@scriptEmbedded",nextEmbedded:"text/javascript"}],[/[ \t\r\n]+/],[/<\/script\s*>/,{token:"@rematch",next:"@pop"}]],scriptAfterTypeEquals:[[/"module"/,{token:"attribute.value",switchTo:"@scriptWithCustomType.text/javascript"}],[/'module'/,{token:"attribute.value",switchTo:"@scriptWithCustomType.text/javascript"}],[/"([^"]*)"/,{token:"attribute.value",switchTo:"@scriptWithCustomType.$1"}],[/'([^']*)'/,{token:"attribute.value",switchTo:"@scriptWithCustomType.$1"}],[/>/,{token:"delimiter",next:"@scriptEmbedded",nextEmbedded:"text/javascript"}],[/[ \t\r\n]+/],[/<\/script\s*>/,{token:"@rematch",next:"@pop"}]],scriptWithCustomType:[[/>/,{token:"delimiter",next:"@scriptEmbedded.$S2",nextEmbedded:"$S2"}],[/"([^"]*)"/,"attribute.value"],[/'([^']*)'/,"attribute.value"],[/[\w\-]+/,"attribute.name"],[/=/,"delimiter"],[/[ \t\r\n]+/],[/<\/script\s*>/,{token:"@rematch",next:"@pop"}]],scriptEmbedded:[[/<\/script/,{token:"@rematch",next:"@pop",nextEmbedded:"@pop"}],[/[^<]+/,""]],style:[[/type/,"attribute.name","@styleAfterType"],[/"([^"]*)"/,"attribute.value"],[/'([^']*)'/,"attribute.value"],[/[\w\-]+/,"attribute.name"],[/=/,"delimiter"],[/>/,{token:"delimiter",next:"@styleEmbedded",nextEmbedded:"text/css"}],[/[ \t\r\n]+/],[/(<\/)(style\s*)(>)/,["delimiter","tag",{token:"delimiter",next:"@pop"}]]],styleAfterType:[[/=/,"delimiter","@styleAfterTypeEquals"],[/>/,{token:"delimiter",next:"@styleEmbedded",nextEmbedded:"text/css"}],[/[ \t\r\n]+/],[/<\/style\s*>/,{token:"@rematch",next:"@pop"}]],styleAfterTypeEquals:[[/"([^"]*)"/,{token:"attribute.value",switchTo:"@styleWithCustomType.$1"}],[/'([^']*)'/,{token:"attribute.value",switchTo:"@styleWithCustomType.$1"}],[/>/,{token:"delimiter",next:"@styleEmbedded",nextEmbedded:"text/css"}],[/[ \t\r\n]+/],[/<\/style\s*>/,{token:"@rematch",next:"@pop"}]],styleWithCustomType:[[/>/,{token:"delimiter",next:"@styleEmbedded.$S2",nextEmbedded:"$S2"}],[/"([^"]*)"/,"attribute.value"],[/'([^']*)'/,"attribute.value"],[/[\w\-]+/,"attribute.name"],[/=/,"delimiter"],[/[ \t\r\n]+/],[/<\/style\s*>/,{token:"@rematch",next:"@pop"}]],styleEmbedded:[[/<\/style/,{token:"@rematch",next:"@pop",nextEmbedded:"@pop"}],[/[^<]+/,""]]}},IUe=Object.freeze(Object.defineProperty({__proto__:null,conf:kUe,language:EUe},Symbol.toStringTag,{value:"Module"})),DUe={comments:{lineComment:"#"},brackets:[["{","}"],["[","]"],["(",")"]],autoClosingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:'"',close:'"'},{open:"'",close:"'"}],surroundingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:'"',close:'"'},{open:"'",close:"'"}]},TUe={defaultToken:"",tokenPostfix:".ini",escapes:/\\(?:[abfnrtv\\"']|x[0-9A-Fa-f]{1,4}|u[0-9A-Fa-f]{4}|U[0-9A-Fa-f]{8})/,tokenizer:{root:[[/^\[[^\]]*\]/,"metatag"],[/(^\w+)(\s*)(\=)/,["key","","delimiter"]],{include:"@whitespace"},[/\d+/,"number"],[/"([^"\\]|\\.)*$/,"string.invalid"],[/'([^'\\]|\\.)*$/,"string.invalid"],[/"/,"string",'@string."'],[/'/,"string","@string.'"]],whitespace:[[/[ \t\r\n]+/,""],[/^\s*[#;].*$/,"comment"]],string:[[/[^\\"']+/,"string"],[/@escapes/,"string.escape"],[/\\./,"string.escape.invalid"],[/["']/,{cases:{"$#==$S2":{token:"string",next:"@pop"},"@default":"string"}}]]}},NUe=Object.freeze(Object.defineProperty({__proto__:null,conf:DUe,language:TUe},Symbol.toStringTag,{value:"Module"})),RUe={wordPattern:/(-?\d*\.\d\w*)|([^\`\~\!\#\%\^\&\*\(\)\-\=\+\[\{\]\}\\\|\;\:\'\"\,\.\<\>\/\?\s]+)/g,comments:{lineComment:"//",blockComment:["/*","*/"]},brackets:[["{","}"],["[","]"],["(",")"]],autoClosingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:'"',close:'"'},{open:"'",close:"'"}],surroundingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:'"',close:'"'},{open:"'",close:"'"},{open:"<",close:">"}],folding:{markers:{start:new RegExp("^\\s*//\\s*(?:(?:#?region\\b)|(?:<editor-fold\\b))"),end:new RegExp("^\\s*//\\s*(?:(?:#?endregion\\b)|(?:</editor-fold>))")}}},AUe={defaultToken:"",tokenPostfix:".java",keywords:["abstract","continue","for","new","switch","assert","default","goto","package","synchronized","boolean","do","if","private","this","break","double","implements","protected","throw","byte","else","import","public","throws","case","enum","instanceof","return","transient","catch","extends","int","short","try","char","final","interface","static","void","class","finally","long","strictfp","volatile","const","float","native","super","while","true","false","yield","record","sealed","non-sealed","permits"],operators:["=",">","<","!","~","?",":","==","<=",">=","!=","&&","||","++","--","+","-","*","/","&","|","^","%","<<",">>",">>>","+=","-=","*=","/=","&=","|=","^=","%=","<<=",">>=",">>>="],symbols:/[=><!~?:&|+\-*\/\^%]+/,escapes:/\\(?:[abfnrtv\\"']|x[0-9A-Fa-f]{1,4}|u[0-9A-Fa-f]{4}|U[0-9A-Fa-f]{8})/,digits:/\d+(_+\d+)*/,octaldigits:/[0-7]+(_+[0-7]+)*/,binarydigits:/[0-1]+(_+[0-1]+)*/,hexdigits:/[[0-9a-fA-F]+(_+[0-9a-fA-F]+)*/,tokenizer:{root:[["non-sealed","keyword.non-sealed"],[/[a-zA-Z_$][\w$]*/,{cases:{"@keywords":{token:"keyword.$0"},"@default":"identifier"}}],{include:"@whitespace"},[/[{}()\[\]]/,"@brackets"],[/[<>](?!@symbols)/,"@brackets"],[/@symbols/,{cases:{"@operators":"delimiter","@default":""}}],[/@\s*[a-zA-Z_\$][\w\$]*/,"annotation"],[/(@digits)[eE]([\-+]?(@digits))?[fFdD]?/,"number.float"],[/(@digits)\.(@digits)([eE][\-+]?(@digits))?[fFdD]?/,"number.float"],[/0[xX](@hexdigits)[Ll]?/,"number.hex"],[/0(@octaldigits)[Ll]?/,"number.octal"],[/0[bB](@binarydigits)[Ll]?/,"number.binary"],[/(@digits)[fFdD]/,"number.float"],[/(@digits)[lL]?/,"number"],[/[;,.]/,"delimiter"],[/"([^"\\]|\\.)*$/,"string.invalid"],[/"""/,"string","@multistring"],[/"/,"string","@string"],[/'[^\\']'/,"string"],[/(')(@escapes)(')/,["string","string.escape","string"]],[/'/,"string.invalid"]],whitespace:[[/[ \t\r\n]+/,""],[/\/\*\*(?!\/)/,"comment.doc","@javadoc"],[/\/\*/,"comment","@comment"],[/\/\/.*$/,"comment"]],comment:[[/[^\/*]+/,"comment"],[/\*\//,"comment","@pop"],[/[\/*]/,"comment"]],javadoc:[[/[^\/*]+/,"comment.doc"],[/\/\*/,"comment.doc.invalid"],[/\*\//,"comment.doc","@pop"],[/[\/*]/,"comment.doc"]],string:[[/[^\\"]+/,"string"],[/@escapes/,"string.escape"],[/\\./,"string.escape.invalid"],[/"/,"string","@pop"]],multistring:[[/[^\\"]+/,"string"],[/@escapes/,"string.escape"],[/\\./,"string.escape.invalid"],[/"""/,"string","@pop"],[/./,"string"]]}},MUe=Object.freeze(Object.defineProperty({__proto__:null,conf:RUe,language:AUe},Symbol.toStringTag,{value:"Module"})),ble={wordPattern:/(-?\d*\.\d\w*)|([^\`\~\!\@\#\%\^\&\*\(\)\-\=\+\[\{\]\}\\\|\;\:\'\"\,\.\<\>\/\?\s]+)/g,comments:{lineComment:"//",blockComment:["/*","*/"]},brackets:[["{","}"],["[","]"],["(",")"]],onEnterRules:[{beforeText:/^\s*\/\*\*(?!\/)([^\*]|\*(?!\/))*$/,afterText:/^\s*\*\/$/,action:{indentAction:H.IndentAction.IndentOutdent,appendText:" * "}},{beforeText:/^\s*\/\*\*(?!\/)([^\*]|\*(?!\/))*$/,action:{indentAction:H.IndentAction.None,appendText:" * "}},{beforeText:/^(\t|(\ \ ))*\ \*(\ ([^\*]|\*(?!\/))*)?$/,action:{indentAction:H.IndentAction.None,appendText:"* "}},{beforeText:/^(\t|(\ \ ))*\ \*\/\s*$/,action:{indentAction:H.IndentAction.None,removeText:1}}],autoClosingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:'"',close:'"',notIn:["string"]},{open:"'",close:"'",notIn:["string","comment"]},{open:"`",close:"`",notIn:["string","comment"]},{open:"/**",close:" */",notIn:["string"]}],folding:{markers:{start:new RegExp("^\\s*//\\s*#?region\\b"),end:new RegExp("^\\s*//\\s*#?endregion\\b")}}},Za={defaultToken:"invalid",tokenPostfix:".ts",keywords:["abstract","any","as","asserts","bigint","boolean","break","case","catch","class","continue","const","constructor","debugger","declare","default","delete","do","else","enum","export","extends","false","finally","for","from","function","get","if","implements","import","in","infer","instanceof","interface","is","keyof","let","module","namespace","never","new","null","number","object","out","package","private","protected","public","override","readonly","require","global","return","satisfies","set","static","string","super","switch","symbol","this","throw","true","try","type","typeof","undefined","unique","unknown","var","void","while","with","yield","async","await","of"],operators:["<=",">=","==","!=","===","!==","=>","+","-","**","*","/","%","++","--","<<","</",">>",">>>","&","|","^","!","~","&&","||","??","?",":","=","+=","-=","*=","**=","/=","%=","<<=",">>=",">>>=","&=","|=","^=","@"],symbols:/[=><!~?:&|+\-*\/\^%]+/,escapes:/\\(?:[abfnrtv\\"']|x[0-9A-Fa-f]{1,4}|u[0-9A-Fa-f]{4}|U[0-9A-Fa-f]{8})/,digits:/\d+(_+\d+)*/,octaldigits:/[0-7]+(_+[0-7]+)*/,binarydigits:/[0-1]+(_+[0-1]+)*/,hexdigits:/[[0-9a-fA-F]+(_+[0-9a-fA-F]+)*/,regexpctl:/[(){}\[\]\$\^|\-*+?\.]/,regexpesc:/\\(?:[bBdDfnrstvwWn0\\\/]|@regexpctl|c[A-Z]|x[0-9a-fA-F]{2}|u[0-9a-fA-F]{4})/,tokenizer:{root:[[/[{}]/,"delimiter.bracket"],{include:"common"}],common:[[/#?[a-z_$][\w$]*/,{cases:{"@keywords":"keyword","@default":"identifier"}}],[/[A-Z][\w\$]*/,"type.identifier"],{include:"@whitespace"},[/\/(?=([^\\\/]|\\.)+\/([dgimsuy]*)(\s*)(\.|;|,|\)|\]|\}|$))/,{token:"regexp",bracket:"@open",next:"@regexp"}],[/[()\[\]]/,"@brackets"],[/[<>](?!@symbols)/,"@brackets"],[/!(?=([^=]|$))/,"delimiter"],[/@symbols/,{cases:{"@operators":"delimiter","@default":""}}],[/(@digits)[eE]([\-+]?(@digits))?/,"number.float"],[/(@digits)\.(@digits)([eE][\-+]?(@digits))?/,"number.float"],[/0[xX](@hexdigits)n?/,"number.hex"],[/0[oO]?(@octaldigits)n?/,"number.octal"],[/0[bB](@binarydigits)n?/,"number.binary"],[/(@digits)n?/,"number"],[/[;,.]/,"delimiter"],[/"([^"\\]|\\.)*$/,"string.invalid"],[/'([^'\\]|\\.)*$/,"string.invalid"],[/"/,"string","@string_double"],[/'/,"string","@string_single"],[/`/,"string","@string_backtick"]],whitespace:[[/[ \t\r\n]+/,""],[/\/\*\*(?!\/)/,"comment.doc","@jsdoc"],[/\/\*/,"comment","@comment"],[/\/\/.*$/,"comment"]],comment:[[/[^\/*]+/,"comment"],[/\*\//,"comment","@pop"],[/[\/*]/,"comment"]],jsdoc:[[/[^\/*]+/,"comment.doc"],[/\*\//,"comment.doc","@pop"],[/[\/*]/,"comment.doc"]],regexp:[[/(\{)(\d+(?:,\d*)?)(\})/,["regexp.escape.control","regexp.escape.control","regexp.escape.control"]],[/(\[)(\^?)(?=(?:[^\]\\\/]|\\.)+)/,["regexp.escape.control",{token:"regexp.escape.control",next:"@regexrange"}]],[/(\()(\?:|\?=|\?!)/,["regexp.escape.control","regexp.escape.control"]],[/[()]/,"regexp.escape.control"],[/@regexpctl/,"regexp.escape.control"],[/[^\\\/]/,"regexp"],[/@regexpesc/,"regexp.escape"],[/\\\./,"regexp.invalid"],[/(\/)([dgimsuy]*)/,[{token:"regexp",bracket:"@close",next:"@pop"},"keyword.other"]]],regexrange:[[/-/,"regexp.escape.control"],[/\^/,"regexp.invalid"],[/@regexpesc/,"regexp.escape"],[/[^\]]/,"regexp"],[/\]/,{token:"regexp.escape.control",next:"@pop",bracket:"@close"}]],string_double:[[/[^\\"]+/,"string"],[/@escapes/,"string.escape"],[/\\./,"string.escape.invalid"],[/"/,"string","@pop"]],string_single:[[/[^\\']+/,"string"],[/@escapes/,"string.escape"],[/\\./,"string.escape.invalid"],[/'/,"string","@pop"]],string_backtick:[[/\$\{/,{token:"delimiter.bracket",next:"@bracketCounting"}],[/[^\\`$]+/,"string"],[/@escapes/,"string.escape"],[/\\./,"string.escape.invalid"],[/`/,"string","@pop"]],bracketCounting:[[/\{/,"delimiter.bracket","@bracketCounting"],[/\}/,"delimiter.bracket","@pop"],{include:"common"}]}},OUe=Object.freeze(Object.defineProperty({__proto__:null,conf:ble,language:Za},Symbol.toStringTag,{value:"Module"})),PUe=ble,FUe={defaultToken:"invalid",tokenPostfix:".js",keywords:["break","case","catch","class","continue","const","constructor","debugger","default","delete","do","else","export","extends","false","finally","for","from","function","get","if","import","in","instanceof","let","new","null","return","set","static","super","switch","symbol","this","throw","true","try","typeof","undefined","var","void","while","with","yield","async","await","of"],typeKeywords:[],operators:Za.operators,symbols:Za.symbols,escapes:Za.escapes,digits:Za.digits,octaldigits:Za.octaldigits,binarydigits:Za.binarydigits,hexdigits:Za.hexdigits,regexpctl:Za.regexpctl,regexpesc:Za.regexpesc,tokenizer:Za.tokenizer},BUe=Object.freeze(Object.defineProperty({__proto__:null,conf:PUe,language:FUe},Symbol.toStringTag,{value:"Module"})),WUe={brackets:[["{","}"],["[","]"],["(",")"]],autoClosingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:'"',close:'"'},{open:"'",close:"'"}],surroundingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:'"',close:'"'},{open:"'",close:"'"}]},HUe={tokenPostfix:".julia",keywords:["begin","while","if","for","try","return","break","continue","function","macro","quote","let","local","global","const","do","struct","module","baremodule","using","import","export","end","else","elseif","catch","finally","mutable","primitive","abstract","type","in","isa","where","new"],types:["LinRange","LineNumberNode","LinearIndices","LoadError","MIME","Matrix","Method","MethodError","Missing","MissingException","Module","NTuple","NamedTuple","Nothing","Number","OrdinalRange","OutOfMemoryError","OverflowError","Pair","PartialQuickSort","PermutedDimsArray","Pipe","Ptr","QuoteNode","Rational","RawFD","ReadOnlyMemoryError","Real","ReentrantLock","Ref","Regex","RegexMatch","RoundingMode","SegmentationFault","Set","Signed","Some","StackOverflowError","StepRange","StepRangeLen","StridedArray","StridedMatrix","StridedVecOrMat","StridedVector","String","StringIndexError","SubArray","SubString","SubstitutionString","Symbol","SystemError","Task","Text","TextDisplay","Timer","Tuple","Type","TypeError","TypeVar","UInt","UInt128","UInt16","UInt32","UInt64","UInt8","UndefInitializer","AbstractArray","UndefKeywordError","AbstractChannel","UndefRefError","AbstractChar","UndefVarError","AbstractDict","Union","AbstractDisplay","UnionAll","AbstractFloat","UnitRange","AbstractIrrational","Unsigned","AbstractMatrix","AbstractRange","Val","AbstractSet","Vararg","AbstractString","VecElement","AbstractUnitRange","VecOrMat","AbstractVecOrMat","Vector","AbstractVector","VersionNumber","Any","WeakKeyDict","ArgumentError","WeakRef","Array","AssertionError","BigFloat","BigInt","BitArray","BitMatrix","BitSet","BitVector","Bool","BoundsError","CapturedException","CartesianIndex","CartesianIndices","Cchar","Cdouble","Cfloat","Channel","Char","Cint","Cintmax_t","Clong","Clonglong","Cmd","Colon","Complex","ComplexF16","ComplexF32","ComplexF64","CompositeException","Condition","Cptrdiff_t","Cshort","Csize_t","Cssize_t","Cstring","Cuchar","Cuint","Cuintmax_t","Culong","Culonglong","Cushort","Cvoid","Cwchar_t","Cwstring","DataType","DenseArray","DenseMatrix","DenseVecOrMat","DenseVector","Dict","DimensionMismatch","Dims","DivideError","DomainError","EOFError","Enum","ErrorException","Exception","ExponentialBackOff","Expr","Float16","Float32","Float64","Function","GlobalRef","HTML","IO","IOBuffer","IOContext","IOStream","IdDict","IndexCartesian","IndexLinear","IndexStyle","InexactError","InitError","Int","Int128","Int16","Int32","Int64","Int8","Integer","InterruptException","InvalidStateException","Irrational","KeyError"],keywordops:["<:",">:",":","=>","...",".","->","?"],allops:/[^\w\d\s()\[\]{}"'#]+/,constants:["true","false","nothing","missing","undef","Inf","pi","NaN","π","ℯ","ans","PROGRAM_FILE","ARGS","C_NULL","VERSION","DEPOT_PATH","LOAD_PATH"],operators:["!","!=","!==","%","&","*","+","-","/","//","<","<<","<=","==","===","=>",">",">=",">>",">>>","\\","^","|","|>","~","÷","∈","∉","∋","∌","∘","√","∛","∩","∪","≈","≉","≠","≡","≢","≤","≥","⊆","⊇","⊈","⊉","⊊","⊋","⊻"],brackets:[{open:"(",close:")",token:"delimiter.parenthesis"},{open:"{",close:"}",token:"delimiter.curly"},{open:"[",close:"]",token:"delimiter.square"}],ident:/π|ℯ|\b(?!\d)\w+\b/,escape:/(?:[abefnrstv\\"'\n\r]|[0-7]{1,3}|x[0-9A-Fa-f]{1,2}|u[0-9A-Fa-f]{4})/,escapes:/\\(?:C\-(@escape|.)|c(@escape|.)|@escape)/,tokenizer:{root:[[/(::)\s*|\b(isa)\s+/,"keyword","@typeanno"],[/\b(isa)(\s*\(@ident\s*,\s*)/,["keyword",{token:"",next:"@typeanno"}]],[/\b(type|struct)[ \t]+/,"keyword","@typeanno"],[/^\s*:@ident[!?]?/,"metatag"],[/(return)(\s*:@ident[!?]?)/,["keyword","metatag"]],[/(\(|\[|\{|@allops)(\s*:@ident[!?]?)/,["","metatag"]],[/:\(/,"metatag","@quote"],[/r"""/,"regexp.delim","@tregexp"],[/r"/,"regexp.delim","@sregexp"],[/raw"""/,"string.delim","@rtstring"],[/[bv]?"""/,"string.delim","@dtstring"],[/raw"/,"string.delim","@rsstring"],[/[bv]?"/,"string.delim","@dsstring"],[/(@ident)\{/,{cases:{"$1@types":{token:"type",next:"@gen"},"@default":{token:"type",next:"@gen"}}}],[/@ident[!?'']?(?=\.?\()/,{cases:{"@types":"type","@keywords":"keyword","@constants":"variable","@default":"keyword.flow"}}],[/@ident[!?']?/,{cases:{"@types":"type","@keywords":"keyword","@constants":"variable","@default":"identifier"}}],[/\$\w+/,"key"],[/\$\(/,"key","@paste"],[/@@@ident/,"annotation"],{include:"@whitespace"},[/'(?:@escapes|.)'/,"string.character"],[/[()\[\]{}]/,"@brackets"],[/@allops/,{cases:{"@keywordops":"keyword","@operators":"operator"}}],[/[;,]/,"delimiter"],[/0[xX][0-9a-fA-F](_?[0-9a-fA-F])*/,"number.hex"],[/0[_oO][0-7](_?[0-7])*/,"number.octal"],[/0[bB][01](_?[01])*/,"number.binary"],[/[+\-]?\d+(\.\d+)?(im?|[eE][+\-]?\d+(\.\d+)?)?/,"number"]],typeanno:[[/[a-zA-Z_]\w*(?:\.[a-zA-Z_]\w*)*\{/,"type","@gen"],[/([a-zA-Z_]\w*(?:\.[a-zA-Z_]\w*)*)(\s*<:\s*)/,["type","keyword"]],[/[a-zA-Z_]\w*(?:\.[a-zA-Z_]\w*)*/,"type","@pop"],["","","@pop"]],gen:[[/[a-zA-Z_]\w*(?:\.[a-zA-Z_]\w*)*\{/,"type","@push"],[/[a-zA-Z_]\w*(?:\.[a-zA-Z_]\w*)*/,"type"],[/<:/,"keyword"],[/(\})(\s*<:\s*)/,["type",{token:"keyword",next:"@pop"}]],[/\}/,"type","@pop"],{include:"@root"}],quote:[[/\$\(/,"key","@paste"],[/\(/,"@brackets","@paren"],[/\)/,"metatag","@pop"],{include:"@root"}],paste:[[/:\(/,"metatag","@quote"],[/\(/,"@brackets","@paren"],[/\)/,"key","@pop"],{include:"@root"}],paren:[[/\$\(/,"key","@paste"],[/:\(/,"metatag","@quote"],[/\(/,"@brackets","@push"],[/\)/,"@brackets","@pop"],{include:"@root"}],sregexp:[[/^.*/,"invalid"],[/[^\\"()\[\]{}]/,"regexp"],[/[()\[\]{}]/,"@brackets"],[/\\./,"operator.scss"],[/"[imsx]*/,"regexp.delim","@pop"]],tregexp:[[/[^\\"()\[\]{}]/,"regexp"],[/[()\[\]{}]/,"@brackets"],[/\\./,"operator.scss"],[/"(?!"")/,"string"],[/"""[imsx]*/,"regexp.delim","@pop"]],rsstring:[[/^.*/,"invalid"],[/[^\\"]/,"string"],[/\\./,"string.escape"],[/"/,"string.delim","@pop"]],rtstring:[[/[^\\"]/,"string"],[/\\./,"string.escape"],[/"(?!"")/,"string"],[/"""/,"string.delim","@pop"]],dsstring:[[/^.*/,"invalid"],[/[^\\"\$]/,"string"],[/\$/,"","@interpolated"],[/@escapes/,"string.escape"],[/\\./,"string.escape.invalid"],[/"/,"string.delim","@pop"]],dtstring:[[/[^\\"\$]/,"string"],[/\$/,"","@interpolated"],[/@escapes/,"string.escape"],[/\\./,"string.escape.invalid"],[/"(?!"")/,"string"],[/"""/,"string.delim","@pop"]],interpolated:[[/\(/,{token:"",switchTo:"@interpolated_compound"}],[/[a-zA-Z_]\w*/,"identifier"],["","","@pop"]],interpolated_compound:[[/\)/,"","@pop"],{include:"@root"}],whitespace:[[/[ \t\r\n]+/,""],[/#=/,"comment","@multi_comment"],[/#.*$/,"comment"]],multi_comment:[[/#=/,"comment","@push"],[/=#/,"comment","@pop"],[/=(?!#)|#(?!=)/,"comment"],[/[^#=]+/,"comment"]]}},VUe=Object.freeze(Object.defineProperty({__proto__:null,conf:WUe,language:HUe},Symbol.toStringTag,{value:"Module"})),zUe={wordPattern:/(-?\d*\.\d\w*)|([^\`\~\!\#\%\^\&\*\(\)\-\=\+\[\{\]\}\\\|\;\:\'\"\,\.\<\>\/\?\s]+)/g,comments:{lineComment:"//",blockComment:["/*","*/"]},brackets:[["{","}"],["[","]"],["(",")"]],autoClosingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:'"',close:'"'},{open:"'",close:"'"}],surroundingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:'"',close:'"'},{open:"'",close:"'"},{open:"<",close:">"}],folding:{markers:{start:new RegExp("^\\s*//\\s*(?:(?:#?region\\b)|(?:<editor-fold\\b))"),end:new RegExp("^\\s*//\\s*(?:(?:#?endregion\\b)|(?:</editor-fold>))")}}},$Ue={defaultToken:"",tokenPostfix:".kt",keywords:["as","as?","break","class","continue","do","else","false","for","fun","if","in","!in","interface","is","!is","null","object","package","return","super","this","throw","true","try","typealias","val","var","when","while","by","catch","constructor","delegate","dynamic","field","file","finally","get","import","init","param","property","receiver","set","setparam","where","actual","abstract","annotation","companion","const","crossinline","data","enum","expect","external","final","infix","inline","inner","internal","lateinit","noinline","open","operator","out","override","private","protected","public","reified","sealed","suspend","tailrec","vararg","field","it"],operators:["+","-","*","/","%","=","+=","-=","*=","/=","%=","++","--","&&","||","!","==","!=","===","!==",">","<","<=",">=","[","]","!!","?.","?:","::","..",":","?","->","@",";","$","_"],symbols:/[=><!~?:&|+\-*\/\^%]+/,escapes:/\\(?:[abfnrtv\\"']|x[0-9A-Fa-f]{1,4}|u[0-9A-Fa-f]{4}|U[0-9A-Fa-f]{8})/,digits:/\d+(_+\d+)*/,octaldigits:/[0-7]+(_+[0-7]+)*/,binarydigits:/[0-1]+(_+[0-1]+)*/,hexdigits:/[[0-9a-fA-F]+(_+[0-9a-fA-F]+)*/,tokenizer:{root:[[/[A-Z][\w\$]*/,"type.identifier"],[/[a-zA-Z_$][\w$]*/,{cases:{"@keywords":{token:"keyword.$0"},"@default":"identifier"}}],{include:"@whitespace"},[/[{}()\[\]]/,"@brackets"],[/[<>](?!@symbols)/,"@brackets"],[/@symbols/,{cases:{"@operators":"delimiter","@default":""}}],[/@\s*[a-zA-Z_\$][\w\$]*/,"annotation"],[/(@digits)[eE]([\-+]?(@digits))?[fF]?/,"number.float"],[/(@digits)?\.(@digits)([eE][\-+]?(@digits))?[fF]?/,"number.float"],[/0[xX](@hexdigits)[uU]?L?/,"number.hex"],[/0[bB](@binarydigits)[uU]?L?/,"number.binary"],[/(@digits)[fF]/,"number.float"],[/(@digits)[uU]?L?/,"number"],[/[;,.]/,"delimiter"],[/"([^"\\]|\\.)*$/,"string.invalid"],[/"""/,"string","@multistring"],[/"/,"string","@string"],[/'[^\\']'/,"string"],[/(')(@escapes)(')/,["string","string.escape","string"]],[/'/,"string.invalid"]],whitespace:[[/[ \t\r\n]+/,""],[/\/\*\*(?!\/)/,"comment.doc","@javadoc"],[/\/\*/,"comment","@comment"],[/\/\/.*$/,"comment"]],comment:[[/[^\/*]+/,"comment"],[/\/\*/,"comment","@comment"],[/\*\//,"comment","@pop"],[/[\/*]/,"comment"]],javadoc:[[/[^\/*]+/,"comment.doc"],[/\/\*/,"comment.doc","@push"],[/\/\*/,"comment.doc.invalid"],[/\*\//,"comment.doc","@pop"],[/[\/*]/,"comment.doc"]],string:[[/[^\\"]+/,"string"],[/@escapes/,"string.escape"],[/\\./,"string.escape.invalid"],[/"/,"string","@pop"]],multistring:[[/[^\\"]+/,"string"],[/@escapes/,"string.escape"],[/\\./,"string.escape.invalid"],[/"""/,"string","@pop"],[/./,"string"]]}},UUe=Object.freeze(Object.defineProperty({__proto__:null,conf:zUe,language:$Ue},Symbol.toStringTag,{value:"Module"})),qUe={wordPattern:/(#?-?\d*\.\d\w*%?)|([@#!.:]?[\w-?]+%?)|[@#!.]/g,comments:{blockComment:["/*","*/"],lineComment:"//"},brackets:[["{","}"],["[","]"],["(",")"]],autoClosingPairs:[{open:"{",close:"}",notIn:["string","comment"]},{open:"[",close:"]",notIn:["string","comment"]},{open:"(",close:")",notIn:["string","comment"]},{open:'"',close:'"',notIn:["string","comment"]},{open:"'",close:"'",notIn:["string","comment"]}],surroundingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:'"',close:'"'},{open:"'",close:"'"}],folding:{markers:{start:new RegExp("^\\s*\\/\\*\\s*#region\\b\\s*(.*?)\\s*\\*\\/"),end:new RegExp("^\\s*\\/\\*\\s*#endregion\\b.*\\*\\/")}}},jUe={defaultToken:"",tokenPostfix:".less",identifier:"-?-?([a-zA-Z]|(\\\\(([0-9a-fA-F]{1,6}\\s?)|[^[0-9a-fA-F])))([\\w\\-]|(\\\\(([0-9a-fA-F]{1,6}\\s?)|[^[0-9a-fA-F])))*",identifierPlus:"-?-?([a-zA-Z:.]|(\\\\(([0-9a-fA-F]{1,6}\\s?)|[^[0-9a-fA-F])))([\\w\\-:.]|(\\\\(([0-9a-fA-F]{1,6}\\s?)|[^[0-9a-fA-F])))*",brackets:[{open:"{",close:"}",token:"delimiter.curly"},{open:"[",close:"]",token:"delimiter.bracket"},{open:"(",close:")",token:"delimiter.parenthesis"},{open:"<",close:">",token:"delimiter.angle"}],tokenizer:{root:[{include:"@nestedJSBegin"},["[ \\t\\r\\n]+",""],{include:"@comments"},{include:"@keyword"},{include:"@strings"},{include:"@numbers"},["[*_]?[a-zA-Z\\-\\s]+(?=:.*(;|(\\\\$)))","attribute.name","@attribute"],["url(\\-prefix)?\\(",{token:"tag",next:"@urldeclaration"}],["[{}()\\[\\]]","@brackets"],["[,:;]","delimiter"],["#@identifierPlus","tag.id"],["&","tag"],["\\.@identifierPlus(?=\\()","tag.class","@attribute"],["\\.@identifierPlus","tag.class"],["@identifierPlus","tag"],{include:"@operators"},["@(@identifier(?=[:,\\)]))","variable","@attribute"],["@(@identifier)","variable"],["@","key","@atRules"]],nestedJSBegin:[["``","delimiter.backtick"],["`",{token:"delimiter.backtick",next:"@nestedJSEnd",nextEmbedded:"text/javascript"}]],nestedJSEnd:[["`",{token:"delimiter.backtick",next:"@pop",nextEmbedded:"@pop"}]],operators:[["[<>=\\+\\-\\*\\/\\^\\|\\~]","operator"]],keyword:[["(@[\\s]*import|![\\s]*important|true|false|when|iscolor|isnumber|isstring|iskeyword|isurl|ispixel|ispercentage|isem|hue|saturation|lightness|alpha|lighten|darken|saturate|desaturate|fadein|fadeout|fade|spin|mix|round|ceil|floor|percentage)\\b","keyword"]],urldeclaration:[{include:"@strings"},[`[^)\r
]+`,"string"],["\\)",{token:"tag",next:"@pop"}]],attribute:[{include:"@nestedJSBegin"},{include:"@comments"},{include:"@strings"},{include:"@numbers"},{include:"@keyword"},["[a-zA-Z\\-]+(?=\\()","attribute.value","@attribute"],[">","operator","@pop"],["@identifier","attribute.value"],{include:"@operators"},["@(@identifier)","variable"],["[)\\}]","@brackets","@pop"],["[{}()\\[\\]>]","@brackets"],["[;]","delimiter","@pop"],["[,=:]","delimiter"],["\\s",""],[".","attribute.value"]],comments:[["\\/\\*","comment","@comment"],["\\/\\/+.*","comment"]],comment:[["\\*\\/","comment","@pop"],[".","comment"]],numbers:[["(\\d*\\.)?\\d+([eE][\\-+]?\\d+)?",{token:"attribute.value.number",next:"@units"}],["#[0-9a-fA-F_]+(?!\\w)","attribute.value.hex"]],units:[["(em|ex|ch|rem|fr|vmin|vmax|vw|vh|vm|cm|mm|in|px|pt|pc|deg|grad|rad|turn|s|ms|Hz|kHz|%)?","attribute.value.unit","@pop"]],strings:[['~?"',{token:"string.delimiter",next:"@stringsEndDoubleQuote"}],["~?'",{token:"string.delimiter",next:"@stringsEndQuote"}]],stringsEndDoubleQuote:[['\\\\"',"string"],['"',{token:"string.delimiter",next:"@popall"}],[".","string"]],stringsEndQuote:[["\\\\'","string"],["'",{token:"string.delimiter",next:"@popall"}],[".","string"]],atRules:[{include:"@comments"},{include:"@strings"},["[()]","delimiter"],["[\\{;]","delimiter","@pop"],[".","key"]]}},GUe=Object.freeze(Object.defineProperty({__proto__:null,conf:qUe,language:jUe},Symbol.toStringTag,{value:"Module"})),KUe={comments:{lineComment:"COMMENT"},brackets:[["(",")"]],autoClosingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:'"',close:'"'},{open:":",close:"."}],surroundingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:"`",close:"`"},{open:'"',close:'"'},{open:"'",close:"'"},{open:":",close:"."}],folding:{markers:{start:new RegExp("^\\s*(::\\s*|COMMENT\\s+)#region"),end:new RegExp("^\\s*(::\\s*|COMMENT\\s+)#endregion")}}},YUe={tokenPostfix:".lexon",ignoreCase:!0,keywords:["lexon","lex","clause","terms","contracts","may","pay","pays","appoints","into","to"],typeKeywords:["amount","person","key","time","date","asset","text"],operators:["less","greater","equal","le","gt","or","and","add","added","subtract","subtracted","multiply","multiplied","times","divide","divided","is","be","certified"],symbols:/[=><!~?:&|+\-*\/\^%]+/,tokenizer:{root:[[/^(\s*)(comment:?(?:\s.*|))$/,["","comment"]],[/"/,{token:"identifier.quote",bracket:"@open",next:"@quoted_identifier"}],["LEX$",{token:"keyword",bracket:"@open",next:"@identifier_until_period"}],["LEXON",{token:"keyword",bracket:"@open",next:"@semver"}],[":",{token:"delimiter",bracket:"@open",next:"@identifier_until_period"}],[/[a-z_$][\w$]*/,{cases:{"@operators":"operator","@typeKeywords":"keyword.type","@keywords":"keyword","@default":"identifier"}}],{include:"@whitespace"},[/[{}()\[\]]/,"@brackets"],[/[<>](?!@symbols)/,"@brackets"],[/@symbols/,"delimiter"],[/\d*\.\d*\.\d*/,"number.semver"],[/\d*\.\d+([eE][\-+]?\d+)?/,"number.float"],[/0[xX][0-9a-fA-F]+/,"number.hex"],[/\d+/,"number"],[/[;,.]/,"delimiter"]],quoted_identifier:[[/[^\\"]+/,"identifier"],[/"/,{token:"identifier.quote",bracket:"@close",next:"@pop"}]],space_identifier_until_period:[[":","delimiter"],[" ",{token:"white",next:"@identifier_rest"}]],identifier_until_period:[{include:"@whitespace"},[":",{token:"delimiter",next:"@identifier_rest"}],[/[^\\.]+/,"identifier"],[/\./,{token:"delimiter",bracket:"@close",next:"@pop"}]],identifier_rest:[[/[^\\.]+/,"identifier"],[/\./,{token:"delimiter",bracket:"@close",next:"@pop"}]],semver:[{include:"@whitespace"},[":","delimiter"],[/\d*\.\d*\.\d*/,{token:"number.semver",bracket:"@close",next:"@pop"}]],whitespace:[[/[ \t\r\n]+/,"white"]]}},ZUe=Object.freeze(Object.defineProperty({__proto__:null,conf:KUe,language:YUe},Symbol.toStringTag,{value:"Module"})),XUe={comments:{lineComment:"--",blockComment:["--[[","]]"]},brackets:[["{","}"],["[","]"],["(",")"]],autoClosingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:'"',close:'"'},{open:"'",close:"'"}],surroundingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:'"',close:'"'},{open:"'",close:"'"}]},QUe={defaultToken:"",tokenPostfix:".lua",keywords:["and","break","do","else","elseif","end","false","for","function","goto","if","in","local","nil","not","or","repeat","return","then","true","until","while"],brackets:[{token:"delimiter.bracket",open:"{",close:"}"},{token:"delimiter.array",open:"[",close:"]"},{token:"delimiter.parenthesis",open:"(",close:")"}],operators:["+","-","*","/","%","^","#","==","~=","<=",">=","<",">","=",";",":",",",".","..","..."],symbols:/[=><!~?:&|+\-*\/\^%]+/,escapes:/\\(?:[abfnrtv\\"']|x[0-9A-Fa-f]{1,4}|u[0-9A-Fa-f]{4}|U[0-9A-Fa-f]{8})/,tokenizer:{root:[[/[a-zA-Z_]\w*/,{cases:{"@keywords":{token:"keyword.$0"},"@default":"identifier"}}],{include:"@whitespace"},[/(,)(\s*)([a-zA-Z_]\w*)(\s*)(:)(?!:)/,["delimiter","","key","","delimiter"]],[/({)(\s*)([a-zA-Z_]\w*)(\s*)(:)(?!:)/,["@brackets","","key","","delimiter"]],[/[{}()\[\]]/,"@brackets"],[/@symbols/,{cases:{"@operators":"delimiter","@default":""}}],[/\d*\.\d+([eE][\-+]?\d+)?/,"number.float"],[/0[xX][0-9a-fA-F_]*[0-9a-fA-F]/,"number.hex"],[/\d+?/,"number"],[/[;,.]/,"delimiter"],[/"([^"\\]|\\.)*$/,"string.invalid"],[/'([^'\\]|\\.)*$/,"string.invalid"],[/"/,"string",'@string."'],[/'/,"string","@string.'"]],whitespace:[[/[ \t\r\n]+/,""],[/--\[([=]*)\[/,"comment","@comment.$1"],[/--.*$/,"comment"]],comment:[[/[^\]]+/,"comment"],[/\]([=]*)\]/,{cases:{"$1==$S2":{token:"comment",next:"@pop"},"@default":"comment"}}],[/./,"comment"]],string:[[/[^\\"']+/,"string"],[/@escapes/,"string.escape"],[/\\./,"string.escape.invalid"],[/["']/,{cases:{"$#==$S2":{token:"string",next:"@pop"},"@default":"string"}}]]}},JUe=Object.freeze(Object.defineProperty({__proto__:null,conf:XUe,language:QUe},Symbol.toStringTag,{value:"Module"})),CX=["area","base","br","col","embed","hr","img","input","keygen","link","menuitem","meta","param","source","track","wbr"],eqe={wordPattern:/(-?\d*\.\d\w*)|([^\`\~\!\@\$\^\&\*\(\)\=\+\[\{\]\}\\\|\;\:\'\"\,\.\<\>\/\s]+)/g,brackets:[["<!--","-->"],["<",">"],["{{","}}"],["{%","%}"],["{","}"],["(",")"]],autoClosingPairs:[{open:"{",close:"}"},{open:"%",close:"%"},{open:"[",close:"]"},{open:"(",close:")"},{open:'"',close:'"'},{open:"'",close:"'"}],surroundingPairs:[{open:"<",close:">"},{open:'"',close:'"'},{open:"'",close:"'"}],onEnterRules:[{beforeText:new RegExp(`<(?!(?:${CX.join("|")}))(\\w[\\w\\d]*)([^/>]*(?!/)>)[^<]*$`,"i"),afterText:/^<\/(\w[\w\d]*)\s*>$/i,action:{indentAction:H.IndentAction.IndentOutdent}},{beforeText:new RegExp(`<(?!(?:${CX.join("|")}))(\\w[\\w\\d]*)([^/>]*(?!/)>)[^<]*$`,"i"),action:{indentAction:H.IndentAction.Indent}}]},tqe={defaultToken:"",tokenPostfix:"",builtinTags:["if","else","elseif","endif","render","assign","capture","endcapture","case","endcase","comment","endcomment","cycle","decrement","for","endfor","include","increment","layout","raw","endraw","render","tablerow","endtablerow","unless","endunless"],builtinFilters:["abs","append","at_least","at_most","capitalize","ceil","compact","date","default","divided_by","downcase","escape","escape_once","first","floor","join","json","last","lstrip","map","minus","modulo","newline_to_br","plus","prepend","remove","remove_first","replace","replace_first","reverse","round","rstrip","size","slice","sort","sort_natural","split","strip","strip_html","strip_newlines","times","truncate","truncatewords","uniq","upcase","url_decode","url_encode","where"],constants:["true","false"],operators:["==","!=",">","<",">=","<="],symbol:/[=><!]+/,identifier:/[a-zA-Z_][\w]*/,tokenizer:{root:[[/\{\%\s*comment\s*\%\}/,"comment.start.liquid","@comment"],[/\{\{/,{token:"@rematch",switchTo:"@liquidState.root"}],[/\{\%/,{token:"@rematch",switchTo:"@liquidState.root"}],[/(<)([\w\-]+)(\/>)/,["delimiter.html","tag.html","delimiter.html"]],[/(<)([:\w]+)/,["delimiter.html",{token:"tag.html",next:"@otherTag"}]],[/(<\/)([\w\-]+)/,["delimiter.html",{token:"tag.html",next:"@otherTag"}]],[/</,"delimiter.html"],[/\{/,"delimiter.html"],[/[^<{]+/]],comment:[[/\{\%\s*endcomment\s*\%\}/,"comment.end.liquid","@pop"],[/./,"comment.content.liquid"]],otherTag:[[/\{\{/,{token:"@rematch",switchTo:"@liquidState.otherTag"}],[/\{\%/,{token:"@rematch",switchTo:"@liquidState.otherTag"}],[/\/?>/,"delimiter.html","@pop"],[/"([^"]*)"/,"attribute.value"],[/'([^']*)'/,"attribute.value"],[/[\w\-]+/,"attribute.name"],[/=/,"delimiter"],[/[ \t\r\n]+/]],liquidState:[[/\{\{/,"delimiter.output.liquid"],[/\}\}/,{token:"delimiter.output.liquid",switchTo:"@$S2.$S3"}],[/\{\%/,"delimiter.tag.liquid"],[/raw\s*\%\}/,"delimiter.tag.liquid","@liquidRaw"],[/\%\}/,{token:"delimiter.tag.liquid",switchTo:"@$S2.$S3"}],{include:"liquidRoot"}],liquidRaw:[[/^(?!\{\%\s*endraw\s*\%\}).+/],[/\{\%/,"delimiter.tag.liquid"],[/@identifier/],[/\%\}/,{token:"delimiter.tag.liquid",next:"@root"}]],liquidRoot:[[/\d+(\.\d+)?/,"number.liquid"],[/"[^"]*"/,"string.liquid"],[/'[^']*'/,"string.liquid"],[/\s+/],[/@symbol/,{cases:{"@operators":"operator.liquid","@default":""}}],[/\./],[/@identifier/,{cases:{"@constants":"keyword.liquid","@builtinFilters":"predefined.liquid","@builtinTags":"predefined.liquid","@default":"variable.liquid"}}],[/[^}|%]/,"variable.liquid"]]}},iqe=Object.freeze(Object.defineProperty({__proto__:null,conf:eqe,language:tqe},Symbol.toStringTag,{value:"Module"})),nqe={comments:{blockComment:["(*","*)"]},brackets:[["{","}"],["[","]"],["(",")"]],autoClosingPairs:[{open:"[",close:"]"},{open:"{",close:"}"},{open:"(",close:")"},{open:"(*",close:"*)"},{open:"<*",close:"*>"},{open:"'",close:"'",notIn:["string","comment"]},{open:'"',close:'"',notIn:["string","comment"]}]},sqe={defaultToken:"",tokenPostfix:".m3",brackets:[{token:"delimiter.curly",open:"{",close:"}"},{token:"delimiter.parenthesis",open:"(",close:")"},{token:"delimiter.square",open:"[",close:"]"}],keywords:["AND","ANY","ARRAY","AS","BEGIN","BITS","BRANDED","BY","CASE","CONST","DIV","DO","ELSE","ELSIF","END","EVAL","EXCEPT","EXCEPTION","EXIT","EXPORTS","FINALLY","FOR","FROM","GENERIC","IF","IMPORT","IN","INTERFACE","LOCK","LOOP","METHODS","MOD","MODULE","NOT","OBJECT","OF","OR","OVERRIDES","PROCEDURE","RAISE","RAISES","READONLY","RECORD","REF","REPEAT","RETURN","REVEAL","SET","THEN","TO","TRY","TYPE","TYPECASE","UNSAFE","UNTIL","UNTRACED","VALUE","VAR","WHILE","WITH"],reservedConstNames:["ABS","ADR","ADRSIZE","BITSIZE","BYTESIZE","CEILING","DEC","DISPOSE","FALSE","FIRST","FLOAT","FLOOR","INC","ISTYPE","LAST","LOOPHOLE","MAX","MIN","NARROW","NEW","NIL","NUMBER","ORD","ROUND","SUBARRAY","TRUE","TRUNC","TYPECODE","VAL"],reservedTypeNames:["ADDRESS","ANY","BOOLEAN","CARDINAL","CHAR","EXTENDED","INTEGER","LONGCARD","LONGINT","LONGREAL","MUTEX","NULL","REAL","REFANY","ROOT","TEXT"],operators:["+","-","*","/","&","^","."],relations:["=","#","<","<=",">",">=","<:",":"],delimiters:["|","..","=>",",",";",":="],symbols:/[>=<#.,:;+\-*/&^]+/,escapes:/\\(?:[\\fnrt"']|[0-7]{3})/,tokenizer:{root:[[/_\w*/,"invalid"],[/[a-zA-Z][a-zA-Z0-9_]*/,{cases:{"@keywords":{token:"keyword.$0"},"@reservedConstNames":{token:"constant.reserved.$0"},"@reservedTypeNames":{token:"type.reserved.$0"},"@default":"identifier"}}],{include:"@whitespace"},[/[{}()\[\]]/,"@brackets"],[/[0-9]+\.[0-9]+(?:[DdEeXx][\+\-]?[0-9]+)?/,"number.float"],[/[0-9]+(?:\_[0-9a-fA-F]+)?L?/,"number"],[/@symbols/,{cases:{"@operators":"operators","@relations":"operators","@delimiters":"delimiter","@default":"invalid"}}],[/'[^\\']'/,"string.char"],[/(')(@escapes)(')/,["string.char","string.escape","string.char"]],[/'/,"invalid"],[/"([^"\\]|\\.)*$/,"invalid"],[/"/,"string.text","@text"]],text:[[/[^\\"]+/,"string.text"],[/@escapes/,"string.escape"],[/\\./,"invalid"],[/"/,"string.text","@pop"]],comment:[[/\(\*/,"comment","@push"],[/\*\)/,"comment","@pop"],[/./,"comment"]],pragma:[[/<\*/,"keyword.pragma","@push"],[/\*>/,"keyword.pragma","@pop"],[/./,"keyword.pragma"]],whitespace:[[/[ \t\r\n]+/,"white"],[/\(\*/,"comment","@comment"],[/<\*/,"keyword.pragma","@pragma"]]}},oqe=Object.freeze(Object.defineProperty({__proto__:null,conf:nqe,language:sqe},Symbol.toStringTag,{value:"Module"})),rqe={comments:{blockComment:["<!--","-->"]},brackets:[["{","}"],["[","]"],["(",")"]],autoClosingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:"<",close:">",notIn:["string"]}],surroundingPairs:[{open:"(",close:")"},{open:"[",close:"]"},{open:"`",close:"`"}],folding:{markers:{start:new RegExp("^\\s*<!--\\s*#?region\\b.*-->"),end:new RegExp("^\\s*<!--\\s*#?endregion\\b.*-->")}}},aqe={defaultToken:"",tokenPostfix:".md",control:/[\\`*_\[\]{}()#+\-\.!]/,noncontrol:/[^\\`*_\[\]{}()#+\-\.!]/,escapes:/\\(?:@control)/,jsescapes:/\\(?:[btnfr\\"']|[0-7][0-7]?|[0-3][0-7]{2})/,empty:["area","base","basefont","br","col","frame","hr","img","input","isindex","link","meta","param"],tokenizer:{root:[[/^\s*\|/,"@rematch","@table_header"],[/^(\s{0,3})(#+)((?:[^\\#]|@escapes)+)((?:#+)?)/,["white","keyword","keyword","keyword"]],[/^\s*(=+|\-+)\s*$/,"keyword"],[/^\s*((\*[ ]?)+)\s*$/,"meta.separator"],[/^\s*>+/,"comment"],[/^\s*([\*\-+:]|\d+\.)\s/,"keyword"],[/^(\t|[ ]{4})[^ ].*$/,"string"],[/^\s*~~~\s*((?:\w|[\/\-#])+)?\s*$/,{token:"string",next:"@codeblock"}],[/^\s*```\s*((?:\w|[\/\-#])+).*$/,{token:"string",next:"@codeblockgh",nextEmbedded:"$1"}],[/^\s*```\s*$/,{token:"string",next:"@codeblock"}],{include:"@linecontent"}],table_header:[{include:"@table_common"},[/[^\|]+/,"keyword.table.header"]],table_body:[{include:"@table_common"},{include:"@linecontent"}],table_common:[[/\s*[\-:]+\s*/,{token:"keyword",switchTo:"table_body"}],[/^\s*\|/,"keyword.table.left"],[/^\s*[^\|]/,"@rematch","@pop"],[/^\s*$/,"@rematch","@pop"],[/\|/,{cases:{"@eos":"keyword.table.right","@default":"keyword.table.middle"}}]],codeblock:[[/^\s*~~~\s*$/,{token:"string",next:"@pop"}],[/^\s*```\s*$/,{token:"string",next:"@pop"}],[/.*$/,"variable.source"]],codeblockgh:[[/```\s*$/,{token:"string",next:"@pop",nextEmbedded:"@pop"}],[/[^`]+/,"variable.source"]],linecontent:[[/&\w+;/,"string.escape"],[/@escapes/,"escape"],[/\b__([^\\_]|@escapes|_(?!_))+__\b/,"strong"],[/\*\*([^\\*]|@escapes|\*(?!\*))+\*\*/,"strong"],[/\b_[^_]+_\b/,"emphasis"],[/\*([^\\*]|@escapes)+\*/,"emphasis"],[/`([^\\`]|@escapes)+`/,"variable"],[/\{+[^}]+\}+/,"string.target"],[/(!?\[)((?:[^\]\\]|@escapes)*)(\]\([^\)]+\))/,["string.link","","string.link"]],[/(!?\[)((?:[^\]\\]|@escapes)*)(\])/,"string.link"],{include:"html"}],html:[[/<(\w+)\/>/,"tag"],[/<(\w+)(\-|\w)*/,{cases:{"@empty":{token:"tag",next:"@tag.$1"},"@default":{token:"tag",next:"@tag.$1"}}}],[/<\/(\w+)(\-|\w)*\s*>/,{token:"tag"}],[/<!--/,"comment","@comment"]],comment:[[/[^<\-]+/,"comment.content"],[/-->/,"comment","@pop"],[/<!--/,"comment.content.invalid"],[/[<\-]/,"comment.content"]],tag:[[/[ \t\r\n]+/,"white"],[/(type)(\s*=\s*)(")([^"]+)(")/,["attribute.name.html","delimiter.html","string.html",{token:"string.html",switchTo:"@tag.$S2.$4"},"string.html"]],[/(type)(\s*=\s*)(')([^']+)(')/,["attribute.name.html","delimiter.html","string.html",{token:"string.html",switchTo:"@tag.$S2.$4"},"string.html"]],[/(\w+)(\s*=\s*)("[^"]*"|'[^']*')/,["attribute.name.html","delimiter.html","string.html"]],[/\w+/,"attribute.name.html"],[/\/>/,"tag","@pop"],[/>/,{cases:{"$S2==style":{token:"tag",switchTo:"embeddedStyle",nextEmbedded:"text/css"},"$S2==script":{cases:{$S3:{token:"tag",switchTo:"embeddedScript",nextEmbedded:"$S3"},"@default":{token:"tag",switchTo:"embeddedScript",nextEmbedded:"text/javascript"}}},"@default":{token:"tag",next:"@pop"}}}]],embeddedStyle:[[/[^<]+/,""],[/<\/style\s*>/,{token:"@rematch",next:"@pop",nextEmbedded:"@pop"}],[/</,""]],embeddedScript:[[/[^<]+/,""],[/<\/script\s*>/,{token:"@rematch",next:"@pop",nextEmbedded:"@pop"}],[/</,""]]}},lqe=Object.freeze(Object.defineProperty({__proto__:null,conf:rqe,language:aqe},Symbol.toStringTag,{value:"Module"})),cqe={comments:{blockComment:["{/*","*/}"]},brackets:[["{","}"]],autoClosingPairs:[{open:'"',close:'"'},{open:"'",close:"'"},{open:"“",close:"”"},{open:"‘",close:"’"},{open:"`",close:"`"},{open:"{",close:"}"},{open:"(",close:")"},{open:"_",close:"_"},{open:"**",close:"**"},{open:"<",close:">"}],onEnterRules:[{beforeText:/^\s*- .+/,action:{indentAction:H.IndentAction.None,appendText:"- "}},{beforeText:/^\s*\+ .+/,action:{indentAction:H.IndentAction.None,appendText:"+ "}},{beforeText:/^\s*\* .+/,action:{indentAction:H.IndentAction.None,appendText:"* "}},{beforeText:/^> /,action:{indentAction:H.IndentAction.None,appendText:"> "}},{beforeText:/<\w+/,action:{indentAction:H.IndentAction.Indent}},{beforeText:/\s+>\s*$/,action:{indentAction:H.IndentAction.Indent}},{beforeText:/<\/\w+>/,action:{indentAction:H.IndentAction.Outdent}},...Array.from({length:100},(s,e)=>({beforeText:new RegExp(`^${e}\\. .+`),action:{indentAction:H.IndentAction.None,appendText:`${e+1}. `}}))]},dqe={defaultToken:"",tokenPostfix:".mdx",control:/[!#()*+.[\\\]_`{}\-]/,escapes:/\\@control/,tokenizer:{root:[[/^---$/,{token:"meta.content",next:"@frontmatter",nextEmbedded:"yaml"}],[/^\s*import/,{token:"keyword",next:"@import",nextEmbedded:"js"}],[/^\s*export/,{token:"keyword",next:"@export",nextEmbedded:"js"}],[/<\w+/,{token:"type.identifier",next:"@jsx"}],[/<\/?\w+>/,"type.identifier"],[/^(\s*)(>*\s*)(#{1,6}\s)/,[{token:"white"},{token:"comment"},{token:"keyword",next:"@header"}]],[/^(\s*)(>*\s*)([*+-])(\s+)/,["white","comment","keyword","white"]],[/^(\s*)(>*\s*)(\d{1,9}\.)(\s+)/,["white","comment","number","white"]],[/^(\s*)(>*\s*)(\d{1,9}\.)(\s+)/,["white","comment","number","white"]],[/^(\s*)(>*\s*)(-{3,}|\*{3,}|_{3,})$/,["white","comment","keyword"]],[/`{3,}(\s.*)?$/,{token:"string",next:"@codeblock_backtick"}],[/~{3,}(\s.*)?$/,{token:"string",next:"@codeblock_tilde"}],[/`{3,}(\S+).*$/,{token:"string",next:"@codeblock_highlight_backtick",nextEmbedded:"$1"}],[/~{3,}(\S+).*$/,{token:"string",next:"@codeblock_highlight_tilde",nextEmbedded:"$1"}],[/^(\s*)(-{4,})$/,["white","comment"]],[/^(\s*)(>+)/,["white","comment"]],{include:"content"}],content:[[/(\[)(.+)(]\()(.+)(\s+".*")(\))/,["","string.link","","type.identifier","string.link",""]],[/(\[)(.+)(]\()(.+)(\))/,["","type.identifier","","string.link",""]],[/(\[)(.+)(]\[)(.+)(])/,["","type.identifier","","type.identifier",""]],[/(\[)(.+)(]:\s+)(\S*)/,["","type.identifier","","string.link"]],[/(\[)(.+)(])/,["","type.identifier",""]],[/`.*`/,"variable.source"],[/_/,{token:"emphasis",next:"@emphasis_underscore"}],[/\*(?!\*)/,{token:"emphasis",next:"@emphasis_asterisk"}],[/\*\*/,{token:"strong",next:"@strong"}],[/{/,{token:"delimiter.bracket",next:"@expression",nextEmbedded:"js"}]],import:[[/'\s*(;|$)/,{token:"string",next:"@pop",nextEmbedded:"@pop"}]],expression:[[/{/,{token:"delimiter.bracket",next:"@expression"}],[/}/,{token:"delimiter.bracket",next:"@pop",nextEmbedded:"@pop"}]],export:[[/^\s*$/,{token:"delimiter.bracket",next:"@pop",nextEmbedded:"@pop"}]],jsx:[[/\s+/,""],[/(\w+)(=)("(?:[^"\\]|\\.)*")/,["attribute.name","operator","string"]],[/(\w+)(=)('(?:[^'\\]|\\.)*')/,["attribute.name","operator","string"]],[/(\w+(?=\s|>|={|$))/,["attribute.name"]],[/={/,{token:"delimiter.bracket",next:"@expression",nextEmbedded:"js"}],[/>/,{token:"type.identifier",next:"@pop"}]],header:[[/.$/,{token:"keyword",next:"@pop"}],{include:"content"},[/./,{token:"keyword"}]],strong:[[/\*\*/,{token:"strong",next:"@pop"}],{include:"content"},[/./,{token:"strong"}]],emphasis_underscore:[[/_/,{token:"emphasis",next:"@pop"}],{include:"content"},[/./,{token:"emphasis"}]],emphasis_asterisk:[[/\*(?!\*)/,{token:"emphasis",next:"@pop"}],{include:"content"},[/./,{token:"emphasis"}]],frontmatter:[[/^---$/,{token:"meta.content",nextEmbedded:"@pop",next:"@pop"}]],codeblock_highlight_backtick:[[/\s*`{3,}\s*$/,{token:"string",next:"@pop",nextEmbedded:"@pop"}],[/.*$/,"variable.source"]],codeblock_highlight_tilde:[[/\s*~{3,}\s*$/,{token:"string",next:"@pop",nextEmbedded:"@pop"}],[/.*$/,"variable.source"]],codeblock_backtick:[[/\s*`{3,}\s*$/,{token:"string",next:"@pop"}],[/.*$/,"variable.source"]],codeblock_tilde:[[/\s*~{3,}\s*$/,{token:"string",next:"@pop"}],[/.*$/,"variable.source"]]}},hqe=Object.freeze(Object.defineProperty({__proto__:null,conf:cqe,language:dqe},Symbol.toStringTag,{value:"Module"})),uqe={wordPattern:/(-?\d*\.\d\w*)|([^\`\~\!\@\#%\^\&\*\(\)\=\$\-\+\[\{\]\}\\\|\;\:\'\"\,\.\<\>\/\?\s]+)/g,comments:{blockComment:["###","###"],lineComment:"#"},folding:{markers:{start:new RegExp("^\\s*#region\\b"),end:new RegExp("^\\s*#endregion\\b")}}},gqe={defaultToken:"",ignoreCase:!1,tokenPostfix:".mips",regEx:/\/(?!\/\/)(?:[^\/\\]|\\.)*\/[igm]*/,keywords:[".data",".text","syscall","trap","add","addu","addi","addiu","and","andi","div","divu","mult","multu","nor","or","ori","sll","slv","sra","srav","srl","srlv","sub","subu","xor","xori","lhi","lho","lhi","llo","slt","slti","sltu","sltiu","beq","bgtz","blez","bne","j","jal","jalr","jr","lb","lbu","lh","lhu","lw","li","la","sb","sh","sw","mfhi","mflo","mthi","mtlo","move"],symbols:/[\.,\:]+/,escapes:/\\(?:[abfnrtv\\"'$]|x[0-9A-Fa-f]{1,4}|u[0-9A-Fa-f]{4}|U[0-9A-Fa-f]{8})/,tokenizer:{root:[[/\$[a-zA-Z_]\w*/,"variable.predefined"],[/[.a-zA-Z_]\w*/,{cases:{this:"variable.predefined","@keywords":{token:"keyword.$0"},"@default":""}}],[/[ \t\r\n]+/,""],[/#.*$/,"comment"],["///",{token:"regexp",next:"@hereregexp"}],[/^(\s*)(@regEx)/,["","regexp"]],[/(\,)(\s*)(@regEx)/,["delimiter","","regexp"]],[/(\:)(\s*)(@regEx)/,["delimiter","","regexp"]],[/@symbols/,"delimiter"],[/\d+[eE]([\-+]?\d+)?/,"number.float"],[/\d+\.\d+([eE][\-+]?\d+)?/,"number.float"],[/0[xX][0-9a-fA-F]+/,"number.hex"],[/0[0-7]+(?!\d)/,"number.octal"],[/\d+/,"number"],[/[,.]/,"delimiter"],[/"""/,"string",'@herestring."""'],[/'''/,"string","@herestring.'''"],[/"/,{cases:{"@eos":"string","@default":{token:"string",next:'@string."'}}}],[/'/,{cases:{"@eos":"string","@default":{token:"string",next:"@string.'"}}}]],string:[[/[^"'\#\\]+/,"string"],[/@escapes/,"string.escape"],[/\./,"string.escape.invalid"],[/\./,"string.escape.invalid"],[/#{/,{cases:{'$S2=="':{token:"string",next:"root.interpolatedstring"},"@default":"string"}}],[/["']/,{cases:{"$#==$S2":{token:"string",next:"@pop"},"@default":"string"}}],[/#/,"string"]],herestring:[[/("""|''')/,{cases:{"$1==$S2":{token:"string",next:"@pop"},"@default":"string"}}],[/[^#\\'"]+/,"string"],[/['"]+/,"string"],[/@escapes/,"string.escape"],[/\./,"string.escape.invalid"],[/#{/,{token:"string.quote",next:"root.interpolatedstring"}],[/#/,"string"]],comment:[[/[^#]+/,"comment"],[/#/,"comment"]],hereregexp:[[/[^\\\/#]+/,"regexp"],[/\\./,"regexp"],[/#.*$/,"comment"],["///[igm]*",{token:"regexp",next:"@pop"}],[/\//,"regexp"]]}},fqe=Object.freeze(Object.defineProperty({__proto__:null,conf:uqe,language:gqe},Symbol.toStringTag,{value:"Module"})),pqe={comments:{lineComment:"//",blockComment:["/*","*/"]},brackets:[["[","]"],["(",")"],["{","}"]],autoClosingPairs:[{open:'"',close:'"',notIn:["string","comment"]},{open:"'",close:"'",notIn:["string","comment"]},{open:"[",close:"]",notIn:["string","comment"]},{open:"(",close:")",notIn:["string","comment"]},{open:"{",close:"}",notIn:["string","comment"]}]},mqe={defaultToken:"",tokenPostfix:".msdax",ignoreCase:!0,brackets:[{open:"[",close:"]",token:"delimiter.square"},{open:"{",close:"}",token:"delimiter.brackets"},{open:"(",close:")",token:"delimiter.parenthesis"}],keywords:["VAR","RETURN","NOT","EVALUATE","DATATABLE","ORDER","BY","START","AT","DEFINE","MEASURE","ASC","DESC","IN","BOOLEAN","DOUBLE","INTEGER","DATETIME","CURRENCY","STRING"],functions:["CLOSINGBALANCEMONTH","CLOSINGBALANCEQUARTER","CLOSINGBALANCEYEAR","DATEADD","DATESBETWEEN","DATESINPERIOD","DATESMTD","DATESQTD","DATESYTD","ENDOFMONTH","ENDOFQUARTER","ENDOFYEAR","FIRSTDATE","FIRSTNONBLANK","LASTDATE","LASTNONBLANK","NEXTDAY","NEXTMONTH","NEXTQUARTER","NEXTYEAR","OPENINGBALANCEMONTH","OPENINGBALANCEQUARTER","OPENINGBALANCEYEAR","PARALLELPERIOD","PREVIOUSDAY","PREVIOUSMONTH","PREVIOUSQUARTER","PREVIOUSYEAR","SAMEPERIODLASTYEAR","STARTOFMONTH","STARTOFQUARTER","STARTOFYEAR","TOTALMTD","TOTALQTD","TOTALYTD","ADDCOLUMNS","ADDMISSINGITEMS","ALL","ALLEXCEPT","ALLNOBLANKROW","ALLSELECTED","CALCULATE","CALCULATETABLE","CALENDAR","CALENDARAUTO","CROSSFILTER","CROSSJOIN","CURRENTGROUP","DATATABLE","DETAILROWS","DISTINCT","EARLIER","EARLIEST","EXCEPT","FILTER","FILTERS","GENERATE","GENERATEALL","GROUPBY","IGNORE","INTERSECT","ISONORAFTER","KEEPFILTERS","LOOKUPVALUE","NATURALINNERJOIN","NATURALLEFTOUTERJOIN","RELATED","RELATEDTABLE","ROLLUP","ROLLUPADDISSUBTOTAL","ROLLUPGROUP","ROLLUPISSUBTOTAL","ROW","SAMPLE","SELECTCOLUMNS","SUBSTITUTEWITHINDEX","SUMMARIZE","SUMMARIZECOLUMNS","TOPN","TREATAS","UNION","USERELATIONSHIP","VALUES","SUM","SUMX","PATH","PATHCONTAINS","PATHITEM","PATHITEMREVERSE","PATHLENGTH","AVERAGE","AVERAGEA","AVERAGEX","COUNT","COUNTA","COUNTAX","COUNTBLANK","COUNTROWS","COUNTX","DISTINCTCOUNT","DIVIDE","GEOMEAN","GEOMEANX","MAX","MAXA","MAXX","MEDIAN","MEDIANX","MIN","MINA","MINX","PERCENTILE.EXC","PERCENTILE.INC","PERCENTILEX.EXC","PERCENTILEX.INC","PRODUCT","PRODUCTX","RANK.EQ","RANKX","STDEV.P","STDEV.S","STDEVX.P","STDEVX.S","VAR.P","VAR.S","VARX.P","VARX.S","XIRR","XNPV","DATE","DATEDIFF","DATEVALUE","DAY","EDATE","EOMONTH","HOUR","MINUTE","MONTH","NOW","SECOND","TIME","TIMEVALUE","TODAY","WEEKDAY","WEEKNUM","YEAR","YEARFRAC","CONTAINS","CONTAINSROW","CUSTOMDATA","ERROR","HASONEFILTER","HASONEVALUE","ISBLANK","ISCROSSFILTERED","ISEMPTY","ISERROR","ISEVEN","ISFILTERED","ISLOGICAL","ISNONTEXT","ISNUMBER","ISODD","ISSUBTOTAL","ISTEXT","USERNAME","USERPRINCIPALNAME","AND","FALSE","IF","IFERROR","NOT","OR","SWITCH","TRUE","ABS","ACOS","ACOSH","ACOT","ACOTH","ASIN","ASINH","ATAN","ATANH","BETA.DIST","BETA.INV","CEILING","CHISQ.DIST","CHISQ.DIST.RT","CHISQ.INV","CHISQ.INV.RT","COMBIN","COMBINA","CONFIDENCE.NORM","CONFIDENCE.T","COS","COSH","COT","COTH","CURRENCY","DEGREES","EVEN","EXP","EXPON.DIST","FACT","FLOOR","GCD","INT","ISO.CEILING","LCM","LN","LOG","LOG10","MOD","MROUND","ODD","PERMUT","PI","POISSON.DIST","POWER","QUOTIENT","RADIANS","RAND","RANDBETWEEN","ROUND","ROUNDDOWN","ROUNDUP","SIGN","SIN","SINH","SQRT","SQRTPI","TAN","TANH","TRUNC","BLANK","CONCATENATE","CONCATENATEX","EXACT","FIND","FIXED","FORMAT","LEFT","LEN","LOWER","MID","REPLACE","REPT","RIGHT","SEARCH","SUBSTITUTE","TRIM","UNICHAR","UNICODE","UPPER","VALUE"],tokenizer:{root:[{include:"@comments"},{include:"@whitespace"},{include:"@numbers"},{include:"@strings"},{include:"@complexIdentifiers"},[/[;,.]/,"delimiter"],[/[({})]/,"@brackets"],[/[a-z_][a-zA-Z0-9_]*/,{cases:{"@keywords":"keyword","@functions":"keyword","@default":"identifier"}}],[/[<>=!%&+\-*/|~^]/,"operator"]],whitespace:[[/\s+/,"white"]],comments:[[/\/\/+.*/,"comment"],[/\/\*/,{token:"comment.quote",next:"@comment"}]],comment:[[/[^*/]+/,"comment"],[/\*\//,{token:"comment.quote",next:"@pop"}],[/./,"comment"]],numbers:[[/0[xX][0-9a-fA-F]*/,"number"],[/[$][+-]*\d*(\.\d*)?/,"number"],[/((\d+(\.\d*)?)|(\.\d+))([eE][\-+]?\d+)?/,"number"]],strings:[[/N"/,{token:"string",next:"@string"}],[/"/,{token:"string",next:"@string"}]],string:[[/[^"]+/,"string"],[/""/,"string"],[/"/,{token:"string",next:"@pop"}]],complexIdentifiers:[[/\[/,{token:"identifier.quote",next:"@bracketedIdentifier"}],[/'/,{token:"identifier.quote",next:"@quotedIdentifier"}]],bracketedIdentifier:[[/[^\]]+/,"identifier"],[/]]/,"identifier"],[/]/,{token:"identifier.quote",next:"@pop"}]],quotedIdentifier:[[/[^']+/,"identifier"],[/''/,"identifier"],[/'/,{token:"identifier.quote",next:"@pop"}]]}},_qe=Object.freeze(Object.defineProperty({__proto__:null,conf:pqe,language:mqe},Symbol.toStringTag,{value:"Module"})),bqe={comments:{lineComment:"--",blockComment:["/*","*/"]},brackets:[["{","}"],["[","]"],["(",")"]],autoClosingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:'"',close:'"'},{open:"'",close:"'"}],surroundingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:'"',close:'"'},{open:"'",close:"'"}]},wqe={defaultToken:"",tokenPostfix:".sql",ignoreCase:!0,brackets:[{open:"[",close:"]",token:"delimiter.square"},{open:"(",close:")",token:"delimiter.parenthesis"}],keywords:["ACCESSIBLE","ADD","ALL","ALTER","ANALYZE","AND","AS","ASC","ASENSITIVE","BEFORE","BETWEEN","BIGINT","BINARY","BLOB","BOTH","BY","CALL","CASCADE","CASE","CHANGE","CHAR","CHARACTER","CHECK","COLLATE","COLUMN","CONDITION","CONSTRAINT","CONTINUE","CONVERT","CREATE","CROSS","CUBE","CUME_DIST","CURRENT_DATE","CURRENT_TIME","CURRENT_TIMESTAMP","CURRENT_USER","CURSOR","DATABASE","DATABASES","DAY_HOUR","DAY_MICROSECOND","DAY_MINUTE","DAY_SECOND","DEC","DECIMAL","DECLARE","DEFAULT","DELAYED","DELETE","DENSE_RANK","DESC","DESCRIBE","DETERMINISTIC","DISTINCT","DISTINCTROW","DIV","DOUBLE","DROP","DUAL","EACH","ELSE","ELSEIF","EMPTY","ENCLOSED","ESCAPED","EXCEPT","EXISTS","EXIT","EXPLAIN","FALSE","FETCH","FIRST_VALUE","FLOAT","FLOAT4","FLOAT8","FOR","FORCE","FOREIGN","FROM","FULLTEXT","FUNCTION","GENERATED","GET","GRANT","GROUP","GROUPING","GROUPS","HAVING","HIGH_PRIORITY","HOUR_MICROSECOND","HOUR_MINUTE","HOUR_SECOND","IF","IGNORE","IN","INDEX","INFILE","INNER","INOUT","INSENSITIVE","INSERT","INT","INT1","INT2","INT3","INT4","INT8","INTEGER","INTERVAL","INTO","IO_AFTER_GTIDS","IO_BEFORE_GTIDS","IS","ITERATE","JOIN","JSON_TABLE","KEY","KEYS","KILL","LAG","LAST_VALUE","LATERAL","LEAD","LEADING","LEAVE","LEFT","LIKE","LIMIT","LINEAR","LINES","LOAD","LOCALTIME","LOCALTIMESTAMP","LOCK","LONG","LONGBLOB","LONGTEXT","LOOP","LOW_PRIORITY","MASTER_BIND","MASTER_SSL_VERIFY_SERVER_CERT","MATCH","MAXVALUE","MEDIUMBLOB","MEDIUMINT","MEDIUMTEXT","MIDDLEINT","MINUTE_MICROSECOND","MINUTE_SECOND","MOD","MODIFIES","NATURAL","NOT","NO_WRITE_TO_BINLOG","NTH_VALUE","NTILE","NULL","NUMERIC","OF","ON","OPTIMIZE","OPTIMIZER_COSTS","OPTION","OPTIONALLY","OR","ORDER","OUT","OUTER","OUTFILE","OVER","PARTITION","PERCENT_RANK","PRECISION","PRIMARY","PROCEDURE","PURGE","RANGE","RANK","READ","READS","READ_WRITE","REAL","RECURSIVE","REFERENCES","REGEXP","RELEASE","RENAME","REPEAT","REPLACE","REQUIRE","RESIGNAL","RESTRICT","RETURN","REVOKE","RIGHT","RLIKE","ROW","ROWS","ROW_NUMBER","SCHEMA","SCHEMAS","SECOND_MICROSECOND","SELECT","SENSITIVE","SEPARATOR","SET","SHOW","SIGNAL","SMALLINT","SPATIAL","SPECIFIC","SQL","SQLEXCEPTION","SQLSTATE","SQLWARNING","SQL_BIG_RESULT","SQL_CALC_FOUND_ROWS","SQL_SMALL_RESULT","SSL","STARTING","STORED","STRAIGHT_JOIN","SYSTEM","TABLE","TERMINATED","THEN","TINYBLOB","TINYINT","TINYTEXT","TO","TRAILING","TRIGGER","TRUE","UNDO","UNION","UNIQUE","UNLOCK","UNSIGNED","UPDATE","USAGE","USE","USING","UTC_DATE","UTC_TIME","UTC_TIMESTAMP","VALUES","VARBINARY","VARCHAR","VARCHARACTER","VARYING","VIRTUAL","WHEN","WHERE","WHILE","WINDOW","WITH","WRITE","XOR","YEAR_MONTH","ZEROFILL"],operators:["AND","BETWEEN","IN","LIKE","NOT","OR","IS","NULL","INTERSECT","UNION","INNER","JOIN","LEFT","OUTER","RIGHT"],builtinFunctions:["ABS","ACOS","ADDDATE","ADDTIME","AES_DECRYPT","AES_ENCRYPT","ANY_VALUE","Area","AsBinary","AsWKB","ASCII","ASIN","AsText","AsWKT","ASYMMETRIC_DECRYPT","ASYMMETRIC_DERIVE","ASYMMETRIC_ENCRYPT","ASYMMETRIC_SIGN","ASYMMETRIC_VERIFY","ATAN","ATAN2","ATAN","AVG","BENCHMARK","BIN","BIT_AND","BIT_COUNT","BIT_LENGTH","BIT_OR","BIT_XOR","Buffer","CAST","CEIL","CEILING","Centroid","CHAR","CHAR_LENGTH","CHARACTER_LENGTH","CHARSET","COALESCE","COERCIBILITY","COLLATION","COMPRESS","CONCAT","CONCAT_WS","CONNECTION_ID","Contains","CONV","CONVERT","CONVERT_TZ","ConvexHull","COS","COT","COUNT","CRC32","CREATE_ASYMMETRIC_PRIV_KEY","CREATE_ASYMMETRIC_PUB_KEY","CREATE_DH_PARAMETERS","CREATE_DIGEST","Crosses","CUME_DIST","CURDATE","CURRENT_DATE","CURRENT_ROLE","CURRENT_TIME","CURRENT_TIMESTAMP","CURRENT_USER","CURTIME","DATABASE","DATE","DATE_ADD","DATE_FORMAT","DATE_SUB","DATEDIFF","DAY","DAYNAME","DAYOFMONTH","DAYOFWEEK","DAYOFYEAR","DECODE","DEFAULT","DEGREES","DES_DECRYPT","DES_ENCRYPT","DENSE_RANK","Dimension","Disjoint","Distance","ELT","ENCODE","ENCRYPT","EndPoint","Envelope","Equals","EXP","EXPORT_SET","ExteriorRing","EXTRACT","ExtractValue","FIELD","FIND_IN_SET","FIRST_VALUE","FLOOR","FORMAT","FORMAT_BYTES","FORMAT_PICO_TIME","FOUND_ROWS","FROM_BASE64","FROM_DAYS","FROM_UNIXTIME","GEN_RANGE","GEN_RND_EMAIL","GEN_RND_PAN","GEN_RND_SSN","GEN_RND_US_PHONE","GeomCollection","GeomCollFromText","GeometryCollectionFromText","GeomCollFromWKB","GeometryCollectionFromWKB","GeometryCollection","GeometryN","GeometryType","GeomFromText","GeometryFromText","GeomFromWKB","GeometryFromWKB","GET_FORMAT","GET_LOCK","GLength","GREATEST","GROUP_CONCAT","GROUPING","GTID_SUBSET","GTID_SUBTRACT","HEX","HOUR","ICU_VERSION","IF","IFNULL","INET_ATON","INET_NTOA","INET6_ATON","INET6_NTOA","INSERT","INSTR","InteriorRingN","Intersects","INTERVAL","IS_FREE_LOCK","IS_IPV4","IS_IPV4_COMPAT","IS_IPV4_MAPPED","IS_IPV6","IS_USED_LOCK","IS_UUID","IsClosed","IsEmpty","ISNULL","IsSimple","JSON_APPEND","JSON_ARRAY","JSON_ARRAY_APPEND","JSON_ARRAY_INSERT","JSON_ARRAYAGG","JSON_CONTAINS","JSON_CONTAINS_PATH","JSON_DEPTH","JSON_EXTRACT","JSON_INSERT","JSON_KEYS","JSON_LENGTH","JSON_MERGE","JSON_MERGE_PATCH","JSON_MERGE_PRESERVE","JSON_OBJECT","JSON_OBJECTAGG","JSON_OVERLAPS","JSON_PRETTY","JSON_QUOTE","JSON_REMOVE","JSON_REPLACE","JSON_SCHEMA_VALID","JSON_SCHEMA_VALIDATION_REPORT","JSON_SEARCH","JSON_SET","JSON_STORAGE_FREE","JSON_STORAGE_SIZE","JSON_TABLE","JSON_TYPE","JSON_UNQUOTE","JSON_VALID","LAG","LAST_DAY","LAST_INSERT_ID","LAST_VALUE","LCASE","LEAD","LEAST","LEFT","LENGTH","LineFromText","LineStringFromText","LineFromWKB","LineStringFromWKB","LineString","LN","LOAD_FILE","LOCALTIME","LOCALTIMESTAMP","LOCATE","LOG","LOG10","LOG2","LOWER","LPAD","LTRIM","MAKE_SET","MAKEDATE","MAKETIME","MASK_INNER","MASK_OUTER","MASK_PAN","MASK_PAN_RELAXED","MASK_SSN","MASTER_POS_WAIT","MAX","MBRContains","MBRCoveredBy","MBRCovers","MBRDisjoint","MBREqual","MBREquals","MBRIntersects","MBROverlaps","MBRTouches","MBRWithin","MD5","MEMBER OF","MICROSECOND","MID","MIN","MINUTE","MLineFromText","MultiLineStringFromText","MLineFromWKB","MultiLineStringFromWKB","MOD","MONTH","MONTHNAME","MPointFromText","MultiPointFromText","MPointFromWKB","MultiPointFromWKB","MPolyFromText","MultiPolygonFromText","MPolyFromWKB","MultiPolygonFromWKB","MultiLineString","MultiPoint","MultiPolygon","NAME_CONST","NOT IN","NOW","NTH_VALUE","NTILE","NULLIF","NumGeometries","NumInteriorRings","NumPoints","OCT","OCTET_LENGTH","OLD_PASSWORD","ORD","Overlaps","PASSWORD","PERCENT_RANK","PERIOD_ADD","PERIOD_DIFF","PI","Point","PointFromText","PointFromWKB","PointN","PolyFromText","PolygonFromText","PolyFromWKB","PolygonFromWKB","Polygon","POSITION","POW","POWER","PS_CURRENT_THREAD_ID","PS_THREAD_ID","PROCEDURE ANALYSE","QUARTER","QUOTE","RADIANS","RAND","RANDOM_BYTES","RANK","REGEXP_INSTR","REGEXP_LIKE","REGEXP_REPLACE","REGEXP_REPLACE","RELEASE_ALL_LOCKS","RELEASE_LOCK","REPEAT","REPLACE","REVERSE","RIGHT","ROLES_GRAPHML","ROUND","ROW_COUNT","ROW_NUMBER","RPAD","RTRIM","SCHEMA","SEC_TO_TIME","SECOND","SESSION_USER","SHA1","SHA","SHA2","SIGN","SIN","SLEEP","SOUNDEX","SOURCE_POS_WAIT","SPACE","SQRT","SRID","ST_Area","ST_AsBinary","ST_AsWKB","ST_AsGeoJSON","ST_AsText","ST_AsWKT","ST_Buffer","ST_Buffer_Strategy","ST_Centroid","ST_Collect","ST_Contains","ST_ConvexHull","ST_Crosses","ST_Difference","ST_Dimension","ST_Disjoint","ST_Distance","ST_Distance_Sphere","ST_EndPoint","ST_Envelope","ST_Equals","ST_ExteriorRing","ST_FrechetDistance","ST_GeoHash","ST_GeomCollFromText","ST_GeometryCollectionFromText","ST_GeomCollFromTxt","ST_GeomCollFromWKB","ST_GeometryCollectionFromWKB","ST_GeometryN","ST_GeometryType","ST_GeomFromGeoJSON","ST_GeomFromText","ST_GeometryFromText","ST_GeomFromWKB","ST_GeometryFromWKB","ST_HausdorffDistance","ST_InteriorRingN","ST_Intersection","ST_Intersects","ST_IsClosed","ST_IsEmpty","ST_IsSimple","ST_IsValid","ST_LatFromGeoHash","ST_Length","ST_LineFromText","ST_LineStringFromText","ST_LineFromWKB","ST_LineStringFromWKB","ST_LineInterpolatePoint","ST_LineInterpolatePoints","ST_LongFromGeoHash","ST_Longitude","ST_MakeEnvelope","ST_MLineFromText","ST_MultiLineStringFromText","ST_MLineFromWKB","ST_MultiLineStringFromWKB","ST_MPointFromText","ST_MultiPointFromText","ST_MPointFromWKB","ST_MultiPointFromWKB","ST_MPolyFromText","ST_MultiPolygonFromText","ST_MPolyFromWKB","ST_MultiPolygonFromWKB","ST_NumGeometries","ST_NumInteriorRing","ST_NumInteriorRings","ST_NumPoints","ST_Overlaps","ST_PointAtDistance","ST_PointFromGeoHash","ST_PointFromText","ST_PointFromWKB","ST_PointN","ST_PolyFromText","ST_PolygonFromText","ST_PolyFromWKB","ST_PolygonFromWKB","ST_Simplify","ST_SRID","ST_StartPoint","ST_SwapXY","ST_SymDifference","ST_Touches","ST_Transform","ST_Union","ST_Validate","ST_Within","ST_X","ST_Y","StartPoint","STATEMENT_DIGEST","STATEMENT_DIGEST_TEXT","STD","STDDEV","STDDEV_POP","STDDEV_SAMP","STR_TO_DATE","STRCMP","SUBDATE","SUBSTR","SUBSTRING","SUBSTRING_INDEX","SUBTIME","SUM","SYSDATE","SYSTEM_USER","TAN","TIME","TIME_FORMAT","TIME_TO_SEC","TIMEDIFF","TIMESTAMP","TIMESTAMPADD","TIMESTAMPDIFF","TO_BASE64","TO_DAYS","TO_SECONDS","Touches","TRIM","TRUNCATE","UCASE","UNCOMPRESS","UNCOMPRESSED_LENGTH","UNHEX","UNIX_TIMESTAMP","UpdateXML","UPPER","USER","UTC_DATE","UTC_TIME","UTC_TIMESTAMP","UUID","UUID_SHORT","UUID_TO_BIN","VALIDATE_PASSWORD_STRENGTH","VALUES","VAR_POP","VAR_SAMP","VARIANCE","VERSION","WAIT_FOR_EXECUTED_GTID_SET","WAIT_UNTIL_SQL_THREAD_AFTER_GTIDS","WEEK","WEEKDAY","WEEKOFYEAR","WEIGHT_STRING","Within","X","Y","YEAR","YEARWEEK"],builtinVariables:[],tokenizer:{root:[{include:"@comments"},{include:"@whitespace"},{include:"@numbers"},{include:"@strings"},{include:"@complexIdentifiers"},{include:"@scopes"},[/[;,.]/,"delimiter"],[/[()]/,"@brackets"],[/[\w@]+/,{cases:{"@operators":"operator","@builtinVariables":"predefined","@builtinFunctions":"predefined","@keywords":"keyword","@default":"identifier"}}],[/[<>=!%&+\-*/|~^]/,"operator"]],whitespace:[[/\s+/,"white"]],comments:[[/--+.*/,"comment"],[/#+.*/,"comment"],[/\/\*/,{token:"comment.quote",next:"@comment"}]],comment:[[/[^*/]+/,"comment"],[/\*\//,{token:"comment.quote",next:"@pop"}],[/./,"comment"]],numbers:[[/0[xX][0-9a-fA-F]*/,"number"],[/[$][+-]*\d*(\.\d*)?/,"number"],[/((\d+(\.\d*)?)|(\.\d+))([eE][\-+]?\d+)?/,"number"]],strings:[[/'/,{token:"string",next:"@string"}],[/"/,{token:"string.double",next:"@stringDouble"}]],string:[[/\\'/,"string"],[/[^']+/,"string"],[/''/,"string"],[/'/,{token:"string",next:"@pop"}]],stringDouble:[[/[^"]+/,"string.double"],[/""/,"string.double"],[/"/,{token:"string.double",next:"@pop"}]],complexIdentifiers:[[/`/,{token:"identifier.quote",next:"@quotedIdentifier"}]],quotedIdentifier:[[/[^`]+/,"identifier"],[/``/,"identifier"],[/`/,{token:"identifier.quote",next:"@pop"}]],scopes:[]}},vqe=Object.freeze(Object.defineProperty({__proto__:null,conf:bqe,language:wqe},Symbol.toStringTag,{value:"Module"})),Cqe={comments:{lineComment:"//",blockComment:["/*","*/"]},brackets:[["{","}"],["[","]"],["(",")"]],autoClosingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:'"',close:'"'},{open:"'",close:"'"}],surroundingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:'"',close:'"'},{open:"'",close:"'"}]},yqe={defaultToken:"",tokenPostfix:".objective-c",keywords:["#import","#include","#define","#else","#endif","#if","#ifdef","#ifndef","#ident","#undef","@class","@defs","@dynamic","@encode","@end","@implementation","@interface","@package","@private","@protected","@property","@protocol","@public","@selector","@synthesize","__declspec","assign","auto","BOOL","break","bycopy","byref","case","char","Class","const","copy","continue","default","do","double","else","enum","extern","FALSE","false","float","for","goto","if","in","int","id","inout","IMP","long","nil","nonatomic","NULL","oneway","out","private","public","protected","readwrite","readonly","register","return","SEL","self","short","signed","sizeof","static","struct","super","switch","typedef","TRUE","true","union","unsigned","volatile","void","while"],decpart:/\d(_?\d)*/,decimal:/0|@decpart/,tokenizer:{root:[{include:"@comments"},{include:"@whitespace"},{include:"@numbers"},{include:"@strings"},[/[,:;]/,"delimiter"],[/[{}\[\]()<>]/,"@brackets"],[/[a-zA-Z@#]\w*/,{cases:{"@keywords":"keyword","@default":"identifier"}}],[/[<>=\\+\\-\\*\\/\\^\\|\\~,]|and\\b|or\\b|not\\b]/,"operator"]],whitespace:[[/\s+/,"white"]],comments:[["\\/\\*","comment","@comment"],["\\/\\/+.*","comment"]],comment:[["\\*\\/","comment","@pop"],[".","comment"]],numbers:[[/0[xX][0-9a-fA-F]*(_?[0-9a-fA-F])*/,"number.hex"],[/@decimal((\.@decpart)?([eE][\-+]?@decpart)?)[fF]*/,{cases:{"(\\d)*":"number",$0:"number.float"}}]],strings:[[/'$/,"string.escape","@popall"],[/'/,"string.escape","@stringBody"],[/"$/,"string.escape","@popall"],[/"/,"string.escape","@dblStringBody"]],stringBody:[[/[^\\']+$/,"string","@popall"],[/[^\\']+/,"string"],[/\\./,"string"],[/'/,"string.escape","@popall"],[/\\$/,"string"]],dblStringBody:[[/[^\\"]+$/,"string","@popall"],[/[^\\"]+/,"string"],[/\\./,"string"],[/"/,"string.escape","@popall"],[/\\$/,"string"]]}},Sqe=Object.freeze(Object.defineProperty({__proto__:null,conf:Cqe,language:yqe},Symbol.toStringTag,{value:"Module"})),xqe={wordPattern:/(-?\d*\.\d\w*)|([^\`\~\!\#\%\^\&\*\(\)\-\=\+\[\{\]\}\\\|\;\:\'\"\,\.\<\>\/\?\s]+)/g,comments:{lineComment:"//",blockComment:["{","}"]},brackets:[["{","}"],["[","]"],["(",")"],["<",">"]],autoClosingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:"<",close:">"},{open:"'",close:"'"}],surroundingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:"<",close:">"},{open:"'",close:"'"}],folding:{markers:{start:new RegExp("^\\s*\\{\\$REGION(\\s\\'.*\\')?\\}"),end:new RegExp("^\\s*\\{\\$ENDREGION\\}")}}},Lqe={defaultToken:"",tokenPostfix:".pascal",ignoreCase:!0,brackets:[{open:"{",close:"}",token:"delimiter.curly"},{open:"[",close:"]",token:"delimiter.square"},{open:"(",close:")",token:"delimiter.parenthesis"},{open:"<",close:">",token:"delimiter.angle"}],keywords:["absolute","abstract","all","and_then","array","as","asm","attribute","begin","bindable","case","class","const","contains","default","div","else","end","except","exports","external","far","file","finalization","finally","forward","generic","goto","if","implements","import","in","index","inherited","initialization","interrupt","is","label","library","mod","module","name","near","not","object","of","on","only","operator","or_else","otherwise","override","package","packed","pow","private","program","protected","public","published","interface","implementation","qualified","read","record","resident","requires","resourcestring","restricted","segment","set","shl","shr","specialize","stored","strict","then","threadvar","to","try","type","unit","uses","var","view","virtual","dynamic","overload","reintroduce","with","write","xor","true","false","procedure","function","constructor","destructor","property","break","continue","exit","abort","while","do","for","raise","repeat","until"],typeKeywords:["boolean","double","byte","integer","shortint","char","longint","float","string"],operators:["=",">","<","<=",">=","<>",":",":=","and","or","+","-","*","/","@","&","^","%"],symbols:/[=><:@\^&|+\-*\/\^%]+/,tokenizer:{root:[[/[a-zA-Z_][\w]*/,{cases:{"@keywords":{token:"keyword.$0"},"@default":"identifier"}}],{include:"@whitespace"},[/[{}()\[\]]/,"@brackets"],[/[<>](?!@symbols)/,"@brackets"],[/@symbols/,{cases:{"@operators":"delimiter","@default":""}}],[/\d*\.\d+([eE][\-+]?\d+)?/,"number.float"],[/\$[0-9a-fA-F]{1,16}/,"number.hex"],[/\d+/,"number"],[/[;,.]/,"delimiter"],[/'([^'\\]|\\.)*$/,"string.invalid"],[/'/,"string","@string"],[/'[^\\']'/,"string"],[/'/,"string.invalid"],[/\#\d+/,"string"]],comment:[[/[^\*\}]+/,"comment"],[/\}/,"comment","@pop"],[/[\{]/,"comment"]],string:[[/[^\\']+/,"string"],[/\\./,"string.escape.invalid"],[/'/,{token:"string.quote",bracket:"@close",next:"@pop"}]],whitespace:[[/[ \t\r\n]+/,"white"],[/\{/,"comment","@comment"],[/\/\/.*$/,"comment"]]}},kqe=Object.freeze(Object.defineProperty({__proto__:null,conf:xqe,language:Lqe},Symbol.toStringTag,{value:"Module"})),Eqe={comments:{lineComment:"//",blockComment:["(*","*)"]},brackets:[["{","}"],["[","]"],["(",")"],["<",">"]],autoClosingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:"<",close:">"},{open:"'",close:"'"}],surroundingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:"<",close:">"},{open:"'",close:"'"}]},Iqe={defaultToken:"",tokenPostfix:".pascaligo",ignoreCase:!0,brackets:[{open:"{",close:"}",token:"delimiter.curly"},{open:"[",close:"]",token:"delimiter.square"},{open:"(",close:")",token:"delimiter.parenthesis"},{open:"<",close:">",token:"delimiter.angle"}],keywords:["begin","block","case","const","else","end","fail","for","from","function","if","is","nil","of","remove","return","skip","then","type","var","while","with","option","None","transaction"],typeKeywords:["bool","int","list","map","nat","record","string","unit","address","map","mtz","xtz"],operators:["=",">","<","<=",">=","<>",":",":=","and","mod","or","+","-","*","/","@","&","^","%"],symbols:/[=><:@\^&|+\-*\/\^%]+/,tokenizer:{root:[[/[a-zA-Z_][\w]*/,{cases:{"@keywords":{token:"keyword.$0"},"@default":"identifier"}}],{include:"@whitespace"},[/[{}()\[\]]/,"@brackets"],[/[<>](?!@symbols)/,"@brackets"],[/@symbols/,{cases:{"@operators":"delimiter","@default":""}}],[/\d*\.\d+([eE][\-+]?\d+)?/,"number.float"],[/\$[0-9a-fA-F]{1,16}/,"number.hex"],[/\d+/,"number"],[/[;,.]/,"delimiter"],[/'([^'\\]|\\.)*$/,"string.invalid"],[/'/,"string","@string"],[/'[^\\']'/,"string"],[/'/,"string.invalid"],[/\#\d+/,"string"]],comment:[[/[^\(\*]+/,"comment"],[/\*\)/,"comment","@pop"],[/\(\*/,"comment"]],string:[[/[^\\']+/,"string"],[/\\./,"string.escape.invalid"],[/'/,{token:"string.quote",bracket:"@close",next:"@pop"}]],whitespace:[[/[ \t\r\n]+/,"white"],[/\(\*/,"comment","@comment"],[/\/\/.*$/,"comment"]]}},Dqe=Object.freeze(Object.defineProperty({__proto__:null,conf:Eqe,language:Iqe},Symbol.toStringTag,{value:"Module"})),Tqe={comments:{lineComment:"#"},brackets:[["{","}"],["[","]"],["(",")"]],autoClosingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:'"',close:'"'},{open:"'",close:"'"},{open:"`",close:"`"}],surroundingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:'"',close:'"'},{open:"'",close:"'"},{open:"`",close:"`"}]},Nqe={defaultToken:"",tokenPostfix:".perl",brackets:[{token:"delimiter.bracket",open:"{",close:"}"},{token:"delimiter.parenthesis",open:"(",close:")"},{token:"delimiter.square",open:"[",close:"]"}],keywords:["__DATA__","else","lock","__END__","elsif","lt","__FILE__","eq","__LINE__","exp","ne","sub","__PACKAGE__","for","no","and","foreach","or","unless","cmp","ge","package","until","continue","gt","while","CORE","if","xor","do","le","__DIE__","__WARN__"],builtinFunctions:["-A","END","length","setpgrp","-B","endgrent","link","setpriority","-b","endhostent","listen","setprotoent","-C","endnetent","local","setpwent","-c","endprotoent","localtime","setservent","-d","endpwent","log","setsockopt","-e","endservent","lstat","shift","-f","eof","map","shmctl","-g","eval","mkdir","shmget","-k","exec","msgctl","shmread","-l","exists","msgget","shmwrite","-M","exit","msgrcv","shutdown","-O","fcntl","msgsnd","sin","-o","fileno","my","sleep","-p","flock","next","socket","-r","fork","not","socketpair","-R","format","oct","sort","-S","formline","open","splice","-s","getc","opendir","split","-T","getgrent","ord","sprintf","-t","getgrgid","our","sqrt","-u","getgrnam","pack","srand","-w","gethostbyaddr","pipe","stat","-W","gethostbyname","pop","state","-X","gethostent","pos","study","-x","getlogin","print","substr","-z","getnetbyaddr","printf","symlink","abs","getnetbyname","prototype","syscall","accept","getnetent","push","sysopen","alarm","getpeername","quotemeta","sysread","atan2","getpgrp","rand","sysseek","AUTOLOAD","getppid","read","system","BEGIN","getpriority","readdir","syswrite","bind","getprotobyname","readline","tell","binmode","getprotobynumber","readlink","telldir","bless","getprotoent","readpipe","tie","break","getpwent","recv","tied","caller","getpwnam","redo","time","chdir","getpwuid","ref","times","CHECK","getservbyname","rename","truncate","chmod","getservbyport","require","uc","chomp","getservent","reset","ucfirst","chop","getsockname","return","umask","chown","getsockopt","reverse","undef","chr","glob","rewinddir","UNITCHECK","chroot","gmtime","rindex","unlink","close","goto","rmdir","unpack","closedir","grep","say","unshift","connect","hex","scalar","untie","cos","index","seek","use","crypt","INIT","seekdir","utime","dbmclose","int","select","values","dbmopen","ioctl","semctl","vec","defined","join","semget","wait","delete","keys","semop","waitpid","DESTROY","kill","send","wantarray","die","last","setgrent","warn","dump","lc","sethostent","write","each","lcfirst","setnetent"],builtinFileHandlers:["ARGV","STDERR","STDOUT","ARGVOUT","STDIN","ENV"],builtinVariables:["$!","$^RE_TRIE_MAXBUF","$LAST_REGEXP_CODE_RESULT",'$"',"$^S","$LIST_SEPARATOR","$#","$^T","$MATCH","$$","$^TAINT","$MULTILINE_MATCHING","$%","$^UNICODE","$NR","$&","$^UTF8LOCALE","$OFMT","$'","$^V","$OFS","$(","$^W","$ORS","$)","$^WARNING_BITS","$OS_ERROR","$*","$^WIDE_SYSTEM_CALLS","$OSNAME","$+","$^X","$OUTPUT_AUTO_FLUSH","$,","$_","$OUTPUT_FIELD_SEPARATOR","$-","$`","$OUTPUT_RECORD_SEPARATOR","$.","$a","$PERL_VERSION","$/","$ACCUMULATOR","$PERLDB","$0","$ARG","$PID","$:","$ARGV","$POSTMATCH","$;","$b","$PREMATCH","$<","$BASETIME","$PROCESS_ID","$=","$CHILD_ERROR","$PROGRAM_NAME","$>","$COMPILING","$REAL_GROUP_ID","$?","$DEBUGGING","$REAL_USER_ID","$@","$EFFECTIVE_GROUP_ID","$RS","$[","$EFFECTIVE_USER_ID","$SUBSCRIPT_SEPARATOR","$\\","$EGID","$SUBSEP","$]","$ERRNO","$SYSTEM_FD_MAX","$^","$EUID","$UID","$^A","$EVAL_ERROR","$WARNING","$^C","$EXCEPTIONS_BEING_CAUGHT","$|","$^CHILD_ERROR_NATIVE","$EXECUTABLE_NAME","$~","$^D","$EXTENDED_OS_ERROR","%!","$^E","$FORMAT_FORMFEED","%^H","$^ENCODING","$FORMAT_LINE_BREAK_CHARACTERS","%ENV","$^F","$FORMAT_LINES_LEFT","%INC","$^H","$FORMAT_LINES_PER_PAGE","%OVERLOAD","$^I","$FORMAT_NAME","%SIG","$^L","$FORMAT_PAGE_NUMBER","@+","$^M","$FORMAT_TOP_NAME","@-","$^N","$GID","@_","$^O","$INPLACE_EDIT","@ARGV","$^OPEN","$INPUT_LINE_NUMBER","@INC","$^P","$INPUT_RECORD_SEPARATOR","@LAST_MATCH_START","$^R","$LAST_MATCH_END","$^RE_DEBUG_FLAGS","$LAST_PAREN_MATCH"],symbols:/[:+\-\^*$&%@=<>!?|\/~\.]/,quoteLikeOps:["qr","m","s","q","qq","qx","qw","tr","y"],escapes:/\\(?:[abfnrtv\\"']|x[0-9A-Fa-f]{1,4}|u[0-9A-Fa-f]{4}|U[0-9A-Fa-f]{8})/,tokenizer:{root:[{include:"@whitespace"},[/[a-zA-Z\-_][\w\-_]*/,{cases:{"@keywords":"keyword","@builtinFunctions":"type.identifier","@builtinFileHandlers":"variable.predefined","@quoteLikeOps":{token:"@rematch",next:"quotedConstructs"},"@default":""}}],[/[\$@%][*@#?\+\-\$!\w\\\^><~:;\.]+/,{cases:{"@builtinVariables":"variable.predefined","@default":"variable"}}],{include:"@strings"},{include:"@dblStrings"},{include:"@perldoc"},{include:"@heredoc"},[/[{}\[\]()]/,"@brackets"],[/[\/](?:(?:\[(?:\\]|[^\]])+\])|(?:\\\/|[^\]\/]))*[\/]\w*\s*(?=[).,;]|$)/,"regexp"],[/@symbols/,"operators"],{include:"@numbers"},[/[,;]/,"delimiter"]],whitespace:[[/\s+/,"white"],[/(^#!.*$)/,"metatag"],[/(^#.*$)/,"comment"]],numbers:[[/\d*\.\d+([eE][\-+]?\d+)?/,"number.float"],[/0[xX][0-9a-fA-F_]*[0-9a-fA-F]/,"number.hex"],[/\d+/,"number"]],strings:[[/'/,"string","@stringBody"]],stringBody:[[/'/,"string","@popall"],[/\\'/,"string.escape"],[/./,"string"]],dblStrings:[[/"/,"string","@dblStringBody"]],dblStringBody:[[/"/,"string","@popall"],[/@escapes/,"string.escape"],[/\\./,"string.escape.invalid"],{include:"@variables"},[/./,"string"]],quotedConstructs:[[/(q|qw|tr|y)\s*\(/,{token:"string.delim",switchTo:"@qstring.(.)"}],[/(q|qw|tr|y)\s*\[/,{token:"string.delim",switchTo:"@qstring.[.]"}],[/(q|qw|tr|y)\s*\{/,{token:"string.delim",switchTo:"@qstring.{.}"}],[/(q|qw|tr|y)\s*</,{token:"string.delim",switchTo:"@qstring.<.>"}],[/(q|qw|tr|y)#/,{token:"string.delim",switchTo:"@qstring.#.#"}],[/(q|qw|tr|y)\s*([^A-Za-z0-9#\s])/,{token:"string.delim",switchTo:"@qstring.$2.$2"}],[/(q|qw|tr|y)\s+(\w)/,{token:"string.delim",switchTo:"@qstring.$2.$2"}],[/(qr|m|s)\s*\(/,{token:"regexp.delim",switchTo:"@qregexp.(.)"}],[/(qr|m|s)\s*\[/,{token:"regexp.delim",switchTo:"@qregexp.[.]"}],[/(qr|m|s)\s*\{/,{token:"regexp.delim",switchTo:"@qregexp.{.}"}],[/(qr|m|s)\s*</,{token:"regexp.delim",switchTo:"@qregexp.<.>"}],[/(qr|m|s)#/,{token:"regexp.delim",switchTo:"@qregexp.#.#"}],[/(qr|m|s)\s*([^A-Za-z0-9_#\s])/,{token:"regexp.delim",switchTo:"@qregexp.$2.$2"}],[/(qr|m|s)\s+(\w)/,{token:"regexp.delim",switchTo:"@qregexp.$2.$2"}],[/(qq|qx)\s*\(/,{token:"string.delim",switchTo:"@qqstring.(.)"}],[/(qq|qx)\s*\[/,{token:"string.delim",switchTo:"@qqstring.[.]"}],[/(qq|qx)\s*\{/,{token:"string.delim",switchTo:"@qqstring.{.}"}],[/(qq|qx)\s*</,{token:"string.delim",switchTo:"@qqstring.<.>"}],[/(qq|qx)#/,{token:"string.delim",switchTo:"@qqstring.#.#"}],[/(qq|qx)\s*([^A-Za-z0-9#\s])/,{token:"string.delim",switchTo:"@qqstring.$2.$2"}],[/(qq|qx)\s+(\w)/,{token:"string.delim",switchTo:"@qqstring.$2.$2"}]],qstring:[[/\\./,"string.escape"],[/./,{cases:{"$#==$S3":{token:"string.delim",next:"@pop"},"$#==$S2":{token:"string.delim",next:"@push"},"@default":"string"}}]],qregexp:[{include:"@variables"},[/\\./,"regexp.escape"],[/./,{cases:{"$#==$S3":{token:"regexp.delim",next:"@regexpModifiers"},"$#==$S2":{token:"regexp.delim",next:"@push"},"@default":"regexp"}}]],regexpModifiers:[[/[msixpodualngcer]+/,{token:"regexp.modifier",next:"@popall"}]],qqstring:[{include:"@variables"},{include:"@qstring"}],heredoc:[[/<<\s*['"`]?([\w\-]+)['"`]?/,{token:"string.heredoc.delimiter",next:"@heredocBody.$1"}]],heredocBody:[[/^([\w\-]+)$/,{cases:{"$1==$S2":[{token:"string.heredoc.delimiter",next:"@popall"}],"@default":"string.heredoc"}}],[/./,"string.heredoc"]],perldoc:[[/^=\w/,"comment.doc","@perldocBody"]],perldocBody:[[/^=cut\b/,"type.identifier","@popall"],[/./,"comment.doc"]],variables:[[/\$\w+/,"variable"],[/@\w+/,"variable"],[/%\w+/,"variable"]]}},Rqe=Object.freeze(Object.defineProperty({__proto__:null,conf:Tqe,language:Nqe},Symbol.toStringTag,{value:"Module"})),Aqe={comments:{lineComment:"--",blockComment:["/*","*/"]},brackets:[["{","}"],["[","]"],["(",")"]],autoClosingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:'"',close:'"'},{open:"'",close:"'"}],surroundingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:'"',close:'"'},{open:"'",close:"'"}]},Mqe={defaultToken:"",tokenPostfix:".sql",ignoreCase:!0,brackets:[{open:"[",close:"]",token:"delimiter.square"},{open:"(",close:")",token:"delimiter.parenthesis"}],keywords:["ALL","ANALYSE","ANALYZE","AND","ANY","ARRAY","AS","ASC","ASYMMETRIC","AUTHORIZATION","BINARY","BOTH","CASE","CAST","CHECK","COLLATE","COLLATION","COLUMN","CONCURRENTLY","CONSTRAINT","CREATE","CROSS","CURRENT_CATALOG","CURRENT_DATE","CURRENT_ROLE","CURRENT_SCHEMA","CURRENT_TIME","CURRENT_TIMESTAMP","CURRENT_USER","DEFAULT","DEFERRABLE","DESC","DISTINCT","DO","ELSE","END","EXCEPT","FALSE","FETCH","FOR","FOREIGN","FREEZE","FROM","FULL","GRANT","GROUP","HAVING","ILIKE","IN","INITIALLY","INNER","INTERSECT","INTO","IS","ISNULL","JOIN","LATERAL","LEADING","LEFT","LIKE","LIMIT","LOCALTIME","LOCALTIMESTAMP","NATURAL","NOT","NOTNULL","NULL","OFFSET","ON","ONLY","OR","ORDER","OUTER","OVERLAPS","PLACING","PRIMARY","REFERENCES","RETURNING","RIGHT","SELECT","SESSION_USER","SIMILAR","SOME","SYMMETRIC","TABLE","TABLESAMPLE","THEN","TO","TRAILING","TRUE","UNION","UNIQUE","USER","USING","VARIADIC","VERBOSE","WHEN","WHERE","WINDOW","WITH"],operators:["AND","BETWEEN","IN","LIKE","NOT","OR","IS","NULL","INTERSECT","UNION","INNER","JOIN","LEFT","OUTER","RIGHT"],builtinFunctions:["abbrev","abs","acldefault","aclexplode","acos","acosd","acosh","age","any","area","array_agg","array_append","array_cat","array_dims","array_fill","array_length","array_lower","array_ndims","array_position","array_positions","array_prepend","array_remove","array_replace","array_to_json","array_to_string","array_to_tsvector","array_upper","ascii","asin","asind","asinh","atan","atan2","atan2d","atand","atanh","avg","bit","bit_and","bit_count","bit_length","bit_or","bit_xor","bool_and","bool_or","bound_box","box","brin_desummarize_range","brin_summarize_new_values","brin_summarize_range","broadcast","btrim","cardinality","cbrt","ceil","ceiling","center","char_length","character_length","chr","circle","clock_timestamp","coalesce","col_description","concat","concat_ws","convert","convert_from","convert_to","corr","cos","cosd","cosh","cot","cotd","count","covar_pop","covar_samp","cume_dist","current_catalog","current_database","current_date","current_query","current_role","current_schema","current_schemas","current_setting","current_time","current_timestamp","current_user","currval","cursor_to_xml","cursor_to_xmlschema","date_bin","date_part","date_trunc","database_to_xml","database_to_xml_and_xmlschema","database_to_xmlschema","decode","degrees","dense_rank","diagonal","diameter","div","encode","enum_first","enum_last","enum_range","every","exp","extract","factorial","family","first_value","floor","format","format_type","gcd","gen_random_uuid","generate_series","generate_subscripts","get_bit","get_byte","get_current_ts_config","gin_clean_pending_list","greatest","grouping","has_any_column_privilege","has_column_privilege","has_database_privilege","has_foreign_data_wrapper_privilege","has_function_privilege","has_language_privilege","has_schema_privilege","has_sequence_privilege","has_server_privilege","has_table_privilege","has_tablespace_privilege","has_type_privilege","height","host","hostmask","inet_client_addr","inet_client_port","inet_merge","inet_same_family","inet_server_addr","inet_server_port","initcap","isclosed","isempty","isfinite","isopen","json_agg","json_array_elements","json_array_elements_text","json_array_length","json_build_array","json_build_object","json_each","json_each_text","json_extract_path","json_extract_path_text","json_object","json_object_agg","json_object_keys","json_populate_record","json_populate_recordset","json_strip_nulls","json_to_record","json_to_recordset","json_to_tsvector","json_typeof","jsonb_agg","jsonb_array_elements","jsonb_array_elements_text","jsonb_array_length","jsonb_build_array","jsonb_build_object","jsonb_each","jsonb_each_text","jsonb_extract_path","jsonb_extract_path_text","jsonb_insert","jsonb_object","jsonb_object_agg","jsonb_object_keys","jsonb_path_exists","jsonb_path_match","jsonb_path_query","jsonb_path_query_array","jsonb_path_exists_tz","jsonb_path_query_first","jsonb_path_query_array_tz","jsonb_path_query_first_tz","jsonb_path_query_tz","jsonb_path_match_tz","jsonb_populate_record","jsonb_populate_recordset","jsonb_pretty","jsonb_set","jsonb_set_lax","jsonb_strip_nulls","jsonb_to_record","jsonb_to_recordset","jsonb_to_tsvector","jsonb_typeof","justify_days","justify_hours","justify_interval","lag","last_value","lastval","lcm","lead","least","left","length","line","ln","localtime","localtimestamp","log","log10","lower","lower_inc","lower_inf","lpad","lseg","ltrim","macaddr8_set7bit","make_date","make_interval","make_time","make_timestamp","make_timestamptz","makeaclitem","masklen","max","md5","min","min_scale","mod","mode","multirange","netmask","network","nextval","normalize","now","npoints","nth_value","ntile","nullif","num_nonnulls","num_nulls","numnode","obj_description","octet_length","overlay","parse_ident","path","pclose","percent_rank","percentile_cont","percentile_disc","pg_advisory_lock","pg_advisory_lock_shared","pg_advisory_unlock","pg_advisory_unlock_all","pg_advisory_unlock_shared","pg_advisory_xact_lock","pg_advisory_xact_lock_shared","pg_backend_pid","pg_backup_start_time","pg_blocking_pids","pg_cancel_backend","pg_client_encoding","pg_collation_actual_version","pg_collation_is_visible","pg_column_compression","pg_column_size","pg_conf_load_time","pg_control_checkpoint","pg_control_init","pg_control_recovery","pg_control_system","pg_conversion_is_visible","pg_copy_logical_replication_slot","pg_copy_physical_replication_slot","pg_create_logical_replication_slot","pg_create_physical_replication_slot","pg_create_restore_point","pg_current_logfile","pg_current_snapshot","pg_current_wal_flush_lsn","pg_current_wal_insert_lsn","pg_current_wal_lsn","pg_current_xact_id","pg_current_xact_id_if_assigned","pg_current_xlog_flush_location","pg_current_xlog_insert_location","pg_current_xlog_location","pg_database_size","pg_describe_object","pg_drop_replication_slot","pg_event_trigger_ddl_commands","pg_event_trigger_dropped_objects","pg_event_trigger_table_rewrite_oid","pg_event_trigger_table_rewrite_reason","pg_export_snapshot","pg_filenode_relation","pg_function_is_visible","pg_get_catalog_foreign_keys","pg_get_constraintdef","pg_get_expr","pg_get_function_arguments","pg_get_function_identity_arguments","pg_get_function_result","pg_get_functiondef","pg_get_indexdef","pg_get_keywords","pg_get_object_address","pg_get_owned_sequence","pg_get_ruledef","pg_get_serial_sequence","pg_get_statisticsobjdef","pg_get_triggerdef","pg_get_userbyid","pg_get_viewdef","pg_get_wal_replay_pause_state","pg_has_role","pg_identify_object","pg_identify_object_as_address","pg_import_system_collations","pg_index_column_has_property","pg_index_has_property","pg_indexam_has_property","pg_indexes_size","pg_is_in_backup","pg_is_in_recovery","pg_is_other_temp_schema","pg_is_wal_replay_paused","pg_is_xlog_replay_paused","pg_jit_available","pg_last_committed_xact","pg_last_wal_receive_lsn","pg_last_wal_replay_lsn","pg_last_xact_replay_timestamp","pg_last_xlog_receive_location","pg_last_xlog_replay_location","pg_listening_channels","pg_log_backend_memory_contexts","pg_logical_emit_message","pg_logical_slot_get_binary_changes","pg_logical_slot_get_changes","pg_logical_slot_peek_binary_changes","pg_logical_slot_peek_changes","pg_ls_archive_statusdir","pg_ls_dir","pg_ls_logdir","pg_ls_tmpdir","pg_ls_waldir","pg_mcv_list_items","pg_my_temp_schema","pg_notification_queue_usage","pg_opclass_is_visible","pg_operator_is_visible","pg_opfamily_is_visible","pg_options_to_table","pg_partition_ancestors","pg_partition_root","pg_partition_tree","pg_postmaster_start_time","pg_promote","pg_read_binary_file","pg_read_file","pg_relation_filenode","pg_relation_filepath","pg_relation_size","pg_reload_conf","pg_replication_origin_advance","pg_replication_origin_create","pg_replication_origin_drop","pg_replication_origin_oid","pg_replication_origin_progress","pg_replication_origin_session_is_setup","pg_replication_origin_session_progress","pg_replication_origin_session_reset","pg_replication_origin_session_setup","pg_replication_origin_xact_reset","pg_replication_origin_xact_setup","pg_replication_slot_advance","pg_rotate_logfile","pg_safe_snapshot_blocking_pids","pg_size_bytes","pg_size_pretty","pg_sleep","pg_sleep_for","pg_sleep_until","pg_snapshot_xip","pg_snapshot_xmax","pg_snapshot_xmin","pg_start_backup","pg_stat_file","pg_statistics_obj_is_visible","pg_stop_backup","pg_switch_wal","pg_switch_xlog","pg_table_is_visible","pg_table_size","pg_tablespace_databases","pg_tablespace_location","pg_tablespace_size","pg_terminate_backend","pg_total_relation_size","pg_trigger_depth","pg_try_advisory_lock","pg_try_advisory_lock_shared","pg_try_advisory_xact_lock","pg_try_advisory_xact_lock_shared","pg_ts_config_is_visible","pg_ts_dict_is_visible","pg_ts_parser_is_visible","pg_ts_template_is_visible","pg_type_is_visible","pg_typeof","pg_visible_in_snapshot","pg_wal_lsn_diff","pg_wal_replay_pause","pg_wal_replay_resume","pg_walfile_name","pg_walfile_name_offset","pg_xact_commit_timestamp","pg_xact_commit_timestamp_origin","pg_xact_status","pg_xlog_location_diff","pg_xlog_replay_pause","pg_xlog_replay_resume","pg_xlogfile_name","pg_xlogfile_name_offset","phraseto_tsquery","pi","plainto_tsquery","point","polygon","popen","position","power","pqserverversion","query_to_xml","query_to_xml_and_xmlschema","query_to_xmlschema","querytree","quote_ident","quote_literal","quote_nullable","radians","radius","random","range_agg","range_intersect_agg","range_merge","rank","regexp_count","regexp_instr","regexp_like","regexp_match","regexp_matches","regexp_replace","regexp_split_to_array","regexp_split_to_table","regexp_substr","regr_avgx","regr_avgy","regr_count","regr_intercept","regr_r2","regr_slope","regr_sxx","regr_sxy","regr_syy","repeat","replace","reverse","right","round","row_number","row_security_active","row_to_json","rpad","rtrim","scale","schema_to_xml","schema_to_xml_and_xmlschema","schema_to_xmlschema","session_user","set_bit","set_byte","set_config","set_masklen","setseed","setval","setweight","sha224","sha256","sha384","sha512","shobj_description","sign","sin","sind","sinh","slope","split_part","sprintf","sqrt","starts_with","statement_timestamp","stddev","stddev_pop","stddev_samp","string_agg","string_to_array","string_to_table","strip","strpos","substr","substring","sum","suppress_redundant_updates_trigger","table_to_xml","table_to_xml_and_xmlschema","table_to_xmlschema","tan","tand","tanh","text","timeofday","timezone","to_ascii","to_char","to_date","to_hex","to_json","to_number","to_regclass","to_regcollation","to_regnamespace","to_regoper","to_regoperator","to_regproc","to_regprocedure","to_regrole","to_regtype","to_timestamp","to_tsquery","to_tsvector","transaction_timestamp","translate","trim","trim_array","trim_scale","trunc","ts_debug","ts_delete","ts_filter","ts_headline","ts_lexize","ts_parse","ts_rank","ts_rank_cd","ts_rewrite","ts_stat","ts_token_type","tsquery_phrase","tsvector_to_array","tsvector_update_trigger","tsvector_update_trigger_column","txid_current","txid_current_if_assigned","txid_current_snapshot","txid_snapshot_xip","txid_snapshot_xmax","txid_snapshot_xmin","txid_status","txid_visible_in_snapshot","unistr","unnest","upper","upper_inc","upper_inf","user","var_pop","var_samp","variance","version","websearch_to_tsquery","width","width_bucket","xml_is_well_formed","xml_is_well_formed_content","xml_is_well_formed_document","xmlagg","xmlcomment","xmlconcat","xmlelement","xmlexists","xmlforest","xmlparse","xmlpi","xmlroot","xmlserialize","xpath","xpath_exists"],builtinVariables:[],pseudoColumns:[],tokenizer:{root:[{include:"@comments"},{include:"@whitespace"},{include:"@pseudoColumns"},{include:"@numbers"},{include:"@strings"},{include:"@complexIdentifiers"},{include:"@scopes"},[/[;,.]/,"delimiter"],[/[()]/,"@brackets"],[/[\w@#$]+/,{cases:{"@operators":"operator","@builtinVariables":"predefined","@builtinFunctions":"predefined","@keywords":"keyword","@default":"identifier"}}],[/[<>=!%&+\-*/|~^]/,"operator"]],whitespace:[[/\s+/,"white"]],comments:[[/--+.*/,"comment"],[/\/\*/,{token:"comment.quote",next:"@comment"}]],comment:[[/[^*/]+/,"comment"],[/\*\//,{token:"comment.quote",next:"@pop"}],[/./,"comment"]],pseudoColumns:[[/[$][A-Za-z_][\w@#$]*/,{cases:{"@pseudoColumns":"predefined","@default":"identifier"}}]],numbers:[[/0[xX][0-9a-fA-F]*/,"number"],[/[$][+-]*\d*(\.\d*)?/,"number"],[/((\d+(\.\d*)?)|(\.\d+))([eE][\-+]?\d+)?/,"number"]],strings:[[/'/,{token:"string",next:"@string"}]],string:[[/[^']+/,"string"],[/''/,"string"],[/'/,{token:"string",next:"@pop"}]],complexIdentifiers:[[/"/,{token:"identifier.quote",next:"@quotedIdentifier"}]],quotedIdentifier:[[/[^"]+/,"identifier"],[/""/,"identifier"],[/"/,{token:"identifier.quote",next:"@pop"}]],scopes:[]}},Oqe=Object.freeze(Object.defineProperty({__proto__:null,conf:Aqe,language:Mqe},Symbol.toStringTag,{value:"Module"})),Pqe={wordPattern:/(-?\d*\.\d\w*)|([^\`\~\!\@\#\%\^\&\*\(\)\-\=\+\[\{\]\}\\\|\;\:\'\"\,\.\<\>\/\?\s]+)/g,comments:{lineComment:"//",blockComment:["/*","*/"]},brackets:[["{","}"],["[","]"],["(",")"]],autoClosingPairs:[{open:"{",close:"}",notIn:["string"]},{open:"[",close:"]",notIn:["string"]},{open:"(",close:")",notIn:["string"]},{open:'"',close:'"',notIn:["string"]},{open:"'",close:"'",notIn:["string","comment"]}],folding:{markers:{start:new RegExp("^\\s*(#|//)region\\b"),end:new RegExp("^\\s*(#|//)endregion\\b")}}},Fqe={defaultToken:"",tokenPostfix:"",tokenizer:{root:[[/<\?((php)|=)?/,{token:"@rematch",switchTo:"@phpInSimpleState.root"}],[/<!DOCTYPE/,"metatag.html","@doctype"],[/<!--/,"comment.html","@comment"],[/(<)(\w+)(\/>)/,["delimiter.html","tag.html","delimiter.html"]],[/(<)(script)/,["delimiter.html",{token:"tag.html",next:"@script"}]],[/(<)(style)/,["delimiter.html",{token:"tag.html",next:"@style"}]],[/(<)([:\w]+)/,["delimiter.html",{token:"tag.html",next:"@otherTag"}]],[/(<\/)(\w+)/,["delimiter.html",{token:"tag.html",next:"@otherTag"}]],[/</,"delimiter.html"],[/[^<]+/]],doctype:[[/<\?((php)|=)?/,{token:"@rematch",switchTo:"@phpInSimpleState.comment"}],[/[^>]+/,"metatag.content.html"],[/>/,"metatag.html","@pop"]],comment:[[/<\?((php)|=)?/,{token:"@rematch",switchTo:"@phpInSimpleState.comment"}],[/-->/,"comment.html","@pop"],[/[^-]+/,"comment.content.html"],[/./,"comment.content.html"]],otherTag:[[/<\?((php)|=)?/,{token:"@rematch",switchTo:"@phpInSimpleState.otherTag"}],[/\/?>/,"delimiter.html","@pop"],[/"([^"]*)"/,"attribute.value"],[/'([^']*)'/,"attribute.value"],[/[\w\-]+/,"attribute.name"],[/=/,"delimiter"],[/[ \t\r\n]+/]],script:[[/<\?((php)|=)?/,{token:"@rematch",switchTo:"@phpInSimpleState.script"}],[/type/,"attribute.name","@scriptAfterType"],[/"([^"]*)"/,"attribute.value"],[/'([^']*)'/,"attribute.value"],[/[\w\-]+/,"attribute.name"],[/=/,"delimiter"],[/>/,{token:"delimiter.html",next:"@scriptEmbedded.text/javascript",nextEmbedded:"text/javascript"}],[/[ \t\r\n]+/],[/(<\/)(script\s*)(>)/,["delimiter.html","tag.html",{token:"delimiter.html",next:"@pop"}]]],scriptAfterType:[[/<\?((php)|=)?/,{token:"@rematch",switchTo:"@phpInSimpleState.scriptAfterType"}],[/=/,"delimiter","@scriptAfterTypeEquals"],[/>/,{token:"delimiter.html",next:"@scriptEmbedded.text/javascript",nextEmbedded:"text/javascript"}],[/[ \t\r\n]+/],[/<\/script\s*>/,{token:"@rematch",next:"@pop"}]],scriptAfterTypeEquals:[[/<\?((php)|=)?/,{token:"@rematch",switchTo:"@phpInSimpleState.scriptAfterTypeEquals"}],[/"([^"]*)"/,{token:"attribute.value",switchTo:"@scriptWithCustomType.$1"}],[/'([^']*)'/,{token:"attribute.value",switchTo:"@scriptWithCustomType.$1"}],[/>/,{token:"delimiter.html",next:"@scriptEmbedded.text/javascript",nextEmbedded:"text/javascript"}],[/[ \t\r\n]+/],[/<\/script\s*>/,{token:"@rematch",next:"@pop"}]],scriptWithCustomType:[[/<\?((php)|=)?/,{token:"@rematch",switchTo:"@phpInSimpleState.scriptWithCustomType.$S2"}],[/>/,{token:"delimiter.html",next:"@scriptEmbedded.$S2",nextEmbedded:"$S2"}],[/"([^"]*)"/,"attribute.value"],[/'([^']*)'/,"attribute.value"],[/[\w\-]+/,"attribute.name"],[/=/,"delimiter"],[/[ \t\r\n]+/],[/<\/script\s*>/,{token:"@rematch",next:"@pop"}]],scriptEmbedded:[[/<\?((php)|=)?/,{token:"@rematch",switchTo:"@phpInEmbeddedState.scriptEmbedded.$S2",nextEmbedded:"@pop"}],[/<\/script/,{token:"@rematch",next:"@pop",nextEmbedded:"@pop"}]],style:[[/<\?((php)|=)?/,{token:"@rematch",switchTo:"@phpInSimpleState.style"}],[/type/,"attribute.name","@styleAfterType"],[/"([^"]*)"/,"attribute.value"],[/'([^']*)'/,"attribute.value"],[/[\w\-]+/,"attribute.name"],[/=/,"delimiter"],[/>/,{token:"delimiter.html",next:"@styleEmbedded.text/css",nextEmbedded:"text/css"}],[/[ \t\r\n]+/],[/(<\/)(style\s*)(>)/,["delimiter.html","tag.html",{token:"delimiter.html",next:"@pop"}]]],styleAfterType:[[/<\?((php)|=)?/,{token:"@rematch",switchTo:"@phpInSimpleState.styleAfterType"}],[/=/,"delimiter","@styleAfterTypeEquals"],[/>/,{token:"delimiter.html",next:"@styleEmbedded.text/css",nextEmbedded:"text/css"}],[/[ \t\r\n]+/],[/<\/style\s*>/,{token:"@rematch",next:"@pop"}]],styleAfterTypeEquals:[[/<\?((php)|=)?/,{token:"@rematch",switchTo:"@phpInSimpleState.styleAfterTypeEquals"}],[/"([^"]*)"/,{token:"attribute.value",switchTo:"@styleWithCustomType.$1"}],[/'([^']*)'/,{token:"attribute.value",switchTo:"@styleWithCustomType.$1"}],[/>/,{token:"delimiter.html",next:"@styleEmbedded.text/css",nextEmbedded:"text/css"}],[/[ \t\r\n]+/],[/<\/style\s*>/,{token:"@rematch",next:"@pop"}]],styleWithCustomType:[[/<\?((php)|=)?/,{token:"@rematch",switchTo:"@phpInSimpleState.styleWithCustomType.$S2"}],[/>/,{token:"delimiter.html",next:"@styleEmbedded.$S2",nextEmbedded:"$S2"}],[/"([^"]*)"/,"attribute.value"],[/'([^']*)'/,"attribute.value"],[/[\w\-]+/,"attribute.name"],[/=/,"delimiter"],[/[ \t\r\n]+/],[/<\/style\s*>/,{token:"@rematch",next:"@pop"}]],styleEmbedded:[[/<\?((php)|=)?/,{token:"@rematch",switchTo:"@phpInEmbeddedState.styleEmbedded.$S2",nextEmbedded:"@pop"}],[/<\/style/,{token:"@rematch",next:"@pop",nextEmbedded:"@pop"}]],phpInSimpleState:[[/<\?((php)|=)?/,"metatag.php"],[/\?>/,{token:"metatag.php",switchTo:"@$S2.$S3"}],{include:"phpRoot"}],phpInEmbeddedState:[[/<\?((php)|=)?/,"metatag.php"],[/\?>/,{token:"metatag.php",switchTo:"@$S2.$S3",nextEmbedded:"$S3"}],{include:"phpRoot"}],phpRoot:[[/[a-zA-Z_]\w*/,{cases:{"@phpKeywords":{token:"keyword.php"},"@phpCompileTimeConstants":{token:"constant.php"},"@default":"identifier.php"}}],[/[$a-zA-Z_]\w*/,{cases:{"@phpPreDefinedVariables":{token:"variable.predefined.php"},"@default":"variable.php"}}],[/[{}]/,"delimiter.bracket.php"],[/[\[\]]/,"delimiter.array.php"],[/[()]/,"delimiter.parenthesis.php"],[/[ \t\r\n]+/],[/(#|\/\/)$/,"comment.php"],[/(#|\/\/)/,"comment.php","@phpLineComment"],[/\/\*/,"comment.php","@phpComment"],[/"/,"string.php","@phpDoubleQuoteString"],[/'/,"string.php","@phpSingleQuoteString"],[/[\+\-\*\%\&\|\^\~\!\=\<\>\/\?\;\:\.\,\@]/,"delimiter.php"],[/\d*\d+[eE]([\-+]?\d+)?/,"number.float.php"],[/\d*\.\d+([eE][\-+]?\d+)?/,"number.float.php"],[/0[xX][0-9a-fA-F']*[0-9a-fA-F]/,"number.hex.php"],[/0[0-7']*[0-7]/,"number.octal.php"],[/0[bB][0-1']*[0-1]/,"number.binary.php"],[/\d[\d']*/,"number.php"],[/\d/,"number.php"]],phpComment:[[/\*\//,"comment.php","@pop"],[/[^*]+/,"comment.php"],[/./,"comment.php"]],phpLineComment:[[/\?>/,{token:"@rematch",next:"@pop"}],[/.$/,"comment.php","@pop"],[/[^?]+$/,"comment.php","@pop"],[/[^?]+/,"comment.php"],[/./,"comment.php"]],phpDoubleQuoteString:[[/[^\\"]+/,"string.php"],[/@escapes/,"string.escape.php"],[/\\./,"string.escape.invalid.php"],[/"/,"string.php","@pop"]],phpSingleQuoteString:[[/[^\\']+/,"string.php"],[/@escapes/,"string.escape.php"],[/\\./,"string.escape.invalid.php"],[/'/,"string.php","@pop"]]},phpKeywords:["abstract","and","array","as","break","callable","case","catch","cfunction","class","clone","const","continue","declare","default","do","else","elseif","enddeclare","endfor","endforeach","endif","endswitch","endwhile","extends","false","final","for","foreach","function","global","goto","if","implements","interface","instanceof","insteadof","namespace","new","null","object","old_function","or","private","protected","public","resource","static","switch","throw","trait","try","true","use","var","while","xor","die","echo","empty","exit","eval","include","include_once","isset","list","require","require_once","return","print","unset","yield","__construct"],phpCompileTimeConstants:["__CLASS__","__DIR__","__FILE__","__LINE__","__NAMESPACE__","__METHOD__","__FUNCTION__","__TRAIT__"],phpPreDefinedVariables:["$GLOBALS","$_SERVER","$_GET","$_POST","$_FILES","$_REQUEST","$_SESSION","$_ENV","$_COOKIE","$php_errormsg","$HTTP_RAW_POST_DATA","$http_response_header","$argc","$argv"],escapes:/\\(?:[abfnrtv\\"']|x[0-9A-Fa-f]{1,4}|u[0-9A-Fa-f]{4}|U[0-9A-Fa-f]{8})/},Bqe=Object.freeze(Object.defineProperty({__proto__:null,conf:Pqe,language:Fqe},Symbol.toStringTag,{value:"Module"})),Wqe={comments:{lineComment:"#"},brackets:[["[","]"],["<",">"],["(",")"]],autoClosingPairs:[{open:"[",close:"]"},{open:"<",close:">"},{open:"(",close:")"}],surroundingPairs:[{open:"[",close:"]"},{open:"<",close:">"},{open:"(",close:")"}]},Hqe={defaultToken:"",tokenPostfix:".pla",brackets:[{open:"[",close:"]",token:"delimiter.square"},{open:"<",close:">",token:"delimiter.angle"},{open:"(",close:")",token:"delimiter.parenthesis"}],keywords:[".i",".o",".mv",".ilb",".ob",".label",".type",".phase",".pair",".symbolic",".symbolic-output",".kiss",".p",".e",".end"],comment:/#.*$/,identifier:/[a-zA-Z]+[a-zA-Z0-9_\-]*/,plaContent:/[01\-~\|]+/,tokenizer:{root:[{include:"@whitespace"},[/@comment/,"comment"],[/\.([a-zA-Z_\-]+)/,{cases:{"@eos":{token:"keyword.$1"},"@keywords":{cases:{".type":{token:"keyword.$1",next:"@type"},"@default":{token:"keyword.$1",next:"@keywordArg"}}},"@default":{token:"keyword.$1"}}}],[/@identifier/,"identifier"],[/@plaContent/,"string"]],whitespace:[[/[ \t\r\n]+/,""]],type:[{include:"@whitespace"},[/\w+/,{token:"type",next:"@pop"}]],keywordArg:[[/[ \t\r\n]+/,{cases:{"@eos":{token:"",next:"@pop"},"@default":""}}],[/@comment/,"comment","@pop"],[/[<>()\[\]]/,{cases:{"@eos":{token:"@brackets",next:"@pop"},"@default":"@brackets"}}],[/\-?\d+/,{cases:{"@eos":{token:"number",next:"@pop"},"@default":"number"}}],[/@identifier/,{cases:{"@eos":{token:"identifier",next:"@pop"},"@default":"identifier"}}],[/[;=]/,{cases:{"@eos":{token:"delimiter",next:"@pop"},"@default":"delimiter"}}]]}},Vqe=Object.freeze(Object.defineProperty({__proto__:null,conf:Wqe,language:Hqe},Symbol.toStringTag,{value:"Module"})),zqe={comments:{lineComment:"//",blockComment:["(*","*)"]},brackets:[["{","}"],["[","]"],["(",")"],["<",">"]],autoClosingPairs:[{open:'"',close:'"',notIn:["string","comment"]},{open:"{",close:"}",notIn:["string","comment"]},{open:"[",close:"]",notIn:["string","comment"]},{open:"(",close:")",notIn:["string","comment"]}]},$qe={tokenPostfix:".pats",defaultToken:"invalid",keywords:["abstype","abst0ype","absprop","absview","absvtype","absviewtype","absvt0ype","absviewt0ype","as","and","assume","begin","classdec","datasort","datatype","dataprop","dataview","datavtype","dataviewtype","do","end","extern","extype","extvar","exception","fn","fnx","fun","prfn","prfun","praxi","castfn","if","then","else","ifcase","in","infix","infixl","infixr","prefix","postfix","implmnt","implement","primplmnt","primplement","import","let","local","macdef","macrodef","nonfix","symelim","symintr","overload","of","op","rec","sif","scase","sortdef","sta","stacst","stadef","static","staload","dynload","try","tkindef","typedef","propdef","viewdef","vtypedef","viewtypedef","prval","var","prvar","when","where","with","withtype","withprop","withview","withvtype","withviewtype"],keywords_dlr:["$delay","$ldelay","$arrpsz","$arrptrsize","$d2ctype","$effmask","$effmask_ntm","$effmask_exn","$effmask_ref","$effmask_wrt","$effmask_all","$extern","$extkind","$extype","$extype_struct","$extval","$extfcall","$extmcall","$literal","$myfilename","$mylocation","$myfunction","$lst","$lst_t","$lst_vt","$list","$list_t","$list_vt","$rec","$rec_t","$rec_vt","$record","$record_t","$record_vt","$tup","$tup_t","$tup_vt","$tuple","$tuple_t","$tuple_vt","$break","$continue","$raise","$showtype","$vcopyenv_v","$vcopyenv_vt","$tempenver","$solver_assert","$solver_verify"],keywords_srp:["#if","#ifdef","#ifndef","#then","#elif","#elifdef","#elifndef","#else","#endif","#error","#prerr","#print","#assert","#undef","#define","#include","#require","#pragma","#codegen2","#codegen3"],irregular_keyword_list:["val+","val-","val","case+","case-","case","addr@","addr","fold@","free@","fix@","fix","lam@","lam","llam@","llam","viewt@ype+","viewt@ype-","viewt@ype","viewtype+","viewtype-","viewtype","view+","view-","view@","view","type+","type-","type","vtype+","vtype-","vtype","vt@ype+","vt@ype-","vt@ype","viewt@ype+","viewt@ype-","viewt@ype","viewtype+","viewtype-","viewtype","prop+","prop-","prop","type+","type-","type","t@ype","t@ype+","t@ype-","abst@ype","abstype","absviewt@ype","absvt@ype","for*","for","while*","while"],keywords_types:["bool","double","byte","int","short","char","void","unit","long","float","string","strptr"],keywords_effects:["0","fun","clo","prf","funclo","cloptr","cloref","ref","ntm","1"],operators:["@","!","|","`",":","$",".","=","#","~","..","...","=>","=<>","=/=>","=>>","=/=>>","<",">","><",".<",">.",".<>.","->","-<>"],brackets:[{open:",(",close:")",token:"delimiter.parenthesis"},{open:"`(",close:")",token:"delimiter.parenthesis"},{open:"%(",close:")",token:"delimiter.parenthesis"},{open:"'(",close:")",token:"delimiter.parenthesis"},{open:"'{",close:"}",token:"delimiter.parenthesis"},{open:"@(",close:")",token:"delimiter.parenthesis"},{open:"@{",close:"}",token:"delimiter.brace"},{open:"@[",close:"]",token:"delimiter.square"},{open:"#[",close:"]",token:"delimiter.square"},{open:"{",close:"}",token:"delimiter.curly"},{open:"[",close:"]",token:"delimiter.square"},{open:"(",close:")",token:"delimiter.parenthesis"},{open:"<",close:">",token:"delimiter.angle"}],symbols:/[=><!~?:&|+\-*\/\^%]+/,IDENTFST:/[a-zA-Z_]/,IDENTRST:/[a-zA-Z0-9_'$]/,symbolic:/[%&+-./:=@~`^|*!$#?<>]/,digit:/[0-9]/,digitseq0:/@digit*/,xdigit:/[0-9A-Za-z]/,xdigitseq0:/@xdigit*/,INTSP:/[lLuU]/,FLOATSP:/[fFlL]/,fexponent:/[eE][+-]?[0-9]+/,fexponent_bin:/[pP][+-]?[0-9]+/,deciexp:/\.[0-9]*@fexponent?/,hexiexp:/\.[0-9a-zA-Z]*@fexponent_bin?/,irregular_keywords:/val[+-]?|case[+-]?|addr\@?|fold\@|free\@|fix\@?|lam\@?|llam\@?|prop[+-]?|type[+-]?|view[+-@]?|viewt@?ype[+-]?|t@?ype[+-]?|v(iew)?t@?ype[+-]?|abst@?ype|absv(iew)?t@?ype|for\*?|while\*?/,ESCHAR:/[ntvbrfa\\\?'"\(\[\{]/,start:"root",tokenizer:{root:[{regex:/[ \t\r\n]+/,action:{token:""}},{regex:/\(\*\)/,action:{token:"invalid"}},{regex:/\(\*/,action:{token:"comment",next:"lexing_COMMENT_block_ml"}},{regex:/\(/,action:"@brackets"},{regex:/\)/,action:"@brackets"},{regex:/\[/,action:"@brackets"},{regex:/\]/,action:"@brackets"},{regex:/\{/,action:"@brackets"},{regex:/\}/,action:"@brackets"},{regex:/,\(/,action:"@brackets"},{regex:/,/,action:{token:"delimiter.comma"}},{regex:/;/,action:{token:"delimiter.semicolon"}},{regex:/@\(/,action:"@brackets"},{regex:/@\[/,action:"@brackets"},{regex:/@\{/,action:"@brackets"},{regex:/:</,action:{token:"keyword",next:"@lexing_EFFECT_commaseq0"}},{regex:/\.@symbolic+/,action:{token:"identifier.sym"}},{regex:/\.@digit*@fexponent@FLOATSP*/,action:{token:"number.float"}},{regex:/\.@digit+/,action:{token:"number.float"}},{regex:/\$@IDENTFST@IDENTRST*/,action:{cases:{"@keywords_dlr":{token:"keyword.dlr"},"@default":{token:"namespace"}}}},{regex:/\#@IDENTFST@IDENTRST*/,action:{cases:{"@keywords_srp":{token:"keyword.srp"},"@default":{token:"identifier"}}}},{regex:/%\(/,action:{token:"delimiter.parenthesis"}},{regex:/^%{(#|\^|\$)?/,action:{token:"keyword",next:"@lexing_EXTCODE",nextEmbedded:"text/javascript"}},{regex:/^%}/,action:{token:"keyword"}},{regex:/'\(/,action:{token:"delimiter.parenthesis"}},{regex:/'\[/,action:{token:"delimiter.bracket"}},{regex:/'\{/,action:{token:"delimiter.brace"}},[/(')(\\@ESCHAR|\\[xX]@xdigit+|\\@digit+)(')/,["string","string.escape","string"]],[/'[^\\']'/,"string"],[/"/,"string.quote","@lexing_DQUOTE"],{regex:/`\(/,action:"@brackets"},{regex:/\\/,action:{token:"punctuation"}},{regex:/@irregular_keywords(?!@IDENTRST)/,action:{token:"keyword"}},{regex:/@IDENTFST@IDENTRST*[<!\[]?/,action:{cases:{"@keywords":{token:"keyword"},"@keywords_types":{token:"type"},"@default":{token:"identifier"}}}},{regex:/\/\/\/\//,action:{token:"comment",next:"@lexing_COMMENT_rest"}},{regex:/\/\/.*$/,action:{token:"comment"}},{regex:/\/\*/,action:{token:"comment",next:"@lexing_COMMENT_block_c"}},{regex:/-<|=</,action:{token:"keyword",next:"@lexing_EFFECT_commaseq0"}},{regex:/@symbolic+/,action:{cases:{"@operators":"keyword","@default":"operator"}}},{regex:/0[xX]@xdigit+(@hexiexp|@fexponent_bin)@FLOATSP*/,action:{token:"number.float"}},{regex:/0[xX]@xdigit+@INTSP*/,action:{token:"number.hex"}},{regex:/0[0-7]+(?![0-9])@INTSP*/,action:{token:"number.octal"}},{regex:/@digit+(@fexponent|@deciexp)@FLOATSP*/,action:{token:"number.float"}},{regex:/@digit@digitseq0@INTSP*/,action:{token:"number.decimal"}},{regex:/@digit+@INTSP*/,action:{token:"number"}}],lexing_COMMENT_block_ml:[[/[^\(\*]+/,"comment"],[/\(\*/,"comment","@push"],[/\(\*/,"comment.invalid"],[/\*\)/,"comment","@pop"],[/\*/,"comment"]],lexing_COMMENT_block_c:[[/[^\/*]+/,"comment"],[/\*\//,"comment","@pop"],[/[\/*]/,"comment"]],lexing_COMMENT_rest:[[/$/,"comment","@pop"],[/.*/,"comment"]],lexing_EFFECT_commaseq0:[{regex:/@IDENTFST@IDENTRST+|@digit+/,action:{cases:{"@keywords_effects":{token:"type.effect"},"@default":{token:"identifier"}}}},{regex:/,/,action:{token:"punctuation"}},{regex:/>/,action:{token:"@rematch",next:"@pop"}}],lexing_EXTCODE:[{regex:/^%}/,action:{token:"@rematch",next:"@pop",nextEmbedded:"@pop"}},{regex:/[^%]+/,action:""}],lexing_DQUOTE:[{regex:/"/,action:{token:"string.quote",next:"@pop"}},{regex:/(\{\$)(@IDENTFST@IDENTRST*)(\})/,action:[{token:"string.escape"},{token:"identifier"},{token:"string.escape"}]},{regex:/\\$/,action:{token:"string.escape"}},{regex:/\\(@ESCHAR|[xX]@xdigit+|@digit+)/,action:{token:"string.escape"}},{regex:/[^\\"]+/,action:{token:"string"}}]}},Uqe=Object.freeze(Object.defineProperty({__proto__:null,conf:zqe,language:$qe},Symbol.toStringTag,{value:"Module"})),qqe={comments:{lineComment:"//",blockComment:["/*","*/"]},brackets:[["[","]"],["(",")"],["{","}"]],autoClosingPairs:[{open:'"',close:'"',notIn:["string","comment","identifier"]},{open:"[",close:"]",notIn:["string","comment","identifier"]},{open:"(",close:")",notIn:["string","comment","identifier"]},{open:"{",close:"}",notIn:["string","comment","identifier"]}]},jqe={defaultToken:"",tokenPostfix:".pq",ignoreCase:!1,brackets:[{open:"[",close:"]",token:"delimiter.square"},{open:"{",close:"}",token:"delimiter.brackets"},{open:"(",close:")",token:"delimiter.parenthesis"}],operatorKeywords:["and","not","or"],keywords:["as","each","else","error","false","if","in","is","let","meta","otherwise","section","shared","then","true","try","type"],constructors:["#binary","#date","#datetime","#datetimezone","#duration","#table","#time"],constants:["#infinity","#nan","#sections","#shared"],typeKeywords:["action","any","anynonnull","none","null","logical","number","time","date","datetime","datetimezone","duration","text","binary","list","record","table","function"],builtinFunctions:["Access.Database","Action.Return","Action.Sequence","Action.Try","ActiveDirectory.Domains","AdoDotNet.DataSource","AdoDotNet.Query","AdobeAnalytics.Cubes","AnalysisServices.Database","AnalysisServices.Databases","AzureStorage.BlobContents","AzureStorage.Blobs","AzureStorage.Tables","Binary.Buffer","Binary.Combine","Binary.Compress","Binary.Decompress","Binary.End","Binary.From","Binary.FromList","Binary.FromText","Binary.InferContentType","Binary.Length","Binary.ToList","Binary.ToText","BinaryFormat.7BitEncodedSignedInteger","BinaryFormat.7BitEncodedUnsignedInteger","BinaryFormat.Binary","BinaryFormat.Byte","BinaryFormat.ByteOrder","BinaryFormat.Choice","BinaryFormat.Decimal","BinaryFormat.Double","BinaryFormat.Group","BinaryFormat.Length","BinaryFormat.List","BinaryFormat.Null","BinaryFormat.Record","BinaryFormat.SignedInteger16","BinaryFormat.SignedInteger32","BinaryFormat.SignedInteger64","BinaryFormat.Single","BinaryFormat.Text","BinaryFormat.Transform","BinaryFormat.UnsignedInteger16","BinaryFormat.UnsignedInteger32","BinaryFormat.UnsignedInteger64","Byte.From","Character.FromNumber","Character.ToNumber","Combiner.CombineTextByDelimiter","Combiner.CombineTextByEachDelimiter","Combiner.CombineTextByLengths","Combiner.CombineTextByPositions","Combiner.CombineTextByRanges","Comparer.Equals","Comparer.FromCulture","Comparer.Ordinal","Comparer.OrdinalIgnoreCase","Csv.Document","Cube.AddAndExpandDimensionColumn","Cube.AddMeasureColumn","Cube.ApplyParameter","Cube.AttributeMemberId","Cube.AttributeMemberProperty","Cube.CollapseAndRemoveColumns","Cube.Dimensions","Cube.DisplayFolders","Cube.Measures","Cube.Parameters","Cube.Properties","Cube.PropertyKey","Cube.ReplaceDimensions","Cube.Transform","Currency.From","DB2.Database","Date.AddDays","Date.AddMonths","Date.AddQuarters","Date.AddWeeks","Date.AddYears","Date.Day","Date.DayOfWeek","Date.DayOfWeekName","Date.DayOfYear","Date.DaysInMonth","Date.EndOfDay","Date.EndOfMonth","Date.EndOfQuarter","Date.EndOfWeek","Date.EndOfYear","Date.From","Date.FromText","Date.IsInCurrentDay","Date.IsInCurrentMonth","Date.IsInCurrentQuarter","Date.IsInCurrentWeek","Date.IsInCurrentYear","Date.IsInNextDay","Date.IsInNextMonth","Date.IsInNextNDays","Date.IsInNextNMonths","Date.IsInNextNQuarters","Date.IsInNextNWeeks","Date.IsInNextNYears","Date.IsInNextQuarter","Date.IsInNextWeek","Date.IsInNextYear","Date.IsInPreviousDay","Date.IsInPreviousMonth","Date.IsInPreviousNDays","Date.IsInPreviousNMonths","Date.IsInPreviousNQuarters","Date.IsInPreviousNWeeks","Date.IsInPreviousNYears","Date.IsInPreviousQuarter","Date.IsInPreviousWeek","Date.IsInPreviousYear","Date.IsInYearToDate","Date.IsLeapYear","Date.Month","Date.MonthName","Date.QuarterOfYear","Date.StartOfDay","Date.StartOfMonth","Date.StartOfQuarter","Date.StartOfWeek","Date.StartOfYear","Date.ToRecord","Date.ToText","Date.WeekOfMonth","Date.WeekOfYear","Date.Year","DateTime.AddZone","DateTime.Date","DateTime.FixedLocalNow","DateTime.From","DateTime.FromFileTime","DateTime.FromText","DateTime.IsInCurrentHour","DateTime.IsInCurrentMinute","DateTime.IsInCurrentSecond","DateTime.IsInNextHour","DateTime.IsInNextMinute","DateTime.IsInNextNHours","DateTime.IsInNextNMinutes","DateTime.IsInNextNSeconds","DateTime.IsInNextSecond","DateTime.IsInPreviousHour","DateTime.IsInPreviousMinute","DateTime.IsInPreviousNHours","DateTime.IsInPreviousNMinutes","DateTime.IsInPreviousNSeconds","DateTime.IsInPreviousSecond","DateTime.LocalNow","DateTime.Time","DateTime.ToRecord","DateTime.ToText","DateTimeZone.FixedLocalNow","DateTimeZone.FixedUtcNow","DateTimeZone.From","DateTimeZone.FromFileTime","DateTimeZone.FromText","DateTimeZone.LocalNow","DateTimeZone.RemoveZone","DateTimeZone.SwitchZone","DateTimeZone.ToLocal","DateTimeZone.ToRecord","DateTimeZone.ToText","DateTimeZone.ToUtc","DateTimeZone.UtcNow","DateTimeZone.ZoneHours","DateTimeZone.ZoneMinutes","Decimal.From","Diagnostics.ActivityId","Diagnostics.Trace","DirectQueryCapabilities.From","Double.From","Duration.Days","Duration.From","Duration.FromText","Duration.Hours","Duration.Minutes","Duration.Seconds","Duration.ToRecord","Duration.ToText","Duration.TotalDays","Duration.TotalHours","Duration.TotalMinutes","Duration.TotalSeconds","Embedded.Value","Error.Record","Excel.CurrentWorkbook","Excel.Workbook","Exchange.Contents","Expression.Constant","Expression.Evaluate","Expression.Identifier","Facebook.Graph","File.Contents","Folder.Contents","Folder.Files","Function.From","Function.Invoke","Function.InvokeAfter","Function.IsDataSource","GoogleAnalytics.Accounts","Guid.From","HdInsight.Containers","HdInsight.Contents","HdInsight.Files","Hdfs.Contents","Hdfs.Files","Informix.Database","Int16.From","Int32.From","Int64.From","Int8.From","ItemExpression.From","Json.Document","Json.FromValue","Lines.FromBinary","Lines.FromText","Lines.ToBinary","Lines.ToText","List.Accumulate","List.AllTrue","List.Alternate","List.AnyTrue","List.Average","List.Buffer","List.Combine","List.Contains","List.ContainsAll","List.ContainsAny","List.Count","List.Covariance","List.DateTimeZones","List.DateTimes","List.Dates","List.Difference","List.Distinct","List.Durations","List.FindText","List.First","List.FirstN","List.Generate","List.InsertRange","List.Intersect","List.IsDistinct","List.IsEmpty","List.Last","List.LastN","List.MatchesAll","List.MatchesAny","List.Max","List.MaxN","List.Median","List.Min","List.MinN","List.Mode","List.Modes","List.NonNullCount","List.Numbers","List.PositionOf","List.PositionOfAny","List.Positions","List.Product","List.Random","List.Range","List.RemoveFirstN","List.RemoveItems","List.RemoveLastN","List.RemoveMatchingItems","List.RemoveNulls","List.RemoveRange","List.Repeat","List.ReplaceMatchingItems","List.ReplaceRange","List.ReplaceValue","List.Reverse","List.Select","List.Single","List.SingleOrDefault","List.Skip","List.Sort","List.StandardDeviation","List.Sum","List.Times","List.Transform","List.TransformMany","List.Union","List.Zip","Logical.From","Logical.FromText","Logical.ToText","MQ.Queue","MySQL.Database","Number.Abs","Number.Acos","Number.Asin","Number.Atan","Number.Atan2","Number.BitwiseAnd","Number.BitwiseNot","Number.BitwiseOr","Number.BitwiseShiftLeft","Number.BitwiseShiftRight","Number.BitwiseXor","Number.Combinations","Number.Cos","Number.Cosh","Number.Exp","Number.Factorial","Number.From","Number.FromText","Number.IntegerDivide","Number.IsEven","Number.IsNaN","Number.IsOdd","Number.Ln","Number.Log","Number.Log10","Number.Mod","Number.Permutations","Number.Power","Number.Random","Number.RandomBetween","Number.Round","Number.RoundAwayFromZero","Number.RoundDown","Number.RoundTowardZero","Number.RoundUp","Number.Sign","Number.Sin","Number.Sinh","Number.Sqrt","Number.Tan","Number.Tanh","Number.ToText","OData.Feed","Odbc.DataSource","Odbc.Query","OleDb.DataSource","OleDb.Query","Oracle.Database","Percentage.From","PostgreSQL.Database","RData.FromBinary","Record.AddField","Record.Combine","Record.Field","Record.FieldCount","Record.FieldNames","Record.FieldOrDefault","Record.FieldValues","Record.FromList","Record.FromTable","Record.HasFields","Record.RemoveFields","Record.RenameFields","Record.ReorderFields","Record.SelectFields","Record.ToList","Record.ToTable","Record.TransformFields","Replacer.ReplaceText","Replacer.ReplaceValue","RowExpression.Column","RowExpression.From","Salesforce.Data","Salesforce.Reports","SapBusinessWarehouse.Cubes","SapHana.Database","SharePoint.Contents","SharePoint.Files","SharePoint.Tables","Single.From","Soda.Feed","Splitter.SplitByNothing","Splitter.SplitTextByAnyDelimiter","Splitter.SplitTextByDelimiter","Splitter.SplitTextByEachDelimiter","Splitter.SplitTextByLengths","Splitter.SplitTextByPositions","Splitter.SplitTextByRanges","Splitter.SplitTextByRepeatedLengths","Splitter.SplitTextByWhitespace","Sql.Database","Sql.Databases","SqlExpression.SchemaFrom","SqlExpression.ToExpression","Sybase.Database","Table.AddColumn","Table.AddIndexColumn","Table.AddJoinColumn","Table.AddKey","Table.AggregateTableColumn","Table.AlternateRows","Table.Buffer","Table.Column","Table.ColumnCount","Table.ColumnNames","Table.ColumnsOfType","Table.Combine","Table.CombineColumns","Table.Contains","Table.ContainsAll","Table.ContainsAny","Table.DemoteHeaders","Table.Distinct","Table.DuplicateColumn","Table.ExpandListColumn","Table.ExpandRecordColumn","Table.ExpandTableColumn","Table.FillDown","Table.FillUp","Table.FilterWithDataTable","Table.FindText","Table.First","Table.FirstN","Table.FirstValue","Table.FromColumns","Table.FromList","Table.FromPartitions","Table.FromRecords","Table.FromRows","Table.FromValue","Table.Group","Table.HasColumns","Table.InsertRows","Table.IsDistinct","Table.IsEmpty","Table.Join","Table.Keys","Table.Last","Table.LastN","Table.MatchesAllRows","Table.MatchesAnyRows","Table.Max","Table.MaxN","Table.Min","Table.MinN","Table.NestedJoin","Table.Partition","Table.PartitionValues","Table.Pivot","Table.PositionOf","Table.PositionOfAny","Table.PrefixColumns","Table.Profile","Table.PromoteHeaders","Table.Range","Table.RemoveColumns","Table.RemoveFirstN","Table.RemoveLastN","Table.RemoveMatchingRows","Table.RemoveRows","Table.RemoveRowsWithErrors","Table.RenameColumns","Table.ReorderColumns","Table.Repeat","Table.ReplaceErrorValues","Table.ReplaceKeys","Table.ReplaceMatchingRows","Table.ReplaceRelationshipIdentity","Table.ReplaceRows","Table.ReplaceValue","Table.ReverseRows","Table.RowCount","Table.Schema","Table.SelectColumns","Table.SelectRows","Table.SelectRowsWithErrors","Table.SingleRow","Table.Skip","Table.Sort","Table.SplitColumn","Table.ToColumns","Table.ToList","Table.ToRecords","Table.ToRows","Table.TransformColumnNames","Table.TransformColumnTypes","Table.TransformColumns","Table.TransformRows","Table.Transpose","Table.Unpivot","Table.UnpivotOtherColumns","Table.View","Table.ViewFunction","TableAction.DeleteRows","TableAction.InsertRows","TableAction.UpdateRows","Tables.GetRelationships","Teradata.Database","Text.AfterDelimiter","Text.At","Text.BeforeDelimiter","Text.BetweenDelimiters","Text.Clean","Text.Combine","Text.Contains","Text.End","Text.EndsWith","Text.Format","Text.From","Text.FromBinary","Text.Insert","Text.Length","Text.Lower","Text.Middle","Text.NewGuid","Text.PadEnd","Text.PadStart","Text.PositionOf","Text.PositionOfAny","Text.Proper","Text.Range","Text.Remove","Text.RemoveRange","Text.Repeat","Text.Replace","Text.ReplaceRange","Text.Select","Text.Split","Text.SplitAny","Text.Start","Text.StartsWith","Text.ToBinary","Text.ToList","Text.Trim","Text.TrimEnd","Text.TrimStart","Text.Upper","Time.EndOfHour","Time.From","Time.FromText","Time.Hour","Time.Minute","Time.Second","Time.StartOfHour","Time.ToRecord","Time.ToText","Type.AddTableKey","Type.ClosedRecord","Type.Facets","Type.ForFunction","Type.ForRecord","Type.FunctionParameters","Type.FunctionRequiredParameters","Type.FunctionReturn","Type.Is","Type.IsNullable","Type.IsOpenRecord","Type.ListItem","Type.NonNullable","Type.OpenRecord","Type.RecordFields","Type.ReplaceFacets","Type.ReplaceTableKeys","Type.TableColumn","Type.TableKeys","Type.TableRow","Type.TableSchema","Type.Union","Uri.BuildQueryString","Uri.Combine","Uri.EscapeDataString","Uri.Parts","Value.Add","Value.As","Value.Compare","Value.Divide","Value.Equals","Value.Firewall","Value.FromText","Value.Is","Value.Metadata","Value.Multiply","Value.NativeQuery","Value.NullableEquals","Value.RemoveMetadata","Value.ReplaceMetadata","Value.ReplaceType","Value.Subtract","Value.Type","ValueAction.NativeStatement","ValueAction.Replace","Variable.Value","Web.Contents","Web.Page","WebAction.Request","Xml.Document","Xml.Tables"],builtinConstants:["BinaryEncoding.Base64","BinaryEncoding.Hex","BinaryOccurrence.Optional","BinaryOccurrence.Repeating","BinaryOccurrence.Required","ByteOrder.BigEndian","ByteOrder.LittleEndian","Compression.Deflate","Compression.GZip","CsvStyle.QuoteAfterDelimiter","CsvStyle.QuoteAlways","Culture.Current","Day.Friday","Day.Monday","Day.Saturday","Day.Sunday","Day.Thursday","Day.Tuesday","Day.Wednesday","ExtraValues.Error","ExtraValues.Ignore","ExtraValues.List","GroupKind.Global","GroupKind.Local","JoinAlgorithm.Dynamic","JoinAlgorithm.LeftHash","JoinAlgorithm.LeftIndex","JoinAlgorithm.PairwiseHash","JoinAlgorithm.RightHash","JoinAlgorithm.RightIndex","JoinAlgorithm.SortMerge","JoinKind.FullOuter","JoinKind.Inner","JoinKind.LeftAnti","JoinKind.LeftOuter","JoinKind.RightAnti","JoinKind.RightOuter","JoinSide.Left","JoinSide.Right","MissingField.Error","MissingField.Ignore","MissingField.UseNull","Number.E","Number.Epsilon","Number.NaN","Number.NegativeInfinity","Number.PI","Number.PositiveInfinity","Occurrence.All","Occurrence.First","Occurrence.Last","Occurrence.Optional","Occurrence.Repeating","Occurrence.Required","Order.Ascending","Order.Descending","Precision.Decimal","Precision.Double","QuoteStyle.Csv","QuoteStyle.None","RelativePosition.FromEnd","RelativePosition.FromStart","RoundingMode.AwayFromZero","RoundingMode.Down","RoundingMode.ToEven","RoundingMode.TowardZero","RoundingMode.Up","SapHanaDistribution.All","SapHanaDistribution.Connection","SapHanaDistribution.Off","SapHanaDistribution.Statement","SapHanaRangeOperator.Equals","SapHanaRangeOperator.GreaterThan","SapHanaRangeOperator.GreaterThanOrEquals","SapHanaRangeOperator.LessThan","SapHanaRangeOperator.LessThanOrEquals","SapHanaRangeOperator.NotEquals","TextEncoding.Ascii","TextEncoding.BigEndianUnicode","TextEncoding.Unicode","TextEncoding.Utf16","TextEncoding.Utf8","TextEncoding.Windows","TraceLevel.Critical","TraceLevel.Error","TraceLevel.Information","TraceLevel.Verbose","TraceLevel.Warning","WebMethod.Delete","WebMethod.Get","WebMethod.Head","WebMethod.Patch","WebMethod.Post","WebMethod.Put"],builtinTypes:["Action.Type","Any.Type","Binary.Type","BinaryEncoding.Type","BinaryOccurrence.Type","Byte.Type","ByteOrder.Type","Character.Type","Compression.Type","CsvStyle.Type","Currency.Type","Date.Type","DateTime.Type","DateTimeZone.Type","Day.Type","Decimal.Type","Double.Type","Duration.Type","ExtraValues.Type","Function.Type","GroupKind.Type","Guid.Type","Int16.Type","Int32.Type","Int64.Type","Int8.Type","JoinAlgorithm.Type","JoinKind.Type","JoinSide.Type","List.Type","Logical.Type","MissingField.Type","None.Type","Null.Type","Number.Type","Occurrence.Type","Order.Type","Password.Type","Percentage.Type","Precision.Type","QuoteStyle.Type","Record.Type","RelativePosition.Type","RoundingMode.Type","SapHanaDistribution.Type","SapHanaRangeOperator.Type","Single.Type","Table.Type","Text.Type","TextEncoding.Type","Time.Type","TraceLevel.Type","Type.Type","Uri.Type","WebMethod.Type"],tokenizer:{root:[[/#"[\w \.]+"/,"identifier.quote"],[/\d*\.\d+([eE][\-+]?\d+)?/,"number.float"],[/0[xX][0-9a-fA-F]+/,"number.hex"],[/\d+([eE][\-+]?\d+)?/,"number"],[/(#?[a-z]+)\b/,{cases:{"@typeKeywords":"type","@keywords":"keyword","@constants":"constant","@constructors":"constructor","@operatorKeywords":"operators","@default":"identifier"}}],[/\b([A-Z][a-zA-Z0-9]+\.Type)\b/,{cases:{"@builtinTypes":"type","@default":"identifier"}}],[/\b([A-Z][a-zA-Z0-9]+\.[A-Z][a-zA-Z0-9]+)\b/,{cases:{"@builtinFunctions":"keyword.function","@builtinConstants":"constant","@default":"identifier"}}],[/\b([a-zA-Z_][\w\.]*)\b/,"identifier"],{include:"@whitespace"},{include:"@comments"},{include:"@strings"},[/[{}()\[\]]/,"@brackets"],[/([=\+<>\-\*&@\?\/!])|([<>]=)|(<>)|(=>)|(\.\.\.)|(\.\.)/,"operators"],[/[,;]/,"delimiter"]],whitespace:[[/\s+/,"white"]],comments:[["\\/\\*","comment","@comment"],["\\/\\/+.*","comment"]],comment:[["\\*\\/","comment","@pop"],[".","comment"]],strings:[['"',"string","@string"]],string:[['""',"string.escape"],['"',"string","@pop"],[".","string"]]}},Gqe=Object.freeze(Object.defineProperty({__proto__:null,conf:qqe,language:jqe},Symbol.toStringTag,{value:"Module"})),Kqe={wordPattern:/(-?\d*\.\d\w*)|([^\`\~\!\@\#%\^\&\*\(\)\=\+\[\{\]\}\\\|\;\:\'\"\,\.\<\>\/\?\s]+)/g,comments:{lineComment:"#",blockComment:["<#","#>"]},brackets:[["{","}"],["[","]"],["(",")"]],autoClosingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:'"',close:'"',notIn:["string"]},{open:"'",close:"'",notIn:["string","comment"]}],surroundingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:'"',close:'"'},{open:"'",close:"'"}],folding:{markers:{start:new RegExp("^\\s*#region\\b"),end:new RegExp("^\\s*#endregion\\b")}}},Yqe={defaultToken:"",ignoreCase:!0,tokenPostfix:".ps1",brackets:[{token:"delimiter.curly",open:"{",close:"}"},{token:"delimiter.square",open:"[",close:"]"},{token:"delimiter.parenthesis",open:"(",close:")"}],keywords:["begin","break","catch","class","continue","data","define","do","dynamicparam","else","elseif","end","exit","filter","finally","for","foreach","from","function","if","in","param","process","return","switch","throw","trap","try","until","using","var","while","workflow","parallel","sequence","inlinescript","configuration"],helpKeywords:/SYNOPSIS|DESCRIPTION|PARAMETER|EXAMPLE|INPUTS|OUTPUTS|NOTES|LINK|COMPONENT|ROLE|FUNCTIONALITY|FORWARDHELPTARGETNAME|FORWARDHELPCATEGORY|REMOTEHELPRUNSPACE|EXTERNALHELP/,symbols:/[=><!~?&%|+\-*\/\^;\.,]+/,escapes:/`(?:[abfnrtv\\"'$]|x[0-9A-Fa-f]{1,4}|u[0-9A-Fa-f]{4}|U[0-9A-Fa-f]{8})/,tokenizer:{root:[[/[a-zA-Z_][\w-]*/,{cases:{"@keywords":{token:"keyword.$0"},"@default":""}}],[/[ \t\r\n]+/,""],[/^:\w*/,"metatag"],[/\$(\{((global|local|private|script|using):)?[\w]+\}|((global|local|private|script|using):)?[\w]+)/,"variable"],[/<#/,"comment","@comment"],[/#.*$/,"comment"],[/[{}()\[\]]/,"@brackets"],[/@symbols/,"delimiter"],[/\d*\.\d+([eE][\-+]?\d+)?/,"number.float"],[/0[xX][0-9a-fA-F_]*[0-9a-fA-F]/,"number.hex"],[/\d+?/,"number"],[/[;,.]/,"delimiter"],[/\@"/,"string",'@herestring."'],[/\@'/,"string","@herestring.'"],[/"/,{cases:{"@eos":"string","@default":{token:"string",next:'@string."'}}}],[/'/,{cases:{"@eos":"string","@default":{token:"string",next:"@string.'"}}}]],string:[[/[^"'\$`]+/,{cases:{"@eos":{token:"string",next:"@popall"},"@default":"string"}}],[/@escapes/,{cases:{"@eos":{token:"string.escape",next:"@popall"},"@default":"string.escape"}}],[/`./,{cases:{"@eos":{token:"string.escape.invalid",next:"@popall"},"@default":"string.escape.invalid"}}],[/\$[\w]+$/,{cases:{'$S2=="':{token:"variable",next:"@popall"},"@default":{token:"string",next:"@popall"}}}],[/\$[\w]+/,{cases:{'$S2=="':"variable","@default":"string"}}],[/["']/,{cases:{"$#==$S2":{token:"string",next:"@pop"},"@default":{cases:{"@eos":{token:"string",next:"@popall"},"@default":"string"}}}}]],herestring:[[/^\s*(["'])@/,{cases:{"$1==$S2":{token:"string",next:"@pop"},"@default":"string"}}],[/[^\$`]+/,"string"],[/@escapes/,"string.escape"],[/`./,"string.escape.invalid"],[/\$[\w]+/,{cases:{'$S2=="':"variable","@default":"string"}}]],comment:[[/[^#\.]+/,"comment"],[/#>/,"comment","@pop"],[/(\.)(@helpKeywords)(?!\w)/,{token:"comment.keyword.$2"}],[/[\.#]/,"comment"]]}},Zqe=Object.freeze(Object.defineProperty({__proto__:null,conf:Kqe,language:Yqe},Symbol.toStringTag,{value:"Module"})),Xqe=["true","false"],Qqe={comments:{lineComment:"//",blockComment:["/*","*/"]},brackets:[["{","}"],["[","]"],["(",")"],["<",">"]],surroundingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:"<",close:">"},{open:'"',close:'"'},{open:"'",close:"'"}],autoClosingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:"<",close:">"},{open:'"',close:'"',notIn:["string"]},{open:"'",close:"'",notIn:["string"]}],autoCloseBefore:`.,=}])>' 
	`,indentationRules:{increaseIndentPattern:new RegExp("^((?!\\/\\/).)*(\\{[^}\"'`]*|\\([^)\"'`]*|\\[[^\\]\"'`]*)$"),decreaseIndentPattern:new RegExp("^((?!.*?\\/\\*).*\\*/)?\\s*[\\}\\]].*$")}},Jqe={defaultToken:"",tokenPostfix:".proto",brackets:[{open:"{",close:"}",token:"delimiter.curly"},{open:"[",close:"]",token:"delimiter.square"},{open:"(",close:")",token:"delimiter.parenthesis"},{open:"<",close:">",token:"delimiter.angle"}],symbols:/[=><!~?:&|+\-*/^%]+/,keywords:["syntax","import","weak","public","package","option","repeated","oneof","map","reserved","to","max","enum","message","service","rpc","stream","returns","package","optional","true","false"],builtinTypes:["double","float","int32","int64","uint32","uint64","sint32","sint64","fixed32","fixed64","sfixed32","sfixed64","bool","string","bytes"],operators:["=","+","-"],namedLiterals:Xqe,escapes:"\\\\(u{[0-9A-Fa-f]+}|n|r|t|\\\\|'|\\${)",identifier:/[a-zA-Z]\w*/,fullIdentifier:/@identifier(?:\s*\.\s*@identifier)*/,optionName:/(?:@identifier|\(\s*@fullIdentifier\s*\))(?:\s*\.\s*@identifier)*/,messageName:/@identifier/,enumName:/@identifier/,messageType:/\.?\s*(?:@identifier\s*\.\s*)*@messageName/,enumType:/\.?\s*(?:@identifier\s*\.\s*)*@enumName/,floatLit:/[0-9]+\s*\.\s*[0-9]*(?:@exponent)?|[0-9]+@exponent|\.[0-9]+(?:@exponent)?/,exponent:/[eE]\s*[+-]?\s*[0-9]+/,boolLit:/true\b|false\b/,decimalLit:/[1-9][0-9]*/,octalLit:/0[0-7]*/,hexLit:/0[xX][0-9a-fA-F]+/,type:/double|float|int32|int64|uint32|uint64|sint32|sint64|fixed32|fixed64|sfixed32|sfixed64|bool|string|bytes|@messageType|@enumType/,keyType:/int32|int64|uint32|uint64|sint32|sint64|fixed32|fixed64|sfixed32|sfixed64|bool|string/,tokenizer:{root:[{include:"@whitespace"},[/syntax/,"keyword"],[/=/,"operators"],[/;/,"delimiter"],[/(")(proto3)(")/,["string.quote","string",{token:"string.quote",switchTo:"@topLevel.proto3"}]],[/(")(proto2)(")/,["string.quote","string",{token:"string.quote",switchTo:"@topLevel.proto2"}]],[/.*?/,{token:"",switchTo:"@topLevel.proto2"}]],topLevel:[{include:"@whitespace"},{include:"@constant"},[/=/,"operators"],[/[;.]/,"delimiter"],[/@fullIdentifier/,{cases:{option:{token:"keyword",next:"@option.$S2"},enum:{token:"keyword",next:"@enumDecl.$S2"},message:{token:"keyword",next:"@messageDecl.$S2"},service:{token:"keyword",next:"@serviceDecl.$S2"},extend:{cases:{"$S2==proto2":{token:"keyword",next:"@extendDecl.$S2"}}},"@keywords":"keyword","@default":"identifier"}}]],enumDecl:[{include:"@whitespace"},[/@identifier/,"type.identifier"],[/{/,{token:"@brackets",bracket:"@open",switchTo:"@enumBody.$S2"}]],enumBody:[{include:"@whitespace"},{include:"@constant"},[/=/,"operators"],[/;/,"delimiter"],[/option\b/,"keyword","@option.$S2"],[/@identifier/,"identifier"],[/\[/,{token:"@brackets",bracket:"@open",next:"@options.$S2"}],[/}/,{token:"@brackets",bracket:"@close",next:"@pop"}]],messageDecl:[{include:"@whitespace"},[/@identifier/,"type.identifier"],[/{/,{token:"@brackets",bracket:"@open",switchTo:"@messageBody.$S2"}]],messageBody:[{include:"@whitespace"},{include:"@constant"},[/=/,"operators"],[/;/,"delimiter"],["(map)(s*)(<)",["keyword","white",{token:"@brackets",bracket:"@open",next:"@map.$S2"}]],[/@identifier/,{cases:{option:{token:"keyword",next:"@option.$S2"},enum:{token:"keyword",next:"@enumDecl.$S2"},message:{token:"keyword",next:"@messageDecl.$S2"},oneof:{token:"keyword",next:"@oneofDecl.$S2"},extensions:{cases:{"$S2==proto2":{token:"keyword",next:"@reserved.$S2"}}},reserved:{token:"keyword",next:"@reserved.$S2"},"(?:repeated|optional)":{token:"keyword",next:"@field.$S2"},required:{cases:{"$S2==proto2":{token:"keyword",next:"@field.$S2"}}},"$S2==proto3":{token:"@rematch",next:"@field.$S2"}}}],[/\[/,{token:"@brackets",bracket:"@open",next:"@options.$S2"}],[/}/,{token:"@brackets",bracket:"@close",next:"@pop"}]],extendDecl:[{include:"@whitespace"},[/@identifier/,"type.identifier"],[/{/,{token:"@brackets",bracket:"@open",switchTo:"@extendBody.$S2"}]],extendBody:[{include:"@whitespace"},{include:"@constant"},[/;/,"delimiter"],[/(?:repeated|optional|required)/,"keyword","@field.$S2"],[/\[/,{token:"@brackets",bracket:"@open",next:"@options.$S2"}],[/}/,{token:"@brackets",bracket:"@close",next:"@pop"}]],options:[{include:"@whitespace"},{include:"@constant"},[/;/,"delimiter"],[/@optionName/,"annotation"],[/[()]/,"annotation.brackets"],[/=/,"operator"],[/\]/,{token:"@brackets",bracket:"@close",next:"@pop"}]],option:[{include:"@whitespace"},[/@optionName/,"annotation"],[/[()]/,"annotation.brackets"],[/=/,"operator","@pop"]],oneofDecl:[{include:"@whitespace"},[/@identifier/,"identifier"],[/{/,{token:"@brackets",bracket:"@open",switchTo:"@oneofBody.$S2"}]],oneofBody:[{include:"@whitespace"},{include:"@constant"},[/;/,"delimiter"],[/(@identifier)(\s*)(=)/,["identifier","white","delimiter"]],[/@fullIdentifier|\./,{cases:{"@builtinTypes":"keyword","@default":"type.identifier"}}],[/\[/,{token:"@brackets",bracket:"@open",next:"@options.$S2"}],[/}/,{token:"@brackets",bracket:"@close",next:"@pop"}]],reserved:[{include:"@whitespace"},[/,/,"delimiter"],[/;/,"delimiter","@pop"],{include:"@constant"},[/to\b|max\b/,"keyword"]],map:[{include:"@whitespace"},[/@fullIdentifier|\./,{cases:{"@builtinTypes":"keyword","@default":"type.identifier"}}],[/,/,"delimiter"],[/>/,{token:"@brackets",bracket:"@close",switchTo:"identifier"}]],field:[{include:"@whitespace"},["group",{cases:{"$S2==proto2":{token:"keyword",switchTo:"@groupDecl.$S2"}}}],[/(@identifier)(\s*)(=)/,["identifier","white",{token:"delimiter",next:"@pop"}]],[/@fullIdentifier|\./,{cases:{"@builtinTypes":"keyword","@default":"type.identifier"}}]],groupDecl:[{include:"@whitespace"},[/@identifier/,"identifier"],["=","operator"],[/{/,{token:"@brackets",bracket:"@open",switchTo:"@messageBody.$S2"}],{include:"@constant"}],type:[{include:"@whitespace"},[/@identifier/,"type.identifier","@pop"],[/./,"delimiter"]],identifier:[{include:"@whitespace"},[/@identifier/,"identifier","@pop"]],serviceDecl:[{include:"@whitespace"},[/@identifier/,"identifier"],[/{/,{token:"@brackets",bracket:"@open",switchTo:"@serviceBody.$S2"}]],serviceBody:[{include:"@whitespace"},{include:"@constant"},[/;/,"delimiter"],[/option\b/,"keyword","@option.$S2"],[/rpc\b/,"keyword","@rpc.$S2"],[/\[/,{token:"@brackets",bracket:"@open",next:"@options.$S2"}],[/}/,{token:"@brackets",bracket:"@close",next:"@pop"}]],rpc:[{include:"@whitespace"},[/@identifier/,"identifier"],[/\(/,{token:"@brackets",bracket:"@open",switchTo:"@request.$S2"}],[/{/,{token:"@brackets",bracket:"@open",next:"@methodOptions.$S2"}],[/;/,"delimiter","@pop"]],request:[{include:"@whitespace"},[/@messageType/,{cases:{stream:{token:"keyword",next:"@type.$S2"},"@default":"type.identifier"}}],[/\)/,{token:"@brackets",bracket:"@close",switchTo:"@returns.$S2"}]],returns:[{include:"@whitespace"},[/returns\b/,"keyword"],[/\(/,{token:"@brackets",bracket:"@open",switchTo:"@response.$S2"}]],response:[{include:"@whitespace"},[/@messageType/,{cases:{stream:{token:"keyword",next:"@type.$S2"},"@default":"type.identifier"}}],[/\)/,{token:"@brackets",bracket:"@close",switchTo:"@rpc.$S2"}]],methodOptions:[{include:"@whitespace"},{include:"@constant"},[/;/,"delimiter"],["option","keyword"],[/@optionName/,"annotation"],[/[()]/,"annotation.brackets"],[/=/,"operator"],[/}/,{token:"@brackets",bracket:"@close",next:"@pop"}]],comment:[[/[^\/*]+/,"comment"],[/\/\*/,"comment","@push"],["\\*/","comment","@pop"],[/[\/*]/,"comment"]],string:[[/[^\\"]+/,"string"],[/@escapes/,"string.escape"],[/\\./,"string.escape.invalid"],[/"/,{token:"string.quote",bracket:"@close",next:"@pop"}]],stringSingle:[[/[^\\']+/,"string"],[/@escapes/,"string.escape"],[/\\./,"string.escape.invalid"],[/'/,{token:"string.quote",bracket:"@close",next:"@pop"}]],constant:[["@boolLit","keyword.constant"],["@hexLit","number.hex"],["@octalLit","number.octal"],["@decimalLit","number"],["@floatLit","number.float"],[/("([^"\\]|\\.)*|'([^'\\]|\\.)*)$/,"string.invalid"],[/"/,{token:"string.quote",bracket:"@open",next:"@string"}],[/'/,{token:"string.quote",bracket:"@open",next:"@stringSingle"}],[/{/,{token:"@brackets",bracket:"@open",next:"@prototext"}],[/identifier/,"identifier"]],whitespace:[[/[ \t\r\n]+/,"white"],[/\/\*/,"comment","@comment"],[/\/\/.*$/,"comment"]],prototext:[{include:"@whitespace"},{include:"@constant"},[/@identifier/,"identifier"],[/[:;]/,"delimiter"],[/}/,{token:"@brackets",bracket:"@close",next:"@pop"}]]}},eje=Object.freeze(Object.defineProperty({__proto__:null,conf:Qqe,language:Jqe},Symbol.toStringTag,{value:"Module"})),tje={comments:{lineComment:"//"},brackets:[["{","}"],["[","]"],["(",")"]],autoClosingPairs:[{open:'"',close:'"',notIn:["string","comment"]},{open:"'",close:"'",notIn:["string","comment"]},{open:"{",close:"}",notIn:["string","comment"]},{open:"[",close:"]",notIn:["string","comment"]},{open:"(",close:")",notIn:["string","comment"]}],folding:{offSide:!0}},ije={defaultToken:"",tokenPostfix:".pug",ignoreCase:!0,brackets:[{token:"delimiter.curly",open:"{",close:"}"},{token:"delimiter.array",open:"[",close:"]"},{token:"delimiter.parenthesis",open:"(",close:")"}],keywords:["append","block","case","default","doctype","each","else","extends","for","if","in","include","mixin","typeof","unless","var","when"],tags:["a","abbr","acronym","address","area","article","aside","audio","b","base","basefont","bdi","bdo","blockquote","body","br","button","canvas","caption","center","cite","code","col","colgroup","command","datalist","dd","del","details","dfn","div","dl","dt","em","embed","fieldset","figcaption","figure","font","footer","form","frame","frameset","h1","h2","h3","h4","h5","h6","head","header","hgroup","hr","html","i","iframe","img","input","ins","keygen","kbd","label","li","link","map","mark","menu","meta","meter","nav","noframes","noscript","object","ol","optgroup","option","output","p","param","pre","progress","q","rp","rt","ruby","s","samp","script","section","select","small","source","span","strike","strong","style","sub","summary","sup","table","tbody","td","textarea","tfoot","th","thead","time","title","tr","tracks","tt","u","ul","video","wbr"],symbols:/[\+\-\*\%\&\|\!\=\/\.\,\:]+/,escapes:/\\(?:[abfnrtv\\"']|x[0-9A-Fa-f]{1,4}|u[0-9A-Fa-f]{4}|U[0-9A-Fa-f]{8})/,tokenizer:{root:[[/^(\s*)([a-zA-Z_-][\w-]*)/,{cases:{"$2@tags":{cases:{"@eos":["","tag"],"@default":["",{token:"tag",next:"@tag.$1"}]}},"$2@keywords":["",{token:"keyword.$2"}],"@default":["",""]}}],[/^(\s*)(#[a-zA-Z_-][\w-]*)/,{cases:{"@eos":["","tag.id"],"@default":["",{token:"tag.id",next:"@tag.$1"}]}}],[/^(\s*)(\.[a-zA-Z_-][\w-]*)/,{cases:{"@eos":["","tag.class"],"@default":["",{token:"tag.class",next:"@tag.$1"}]}}],[/^(\s*)(\|.*)$/,""],{include:"@whitespace"},[/[a-zA-Z_$][\w$]*/,{cases:{"@keywords":{token:"keyword.$0"},"@default":""}}],[/[{}()\[\]]/,"@brackets"],[/@symbols/,"delimiter"],[/\d+\.\d+([eE][\-+]?\d+)?/,"number.float"],[/\d+/,"number"],[/"/,"string",'@string."'],[/'/,"string","@string.'"]],tag:[[/(\.)(\s*$)/,[{token:"delimiter",next:"@blockText.$S2."},""]],[/\s+/,{token:"",next:"@simpleText"}],[/#[a-zA-Z_-][\w-]*/,{cases:{"@eos":{token:"tag.id",next:"@pop"},"@default":"tag.id"}}],[/\.[a-zA-Z_-][\w-]*/,{cases:{"@eos":{token:"tag.class",next:"@pop"},"@default":"tag.class"}}],[/\(/,{token:"delimiter.parenthesis",next:"@attributeList"}]],simpleText:[[/[^#]+$/,{token:"",next:"@popall"}],[/[^#]+/,{token:""}],[/(#{)([^}]*)(})/,{cases:{"@eos":["interpolation.delimiter","interpolation",{token:"interpolation.delimiter",next:"@popall"}],"@default":["interpolation.delimiter","interpolation","interpolation.delimiter"]}}],[/#$/,{token:"",next:"@popall"}],[/#/,""]],attributeList:[[/\s+/,""],[/(\w+)(\s*=\s*)("|')/,["attribute.name","delimiter",{token:"attribute.value",next:"@value.$3"}]],[/\w+/,"attribute.name"],[/,/,{cases:{"@eos":{token:"attribute.delimiter",next:"@popall"},"@default":"attribute.delimiter"}}],[/\)$/,{token:"delimiter.parenthesis",next:"@popall"}],[/\)/,{token:"delimiter.parenthesis",next:"@pop"}]],whitespace:[[/^(\s*)(\/\/.*)$/,{token:"comment",next:"@blockText.$1.comment"}],[/[ \t\r\n]+/,""],[/<!--/,{token:"comment",next:"@comment"}]],blockText:[[/^\s+.*$/,{cases:{"($S2\\s+.*$)":{token:"$S3"},"@default":{token:"@rematch",next:"@popall"}}}],[/./,{token:"@rematch",next:"@popall"}]],comment:[[/[^<\-]+/,"comment.content"],[/-->/,{token:"comment",next:"@pop"}],[/<!--/,"comment.content.invalid"],[/[<\-]/,"comment.content"]],string:[[/[^\\"'#]+/,{cases:{"@eos":{token:"string",next:"@popall"},"@default":"string"}}],[/@escapes/,{cases:{"@eos":{token:"string.escape",next:"@popall"},"@default":"string.escape"}}],[/\\./,{cases:{"@eos":{token:"string.escape.invalid",next:"@popall"},"@default":"string.escape.invalid"}}],[/(#{)([^}]*)(})/,["interpolation.delimiter","interpolation","interpolation.delimiter"]],[/#/,"string"],[/["']/,{cases:{"$#==$S2":{token:"string",next:"@pop"},"@default":{token:"string"}}}]],value:[[/[^\\"']+/,{cases:{"@eos":{token:"attribute.value",next:"@popall"},"@default":"attribute.value"}}],[/\\./,{cases:{"@eos":{token:"attribute.value",next:"@popall"},"@default":"attribute.value"}}],[/["']/,{cases:{"$#==$S2":{token:"attribute.value",next:"@pop"},"@default":{token:"attribute.value"}}}]]}},nje=Object.freeze(Object.defineProperty({__proto__:null,conf:tje,language:ije},Symbol.toStringTag,{value:"Module"})),sje={comments:{lineComment:"#",blockComment:["'''","'''"]},brackets:[["{","}"],["[","]"],["(",")"]],autoClosingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:'"',close:'"',notIn:["string"]},{open:"'",close:"'",notIn:["string","comment"]}],surroundingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:'"',close:'"'},{open:"'",close:"'"}],onEnterRules:[{beforeText:new RegExp("^\\s*(?:def|class|for|if|elif|else|while|try|with|finally|except|async|match|case).*?:\\s*$"),action:{indentAction:H.IndentAction.Indent}}],folding:{offSide:!0,markers:{start:new RegExp("^\\s*#region\\b"),end:new RegExp("^\\s*#endregion\\b")}}},oje={defaultToken:"",tokenPostfix:".python",keywords:["False","None","True","_","and","as","assert","async","await","break","case","class","continue","def","del","elif","else","except","exec","finally","for","from","global","if","import","in","is","lambda","match","nonlocal","not","or","pass","print","raise","return","try","type","while","with","yield","int","float","long","complex","hex","abs","all","any","apply","basestring","bin","bool","buffer","bytearray","callable","chr","classmethod","cmp","coerce","compile","complex","delattr","dict","dir","divmod","enumerate","eval","execfile","file","filter","format","frozenset","getattr","globals","hasattr","hash","help","id","input","intern","isinstance","issubclass","iter","len","locals","list","map","max","memoryview","min","next","object","oct","open","ord","pow","print","property","reversed","range","raw_input","reduce","reload","repr","reversed","round","self","set","setattr","slice","sorted","staticmethod","str","sum","super","tuple","type","unichr","unicode","vars","xrange","zip","__dict__","__methods__","__members__","__class__","__bases__","__name__","__mro__","__subclasses__","__init__","__import__"],brackets:[{open:"{",close:"}",token:"delimiter.curly"},{open:"[",close:"]",token:"delimiter.bracket"},{open:"(",close:")",token:"delimiter.parenthesis"}],tokenizer:{root:[{include:"@whitespace"},{include:"@numbers"},{include:"@strings"},[/[,:;]/,"delimiter"],[/[{}\[\]()]/,"@brackets"],[/@[a-zA-Z_]\w*/,"tag"],[/[a-zA-Z_]\w*/,{cases:{"@keywords":"keyword","@default":"identifier"}}]],whitespace:[[/\s+/,"white"],[/(^#.*$)/,"comment"],[/'''/,"string","@endDocString"],[/"""/,"string","@endDblDocString"]],endDocString:[[/[^']+/,"string"],[/\\'/,"string"],[/'''/,"string","@popall"],[/'/,"string"]],endDblDocString:[[/[^"]+/,"string"],[/\\"/,"string"],[/"""/,"string","@popall"],[/"/,"string"]],numbers:[[/-?0x([abcdef]|[ABCDEF]|\d)+[lL]?/,"number.hex"],[/-?(\d*\.)?\d+([eE][+\-]?\d+)?[jJ]?[lL]?/,"number"]],strings:[[/'$/,"string.escape","@popall"],[/f'{1,3}/,"string.escape","@fStringBody"],[/'/,"string.escape","@stringBody"],[/"$/,"string.escape","@popall"],[/f"{1,3}/,"string.escape","@fDblStringBody"],[/"/,"string.escape","@dblStringBody"]],fStringBody:[[/[^\\'\{\}]+$/,"string","@popall"],[/[^\\'\{\}]+/,"string"],[/\{[^\}':!=]+/,"identifier","@fStringDetail"],[/\\./,"string"],[/'/,"string.escape","@popall"],[/\\$/,"string"]],stringBody:[[/[^\\']+$/,"string","@popall"],[/[^\\']+/,"string"],[/\\./,"string"],[/'/,"string.escape","@popall"],[/\\$/,"string"]],fDblStringBody:[[/[^\\"\{\}]+$/,"string","@popall"],[/[^\\"\{\}]+/,"string"],[/\{[^\}':!=]+/,"identifier","@fStringDetail"],[/\\./,"string"],[/"/,"string.escape","@popall"],[/\\$/,"string"]],dblStringBody:[[/[^\\"]+$/,"string","@popall"],[/[^\\"]+/,"string"],[/\\./,"string"],[/"/,"string.escape","@popall"],[/\\$/,"string"]],fStringDetail:[[/[:][^}]+/,"string"],[/[!][ars]/,"string"],[/=/,"string"],[/\}/,"identifier","@pop"]]}},rje=Object.freeze(Object.defineProperty({__proto__:null,conf:sje,language:oje},Symbol.toStringTag,{value:"Module"})),aje={comments:{lineComment:"//"},brackets:[["{","}"],["[","]"],["(",")"]],autoClosingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:'"',close:'"',notIn:["string","comment"]}],surroundingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:'"',close:'"'}]},lje={keywords:["namespace","open","import","export","as","operation","function","body","adjoint","newtype","struct","controlled","if","elif","else","repeat","until","fixup","for","in","while","return","fail","within","apply","Adjoint","Controlled","Adj","Ctl","is","self","auto","distribute","invert","intrinsic","let","set","w/","new","not","and","or","use","borrow","using","borrowing","mutable","internal"],typeKeywords:["Unit","Int","BigInt","Double","Bool","String","Qubit","Result","Pauli","Range"],invalidKeywords:["abstract","base","bool","break","byte","case","catch","char","checked","class","const","continue","decimal","default","delegate","do","double","enum","event","explicit","extern","finally","fixed","float","foreach","goto","implicit","int","interface","lock","long","null","object","operator","out","override","params","private","protected","public","readonly","ref","sbyte","sealed","short","sizeof","stackalloc","static","string","switch","this","throw","try","typeof","unit","ulong","unchecked","unsafe","ushort","virtual","void","volatile"],constants:["true","false","PauliI","PauliX","PauliY","PauliZ","One","Zero"],builtin:["X","Y","Z","H","HY","S","T","SWAP","CNOT","CCNOT","MultiX","R","RFrac","Rx","Ry","Rz","R1","R1Frac","Exp","ExpFrac","Measure","M","MultiM","Message","Length","Assert","AssertProb","AssertEqual"],operators:["and=","<-","->","*","*=","@","!","^","^=",":","::",".","..","==","...","=","=>",">",">=","<","<=","-","-=","!=","or=","%","%=","|","+","+=","?","/","/=","&&&","&&&=","^^^","^^^=",">>>",">>>=","<<<","<<<=","|||","|||=","~~~","_","w/","w/="],namespaceFollows:["namespace","open"],importsFollows:["import"],symbols:/[=><!~?:&|+\-*\/\^%@._]+/,escapes:/\\[\s\S]/,tokenizer:{root:[[/[a-zA-Z_$][\w$]*/,{cases:{"@namespaceFollows":{token:"keyword.$0",next:"@namespace"},"@importsFollows":{token:"keyword.$0",next:"@imports"},"@typeKeywords":"type","@keywords":"keyword","@constants":"constant","@builtin":"keyword","@invalidKeywords":"invalid","@default":"identifier"}}],{include:"@whitespace"},[/[{}()\[\]]/,"@brackets"],[/@symbols/,{cases:{"@operators":"operator","@default":""}}],[/\d*\.\d+([eE][\-+]?\d+)?/,"number.float"],[/\d+/,"number"],[/[;,.]/,"delimiter"],[/"/,{token:"string.quote",bracket:"@open",next:"@string"}]],string:[[/[^\\"]+/,"string"],[/@escapes/,"string.escape"],[/"/,{token:"string.quote",bracket:"@close",next:"@pop"}]],namespace:[{include:"@whitespace"},[/[A-Za-z]\w*/,"namespace"],[/[\.]/,"delimiter"],["","","@pop"]],imports:[{include:"@whitespace"},[/[A-Za-z]\w*(?=\.)/,"namespace"],[/[A-Za-z]\w*/,"identifier"],[/\*/,"wildcard"],[/[\.,]/,"delimiter"],["","","@pop"]],whitespace:[[/[ \t\r\n]+/,"white"],[/(\/\/).*/,"comment"]]}},cje=Object.freeze(Object.defineProperty({__proto__:null,conf:aje,language:lje},Symbol.toStringTag,{value:"Module"})),dje={comments:{lineComment:"#"},brackets:[["{","}"],["[","]"],["(",")"]],autoClosingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:'"',close:'"'}],surroundingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:'"',close:'"'}]},hje={defaultToken:"",tokenPostfix:".r",roxygen:["@alias","@aliases","@assignee","@author","@backref","@callGraph","@callGraphDepth","@callGraphPrimitives","@concept","@describeIn","@description","@details","@docType","@encoding","@evalNamespace","@evalRd","@example","@examples","@export","@exportClass","@exportMethod","@exportPattern","@family","@field","@formals","@format","@import","@importClassesFrom","@importFrom","@importMethodsFrom","@include","@inherit","@inheritDotParams","@inheritParams","@inheritSection","@keywords","@md","@method","@name","@noMd","@noRd","@note","@param","@rawNamespace","@rawRd","@rdname","@references","@return","@S3method","@section","@seealso","@setClass","@slot","@source","@template","@templateVar","@title","@TODO","@usage","@useDynLib"],constants:["NULL","FALSE","TRUE","NA","Inf","NaN","NA_integer_","NA_real_","NA_complex_","NA_character_","T","F","LETTERS","letters","month.abb","month.name","pi","R.version.string"],keywords:["break","next","return","if","else","for","in","repeat","while","array","category","character","complex","double","function","integer","list","logical","matrix","numeric","vector","data.frame","factor","library","require","attach","detach","source"],special:["\\n","\\r","\\t","\\b","\\a","\\f","\\v","\\'",'\\"',"\\\\"],brackets:[{open:"{",close:"}",token:"delimiter.curly"},{open:"[",close:"]",token:"delimiter.bracket"},{open:"(",close:")",token:"delimiter.parenthesis"}],tokenizer:{root:[{include:"@numbers"},{include:"@strings"},[/[{}\[\]()]/,"@brackets"],{include:"@operators"},[/#'$/,"comment.doc"],[/#'/,"comment.doc","@roxygen"],[/(^#.*$)/,"comment"],[/\s+/,"white"],[/[,:;]/,"delimiter"],[/@[a-zA-Z]\w*/,"tag"],[/[a-zA-Z]\w*/,{cases:{"@keywords":"keyword","@constants":"constant","@default":"identifier"}}]],roxygen:[[/@\w+/,{cases:{"@roxygen":"tag","@eos":{token:"comment.doc",next:"@pop"},"@default":"comment.doc"}}],[/\s+/,{cases:{"@eos":{token:"comment.doc",next:"@pop"},"@default":"comment.doc"}}],[/.*/,{token:"comment.doc",next:"@pop"}]],numbers:[[/0[xX][0-9a-fA-F]+/,"number.hex"],[/-?(\d*\.)?\d+([eE][+\-]?\d+)?/,"number"]],operators:[[/<{1,2}-/,"operator"],[/->{1,2}/,"operator"],[/%[^%\s]+%/,"operator"],[/\*\*/,"operator"],[/%%/,"operator"],[/&&/,"operator"],[/\|\|/,"operator"],[/<</,"operator"],[/>>/,"operator"],[/[-+=&|!<>^~*/:$]/,"operator"]],strings:[[/'/,"string.escape","@stringBody"],[/"/,"string.escape","@dblStringBody"]],stringBody:[[/\\./,{cases:{"@special":"string","@default":"error-token"}}],[/'/,"string.escape","@popall"],[/./,"string"]],dblStringBody:[[/\\./,{cases:{"@special":"string","@default":"error-token"}}],[/"/,"string.escape","@popall"],[/./,"string"]]}},uje=Object.freeze(Object.defineProperty({__proto__:null,conf:dje,language:hje},Symbol.toStringTag,{value:"Module"})),yX=["area","base","br","col","embed","hr","img","input","keygen","link","menuitem","meta","param","source","track","wbr"],gje={wordPattern:/(-?\d*\.\d\w*)|([^\`\~\!\@\$\^\&\*\(\)\-\=\+\[\{\]\}\\\|\;\:\'\"\,\.\<\>\/\s]+)/g,comments:{blockComment:["<!--","-->"]},brackets:[["<!--","-->"],["<",">"],["{","}"],["(",")"]],autoClosingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:'"',close:'"'},{open:"'",close:"'"}],surroundingPairs:[{open:'"',close:'"'},{open:"'",close:"'"},{open:"<",close:">"}],onEnterRules:[{beforeText:new RegExp(`<(?!(?:${yX.join("|")}))(\\w[\\w\\d]*)([^/>]*(?!/)>)[^<]*$`,"i"),afterText:/^<\/(\w[\w\d]*)\s*>$/i,action:{indentAction:H.IndentAction.IndentOutdent}},{beforeText:new RegExp(`<(?!(?:${yX.join("|")}))(\\w[\\w\\d]*)([^/>]*(?!/)>)[^<]*$`,"i"),action:{indentAction:H.IndentAction.Indent}}]},fje={defaultToken:"",tokenPostfix:"",tokenizer:{root:[[/@@@@/],[/@[^@]/,{token:"@rematch",switchTo:"@razorInSimpleState.root"}],[/<!DOCTYPE/,"metatag.html","@doctype"],[/<!--/,"comment.html","@comment"],[/(<)([\w\-]+)(\/>)/,["delimiter.html","tag.html","delimiter.html"]],[/(<)(script)/,["delimiter.html",{token:"tag.html",next:"@script"}]],[/(<)(style)/,["delimiter.html",{token:"tag.html",next:"@style"}]],[/(<)([:\w\-]+)/,["delimiter.html",{token:"tag.html",next:"@otherTag"}]],[/(<\/)([\w\-]+)/,["delimiter.html",{token:"tag.html",next:"@otherTag"}]],[/</,"delimiter.html"],[/[ \t\r\n]+/],[/[^<@]+/]],doctype:[[/@[^@]/,{token:"@rematch",switchTo:"@razorInSimpleState.comment"}],[/[^>]+/,"metatag.content.html"],[/>/,"metatag.html","@pop"]],comment:[[/@[^@]/,{token:"@rematch",switchTo:"@razorInSimpleState.comment"}],[/-->/,"comment.html","@pop"],[/[^-]+/,"comment.content.html"],[/./,"comment.content.html"]],otherTag:[[/@[^@]/,{token:"@rematch",switchTo:"@razorInSimpleState.otherTag"}],[/\/?>/,"delimiter.html","@pop"],[/"([^"]*)"/,"attribute.value"],[/'([^']*)'/,"attribute.value"],[/[\w\-]+/,"attribute.name"],[/=/,"delimiter"],[/[ \t\r\n]+/]],script:[[/@[^@]/,{token:"@rematch",switchTo:"@razorInSimpleState.script"}],[/type/,"attribute.name","@scriptAfterType"],[/"([^"]*)"/,"attribute.value"],[/'([^']*)'/,"attribute.value"],[/[\w\-]+/,"attribute.name"],[/=/,"delimiter"],[/>/,{token:"delimiter.html",next:"@scriptEmbedded.text/javascript",nextEmbedded:"text/javascript"}],[/[ \t\r\n]+/],[/(<\/)(script\s*)(>)/,["delimiter.html","tag.html",{token:"delimiter.html",next:"@pop"}]]],scriptAfterType:[[/@[^@]/,{token:"@rematch",switchTo:"@razorInSimpleState.scriptAfterType"}],[/=/,"delimiter","@scriptAfterTypeEquals"],[/>/,{token:"delimiter.html",next:"@scriptEmbedded.text/javascript",nextEmbedded:"text/javascript"}],[/[ \t\r\n]+/],[/<\/script\s*>/,{token:"@rematch",next:"@pop"}]],scriptAfterTypeEquals:[[/@[^@]/,{token:"@rematch",switchTo:"@razorInSimpleState.scriptAfterTypeEquals"}],[/"([^"]*)"/,{token:"attribute.value",switchTo:"@scriptWithCustomType.$1"}],[/'([^']*)'/,{token:"attribute.value",switchTo:"@scriptWithCustomType.$1"}],[/>/,{token:"delimiter.html",next:"@scriptEmbedded.text/javascript",nextEmbedded:"text/javascript"}],[/[ \t\r\n]+/],[/<\/script\s*>/,{token:"@rematch",next:"@pop"}]],scriptWithCustomType:[[/@[^@]/,{token:"@rematch",switchTo:"@razorInSimpleState.scriptWithCustomType.$S2"}],[/>/,{token:"delimiter.html",next:"@scriptEmbedded.$S2",nextEmbedded:"$S2"}],[/"([^"]*)"/,"attribute.value"],[/'([^']*)'/,"attribute.value"],[/[\w\-]+/,"attribute.name"],[/=/,"delimiter"],[/[ \t\r\n]+/],[/<\/script\s*>/,{token:"@rematch",next:"@pop"}]],scriptEmbedded:[[/@[^@]/,{token:"@rematch",switchTo:"@razorInEmbeddedState.scriptEmbedded.$S2",nextEmbedded:"@pop"}],[/<\/script/,{token:"@rematch",next:"@pop",nextEmbedded:"@pop"}]],style:[[/@[^@]/,{token:"@rematch",switchTo:"@razorInSimpleState.style"}],[/type/,"attribute.name","@styleAfterType"],[/"([^"]*)"/,"attribute.value"],[/'([^']*)'/,"attribute.value"],[/[\w\-]+/,"attribute.name"],[/=/,"delimiter"],[/>/,{token:"delimiter.html",next:"@styleEmbedded.text/css",nextEmbedded:"text/css"}],[/[ \t\r\n]+/],[/(<\/)(style\s*)(>)/,["delimiter.html","tag.html",{token:"delimiter.html",next:"@pop"}]]],styleAfterType:[[/@[^@]/,{token:"@rematch",switchTo:"@razorInSimpleState.styleAfterType"}],[/=/,"delimiter","@styleAfterTypeEquals"],[/>/,{token:"delimiter.html",next:"@styleEmbedded.text/css",nextEmbedded:"text/css"}],[/[ \t\r\n]+/],[/<\/style\s*>/,{token:"@rematch",next:"@pop"}]],styleAfterTypeEquals:[[/@[^@]/,{token:"@rematch",switchTo:"@razorInSimpleState.styleAfterTypeEquals"}],[/"([^"]*)"/,{token:"attribute.value",switchTo:"@styleWithCustomType.$1"}],[/'([^']*)'/,{token:"attribute.value",switchTo:"@styleWithCustomType.$1"}],[/>/,{token:"delimiter.html",next:"@styleEmbedded.text/css",nextEmbedded:"text/css"}],[/[ \t\r\n]+/],[/<\/style\s*>/,{token:"@rematch",next:"@pop"}]],styleWithCustomType:[[/@[^@]/,{token:"@rematch",switchTo:"@razorInSimpleState.styleWithCustomType.$S2"}],[/>/,{token:"delimiter.html",next:"@styleEmbedded.$S2",nextEmbedded:"$S2"}],[/"([^"]*)"/,"attribute.value"],[/'([^']*)'/,"attribute.value"],[/[\w\-]+/,"attribute.name"],[/=/,"delimiter"],[/[ \t\r\n]+/],[/<\/style\s*>/,{token:"@rematch",next:"@pop"}]],styleEmbedded:[[/@[^@]/,{token:"@rematch",switchTo:"@razorInEmbeddedState.styleEmbedded.$S2",nextEmbedded:"@pop"}],[/<\/style/,{token:"@rematch",next:"@pop",nextEmbedded:"@pop"}]],razorInSimpleState:[[/@\*/,"comment.cs","@razorBlockCommentTopLevel"],[/@[{(]/,"metatag.cs","@razorRootTopLevel"],[/(@)(\s*[\w]+)/,["metatag.cs",{token:"identifier.cs",switchTo:"@$S2.$S3"}]],[/[})]/,{token:"metatag.cs",switchTo:"@$S2.$S3"}],[/\*@/,{token:"comment.cs",switchTo:"@$S2.$S3"}]],razorInEmbeddedState:[[/@\*/,"comment.cs","@razorBlockCommentTopLevel"],[/@[{(]/,"metatag.cs","@razorRootTopLevel"],[/(@)(\s*[\w]+)/,["metatag.cs",{token:"identifier.cs",switchTo:"@$S2.$S3",nextEmbedded:"$S3"}]],[/[})]/,{token:"metatag.cs",switchTo:"@$S2.$S3",nextEmbedded:"$S3"}],[/\*@/,{token:"comment.cs",switchTo:"@$S2.$S3",nextEmbedded:"$S3"}]],razorBlockCommentTopLevel:[[/\*@/,"@rematch","@pop"],[/[^*]+/,"comment.cs"],[/./,"comment.cs"]],razorBlockComment:[[/\*@/,"comment.cs","@pop"],[/[^*]+/,"comment.cs"],[/./,"comment.cs"]],razorRootTopLevel:[[/\{/,"delimiter.bracket.cs","@razorRoot"],[/\(/,"delimiter.parenthesis.cs","@razorRoot"],[/[})]/,"@rematch","@pop"],{include:"razorCommon"}],razorRoot:[[/\{/,"delimiter.bracket.cs","@razorRoot"],[/\(/,"delimiter.parenthesis.cs","@razorRoot"],[/\}/,"delimiter.bracket.cs","@pop"],[/\)/,"delimiter.parenthesis.cs","@pop"],{include:"razorCommon"}],razorCommon:[[/[a-zA-Z_]\w*/,{cases:{"@razorKeywords":{token:"keyword.cs"},"@default":"identifier.cs"}}],[/[\[\]]/,"delimiter.array.cs"],[/[ \t\r\n]+/],[/\/\/.*$/,"comment.cs"],[/@\*/,"comment.cs","@razorBlockComment"],[/"([^"]*)"/,"string.cs"],[/'([^']*)'/,"string.cs"],[/(<)([\w\-]+)(\/>)/,["delimiter.html","tag.html","delimiter.html"]],[/(<)([\w\-]+)(>)/,["delimiter.html","tag.html","delimiter.html"]],[/(<\/)([\w\-]+)(>)/,["delimiter.html","tag.html","delimiter.html"]],[/[\+\-\*\%\&\|\^\~\!\=\<\>\/\?\;\:\.\,]/,"delimiter.cs"],[/\d*\d+[eE]([\-+]?\d+)?/,"number.float.cs"],[/\d*\.\d+([eE][\-+]?\d+)?/,"number.float.cs"],[/0[xX][0-9a-fA-F']*[0-9a-fA-F]/,"number.hex.cs"],[/0[0-7']*[0-7]/,"number.octal.cs"],[/0[bB][0-1']*[0-1]/,"number.binary.cs"],[/\d[\d']*/,"number.cs"],[/\d/,"number.cs"]]},razorKeywords:["abstract","as","async","await","base","bool","break","by","byte","case","catch","char","checked","class","const","continue","decimal","default","delegate","do","double","descending","explicit","event","extern","else","enum","false","finally","fixed","float","for","foreach","from","goto","group","if","implicit","in","int","interface","internal","into","is","lock","long","nameof","new","null","namespace","object","operator","out","override","orderby","params","private","protected","public","readonly","ref","return","switch","struct","sbyte","sealed","short","sizeof","stackalloc","static","string","select","this","throw","true","try","typeof","uint","ulong","unchecked","unsafe","ushort","using","var","virtual","volatile","void","when","while","where","yield","model","inject"],escapes:/\\(?:[abfnrtv\\"']|x[0-9A-Fa-f]{1,4}|u[0-9A-Fa-f]{4}|U[0-9A-Fa-f]{8})/},pje=Object.freeze(Object.defineProperty({__proto__:null,conf:gje,language:fje},Symbol.toStringTag,{value:"Module"})),mje={brackets:[["{","}"],["[","]"],["(",")"]],autoClosingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:'"',close:'"'},{open:"'",close:"'"}],surroundingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:'"',close:'"'},{open:"'",close:"'"}]},_je={defaultToken:"",tokenPostfix:".redis",ignoreCase:!0,brackets:[{open:"[",close:"]",token:"delimiter.square"},{open:"(",close:")",token:"delimiter.parenthesis"}],keywords:["APPEND","AUTH","BGREWRITEAOF","BGSAVE","BITCOUNT","BITFIELD","BITOP","BITPOS","BLPOP","BRPOP","BRPOPLPUSH","CLIENT","KILL","LIST","GETNAME","PAUSE","REPLY","SETNAME","CLUSTER","ADDSLOTS","COUNT-FAILURE-REPORTS","COUNTKEYSINSLOT","DELSLOTS","FAILOVER","FORGET","GETKEYSINSLOT","INFO","KEYSLOT","MEET","NODES","REPLICATE","RESET","SAVECONFIG","SET-CONFIG-EPOCH","SETSLOT","SLAVES","SLOTS","COMMAND","COUNT","GETKEYS","CONFIG","GET","REWRITE","SET","RESETSTAT","DBSIZE","DEBUG","OBJECT","SEGFAULT","DECR","DECRBY","DEL","DISCARD","DUMP","ECHO","EVAL","EVALSHA","EXEC","EXISTS","EXPIRE","EXPIREAT","FLUSHALL","FLUSHDB","GEOADD","GEOHASH","GEOPOS","GEODIST","GEORADIUS","GEORADIUSBYMEMBER","GETBIT","GETRANGE","GETSET","HDEL","HEXISTS","HGET","HGETALL","HINCRBY","HINCRBYFLOAT","HKEYS","HLEN","HMGET","HMSET","HSET","HSETNX","HSTRLEN","HVALS","INCR","INCRBY","INCRBYFLOAT","KEYS","LASTSAVE","LINDEX","LINSERT","LLEN","LPOP","LPUSH","LPUSHX","LRANGE","LREM","LSET","LTRIM","MGET","MIGRATE","MONITOR","MOVE","MSET","MSETNX","MULTI","PERSIST","PEXPIRE","PEXPIREAT","PFADD","PFCOUNT","PFMERGE","PING","PSETEX","PSUBSCRIBE","PUBSUB","PTTL","PUBLISH","PUNSUBSCRIBE","QUIT","RANDOMKEY","READONLY","READWRITE","RENAME","RENAMENX","RESTORE","ROLE","RPOP","RPOPLPUSH","RPUSH","RPUSHX","SADD","SAVE","SCARD","SCRIPT","FLUSH","LOAD","SDIFF","SDIFFSTORE","SELECT","SETBIT","SETEX","SETNX","SETRANGE","SHUTDOWN","SINTER","SINTERSTORE","SISMEMBER","SLAVEOF","SLOWLOG","SMEMBERS","SMOVE","SORT","SPOP","SRANDMEMBER","SREM","STRLEN","SUBSCRIBE","SUNION","SUNIONSTORE","SWAPDB","SYNC","TIME","TOUCH","TTL","TYPE","UNSUBSCRIBE","UNLINK","UNWATCH","WAIT","WATCH","ZADD","ZCARD","ZCOUNT","ZINCRBY","ZINTERSTORE","ZLEXCOUNT","ZRANGE","ZRANGEBYLEX","ZREVRANGEBYLEX","ZRANGEBYSCORE","ZRANK","ZREM","ZREMRANGEBYLEX","ZREMRANGEBYRANK","ZREMRANGEBYSCORE","ZREVRANGE","ZREVRANGEBYSCORE","ZREVRANK","ZSCORE","ZUNIONSTORE","SCAN","SSCAN","HSCAN","ZSCAN"],operators:[],builtinFunctions:[],builtinVariables:[],pseudoColumns:[],tokenizer:{root:[{include:"@whitespace"},{include:"@pseudoColumns"},{include:"@numbers"},{include:"@strings"},{include:"@scopes"},[/[;,.]/,"delimiter"],[/[()]/,"@brackets"],[/[\w@#$]+/,{cases:{"@keywords":"keyword","@operators":"operator","@builtinVariables":"predefined","@builtinFunctions":"predefined","@default":"identifier"}}],[/[<>=!%&+\-*/|~^]/,"operator"]],whitespace:[[/\s+/,"white"]],pseudoColumns:[[/[$][A-Za-z_][\w@#$]*/,{cases:{"@pseudoColumns":"predefined","@default":"identifier"}}]],numbers:[[/0[xX][0-9a-fA-F]*/,"number"],[/[$][+-]*\d*(\.\d*)?/,"number"],[/((\d+(\.\d*)?)|(\.\d+))([eE][\-+]?\d+)?/,"number"]],strings:[[/'/,{token:"string",next:"@string"}],[/"/,{token:"string.double",next:"@stringDouble"}]],string:[[/[^']+/,"string"],[/''/,"string"],[/'/,{token:"string",next:"@pop"}]],stringDouble:[[/[^"]+/,"string.double"],[/""/,"string.double"],[/"/,{token:"string.double",next:"@pop"}]],scopes:[]}},bje=Object.freeze(Object.defineProperty({__proto__:null,conf:mje,language:_je},Symbol.toStringTag,{value:"Module"})),wje={comments:{lineComment:"--",blockComment:["/*","*/"]},brackets:[["{","}"],["[","]"],["(",")"]],autoClosingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:'"',close:'"'},{open:"'",close:"'"}],surroundingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:'"',close:'"'},{open:"'",close:"'"}]},vje={defaultToken:"",tokenPostfix:".sql",ignoreCase:!0,brackets:[{open:"[",close:"]",token:"delimiter.square"},{open:"(",close:")",token:"delimiter.parenthesis"}],keywords:["AES128","AES256","ALL","ALLOWOVERWRITE","ANALYSE","ANALYZE","AND","ANY","ARRAY","AS","ASC","AUTHORIZATION","AZ64","BACKUP","BETWEEN","BINARY","BLANKSASNULL","BOTH","BYTEDICT","BZIP2","CASE","CAST","CHECK","COLLATE","COLUMN","CONSTRAINT","CREATE","CREDENTIALS","CROSS","CURRENT_DATE","CURRENT_TIME","CURRENT_TIMESTAMP","CURRENT_USER","CURRENT_USER_ID","DEFAULT","DEFERRABLE","DEFLATE","DEFRAG","DELTA","DELTA32K","DESC","DISABLE","DISTINCT","DO","ELSE","EMPTYASNULL","ENABLE","ENCODE","ENCRYPT","ENCRYPTION","END","EXCEPT","EXPLICIT","FALSE","FOR","FOREIGN","FREEZE","FROM","FULL","GLOBALDICT256","GLOBALDICT64K","GRANT","GROUP","GZIP","HAVING","IDENTITY","IGNORE","ILIKE","IN","INITIALLY","INNER","INTERSECT","INTO","IS","ISNULL","JOIN","LANGUAGE","LEADING","LEFT","LIKE","LIMIT","LOCALTIME","LOCALTIMESTAMP","LUN","LUNS","LZO","LZOP","MINUS","MOSTLY16","MOSTLY32","MOSTLY8","NATURAL","NEW","NOT","NOTNULL","NULL","NULLS","OFF","OFFLINE","OFFSET","OID","OLD","ON","ONLY","OPEN","OR","ORDER","OUTER","OVERLAPS","PARALLEL","PARTITION","PERCENT","PERMISSIONS","PLACING","PRIMARY","RAW","READRATIO","RECOVER","REFERENCES","RESPECT","REJECTLOG","RESORT","RESTORE","RIGHT","SELECT","SESSION_USER","SIMILAR","SNAPSHOT","SOME","SYSDATE","SYSTEM","TABLE","TAG","TDES","TEXT255","TEXT32K","THEN","TIMESTAMP","TO","TOP","TRAILING","TRUE","TRUNCATECOLUMNS","UNION","UNIQUE","USER","USING","VERBOSE","WALLET","WHEN","WHERE","WITH","WITHOUT"],operators:["AND","BETWEEN","IN","LIKE","NOT","OR","IS","NULL","INTERSECT","UNION","INNER","JOIN","LEFT","OUTER","RIGHT"],builtinFunctions:["current_schema","current_schemas","has_database_privilege","has_schema_privilege","has_table_privilege","age","current_time","current_timestamp","localtime","isfinite","now","ascii","get_bit","get_byte","set_bit","set_byte","to_ascii","approximate percentile_disc","avg","count","listagg","max","median","min","percentile_cont","stddev_samp","stddev_pop","sum","var_samp","var_pop","bit_and","bit_or","bool_and","bool_or","cume_dist","first_value","lag","last_value","lead","nth_value","ratio_to_report","dense_rank","ntile","percent_rank","rank","row_number","case","coalesce","decode","greatest","least","nvl","nvl2","nullif","add_months","at time zone","convert_timezone","current_date","date_cmp","date_cmp_timestamp","date_cmp_timestamptz","date_part_year","dateadd","datediff","date_part","date_trunc","extract","getdate","interval_cmp","last_day","months_between","next_day","sysdate","timeofday","timestamp_cmp","timestamp_cmp_date","timestamp_cmp_timestamptz","timestamptz_cmp","timestamptz_cmp_date","timestamptz_cmp_timestamp","timezone","to_timestamp","trunc","abs","acos","asin","atan","atan2","cbrt","ceil","ceiling","checksum","cos","cot","degrees","dexp","dlog1","dlog10","exp","floor","ln","log","mod","pi","power","radians","random","round","sin","sign","sqrt","tan","to_hex","bpcharcmp","btrim","bttext_pattern_cmp","char_length","character_length","charindex","chr","concat","crc32","func_sha1","initcap","left and rights","len","length","lower","lpad and rpads","ltrim","md5","octet_length","position","quote_ident","quote_literal","regexp_count","regexp_instr","regexp_replace","regexp_substr","repeat","replace","replicate","reverse","rtrim","split_part","strpos","strtol","substring","textlen","translate","trim","upper","cast","convert","to_char","to_date","to_number","json_array_length","json_extract_array_element_text","json_extract_path_text","current_setting","pg_cancel_backend","pg_terminate_backend","set_config","current_database","current_user","current_user_id","pg_backend_pid","pg_last_copy_count","pg_last_copy_id","pg_last_query_id","pg_last_unload_count","session_user","slice_num","user","version","abbrev","acosd","any","area","array_agg","array_append","array_cat","array_dims","array_fill","array_length","array_lower","array_ndims","array_position","array_positions","array_prepend","array_remove","array_replace","array_to_json","array_to_string","array_to_tsvector","array_upper","asind","atan2d","atand","bit","bit_length","bound_box","box","brin_summarize_new_values","broadcast","cardinality","center","circle","clock_timestamp","col_description","concat_ws","convert_from","convert_to","corr","cosd","cotd","covar_pop","covar_samp","current_catalog","current_query","current_role","currval","cursor_to_xml","diameter","div","encode","enum_first","enum_last","enum_range","every","family","format","format_type","generate_series","generate_subscripts","get_current_ts_config","gin_clean_pending_list","grouping","has_any_column_privilege","has_column_privilege","has_foreign_data_wrapper_privilege","has_function_privilege","has_language_privilege","has_sequence_privilege","has_server_privilege","has_tablespace_privilege","has_type_privilege","height","host","hostmask","inet_client_addr","inet_client_port","inet_merge","inet_same_family","inet_server_addr","inet_server_port","isclosed","isempty","isopen","json_agg","json_object","json_object_agg","json_populate_record","json_populate_recordset","json_to_record","json_to_recordset","jsonb_agg","jsonb_object_agg","justify_days","justify_hours","justify_interval","lastval","left","line","localtimestamp","lower_inc","lower_inf","lpad","lseg","make_date","make_interval","make_time","make_timestamp","make_timestamptz","masklen","mode","netmask","network","nextval","npoints","num_nonnulls","num_nulls","numnode","obj_description","overlay","parse_ident","path","pclose","percentile_disc","pg_advisory_lock","pg_advisory_lock_shared","pg_advisory_unlock","pg_advisory_unlock_all","pg_advisory_unlock_shared","pg_advisory_xact_lock","pg_advisory_xact_lock_shared","pg_backup_start_time","pg_blocking_pids","pg_client_encoding","pg_collation_is_visible","pg_column_size","pg_conf_load_time","pg_control_checkpoint","pg_control_init","pg_control_recovery","pg_control_system","pg_conversion_is_visible","pg_create_logical_replication_slot","pg_create_physical_replication_slot","pg_create_restore_point","pg_current_xlog_flush_location","pg_current_xlog_insert_location","pg_current_xlog_location","pg_database_size","pg_describe_object","pg_drop_replication_slot","pg_export_snapshot","pg_filenode_relation","pg_function_is_visible","pg_get_constraintdef","pg_get_expr","pg_get_function_arguments","pg_get_function_identity_arguments","pg_get_function_result","pg_get_functiondef","pg_get_indexdef","pg_get_keywords","pg_get_object_address","pg_get_owned_sequence","pg_get_ruledef","pg_get_serial_sequence","pg_get_triggerdef","pg_get_userbyid","pg_get_viewdef","pg_has_role","pg_identify_object","pg_identify_object_as_address","pg_index_column_has_property","pg_index_has_property","pg_indexam_has_property","pg_indexes_size","pg_is_in_backup","pg_is_in_recovery","pg_is_other_temp_schema","pg_is_xlog_replay_paused","pg_last_committed_xact","pg_last_xact_replay_timestamp","pg_last_xlog_receive_location","pg_last_xlog_replay_location","pg_listening_channels","pg_logical_emit_message","pg_logical_slot_get_binary_changes","pg_logical_slot_get_changes","pg_logical_slot_peek_binary_changes","pg_logical_slot_peek_changes","pg_ls_dir","pg_my_temp_schema","pg_notification_queue_usage","pg_opclass_is_visible","pg_operator_is_visible","pg_opfamily_is_visible","pg_options_to_table","pg_postmaster_start_time","pg_read_binary_file","pg_read_file","pg_relation_filenode","pg_relation_filepath","pg_relation_size","pg_reload_conf","pg_replication_origin_create","pg_replication_origin_drop","pg_replication_origin_oid","pg_replication_origin_progress","pg_replication_origin_session_is_setup","pg_replication_origin_session_progress","pg_replication_origin_session_reset","pg_replication_origin_session_setup","pg_replication_origin_xact_reset","pg_replication_origin_xact_setup","pg_rotate_logfile","pg_size_bytes","pg_size_pretty","pg_sleep","pg_sleep_for","pg_sleep_until","pg_start_backup","pg_stat_file","pg_stop_backup","pg_switch_xlog","pg_table_is_visible","pg_table_size","pg_tablespace_databases","pg_tablespace_location","pg_tablespace_size","pg_total_relation_size","pg_trigger_depth","pg_try_advisory_lock","pg_try_advisory_lock_shared","pg_try_advisory_xact_lock","pg_try_advisory_xact_lock_shared","pg_ts_config_is_visible","pg_ts_dict_is_visible","pg_ts_parser_is_visible","pg_ts_template_is_visible","pg_type_is_visible","pg_typeof","pg_xact_commit_timestamp","pg_xlog_location_diff","pg_xlog_replay_pause","pg_xlog_replay_resume","pg_xlogfile_name","pg_xlogfile_name_offset","phraseto_tsquery","plainto_tsquery","point","polygon","popen","pqserverversion","query_to_xml","querytree","quote_nullable","radius","range_merge","regexp_matches","regexp_split_to_array","regexp_split_to_table","regr_avgx","regr_avgy","regr_count","regr_intercept","regr_r2","regr_slope","regr_sxx","regr_sxy","regr_syy","right","row_security_active","row_to_json","rpad","scale","set_masklen","setseed","setval","setweight","shobj_description","sind","sprintf","statement_timestamp","stddev","string_agg","string_to_array","strip","substr","table_to_xml","table_to_xml_and_xmlschema","tand","text","to_json","to_regclass","to_regnamespace","to_regoper","to_regoperator","to_regproc","to_regprocedure","to_regrole","to_regtype","to_tsquery","to_tsvector","transaction_timestamp","ts_debug","ts_delete","ts_filter","ts_headline","ts_lexize","ts_parse","ts_rank","ts_rank_cd","ts_rewrite","ts_stat","ts_token_type","tsquery_phrase","tsvector_to_array","tsvector_update_trigger","tsvector_update_trigger_column","txid_current","txid_current_snapshot","txid_snapshot_xip","txid_snapshot_xmax","txid_snapshot_xmin","txid_visible_in_snapshot","unnest","upper_inc","upper_inf","variance","width","width_bucket","xml_is_well_formed","xml_is_well_formed_content","xml_is_well_formed_document","xmlagg","xmlcomment","xmlconcat","xmlelement","xmlexists","xmlforest","xmlparse","xmlpi","xmlroot","xmlserialize","xpath","xpath_exists"],builtinVariables:[],pseudoColumns:[],tokenizer:{root:[{include:"@comments"},{include:"@whitespace"},{include:"@pseudoColumns"},{include:"@numbers"},{include:"@strings"},{include:"@complexIdentifiers"},{include:"@scopes"},[/[;,.]/,"delimiter"],[/[()]/,"@brackets"],[/[\w@#$]+/,{cases:{"@keywords":"keyword","@operators":"operator","@builtinVariables":"predefined","@builtinFunctions":"predefined","@default":"identifier"}}],[/[<>=!%&+\-*/|~^]/,"operator"]],whitespace:[[/\s+/,"white"]],comments:[[/--+.*/,"comment"],[/\/\*/,{token:"comment.quote",next:"@comment"}]],comment:[[/[^*/]+/,"comment"],[/\*\//,{token:"comment.quote",next:"@pop"}],[/./,"comment"]],pseudoColumns:[[/[$][A-Za-z_][\w@#$]*/,{cases:{"@pseudoColumns":"predefined","@default":"identifier"}}]],numbers:[[/0[xX][0-9a-fA-F]*/,"number"],[/[$][+-]*\d*(\.\d*)?/,"number"],[/((\d+(\.\d*)?)|(\.\d+))([eE][\-+]?\d+)?/,"number"]],strings:[[/'/,{token:"string",next:"@string"}]],string:[[/[^']+/,"string"],[/''/,"string"],[/'/,{token:"string",next:"@pop"}]],complexIdentifiers:[[/"/,{token:"identifier.quote",next:"@quotedIdentifier"}]],quotedIdentifier:[[/[^"]+/,"identifier"],[/""/,"identifier"],[/"/,{token:"identifier.quote",next:"@pop"}]],scopes:[]}},Cje=Object.freeze(Object.defineProperty({__proto__:null,conf:wje,language:vje},Symbol.toStringTag,{value:"Module"})),yje={brackets:[["{","}"],["[","]"],["(",")"]],autoClosingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:"<",close:">",notIn:["string"]}],surroundingPairs:[{open:"(",close:")"},{open:"[",close:"]"},{open:"`",close:"`"}],folding:{markers:{start:new RegExp("^\\s*<!--\\s*#?region\\b.*-->"),end:new RegExp("^\\s*<!--\\s*#?endregion\\b.*-->")}}},Sje={defaultToken:"",tokenPostfix:".rst",control:/[\\`*_\[\]{}()#+\-\.!]/,escapes:/\\(?:@control)/,empty:["area","base","basefont","br","col","frame","hr","img","input","isindex","link","meta","param"],alphanumerics:/[A-Za-z0-9]/,simpleRefNameWithoutBq:/(?:@alphanumerics[-_+:.]*@alphanumerics)+|(?:@alphanumerics+)/,simpleRefName:/(?:`@phrase`|@simpleRefNameWithoutBq)/,phrase:/@simpleRefNameWithoutBq(?:\s@simpleRefNameWithoutBq)*/,citationName:/[A-Za-z][A-Za-z0-9-_.]*/,blockLiteralStart:/(?:[!"#$%&'()*+,-./:;<=>?@\[\]^_`{|}~]|[\s])/,precedingChars:/(?:[ -:/'"<([{])/,followingChars:/(?:[ -.,:;!?/'")\]}>]|$)/,punctuation:/(=|-|~|`|#|"|\^|\+|\*|:|\.|'|_|\+)/,tokenizer:{root:[[/^(@punctuation{3,}$){1,1}?/,"keyword"],[/^\s*([\*\-+‣•]|[a-zA-Z0-9]+\.|\([a-zA-Z0-9]+\)|[a-zA-Z0-9]+\))\s/,"keyword"],[/([ ]::)\s*$/,"keyword","@blankLineOfLiteralBlocks"],[/(::)\s*$/,"keyword","@blankLineOfLiteralBlocks"],{include:"@tables"},{include:"@explicitMarkupBlocks"},{include:"@inlineMarkup"}],explicitMarkupBlocks:[{include:"@citations"},{include:"@footnotes"},[/^(\.\.\s)(@simpleRefName)(::\s)(.*)$/,[{token:"",next:"subsequentLines"},"keyword","",""]],[/^(\.\.)(\s+)(_)(@simpleRefName)(:)(\s+)(.*)/,[{token:"",next:"hyperlinks"},"","","string.link","","","string.link"]],[/^((?:(?:\.\.)(?:\s+))?)(__)(:)(\s+)(.*)/,[{token:"",next:"subsequentLines"},"","","","string.link"]],[/^(__\s+)(.+)/,["","string.link"]],[/^(\.\.)( \|)([^| ]+[^|]*[^| ]*)(\| )(@simpleRefName)(:: .*)/,[{token:"",next:"subsequentLines"},"","string.link","","keyword",""],"@rawBlocks"],[/(\|)([^| ]+[^|]*[^| ]*)(\|_{0,2})/,["","string.link",""]],[/^(\.\.)([ ].*)$/,[{token:"",next:"@comments"},"comment"]]],inlineMarkup:[{include:"@citationsReference"},{include:"@footnotesReference"},[/(@simpleRefName)(_{1,2})/,["string.link",""]],[/(`)([^<`]+\s+)(<)(.*)(>)(`)(_)/,["","string.link","","string.link","","",""]],[/\*\*([^\\*]|\*(?!\*))+\*\*/,"strong"],[/\*[^*]+\*/,"emphasis"],[/(``)((?:[^`]|\`(?!`))+)(``)/,["","keyword",""]],[/(__\s+)(.+)/,["","keyword"]],[/(:)((?:@simpleRefNameWithoutBq)?)(:`)([^`]+)(`)/,["","keyword","","",""]],[/(`)([^`]+)(`:)((?:@simpleRefNameWithoutBq)?)(:)/,["","","","keyword",""]],[/(`)([^`]+)(`)/,""],[/(_`)(@phrase)(`)/,["","string.link",""]]],citations:[[/^(\.\.\s+\[)((?:@citationName))(\]\s+)(.*)/,[{token:"",next:"@subsequentLines"},"string.link","",""]]],citationsReference:[[/(\[)(@citationName)(\]_)/,["","string.link",""]]],footnotes:[[/^(\.\.\s+\[)((?:[0-9]+))(\]\s+.*)/,[{token:"",next:"@subsequentLines"},"string.link",""]],[/^(\.\.\s+\[)((?:#@simpleRefName?))(\]\s+)(.*)/,[{token:"",next:"@subsequentLines"},"string.link","",""]],[/^(\.\.\s+\[)((?:\*))(\]\s+)(.*)/,[{token:"",next:"@subsequentLines"},"string.link","",""]]],footnotesReference:[[/(\[)([0-9]+)(\])(_)/,["","string.link","",""]],[/(\[)(#@simpleRefName?)(\])(_)/,["","string.link","",""]],[/(\[)(\*)(\])(_)/,["","string.link","",""]]],blankLineOfLiteralBlocks:[[/^$/,"","@subsequentLinesOfLiteralBlocks"],[/^.*$/,"","@pop"]],subsequentLinesOfLiteralBlocks:[[/(@blockLiteralStart+)(.*)/,["keyword",""]],[/^(?!blockLiteralStart)/,"","@popall"]],subsequentLines:[[/^[\s]+.*/,""],[/^(?!\s)/,"","@pop"]],hyperlinks:[[/^[\s]+.*/,"string.link"],[/^(?!\s)/,"","@pop"]],comments:[[/^[\s]+.*/,"comment"],[/^(?!\s)/,"","@pop"]],tables:[[/\+-[+-]+/,"keyword"],[/\+=[+=]+/,"keyword"]]}},xje=Object.freeze(Object.defineProperty({__proto__:null,conf:yje,language:Sje},Symbol.toStringTag,{value:"Module"})),Lje={comments:{lineComment:"#",blockComment:["=begin","=end"]},brackets:[["(",")"],["{","}"],["[","]"]],autoClosingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:'"',close:'"'},{open:"'",close:"'"}],surroundingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:'"',close:'"'},{open:"'",close:"'"}],indentationRules:{increaseIndentPattern:new RegExp(`^\\s*((begin|class|(private|protected)\\s+def|def|else|elsif|ensure|for|if|module|rescue|unless|until|when|while|case)|([^#]*\\sdo\\b)|([^#]*=\\s*(case|if|unless)))\\b([^#\\{;]|("|'|/).*\\4)*(#.*)?$`),decreaseIndentPattern:new RegExp("^\\s*([}\\]]([,)]?\\s*(#|$)|\\.[a-zA-Z_]\\w*\\b)|(end|rescue|ensure|else|elsif|when)\\b)")}},kje={tokenPostfix:".ruby",keywords:["__LINE__","__ENCODING__","__FILE__","BEGIN","END","alias","and","begin","break","case","class","def","defined?","do","else","elsif","end","ensure","for","false","if","in","module","next","nil","not","or","redo","rescue","retry","return","self","super","then","true","undef","unless","until","when","while","yield"],keywordops:["::","..","...","?",":","=>"],builtins:["require","public","private","include","extend","attr_reader","protected","private_class_method","protected_class_method","new"],declarations:["module","class","def","case","do","begin","for","if","while","until","unless"],linedecls:["def","case","do","begin","for","if","while","until","unless"],operators:["^","&","|","<=>","==","===","!~","=~",">",">=","<","<=","<<",">>","+","-","*","/","%","**","~","+@","-@","[]","[]=","`","+=","-=","*=","**=","/=","^=","%=","<<=",">>=","&=","&&=","||=","|="],brackets:[{open:"(",close:")",token:"delimiter.parenthesis"},{open:"{",close:"}",token:"delimiter.curly"},{open:"[",close:"]",token:"delimiter.square"}],symbols:/[=><!~?:&|+\-*\/\^%\.]+/,escape:/(?:[abefnrstv\\"'\n\r]|[0-7]{1,3}|x[0-9A-Fa-f]{1,2}|u[0-9A-Fa-f]{4})/,escapes:/\\(?:C\-(@escape|.)|c(@escape|.)|@escape)/,decpart:/\d(_?\d)*/,decimal:/0|@decpart/,delim:/[^a-zA-Z0-9\s\n\r]/,heredelim:/(?:\w+|'[^']*'|"[^"]*"|`[^`]*`)/,regexpctl:/[(){}\[\]\$\^|\-*+?\.]/,regexpesc:/\\(?:[AzZbBdDfnrstvwWn0\\\/]|@regexpctl|c[A-Z]|x[0-9a-fA-F]{2}|u[0-9a-fA-F]{4})?/,tokenizer:{root:[[/^(\s*)([a-z_]\w*[!?=]?)/,["white",{cases:{"for|until|while":{token:"keyword.$2",next:"@dodecl.$2"},"@declarations":{token:"keyword.$2",next:"@root.$2"},end:{token:"keyword.$S2",next:"@pop"},"@keywords":"keyword","@builtins":"predefined","@default":"identifier"}}]],[/[a-z_]\w*[!?=]?/,{cases:{"if|unless|while|until":{token:"keyword.$0x",next:"@modifier.$0x"},for:{token:"keyword.$2",next:"@dodecl.$2"},"@linedecls":{token:"keyword.$0",next:"@root.$0"},end:{token:"keyword.$S2",next:"@pop"},"@keywords":"keyword","@builtins":"predefined","@default":"identifier"}}],[/[A-Z][\w]*[!?=]?/,"constructor.identifier"],[/\$[\w]*/,"global.constant"],[/@[\w]*/,"namespace.instance.identifier"],[/@@@[\w]*/,"namespace.class.identifier"],[/<<[-~](@heredelim).*/,{token:"string.heredoc.delimiter",next:"@heredoc.$1"}],[/[ \t\r\n]+<<(@heredelim).*/,{token:"string.heredoc.delimiter",next:"@heredoc.$1"}],[/^<<(@heredelim).*/,{token:"string.heredoc.delimiter",next:"@heredoc.$1"}],{include:"@whitespace"},[/"/,{token:"string.d.delim",next:'@dstring.d."'}],[/'/,{token:"string.sq.delim",next:"@sstring.sq"}],[/%([rsqxwW]|Q?)/,{token:"@rematch",next:"pstring"}],[/`/,{token:"string.x.delim",next:"@dstring.x.`"}],[/:(\w|[$@])\w*[!?=]?/,"string.s"],[/:"/,{token:"string.s.delim",next:'@dstring.s."'}],[/:'/,{token:"string.s.delim",next:"@sstring.s"}],[/\/(?=(\\\/|[^\/\n])+\/)/,{token:"regexp.delim",next:"@regexp"}],[/[{}()\[\]]/,"@brackets"],[/@symbols/,{cases:{"@keywordops":"keyword","@operators":"operator","@default":""}}],[/[;,]/,"delimiter"],[/0[xX][0-9a-fA-F](_?[0-9a-fA-F])*/,"number.hex"],[/0[_oO][0-7](_?[0-7])*/,"number.octal"],[/0[bB][01](_?[01])*/,"number.binary"],[/0[dD]@decpart/,"number"],[/@decimal((\.@decpart)?([eE][\-+]?@decpart)?)/,{cases:{$1:"number.float","@default":"number"}}]],dodecl:[[/^/,{token:"",switchTo:"@root.$S2"}],[/[a-z_]\w*[!?=]?/,{cases:{end:{token:"keyword.$S2",next:"@pop"},do:{token:"keyword",switchTo:"@root.$S2"},"@linedecls":{token:"@rematch",switchTo:"@root.$S2"},"@keywords":"keyword","@builtins":"predefined","@default":"identifier"}}],{include:"@root"}],modifier:[[/^/,"","@pop"],[/[a-z_]\w*[!?=]?/,{cases:{end:{token:"keyword.$S2",next:"@pop"},"then|else|elsif|do":{token:"keyword",switchTo:"@root.$S2"},"@linedecls":{token:"@rematch",switchTo:"@root.$S2"},"@keywords":"keyword","@builtins":"predefined","@default":"identifier"}}],{include:"@root"}],sstring:[[/[^\\']+/,"string.$S2"],[/\\\\|\\'|\\$/,"string.$S2.escape"],[/\\./,"string.$S2.invalid"],[/'/,{token:"string.$S2.delim",next:"@pop"}]],dstring:[[/[^\\`"#]+/,"string.$S2"],[/#/,"string.$S2.escape","@interpolated"],[/\\$/,"string.$S2.escape"],[/@escapes/,"string.$S2.escape"],[/\\./,"string.$S2.escape.invalid"],[/[`"]/,{cases:{"$#==$S3":{token:"string.$S2.delim",next:"@pop"},"@default":"string.$S2"}}]],heredoc:[[/^(\s*)(@heredelim)$/,{cases:{"$2==$S2":["string.heredoc",{token:"string.heredoc.delimiter",next:"@pop"}],"@default":["string.heredoc","string.heredoc"]}}],[/.*/,"string.heredoc"]],interpolated:[[/\$\w*/,"global.constant","@pop"],[/@\w*/,"namespace.class.identifier","@pop"],[/@@@\w*/,"namespace.instance.identifier","@pop"],[/[{]/,{token:"string.escape.curly",switchTo:"@interpolated_compound"}],["","","@pop"]],interpolated_compound:[[/[}]/,{token:"string.escape.curly",next:"@pop"}],{include:"@root"}],pregexp:[{include:"@whitespace"},[/[^\(\{\[\\]/,{cases:{"$#==$S3":{token:"regexp.delim",next:"@pop"},"$#==$S2":{token:"regexp.delim",next:"@push"},"~[)}\\]]":"@brackets.regexp.escape.control","~@regexpctl":"regexp.escape.control","@default":"regexp"}}],{include:"@regexcontrol"}],regexp:[{include:"@regexcontrol"},[/[^\\\/]/,"regexp"],["/[ixmp]*",{token:"regexp.delim"},"@pop"]],regexcontrol:[[/(\{)(\d+(?:,\d*)?)(\})/,["@brackets.regexp.escape.control","regexp.escape.control","@brackets.regexp.escape.control"]],[/(\[)(\^?)/,["@brackets.regexp.escape.control",{token:"regexp.escape.control",next:"@regexrange"}]],[/(\()(\?[:=!])/,["@brackets.regexp.escape.control","regexp.escape.control"]],[/\(\?#/,{token:"regexp.escape.control",next:"@regexpcomment"}],[/[()]/,"@brackets.regexp.escape.control"],[/@regexpctl/,"regexp.escape.control"],[/\\$/,"regexp.escape"],[/@regexpesc/,"regexp.escape"],[/\\\./,"regexp.invalid"],[/#/,"regexp.escape","@interpolated"]],regexrange:[[/-/,"regexp.escape.control"],[/\^/,"regexp.invalid"],[/\\$/,"regexp.escape"],[/@regexpesc/,"regexp.escape"],[/[^\]]/,"regexp"],[/\]/,"@brackets.regexp.escape.control","@pop"]],regexpcomment:[[/[^)]+/,"comment"],[/\)/,{token:"regexp.escape.control",next:"@pop"}]],pstring:[[/%([qws])\(/,{token:"string.$1.delim",switchTo:"@qstring.$1.(.)"}],[/%([qws])\[/,{token:"string.$1.delim",switchTo:"@qstring.$1.[.]"}],[/%([qws])\{/,{token:"string.$1.delim",switchTo:"@qstring.$1.{.}"}],[/%([qws])</,{token:"string.$1.delim",switchTo:"@qstring.$1.<.>"}],[/%([qws])(@delim)/,{token:"string.$1.delim",switchTo:"@qstring.$1.$2.$2"}],[/%r\(/,{token:"regexp.delim",switchTo:"@pregexp.(.)"}],[/%r\[/,{token:"regexp.delim",switchTo:"@pregexp.[.]"}],[/%r\{/,{token:"regexp.delim",switchTo:"@pregexp.{.}"}],[/%r</,{token:"regexp.delim",switchTo:"@pregexp.<.>"}],[/%r(@delim)/,{token:"regexp.delim",switchTo:"@pregexp.$1.$1"}],[/%(x|W|Q?)\(/,{token:"string.$1.delim",switchTo:"@qqstring.$1.(.)"}],[/%(x|W|Q?)\[/,{token:"string.$1.delim",switchTo:"@qqstring.$1.[.]"}],[/%(x|W|Q?)\{/,{token:"string.$1.delim",switchTo:"@qqstring.$1.{.}"}],[/%(x|W|Q?)</,{token:"string.$1.delim",switchTo:"@qqstring.$1.<.>"}],[/%(x|W|Q?)(@delim)/,{token:"string.$1.delim",switchTo:"@qqstring.$1.$2.$2"}],[/%([rqwsxW]|Q?)./,{token:"invalid",next:"@pop"}],[/./,{token:"invalid",next:"@pop"}]],qstring:[[/\\$/,"string.$S2.escape"],[/\\./,"string.$S2.escape"],[/./,{cases:{"$#==$S4":{token:"string.$S2.delim",next:"@pop"},"$#==$S3":{token:"string.$S2.delim",next:"@push"},"@default":"string.$S2"}}]],qqstring:[[/#/,"string.$S2.escape","@interpolated"],{include:"@qstring"}],whitespace:[[/[ \t\r\n]+/,""],[/^\s*=begin\b/,"comment","@comment"],[/#.*$/,"comment"]],comment:[[/[^=]+/,"comment"],[/^\s*=begin\b/,"comment.invalid"],[/^\s*=end\b.*/,"comment","@pop"],[/[=]/,"comment"]]}},Eje=Object.freeze(Object.defineProperty({__proto__:null,conf:Lje,language:kje},Symbol.toStringTag,{value:"Module"})),Ije={comments:{lineComment:"//",blockComment:["/*","*/"]},brackets:[["{","}"],["[","]"],["(",")"]],autoClosingPairs:[{open:"[",close:"]"},{open:"{",close:"}"},{open:"(",close:")"},{open:'"',close:'"',notIn:["string"]}],surroundingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:'"',close:'"'},{open:"'",close:"'"}],folding:{markers:{start:new RegExp("^\\s*#pragma\\s+region\\b"),end:new RegExp("^\\s*#pragma\\s+endregion\\b")}}},Dje={tokenPostfix:".rust",defaultToken:"invalid",keywords:["as","async","await","box","break","const","continue","crate","dyn","else","enum","extern","false","fn","for","if","impl","in","let","loop","match","mod","move","mut","pub","ref","return","self","static","struct","super","trait","true","try","type","unsafe","use","where","while","catch","default","union","static","abstract","alignof","become","do","final","macro","offsetof","override","priv","proc","pure","sizeof","typeof","unsized","virtual","yield"],typeKeywords:["Self","m32","m64","m128","f80","f16","f128","int","uint","float","char","bool","u8","u16","u32","u64","f32","f64","i8","i16","i32","i64","str","Option","Either","c_float","c_double","c_void","FILE","fpos_t","DIR","dirent","c_char","c_schar","c_uchar","c_short","c_ushort","c_int","c_uint","c_long","c_ulong","size_t","ptrdiff_t","clock_t","time_t","c_longlong","c_ulonglong","intptr_t","uintptr_t","off_t","dev_t","ino_t","pid_t","mode_t","ssize_t"],constants:["true","false","Some","None","Left","Right","Ok","Err"],supportConstants:["EXIT_FAILURE","EXIT_SUCCESS","RAND_MAX","EOF","SEEK_SET","SEEK_CUR","SEEK_END","_IOFBF","_IONBF","_IOLBF","BUFSIZ","FOPEN_MAX","FILENAME_MAX","L_tmpnam","TMP_MAX","O_RDONLY","O_WRONLY","O_RDWR","O_APPEND","O_CREAT","O_EXCL","O_TRUNC","S_IFIFO","S_IFCHR","S_IFBLK","S_IFDIR","S_IFREG","S_IFMT","S_IEXEC","S_IWRITE","S_IREAD","S_IRWXU","S_IXUSR","S_IWUSR","S_IRUSR","F_OK","R_OK","W_OK","X_OK","STDIN_FILENO","STDOUT_FILENO","STDERR_FILENO"],supportMacros:["format!","print!","println!","panic!","format_args!","unreachable!","write!","writeln!"],operators:["!","!=","%","%=","&","&=","&&","*","*=","+","+=","-","-=","->",".","..","...","/","/=",":",";","<<","<<=","<","<=","=","==","=>",">",">=",">>",">>=","@","^","^=","|","|=","||","_","?","#"],escapes:/\\([nrt0\"''\\]|x\h{2}|u\{\h{1,6}\})/,delimiters:/[,]/,symbols:/[\#\!\%\&\*\+\-\.\/\:\;\<\=\>\@\^\|_\?]+/,intSuffixes:/[iu](8|16|32|64|128|size)/,floatSuffixes:/f(32|64)/,tokenizer:{root:[[/r(#*)"/,{token:"string.quote",bracket:"@open",next:"@stringraw.$1"}],[/[a-zA-Z][a-zA-Z0-9_]*!?|_[a-zA-Z0-9_]+/,{cases:{"@typeKeywords":"keyword.type","@keywords":"keyword","@supportConstants":"keyword","@supportMacros":"keyword","@constants":"keyword","@default":"identifier"}}],[/\$/,"identifier"],[/'[a-zA-Z_][a-zA-Z0-9_]*(?=[^\'])/,"identifier"],[/'(\S|@escapes)'/,"string.byteliteral"],[/"/,{token:"string.quote",bracket:"@open",next:"@string"}],{include:"@numbers"},{include:"@whitespace"},[/@delimiters/,{cases:{"@keywords":"keyword","@default":"delimiter"}}],[/[{}()\[\]<>]/,"@brackets"],[/@symbols/,{cases:{"@operators":"operator","@default":""}}]],whitespace:[[/[ \t\r\n]+/,"white"],[/\/\*/,"comment","@comment"],[/\/\/.*$/,"comment"]],comment:[[/[^\/*]+/,"comment"],[/\/\*/,"comment","@push"],["\\*/","comment","@pop"],[/[\/*]/,"comment"]],string:[[/[^\\"]+/,"string"],[/@escapes/,"string.escape"],[/\\./,"string.escape.invalid"],[/"/,{token:"string.quote",bracket:"@close",next:"@pop"}]],stringraw:[[/[^"#]+/,{token:"string"}],[/"(#*)/,{cases:{"$1==$S2":{token:"string.quote",bracket:"@close",next:"@pop"},"@default":{token:"string"}}}],[/["#]/,{token:"string"}]],numbers:[[/(0o[0-7_]+)(@intSuffixes)?/,{token:"number"}],[/(0b[0-1_]+)(@intSuffixes)?/,{token:"number"}],[/[\d][\d_]*(\.[\d][\d_]*)?[eE][+-][\d_]+(@floatSuffixes)?/,{token:"number"}],[/\b(\d\.?[\d_]*)(@floatSuffixes)?\b/,{token:"number"}],[/(0x[\da-fA-F]+)_?(@intSuffixes)?/,{token:"number"}],[/[\d][\d_]*(@intSuffixes?)?/,{token:"number"}]]}},Tje=Object.freeze(Object.defineProperty({__proto__:null,conf:Ije,language:Dje},Symbol.toStringTag,{value:"Module"})),Nje={comments:{lineComment:"'"},brackets:[["(",")"],["[","]"],["If","EndIf"],["While","EndWhile"],["For","EndFor"],["Sub","EndSub"]],autoClosingPairs:[{open:'"',close:'"',notIn:["string","comment"]},{open:"(",close:")",notIn:["string","comment"]},{open:"[",close:"]",notIn:["string","comment"]}]},Rje={defaultToken:"",tokenPostfix:".sb",ignoreCase:!0,brackets:[{token:"delimiter.array",open:"[",close:"]"},{token:"delimiter.parenthesis",open:"(",close:")"},{token:"keyword.tag-if",open:"If",close:"EndIf"},{token:"keyword.tag-while",open:"While",close:"EndWhile"},{token:"keyword.tag-for",open:"For",close:"EndFor"},{token:"keyword.tag-sub",open:"Sub",close:"EndSub"}],keywords:["Else","ElseIf","EndFor","EndIf","EndSub","EndWhile","For","Goto","If","Step","Sub","Then","To","While"],tagwords:["If","Sub","While","For"],operators:[">","<","<>","<=",">=","And","Or","+","-","*","/","="],identifier:/[a-zA-Z_][\w]*/,symbols:/[=><:+\-*\/%\.,]+/,escapes:/\\(?:[abfnrtv\\"']|x[0-9A-Fa-f]{1,4}|u[0-9A-Fa-f]{4}|U[0-9A-Fa-f]{8})/,tokenizer:{root:[{include:"@whitespace"},[/(@identifier)(?=[.])/,"type"],[/@identifier/,{cases:{"@keywords":{token:"keyword.$0"},"@operators":"operator","@default":"variable.name"}}],[/([.])(@identifier)/,{cases:{$2:["delimiter","type.member"],"@default":""}}],[/\d*\.\d+/,"number.float"],[/\d+/,"number"],[/[()\[\]]/,"@brackets"],[/@symbols/,{cases:{"@operators":"operator","@default":"delimiter"}}],[/"([^"\\]|\\.)*$/,"string.invalid"],[/"/,"string","@string"]],whitespace:[[/[ \t\r\n]+/,""],[/(\').*$/,"comment"]],string:[[/[^\\"]+/,"string"],[/@escapes/,"string.escape"],[/\\./,"string.escape.invalid"],[/"C?/,"string","@pop"]]}},Aje=Object.freeze(Object.defineProperty({__proto__:null,conf:Nje,language:Rje},Symbol.toStringTag,{value:"Module"})),Mje={wordPattern:/(unary_[@~!#%^&*()\-=+\\|:<>\/?]+)|([a-zA-Z_$][\w$]*?_=)|(`[^`]+`)|([a-zA-Z_$][\w$]*)/g,comments:{lineComment:"//",blockComment:["/*","*/"]},brackets:[["{","}"],["[","]"],["(",")"]],autoClosingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:'"',close:'"'},{open:"'",close:"'"}],surroundingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:'"',close:'"'},{open:"'",close:"'"}],folding:{markers:{start:new RegExp("^\\s*//\\s*(?:(?:#?region\\b)|(?:<editor-fold\\b))"),end:new RegExp("^\\s*//\\s*(?:(?:#?endregion\\b)|(?:</editor-fold>))")}}},Oje={tokenPostfix:".scala",keywords:["asInstanceOf","catch","class","classOf","def","do","else","extends","finally","for","foreach","forSome","if","import","isInstanceOf","macro","match","new","object","package","return","throw","trait","try","type","until","val","var","while","with","yield","given","enum","then"],softKeywords:["as","export","extension","end","derives","on"],constants:["true","false","null","this","super"],modifiers:["abstract","final","implicit","lazy","override","private","protected","sealed"],softModifiers:["inline","opaque","open","transparent","using"],name:/(?:[a-z_$][\w$]*|`[^`]+`)/,type:/(?:[A-Z][\w$]*)/,symbols:/[=><!~?:&|+\-*\/^\\%@#]+/,digits:/\d+(_+\d+)*/,hexdigits:/[[0-9a-fA-F]+(_+[0-9a-fA-F]+)*/,escapes:/\\(?:[btnfr\\"']|x[0-9A-Fa-f]{1,4}|u[0-9A-Fa-f]{4}|U[0-9A-Fa-f]{8})/,fstring_conv:/[bBhHsScCdoxXeEfgGaAt]|[Tn](?:[HIklMSLNpzZsQ]|[BbhAaCYyjmde]|[RTrDFC])/,tokenizer:{root:[[/\braw"""/,{token:"string.quote",bracket:"@open",next:"@rawstringt"}],[/\braw"/,{token:"string.quote",bracket:"@open",next:"@rawstring"}],[/\bs"""/,{token:"string.quote",bracket:"@open",next:"@sstringt"}],[/\bs"/,{token:"string.quote",bracket:"@open",next:"@sstring"}],[/\bf""""/,{token:"string.quote",bracket:"@open",next:"@fstringt"}],[/\bf"/,{token:"string.quote",bracket:"@open",next:"@fstring"}],[/"""/,{token:"string.quote",bracket:"@open",next:"@stringt"}],[/"/,{token:"string.quote",bracket:"@open",next:"@string"}],[/(@digits)[eE]([\-+]?(@digits))?[fFdD]?/,"number.float","@allowMethod"],[/(@digits)\.(@digits)([eE][\-+]?(@digits))?[fFdD]?/,"number.float","@allowMethod"],[/0[xX](@hexdigits)[Ll]?/,"number.hex","@allowMethod"],[/(@digits)[fFdD]/,"number.float","@allowMethod"],[/(@digits)[lL]?/,"number","@allowMethod"],[/\b_\*/,"key"],[/\b(_)\b/,"keyword","@allowMethod"],[/\bimport\b/,"keyword","@import"],[/\b(case)([ \t]+)(class)\b/,["keyword.modifier","white","keyword"]],[/\bcase\b/,"keyword","@case"],[/\bva[lr]\b/,"keyword","@vardef"],[/\b(def)([ \t]+)((?:unary_)?@symbols|@name(?:_=)|@name)/,["keyword","white","identifier"]],[/@name(?=[ \t]*:(?!:))/,"variable"],[/(\.)(@name|@symbols)/,["operator",{token:"@rematch",next:"@allowMethod"}]],[/([{(])(\s*)(@name(?=\s*=>))/,["@brackets","white","variable"]],[/@name/,{cases:{"@keywords":"keyword","@softKeywords":"keyword","@modifiers":"keyword.modifier","@softModifiers":"keyword.modifier","@constants":{token:"constant",next:"@allowMethod"},"@default":{token:"identifier",next:"@allowMethod"}}}],[/@type/,"type","@allowMethod"],{include:"@whitespace"},[/@[a-zA-Z_$][\w$]*(?:\.[a-zA-Z_$][\w$]*)*/,"annotation"],[/[{(]/,"@brackets"],[/[})]/,"@brackets","@allowMethod"],[/\[/,"operator.square"],[/](?!\s*(?:va[rl]|def|type)\b)/,"operator.square","@allowMethod"],[/]/,"operator.square"],[/([=-]>|<-|>:|<:|:>|<%)(?=[\s\w()[\]{},\."'`])/,"keyword"],[/@symbols/,"operator"],[/[;,\.]/,"delimiter"],[/'[a-zA-Z$][\w$]*(?!')/,"attribute.name"],[/'[^\\']'/,"string","@allowMethod"],[/(')(@escapes)(')/,["string","string.escape",{token:"string",next:"@allowMethod"}]],[/'/,"string.invalid"]],import:[[/;/,"delimiter","@pop"],[/^|$/,"","@pop"],[/[ \t]+/,"white"],[/[\n\r]+/,"white","@pop"],[/\/\*/,"comment","@comment"],[/@name|@type/,"type"],[/[(){}]/,"@brackets"],[/[[\]]/,"operator.square"],[/[\.,]/,"delimiter"]],allowMethod:[[/^|$/,"","@pop"],[/[ \t]+/,"white"],[/[\n\r]+/,"white","@pop"],[/\/\*/,"comment","@comment"],[/(?==>[\s\w([{])/,"keyword","@pop"],[/(@name|@symbols)(?=[ \t]*[[({"'`]|[ \t]+(?:[+-]?\.?\d|\w))/,{cases:{"@keywords":{token:"keyword",next:"@pop"},"->|<-|>:|<:|<%":{token:"keyword",next:"@pop"},"@default":{token:"@rematch",next:"@pop"}}}],["","","@pop"]],comment:[[/[^\/*]+/,"comment"],[/\/\*/,"comment","@push"],[/\*\//,"comment","@pop"],[/[\/*]/,"comment"]],case:[[/\b_\*/,"key"],[/\b(_|true|false|null|this|super)\b/,"keyword","@allowMethod"],[/\bif\b|=>/,"keyword","@pop"],[/`[^`]+`/,"identifier","@allowMethod"],[/@name/,"variable","@allowMethod"],[/:::?|\||@(?![a-z_$])/,"keyword"],{include:"@root"}],vardef:[[/\b_\*/,"key"],[/\b(_|true|false|null|this|super)\b/,"keyword"],[/@name/,"variable"],[/:::?|\||@(?![a-z_$])/,"keyword"],[/=|:(?!:)/,"operator","@pop"],[/$/,"white","@pop"],{include:"@root"}],string:[[/[^\\"\n\r]+/,"string"],[/@escapes/,"string.escape"],[/\\./,"string.escape.invalid"],[/"/,{token:"string.quote",bracket:"@close",switchTo:"@allowMethod"}]],stringt:[[/[^\\"\n\r]+/,"string"],[/@escapes/,"string.escape"],[/\\./,"string.escape.invalid"],[/"(?=""")/,"string"],[/"""/,{token:"string.quote",bracket:"@close",switchTo:"@allowMethod"}],[/"/,"string"]],fstring:[[/@escapes/,"string.escape"],[/"/,{token:"string.quote",bracket:"@close",switchTo:"@allowMethod"}],[/\$\$/,"string"],[/(\$)([a-z_]\w*)/,["operator","identifier"]],[/\$\{/,"operator","@interp"],[/%%/,"string"],[/(%)([\-#+ 0,(])(\d+|\.\d+|\d+\.\d+)(@fstring_conv)/,["metatag","keyword.modifier","number","metatag"]],[/(%)(\d+|\.\d+|\d+\.\d+)(@fstring_conv)/,["metatag","number","metatag"]],[/(%)([\-#+ 0,(])(@fstring_conv)/,["metatag","keyword.modifier","metatag"]],[/(%)(@fstring_conv)/,["metatag","metatag"]],[/./,"string"]],fstringt:[[/@escapes/,"string.escape"],[/"(?=""")/,"string"],[/"""/,{token:"string.quote",bracket:"@close",switchTo:"@allowMethod"}],[/\$\$/,"string"],[/(\$)([a-z_]\w*)/,["operator","identifier"]],[/\$\{/,"operator","@interp"],[/%%/,"string"],[/(%)([\-#+ 0,(])(\d+|\.\d+|\d+\.\d+)(@fstring_conv)/,["metatag","keyword.modifier","number","metatag"]],[/(%)(\d+|\.\d+|\d+\.\d+)(@fstring_conv)/,["metatag","number","metatag"]],[/(%)([\-#+ 0,(])(@fstring_conv)/,["metatag","keyword.modifier","metatag"]],[/(%)(@fstring_conv)/,["metatag","metatag"]],[/./,"string"]],sstring:[[/@escapes/,"string.escape"],[/"/,{token:"string.quote",bracket:"@close",switchTo:"@allowMethod"}],[/\$\$/,"string"],[/(\$)([a-z_]\w*)/,["operator","identifier"]],[/\$\{/,"operator","@interp"],[/./,"string"]],sstringt:[[/@escapes/,"string.escape"],[/"(?=""")/,"string"],[/"""/,{token:"string.quote",bracket:"@close",switchTo:"@allowMethod"}],[/\$\$/,"string"],[/(\$)([a-z_]\w*)/,["operator","identifier"]],[/\$\{/,"operator","@interp"],[/./,"string"]],interp:[[/{/,"operator","@push"],[/}/,"operator","@pop"],{include:"@root"}],rawstring:[[/[^"]/,"string"],[/"/,{token:"string.quote",bracket:"@close",switchTo:"@allowMethod"}]],rawstringt:[[/[^"]/,"string"],[/"(?=""")/,"string"],[/"""/,{token:"string.quote",bracket:"@close",switchTo:"@allowMethod"}],[/"/,"string"]],whitespace:[[/[ \t\r\n]+/,"white"],[/\/\*/,"comment","@comment"],[/\/\/.*$/,"comment"]]}},Pje=Object.freeze(Object.defineProperty({__proto__:null,conf:Mje,language:Oje},Symbol.toStringTag,{value:"Module"})),Fje={comments:{lineComment:";",blockComment:["#|","|#"]},brackets:[["(",")"],["{","}"],["[","]"]],autoClosingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:'"',close:'"'}],surroundingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:'"',close:'"'}]},Bje={defaultToken:"",ignoreCase:!0,tokenPostfix:".scheme",brackets:[{open:"(",close:")",token:"delimiter.parenthesis"},{open:"{",close:"}",token:"delimiter.curly"},{open:"[",close:"]",token:"delimiter.square"}],keywords:["case","do","let","loop","if","else","when","cons","car","cdr","cond","lambda","lambda*","syntax-rules","format","set!","quote","eval","append","list","list?","member?","load"],constants:["#t","#f"],operators:["eq?","eqv?","equal?","and","or","not","null?"],tokenizer:{root:[[/#[xXoObB][0-9a-fA-F]+/,"number.hex"],[/[+-]?\d+(?:(?:\.\d*)?(?:[eE][+-]?\d+)?)?/,"number.float"],[/(?:\b(?:(define|define-syntax|define-macro))\b)(\s+)((?:\w|\-|\!|\?)*)/,["keyword","white","variable"]],{include:"@whitespace"},{include:"@strings"},[/[a-zA-Z_#][a-zA-Z0-9_\-\?\!\*]*/,{cases:{"@keywords":"keyword","@constants":"constant","@operators":"operators","@default":"identifier"}}]],comment:[[/[^\|#]+/,"comment"],[/#\|/,"comment","@push"],[/\|#/,"comment","@pop"],[/[\|#]/,"comment"]],whitespace:[[/[ \t\r\n]+/,"white"],[/#\|/,"comment","@comment"],[/;.*$/,"comment"]],strings:[[/"$/,"string","@popall"],[/"(?=.)/,"string","@multiLineString"]],multiLineString:[[/[^\\"]+$/,"string","@popall"],[/[^\\"]+/,"string"],[/\\./,"string.escape"],[/"/,"string","@popall"],[/\\$/,"string"]]}},Wje=Object.freeze(Object.defineProperty({__proto__:null,conf:Fje,language:Bje},Symbol.toStringTag,{value:"Module"})),Hje={wordPattern:/(#?-?\d*\.\d\w*%?)|([@$#!.:]?[\w-?]+%?)|[@#!.]/g,comments:{blockComment:["/*","*/"],lineComment:"//"},brackets:[["{","}"],["[","]"],["(",")"]],autoClosingPairs:[{open:"{",close:"}",notIn:["string","comment"]},{open:"[",close:"]",notIn:["string","comment"]},{open:"(",close:")",notIn:["string","comment"]},{open:'"',close:'"',notIn:["string","comment"]},{open:"'",close:"'",notIn:["string","comment"]}],surroundingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:'"',close:'"'},{open:"'",close:"'"}],folding:{markers:{start:new RegExp("^\\s*\\/\\*\\s*#region\\b\\s*(.*?)\\s*\\*\\/"),end:new RegExp("^\\s*\\/\\*\\s*#endregion\\b.*\\*\\/")}}},Vje={defaultToken:"",tokenPostfix:".scss",ws:`[ 	
\r\f]*`,identifier:"-?-?([a-zA-Z]|(\\\\(([0-9a-fA-F]{1,6}\\s?)|[^[0-9a-fA-F])))([\\w\\-]|(\\\\(([0-9a-fA-F]{1,6}\\s?)|[^[0-9a-fA-F])))*",brackets:[{open:"{",close:"}",token:"delimiter.curly"},{open:"[",close:"]",token:"delimiter.bracket"},{open:"(",close:")",token:"delimiter.parenthesis"},{open:"<",close:">",token:"delimiter.angle"}],tokenizer:{root:[{include:"@selector"}],selector:[{include:"@comments"},{include:"@import"},{include:"@variabledeclaration"},{include:"@warndebug"},["[@](include)",{token:"keyword",next:"@includedeclaration"}],["[@](keyframes|-webkit-keyframes|-moz-keyframes|-o-keyframes)",{token:"keyword",next:"@keyframedeclaration"}],["[@](page|content|font-face|-moz-document)",{token:"keyword"}],["[@](charset|namespace)",{token:"keyword",next:"@declarationbody"}],["[@](function)",{token:"keyword",next:"@functiondeclaration"}],["[@](mixin)",{token:"keyword",next:"@mixindeclaration"}],["url(\\-prefix)?\\(",{token:"meta",next:"@urldeclaration"}],{include:"@controlstatement"},{include:"@selectorname"},["[&\\*]","tag"],["[>\\+,]","delimiter"],["\\[",{token:"delimiter.bracket",next:"@selectorattribute"}],["{",{token:"delimiter.curly",next:"@selectorbody"}]],selectorbody:[["[*_]?@identifier@ws:(?=(\\s|\\d|[^{;}]*[;}]))","attribute.name","@rulevalue"],{include:"@selector"},["[@](extend)",{token:"keyword",next:"@extendbody"}],["[@](return)",{token:"keyword",next:"@declarationbody"}],["}",{token:"delimiter.curly",next:"@pop"}]],selectorname:[["#{",{token:"meta",next:"@variableinterpolation"}],["(\\.|#(?=[^{])|%|(@identifier)|:)+","tag"]],selectorattribute:[{include:"@term"},["]",{token:"delimiter.bracket",next:"@pop"}]],term:[{include:"@comments"},["url(\\-prefix)?\\(",{token:"meta",next:"@urldeclaration"}],{include:"@functioninvocation"},{include:"@numbers"},{include:"@strings"},{include:"@variablereference"},["(and\\b|or\\b|not\\b)","operator"],{include:"@name"},["([<>=\\+\\-\\*\\/\\^\\|\\~,])","operator"],[",","delimiter"],["!default","literal"],["\\(",{token:"delimiter.parenthesis",next:"@parenthizedterm"}]],rulevalue:[{include:"@term"},["!important","literal"],[";","delimiter","@pop"],["{",{token:"delimiter.curly",switchTo:"@nestedproperty"}],["(?=})",{token:"",next:"@pop"}]],nestedproperty:[["[*_]?@identifier@ws:","attribute.name","@rulevalue"],{include:"@comments"},["}",{token:"delimiter.curly",next:"@pop"}]],warndebug:[["[@](warn|debug)",{token:"keyword",next:"@declarationbody"}]],import:[["[@](import)",{token:"keyword",next:"@declarationbody"}]],variabledeclaration:[["\\$@identifier@ws:","variable.decl","@declarationbody"]],urldeclaration:[{include:"@strings"},[`[^)\r
]+`,"string"],["\\)",{token:"meta",next:"@pop"}]],parenthizedterm:[{include:"@term"},["\\)",{token:"delimiter.parenthesis",next:"@pop"}]],declarationbody:[{include:"@term"},[";","delimiter","@pop"],["(?=})",{token:"",next:"@pop"}]],extendbody:[{include:"@selectorname"},["!optional","literal"],[";","delimiter","@pop"],["(?=})",{token:"",next:"@pop"}]],variablereference:[["\\$@identifier","variable.ref"],["\\.\\.\\.","operator"],["#{",{token:"meta",next:"@variableinterpolation"}]],variableinterpolation:[{include:"@variablereference"},["}",{token:"meta",next:"@pop"}]],comments:[["\\/\\*","comment","@comment"],["\\/\\/+.*","comment"]],comment:[["\\*\\/","comment","@pop"],[".","comment"]],name:[["@identifier","attribute.value"]],numbers:[["(\\d*\\.)?\\d+([eE][\\-+]?\\d+)?",{token:"number",next:"@units"}],["#[0-9a-fA-F_]+(?!\\w)","number.hex"]],units:[["(em|ex|ch|rem|fr|vmin|vmax|vw|vh|vm|cm|mm|in|px|pt|pc|deg|grad|rad|turn|s|ms|Hz|kHz|%)?","number","@pop"]],functiondeclaration:[["@identifier@ws\\(",{token:"meta",next:"@parameterdeclaration"}],["{",{token:"delimiter.curly",switchTo:"@functionbody"}]],mixindeclaration:[["@identifier@ws\\(",{token:"meta",next:"@parameterdeclaration"}],["@identifier","meta"],["{",{token:"delimiter.curly",switchTo:"@selectorbody"}]],parameterdeclaration:[["\\$@identifier@ws:","variable.decl"],["\\.\\.\\.","operator"],[",","delimiter"],{include:"@term"},["\\)",{token:"meta",next:"@pop"}]],includedeclaration:[{include:"@functioninvocation"},["@identifier","meta"],[";","delimiter","@pop"],["(?=})",{token:"",next:"@pop"}],["{",{token:"delimiter.curly",switchTo:"@selectorbody"}]],keyframedeclaration:[["@identifier","meta"],["{",{token:"delimiter.curly",switchTo:"@keyframebody"}]],keyframebody:[{include:"@term"},["{",{token:"delimiter.curly",next:"@selectorbody"}],["}",{token:"delimiter.curly",next:"@pop"}]],controlstatement:[["[@](if|else|for|while|each|media)",{token:"keyword.flow",next:"@controlstatementdeclaration"}]],controlstatementdeclaration:[["(in|from|through|if|to)\\b",{token:"keyword.flow"}],{include:"@term"},["{",{token:"delimiter.curly",switchTo:"@selectorbody"}]],functionbody:[["[@](return)",{token:"keyword"}],{include:"@variabledeclaration"},{include:"@term"},{include:"@controlstatement"},[";","delimiter"],["}",{token:"delimiter.curly",next:"@pop"}]],functioninvocation:[["@identifier\\(",{token:"meta",next:"@functionarguments"}]],functionarguments:[["\\$@identifier@ws:","attribute.name"],["[,]","delimiter"],{include:"@term"},["\\)",{token:"meta",next:"@pop"}]],strings:[['~?"',{token:"string.delimiter",next:"@stringenddoublequote"}],["~?'",{token:"string.delimiter",next:"@stringendquote"}]],stringenddoublequote:[["\\\\.","string"],['"',{token:"string.delimiter",next:"@pop"}],[".","string"]],stringendquote:[["\\\\.","string"],["'",{token:"string.delimiter",next:"@pop"}],[".","string"]]}},zje=Object.freeze(Object.defineProperty({__proto__:null,conf:Hje,language:Vje},Symbol.toStringTag,{value:"Module"})),$je={comments:{lineComment:"#"},brackets:[["{","}"],["[","]"],["(",")"]],autoClosingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:'"',close:'"'},{open:"'",close:"'"},{open:"`",close:"`"}],surroundingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:'"',close:'"'},{open:"'",close:"'"},{open:"`",close:"`"}]},Uje={defaultToken:"",ignoreCase:!0,tokenPostfix:".shell",brackets:[{token:"delimiter.bracket",open:"{",close:"}"},{token:"delimiter.parenthesis",open:"(",close:")"},{token:"delimiter.square",open:"[",close:"]"}],keywords:["if","then","do","else","elif","while","until","for","in","esac","fi","fin","fil","done","exit","set","unset","export","function"],builtins:["ab","awk","bash","beep","cat","cc","cd","chown","chmod","chroot","clear","cp","curl","cut","diff","echo","find","gawk","gcc","get","git","grep","hg","kill","killall","ln","ls","make","mkdir","openssl","mv","nc","node","npm","ping","ps","restart","rm","rmdir","sed","service","sh","shopt","shred","source","sort","sleep","ssh","start","stop","su","sudo","svn","tee","telnet","top","touch","vi","vim","wall","wc","wget","who","write","yes","zsh"],startingWithDash:/\-+\w+/,identifiersWithDashes:/[a-zA-Z]\w+(?:@startingWithDash)+/,symbols:/[=><!~?&|+\-*\/\^;\.,]+/,tokenizer:{root:[[/@identifiersWithDashes/,""],[/(\s)((?:@startingWithDash)+)/,["white","attribute.name"]],[/[a-zA-Z]\w*/,{cases:{"@keywords":"keyword","@builtins":"type.identifier","@default":""}}],{include:"@whitespace"},{include:"@strings"},{include:"@parameters"},{include:"@heredoc"},[/[{}\[\]()]/,"@brackets"],[/@symbols/,"delimiter"],{include:"@numbers"},[/[,;]/,"delimiter"]],whitespace:[[/\s+/,"white"],[/(^#!.*$)/,"metatag"],[/(^#.*$)/,"comment"]],numbers:[[/\d*\.\d+([eE][\-+]?\d+)?/,"number.float"],[/0[xX][0-9a-fA-F_]*[0-9a-fA-F]/,"number.hex"],[/\d+/,"number"]],strings:[[/'/,"string","@stringBody"],[/"/,"string","@dblStringBody"]],stringBody:[[/'/,"string","@popall"],[/./,"string"]],dblStringBody:[[/"/,"string","@popall"],[/./,"string"]],heredoc:[[/(<<[-<]?)(\s*)(['"`]?)([\w\-]+)(['"`]?)/,["constants","white","string.heredoc.delimiter","string.heredoc","string.heredoc.delimiter"]]],parameters:[[/\$\d+/,"variable.predefined"],[/\$\w+/,"variable"],[/\$[*@#?\-$!0_]/,"variable"],[/\$'/,"variable","@parameterBodyQuote"],[/\$"/,"variable","@parameterBodyDoubleQuote"],[/\$\(/,"variable","@parameterBodyParen"],[/\$\{/,"variable","@parameterBodyCurlyBrace"]],parameterBodyQuote:[[/[^#:%*@\-!_']+/,"variable"],[/[#:%*@\-!_]/,"delimiter"],[/[']/,"variable","@pop"]],parameterBodyDoubleQuote:[[/[^#:%*@\-!_"]+/,"variable"],[/[#:%*@\-!_]/,"delimiter"],[/["]/,"variable","@pop"]],parameterBodyParen:[[/[^#:%*@\-!_)]+/,"variable"],[/[#:%*@\-!_]/,"delimiter"],[/[)]/,"variable","@pop"]],parameterBodyCurlyBrace:[[/[^#:%*@\-!_}]+/,"variable"],[/[#:%*@\-!_]/,"delimiter"],[/[}]/,"variable","@pop"]]}},qje=Object.freeze(Object.defineProperty({__proto__:null,conf:$je,language:Uje},Symbol.toStringTag,{value:"Module"})),jje={comments:{lineComment:"//",blockComment:["/*","*/"]},brackets:[["{","}"],["[","]"],["(",")"],["<",">"]],autoClosingPairs:[{open:'"',close:'"',notIn:["string","comment"]},{open:"{",close:"}",notIn:["string","comment"]},{open:"[",close:"]",notIn:["string","comment"]},{open:"(",close:")",notIn:["string","comment"]}]},Gje={defaultToken:"",tokenPostfix:".sol",brackets:[{token:"delimiter.curly",open:"{",close:"}"},{token:"delimiter.parenthesis",open:"(",close:")"},{token:"delimiter.square",open:"[",close:"]"},{token:"delimiter.angle",open:"<",close:">"}],keywords:["pragma","solidity","contract","library","using","struct","function","modifier","constructor","address","string","bool","Int","Uint","Byte","Fixed","Ufixed","int","int8","int16","int24","int32","int40","int48","int56","int64","int72","int80","int88","int96","int104","int112","int120","int128","int136","int144","int152","int160","int168","int176","int184","int192","int200","int208","int216","int224","int232","int240","int248","int256","uint","uint8","uint16","uint24","uint32","uint40","uint48","uint56","uint64","uint72","uint80","uint88","uint96","uint104","uint112","uint120","uint128","uint136","uint144","uint152","uint160","uint168","uint176","uint184","uint192","uint200","uint208","uint216","uint224","uint232","uint240","uint248","uint256","byte","bytes","bytes1","bytes2","bytes3","bytes4","bytes5","bytes6","bytes7","bytes8","bytes9","bytes10","bytes11","bytes12","bytes13","bytes14","bytes15","bytes16","bytes17","bytes18","bytes19","bytes20","bytes21","bytes22","bytes23","bytes24","bytes25","bytes26","bytes27","bytes28","bytes29","bytes30","bytes31","bytes32","fixed","fixed0x8","fixed0x16","fixed0x24","fixed0x32","fixed0x40","fixed0x48","fixed0x56","fixed0x64","fixed0x72","fixed0x80","fixed0x88","fixed0x96","fixed0x104","fixed0x112","fixed0x120","fixed0x128","fixed0x136","fixed0x144","fixed0x152","fixed0x160","fixed0x168","fixed0x176","fixed0x184","fixed0x192","fixed0x200","fixed0x208","fixed0x216","fixed0x224","fixed0x232","fixed0x240","fixed0x248","fixed0x256","fixed8x8","fixed8x16","fixed8x24","fixed8x32","fixed8x40","fixed8x48","fixed8x56","fixed8x64","fixed8x72","fixed8x80","fixed8x88","fixed8x96","fixed8x104","fixed8x112","fixed8x120","fixed8x128","fixed8x136","fixed8x144","fixed8x152","fixed8x160","fixed8x168","fixed8x176","fixed8x184","fixed8x192","fixed8x200","fixed8x208","fixed8x216","fixed8x224","fixed8x232","fixed8x240","fixed8x248","fixed16x8","fixed16x16","fixed16x24","fixed16x32","fixed16x40","fixed16x48","fixed16x56","fixed16x64","fixed16x72","fixed16x80","fixed16x88","fixed16x96","fixed16x104","fixed16x112","fixed16x120","fixed16x128","fixed16x136","fixed16x144","fixed16x152","fixed16x160","fixed16x168","fixed16x176","fixed16x184","fixed16x192","fixed16x200","fixed16x208","fixed16x216","fixed16x224","fixed16x232","fixed16x240","fixed24x8","fixed24x16","fixed24x24","fixed24x32","fixed24x40","fixed24x48","fixed24x56","fixed24x64","fixed24x72","fixed24x80","fixed24x88","fixed24x96","fixed24x104","fixed24x112","fixed24x120","fixed24x128","fixed24x136","fixed24x144","fixed24x152","fixed24x160","fixed24x168","fixed24x176","fixed24x184","fixed24x192","fixed24x200","fixed24x208","fixed24x216","fixed24x224","fixed24x232","fixed32x8","fixed32x16","fixed32x24","fixed32x32","fixed32x40","fixed32x48","fixed32x56","fixed32x64","fixed32x72","fixed32x80","fixed32x88","fixed32x96","fixed32x104","fixed32x112","fixed32x120","fixed32x128","fixed32x136","fixed32x144","fixed32x152","fixed32x160","fixed32x168","fixed32x176","fixed32x184","fixed32x192","fixed32x200","fixed32x208","fixed32x216","fixed32x224","fixed40x8","fixed40x16","fixed40x24","fixed40x32","fixed40x40","fixed40x48","fixed40x56","fixed40x64","fixed40x72","fixed40x80","fixed40x88","fixed40x96","fixed40x104","fixed40x112","fixed40x120","fixed40x128","fixed40x136","fixed40x144","fixed40x152","fixed40x160","fixed40x168","fixed40x176","fixed40x184","fixed40x192","fixed40x200","fixed40x208","fixed40x216","fixed48x8","fixed48x16","fixed48x24","fixed48x32","fixed48x40","fixed48x48","fixed48x56","fixed48x64","fixed48x72","fixed48x80","fixed48x88","fixed48x96","fixed48x104","fixed48x112","fixed48x120","fixed48x128","fixed48x136","fixed48x144","fixed48x152","fixed48x160","fixed48x168","fixed48x176","fixed48x184","fixed48x192","fixed48x200","fixed48x208","fixed56x8","fixed56x16","fixed56x24","fixed56x32","fixed56x40","fixed56x48","fixed56x56","fixed56x64","fixed56x72","fixed56x80","fixed56x88","fixed56x96","fixed56x104","fixed56x112","fixed56x120","fixed56x128","fixed56x136","fixed56x144","fixed56x152","fixed56x160","fixed56x168","fixed56x176","fixed56x184","fixed56x192","fixed56x200","fixed64x8","fixed64x16","fixed64x24","fixed64x32","fixed64x40","fixed64x48","fixed64x56","fixed64x64","fixed64x72","fixed64x80","fixed64x88","fixed64x96","fixed64x104","fixed64x112","fixed64x120","fixed64x128","fixed64x136","fixed64x144","fixed64x152","fixed64x160","fixed64x168","fixed64x176","fixed64x184","fixed64x192","fixed72x8","fixed72x16","fixed72x24","fixed72x32","fixed72x40","fixed72x48","fixed72x56","fixed72x64","fixed72x72","fixed72x80","fixed72x88","fixed72x96","fixed72x104","fixed72x112","fixed72x120","fixed72x128","fixed72x136","fixed72x144","fixed72x152","fixed72x160","fixed72x168","fixed72x176","fixed72x184","fixed80x8","fixed80x16","fixed80x24","fixed80x32","fixed80x40","fixed80x48","fixed80x56","fixed80x64","fixed80x72","fixed80x80","fixed80x88","fixed80x96","fixed80x104","fixed80x112","fixed80x120","fixed80x128","fixed80x136","fixed80x144","fixed80x152","fixed80x160","fixed80x168","fixed80x176","fixed88x8","fixed88x16","fixed88x24","fixed88x32","fixed88x40","fixed88x48","fixed88x56","fixed88x64","fixed88x72","fixed88x80","fixed88x88","fixed88x96","fixed88x104","fixed88x112","fixed88x120","fixed88x128","fixed88x136","fixed88x144","fixed88x152","fixed88x160","fixed88x168","fixed96x8","fixed96x16","fixed96x24","fixed96x32","fixed96x40","fixed96x48","fixed96x56","fixed96x64","fixed96x72","fixed96x80","fixed96x88","fixed96x96","fixed96x104","fixed96x112","fixed96x120","fixed96x128","fixed96x136","fixed96x144","fixed96x152","fixed96x160","fixed104x8","fixed104x16","fixed104x24","fixed104x32","fixed104x40","fixed104x48","fixed104x56","fixed104x64","fixed104x72","fixed104x80","fixed104x88","fixed104x96","fixed104x104","fixed104x112","fixed104x120","fixed104x128","fixed104x136","fixed104x144","fixed104x152","fixed112x8","fixed112x16","fixed112x24","fixed112x32","fixed112x40","fixed112x48","fixed112x56","fixed112x64","fixed112x72","fixed112x80","fixed112x88","fixed112x96","fixed112x104","fixed112x112","fixed112x120","fixed112x128","fixed112x136","fixed112x144","fixed120x8","fixed120x16","fixed120x24","fixed120x32","fixed120x40","fixed120x48","fixed120x56","fixed120x64","fixed120x72","fixed120x80","fixed120x88","fixed120x96","fixed120x104","fixed120x112","fixed120x120","fixed120x128","fixed120x136","fixed128x8","fixed128x16","fixed128x24","fixed128x32","fixed128x40","fixed128x48","fixed128x56","fixed128x64","fixed128x72","fixed128x80","fixed128x88","fixed128x96","fixed128x104","fixed128x112","fixed128x120","fixed128x128","fixed136x8","fixed136x16","fixed136x24","fixed136x32","fixed136x40","fixed136x48","fixed136x56","fixed136x64","fixed136x72","fixed136x80","fixed136x88","fixed136x96","fixed136x104","fixed136x112","fixed136x120","fixed144x8","fixed144x16","fixed144x24","fixed144x32","fixed144x40","fixed144x48","fixed144x56","fixed144x64","fixed144x72","fixed144x80","fixed144x88","fixed144x96","fixed144x104","fixed144x112","fixed152x8","fixed152x16","fixed152x24","fixed152x32","fixed152x40","fixed152x48","fixed152x56","fixed152x64","fixed152x72","fixed152x80","fixed152x88","fixed152x96","fixed152x104","fixed160x8","fixed160x16","fixed160x24","fixed160x32","fixed160x40","fixed160x48","fixed160x56","fixed160x64","fixed160x72","fixed160x80","fixed160x88","fixed160x96","fixed168x8","fixed168x16","fixed168x24","fixed168x32","fixed168x40","fixed168x48","fixed168x56","fixed168x64","fixed168x72","fixed168x80","fixed168x88","fixed176x8","fixed176x16","fixed176x24","fixed176x32","fixed176x40","fixed176x48","fixed176x56","fixed176x64","fixed176x72","fixed176x80","fixed184x8","fixed184x16","fixed184x24","fixed184x32","fixed184x40","fixed184x48","fixed184x56","fixed184x64","fixed184x72","fixed192x8","fixed192x16","fixed192x24","fixed192x32","fixed192x40","fixed192x48","fixed192x56","fixed192x64","fixed200x8","fixed200x16","fixed200x24","fixed200x32","fixed200x40","fixed200x48","fixed200x56","fixed208x8","fixed208x16","fixed208x24","fixed208x32","fixed208x40","fixed208x48","fixed216x8","fixed216x16","fixed216x24","fixed216x32","fixed216x40","fixed224x8","fixed224x16","fixed224x24","fixed224x32","fixed232x8","fixed232x16","fixed232x24","fixed240x8","fixed240x16","fixed248x8","ufixed","ufixed0x8","ufixed0x16","ufixed0x24","ufixed0x32","ufixed0x40","ufixed0x48","ufixed0x56","ufixed0x64","ufixed0x72","ufixed0x80","ufixed0x88","ufixed0x96","ufixed0x104","ufixed0x112","ufixed0x120","ufixed0x128","ufixed0x136","ufixed0x144","ufixed0x152","ufixed0x160","ufixed0x168","ufixed0x176","ufixed0x184","ufixed0x192","ufixed0x200","ufixed0x208","ufixed0x216","ufixed0x224","ufixed0x232","ufixed0x240","ufixed0x248","ufixed0x256","ufixed8x8","ufixed8x16","ufixed8x24","ufixed8x32","ufixed8x40","ufixed8x48","ufixed8x56","ufixed8x64","ufixed8x72","ufixed8x80","ufixed8x88","ufixed8x96","ufixed8x104","ufixed8x112","ufixed8x120","ufixed8x128","ufixed8x136","ufixed8x144","ufixed8x152","ufixed8x160","ufixed8x168","ufixed8x176","ufixed8x184","ufixed8x192","ufixed8x200","ufixed8x208","ufixed8x216","ufixed8x224","ufixed8x232","ufixed8x240","ufixed8x248","ufixed16x8","ufixed16x16","ufixed16x24","ufixed16x32","ufixed16x40","ufixed16x48","ufixed16x56","ufixed16x64","ufixed16x72","ufixed16x80","ufixed16x88","ufixed16x96","ufixed16x104","ufixed16x112","ufixed16x120","ufixed16x128","ufixed16x136","ufixed16x144","ufixed16x152","ufixed16x160","ufixed16x168","ufixed16x176","ufixed16x184","ufixed16x192","ufixed16x200","ufixed16x208","ufixed16x216","ufixed16x224","ufixed16x232","ufixed16x240","ufixed24x8","ufixed24x16","ufixed24x24","ufixed24x32","ufixed24x40","ufixed24x48","ufixed24x56","ufixed24x64","ufixed24x72","ufixed24x80","ufixed24x88","ufixed24x96","ufixed24x104","ufixed24x112","ufixed24x120","ufixed24x128","ufixed24x136","ufixed24x144","ufixed24x152","ufixed24x160","ufixed24x168","ufixed24x176","ufixed24x184","ufixed24x192","ufixed24x200","ufixed24x208","ufixed24x216","ufixed24x224","ufixed24x232","ufixed32x8","ufixed32x16","ufixed32x24","ufixed32x32","ufixed32x40","ufixed32x48","ufixed32x56","ufixed32x64","ufixed32x72","ufixed32x80","ufixed32x88","ufixed32x96","ufixed32x104","ufixed32x112","ufixed32x120","ufixed32x128","ufixed32x136","ufixed32x144","ufixed32x152","ufixed32x160","ufixed32x168","ufixed32x176","ufixed32x184","ufixed32x192","ufixed32x200","ufixed32x208","ufixed32x216","ufixed32x224","ufixed40x8","ufixed40x16","ufixed40x24","ufixed40x32","ufixed40x40","ufixed40x48","ufixed40x56","ufixed40x64","ufixed40x72","ufixed40x80","ufixed40x88","ufixed40x96","ufixed40x104","ufixed40x112","ufixed40x120","ufixed40x128","ufixed40x136","ufixed40x144","ufixed40x152","ufixed40x160","ufixed40x168","ufixed40x176","ufixed40x184","ufixed40x192","ufixed40x200","ufixed40x208","ufixed40x216","ufixed48x8","ufixed48x16","ufixed48x24","ufixed48x32","ufixed48x40","ufixed48x48","ufixed48x56","ufixed48x64","ufixed48x72","ufixed48x80","ufixed48x88","ufixed48x96","ufixed48x104","ufixed48x112","ufixed48x120","ufixed48x128","ufixed48x136","ufixed48x144","ufixed48x152","ufixed48x160","ufixed48x168","ufixed48x176","ufixed48x184","ufixed48x192","ufixed48x200","ufixed48x208","ufixed56x8","ufixed56x16","ufixed56x24","ufixed56x32","ufixed56x40","ufixed56x48","ufixed56x56","ufixed56x64","ufixed56x72","ufixed56x80","ufixed56x88","ufixed56x96","ufixed56x104","ufixed56x112","ufixed56x120","ufixed56x128","ufixed56x136","ufixed56x144","ufixed56x152","ufixed56x160","ufixed56x168","ufixed56x176","ufixed56x184","ufixed56x192","ufixed56x200","ufixed64x8","ufixed64x16","ufixed64x24","ufixed64x32","ufixed64x40","ufixed64x48","ufixed64x56","ufixed64x64","ufixed64x72","ufixed64x80","ufixed64x88","ufixed64x96","ufixed64x104","ufixed64x112","ufixed64x120","ufixed64x128","ufixed64x136","ufixed64x144","ufixed64x152","ufixed64x160","ufixed64x168","ufixed64x176","ufixed64x184","ufixed64x192","ufixed72x8","ufixed72x16","ufixed72x24","ufixed72x32","ufixed72x40","ufixed72x48","ufixed72x56","ufixed72x64","ufixed72x72","ufixed72x80","ufixed72x88","ufixed72x96","ufixed72x104","ufixed72x112","ufixed72x120","ufixed72x128","ufixed72x136","ufixed72x144","ufixed72x152","ufixed72x160","ufixed72x168","ufixed72x176","ufixed72x184","ufixed80x8","ufixed80x16","ufixed80x24","ufixed80x32","ufixed80x40","ufixed80x48","ufixed80x56","ufixed80x64","ufixed80x72","ufixed80x80","ufixed80x88","ufixed80x96","ufixed80x104","ufixed80x112","ufixed80x120","ufixed80x128","ufixed80x136","ufixed80x144","ufixed80x152","ufixed80x160","ufixed80x168","ufixed80x176","ufixed88x8","ufixed88x16","ufixed88x24","ufixed88x32","ufixed88x40","ufixed88x48","ufixed88x56","ufixed88x64","ufixed88x72","ufixed88x80","ufixed88x88","ufixed88x96","ufixed88x104","ufixed88x112","ufixed88x120","ufixed88x128","ufixed88x136","ufixed88x144","ufixed88x152","ufixed88x160","ufixed88x168","ufixed96x8","ufixed96x16","ufixed96x24","ufixed96x32","ufixed96x40","ufixed96x48","ufixed96x56","ufixed96x64","ufixed96x72","ufixed96x80","ufixed96x88","ufixed96x96","ufixed96x104","ufixed96x112","ufixed96x120","ufixed96x128","ufixed96x136","ufixed96x144","ufixed96x152","ufixed96x160","ufixed104x8","ufixed104x16","ufixed104x24","ufixed104x32","ufixed104x40","ufixed104x48","ufixed104x56","ufixed104x64","ufixed104x72","ufixed104x80","ufixed104x88","ufixed104x96","ufixed104x104","ufixed104x112","ufixed104x120","ufixed104x128","ufixed104x136","ufixed104x144","ufixed104x152","ufixed112x8","ufixed112x16","ufixed112x24","ufixed112x32","ufixed112x40","ufixed112x48","ufixed112x56","ufixed112x64","ufixed112x72","ufixed112x80","ufixed112x88","ufixed112x96","ufixed112x104","ufixed112x112","ufixed112x120","ufixed112x128","ufixed112x136","ufixed112x144","ufixed120x8","ufixed120x16","ufixed120x24","ufixed120x32","ufixed120x40","ufixed120x48","ufixed120x56","ufixed120x64","ufixed120x72","ufixed120x80","ufixed120x88","ufixed120x96","ufixed120x104","ufixed120x112","ufixed120x120","ufixed120x128","ufixed120x136","ufixed128x8","ufixed128x16","ufixed128x24","ufixed128x32","ufixed128x40","ufixed128x48","ufixed128x56","ufixed128x64","ufixed128x72","ufixed128x80","ufixed128x88","ufixed128x96","ufixed128x104","ufixed128x112","ufixed128x120","ufixed128x128","ufixed136x8","ufixed136x16","ufixed136x24","ufixed136x32","ufixed136x40","ufixed136x48","ufixed136x56","ufixed136x64","ufixed136x72","ufixed136x80","ufixed136x88","ufixed136x96","ufixed136x104","ufixed136x112","ufixed136x120","ufixed144x8","ufixed144x16","ufixed144x24","ufixed144x32","ufixed144x40","ufixed144x48","ufixed144x56","ufixed144x64","ufixed144x72","ufixed144x80","ufixed144x88","ufixed144x96","ufixed144x104","ufixed144x112","ufixed152x8","ufixed152x16","ufixed152x24","ufixed152x32","ufixed152x40","ufixed152x48","ufixed152x56","ufixed152x64","ufixed152x72","ufixed152x80","ufixed152x88","ufixed152x96","ufixed152x104","ufixed160x8","ufixed160x16","ufixed160x24","ufixed160x32","ufixed160x40","ufixed160x48","ufixed160x56","ufixed160x64","ufixed160x72","ufixed160x80","ufixed160x88","ufixed160x96","ufixed168x8","ufixed168x16","ufixed168x24","ufixed168x32","ufixed168x40","ufixed168x48","ufixed168x56","ufixed168x64","ufixed168x72","ufixed168x80","ufixed168x88","ufixed176x8","ufixed176x16","ufixed176x24","ufixed176x32","ufixed176x40","ufixed176x48","ufixed176x56","ufixed176x64","ufixed176x72","ufixed176x80","ufixed184x8","ufixed184x16","ufixed184x24","ufixed184x32","ufixed184x40","ufixed184x48","ufixed184x56","ufixed184x64","ufixed184x72","ufixed192x8","ufixed192x16","ufixed192x24","ufixed192x32","ufixed192x40","ufixed192x48","ufixed192x56","ufixed192x64","ufixed200x8","ufixed200x16","ufixed200x24","ufixed200x32","ufixed200x40","ufixed200x48","ufixed200x56","ufixed208x8","ufixed208x16","ufixed208x24","ufixed208x32","ufixed208x40","ufixed208x48","ufixed216x8","ufixed216x16","ufixed216x24","ufixed216x32","ufixed216x40","ufixed224x8","ufixed224x16","ufixed224x24","ufixed224x32","ufixed232x8","ufixed232x16","ufixed232x24","ufixed240x8","ufixed240x16","ufixed248x8","event","enum","let","mapping","private","public","external","inherited","payable","true","false","var","import","constant","if","else","for","else","for","while","do","break","continue","throw","returns","return","suicide","new","is","this","super"],operators:["=",">","<","!","~","?",":","==","<=",">=","!=","&&","||","++","--","+","-","*","/","&","|","^","%","<<",">>",">>>","+=","-=","*=","/=","&=","|=","^=","%=","<<=",">>=",">>>="],symbols:/[=><!~?:&|+\-*\/\^%]+/,escapes:/\\(?:[abfnrtv\\"']|x[0-9A-Fa-f]{1,4}|u[0-9A-Fa-f]{4}|U[0-9A-Fa-f]{8})/,integersuffix:/(ll|LL|u|U|l|L)?(ll|LL|u|U|l|L)?/,floatsuffix:/[fFlL]?/,tokenizer:{root:[[/[a-zA-Z_]\w*/,{cases:{"@keywords":{token:"keyword.$0"},"@default":"identifier"}}],{include:"@whitespace"},[/\[\[.*\]\]/,"annotation"],[/^\s*#\w+/,"keyword"],[/int\d*/,"keyword"],[/[{}()\[\]]/,"@brackets"],[/[<>](?!@symbols)/,"@brackets"],[/@symbols/,{cases:{"@operators":"delimiter","@default":""}}],[/\d*\d+[eE]([\-+]?\d+)?(@floatsuffix)/,"number.float"],[/\d*\.\d+([eE][\-+]?\d+)?(@floatsuffix)/,"number.float"],[/0[xX][0-9a-fA-F']*[0-9a-fA-F](@integersuffix)/,"number.hex"],[/0[0-7']*[0-7](@integersuffix)/,"number.octal"],[/0[bB][0-1']*[0-1](@integersuffix)/,"number.binary"],[/\d[\d']*\d(@integersuffix)/,"number"],[/\d(@integersuffix)/,"number"],[/[;,.]/,"delimiter"],[/"([^"\\]|\\.)*$/,"string.invalid"],[/"/,"string","@string"],[/'[^\\']'/,"string"],[/(')(@escapes)(')/,["string","string.escape","string"]],[/'/,"string.invalid"]],whitespace:[[/[ \t\r\n]+/,""],[/\/\*\*(?!\/)/,"comment.doc","@doccomment"],[/\/\*/,"comment","@comment"],[/\/\/.*$/,"comment"]],comment:[[/[^\/*]+/,"comment"],[/\*\//,"comment","@pop"],[/[\/*]/,"comment"]],doccomment:[[/[^\/*]+/,"comment.doc"],[/\*\//,"comment.doc","@pop"],[/[\/*]/,"comment.doc"]],string:[[/[^\\"]+/,"string"],[/@escapes/,"string.escape"],[/\\./,"string.escape.invalid"],[/"/,"string","@pop"]]}},Kje=Object.freeze(Object.defineProperty({__proto__:null,conf:jje,language:Gje},Symbol.toStringTag,{value:"Module"})),Yje={comments:{lineComment:"//",blockComment:["/*","*/"]},brackets:[["{","}"],["[","]"],["(",")"],["<",">"]],autoClosingPairs:[{open:'"',close:'"',notIn:["string","comment"]},{open:"{",close:"}",notIn:["string","comment"]},{open:"[",close:"]",notIn:["string","comment"]},{open:"(",close:")",notIn:["string","comment"]}]},Zje={defaultToken:"",tokenPostfix:".aes",brackets:[{token:"delimiter.curly",open:"{",close:"}"},{token:"delimiter.parenthesis",open:"(",close:")"},{token:"delimiter.square",open:"[",close:"]"},{token:"delimiter.angle",open:"<",close:">"}],keywords:["contract","library","entrypoint","function","stateful","state","hash","signature","tuple","list","address","string","bool","int","record","datatype","type","option","oracle","oracle_query","Call","Bits","Bytes","Oracle","String","Crypto","Address","Auth","Chain","None","Some","bits","bytes","event","let","map","private","public","true","false","var","if","else","throw"],operators:["=",">","<","!","~","?","::",":","==","<=",">=","!=","&&","||","++","--","+","-","*","/","&","|","^","%","<<",">>",">>>","+=","-=","*=","/=","&=","|=","^=","%=","<<=",">>=",">>>="],symbols:/[=><!~?:&|+\-*\/\^%]+/,escapes:/\\(?:[abfnrtv\\"']|x[0-9A-Fa-f]{1,4}|u[0-9A-Fa-f]{4}|U[0-9A-Fa-f]{8})/,integersuffix:/(ll|LL|u|U|l|L)?(ll|LL|u|U|l|L)?/,floatsuffix:/[fFlL]?/,tokenizer:{root:[[/[a-zA-Z_]\w*/,{cases:{"@keywords":{token:"keyword.$0"},"@default":"identifier"}}],{include:"@whitespace"},[/\[\[.*\]\]/,"annotation"],[/^\s*#\w+/,"keyword"],[/int\d*/,"keyword"],[/[{}()\[\]]/,"@brackets"],[/[<>](?!@symbols)/,"@brackets"],[/@symbols/,{cases:{"@operators":"delimiter","@default":""}}],[/\d*\d+[eE]([\-+]?\d+)?(@floatsuffix)/,"number.float"],[/\d*\.\d+([eE][\-+]?\d+)?(@floatsuffix)/,"number.float"],[/0[xX][0-9a-fA-F']*[0-9a-fA-F](@integersuffix)/,"number.hex"],[/0[0-7']*[0-7](@integersuffix)/,"number.octal"],[/0[bB][0-1']*[0-1](@integersuffix)/,"number.binary"],[/\d[\d']*\d(@integersuffix)/,"number"],[/\d(@integersuffix)/,"number"],[/[;,.]/,"delimiter"],[/"([^"\\]|\\.)*$/,"string.invalid"],[/"/,"string","@string"],[/'[^\\']'/,"string"],[/(')(@escapes)(')/,["string","string.escape","string"]],[/'/,"string.invalid"]],whitespace:[[/[ \t\r\n]+/,""],[/\/\*\*(?!\/)/,"comment.doc","@doccomment"],[/\/\*/,"comment","@comment"],[/\/\/.*$/,"comment"]],comment:[[/[^\/*]+/,"comment"],[/\*\//,"comment","@pop"],[/[\/*]/,"comment"]],doccomment:[[/[^\/*]+/,"comment.doc"],[/\*\//,"comment.doc","@pop"],[/[\/*]/,"comment.doc"]],string:[[/[^\\"]+/,"string"],[/@escapes/,"string.escape"],[/\\./,"string.escape.invalid"],[/"/,"string","@pop"]]}},Xje=Object.freeze(Object.defineProperty({__proto__:null,conf:Yje,language:Zje},Symbol.toStringTag,{value:"Module"})),Qje={comments:{lineComment:"#"},brackets:[["{","}"],["[","]"],["(",")"]],autoClosingPairs:[{open:"'",close:"'",notIn:["string"]},{open:'"',close:'"',notIn:["string"]},{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"}]},Jje={defaultToken:"",tokenPostfix:".rq",brackets:[{token:"delimiter.curly",open:"{",close:"}"},{token:"delimiter.parenthesis",open:"(",close:")"},{token:"delimiter.square",open:"[",close:"]"},{token:"delimiter.angle",open:"<",close:">"}],keywords:["add","as","asc","ask","base","by","clear","construct","copy","create","data","delete","desc","describe","distinct","drop","false","filter","from","graph","group","having","in","insert","limit","load","minus","move","named","not","offset","optional","order","prefix","reduced","select","service","silent","to","true","undef","union","using","values","where","with"],builtinFunctions:["a","abs","avg","bind","bnode","bound","ceil","coalesce","concat","contains","count","datatype","day","encode_for_uri","exists","floor","group_concat","hours","if","iri","isblank","isiri","isliteral","isnumeric","isuri","lang","langmatches","lcase","max","md5","min","minutes","month","now","rand","regex","replace","round","sameterm","sample","seconds","sha1","sha256","sha384","sha512","str","strafter","strbefore","strdt","strends","strlang","strlen","strstarts","struuid","substr","sum","timezone","tz","ucase","uri","uuid","year"],ignoreCase:!0,tokenizer:{root:[[/<[^\s\u00a0>]*>?/,"tag"],{include:"@strings"},[/#.*/,"comment"],[/[{}()\[\]]/,"@brackets"],[/[;,.]/,"delimiter"],[/[_\w\d]+:(\.(?=[\w_\-\\%])|[:\w_-]|\\[-\\_~.!$&'()*+,;=/?#@%]|%[a-f\d][a-f\d])*/,"tag"],[/:(\.(?=[\w_\-\\%])|[:\w_-]|\\[-\\_~.!$&'()*+,;=/?#@%]|%[a-f\d][a-f\d])+/,"tag"],[/[$?]?[_\w\d]+/,{cases:{"@keywords":{token:"keyword"},"@builtinFunctions":{token:"predefined.sql"},"@default":"identifier"}}],[/\^\^/,"operator.sql"],[/\^[*+\-<>=&|^\/!?]*/,"operator.sql"],[/[*+\-<>=&|\/!?]/,"operator.sql"],[/@[a-z\d\-]*/,"metatag.html"],[/\s+/,"white"]],strings:[[/'([^'\\]|\\.)*$/,"string.invalid"],[/'$/,"string.sql","@pop"],[/'/,"string.sql","@stringBody"],[/"([^"\\]|\\.)*$/,"string.invalid"],[/"$/,"string.sql","@pop"],[/"/,"string.sql","@dblStringBody"]],stringBody:[[/[^\\']+/,"string.sql"],[/\\./,"string.escape"],[/'/,"string.sql","@pop"]],dblStringBody:[[/[^\\"]+/,"string.sql"],[/\\./,"string.escape"],[/"/,"string.sql","@pop"]]}},eGe=Object.freeze(Object.defineProperty({__proto__:null,conf:Qje,language:Jje},Symbol.toStringTag,{value:"Module"})),tGe={comments:{lineComment:"--",blockComment:["/*","*/"]},brackets:[["{","}"],["[","]"],["(",")"]],autoClosingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:'"',close:'"'},{open:"'",close:"'"}],surroundingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:'"',close:'"'},{open:"'",close:"'"}]},iGe={defaultToken:"",tokenPostfix:".sql",ignoreCase:!0,brackets:[{open:"[",close:"]",token:"delimiter.square"},{open:"(",close:")",token:"delimiter.parenthesis"}],keywords:["ABORT","ABSOLUTE","ACTION","ADA","ADD","AFTER","ALL","ALLOCATE","ALTER","ALWAYS","ANALYZE","AND","ANY","ARE","AS","ASC","ASSERTION","AT","ATTACH","AUTHORIZATION","AUTOINCREMENT","AVG","BACKUP","BEFORE","BEGIN","BETWEEN","BIT","BIT_LENGTH","BOTH","BREAK","BROWSE","BULK","BY","CASCADE","CASCADED","CASE","CAST","CATALOG","CHAR","CHARACTER","CHARACTER_LENGTH","CHAR_LENGTH","CHECK","CHECKPOINT","CLOSE","CLUSTERED","COALESCE","COLLATE","COLLATION","COLUMN","COMMIT","COMPUTE","CONFLICT","CONNECT","CONNECTION","CONSTRAINT","CONSTRAINTS","CONTAINS","CONTAINSTABLE","CONTINUE","CONVERT","CORRESPONDING","COUNT","CREATE","CROSS","CURRENT","CURRENT_DATE","CURRENT_TIME","CURRENT_TIMESTAMP","CURRENT_USER","CURSOR","DATABASE","DATE","DAY","DBCC","DEALLOCATE","DEC","DECIMAL","DECLARE","DEFAULT","DEFERRABLE","DEFERRED","DELETE","DENY","DESC","DESCRIBE","DESCRIPTOR","DETACH","DIAGNOSTICS","DISCONNECT","DISK","DISTINCT","DISTRIBUTED","DO","DOMAIN","DOUBLE","DROP","DUMP","EACH","ELSE","END","END-EXEC","ERRLVL","ESCAPE","EXCEPT","EXCEPTION","EXCLUDE","EXCLUSIVE","EXEC","EXECUTE","EXISTS","EXIT","EXPLAIN","EXTERNAL","EXTRACT","FAIL","FALSE","FETCH","FILE","FILLFACTOR","FILTER","FIRST","FLOAT","FOLLOWING","FOR","FOREIGN","FORTRAN","FOUND","FREETEXT","FREETEXTTABLE","FROM","FULL","FUNCTION","GENERATED","GET","GLOB","GLOBAL","GO","GOTO","GRANT","GROUP","GROUPS","HAVING","HOLDLOCK","HOUR","IDENTITY","IDENTITYCOL","IDENTITY_INSERT","IF","IGNORE","IMMEDIATE","IN","INCLUDE","INDEX","INDEXED","INDICATOR","INITIALLY","INNER","INPUT","INSENSITIVE","INSERT","INSTEAD","INT","INTEGER","INTERSECT","INTERVAL","INTO","IS","ISNULL","ISOLATION","JOIN","KEY","KILL","LANGUAGE","LAST","LEADING","LEFT","LEVEL","LIKE","LIMIT","LINENO","LOAD","LOCAL","LOWER","MATCH","MATERIALIZED","MAX","MERGE","MIN","MINUTE","MODULE","MONTH","NAMES","NATIONAL","NATURAL","NCHAR","NEXT","NO","NOCHECK","NONCLUSTERED","NONE","NOT","NOTHING","NOTNULL","NULL","NULLIF","NULLS","NUMERIC","OCTET_LENGTH","OF","OFF","OFFSET","OFFSETS","ON","ONLY","OPEN","OPENDATASOURCE","OPENQUERY","OPENROWSET","OPENXML","OPTION","OR","ORDER","OTHERS","OUTER","OUTPUT","OVER","OVERLAPS","PAD","PARTIAL","PARTITION","PASCAL","PERCENT","PIVOT","PLAN","POSITION","PRAGMA","PRECEDING","PRECISION","PREPARE","PRESERVE","PRIMARY","PRINT","PRIOR","PRIVILEGES","PROC","PROCEDURE","PUBLIC","QUERY","RAISE","RAISERROR","RANGE","READ","READTEXT","REAL","RECONFIGURE","RECURSIVE","REFERENCES","REGEXP","REINDEX","RELATIVE","RELEASE","RENAME","REPLACE","REPLICATION","RESTORE","RESTRICT","RETURN","RETURNING","REVERT","REVOKE","RIGHT","ROLLBACK","ROW","ROWCOUNT","ROWGUIDCOL","ROWS","RULE","SAVE","SAVEPOINT","SCHEMA","SCROLL","SECOND","SECTION","SECURITYAUDIT","SELECT","SEMANTICKEYPHRASETABLE","SEMANTICSIMILARITYDETAILSTABLE","SEMANTICSIMILARITYTABLE","SESSION","SESSION_USER","SET","SETUSER","SHUTDOWN","SIZE","SMALLINT","SOME","SPACE","SQL","SQLCA","SQLCODE","SQLERROR","SQLSTATE","SQLWARNING","STATISTICS","SUBSTRING","SUM","SYSTEM_USER","TABLE","TABLESAMPLE","TEMP","TEMPORARY","TEXTSIZE","THEN","TIES","TIME","TIMESTAMP","TIMEZONE_HOUR","TIMEZONE_MINUTE","TO","TOP","TRAILING","TRAN","TRANSACTION","TRANSLATE","TRANSLATION","TRIGGER","TRIM","TRUE","TRUNCATE","TRY_CONVERT","TSEQUAL","UNBOUNDED","UNION","UNIQUE","UNKNOWN","UNPIVOT","UPDATE","UPDATETEXT","UPPER","USAGE","USE","USER","USING","VACUUM","VALUE","VALUES","VARCHAR","VARYING","VIEW","VIRTUAL","WAITFOR","WHEN","WHENEVER","WHERE","WHILE","WINDOW","WITH","WITHIN GROUP","WITHOUT","WORK","WRITE","WRITETEXT","YEAR","ZONE"],operators:["ALL","AND","ANY","BETWEEN","EXISTS","IN","LIKE","NOT","OR","SOME","EXCEPT","INTERSECT","UNION","APPLY","CROSS","FULL","INNER","JOIN","LEFT","OUTER","RIGHT","CONTAINS","FREETEXT","IS","NULL","PIVOT","UNPIVOT","MATCHED"],builtinFunctions:["AVG","CHECKSUM_AGG","COUNT","COUNT_BIG","GROUPING","GROUPING_ID","MAX","MIN","SUM","STDEV","STDEVP","VAR","VARP","CUME_DIST","FIRST_VALUE","LAG","LAST_VALUE","LEAD","PERCENTILE_CONT","PERCENTILE_DISC","PERCENT_RANK","COLLATE","COLLATIONPROPERTY","TERTIARY_WEIGHTS","FEDERATION_FILTERING_VALUE","CAST","CONVERT","PARSE","TRY_CAST","TRY_CONVERT","TRY_PARSE","ASYMKEY_ID","ASYMKEYPROPERTY","CERTPROPERTY","CERT_ID","CRYPT_GEN_RANDOM","DECRYPTBYASYMKEY","DECRYPTBYCERT","DECRYPTBYKEY","DECRYPTBYKEYAUTOASYMKEY","DECRYPTBYKEYAUTOCERT","DECRYPTBYPASSPHRASE","ENCRYPTBYASYMKEY","ENCRYPTBYCERT","ENCRYPTBYKEY","ENCRYPTBYPASSPHRASE","HASHBYTES","IS_OBJECTSIGNED","KEY_GUID","KEY_ID","KEY_NAME","SIGNBYASYMKEY","SIGNBYCERT","SYMKEYPROPERTY","VERIFYSIGNEDBYCERT","VERIFYSIGNEDBYASYMKEY","CURSOR_STATUS","DATALENGTH","IDENT_CURRENT","IDENT_INCR","IDENT_SEED","IDENTITY","SQL_VARIANT_PROPERTY","CURRENT_TIMESTAMP","DATEADD","DATEDIFF","DATEFROMPARTS","DATENAME","DATEPART","DATETIME2FROMPARTS","DATETIMEFROMPARTS","DATETIMEOFFSETFROMPARTS","DAY","EOMONTH","GETDATE","GETUTCDATE","ISDATE","MONTH","SMALLDATETIMEFROMPARTS","SWITCHOFFSET","SYSDATETIME","SYSDATETIMEOFFSET","SYSUTCDATETIME","TIMEFROMPARTS","TODATETIMEOFFSET","YEAR","CHOOSE","COALESCE","IIF","NULLIF","ABS","ACOS","ASIN","ATAN","ATN2","CEILING","COS","COT","DEGREES","EXP","FLOOR","LOG","LOG10","PI","POWER","RADIANS","RAND","ROUND","SIGN","SIN","SQRT","SQUARE","TAN","APP_NAME","APPLOCK_MODE","APPLOCK_TEST","ASSEMBLYPROPERTY","COL_LENGTH","COL_NAME","COLUMNPROPERTY","DATABASE_PRINCIPAL_ID","DATABASEPROPERTYEX","DB_ID","DB_NAME","FILE_ID","FILE_IDEX","FILE_NAME","FILEGROUP_ID","FILEGROUP_NAME","FILEGROUPPROPERTY","FILEPROPERTY","FULLTEXTCATALOGPROPERTY","FULLTEXTSERVICEPROPERTY","INDEX_COL","INDEXKEY_PROPERTY","INDEXPROPERTY","OBJECT_DEFINITION","OBJECT_ID","OBJECT_NAME","OBJECT_SCHEMA_NAME","OBJECTPROPERTY","OBJECTPROPERTYEX","ORIGINAL_DB_NAME","PARSENAME","SCHEMA_ID","SCHEMA_NAME","SCOPE_IDENTITY","SERVERPROPERTY","STATS_DATE","TYPE_ID","TYPE_NAME","TYPEPROPERTY","DENSE_RANK","NTILE","RANK","ROW_NUMBER","PUBLISHINGSERVERNAME","OPENDATASOURCE","OPENQUERY","OPENROWSET","OPENXML","CERTENCODED","CERTPRIVATEKEY","CURRENT_USER","HAS_DBACCESS","HAS_PERMS_BY_NAME","IS_MEMBER","IS_ROLEMEMBER","IS_SRVROLEMEMBER","LOGINPROPERTY","ORIGINAL_LOGIN","PERMISSIONS","PWDENCRYPT","PWDCOMPARE","SESSION_USER","SESSIONPROPERTY","SUSER_ID","SUSER_NAME","SUSER_SID","SUSER_SNAME","SYSTEM_USER","USER","USER_ID","USER_NAME","ASCII","CHAR","CHARINDEX","CONCAT","DIFFERENCE","FORMAT","LEFT","LEN","LOWER","LTRIM","NCHAR","PATINDEX","QUOTENAME","REPLACE","REPLICATE","REVERSE","RIGHT","RTRIM","SOUNDEX","SPACE","STR","STUFF","SUBSTRING","UNICODE","UPPER","BINARY_CHECKSUM","CHECKSUM","CONNECTIONPROPERTY","CONTEXT_INFO","CURRENT_REQUEST_ID","ERROR_LINE","ERROR_NUMBER","ERROR_MESSAGE","ERROR_PROCEDURE","ERROR_SEVERITY","ERROR_STATE","FORMATMESSAGE","GETANSINULL","GET_FILESTREAM_TRANSACTION_CONTEXT","HOST_ID","HOST_NAME","ISNULL","ISNUMERIC","MIN_ACTIVE_ROWVERSION","NEWID","NEWSEQUENTIALID","ROWCOUNT_BIG","XACT_STATE","TEXTPTR","TEXTVALID","COLUMNS_UPDATED","EVENTDATA","TRIGGER_NESTLEVEL","UPDATE","CHANGETABLE","CHANGE_TRACKING_CONTEXT","CHANGE_TRACKING_CURRENT_VERSION","CHANGE_TRACKING_IS_COLUMN_IN_MASK","CHANGE_TRACKING_MIN_VALID_VERSION","CONTAINSTABLE","FREETEXTTABLE","SEMANTICKEYPHRASETABLE","SEMANTICSIMILARITYDETAILSTABLE","SEMANTICSIMILARITYTABLE","FILETABLEROOTPATH","GETFILENAMESPACEPATH","GETPATHLOCATOR","PATHNAME","GET_TRANSMISSION_STATUS"],builtinVariables:["@@DATEFIRST","@@DBTS","@@LANGID","@@LANGUAGE","@@LOCK_TIMEOUT","@@MAX_CONNECTIONS","@@MAX_PRECISION","@@NESTLEVEL","@@OPTIONS","@@REMSERVER","@@SERVERNAME","@@SERVICENAME","@@SPID","@@TEXTSIZE","@@VERSION","@@CURSOR_ROWS","@@FETCH_STATUS","@@DATEFIRST","@@PROCID","@@ERROR","@@IDENTITY","@@ROWCOUNT","@@TRANCOUNT","@@CONNECTIONS","@@CPU_BUSY","@@IDLE","@@IO_BUSY","@@PACKET_ERRORS","@@PACK_RECEIVED","@@PACK_SENT","@@TIMETICKS","@@TOTAL_ERRORS","@@TOTAL_READ","@@TOTAL_WRITE"],pseudoColumns:["$ACTION","$IDENTITY","$ROWGUID","$PARTITION"],tokenizer:{root:[{include:"@comments"},{include:"@whitespace"},{include:"@pseudoColumns"},{include:"@numbers"},{include:"@strings"},{include:"@complexIdentifiers"},{include:"@scopes"},[/[;,.]/,"delimiter"],[/[()]/,"@brackets"],[/[\w@#$]+/,{cases:{"@operators":"operator","@builtinVariables":"predefined","@builtinFunctions":"predefined","@keywords":"keyword","@default":"identifier"}}],[/[<>=!%&+\-*/|~^]/,"operator"]],whitespace:[[/\s+/,"white"]],comments:[[/--+.*/,"comment"],[/\/\*/,{token:"comment.quote",next:"@comment"}]],comment:[[/[^*/]+/,"comment"],[/\*\//,{token:"comment.quote",next:"@pop"}],[/./,"comment"]],pseudoColumns:[[/[$][A-Za-z_][\w@#$]*/,{cases:{"@pseudoColumns":"predefined","@default":"identifier"}}]],numbers:[[/0[xX][0-9a-fA-F]*/,"number"],[/[$][+-]*\d*(\.\d*)?/,"number"],[/((\d+(\.\d*)?)|(\.\d+))([eE][\-+]?\d+)?/,"number"]],strings:[[/N'/,{token:"string",next:"@string"}],[/'/,{token:"string",next:"@string"}]],string:[[/[^']+/,"string"],[/''/,"string"],[/'/,{token:"string",next:"@pop"}]],complexIdentifiers:[[/\[/,{token:"identifier.quote",next:"@bracketedIdentifier"}],[/"/,{token:"identifier.quote",next:"@quotedIdentifier"}]],bracketedIdentifier:[[/[^\]]+/,"identifier"],[/]]/,"identifier"],[/]/,{token:"identifier.quote",next:"@pop"}]],quotedIdentifier:[[/[^"]+/,"identifier"],[/""/,"identifier"],[/"/,{token:"identifier.quote",next:"@pop"}]],scopes:[[/BEGIN\s+(DISTRIBUTED\s+)?TRAN(SACTION)?\b/i,"keyword"],[/BEGIN\s+TRY\b/i,{token:"keyword.try"}],[/END\s+TRY\b/i,{token:"keyword.try"}],[/BEGIN\s+CATCH\b/i,{token:"keyword.catch"}],[/END\s+CATCH\b/i,{token:"keyword.catch"}],[/(BEGIN|CASE)\b/i,{token:"keyword.block"}],[/END\b/i,{token:"keyword.block"}],[/WHEN\b/i,{token:"keyword.choice"}],[/THEN\b/i,{token:"keyword.choice"}]]}},nGe=Object.freeze(Object.defineProperty({__proto__:null,conf:tGe,language:iGe},Symbol.toStringTag,{value:"Module"})),sGe={comments:{lineComment:"//",blockComment:["(*","*)"]},brackets:[["{","}"],["[","]"],["(",")"],["var","end_var"],["var_input","end_var"],["var_output","end_var"],["var_in_out","end_var"],["var_temp","end_var"],["var_global","end_var"],["var_access","end_var"],["var_external","end_var"],["type","end_type"],["struct","end_struct"],["program","end_program"],["function","end_function"],["function_block","end_function_block"],["action","end_action"],["step","end_step"],["initial_step","end_step"],["transaction","end_transaction"],["configuration","end_configuration"],["tcp","end_tcp"],["recource","end_recource"],["channel","end_channel"],["library","end_library"],["folder","end_folder"],["binaries","end_binaries"],["includes","end_includes"],["sources","end_sources"]],autoClosingPairs:[{open:"[",close:"]"},{open:"{",close:"}"},{open:"(",close:")"},{open:"/*",close:"*/"},{open:"'",close:"'",notIn:["string_sq"]},{open:'"',close:'"',notIn:["string_dq"]},{open:"var_input",close:"end_var"},{open:"var_output",close:"end_var"},{open:"var_in_out",close:"end_var"},{open:"var_temp",close:"end_var"},{open:"var_global",close:"end_var"},{open:"var_access",close:"end_var"},{open:"var_external",close:"end_var"},{open:"type",close:"end_type"},{open:"struct",close:"end_struct"},{open:"program",close:"end_program"},{open:"function",close:"end_function"},{open:"function_block",close:"end_function_block"},{open:"action",close:"end_action"},{open:"step",close:"end_step"},{open:"initial_step",close:"end_step"},{open:"transaction",close:"end_transaction"},{open:"configuration",close:"end_configuration"},{open:"tcp",close:"end_tcp"},{open:"recource",close:"end_recource"},{open:"channel",close:"end_channel"},{open:"library",close:"end_library"},{open:"folder",close:"end_folder"},{open:"binaries",close:"end_binaries"},{open:"includes",close:"end_includes"},{open:"sources",close:"end_sources"}],surroundingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:'"',close:'"'},{open:"'",close:"'"},{open:"var",close:"end_var"},{open:"var_input",close:"end_var"},{open:"var_output",close:"end_var"},{open:"var_in_out",close:"end_var"},{open:"var_temp",close:"end_var"},{open:"var_global",close:"end_var"},{open:"var_access",close:"end_var"},{open:"var_external",close:"end_var"},{open:"type",close:"end_type"},{open:"struct",close:"end_struct"},{open:"program",close:"end_program"},{open:"function",close:"end_function"},{open:"function_block",close:"end_function_block"},{open:"action",close:"end_action"},{open:"step",close:"end_step"},{open:"initial_step",close:"end_step"},{open:"transaction",close:"end_transaction"},{open:"configuration",close:"end_configuration"},{open:"tcp",close:"end_tcp"},{open:"recource",close:"end_recource"},{open:"channel",close:"end_channel"},{open:"library",close:"end_library"},{open:"folder",close:"end_folder"},{open:"binaries",close:"end_binaries"},{open:"includes",close:"end_includes"},{open:"sources",close:"end_sources"}],folding:{markers:{start:new RegExp("^\\s*#pragma\\s+region\\b"),end:new RegExp("^\\s*#pragma\\s+endregion\\b")}}},oGe={defaultToken:"",tokenPostfix:".st",ignoreCase:!0,brackets:[{token:"delimiter.curly",open:"{",close:"}"},{token:"delimiter.parenthesis",open:"(",close:")"},{token:"delimiter.square",open:"[",close:"]"}],keywords:["if","end_if","elsif","else","case","of","to","__try","__catch","__finally","do","with","by","while","repeat","end_while","end_repeat","end_case","for","end_for","task","retain","non_retain","constant","with","at","exit","return","interval","priority","address","port","on_channel","then","iec","file","uses","version","packagetype","displayname","copyright","summary","vendor","common_source","from","extends","implements"],constant:["false","true","null"],defineKeywords:["var","var_input","var_output","var_in_out","var_temp","var_global","var_access","var_external","end_var","type","end_type","struct","end_struct","program","end_program","function","end_function","function_block","end_function_block","interface","end_interface","method","end_method","property","end_property","namespace","end_namespace","configuration","end_configuration","tcp","end_tcp","resource","end_resource","channel","end_channel","library","end_library","folder","end_folder","binaries","end_binaries","includes","end_includes","sources","end_sources","action","end_action","step","initial_step","end_step","transaction","end_transaction"],typeKeywords:["int","sint","dint","lint","usint","uint","udint","ulint","real","lreal","time","date","time_of_day","date_and_time","string","bool","byte","word","dword","array","pointer","lword"],operators:["=",">","<",":",":=","<=",">=","<>","&","+","-","*","**","MOD","^","or","and","not","xor","abs","acos","asin","atan","cos","exp","expt","ln","log","sin","sqrt","tan","sel","max","min","limit","mux","shl","shr","rol","ror","indexof","sizeof","adr","adrinst","bitadr","is_valid","ref","ref_to"],builtinVariables:[],builtinFunctions:["sr","rs","tp","ton","tof","eq","ge","le","lt","ne","round","trunc","ctd","сtu","ctud","r_trig","f_trig","move","concat","delete","find","insert","left","len","replace","right","rtc"],symbols:/[=><!~?:&|+\-*\/\^%]+/,escapes:/\\(?:[abfnrtv\\"']|x[0-9A-Fa-f]{1,4}|u[0-9A-Fa-f]{4}|U[0-9A-Fa-f]{8})/,tokenizer:{root:[[/(\.\.)/,"delimiter"],[/\b(16#[0-9A-Fa-f\_]*)+\b/,"number.hex"],[/\b(2#[01\_]+)+\b/,"number.binary"],[/\b(8#[0-9\_]*)+\b/,"number.octal"],[/\b\d*\.\d+([eE][\-+]?\d+)?\b/,"number.float"],[/\b(L?REAL)#[0-9\_\.e]+\b/,"number.float"],[/\b(BYTE|(?:D|L)?WORD|U?(?:S|D|L)?INT)#[0-9\_]+\b/,"number"],[/\d+/,"number"],[/\b(T|DT|TOD)#[0-9:-_shmyd]+\b/,"tag"],[/\%(I|Q|M)(X|B|W|D|L)[0-9\.]+/,"tag"],[/\%(I|Q|M)[0-9\.]*/,"tag"],[/\b[A-Za-z]{1,6}#[0-9]+\b/,"tag"],[/\b(TO_|CTU_|CTD_|CTUD_|MUX_|SEL_)[A_Za-z]+\b/,"predefined"],[/\b[A_Za-z]+(_TO_)[A_Za-z]+\b/,"predefined"],[/[;]/,"delimiter"],[/[.]/,{token:"delimiter",next:"@params"}],[/[a-zA-Z_]\w*/,{cases:{"@operators":"operators","@keywords":"keyword","@typeKeywords":"type","@defineKeywords":"variable","@constant":"constant","@builtinVariables":"predefined","@builtinFunctions":"predefined","@default":"identifier"}}],{include:"@whitespace"},[/[{}()\[\]]/,"@brackets"],[/"([^"\\]|\\.)*$/,"string.invalid"],[/"/,{token:"string.quote",bracket:"@open",next:"@string_dq"}],[/'/,{token:"string.quote",bracket:"@open",next:"@string_sq"}],[/'[^\\']'/,"string"],[/(')(@escapes)(')/,["string","string.escape","string"]],[/'/,"string.invalid"]],params:[[/\b[A-Za-z0-9_]+\b(?=\()/,{token:"identifier",next:"@pop"}],[/\b[A-Za-z0-9_]+\b/,"variable.name","@pop"]],comment:[[/[^\/*]+/,"comment"],[/\/\*/,"comment","@push"],["\\*/","comment","@pop"],[/[\/*]/,"comment"]],comment2:[[/[^\(*]+/,"comment"],[/\(\*/,"comment","@push"],["\\*\\)","comment","@pop"],[/[\(*]/,"comment"]],whitespace:[[/[ \t\r\n]+/,"white"],[/\/\/.*$/,"comment"],[/\/\*/,"comment","@comment"],[/\(\*/,"comment","@comment2"]],string_dq:[[/[^\\"]+/,"string"],[/@escapes/,"string.escape"],[/\\./,"string.escape.invalid"],[/"/,{token:"string.quote",bracket:"@close",next:"@pop"}]],string_sq:[[/[^\\']+/,"string"],[/@escapes/,"string.escape"],[/\\./,"string.escape.invalid"],[/'/,{token:"string.quote",bracket:"@close",next:"@pop"}]]}},rGe=Object.freeze(Object.defineProperty({__proto__:null,conf:sGe,language:oGe},Symbol.toStringTag,{value:"Module"}));const aGe={comments:{lineComment:"//",blockComment:["/*","*/"]},brackets:[["{","}"],["[","]"],["(",")"]],autoClosingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:'"',close:'"'},{open:"'",close:"'"},{open:"`",close:"`"}],surroundingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:'"',close:'"'},{open:"'",close:"'"},{open:"`",close:"`"}]},lGe={defaultToken:"",tokenPostfix:".swift",identifier:/[a-zA-Z_][\w$]*/,attributes:["@GKInspectable","@IBAction","@IBDesignable","@IBInspectable","@IBOutlet","@IBSegueAction","@NSApplicationMain","@NSCopying","@NSManaged","@Sendable","@UIApplicationMain","@autoclosure","@actorIndependent","@asyncHandler","@available","@convention","@derivative","@differentiable","@discardableResult","@dynamicCallable","@dynamicMemberLookup","@escaping","@frozen","@globalActor","@inlinable","@inline","@main","@noDerivative","@nonobjc","@noreturn","@objc","@objcMembers","@preconcurrency","@propertyWrapper","@requires_stored_property_inits","@resultBuilder","@testable","@unchecked","@unknown","@usableFromInline","@warn_unqualified_access"],accessmodifiers:["open","public","internal","fileprivate","private"],keywords:["#available","#colorLiteral","#column","#dsohandle","#else","#elseif","#endif","#error","#file","#fileID","#fileLiteral","#filePath","#function","#if","#imageLiteral","#keyPath","#line","#selector","#sourceLocation","#warning","Any","Protocol","Self","Type","actor","as","assignment","associatedtype","associativity","async","await","break","case","catch","class","continue","convenience","default","defer","deinit","didSet","do","dynamic","dynamicType","else","enum","extension","fallthrough","false","fileprivate","final","for","func","get","guard","higherThan","if","import","in","indirect","infix","init","inout","internal","is","isolated","lazy","left","let","lowerThan","mutating","nil","none","nonisolated","nonmutating","open","operator","optional","override","postfix","precedence","precedencegroup","prefix","private","protocol","public","repeat","required","rethrows","return","right","safe","self","set","some","static","struct","subscript","super","switch","throw","throws","true","try","typealias","unowned","unsafe","var","weak","where","while","willSet","__consuming","__owned"],symbols:/[=(){}\[\].,:;@#\_&\-<>`?!+*\\\/]/,operatorstart:/[\/=\-+!*%<>&|^~?\u00A1-\u00A7\u00A9\u00AB\u00AC\u00AE\u00B0-\u00B1\u00B6\u00BB\u00BF\u00D7\u00F7\u2016-\u2017\u2020-\u2027\u2030-\u203E\u2041-\u2053\u2055-\u205E\u2190-\u23FF\u2500-\u2775\u2794-\u2BFF\u2E00-\u2E7F\u3001-\u3003\u3008-\u3030]/,operatorend:/[\u0300-\u036F\u1DC0-\u1DFF\u20D0-\u20FF\uFE00-\uFE0F\uFE20-\uFE2F\uE0100-\uE01EF]/,operators:/(@operatorstart)((@operatorstart)|(@operatorend))*/,escapes:/\\(?:[abfnrtv\\"']|x[0-9A-Fa-f]{1,4}|u[0-9A-Fa-f]{4}|U[0-9A-Fa-f]{8})/,tokenizer:{root:[{include:"@whitespace"},{include:"@comment"},{include:"@attribute"},{include:"@literal"},{include:"@keyword"},{include:"@invokedmethod"},{include:"@symbol"}],whitespace:[[/\s+/,"white"],[/"""/,"string.quote","@endDblDocString"]],endDblDocString:[[/[^"]+/,"string"],[/\\"/,"string"],[/"""/,"string.quote","@popall"],[/"/,"string"]],symbol:[[/[{}()\[\]]/,"@brackets"],[/[<>](?!@symbols)/,"@brackets"],[/[.]/,"delimiter"],[/@operators/,"operator"],[/@symbols/,"operator"]],comment:[[/\/\/\/.*$/,"comment.doc"],[/\/\*\*/,"comment.doc","@commentdocbody"],[/\/\/.*$/,"comment"],[/\/\*/,"comment","@commentbody"]],commentdocbody:[[/\/\*/,"comment","@commentbody"],[/\*\//,"comment.doc","@pop"],[/\:[a-zA-Z]+\:/,"comment.doc.param"],[/./,"comment.doc"]],commentbody:[[/\/\*/,"comment","@commentbody"],[/\*\//,"comment","@pop"],[/./,"comment"]],attribute:[[/@@@identifier/,{cases:{"@attributes":"keyword.control","@default":""}}]],literal:[[/"/,{token:"string.quote",next:"@stringlit"}],[/0[b]([01]_?)+/,"number.binary"],[/0[o]([0-7]_?)+/,"number.octal"],[/0[x]([0-9a-fA-F]_?)+([pP][\-+](\d_?)+)?/,"number.hex"],[/(\d_?)*\.(\d_?)+([eE][\-+]?(\d_?)+)?/,"number.float"],[/(\d_?)+/,"number"]],stringlit:[[/\\\(/,{token:"operator",next:"@interpolatedexpression"}],[/@escapes/,"string"],[/\\./,"string.escape.invalid"],[/"/,{token:"string.quote",next:"@pop"}],[/./,"string"]],interpolatedexpression:[[/\(/,{token:"operator",next:"@interpolatedexpression"}],[/\)/,{token:"operator",next:"@pop"}],{include:"@literal"},{include:"@keyword"},{include:"@symbol"}],keyword:[[/`/,{token:"operator",next:"@escapedkeyword"}],[/@identifier/,{cases:{"@keywords":"keyword","[A-Z][a-zA-Z0-9$]*":"type.identifier","@default":"identifier"}}]],escapedkeyword:[[/`/,{token:"operator",next:"@pop"}],[/./,"identifier"]],invokedmethod:[[/([.])(@identifier)/,{cases:{$2:["delimeter","type.identifier"],"@default":""}}]]}},cGe=Object.freeze(Object.defineProperty({__proto__:null,conf:aGe,language:lGe},Symbol.toStringTag,{value:"Module"})),dGe={comments:{lineComment:"//",blockComment:["/*","*/"]},brackets:[["{","}"],["[","]"],["(",")"],["begin","end"],["case","endcase"],["casex","endcase"],["casez","endcase"],["checker","endchecker"],["class","endclass"],["clocking","endclocking"],["config","endconfig"],["function","endfunction"],["generate","endgenerate"],["group","endgroup"],["interface","endinterface"],["module","endmodule"],["package","endpackage"],["primitive","endprimitive"],["program","endprogram"],["property","endproperty"],["specify","endspecify"],["sequence","endsequence"],["table","endtable"],["task","endtask"]],autoClosingPairs:[{open:"[",close:"]"},{open:"{",close:"}"},{open:"(",close:")"},{open:"'",close:"'",notIn:["string","comment"]},{open:'"',close:'"',notIn:["string"]}],surroundingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:'"',close:'"'},{open:"'",close:"'"}],folding:{offSide:!1,markers:{start:new RegExp("^(?:\\s*|.*(?!\\/[\\/\\*])[^\\w])(?:begin|case(x|z)?|class|clocking|config|covergroup|function|generate|interface|module|package|primitive|property|program|sequence|specify|table|task)\\b"),end:new RegExp("^(?:\\s*|.*(?!\\/[\\/\\*])[^\\w])(?:end|endcase|endclass|endclocking|endconfig|endgroup|endfunction|endgenerate|endinterface|endmodule|endpackage|endprimitive|endproperty|endprogram|endsequence|endspecify|endtable|endtask)\\b")}}},hGe={defaultToken:"",tokenPostfix:".sv",brackets:[{token:"delimiter.curly",open:"{",close:"}"},{token:"delimiter.parenthesis",open:"(",close:")"},{token:"delimiter.square",open:"[",close:"]"},{token:"delimiter.angle",open:"<",close:">"}],keywords:["accept_on","alias","always","always_comb","always_ff","always_latch","and","assert","assign","assume","automatic","before","begin","bind","bins","binsof","bit","break","buf","bufif0","bufif1","byte","case","casex","casez","cell","chandle","checker","class","clocking","cmos","config","const","constraint","context","continue","cover","covergroup","coverpoint","cross","deassign","default","defparam","design","disable","dist","do","edge","else","end","endcase","endchecker","endclass","endclocking","endconfig","endfunction","endgenerate","endgroup","endinterface","endmodule","endpackage","endprimitive","endprogram","endproperty","endspecify","endsequence","endtable","endtask","enum","event","eventually","expect","export","extends","extern","final","first_match","for","force","foreach","forever","fork","forkjoin","function","generate","genvar","global","highz0","highz1","if","iff","ifnone","ignore_bins","illegal_bins","implements","implies","import","incdir","include","initial","inout","input","inside","instance","int","integer","interconnect","interface","intersect","join","join_any","join_none","large","let","liblist","library","local","localparam","logic","longint","macromodule","matches","medium","modport","module","nand","negedge","nettype","new","nexttime","nmos","nor","noshowcancelled","not","notif0","notif1","null","or","output","package","packed","parameter","pmos","posedge","primitive","priority","program","property","protected","pull0","pull1","pulldown","pullup","pulsestyle_ondetect","pulsestyle_onevent","pure","rand","randc","randcase","randsequence","rcmos","real","realtime","ref","reg","reject_on","release","repeat","restrict","return","rnmos","rpmos","rtran","rtranif0","rtranif1","s_always","s_eventually","s_nexttime","s_until","s_until_with","scalared","sequence","shortint","shortreal","showcancelled","signed","small","soft","solve","specify","specparam","static","string","strong","strong0","strong1","struct","super","supply0","supply1","sync_accept_on","sync_reject_on","table","tagged","task","this","throughout","time","timeprecision","timeunit","tran","tranif0","tranif1","tri","tri0","tri1","triand","trior","trireg","type","typedef","union","unique","unique0","unsigned","until","until_with","untyped","use","uwire","var","vectored","virtual","void","wait","wait_order","wand","weak","weak0","weak1","while","wildcard","wire","with","within","wor","xnor","xor"],builtin_gates:["and","nand","nor","or","xor","xnor","buf","not","bufif0","bufif1","notif1","notif0","cmos","nmos","pmos","rcmos","rnmos","rpmos","tran","tranif1","tranif0","rtran","rtranif1","rtranif0"],operators:["=","+=","-=","*=","/=","%=","&=","|=","^=","<<=",">>+","<<<=",">>>=","?",":","+","-","!","~","&","~&","|","~|","^","~^","^~","+","-","*","/","%","==","!=","===","!==","==?","!=?","&&","||","**","<","<=",">",">=","&","|","^",">>","<<",">>>","<<<","++","--","->","<->","inside","dist","::","+:","-:","*>","&&&","|->","|=>","#=#"],symbols:/[=><!~?:&|+\-*\/\^%#]+/,escapes:/%%|\\(?:[antvf\\"']|x[0-9A-Fa-f]{1,2}|[0-7]{1,3})/,identifier:/(?:[a-zA-Z_][a-zA-Z0-9_$\.]*|\\\S+ )/,systemcall:/[$][a-zA-Z0-9_]+/,timeunits:/s|ms|us|ns|ps|fs/,tokenizer:{root:[[/^(\s*)(@identifier)/,["",{cases:{"@builtin_gates":{token:"keyword.$2",next:"@module_instance"},table:{token:"keyword.$2",next:"@table"},"@keywords":{token:"keyword.$2"},"@default":{token:"identifier",next:"@module_instance"}}}]],[/^\s*`include/,{token:"keyword.directive.include",next:"@include"}],[/^\s*`\s*\w+/,"keyword"],{include:"@identifier_or_keyword"},{include:"@whitespace"},[/\(\*.*\*\)/,"annotation"],[/@systemcall/,"variable.predefined"],[/[{}()\[\]]/,"@brackets"],[/[<>](?!@symbols)/,"@brackets"],[/@symbols/,{cases:{"@operators":"delimiter","@default":""}}],{include:"@numbers"},[/[;,.]/,"delimiter"],{include:"@strings"}],identifier_or_keyword:[[/@identifier/,{cases:{"@keywords":{token:"keyword.$0"},"@default":"identifier"}}]],numbers:[[/\d+?[\d_]*(?:\.[\d_]+)?[eE][\-+]?\d+/,"number.float"],[/\d+?[\d_]*\.[\d_]+(?:\s*@timeunits)?/,"number.float"],[/(?:\d+?[\d_]*\s*)?'[sS]?[dD]\s*[0-9xXzZ?]+?[0-9xXzZ?_]*/,"number"],[/(?:\d+?[\d_]*\s*)?'[sS]?[bB]\s*[0-1xXzZ?]+?[0-1xXzZ?_]*/,"number.binary"],[/(?:\d+?[\d_]*\s*)?'[sS]?[oO]\s*[0-7xXzZ?]+?[0-7xXzZ?_]*/,"number.octal"],[/(?:\d+?[\d_]*\s*)?'[sS]?[hH]\s*[0-9a-fA-FxXzZ?]+?[0-9a-fA-FxXzZ?_]*/,"number.hex"],[/1step/,"number"],[/[\dxXzZ]+?[\dxXzZ_]*(?:\s*@timeunits)?/,"number"],[/'[01xXzZ]+/,"number"]],module_instance:[{include:"@whitespace"},[/(#?)(\()/,["",{token:"@brackets",next:"@port_connection"}]],[/@identifier\s*[;={}\[\],]/,{token:"@rematch",next:"@pop"}],[/@symbols|[;={}\[\],]/,{token:"@rematch",next:"@pop"}],[/@identifier/,"type"],[/;/,"delimiter","@pop"]],port_connection:[{include:"@identifier_or_keyword"},{include:"@whitespace"},[/@systemcall/,"variable.predefined"],{include:"@numbers"},{include:"@strings"},[/[,]/,"delimiter"],[/\(/,"@brackets","@port_connection"],[/\)/,"@brackets","@pop"]],whitespace:[[/[ \t\r\n]+/,""],[/\/\*/,"comment","@comment"],[/\/\/.*$/,"comment"]],comment:[[/[^\/*]+/,"comment"],[/\*\//,"comment","@pop"],[/[\/*]/,"comment"]],strings:[[/"([^"\\]|\\.)*$/,"string.invalid"],[/"/,"string","@string"]],string:[[/[^\\"]+/,"string"],[/@escapes/,"string.escape"],[/\\./,"string.escape.invalid"],[/"/,"string","@pop"]],include:[[/(\s*)(")([\w*\/*]*)(.\w*)(")/,["","string.include.identifier","string.include.identifier","string.include.identifier",{token:"string.include.identifier",next:"@pop"}]],[/(\s*)(<)([\w*\/*]*)(.\w*)(>)/,["","string.include.identifier","string.include.identifier","string.include.identifier",{token:"string.include.identifier",next:"@pop"}]]],table:[{include:"@whitespace"},[/[()]/,"@brackets"],[/[:;]/,"delimiter"],[/[01\-*?xXbBrRfFpPnN]/,"variable.predefined"],["endtable","keyword.endtable","@pop"]]}},wle=Object.freeze(Object.defineProperty({__proto__:null,conf:dGe,language:hGe},Symbol.toStringTag,{value:"Module"})),uGe={brackets:[["{","}"],["[","]"],["(",")"]],autoClosingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:'"',close:'"'},{open:"'",close:"'"}],surroundingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:'"',close:'"'},{open:"'",close:"'"}]},gGe={tokenPostfix:".tcl",specialFunctions:["set","unset","rename","variable","proc","coroutine","foreach","incr","append","lappend","linsert","lreplace"],mainFunctions:["if","then","elseif","else","case","switch","while","for","break","continue","return","package","namespace","catch","exit","eval","expr","uplevel","upvar"],builtinFunctions:["file","info","concat","join","lindex","list","llength","lrange","lsearch","lsort","split","array","parray","binary","format","regexp","regsub","scan","string","subst","dict","cd","clock","exec","glob","pid","pwd","close","eof","fblocked","fconfigure","fcopy","fileevent","flush","gets","open","puts","read","seek","socket","tell","interp","after","auto_execok","auto_load","auto_mkindex","auto_reset","bgerror","error","global","history","load","source","time","trace","unknown","unset","update","vwait","winfo","wm","bind","event","pack","place","grid","font","bell","clipboard","destroy","focus","grab","lower","option","raise","selection","send","tk","tkwait","tk_bisque","tk_focusNext","tk_focusPrev","tk_focusFollowsMouse","tk_popup","tk_setPalette"],symbols:/[=><!~?:&|+\-*\/\^%]+/,brackets:[{open:"(",close:")",token:"delimiter.parenthesis"},{open:"{",close:"}",token:"delimiter.curly"},{open:"[",close:"]",token:"delimiter.square"}],escapes:/\\(?:[abfnrtv\\"'\[\]\{\};\$]|x[0-9A-Fa-f]{1,4}|u[0-9A-Fa-f]{4}|U[0-9A-Fa-f]{8})/,variables:/(?:\$+(?:(?:\:\:?)?[a-zA-Z_]\w*)+)/,tokenizer:{root:[[/[a-zA-Z_]\w*/,{cases:{"@specialFunctions":{token:"keyword.flow",next:"@specialFunc"},"@mainFunctions":"keyword","@builtinFunctions":"variable","@default":"operator.scss"}}],[/\s+\-+(?!\d|\.)\w*|{\*}/,"metatag"],{include:"@whitespace"},[/[{}()\[\]]/,"@brackets"],[/@symbols/,"operator"],[/\$+(?:\:\:)?\{/,{token:"identifier",next:"@nestedVariable"}],[/@variables/,"type.identifier"],[/\.(?!\d|\.)[\w\-]*/,"operator.sql"],[/\d+(\.\d+)?/,"number"],[/\d+/,"number"],[/;/,"delimiter"],[/"/,{token:"string.quote",bracket:"@open",next:"@dstring"}],[/'/,{token:"string.quote",bracket:"@open",next:"@sstring"}]],dstring:[[/\[/,{token:"@brackets",next:"@nestedCall"}],[/\$+(?:\:\:)?\{/,{token:"identifier",next:"@nestedVariable"}],[/@variables/,"type.identifier"],[/[^\\$\[\]"]+/,"string"],[/@escapes/,"string.escape"],[/"/,{token:"string.quote",bracket:"@close",next:"@pop"}]],sstring:[[/\[/,{token:"@brackets",next:"@nestedCall"}],[/\$+(?:\:\:)?\{/,{token:"identifier",next:"@nestedVariable"}],[/@variables/,"type.identifier"],[/[^\\$\[\]']+/,"string"],[/@escapes/,"string.escape"],[/'/,{token:"string.quote",bracket:"@close",next:"@pop"}]],whitespace:[[/[ \t\r\n]+/,"white"],[/#.*\\$/,{token:"comment",next:"@newlineComment"}],[/#.*(?!\\)$/,"comment"]],newlineComment:[[/.*\\$/,"comment"],[/.*(?!\\)$/,{token:"comment",next:"@pop"}]],nestedVariable:[[/[^\{\}\$]+/,"type.identifier"],[/\}/,{token:"identifier",next:"@pop"}]],nestedCall:[[/\[/,{token:"@brackets",next:"@nestedCall"}],[/\]/,{token:"@brackets",next:"@pop"}],{include:"root"}],specialFunc:[[/"/,{token:"string",next:"@dstring"}],[/'/,{token:"string",next:"@sstring"}],[/\S+/,{token:"type",next:"@pop"}]]}},fGe=Object.freeze(Object.defineProperty({__proto__:null,conf:uGe,language:gGe},Symbol.toStringTag,{value:"Module"})),pGe={wordPattern:/(-?\d*\.\d\w*)|([^\`\~\!\@\$\^\&\*\(\)\=\+\[\{\]\}\\\|\;\:\'\"\,\.\<\>\/\s]+)/g,comments:{blockComment:["{#","#}"]},brackets:[["{#","#}"],["{%","%}"],["{{","}}"],["(",")"],["[","]"],["<!--","-->"],["<",">"]],autoClosingPairs:[{open:"{# ",close:" #}"},{open:"{% ",close:" %}"},{open:"{{ ",close:" }}"},{open:"[",close:"]"},{open:"(",close:")"},{open:'"',close:'"'},{open:"'",close:"'"}],surroundingPairs:[{open:'"',close:'"'},{open:"'",close:"'"},{open:"<",close:">"}]},mGe={defaultToken:"",tokenPostfix:"",ignoreCase:!0,keywords:["apply","autoescape","block","deprecated","do","embed","extends","flush","for","from","if","import","include","macro","sandbox","set","use","verbatim","with","endapply","endautoescape","endblock","endembed","endfor","endif","endmacro","endsandbox","endset","endwith","true","false"],tokenizer:{root:[[/\s+/],[/{#/,"comment.twig","@commentState"],[/{%[-~]?/,"delimiter.twig","@blockState"],[/{{[-~]?/,"delimiter.twig","@variableState"],[/<!DOCTYPE/,"metatag.html","@doctype"],[/<!--/,"comment.html","@comment"],[/(<)((?:[\w\-]+:)?[\w\-]+)(\s*)(\/>)/,["delimiter.html","tag.html","","delimiter.html"]],[/(<)(script)/,["delimiter.html",{token:"tag.html",next:"@script"}]],[/(<)(style)/,["delimiter.html",{token:"tag.html",next:"@style"}]],[/(<)((?:[\w\-]+:)?[\w\-]+)/,["delimiter.html",{token:"tag.html",next:"@otherTag"}]],[/(<\/)((?:[\w\-]+:)?[\w\-]+)/,["delimiter.html",{token:"tag.html",next:"@otherTag"}]],[/</,"delimiter.html"],[/[^<{]+/]],commentState:[[/#}/,"comment.twig","@pop"],[/./,"comment.twig"]],blockState:[[/[-~]?%}/,"delimiter.twig","@pop"],[/\s+/],[/(verbatim)(\s*)([-~]?%})/,["keyword.twig","",{token:"delimiter.twig",next:"@rawDataState"}]],{include:"expression"}],rawDataState:[[/({%[-~]?)(\s*)(endverbatim)(\s*)([-~]?%})/,["delimiter.twig","","keyword.twig","",{token:"delimiter.twig",next:"@popall"}]],[/./,"string.twig"]],variableState:[[/[-~]?}}/,"delimiter.twig","@pop"],{include:"expression"}],stringState:[[/"/,"string.twig","@pop"],[/#{\s*/,"string.twig","@interpolationState"],[/[^#"\\]*(?:(?:\\.|#(?!\{))[^#"\\]*)*/,"string.twig"]],interpolationState:[[/}/,"string.twig","@pop"],{include:"expression"}],expression:[[/\s+/],[/\+|-|\/{1,2}|%|\*{1,2}/,"operators.twig"],[/(and|or|not|b-and|b-xor|b-or)(\s+)/,["operators.twig",""]],[/==|!=|<|>|>=|<=/,"operators.twig"],[/(starts with|ends with|matches)(\s+)/,["operators.twig",""]],[/(in)(\s+)/,["operators.twig",""]],[/(is)(\s+)/,["operators.twig",""]],[/\||~|:|\.{1,2}|\?{1,2}/,"operators.twig"],[/[^\W\d][\w]*/,{cases:{"@keywords":"keyword.twig","@default":"variable.twig"}}],[/\d+(\.\d+)?/,"number.twig"],[/\(|\)|\[|\]|{|}|,/,"delimiter.twig"],[/"([^#"\\]*(?:\\.[^#"\\]*)*)"|\'([^\'\\]*(?:\\.[^\'\\]*)*)\'/,"string.twig"],[/"/,"string.twig","@stringState"],[/=>/,"operators.twig"],[/=/,"operators.twig"]],doctype:[[/[^>]+/,"metatag.content.html"],[/>/,"metatag.html","@pop"]],comment:[[/-->/,"comment.html","@pop"],[/[^-]+/,"comment.content.html"],[/./,"comment.content.html"]],otherTag:[[/\/?>/,"delimiter.html","@pop"],[/"([^"]*)"/,"attribute.value.html"],[/'([^']*)'/,"attribute.value.html"],[/[\w\-]+/,"attribute.name.html"],[/=/,"delimiter.html"],[/[ \t\r\n]+/]],script:[[/type/,"attribute.name.html","@scriptAfterType"],[/"([^"]*)"/,"attribute.value.html"],[/'([^']*)'/,"attribute.value.html"],[/[\w\-]+/,"attribute.name.html"],[/=/,"delimiter.html"],[/>/,{token:"delimiter.html",next:"@scriptEmbedded",nextEmbedded:"text/javascript"}],[/[ \t\r\n]+/],[/(<\/)(script\s*)(>)/,["delimiter.html","tag.html",{token:"delimiter.html",next:"@pop"}]]],scriptAfterType:[[/=/,"delimiter.html","@scriptAfterTypeEquals"],[/>/,{token:"delimiter.html",next:"@scriptEmbedded",nextEmbedded:"text/javascript"}],[/[ \t\r\n]+/],[/<\/script\s*>/,{token:"@rematch",next:"@pop"}]],scriptAfterTypeEquals:[[/"([^"]*)"/,{token:"attribute.value.html",switchTo:"@scriptWithCustomType.$1"}],[/'([^']*)'/,{token:"attribute.value.html",switchTo:"@scriptWithCustomType.$1"}],[/>/,{token:"delimiter.html",next:"@scriptEmbedded",nextEmbedded:"text/javascript"}],[/[ \t\r\n]+/],[/<\/script\s*>/,{token:"@rematch",next:"@pop"}]],scriptWithCustomType:[[/>/,{token:"delimiter.html",next:"@scriptEmbedded.$S2",nextEmbedded:"$S2"}],[/"([^"]*)"/,"attribute.value.html"],[/'([^']*)'/,"attribute.value.html"],[/[\w\-]+/,"attribute.name.html"],[/=/,"delimiter.html"],[/[ \t\r\n]+/],[/<\/script\s*>/,{token:"@rematch",next:"@pop"}]],scriptEmbedded:[[/<\/script/,{token:"@rematch",next:"@pop",nextEmbedded:"@pop"}],[/[^<]+/,""]],style:[[/type/,"attribute.name.html","@styleAfterType"],[/"([^"]*)"/,"attribute.value.html"],[/'([^']*)'/,"attribute.value.html"],[/[\w\-]+/,"attribute.name.html"],[/=/,"delimiter.html"],[/>/,{token:"delimiter.html",next:"@styleEmbedded",nextEmbedded:"text/css"}],[/[ \t\r\n]+/],[/(<\/)(style\s*)(>)/,["delimiter.html","tag.html",{token:"delimiter.html",next:"@pop"}]]],styleAfterType:[[/=/,"delimiter.html","@styleAfterTypeEquals"],[/>/,{token:"delimiter.html",next:"@styleEmbedded",nextEmbedded:"text/css"}],[/[ \t\r\n]+/],[/<\/style\s*>/,{token:"@rematch",next:"@pop"}]],styleAfterTypeEquals:[[/"([^"]*)"/,{token:"attribute.value.html",switchTo:"@styleWithCustomType.$1"}],[/'([^']*)'/,{token:"attribute.value.html",switchTo:"@styleWithCustomType.$1"}],[/>/,{token:"delimiter.html",next:"@styleEmbedded",nextEmbedded:"text/css"}],[/[ \t\r\n]+/],[/<\/style\s*>/,{token:"@rematch",next:"@pop"}]],styleWithCustomType:[[/>/,{token:"delimiter.html",next:"@styleEmbedded.$S2",nextEmbedded:"$S2"}],[/"([^"]*)"/,"attribute.value.html"],[/'([^']*)'/,"attribute.value.html"],[/[\w\-]+/,"attribute.name.html"],[/=/,"delimiter.html"],[/[ \t\r\n]+/],[/<\/style\s*>/,{token:"@rematch",next:"@pop"}]],styleEmbedded:[[/<\/style/,{token:"@rematch",next:"@pop",nextEmbedded:"@pop"}],[/[^<]+/,""]]}},_Ge=Object.freeze(Object.defineProperty({__proto__:null,conf:pGe,language:mGe},Symbol.toStringTag,{value:"Module"})),vle=s=>`\\b${s}\\b`,SX=s=>`(?!${s})`,bGe="[_a-zA-Z]",wGe="[_a-zA-Z0-9]",xX=vle(`${bGe}${wGe}*`),vGe=vle("[_a-zA-Z-0-9]+"),CGe=["import","model","scalar","namespace","op","interface","union","using","is","extends","enum","alias","return","void","if","else","projection","dec","extern","fn"],yGe=["true","false","null","unknown","never"],SGe="[ \\t\\r\\n]",xGe="[0-9]+",LGe={comments:{lineComment:"//",blockComment:["/*","*/"]},brackets:[["{","}"],["[","]"],["(",")"]],autoClosingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:'"',close:'"'},{open:"/**",close:" */",notIn:["string"]}],surroundingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:'"',close:'"'}],indentationRules:{decreaseIndentPattern:new RegExp("^((?!.*?/\\*).*\\*/)?\\s*[\\}\\]].*$"),increaseIndentPattern:new RegExp("^((?!//).)*(\\{([^}\"'`/]*|(\\t|[ ])*//.*)|\\([^)\"'`/]*|\\[[^\\]\"'`/]*)$"),unIndentedLinePattern:new RegExp("^(\\t|[ ])*[ ]\\*[^/]*\\*/\\s*$|^(\\t|[ ])*[ ]\\*/\\s*$|^(\\t|[ ])*[ ]\\*([ ]([^\\*]|\\*(?!/))*)?$")}},kGe={defaultToken:"",tokenPostfix:".tsp",brackets:[{open:"{",close:"}",token:"delimiter.curly"},{open:"[",close:"]",token:"delimiter.square"},{open:"(",close:")",token:"delimiter.parenthesis"}],symbols:/[=:;<>]+/,keywords:CGe,namedLiterals:yGe,escapes:'\\\\(u{[0-9A-Fa-f]+}|n|r|t|\\\\|"|\\${)',tokenizer:{root:[{include:"@expression"},{include:"@whitespace"}],stringVerbatim:[{regex:'(|"|"")[^"]',action:{token:"string"}},{regex:`"""${SX('"')}`,action:{token:"string",next:"@pop"}}],stringLiteral:[{regex:"\\${",action:{token:"delimiter.bracket",next:"@bracketCounting"}},{regex:'[^\\\\"$]+',action:{token:"string"}},{regex:"@escapes",action:{token:"string.escape"}},{regex:"\\\\.",action:{token:"string.escape.invalid"}},{regex:'"',action:{token:"string",next:"@pop"}}],bracketCounting:[{regex:"{",action:{token:"delimiter.bracket",next:"@bracketCounting"}},{regex:"}",action:{token:"delimiter.bracket",next:"@pop"}},{include:"@expression"}],comment:[{regex:"[^\\*]+",action:{token:"comment"}},{regex:"\\*\\/",action:{token:"comment",next:"@pop"}},{regex:"[\\/*]",action:{token:"comment"}}],whitespace:[{regex:SGe},{regex:"\\/\\*",action:{token:"comment",next:"@comment"}},{regex:"\\/\\/.*$",action:{token:"comment"}}],expression:[{regex:'"""',action:{token:"string",next:"@stringVerbatim"}},{regex:`"${SX('""')}`,action:{token:"string",next:"@stringLiteral"}},{regex:xGe,action:{token:"number"}},{regex:xX,action:{cases:{"@keywords":{token:"keyword"},"@namedLiterals":{token:"keyword"},"@default":{token:"identifier"}}}},{regex:`@${xX}`,action:{token:"tag"}},{regex:`#${vGe}`,action:{token:"directive"}}]}},EGe=Object.freeze(Object.defineProperty({__proto__:null,conf:LGe,language:kGe},Symbol.toStringTag,{value:"Module"})),IGe={comments:{lineComment:"'",blockComment:["/*","*/"]},brackets:[["{","}"],["[","]"],["(",")"],["<",">"],["addhandler","end addhandler"],["class","end class"],["enum","end enum"],["event","end event"],["function","end function"],["get","end get"],["if","end if"],["interface","end interface"],["module","end module"],["namespace","end namespace"],["operator","end operator"],["property","end property"],["raiseevent","end raiseevent"],["removehandler","end removehandler"],["select","end select"],["set","end set"],["structure","end structure"],["sub","end sub"],["synclock","end synclock"],["try","end try"],["while","end while"],["with","end with"],["using","end using"],["do","loop"],["for","next"]],autoClosingPairs:[{open:"{",close:"}",notIn:["string","comment"]},{open:"[",close:"]",notIn:["string","comment"]},{open:"(",close:")",notIn:["string","comment"]},{open:'"',close:'"',notIn:["string","comment"]},{open:"<",close:">",notIn:["string","comment"]}],folding:{markers:{start:new RegExp("^\\s*#Region\\b"),end:new RegExp("^\\s*#End Region\\b")}}},DGe={defaultToken:"",tokenPostfix:".vb",ignoreCase:!0,brackets:[{token:"delimiter.bracket",open:"{",close:"}"},{token:"delimiter.array",open:"[",close:"]"},{token:"delimiter.parenthesis",open:"(",close:")"},{token:"delimiter.angle",open:"<",close:">"},{token:"keyword.tag-addhandler",open:"addhandler",close:"end addhandler"},{token:"keyword.tag-class",open:"class",close:"end class"},{token:"keyword.tag-enum",open:"enum",close:"end enum"},{token:"keyword.tag-event",open:"event",close:"end event"},{token:"keyword.tag-function",open:"function",close:"end function"},{token:"keyword.tag-get",open:"get",close:"end get"},{token:"keyword.tag-if",open:"if",close:"end if"},{token:"keyword.tag-interface",open:"interface",close:"end interface"},{token:"keyword.tag-module",open:"module",close:"end module"},{token:"keyword.tag-namespace",open:"namespace",close:"end namespace"},{token:"keyword.tag-operator",open:"operator",close:"end operator"},{token:"keyword.tag-property",open:"property",close:"end property"},{token:"keyword.tag-raiseevent",open:"raiseevent",close:"end raiseevent"},{token:"keyword.tag-removehandler",open:"removehandler",close:"end removehandler"},{token:"keyword.tag-select",open:"select",close:"end select"},{token:"keyword.tag-set",open:"set",close:"end set"},{token:"keyword.tag-structure",open:"structure",close:"end structure"},{token:"keyword.tag-sub",open:"sub",close:"end sub"},{token:"keyword.tag-synclock",open:"synclock",close:"end synclock"},{token:"keyword.tag-try",open:"try",close:"end try"},{token:"keyword.tag-while",open:"while",close:"end while"},{token:"keyword.tag-with",open:"with",close:"end with"},{token:"keyword.tag-using",open:"using",close:"end using"},{token:"keyword.tag-do",open:"do",close:"loop"},{token:"keyword.tag-for",open:"for",close:"next"}],keywords:["AddHandler","AddressOf","Alias","And","AndAlso","As","Async","Boolean","ByRef","Byte","ByVal","Call","Case","Catch","CBool","CByte","CChar","CDate","CDbl","CDec","Char","CInt","Class","CLng","CObj","Const","Continue","CSByte","CShort","CSng","CStr","CType","CUInt","CULng","CUShort","Date","Decimal","Declare","Default","Delegate","Dim","DirectCast","Do","Double","Each","Else","ElseIf","End","EndIf","Enum","Erase","Error","Event","Exit","False","Finally","For","Friend","Function","Get","GetType","GetXMLNamespace","Global","GoSub","GoTo","Handles","If","Implements","Imports","In","Inherits","Integer","Interface","Is","IsNot","Let","Lib","Like","Long","Loop","Me","Mod","Module","MustInherit","MustOverride","MyBase","MyClass","NameOf","Namespace","Narrowing","New","Next","Not","Nothing","NotInheritable","NotOverridable","Object","Of","On","Operator","Option","Optional","Or","OrElse","Out","Overloads","Overridable","Overrides","ParamArray","Partial","Private","Property","Protected","Public","RaiseEvent","ReadOnly","ReDim","RemoveHandler","Resume","Return","SByte","Select","Set","Shadows","Shared","Short","Single","Static","Step","Stop","String","Structure","Sub","SyncLock","Then","Throw","To","True","Try","TryCast","TypeOf","UInteger","ULong","UShort","Using","Variant","Wend","When","While","Widening","With","WithEvents","WriteOnly","Xor"],tagwords:["If","Sub","Select","Try","Class","Enum","Function","Get","Interface","Module","Namespace","Operator","Set","Structure","Using","While","With","Do","Loop","For","Next","Property","Continue","AddHandler","RemoveHandler","Event","RaiseEvent","SyncLock"],symbols:/[=><!~?;\.,:&|+\-*\/\^%]+/,integersuffix:/U?[DI%L&S@]?/,floatsuffix:/[R#F!]?/,tokenizer:{root:[{include:"@whitespace"},[/next(?!\w)/,{token:"keyword.tag-for"}],[/loop(?!\w)/,{token:"keyword.tag-do"}],[/end\s+(?!for|do)(addhandler|class|enum|event|function|get|if|interface|module|namespace|operator|property|raiseevent|removehandler|select|set|structure|sub|synclock|try|while|with|using)/,{token:"keyword.tag-$1"}],[/[a-zA-Z_]\w*/,{cases:{"@tagwords":{token:"keyword.tag-$0"},"@keywords":{token:"keyword.$0"},"@default":"identifier"}}],[/^\s*#\w+/,"keyword"],[/\d*\d+e([\-+]?\d+)?(@floatsuffix)/,"number.float"],[/\d*\.\d+(e[\-+]?\d+)?(@floatsuffix)/,"number.float"],[/&H[0-9a-f]+(@integersuffix)/,"number.hex"],[/&0[0-7]+(@integersuffix)/,"number.octal"],[/\d+(@integersuffix)/,"number"],[/#.*#/,"number"],[/[{}()\[\]]/,"@brackets"],[/@symbols/,"delimiter"],[/["\u201c\u201d]/,{token:"string.quote",next:"@string"}]],whitespace:[[/[ \t\r\n]+/,""],[/(\'|REM(?!\w)).*$/,"comment"]],string:[[/[^"\u201c\u201d]+/,"string"],[/["\u201c\u201d]{2}/,"string.escape"],[/["\u201c\u201d]C?/,{token:"string.quote",next:"@pop"}]]}},TGe=Object.freeze(Object.defineProperty({__proto__:null,conf:IGe,language:DGe},Symbol.toStringTag,{value:"Module"})),NGe={comments:{lineComment:"//",blockComment:["/*","*/"]},brackets:[["{","}"],["[","]"],["(",")"]],autoClosingPairs:[{open:"[",close:"]"},{open:"{",close:"}"},{open:"(",close:")"}],surroundingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"}]};function Bh(s){let e=[];const t=s.split(/\t+|\r+|\n+| +/);for(let i=0;i<t.length;++i)t[i].length>0&&e.push(t[i]);return e}const RGe=Bh("true false"),AGe=Bh(`
			  alias
			  break
			  case
			  const
			  const_assert
			  continue
			  continuing
			  default
			  diagnostic
			  discard
			  else
			  enable
			  fn
			  for
			  if
			  let
			  loop
			  override
			  requires
			  return
			  struct
			  switch
			  var
			  while
			  `),MGe=Bh(`
			  NULL
			  Self
			  abstract
			  active
			  alignas
			  alignof
			  as
			  asm
			  asm_fragment
			  async
			  attribute
			  auto
			  await
			  become
			  binding_array
			  cast
			  catch
			  class
			  co_await
			  co_return
			  co_yield
			  coherent
			  column_major
			  common
			  compile
			  compile_fragment
			  concept
			  const_cast
			  consteval
			  constexpr
			  constinit
			  crate
			  debugger
			  decltype
			  delete
			  demote
			  demote_to_helper
			  do
			  dynamic_cast
			  enum
			  explicit
			  export
			  extends
			  extern
			  external
			  fallthrough
			  filter
			  final
			  finally
			  friend
			  from
			  fxgroup
			  get
			  goto
			  groupshared
			  highp
			  impl
			  implements
			  import
			  inline
			  instanceof
			  interface
			  layout
			  lowp
			  macro
			  macro_rules
			  match
			  mediump
			  meta
			  mod
			  module
			  move
			  mut
			  mutable
			  namespace
			  new
			  nil
			  noexcept
			  noinline
			  nointerpolation
			  noperspective
			  null
			  nullptr
			  of
			  operator
			  package
			  packoffset
			  partition
			  pass
			  patch
			  pixelfragment
			  precise
			  precision
			  premerge
			  priv
			  protected
			  pub
			  public
			  readonly
			  ref
			  regardless
			  register
			  reinterpret_cast
			  require
			  resource
			  restrict
			  self
			  set
			  shared
			  sizeof
			  smooth
			  snorm
			  static
			  static_assert
			  static_cast
			  std
			  subroutine
			  super
			  target
			  template
			  this
			  thread_local
			  throw
			  trait
			  try
			  type
			  typedef
			  typeid
			  typename
			  typeof
			  union
			  unless
			  unorm
			  unsafe
			  unsized
			  use
			  using
			  varying
			  virtual
			  volatile
			  wgsl
			  where
			  with
			  writeonly
			  yield
			  `),OGe=Bh(`
		read write read_write
		function private workgroup uniform storage
		perspective linear flat
		center centroid sample
		vertex_index instance_index position front_facing frag_depth
			local_invocation_id local_invocation_index
			global_invocation_id workgroup_id num_workgroups
			sample_index sample_mask
		rgba8unorm
		rgba8snorm
		rgba8uint
		rgba8sint
		rgba16uint
		rgba16sint
		rgba16float
		r32uint
		r32sint
		r32float
		rg32uint
		rg32sint
		rg32float
		rgba32uint
		rgba32sint
		rgba32float
		bgra8unorm
`),PGe=Bh(`
		bool
		f16
		f32
		i32
		sampler sampler_comparison
		texture_depth_2d
		texture_depth_2d_array
		texture_depth_cube
		texture_depth_cube_array
		texture_depth_multisampled_2d
		texture_external
		texture_external
		u32
		`),FGe=Bh(`
		array
		atomic
		mat2x2
		mat2x3
		mat2x4
		mat3x2
		mat3x3
		mat3x4
		mat4x2
		mat4x3
		mat4x4
		ptr
		texture_1d
		texture_2d
		texture_2d_array
		texture_3d
		texture_cube
		texture_cube_array
		texture_multisampled_2d
		texture_storage_1d
		texture_storage_2d
		texture_storage_2d_array
		texture_storage_3d
		vec2
		vec3
		vec4
		`),BGe=Bh(`
		vec2i vec3i vec4i
		vec2u vec3u vec4u
		vec2f vec3f vec4f
		vec2h vec3h vec4h
		mat2x2f mat2x3f mat2x4f
		mat3x2f mat3x3f mat3x4f
		mat4x2f mat4x3f mat4x4f
		mat2x2h mat2x3h mat2x4h
		mat3x2h mat3x3h mat3x4h
		mat4x2h mat4x3h mat4x4h
		`),WGe=Bh(`
  bitcast all any select arrayLength abs acos acosh asin asinh atan atanh atan2
  ceil clamp cos cosh countLeadingZeros countOneBits countTrailingZeros cross
  degrees determinant distance dot exp exp2 extractBits faceForward firstLeadingBit
  firstTrailingBit floor fma fract frexp inverseBits inverseSqrt ldexp length
  log log2 max min mix modf normalize pow quantizeToF16 radians reflect refract
  reverseBits round saturate sign sin sinh smoothstep sqrt step tan tanh transpose
  trunc dpdx dpdxCoarse dpdxFine dpdy dpdyCoarse dpdyFine fwidth fwidthCoarse fwidthFine
  textureDimensions textureGather textureGatherCompare textureLoad textureNumLayers
  textureNumLevels textureNumSamples textureSample textureSampleBias textureSampleCompare
  textureSampleCompareLevel textureSampleGrad textureSampleLevel textureSampleBaseClampToEdge
  textureStore atomicLoad atomicStore atomicAdd atomicSub atomicMax atomicMin
  atomicAnd atomicOr atomicXor atomicExchange atomicCompareExchangeWeak pack4x8snorm
  pack4x8unorm pack2x16snorm pack2x16unorm pack2x16float unpack4x8snorm unpack4x8unorm
  unpack2x16snorm unpack2x16unorm unpack2x16float storageBarrier workgroupBarrier
  workgroupUniformLoad
`),HGe=Bh(`
					 &
					 &&
					 ->
					 /
					 =
					 ==
					 !=
					 >
					 >=
					 <
					 <=
					 %
					 -
					 --
					 +
					 ++
					 |
					 ||
					 *
					 <<
					 >>
					 +=
					 -=
					 *=
					 /=
					 %=
					 &=
					 |=
					 ^=
					 >>=
					 <<=
					 `),VGe=/enable|requires|diagnostic/,LX=new RegExp("[_\\p{XID_Start}]\\p{XID_Continue}*","u"),db="variable.predefined",zGe={tokenPostfix:".wgsl",defaultToken:"invalid",unicode:!0,atoms:RGe,keywords:AGe,reserved:MGe,predeclared_enums:OGe,predeclared_types:PGe,predeclared_type_generators:FGe,predeclared_type_aliases:BGe,predeclared_intrinsics:WGe,operators:HGe,symbols:/[!%&*+\-\.\/:;<=>^|_~,]+/,tokenizer:{root:[[VGe,"keyword","@directive"],[LX,{cases:{"@atoms":db,"@keywords":"keyword","@reserved":"invalid","@predeclared_enums":db,"@predeclared_types":db,"@predeclared_type_generators":db,"@predeclared_type_aliases":db,"@predeclared_intrinsics":db,"@default":"identifier"}}],{include:"@commentOrSpace"},{include:"@numbers"},[/[{}()\[\]]/,"@brackets"],["@","annotation","@attribute"],[/@symbols/,{cases:{"@operators":"operator","@default":"delimiter"}}],[/./,"invalid"]],commentOrSpace:[[/\s+/,"white"],[/\/\*/,"comment","@blockComment"],[/\/\/.*$/,"comment"]],blockComment:[[/[^\/*]+/,"comment"],[/\/\*/,"comment","@push"],[/\*\//,"comment","@pop"],[/[\/*]/,"comment"]],attribute:[{include:"@commentOrSpace"},[/\w+/,"annotation","@pop"]],directive:[{include:"@commentOrSpace"},[/[()]/,"@brackets"],[/,/,"delimiter"],[LX,"meta.content"],[/;/,"delimiter","@pop"]],numbers:[[/0[fh]/,"number.float"],[/[1-9][0-9]*[fh]/,"number.float"],[/[0-9]*\.[0-9]+([eE][+-]?[0-9]+)?[fh]?/,"number.float"],[/[0-9]+\.[0-9]*([eE][+-]?[0-9]+)?[fh]?/,"number.float"],[/[0-9]+[eE][+-]?[0-9]+[fh]?/,"number.float"],[/0[xX][0-9a-fA-F]*\.[0-9a-fA-F]+(?:[pP][+-]?[0-9]+[fh]?)?/,"number.hex"],[/0[xX][0-9a-fA-F]+\.[0-9a-fA-F]*(?:[pP][+-]?[0-9]+[fh]?)?/,"number.hex"],[/0[xX][0-9a-fA-F]+[pP][+-]?[0-9]+[fh]?/,"number.hex"],[/0[xX][0-9a-fA-F]+[iu]?/,"number.hex"],[/[1-9][0-9]*[iu]?/,"number"],[/0[iu]?/,"number"]]}},$Ge=Object.freeze(Object.defineProperty({__proto__:null,conf:NGe,language:zGe},Symbol.toStringTag,{value:"Module"})),UGe={comments:{blockComment:["<!--","-->"]},brackets:[["<",">"]],autoClosingPairs:[{open:"<",close:">"},{open:"'",close:"'"},{open:'"',close:'"'}],surroundingPairs:[{open:"<",close:">"},{open:"'",close:"'"},{open:'"',close:'"'}],onEnterRules:[{beforeText:new RegExp("<([_:\\w][_:\\w-.\\d]*)([^/>]*(?!/)>)[^<]*$","i"),afterText:/^<\/([_:\w][_:\w-.\d]*)\s*>$/i,action:{indentAction:H.IndentAction.IndentOutdent}},{beforeText:new RegExp("<(\\w[\\w\\d]*)([^/>]*(?!/)>)[^<]*$","i"),action:{indentAction:H.IndentAction.Indent}}]},qGe={defaultToken:"",tokenPostfix:".xml",ignoreCase:!0,qualifiedName:/(?:[\w\.\-]+:)?[\w\.\-]+/,tokenizer:{root:[[/[^<&]+/,""],{include:"@whitespace"},[/(<)(@qualifiedName)/,[{token:"delimiter"},{token:"tag",next:"@tag"}]],[/(<\/)(@qualifiedName)(\s*)(>)/,[{token:"delimiter"},{token:"tag"},"",{token:"delimiter"}]],[/(<\?)(@qualifiedName)/,[{token:"delimiter"},{token:"metatag",next:"@tag"}]],[/(<\!)(@qualifiedName)/,[{token:"delimiter"},{token:"metatag",next:"@tag"}]],[/<\!\[CDATA\[/,{token:"delimiter.cdata",next:"@cdata"}],[/&\w+;/,"string.escape"]],cdata:[[/[^\]]+/,""],[/\]\]>/,{token:"delimiter.cdata",next:"@pop"}],[/\]/,""]],tag:[[/[ \t\r\n]+/,""],[/(@qualifiedName)(\s*=\s*)("[^"]*"|'[^']*')/,["attribute.name","","attribute.value"]],[/(@qualifiedName)(\s*=\s*)("[^">?\/]*|'[^'>?\/]*)(?=[\?\/]\>)/,["attribute.name","","attribute.value"]],[/(@qualifiedName)(\s*=\s*)("[^">]*|'[^'>]*)/,["attribute.name","","attribute.value"]],[/@qualifiedName/,"attribute.name"],[/\?>/,{token:"delimiter",next:"@pop"}],[/(\/)(>)/,[{token:"tag"},{token:"delimiter",next:"@pop"}]],[/>/,{token:"delimiter",next:"@pop"}]],whitespace:[[/[ \t\r\n]+/,""],[/<!--/,{token:"comment",next:"@comment"}]],comment:[[/[^<\-]+/,"comment.content"],[/-->/,{token:"comment",next:"@pop"}],[/<!--/,"comment.content.invalid"],[/[<\-]/,"comment.content"]]}},jGe=Object.freeze(Object.defineProperty({__proto__:null,conf:UGe,language:qGe},Symbol.toStringTag,{value:"Module"})),GGe={comments:{lineComment:"#"},brackets:[["{","}"],["[","]"],["(",")"]],autoClosingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:'"',close:'"'},{open:"'",close:"'"}],surroundingPairs:[{open:"{",close:"}"},{open:"[",close:"]"},{open:"(",close:")"},{open:'"',close:'"'},{open:"'",close:"'"}],folding:{offSide:!0},onEnterRules:[{beforeText:/:\s*$/,action:{indentAction:H.IndentAction.Indent}}]},KGe={tokenPostfix:".yaml",brackets:[{token:"delimiter.bracket",open:"{",close:"}"},{token:"delimiter.square",open:"[",close:"]"}],keywords:["true","True","TRUE","false","False","FALSE","null","Null","Null","~"],numberInteger:/(?:0|[+-]?[0-9]+)/,numberFloat:/(?:0|[+-]?[0-9]+)(?:\.[0-9]+)?(?:e[-+][1-9][0-9]*)?/,numberOctal:/0o[0-7]+/,numberHex:/0x[0-9a-fA-F]+/,numberInfinity:/[+-]?\.(?:inf|Inf|INF)/,numberNaN:/\.(?:nan|Nan|NAN)/,numberDate:/\d{4}-\d\d-\d\d([Tt ]\d\d:\d\d:\d\d(\.\d+)?(( ?[+-]\d\d?(:\d\d)?)|Z)?)?/,escapes:/\\(?:[btnfr\\"']|[0-7][0-7]?|[0-3][0-7]{2})/,tokenizer:{root:[{include:"@whitespace"},{include:"@comment"},[/%[^ ]+.*$/,"meta.directive"],[/---/,"operators.directivesEnd"],[/\.{3}/,"operators.documentEnd"],[/[-?:](?= )/,"operators"],{include:"@anchor"},{include:"@tagHandle"},{include:"@flowCollections"},{include:"@blockStyle"},[/@numberInteger(?![ \t]*\S+)/,"number"],[/@numberFloat(?![ \t]*\S+)/,"number.float"],[/@numberOctal(?![ \t]*\S+)/,"number.octal"],[/@numberHex(?![ \t]*\S+)/,"number.hex"],[/@numberInfinity(?![ \t]*\S+)/,"number.infinity"],[/@numberNaN(?![ \t]*\S+)/,"number.nan"],[/@numberDate(?![ \t]*\S+)/,"number.date"],[/(".*?"|'.*?'|[^#'"]*?)([ \t]*)(:)( |$)/,["type","white","operators","white"]],{include:"@flowScalars"},[/.+?(?=(\s+#|$))/,{cases:{"@keywords":"keyword","@default":"string"}}]],object:[{include:"@whitespace"},{include:"@comment"},[/\}/,"@brackets","@pop"],[/,/,"delimiter.comma"],[/:(?= )/,"operators"],[/(?:".*?"|'.*?'|[^,\{\[]+?)(?=: )/,"type"],{include:"@flowCollections"},{include:"@flowScalars"},{include:"@tagHandle"},{include:"@anchor"},{include:"@flowNumber"},[/[^\},]+/,{cases:{"@keywords":"keyword","@default":"string"}}]],array:[{include:"@whitespace"},{include:"@comment"},[/\]/,"@brackets","@pop"],[/,/,"delimiter.comma"],{include:"@flowCollections"},{include:"@flowScalars"},{include:"@tagHandle"},{include:"@anchor"},{include:"@flowNumber"},[/[^\],]+/,{cases:{"@keywords":"keyword","@default":"string"}}]],multiString:[[/^( +).+$/,"string","@multiStringContinued.$1"]],multiStringContinued:[[/^( *).+$/,{cases:{"$1==$S2":"string","@default":{token:"@rematch",next:"@popall"}}}]],whitespace:[[/[ \t\r\n]+/,"white"]],comment:[[/#.*$/,"comment"]],flowCollections:[[/\[/,"@brackets","@array"],[/\{/,"@brackets","@object"]],flowScalars:[[/"([^"\\]|\\.)*$/,"string.invalid"],[/'([^'\\]|\\.)*$/,"string.invalid"],[/'[^']*'/,"string"],[/"/,"string","@doubleQuotedString"]],doubleQuotedString:[[/[^\\"]+/,"string"],[/@escapes/,"string.escape"],[/\\./,"string.escape.invalid"],[/"/,"string","@pop"]],blockStyle:[[/[>|][0-9]*[+-]?$/,"operators","@multiString"]],flowNumber:[[/@numberInteger(?=[ \t]*[,\]\}])/,"number"],[/@numberFloat(?=[ \t]*[,\]\}])/,"number.float"],[/@numberOctal(?=[ \t]*[,\]\}])/,"number.octal"],[/@numberHex(?=[ \t]*[,\]\}])/,"number.hex"],[/@numberInfinity(?=[ \t]*[,\]\}])/,"number.infinity"],[/@numberNaN(?=[ \t]*[,\]\}])/,"number.nan"],[/@numberDate(?=[ \t]*[,\]\}])/,"number.date"]],tagHandle:[[/\![^ ]*/,"tag"]],anchor:[[/[&*][^ ]+/,"namespace"]]}},YGe=Object.freeze(Object.defineProperty({__proto__:null,conf:GGe,language:KGe},Symbol.toStringTag,{value:"Module"}));export{XGe as F,sQe as W,Se as _,nQe as a,oQe as g,TX as l,iQe as m};
